// @(#)$Id: AbstractFileSetTask.java,v 1.3 2004/05/26 19:02:49 leavens Exp $

// Copyright (C) 2002 Iowa State University

// This file is part of JML

// JML is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.

// JML is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with JML; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

package org.jmlspecs.ant.tasks;

import java.io.File;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.Task;

/**
 * <p>An abstract class for creating Tasks that work with <b>Ant</b>
 * filesets. It provides an inner class FileSetVisitor to conveniently
 * apply a certain action on all the files in all the FileSets of an
 * AbstractFileSetTask.</p>
 *
 * <p>The FileSetVisitor is typically used as follows in a subclass</p>
 *<pre><code>
 * new FileSetVisitor() {
 *   public void visit(File file) {
 *     //action code
 *   }
 * }.visit();
 * 
 *</code></pre>
 *
 * <p>This task works with Ant filesets in order to keep things simple!</p>
 *
 * @version $Revision: 1.3 $
 * @author Marko van Dooren
 */
public class AbstractFileSetTask extends Task {

    private HashSet _filesets = new HashSet();

    /*@
      @ private invariant ! _filesets.contains(null);
      @ private invariant (\forall Object o; _filesets.contains(o);
      @                                      o instanceof FileSet);
      @*/

    /**
     * Adds a set of files (nested fileset attribute).
     *
     * @param fileset
     *        The fileset to be added.
     */
    /*@ 
      @   pre fileset != null;
      @   post getFileSets().contains(fileset);
      @*/
    public void addFileset(FileSet fileset) {
        _filesets.add(fileset);
    }

    /**
     * Return the filesets of the AbstractFileSetTask
     */
    /*@
      @ public normal_behavior
      @   post ! \result.contains(null);
      @   post (\forall Object o; \result.contains(o); o instanceof FileSet);
      @*/
    public /*@ pure @*/ Set getFileSets() {
        return new HashSet(_filesets);
    }

    public abstract class FileSetVisitor {
        /**
         * Perform an action on the given file.
         *
         * @param file
         *        The file to perform the action on.
         */
        /*@
          @   pre file != null;
          @*/
        public abstract void visit(File file) throws Exception;

        /**
         * Perform the action defined in visit(File) on all
         * files in the filesets of this FileSetVisitor.
         */  
        public void visit() throws Exception {
            Iterator iter = _filesets.iterator();
            while(iter.hasNext()) {
                FileSet fileset = (FileSet)iter.next();
                FileSetWrapper wrapper = new FileSetWrapper(fileset);
                FileSetWrapper.Visitor visitor = wrapper.new Visitor() {
                        public void visit(File file) throws Exception {
                            (FileSetVisitor.this).visit(file);
                        }
                    };
                visitor.visit(getProject());
            }
        }    
    }

} 
