// @(#)$Id: CommonOptionsTask.java,v 1.3 2003/10/12 18:02:23 leavens Exp $

// Copyright (C) 2002 Iowa State University

// This file is part of JML

// JML is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.

// JML is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with JML; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

package org.jmlspecs.ant.tasks;

import java.util.ArrayList;
import java.util.List;
import java.io.File;
import org.apache.tools.ant.BuildException;

/** An abstract Ant task that holds common options for the JML tools,
 * including FileSets.
 *
 * @version $Revision: 1.3 $
 * @author  Marko van Dooren
 * @author Gary T. Leavens
 */
public class CommonOptionsTask extends AbstractFileSetTask {
  
    //@ post !getVerbose() && !getRecursive();
    //@ post getFileSets() != null && getFileSets().isEmpty();
    public CommonOptionsTask() {
        // We don't want those ugly messages all the time.
        _verbose = false;
        _recursive = false;
    }

    /**
     * Return an array of the argument strings for running a JML tool.
     */
    public /*@ non_null @*/ String[] getArguments() throws Exception {
            final ArrayList args = new ArrayList();

            if(! getVerbose()) {
                args.add("-Q");
            }

            if(getRecursive()) {
                args.add("-R");
            }

            if(getClasspath() != null) {
                args.add("-C"+getClasspath());
            }

            if(getSourcepath() != null) {
                args.add("-S"+getSourcepath());
            }

            if(getDestination() != null) {
                args.add("-d"+getDestination());
            }

            args.addAll(getToolSpecificOptions());
            
            args.addAll(getFilesAndDirectories());

            return (String[]) args.toArray(new String[0]);
    }

    /**
     * Return a list of the tool-specific options for running a JML tool.
     */
    public /*@ pure non_null @*/ List getToolSpecificOptions() {
        return new ArrayList();
    }

    /**
     * Return a list of the files and directories needed for running a tool.
     */
    public /*@ non_null @*/ List getFilesAndDirectories()
        throws Exception
    {
        final ArrayList args = new ArrayList();

        new FileSetVisitor() {
                public void visit(File file) {
                    args.add(file.getAbsolutePath());
                }
            }.visit();

        return args;
    }

    /**
     * Check whether or not this task works recursive.
     */
    public /*@ pure @*/ boolean getRecursive() {
        return _recursive;
    }

    /**
     * Set the verbosity
     *
     * @param recursive
     *        True if the task should work recursive, false otherwise.
     */
    /*@
      @   post getRecursive() == recursive;
      @*/
    public void setRecursive(boolean recursive) {
        _recursive = recursive;
    }

    private boolean _recursive;

    /**
     * Check whether or not this task should be verbose.
     */
    public /*@ pure @*/ boolean getVerbose() {
        return _verbose;
    }

    /**
     * Set the verbosity
     *
     * @param verbose
     *        True if the task should be verbose, false otherwise.
     */
    /*@ 
      @   post getVerbose() == verbose;
      @*/
    public void setVerbose(boolean verbose) {
        _verbose = verbose;
    }

    private boolean _verbose;

    /**
     * Set the CLASSPATH.
     *
     * @param classpath
     *        The CLASSPATH.
     */
    /*@ 
      @   post getClasspath() == classpath;
      @*/
    public void setClasspath(String classpath) {
        _classPath=classpath;
    }

    /**
     * Return the classpath of this CheckTask.
     */
    public /*@ pure @*/ String getClasspath() {
        return _classPath;
    }

    private String _classPath;

    /**
     * Set the destination of this CompileTask.
     *
     * @param destination
     *        A String containing the directory where the files
     *        should be put.
     */
    /*@ 
      @   post getDestination() == destination;
      @*/
    public void setDestination(String destination) {
        _destination = destination;
    }

    /**
     * Return the destination of this CompileTask.
     */
    public /*@ pure @*/ String getDestination() {
        return _destination;
    }

    private String _destination;

    /**
     * Set the sourcepath for the JML checker.
     *
     * @param sourcepath
     *        The sourcepath.
     */
    /*@
      @ post getSourcepath() == sourcepath;
      @*/
    public void setSourcepath(String sourcepath) {
        _sourcePath=sourcepath;
    }

    /**
     * Return the sourcepath of this CheckTask.
     */
    public /*@ pure @*/ String getSourcepath() {
        return _sourcePath;
    }

    private String _sourcePath;

}
