// @(#)$Id: FileSetWrapper.java,v 1.3 2003/09/02 19:27:20 cclifton Exp $

// Copyright (C) 2002 Iowa State University

// This file is part of JML

// JML is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.

// JML is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with JML; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

package org.jmlspecs.ant.tasks;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.FileScanner;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.types.FileSet;

import java.io.File;

/**
 * <p>The inner class Visitor is provided to conveniently perform a
 * certain action on all element in a FileSet. It is typically used as
 * follows in a task:</p>
 *
 *<pre><code>
 * fileset.new Visitor() {
 *   public void visit(File file) {
 *     //action code
 *   }
 * }.visit(project);
 *</code></pre>
 *
 * @version $Revision: 1.3 $
 * @author Marko van Dooren
 */
public class FileSetWrapper {

    /*@ public normal_behavior
      @   pre fileset != null;
      @   assignable \not_specified;
      @   post getFileSet() == fileset;
      @*/
    public FileSetWrapper(FileSet fileset) {
        _fileset = fileset;
    }

    /*@ public behavior
      @   post \result != null;
      @*/
    public /*@ pure non_null @*/ FileSet getFileSet() {
        return _fileset;
    }

    private /*@ non_null @*/FileSet _fileset;

    public abstract class Visitor {
    
        /**
         * Perform an action on the given file.
         *
         * @param file
         *        The file to perform the action on.
         */
        /*@ public behavior
          @   pre file != null;
          @   assignable \not_specified;
          @*/
        public abstract void visit(File file) throws Exception;
    
        /**
         * Perform the action defined in visit(File) on all
         * files in the fileset of this Visitor for the given project.
         *
         * @param project
         *        The project
         */
        /*@ public behavior
          @   pre project != null;
          @   assignable \not_specified;
          @*/
        public void visit(Project project) throws Exception {
            DirectoryScanner dir = _fileset.getDirectoryScanner(project);
            String[] files = dir.getIncludedFiles();
            for (int i=0; i <files.length; i++) {
                File file = new File(dir.getBasedir().getAbsolutePath()+File.separator+files[i]);
                visit(file);
            }
        }
    }

}
