// @(#)$Id: Sequence.java,v 1.3 2006/09/12 04:58:29 ye-cui Exp $

// Copyright (C) 2005 Iowa State University

// This file is part of JML

// JML is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.

// JML is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with JML; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

package org.jmlspecs.checker.testcase.typecheck;

import java.util.*;

/** Uses of the model program constructs.
 * @author Steve Shaner
 */
public abstract class Sequence {
    //@ public model instance JMLDataGroup myState;
    //@ assignable myState;
    abstract int nextValue(int currentValue);

    private /*@ spec_public @*/ int lastValue;

    /*@ public model_program {
      @   normal_behavior
      @     assignable lastValue;
      @     ensures lastValue == firstValue;
      @ }
      @ also
      @ assignable lastValue;
      @*/
    public Sequence(int firstValue) {
        //@ assume true;
        lastValue = firstValue;
        //@ assert lastValue == firstValue;
    }

    /*@ public model_program {
      @ normal_behavior
      @   assignable lastValue;
      @   ensures lastValue == 0;
      @ }
      @ also
      @ assignable lastValue;
      @*/
    public Sequence() {
        // JML doesn't allow anything in front of the constructor call,
        // so can't do the following:
        // //@ assume true;        
        this(0);
        //@ assert lastValue == 0;
    }

    /*@ public model_program
      @ {
      @   ghost int nextValue = nextValue(lastValue);
      @   normal_behavior
      @     assignable lastValue, myState;
      @     ensures lastValue == nextValue;
      @   return nextValue;
      @ }
      @ also
      @ assignable lastValue, myState;
      @*/
    public int next() {
        //@ ghost int nextValue = nextValue(lastValue);
        //@ assume true;
        lastValue = nextValue(lastValue);
        //@ assert lastValue == nextValue;
        
        return lastValue;
    }
    
    /*@ public model_program
      @ {
      @   ghost int[] twoValues = new int[2];
      @   ghost int firstValue = nextValue(lastValue);
      @   ghost int secondValue = nextValue(firstValue);
      @   normal_behavior
      @     assignable lastValue;
      @     ensures (twoValues[0]==firstValue && twoValues[1]==secondValue
      @              && lastValue==secondValue);
      @   return twoValues;
      @ }
      @ also
      @ assignable lastValue, myState;
      @*/
    public int[] nextTwo() {
        //@ ghost int[] twoValues = new int[2];
        //@ ghost int firstValue = nextValue(lastValue);
        //@ ghost int secondValue = nextValue(firstValue);
        int[] result = new int[2];
        
        //@ assume true;
        result[0] = next();
        result[1] = next();
        //@ assert (twoValues[0]==firstValue && twoValues[1]==secondValue);
        
        return result;
    }

    /*@ public model_program
      @ {
      @   ghost int firstValue;
      @   ghost int newValue;
      @
      @   do {
      @     set firstValue = lastValue;
      @     set newValue = nextValue(firstValue);
      @   } while (firstValue != newValue);
      @   normal_behavior
      @     assignable lastValue;
      @     ensures lastValue==newValue;
      @
      @   assert lastValue == newValue;
      @
      @   return newValue;
      @ }
      @ also
      @ assignable lastValue, myState;
      @*/
    public int fixpoint() {
        //@ ghost int firstValue;
        //@ ghost int newValue;
        int oldLast;
        
        do {
            //@ set firstValue = lastValue;
            //@ set newValue = nextValue(firstValue);
            oldLast = lastValue;
        }
        while(oldLast != next());
        //@ assert lastValue == newValue;
        
        return lastValue;
    }

    /*@ public model_program
      @ {
      @   ghost int firstValue;
      @   ghost int newValue;
      @
      @   do {
      @     set firstValue = lastValue;
      @     set newValue = nextValue(firstValue);
      @   } while (firstValue != newValue);
      @   normal_behavior
      @     assignable lastValue;
      @     ensures lastValue==newValue;
      @
      @   assert lastValue == newValue;
      @
      @   return newValue;
      @ }
      @ also
      @ assignable lastValue, myState;
      @*/
    public int fixpoint2() {return recursiveFixpoint(lastValue);}

    /*@ private model_program
      @ {
      @   ghost int firstValue = lastValue;
      @   ghost int newValue = nextValue(firstValue);
      @
      @   normal_behavior
      @     assignable lastValue;
      @     ensures lastValue==newValue;
      @
      @   if (firstValue == newValue)
      @     return firstValue;
      @   return recursiveFixpoint(lastValue);
      @ }
      @ also
      @ assignable lastValue, myState;
      @*/
    private int recursiveFixpoint(int oldValue) {
	if (oldValue==next())
	    return oldValue;
	return recursiveFixpoint(lastValue);
    }

    /*@ public model_program
      @ {
      @   ghost int newValue = nextValue(lastValue);
      @
      @   normal_behavior
      @     assignable lastValue;
      @     ensures lastValue == newValue;
      @
      @   if (newValue == 0) {
      @     normal_behavior
      @       assignable lastValue;
      @       ensures lastValue == zeroReplacer;
      @   }
      @   return lastValue;
      @ }
      @ also
      @ assignable lastValue, myState;
      @*/
    public int nextUnlessZero(int zeroReplacer) {
        //@ ghost int newValue = nextValue(lastValue);

        //@ assume true;
        next();
        //@ assert lastValue == newValue;
	
        if (lastValue==0) {
            //@ assume true;
            lastValue = zeroReplacer;
            //@ assert lastValue == zeroReplacer;
        }

	return lastValue;
    }

    /*@ public model_program
      @ {
      @   ghost int jumpSize;
      @   ghost int j;
      @   ghost int jumpResult = lastValue;
      @   choose { set jumpSize = 0; }
      @       or { set jumpSize = 1; }
      @       or { set jumpSize = 2; }
      @       or { set jumpSize = 3; }
      @       or { set jumpSize = 4; }
      @   for(j = 0; j <= jumpSize; j++) {
      @     set jumpResult = nextValue(jumpResult);
      @     normal_behavior
      @       assignable lastValue;
      @       ensures lastValue == jumpResult;
      @   }
      @   return jumpResult;
      @ }
      @ also
      @ assignable lastValue, myState;
      @*/
    public int shortRandomJump() {
        //@ ghost int jumpSize;
        //@ ghost int j;
        //@ ghost int jumpResult = lastValue;
        
        Random r = new Random();
        for(int i=0; i<=r.nextInt(5); i++){
            //@ assume true;
            next();   
            int nvlv = nextValue(lastValue);
            //@ set jumpResult = nvlv;
            //@ assert lastValue == jumpResult;
        }
        
        return lastValue;
    }
}
