/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JmlHTML.java,v 1.7 2007/02/08 14:05:50 leavens Exp $
 */

/* Author: David Cok */


package org.jmlspecs.jmldoc.jmldoc_141;

import org.jmlspecs.jmldoc.*;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.compiler.TabbedPrintWriter;
import org.multijava.mjc.JCompilationUnitType;
import org.multijava.mjc.CompilerPassEnterable;
import org.multijava.mjc.JTypeDeclarationType;
import org.multijava.mjc.JFormalParameter;
import org.multijava.mjc.CClass;
import org.multijava.mjc.CType;
import org.multijava.mjc.CArrayType;
import org.multijava.mjc.CClassType;
import org.multijava.mjc.JVariableDefinition;
import org.multijava.mjc.JFieldDeclarationType;
import org.multijava.util.MessageDescription;
import org.jmlspecs.checker.Constants;
import org.multijava.mjdoc.mjdoc_141.*;


import org.multijava.mjdoc.mjdoc_141.MjdocWrapper;
import com.sun.tools.doclets.standard.HtmlStandardWriter;

import org.multijava.javadoc.JavadocComment;
import org.jmlspecs.checker.*;
import org.multijava.util.Utils;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;

import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.List;
import java.util.LinkedList;

/**
 * This class contains functions to generate html, fitting in with javadoc comments,
 * that documents jml annotations.  The current implementation requires that a
 * javadoc file already exist; the code here will insert additional html into the
 * existing javadoc files.  It is ok to run this operation of inserting additional
 * html multiple times since the inserted material is marked and removed prior to 
 * inserting regenerated material.
 * <P>
 * An improved implementation will generate the html from scratch, by calling
 * appropriate methods in the javadoc implementation, potentially as a doclet.  Note
 * that the current implementation requires that javadoc be runnable on the file;
 * thus the files must be legal java, without mjc extensions.
 * <P>
 * Not currently implemented: 
 * <BR> reading of spec/.jml files (and refines clauses and files)
 <BR> implement weakly, jml_var_assertion, static_initializer, initializer
 <BR> model programs
 <BR>
 <BR>Other comments and bugs:
 <BR> Better wrapping of long specs: mostly ok, but would like to have hanging indents
 <BR> format specifications: can we get rid of blank lines near {| and |}?
 <BR> Add comments, javadoc comments, specifications to this code
 <BR> Better error reporting when a ; is missing from a jml spec.
 <BR> Interleave parsing and jml generation
 <BR> How to record http: links in the docpath and associate them with packages - use multiple -link options ?
 <BR> Resolve all ??? comments in the code, and these comments here.
 <BR> The <- does not appear very elegantly (also _ and \ )
 <BR> Loosen up the syntactic requirements on PRE JML (as tags)
 <BR> sort out the finding of target html files (package-lists etc.)
 <BR> find a better way to find information from CClass objects rather than having
 to find the JmlTypeDeclaration (and parse the relevant file)
 <BR>

 CAUTION: There are static values that are not protected.  If any thought is given
 to multi-threading the generation of html, some work will need to be done here.
*/

public class JmlHTML {

    /** The string that terminates lines on this platform. */
    private static final String eol = System.getProperty("line.separator");

    // These strings are used to bracket the material added to javadoc files by
    // these routines.  That way the material can be found and removed.
    // If you change these strings, you should regenerate the javadoc files
    // that jml has touched, in order to remove any instances of the previous
    // versions of these strings.
    /** Used to begin any section of material added to an HTML file. */
    private static final String jmlbegin = ""; // "<!-- JML -->";
    /** Used to end any section of material added to an HTML file. */
    private static final String jmlend   = ""; // "<!-- JML-END -->" + eol;

    private static final String jmlcomment1 = ""; // "<!-- JMLCOMMENT " + eol;
    private static final String jmlcomment2 = ""; // eol + " ENDJMLCOMMENT -->";
	
    /** This string is used as a suffix for the backup files - the original 
	html files before begin altered by adding jml annotations.
    */
    private static final String oldsuffix = ".old";
	
    /** This string is used as a suffix for the new html file, temporarily;
	Once the file is complete it is moved on top of the old one.
    */
    private static final String newsuffix = ".new";

	

    // The setting of options can be class static because the options 
    // are determined by the command line, and thus
    // are fixed for the entire run of the app.
    /** This is a cached value of the options object supplied here 
	from Main.
    */
    /*@ spec_public @*/ private static JmldocOptions options = null;

    /** This gives the absolute path of the destination for generated html files.
	It is the same as the command-line option, unless the destination is
	not specified on the command-line, in which case this variable is set
	to where the output files are found to be located.
    */
    /*@ spec_protected @*/ private static String adjustedDestination = null;

    /** This field is true if html files generated by javadoc use dollar signs 
	to separate components in the names of HTML files for inner classes 
	(this is not the same as using dollar signs in .class files).
	This is platform dependent.  On Mac OS X, the value is false.  If the
	files cannot be found, the functions below will try alternate values
	of this variable in an attempt to find the files.
    */
    /*@ spec_protected @*/ private static boolean useDollar = false;
	
    /** The value of the suffix that javadoc uses on this platform for 
	html files, with the leading period.
	On Mac OS X: html
    */
    /*@ spec_protected @*/ private static String suffix = ".html";

    /** A second suffix for html files that should be tried. */
    /*@ spec_protected @*/ private static String alternateSuffix = ".htm";
	

    protected File currentFile;
    protected String currentFilePath;

    //--------------------------------------------------------------------------------------------------------------
	
    //@ requires opt != null;	
    public JmlHTML(JmldocOptions opt) {
	if (options == null) adjustedDestination = opt.destination();
	options = opt;
    }
	
    /** This is the entry point into these routines for adding html
	material representing the jml annotations to an html file.
	It does this work for all the classes defined in the
	compilation unit represented by 'tree'.
    */
    //@ requires tree != null;
    public void jmlize(CompilerPassEnterable tree) 
	throws Exception {
	if (tree instanceof JClassDeclarationWrapper) {
	    // Trees of this class are caused by tasks that are processing anonymous classes.
	    // We don't generate any html for those.
	    return;
	}
	if (!(tree instanceof JCompilationUnitType)) {
	    throw new Exception("Method JmlHtml.jmlize was called with "
				+ "an argument that is a " + tree.getClass().toString()
				+ " rather than an instance of JCompilationUnitType.");
	}
	JCompilationUnitType unit = (JCompilationUnitType)tree;
		
	// From the filename of the compilation unit and the package name 
	// (in the file), determine the path
	// to the package root (with a terminating separator).
		
	String packageRoot = null;
	if (unit.packageNameAsString().equals("")) {
	    // This happens when the compilation unit is not part of a package.
	    packageRoot = "." + java.io.File.separator;
	} else {
	    // The absolute path of the unit's file contains the package root
	    // followed by the package name followed by the file name.
	    // So the package root is everything before the package name.
	    int p;
	    if (File.separatorChar == '/') {
		p = unit.file().getAbsolutePath().indexOf(unit.packageNameAsString());
	    } else {
		String pname = unit.packageNameAsString().replace('/',File.separatorChar);
		p = unit.file().getAbsolutePath().indexOf(pname);
	    }
	    if (p == -1) {
		throw new Exception("Compilation unit " + unit.file() +
				    " does not contain the package name (" +
				    unit.packageNameAsString() +
				    "), so the package root cannot be found.");
	    }
	    packageRoot = unit.file().getAbsolutePath().substring(0,p);
	}
		
	// Now loop over all the classes and interfaces declared in 
	// the compilation unit.  Note that there can
	// be only one public class (named in a way corresponding 
	// to the name of the file), but there can be
	// other non-public classes.
		
	JTypeDeclarationType[] jtypes = unit.typeDeclarations();
	for (int j=0; j<jtypes.length; ++j) {
	    // Catch exceptions so that we can go on with other classes in the
	    // file.
	    try {
		jmlize(jtypes[j],packageRoot);
	    } catch (Exception e) {
		System.err.println(unit.file() + "::" + jtypes[j].ident() + " - " + e);
		e.printStackTrace(System.err);
	    } catch (Error e) {
		System.err.println(unit.file() + "::" + jtypes[j].ident() + " - " + e);
		e.printStackTrace(System.err);
	    }
	}
    }
	
    //--------------------------------------------------------------------------------------------------------------

    /** This method generates the html file for the class represented by 'jtype'. */
    //@ requires jtype != null && packageRoot != null;
    //@ modifies adjustedDestination, suffix, alternateSuffix, useDollar;
    //@ modifies currentFile, currentFilePath;
    //@ modifies printedClassSpecsHeader;
    void jmlize(JTypeDeclarationType jtype, String packageRoot) throws Exception {

	// Construct the path name to the html file
		
	String cpath = jtype.getCClass().toString(); 
	// This is the full path name of the class, including the
	// package name and enclosing classes.  The nested class
	// components of the name are separated by $ (as they would
	// be in the name of the compiled class file).

	if (!checkModifiers(jtype.modifiers())) {
	    // Do nothing because the access permission is too high for the given privacy option
	    return;
	}
		
	String rpath = cpath;
	if (!useDollar) rpath = cpath.replace('$','.');
	String dpath;
	if (adjustedDestination == null) {
	    dpath = packageRoot;
	} else {
	    dpath = adjustedDestination;
	}		
	String htmlPath = dpath + java.io.File.separator + rpath + suffix;

	// Now check if a file by that name exists.
	File f = new File(htmlPath);
	if (!f.exists()) {
	    // The current settings of useDollar, suffix, and adjustedDestination did not locate the html file
	    // for this class.  So look at other combinations to see if we can figure out the right values for
	    // this platform.

	    //System.out.println("LOOKING FOR FILE " + htmlPath);
	    if (!findFile(cpath,packageRoot)) {
		System.out.println("ERROR: No html file could be found: " + htmlPath);
		return;
	    }
			
	    // If we were successful in finding a file, then adjustedDestination, useDollar, and suffix have been
	    // changed to match the successful configuration.
	    if (!useDollar) rpath = cpath.replace('$','.');
	    htmlPath = adjustedDestination + java.io.File.separator + rpath + suffix;
	    f = new File(htmlPath);
	    if (!f.exists()) {
		// Some internal logic is wrong if we get here.  Supposedly, findFile constructed this same path
		// and verified that it existed.
		System.out.println("INTERNAL ERROR: FILE DOES NOT EXIST (B): " + htmlPath);
		return;
	    }
	} else {
	    adjustedDestination = dpath;
	}

	currentFile = f;
	currentFilePath = f.getAbsolutePath();
		
	// Now that the file is found, call the following to do the work 
	// of putting in the added html content.
	jmlizeFile(f,jtype);
		
	// Now recursively call this for any nested classes
		
	Iterator itr = jtype.inners().iterator();
	while (itr.hasNext()) { 
	    JTypeDeclarationType intype = (JTypeDeclarationType)(itr.next());
	    // Catch the exception so that we can go on to try the other
	    // inner classes
	    try {
		jmlize(intype,packageRoot);
	    } catch (Exception e) {
		System.err.println(rpath + " - " + e);
		e.printStackTrace(System.err);
	    } catch (Error e) {
		System.err.println(rpath + " - " + e);
		e.printStackTrace(System.err);
	    }
	}
    }

    //--------------------------------------------------------------------------------------------------------------
    /** This function compares the argument to the privacy level set in the command-line options ;
	it returns true if the argument is enabled to be printed per the option.
    */
    static boolean checkModifiers(long mods) {
	return MjdocWrapper.privacyChecker.isVisible(mods);
	/*
	  int level = MjdocWrapper.privacy(); //options.htmlLevel;
	  // level == 0 ==> only public specifications
	  // level == 1 ==> public and protected specifications
	  // level == 2 ==> public, protetcted, package specifications
	  // level == 3 ==> all specifications
	  // Question: does this apply to specifications, to classes, to members?
	  if (level == 3) return true; // level == 3 - allow everything
	  if ((mods & (Constants.ACC_PUBLIC|Constants.ACC_SPEC_PUBLIC)) != 0) return true; // it is public
	  if (level == 0) return false; // it is not public and level == 0
	  if ((mods & (Constants.ACC_PROTECTED|Constants.ACC_SPEC_PROTECTED)) != 0) return true; // it is protected and level > 0
	  if ((mods & Constants.ACC_PRIVATE) != 0) return false; // it is private and level != 3
	  // Note: private requires not spec_public and not spec_protected
	  if (level == 1) return false;  // it is package access and level == 1
	  return true; // it is package access and level == 2
	*/
    }
	
    //--------------------------------------------------------------------------------------------------------------
    /** This function checks whether a html file exists with a name 
	constructed from the given class name (cp), and the given 
	settings of the destination directory (dp), useDollar (useD) and 
	suffix to use (suf).  It returns true if the file could be found.
    */
    //@ requires dp != null && cp != null && suf != null;
    //@ modifies adjustedDestination, suffix, alternateSuffix, useDollar;
    protected boolean checkFile(String dp, boolean useD, String cp, String suf) {
	String path = dp + java.io.File.separator + (useD?cp:cp.replace('$','.')) + suf;
	if (!(new File(path)).exists()) return false;
		
	// Found a file, so set the static values of useDollar, adjustedDestination and
	// suffix to the successful values.
		
	useDollar = useD;
	if (!suf.equals(suffix)) {
	    alternateSuffix = suffix;
	    suffix = suf;
	}
	adjustedDestination = dp;
	//System.out.println("For this platform, found useDollar = " + useDollar + ", suffix = " + suffix 
	//			+ ", destination = " + adjustedDestination + " " + path);
	return true;
    }

    //--------------------------------------------------------------------------------------------------------------

    /** This function finds an html file for the given class name (cpath) and directory
	containing the package root (packageRoot).  It tries various options for
	adjustedDestination, useDollar, and suffix until it finds a successful combination.
	Returns true if some combination is successful, and false if not.  If successful
	adjustedDestination, useDollar, and suffix are set to the successful values.
    */
    //@ requires cpath != null && packageRoot != null;
    //@ modifies adjustedDestination, suffix, alternateSuffix, useDollar;
    protected boolean findFile(String cpath, String packageRoot) {
	String dp;
	if (options.destination() != null) {
	    dp = options.destination();
	    if (checkFile(dp,false,cpath,alternateSuffix)) return true;		
	    if (checkFile(dp,true,cpath,suffix)) return true;
	    if (checkFile(dp,true,cpath,alternateSuffix)) return true;
	    if (checkFile(dp,false,cpath,suffix)) return true;
	} else {
	    if (packageRoot != null) {
		dp = packageRoot;
		if (checkFile(dp,false,cpath,alternateSuffix)) return true;
		if (checkFile(dp,true,cpath,suffix)) return true;
		if (checkFile(dp,true,cpath,alternateSuffix)) return true;
		if (checkFile(dp,false,cpath,suffix)) return true;
		dp = packageRoot + "javadocs";
		if (checkFile(dp,false,cpath,alternateSuffix)) return true;
		if (checkFile(dp,true,cpath,suffix)) return true;
		if (checkFile(dp,true,cpath,alternateSuffix)) return true;
		if (checkFile(dp,false,cpath,suffix)) return true;
	    }

	    dp = ".";
	    if (checkFile(dp,true,cpath,alternateSuffix)) return true;
	    if (checkFile(dp,false,cpath,suffix)) return true;
	    if (checkFile(dp,true,cpath,suffix)) return true;
	    if (checkFile(dp,false,cpath,alternateSuffix)) return true;
	    dp = "javadocs" ;
	    if (checkFile(dp,true,cpath,alternateSuffix)) return true;
	    if (checkFile(dp,false,cpath,suffix)) return true;
	    if (checkFile(dp,true,cpath,suffix)) return true;
	    if (checkFile(dp,false,cpath,alternateSuffix)) return true;
			
	}
	return false;
    }

    //--------------------------------------------------------------------------------------------------------------
    /** This class holds a pair of integers, used here to mark the start and (one past the) end of a 
	section of a text string.
    */
    public static class IntPair implements Comparable {

	//@ invariant offset >= 0;
	/** The beginning of the marked text substring */
	public int offset;
		
	/** One past the end of the marked text substring */
	//@ invariant offset < end;
	public int end;

	/** The constructor for the pair of integers */
	//@ requires offset >= 0 & end > offset ;
	public IntPair(int offset, int end) {
	    this.offset = offset;
	    this.end = end;
	}

	/** The comparison function that implements the Comparable interface.  Object A is less than Object B
	    if the start position is less; if the start positions are equal, then if the end position is less.
	*/
	//@ also requires p != null;
	//@ requires p instanceof IntPair;
	public int compareTo(/*@ non_null @*/ Object p) {
	    IntPair ip = (IntPair)p;
	    if (offset < ip.offset) return -1;
	    if (offset > ip.offset) return 1;
	    if (end < ip.end) return -1;
	    if (end > ip.end) return 1;
	    return 0;
	}

	/** The equality method that implements the Comparable interface.  
	    Two objects are equal if both start and
	    end positions are equal, repectively.
	*/
	//@ also requires p != null;
	//@ requires p instanceof IntPair;
	public boolean equals(/*@ nullable @*/ Object p) {
	    IntPair ip = (IntPair)p;
	    return offset == ip.offset && end == ip.end;
	}
    }

    //--------------------------------------------------------------------------------------------------------------
    /** This class holds an integer and a String, corresponding to a 
	position at which to insert text, 
	and the text to be inserted at that position.
    */
    public static class IntString implements Comparable {
	
	/** The position at which to insert some text. */
	//@ invariant offset >= 0;
	public int offset;
		
	/** The text to be inserted at the given position. */
	//@ invariant text != null;
	public String text;

	/** A constructor, taking an integer position and the text to 
	    be inserted at that position. */
	//@ requires offset >= 0;
	//@ requires text != null;
	public IntString(int offset, String text) {
	    this.offset = offset;
	    this.text = text;
	}

	/** The comparison method that implements the Comparable interface.  
	    The objects are compared first on the
	    'offset' values and then, if those are equal, on the 'text' values.
	*/
        // specification inherited
	public int compareTo(/*@ non_null @*/ Object p) {
	    IntString ip = (IntString)p;
	    if (offset < ip.offset) return -1;
	    if (offset > ip.offset) return 1;
	    return text.compareTo(ip.text);
	}

	/** The equality method that implements the Comparable interface.  The objects are equal if corresponding
	    fields are equal in the two objects.
	*/
	//@ also requires p != null;
	//@ requires p instanceof IntString;
	public boolean equals(/*@ nullable @*/ Object p) {
	    IntString ip = (IntString)p;
	    return offset == ip.offset && text.equals(ip.text);
	}
    }

    //--------------------------------------------------------------------------------------------------------------
    /** This function finds all occurrences of text in 'fileString' bracketed by the given begin and end Strings.  
	The beginning and end of that text segment (including the bracketing strings) is added as an element 
	to the 'extracts' ArrayList.
    */
    public void findSections(String fileString, String begin, String end, ArrayList extracts) throws Exception {
	int n=0;
	int nn;
	while ((nn  = fileString.indexOf(begin,n))!= -1) {
	    n = fileString.indexOf(end,nn);
	    if (n == -1) {
		throw new Exception("The string '" + begin + "' is not matched by a closing '" + end + "' (" + currentFilePath + ")");
	    }
	    n += end.length();
	    extracts.add(new IntPair(nn,n));
	}
	return;
    }
	
    //--------------------------------------------------------------------------------------------------------------
    /** This function finds all occurrences of the string begin in 'fileString' .
	The beginning and end of that text segment (including the bracketing strings) is added as an element 
	to the 'extracts' ArrayList.
    */
    public void findSections(String fileString, String begin, ArrayList extracts) throws Exception {
	int n=0;
	int nn;
	while ((nn  = fileString.indexOf(begin,n))!= -1) {
	    n = nn + begin.length();
	    extracts.add(new IntPair(nn,n));
	}
	return;
    }
	
    //--------------------------------------------------------------------------------------------------------------
    /** This method generates the html describing the specifications of class/interface jmltype.
	The text is appended to the given StringBuffer (without jmlbegin and
	jmlend).  If this is a super class/interface, then 'inherited' should be true.
	'jmltype' is null iff the specs for 'title' cannot be located.
    */
    //@ requires sb != null && title != null;
    //@ modifies (* contents of sb *);
    protected void classSpecs(JmlTypeDeclaration jmltype, 
			      StringBuffer sb,
			      boolean inherited, String title) {
			
	if (jmltype == null && inherited) {
	    // Uncomment the following if you want to list classes/interfaces for which the specs are not
	    // known (because a .java or .jml file cannot be found).
	    //sb.append("<TABLE BORDER=\"1\" CELLPADDING=\"3\" CELLSPACING=\"0\" WIDTH=\"100%\">" + eol);
	    //sb.append("<TR BGCOLOR=\"#EEEEFF\" CLASS=\"TableSubHeadingColor\">" + eol);
	    //sb.append("<TD<B>" + title + "</B> - Unknown specifications</TD>" + eol);
	    //sb.append("</TR></TABLE><P>" + eol);
	    return;
	}

	StringBuffer classspec = new StringBuffer();
	    
	printClassSpecsHeader(classspec,inherited,title);
	    
	boolean anySpecs = false;
	    
	JmlAxiom[] axioms = jmltype.axioms();
	if (axioms.length > 0) anySpecs = true;
	for (int mm=0; mm<axioms.length; ++mm) {
	    JmlAxiom iv = axioms[mm];
	    new SpecWriter(classspec,iv);
	}
	JmlInvariant[] inv = jmltype.invariants();
	for (int mm=0; mm<inv.length; ++mm) {
	    JmlInvariant iv = inv[mm];
	    if (!checkModifiers(iv.modifiers())) continue;
	    new SpecWriter(classspec,iv);
	    anySpecs = true;
	}
	JmlConstraint[] constraints = jmltype.constraints();
	for (int mm=0; mm<constraints.length; ++mm) {
	    JmlConstraint iv = constraints[mm];
	    if (!checkModifiers(iv.modifiers())) continue;
	    new SpecWriter(classspec,iv);
	    anySpecs = true;
	}
	JmlRepresentsDecl[] represents = jmltype.representsDecls();
	for (int mm=0; mm<represents.length; ++mm) {
	    JmlRepresentsDecl iv = represents[mm];
	    if (!checkModifiers(iv.modifiers())) continue;
	    new SpecWriter(classspec,iv);
	    anySpecs = true;
	}
	JmlVarAssertion[] varassertions = jmltype.varAssertions();
	for (int mm=0; mm<varassertions.length; ++mm) {
	    JmlVarAssertion iv = varassertions[mm];
	    if (!checkModifiers(iv.modifiers())) continue;
	    new SpecWriter(classspec,iv);
	    anySpecs = true;
	}
	
	SpecWriter.removeBR(classspec);

	printClassSpecsFooter(classspec);
	    
	// Do the following append unconditionally if you want to list super classes and interfaces
	// for which there are no specifications given in the corresponding .java/jml file
	if (anySpecs) {
	    sb.append(classspec.toString());
	    printedClassSpecsHeader = true;
	}

    }
    
    public void printClassSpecsHeader(StringBuffer classspec, boolean inherited, String title) {
	if (inherited) printClassSpecsHeaderInherited(classspec,title);
	else           printClassSpecsHeaderMain(classspec);
    }
    
    private final String classSpecsTitle = "Class Specifications";
    
    public void printClassSpecsHeaderMain(StringBuffer classspec) {
	classspec.append("<TABLE BORDER=\"1\" CELLPADDING=\"3\" CELLSPACING=\"0\" WIDTH=\"100%\">" + eol);
	classspec.append("<TR BGCOLOR=\"#CCCCFF\" CLASS=\"TableHeadingColor\">" + eol);
	classspec.append("<TD><FONT SIZE=\"+2\"><B>" + classSpecsTitle + "</B></FONT></TD></TR>" + eol);
	classspec.append("<TR BGCOLOR=\"white\" CLASS=\"TableRowColor\"><TD>" + eol);
    }
    
    public void printClassSpecsHeaderShort(StringBuffer classspec) {
	classspec.append("<TABLE BORDER=\"1\" CELLPADDING=\"3\" CELLSPACING=\"0\" WIDTH=\"100%\">" + eol);
	classspec.append("<TR BGCOLOR=\"#CCCCFF\" CLASS=\"TableHeadingColor\">" + eol);
	classspec.append("<TD><FONT SIZE=\"+2\"><B>" + classSpecsTitle + "</B></FONT></TD></TR>" + eol);
	classspec.append("</DL></TABLE><P>" + eol);
    }
    
    public void printClassSpecsHeaderInherited(StringBuffer classspec, String title) {
	if (!printedClassSpecsHeader) {
	    printClassSpecsHeaderShort(classspec);
	}
	classspec.append("<TABLE BORDER=\"1\" CELLPADDING=\"3\" CELLSPACING=\"0\" WIDTH=\"100%\">" + eol);
	classspec.append("<TR BGCOLOR=\"#EEEEFF\" CLASS=\"TableSubHeadingColor\">" + eol);
	classspec.append("<TD><B>" + title + "</B></TD></TR>" + eol);
	classspec.append("<TR BGCOLOR=\"white\" CLASS=\"TableRowColor\"><TD>" + eol);
    }
    
    public void printClassSpecsFooter(StringBuffer classspec) {
	classspec.append("</TD></TR></DL></TABLE><P>" + eol);
    }

    //@ spec_protected
    private boolean printedClassSpecsHeader;

	
    //--------------------------------------------------------------------------------------------------------------
    /** This method inserts into the given File all the jml annotations corresponding to the given
	JTypeDeclarationType.  The file must be the html file for the given class/interface.
    */
    //@ requires f != null && f.exists();
    //@ requires jtype != null;
    //@ requires (* f is the html file for jtype *);
    //@ modifies printedClassSpecsHeader;
    protected void jmlizeFile(File f, JTypeDeclarationType jtype) throws Exception {
	//System.out.println("PROCESSING " + f.toString());
	String fileString; // Will hold the entire contents of the file
	int n;
	int nn;
	int m;

	// Read the file into a buffer

	StringBuffer fileBuffer = new StringBuffer();
	try {
	    FileReader ff = new FileReader(f);
	    char[] cbuf = new char[10000];
	    while ((n=ff.read(cbuf,0,cbuf.length)) != -1) {
		fileBuffer.append(cbuf,0,n);
	    }
	    ff.close();
	    fileString = fileBuffer.toString();
	} catch (Exception e) {
	    //System.out.println(e);
	    throw e;
	}

	ArrayList extracts = new ArrayList();
/*
	// Find any jml stuff already present

	findSections(fileString,jmlbegin,jmlend,extracts);
	findSections(fileString,jmlcomment1,extracts);
	findSections(fileString,jmlcomment2,extracts);
		
	java.util.Collections.sort(extracts);

	// Now reform the fileString leaving out jml stuff

	m = extracts.size();
	nn = 0;
	fileBuffer.setLength(0);
	for (n=0; n<m; ++n) {
	    IntPair p = (IntPair)(extracts.elementAt(n));
	    fileBuffer.append(fileString.substring(nn,p.offset));
	    nn = p.end;
	}
	fileBuffer.append(fileString.substring(nn));
	fileString = fileBuffer.toString();
*/
		
	// Now we figure out what to insert

	ArrayList insertions = new ArrayList();

	// Put any annotations embbeded in javadoc into comments, if requested
		
	/*
	  if (true) { // Don't do this block if you want to keep jml annotations
	  // that are inside a javadoc block in the javadoc output.
	  // They will be also parsed and included in the Specifications
	  // along with other specifications that are not inside a
	  // javadoc comment.
	  // Note that with the mjc parser, these annotations
	  // do not end up in the javadoc at all anyway
	  ArrayList marks = new ArrayList();
	  findSections(fileString,"<pre><jml>","</jml></pre>",marks);
	  findSections(fileString,"<PRE><JML>","</JML></PRE>",marks);
	  findSections(fileString,"<pre><esc>","</esc></pre>",marks);
	  findSections(fileString,"<PRE><ESC>","</ESC></PRE>",marks);
				
	  // ??? Should we do some sort of parsing or regular expression matching
	  // here (or in findSections) - the above requires an exact string match.
				
	  Iterator i = marks.iterator();
	  while (i.hasNext()) {
	  IntPair ip = (IntPair)(i.next());
	  insertions.add(new IntString(ip.offset,jmlcomment1));
	  insertions.add(new IntString(ip.end,jmlcomment2));
	  }
	  }
	*/
	// Change the title line slightly

	// ??? FIXME - this is not robust - it only works when the header
	// is the given stitle string

	String stitle = "<strong>mjc</strong></EM>";
	int nnn = fileString.indexOf(stitle);
	if ( nnn != -1) {
	    nnn = fileString.indexOf("</TD>",nnn);
	    insertions.add(new IntString(nnn,jmlbegin + "<EM><strong>- jml</strong></EM>" + eol + jmlend));
	}
	nnn = fileString.indexOf(stitle,nnn);
	if ( nnn != -1) {
	    nnn = fileString.indexOf("</TD>",nnn);
	    insertions.add(new IntString(nnn,jmlbegin + "<EM><strong>- jml</strong></EM>" + eol + jmlend));
	}
		
	if (!(jtype instanceof JmlTypeDeclaration)) {
	    System.err.println("ERROR: Expected a JTypeDeclarationType to be a JmlTypeDeclaration: " + jtype.toString() + " " + jtype.getClass()); 
	    return;
	    // Not a JML compilation unit, so there is no jml stuff to add - the file remains unchanged.
	}

	JmlTypeDeclaration jmltype = (JmlTypeDeclaration)jtype;

	// Put in any weakly modifiers
		
	{
		
	    // The following is very sensitive to the format of the html file.  It presumes that
	    // the list of implemented interfaces is listed the html like this:
	    //	implements X(, X)*
	    //  X = interfacename | <A HREF="htmllocation">interfacename</A>
	    // The code below searchs for 'implements' and then searches for the interfacename.
	    // If there is a period after the interfacename, then it is presumed that it is part of
	    // the file name in the htmllocation and searchs for the nexxt instance of the 
	    // interfacename.  This is fragile against the interfacename being a substring of another
	    // interfacename or of some other syntax in the html file (e.g. 'HREF' or part of a filename). (???)
	    // It also puts a space between the inserted word 'weakly' and a following comma because of
	    // the eol on the end of jmlend.

	    if (jmltype.getCClass().getSuperClass() != null && jmltype instanceof JmlClassDeclaration && ((JmlClassDeclaration)jmltype).isWeakSubtype()) {
		int q = fileString.indexOf("extends");
		if (q == -1) { // error !!! ???
		}
		q = fileString.indexOf("<DT>",q);
		if (q == -1) { // error !!! ???
		}
		insertions.add(new IntString(q,jmlbegin + " <B>weakly</B>" + jmlend));
	    }
			
	    int q;
	    if (jmltype instanceof JmlClassDeclaration)
		q = fileString.indexOf("implements");
	    else
		q = fileString.indexOf("extends");
	    if (q == -1) { // error !!! ???
	    }
	    CClassType[] interfaces = jmltype.interfaces();
	    boolean[] weaklyFlags = jmltype.interfaceWeaklyFlags();
			
	    for (int i=0; i<weaklyFlags.length; ++i) {
		if (!weaklyFlags[i]) continue;
		String nm = interfaces[i].toString();
		int p = fileString.indexOf(nm,q);
		// System.out.println("WEAKLY FOUND " + nm + " AT " + p);
		if (p == -1) { // ??? error
		}
		p += nm.length();
		if (fileString.charAt(p) == '.') {
		    p = fileString.indexOf(nm,p);
		    p += nm.length() + 4; // 4 is length of "</A>"
		} // ??? com ma is wrong
		insertions.add(new IntString(p,jmlbegin + " <B>weakly</B>" + jmlend));
	    }
	    // ??? what about indirect weakly's
	    // ??? what about conflicting weakly's
	}
			

	// Construct class specification section

	{
	    String csp = generateClassSpecification(jmltype, null);

	    // Where does it go
	    //  Just before the first Inner class, field, Constructor, or Method summary
	
	    int p = fileString.indexOf("SUMMARY ======");
	    p = fileString.lastIndexOf("<!--",p);
	
	    insertions.add(new IntString(p,csp));

	}

	insertFieldSpecs(jmltype,fileString,insertions);
	insertMethodSpecs(jmltype,fileString,insertions);

	// Insert all the new stuff

	java.util.Collections.sort(insertions);

	nn = 0;
	fileBuffer.setLength(0);
	Iterator itr = insertions.iterator();
	while (itr.hasNext()) {
	    IntString p = (IntString)(itr.next());
	    fileBuffer.append(fileString.substring(nn,p.offset));
	    nn = p.offset;
	    if (false && p.text == jmlcomment1) {
		p = (IntString)(itr.next());
		nn = p.offset;
	    } else {
		fileBuffer.append(p.text);
	    }
	}
	fileBuffer.append(fileString.substring(nn));
	fileString = fileBuffer.toString();


	// And finally write the fileString back out to the file

	String pathname = f.toString();
	String oldname = pathname.substring(0,pathname.lastIndexOf('.')); // just the prefix
	String newname = oldname + newsuffix;
	oldname = oldname + oldsuffix;
		
	try {
	    FileWriter ff = new FileWriter(newname);
	    ff.write(fileString,0,fileString.length());
	    ff.close();
	} catch (Exception e) {
	    // Delete it just in case.
	    (new File(newname)).delete();
	    throw e;
	}
	if (!options.Quiet()) {
	    // ??? !FIXME! Would like to use inform but it is not in scope
	    //inform(new MessageDescription("generated {0}","JMLDOC",0),pathname);
	    System.out.println("generated " + pathname);
	}
		
	// Save this renaming until now, so that any exceptions up to this point will leave the original
	// file untouched.
		
	// move original to old
		
	// Uncomment this if you really want the backup, or for testing
	//(new File(pathname)).renameTo(new File(oldname));
		


        // The call of renameTo below fails under Windows OSes if a file exists 
        // with the new file name.  Therefore we have to delete it first. 

        (new File(pathname)).delete(); 

	// move new to original name
		
	(new File(newname)).renameTo(new File(pathname));
		
    }
	
    //--------------------------------------------------------------------------------------------------------------
    /** This generates the full html insert for the class level specifications, including those for any 
	super classes and including the leading jmlbegin and jmlend strings.
    */
    //@ requires jmltype != null;
    //@ modifies printedClassSpecsHeader;
    //@ ensures \result != null;
    protected String generateClassSpecification(JmlTypeDeclaration jmltype, HtmlStandardWriter writer) {

	StringBuffer classspec = new StringBuffer();
	//classspec.append(jmlbegin).append(eol);
	classspec.append("<A NAME=\"" + "class_specifications" + "\"><!-- --></A>" + eol);
		
	// Add the specification for this class

	printedClassSpecsHeader = false;
	classSpecs(jmltype,classspec,false,"Class Specifications");
			
	// Now add the specifications for any super classes

	if (true) {
		
	    CClass stype = jmltype.getCClass();
	    while ((stype = stype.getSuperClass()) != null) {
		//JmlTypeDeclaration t = 
		    //JmlTypeLoader.getJmlSingleton().typeDeclarationOf(stype);
		JmlTypeDeclaration t = null;
		MjClassDoc mjd = MjdocWrapper.lookup(stype);
		if (mjd != null) {
		    JTypeDeclarationType jt = mjd.jtype();
		    if (jt instanceof JmlClassDeclaration) t = (JmlClassDeclaration)jt;
		}
		classSpecs(t,classspec,true,
			   "Specifications inherited from class " 
			   + linkString(stype,false,writer));
	    }
	
				
	    // Now add implemented interface specifications
			
	    List interfaces = getAllInterfaces(jmltype.getCClass());
	    Iterator it = interfaces.iterator();
	    while (it.hasNext()) {
		CClass cclass = ((CClassType)(it.next())).getCClass();					
		JmlTypeDeclaration t = 
		    JmlTypeLoader.getJmlSingleton().typeDeclarationOf(cclass);
		classSpecs(t,classspec,true,"Specifications inherited from interface <I>" 
			   + linkString(cclass,false,writer)
			   + "</I>");
	    }
	}
	//classspec.append(jmlend);
	return classspec.toString();
    }

    //--------------------------------------------------------------------------------------------------------------
    /** The result is a List of CClassType objects containing all the interfaces, with no duplicates. */
    //@ requires cc != null;
    //@ ensures \result != null;
    static List getAllInterfaces(CClass cc) { 
	//System.out.println("GETTING INTERFACES FOR " + jtype.getCClass().toString());
	List result = new LinkedList();

	//CClass cc = jtype.getCClass();
	do {
	    getAllInterfaces(cc.getInterfaces(),result);
	} while ((cc = cc.getSuperClass()) != null);		
		
	return result;
    }
	
	
    //--------------------------------------------------------------------------------------------------------------
    /** This function adds to 'interfaces' any interfaces implemented 
	by the classes or interfaces in 'ints'. 
    */
    //@ requires interfaces != null && (* objects in interfaces are of type CClassType *);
    //@ ensures (* objects in interfaces are of type CClassType *);
    //@ modifies \nothing;
    static private void getAllInterfaces(CClassType[] ints, List interfaces) {
	for (int i=0; i<ints.length; ++i) {
	    CClassType cc = ints[i];
	    //System.out.println(" FOUND " + cc + " " + interfaces.contains(cc));
	    if (!interfaces.contains(cc)) { interfaces.add(cc); }
	}
	// We use two separate loops here so that the interfaces are added to the list in a
	// breadth-first manner, rather than a depth-first manner.
	for (int i=0; i<ints.length; ++i) {
	    getAllInterfaces(ints[i].getCClass().getInterfaces(),interfaces);
	}
    }
	

    //------------------------------------------------------------------------------------------------------------
    /** This function returns a String giving the signature of a method - the method name and fully qualified
	argument types, but not formal parameter names.  It must be called after type checking has been run on
	the parse tree.
    */
    //@ modifies \nothing;
    static String makeSig(JmlMethodDeclaration m, JmlTypeDeclaration jmltype) {
	JFormalParameter[] parameters = m.parameters();
	String s;
	if (m instanceof JmlConstructorDeclaration) {
	    // Need to be sure to include the name of the enclosing class, if this is a nested class
	    s = jmltype.getCClass().toString();
	    s = s.substring(1+s.lastIndexOf(java.io.File.separatorChar)).replace('$','.') + "(";
	} else {
	    // Here we just need the method name
	    s = m.ident() + "("; 
	}
	if (parameters.length > 0) {
	    s = s + parameters[0].getType().toString();
	    for (int i=1; i<parameters.length; ++i) {
		s = s + ", ";
		s = s + parameters[i].getType().toString();
	    }
	}
	s = s + ")";
	return s;
    }
	
    //-----------------------------------------------------------------------------------------------------------
    // ??? Would like to find the method without resorting to the java files - then match back to the
    // right method in the JmlMethodDeclaration
    /** Looks for method jm in class/interface jt; returns null if not found, 
	otherwise returns the corresponding JmlMethodDeclaration.
    */
    //@ modifies \nothing;
    static JmlMethodDeclaration findMatchingMethod(JmlTypeDeclaration jt, JmlMethodDeclaration jm) {
	if (jt == null) return null;
	//System.out.println("SEARCHING FOR " + jm.ident() + " in " + jt.getCClass());
		
	// Iterate through all the methods of jt, looking for one that matches.
	Iterator v = jt.methods().iterator();
	while (v.hasNext()) {
	    JmlMethodDeclaration jmd = (JmlMethodDeclaration)(v.next());
	    //System.out.println("TRYING " + jm.getMethod().toString() + " :: " + jmd.getMethod().toString());
	    if (jm.getMethod().isMoreSpecificThan(jmd.getMethod())) { return jmd; }
	}
	return null;
    }


    //-----------------------------------------------------------------------------------------------------------
    /** This method generates the html text for the specifications of the methods for type
	jmlType, with reference to the original file content in fileString.  The insertion
	points and text are recorded in the insertions ArrayList for later insertion.
    */
    //@ modifies (* the content of insertions *);
    protected void insertMethodSpecs(JmlTypeDeclaration jmltype, String fileString, 
				     ArrayList insertions) throws Exception {
	Iterator i = jmltype.methods().iterator();
	while (i.hasNext()) {
	    JmlMethodDeclaration jm = (JmlMethodDeclaration)(i.next());

	    // Need to add specs for super classes even if this class does not have any specs of its own

	    if (checkModifiers(jm.modifiers())) {
		insertMethodSpecs(jm,fileString,insertions,makeSig(jm,jmltype),jmltype);			
	    }
	}
    }

    //-----------------------------------------------------------------------------------------------------------
    /** This method generates the html text for the specifications of the fields for type
	jmlType, with reference to the original file content in fileString.  The insertion
	points and text are recorded in the insertions ArrayList for later insertion.
    */
    //@ modifies (* the content of insertions *);

    protected void insertFieldSpecs(JmlTypeDeclaration jmltype, String fileString, 
				    ArrayList insertions) throws Exception {
	int pos = fileString.indexOf("Field Detail");
	JFieldDeclarationType[] fields = jmltype.fields();
	for (int i = 0; i<fields.length; ++i) {
	    JFieldDeclarationType jm = fields[i];

	    // FIXME - Need to add specs for super classes even if this class does not have any specs of its own
	    long mods = jm.variable().modifiers();
	    if (checkModifiers(mods)) {
		String s = generateFieldSpecification(jm,null);
		if (s != null) {
		    String name = jm.variable().ident();
		    int nnn = fileString.indexOf("A NAME=\"" + name,pos);
		    nnn = fileString.indexOf("</DL>",nnn);

		    insertions.add(new IntString(nnn,s));
		}
	    }

	}
    }

    public String generateFieldSpecification(JFieldDeclarationType jm,
				JmldocFieldSubWriter fsw) {
	StringBuffer sb = new StringBuffer();
	long mods = jm.variable().modifiers();
	String s = SpecWriter.jmlModifiers(mods);

	if (jm instanceof JmlFieldDeclaration) {
	    SpecWriter sw = new SpecWriter(sb);
	    JmlFieldDeclaration jmm = (JmlFieldDeclaration)jm;
	    if (jmm.hasAssertions()) {
                JmlVarAssertion[] vars = jmm.varAssertions();
                for (int i = 0; i < vars.length; i++) {
                    vars[i].accept(new SpecWriter(sb));
                }
		/* !FIXME!
		  Need to handle data group clauses
		 */
		sb.append("<BR>"+eol);
		//sb.append(jmlend);
	    }
	    JmlInGroupClause[] clauses = jmm.getCombinedInGroupClauses();
	    if (clauses != null && clauses.length > 0) {
		sb.append("<B>is in groups:</B>");
		for (int k=0; k<clauses.length; ++k) {
		    if (clauses[k] == null) { System.out.println("CLAUSESNULL "); continue; }
		    JmlStoreRefExpression[] exprs = clauses[k].groupList();
		    for (int kk=0; kk<exprs.length; ++kk) {
			sb.append(" ");
			if (exprs[kk] == null) System.out.println("NULLEXPR");
			else {
			    addLink(jmm.getField().owner(),exprs[kk],fsw,sw);
			}
		    }
		}
		sb.append("<BR>"+eol);
	    }
	    JmlMapsIntoClause[] mclauses = jmm.getCombinedMapsIntoClauses();
	    if (mclauses != null && mclauses.length > 0) {
		sb.append("<B>maps </B>");
		boolean needsComma = false;
		for (int k=0; k<mclauses.length; ++k) {
		    if (mclauses[k] == null) { System.out.println("MCLAUSESNULL "); continue; }
		    if (needsComma) sb.append(", "); else needsComma = true;
		    mclauses[k].memberRef().accept(sw);
		    sb.append(" <B>\\into</B>");
		    JmlStoreRefExpression[] exprs = mclauses[k].groupList();
		    for (int kk=0; kk<exprs.length; ++kk) {
			sb.append(" ");
			if (exprs[kk] == null) System.out.println("NULLEXPR2");
			else addLink(jmm.getField().owner(),exprs[kk],fsw,sw);
		    }
	 	}
		sb.append("<BR>"+eol);
	    }
	    if (jmm.datagroupContents.size() > 0) {
		// FIXME - the links generated here can be improved -
		// perhaps they should link to the field that declared
		// the in or maps.
		sb.append("<B>datagroup contains:</B>");
		java.util.Iterator ii = jmm.datagroupContents.iterator();
		while (ii.hasNext()) {
		    Object o = ii.next();
		    if (o instanceof JmlFieldDeclaration) {
			sb.append(" " + fsw.getLink(jmm.getField().owner(),((JmlFieldDeclaration)o).getField(),null));
		     } else if (o instanceof JmlStoreRef) {
			sb.append(" ");
			((JmlStoreRef)o).accept(sw);
		    } else {
			System.out.println("UNIMPLEMENTED TYPE " + o.getClass());
		    }
	        }
		sb.append("<BR>"+eol);
	    }
		    
	}
	String ss = sb.toString();

	if (s.length() == 0 && ss.length() == 0) return null;
	return ("<DD><B>Specifications:</B>" + s + "<BR>" + eol + ss);
    }
	
    public void addLink(CClass container, JmlStoreRefExpression sre, 
			JmldocFieldSubWriter fsw, SpecWriter sw) {
	org.multijava.mjc.CFieldAccessor cfa = sre.getField();
	if (cfa == null) {
	    sre.accept(sw);
	} else {
	    org.multijava.mjc.CField cf = cfa.getField();
	    // Need a space (instead of an empty string) to get a link with
	    // no content.
	    String s = fsw.getLink(container,cf," ");
	    if (s.endsWith("</A>")) {
		// The -5 removes the </A> and the single space above.
		sw.append(s.substring(0,s.length()-5));
		sre.accept(sw);
		sw.append("</A>");
	    } else {
		sre.accept(sw);
	    }
	}
    }

    //-----------------------------------------------------------------------------------------------------------
    /** This method generates the specification text for the method jm with signature string sig belonging
	to class jmltype, and with reference to the original html file in fileString.  The jm method may be
	a method of a super class/interface that is being overridden by a method in jmltype.  The new text
	and its insertion point are recorded in the insertions ArrayList.
    */
    //@ modifies \nothing;
    protected void insertMethodSpecs(JmlMethodDeclaration jm, String fileString, ArrayList insertions, String sig,
				     JmlTypeDeclaration jmltype) throws Exception {
	// The following presumes a certain structure to the HTML file.
	// The specifications are inserted in the area of a Method or
	// Constructor detail, just like javadoc tags are inserted.
	// Early in a method detail is a <A NAME=... construct that provides
	// target for a link to this method; we search on the name (which is
	// the signature of the method) to get us into the ballpark.
	// Each method detail is ended by an <HR>, except possibly the last one
	// in a section, which would be terminated by a comment containing ====
	// If there are no javadoc tags, then the text of a javadoc comment is
	// simply terminated by a </DL>.  If there are javadoc tags, the 
	// portion containing the tags is terminated by an earlier </DL>.

	// Find the position at which to insert stuff for this 
	// particular method (in the METHOD DETAIL or
	// CONSTRUCTOR DETAIL section of the html file.
		
	String search = "NAME=\"" + sig;

	boolean foundOK = false; 
	int position = -1;
	boolean oneDL = true;

	int n = fileString.indexOf(search);
	if (n != -1 ) {
	    foundOK = true;
	    int nn = fileString.indexOf("<HR>",n);
	    int nnn = fileString.indexOf("=====",n);
	    if (nn == -1) nn = nnn;
	    else if (nn>nnn && nnn != -1) nn = nnn;
	    if (nn == -1) {
		throw new Exception("Unexpected html file layout encountered in parsing "); 
	    } 

	    nn = fileString.lastIndexOf("</DL>",nn);

	    position = fileString.lastIndexOf("</DL>",nn-1); 
	    oneDL = position < n;
	    if (oneDL) position = nn;
		    
	    // The position in the file is at character 'position' 
	}

	String s = generateMethodSpecification(jm,jmltype.getCClass(),sig,null); //oneDL,null);
		
	if (s != null) {
	    if (!foundOK) { 
		// Turn off this warning for now -- !FIXME! - DRCok
		//System.err.println("WARNING: (Are original javadoc files up to date?) Could not find a method description in the html file in order to insert a specification - " + sig + " in " + jmltype.getCClass().getType().toString());
		return;
	    }
	    insertions.add(new IntString(position,s));
	}
    }
	
    public String generateMethodSpecification(JmlMethodDeclaration jm, 
					      CClass cclass, String sig,
					      HtmlStandardWriter writer) {
	boolean anythingToSay = false;
	StringBuffer s = new StringBuffer();
	//s.append(jmlbegin + eol);
	s.append("<DD><DL>"+eol);
	    
	long mods = jm.getMethod().modifiers();
	String smods = SpecWriter.jmlModifiers(mods);
	if (!Utils.hasFlag(mods,Constants.ACC_PURE)) {
	    if (((JmlMemberAccess)jm.getMethod().owner().access()).isPure())
		smods += " (class)pure";
	    else if (((JmlSourceMethod) jm.getMethod()).isPure())
		smods += " (inherited)pure";
	}

	if (jm.hasSpecification() || smods.length() != 0 ) {
	    //if (needsBR) s.append("<P>"); needsBR = true;
	    s.append("<DT><B>Specifications:</B>");
	    s.append(smods);
	    s.append(eol); 
	    insertSpecification(s, jm.methodSpecification());
	    anythingToSay = true;
//	} else {
//	    needsBR = false;
	}
		
	// Add any method specifications from super classes and interfaces
	// but not for constructors or static members
	    
	if (! (jm instanceof JmlConstructorDeclaration) 
	    && !jm.isStatic()) {

	    CClass sclass = cclass;
	    while ((sclass = sclass.getSuperClass()) != null) {
		boolean b = superSpecification(s,sclass,jm,sig,true,writer);
		if (b) anythingToSay = true;
//		if (b) needsBR = false;
	    }
		
	    List ints = getAllInterfaces(cclass);
	    Iterator it = ints.iterator();
	    while (it.hasNext()) {
		CClass nclass = ((CClassType)(it.next())).getCClass();
		boolean b = superSpecification(s,nclass,jm,sig,false,writer);
		if (b) anythingToSay = true;
//		if (b) needsBR = false;
	    }			
	}
		    
	s.append("</DL></DD>" + eol);
	//s.append(jmlend);
	if (!anythingToSay) return null;
	return s.toString();
	
    }
	
    //-----------------------------------------------------------------------------------------------------------
    /** Generates the html for any specifications for the method 'jm' in the super class 'sclass'.
	'isclass' is true if 'sclass' is a class, false if it is an interface.
	The function returns true if any text was generated, false if no text was generated.
	Any html text generated is appended to the given StringBuffer.
	The String 'sig' is the textual signature of the method.
    */
    //@ modifies \nothing;
    //@ modifies (* just the contents of the StringBuffer s *);
    protected boolean superSpecification(StringBuffer s, CClass sclass, JmlMethodDeclaration jm, String sig, boolean isclass, HtmlStandardWriter writer) {
	JmlTypeDeclaration jt = 
	    JmlTypeLoader.getJmlSingleton().typeDeclarationOf(sclass);
	JmlMethodDeclaration jms = findMatchingMethod(jt,jm);
	boolean anythingToSay = false;
	String start = (isclass ? "class" : "interface");
	String id = sclass.ident();
	if (!isclass) id = "<I>" + id + "</I>"; 
	String ss = writer.getClassLink(MjdocWrapper.lookup(sclass),sig,id,true);
	String methodsig = "";
	if (jt != null && jms != null && jms.hasSpecification()) {
	    MjClassDoc mcd = MjdocWrapper.lookup(jt.getCClass());
	    if (mcd != null) {
		MjMethodDoc mm = mcd.findMethod(jms.getMethod());
		if (mm.parameters().length > 0) methodsig = " " + mm.name() + mm.parameterSignature();
	    }
	}
	start = ("<DT><B>Specifications inherited from overridden method"
		 + methodsig
		 + " in " + start + "</B> " + ss + ":" + eol);
// XXX
	String smods = "";
	if (jms != null) {
	    long mods = jms.getMethod().modifiers();
	    smods = SpecWriter.jmlModifiers(mods);
	    if (!Utils.hasFlag(mods,Constants.ACC_PURE) &&
		((JmlMemberAccess)jms.getMethod().owner().access()).isPure())
		    smods += " (class)pure";
	    if (smods.length() != 0) {
		    start +=  "<BR>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;" + smods;
		    anythingToSay = true;
	    }
	}
	if (jt == null) {
	    // No html file for this class, so don't know the methods or specifications
	    //s.append(start + "<DD>Unknown" + eol);
	    //anythingToSay = true;
	} else if (jms == null) {
	    // The stated class does not have this method
	    //s.append(start + "<DD>No matching method" + eol);
	    //anythingToSay = true;
	} else if (!jms.hasSpecification()) {
	    // There is a corresponding method, but it has no specifications.
	    s.append(start);
	    if (smods.length() == 0) {
		s.append("<BR>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;--- None ---<BR>");
		s.append(eol);
		anythingToSay = true;
	    }
	} else {
	    // There is a method and it has specifications.
	    //if (needsBR) s.append("<BR>");
	    s.append(start);
	    insertSpecification(s,jms.methodSpecification());
	    anythingToSay = true;
	}
	return anythingToSay;
    }
	
    //-----------------------------------------------------------------------------------------------------------
    /** This method actually generates a string representing the given
	JmlMethodSpecification and appends it to the given
	StringBuffer.  This happens by walking the AST that represents
	the method specification.
    */
    protected void insertSpecification(StringBuffer s, JmlMethodSpecification jms) {
	if (jms == null) return;
	String ss = (new SpecWriter(jms)).toString();
	//s.append("<DD>");
	s.append(ss);
    }
	
    //-----------------------------------------------------------------------------------------------------------
    /** This function generates a String giving the name of the class corresponding to 'cclass';
	if an html file can be located for this class, then the string consists of an HTML link
	reference to that class.  If 'fullname' is true, then the name is given with the package
	prefix; otherwise just the name (with enclosing classes) is given.
    */
    //@ modifies \nothing;
    public String linkString(CClass cclass, boolean fullname, HtmlStandardWriter writer) {
	String path = cclass.toString();
	if (!fullname) {
	    int k = path.lastIndexOf('/') + 1;
	    path = path.substring(k);
	}
	if (!useDollar) path = path.replace('$','.'); 
	return  writer.getClassLink(MjdocWrapper.lookup(cclass),"",path,false);
    }
}
	
//-----------------------------------------------------------------------------------------------------------------
/** This class is just used to create a factory instance of the
    JmlHTML class, from which one can call jmlize to do the actual
    work on a real file.  The purpose of this is to allow easy
    subclassing if necessary.  If you want to generate html using a
    different class than JmlHTML, then create that new class (say
    NewHTML - does not need to be derived from JmlHTML), create a
    derived class of JmlHTMLFactory (say NewJmlHTMLFactory), be sure
    that NewJmlHTMLFactory.jmlize instantiates an instance of NewHTML,
    and in Main (or a derived class of Main) assign to htmlGenerator
    an instance of NewJmlHTMLFactory.  An alternate design would be to
    put this factory method in jmlHTML, but then any new Html
    generation class would have to subclass from JmlHtml.
*/
	
// not public
class JmlHtmlFactory {
    public JmlHtmlFactory() {}

    /** This method creates an instance of JmlHTML and calls it to do the
	html generation for the given set of files (in tree) and command
	line options.
		
	@param tree The tree of parsed classes
	@param options The set of command-line options
    */
    //@ requires tree != null;
    //@ requires options != null;
    //@ ensures SpecWriter.options == options; 		
    public void jmlize(CompilerPassEnterable tree, JmldocOptions options) 
	throws Exception {
	SpecWriter.options = options;
	(new JmlHTML(options)).jmlize(tree);
    }
}
