/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.12 2005/04/26 02:40:22 kui_dai Exp $
 * Author: David R. Cok
 */

package org.jmlspecs.jmldoc.jmldoc_142;

import org.jmlspecs.jmldoc.*;

import java.io.File;
import java.io.Reader;
import java.io.FileReader;
import java.io.IOException;
import java.io.FilenameFilter;
import java.io.OutputStream;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import org.multijava.util.classfile.ClassFileFormatException;
import org.multijava.util.compiler.*;
import org.multijava.mjc.*;
import org.multijava.mjdoc.mjdoc_142.MjdocWrapper;
import org.multijava.mjdoc.mjdoc_142.MjClassDoc;
import org.jmlspecs.checker.*;
import org.multijava.util.Destination;

import org.multijava.mjdoc.MjdocMessages;
import org.jmlspecs.checker.Constants;
import org.jmlspecs.checker.JmlOptions; // FIXME - remove when @see lookup fixed

/**
 * This class implements the entry point of the JML compiler.  Use
 * <code>java org.jmlspecs.checker.Main --help</code> to get usage
 * options.  */
public class Main extends org.jmlspecs.checker.Main {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /** The principal constructor for an object to do the work of the
        Jml compiler.*/
    public Main() {
        super();
	init();
        appName = "jmldoc";
        htmlGenerator = new JmlHtmlFactory();
        MjdocWrapper.setWrapper(new JmldocWrapper(this));
        // In javadoc 1.4, stubs are allowed in java files
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    return true;
                }
            }
        );       
    }
    
    /** Construct a JML compiler (an object of <code>Main</code>) that
     * uses the given <code>ModifierUtility</code>. This constructor
     * allows the subclasses to have their own version of modifier 
     * utilities.
     */
    protected Main( ModifierUtility modUtil ) {
        super( modUtil );
	init();
        appName = "jmldoc";
        htmlGenerator = new JmlHtmlFactory();
        MjdocWrapper.setWrapper(new JmldocWrapper(this));
        // In javadoc 1.4, stubs are allowed in java files
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    return true;
                }
            }
        );       
    }

    // ----------------------------------------------------------------------
    // RUN FROM COMMAND LINE
    // ----------------------------------------------------------------------

    /**
     * The entry point when starting this program from the command line.
     *
     * @param   args            the command line arguments
     */
    public static void main(String[] args) {

        boolean success;

        success = compile(args);

        System.exit(success ? 0 : 1);
    }

    /**
     * Second entry point.  This function is overridden from mjc.Main
     * so that we can * instantiate a derived class Main object.  */
    public static boolean compile(String[] args) {
        return new Main().run(args);
    }
    
    // ----------------------------------------------------------------------
    // GUI COMMAND LINE PARSER
    // ----------------------------------------------------------------------
        
    /**
     *  Runs the argument parser only so the GUI can process options from
     *  the command line
     */
    public boolean runParser(String[] args, MjcCommonOptions opt, 
                             ArrayList infiles) {

        args = MjdocWrapper.wrapper().setOptions(args, (JmldocOptions)opt);
        if (args == null) return true;

        if (! super.runParser(args, opt, infiles)) return false;

        return true;
    }
    
    // ----------------------------------------------------------------------
    // RUN FROM GUI
    // ----------------------------------------------------------------------
    
    /**
     * Entry point for the GUI
     *
     * @param   args    the file, package, and directory names
     * @param   opt     the options for the tool
     * @param   os      the output stream for the compiler messages
     */
    public static boolean compile(String[] args, JmldocOptions opt, 
                                  OutputStream os) {
        init();
        return new Main().run(args, opt, os);
    }

    // ----------------------------------------------------------------------
    // HELPER METHODS
    // ----------------------------------------------------------------------

    static void init() {
        MjdocWrapper.setAppName("jmldoc");
        MjClassDoc.memberFilter = new JmldocClassWriter.JmlMemberFilter();
    }

    /** This function creates an object to do the parsing of the
     *  command line arguments and to store the values of the flags
     *  and options so obtained (it does not actually do the argument
     *  parsing).  It is overridden here from mjc.Main so that we can
     *  instantiate a JmlOptions object, in order to parse some new
     *  options specific to this compiler (cf. JmlOptions.opt).
     *  @see JmlOptions */
    protected MjcCommonOptions makeOptionsInstance() {
        JmldocOptions opt = new JmldocOptions();
        exposeOptions(opt);
        return opt;
    }
        
        /** Assigns opt to the local instance of RacOptions and returns
     *  opt as an instance of MjcCommonOptions so it can be assigned to
     *  the options variable in mjc's Main.  This is done so we can access
     *  options that are only processed in MjcCommonOptions.
     */
    //@ also
    //@ requires opt instanceof JmldocOptions;
    protected MjcCommonOptions getOptionsInstance(MjcCommonOptions opt) {
        exposeOptions((JmldocOptions)opt);
        return opt;
    }
    
    /**
     * Parses the argument list.  Mutates infiles, options.  Returns true
     * if there were no problems parsing the arguments.  The list of 
     * command line arguments is parsed per the options defined in 
     * JmlOptions.opt.  Values of given options are stored in the protected
     * variable super.options.  Remaining command-line arguments are 
     * considered to be filenames or directories.  Any directories are 
     * expanded to be all the *.java files in the directory (recursively 
     * including *.java files in subdirectories if the --recursive flag 
     * is set).  The resulting list of files to be processed is stored in
     * the ArrayList infiles.
     */
    //@ also 
    //@ requires args != null && infiles != null;
    //@ modifies options; // (* and the object that is referenced by infiles *)
    //@ ensures options instanceof JmlOptions;
    //@ ensures (* values set in the fields of options correspond to the contents of args *);
    //@ ensures (* items in infiles are non-null Strings corresponding to files that exist and were on the command line *);
    //@ ensures (* \result is true if arguments parsed successfully *);
    public boolean parseArguments(String[] args, ArrayList infiles) {
        
        // First we have to parse arguments that the standard option
        // handler cannot, namely the -link (--link, -L) options.  They
        // are a problem because (a) you can have multiple -link options
        // and the values accumulate and (b) javadoc uses just a single
        // hyphen (-link) and we want to duplicate that functionality.
        
        // Also, for the standalone parser, we need to extract all the 
        // javadoc (single -) style comments.

        args = MjdocWrapper.wrapper().setOptions(args, (JmldocOptions)options);
        if (args == null) return true;

        // Now parse the remaining options in 'args',  leaving in 'infiles' all
        // non-option command-line arguments.
        if (! super.parseArguments(args, infiles)) return false;

        MjdocWrapper.quiet(((JmlCommonOptions)options).Quiet()); 
        SpecWriter.options = (JmldocOptions)jmlOptions;

        if (infiles.size() > 0) filesFound = true;
        
        return true;
    }
	
    /** 
     *  Take care of some interactions among options.
     */
    //@    modifies options;
    /*@    ensures (* option interactions are taken care of *);
      @*/
    protected void handleOptionInteractions() {
        // Turn Quiet off if verbose is on
        if (options.verbose()) ((JmldocOptions)options).set_Quiet(false);
        
        // Turn quiet on if Quiet is on
        if (((JmldocOptions)options).Quiet()) options.set_quiet(true);
    }

    protected ArrayList packageList;

    public void handlePackages(ArrayList packages, ArrayList dirs, ArrayList infiles) {
        packageList = new ArrayList(10);
        Iterator i = packages.iterator();
        while (i.hasNext()) {
            String packageName = (String)(i.next());
	    File d = handlePackageName(packageName,dirs);

            String nn = packageName.replace('.','/') + '/';
            packageList.add(nn);

            if (options.recursive()) addRecursivePackages(nn,d,dirs);

        }

        MjdocWrapper.selectPackages(packageList);

    }
    
    /** This adds all subdirectories (recursively) as packages to be processed. */
    // FIXME - One might rather only add those pakcages that have the same package root, or actually have java files in them,
    // but for now all subdirectories are added.

    protected void addRecursivePackages(String packageName, File dir, ArrayList dirs) {
        String[] ds = dir.list(subdirectoryFilter());
        for (int i=0; i<ds.length; ++i) {
            String n = ds[i];
            String pn = packageName + n + '/';
            File dn = new File(dir,n);
            packageList.add(pn);
            dirs.add(dn);
            addRecursivePackages(pn,dn,dirs);
        }
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------

    // -------------------------------------------------------------
    // TASK AND TASK SEQUENCING CODE
    // -------------------------------------------------------------

    /* !FIXME! This should be a doc comment but there is a jml bug
     * This method uses the dynamic type of <code>oldTask</code> along
     * with the command line options to determine what task to add to
     * the task queue after the given task completes.  
     *
     * <pre><jml>
     * requires oldTask != null && oldTask.completed;
     * ensures \result == null || 
     *         (\result.sequenceID() == oldTask.sequenceID());
     * </jml></pre>
     *
     */
    protected Task createTaskAfter( Task oldTask ) { // !CONVERT! multimethods
        if (oldTask instanceof MjdocTask) {
            return null;
        }
        
        if (oldTask instanceof PreprocessTask) {
	    // Continue at least up through the CheckInitializerTask.
	    // JML often stops here.
            return new CheckInterfaceTask(((PreprocessTask) oldTask).trees(),
                                          oldTask.sequenceID() );
        } else if (oldTask instanceof JmlTypecheckTask) { 
	    // stop JML processing here, i.e., don't continue with 
	    // checking of assignable clauses.
        } else {
            // The oldTask is one of the superclass' tasks.  Do
            // whatever it says next.  If it returns null, then it is
            // finished.
            Task t = super.createTaskAfter( oldTask );
            if (t != null) return t;
        }

        try {
            // All classes need to be registered with the doclet code,
            // so that appropriate links, qulaified names etc. can be
            // used in the documentation.  Only classes in files
            // specified on the command line are 'specified', that is,
            // have html files written out for them.  These are the
            // ones that have the same sequenceID as the mainSequence.
            return new MjdocTask(((TreeProcessingTask) oldTask).trees(),
                                 oldTask.sequenceID());
            

        } catch (Exception e) {
            System.out.println(e); //!FIXME! do something better with this error
            e.printStackTrace(System.out);
        }
        return null;
                
    }

    /**
     * Implemented in jmldoc to cause the generation of all the html
     * files by the appropriate call into the doclet api.  This method
     * is called by {@link #run(String[])} when all reentrant
     * invocations of {@link #processTaskQueue()} have been exited and
     * in this implementation the method finalizes the generation of
     * the HTML.  This method should be overridden by subclasses to
     * perform final work after completion of all tasks.  */
    protected void taskQueueEmptied() {
        long startTime = System.currentTimeMillis();
        MjdocWrapper.setDoclet("org.jmlspecs.jmldoc.jmldoc_142.JmldocStandard");
        MjdocWrapper.privacyChecker = new JmlPrivacyChecker(MjdocWrapper.privacy());
        boolean b = MjdocWrapper.complete();
        if (!b) errorFound = true;
        if (!quietMode()) {
            inform( MjdocMessages.HTML_GEN,
                    new Long(System.currentTimeMillis() - startTime) );
        }
    }

    static public class JmlPrivacyChecker extends MjdocWrapper.MjPrivacyChecker {
        public JmlPrivacyChecker(long privacy) {
                super(privacy);
        }

        public boolean isVisible(long mods) {
            if ((mods & Constants.ACC_SPEC_PUBLIC) != 0) return true;
            if ((mods & Constants.ACC_SPEC_PROTECTED) != 0
                        && privacy > 0) return true;

            return super.isVisible(mods);
        }
    }   

    // FIXME - can we avoid replicating this code between here and mjdoc
    /**
     * This class drives the generation of html files.  */
    public class MjdocTask extends TreeProcessingTask {

        /**
        * Constructs a task for javadocizing the trees in the given
        * forest.  Stores an alias to trees.  */
        public MjdocTask( CompilerPassEnterable[] trees, 
                                Object sequenceID) {
            super( PRI_PRETTY_PRINT, sequenceID, trees, MjdocMessages.REGISTER ); 
        }

        protected void processTree( CompilerPassEnterable tree ) {
            try {
                MjdocWrapper.wrapper().register(tree);
                if (sequenceID() == mainSequenceID()) {
                    MjdocWrapper.wrapper().selected(tree);
                }

            } catch (Exception e) {
                // Report a problem, but continue to process other files
                // !FIXME! How to report it in the process return value???
                //errorFound = true; // Does this allow continuing with other files???
                reportTrouble(e);
            }
        }
    }

    /**
     * This class jmlizes java files for each listed java file.  */
    public class JmlHtmlTask extends TreeProcessingTask {

        /** This flag is true if the files identified by this task are to have
            html written out for them.  If not, the classes contained in the file
            are simply registered in the Doclet code so that their characteristics
            (e.g. qualified name) can be properly written in the html files that
            are generated.
        */  
        private boolean specify;

        /**
        * Constructs a task for javadocizing the trees in the given
        * forest.  Stores an alias to trees.  */
        public JmlHtmlTask( CompilerPassEnterable[] trees, 
                                Object sequenceID, boolean specify ) {
            super( PRI_PRETTY_PRINT, sequenceID, trees, MjdocMessages.HTML_GEN );
            this.specify = specify; 
        }
        
        protected void processTree( CompilerPassEnterable tree ) {
            try {
                if (specify) htmlGenerator.jmlize(tree,(JmldocOptions)options);
            } catch (Exception e) {
                // Report a problem, but continue to process other files
                // !FIXME! How to report it in the process return value???

                reportTrouble(e);
            }
        }
    }

    //---------------------------------------------------------------------
    // Fields
    //---------------------------------------------------------------------
 
    /** This object is used to generate html files.  Derived classes
        may initialize this variable with an alternate derivative of
        JmlHTML, if desired (do so in the derived class constructor).  */
    //@ invariant htmlGenerator != null;
    protected JmlHtmlFactory htmlGenerator;
}
