/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: DefaultOptionsTestSuite.java,v 1.10 2004/03/21 19:02:50 davidcok Exp $
 * Author: David R. Cok
 */
package org.jmlspecs.jmldoc.testcase;

import org.multijava.mjdoc.ToolsVersion;
import org.multijava.util.testing.TestCase;
import org.multijava.util.testing.*;
import junit.framework.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;

/** This class runs the test cases that have a common set of command-line
    options for each test.  Thus there is just one execution of compile().
    There is one TestCase for the compile and then individual TestCases
    for each file comparison with the expected output.  Because of this
    separation, the file comparison test cases are dependent on the compile
    test case having been run, which is not really how JUnit prefers tests
    to be - they should be independent.  However, this way we get separate 
    comparison output for each file tested.
<P>
    The set of external test files is obtained by running the command
    'listTestsWithDefaultOptions'.
<P>
    For a file named x.java, the generated html is put in javadocs/x.html; the
    reference output is in x.html-expected.  If more than one output file needs
    to be checked, the files are named in a line containing JMLOUTPUT in x.java.
    If the test fails, the actual output is moved to x.html-ckd, so that it can
    later be reviewed against x.html-expected.
<P>
    The test must be run from the directory in which it resides - because
    it runs an external process that is expected to be in the current
    directory and opens/creates files in the current directory.
<P>
    Note that since the file comparisons are in a separate
    @author David R. Cok
*/

// NOTE: Most of this code is reused in the derived class NonDefaultOptionsTestSuite
public class DefaultOptionsTestSuite  extends TestSuite {

    /** Returns a Test that is this test suite. */
    public static Test suite() {
        return (new DefaultOptionsTestSuite("defaultOptions"));
    
    }

    /** Accumulates the list of files to be tested. */
    ArrayList filelist = new ArrayList(100);

    /** The constructor called by the JUnit framework. */
    public DefaultOptionsTestSuite(String name) {
	super(name);

	try {
	
	    // Now get all the files that are to be tested.
	    Iterator i = new FileIterator("listTestsWithDefaultOptions");
	    while (i.hasNext()) {
		filelist.add(i.next());
	    }
	    
	    // Create one test for the compile
	    addTest(new Helper());
	    
	    // Create tests for all the file compare operations
	    i = filelist.iterator();
	    while (i.hasNext()) {
		String s = (String)i.next();
		addTestsForFileComparisons(s);
	    }
	
	} catch (java.io.IOException e) {
	    throw new RuntimeException(e.toString());
	}
    }

    /** This constructor is simply here so that derived classes do not
	need to inherit all the tests of the parent class.
    */
    protected DefaultOptionsTestSuite(int i) {
    }

    /** This method adds on the TestCases for file comparisons.  If the
	external test file contains a line containing JMLOUTPUT, then all
	white-space separated tokens in that line, after the first, are
	interpreted as file names that need to be checked.  If there is no
	line containing JMLOUTPUT, then just the .html file corresponding
	to the .java test file is checked.
    */
    //@ requires filename != null;
    //@ requires filename.endsWith(".java");
    void addTestsForFileComparisons(String filename) throws java.io.IOException {
	// Find the files that should be tested
	ArrayList a = Utils.parseFoundLine("JMLOUTPUT",filename);
	if (a == null || a.size() == 0) {
	    String root = filename.substring(0,filename.length()-5); // Removing 5 for .java
	    addTest(new FileCompare(root));
	} else {
	    Iterator ia = a.iterator();
	    ia.next();
	    while (ia.hasNext()) {
		String ss = (String)(ia.next());
		String root = ss.substring(0,ss.length()-5); // Removing 5 for .html
		addTest(new FileCompare(root));
	    }
	}
    }
    
    /** This must match the flags used in the Makefile (JMLFLAGS). */
    String[] defaultFlags = { "-private", "-nonavbar", "-Q",
			      "-classpath", ".",
			      "-f","org.jmlspecs.checker.JMLWarningFilter",
			      "-xnodate", "-d", JAVADOC_LOCATION };

    // A file named x.java produces an html output in JAVADOC_LOCATION/x + ORIG_SUFFIX
    // The expected output is in x + ORACLE_SUFFIX
    // If the result is saved, it is saved in x + SAVED_SUFFIX
    final static String DIFF_COMMAND = "diff -b "; 
    final static String ORIG_SUFFIX = ".html";
    final static String SAVED_SUFFIX = ".html-ckd";
    final static String ORACLE_PREFIX = "expected-" 
		+ ToolsVersion.version() + java.io.File.separator;
    final static String ORACLE_SUFFIX = ".html-expected";
    final static String JAVADOC_LOCATION = "javadocs-" + ToolsVersion.version();

    /** This test exercises all the tests that are supposed to execute
	without producing any output.  They are all run at once.
	The result of Main.compile() has to be true with no output.
    */
    class Helper extends TestCase {
    
	public Helper() { super("TestsWithDefaultOptions-compile"); }
    
	public void runTest() throws java.io.IOException {
	    //System.out.println("\nJML tests with default options");
    
	    // Need to empty out the javadocs area prior to each test
	    if (! Utils.recursivelyRemoveDirectory(new File(JAVADOC_LOCATION),true)) 
		fail("Failed to delete target directory " + JAVADOC_LOCATION);
	    
	    // Run the external process, which must produce a list of the
	    // test filenames, one per line.  A reader reads them into
	    // an ArrayList. 
    
	    // Now create a command-line with the options and all the filenames.
	    String[] args = new String[defaultFlags.length+filelist.size()];
	    System.arraycopy(defaultFlags,0,args,0,defaultFlags.length);
	    int j = defaultFlags.length;
	    Iterator i = filelist.iterator();
	    while (i.hasNext()) args[j++] = (String)(i.next());
    
	    // Set the output streams to write into the ByteArray
	    // and then execute the compile
	    ByteArrayOutputStream ba = new ByteArrayOutputStream();
	    boolean b = false;
	    try {
		Utils.setStreams(new PrintStream(ba));
		b = org.jmlspecs.jmldoc.Main.compile(args);
	    } finally {
		Utils.restoreStreams();
	    }
	    String s = ba.toString();
    
	    // Check that there is no output
	    assertTrue("Unexpected error or warning output has been produced: " + s, s.length() == 0);
    
	    // Check the the result value is true 
	    assertTrue("Running jmldoc should produce a success result (Main.compile() returns false when it should return true)",b);
    
	}
    }

    // Note - shared with NonDefaultOptionsTestSuite.java
    /** A TestCase that compares with actual and expected html outputs, using the DIFF_COMMAND.
	The test fails if the outputs do not match; if the outputs do not match, a -ckd file
	is produced with the actual output.
    */
    static public class FileCompare extends FileCompareTestCase {
    
	private String root;
	
	public FileCompare(String root) {
		// Note: Using File.separator here fails on Windows under 1.4.2
	    super("Comparing " + root + ORIG_SUFFIX,
                  JAVADOC_LOCATION + "/" + root + ORIG_SUFFIX,
                  ORACLE_PREFIX + root + ORACLE_SUFFIX, 
                  DIFF_COMMAND);
	    this.root = root;
	}
	
	public void action(int result, String outputText, String errorText) {
	    if (result != 0) { 
		(new File(JAVADOC_LOCATION + java.io.File.separator + root + ORIG_SUFFIX)).renameTo(new File(root + SAVED_SUFFIX)); 
		//System.out.println(root + " " + result); 
		assertTrue("Output did not match the expected output for " + root + " - exitvalue=" + result 
			    + System.getProperty("line.separator") + outputText + errorText,result==0);
	    } else {
		(new File(root+SAVED_SUFFIX)).delete();
	    }
	}
    }

}


