/*
 * @(#)$Id: Main.java,v 1.2 2008/03/05 00:18:27 wahlst Exp $
 *
 * Copyright (C) 2006 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlexec;

import org.multijava.mjc.*;
import org.jmlspecs.checker.*;
import org.jmlspecs.checker.Constants;
import org.multijava.util.compiler.*;
import org.multijava.util.Utils;
import java.io.*;
import java.util.*;
import org.jmlspecs.jmlrac.*;

/**
 * A class implementing the entry point of the JML EXEC compiler.
 *
 * @see org.jmlspecs.rac.Main
 */
public class Main extends org.jmlspecs.jmlrac.Main {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Constructs a JML EXEC compiler with a default modifier utility. */
    public Main() {
        this(modUtil);
    }

    /** Constructs a JML EXEC compiler with the given modifier utility,
     * <code>modUtil</code>. */
    protected Main(ModifierUtility modUtil) {
        super(modUtil);
    }

    // ----------------------------------------------------------------------
    // RUN FROM COMMAND LINE
    // ----------------------------------------------------------------------

    /** Runs a JML EXEC compiler on the command line arguments. */
    public static void main(String[] args) {
        boolean success = compile(args);
        System.exit(success ? 0 : 1);
    }

    /** Runs a JML EXEC compiler on the given arguments. */
    public static boolean compile(String[] args) {
        return new Main().run(args);
    }
        
    // ----------------------------------------------------------------------
    // RUN FROM GUI
    // ----------------------------------------------------------------------
    
    /**
     * Entry point for the GUI
     *
     * @param   args    the file, package, and directory names
     * @param   opt     the options for the tool
     * @param   os      the output stream for the compiler messages
     */
    public static boolean compile(String[] args, ExecOptions opt,
                                  OutputStream os) {
        return new Main().run(args, opt, os);
    }

        /** Creates an object to do the parsing of the command line arguments
     * and to store the values of the flags and options so obtained 
     * (it does not actually do the argument parsing).  
     * It is overridden here from mjc.Main so that we can
     * instantiate a RacOptions object, in order to parse some new
     * options specific to this compiler (cf. RacOptions.opt).
     * @see RacOptions */
    protected MjcCommonOptions makeOptionsInstance() {
        execOptions = new ExecOptions("jmle"); // prefix to error messages
        exposeOptions(execOptions);
        //opt.Quiet = true; // enforce quiet mode.
        return execOptions;
    }

    /** Assigns opt to the local instance of RacOptions, assigns opt to
         *  an instance of jmlOptions in the checker, and returns opt as an 
         *  instance of MjcCommonOptions so it can be assigned to
     *  the options variable in mjc's Main.  This is done so we can access
     *  options that are only processed in MjcCommonOptions.
     */
    //@ also
    //@ requires opt instanceof RacOptions;
    protected MjcCommonOptions getOptionsInstance(MjcCommonOptions opt) {
        execOptions = (ExecOptions)opt;
        exposeOptions((ExecOptions)opt);
        return opt;
    }

    /**
     * Sets the static field, <code>jmlOptions</code>, to the given
     * arugment so that typechecking routines can access command-line
     * options.
     * 
     * <pre><jml>
     * also
     *   assignable jmlOptions;
     *   ensures jmlOptions == opt;
     * </jml></pre>
     */
    protected void exposeOptions(/*@ non_null @*/ JmlCommonOptions opt) {
        super.exposeOptions(opt);
        execOptions = (ExecOptions) opt;
    }


    /**
     * Returns the next task to be performed after <code>oldTask</code>.
     * This method uses the dynamic type of <code>oldTask</code> along
     * with the command line options to determine what task to add to
     * the task queue after the given task completes. It is overriden
     * here to add JML EXEC specific tasks.
     *
     * <pre><jml>
     * also
     *   requires oldTask != null;
     * </jml></pre>
     */

   protected Task createTaskAfter(Task oldTask) {
       if ((oldTask instanceof TypecheckTask) && !hasOptionXsymw()
           && !canStopSequenceBeforeAllPasses(oldTask)) {
          // 1st round and not recursively referenced types,
          // so generate EXEC code from AST
          JmlCompilationUnit[] trees = filterOut(
              ((TypecheckTask) oldTask).trees());
          return (trees.length > 0) ?
              new JmlTranslateAssertionTask(trees,oldTask.sequenceID())
             : null;
        } else if ((oldTask instanceof TranslateMJTask) && 
                   !canStopSequenceBeforeAllPasses(oldTask)) {
          // 1st round and not recursive type
          JmlCompilationUnit[] trees = filterOut(
              ((TranslateMJTask) oldTask).trees());
          return (trees.length > 0) ?
              new JmlTranslateAssertionTask(trees,oldTask.sequenceID())
              : null;                
        } else if (oldTask instanceof JmlTranslateAssertionTask) {
            // start the second round of compilation, i.e.,
            // compilation of RAC code
            prepareSecondRound();
            return firstTask(((JmlTranslateAssertionTask)oldTask).racFiles());
        }

        return super.createTaskAfter(oldTask);
    }

/*
  protected Task createTaskAfter(Task oldTask) {

       if (oldTask instanceof JmlTypecheckTask) {
            if (!hasOptionXsymw() &&
                canStopSequenceBeforeAllPasses( oldTask )) {
                // For 2nd round, next task should be TranslateMJTask
                return ((RacOptions)options).nowrite() ? null :
                    new TranslateMJTask(((TypecheckTask) oldTask).trees(),
                                        oldTask.sequenceID());
            } else if (hasOptionXsymw()) {
                // Experimental feature to generate symbol (.sym) files;
                // The .sym files don't support checking assignable
                // clauses yet, so the JmlCheckAssignableTask is skipped.
                if(hasOptionXsymw()){
                    if(isSecondRound){
                        return ((RacOptions)options).nowrite() ? null :
                            new TranslateMJTask(((TypecheckTask) oldTask).trees(),
                                                oldTask.sequenceID());
                    }else{
                        CompilerPassEnterable[] trees
                        = ((TypecheckTask) oldTask).trees();
                        return new TranslateMJTask( trees,
                                                       oldTask.sequenceID() );   
                    }
                }
            } else {
                CompilerPassEnterable[] trees
                    = ((TypecheckTask) oldTask).trees();
                return new JmlCheckAssignableTask( trees,
                                                   oldTask.sequenceID() );
            }
        } else if (oldTask instanceof TypecheckTask) {
           if (hasOptionXsymw()) {
                if (isSecondRound) {
                    // translate RAC code into Java bytecode
                    return ((RacOptions)options).nowrite() ? null :
                        new TranslateMJTask(((TypecheckTask) oldTask).trees(),
                                            oldTask.sequenceID());
                } else {
                    // generate symbol files
                    return
                        new TranslateMJTask(((TypecheckTask) oldTask).trees(),
                                            oldTask.sequenceID());
                }
            } else {
                if (!canStopSequenceBeforeAllPasses(oldTask)) {
                    // 1st round and not recursively referenced types,
                    // so generate RAC code from AST
                    JmlCompilationUnit[] trees = filterOut(
                        ((TypecheckTask) oldTask).trees());
                    return (trees.length > 0) ?
                       new JmlTranslateAssertionTask(trees,oldTask.sequenceID())
                       : null;
                } else {
                    // 2nd round, translate RAC code into Java bytecode
                    return ((RacOptions)options).nowrite() ? null :
                        new TranslateMJTask(((TypecheckTask) oldTask).trees(),
                                            oldTask.sequenceID());
                }
            }
        } else if (oldTask instanceof TranslateMJTask) {
            if (canStopSequenceBeforeAllPasses(oldTask)) {
                // 1st round, recursive type and hasOptionXsymw()
                // or 2nd round
                return null;
            } else {
                // 1st round and not recursive type
                JmlCompilationUnit[] trees = filterOut(
                    ((TranslateMJTask) oldTask).trees());
                return (trees.length > 0) ?
                    new JmlTranslateAssertionTask(trees,oldTask.sequenceID())
                    : null;
            }
        } else if (oldTask instanceof JmlTranslateAssertionTask) {
            // pretty print or write RAC code to internal file
            CompilerPassEnterable[] trees =
                ((JmlTranslateAssertionTask) oldTask).trees();
            if (((RacOptions)options).print())
                return new JmlPrettyPrintTask(trees, oldTask.sequenceID());
            else
                return new JmlWriteAssertionTask(trees, oldTask.sequenceID());

        } else if (oldTask instanceof JmlWriteAssertionTask) {
            // start the second round of compilation, i.e.,
            // compilation of RAC code
            prepareSecondRound();
            return firstTask(((JmlWriteAssertionTask)oldTask).racFiles());
        }

        return super.createTaskAfter(oldTask);
    }
*/



    /** Filter out files that aren't to be processed; we do however keep
        non-.java files, whatever their suffix.
    */
    /** Filters out non-RACable files. E.g., ".jml" files. */
    private static JmlCompilationUnit[] filterOut(
        CompilerPassEnterable[] trees) {
        java.util.ArrayList racables = new java.util.ArrayList(trees.length);
        for (int i = 0; i < trees.length; i++) {
            if ((trees[i] instanceof JmlCompilationUnit)) {
                //String path = ((JmlCompilationUnit)trees[i]).file().getPath();
                JmlCompilationUnit jcu = (JmlCompilationUnit) trees[i];
                if (! jcu.isRefined()) {
                    racables.add(trees[i]);
                }
            }
        }
        return (JmlCompilationUnit[]) racables.toArray(
            new JmlCompilationUnit[racables.size()]);
    }

    // -------------------------------------------------------------
    // NESTED CLASSES
    // -------------------------------------------------------------

    public class JmlTranslateAssertionTask extends JmlPrettyPrintTask {
    
       public JmlTranslateAssertionTask( CompilerPassEnterable[] trees,
                                           Object sequenceID ) {
           super( trees, sequenceID );
           racFiles = new ArrayList();
       }
    
       protected void processTree( CompilerPassEnterable tree ) {
         // tree can be an anonymous or local inner class.
         if (tree == null || !(tree instanceof JmlCompilationUnit))
             return;
         try {
             File tmpFile = File.createTempFile("jmlexec", ".java");
             tmpFile.deleteOnExit();
             final JmlCompilationUnit cunit = (JmlCompilationUnit) tree;
             prettyPrint(tmpFile, cunit);
//             System.out.println(cunit.packageNameAsString());
             // if (cunit.packageNameAsString().startsWith("
             racFiles.add(tmpFile.getAbsolutePath());
             getFileMap().put(tmpFile, cunit.file());
         }
         catch (IOException e) {
             noteError();
             inform(e);
         }
       }
    
       /** Pretty-print the AST tree to the external file */
       private void prettyPrint(File file, JmlCompilationUnit tree) {
           ExecPrettyPrinter pp =
               new ExecPrettyPrinter(file, (JmlModifier)modUtil);
           tree.accept(pp);
           pp.close();
       }
    
      public ArrayList racFiles() {
                 return racFiles;
      }
    
      private ArrayList racFiles;
    }

    public static ExecOptions execOptions;
}
