package org.jmlspecs.jmlexec.jack.evaluator;

/**
   Class Constraint
   @version 1 (26/05/1999)
   @author Matthias Schmau
*/

/**
   This class is the superclass for all constraints in the package (i.e. user defined constraints, 
   methods, expressions)
*/

public abstract class Constraint {
    
    /**
       A constraint needs an identifier, to differ it from other constraints ;-) .
    */
    
    protected String identifier;
    
    /**
       The arguments of the constraint are stored in an ObjectContainer.
    */
    
    protected ObjectContainer argList;
    
    /**
       The arity of a constraint is the second characteristic to differ it from other constraints.
    */
    
    protected int arity;
    
    /**
       The variable table administrates the variables.
    */
    
    protected VariableTable vTable;
    

    //////////////////////////////////////
    // methods
    
    public abstract Object clone();
    public Object clone(java.util.Map m) {return clone();};
    
    public abstract String toString();
    
    public abstract void setArguments(ObjectContainer oc) throws SecurityException, VariableUnboundException;
    
    public abstract boolean checkTypes()
	throws WrongArgumentTypeException, UnknownConstraintException;
    
    /**
       This method set the variable table, which administrates the bindings of the variables.
       Each constraint with variables (@see java.lang.Object) has to know the global variable 
       table (@see VariableTable).
    */
    
    public void setVTable(VariableTable vt) {
	
	vTable = vt;
	final int size=argList.size();
	for (int i=0; i<size; i++) {
	    
	    Object o=argList.get(i);
	    if (o instanceof Constraint)
		((Constraint)o).setVTable(vt);
	    
	}
	
    }

    // with this method, some tables are updated
    // goals (G): reset goal variables of a cloned constraint
    // observers (O): add variables and constraint to tables
    // callers: ConstraintSystem, Rule

    public void updateGO(boolean goals, boolean observers) {

	ObjectContainer variables = new ObjectContainer();
	
	final int size=argList.size();
	for (int i=0; i<size; i++) {

	    final Object o=argList.get(i);

	    if (vTable.isVariable(o)) {
		
		if (goals) {

		    if (!vTable.isGoalVar(o))
			vTable.addGoalVar(o);
		    
		}

		if (observers) {

		    // register all variables used in this constraint (I)	
		    
		    // XXX performance - better use a Hash here
		    if (!variables.contains(o))
			variables.add(o);
		    
		    // register constraint as variable observer
		    vTable.addAsVariableObserver(o, i, this);

		}

	    }	       

	}

	// register all variables used in this constraint (II)
	if (observers) {

	    if (variables.size() > 0)
		vTable.addUsedVariables(this, variables);
	    
	}
	    

    }

    // Undoes the effect of updateGO
    public void unUpdateGO()
    {
	final int size=argList.size();
	for (int i=0; i<size; i++) {

	    final Object o=argList.get(i);

	    if (vTable.isVariable(o)) {
		

		// XXX TODO 
		// this cannot be undone yet, since we don't know whether
		// this here is the last constraint in which the variable
		// occured. Introduce a variable count for that.
		// NOTE: if this is implemented, check the trailstack stuff -
		// there, updateGo(FALSE,true) is used.
		// vTable.addGoalVar(o);

		// register constraint as variable observer
		vTable.removeFromObserverList(o, this);
	    }	       

	}	

	vTable.removeVariableUser(this);

    }

    public void setAt(Object o, int i) {

	argList.setAt(o, i);

    }

    public String getIdentifier() {
	
	return identifier;
	
    }
    
    public Object getArgument(int i) {
	
	return argList.get(i);
	
    }

    public int getArity() {
	
	return arity;

    }

    /**
       This method returns true, if one argument of the current
       constraint is a variable.  Constraints with variables must not
       be assigned to variables.
    */
  
    public boolean hasVariableArgument() {
	
	return vTable.isAnyVariable(argList);
	
    }

    /**
       This method removes the bindings of all variables used in the
       constraint's argument list. It is needed to reinitialize a
       constraint used in a constraint handling rule.  
    */
    
    public void unbindAllVariables(TrailStack trailStack) {
	
	if (argList != null)
	    vTable.unbindVarsWithCheck(argList,trailStack);
    }

}


