package org.jmlspecs.jmlexec.jack.evaluator;

import java.util.Hashtable;
// import java.util.Enumeration;
import java.util.Iterator;


/**
 * ConstraintSystem.java <br>
 * $Id: ConstraintSystem.java,v 1.1 2008/03/05 00:08:34 wahlst Exp $
 *
 * @author Matthias Schmau (last modified by $Author: wahlst $)
 * @version $Date: 2008/03/05 00:08:34 $, $Revision: 1.1 $
 */
public class ConstraintSystem {

    // ------------------------------------------------------------------------------
    // fields

    // memory for the CHR; it provides method for adding rules and 
    // running through the set of rules
    private RuleMemory rules;

    // memory for managing user defined constraints; it provides methods
    // for searching instance candiates and for handling active constraints
    private UserDefinedConstraintMemory udConstraints;

    // this memory simly stores executed builtin constraints;
    // it is used to document evaluation runs.
    private BuiltInConstraintMemory builtins;

    // this data structure stores several information about logic variables,
    // e.g. type, name, ...
    private VariableTable vTable;

    // the type of backtracking to be used
    public final static int 
	BACK_INVALID=0,
	BACK_COPY=1,		// copying; this doesn't work correctly yet XXX
	BACK_TRAIL=2,		// trailing stack
	BACK_TESTCASE=3,	// for test cases
	BACK_DEFAULT=BACK_COPY;	// default; when not specified

    private int backtrackingType=BACK_INVALID;

    // BACK_TRAIL:
    // trailing stack, keeps track of all changes to the state. Can be null.
    private TrailStack trailStack;

    // counters for calls of getEnvironment() and setEnvironment()
    public int getEnvironmentCalls=0;
    public int setEnvironmentCalls=0;

    // variables to keep track of the time spent in setEnvironment*
    private long timerTime=0;	// ms


    // ------------------------------------------------------------------------------
    // constrcutors
    
    public ConstraintSystem() {

	this(BACK_DEFAULT);
	
    }
    
    public ConstraintSystem(int backtrackingType) {
	
	// To test a backtracking mechanism against all the regression
	// examples, only this definition needs to be changed; and
	// only the evaluator needs to be compiled (this is only for
	// convenience, so we don't need to recompile all of the
	// examples)

	if (backtrackingType==BACK_TESTCASE)
	    backtrackingType=BACK_TRAIL;

	this.backtrackingType=backtrackingType;

        vTable = new VariableTable();
	rules = new RuleMemory();
	udConstraints = new UserDefinedConstraintMemory(vTable);
	builtins = new BuiltInConstraintMemory();	

	switch (backtrackingType)
	{
	case BACK_COPY:
	    break;
	case BACK_TRAIL:
	    trailStack=new TrailStack();
	    break;
	default:
	    throw new RuntimeException("ConstraintSystem: backtracking type "+
				       backtrackingType+" not implemented");
	}
    }

/*
  public void addVariables(VariableTable table) {
    if (table != null) {
      Iterator vars = table.getVariables();
      while (vars.hasNext()) {
        Object var = vars.next();
        if (!vTable.isVariable(var)) {
          addVariable(var, "java.lang.Object");
        }
        if (table.isBound(var)) {
          Object val = table.getBinding(var);
          if (var != val && val != null && !vTable.isBound(var)) {
            setVarValue(var, val);
          }
        }
      }
    }
  }
*/

    // creates a matching SMarkedConstraintSystem
    public SMarkedConstraintSystem createMarkedSystem()
    {
	switch (backtrackingType)
	{
	case ConstraintSystem.BACK_TRAIL:
	    return SMarkedConstraintSystem.createTrailing(this);
	case ConstraintSystem.BACK_COPY:
	    return SMarkedConstraintSystem.createCopying(this);
	default:
	    throw new RuntimeException("ConstraintSystem: backtracking type "+
				       backtrackingType+" not implemented");
	}
    }

    // ------------------------------------------------------------------------------
    // initialization/reinitialization

    // this method does the same like the constructor
    // it resets the constraint system to an initial state
    public void init() {
	
	vTable.unbindAll(null /* trailStack is reset anyway */);
	vTable.removeAllObservers();
	udConstraints = new UserDefinedConstraintMemory(vTable);
	builtins = new BuiltInConstraintMemory();
	rules.reInitRules();
        vTable.removeVariables();

	if (trailStack!=null) trailStack.reset();
		
    }
    
    // this method resets only the rule memory to an initial state
    // (needed for backtracking algorithms)
    public void reInitRules() {

	// XXX -> TrailStack?

	rules.reInitRules();
	
    }
    
    // this method reactivates all user defined constraints (needed
    // for backtracking algorithms) XXX TODO: is this really needed?
    // setAllConstraintsActive() is automatically called in
    // callGoal(), anyway!
    public void reactivateConstraints() {

	// XXX -> TrailStack?

	udConstraints.setAllConstraintsActive();
	
    }
    
    // ------------------------------------------------------------------------------
    // variable handling and constraint retrieval
    
    // with this method, the user as the possibilty to set
    // the value of a logic variable (needed for result retrieval)
    public void setVarValue(Object var, Object val) {
	
	vTable.bind(var, val, trailStack);

    }

    // this method returns the value of a logic variable
    // (needed for result retrieval)
    public Object getVarValue(Object var) {
// modified by T. Wahls 6/7/2004
                                                                                
/*
        return vTable.getBinding(var);
*/
        boolean goon = true;
        while (goon) {
           if ((vTable.isVariable(var)) && (vTable.isBound(var)))
                var = vTable.getBinding(var);
            else
                goon = false;
                                                                                
        }

        if (var == org.jmlspecs.jmlexec.runtime.Null.NullObj) return null;
// modified by T. Wahls 7/4/2005
        else if (vTable.isVariable(var)) {
          throw new org.jmlspecs.jmlexec.runtime.NoValueException(
            "Post-state value is not bound");
        } else return var;
    }
                                                                                

    // This method returns the value of some variables (needed for
    // result retrieval). If beautiful is true, then if a variable is
    // bound to another variable, instead of that variable, a string
    // describing it is returned.
    public ObjectContainer getVarValues(ObjectContainer vars,
					boolean beautiful) {

	int size=vars.size();
	ObjectContainer result=new ObjectContainer();

	for (int t=0; t<size; t++)
	{
	    Object val=vTable.getBinding(vars.get(t));
	    if (beautiful)
		val=vTable.variableToString(val);
	    result.add(val);
	}
	
	return result;

    }

    // this method tests, whether the given logic variable
    // is bound (needed for result retrieval)
    public boolean hasVarValue(Object var) {

	return vTable.isBound(var);

    }

    // returns the variable object represented by the
    // given String
    public Object getVariableObject(String v) {

	return vTable.getVariableObject(v);

    }

    // added by TAW 7/11/2006
    public Class getVariableType(Object var) {

	return vTable.getVariableClass(var);

    }

    // with this method a logic variable is declared for the
    // constraint system (object, type, identifer (debug)). 
    public void addVariable(Object var, String vClass, String vName) {
	
	// TrailStack is handled in VariableTable.declare()

	vTable.declare(var, vClass, vName);

    }
    
    // with this method a logic variable is declared for the
    // constraint system (object, type). 
    public void addVariable(Object var, String vClass) {
	
	// TrailStack is handled in VariableTable.declare()

	vTable.declare(var, vClass);

    }

    // this method returns possible instances of the give 
    // user defined constraint (needed for result retrieval)
    public ObjectContainer getInstances(UDConstraint c) {
	
	ObjectContainer oc=udConstraints.getInstanceCandidates(c);
	if (oc==null) oc=new ObjectContainer();	// XXX
	return oc;

    }

    public ObjectContainer getUsedVariables(Constraint c) {       

	ObjectContainer vars=vTable.getUsedVariables(c);

	// XXX we should simply return null, but then the caller has
	// to handle a null result
	if (vars==null) vars=new ObjectContainer();

	return vars;

    }

    public ObjectContainer getVariableObservers(Object var) {

	ObjectContainer observers=vTable.getVariableObservers(var);

	// XXX we should simply return null, but then the caller has
	// to handle a null result
	if (observers==null) observers=new ObjectContainer();

	return observers;
    }

    public void removeVariableObserver(Object var, Constraint c) {

	// TrailStack is handled in VariableTable.removeVariableObserver()

	vTable.removeFromObserverList(var, c);

    }

    public String getVariableName(Object var) {

	return vTable.getVariableName(var);

    }
    
    public ObjectContainer getVariableNames(ObjectContainer oc) {

	ObjectContainer result = new ObjectContainer();
	java.util.Iterator enu = oc.iterator();

	while(enu.hasNext()) 
	    result.add(vTable.getVariableName(enu.next()));

	return result;

    }

    // Returns all occurences of a constraint q(...,V,...), where q is
    // the identifier, V is the variable, and V is at the given
    // position. Does NOT return constraints q(...,W,...), where V=W.
    // XXX TODO this should probably be private.
    // XXX TODO ...and if made private, the call to variablesAreEqualOrBound
    //            should be removed (if getConstraints() is only used together
    //            with getEqualConstraints())
    public ObjectContainer getConstraints(String identifier,
					  Object var,int position)
    {
	ObjectContainer result=new ObjectContainer();

	// Get all constraints which contain the variable

	ObjectContainer observers = getVariableObservers(var);

	if (observers!=null)
	{
	    int size=observers.size();
	    for (int i=0; i<size; i++) 
	    {
		VarObserver vo = (VarObserver)observers.get(i);
		Constraint constraint=((Constraint)(vo.observer));

		// The variable must be at the requested position

		if (vo.position==position &&
		    constraint.getIdentifier().equals(identifier))
		{
		    Object theVar=constraint.getArgument(position);

		    if (vTable.variablesAreEqualOrBound(var,theVar))
			result.add(constraint);
		}

	    }
	}

	return result;
    }

    
    // Returns all occurences of a constraint q(...,V,...), where q is
    // the identifier, V is the variable, and V is at the given
    // position. DOES return constraints q(...,W,...), where V is equal to W.
    // XXX TODO probably the name should be changed (the method has nothing
    //    to do with equality constraints)
    public ObjectContainer getEqualConstraints(String identifier,
					       Object var,int position)
    {
	ObjectContainer equalVars=vTable.getEqualVars(var);

	ObjectContainer result=new ObjectContainer();
	int size=equalVars.size();
	for (int t=0; t<size; t++)
	    result.concatenate(getConstraints(identifier,
					       equalVars.get(t),
					       position));

	// XXX TODO: optimize this - pass "result" to
	// getConstraints(), so no new ObjectContainers have to be
	// created and their contents copied for each variable.

	return result;
    }

    // ------------------------------------------------------------------------------
    // addition of rules and constraints


    // this method adds a rule to the constraint system
    // (needed for handler definition)
    public void addRule(Rule r) {

	// Rules are not kept in the TrailStack (yet?), since it is
	// assumed that they will not change during evaluation of the
	// problem.
	if (trailStack!=null) trailStack.assertNotSearching();

	// give rule information about the variables
	r.setVTable(vTable);
	
	try {
	    
	    if (r.checkTypes())
		rules.add(r);

	} catch (WrongArgumentTypeException e) {
	    
	    System.out.print(""+e+"\n");
	    System.out.print("Rule will not be added to rule memory\n");

	} catch (UnknownConstraintException e) {
	    
	    System.out.print(""+e+"\n");
	    System.out.print("Rule will not be added to rule memory\n");
	}

    }
    
    // this method adds constraint to the constraint system;
    // it sorts builtin and user defined constraints into the correct
    // constraint memery (needed for handler definition)
    public void addGoalConstraint(Constraint c)
    {
	addGoalConstraint(c,false);
    }

    public void addGoalConstraint(Constraint c,boolean trace) {

// 	// It is assumed that this will not be called during
// 	// evaluation of the problem.
// 	if (trailStack!=null) trailStack.assertNotSearching();
	
	c.setVTable(vTable);
	c.updateGO(true, true);

	try {
	    
	    if (c.checkTypes()) {
		
		if (c instanceof UDConstraint) {
		    
		    udConstraints.add((UDConstraint)c,true,trailStack);
		    
		} else if (c instanceof BuiltInConstraint) {
		    
		    if (!((BuiltInConstraint)c).execute(false, null /* see above */)) {
			
			if (trace)
			    System.out.println("\t'false' has been inserted into constraint memory\n"+
					       "\t  reason: "+c+"");

			builtins.add(new Bool("false"), c, trailStack);
			
		    } else

			builtins.add((BuiltInConstraint)c, trailStack);
		    
		} 
		
	    }
	    
	} catch (UnknownConstraintException e) {
	    
	    System.out.print(""+e+"\n");
	    System.out.print("Constraint will not be added to goal!\n");
	    
	} catch (WrongArgumentTypeException e) {
	    
	    System.out.print(""+e+"\n");
	    System.out.print("Constraint will not be added to goal!\n");
	    
	}
    }

    // this method adds an ObjectContainer (containing constraints)
    // to the constraint system
    public void addGoalConstraints(ObjectContainer oc) {
	
		for (int i = 0; i < oc.size(); i++)
			addGoalConstraint((Constraint)oc.get(i));

    }

    // remove a constraint from the memory (builtin constraints cannot
    // be removed since they are "executed" immediately when being
    // inserted into the store)
    public void removeGoalConstraint(UDConstraint c)
    {
	c.unUpdateGO();
	udConstraints.remove(c,trailStack);

    }

    // ------------------------------------------------------------------------------
    // activation of the evaluator

    public boolean callGoal() {

	return callGoal(true);

    }

    // this method activates the evaluation of a goal
    public boolean callGoal(boolean trace) {
    
	// set all user defined constraints as active
	
	UDConstraint activeConstraint = null;
	udConstraints.setAllConstraintsActive();
        boolean firstTime = true;
	
	// get the intial active user defined constraint; the active
	// constraint list is not decreased; the method getNextActive
	// will have no effect on the call of
	// hasMoreActiveConstraints, until the method setPassive is
	// call for the active constraint
	if (udConstraints.hasMoreActiveConstraints()) 
	    activeConstraint = udConstraints.getNextActive();
	
	// start evaluation loop
	// loop as long as there are one or more active constraints left	
	while (udConstraints.hasMoreActiveConstraints()) { // still more active constraints
	    
	    // builtin constraints are executed as soon as they are
	    // inserted into the buitlin constraint memory, so it is
	    // possible that the constraint false is contained in the
	    // builtin constraint memory
	    if (builtins.containsFalse())
		return false;
	    

	    if (trace) {

		System.out.println("--------------------------------");
		System.out.println(""+getCopyEnvironment());
		System.out.println("--------------------------------\n");
		System.out.println("active constraint: "+activeConstraint);

	    }

 
	    // the rule memory is reninitialized in order to test all
	    // rules; the state of the constraint memories may have
	    // changed during the last run; so the initialization is
	    // neccessary No effect on TrailStack

	    rules.init();
	    
	    // if a rule fires the loop which runs through the rules
	    // is interrupted; the boolean variable success is set, in
	    // order to differ the case, when the rule loop is
	    // finished because of a lack of untested rules.
	    
	    boolean success = false;	    
	    // run through the following loop as long as there are
	    // untested rules or a rule fires.
	    
	    while (rules.hasMoreRules()) {
		
		// get next rule, no problem, because there are rules left
		Rule r = rules.nextRule();
		
		// decide, whether the current rule is a rule with
		// that constraints have possibly to be removed from
		// the constraint memory (simpagation or
		// simplification)
		boolean simpRule = false;
		if ((r instanceof SimplificationRule) || 
		    (r instanceof SimpagationRule))
		    simpRule = true;
		
		// test whether rule fires with the current active
		// constraint i.e.: eventual partner constraints have
		// to be found and the execution of the guard has to
		// be successful

 /*
		if (trace)
		    System.out.println("trying rule: "+r);
 */

		if (r.apply(udConstraints, builtins, 
			    activeConstraint, simpRule, trailStack,
			    trace)) {
		    
		    if (trace) {
		/*System.out.println("--------------------------------");
		System.out.println(""+getCopyEnvironment());
		System.out.println("--------------------------------\n");
		System.out.println("active constraint: "+activeConstraint);*/
			System.out.println("rule " + r + " fires");
                     }
		    
		    // if the rule fires, add to the documentation
		    // data structure
		    // XXX should we keep track of this in TrailStack?
		    rules.addRuleHistory(activeConstraint, r);

// added by TAW 5/25/2007
// makes sure that an invariant constraint gets activated when a precond
//   constraint is solved
/*
                    if (activeConstraint.getIdentifier().equals("precond")) {
	                udConstraints.setAllConstraintsActive();
                    }
*/
		    
		    // decide whether a new active constraint has to
		    // be chosen; of course this works only if there
		    // are active constraints left; another premise is
		    // that the current active constraint is set
		    // passive (see below) or if there are no more
		    // rules left that can be tested
		    if ((!udConstraints.isActive(activeConstraint)) || 
			(!rules.hasMoreRules())) {

			if (udConstraints.hasMoreActiveConstraints()) {
			    activeConstraint = udConstraints.getNextActive();
                        }
			
		    }
		    
		    // success is a testing variable in order to
		    // differ between the case of no more left rules
		    // and a successful firing of a rule
		    success = true;
					
		    // finish loop "nonnatural"
		    rules.finishLoop();
		    
					
		} else {
				    
 /*
		    if (trace)
			System.out.println("\tthe rule does not fire");
 */
				    
		}
		
		
	    }
		
	    // when the "rule"-loop was finished "natural" (i.e. there
	    // were no testing rules left), the current active
	    // constraint is set passive and a new active constraint
	    // is set (of course only, if there are actives left)
	    
//System.out.println("done with constraint: " + activeConstraint);
	    if (!success) {
		
//System.out.println("setting constraint passive: " + activeConstraint);
//System.out.println(udConstraints.hasMoreActiveConstraints());
		udConstraints.setPassive(activeConstraint);
		if (udConstraints.hasMoreActiveConstraints())
		    activeConstraint = udConstraints.getNextActive();
// System.out.println("new active constraint: " + activeConstraint);
		
	    }

	    // Has any variable been bound since the last call of
	    // isUpdated()?  If so, then all constraints are set
	    // active. TODO XXX: We only need to set those constraints
	    // active that are affected by the bound variables (can we
	    // use VariableObservers for this purpose?)
	    
	    if (vTable.isUpdated()) {
//System.out.println("vtable updated ");
		
		udConstraints.setAllConstraintsActive();
		if (udConstraints.hasMoreActiveConstraints())
		    activeConstraint = udConstraints.getNextActive();
		
	    } else if (success) {
                 udConstraints.setExceptionConstraintsActive();
            }
	    	    
	} //end of while
	
	// return false if the execution of a builtin constraint failed
	// and false has been inserted into the builtin constrant memory
	
	if (builtins.containsFalse())
	    return false;
      /* modified by T. Wahls 7/29/2004
	return true;	
      */
        //else if (udConstraints.isEmpty()) {
          return true;	
        /*} else {
	   System.out.println("UDConstraints: " + udConstraints);
	   throw new 
           org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
             "Insufficient information to complete execution.");*/
        //}
    }

    // -----------------------------------------------------------------------------------------
    // debug functions (be careful while using these methods
    // internal changement may cause false results

    // return the current variable table
    public VariableTable getVariableTable() {

	return vTable;

    }
 
    // reset the current variable table
    public void setVariableTable(VariableTable vt) {

	if (trailStack!=null) trailStack.setInvalid();
	
	vTable = vt;

    }

    // return the user defined constraint memory
    public ObjectContainer getUserDefinedConstraintMemory() {

	return udConstraints.elements();
    }

    // return the active user defined constraints
    public ObjectContainer getActiveConstraints() {

	udConstraints.setAllConstraintsActive();
	return udConstraints.getActives();
    }

    // return the bultin constraint memory
    public ObjectContainer getBuiltinConstraintMemory() {

	return builtins.elements();
		
    }

    public Constraint getFailer() {

	return builtins.failer;

    }

    // return the history of fired rules
    public Hashtable getRuleHistory() {
	
		return rules.getRuleHistory();

    }

    // --------------------------------------------------------------------------
    // managing states

    // method for getting the system state
    public Environment getEnvironment()
    {
	switch (backtrackingType)
	{
	case BACK_COPY:
	    return getCopyEnvironment();
	case BACK_TRAIL:
	    return getTrailEnvironment();
	default:
	    throw new RuntimeException("ConstraintSystem.getEnvironment() not implemented for backtracking type "+backtrackingType);
	}
    }

    // reset the Environment (needed for backtracking algorithms) if
    // the type of env is known, better use setCopyEnvironment(),
    // setTrailEnvironment(), etc.)
    public void setEnvironment(Environment env)
    {
	if (backtrackingType==BACK_COPY && 
	    env instanceof CopyEnvironment) 
	{
	    setCopyEnvironment((CopyEnvironment)env);
	}
	else if (backtrackingType==BACK_TRAIL && 
		 env instanceof TrailEnvironment)
	{ 
	    setTrailEnvironment((TrailEnvironment)env);
	}
	else 
	{
	    throw new 
		RuntimeException("Invalid environment class while using "+
				 "backtracking method "+backtrackingType+
				 ": "+env.getClass());
	}
    }

    // return the whole environment (udc memory, bic memory, rule memory)
    public CopyEnvironment getCopyEnvironment() {
	
	getEnvironmentCalls++;

	// vTable is contained in udConstraints.getUDCMState()

	return new CopyEnvironment(udConstraints.getUDCMState(), 
				   (BuiltInConstraintMemory)builtins.clone());
    }

    // reset the Environment (needed for backtracking algorithms
    public void setCopyEnvironment(CopyEnvironment env) {

	setEnvStartTimer();

	setEnvironmentCalls++;

	udConstraints.setUDCMState(env.udcms);
	builtins = (BuiltInConstraintMemory)(env.bicm.clone());	

	reInitRules();

	setEnvEndTimer();
    }

    // return the whole environment (udc memory, bic memory, rule memory)
    public TrailEnvironment getTrailEnvironment() {
	
	getEnvironmentCalls++;

	return new TrailEnvironment(trailStack);
    }

    // reset the Environment (needed for backtracking algorithms
    public void setTrailEnvironment(TrailEnvironment env) {

	setEnvStartTimer();

	setEnvironmentCalls++;

	trailStack.undo(env.mark);
	reInitRules(); // XXX maybe make this unnecessary by trailing the changes

	setEnvEndTimer();
    }

    public Object clone() {

	return this;

    }

    // Returns the average time for one "set" operation, in milliseconds
    public long setEnvAvgTime()
    {
	return timerTime/setEnvironmentCalls;
    }

    // Start one "set" operation
    private void setEnvStartTimer()
    {
	timerTime-=System.currentTimeMillis();
    }

    // End one "set" operation
    private void setEnvEndTimer()
    {
	timerTime+=System.currentTimeMillis();
    }

    // --------------------------------------------------------------------------
    // debug extensions START

    // XXX TODO: this stuff duplicates the callGoal() method. Combine
    // those two in a better way...
	

	/** contains current active constraint (for stepwise call) */
	private UDConstraint activeConstraint = null;
	/** contains the last rule fired (needed for debugging) */
	private Rule currentRule;


	/** 
	 * get the rule memory
	 *
	 * @return the rules object
	 */
	public RuleMemory getRules() {
		return rules;
	}//end of getRules


    /** 
     * initialize stepwise call
     *
     * @note must be called every time start debugging 
     *
     * @note cannot be used with a TrailStack
     */
    public void initializeStepwiseCall() {

	
	if (trailStack!=null)
	{
	    throw new RuntimeException("Disable the trailing stack first.");
	}

	// set all user defined constraints as active
	activeConstraint = null;
	udConstraints.setAllConstraintsActive();
	
	// get the intial active user defined constraint;
	// the active constraint list is not decreased;
	// the method getNextActive will have no effect on
	// the call of hasMoreActiveConstraints, until the
	// method setPassive is call for the active constraint
	if(udConstraints.hasMoreActiveConstraints()) 
	    activeConstraint = udConstraints.getNextActive();

	return;
    }//end of initializeStepwiseCall
    
    
    /**
     * test if there are more steps
     *
     * @return true, if there are more steps; false otherwise
     */
    public boolean hasNextStep() {
	return udConstraints.hasMoreActiveConstraints();
    }//end of hasNextStep
    
    
    /**
     * evaluate next step
     *
     * @return true, if step success; false otherwise
     */
    public boolean nextStep() {
	boolean emptyStep = true;
	if(hasNextStep()) {
	    
//    	    System.out.println("--------------------------------");
//     	    System.out.println(""+getCopyEnvironment());
//     	    System.out.println("--------------------------------\n");	    
//     	    System.out.println("active constraint: "+activeConstraint);
	    
	    // builtin constraints are executed as soon as they are inserted into the
	    // buitlin constraint memory, so it is possible, that the constraint false
	    // is contained in the builtin constraint memory
	    if (builtins.containsFalse())		
		return false;
	    
       	    // the rule memory is reninitialized in order to test
	    // all rules; the state of the constraint memories may have changed
	    // during the last run; so the initialization is neccessary	    
	    rules.init();
	    
	    // if a rule fires the loop, which runs through the rules is 
	    // interrupted; the boolean variable success is set, in order
	    // to differ the case, when the rule loop is finished because of
	    // a lack of untested rules.
	    boolean success = false;
	    
			// run through the following loop as long as there are untested rules
	    // or a rule fires.
	    while (rules.hasMoreRules()) {
		
				// get ntext rule, no problem, because there are rules left
		currentRule = rules.nextRule();
		
				// decide, whether the current rule is a rule, with that 
				// constraints have possibly to be removed from the constraint
				// memory (simpagation or simplification)
		boolean simpRule = false;
		if ((currentRule instanceof SimplificationRule) || 
		    (currentRule instanceof SimpagationRule))
		    simpRule = true;
		    
				// test, whether rule fires with the current active constraint
				// i.e.: eventual partner constraints have to be found and the
				// execution of the guard has to be successful
		
//    		System.out.println("trying rule: "+r);


		if(currentRule.apply(udConstraints, builtins, activeConstraint, simpRule, null /* this debugging stuff cannot be used together with a TrailStack */, false )) {
		    
 		    System.out.println("rule " + currentRule + " fires");
		    
		    // if the rule fires, add to the documentation
		    // data structure
		    rules.addRuleHistory(activeConstraint, currentRule);
		    
		    // decide, whether a new active constraint has to be choosen;
		    // of course this works only, if there are active constraints left;
					// another premise is, that the current active constraint is set
		    // passive (see below) or if there are no more rules left, which
		    // could be tested
		    if ((!udConstraints.isActive(activeConstraint)) || 
						(!rules.hasMoreRules())) {
			
			if (udConstraints.hasMoreActiveConstraints())
			    activeConstraint = udConstraints.getNextActive();

		    }//end of if
		    
		    // success is a testing variable in order to differ between the case
		    // of no more left rules and a successful firing of a rule
		    success = true;
		    
		    // finish loop "nonnatural"
		    rules.finishLoop();

		    // something has changed
		    emptyStep = false;
		} else {

// 		    System.out.println("rule does not fire");
		    
		}//end of if
		
	    }//end of while(rules.hasMoreRules())
		
	    // when the "rule"-loop was finished "natural" (i.e. there were no testing
	    // rulesleft), the current active constraint is set passive and a new active 
	    // constraint is set (of course only, if there are actives left)
	    if(!success) {
		udConstraints.setPassive(activeConstraint);
		if (udConstraints.hasMoreActiveConstraints())
		    activeConstraint = udConstraints.getNextActive();
	    }//end if if(!success)

	    if(vTable.isUpdated()) {
		udConstraints.setAllConstraintsActive();
		if(udConstraints.hasMoreActiveConstraints())
		    activeConstraint = udConstraints.getNextActive();
	    }//end of if(vTable.isUpdated())
	    
		} else {
		    // no more to do
		    emptyStep = false;
		}//end of if(hasNextStep())
	
	// if no changes in constraint store do next step
	if(emptyStep) {
	    return nextStep();
	}//end of if
	
	return !builtins.containsFalse();
    }//end of nextStep

    
    /** 
     * get the last rule fired
     *
     * @return the currentRule object
	 */
    public Rule getCurrentRule() {
	return currentRule;
    }//end of getCurrentRule

    public String getVariableString(Object var) {
      return vTable.variableToString(var);
    }
    
    
    // debug extensions END
    // --------------------------------------------------------------------------

   // added by T. Wahls 7/17/2004
   public boolean isVariable(Object obj) {
     return vTable.isVariable(obj);
   }

}
