package org.jmlspecs.jmlexec.jack.evaluator;

/**
 * SDiscrepancyExploration.java <br>
 * $Id: SDiscrepancyExploration.java,v 1.1 2008/03/05 00:08:35 wahlst Exp $
 *
 * @author Ekkehard Kramer (last modified by $Author: wahlst $)
 * @version $Date: 2008/03/05 00:08:35 $, $Revision: 1.1 $
 */

// Performs LDS exploration of a search tree

// TODO: we should generalise this to allow for more than binary nodes

public class SDiscrepancyExploration implements SExploration
{
    final private int discLimit; // absolute maximum discrepancy count
    private int maxDisc;	// current maximum allowed discrepancy count
    private int maxOccuredDisc;	// the maximum occured discrepancy
    private boolean isFreshNode; // current node hasn't been processed before

    private ObjectContainer stack = new ObjectContainer();

    private SChoice rootChoice;	// for reset()
    private SMarkedConstraintSystem rootSystem; // for reset
   
    public SDiscrepancyExploration(ConstraintSystem system,SChoice choice,
				   int startingMaxDisc,int discLimit)
    {       
	this.discLimit=discLimit;
	this.maxDisc=startingMaxDisc;
	this.maxOccuredDisc=0;

	rootChoice=choice;
	reset(system);
    }

    // During SNode.createNextChild(), and thereby SChoice.choose(),
    // this method can be used to test whether the current node has
    // been tried before (when exploring the search tree with a lower
    // maximum discrepancy count). This can be used to ignore
    // duplicate solutions.
    // At other times, the return value of this method is undefined.
    public boolean isFreshNode()
    {
	return isFreshNode;
    }

    // Reset everything
    public void reset(ConstraintSystem system)
    {
	maxDisc=0;
	maxOccuredDisc=0;

	// restore the constraint system and re-initialise the root choice
	rootSystem=system.createMarkedSystem(); 
	rootChoice.reset();

	// Push the root node onto the stack
	stack.removeAll();
	stack.add(new SDiscrepancyNode(rootSystem,rootChoice));
    }
    
    // Returns null when finished
    public SNode performOneStep()
    {

	int stackSize=stack.size();

	if (stackSize==0) 
	{
	    // The search is finished.

	    // If the maximum occured discrepancy count is less than the
	    // maximum allowed discrepancy count, we are truly done.

	    if (maxOccuredDisc<maxDisc)
		return null;
	    
	    // Increase the maximum discrepancy count, if possible
	    
	    int newMaxDisc=maxDisc+1;

	    if (newMaxDisc>discLimit)
		return null;

	    // Reset everything

	    reset(rootSystem.getSystem());

	    maxDisc=newMaxDisc;
	    
	    // ... and continue; stackSize=1 now

	    stackSize=stack.size();
	}

	// Pop the current element from the stack

	SDiscrepancyNode node=(SDiscrepancyNode)(stack.get(stackSize-1));
	stack.remove(stackSize-1);
	isFreshNode=node.discrepancy==maxDisc; 

	// If this is one of the special nodes, return it
	// (it has no children)

	if (node.node.isSuccess() || node.node.isFailure())
	    return node.node;

	// Create the child

	SDiscrepancyNode nextNode=node.createNextChild();	

	if (nextNode.discrepancy>maxOccuredDisc)
	    maxOccuredDisc=nextNode.discrepancy;

	// Push the same node again, if this wasn't the last child
	// and the discrepancy count is not at its maximum
	
	if (node.discrepancy<maxDisc &&
	    node.node.hasMoreChildren())
	    stack.add(node);

	// Push the child (since the discrepancy count of the parent
	// was lower than maxDisc, the discrepancy count of this node
	// can at most be maxDisc)

	stack.add(nextNode);

	// XXX TODO remove this test
	if (nextNode.discrepancy>maxDisc)
	    throw 
		new RuntimeException("Paranoia: nextNode.discrepancy>maxDisc");

	return nextNode.node;
    }
}

class SDiscrepancyNode
{
    SNode node;
    int discrepancy;

    SDiscrepancyNode(SMarkedConstraintSystem markedSystem,SChoice choice)
    {
	this(new SNode(markedSystem,choice),0);
    }

    SDiscrepancyNode(SNode node,int discrepancy)
    {
	this.node=node;
	this.discrepancy=discrepancy;
    }

    public SDiscrepancyNode createNextChild()
    {
	// Calculate the new discrepancy count

	int newDisc=discrepancy;
	if (node.getCurrentChild()>0)
	    newDisc++;

	// Create the new node

	return new SDiscrepancyNode(node.createNextChild(),
				    newDisc);
    }
}
