package org.jmlspecs.jmlexec.jack.evaluator;

/**
 * TrailStack.java <br>
 * $Id: TrailStack.java,v 1.1 2008/03/05 00:08:35 wahlst Exp $
 *
 * @author Ekkehard Kramer (last modified by $Author: wahlst $)
 * @version $Date: 2008/03/05 00:08:35 $, $Revision: 1.1 $
 */

public class TrailStack
{    
    // true -> something has gone wrong; this is no problem, as long
    // as the changes are to be undone, in which case an exception is
    // thrown
    private boolean isInvalid = false;

    // true -> getMark() has been called at least once, and changes
    // are actually traced (else, the add() methods are no-ops)
    private boolean getMarkHasBeenCalled = false;

    // the actual stack
    private ObjectContainer stack = new ObjectContainer();

    // output error messages etc.
    public static int 
	DBG_NONE = 0,
	DBG_UNDO = 1,
	DBG_MARK = 2,
	DBG_ADD = 3,		// if debug is >=1, debug calls to add() 
	DBG_ALL = 99;

    public static int debug = DBG_NONE;

    // just for error catching: contains all marks that are "active",
    // i.e., which have been requested with getMark(), but not yet
    // returned with setMark()
    // XXX TODO: remove this, it's just a paranoid check
    // private ObjectContainer marks = new ObjectContainer();

    // encapsulates a position in the stack
    class Mark
    {
	private int position;
	
	private Mark(int position)
	{
	    this.position=position;
	}

	public String toString()
	{
	    return "Mark [ "+position+" ]";
	}
    }
    
    // No special constructor
    public TrailStack()
    {	
    }

    // After this call, the stack looks the same as after creating it
    public void reset()
    {
	isInvalid=false;
	getMarkHasBeenCalled=false;
	stack.removeAll();
// 	marks.removeAll();
    }

    // This call is used to mark the whole stack as invalid. Certain
    // methods will throw exceptions after it.
    public void setInvalid()
    {
	isInvalid=true;
    }

    // When this method is called, no call do getMark() must have been done
    // before. This is a way to make sure that we do not accidently call
    // certain methods of ConstraintSystem etc. which are not expected
    // to be called while searching is performed.
    public void assertNotSearching()
    {
	if (getMarkHasBeenCalled)
	    throw new RuntimeException("Assertion failed! We are searching.");
    }

    // This is the core method of this class: it adds a Runnable
    // object which, if it's run() method is called, undoes a change
    // to the system state. Calls of this method should be minimized
    // as much as possible for obvious performance reasons.
    public void add(Runnable undo)
    {
	//	add("?",undo);

	// We only have to keep track of changes when there is a
	// chance that they will be undone.
	if (!getMarkHasBeenCalled) return;

	if (debug>=DBG_ADD)
	    System.out.println("TS #"+stack.size()+" (no comment)");

	stack.add(undo);

    }

//     public void add(String comment,Runnable undo)
//     {
// 	// We only have to keep track of changes when there is a
// 	// chance that they will be undone.
// 	if (!getMarkHasBeenCalled) return;

// 	if (debug>=DBG_ADD)
// 	    System.out.println("TS #"+stack.size()+": "+comment);

// 	stack.add(undo);
//     }

    // Get a "bookmark" of the current state
    public Mark getMark()
    {
	getMarkHasBeenCalled=true;

	if (isInvalid) throw new RuntimeException("TrailStack is invalid");

 	Mark mark=new Mark(stack.size());
//  	marks.add(mark);
	
	if (debug>=DBG_MARK)
	    System.out.println("TS: created "+mark);

	return mark;
    }

    // Restore the state denoted by the "bookmark"
    public void undo(Mark mark)
    {
	if (debug>=DBG_UNDO)
	    System.out.println("TS: undoing up to "+mark);

	if (isInvalid) 
	    throw new RuntimeException("TrailStack is invalid");

	int markPos=mark.position;

	// XXX TODO: remove this, it's just a paranoid check
// 	while (marks.size()>0)
// 	{
// 	    Mark lastMark=(Mark)(marks.get(marks.size()-1));

// 	    if (lastMark.position>markPos)
// 		marks.remove(marks.size()-1);

// 	    if (debug>=DBG_MARK)
// 		System.out.println("TS: last valid mark is "+lastMark);

// 	    if (lastMark.position==markPos)
// 		break;

// 	    if (marks.size()==0)   
// 		throw new RuntimeException("Mark is invalid");
// 	}

	int size=stack.size();

	while (size>markPos)
	{
	    Runnable run=(Runnable)(stack.get(size-1));
	    stack.remove(size-1);
	    size--;
	    run.run();
	}

// 	if (marks.size()==0)
// 	    getMarkHasBeenCalled=false;
    }
}
