package org.jmlspecs.jmlexec.jack.evaluator;

/**
   Class UDCConstraint <br>
   $Id: UDConstraint.java,v 1.1 2008/03/05 00:08:35 wahlst Exp $
   @version 1 (28/05/1999), $Revision: 1.1 $
   @author Matthias Schmau (last modified by $Author: wahlst $)
*/

/**
   This class provides the functionality for user defined constraints
   (e.g. declaration, matching with other constraint, etc).
*/

public class UDConstraint extends Constraint {
    
    //////////////////////////////////////
    // fields

    /**
       This member is needed for a test similar to the occur-check (in
       the unification algorithm) for the match method
    */
    
    private boolean isInChain;
    
    /**
       In a rule, a constraint is marked "remove", when it has to be removed 
       from the constraint memory after the rule has fired.
    */

    private boolean remove;

    /**
       This contains the classes of the constraint's arguments, in order
       to keep the system well typed.
    */
    
    private ObjectContainer argClasses;


    /**
     * contains all variable-assignments needed for
     * matching the constraint's arguments
     */
    private ObjectContainer usedVariableBindings;

   /** added by T. Wahls 7/21/2005
     * was this constraint generated in an assertion context?
     * does the result (last argument) need to be ground?
   */
    private boolean inAssert = false;
    private boolean lastGround = true;

    /**
     * to use the *UsedVariableBindings() methods, this here must be set
     * to true at the start of the main program.
     */
    public static boolean supportUsedVariableBindings=false;


    //////////////////////////////////////
    // constructor

    public UDConstraint(String id, int a) {
	
	identifier = id;
	arity = a;
	argList = new ObjectContainer();
	argClasses = new ObjectContainer();
	usedVariableBindings = null;//new ObjectContainer();

	remove = false;
	isInChain = false;
	
    }

    //////////////////////////////////////
    // methods

    public Object getArgAt(int i) {

	return argList.get(i);

    }

    public void setRemove() {

	remove = true;
	
    }

    public boolean isToRemove() {
	
	return remove;
    
    }

    public int getArity() {
	
	return arity;

    }

    /**
       Returns all unbound variables used in the constraint
    */
    public ObjectContainer getChangeableVars() {
	
	ObjectContainer result = new ObjectContainer();

	final int size=getArity();
	for (int i = 0; i < size; i++) {
	    
	    Object o=argList.get(i);

	    if (vTable.isVariable(o) &&
		!vTable.isBound(o)) {
		
		result.add(o);

	    } 
	    
	}

	return result;
	
    }

    /**
       User defined constraints are typed. To declare a user defined
       constraint, an ObjectContainer with the class names of the
       arguments has to be given to this method. This has to be
       repeated, everytime a new user defined constraint is created.
    */

    public void declareConstraint(ObjectContainer argClassNames) 
	throws ClassNotFoundException {
	
	final int size=argClassNames.size();
	for (int i = 0; i < size; i++) {
	    
	    argClasses.add(Class.forName((String)argClassNames.get(i)));
	
	}

    }

    public void setDeclaration(ObjectContainer argClasses) {

	this.argClasses = argClasses;

    }

    public void setArguments(ObjectContainer oc) {

	if (oc.size() == getArity()) {

	    argList = oc;

	}
	else
	    throw new RuntimeException("Wrong number of arguments ("+
				       getArity()+" expected, "+oc.size()+
				       " found");

    }

    /**
       When a constraint is added to the goal memory or a rule is
       added to the rule memory, this method is called for each
       constraint in order to to check whether the classes of the
       arguments fit to the declaration of the constraint.
    */

    public boolean checkTypes() 
	throws WrongArgumentTypeException, UnknownConstraintException {
 
	Class argClass;

	final int argListSize=argList.size();
	if (argListSize > 0) {

	    if ((argClasses != null) && (vTable != null)) {
	    
		for (int i=0; i < argListSize; i++) {
		
		    Object o=argList.get(i);

		    if (vTable.isVariable(o)) {
			argClass = vTable.getVariableClass(o);
		    } else {
			argClass = o.getClass();
		    }

		    Object c = argClasses.get(i);
/* removed by T. Wahls 7.15.2004
		    if (!c.getClass().equals(Object.class))		
			if (!c.equals(argClass))
			    throw new WrongArgumentTypeException(this, (Class)c, argClass, (i+1));
		    
*/
		}
		
		return true;

	    } else
		throw new UnknownConstraintException(identifier);
	
	} else 
	    return true;

    }

  /**
     This methods tests whether the give constraint is an instance of the 
     current constraint.
  */


    public boolean match(UDConstraint c,TrailStack trailStack) {

	if (supportUsedVariableBindings)
	    usedVariableBindings = new ObjectContainer();

	final int arity=getArity();
	if ((arity == c.getArity()) && 
	    (identifier == c.getIdentifier())) {

	    for (int i = 0; i < arity; i++) {

		if (!matchArguments(argList.get(i), 
				    c.getArgument(i), 
				    trailStack)) {

		    // rueckgaengig machen der letzten Bindung. XXX TODO
		    return false;

		} 

	    }

	} else {

	    return false;

	}

	// "this" constraint is a "rule constraint", so set the used
	// variable bindings for the "real" constraint
	if (supportUsedVariableBindings)
	    c.setUsedVariableBindings(usedVariableBindings);
	return true;

    }

  /**
     Matching of arguments. 
     @see UDConstraint@match.
  */

    public boolean matchArguments(Object general, Object instance, 
				  TrailStack trailStack) {

	// XXX TrailStack!

	Object tmpInstance, tmpGeneral;
	isInChain = false;

	if (vTable.isVariable(general)) { // general is a variable
	
	    if (vTable.isBound(general)) { // general is bound ------------------------------------------

		if (vTable.isVariable(instance) && vTable.isBound(instance)) { // instance is a variable
		    
		    tmpInstance = evalValue(instance, null);
			
		} else { // instance is not bound or instance is not a variable

		    tmpInstance = instance;
	  
		}

		tmpGeneral = evalValue(general, tmpInstance); 
		return isInChain;
	  
	    } else { // general is not bound  -------------------------------------------------------------

		Class generalClass = vTable.getVariableClass(general);
		Class instanceClass;

		if (vTable.isVariable(instance)) { // instance is a variable

		    instanceClass = vTable.getVariableClass(instance);

		} else { // instance is not a variable

		    instanceClass = instance.getClass();
		    
		}

/* removed/modified by T. Wahls 7/15/2004
		if (generalClass.equals(instanceClass)) {
		    
		    if ((instance instanceof Constraint) && (((Constraint)instance).hasVariableArgument()))
			return false;

		    vTable.bind(general, instance, trailStack ); // trailStack correct?
		    return true;
		    
		} else {
		    
		    return false;
		    
		}
*/
                if ((instance instanceof Constraint) && 
                    ((Constraint)instance).hasVariableArgument())
                   return false;
                vTable.bind(general, instance, trailStack );
		return true;
	
	    }
      
	} else { // general is not a variable  --------------------------------------------------

	    if (vTable.isVariable(instance)) {
	    
		if (vTable.isBound(instance))
		   return general.equals(evalValue(instance, null));
		else
		    return false;

	    } else 	  
		 return general.equals(instance);
	    
	}
	
    }

    /**
       this method expects a bound variable as argument
    */

    private Object evalValue(Object var, Object testObj) {

		Object binding = vTable.getBinding(var);

		if (supportUsedVariableBindings)
		    usedVariableBindings.add(vTable.getVariableName(var) + 
					     " = " + 
					     vTable.getVariableName(binding));

		if (testObj != null)
			if (testObj.equals(binding))
				isInChain = true;

		if (vTable.isVariable(binding) && vTable.isBound(binding)) {

			return evalValue(binding, testObj);
		
		} 
	
		return binding;

    }

    public String toString() {

		String s = identifier;
		boolean first = true;

		if (getArity() > 0) {

			s += "(";
	
			for (int i = 0; i < getArity(); i++) {
	  
				if (first)
					first = false;
				else 
					s += ", ";
	    
				Object o = getArgument(i);
				if (vTable.isVariable(o)) {

				    s+=vTable.variableToString(o);

// 					if (vTable.isBound(o))
// 					{
// 					    Object other=vTable.getBinding(o);
// 					    if (vTable.isVariable(other))
// 						other=vTable.getVariableName(other);

// 					    s+= vTable.getVariableName(o) +"(=" + other +")";
// 					}
// 					else 
// 						s+= vTable.getVariableName(o);

				} else if (o.getClass().isArray()) {
                                    Object [] ar = (Object []) o;
                                    s += "[";
                                    boolean firstVal = true;
                                    for (int j = 0; j < ar.length; j++) {
                                      if (firstVal) {
                                        firstVal = false;
                                      } else {
                                        s += ",";
                                      }
                                      Object ob = ar[j];
				      if (vTable.isVariable(ob)) {
				        s+=vTable.variableToString(ob);
                                      } else {
                                        s += Util.JMLEString(ob);
                                      }
                                    }
                                    s += "]";
                                } else {
					s += ""+Util.JMLEString(getArgument(i));
                                }

			}

			return s+")" + " " + inAssert;

		} else
			return s + " " + inAssert;
      
    }

    public void cloneInwards(UDConstraint tmp, java.util.Map ht) {

		tmp.setDeclaration(argClasses);
	
		// temporay variable for the arguments
		Object o;

		// the argument list
		ObjectContainer args = new ObjectContainer();
	
		final int iSize=argList.size();
		for (int i = 0; i < iSize; i++) {
	    
			o = argList.get(i);
      
			if ((vTable.isVariable(o)) && (vTable.isBound(o))) {

				o = vTable.getBinding(o);

				boolean goon = true;
				while(goon) {

					if ((vTable.isVariable(o)) && (vTable.isBound(o)))
						o = vTable.getBinding(o);
					else
						goon = false;

				}
		

         /* modified by Tim Wahls 7/15/2004
				if (vTable.isVariable(o))
					args.add(o);
				else
					args.add(ObjectContainer.generalClone(o));
        */
                             args.add(o);

			} else if (vTable.isVariable(o)) {

         // modified by T. Wahls 7/14/2004
         if (ht != null && ht.containsKey(o)) {
            args.add(ht.get(o));
         } else {
				Object v = new Object(); vTable.declare(v, vTable.getVariableClass(o).getName());
				args.add(v);
           if (ht != null) ht.put(o, v);
         }
         // end modified code
		
			} else
         /* modified by T. Wahls 7/15/2004
				args.add(ObjectContainer.generalClone(o));
         */
                    args.add(o);
      
		}

		if (isToRemove())
			tmp.setRemove();
	
		tmp.setVTable(vTable);
		tmp.setArguments(args);
	
    }

    public void cloneInwards(UDConstraint tmp) {
      cloneInwards(tmp, null);
    }

    public Object clone() {

		// temporay object; the clone
		UDConstraint tmp = new UDConstraint(identifier, arity); 

		cloneInwards(tmp, null);

		return tmp;

    }

    public Object clone(java.util.Map ht) {
		// temporay object; the clone
		UDConstraint tmp = new UDConstraint(identifier, arity); 

		cloneInwards(tmp, ht);

		return tmp;
    }


	/**
	 * determine, if for the matching of the constraint's arguments some
	 * variable-assignments were needed
	 *
	 * @return true, if there were some variable bindings were needed; 
	 *         false in any other case
	 */
	public boolean hasUsedVariableBindings() {

	    if (!supportUsedVariableBindings)
		throw new RuntimeException("Set UDConstraint.supportUsedVariableBindings=true if you want to use this method");

		return (usedVariableBindings.size() > 0);
	}//end of hasUsedVariableBindings


	/**
	 * get the variable-assignments needed for matching the constraint's
	 * arguments
	 *
	 * @return an ObjectContainer with entries in form of "var = binding"
	 */
	public ObjectContainer getUsedVariableBindings() {

	    if (!supportUsedVariableBindings)
		throw new RuntimeException("Set UDConstraint.supportUsedVariableBindings=true if you want to use this method");

		return usedVariableBindings;
	}//end of getUsedVariableBindings


	/**
	 * set the variable-assignments needed for matching the constraint's
	 * arguments
	 *
	 * @return an ObjectContainer with entries in form of "var = binding"
	 */
	public void setUsedVariableBindings(ObjectContainer usedVariableBindings) {

	    if (!supportUsedVariableBindings)
		throw new RuntimeException("Set UDConstraint.supportUsedVariableBindings=true if you want to use this method");

		this.usedVariableBindings = usedVariableBindings;
		return;
	}//end of setUsedVariableBindings

   /** added by T. Wahls 7/21/2005 */
   public UDConstraint setInAssert(boolean inAssert) {
     this.inAssert = inAssert;
     return this;
   }

   public UDConstraint setInAssert(boolean inAssert, boolean lastGround) {
     this.inAssert = inAssert;
     this.lastGround = lastGround;
     return this;
   }

   public boolean inAssert() { return inAssert;}
   public boolean lastGround() { return lastGround;}

}
