/*
 * @(#)$Id: FDomain.java,v 1.2 2008/03/07 21:48:46 wahlst Exp $
 *
 * Copyright (C) 2006 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlexec.runtime;

import java.util.ArrayList;
import java.util.ListIterator;
import java.util.Iterator;

public class FDomain {
  private static class Range {
    /*@ public invariant low <= hi; */
    public final long low;
    public final long hi;

    public Range(long l, long h) {
      low = l;
      hi = h;
    }

    public String toString() {
      return "" + low + "#" + hi;
    }

    public boolean equals(Object other) {
      if (other == null || !(other instanceof Range)) {
        return false;
      } else {
        Range r = (Range) other;
        return low == r.low && hi == r.hi;
      }
    }

    public double size() {
      return (hi - (double) low) + 1;
    }
  }

  // ranges do not overlap and are stored in ascending order
  private ArrayList ranges;
  private long oldValue; // prestate value of variable

  public FDomain() {
    ranges = new ArrayList();
    oldValue = 0;
  }

  public FDomain(long low, long hi, Object old) {
    ranges = new ArrayList();
    ranges.add(new Range(low, hi));
    if (old == null || !(old instanceof MyIntegerType)) {
      oldValue = 0;
    } else {
      oldValue = ((MyIntegerType) old).longValue();
    }
  }

  public long oldValue() { return oldValue; }

  public static boolean inRange(MyIntegerType mi, FDomain fd) {
    return fd.inRange(mi.longValue());
  }

  public boolean inRange(long l) {
    Iterator it = ranges.iterator();
    while (it.hasNext()) {
      Range r = (Range) it.next();
      if (l >= r.low && l <= r.hi) {
        return true;
      }
    }
    return false;
  }

  public double size() {
    double sum = 0;
    Iterator it = ranges.iterator();
    while (it.hasNext()) {
      Range r = (Range) it.next();
      sum += r.size();
    }
    return sum;
  }

  public void remove(long l) {
    ListIterator iter = ranges.listIterator();
    while (iter.hasNext()) {
      Range r = (Range) iter.next();
      if (l == r.low) {
        if (l == r.hi) {
          iter.remove();
        } else {
          r = new Range(l + 1, r.hi);
          iter.set(r);
        }
      } else if (l == r.hi) {
        r = new Range(r.low, l - 1);
        iter.set(r);
      } else if (l >= r.low && l <= r.hi) {
        Range rl = new Range(r.low, l - 1);
        Range rh = new Range(l + 1, r.hi);
        iter.set(rl);
        iter.add(rh);
      }
    }
  }

  public static FDomain remove(MyNumber mi, FDomain fd) {
    double d = mi.value();
    FDomain fd2 = fd.copy();
    if (d == (long) d) {
      fd2.remove((long) d);
    }
    return fd2;
  }
  
  public static FDomain intersect(FDomain f1, FDomain f2) {
    FDomain res = new FDomain();
    ListIterator iter1 = f1.ranges.listIterator();
    ListIterator iter2 = f2.ranges.listIterator();
    Range r1 = null, r2 = null;
    if (iter1.hasNext() && iter2.hasNext()) {
      r1 = (Range) iter1.next();
      r2 = (Range) iter2.next();
      boolean done = false;
      while (!done) {
        if (r1.hi < r2.low) {
          if (iter1.hasNext()) {
            r1 = (Range) iter1.next();
          } else {
            done = true;
          }
        } else if (r2.hi < r1.low) {
          if (iter2.hasNext()) {
            r2 = (Range) iter2.next();
          } else {
            done = true;
          }
        } else if (r2.hi <= r1.hi) {
          long low = Math.max(r1.low, r2.low);
          if (low <= r2.hi) {
            res.ranges.add(new Range(low, r2.hi));
          }
          if (iter2.hasNext()) {
            r2 = (Range) iter2.next();
          } else {
            done = true;
          }
        } else { // r2.hi > r1.hi
          long low = Math.max(r1.low, r2.low);
          if (low <= r1.hi) {
            res.ranges.add(new Range(low, r1.hi));
          }
          if (iter1.hasNext()) {
            r1 = (Range) iter1.next();
          } else {
            done = true;
          }
        }  
      }  
    } 
    return res;
  }

  public void upperBound(double l) {
    ListIterator iter = ranges.listIterator();
    while (iter.hasNext()) {
      Range r = (Range) iter.next();
      if (r.low > l) {
        iter.remove();
      } else if (r.hi > l) {
        Range nr = new Range(r.low, (long) l);
        iter.set(nr);
      }
    }
  }

  public boolean empty() {
    return ranges.size() == 0;
  }

  public static boolean empty(FDomain fd) {
    return fd.empty();
  }

  public boolean determined() {
    if (ranges.size() != 1) {
      return false;
    } else {
      ListIterator iter = ranges.listIterator();
      Range r = (Range) iter.next();
      return r.low == r.hi;
    }
  }
      
  public static boolean determined(FDomain fd) {
    return fd.determined();
  }
  
  public static FDomain upperBound(MyNumber mi, FDomain fd) {
    double d = mi.value();
    if (mi instanceof MyIntegerType) {
      d = d - 1;
    }
    FDomain fd2 = fd.copy();
    fd2.upperBound(d);
//    return !fd.empty();
    return fd2;
  }

  public FDomain copy() {
    FDomain res = new FDomain();
    Iterator it = ranges.iterator();
    while (it.hasNext()) {
      Range r = (Range) it.next();
      res.ranges.add(r);
    }
    res.oldValue = oldValue;
    return res;
  }

  public static FDomain upperBoundEQ(MyNumber mi, FDomain fd) {
    double d = mi.value();
    FDomain fd2 = fd.copy();
    fd2.upperBound(d);
//    return !fd.empty();
    return fd2;
  }

  public void lowerBound(double l) {
    ListIterator iter = ranges.listIterator();
    while (iter.hasNext()) {
      Range r = (Range) iter.next();
      if (r.hi < l) {
        iter.remove();
      } else if (r.low < l) {
       if (l > (long) l) {
         l = l + 1;
       }
       Range nr = new Range((long) l, r.hi);
       iter.set(nr);
      }
    }
  }

  public static FDomain lowerBound(MyNumber mi, FDomain fd) {
    double d = mi.value();
    if (mi instanceof MyIntegerType) {
      d = d + 1;
    }
    FDomain fd2 = fd.copy();
    fd2.lowerBound(d);
//    return !fd.empty();
    return fd2;
  }

  public static FDomain lowerBoundEQ(MyNumber mi, FDomain fd) {
    double d = mi.value();
    FDomain fd2 = fd.copy();
    fd2.lowerBound(d);
//    return !fd.empty();
    return fd2;
  }

  public boolean hasSucc(long l) {
    ListIterator iter = ranges.listIterator();
    while (iter.hasNext()) {
      Range r = (Range) iter.next();
      if (l >= r.low && l < r.hi) {
        return true;
      } else if (l == r.hi) {
        return iter.hasNext();
      }
    }
    return false;
  }

  public long getSucc(long l) {
    ListIterator iter = ranges.listIterator();
    while (iter.hasNext()) {
      Range r = (Range) iter.next();
      if (l >= r.low && l < r.hi) {
        return l + 1;
      } else if (l == r.hi && iter.hasNext()) {
        r = (Range) iter.next();
        return r.low;
      }
    }
    throw new NoValueException("Error: no value for fd variable!");
  }

  public long getFirst() {
   if (inRange(oldValue)) {
      return oldValue;
    }
    ListIterator iter = ranges.listIterator();
    if (iter.hasNext()) {
      Range r = (Range) iter.next();
      return r.low;
    } else {
      throw new NoValueException("Error: no value for fd variable!");
    }
  }

  public long getLow() {
    ListIterator iter = ranges.listIterator();
    if (iter.hasNext()) {
      Range r = (Range) iter.next();
      return r.low;
    }
    throw new NoValueException("Error: no value for fd variable!");
  }

  public static MyLong getLow(FDomain fd) {
    return new MyLong(fd.getLow());
  }

  public long getHi() {
    ListIterator iter = ranges.listIterator();
    Range r = null;
    while (iter.hasNext()) {
      r = (Range) iter.next();
    }
    if (r == null) {
      throw new NoValueException("Error: no value for fd variable!");
    } else {
      return r.hi;
    }
  }

  public static MyLong getHi(FDomain fd) {
    return new MyLong(fd.getHi());
  }

  public static boolean changesUpperBound(FDomain fd, MyNumber n) {
    try {
      double d = n.value();
      long top = fd.getHi();
      return d <= top;
    } catch (NoValueException e) {
      return false;
    }
  }

  public static boolean changesUpperBoundEQ(FDomain fd, MyNumber n) {
    try {
      double d = n.value();
      long top = fd.getHi();
      return d < top;
    } catch (NoValueException e) {
      return false;
    }
  }

  public static boolean changesLowerBound(FDomain fd, MyNumber n) {
    try {
      double d = n.value();
      long bot = fd.getLow();
      return d >= bot;
    } catch (NoValueException e) {
      return false;
    }
  }

  public static boolean changesLowerBoundEQ(FDomain fd, MyNumber n) {
    try {
      double d = n.value();
      long bot = fd.getLow();
    return d > bot;
    } catch (NoValueException e) {
      return false;
    }
  }

  // fd1 + fd2 = n
  public static boolean addChangesDomain1(FDomain fd1, FDomain fd2, 
                                          MyNumber n) {
    try {
      double res = n.value();
      long bot1 = fd1.getLow();
      long bot2 = fd2.getLow();
      long hi1 = fd1.getHi();
      long hi2 = fd2.getHi();
      return bot1 + hi2 < res || hi1 + bot2 > res || 
             bot2 + hi1 < res || hi2 + bot1  > res;
    } catch (NoValueException e) {
      return false;
    }
  }

  // fd1 + fd2 = n
  public static ArrayList addBoundDomain1(FDomain fd1, FDomain fd2,
                                          MyNumber n) {
    double res = n.value();
    long bot1 = fd1.getLow();
    long hi1 = fd1.getHi();
    long bot2 = fd2.getLow();
    long hi2 = fd2.getHi();
    FDomain newFD1 = fd1.copy();
    newFD1.lowerBound(res - hi2);
    newFD1.upperBound(res - bot2);
    FDomain newFD2 = fd2.copy();
    newFD2.lowerBound(res - hi1);
    newFD2.upperBound(res - bot1);
    ArrayList resFDs = new ArrayList(2);
    resFDs.add(newFD1);
    resFDs.add(newFD2);
    return resFDs;
  }

  // fd1 + n = fd2 
  public static boolean addChangesDomain2(FDomain fd1, MyNumber n,
                                          FDomain fd2) {
    try {
      double nVal = n.value();
      long bot1 = fd1.getLow();
      long bot2 = fd2.getLow();
      long hi1 = fd1.getHi();
      long hi2 = fd2.getHi();
      return bot1 + nVal < bot2 || hi1 + nVal > hi2 || bot2 - nVal < bot1 ||
             hi2 - nVal > hi1;
    } catch (NoValueException e) {
      return false;
    }
  }

  // fd1 + n = fd2 
  public static ArrayList addBoundDomain2(FDomain fd1, MyNumber n,
                                          FDomain fd2) {
    double nVal = n.value();
    long bot1 = fd1.getLow();
    long hi1 = fd1.getHi();
    long bot2 = fd2.getLow();
    long hi2 = fd2.getHi();
    FDomain newFD1 = fd1.copy();
    newFD1.lowerBound(bot2 - nVal);
    newFD1.upperBound(hi2 - nVal);
    FDomain newFD2 = fd2.copy();
    newFD2.lowerBound(bot1 + nVal);
    newFD2.upperBound(hi1 + nVal);
    ArrayList resFDs = new ArrayList(2);
    resFDs.add(newFD1);
    resFDs.add(newFD2);
    return resFDs;
  }

  // fd1 + fd2 = fd3 
  public static boolean addChangesDomain3(FDomain fd1, FDomain fd2,
                                          FDomain fd3) {
    try {
      long bot1 = fd1.getLow();
      long bot2 = fd2.getLow();
      long bot3 = fd3.getLow();
      long hi1 = fd1.getHi();
      long hi2 = fd2.getHi();
      long hi3 = fd3.getHi();
      return bot1 + bot2 > bot3 || hi1 + hi2 < hi3 ||
             bot3 - hi2 > bot1 || hi3 - bot2 < hi1 ||
             bot3 - hi1 > bot2 || hi3 - bot1 < hi2;
    } catch (NoValueException e) {
      return false;
    }
  }

  // fd1 + fd2 = fd3 
  public static ArrayList addBoundDomain3(FDomain fd1, FDomain fd2,
                                          FDomain fd3) {
    long bot1 = fd1.getLow();
    long hi1 = fd1.getHi();
    long bot2 = fd2.getLow();
    long hi2 = fd2.getHi();
    long bot3 = fd3.getLow();
    long hi3 = fd3.getHi();
    FDomain newFD1 = fd1.copy();
    newFD1.lowerBound(bot3 - hi2);
    newFD1.upperBound(hi3 - bot2);
    FDomain newFD2 = fd2.copy();
    newFD2.lowerBound(bot3 - hi1);
    newFD2.upperBound(hi3 - bot1);
    FDomain newFD3 = fd3.copy();
    newFD3.lowerBound(bot1 + bot2);
    newFD3.upperBound(hi1 + hi2);
    ArrayList resFDs = new ArrayList(3);
    resFDs.add(newFD1);
    resFDs.add(newFD2);
    resFDs.add(newFD3);
    return resFDs;
  }

  // fd1 * n = fd2
  public static boolean multChangesDomain1(FDomain fd1, MyNumber n, 
                                           FDomain fd2) {
    try {
      double mult = n.value();
      if (mult == 0) return false;
      long bot1 = fd1.getLow();
      long hi1 = fd1.getHi();
      long bot2 = fd2.getLow();
      long hi2 = fd2.getHi();
      if (mult > 0) {
        return bot1 * mult > bot2 || hi1 * mult < hi2 || bot1 < bot2 / mult ||
               hi1 > hi2 / mult;
      } else {
        double minLeft = Math.min(bot1 * mult, hi1 * mult);
        double maxLeft = Math.max(bot1 * mult, hi1 * mult);
        double minRight = Math.min(bot2 / mult, hi2 / mult);
        double maxRight = Math.max(bot2 / mult, hi2 / mult);
        return minLeft > bot2 || maxLeft < hi2 || minRight > bot1 || 
               maxRight < hi1;
      }
    } catch (NoValueException e) {
      return false;
    }
  }

  // pre: n != 0
  // fd1 * n = fd2
  public static ArrayList multBoundDomain1(FDomain fd1, MyNumber n, 
                                           FDomain fd2) {
    FDomain newFD1 = fd1.copy();
    FDomain newFD2 = fd2.copy();
    double mult = n.value();
    long bot1 = fd1.getLow();
    long hi1 = fd1.getHi();
    long bot2 = fd2.getLow();
    long hi2 = fd2.getHi();
    if (mult > 0) {
      newFD1.lowerBound(bot2 / mult);
      newFD1.upperBound(hi2 / mult);
      newFD2.lowerBound(bot1 * mult);
      newFD2.upperBound(hi1 * mult);
    } else { // mult < 0
      double minLeft = Math.min(bot1 * mult, hi1 * mult);
      double maxLeft = Math.max(bot1 * mult, hi1 * mult);
      double minRight = Math.min(bot2 / mult, hi2 / mult);
      double maxRight = Math.max(bot2 / mult, hi2 / mult);
      newFD1.lowerBound(minRight);
      newFD1.upperBound(maxRight);
      newFD2.lowerBound(minLeft);
      newFD2.upperBound(maxLeft);
    }
    ArrayList resFDs = new ArrayList(2);
    resFDs.add(newFD1);
    resFDs.add(newFD2);
    return resFDs;
  }

  // fd1 * fd2 = n3
  public static boolean multChangesDomain2(FDomain fd1, FDomain fd2, 
                                           MyNumber n3) {
    try {
      long bot1 = fd1.getLow();
      long hi1 = fd1.getHi();
      long bot2 = fd2.getLow();
      long hi2 = fd2.getHi();
      double res = n3.value();
      if (bot1 >= 0 && bot2 >= 0 && res >= 0) {
        if (hi1 > 0 && bot2 < res / hi1) {
          return true;
        } else if (hi2 > 0 && bot1 < res / hi2) {
          return true;
        } else if (bot1 > 0 && hi2 > res / bot1) {
          return true;
        } else if (bot2 > 0 && hi1 > res / bot2) {
          return true;
        }
      }
      return false;
    } catch (NoValueException e) {
      return false;
    }
  }

  // fd1 * fd2 = n3
  public static ArrayList multBoundDomain2(FDomain fd1, FDomain fd2,
                                           MyNumber n3) {
    FDomain newFD1 = fd1.copy();
    FDomain newFD2 = fd2.copy();
    double res = n3.value();
    long bot1 = fd1.getLow();
    long hi1 = fd1.getHi();
    long bot2 = fd2.getLow();
    long hi2 = fd2.getHi();
    if (bot1 >= 0 && bot2 >= 0 && res >= 0) {
      if (hi2 > 0) {
        newFD1.lowerBound(res / hi2);
      }
      if (hi1 > 0) {
        newFD2.lowerBound(res / hi1);
      }
      if (bot2 > 0) {
        newFD1.upperBound(res / bot2);
      }
      if (bot1 > 0) {
        newFD2.upperBound(res / bot1);
      }
    }
    ArrayList resFDs = new ArrayList(2);
    resFDs.add(newFD1);
    resFDs.add(newFD2);
    return resFDs;
  }


  // fd1 * fd2 = fd3
  public static boolean multChangesDomain3(FDomain fd1, FDomain fd2, 
                                           FDomain fd3) {
    try {
      long bot1 = fd1.getLow();
      long hi1 = fd1.getHi();
      long bot2 = fd2.getLow();
      long hi2 = fd2.getHi();
      long bot3 = fd3.getLow();
      long hi3 = fd3.getHi();
      if (bot1 >= 0 && bot2 >= 0 && bot3 >= 0) {
        if (bot3 < bot1 * (double) bot2 || hi3 > hi1 * (double) hi2) {
          return true;
        } else if (hi2 > 0 && bot1 < bot3 / (double) hi2) {
          return true; 
        } else if (bot2 > 0 && hi1 > hi3 / (double) bot2) {
          return true; 
        } else if (hi1 > 0 && bot2 < bot3 / (double) hi1) {
          return true; 
        } else if (bot1 > 0 && hi2 > hi3 / (double) bot1) {
          return true; 
        }
      }
      return false;
    } catch (NoValueException e) {
      return false;
    }
  }

  // fd1 * fd2 = fd3
  public static ArrayList multBoundDomain3(FDomain fd1, FDomain fd2,
                                           FDomain fd3) {
    long bot1 = fd1.getLow();
    long hi1 = fd1.getHi();
    long bot2 = fd2.getLow();
    long hi2 = fd2.getHi();
    long bot3 = fd3.getLow();
    long hi3 = fd3.getHi();
    FDomain newFD1 = fd1.copy();
    FDomain newFD2 = fd2.copy();
    FDomain newFD3 = fd3.copy();
    newFD3.lowerBound(bot1 * (double) bot2);
    newFD3.upperBound(hi1 * (double) hi2);
    if (hi1 > 0) {
      newFD2.lowerBound(bot3 / (double) hi1);
    }
    if (hi2 > 0) {
      newFD1.lowerBound(bot3 / (double) hi2);
    }
    if (bot1 > 0) {
      newFD2.upperBound(hi3 / (double) bot1);
    }
    if (bot2 > 0) {
      newFD1.upperBound(hi3 / (double) bot2);
    }
    ArrayList resFDs = new ArrayList(3);
    resFDs.add(newFD1);
    resFDs.add(newFD2);
    resFDs.add(newFD3);
    return resFDs;
  }

  public String toString() {
    return ranges.toString();
  }

  public boolean equals(Object other) {
    if (other == null || !(other instanceof FDomain)) {
      return false;
    } else {
      return ranges.equals(((FDomain) other).ranges);
    }
  }

  public static boolean isSmall(FDomain fd) {
    return fd.size() < 50;
  }

  private static MyIntegerType wrap(long l, String type) {
     if (type.equals("byte")) {
        return new MyByte(l);
     } else if (type.equals("char")) {
        return new MyChar(l);
     } else if (type.equals("short")) {
        return new MyShort(l);
     } else if (type.equals("int")) {
        return new MyInteger(l);
     } else if (type.equals("long")) {
        return new MyLong(l);
     } else {
        throw new RuntimeException("Error: invalid ordinal type " + type);
     }
  }

  public static org.jmlspecs.models.JMLObjectSet boundSet(FDomain fd,
                                                          String type) {
    org.jmlspecs.models.JMLObjectSet res = 
       new org.jmlspecs.models.JMLObjectSet();
    Iterator it = fd.ranges.iterator();
    while (it.hasNext()) {
      Range r = (Range) it.next();
      for (long l = r.low; l <= r.hi; l++) {
        res = res.insert(wrap(l, type));     
      }
    }

    return res;
  }

  public static void main(String argv[]) {
    FDomain f1 = new FDomain();
    FDomain f2 = new FDomain();
    f1.ranges.add(new Range(-5, 10));
    f1.upperBound(1);
    f1.lowerBound(-1.1);
    System.out.println(f1);
/*
    f1.ranges.add(new Range(5, 10));
    f1.ranges.add(new Range(11, 13));
    f1.ranges.add(new Range(15, 20));
    f2.ranges.add(new Range(1, 3));
    f2.ranges.add(new Range(4, 7));
    f2.ranges.add(new Range(8, 12));
    f2.ranges.add(new Range(13, 21));
    FDomain f3 = intersect(f1, f2);
    System.out.println(f1);
    System.out.println(f2);
    System.out.println(f3);
*/
  }

}
