/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: AssertionMethod.java,v 1.9 2007/06/30 02:32:40 chalin Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * An abstract class for generating <em>assertion check methods</em>
 * for specific kinds of assertions such as preconditions, normal and
 * exceptional postconditions, invariants and (history) constraints.
 * The generated assertion check methods inherit the superclass's
 * assertions by dynamically invoking the superclass's corresponding
 * assertion check methods. For example, the precondition check method
 * of a method, <code>m</code>, invokes dynamically the
 * <code>m</code>'s precondition method of the superclass if the
 * method <code>m</code> is overriden in the subclass.  In general,
 * the following rules are applied for inheriting assertions.
 *
 * <p>
 * <ul><li>
 * Method specifications such as preconditions and postconditions are 
 * inherited by the subclass, but only for intance (non-static) methods.
 * That is, class (static) method specifications are not inherited by
 * subclasses. In addition, method specifications for constructors are not
 * inherited by subclasses.
 * </li></ul>
 *
 * <ul><li>
 * Both static and non-static invariants are inherited by subclasses.
 * </li></ul>
 *
 * <ul><li>
 * Both static and non-static constraints are inherited by subclasses.
 * However, the inheritance of constraints is subjected to the
 * <em>strong</em> and <em>weak</em> behavioral subtyping, specified by
 * the <code>weakly</code> annotation in the <code>extends</code> clause.
 * Not implemented yet!
 * </li></ul>
 *
 * All inheritances are subjected to the accessibility; i.e., only public
 * and protected assertions are inherited. Not implemented yet!
 * </p>
 *
 * <p>
 * The class is implemented with a variant of the <em>Template Pattern</em>
 * [GoF95].
 * The abstract method <code>generate</code> is supposed to call the
 * methods <code>buildHeader</code> and <code>checker</code>,
 * respectively, to compose the header and the body of the assertion method.
 * The method <code>checker</code> calls the query method 
 * <code>canInherit</code> to decide whether to generate code for inheriting
 * specifications or not. The actual code for inheritance is generated
 * by calling the method <code>inheritAssertion</code>. 
 *
 * <pre>
 *  generate {abstract}
 *    buildHeader
 *    checker
 *      canInherit {abstract}
 *      inheritAssertion
 * </pre>
 * </p>
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.9 $
 */

public abstract class AssertionMethod extends TransUtils {
    // ----------------------------------------------------------------------
    // METHOD GENERATION
    // ----------------------------------------------------------------------

    /** Generates and returns an assertion checking method. Append to
     * the body (<code>stmt</code>) code to check the inherited
     * assertions if any, and code to throw an appropriate exception
     * to notify an assertion violation if it happens at runtime.
     * 
     * @param stmt code to evaluate the assertions; the result is supposed
     *             to be stored in the variable <tt>VN_ASSERTION</tt>. 
     *             A <tt>null</tt> value means that no assertion is 
     *             specified or the assertion is not executable.
     */
    public abstract JMethodDeclarationType generate(RacNode stmt);


    /** Returnd the assertion check code (<code>stmt</code>) wrapped with
     * code to inherit superclass's assertions if any, and to signal
     * an assertion violation if it happens at runtime. The assertion
     * check code (<code>stmt</code>), if non-<code>null</code>,
     * evaluates all assertions specified in the class, conjoin the results, 
     * and store the final result to the variable <code>VN_ASSERTION</code>, 
     * called the <em>assertion variable</em>. The assertion variable
     * is supposed to be declared in the block that encloses the assertion
     * check code.
     *
     * @param initValue initial value of the assertion variable.
     * @param stmt assertion check code to wrap
     * @param params parameters to call the inherited check method if necessary.
     */
    protected String checker(boolean initValue, RacNode stmt, 
	JFormalParameter[] params)
    {
	// done if has no check code and nothing to inherit?
	if (stmt == null && !canInherit())
	    return "";

	// code for defining local variables, e.g., the assertion var
	StringBuffer code = new StringBuffer();
	code.append("  java.util.Set ");
	code.append(     VN_ERROR_SET);
	code.append(     " = new java.util.HashSet();\n");
	code.append("  boolean " + VN_ASSERTION + " = " + initValue + ";\n");

	// code for evaluating assertion, i.e., assertion check code
	if (stmt != null) {
	    code.append("$0\n");
	    stmt.incrIndent();
	}

	// code for evaluating inherited assertion
	if (canInherit())
	    code.append(inheritAssertion(params));

	// code for checking and signaling assertion violation
	code.append("  if (!" + VN_ASSERTION + ") {\n");
	code.append("    JMLChecker.exit();\n");
	code.append("    throw new " + exceptionToThrow + "(\"");
        code.append(       typeDecl.ident());
 	code.append(       "\", ");
 	code.append(       methodIdentification);
	code.append(       ", " + VN_ERROR_SET +");\n");
	code.append("  }\n");
	
	return code.toString();
    }

    /** Returnd the code to inherit the superclass's assertions, i.e.,
     * to call the superclass's assertion check method and
     * appropriately accumulate the result to the assertion variable
     * (e.g., conjunction).
     *
     * @param parameters parameters of the superclass's assertion check method
     */
    protected abstract String inheritAssertion(JFormalParameter[] parameters);


    /** Builds and returns a method header as a string.
     * 
     * @param returnType return type
     * @param name method name
     * @param parameters formal parameters
     * @param exceptions checked exceptions
     */
    protected StringBuffer buildHeader(String returnType, String name,
	JFormalParameter[] parameters, CClassType[] exceptions) 
    {
	final StringBuffer code = new StringBuffer();

        // heading if any
        code.append("\n");
        if (javadoc != null) {
            code.append(javadoc);
        } else {
            code.append("/** Generated by JML to check assertions. */");
        }
	code.append("\npublic ");

	// return type and name
	if (isStatic)
	    code.append("static ");
	code.append(returnType).append(" ").append(name);

	// foraml parameters
	code.append("(");
	if (parameters != null) {
	    for (int i = 0; i < parameters.length; i++) {
		if (i != 0)
		    code.append(", ");
		parameters[i].accept( new JmlAbstractVisitor() {
		    public void	visitJmlFormalParameter(
		        JmlFormalParameter self) {
			CType actualType = self.getType();
			String ident = self.ident();
			if( self.isFinal() ) {
			    code.append("final ");
			}
			// FRioux - Experimental test for efficiency
			// In order to use the inner class approach, the params
			// have to be final.
			else if(!Main.racOptions.oldSemantics()) {
				code.append("final ");
			}
			code.append(TransUtils.toString(actualType));
			code.append(" ");
			code.append(ident);
		    }
	       });
	    }
	}
	code.append(")");

	// checked-exceptions (throws list)
	if (exceptions != null) {
	    for (int i = 0; i < exceptions.length; i++) {
		code.append(i == 0 ? " throws " : ", ");
		code.append(exceptions[i].toString());
	    }
	}

	return code;
    }

    /** Indicates whether the generated assertion method should try 
     * to dynamically inherit the corresponding assertion method of 
     * the superclass. */
    protected abstract /*@ pure @*/ boolean canInherit();

    
    /**
     * Returns true if the type being translated has an explicitly
     * specified superclass.
     */
    protected /*@ pure @*/ boolean hasExplicitSuperClass() {
        return (typeDecl instanceof JmlClassDeclaration) &&
	    ((JmlClassDeclaration)typeDecl).superName() != null;
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** target type declaration for generating assertion methods */
    protected JmlTypeDeclaration typeDecl;

    /** Name prefix of the method to be inherited, e.g., MN_CHECK_PRE */
    protected String prefix; 

    /** static-ness of assertion method */
    protected boolean isStatic; 

    /** name of assertion method */
    protected /*@ non_null @*/ String methodName;
    //@ protected invariant methodName.length() > 0;

    /** name of exception class to be thrown if assertions are violated.
     */
    protected /*@ non_null @*/ String exceptionToThrow;
    //@ protected invariant exceptionToThrow.length() > 0;

    /** Javadoc comment to be added to the generated method. */
    protected /*@ nullable */ String javadoc;

    /** name of the target method for printing assertion violation error. */
    protected String methodIdentification;
}

