/*
 * Copyright (C) 2001-2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ConstraintMethod.java,v 1.12 2005/07/30 02:45:41 cheon Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import java.util.*;

/**
 * A class for generating assertion check methods for (history)
 * constraints.  There are two types of history constraints:
 * <em>instance</em> (<em>non-static</em>) <em>constraints</em> and
 * <em>class</em> (<em>static</em>) <em>constraints</em>.  As thus,
 * two types of constraint check methods are generated.  An instance
 * constraint method checks both the instance and class constraints
 * while a class constraint method checks only the class constraints.
 * The generated constraint check methods inherit its superclass's
 * constraints by dynamically invoking the corresponding assertion
 * methods.  The inheritance of history constrains is interpreted
 * differently between <em>strong</em> and <em>weak</em> behavioral
 * subtyping.  For strong behavioral subtyping, the suptype's history
 * constraints are inherited by all methods of the subtype, while for
 * weak behavior subtyping, the supertype's constraints are inhertied
 * only by overriding methods of the subtype.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 * @author Yoonsik Cheon
 * @version $Revision: 1.12 $
 */

public abstract class ConstraintMethod extends InvariantLikeMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new instance.
     *
     * @param isStatic the kind of constraint method to generate;
     *                 <code>true</code> for static and <code>false</code> 
     *                 for non-static (instance) constraint method.
     * @param typeDecl the target type declaration whose constraint methods
     *                 are to be generated.
     */
    protected ConstraintMethod(boolean isStatic, 
                               /*@ non_null @*/ JmlTypeDeclaration typeDecl,
                               /*@ non_null @*/ String postfix) {
	super(isStatic, typeDecl);
	this.exceptionToThrow = "JMLHistoryConstraintError";
        this.prefix = MN_CHECK_HC;
	this.methodName = prefix + postfix;

        // javadoc to be added to the generated constraint method
        javadoc = "/** Generated by JML to check " + 
            (isStatic ? "static" : "non-static") + " constraints of \n" + 
            ((typeDecl instanceof JmlClassDeclaration) ? 
             " * class " : " * interface ") +
            typeDecl.ident() + ". */";
    }

    /** Returns the formal parameters of the assertion check method. */
    protected /*@ non_null @*/ String formalParameters() {
        return "java.lang.String rac$msg, \n" +
            "  java.lang.String rac$name, java.lang.Class[] rac$params";
    }    

    /** Returns, in the string form, the formal parameter types of the
     * supertype's constraint check method. The returned value is used
     * to look up the supertype's constraint check method using the
     * Java's reflection facility.
     *
     * @see #reflectiveCall(boolean, CClass)
     * @see #nonReflectiveCall(boolean, CClass)
     */
    protected /*@ non_null @*/ String inheritCallArgTypes() {
        return "java.lang.String.class, " +
            "java.lang.String.class, java.lang.Class[].class";
    }

    /** Returns, in the string form, the arguments for calling
     * supertype's constraint check method. 
     *
     * @see #reflectiveCall(boolean, CClass)
     * @see #nonReflectiveCall(boolean, CClass)
     */
    protected /*@ non_null @*/ String inheritCallArgs() {
        return "rac$msg, rac$name, rac$params";
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------
}

