/* $Id: ConstructorWrapper.java,v 1.24 2006/12/11 05:59:03 chalin Exp $
 * 
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * A class for generating constructor wrappers.  A <em>constructor
 * wrapper</em> is a constructor that wraps a constructor with
 * assertion check methods such as pre and postcondition check
 * methods. The constructor under wrapping becomes a private method
 * and called by the wrapper.
 *
 * @see TransConstructor
 * @author Yoonsik Cheon
 * @version $Revision: 1.24 $
 */

public class ConstructorWrapper extends WrapperMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new instance.
     * 
     * <pre><jml>
     * requires typeName != null && mdecl != null;
     * </jml></pre>
     *
     * @param mdecl constructor to be wrapped
     */
    public ConstructorWrapper(String typeName, 
                              JmlConstructorDeclaration mdecl) {
	super(typeName, mdecl);
    }

    // ----------------------------------------------------------------------
    // TRANSLATION
    // ----------------------------------------------------------------------

    /** 
     * Returns a constructor wrapper for the given constructor
     * declaration. The constructor body is replaced by a new block of
     * code that forwards constructor calls to the orginal constructor
     * (which now becomes a private method) and performs appropriate
     * assertion checks (such as pre- and postoncodition checks)
     * before and after forwarding. The new body looks something like
     * the following:
     *
     * <pre>
     * C(T1 x1, ..., Tn xn) {
     *   checkPre(x1, ..., xn);
     *   try {
     *     internal(x1, ..., xn);
     *     checkPost(x1, ..., xn);
     *   }
     *   catch (JMLEntryPreconditionError e) {
     *     throw new JMLInternalPreconditionError(e); 
     *   }
     *   catch (JMLAssertionError e) {
     *     throw e;
     *   }
     *   catch (Throwable e) {
     *     checkXPost(x1, ..., xn, e);
     *   }
     *   finally {
     *     checkInv(x1, ..., xn);
     * }
     * </pre>
     *
     * <pre><jml>
     * also
     *   modifies methodDecl.*;
     *   accessible methodDecl;
     *   ensures \result != null;
     * </jml></pre>
     */
    public /*@ non_null @*/ JMethodDeclarationType generate() {
	// build a new constructor body with no explicit super call. 
	JStatement[] stmts = methodDecl.body().body();
	if (stmts != null && stmts.length > 0 && 
	    stmts[0] instanceof JEmptyStatement) { 
            // Remove an empty statement introduced by the typechecker;
            // the type checker replaces an explicit super call with 
            // a JEmptyStatement. This cleanup is not necessary but 
            // it looks nicer without the empty statement (;).
	    JStatement[] orig = stmts;
	    stmts = new JStatement[orig.length-1];
	    System.arraycopy(orig, 1, stmts, 0, orig.length-1);
	}
	JBlock body = new JBlock(org.multijava.util.compiler.TokenReference.NO_REF, stmts, null);

        JMethodDeclaration mdecl = null;
        if(JmlRacGenerator.checking_mode != JmlRacGenerator.WRAPPER) {	
            // create a private method with the modified constructor body
            mdecl = new JMethodDeclaration(
               methodDecl.getTokenReference(),
               ACC_PRIVATE,
	       CTypeVariable.EMPTY,
               CStdType.Void,
               MN_INTERNAL + MN_INIT,
               methodDecl.parameters(), 
               methodDecl.getExceptions(),
               body,
               null, // methodDecl.javadocComment(),
               null);
        } else {
            //With wrappers, we don't need the internal initializer
            //but we need to change the name of the method
            methodDecl.setIdent("$chx_Wrap_" + methodDecl.ident());
        }

	// turn the constructor as a wrapper method
	JExpression explicitSuper = 
	    ((JConstructorBlock)methodDecl.body()).explicitSuper();
	methodDecl.setBody(buildWrapperBody(explicitSuper));

        // return the newly created internal method
        if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
            //We don't need the internal method for wrappers.
            return RacParser.parseMethod("",null);
        } else {
            return RacParser.parseMethod(
                "\n/** Generated by JML to wrap assertion check code to\n" +
                " * the constructor of the same signature. */$0", mdecl);
        }
    }

    /** 
     * Returns a body for the constructor wrapper being built. Refers
     * to the method <code>generate()</code> for the structure of the
     * constructor body.
     *
     * <pre><jml>
     * assignable args, postArgs, xpostArgs;
     * accessible methodDecl;
     * ensures \result != null;
     * </jml></pre>
     *
     * @param explicitSuper explicit super/this call to be added
     *  as the first statement of the body to be built; 
     *  if <code>null</code>, no such call is added.
     *
     * @see #generate()
     * @see WrapperMethod#buildWrapperBody
     */
    private /*@ non_null @*/ JBlock buildWrapperBody(JExpression explicitSuper)
    {
	final StringBuffer code = 
	    new StringBuffer(explicitSuper == null ? "{\n" : "{$0\n");

        buildArguments();
        code.append("  " + VN_CONSTRUCTED + " = true;\n");
        code.append(   nocheckDuringInit());
        code.append("  if (!(" + isActive("PRECONDITION") + ")) {\n");
	code.append(     callOriginalMethod());
	code.append("    return;\n");
	code.append("  }\n");
	code.append(   checkPreInvariant());
	code.append(   checkPrecondition());
	code.append(   evalOldInConstraint());
	code.append("  boolean rac$ok = true;\n");
	code.append("  boolean rac$inv = true;\n");
	code.append("  try {\n");
	code.append(     callOriginalMethod());
	code.append(     checkPostcondition());
	code.append("  }\n");
	code.append("  catch (JMLEntryPreconditionError rac$e) {\n");
	code.append("    rac$ok = false;\n");
	code.append("    throw new JMLInternalPreconditionError(rac$e);\n");
	code.append("  }\n");
	code.append("  catch (JMLAssertionError rac$e) {\n");
	code.append("    rac$ok = false;\n");
	code.append("    throw rac$e;\n");
	code.append("  }\n");
	code.append("  catch (java.lang.Throwable rac$e) {\n");
	code.append("    rac$inv = false;\n");
	code.append("    try {\n");
	code.append(       checkXPostcondition());
	code.append("    }\n");
	code.append("    catch (JMLAssertionError rac$e1) {\n");
	code.append("      rac$ok = false;\n");
	code.append("      throw rac$e1;\n");
	code.append("    }\n");
	code.append("  }\n");
	code.append("  finally {\n");
	code.append("    if (rac$ok && rac$inv) {\n");
	code.append(       checkPostInstanceInvariant());
	code.append("    }\n");
	code.append("    if (rac$ok) {\n");
	code.append(       checkPostStaticInvariant());
	code.append(       checkConstraint());
	code.append("    }\n");
	code.append("  }\n");
	code.append("}");

	RacNode node = null;
	if (explicitSuper != null) 
	    node = RacParser.parseStatement("$0", explicitSuper).incrIndent();

	return RacParser.parseBlock(code.toString(), node); // supercall
    }

    /** Returns code for bypassing assertion checks during the class
     * and instance initialization. The returned code has the
     * following structure.
     *
     * <pre>
     *  if (!VN_CLASS_INIT || !VN_INIT) {
     *     T(x1, ..., xn);
     *     return;
     *  }
     * </pre>
     */
    protected /*@ non_null @*/ String nocheckDuringInit() {
        return 
            "  // skip assertion checking during initialization\n" +
            "  if (!" + VN_CLASS_INIT + " || !" + VN_INIT +") {\n" +
            callOriginalMethod() +
            "    return;\n" +
            "  }\n";
    }

    /** Returns code for checking precondition. The precondition is
     * always checked. The returned code has the following structure.
     *
     * <pre>
     *  // check precondition
     *  if (JMLChecker.isActive(JMLChecker.PRECONDITION)) {
     *    JMLChecker.enter();
     *    checkPre$$init$$T(x1,...,xn);
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>T</code> is the name of target type.
     *
     * @see WrapperMethod#checkPrecondition()
     */	    
    protected /*@ non_null @*/ String checkPrecondition() {
	return
	    "  // check precondition\n" +
	    "  if (JMLChecker.isActive(JMLChecker.PRECONDITION)) {\n" +
	    "    JMLChecker.enter();\n" + 
	    "    " + MN_CHECK_PRE + MN_INIT + "$" + typeName + args + ";\n" +
	    "    JMLChecker.exit();\n" +
	    "  }\n";
    }

    /** Returns code for checking invariant in the
     * pre-state. Constructors should preserve static invariants, and
     * have to establish non-static invariants. However, helpers are
     * exempted from the obligation. Thus, for helpers, an empty
     * string is returned, and for non-helpers, the following form of
     * code is returned.
     *
     * <pre>
     *  // check static invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT) && rac$constructed()) {
     *    JMLChecker.enter();
     *    checkInv$static();
     *    JMLChecker.exit();
     *  }
     * </pre>
     *
     * where <code>T</code> is the name of type being translated.
     */
    protected /*@ non_null @*/ String checkPreInvariant() {
        if (isHelper) {
            return "";
        }

	return wrap("INVARIANT", "check static invariant", 
                    "checkInv$static(\"" + identification("pre") + "\")");
    }

    /** Returns code for checking invariant in the
     * post-state. Constructors should preserve static invariants, and
     * have to establish non-static invariants. However, helpers are
     * exempted from the obligation. Thus, for helpers, an empty
     * string is returned, and for non-helpers, the following form of
     * code is returned.
     *
     * <pre>
     *  // check static invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT) && rac$constructed()) {
     *    JMLChecker.enter();
     *    checkInv$static();
     *    JMLChecker.exit();
     *  }
     *  // check instance invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT) && rac$constructed()) {
     *    JMLChecker.enter();
     *    checkInv$instance$T();
     *    JMLChecker.exit();
     *  }
     * </pre>
     *
     * where <code>T</code> is the name of type being translated.
     */
    protected /*@ non_null @*/ String checkPostInvariant() {
        if (isHelper) {
            return "";
        }

	String result = wrap3("INVARIANT", "check static invariant", 
                              "checkInv$static(\"" +
                              identification("post") + "\")");

        result += wrap3("INVARIANT", 
                        "check instance invariant",
                        "checkInv$instance$" + typeName + "(\"" +
                        identification("post") + "\")");
	return result;
    }

    /** Returns code for checking instance invariant in the
     * post-state. Constructors have to establish non-static
     * invariants. However, helpers are exempted from this
     * obligation. Thus, for helpers, an empty string is returned, and
     * for non-helpers, the following form of code is returned.
     *
     * <pre>
     *  // check instance invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT) && rac$constructed()) {
     *    JMLChecker.enter();
     *    checkInv$instance$T();
     *    JMLChecker.exit();
     *  }
     * </pre>
     *
     * where <code>T</code> is the name of type being translated.
     */
    protected /*@ non_null @*/ String checkPostInstanceInvariant() {
        if (isHelper) {
            return "";
        }
        return wrap3("INVARIANT", 
                     "check instance invariant",
                     "checkInv$instance$" + typeName + "(\"" +
                     identification("post") + "\")");
    }

    /** Returns code for checking static invariant in the
     * post-state. Constructors should preserve static
     * invariants. However, helpers are exempted from this
     * obligation. Thus, for helpers, an empty string isq returned, and
     * for non-helpers, the following form of code is returned.
     *
     * <pre>
     *  // check static invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT) && rac$constructed()) {
     *    JMLChecker.enter();
     *    checkInv$static();
     *    JMLChecker.exit();
     *  }
     * </pre>
     */
    protected /*@ non_null @*/ String checkPostStaticInvariant() {
        if (isHelper) {
            return "";
        }

	return wrap3("INVARIANT", "check static invariant", 
                     "checkInv$static(\"" +
                     identification("post") + "\")");
    }

    /** Returns code for evaluating old expressions in static
     * constraint. All non-helper constructors must preserve static
     * constraints, but not instance constraints. Thus, an empty
     * string is returned for helpers, and the following form of code
     * for non-helpers.
     *
     * <pre>
     *  // eval old exprs in static constraint
     *  if (JMLChecker.isActive(JMLChecker.CONSTRAINT)) {
     *    JMLChecker.enter();
     *    evalOldExprInHC$satic(); 
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>T</code> is the name of type being translated.
     */
    protected /*@ non_null @*/ String evalOldInConstraint() {
        if (isHelper) {
            return "";
        }

	return wrap("CONSTRAINT", "eval old exprs in constraint", 
                    evalOldName(typeName,true) + "()"); // true for static
    }

    /** Returns code for checking history constraints. All non-helper
     * constructors must preserve static constraints, but not instance
     * constraints. Thus, an empty string is returned for helpers, and
     * the following form of code for non-helpers.
     *
     * <pre>
     *  // check static constraint
     *  if (JMLChecker.isActive(JMLChecker.CONSTRAINT)) {
     *    JMLChecker.enter();
     *    checkHC$static();
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>T</code> is the name of type being translated. 
     */
    protected /*@ non_null @*/ String checkConstraint() {
        if (isHelper) {
            return "";
        }

        return wrap3("CONSTRAINT", "check static constraint", 
                     "checkHC$static(\"" +
                     identification("post") + "\", \"<init>\", " +
                     constraintCallArg()+ ")");
    }

    /** Returns code for invoking the now-renamed original constructor.
     *  I.e., <code>internal$m(x1,...,xn);</code>.
     */
    protected /*@ non_null @*/ String callOriginalMethod() {  

        if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
            return "    _wrapped_object = new $chx_Orig_" + typeName + args + 
                   ";\n" + "    _wrapped_object._chx_this = this;\n";
        } else {
            return "    " +  MN_INTERNAL + MN_INIT + args + ";\n";
        }
    }
    
    /** Returns code for checking normal postcondition. The
     * postcondition is checked only if the object is not being
     * constructed. The returned code has the following structure.
     *
     * <pre>
     *  // check normal postcondition
     *  if (JMLChecker.isActive(JMLChecker.POSTCONDITION) 
     *      && rac$constructed()) {
     *    JMLChecker.enter();
     *    checkPost$$init$$T(x1,...,xn);
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>T</code> is the name of the target type.
     */
    protected /*@ non_null @*/ String checkPostcondition() {
	return wrap2("POSTCONDITION",
		     "check normal postcondition",
                     MN_CHECK_POST + MN_INIT + "$" + typeName + postArgs);
    }

    /** Returns code for checking exceptional postcondition. The
     * exceptional postcondition is checked only if the object is not being
     * constructed. The returned
     * code has the following structure.
     *
     * <pre>
     *  // check exceptional postcondition
     *  if (JMLChecker.isActive(JMLChecker.POSTCONDITION)
     *      && rac$constructed()) {
     *    JMLChecker.enter();
     *    checkXPost$$init$T(x1,...,xn,e);
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>T</code> is the name of the target type.
     */
    protected /*@ non_null @*/ String checkXPostcondition() {
	return wrap3("POSTCONDITION", 
		     "check exceptional postcondition",
		     MN_CHECK_XPOST + MN_INIT +"$"+ typeName + xpostArgs);
    }

    /**
     * Returns a string representation of the condition that tests if
     * the given type of assertion is active. Assertion checking is
     * active only when the instance has completed its construction.
     */
    protected /*@ non_null @*/ String isActive(/*@ non_null @*/ String type) {
       if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
            String equiv_method = null;
            if(type == "PRECONDITION") {
                equiv_method = "Precondition";
            } else if(type == "POSTCONDITION") {
                equiv_method = "Postcondition";
            } else if(type == "INVARIANT") {
                equiv_method = "Invariant";
            }
            if(equiv_method != null) {
                return "$chx_Statics_" + typeName + ".access.isCheck" 
                        + equiv_method + "() && " + VN_CONSTRUCTED + "()";
            }
        }
        return "JMLChecker.isActive(JMLChecker." + type + ") && " +
            VN_CONSTRUCTED + "()";
    }

    /** Returns a string that identifies the method being translated. */
    protected /*@ non_null @*/ String identification(String state) {
        return "<init>@" + state + "<" + 
            escapeString(methodDecl.getTokenReference().toString()) + ">";
    }
}
