/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ExceptionalPostconditionMethod.java,v 1.9 2005/01/14 23:42:10 cheon Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * A class for generating exceptional postcondition check methods. 
 * The excpetional postcondition check code is wrapped with code 
 * to check the inherited exceptional postconditions if any, and 
 * to throw an appropriate exception to signal a violation
 * if the exceptional postcondition is violated at runtime.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.9 $
 */

public class ExceptionalPostconditionMethod extends PostconditionMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new <tt>ExceptionalPostconditionMethod</tt> object. 
     *
     *  @param mdecl method for which the checking method is generated
     */
    public ExceptionalPostconditionMethod(JmlTypeDeclaration typeDecl,
					  JmlMethodDeclaration mdecl,
                                          String restoreMethod)
    {
	super(typeDecl, mdecl, restoreMethod);
	exceptions = methodDecl.getExceptions();
	prefix = MN_CHECK_XPOST;
	methodName = prefix + methodName(mdecl) + "$" + typeDecl.ident();
	exceptionToThrow = "JMLExitExceptionalPostconditionError";

        // javadoc to be added to the generated method
        javadoc = "/** Generated by JML to check the exceptional " + 
            "postcondition of\n" + 
            " * method " + mdecl.ident() + ". */";
    }

    // ----------------------------------------------------------------------
    // GENERATION
    // ----------------------------------------------------------------------

    /** Returns an exceptional postcondition checking method.  This
     * method appends to the body, <code>stmt</code>, (1) code to
     * check the inherited exceptional postconditions if any, and (2)
     * code to throw an appropriate exception to notify a violation if
     * the condition is violated at runtime.
     * 
     * @param stmt code to evaluate the exceptional postcondition; 
     *             the result is supposed to be stored in the variable
     *             <code>VN_ASSERTION</code>. 
     *             A <code>null</code> value means that no exceptional
     *             postcondition is specified or it is not executable.
     */
    public JMethodDeclarationType generate(RacNode stmt)
    {
	JFormalParameter[] params = 
	    mkFormals(parameters, CStdType.Throwable, VN_EXCEPTION);
	StringBuffer code = 
	    buildHeader("void", methodName, params, exceptions);

	code.append(" {\n");
        // code for restoring pre-values for use in the post-state
        if (stackMethod != null) {
            code.append("  " + stackMethod + "();\n");
        }
	code.append(checker(true, stmt, params));

	// build exception rethrow statements to be executed if the 
	// exceptional condition be satisfied
	code.append("  JMLChecker.exit();\n");
	code.append("  if (" + VN_EXCEPTION);
	code.append(       " instanceof java.lang.RuntimeException) {\n");
	code.append("      throw (java.lang.RuntimeException) " + 
                               VN_EXCEPTION + ";\n");
	code.append("  }\n");
	code.append("  if (" + VN_EXCEPTION);
	code.append(       " instanceof java.lang.Error) {\n");
	code.append("      throw (java.lang.Error) " + VN_EXCEPTION + ";\n");
	code.append("  }\n");
	if (exceptions != null)
	    for (int i = 0; i < exceptions.length; i++) {
	      code.append("  if (");
	      code.append(   VN_EXCEPTION);
	      code.append(   " instanceof " + exceptions[i] + ") {\n");
	      code.append("      throw (" + exceptions[i] + ")");
	      code.append(         VN_EXCEPTION + ";\n");
	      code.append("  }\n");
	    }

	code.append("}\n");
	return RacParser.parseMethod(code.toString(), stmt);
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    protected CClassType[] exceptions;
}

