/* $Id: FinalizerWrapper.java,v 1.1 2002/11/07 05:10:56 cheon Exp $
 *
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * A class for generating wrapper methods for finalizers. For
 * finalizers, the post-state instance invariants and the post-state
 * instance constraints are not checked, but all other assertions are
 * checked in the same fashion as in regulrar methods.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.1 $
 */

public class FinalizerWrapper extends WrapperMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new instance.
     *
     * <pre><jml>
     * requires typeName != null && mdecl != null && mdecl.body() != null;
     * </jml></pre>
     * 
     * @param mdecl method to be wrapped; must not be abstract.
     */
    public FinalizerWrapper(String typeName, JmlMethodDeclaration mdecl) {
        super(typeName, mdecl);
    }

    /** Returns code for checking invariant in the post-state. If the
     * target method being translated is a helper method, an empty
     * string is returned; otherwise, the following form of code is
     * returned. 
     *
     * <pre>
     *  // check static invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT)) {
     *    JMLChecker.enter();
     *    checkInv$static();
     *    JMLChecker.exit();
     *  }
     * </pre>
     *
     * As shown by the return code, post-state instance invariants are
     * not checked for finalizers.
     */
    protected /*@ non_null @*/ String checkPostInvariant() {
        if (isHelper) {
            return "";
        }

        return wrap3("INVARIANT", "check static invariant", 
                     "checkInv$static(\"" + 
                     identification("post") + "\")");
    }

    /** Returns code for checking history constraint. If the target
     * method being translated is a helper method, an empty string is
     * returned; otherwise, the following form of code is returned.
     *
     * <pre>
     *  // check static constraint
     *  if (JMLChecker.isActive(JMLChecker.CONSTRAINT)) {
     *    JMLChecker.enter();
     *    checkHC$static();
     *    JMLChecker.exit();
     *  }
     * </pre>
     *
     * As shown by the return code, post-state instance invariants are
     * not checked for finalizers.
     */
    protected /*@ non_null @*/ String checkConstraint() {
        if (isHelper) {
            return "";
        }

        return wrap3("CONSTRAINT", "check static constraint", 
                     "checkHC$static(\"" +
                     identification("post") + "\", \"" +
                     ident + "\", " +constraintCallArg()+ ")");
    }
}

