/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: InvariantLikeMethod.java,v 1.18 2005/07/30 02:45:41 cheon Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * A class for generating assertion check methods for class-level
 * assertions such as invariants and history constraints.
 * There are two types of class-level assertions:
 * <em>instance</em> (<em>non-static</em>) <em>assertions</em> and
 * <em>class</em> (<em>static</em>) <em>assertions</em>.
 * As thus, two types of assertion check methods are generated. 
 * An instance assertion check method checks both the instance and class 
 * assertions while a class assertion check method checks only the class 
 * assertionss. 
 * The generated assertion check methods inherit assertions of its superclass
 * interfaces by dynamically invoking the corresponding assertion methods.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.18 $
 */

public abstract class InvariantLikeMethod extends AssertionMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new <code>InvariantLikeMethod</code> object. 
     *
     * @param isStatic the kind of assertion check method to generate;
     *                 <tt>true</tt> for static and <tt>false</tt> for 
     *                 non-static (instance) assertion check method.
     * @param typeDecl the target type declaration whose assertion check
     *                 methods are to be generated.
     */
    public InvariantLikeMethod(boolean isStatic, JmlTypeDeclaration typeDecl)
    { 
	this.isStatic = isStatic;
	this.typeDecl = typeDecl;
        
        // method id for assertion violation reporting comes from
        // argument named rac$msg (refer to method buildHeader and classes
        // WrapperMethod and ConstructorWrapper)
        this.methodIdentification = "rac$msg"; 
    }

    // ----------------------------------------------------------------------
    // GENERATION
    // ----------------------------------------------------------------------

    /** Generate and return a type-level assertion check method such
     * as invariants and history constraints.  Append to the body
     * (<code>stmt</code>): (1) code to check the inherited assertions
     * if any, and (2) code to throw an appropriate exception to
     * notify an assertion violation. 
     * 
     * @param stmt code to evaluate the assertions; the result is supposed
     *             to be stored in the variable <code>VN_ASSERTION</code>. 
     *             A <code>null</code> value means that no assertion is 
     *             specified or the assertion is not executable.
     */
    public JMethodDeclarationType generate(RacNode stmt)
    {
	StringBuffer code = buildHeader("void", methodName, null, null);
	code.append(" {\n");
	code.append(checker(true, stmt, null));
	code.append("}\n");

	return RacParser.parseMethod(code.toString(), stmt);
    }

    /** Builds and returns the method header of the assertion check
     * method as a string.
     * 
     * @param returnType return type
     * @param name method name
     * @param parameters formal parameters
     * @param exceptions checked exceptions
     */
    protected StringBuffer buildHeader(String returnType, String name,
	JFormalParameter[] parameters, CClassType[] exceptions) 
    {
	final StringBuffer code = new StringBuffer();

        // heading if any
        code.append("\n");
        if (javadoc != null) {
            code.append(javadoc);
        } else {
            code.append("/** Generated by JML to check assertions. */");
        }
	code.append("\npublic ");

	// return type and name
	if (isStatic)
	    code.append("static ");
	code.append(returnType).append(" ").append(name);

	code.append("(");
        code.append(formalParameters());
	code.append(")");

	return code;
    }

    /** Returns the formal parameters of the assertion check method. */
    protected /*@ non_null @*/ String formalParameters() {
        return "java.lang.String rac$msg";
    }    

    /** Return the code to inherit assertions of the superclass and
     * interfaces, i.e., to call the assertion check method of
     * the superclass and interfaces and appropriately accumulate 
     * the result to the assertion variable (e.g., conjunction). 
     *
     * @param parameters parameters to the inherited assertion check method;
     *                   not used in this specialization.
     */
    protected String inheritAssertion(JFormalParameter[] parameters) 
    {
	StringBuffer code = new StringBuffer();
	code.append("  boolean rac$bSuper = true;\n");

	// inherit the invariant of the superclass if exists
	if (hasExplicitSuperClass()) {
            code.append(inheritFrom(typeDecl.getCClass().getSuperClass()));
	}

	// inherit the invariants of the interfaces if any
	CClassType[] interfaces = typeDecl.interfaces();
	for (int i = 0; i < interfaces.length; i++) {
            code.append(inheritFrom(interfaces[i].getCClass()));
	}
	return code.toString();
    }

    /** 
     * Return the code that, if executed, inherits invariant-like
     * assertions such as invariants and constraints from the given
     * class, <code>clazz</code>. Because dynamic delegation is
     * expensive, the inherting code is short-circuited in that it is
     * executed only if the asssertion is not violated yet; i.e., not
     * executed if the assertion is already violated in the inheriting
     * class or in other classes whose assertions already inherited.
     *
     * @see #reflectiveCall(CClass)
     * @see #nonReflectiveCall(CClass)
     */
    protected /*@ non_null @*/ String inheritFrom(/*@ non_null @*/CClass clazz)
    {
        // no need to attempt, e.g., JDK classes?
        if (excludedFromInheritance(clazz)) {
            return "";
        }

	// build code for dynamic method invocation of the form:
	//   rac$check("C*", null, "prefix$static", null, null) or
	//   rac$check("C*", this, "prefix$instance$C", null, null)
	// where C* is either C or C$JmlSurrogate.

	StringBuffer code = new StringBuffer();
        code.append("  if (" + VN_ASSERTION + ") {\n");
	code.append("    rac$bSuper = true;\n");
	code.append("    try {\n");

        // call to the supertype's assertion method
        code.append(useReflection() ? reflectiveCall(clazz) :
                    nonReflectiveCall(clazz));

        // interpret the result of call
	code.append("    }\n");
	code.append("    catch (JMLAssertionError rac$e) {\n");
        code.append("      JMLChecker.enter();\n"); // nullify super's exit
	code.append("      rac$bSuper = false;\n");
	code.append("      " + VN_ERROR_SET + ".addAll(rac$e.locations());\n");
	code.append("    }\n");
	code.append("    " + VN_ASSERTION + " = " + VN_ASSERTION);
	code.append(       " && rac$bSuper;\n");
        code.append("  }\n"); // if (VN_ASSERTION)
        //             else { pop the old values stack of supertype; }
	return code.toString();
    }
        
    /** Return code that calls reflectively the assertion check method
     * of the given supertype, <code>claszz</code>.
     *
     * @see #inheritFrom(CClass)
     */
    protected String reflectiveCall(/*@ non_null @*/ final CClass clazz) {
        final String tab = "      ";
        return 
            tab+ "rac$check(\"" +dynamicCallName(clazz)+ "\", " +
                        (isStatic ? "null, \n" : "this, \n") + 
            tab+ "\"" +invariantLikeName(prefix, clazz, isStatic)+ "\",\n" +
            tab+ "  new java.lang.Class[] { " +inheritCallArgTypes()+ " },\n"+
            tab+ "  new java.lang.Object[] { " +inheritCallArgs()+ " });\n";
    }

    /** Return code that calls non-reflectively the assertion check method
     * of the given supertype, <code>claszz</code>.
     *
     * @see #inheritFrom(CClass)
     */
    protected String nonReflectiveCall(/*@ non_null @*/ CClass clazz) {
        String tab = "      ";
        StringBuffer code = new StringBuffer();

        // determine the receiver of the call. If clazz is a class, it
        // is "". If clazz is an interface, it is "T.TN_SURROGATE" for
        // a static method and "((T.TN_SURROGATE) surrogate)" for a
        // non-static method.
        String receiver = null;
        if (clazz.isInterface()) {
            receiver = clazz.qualifiedName().replace('/','.') + 
                "$" + TN_SURROGATE;
            if (!isStatic) {
                // use DOT-separate names for dynamic lookup of classes
                String target = "\"" + receiver + "\"";

                // find or create a surrogate
                String delegee = (typeDecl instanceof JmlClassDeclaration) ?
                    "this" : VN_DELEGEE;
                code.append(tab);
                code.append("Object surObj = " +delegee +
                            ".rac$getSurrogate(" +target+ ");\n");
                code.append(tab);
                code.append("if (surObj == null) {\n");
                code.append(tab);                
                code.append("  surObj = new " +receiver+ "(" +delegee+ ");\n");
                code.append(tab);                
                code.append("}\n");
                code.append(tab);
                code.append(delegee+ ".rac$setSurrogate(" +target+
                            ", (" +TN_JMLSURROGATE+ ") surObj);\n");
                receiver = "((" +receiver+ ") surObj)";
            }
        } else {
            if (isStatic) 
                receiver = clazz.qualifiedName().replace('/','.');
        }
        
        // method call statement, e.g., receiver.name(x1, ..., xn)
        code.append(tab);
        code.append(receiver == null ? "" : receiver + ".");
        code.append(invariantLikeName(prefix, clazz, isStatic));
        code.append("(" + inheritCallArgs() + ");\n");
        
        return code.toString();
    }

    /** Returns, in the string form, the parameter types of the
     * supertype's assertion check method. The returned value is used
     * to look up the supertype's assertion check method using the
     * Java's reflection facility. 
     *
     * @see #reflectiveCall(CClass)
     */
    protected /*@ non_null @*/ String inheritCallArgTypes() {
        return "java.lang.String.class";
    }

    /** Returns, in the string form, the arguments for calling
     * supertype's constraint check method. 
     *
     * @see #reflectiveCall(CClass)
     */
    protected /*@ non_null @*/ String inheritCallArgs() {
        return "rac$msg";
    }

    /** Indicates whether the generated assertion method should try 
     * to dynamically inherit the corresponding assertion method of 
     * the superclass or interfaces.
     *
     * <pre><jml>
     * also
     *   protected normal_behavior
     *     ensures \result == (!isStatic &&
     *      (hasExplicitSuperClass() || typeDecl.interfaces().length > 0));
     * </jml></pre>
     */
    protected /*@ pure @*/ boolean canInherit() {
	return !isStatic && 
            (hasExplicitSuperClass() || typeDecl.interfaces().length > 0);
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------
}

