/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: InvariantMethod.java,v 1.5 2002/10/27 20:49:10 leavens Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;

/**
 * A class for generating assertion check methods for invariants.
 * There are two types of invariants: 
 * <em>instance</em> (<em>non-static</em>) <em>invariants</em> and
 * <em>class</em> (<em>static</em>) <em>invariants</em>.
 * As thus, two types of invariant check methods are generated. 
 * An instance invariant method checks both the instance and class 
 * invariants while a class invariant method checks only the class 
 * invariants. 
 * The generated invariant methods inherit invariants of its superclass
 * interfaces by dynamically invoking the corresponding assertion methods.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 * @author Yoonsik Cheon
 * @version $Revision: 1.5 $
 */

public class InvariantMethod extends InvariantLikeMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new <code>InvariantMethod</code> object. 
     *
     * @param isStatic the kind of invariant method to generate;
     *                 <tt>true</tt> for static and <tt>false</tt> for 
     *                 non-static (instance) invariant method.
     * @param typeDecl the target type declaration whose invariant methods
     *                 are to be generated.
     */
    public InvariantMethod(boolean isStatic, JmlTypeDeclaration typeDecl)
    { 
	super(isStatic, typeDecl);

	exceptionToThrow = "JMLInvariantError";
	prefix = MN_CHECK_INV;
	if (isStatic) {
	    methodName = prefix + "static";
	} else {
	    methodName = prefix + "instance$" + typeDecl.ident();
	}

        // javadoc to be added to the generated invariant method
        javadoc = "/** Generated by JML to check " + 
            (isStatic ? "static" : "non-static") + " invariants of \n" + 
            ((typeDecl instanceof JmlClassDeclaration) ? 
             " * class " : " * interface ") +
            typeDecl.ident() + ". */";
    }

}

