/*
 * Copyright (C) 2001-2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: LocalConstraintMethod.java,v 1.3 2002/10/20 15:24:40 cheon Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * A class for generating constaint check methods that check locally
 * specified type constraints without inheriting any constraints from
 * supertypes. Two kinds of constraint check methods can be generated:
 * instance constraint check methods and static constraint check
 * methods. For details on constraint check methods, refer to the class
 * {@link ConstraintMethod}.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 * @see ConstraintMethod
 * @see #generate(RacNode)
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.3 $
 */

public class LocalConstraintMethod extends ConstraintMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new instance.
     *
     * @param isStatic the kind of constraint method to generate;
     * <code>true</code> for static and <code>false</code> for
     * non-static (instance) constraint method.
     * @param typeDecl the target type declaration whose constraint methods
     *                 are to be generated.
     * @param names method names that the generated constraint check
     * method is applicable to.
     * @param suffix a unique sequence number to be used as the suffix of the
     * constraint check method's name.
     */
    public LocalConstraintMethod(/*@ non_null @*/ VarGenerator varGen,
                                 boolean isStatic, 
                                 /*@ non_null @*/ JmlTypeDeclaration typeDecl,
                                 JmlMethodName[] names,
                                 int suffix) {
        super(isStatic, typeDecl, postfix(isStatic, typeDecl, suffix));
        this.names = names;
        this.varGen = varGen;
    }

    /** Returns the postfix to be used for constraint check method's
     * name. */
    public static String postfix(boolean isStatic, 
                                 JmlTypeDeclaration typeDecl, 
                                 int suffix) {
        return isStatic ? "static" + suffix :
            "instance" + suffix + "$" + typeDecl.ident();
    }

    /** Generates a constrain check method by using the given block of
     * code, <code>stmt</code>, as the main body. This method prefixes
     * the body with the code that tests if the constraint is
     * applicable to the argument method and appends the code to throw
     * a constraint violation error. Thus, the returned method has the
     * following structure.
     *
     * <pre>
     * public void checkHC$instance1$Constraint1(String rac$msg,
     *   String rac$name, java.lang.Class[] rac$params) {
     *   boolean b = false;
     *   b = ...; // test if applicable to method <rac$name,rac$params> 
     *   if (!b) {
     *     return;
     *   }
     *   stmt
     *   if (!constraint_hold) {
     *     throw new JMLHistoryConstraintError(...);
     *   }
     * }
     * </pre>
     * 
     * @param stmt code to evaluate the constraint; the result is supposed
     *             to be stored in the variable <code>VN_ASSERTION</code>. 
     *             A <code>null</code> value means that no assertion is 
     *             specified or the assertion is not executable.
     */
    public /*@ non_null @*/ JMethodDeclarationType generate(RacNode stmt)
    {
	StringBuffer code = buildHeader("void", methodName, null, null);
	code.append(" {\n");
        RacNode test = null;
        if (names != null && names.length > 0) {
            test = isApplicable(names).incrIndent();
            code.append("$1");
        }
	code.append(checker(true, stmt, null));
	code.append("}\n");

	return RacParser.parseMethod(code.toString(), stmt, test);
    }

    /**
     * Returns code that tests if the given JML method name,
     * <code>name</code>, is applicable to the method name given as a
     * parameter (<code>rac$name</code> and <code>rac$params</code>)
     * to the constraint check method. 
     */
    private RacNode isApplicable(JmlMethodName[] names) {
        //@ assume names != null && names.length > 0;
        String var = varGen.freshVar();
        String code = "boolean " + var + " = false;\n";
        for (int i = 0; i < names.length; i++) {
            code += isApplicable(names[i], var);
        }        
        code += "if (!" + var + ") {\n" +
            "  return;\n" +
            "}\n";
        return RacParser.parseStatement(code);
    }

    /**
     * Returns code that tests if the given JML method name,
     * <code>name</code>, is applicable to the method name given as a
     * parameter (<code>rac$name</code> and <code>rac$params</code>)
     * to the constraint check method. The code sets the given boolean
     * variable, <code>var</code> to the result.
     */
    private String isApplicable(JmlMethodName name, String var) {
        //@ assume name != null && var != null;
        return 
            "if (!" + var + ") {\n" +
              nameMatched(name, var) +  // check first the name part
              paramMatched(name, var) +  // check formal paramters
            "}\n";
    }

    /**
     * Returns code that compares the given JML method name,
     * <code>name</code>, with the method name given as a
     * parameter (<code>rac$name</code>) to the constraint check
     * method. The code sets the given boolean variable,
     * <code>var</code> to the result of the comparison.
     */
    private String nameMatched(JmlMethodName name, String var) {
        String sname = "";

        if (name instanceof JmlConstructorName) {
            sname = "<init>";
        } else {
            sname = simpleName(name.subnames());
        }

        if (sname != null) {
            return "  " + var + " = \"" +sname+ "\".equals(rac$name);\n";
        }
        return "";
    }

    /**
     * Returns code that compares the foraml parameter types of the
     * given JML method, <code>name</code>, with those of the
     * constraint check method. The code sets the given boolean
     * variable, <code>var</code> to the result of the comparison.
     */
    private String paramMatched(JmlMethodName name, String var) {

        // if parameters not specified, matches by default, e.g.,
        // constraint x > \old(x) for m; 
        final String tab = "  ";
        if (!name.hasParamDisambigList()) {
            return "";
        }
        
        CType[] types = name.paramDisambigList();
        String code = // has the same length?
            tab + "if (" +var+ " && " 
                         +types.length+ " != rac$params.length) {\n" +
            tab + "  " + var + " = false;\n" +
            tab + "}\n";

        // does each type match?
        for (int i = 0; i < types.length; i++) {
            code += tab + "if (" +var+ " && !rac$params[" +i+ "].equals(" +
                typeToClass(types[i]) + ")) {\n" +
                tab + "  " + var + " = false;\n" +
                tab + "}\n";
        }
        return code;
    }

    /**
     * Returns a simplied method name of the given JML names, that can
     * be compared with the method name given as a parameter of the
     * constraint check method. E.g., n for n or this.n, and null for
     * all other cases.
     */
    private String simpleName(/*@ non_null @*/ JmlName[] subnames) {
        String n = null;
	for (int i = 0; i < subnames.length; i++) {
	    //@ assert (* subnames[i] is ident, this, super, or \other *);
            if (subnames[i].isThis())
		continue;
	    if (subnames[i].isIdent()) {
                if (n != null)
                    return null; // can't handle names such as n1.n2
                else
                    n = subnames[i].ident();
            }
	    else 
                return null; // super or \other
	}
        return n;
    }
    
    /** Indicates whether the generated assertion method should try 
     * to dynamically inherit the corresponding assertion method of 
     * the superclass or interfaces.
     *
     * <pre><jml>
     * also
     *   protected normal_behavior
     *     ensures \result == false;
     * </jml></pre>
     */
    protected /*@ pure @*/ boolean canInherit() {
        return false;
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** The variable generator to be used in the translation. If
     * necessary, fresh variables are generated by this translator
     * using this variable generator.
     */
    private final /*@ spec_public non_null @*/ VarGenerator varGen;

    /** The names of methods to which the generated constraint check
     * method is applicable; it is null if the constraint is
     * applicable to all methods. */
    private JmlMethodName[] names;
}
