/*
 * Copyright (C) 2001-2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: MotherConstraintMethod.java,v 1.5 2004/01/22 19:59:34 leavens Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import java.util.*;

/**
 * A class for generating constraint check methods that check locally
 * specified type constraints and inherit constraints from
 * supertypes. The generated constraint methods are supposed to be
 * called by wrapper methods and constructors. This class can generate
 * both instance and static constraint check methods. For details on
 * constraint check methods, refer to the class 
 * {@link ConstraintMethod}.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 * @author Yoonsik Cheon
 * @version $Revision: 1.5 $
 */

public class MotherConstraintMethod extends ConstraintMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new instance.
     * 
     * <pre><jml>
     *  requires (\forall Object o; restoreMethods.contains(o);
     *    o instanceof String);
     * </jml></pre>
     *
     * @param isStatic the kind of constraint method to generate;
     *                 <code>true</code> for static and <code>false</code> 
     *                 for non-static (instance) constraint method.
     * @param typeDecl the target type declaration whose constraint methods
     *                 are to be generated.
     * @param restoreMethods the names of resore methods to be used to restore
     *                 the values of old expression fields.
     */
    public MotherConstraintMethod(boolean isStatic, 
                            /*@ non_null @*/ JmlTypeDeclaration typeDecl,
                            List restoreMethods)
    { 
        this(isStatic, typeDecl, postfix(isStatic, typeDecl), restoreMethods);
    }

    /** Construct a new instance.
     * 
     * <pre><jml>
     *  requires typeDecl != null;
     *  requires (\forall Object o; restoreMethods.contains(o);
     *    o instanceof String);
     * </jml></pre>
     *
     * @param isStatic the kind of constraint method to generate;
     *                 <code>true</code> for static and <code>false</code> 
     *                 for non-static (instance) constraint method.
     * @param typeDecl the target type declaration whose constraint methods
     *                 are to be generated.
     * @param restoreMethods the names of resore methods to be used to restore
     *                 the values of old expression fields.
     */
    protected MotherConstraintMethod(boolean isStatic, 
                                     JmlTypeDeclaration typeDecl,
                                     /*@ non_null @*/ String postfix,
                                     List restoreMethods) {
        super(isStatic, typeDecl, postfix);
        this.restoreMethods = 
            (restoreMethods == null) ? new ArrayList(0) : restoreMethods;
    }

    /** Returns the postfix to be used for constraint check method's
     * name. The return value is of the form:
     * <code>[static|instance$T]</code>.*/
    private static String postfix(boolean isStatic, 
                                 /*@ non_null @*/ JmlTypeDeclaration typeDecl) 
    {
        return isStatic ? "static" : "instance$" + typeDecl.ident();
    }

    /** Returns a constraint check method with the given body,
     * <code>stmt</code>. Appended to the body (<code>stmt</code>) are
     * (1) code to check the inherited assertions if any, and (2)
     * code to throw an appropriate exception to notify an assertion
     * violation.
     * 
     * @param stmt code to evaluate the assertions; the result is supposed
     *             to be stored in the variable <code>VN_ASSERTION</code>. 
     *             A <code>null</code> value means that no assertion is 
     *             specified or the assertion is not executable.
     */
    public JMethodDeclarationType generate(RacNode stmt)
    {
	StringBuffer code = buildHeader("void", methodName, null, null);
	code.append(" {\n");

        // code for restoring pre-values for use in the post-state
        if (restoreMethods != null) {
            for (Iterator i = restoreMethods.iterator(); i.hasNext(); ) {
                code.append("  " + i.next() + "();\n");
            }
        }

	code.append(checker(true, stmt, null));
	code.append("}\n");

	return RacParser.parseMethod(code.toString(), stmt);
    }

    /** Builds, in a string form, the header of constraint method. The 
     * header of a constraint method has the following form.
     *
     * <pre>
     *  public [static] void name(String rac$msg, String rac$name,
     *    String rac$params)
     * </pre>
     * 
     * @param returnType return type
     * @param name method name
     * @param parameters formal parameters
     * @param exceptions checked exceptions
     */
    protected StringBuffer buildHeader(String returnType, String name,
	JFormalParameter[] parameters, CClassType[] exceptions) 
    {
	final StringBuffer code = new StringBuffer();

        // heading if any
        code.append("\n");
        if (javadoc != null) {
            code.append(javadoc);
        } else {
            code.append("/** Generated by JML to check assertions. */");
        }
	code.append("\npublic ");

	// return type and name
	if (isStatic)
	    code.append("static ");
	code.append(returnType).append(" ").append(name);

        code.append("(java.lang.String rac$msg,\n  ");
        if (!isStatic) {
            code.append("boolean rac$private, ");
        }
        code.append("java.lang.String rac$name, java.lang.Class[] rac$params)");

	return code;
    }

    /** Returns the given constraint check code, <code>stmt</code>,
     * wrapped with code to inherit supertypes's constraints. For each
     * supertype, its strong or weak form of constraint check method is
     * called to inherit its constraints.
     *
     * @param initValue not used
     * @param stmt constraint check code to be wrapped
     * @param params not used
     */
    protected String checker(boolean initValue, RacNode stmt, 
                             JFormalParameter[] params) {
	StringBuffer code = new StringBuffer();

	// code for evaluating local constraints
	if (stmt != null) {
	    code.append("$0\n");
	    stmt.incrIndent();
	}
        
	// code for evaluating inherited assertion
	if (canInherit()) {
            code.append("  rac$name = rac$private ? null : rac$name;\n");
	    code.append(inheritAssertion(params));
        }
	
	return code.toString();
    }

    /** Return the code to inherit constraints of the superclass and
     * interfaces, i.e., to call the constraint methods of
     * the superclass and interfaces.
     *
     * @param parameters parameters to the inherited assertion check method;
     *                   not used in this specialization.
     */
    protected String inheritAssertion(JFormalParameter[] parameters) {
	StringBuffer code = new StringBuffer();

	// inherit from the superclass
	if (hasExplicitSuperClass()) {
            JmlClassDeclaration cdecl = (JmlClassDeclaration) typeDecl;
            code.append(inheritFrom(cdecl.isWeakSubtype(),
                                    cdecl.getCClass().getSuperClass()));
	}

	// inherit from the interfaces
	CClassType[] interfaces = typeDecl.interfaces();
        boolean[] weakly = typeDecl.interfaceWeaklyFlags();
	for (int i = 0; i < interfaces.length; i++) {
            code.append(inheritFrom(weakly[i],
                                    interfaces[i].getCClass()));
	}
	return code.toString();
    }

    /** 
     * Returns code that, if executed, inherits constraints from
     * the given type, <code>clazz</code>. 
     *
     * @see #reflectiveCall(CClass)
     * @see #nonReflectiveCall(CClass)
     */
    protected /*@ non_null @*/ String inheritFrom(boolean isWeaksubtype,
                                                  CClass clazz) {
        // no need to attempt, e.g., JDK classes?
        if (excludedFromInheritance(clazz)) {
            return "";
        }

        return useReflection() ? reflectiveCall(isWeaksubtype, clazz) :
            nonReflectiveCall(isWeaksubtype, clazz);
    }

    /** Return code, in the string form, that calls reflectively the
     * constraint check method of the given supertype,
     * <code>claszz</code>. If the argument, <code>isWeakly</code> is
     * true, the code calls the supertype's constraint method that
     * checks constraints according to the weak subtyping; otherwise,
     * called is the one that checks constraints according to the
     * strong form of subtyping.
     *
     * @see #inheritFrom(boolean, CClass)
     */
    protected String reflectiveCall(boolean isWeakly,
                                    /*@ non_null @*/ final CClass clazz) {
        final String tab = "  ";
        return 
            tab+ "rac$check(\"" +dynamicCallName(clazz)+ "\", this, \"" +
                 subtypeConstraintName(prefix, clazz, isWeakly)+ "\",\n" +
            tab+ "  new java.lang.Class[] { " +inheritCallArgTypes()+ " },\n"+
            tab+ "  new java.lang.Object[] { " +inheritCallArgs()+ " });\n";
    }

    /** Return code, in the string form, that calls non-reflectively
     * the constraint check method of the given supertype,
     * <code>claszz</code>. If the argument, <code>isWeakly</code> is
     * true, the code calls the supertype's constraint method that
     * checks constraints according to the weak subtyping; otherwise,
     * called is the one that checks constraints according to the
     * strong form of subtyping.
     *
     * @see #inheritFrom(boolean, CClass)
     */
    protected String nonReflectiveCall(boolean isWeakly,
                                       /*@ non_null @*/ final CClass clazz) {
        final String typeName = clazz.qualifiedName().replace('/','.');
        final String tab = "    ";
        final StringBuffer code = new StringBuffer();

        code.append("  // inheriting from type " + typeName + "\n");
        code.append("  {\n");

        // determine the receiver of the call. If clazz is a class, it
        // is "". If clazz is an interface, it is "T.TN_SURROGATE" for
        // a static method and "((T.TN_SURROGATE) surrogate)" for a
        // non-static method.
        String receiver = null;
        if (clazz.isInterface()) {
            receiver = typeName + "$" + TN_SURROGATE;
            if (!isStatic) {
                // use DOT-separate names for dynamic lookup of classes
                String target = "\"" + receiver + "\"";

                // find or create a surrogate
                String delegee = (typeDecl instanceof JmlClassDeclaration) ?
                    "this" : VN_DELEGEE;
                code.append(tab);
                code.append("Object surObj = " +delegee +
                            ".rac$getSurrogate(" +target+ ");\n");
                code.append(tab);
                code.append("if (surObj == null) {\n");
                code.append(tab);                
                code.append("  surObj = new " +receiver+ "(" +delegee+ ");\n");
                code.append(tab);                
                code.append("}\n");
                code.append(tab);
                code.append(delegee+ ".rac$setSurrogate(" +target+
                            ", (" +TN_JMLSURROGATE+ ") surObj);\n");
                receiver = "((" +receiver+ ") surObj)";
            }
        } else {
            if (isStatic) 
                receiver = typeName;
        }
        
        // method call statement, e.g., receiver.name(x1, ..., xn)
        code.append(tab);
        code.append(receiver == null ? "" : receiver + ".");
        code.append(subtypeConstraintName(prefix, clazz, isWeakly));
        code.append("(" + inheritCallArgs() + ");\n");
        code.append("  }\n");
        
        return code.toString();
    }

    /** Returns, in the string form, the formal parameter types of the
     * supertype's constraint check method. The returned value is used
     * to look up the supertype's constraint check method using the
     * Java's reflection facility.
     *
     * @see #reflectiveCall(boolean, CClass)
     * @see #nonReflectiveCall(boolean, CClass)
     */
    protected /*@ non_null @*/ String inheritCallArgTypes() {
        return "java.lang.String.class, " +
            "java.lang.String.class, java.lang.Class[].class";
    }

    /** Returns, in the string form, the arguments for calling
     * supertype's constraint check method. 
     *
     * @see #reflectiveCall(boolean, CClass)
     * @see #nonReflectiveCall(boolean, CClass)
     */
    protected /*@ non_null @*/ String inheritCallArgs() {
        return "rac$msg, rac$name, rac$params";
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** The names of the restoration methods that can restore the
     * correct values for old expressions appearing in the constraint
     * clause in the presence of recursive method calls. Each entry of
     * this list is a string.
     *
     * <pre><jml>
     * private invariant (\forall Object o; restoreMethods.contains(o);
     *   o instanceof String);
     * </jml></pre>
     */
    private List restoreMethods;
}

