/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: PostconditionMethod.java,v 1.18 2006/02/11 03:06:44 chalin Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;  
import org.multijava.util.compiler.TokenReference;

/**
 * A class for generating postcondition check methods. The postcondition 
 * check code is wrapped with code to check the inherited postconditions
 * if any, and to throw an appropriate exception to signal a violation
 * if the postcondition is violated at runtime.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.18 $
 */

public class PostconditionMethod extends PreOrPostconditionMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new <tt>PostconditionMethod</tt> object.
     * @param mdecl method for which the postcondition method is generated
     */
    public PostconditionMethod(JmlTypeDeclaration typeDecl,
			       JmlMethodDeclaration mdecl,
                               String restoreMethod)
    { 
	super(typeDecl, mdecl, restoreMethod);
	prefix = MN_CHECK_POST;
	methodName = prefix + methodName(mdecl) + "$" + typeDecl.ident();
	exceptionToThrow = "JMLExitNormalPostconditionError";
	returnType = mdecl.returnType();

        // javadoc to be added to the generated method
        javadoc = "/** Generated by JML to check the normal " + 
            "postcondition of\n" + 
            " * method " + mdecl.ident() + ". */";
    }

    // ----------------------------------------------------------------------
    // TRANSLATION
    // ----------------------------------------------------------------------

    /** Generate and return a postcondition checking method. Append to the 
     * body (<tt>stmt</tt>) (1) code to check the inherited postcondition
     * if any, and (2) code to throw an appropriate exception to notify a
     * violation if the postcondition is violated at runtime.
     * 
     * @param stmt code to evaluate the postcondition; the result is supposed
     *             to be stored in the variable <code>VN_ASSERTION</code>. 
     *             A <code>null</code> value means that no postcondition is 
     *             specified or it is not executable.
     */
    public JMethodDeclarationType generate(RacNode stmt)
    {
	JFormalParameter[] params = 
	    mkFormals(parameters, returnType, VN_RESULT);
	
	StringBuffer code = buildHeader("void", methodName, params, null);
	
	code.append(" {\n");
        // code for restoring pre-values for use in the post-state
        if (stackMethod != null) {
            code.append("  " + stackMethod + "();\n");
        }
	code.append(checker(true, stmt, params));
	code.append("}\n");

	return RacParser.parseMethod(code.toString(), stmt);
    }

    /** Builds new formal parameters by attaching a new formal
     * parameter <code>name</code> of type <code>type</code> to the
     * formal paramters <code>params</code>.
     *
     * @param params original formal parameters
     * @param type type of the new formal parameter to be added
     * @param name name of the new formal parameter to be added
     */
    protected static JFormalParameter[] mkFormals(JFormalParameter params[], 
                                                  CType type, String name) {
	int l = params == null ? 0 : params.length;
	JFormalParameter[] newParams = new JmlFormalParameter[l + 1];
	if (l > 0) {
	    System.arraycopy(params, 0, newParams, 0, l);
        }

        // if type is CStdType.Void, add a dummy parameter of type
        // Object at the end.
	newParams[l] = new JmlFormalParameter(TokenReference.NO_REF, 
	      0,
	      0,
	      new CSpecializedType(type.isVoid() ? CStdType.Object : type), 
	      name);
	return newParams;
    }

    /** 
     * Return the code to check inherited postconditions (from both
     * the superclass and superinterfaces). The generated code calls
     * the proper postcondition check methods of supertypes and
     * conjoins the results to the assertion variable. However,
     * because dynamic delegation is expensive, the inherting code is
     * short-circuited in that it is executed only if the
     * postcondition is not violated yet; i.e., not executed if the
     * the postcondition is already violated in the inheriting type
     * or in other types whose postconditions already inherited.
     *
     * @param parameters parameters to the precondition check methods
     */
    protected String inheritAssertion(JFormalParameter[] parameters) 
    {
	StringBuffer code = new StringBuffer();
        code.append("  if (" + VN_ASSERTION + ") {\n");
	code.append("    boolean rac$superPost = true;\n");

	CMethodSet methodSet = methodDecl.overriddenMethods();
	for (int i = 0; i < methodSet.size(); i++) {
	    CClass cls = methodSet.getMethod(i).owner();
            if (excludedFromInheritance(cls)) {
                continue;
            }
	    code.append("    try {\n");

	    String n = dynamicCallName(cls);
	    String m = prefix + name + "$" + cls.ident();

            // call to supertype's assertion check method
            code.append(useReflection() ?
                        reflectiveCall(cls, parameters, null) :
                        nonReflectiveCall(cls, parameters, null));
	    code.append("    }\n");
            // The name rac$e is reserved for use (as formal paramter
            // name of) in exceptional postcondition check method, so
            // use name such as rac$e1 here.
	    code.append("    catch (JMLAssertionError rac$e1) {\n" );
            code.append("      JMLChecker.enter();\n"); // nullify super's exit
	    code.append("      rac$superPost = false;\n");
	    code.append("      "+VN_ERROR_SET+".addAll(rac$e1.locations());\n");
	    code.append("    }\n");
	    code.append("    catch (java.lang.Throwable rac$e1) {\n" );
	    code.append("    }\n");
	    code.append("    " + VN_ASSERTION + " = " + VN_ASSERTION);
	    code.append(        " && rac$superPost;\n");
	}
        code.append("  }\n"); // if (VN_ASSERTION)
        //             else { pop the old values stack of supertype; }
	return code.toString();
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** Return type of the method declaration */
    protected CType returnType;
}

