/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: PreOrPostconditionMethod.java,v 1.13 2005/12/09 19:46:03 f_rioux Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * An abstract class for generating <em>precondition</em> or
 * <em>postconditin check methods</em> for the given methods.
 * The generated assertion methods inherit its superclass's
 * assertions by dynamically invoking the corresponding assertion methods. 
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.13 $
 */

public abstract class PreOrPostconditionMethod extends AssertionMethod
{
    /** Construct a new <code>PreOrPostconditionMethod</code> object. */
    protected PreOrPostconditionMethod(JmlTypeDeclaration typeDecl,
				       JmlMethodDeclaration mdecl,
                                       String stackMethod) { 
	this.typeDecl = typeDecl;
	methodDecl = mdecl;
        this.stackMethod = stackMethod;
	modifiers = mdecl.modifiers();
        name = mdecl.ident();
	methodIdentification = "\"" + name + "\"";
	parameters = mdecl.parameters();
	
	isStatic = mdecl.isStatic();
	exceptionToThrow = "JMLAssertionError";
    }

    // ----------------------------------------------------------------------
    // GENERATION
    // ----------------------------------------------------------------------

    /** Indicates whether the generated assertion method should try to
     * dynamically inherit the corresponding assertion method of the
     * superclass or interfaces.  Method specifications are inherited
     * only for instance (non-static) methods; i.e., class (static)
     * method specifications are not inherited by subclasses. Also,
     * method specifications for constructors are not inherited by
     * subclasses.  Dynamic inheritance is costly, so attempted only
     * if the method has a superclass (or interfaces) and is
     * overriding any of its superclass methods.
     *
     * <pre><jml>
     * also
     * protected normal_behavior
     *   ensures \result == !isStatic 
     *     && !(methodDecl instanceof JmlConstructorDeclaration)
     *     && methodDecl.isOverriding();
     * </jml></pre>
     */
    protected /*@ pure @*/ boolean canInherit() {
	return 
            // don't try to inherit from Object
            (hasExplicitSuperClass() || typeDecl.interfaces().length > 0) 
            && !isStatic
 	    && !(methodDecl instanceof JmlConstructorDeclaration)
 	    && methodDecl.isOverriding();
    }

    /**
     * Returns the name of assertion check method to be generated for
     * the given method declaration. For constructors, the result is 
     * MN_INIT; for non-constructors, the result is the same as the name
     * of the given method declaration.
     */    
    protected static String methodName(JmlMethodDeclaration mdecl) {
        return (mdecl instanceof JmlConstructorDeclaration) ? 
            MN_INIT : mdecl.ident();
    }

    /** Return code that calls reflectively the pre or postcondition
     * method of the given supertype, <code>claszz</code>.
     */
    protected String reflectiveCall(/*@ non_null @*/ CClass clazz,
                                  JFormalParameter[] parameters,
                                  /*@ non_null @*/ String resultVar) {
	// params to call the precond methods using reflection
	final String self = isStatic ? "null" : "this";
        String m = prefix + name + "$" + clazz.ident();
        String lexpr = 
            resultVar == null ? "" : resultVar + " = ";
        String tab = resultVar == null ? "      " : "    ";
        StringPair p = reflectiveCallArguments(parameters);
        return 
            tab + "java.lang.Class[] rac$types = " + p.first + ";\n" +
            tab + "java.lang.Object[] rac$args = " + p.second + ";\n" +
            tab + lexpr + "rac$check(\"" +
                 dynamicCallName(clazz) +
                 "\", " + self + ", \"" +
                 m + "\", rac$types, rac$args);\n";
    }

    /** Returns a pair of strings suitable for arguments to invoke
     * dynamically a method with the given formal parameter,
     * <code>parameters</code>. The first of the pair is of the form,
     * <code>"new java.lang.Class[] { T1.class, ..., Tn.class }"<code>,
     * and the second of the form
     * <code>"new java.lang.Object[] { x1, ..., xn }"</code>.
     */
    private static /*@ non_null @*/ StringPair reflectiveCallArguments(
        /*@ non_null @*/ JFormalParameter[] parameters) {
        String tcode = "new java.lang.Class[] { ";
        String acode = "new java.lang.Object[] { ";
        for (int i = 0; i < parameters.length; i++) {
            CType type = parameters[i].getType();
            tcode += typeToClass(type) + ", ";
            acode += wrappedValue(type, parameters[i].ident()) + ", ";
        }
        tcode += "}";
        acode += "}";

	return new StringPair(tcode, acode);
    }

    /** A record class that can store a pair of strings. */
    private static class StringPair {
	public StringPair(String fst, String snd) {
	    first = fst;
	    second = snd;
	}
	public String first;
	public String second;
    }

    /** Return code that calls non-reflectively the pre or postcondition
     * method of the given supertype, <code>claszz</code>.
     */
    protected String nonReflectiveCall(/*@ non_null @*/ CClass clazz,
                                     JFormalParameter[] parameters,
                                     /*@ non_null @*/ String resultVar) {
        String tab = resultVar == null ? "      " : "    ";
        StringBuffer code = new StringBuffer();

        // determine the receiver of the call. If clazz is a class, it
        // is "". If clazz is an interface, it is "T.TN_SURROGATE" for
        // a static method and "((T.TN_SURROGATE) surrogate)" for a
        // non-static method.
        String receiver = null;
        if (clazz.isInterface()) {
            receiver = clazz.qualifiedName().replace('/','.') + 
                "$" + TN_SURROGATE; // .
            if (!methodDecl.isStatic()) {
                // use DOT-separate names for dynamic lookup of classes
                String target = "\"" + receiver + "\"";

                // find or create a surrogate
                String delegee = (typeDecl instanceof JmlClassDeclaration) ?
                    "this" : VN_DELEGEE;
                code.append(tab);
                code.append("Object surObj = " +delegee +
                            ".rac$getSurrogate(" +target+ ");\n");
                code.append(tab);
                code.append("if (surObj == null) {\n");
                code.append(tab);                
                code.append("  surObj = new " +receiver+ "(" +delegee+ ");\n");
                code.append(tab);                
                code.append("}\n");
                code.append(tab);
                code.append(delegee+ ".rac$setSurrogate(" +target+
                            ", (" +TN_JMLSURROGATE+ ") surObj);\n");
                receiver = "((" +receiver+ ") surObj)";
            }
        }
        
        // method call statement, e.g., receiver.name(x1, ..., xn)
        code.append(tab);
        if (resultVar != null)
            code.append(resultVar + " = ");
        code.append(receiver == null ? "" : receiver + ".");
        code.append(prefix + name + "$" + clazz.ident()); // method name
        code.append(nonreflectiveCallArguments(parameters));
        code.append(";\n");
        
        return code.toString();
    }

    /** Return code that represents the arguments corresponding to
     * the given formal parameters, <code>parameters</code>. */
    private static /*@ non_null @*/ String nonreflectiveCallArguments(
        /*@ non_null @*/ JFormalParameter[] parameters) {
	StringBuffer args = new StringBuffer("(");
	for (int i = 0; i < parameters.length; i++) {
	    if (i != 0)
		args.append(", ");
	    args.append(parameters[i].ident());
	}
	args.append(")");
        return args.toString();
    }


    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** target method for which assertion method is generated */
    protected final JmlMethodDeclaration methodDecl;
    
    /** modifiers of the target method */
    protected final long modifiers;

    /** parameters of the target method */
    protected final JFormalParameter[] parameters;

    protected String stackMethod;

    /** name of the target method */
    protected String name;
}

