/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: PreconditionMethod.java,v 1.13 2007/07/19 10:51:36 f_rioux Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import java.util.List;

/**
 * A class for generating precondition check methods. The precondition 
 * check code is wrapped with code to check the inherited preconditions 
 * if any, and to throw an appropriate exception to signal a violation
 * if the precondition is violated.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.13 $
 */

public class PreconditionMethod extends PreOrPostconditionMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new <tt>PreconditionMethod</tt> object. 
     *
     * @param mdecl method for which the precondition method is generated
     * @param hasAssertion <code>true</code> if <code>mdecl</code> has 
     *                     a precondition specification
     */
    public PreconditionMethod(JmlTypeDeclaration typeDecl,
			      JmlMethodDeclaration mdecl,
			      boolean hasAssertion,
                              String saveMethod)
    {
	super(typeDecl, mdecl, saveMethod);
	
	this.hasAssertion = hasAssertion;
	prefix = MN_CHECK_PRE;
	methodName = prefix + methodName(mdecl) + "$" + typeDecl.ident();
	exceptionToThrow = "JMLEntryPreconditionError";

        // javadoc to be added to the generated method
        javadoc = "/** Generated by JML to check the precondition of\n" + 
             " * method " + mdecl.ident() + ". */";
    }

    // ----------------------------------------------------------------------
    // GENERATION
    // ----------------------------------------------------------------------

    public JMethodDeclarationType generate(RacNode stmt) {
	throw new UnsupportedOperationException();
    }

    /** Generates and returns a precondition checking method. Append
     * to the body (<code>stmt</code>) (1) code to check the inherited
     * precondition if any, and (2) code to throw an appropriate
     * exception to notify a violation if the precondition is violated.
     * 
     * @param stmt code to evaluate the precondition; the result is supposed
     *             to be stored in the variable <code>VN_ASSERTION</code>. 
     *             A <code>null</code> value means that no precondition is 
     *             specified or the precondition is not executable.
     * @param preExprs a list of <code>RacNode</code>s denoting expressions
     *             to be executed in the pre-sate, before the precondition.
     * @param oldExprs a list of <code>RacNode</code>s denoting expressions
     *             to be executed in the pre-sate, after the precondition.
     */
    public JMethodDeclarationType generate(RacNode stmt, List preExprs, List oldExprs)
    {
	StringBuffer code = 
	    buildHeader( "boolean", methodName, parameters, null);
	code.append(" {\n");

	// add pre-state expressions, e.g., spec vars and old exprs
	RacNode node1 = processList(preExprs);
	code.append(node1 != null ? "$1" : "");

	// add pre-state expressions, e.g., spec vars and old exprs
	RacNode node2 = processList(oldExprs);
	code.append(checker(false, stmt, null, node2 != null ? "$1" : ""));
	code.append("}\n");

	List allExprs = new java.util.ArrayList();
	allExprs.addAll(preExprs);
	allExprs.addAll(oldExprs);
	RacNode node3 = processList(allExprs);

	return RacParser.parseMethod(code.toString(), stmt, node3);
    }

	private /*@nullable*/ RacNode processList(/*@nullable*/  List list){
		RacNode node = null;
		if (list != null && list.size() > 0) {
		    String str = "";
		    for (int i = 0; i < list.size(); i++) {
			str = str + "$" + i + "\n";
		    }
		    node = RacParser.parseStatement(str, list.toArray());
		    node.incrIndent();
		}
		return node;
	}

    /** Returns the precondition check code (<code>stmt</code>)
     * wrapped with code to inherit superclass's preconditions if any,
     * and to signal an assertion violation if it happens at
     * runtime. The precondition check code (<code>stmt</code>), if
     * non-<code>null</code>, evaluates all preconditions specified in
     * the class, disconjoin the results, and store the final result
     * to the variable <code>VN_ASSERTION</code>, called the
     * <em>assertion variable</em>.  The assertion variable is
     * supposed to be declared in the block that encloses the
     * precondition check code.
     *
     * @param start initial value of the assertion variable.
     * @param stmt precondition check code to wrap
     * @param params parameters to call the inherited check method if necessary.
     * @param oldExprEvaluation evaluation of old expression 
     */
    protected String checker(boolean start, RacNode stmt,
	JFormalParameter[] params, String oldExprEvaluation)
    {
	// done if no precondition and nothing to inherit?
	if (stmt == null && !canInherit()) {
		return oldExprEvaluation + preValueSaveCode() + "  return false;\n";
	}

	// code for local variables
	StringBuffer code = new StringBuffer();
	code.append("  java.util.Set ");
	code.append(     VN_ERROR_SET);
	code.append(     " = new java.util.HashSet();\n");
	code.append("  boolean " + VN_ASSERTION + " = false;\n");

	// code for evaluating precondition
	if (stmt != null) {
	    code.append("$0\n");
	    stmt.incrIndent();
	}

	code.append(oldExprEvaluation);

	// code for evaluating inherited preconditions
	if (canInherit())
	    code.append(inheritAssertion(parameters));

	// code for checking precondition
	code.append("  if (!" + VN_ASSERTION + ") {\n");

	// we still need to preserve pre-values at this point due to
	// multiple inheritance, i.e., the precondition may be
	// satisfied by those inherited from others, and the
	// postcondition corresponding to this one can still be
	// called by subclasses.
	code.append(     preValueSaveCode2());

	code.append("    JMLChecker.exit();\n");
	code.append("    throw new " + exceptionToThrow + "(\"");
	code.append(       typeDecl.ident());
	code.append(       "\", \"");
	code.append(       name);
	code.append(       "\", " + VN_ERROR_SET + ");\n");
	code.append("  }\n");

	code.append(preValueSaveCode()); // save pre-values to stack

	// code for returning result
	if (hasAssertion) 
	    code.append("  return true;\n");
	else 
	    code.append("  return rac$hasPre;\n");

	return code.toString(); 
    } 

    /** Returns code for saving pre-values to internal stacks to be
     * fetched by the post-state assertion check method such as the
     * postcondition method in the post-state. 
     *
     * <pre><jml>
     *   requires stackMethod == null;
     *   ensures \result.equals("");
     * also
     *   requires stackMethod != null;
     *   ensures (* \result is stack save code *);
     * </jml></pre>
     */
    private /*@ non_null @*/ String preValueSaveCode() {
        return (stackMethod == null) ? "" :
            "  if (JMLChecker.getLevel() > JMLChecker.PRECONDITION) {\n" +
            "    " + stackMethod + "();\n" +
            "  }\n";
    }

    /** Returns code for saving pre-values to internal stacks to be
     * fetched by the post-state assertion check method such as the
     * postcondition method in the post-state. The returned code has
     * the same structure of that of the method
     * <code>preValueSaveCode</code>, but indented to the two tab
     * position. 
     *
     * @see #preValueSaveCode()
     */
    private /*@ non_null @*/ String preValueSaveCode2() {
        return (stackMethod == null) ? "" :
            "    if (JMLChecker.getLevel() > JMLChecker.PRECONDITION) {\n" +
            "      " + stackMethod + "();\n" +
            "    }\n";
    }

    /** 
     * Return the code to check inherited preconditions (from both the
     * superclass and interfaces), i.e., to call the proper precondition
     * check methods and appropriately disjoin the result to 
     * the assertion variable.
     *
     * @param parameters parameters to the precondition check methods
     */
    protected String inheritAssertion(JFormalParameter[] parameters) 
    {
	StringBuffer code = new StringBuffer();
	code.append("  boolean rac$hasPre = " + hasAssertion + ";\n");
	code.append("  boolean rac$hasSuperPre = true;\n");
	code.append("  boolean rac$superPre = true;\n");

	CMethodSet methodSet = methodDecl.overriddenMethods();
	for (int i = 0; i < methodSet.size(); i++) {

	    CClass cls = methodSet.getMethod(i).owner();
            if (excludedFromInheritance(cls)) {
                continue;
            }

	    if (i != 0)
		code.append("  rac$superPre = true;\n");

	    code.append("  try {\n");

            // call to supertype's assertion method
            code.append(useReflection() ?
                        reflectiveCall(cls, parameters, "rac$hasSuperPre") :
                        nonReflectiveCall(cls, parameters, "rac$hasSuperPre"));

	    code.append("  }\n");
	    code.append("  catch (JMLAssertionError rac$e) {\n" );
            code.append("    JMLChecker.enter();\n"); // nullify super's exit
	    code.append("    rac$superPre = false;\n");
	    code.append("    "+VN_ERROR_SET+".addAll(rac$e.locations());\n");
	    code.append("  }\n");
	    code.append("  if (rac$hasSuperPre) {\n");
	    code.append("    " + VN_ASSERTION + " = " + VN_ASSERTION);
	    code.append(        " || rac$superPre;\n");
	    code.append("  }\n");
	    code.append("  rac$hasPre = rac$hasPre || rac$hasSuperPre;\n");
	}
	code.append("  if (!rac$hasPre) {\n");
	code.append("     " + VN_ASSERTION + " = true;\n");
	code.append("  }\n");

	return code.toString();
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    private boolean hasAssertion;
}
