/*
 * Copyright (C) 2001-2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: SubtypeConstraintMethod.java,v 1.4 2004/01/22 19:59:34 leavens Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import java.util.*;

/**
 * A class for generating constaint check methods that check locally
 * specified type constraints and inherit constraints from
 * supertypes. The generated constraint methods are supposed to be
 * called by subtypes. Thus, this class can generate two kinds of
 * constraint check methods, for weak subtypes and for strong
 * subtypes. Both kinds of constraint methods are generated only for
 * instance constraints. For details on constraint check methods,
 * refer to the class {@link ConstraintMethod}.
 *
 * <p>
 * The class implements a variant of the <em>Template Pattern</em>
 * [GoF95], prescribed in the class {@link AssertionMethod}.
 * </p>
 *
 * @see AssertionMethod
 * @see ConstraintMethod
 * @see #generate(RacNode)
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.4 $
 */

public class SubtypeConstraintMethod extends MotherConstraintMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new instance.
     *
     * <pre><jml>
     *  requires (\forall Object o; restoreMethods.contains(o);
     *    o instanceof String);
     * </jml></pre>
     *
     * @param tdecl the target type declaration whose constraint methods
     *                 are to be generated.
     * @param forWeakSubtype true if the constraint method is for weak
     * subtypes; false if it is for strong subtypes.
     * @param restoreMethods the names of resore methods to be used to restore
     *                 the values of old expression fields.
     */
    public SubtypeConstraintMethod(/*@ non_null @*/ JmlTypeDeclaration tdecl,
                                   boolean forWeakSubtype,
                                   List restoreMethods) {
        super(false, tdecl, postfix(tdecl, forWeakSubtype), restoreMethods);
        this.forWeakSubtype = forWeakSubtype;
    }

    /** Returns the postfix to be used for constraint check method's
     * name.  The return value is of the form:
     * <code>instance[W|S]$T</code>.*/
    public static String postfix(/*@ non_null @*/ JmlTypeDeclaration tdecl,
                                 boolean forWeakSubtype) {
        return postfix(tdecl.getCClass(), forWeakSubtype);
    }

    /** Returns the postfix to be used for constraint check method's
     * name. The return value is of the form:
     * <code>instance[W|S]$T</code>.
     */
    public static String postfix(/*@ non_null @*/ CClass clazz, 
                                 boolean forWeakSubtype) {
        return "instance" + (forWeakSubtype ? "W" : "S") + "$" + clazz.ident();
    }

    /** Returns the local constraint check code (<code>stmt</code>)
     * wrapped with code to inherit supertypes's constraints.
     *
     * @param initValue initial value of the assertion variable.
     * @param stmt assertion check code to wrap
     * @param params parameters to call the inherited check method if necessary.
     */
    protected String checker(boolean initValue, RacNode stmt, 
                             JFormalParameter[] params) {
	StringBuffer code = new StringBuffer();

	// code to evaluate local constraints
	if (stmt != null) {
            stmt.incrIndent();
            
            if (forWeakSubtype) {
                // for weak subtype, constraints are checked only for
                // inherited methods
                String name = typeDecl.getCClass().
                    qualifiedName().replace('/','.');
                code.append("  java.lang.Class rac$cls = null;\n");
                code.append("  try {\n");
                code.append("    rac$cls = rac$forName(\"" + 
                                   name + "\");\n");
                code.append("  } catch (java.lang.Exception rac$e) {\n");
                code.append("  }\n");
                code.append("  if (JMLChecker.inheritedFrom(rac$cls, " + 
                                   "rac$name, rac$params)) {\n");
                code.append(     "$0\n"); 
                code.append("  }\n");
                stmt.incrIndent();
            } else {

                // strong subtype, constraint are always checked
                code.append("$0\n"); 
            }
	}
        
	// code to call supertype's constraint methods
	if (canInherit()) {
	    code.append(inheritAssertion(params));
        }
	
	return code.toString();
    }

    /** Builds, in a string form, the header of constraint method. The 
     * header of a constraint method has the following form.
     *
     * <pre>
     *  public [static] void name(String rac$msg, String rac$name,
     *    String rac$params)
     * </pre>
     * 
     * @param returnType return type
     * @param name method name
     * @param parameters formal parameters
     * @param exceptions checked exceptions
     */
    protected StringBuffer buildHeader(String returnType, String name,
	JFormalParameter[] parameters, CClassType[] exceptions) 
    {
	final StringBuffer code = new StringBuffer();

        // add javadoc comment if any
        code.append("\n");
        if (javadoc != null) {
            code.append(javadoc);
        } else {
            code.append("/** Generated by JML to check assertions. */");
        }
	code.append("\npublic ");

	// add return type and name
	if (isStatic)
	    code.append("static ");
	code.append(returnType).append(" ").append(name);

        code.append("(java.lang.String rac$msg,\n");
        code.append("  java.lang.String rac$name, java.lang.Class[] rac$params)");

	return code;
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * Indicates whether the constraint method to be generated should
     * be called by weak subtypes or strong subtypes.
     */
    private final boolean forWeakSubtype;
}
