/*
 * Copyright (C) 2001-2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: TransClass.java,v 1.77 2007/02/03 02:04:49 delara Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import org.multijava.util.compiler.*;
import java.util.*;

/**
 * A class for translating JML class declarations. The translation 
 * produces a set of assertion check methods for the class and any 
 * necessary fields, e.g., to hold old values. This class implements
 * the <em>Template Method Pattern</em> layed out by the abstract
 * superclass {@link TransType}.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.77 $
 */

public class TransClass extends TransType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a <code>TransClass</code> object.
     *
     * @param classDecl target class declaration to be translated
     */
    public TransClass(JmlClassDeclaration classDecl) {
        super(classDecl);
	this.classDecl = classDecl;
    }

    // ----------------------------------------------------------------------
    // TRANSLATION
    // ----------------------------------------------------------------------

    /**
     * Translates a JML field declaration, <code>fieldDecl</code>, e.g.,
     * handling final, model, spec_public, and spec_protected.
     *
     * <p>
     * If this is a model field and it has no accessor method generated by
     * <code>represents</code> clauses, generate the following form of
     * default model accessor method:
     * 
     * <pre>
     * [[... model T m ...;]] ==
     *   protected T m() {
     *     throw new JMLNonExecutableException();
     *   }
     * </pre>
     * The generated method declaration is added to the current class,
     * <code>classDecl</code> and its name is added to the field 
     * <code>modelMethods</code>.
     * 
     * <p> If the given declaration is a ghost field declaration,
     * generates a pair of ghost field accessor methods (getter and
     * setter), a private field for storing ghost values, and possibly
     * an initialization block. The generated code has the following
     * structure:
     *
     * <pre>
     *  private [static] T v;
     *  [static] {
     *     v = initializer;
     *  }
     *  public [static] T ghost$v$C() {
     *     return v;
     *  } 
     *  public [static] void ghost$v$C(T x) {
     *     v = x;
     *  } 
     * </pre>
     *
     * The generated accessors are added to the current class
     * declaration, and the new field declaration (with its
     * initialization block) is stored into the input field
     * declaration for later pretty-printing in the proper order.
     *
     * <pre><jml>
     * also
     *   requires fieldDecl != null;
     *   modifies modelMethods, classDecl.*;
     * </jml></pre>
     *
     * @see #translateRepresents(JmlRepresentsDecl[])
     * @see #modelMethods
     */
    protected JmlFieldDeclaration translateField(
        JmlFieldDeclaration fieldDecl) 
    {
	JmlFieldDeclaration fdecl = super.translateField(fieldDecl);
	final CField field = fdecl.getField();
	final long modifiers = field.modifiers();
	if (hasFlag(modifiers, ACC_MODEL)) {
	    if (!isAccessorGenerated(field)) {
		JmlMemberAccess access = classDecl.jmlAccess();
		if ( !classDecl.isInterface() && !access.isAbstract() ) {
		    JmlRacGenerator.warn(fieldDecl.getTokenReference(), 
					 RacMessages.MAY_NOT_EXECUTABLE,
					 "References to model field \"" +
					 fieldDecl.ident() + "\"");
		}
                addNewMethod(defaultModelFieldAccessor(field));
		setAccessorGenerated(field);
	    }
	} else if (hasFlag(modifiers, ACC_GHOST)) {
            addNewMethod(ghostFieldAccessors(fdecl));
        } else if (specAccessorNeeded(modifiers)) {
            addNewMethod(specPublicAccessor(field));
	    setAccessorGenerated(field);
	}
	return fdecl;
    }

    /** 
     * Translates invariant clauses of this class declaration. This
     * method produces a couple of <em>invariant check methods</em>,
     * one for the static invariants and the other for instance
     * invariants.
     * 
     * <pre><jml>
     * also
     *   modifies classDecl.*;
     * </jml></pre>
     */
    protected void translateInvariant() {
	TransInvariant tinv = new TransInvariant(classDecl, varGen);
	RacNode mdecl = tinv.translate(classDecl.invariants());
	addNewMethod((JmlMethodDeclaration)mdecl);
    }

    /** 
     * Translates constraint clauses of the target class. This method
     * produces <em>constraint check methods</em>, <em>old expression
     * evaluation methods</em>, and <em>old expression fields</em>, to
     * be added to the translated class.
     * 
     * <pre><jml>
     * also
     *   assignable classDecl.*;
     * </jml></pre>
     *
     * @see TransConstraint
     */
    protected void translateConstraint() {
        // add static/non-static constraint check methods and old
        // expression evaluation methods
	TransConstraint trans = new TransConstraint(classDecl, varGen);
	RacNode mdecl = trans.translate(classDecl.constraints());
	addNewMethod((JmlMethodDeclaration)mdecl);

        PreValueVars.Entry[] staticFields = trans.oldExprFields(true);
	PreValueVars.Entry[] instanceFields = trans.oldExprFields(false);
	if (staticFields.length > 0 || instanceFields.length > 0) {
	    String fieldDecls = "";

	    String mod = "private ";
            for (int i = 0; i < staticFields.length; i++) 
		fieldDecls += mod + staticFields[i] + "\n";

            for (int i = 0; i < instanceFields.length; i++) 
		fieldDecls += mod + instanceFields[i] + "\n";;

	    fieldDecls = "\n// Generated by JML\n" + fieldDecls;
	    addNewMethod(RacParser.parseMethod(fieldDecls));
	}
    }

    /**
     * Translates a class body of the target class declaration. 
     * This method is overridden here to generate a default constructor
     * with assertion checking capability if the class has no constructor. 
     *
     * @param inners inner classes
     * @param methods mthod declarations
     * @param fieldsAndInits field declarations
     *
     * <pre><jml>
     * also
     *   requires inners != null && methods != null && fieldsAndInits != null;
     * </jml></pre>     
     */
    protected void translateBody(ArrayList inners, ArrayList methods, 
	JPhylum[] fieldsAndInits ) 
    {
	boolean hasConstructor = false;
	Iterator iter = methods.iterator();
	while (!hasConstructor && iter.hasNext()) { 
           hasConstructor = 
                iter.next() instanceof JConstructorDeclarationType;
	}

        // generate a default constructor (JLS 8.8.7) and translate it
        if (!hasConstructor && !genSpecTestFile) {
	    // We don't do spec files because we don't know whether
	    // the class under test actually has a constructor or not
	    // (it might be private).  So essentially we require that
	    // a spec file declare any constructors (it should be
	    // giving them specs in any case).
	    
	    JmlConstructorDeclaration decl = defaultConstructor();
	    addNewMethod(decl);
	    translateMethod(decl);
        }

        super.translateBody(inners, methods, fieldsAndInits);
    }

    /**
     * Translates a JML method declaration. A translation of JML method
     * declaration produces four new methods, and the original method
     * is renamed and made <tt>private</tt>-accessible:
     * 
     * <p>
     * <ul><li>
     * <em>Precondition checking method</em>. A <tt>protected</tt> method
     * to check the precondition of the original method (refer to the class
     * <code>PreconditionMethod</code>).
     * </li></ul>
     *
     * <ul><li>
     * <em>Normal and exceptional postcondition checking methods</em>. 
     * A couple of <tt>protected</tt> methods to check the normal and
     * exceptional postconditions of the original method
     * (refer to the classes <code>PostconditionMethod</code> and
     * <code>ExceptionalPostconditionMethod</code>).
     * </li></ul>
     *
     * <ul><li>
     * <em>Wrapper method</em>. A wrapper method that wraps the original
     * method with pre and postcondition checking (refer to the class
     * <code>WappersMethod</code>).
     *
     * <p>
     * The translation also generate several new instance fields to store
     * temporally values used by the assertion checking methods. 
     * There are three kinds of variables that are generated and declared
     * as instance fields.
     * 
     * <p>
     * <ul><li>
     * <em>Precondition variables</em>.
     * </li></ul>
     *
     * <ul><li>
     * <em>Pre (\pre) expression variables</em>.
     * </li></ul>
     *
     * <ul><li>
     * <em>Old (\old) expression variables</em>.
     * </li></ul>
     *
     * <ul><li>
     * <em>Specification variables</em>.
     * </ul></li>
     *
     * <p>
     * Similarly, translation of JML constructor declaration also produces 
     * four new methods, and the original constructor becomes a wrapper
     * method (refer to the class <code>TransConstructor</code>).
     */
    protected void translateMethod(JmlMethodDeclaration mdecl) {
        // no translation for model or main methods
        //if (isMain(mdecl)) {
            //return;
        //}
        if (mdecl.isModel()) {
            // !FIXME! no translation of in-line assertions for medel methods
            translateModelMethod(mdecl);
            return;
        }

	// !FIXME! no visibility checking yet 
	// modUtil.asString(self.modifiers()));

	Object body = mdecl.body();
	if (genSpecTestFile && mdecl.isConstructor()) body = ((JBlock)body).body();
	if (body != null) {
	    VarGenerator varGenBody = VarGenerator.forMethod(varGen);
	    TransMethodBody transBody = null;
	    if (mdecl.isConstructor())
		transBody = new TransConstructorBody(varGenBody, mdecl, classDecl);
	    else
		transBody = new TransMethodBody(varGenBody, mdecl, classDecl);
	    mdecl.setBody(transBody.translate());
        }

        // generate accessor method for spec_public and spec_protected
        if (mdecl.isSpecPublic() || mdecl.isSpecProtected()) {
            if (mdecl.isConstructor())
                typeDecl.addMember(specPublicConstructorAccessor(
                                       (JmlConstructorDeclaration) mdecl));
            else
                typeDecl.addMember(specPublicMethodAccessor(mdecl));
        }
        
	// translate method specification
	TransMethod trans = createMethodTranslator(mdecl);
	trans.perform(mdecl);

	// add newly created methods if any
	Iterator iter = trans.newMethods().iterator();
	while (iter.hasNext())
	    typeDecl.addMember((JMethodDeclarationType)iter.next());
	
	// add newly created fields if any
	if (trans.hasNewFields()) {
	    iter = trans.newFields().iterator();
	    String fieldDecls = "";
	    while (iter.hasNext())
		fieldDecls = fieldDecls + iter.next();
	    fieldDecls = "\n// Generated by JML\n" + fieldDecls;
	    typeDecl.addMember(RacParser.parseMethod(fieldDecls));
	}
	
    }

    /** Creates a method translator for the given method declaration. */
    protected TransMethod createMethodTranslator(JmlMethodDeclaration mdecl) {
        return mdecl.isConstructor() ?
            new TransConstructor(typeDecl, varGen) :
	    new TransMethod(typeDecl, varGen);
    }

    /**
     * Translates the given model method (or constructor),
     * <code>mdecl</code>. !FIXME!describe translation rules.
     */
    protected void translateModelMethod(JmlMethodDeclaration mdecl) {
        //@ assume mdecl.isModel();

        // if not executable, we are done.
        if (!mdecl.isExecutableModel()) {
            return;
        }

        // turn it into a Java method or constructor. Perhaps,
        // generate new one instead of mutating it?
	mdecl.setModifiers(mdecl.modifiers() & (~ACC_MODEL));

	mdecl.setModifiers2(mdecl.modifiers2() | ACC2_RAC_METHOD);
        
        // for a constructor, we are almost done
        if (mdecl.isConstructor()) {
            // if protected or package-visivible, make it public so
            // that dynamic call can be maded to it by subclasses or
            // other classes in the same package.
            if (!(mdecl.isPublic() || mdecl.isPrivate())) { 
                long mod = mdecl.modifiers();
                mod = (mod | ACC_PUBLIC) & (~ACC_PROTECTED) & (~ACC_PRIVATE);
                mdecl.setModifiers(mod);
            }

	    mdecl.setModifiers2(mdecl.modifiers2() | ACC2_RAC_METHOD);
            return;
        }

        // for a public or private method, we are done.
        if (mdecl.isPublic() || mdecl.isPrivate()) {
            return;
        }

        // for a protected and package-visible method, build an access
        // method so that subclasses or other classes in the same
        // package can call it dynamically.
	JmlMethodDeclaration decl = JmlMethodDeclaration.makeInstance( 	// FIXME, This one can changed back to JmlMethodDeclaration if the member variable modifier2 is moved
           mdecl.getTokenReference(),
	   (mdecl.modifiers() | ACC_PUBLIC) & (~ACC_PROTECTED),
           mdecl.typevariables(),
	   mdecl.returnType(),
	   TransUtils.modelPublicAccessorName(mdecl.getMethod()),
	   mdecl.parameters(), 
	   mdecl.getExceptions(),
	   mdecl.body(), 
	   null,
	   null,
	   null);

	decl.setModifiers2(decl.modifiers2() | ACC2_RAC_METHOD);

        // finally, add the accessor to the host class
        addNewMethod(RacParser.parseMethod(
           "\n/** Generated by JML to access the model method " +
           mdecl.ident() + ". */$0", decl));
    }

    /** Returns an access method for the given spec_public (or
     * spec_protected) method declaration, <code>mdecl</code>. The
     * returned method has the following code structure.
     *
     * <pre>
     * public [static] T specPublic$m(x1, ..., xn) {
     *   [return] m(x1, ..., xn);
     * }
     * </pre>
     * 
     * <pre><jml>
     * requires mdecl.isSpecPublic() || mdecl.isSpecPublic();
     * </jml></pre>
     */
    private /*@ non_null @*/ JmlMethodDeclaration specPublicMethodAccessor(
        /*@ non_null @*/ JmlMethodDeclaration mdecl) {
	JFormalParameter[] parameters = mdecl.parameters();
        String ident = mdecl.ident();
        CType returnType = mdecl.returnType();
	boolean hasReturn = returnType != null && !returnType.isVoid();

        // build arguments for delegation call of the form: (x1, ..., xn)
	StringBuffer args = new StringBuffer("(");
	for (int i = 0; i < parameters.length; i++) {
	    if (i != 0)
		args.append(", ");
	    args.append(parameters[i].ident());
	}
	args.append(")");

        // build the body
        JBlock body = RacParser.parseBlock(
            "{\n" +
               (hasReturn ? "  return " : "  ") + ident + args + ";\n" +
            "}");

        // build accessor method
        JMethodDeclaration m = new JMethodDeclaration(
           mdecl.getTokenReference(),
	   ACC_PUBLIC | (mdecl.isStatic() ? ACC_STATIC : 0),
	   CTypeVariable.EMPTY,
	   returnType,
	   specPublicAccessorName(ident),
	   parameters,
	   mdecl.getExceptions(),
	   body,
	   null,
	   null);

        // add javadoc and return the accessor
        return RacParser.parseMethod(
            "\n/** Generated by JML for the " +
            (mdecl.isSpecPublic() ? "spec_public" : "spec_protected") +
            " method " + ident + ". */$0", m);
    }

    /** Returns an access method for the given spec_public (or
     * spec_protected) constructor declaration. The
     * returned method has the following code structure.
     *
     * <pre>
     * public static T specPublic$$init$(x1, ..., xn) {
     *   return new T(x1, ..., xn);
     * }
     * </pre>
     *
     * where <code>T</code> is the name of type currently being
     * translated.
     * 
     * <pre><jml>
     * requires mdecl.isSpecPublic() || mdecl.isSpecPublic();
     * </jml></pre>
     */
    private /*@ non_null @*/ JmlMethodDeclaration 
        specPublicConstructorAccessor(
            /*@ non_null @*/ JmlConstructorDeclaration mdecl) {

	JFormalParameter[] parameters = mdecl.parameters();
        // String ident = MN_INIT;
        CType returnType = 
            CTopLevel.getTypeRep(typeDecl.getCClass().qualifiedName(), true);

        // build arguments for delegation call of the form: (x1, ..., xn)
	StringBuffer args = new StringBuffer("(");
	for (int i = 0; i < parameters.length; i++) {
	    if (i != 0)
		args.append(", ");
	    args.append(parameters[i].ident());
	}
	args.append(")");

        // build the body
        JBlock body = RacParser.parseBlock(
            "{\n" +
            "  return new " + typeDecl.ident() + args + ";\n" +
            "}");

        // build accessor method
        JMethodDeclaration m = new JMethodDeclaration(
           mdecl.getTokenReference(),
	   ACC_PUBLIC | ACC_STATIC,
	   CTypeVariable.EMPTY,
	   returnType,
	   specPublicAccessorName(MN_INIT),
	   parameters,
	   mdecl.getExceptions(),
	   body,
	   null,
	   null);

        // add javadoc and return the accessor
        return RacParser.parseMethod(
            "\n/** Generated by JML for the " +
            (mdecl.isSpecPublic() ? "spec_public" : "spec_protected") +
            " constructor. */$0", m);
    }

    /** 
     * Performs a finialization of translation, e.g., generation of spec 
     * inheritance mechanism. 
     */
    protected void finalizeTranslation() {
	// generate field accessors for model fields inherited from
	// the interfaces

	boolean modelInheritanceNeeded = false;
        modelInheritanceNeeded = generateAccessorsForInterfaceModelFields();

	// generate specification inheritance methods, if needed.
        if (specInheritanceMethodNeeded || // see trans of \invariant_for
            (TransUtils.useReflection() &&
            (classDecl.superName() != null || 
	    classDecl.interfaces().length > 0))) {
	    addNewMethod(assertionInheritanceMethod());
            specInheritanceMethodNeeded = false;
	}
	if (dynamicInvocationMethodNeeded) {
	    addNewMethod(dynamicInvocationMethod());
            dynamicInvocationMethodNeeded = false;
        }

	// generate a interface surrogate chache and its accessors
        // !FIXME! can clear to generate surrogate cache when it is
        // absolutely needed.
	if (surrogateCacheNeeded ||
            true ) { // classDecl.interfaces().length > 0) {
	    addNewMethod(surrogateCacheDeclaration());

            // indicates to the pretty-printer that this class implements
            // the interface runtime.JMLCheckable.
            classDecl.setRacable();
	}

        // add a flag and accessor that indicates whether the "this"
        // object has completed its construction. This flag is used to
        // disable assertion checks while the object is being
        // constructed by the constructor.
        addNewMethod(constructionStatusAccessor());

        // finalization common to all type declarations, e.g.,
        // adding class/instance initialzation flags.
        super.finalizeTranslation();

        // add a static initializer that checks static invariants are
        // established at the end of initialization of this class.
        addNewMethod(staticInvariantCheckInitializer());

    }
 
    /**
     * Returns a static initializer that checks whether static
     * invariants are established by the initialization of this class.
     * The returned static initializer simply calls the static
     * invariant check method of this class.
     */
    protected JmlMethodDeclaration staticInvariantCheckInitializer() {
        String cn = TransUtils.dynamicCallName(classDecl.getCClass());
	return RacParser.parseMethod(
	    "\n/** Generated by JML to check the establishment of static\n" +
            "   * invariants. */\n" +
	    "static {\n" +
            "  // cache this class for dynamic calls.\n" +
            "  org.jmlspecs.jmlrac.runtime.JMLChecker.classes.put(" +
               "\"" + cn + "\", " + cn + ".class);\n" +
            "  // check static invariant.\n" +
	    "  " + MN_CHECK_INV + "static(\"<clinit>\");\n" +
            "}\n");
    }

    /**
     * Returns declarations of a private instance field and a
     * protected access method for storing and accessing the status of
     * an instance construction. The returned field and method are
     * used to disable assertion checks while the "this" object is
     * being constructed by the constructor.
     */
    protected JmlMethodDeclaration constructionStatusAccessor() {
	return RacParser.parseMethod(
	    "\n/** Generated by JML to indicate if an instance completes\n" +
            "   * its construction. */\n" +
	    "private transient boolean " + VN_CONSTRUCTED + " = false;\n\n" +
            
	    "/** Generated by JML to query an instance's construction\n" +
            "  * status. */\n" +
	    "protected boolean " + VN_CONSTRUCTED + "() {\n" +
            "  return " + VN_CONSTRUCTED + ";\n" +
            "}\n");
    }

    // ----------------------------------------------------------------------
    // MODEL AND GHOST FIELDS
    // ----------------------------------------------------------------------

    /** 
     * Generates access methods for all the model fields inherited from 
     * the interfaces of this class. The generated access method is 
     * either a default access method or a dynamic delegation method.
     * If the inherited model field has no represents clauses in the 
     * interface hiearchy, a default access method is generated; otherwise,
     * a dynamic delegation method is generated which makes a dynamic
     * call to the corresponding access method of the interfeace's surrogate
     * class.
     *
     * @see #defaultModelFieldAccessor(CField)
     * @see #modelFieldDelegationMethod(CField, JmlTypeDeclaration)
     */
    private boolean generateAccessorsForInterfaceModelFields() {
	boolean inheritanceNeeded = false;
	JFieldDeclarationType[] fields = interfaceModelFields();
	for (int i = 0; i < fields.length; i++) {
	    CField field = fields[i].getField();
	    if (!isAccessorGenerated(field)) { 
		JmlTypeDeclaration tdecl 
                    = findTypeWithRepresentsFor(classDecl, field);
		JmlMethodDeclaration mdecl = null;
		if (tdecl == null) {
		    JmlMemberAccess access = classDecl.jmlAccess();
		    if ( !classDecl.isInterface() && !access.isAbstract() ) {
			JmlRacGenerator.warn(fields[i].getTokenReference(), 
					     RacMessages.MAY_NOT_EXECUTABLE,
					     "References to model field \"" +
					     field.ident() + "\"");
		    }
		    mdecl = defaultModelFieldAccessor(field);
		} else {
		    inheritanceNeeded = true;
		    mdecl = modelFieldDelegationMethod(field, tdecl);
		}
		addNewMethod(mdecl);
		setAccessorGenerated(field);
	    }
	}
	return inheritanceNeeded;
    }
    
    /**
     * Returns all model fields inherited from the interfaces of this
     * class, but excluding those also inherited from the superclass
     * chain.
     */
    private /*@ non_null @*/ JFieldDeclarationType[] interfaceModelFields() {
	JFieldDeclarationType[] fields = 
            classDecl.getAllInterfaceModelFields();
        
	JmlClassDeclaration superDecl 
            = JmlTypeLoader.getJmlSingleton().superClassOf(classDecl);

        if (superDecl == null) {
            return fields;
        }

        JFieldDeclarationType[] superFields = 
            superDecl.getAllInterfaceModelFields();
        if (superFields.length == 0) {
            return fields;
        }

        // filter those inherited through super (too)
        Set all = toSet(fields);
        all.removeAll(toSet(superFields));

        return (JFieldDeclarationType[]) 
            all.toArray(new JFieldDeclarationType[0]);
    }

    /** Returns a set containing elements from the given array. */
    private /*@ non_null @*/ Set toSet(JFieldDeclarationType[] fdecls) {
        Set result = new HashSet();
        for (int i = 0; i < fdecls.length; i++) {
            if (fdecls[i] != null) {
                result.add(fdecls[i]);
            }
        }
        return result;
    }

    /** 
     * Returns a default accessor method for a model field. 
     * The returned method has the following form:
     * <pre>
     *  public [static] T model$n() {
     *    throw new JMLNonExecutableException();
     *  }
     * </pre>
     */
    private JmlMethodDeclaration defaultModelFieldAccessor(CField field)
    {
	final String mod = field.isStatic() ? "static " : "";
	final CType type = field.getType();
	String mn = MN_MODEL + field.ident() + "$" + field.owner().ident();
	return RacParser.parseMethod(
            "\n\n/** Generated by JML to access the model field " + 
            field.ident() + ". */\n" +
            "public " + mod + toString(type) + " " + mn + "() {\n" +
            "  throw JMLChecker.ANGELIC_EXCEPTION;\n" +
            "}\n");
    }

    /** 
     * Returns a model field delegation method for model
     * field, <code>field</code>, whose <code>represents</code> clause 
     * is found in type <code>tdecl</code>.
     * The returned method has the following form:
     * <pre>
     *  public [static] T model$n() {
     *    Object obj = dynamic_delegation_to_accessor_method_in_tdecl();
     *    return unwrapped_obj;
     *  }
     * </pre>
     */
    private JmlMethodDeclaration modelFieldDelegationMethod(CField field, 
        JmlTypeDeclaration tdecl) 
    {
        dynamicInvocationMethodNeeded = true;

	String mod = field.isStatic() ? "static " : "";
        String rn = field.isStatic() ? "null" : "this";
	CType type = field.getType();
	String mn = MN_MODEL + field.ident() + "$" + field.owner().ident();
	String cn = TransUtils.dynamicCallName(tdecl.getCClass());
	return RacParser.parseMethod(
            "\n\n/** Generated by JML to access the model field " + 
            field.ident() + ". */\n" +
            "public " + mod + toString(type) + " " + mn + "() {\n" +
            "  java.lang.Object obj = rac$invoke(\"" + cn + "\"," + rn + ", \"" +
               mn + "\", null, null);\n" +
            "  return " + unwrapObject(type, "obj") + ";\n" +
            "}\n");
    }

    /** 
     * Returns a default access method for a <code>spec_public</code> field. 
     * The returned method has the following form:
     * <pre>
     *  public [static] T spec$n() {
     *    return x;
     *  }
     * </pre>
     * where <code>x</code> is the name of the given field.
     */
    protected JmlMethodDeclaration specPublicAccessor(CField field)
    {
	String mod = field.isStatic() ? "static " : "";
	CType type = field.getType();
	String mn = MN_SPEC + field.ident() + (genSpecTestFile?"":"$" + classDecl.ident());
	
        String field_accessor;
        if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
            field_accessor = "_chx_get_" + field.ident() + "()";
        } else {
            field_accessor = field.ident();
        } 


        return RacParser.parseMethod(
            "\n\n/** Generated by JML to access the spec_public field " + 
            field.ident() + ". */\n" +
	    "public " + mod + toString(type) + " " + mn + "() {\n" +
	    "  return " + field_accessor + ";\n" +
	    "}\n");
    }
 
    /** Adds a new method declaration, <code>mdecl</code>, to the 
     * instrumented class.
     */
    //@ also
    //@   assignable classDecl.*;
    protected void addNewMethod(JmlMethodDeclaration mdecl) {
        classDecl.addMember(mdecl);
    }

    /** Returns a default constructor for the target class declaration. 
     */
    private JmlConstructorDeclaration defaultConstructor() {
        // The access modifier of the default constructor is the same
        // as that of the class [JLS 8.8.7].
        long mod = 0;
        final long cmod = classDecl.modifiers();
        if (hasFlag(cmod, ACC_PUBLIC)) {
            mod = ACC_PUBLIC;
        } else if (hasFlag(cmod, ACC_PROTECTED)) {
            mod = ACC_PROTECTED;
        } else if (hasFlag(cmod, ACC_PRIVATE)) {
            mod = ACC_PRIVATE;
        }

        return JmlConstructorDeclaration.makeInstance(
            TokenReference.NO_REF,
            mod,
            classDecl.ident(),
            JFormalParameter.EMPTY,
            CClassType.EMPTY,
            new JConstructorBlock(TokenReference.NO_REF, genSpecTestFile ? null : new JStatement[0]),
            null,  // Javadoc comment
            null,  // comments
            null); // method spec
    }

    /** 
     * Returns a method declaration implementing dynamic inheritance
     * using the Java's reflection facility. The returned method is
     * often called an <em>inheritance method</em>. The inheritance
     * method makes dynamic invocations to assertion check methods,
     * given an assertion check method's class name, method name,
     * formal parameter types, and actual arguments. If the target
     * assertion check method returns a boolean value, that value is
     * returned; otherwise, true is returned. An exception thrown by
     * the targe method is transparently propagated to the caller.
     *
     * <pre><jml>
     * ensures \result != null;
     * </jml></pre>
     */
    protected JmlMethodDeclaration assertionInheritanceMethod() {
	surrogateCacheNeeded = true;
	return RacParser.parseMethod(
           "\n/** Generated by JML to make dynamic calls to an assertion\n" +
           " * check methoda, given its class name (className),\n" +
           " * method name (name), parameter types (types),\n" +
           " * and actual arguments (args).\n" +
           " * If the assertion method returns a boolean value,\n" +
           " * that value is returned; otherwise, true is returned.\n" +
           " * An exception thrown by the assertion method is transparently\n"+
           " * propagated to the caller. */\n" +
	   "private static boolean rac$check(java.lang.String cn, JMLCheckable self,\n" +
           "  java.lang.String name, java.lang.Class types[], java.lang.Object args[]) {\n" +
	   "  try {\n" +
	   "    java.lang.Class cls = rac$forName(cn);\n" +
           "    java.lang.Object inst = rac$receiver(cn, self);\n" +
	   "    java.lang.reflect.Method meth =\n" +
           "      " +TN_JMLSURROGATE+ ".getMethod(cls, name, types);\n" +
	   "    java.lang.Object result = meth.invoke(inst, args);\n" +
           "    return (result instanceof java.lang.Boolean) ?\n" +
           "     ((java.lang.Boolean) result).booleanValue() : true;\n" +
	   "  }\n" +
	   "  catch (java.lang.reflect.InvocationTargetException e) {\n" +
           "    // ok. The invoked method throws an exception, possibly an\n"+
           "    // assertion violation exception. Rethrow it transparently.\n"+
	   "    Throwable thr = e.getTargetException();\n" +
	   "    if (thr instanceof java.lang.RuntimeException)\n" +
	   "       throw (java.lang.RuntimeException) thr;\n" +
	   "    else if (thr instanceof java.lang.Error)\n" +
	   "       throw (java.lang.Error) thr;\n" +
	   "    else\n" +
	   "       throw new java.lang.RuntimeException(e.getMessage());\n" +
	   "  }\n" +

           // Ignore all other exceptions such as NoClassDefFoundError.
	   "  catch (java.lang.Throwable e) {\n" +
           "     //System.out.println(e.getClass().getName());\n" +
	   "     return false;\n" +
	   "  }\n" +
	   "}\n");
    }

    /** 
     * Returns a field declaration for an interface surroagte cache
     * with a cache lookup method and a cache update method.  The
     * returned value is of type <code>JmlMethodDeclaration</code> so
     * that it can be added to the current type declaration by using
     * <code>addMember</code> method.
     */
    protected JmlMethodDeclaration surrogateCacheDeclaration() {
	return RacParser.parseMethod(
	 "\n/** Generated by JML for caching interface surrogates.\n" +
         " * It is a mapp from fully qualified interface names\n" +
         " * to their surrogate objects. */\n" +

         // It is *critical* not to initialize this field to, e.g.,
         // new java.util.HashMap(). Otherwise, the surrogate objects
         // created during the explicit constructor invocation (e.g.,
         // super()) in the constructor will be all lost, as the field
         // becomes initialized after the explicit constructor
         // invocation call.

         "public transient java.util.Map rac$surrogates;\n" + 
         
         // The if-conditions below are needed to guard against null
         // pointer exceptions when called by the superclass's
         // constructor (by super call) while the subclass's object is
         // still being constructed.

	 "\n/** Generated by JML to returns the surrogate of\n" +
         " * the given interface, <code>name</code>. */\n" +
         "public java.lang.Object rac$getSurrogate(java.lang.String name) {\n" +
         "  if (rac$surrogates == null) {\n" +
         "    rac$surrogates = new java.util.HashMap(37);\n" +
         "  }\n" +
         "  return rac$surrogates.get(name);\n" +
         "}\n" +

         "\n/** Generated by JML to set the surrogate of\n" +
         " * the given interface, name, to the object obj. */\n" +
         "public void rac$setSurrogate(java.lang.String name, " +
                                       TN_JMLSURROGATE + " obj) {\n" +
         "  if (rac$surrogates == null) {\n" +
         "    rac$surrogates = new java.util.HashMap(37);\n" +
         "  }\n" +
         "  rac$surrogates.put(name, obj);\n" +
         "}\n" +

         "\n/** Generated by JML to return the actual receiver (possibly\n" +
         " * a surrogate) for dynamic calls to the class, name. */\n" +
         "public static java.lang.Object rac$receiver(java.lang.String name, java.lang.Object forObj) {\n" +
         "  if (forObj == null || !name.endsWith(\"$" +
                                                 TN_SURROGATE+ "\")) {\n" +
         "    return forObj;\n" +
         "  }\n" +
         "  //@ assume forObj instanceof " +TN_JMLCHECKABLE+ ";\n" +
         "  " +TN_JMLCHECKABLE+ " cobj = (" +TN_JMLCHECKABLE+ ") forObj;\n" +
         "  try {\n" +
         "    java.lang.Object surObj = cobj.rac$getSurrogate(name);\n" +
         "    if (surObj == null) {\n" +
         "      java.lang.Class[] types =\n" +
         "        new java.lang.Class[] { " +TN_JMLCHECKABLE+ ".class };\n" +
         "      java.lang.Class clazz = rac$forName(name);\n" +
         "      java.lang.reflect.Constructor constr =\n" +
         "        clazz.getConstructor(types);\n" +
         "      java.lang.Object[] args = new java.lang.Object[] { cobj };\n" +
         "      surObj = constr.newInstance(args);\n" +
         "      cobj.rac$setSurrogate(name,(" +TN_JMLSURROGATE+ ")surObj);\n" +
         "    }\n" +
         "    return surObj;\n" +
         "  }\n" +
         "  catch (Exception e) {\n" +
         "    java.lang.System.err.println(\"Internal error in getSurrogate()!\");\n" +
         "    //e.printStackTrace();\n" +
         "    System.exit(1);\n" +
         "  }\n" +
         "  return null;\n" +
         "}\n");
    }

    /** 
     * Returns a method declaration that makes a dynamic call to the
     * given method using the Java's reflection facility. The
     * inheritance method makes dynamic invocations to a target method
     * to is specified by its name, its owner's class name, formal
     * parameter types, and actual arguments. The target method is
     * typically an access method generated by the runtime assertion
     * checker, such as an accessor for a model, ghost, spec_public,
     * or spec_protected field, a setter for a ghost field, and an
     * accessor for a spec_public or spec_protected method. The return
     * value is that of the target method, if necessary, wrapped in an
     * object. If the target method is not found, a instance of
     * JMLNonExecutableException is thrown. If the target method
     * throws a RuntimeException, including JMLNonExecutableException,
     * the exception is rethrown; for other types of exceptions, an
     * instanceof RuntimeException is thrown.
     */
    protected JmlMethodDeclaration dynamicInvocationMethod() {
	surrogateCacheNeeded = true;
        return super.dynamicInvocationMethod();
    }

    /** Returns a string form of code that, if executed, returns the
     * receiver of for a dynamic call. This method is used by the
     * method <code>dynamicInvocationMethod</code> to define a
     * subclass-specific dynamic invocation method. 
     *
     * <pre><jml>
     * also
     * ensures \result.equals("rac$receiver(" +clsName+ ", " +receiver+ ")");
     * </jml></pre>
     *
     * @see TransType#dynamicInvocationMethod
     */
    protected String receiver(String clsName, String clazz, String receiver) {
        return "rac$receiver(" + clsName + ", " + receiver + ")";
    }


    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** Target class declaration to be translated.
     *
     * <pre><jml>
     * private invariant classDecl == (JmlClassDeclaration)typeDecl;
     * </jml></pre>
     */
    private final /*@ spec_protected non_null @*/JmlClassDeclaration classDecl;

    /** True if a cache for interface surrogates need be generated for
     * the current class being translated.
     * 
     * <pre><jml>
     * private invariant dynamicInvocationMethodNeeded 
     *                     ==> surrogateCacheNeeded;
     * </jml></pre>
     *
     * @see #dynamicInvocationMethod
     * @see #surrogateCacheDeclaration
     */
    private boolean surrogateCacheNeeded = false;
}
