/* $Id: TransConstraint.java,v 1.30 2006/12/11 05:59:03 chalin Exp $
 *
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import java.util.*;

/**
 * A class for translating JML (history) constraints. A history
 * constraint is translated into a couple of <em>constraint check
 * methods</em>, one for static (class) constraints and the other for
 * non-static (instance) constraints.  The translation also produce a
 * couple of <em>old expression evaluation methods</em>, one for
 * static constraints and the other for both static and non-static
 * constraints. The latter is supposed to be called by non-static
 * methods. The old expression evaluation methods, if called in the
 * pre-state, evaluate old expressions appearing in the constraints
 * and store the results to newly introduced private fields, called
 * <em>old expression fields</em>. An old expression field is a
 * temporary location to hold the value of an old expression
 * (evaluated in the pre-state). The constraint check methods use the
 * old variable in place of the original old expression.  The client
 * of this class is responsible for (1) adding constraint check
 * methods and old expression evaluation methods to the host type
 * declaration containing the target constraint clauses, (2) declaring
 * appropriate old expression fields to the host type declaration, and
 * (3) making calls to the old expression evaluation methods in the
 * appropriate pre-state (e.g., in the wrapper method).
 *
 * @see ConstraintMethod
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.30 $
 */

public class TransConstraint extends TransUtils {

    /**
     * Constructs a <code>TransConstraint</code> object.
     * 
     * <pre><jml>
     *  requires typeDecl != null && varGen != null;
     * </jml></pre>
     *
     * @param typeDecl host type declaration containing the target
     *                 constraint clause to be translated
     * @param varGen variable generator to be used for the translation
     */
    public TransConstraint(JmlTypeDeclaration typeDecl, VarGenerator varGen)
    {
	this.typeDecl = typeDecl;
	this.varGen = varGen;
	oldExprsStatic = new ArrayList();
	oldExprsInstance = new ArrayList();
        stackVarsStatic = new ArrayList();
        stackVarsInstance = new ArrayList();
    }

    /** Returns field declarations for the old expressions appearing
     * in the constraint clause. Each field declaration has the
     * following string form: "<code>[static] T n;\n</code>", where
     * <code>T</code> is the type and <code>n</code> is the field
     * name. Note that the old expression evaluation methods, returned
     * by the method <code>translate</code>, if called, evaluate old
     * expressions and initialize the declared fields.
     * These fields are called <em>old expression fields</em>.
     * This method must be called after the translation.
     *
     * <pre><jml>
     * ensures \result != null;
     * </jml></pre>
     *
     * @param forStatic <code>true</code> for static old expressions;
     *                  <code>false</code> for instance old expressions.
     *
     * @see #translate(JmlConstraint[])
     * @see #oldExprsStatic
     * @see #oldExprsInstance
     * @see PreValueVars.Entry
     */
    public PreValueVars.Entry[] oldExprFields(boolean forStatic) {
        List exprs = forStatic ? oldExprsStatic : oldExprsInstance;
        PreValueVars.Entry[] result = new PreValueVars.Entry[exprs.size()];
        int j = 0;
        for (Iterator i = exprs.iterator(); i.hasNext(); ) 
            result[j++] = ((RacNode) i.next()).varDecl();
        return result;
    }

    /** 
     * Translates the given history constraints,
     * <code>constraints</code>, and returns the translation result.
     * This method translates the given history constraints and
     * produces a couple of new field declarations and several new
     * methods:
     *
     * <p>
     * <ul>
     * <li> A static constraint check method (see 
     *      {@link org.jmlspecs.jmlrac.ConstraintMethod}).</li>
     * <li> A non-static (instance) constraint check method (see
     *      {@link org.jmlspecs.jmlrac.ConstraintMethod}).</li>
     * <li> A static old-expression evaluation method.</li>
     * <li> A static stack field declaration.</li>
     * <li> A static old values save method.</li>
     * <li> A static old values restoration method.</li>
     * <li> A non-static stack field declaration.</li>
     * <li> A non-static old values save method.</li>
     * <li> A non-static old values restoration method.</li>
     * </ul>
     * </p>
     *
     * <p>
     * The old expression evaluation methods, if called, evaluate the
     * old expressions appearing in the constraints and store the
     * results into newly introduced fields, called <em>old expression
     * fields</em>.  The old expression fields are used instead of the
     * original old expressions by the constraint check methods.  The
     * generated methods are concatenated and returned as the
     * translation result.
     * </p>
     *
     * <p> The stack variables and their access methods are used to
     * correctly pass, in the presence recursive method calls, old
     * expression values evaluated in the pre=state to the constraint
     * check methods executed in the post-state.  
     * </p>
     *
     * <pre><jml>
     *  requires constraints != null;
     *  assignable oldExprsStatic, oldExprsInstance;
     *  assignable stackVarsStatic, stackVarsInstance, restoreMethods;
     *  ensures \result != null;
     * </jml></pre>
     *
     * @param constraints history constraints to be translated
     * @see #oldExprFields(boolean)
     */
    public RacNode translate(JmlConstraint[] constraints) 
    {
	// generate static/non-static constraint check methods
	RacNode result = constraintMethod(constraints, true);  // static
	result = RacParser.parseMethod("$0$1", result,  // non-static
				       constraintMethod(constraints, false));

	// generate static/non-static old exprs evalutation methods
	result = RacParser.parseMethod("$0$1", result,  // static
				       oldMethod(oldExprsStatic, true));
	List exprs = 
            new ArrayList(oldExprsStatic.size() + oldExprsInstance.size());
	exprs.addAll(oldExprsStatic);
	exprs.addAll(oldExprsInstance);
	result = RacParser.parseMethod("$0$1", result,  // non-static
				       oldMethod(exprs, false));

        return result;
    }

    /** 
     * Returns a static or an instance constraint check method for the
     * given history constraints, <code>constraints</code>.  As a
     * side-effect, this method also translates and accumulates into
     * private fields old expressions appearing in the constraints for
     * future use.
     *
     * <pre><jml>
     *  requires constraints != null;
     *  assignable oldExprsStatic, oldExprsInstance;
     *  assignable stackVarsStatic, stackVarsInstance, restoreMethods;
     *  ensures \result != null;
     * </jml></pre>
     *
     * @param constraints history constraints to be translated
     * @param forStatic type of constraints to be translated 
     *                  (static or non-static)
     * @return object of <code>RacNode</code> representing a 
     *         constraint check method.
     * @see #oldExprsStatic
     * @see #oldExprsInstance
     */
    private RacNode constraintMethod(JmlConstraint[] constraints,
                                     boolean forStatic) 
    {
        restoreMethods = new ArrayList();

        // conjoin universal constraints while translating
        // non-universal (i.e., method-specific) ones separately
	JExpression left = null;
        int cnt = 0; // num of non-universal constraints
        RacNode nonUniv = null; // translated non-universal check methods
	for (int i = 0; i < constraints.length; i++) {
            // skip uncheckable constraints (e.g., redundant,
            // not requested, or trivially true.
	    if (!isCheckable(constraints[i]) ||
		hasFlag(constraints[i].modifiers(), ACC_STATIC) != forStatic)
                continue;

            // translate method-specific constraints
            if (!constraints[i].isForEverything()) {
                if (nonUniv == null) {
                    nonUniv = constraintMethod(constraints[i], ++cnt);
                } else {
                    nonUniv = RacParser.parseMethod("$0$1", nonUniv,
                       constraintMethod(constraints[i], ++cnt));
                }
		continue;
            }

            // conjoin universal constraints
	    JExpression p = new RacPredicate(constraints[i].predicate());
	    left = left == null ? 
		p : new JConditionalAndExpression(org.multijava.util.compiler.TokenReference.NO_REF, left, p);
	}
	
        // translate conjoined universal constraints
        RacNode univ = null;
        if (left != null) {
            univ = constraintMethod(forStatic, 0, left, null);
        }

        // generate the mother constraint check method
        RacNode result = null;
        String code = "";
        { 
            if (univ != null) {
                code += MN_CHECK_HC + LocalConstraintMethod.postfix(forStatic, 
                        typeDecl, 0) + "(rac$msg, rac$name, rac$params);\n";
            }
            for (int i = 1; i <= cnt; i++) {
                code += MN_CHECK_HC + LocalConstraintMethod.postfix(forStatic, 
                        typeDecl, i) + "(rac$msg, rac$name, rac$params);\n";
            }
            RacNode stmt = code.length() == 0 ? null :
                RacParser.parseStatement(code);
            result = (RacNode) new MotherConstraintMethod(forStatic, 
                typeDecl, restoreMethods).generate(stmt);
        }

        // combine with univ and non-univ constraint check methods
        if (univ != null) {
            result = RacParser.parseMethod("$0$1", result, univ);
        }
        if (nonUniv != null) {
            result = RacParser.parseMethod("$0$1", result, nonUniv);
        }
        
        // generate weak/strong subtype constraint check methods
        if (!forStatic && !(typeDecl.isClass() && typeDecl.isFinal())) {
            RacNode stmt = code.length() == 0 ? null :
                RacParser.parseStatement(code);
            RacNode subtype = (RacNode) new SubtypeConstraintMethod(
                typeDecl, true, restoreMethods).generate(stmt);
            result = RacParser.parseMethod("$0$1", result, subtype);
            stmt = stmt == null ? null : RacParser.parseStatement(code);
            subtype = (RacNode) new SubtypeConstraintMethod(
                typeDecl, false, restoreMethods).generate(stmt);
            result = RacParser.parseMethod("$0$1", result, subtype);
        }

        return result;
    }

    /**
     * Returns a constraint check method for the given hisotry
     * constraint, <code>constraint</code>. The parameter
     * <code>cnt</code> is used as a suffix for the name of generated
     * check method. As a side-effect, this method also translates and
     * accumulates into private fields all the old expressions
     * appearing in the given constraint.
     *
     * <pre><jml>
     *  requires constraint != null;
     *  assignable oldExprsStatic, oldExprsInstance;
     *  assignable stackVarsStatic, stackVarsInstance;
     *  ensures \result != null;
     * </jml></pre>
     *
     * @param constraint history constraint to be translated
     * @return object of <code>RacNode</code> representing a 
     *         constraint check method.
     * @see #oldExprsStatic
     * @see #oldExprsInstance
     * @see #stackVarsStatic
     * @see #stackVarsInstance
     */
    private RacNode constraintMethod(JmlConstraint constraint, int cnt) {
        //@ assume !constraint.isForEverything();
        return constraintMethod(constraint.isStatic(), 
                                cnt, 
                                constraint.predicate(),
                                constraint.methodNames().methodNameList());
    }

    /**
     * Returns a constraint check method that checks the given
     * predicate, <code>pred</code> as a hisotry constraint.  If the
     * parameter <code>forStatic</code> is true, the given predicate
     * is regarded as a static constraint; otherwise, it is regared as
     * a non-static constraint. The parameter <code>cnt</code> is
     * used as a suffix for the name of generated check method. As a
     * side-effect, this method also translates and accumulates into
     * private fields all the old expressions appearing in the given
     * constraint.
     *
     * <pre><jml>
     *  assignable oldExprsStatic, oldExprsInstance;
     *  assignable stackVarsStatic, stackVarsInstance;
     *  ensures \result != null;
     * </jml></pre>
     *
     * @return object of <code>RacNode</code> representing a 
     *         constraint check method.
     * @see #oldExprsStatic
     * @see #oldExprsInstance
     * @see #stackVarsStatic
     * @see #stackVarsInstance
     * @see #restoreMethods
     */
    private RacNode constraintMethod(boolean forStatic, int cnt, 
                                     JExpression pred,
                                     JmlMethodName[] names) {
        // translate pred into Java statements
        VarGenerator mvarGen = VarGenerator.forMethod(varGen);
        RacNode stmt = null;
        List oldExprs = null;
        if (pred != null) {
            RacContext ctx = RacContext.createPositive();
            TransPostcondition trans = new TransPostcondition(
                mvarGen, ctx,  mvarGen, forStatic, pred, "rac$b", typeDecl);
            stmt = trans.wrappedStmt();
            oldExprs = trans.oldExprs(); // old expressions in pred
            // assign to global variable
            (forStatic ? oldExprsStatic : oldExprsInstance).addAll(oldExprs);
        }

        // generate a fresh stack variable for storing old values and
        // also stack var and access methods
        RacNode stackMethod = null;
        if (oldExprs != null && !oldExprs.isEmpty()) {
            String stackVar = varGen.freshStackVar();
            (forStatic ? stackVarsStatic : stackVarsInstance).add(stackVar);
            restoreMethods.add(MN_RESTORE_FROM + stackVar);
            // generate stack var and access methods
            stackMethod = genStackMethods(forStatic, stackVar, oldExprs);
        }

        // generate a constraint check method
	RacNode result = (RacNode) new LocalConstraintMethod(mvarGen,
           forStatic, 
           typeDecl, names, cnt).generate(stmt);

        // combine with stack method if exists
        if (stackMethod != null) {
            result = RacParser.parseMethod("$0$1", result, stackMethod);
        }
        
        return result;
    }

    /**
     * Returns true if the given constraint clause is checkable. The
     * clause is checkable if it is not a redundant specification or
     * the command line option "noredundancy" is not turned on.
     */
    private static boolean isCheckable(JmlConstraint cons) {
        return !cons.isRedundantly() || !Main.racOptions.noredundancy();
    }

    /** 
     * Returns an old-expression evaluation method for the given old
     * expressions, <code>exprs</code>. The resulting method, if
     * called, evaluates old expressions given by <code>exprs</code>
     * and stores the results into the so-called old expression
     * fields.
     *
     * <pre><jml>
     *  requires exprs != null && 
     *    (\forall Object o; exprs.contains(o);
     *       o instanceof RacNode);
     *  ensures \result != null;
     * </jml></pre>
     *
     * @param exprs a list of (translated) old expressions
     *              (<code>RacNode</code>)
     * @param forStatic type of constraint (static or non-static)
     * @return object of <code>RacNode</code> representing an old
     *         expression evaluation method.
     * @see #oldExprsStatic
     * @see #oldExprsInstance
     */
    private RacNode oldMethod(List exprs, boolean forStatic) 
    {
	StringBuffer code = new StringBuffer();

	// build the header, i.e., [static] void name().
	code.append("\n/** Generated by JML to evaluate, in the pre-state,\n" +
                    " * the old expressions appearing in the " +
                    (forStatic ? "static" : "non-static") + " constraints\n" +
                    " * of " +
                    ((typeDecl instanceof JmlClassDeclaration) ? 
                     "class " : "interface ") + 
                    typeDecl.ident() + ". */\n");
        code.append("public ");
	if (forStatic)
	    code.append("static ");
	code.append("void ");
	code.append(evalOldName(typeDecl.ident(), forStatic));

	// opening curly brace for body
	code.append("() {\n");

	// evaluate and store each old expression in exprs
	RacNode node = null;
 	if (exprs != null && exprs.size() > 0) {
 	    String str = "";
 	    for (int i = 0; i < exprs.size(); i++) {
 		str = str + "$" + i + "\n";
 	    }
 	    node = RacParser.parseStatement(str, exprs.toArray());
 	    node.incrIndent();
	    code.append("$0");
 	}

        // call to stack save methods. An instance old-expression
        // evaluation method computes values both for static and
        // non-static old expressions, so we should save both for
        // an instance case.
        for (Iterator i = stackVarsStatic.iterator(); i.hasNext(); ) {
            code.append("  " + MN_SAVE_TO + i.next() + "();\n");
        }
        if (!forStatic) {
            for (Iterator i = stackVarsInstance.iterator(); i.hasNext(); ) {
                code.append("  " + MN_SAVE_TO + i.next() + "();\n");
            }
        }

        VarGenerator mvarGen = VarGenerator.forMethod(varGen);

	// dynamically call the evaluation method of superclass
	if (typeDecl instanceof JmlClassDeclaration) {
	    String cls = ((JmlClassDeclaration)typeDecl).superName();
	    if (cls != null) {
                code.append(evalSuperHC(mvarGen, forStatic,
                                        typeDecl.getCClass().getSuperClass()));
	    }
	}

	// dynamically call the evaluation methods of interfaces
	CClassType[] interfaces = typeDecl.interfaces();	
	for (int i = 0; i < interfaces.length; i++) {
            code.append(evalSuperHC(mvarGen, forStatic,
                                    interfaces[i].getCClass()));
	}

	// closing curly brace for body
	code.append("}\n");

	return RacParser.parseMethod(code.toString(), node);
    }

    /** 
     * Return code that evaluates the old expressions of the given
     * supertype, <code>clazz</code>, by making an appropriate call to
     * its old expression evaluation method.
     *
     * @see #reflectiveCall(CClass)
     */
    private String evalSuperHC(VarGenerator varGen,
                               boolean forStatic,
                               /*@ non_null @*/ CClass clazz) {

        // no need to attempt, e.g., JDK classes?
        if (excludedFromInheritance(clazz)) 
            return "";

        // use dynamic call?
        if (useReflection()) {
	    String dname = dynamicCallName(clazz);
	    return "  rac$check(\"" + dname + "\", " + 
		(forStatic ? "null" : "this") + ", \"" +
                evalOldName(clazz, forStatic) + "\", null, null);\n";
        }
        
        // nope, let's go with non-reflective call
        String tab = "  ";
        StringBuffer code = new StringBuffer();

        // determine the receiver of the call. If clazz is a class, it
        // is "". If clazz is an interface, it is "T.TN_SURROGATE" for
        // a static method and "((T.TN_SURROGATE) surrogate)" for a
        // non-static method.
        String receiver = null;
        if (clazz.isInterface()) {
            receiver = clazz.qualifiedName().replace('/','.') + 
                "$" + TN_SURROGATE;
            if (!forStatic) {
                String target = "\"" +receiver+ "\"";
                // find or create a surrogate
                String delegee = (typeDecl instanceof JmlClassDeclaration) ?
                    "this" : VN_DELEGEE;
                String var = varGen.freshVar();
                code.append(tab);
                code.append("java.lang.Object " +var+ " = " +delegee+
                            ".rac$getSurrogate(" +target+ ");\n");
                code.append(tab);
                code.append("if (" +var+ " == null) {\n");
                code.append(tab);                
                code.append("  " +var+ " = new " +receiver+ 
                            "(" +delegee+ ");\n");
                code.append(tab);                
                code.append("}\n");
                code.append(tab);
                code.append(delegee + ".rac$setSurrogate(" +target+ 
                            ", (" +TN_JMLSURROGATE+ ") " +var+ ");\n");
                receiver = "((" +receiver+ ") " +var+ ")";
            }
        } else {
            if (forStatic) 
                receiver = clazz.qualifiedName().replace('/','.');
        }
        
        // call statement
        code.append(tab);
        code.append(receiver == null ? "" : receiver + ".");
        code.append(evalOldName(clazz, forStatic));
        code.append("();\n");
        
        return code.toString();
    }

    /** 
     * Returns a private stack field declaration with a pair of push
     * and pop methods. The stack is used to save into the pre-state
     * old expression values, and restore them from. The reason is to
     * pass correct pre-state values to the post-state assertion check
     * in the presence of recursive method calls. The returned code
     * has the following form.
     *
     * <pre>
     * private [static] Stack var = new Stack();
     *
     * private [static] void saveTo$var() {
     *   Object val[] = new Object[n];
     *   val[1] = preVar1;
     *   ...
     *   val[n] = preVarn;
     *   var.push(val);
     * }
     *
     * private [static] void restoreFrom$var() {
     *   Object val[] = (Object[]) var.pop();
     *   preVar1 = val[1];
     *   ...
     *   preVarn = val[n];
     * }
     * </pre>
     * 
     * <pre><jml>
     *  requires (\forall Object o; oldExprs.contains(o);
     *             o instanceof PreValueVars.Entry);
     * assignable \nothing;
     * </jml></pre>
     *
     * @param forStatic true if the stack and methods should be static
     * @param stackVar name of the stack field to be declared
     * @param oldExprs fields to store into and restore from the stack
     */
    private RacNode genStackMethods(boolean forStatic, String stackVar,
                                    /*@ non_null @*/ List oldExprs) {
        if (stackVar == null) {
            return null;
        }

        String saveMethod = MN_SAVE_TO + stackVar;
        String restoreMethod = MN_RESTORE_FROM + stackVar;
        // String ident = typeDecl.ident();
        String mod = forStatic ? " static " : " ";

        // generate a private stack field declaration to be used for
        // saving into, and restoring from old expression values to
        // guard againt potential recursive method calls.
        RacNode varNode = RacParser.parseMethod(
               "\n/** Generated by JML to save and restore the old " + 
               "expression values\n" +
               " * of the" + mod + "history constraints. */\n" +
               "private" + mod + "transient java.util.Stack " + stackVar  +
                  " = new java.util.Stack();");

        PreValueVars vars = new PreValueVars();
        for (Iterator i = oldExprs.iterator(); i.hasNext(); ) {
            vars.add((PreValueVars.Entry) ((RacNode)i.next()).varDecl());
        }

        // generate a save (stack push) method
        RacNode saveNode = RacParser.parseMethod(
            "\n\n/** Generated by JML to save the old expression values of\n" +
            " * the" + mod + "history constraints. */\n" +
            "private" + mod + "void " + saveMethod + "() {\n" +
            "$0" +
            "}", 
            vars.saveCode(stackVar));

        // generate a restoration (stack pop) method
        RacNode restoreNode = RacParser.parseMethod(
          "\n\n/** Generated by JML to restore the old expression values of\n"+
            " * the" + mod + "history constraints. */\n" +
            "private" + mod + "void " + restoreMethod + "() {\n" +
            "$0" +
            "}\n", 
            vars.restoreCode(stackVar));
        
        return RacParser.parseStatement("$0$1$2",varNode,saveNode,restoreNode);
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** The host type declaration containing the target constraint
     * clause to be translated by this translator.
     */
    private final /*@ spec_public non_null @*/ JmlTypeDeclaration typeDecl;
    
    /** The variable generator to be used in the translation. If
     * necessary, fresh variables are generated by this translator
     * using this variable generator during the translation.
     */
    private final /*@ spec_public non_null @*/ VarGenerator varGen;

    /** The set of (translated) old expressions appearing in the
     * static constraints.  Each old expression appearing in the
     * static constraints is translated into a sequence of statements
     * (i.e., an instance of {@link org.jmlspecs.jmlrac.RacNode}),
     * which, if executed, evaluate the old expression and store the
     * result into a freshly introduced field, called an <em>old
     * expression field</em>.  The name of the old expression field is
     * given by the <code>name</code> field of the class
     * {@link org.jmlspecs.jmlrac.RacNode} in the form of 
     * "<code>T n;\n</code>", where <code>T</code> is the type and
     * <code>n</code> is the field name.
     *
     * <pre><jml> 
     * public invariant (\forall Object o; oldExprsStatic.contains(o); 
     *   o instanceof RacNode);
     * </jml></pre>
     *
     * @see TransPostcondition#visitJmlPreExpression(JmlPreExpression)
     * @see TransPostcondition#visitJmlOldExpression(JmlOldExpression)
     */
    private /*@ spec_public non_null @*/ List oldExprsStatic;

    /** The set of (translated) old expressions appearing in the
     * non-static constraints.  Each old expression appearing in the
     * non-static constraints is translated into a sequence of statements
     * (i.e., an instance of {@link org.jmlspecs.jmlrac.RacNode}),
     * which, if executed, evaluate the old expression and store the
     * result into a freshly introduced field, called an <em>old
     * expression field</em>.  The name of the old expression field is
     * given by the <code>name</code> field of the class
     * {@link org.jmlspecs.jmlrac.RacNode} in the form of 
     * "<code>T n;\n</code>", where <code>T</code> is the type and
     * <code>n</code> is the field name.
     *
     * <pre><jml> 
     * public invariant (\forall Object o; oldExprsInstance.contains(o); 
     *   o instanceof RacNode);
     * </jml></pre>
     *
     * @see TransPostcondition#visitJmlPreExpression(JmlPreExpression)
     * @see TransPostcondition#visitJmlOldExpression(JmlOldExpression)
     */
    private /*@ spec_public non_null @*/ List oldExprsInstance;

    /** The set of private stack field names used to save and restore
     * old expression values appearing in static constraint
     * clauses. The set becomes empty if no such stack field is
     * needed, e.g., no old expression in any static constraints.
     *
     * <pre><jml> 
     * private invariant (\forall Object o; stackVarsStatic.contains(o); 
     *   o instanceof String);
     * </jml></pre>
     */
    private /*@ spec_public non_null @*/ List stackVarsStatic;

    /** The set of private stack field names used to save and restore
     * old expression values appearing in non-static constraint
     * clauses. The set becomes empty if no such stack field is
     * needed, e.g., no old expression in any non-static constraints.
     *
     * <pre><jml> 
     * private invariant (\forall Object o; stackVarsInstance.contains(o); 
     *   o instanceof String);
     * </jml></pre>
     */
    private /*@ spec_public non_null @*/ List stackVarsInstance;

    /** The names of restore methods generated so far.
     *
     * <pre><jml> 
     * private invariant (\forall Object o; restoreMethods.contains(o); 
     *   o instanceof String);
     * </jml></pre>
     */
    private /*@ spec_public non_null @*/ List restoreMethods;
}
