/*
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: TransConstructor.java,v 1.12 2002/12/14 14:53:57 davidcok Exp $
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import java.util.ArrayList;
import org.multijava.mjc.JStatement;
import org.multijava.mjc.JConstructorBlock;

/**
 * A class for translating JML annotated constructor into a RAC-enabled
 * method. The translation produces new methods and fields that must
 * be added to the target class, and also modify the body to wrap it with
 * pre and postcondition checks.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.12 $
 */

public class TransConstructor extends TransMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new <tt>TransConstructor</tt> object.
     *
     * @param varGen fresh variable name generator
     *
     * <pre><jml>
     * requires typeDecl != null && varGen != null;
     * </jml></pre>
     */
    public TransConstructor(JmlTypeDeclaration typeDecl, VarGenerator varGen) 
    {
	super(typeDecl, varGen);
    }

    // ----------------------------------------------------------------------
    // TRANSLATION
    // ----------------------------------------------------------------------

    /** 
     * Wraps the constructor with pre and postcondition checking
     * methods. The constructor body is replaced by a new block of
     * code that forwards the constructor call to a generated 
     * <tt>private</tt> method with the original constructor body performing
     * pre and postcondition checking before and after the call forwarding.
     * The new body looks something like the following:
     *
     * <pre>
     * C(T1 x1, ..., Tn xn) {
     *   checkPre$C(x1, ..., xn);
     *   try {
     *     internal$init$(x1, ..., xn);
     *   }
     *   catch (Exception e) {
     *     checkExcept$C(x1, ..., xn, e);
     *   }
     *   checkPost$C(x1, ..., xn);
     * }
     * </pre>
     *
     * <pre><jml>
     * also
     *   modifies methodDecl, newMethods;
     * </jml></pre>
     *
     */
    protected void genWrapperMethod() {
	if (genSpecTestFile && methodDecl.body().body() == null) {
	    JStatement[] js = genDelegatedMethod();
	    methodDecl.setBody(new JConstructorBlock(null,js));
	}
	ConstructorWrapper m = new ConstructorWrapper(
	    typeDecl.ident(), (JmlConstructorDeclaration)methodDecl);
	newMethods.add(m.generate());
    }



    /** This method is overwritten here. */
    protected void genStackMethod(PreValueVars vars) {
    }
}
