/* $Id: TransInvariant.java,v 1.20 2007/06/30 02:32:39 chalin Exp $
 *
 * Copyright (C) 2001, 2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import org.multijava.util.compiler.TokenReference;

/**
 * A class for translating JML invariants. The translation produces
 * two <em>invariant check methods</em>, one for static (class)
 * invariants and the other for non-static (instance) invariants.
 * Multiple invariant clauses are conjoined and <code>non_null</code>
 * fields, if exist any, are also conjoined to the invariant clauses;
 * a <code>non_null</code> annotation for a field <code>x</code> is
 * translated into a predicate <code>x != null</code>.  The generated
 * invariant check methods are returned as an object of
 * <code>RacNode</code>.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.20 $
 */

public class TransInvariant extends TransUtils {

    /**
     * Constructs a <code>TransInvariant</code> object.
     *
     * @param typeDecl host type declaration that contains the
     * invariant clauses to be translated.
     * @param varGen variable generator to be used in the translation
     *
     * <pre><jml>
     *  requires typeDecl != null && varGen != null;
     * </jml></pre>
     */
    public TransInvariant(JmlTypeDeclaration typeDecl, VarGenerator varGen)
    {
	this.typeDecl = typeDecl;
	this.varGen = varGen;
    }

    /** 
     * Generates both static and instance invariant check methods and
     * return them.
     *
     * @param invariants Invariants to be translated
     * @return an object of <code>RacNode</code> representing static and
     *         non-static invariant check methods.
     *
     * <pre><jml>
     *  requires invariants != null;
     *  ensures \result != null;
     * </jml></pre>
     */
    public RacNode translate(JmlInvariant[] invariants) 
    {
	RacNode result = translate(invariants, true);  // static inv
	result = RacParser.parseMethod("$0$1", result, // instance inv
				       translate(invariants, false)); 
	return result;
    }

    /** 
     * Generates and returns a static or an instance invariant check
     * method for the given invariant clauses,
     * <code>invariants</code>.  All invariants and
     * <code>non_null</code> annotations are conjoined and then
     * translated; a <code>non_null</code> annotation for a field
     * <code>x</code> is translated into <code>x != null</code> to be
     * conjoined to the invariant clauses.
     *
     * @param invariants Invariants to be translated
     * @param forStatic type of invariants to be translated 
     *                  (static or non-static)
     * @return an object of <code>RacNode</code> representing 
     *         the invariant check method
     *
     * <pre><jml>
     *  requires invariants != null;
     *  ensures \result != null;
     * </jml></pre>
     */
    private RacNode translate(JmlInvariant[] invariants, boolean forStatic) 
    {
	// conjoin all invariants
	JExpression root = null;
	for (int i = 0; i < invariants.length; i++) {
	    if (!isCheckable(invariants[i]) ||
                (hasFlag(invariants[i].modifiers(), ACC_STATIC) != forStatic))
		continue;
	    JExpression p = new RacPredicate(invariants[i].predicate());
	    root = root == null ? 
		p : new JConditionalAndExpression(org.multijava.util.compiler.TokenReference.NO_REF, 
						  root, p);
	}
	
	// conjoin non-null annotations if any
	JFieldDeclarationType[] fields = typeDecl.fields();
	for (int i = 0; i < fields.length; i++) {
	    JVariableDefinition varDef = fields[i].variable();
	    if ( !varDef.isDeclaredNonNull() || 
		 varDef.isStatic() != forStatic ) {
		continue;
	    }
            // build an expression of the form: x != null;
            TokenReference tref = fields[i].getTokenReference();
            JExpression left = null; // left hand of expr != (e.g. x).

            // for a model field, build model$x$T() != null; otherwise
            // (including ghost field), build x != null.
                
            class CallExpr extends JMethodCallExpression {
                CallExpr(TokenReference ref, JExpression expr, CType type) {
                    super(ref, expr, null);
                    returnType = type;
                }

                public CType getType() {
                    return returnType;
                }

                private CType returnType;
            };

	    if (hasFlag(varDef.modifiers(), ACC_MODEL)) {
                String mn = MN_MODEL + varDef.ident() +"$"+ typeDecl.ident();
                left = new CallExpr(tref, new JNameExpression(tref, mn),
                                    varDef.getType());
            } else {
                class CallExpr2 extends JMethodCallExpression {
                    CallExpr2(TokenReference ref, JExpression expr, 
                      CType type) {
                        super(ref, expr, null);
                        returnType = type;
                    }

                    public CType getType() {
                        return returnType;
                    }

                    private CType returnType;
                };
                if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
                        String fieldAcc = "_chx_get_" + varDef.ident();
                    left = new CallExpr2(tref, 
                                        new JNameExpression(tref, fieldAcc), 
                                        varDef.getType());
                } else {
                    left = new JLocalVariableExpression(tref, varDef);
                }
            }

	    // build expression of the form: x != null
	    JExpression expr = new JEqualityExpression(tref, OPE_NE, left,
                                                       new JNullLiteral(tref));
		
	    // build rac predicate (for error reporting) and conjoin it
	    expr = new RacPredicate(expr,
                                    "non_null for field '" + 
                                    fields[i].ident() + "'");
	    root = root == null ? 
		expr : new JConditionalAndExpression(tref, root, expr);
	}
	
	// translate conjoined invariants
	RacNode stmt = null;
	if (root != null) {
            RacContext ctx = RacContext.createPositive();
            
            // FRioux - Experimental test for efficiency (begin)
            
            if(!Main.racOptions.oldSemantics()){
                TransExpression2 transx = new TransExpression2(
                        VarGenerator.forMethod(varGen), ctx, root, "rac$b", typeDecl, "JMLInvariantError");
                stmt = transx.stmt(true);
            }
            else{
                TransPredicate trans = new TransPredicate(
                        VarGenerator.forMethod(varGen), ctx, root, "rac$b", typeDecl);
                stmt = trans.wrappedStmt();
            }
            
            // FRioux - Experimental test for efficiency (end)          
            
	}

	// generate invariant method
	InvariantMethod m = new InvariantMethod(forStatic, typeDecl);
	return (RacNode)m.generate(stmt);
    }

    /**
     * Returns true if the given invariant clause is checkable. The
     * clause is checkable if it is not a redundant specification or
     * the command line option "noredundancy" is not turned on.
     */
    private static boolean isCheckable(JmlInvariant inv) {
        return !inv.isRedundantly() || !Main.racOptions.noredundancy();
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------
    
    /** variable generator to be used in the translation */
    private final /*@ non_null @*/ VarGenerator varGen;

    /** taget type declaration that has this invariant specification. */
    private final /*@ non_null @*/ JmlTypeDeclaration typeDecl;
}

