/* $Id: TransMethod.java,v 1.68 2007/07/19 10:51:36 f_rioux Exp $
 *
 * Copyright (C) 2001, 2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;
import org.multijava.util.compiler.TokenReference;
import java.util.*;

/**
 * A class for translating JML annotated Java methods into RAC-enabled
 * methods. The translation produces new methods and fields that must
 * be added to the target class, and also renames the original method.
 * 
 * <p> A <code>non_null</code> annotation to a formal parameter is
 * interpreted as a part of preconditions. That is, a
 * <code>non_null</code> annotation to a formal parameter, say
 * <code>x</code>, has the same effect as conjoining the predicate
 * <code>x != null</code> to the precondition of the method.  </p>
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.68 $
 */

// !FIXME! replace rac$e and rac$b with constants!

public class TransMethod extends TransUtils {

    /** True if we are generating a test wrapper for a spec file (and we do 
	not have the original source).
    */
    protected boolean genSpecTestFile = false;
    
    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Constructs a new <code>TransMethod</code> object.
     * 
     * @param typeDecl target type declaration whose methods are to be 
     *                 translated.
     * @param varGen variable name generator
     * 
     * <pre><jml>
     * requires typeDecl != null && varGen != null;
     * </jml></pre>
     */
    public TransMethod(JmlTypeDeclaration typeDecl, VarGenerator varGen) 
    {
	this.typeDecl = typeDecl; 
	this.varGen = varGen;
	newMethods = new ArrayList();
	newFields = new ArrayList();
	genSpecTestFile = Main.isSpecMode(typeDecl);
    }

    // ----------------------------------------------------------------------
    // ACCESSOR METHODS
    // ----------------------------------------------------------------------

    /** Returns true if any new field declarations has been generated
     * as the result of translation.
     *
     * <pre><jml>
     * ensures  \result == newFields.size() > 0;
     * </jml></pre>
     */
    public boolean hasNewFields() {
	return newFields.size() > 0;
    }

    /** Returns true if any new method declarations has been generated
     * as the result of translation.
     *
     * <pre><jml>
     * ensures  \result == newMethods.size() > 0;
     * </jml></pre>
     */
    public boolean hasNewMethods() {
	return newMethods.size() > 0;
    }

    /** Returns a list of new field declarations created by this
     * translation, which need be added to the target class or interface.
     *
     * <pre><jml>
     * ensures  \result != null && (\forall Object o; \result.contains(o);
     *   (o instanceof String) && 
     *   (* o has the form: "[static] T v = init;\n" *));
     * </jml></pre>
     */
    public List newFields() {
	return newFields;
    }

    /** Returns a list of new method declaratons created by this
     * transformation, and need be added to the target class or instance. 
     *
     * <pre><jml>
     * ensures  \result != null && (\forall Object o; \result.contains(o);
     *   (o instanceof JmlMethodDeclaration));
     * </jml></pre>
     */
    public List newMethods() {
	return newMethods;
    }

    // ----------------------------------------------------------------------
    // TRANSLATION
    // ----------------------------------------------------------------------

    /** Peforms translation leaving the results in appropriate instance
     * variables.
     *
     * @param methodDecl method to translate
     *
     * <pre><jml>
     * requires methodDecl != null && typeDecl.methods().contains(methodDecl);
     * assignable newMethods, newFields, this.methodDecl, desugaredSpec;
     * assignable stackVar;
     * </jml></pre>
     */
    public void perform(JmlMethodDeclaration methodDecl) 
    {
	this.methodDecl = methodDecl;

	// desugar method specification 
	if (methodDecl.hasSpecification()) {
	    DesugarSpec desugar = new DesugarSpec();
	    desugaredSpec = 
		desugar.perform(methodDecl.methodSpecification(),
                                methodDecl.getExceptions());
	}
        // generate assertion check methods such as pre and postcondition
        // check methods.
	PreValueVars vars = genAssertionMethods();

        // generate a pair of (pre-values) stack manipulation methods for
        // potential recusive method calls.
        genStackMethod(vars);

        // This method must be called last because it produces some
        // side-effects, e.g., making it a private internal method.
	genWrapperMethod();
    }

    /** 
     * Generates pre/postcondition check methods. The generated
     * methods and any fields (generated as side-effects) are added to
     * the instance fields <code>newMethods</code> and
     * <code>newFields</code> respectively. Returned is the set of
     * (generated) private fields that need to be saved before, and
     * restored after the execution of the method body, due to
     * the possiblity of recursive method calls.
     * 
     * <pre><jml>
     * requires (* desugaredSpec has desugared method spec *);
     * //assignable newMethods, newFields, stackVar;
     * ensures \result != null;
     * </jml></pre>
     */
    protected PreValueVars genAssertionMethods()
    {
	// fresh variable generators
	VarGenerator preVarGen = VarGenerator.forMethod(varGen);
	VarGenerator postVarGen = VarGenerator.forMethod(varGen);
	VarGenerator epostVarGen = VarGenerator.forMethod(varGen);

	// translation result accumulators
	RacNode preStmt = null;
	RacNode postStmt = null;
	RacNode epostStmt = null;

	// A list of String, representing the names of private boolean
	// fields to be generated to store the result of evaluating
	// preconditions in the pre-state.
	List preVars = new ArrayList();

	// expression to check non_null annotations of parameters and result.
	JExpression preNonNullExpr = null;
	JExpression postNonNullExpr = null;
	if(!methodDecl.isOverriding()) {
	    // The method is not overriding. Desugar nullity modifiers
	    // into the assertions that they will contribute to the
	    // pre- and post-conditions.
	    preNonNullExpr = preNonNullAnnotations();
	    postNonNullExpr = postNonNullAnnotation();
	} else {
	    // This method is overriding. If it does not have a
	    // specification and it does not have explicit nullity
	    // modifiers (either in the parameters or the return type)
	    // then we inherit the supertype contract, otherwise we
	    // must extend the contract by adding a specification case
	    // containing the desugaring of the nullity modifiers.
	    // Note that we need to consider the special case where
	    // the only explicit nullity modifiers are "nullable": in
	    // this case we need to ensure that an explicit
	    // precondition of "true" is given instead of leaving
	    // preNonNullExpr == null and postNonNullExpr == null.
	    // Should we leave preNonNullExpr == null and
	    // postNonNullExpr == null then we would fail to extend
	    // this methods specification.
	    //
	    // For overriding methods we generate a precondition (to
	    // be added to the method contract) based on the nullity
	    // modifiers iff:
	    // - this method has an explicit specification, or
	    // - nullity modifiers were explicitly given for the parameters
	    //   or return type.
	    boolean paramsHaveExplicitNullityMods = paramsHaveExplicitNullityModifiers();

	    if(methodHasRealExplicitSpecs()
	       || methodReturnTypeHasExplicitNullityModifiers()
	       || paramsHaveExplicitNullityMods)
	    {
		preNonNullExpr = preNonNullAnnotations();
		postNonNullExpr = postNonNullAnnotation();

		if(preNonNullExpr == null 
		   // && postNonNullExpr == null 
		   && (methodReturnTypeHasExplicitNullityModifiers()
		       || paramsHaveExplicitNullityMods)) 
		{
		    // Ensure that an extending spec case will be
		    // created by setting the precondition to "true".
		    preNonNullExpr = new JBooleanLiteral(methodDecl.getTokenReference(), true);
		    // JExpression expr = new JBooleanLiteral(methodDecl.getTokenReference(), true);
		    // preNonNullExpr = new RacPredicate(expr, "nullity for " + methodDecl.ident());
		}
	    }
	}

	// A list of RacNode corresponding to old expressions found in
	// normal and exceptional postconditions of this method
	// specification, collected to evaluate them in the pre-state
	// and store the results to the corresponding, generated
	// private fields. The declaration of the private field (of
	// the type PreValueVars.Entry) is piggybacked through the
	// varDecl field of the RacNode.
	List oldExprs = new ArrayList(); 

	// A list of RacNode corresponding to spec variables found in
	// the method specification, collected to declare the
	// corresponding private fields and to evaluate the
	// initializers in pre-state. Each entry (i.e., an object of
	// RacNode), if executed, evaluates the initializer and sets
	// the corresponding private field. The declaration of the
	// private field (of type PreValueVars.Entry) is piggybacked
	// through the varDecl field of RacNode.
	List specVars = new ArrayList();

	boolean hasPrecondition = false;
	boolean hasPostcondition = false;
        RacContext ctx = RacContext.createPositive();

        SpecCaseCollector specs = new SpecCaseCollector(desugaredSpec);
        Iterator iter = specs.iterator();
        while (iter.hasNext()) {
            SpecCase c = (SpecCase) iter.next();
            
            // translate spec vars if any
            if (c.hasSpecVarDecls()) {
                specVars.addAll(translateSpecVarDecls(
                                    c.specVarDecls(), preVarGen));
            }

            // translate precondition if exists
            String preVar = null;
            JExpression pred = c.precondition();
            if (preNonNullExpr != null) {
                // make non_null annotations a part of precondition
		if ( pred instanceof RacPredicate) {
		    if (!((RacPredicate)pred).isTrueLiteral())
			((RacPredicate)pred).countCoverage();
		}
		if (preNonNullExpr instanceof RacPredicate) ((RacPredicate)preNonNullExpr).countCoverage();
                pred = pred == null ? preNonNullExpr :
                    new JConditionalAndExpression(pred.getTokenReference(),
                                                  preNonNullExpr, pred);
            }
            if (pred != null) {
		if ( pred instanceof RacPredicate) {
		    if (!((RacPredicate)pred).isTrueLiteral())
			((RacPredicate)pred).countCoverage();
		}
                hasPrecondition = true;
                preVar = preVarGen.freshPreVar();
                preVars.add(preVar);
                
                // FRioux - Experimental test for efficiency (begin)
                             
                if(!Main.racOptions.oldSemantics()){
                    TransExpression2 transx = new TransExpression2(preVarGen, ctx, pred, preVar, typeDecl, "JMLEntryPreconditionError");
                    preStmt = RacParser.parseStatement(
                            (preStmt == null ? "" : "$0\n") +
                            "$1\n" +
                            "rac$b = rac$b || " + preVar + ";",
                            preStmt, transx.stmt(true));     
                }
                else{
                    TransPredicate trans = 
                        new TransPredicate(preVarGen, ctx, pred, preVar, typeDecl);
                    preStmt = RacParser.parseStatement(
                            (preStmt == null ? "" : "$0\n") +
                            "$1\n" +
                            "rac$b = rac$b || " + preVar + ";",
                            preStmt, trans.wrappedStmt());
                }
                
                // FRioux - Experimental test for efficiency (end)
            }
            
            // translate postcondition if exists
            pred = c.postcondition();
            if (postNonNullExpr != null) {
                pred = pred == null ? postNonNullExpr :
                    new JConditionalAndExpression(pred.getTokenReference(),
                                                  postNonNullExpr, pred);
            }
            if (pred != null) {
                hasPostcondition = true;
                String var = postVarGen.freshPostVar();
                
                // FRioux - Experimental test for efficiency (begin)
                
                if(!Main.racOptions.oldSemantics()){
                	TransPostExpression2 trans = new TransPostExpression2(
                			postVarGen, ctx, preVarGen, methodDecl.isStatic(),
                			pred, var, typeDecl, "JMLExitNormalPostconditionError");		
                	RacNode stmt = trans.stmt(true); // !FIXME! ctx?
                	postStmt = RacParser.parseStatement(
                			(postStmt == null ? "" : "$0\n") + // prev. post
                			"boolean " + var + " = true;\n" +  // result var for this
                			(preVar == null ?             // if no precond?
                					"$1\n" :                     //   always eval post
                						("if (" + preVar + ") {\n" + // else
                								"$1\n" +                    //   eval only if pre holds 
                						"}\n")) +
                						"rac$b = rac$b && " + var + ";", // accumulate result
                						postStmt, 
                						(preVar == null ? stmt : stmt.incrIndent())); 
                	oldExprs.addAll(trans.oldExprs());
                }
                else{
                	TransPostcondition trans = createPostconditionTranslator(
                			postVarGen, ctx, preVarGen, methodDecl.isStatic(),
                			pred, var);
                	RacNode stmt = trans.wrappedStmt(); // !FIXME! ctx?
                	postStmt = RacParser.parseStatement(
                			(postStmt == null ? "" : "$0\n") + // prev. post
                			"boolean " + var + " = true;\n" +  // result var for this
                			(preVar == null ?             // if no precond?
                					"$1\n" :                     //   always eval post
                						("if (" + preVar + ") {\n" + // else
                								"$1\n" +                    //   eval only if pre holds 
                						"}\n")) +
                						"rac$b = rac$b && " + var + ";", // accumulate result
                						postStmt, 
                						(preVar == null ? stmt : stmt.incrIndent())); 
                	oldExprs.addAll(trans.oldExprs());
                }
                
                // FRioux - Experimental test for efficiency (end)
            }
            
            // translate exceptional postcondition if exists
            if (c.hasExceptionalPostcondition()) {
                JmlSignalsClause[] s = c.exceptionalPostcondition();
                for (int i = 0; i < s.length; i++) {
                    if (s[i].hasPredicate() && !s[i].isNotSpecified()) {
                        epostStmt = translateSignalsClause(postVarGen, 
                            preVarGen, s[i], preVar, epostStmt, oldExprs);
                    }
                    s[i].type();
                }
            }
        }

        // handle the case when no pre/postcondition are specified
        // but non_null annotations are specified.
        String preVar = null;
        if (!hasPrecondition && preNonNullExpr != null) {
            hasPrecondition = true;
            preVar = preVarGen.freshPreVar();
            preVars.add(preVar);
	    if (preNonNullExpr instanceof RacPredicate) ((RacPredicate)preNonNullExpr).countCoverage();
        
	    // FRioux - Experimental test for efficiency (begin)
	    
	    if(!Main.racOptions.oldSemantics()){
	        TransExpression2 transx = new TransExpression2(preVarGen, ctx, preNonNullExpr, preVar, typeDecl, "JMLEntryPreconditionError");      
	        preStmt = RacParser.parseStatement(
	                "$0\n" +
	                "rac$b = " + preVar + ";",
	                transx.stmt(true));
	    }
	    else{
	        TransPredicate trans = 
	            new TransPredicate(preVarGen, ctx, preNonNullExpr, preVar, typeDecl);
	        preStmt = RacParser.parseStatement(
	                "$0\n" +
	                "rac$b = " + preVar + ";",
	                trans.wrappedStmt());
	    }                      
	    // FRioux - Experimental test for efficiency (end)
           
        }
        if (!hasPostcondition && postNonNullExpr != null) {
            String var = postVarGen.freshPostVar();
            
            // FRioux - Experimental test for efficiency (begin)
            
            if(!Main.racOptions.oldSemantics()){
            	TransPostExpression2 trans = 
            		new TransPostExpression2(postVarGen, ctx, preVarGen, 
            				methodDecl.isStatic(), 
            				postNonNullExpr, var, typeDecl, "JMLExitNormalPostconditionError");
            	postStmt = RacParser.parseStatement(
            			"boolean " + var + " = false;\n" +
            			"$0\n" +
            			(preVar == null ?  "" :
            				var + " = " + preVar + " ? " + var + " : true;\n") +
            				"rac$b = " + var + ";",
            				trans.stmt(true));
            }
            else{
            	TransPostcondition trans = 
            		new TransPostcondition(postVarGen, ctx, preVarGen, 
            				methodDecl.isStatic(), 
            				postNonNullExpr, var, typeDecl);
            	postStmt = RacParser.parseStatement(
            			"boolean " + var + " = false;\n" +
            			"$0\n" +
            			(preVar == null ?  "" :
            				var + " = " + preVar + " ? " + var + " : true;\n") +
            				"rac$b = " + var + ";",
            				trans.wrappedStmt());	
            }
            
            // FRioux - Experimental test for efficiency (end)
        }

        // accumlate pre-state variables that need to be saved and
        // restored before calling the original method, due to
        // potential recursive calls.
        PreValueVars backupVars = new PreValueVars();
        for (Iterator i = specVars.iterator(); i.hasNext(); ) { // old vars
            backupVars.add(((RacNode)i.next()).varDecl());
	}
        for (Iterator i = oldExprs.iterator(); i.hasNext(); ) { // old exprs
            backupVars.add(((RacNode)i.next()).varDecl());
	}
        boolean isStatic = methodDecl.isStatic();
	for (Iterator i = preVars.iterator(); i.hasNext(); ) { // preconditions
            backupVars.addBoolean(isStatic, (String) i.next());
	}
        // note that old expressions appearing in constraint clauses
        // are handled separately by constraint check methods and old
        // expression evaluation methods; and thus, they have separate
        // stack variables and access methods.

        // determine the names of stack manipulation methods
        String saveMethod = null;
        String restoreMethod = null;
        if (!(backupVars.isEmpty() ||
              (methodDecl instanceof JmlConstructorDeclaration))) {
            stackVar = varGen.freshStackVar();
            saveMethod = MN_SAVE_TO + stackVar;
            restoreMethod = MN_RESTORE_FROM + stackVar;
        }

	// generate a prcondition check method
	PreconditionMethod pm = new PreconditionMethod(typeDecl, 
            methodDecl, hasPrecondition, saveMethod);
	newMethods.add(pm.generate(preStmt, specVars, oldExprs));
	specVars.addAll(oldExprs);

	// generate a normal postcondition check method
	AssertionMethod m = new PostconditionMethod(typeDecl, 
	    methodDecl, restoreMethod);
	newMethods.add(m.generate(postStmt));

	// generate an exceptional postcondition check method
	m = new ExceptionalPostconditionMethod(typeDecl, methodDecl, 
            restoreMethod);
	newMethods.add(m.generate(epostStmt));
	
	// add any new field declarations (e.g., for preconditions,
        // spec vars, old expressions) to the type declaration.
	String mod = isStatic ? "private static " : "private ";
	for (Iterator i = preVars.iterator(); i.hasNext(); ) {
            String name = (String) i.next();
	    newFields.add(mod + "transient boolean " + name + ";\n");
	}
	for (Iterator i = specVars.iterator(); i.hasNext(); ) {
            newFields.add(((RacNode)i.next()).varDecl() + "\n");
	}
        return backupVars;
    }

    /** 
     * Creates an appropriate postcondition translator. This is a hook
     * method for subclasses to override to use different translators.
     */
    protected TransPostcondition createPostconditionTranslator(
        VarGenerator postVarGen,
        RacContext ctx,
        VarGenerator preVarGen,
        boolean isStatic,
        JExpression pred,
        String var) {

        return new TransPostcondition(postVarGen, ctx, preVarGen, 
                                      isStatic, pred, var, typeDecl);
    }
    
    /** Translates specification variable declarations. Initializers
     * are translated and returned with their new private field
     * declarations. The caller is responsible for properly declaring
     * the fields for the specification variables and execute the
     * translated initializers in the appropriate place, e.g., in the
     * method prolog.
     *
     * <pre><jml>
     * requires varDecls != null && varGen != null;
     * ensures \fresh(\result) && (\forall Object o; \result.contains(o);
     *         o instanceof RacNode);
     * </jml></pre>
     *
     * @param varDecls declarations to be translated
     * @param varGen variable generator for fresh variable names
     * @return list of <code>RacNode</code>s, representing translated
     *         initializers with the corresponding new field 
     *         declarations. The field declarations are
     *         piggybacked in the <code>name</code> fields of nodes.
     * @see org.jmlspecs.jmlrac.runtime.JMLRacValue     
     */
    protected List translateSpecVarDecls(JmlSpecVarDecl[] varDecls,
                                         VarGenerator varGen) 
    {
	// list of RacNode's representing translated initializers
	List result = new ArrayList();

	for (int i = 0; i < varDecls.length; i++) {
	    // already processed, i.e., when desugared?
	    if (varDecls[i].racGenerated()) {
		continue;
            }
	    varDecls[i].setRacGenerated();

            if (varDecls[i] instanceof JmlLetVarDecl) {
                result.addAll(translateLetVarDecl(varDecls[i], varGen));
            } else {
		result.addAll(translateForAllVarDecl(varDecls[i], varGen));
            } 
        }

        return result;
    }

    /** Translates the given old variable declarations. The
     * initializers are translated and and returned with new instance
     * variable declarations. The caller is responsible for properly
     * declaring the new instance variables and execute the code to
     * stores the values of initializers to the new fields.
     *
     * <pre><jml>
     * requires varDecl != null && varGen != null;
     * ensures \fresh(\result) && (\forall Object o; \result.contains(o);
     *         o instanceof RacNode);
     * </jml></pre>
     *  
     * @param varDecl declaration to be translated
     * @param varGen variable generator for fresh variable names
     * @return list of <code>RacNode</code>s representing translated
     *         code with the new instance field declarations. 
     *         The field declarations are piggybacked in the 
     *         <code>name</code> fields of RacNode.
     * @see org.jmlspecs.jmlrac.runtime.JMLRacValue
     */
    protected List translateLetVarDecl(JmlSpecVarDecl varDecl,
                                       VarGenerator varGen) {
	// list of RacNodes representing translated code
	List result = new ArrayList();

        JmlVariableDefinition[] varDefs = (JmlVariableDefinition[])
            ((JmlLetVarDecl) varDecl).specVariableDeclarators();

        for (int j = 0; j < varDefs.length; j++) {
            if (!varDefs[j].hasInitializer()) {
                // jml gurantees this not to reach!
                //@ unreachable;
                fail("Reached unreachable!");
            }

            // create a fresh variable unique in the class scope. This
            // prevents name crashes when the variable becomes a new
            // privat field; i.e, name crashes among let variables
            // and existing fields.
            String oldVar = varGen.freshOldVar();
            CType type = varDefs[j].getType();

            // change the identifier so that from now on it is printed as
            // code for unwrapping the new field.
//             varDefs[j].setIdent(
//                 TransUtils.unwrapObject(type, oldVar + ".value()"));
            varDefs[j].setRacField(oldVar);

            // generates code that evaluates the initializer

            // disable contextual interpretation so that it is handled
            // when the field variable is actually referenced.
            RacContext ctx = RacContext.createNeutral();
            String var = varGen.freshVar();
            String val = TransExpression.defaultValue(type);
            TransExpression trans = new TransExpression(varGen, ctx,
                varDefs[j].expr(), var, typeDecl);
            RacNode node = trans.stmt();

            // add code that stores the initializer value to the new field
            node = RacParser.parseStatement(
                "try {\n" +
                "  " + toString(type) + " " + var + " = " + val + ";\n" +
                "$0\n" +
                "  $1 = " +TN_JMLVALUE+ ".ofObject(" +
                    TransUtils.wrapValue(type, var)+ ");\n" +
                "}\n" +
                "catch (JMLNonExecutableException jml$e0) {\n" +
                "  $1 = " +TN_JMLVALUE+ ".ofNonExecutable();\n" +
                "}" +
                "catch (java.lang.Exception jml$e0) {\n" +
                "  $1 = " +TN_JMLVALUE+ ".ofUndefined();\n" +
                "}", 
                node.incrIndent(), oldVar);

            // piggy-back a new field declaration to the translated code
            node.setVarDecl(PreValueVars.createJmlValueVar(
                                methodDecl.isStatic(), oldVar));
            result.add(node);
	}

	return result;
    }

    /** Translates the given forall variable declarations. They are 
     * translated into the non-executable values, and and returned with
     * their corresponding instance variable declarations. The caller
     * is responsible for properly declaring the instance variables and
     * execute the code stores non-executable values to the fields.
     *
     * <pre><jml>
     * requires varDecl != null && varGen != null;
     * ensures \fresh(\result) && (\forall Object o; \result.contains(o);
     *         o instanceof RacNode);
     * </jml></pre>
     *  
     * @param varDecl declaration to be translated
     * @param varGen variable generator for fresh variable names
     * @return list of <code>RacNode</code>s representing translated
     *         code with the corresponding instance field declarations. 
     *         The field declarations are piggybacked in the 
     *         <code>name</code> fields of RacNode.
     * @see org.jmlspecs.jmlrac.runtime.JMLRacValue
     */
    protected List translateForAllVarDecl(JmlSpecVarDecl varDecl,
                                          VarGenerator varGen) {
	// list of RacNodes representing translated code
	List result = new ArrayList();

        JVariableDefinition[] varDefs = 
            ((JmlForAllVarDecl)varDecl).quantifiedVarDecls();
        for (int j = 0; j < varDefs.length; j++) {
            // create a fresh variable unique in the class scope. This
            // prevents name crashes when the variable becomes a new
            // privat field; i.e, name crashes among forall variables
            // and existing fields.
            String oldVar = varGen.freshOldVar();

            // generates code that stores the non-executable value to the
            // fresh field.
            CType type = varDefs[j].getType();
            RacNode node = RacParser.parseStatement(
                "if (true) {\n" +
                "  " + oldVar + " = " +TN_JMLVALUE+ ".ofNonExecutable();\n" +
                "}");

            // piggy-back a new field declaration to the code
            node.setVarDecl(PreValueVars.createJmlValueVar(
                                methodDecl.isStatic(), oldVar));

            // change the identifier so that it is now printed as
            // code for unwrapping the new field.
            varDefs[j].setIdent(
                TransUtils.unwrapObject(type, oldVar + ".value()"));

            result.add(node);
        }
        return result;
    }

    /** Translates the given <code>signals</code> clause. The overall
     * translation rule for exceptional postconditions is roughly
     * defined as follows.
     * 
     * <pre>
     *  [[signals (Ei ei) Pi]] =
     *    if (e instanceof Ei) {
     *       Ei ei = (Ei) e;
     *       boolean vi = false;
     *       [[Pi, vi]]
     *       b = b && vi;
     *    }
     *    check(b);
     *    if (e instanceof java.lang.RuntimeException)
     *       throw (java.lang.RuntimeException) e;
     *    else if (e instanceof java.lang.Error)
     *       throw (java.lang.Error) e;
     *    if (e instanceof CEi)
     *       throw (CEi) e;
     * </pre>
     *
     * where the variable <code>e</code> refers to actual exceptions thrown
     * and <code>b</code> conjoins the results of evaluating all exceptional
     * predicates. Both variables are defined in the body of exceptional
     * postcondition checking method. If the exceptional postcondition
     * is satisfied, the original exception is rethrown at the end; 
     * <code>CEi</code> denotes a checked exception that is listed in the
     * method declaration.
     *
     * <p>
     * This method method produces the first <code>if</code> statement
     * (of the translation rule) corresponding to the given 
     * <code>signals</code> clause. The method accumulates the generated
     * <code>if</code> statement by appending it to the argument 
     * <code>stmt</code>. The accumulated <code>if</code> statements is
     * is returned as the result. The method also accumulates the translated 
     * old expressions appearing in the given <code>signals</code> clause
     * to the argument <code>oldExprs</code>. For the translation, the 
     * method uses the variable generator <code>varGen</code> to generate
     * fresh variables and the variable generator <code>oldVarGen</code>
     * for fresh old variables.
     * </p>
     *
     * @param varGen variable generator for fresh variables
     * @param oldVarGen variable generator for old variables
     * @param sigClause the signals clause to translate
     * @param preVar precondition variable
     * @param stmt accumlated result of signals clauses translation
     * @param oldExprs old expressions accumulated so far
     *
     * @see ExceptionalPostconditionMethod#generate(RacNode)
     * 
     * <pre><jml>
     * requires varGen != null && oldVarGen != null && sigClause != null
     *   && preVar != null && oldExprs != null;
     * modifies oldExprs.theList;
     * ensures \result != null;
     * </jml></pre>
     *
     */
    protected RacNode translateSignalsClause(VarGenerator varGen,
    		VarGenerator oldVarGen, JmlSignalsClause sigClause, 
    		String preVar, RacNode stmt, List oldExprs)
    {
    	CType type = sigClause.type();
    	String ident = sigClause.ident();
    	JmlPredicate pred = sigClause.predOrNot();
    	String var = varGen.freshPostVar();
    	RacContext ctx = RacContext.createPositive();
    	
    	// translate predicate of signals clause
    	
    	// FRioux - Experimental test for efficiency (begin)
    	// TODO lower the amount of code duplication
    	
    	if(!Main.racOptions.oldSemantics()){
    		TransPostExpression2 trans = 
    			new TransPostExpression2(varGen, ctx, oldVarGen,
    					methodDecl.isStatic(), pred, var, typeDecl,  "JMLExitExceptionalPostconditionError");
    		RacNode thisStmt = trans.stmt(true).incrIndent();
    		
    		// compose error message for later use
    		String msg = "\"\\t" +
    		escapeString(pred.getTokenReference().toString()) +
    		", when\" + ";
    		String dterms = "";
//    		if (trans.hasDiagTerms()) {
//    			String dvar = varGen.freshVar();
//    			dterms = trans.diagTerms(dvar);
//    			msg += dvar + " + ";
//    		}
    		msg += "\"\\n\\t\\t" + // tabbing
    		((ident == null) ? "thrown" : ("'" + ident + "'")) + 
    		" is \" + rac$e";
    		
    		stmt = RacParser.parseStatement(
    				// concatenate accumulated signals (xpost) if any
    				(stmt == null ? "" : "$0\n") + 
    				
    				// evaludate predicate only if the thrown exception 
    				// (rac$e) is of proper type, i.e., this clause's type
    				"if (rac$e instanceof " + type + ") {\n" + 
    				
    				// if necessary, initialize exception variable (ident).
    				(ident == null ? "" :
    					"  " + type + " " + ident + " = (" + type + ") rac$e;\n") +
    					
    					// result var for this xpost
    					"  boolean " + var + " = true;\n" + 
    					
    					// if has no precond, then always eval xpost
    					(preVar == null ?  
    							"$1\n" :           
    								// else eval xpost only if precond holds
    								("  if (" + preVar + ") {\n" + 
    										"$1\n" +                   
    								"  }\n")) + 
    								
    								// if xpost fails, add error messages
    								"  if (!" + var + ") {\n" +   
    								dterms +
    								"     " + VN_ERROR_SET + ".add(" + msg + ");\n" +
    								"  }\n" +
    								
    								// accumulates the result
    								"  rac$b = rac$b && " + var + ";\n" +
    								"}",
    								stmt, (preVar == null ? thisStmt : thisStmt.incrIndent()));
    		
    		oldExprs.addAll(trans.oldExprs());     	
    	}
    	else{
    		TransPostcondition trans = 
    			new TransPostcondition(varGen, ctx, oldVarGen,
    					methodDecl.isStatic(), pred, var, typeDecl);
    		RacNode thisStmt = trans.stmt().incrIndent();
    		
    		// compose error message for later use
    		String msg = "\"\\t" +
    		escapeString(pred.getTokenReference().toString()) +
    		", when\" + ";
    		String dterms = "";
    		if (trans.hasDiagTerms()) {
    			String dvar = varGen.freshVar();
    			dterms = trans.diagTerms(dvar);
    			msg += dvar + " + ";
    		}
    		msg += "\"\\n\\t\\t" + // tabbing
    		((ident == null) ? "thrown" : ("'" + ident + "'")) + 
    		" is \" + rac$e";
    		
    		stmt = RacParser.parseStatement(
    				// concatenate accumulated signals (xpost) if any
    				(stmt == null ? "" : "$0\n") + 
    				
    				// evaludate predicate only if the thrown exception 
    				// (rac$e) is of proper type, i.e., this clause's type
    				"if (rac$e instanceof " + type + ") {\n" + 
    				
    				// if necessary, initialize exception variable (ident).
    				(ident == null ? "" :
    					"  " + type + " " + ident + " = (" + type + ") rac$e;\n") +
    					
    					// result var for this xpost
    					"  boolean " + var + " = true;\n" + 
    					
    					// if has no precond, then always eval xpost
    					(preVar == null ?  
    							"$1\n" :           
    								// else eval xpost only if precond holds
    								("  if (" + preVar + ") {\n" + 
    										"$1\n" +                   
    								"  }\n")) + 
    								
    								// if xpost fails, add error messages
    								"  if (!" + var + ") {\n" +   
    								dterms +
    								"     " + VN_ERROR_SET + ".add(" + msg + ");\n" +
    								"  }\n" +
    								
    								// accumulates the result
    								"  rac$b = rac$b && " + var + ";\n" +
    								"}",
    								stmt, (preVar == null ? thisStmt : thisStmt.incrIndent()));
    		
    		oldExprs.addAll(trans.oldExprs());     	
    	}
    	
    	// FRioux - Experimental test for efficiency (end)
    	
    	return stmt;
    }

    /** 
     * Renames the original method and generate a wrapper method that 
     * calls the original method wrapped with pre and postcondition
     * checking. The body of the wrapper method looks like the 
     * following.
     *
     * <pre>
     * T m(T1 x1, ..., Tn xn) {
     *   checkPre$m(x1, ..., xn);
     *   T r = init_T;
     *   try {
     *     r = orig$m(x1, ..., xn);
     *   }
     *   catch (Exception e) {
     *     checkExcept$m(x1, ..., xn, e);
     *   }
     *   checkPost$m(x1, ..., xn, r);
     * }
     * </pre>
     */

    /*@ old boolean isAbstract = methodDecl.body() == null;
      @ {|
      @   requires isAbstract;
      @   modifies \nothing;
      @   ensures true; 
      @ also
      @   old String mn = methodDecl.ident();
      @   requires !isAbstract;
      @   modifies methodDecl, newMethods;
      @   ensures methodDecl.ident().equals(MN_INTERNAL + mn) &&
      @           newMethods.size() == \old(newMethods.size() + 1);
      @ |}
      @*/
    protected void genWrapperMethod()
    {
	if (genSpecTestFile && methodDecl.body() == null && !hasFlag(methodDecl.modifiers(),ACC_ABSTRACT)) {
	    genDelegatedMethod();
	}
	if (methodDecl.body() != null) {
	    WrapperMethod m = methodDecl.isFinalizer() ?
		new FinalizerWrapper(typeDecl.ident(), methodDecl) :
		new WrapperMethod(typeDecl.ident(), methodDecl);
	    newMethods.add(m.generate());
	}
    }

    /** Generates a method body for a non-abstract method that does not have
	an implementation; it is used to generate test fixtures for
	specification files in the absence of .java implementations.
    */
    //@ assignable methodDecl;
    protected JStatement[] genDelegatedMethod() {
	String typename = typeDecl.getCClass().qualifiedName().replace('/','.');
	JFormalParameter[] params = methodDecl.parameters();
	JExpression[] args = new JExpression[params.length];
	for (int i = 0; i< params.length; ++i) {
		String n = params[i].ident();
		args[i] = new JNameExpression(null,n);
	}
	String qn = typeDecl.getCClass().qualifiedName();
	JExpression prefix = null;
	JExpression m;
	if (methodDecl.isConstructor()) {
	    if (!TransType.genWrapperClass) 
	       m = new JAssignmentExpression(null,
		    new JNameExpression(null,"delegee"),
		    new JNewObjectExpression(null,typeDecl.getCClass().getType(),null,args)
		);
	    else
	       m = new JAssignmentExpression(null,
		    new JNameExpression(null,"delegee"),
		    new JNewObjectExpression(null,CTopLevel.getTypeRep("Wrapper",false),null,args)
		);
	} else if (methodDecl.isStatic()) {
	    int p;
	    while ((p=qn.indexOf('/')) != -1 ) {
		    prefix = new JNameExpression(null,prefix,qn.substring(0,p));
		    qn = qn.substring(p+1);
	    }
	    prefix = new JNameExpression(null,prefix,qn);
	    m = new JMethodCallExpression(null,
		new JNameExpression(null,prefix,methodDecl.ident()),args);
	} else {
	    prefix = new JMethodCallExpression(null,
		new JNameExpression(null,"delegee_" + typeDecl.ident()),new JExpression[]{});
	    //prefix = new JNameExpression(null,"delegee");
	    m = new JMethodCallExpression(null,
		new JNameExpression(null,prefix,methodDecl.ident()),args);
	}
		
	JStatement[] js;
	if (methodDecl.returnType().isVoid()) {
		js = new JStatement[]{new JExpressionStatement(null,m,null) };
	} else if (false && methodDecl.returnType().toString().equals(typename)) {
		m = new JMethodCallExpression(null,
			new JNameExpression(null,"make"),new JExpression[]{m});
		js = new JReturnStatement[]{new JReturnStatement(null,m,null) };
	} else {
		js = new JReturnStatement[]{new JReturnStatement(null,m,null) };
	}
	methodDecl.setBody(new JBlock(null,js, null));
	methodDecl.setModifiers(methodDecl.modifiers() & ~RacConstants.ACC_NATIVE);
	return js;
    }
    
    /** 
     * Generates a private stack field declaration and a pair of push
     * and pop methods. The stack is used to save the pre-state values
     * such as preconditions, old variables, and old expressions into,
     * and restore them from. The reason is to pass correct pre-state
     * values to the post-state assertion check methods such as the
     * postcondition check method in the presence of recursive method
     * calls. The generated stack field and methods are added to the
     * type declaration being translated. The code has the following
     * form.
     *
     * <pre>
     * private [static] Stack var = new Stack();
     *
     * private [static] void saveTo$var() {
     *   Object val[] = new Object[n];
     *   val[1] = preVar1;
     *   ...
     *   val[n] = preVarn;
     *   var.push(val);
     * }
     *
     * private [static] void restoreFrom$var() {
     *   Object val[] = (Object[]) var.pop();
     *   preVar1 = val[1];
     *   ...
     *   preVarn = val[n];
     * }
     * </pre>
     */
    //@ assignable newMethods;
    protected void genStackMethod(/*@ non_null @*/ PreValueVars vars) {
        if (vars.isEmpty() ||
            (methodDecl instanceof JmlConstructorDeclaration)) {
            return;
        }

        //@ assert stackVar != null;
        String saveMethod = MN_SAVE_TO + stackVar;
        String restoreMethod = MN_RESTORE_FROM + stackVar;
        String ident = methodDecl.ident();
        String mod = methodDecl.isStatic() ? " static " : " ";

        // add a private stack field declaration to be used for saving
        // into, and restoring from the pre-values such as
        // preconditions, old variables, and old expressions, to guard
        // againt potential recursive method calls.
        newMethods.add(RacParser.parseMethod(
               "\n/** Generated by JML to save pre-values against potential " +
               "recursive\n  * method calls to the method " +ident+ ". */\n" +
               "private" + mod + "transient java.util.Stack " + stackVar  +
                  " = new java.util.Stack();"));

        // add a save (push) method
        newMethods.add(RacParser.parseMethod(
            "\n\n/** Generated by JML to save pre-values for the method " + 
            ident + ". */\n" +
            "private" + mod + "void " + saveMethod + "() {\n" +
            "$0" +
            "}", 
            vars.saveCode(stackVar)));

        // add a restore (pop) method
        newMethods.add(RacParser.parseMethod(
            "\n\n/** Generated by JML to restore pre-values for the method " + 
            ident + ". */\n" +
            "private" + mod + "void " + restoreMethod + "() {\n" +
            "$0" +
            "}\n", 
            vars.restoreCode(stackVar)));
    }

    private boolean methodReturnTypeHasExplicitNullityModifiers() {
	return hasFlag(methodDecl.modifiers(), NULLITY_MODS);
    }

    private boolean paramsHaveExplicitNullityModifiers() {
	JFormalParameter[] params = methodDecl.parameters();
	for (int i = 0; i < params.length; i++) {
	    if (hasFlag(params[i].modifiers(), NULLITY_MODS)) {
		return true;
	    }
	}
	return false;
    }

    private JExpression preNonNullAnnotationsDefaultToTrueExpr() {
	JExpression root = preNonNullAnnotations();
	return root != null ? root
	    : new JBooleanLiteral(methodDecl.getTokenReference(), true);
    }
	
    /** Composes the (implicit or explicit) <code>non_null</code> annotations of formal 
     * parameters, if any, into a JML expression and return it. 
     * If all parameters are nullable then,
     * the method returns <code>null</code>.
     *
     * <pre><jml>
     * ensures (* \result may be null *);
     * </jml></pre>
     */
    protected JExpression preNonNullAnnotations() {
	JExpression root = null;
	JFormalParameter[] params = methodDecl.parameters();
	for (int i = 0; i < params.length; i++) {
	    if (params[i].isDeclaredNonNull()) { //NonNull()) {
		// build expression of the form: x != null
		TokenReference tref = params[i].getTokenReference();
		JExpression expr = new JEqualityExpression(tref, OPE_NE,
		    new JLocalVariableExpression(tref, params[i]),
		    new JNullLiteral(tref));
		
		// build rac predicate (for error reporting) and conjoin it
		expr = new RacPredicate(expr, 
                                        "non_null for parameter '" + 
                                        params[i].ident()
                                        + "'");
		root = root == null ? 
		    expr : new JConditionalAndExpression(tref, root, expr);
	    }
	}
	return root;
    }

    /** Returns an expression conjoinable to the normal postcondition,
     * that checks the <code>non_null</code> annotation of the method
     * declaration. If the method declaration is <code>nullable</code> then
     * <code>null</code> is returned.
     */
    protected /*@ nullable @*/ JExpression postNonNullAnnotation() {
	if( !methodDecl.isDeclaredNonNull()) { //NonNull() ) {
            return null;
        }
	// Return type declared non-null either explicitly or implicitly ...

        // build AST node for \result with typecheck information
        TokenReference tref = methodDecl.getTokenReference();
        JmlResultExpression rexpr = new JmlResultExpression(tref);
        rexpr.setType(methodDecl.returnType());
        
        // build expression of the form: \result != null
        JExpression eqexpr = new JEqualityExpression(tref, OPE_NE, rexpr,
                                                     new JNullLiteral(tref));

        // build and return a RacPredicate for the error printing purpose
        // (see TransExpression.visitRacPredicate(RacPredicate)).
        return new RacPredicate(eqexpr, "non_null for result");
    }

    /** @return true iff methodDecl has explicit spec cases at least
     * one of which is not a redundant spec case (e.g. imples_that
     * introduces a redundant spec case. */
    private boolean methodHasRealExplicitSpecs() {
	if(!methodDecl.hasSpecification())
	    return false;
	//@ assert methodDecl.methodSpecification() != null;
	JmlSpecCase[] specCases = methodDecl.methodSpecification().specCases();
	return specCases != null && specCases.length > 0;
	// It would seem that all elements in specCases are
	// non-redundant, hence not need to inspec the content of this
	// array.
    }

    /** Returns the name of the targe method declaration being
     * translated. */
    protected String methodName() {
        return methodDecl.ident();
    }

    // ----------------------------------------------------------------------
    // INNER CLASSES
    // ----------------------------------------------------------------------

    /**
     * A class for collecting all specification cases from a desugared
     * method specification. A <em>desugared method specification</em>
     * is a method specification that consists of only <em>general
     * behavior specification cases</em>.
     *
     * @see DesugarSpec
     * @see org.jmlspecs.checker.JmlGenericSpecCase
     * @author Yoonsik Cheon
     * @version $Revision: 1.68 $
     */
    protected static class SpecCaseCollector extends JmlAbstractVisitor
	implements JmlVisitor {

        /** Constructs a new instance.
         *
         * <pre><jml>
         * requires mspec != null && (* mspec is desugared *);
         * </jml></pre>
         */
	protected SpecCaseCollector(JmlMethodSpecification mspec) {
	    super();
	    specCases = new ArrayList();
	    if (mspec != null)
		mspec.accept(this);
	}

	public void visitJmlExtendingSpecification(
	    JmlExtendingSpecification self) {
	    JmlSpecCase[] specCases = self.specCases();
	    if (specCases != null)
		for (int i = 0; i < specCases.length; i++)
		    specCases[i].accept(this);
	}

	public void visitJmlSpecification(JmlSpecification self) {
	    JmlSpecCase specCases[] = self.specCases();
	    if (specCases != null)
		for (int i = 0; i < specCases.length; i++) {
		    specCases[i].accept(this);
		}
	}

	public void visitJmlBehaviorSpec(JmlBehaviorSpec self) {
	    specCases.add(new GeneralSpecCase(self.specCase()));
	}
	
	/** Returns an iterator over all collected spec cases. 
         *
         * <pre><jml>
         * ensures \result != null && (* each element of \result is
         * instanceof GeneralSpecCase *);
         * </jml></pre>
         *
         */
	public Iterator iterator() {
	    return specCases.iterator();
	}

	/** List of desugared JmlGenericSpecCase. */
	private /*@ non_null @*/ List specCases;
    }

    /**
     * An abstract superclass for conjoining multiple specification
     * clauses, such as <code>requires</code> and <code>ensures</code>
     * clauses, in a specification case.
     *
     * @author Yoonsik Cheon
     * @version $Revision: 1.68 $
     */
    protected static abstract class SpecCase {
	protected SpecCase(JmlSpecBodyClause[] bc) {
	    bodyClauses = bc;
	    if (bodyClauses == null)
		bodyClauses = new JmlSpecBodyClause[0];
	}
	
        /** Returns true if this spec case has any specification
         * variable declarations. */
	public abstract boolean hasSpecVarDecls();

        /** Returns true if this spec case has the precondition
         * specified. */
	public abstract boolean hasPrecondition();

        /** Returns true if this spec case has the normal
         * postcondition specified. */
	public boolean hasPostcondition() {
	    for (int i = 0; i < bodyClauses.length; i++)
		if ((bodyClauses[i] instanceof JmlEnsuresClause) &&
                    isCheckable(bodyClauses[i]))
		    return true;
	    return false;
	}

        /** Returns true if this spec case has the postcondition. */
	public boolean hasExceptionalPostcondition() {
	    for (int i = 0; i < bodyClauses.length; i++)
		if (isCheckable(bodyClauses[i]) &&
                    (bodyClauses[i] instanceof JmlSignalsClause ||
                     bodyClauses[i] instanceof JmlSignalsOnlyClause)) 
		    return true;
	    return false;
	}

        /** Returns the specification variable declarations of this
         * spec case. */ 
	public abstract JmlSpecVarDecl[] specVarDecls();

        /** Returns the precondition of this spec case. If this spec
         * case has multiple <code>requires</code> clauses, the
         * clauses are conjoined.
         * @see #conjoin(List)
         */
	public abstract JExpression precondition();

        /** Returns the normal postcondition of this spec case. If
         * this spec case has multiple <code>ensures</code> clauses,
         * the clauses are conjoined.
         * @see #conjoin(List) */
	public JExpression postcondition() {
	    List l = new ArrayList();
	    for (int i = 0; i < bodyClauses.length; i++) {
		if ((bodyClauses[i] instanceof JmlEnsuresClause)
                    && isCheckable(bodyClauses[i])) {
		    l.add(bodyClauses[i]);
                }
            }
	    if (l.size() > 0)
		return conjoin(l);
	    return null;
	}
       
        /**
         * Returns true if the given spec body clause is
         * checkable. The clause is checkable if it is not a redundant
         * specification or the command line option "noredundancy" is
         * not turned on.
         */
        protected boolean isCheckable(JmlSpecBodyClause clause) {
            return !clause.isRedundantly() || !Main.racOptions.noredundancy();
        }

        /** Returns the exceptional postcondition of this spec
         * case. If this spec case has multiple <code>signals</code>
         * clauses, the clauses are conjoined. If this spec case has a
         * signals_only clause, it is first desugared into a signals
         * clause.
         *
         * <pre><jml>
         * ensures \result != null ==> \result.length > 0;
         * </jml></pre>
         *
         * @see #conjoin(List)*/
	public JmlSignalsClause[] exceptionalPostcondition() {
	    List l = new ArrayList();
	    for (int i = 0; i < bodyClauses.length; i++) {
                if (!isCheckable(bodyClauses[i])) {
                    continue;
                }
		if (bodyClauses[i] instanceof JmlSignalsClause) {
		    l.add(bodyClauses[i]);
                } else if (bodyClauses[i] instanceof JmlSignalsOnlyClause) {
		    l.add(desugar((JmlSignalsOnlyClause) bodyClauses[i]));
                }
            }
	    if (l.size() > 0) {
		JmlSignalsClause[] sc = new JmlSignalsClause[l.size()];
		l.toArray(sc);
		return sc;
	    }
	    return null;
	}

        /** Desugars a signals_only clause into a signals clause. A
         * signals_only clause of the form:
         * 
         * <pre>
         *  signals_only \nothing;
         * </pre>
         *
         * is desugared into the following signals clause:
         *
         * <pre>
         *  signals (java.lang.Exception e) false;
         * </pre>
         *
         * Furthermore, a signals_only clause of the form
         *
         * <pre>
         *  signals_only E1, E2, ..., En;
         * </pre>
         *
         * is desugared into the following signals clause:
         *
         * <pre>
         *  signals (java.lang.Exception e) 
         *       e instanceof E1
         *    || e instanceof E2
         *       ...
         *    || e instanceof En;
         * </pre>
         */
        private  /*@ non_null @*/ JmlSignalsClause desugar(
            /*@ non_null @*/ JmlSignalsOnlyClause sig) {

            TokenReference tref = sig.getTokenReference();
            if (sig.isNothing()) {
                JExpression expr = new JBooleanLiteral(tref, false);
                JmlPredicate pred
                    = new JmlPredicate(new JmlSpecExpression(expr));
                return new JmlSignalsClause(tref, 
                                            sig.isRedundantly(),
                                            CStdType.Exception,
                                            "jml$e",
                                            pred,
                                            false);
            }
                
            CClassType[] exceptions = sig.exceptions();
            //@ assume exceptions.length > 0;

            JExpression expr = null;
            for (int i = 0; i < exceptions.length; i++) {
                JVariableDefinition var = new JVariableDefinition(
                    tref, 0, CStdType.Exception, "jml$e", null);
                JLocalVariableExpression varExpr =
                    new JLocalVariableExpression(tref, var);
                JExpression inst = new JInstanceofExpression(
                    tref, varExpr, exceptions[i]);
                expr = (expr == null) ? inst :
                    new JConditionalOrExpression(tref, expr, inst);
            }

	    JmlPredicate pred = new JmlPredicate(new JmlSpecExpression(expr));
            return new JmlSignalsClause(tref, 
                                        sig.isRedundantly(),
                                        CStdType.Exception,
                                        "jml$e",
                                        pred,
                                        false);
        }

        /** Returns a new conjoined expression made out of predicates
         * of the given specification clauses. A null is returned if
         * there is nothing to be conjoined in the given list of
         * specification clauses. A specification clause corresponding
         * to <code>\not_specified</code> is desugared into true.
         * 
         * <pre><jml>
         * requires clauses != null && 
         *   (\forall Object o; clauses.contains(o); 
         *     (o instanceof JmlRequiresClause) ||
         *     (o instanceof JmlEnsuresClause));
         * </jml></pre>
         */
	protected static JExpression conjoin(List clauses) {
	    JExpression left = null;
	    Iterator iter = clauses.iterator();
            JmlPredicateClause notSpecified = null;
	    while (iter.hasNext()) {
		JmlPredicateClause c = (JmlPredicateClause)iter.next();
                if (!c.isNotSpecified()) {
		    RacPredicate p = new RacPredicate(c.predOrNot());
		    if (left == null)
			left = p;
		    else
			left = new JConditionalAndExpression(org.multijava.util.compiler.TokenReference.NO_REF,
							     left, p);
		} else if (notSpecified == null) {
                    notSpecified = c;
                }
	    }

            // desugar \not_specified as true; 
            // see testcase AlsoAndLightweight.java
            if (left == null && notSpecified != null) {
                return new JBooleanLiteral(
                    notSpecified.getTokenReference(), true);
            }
	    return left;
	}

	protected /*@ non_null @*/ JmlSpecBodyClause[] bodyClauses;
    }

    /**
     * A concrete wrapper class to <code>JmlGeneralSpecCase</code> for
     * conjoining multiple specification clauses in a specification
     * case.
     *
     * @see org.jmlspecs.checker.JmlGeneralSpecCase
     * @author Yoonsik Cheon
     * @version $Revision: 1.68 $
     */
    protected static class GeneralSpecCase extends SpecCase {
	/** Constructs a new instance.
         * 
         * <pre><jml>
         * requires spec != null && (* spec is desugared *);
         * </jml></pre>
         */
	public GeneralSpecCase(JmlGeneralSpecCase spec) {
	    super((spec.hasSpecBody() && spec.specBody().isSpecClauses()) ?
		  spec.specBody().specClauses() : null);
	    this.spec = spec;
	}

        /** Returns true if this spec case has any specification
         * variable declarations. */
	public boolean hasSpecVarDecls() {
	    return spec.hasSpecVarDecls();
	}

        /** Returns true if this spec case has the precondition
         * specified. */
	public boolean hasPrecondition() {
	    if (spec.hasSpecHeader()) {
                JmlRequiresClause[] clauses = spec.specHeader();
                for (int i = 0; i < clauses.length; i++) {
                    if (isCheckable(clauses[i]))
                        return true;
                }
            }
	    return false;
	}

        /** Returns the specification variable declarations of this
         * spec case. */ 
	public JmlSpecVarDecl[] specVarDecls() {
	    return spec.specVarDecls();
	}

        /** Returns the precondition of this spec case. If this spec
         * case has multiple <code>requires</code> clauses, the
         * clauses are conjoined.
         */
	public JExpression precondition() {
	    if (spec.hasSpecHeader()) {
                JmlRequiresClause[] clauses = spec.specHeader();
                List l = new ArrayList();
                for (int i = 0; i < clauses.length; i++) {
                    if (isCheckable(clauses[i]))
                        l.add(clauses[i]);
                }
                if (l.size() > 0)
                    return conjoin(l);
            }
	    return null;
	}

        /** The spec case whose specification clauses to be conjoined. */
	private /*@ non_null @*/ JmlGeneralSpecCase spec;
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** Target type declaration whose methods are to be translated.
     * 
     * <pre><jml>
     * protected invariant typeDecl != null;
     * </jml></pre>
     */
    protected final /*@ spec_public @*/ JmlTypeDeclaration typeDecl;
    

    /** target method to be transformed.
     * 
     * <pre><jml>
     * protected invariant methodDecl != null ==>
     *   typeDecl.methods().contains(methodDecl);
     * </jml></pre>
     */
    protected /*@ spec_public @*/ JmlMethodDeclaration methodDecl;

    /** newly created fields (as the result of translation) that
     * need be added to the target class, e.g., old and precondition
     * variables */ 
    protected /*@ spec_public non_null @*/ List newFields;

    /** newly created methods (as the result of translation) that
     * need be added to the target class */
    protected /*@ spec_public non_null @*/ List newMethods;

    /** desugared specification of the target method to be translated */
    protected /*@ spec_public @*/ JmlMethodSpecification desugaredSpec;

    /** for generating fresh variables needed in the generated Java code */
    protected /*@ non_null @*/ VarGenerator varGen;

    private /*@ spec_public @*/ String stackVar;
}

