/* $Id: TransPostExpression2.java,v 1.14 2007/07/19 10:51:37 f_rioux Exp $
 *
 * Copyright (C) 2005 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.jmlspecs.jmlrac.qexpr.*;
import org.multijava.mjc.*;
import java.util.*;

/**
 * A RAC visitor class to translate JML expressions into Java source code.
 * 
 * @see TransExpression2
 * 
 * @author Frederic Rioux (based on Yoonsik Cheon's work)
 * @version $Revision: 1.14 $
 */
public/*@ non_null_by_default */ class TransPostExpression2 extends TransExpression2 {
    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a new instance and translates the given
     * post-expression.
     *
     * @param resultVar variable to store the result
     */
	public TransPostExpression2( 
			VarGenerator varGen,
			RacContext ctx,
			VarGenerator oldVarGen,
			boolean forStatic,
			JExpression pred,
			String resultVar,
			JTypeDeclarationType typeDecl,
			String errorType) {
		super(varGen, ctx, pred, resultVar, typeDecl, errorType);
		this.oldVarGen = oldVarGen;
		oldExprs = new ArrayList();
		this.forStatic = forStatic;
	}
	
	protected void handleExceptionalTranslation(PositionnedExpressionException e){
		quantifiers = new Stack();	//reset static quantifier stack
		super.handleExceptionalTranslation(e);
	}

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /** Returns a list of old expressions (as <code>RacNode</code>) that
     * must be executed in the pre-state. The declaration of a local 
     * variable to hold the result of the expression is stored in the
     * <code>name</code> field of class <code>RacNode</code>. This method
     * must be called after translation.
     */
    public List oldExprs() {
        perform();
	return oldExprs;
    }

    // ----------------------------------------------------------------------
    // TRANSLATION
    // ----------------------------------------------------------------------

    /**
	 * Translates a JML pre expression.
	 */
	public void visitJmlPreExpression(JmlPreExpression self) {
		transPreExpression(self);
	}

	/**
	 * Translates a JML old expression.
	 */
	public void visitJmlOldExpression(JmlOldExpression self) {
		transPreExpression(self);
	}

    public void transPreExpression( JmlSpecExpressionWrapper self ) {
        
    	// does the pre expression have any quantified variables?
        if (hasQuantifiedVar(self)) {
        	//throw(new NotImplementedExpressionException(self.getTokenReference(), "Quantified variables in JML pre-state expression"));
        	oldExpressionInQuantifiers(self);
            return;
        } 

        // Translates the pre expression such that it is to be
        // evaluated in the pre-state and stores the result into a
        // special private field of type TN_JMLVALUE.

        JExpression expr = self.specExpression();
        CType type = expr.getApparentType();

        // translate the pre expr with a fresh pre-state variable
        String var = oldVarGen.freshVar();
        TransExpression2 tr = new TransExpression2(oldVarGen, context, expr, var, typeDecl, errorType);
        tr.isInner = true;
        tr.thisIs = thisIs; //"this" issue for anonymous class

        String oldVar = varGen.freshOldVar();
        String oldStmt = tr.stmtAsString();
        
        // There is always the possibility that the evaluation of the old expression
        // failed. If so, rethrow the PositionnedExpressionException.
	    if(!tr.isProperlyEvaluated()){
	    	throw tr.reportedException;
	    }
        
        // like in: \old(null)
        if("null".equals(oldStmt)){
        	RETURN_RESULT("null");
        	return;
        }
     
        String resultStr = TN_JMLVALUE + ".ofObject(" + TransUtils.wrapValue(type, oldStmt)+ ")";
        RacNode node = tr.wrap(RacParser.parseStatement(oldVar + " = " + resultStr + ";"));

        // piggy-back the name of new old variable in the name field
        // of the RAC node; the node will be evaluated by the precondition
        // check method.
        
        node.setVarDecl(PreValueVars.createJmlValueVar(forStatic, oldVar));
        oldExprs.add(node);
        
	RETURN_RESULT(TransUtils.unwrapObject(type, oldVar + ".value()"));
    }

    /** Translates the given JML quantified expression. It is overridden
     * here to keep track of the set of quantifiers that enclose the old 
     * expression currently being translated.
     *
     * @see #visitJmlPreExpression(JmlPreExpression)
     * @see #visitJmlOldExpression(JmlOldExpression)
     */

    public void visitJmlSpecQuantifiedExpression( 
       JmlSpecQuantifiedExpression self )
    {	       
    	//super.visitJmlSpecQuantifiedExpression(self);       
        TransPostExpression2 t = new TransPostExpression2(varGen, context, oldVarGen, forStatic, expression, resultVar, typeDecl, errorType);
        quantifiers.push(self);
        LOG(this.toString() + " QStack push " + quantifiers.size());
        translateSpecQuantifiedExpression(t, self);	
		oldExprs.addAll(t.oldExprs);
		quantifiers.pop();
        LOG(this.toString() + " QStack pop " + quantifiers.size());
    }

    /**
     * Translates a JML result expression.
     */
    public void visitJmlResultExpression( JmlResultExpression self ) {
    	RETURN_RESULT(VN_RESULT);
    }
        
    // ----------------------------------------------------------------------
    // HELPER METHODS
    // ----------------------------------------------------------------------

    /** Does the given old or pre expression, <code>expr</code>, have any
     * free quantified variables? */

    private boolean hasQuantifiedVar(JmlSpecExpressionWrapper expr) {
        return new QVarChecker().hasQVar(expr);
    }
    
    /**
     * Translates a JML old expression, <code>self</code>, enclosed in
     * quantifiers. The translation generates code that evaluates the
     * old expression for each combination of bound variables of
     * quantifiers and stores the result to a private cache. The original
     * old expression is replaced with cache lookup code.
     * 
     * <pre><jml>
     * requires !quantifiers.isEmpty();
     * //assignable oldExprs;
     * </jml></pre>
     */
    private void oldExpressionInQuantifiers(JmlSpecExpressionWrapper self) {
    	JExpression expr = self.specExpression();
    	CType type = expr.getApparentType();
    	String mapVar = varGen.freshOldVar(); // cache variable
    	
    	// statements for evaluating old expression, E, of type T.
    	//  T v = T_init;
    	//  v = [[E]]
    	//  mapVar.put(key, v);
    	String var = oldVarGen.freshVar();
    	String key = buildKey();
    	
    	// coerce to type TN_JMLVALUE with appropriate guard against 
    	// undefinedness.
//    	RacNode stmt = new TransOldExpression(oldVarGen,context,expr,var,typeDecl).stmt();
    	
    	TransExpression2 tr = new TransExpression2(oldVarGen,context,expr,var,typeDecl, errorType);
    	RacNode stmt = tr.stmt(false);
    	
	    if(!tr.isProperlyEvaluated()){
	    	throw tr.reportedException;
	    }
	    
    	stmt = RacParser.parseStatement(
    			//"try {\n" +
    			"  " + TransUtils.toString(type) + " " + var + " = " +
    			TransUtils.defaultValue(type) + ";\n" +
    			"$0\n" +
    			"  " + mapVar + ".put(" + key + ", " +TN_JMLVALUE+ ".ofObject(" +
    			TransUtils.wrapValue(type, var) + "));"/* +
    			"}\n" +
    			"catch (JMLNonExecutableException jml$e0) {\n" +
    			"  " + mapVar + ".put(" + key + 
    			", " +TN_JMLVALUE+ ".ofNonExecutable());\n" +
    			"}" +
    			"catch (java.lang.Exception jml$e0) {\n" +
    			"  " + mapVar + ".put(" + key + 
    			", " +TN_JMLVALUE+ ".ofUndefined());\n" +
    			"}"*/, 
    			stmt.incrIndent());
    	
    	// iteration of the statement over bound variables
    	try {
    		TransQuantifiedExpression trans = null;
    		// As the variable quantifiers is a stack, the translation
    		// should be done in the reverse order, i.e., from the
    		// inner to outer quantifiers. Thanks to Peter Chan for
    		// reporting this error.
    		for (int i = quantifiers.size() - 1; i >= 0; i--) {
    			trans = new TransQuantifiedExpression(varGen, context,
    					(JmlSpecQuantifiedExpression)quantifiers.get(i), 
    					null, this);
    			stmt = trans.generateLoop(stmt);
    		}
    	} catch (NonExecutableQuantifierException e) {
    		/*
    		// contextual interpretation of non-executable expression
    		if (context.enabled() && type.isBoolean()) {
    			RETURN_RESULT(RacParser.parseStatement(
    					"// from a non_executable, boolean, JML clause\n" +
    					context.angelicValue(GET_ARGUMENT()) +  ";"));
    		} else {
    			nonExecutableExpression();
    		}
    		return;*/
    		throw(new NotImplementedExpressionException(self.getTokenReference(), "The old expression in this quantifier"));
    	}
    	
    	// piggyback cache declaration in the name field
    	stmt.setVarDecl(PreValueVars.createVar(forStatic, 
    			"JMLOldExpressionCache",
    			mapVar,
    	"new JMLOldExpressionCache()"));
    	oldExprs.add(stmt);
    	
    	// replace the old expr with cache lookup code
//    	RacNode result = RacParser.parseStatement(
//    			"if (" + mapVar + ".containsKey(" + key + ")) {\n" +
//    			"  " + GET_ARGUMENT() + " = " +
//    			TransUtils.unwrapObject(type, 
//    					"((" +TN_JMLVALUE+ ")" +mapVar+ ".get(" +key+ ")).value()") 
//    					+ ";\n" +
//    					"} else {\n" +
//    					"  throw JMLChecker.ANGELIC_EXCEPTION;\n" +
//    	"}");
    	String result = TransUtils.unwrapObject(type,"((" +TN_JMLVALUE+ ")" +mapVar+ ".get(" +key+ ")).value()");
    					
    	RETURN_RESULT(result);        
    }            
    
    /** Returns a key for retrieving the value of old expression currently
     * being translated. We store the old expression values as a mapping
     * from quantified (bound) variables to values, and the quantified
     * variables beomes a key. For example, if the quantifiers are 
     * <code>x1</code>, <code>x2</code>, ..., <code>xn</code>, the key
     * is <code>new Object[] { x1, x2, ..., xn }</code>, where 
     * <code>xi</code> is wrapped if necessary.
     *
     * <pre><jml>
     * requires !quantifiers.isEmpty();
     * ensures \result != null;
     * </jml></pre>
     */
    private String buildKey() {
    	StringBuffer code = new StringBuffer();
    	code.append("new java.lang.Object[] { " );
    	boolean isFirst = true;
    	for (Iterator i = quantifiers.iterator(); i.hasNext(); ) {
    		JVariableDefinition[] vars = 
    			((JmlSpecQuantifiedExpression)i.next()).quantifiedVarDecls();
    		for (int j = 0; j < vars.length; j++) {
    			if (isFirst) {
    				isFirst = false;
    			} else {
    				code.append(", ");
    			}
    			code.append(TransUtils.wrapValue(vars[j].getType(), 
    					vars[j].ident()));
    		}            
    	}
    	code.append(" }");
    	return code.toString();
    }
    
    
    // ----------------------------------------------------------------------
    // INNER CLASS
    // ----------------------------------------------------------------------

    /**
     * A class to check whether an expression has any references to 
     * quantified variables.
     */
    
    private class QVarChecker extends AbstractExpressionVisitor {
    	public QVarChecker() {}
    	
    	/** Returns <code>true</code> if the expression <code>expr</code>
    	 * contains any references to quantified variables.
    	 */
    	
    	public boolean hasQVar(JExpression expr) {
    		hasQVar = false;
    		this.qvars = new HashSet();
    		for (Iterator i = quantifiers.iterator(); i.hasNext(); ) {
    			JVariableDefinition[] vars = 
    				((JmlSpecQuantifiedExpression)i.next()).quantifiedVarDecls();
    			for (int j = 0; j < vars.length; j++) {
    				qvars.add(vars[j].ident());
    			}
    		}
    		expr.accept(this);
    		return hasQVar;
    	}
    	public void visitLocalVariableExpression( 
    			JLocalVariableExpression self ) {
    		if (qvars.contains(self.ident())) 
    			hasQVar = true;
    	}
    	private boolean hasQVar;
    	private Set qvars;
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** Indicates whether this translation is for a static method
     * declaration. */
    // TODO useful???
    private boolean forStatic;

    /** A list of <code>RacNode</code>'s representing old expressions that 
     * must be executed in the pre-state. */
    private List oldExprs;

    /** Variable generator to be used in the translation of old expression. 
     */
    private VarGenerator oldVarGen;

    /** The set of quantifiers that enclose the old expression currently
     * being translated.
     * 
     * <pre><jml>
     * private invariant (\forall Object o; quantifiers.contains(o);
     *   o instanceof JmlSpecQuantifiedExpression);
     * </jml></pre>
     */
    private static Stack quantifiers = new Stack();
}
