/* $Id: WrapperMethod.java,v 1.31 2006/12/13 21:09:05 wdietl Exp $
 *
 * Copyright (C) 2001 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlrac;

import org.jmlspecs.checker.*;
import org.multijava.mjc.*;

/**
 * A class for generating wrapper methods. A <em>wrapper method</em>
 * is a method that wraps a method with assertion check methods such
 * as pre and postcondition check methods. The original method is
 * renamed and called by the wrapper method inside the wrap. Wrapper
 * methods are generated for only concrete methods; no wrapper methods
 * are generated for abstract methods.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.31 $
 */

public class WrapperMethod extends TransUtils {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a new instance.
     *
     * <pre><jml>
     * requires typeName != null && mdecl != null && mdecl.body() != null;
     * </jml></pre>
     * 
     * @param mdecl method to be wrapped; must not be abstract.
     */
    public WrapperMethod(String typeName, JmlMethodDeclaration mdecl) {
        // initialize private short-cut fields
	this.typeName = typeName;
	this.methodDecl = mdecl;
	returnType = mdecl.returnType();
	hasReturn = returnType != null && !returnType.isVoid();
	ident = mdecl.ident();
        isHelper = mdecl.isHelper();
        isStatic = mdecl.isStatic();
    }

    /** 
     * Returns a wrapper method. The original method is renamed 
     * and called by the wrapper method enclosed in assertion check code
     * such as pre and postcondition check methods. The body of the wrapper
     * method has the following general structure:
     *
     * <pre>
     * T m(T1 x1, ..., Tn xn) {
     *   checkPre$m(x1, ..., xn);
     *   checkInv();
     *   T r = init_T;
     *   boolean ok = true;
     *   try {
     *     r = orig$m(x1, ..., xn);
     *     checkPost$m(x1, ..., xn, r);
     *   }
     *   catch (JMLEntryPreconditionError e) {
     *     ok = false;
     *     throw new JMLInternalPreconditionError(e);
     *   }
     *   catch (JMLExitPostconditionError e) {
     *     ok = false;
     *     throw new JMLInternalPostconditionError(e);
     *   }
     *   catch (JMLAssertionError e) {
     *     ok = false;
     *     throw e;
     *   }
     *   catch (Throwable e) {
     *     try {
     *       checkXPost$m(x1, ..., xn, r, e);
     *     } 
     *     catch (JMLAssertionError e1) {
     *       ok = false;
     *       throw e1;
     *     }
     *   }
     *   finally {
     *     if (ok) {
     *        checkInv();
     *        checkConstraint();
     *     }
     *   }
     * }
     * </pre>
     * 
     * <pre><jml>
     * old String n = methodDecl.ident();
     * modifies methodDecl.*;
     * assignable args, postArgs, xpostArgs;
     * ensures methodDecl.ident().equals(MN_INTERNAL + n) &&
     *         \fresh(\result);
     * </jml></pre>
     */
    public /*@ non_null @*/ JMethodDeclarationType generate()
    {
	// build a wrapper method
	String name = methodDecl.ident();
	buildArguments();
	JMethodDeclaration mdecl = new JMethodDeclaration(
           methodDecl.getTokenReference(),
	   methodDecl.modifiers(),
	   CTypeVariable.EMPTY,
	   methodDecl.returnType(),
	   name,
	   methodDecl.parameters(), 
	   methodDecl.getExceptions(),
	   buildWrapperBody(),
	   null, // methodDecl.javadocComment(),
	   null);

        if(JmlRacGenerator.checking_mode != JmlRacGenerator.WRAPPER) {
            // rename the original method and make it private
            methodDecl.setIdent(MN_INTERNAL + name);
            long mod = methodDecl.modifiers();
            mod = mod & (~ACC_PUBLIC);
            mod = mod & (~ACC_PROTECTED);
            mod = mod | (ACC_PRIVATE);
            methodDecl.setModifiers(mod);
        } else { //if tthe mode is WRAPPER
            //We need to change all private and protected members
            //of the wrapper method to public.
            long mod = mdecl.modifiers();
            mod = mod & (~ACC_PRIVATE);
            mod = mod & (~ACC_PROTECTED);
            mod = mod | (ACC_PUBLIC);
            mdecl.setModifiers(mod);
            //We Have a problem here!
            //methodDecl is final, and should never be null
            //However, methodDecl is the original method,
            //and we should not have it in the class at all!
            //The solution is to set the ident to begin with $chx_IGNORE_THIS,
            //which tells the pretty printer *not* to print it out.
            methodDecl.setIdent("$chx_IGNORE_THIS_" + name);
        }

        // add javadoc and return the accessor
        return RacParser.parseMethod(
            "\n/** Generated by JML to wrap assertion check code to\n" +
            " * the method " + name + ". */$0", mdecl);
    }

    /** 
     * Returns code for the body of the wrapper method being created.
     * For the structure of the returned code refers to the method
     * <code>generate()</code>.
     *
     * <pre><jml>
     * ensures \result != null;
     * </jml></pre>
     *
     * @see #generate()
     */
    protected /*@ non_null @*/ JBlock buildWrapperBody() {
	final StringBuffer code = new StringBuffer("{\n");
        code.append(   nocheckDuringInit());
	code.append(   resultVarDecl());
        code.append("  if (!(" + isActive("PRECONDITION") + ")) {\n");
	code.append(     callOriginalMethod());
	code.append(     mandatoryReturn());
	code.append("  }\n");
	code.append(   evalOldInConstraint());
	code.append(   checkPreInvariant());
	code.append(   checkPrecondition());
	code.append("  boolean rac$ok = true;\n");
	code.append("  try {\n");
	code.append("    // execute original method\n");
	code.append(     callOriginalMethod());
	code.append(     checkPostcondition());
	code.append("  }\n");
	code.append("  catch (JMLEntryPreconditionError rac$e) {\n");
	code.append("    rac$ok = false;\n");
	code.append("    throw new JMLInternalPreconditionError(rac$e);\n");
	code.append("  }\n");
	code.append("  catch (JMLExitNormalPostconditionError rac$e) {\n");
	code.append("    rac$ok = false;\n");
        code.append("    throw new JMLInternalNormalPostconditionError(rac$e);\n");
        code.append("  }\n");
	code.append("  catch (JMLExitExceptionalPostconditionError rac$e) {\n");
	code.append("    rac$ok = false;\n");
        code.append("    throw new JMLInternalExceptionalPostconditionError(rac$e);\n");
	code.append("  }\n");
	code.append("  catch (JMLAssertionError rac$e) {\n");
	code.append("    rac$ok = false;\n");
	code.append("    throw rac$e;\n");
	code.append("  }\n");
	code.append("  catch ( java.lang.Throwable rac$e) {\n");
	code.append("    try {\n");
	code.append(       checkXPostcondition());
	code.append("    }\n");
	code.append("    catch (JMLAssertionError rac$e1) {\n");
	code.append("      rac$ok = false;\n");
	code.append("      throw rac$e1;\n");
	code.append("    }\n");
	code.append("  }\n");
	code.append("  finally {\n");
	code.append("    if (rac$ok) {\n");
	code.append(       checkPostInvariant());
	code.append(       checkConstraint());
	code.append("    }\n");
	code.append("  }\n");
	code.append(   optionalReturn());
	code.append("}");
	return RacParser.parseBlock(code.toString());
    }

    /** Returns code for bypassing assertion checks during the class
     * and instance initialization. Assertions are also not checked
     * while the object is being constructed. The returned code has
     * the following structure.
     *
     * <pre>
     *  if (!VN_CLASS_INIT [|| !VN_INIT] [|| !VN_CONSTRUCTED()]) {
     *     [return] m(x1, ..., xn);
     *     [return;]
     *  }
     * </pre>
     */
    protected /*@ non_null @*/ String nocheckDuringInit() {
        String prefix = MN_INTERNAL;
        if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
            if(methodDecl.isStatic()) {
                prefix = "$chx_Orig_" + typeName + ".";
            } else {
                prefix = "(($chx_Orig_" + typeName + ") _wrapped_object).";
            }
        }
        return 
            "  // skip assertion checks during initialization\n" +
            "  if (!" + VN_CLASS_INIT + 
                   (isStatic? "" : " || !" + VN_INIT) + 
                   (isStatic? "" : " || !" + VN_CONSTRUCTED + "()") + ") {\n" +
            "    " + (hasReturn ? "return " : "") + // return for non-void
                   prefix + ident + args + ";\n" + 
            (hasReturn ? "" : "    return;\n") +    // return for void
            "  }\n";
    }

    /** Returns code for checking invariant in the pre-state. If the
     * method being translated is a helper method, an empty string is
     * returned; otherwise, the following form of code is returned.
     *
     * <pre>
     *  // check static invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT)) {
     *    JMLChecker.enter();
     *    checkInv$static();
     *    JMLChecker.exit();
     *  }
     *  // check instance invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT)) {
     *    JMLChecker.enter();
     *    checkInv$instance$T();
     *    JMLChecker.exit();
     *  }
     * </pre>
     *
     * where <code>T</code> is the name of type being translated. The
     * code for checking instance invariant is generated only if the
     * target method is an instance method.
     *
     * <pre><jml>
     *   requires isHelper;
     *   ensures \fresh(\result) && "".equals(\result);
     * also
     *   requires !isHelper;
     *   ensures \fresh(\result) && \result.length() > 0 &&
     *     (* \result is of the above form of code *);
     * </jml></pre>
     *
     */
    protected /*@ non_null@*/ String checkPreInvariant() {
        if (isHelper) {
            return "";
        }

	String code = wrap("INVARIANT", "check static invariant", 
			   "checkInv$static(\"" +
                           identification("pre") + "\")");
	if (!isStatic)
	    code += wrap("INVARIANT", 
			 "check instance invariant",
			 "checkInv$instance$" + typeName + "(\"" +
                           identification("pre") + "\")");
	return code;
    }

    /** Returns code for checking invariant in the post-state. If the
     * target method being translated is a helper method, an empty
     * string is returned; otherwise, the following form of code is
     * returned.
     *
     * <pre>
     *  // check static invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT)) {
     *    JMLChecker.enter();
     *    checkInv$static();
     *    JMLChecker.exit();
     *  }
     *  // check instance invariant
     *  if (JMLChecker.isActive(JMLChecker.INVARIANT)) {
     *    JMLChecker.enter();
     *    checkInv$instance$T();
     *    JMLChecker.exit();
     *  }
     * </pre>
     *
     * where <code>T</code> is the name of type being translated. The
     * code for checking instance invariant is generated only if the
     * target method is an instance method.
     *
     * <pre><jml>
     *   requires isHelper;
     *   ensures \fresh(\result) && "".equals(\result);
     * also
     *   requires !isHelper;
     *   ensures \fresh(\result) && \result.length() > 0 &&
     *     (* \result is of the above form of code *);
     * </jml></pre>
     */
    protected /*@ non_null @*/ String checkPostInvariant() {
        if (isHelper) {
            return "";
        }

	String code = wrap3("INVARIANT", "check static invariant", 
			   "checkInv$static(\"" + 
                            identification("post") + "\")");
	if (!isStatic)
	    code += wrap3("INVARIANT", 
                          "check instance invariant",
                          "checkInv$instance$" + typeName + 
                          "(\"" + identification("post") + "\")");
	return code;
    }

    /** Returns code for evaluating old expressions in constraints.  A
     * helper need not establish the constraints. Thus, an empty
     * string is returned for a helper and the following form of code
     * for a non-helper.
     *
     * <pre>
     *  // eval old exprs in static constraint
     *  if (JMLChecker.isActive(JMLChecker.CONSTRAINT)) {
     *    JMLChecker.enter();
     *    evalOldExprInHC$satic(); 
     *    JMLChecker.exit();
     *  }
     *  // eval old exprs in non-static constraint
     *  if (JMLChecker.isActive(JMLChecker.CONSTRAINT)) {
     *    JMLChecker.enter();
     *    evalOldExprInHC$instance$T();
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>T</code> is the name of type being translated. The
     * code for instance constraints is generated only if the
     * target method is an instance method.
     */
    protected /*@ non_null @*/ String evalOldInConstraint() {
        if (isHelper) {
            return "";
        }

	return wrap("CONSTRAINT", "eval old exprs in constraint", 
                    evalOldName(typeName,isStatic) + "()");
    }

    /** Returns code for checking history constraint. The constraint
     * is not checked for constructors. If the target method being
     * translated is a helper method, an empty string is returned;
     * otherwise, the following form of code is returned.
     *
     * <pre>
     *  // check static constraint
     *  if (JMLChecker.isActive(JMLChecker.CONSTRAINT)) {
     *    JMLChecker.enter();
     *    checkHC$static();
     *    JMLChecker.exit();
     *  }
     *  // check instance constraint
     *  if (JMLChecker.isActive(JMLChecker.CONSTRAINT)) {
     *    JMLChecker.enter();
     *    checkHC$instance$T();
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>T</code> is the name of type being translated. The
     * code for checking instance constraint is generated only if the
     * target method is an instance method.
     */
    protected /*@ non_null @*/ String checkConstraint() {
        if (isHelper) {
            return "";
        }

	String code = wrap3("CONSTRAINT", "check static constraint", 
                            "checkHC$static(\"" +
                            identification("post") + "\", \"" +
                            ident + "\", " +constraintCallArg()+ ")");
	if (!isStatic) {
            boolean isPrivate = !hasFlag(methodDecl.modifiers(), 
                                         ACC_PUBLIC | ACC_PROTECTED);
	    code += wrap3("CONSTRAINT", "check instance constraint",
			  "checkHC$instance$" + typeName + "(\"" + 
                            identification("post") + "\", " +
                          isPrivate + ", \"" +ident+ "\", " + 
                          constraintCallArg()+ ")");
        }
	return code;
    }

    /** Returns, in the string form, the argument for calling
     * a constraint check method. */
    protected /*@ non_null @*/ String constraintCallArg() {
        String code = "new java.lang.Class[] { ";
        JFormalParameter[] params = methodDecl.parameters();
        for (int i = 0; i < params.length; i++) {
            code += typeToClass(params[i].getType()) + ", ";
        }
        code += "}";
        return code;
    }

    /** Builds arguments to various assertion check methods such as
     * pre- and postcondition check methods. If the target method has
     * arguments x1, x2, ..., xn, and throws some exceptions,
     * this method sets the field
     * <code>args</code> to <code>"(x1,...,xn)"</code>, 
     * <code>postArgs</code> to <code>"(x1,...,xn,VN_RESULT)"</code>, and
     * <code>xpostArgs</code> to <code>"(x1,...,xn,VN_XRESULT)"</code>.
     *
     * <pre><jml>
     * assignable args, postArgs, xpostArgs;
     * ensures args != null && postArgs != null && xpostArgs != null;
     * </jml></pre>
     */
    protected void buildArguments() {
	JFormalParameter[] parameters = methodDecl.parameters();
	args = new StringBuffer("(");
	for (int i = 0; i < parameters.length; i++) {
	    if (i != 0)
		args.append(", ");
	    args.append(parameters[i].ident());
	}
	
	postArgs = new StringBuffer(args.toString());
        postArgs.append(parameters.length > 0 ? "," : "");
        postArgs.append(hasReturn ? VN_RESULT : "null");
	postArgs.append(")");
	
	xpostArgs = new StringBuffer(args.toString());
	xpostArgs.append(parameters.length > 0 ? "," : "");
	xpostArgs.append(VN_XRESULT);
	xpostArgs.append(")"); 
	args.append(")");
    }

    /** Returns code for invoking the original method.
     *  I.e., <code>[ VN_RESULT = ] internal$m(x1,...,xn);</code>.
     */
    protected /*@ non_null @*/ String callOriginalMethod() {  
        String prefix = MN_INTERNAL;
        if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
            if(methodDecl.isStatic()) {
                prefix = "$chx_Orig_" + typeName + ".";
            } else {
                prefix = "(($chx_Orig_" + typeName + ") _wrapped_object).";
            }
        }

 	String code = "    ";
	code = code + (hasReturn ? VN_RESULT + " = " : "");
	code = code + prefix + ident + args + ";\n";
	return code;
    }

    /** Returns code for declaring result variable. I.e.,
     * <code>T VN_RESULT = init_value_of_T;</code>, 
     * where <code>T</code> is the return type.
     */
    protected /*@ non_null @*/ String resultVarDecl() {
	String result = "";
	if (hasReturn)
	    result = "  " + TransUtils.toString(returnType) + " " + VN_RESULT + " = " + 
		TransExpression.defaultValue(returnType) + ";\n";
	return result;
    }

    /** Returns code for checking precondition. The
     * returned code has the following structure.
     *
     * <pre>
     *  // check precondition
     *  if (JMLChecker.isActive(JMLChecker.PRECONDITION)) {
     *    JMLChecker.enter();
     *    checkPre$m$T(x1,...,xn);
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>m</code> is the name of method being translated and
     * <code>T</code> is the name of the owning type.
     */
    protected /*@ non_null @*/ String checkPrecondition() {
	return wrap("PRECONDITION", 
		    "check precondition", 
                    racMethodName(MN_CHECK_PRE) + args);
    }

    /** Returns code for checking normal postcondition.  The returned
     * code has the following structure.
     *
     * <pre>
     *  // check normal postcondition
     *  if (JMLChecker.isActive(JMLChecker.POSTCONDITION)) {
     *    JMLChecker.enter();
     *    checkPost$m$T(x1,...,xn,r);
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>m</code> is the name of method being translated and
     * <code>T</code> is the name of the owning type.
     */
    protected /*@ non_null @*/ String checkPostcondition() {
	return wrap2("POSTCONDITION",
		     "check normal postcondition",
                    racMethodName(MN_CHECK_POST) + postArgs);
    }

    /** Returns code for checking exceptional postcondition.  The returned
     * code has the following structure.
     *
     * <pre>
     *  // check exceptional postcondition
     *  if (JMLChecker.isActive(JMLChecker.POSTCONDITION)) {
     *    JMLChecker.enter();
     *    checkXPost$m$T(x1,...,xn,r,e);
     *    JMLChecker.exit();
     *  }
     * </pre>
     * where <code>m</code> is the name of method being translated and
     * <code>T</code> is the name of the owning type.
     */
    protected /*@ non_null @*/ String checkXPostcondition() {
	return wrap3("POSTCONDITION", 
		     "check exceptional postcondition",
		     racMethodName(MN_CHECK_XPOST) + xpostArgs);
    }

    /** Returns a mandatory <code>return</code> statement. I.e,
     * <code>return [ VN_RESULT ];</code>. The optional result
     * variable is generated only if the method being translated has a
     * non-void return type.
     */
    protected /*@ non_null @*/ String mandatoryReturn() {
	if (hasReturn)
	    return "    return " + VN_RESULT + ";\n";
	else
	    return "    return;\n";
    }

    /** Returns an optional <code>return</code> statement. I.e,
     * <code>[ return VN_RESULT ; ]</code>. If the method being
     * translated has a void return type, the result is an empty
     * string.
     */
    protected /*@ non_null @*/ String optionalReturn() {
	if (hasReturn)
	    return "  return " + VN_RESULT + ";\n";
	else
	    return "";
    }

    /** Returns the given statement, <code>stmt</code>, wrapped in
     * assertion check demarcation. The returned code has the
     * following structure.
     *
     * <pre>
     *  // comment
     *  if (JMLChecker.isActive(JMLChecker.type)) {
     *    JMLChecker.enter();
     *    stmt;
     *    JMLChecker.exit();
     *  }
     * </post>
     *
     * @param type assertion type, e.g., PRECONDITION, etc.
     * @param comment header comment
     * @param stmt statement to be wrapped
     */
    protected /*@ non_null @*/ String wrap(String type, String comment, 
                                           String stmt) {
	return
	    "  // " + comment + "\n" +
	    "  if (" + isActive(type) + ") {\n" +
	    "    JMLChecker.enter();\n" + 
	    "    " + stmt + ";\n" +
	    "    JMLChecker.exit();\n" +
	    "  }\n";
    }
    
    /** Returns the given statement, <code>stmt</code>, wrapped in
     * assertion check demarcation. The whole code is indented two
     * tabs. The returned code has the following structure.
     *
     * <pre>
     *  // comment
     *  if (JMLChecker.isActive(JMLChecker.type)) {
     *    JMLChecker.enter();
     *    stmt;
     *    JMLChecker.exit();
     *  }
     * </post>
     *
     * @param type assertion type, e.g., PRECONDITION, etc.
     * @param comment header comment
     * @param stmt statement to be wrapped
     */
    protected /*@ non_null @*/ String wrap2(String type, String comment, 
                                            String stmt) {
	return
	    "    // " + comment + "\n" +
	    "    if (" + isActive(type) + ") {\n" +
	    "      JMLChecker.enter();\n" + 
	    "      " + stmt + ";\n" +
	    "      JMLChecker.exit();\n" +
	    "    }\n";
    }

    /** Returns the given statement, <code>stmt</code>, wrapped in
     * assertion check demarcation. The whole code is indented three
     * tabs. The returned code has the following structure.
     *
     * <pre>
     *  // comment
     *  if (JMLChecker.isActive(JMLChecker.type)) {
     *    JMLChecker.enter();
     *    stmt;
     *    JMLChecker.exit();
     *  }
     * </post>
     *
     * @param type assertion type, e.g., PRECONDITION, etc.
     * @param comment header comment
     * @param stmt statement to be wrapped
     */
    protected /*@ non_null @*/ String wrap3(String type, String comment,
                                            String stmt) {
	return
	    "      // " + comment + "\n" +
	    "      if (" + isActive(type) + ") {\n" +
	    "        JMLChecker.enter();\n" + 
	    "        " + stmt + ";\n" +
	    "        JMLChecker.exit();\n" +
	    "      }\n";
    }

    /**
     * Returns a string representation of the condition that tests if
     * the given type of assertion is active. 
     *
     * <pre><jml>
     * ensures \result.equals("JMLChecker.isActive(JMLChecker." + type +")"); 
     * </jml></pre>
     *
     */
    protected /*@ non_null @*/ String isActive(/*@ non_null @*/ String type) {
        if(JmlRacGenerator.checking_mode == JmlRacGenerator.WRAPPER) {
            String equiv_method = null;
            if(type == "PRECONDITION") {
                equiv_method = "Precondition";
            } else if(type == "POSTCONDITION") {
                equiv_method = "Postcondition";
            } else if(type == "INVARIANT") {
                equiv_method = "Invariant";
            }
            if(equiv_method != null) {
                return "$chx_Statics_" + this.typeName +  
                            ".access.isCheck" + equiv_method + "()";
            }
        }
        return "JMLChecker.isActive(JMLChecker." + type + ")";
    }

    /** Returns a string that identifies the method being translated. */
    protected /*@ non_null @*/ String identification(String state) {
        return methodDecl.ident() + "@" + state + "<" + 
            escapeString(methodDecl.getTokenReference().toString()) + ">";
    }
    
    /**
     * Returns the name of runtime assertion checker method for the
     * given type. 
     * 
     * <pre><jml>
     * requires (* type is MN_CHECK_PRE, MN_CHECK_POST, MN_CHECK_XPOST, etc.*);
     * ensures \result.equals(type + ident + "$" + typeName);
     * </jml></pre>
     */
    protected /*@ non_null @*/ String racMethodName(String type) {
        return type + ident + "$" + typeName;
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** type name whose methods are to be wrapped. This field is used
     * to generate the names of various assertion check methods. */
    protected final String typeName;

    /** target method to be wrapped.
     *
     * <pre><jml>
     * protected invariant methodDecl != null;
     * </jml></pre>
     */
    /*@ spec_public @*/ protected final JmlMethodDeclaration methodDecl;


    /** true if the target method is a helper.
     *
     * <pre><jml>
     * protected invariant isHelper == methodDecl.isHelper();
     * </jml></pre>
     */
    protected final boolean isHelper;

    /** true if the target method is a static method.
     *
     * <pre><jml>
     * protected invariant isStatic == methodDecl.isStatic();
     * </jml></pre>
     */
    protected final boolean isStatic;

    /** true if the target method has return type.
     *
     * <pre><jml>
     * private invariant hasReturn == 
     *   (returnType != null && !returnType.isVoid());
     * </jml></pre>
     */
    private final boolean hasReturn;

    /** return type of the target method.
     *
     * <pre><jml>
     * private invariant returnType == methodDecl.returnType();
     * </jml></pre>
     */
    protected final CType returnType;

    /** name of the target method.
     *
     * <pre><jml>
     * private invariant ident == methodDecl.ident();
     * </jml></pre>
     */
    protected final String ident;

    /** arguments to pre/orig, post, or xpost method calls */
    /*@ spec_public @*/ protected StringBuffer args, postArgs, xpostArgs;
}

