// @(#)$Id: JMLAssertionError.java,v 1.8 2006/06/07 02:11:02 f_rioux Exp $

// Copyright (C) 2005 Iowa State University
//
// This file is part of the runtime library of the Java Modeling Language.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with JML; see the file LesserGPL.txt.  If not, write to the Free
// Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA.

package org.jmlspecs.jmlrac.runtime;

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

/**
 * An abstract error class to notify all kinds of runtime assertion
 * violations.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.8 $
 */
public abstract class JMLAssertionError extends Error
{
    /** Name of the class or interface that contains the violated assertion.
     *
     * <pre><jml>
     * invariant className != null;
     * </jml></pre>
     */
    /*@ spec_public @*/ protected String className;
    
    /** Name of the method that contains the violated assertion. 
     * It can be <code>null</code> if the assertion is not associated
     * with any particular methods, e.g., invariants or history constraints.
     */
    /*@ spec_public @*/ protected String methodName;

    /** The locations (strings of the form: "file_name:line_num:column_num")
     * of the violated assertion. The violated assertion might be specified
     * in several different places due to specification inheritance.
     *
     * <pre><jml>
     * invariant locations != null && 
     *  (\forall Object o; locations.contains(o); o instanceof String);
     * </jml></pre>
     */
    /*@ spec_public @*/ protected /*@ non_null @*/ Set locations;

    /**
     * Creates a new <code>JMLAssertionError</code> instance with given
     * message.
     *
     * @param message an informative message; can be <code>null</code>.
     */
    protected JMLAssertionError (String message)
    {
	super(message);
	locations = new HashSet();
	className = "";
	methodName = "";
    }
    
    protected JMLAssertionError(String message, Throwable cause) {
		super(message, cause);
		locations = new HashSet();
		className = "";
		methodName = "";
	}
    
    protected JMLAssertionError(Throwable cause) {
		super(cause.getMessage(), cause);
		locations = new HashSet();
		className = "";
		methodName = "";
	}

    /**
     * Return the name of class that contains the violated assertion.
     * The result is not <code>null</code>.
     *
     * <pre><jml>
     *  ensures \result.equals(className);
     * </jml></pre>
     *
     * @return a class name
     */
    public String className() {
	return className;
    }
    
    /**
     * Return the name of method that contains the violated assertion.
     * The result may be <code>null</code>, e.g., in case of invariant
     * or history constraint violations.
     *
     * <pre><jml>
     *  ensures \result == null || \result.equals(methodName);
     * </jml></pre>
     *
     * @return a method name
     */
    public String methodName() {
	return methodName;
    }

    /**
     * Return the associated message.
     *
     * @return a <code>String</code> message
     */
    public String message() {
	return super.getMessage();
    }

    /**
     * Return the locations. The result is not <code>null</code> and
     * contains strings of the form: "file_name:line_no:column_no".
     *
     * <pre><jml>
     *  ensures \result == locations;
     * </jml></pre>
     *
     * @return locations
     */
    public Set locations() {
	return locations;
    }
    
    /**
     * Return the associated message concatenated with location information.
     *
     * <pre><jml>
     * also
     *   ensures \result != null;
     * </jml></pre>
     * @return a <code>String</code> message
     */
    public String getMessage() {
	String msg = message() == null ? "" : (message() + ": ");
	msg = msg + "by method " + className ;
        msg = msg + "." + methodName;

	if (!locations.isEmpty()) {
	    msg = msg + " regarding specifications at";
	    Iterator iter = locations.iterator();
	    while (iter.hasNext())
		msg = msg + "\n" + iter.next();
	}
	return msg;
    }

}
