/* $Id: JMLChecker.java,v 1.25 2007/02/03 02:04:50 delara Exp $
 *
// Copyright (C) 2005 Iowa State University
//
// This file is part of the runtime library of the Java Modeling Language.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with JML; see the file LesserGPL.txt.  If not, write to the Free
// Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA.
 */

package org.jmlspecs.jmlrac.runtime;

import java.lang.reflect.*;
import java.util.*;

/** 
 * A class to set various runtime options and to check and report
 * runtime assertion violations.
 *
 *  @author Yoonsik Cheon
 *  @version $Revision: 1.25 $
 */

public abstract class JMLChecker implements JMLOption {

    // ----------------------------------------------------------------------
    // STATIC METHODS
    // ----------------------------------------------------------------------

    /** Returns true only if the given class, <code>clazz</code>, is
     * compiled with runtime assertion check code. 
     */
    public static boolean isRACCompiled(/*@ non_null @*/ Class clazz) {
        // Be warned that the field name here should be the same as
        // the constant RacConstants.VN_RAC_COMPILED.
        try {
            return clazz.getDeclaredField("rac$RAC_COMPILED") != null;
        } catch (NoSuchFieldException e) {
            return false;
        } catch (SecurityException e) {
            return false;
        }
    }

    /** Sets the global runtime assertion check level to the level
     * <code>l</code>.
     *
     * <pre><jml>
     * normal_behavior
     *   requires l == NONE || l == PRECONDITIONS_ONLY || l == ALL;
     *   assignable level;
     *   ensures level == l;
     * </jml></pre>
     */
    public static void setLevel(int l) {
	level = l;
    }

    /** Returns the current global runtime assertion check level.  The
     * returned value is <code>NONE</code>,
     * <code>PRECONDITIONS_ONLY</code>, or <code>ALL</code>.
     *
     * <pre><jml>
     * normal_behavior
     *   assignable \nothing;
     *   ensures \result == level;
     * </jml></pre>
     */
    public static int getLevel() {
      return level;
    }

    /** Sets the runtime assertion check level of the class,
     * <code>clazz</code>, to the level, <code>l</code>. If no such
     * class exists, this method has no effect.
     *
     * <pre><jml>
     * normal_behavior
     *   requires l == NONE || l == PRECONDITIONS_ONLY || l == ALL;
     *   assignable \not_specified;
     * </jml></pre>
     */
    public static void setLevel(/*@ non_null @*/ Class clazz, int l) {
        // Be warned that the field name here should be the same as
        // the constant RacConstants.VN_RAC_LEVEL.
        try {
            Class cls = getRacClass(clazz);
            java.lang.reflect.Field f = cls.getDeclaredField("rac$level");
            f.setInt(cls, l);
        } catch (Exception e) {
        }
    }
    
    /** Returns the runtime assertion checker class of the given class,
     * <code>clazz</code>. If the given class is an interface, its
     * surrogate class is returned; otherwise, the given class itself
     * is returned.
     */
    private static Class getRacClass(/*@ non_null @*/ Class clazz) {
        if (clazz.isInterface()) {
            Class[] classes = clazz.getDeclaredClasses();
            clazz = null;
            for (int i = 0; i < classes.length; i++) {
                if (classes[i].getName().endsWith("$JmlSurrogate")) {
                    clazz = classes[i];
                    break;
                }
            }
        }
        return clazz;
    }

    /**
     * Returns true if a method named <code>name</code> with the
     * formal parameter types, <code>params</code>, can be inherited
     * from the type, <code>class</code>, or its supertypes.  If
     * either <code>clazz</code> or <code>name</code> is null, false
     * is returned.
     */
    public static boolean inheritedFrom(Class clazz, String name, 
                                        /*@ non_null @*/ Class[] params) {
        if (clazz == null || name == null) {
            return false;
        }

        // does the class declare a non-static and public or proected method?
        try {
            Method meth = clazz.getDeclaredMethod(name, params);
            if (meth != null) {
                int mod = meth.getModifiers();
                if (!Modifier.isStatic(mod) && 
                    (Modifier.isPublic(mod) || Modifier.isProtected(mod))) {
                    return true;
                }
            }        
        } catch (Exception e) {
        }

        // check superclass
        if (inheritedFrom(clazz.getSuperclass(), name, params)) {
            return true;
        }

        // check interfaces
        Class[] interfaces = clazz.getInterfaces();
        for (int i = 0; i < interfaces.length; i++) {
            if (inheritedFrom(interfaces[i], name, params)) 
                return true;
        }

        return false;
    }

    /** Returns the current runtime assertion check level of the given
     * class, <code>clazz</code>. The returned value is
     * <code>NONE</code>, <code>PRECONDITIONS_ONLY</code>, or
     * <code>ALL</code>. If no such class is found or the class is not
     * compiled with runtime assertion check code, <code>NONE</code>
     * is returned.
     */
    public static int getLevel(/*@ non_null @*/ Class clazz) {
        try {
            return getRacClass(clazz).
                getDeclaredField("rac$level").getInt(clazz);
        } catch (Exception e) {
            return NONE;
        }
    }

    /**
     * Indicates whether an assumption violation should be reported to
     * the user by throwing an appropriate exception (i.e., 
     * <code>JMLIntraconditionError</code>). 
     * The default is to report assumption violations.
     *
     * @see #reportAssumptionViolation(boolean)
     */
    public static boolean reportAssumptionViolation() {
	return reportAssumptionViolation;
    }

    /**
     * Sets whether an assumption violation should be reported to
     * the user by throwing an appropriate exception (i.e., 
     * <code>JMLIntraconditionError</code>). 
     * The default is to report assumption violations.
     *
     * @see #reportAssumptionViolation()
     */
    public static void reportAssumptionViolation(boolean flag) {
	reportAssumptionViolation = flag;
    }

    /** Indicates the start of assertion check by the current thread.
     * <pre><jml>
     * normal_behavior
     *   old Thread cur = Thread.currentThread();
     *   requires !threadsBeingChecked.contains(cur);
     *   assignable threadsBeingChecked;
     *   ensures (* threadsBeingChecked.add(cur) *);
     * </jml></pre>
     */
    public synchronized static void enter() {
	threadsBeingChecked.add(Thread.currentThread());
    }

    /** Indicates the end of assertion check for the current thread.
     *
     * <pre><jml>
     * normal_behavior
     *   old Thread cur = Thread.currentThread();
     *   requires threadsBeingChecked.contains(cur);
     *   assignable threadsBeingChecked;
     *   ensures (* threadsBeingChecked.remove(cur) *);
     * </jml></pre>
     */
    public synchronized static void exit() {
	threadsBeingChecked.remove(Thread.currentThread());
    }

    /** Determines if the runtime code for checking this type of assertion
     *  should be executed. It depends on three factors
     *  1. Assertion check is not turned off at runtime 
     *  2. Runtime option permits you to execute this code
     *  3. This method is not a part of another assertion already being 
     *     checked in this thread
     *
     *  @param type Type of the current assertion check (e.g., PRECONDITION)
     *  @return true if the current assertion can be checked; false otherwise.
     */ 
    public synchronized static boolean isActive(int type) {
	return (level != NONE && level >= type
		&& !threadsBeingChecked.contains(Thread.currentThread()));
    }

    /**
     * Returns a surrogate object, an instance of the given surrogate
     * class <code>clazz</code>, that is suitable for statically
     * calling JML access methods on the object <code>thisObj</code>.
     * The argument <code>clazz</code> must be a surrogate class.  If
     * the object <code>thisObj</code> itself is a surrogate, the
     * delegee's surrogate cache is looked upon for an existing
     * surrogate object or creating a new one; otherwise, the object's
     * surrogate cache is looked upon. The arument <code>name</code>
     * equals to clazz.getName(); it is passed in for the performance
     * reason, i.e., to prevent extra calls to the method getName().
     */
    public static Object getSurrogate(/* non_null @*/ String name, 
                                      /* non_null @*/ Class clazz, 
                                      /* non_null @*/ Object thisObj) {

        //@ assume JMLSurrogate.class.isAssignableFrom(clazz);
        //@ assume_redundantly (* clazz is a surrogate class *);

        // decide the delegee object
        Object delegee = null;
        if (thisObj instanceof JMLSurrogate) {
            delegee = ((JMLSurrogate) thisObj).getSelf();
        } else {
            delegee = thisObj;
        }

        // delegee's class not compiled with jmlc?
        if (!(delegee instanceof JMLCheckable)) {
            throw new JMLNonExecutableException();
        }

        try {
            //@ assume delegee instanceof JMLCheckable;
            JMLCheckable cobj = (JMLCheckable) delegee;

            // lookup delegee's surrogate cache
            Object surObj = cobj.rac$getSurrogate(name);

            if (surObj == null) {
                // create a new surrogate object for thisObj
                Class[] types = new Class[] { JMLCheckable.class };
                Constructor constr = clazz.getConstructor(types);
                Object[] args = new Object[] { cobj };
                surObj = constr.newInstance(args);

                // store the surrogate into thisObj's cache
                cobj.rac$setSurrogate(name, (JMLSurrogate)surObj);
            }
            return surObj;
        }
        catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }

        //@ unreachable;
        return null;
    }

    /**
     * Returns the result of invoking the <code>toString</code> method
     * to the argument, <code>obj</code>. If the argument is null, "null"
     * is returned. If the invocation encounters an exception,
     * "UNKNOWN" is returned; otherwise, the result of invocation
     * is returned.
     */
    public static /*@ non_null @*/ String toString(Object obj) {
        if (obj == null) {
            return "null";
        }

        String result = "<UNKNOWN-to-JML>";
        try {
            result = obj.toString();
        } 
        catch (java.lang.Throwable e1) {
            String addr = "<UNKNOWN-to-JML>";
            try {
                addr = "" + obj.hashCode();
            }
            catch (java.lang.Throwable e2) {
            }
            result = obj.getClass().getName() + "@" + addr;
        }
        return result;
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(boolean i) {
        return "" + i;
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(byte i) {
        return "" + i;
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(char i) {
        return "" + i;
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(int i) {
        return "" + i;
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(short i) {
        return "" + i;
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(long i) {
        return "" + i;
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(float value) {
        if (value == Float.POSITIVE_INFINITY) {
            return "java.lang.Float.POSITIVE_INFINITY";
        } else if (value == Float.NEGATIVE_INFINITY) {
            return "java.lang.Float.NEGATIVE_INFINITY";
        } else if (Float.isNaN(value)) {
            return "java.lang.Float.NaN";
        } else if (value == 0.0F && (new Float(value).equals(new Float("-0.0")))) {
            return "-0.0F";
        } else if (value == Float.MAX_VALUE) {
            return "java.lang.Float.MAX_VALUE";
        } else if (value == Float.MIN_VALUE) {
            return "java.lang.Float.MIN_VALUE";
        } else if (value == 0.0F) {
            return "+0.0F";
        } else {
            return value + "F";
        }
    }

    /** Returns the string representation of the given argument. 
     */
    public static /*@ non_null @*/ String toString(double value) {
        if (value == Double.POSITIVE_INFINITY) {
            return "java.lang.Double.POSITIVE_INFINITY";
        } else if (value == Double.NEGATIVE_INFINITY) {
            return "java.lang.Double.NEGATIVE_INFINITY";
        } else if (Double.isNaN(value)) {
            return "java.lang.Double.NaN";
        } else if (value == Double.MAX_VALUE) {
            return "java.lang.Double.MAX_VALUE";
        } else if (value == Double.MIN_VALUE) {
            return "java.lang.Double.MIN_VALUE";
        } else if (value == 0.0D && (new Double(value).equals(new Double("-0.0")))) {
            return "-0.0D";
        } else if (value == 0.0D) {
            return "+0.0D";
        } else {
            return value + "D";
        }
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** 
     * Runtime option indicating the assertion check level, 
     * the kinds of assertions to be checked. 
     * The default is <code>JMLOption.ALL</code>.
     *
     * <pre><jml>
     *  public invariant level == JMLOption.NONE || 
     *    level == JMLOption.PRECONDITIONS_ONLY ||
     *    level == JMLOption.ALL;
     * </jml></pre>
     *
     */
    protected static /*@ spec_public @*/ int level = ALL;

    /** Stores a flag per thread to turn assertion check on/off for every 
     *  thread to avoid recursive assertion checking. The flag is "created"
     *  for every thread on demand basis taking default value to be true.
     */
    protected static /*@ spec_public non_null @*/ Set threadsBeingChecked 
        = new HashSet(357);

    /** Indicates whether an assumption violation should be reported to
     * the user by throwing an appropriate exception */
    private static boolean reportAssumptionViolation = true;

    /** A map from fully quallified class names to class objects. If a
     * class name is mapped to the object <code>NO_CLASS</code>, it
     * means that there is no such named class. This map is used by
     * the generated method <code>rac$forName(String)</code>, to save
     * class lookup and load time. */
    public static HashMap classes = new HashMap(137);

    /** A unique object to indicate non-existing class. This constant
     * is used as an entry in the map <code>classes</code>. */
    public static final Object NO_CLASS = new Object();

    /** An exception object indicating angelic undefinedness. Instead
     * of creating a new exception object for each occurrence of
     * angelic undefinedness, this global constant is used. Creating
     * an exception object is costly. */
    public static final JMLNonExecutableException ANGELIC_EXCEPTION 
        = new JMLNonExecutableException();

    /** An exception object indicating demonic undefinedness. Instead
     * of creating a new exception object for each occurrence of
     * demonic undefinedness, this global constant is used. Creating
     * an exception object is costly. */
    public static final RuntimeException DEMONIC_EXCEPTION 
        = new RuntimeException();

    public static final int PRECONDITION = 1; 
    public static final int POSTCONDITION = 2;
    public static final int INVARIANT = 3;
    public static final int CONSTRAINT = 4;
    public static final int INTRACONDITION = 5; 
    public static final int PROTOCOL = 6; 
    
    // Coverage methods
    
        /** A map to hold information about coverage */
    static protected java.util.Map coverage = new java.util.TreeMap();

    static public class CoverageCount {
	public String location;
	public int trueCount;
	public int falseCount;
	public CoverageCount(String loc) { location = loc; }
	public void incr(boolean v) {
	    if (v) ++trueCount; else ++falseCount;
	}
    }
    
    static public void clearCoverage() { coverage = new java.util.TreeMap(); }
    
    static public void addCoverage(String location, boolean value) {
	//java.lang.System.out.println("Adding " + location + " " + value);
	Object o = coverage.get(location);
	if ( o == null) {
	    o = new CoverageCount(location);
	    coverage.put(location,o);
	}
	((CoverageCount)o).incr(value);
    }
    
    static public void reportCoverage(java.io.PrintStream writer) {
	java.util.Iterator i = coverage.values().iterator();
	while (i.hasNext()) {
	    CoverageCount c = (CoverageCount)(i.next());
	    if (c.trueCount == 0 && c.falseCount == 0) {
		writer.println(c.location + ": The requires clause is NEVER executed");
	    } else if (c.trueCount == 0 && c.falseCount > 0) {
		writer.println(c.location + ": The requires clause is always FALSE ("
			+ c.falseCount + " tests)");
	    } else if (c.trueCount > 0 && c.falseCount == 0) {
		writer.println(c.location + ": The requires clause is always TRUE ("
			+ c.trueCount + " tests)");
	    } else {
		//writer.println(c.location + " has " + c.trueCount + " true and " + c.falseCount + " false tests");
	    }
	}
    }


}
