/* $Id: JMLOldExpressionCache.java,v 1.3 2005/12/23 17:02:07 chalin Exp $
 *
// Copyright (C) 2005 Iowa State University
//
// This file is part of the runtime library of the Java Modeling Language.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with JML; see the file LesserGPL.txt.  If not, write to the Free
// Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA.
 */

package org.jmlspecs.jmlrac.runtime;

import java.util.*;

/**
 * An abstraction of caches for JML old expressions. This class implements 
 * a simple map that is suitable for storing the values of old expressions 
 * evaluated in the pre-state. If an old expression appears in 
 * a quantifier, we can view it as a mapping from quantified (bound)
 * variables to values. An object of this class is used to store this
 * mapping.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.3 $
 */

public class JMLOldExpressionCache {
    
    /** Constructs a new, empty cache object. */
    public JMLOldExpressionCache() {
        cache = new HashMap();
    }

    /** Does this cache contain a value for the key, <code>key</code>? 
     * <pre><jml>
     * requires key != null;
     * </jml></pre>
     */
    public boolean containsKey(Object key) {
        return containsKey(new Object[] { key });
    }

    /** Does this cache contain a value for the key, <code>key</code>? 
     * <pre><jml>
     * requires key != null;
     * </jml></pre>
     */
    public boolean containsKey(Object[] key) {
        return cache.containsKey(new Key(key));
    }

    /** Clears this cache. */
    public void clear() {
        cache.clear();
    }

    /** Puts the key/value pair to this cache. 
     * <pre><jml>
     * requires key != null && value != null;
     * </jml></pre>
     */
    public void put(Object key, Object value) {
        put(new Object[] { key }, value);
    }

    /** Puts the key/value pair to this cache. 
     * <pre><jml>
     * requires key != null && value != null;
     * </jml></pre>
     */
    public void put(Object[] key, Object value) {
        cache.put(new Key(key), value);
    }

    /** Returns the value for the key, <code>key</code>. If no value
     * is found for the given key, <code>null</code> is returned.
     * <pre><jml>
     * requires key != null;
     * </jml></pre>
     */
    public Object get(Object key) {
        return get(new Object[] { key });
    }

    /** Returns the value for the key, <code>key</code>. If no value
     * is found for the given key, <code>null</code> is returned.
     * <pre><jml>
     * requires key != null;
     * </jml></pre>
     */
    public Object get(Object[] key) {
        return cache.get(new Key(key));
    }

    // ----------------------------------------------------------------------
    // INNER CLASS
    // ----------------------------------------------------------------------

    /** A class for representing keys for cache objects. We use arrays for
     * keys, but we treat two arrays equal if both have the same elements.
     */
    public static class Key {
        /** Constructs a new key object from an array. */
        public Key(Object[] values) {
            this.values = values;
        }
        
        /** Does the argument equal to this key object? */
        public boolean equals(/*@ nullable @*/ Object obj) {
            if (this == obj) {
                return true;
            } else if (obj == null || getClass() != obj.getClass()) {
                return false;
            }
            Key other = (Key) obj;
            if (values.length != other.values.length) {
                return false;
            }
            for (int i = 0; i < values.length; i++) {
                if (!values[i].equals(other.values[i]))
                    return false;
            }
            return true;
        }

        /** Returns the hash code for this key object. */
        public int hashCode () {
            int result = 17;
            for (int i = 0; i < values.length; i++) {
                result = 37 * result + values[i].hashCode();
            }
            return result;
        }

        public Object[] values;
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** Map for (key, value) pairs. */
    private /*@ non_null @*/ Map cache;
}
