/* $Id: JMLRacValue.java,v 1.4 2005/07/07 21:03:03 leavens Exp $
 *
// Copyright (C) 2005 Iowa State University
//
// This file is part of the runtime library of the Java Modeling Language.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with JML; see the file LesserGPL.txt.  If not, write to the Free
// Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA.
 */

package org.jmlspecs.jmlrac.runtime;

/**
 * A class for denoting JML expressible values for the runtime
 * assertion checker. In addition to regular Java values and objects,
 * this class can represent two special JML values: undefinedness and
 * non-executability. An undefined value denotes the value of an
 * expression that throws an exception during evaluation. Similarly, a
 * non-executable value denotes a non-executable JML expression.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.4 $
 */

public class JMLRacValue {

    // ----------------------------------------------------------------------
    // CONSRUCTORS AND FACTORY METHODS
    // ----------------------------------------------------------------------
    
    /**
     * Constructs a new instance. Note that the argument,
     * <code>value</code>, may be <code>null</code>.
     */
    private JMLRacValue(Object value) {
        this.value = value;
    }

    /**
     * Constructs a new instance. */
    private JMLRacValue() {
    }

    /**
     * Returns an instance that denotes an angelic undefinedness.
     */
    public static /*@ non_null @*/ JMLRacValue ofUndefined() {
        return DEMON;
    }

    /**
     * Returns an instance that denotes a demonic undefinedness.
     */
    public static /*@ non_null @*/ JMLRacValue ofNonExecutable() {
        return ANGEL;
    }

    /**
     * Returns a new JMLRacValue object that wraps the given argument.
     */
    public static /*@ non_null @*/ JMLRacValue ofObject(Object v) {
        return new JMLRacValue(v);
    }

    /**
     * Returns a new JMLRacValue object that wraps the given boolean
     * value, <code>v</code>.
     */
    public static /*@ non_null @*/ JMLRacValue ofBoolean(boolean v) {
        return new JMLRacValue(new Boolean(v));
    }

    /**
     * Returns a new JMLRacValue object that wraps the given int
     * value, <code>v</code>.
     */
    public static /*@ non_null @*/ JMLRacValue ofInt(int v) {
        return new JMLRacValue(new Integer(v));
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the wrapped value of this JMLRacValue object .
     */
    public /*@ pure @*/ Object value() {
        checkUndefinedness();
        return value;
    }
    
    /**
     * Returns true if this object represents an angelic undefinedness.
     */
    public /*@ pure @*/ boolean isAngel() {
        return this == ANGEL;
    }

    /**
     * Returns true if this object represents a demonic undefinedness.
     */
    public /*@ pure @*/ boolean isDemon() {
        return this == DEMON;
    }

    /**
     * Throws appropriate exception if this object represents either
     * angelic or demonic undefinedness.
     */
    private /*@ pure @*/ void checkUndefinedness() {
        if (this == DEMON) {
            throw new java.lang.RuntimeException("JML undefined value");
        }
        if (this == ANGEL) {
            throw new JMLNonExecutableException("JML non-executable value");
        }
    }
    
    /**
     * Returns the string representation of this object.
     */
    public /*@ non_null pure @*/ String toString() {
        if (isAngel()) {
            return "ANGELIC_UNDEF";
        } else if (isDemon()) {
            return "DEMONIC_UNDEF";
        }
        return value == null ? "null" : value.toString();
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------
    
    /**
     * A unique object to represent an angelic undefinedness such as
     * non-executable specification constructs.
     */
    private final static JMLRacValue ANGEL = new JMLRacValue(); 

    /**
     * A unique object to represent a demonic undefinedness such as
     * runtime exceptions.
     */   
    private final static JMLRacValue DEMON = new JMLRacValue();

    /**
     * The object wrapped by this instance of class
     * <code>JMLRacValue</code>. Note that it may be null.
     */
    private Object value = null;
}
