/* $Id: JMLSurrogate.java,v 1.11 2005/12/23 17:02:07 chalin Exp $
 *
// Copyright (C) 2005 Iowa State University
//
// This file is part of the runtime library of the Java Modeling Language.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with JML; see the file LesserGPL.txt.  If not, write to the Free
// Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA.
 */

package org.jmlspecs.jmlrac.runtime;

import java.lang.reflect.*;
import java.util.*;

/**
 * The common behavior of all surrogate classes. Each interface
 * surrogate class, generated by <code>jmlc</code>, is supposed to be
 * inherit behavior specified in this class.
 *
 * @author Yoonsik Cheon
 * @version $Revision: 1.11 $
 */

public abstract class JMLSurrogate {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a new surrogate for the given object, <code>obj</code>.
     */
    protected JMLSurrogate(/*@ non_null @*/ JMLCheckable self) {
        this.self = self;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /** Returns the delegee of this object. The delegee is the object
     * that this surrogate object exists for. */
    public /* non_null */ JMLCheckable getSelf() { 
        return self;
    }

    // ----------------------------------------------------------------------
    // ASSERTION CHECKING
    // ----------------------------------------------------------------------

    /**
     * Returns the receiver object for dynamic calls to methods of the
     * class <code>clazz</code> from the object <code>thisObj</code>.
     * If the object <code>thisObj</code> is a surrogate object, its
     * delegee is referenced to determine an existing, appropriate
     * object or create a new one; otherwise, the object itself is the
     * receiver. */
    protected static Object getReceiver(/* non_null @*/ Class clazz, 
                                     Object thisObj) {
        // dynamic call to a static method?
        if (thisObj == null) {
            return null;
        }

        // decide the delegee object
        Object delegee = null;
        if (thisObj instanceof JMLSurrogate) {
            // e.g., this.m(), m's owner better be is an interface
            //@ assert JMLSurrogate.class.isAssignableFrom(clazz);
            //@ assert_redundantly (* clazz is a surrogate class *);
            delegee = ((JMLSurrogate) thisObj).getSelf();
        } else {
            if (JMLSurrogate.class.isAssignableFrom(clazz)) {
                // e.g., other.m(), where m's owner is an interface
                delegee = thisObj;
            } else {
                // e.g., other.m(), where m's owner is a class
                return thisObj;
            }            
        }

        //@ assume delegee instanceof JMLCheckable;
        if (!(delegee instanceof JMLCheckable)) {
            // delegee not RAC-compiled, e.g., an instance of an inner class.
            throw new JMLNonExecutableException();
        }
        JMLCheckable cobj = (JMLCheckable) delegee;

        try {
            // lookup thisObj's surrogate cache
            Object surObj = cobj.rac$getSurrogate(clazz.getName());

            if (surObj == null) {
                // create a new surrogate object for thisObj
                Class[] types = new Class[] { JMLCheckable.class };
                Constructor constr = clazz.getConstructor(types);
                Object[] args = new Object[] { cobj };
                surObj = constr.newInstance(args);

                // store the surrogate into thisObj's cache
                cobj.rac$setSurrogate(clazz.getName(), (JMLSurrogate)surObj);
            }
            return surObj;
        }
        catch (Exception e) {
            System.err.println("Internal error!" +
                               "JMLSurrogate.getreceiver failed.");
            System.exit(1);
        }

        return null;
    }
    
    /**
     * Returns a method object that reflects the specified public
     * member method of the class or interface, <code>clazz</code>.
     * The name parameter is a string specifying the simple name of
     * the desired method. The types parameter is an array of class
     * objects that identify the method's formal parameter types, in
     * declared order. If it is null, it is treated as if it were an
     * empty array. */
    public static /*@ non_null @*/ Method getMethod(
        /*@ non_null @*/ Class clazz, 
        /*@ non_null @*/ String name,
        Class[] types) 
        throws NoSuchMethodException, SecurityException {

        // key for assertion check and accessor methods
        MapKey key = MapKey.create(clazz, name, types);
        Object result = methods.get(key);

        if (result == NULL_METHOD) {
            throw new NoSuchMethodException(name);
        } else if (result != null) {
            return (Method) result;
        } else {
            try {
                Method meth = clazz.getMethod(name, types);            
                methods.put(key, meth);
                return meth;
            }
            catch (Exception e) {
                methods.put(key, NULL_METHOD);
                if (e instanceof RuntimeException) {
                    throw (RuntimeException) e;
                } else {
                    throw new NoSuchMethodException(name);
                }
            }
        }
    }

    /**
     * Returns the accessor (getter) method of a ghost field or a
     * model field, <code>name</code>, declared in the type,
     * <code>clazz</code>.  This method looks for the method first at
     * the local cache, and then, if fails, the class object
     * itself. */
    public static /*@ non_null @*/ Method getAccessor(
        /*@ non_null @*/ Class clazz, 
        /*@ non_null @*/ String name)
        throws NoSuchMethodException, SecurityException {
        return getMethod(clazz, name, null);
    }

    /**
     * Returns the setter method of the ghost field,
     * <code>name</code>, of type <code>type</code>, declared in
     * class, <code>clazz</code>.  This method looks for the method
     * first at the local cache, and then, if fails, the class object
     * itself. */
    public static /*@ non_null @*/ Method getSetter(
        /*@ non_null @*/ Class clazz, 
        /*@ non_null @*/ String name,
        /*@ non_null @*/ Class type)
        throws NoSuchMethodException, SecurityException {

        return getMethod(clazz, name, new Class[] { type });
    }

    // ----------------------------------------------------------------------
    // INNER CLASSES
    // ----------------------------------------------------------------------

    /** A class for implementing keys for the method cache. */
    /*@ spec_public @*/ private static class MapKey {

        /** Constructs a new key for a method named <code>name</code>
         * with formal parameter types, <code>types</code>, declared in
         * the class, <code>clazz</code>. 
         *
         * <pre><jml>
         * requires clazz != null && name != null;
         * </jml></pre>
         */
        private MapKey(Class clazz, String name, Class types[]) {
            this.clazz = clazz;
            this.name = name;
            this.types = types;
        }

        /** Returns a new key for a method named <code>name</code>
         * with formal parameter types, <code>types</code>, declared in
         * the class, <code>clazz</code>. 
         *
         * <pre><jml>
         * requires clazz != null && name != null;
         * </jml></pre>
         */
        public static MapKey create(Class clazz, String name, 
                                    Class types[]) {
            return new MapKey(clazz, name, types);
        }

        /** Returns the hash code for this key object. */
        public int hashCode() {
            int hash = 0;
            if (types != null) {
                for (int i = 0; i < types.length; i++) 
                    hash += types[i].hashCode();
            }
            return clazz.hashCode() + name.hashCode() + hash;
        }

        /** Does the argument equal to this object? */
        public boolean equals(/*@ nullable @*/ Object obj) {
            if (this == obj) {
                return true;
            } else if (obj == null || MapKey.class != obj.getClass()) {
                return false;
            }
            MapKey other = (MapKey) obj;
            try {
                if (clazz != other.clazz || !name.equals(other.name)) {
                    return false;
                } else if (types == null && other.types == null) {
                    return true;
                } else if (types.length != other.types.length) {
                    return false;
                }
                for (int i = 0; i < types.length; i++) 
                    if (types[i] != other.types[i])
                        return false;
            } catch (NullPointerException e) {
                return false;
            }
            return true;
        }

        /*@ non_null @*/ Class clazz;
        /*@ non_null @*/ String name;
        Class types[];            
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** The shared cache for dynamically looking up methods. This map
     * contains assertion check methods, accessor methods, and setter
     * methods that are dynamically looked up for runtime assertion
     * checking. The keys are instances of the class KeyMap.
     *
     * <pre><jml>
     * public invariant methods != null &&
     *   (\forall MapKey k; methods.containsKey(k); 
     *      methods.get(k) instanceof Method);
     * </jml></pre>
     */
    public final static java.util.Map methods = new java.util.HashMap();

    /** object that this object is surrogate for. */
    protected /*@ non_null @*/ JMLCheckable self;

    /** A special object denoting no method. */
    private final static Object NULL_METHOD = new Object();

}
