// @(#)$Id: PriorityQueue.java,v 1.2 2005/12/23 17:02:07 chalin Exp $

// Copyright (C) 1998, 1999 Iowa State University

// This file is part of JML

// JML is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.

// JML is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with JML; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


package org.jmlspecs.jmlrac.testcase.bugs.priorityqueue;

//@ refine "PriorityQueue.jml-refined";

import org.jmlspecs.models.*;

/** An implementation of priority queues.
 * @author Albert L. Baker.
 */
public class PriorityQueue {
  // The non-defensive version - relies on caller to ensure
  // preconditions to protect the invariants.

  //**************************************************************
  // Implementation data structures

    public class EntryNode {
	//@ public model JMLDataGroup entryData;
	public String        iD;
	//@                    in entryData;

	public EntryNode     next;
	/*@                    in entryData;
	  @                    maps next.entryData \into entryData;
	  @*/

	//@ public ghost int priority;

	//@ public ghost int timeStamp;

	//@ public model int numberOfEntries;       in entryData;

	/*@ public represents numberOfEntries <- 
	  @        (next == null)? 1 : (int)(next.numberOfEntries + 1);
	  @*/
    }

    protected class PriorityListNode {
	//@ public model JMLDataGroup priorityData;

	/*@ spec_public @*/
	int      priorityLevel;
	//@              in priorityData;

	/*@ spec_public @*/ 
	EntryNode  entries;
	//@            maps entries.entryData \into priorityData;

	/*@ spec_public @*/ 
	PriorityListNode next;
	/*@                in priorityData;
	  @                maps next.priorityData \into priorityData;
	  @*/

	//@ public model int numberOfLevels;

	/*@ public represents numberOfLevels <- 
	  @        (next == null)? 1 : (int)(next.numberOfLevels + 1);
	  @*/

	//@ public model int numberOfEntries;             in priorityData;

	/*@ public represents numberOfEntries <- 
	  @        (entries == null)? 0 : (int)(entries.numberOfEntries + 1);
	  @*/

    }

  protected PriorityListNode theEntries;
  //@            in entries;
  //@            maps theEntries.priorityData \into entries;

  // The following model field is used in "measured_by" clause of recursive
  // model method isPriorityNode:
  //@ public model int numberOfLevels;

  /*@ protected represents numberOfLevels <- 
    @       (theEntries == null)? 0 : theEntries.numberOfLevels;
    @*/

  /*@ protected represents entries \such_that
    @   entries.equals(makeQueueEntrySet(theEntries, 0));
    @*/

  /*@ protected invariant
    @   orderedLegalLevels()
    @   && noEmptyLevels()
    @   && uniqueIDs()
    @   && timeStampsInOrderEachLevel()
    @   && singlyLinkedLevelsList()
    @   && (\forall PriorityListNode pln;
    @        isPriorityNode(theEntries, pln, new JMLObjectSet());
    @        pln.entries != null && singlyLinkedEntries(pln, pln.entries)
    @        && pln.priorityLevel == pln.entries.priority);
    @*/

  //**************************************************************
  // model methods used in "concrete" invariants, represents and
  // asserts clauses

  /*@  protected normal_behavior
    @     requires pln == null;
    @     ensures \result.equals(new JMLValueSet());
    @   also
    @     requires pln != null;
    @     ensures
    @         \result.equals(
    @           oneLevelEntries(pln.entries, pln.numberOfEntries).
    @             union(makeQueueEntrySet(pln.next, 
    @                                     (int)(numLevelsVisited + 1))));
    @ protected model pure 
    @     JMLValueSet makeQueueEntrySet(PriorityListNode pln, 
    @                                   int numLevelsVisited);
    @*/

  /*@ 
    @   protected normal_behavior
    @     requires en == null;
    @     ensures \result.equals(new JMLValueSet());
    @   also
    @     requires en != null;
    @     ensures
    @         \result.equals(oneLevelEntries(en.next, (int)(numEntriesToGo - 1))
    @           .insert(new QueueEntry(en.iD, en.priority, en.timeStamp)));
    @ protected model pure static JMLValueSet oneLevelEntries(
    @                      EntryNode en, int numEntriesToGo);
    @*/

  /*@ 
    @   protected normal_behavior
    @     requires theEntries == null;
    @     ensures \result == true;
    @   also
    @     requires theEntries != null;
    @     ensures
    @           (\forall PriorityListNode pln;
    @             isPriorityNode(theEntries, pln, new JMLObjectSet());
    @             pln.priorityLevel >= 0
    @             && (pln.next != null ==> 
    @               pln.priorityLevel > pln.next.priorityLevel) );
    @ protected model pure boolean orderedLegalLevels();
    @*/

  /*@ 
    @   protected normal_behavior
    @     requires theEntries == null;
    @     ensures \result == true;
    @   also
    @     requires theEntries != null;
    @     ensures 
    @           (\forall PriorityListNode pln;
    @             isPriorityNode(theEntries, pln, new JMLObjectSet());
    @             pln.entries != null );
    @ protected model pure boolean noEmptyLevels();
    @*/

  /*@ 
    @   protected normal_behavior
    @     ensures \result ==
    @       (\forall EntryNode en1; isEntry(en1);
    @         (\forall EntryNode en2; 
    @            isEntry(en2) && en2 != en1;
    @            !(en2.iD.equals(en1.iD)) ));
    @ protected model pure boolean uniqueIDs();
    @*/

  /*@ 
    @   protected normal_behavior
    @     ensures \result ==
    @       (\forall PriorityListNode pln;
    @             isPriorityNode(theEntries, pln, new JMLObjectSet());
    @         (\forall EntryNode en1;
    @              isEntryNode(pln, pln.entries, en1, new JMLObjectSet())
    @                && en1.next != null;
    @              en1.timeStamp < en1.next.timeStamp));
    @ protected model pure boolean timeStampsInOrderEachLevel();
    @*/

  /*@ 
    @ protected normal_behavior
    @   ensures \result ==
    @     (\exists PriorityListNode pln;
    @         isPriorityNode(theEntries, pln, new JMLObjectSet());
    @         pln.next == null);
    @ protected model pure boolean singlyLinkedLevelsList();
    @*/

  /*@   protected normal_behavior
    @     ensures
    @       (hd == null ==> \result == false)
    @       && (hd != null ==>
    @         (hd == pln ==> \result == true)
    @         && (hd != pln ==> \result == 
    @           isPriorityNode(hd.next, pln, nodesTraversed.insert(hd))));
    @ protected model pure boolean isPriorityNode(PriorityListNode hd,
    @                                             PriorityListNode pln,
    @                                             JMLObjectSet nodesTraversed);
    @*/

  /*@ 
    @   protected normal_behavior
    @     ensures \result ==
    @       (\exists EntryNode en2;
    @          isEntryNode(pln, en, en2, new JMLObjectSet());
    @          en2.next == null);
    @ protected model pure static boolean singlyLinkedEntries(
    @    PriorityListNode pln, EntryNode en);
    @*/

  /*@ 
    @   protected normal_behavior
    @     ensures
    @       (hd == null ==> \result == false)
    @       && (hd != null ==>
    @         (hd == en ==> \result == true)
    @         && (hd != en ==> \result == 
    @           isEntryNode(pln, hd.next, en, nodesTraversed.insert(hd))));
    @ protected model pure static boolean isEntryNode(
    @     PriorityListNode pln, EntryNode hd,
    @     EntryNode en, JMLObjectSet nodesTraversed);
    @*/

  /*@ also
    @   protected normal_behavior
    @     requires theEntries != null;
    @     ensures
    @       (\exists EntryNode en1; isEntry(en1);
    @         (\forall EntryNode en2; 
    @              isEntry(en2) && en2 != en1;
    @              en1.timeStamp > en2.timeStamp && \result == en1.timeStamp));
    @ public model pure int maxTimeStamp();
    @*/

  /*@ 
    @   protected normal_behavior
    @     ensures \result ==
    @       (\exists PriorityListNode pln;
    @          isPriorityNode(theEntries, pln, new JMLObjectSet());
    @          isEntryNode(pln, pln.entries, en, new JMLObjectSet()));
    @ protected model pure boolean isEntry(EntryNode en);
    @*/

  //***************************************************************
  // public methods for class PriorityQueue

  public PriorityQueue()
  {
    theEntries = null;
    //@ assert theEntries == null && numberOfLevels == 0;
  }

  public void addEntry(String argID, int argPriorityLevel)
  {
    /*@ assert argID != null && !argID.equals(new String())
      @   && argPriorityLevel >= 0;
      @*/

    PriorityListNode matcher = theEntries;
    PriorityListNode predOfMatcher = null;

    while (matcher != null && matcher.priorityLevel > argPriorityLevel) {
      predOfMatcher = matcher;
      matcher = matcher.next;
    }
    if (matcher ==  null && predOfMatcher == null) {
      theEntries = new PriorityListNode();
      theEntries.next = null;
      theEntries.priorityLevel = argPriorityLevel;
      theEntries.entries = new EntryNode();
      theEntries.entries.iD = new String(argID);
      theEntries.entries.next = null;
      /*@ set theEntries.entries.priority = argPriorityLevel;
        @ set theEntries.entries.timeStamp = (int)(maxTimeStamp() + 1);
        @*/
    } else if (matcher == null && predOfMatcher != null) {
      predOfMatcher.next = new PriorityListNode();
      predOfMatcher.next.priorityLevel = argPriorityLevel;
      predOfMatcher.next.entries = new EntryNode();
      predOfMatcher.next.entries.iD = new String(argID);
      predOfMatcher.next.entries.next = null;
      /*@ set predOfMatcher.next.entries.priority = argPriorityLevel;
        @ set predOfMatcher.next.entries.timeStamp = (int)(maxTimeStamp() + 1);
        @*/
    } else if (   matcher != null 
               && predOfMatcher == null 
               && matcher.priorityLevel != argPriorityLevel) {
      PriorityListNode saveTheEntries = theEntries;
      theEntries = new PriorityListNode();
      theEntries.next = saveTheEntries;
      theEntries.priorityLevel = argPriorityLevel;
      theEntries.entries = new EntryNode();
      theEntries.entries.iD = new String(argID);
      theEntries.entries.next = null;
      /*@ set theEntries.entries.priority = argPriorityLevel;
	@ set theEntries.entries.timeStamp = (int)(maxTimeStamp() + 1);
	@*/
    } else {
      if (argPriorityLevel == matcher.priorityLevel) {
        EntryNode entryWalker = matcher.entries;
        while (entryWalker.next != null) {
          entryWalker = entryWalker.next;
        }
        entryWalker.next = new EntryNode();
        entryWalker.next.iD = new String(argID);
        entryWalker.next.next = null;
	/*@ set entryWalker.next.priority = argPriorityLevel;
	  @ set entryWalker.next.timeStamp = (int)(maxTimeStamp() + 1);
          @*/
      } else {
        PriorityListNode savePredsNext = predOfMatcher.next;
        predOfMatcher.next = new PriorityListNode();
        predOfMatcher.next.next = savePredsNext;
        predOfMatcher.next.priorityLevel = argPriorityLevel;
        predOfMatcher.next.entries = new EntryNode();
        predOfMatcher.next.entries.iD = new String(argID);
        predOfMatcher.next.entries.next = null;
	/*@ set predOfMatcher.next.entries.priority = argPriorityLevel;
	  @ set predOfMatcher.next.entries.timeStamp = (int)(maxTimeStamp()+1);
          @*/
      }
    }
    // 'What' is not always more succinct than 'how'!!!
    /*@ assert
      @   (\old(theEntries == null) ==> \fresh(theEntries)
      @     && theEntries.priorityLevel == argPriorityLevel
      @     && theEntries.next == null
      @     && theEntries.numberOfEntries == 1
      @     && \fresh(theEntries.entries) 
      @     && theEntries.entries.iD.equals(argID)
      @     && theEntries.entries.priority == argPriorityLevel
      @     && theEntries.entries.timeStamp == 1
      @     && theEntries.entries.next == null
      @     && numberOfLevels == 1)
      @   && (\old(theEntries != null) ==>
      @     ((\exists PriorityListNode pln;
      @       \old(isPriorityNode(theEntries, pln, new JMLObjectSet()) 
      @       && pln.priorityLevel == argPriorityLevel)) ==>
      @         ((\exists PriorityListNode pln;
      @           \old(isPriorityNode(theEntries, pln, new JMLObjectSet()) 
      @           && pln.priorityLevel == argPriorityLevel)
      @           && pln.numberOfEntries == \old(pln.numberOfEntries + 1)
      @           && (\exists EntryNode en;
      @             \old(isEntryNode(pln, pln.entries, en, new JMLObjectSet())
      @             && en.next == null) && \fresh(en.next) 
      @             && en.next.iD.equals(argID) && en.next.next == null 
      @             && en.next.priority == argPriorityLevel 
      @             && en.next.timeStamp == \old(maxTimeStamp() + 1) 
      @             && pln.numberOfEntries == \old(pln.numberOfEntries + 1))))
      @     && (!(\exists PriorityListNode pln;
      @       \old(isPriorityNode(theEntries, pln, new JMLObjectSet())
      @       && pln.priorityLevel == argPriorityLevel)) ==>
      @         (argPriorityLevel < \old(theEntries.priorityLevel) ==>
      @           \fresh(theEntries) && theEntries.next == \old(theEntries)
      @           && theEntries.priorityLevel == argPriorityLevel
      @           && theEntries.numberOfEntries == 1
      @           && \fresh(theEntries.entries)
      @           && theEntries.entries.iD.equals(argID)
      @           && theEntries.entries.next == null
      @           && theEntries.entries.priority == argPriorityLevel
      @           && theEntries.entries.timeStamp == \old(maxTimeStamp() + 1))
      @         && ((\exists PriorityListNode pln; 
      @            \old(isPriorityNode(theEntries, pln, new JMLObjectSet()) 
      @            && pln.next == null 
      @            && pln.priorityLevel > argPriorityLevel)) ==>
      @              (\exists PriorityListNode pln;
      @               \old(isPriorityNode(theEntries, pln, new JMLObjectSet()) 
      @                && pln.next == null 
      @                && pln.priorityLevel > argPriorityLevel)
      @                && \fresh(pln.next) 
      @                && pln.next.priorityLevel == argPriorityLevel
      @                && pln.next.next == null
      @                && pln.next.numberOfEntries == 1
      @                && \fresh(pln.entries)
      @                && pln.entries.iD.equals(argID)
      @                && pln.entries.next == null
      @                && pln.entries.priority == argPriorityLevel
      @                && pln.entries.timeStamp == \old(maxTimeStamp() + 1)))
      @         && ((\exists PriorityListNode pln1;
      @           \old(isPriorityNode(theEntries, pln1, new JMLObjectSet())
      @             && pln1.priorityLevel > argPriorityLevel)
      @             && (\exists PriorityListNode pln2;
      @               \old(isPriorityNode(theEntries, pln2, new JMLObjectSet())
      @               && pln2.priorityLevel < argPriorityLevel))) ==>
      @                 (\exists PriorityListNode pln1;
      @                 \old(isPriorityNode(theEntries,pln1,new JMLObjectSet())
      @                     && pln1.priorityLevel > argPriorityLevel)
      @                     && (\exists PriorityListNode pln2;
      @                       \old(isPriorityNode(
      @                           theEntries, pln2, new JMLObjectSet())
      @                       && pln2.priorityLevel < argPriorityLevel)
      @                       && \fresh(pln1.next)
      @                       && pln1.next.next == \old(pln1.next)
      @                       && pln1.next.priorityLevel == argPriorityLevel
      @                       && pln1.next.numberOfEntries == 1
      @                       && \fresh(pln1.next.entries)
      @                       && pln1.next.entries.iD.equals(argID)
      @                       && pln1.next.entries.next == null
      @                       && pln1.next.entries.priority == 
      @                           argPriorityLevel
      @                       && pln1.next.entries.timeStamp ==
      @                           \old(maxTimeStamp() + 1)))))));
      @*/
  }
 
  public /*@ pure @*/ String next()
  {
    return theEntries.entries.iD;
  }
 
  public void remove(String argID)
  {
    PriorityListNode priorityListWalker = theEntries;
    PriorityListNode predPriorityListWalker = null;
    EntryNode entryWalker;
    EntryNode predEntryWalker;

    while (priorityListWalker != null) {
      entryWalker = priorityListWalker.entries;
      predEntryWalker = null;
      while (entryWalker != null) {
        if (entryWalker.iD.equals(argID)) {
          if (predEntryWalker != null) {
            predEntryWalker.next = entryWalker.next;
            return;
          } else if (entryWalker.next != null) {
            priorityListWalker.entries = entryWalker.next;
            return;
          } else if (predPriorityListWalker != null) {
            predPriorityListWalker.next = priorityListWalker.next;
            return;
          } else if (priorityListWalker.next != null) {
            theEntries = priorityListWalker.next;
            return;
          } else {
            theEntries = null;
            return;
          }
        } else {
          predEntryWalker = entryWalker;
          entryWalker = entryWalker.next;
        }
      }
      predPriorityListWalker = priorityListWalker;
      priorityListWalker = priorityListWalker.next;
    }
  }
 
  public /*@ pure @*/ boolean isUsedID(String argID)
  {
    PriorityListNode priorityListWalker = theEntries;
    EntryNode entryWalker;

    while (priorityListWalker != null) {
      entryWalker = priorityListWalker.entries;
      while (entryWalker != null) {
        if (entryWalker.iD.equals(argID)) {
          return true;
        } else {
          entryWalker = entryWalker.next;
        }
      }
      priorityListWalker = priorityListWalker.next;
    }
    //@ assert ! (\exists EntryNode en; isEntry(en); en.iD.equals(argID));
    return false;

  }

  public /*@ pure @*/ boolean equals(/*@ nullable @*/ Object pq2) {
        if (this == pq2) {
            return true;
        }

        if (!(pq2 instanceof PriorityQueue)) {
            return false;
        }

        PriorityQueue q2 = (PriorityQueue) pq2;
        PriorityListNode myplWalker = theEntries;
        PriorityListNode plWalker2 = q2.theEntries;
        EntryNode entryWalker;
        EntryNode entryWalker2;
    
        while (myplWalker != null && plWalker2 != null) {
            if (myplWalker.priorityLevel != plWalker2.priorityLevel) {
                return false;
            }
            entryWalker = myplWalker.entries;
            entryWalker2 = plWalker2.entries;
            if (entryWalker.iD != entryWalker2.iD) {
                return false;
            }
            entryWalker = entryWalker.next;
            entryWalker2 = entryWalker2.next;
            while (entryWalker != null && entryWalker2 != null) {
                if (entryWalker.iD != entryWalker2.iD) {
                    return false;
                }
                entryWalker = entryWalker.next;
                entryWalker2 = entryWalker2.next;
            }
            if (entryWalker != null || entryWalker2 != null) {
                return false;
            }
            myplWalker = myplWalker.next;
            plWalker2 = plWalker2.next;
        }
        if (myplWalker != null || plWalker2 != null) {
            return false;
        }
        return true;
  }

  public /*@ pure @*/ String toString()
  {
    String retValue = new String();
    PriorityListNode priorityListWalker = theEntries;
    EntryNode entryWalker;
    
    while (priorityListWalker != null) {
      retValue = retValue + "Entries at level " 
	  + priorityListWalker.priorityLevel
          + ": ";
      entryWalker = priorityListWalker.entries;
      retValue = retValue +  entryWalker.iD;
      entryWalker = entryWalker.next;
      while (entryWalker != null) {
        retValue = retValue + ", " + entryWalker.iD;
        entryWalker = entryWalker.next;
      }
      retValue = retValue + "\n";
      priorityListWalker = priorityListWalker.next;
    }
    /*@ assert (* retValue is a String representation of this
      @   with one line for each priority level containing all entries
      @   at that level. *);
      @*/
    return retValue;
  }

}  // end of class PriorityQueue
