/* $Id: JspComparator.java,v 1.15 2005/04/26 02:40:29 kui_dai Exp $
 *
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlspec;
import org.multijava.mjc.*;
import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.Utils;
import java.util.ArrayList;
import java.util.Iterator;
import java.lang.reflect.*;
 
/** This class compares two compilation units to be sure that the specifications match appropriately. */

public class JspComparator {

    final String eol = System.getProperty("line.separator");
    
    boolean printPath = false;
    boolean showMissing = true;
    boolean showInherited = true;
    boolean synch = false;
    int privacyLevel = 3;
    boolean verbose = false;
    boolean veryQuiet = false;
    final ModifierUtility modutil = new CModifier();

    public boolean checkPrivacy(long mods) {
	int lev = 2;
	if (Utils.hasFlag(mods,Constants.ACC_PUBLIC)) lev = 0;
	else if (Utils.hasFlag(mods,Constants.ACC_PROTECTED)) lev = 1;
	else if (Utils.hasFlag(mods,Constants.ACC_PRIVATE)) lev = 3;
	else  lev = 2;
	return lev <= privacyLevel;
    }
    
    public JspComparator(JspOptions options) {
	privacyLevel = options.specLevel();
	verbose = options.verbose();
	veryQuiet = options.Quiet();
	showMissing = ! options.suppressMissing();
	showInherited = ! options.suppressInherited();
	printPath = options.path();
	synch = options.synch();
    }

    public String compare(JCompilationUnitType cu, JCompilationUnitType cujava) {

	StringBuffer sb = new StringBuffer();
	String filename = cu.file().getName();
	String jfilename = cujava.file().getName();

	// Compare package
	
	if (! cu.packageNameAsString().equals(cujava.packageNameAsString()) ) {
	    sb.append(cu.packageName().getTokenReference() + ": " + "Package names differ: \"" + cu.packageNameAsString() + "\" vs. \""
		+cujava.packageNameAsString() + "\"" + eol);
	}
	
	// Compare imports? - they may be different
	    
	// Compare top-level declarations
	JTypeDeclarationType[] types = cu.typeDeclarations();
	JTypeDeclarationType[] jtypes = (JTypeDeclarationType[]) cujava.typeDeclarations().clone();
	sb.append( compareTypes(types,jtypes,filename,jfilename) );
	
	// compare GFs
	ArrayList tlmethods = cu.tlMethods();
	ArrayList jtlmethods = (ArrayList) cujava.tlMethods().clone();
	sb.append( compareGFs(tlmethods,jtlmethods,filename,jfilename) );

	if (printPath) {
	    filename = cu.file().getAbsolutePath();
	    jfilename = cujava.file().getAbsolutePath();
	}
	
	return  ((!veryQuiet || sb.length() > 0) ? "Comparing " + filename + " and " + jfilename + eol : "" )
		    + (sb.toString());
	
    }

    public String compareToClass(JCompilationUnitType cu) {

	StringBuffer sb = new StringBuffer();
	String filename = cu.file().getName();
	String longfilename = cu.file().getAbsolutePath();
	JTypeDeclarationType[] types = cu.typeDeclarations();
	for (int i=0; i<types.length; ++i) {
	    JTypeDeclarationType type = types[i];

	    String qualifiedName = cu.packageNameAsString().replace('/','.') + type.ident();
	    Class cc = null;
	    try {
	    	cc = Class.forName(qualifiedName);
	    } catch (ClassNotFoundException e) {}
	    
	    String jfilename = type.ident() + ".class";
	    
	    String result;
	    if (cc == null) {
		result = (type.getTokenReference() + ": No class file for " + qualifiedName + eol);
	    } else {
		result = compareType(type,cc,filename,jfilename);
	    }
	    sb.append(  ((!veryQuiet || result.length() > 0) ? "Comparing " + type.ident() + " in " + 
		    (printPath ? longfilename : filename) + " and " + jfilename + eol : "" ) + result);
	}
	
	// compare GFs
	ArrayList tlmethods = cu.tlMethods();
	//sb.append( compareGFs(tlmethods,jtlmethods,filename,jfilename) );

	return sb.toString();
	
    }


    public String compareGFs(ArrayList tlmethods, ArrayList jtlmethods, String filename, String jfilename) {
	    // FIXME -- check and compare privacy ???

	StringBuffer sb = new StringBuffer();
	
	    java.util.Iterator i = tlmethods.iterator();
	    while (i.hasNext()) {
		MJTopLevelMethodDeclaration jm = (MJTopLevelMethodDeclaration)(i.next());
		String ident = jm.ident();
		java.util.Iterator j = jtlmethods.iterator();
		boolean found = false;
		while (j.hasNext()) {
		    MJTopLevelMethodDeclaration jjm = (MJTopLevelMethodDeclaration)(j.next());
		    // FIXME - have to compare entire signature
		    if (ident.equals(jjm.ident())) {
			// COMPARE!!!
			j.remove();
			found = true;
			break;
		    }
		}
		if (!found) {
		    sb.append(jm.getTokenReference() + ": " + "GF " + ident + " is in " + filename + ", but is not in " + jfilename + eol);
		}
			
	    }
	    java.util.Iterator j = jtlmethods.iterator();
	    while (j.hasNext()) {
		MJTopLevelMethodDeclaration jjm = (MJTopLevelMethodDeclaration)(j.next());
		if (!checkPrivacy(jjm.modifiers())) continue;
		if (showMissing) sb.append(jjm.getTokenReference() + ": " + "GF " + jjm.ident() + " is not in " + filename + ", but is in " + jfilename + eol);
	    }
	    
	return sb.toString();
    }
    
    public String compareTypes(JTypeDeclarationType[] types, JTypeDeclarationType[] jtypes, String filename, String jfilename) {
	StringBuffer sb = new StringBuffer();
	
	for (int i = 0; i < types.length; ++i ) {
	    String ident = types[i].ident();
	    boolean found = false;
	    for (int j=0; j< jtypes.length; ++j ) {
		if (jtypes[j] == null) continue;
		if (jtypes[j].ident().equals(ident)) {
		    sb.append( compareType(types[i],jtypes[j],filename,jfilename) );
		    jtypes[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found) {
	    	sb.append(types[i].getTokenReference() + ": " + "Type " + ident + " is in " + filename + ", but is not in " + jfilename + eol);
	    }
	}
	for (int j=0; j< jtypes.length; ++j ) {
	    if (jtypes[j] == null) continue;
	    if (!checkPrivacy(jtypes[j].modifiers())) continue;
	    if (showMissing) sb.append(jtypes[j].getTokenReference() + ": " + "Type " + jtypes[j].ident() + " is not in " + filename + ", but is in " + jfilename + eol);
	}
	
	return sb.toString();
    }

    
    public String compareTypes(JTypeDeclarationType[] types, Class[] jtypes, String filename, String jfilename) {
	StringBuffer sb = new StringBuffer();
	
	for (int i = 0; i < types.length; ++i ) {
	    String ident = types[i].ident();
	    boolean found = false;
	    for (int j=0; j< jtypes.length; ++j ) {
		if (jtypes[j] == null) continue;
		if (typeNamesMatch(ident,toString(jtypes[j]))) {
		    sb.append( compareType(types[i],jtypes[j],filename,toString(jtypes[j])+".class") );
		    jtypes[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found) {
	    	sb.append(types[i].getTokenReference() + ": " + "Type " + ident + " is in " + filename + ", but is not in " + jfilename + eol);
	    }
	}
	for (int j=0; j< jtypes.length; ++j ) {
	    if (jtypes[j] == null) continue;
	    if (toString(jtypes[j]).indexOf('$') != -1 ) continue; // Synthetic - e.g. anonymous class
	    if (!checkPrivacy(jtypes[j].getModifiers())) continue;
	    if (showMissing) sb.append(filename + ": " + "Type " + toString(jtypes[j]) + " is not in " + filename + ", but is in " + jfilename + eol);
	}
	
	return sb.toString();
    }
    
    public boolean isInterface(JTypeDeclarationType t) {
	return t instanceof JInterfaceDeclarationType;
    }
    
    public String compareType(JTypeDeclarationType ty, JTypeDeclarationType jty, String filename, String jfilename) {


	StringBuffer sb = new StringBuffer();
	String ident =  ty.ident();
	
	// Same kind of type (class or interface)?
	if (isInterface(ty) != isInterface(jty)) {
	    sb.append(ty.getTokenReference() + ": " + "Type " + ident + " is " 
		+ (isInterface(ty)?"an interface" : "a class")
		+ " in " + filename + ", but is " 
		+ (isInterface(jty)?"an interface" : "a class")
		+ " in " + jfilename + eol);
	}
	
	if (!isInterface(ty)) {
	    // Compare the super class
	    
	    String s1 = ((JClassDeclarationType)ty).superName();
	    String s2 = ((JClassDeclarationType)jty).superName();
	    if (s1 == null && s2 != null) {
		sb.append(ty.getTokenReference() + ": " + "Class " + ident + " extends " + s2 + " in " + jfilename
			+ ", but does not extend anything in " + filename + eol);
	    } else if (s2 == null && s1 != null) {
		sb.append(ty.getTokenReference() + ": " + "Class " + ident + " extends " + s1 + " in " + filename
			+ ", but does not extend anything in " + jfilename + eol);
	    } else if (s1 != s2 && ! typeNamesMatch(s1,s2)) {
		sb.append(ty.getTokenReference() + ": " + "Class " + ident + " extends " + s1 + " in " + filename
			+ ", but extends " + s2 + " in " + jfilename + eol);
	    }
	    
	}

	// Compare deprecation
	if (ty.isDeprecated() != jty.isDeprecated()) {
	    sb.append(ty.getTokenReference() + ": " + "Type " + ident + (ty.isDeprecated()?" is " : " is not ")
		+ "deprecated in " + filename + ", but" + (jty.isDeprecated()?" is " : " is not ")
		+ "in " + jfilename + eol);
	}

	// Compare modifiers
	long modsToIgnore = Constants.ACC_STRICT | Constants.ACC_SUPER;
	if ((ty.modifiers()&~modsToIgnore) != (jty.modifiers()&~modsToIgnore)) {
	    sb.append(ty.getTokenReference() + ": " + "Modifiers for type " + ident + " are different: " + modutil.asString(ty.modifiers()) + "vs. " + modutil.asString(jty.modifiers()) + eol);
	}
	
	// Compare interfaces (implemented interfaces for a class, extended interfaces for an interface)
	
	{
	    CClassType[] ints = ty.interfaces();
	    CClassType[] jints = (CClassType[]) jty.interfaces().clone();
	    for (int i=0; i<ints.length; ++i) {
		CClassType iint = ints[i];
		boolean found = false;
		for (int j=0; j<jints.length; ++j) {
		    CClassType jint = jints[j];
		    if (jint == null) continue;
		    if (iint.toString().equals(jint.toString())) {
			jints[j] = null;
			found = true;
			break;
		    }
		}
		if (!found) {
		    sb.append(ty.getTokenReference() + ": " + iint.toString() + " is an interface for " + ident
			+ " in " + filename + ", but not in " + jfilename + eol);
		}
	    }
	    for (int j = 0; j<jints.length; ++j) {
		CClassType jint = jints[j];
		if (jint != null) sb.append(ty.getTokenReference() + ": " + jint.toString() + " is an interface for " + ident
			    + " in " + jfilename + ", but not in " + filename + eol);
	    }
	}

	modsToIgnore = Constants.ACC_NATIVE | Constants.ACC_STRICT;
	if (isInterface(ty)) modsToIgnore |= Constants.ACC_ABSTRACT;
	if (isInterface(ty) && Utils.hasFlag(ty.modifiers(),Constants.ACC_PUBLIC)) modsToIgnore |= Constants.ACC_PUBLIC;
	sb.append(  compareMethods(
		    (JMethodDeclarationType[]) ty.methods().toArray(new JMethodDeclarationType[ty.methods().size()]),
		    (JMethodDeclarationType[]) jty.methods().toArray(new JMethodDeclarationType[jty.methods().size()]),
		    filename,jfilename,modsToIgnore)  );
	
	modsToIgnore = 0;
	if (isInterface(ty) && Utils.hasFlag(ty.modifiers(),Constants.ACC_PUBLIC)) modsToIgnore |= Constants.ACC_PUBLIC;
	sb.append(  compareFields(ty.fields(),
		    (JFieldDeclarationType[]) jty.fields().clone(),filename,jfilename,modsToIgnore)  );
	
	sb.append(  compareTypes((JTypeDeclarationType[])ty.inners().toArray(new JTypeDeclarationType[0]),
			(JTypeDeclarationType[])jty.inners().toArray(new JTypeDeclarationType[0]),
			filename,jfilename)  );

	return sb.toString();

    }
    
    public boolean typeNamesMatch(String s, String ss) {
	// The second string should be the one from a class file, if either is
	if (s.equals(ss)) return true;
	ss = ss.replace('$','.');
	return s.equals(ss) || s.endsWith(ss) || ss.endsWith(s);
    }
    
    public String compareType(JTypeDeclarationType ty, Class cc, String filename, String jfilename) {


	StringBuffer sb = new StringBuffer();
	String ident =  ty.ident();
	
	// Same kind of type (class or interface)?
	if (isInterface(ty) != cc.isInterface()) {
	    sb.append(ty.getTokenReference() + ": " + "Type " + ident + " is " 
		+ (isInterface(ty)?"an interface" : "a class")
		+ " in " + filename + ", but is " 
		+ (cc.isInterface()?"an interface" : "a class")
		+ " in " + jfilename + eol);
	}
	
	if (!isInterface(ty)) {
	    // Compare the super class

	    String s1 = ((JClassDeclarationType)ty).superName();
	    if (cc.getSuperclass() == null) {
		if (s1 != null) {
		    sb.append(ty.getTokenReference() + ": " + "Class " + ident + " extends " + s1 + " in " + filename
			+ ", but does not extend anything in " + jfilename + eol);		
		}		
	    } else {
		String s2 = cc.getSuperclass().getName();
		if (s1 == null) s1 = "java.lang.Object";
		else s1 = s1.replace('/','.');
		if (s1 == null && s2 != null) {
		    sb.append(ty.getTokenReference() + ": " + "Class " + ident + " extends " + s2 + " in " + jfilename
			    + ", but does not extend anything in " + filename + eol);
		} else if (s2 == null && s1 != null) {
		    sb.append(ty.getTokenReference() + ": " + "Class " + ident + " extends " + s1 + " in " + filename
			    + ", but does not extend anything in " + jfilename + eol);
		} else if (s1 != s2 && ! typeNamesMatch(s1,s2)) {
		    sb.append(ty.getTokenReference() + ": " + "Class " + ident + " extends " + s1 + " in " + filename
			    + ", but extends " + s2 + " in " + jfilename + eol);
		}
	    }
	    
	}

	// Compare deprecation
	/* FIXME
	if (ty.isDeprecated() != cc.isDeprecated()) {
	    sb.append(ty.getTokenReference() + ": " + "Type " + ident + (ty.isDeprecated()?" is " : " is not ")
		+ "deprecated in " + filename + ", but" + (cc.isDeprecated()?" is " : " is not ")
		+ "in " + jfilename + eol);
	}
	*/

	// Compare modifiers
	long modsToIgnore = Constants.ACC_STRICT | Constants.ACC_SUPER;
	if ((ty.modifiers()&~modsToIgnore) != (cc.getModifiers()&~modsToIgnore)) {
	    sb.append(ty.getTokenReference() + ": " + "Modifiers for type " + ident + " are different: " + modutil.asString(ty.modifiers()) + "vs. " + modutil.asString(cc.getModifiers()) + eol);
	}
	
	// Compare interfaces (implemented interfaces for a class, extended interfaces for an interface)
	
	{
	    CClassType[] ints = ty.interfaces();
	    Class[] jints = (Class[]) cc.getInterfaces().clone();
	    for (int i=0; i<ints.length; ++i) {
		CClassType iint = ints[i];
		boolean found = false;
		for (int j=0; j<jints.length; ++j) {
		    Class jint = jints[j];
		    if (jint == null) continue;
		    if (typeNamesMatch(iint.toString(),jint.getName())) {
			jints[j] = null;
			found = true;
			break;
		    }
		}
		if (!found) {
		    sb.append(ty.getTokenReference() + ": " + iint.toString() + " is an interface for " + ident
			+ " in " + filename + ", but not in " + jfilename + eol);
		}
	    }
	    for (int j = 0; j<jints.length; ++j) {
		Class jint = jints[j];
		if (jint != null) sb.append(ty.getTokenReference() + ": " + jint.getName() + " is an interface for " + ident
			    + " in " + jfilename + ", but not in " + filename + eol);
	    }
	}


	modsToIgnore = Constants.ACC_NATIVE | Constants.ACC_STRICT;
	if (isInterface(ty)) modsToIgnore |= Constants.ACC_ABSTRACT;
	if (isInterface(ty) && Utils.hasFlag(ty.modifiers(),Constants.ACC_PUBLIC)) modsToIgnore |= Constants.ACC_PUBLIC;
	Member[] members = new Member[cc.getDeclaredMethods().length + cc.getDeclaredConstructors().length];
	System.arraycopy(cc.getDeclaredMethods(),0,members,0,cc.getDeclaredMethods().length);
	System.arraycopy(cc.getDeclaredConstructors(),0,members,cc.getDeclaredMethods().length,cc.getDeclaredConstructors().length);
	sb.append(  compareMethods(
		    (JMethodDeclarationType[]) ty.methods().toArray(new JMethodDeclarationType[ty.methods().size()]),
		    members,
		    filename,jfilename,modsToIgnore,cc)  );
	
	modsToIgnore = 0;
	if (isInterface(ty) && Utils.hasFlag(ty.modifiers(),Constants.ACC_PUBLIC)) modsToIgnore |= Constants.ACC_PUBLIC;
	sb.append(  compareFields(ty.fields(),
		    (Field[]) cc.getDeclaredFields().clone(),filename,jfilename,modsToIgnore)  );
	
	sb.append(  compareTypes((JTypeDeclarationType[])ty.inners().toArray(new JTypeDeclarationType[0]),
			(Class[])cc.getDeclaredClasses().clone(),
			filename,jfilename)  );

	return sb.toString();

    }
    
    public String compareFields(JFieldDeclarationType[] fields, JFieldDeclarationType[] jfields,
		    String filename, String jfilename, long modsToIgnore) {
		    
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < fields.length; ++i ) {
	    String ident = fields[i].ident();
	    boolean found = false;
	    for (int j=0; j< jfields.length; ++j ) {
		if (jfields[j] == null) continue;
		if (jfields[j].ident().equals(ident)) {
		    sb.append(compareField(fields[i],jfields[j],filename,jfilename,modsToIgnore));
		    jfields[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found) {
	    	sb.append(fields[i].getTokenReference() + ": " + "Field " + ident + " is in " + filename + ", but is not in " + jfilename + eol);
	    }
	}
	for (int j=0; j< jfields.length; ++j ) {
	    if (jfields[j] == null) continue;
	    if (!checkPrivacy(jfields[j].modifiers())) continue;
	    if (showMissing) sb.append(jfields[j].getTokenReference() + ": "  + "Field " + jfields[j].ident() + " is not in " + filename + ", but is in " + jfilename + eol);
	}
	return sb.toString();
    }
    
    public String compareFields(JFieldDeclarationType[] fields, Field[] jfields,
		    String filename, String jfilename, long modsToIgnore) {
		    
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < fields.length; ++i ) {
	    String ident = fields[i].ident();
	    boolean found = false;
	    for (int j=0; j< jfields.length; ++j ) {
		if (jfields[j] == null) continue;
		if (jfields[j].getName().equals(ident)) {
		    sb.append(compareField(fields[i],jfields[j],filename,jfilename,modsToIgnore));
		    jfields[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found) {
	    	sb.append(fields[i].getTokenReference() + ": " + "Field " + ident + " is in " + filename + ", but is not in " + jfilename + eol);
	    }
	}
	for (int j=0; j< jfields.length; ++j ) {
	    if (jfields[j] == null) continue;
	    if (isSynthetic(jfields[j])) continue;
	    if (!checkPrivacy(jfields[j].getModifiers())) continue;
	    if (showMissing) sb.append(filename + ": "  + "Field " + jfields[j].getName() + " is not in " + filename + ", but is in " + jfilename + eol);
	}
	return sb.toString();
    }
    
    
    public String compareField(JFieldDeclarationType field, JFieldDeclarationType jfield,
		    String filename, String jfilename, long modsToIgnore) {

	StringBuffer sb = new StringBuffer();
	String ident = field.ident();

	// Compare deprecated
	if (field.isDeprecated() != jfield.isDeprecated()) {
	    sb.append(field.getTokenReference() + ": " + "Field " + ident + (field.isDeprecated()?" is " : " is not ")
		+ "deprecated in " + filename + ", but" + (jfield.isDeprecated()?" is " : " is not ")
		+ "in " + jfilename + eol);
	}

	// Compare modifiers
	if ((field.modifiers()&~modsToIgnore) != (jfield.modifiers()&~modsToIgnore)) {
	    sb.append(field.getTokenReference() + ": " + "Modifiers for field " + ident + " are different: " + modutil.asString(field.modifiers()) + "vs. " + modutil.asString(jfield.modifiers()) + eol);
	}

	// Compare type
	String t1 = field.variable().getType().toString();
	String t2 = jfield.variable().getType().toString();
	
	if (!t1.equals(t2)) {
	    sb.append(field.getTokenReference() + ": " + "Type of field " + ident + " is " + t1 + " in " + filename
				+ ", but " + t2 + " in " + jfilename + eol);
	}

	return sb.toString();
    }
    
     
    public String compareField(JFieldDeclarationType field, Field jfield,
		    String filename, String jfilename, long modsToIgnore) {

	StringBuffer sb = new StringBuffer();
	String ident = field.ident();

	// Compare deprecated - FIXME
/*	if (field.isDeprecated() != jfield.isDeprecated()) {
	    sb.append(field.getTokenReference() + ": " + "Field " + ident + (field.isDeprecated()?" is " : " is not ")
		+ "deprecated in " + filename + ", but" + (jfield.isDeprecated()?" is " : " is not ")
		+ "in " + jfilename + eol);
	}
*/
	// Compare modifiers
	if ((field.modifiers()&~modsToIgnore) != (jfield.getModifiers()&~modsToIgnore)) {
	    sb.append(field.getTokenReference() + ": " + "Modifiers for field " + ident + " are different: " + modutil.asString(field.modifiers()) + "vs. " + modutil.asString(jfield.getModifiers()) + eol);
	}

	// Compare type
	String t1 = field.variable().getType().toString();
	String t2 = toString(jfield.getType());
	
	if (!typeNamesMatch(t1,t2)) {
	    sb.append(field.getTokenReference() + ": " + "Type of field " + ident + " is " + t1 + " in " + filename
				+ ", but " + t2 + " in " + jfilename + eol);
	}

	return sb.toString();
    }
    
   
    public String compareMethods(JMethodDeclarationType[] methods, JMethodDeclarationType[] jmethods,
		    String filename, String jfilename, long modsToignore) {
		    
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < methods.length; ++i ) {
	    String ident = signature(methods[i]);
	    boolean found = false;
	    for (int j=0; j< jmethods.length; ++j ) {
		if (jmethods[j] == null) continue;
		if (matches(methods[i],jmethods[j])) {
		    sb.append(compareMethod(methods[i],jmethods[j],filename,jfilename,modsToignore));
		    jmethods[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found) { 
		boolean isConstructor = methods[i] instanceof JConstructorDeclarationType;
		boolean isDefaultConstructor = checkDefaultConstructor(methods[i],methods,jmethods);
		if (!(isDefaultConstructor && jfilename.endsWith(".java"))) {
		    sb.append(methods[i].getTokenReference() + ": " + (isConstructor?"Constructor ":"Method ") 
			+ ident + " is in " + filename + ", but is not in " + jfilename + eol);
		}
	    }
	}
	for (int j=0; j< jmethods.length; ++j ) {
	    if (jmethods[j] == null) continue;
	    if (!checkPrivacy(jmethods[j].modifiers())) continue;
	    boolean isConstructor = jmethods[j] instanceof JConstructorDeclarationType;
	    if (showMissing) sb.append(jmethods[j].getTokenReference() + ": "  + (isConstructor?"Constructor ":"Method ") 
			+ signature(jmethods[j]) + " is not in " + filename + ", but is in " + jfilename + eol);
	}
	return sb.toString();
    }

    public String compareMethods(JMethodDeclarationType[] methods, Member[] jmethods,
		    String filename, String jfilename, long modsToignore, Class cc) {
		    
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < methods.length; ++i ) {
	    String ident = signature(methods[i]);
	    boolean found = false;
	    for (int j=0; j< jmethods.length; ++j ) {
		if (jmethods[j] == null) continue;
		if (matches(methods[i],jmethods[j])) {
		    sb.append(compareMethod(methods[i],jmethods[j],filename,jfilename,modsToignore));
		    jmethods[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found) { 
		boolean isConstructor = methods[i] instanceof JConstructorDeclarationType;
		boolean print = isConstructor || ! checkInheritance(methods[i],cc);
		if (print) sb.append(methods[i].getTokenReference() + ": " + (isConstructor?"Constructor ":"Method ") 
			+ ident + " is in " + filename + ", but is not in " + jfilename + eol);
		else if (showInherited) {
		    sb.append(methods[i].getTokenReference() + ": " + (isConstructor?"Constructor ":"Method ") 
			+ ident + " is in " + filename + ", but is inherited by but not declared in " + jfilename 
			+ " (use -i to suppress this difference)" + eol);
		}
	    }
	}
	for (int j=0; j< jmethods.length; ++j ) {
	    if (jmethods[j] == null) continue;
	    if (isSynthetic(jmethods[j])) continue;
	    //sb.append("MOD " + jmethods[j].getModifiers() + " " + Constants.ACC_PRIVATE + " " + checkPrivacy(jmethods[j].getModifiers()) + eol);
	    if (!checkPrivacy(jmethods[j].getModifiers())) continue;
	    boolean isConstructor = jmethods[j] instanceof Constructor;
	    if (showMissing) sb.append(filename + ": "  + (isConstructor?"Constructor ":"Method ") 
			+ signature(jmethods[j]) + " is not in " + filename + ", but is in " + jfilename + eol);
	}
	return sb.toString();
    }

    public boolean checkInheritance(JMethodDeclarationType m, Class cc) {
	Method[] meths = cc.getMethods(); // includes inherited methods
	for (int j=0; j<meths.length; ++j) {
	    Method mm = meths[j];
	    if (matches(m,mm)) return true;
	}
	if (cc.isInterface() && !showInherited) {
	    Class obj = Object.class;
	    meths = obj.getMethods();
	    for (int j=0; j<meths.length; ++j) {
	        if (matches(m,meths[j])) return true;
	    }
	}
	return false;
    }
    
    // FIXME - is there some better way?
    public boolean isSynthetic(Member m) {
	return (m.getName().indexOf('$') != -1);
    }

    /** Returns true if method is the one constructor in methods, has no arguments, and jmethods contains no constructors. */
    public boolean checkDefaultConstructor(JMethodDeclarationType method,JMethodDeclarationType[] methods,JMethodDeclarationType[] jmethods) {
	if ( !(method instanceof JConstructorDeclarationType)) return false;
	if (method.parameters().length != 0) return false;
	for (int j = 0; j < jmethods.length; ++j) {
	    if (jmethods[j] instanceof JConstructorDeclarationType) return false;
	}
	for (int i = 0; i < methods.length; ++i) {
	    if (method == methods[i]) continue;
	    if (methods[i] instanceof JConstructorDeclarationType) return false;
	}
	return true;
    }
    
    public boolean matches(JMethodDeclarationType method, JMethodDeclarationType jmethod) {
	if (!method.ident().equals(jmethod.ident())) return false;
	if (method.parameters().length != jmethod.parameters().length) return false;
	JFormalParameter[] params = method.parameters();
	JFormalParameter[] jparams = jmethod.parameters();
	for (int i = 0; i< params.length; ++i ) {
	    if (! params[i].typeToString().equals(jparams[i].typeToString())) return false;
	}
	return true;
    }
    
    public boolean matches(JMethodDeclarationType method, Member jmethod) {
	if ((jmethod instanceof Method) && !method.ident().equals(jmethod.getName())) return false;
	if ((jmethod instanceof Constructor) && !(method instanceof JConstructorDeclarationType)) return false;
	Class[] jparams = (jmethod instanceof Method) ? ((Method)jmethod).getParameterTypes() : ((Constructor)jmethod).getParameterTypes();
	if (method.parameters().length != jparams.length) return false;
	JFormalParameter[] params = method.parameters();
	for (int i = 0; i< params.length; ++i ) {
	    if (! typeNamesMatch(params[i].typeToString(),toString(jparams[i]))) return false;
	}
	return true;
    }
    
    public String signature(JMethodDeclarationType m) {
	String s = m.ident() + "(";
	JFormalParameter[] params = m.parameters();
	for (int i=0; i<params.length; ++i) {
	    if (i != 0) s = s + ",";
	    s = s + params[i].typeToString() + params[i].paramToString();
	}
	s = s + ")";
	return s;
    }

    public String signature(Member m) {
	String s = m.getName();
	int p = s.lastIndexOf('.');
	s = s.substring(p+1) + "(";
	Class[] params = (m instanceof Method) ? ((Method)m).getParameterTypes() : ((Constructor)m).getParameterTypes();
	for (int i=0; i<params.length; ++i) {
	    if (i != 0) s = s + ",";
	    s = s + toString(params[i]);
	}
	s = s + ")";
	return s;
    }

    
    public String compareMethod(JMethodDeclarationType method, JMethodDeclarationType jmethod,
		    String filename, String jfilename, long modsToIgnore) {
	// Compare members - mods, ident, return type, arg types, arg idents, arg mods, throws, deprecated

	StringBuffer sb = new StringBuffer();
	String ident = signature(method);

	// Compare deprecated
	if (method.isDeprecated() != jmethod.isDeprecated()) {
	    sb.append(method.getTokenReference() + ": " + "Method " + ident + (method.isDeprecated()?" is " : " is not ")
		+ "deprecated in " + filename + ", but" + (jmethod.isDeprecated()?" is " : " is not ")
		+ "in " + jfilename + eol);
	}

	// Compare modifiers
	if (!synch) modsToIgnore |= Constants.ACC_SYNCHRONIZED;
	if ((method.modifiers()&~modsToIgnore) != (jmethod.modifiers()&~modsToIgnore)) {
	    sb.append(method.getTokenReference() + ": " + "Modifiers for method " + ident + " are different: " + modutil.asString(method.modifiers()) + "vs. " + modutil.asString(jmethod.modifiers()) + eol);
	} else if ((method.modifiers()&Constants.ACC_SYNCHRONIZED)!=0 && (jmethod.modifiers()&Constants.ACC_SYNCHRONIZED)==0) {
	    sb.append(method.getTokenReference() + ": " + "The synchronized modifier may not be in the specification if it is not in the implementation"+eol);
        }

	// Compare arg types - number and types of arguments are the same to get this far
	
	// Compare arg idents and mods
	JFormalParameter[] params = method.parameters();
	JFormalParameter[] jparams = (JFormalParameter[]) jmethod.parameters().clone();
	for (int i = 0; i< params.length; ++i ) {
	    if (!params[i].ident().equals(jparams[i].ident())) {
		sb.append(params[i].getTokenReference() + ": " + "Parameter " + (i+1) + " has identifier " + params[i].ident() + " in " + filename + 
		    ", but " + jparams[i].ident() + " in " + jfilename + eol);
	    }
	    long ignore = Constants.ACC_FINAL;
	    if ((params[i].modifiers()&~ignore) != (jparams[i].modifiers()&~ignore)) {
		sb.append(params[i].getTokenReference() + ": " + "Modifiers for parameter " + (i+1) + "(" + params[i].ident() + ") are different: \"" + modutil.asString(params[i].modifiers()) + "\" vs. \"" + modutil.asString(jparams[i].modifiers()) + "\"" + eol);
	    }
	}
	
	
	// Compare throws
	CClassType[] exc = method.getExceptions();
	CClassType[] jexc = (CClassType[]) jmethod.getExceptions().clone();
	sb.append(compareExceptions(exc,jexc,ident,filename,jfilename,method.getTokenReference().toString()));
	

	// Compare return type
	String t1 = method.returnType().toString();
	String t2 = jmethod.returnType().toString();
	
	if (!typeNamesMatch(t1,t2)) {
	    sb.append(method.getTokenReference() + ": " + "Return type of method " + ident + " is " + t1 + " in " + filename
				+ ", but " + t2 + " in " + jfilename + eol);
	}

	return sb.toString();
    }

    
    public String compareMethod(JMethodDeclarationType method, Member jmethod,
		    String filename, String jfilename, long modsToIgnore) {
	// Compare members - mods, ident, return type, arg types, arg idents, arg mods, throws, deprecated

	StringBuffer sb = new StringBuffer();
	String ident = signature(method);

	// Compare deprecated - FIXME
	/*if (method.isDeprecated() != jmethod.isDeprecated()) {
	    sb.append(method.getTokenReference() + ": " + "Method " + ident + (method.isDeprecated()?" is " : " is not ")
		+ "deprecated in " + filename + ", but" + (jmethod.isDeprecated()?" is " : " is not ")
		+ "in " + jfilename + eol);
	}*/

	// Compare modifiers
	if (!synch) modsToIgnore |= Constants.ACC_SYNCHRONIZED;
	if ((method.modifiers()&~modsToIgnore) != (jmethod.getModifiers()&~modsToIgnore)) {
	    sb.append(method.getTokenReference() + ": " + "Modifiers for method " + ident + " are different: " + modutil.asString(method.modifiers()) + "vs. " + modutil.asString(jmethod.getModifiers()) + eol);
	} else if ((method.modifiers()&Constants.ACC_SYNCHRONIZED)!=0 && (jmethod.getModifiers()&Constants.ACC_SYNCHRONIZED)==0) {
	    sb.append(method.getTokenReference() + ": " + "The synchronized modifier may not be in the specification if it is not in the implementation"+eol);
        }

	// Compare arg types - number and types of arguments are the same to get this far
	
	// Compare arg idents and mods - no idents in a class file
	
	// Compare throws
	CClassType[] exc = method.getExceptions();
	Class[] jexc = (Class[])( (jmethod instanceof Method) ? ((Method)jmethod).getExceptionTypes().clone() : ((Constructor)jmethod).getExceptionTypes().clone() );
	sb.append(compareExceptions(exc,jexc,ident,filename,jfilename,method.getTokenReference().toString()));
	

	// Compare return type
	if (jmethod instanceof Method) {
	
	String t1 = method.returnType().toString();
	String t2 = toString(((Method)jmethod).getReturnType());
	
	if (!typeNamesMatch(t1,t2)) {
	    sb.append(method.getTokenReference() + ": " + "Return type of method " + ident + " is " + t1 + " in " + filename
				+ ", but " + t2 + " in " + jfilename + eol);
	}
	}

	return sb.toString();
    }

    public String compareExceptions(CClassType[] exc, CClassType[] jexc,
		    String method, String filename, String jfilename, String token) {
		    
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < exc.length; ++i ) {
	    String ident = exc[i].toString();
	    boolean found = false;
	    for (int j=0; j< jexc.length; ++j ) {
		if (jexc[j] == null) continue;
		if (typeNamesMatch(ident,jexc[j].toString())) {
		    jexc[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found) {
	    	sb.append(token + ": " + "Method " + method + " has exception " 
			+ ident + " in " + filename + ", but not in " + jfilename + eol);
	    }
	}
	for (int j=0; j< jexc.length; ++j ) {
	    if (jexc[j] == null) continue;
	    if (showMissing) sb.append(token + ": " + "Method " + method + " does not have exception " 
			+ jexc[j].toString() + " in " + filename + ", but does in " + jfilename + eol);
	}
	return sb.toString();
    }
    
    public String compareExceptions(CClassType[] exc, Class[] jexc,
		    String method, String filename, String jfilename, String token) {
		    
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < exc.length; ++i ) {
	    String ident = exc[i].toString();
	    boolean found = false;
	    for (int j=0; j< jexc.length; ++j ) {
		if (jexc[j] == null) continue;
		if (typeNamesMatch(ident,toString(jexc[j]))) {
		    jexc[j] = null;
		    found = true;
		    break;
		}
	    }
	    if (!found && !isARuntimeException(ident)) {
	    	sb.append(token + ": " + "Method " + method + " has exception " 
			+ ident + " in " + filename + ", but not in " + jfilename + eol);
	    }
	}
	for (int j=0; j< jexc.length; ++j ) {
	    if (jexc[j] == null) continue;
	    if (showMissing) sb.append(token + ": " + "Method " + method + " does not have exception " 
			+ toString(jexc[j]) + " in " + filename + ", but does in " + jfilename + eol);
	}
	return sb.toString();
    }

    /** Returns true if the argument is the name of a class that inherits from RuntimeException. */
    // FIXME - this is not very robust - hard to do without resolving type names
    public boolean isARuntimeException(String ident) {
	Class cc = null;
	try {
	    cc = Class.forName(ident);
	    return RuntimeException.class.isAssignableFrom(cc);
	} catch (ClassNotFoundException e) {}
	try {
	    cc = Class.forName("java.lang." + ident);
	    return RuntimeException.class.isAssignableFrom(cc);
	} catch (ClassNotFoundException e) {}
	return false;
    }

    public String toString(Class c) {
	if (c.isArray()) return toString(c.getComponentType()) + "[]";
	return c.toString();
    }

}
