/* $Id: Main.java,v 1.27 2005/06/24 12:57:52 wdietl Exp $
 *
 * Copyright (C) 2002-2004 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* TODO:

Need to implement jmlspec for external methods, both generation and -diff

Would like package names to generate all specs for all class files in that
package

Would also like to be able to use wild-card package names

If a file is repeated on the command line double parsing occurs

Try jmlspec -diff . in testdiffs - would prefer that the tool continue to
give information even if some files have a bad parse.

Inner class constructors not handled correctly (e.g. testdiffs/Inner.spec)
(fix and verify for both java and class inputs)

Check that deprecation in a spec file is compared against the java/class file

Document

What sort of checking for RuntimeExceptions is needed? e.g may a spec file
declare such an exception even if the source does not?

Does the isSynthetic need improvement?
    
-diff needs to report on mismatches of inner classes when comparing against
a .class file (can it?)

Can we know to include a static_initializer when generating a spec file from
a java file?

Generation from binary class files:
	- Formatting should be the same as from a class file
*/

package org.jmlspecs.jmlspec;

import org.multijava.mjc.*;
import org.multijava.util.Utils;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.classfile.ClassPath;
import org.multijava.util.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;

/**
 * A class implementing the entry point of the jmlspec specification
 * skeleton generator/comparator.
 * Use <code>java org.jmlspecs.jmlspec.Main --help</code>
 * to get usage options.  
 *
 * @author David Cok
 * @version $Revision: 1.27 $
 */
public class Main extends org.multijava.mjc.Main implements Constants {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    public Main() {
        super();
        parseJavadoc = true; // for deprecated flag
        appName = "jmlspec"; // for error messages
    }
    
    // ----------------------------------------------------------------------
    // RUN FROM COMMAND LINE
    // ----------------------------------------------------------------------

    /**
     * An entry point when starting from the command line.
     *
     * @param args The command line arguments
     */
    public static void main(String[] args) {
        boolean success;
        success = compile(args);
        System.exit(success ? 0 : 1);
    }

    /**
     * Generates JML skeleton with the given argument,
     * <code>args</code>, and returns <code>true</code> if all the
     * skeletons are successfully generated.
     */
    public static boolean compile(String[] args) {
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    // Allows methods with no bodies
                    return true;
                }
            }
        );

        return new Main().run(args);
    }
        
    // ----------------------------------------------------------------------
    // RUN FROM GUI
    // ----------------------------------------------------------------------
    
    /**
     * Entry point for the GUI
     *
     * @param   args    the file, package, and directory names
     * @param   opt     the options for the tool
     * @param   os      the output stream for the compiler messages
     */
    public static boolean compile(String[] args, JspOptions opt, 
                                  OutputStream os) {
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    // Allows methods with no bodies
                    return true;
                }
            }
        );
        return new Main().run(args, opt, os);
    }
        
    // ----------------------------------------------------------------------
    // HELPER METHODS
    // ----------------------------------------------------------------------
    
    protected boolean runInitialization(String[] args) {
        if (!super.runInitialization(args)) return false;
	setTemplates((JspOptions)options);
	setAllowUniverses(options.universesx()); // AHS
        String suffix = ((JspOptions)options).suffix();
        for (int i=0; i<validSuffixes.length; ++i) {
            if (validSuffixes[i].equals(suffix)) return true;
        }
        reportTrouble(new CWarning(null,JspMessages.NON_STANDARD_SUFFIX, suffix));
        return true;
    }

    ArrayList binaries = new ArrayList();

    public void handlePackages(ArrayList packages, ArrayList dirs, ArrayList infiles) {
        Iterator i = packages.iterator();
        while (i.hasNext()) {
            String s = (String)(i.next());
	    String nm = s.replace('.','/');
	    File loc = CTopLevel.findSourceFile(nm);
	    if (loc != null) {
		infiles.add( loc.getAbsolutePath() );
	    } else {
		binaries.add(nm);
	    }
        }
    }

    public String checkPackageName(String packageName, boolean quiet) {
        String location = ClassPath.findPackageSource(packageName);
        if (location == null) location = ClassPath.findPackage(packageName);
        if (location == null) {
	    if (CTopLevel.loadClass(packageName.replace('.','/')) == CClass.CLS_UNDEFINED) {
		if (!quiet) { 
		    reportTrouble(new FormattedException("Error: " + appName + 
			   " could not find file or package or class " + packageName));
		}
		return null;
	    }
	    return null;
        }
 
        if (! ((new File(location)).isDirectory())) {
            reportTrouble(new FormattedException("Not implemented by " + appName +
                   ": processing of source files within a non-directory " +
                   "(e.g. jar file): " + location));
            return null;
        }
        return location;

    }


    /** This filter returns all the files that are processed when a 
        directory or package is named on the command-line.
    */
    //@ also ensures \result != null;
    protected FilenameFilter filenameFilter() {
        if (((JspOptions)options).diff()) {
            return new SuffixFilter(validSuffixes);
        } else {
            return new Filter();
        }
    }

    /** This class is used with the Directory.list method to list
        those files in a directory that this program is interested in
        processing - in this case, all those that end in a valid suffix.
    */
    static class SuffixFilter implements FilenameFilter {
        protected String[] suffixes;
        public SuffixFilter(String[] suf) { suffixes = suf; }

        //@ also
        //@ requires name != null && dir != null;
        //@ requires (* dir is a Directory, name is an object in dir *);
        //@ ensures (* \result is true iff name ends with a legal suffix *);
        public boolean accept(File dir, String name) {
            for (int i=0; i<suffixes.length; ++i) {
                if (name.endsWith(suffixes[i])) return true;
            }
            return false;
        }
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------
    
    /** This function creates an object to do the parsing of the
     *  command line arguments and to store the values of the flags
     *  and options so obtained (it does not actually do the argument
     *  parsing).  It is overridden here from mjc.Main so that we can
     *  instantiate a JspOptions object, in order to parse some new
     *  options specific to this compiler (cf. JspOptions.opt).
     *  @see JspOptions */
    protected MjcCommonOptions makeOptionsInstance() {
        return new JspOptions("jmlspec");
    }
        
    /** Returns opt as an instance of MjcCommonOptions so it can be assigned 
     *  to the options variable in mjc's Main.  This is done so we can access
     *  options that are only processed in MjcCommonOptions.
     */
    //@ also
    //@ requires opt instanceof JspOptions;
    protected MjcCommonOptions getOptionsInstance(MjcCommonOptions opt) {
        return opt;
    }

    protected String getWarningFilterNameFromOptions(MjcCommonOptions opts)
    {
        return ((JspOptions)opts).filter();
    }

    // -------------------------------------------------------------
    // TASK AND TASK SEQUENCING
    // -------------------------------------------------------------

    public ParseTask firstTask( ArrayList infiles ) {
        return new JspParseTask( infiles );
    }

    protected void runSetInitialTasks(ArrayList infiles) {
	taskQueue.add( new JspBinaryPrinterTask());
        ParseTask first = firstTask( infiles );
        setMainSequenceID( first.sequenceID() );
        taskQueue.add( first );
    }

    /**
     * Adds JspPrettyPrinterTask or JspCompareTask after the regular parsing.
     * Note that we only do the parsing, no other checking.
     *
     */
    protected Task createTaskAfter(Task oldTask) {
        if (oldTask instanceof ParseTask) {
            if (! ((JspOptions)options).diff() ) {
                if (oldTask.sequenceID() != mainSequenceID()) {
                    // other files needed indirectly
                    return null;
                } else {
                    // explicitly mentioned files
                    return new JspPrettyPrinterTask(
                        ((Trees)oldTask).trees(), oldTask.sequenceID());
                }
            } else {
                if (oldTask.sequenceID() != mainSequenceID()) {
                    map = null;
                    return null;
                } else {
                    return new JspCompareTask(
                        ((Trees)oldTask).trees(), oldTask.sequenceID());
                }
            }
        } else if (oldTask instanceof JspPrettyPrinterTask) {
            return null;
        } else if (oldTask instanceof JspBinaryPrinterTask) {
            return null;
        } else if (oldTask instanceof JspCompareTask) {
            return null;
        }
        return super.createTaskAfter(oldTask);
    }

    public void doBinaries() {
	Iterator i = binaries.iterator();
	while (i.hasNext()) {
	    String nm = (String)i.next();
	    CClass cclass = CTopLevel.loadClass(nm);
	    if (cclass != CClass.CLS_UNDEFINED) {
		(new JspBinaryPrinter(cclass,((JspOptions)options).specLevel())).print(options.destination(),
				((JspOptions)options).suffix());
	    } else {
		String location = ClassPath.findPackage(nm);
		if (location == null) {
		    
		    reportTrouble(new FormattedException("Error: " + appName + 
			   " could not find file or package or class " +
				nm.replace('/','.') ));
		} else {
		    String dir = location + File.separator + nm;
		    ArrayList ds = new ArrayList(1);
		    ArrayList fs = new ArrayList();
		    ds.add(new File(dir));
		    handleDirectories(ds,fs);
		    taskQueue.add(new JspParseTask(fs).setToMainSequenceID());
		}
	    }
	}
    }

    public class JspBinaryPrinterTask extends Task {
	public JspBinaryPrinterTask() {
	    super(PRI_SKELETON_GENERATION, mainSequenceID());
	}
	public void execute() {
	    doBinaries();
	}
    }

    /**
     * A task for generating specification skeletons.
     */
    public class JspPrettyPrinterTask extends TreeProcessingTask {
        public JspPrettyPrinterTask(CompilerPassEnterable[] trees,
                                       Object sequenceID) {
            super(PRI_SKELETON_GENERATION, sequenceID, trees, 
                  JspMessages.SKELETON_GENERATED);
        }


        protected void processTree(CompilerPassEnterable tree) {
            if (tree == null || !(tree instanceof JCompilationUnit)) {
                return;
            }

            JCompilationUnit cunit = (JCompilationUnit) tree;
            File sourceFile = cunit.file();
            String originalName = sourceFile.getName();
            // extract the base file name of the compilation unit
            String name = originalName;
            int p = name.lastIndexOf('.');
            if (p == -1) {
                reportTrouble(new PositionedError(null,JspMessages.NO_BASE_NAME, 
                                Utils.relativePathTo(sourceFile)) );
                return;
            } else {
                name = name.substring(0,p);
            }
            
            String packageName = cunit.packageNameAsString();
            File targetDir = 
                destination.classFileTargetDirectory(packageName,sourceFile);
            File file = new File(
                targetDir, name + ((JspOptions) options).suffix() );
            
            //System.out.println("GENERATING " + file.getAbsolutePath());
            if (file.exists() && ! ((JspOptions)options).overwrite()) {
                reportTrouble(new CWarning(TokenReference.build(file,0),
					   JspMessages.FILE_NOT_GENERATED,
                                           Utils.relativePathTo(file)) );
                return;
            }
            
            // create any needed directories
            targetDir.mkdirs();
            
            // do the file generation
            JspPrettyPrinter pp = 
                    new JspPrettyPrinter(file, new CModifier(),
                                         (JspOptions) options, originalName);

            cunit.accept(pp);
            pp.close();
 
        }
     }

    /**
     * A task for generating specification skeletons.
     */
    public class JspParseTask extends ParseTask {
            // Extends ParseTask in order to be able to use parseFile

        protected ArrayList additionalFiles = new ArrayList();
        
        public JspParseTask(ArrayList infiles) {
            super(infiles);
            map = new HashMap(); 
        }

        protected JCompilationUnitType parseFile(File file) {
            JCompilationUnitType unit = super.parseFile(file);
            if (unit == null) return unit;
            if (!((JspOptions)options).diff()) return unit;

            // get the package and file name

            // This is a bit of an odd design.  But we put all this parsing of related
            // specification files here because we need parseFile from the parent class
            // to do it.  Also we would rather do all the extra parsing after all the 
            // files on the command-line are parsed, but we aren't supposed to override
            // the execute method.
            
            // There is an outstanding problem that soem files get parsed twice, if one
            // lists more than one file of a refinement sequence on the command line. (FIXME)
            
            String originalName = unit.file().getName();
            String name = originalName;
            int p = name.lastIndexOf('.');
            if (p == -1) { // A non-fatal error
                reportTrouble(new CWarning(TokenReference.build(unit.file(),0),JspMessages.NO_BASE_NAME, Utils.relativePathTo(unit.file())) );
                return unit;
            }
            
            String suffix = name.substring(p);
            name = name.substring(0,p);
            int suffixIndex = findSuffixIndex(suffix);
            if (suffixIndex == -1) { // A non-fatal error
                reportTrouble(new CWarning(TokenReference.build(unit.file(),0),JspMessages.INVALID_SUFFIX, originalName) );
                return unit;
            } 

            // Get package name
            
            String pname = unit.packageNameAsString();
            String qualifiedName = pname + name; // with separation by /
            File ff = findFile(qualifiedName,suffix);
            try { // various non-fatal errors
                if (ff == null) { 
                    reportTrouble(new CWarning(TokenReference.build(unit.file(),0),JspMessages.BAD_CLASSPATH, originalName, qualifiedName.replace('/','.')) );
                } else if (!ff.getCanonicalPath().equals(unit.file().getCanonicalPath())) {
                    reportTrouble(new CWarning(TokenReference.build(ff,0),JspMessages.BAD_LOCATION, originalName, Utils.relativePathTo(ff)) );
                }
            } catch (IOException e) {
                    reportTrouble(new CWarning(TokenReference.build(ff,0),JspMessages.IOEXCEPTION, e.toString()) );
                    return unit;
            }

            // Enter the ast from parsing into the map
            // Parse any other related specification files
            
            JCompilationUnitType[] asts = (JCompilationUnitType[]) map.get(qualifiedName);
            if (asts == null) {
                asts = new JCompilationUnitType[validSuffixes.length];
                map.put(qualifiedName,asts);
                asts[suffixIndex] = unit;
                for (int i=0; i<validSuffixes.length; ++i) {
                    if (asts[i] != null) continue;
                    File f = findFile(qualifiedName,validSuffixes[i]);
                    if (f == null) continue;
                    asts[i] = parseFile(f);
                }
            } else {
                asts[suffixIndex] = unit;
            }
            
            return unit;
        }
        
     }

    // Utility method to find files on the classpath
    File findFile(String qname, String suffix) {
        ClassPath.ClassDescription d = ClassPath.getFile(qname,suffix);
        if (d == null) return null;
        
        if (! (d instanceof ClassPath.FileClassDescription)) {
            // a caution - the source is presumably in a zip file - the compiler can't read from zip
            reportTrouble(new CWarning(null,JspMessages.NON_FILE, qname+suffix ) );
            return null;
        }
        
        return ((ClassPath.FileClassDescription)d).file();
    
    }

    /**
     * A task for generating specification skeletons.
     */
    public class JspCompareTask extends TreeProcessingTask {
            // Extends ParseTask in order to be able to use parseFile
        
        public JspCompareTask(CompilerPassEnterable[] trees,
                                       Object sequenceID) {
            super(PRI_SKELETON_GENERATION, sequenceID, trees, 
                  JspMessages.SKELETON_GENERATED); // FIXME - Wrong message
        }

        protected void processTree(CompilerPassEnterable tree) {
            if (tree == null || !(tree instanceof JCompilationUnit)) {
                return;
            }

            JCompilationUnit cunit = (JCompilationUnit) tree;
            File sourceFile = cunit.file();
            String originalName = sourceFile.getName();
            // extract the base file name of the compilation unit
            String name = originalName;
            int p = name.lastIndexOf('.');
            String suffix = name.substring(p);
            name = name.substring(0,p);
            String pname = cunit.packageNameAsString();
            String qualifiedName = (pname + name); // with separation by /

            JCompilationUnitType[] asts = (JCompilationUnitType[]) map.get(qualifiedName);
            if (asts == null) return;
            String qualifiedNameDots = qualifiedName.replace('/','.');
            
            boolean isAJavaFile = asts[javaSuffixIndex] != null;
            int referenceIndex = findSuffixIndex(suffix);
            Class cc = null;
            if (!isAJavaFile) {
                // Look for a class file if no java file
                try {
                    cc = Class.forName(qualifiedNameDots);
                } catch (ClassNotFoundException e) {}
            }

            // Complain if there is only one file - nothing to compare against
            int count = 0;
            if (cc != null) count++;
            for (int i=0; i<validSuffixes.length; ++i) {
                if (asts[i] == null) continue;
                ++count;
                if (options.verbose()) inform(" FOUND " + qualifiedName + validSuffixes[i]);
            }
            if (count <= 1) reportTrouble(new CWarning(TokenReference.build(sourceFile,0),JspMessages.SINGLE, qualifiedName+suffix ) );


            if (cc == null) { 
                // If there is a java file: compare all other files to the java file
                // Else: compare all spec files to the one listed on the command line
                if (isAJavaFile) referenceIndex = javaSuffixIndex;
                for (int i=0; i<validSuffixes.length; ++i) {
                    if (asts[i] == null) continue;
                    if (i == referenceIndex) continue;
                    String out = (new JspComparator((JspOptions)options)).compare(asts[i],asts[referenceIndex]);
                    if (out.length() != 0) {
                        errorFound = true;
                        System.out.print(out); // out has its own line endings
                    }
                }
            } else {
                // Compare everything to the class file
                for (int i=0; i<validSuffixes.length; ++i) {
                    if (asts[i] == null) continue;
                    String out = (new JspComparator((JspOptions)options)).compareToClass(asts[i]);
                    if (out.length() != 0) {
                        errorFound = true;
                        System.out.print(out); // out has its own line endings
                    }
                }
            }
            
            // Erase the item we have just processed
            map.put(qualifiedName,null);
            
        }
     }

    /** Generates an int corresponding to the suffix; no particular order, just has to be unique
        - used as the index into an array.
    */
    //@ requires s != null;
    //@ ensures \result >= -1 && \result < validSuffixes.length;
    //@ ensures \result == -1 || validSuffixes[\result].equals(s);
    //@ pure
    static public int findSuffixIndex(String s) {
        for (int i=0; i<validSuffixes.length; ++i) {
            if (validSuffixes[i].equals(s)) return i;
        }
        return -1;
    }
    
    // -------------------------------------------------------------
    // DATA MEMBERS
    // -------------------------------------------------------------

    /** Map of fully-qualified class names (/-separated) to the arrays of ASTs that hold the
        parsed versions of all of the refinement files.
    */
    Map map = null;
    
    /** The task priority for generating test class. */
    public final static int PRI_SKELETON_GENERATION = 550;
    
    /** Valid suffixes for specification files. */
    public final static String[] validSuffixes =
        { ".refines-java", ".refines-spec", ".refines-jml", ".java", ".spec", ".jml",
            ".java-refined", ".spec-refined", ".jml-refined" };

    public final static int javaSuffixIndex = findSuffixIndex(".java");

    public void setTemplates(JspOptions options) {
        if (options.annotation().equals("none") || options.annotation().equals("")) {
            annotation = nocomment;
            classAnnotation = nocomment;
        } else if (options.annotation().equals("comment")) {
            annotation = comment;
            classAnnotation = comment;
        } else if (options.annotation().equals("also")) {
            annotation = also;
            classAnnotation = comment; // Don't have also in the class comment
        } else {
            ArrayList list = new ArrayList();
            String s = options.annotation();
            char separator = s.charAt(0);
            int pos = 1;
            int i;
            while ((i = s.indexOf(separator,pos)) != -1) {
                list.add(s.substring(pos,i));
                pos = i+1;
            }
            list.add(s.substring(pos));
            annotation = new String[list.size()];
            annotation = (String[])list.toArray(annotation);
            classAnnotation = comment;
        }

        if (options.classAnnotation() == null) { // nothing
        } else if (options.classAnnotation().equals("none") || options.classAnnotation().equals("")) {
            classAnnotation = nocomment;
        } else if (options.classAnnotation().equals("comment")) {
            classAnnotation = comment;
        } else {
            ArrayList list = new ArrayList();
            String s = options.classAnnotation();
            char separator = s.charAt(0);
            int pos = 1;
            int i;
            while ((i = s.indexOf(separator,pos)) != -1) {
                list.add(s.substring(pos,i));
                pos = i+1;
            }
            list.add(s.substring(pos));
            classAnnotation = new String[list.size()];
            classAnnotation = (String[])list.toArray(classAnnotation);
        }

    }
    static protected String[] annotation;
    static protected String[] classAnnotation;    

    static final String[] nocomment = new String[]{};
    static final String[] comment   = new String[]{"/*@","  @ ","  @*/"};
    static final String[] also      = new String[]{"/*@ also","  @ ","  @*/"};

}

