/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: NonDefaultOptionsTestSuite.java,v 1.4 2003/03/08 17:06:28 davidcok Exp $
 */

package org.jmlspecs.jmlspec.testcase;

import org.multijava.util.testing.TestCase;
import org.multijava.util.testing.*;
import junit.framework.*;
import java.io.*;
import java.util.Iterator;
import java.util.ArrayList;

/** This class runs the test cases that have a different set of command-line
    options for each test.  Thus each one needs its own execution of compile().
    For each one, the options are determined (they are specified in the test file
    by a line containing JMLFLAGS), the compile is run and the output is
    compared to the expected output.  There is a TestCase for each compile and a
    separate TestCase for each file comparison.
<P>
    The set of external test files is obtained by running the command
    'listTestsWithNonDefaultOptions'.
<P>
    The test must be run from the directory in which it resides - because
    it runs an external process that is expected to be in the current
    directory and opens/creates files in the current directory.

    @author David R. Cok
*/
public class NonDefaultOptionsTestSuite  extends DefaultOptionsTestSuite {

    public static Test suite() {
        return (new NonDefaultOptionsTestSuite("NonDefaultOptionsTestSuite"));
    }

    /** Standard constructor used by the JUnit framework. 
	It fills the TestSuite with all the tests.
    */
    public NonDefaultOptionsTestSuite(String name) { 
	super(0); // Create the superclass empty
	
	try {
	
	    Iterator i = new FileIterator("listTestsWithNonDefaultOptions");
	    while (i.hasNext()) {
		String s = (String)i.next();
		addTest(new Helper(s)); // the TestCase for the compile
		addTestsForFileComparisons(s);
	    }
	
	} catch (java.io.IOException e) {
	    throw new RuntimeException(e.toString());
	}
     }

    public NonDefaultOptionsTestSuite(int i) { 
	super(0); // Create the superclass empty
    }

    public boolean expectedStatus(String filename, String output) {
	return output.length() == 0 || filename.equals("Help.java") || filename.equals("BadSuffix.java");
    }
    
    /** This is the TestCase that does the compile() call. */
    public class Helper extends org.multijava.util.testing.TestCase {

	/** The constructor called by the JUnit framework. 
	    The argument is used as the name of the test as well as 
	    the name of the file to be tested.
	*/
	public Helper(String filename) {
	    super(filename);
	    fileToTest = filename;
	}
	
	/** The name of the file to be tested. */
	private String fileToTest;
    
	/** A holder for the command-line arguments for the call of compile();
	    the complete list of options consists of those specified in the
	    external test file, those listed here, and the appropriate filename.
	*/
	final String[] args = { };

	public void runTest() throws java.io.IOException {
	    //System.out.println("JMLDOC testfiles " + fileToTest);

	    // Find the flags that this test should use
	    ArrayList a = Utils.parseFoundLine("JMLFLAGS",fileToTest); 
	    a.remove(0);
	    // Add on the ones always used
	    for (int ia=0; ia<args.length; ++ia) a.add(args[ia]);
	    // Add on the filename
	    a.add(fileToTest);

	    // Execute the compile
	    // Set the output streams to write into the ByteArray
	    // and then execute the compile
	    ByteArrayOutputStream ba = new ByteArrayOutputStream();
	    boolean b = false;
	    try {
		Utils.setStreams(new PrintStream(ba));
		b = org.jmlspecs.jmlspec.Main.compile((String[])a.toArray(new String[0]));
	    } finally {
		Utils.restoreStreams();
	    }
	    String s = ba.toString();

	    if ((new File(fileToTest + "-expected")).exists()) {

		String ss = Utils.readFile(fileToTest + "-expected");
		Diff df = new Diff("expected", ss, "actual", s);
                if (!df.areDifferent()) {
                    // If the two strings match, the test succeeds and we make sure
                    // that there is no -ckd file to confuse anyone.
                    (new File(fileToTest+"-ckd")).delete();
                } else {
                    // If the strings do not match, we save the actual string and
                    // fail the test.
                    FileWriter f = new FileWriter(fileToTest+"-ckd");
                    f.write(s);
                    f.close();

                    fail(df.result());
                }

		// Check that the result value
		boolean exp = expectedStatus(fileToTest,s);
		assertTrue("Return value of the compile was not " + exp,b == exp);

	    } else {

		if (s.length() != 0) {
                    FileWriter f = new FileWriter(fileToTest+"-ckd");
                    f.write(s);
                    f.close();
		}

		// Check that there is no output
		assertTrue("Unexpected error or warning output has been produced: " + s, s.length() == 0);

		// Check that the result value is true
		assertTrue("Return value of the compile was not true",b);

	    }
	    
	}

    }
    
}


