// @(#)$Id: FancyTabbedPrintWriter.java,v 1.1 2004/02/03 23:12:31 leavens Exp $

// Copyright (C) 2004 Iowa State University

// This file is part of JML

// JML is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.

// JML is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with JML; see the file COPYING.  If not, write to
// the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

package org.jmlspecs.jmlunit;

import org.multijava.util.compiler.TabbedPrintWriter;
import java.io.*;

/**
 * A more convenient print writer.
 *
 * @author Gary T. Leavens
 * @version $Revision: 1.1 $
 */
public class FancyTabbedPrintWriter {

    //@ public model int TAB_SIZE;
    //@ public invariant TAB_SIZE > 0;

    //@ public model int leftMargin;
    //@ public invariant 0 <= leftMargin;

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Initialize this fancy tabbed print writer.  The default tab
     * size, 4, is appropriate for emacs. (:-)
     */
    //@ assignable objectState, TAB_SIZE, leftMargin;
    //@ ensures TAB_SIZE == 4 && leftMargin == 0;
    public FancyTabbedPrintWriter(/*@ non_null @*/ Writer writer) {
        this(writer, 4);
    }

    /**
     * Initialize this fancy tabbed print writer.
     */
    //@ requires tabSize > 0;
    //@ assignable objectState, TAB_SIZE, leftMargin;
    //@ ensures TAB_SIZE == tabSize && leftMargin == 0;
    public FancyTabbedPrintWriter(/*@ non_null @*/ Writer writer, int tabSize)
    {
        this.writer = new TabbedPrintWriter(writer);
        this.tabSize = tabSize;
    }

    /** Indents one TAB_SIZE. */
    //@ requires leftMargin < Integer.MAX_VALUE - TAB_SIZE;
    //@ assignable leftMargin;
    //@ ensures leftMargin == \old(leftMargin + TAB_SIZE);
    public void indent() {
        pos += tabSize;
    }

    /** Und-ndents one tabSize. */
    //@ requires 0 <= leftMargin;
    //@ assignable leftMargin;
    //@ ensures leftMargin == \old(leftMargin - TAB_SIZE);
    public void undent() {
        pos -= tabSize;
    }

    /** Prints the given object. */
    //@ assignable objectState;
    public void print(Object s) {
	print(s.toString());
    }

    /** Prints the given string. */
    //@ assignable objectState;
    public void print(String s) {
	writer.setPos(pos);
	writer.print(s);
    }

    /** Prints the given object with a trailing line separator. */
    //@ assignable objectState;
    public void println(Object s) {
	print(s.toString());
        writer.println();
    }

    /** Prints the given object with one tabSize indentation
     * and a trailing line separator. */
    //@ assignable objectState, leftMargin;
    //@ ensures \not_modified(leftMargin);
    public void printlnIn(Object s) {
        pos += tabSize;
	println(s.toString());
        pos -= tabSize;
    }

    /** Prints a line separator. */
    //@ assignable objectState;
    public void newLine() {
	writer.println();
    }

    /** Set the tab size to the given integer.
     */
    //@ requires n > 0;
    //@ assignable TAB_SIZE;
    //@ ensures TAB_SIZE == n;
    public void setTabSize(int n) {
        tabSize = n;
    }

    /** Close this file. */
    //@ assignable objectState;
    public void close() {
        writer.close();
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** The writer to which the output is to be printed. */
    private /*@ non_null @*/ TabbedPrintWriter writer;
    //@ in objectState; maps writer.objectState \into objectState;

    /** The depth of indentation.
     * <pre><jml>
     * private invariant_redundantly tabSize > 0;
     * </jml></pre>
     */
    private int tabSize;
    //@ in TAB_SIZE;
    //@ private represents TAB_SIZE <- tabSize;

    /** The column number where the next text should be written by the
     * writer.
     * <pre><jml>
     * private invariant pos >= 0;
     * </jml></pre>
     */
    private int pos = 0;
    //@ in leftMargin;
    //@ private represents leftMargin <- pos;

}
