// @(#)$Id: JMLTestResult.java,v 1.8 2005/07/07 21:03:03 leavens Exp $
//
// Copyright (C) 2005 Iowa State University
//
// This file is part of the runtime library of the Java Modeling Language.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with JML; see the file LesserGPL.txt.  If not, write to the Free
// Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA.

package org.jmlspecs.jmlunit;

import junit.framework.*;
import java.util.HashSet;
//@ model import org.jmlspecs.models.*;

/**
 * A class for collecting the results of executing test cases. 
 * In addition to test <i>failures</i> and <i>errors</i>
 * collected by the superclass, this class accumulates JML-specific 
 * test information such as the number of meaningless test runs
 * (e.g., test inputs causing precondition violations).
 *
 * @author Yoonsik Cheon
 * @author Gary T. Leavens
 * @version $Revision: 1.8 $
 */
public class JMLTestResult extends TestResult {
    
    /**
     * Constructs a new JMLTestResult object.
     */
    public JMLTestResult() {
        super();
        auditors = new HashSet();
    }

    //@ public non_null model JMLObjectSet jmlListeners;

    private /*@ non_null @*/ HashSet auditors; //@ in jmlListeners;
    //@ private represents jmlListeners <- JMLObjectSet.convertFrom(auditors);

    /** Add the argument to the set of JML listeners (only) Such
     * listeners will only be notified also of meaningless test
     * executions, and will not receive the notifications normal
     * TestListeners receive.
     * @see #addListener(JMLTestListener)
     */
    //@ requires jmlListener != null;
    //@ assignable jmlListeners;
    //@ ensures jmlListeners.has(jmlListener);
    public void addJMLListener(JMLTestListener jmlListener) {
        auditors.add(jmlListener);
    }

    /** Add the argument to the set of JML listeners and the set of
     * listeners maintained by the superclass.  Such listeners will be
     * notified also of meaningless test executions, in addition to
     * the notifications normal TestListeners receive.
     * @see #addJMLListener(JMLTestListener)
     */
    //@ requires jmlListener != null;
    //@ assignable objectState, jmlListeners;
    //@ ensures jmlListeners.has(jmlListener);
    public void addListener(JMLTestListener jmlListener) {
        addJMLListener(jmlListener);
        super.addListener(jmlListener);
    }

    /**
     * Increments the total number of JML test runs that are
     * meaningless (i.e., those with entry precondition violations),
     * and notifies any JMLTestListeners.
     *
     * <pre><jml>
     * ensures meaninglessCount == \old(meaninglessCount + 1);
     * </jml></pre>
     */
    // !FIXME! assignable below should allow the listener's state to change
    // this should be all of the "changable state" in each listener.
    //@ assignable (* the changable state in each listener *);
    //@ assignable meaninglessCount;
    //@ ensures meaninglessCount == \old(meaninglessCount + 1);
    public synchronized void addMeaningless(Test test) {
        meaninglessCount++;
	append(test.toString());
        java.util.Iterator li = auditors.iterator();
        while (li.hasNext()) {
            JMLTestListener jmlListener = (JMLTestListener) li.next();
            jmlListener.addMeaningless(test);
        }
    }

    private static final String line_sep
        = System.getProperty("line.separator");

    /** Appends information about meaningless test cases.
     */
    public synchronized void append(java.lang.String msg) {
	meaninglessInformation.append(msg);
	meaninglessInformation.append(line_sep);
    }

    /** Returns information about all of the meaningless test cases.
     */
    public synchronized java.lang.String meaninglessTestCaseInfo() {
	return meaninglessInformation.toString();
    }

    /**
     * Returns the total number of JML meaningless test runs.
     * <pre><jml>
     * ensures \result == meaninglessCount;
     * </jml></pre>
     */
    public synchronized int meaninglessCount() {
        return meaninglessCount;
    }

    /** Counter for the total number of meaningless JML test runs. 
     *
     * <pre><jml>
     * public invariant meaninglessCount >= 0;
     * </jml></pre>
     */
    private /*@ spec_public @*/ int meaninglessCount;

    /** StringBuffer that contains information on all of the
     * meaningless test runs.
     */
    private /*@ spec_public @*/ StringBuffer meaninglessInformation 
        = new StringBuffer();
}
