// @(#)$Id: JMLTestRunner.java,v 1.26 2005/07/07 21:03:03 leavens Exp $
//
// Copyright (C) 2005 Iowa State University
//
// This file is part of the runtime library of the Java Modeling Language.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with JML; see the file LesserGPL.txt.  If not, write to the Free
// Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
// 02110-1301  USA.

package org.jmlspecs.jmlunit;

import java.io.PrintStream;
import java.util.Enumeration;
import java.lang.reflect.*;
import java.util.Stack;

import junit.framework.*;
import junit.runner.*;
import junit.textui.*;

/**
 * A JML/JUnit test runner class. This class refines JUnit's 
 * text-based test runner class <code>TestRunner</code> and provides
 * a command line based tool to run tests.
 * 
 * <pre>
 * java org.jmlspecs.jmlunit.JMLTestRunner [-M] [-R] [-testMode] [-list] [-wait] TestCaseClass
 * </pre>
 *
 * JMLTestRunner expects the name of a TestCase class as argument.
 * If this class defines a static <code>suite</code> method it 
 * will be invoked and the returned test is run. Otherwise all 
 * the methods starting with "test" having no arguments are run.
 *
 * <p>
 * When the -wait command line argument is given JMLTestRunner
 * waits until the users types RETURN.
 *
 * <p>
 * JMLTestRunner prints a trace as the tests are executed followed by a
 * summary at the end. In addition to the summary by TestRunner, 
 * it produces JML/JUnit specific test results such as the total 
 * number of test cases and the number of meaningful test cases.
 *
 * <p>
 * More information about meaningless test cases is printed by
 * using the -M flag.  The -R flag can be used to suppress reporting
 * coverage information which is normally printed after successful
 * runs.
 * <p>
 * The -list option simply lists the tests by name, in order, without
 * running them.  The -testMode option does not print out time information
 * so that test output is consistent from run to run.
 *
 * @author Yoonsik Cheon
 * @author Gary T. Leavens
 * @author David Cok
 * @version $Revision: 1.26 $
 */

public class JMLTestRunner extends TestRunner {

    // ----------------------------------------------------------------------
    // FIELDS
    // ----------------------------------------------------------------------

    /** If true, then information about the cases declared meaningless is
	printed at the end of the JUnit run.
    */
    protected static boolean showMeaninglessCases = false;

    /** If true, then information about the coverage of tests is
	printed at the end of the JUnit run.
    */
    protected static boolean showCoverage = true;

    /** If true, then no information that varies from test run to test run
	(such as elapsed time) is printed.
    */
    protected static boolean testMode = false;

    /** If true, then instead of running tests, they are simply listed by name
	in order.
    */
    protected static boolean listMode = false;

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** 
     * Constructs a JMLTestRunner using System.out for all output. 
     */
    public JMLTestRunner() {
        this(System.out);
    }

    /** 
     * Constructs a JMLTestRunner using the given stream for all
     * the output. 
     */
    public JMLTestRunner(PrintStream writer) {
        this(new JmlResultPrinter(writer));
    }
	
    /**
     * Constructs a TestRunner using the given ResultPrinter all the output
     */
    private JMLTestRunner(JmlResultPrinter printer) {
        super(printer);
        this.printer = printer;
    }

    // ----------------------------------------------------------------------
    // STATIC METHODS
    // ----------------------------------------------------------------------

    /**
     * Performs test execution for the user-specified testcase class.
     * The name of a TestData class (a subclass of junit.TestCase)
     * must be given as an argument.  If this class defines a static
     * <code>suite</code> method it will be invoked and the returned
     * test is run. Otherwise all the methods starting with "test"
     * having no arguments are run.
     *
     * <pre> 
     * java org.jmlspecs.jmlunit.JMLTestRunner [-M] [-R] [-testMode] [-wait] TestCaseClass 
     * </pre>
     */	
    public static void main(String args[]) {
	String[] newargs = args;
        int i = 0;
        while (i < newargs.length) {
            if (newargs[i].equals("-M")) {
		showMeaninglessCases = true;
                args = newargs;
		newargs = new String[args.length-1];
		System.arraycopy(args,0,newargs,0,i);
		System.arraycopy(args,i+1,newargs,i,newargs.length);
            } else if (newargs[i].equals("-R")) {
                showCoverage = false;
                args = newargs;
		newargs = new String[args.length-1];
		System.arraycopy(args,0,newargs,0,i);
		System.arraycopy(args,i+1,newargs,i,newargs.length);
            } else if (newargs[i].equals("-testMode")) {
                testMode = true;
                args = newargs;
		newargs = new String[args.length-1];
		System.arraycopy(args,0,newargs,0,i);
		System.arraycopy(args,i+1,newargs,i,newargs.length);
            } else if (newargs[i].equals("-list")) {
                listMode = true;
                args = newargs;
		newargs = new String[args.length-1];
		System.arraycopy(args,0,newargs,0,i);
		System.arraycopy(args,i+1,newargs,i,newargs.length);
            } else {
                i++;
            }
        }
	if (listMode) {
	    Stack stack = new Stack();
	    for (i=0; i<newargs.length; ++i) {
		String name = newargs[i];
		try {
		    Class c = Class.forName(name);
		    Method m = c.getDeclaredMethod("suite",new Class[0]);
		    junit.framework.TestSuite suite = (junit.framework.TestSuite)m.invoke(null,new Object[0]);
		    Enumeration en = suite.tests();
		    while(true) {
			while (en.hasMoreElements()) {
			    Object t = en.nextElement();
			    if (t instanceof TestCase) {
				System.out.println(((TestCase)t).getName());
			    } else if (t instanceof junit.framework.TestSuite) {
				stack.push(en);
				en = ((junit.framework.TestSuite)t).tests();
			    } else {
				System.out.println( "UNKNOWN " + t.getClass());
			    }
			}
			if (stack.isEmpty()) break;
			en = (Enumeration)stack.pop();
		    }
		} catch (Exception e) {
		    System.out.println("Exception on " + name + ": " + e);
		}
	    }
	    return;
	}
        JMLTestRunner testRunner= new JMLTestRunner();
        try {
            TestResult result = testRunner.start(newargs);
            if (!result.wasSuccessful()) {
                System.exit(FAILURE_EXIT);
            }
            System.exit(SUCCESS_EXIT);
        } catch(Exception e) {
            System.err.println(e.getMessage());
            System.exit(EXCEPTION_EXIT);
        }
    }

    /**
     * Runs a single test and collects its results. This method can 
     * be used to start a test run from a program, e.g.
     * <pre>
     * public static void main (String[] args) {
     *     org.jmlspecs.jmlunit.JMLTestRunner.run(suite());
     * }
     * </pre>
     */
    static public TestResult run(Test suite) {
        TestRunner testRunner= new JMLTestRunner();
        return testRunner.doRun(suite, false);
    }

    /**
     * Runs a single test and waits until the user types RETURN.
     */
    static public void runAndWait(Test suite) {
        TestRunner testRunner= new JMLTestRunner();
        testRunner.doRun(suite, true);
    }

    // ----------------------------------------------------------------------
    // OVERRIDDEN METHODS
    // ----------------------------------------------------------------------
    
    /**
     * Performs test execution and print test report. This method is
     * overridden here to produce a JML-specific summary report.  The
     * method print of the class ResultPrinter is package-visible,
     * thus all these troubles of overriding.
     */
    public TestResult doRun(Test suite, boolean wait) {
        TestResult result= createTestResult();
        result.addListener(printer);
        long startTime= System.currentTimeMillis();
        suite.run(result);
        long endTime= System.currentTimeMillis();
        long runTime= endTime-startTime;
        if (testMode) runTime = 0;
	printer.print(result, runTime);

        pause(wait);
        return result;
    }

    /**
     * Creates a test result object to be used for a test run.
     */
    protected TestResult createTestResult() {
        return new JMLTestResult();
    }

    //@ invariant (* printer == super.fPrinter *);
    private JmlResultPrinter printer;

    // ----------------------------------------------------------------------
    // INNER CLASSES
    // ----------------------------------------------------------------------

    public static class JmlResultPrinter extends ResultPrinter {

        /** A number of blanks the size of the chr string "WARNING: ". */
        private static final String sizeOfWarningBlanks = "         ";

        //@ invariant (* writer == super.fWriter *);
        private PrintStream writer;

	JmlResultPrinter(PrintStream writer) {
		super(writer);
                this.writer = writer;                
	}

        /** 
         * Prints test report inluding a JML-specific summary. This
         * method is redefined here as the superclass's is
         * package-visible.
         */
	public synchronized void print(TestResult result, long runTime) {
            this.printHeader(runTime);
	    printErrors(result);
	    printFailures(result);
	    printFooter(result);
	}

        /**
         * Prints the footer of the test report. This method is
         * overridden here to print JML-specific summary footer.
         */
        public void printFooter(TestResult result) {
            super.printFooter(result);
            if (!(result instanceof JMLTestResult)) {
                return;
            }

            JMLTestResult jresult = (JMLTestResult) result;

            writer.println("JML test runs: " + 
                          (jresult.runCount() - jresult.meaninglessCount()) +
                           "/" + jresult.runCount() + 
                           " (meaningful/total)");

            if (showMeaninglessCases && jresult.meaninglessCount() != 0) {
		writer.println(jresult.meaninglessCount() + 
                            " test cases declared meaningless:");
		writer.println(jresult.meaninglessTestCaseInfo());
	    }

	    // Print test coverage
            if (showCoverage) {
                org.jmlspecs.jmlrac.runtime.JMLChecker.reportCoverage(writer);
            }
        }
    }
}
