/* $Id: Main.java,v 1.31 2006/08/01 23:22:14 chalin Exp $
 *
 * Copyright (C) 2002-2004 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlunit;

import org.multijava.mjc.*;
import org.multijava.util.Utils;
import org.multijava.util.compiler.CWarning;
import java.io.*;

/**
 * A class implementing the entry point of the JML/JUnit test oracle
 * generator. 
 * Use <code>java org.jmlspecs.jmlunit.Main --help</code>
 * to get usage options.  
 *
 * @author Yoonsik Cheon
 * @author David R. Cok
 * @author Kristina Boysen
 * @author Gary T. Leavens
 * @version $Revision: 1.31 $
 */
public class Main extends org.multijava.mjc.Main implements Constants {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /** Constructs an object for generating test oracles. */
    public Main() {
        super();
        parseJavadoc = true; // for deprecated flag
        appName = "jmlunit"; // for error messages
    }
    
    // ----------------------------------------------------------------------
    // EXPORTED APIS
    // ----------------------------------------------------------------------

    /**
     * An entry point when starting from the command line.
     *
     * @param args The command line arguments
     */
    public static void main(String[] args) {
        boolean success;
        success = compile(args);
        System.exit(success ? 0 : 1);
    }

    /**
     * Generates JML/JUnit test oracle classes with the given argument,
     * <code>args</code>, and returns <code>true</code> if all the test
     * classes are successfully generated.
     */
    public static boolean compile(String[] args) {
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    // We will allow declarations without bodies in order to
                    // allow the generation of tests from .spec files.  We
                    // could try to figure out if the file is indeed a .spec
                    // and still require method bodies for .java files, but
                    // it is a pain to find that out given the information
                    // available here, and does not seem worth the complexity
                    // it would introduce. -- DRCok
                    return true;
                }
            }
        );

        return new Main().run(args);
    }
        
    // ----------------------------------------------------------------------
    // RUN FROM GUI
    // ----------------------------------------------------------------------
    
    /**
     * Entry point for the GUI
     *
     * @param   args    the file, package, and directory names
     * @param   opt     the options for the tool
     * @param   os      the output stream for the compiler messages
     */
    public static boolean compile(String[] args, JntOptions opt,
                                  OutputStream os) {
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    // We will allow declarations without bodies in order to
                    // allow the generation of tests from .spec files.  We
                    // could try to figure out if the file is indeed a .spec
                    // and still require method bodies for .java files, but
                    // it is a pain to find that out given the information
                    // available here, and does not seem worth the complexity
                    // it would introduce. -- DRCok
                    return true;
                }
            }
        );
        return new Main().run(args, opt, os);
    }
        
    // ----------------------------------------------------------------------
    // HELPER METHODS
    // ----------------------------------------------------------------------

    static String genPackageName(JCompilationUnitType cunit, 
                                        JntOptions options) {
        String testPackageName = options.packageName();
        JPackageName pkg = cunit.packageName();
        String pname = pkg.getName().replace('/','.');
        if (testPackageName == null) {
                // Use package name as is
        } else if (testPackageName.startsWith("+")) {
                String suffix = testPackageName.substring(1);
                if (!suffix.startsWith(".")) suffix = "." + suffix;
                pname = pname + suffix;
        } else if (testPackageName.endsWith("+")) {
                String prefix = testPackageName.substring(0,
                                                testPackageName.length()-1);
                if (!prefix.endsWith(".")) prefix = prefix + ".";
                pname = prefix + pname;
        } else {
                pname = testPackageName;
        }
        return pname;
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------
    
    /** This function creates an object to do the parsing of the
     *  command line arguments and to store the values of the flags
     *  and options so obtained (it does not actually do the argument
     *  parsing).  It is overridden here from mjc.Main so that we can
     *  instantiate a JntOptions object, in order to parse some new
     *  options specific to this compiler (cf. JntOptions.opt).
     *  @see JntOptions */
    protected MjcCommonOptions makeOptionsInstance() {
        return new JntOptions("jmlunit");
    }
        
    /** Returns opt as an instance of MjcCommonOptions so it can be assigned 
     *  to the options variable in mjc's Main.  This is done so we can access
     *  options that are only processed in MjcCommonOptions.
     */
    //@ also
    //@ requires opt instanceof JntOptions;
    protected MjcCommonOptions getOptionsInstance(MjcCommonOptions opt) {
        return opt;
    }

    protected String getWarningFilterNameFromOptions(MjcCommonOptions opts)
    {
        return ((JntOptions)opts).filter();
    }

    // -------------------------------------------------------------
    // TASK AND TASK SEQUENCING
    // -------------------------------------------------------------

    /**
     * Returns the next task to be added to the task queue after the given
     * task, <code>oldTask</code>, completes.
     *
     * <pre><jml>
     * also
     *   requires oldTask != null && oldTask.completed;
     *   ensures \result == null || 
     *     (\result.sequenceID() == oldTask.sequenceID());
     * </jml></pre>
     */
    protected Task createTaskAfter(Task oldTask) {
        if (oldTask instanceof TypecheckTask) {
            // test class generation only for explicitly mentioned files.
            if (oldTask.sequenceID() == mainSequenceID()) {
                return new TestClassGenerationTask(
                    ((TypecheckTask)oldTask).trees(), oldTask.sequenceID());
            } else {
                return null;
            }
        } else if (oldTask instanceof TestClassGenerationTask) {
            return null;
        }
        return super.createTaskAfter(oldTask);
    }

    /**
     * A task for generating JML/JUnit test oracle classes.
     */
    public class TestClassGenerationTask extends TreeProcessingTask {
        public TestClassGenerationTask(CompilerPassEnterable[] trees,
                                       Object sequenceID) {
            super(PRI_TEST_CLASS_GENERATION, sequenceID, trees, 
                  JntMessages.TEST_CLASS_GENERATED);
        }

        /** Returns true if it is possible to generate test classes
         * for the given compilation unit. I.e., it has at least one
         * type declaration (class or interface declaration).
         */
        private boolean isTestable(JCompilationUnit cunit) {
            JTypeDeclarationType[] typeDecls = cunit.typeDeclarations();
            return typeDecls != null && typeDecls.length > 0;
        }

        protected void processTree(CompilerPassEnterable tree) {
            if (tree == null || !(tree instanceof JCompilationUnit)) {
                return;
            }

            // see if the compilation unit has at least one type declaration
            JCompilationUnit cunit = (JCompilationUnit) tree;
            if (!isTestable(cunit)) {
                System.err.println("No test class is generated for the file" +
                  Utils.relativePathTo(cunit.file()) + "!");
                noteError();
                return;
            }

            File sourceFile = cunit.file();

            // extract the base file name of the compilation unit
            // Note: we allow any suffix.  Files with suffixes other than .java
            // may omit bodies of methods and constructors, but must otherwise be
            // valid Java (jmlunit parses Java, not JML).
            String name = sourceFile.getName();
            int p = name.lastIndexOf('.');
            if (p == -1) {
                System.err.println("No test class is generated for the file " +
                  Utils.relativePathTo(sourceFile) + "!");
                noteError();
                return;
            } else {
                if (!name.substring(p).equals(".java"))
                    reportTrouble(
                        new CWarning(null, JntMessages.REFINEMENT_SUFFIX, name));

                name = name.substring(0,p);
            }

            // no further processing for jmlunit generated files
            if (name.endsWith(TEST_CLASS_NAME_POSTFIX) ||
                name.endsWith(TEST_DATA_NAME_POSTFIX)) {
                System.err.println("No test class is generated for the file " +
                  Utils.relativePathTo(sourceFile) + "!");
                noteError();
                return;         
            }

            File file = null;
            // packageName and targetDir are used to handle the -d option
            // and putting the output in the correct relative directory.
            String packageName = genPackageName(cunit, (JntOptions)options);
            packageName = packageName.replace('.','/');
            File targetDir = 
                destination.classFileTargetDirectory(packageName,sourceFile);
            try {
                // generate test oracle class, i.e., JUnit test class.
                file = new File(targetDir, name + TEST_CLASS_FILE_NAME_POSTFIX);
                // make any necessary directories
                targetDir.mkdirs();
                Writer writer = new BufferedWriter(new FileWriter(file));
                TestClassGenerator tgen = 
                    new TestClassGenerator((JntOptions) options);
                try {
                    tgen.perform(writer, cunit, (JntOptions)options);
                } finally {
                    writer.close();
                }

                // generate skeleton test case if requested
                file = new File(targetDir, name + TEST_DATA_FILE_NAME_POSTFIX);
                if (testDataClassRequested(file)) {
                    tgen = new TestDataClassGenerator((JntOptions) options);
                    writer = new BufferedWriter(new FileWriter(file));
                    try {
                        tgen.perform(writer, cunit, (JntOptions)options);
                    } finally {
                        writer.close();
                    }
                }
            } 
            catch (IOException e) {
                System.err.println("Can't create or write to file " + 
                                   Utils.relativePathTo(sourceFile) + "!");
                noteError();
            }
        }

        private boolean testDataClassRequested(File file) {
            return ((JntOptions)options).testdata() && 
		(!file.exists() || ((JntOptions)options).useGenerator());
        }
    }

    // -------------------------------------------------------------
    // DATA MEMBERS
    // -------------------------------------------------------------

    /** The task priority for generating test class. */
    public final static int PRI_TEST_CLASS_GENERATION = 550;

}
