/* $Id: TestClassGenerator.java,v 1.61 2006/07/31 21:24:27 perryjames Exp $
 *
 * Copyright (C) 2002-2004 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlunit;

import org.multijava.mjc.*;
import org.multijava.util.*;
import org.multijava.util.compiler.*;

import java.io.*;
import java.util.*;

/**
 * A class for generating JML/JUnit test driver classes. 
 * An object of this class can generate a JML/JUnit test oracle class
 * from an object of {@link JCompilationUnit}.
 *
 * @author Yoonsik Cheon
 * @author Gary T. Leavens
 * @version $Revision: 1.61 $
 */
public class TestClassGenerator implements Constants {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a new test oracle generator.
     */
    public TestClassGenerator(/*@ non_null @*/ JntOptions options) {
        switch (options.testLevel()) {
        case 0:
            requestedVisibility = ACC_PUBLIC;
            break;            
        case 1:
            requestedVisibility = ACC_PROTECTED;
            break;
        default:
            requestedVisibility = ACC_ALL;
        }
        requestedDeprecation = options.deprecation();
        includeInherited = options.inherit();
        useGenerator=options.useGenerator();
        generatorPackageName=options.generatorPackageName();
        generatorClassName=options.generatorClassName();
    }

    // ----------------------------------------------------------------------
    // PUBLIC METHODS
    // ----------------------------------------------------------------------

    /**
     * Generates a JML/JUnit test oracle class for the given compilation 
     * unit. The generated test oracle class is written in the Java
     * source code form to the given writer object.
     *
     * <pre><jml>
     * requires cunit != null && writer != null;
     * </jml></pre>
     */
    //@ modifies (* the contents of fixtureTypes *);
    public void perform(/*@ non_null @*/ Writer writer,
                        /*@ non_null @*/ JCompilationUnit cunit,
			/*@ non_null @*/ JntOptions options) 
    {
        this.writer = new FancyTabbedPrintWriter(writer);
        this.packageString = Main.genPackageName(cunit,options);
        this.testedPackageString
            = cunit.packageName().getName().replace('/','.');

        // print file header
        printFileHeader();

        // print package and import statements
        printPackageStatement();
        printImportStatements(cunit);

        // generate test oracles for classes
	JTypeDeclarationType[] typeDecls = cunit.typeDeclarations();
	for (int i = 0; i < typeDecls.length ; i++) {
            if (isTestable(typeDecls[i], cunit.fileNameIdent())) {
                generateTestClass(cunit, typeDecls[i]);
            }
	}

        this.writer.close();
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------

    /** Prints a file header message. */
    protected void printFileHeader() {
        println("// This file was generated by jmlunit on " + new Date() + ".");
    }

    /** Prints import statements. */
    protected void printImportStatements(JCompilationUnit cunit) {
	JPackageImportType[] pkgs = cunit.importedPackages();
        boolean isFirstImport = true;
        for (int i = 0; i < pkgs.length ; i++) {
            if (pkgs[i].getName().equals("java/lang")) {
                continue;
            }
            if (isFirstImport) {
                newLine();
                isFirstImport = false;
            }
            println("import " + pkgs[i].getName().replace('/', '.') + ".*;");
        }

        // for the -N option, need to import the package being tested
        if (!(packageString.equals(testedPackageString))) {
            if (isFirstImport) {
                newLine();
                isFirstImport = false;
            }
            println("import " + testedPackageString + ".*;");
        }

	JClassOrGFImportType[] classes = cunit.importedClasses();
        for (int i = 0; i < classes.length ; i++) {
            if (isFirstImport) {
                newLine();
                isFirstImport = false;
            }
            println("import " + classes[i].getName().replace('/', '.') + ";");
        }

        if (PKG_JUNIT.length() == 0) {
            if (isFirstImport) {
                newLine();
                isFirstImport = false;
            }
            println("import junit.framework.*;");
        }
        if (PKG_JMLRAC.length() == 0) {
            if (isFirstImport) {
                newLine();
                isFirstImport = false;
            }
            println("import org.jmlspecs.jmlrac.runtime.*;");
        }
        if (PKG_JMLUNIT.length() == 0) {
            if (isFirstImport) {
                newLine();
                isFirstImport = false;
            }
            println("import org.jmlspecs.jmlunit.*;");
        }
    }

    /** Prints a javadoc comment for the current class. */
    protected void printClassJavadoc() {
        String name = typeDecl.ident();
        println("/** Automatically-generated test driver"
                + " for JML and JUnit based");
        println(" * testing of " + name
                + ". The superclass of this class should be edited");
        println(" * to supply test data."
                + " However it's best not to edit this class");
        println(" * directly; instead use the command");
        println(" * <pre>");
        println(" *  jmlunit " + name + ".java");
        println(" * </pre>");
        println(" * to regenerate this class whenever "
                + name + ".java changes.");
        println(" */");
    }

    /** Prints the class header of test oracle class. */
    protected void printClassHeader() {
        // if the class is not public, then JUnit complains about testing it
        print("public ");
	println("class " + testClassName);
        indent();
        println(" extends " + testClassName + "Data");
        undent();
        println("{");
    }

    /** 
     * Generates a test oracle class for the given class
     * declaration, <code>cdecl</code>. */
    //@ modifies (* the contents of fixtureTypes *);
    protected void generateTestClass(/*@ non_null @*/ JCompilationUnit cunit,
                                     /*@ non_null @*/ JTypeDeclarationType cdecl)
    {
        // initialize shared fields
        this.typeDecl = cdecl;
        testClassName = cdecl.ident() + TEST_CLASS_NAME_POSTFIX;
        initializeFixture();

        // javadoc
	newLine();
        printClassJavadoc();

        // class header
        printClassHeader();
        indent();

        // boilerplate methods
        printConstructor();
        printMain();

        printTestIsRACCompiled(cdecl);

        // print methods to construct the test suite and OneTest
        printSuite();
        printOneTest();
        // print methods to add each methods' tests and test object class
        printAddTestSuiteForEachMethod();

        // other helpers
        printPrivateHelperMethods();

        undent();
        println("}"); // end of class
    }

    /** Initialize the fixture to contain all of the types used as
     * paramaeters in all the methods. */
    //@ assignable fixtureTypes.objectState;
    protected void initializeFixture() {
        // fixture for methods declared in this class
        MethodsIterator mi = methodsIter();
        while (mi.hasNext()) {
            MethodInfo mdecl = mi.nextMethod();
            Parameter[] ps = mdecl.parameters();
            for (int n = 0; n < ps.length; n++) {
                fixtureTypes.add(ps[n].ctype);
            }
        }
    }

    /** Return the kind of method this is: either "method" or "constructor". */
    private String methodKind(MethodInfo mdecl) {
        if (mdecl.isConstructor()) {
            return "contructor";
        } else {
            return "method";
        }
    }

    /** Prints a boilerplate constructor. */
    protected void printConstructor() {
        println("/** Initialize this class. */");
        println("public " + testClassName + "(java.lang.String name) {");
        indent(); 
        println("super(name);");
        undent();
        println("}");
    }

    /** Prints a boilerplate main method. */
    protected void printMain() {
        newLine();
        println("/** Run the tests. */");
        println("public static void main(java.lang.String[] args) {");
        indent();
        println("org.jmlspecs.jmlunit.JMLTestRunner.run(suite());");
        println("// You can also use a JUnit test runner such as:");
        println("// junit.textui.TestRunner.run(suite());");
        undent();
        println("}");
    }

    /** Prints the <kbd>suite()</kbd> method. */
    protected void printSuite() {
        newLine();
        println("/** Return the test suite for this test class.  This will have");
        println("* added to it at least test$IsRACCompiled(), and any test methods");
        println("* written explicitly by the user in the superclass.  It will also");
        println("* have added test suites for each testing each method and");
        println("* constructor.");
        println("*/");
        println("//@ ensures \\result != null;");
        println("public static "+ PKG_JUNIT + "Test suite() {");
        indent();
        println(testClassName + " testobj");
        printlnIn("= new " + testClassName + "(\"" + testClassName + "\");");
        println(PKG_JUNIT + "TestSuite testsuite = testobj.overallTestSuite();");
        println("// Add instances of Test found by the reflection mechanism.");
        println("testsuite.addTestSuite(" + testClassName + ".class);");
        println("testobj.addTestSuiteForEachMethod(testsuite);");
        println("return testsuite;");
        undent();
        println("}");
    }

    /** Prints the nested <kbd>OneTest</kbd> class. */
    private void printOneTest() {
        newLine();
        println("/** A JUnit test object that can run a single test method.  This");
        println(" * is defined as a nested class solely for convenience; it can't");
        println(" * be defined once and for all because it must subclass its");
        println(" * enclosing class.");
        println(" */");
        print("protected static abstract class OneTest extends ");
        print(testClassName);
        println(" {");

        indent();
        newLine();

        println("/** Initialize this test object. */");
        println("public OneTest(String name) {");
        printlnIn("super(name);");
        println("}");

        newLine();

        println("/** The result object that holds information about testing. */");
        print("protected ");
        println(PKG_JUNIT + "TestResult result;");

        newLine();

        println("//@ also");
        println("//@ requires result != null;");
        print("public void run(");
        println(PKG_JUNIT + "TestResult result) {");
        printlnIn("this.result = result;");
        printlnIn("super.run(result);");
        println("}");

        newLine();

        println("/* Run a single test and decide whether the test was");
        println(" * successful, meaningless, or a failure.  This is the");
        println(" * Template Method pattern abstraction of the inner loop in a");
        println(" * JML/JUnit test. */");
        println("public void runTest() throws java.lang.Throwable {");
        indent();
        println("try {");
        printlnIn("// The call being tested!");
        printlnIn("doCall();");
        println("}");
        println("catch (" + PKG_JMLRAC + "JMLEntryPreconditionError e) {");
        printlnIn("// meaningless test input");
        printlnIn("addMeaningless();");
        println("} catch (" + PKG_JMLRAC + "JMLAssertionError e) {");
        indent();
        println("// test failure");
        println("int l = " + PKG_JMLRAC + "JMLChecker.getLevel();");
        println(PKG_JMLRAC + "JMLChecker.setLevel");
        printlnIn("(" + PKG_JMLRAC + "JMLOption.NONE);");
        println("try {");
        indent();
        println("java.lang.String failmsg = this.failMessage(e);");
        println(PKG_JUNIT + "AssertionFailedError err");
        printlnIn("= new " + PKG_JUNIT + "AssertionFailedError(failmsg);");
        
        println("err.setStackTrace(new java.lang.StackTraceElement[]{});");
        println("err.initCause(e);");
        println("result.addFailure(this, err);");
        undent();
        println("} finally {");
        printlnIn(PKG_JMLRAC + "JMLChecker.setLevel(l);");
        println("}");
        undent();
        println("} catch (java.lang.Throwable e) {");
        printlnIn("// test success");
        println("}");
        undent();
        println("}");

        newLine();

        println("/** Call the method to be tested with the appropriate arguments. */");
        println("protected abstract void doCall() throws java.lang.Throwable;");

        newLine();

        println("/** Format the error message for a test failure, based on the");
        println(" * method's arguments. */");
        println("protected abstract java.lang.String failMessage");
        printlnIn("(" + PKG_JMLRAC + "JMLAssertionError e);");

        newLine();

        println("/** Inform listeners that a meaningless test was run. */");
        println("private void addMeaningless() {");
        indent();
        println("if (result instanceof " + PKG_JMLUNIT + "JMLTestResult) {");
        indent();
        println("((" + PKG_JMLUNIT + "JMLTestResult)result)");
        printlnIn(".addMeaningless(this);");
        undent();
        println("}");
        undent();
        println("}");

        undent();
        println("}");
    }

    /** Prints the <kbd>addTestSuiteForEachMethod</kbd> method, and
     * the methods directly called by it and supporting nested classes. */
    private void printAddTestSuiteForEachMethod() {
        newLine();
        println("/** Create the tests that are to be run for testing the class");
        println(" * " + typeDecl.ident() + ".  The framework will then run them.");
        println(" * @param overallTestSuite$ The suite accumulating all of the tests");
        println(" * for this driver class.");
        println(" */");
        println("//@ requires overallTestSuite$ != null;");
        println("public void addTestSuiteForEachMethod");
        printlnIn("(" + PKG_JUNIT + "TestSuite overallTestSuite$)");
        println("{");
        indent();

        MethodsIterator mi = methodsIter();
        while (mi.hasNext()) {
            MethodInfo mdecl = mi.nextMethod();
            String mname = mdecl.ident();
            String className = testMethodClassName(mname);
            printAddCallFor(className);
        }
        undent();
        println("}");

        nameGen.reset();

        Parameter receiver = new Parameter(typeDecl.getCClass().getType(),
                                               "receiver$");
        mi = methodsIter();
        while (mi.hasNext()) {
            MethodInfo mdecl = mi.nextMethod();
            String mname = mdecl.ident();
            String className = testMethodClassName(mname);
            Parameter [] params = mdecl.parameters();
            boolean isInstance = !isStatic(mdecl.modifiers());
            boolean isConstructor = mdecl.isConstructor();
            printAddMethodFor(mdecl, mname, className, params, receiver,
                              isInstance, isConstructor);
            generateNestedTestClassFor(mdecl, mname, className,
                                       params, receiver,
                                       isInstance, isConstructor);
        }
    }

    /** Prints a call to the appropriateaddTestSuiteFor$X method. */
    //@ requires name != null;
    private void printAddCallFor(String name) {
        println("try {");
        indent();
        print("this.addTestSuiteFor$");
        print(name);
        println("(overallTestSuite$);");
        undent();
        println("} catch (java.lang.Throwable ex) {");
        indent();
        println("overallTestSuite$.addTest");
        indent();
        print("(new ");
        print(PKG_STRATEGIES);
        println("ConstructorFailed(ex));");
        undent();
        undent();
        println("}");
    }

    /** Prints a method to add the test cases for the given method to
     * a test suite. */
    //@ requires mdecl != null;
    private void printAddMethodFor(MethodInfo mdecl,
                                   String mname,
                                   String className,
                                   Parameter [] params,
                                   Parameter receiver,
                                   boolean isInstance,
                                   boolean isConstructor)
    {
        newLine();
        print("/** Add tests for the " + mname + " ");
        println(methodKind(mdecl));
        println(" * to the overall test suite. */");
        print("private void addTestSuiteFor$");
        println(className);
        printlnIn("(" + PKG_JUNIT + "TestSuite overallTestSuite$)");
        println("{");
        indent();
        println(PKG_JUNIT + "TestSuite methodTests$");
        printlnIn("= this.emptyTestSuiteFor(\"" + mname + "\");");
        println("try {");
        indent();

        // print while loop heads
        if (isInstance && !isConstructor) {
            printBodyLoopControl(mname, receiver,
                                 0, params.length);
        }
        int i;
        for (i = 1; i <= params.length; i++) {
            printBodyLoopControl(mname, params[i-1],
                                 i, params.length-i);
        }

        // print inner variable declarations
        if (isInstance && !isConstructor) {
            printInnerVarDecl(mname, receiver, 0);
        }
        for (i = 1; i <= params.length; i++) {
            printInnerVarDecl(mname, params[i-1], i);
        }

        // print code to add test to the test suite
        println("methodTests$.addTest");
        indent();
        print("(new ");
        print(className);
        print("(");
        if (isInstance && !isConstructor) {
            print("receiver$");
            if (params.length > 0) {
                print(", ");
            }
        }
        for (i = 1; i <= params.length; i++) {
            print(params[i-1].ident);
            if (i < params.length) {
                print(", ");
            }
        }
        println("));");
        undent();

        // print advances and ends of loops
        for (i = params.length; i >= 1; i--) {
            printAdvance(params[i-1].tvar() + "$" + i);
        }
        if (isInstance && !isConstructor) {
            printAdvance("receivers");
        }

        // print end of try block
        undent();
        print("} catch (");
        println(PKG_STRATEGIES + "TestSuiteFullException e$) {");
        printlnIn("// methodTests$ doesn't want more tests");
        println("}");
        println("overallTestSuite$.addTest(methodTests$);");

        undent();
        println("}");
    }

    /** Print the loop control statements for the body of adding tests
     * for a method. */
    private void printBodyLoopControl(String mname,
                                      Parameter param,
                                      int paramNum,
                                      int loopsThisSurrounds)
    {
        boolean isReceiver = (paramNum == 0);
        String thisType = param.type;
        String iterCall
            = "this." + param.tvar() + "Iter(\""
            + mname + "\", " + loopsThisSurrounds
            + ")";
        print(PKG_STRATEGIES);
        if (param.ctype.isPrimitive()) {
            print(capitalize(param.type.toString()));
        } else {
            print("Indefinite");
        }
        println("Iterator");
        String iterVarName
            = (isReceiver ? "receivers" : param.tvar() + "$" + paramNum)
            + "$iter";
        indent();
        println(iterVarName);
        print("= ");
        if (isReceiver) {
            println("new " + PKG_STRATEGIES + "NonNullIteratorDecorator");
            printlnIn("(" + iterCall + ");");
        } else {
            println(iterCall + ";");
        }
        undent();
        println("this.check_has_data");
        indent();
        println("(" + iterVarName + ",");
        print(" \"");
        if (isReceiver) {
            print("new NonNullIteratorDecorator(");
        }
        print(iterCall.replaceAll("\"", "\\\\\""));
        if (isReceiver) {
            print(")");
        }
        println("\");");
        undent();
        println("while (!" + iterVarName + ".atEnd()) {");
        indent();
    }

    /** Print the declaration of a final variable for this parameter. */
    private void printInnerVarDecl(String mname,
                                   Parameter param, int paramNum)
    {
        println("final " + param.type + " " + param.ident);
        indent();
        print("= ");
        if (!param.ctype.isPrimitive()) {
            print("(" + param.type.toString() + ") ");
        }
        if (paramNum == 0) {
            print("receivers");
        } else {
            print(param.tvar() + "$" + paramNum);
        }
        print ("$iter.get");
        if (param.ctype.isPrimitive()) {
            print(capitalize(param.type.toString()));
        }
        println("();");
        undent();
    }

    /** Print the advance of the given named iterator and closing bracket */
    private void printAdvance(String iterVarStart) {
        println(iterVarStart + "$iter.advance();");
        undent();
        println("}");
    }

    /** 
     * Generates and prints a test class for the given method
     * declaration, <code>mdecl</code>. 
     */
    private void generateNestedTestClassFor(MethodInfo mdecl,
                                            String mname,
                                            String className,
                                            Parameter [] params,
                                            Parameter receiver,
                                            boolean isInstance,
                                            boolean isConstructor)
    {
	newLine();
        print("/** Test for the " + mname + " ");
        print(methodKind(mdecl));
        println(". */");
	println("protected static class " + className + " extends OneTest {");
        indent();

        int i;
        if (isInstance && !isConstructor) {
            printTestClassFieldDecl(receiver, 0);
        }
        for (i = 1; i <= params.length; i++) {
            printTestClassFieldDecl(params[i-1], i);
        }

        // print the constructor
        newLine();
        println("/** Initialize this instance. */");
        print("public ");
        print(className);
        print("(");
        if (isInstance && !isConstructor) {
            printTestClassArgumentDecl(receiver);
            if (params.length > 0) {
                print(", ");
            }
        }
        for (i = 1; i <= params.length; i++) {
            printTestClassArgumentDecl(params[i-1]);
            if (i < params.length) {
                print(", ");
            }
        }
        println(") {");
        indent();

        print("super(\"" + mname +"\"");
	if (params.length > 0) {
	    for (i = 0; i<params.length; ++i) {
		print( i==0 ? "+ \":\" + " : "+ \",\" +");
		CType t = params[i].ctype;
		String typename = params[i].type;
		String id = params[i].ident;
		if (!t.isReference()) { // primitive type
		    if (typename.equals("char")) {
			// FIXME - this can be improved for character values
			//print("\"\\\'\"+" + "Integer.toHexString(" + id + ")"  + "+\"\\\'\"");
			print("\"\\\'\"+charToString(" + id + ")+\"\\\'\"");
		    } else 
			print(id);
		} else if (t.isArrayType()) {
		    int k = typename.indexOf('[');
		    print("("+id+"==null?\"null\":(\"{" + typename.substring(0,k+1) + "\"+"+id+".length + \"" + typename.substring(k+1) + "\"+\"}\"))");
		} else if (typename.equals("String") || typename.equals("java.lang.String")) {
		    print("("+id+"==null? \"null\" :(\"\\\"\"+"+id+"+\"\\\"\"))");
		} else {
		    print ("("+id+"==null? \"null\" :\"{"+typename+"}\")");
		}
	    }
	}
        println(");");

        if (isInstance && !isConstructor) {
            printTestClassConstrAssign(receiver);
        }
        for (i = 1; i <= params.length; i++) {
            printTestClassConstrAssign(params[i-1]);
        }

        undent();
        println("}");
        
        // print the doCall method
        printTestClassDoCall(mname, params, receiver,
                             isInstance, isConstructor);

        // print the failMessage method
        printTestClassFailMessage(mdecl, mname, params, receiver,
                                  isInstance, isConstructor);
        undent();
	println("}");
    }

    /** Print the declaration of a field for this parameter. */
    private void printTestClassFieldDecl(Parameter param, int paramNum)
    {
        print("/** ");
        if (paramNum == 0) {
            print("The receiver");
        } else {
            print("Argument " + param.ident);
        }
        println(" */");
        println("private " + param.type + " " + param.ident + ";");
    }

    /** Print declaration of an argument. */
    private void printTestClassArgumentDecl(Parameter param)
    {
        print(param.type + " " + param.ident);
    }

    /** Print the assignment of a parameter to a field. */
    private void printTestClassConstrAssign(Parameter param)
    {
        print("this.");
        print(param.ident);
        print(" = ");
        print(param.ident);
        println(";");
    }

    /** Print the doCall method. */
    private void printTestClassDoCall(String mname,
                                      Parameter [] params,
                                      Parameter receiver,
                                      boolean isInstance,
                                      boolean isConstructor)
    {
        newLine();
        println("protected void doCall() throws java.lang.Throwable {");
        indent();
        if (isInstance && !isConstructor) {
            print(receiver.ident);
            print(".");
            print(mname);
        } else if (isConstructor) {
            print("new ");
            print(typeDecl.ident());
        } else if (!isInstance) {
            print(typeDecl.ident());
            print(".");
            print(mname);
        }
        print("(");
        for (int i = 1; i <= params.length; i++) {
            print(params[i-1].ident);
            if (i < params.length) {
                print(", ");
            }
        }
        println(");");
        undent();
        println("}");
    }

    /** Print the doCall method. */
    private void printTestClassFailMessage(MethodInfo mdecl,
                                           String mname,
                                           Parameter [] params,
                                           Parameter receiver,
                                           boolean isInstance,
                                           boolean isConstructor)
    {
        newLine();
        println("protected java.lang.String failMessage");
        indent();
        print("(");
        print(PKG_JMLRAC);
        println("JMLAssertionError e$)");
        undent();
        println("{");
        indent();
	composeFailMessage(mdecl, mname, params, receiver,
                           isInstance, isConstructor);
	println("return msg;");
        undent();
        println("}");
    }

    /** return the package string, if present. */
    protected String getPackageString() {
        return packageString != null && packageString.length() > 0
            ? packageString + "." : "";
    }

    /** return the package string, if present. */
    protected String getTestedPackageString() {
        return testedPackageString != null && testedPackageString.length() > 0
            ? testedPackageString + "." : "";
    }

    /** Prints test$IfRACCompiled method. The purpose of this method is
     * to test to see if the code for the class under test, at least
     * has been compiled with the JML runtime assertion checker.  If
     * not, then no testing is going to happen. */
    private void printTestIsRACCompiled(JTypeDeclarationType cdecl) {
        String cname = cdecl.ident();
        String whatKind_space;
        if (cdecl instanceof JClassDeclarationType) {
            whatKind_space = "class ";
        } else {
            whatKind_space = "interface ";
        }
        newLine(); 
        print("/** Test to see if the code for ");
        print(whatKind_space);
        println(cname);
        println(" * has been compiled with runtime assertion checking" +
                " (i.e., by jmlc).");
        println(" * Code that is not compiled with jmlc would not make" +
                " an effective test,");
        println(" * since no assertion checking would be done." +
                " */");
        
        println("public void test$IsRACCompiled() {");
        indent();

        print(PKG_JUNIT + "Assert.assertTrue(\"code for ");
        print(whatKind_space);
        print(cname);
        println("\"");
        indent();
        indent();
        println("+ \" was not compiled with jmlc\"");
        println("+ \" so no assertions will be checked!\",");
        undent();
        println(PKG_JMLRAC + "JMLChecker.isRACCompiled(" + cname + ".class)");
        println(");");
        undent();

        undent();
        println("}");
    }

    /** Prints a package statement. */
    private void printPackageStatement() {
	if (packageString != null && packageString.length() > 0) {
            newLine();
            println("package " + packageString + ";");
        }
    }

    /** 
     * Returns a fixture variable name for the type named, 
     * <code>type</code>. The convention is to prefix the string
     * "v" to the type name.
     */
    protected static final String fixtureVariableName(CType type) {
        String name = type.toString();
        name = name.replace('[','$').replace(']', '_');
        return "v" + name.replace('.', '_');
    }

    // ----------------------------------------------------------------------
    // PRIVATE HELPER METHODS
    // ----------------------------------------------------------------------

    /** Returns <code>true</code> if the given type declaration is
     * testable using the JML/JUnit approach. All non-private classes
     * and interfaces are are testable. */
    private boolean isTestable(JTypeDeclarationType cdecl, String fileNameId) {
        return cdecl.ident().equals(fileNameId) &&
            !isPrivate(cdecl.modifiers());
    }

    /** Return true if the given method declaration can be tested.
     *  Static and non-static (instance) methods of public or package
     *  visibility and public constructors of non-abstract classes are
     *  testable. However particular, the main method and non-public
     *  constructors are not testable.
     */
    private boolean isTestable(JMethodDeclarationType mdecl) {
        long modifiers = mdecl.modifiers();
        
        // has the same or higher visibility than requested?
        if (!hasRequestedVisibility(modifiers)) {
            return false;
        }

        if (!requestedDeprecation && mdecl.isDeprecated()) {
            return false;
        }

        // only normal instance methods and constructors are testable.
        if (!(mdecl instanceof JMethodDeclaration
              || mdecl instanceof JConstructorDeclaration)
            || (mdecl instanceof JInitializerDeclaration)
            || (mdecl instanceof MJTopLevelMethodDeclaration)) {
            return false;
        }

        // only constructors for non-abstract classes are testable
        if (mdecl instanceof JConstructorDeclaration
            && isAbstract(typeDecl.modifiers())) {
            return false;
        }

        // we don't test "public static void main(String[])" method
        if (isPublic(modifiers) && isStatic(modifiers) &&
            mdecl.returnType().isVoid() &&  "main".equals(mdecl.ident()) &&
            mdecl.parameters().length == 1) {
            CType p0type = mdecl.parameters()[0].getType();
            if (p0type.isArrayType() &&
                ((CArrayType) p0type).getBaseType() == CStdType.String) {
                return false;
            }
        }

        return true;
    }

    /** Return true if the given inherited method can be tested.
     */
    private boolean isTestable(CMethod mdecl) {
        long modifiers = mdecl.modifiers();
        
        // has the same or higher visibility than requested?
        if (!hasRequestedVisibility(modifiers)) {
            return false;
        }

        if (!requestedDeprecation && mdecl.isDeprecated()) {
            return false;
        }

        return true;
    }

    /**
     * Returns true if the given modifiers include an access modifier
     * that has the same or widely visible than requested by the user.
     */
    private boolean hasRequestedVisibility(long modifiers) {
        /*switch (requestedVisibility) {
        case ACC_PUBLIC: 
            return isPublic(modifiers);
        case ACC_PROTECTED:
            return isPublic(modifiers) || isProtected(modifiers);
        default:
            return !isPrivate(modifiers);
	    }*/
	if(requestedVisibility == ACC_PUBLIC) {
            return isPublic(modifiers);	    
	} else if(requestedVisibility == ACC_PROTECTED) {
            return isPublic(modifiers) || isProtected(modifiers);
	} else {
            return !isPrivate(modifiers);
	}
    }

    /** Emit methods private to the generated class.
     */
    private void printPrivateHelperMethods() {
        printCheckHasData();
	printCharToString();
    }

    /** Emit the a private method to check that the result of a call
     * to get_receivers returned a non-null value (and note a test failure
     * otherwise).
     */
    private void printCheckHasData() {
        newLine();
        println("/** Check that the iterator is non-null and not empty. */");
        println("private void");
        println("check_has_data(" + PKG_STRATEGIES + "IndefiniteIterator iter,");
        println("               String call)");
        println("{");
        indent();
        println("if (iter == null) {");
        printlnIn(PKG_JUNIT + "Assert.fail(call + \" returned null\");");
        println("}");
        println("if (iter.atEnd()) {");
        printlnIn(PKG_JUNIT + "Assert.fail(call + \" returned an empty iterator\");");
        println("}");
        undent();
        println("}");
    }

    private void printCharToString() {
	newLine();
	println("/** Converts a char to a printable String for display */");
	println("public static String charToString(char c) {");
	println("    if (c == '\\n') {");
	println("        return \"NL\";");
	println("    } else if (c == '\\r') {");
	println("        return \"CR\";");
	println("    } else if (c == '\\t') {");
	println("        return \"TAB\";");
	println("    } else if (Character.isISOControl(c)) {");
	println("        int i = (int)c;");
	println("        return \"\\\\u\"");
	println("                + Character.forDigit((i/2048)%16,16)");
	println("                + Character.forDigit((i/256)%16,16)");
	println("                + Character.forDigit((i/16)%16,16)");
	println("                + Character.forDigit((i)%16,16);");
	println("    }");
        println("    return Character.toString(c);");
	println("}");
    }

    /** Generates code that composes a message about the method and parameters.
     * The code declares a String variable msg and the generated message is
     * contained in that variable in the generated code. */
    protected void composeFailMessage(MethodInfo mdecl,
                                      String mname,
                                      Parameter[] params,
                                      Parameter receiver,
                                      boolean isInstance,
                                      boolean isConstructor) {
        print("java.lang.String msg = \"\\n\\t");
        print(capitalize(methodKind(mdecl)));
        print(" '");
        print(mname);
        if (isConstructor && params.length == 0) {
            print("'");
        } else {
            print("' applied to");
        }
        println("\";");
        if (!isConstructor) {
            print("msg += \"\\n\\tReceiver");
            if (isInstance) {
                print(": \" + this.");
                print(receiver.ident);
                println(";");
            } else {
                println(" class " + typeDecl.ident() + "\";");
            }
        }
        for (int i = 1; i <= params.length; i++) {
           String v = params[i-1].ident;
           print("msg += \"\\n\\tArgument ");
           print(params[i-1].ident);
           print(": \" + this.");
           print(params[i-1].ident);
           println(";");
        }
    }
    
    /** 
     * Returns an array of original formal parameters, i.e., those not
     * generated by the type checker.
     *
     * <pre><jml>
     *  requires p != null;
     *  ensures \result != null;
     * </jml></pre>
     */
    private Parameter[] filterParameters(JFormalParameter[] p) {
        List l = new ArrayList(p.length);
        for (int i = 0; i < p.length; i++) {
            if (!p[i].isGenerated()) {
                l.add(new Parameter(p[i].getType(), 
                                    p[i].ident()));
            }
        }
        return (Parameter[]) l.toArray(new Parameter[0]);
    }


    /** 
     * Returns an array of formal parameters for given parameter
     * types, <code>p</code>.
     *
     * <pre><jml>
     *  requires p != null;
     *  ensures \result != null;
     * </jml></pre>
     */    
    private Parameter[] generateParameters(CSpecializedType[] p) {
        List l = new ArrayList(p.length);
        for (int i = 0; i < p.length; i++) {
            l.add(new Parameter(p[i].staticType(), "arg" + i));
        }
        return (Parameter[]) l.toArray(new Parameter[0]);
    }

    /** Return a capitalized version of the argument */
    protected String capitalize(/*@ non_null @*/ String s) {
        return Character.toUpperCase(s.charAt(0)) + s.substring(1);
    }

    /** 
     * Returns a unique test method name for the given method name.
     */
    private String testMethodName(String ident) {
        String name = nameGen.nameFor(ident);
        return TEST_METHOD_NAME_PREFIX + capitalize(name);
    }

    /** 
     * Returns a unique test class name for the given method name.
     */
    private String testMethodClassName(String ident) {
        return capitalize(testMethodName(ident));
    }

    /** Returns true if the argument has a static modifier. */
    protected static final boolean isStatic(long modifiers) {
        return Utils.hasFlag(modifiers, ACC_STATIC);
    }

    /** Returns true if the argument has a private modifier. */
    private static final boolean isPrivate(long modifiers) {
        return Utils.hasFlag(modifiers, ACC_PRIVATE);
    }

    /** Returns true if the argument has a public modifier. */
    private static final boolean isPublic(long modifiers) {
        return Utils.hasFlag(modifiers, ACC_PUBLIC);
    }

    /** Returns true if the argument has a protected modifier. */
    private static final boolean isProtected(long modifiers) {
        return Utils.hasFlag(modifiers, ACC_PROTECTED);
    }

    /** Returns true if the argument has a public modifier. */
    private static final boolean isAbstract(long modifiers) {
        return Utils.hasFlag(modifiers, ACC_ABSTRACT);
    }

    // ----------------------------------------------------------------------
    // PRIVATE INNER CLASSES
    // ----------------------------------------------------------------------

    /** A class for generating unique names for test methods. */
    private static class NameGenerator {
        String nameFor(String n) {
            if (names.containsKey(n)) {
                int cnt = ((Integer)names.get(n)).intValue();
                names.put(n, new Integer(++cnt));
                return n + "$" + cnt;
            } else {
                names.put(n, new Integer(0));
                return n;
            }
        }
        private Map names = new HashMap();

        void reset() {
            names = new HashMap();
        }
    }

    /** A simple data structure class for storing information about
     * a formal parameter. */
    private static class Parameter {
        Parameter(CType type, String ident) {
            this.ctype = type;
            this.type = type.toString();
            this.ident = ident;
        }

        /** The ctype of this parameter */
        CType ctype;
        /** The string name of this type */
        String type;

        /** The variable's name */
        String ident;

        String tvar() {
            if (tvar == null) {
                tvar = fixtureVariableName(ctype);
            }
            //@ assume tvar != null;
            return tvar;
        }

        /* Fixture variable name for the type ctype. */
        private String tvar;
    }

    /** Common information needed from JMethodDeclarationType and
     * CMethod in this class. */
    protected class MethodInfo {
        /** Initialize this object from the given CMethod */
        public MethodInfo(CMethod m) {
            retTyp = m.returnType();
            mods = m.modifiers();
            name = m.ident();
            CSpecializedType[] ts = m.parameters();
            params = new Parameter[ts.length];
            for (int i = 0; i < ts.length; i++) {
                params[i] = new Parameter(ts[i].staticType(),
                                          "arg" + (i+1));
            }
            isConstr = m.isConstructor();
            isDepr = m.isDeprecated();
        }

        /** Initialize this object from the given JMethodDeclarationType */
        public MethodInfo(JMethodDeclarationType mdecl) {
            retTyp = mdecl.returnType();
            mods = mdecl.modifiers();
            name = mdecl.ident();
            params = filterParameters(mdecl.parameters());
            isConstr = mdecl instanceof JConstructorDeclarationType;
            isDepr = mdecl.isDeprecated();
        }

        /** The method's return type */
        private CType retTyp;
        /** Return the modifiers of this method. */
        public final CType returnType() { return retTyp; }

        /** The method's modifiers */
        private long mods;
        /** Return the modifiers of this method. */
        public final long modifiers() { return mods; }

        /** The method's name */
        private String name;
        /** Return the name of this method. */
        public final String ident() { return name; }

        /** The method's parameters */
        private Parameter[] params;
        /** Return this method's parameters */
        public final Parameter[] parameters() {
            return params;
        }

        /** Is this a constructor? */
        private boolean isConstr;
        /** Is this a constructor? */
        public final boolean isConstructor() { return isConstr; }

        /** Is this deprecated? */
        private boolean isDepr;
        /** Is this method deprecated? */
        public final boolean isDeprecated() { return isDepr; }
    }

    /** A composite control abstraction useful in this class. */
    protected class MethodsIterator implements Iterator {
        /** The non-inherited methods. */
        private /*@ non_null @*/ Iterator methods;
        //@ in objectState; maps methods.objectState \into objectState;
        /** The inherited methods. */
        private /*@ non_null @*/ Iterator inheritedMethods;
        /*@       in objectState;
          @       maps inheritedMethods.objectState \into objectState;
          @*/

        /** Cached method to return next, null if none */
        private MethodInfo cache = null; //@ in objectState;

        //@ private represents moreElements <- cache != null;

        /** Initialize this methods iter. */
        //@ requires methods != null;
        //@ assignable objectState;
        public MethodsIterator(List methods, List inhMeths) {
            this.methods = methods.iterator();
            if (inhMeths != null) {
                this.inheritedMethods = inhMeths.iterator();
            } else {
                this.inheritedMethods = new ArrayList().iterator();
            }
            cacheNext();
        }

        /** Does the composition have more elements? */
        /*@ assignable objectState;
          @ ensures \result
          @      <==> methods.moreElements || inheritedMethods.moreElements;
          @*/
        private boolean compositeHasNext() {
            return methods.hasNext() || inheritedMethods.hasNext();
        }

        /** Return the next element in the composition */
        //@ assignable methods.objectState, inheritedMethods.objectState;
        private Object compositeNextMethod() {
            if (methods.hasNext()) {
                return methods.next();
            } else {
                return inheritedMethods.next();
            }
        }

        /** Advance the state of the iteration to point to the next
         * testable method declaration. */
        //@ assignable objectState;
        private void cacheNext() {
            Object o = null;
            JMethodDeclarationType mdecl = null;
            CMethod m = null;
            boolean testable = false;
            if (!compositeHasNext()) {
                cache = null;
                return;
            } else {
                do {
                    mdecl = null;
                    m = null;
                    o = compositeNextMethod();
                    if (o instanceof JMethodDeclarationType) {
                        mdecl = (JMethodDeclarationType) o;
                        testable = isTestable(mdecl);
                    } else if (o instanceof CMethod) {
                        m = (CMethod) o;
                        testable = isTestable(m);
                    } else {
                        // could be an initializer block or top level method
                        testable = false;
                    }
                } while (!testable && compositeHasNext());
            }
            /*@ assert testable
              @    || !(methods.moreElements || inheritedMethods.moreElements);
              @ hence_by (* logic *);
              @ assert testable
              @   ==> !(methods.moreElements || inheritedMethods.moreElements);
              @*/
            if (!testable) {
                cache = null;
            } else if (mdecl != null) {
                cache = new MethodInfo(mdecl);
            } else {
                //@ assume m != null;
                cache = new MethodInfo(m);
            }
        }

        public /*@ pure @*/ boolean hasNext() {
            return cache != null;
        }

        //@ also
        //@ assignable objectState;
        public Object next() {
            return nextMethod();
        }

        /** Return the next testable method */
        //@ assignable objectState;
        public MethodInfo nextMethod() {
            if (!hasNext()) {
                throw new NoSuchElementException();
            }
            MethodInfo ret = cache;
            cacheNext();
            return ret;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }


    /** Return a new methods iterator for the methods that should be checked. */
    //@ assignable \nothing;
    //@ ensures \fresh(\result);
    protected MethodsIterator methodsIter() {
        return new MethodsIterator(typeDecl.methods(),
                               includeInherited
                               ? typeDecl.getCClass().getAllInheritedMethods()
                               : null);
    }

    // ----------------------------------------------------------------------
    // HELPER METHODS FOR PRINTING
    // ----------------------------------------------------------------------

    /** Indents one TAB_SIZE. */
    protected final void indent() {
        writer.indent();
    }

    /** Und-ndents one TAB_SIZE. */
    protected final void undent() {
        writer.undent();
    }

    /** Prints the given object. */
    protected void print(Object s) {
	writer.print(s);
    }

    /** Prints the given string. */
    protected void print(String s) {
	writer.print(s);
    }

    /** Prints the given object with a trailing line separator. */
    protected void println(Object s) {
	writer.println(s);
    }

    /** Prints the given object with one TAB_SIZE indentation
     * and a trailing line separator. */
    protected void printlnIn(Object s) {
	writer.printlnIn(s);
    }

    /** Prints a line separator. */
    protected void newLine() {
	writer.newLine();
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /** The writer where the generated test oracle classes are to be
     * printed to.
     */
    private /*@ non_null @*/ FancyTabbedPrintWriter writer;

    /** The type declaration currently being translated. */
    protected JTypeDeclarationType typeDecl;

    /** The test class name currently being generated.
     *
     * <pre><jml>
     * invariant typeDecl != null ==> 
     *   testClassName.equals(typeDecl.ident() + TEST_CLASS_NAME_POSTFIX);
     * </jml></pre>
     */
    protected String testClassName;

    /** The name of the package the generated class is in. */
    protected String packageString = null;

    /** The name of the package the type being tested is in. */
    protected String testedPackageString = null;

    /** The set of fixture types. This is the set of all formal parameters
     * of the methods to be tested. 
     *
     * <pre><jml>
     * invariant fixtureTypes != null && 
     *   (\forall Object o; fixtureTypes.contains(o);
     *                      o != null && o instanceof CType);
     * </jml></pre>
     */
    protected Set fixtureTypes = new HashSet();

    /** A name generator to generate unique names for test methods. */
    private NameGenerator nameGen = new NameGenerator();

    /** The requested visibility of test methods. Test methods are
     * generated only for those methods that have the same or higher
     * visibility than the requested visibility.
     *
     * <pre><jml>
     * private invariant requestedVisibility == ACC_PUBLIC ||
     *                   requestedVisibility == ACC_PROTECTED ||
     *                   requestedVisibility == ACC_ALL;
     * </jml></pre>
     *
     * @see #hasRequestedVisibility(int)
     * @see #isTestable(JMethodDeclarationType)
     * @see #isTestable(CMethod)
     */
    private final long requestedVisibility;

    /** Tell whether tests should be generated for deprecated methods.
     * True means that tests for deprecated methods and constructors
     * should be generated (the default option is false).
     *
     * @see #isTestable(JMethodDeclarationType)
     * @see #isTestable(CMethod)
     */
    private final boolean requestedDeprecation;

    private final static long ACC_ALL = ACC_PUBLIC | ACC_PROTECTED;
    
    /**
     * Indicates whether tests are generated for all the inherited
     * methods too.
     */
    protected final boolean includeInherited;
    protected boolean useGenerator;
    protected String  generatorPackageName;
    protected /*@non_null*/ String  generatorClassName;
}
