/* $Id: TestDataClassGenerator.java,v 1.10 2006/08/01 23:23:00 chalin Exp $
 *
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlunit;

import org.multijava.mjc.*;
import org.multijava.util.compiler.*;
import org.multijava.util.classfile.*;

import java.util.Iterator;

/**
 * A class for generating JML/JUnit test data classes. A test data class
 * must be used to specify test data for a JML/JUnit test oracle class
 * generated by the class {@link TestClassGenerator}. 
 *
 * @author Yoonsik Cheon
 * @author Gary T. Leavens
 * @version $Revision: 1.10 $
 */
public class TestDataClassGenerator extends TestClassGenerator
    implements Constants {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a new test data class generator object.
     */
    public TestDataClassGenerator(JntOptions options) {
        super(options);
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------

    /** Prints import statements. */
    protected void printImportStatements(JCompilationUnit cunit) {
	super.printImportStatements(cunit);

        // for the -g option, need to import the generator class, 
	// if it's not in the current package
	if (useGenerator) {
	   if (generatorPackageName!=null && !generatorPackageName.equals(packageString)) {
              println("import "+generatorPackageName+"."+generatorClassName+";");
	   }
	}
    }

    /** Prints a javadoc comment for the current class. */
    protected void printClassJavadoc() {
        String name = typeDecl.ident();
        println("/** Supply test data for the JML and JUnit based testing of ");
        print(" *  ");
        print(name);
        println(".");
        println(" *");
        println(" *  <p>Test data is supplied by overriding methods in this class.  See");
        println(" *  the JML documentation and the comments below about how to do this.");
        println(" *");
        println(" *  <p>This class is also the place to override the <kbd>setUp()</kbd>");
        println(" *  and <kbd>tearDown()</kbd> methods if your testing needs some");
        println(" *  actions to be taken before and after each test is executed.");
        println(" *");
        println(" *  <p>This class is never rewritten by jmlunit.");
        println(" */");
    }

    /** 
     * Generates a test data class for the given class, <code>cdecl</code>. 
     * This method is overridden here to generate a data case class
     * instead of a test oracle class.
     */
    protected void generateTestClass(/*@ non_null @*/ JCompilationUnit cunit,
                                     /*@ non_null @*/ JTypeDeclarationType cdecl)
    {
        // initialize shared fields: class decl and name.
        this.typeDecl = cdecl;
        testClassName = cdecl.ident() + TEST_DATA_NAME_POSTFIX;

        // javadoc
	newLine();
        printClassJavadoc();

        // class header
        printClassHeader();
        indent();

        // boilerplate methods
        printConstructor();
        printSuiteFactoryMethods();

        // print template for setUp method
        initializeFixture();
        printBigComment();
        printDataProvisionMethods();

        undent();
        println("}"); // end of class
    }

    /** Prints the class header of test case class. */
    protected void printClassHeader() {
        // If the class is not public, then the code doesn't run
	println("public abstract class " + testClassName);
        indent();
        print("extends ");
        println(PKG_JUNIT + "TestCase");
        undent();
        println("{");
    }

    /** Print the factory method declarations. */
    protected void printSuiteFactoryMethods() {
        printOverallTestSuiteMethod();
        printEmptyTestSuiteForMethod();
    }

    /** Print the declaration of the overallTestSuite method. */
    private void printOverallTestSuiteMethod() {
        newLine();
        println("/** Return the overall test suite for accumulating tests; the");
        println(" * result will hold every test that will be run.  This factory");
        println(" * method can be altered to provide filtering of test suites, as");
        println(" * they are added to this overall test suite, based on various");
        println(" * criteria.  The test driver will first call the method");
        println(" * addTestSuite to add a test suite formed from custom programmed");
        println(" * test methods (named testX for some X), which you can add to");
        println(" * this class; this initial test suite will also include a method");
        println(" * to check that the code being tested was compiled with jmlc.");
        println(" * After that, for each method to be tested, a test suite");
        println(" * containing tests for that method will be added to this overall");
        println(" * test suite, using the addTest method.  Test suites added for a");
        println(" * method will have some subtype of TestSuite and that method's");
        println(" * name as their name. So, if you want to control the overall");
        println(" * suite of tests for testing some method, e.g., to limit the");
        println(" * number of tests for each method, return a special-purpose");
        println(" * subclass of {@link junit.framework.TestSuite} in which you override the");
        println(" * addTest method.");
        println(" * @see junit.framework.TestSuite");
        println(" */");
        println("//@ assignable objectState;");
        println("//@ ensures \\result != null;");
        println("public " + PKG_JUNIT + "TestSuite overallTestSuite() {");
        printlnIn("return new " + PKG_JUNIT + "TestSuite(\"Overall tests for "
                  + typeDecl.ident() + "\");");
        println("}");
    }

    /** Print the declaration of the emptyTestSuiteFor method. */
    private void printEmptyTestSuiteForMethod() {
        newLine();
        println("/** Return an empty test suite for accumulating tests for the");
        println(" * named method.  This factory method can be altered to provide");
        println(" * filtering or limiting of the tests for the named method, as");
        println(" * they are added to the test suite for this method.  The driver");
        println(" * will add individual tests using the addTest method.  So, if you");
        println(" * want to filter individual tests, return a subclass of TestSuite");
        println(" * in which you override the addTest method.");
        println(" * @param methodName The method the tests in this suite are for.");
        println(" * @see junit.framework.TestSuite");
        println(" * @see org.jmlspecs.jmlunit.strategies.LimitedTestSuite");
        println(" */");
        println("//@ assignable objectState;");
        println("//@ ensures \\result != null;");
        println("public " + PKG_JUNIT + "TestSuite emptyTestSuiteFor");
        printlnIn("(java.lang.String methodName)");
        println("{");
        printlnIn("return new " + PKG_JUNIT + "TestSuite(methodName);");
        println("}");
    }

    /** Print the big comment in the body explaining how to edit the
     * code to provide test data. */
    protected void printBigComment() {
        newLine();
        println("// TEST DATA SUPPLY SECTION");
        newLine();
        println("// You should edit the following code to supply test data.  In the");
        println("// skeleton originally supplied below, the jmlunit tool made a");
        println("// guess as to a minimal strategy for generating test data for");
        println("// each type of object used as a receiver, and each type used as");
        println("// an argument.  There is a library of strategies for generating");
        println("// test data in org.jmlspecs.jmlunit.strategies, which are used in");
        println("// the tool's guesses.  See the documentation for JML and in");
        println("// particular for the org.jmlspecs.jmlunit.strategies package for");
        println("// a general discussion of how to do this.  (This package's");
        println("// documentation is available through the JML.html file in the top");
        println("// of the JML release, and also in the package.html file that");
        println("// ships with the package.)");
        println("//");
        println("// In the code below, you can change the strategies from those");
        println("// that were guessed by the jmlunit tool, and you can also define");
        println("// new ones to suit your needs.  You can also delete any useless");
        println("// sample test data that has been generated for you to show you");
        println("// the pattern of how to add your own test data.  The only");
        println("// requirement is that you implement the methods below.");
        println("//");
        println("// If you change the type being tested in a way that introduces");
        println("// new types of arguments for some methods, then you will have to");
        println("// introduce (by hand) definitions that are similar to the ones");
        println("// below, because jmlunit never rewrites this file.");
    }

    protected void initializeFixture() {
        super.initializeFixture();
        // Add the class itself in if there are any instance methods
        CType thisClassType = typeDecl.getCClass().getType();
        TestClassGenerator.MethodsIterator mi = methodsIter();
        boolean hasInstanceMethod = false;
        while (mi.hasNext()) {
            MethodInfo mdecl = mi.nextMethod();
            if (!isStatic(mdecl.modifiers())
                && !(mdecl.isConstructor())) {
                hasInstanceMethod = true;
                break;
            }
        }
        if (hasInstanceMethod) {
            fixtureTypes.add(thisClassType);
        }
    }

    /** Prints declarations of the vTiter methods, as well
     * as sample strategy declarations for. */
    protected void printDataProvisionMethods() {
        for (Iterator i = fixtureTypes.iterator(); i.hasNext(); ) {
            CType type = (CType) i.next();
            printDataProvisionCode(type);
        }
    }

    /** Prints the declaration of the vTiter method as well as the
     * sample strategy for the given type. */
    //@ requires type != null;
    private void printDataProvisionCode(CType type) {
        String varName = fixtureVariableName(type);
        String typeName = type.toString();
        boolean isPrimitive = type.isPrimitive();
	boolean isImmutableType = isImmutableType(type, typeName);
        boolean hasCloneMethod = (type instanceof CArrayType)
            || (type.isReference() && hasCloneMethod(type.getCClass()));
        printIteratorProvisionMethod(varName, typeName, isPrimitive,
                                     isImmutableType, hasCloneMethod);
        printStrategyGuess(type, varName, typeName, isPrimitive,
                           isImmutableType, hasCloneMethod);
    }

    /** Print the declaration of the vTiter method for the given type. */
    private void printIteratorProvisionMethod
        (String varName, String typeName, boolean isPrimitive,
         boolean isImmutableType, boolean hasCloneMethod)
    {
        newLine();
        printProvideIteratorComment(typeName);
        print("protected ");
        print(PKG_STRATEGIES);
        String capTypeName;
        if (!isPrimitive) {
            capTypeName = typeName;
        } else {
            capTypeName = capitalize(typeName);
        }
        if (!isPrimitive) {
            print("Indefinite");
        } else {
            print(capTypeName);
        }
        println("Iterator");
        indent();
        print(varName);
        println("Iter");
        println("(java.lang.String methodName, int loopsThisSurrounds)");
        undent();
        println("{");
        indent();
        print("return ");
        print(varName);
        print("Strategy.");
        if (isPrimitive) {
            print(typeName);
            print("Iterator");
        } else {
            print("iterator");
        }
        println("();");
        undent();
        println("}");
    }

    private void printProvideIteratorComment(String typeName) {
        println("/** Return a new, freshly allocated indefinite iterator that");
        println(" * produces test data of type ");
        print(" * ");
        println(typeName);
        println(" * for testing the method named by the String methodName in");
        println(" * a loop that encloses loopsThisSurrounds many other loops.");
        println(" * @param methodName name of the method for which this");
        println(" *                      test data will be used.");
        println(" * @param loopsThisSurrounds number of loops that the test");
        println(" *                           contains inside this one.");
        println(" */");
        println("//@ requires methodName != null && loopsThisSurrounds >= 0;");
        println("//@ ensures \\fresh(\\result);");
    }

    /** Print the declaration of the default strategy for the given type. */
    private void printStrategyGuess
        (CType type, String varName, String typeName, boolean isPrimitive,
         boolean isImmutable, boolean hasCloneMethod)
    {
        newLine();
        println("/** The strategy for generating test data of type");
        print(" * ");
        print(typeName);
        println(". */");
        print("private ");
        print(PKG_STRATEGIES);
        if (isPrimitive) {
            print(capitalize(typeName));
        }
        println("StrategyType");
        indent();
        print(varName);
        println("Strategy");
        print("= new ");
        print(PKG_STRATEGIES);
        if (isPrimitive) {
            print(capitalize(typeName));
            print("Strategy()");
            if (type == CStdType.Boolean) {
                println(";");
            } else {
                newLine();
            }
        } else if (isImmutable) {
            if (typeName.equals("java.lang.String")) {
                println("StringStrategy()");
            } else {
                println("ImmutableObjectAbstractStrategy()");
            }
        } else if (hasCloneMethod) {
                println("CloneableObjectAbstractStrategy()");
        } else {
            // mutable objects, no public clone method
            if (typeName.equals("java.lang.Object")) {
                println("ObjectStrategy()");
            } else {
                println("NewObjectAbstractStrategy()");
            }
        }
        indent();
        if (type != CStdType.Boolean) {
            println("{");
            indent();
            if (isPrimitive || isImmutable || hasCloneMethod) {
                print("protected ");
                if (isPrimitive) {
                    print(typeName);
                } else {
                    print("java.lang.Object");
                }
                println("[] addData() {");
                indent();
	        if (useGenerator) {
		   print("return "+generatorClassName+".get");
		   if (isPrimitive)
		      print(capitalize(typeName));
		   else if (typeName.indexOf(".")>0) {
		      String name=typeName.substring(typeName.lastIndexOf(".")+1);
		      if (name.endsWith("[]")) {
			 name=name.substring(0, name.length()-2);
		      }
		      print(name);
		   }
		   else
		      print(typeName);
		   println("Array();");
		} else {
                   print("return new ");
                   print(typeName);
                   println("[] {");
                   indent();
                   println("// replace this comment with test data if desired");
                   undent();
                   println("};");
		}
                undent();
                println("}");

                if (!isPrimitive && !isImmutable) {
                    //@ assume hasCloneMethod;
                    newLine();
                    println("//@ also");
                    println("//@ requires o$ != null;");
                    println("protected Object cloneElement(java.lang.Object o$) {");
                    indent();
                    print(typeName);
                    println(" down$");
                    indent();
                    print("= (");
                    print(typeName);
                    println(") o$;");
                    undent();
                    println("return down$.clone();");
                    undent();
                    println("}");
                }
            } else {
                println("protected Object make(int n) {");
                indent();
                if (useGenerator) {
		   print("return "+generatorClassName+".get");
		   if (isPrimitive)
		      print(capitalize(typeName));
		   else if (typeName.indexOf(".")>0)
		      print(typeName.substring(typeName.lastIndexOf(".")+1));
		   else
		      print(typeName);
		   println("(n);");
                   undent();
                   println("}");
		} else {
                   println("switch (n) {");
                   println("// replace this comment with test data if desired");
                   print("default");  // split this so it formats (here :-) in emacs
                   println(":");
                   printlnIn("break;");
                   println("}");
                   println("throw new java.util.NoSuchElementException();");
                   undent();
                   println("}");
		}
            }
            undent();
            println("};");
        }
        undent();
        undent();
    }

    /** Prints the declaration of the size_X method. */
    //@ requires varName != null && typeName != null;
    private void print_size_X_MethodDecl(String varName, String typeName,
                                         boolean isImmutableType,
                                         boolean hasCloneMethod)
    {
        newLine();
        println("protected int size_" + varName);
        indent();
        println("(java.lang.String forMethodName)");
        undent();
        println("{");
        indent();
        if (isImmutableType || hasCloneMethod) {
            println("return " + varName + ".length;");
        } else {
            println("// change the following when you change the get_"
                    + varName + " method.");
            println("return 0;");
        }
        undent();
        println("}");
    }

    /** Prints the declaration of the get_X method. */
    //@ requires varName != null && typeName != null;
    private void print_get_X_MethodDecl(String varName, String typeName,
                                        String defaultValue,
                                        boolean isImmutableType,
                                        boolean hasCloneMethod) {
        newLine();
        println("protected " + typeName + " get_" + varName);
        indent();
        println("(java.lang.String forMethodName, int i)");
        undent();
        println("{");
        indent();
        String ithElem = varName + "[i]";
        if (isImmutableType) {
            println("return " + ithElem + ";");
        } else if (hasCloneMethod) {
            println("if (" + ithElem + " == null) {");
            printlnIn("return null;");
            println("} else {");
            printlnIn("return (" + typeName + ") " + ithElem + ".clone();");
            println("}");
        } else {
            println("switch (i) {");
            println("// case 0:");
            indent();
            println("// return " + defaultValue + ";");
            undent();
            println("default:");
            indent();
            println("break;");
            undent();
            println("}");
            println("throw new java.lang.IndexOutOfBoundsException");
            indent();
            String getXname = "get_" + varName;
            println("(\"" + getXname + " has a missing case.  \"");
            println(" + \"It is not defined for " + getXname + "\"");
            println(" + \"(\\\"\" + forMethodName + \"\\\", \" + i + \")\");");
            undent();
        }
        undent();
        println("}");
    }

    /** Is the given type an immutable type? */
    //@ requires (* typeName is the same as type.toString *);
    private boolean isImmutableType(/*@ non_null @*/ CType type,
                            /*@ non_null @*/ String typeName) {
        return
            (type instanceof CNumericType
             || type == CStdType.Boolean
             || (type instanceof CClassType
                 && (   
                     typeName.equals("java.lang.String")
                     || typeName.equals("java.lang.Boolean")
                     || typeName.equals("java.lang.Byte")
                     || typeName.equals("java.lang.Character")
                     || typeName.equals("java.lang.Class")
                     || typeName.equals("java.lang.Double")
                     || typeName.equals("java.lang.Float")
                     || typeName.equals("java.lang.Integer")
                     || typeName.equals("java.lang.Long")
                     || typeName.equals("java.lang.Short")
                     || typeName.equals("java.lang.Package")

                     || typeName.equals("java.math.BigDecimal")
                     || typeName.equals("java.math.BigInteger")

                     || typeName.equals("org.jmlspecs.models.JMLByte")
                     || typeName.equals("org.jmlspecs.models.JMLChar")
                     || typeName.equals("org.jmlspecs.lang.JMLDataGroup")
                     || typeName.equals("org.jmlspecs.models.JMLDouble")
                     || typeName.equals("org.jmlspecs.models.JMLFloat")
                     || typeName.equals("org.jmlspecs.models.JMLInfiniteInteger")
                     || typeName.equals("org.jmlspecs.models.JMLInteger")
                     || typeName.equals("org.jmlspecs.models.JMLLong")
                     || typeName.equals("org.jmlspecs.models.JMLMath")
                     || typeName.equals("org.jmlspecs.models.JMLObjectBag")
                     || typeName.equals("org.jmlspecs.models.JMLObjectSequence")
                     || typeName.equals("org.jmlspecs.models.JMLObjectSet")
                     || typeName.equals("org.jmlspecs.models.JMLObjectToObjectMap")
                     || typeName.equals("org.jmlspecs.models.JMLObjectToObjectRelation")
                     || typeName.equals("org.jmlspecs.models.JMLObjectToValueMap")
                     || typeName.equals("org.jmlspecs.models.JMLObjectToValueRelation")
                     || typeName.equals("org.jmlspecs.models.JMLShort")
                     || typeName.equals("org.jmlspecs.models.JMLString")
                     || typeName.equals("org.jmlspecs.models.JMLValueBag")
                     || typeName.equals("org.jmlspecs.models.JMLValueSequence")
                     || typeName.equals("org.jmlspecs.models.JMLValueSet")
                     || typeName.equals("org.jmlspecs.models.JMLValueToObjectMap")
                     || typeName.equals("org.jmlspecs.models.JMLValueToObjectRelation")
                     || typeName.equals("org.jmlspecs.models.JMLValueToValueMap")
                     || typeName.equals("org.jmlspecs.models.JMLValueToValueRelation")

                     || typeName.equals("org.jmlspecs.models.resolve.NaturalNumber")
                     || typeName.equals("org.jmlspecs.models.resolve.StringOfObject")

                     || typeName.equals("java.net.URL")
                     || typeName.equals("java.util.Locale")

                     )));
    }

    /** Compute a context for use in testing whether a method
     *  exists in the given class. */
    private CExpressionContextType expr_context(CClass selfClass) {
        org.multijava.mjc.Main m = new org.multijava.mjc.Main();
        CCompilationUnitContextType compUnit = 
	    m.createCompilationUnitContext( null, 
					    new CCompilationUnit("Unnamed") );
	CClassContextType clazz
            = compUnit.createClassContext( new CBadClass("Unnamed")) ;
        CMethod methodSig
            = new CBinaryMethod(selfClass,
                                new org.multijava.util.classfile.MethodInfo
                                ((short)ACC_PUBLIC, "test",
                                 "()Z", "()Z",new String[]{},
                                 CodeInfo.DUM_INFO,
                                 false, true),
                                clazz);
	CMethodContextType method = clazz.createMethodContext( methodSig );
	CFlowControlContextType methodBlock
            = method.createFlowControlContext(0,null);
	return methodBlock.createExpressionContext();
    }

    /** Does the given CClass have a clone method?
     *  (Thanks to Curtis Clifton for help with this.)
     */
    public boolean hasCloneMethod(CClass selfClass) {
        CMethod result;
        try {
            result = selfClass.lookupMethod("clone", new CType[] {},
					    CClassType.EMPTY,
                                            expr_context(selfClass));
        } catch (UnpositionedError e) {
            return false;
        } catch (NullPointerException e) {
            // !FIXME! should be a better way to handle this!
            // The exception comes from isDefinedInside in CClass
            // when the type doesn't have a clone method.
            return false;
        }
        return result != null;
    }

    /** Returns the default value of the type <code>type</code>. I.e.,
     * 0 for numerical types, false for boolean, empty array for array
     * types, and null for other types. */
    public static String defaultValue(/*@ non_null @*/ CType type) {
	if (type instanceof CNumericType) {
	    return "0";
        }
	else if (type == CStdType.Boolean) {
	    return "false";
        }
	else if (type instanceof CArrayType) {
            CArrayType atype = (CArrayType) type;
            String res = "new " + atype.getBaseType().toString() + "[0]";
            for (int j = 1; j < atype.getArrayBound(); j++) {
                res += "[]";
            }
            return res;
        }
        else {
	    return "null";
        }
    }

    // ----------------------------------------------------------------------
    // HELPER METHODS FOR PRINTING
    // ----------------------------------------------------------------------

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------
}
