/* 
 * Copyright (C) 2000-2001 Virginia Tech
 *
 * This file is part of JML
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: FactoryPrinter.java,v 1.5 2003/09/02 19:27:23 cclifton Exp $
 */

package org.jmlspecs.racwrap;

import java.io.*;
import java.util.*;

import org.multijava.mjc.*;
import org.jmlspecs.jmlrac.*;
import org.jmlspecs.checker.*;

/**
*   The InterfacePrinter is an object that is used to output the
*   interface definition of a class to a file, given an abstract syntax tree.
*   This class extends JMLAbstractVisitor, so we only need to implement
*   the parts that we need.
**/
public class FactoryPrinter {

    public PrintStream out = null;

    /**
    *   Constructor. Initialize a new factory printer.
    *   @param out the PrintStream object to output the result to.
    **/
    //@requires out != null;
    public FactoryPrinter(PrintStream out) {
        this.out = out;
    }

    /**
    *   Parameterless constructor. Defaults to having out as System.out
    */
    public FactoryPrinter() {
        this.out = System.out;
    }

    /**
    *   Prints the factory methods for a class. This generates
    *   a $chx_Statics_xxx class with a factory method for every constructor.
    *   Also, this wraps the static methods of the original class.
    *   This is limited to:
    *       files with only one toplevel class
    */
    //@ requires compileUnit.typeDeclarations().length > 0;
    public void print( JmlCompilationUnit compileUnit) {

        //First, check if this contains only 1 toplevel class declaration:
        JTypeDeclarationType[] classes = compileUnit.typeDeclarations();

        this.out.println("//-----");
        this.out.println("//Factory file generated for:");
        this.out.println("//" + compileUnit.fileNameIdent());
        this.out.println("//-----");

        //Output package name
        if(!compileUnit.packageNameAsString().equals("")) {
            
            String packageName;

            //Format the packagename using dot separators instead of "/"
            packageName = compileUnit.packageNameAsString().replace('/','.');
            packageName = packageName.substring(0, packageName.length() -1);

            this.out.println();
            this.out.println("package " + packageName + ";");
        }

        
        this.out.println();
        //output every imported class
        JClassOrGFImportType[] importClasses = compileUnit.importedClasses();
        for(int i = 0; i < importClasses.length; i++) {
            JClassOrGFImportType importClass = importClasses[i];
            this.out.print("import ");
            this.out.println(importClass.getName().replace('/','.') + ";");
        }

        //output every imported package
        JPackageImportType[] importPackages = compileUnit.importedPackages();
        for(int i = 0; i < importPackages.length; i++) {
            JPackageImportType importPackage = importPackages[i];
            this.out.print("import ");
            this.out.println(importPackage.getName().replace('/','.') + ".*;");
        }

        //output our imported package
        this.out.println("import org.jmlspecs.racwrap.runner.*;");


        for(int i = 0; i < classes.length; i++) {
            JmlTypeDeclaration clazz = (JmlTypeDeclaration) classes[i]; 
            if(clazz.isInterface()) {
                System.out.println( clazz.ident() + " not a class, ignoring...");
            } else {
                print_class(clazz);
            }
        }
    }

    public void print_class(JmlTypeDeclaration clazz) {

        this.out.println();
        //Output class name
        if(clazz.hasFlag(clazz.modifiers(), clazz.ACC_PUBLIC)) {
            this.out.print("public ");
        }

        this.out.println("class $chx_Statics_" + clazz.ident() + " {");

        this.out.println();
        this.out.println("public static Node access = null;");
       
        //Wrap every static method, every constructor
        int constructorCtr = 0;
        ArrayList methods = clazz.methods();
        JmlMethodDeclaration method;
        for(int i = 0; i < methods.size(); i++) {
            method = (JmlMethodDeclaration) methods.get(i);
            if(method.isStatic()) {
            /*
                JFormalParameter[] params = method.parameters();
                CType type = method.returnType();
                this.out.print("    public static ");
                this.out.print(type.toString() + " ");
                this.out.print(method.ident() + "(");
                Util.printParams(out, params);
                this.out.println(");");
            */
            } else if(method.isConstructor()) {
                constructorCtr++;
                if(! clazz.hasFlag(clazz.modifiers(), clazz.ACC_ABSTRACT)) {
                    printConstructor(clazz, method);
                }
            }
        }//end of outputting every non-static method

        if(constructorCtr == 0) {
            printConstructor(clazz, clazz.getDefaultConstructor());
        }
        this.out.println("} //end of statics");
    }


    /**
    *   This prints out the factory method that corresponds to a
    *   constructor in the original object.
    */
    public void printConstructor(JmlTypeDeclaration clazz,
                        JMethodDeclarationType method) {

        JFormalParameter[] params = method.parameters();

        this.out.print("public static " + method.ident() + " newObject(");
        Util.printParams(out, params);
        this.out.println(") {");
        this.out.print("    ");
        this.out.println("if(access == null || !access.isWrap()) {");
        this.out.print("        ");
        this.out.println("$chx_Orig_" + clazz.ident() + " result = null;");
        this.out.print("        result = " + "new $chx_Orig_");
        this.out.print(clazz.ident() + "(");

        for(int j = 0; j < params.length; j ++) {
            this.out.print(params[j].ident());
            if(j < params.length - 1) {
                this.out.print(",");
            }
        }

        this.out.println(");");
        this.out.print("        ");
        this.out.println("result._chx_this = result;");
        this.out.print("        ");
        this.out.println("return result;");

        this.out.println("    } else {");
        this.out.print("        ");
        this.out.println("$chx_Wrap_" + clazz.ident() + " result;");
        this.out.print("result = new $chx_Wrap_" + clazz.ident() + "(");
        for(int j = 0; j < params.length; j ++) {
            this.out.print(params[j].ident());
            if(j < params.length - 1) {
                this.out.print(",");
            }
        }
        this.out.println(");");
        this.out.println("result.access = access;");
        this.out.print("        return result;");
        this.out.println("    }");
        this.out.println("}");
    }
}
