/* 
 * 
 * Copyright (C) 2000-2001 Virginia Tech
 *
 * This file is part of JML
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: InterfacePrinter.java,v 1.5 2003/11/22 06:59:15 leavens Exp $
 */

package org.jmlspecs.racwrap;

import java.io.*;
import java.util.*;

import org.multijava.mjc.*;
import org.jmlspecs.jmlrac.*;
import org.jmlspecs.checker.*;

/**
*   The InterfacePrinter is an object that is used to output the
*   interface definition of a class to a file, given an abstract syntax tree.
*   This class extends JmlAbstractVisitor, so we only need to implement
*   the parts that we need.
**/
public class InterfacePrinter {

    public PrintStream out = null;

    /**
    *   Constructor. Initialize a new Interface printer.
    *   @param out the File object to output the result to.
    **/
    //@requires out != null;
    public InterfacePrinter(PrintStream out) {
        this.out = out;
    }

    /**
    *   Parameterless constructor. Defaults to having out as System.out
    */
    public InterfacePrinter() {
        this.out = System.out;
    }

    /**
    *   Prints the equivalent interface of a class.
    *   This is limited to:
    *       files with only one toplevel class
    */
    //@ requires compileUnit.typeDeclarations().length > 0;
    public void print( JmlCompilationUnit compileUnit) {

        JTypeDeclarationType[] classes = compileUnit.typeDeclarations();

/*
        if(classes.length > 1) {
            //FIXME handle this case...
            System.out.println( "Cannot handle file with more than one class." 
                                + compileUnit.fileNameIdent());
        
            return;
        }
*/


        this.out.println("//-----");
        this.out.println("//Interface file generated for:");
        this.out.println("//" + compileUnit.fileNameIdent());
        this.out.println("//-----");

        //Output package name
        if(!compileUnit.packageNameAsString().equals("")) {
            
            String packageName;

            //Format the packagename using dot separators instead of "/"
            packageName = compileUnit.packageNameAsString().replace('/','.');
            packageName = packageName.substring(0, packageName.length() -1);

            this.out.println();
            this.out.println("package " + packageName + ";");
        }

        
        this.out.println();
        //output every imported class
        JClassOrGFImportType[] importClasses = compileUnit.importedClasses();
        for(int i = 0; i < importClasses.length; i++) {
            JClassOrGFImportType importClass = importClasses[i];
            this.out.print("import ");
            this.out.println(importClass.getName().replace('/','.') + ";");
        }

        //output every imported package
        JPackageImportType[] importPackages = compileUnit.importedPackages();
        for(int i = 0; i < importPackages.length; i++) {
            JPackageImportType importPackage = importPackages[i];
            this.out.print("import ");
            this.out.println(importPackage.getName().replace('/','.') + ".*;");
        }


        for(int i = 0; i < classes.length; i++) {
            JmlTypeDeclaration clazz = (JmlTypeDeclaration) classes[i]; 
            if(clazz.isInterface()) {
                //TODO: actually print the interface as well.
                System.out.println( clazz.ident() + " Not a class, ignoring...");
            } else {
                print_class(clazz);
            }
        }
    }

    private void print_class(JmlTypeDeclaration clazz) {
        this.out.println();
        //Output class name
        if(clazz.hasFlag(clazz.modifiers(), clazz.ACC_PUBLIC)) {
            this.out.print("public ");
        }
        this.out.print("interface " + clazz.ident());

        //if the original class has a superclass, extend that...
        String superName = ((JmlClassDeclaration) clazz).superName();
        if( superName != null 
            && !superName.equals("Object")
            && !superName.equals("java/lang/Object")) {
            this.out.print(" extends " + superName.replace('/','.'));
        }
        this.out.println(" { ");
        
        //Put every non-static, non-constructor  method.
        ArrayList methods = clazz.methods();
        for(int i = 0; i < methods.size(); i++) {
            JmlMethodDeclaration method;
            method = (JmlMethodDeclaration) methods.get(i);
            if(!(method.isStatic() || method.isConstructor())) {
                CType type = method.returnType();
                this.out.print("    public ");
                this.out.print(type.toString() + " ");
                this.out.print(method.ident() + "(");
                JFormalParameter[] params = method.parameters();
                Util.printParams(out, params);
                this.out.println(");");
            }//end of if !method.isStatic
        }//end of outputting every non-static method


        //For every public or protected
        //data member, add getter and setter methods.
        JFieldDeclarationType[] fields = clazz.fields();

        for(int i = 0; i < fields.length; i++) {

            if( fields[i].getField().isStatic() 
                || fields[i].getField().isPrivate()
                || fields[i].getField().hasFlag(fields[i].modifiers(), clazz.ACC_MODEL)
            ) {
                continue; //<-- just to make the nesting not so deep.
            }
            this.out.println();
            this.out.print("    ");
            this.out.print("public " + fields[i].getType().toString() + " ");
            this.out.println("_chx_get_" + fields[i].ident() + "();");

            this.out.print("    ");
            this.out.print("public void _chx_set_" + fields[i].ident());
            this.out.println("( " + fields[i].getType().toString() + " obj);");
        }
        
        this.out.println("} //End of interface");
    }

}
