/* 
 * Copyright (C) 2003 Virginia Tech 
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: Main.java,v 1.7 2004/07/30 15:07:10 wdietl Exp $
 */

package org.jmlspecs.racwrap;

import org.multijava.mjc.*;
import org.multijava.util.Utils;
import org.jmlspecs.checker.*;
import org.jmlspecs.jmlrac.*;

import java.io.*;
import java.util.*;

/**
This is the Main class for generating assertion code from JML specs
that goes into wrappers, rather that inline. 

There are two tasks that need to be done here. One is to compile the
original code normally, as if not processed by JML. The other is
to compile the wrapper generated code.
*/
public class Main extends org.jmlspecs.jmlrac.Main {

    //-----
    //  DATA MEMBERS
    //----

    private String[] args = null;

    public void setArgs(String[] args) {
        this.args = args;
    }

    /** these are the files that the wrapper version generated. */
    private java.util.ArrayList wrapFiles = new ArrayList();

    /** maps directories to classes in that directory for which racwrap has
        generated files. */
    private Map wrapMap = new HashMap();

    /** these are the original source files. */
    private java.util.List origFiles = new ArrayList();
    
    //-----
    //  CONSTRUCTORS
    //-----

    /**
    *   The the constructor for the wrapper based runtime assertion check 
    *   generator.
    **/
    public Main() {
        super();
    }

    //-----
    //  ENTRY POINTS
    //-----

    /**
    *   The main entry point when starting this program from the command line.
    *   @param  args    The command line arguments.
    */
    public static void main(String args[]) {
        boolean success;
        success = compile(args);
        System.exit(success ? 0 : 1);
    }

    /**
    *   Entry point for other programs wishing to use the wrapper generator.
    *   @param args The command line arguments.
    *   @return true is the compilation was successful
    **/
    public static boolean compile(String[] args) {
        Main compiler = new Main();
        compiler.setArgs(args); 
        boolean success = compiler.run(args);

        if(success) {
            Iterator i = compiler.wrapMap.entrySet().iterator();
            while(i.hasNext()) {
                Map.Entry e = (Map.Entry) i.next();
                String dir = (String) e.getKey();
                List names = (List) e.getValue();
                Iterator j = names.iterator();
                while(j.hasNext()) {
                    String name = (String) j.next();

                    File source = new File(dir, name + ".class");
                    File target = new File(dir, name + ".iface.chx"); 
                    success = success && source.renameTo(target);
    
                    source = new File(dir, "$chx_Statics_" + name + ".class");
                    target = new File(dir, name + ".statics.chx"); 
                    success = success && source.renameTo(target);

                    source = new File(dir, "$chx_Orig_" + name + ".class");
                    target = new File(dir, name + ".shadow"); 
                    success = success && source.renameTo(target);

                    source = new File(dir, "$chx_Wrap_" + name + ".class");
                    target = new File(dir, name + ".wrap.chx"); 
                    success = success && source.renameTo(target);
                 }
            }

            if(success) {
                String[] files = new String[compiler.origFiles.size()];
                org.multijava.mjc.Main.compile(
                    (String[]) compiler.origFiles.toArray(files));
            }
        }

        return success;
    }

    //-----
    //  TASK SEQUENCING
    //-----
    
    /**
    *   This method uses the type of oldTask and the command line options
    *   to sequence tasks. The first task is created by the createFirstTask
    *   method of the parent.
    *   @param oldTask the previous task.
    *   @return The next task that needs to be done
    *
    **/
    protected Task createTaskAfter(Task oldTask) {
        //We preempt the sequencing of the parent method
        //once the files are typechecked.

        //if(oldTask instanceof PrintWrapperTask) {
        //    return null;
        //} else 
        if(oldTask instanceof JmlTypecheckTask) {
            if(isSecondRound) {
                return ((RacOptions)options).nowrite() ? null :
                new TranslateMJTask(((TypecheckTask) oldTask).trees(),
                                            oldTask.sequenceID());    
            } else {
                return new CollateFilesTask(((JmlTypecheckTask) 
                            oldTask).trees(), oldTask.sequenceID()); 
            }
        } else if(oldTask instanceof CollateFilesTask) {
            return new PrintInterfaceTask(((CollateFilesTask) oldTask).trees(),
                            oldTask.sequenceID()); 
         } else if(oldTask instanceof PrintInterfaceTask) {
            return new PrintFactoryTask(((PrintInterfaceTask) oldTask).trees(),
                            oldTask.sequenceID());
        } else if(oldTask instanceof PrintFactoryTask) {
            //System.out.println("Printing orig:");
            return new PrintOrigTask(((PrintFactoryTask) oldTask).trees(),
                            oldTask.sequenceID());
         } else if(oldTask instanceof PrintOrigTask) {
            //JmlGenerateAssertionTask is an inner class of ..jmlrac.Main
            JmlRacGenerator.checking_mode = JmlRacGenerator.WRAPPER;
            return new JmlGenerateAssertionTask(
                            ((PrintOrigTask) oldTask).trees(),
                            oldTask.sequenceID());
        } else if(oldTask instanceof JmlGenerateAssertionTask) {
            return new PrintWrapperTask(
                            ((JmlGenerateAssertionTask)
                            oldTask).trees(),
                            oldTask.sequenceID());
        } else if(oldTask instanceof PrintWrapperTask) {
            prepareSecondRound();
            return firstTask(wrapFiles);
        } else {
            return super.createTaskAfter(oldTask);
        }
    }

    //-----
    // NESTED CLASSES FOR TASK SCHEDULING
    //----

    /**
    * CollateFilesTask is used to put all the source files into a list.
    **/
    public class CollateFilesTask extends TreeProcessingTask {

        public CollateFilesTask( CompilerPassEnterable[] trees,
            Object sequenceID) {

            super( PRI_PRETTY_PRINT, sequenceID,
                    trees, MjcMessages.PRINT_TIME);
        }

        /**
        * collate the files to compile into origFiles
        **/
        public void processTree(CompilerPassEnterable tree) {

            if(!(tree instanceof JmlCompilationUnit) || (tree == null)) {
                return;
            }

            PrintStream p = null;
            JmlCompilationUnit cunit = (JmlCompilationUnit) tree;
            origFiles.add(cunit.file().getAbsolutePath());
        }

    } //end of inner class CollateFilesTask 

    /**
    * PrintInterfaceTask is used for outputting the interface for the
    * input file.
    **/
    public class PrintInterfaceTask extends TreeProcessingTask {

        public PrintInterfaceTask( CompilerPassEnterable[] trees,
            Object sequenceID) {

            super( PRI_PRETTY_PRINT, sequenceID,
                    trees, MjcMessages.PRINT_TIME);
        }

        /**
        *   This will create the temporary file for the interface, and
        *   call on InterfacePrinter to print the interface file.
        **/
        public void processTree(CompilerPassEnterable tree) {

            if(!(tree instanceof JmlCompilationUnit) || (tree == null)) {
                return;
            }

            PrintStream p = null;
            JmlCompilationUnit cunit = (JmlCompilationUnit) tree;

            p = createTempFile(cunit, cunit.fileNameIdent() + ".java");

            //InterfacePrinter printer = new InterfacePrinter(System.out);
            InterfacePrinter printer = new InterfacePrinter(p);
            
            printer.print((JmlCompilationUnit) tree);
        }

    } //end of inner class PrintInterfaceTask 


    /**
    * PrintFactoryTask actually also prints the static members of the class
    */
    public class PrintFactoryTask extends TreeProcessingTask {

        /**
        *   Just call the superclass's constructor.
        */
        public PrintFactoryTask( CompilerPassEnterable[] trees,
            Object sequenceID) {

            super( PRI_PRETTY_PRINT, sequenceID,
                    trees, MjcMessages.PRINT_TIME);
        }

        /**
        *   This will call on FactoryPrinter to print the tree.
        **/
        public void processTree(CompilerPassEnterable tree) {

            if(!(tree instanceof JmlCompilationUnit) || (tree == null)) {
                return;
            }

            PrintStream p = null;
            JmlCompilationUnit cunit = (JmlCompilationUnit) tree;

            String filename = "$chx_Statics_";
            filename = filename + cunit.fileNameIdent() + ".java";
            p = createTempFile(cunit, filename);

            FactoryPrinter printer = new FactoryPrinter(p);
            
            printer.print((JmlCompilationUnit) tree);
        }
    }

    /**
    * PrintOrigTask is used to print the modified original class
    */
    public class PrintOrigTask extends TreeProcessingTask {

        /**
        *   Just call the superclass's constructor.
        */
        public PrintOrigTask( CompilerPassEnterable[] trees,
            Object sequenceID) {

            super( PRI_PRETTY_PRINT, sequenceID,
                    trees, MjcMessages.PRINT_TIME);
        }

        /**
        *   This will call on OrigPrettyPrinter to print the tree.
        **/
        public void processTree(CompilerPassEnterable tree) {
            if(tree == null || !(tree instanceof JmlCompilationUnit)) {
                return;
            }
            //System.out.println("Printing tree:");
            PrintStream p = null;
            JmlCompilationUnit cunit = (JmlCompilationUnit) tree;

            JmlTypeDeclaration clazz;
            clazz = (JmlTypeDeclaration) cunit.typeDeclarations()[0];
            RacPrettyPrinter printer;
            PrintWriter writer;
            
            if(clazz.isInterface()) {
                //Do nothing.
                //TODO: The compilation unit might
                //define other package-level classes as well.
                //We are currently ignoring these other classes.
            } else {

                String filename = "$chx_Orig_";
                filename = filename + cunit.fileNameIdent() + ".java";
                p = createTempFile(cunit, filename);
            
                writer = new PrintWriter(p);
                printer = new OrigPrettyPrinter(writer, new JmlModifier());
                tree.accept(printer);
                printer.close();
            }
        }
    }

    /**
    * PrintWrapperTask is used to print the wrappers class
    */
    public class PrintWrapperTask extends TreeProcessingTask {

        /**
        *   Just call the superclass's constructor.
        */
        public PrintWrapperTask( CompilerPassEnterable[] trees,
            Object sequenceID) {

            super( PRI_PRETTY_PRINT, sequenceID,
                    trees, MjcMessages.PRINT_TIME);
        }

        /**
        *   This will call on WrapperPrettyPrinter to print the tree.
        **/
        public void processTree(CompilerPassEnterable tree) {
            if(tree == null || !(tree instanceof JmlCompilationUnit)) {
                return;
            }
            //System.out.println("Printing tree:");
            PrintStream p = null;
            JmlCompilationUnit cunit = (JmlCompilationUnit) tree;

            JmlTypeDeclaration clazz;
            clazz = (JmlTypeDeclaration) cunit.typeDeclarations()[0];
            RacPrettyPrinter printer;
            PrintWriter writer;
            
            if(clazz.isInterface()) {
                p = createTempFile(cunit, cunit.fileNameIdent() + ".java");
                writer = new PrintWriter(p);
                printer = new RacPrettyPrinter(writer, new JmlModifier());
                tree.accept(printer);
                printer.close();
            } else {

                String filename = "$chx_Wrap_";
                filename = filename + cunit.fileNameIdent() + ".java";
                p = createTempFile(cunit, filename);

                //add this class to wrapMap for post-proecessing.
                String dir = cunit.file().getParent();
                List classes = (List) wrapMap.get(dir);
                if(classes == null) {
                    classes = new ArrayList();
                }
                classes.add(cunit.fileNameIdent());
                wrapMap.put(dir, classes); 
            
                writer = new PrintWriter(p);
                printer = new WrapperPrettyPrinter(writer, new JmlModifier());
                tree.accept(printer);
                printer.close();
            }
        }
    }



    //-----
    //  Utility methods.
    //-----
    public PrintStream createTempFile(JmlCompilationUnit cunit, String filename) {
        String dirname = System.getProperty("java.io.tmpdir");
        dirname = dirname + "/racwrap/" + cunit.packageNameAsString();

        //System.out.println("creating:" + dirname + filename);
        PrintStream p = null;
        try {
            boolean success = false;
            File tmpDir = new File(dirname);
	    tmpDir.deleteOnExit();
            success = tmpDir.mkdirs();

            File tmpFile = new File(tmpDir, filename);
            success = tmpFile.createNewFile();
            //System.out.println(tmpFile);
            tmpFile.deleteOnExit();
            p = new PrintStream(new FileOutputStream(tmpFile));
            String orig_dir = cunit.file().getParent();
            File pretend = new File(orig_dir, filename);
            getFileMap().put(tmpFile, pretend);
            wrapFiles.add(tmpFile.getAbsolutePath());
            
        } catch (Exception e) {
            String errMsg = "Cannot create temp file: ";
            errMsg = errMsg + filename + "\n" + e.toString();
            System.out.println(errMsg);
            System.out.println("output will go to System.out");
            p = System.out;
        }

        return p;
    }
}
