/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Disassembler.java,v 1.17 2007/12/20 12:25:26 chalin Exp $
 */

package org.multijava.dis;

import java.io.*;
import java.util.*;

import org.multijava.util.classfile.*;
import org.multijava.util.*;
import org.multijava.util.compiler.*;
import org.multijava.mjc.CUniverseTypeAnnotation;
import org.multijava.mjc.CUniverseMethodAnnotation;

/**
 * This class prints the classfile in ksm syntaxt
 */
public class Disassembler 
    implements org.multijava.util.classfile.Constants, Constants 
{

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Create a disassembler object from a classinfo
     */
    //x@ modifies classInfo,options,this.*;
    protected Disassembler( ClassInfo classInfo, DisOptions options ) {
	this.classInfo = classInfo;
	this.options = options;
    }

    /**
     * Disassembles a class file.
     */
    public static void disassemble( String sourceFile, 
				    String destination, 
				    DisOptions options ) 
	throws UnpositionedError {
        disassemble(sourceFile, destination, options, helper);
    }

    /**
     * Disassembles a class file by using the given helper.
     */
    protected static void disassemble( String sourceFile, 
				       String destination, 
				       DisOptions options,
				       DisassemblerHelper helper)
	throws UnpositionedError
    {
	ClassInfo		classInfo;

	if (sourceFile.endsWith(helper.inputExtension())) {
	    try {
		classInfo = helper.createClassInfo(
		    new DataInputStream(
			new BufferedInputStream(
			    new FileInputStream(sourceFile), 2048)));
	    } catch (Exception e) {
		throw new UnpositionedError(DisMessages.FILE_READ_PROBLEM2, 
					    sourceFile, e);
	    }
	} else {
	    try {
		classInfo = 
		    helper.createClassInfo(sourceFile.replace('.', '/'));
	    } catch (ClassFileReadException e) {
		throw new UnpositionedError(DisMessages.FILE_READ_PROBLEM,
					    e.getMessage());
	    }
	}

	if (classInfo == null) {
	    throw new UnpositionedError(DisMessages.CLASS_NOT_FOUND, 
					new Object[] { sourceFile });
	}

	writeAssemblerFile(classInfo, destination, options, helper);
    }

    /**
     * Creates a class file from class info
     */
    private static void writeAssemblerFile( ClassInfo classInfo, 
					    String destination, 
					    DisOptions options,
                                            DisassemblerHelper helper )
	throws UnpositionedError
    {
	String[] classPath = Utils.splitQualifiedName(classInfo.getName());

	if (classPath[0] != null) { // the class is part of a package
	    String	classDir =
		classPath[0].replace('.', File.separatorChar)
		.replace('/', File.separatorChar);

	    if (destination != null) {
		destination += File.separator + classDir;
	    }
	}

 	if (!options.stdout() && options.destination() != null) {
	    // check that destination exists or else create it
	    File destDir = new File(options.destination());

	    if (! destDir.exists()) {
		destDir.mkdirs();
	    }

	    if (! destDir.isDirectory()) {
		throw new UnpositionedError(CompilerMessages.IO_EXCEPTION, 
					    new Object[] {options.destination()});
	    }
	}

	File outputFile = null;
	try {
	    PrintWriter	out;

	    if (options.stdout()) {
		outputFile = null;
		out = new PrintWriter(System.out);
	    } else {
		outputFile = new File(destination, classPath[1] + 
                                      helper.outputExtension());
		out = new PrintWriter(
		    new BufferedWriter(new FileWriter(outputFile)));
	    }

	    Disassembler dis = helper.createDisassembler( classInfo, options );
	    dis.writeClass(new IndentingWriter(out) );

	    out.flush();
	    out.close();
	} catch (java.io.IOException e) {
	    throw new UnpositionedError(CompilerMessages.IO_EXCEPTION, 
					new Object[]{ outputFile.getPath(), 
						      e.getMessage() });
	}
    }

    /** 
     * Prints the modifiers of the given class info. Factored out to
     * allow subclasses to print non-Java modifiers, e.g., those of
     * JML.
     */
    protected void writeModifiers(IndentingWriter out, ClassInfo info) {
        writeModifiers(out, info.getModifiers() & ~ACC_SUPER);
    }

    /** 
     * Prints the modifiers of the given field info. Factored out to
     * allow subclasses to print non-Java modifiers, e.g., those of
     * JML.
     */
    protected void writeModifiers(IndentingWriter out, FieldInfo info) {
        writeModifiers(out, info.getModifiers());
    }

    /** 
     * Prints the modifiers of the given method info. Factored out to
     * allow subclasses to print non-Java modifiers, e.g., those of
     * JML.
     */
    protected void writeModifiers(IndentingWriter out, MethodInfo info) {
        writeModifiers(out, info.getModifiers());
    }

    /** 
     * Prints the modifiers of the given inner class info. Factored
     * out to allow subclasses to print non-Java modifiers, e.g.,
     * those of JML.
     */
    protected void writeModifiers(IndentingWriter out, InnerClassInfo info) {
        writeModifiers(out, info.getModifiers());
    }

    /**
     * Prints the class file
     */
    private void writeClass(IndentingWriter out) {
	//      out.println("// compiler version: " + classInfo.getMajorVersion() + "." + classInfo.getMinorVersion());
	//      out.println();

	// Headers
	if (classInfo.sourceFileName() != null) {
	    out.println("@source \"" + classInfo.sourceFileName() + "\"");
	}

	if (options.multijava() && classInfo.isAnchor()) {
	    out.println("// generic function anchor");
	}
	
	// AHS: output universe type system version
	if (options.universes() &&
            classInfo.getUniverseVersion() != null) {
	    out.println("/* JUTS version: " + classInfo.getUniverseVersion() + " */");
	}

	// super is always set (and would print "synchronized")
	writeModifiers(out, classInfo);
	if ((classInfo.getModifiers() & ACC_INTERFACE) > 0) {
	    out.print("interface ");
	} else {
	    out.print("class ");
	}
	out.print(convertQualifiedName(classInfo.getName()));

	if (classInfo.getSuperClass() != null) {
	    out.print(" extends ");
	    out.print(convertQualifiedName(classInfo.getSuperClass()));
	}

	String[] interfaces = classInfo.getInterfaces();
	if (interfaces != null && interfaces.length > 0) {
	    out.print(" implements");
	    for (int i = 0; i < interfaces.length; i++) {
		out.print((i == 0 ? " " : ", ") + 
			  convertQualifiedName(interfaces[i]));
	    }
	}

	out.print(" {");

	FieldInfo[] fields = classInfo.getFields();
	if (options.sorted()) {
	    Arrays.sort(fields);
	}
	for (int i = 0; i < fields.length; i++) {
	    writeField(out, fields[i]);
	}

	out.println();

	MethodInfo[] methods = null;
	if (options.multijava()) {
	    methods = classInfo.getMultimethods();
	} else {
	    methods = classInfo.getMethods();
	}

	if (options.sorted()) {
	    Arrays.sort(methods);
	}
	for (int i = 0; i < methods.length; i++) {
	    writeMethod( out, methods[i] );
	    if (i != methods.length - 1) {
		out.println();
	    }
	}

	// !FIXME! Need to filter the inner class information as in
	// CBinaryClass.processInnerClasses

	InnerClassInfo[] inners = classInfo.getInnerClasses();
	if (inners != null && inners.length > 0 ) {
	    out.incrementLevel();
	    out.println();
	    out.println( "/** InnerClasses attribute */" );
	    out.decrementLevel();

	    if (options.sorted()) {
		Arrays.sort(inners);
	    }

	    for (int i = 0; i < inners.length; i++) {
		writeInner( out, inners[i] );
		if (i != inners.length - 1) {
		    out.println();
		}
	    }
	}

	out.println();
	out.println("}");
    }

    /**
     * Prints inner class info
     */
    private void writeInner( IndentingWriter out, InnerClassInfo info ) {
	out.incrementLevel();
	out.println();

	writeModifiers( out, info );
	out.print( "class " );
	out.print( info.getQualifiedName() );
	out.println(";");
	out.println(info.toString());

	out.decrementLevel();
    }

    /**
     * Prints fields
     */
    private void writeField(IndentingWriter out, FieldInfo info) {
	out.incrementLevel();
	out.println();

	writeModifiers(out, info);
	
	// AHS: output universe field annotation
	if (options.universes() &&
            info.getUniverseAnnotation() != null)
	    out.print(convertFieldSignature(info.getSignature(), null,
	                                    info.getUniverseAnnotation())+" ");
	else
	    out.print(convertFieldSignature(info.getSignature()) + " ");
        
	out.print(info.getName());

	Object	value = info.getConstantValue();
	if (value != null) {
	    out.print(" = " + convertLiteral(value));
	}
	out.print(";");
	out.decrementLevel();
    }

    /**
     * Prints methods
     */
    private void writeMethod( IndentingWriter out, MethodInfo info ) {
	out.incrementLevel();
	out.println();

	if( !options.inter() ) {
	    out.println("/**");
	    out.println(" * " + info.getName());
	    out.println(" *");
	    if (info.getCodeInfo() != null) {
		out.println(" * @stack  " + info.getCodeInfo().getMaxStack());
		out.println(" * @locals  " + info.getCodeInfo().getMaxLocals());
	    }
	    out.println(" */");
	}

	writeModifiers( out, info );
	
	// AHS: output purity
	if (options.universes() &&
            info.getUniverseAnnotation() != null)
	    out.print(info.getUniverseAnnotation().isPure() ? "pure " : "");
	
	// AHS: output universe annotations for argument and return types
	String[] type;
	if (options.universes() &&
            info.getUniverseAnnotation() != null)
	    type = convertMethodSignature(info.getSignature(), info.getUniverseAnnotation());
	else
	    type = convertMethodSignature(info.getSignature());
        
	out.print(type[1]);
	out.print(" ");
	out.print(info.getName());
	out.print(type[0]);

	String[]	exceptions = info.getExceptions();
	if (exceptions != null) {
	    out.print(" throws");
	    for (int i = 0; i < exceptions.length; i++) {
		out.print((i == 0 ? " " : ", ") + exceptions[i].replace('/', '.'));
	    }
	}
	if( info.getCodeInfo() == null || options.inter() ) {
	    out.print(";");
	} else {
	    out.print(" {");
	    writeCodeInfo(out, info.getCodeInfo());
	    out.println();
	    out.print("}");
	}
	out.decrementLevel();
    }

    /**
     * Prints code
     */
    private void writeCodeInfo(IndentingWriter out, CodeInfo info) {
	Instruction[]		insns = info.getInstructions();

	InstructionHandle[]		handles = new InstructionHandle[insns.length];
	for (int i = 0; i < handles.length; i++) {
	    handles[i] = new InstructionHandle(insns[i], i);
	}

	try {
	    info.transformAccessors(new HandleCreator(insns, handles));
	} catch (BadAccessorException e) {
	    throw new RuntimeException(e.getMessage());
	}

	out.incrementLevel();
	for (int i = 0; i < handles.length; i++) {
	    handles[i].write(out, options.stack());
	}
	writeHandlerInfo(out, info);
	writeLocalVariableInfo(out, info);
	out.decrementLevel();
    }

    private void writeModifiers(IndentingWriter out, long modifiers) {
	if ((modifiers & ACC_PUBLIC) != 0) {
	    out.print("public ");
	} else if ((modifiers & ACC_PROTECTED) != 0) {
	    out.print("protected ");
	} else if ((modifiers & ACC_PRIVATE) != 0) {
	    out.print("private ");
	}

	if ((modifiers & ACC_STATIC) != 0) {
	    out.print("static ");
	}
	if ((modifiers & ACC_ABSTRACT) != 0) {
	    out.print("abstract ");
	}
	if ((modifiers & ACC_FINAL) != 0) {
	    out.print("final ");
	}
	if ((modifiers & ACC_NATIVE) != 0) {
	    out.print("native ");
	}
	if ((modifiers & ACC_SYNCHRONIZED) != 0) {
	    out.print("synchronized ");
	}
	if ((modifiers & ACC_TRANSIENT) != 0) {
	    out.print("transient ");
	}
	if ((modifiers & ACC_VOLATILE) != 0) {
	    out.print("volatile ");
	}
    }

    // --------------------------------------------------------------------

    /**
     * Prints exception handlers
     */
    private void writeHandlerInfo(IndentingWriter out, CodeInfo info) {
	HandlerInfo[]	handlers = info.getHandlers();

	for (int i = 0; i < handlers.length; i++) {
	    out.println();
	    out.print("@catch  " + ((InstructionHandle)handlers[i].getHandler()).getLabel());
	    if (handlers[i].getThrown() != null) {
		out.print(" " + handlers[i].getThrown().replace('/', '.'));
	    }
	    out.print(" [" + ((InstructionHandle)handlers[i].getStart()).getLabel() + ", " + ((InstructionHandle)handlers[i].getEnd()).getLabel() + "]");
	}
    }

    /**
     * Prints local variables
     */
    private void writeLocalVariableInfo(IndentingWriter out, CodeInfo info) {
	LocalVariableInfo[]	localVariables = info.getLocalVariables();

	if (localVariables == null) {
	    return;
	}

	for (int i = 0; i < localVariables.length; i++) {
	    if (i == 0) {
		out.println();
	    }

	    out.println("@var  " + localVariables[i].getSlot()
			+ ": " + localVariables[i].getName()
			+ " " + convertFieldSignature(localVariables[i].getType())
			+ " [ " + ((InstructionHandle)localVariables[i].getStart()).getLabel()
			+ ", " + ((InstructionHandle)localVariables[i].getEnd()).getLabel()
			+ "]");
	}
    }

    // ----------------------------------------------------------------------
    // UTILITY METHODS WITH PACKAGE SCOPE
    // ----------------------------------------------------------------------

    /**
     * Converts a field signature into ksm syntax.
     *
     * @param	signature	the signature to convert.
     * @return	a string in ksm syntax.
     */
    /*package*/ static String convertFieldSignature(String signature) {
	return convertFieldSignature(signature, null);
    }
    
    /**
     * Converts a field signature into ksm syntax.
     *
     * @param	signature	the signature to convert.
     * @param 	staticType 	the name of the associated static type, if
     *				this signature represents a value specializer
     * @return	a string in ksm syntax.
     */
    private static String convertFieldSignature(String signature,
            					String staticType) {
        // AHS: call the other method
        return convertFieldSignature(signature, staticType, null);
    }
	
    /**
     * Converts a field signature into ksm syntax.
     *
     * @param	signature	the signature to convert.
     * @param 	staticType 	the name of the associated static type, if
     *				this signature represents a value specializer
     * @return	a string in ksm syntax.
     */
    private static String convertFieldSignature(String signature,
						String staticType,
						CUniverseTypeAnnotation uta) {
	int brackets = 0;
	while (signature.charAt(brackets) == '[') {
	    brackets++;
	}

	signature = signature.substring(brackets);

	
	switch (signature.charAt(0)) {
	case 'Z':
	    signature = "boolean";
	    break;
	    
	case 'B':
	    signature = "byte";
	    break;

	case 'C':
	    signature = "char";
	    break;

	case 'D':
	    signature = "double";
	    break;

	case 'F':
	    signature = "float";
	    break;

	case 'I':
	    signature = "int";
	    break;

	case 'J':
	    signature = "long";
	    break;

	case 'S':
	    signature = "short";
	    break;

	case 'V':
	    signature = "void";
	    break;

	case 'L':
	    // AHS: add universe annotation to reference type
	    if (uta == null) {
	        signature = signature.substring(1, signature.length() - 1);
	    } else {
	        if (brackets > 0) {
	            // first component universe, then later array universe
	            signature = uta.getBaseUniverse().toString() + " " +
	            		signature.substring(1, signature.length() - 1);
	        } else {
	            signature = uta.getUniverse().toString() + " " + 
	            		signature.substring(1, signature.length() - 1);
	        }
	    }
	    break;

	case 'X':
	    /* a value specializer!  */
	    // hacky to hard-code the qualified name of String
	    if (staticType.equals("java.lang.String")) {
		// the String has the form
		// X\"<stuff>\""
		// we strip the leading 2 characters, strip the
		// trailing 3 characters, and add a trailing quote, giving
		// "<stuff>"
		signature = signature.substring(2, signature.length() - 3)
		    + "\"";
	    } else {
		// strip the leading X and trailing quote
		signature = signature.substring(1, signature.length() - 1);
	    }
	    break;

	default:
	    throw new RuntimeException("UNEXPECTED TYPE " + signature.charAt(0));
	}
	
	// AHS: add array annotation
	if (uta !=  null && brackets > 0) {
	    signature = uta.getUniverse() + " " + signature;
	}
	
	for (; brackets > 0; brackets -= 1) {
	    signature += "[]";
	}

	return signature.replace('/', '.');
    }

    
    /**
     * Converts a method signature into ksm syntax.
     *
     * @param	signature	the signature to convert.
     * @return	an array of two string in ksm syntax:
     *		element 0: the argument types
     *		element 1: the return type
     */
    /* package */ static String[] convertMethodSignature(String signature) {
        // AHS: call the other method
        return convertMethodSignature(signature, null);
    }
    
    /**
     * Converts a method signature into ksm syntax. Prints universe annotations
     * for argument and return types.
     * 
     * @param	signature	the signature to convert
     * @param	uma		the method's universe annotation
     * @return	an array of two strings in ksm syntax:
     * 		element 0: the argument types
     * 		element 1: the return type
     */
    /* package */ static String[] convertMethodSignature(String signature,
            				CUniverseMethodAnnotation uma) {
        if (signature.charAt(0) != '(') {
	    throw new RuntimeException("UNEXPECTED TYPE " + signature);
	}

	String method = "(";

	int paramno = 0;
	signature = signature.substring(1);
	while (signature.charAt(0) != ')') {
	    int end;
	    if (signature.charAt(0) == '@') {
		// got a specialized argument
		int start;
		String sep;
		if (signature.charAt(1) == '@') {
		    // got a value specializer
		    start = 2;
		    sep = "@@";
		} else {
		    // got a regular specializer
		    start = 1;
		    sep = "@";
		}
		
		end = findParamDescriptionEnd(start,signature);
		String staticType =
		    convertFieldSignature(signature.substring(start, end));
		method += staticType + sep;
		signature = signature.substring(end);
		end = findParamDescriptionEnd(0,signature);
		method += convertFieldSignature(signature.substring(0,end),
						staticType);
		
	    } else {
		end = findParamDescriptionEnd(0, signature);
		
		// AHS: add universe annotation to param
		if (uma != null) {
		    method += convertFieldSignature(signature.substring(0,end),
		            		null, uma.getParamAnnotation(paramno));
		} else {
		    method += convertFieldSignature(signature.substring(0,end));
		}
	    }
	    signature = signature.substring(end);
	    if (signature.charAt(0) != ')') {
		method += ", ";
	    }
	    paramno++;
	}
	method += ")";
	
	String rettype;
	// AHS: add universe annotation to return type
	if (uma != null) {
	    rettype = convertFieldSignature(signature.substring(1), null,
	              uma.getRettypeAnnotation());
	} else {
	    rettype = convertFieldSignature(signature.substring(1));
	}

	return new String[] {method, rettype};
    }


    /**
     * Finds the end of the next parameter descriptor in
     * <CODE>signature</CODE> starting with the character in position
     * <CODE>start</CODE>.  */
    private static int findParamDescriptionEnd( int start, String signature ) {
	int result = start;
	// skip brackets
	while (signature.charAt(result) == '[') {
	    result += 1;
	}
	if (signature.charAt(result) == 'L') {
	    // a reference type, so skip to the semicolon
	    while (signature.charAt(result) != ';') {
		result += 1;
	    }
	    result += 1;
	} else if (signature.charAt(result) == 'X') {
	    // a value specializer.
	    // we're at the end when we see a quote symbol with
	    // an even number of backslashes preceding it.
	    boolean evenBackslashes = true;
	    boolean notDone = true;
	    while (notDone) {
		while (signature.charAt(result) != '\"') {
		    if (signature.charAt(result) == '\\')
			evenBackslashes = !evenBackslashes;
		    else
			evenBackslashes = true;
		    result += 1;
		}
		if (evenBackslashes) {
		    notDone = false;
		} else {
		    evenBackslashes = true;
		}
		result += 1;
	    }		
	} else {
	    // a primitive type
	    result += 1;
	}
	return result;
    }

    /**
     * Converts a qualified identifier into ksm syntax.
     *
     * @param	ident		the identifier to convert.
     * @return	a string in ksm syntax.
     */
    /*package*/ static String convertQualifiedName(String ident) {
	StringTokenizer token = 
	    new StringTokenizer(ident.replace('/', '.'), ".");
	String result;

	result = token.nextToken();
	result = isJavaIdentifier(result) ? result : "\'" + result + "\'";
	while (token.hasMoreElements()) {
	    String current = token.nextToken();
	    result += "." + (isJavaIdentifier(current) ? 
			     current : "\'" + current + "\'");
	}
	return result;
    }

    private static boolean isJavaIdentifier(String ident) {
	if (ident.equals("<init>") || ident.equals("<clinit>")) {
	    return true;
	} else if (!Character.isJavaIdentifierStart(ident.charAt(0))) {
	    return false;
	} else {
	    for (int i = 1; i < ident.length(); i++) {
		if (!Character.isJavaIdentifierPart(ident.charAt(i))) {
		    return false;
		}
	    }
	    return true;
	}
    }

    /**
     * Converts a literal into ksm syntax.
     */
    /*package*/ static String convertLiteral(Object t) {
	if (t instanceof String) {
	    return convertStringLiteral((String)t);
	} else if (t instanceof Float) {
	    return convertFloatLiteral((Float)t);
	} else if (t instanceof Double) {
	    return convertDoubleLiteral((Double)t);
	} else if (t instanceof Long) {
	    return convertLongLiteral((Long)t);
	} else {
	    return t.toString();
	}
    }

    private static String convertDoubleLiteral(Double t) {
	if (t.isNaN()) {
	    return "<NaN>D";
	} else if (t.isInfinite()) {
	    return t.doubleValue() > 0 ? "<Inf>D" : "-<Inf>D";
	} else {
	    return t + "D";
	}
    }

    private static String convertFloatLiteral(Float t) {
	if (t.isNaN()) {
	    return "<NaN>F";
	} else if (t.isInfinite()) {
	    return t.floatValue() > 0 ? "<Inf>F" : "-<Inf>F";
	} else {
	    return t + "F";
	}
    }

    private static String convertLongLiteral(Long t) {
	return t + "L";
    }

    private static String convertStringLiteral(String t) {
	StringBuffer buffer = new StringBuffer();

	buffer.append("\"");
	for (int i = 0; i < t.length(); i++) {
	    switch (t.charAt(i)) {
	    case '\t':
		buffer.append("\\t");
		break;
	    case '\r':
		buffer.append("\\r");
		break;
	    case '\n':
		buffer.append("\\n");
		break;
	    case '\"':
		buffer.append("\\\"");
		break;
	    case '\\':
		buffer.append("\\\\");
		break;
	    case '\'':
		buffer.append("\\'");
		break;
	    default:
		buffer.append(t.charAt(i));
	    }
	}
	buffer.append("\"");
	return buffer.toString();
    }

    // ----------------------------------------------------------------------
    // PRIVATE METHODS
    // ----------------------------------------------------------------------

    // --------------------------------------------------------------------
    // INNER CLASSES
    // --------------------------------------------------------------------

    /**
     * A helper class to tune the disassembler for languages that
     * extend Java, such as JML.
     */
    public static class DisassemblerHelper {
        public String inputExtension() {
            return ".class";
        }

        public String outputExtension() {
            return ".ksm";
        }

        public Disassembler createDisassembler(ClassInfo classInfo, 
                                               DisOptions options) {
            return new Disassembler(classInfo, options);
        }

        public ClassInfo createClassInfo(DataInput in) 
            throws IOException, ClassFileFormatException {
            return new ClassInfo(in, false);
        }

        public ClassInfo createClassInfo(String name) 
	    throws ClassFileReadException
	{
            return ClassPath.getClassInfo(name, false);
        }
    }

    private static final DisassemblerHelper helper = new DisassemblerHelper();

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    protected final ClassInfo	classInfo;
    protected final DisOptions	options;
}
