/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.15 2007/12/20 12:25:26 chalin Exp $
 */

package org.multijava.dis;

import java.text.MessageFormat;

import org.multijava.util.classfile.AttributeList;
import org.multijava.util.classfile.ClassPath;
import org.multijava.util.classfile.MJAttributeParser;
import org.multijava.util.classfile.UniverseAttributeParser;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CompilationAbortedException;
/**
 * This class is the entry point for the KOPI disassembler.
 */
public class Main implements Constants {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Only main and subclasses can construct Main
     */
    //x@ modifies args,this.*; // FIXME - this line does not work
    protected Main(String[] args) {
	this.args = args;
    }

    /**
     * Disassembles the files specified by the instantiation arguments
     *
     * @return	true iff disassembly succeeded
     */
    protected boolean disasm() {
	try {
	    if( !parseArguments(args) ) {
		return false;
	    }
	} catch (CompilationAbortedException e) {
	    // Assume an argument problem, and that usage info was printed
	    // by the options processor
	    return false;
	}
	
	if (options.classpath() == null) {
	    String path = System.getProperty( "java.class.path" );
	    ClassPath.init(path == null ? "." : path,null);
	} else {
	    ClassPath.init(options.classpath(),null);
	}

	String[] infiles = options.nonOptions();

	if (infiles.length == 0) {
	    options.usage();
	    System.err.println(DisMessages.NO_SOURCE_FILE);
	    return false;
	} else {
	    if (options.multijava()) {
		AttributeList.addParser(new MJAttributeParser());
	    }
	    
	    // AHS: insert universe attribute parser into parsing chain
	    if (options.universes()) {
	        AttributeList.addParser(new UniverseAttributeParser());
	    }

	    boolean	errorsFound = false;

	    for (int i = 0; i < infiles.length; i++) {
		if (options.verbose()) {
		    System.err.println( 
			MessageFormat.format( 
			    DisMessages.PROCESSING.getFormat(), 
			    new String[] {infiles[i]} ));
		}

		try {
		    disassembleClass( infiles[i] );
		} catch( UnpositionedError e ) {
		    System.err.println("Error: " + e.getMessage());
		    errorsFound = true;
		}
	    }

	    return !errorsFound;
	}
    }

    // --------------------------------------------------------------------
    // ACTIONS
    // --------------------------------------------------------------------

    /**
     * Reads, and disassembles a class file
     * @exception	UnpositionedError	an error occurred
     */
    protected void disassembleClass( String fileName ) 
        throws UnpositionedError {
	Disassembler.disassemble( fileName, options.destination(), options );
    }

    // --------------------------------------------------------------------
    // ENTRY POINT
    // --------------------------------------------------------------------

    /**
     * Entry point to the assembler
     */
    public static void main(String[] args) {
	boolean success = compile(args);
	System.exit( success ? 0 : 1 );
    }

    /** This static method is inserted simply to facilitate testing. */
    public static boolean compile(String[] args) {
	return new Main(args).disasm();
    }

    public static boolean run( String[] args ) {
	return new Main( args ).disasm();
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /*
     * Parse command line arguments.
     */
    //@ modifies options;
    protected boolean parseArguments(String[] args) {
	options = new DisOptions();
	return options.parseCommandLine(args);
    }

    private String[] args;

    protected DisOptions		options;
}
