/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.launcher;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JPanel;

import org.multijava.util.gui.GUIWindowPreferences;

/**
 *  The implementation for a launcher that only provides a list of buttons
 *  for access to the tools.  
 *
 *@author     Kristina Boysen
 *@version    $Revision: 1.8 $
 */

public class CompactLauncher extends ImpLauncher {
        
    // ----------------------------------------------------------------------
    // CONSTRUCTOR
    // ----------------------------------------------------------------------
        
    /**
     *  Constructor for the CompactLauncher
     */
    public CompactLauncher(String name, ResettableIterator tools, 
                               String logoName) {
        super(name, tools, logoName);
                
        // Initialize preference system for the full launcher
        compactLauncherPrefs = new GUIWindowPreferences(name, "Compact", this);
        compactLauncherPrefs.setAllPreferences();
                
        setResizable(true);
        setVisible(true);
    }
        
    // ----------------------------------------------------------------------
    // LAYOUT CREATION METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Creates the CompactLauncher panel layout by placing a logo on top of
     *  a panel with a series of buttons with access to the tools.
     */
    protected void createLauncherLayout() {
        getContentPane().setLayout(new BoxLayout(
            getContentPane(), BoxLayout.Y_AXIS));
        getContentPane().add(createLogoPanel());
        getContentPane().add(Box.createVerticalStrut(15));
        getContentPane().add(createSelectPanel());
        getContentPane().add(Box.createVerticalStrut(15));
        getContentPane().add(createSwitchButtonPanel());
    }
        
    /**
     *  Creates the panel that holds the logo and tool information
     */
    private JPanel createLogoPanel() {
        return new ImagePanel(loadIcon(logoName));
    }
        
    /**
     *  Loads the logo from disk, blocking for up to MAX_LOGO_DELAY ms
     *  to allow the image to fully load.  Fully loading the image
     *  ensures that the Image size is known.
     */
    private Image loadIcon(String filename) {
        MediaTracker mt = new MediaTracker(this);
        int logoID = 0;
        Image logo = Toolkit.getDefaultToolkit().createImage(
                     ClassLoader.getSystemClassLoader().getResource(filename));
        mt.addImage(logo, logoID);
        try {
            mt.waitForID(logoID, MAX_LOGO_DELAY);            
        } catch (InterruptedException e) {
            /* !FIXME! Is there something better we can do if the
               image loading is interrupted? */
        }
        return logo;
    }
        
    /**
     *  Creates the panel that holds all of the tool buttons
     */
    private JPanel createSelectPanel() {
        // Initialize the panel
        JPanel selectPanel = new JPanel();
        selectPanel.setLayout(new BoxLayout(selectPanel, BoxLayout.Y_AXIS));
        selectPanel.setBorder(BorderFactory.createTitledBorder(
                                  "Please select desired tool"));
                
        // Dynamically add each tool button to the panel
        while (tools.hasNext()) {
            Map.Entry tool = (Map.Entry)(tools.next());
            JButton button = new JButton((String)tool.getKey());
            button.setAlignmentX(Component.CENTER_ALIGNMENT);
            button.addActionListener((ActionListener)tool.getValue());
            selectPanel.add(button);
        }
        
        return selectPanel;
    }
        
    /**
     *  Returns a panel with the compact launcher switching button
     */
    private JButton createSwitchButtonPanel() {
        JButton button = new JButton("Switch to Full Launcher");
        button.setAlignmentX(Component.CENTER_ALIGNMENT);
        button.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    setVisible(false);
                    LauncherFactory.instance().getFullLauncher()
                        .setVisible(true);
                }
            });
        button.setEnabled(false);
            
        return button;
    }
        
    // ----------------------------------------------------------------------
    // PREFERENCE METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Returns the user's preferred size for this window.
     */
    public Dimension getPreferredSize() {
        return getMinimumSize();
    }
        
    // ----------------------------------------------------------------------
    // PROTECTED CLASSES
    // ----------------------------------------------------------------------

    /**
     *  Creates the panel that draws the logo
     */
    protected static class ImagePanel extends JPanel {
        private Image image;

        /**
         *  Constructor for ImagePanel
         */
        public ImagePanel(Image image) {
            this.image = image;
        }

        /**
         *  Overrides paintComponent(g) in JPanel so ImagePanel
         *  paints the image at initialization
         */
        public void paintComponent(Graphics g) {
            super.paintComponent(g);
            g.drawImage(image, 0, 0, this);
        }

        public Dimension getMinimumSize() {
            return new Dimension(image.getWidth(this), image.getHeight(this));
        }

        public Dimension getPreferredSize() {
            return getMinimumSize();
        }
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------

    /**
     *  Window size and location preferences for the compact launcher window.
     */
    private GUIWindowPreferences compactLauncherPrefs = null;
        
    /**
     * Records the number of milliseconds to wait for the logo image
     * to load before proceeding.
     */
    private long MAX_LOGO_DELAY = 5000;
}
