/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
 
package org.multijava.launcher;

import java.awt.BorderLayout;
import java.awt.Dimension;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.util.Map;

import javax.swing.DefaultListModel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;

import org.multijava.util.gui.AbstractOpenHandler;
import org.multijava.util.gui.GUIFileListPanel;
import org.multijava.util.gui.GUIWindowPreferences;

/**
 *  The implementation for a launcher that provides access to the tools
 *  as well as keeping track of common preferences and lists of files between
 *  tools.
 *
 *@author     Kristina Boysen
 *@version    $Revision: 1.3 $
 */

public class FullLauncher extends ImpLauncher {
        
    // ----------------------------------------------------------------------
    // CONSTRUCTOR
    // ----------------------------------------------------------------------
        
    /**
     *  Constructor for the FullLauncher
     */
    public FullLauncher(String name, ResettableIterator tools, 
                            String logoName) {
        super(name, tools, logoName);
                
        // Initialize preference system for the full launcher
        fullLauncherPrefs = new GUIWindowPreferences(name, "Full", this);
        fullLauncherPrefs.setAllPreferences();
                
        setResizable(true);
        setVisible(true);
    }
        
    // ----------------------------------------------------------------------
    // LAYOUT CREATION METHODS
    // ----------------------------------------------------------------------
    
    /**
     *  Creates the environment layout and menus in which other tools can run
     */
    protected void createLauncherLayout() {
        fileModel = new DefaultListModel();
        fileList = new JList(fileModel);
                
        setJMenuBar(createMenus());
        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(new GUIFileListPanel(
            fileList, new OpenHandler()), BorderLayout.NORTH);
    }
        
    /**
     *  Creates the menu bar
     */
    private JMenuBar createMenus() {
        JMenuBar menuBar = new JMenuBar();

        menuBar.add(createFileMenu());
        menuBar.add(createViewMenu());
        menuBar.add(createRunMenu());
                
        return menuBar;
    }
        
    /**
     *  Creates the File menu
     */
    private JMenu createFileMenu() {
        JMenu fileMenu = new JMenu("File");
                
        // Open button
        JMenuItem openItem = new JMenuItem("Open File...");
        openItem.addActionListener(new OpenHandler());
                
        // Exit button
        JMenuItem exitItem = new JMenuItem("Exit");
        exitItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    System.exit(0);
                }
            });
                
        fileMenu.add(openItem);
        fileMenu.addSeparator();
        fileMenu.add(exitItem);
                
        return fileMenu;
    }
        
    /**
     *  Creates the View menu
     */
    private JMenu createViewMenu() {
        JMenu viewMenu = new JMenu("View");
                
        // Switch button
        JMenuItem switchItem = new JMenuItem("Switch to Compact Launcher");
        switchItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    setVisible(false);
                    LauncherFactory.instance().getCompactLauncher()
                        .setVisible(true);
                }
            });

        viewMenu.add(switchItem);
                
        return viewMenu;
    }
        
    /**
     *  Creates the Run menu
     */
    private JMenu createRunMenu() {
        JMenu runMenu = new JMenu("Run Tool");
        
        // Dynamically add each tool button to the panel
        while (tools.hasNext()) {
            Map.Entry tool = (Map.Entry)(tools.next());
            JMenuItem menu = new JMenuItem((String)tool.getKey());
            menu.addActionListener((ActionListener)tool.getValue());
            runMenu.add(menu);
        }
        
        return runMenu;
    }

    // ----------------------------------------------------------------------
    // PREFERENCE METHODS
    // ----------------------------------------------------------------------
        
    /**
     * Returns the window's minimum size.
     */
    public Dimension getMinimumSize() {
        int width = 300;
        int height = 200;
        return new Dimension(width, height);
    }
        
    /**
     *  Returns the user's preferred size for this window.
     */
    public Dimension getPreferredSize() {
        return fullLauncherPrefs.getPreferredSize();
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASSES
    // ----------------------------------------------------------------------
        
    /**
     *  Handles the selection of files in an open file window
     */
    private class OpenHandler extends AbstractOpenHandler {
                
        /**
         *  Constructs the "Open File" dialog
         */
        protected OpenHandler() {
            super();
        }
                
        /**
         *  Returns the starting directory preferred by the user
         */
        protected String getCurrentDirectory() {
            return ".";
        }
                
        /**
         *  Returns the list of files that the tool is going to process
         */
        protected JList getFileList() {
            return fileList;
        }
                
        /**
         *  Set current directory location to last location accessed
         */
        protected void setDirectory(String file) {
        }
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------
    
    /**
     *  Window size and location preferences for the full launcher window.
     */
    private GUIWindowPreferences fullLauncherPrefs = null;
        
    /**
     *  List of files to process
     */
    private /*@ spec_public @*/ JList fileList;
        
    /**
     *  Represents the contents of fileList
     */
    private /*@ spec_public @*/ DefaultListModel fileModel;
}
