/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.launcher;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

/**
 *  The superclass for all of the tool set definitions
 *
 *@author     Kristina Boysen
 *@version    $Revision: 1.9 $
 */

public abstract class Launcher {

    // ----------------------------------------------------------------------
    // BUILDER METHODS
    // ----------------------------------------------------------------------

    /**
     *  Returns a Singleton instance of the ToolLauncher
     */
    protected ImpLauncher getImpLauncher(String name, String logoName) {
        if (impLauncher == null) {
            impLauncher = LauncherFactory.instance().getImpLauncherInstance(
                name, getTools(), logoName);
        }       
                
        return impLauncher;
    }
        
    // ----------------------------------------------------------------------
    // ABSTRACT METHODS
    // ----------------------------------------------------------------------
    
    /**
     *  Returns the Iterator that cycles through the tool descriptions and 
     *  AbstractToolLaunchers
     */
    protected abstract ToolIterator getTools();
        
    // ----------------------------------------------------------------------
    // PROTECTED CLASSES
    // ----------------------------------------------------------------------
        
    /**
     *  The Iterator that uses an ArrayList to cycle through Hashtables
     *  that hold the tool descriptions and AbstractToolLaunchers to enable
     *  the launchers to run the various tools.  The Iterator enables these
     *  launchers to visually present the tools in the order they were added
     *  to the iterator.
     */
    protected static class ToolIterator implements ResettableIterator {
        ArrayList tools;
        int currentIndex;
        //@            in objectState;

        //@ public represents moreElements <- hasNext();

        /**
         *  Constructor for the ToolIterator
         */
        public ToolIterator() {
            tools = new ArrayList();
            currentIndex = 0;
        }

        /**
         *  Checks to see whether the iterator has a next element.
         */
        public /*@ pure @*/ boolean hasNext() {
            return (currentIndex < tools.size());
        }
                
        /**
         *  Returns the next element in the iterator
         */
        public Object next() {
            //@ set remove_called_since = false;
            try {
                return (Object)(tools.get(currentIndex++));
            } catch (ArrayIndexOutOfBoundsException e) {
                throw new NoSuchElementException();
            }
        }
                
        /**
         *  Removes the current element from the iterator
         */
        public void remove() {
            //@ set remove_called_since = true;
            tools.remove(currentIndex);
            if (currentIndex == tools.size()) {
                currentIndex--;
            }
        }
                
        /**
         *  Adds an element to the iterator
         */
        public void add(Object newItem) {
            tools.add((ToolIteratorPair)(newItem));
        }
                
        /**
         *  Returns the current index of the iterator
         */
        public int index() {
            return currentIndex;
        }
                
        /**
         *  Resets the iterator back to its initial index
         */
        public void reset() {
            currentIndex = 0;
        }
    }
        
    /**
     *  Collects the name of the tool and the launcher of the tool into
     *  a mapping that can be accessed by outside classes because of the
     *  common interface
     */
    protected static class ToolIteratorPair implements Map.Entry {
        String toolName; //@ in abstractKey;
        AbstractLaunchTool launcher; //@ in abstractValue;

        //@ represents abstractKey <- toolName;
        //@ represents abstractValue <- launcher;

        /**
         *  Constructor
         */
        public ToolIteratorPair(String toolName, AbstractLaunchTool launcher) {
            this.toolName = toolName;
            this.launcher = launcher;
        }
                
        /**
         *  Returns the key for this pair
         */
        public Object getKey() {
            return toolName;
        }
                
        /**
         *  Returns the value for this pair
         */
        public Object getValue() {
            return launcher;
        }
                
        /**
         *  Replaces the old value with the new value and returns this new
         *  value
         */
        public Object setValue(Object value) {
            launcher = (AbstractLaunchTool)value;
            return launcher;
        }
    }
    
    // ----------------------------------------------------------------------
    // PROTECTED CLASS VARIABLES
    // ----------------------------------------------------------------------
        
    /**
     *  Keeps an instance of the ImpLauncher class so it can "bridge" the
     *  abstraction of the various tool launchers to the implementation
     *  of how they are displayed.
     */
    protected ImpLauncher impLauncher = null;
}


