/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CAbstractMethodSet.java,v 1.8 2008/05/14 18:16:30 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;


import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.classfile.MethodInfo;

/**
 * This class encapsulates a collection of method signatures.  It is
 * used to track the type signatures of methods declared in a class
 * and to track the apparent methods of a class as augmented by
 * external methods as viewed by another compilation unit.  */
public abstract class CAbstractMethodSet extends org.multijava.util.Utils {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    protected CAbstractMethodSet() {
	this.methods = new ArrayList();
    }
  
    protected CAbstractMethodSet( CMethod[] methodArray ) {
	this.methods = new ArrayList( Arrays.asList( methodArray ) );
    }


    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Indicates the number of methods in this collection.
     *
     * @return	the number of methods
     */
    public int size() {
	return methods.size();
    }
    
    /**
     * Indicates whether this set is empty.
     *
     * @return	true iff this set is empty
     */
    public boolean isEmpty() {
	return methods.isEmpty();
    }


    /**
     * Indicates whether this set contains the given method.
     */
    public boolean contains( CMethod m ) {
	return methods.contains(m);
    }

    /**
     * Returns the first method in this collection or null if the
     * collection is empty.
     *
     * @return the first method or null */
    public CMethod getFirst() {
	if( methods.size() == 0 ) {
	    return null;
	} else {
	    return getMethod( 0 );
	}
    }

    public String toString() {
	return "{ " + methods.toString() + " }";
    }

    /**
     * Adds the given method to this method set.
     *
     * @param	m	the method to be added to this set
     */
    public void addMethod( CMethod m ) {
	methods.add( m );
    }
    
    /**
     * Returns the method with the given index.
     *
     * @param	i	the index of the method to return
     */
    public CMethod getMethod( int i ) {
	return (CMethod) methods.get( i );
    }
    
    public CAbstractMethodSet.Iterator iterator() {
	return new Iterator();
    }

    //---------------------------------------------------------------------
    // TYPE-SPECIFIC ITERATOR
    //---------------------------------------------------------------------

    /**
     * This nested class mimics the <code>java.util.Iterator</code>
     * interface but is specific to collections of
     * <code>CMethod</code>. 
     *
     * @see java.util.Iterator
     * @see #iterator()
     */
    public final class Iterator {
	private Iterator() { 
	    self = ((/*@\peer@*/ ArrayList) methods).iterator();
	}

	public boolean hasNext() {
	    return self.hasNext();
	}

	public CMethod next() {
	    return (CMethod) self.next();
	}

	public void remove() {
	    self.remove();
	}

	private java.util.Iterator self;
    }
    

    // ----------------------------------------------------------------------
    // PRIVATE AND PROTECTED DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * The signatures of the methods in this collection.
     */
    private final ArrayList /* of CMethod */ methods;
}
