/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CAnchorInitializer.java,v 1.9 2005/04/26 02:39:30 kui_dai Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.MethodRefInstruction;

/**
 * This is used as a wrapper around the normal static or instance
 * initializer method of a class that includes nested
 * dispatcher classes (e.g., an anchor class).  The wrapper generates 
 * a call to the method that initializes the chain of responsibility.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.9 $ */

public class CAnchorInitializer extends CSourceMethod {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Creates a new <code>CAnchorInitializer</code> instance.
     *
     * <pre><jml>
     * {|
     *   requires owner != null;
     *   requires regularInit == null;
     *   ensures (* generated code just includes chain-of-responsibility
     *              initialization *);
     * also
     *   requires owner != null;
     *   requires regularInit != null;
     *   ensures (* generated code includes chain-of-responsibility
     *              initialization followed by regular initialization *);
     * |}
     * </jml></pre>
     *
     * @param regularInit       a regular initializer to which a call to the 
     *                          chain-of-responsibility initialization code 
     *                          will be prepended
     */
    public CAnchorInitializer(CSourceClass owner, CSourceMethod regularInit,
                              boolean isStatic ) {
        // can't just pull data from regularInit; it might be null
        super( new MemberAccess(owner, isStatic ? ACC_STATIC : 0L),
               isStatic ? JAV_STATIC_INIT : JAV_INIT, CStdType.Void, 
               new CSpecializedType[0], CClassType.EMPTY, 
               CTypeVariable.EMPTY, false, null, null, null );
        this.regularInit = regularInit;
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // DERIVED ATTRIBUTES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // INTERFACE CHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // TYPECHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    public ArrayList genMethodInfo() {
        ArrayList result = new ArrayList( 1 );
        result.add(nonEmptyMethodInfo( ident() ));
        return result;
    }

    /**
     * Plants the bytecode for the body of this method, including code
     * to initialize the chain-of-responsibility for the nested
     * dispatcher classes.  */
    protected void plantBodyBytecode( CodeSequence code ) {
        // initialize the chains-of-responsibility
        code.plantInstruction(
            new MethodRefInstruction(opc_invokestatic, owner().qualifiedName(),
                                     CCORInitializer.initMethodName,
                                     CType.genMethodSignature(CStdType.Void,
                                                              CType.EMPTY) ));


        // add the rest of the initialization if necessary
        if (regularInit != null) {
            regularInit.plantBodyBytecode( code );
        } else {
            code.plantInstruction(new NoArgInstruction(opc_return));
        }
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * The regular initializer wrapped by this.
     */
    private CSourceMethod regularInit;
}
