/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CArrayType.java,v 1.38 2008/05/14 18:16:29 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.Collection;
import java.util.Collections;

import org.multijava.util.classfile.UniverseByteConstants;
import org.multijava.util.compiler.FastStringBuffer;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the array type in the type structure.
 * @see	CType
 */
public class CArrayType extends CClassType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    // /*
    //  * Constructs an array type
    //  * @param	baseType	the base type of the array type
    //  * @param	arrayBound	the dimension of the array type
    //  *
    //  * WMD: the only time this constructor should be called directly,
    //  * is for old-style array declarations. We don't support
    //  * two universe specifiers in the old-style declaration!
    //  *
    //  * @deprecated
    //  */
    // public CArrayType(CType baseType, int arrayBound) {
    //    this( baseType, arrayBound, null );
    // }

    /**
     * Constructs an array type with universe modifier for the whole array. WMD
     *
     * @param	baseType	the base type of the array type
     * @param	arrayBound	the dimension of the array type
     * @param   array_univ      the universe for the whole array
     *
     */
    public CArrayType(CType baseType, int arrayBound, CUniverse array_univ) {
	super();
	type = TID_ARRAY;
	assertTrue(baseType != null);
	assertTrue(arrayBound > 0);

	if (baseType.isArrayType()) {
	    this.arrayBound = arrayBound + ((CArrayType)baseType).arrayBound;
	    this.baseType = ((CArrayType)baseType).baseType;
	} else {
	    this.arrayBound = arrayBound;
	    this.baseType = baseType;
	}

	if( array_univ == null ) {
	    // WMD TODO log a warning
	    this.universe = CUniverseServices.getDefaultVariable();
	} else {
	    this.universe = array_univ;
	}
        
        // this constructor should only be used for arrays of
        // class types or of one-dimensional arrays of primitive type
        // for compatibility with all the places with old-style arrays
        // in Mjc.g and Jml.g, we check for Universes too.
        if( array_univ != null )
            assertTrue(!baseType.isPrimitive() || arrayBound == 1);
        
        // not used
        elem_univ = null;
    }

    
    /**
     * Constructs an array type of a primitive type with an Universe
     * modifier for the whole array and an modifier for the element type;
     * for arrays of primitive type with more than one dimension.
     *
     * @param	baseType	the base type of the array type
     * @param	arrayBound	the dimension of the array type
     * @param   array_univ      the universe for the whole array
     * @param   elem_univ       the universe for the subarrays.
     *
     */
    public CArrayType(CType baseType, int arrayBound,
            CUniverse array_univ, CUniverse elem_univ) {
	super();
	type = TID_ARRAY;
	assertTrue(baseType != null);
        assertTrue(baseType.isPrimitive());
	assertTrue(arrayBound > 1);

        this.arrayBound = arrayBound;
	this.baseType = baseType;

	if( array_univ == null ) {
	    // WMD TODO log a warning
	    this.universe = CUniverseServices.getDefaultVariable();
	} else {
	    this.universe = array_univ;
	}
        if( elem_univ == null ) {
            this.elem_univ = CUniverseServices.getDefaultVariable();
        } else {
            this.elem_univ = elem_univ;
        }
    }


    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Transforms this type to a string
     * @return	the ksm form of this type
     */
    public String toString() {
	String	res = baseType.toString();
	for (int i = 0; i < arrayBound; i++) {
	    res += "[]";
	}
	return res;
    }


    /**
     * Return the universe modifier plus the class type as string.
     * 
     * by WMD
     * by RZU
     */
    public String toVerboseString() {
    	if (CTopLevel.getCompiler().universeChecks()) {
    		return getCUniverseString() + " " + toString();
    	} else {
    		return toString();
    	}
    }

    
    /** Return the universe modifiers as string.
     * Always returns a Universe string, even if the Universe checks
     * are turned of!
     * By WMD.
     */
    public String getCUniverseString() {
	String	res = super.getCUniverseString();

	if( baseType.isClassType() ) {
	    res += " " + ((CClassType) baseType).getCUniverseString();
	}
        if( baseType.isPrimitive() && arrayBound > 1 ) {
            res += " " + elem_univ.toString();
        }
	
	return res;
    }
    
    /** Return the universe modifier as string, for output in JML
     * specifications. The only difference is that implicit annotations
     * are escaped with a backslash.
     * By AHS, WMD.
     */ 
    public String getCUniverseJMLString() {
        String res = super.getCUniverseJMLString();
        
        if (baseType.isClassType()) {
            res += " " + ((CClassType) baseType).getCUniverseJMLString();
        }
        if( baseType.isPrimitive() && arrayBound > 1 ) {
            res += " " + elem_univ.toJMLString();
        }
        
        return res;
    }
    

    /**
     * Returns the universe modifier as String, for output in source
     * code parsed by MultiJava, i.e. the modifiers are written as is.
     * By RZU, WMD.
     */
    public String getCUniverseMJString() {
        String res = super.getCUniverseMJString();
        
        if (baseType.isClassType()) {
            res += " " + ((CClassType) baseType).getCUniverseMJString();
        }
        if( baseType.isPrimitive() && arrayBound > 1 ) {
            res += " " + elem_univ.toMJString();
        }

        return res;
    }


    /**
     *
     */
    public String qualifiedName() {
	return getSignature();
    }

    /**
     * Transforms this type to a string
     * @return	the class file form of thid type
     */
    public void appendSignature(FastStringBuffer buff) {
	for (int i = 0; i < arrayBound; i++) {
	    buff.append('[');
	}

	baseType.appendSignature(buff);
    }

    public void appendGenericSignature(FastStringBuffer buff) {
	for (int i = 0; i < arrayBound; i++) {
	    buff.append('[');
	}
	baseType.appendGenericSignature(buff);
    }

    /**
     * @return	the size used in stack by value of this type
     */
    public int getSize() {
	return 1;
    }

    /**
     * @return	is this type an array ?
     */
    //@ pure
    public boolean isArrayType() {
	return true;
    }

    public boolean isReifiableType() {
        return baseType.isReifiableType();
    }

    /** Determine and return the base type of the array.
     *
     * @return	The base class of this type
     */
    public CType getBaseType() {
		assertTrue(baseType != null);
		assertTrue(!(baseType instanceof CArrayType));
	    return baseType;
    }
    
    /**
     * Sets the base type of the array.
     * 
     * @param baseType The new base type
     */
    public void setBaseType(CType baseType) {
    	this.baseType = baseType;
    }


    /**
     * @return	the base class of this type
     *
     * WMD: this is either the same as the result of getBaseType
     * or a new CArrayType with the same basetype and universe
     * but one fewer dimension.
     */
    public CType getIndexedType() {
	assertTrue(baseType != null);
	assertTrue(!(baseType instanceof CArrayType));

	if (arrayBound == 1) {
	    // return baseType;
	    // WMD should we combine this with the array universe?
	    // WMD yes -> use the getBaseType method!
	    return getBaseType();
	} if (baseType.isPrimitive() ) {
            // don't expect an element modifier if there is no main modifier.
            // WMD TODO: the old-style arrays currently break this check.
            // assertTrue( this.universe==null ||  elem_univ!=null );
            if (arrayBound == 2) {
                // Going from 2 to 1 is special
                return new CArrayType(baseType, 1, this.elem_univ);
            } else {
                return new CArrayType(baseType, arrayBound-1,
                        this.universe, this.elem_univ);
            }
        } else {
	    // WMD make sure that the array universe is preserved
	    return new CArrayType(baseType, arrayBound - 1, this.universe);
	}
    }


    /** Return the Universe modifier that is used for multi-dimensional
     * arrays of primitive types.
     * 
     * @return Universe modifier for the element array.
     */
    public CUniverse getCUniverseElem() {
        assertTrue(arrayBound > 1 && baseType.isPrimitive());
        return elem_univ;
    }
    
    /** Set the Universe modifier that is used for multi-dimensional
     * arrays of primitive types.
     * 
     * @param u The new Universe modifier.
     */
    public void setCUniverseElem(CUniverse u) {
        assertTrue(arrayBound > 1 && baseType.isPrimitive());
        elem_univ = u;
    }
    
    
    /** Return the Universe bytecode constant of this class type.
     * By AHS.
     */
    public byte getUniverseByteConstant() {
        return baseType.isClassType() ?
            UniverseByteConstants.buildConstant(universe, ((CClassType)baseType).getCUniverse())
          : UniverseByteConstants.buildConstant(universe, CUniverseImplicitPeer.getUniverse());
    }


    /**
     * @return	the number of array bracket of this type
     */
    public int getArrayBound() {
	return arrayBound;
    }

    /**
     * @return	true if this type is valid
     */
    //@ pure
    public boolean checked() {
	return baseType.checked();
    }

    /**
     * Compares two objects for equality.
     *
     * @param other	the object to compare with
     * @return true iff <code>other</code> is a <code>CType</code> object 
     *		representing the same Java type as this represents
     *
     * WMD: also compare the universe modifiers.
     */
    public boolean equals(/*@ nullable @*/  Object other ) {
	return equals( other, CTopLevel.getCompiler().universeChecks(), false );
    }

    public boolean equals(/*@ nullable @*/ Object other, boolean enableUniv,
    		boolean allowUniversesCovariance) {
	if( other instanceof CArrayType ) {
	    CArrayType arr = (CArrayType)other;

	    return baseType.equals(arr.baseType) 
		&& arrayBound == arr.arrayBound
		// WMD
		&& ( !enableUniv ||
		     ( getCUniverse().equals(arr.getCUniverse()) &&
                       (elem_univ==null || elem_univ.equals(arr.elem_univ)) ) );

	}
	return false;
    }
    

    public boolean equals(/*@ nullable @*/ CType other,
			  /*@ non_null @*/ CClassType[] substitution) {
	return equals(other, substitution, true);
    }

    public boolean equals(/*@ nullable @*/ CType other,
			  /*@ non_null @*/ CClassType[] substitution,
			  boolean enableUniv ) {
	if (!other.isArrayType()) {
	    return false;
	} else {
	    CArrayType arr = (CArrayType)other;
	    
	    return baseType.equals(arr.baseType, substitution) &&
		arrayBound == arr.arrayBound &&
		( !enableUniv ||
		  (getCUniverse().equals(arr.getCUniverse()) &&
                       (elem_univ==null || elem_univ.equals(arr.elem_univ))) );
	}
    }


    /**
     * Calculates the hash code for this
     *
     * @return	the hash code for this, which is the same as the hash code for
     *		the corresponding CClass singleton * 31 + array bound
     */
    public int hashCode() {
	
	return baseType.hashCode() * 31 + arrayBound;
    }

    /**
     * Returns the set of types (excluding non-reference types) that
     * are directly visible from this type. */
    public Collection directlyVisibleTypes() {
	return Collections.EMPTY_SET;
    }

    /**
     * Indicates whether this type is concrete.  All types are
     * concrete except abstract class types and interface types.
     *
     * @return	whether this type is concrete */
    public boolean isConcrete() {
	return true;
    }

    /**
     * Returns the class object associated with this type.
     *
     * <pre><jml>
     * also 
     * requires isChecked();
     * </jml></pre>
     *
     *
     * @return the class object associated with this type
     */
    public CClass getCClass() {
	return CStdType.Object.getCClass();
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks that this type is valid, throwing an exception if it is
     * not.  For class types this check resolves type names into their
     * fully qualified versions and sets the class parameter so that
     * getCClass returns the appropriate singleton.
     *
     * @param		context		the context in which this type was
     *					declared, may be null
     * @exception	UnpositionedError	this unpositioned error
     *						should be positioned
     *						by the calling node of
     *						the AST, typically by
     *						calling the
     *						addPosition method of
     *						UnpositionedError 
     */
    public CType checkType( CContextType context ) throws UnpositionedError {
	if( !isChecked() ) {
	    setClass( CStdType.Object.getCClass() );
	    baseType = baseType.checkType( context );
	    context.registerVisibleType( this );
	}
	return this;
    }
    
    // ----------------------------------------------------------------------
    // BODY CHECKING
    // ----------------------------------------------------------------------

    public boolean changesByErasure() {
        return baseType.changesByErasure();
    }

    public CType getErasure() {
	if (baseType.isClassType()) {
	    return new CArrayType(baseType.getErasure(), arrayBound, universe);
	} else {
	    return this;
	}
    }
    
    public boolean isAlwaysAssignableTo(CType dest, boolean inst) {
	return isAlwaysAssignableTo(dest);
    }
    
    public boolean isAlwaysAssignableTo(CType dest, CClassType[] subst) {
	return isAlwaysAssignableTo(dest);
    }
    
    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2] [JLS2 5.1.4]
     * @param	dest		the destination type
     * @return true iff assignment is valid
     *
     * WMD: make sure that the universe modifiers are also assignable.
     */
    public boolean isAlwaysAssignableTo(CType dest) {
        if (dest.isCapture()) {
            // A type is a subtype of a capture type variable iff it is a
            // subtype of the capture's lower bound (see end of JLS3 4.10.2).
            CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
            if (lowerBound == null) {
                return false;
            } else {
                dest = lowerBound;
            }
        }

	if( dest.isArrayType() ) {
	    // Have two array types.  Would like to simply reduce the
	    // bound of each by one and make a recursive call.  But
	    // that isn't runtime-efficient.  So...
            
	    if( arrayBound == ((CArrayType)dest).arrayBound &&
		baseType.isAlwaysAssignableTo(((CArrayType)dest).baseType) ){
		
                // either Universe checks are disabled or they are
                // assignalbe to each other
                return !CTopLevel.getCompiler().universeChecks() ||
                (this.universe.isAlwaysAssignableTo(((CArrayType) dest).getCUniverse()) &&
                  ( !(this.baseType.isPrimitive() && this.arrayBound > 1) ||
                    this.getCUniverseElem().isAlwaysAssignableTo(((CArrayType) dest).getCUniverseElem())                             
                  ) ); 

	    } else {
		if (arrayBound < ((CArrayType)dest).arrayBound) {
		    return false;
		}
		if( ((CArrayType)dest).baseType.isClassType() &&
		    this.baseType.isClassType() ) {

		    // May be the unusal case Object[] o = new String[][]...;
		    return ((CArrayType)dest).baseType.equals(CStdType.Object, false, false) &&
                        (!CTopLevel.getCompiler().universeChecks() ||
                          (this.universe.isAlwaysAssignableTo(((CArrayType) dest).getCUniverse()) && // WMD
                          ((CClassType) baseType).universe.isAlwaysAssignableTo(((CClassType) ((CArrayType) dest).baseType).universe)));
		}

                if( this.baseType.isPrimitive() && this.arrayBound > 1 &&
                    ((CArrayType) dest).baseType.isClassType() &&
                    ((CArrayType)dest).baseType.equals(CStdType.Object, false, false ) &&
                    CTopLevel.getCompiler().universeChecks() &&
                    this.universe.isAlwaysAssignableTo(((CArrayType) dest).getCUniverse()) &&
                    this.elem_univ.isAlwaysAssignableTo( ((CClassType) ((CArrayType) dest).baseType).getCUniverse()) ) {
                        return true;
                }
                
                /* WMD: I would prefer if the following would be just:
		  return false;
		  but to keep compatible with Testcase test_badSpec_8
		  I kept it like this.
		  FIXME!
		 */
 		return (((CArrayType)dest).baseType).equals(CStdType.Object);
	    }
	} else {
	    if( dest.isClassType() ) {
		/* WMD first only compare the class and
		 * then compare the universes. */
		return ( dest.equals( CStdType.Object, false, false )
			 || dest.equals( CStdType.Cloneable, false, false )
			 || dest.equals( CStdType.Serializable, false, false ) ) &&
		    (!CTopLevel.getCompiler().universeChecks() ||
		     this.universe.isAlwaysAssignableTo(((CClassType)dest).getCUniverse())); // WMD
	    } else {
		return false;
	    }
	}
    }

    /** Forward the call, don't use the CClassType implementation.
     *
     */
    public boolean isAlwaysAssignableToIgnoreUniverses( CType dest ) {
    	return isAlwaysAssignableToIgnoreUniverses(dest, null);
    }

    /** A version of the Assignable check without considering the Universe types.
     * This is needed to determine e.g. the correct method to invoke, where we
     * do not look at the Universe types.
     * There were already many versions of "isAlwaysAssignableTo" and also with
     * boolean flags; therefore I decided to change the name.
     * By WMD.
     */
    public boolean isAlwaysAssignableToIgnoreUniverses( CType dest,
						    CClassType[] substitution) {
        if (dest.isClassTypeVariable()) {
            dest = substitution[((CTypeVariable)dest).getIndex()];
        }

        if (dest.isCapture()) {
            // A type is a subtype of a capture type variable iff it is a
            // subtype of the capture's lower bound (see end of JLS3 4.10.2).
            CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
            if (lowerBound == null) {
                return false;
            } else {
                dest = lowerBound;
            }
        }

	if( dest.isArrayType() ) {
	    // Have two array types.  Would like to simply reduce the
	    // bound of each by one and make a recursive call.  But
	    // that isn't runtime-efficient.  So...
	    if( arrayBound == ((CArrayType)dest).arrayBound &&
		baseType.isAlwaysAssignableTo(((CArrayType)dest).baseType) ) {
		return true;
	    } else {
		if (arrayBound < ((CArrayType)dest).arrayBound) {
		    return false;
		}
		if( ((CArrayType)dest).baseType.isClassType() &&
		    this.baseType.isClassType() ) {

		    // May be the unusal case Object[] o = new String[][]...;
		    return (((CArrayType)dest).baseType).equals(CStdType.Object, false, false);
		}
 		return (((CArrayType)dest).baseType).equals(CStdType.Object, false, false);
	    }
	} else {
	    if( dest.isClassType() ) {
                    return ( dest.equals( CStdType.Object, false, false )
			 || dest.equals( CStdType.Cloneable, false, false )
			 || dest.equals( CStdType.Serializable, false, false ) );
            } else {
		return false;
	    }
	}
    }
    
        
    /**
     * Can this type be converted to the specified type by casting conversion (JLS 5.5) ?
     * @param	dest		the destination type
     * @return	true iff the casting conversion is valid
     *
     * WMD: make sure the universe modifiers are also castable.
     */
    public boolean isCastableTo(CType dest) {
	// test for array first because array types are classes

	if (dest.isArrayType()) {
	    // optimizes recursive call with base types (indexed types)

	    CType       destBase = ((CArrayType)dest).baseType;
	    CUniverse   destUniv = ((CArrayType)dest).getCUniverse();
	    int		destBound = ((CArrayType)dest).arrayBound;

	    if (arrayBound == destBound) {
		if (baseType.isPrimitive()) {
		    return (baseType == destBase) &&
			(!CTopLevel.getCompiler().universeChecks() ||
			 getCUniverse().isAlwaysAssignableTo(destUniv) ||
			 // WMD: upcast of universe type
			 destUniv.isAlwaysAssignableTo(getCUniverse()));
		         // WMD: FIXME downcast: runtime checks or warning
		} else {
                    return (baseType.isCastableTo(destBase)) &&
			(!CTopLevel.getCompiler().universeChecks() ||
			 getCUniverse().isAlwaysAssignableTo(destUniv) ||
			 // WMD: upcast of universe type
			 destUniv.isAlwaysAssignableTo(getCUniverse()));
		    // WMD: downcast of universe type
		}
	    } else if (arrayBound < destBound) {
		return (baseType.isCastableTo(new CArrayType(destBase, destBound - arrayBound, destUniv))) &&
		    (!CTopLevel.getCompiler().universeChecks() ||
		     getCUniverse().isAlwaysAssignableTo(destUniv) ||
		     // WMD: upcast of universe type
		     destUniv.isAlwaysAssignableTo(getCUniverse()));
		     // WMD: FIXME downcast: runtime checks or warning
	    } else {
		// arrayBound > destBound
		return (new CArrayType(baseType, arrayBound - destBound, this.universe).isCastableTo(destBase)) &&
		    (!CTopLevel.getCompiler().universeChecks() ||
		     getCUniverse().isAlwaysAssignableTo(destUniv) ||
		     // WMD: upcast of universe type
		     destUniv.isAlwaysAssignableTo(getCUniverse()));
		     // WMD: FIXME downcast: runtime checks or warning
	    }
	} else if (dest.isClassType()) {
	    CUniverse   destUniv = ((CClassType)dest).getCUniverse();

	    if (dest.equals(CStdType.Object, false, false)) {
		// if T is a class type, then if T is not Object,
		// then a compile-time error occurs
		//		return true;
		return !CTopLevel.getCompiler().universeChecks() ||
			getCUniverse().isAlwaysAssignableTo(destUniv) ||
			// WMD: upcast of universe type
			destUniv.isAlwaysAssignableTo(getCUniverse());
		// WMD: downcast: runtime checks or warning
	    } else if (dest.getCClass().qualifiedName().equals(JAV_CLONEABLE)) {
		// if T is an interface type, then a compile-time error
		// occurs unless T is the interface type Cloneable
		// return true;
		return !CTopLevel.getCompiler().universeChecks() ||
			getCUniverse().isAlwaysAssignableTo(destUniv) ||
			// WMD: upcast of universe type
			destUniv.isAlwaysAssignableTo(getCUniverse());
		// WMD: downcast: runtime checks or warning
	    } else {
		return false;
	    }
	} else {
	    return false;
	}
    }

    /**
     * implicitlyCastTo
     * @param	dest	the expression type to cast to
     * @return	cast is valid ?
     */
    public boolean implicitlyCastTo(CType dest) {
	return isAlwaysAssignableTo(dest);
    }

    public boolean needsUncheckedConversion(CType dest) {
        if (!dest.isArrayType()) {
            return false;
        }

        return getBaseType().needsUncheckedConversion(((CArrayType) dest).getBaseType());
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private CType	baseType; //@ in thisClassInfo;
    private int		arrayBound;
    
    // The Universe modifier for an array of primitive type
    // with more than one dimension. Otherwise not used.
    // This modifier can not be stored in the baseType, because
    // primitive types do not take Universe modifiers.
    private CUniverse   elem_univ;
}
