/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package org.multijava.mjc;

import org.multijava.util.classfile.*;

/**
 * This class represents a synthetic method used in the implementation
 * of Java's assertion mechanism.  The method simply uses
 * java.lang.Class's forName method to convert a String into a Class
 * instance.  That instance will then be consulted to determine if assertions
 * are to be enabled or disabled for the associated class. */

public class CAssertHelperMethod extends CSourceMethod {
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Constructs an assertion helper method for the given method.
     * @param	owner	the receiver of the new method
     */
    public CAssertHelperMethod( CClass owner ) {
	super( new MemberAccess(owner, owner, ACC_STATIC),
	       JAssertStatement.assertSyntheticPrefix, CStdType.Class,
	       new CSpecializedType[1],
	       CClassType.EMPTY, CTypeVariable.EMPTY,
	       false, JBlock.DUM_BLOCK, null, null );
	    // the first instruction has to be the super send,
	    // so we set the appropriate parameters afterwards
	CSpecializedType[] parms = parameters();
	parms[0] = new CSpecializedType(CStdType.String);

	this.topConcreteMethod = this;
    }

    
    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Plants the bytecode for the body of this method.  Separated
     * from <code>genCode()</code> so subclasses may replace standard
     * body bytecode with specialized code.  
     *
     * @see #genCode()
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	    // invoke java.lang.Class.forName on the given String arg
	    // and return the result
	int startPC = code.getPC();
	code.plantInstruction( new LocalVarInstruction(opc_aload, 0) );
	CType[] parms = { CStdType.String };
	code.plantInstruction( new MethodRefInstruction(
				   opc_invokestatic, JAV_CLASS,
				   "forName",
				   CType.genMethodSignature( CStdType.Class,
							     parms ) ) );
	int endPC = code.getPC();
	code.plantInstruction( new NoArgInstruction( opc_areturn ) );
	int catchPC = code.getPC();
	
	    // now handle the potential ClassNotFoundException
	    // by throwing a NoClassDefFoundError with an appropriate message
	code.plantInstruction( new LocalVarInstruction(opc_astore, 1) );
	code.plantInstruction( new ClassRefInstruction(
				   opc_new,
				   JAV_NOCLASSDEFFOUND_ERROR )
	    );
	code.plantInstruction( new NoArgInstruction(opc_dup) );
	code.plantInstruction( new LocalVarInstruction(opc_aload, 1) );
	    // get the thrown ClassNotFoundException's String message
	code.plantInstruction( new MethodRefInstruction(
				   opc_invokevirtual,
				   JAV_CLASSNOTFOUND_EXCEPTION,
				   "getMessage",
				   CType.genMethodSignature( CStdType.String,
							     new CType[0] ) )
	    );
	CType[] parms2 = { CStdType.String };
	code.plantInstruction( new MethodRefInstruction(
				   opc_invokespecial,
				   JAV_NOCLASSDEFFOUND_ERROR,
				   JAV_CONSTRUCTOR,
				   CType.genMethodSignature( CStdType.Void,
							     parms2 ) ) );
	code.plantInstruction(new NoArgInstruction( opc_athrow ));
	    // add the catch clause to the exception table
	code.addExceptionHandler(new HandlerInfo(
				     code.getInstructionAt(startPC),
				     code.getInstructionAt(endPC),
				     code.getInstructionAt(catchPC),
				     JAV_CLASSNOTFOUND_EXCEPTION ));
    }
    

    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------

}
