/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package org.multijava.mjc;

import org.multijava.util.classfile.*;

/**
 * This class represents a static initializer that begins with code to
 * figure out if the associated class has assertions enabled or not.
 */

public class CAssertStaticInitMethod extends CSourceMethod {
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Constructs a new static initializer for a class, with code for
     * handling assertions.
     * @param	owner	the class to which the static initializer belongs
     * @param topType 	the top-level enclosing type of the static
     * 			initializer's owner
     * @param helperCls	the class containing helper methods for getting
     *			the topType's associated Class instance
     */
    public CAssertStaticInitMethod(CClass owner, CClass topType,
				   CClass helperCls) {
	super( new MemberAccess(owner, owner, ACC_STATIC), 
	       JAV_STATIC_INIT, CStdType.Void,
	       CSpecializedType.EMPTY, CClassType.EMPTY,
	       CTypeVariable.EMPTY, false,
	       JBlock.DUM_BLOCK, null, null );
	this.topType = topType;
	this.helperCls = helperCls;
	this.topConcreteMethod = this;
    }

    
    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    public void setOldStaticInit(CSourceMethod statInitMeth) {
	staticInit = statInitMeth;
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Plants the bytecode for the body of this method.  Separated
     * from <code>genCode()</code> so subclasses may replace standard
     * body bytecode with specialized code.  
     *
     * @see #genCode()
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	    // first load the surrounding top-level class's associated
	    // java.lang.Class instance from the helper class's appropriate
	    // synthetic field; if the field is null then first compute the
	    // java.lang.Class instance and store it in the field
	String helperClsName = helperCls.qualifiedName();
	String synthFieldName = JAssertStatement.assertSyntheticPrefix +
	    topType.getQualNameWithSeparator('$');
	code.plantInstruction(new FieldRefInstruction(
				  opc_getstatic, helperClsName,
				  synthFieldName,
				  CStdType.Class.getSignature() ));
	CodeLabel nonnullLabel = new CodeLabel();
	code.plantLabelRef(opc_ifnonnull, nonnullLabel);
	code.plantInstruction(new PushLiteralInstruction(
				  topType.getJavaName()));
	CType[] parms = { CStdType.String };
	code.plantInstruction(new MethodRefInstruction(
				  opc_invokestatic, helperClsName,
				  JAssertStatement.assertSyntheticPrefix,
				  CType.genMethodSignature( CStdType.Class,
							    parms ) ));
	code.plantInstruction( new NoArgInstruction(opc_dup) );
	code.plantInstruction( new FieldRefInstruction(
				   opc_putstatic, helperClsName,
				   synthFieldName,
				   CStdType.Class.getSignature() ));
	CodeLabel classInstanceLoadedLabel = new CodeLabel();
	code.plantLabelRef(opc_goto, classInstanceLoadedLabel);
	code.plantLabel(nonnullLabel);
	code.plantInstruction(new FieldRefInstruction(
				  opc_getstatic, helperClsName,
				  synthFieldName,
				  CStdType.Class.getSignature() ));

	    // now set the directly-enclosing class's $assertionsDisabled
	    // field according to whether or not assertions in the top-level
	    // enclosing class or disabled (using the new java.lang.Class
	    // method desiredAssertionStatus()
	code.plantLabel(classInstanceLoadedLabel);
	code.plantInstruction(new MethodRefInstruction(
				  opc_invokevirtual, JAV_CLASS,
				  "desiredAssertionStatus",
				  CType.genMethodSignature( CStdType.Boolean,
							    CType.EMPTY ) ));
	CodeLabel assertionsEnabledLabel = new CodeLabel();
	code.plantLabelRef(opc_ifne, assertionsEnabledLabel);
	code.plantInstruction(new PushLiteralInstruction(1));
	CodeLabel truthValueLoaded = new CodeLabel();
	code.plantLabelRef(opc_goto, truthValueLoaded);
	code.plantLabel(assertionsEnabledLabel);
	code.plantInstruction(new PushLiteralInstruction(0));
	code.plantLabel(truthValueLoaded);
	code.plantInstruction( new FieldRefInstruction(
				   opc_putstatic, owner().qualifiedName(),
				   JAssertStatement.assertDisabledName,
				   CStdType.Boolean.getSignature() ));

	    // if there is no other static init code, generate a 'return'
	    // stmt.  otherwise generate the rest of the code.
	if (staticInit == null) {
	    code.plantInstruction( new NoArgInstruction(opc_return) );
	} else {
	    staticInit.plantBodyBytecode(code);
	}	
    }
    

    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------

    /**
     * The top-level type enclosing the one containing this static initializer
     */
     private CClass topType;


    /**
     * The class containing helpers for getting the top-level type's Class
     * instance.
     */
     private CClass helperCls;

    
    /**
     * The existing static initializer being prepended to (if any).
     */
    private CSourceMethod staticInit;
    
}
