/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CBinaryClass.java,v 1.24 2006/07/16 20:35:14 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.*;

import org.multijava.util.compiler.UnpositionedError;

import java.util.Collection;
import java.util.Hashtable;
import java.util.ArrayList;
import org.multijava.util.InconsistencyException;

/**
 * This class represents a class read from a *.class file.  It is
 * primarily just a data structure, apart from methods for generating
 * the qualified name and for determining whether the member is
 * accessible from some class.
 *
 * @see	CMember
 */
public class CBinaryClass extends CClass {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a class export from file. */
    public CBinaryClass(Main compiler, ClassInfo classInfo) {
        this(compiler, classInfo, ClassCreator.getInstance(),
	     MjcSignatureParser.getInstance());
    }

    /**
     * Constructs a class export from file. */
    protected CBinaryClass(Main compiler, ClassInfo classInfo,
                           ClassCreator creator,
			   SignatureParser signatureparser) {
	super( compiler,
               creator.createMemberAccess(
                   getOuterClassFrom( classInfo.getName() ),
                   getHostFrom( classInfo.getName() ),
                   classInfo ),
	       classInfo.sourceFile(),
	       getIdentFrom( classInfo.getName() ),
	       classInfo.getName(),
	       //classInfo.getSuperClass() == null ?
	       //null : CTopLevel.getTypeRep( classInfo.getSuperClass(),true ),
	       null,
	       classInfo.isDeprecated() );
	registerInstanceWithCache();
	CClassType[] interfaces;
	String genericSignature = classInfo.getGenericSignature();
        // If we have a generic signature but Java generics are turned off in
        // the current compilation session, we simply discard the generic
        // signature and use the non-generic one instead.
	if ((genericSignature == null)
             // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
             || !CTopLevel.getCompiler().Generic(this)) {
	    setSuperClass(classInfo.getSuperClass() == null?
			  null:CTopLevel.getTypeRep(classInfo.getSuperClass(),true));
	    interfaces = loadInterfaces( classInfo.getInterfaces() );
	    setTypeVariables(CTypeVariable.EMPTY);
	} else {
	    SignatureParser.ClassSignature classsignature = signatureparser.parseClassSignature(genericSignature);
	    setSuperClass(classsignature.superType);
	    interfaces = classsignature.interfaces;
	    setTypeVariables(classsignature.typeVariables);//	SignatureParser.
	}
	//	CClassType[] interfaces = loadInterfaces( classInfo.getInterfaces() );

	FieldInfo[]	fields = classInfo.getFields();
	Hashtable	fields_H = null;
	CField[]	fields_V = null;

	// cook a context for use in resolving top methods
	CCompilationUnitContextType cuc = 
	    compiler.createCompilationUnitContext( null, 
						    (CCompilationUnit) host() );
	declContext = cuc.createBinaryClassContext( this );
	compiler.adoptCompilationUnitContext( cuc );

	try {
	    if (fields.length <= 10) {
		fields_V = new CField[fields.length];
		for (int i = 0; i < fields.length; i++) {
		    fields_V[i] = creator.createBinaryField(this, fields[i]);
		    CType fieldtype = fields_V[i].getType();
		    checkTypes(fields_V[i], fieldtype, declContext);
		}
	    } else {
		fields_H = new Hashtable(fields.length + 1, 1.0f);
		
		for (int i = 0; i < fields.length; i++) {
		    CField field = creator.createBinaryField(this, fields[i]);
		    CType fieldtype = field.getType();
		    checkTypes(field,fieldtype, declContext);
		    fields_H.put(field.ident(), field);
		}
	    }
	} catch(UnpositionedError e) {
	    throw new InconsistencyException("Error while reading class");
	}
	// cook a context for use in resolving top methods
	//	CCompilationUnitContextType cuc = 
	//    compiler.createCompilationUnitContext( null, 
	//					    (CCompilationUnit) host() );
	//declContext = cuc.createBinaryClassContext( this );
	//	compiler.adoptCompilationUnitContext( cuc );
	MethodInfo[] methods = classInfo.getMultimethods();
	CBinaryMethod[] methods_A = new CBinaryMethod[methods.length];
	try {
	    for (int i = 0; i < methods.length; i++) {
		methods_A[i] = creator.createBinaryMethod(this, methods[i],
							  declContext);
		CTypeVariable[] tvs = methods_A[i].getTypeVariable();
		for(int k = 0 ; k <tvs.length; k++ ) {
		    tvs[k].setMethodTypeVariable(true);
		}
		CContextType surroundingContext = createExtendedClassContext
		    (declContext, tvs);
		
		checkTypes(methods_A[i], surroundingContext);
	    }
	} catch (UnpositionedError e) {
	    throw new InconsistencyException("error while reading class");
	}
	
	processInnerClassesAttr(classInfo.getInnerClasses());

	setCheckedInterfaces( interfaces );
	setCheckedMembers( fields_V, fields_H, methods_A );
	
	// AHS: set the universe version of this class, this is null if the
	// class was not compiled with the universe type system.
	this.universeVersion = classInfo.getUniverseVersion();
	this.universeRuntimeSupport = classInfo.hasUniverseRuntimeSupport();
	
	// -------------------------------------------------------------
	// construct set of directly visible types
	
	// super class
	CClass sc = getSuperClass();
	if (sc != null) {
	    // not java.lang.Object
	    declContext.registerVisibleType(sc.getType());
	}
	
	// interfaces
	for (int i = 0; i < interfaces.length; i++) {
	    declContext.registerVisibleType(interfaces[i]);
	}
	
	// field types
	Collection fieldColl = fields();
	CField[] fieldArr = new CField[fieldColl.size()];
	fieldColl.toArray(fieldArr);
	for (int i = 0; i < fieldArr.length; i++) {
	    declContext.registerVisibleType(fieldArr[i].getType());
	}

	// types in method signatures
	for (int i = 0; i < methods_A.length; i++) {
	    methods_A[i].registerVisibleTypes();
	}
	
	close( methods_A, new CClassType[0] );
    }
    
    public CExtendedClassContext createExtendedClassContext(CContextType parent,
							    CTypeVariable[] tvs)
    {
	return new CExtendedClassContext(parent, tvs);
    }
    //---------------------------------------------------------------------
    // HELPER METHODS
    //---------------------------------------------------------------------

    /**
     * Adds this instance to the appropriate cache to prevent an
     * infinite regress if the initialization of this results in
     * another request to re-read the class file.  This method should
     * be called at the start of the constructor before any recursive
     * initialization the other types referenced by the class file.
     * This method should also be overridden as necessary in
     * subclasses so that their instances are added to the appropriate
     * cache.
     */
    protected void registerInstanceWithCache() {
	CTopLevel.addClass(this);
    }
    
    public void checkTypes(CMethod method, CContextType context) 
	throws UnpositionedError {
	((CBinaryMethod)method).checkTypes(context);
    }
    public void checkTypes(CField field, CType fieldtype, CContextType context) throws UnpositionedError{
	if(fieldtype instanceof CTypeVariableAlias ||
	   ((fieldtype instanceof CArrayType) && 
	    ((CArrayType)fieldtype).getBaseType() instanceof CTypeVariableAlias)){
	    ((CBinaryField)field).checkTypes(context);
	} 
    }
    /**
     * Loads the interfaces specified by the Strings in the argument
     * array (whether from other declarations in this compilation pass
     * or from *.class files.)
     *
     * @param	interfaces	the names of the interfaces to load
     * @return	an array of the types of the interfaces
     */
    protected CClassType[] loadInterfaces(String[] interfaces) {
	if (interfaces != null) {
	    CClassType[]	ret;
	    ret = new CClassType[interfaces.length];
	    for (int i = 0; i < interfaces.length; i++) {
		ret[i] = CTopLevel.getTypeRep(interfaces[i],true);
	    }
	    return ret;
	} else {
	    return CClassType.EMPTY;
	}
    }

    /**
     * Loads the inner classes specified by the info in the argument
     * array (whether from other declarations in this compilation pass
     * or from *.class files.)
     *
     * @param	inners		the info of the inner classes to load
     */
    protected final void processInnerClassesAttr(InnerClassInfo[] inners) {
	// the elements of inners fall into three categories: 
	// 1. nested and inner types of this, 
	// 2. self-referential nested type carrying additional modifier 
	//    information, and 
	// 3. other inner types visible within this, but whose
	//    information we currently ignore
	if (inners != null) {
	    ArrayList innerClassesAL = new ArrayList(inners.length);
	    for (int i = 0; i < inners.length; i++) {
		if (inners[i].getQualifiedName().equals(qualifiedName())) {
		    // category 2: self-referential nested type, grab
		    // its modifiers
		    access().setModifiers( access().modifiers() 
					   | inners[i].getModifiers() );
		    isAnonymous = inners[i].isAnonymous();
		    isMember = inners[i].isMember();
		} else if(inners[i].outerClassName().equals(qualifiedName())) {
		    // category 1: add to innerClasses of this
		    innerClassesAL.add(
			CTopLevel.getTypeRep( inners[i].getQualifiedName(),
					      true ));
		} // else category 3, ignored
	    }
	    CClassType[] innerClasses = new CClassType[innerClassesAL.size()];
	    innerClassesAL.toArray(innerClasses);
	    setInnerClasses( innerClasses );
	}
    }

    /**
     * Returns the surrounding class for the inner class named by the argument.
     *
     * @param	clazz		the name of the inner class
     * @return	the class representing the surrounding class
     */
    private static CClass getOuterClassFrom(String clazz) {
	int index = clazz.lastIndexOf("$");
	
	return index == -1 ? 
	    null : CTopLevel.loadClass( clazz.substring( 0, index ) );	
    }

    /**
     * Returns the host for the class named by the argument.
     *
     * @param	clazz		the name of the class
     * @return	the host for the named class
     */
    private static CMemberHost getHostFrom( String clazz ) {
	int index = clazz.lastIndexOf('/');
	return new CCompilationUnit( index == -1 ? "" : 
				     (clazz.substring(0, index) + "/") );
    }

    //---------------------------------------------------------------------
    // PUBLIC METHODS
    //---------------------------------------------------------------------

    /**
     * @return true if this class is anonymous
     */
    public boolean isAnonymous() {
	return isAnonymous;
    }

    /**
     * Indicates whether the class represented by this is a member
     * type.  A member type is a type that is declared inside another
     * type but that is not an anonymous class or a local type
     * declaration.  */
    public boolean isMember() {
	return isMember;
    }
    
    /**
     * Return the version of the encoding used to encode the universe
     * type modifiers.
     * By AHS.
     */
    public String getUniverseVersion() {
        return universeVersion;
    }
    
    /**
     * Return true if the class was compiled with run-time support
     * enabled.
     * By AHS.
     */
    public boolean hasUniverseRuntimeSupport() {
        return universeRuntimeSupport;
    }
    
    

    //---------------------------------------------------------------------
    // DATA MEMBERS
    //---------------------------------------------------------------------
    
    /**
     * Indicates whether the class represented by this is anonymous.  */
    private boolean isAnonymous;


    /**
     * Indicates whether the class represented by this is a member
     * type.  A member type is a type that is declared inside another
     * type but that is not an anonymous class or a local type
     * declaration.  */
    private boolean isMember;


    /**
     * This declaration context is used in top method searches.  It is
     * needed within CBinaryMethod instances and is stored here so
     * that it is available to subclasses that must instantiates such
     * instances.  */
    protected final CClassContextType declContext;
    
    

    /** AHS: The version of the encoding used to encode the universe type
     * modifiers.
     */
    private String universeVersion;
    
    /** AHS: indicates if the class has been compiled with runtime support
     * enabled.
     */
    private boolean universeRuntimeSupport;
}
