/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CBinaryMethod.java,v 1.27 2006/07/16 20:35:14 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.MethodInfo;
import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.MultimethodInfo;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a loaded (already compiled) class method.
 */
public class CBinaryMethod extends CMethod {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a method
     * @param   host            the host of this method
     * @param   methodInfo      a multimethod info from a class file
     * @param   declarationContext the context in which this method is declared
     */
    public CBinaryMethod( CClass host, MethodInfo methodInfo, 
                          CContextType declarationContext ) {

        this(new MemberAccess(buildReceiverClass( host, methodInfo ), 
                              host, methodInfo.getModifiers()),
             methodInfo, declarationContext);
    }

    /**
     * Constructs a method
     * @param   acc             the member access for this method
     * @param   methodInfo      a multimethod info from a class file
     * @param   declarationContext the context in which this method is declared
     */
    protected CBinaryMethod( MemberAccess acc, MethodInfo methodInfo, 
                             CContextType declarationContext ) {
        super( acc,
               methodInfo.getName(),
               // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
               buildReturnType( acc.owner(), methodInfo ),
               // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
               buildParameterTypes( acc.owner(), methodInfo ),
               buildExceptionTypes( methodInfo ),
               // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
               buildTypeParameter( acc.owner(), methodInfo ),
               methodInfo.isDeprecated(),
               declarationContext );

        setSynthetic(methodInfo.isSynthetic());

        if (methodInfo instanceof MultimethodInfo) { // !CONVERT!
            MultimethodInfo mmInfo = ((MultimethodInfo) methodInfo);
            this.functionNumber = mmInfo.functionNumber();
            String topConcreteMethodCollection = 
                mmInfo.topConcreteMethodCollection();
            if (topConcreteMethodCollection != null) {
                CGenericFunctionCollection tmc = 
                    CTopLevel.lookupGFCollection(topConcreteMethodCollection);
                tmc.applyAugmentationFor( declarationContext );
            }
        } else {
            this.functionNumber = Short.MAX_VALUE;
        }

	// AHS
	if (CTopLevel.getCompiler().universeBytecode() &&
            methodInfo.getUniverseAnnotation() != null &&
            methodInfo.getUniverseAnnotation().isPure()) {
                access().setModifiers(access().modifiers() | Constants.ACC_PURE);
	}
        
        registerNew(this);
    }

    private static CClass buildReceiverClass(CClass host, 
                                             MethodInfo methodInfo) 
    {
        CType.MethodSignature mSig = 
            CType.parseMethodSignature(getCorrectSignature(host, methodInfo));
        CClassType recType = mSig.receiverType;
        if (recType == null || 
            recType.qualifiedName().equals(host.qualifiedName())) {
            // either a regular Java method signature with no explicit
            // receiver, so use host; or the receiver is the same as
            // the host class that is currently being built, so use
            // host to avoid an infinite regress
            return host;
        } else {
            return recType.getCClass();
        }
    }

    /**
     * Simple convenience method which returns the appropriate method signature
     * stored in the given <code>methodInfo</code> depending on whether Java
     * generics are turned on or off in the current compilation session.
     *
     * @param methodInfo  The information representing this method as parsed
     *                    from a *.class file.
     * @return            The signature to be used for this method.
     */
    // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
    private static String getCorrectSignature(CClass host, MethodInfo methodInfo) {
        // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
        if (CTopLevel.getCompiler().Generic(host)) {
            return methodInfo.getGenericSignature();
        } else {
            return methodInfo.getSignature();
        }
    }

    // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
    private static CType buildReturnType(CClass host, MethodInfo methodInfo) {
	// AHS: call the extended method, with the universe annotation
        CType.MethodSignature mSig =
            // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
            CType.parseMethodSignature(getCorrectSignature(host, methodInfo),
                                       methodInfo.getUniverseAnnotation());
        
	return mSig.returnType;
    }

    // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
    private static CTypeVariable[] buildTypeParameter(CClass host, MethodInfo methodInfo) {
        CType.MethodSignature mSig =
            // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
            CType.parseMethodSignature(getCorrectSignature(host, methodInfo));
	return mSig.tvs;
    }

    // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
    private static CSpecializedType[] buildParameterTypes(CClass host, MethodInfo methodInfo) {
        // AHS: call the extended method, with the universe annotation
	CType.MethodSignature mSig = 
            // TODO[OJM]: Remove CClass parameter as soon as Java generics are supported!!!
	    CType.parseMethodSignature(getCorrectSignature(host, methodInfo),
	                               methodInfo.getUniverseAnnotation());

        //  If methodInfo is for internal method of external GF, then
        //  we need to yank out the synthetic receiver parameter.
        if (methodInfo.hasSyntheticReceiver()) {
            CSpecializedType[] allParms = mSig.parameterTypes;
            CSpecializedType[] nonReceiverParms = 
                new CSpecializedType[allParms.length - 1];
            System.arraycopy( allParms, 1, nonReceiverParms, 0, 
                              nonReceiverParms.length);
            return nonReceiverParms;
        } else {
            return mSig.parameterTypes;
        }
    }

    private static CClassType[] buildExceptionTypes(MethodInfo methodInfo) {
        String[] exceptions = methodInfo.getExceptions();

        if (exceptions == null) {
            return new CClassType[0];
        } else {
            CClassType[] types = new CClassType[exceptions.length];

            for (int i = 0; i < exceptions.length; i++) {
                types[i] = CTopLevel.getTypeRep( exceptions[i], true );
            }
            return types;
        }
    }

    /**
     * Registers the types visible in the signature of this.  */
    public void registerVisibleTypes() {
        declarationContext.registerVisibleType( receiverType() );
        CSpecializedType[] parms = parameters();
        for (int j = 0; j < parms.length; j++) {
            declarationContext.registerVisibleType(parms[j].dynamicType());
            declarationContext.registerVisibleType(parms[j].staticType());
        }
    }

    // -------------------------------------------------------------
    // CODE GENERATION
    // -------------------------------------------------------------
    
    /**
     * Generates bytecode to put a reference to the external
     * <code>function</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     *   also 
     *   requires this.isExternal();
     * </jml></pre> */
    public void plantFunctionRef( CodeSequence code ) {
        code.plantInstruction( 
                new MethodRefInstruction(opc_invokestatic,
                                         ((CClass) host()).qualifiedName(),
                                         // !FIXME!function
                                         "getFunction" + functionNumber,
                                          CType.genMethodSignature(
                                              dispatcherSignature().getType(),
                                              CType.EMPTY) ));
    }

    /**
     * Generates bytecode to put a reference to the
     * <code>oldFunction</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     *   also
     *   requires this.isExternal();
     * </jml></pre>
     * @param code the code sequence to add the bytecode to
     * @param forSuper if true, then this is for a super send, which,
     *                 under RMJ, should start from the oldSuperFunction
     *                 rather than oldFunction
     * @param fromOuter if true, then generates a reference that works from the
     *                  outer class context, otherwise generates a reference that
     *                  works from within the dispatcher class
     */
    public void plantOldFunctionRef( CodeSequence code, 
                                     boolean forSuper, boolean fromOuter ) {
        throw new RuntimeException( "Attempting to plant a reference to the " +
                                    "oldFunction field in a chain of " + 
                                    "responsibility.  However, this is the " +
                                    "last entry in the chain.  If this " +
                                    "is reached there is a bug in the code " +
                                    "generation of mjc." );
    }

    /**
     * Returns the $signature class for this.
     *
     * <pre><jml>
     *   also 
     *   requires this.isExternal();
     * </jml></pre> */
    public CDispatcherSignature dispatcherSignature() {
        try {
            CMethod top = topConcreteMethod();
            if (top != this) {
                return top.dispatcherSignature();
            }
            if( dispatcherSig != null ) {
                return dispatcherSig;
            }
        
            //@ assert top == this;
            final String qualifiedName = 
                ((CClass) host()).qualifiedName() + "$signature" + 
                functionNumber;
            final CClassType sigType = 
                CTopLevel.getTypeRep(qualifiedName, true);

            dispatcherSig = new CDispatcherSignature() {
                    public String qualifiedName() {
                        return qualifiedName;
                    }

                    public CClassType getType() {
                        return sigType;
                    }
                };

            return dispatcherSig;
        } catch (UnpositionedError e) {
            // ostensibly the call to topConcreteMethod() above can throw an
            // exception, however topConcreteMethod should have been resolved
            // before reaching this stage of execution (I think :-)
            throw new RuntimeException( e.getMessage() );
        }
    }

    /**
     * Returns the ambiguous-method dispatcher class for this.  (RMJ only.)
     *
     * <pre><jml>
     *   also
     *   requires this.isExternal();
     * </jml></pre> */
    public CAmbiguousDispatcherClass ambiguousDispatcherClass() {
        try {
            CMethod top = topConcreteMethod();
            if (top != this) {
                return top.ambiguousDispatcherClass();
            }
            if( ambiguousDispatcherClass != null ) {
                return ambiguousDispatcherClass;
            }

            // no properties actually needed...
            ambiguousDispatcherClass = new CAmbiguousDispatcherClass() { };

            return ambiguousDispatcherClass;
        
        } catch (UnpositionedError e) {
            // ostensibly the call to topConcreteMethod() above can throw an
            // exception, however topConcreteMethod should have been resolved
            // before reaching this stage of execution (I think :-)
            throw new RuntimeException( e.getMessage() );
        }
    }

    /**
     * Returns the $anchor class for this generic function.
     *
     * <pre><jml>
     * also
     * requires this.isExternal();
     * </jml></pre>
     *
     */
    public CClass anchorClass() {
        try {
            CMethod top = topConcreteMethod();
            if (top != this) {
                return top.anchorClass();
            }

            //@ assert top == this;
            return (CBinaryGFCollection) host();
        } catch (UnpositionedError e) {
            // ostensibly the call to topConcreteMethod() above can throw an
            // exception, however topConcreteMethod should have been resolved
            // before reaching this stage of execution (I think :-)
            throw new RuntimeException( e.getMessage() );
        }
    }

    /**
     * Returns the function field number for this generic function
     * within its $anchor class.
     *
     * <pre><jml>
     * also
     * requires this.isExternal();
     * </jml></pre>
     * */
    public int functionNumber() {
        return functionNumber;
    }

    // -------------------------------------------------------------
    // RESOLVE TOP METHODS
    // -------------------------------------------------------------
    // This is a hack to force the top methods of all binary methods
    // to be resolved.  The resolveTops method is called at the end of
    // every ResolveTopMethodTask to make sure no unresolved methods
    // remain.

    /**
     * Registers that the newly created binary method
     * <Code>self</code> has not had its top method resolved or its
     * receiver swallowed.  Called from the constructor of this
     * class. */
    private static void registerNew( CBinaryMethod self ) {
        if (topsNextPos >= unresolvedTops.length) {
            CBinaryMethod[] bigger = new CBinaryMethod[unresolvedTops.length*2];
            System.arraycopy(unresolvedTops, 0, bigger, 0, 
                             unresolvedTops.length);
            unresolvedTops = bigger;
        }
        unresolvedTops[topsNextPos] = self;
        topsNextPos++;

        if (unswNextPos >= unswallowed.length) {
            CBinaryMethod[] bigger = new CBinaryMethod[unswallowed.length*2];
            System.arraycopy(unswallowed, 0, bigger, 0, 
                             unswallowed.length);
            unswallowed = bigger;
        }
        unswallowed[unswNextPos] = self;
        unswNextPos++;
    }

    /**
     * Resolves the top methods of all binary methods.  */
    protected static void resolveTops() throws PositionedError {
        try {
            for (int i = 0; i < topsNextPos; i++) {
                unresolvedTops[i].topConcreteMethod();
            }
        } catch (UnpositionedError e) {
            throw e.addPosition(TokenReference.NO_REF);
        }
        
        // clear list
        topsNextPos = 0;
        if (unresolvedTops.length > TOPS_INIT_SIZE * 4) {
            unresolvedTops = new CBinaryMethod[TOPS_INIT_SIZE];
        }
        
    }
    
    public void checkTypes(CContextType context)  
        throws UnpositionedError
    {
        if(returnType() instanceof CTypeVariableAlias ||
           ((returnType() instanceof CArrayType) && 
            (((CArrayType)returnType()).getBaseType() instanceof CTypeVariableAlias))) {
            setReturnType(returnType().checkType(context));
        }
        CSpecializedType[] temp = parameters();
        for (int i = 0 ; i< temp.length; i++ ){
            CType t = temp[i].staticType();
            if(t instanceof CTypeVariableAlias ||
               ((t instanceof CArrayType) && 
                (((CArrayType)t).getBaseType() instanceof CTypeVariableAlias))) {
                t = t.checkType(context); 
                temp[i] = new CSpecializedType(t);
            }
            setParameters(temp);
        }
    }
    
    /**
     * Swallows the receivers of all external methods read from
     * bytecode.  Called immediately before code generation (synthetic
     * receivers are kept outside the parameter lists until then for
     * typechecking purposes. */
    protected static void swallowReceivers() {
        for (int i = 0; i < unswNextPos; i++) {
            CBinaryMethod u = unswallowed[i];
            try {
                if ( u.isInExternalGF() ) {
                    u.swallowReceiver();
                }
            } catch (UnpositionedError e) {
                throw new RuntimeException("unreachable");
            }
        }
        
        // clear list
        unswNextPos = 0;
        if (unswallowed.length > UNSW_INIT_SIZE * 4) {
            unswallowed = new CBinaryMethod[UNSW_INIT_SIZE];
        }
    }

    private static final int TOPS_INIT_SIZE = 50;
    private static int topsNextPos = 0;
    private static CBinaryMethod[] unresolvedTops = 
        new CBinaryMethod[TOPS_INIT_SIZE];
    
    private static final int UNSW_INIT_SIZE = 100;
    private static int unswNextPos = 0;
    private static CBinaryMethod[] unswallowed = 
        new CBinaryMethod[UNSW_INIT_SIZE];

    // -------------------------------------------------------------
    // PRIVILEGED DATA
    // -------------------------------------------------------------

    /**
     * The unique number of the overloaded external generic function
     * to which this belongs, or Short.MAX_VALUE if this does not
     * belong to an external generic function. */
    private final int functionNumber;

    private CDispatcherSignature dispatcherSig;

    private CAmbiguousDispatcherClass ambiguousDispatcherClass;
}
