/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CCORInitializer.java,v 1.3 2006/12/10 22:51:21 chalin Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import org.multijava.util.classfile.FieldRefInstruction;
import org.multijava.util.classfile.HandlerInfo;
import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.PushLiteralInstruction;

/**
 * This generates code to add external generic function
 * dispatcher classes to the chain of responsibility for the
 * appropriate generic function anchor class.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.3 $ */

public class CCORInitializer extends CSourceMethod {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Creates a new <code>CCORInitializer</code> instance.
     *
     * <pre><jml>
     *   requires owner != null;
     *   requires dispClasses != null && dispClasses.length > 0;
     * </jml></pre>
     *
     * @param owner             the class for which this is the static 
     *                          initializer
     * @param dispClasses       the nested dispatcher classes to be add to
     *                          the chains-of-responsibility (which would be
     *                          a great name for a band, with apologies to 
     *                          Dave Barry).
     */
    public CCORInitializer( CClass owner, 
                               CSourceDispatcherClass[] dispClasses ) {
        super( new MemberAccess(owner, owner, ACC_STATIC), initMethodName, 
               CStdType.Void, new CSpecializedType[0], CClassType.EMPTY, 
               CTypeVariable.EMPTY, false, null, null, null );
        this.dispClasses = dispClasses;
        
        this.isInitializedFlag = 
            new CSourceField(new MemberAccess(owner, ACC_STATIC | ACC_PRIVATE),
                             isInitializedFlagName, CStdType.Boolean, false );
                             
        this.isInitializedLock = 
            new CSourceField(new MemberAccess(owner, ACC_STATIC | ACC_PRIVATE),
                             isInitializedLockName, CStdType.Class, false );
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    CSourceField isInitializedFlag() {
        return isInitializedFlag;
    }
    
    CSourceField isInitializedLock() {
        return isInitializedLock;
    }
    
    //---------------------------------------------------------------------
    // DERIVED ATTRIBUTES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // INTERFACE CHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // TYPECHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    public ArrayList genMethodInfo() {
        ArrayList result = new ArrayList( 1 );
        result.add(nonEmptyMethodInfo( ident() ));
        return result;
    }

    /**
     * Plants the bytecode for the body of this method, including code
     * to initialize the chain-of-responsibility for the nested
     * dispatcher classes.  */
    protected void plantBodyBytecode( CodeSequence code ) {
        // Guard chains-of-responsibility initialization check to prevent 
        // multiple updates
        
        // see if lock object has been cached
        code.plantInstruction( 
            new FieldRefInstruction(opc_getstatic, owner().qualifiedName(),
                                    isInitializedLockName, 
                                    CStdType.Class.getSignature() ));
        CodeLabel afterLockObjectLookup = new CodeLabel();
        code.plantLabelRef(opc_ifnonnull, afterLockObjectLookup);
        
        // lock object not cached, so cache it
        code.plantInstruction(
            new PushLiteralInstruction(dispClasses[0].qualifiedName().replace('/', '.')));
		code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
							   "java/lang/Class",
							   "forName",
							   "(Ljava/lang/String;)Ljava/lang/Class;"));
        code.plantInstruction(new NoArgInstruction( opc_dup ));
        code.plantInstruction( 
            new FieldRefInstruction(opc_putstatic, owner().qualifiedName(),
                                    isInitializedLockName, 
                                    CStdType.Class.getSignature() ));
        // skip lock object reload, since one is on the stack
        CodeLabel afterLockObjectReload = new CodeLabel();
        code.plantLabelRef(opc_goto, afterLockObjectReload); 
        
        // reload the non-null lock object that was consumed from the stack
        code.plantLabel(afterLockObjectLookup);
        code.plantInstruction( 
            new FieldRefInstruction(opc_getstatic, owner().qualifiedName(),
                                    isInitializedLockName, 
                                    CStdType.Class.getSignature() ));
        code.plantLabel(afterLockObjectReload);
        
        // squirrel away the lock object for the monitor exit, and enter the 
        // monitor region
        code.plantInstruction(new NoArgInstruction( opc_dup ));
        JLocalVariable lockVar = 
			new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF, 
						     0, CStdType.Class, "localLock", null ) {
				public int getPosition() {
					// Just assume we can use the first slot.  This works
					// because we're done with the slot before any code that
					// might use it tries to do so.
					return 0;
				}
			};
        code.plantLocalVar(opc_astore, lockVar);
		code.plantInstruction(new NoArgInstruction(opc_monitorenter));
		int startPC = code.getPC(); // save start of critical region
        
        { // synchronized on Class object of first dispatcher
            // check whether the chains-of-responsibility have been initialized
            code.plantInstruction( 
                new FieldRefInstruction(opc_getstatic, owner().qualifiedName(),
                                        isInitializedFlagName, 
                                        CStdType.Boolean.getSignature() ));
            CodeLabel afterInit = new CodeLabel();
            code.plantLabelRef(opc_ifne, afterInit);
    
            { // Code guarded by ifInitializedFlag        
                // initialize the chains-of-responsibility
                for (int i = 0; i < dispClasses.length; i++) {
                    dispClasses[i].plantCORInitialization( code );
                }
                
                code.plantInstruction(new PushLiteralInstruction(1));
                code.plantInstruction( 
                    new FieldRefInstruction(opc_putstatic, 
                                            owner().qualifiedName(),
                                            isInitializedFlagName, 
                                            CStdType.Boolean.getSignature() ));            
            }
    
            code.plantLabel(afterInit);
        }
        
        // Exit critical region
		code.plantLocalVar(opc_aload, lockVar);
		code.plantInstruction(new NoArgInstruction(opc_monitorexit));
		int endPC = code.getPC();
		
		CodeLabel afterHandler = new CodeLabel();
		code.plantLabelRef(opc_goto, afterHandler);

		int	errorPC = code.getPC();
		code.plantLocalVar(opc_aload, lockVar);
		code.plantInstruction(new NoArgInstruction(opc_monitorexit));
		code.plantInstruction(new NoArgInstruction(opc_athrow));

		code.plantLabel(afterHandler);

		code.addExceptionHandler(new HandlerInfo(code.getInstructionAt(startPC),
								code.getInstructionAt(endPC),
								code.getInstructionAt(errorPC),
								null));
        
        code.plantInstruction(new NoArgInstruction(opc_return));
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * The dispatcher classes for which the chains-of-responsibility
     * need to be initialized.  */
    private CSourceDispatcherClass[] dispClasses;
    
    private final CSourceField isInitializedFlag;
    private final CSourceField isInitializedLock;

    public static final String initMethodName = "mjc$initializeCOR";
    public static final String isInitializedFlagName = "mjc$CORinitialized";
    public static final String isInitializedLockName = "mjc$CORinitializerLock";

}
