/*
 * Copyright (C) 2006 Swiss Federal Institute of Technology Zurich
 *
 * Part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2006 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CCaptureType.java,v 1.3 2008/03/16 15:52:41 wdietl Exp $
 */
package org.multijava.mjc;


/**
 * Type resulting from applying capture conversion (JLS3 5.1.10) to a wildcard.
 *
 * This type is basically a normal <code>CTypeVariable</code> whose upper bounds
 * are a combination of the wildcard's optional bound and the bounds of the
 * formal type variable for which the wildcard is a type argument. Additionally,
 * a type variable resulting from capture conversion may have a lower bound
 * associated to it which is used for checking subtype relationships.
 */
public class CCaptureType extends CTypeVariable {

    /**
     * The lower bound of this capture type variable or <code>null</code> if the
     * lower bound is of type <code>CNullType</code>.
     * <p>
     * See JLS3 5.1.10 for details on what the lower bound of a wildcard capture
     * type is.
     * </p>
     */
    private CClassType lowerBound = null;

    /**
     * Private constructor.
     *
     * Initializes this wildcard capture type by giving it a name and a set
     * of upper bounds.
     *
     * @param name         The name of the wildcard capture type.
     * @param upperBounds  The upper bounds of the wildcard capture type.
     * @param lowerBound   The lower bound of the wildcard capture type or
     *                     <code>null</code> if it has no lower bound.
     */
    private CCaptureType(
            /*@ non_null @*/ String name,
            /*@ non_null @*/ CClassType[] upperBounds,
            /*@ nullable @*/ CClassType lowerBound) {
        super(name, upperBounds);
        this.lowerBound = lowerBound;
    }

    /**
     * Creates a capture type for the given <code>wildcard</code> which must
     * have been type checked.
     * <p>
     * The returned wildcard capture type is guaranteed to be type checked.
     * </p>
     *
     * @param wildcard  The actual wildcard.
     * @return          The type resulting from applying capture conversion
     *                  on the given wildcard. Is guaranteed to be type checked.
     *                  Is never <code>null</code>.
     */
    public static /*@ non_null @*/ CCaptureType create(
            /*@ non_null @*/ CWildcardType wildcard) {
        // Construct the name of the capture type.
        String captureName = "capture-of " + 
        	(!(CTopLevel.getCompiler() != null &&
        		CTopLevel.getCompiler().universeChecks()) ? wildcard.toString() :
        				wildcard.toVerboseString());

        // Construct the upper bounds and the lower bound of the wildcard
        // capture type.
        CTypeVariable typeVariable = wildcard.getTypeVariable();
        CClassType[] variableBounds = typeVariable.getBounds();
        CClassType[] captureBounds;
        CClassType captureLowerBound;
        if (wildcard.getBound() == null) {
            // An unbounded wildcard has the same upper bounds as the formal
            // type variable.
            captureBounds = new CClassType[variableBounds.length];
            System.arraycopy(variableBounds, 0, captureBounds, 0, variableBounds.length);
            // No lower bound (means CNullType as lower bound).
            captureLowerBound = null;
        } else if (wildcard.isUpperBounded()) {
            // An upper bounded wildcard is a bit more complicated.
            CClassType wildcardBound = wildcard.getBound();
            if (variableBounds.length == 0) {
                // If the formal type variable has no upper bounds, the only
                // bound is that of the wildcard.
                captureBounds = new CClassType[] { wildcardBound };
            } else {
                // The formal type variable does have some upper bounds.
                if (wildcardBound.isInterface()) {
                    // If the wildcard bound is an interface, we just append it
                    // to the bounds of the formal type variable.
                    captureBounds = new CClassType[variableBounds.length + 1];
                    System.arraycopy(variableBounds, 0, captureBounds, 0, variableBounds.length);
                    captureBounds[captureBounds.length - 1] = wildcardBound;
                } else {
                    // The wildcard bound is a class.
                    if (variableBounds[0].isInterface()) {
                        // If the formal type variable has no class in its upper
                        // bounds, we prepend the wildcard bound class.
                        captureBounds = new CClassType[variableBounds.length + 1];
                        System.arraycopy(variableBounds, 0, captureBounds, 1, variableBounds.length);
                        captureBounds[0] = wildcardBound;
                    } else {
                        // If the formal type variable does have a class in its
                        // upper bounds, we replace it by the most specific
                        // class.
                        captureBounds = new CClassType[variableBounds.length];
                        System.arraycopy(variableBounds, 0, captureBounds, 0, variableBounds.length);
                        captureBounds[0] = wildcardBound;
                        // If the bounds class of the type parameter is more
                        // specific than the upper bound of the wildcard (that's
                        // allowed by the JLS3!), we set it as the bounds class
                        // of the capture type variable.
                        if (variableBounds[0].isAlwaysAssignableTo(wildcardBound)) {
                            captureBounds[0] = variableBounds[0];
                        }
                    }
                }
            }
            // No lower bound (means CNullType as lower bound).
            captureLowerBound = null;
        } else {
            // A lower bounded wildcard has the same upper bounds as the formal
            // type variable.
            captureBounds = new CClassType[variableBounds.length];
            System.arraycopy(variableBounds, 0, captureBounds, 0, variableBounds.length);
            // Same lower bound as the wildcard's lower bound.
            captureLowerBound = wildcard.getBound();
        }

        CCaptureType captureType =
            new CCaptureType(captureName, captureBounds, captureLowerBound);

        // Set the CClass of the capture type variable.
        if (captureType.getBounds().length == 0) {
            captureType.setClass(CStdType.Object.getCClass());
        } else {
            captureType.setClass(captureType.getBounds()[0].getCClass());
        }

        return captureType;
    }

    /**
     * Returns the lower bound of this type variable or <code>null</code> if
     * the type variable has no lower bound.
     *
     * @return  The lower bound of this type variable or <code>null</code> if
     *          the type variable has no lower bound.
     */
    public /*@ pure @*/ /*@ nullable @*/ CClassType getLowerBound() {
        return lowerBound;
    }

    public boolean isTypeVariable() {
        return true;
    }

    public boolean isClassTypeVariable() {
        return false;
    }

    public boolean isMethodTypeVariable() {
        return false;
    }

    public boolean isCapture() {
        return true;
    }

    public String toString() {
        return getIdent();
    }
}
