/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClass.java,v 1.100 2008/03/17 16:59:46 wdietl Exp $
 */

package org.multijava.mjc;

import java.io.File;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import org.multijava.util.classfile.*;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.FastStringBuffer;
import org.multijava.relaxed.util.RMJAnnotation;


import org.multijava.util.Destination;

/**
 * This class represents a class declaration in the signature hierarchy.
 *
 * @see CMember
 */
public abstract class CClass extends CMember 
    implements CMemberHost, Comparable
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a class signature.
     *
     * <pre><jml>
         ensures owner != null ==> owner == host();
         ensures owner == null ==> host().isCompilationUnit();
     * </jml></pre>
     */
    public CClass( Main compiler,
                   MemberAccess access,
                   File sourceFile,
                   String ident,
                   String qualifiedName,
                   CClassType superClass,
                   boolean deprecated )
    {
        super(access, ident, deprecated);

        this.compiler = compiler;
        this.sourceFile = sourceFile;
        this.qualifiedName = qualifiedName;
        this.superClass = superClass;
    }

    /**
     * Sets the interfaces extended or implemented by this type.
     * Called during checkInterface pass <em>before checking the
     * signatures of members</em>, since the super-interfaces must be
     * known to resolve type signatures of members.  Subtyping
     * relationships can be tested after this information has been set
     * (for both classes and all their ancestors).
     *
     * <pre><jml>
        requires interfaces != null && 
                 (\forall int i; 0 <= i && i < interfaces.length;
                          interfaces[i].isChecked());
     * </jml></pre>
     *
     */
    public void setCheckedInterfaces( CClassType[] interfaces ) {
        this.interfaces = interfaces;
        assertTrue(interfaces != null);
    } 
    
    /**
     * Stores information on member signatures gathered during
     * checkInterface pass prior to closing in the translateMJ
     * pass.  Method and field lookup is possible after this method
     * has been called, however, code generation must wait until after
     * the <code>close</code> method has been called. 
     *
     * <pre><jml>
     * requires (fields_A != null || fields_H != null);
     * requires protoMethods_A != null;
     * ensures observerSet != null ==>
     *         (\forall Observer o; observerSet.contains(o); 
     *                              (* calls o.signatureCompleted(this) *) );
     * </jml></pre>
     *
     */
    public final void setCheckedMembers( CField[] fields_A,
                                         Hashtable fields_H,
                                         CMethod[] protoMethods_A ) 
    {
        this.fields = new CFieldTable(fields_A, fields_H);
        this.protoMethods = new CMethodSet( protoMethods_A );
        if (observerSet != null) {
            Iterator i = observerSet.iterator();
            while (i.hasNext()) {
                ((Observer) i.next()).signatureCompleted(this);
            }
        }
    }

    /**
     * Ends the definition of this class by adding the
     * dispatcher-wrapped methods and inner-dispatcher classes.  
     *
     * <pre><jml>
     * requires methods_A != null && dispClassTypes != null;
     * assignable methods, innerClasses, this.dispClassTypes;
     * ensures (* getInnerClasses() will return the union of 
     *            \old(getInnerClasses()) and dispClassTypes *);
     *                  // !FIXME! Why does JML checker require the signature
     *                  // for setInnerClasses here even though there is no 
     *                  // overloading?
     * callable getInnerClasses, setInnerClasses(CClassType[]);
     * </jml></pre>
     *
     */
    public void close( CMethod[] methods_A, CClassType[] dispClassTypes )
    {
        // cache the dispatcher classes for generating bytecode
        // signature attributes
        this.dispClassTypes = dispClassTypes;

        this.methods = new CMethodSet( methods_A );

        // add the wrapper methods
        CAbstractMethodSet.Iterator i = wrapperMethods.iterator();
        while( i.hasNext() )
            this.methods.addMethod(i.next());

        // add the field getter methods
        i = fieldGetterMethods.iterator();
        while( i.hasNext() )
            this.methods.addMethod(i.next());

        // add the field setter methods
        i = fieldSetterMethods.iterator();
        while( i.hasNext() )
            this.methods.addMethod(i.next());

        // add the redirector methods
        i = redirectorMethods.iterator();
        while( i.hasNext() )
            this.methods.addMethod(i.next());

        // add the assert methods
        if (assertHelperMethod != null) {
            this.methods.addMethod(assertHelperMethod);
        }
        if (assertStaticInitMethod != null) {
                // replace the existing static initializer
            i = methods.iterator();
            while ( i.hasNext() ) {
                CMethod currMeth = i.next();
                if (currMeth.ident().equals(JAV_STATIC_INIT)) {
                    assertStaticInitMethod.setOldStaticInit(
                        (CSourceMethod) currMeth );
                    i.remove();
                    break;
                }
            }
            this.methods.addMethod(assertStaticInitMethod);
        }

        CClassType[] oldInnerClasses = getInnerClasses();
        CClassType[] updatedInnerClasses = 
            new CClassType[ oldInnerClasses.length + dispClassTypes.length ];
        System.arraycopy( oldInnerClasses, 0, updatedInnerClasses, 0, 
                          oldInnerClasses.length );
        System.arraycopy( dispClassTypes, 0, updatedInnerClasses, 
                          oldInnerClasses.length, dispClassTypes.length );
        setInnerClasses( updatedInnerClasses );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------
    /**
     * Returns the compiler 
     */
    public /*@ pure @*/ Main getCompiler() {
        return compiler;
    }

    /**
     * @return  the interface
     */
    public /*@ pure @*/ CClass getCClass() {
        return this;
    }

    /**
     * set the type variables of the class 
     */
    public void setTypeVariables(CTypeVariable[] typevariables){
        this.typevariables = typevariables;
    }
    /**
     * return the type variables of the class 
     */
    public CTypeVariable[] getTypeVariables(){
        return typevariables;
    }
    /**
     * Indicates whether this host is a class.
     *
     * @return  true iff this host represents a class
     */
    public /*@ pure @*/ boolean isClass() { return true; }

    /**
     * Indicates whether this host is a nested type, i.e., a member
     * type.  */
    public /*@ pure @*/ boolean isNestedType() {
        return owner() != null;
    }

    /**
     * Indicates whether this host is an inner class, i.e., a
     * non-static member class.  */
    public /*@ pure @*/ boolean isInnerClass() {
        return isNestedType() && !isStatic();
    }

    /**
     * Indicates whether this host is a compilation unit
     *
     * @return  true iff this host represents a compilation unit
     */
    public /*@ pure @*/ boolean isCompilationUnit() { return false; }

    /**
     * @return  true if this member is abstract
     */
    public /*@ pure @*/ boolean isAbstract() {
        return access().isAbstract();
    }

    /**
     * @return true if this class is anonymous
     */
    public abstract /*@ pure @*/ boolean isAnonymous();

    /**
     * Indicates whether the class represented by this is a member
     * type.  A member type is a type that is declared inside another
     * type but that is not an anonymous class or a local type
     * declaration.  */
    public abstract /*@ pure @*/ boolean isMember();

    /**
     * @return true if this class is an interface
     */
    public /*@ pure @*/ boolean isInterface() {
        return access().isInterface();
    }


        /**
         * Records the fact that even though the class is an inner
         * class, we should not generate an outer this for it, because
         * the outer this is not accessible.
         */
    public void syntheticOuterThisInaccessible() {
        syntheticOuterThisInaccessible = true;
    }
    
    /**
     * Indicates whether the class represented by this should have a
     * synthetic outer this field.  That is true if the class is an inner
     * class and the outer this is accessible.
     */
    public boolean hasSyntheticOuterThis() {
        return isInnerClass() && !syntheticOuterThisInaccessible;
    }
    
    /**
     * Returns the supertype of this class or <code>null</code> if this is
     * the <code>java.lang.Object</code> class.
     *
     * @return  The supertype of this class or <code>null</code> if this is
     *          the <code>java.lang.Object</code> class.
     */
    public /*@ pure @*/ /*@ nullable @*/ CClassType getSuperType() {
        return superClass;
    }

    /**
     * @return an array of the interfaces implemented by this class
     */
    //@ ensures \result != null;
    public /*@ pure @*/ CClassType[] getInterfaces() {
        return interfaces;
    }
    
    /**
     * @return the full name of this class
     */
    public /*@ pure @*/ String qualifiedName() {
        return qualifiedName;
    }

    /**
     * @return the name of the package this class belongs to
     */
    public /*@ pure @*/ String packageName() {
        return host().packageName();
    }

    /**
     * Returns the super class of this class
     */
    public /*@ pure @*/ CClass getSuperClass() {
        return superClass == null ? null : superClass.getCClass();
    }

    public String getGenericSignature() {
        FastStringBuffer        buffer;
        String          result;
        
        buffer = CType.StringBuffers.getStringBuffer();
        if (typevariables.length > 0) {
            buffer.append('<');
            for (int i = 0; i < typevariables.length; i++) {
                typevariables[i].appendDefinitionSignature(buffer);
            }
            buffer.append('>');
        }
        superClass.appendGenericSignature(buffer);
        for (int i = 0; i < interfaces.length; i++) {
            interfaces[i].appendGenericSignature(buffer);
        }
        result = buffer.toString();
        CType.StringBuffers.release(buffer);
        return result;
    }
    
    /**
     * Sets the super class of this class
     */
    public void setSuperClass(CClassType superClass) {
        this.superClass = superClass;
    }

    /**
     * Returns the type of this class.  Ensures that the type has been
     * checked. */
    public CClassType getType() {
        if (type == null) {
            type = CTopLevel.getTypeRep(qualifiedName(),true);
            type.setAllArguments(new CClassType[][]{typevariables});
            type.setClass( this );
        } else {
            type.setAllArguments(new CClassType[][]{typevariables});
        }

        return type;
    }

    /**
     * @return the source file of this class
     */
    public /*@ pure @*/ File sourceFile() {
        return sourceFile;
    }

    /**
     * Returns the signature of the field with the given name declared
     * in this class, or null if this class does not declare a field
     * with the given name. */
    protected /*@ pure @*/ CField getDeclaredField( String ident) {
        return fields.lookup(ident);
    }

    /** Returns a Collection of CField values, one for each field in this class. */
    public /*@ pure @*/ Collection fields() {
        return fields.fields();
    }
    
    /** Returns the number of fields in this class. */
    public /*@ pure @*/ int getFieldCount() {
        return fields.size();
    }

    /**
     * Returns an array of the types of the inner classes of the class
     * represented by this.
     * 
     * <pre><jml>
     * ensures \result != null;
     * </jml></pre>
     * */
    public /*@ pure @*/ CClassType[] getInnerClasses() {
        return innerClasses != null ? innerClasses : new CClassType[0];
    }


    /**
     * End of first pass, we need the inner classes's signatures
     */
    //@ assignable innerClasses;
    public void setInnerClasses(CClassType[] inners) {
        this.innerClasses = inners;
    }

    public String toString() {
        return qualifiedName();
    }

    /**
     * Compares this to a given object.
     *
     * <pre><jml>
       also
         requires this.equals(o);
         ensures \result == 0;
       also
         requires o != null && !(o instanceof CClass);
         signals_only ClassCastException;
     * </jml></pre>
     *
     *
     * @param o an <code>Object</code> value to compare to this
     * @return 0 if this.equals(o)
     * @exception ClassCastException if <code>o</code> is incomparable to this
     */
    public /*@ pure @*/ int compareTo(/*@ non_null @*/ Object o) throws ClassCastException {
        CClass other = (CClass) o;
        int res = qualifiedName().compareTo(other.qualifiedName());
        if (res != 0) {
            return res;
        }
        return System.identityHashCode(this) - System.identityHashCode(other);
    }

    /**
     * Compares the fully qualified name of <code>this</code> with the
     * fully qualified name of <code>o</code>.
     * @exception ClassCastException if <code>o</code> is not an instance of <code>CClass</code> 
     *
     * <pre><jml>
      also
         requires (o instanceof CClass);
         ensures this.qualifiedName().equals(((CClass)o).qualifiedName());
       also
         requires !(o instanceof CClass);
         signals_only ClassCastException;
     * </jml></pre>
     *
     *
     */
    public /*@ pure @*/ boolean equals(/*@ nullable @*/  Object o )
        throws ClassCastException
    {
        CClass other = (CClass) o;
        return qualifiedName().equals(other.qualifiedName());
    }

    // doc comment and specification inherited
    public int hashCode() {
        // !FIXME! the following doesn't seem right,
        // but is relied on by JML refinement code
        return super.hashCode();
    }

    /**
     * Indicates whether this host is a subclass of the given class,
     * where "subclass" is the reflexive, transitive closure of the
     * extends relation.
     *
     * @param   maybeSuper      the potential superclass
     */
    public /*@ pure @*/ boolean descendsFrom( CClass maybeSuper ) {
    	// Check for equality, there may be multiple instances in JML
    	// for some weird reason
    	if (equals(maybeSuper)) {
            return true;
        } else if (superClass == null) {
            // java/lang/object
            return false;
        } else if (getSuperClass().descendsFrom(maybeSuper)) {
            return true;
        } else {
            if (interfaces != null) {
                for (int i = 0; i < interfaces.length; i++) {
                    if (interfaces[i].getCClass().descendsFrom(maybeSuper)) {
                        return true;
                    }
                }
            }

            return false;
        }
    }

    /**
     * <p>
     * Returns whether the given <code>maybeSuper</code> type is a supertype
     * of the <code>maybeSub</code> type. These <code>CClassType</code>s are
     * thereby necessary when checking for a subtype relationship on generic
     * classes since in such cases the actual type arguments are required in
     * order to check for the subtype relationship. If the classes involved
     * are not generic, the overloaded method <code>descendsFrom(CClass)</code>
     * may be used instead.
     * </p>
     * <p>
     * Note that when invoking this method from outside this class, the
     * <code>maybeSub</code> type should be a type of this class.
     * </p>
     *
     * @param maybeSuper The potential supertype.
     * @param maybeSub The potential subtype.
     * @param enableUniv Take universe modifier into account iff true.
     * @return Whether <code>maybeSuper</code> is a supertype of
     *         <code>maybeSub</code>.
     */
    public boolean descendsFrom(CClassType maybeSuper,
    		CClassType maybeSub,
    		boolean enableUniv,
    		boolean allowUniversesCovariance) {
    	
    	// Check universe modifier if desired
        if (enableUniv&& !maybeSub.getCUniverse()
        		.isAlwaysAssignableTo(maybeSuper.getCUniverse())) {
            return false;
        }

        // Even if both classes are the same, we still have to check for the
        // generic type arguments, if any.
        if (maybeSuper.getCClass() == this) {
            // If this is no generic class or any of the types involved is a
            // raw type, then we are done and the subtype relationship holds.
            if (!isGenericClass() || maybeSuper.isRawType() || maybeSub.isRawType()) {
                return true;
            }

            // Check whether the type arguments match.
            CClassType[] maybeSuperArgs = maybeSuper.getArguments();
            CClassType[] maybeSubArgs = maybeSub.getArguments();

            assertTrue(maybeSuperArgs.length == maybeSubArgs.length);

            // Check for type argument containment (JLS3 4.5.1.1) for the
            // subtype relationship between parametrized types (JLS3 4.10.2).
            for (int i = 0; i < maybeSuperArgs.length; i++) {
                if (!maybeSuperArgs[i].contains(maybeSubArgs[i],
                		enableUniv,
                		allowUniversesCovariance)) {
                	return false;
                }
            }
            return true;
        }

        // All types descend from any Object
        if (maybeSuper.getCClass().equals(CStdType.Object.getCClass())
        		&& (!enableUniv
        		|| CUniverseReadonly.getUniverse()
        				.isAlwaysAssignableTo(maybeSuper.getCUniverse()))) {
        	return true;
        }
        
        // If the types are not the same (previous check) and this type is
        // java.lang.Object, this type cannot descend from any other type.
        if (getSuperClass() == null) {
            return false;
        }

        // If the classes are not the same nor is this type at the top of
        // the hierarchy tree, we simply call CClass.descendsFrom() recursively
        // on all super types of this type.

        // WMD wanted to change the below to take the type arguments into account.
        
        // Recurse on the super class.
        if (getSuperClass().descendsFrom(maybeSuper.getCClass())) {
            return true;
        }
        // Recurse on all super interfaces.
        if (interfaces != null) {
            for (int i = 0; i < interfaces.length; i++) {
                if (interfaces[i].getCClass().descendsFrom(maybeSuper.getCClass())) {
                    return true;
                }
            }
        }
        
        /* WMD wrote some examples, where something closer to the below would
         * be better. However, just taking this breaks "mjc/testcases/gj/rule2"
         
        // Recurse on the super class.  
        if ( getSuperClass().descendsFrom(maybeSuper,
        		maybeSub.getSuperClassType(),
        		enableUniv,
        		allowUniversesCovariance) ) {
            return true;
        }
        // Recurse on all super interfaces.
        CClassType[] sinterf = maybeSub.getSuperInterfaceTypes();
        
        if (sinterf != null) {
            for (int i = 0; i < sinterf.length; i++) {
                if (sinterf[i].getCClass().descendsFrom(maybeSuper,
                		sinterf[i],
                		enableUniv,
                		allowUniversesCovariance) ) {
                    return true;
                }
            }
        }*/
        
        return false;
    }

    /**
     * Returns true if this host is a subclass of the given class or
     * if some host of this host is a subclass of the given class.
     *
     * @param maybeOuterSuper the potential superclass */
    public /*@ pure @*/ boolean nestedDescendsFrom( CClass maybeOuterSuper ) {
        return descendsFrom( maybeOuterSuper ) ||
            host().nestedDescendsFrom( maybeOuterSuper );
    }

    /**
     * Indicates whether this is defined inside the given host.
     *
     * @param   maybeOuter      the host that this may be inside
     * @return  true if the given host is a (possibly transitive) host of this
     */
    public /*@ pure @*/ boolean isDefinedInside( CMemberHost maybeOuter) {
        if( this == maybeOuter ) {
            return true;
        } else {
            return host().isDefinedInside( maybeOuter );
        }
    }

    /**
     * Returns the set of types (excluding non-reference types) that
     * are directly visible from this class.  The directly visible
     * types are those referenced in the signature of this class. */
    public /*@ pure @*/ Collection directlyVisibleTypes() {
        return getCCompilationUnit().directlyVisibleTypes();
    }

    /**
     * Returns the data structure representing the methods that will
     * be used for generating bytecode.  The returned data structure
     * is not cloned but is aliased by this; handle with care.
     *
     * @see #methods */
    protected /*@ pure @*/ CMethodSet methodsForCodeGen() {
        return methods;
    }
    
    /**
     * Returns the data structure representing the methods that will
     * be used for generating bytecode.  The returned data structure
     * is not cloned but is aliased by this; handle with care.
     *
     * @see #methods */
    public /*@ pure @*/ CMethodSet methods() { return methods; }


    /**
     * Returns the data structure representing the methods of this
     * class as generated from the declarations after the
     * checkInterface pass but before dispatcher wrapping.  The
     * returned data structure is not cloned but is aliased by this;
     * thus, it should be handled with care.
     * dispatcher wrapping.
     *
     * @see #protoMethods */
    public /*@ pure @*/ CMethodSet protoMethods() { 
         return protoMethods;
     }

    /**
     * Returns the data structure representing the fields that will be
     * used for generating bytecode.  The returned data structure is
     * not cloned but is aliased by this; handle with care.
     *
     * @see #fields */
    protected /*@ pure @*/ CFieldTable fieldsForCodeGen() {
        return fields;
    }

        // Returns the next access$ method name to use.
    public static String getNextAccessName() {
        return "access$" + ++accessID;
    }

    /** Initializes any non-final static items. */
    public static void initSession() {
        accessID = 0;
    }

    
    /**
     * Returns a unique int for use in naming synthetic classes
     * (e.g. anonymous classes).
     * @return an unique int value incrementaly
     */
    public int getNextSyntheticIndex() {
        return ++syntheticIndex;
    }


        /**
         * Set the synthetic helper method used to implement asserts.
         */
    public void setAssertHelperMethod(CMethod m) {
        assertHelperMethod = m;
    }

        /**
         * Set the static initializer for handling asserts.  If there already
         * exists a static initializer method for this class, remove it.
         */
    public void setAssertStaticInitMethod(CAssertStaticInitMethod m) {
        assertStaticInitMethod = m;
    }
    
        /** Returns a boolean indicating whether or not assertions
            have been found in this class as of yet. */
    public /*@ pure @*/ boolean noAssertionsYet() {
        return fields.lookup(JAssertStatement.assertDisabledName) == null;
    }

        /** Returns a boolean indicating whether or not assertions
            have been found in this class or any inner classes as of yet. */
    public /*@ pure @*/ boolean noPossiblyInnerAssertionsYet() {
        return assertHelperMethod == null;
    }

    
    
    // ----------------------------------------------------------------------
    // LOOKUP
    // ----------------------------------------------------------------------

    /** This class is a specialization of CMethodSet.Strategy that
     *  keeps track of which CClass instances have already been
     *  searched for appropriate methods.  The various lookup methods
     *  below will create appropriate subclasses of this class, using it
     *  to ensure that they do not search a class more than once. */
    protected abstract class NoDupStrategy implements CMethodSet.Strategy {
        protected Set already_searched = new HashSet();
        public boolean alreadySearched( CClass cls ) {
            return !(already_searched.add(cls));
        }
    }
    
    /**
     * Searches for a class with the given simple name according the
     * procedure in JLS2 6.5.5.  This method implements the second and
     * third bullet points of the procedure.  That is "Otherwise, if
     * the simple type name occurs within the scope of exactly one
     * visible member type, then the simple type name denotes that
     * member type;" and "Otherwise, if the simple type name occurs
     * within the scope of more than one visible member type, then the
     * name is ambiguous as a type name; a compile-time error occurs."
     * 
     * @param   name    the class name, without qualifiers
     * @param   context         the referencing context, used for accessibility
     *                          checks
     * @return  the class if found, null otherwise
     * @exception UnpositionedError if search fails */
    /* !FIXME! The context may be null, because this method may be
       called from CClassNameType.checkType, which allows null
       contexts.  In that case, the code below disables the access
       checks on the class being looked up.  If null contexts are
       meaningful, we should figure out the right access checks in
       that case.  Otherwise, we should fix things so that null
       contexts are never sent to checkType.  */
    public CClass lookupClass( String name, CContextType context ) 
        throws UnpositionedError
    {
        ArrayList possibleClasses = new ArrayList();

        // See if there is a local inner class with the given name.
        // There can only be one such inner class; this is verified by
        // the checks against duplicate declarations during the
        // typechecking of the outer class.
        if( innerClasses != null ) {
            for( int i = 0; i < innerClasses.length; i++ ) {
                CClass maybeResult = innerClasses[i].getCClass();
                if( maybeResult.ident().equals(name) ) {
                    if( context == null ||
                        maybeResult.isAccessibleFrom( context.
                                                      findNearestHost() )) {
                        possibleClasses.add( maybeResult );
                    } else {
                        context.reportTrouble( 
                            new CWarning( NO_TOKEN,
                                          MjcMessages.CLASS_NO_ACCESS,
                                          name, maybeResult ));
                    }
                    break;
                }
            }
        }

            // only check superclasses/superinterfaces if we didn't
            // find a class of the right name above -- as per JLS 8.5,
            // the locally declared type hides any member types of the
            // same name in superclasses and superinterfaces.
        if (possibleClasses.size() == 0) {
        
            // See if we inherited an inner class with the given name
            if( superClass != null ) {
                // superclass should only be null for java.lang.Object
                CClass maybeResult = 
                    superClass.getCClass().lookupClass( name, context );
                if( maybeResult != null ) {
                    possibleClasses.add( maybeResult );
                }
            }
            
            if( interfaces != null ) {
                for( int i = 0; i < interfaces.length; i++ ) {
                    CClass maybeResult = 
                        interfaces[i].getCClass().lookupClass( name, context );
                    if( maybeResult != null ) {
                        possibleClasses.add( maybeResult );
                    }
                }
            }
        }
        
        // We either found no matches, exactly one match, or the reference
        // is ambiguous.
        if( possibleClasses.size() == 0 ) {
            return null;
        } else {
            Iterator i = possibleClasses.iterator();
            Object first = i.next();
            while (i.hasNext()) {
                Object second = i.next();
                if (!first.equals(second)) {
                    throw new UnpositionedError( MjcMessages.TYPE_AMBIG,
                                                 first, second );
                }
            }
            
            CClass result = (CClass)first;
            return result;
        }
    }

    public /*@nullable*/ CTypeVariable lookupTypeVariable(/*@non_null*/ String ident) {
        for (int i=0; i < typevariables.length; i++) {
            if(ident.equals(typevariables[i].getIdent())){
                return typevariables[i];
            }
        }
        if (owner() == null || isStatic()) {
            return null;   
        } else {
            return owner().lookupTypeVariable(ident);
        }
    }

    /**
     * Performs the typechecking for the type variables of this class.
     *
     * @param context  The type checking context.
     *
     * @throws UnpositionedError if some error occurs during type checking.
     */
    public void checkTypeVariables(CContextType context) throws UnpositionedError {
    	if (context != null) {
    		context = context.createClassContext(this);
    	}
    	
        for (int i = 0 ; i < typevariables.length; i++) {
            typevariables[i] =(CTypeVariable) typevariables[i].checkType(context);
        }
    }

    /**
     * When a generic type is instantiated, the type variables should be substituted
     * with the type arguments, This method returns the corresponding type arguments
     * for the type variable
     */
    public CClassType getSubstitution(CTypeVariable typevariable,
    		CClassType[][] sub) {
    	return getSubstitution(typevariable, sub, false);
    }
    
    /**
     * When a generic type is instantiated, the type variables should be substituted
     * with the type arguments, This method returns the corresponding type arguments
     * for the type variable
     * 
     * @param notOwnedTypeVariablesOnly Set to true, only type variables not
     * 									owned by the class itself are
     * 									replaced. This is very specific
     * 									to Generic Universe Types.
     */
    public CClassType getSubstitution(CTypeVariable typevariable,
    		CClassType[][] sub,
    		boolean notOwnedTypeVariablesOnly){
        CClassType substitution = getSubstitution(typevariable,
        		sub[sub.length-1],
        		notOwnedTypeVariablesOnly);
        
        return substitution;
    }
    
    
    public CClassType getSubstitution(CTypeVariable typevariable,
    		CClassType[] sub,
    		boolean notOwnedTypeVariablesOnly){
        assertTrue(sub != null);
        assertTrue(sub.length == typevariables.length);
        CClassType substitution = typevariable;
        if(ownerOf(typevariable)) {
        	if (!notOwnedTypeVariablesOnly) {
        		return sub[typevariable.getIndex()];
        	} else {
        		return typevariable;
        	}
        }
        else {
            if (superClass != null ) {
                substitution = getSuperSubstitution(superClass, typevariable);
                if(substitution != null) {
                    return substitution;
                }
            }
            for (int i=0 ; i < interfaces.length; i++) {
                substitution = getSuperSubstitution(interfaces[i],typevariable);
                if( substitution != null) {
                    break;
                }
            }
            return substitution;
        }
    }
    
    public CClassType getSuperSubstitution(CClassType supertype, CTypeVariable typeVar){
        CClassType[] newSubstitution = null;
        if (!supertype.isGenericType()) {
            newSubstitution = CClassType.EMPTY;
        } else {
            newSubstitution = ((CClassNameType)supertype).getArguments();
            //newSubstitution = new CClassType[typearguments.length];
            //for (int i= 0; i < typearguments.length; i++) {
            // if(typearguments[i].isTypeVariable()) {
            // newSubstitution[i] = substitution[typearguments[i].getIndex()];
            //          } else {
            //              newSubstitution[i] = typearguments[i];
            //          }
            //}
        }
        return supertype.getCClass().getSubstitution(typeVar,
        		newSubstitution,
        		false);
    }
    
    /**
     * check whether this class si the owner of the given type variable
     */
    private boolean ownerOf(CTypeVariable typevariable){
        if (typevariables == null) {
            return false;
        } else {
            for(int i=0; i < typevariables.length; i++){
                if (typevariables[i].getIdent().equals(typevariable.getIdent())) {
                    return true;
                }  
            }
            return false;
        }
    }
    
    /**
     * Searches a field in current class and parent hierarchy as
     * needed Creates an appropriate getter or setter method (or both)
     * if necessary for privileged access to the field.
     * @param   name    the simple name of the field
     * @param   context the context in which the reference appears, used 
     *                  for accessibility checks        
     * @exception UnpositionedError this error will be positioned soon */
    public CFieldAccessor lookupField( String name,
                                       CExpressionContextType context ) 
        throws UnpositionedError 
    {
        CField field = lookupFieldHelper( name, context );
        
        /* if this is a call to a private method in an enclosing
           or nested class, strip off the privacy modifier to
           hack around the java verifier */
        if ( memberNeedsPrivilegedAccess( field, context ) ) {
            //                  field.needsPrivacyModifierRemoved = true;
            return getterOrSetterMethodFor( field, context);
        }
        return field;
    }
    
    /**
     * Searches a field in current class and parent hierarchy as needed
     * @param   name    the simple name of the field
     * @param   context the context in which the reference appears, used 
     *                  for accessibility checks        
     * @exception UnpositionedError     this error will be positioned soon
     */
    protected CField lookupFieldHelper( String name,
                                     CExpressionContextType context ) 
        throws UnpositionedError 
    {
        CField field = getDeclaredField( name );
        
        if( field != null ) {
            if(!field.isAccessibleFrom( context.findNearestHost() ) ) {
                context.
                    reportTrouble( new CWarning( NO_TOKEN,
                                                 MjcMessages.FIELD_NO_ACCESS,
                                                 name, field.host() ));
                field = lookupSuperField( name, context );
            }
        } else {
            field = lookupSuperField( name, context );
        }
        return field;
    }
    
    /**
     * Searches a field in current class and parent hierarchy as needed
     * @param   name            the simple name of the field
     * @param   context the context in which the reference appears, used 
     *                  for accessibility checks        
     * @exception UnpositionedError     this error will be positioned soon
     */
    public CField lookupSuperField( String name, CExpressionContextType context )
        throws UnpositionedError 
    {
        // CField       field = getField( name );
        CField field = null;

        if (superClass != null) {
            // java.lang.object / interfaces
            field = getSuperClass().lookupFieldHelper( name, context );
        }

        for (int i = 0; i < interfaces.length; i++) {
            CField newField = 
                interfaces[i].getCClass().lookupFieldHelper( name, context );

            if ((field != null) && (newField != null) 
                && (field.owner() != newField.owner())) {
                throw new UnpositionedError(MjcMessages.FIELD_AMBIGUOUS, 
                                            new Object[] { 
                                                name,
                                                field.owner().getType(), 
                                                newField.owner().getType() }
                                            );
            }

            if (newField != null) {
                field = newField;
            }
        }

        return field;
    }

    /**
     * Returns true iff a field with same name is already defined in a 
     * superclass or an implemented interface.
     *
     * @param   ident           the name of the field
     */
    protected boolean isFieldRedefined( String ident, 
                                        CExpressionContextType dummyContext ) 
        throws UnpositionedError 
    {
        try {
            return this.lookupSuperField( ident, dummyContext ) != null;
        } catch (UnpositionedError e) {
            if (! e.hasDescription(MjcMessages.FIELD_AMBIGUOUS)) {
                throw e;
            }
            // defined more than once!
            return true;
        }
    }

   
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    public void augmentWithMethod( CMethod m, CCompilationUnitContextType c ) {
        augMap.addMethod( m, c );
    }

    /**
     * Searches for the generic function applicable to the given
     * identifier and argument type tuple, in the given context,
     * looking in parent hierarchy and augmentation map as needed.  The
     * receiver is assumed to be the type represented by this.
     *
     * @param   name            method name
     * @param   params          method parameter types
     * @param   context         the context in which the method reference
     *                          should be resolved, used to identify the
     *                          external methods augmenting this class
     * @return a method of the most specific applicable generic function
     * @exception UnpositionedError if the result is ambiguous */
    public CMethod lookupMethod( final String name, final CType[] params, 
                                 final CClassType[] args,
                                 final CContextType context) 
        throws UnpositionedError 
    {
        CMethodSet accum = lookupMethodOrSet(name, params,args, context);

        if (accum == null) {
            return null;
        } else if( accum.size() > 1 ) {
            throw new UnpositionedError( MjcMessages.METHOD_AMBIG, 
                                         accum.getMethod(0),
                                         accum.getMethod(1) );
        } else {
            return accum.getFirst();
        }
    }

    /**
     * Searches for the generic function applicable to the given
     * identifier and argument type tuple, in the given context,
     * looking in parent hierarchy and augmentation map as needed.  The
     * receiver is assumed to be the type represented by this.
     *
     * @param   name            method name
     * @param   params          method parameter types
     * @param   context         the context in which the method reference
     *                          should be resolved, used to identify the
     *                          external methods augmenting this class
     * @return  set of methods of the most specific applicable generic function
     * @exception UnpositionedError if the result is ambiguous */
    public CMethodSet lookupMethodOrSet( final String name, 
                                         final CType[] params, 
                                         final CClassType[] argss,
                                         final CContextType context) 
        throws UnpositionedError 
    {
        final CMethodSet accum = new CMethodSet();
        final CType recv = getType();

        NoDupStrategy actor =
            new NoDupStrategy() {
                    public boolean maybeInclude( CMethod candidate,
                                                 CClassType[] args) {
                        if( candidate.isApplicable( name, recv, params, args )) {
                            if( !candidate.
                                isAccessibleFrom(context.findNearestHost())) {
                                // applicable method is not accessible
                                // so issue a warning
                                context.reportTrouble( 
                                   new CWarning( NO_TOKEN,
                                                 MjcMessages.INVOKE_NO_ACCESS,
                                                 name, candidate.host() ));
                                return false;
                            } else {
                                return true;
                            }
                        } else {
                            return false;
                        }
                    }
                
                    public byte resultFor( CMethod candidate,
                                           CMethod accumMethod )
                        throws UnpositionedError
                    {
                        // By the invariant on accum, candidate is
                        // (i)   incomparable to all methods in accum,
                        // (ii)  less specific than some (or all) methods 
                        //       in accum and incomparable to the others, 
                        // (iii) more specific than some (or all) methods in 
                        //       accum and incomparable to the others,
                        // (iv)  equal to exactly one method in accum (i.e., 
                        //       it has the same signature), or
                        boolean ignoreCurrent = false;
                        boolean done = false;
                        boolean remove = false;
                            // the 'true' argument to isMoreSpecificThan
                            // signifies that we should treat interfaces
                            // as not inheriting from Object
                        if( accumMethod.isMoreSpecificThan(candidate, true) ){
                            // if current is less specific than some
                            // method in accum then ignore current,
                            // (case ii)
                            ignoreCurrent = true; done = true; remove = false;
                        } else if(
                            candidate.isMoreSpecificThan(accumMethod, true) ){
                            // else if current is more specific than some
                            // method in accum then add current to accum
                            // and remove the methods that are less
                            // specific than current (case iii)
                            ignoreCurrent = false; done = false; remove = true;
                        } else {
                                // the methods are either incomparable (case i)
                                // or they are equivalent (case iv)
                            boolean eq = candidate.equals(accumMethod);
                            
                                // if either method is abstract then it should
                                // be removed
                            if( candidate.isAbstract() ) {
                                ignoreCurrent = true; remove = false;
                            } else if( accumMethod.isAbstract() ) {
                                ignoreCurrent = false; remove = true;
                            } else if (eq) {
                                    // if neither is abstract and the methods
                                    // are equivalent then we have an ambiguity
                                throw new UnpositionedError( MjcMessages
                                                             .METHOD_AMBIG,
                                                             candidate,
                                                             accumMethod );
                            }
                            if (eq) {
                                    // by the invariant no methods other than
                                    // accumMethod in accum can have the same
                                    // signature as current, therefore it must
                                    // be the case that no other methods in
                                    // accum are comparable to current or
                                    // accumMethod, so...
                                done = true;
                            }
                        }
                        return CMethodSet.makeResult( ignoreCurrent, remove, 
                                                      done );

                    }
                };
        
        
        accumMostSpecificMethods( name, actor, accum, argss, context );
        
        CMethod m = accum.getFirst();
        if( accum.size() > 1 ) {
            if (context.getCompiler().RMJ() && m.isInExternalGF()) {
                // we can allow ambiguous external methods, since glue
                // methods can be defined that resolve the ambiguity
            } else {
                throw new UnpositionedError( MjcMessages.METHOD_AMBIG, 
                                             m, accum.getMethod(1) );
            }
        } else {
                /* if this is a call to a private method in an
                   enclosing or nested class, strip off the privacy
                   modifier to hack around the java verifier */
            if ( memberNeedsPrivilegedAccess( m, context ) ) {
                    // just strip the privacy modifier on external methods
                    // and constructors, rather than creating a wrapper method
                    /* !FIXME! We may want to handle constructors as javac
                       does, creating a new package-level constructor that has
                       an extra dummy parameter.
                    */
                    /* !FIXME! This scheme is not sufficient for
                       handling protected external methods, only
                       private ones.  */
                if ( m.isConstructor() || m.isInExternalGF() ) {
                    m.topConcreteMethod().needsPrivacyModifierRemoved = true;
                } else {
                    CMethodSet accum2 = new CMethodSet();
                    accum2.addMethod( wrapCallTo(m,false) );
                    return accum2;
                }
            }
        }

        return accum;
    }


    /**
     * Searches for the methods specialized by a given method, in the
     * given context, looking in parent hierarchy and augmentation map
     * as needed.
     *
     * @param   specMethod      the method that is doing the specializing
     * @param   context         the context in which the method reference
     *                          should be resolved, used to identify the
     *                          external methods augmenting this class
     * @return  A set of methods, all with the same static type tuple and 
     *          ident as <code>specMethod</code>, such that 
     *          <code>specMethod</code> <em>immediately specializes</em> 
     *          each method in the set in the given context.
     * @exception UnpositionedError if the result is ambiguous */
    public CMethodSet 
        lookupMethodsSpecializedByMethod( final CMethod specMethod,
                                          final CClassType[] args,
                                          final ArrayList needbridge,
                                          final CContextType context ) 
        throws UnpositionedError 
    {
        final String name = specMethod.ident();
        CMethodSet accum = new CMethodSet();
        // accum satisfies the invariants that <code>specMethod</code>
        // apparentlySpecializes all methods in accum and any two
        // distinct methods in <code>accum</code> are either
        // incomparable by the apparentlySpecializes relationship, or
        // have the same signature (the same signature is allowed for
        // methods that are inherited from object through both a
        // superclass and an interface

        // The apparentlySpecializes relationship is used because we
        // want the resulting set to include implemented interface
        // methods as well as overridden concrete methods.

        NoDupStrategy actor = 
            new NoDupStrategy() {
                    public boolean maybeInclude( CMethod candidate,
                                                 CClassType[] args) 
                        throws UnpositionedError
                    {
			// WMD: disregard the Universe part for overriding
			// resolution!
                        boolean apparentlySpecializes = 
                            specMethod.apparentlySpecializes (candidate,args, false);
                        if(context.getCompiler().Generic()){
			    // WMD: FIXME check whether to disregard
			    // the universes here!
                            boolean hasSameSig =
                                specMethod.hasSameSignature(candidate);
                            // If both methods are static, they are totally
                            // unrelated (hiding, not overriding).
                            boolean bothStatic =
                                specMethod.isStatic() && candidate.isStatic();
                            String ident = specMethod.ident();
                            boolean initStyleCheck =(ident==JAV_CONSTRUCTOR
                                                     ||ident==JAV_INIT
                                                     ||ident==JAV_STATIC_INIT);
                            if( !apparentlySpecializes && !bothStatic
                                && !initStyleCheck && hasSameSig &&
                                !specMethod.receiverType().equals
                                (candidate.receiverType())){
                                //has same signature but does not override it
                                //thus violate generic method override rule 2
                                throw new UnpositionedError(
                                		MjcMessages.GENERIC_OVER_RULE2,
                                		specMethod.toString(),
                                		candidate.toString());
                            }
                        }
                        if( apparentlySpecializes ) {
                            if(context.getCompiler().Generic()){
                                boolean sameReturn = true;
                                CType r1 = specMethod.returnType();
                                CType r2 = candidate.returnType();
                                if(r1.isClassType()
                                   && r2.isClassType()
                                   && !(r1.getCClass().equals(r2.getCClass()))){
                                    sameReturn = false;
                                }
                                if((!specMethod.hasSameSignatureNoUniverses(candidate)
                                    || !sameReturn)
                                   && !candidate.isPrivate()){
                                    needbridge.add(candidate);
                                }
                            }
                            if( !candidate.
                                isAccessibleFrom(context.findNearestHost())) {
                                // applicable method is not accessible
                                // so issue a warning
                                context.reportTrouble( 
                                   new CWarning( NO_TOKEN,
                                                 MjcMessages.OVERRIDE_NO_ACC,
                                                 name, candidate.host() ));
                                return false;
                            } else {
                                return true;
                            }
                        } else {
                            return false;
                        }
                    }

                    public byte resultFor( CMethod candidate,
                                           CMethod accumMethod )
                        throws UnpositionedError
                    {
                        // By the invariant on accum, current is
                        // (i)   incomparable to all methods in accum,
                        // (ii)  less specific than some (or all) methods in 
                        //       accum and incomparable to the others, 
                        // (iii) more specific than some (or all) methods in 
                        //       accum and incomparable to the others,
                        // (iv)  equal to one or more methods in accum (i.e.,
                        //       it has the same signature), or
                        boolean ignoreCurrent = false;
                        boolean done = false;
                        boolean remove = false;
                        if( accumMethod.isMoreSpecificThan(candidate)) {
                            // if current is less specific than some
                            // method in accum then ignore current,
                            // (case ii)
                            ignoreCurrent = true; done = true; remove = false;
                        } else if( candidate.isMoreSpecificThan(accumMethod)) {
                            // else if current is more specific than some
                            // method in accum then add current to accum
                            // and remove the methods that are less
                            // specific than current (case iii)
                            ignoreCurrent = false; done = false; remove = true;
                        } else if( candidate.equals( accumMethod ) ) {
                            // by the invariant it must
                            // be the case that no other methods in accum
                            // are more or less specific than current, so...
                            ignoreCurrent = false; done = true; remove = false;
                        }
                        return CMethodSet.makeResult( ignoreCurrent, remove, 
                                                      done );

                    }
                };

        accumLocalExtAndInheritedMethods( name, actor, accum, args, context );
        
        return accum;
    }

    /**
     * Searches for the top <em>concrete</em> method of the generic
     * function to which the given method belongs, in the given
     * context, looking in the parent hierarchy and augmentation map as
     * needed.
     *
     * @param   method          the method from which to begin the search
     * @param   context         the context in which the method reference
     *                          should be resolved.  Used to identify the
     *                          external methods augmenting this class and 
     *                          to determine the accessibility of the 
     *                          methods found.
     * @return  the top concrete method of the generic function
     * @exception UnpositionedError if the result is ambiguous */
    public CMethod lookupTopConcreteMethod( final CMethod method, 
                                            final CClassType[] args,
                                            final CContextType context ) 
        throws UnpositionedError
    {
        final String name = method.ident();
            // the types already searched for an appropriate top method
        CMethodSet accum = new CMethodSet();
        
        // accum satisfies the invariants that (1)
        // <code>specMethod</code> apparentlySpecializes all methods
        // in accum (see CMethod), (2) any two distinct methods in
        // <code>accum</code> are incomparable by the specializes
        // relationship, and (3) every method in accum is concrete

        NoDupStrategy actor = 
            new NoDupStrategy() {
                    public boolean maybeInclude( CMethod candidate,
                                                 CClassType[] args) {
                        // The use of apparentlySpecializes here
                        // imposes a partial order on the methods so
                        // that the calls to specialize below cannot
                        // (recursively) invoke
                        // <code>lookupTopConcreteMethod</code> on
                        // <code>method</code>.  That must be
                        // guaranteed or we would have an infinite
                        // loop.
                        return method.apparentlySpecializes( candidate,
                                                             args)
                            && !candidate.isAbstract()
                            && candidate.
                            isAccessibleFrom( context.findNearestHost() );
                    }

                    public byte resultFor( CMethod candidate,
                                           CMethod accumMethod )
                        throws UnpositionedError
                    {
                        // By the invariant on accum, current is
                        // (i)   incomparable to all methods in accum,
                        // (ii)  less specific than some (or all) methods in
                        //       accum and incomparable to the others, 
                        // (iii) more specific than some (or all) methods in 
                        //       accum and incomparable to the others, or
                        // (iv)  equal to exactly one method in accum (i.e., 
                        //       it has the same signature).
                        boolean ignoreCurrent = false;// default to case i
                        boolean done = false;
                        boolean remove = false;
                        if( accumMethod.specializes(candidate)) {
                            // if current is less specific than some
                            // method in accum add current to accum and
                            // remove methods that are more specific than
                            // current (case ii)
                            ignoreCurrent = false; done = false; remove = true;
                        } else if( candidate.specializes(accumMethod)) {
                            // else if current is more specific than some
                            // method in accum then ignore current (case
                            // iii)
                            ignoreCurrent = true; done = true; remove = false;
                        } else if( candidate.equals( accumMethod ) ) {
                            // by the invariant it must
                            // be the case that no other methods in accum
                            // are more or less specific than current, so...
                            ignoreCurrent = false; done = true; remove = false;
                        }
                        return CMethodSet.makeResult( ignoreCurrent, remove, 
                                                      done );

                    }
                };

        accumLocalExtAndInheritedMethods( name, actor, accum, args, context );

        if( accum.size() > 1 ) {
            CMethod m1 = accum.getMethod(0);
            CMethod m2 = accum.getMethod(1);
            CMethod lessSpec, moreSpec;
            // Determine which method is more specific.  Because of
            // the invariants on the accumulator and Java's single
            // inheritance:
            /*@ assert( m1.apparentlySpecializes(m2) ||
              @         m2.apparentlySpecializes(m1) ); @*/
            if (m1.apparentlySpecializes(m2)) {
                moreSpec = m1;
                lessSpec = m2;
            } else {
                moreSpec = m2;
                lessSpec = m1;
            } // end of else
            if (lessSpec.isExternal() ) {
                if (moreSpec.isExternal()) {
                    throw new UnpositionedError( MjcMessages.
                                                 TOP_METHOD_AMBIG_IMPORT2, 
                                                 moreSpec, lessSpec );
                } else {
                    throw new UnpositionedError( MjcMessages.
                                                 TOP_METHOD_AMBIG_IMPORT1, 
                                                 moreSpec, lessSpec );
                }
            }
            // issue a general ambiguity error, if we come up with a
            // test case that triggers this then perhaps we can
            // improve the message
            throw new UnpositionedError( MjcMessages.TOP_METHOD_AMBIG, 
                                         moreSpec, lessSpec ); 
        } else {
            return accum.getFirst();
        }
    }

    /**
     * Searches for the most specific applicable method for the given
     * argument tuple in the generic function to which the given
     * method belongs in the given context, looking in parent
     * hierarchy and augmentation map as needed.
     *
     * @param   top             the method that denotes the generic function
     * @param   argTuple        the tuple of argument types
     * @param   context         the context in which the method reference
     *                          should be resolved, used to identify the
     *                          external methods augmenting this class
     * @return  the set of most specific applicable methods, or null if none
     * @exception UnpositionedError if the result is ambiguous and it
     * can't be resolved through external means (e.g. via glue
     * methods) */
    public CMethodSet lookupMSAMethod( final CMethod top, CType[] argTuple, 
                                       final CClassType[] argss,
                                       final CContextType context )  
        throws UnpositionedError
    {
        assertTrue( argTuple.length >= 1 );
        final String name = top.ident();
        final CType recvType = argTuple[0];
        final CType[] actuals = new CType[ argTuple.length - 1 ];
        for( int i=0; i<actuals.length; i++ ) {
            actuals[i] = argTuple[i+1];
        }
        final CMemberHost host = context.findNearestHost();

        CMethodSet accum = new CMethodSet();
        // accum satisfies the invariants that all methods in accum
        // are applicable and any two distinct methods in accum are
        // incomparable by the specializes relationship

        NoDupStrategy actor = 
            new NoDupStrategy() {
                    public boolean maybeInclude( CMethod candidate,
                                                 CClassType[] arg1) 
                        throws UnpositionedError
                        {
                            return candidate.topConcreteMethod().equals(top)
                                && candidate.isApplicable( name, 
                                                           recvType,
                                                           actuals,
                                                           arg1);
                        }

                    public byte resultFor( CMethod candidate,
                                           CMethod accumMethod )
                        throws UnpositionedError
                    {

                        // By the invariant on accum, current is
                        // (i)   incomparable to all methods in accum,
                        // (ii)  less specific than some (or all) methods in 
                        //       accum and incomparable to the others, 
                        // (iii) more specific than some (or all) methods in 
                        //       accum and incomparable to the others,
                        // (iv)  equal to exactly one method in accum (i.e., 
                        //       it has the same signature), or
                        boolean ignoreCurrent = false;
                        boolean done = false;
                        boolean remove = false;
                            // the 'true' argument to isMoreSpecificThan
                            // signifies that we should treat interfaces
                            // as not inheriting from Object
                        if( accumMethod.isMoreSpecificThan(candidate, true)) {
                            // if current is less specific than some
                            // method in accum then ignore current, (case ii)
                            ignoreCurrent = true; done = true; remove = false;
                        } else if( candidate.isMoreSpecificThan(accumMethod,
                                                                true)) {
                            // else if current is more specific than some
                            // method in accum then add current to accum
                            // and remove the methods that are less
                            // specific than current (case iii)
                            ignoreCurrent = false; done = false; remove = true;
                        } else {
                                // the methods are either incomparable (case i)
                                // or they are equivalent (case iv)

                            boolean eq = candidate.equals(accumMethod);
                            
                                // if either method is abstract then it should
                                // be removed
                            if( candidate.isAbstract() ) {
                                ignoreCurrent = true; remove = false;
                            } else if( accumMethod.isAbstract() ) {
                                ignoreCurrent = false; remove = true;
                            } else if (eq) {
                                    // if neither is abstract and the methods
                                    // are equivalent then we have an ambiguity
                                Object[] args = 
                                    { name,
                                      CType.tupleToString( recvType, actuals ),
                                      candidate, accumMethod };
                                throw new UnpositionedError( 
                                        MjcMessages.MM_AMBIG, args );
                            }
                                
                            if (eq) {
                                    // by the invariant no methods other than
                                    // accumMethod in accum can have the same
                                    // signature as current, therefore it must
                                    // be the case that no other methods in
                                    // accum are comparable to current or
                                    // accumMethod, so...
                                done = true;
                            }
                        }
                        return CMethodSet.makeResult( ignoreCurrent, remove, 
                                                      done );
                        
                    }
            };

        accumLocalExtAndInheritedMethods( name, actor, accum,argss, context );

        if( accum.size() > 1 ) {
            Object[] args = 
                new Object[] { name, 
                               CType.tupleToString(recvType, actuals),
                               accum.getMethod(0), 
                               accum.getMethod(1) };
            if (context.getCompiler().RMJ() && top.isInExternalGF()) {
                // we can allow ambiguous external methods, since glue
                // methods can be defined that resolve the ambiguity
                context.reportTrouble( new CWarning( NO_TOKEN,
                                                     MjcMessages.RMJ_MM_AMBIG,
                                                     args ));
            } else {
                throw new UnpositionedError( MjcMessages.MM_AMBIG, args );
            }
        }

        return accum;
    }

    /**
     * Collects all the methods in the same GF as the argument method,
     * visible in the given context, looking in parent hierarchy and
     * augmentation map as needed.
     *
     * @param   m               a method in the generic function being gathered
     * @param   context         the context in which the method reference
     *                          should be resolved, used to identify the
     *                          external methods augmenting this class
     * @return  the set of methods in the GF, or null if none
     * @exception UnpositionedError if there was an error
     */
    public CMethodSet lookupAllMethods( final CMethod m, 
                                        final CClassType[] args,
                                        final CContextType context )  
        throws UnpositionedError
    {
        final CMethod top = m.topConcreteMethod();

        CMethodSet accum = new CMethodSet();

        NoDupStrategy actor = 
            new NoDupStrategy() {
                public boolean maybeInclude( CMethod candidate,
                                             CClassType[] args) 
                    throws UnpositionedError
                {   //maybe need fixed, TJ
                    return candidate.topConcreteMethod().equals(top);
                }
                
                public byte resultFor( CMethod candidate,
                                       CMethod accumMethod )
                {
                    return CMethodSet.makeResult( false, false, false );
                    
                }
            };
        
        String name = m.ident();
        accumLocalExtAndInheritedMethods( name, actor, accum, args,context );
        return accum;
    }


    /**
     * If overloader could be rewritten to dynamically specialize a
     * method instead of statically overloading it, then this return
     * said method.  Otherwise this returns null.
     * @param overloader the method to be checked
     * @param context the context in which the method is declared
     */
    public CMethod lookupOverloadedMethod(final CMethod overloader,
                                          final CClassType[] args,
                                          final CContextType context) 
        throws UnpositionedError
    {
        final String name = overloader.ident();
        final CSpecializedType[] overloaderParms = overloader.parameters();
        CMethodSet accum = new CMethodSet();
        
        NoDupStrategy actor = 
            new NoDupStrategy() {
                public boolean maybeInclude( CMethod candidate,
                                             CClassType[] args) 
                    throws UnpositionedError
                {
                    if (candidate == overloader) {
                        return false;
                    }
                    
                    CSpecializedType[] candidateParms = candidate.parameters();
                    if (overloaderParms.length != candidateParms.length
                        || !candidate.ident().equals(name)) {
                        return false;
                    }

                    // The walk up the class hierarchy ensures the
                    // overloader's receiver type is a subtype of
                    // candidate's receiver type, so just check the
                    // parameters:
                    return couldOverload(overloaderParms, candidateParms);
                }

                public byte resultFor( CMethod candidate,
                                       CMethod accumMethod )
                {
                    boolean ignoreCandidate;
                    boolean removeAccum;
                    
                    // assert(candidate.ident().equals(accumMethod.ident()));
                    CSpecializedType[] candidateParms = candidate.parameters();
                    CSpecializedType[] accumParms = accumMethod.parameters();
                    if (couldOverload(candidateParms, accumParms)) {
                        ignoreCandidate = true;
                        removeAccum = false;
                    } else if (couldOverload(accumParms, candidateParms)) {
                        ignoreCandidate = false;
                        removeAccum = true;
                    } else {
                        ignoreCandidate = false;
                        removeAccum = false;
                    }

                    final boolean done = false;
                    return CMethodSet.makeResult(ignoreCandidate, 
                                                 removeAccum, done);
                }

                private boolean couldOverload(CSpecializedType[] ss,
                                              CSpecializedType[] ts) 
                {
                    boolean atLeastOneDifference = false;
                    for (int i = 0; i < ss.length; i++) {
                        if (!ss[i].couldOverload(ts[i])) {
                            return false;
                        }
                        atLeastOneDifference = atLeastOneDifference 
                            || !ss[i].staticType().equals(ts[i].staticType());
                    }
                    return atLeastOneDifference;
                }
            };

        accumLocalExtAndInheritedMethods( name, actor, accum, args,context );

        if (accum.size() == 0) {
            return null;
        } else {
            return accum.getFirst();
        }
        
    }
    

    /**
     * Accumulates the set of methods with identifier
     * <code>name</code> declared in the type represented by this,
     * <em>or added to the type by external methods</em>, using the
     * strategy <code>actor</code>.  Only searches supertypes if no
     * matches have been accumulated while searching the type
     * represented by this.
     *
     * @param   name            method name
     * @param   actor           the strategy for selecting methods
     * @param   accum           a method set in which to accumulate the 
     *                          results
     * @param   context         the context in which the method reference
     *                          appears, used for resolving augmenting
     * @exception UnpositionedError at the discretion of the strategy
     *                          <code>actor</code> */
    
    protected void accumMostSpecificMethods( String name, 
                                             NoDupStrategy actor,
                                             CMethodSet accum,
                                             CClassType[] args,
                                             CContextType context )
        throws UnpositionedError 
    {

        if (actor.alreadySearched(this))
            return;

        accumLocalInternalMethods( name, actor, accum, args );

        augMap.getAddedMethods( context.getCompilationUnit() )
            .accumMethodSet( actor, name, accum,args );

        if (accum.size() == 0) {
            if( superClass != null ) {
                CClassType[] superArgs= 
                    superClass.isGenericType()?superClass.getArguments():
                    CClassType.EMPTY;
                getSuperClass().accumMostSpecificMethods( name, actor, accum, 
                                                          superArgs,
                                                          context );
            }

            for (int i = 0; i < interfaces.length; i++) {
                CClassType[] interfaceArgs= 
                    interfaces[i].isGenericType()?interfaces[i].getArguments():
                    CClassType.EMPTY;
                interfaces[i].getCClass().
                    accumMostSpecificMethods( name, actor, accum,
                                              interfaceArgs,
                                              context );
            }
        }
    }

  
    /**
     * Accumulates the set of methods with identifier
     * <code>name</code> declared in the type represented by this,
     * using the strategy <code>actor</code>.  Searches neither
     * external methods nor supertypes.
     *
     * @param   name            method name
     * @param   actor           the strategy for selecting methods
     * @param   accum           a method set in which to accumulate the 
     *                          results
     * @exception UnpositionedError at the discretion of the strategy
     *                          <code>actor</code> */
    
    protected void accumLocalInternalMethods( String name, 
                                              NoDupStrategy actor,
                                              CMethodSet accum,
                                              CClassType[] args )
                                              
        throws UnpositionedError 
    {
        protoMethods.accumMethodSet( actor, name, accum, args );
    }
    /**
     * Accumulates the set of methods with identifier
     * <code>name</code> declared in the type represented by this,
     * <em>or added to the type by external methods</em>, using the
     * strategy <code>actor</code>.  <em>Always</em> searches supertypes.
     *
     * @param   name            method name
     * @param   actor           the strategy for selecting methods
     * @param   accum           a method set in which to accumulate the 
     *                          results
     * @param   context         the context in which the method reference
     *                          appears, used for resolving augmenting
     * @exception UnpositionedError at the discretion of the strategy
     *                          <code>actor</code> */
    protected final void accumLocalExtAndInheritedMethods(String name,
                                                          NoDupStrategy actor,
                                                          CMethodSet accum,
                                                          CClassType[] args,
                                                          CContextType context)
        throws UnpositionedError 
    {
        if (actor.alreadySearched(this))
            return;

        accumLocalInternalMethods( name, actor, accum, args );

        // check the augmenting methods of this class
        augMap.getAddedMethods( context.getCompilationUnit() )
            .accumMethodSet( actor, name, accum ,args);

        if (superClass != null) {
            CClassType[] superArgs = 
                superClass.isGenericType()? superClass.getArguments():CClassType.EMPTY;
            getSuperClass().accumLocalExtAndInheritedMethods( name, actor, 
                                                              accum,superArgs,
                                                              context );
        }

        for (int i = 0; i < interfaces.length; i++) {
            CClassType[] interfaceArgs =
                interfaces[i].isGenericType()? interfaces[i].getArguments():
                CClassType.EMPTY;
            interfaces[i].getCClass().
                accumLocalExtAndInheritedMethods( name, actor, 
                                                  accum,interfaceArgs,
                                                  context );
        }
    }
    
    /**
     * Returns a list of abstract methods
     */
    public CMethodSet.MethodArgsPair[] getAbstractMethods() {
        ArrayList v = new ArrayList();
        CClassType[] args = getType().getArguments();
        collectAbstractMethods(v, args);
        CMethodSet.MethodArgsPair[] methods =
            new CMethodSet.MethodArgsPair[v.size()];
        v.toArray(methods);
        
        return methods;
    }

    /**
     * Returns a list of interface methods
     */
    public CMethodSet.MethodArgsPair[] getInterfaceMethods() {
        ArrayList v = new ArrayList();
        CClassType[] args = getType().getArguments();
        collectInterfaceMethods(v, args);
        CMethodSet.MethodArgsPair[] methods = new CMethodSet.MethodArgsPair[v.size()];
        v.toArray(methods);
        
        return methods;
    }
    
    public void addBridgeMethod(CMethod me,
                                CContextType context)
        throws UnpositionedError
    {
        int size = protoMethods.size();
        boolean alreadycontain = false;
        for (int i =0 ; i< size; i++) {
            CMethod method = protoMethods.getMethod(i);
            if(method.hasSameSignature(me)&& 
               method.returnType().equals(me.returnType())) {
                if(method.isSynthetic()&&
                   method.getGenerateFrom() != me.getGenerateFrom() ){
                    throw new UnpositionedError(MjcMessages.GENERIC_OVER_RULE3,
                                                me.getGenerateFrom(),
                                                method.getGenerateFrom());
                } else {
                    alreadycontain = true;
                }           
            }
        }
        CClassType[] args = getType().getArguments();
        checkBridgeMethod(me,args,context);
        
        if(!alreadycontain) {
            this.protoMethods.addMethod(me);
        }

    }
    
    public void checkBridgeMethod(CMethod me,
                                  CClassType[] args,
                                  CContextType context) 
        throws UnpositionedError 
    {
        CMethod generatefrom = me.getGenerateFrom();
        int size = protoMethods.size();
        for (int i =0 ; i< size; i++) {
            CMethod method = protoMethods.getMethod(i);
            boolean apparentlySpeciaze = 
                generatefrom.apparentlySpecializes(method,args);
            boolean hasSameSig =
                method.hasSameSignature(me);
            if(!apparentlySpeciaze && hasSameSig && 
               method.receiverType() != me.receiverType()) {
                throw new UnpositionedError(MjcMessages.GENERIC_OVER_RULE4,
                                            me.getGenerateFrom(),
                                            method);
            }
        }
        
        if (superClass != null) {
            // java.lang.object
            CClassType[] superArgs= 
                superClass.isGenericType()?superClass.getArguments():
                CClassType.EMPTY;
            getSuperClass().checkBridgeMethod(me, superArgs,context);
        }
        if (interfaces != null) {
            for (int i = 0; i < interfaces.length; i++) {
                CClass parent;
                parent = interfaces[i].getCClass();
                CClassType[] interfaceArgs= 
                    interfaces[i].isGenericType()?interfaces[i].getArguments():
                    CClassType.EMPTY;
                assertTrue(parent != null);
                parent.checkBridgeMethod(me , interfaceArgs,context);
            }
        }
        
    }

    /**
     * Walks up the extends hierarchy and adds all methods
     * to the given list.
     *
     * @param   v       mutated to include the CMethods for all 
     *                  methods in this class and all direct ancestors
     */
    public void getAllMethods(ArrayList v) {
        protoMethods.appendMethodsTo( v );

        if (superClass != null) {
            // java.lang.object
            getSuperClass().getAllMethods(v);
        }
    }

    /**
     * Return all methods that are inherited from supertypes, except
     * for those overridden in this class.
     */
    public /*@ pure non_null @*/ ArrayList getAllInheritedMethods() {
        ArrayList methods = new ArrayList();
        getAllInheritableMethods(methods);

        ArrayList result = new ArrayList();
        for (Iterator i = methods.iterator(); i.hasNext(); ) {
            CMethod m = (CMethod) i.next();
            if (!protoMethods.contains(m)) {
                result.add(m);
            }
        }
        
        return result;
    }
    
    /**
     * Walks up the extends and implements hierarchies and adds all
     * inheritable methods to the given list, <code>v</code>, if
     * methods of the same signature are not already contained in
     * <code>v</code>.
     */
    public void getAllInheritableMethods(/*@ non_null @*/ ArrayList v) {
        // exclude methods inherited from java.lang.Object
        if (getType().equals(CStdType.Object)) {
            return;
        }

        // add inheritable methods from this class or interface
        for (CAbstractMethodSet.Iterator i = protoMethods.iterator(); 
             i.hasNext(); ) {
            CMethod m = (CMethod) i.next();
            if ((m.isPublic() || m.isProtected()) &&
                !m.isStatic() &&
                !m.isConstructor()) {
                boolean found = false;
                for (Iterator j = v.iterator(); j.hasNext(); ) {
                    CMethod n = (CMethod) j.next();
                    if (m.ident().equals(n.ident()) &&
                        m.equalParameters(n)) {
                        found = true;
                        break;
                    }
                }
                if (!found) {
                    v.add(m);
                }
            }
        }
        
        // accumulate inheritable methods of the superclass
        if (superClass != null) {
            getSuperClass().getAllInheritableMethods(v);
        }

        // accumulate inheritable methods of superinterfaces
        if (isInterface()) {
            for (int i = 0; i < interfaces.length; i++) {
                CClass parent = interfaces[i].getCClass();
                parent.getAllInheritableMethods(v);
            }
        }
    }
    
    /**
     * Get the version of the encoding used to encode the universe
     * type modifiers. This is only meaningful in CSourceClass (where it
     * indicates which version the compiler is using) and CBinaryClass
     * (where it indicates which version has been used in the class file).
     * By AHS.
     */
    public String getUniverseVersion() {
        return null;
    } 

    /**
     * Walks up the extends hierarchy and adds all abstract methods
     * to the given list.
     *
     * @param   v       mutated to include the CMethods for all abstract 
     *                  methods in this class and all direct ancestors
     */
    public void collectAbstractMethods(ArrayList v, CClassType[] args) {
        protoMethods.appendAbstractMethodsTo( v , args );

        if (superClass != null) {
            CClassType[] superArgs= 
                superClass.isGenericType()?superClass.getArguments():
                CClassType.EMPTY;
            // java.lang.object
            getSuperClass().collectAbstractMethods(v, superArgs);
        }
    }

    /**
     * Walks up the implements and extends hierarchies and adds all methods
     * of interfaces to the given list.
     *
     * @param   v       mutated to include the CMethods for all interface 
     *                  methods in this class and all its ancestors
     */
    public void collectInterfaceMethods(ArrayList v, CClassType[] args) {
        if (isInterface()) {
            protoMethods.appendMethodsTo( v, args );
        }

        if (interfaces != null) {
            for (int i = 0; i < interfaces.length; i++) {
                CClass parent;
                parent = interfaces[i].getCClass();
                CClassType[] interfaceArgs= 
                    interfaces[i].isGenericType()?interfaces[i].getArguments():
                    CClassType.EMPTY;
                assertTrue(parent != null);
                parent.collectInterfaceMethods(v , interfaceArgs);
            }
        }

        if (superClass != null) {
            // java.lang.object
            CClassType[] superArgs= 
                superClass.isGenericType()?superClass.getArguments():
                CClassType.EMPTY;
            getSuperClass().collectInterfaceMethods(v, superArgs);
        }
    }

    /**
     * @return the short name of this class
     */
    protected static String getIdentFrom(String qualifiedName) {
        assertTrue(qualifiedName != null);
        String  syntheticName;

        int             cursor = qualifiedName.lastIndexOf('/');
        if (cursor > 0) {
            syntheticName = qualifiedName.substring(cursor + 1);
        } else {
            syntheticName = qualifiedName;
        }

        cursor = syntheticName.lastIndexOf('$');
        if (cursor > 0) {
            return syntheticName.substring(cursor + 1);
        } else {
            return syntheticName;
        }
    }

    /*
     * Indicates whether this is accessible from the given host.
     *
     * @param   from    the host that wants access
     * @return  true iff the given host is allowed access to this host's 
     *          members
    public boolean isAccessibleFrom( CMemberHost from ) {
        // public members and lexically nested members are accessible
        if( isPublic() 
            || from.isDefinedInside( this ) 
            || isDefinedInside( from ) ) {
            return true;
        }
        
        // nested types have the same access privileges as their host class
        if( from.isNestedType() && isAccessibleFrom( from.host() )) {
            return true;
        }

        // non-private members are accessible throughout their home package
        if( !isPrivate() && packageName() == from.packageName() ) {
            return true;
        } 
        
            // protected classes are accessible from subclasses of their
            // lexically enclosing classes
        if( isProtected() && isNestedType() && host().isClass() &&
            from.descendsFrom((CClass) host()) ) {
            return true;
        }

        return false;
    }
     */

    /**
     * Indicates whether this member is accessible from the given host 
     * when this member has protected visibility.
     * In this case, protected classes are accessible from subclasses of their
     * lexically enclosing classes.
     *
     * @param   from    the host that wants access
     * @return  true iff the given host is allowed access to this member
     */
    public /*@ pure @*/ boolean hasProtectedVisibilityIn( CMemberHost from ) {
        return isNestedType() 
            && host().isClass() 
            && from.descendsFrom((CClass) host());
    }

    
    /**
     * Add inner fields
     */
    public void addField(CSourceField field) {
        fields.add(field);
    }

    /**
     * Register a nested type visible inside this class.
     *
     * <pre><jml>
     * requires type instanceof CClassType && type.isNestedType();
     * </jml></pre>
     */
    public void registerNestedType( CType type ) {
        visibleNestedTypes.add( type );
    }

    // ----------------------------------------------------------------------
    // PRIVILEGED ACCESS BY INNER/ENCLOSING CLASSES
    // ----------------------------------------------------------------------

    /**
     * Indicates whether the given member needs a wrapper method
     *
     * @param   m       the member to check
     * @param   context the context of the method access
     */
    protected boolean memberNeedsPrivilegedAccess(
        CMember m, CContextType context ) {
        if (m == null)
            return false;
        CMemberHost mClass = m.host();
        CClass callerClass = context.getClassContext().getHostClass();
            // either the method is private and in a different class
            // from the caller ...
        return (m.isPrivate() && !mClass.equals(callerClass))
                // or the method is protected and in a different package
                // from the caller and has a receiver that is
                // a superclass of the caller's class
            || (m.isProtected()
                && !m.getCCompilationUnit().packageName().equals(
                    callerClass.getCCompilationUnit().packageName()
                    )
                && !callerClass.descendsFrom( m.owner() ));
    }


    /**
     * Returns a method that allows getting or setting the value of
     * <code>field</code> from context <code>context</code>.  Called
     * when a legal access to the field is made from a context that
     * does not have access to the field in the JVM, i.e., an inner or
     * outer class accessing a private field or a superclass field.  */
    public CFieldAccessorMethod 
        getterOrSetterMethodFor(CField field, CExpressionContextType context) {
        if (context.isLeftSide()) {
            // if this field access is part of an increment or
            // decrement expression, find (but don't return) a
            // getter too.  NOTE: the assumption here is that
            // isIncDec => isLeftSide
            if (context.isIncDec()) {
                CFieldAccessor f = findOrCreateGetterMethod( field );
            }               
            return findOrCreateSetterMethod( field );
        }
        else {
            return findOrCreateGetterMethod( field );
        }
    }

    /* Find an applicable getter/setter method for the given field in
       the given set of methods.  Return null if none exists.  The
       given CMethodSet should contain only CFieldAccessorMethods. */
    private CFieldAccessorMethod findAccessorMethod( CField f,
                                                     CMethodSet ms ) {
        CAbstractMethodSet.Iterator i = ms.iterator();
        while( i.hasNext() ) {
            CFieldAccessorMethod fa = (CFieldAccessorMethod) i.next();
            if (fa.isApplicableTo(f)) {
                return fa;
            }
        }
        return null;
    }
    
    /* Find a getter method for the given field, returning null if
       none is found. This method is called instead of
       findOrCreateGetterMethod (see below) when it's too late to
       create a new method in the class, so the getter is expected to
       have already been created.  */
    public CFieldAccessorMethod findGetterMethod( CField f ) {
        return findAccessorMethod( f, fieldGetterMethods );
    }

    /* Find a getter method for the given field, or create one for it
       if none exists. */
    public CFieldAccessorMethod findOrCreateGetterMethod( CField f ) {
        CFieldAccessorMethod fa = findAccessorMethod( f, fieldGetterMethods );
        if (fa == null) {
                // make a new field getter method
            CFieldGetterMethod getter =
                new CFieldGetterMethod(f, this);
            fieldGetterMethods.addMethod(getter);
            return getter;
        }
        else
            return fa;
    }

    /* Find a setter method for the given field, or create one for it
       if none exists. */
    public CFieldAccessorMethod findOrCreateSetterMethod( CField f ) {
        CFieldAccessorMethod fa = findAccessorMethod( f, fieldSetterMethods );
        if (fa == null) {
                // make a new field setter method
            CFieldSetterMethod setter =
                new CFieldSetterMethod(f, this);
            fieldSetterMethods.addMethod(setter);
            return setter;
        }
        else
            return fa;
    }


    /**
     * Returns a wrapper method for forwarding qualified super method
     * invocations from an inner class to the super class of this.
     * Creates and adds the wrapper method to this if it does not
     * already exist.
     *
     * <pre><jml>
     * requires this.getSuperClass() == method.owner();
     * </jml></pre>
     * */
    public CMethod wrapSuperCallTo(CMethod method) throws UnpositionedError {
        return wrapCallTo(method,true);
    }
    
    /**
     * Returns a package-protected wrapper method for forwarding calls
     * to the given method.  The wrapper is needed for inner or outer
     * classes for which the virtual machine denies access to
     * <code>method</code>.  These denials stem from the virtual
     * machine not understanding inner classes.  For example, private
     * methods of an outer class do not appear to be accessible from
     * an inner class in the JVM. */
    private CMethod wrapCallTo(CMethod method, boolean isSuper) 
        throws UnpositionedError 
    {
        CAbstractMethodSet.Iterator i = wrapperMethods.iterator();
        while( i.hasNext() ) {
            CWrapperMethod wm = (CWrapperMethod) i.next();
            if (wm.isApplicableTo(method,isSuper)) {
                // found an existing wrapper for the method
                return wm;
            }
        }
        // make a new wrapper method
        CWrapperMethod wrapper = new CWrapperMethod(method, this, isSuper);
        wrapperMethods.addMethod(wrapper);
        return wrapper;
    }

    /**
     * Adds a synthetic redirector method to this.
     */
    public void addRedirector(CSourceRedirectorMethod redir) {
        redirectorMethods.addMethod(redir);
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Generate the code in a class file.  Uses several protected
     * helper methods so subclasses can override portions of the
     * behavior.
     *
     * @param   dest    the destination directory of the class file
     * @see #genInterfaces()
     * @see #genMethods()
     * @see #genFields()
     * @see #genInners() */
    public void genClassFile( Destination dest ) 
        throws IOException, ClassFileFormatException 
    {
        long mods = access().modifiers() & ~ACC_STATIC;
        if (isNestedType()) {
            if (hasFlag(mods, ACC_PROTECTED)) {
                // protected nested classes are made public since the
                // JVM does not understand the protected modifier on a
                // class
                mods &= ~ACC_PROTECTED;
                mods |= ACC_PUBLIC;
            } else if (hasFlag(mods, ACC_PRIVATE)) {
                // private nested classes are made package-accessible
                // since the JVM does not understand the private
                // modifier on a class
                mods &= ~ACC_PRIVATE;
            }
        }
        ClassInfo classInfo = 
            createClassInfo(mods,
                            superClass == null ? 
                            null : superClass.getCClass().qualifiedName(),
                            sourceFile);

        classInfo.write( dest );
   }

    /**
     * Creates an instance of <code>ClassInfo</code>. This is a
     * factory method for subclasses to generate a refined ClassInfo
     * object.
     */
    protected ClassInfo createClassInfo(long modifiers,
                                        String superClass,
                                        File sourceFile) {
        return new ClassInfo((short) modifiers,
                             qualifiedName(),
                             superClass,
                             getSuperClass()==null?null:getGenericSignature(),
                             genInterfaces(),
                             genFields(),
                             genMethods(),
                             genInners(),
                             genCustomAttributes(),
                             sourceFile,
                             false);
    }

    /**
     * Returns an array representing (for bytecode) the qualified
     * names of the implemented interfaces.  */
    protected ClassConstant[] genInterfaces() {
        ClassConstant[] interfaces = new ClassConstant[this.interfaces.length];

        for (int i = 0; i < this.interfaces.length; i++) {
            interfaces[i] = 
                new ClassConstant( this.interfaces[i].qualifiedName() );
        }

        return interfaces;
    }

    /**
     * Returns an array representing all the methods for bytecode. */
    protected MethodInfo[] genMethods() {
        return methods.buildMethodInfo();
    }

    /**
     * Returns an array representing all the fields for bytecode. */
    protected FieldInfo[] genFields() {
        return fields.buildFieldInfo();
    }


    /**
     * Returns an attribute list containing all the custom attributes
     * to be added to the top-level classinfo generated for this.
     * Regular Java attributes will be added later.  */
    protected final AttributeList genCustomAttributes() {
        AttributeList accum = new AttributeList();
        // call a helper so we can use downcalls
        appendCustomAttributes(accum);
        return accum;
    }


    /**
     * Appends any custom attributes for this class to the attribute
     * list represented by <code>accum</code>.  */
    protected void appendCustomAttributes(AttributeList accum) {
	ArrayList gfInfo = new ArrayList();
	genGenericFunctions(gfInfo);
	if (gfInfo.size() > 0) {
	    GenericFunctionInfo[] gfInfoArray = 
		new GenericFunctionInfo[gfInfo.size()];
	    gfInfo.toArray(gfInfoArray);
	    accum.add( new GenericFunctionsAttribute(gfInfoArray) );
	}
	
	
	// AHS: Add JUTS version
	if (CTopLevel.getCompiler().universeBytecode()) {
                // Use normal class attribute
	        accum.add(new UniverseClassAttribute(
                                  CTopLevel.getCompiler().universeVersion(),
                                  CTopLevel.getCompiler().universeDynChecks()));	        
        }

        if (CTopLevel.getCompiler().universeBytecodeAnnotations()) {
                // use Java 5 annotations
	        Annotation version = new Annotation("org.multijava.universes.UniverseClass");
	        version.addElement("version", new AnnotStringElementValue(
	                CTopLevel.getCompiler().universeVersion()));
	        version.addElement("hasRuntimeSupport", new AnnotBooleanElementValue(
	                CTopLevel.getCompiler().universeDynChecks()));
                
	        ArrayList annots = new ArrayList();
	        annots.add(version);
	        accum.add(new RuntimeVisibleAnnotationsAttribute(annots));
	}
    }

    /**
     * Accumulates an array of all the required gf_info data
     * structures for this.  */
    protected void genGenericFunctions(ArrayList accum) {
        for (int i = 0; i < dispClassTypes.length; i++) {
            dispClassTypes[i].getCClass().genGenericFunctions(accum);
        }       
        
        methods.buildGenericFunctionInfo(accum);
    }

    // ----------------------------------------------------------------------
    // INNER CLASS SUPPORT
    // ----------------------------------------------------------------------

    /**
     * Returns an array representing (for bytecode) the type
     * signatures of the inner and nested classes.  */
    protected InnerClassInfo[] genInners() {
        // construct a set containing all the inners that must be
        // added to the InnerClasses attribute (JVM2 4.7.5)

        CClassType[] inners = innerClassesForAttribute();
        for (int i = 0; i < inners.length; i++) {
            visibleNestedTypes.add( inners[i] );
        }

        if (isNestedType()) {
            // call method instead of using field since the method
            // ensures that the field is initialized
            visibleNestedTypes.add( getType() );
        }

        InnerClassInfo[] result = 
            new InnerClassInfo[ visibleNestedTypes.size() ];
        Iterator iter = visibleNestedTypes.iterator();
        int i = 0;
        while (iter.hasNext()) {
            result[i] = ((CClassType) iter.next()).getCClass().genInnerInfo();
            i++;
        }

        return result;
    }
    
    /**
     * Collects all the inner classes that must be added to the
     * InnerClasses attribute. This is a hook method so that a
     * subclass can override it; e.g., JML overrides this method to
     * combine inner classes through the refinement chain.
     */
    protected CClassType[] innerClassesForAttribute() {
        return innerClasses;
    }

    /**
     * Generates the inner class info record for this.
     *
     * <pre><jml>
     * requires isNestedType();
     * </jml></pre> */
    public InnerClassInfo genInnerInfo() {
        return new InnerClassInfo( qualifiedName(),
                                   isMember() ?
                                   owner().qualifiedName() : null,
                                   isAnonymous() ? null : ident(),
                                   (short) access().modifiers() );
    }
            
    /**
     * Mutates the given array to add any synthetic parameters for
     * outer local variables referenced within the class represented
     * by this.
     *
     * <pre><jml>
     * requires (* this represents an anonymous class or 
     *             a local class declaration *);
     * requires (* params is the parameter array for a constructor of this *);
     * ensures (* params includes the types of any outer local variables 
     *            required to instantiate this *);
     * </jml></pre>
     *
     */
    public CType[] genConstructorArray(CType[] params) {
        throw new RuntimeException();
    }

    /**
     * Adds synthetic parameters to the code sequence for a call to
     * this class's constructor.  In other words, it does nothing,
     * except for anonymous classes and inner class declarations that
     * reference final local variables.  In those cases, this adds
     * code to push the local variable values onto the stack.
     * 
     * @param code the code sequence being generated */
    public void genSyntheticParams( CodeSequence code ) {
        // Nothing to do here.  Outer local access is only needed for
        // calls to anonymous or local class constructors, which must
        // necessarily originate from the same compilation unit.  Thus
        // if we are compiling such a call we must have a CSourceClass
        // representing the target of the call.
    }

    /**
     * Adds synthetic parameters to the code sequence for an explicit
     * constructor invocation to this class's constructor.  In other
     * words, it does nothing, except for anonymous classes and inner
     * class declarations that reference final local variables.  In
     * those cases, this adds code to push the local variable values
     * onto the stack.
     *
     * @param code              the code sequence being generated 
     * @param constructor       the signature of the calling constructor
     */
    public void genSyntheticParamsForExplicitThis(CodeSequence code,
                                                  final CMethod constructor) {
        // Nothing to do here.  Outer local access is only needed for
        // calls to anonymous or local class constructors, which must
        // necessarily originate from the same compilation unit.  Thus
        // if we are compiling such a call we must have a CSourceClass
        // representing the target of the call.
    }

    /**
     * For an explicit super class constructor invocation, returns the
     * expressions that must be evaluated in the body of the
     * constructor <code>caller</code> of the class
     * <code>callerClass</code> to add the outer local variable values
     * required by this to the stack.
     *
     * <pre><jml>
     * requires callerClass.descendsFrom(this);
     * requires caller.isConstructor();
     * ensures \result != null;
     * </jml></pre>
     * */
    public JExpression[] 
        getSyntheticParamsForExplicitSuper( CSourceClass callerClass,
                                            CMethod caller ) {
        // Outer local access is only needed for calls to anonymous or
        // local class constructors, which must necessarily originate
        // from the same compilation unit.  Thus if we are compiling
        // such a call we must have a CSourceClass representing the
        // target of the call.
        return new JExpression[0];
    }

    /** Generate any RMJ annotations */
    public void rmjAnnotate(RMJAnnotation rmjInfo)
    {
        // by default, no annotations to generate
    }

    // -------------------------------------------------------------
    // OBSERVER SUPPORT
    // -------------------------------------------------------------

    /**
     * Registers the given observer to be notified when 
     * {@link #setCheckedMembers()} is called.
     */
    public void register( Observer o ) {
        if (observerSet == null) {
            observerSet = new HashSet(2);
        }
        
        observerSet.add(o);
    }

    public interface Observer {
        void signatureCompleted(CClass self);
    }
 
    /** 
     * Indicates whether this class is Generic Class
     */
    public boolean isGenericClass() {
        return typevariables.length > 0;
    }

    // ----------------------------------------------------------------------
    // PRIVATE INITIALIZATIONS
    // ----------------------------------------------------------------------

    public final static CClass  CLS_UNDEFINED = 
        new CSourceClass( null, new MemberAccess(null, null, 0L),  
                          TokenReference.NO_REF, 
                          "<gen>", "<gen>", CTypeVariable.EMPTY,false, false, false);

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private final Main  compiler;
    private final File sourceFile;

    //@ spec_public
    private final String        qualifiedName;

    //@ spec_public
    private CClassType[]        interfaces;

    //@ spec_public
    private CClassType          type;

    //@ spec_public
    private CClassType[]        innerClasses;

    //@ spec_public
    private CClassType          superClass;

    /**
     * Tracks the nested and inner classes referred to within this
     * class (whether or not their declarations are lexically enclosed
     * in this class).
     *
     * <pre><jml>
     * private invariant (\forall Object o; visibleNestedTypes.contains(o);
     *                    o instanceof CClassType && ((CClassType)o).isNestedType() );
     * </jml></pre>
     * */
    private HashSet visibleNestedTypes = new HashSet();

    /**
     * Maps names to field signature for all the fields of this class.
     * */
    private CFieldTable         fields;

    /**
     * The methods of this class after dispatcher wrapping. */
    //@ spec_public
    protected CMethodSet methods;

    /**
     * The methods of this class as generated from the declarations
     * after the checkInterface pass but before dispatcher wrapping.  */
    private CMethodSet          protoMethods;

    /**
     * The wrapper methods of this class
     */
    private CMethodSet          wrapperMethods = new CMethodSet();

    /**
     * The field getter methods of this class
     */
    private CMethodSet          fieldGetterMethods = new CMethodSet();

    /**
     * The field setter methods of this class
     */
    private CMethodSet fieldSetterMethods = new CMethodSet();

    /**
     * Redirectors for inherited external methods that implement a regular
     * Java interface. */
    private CMethodSet redirectorMethods = new CMethodSet();

    /**
     * The synthetic methods used to implement assertions
     */
    private CMethod assertHelperMethod = null;
    private CAssertStaticInitMethod assertStaticInitMethod = null;

    /**
     * Contains a mapping from contexts to the the set of augmenting
     * methods for this class in each context.  */
    private CAugmentationMap    augMap = new CAugmentationMap();

    /**
     * Used for error reporting.  */
    private static final TokenReference NO_TOKEN = TokenReference.NO_REF;

        // Used for giving each access$ method a unique name.
    private static int accessID = 0;

    private int syntheticIndex = 0;
    
    /**
     * Stores the dispatcher classes contained within the class or
     * generic function anchor represented by this.  This information
     * is also included in <code>innerClasses</code>, but is stored
     * separately here for use in generating MultiJava-specific
     * bytecode attributes.  */
    //@ spec_public
    protected CClassType[] dispClassTypes;

    /**
     * Stores the observers that are to be notified when 
     * {@link #setCheckedMembers()} is called.  Lazily initialized.
     */
    private /*@ spec_public @*/ HashSet observerSet;
    private CTypeVariable[]  typevariables = CTypeVariable.EMPTY;

     /**
     * Flag is true if we are not generating a synthetic outer 'this' field
     * for this inner class, because the outer this is not accessible.  This is
     * the case when the outer this has not yet been fully constructed --
     * the inner class is created in the outer class's constructor, before the
     * superclass constructor has been called.
     */
     private boolean syntheticOuterThisInaccessible = false;

}
