/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClassContext.java,v 1.24 2006/12/11 05:55:56 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the context for a class during checking
 * passes (checkInterface, checkInitializers, typecheck).
 * @see org.multijava.mjc.CContextType 
 */
public class CClassContext extends CContext implements CClassContextType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * @param	parent		the parent context or null at top level
     * @param	self		the corresponding clazz
     */
    public CClassContext( CContextType parent, CClass self ) {
	super(parent);

	this.self = self;
    }

    /**
     * Verifies that all final and all non-null fields are initialized 
     * and all abstract methods are appropriately implemented.
     *
     * @exception UnpositionedError if any checks fail */
    public void checkingComplete( JTypeDeclarationType decl,
				  CVariableInfoTable staticC,
				  CVariableInfoTable instanceC,
				  CVariableInfoTable[] constructorsC )
	throws UnpositionedError
    {
	JFieldDeclarationType[]	fields = decl.fields();
	CExpressionContextType dummyContext =
		new CExpressionContext(
		    new CFlowControlContext(
			new CMethodContext(this, null),
			0, null ) );

	for (int i = 0; i < fields.length; i++) {
	    CSourceField	field = (CSourceField)fields[i].getField();

		// create a dummy expression context for use in this lookup
		// defined once in super classes ?
	    if( self.isFieldRedefined( field.ident(), dummyContext ) ) {
		reportTrouble( new CWarning( decl.getTokenReference(),
					     MjcMessages.FIELD_RENAME_SUPER,
					     field.ident(),
					     null ));
	    }

	    if( field.isFinal() || (CTopLevel.getCompiler().nonNullTypes() && field.isDeclaredNonNull()) ) {
	        if (Main.contextBehavior().noBodyOK(this)) {
		    // It is ok to omit the initializer if bodies are not required for methods
		} else if( field.isStatic() ) {
		    check( staticC != null &&
			   staticC.getState(field.index()).
			       isDefinitelyAssigned(),
			   (field.isFinal() ? MjcMessages.UNINITIALIZED_FINAL_FIELD : MjcMessages.UNINITIALIZED_NONNULL_FIELD),
			   field.ident() );
		} else {
		    if( !(instanceC != null &&
			  instanceC.getState(field.index()).
			      isDefinitelyAssigned())) 
		    {
			// if field is a blank final instance field
			// then it must be initialized in all constructors
			for( int j = 0; j < constructorsC.length; j++ ) {
			    check( constructorsC[j].getState( field.index() ).
				           isDefinitelyAssigned(),
			   	   (field.isFinal() ? MjcMessages.UNINITIALIZED_FINAL_FIELD : MjcMessages.UNINITIALIZED_NONNULL_FIELD),
				   field.ident() );
			}
		    }
		}
	    } else {
		if( field.isStatic() ) {
		    if( staticC != null &&
			staticC.getState(field.index()).isDefinitelyAssigned()) {
			reportTrouble(
			     new CWarning( decl.getTokenReference(),
					   MjcMessages.UNINITIALIZED_FIELD,
					   field.ident() ));
		    }
		} else {
		    if( !(instanceC != null && 
			  instanceC.getState( field.index() ).
			      isDefinitelyAssigned()) )
			{
			// blank field should probably be initialized
			// in all constructors
			for( int j = 0; j < constructorsC.length; j++ ) {
			    if( !constructorsC[j].
				    getState( field.index() ).
				        isDefinitelyAssigned() ) {
				reportTrouble(
				  new CWarning(decl.getTokenReference(),
					       MjcMessages.UNINITIALIZED_FIELD,
					       field.ident() ));
				break;
			    }
			}
		    }
		}
	    }
	}

	if( !self.isAbstract() ) { // !HERE!note6
	    // check that all (internal) abstract methods are implemented 

	    // first get all the internal abstract methods of this
	    // class and all superclasses
	    CMethodSet.MethodArgsPair[]	pairs = self.getAbstractMethods();

	    // now for each of the abstract methods we verify that
	    // there is a concrete implementation
	    for (int i = 0; i < pairs.length; i++) {
		CMethod m = pairs[i].getMethod();
		CClassType[] abstractargs = pairs[i].getArgs();
		//	CType[] specializers = 
		//    CSpecializedType.tupleOfSpecializersFrom
		//    ( m.parameters(), abstractargs );
		//	CClassType[] argss = self.getType().getArguments();
		// CType[] actuals = new CType[specializers.length];
// 		for (int j = 0; j<specializers.length; j++) {
// 		    if(specializers[j].isTypeVariable()){
// 			actuals[j]=self.getSubstitution
// 			    ((CTypeVariable)specializers[j],args);
// 		    }else {
// 			actuals[j]=specializers[j];
// 		    }
// 		}
		CType[] specializers = CSpecializedType.tupleOfSpecializersFrom
		    (m.parameters(), abstractargs);
		CMethodSet ms = 
		    self.lookupMethodOrSet( m.ident(), specializers,
					    self.getType().getArguments(),
					    this );

		// (if there are multiple implementations, e.g. in the
		// case of ambiguously defined RMJ methods, then each
		// of them should be non-abstract.)
		for (int j = 0; j < ms.size(); j++) {
		    CMethod m2 = ms.getMethod(j);
		    if (m2.isAbstract()) {
			// XXX disabled for now, since it's not fully
			// implemented
			if (false && getCompiler().RMJ()) {
			    // this might only be a warning, if there
			    // are concrete multimethods overriding
			    // this abstract method, which cover all
			    // the cases.  first, it must be the case
			    // that this method has at least one
			    // abstract argument, so that other
			    // multimethods can override it.
			    boolean foundAbstract = false;
			    for (int k = 0; k < specializers.length; k++) {
				CType spec = specializers[k];
				if (spec.isClassType() &&
				    !((CClassType) spec).isConcrete()) {
				    // we found an abstract argument
				    foundAbstract = true;
				    break;
				}
			    }
			    if (foundAbstract) {
				// warn that an abstract method is
				// being declared in a concrete class,
				// and that it needs to be overridden
				// by concrete multimethods for all
				// concrete argument tuples
				reportTrouble(
				    new CWarning(
					decl.getTokenReference(),
					MjcMessages
					.RMJ_ABSTRACT_METHOD_IN_CONCRETE_CLASS,
					self.ident(), 
					m2) );
				// at this point, we should identify
				// that concrete overriding methods
				// exist in this class to cover all
				// visible concrete tuples of the
				// argument specializers.  but we
				// don't *have* to, since we can rely
				// on the load-time checking to check
				// all the cases.  this would only be
				// early warning about a subset of the
				// combinations.
			    } else {
				// none of the arguments were
				// abstract; we really do need to
				// reject this class
				fail(MjcMessages
				     .ABSTRACT_METHOD_IN_CONCRETE_CLASS, 
				     self.ident(), 
				     m2);
			    }
			} else {
			    fail(MjcMessages.ABSTRACT_METHOD_IN_CONCRETE_CLASS,
				 self.ident(), 
				 m2);
			}
		    }
		}
	    }
	}
	// check all interfaces methods are implemented
	//CMethod[] methods = self.getInterfaceMethods();
	CMethodSet.MethodArgsPair[] pairs = self.getInterfaceMethods();
	for( int i = 0; i < pairs.length; i++ ) {
	    CMethod m = pairs[i].getMethod();
	    CClassType[] argss = pairs[i].getArgs();
	    String interfaceMethodName = m.ident();
	    CType[] arg = CSpecializedType.tupleOfSpecializersFrom
		(m.parameters(), argss);
	    if( interfaceMethodName != JAV_STATIC_INIT  &&
		getCompiler().options().multijava()) {
		resolveMaybeExtMethodRef( interfaceMethodName );

		CMethodSet ms =
		    self.lookupMethodOrSet(
			interfaceMethodName,
			arg,
			self.getType().getArguments(),
			this );

		// (if there are multiple implementations, e.g. in the
		// case of ambiguously defined RMJ methods, then each
		// of them should be non-abstract.)
		for (int j = 0; j < ms.size(); j++) {
		    CMethod m2 = ms.getMethod(j);
		    //XXX need to extend this to allow RMJ warning
		    check( !m2.isAbstract() || self.isAbstract(), 
			   MjcMessages.CLASS_SHOULD_IMPLEMENT, 
			   self.ident(), 
			   m2.ident() );
		    // if an interface is implemented by inheriting an
		    // external method, then the inherited method
		    // becomes pleomorphic and a redirector is needed:
		    if (m2.isExternal() && m2.owner() != self) {
			 CSourceRedirectorMethod redir =
			     new CSourceRedirectorMethod(m2, decl.
							 getTokenReference(),
							 this);
			 self.addRedirector(redir); 
		    }
		    
		}
	    }
	}
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (NEW CONTEXT)
    // ----------------------------------------------------------------------

    public CMethodContextType createMethodContext(CMethod self) {
	return new CMethodContext(this, self);
    }

    public CConstructorContextType createConstructorContext(CMethod self) {
	return new CConstructorContext(this, self);
    }

    public CInitializerContextType createInitializerContext(CMethod self) {
	return new CInitializerContext(this, self);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Indicates whether this context is in a loop.
     */
    public final boolean isInLoop() {
	// class contexts inside loop contexts are typechecked as if
	// they are not in a loop (because the local or anonymous
	// classes that are declared inside loops are really only
	// declared once)
	return false;
    }

    /**
     * Add an initializer to this context
     */
    public void addInitializer() {
	hasInitializer = true;
    }

    /**
     * Returns true if this class need initializers
     */
    public boolean hasInitializer() {
	return hasInitializer;
    }


    /**
     * Returns the fieldInfo table.
     *
     * @return an field into table.
     */
    public CVariableInfoTable fieldInfo() {
	return fieldInfo;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (INFOS)
    // ----------------------------------------------------------------------

    /**
     * Replaces the local field info for fields in positions 0 up to
     * <code>pos</code> with the info in <code>replacement</code>.  */
    public void replaceFieldInfoUpTo( int pos, 
				      CVariableInfoTable replacement ) {
	fieldInfo.replaceUpTo( pos, replacement );
    }

    /**
     * Marks the field with the given descriptor as definitely assigned
     * to in this context.  */
    public void initializeField(VariableDescriptor varDesc) {
	fieldInfo.initialize(varDesc);
    }

    /**
     * Indicates whether the field in the given position is
     * definitely assigned to in this context.  */
    public boolean isFieldDefinitelyAssigned(int pos) {
	return fieldInfo.getState(pos).isDefinitelyAssigned();
    }

    /**
     * Returns the current field information for the most recently
     * checked member of this context.  It then clears the field info
     * stored in this object.  The pattern of use is to pass this
     * context into a <code>typecheck</code> method.  That method mutates this
     * context to include the variable information extracted from the
     * statement being checked.  Upon return from the
     * <code>typecheck</code> the caller calls
     * <code>fieldInfoTable</code> to extract the information.  This
     * is kludgey and one wonders why the <code>typecheck</code>
     * methods doesn't just return the <code>CVariableInfoTable</code>
     * instance. !FIXME! */
    public CVariableInfoTable fieldInfoTable() {
	CVariableInfoTable tmp = fieldInfo;

	fieldInfo = new CVariableInfoTable(); // KOPI: rename it reset

	return tmp;
    }

    /**
     * Sets the field information for this to match that given by
     * <code>fieldInfo</code>.  Called before typechecking
     * constructors with the argument equal to the field assignment
     * information after checking initializers.  */
    public void setFieldInfoTable(CVariableInfoTable fieldInfo) {
	if (fieldInfo != null) {
	    this.fieldInfo = fieldInfo.duplicate();	     
	}
    }

    public void markAllFieldsAsInitialized() {
	for (int i = getHostClass().getFieldCount() - 1; i >= 0; i--) {
	    fieldInfo.getState(i).initialize();
	}
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------

    /**
     * Searches for a class with the given simple name according the
     * procedure in JLS2 6.5.5.  This method implements the second and
     * third bullet points of the procedure.  That is "Otherwise, if
     * the simple type name occurs within the scope of exactly one
     * visible member type, then the simple type name denotes that
     * member type;" and "Otherwise, if the simple type name occurs
     * within the scope of more than one visible member type, then the
     * name is ambiguous as a type name; a compile-time error occurs."
     * If the class is not found according to these constraints, then 
     * the surrounding context is called to check the remaining bullets.
     * 
     * @param	name	the class name, without qualifiers
     * @return	the class if found, null otherwise
     * @exception UnpositionedError if search fails */
    public CClass lookupClass(String name) throws UnpositionedError {
	CClass clazz = self.lookupClass( name, this );
	return (clazz != null) ? clazz : super.lookupClass(name);
    }
    
    public /*@nullable*/ CTypeVariable lookupTypeVariable(/*@non_null*/ String ident) 
	throws UnpositionedError {
	CTypeVariable tv = getOwnerClass().lookupTypeVariable(ident);
	if (tv!=null) {
	    return tv ;
	} else {
	    return getParentContext().lookupTypeVariable(ident);
	}
    }
    
    /**
     * Searches for the most specific method <em>when no receiver is
     * explicit at the call site</em>.
     *
     * @param	ident		method name
     * @param	params		method parameters
     * @param	context		the context of the class containing
     * 				the method call
     * @return	the method or null if not found
     * @exception UnpositionedError this error will be positioned soon */
    public CMethod lookupMethod( String ident, CType[] params,
				 CClassContextType context ) 
	throws UnpositionedError 
    {
	// see if the method is in this class
	CMethod	method = self.lookupMethod( ident, params,CClassType.EMPTY,
					    context );
	if( method != null ) {
	    return method;
	}

	// method is not in this class so try surrounding classes
	return super.lookupMethod( ident, params, context );
    }

    /**
     * Searches for the most specific method(s) <em>when no receiver is
     * explicit at the call site</em>.
     *
     * @param	ident		method name
     * @param	params		method parameters
     * @param	context		the context of the class containing
     * 				the method call
     * @return	the methods or null if not found
     * @exception UnpositionedError this error will be positioned soon */
    public CMethodSet lookupMethodOrSet( String ident, CType[] params,
					 CClassContextType context ) 
	throws UnpositionedError 
    {
	// see if the method(s) are in this class
	CMethodSet methods = self.lookupMethodOrSet( ident, params, 
						     self.getType().getArguments(),
						     context );
	if( methods != null && methods.size() > 0 ) {
	    return methods;
	}

	// method is not in this class so try surrounding classes
	return super.lookupMethodOrSet( ident, params, context );
    }

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupOuterField(String ident,
					   CExpressionContextType context)
	throws UnpositionedError {
	CFieldAccessor field =  parent.lookupField(ident, context);
	return field != null ? field : parent.lookupOuterField(ident, context);
    }

    /**
     * lookupField
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from an ident in current context
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupField(String ident,
				      CExpressionContextType context)
	throws UnpositionedError {
	return self.lookupField( ident, context );
    }

    /**
     * Finds a local variable with the given name that appears outside
     * the current lexical contour.
     *
     * @param	ref		a token reference used to build a new
     *				<code>JOuterLocalVariableExpression</code>
     * @param	ident		the name of the outer variable
     * @return a variable from an ident in the surrounding context, or null
     *		if not found */
    public JExpression lookupOuterLocalVariable(TokenReference ref, 
						String ident) {
	if (parent instanceof CFlowControlContextType 
	    || parent instanceof CExpressionContextType) {
	    JLocalVariable var = parent.lookupLocalVariable(ident);
	    if (var != null) {
		return new JOuterLocalVariableExpression(ref, var, 
							 getHostClass());
	    } else {
		return super.lookupOuterLocalVariable(ref, ident);
	    }
	}
	return null;
    }

    /**
     * lookupLocalVariable
     * @param	ident		the name of the local variable
     * @return	a variable from an ident in current context
     */
    public JLocalVariable lookupLocalVariable(String ident) {
	return null;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (TREE HIERARCHY)
    // ----------------------------------------------------------------------

    /**
     * getClassContext
     * @return	the near parent of type CClassContextType
     */
    public CClassContextType getClassContext() {
	return this;
    }

    /**
     * Returns the CClass representing the signature of the class
     * containing this context.
     *
     * @return	the CClass representing the signature of the class
     */
    public CClass getHostClass() {
	return self;
    }

    /**
     * Returns the CClass representing the signature of the class
     * that is the logical owner of this context.  Useful for external
     * methods where the host is the generic function anchor but the
     * owner is the method's receiver.
     *
     * @return	the CClass representing the signature of the class
     * @see org.multijava.mjc.CClassContextType#getHostClass()
     * @see org.multijava.mjc.CExtMethodContext#getHostClass()
     */
    public CClass getOwnerClass() {
	return self;
    }

    /**
     * Returns the signature of the nearest lexically enclosing
     * context that can host member declarations (i.e., a
     * <code>CClass</code> or a <code>CCompilationUnit</code>).  Used
     * for access checks, this is slightly different than
     * <code>getHostClass()</code>.  The later method returns an
     * anchor class signature for external methods, which is useful in
     * code generation and in typechecking external method bodies.  On
     * the other hand, this method returns a
     * <code>CCompilationUnit</code> for external methods, which is
     * useful in checking access between two differently named generic
     * functions declared in the same compilation unit.  !FIXME! It
     * may be possible to eliminate this method and just use
     * <code>getHostClass</code> now that private external methods are
     * duplicated in every anchor.
     *
     * @return	the nearest parent that can be coerced to a 
     * <code>CMemberHost</code>, i.e., <code>this.self</code> */
    public CMemberHost findNearestHost() {
	return self;
    }

    // ----------------------------------------------------------------------
    // CLASS AND EXTERNAL GF HANDLING
    // ----------------------------------------------------------------------

    /**
     * Registers with the surrounding context that a reference to
     * the given type occurs in this context. */
    public void registerVisibleType( CType type ) {
	// record any inner classes used in this class (for generating
	// the InnerClasses attribute in bytecode, see JVM2 4.7.5)
	if (type.isNestedType()) {
	    self.registerNestedType( type );
	}
	
	// can skip possible outer class contexts and just register
	// the type with the compilation unit
	getCompilationUnit().registerVisibleType( type );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	CVariableInfoTable	fieldInfo = new CVariableInfoTable();
    private	boolean		hasInitializer;
    protected final CClass self;
}
