/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClassContextType.java,v 1.15 2006/12/11 05:55:56 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the context for a class during checking
 * passes (checkInterface, checkInitializers, typecheck).
 * @see org.multijava.mjc.CContextType 
 */
public interface CClassContextType extends CContextType {

    /**
     * Verifies that all final fields are initialized and all abstract
     * methods are appropriately implemented.
     *
     * @exception UnpositionedError if any checks fail */
    void checkingComplete( JTypeDeclarationType decl,
				  CVariableInfoTable staticC,
				  CVariableInfoTable instanceC,
				  CVariableInfoTable[] constructorsC )
	throws UnpositionedError;

    // ----------------------------------------------------------------------
    // ACCESSORS (NEW CONTEXT)
    // ----------------------------------------------------------------------

    CMethodContextType createMethodContext(CMethod self);

    CConstructorContextType createConstructorContext(CMethod self);

    CInitializerContextType createInitializerContext(CMethod self);

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Add an initializer to this context
     */
    void addInitializer();

    /**
     * Returns true if this class need initializers
     */
    /*@ pure @*/ boolean hasInitializer();

    /**
     * Returns the fieldInfo table.
     *
     * @return an field into table.
     */
    CVariableInfoTable fieldInfo() ;

    // ----------------------------------------------------------------------
    // ACCESSORS (INFOS)
    // ----------------------------------------------------------------------

    /**
     * Replaces the local field info for fields in positions 0 up to
     * <code>pos</code> with the info in <code>replacement</code>.  */
    void replaceFieldInfoUpTo( int pos, 
				      CVariableInfoTable replacement );

    /**
     * Marks the field with the given descriptor as definitely assigned
     * to in this context.  */
    void initializeField(VariableDescriptor varDesc);

    /**
     * Indicates whether the field in the given position is
     * definitely assigned to in this context.  */
    boolean isFieldDefinitelyAssigned(int pos);

    /**
     * Returns the current field information for the most recently
     * checked member of this context.  It then clears the field info
     * stored in this object.  The pattern of use is to pass this
     * context into a <code>typecheck</code> method.  That method mutates this
     * context to include the variable information extracted from the
     * statement being checked.  Upon return from the
     * <code>typecheck</code> the caller calls
     * <code>fieldInfoTable</code> to extract the information.  This
     * is kludgey and one wonders why the <code>typecheck</code>
     * methods doesn't just return the <code>CVariableInfoTable</code>
     * instance. !FIXME! */
    CVariableInfoTable fieldInfoTable();

    /**
     * Sets the field information for this to match that given by
     * <code>fieldInfo</code>.  Called before typechecking
     * constructors with the argument equal to the field assignment
     * information after checking initializers.  */
    void setFieldInfoTable(CVariableInfoTable fieldInfo);

    void markAllFieldsAsInitialized();

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------

    /**
     * Searches for a class with the given simple name according the
     * procedure in JLS2 6.5.5.  This method implements the second and
     * third bullet points of the procedure.  That is "Otherwise, if
     * the simple type name occurs within the scope of exactly one
     * visible member type, then the simple type name denotes that
     * member type;" and "Otherwise, if the simple type name occurs
     * within the scope of more than one visible member type, then the
     * name is ambiguous as a type name; a compile-time error occurs."
     * If the class is not found according to these constraints, then 
     * the surrounding context is called to check the remaining bullets.
     * 
     * @param	name	the class name, without qualifiers
     * @return	the class if found, null otherwise
     * @exception UnpositionedError if search fails */
    /*@nullable*/ CClass lookupClass(/*@non_null@*/ String name) throws UnpositionedError;

    /**
     * search for the type variable that are accessible
     */
    /*@nullable*/ CTypeVariable lookupTypeVariable(/*@non_null@*/ String ident) throws UnpositionedError;

    /**
     * Searches for the most specific method <em>when no receiver is
     * explicit at the call site</em>.
     *
     * @param	ident		method name
     * @param	params		method parameters
     * @param	context		the context of the class containing
     * 				the method call
     * @return	the method or null if not found
     * @exception UnpositionedError this error will be positioned soon */
    /*@nullable*/ CMethod lookupMethod( /*@non_null@*/ String ident,  CType[] params,
			  /*@non_null@*/ CClassContextType context ) 
	throws UnpositionedError;

    /**
     * Searches for the most specific method(s) <em>when no receiver is
     * explicit at the call site</em>.
     *
     * @param	ident		method name
     * @param	params		method parameters
     * @param	context		the context of the class containing
     * 				the method call
     * @return	the method or null if not found
     * @exception UnpositionedError this error will be positioned soon */
    /*@nullable*/ CMethodSet lookupMethodOrSet( /*@non_null@*/ String ident, CType[] params,
				  /*@non_null@*/ CClassContextType context ) 
	throws UnpositionedError;

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    /*@nullable*/ CFieldAccessor lookupOuterField(/*@non_null@*/ String ident,
				    /*@non_null@*/ CExpressionContextType context)
	throws UnpositionedError;

    /**
     * lookupField
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from an ident in current context
     * @exception UnpositionedError	this error will be positioned soon
     */
    CFieldAccessor lookupField(/*@non_null@*/ String ident, /*@non_null@*/ CExpressionContextType context)
	throws UnpositionedError;

    /**
     * Finds a local variable with the given name that appears outside
     * the current lexical contour.
     *
     * @param	ref		a token reference used to build a new
     *				<code>JOuterLocalVariableExpression</code>
     * @param	ident		the name of the outer variable
     * @return a variable from an ident in the surrounding context, or null
     *		if not found */
    /*@nullable*/ JExpression lookupOuterLocalVariable(/*@non_null@*/ TokenReference ref, 
					 /*@non_null@*/ String ident);
    
    // ----------------------------------------------------------------------
    // ACCESSORS (TREE HIERARCHY)
    // ----------------------------------------------------------------------

    /**
     * getClassContext
     * @return	the near parent of type CClassContextType
     */
    /*@ pure non_null @*/ CClassContextType getClassContext();

    /**
     * Returns the CClass representing the signature of the class
     * containing this context.
     *
     * @return	the CClass representing the signature of the class
     */
    /*@ pure non_null @*/ CClass getHostClass();

    /**
     * Returns the CClass representing the signature of the class
     * that is the logical owner of this context.  Useful for external
     * methods where the host is the generic function anchor but the
     * owner is the method's receiver.
     *
     * @return	the CClass representing the signature of the class
     * @see org.multijava.mjc.CClassContextType#getHostClass()
     * @see org.multijava.mjc.CExtMethodContext#getHostClass()
     */
    /*@ pure non_null@*/ CClass getOwnerClass();

    /**
     * Returns the signature of the nearest lexically enclosing
     * context that can host member declarations (i.e., a
     * <code>CClass</code> or a <code>CCompilationUnit</code>).  Used
     * for access checks, this is slightly different than
     * <code>getHostClass()</code>.  The later method returns an
     * anchor class signature for external methods, which is useful in
     * code generation and in typechecking external method bodies.  On
     * the other hand, this method returns a
     * <code>CCompilationUnit</code> for external methods, which is
     * useful in checking access between two differently named generic
     * functions declared in the same compilation unit.  !FIXME! It
     * may be possible to eliminate this method and just use
     * <code>getHostClass</code> now that private external methods are
     * duplicated in every anchor.
     *
     * @return	the nearest parent that can be coerced to a 
     * <code>CMemberHost</code>, i.e., <code>this.self</code> */
    /*@ pure @*/ CMemberHost findNearestHost();

    // ----------------------------------------------------------------------
    // CLASS AND EXTERNAL GF HANDLING
    // ----------------------------------------------------------------------

    /**
     * Searches for any imported or private external generic functions.
     *
     * @param	ident	the name of the generic function to search for
     */
    void resolveMaybeExtMethodRef( /*@non_null@*/ String ident );

    /**
     * Registers with the surrounding context that a reference to
     * the given type occurs in this context. */
    void registerVisibleType( /*@non_null@*/ CType type );
}
