/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClassFQNameType.java,v 1.15 2008/03/16 15:52:42 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a class type when the class name first
 * appears as a forward reference <em>and the name is definitely fully
 * qualified</em> as in an import clause.  The internal structure must
 * be filled in later when enough information is available to find the
 * corresponding class declaration.
 * @see CType */
public class CClassFQNameType extends CClassNameType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a class type.  Clients
     * should call CTopLevel.getTypeRep(name,isFullyQualified) to get
     * a singleton for the given name, but made protected so that
     * JML can inherit and slightly alter behavior.
     * @param fullyQualifiedName the fully qualified name of the
     * class, with '/' not '.' separators
     * @see CClassType#lookup(String,boolean) */
    protected CClassFQNameType(String fullyQualifiedName) {
	//super(fullyQualifiedName);
	//this.fullyQualifiedName = fullyQualifiedName;
    // WMD the superclass will replace null by the defauls
	this(fullyQualifiedName,null,EMPTY_ARG);
    }

    protected CClassFQNameType(String fullyQualifiedName,
			       CUniverse univ, // WMD
                               CClassType[][] arguments){
	super(fullyQualifiedName,univ,arguments);
	this.fullyQualifiedName = fullyQualifiedName;
    }
    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the qualified name of this type.  Overrides the
     * superclass method to return the name without requiring
     * isChecked().  */
    public String qualifiedName() {
	return fullyQualifiedName;
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks that this type is valid, throwing an exception if it is
     * not.  For class types this check resolves type names into their
     * fully qualified versions and sets the class parameter so that
     * getCClass returns the appropriate singleton.
     *
     * @param		context		the context in which this type was
     *					declared
     * @exception	UnpositionedError	this unpositioned error
     *						should be positioned
     *						by the calling node of
     *						the AST, typically by
     *						calling the
     *						addPosition method of
     *						UnpositionedError 
     */
    public CType checkType(CContextType context) throws UnpositionedError {
	// maybeQualifiedName is set to null when the appropriate CClass
	// has been found, so the finding only needs to happen here if
	// maybeQualifiedName is non-null
	if (maybeQualifiedName == null) {
	    // The context may be different than the first time
	    // checkType() was called on this.  If so then this
	    // isn't registered as a visible type in the given
	    // context.  We resolve that here.
            checkInstantiation(context);
	    registerAsVisibleIn(context);
	    return this;
	}

	// Name resolution follows the procedure in JLS2, section
	// 6.5.5.
    name_res: 
	{
	    int index = maybeQualifiedName.lastIndexOf('/');
	    if( index < 0 ) {
		if (maybeQualifiedName.equals("java")) {
		    // !FIXME! There is currently a bug in name space
		    // resolution; One place it comes up is in
		    // repeated searching of the entire super class
		    // and interface hierarchy for a class named
		    // 'java'.  There is not likely to be such, so we
		    // short circuit that test.  What really is needed
		    // is to fix the problem.
		    if (!ignoredJavaAlready) {
			ignoredJavaAlready = true;
			//System.out.println("IGNORING JAVA; context? " + (context != null));
		    }
                    throw new UnpositionedError( MjcMessages.TYPE_UNKNOWN,
                                                 this.toString() );

		}
		if( context==null ) {
		    // if the context is null, we have no option other
		    // than to hope that the class was defined at the
		    // top level.
		    setClassFromNameOrDieTrying();
                    checkInstantiation(context);
		    return this; 
		}
		// JLS2 6.5.5.1 simple name
		String identifier = maybeQualifiedName;
		CClass clazz = context.lookupClass( identifier );
		if( clazz == null ) {
		    throw new UnpositionedError( MjcMessages.TYPE_UNKNOWN, 
						 this.toString() );
		}
		setClass( clazz );
		break name_res;
	    } else {
		// JLS2 6.5.5.2 qualified name 
	    
		// The prefix either names a package or an outer class (or
		// both, so there could be an ambiguity).
		String prefix = maybeQualifiedName.substring( 0, index );
		CClassType outerType;
		try {
		    // for a fully qualified name the prefix is fully qual.
		    outerType = CTopLevel.getTypeRep(prefix,true); 
		    outerType = (CClassType)(outerType.checkType(context));
		    if (maybeQualifiedName == null) {
			    // resolved this name recursively
			break name_res;
		    }
		} catch (UnpositionedError ue) {
		    // couldn't find an outer class, so we look in a
		    // package
		    setClassFromNameOrDieTrying();
		    break name_res;
		}

		// maybeQualifiedName may have been set to null by now

		CClass outer = outerType.getCClass();
		if( outer != null && maybeQualifiedName != null) {
		    // found an outer class with the right name :-|
		    CClass maybeResult = 
			outer.lookupClass( maybeQualifiedName.substring( index+1 ),
					   context );
		    if( maybeResult != null ) {
			// found an inner class with the right name :-)
			if( CTopLevel.classExists( maybeQualifiedName ) ) {
			    // found an ambiguity :-(
			    throw 
				new UnpositionedError( MjcMessages.TYPE_AMBIG2,
						       maybeQualifiedName, prefix );
			} else {
			    // no ambiguity <:-)
			    setClass( maybeResult );
			    break name_res;
			}
		    }
		}

		// did not find an inner class so our last hope is to
		// look in a package
		setClassFromNameOrDieTrying();
	    }
	} // end of name_res block

        checkInstantiation(context);
	registerAsVisibleIn(context);
	return this;
    }

    // ----------------------------------------------------------------------
    // PRIVILEGED DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * Cached version of the fully qualified name.  */
    private String fullyQualifiedName;

}
