/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClassNameType.java,v 1.31 2008/03/16 15:52:42 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.StringTokenizer;
import java.util.Vector;

import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CompilationAbortedError;
import org.multijava.util.compiler.FastStringBuffer;

/**
 * This class represents a class type when the class name first
 * appears as a forward reference.  The internal structure must be
 * filled in later when enough information is available to find the
 * corresponding class declaration.
 * @see CType */
public class CClassNameType extends CClassType {

    static boolean ignoredJavaAlready = false; // !FIXME!
    // Remove this when the !FIXME! below is removed

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /** Construct a class type.
     */
    CClassNameType(String maybeQualifiedName){
	this(maybeQualifiedName, null, EMPTY_ARG);
    }
    /**
     * Construct a class type.  Package accessible since clients
     * should call CTopLevel.getTypeRep(name,isFullyQualified) to get
     * a singleton for the given name.
     * @param maybeQualifiedName the name of the class, perhaps just
     * partially qualified or not qualified at all, but with '/' not
     * '.' separators if any separators are present
     * @see CClassType#lookup(String,boolean) */
    CClassNameType(String maybeQualifiedName, CUniverse univ,
                   CClassType[][] arguments) {
	super();

	// WMD
	if( univ == null )
	    this.universe = CUniverseServices.getDefaultVariable();
	else
	    this.universe = univ;

	this.arguments = arguments;
	assertTrue(maybeQualifiedName.indexOf('.') < 0);

	this.maybeQualifiedName = maybeQualifiedName.intern();
	this.originalQualifiedName = this.maybeQualifiedName;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Transforms this type to a string
     */
    public String toString() {
	return maybeQualifiedName != null ?
	    maybeQualifiedName.replace('/', '.') + printArgs(false) : 
	    super.toString();
    }
    
    /**
     * Print the type arguments of this class type
	 * 
	 * @param includeUniverseModifier Universe modifiers are printed iff true
     */
    public String printArgs(boolean includeUniverseModifier) {
	if (arguments == null 
	    || arguments.length == 0 
	    ||  arguments[arguments.length-1] == null 
	    || arguments[arguments.length-1].length == 0
            || isRawType()) {
	    return "";
	}
	StringBuffer        buffer = new StringBuffer();
	
	buffer.append('<');
	for (int i=0; i<arguments[arguments.length-1].length; i++) {
	    if (i > 0) {
		buffer.append(", ");
	    }
	    buffer.append(includeUniverseModifier
	    		? arguments[arguments.length-1][i].toVerboseString()
	    		: arguments[arguments.length-1][i].toString());
	}
	buffer.append('>');
	return buffer.toString();
    }

    public void appendGenericSignature(FastStringBuffer buff) {
	buff.append('L');
	buff.append(qualifiedName());
	if(isGenericType()){
	    buff.append('<');
	    for (int i=0; i<arguments[arguments.length-1].length; i++) {
		arguments[arguments.length-1][i].appendGenericSignature(buff);
	    }
	    buff.append('>');
	}
	buff.append(';');
    }
    /**
     *
     */
    public String qualifiedName() {
	return maybeQualifiedName == null ? super.qualifiedName() : maybeQualifiedName;
    }

    public String originalQualifiedName() {
	return originalQualifiedName == null ? super.originalQualifiedName()
	    : originalQualifiedName.replace('/', '.');
    }

    /**
     * Returns the class object associated with this type.
     *
     * If this type was never checked (i.e., associated with a CClass
     * singleton), then check it!
     *
     * @return the class object associated with this type */
    public CClass getCClass() { 
	if( maybeQualifiedName != null ) { // class of this hasn't been checked
	    try {
		CType dummy = checkType(null);
	    } catch (UnpositionedError cue) {
		throw new CompilationAbortedError("Fatal error - Unable to find"
						  + " a class for " + 
						  maybeQualifiedName + ": " + 
						  cue.getMessage());
	    }
	}

	CClass c = super.getCClass();
	return c;
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks that this type is valid, throwing an exception if it is
     * not.  For class types this check resolves type names into their
     * fully qualified versions and sets the class parameter so that
     * getCClass returns the appropriate singleton.
     *
     * @param		context		the context in which this type was
     *					declared
     * @exception	UnpositionedError	this unpositioned error
     *						should be positioned
     *						by the calling node of
     *						the AST, typically by
     *						calling the
     *						addPosition method of
     *						UnpositionedError 
     */
    public CType checkType(CContextType context) throws UnpositionedError {
	// maybeQualifiedName is set to null when the appropriate CClass
	// has been found, so the finding only needs to happen here if
	// maybeQualifiedName is non-null
	if (maybeQualifiedName == null) {
	    // The context may be different than the first time
	    // checkType() was called on this.  If so then this
	    // isn't registered as a visible type in the given
	    // context.  We resolve that here.
            checkInstantiation(context);
	    registerAsVisibleIn(context);
	    return this;
	}
	if (context == null) {
	    throw new CompilationAbortedError(
		"You have encountered a bug in the compiler.  " +
		"We apologize for the inconvenience.  Please " +
		"report the following information to " +
		"www.multijava.org: " + System.getProperty("line.separator") +
		"Cannot resolve a maybe-qualified " +
		"class name without context information:"
		+ maybeQualifiedName);
	}

	// Name resolution follows the procedure in JLS2, section
	// 6.5.5.
    name_res: 
	{
	    int index = maybeQualifiedName.lastIndexOf('/');
	    if( index < 0 ) {
		if (maybeQualifiedName.equals("java")) {
			// !FIXME! There is currently a bug in name
			// space resolution; One place it comes up is
			// in repeated searching of the entire super
			// class and interface hierarchy for a class
			// named 'java'.  There is not likely to be
			// such, so we short circuit that test.  What
			// really is needed it to fix the problem.
			if (!ignoredJavaAlready) {
			    ignoredJavaAlready = true;
			    //System.out.println("IGNORING JAVA; context? " + (context != null));
			}
                    throw new UnpositionedError( MjcMessages.TYPE_UNKNOWN,
                                                 this.toString() );

		}
		if( context==null ) {
			// if the context is null, we have no option other
			// than to hope that the class was defined at the
			// top level.
		    setClassFromNameOrDieTrying();
                    checkInstantiation(context);
		    return this;
		}
		// JLS2 6.5.5.1 simple name
		String identifier = maybeQualifiedName;
		CTypeVariable typevariable = context.lookupTypeVariable(identifier);
		if(typevariable != null) {
            // We cannot check in the grammar, whether we have a type variable or not.
			// Now we know that we are a TV and check that there is no Universe modifier.
			// However, implicit modifiers might have been added, because the defaulting also
			// could not decide whether it's a class or TV.
           	if( universe != null &&
               	!(universe instanceof CUniverseThis) &&
               	!(universe instanceof CUniverseImplicitPeer) &&
           		!(universe instanceof CUniverseImplicitReadonly)) {
           		throw new UnpositionedError(CUniverseMessages.TYPEVAR_WITH_UNIVERSES_FORBIDDEN,
           					this.toVerboseString());                    
            }

		    return typevariable.checkType(context);
		}
		CClass clazz = context.lookupClass( identifier );
		if( clazz == null ) {
		    throw new UnpositionedError( MjcMessages.TYPE_UNKNOWN, 
						 this.toString());
		}

                setClass(clazz);
		break name_res;
	    } else {
		// JLS2 6.5.5.2 qualified name 
	    
		// The prefix either names a package or an outer class (or
		// both, so there could be an ambiguity).
		String prefix = maybeQualifiedName.substring( 0, index );
		CClassType outerType;
		try {
		    CClassType[][] args;
		    if (arguments.length < 2) {
			args = CClassType.EMPTY_ARG;
		    } else {
			args = new  CClassType[arguments.length-1][];
			System.arraycopy(arguments, 0, args, 0, args.length);
		    }
		    outerType = CTopLevel.getTypeRep(prefix,args,false);
		    outerType = (CClassType)(outerType.checkType(context));
		    if (maybeQualifiedName == null) {
			    // resolved this name recursively
			break name_res;
		    }
		} catch (UnpositionedError ue) {
		    // couldn't find an outer class, so we look in a
		    // package
		    setClassFromNameOrDieTrying();
		    break name_res;
		}

		// maybeQualifiedName may have been set to null by now

		CClass outer = outerType.getCClass();
		if( outer != null && maybeQualifiedName != null) {
		    // found an outer class with the right name :-|
		    CClass maybeResult = 
			outer.lookupClass( maybeQualifiedName.substring( index+1 ),
					   context );
		    if( maybeResult != null ) {
			// found an inner class with the right name :-)
			if( CTopLevel.classExists( maybeQualifiedName ) ) {
			    // found an ambiguity :-(
			    throw 
				new UnpositionedError( MjcMessages.TYPE_AMBIG2,
						       maybeQualifiedName, prefix );
			} else {
			    // no ambiguity <:-)
			    setClass( maybeResult );
			    break name_res;
			}
		    }
		}

		// did not find an inner class so our last hope is to
		// look in a package
		setClassFromNameOrDieTrying();
	    }
	} // end of name_res block
	checkInstantiation(context);
	registerAsVisibleIn(context);
	return this;
    }

    /**
     * Checks the instantiation of a generic type by performing the following
     * checks on this type's <code>CClass</code> and all its owners:
     * <ul>
     *   <li>
     *     Type check the type variables of the class.
     *   </li>
     *   <li>
     *    Check that the numer of actual type arguments matches that of
     *    the number of formal type variables.
     *   </li>
     *   <li>
     *     Check that the actual type arguments are valid for the formal
     *     type variables.
     *   </li>
     * </ul>
     *
     * @param context  The type checking context.
     *
     * @throws UnpositionedError if some error occurs during type checking.
     */
    protected void checkInstantiation(CContextType context) throws UnpositionedError {
        // Make sure we always have entries which are not null.
        for (int i = 0; i < arguments.length; i++) {
            if (arguments[i] == null) {
                arguments[i] = CClassType.EMPTY;
            }
        }

        checkPossibleRawType(context);

        // We go from the innermost to the outermost class of a possibly nested
        // class and perform the appropriate checks.
        CClass clazz = getCClass();
        for (int i = arguments.length - 1; (i >= 0) && (clazz != null); i--) {
            // Make sure the type variables have been type checked before
            // extracting them.
            clazz.checkTypeVariables(context);
            CTypeVariable[] typeVariables = clazz.getTypeVariables();

            // Make sure that the number of type arguments matches that of
            // the number of type variables.
            if (arguments[i].length < typeVariables.length) {
                throw new UnpositionedError(MjcMessages.LESS_PARAMETER,
                                            clazz.qualifiedName(),
                                            String.valueOf(typeVariables.length));
            }

            if (arguments[i].length > typeVariables.length) {
                throw new UnpositionedError(MjcMessages.TOO_MANY_PARAMETER,
                                            clazz.qualifiedName(),
                                            String.valueOf(typeVariables.length));
            }

            // The raw types's type arguments are set to the erasures of the
            // corresponding type variables (see checkPossibleRawType()).
            // Even though this conforms to the semantics of raw types, those
            // generated type arguments are not necessarily within the bounds
            // of the type variable. Therefore, we avoid checking the type
            // arguments on raw types. [hook :-)]
            if (!isRawType()) {
                checkTypeArguments(typeVariables, arguments[i], context);
            }

            // We have checked the instantiation for this class, so let's move
            // to its possible owner class.
            clazz = clazz.owner();
        }
    }

    /**
     * Performs the necessary type checking for a raw type.
     * In particular, the following tasks are performed:
     * <ul>
     *   <li>
     *     It ensures that the method <code>isRawType()</code> correctly
     *     returns whether this is a raw type or not.
     *   </li>
     *   <li>
     *     If this is a raw type, it is given type arguments which correspond
     *     to the erasures of its class' type variables. This reflects the
     *     semantics behind raw types and at the same time ensures that all
     *     generic types have type arguments.
     *   </li>
     * </ul>
     *
     * @param context  The type checking context.
     *
     * @throws UnpositionedError if some error occurs during type checking.
     */
    private void checkPossibleRawType(CContextType context) throws UnpositionedError {
        // If we have already determined that this is a raw type, we are done.
        if (isRawType) {
            return;
        }

        isRawType = isGenericType()
                    && ((arguments.length == 0)
                        || (arguments[arguments.length - 1].length == 0));

        // If it is a raw type, we set its type arguments to the erasures of the
        // class' type variables. That's basically the semantics of a raw type.
        if (isRawType) {
            Vector rawArgs = new Vector();
            for (CClass clazz = getCClass(); clazz != null; clazz = clazz.owner()) {
                clazz.checkTypeVariables(context);
                CTypeVariable[] typeVariables = clazz.getTypeVariables();
                CClassType[] args = new CClassType[typeVariables.length];
                for (int i = 0; i < args.length; i++) {
                    args[i] = (CClassType) typeVariables[i].getErasure();
                }
                rawArgs.add(0, args);
            }
            arguments = (CClassType[][]) rawArgs.toArray(new CClassType[rawArgs.size()][]);
        }
    }

    /**
     * Performs the type checking on all the <code>typeArguments</code> (the
     * type checked versions are re-stored into the same array) and checks
     * whether the <code>typeArguments</code> are valid for the given
     * <code>typeVariables</code>.
     *
     * @param typeVariables  The formal type variables.
     * @param typeArguments  The actual type arguments for the type variables.
     * @param context        The type checking context.
     *
     * @throws UnpositionedError if some error occurs during type checking.
     */
    private void checkTypeArguments(
            CTypeVariable[] typeVariables,
            CClassType[] typeArguments,
            CContextType context) throws UnpositionedError {
        for (int i = 0; i < typeArguments.length; i++) {
            // Wildcards need to know the formal type variable for which they
            // are a type argument for later type checking. Therefore, we must
            // set it on the wildcard before checking its type as specified
            // by the wildcard's interface.
            if (typeArguments[i].isWildcard()) {
                ((CWildcardType) typeArguments[i]).setTypeVariable(typeVariables[i]);
            }

            // We have to type check the type argument before checking whether
            // it is a valid type argument for the matching type variable.
            typeArguments[i] = (CClassType) typeArguments[i].checkType(context);

            if (!typeArguments[i].isValidTypeArgumentFor(typeVariables[i],
            		typeArguments,
            		this)) {
                throw new UnpositionedError(MjcMessages.TYPE_NOT_MATCH,
                		typeArguments[i].toVerboseString(),
                		typeVariables[i].getIdent());
            } 
        }
    }

    /**
     * Registers the type represented by this as being visible in the
     * given context.  Does nothing if context == null.  
     *
     * <pre><jml>
     * requires maybeQualifiedName == null;
     * </jml></pre>
     */
    protected final void registerAsVisibleIn(CContextType context) 
	throws UnpositionedError
    {
	if( context != null ) {
	    // make sure the class is accessible from current context
	    //@ assert getClass() != null;
	    if (!getCClass().isAccessibleFrom(context.findNearestHost())) {
		throw new UnpositionedError(MjcMessages.TYPE_UNKNOWN,
					    this.toString());
	    }
	    
	    context.registerVisibleType(this);
	}
    }

    /**
     * Attempts to associate this type with a <code>CClass</code>
     * singleton having the name given by <code>maybeQualifiedName</code>.
     * Called after all other options for finding the class have been
     * exhausted.
     *
     * @exception UnpositionedError if singleton cannot be found 
     * @see #maybeQualifiedName
     * @see CTopLevel#loadClass(String)
     */
    protected void setClassFromNameOrDieTrying()
	throws UnpositionedError
    {
	if( !CTopLevel.classExists( maybeQualifiedName ) ) {
	    // We do need to throw an error here. It is used to signal
	    // errors on prefixes of qualified names, until we get
	    // enough of the qualified name to find a class.
	    throw new UnpositionedError( MjcMessages.TYPE_UNKNOWN, 
					 this.toString() );
	}
	String qName = maybeQualifiedName;
	CClass cl = CTopLevel.loadClass( maybeQualifiedName );

	// It is possible that 'classExists' returns true above, but the class
	// does not actually exist.  This can happen legitimately on a 
	// case-insensitive filesystem.
	if (cl == CClass.CLS_UNDEFINED) {
            throw new UnpositionedError( MjcMessages.TYPE_UNKNOWN,
                                         qName.replace('/', '.') );
	}

	setClass( cl );
    }

    /**
     * Sets the CClass singleton that goes with this class type and
     * registers that this name type has been checked.
     *
     * @param clazz the CClass singleton that goes with this class type */
    protected void setClass(CClass clazz) {
	super.setClass( clazz );
	maybeQualifiedName = null;
    }
    
    void forgetChecking() {
	maybeQualifiedName = originalQualifiedName;
    }

    public CType getErasure() {
        return !isParameterizedType() ? this : super.getErasure();
    }

    public /*@ non_null @*/ CType getCapture(CContextType context) throws UnpositionedError {
        if (!hasWildcardTypeArgument()) {
            return this;
        }

        CClassType[][] args = getAllArguments();
        CClassType[][] captureArgs = new CClassType[args.length][];
        for (int i = 0; i < args.length; i++) {
            captureArgs[i] = new CClassType[args[i].length];
            for (int j = 0; j < args[i].length; j++) {
                captureArgs[i][j] = args[i][j].capture();
            }
        }

        CClassType capture = CTopLevel.getTypeRep(
                getCClass().qualifiedName(),
                getCUniverse(),
                captureArgs,
                true);

        return (CClassType) capture.checkType(context);
    }

    /**
     * check to see whether this class type is Generic or Not
     */
    public boolean isGenericType() {
	 return getCClass().isGenericClass();
     }

    public boolean isRawType() {
        return isRawType;
    }
    
    public boolean isReifiableType() {
        return !isParameterizedType()
               || isRawType()
               || hasUnboundedWildcardTypeArgumentsOnly();
    }

    public boolean changesByErasure() {
        return isParameterizedType() && !isRawType();
    }

    public CClassType[] getArguments() {
    	// WMD added this sanity check as otherwise I got ArrayIndexOutOfBound
    	// errors in some cases
    	if(arguments == null || arguments.length==0)
    		return CClassType.EMPTY;
	return arguments[arguments.length-1];
    }
    
    public CClassType[][] getAllArguments() {
    	return arguments;
    }
    
    public void setAllArguments(CClassType[][] args) {
    	this.arguments = args;
    }
    
    public boolean isTypeVariable() {
	return isTypeVariable;
    }

    public boolean isMethodTypeVariable() {
	return isMethodTypeVariable;
    }
    
    public String getIdent() {
	return ident;
    }
    
    public void setIdent(String id) {
	ident = id ;
    }
            


    // ----------------------------------------------------------------------
    // PRIVILEGED DATA MEMBERS
    // ----------------------------------------------------------------------

    private String ident = "";
    protected String maybeQualifiedName; // null ==> checked
    //@ in thisClassInfo;
    /*@ private invariant maybeQualifiedName != null 
      @    ==> maybeQualifiedName.equals(originalQualifiedName);
      @*/
    private final String originalQualifiedName;
    //this is additional field to contain the typevariable when
    //the ClassNameType is actually is a type variable;

    private boolean isRawType = false;
}
