/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CClassType.java,v 1.67 2008/05/14 18:16:20 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.multijava.util.classfile.ClassRefInstruction;
import org.multijava.util.classfile.UniverseByteConstants;
import org.multijava.util.compiler.FastStringBuffer;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;

//@ model import org.jmlspecs.lang.JMLDataGroup;

/**
 * This class represents a class type in the type structure.  
 * @see	CType
 */
public abstract class CClassType extends CType implements Cloneable {

    /*@ public static model non_null JMLDataGroup loadedClassInfo;
      @ public model non_null JMLDataGroup thisClassInfo;
      @                               in objectState;
      @*/

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a class type
     */
    protected CClassType() {
	super(TID_CLASS);

	this.clazz = BAC_CLASS;
	this.arguments = new CClassType[][]{EMPTY};

	this.universe = CUniverseServices.getDefaultVariable();
    }

    /** Construct a class type with the default universe.
     *  By WMD (universe part).
     */
    public CClassType(CClass clazz) {
	this(clazz, CUniverseServices.getDefaultVariable());
    }

    /**
     * Construct a class type with a universe.
     * By WMD (universe part).
     *
     * @param	clazz		the class that will represent this type
     */
    public CClassType(CClass clazz, CUniverse univ) {
	super(TID_CLASS);

	this.clazz = clazz;
	this.arguments = new CClassType[][]{EMPTY};

	// WMD
	if( univ == null )
	    // this makes sure that we never get null de-references.
	    this.universe = CUniverseServices.getDefaultVariable();
	else
	    this.universe = univ;

	if (!((this instanceof CArrayType)
              || (this instanceof CErasedClassType))) {
	    //Debug.breakMe();
	    CTopLevel.addTypeRep(clazz.qualifiedName(),this);
	}
    }

    /**
     * Returns true if this class type has been resolved to a CClass 
     * singleton.
     *
     * @return	true if this class type has been resolved
     */
    //@ spec_public
    protected /*@ pure @*/ boolean isChecked() {
	return clazz != BAC_CLASS;
    }
    
    public int getIndex() {
	fail();
	return -1;
    }

    /**
     *set the type variable to the given index
     */
    public void setIndex(int i) {
	//index = i;
	fail();
    }

    /**
     * Sets the CClass singleton that goes with this class type.
     *
     * @param	clazz	the CClass singleton that goes with this class type
     */
    //@ assignable thisClassInfo;
    protected void setClass(CClass clazz) {
	this.clazz = clazz;
	cachedString = null;
    }


    /** Set the universe of this class type.
     * By WMD.
     *
     * @param univ The new universe modifier.
     *        If the parameter is null, then the peer modifier is used.
     */
    public void setCUniverse( CUniverse univ ) {
	if( univ == null ) {
	    // WMD TODO log a warning
	    univ = CUniverseServices.getDefaultVariable();
	}
	
	universe = univ;
    }


    /** We need to allow cloning of CClassType and its subtypes,
     * in order to be able to create a new CClassType with the correct
     * universe modifier.
     * By WMD.
     */
    public Object clone() throws CloneNotSupportedException {
	return super.clone();
    }


    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Compares two objects for equality.
     *
     * @param other	the object to compare with
     * @return true iff <code>other</code> is a <code>CType</code> object 
     *		representing the same Java type as this represents
     *
     * WMD: Forward the call to the universe enabled version below.
     */
    // !FIXME! the following JML indicates a problem that we are
    // unable to fix easily.  This method really does have side
    // effects and the assignable clauses below are accurate.
    // However, equals cannot have side effects.  It also seems
    // difficult to refactor this code, because of the use of
    // CClassType objects in collections that are based on equals and
    // the desire for lazy loading of classes, although that should be
    // done.  Hence the following is *not* a JML annotation currently.
    /*
      @ also
      @   assignable other.objectState if other != null;
      @   assignable loadedClassInfo, thisClassInfo;
      */
    public boolean equals(/*@ nullable @*/  Object other ) {
	return equals( other, CTopLevel.getCompiler().universeChecks(), false );
    }

    /**
     * Compares two objects for equality.
     *
     * @param other	the object to compare with
     * @param enableUniv whether to also check the universe part
     *          this is needed sometimes to allow relaxed checks
     * @param allowUniversesCovariance Allow limited covariance from GUT iff true. 
     * @return true iff <code>other</code> is a <code>CType</code> object 
     *		representing the same Java type as this represents
     */
    public boolean equals(/*@ nullable @*/ Object other, boolean enableUniv,
    		boolean allowUniversesCovariance) {
	if (!(other instanceof CClassType)) {
	    return false;
	} else {
	    CClassType ot = (CClassType)other;
	    
		// This is a hack for raw types so we don't end up being
		// stuck in an infinite loop
	    if (qualifiedName().equals(((CClassType) other).qualifiedName())
				&& getAllArguments() != null
				&& ot.getAllArguments() != null
				&& getAllArguments().length == 0
				&& ot.getAllArguments().length == 0) {
			return true;
		}
	    
	    if ( !(ot.isClassType())
		|| ot.isArrayType()
		//||ot.isTypeVariable()
		|| ot.getCClass() != getCClass()
		|| (isTypeVariable() != ot.isTypeVariable())
                || (isWildcard() != ot.isWildcard())
		||ot.isGenericType() != isGenericType()
		// WMD
		|| (enableUniv && !allowUniversesCovariance && !(ot.getCUniverse().equals(this.getCUniverse())))
		|| (enableUniv && allowUniversesCovariance && !(ot.getCUniverse().isAlwaysAssignableTo(this.getCUniverse())))
		) {
		return false;
	    }
	    if(ot.isGenericType() && isGenericType()) {
		CClassType[] otherArgs = ((CClassType)other).getArguments();
		if(otherArgs.length!= arguments[arguments.length-1].length) {
		    return false;
		}
		
		for (int i=0; i<arguments[arguments.length-1].length;i++) {
		    if(!otherArgs[i].equals(arguments[arguments.length-1][i], enableUniv, allowUniversesCovariance)) {
			//System.out.println("here1");
			return false;
		    }
		}
	    }
	    return true;
	}
    }
    
    public boolean equals(/*@ nullable @*/ CType other,
			  /*@ non_null @*/ CClassType[] substitution) {
	return equals(other, substitution, CTopLevel.getCompiler().universeChecks());
    }

    public boolean equals(/*@ nullable @*/ CType other,
			  /*@ non_null @*/ CClassType[] substitution,
			  boolean enableUniv ) {
	if (other.isClassTypeVariable()) {
	    other = substitution[((CTypeVariable)other).getIndex()];
	}
	
	if (!(other instanceof CClassType)) {
	    return false;
	}
 	else {
	    CClassType ot = (CClassType)other;
	    if (!(ot.isClassType())
		||ot.isArrayType()
		||ot.getCClass() != getCClass()
                || (isWildcard() != ot.isWildcard())
		//||ot.isGenericType() != isGenericType()
		// WMD
		|| (enableUniv && !(getCUniverse().equals(ot.getCUniverse())))
		) {
		return false;
	    }
	    if(ot.isGenericType() && isGenericType()) {
		CClassType[] otherArgs = ((CClassType)other).getArguments();
		if(otherArgs.length!= arguments[arguments.length-1].length) {
		    return false;
		}
		
		for (int i=0; i<arguments[arguments.length-1].length;i++) {
		    if(!otherArgs[i].equals(arguments[arguments.length-1][i], enableUniv, false)) {
			return false;
		    }
		}
	    }
	    return true;
	}
    }
    

    /**
     * Calculates the hash code for this
     *
     * @return	the hash code for this, which is the same as the hash code for
     *		the corresponding CClass singleton
     *
     * <pre><jml>
     * also
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     */
    public int hashCode() {
	// WMD: maybe add the CUniverse hashcode too, but not required.
	return getCClass().hashCode();
    }

    /**
     * Transforms this type to a string
     *
     * <pre><jml>
     * also
     * assignable loadedClassInfo, thisClassInfo;
     * ensures \not_modified(loadedClassInfo);
     * </jml></pre>
     */
    public String toString() {
		if (cachedString == null) {
			createCachedString();
		}
		
    	return cachedString + printArgs(false);
    }

	/**
	 * Caches the string of the qualified name
	 */
    private void createCachedString() {
    	cachedString = qualifiedName().replace('/', '.');
    	if (clazz.isNestedType()) {
    		cachedString = cachedString.replace('$', '.');
    	}
    }
    
    /**
     * Print the type arguments of this class type
	 * 
	 * @param includeUniverseModifier Universe modifiers are printed iff true
     */
    //@ assignable \nothing;
    public String printArgs(boolean includeUniverseModifier) {
    	if (arguments == null || !getCClass().isGenericClass()) {
    		return "";
    	}

    	StringBuffer buffer = new StringBuffer();
    	buffer.append('<');
    	for(int i=0; i<arguments[arguments.length-1].length; i++) {
    		if (i>0) {
    			buffer.append(", ");
    		}
    		buffer.append(includeUniverseModifier
    				? arguments[arguments.length-1][i].toVerboseString()
    				: arguments[arguments.length-1][i].toString());
    	}
    	buffer.append('>');
    	return buffer.toString();
    }
    
    public String originalQualifiedName() {
	return toString();
    }


    /**
     * Return the universe modifier plus the class type as string.
     * 
     * by WMD
     * by RZU
     */
    public String toVerboseString() {
    	if (CTopLevel.getCompiler().universeChecks()) {
    		if (cachedString == null) {
    			createCachedString();
    		}
    		
			return getCUniverseString() 
					+ " "
					+ cachedString
					+ printArgs(true);
    	} else {
    		return toString();
    	}
    }


    /** Return the universe modifier as string.
     * Always returns a Universe string, even if the Universe checks
     * are turned off!
     * By WMD.
     */
    public String getCUniverseString() {
	return getCUniverse().toString();
    }
    
    /** Return the universe modifier as string, for output in JML
     * specifications. The only difference is that implicit annotations
     * are escaped with a backslash.
     * By AHS, WMD.
     */
    public String getCUniverseJMLString() {
	return getCUniverse().toJMLString();
    }

    /**
     * Returns the universe modifier as String, for output in source
     * code parsed by MultiJava, i.e. the modifiers are written as is.
     * By RZU, WMD.
     */
    public String getCUniverseMJString() {
	return getCUniverse().toMJString();
    }

    
    /**
     * Indicates whether this type is a nested type, i.e., a member
     * type.  */
    //@ pure
    public boolean isNestedType() {
	return clazz.isNestedType();
    }

    /**
     * Transforms this type to a string
     */
    public void appendSignature(FastStringBuffer buff) {
	buff.append('L');
	buff.append(qualifiedName());
	buff.append(';');
    }

    /**
     * @return the short name of this class
     *
     * <pre><jml>
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     */
    public String ident() {
	return getCClass().ident();
    }

    /**
     * @return the qualified name of this class.
     */
    // <pre><jml>
    // assignable loadedClassInfo, thisClassInfo;
    // </jml></pre>
  
    public String qualifiedName() {
	return getCClass().qualifiedName();
    }

    /**
     * Returns the size used in stack by value of this type
     */
    public /*@ pure @*/ int getSize() {
	return 1;
    }

    /**
     * Check if a type is a reference
     * @return	is it a type that accept null value
     */
    public /*@ pure @*/ boolean isReference() {
	return true;
    }

    /**
     * Check if a type is a class type
     * @return	is it a subtype of ClassType ?
     */
    public /*@ pure @*/ boolean isClassType() {
	return true;
    }

    /**
     * Check if a type is an interface
     * @return is it a class type representing an interface ?
     *
     * <pre><jml>
     * also
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     */
    public boolean isInterface() {
	CClass clazz = getCClass();
	return clazz.isInterface();
    }
    
    /**
     * Indicates whether this type is concrete.  All types are
     * concrete except abstract class types and interface types.
     *
     * @return	whether this type is concrete
     *
     * <pre><jml>
     * also
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     */
    public boolean isConcrete() {
	CClass clazz = getCClass();
	return !clazz.isAbstract() && !clazz.isInterface();
    }

    /**
     * Indicates whether this type is declared local to the given
     * context.  Can only be true of reference types.
     *
     * @param	context	the context in which to check
     * @return whether this type is declared local to the given context.
     *
     * <pre><jml>
     * also
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     */
    public boolean isLocalTo( CContextType context ) {
	return getCClass().isDefinedInside( context.getCompilationUnit()
					    .findNearestHost() );
    }

     /* <pre><jml>
     * also 
     * requires isChecked();
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     *
     */
    /**
     * Returns the class object associated with this type.
     *
     * @return the class object associated with this type
     */
    public CClass getCClass() {
	if (!isChecked()) {
	    throw new RuntimeException( "Class type " + toString() +
					" not checked." );
	}
	
	//@ assert(isChecked());
	return getCachedCClass();
    }


    /** Return the universe of this class type.
     * By WMD.
     */
    public CUniverse getCUniverse() {
	// WMD do we need something like the isChecked above?

	return universe;
    }
    
    /** Return the universe byte constant for this class type.
     * By AHS.
     */
    public byte getUniverseByteConstant() {
        return UniverseByteConstants.buildConstant(universe);
    }

    public CClassType createSubstitutedType(CClass local,
    		CClassType prefixType,
		    CClassType[][] substitution) {
    	return createSubstitutedType(local,
    			prefixType,
    			substitution,
    			false);
    }
    
    public CClassType createSubstitutedType(CClass local,
    		CClassType prefixType,
    		CClassType[][] substitution,
    		boolean notOwnedTypeVariablesOnly) {
	CClassType[][]          temparguments;
	CClassType[][]          prefixArgs = prefixType == null ? CClassType.EMPTY_ARG : prefixType.getAllArguments();
	
	if (!getCClass().isNestedType() || prefixType == null) {
	    temparguments = new CClassType[][]{getArguments()};
	} else {
	    temparguments = new CClassType[prefixArgs.length+1][];
	    System.arraycopy(prefixArgs, 0, temparguments, 0, prefixArgs.length);
	    temparguments[prefixArgs.length] = getArguments();
	}
	
	if (prefixType != null) {
	    substitution = prefixType.getAllArguments();
	    local = prefixType.getCClass();
	}
	
	if (temparguments.length > 0) {
	    int                       index = temparguments.length-1;
	    CClassType[]          subArgs = new CClassType[temparguments[index].length];
	    
	    for (int i=0; i < temparguments[index].length; i++) {
		CClassType          type = temparguments[index][i];
		
                // Only class type variables can be substituted.
		if (type.isClassTypeVariable()) {
		    subArgs[i] = local.getSubstitution((CTypeVariable) type,
		    		substitution,
		    		notOwnedTypeVariablesOnly);
		} else if (type.isGenericType()) {
		    subArgs[i] = ((CClassType) type).createSubstitutedType(local,
		    		null,
		    		substitution,
		    		notOwnedTypeVariablesOnly);
		} else {
		    subArgs[i] = type;
		}
	    }
	       
	    temparguments[index] = subArgs;
	}
	CClassType newType = null;
	try{
	    newType = (CClassType)clone();
				
	    // CClassNameType       type = new CClassNameType(getCClass(), arguments);
	    newType.setAllArguments(temparguments);
	} catch(CloneNotSupportedException e) {
	  fail("Clone exception: " + e);
	}
	
	return newType;
    }


    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks that this type is valid, throwing an exception if it is
     * not.  For class types this check resolves type names into their
     * fully qualified versions and sets the class parameter so that
     * getCClass returns the appropriate singleton.
     *
     * @param		context		the context in which this type was
     *					declared, may be null
     * @exception	UnpositionedError	this unpositioned error
     *						should be positioned
     *						by the calling node of
     *						the AST, typically by
     *						calling the
     *						addPosition method of
     *						UnpositionedError 
     * <pre><jml>
     * //also
     * //assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     */
    public CType checkType(CContextType context)
	throws UnpositionedError {
	context.registerVisibleType( this );
	return this;
    }

    /**
     * Compute the value of a specializer expression used to dispatch on
     * a compile-time constant.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    public void resolveValueType(CExpressionContextType context)
	throws PositionedError {}
    
    
    // ----------------------------------------------------------------------
    // BODY CHECKING
    // ----------------------------------------------------------------------

    /**
     * Returns whether this type descends from the given <code>maybeSuper</code>
     * type.
     * 
     * Universes are taken into account iff universe checks are on.
     *
     * @param maybeSuper  The potential supertype of this type.
     * @return            Whether this type is a subtype of the given
     *                    <code>maybeSuper</code>.
     */
    private boolean descendsFrom(/*@ non_null @*/ CClassType maybeSuper) {
    	return descendsFrom(maybeSuper,
    			CTopLevel.getCompiler().universeChecks());
    }
    
    /**
     * Checks if this type descends from <code>maybeSuper</code>.
     * 
     * @param maybeSuper The potential supertype to check.
     * @param considerUniverses Universe are taken into account iff true.
     * @return True iff <code>this</code> (indirectly) inherits from
     *         <code>maybeSuper</code>.
     * 
     * by WMD
     * by RZU
     */
    private boolean descendsFrom(/*@ non_null @*/ CClassType maybeSuper,
    		boolean considerUniverses) {
        // Check universe modifier if desired
        if (considerUniverses && !getCUniverse()
        		.isAlwaysAssignableTo(maybeSuper.getCUniverse())) {
            return false;
        }
    	
        // WMD: whoever made this case distinction should think of the case,
        // where we have a subclass "C extends D<Data>" and we need to check whether
        // D<Object> d = new C()" is valid. We can not just ignore the arguments
        // of D, just because C is not generic.
        // However, without it, some other testcases fail :-(
        // Only the else-part should remain eventually.
        if (!isGenericType() ) {
        	return getCClass().descendsFrom(maybeSuper.getCClass());
        } else {        	
        	return getCClass().descendsFrom(maybeSuper,
           		this,
           		considerUniverses,
           		maybeSuper.getCUniverse() instanceof CUniverseReadonly);
        }
    }

    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2] [JLS2 5.1.4]
     * @param	dest		the destination type
     * @return true iff assignment is valid
     *
     * <pre><jml>
     * //also
     *   //assignable_redundantly thisClassInfo;
     * </jml></pre>
     *
     * WMD: In addition the universes have to be assignable, too.
     */
    public boolean isAlwaysAssignableTo(CType dest, boolean inst) {
	if (!(dest.isClassType() && !dest.isArrayType())) {
	    return false;
	}

        if (dest.isCapture()) {
            // A type is a subtype of a capture type variable iff it is a
            // subtype of the capture's lower bound (see end of JLS3 4.10.2).
            CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
            if (lowerBound == null) {
                return false;
            } else {
                return isAlwaysAssignableTo(lowerBound, inst);
            }
        }

        if (dest.isTypeVariable()) {
            if (!inst) {
                return false;
            } 
        }

        return descendsFrom((CClassType) dest);
    }

    public boolean isAlwaysAssignableTo(CType dest, 
					CClassType[] substitution) {
	if (!(dest.isClassType() && !dest.isArrayType())) {
	    return false;
	}

	// WMD assumes that "substitution" is actually only meant for the method
	// type variables... but the documentation of what gets passed in is
	// non-existent :-(
	// WMD TODO But when I do these changes, i.e. make a distinction between method TVs
	// and class TVs, some of the gj testcases fail :-(
	if (dest.isClassTypeVariable()) {
//    	dest = this.createSubstitutedType(this.getCClass(), (CClassType) dest, getAllArguments());
//	}

//	if (dest.isMethodTypeVariable()) {
		if( substitution==null ||
			substitution.length <= 0 ) {
			fail("Sorry, something wrong with type variable substitution!");
		}
		dest = substitution[((CTypeVariable)dest).getIndex()];		
	}

        if (dest.isCapture()) {
            // A type is a subtype of a capture type variable iff it is a
            // subtype of the capture's lower bound (see end of JLS3 4.10.2).
            CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
            if (lowerBound == null) {
                return false;
            } else {
                dest = lowerBound;
            }
        }

        return descendsFrom((CClassType) dest);
    }

    /** A version of the Assignable check without considering the Universe types.
     * This is needed to determine e.g. the correct method to invoke, where we
     * do not look at the Universe types.
     * There were already many versions of "isAlwaysAssignableTo" and also with
     * boolean flags; therefore I decided to change the name.
     * By WMD.
     */
    public boolean isAlwaysAssignableToIgnoreUniverses( CType dest, 
						    CClassType[] substitution ) {
	if (!(dest.isClassType() && !dest.isArrayType())) {
	    return false;
	}

	// WMD assumes that "substitution" is actually only meant for the method
	// type variables... but the documentation of what gets passed in is
	// non-existent :-(
	// WMD TODO But when I do these changes, i.e. make a distinction between method TVs
	// and class TVs, some of the gj testcases fail :-(
	if (dest.isClassTypeVariable()) {
//    	dest = this.createSubstitutedType(this.getCClass(), (CClassType) dest, getAllArguments());
//	}

//	if (dest.isMethodTypeVariable()) {
		
		// System.out.println("dest1: " + dest.toVerboseString());
		if( substitution==null ||
			substitution.length <= 0 ) {
			fail("Sorry, something wrong with type variable substitution!");
		}
		dest = substitution[((CTypeVariable)dest).getIndex()];
		// System.out.println("dest2: " + dest.toVerboseString());
	}
	
	
        if (dest.isCapture()) {
            // A type is a subtype of a capture type variable iff it is a
            // subtype of the capture's lower bound (see end of JLS3 4.10.2).
            CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
            if (lowerBound == null) {
                return false;
            } else {
                dest = lowerBound;
            }
        }

        if (!isGenericType() ) {
	    return getCClass().descendsFrom(dest.getCClass());
        } else {
	    return getCClass().descendsFrom((CClassType) dest, this, false, false);
        }
        // WMD TODO I think the above should be changed to the below:
        // return descendsFrom((CClassType) dest, false);
    }

    public boolean isAlwaysAssignableTo(CType dest) {
	return isAlwaysAssignableTo(dest,false);
    }
    
    public boolean isAlwaysAssignableToIgnoreUniverses(CType dest) {
    	if (!(dest.isClassType() && !dest.isArrayType())) {
    		return false;
    	}

    	if (dest.isCapture()) {
    		// A type is a subtype of a capture type variable iff it is a
    		// subtype of the capture's lower bound (see end of JLS3 4.10.2).
    		CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
    		if (lowerBound == null) {
    			return false;
    		} else {
    			return isAlwaysAssignableTo(lowerBound);
    		}
    	}

    	// WMD TODO: why is this here, only in this version, without inst?
    	if (dest.isTypeVariable()) {
    	 	return false;
    	}

    	return descendsFrom((CClassType) dest, false);
    }

    /**
     * Returns whether this type is a valid type argument for the given
     * <code>typeVariable</code>.
     *
     * @param typeVariable  The type variable for which this type is the actual
     *                      type parameter.
     * @param substitution  All the actual type arguments of the parametrized type.
     * @param type			The type whose type variable it is
     * @return              Whether this type is a valid type argument for the
     *                      given <code>typeVariable</code>.
     */
    public boolean isValidTypeArgumentFor(
            CTypeVariable typeVariable,
            CClassType[] substitution,
            CClassNameType type) {
        CClassType[] destBounds = typeVariable.getBounds();
        // Handle the special case of having the implicit java.lang.Object bound
        // for the type variable.
        if (destBounds.length == 0) {
            return true; // bound == java.lang.Object
        }

        for (int i = 0 ; i < destBounds.length; i++) {
        	CClassType currentBound = destBounds[i];

        	// Adapt viewpoint of upper bound for Generic Universe Types
        	if (CTopLevel.getCompiler().universeChecks()) {
        		currentBound = CUniverseServices
        				.combine(type, currentBound);
        	}
        	
			if (!isAlwaysAssignableTo(currentBound, substitution)
                || needsUncheckedConversion(currentBound, substitution)) {
                return false;
            }
        }
        return true;
    }

    // WMD this method was never adapted to generics... why?
    public boolean isAssignableTo(CType dest) {
	if (!(dest.isClassType() && !dest.isArrayType())) {
	    return false;	
	} else {
	    return getCClass().descendsFrom(dest.getCClass()) &&
		(!CTopLevel.getCompiler().universeChecks() ||
		 getCUniverse().isAlwaysAssignableTo(((CClassType)dest).getCUniverse())); // WMD FIXME is this ok???
	}
    }

    public boolean needsUncheckedConversion(CType dest) {
        if (!dest.isClassType()) {
            return false;
        }

        // Assigning to a capture type variable is equivalent to assigning to
        // its lower bound.
        if (dest.isCapture()) {
            dest = ((CCaptureType) dest).getLowerBound();
            if (dest == null) {
                return false;
            }
        }

        CClassType classType = (CClassType) dest;

        // If we are converting to a non-parameterized type or a raw type,
        // no unchecked conversion is needed.
        if (!classType.isParameterizedType() || classType.isRawType()) {
            return false;
        }

        // If we are converting to a parameterized type whose type arguments are
        // all unbounded wildcards, no unchecked conversion is needed (JLS3 5.1.9).
        if (classType.hasUnboundedWildcardTypeArgumentsOnly()) {
            return false;
        }

        return isRawTypeInvocationOf(classType.getCClass());
    }

    public boolean needsUncheckedConversion(
            /*@ non_null @*/ CType dest,
            /*@ non_null @*/ CClassType[] substitution) {
        if (dest.isClassTypeVariable()) {
            dest = substitution[((CTypeVariable) dest).getIndex()];
        }

        return needsUncheckedConversion(dest);
    }

    /**
     * Returns whether this type is a raw type invocation of the given
     * <code>superClazz</code>.
     *
     * @param superClazz  The target super class.
     * @return            Whether this type is a raw type invocation of the
     *                    given <code>superClazz</code>.
     */
    public boolean isRawTypeInvocationOf(CClass superClazz) {
        return isRawTypeInvocationOfRec(superClazz, this, isRawType());
    }

    /**
     * Helper method which goes up the supertype hierarchy in order to find out
     * whether the given <code>type</code> is a raw type invocation of the
     * given <code>superClazz</code>.
     *
     * @param superClazz  The target super class.
     * @param type        The current type in the supertype hierarchy.
     * @param isRawType   Whether <code>type</code> is a raw type.
     * @return            Whether <code>type</code> is a raw type invocation of
     *                    the given <code>superClazz</code>.
     */
    private boolean isRawTypeInvocationOfRec(
            CClass superClazz,
            CClassType type,
            boolean isRawType) {
        CClass clazz = type.getCClass();

        if (superClazz == clazz) {
            return isRawType;
        }

        if (isRawTypeInvocationOfSuper(superClazz, clazz.getSuperType(), isRawType)) {
            return true;
        }

        CClassType[] interfaces = clazz.getInterfaces();
        for (int i = 0; i < interfaces.length; i++) {
            if (isRawTypeInvocationOfSuper(superClazz, interfaces[i], isRawType)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Helper method which takes the step from a type to its supertype in order
     * to find out whether the latter is a raw type invocation of the given
     * <code>superClazz</code>.
     *
     * @param superClazz  The target super class.
     * @param superType   The supertype to which to move.
     * @param isRawType   Whether the current type is a raw type.
     * @return            Whether the given <code>superType</code> is a raw
     *                    type invocation of <code>superClazz</code>.
     */
    private boolean isRawTypeInvocationOfSuper(
            CClass superClazz,
            CClassType superType,
            boolean isRawType) {
        if (superType != null) {
            if (!superType.isParameterizedType()) {
                isRawType = false;
            } else {
                isRawType |= superType.isRawType();
            }
            return isRawTypeInvocationOfRec(superClazz, superType, isRawType);
        }
        return false;
    }

    /**
     * Returns whether this type contains the given type as defined in
     * JLS3 4.5.1.1.
     * This test is used when checking for the subtype relationship between
     * two parametrized types. In order for a parametrized type to be
     * a subtype of another, all type arguments of the former must be
     * contained by the corresponding type arguments of the latter.
     *
     * @param type The type possibly contained by this type.
     * @param considerUniverses Take universe modifier into account iff true.
	 * @param allowUniversesCovariance Allow limited covariance from GUT iff true.
     * @return Whether this type contains the given <code>type</code>.
     */
    public boolean contains(/*@ non_null @*/ CClassType type,
    		boolean considerUniverses,
    		boolean allowUniversesCovariance) {
    	// Check universe modifiers if desired
    	if (considerUniverses) {
    		if (allowUniversesCovariance) {
    			if (!type.getCUniverse()
    					.isAlwaysAssignableTo(getCUniverse())) {
					return false;
    			}
    		} else {
    			if (!type.getCUniverse().equals(getCUniverse())) {
					return false;
    			}    			
    		}
    	}

        // Any class type except wildcards merely contains itself.
    	// RZU: Universe modifiers are not taken into account since we
    	// have already dealt with them
    	// WMD: Wrong, we still have to check the arguments recursively.
        return equals(type, considerUniverses, allowUniversesCovariance);
    }

    public CType getErasure() {
	return CTopLevel.getTypeRep(getCClass().qualifiedName(), universe, true);
    }

    /**
     * Returns the type captured by a type argument.
     *
     * Only wildcards have a capture which differs from the type itself as
     * described in JLS3 5.1.10.
     *
     * @return  The capture of this type.
     *          Is never <code>null</code>.
     */
    public /*@ non_null @*/ CClassType capture() {
        return this;
    }

    /**
     * Returns whether this is a parameterized type (JLS3 4.5).
     *
     * @return  Whether this is a parameterized type.
     */
    public /*@ pure @*/ boolean isParameterizedType() {
        CClassType[][] args = getAllArguments();
        for (int i = 0; i < args.length; i++) {
            if ((args[i] != null) && (args[i].length > 0)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns whether any of the type arguments of this type is a wildcard.
     *
     * @return  Whether any of the type arguments of this type is a wildcard.
     */
    public /*@ pure @*/ boolean hasWildcardTypeArgument() {
        CClassType[][] args = getAllArguments();
        for (int i = 0; i < args.length; i++) {
            if (args[i] != null) {
                for (int j = 0; j < args[i].length; j++) {
                    if (args[i][j].isWildcard()) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    /**
     * Returns <code>true</code> if and only if this is a parameterized type
     * <i>and</i> all its type arguments are unbounded wildcards.
     *
     * @return  Whether this is a parameterized type <i>and</i> all its type
     *          arguments are unbounded wildcards.
     */
    public /*@ pure @*/ boolean hasUnboundedWildcardTypeArgumentsOnly() {
        if (!isParameterizedType()) {
            return false;
        }

        CClassType[][] args = getAllArguments();
        for (int i = 0; i < args.length; i++) {
            if (args[i] != null) {
                for (int j = 0; j < args[i].length; j++) {
                    // If it is no wildcard at all, we are done.
                    if (!args[i][j].isWildcard()) {
                        return false;
                    }
                    // If it is a bounded wildcard, we are done.
                    CWildcardType wildcard = (CWildcardType) args[i][j];
                    if (wildcard.getBound() != null) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    /**
     * Can this type be converted to the specified type by casting
     * conversion (JLS 5.5) ?
     *
     * @param	dest		the destination type
     * @return true iff the casting conversion is valid
     *
     * <pre><jml>
     * also
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     *
     * WMD: added additional conditions for the universe modifiers
     */
    public boolean isCastableTo(CType dest) {
    	// test for array first because array types are classes

    	if (getCClass().isInterface()) {
    		if (! dest.isClassType()) {
    			return false;
    		} else if (dest.getCClass().isInterface()) {
    			// if T is an interface type and if T and S contain methods
    			// with the same signature (8.4.2) but different return types,
    			// then a compile-time error occurs.
    			//!!! graf 000512: FIXME: implemt this test
    			return true;
    		} else if (! dest.getCClass().isFinal()) {
    			return true;
    		} else {
    			return dest.isClassType()
    			? ((CClassType) dest).descendsFrom(this)
    					: dest.getCClass().descendsFrom(getCClass());
    		}
    	} else {
    		// this is a class type
    		if (dest.isArrayType()) {
    			// WMD compare the universe separately
    			return this.getCClass().equals(CStdType.Object.getCClass());
    		} else if (! dest.isClassType()) {
    			return false;
    		} else if (dest.getCClass().isInterface()) {
    			if (! getCClass().isFinal()) {
    				return true;
    			} else {
    				return getCClass().descendsFrom(dest.getCClass());
    			}
    		} else if (dest.isTypeVariable()) {
    			CTypeVariable destTV = (CTypeVariable) dest;
    			
    			// Check that the given type can be cast to every bound
    			for (int i = 0; i < destTV.getBounds().length; i++) {
    				if (!isCastableTo(destTV.getBounds()[i])) {
    					return false;
    				}
    			}
    				
    			return true;
    		} else { 
    			boolean classpart = dest.isClassType() 
    					? ((CClassType) dest).descendsFrom(this, false)
    							|| descendsFrom((CClassType) dest, false)
    					: dest.getCClass().descendsFrom(getCClass())
    							|| getCClass().descendsFrom(getCClass());
                        return classpart & (dest.isClassType() ?
                            (this.universe.isAlwaysAssignableTo(((CClassType) dest).universe) ||
                            ((CClassType)dest).universe.isAlwaysAssignableTo(this.universe) )
                            : true);
    		}
    	}
    }

    /**
     * implicitlyCastTo
     * @param	dest	the expression type to cast to
     * @return	cast is valid ?
     */
    public boolean implicitlyCastTo(CType dest) {
	return isAlwaysAssignableTo(dest);
    }

    /**
     * Returns true if this represents the type of a checked exception.
     */
    public boolean isCheckedException() {
	return !isAlwaysAssignableTo(CStdType.RuntimeException) &&
	    !isAlwaysAssignableTo(CStdType.Error);
    }

    /**
     * Returns the set of types (excluding non-reference types) that
     * are directly visible from this type.
     *
     * <pre><jml>
     * assignable loadedClassInfo, thisClassInfo;
     * </jml></pre>
     */
    public Collection directlyVisibleTypes() {
	return getCClass().directlyVisibleTypes();
    }

    /**
     * Returns the set of visible types by finding the reflexive,
     * transitive closure of directlyVisibleTypes() on each type in
     * the given collection.  Requires the every member of the given
     * collection is an instance of CType.  
     * @param dirVisTypes an initial collection of directly visible types
     *
     * <pre><jml>
     * requires (\forall Object o; dirVisTypes.contains(o);
     *				   o instanceof CType);
     * </jml></pre>
     *
     */
    public static Collection getVisibleTypesFrom( Collection dirVisTypes ) {
	// ----------------------------------------------------------------
	// Algorithm is a breadth-first search of the graph of
	// directly visible types, with cycle detection.
	// ----------------------------------------------------------------
	
	// An empirical estimate of the ratio of visible types to
	// directly visible types for a typical compilation unit
	final int RATIO = 20;	// !FIXME! SWAG, adjust after checking
				// of visible types in CBinaryClass is
				// completed (see estimating output below)

	Set result = new HashSet( RATIO * dirVisTypes.size() );
	result.addAll( dirVisTypes );
	
	// Used for cycle detection
	Set searched = new HashSet( RATIO * dirVisTypes.size() );

	Set toBeSearched = new HashSet( dirVisTypes );
	Set nextDepth = new HashSet();
	while( toBeSearched.size() > 0 ) {
	    // find the next depth in the graph of directly visible types
	    Iterator iter = toBeSearched.iterator();
	    while( iter.hasNext() ) {
		CType type = ((CType) iter.next());
		searched.add( type );
		if (type.isClassType()) {
		    Collection dirVisFromType =
			((CClassType) type).directlyVisibleTypes();
		    nextDepth.addAll( dirVisFromType );
		    result.addAll( dirVisFromType );
		}
	    }
	    
	    // Find the types in nextDepth that haven't been searched
	    // yet.  Swap the sets to reduce the number of
	    // instantiations.
	    Set swap = toBeSearched;
	    toBeSearched = nextDepth;
	    toBeSearched.removeAll( searched );
	    nextDepth = swap;
	    nextDepth.clear();
	}

	// Output data to estimate RATIO
//  	System.out.println( "RATIO: " + 
//  			    ( result.size() / (double) dirVisTypes.size() ) );

	return result;
    }


    /** Return the supertype of this type, with all type variables of the subtype
     * substituted in the supertype.
     * 
     * @return The substituted supertype.
     * 
     * by WMD.
     */
    public CClassType getSuperClassType() {
    	CClassType sc = this.getCClass().getSuperType();
    	if( sc==null ) return null;
    	
    	CClassType ssc;
    	ssc = sc.createSubstitutedType(this.getCClass(), this, getAllArguments(), true);
    	ssc = CUniverseServices.combine(this.getCUniverse(), ssc);
    	ssc = ssc.createSubstitutedType(this.getCClass(), this, getAllArguments(), false);
    	
    	return ssc;
    }
    
    /** Return all the interfaces of this type, with all type variables of the subtype
     * substituted in the supertype.
     * 
     * @return The substituted interface types.
     * 
     * by WMD.
     */
    public CClassType[] getSuperInterfaceTypes() {
    	CClassType[] in = this.getCClass().getInterfaces();
    	
    	if( in!=null ) {
    		CClassType[] res = new CClassType[in.length];
    		for(int i=0; i<in.length; ++i) {
    			res[i] = in[i].createSubstitutedType(this.getCClass(), this, getAllArguments(), true);
    	    	res[i] = CUniverseServices.combine(this.getCUniverse(), res[i]);
    	    	res[i] = res[i].createSubstitutedType(this.getCClass(), this, getAllArguments(), false);
    		}
    		return res;
    	}
    	return null;
    }
    
    
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

	/* Performs the instanceof / equality test for specializers of this
	 * type.  We assume that the actual argument has already been loaded.
	 *
	 * @param	code		the bytecode sequence to append to
	 * @param	nextInstruction	the instruction to branch to if the
	 *				dispatch test fails
	 */
    public void plantDispatchTest( CodeSequence code,
				   CodeLabel nextInstruction ) {
	code.plantInstruction( new ClassRefInstruction( opc_instanceof,
							qualifiedName() ) );
	// Note that instanceof tests in the virtual machine leave a 1
	// on the stack if the instanceof tests yields true and a 0
	// otherwise.  Thus an ifeq opcode following an instanceof
	// test will branch if the instanceof test yielded false.
	code.plantLabelRef( opc_ifeq, nextInstruction );
    }
    
    // ----------------------------------------------------------------------
    // INITIALIZERS METHODS
    // ----------------------------------------------------------------------

    /**
     * Get source class from cached info of all loaded classes in CTopLevel.
     *
     * <pre><jml>
     * assignable loadedClassInfo, thisClassInfo, cachedString, clazz;
     * </jml></pre>
     */
    private CClass getCachedCClass() {
	CClass srcType = CTopLevel.reloadClass(clazz);
	if (srcType != clazz) {
	    // avoids unnecessary nullifying of cachedString.
	    setClass( srcType );
	}
	    
	if (clazz == null) {
	    if (this == CStdType.Object) {
		throw new RuntimeException("java.lang.Object is not on " 
					   + "the classpath !!!");
	    }
	    fail( "We're sorry, but you've found a bug in the type name "
		  + "resolution procedure. Please file a bug report at "
		  + "http://www.multijava.org.  Thanks!" );
	}

	return clazz;
    }

     public CClassType[] getArguments() {
	return CClassType.EMPTY;
    }
    
    /**
     * return the arguments of this type
     */
    public CClassType[][] getAllArguments() {
	return new CClassType[][]{CClassType.EMPTY};
    }
    
    public void setAllArguments(CClassType[][] arguments) {
	this.arguments = arguments;
    }
    // ----------------------------------------------------------------------
    // PUBLIC CONSTANTS
    // ----------------------------------------------------------------------

    public static final CClassType[]		EMPTY = new CClassType[0];
    public static final CClassType[][]          EMPTY_ARG = new CClassType[0][];

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private static final CClass	BAC_CLASS = new CBadClass("<NOT YET DEFINED>");

    private CClass		clazz;
    //@                            in thisClassInfo;
    protected int index;
    protected CClassType[][]    arguments;
    //@ private represents thisClassInfo <- JMLDataGroup.IT;

    /** Stores the universe modifier that is attached with this class type.
     *  By WMD.
     */
    protected CUniverse           universe;

    // We cache the value of toString in order to improve performance;
    // profiling said toString was called often.
    private String cachedString = null; //@ in thisClassInfo;
}
