/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CCompilationUnit.java,v 1.20 2005/05/09 16:25:36 f_rioux Exp $
 */

package org.multijava.mjc;

import java.util.Collection;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.MessageDescription;
import org.multijava.util.Message;

/**
 * This class handles namespace management for a compilation unit.
 * Based on the imports and package of a compilation unit, this class
 * includes methods for resolving partially-qualified names.  This
 * class also tracks the external members declared in a compilation
 * unit.  */
public class CCompilationUnit implements CMemberHost {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a minimal compilation unit context for hosting
     * <code>CBinaryClass</code>es.
     *
     * @param packageName the packageName of the compilation unit */
    public CCompilationUnit( /*@non_null@*/ String packageName ) {
	this.packageName = packageName.intern();
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------

    /**
     * Searches for a class with the given simple name according the
     * procedure in JLS2 6.5.5.  By the time the search reaches this
     * method the first three bullet points of the procedure have been
     * checked.  This method checks the remaining bullet points:
     *
     * <UL>
     *  <LI>
     *   Otherwise, if a type with that name is declared in the current 
     *   compilation unit, either by a single-type-import declaration or 
     *   by a declaration of a class or interface type, then the simple type 
     *   name denotes that type. 
     *  </LI>
     *  <LI>
     *   Otherwise, if a type with that name is declared in another 
     *   compilation unit of the package containing the identifier, then the 
     *   identifier denotes that type.
     *  </LI>
     *  <LI>
     *   Otherwise, if a type of that name is declared by exactly one type-
     *   import-on-demand declaration of the compilation unit containing the 
     *   identifier, then the simple type name denotes that type. 
     *  </LI>
     *  <LI>
     *   Otherwise, if a type of that name is declared by more than one 
     *   type-import-on-demand declaration of the compilation unit, then the 
     *   name is ambiguous as a type name; a compile-time error occurs. 
     *  </LI>
     * </UL>
     * 
     * @param	name	the class name, without qualifiers
     * @param	context	used for error reporting
     * @return	the class if found, null otherwise
     * @exception UnpositionedError if search fails */
    public CClass lookupClass( /*@non_null@*/ String name, /*@non_null@*/ CContextType context ) 
	throws UnpositionedError 
    {
	CClassType type;

	// First look for a type declared by a single-type-import or
	// by a type declaration.  We don't have to worry about
	// duplicates since those are caught while typechecking the
	// compilation unit.
	if ((type = (CClassType)loadedClasses.get(name)) != null) {
	    // If type is declared by a single-type-import, mark
	    // it as used (max. 1)
	    for( int i = 0; i < importedClasses.length; i++ ) {
		if( name == importedClasses[i].ident() ) {
		    importedClasses[i].setUsed();
		    break;
		}
	    }

	    return type.getCClass();
	}

	// Otherwise, look for a type declared in another compilation
	// unit of this package.
	CClass clazz = null;
	String temp = packageName + name;
	clazz = CTopLevel.classExists(temp) ? CTopLevel.loadClass(temp) : null;
	if (clazz == CClass.CLS_UNDEFINED) {
	    clazz = null;
	}

	if (clazz != null) {
	    // We found a match in the same package, so we note that the class
	    // is visible.  This also makes for more efficient location of the
	    // class on subsequent searches.
	    type = clazz.getType();
	    loadedClasses.put( name, type );
	    return clazz;
	}

	// Otherwise, look for a type declared by EXACTLY ONE
	// import-on-demand declaration
	for( int i = 0; i < importedPackages.length; i++ ) {
	    String qualifiedName = 
		(importedPackages[i].getName() + '/' + name).intern();

	    if( CTopLevel.classExists( qualifiedName ) ) {
		CClass candidate = CTopLevel.loadClass( qualifiedName );
		if (candidate == null || candidate == CClass.CLS_UNDEFINED) {
		    // If the class exists but cannot be loaded there
		    // is probably either a parsing error in the java
		    // file or some file io error in reading either
		    // the class or java file.
		    throw new UnpositionedError(//  !FIXME! make this a real message in MjcMessages.msg
			new Message(
			    new MessageDescription("Failed to load {0}","",MessageDescription.LVL_ERROR),
			    qualifiedName));
		}
		
		if (!candidate.qualifiedName().equals(qualifiedName)) {
		    // This candidate does not match the name we are
		    // looking for, probably because we are running on
		    // a case insensitive file system and we loaded a
		    // type Foo from a file named 'foo.class'.
		    continue;
		}
		
		if ( !candidate.isAccessibleFrom( this ) ) {
			// the class is not accessible, so it's not a match
			context.reportTrouble( 
			    new CWarning( NO_TOKEN,
					  MjcMessages.CLASS_NO_ACCESS,
					  name, qualifiedName ));
		} else if ( clazz != null ) {
		    // We already have a match from a previous
		    // iteration of the loop, so make sure the two
		    // classes are one in the same (this happens if
		    // the source code includes duplicate type-import-
		    // on-demand declarations).
		    if (!clazz.qualifiedName().equals(qualifiedName)) {
			throw new 
			    UnpositionedError( MjcMessages.CUNIT_RENAME2, 
					       qualifiedName,
					       clazz.qualifiedName() );
		    } else {
			importedPackages[i].setClassUsed ( name );
		    }
		} else {
			// we've got a unique match
		    clazz = candidate;
		    importedPackages[i].setClassUsed( name );
		}
	    }
	}

	if( clazz != null ) {
	    type = clazz.getType();
	    loadedClasses.put( name, type );
	    return clazz;
	}

	return null;		// not found
    }

    /**
     * Searches for any imported external generic functions.  Handles
     * implicit imports (generic functions in the current package or in
     * imported packages.)
     *
     * @param	context	the compilation unit context in which the generic
     *			functions found are to be visible
     * @param	name	the name of the generic function to search for
     */
    public void resolveMaybeExtMethodRef( /*@non_null@*/ CCompilationUnitContextType context, 
					  /*@non_null@*/ String name ) 
    {
	// Unlike the search for classes above it is possible for
	// multiple generic functions of the same name to be in scope
	// as long as the type tuples of their top methods are
	// distinct.  We just apply the augmentation for all of them
	// and let the method lookup code flag any ambiguities.
	if (!resolvedGFNames.contains(name)) {
	    // find the collections of the local package
	    CGenericFunctionCollection packageCollection = 
		CTopLevel.lookupGFCollection(packageName + name);
	    if (packageCollection != CGenericFunctionCollection.GF_UNDEFINED) {
		loadedGFCollections.register(name,packageCollection);
	    }

	    // find the import-on-demand collections
	    for( int i = 0; i < importedPackages.length; i++ ) {
		String qualifiedName = 
		    importedPackages[i].getName() + '/' + name;
		CGenericFunctionCollection coll =
		    CTopLevel.lookupGFCollection( qualifiedName );
		if (coll != CGenericFunctionCollection.GF_UNDEFINED) {
		    loadedGFCollections.register(name,coll);
		}
	    }

	    // done resolving name for this compilation unit, no need
	    // to do it again
	    resolvedGFNames.add(name);
	}

	CGenericFunctionCollection[] collections = 
	    loadedGFCollections.getCollectionsFor( name );
	for( int i=0; i < collections.length; i++ ) {
	    collections[i].applyAugmentationFor( context );
	}
    }

    /**
     * Registers the declaration of an external generic function in
     * this compilation unit. */
    void registerGFDecl(/*@non_null@*/ String ident, /*@non_null@*/ CGenericFunctionCollection coll) {
	loadedGFCollections.register(ident, coll);
    }

    /**
     * Registers that a declaration of the given method occurs in the
     * compilation unit represented by this.*/
    public void registerVisibleMethod( /*@non_null@*/ CMethod method ) {
	visibleMethods.add( method );
    }

    /**
     * Registers that a reference to the given type occurs in the
     * compilation unit represented by this.*/
    public void registerVisibleType( /*@non_null@*/ CType type ) {
	directlyVisibleTypes.add( type );
    }

    //---------------------------------------------------------------------
    // ACCESSORS (UPDATE)
    //---------------------------------------------------------------------

    /**
     * Updates the compilation unit after completion of grouping
     * open methods.
     *
     * @param	importedClasses		the single-type imports
     * @param	importedGFs		the single-GF imports
     * @param	importedPackages	the import-on-demand imports
     * @param	loadedClasses		the classes actually loaded by this,
     *					initialized in <code>JCompilationUnit
     *					</code> based on declarations and
     *					imports, and then maintained here for
     *					other references
     * @param	loadedGFCollections	the external GFs actually loaded
     * @see JCompilationUnitType#checkInterface( Main )
     */
    public void update( JClassOrGFImportType[] importedClasses,
			JClassOrGFImportType[] importedGFs,
			JPackageImportType[] importedPackages,
			Hashtable loadedClasses,
			CGFCollectionMap loadedGFCollections ) {
	this.importedClasses = importedClasses;
	this.importedGFs = importedGFs;
	this.importedPackages = importedPackages;
	this.loadedClasses = loadedClasses;
	this.loadedGFCollections = loadedGFCollections;
    }

    //---------------------------------------------------------------------
    // CMemberHost INTERFACE
    //---------------------------------------------------------------------

    /**
     * Indicates whether this is accessible from the given host.
     *
     * @param	from	the host that wants access
     * @return	true iff the given host is allowed access to this host's 
     *		members
     */
    public boolean isAccessibleFrom( CMemberHost from ) {
	return true;
    }

    /**
     * Indicates whether this is defined inside the given host.
     *
     * @param	maybeOutside	the host that this may be inside
     * @return true if the given host is a (possibly reflexive and
     * transitive) host of this */
    public boolean isDefinedInside( CMemberHost maybeOutside ) {
	return this == maybeOutside;
    }

    /**
     * Indicates whether this host is a subclass of the given class,
     * where "subclass" is the reflexive, transitive closure of the
     * extends relation.
     *
     * @param	maybeSuper	the potential superclass
     */
    public boolean descendsFrom( CClass maybeSuper ) {
	return false;
    }

    /**
     * Returns true if this host is a subclass of the given class or
     * if some host of this host is a subclass of the given class.
     *
     * @param maybeOuterSuper the potential superclass */
    public boolean nestedDescendsFrom( CClass maybeOuterSuper ) {
	return false;
    }

    /**
     * Indicates whether this host is a class.
     *
     * @return	true iff this host represents a class
     */
    public boolean isClass() { return false; }

    /**
     * Indicates whether this host is a nested type, i.e., a member
     * type, a type declared within another type.  */
    public boolean isNestedType() { return false; }

    /**
     * Indicates whether this host is an inner class, i.e., a
     * non-static member class.  */
    public boolean isInnerClass() { return false; }

    /**
     * Indicates whether this host is a compilation unit
     *
     * @return	true iff this host represents a compilation unit
     */
    public boolean isCompilationUnit() { return true; }

    /**
     * Gives the identifier for this host, or null for a compilation unit.
     */
    public String ident() { return null; }

    /**
     * Gives the package name for this host.
     */
    public String packageName() { return packageName; }

    /**
     * Returns the host of this host, that is, the context in which
     * this host is declared.  This value is null except for inner
     * classes.
     *
     * @return the host of this host */
    public CMemberHost host() { return null; }

    //---------------------------------------------------------------------
    // ACCESSORS (QUICK)
    //---------------------------------------------------------------------

    /**
     * Returns a string representation of this.
     *
     * @return	a string representation of this
     */
//      public String toString() {
//  	return "compilation unit " + packageName + "." + "???";
//  	// !FIXME! replace ??? with filename which we can get from the
//  	// token reference that should be stored with this
//      }

    /**
     * Returns the methods that are declared within this compilation unit.
     */
    public /*@non_null@*/ Collection visibleMethods() {
	return (Collection) visibleMethods.clone();
    }

    /**
     * Returns the set of (non-reference) types that are directly
     * visible in this compilation unit.  */
    public /*@non_null@*/ Collection directlyVisibleTypes() {
	return (Collection) directlyVisibleTypes.clone();
    }

    /**
     * Returns the set of (non-reference) types that are directly or
     * transitively visible in this compilation unit.  For efficiency
     * this method caches its result.  The method should only be
     * called after all members of the compilation unit have been
     * typechecked. */
    public Collection visibleTypes() {
	return CClassType.getVisibleTypesFrom( directlyVisibleTypes );
    }

    public JClassOrGFImportType[] importedClasses() { return importedClasses; }

    public JPackageImportType[]    importedPackages() { return importedPackages; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private final String		packageName;

    private /* final */ JClassOrGFImportType[]	importedClasses;
    private /* final */ JClassOrGFImportType[]	importedGFs;
    private /* final */ JPackageImportType[]	importedPackages;

    /**
     * Maps from type names to classes.  Tracks all classes referenced
     * within this compilation unit.  Used to efficiently resolve
     * short class names to their fully qualified types. Invariant:
     * all values stored in table are of type <code>CClassType</code>
     * and have had <code>checkType</code> called on them.
     *
     * @see CClassType
     * @see JCompilationUnit
     * @see JCompilationUnit#allLoadedClasses 
     */
    private Hashtable loadedClasses;

    /**
     * Maps from identifiers to sets of external generic function
     * collections that are in scope and loaded for this compilation unit.
     * Single-gf-import-on-demand clauses and local external generic
     * function declarations are loaded as early as possible.  Generic
     * functions that are in scope because of the package declaration
     * or import-on-demand clauses are added to the map lazily.
     * @see JCompilationUnit#inScopeGenFuncMap alias to this field in 
     *						JCompilatoinUnit
     */
    private CGFCollectionMap loadedGFCollections;


    /**
     * Tracks the names of the generic functions for which we have
     * completed the package and import-on-demand search.  
     * @see #resolveMaybeExtMethodRef(CCompilationUnitContextType,String)
     */
    private /*@non_null@*/ HashSet resolvedGFNames = new HashSet();

    /**
     * Tracks the methods that are referenced within this.  */
    private /*@non_null@*/ HashSet visibleMethods = new HashSet();

    /**
     * The set of reference types that are directly visible in this
     * compilation unit.  It's important for this field to be an
     * ordered collection, because otherwise we see some
     * cross-platform discrepancies in the error messages produced.
     */
    private /*@non_null@*/ ArrayList directlyVisibleTypes = new ArrayList();

    /**
     * Used for error reporting.  */
    private static final /*@non_null@*/ TokenReference NO_TOKEN = TokenReference.NO_REF;
}


