/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CCompilationUnitContextType.java,v 1.15 2005/12/23 18:14:20 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents the context for a compilation unit during checking
 * passes (checkInterface, checkInitializers, typecheck).
 * @see CContextType
 */
public interface CCompilationUnitContextType extends CContextType {

    // ----------------------------------------------------------------------
    // ACCESSORS (INFOS)
    // ----------------------------------------------------------------------

    /**
     * Fields cannot be declared in the context of a compilation unit
     * (yet), consequently we do not expect to invoke this method on a
     * compilation unit context.
     *
     * @exception	UnsupportedOperationException */
    CVariableState fieldInfo(int pos);

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------

    /**
     * Searches for a class with the given simple name according the procedure
     * in JLS2 6.5.5.
     * 
     * @param	name	the class name, without qualifiers
     * @return	the class if found, null otherwise
     * @exception	UnpositionedError	if search fails
     */
    CClass lookupClass( String name ) throws UnpositionedError;

     /**
     * Searches for the most specific method applicable to the given
     * identifier and argument type tuple, in the current context.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @param	context		the context of the class containing
     * 				the method call
     * @exception UnpositionedError if the result is ambiguous */
    CMethod lookupMethod( String name, CType[] params,
				 CClassContextType context ) 
	throws UnpositionedError;

     /**
     * Searches for the most specific method(s) applicable to the given
     * identifier and argument type tuple, in the current context.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @param	context		the context of the class containing
     * 				the method call
     * @exception UnpositionedError if the result is ambiguous */
    CMethodSet lookupMethodOrSet( String name, CType[] params,
				  CClassContextType context ) 
	throws UnpositionedError;

   /**
     * searches for a field with the given identifier
     * @return a field from an ident in current context
     * @exception	UnpositionedError	this error will be positioned soon
     */
    CFieldAccessor lookupField(String ident, CExpressionContextType context)
	throws UnpositionedError;

    /**
     * Finds a local variable with the given name that appears outside
     * the current lexical contour.
     *
     * @param	ref		a token reference used to build a new
     *				<code>JOuterLocalVariableExpression</code>
     * @param	ident		the name of the outer variable
     * @return a variable from an ident in the surrounding context, or null
     *		if not found */
    JExpression lookupOuterLocalVariable( TokenReference ref, 
						 String ident );

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    CFieldAccessor lookupOuterField(String ident,
				    CExpressionContextType context)
	throws UnpositionedError;

    /**
     * Searches for any imported external generic functions.
     *
     * @param	ident	the name of the generic function to search for
     */
    void resolveMaybeExtMethodRef( String ident );

    /**
     * Registers that a declaration of the given method occurs within
     * the compilation unit. */
    void registerVisibleMethod( CMethod method );

    /**
     * Registers that a reference to the given type occurs within
     * the compilation unit. */
    void registerVisibleType( CType type );

    // ----------------------------------------------------------------------
    // ACCESSORS (TREE HIERARCHY)
    // ----------------------------------------------------------------------

    /**
     * getParentContext
     * @return	the parent
     */
    CContextType getParentContext();

    /**
     * getClass
     * @return	the near parent of type CClassContextType
     */
    CClassContextType getClassContext();

    /**
     * getMethod
     * @return	the near parent of type CMethodContextType
     */
    CMethodContextType getMethodContext();

    /**
     * @return	the compilation unit
     */
    CCompilationUnitContextType getCompilationUnit();

    /** 
     * Returns the nearest control flow context (where control flow
     * information is stored.)
     * @return the nearest parent of type CFlowControlContextType
     */
    CFlowControlContextType getFlowControlContext();

    /**
     * Returns the signature of the nearest lexically enclosing
     * context that can host member declarations (i.e., a
     * <code>CClass</code> or a <code>CCompilationUnit</code>).  Used
     * for access checks, this is slightly different than
     * <code>getHostClass()</code>.  The later method returns an
     * anchor class signature for external methods, which is useful in
     * code generation and in typechecking external method bodies.  On
     * the other hand, this method returns a
     * <code>CCompilationUnit</code> for external methods, which is
     * useful in checking access between two differently named generic
     * functions declared in the same compilation unit.  !FIXME! It
     * may be possible to eliminate this method and just use
     * <code>getHostClass</code> now that private external methods are
     * duplciated in every anchor.
     *
     * @return	the nearest parent that can be coerced to a 
     * <code>CMemberHost</code>, i.e., <code>this.cunit</code> */
    CMemberHost findNearestHost();

    // ----------------------------------------------------------------------
    // ACCESSORS (NEW CONTEXT)
    // ----------------------------------------------------------------------

    CBinaryClassContext createBinaryClassContext(CBinaryClass clazz);

    // ----------------------------------------------------------------------
    // ERROR HANDLING
    // ----------------------------------------------------------------------

    /**
     * Adds an trouble into the list and eats it This method should be
     * called after a try catch block after catching exception or
     * directly without exception thrown
     * @param trouble the trouble
     */
    void reportTrouble(Exception trouble);

    /**
     * Gets the compiler
     */
    org.multijava.mjc.Main getCompiler();
  
    // ----------------------------------------------------------------------
    // CLASS HANDLING
    // ----------------------------------------------------------------------

    /**
     * Adds a class to generate
     */
    void classToGenerate( CSourceClass clazz );

    /**
     * Indicates whether this is equal to a given object.
     *
     * @param	o		the object to compare against
     * @return	true iff <code>o</code> is another 
     *		<code>CCompilationUnitContextType</code> formed around an
     *		identical (i.e., <code>==</code>) <code>CCompilationUnit</code>
     */
    boolean equals(/*@ nullable @*/  Object o );

    /**
     * Returns the hash code for this, equivalent to 
     * <code>this.cunit.hashCode()</code>.
     *
     * @return	the hash code for this
     */
    int hashCode();
}
