/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CContext.java,v 1.34 2006/12/11 05:55:56 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.mjc.MJMathMode;

import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

import org.multijava.util.MessageDescription;
import org.multijava.util.Utils;

/**
 * <p>Descendents of this class represent local contexts during
 * checking passes (checkInterface, checkInitializers, typecheck).
 * The context classes follow the control flow and maintain
 * information about variables (initialized, used, allocated) and
 * exceptions (thrown, caught).  They also verify that a context is
 * still reachable.  The contexts are used for name resolution (JLS2,
 * 6 and 14.4), checking for unreachable statements (JLS2 14.20),
 * checking for definite assignment (JLS2, 16), and for checking for
 * proper subclassing and overriding (JLS2, 8).</p>
 *
 * <p>The contexts are hierarchical. To wit, a compilation unit context
 * is created and passed to the method that checks a class.  This
 * method creates a class context by passing the compilation unit
 * context to the CClassContext constructor.  When checking methods of
 * the class the CClassContext instance is passed to the CMethodContext
 * constructor.  These context classes are stored in fields of the AST
 * nodes.</P>
 *
 * <P>During typechecking the context hierarchy is mutated in a way
 * that mimics the actual run-time control flow.  For example,
 * variable declarations cause the context to be mutated to record the
 * variable's name and type.  Assignment to a variable causes the
 * context to be mutated to record that the variable is definitely
 * assigned.  Reference a variable's value triggers a check that the
 * variable has been definitely assigned.</P>
 *
 * */
public abstract class CContext extends Utils 
    implements CContextType, Constants 
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a new context.
     *
     * <pre><jml>
     * requires this instanceof CCompilationUnitContextType;
     * </jml></pre>
     */
    CContext() {
	cunit = (CCompilationUnitContextType)this;
    }

    /**
     * Constructs a nested context.
     * @param	parent		the parent context, it must be different
     *				than null except if called by the top level
     */
    protected CContext(CContextType parent) {
	this.parent = parent;
	cunit = parent.getCompilationUnit();

        // Nullity info does not evolve monotonically (e.g. consider 
	// typechecking of JIfStatement) and over time we want this context 
	// to be unaffected by nullity info changes of the parent, hence 
	// we make a copy now:

	addFANonNulls(parent.getFANonNulls());
	addFANulls(parent.getFANulls());
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------

    /**
     * Searches for a class with the given simple name according the procedure
     * in JLS2 6.5.5.
     * 
     * @param	name	the class name, without qualifiers
     * @return	the class if found, null otherwise
     * @exception	UnpositionedError	if search fails
     */
    public CClass lookupClass(String name) throws UnpositionedError {
	return parent.lookupClass(name);
    }
    
    public /*@nullable*/ CTypeVariable lookupTypeVariable(String ident) 
	throws UnpositionedError {
	return parent.lookupTypeVariable(ident);
    }
    /**
     * Searches for the most specific method in the current context
     * that is applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @exception UnpositionedError if the result is ambiguous */
    public CMethod lookupMethod( String name, CType[] params ) 
	throws UnpositionedError 
    {
	return lookupMethod( name, params, getClassContext() );
    }

    /**
     * Searches for the most specific method in the current context
     * that is applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @param	context		the context of the class containing
     * 				the method call
     * @exception UnpositionedError if the result is ambiguous */
    public CMethod lookupMethod( String name, CType[] params,
				 CClassContextType context ) 
	throws UnpositionedError 
    {
	return parent.lookupMethod( name, params, context );
    }

    /**
     * Searches for the most specific method(s) in the current context
     * that is applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @exception UnpositionedError if the result is ambiguous */
    public CMethodSet lookupMethodOrSet( String name, CType[] params ) 
	throws UnpositionedError 
    {
	return lookupMethodOrSet( name, params, getClassContext() );
    }

    /**
     * Searches for the most specific method(s) in the current context
     * that are applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @param	context		the context of the class containing
     * 				the method call
     * @exception UnpositionedError if the result is ambiguous */
    public CMethodSet lookupMethodOrSet( String name, CType[] params,
					 CClassContextType context ) 
	throws UnpositionedError 
    {
	return parent.lookupMethodOrSet( name, params, context );
    }

    /**
     * searches for a field with the given identifier
     * @return a field from an ident in current context
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupField(String ident) throws UnpositionedError {
	return parent.lookupField(ident );
    }

    /**
     * searches for a field with the given identifier
     * @return a field from an ident in current context
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupField(String ident,
				      CExpressionContextType context)
	throws UnpositionedError {
	return parent.lookupField(ident, context);
    }

    /**
     * searches for a local variable with the given identifier
     * @param	ident		the name of the local variable
     * @return	a variable from an ident in current context
     */
    public JLocalVariable lookupLocalVariable(String ident) {
	return parent.lookupLocalVariable(ident);
    }

    /**
     * Finds a local variable with the given name that appears outside
     * the current lexical contour.
     *
     * @param	ref		a token reference used to build a new
     *				<code>JOuterLocalVariableExpression</code>
     * @param	ident		the name of the outer variable
     * @return a variable from an ident in the surrounding context, or null
     *		if not found */
    public JExpression lookupOuterLocalVariable( TokenReference ref, 
						 String ident ) {
	return parent.lookupOuterLocalVariable( ref, ident );
    }

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupOuterField(String ident)
	throws UnpositionedError {
	return parent.lookupOuterField( ident );
    }

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupOuterField(String ident,
					   CExpressionContextType context)
	throws UnpositionedError {
	return parent.lookupOuterField( ident, context );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (INFOS)
    // ----------------------------------------------------------------------

    /**
     * Replaces the local field info for fields in positions 0 up to
     * <code>pos</code> with the info in <code>replacement</code>.  */
    public void replaceFieldInfoUpTo( int pos, 
				      CVariableInfoTable replacement ) {
	parent.replaceFieldInfoUpTo( pos, replacement );
    }

    /**
     * Marks the field with the given descriptor as definitely assigned
     * to in this context.  */
    public void initializeField(VariableDescriptor varDesc) {
	parent.initializeField(varDesc);
    }

    /**
     * Indicates whether the field in the given position is
     * definitely assigned to in this context.  */
    public boolean isFieldDefinitelyAssigned(int pos) {
	return parent.isFieldDefinitelyAssigned(pos);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (TREE HIERARCHY)
    // ----------------------------------------------------------------------

    /**
     * Returns the parent context for this context
     * @return	the parent
     */
    public CContextType getParentContext() {
	return parent;
    }

    /**
     * Returns the compilation unit context for this context.  This
     * only makes sense for contexts that do not themselves represent
     * compilation units.  I.e., in classes that represent compilation
     * unit context this should be overridden.
     * @return the compilation unit 
     */
    public CCompilationUnitContextType getCompilationUnit() {
	return cunit;
    }

    /**
     * Returns the class context for this context.  This only makes
     * sense for contexts that can appear inside classes.  I.e., This
     * should be overridden in classes representing other sorts of contexts.
     * @return the nearest parent of type CClassContextType.
     */
    public CClassContextType getClassContext() {
	return parent.getClassContext();
    }

    /**
     * Returns the method context for this context.  This only makes
     * sense for contexts that can appear inside methods.  I.e., This
     * should be overridden in classes representing other sorts of contexts.
     * @return	the nearest parent of type CMethodContextType.
     */
    public CMethodContextType getMethodContext() {
	return parent.getMethodContext();
    }

    /** 
     * Returns the nearest control flow context (where control flow
     * information is stored.)
     * @return the nearest parent of type CFlowControlContextType
     */
    public CFlowControlContextType getFlowControlContext() {
	return parent.getFlowControlContext();
    }

    /**
     * Returns the signature of the method declaration in which this
     * context is enclosed, or null if this context is not enclosed in
     * a method declaration.  */
    //@ pure
    public CMethod getCMethod() {
	CMethodContextType methodContext = getMethodContext();
	return methodContext == null ? null : methodContext.getCMethod();
    }

    /**
     * Returns the signature of the nearest lexically enclosing
     * context that can host member declarations (i.e., a
     * <code>CClass</code> or a <code>CCompilationUnit</code>).  Used
     * for access checks, this is slightly different than
     * <code>getHostClass()</code>.  The later method returns an
     * anchor class signature for external methods, which is useful in
     * code generation and in typechecking external method bodies.  On
     * the other hand, this method returns a
     * <code>CCompilationUnit</code> for external methods, which is
     * useful in checking access between two differently named generic
     * functions declared in the same compilation unit.  !FIXME! It
     * may be possible to eliminate this method and just use
     * <code>getHostClass</code> now that private external methods are
     * duplciated in every anchor.
     *
     * @return	the nearest parent that can be coerced to a 
     * <code>CMemberHost</code> */
    public CMemberHost findNearestHost() {
	return parent.findNearestHost();
    }


    /**
     * Creates a class context with this context as its parent.
     * @param	self		the corresponding class
     */
    public CClassContextType createClassContext( CClass self ) {
	return new CClassContext( this, self );
    }

    /**
     * Creates an interface context with this context as its parent.
     * @param	self		the corresponding interface signature
     */
    public CInterfaceContextType createInterfaceContext( CClass self ) {
	return new CInterfaceContext( this, self );
    }


    /**
     * Create a new child of this context representing the context in
     * which an external method declaration is typechecked. 
     * @param	host		the host class
     * @param	owner		the receiver class
     */
    public CExtMethodContext createExtMethodContext( CSourceClass host,
						     CClass owner ) {
	return new CExtMethodContext( this, host, owner );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (NULLITY)
    // ----------------------------------------------------------------------

    /**
     * Indicates whether expr (or member) is conditionally NonNull is this context.
     */
    /*@ also 
      @   requires expr instanceof JPhylum 
      @         || expr instanceof JClassFieldExpression;
      @*/
    public /*@pure*/ boolean isFANonNull(/*@non_null*/Object expr) {
       if (expr instanceof CFieldAccessor)
	  return contextNullity.isFANonNull(((CFieldAccessor)expr).getField());
       if (expr instanceof CField)
	  return contextNullity.isFANonNull((CField)expr);
       if (expr instanceof JClassFieldExpression)
          return contextNullity.isFANonNull(((JClassFieldExpression)expr).getField());
       else 
          return contextNullity.isFANonNull((JPhylum)expr);
    }
 
    /**
     * returns the JPhyla that are known to be non-null in this context.
     */
     public /*@pure non_null*/ Object[] getFANonNulls() {
       return contextNullity.getFANonNulls();
     }
     public /*@pure non_null*/ Object[] getFANulls() {
       return contextNullity.getFANulls();
     }
    /**
     * Mark expr (or member) as NonNull in this context
     */
    public void addFANonNull(/*@non_null*/ Object expr) {
       contextNullity.addFANonNull(expr);
    }
    public void addFANull(/*@non_null*/ Object expr) {
       contextNullity.addFANull(expr);
    }
    public void removeFANonNull(/*@non_null*/ Object expr) {
       contextNullity.removeFANonNull(expr);
    }
    public void removeAllFANullity() {
       contextNullity.removeAllFANullity();
    }
    /**
     * adds exprs (or members) as NonNull in this context
     */
    public void addFANonNulls(/*@non_null*/ Object[] exprs) {
       contextNullity.addFANonNulls(exprs);
    }
    public void addFANulls(/*@non_null*/ Object[] exprs) {
       contextNullity.addFANulls(exprs);
    }
    /**
     * Merge the list of nulls in this context with that of the given context.
     * This is done by taking a union of the two sets. 
     */
    public void mergeNullityInfo(/*@non_null*/ CContextType other) {
       contextNullity.mergeNullityInfo(other);
    }
    public void adoptNullityInfo(/*@non_null*/ CContextType other) {
       contextNullity.adoptNullityInfo(other);
    }
    public void dumpNonNulls(/*@non_null*/ String msg) {
       contextNullity.dumpNonNulls(msg);
    }

 
    // ----------------------------------------------------------------------
    // CLASS AND EXTERNAL GF HANDLING
    // ----------------------------------------------------------------------

    /**
     * Adds a source class to be generated
     *
     * @param clazz	the class to be generated
     */
    public void classToGenerate(CSourceClass clazz) {
	getCompilationUnit().classToGenerate(clazz);
    }

    /**
     * Searches for any imported external generic functions.
     *
     * @param	ident	the name of the generic function to search for
     */
    public void resolveMaybeExtMethodRef( String ident ) {
	parent.resolveMaybeExtMethodRef( ident );
    }

    /**
     * Registers the declaration of an external generic function in
     * this context. */
    public void registerGFDecl(String methodIdent, 
			       CGenericFunctionCollection coll) {
	getCompilationUnit().registerGFDecl(methodIdent,coll);
    }
    
    /**
     * Registers with the surrounding context that a declaration of
     * the given method occurs in this context. */
    public void registerVisibleMethod( CMethod method ) {
	// jump immediately to the top of the context hierarchy
	getCompilationUnit().registerVisibleMethod( method );
    }

    /**
     * Registers with the surrounding context that a reference to
     * the given type occurs in this context. */
    public void registerVisibleType( CType type ) {
	// Unlike with registerVisibleMethod above, we climb out of
	// the contexts one-by-one so the inner-most class context can
	// track the inner class used inside it.  These are needed for
	// code generation, particularly for the InnerClasses
	// attribute (see JVM2, 4.7.5).
	parent.registerVisibleType( type );
    }

    // ----------------------------------------------------------------------
    // ERROR HANDLING
    // ----------------------------------------------------------------------

    /**
     * Reports the given error and the "swallows" it, allowing
     * compilation to continue.  This method is used to report
     * non-fatal errors; for fatal errors the exception is just
     * thrown.
     *
     * @param trouble the error */
    public void reportTrouble(Exception trouble) {
	getCompilationUnit().reportTrouble(trouble);
    }

    /**
     * Generates an UnpositionedError with a given message.
     * @param	mess		the error message
     * @param	param1		the first message parameter
     * @param	param2		the second message parameter
     * @return	never, i.e., this method always throws an exception
     * @exception	UnpositionedError	the error exception
     */
    public void fail(MessageDescription mess, Object param1, Object param2)
	throws UnpositionedError
    {
	throw new UnpositionedError(mess, param1, param2);
    }

    /**
     * Generates an UnpositionedError with a given message.
     * @param	mess		the error message
     * @param	params		the message parameters
     * @return	never, i.e., this method always throws an exception
     * @exception	UnpositionedError	the error exception
     */
    public void fail(MessageDescription mess, Object[] params)
	throws UnpositionedError
    {
	throw new UnpositionedError(mess, params);
    }

    /**
     * Verifies an expression and if false signals an error.
     * @param	expr		condition to verify
     * @param	mess		the message to be displayed
     * @return	true iff expr is true
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public boolean check(boolean expr,
			 MessageDescription mess) throws UnpositionedError {
	if (!expr) {
	    fail(mess, null, null);
	}
	return expr;
    }

    /**
     * Verifies an expression and if false signals an error.
     * @param	expr		condition to verify
     * @param	mess		the message to be displayed
     * @param	param1		the first parameter
     * @return	true iff expr is true
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public boolean check(boolean expr,
			 MessageDescription mess,
			 Object param1) throws UnpositionedError {
	if (!expr) {
	    fail(mess, param1, null);
	}
	return expr;
    }

    /**
     * Verifies an expression and if false signals an error.
     * @param	expr		condition to verify
     * @param	mess		the message to be displayed
     * @param	param1		the first parameter
     * @param	param2		the second parameter
     * @return	true iff expr is true
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public boolean check(boolean expr,
			 MessageDescription mess,
			 Object param1,
			 Object param2) throws UnpositionedError {
	if (!expr) {
	    fail(mess, param1, param2);
	}
	return expr;
    }

    public boolean check(boolean expr,
                         MessageDescription mess,
			 Object[] params) throws UnpositionedError {
	if (!expr) {
	    fail(mess, params);
        }
	return expr;
    }

    // -------------------------------------------------------------
    // UTILITY METHODS
    // -------------------------------------------------------------

    /**
     * Gets the compiler for this context
     */
    public Main getCompiler() {
	return parent.getCompiler();
    }

    /**
     * Returns the modifier utility for this.  */
    public final ModifierUtility modUtil() {
	return getCompiler().modUtil();
    }

    /**
     * Indicates whether this context is enclosed in a loop. */
    public boolean isInLoop() {
	return parent != null && parent.isInLoop();
    }

    /**
     * Indicates whether this context is enclosed in a loop and the
     * given variable is declared outside the inner-most loop
     * context. */
    public boolean declaredOutsideOfLoop(JLocalVariable var) {
	return parent != null && parent.declaredOutsideOfLoop(var);
    }

    /**
     * Returns true if the given local variable is declared exactly in
     * this context, i.e., it is not declared in an outer context. */
    public boolean declares(JLocalVariable var) {
	// Local vars are only declared in flow control contexts.
	return false;
    }

    /**
     * Indicates whether this context is enclosed in an instance or
     * static initializer.  */
    public boolean isInInitializer() {
	return parent != null && parent.isInInitializer();
    }

    /**
     * Indicates whether this context is enclosed in a constructor.  */
    public boolean isInConstructor() {
	return parent != null && parent.isInConstructor();
    }

    /**
     * Indicates whether this context is enclosed in a constructor and occurs
     * before the constructor has invoked the superclass constructor.  */
    public boolean isBeforeSuperConstructorCall() {
	return isInConstructor()
	    && !(((CConstructorContextType) getMethodContext())
		 .isSuperConstructorCalled());
    }    

    /**
     * Indicates whether this context is "static".
     *
     * @return true iff this context is enclosed in a context for a
     * static initializer or static method.  */
    public boolean isStatic() {
	return parent != null && parent.isStatic();
    }


    /**
     * Indicates whether this context is "pure".
     *
     * @return true iff this context is enclosed in a pure method.
     */
    //@ also
    //@ assignable objectState;
    public boolean isPure() {
	return parent != null && parent.isPure();
    }

    /**
     * Calls back to the compiler for this context and requests that
     * the compiler catch-up decl (which is necessary the AST for an
     * anonymous type or a type declared in a type declaration
     * statement) to the same pass as the currently active task.  */
    public void catchUp(JTypeDeclarationType decl) {
	decl.cachePassParameters(this);
	getCompiler().catchUp(decl);
    }

    /**
     * Indicates the integral arithmetic mode that should be used.
     */
    public MJMathMode arithmeticMode(){
    	if(parent != null){
    		//return parent.arithmeticMode();
    		return parent.arithmeticMode();
    	}
	else{
		//return Constants.AMID_JAVA_MATH;	//default
		return MJMathMode.newMJMathMode();
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	CContextType		parent; //@ in objectState;
                                //@ maps parent.objectState \into objectState;
    protected	CCompilationUnitContextType	cunit;

    protected	final /*@non_null*/ CContextNullity contextNullity=new CContextNullity();
}
