/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CContextType.java,v 1.35 2008/05/14 18:16:25 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

import org.multijava.util.MessageDescription;
import org.multijava.util.Utils;

/**
 * <p>Descendents of this class represent local contexts during
 * checking passes (checkInterface, checkInitializers, typecheck).
 * The context classes follow the control flow and maintain
 * information about variables (initialized, used, allocated) and
 * exceptions (thrown, caught).  They also verify that a context is
 * still reachable.  The contexts are used for name resolution (JLS2,
 * 6 and 14.4), checking for unreachable statements (JLS2 14.20),
 * checking for definite assignment (JLS2, 16), and for checking for
 * proper subclassing and overriding (JLS2, 8).</p>
 *
 * <p>The contexts are hierarchical. To wit, a compilation unit context
 * is created and passed to the method that checks a class.  This
 * method creates a class context by passing the compilation unit
 * context to the CClassContext constructor.  When checking methods of
 * the class the CClassContext instance is passed to the CMethodContext
 * constructor.  These context classes are stored in fields of the AST
 * nodes.</P>
 *
 * <P>During typechecking the context hierarchy is mutated in a way
 * that mimics the actual run-time control flow.  For example,
 * variable declarations cause the context to be mutated to record the
 * variable's name and type.  Assignment to a variable causes the
 * context to be mutated to record that the variable is definitely
 * assigned.  Reference a variable's value triggers a check that the
 * variable has been definitely assigned.</P>
 *
 * */
public interface CContextType { 

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------

    /**
     * Searches for a class with the given simple name according the procedure
     * in JLS2 6.5.5.
     * 
     * @param	name	the class name, without qualifiers
     * @return	the class if found, null otherwise
     * @exception	UnpositionedError	if search fails
     */
    /*@nullable*/ CClass lookupClass(/*@non_null@*/ String name) throws UnpositionedError;

    /**
     * search for type variable that are accessible
     */
    /*@nullable*/ CTypeVariable lookupTypeVariable(/*@non_null@*/ String Ident) throws UnpositionedError;

    /**
     * Searches for the most specific method in the current context
     * that is applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @exception UnpositionedError if the result is ambiguous */
    /*@nullable*/ CMethod lookupMethod(/*@non_null@*/ String name, CType[] params ) 
	throws UnpositionedError;

    /**
     * Searches for the most specific method in the current context
     * that is applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @param	context		the context of the class containing
     * 				the method call
     * @exception UnpositionedError if the result is ambiguous */
    /*@nullable*/ CMethod lookupMethod( /*@non_null@*/ String name, CType[] params,
			  /*@non_null@*/ CClassContextType context ) 
	throws UnpositionedError;

    /**
     * Searches for the most specific method(s) in the current context
     * that is applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @exception UnpositionedError if the result is ambiguous */
    /*@nullable*/ CMethodSet lookupMethodOrSet( /*@non_null@*/ String name, CType[] params ) 
	throws UnpositionedError;

    /**
     * Searches for the most specific method(s) in the current context
     * that is applicable to the given identifier and argument type
     * tuple.
     *
     * @param	name		method name
     * @param	params		method parameter types
     * @param	context		the context of the class containing
     * 				the method call
     * @exception UnpositionedError if the result is ambiguous */
    /*@nullable*/ CMethodSet lookupMethodOrSet( /*@non_null@*/ String name, CType[] params,
				  /*@non_null@*/ CClassContextType context ) 
	throws UnpositionedError;

    /**
     * searches for a field with the given identifier
     * @return a field from an ident in current context
     * @exception	UnpositionedError	this error will be positioned soon
     */
    /*@nullable*/ CFieldAccessor lookupField(/*@non_null@*/ String ident) throws UnpositionedError;

    /**
     * searches for a field with the given identifier
     * @return a field from an ident in current context
     * @exception	UnpositionedError	this error will be positioned soon
     */
    /*@nullable*/ CFieldAccessor lookupField(/*@non_null@*/ String ident, /*@non_null@*/ CExpressionContextType context)
	throws UnpositionedError;

    /**
     * searches for a local variable with the given identifier
     * @param	ident		the name of the local variable
     * @return	a variable from an ident in current context
     */
    /*@nullable*/ JLocalVariable lookupLocalVariable(/*@non_null@*/ String ident);

    /**
     * Finds a local variable with the given name that appears outside
     * the current lexical contour.
     *
     * @param	ref		a token reference used to build a new
     *				<code>JOuterLocalVariableExpression</code>
     * @param	ident		the name of the outer variable
     * @return a variable from an ident in the surrounding context, or null
     *		if not found */
    /*@nullable*/ JExpression lookupOuterLocalVariable( /*@non_null@*/ TokenReference ref, 
					  /*@non_null@*/ String ident );

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    /*@nullable*/ CFieldAccessor lookupOuterField(/*@non_null@*/ String ident) throws UnpositionedError;

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    /*@nullable*/ CFieldAccessor lookupOuterField(/*@non_null@*/ String ident,
				    /*@non_null@*/ CExpressionContextType context)
	throws UnpositionedError;

    /**
     * Replaces the local field info for fields in positions 0 up to
     * <code>pos</code> with the info in <code>replacement</code>.  */
    void replaceFieldInfoUpTo( int pos, CVariableInfoTable replacement );
    /**
     * Marks the field with the given descriptor as definitely assigned
     * to in this context.  */
    void initializeField(VariableDescriptor varDesc);

    /**
     * Indicates whether the field in the given position is
     * definitely assigned to in this context.  */
    boolean isFieldDefinitelyAssigned(int pos);

    // ----------------------------------------------------------------------
    // ACCESSORS (TREE HIERARCHY)
    // ----------------------------------------------------------------------

    /**
     * Returns the parent context for this context
     * @return	the parent
     */
    /*@ pure non_null @*/ CContextType getParentContext();

    /**
     * Returns the compilation unit context for this context.  This
     * only makes sense for contexts that do not themselves represent
     * compilation units.  I.e., in classes that represent compilation
     * unit context this should be overridden.
     * @return the compilation unit 
     */
    /*@ pure non_null @*/ CCompilationUnitContextType getCompilationUnit();

    /**
     * Returns the class context for this context.  This only makes
     * sense for contexts that can appear inside classes.  I.e., This
     * should be overridden in classes representing other sorts of contexts.
     * @return the nearest parent of type CClassContextType.
     */
    /*@ pure non_null @*/ CClassContextType getClassContext();

    /**
     * Returns the method context for this context.  This only makes
     * sense for contexts that can appear inside methods.  I.e., This
     * should be overridden in classes representing other sorts of contexts.
     * @return	the nearest parent of type CMethodContextType.
     */
    /*@ pure non_null@*/ CMethodContextType getMethodContext();

    /** 
     * Returns the nearest expression context.
     * @return the nearest parent of type CExpressionContextType
     */
    /*@ pure non_null@*/ CFlowControlContextType getFlowControlContext();

    /**
     * Returns the signature of the method declaration in which this
     * context is enclosed, or null if this context is not enclosed in
     * a method declaration.  */
    /*@ pure @*/ CMethod getCMethod();

    /**
     * Returns the signature of the nearest lexically enclosing
     * context that can host member declarations (i.e., a
     * <code>CClass</code> or a <code>CCompilationUnit</code>).  Used
     * for access checks, this is slightly different than
     * <code>getHostClass()</code>.  The later method returns an
     * anchor class signature for external methods, which is useful in
     * code generation and in typechecking external method bodies.  On
     * the other hand, this method returns a
     * <code>CCompilationUnit</code> for external methods, which is
     * useful in checking access between two differently named generic
     * functions declared in the same compilation unit.  !FIXME! It
     * may be possible to eliminate this method and just use
     * <code>getHostClass</code> now that private external methods are
     * duplciated in every anchor.
     *
     * @return	the nearest parent that can be coerced to a 
     * <code>CMemberHost</code> */
    /*@ pure @*/ CMemberHost findNearestHost();


    /**
     * Creates a class context with this context as its parent.
     * @param	self		the corresponding class
     */
    CClassContextType createClassContext( /*@non_null@*/ CClass self );

    /**
     * Creates an interface context with this context as its parent.
     * @param	self		the corresponding interface signature
     */
    CInterfaceContextType createInterfaceContext( /*@non_null@*/ CClass self );


    /**
     * Create a new child of this context representing the context in
     * which an external method declaration is typechecked. 
     * @param	host		the host class
     * @param	owner		the receiver class
     */
    CExtMethodContext createExtMethodContext( /*@non_null@*/ CSourceClass host,
					      /*@non_null@*/ CClass owner );
    
    // ----------------------------------------------------------------------
    // CLASS AND EXTERNAL GF HANDLING
    // ----------------------------------------------------------------------

    /**
     * Adds a source class to be generated
     *
     * @param clazz	the class to be generated
     */
    void classToGenerate(/*@non_null@*/ CSourceClass clazz) ;

    /**
     * Searches for any imported external generic functions.
     *
     * @param	ident	the name of the generic function to search for
     */
    void resolveMaybeExtMethodRef( /*@non_null@*/ String ident );

    /**
     * Registers the declaration of an external generic function in
     * this context. */
    void registerGFDecl(/*@non_null@*/ String methodIdent, CGenericFunctionCollection coll);
    
    /**
     * Registers with the surrounding context that a declaration of
     * the given method occurs in this context. */
    void registerVisibleMethod( /*@non_null@*/ CMethod method );

    /**
     * Registers with the surrounding context that a reference to
     * the given type occurs in this context. */
    void registerVisibleType( /*@non_null@*/ CType type );

    // ----------------------------------------------------------------------
    // ERROR HANDLING
    // ----------------------------------------------------------------------

    /**
     * Reports the given error and the "swallows" it, allowing
     * compilation to continue.  This method is used to report
     * non-fatal errors; for fatal errors the exception is just
     * thrown.
     *
     * @param trouble the error */
    //@ assignable \nothing;
    void reportTrouble(/*@\readonly@*/ Exception trouble);

    /**
     * Generates an UnpositionedError with a given message.
     * @param	mess		the error message
     * @param	param1		the first message parameter
     * @param	param2		the second message parameter
     * @return	never, i.e., this method always throws an exception
     * @exception	UnpositionedError	the error exception
     */
    //@ ensures false;
    void fail(MessageDescription mess, /*@\readonly@*/ Object param1, /*@\readonly@*/ Object param2)
	throws UnpositionedError;

    //@ ensures false;
    void fail(MessageDescription mess, /*@\readonly \readonly@*/ Object[] params)
	throws UnpositionedError;

    /**
     * Verifies an expression and if false signals an error.
     * @param	expr		condition to verify
     * @param	mess		the message to be displayed
     * @return	true iff expr is true
     * @exception	UnpositionedError	this error will be positioned soon
     */
    boolean check(boolean expr,
		  MessageDescription mess) throws UnpositionedError;

    boolean check(boolean expr,
		  MessageDescription mess,
		  /*@\readonly \readonly@*/ Object[] params) throws UnpositionedError;
    /**
     * Verifies an expression and if false signals an error.
     * @param	expr		condition to verify
     * @param	mess		the message to be displayed
     * @param	param1		the first parameter
     * @return	true iff expr is true
     * @exception	UnpositionedError	this error will be positioned soon
     */
    boolean check(boolean expr,
		  MessageDescription mess,
		  /*@\readonly@*/ Object param1) throws UnpositionedError;

    /**
     * Verifies an expression and if false signals an error.
     * @param	expr		condition to verify
     * @param	mess		the message to be displayed
     * @param	param1		the first parameter
     * @param	param2		the second parameter
     * @return	true iff expr is true
     * @exception	UnpositionedError	this error will be positioned soon
     */
    boolean check(boolean expr,
		  MessageDescription mess,
		  /*@\readonly@*/ Object param1,
		  /*@\readonly@*/ Object param2) throws UnpositionedError;


    // -------------------------------------------------------------
    // UTILITY METHODS
    // -------------------------------------------------------------

    /**
     * Gets the compiler for this context
     */
    /*@ pure non_null@*/ Main getCompiler();

    /**
     * Returns the modifier utility for this.  */
    /*@ pure non_null@*/ ModifierUtility modUtil();

    /**
     * Indicates whether this context is enclosed in a loop. */
    /*@ pure @*/ boolean isInLoop();

    /**
     * Indicates whether this context is enclosed in a loop and the
     * given variable is declared outside the inner-most loop
     * context. */
    boolean declaredOutsideOfLoop(JLocalVariable var);

    /**
     * Returns true if the given local variable is declared exactly in
     * this context, i.e., it is not declared in an outer context. */
    public boolean declares(JLocalVariable var);

    /**
     * Indicates whether this context is enclosed in an instance or
     * static initializer.  */
    /*@ pure @*/ boolean isInInitializer();

    /**
     * Indicates whether this context is enclosed in a constructor.  */
    /*@ pure @*/ boolean isInConstructor();

    /**
     * Indicates whether this context is enclosed in a constructor and occurs
     * before the constructor has invoked the superclass constructor.  */
    /*@ pure @*/ boolean isBeforeSuperConstructorCall();
    
    /**
     * Indicates whether this context is "static".
     *
     * @return true iff this context is enclosed in a context for a
     * static initializer or static method.  */
    /*@ pure @*/ boolean isStatic();


    /**
     * Indicates whether this context is "pure".
     *
     * @return true iff this context is enclosed in a pure method.
     */
    //@ assignable objectState;
    boolean isPure();


    /**
     * Calls back to the compiler for this context and requests that
     * the compiler catch-up decl to the same pass as the currently
     * active task.  */
    void catchUp(JTypeDeclarationType decl);
  
    /**
     * Indicates the integral arithmetic mode that should be used.
     */
    /*@non_null@*/ MJMathMode arithmeticMode();

    /**
     * Indicates whether expr (or member) is conditionally NonNull is this context.
     */
    /*@pure*/ boolean isFANonNull(/*@non_null*/ Object expr);
 
    /**
     * returns the JPhyla that are known to be non-null in this context.
     */
     /*@pure non_null*/ Object[] getFANonNulls();
     /*@pure non_null*/ Object[] getFANulls();
    /**
     * Mark expr (or member) as NonNull in this context
     */
    void addFANonNull(/*@non_null*/Object expr);
    void addFANull(/*@non_null*/Object expr);
    void removeFANonNull(/*@non_null*/Object expr);
    void removeAllFANullity();
    /**
     * adds exprs (or members) as NonNull in this context
     */
    void addFANonNulls(/*@non_null*/Object[] exprs);
    void addFANulls(/*@non_null*/Object[] exprs);
    /**
     * Merge the list of nulls in this context with that of the given context.
     * This is done by taking a union of the two sets. 
     */
    void mergeNullityInfo(/*@non_null*/CContextType other);
    void adoptNullityInfo(/*@non_null*/CContextType other);
    void dumpNonNulls(/*@non_null*/String msg);
}
