/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CExpressionContext.java,v 1.12 2006/08/08 17:18:43 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the context for an expression during checking
 * passes (checkInterface, checkInitializers, typecheck).
 * @see CContextType 
 */
public class CExpressionContext extends CContext 
    implements CExpressionContextType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement.
     * Clients should not call this but should instead call
     * <code>CFlowControlContextType.createExpressionContext</code>.
     *
     * <pre><jml>
     * requires (* \caller instanceof CContextType *);
     * </jml></pre>
     * 
     * @param	parent		the parent context, it must be different
     *				than null except if called by the top level
     *
     * @see CFlowControlContextType#createExpressionContext
     */
    public CExpressionContext(CFlowControlContextType parent) { //  !FIXME!12/13 should be package access once CJmlExpressionContext is a decorator
	super(parent);
	arithmeticMode = super.arithmeticMode();
    }
    public CExpressionContext(CExpressionContextType parent) { 
	super(parent);
	arithmeticMode = super.arithmeticMode();
    }


    // ----------------------------------------------------------------------
    // FACTORY METHODS
    // ----------------------------------------------------------------------

	/**
	 * Create a copy of this context to handle divergent paths, e.g., in 
	 * conditional expressions
	 */
	public CExpressionContextType createExpressionContext_() {
		return new CExpressionContext(this);
	}



    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * searches for a field with the given identifier
     * @return a field from an ident in current context
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupField(String ident) throws UnpositionedError {
	return lookupField(ident, this);
    }

   /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupOuterField(String ident)
	throws UnpositionedError {
	return lookupOuterField( ident, this );
    }

    /**
     * Set this context as an increment or decrement expression, so
     * we can generate appropriate getters and setters if necessary
     * (and also generate proper code in the future).
     * @param	incDec
     */
    public void setIncDec(boolean incDec) {
	this.incDec = incDec;
    }

    /**
     * @return	true if this context is part of an increment or decrement
     * 		expression
     */
    public boolean isIncDec() {
	return incDec;
    }


    /**
     * Set this context as a left side in an assignment, so access to vars
     * may be uninitialized
     * @param	leftSide		set the side of current code flow
     */
    public void setLeftSide(boolean leftSide) {
	this.leftSide = leftSide;
    }

    /**
     * @return	true if this context is on a left side of an assignement
     */
    public boolean isLeftSide() {
	return leftSide;
    }

    /**
     * set this context as a discarded value so result is ignored
     * @param	discardValue	is the value of this expression in 
     *				this context discarded?
     */
    public void setDiscardValue(boolean discardValue) {
	this.discardValue = discardValue;
    }

    /**
     * @return	true if the result value of this expression is discarded
     */
    public boolean discardValue() {
	return discardValue;
    }

    public MJMathMode arithmeticMode() {
	return arithmeticMode;
    }

    public void setArithmeticMode(byte mode) {
	arithmeticMode = MJMathMode.newMJMathMode(mode);
    }

    // ----------------------------------------------------------------------
    // PRIVATE UTILITIES
    // ----------------------------------------------------------------------

    protected	boolean		discardValue;
    protected	boolean		leftSide;
	// note that incDec ==> leftSide
    protected	boolean		incDec;
    protected	MJMathMode	arithmeticMode;
}
