/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CExtMethodContext.java,v 1.8 2004/03/06 10:03:10 tongjie Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This represents the context in which an external method declaration
 * is typechecked.  It subclasses <code>CClassContextType</code> and so
 * can be returned when a client calls <code>getClassContext()</code>
 * on a local context.  This context provides access to the external
 * method's receiver <Code>CClass</code> singleton. */
public class CExtMethodContext extends CClassContext {
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------
    /**
     * @param	parent		the parent context or null at top level
     * @param	host		the host class
     * @param	owner		the receiver class
     */
    public CExtMethodContext( CContextType parent, CSourceClass host, 
		       CClass owner ) {
	super( parent, host );

	this.owner = owner;
    }

    /**
     * Searches for a class with the given name.
     *
     * @param	name		method name
     * @return	the class type if found, null otherwise
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CClass lookupClass( String name ) throws UnpositionedError {
	// redirect search into logical owner before searching the host
	CClass clazz = owner.lookupClass( name, this );
	return (clazz != null) ? clazz : super.lookupClass(name);
    }

    /**
     * Searches for the most specific method <em>when no receiver is
     * explicit at the call site</em>.
     *
     * @param	ident		method name
     * @param	params		method parameters
     * @param	context		the con	text of the class containing
     * 				the method call
     * @return	the method or null if not found
     * @exception UnpositionedError this error will be positioned soon */
    public CMethod lookupMethod( String ident, CType[] params,
				 CClassType[] args,
				 CClassContextType context ) 
	throws UnpositionedError 
    {
	// redirect search into logical owner instead of searching the host
	return owner.lookupMethod( ident, params,args, context );
    }

    /**
     * Searches for the most specific method(s) <em>when no receiver is
     * explicit at the call site</em>.
     *
     * @param	ident		method name
     * @param	params		method parameters
     * @param	context		the con	text of the class containing
     * 				the method call
     * @return	the method or null if not found
     * @exception UnpositionedError this error will be positioned soon */
    public CMethodSet lookupMethodOrSet( String ident, CType[] params,
					 CClassContextType context ) 
	throws UnpositionedError 
    {
	// redirect search into logical owner instead of searching the host
	return owner.lookupMethodOrSet( ident, params, CClassType.EMPTY,context );
    }

    /**
     * Searches for a field with the given name in this context.
     *
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a field from the current context
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupField(String ident,
				      CExpressionContextType context)
	throws UnpositionedError {
	// redirect search into logical owner instead of searching the host
	return owner.lookupField( ident, context );
    }

    /**
     * Searches for a field of the given name in the context surrounding the
     * current lexical contour.
     *
     * @param	ident		the name of the field
     * @param	context		the context of the field access
     * @return	a variable from a field in the surrounding context, or null
     *		if none is found
     * @exception UnpositionedError	this error will be positioned soon
     */
    public CFieldAccessor lookupOuterField(String ident,
				   CExpressionContextType context)
	throws UnpositionedError {
	return null;
    }

    /**
     * Finds a local variable with the given name that appears outside
     * the current lexical contour.
     *
     * @param	ref		a token reference used to build a new
     *				<code>JOuterLocalVariableExpression</code>
     * @param	ident		the name of the outer variable
     * @return a variable from an ident in the surrounding context, or null
     *		if not found */
    public JExpression lookupOuterLocalVariable( TokenReference ref, 
						 String ident ) {
	return null;
    }

    /**
     * Returns the CClass representing the signature of the class
     * that is the logical owner of this context.  Useful for external
     * methods where the host is the generic function anchor but the
     * owner is the method's receiver.
     *
     * @return	the CClass representing the signature of the class
     * @see org.multijava.mjc.CClassContextType#getHostClass()
     */
    public CClass getOwnerClass() {
	return owner;
    }

    /**
     * Returns the signature of the nearest lexically enclosing
     * context that can host member declarations (i.e., a
     * <code>CClass</code> or a <code>CCompilationUnit</code>).  Used
     * for access checks, this is slightly different than
     * <code>getHostClass()</code>.  The later method returns an
     * anchor class signature for external methods, which is useful in
     * code generation and in typechecking external method bodies.  On
     * the other hand, this method returns a
     * <code>CCompilationUnit</code> for external methods, which is
     * useful in checking access between two differently named generic
     * functions declared in the same compilation unit.  !FIXME! It
     * may be possible to eliminate this method and just use
     * <code>getHostClass</code> now that private external methods are
     * duplciated in every anchor.
     *
     * @return the nearest parent that can be coerced to a
     * <code>CMemberHost</code> */
    public CMemberHost findNearestHost() {
	return getParentContext().findNearestHost();
    }

    // ----------------------------------------------------------------------
    // CLASS AND EXTERNAL GF HANDLING
    // ----------------------------------------------------------------------

    //---------------------------------------------------------------------
    // PRIVATE MEMBERS
    //---------------------------------------------------------------------

    /**
     * Represents the logical owner of this context, i.e., the
     * receiver of the method.*/
    private CClass owner;
}
