/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CField.java,v 1.15 2008/03/17 16:59:46 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.*;
import java.util.ArrayList;

/**
 * This class represents an exported member of a class (fields)
 */
public abstract class CField extends CMember implements CFieldAccessor {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a field export
     * @param	access		contains the info used to determine accessibility of this field
     * @param	ident		the name of this field
     * @param	type		the type of this field
     * @param	deprecated	is this field deprecated ?
     */
    public CField( MemberAccess access, String ident, CType type, 
		   boolean deprecated ) {
	super( access, ident, deprecated );
	this.type = type;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * @return	the interface
     */
    public /*@ pure @*/ CField getField() {
	return this;
    }

    /**
     * Return the type of this field.
     * This method used to be pure, but WMD changed it in order to check whether
     * the field type was already queried. Now we have only the assignable clause.
     * 
     * @return the type of this field
     */
    //@ also assignable wasTypeAlreadyAsked;
    public CType getType() {
    	wasTypeAlreadyAsked = true;
	return type;
    }

    // WMD wasTypeAlreadyAsked is used to determine whether the Universe modifier
    // of a field can still be changed safely. If the type was not used to determine
    // the type of an other field, we can still propagate the field type from
    // the initializer expression. Otherwise, we issue a warning.
    //@ spec_public
    private boolean wasTypeAlreadyAsked = false;
    public boolean wasTypeAlreadyAsked() {
    	return wasTypeAlreadyAsked;
    }
    
    public void setType(CType type) {
	this.type = type;
    }

    /**
     * @param	value		the value known at third pass
     */
    public void setValue(JExpression value) {
	this.value = value;
    }

    /**
     * @return	the value of initializer or null
     */
    public JExpression getValue() {
        //!FIXME! I commented this out as I/O like this is technically illegal in a pure method -- Gary
	//!!!TEST
	// {
	//     if (value != null && ! value.getType().equals(type)) {
	// 	System.err.println("<<< get " + getOwnerName() + "." + ident() + ": bad value type " + value.getType() + "(" + type + " expected)");
	//     }
	// }
	//!!!TEST
	return value;
    }

    /** This method does the same thing as isStatic().  It is used by
        the field getter and setter classes to distinguish between
        checking that method and checking its associated field for
        being static. */
    public /*@ pure @*/ boolean isFieldStatic() {
	return isStatic();
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Generates a sequence of bytecodes to load
     * @param	code		the code list
     */
    public void genLoad(CodeSequence code) {
	code.plantInstruction(new FieldRefInstruction(isStatic() ? opc_getstatic : opc_getfield,
						      getOwnerName(),
						      ident(),
						      getType().getSignature()));
    }

    /**
     * Generates a sequence of bytecodes to load
     * @param	code		the code list
     */
    public void genStore(CodeSequence code) {
	code.plantInstruction(new FieldRefInstruction(isStatic() ? opc_putstatic : opc_putfield,
						      getOwnerName(),
						      ident(),
						      getType().getSignature()));
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Generate the code in a class file
     */
    public FieldInfo genFieldInfo() {
	Object		value = null;

	    // remove the privacy modifier if necessary
//  	if ( needsPrivacyModifierRemoved )
//  	    this.modifiers = access().removePrivateModifier();
	
	if (this.value != null &&
	    this.value.isConstant() &&
	    isFinal() &&
	    isStatic() &&
	    !(this.value instanceof JNullLiteral)) {
	    CType t = this.value.getType();
	    value = this.value.getLiteral().getValue();
	    if (t.isOrdinal()) {
		if (t == CStdType.Long)
		    value = new Long(this.value.getOrdinalLiteral()
				     .numberValue().longValue());
		    // treat ints, shorts, bytes, chars all like ints
		    // hopefully, narrowing conversions will do the right thing
		    // we have to treat Long specially, or else there is
		    // an error with widening conversions. (tdm)
		    // kopi 1.5b actually separates out the various kinds
		    // of ordinals into distinct subclasses.
		else
		    value = new Integer(this.value.getOrdinalLiteral()
					.numberValue().intValue());
		
	    } else if (this.value.getType() == CStdType.Float) {
		value = new Float(((Number)value).floatValue());
	    } else if (this.value.getType() == CStdType.Double) {
		value = new Double(((Number)value).doubleValue());
	    }
	}
	
	FieldInfo fi = createFieldInfo(access().modifiers(),
                               ident(),
                               type.getSignature(),
			       type.getGenericSignature(),
			       value,
                               isDeprecated(),
                               ident().indexOf('$') >= 0);
	
	// AHS: add Universe annotation to classfile, either through a field
	//      attribute or a Java 5 annotation
	if (type.isClassType() ) {
	    byte univConst = ((CClassType)type).getUniverseByteConstant();

            if (CTopLevel.getCompiler().universeBytecode()) {
                fi.addAttribute(new UniverseFieldAttribute(
                          ((CClassType)type).getUniverseByteConstant()));	        
	    } else if (CTopLevel.getCompiler().universeBytecodeAnnotations()) {
                Annotation typeAnnot = new Annotation("org.multijava.universes.UniverseType");
	        typeAnnot.addElement("value", new AnnotByteElementValue(univConst));
                
	        ArrayList annots = new ArrayList();
	        annots.add(typeAnnot);
                
	        fi.addAttribute(new RuntimeVisibleAnnotationsAttribute(annots));
	    }
	}
	
	return fi;
    }

    
    /**
     * Creates a new field info object. This is a factory method to
     * create a subclass-specific field info object, e.g., a refined
     * field info object for JML.
     */
    protected FieldInfo createFieldInfo(long modifiers,
                                        String name,
                                        String type,
					String generictype,
                                        Object value,
                                        boolean deprecated,
                                        boolean synthetic) {
        return new FieldInfo((short) modifiers, name, type, generictype, value,
                             deprecated, synthetic);
    }
                                        
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	CType		type;
    private	JExpression	value;

    /**
     * NO LONGER USED -- replaced by CFieldGetterMethod and CFieldSetterMethod.
     * Indicates whether this field should have its privacy modifier removed
     * for the purposes of code generation.  This is the case if the field
     * is private but also accessed by an enclosing or nested class.
     * */
//    public boolean needsPrivacyModifierRemoved = false;

}
