/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.mjc;

import org.multijava.util.classfile.LocalVarInstruction;
import org.multijava.util.classfile.NoArgInstruction;

/**
 * This class represents a getter method for a field that an inner
 * or enclosing class has privileged access to.  The getter method is
 * necessary to work around the Java verifier, which doesn't know that
 * the inner/enclosing class has privileged access.  */

public class CFieldGetterMethod extends CFieldAccessorMethod
implements CFieldAccessor {
    
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Constructs a getter method for the given method.
     * @param	target	the field
     * @param	owner	the receiver of the getter method
     */
    public CFieldGetterMethod( CField target, CClass owner ) {
	super( target, owner, target.getType(), createParms(target, owner) );
    }

    
  	// create the formal parameters for the getter method
      private static CSpecializedType[] createParms( CField target,
						       CClass owner ) {
	  if (target.isStatic())
	      return new CSpecializedType[0];
	  else {
	      CSpecializedType[] parms = new CSpecializedType[1];
	      parms[0] = new CSpecializedType(owner.getType());
	      return parms;
	  }
      }

    
    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Plants the bytecode for the body of this method.  Separated
     * from <code>genCode()</code> so subclasses may replace standard
     * body bytecode with specialized code.  
     *
     * @see #genCode()
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	if ( !target.isStatic() ) {
		// put the field's receiver on the stack
	    code.plantInstruction(new LocalVarInstruction(
		parameters()[0].staticType().getLoadOpcode(), 0 )
		);
	}
	    // get the field's value
	target.genLoad( code );
	    // return
	code.plantInstruction(
	    new NoArgInstruction( returnType().getReturnOpcode() ) );
    }

	// Substitute a load of the field by a call to this getter method. 
    public void genLoad(CodeSequence code) {
	genGenFuncInvocation(code, false);
    }

    	// A getter method doesn't know how to store into its field.
    public void genStore(CodeSequence code) {
	fail("method genStore not defined for CFieldGetterMethod");
	    //@ unreachable;
	return;
    }

    

}
