/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.mjc;

import org.multijava.util.classfile.LocalVarInstruction;
import org.multijava.util.classfile.NoArgInstruction;

/**
 * This class represents a setter method for a field that an inner
 * or enclosing class has privileged access to.  The setter method is
 * necessary to work around the Java verifier, which doesn't know that
 * the inner/enclosing class has privileged access.  */

public class CFieldSetterMethod extends CFieldAccessorMethod
implements CFieldAccessor {
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Constructs a setter method for the given method.
     * @param	target	the field
     * @param	owner	the receiver of the setter method
     */
    public CFieldSetterMethod( CField target, CClass owner ) {
	super( target, owner, CStdType.Void, createParms(target, owner) );
    }

    
  	// create the formal parameters for the setter method
      private static CSpecializedType[] createParms( CField target,
						       CClass owner ) {
	  CSpecializedType[] parms;
	  int i = 0;
	  if (target.isStatic())
	      parms = new CSpecializedType[1];
	  else {
	      parms = new CSpecializedType[2];
	      parms[0] =  new CSpecializedType(owner.getType());
	      i++;
	  }
	  parms[i] = new CSpecializedType(target.getType());
	  return parms;
      }
    
    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Plants the bytecode for the body of this method.  Separated
     * from <code>genCode()</code> so subclasses may replace standard
     * body bytecode with specialized code.  
     *
     * @see #genCode()
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	    // put the args on the stack
	for (int i = 0; i < parameters().length; i++) {
	    code.plantInstruction(new LocalVarInstruction(
		parameters()[i].staticType().getLoadOpcode(), i )
		);
	}
	    // update the field's value
	target.genStore( code );
	    // return
	code.plantInstruction(
	    new NoArgInstruction( returnType().getReturnOpcode() ) );
    }
    

	/* Generate a load of the field.  We ask the current class for
	   an appropriate getter method, and then invoke it. */
    public void genLoad(CodeSequence code) {
	CFieldAccessorMethod g = owner().findGetterMethod( target );
	if (g != null)
	    g.genGenFuncInvocation(code, false);
	else
	    fail("CFieldSetterMethod::genLoad could not find an appropriate getter method");
    }

	// A store of the field is replaced by a call to this setter method.
    public void genStore(CodeSequence code) {
	genGenFuncInvocation(code, false);
    }

}
