/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CFieldTable.java,v 1.9 2005/03/27 14:50:24 leavens Exp $
 */

package org.multijava.mjc;


import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Collection;

import org.multijava.util.classfile.FieldInfo;

/**
 * This class encapsulates the fields of a class.
 */
public class CFieldTable extends org.multijava.util.Utils {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a field table using <em>one of</em> array or hashed.  The
     * other parameter should be null.
     *
     */
    public CFieldTable(CField[] array, Hashtable hashed) {
	assertTrue((array != null) ^ (hashed != null));
	if( hashed == null ) {
	    this.hashed = buildHash( array );
	} else {
	    this.hashed = hashed;
	    expandable = true;
	}
    }

    /**
     * Utility methods builds a hash table from the array of fields.
     *
     * <pre><jml>
         requires array != null && 
	          (\forall int i; 0<=i && i < array.length; 
		           array[i] instanceof CSourceField );
         assignable expandable;
		 ensures expandable;
       also
         requires array != null && 
	          (\exists int i; 0<=i && i < array.length; 
		           !(array[i] instanceof CSourceField) );
		 ensures !expandable;
     * </jml></pre>
     */
    private Hashtable buildHash( CField[] array ) {
	expandable = true;
	Hashtable result = new Hashtable();
	for (int i = 0; i < array.length; i++) {
	    CField f = array[i];
	    if( f instanceof CSourceField ) {
		((CSourceField) f).setIndex( i );
	    } else {
		expandable = false;
	    }
	    result.put( array[i].ident(), array[i] );
	}
	return result;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the field with specified name
     * @param	ident		the simple name of the field
     */
    public /*@ pure @*/ CField lookup(String ident) {
	return (CField) hashed.get(ident);
    }

    /**
     * Returns the number of fields.
     */
    public int size() {
	return hashed.size();
    }

    /**
     * Returns a Collection of CField objects contained in this CFieldTable.
     */
    public Collection fields() {
	return hashed.values();
    }

    /**
     * Adds a field.
     */
    public void add( CSourceField field ) {
	assertTrue( expandable );
	// !FIXME!function this is a hack to suppress errors on
	// multiple generic functions in a single collection
	String id = field.ident();
	if( hashed.containsKey( id ) ) {
	    id = id + uniquifier;
	    uniquifier++;
	}
	field.setIndex( hashed.size() );
	hashed.put( id, field );
    }

    /**
     * Returns the field infos
     */
    public FieldInfo[] buildFieldInfo() {
	try {
	assertTrue( expandable );
	FieldInfo[] fields = new FieldInfo[hashed.size()];
	    
	for( Enumeration elems = hashed.elements(); 
	     elems.hasMoreElements(); ) {
	    CSourceField field = (CSourceField)elems.nextElement();
	    fields[field.index()] = field.genFieldInfo();
	}
	    
	return fields;
	} catch( ArrayIndexOutOfBoundsException e ) {
	    throw new RuntimeException( "out of bounds: " + 
					hashed.toString() + ", hashed.size()="
					+ hashed.size() );
	}
    }


    /**
     * Returns an array of the CSourceFields stored in this field
     * table.  The array returned is empty if this field table is not
     * expandable.  */
    public CSourceField[] sourceFields() {
	if (!expandable) {
	    return new CSourceField[0];
	}
	CSourceField[] results = new CSourceField[ hashed.size() ];
	hashed.values().toArray(results);
	return results;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private Hashtable hashed;

    /**
     * Indicates whether additional fields can be added to this field
     * table.  */
    private boolean expandable;

    private static int uniquifier = 0; // !FIXME!function
}
