/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CFlowControlContext.java,v 1.21 2006/08/08 17:18:43 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.MessageDescription;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

/**
 * <P>This class is used during typechecking for control flow analysis that
 * maintains local variable definite assignment (JLS2, 16), throwable, and
 * reachability information (JLS2, 14.20).</P>
 */
public class CFlowControlContext extends CContext
		implements CFlowControlContextType {

	// ----------------------------------------------------------------------
	// CONSTRUCTORS
	// ----------------------------------------------------------------------

	/**
	 * Construct a nested flow control context. Clients should not call this but
	 * should instead call <code>CCompilationUnitContextType.createFlowControlContext</code>.
	 * <p/>
	 * <pre><jml>
	 * requires (* \caller instanceof CContextType *);
	 * ensures nestedFlowControlContext;
	 * </jml></pre>
	 *
	 * @param where a source-code location for the start of this context
	 *
	 * @param	parent		the parent context, it must be different than null except if
	 * called by the top level
	 * @see #createFlowControlContext(org.multijava.util.compiler.TokenReference)
	 */
	public CFlowControlContext(CFlowControlContextType parent,
							   TokenReference where) {
		this(parent, DEFAULT_VAR_ESTIMATE, where);
	}

	/**
	 * Construct a nested flow control context. Clients should not call this but
	 * should instead call <code>createFlowControlContext(int)</code>.
	 * <p/>
	 * <pre><jml>
	 * requires (* \caller instanceof CContextType *);
	 * ensures nestedFlowControlContext;
	 * </jml></pre>
	 *
	 * @param parent      the parent context, it must be different than null
	 * @param varEstimate an estimate of the number of variable slots to reserve
	 * @param where       a source-code location for the start of this context
	 *
	 * @see #createFlowControlContext(int, org.multijava.util.compiler.TokenReference)
	 */
	public CFlowControlContext(CFlowControlContextType parent, int varEstimate,
							   TokenReference where) {
		super(parent);

		cachedParent = parent;
		methodContext = null;

		isReachable = parent.isReachable();
		variableInfo = parent.variableInfo().duplicate();
		fieldInfo = parent.fieldInfo().duplicate();

		this.parentIndex = parent.numberOfLocalVars();
		this.localsPosition = parent.localsPosition();
		initializeLocalVars(varEstimate);

		this.where = where;
	}

	/**
	 * Construct an outer-most flow control context. Clients should not call this
	 * but should instead call <code>CMethodContextType.createFlowControlContext(int)</code>.
	 * <p/>
	 * <pre><jml>
	 * requires (* \caller instanceof CContextType *);
	 * ensures !nestedFlowControlContext;
	 * </jml></pre>
	 *
	 * @param parent         the parent context, it must be different than null
	 * @param varEstimate    an estimate of the number of variable slots to
	 *                       reserve
	 * @param isInExternalGF true if this context is in a method that is part of an
	 *                       external generic function
	 *
	 * @see CMethodContextType#createFlowControlContext(int, boolean,
			*      org.multijava.util.compiler.TokenReference)
	 */
	public CFlowControlContext(CMethodContextType parent, int varEstimate,
							   boolean isInExternalGF, TokenReference where) {
		super(parent);

		cachedParent = null;
		methodContext = parent;

		variableInfo = new CVariableInfoTable();
		try {
			fieldInfo = parent.getClassContext().fieldInfo().duplicate();
		} catch (NullPointerException e) {
			fieldInfo = new CVariableInfoTable();
		}

		// if we are in an external gf, start at position 1 instead of
		// zero, so we can later stick the receiver as the first
		// argument
		this.parentIndex = 0;
		if (isInExternalGF) {
			this.localsPosition = 1;
			varEstimate++;
		} else {
			this.localsPosition = 0;
		}
		initializeLocalVars(varEstimate);

		this.where = where;
	}

	public CFlowControlContext(CMethodContextType parent, int varEstimate,
							   TokenReference where) {
		this(parent, varEstimate, false, where);
	}

	public String toString() {
		return "Context is " + (isReachable() ? "" : "not ") + "reachable. \n"
			   + "          Locals: " + variableInfo.toString() + "\n" +
			   "          Fields: " + fieldInfo.toString();
	}

	/**
	 * Initializes local vars based on the value of varEstimate.
	 */
	private void initializeLocalVars(int varEstimate) {
		localVars = varEstimate == 0 ? null : new ArrayList(varEstimate);
	}

	/**
	 * Used to clone this flow control context in subclasses.
	 * <p/>
	 * <pre><jml>
	 * requires (* \caller instanceof CContextType *);
	 * ensures nestedFlowControlContext;
	 * </jml></pre>
	 *
	 * @param clone the context whose state is to be adopted
	 *
	 * @param	parent		the parent context, it must be different than null
	 */
	public CFlowControlContext(CFlowControlContextType parent,
							   CFlowControlContextType clone) {
		super(parent);
		cachedParent = parent;
		methodContext = null;

		isReachable = clone.isReachable();

		variableInfo = clone.variableInfo().duplicate();
		fieldInfo = clone.fieldInfo().duplicate();

		parentIndex = clone.parentIndex();
		localsPosition = clone.localsPosition();

		ArrayList clonesVars = clone.localVars();
		localVars = clonesVars == null ? null : (ArrayList) clonesVars.clone();
		localsIndex = clone.localsIndex();

		this.where = clone.getTokenReference();
	}

	// -------------------------------------------------------------
	// "DESTRUCTOR"
	// -------------------------------------------------------------

	/**
	 * <P>Registers that this context is no longer needed.  Passes information
	 * collected by this context to the parent context.</P>
	 * <p/>
	 * <p>Passes this context's field information to the surrounding context;
	 * passes local variable information to the surrounding context if that context
	 * is one which maintains local variable information.</p>
	 * <p/>
	 * <P>Checks that local variables are used and issues warnings if not.</P>
	 * <p/>
	 * <P>This should only be called on a linearly nested context! !FIXME!10/26 If
	 * we refactor the context creation code as planned (to lock the parent
	 * context) then this method is called to unlock the parent context when we
	 * exit a linearly nested context.</P>
	 */
	public void checkingComplete() {
		checkLocalVarUsage();

		// pass reachability information and definite assignment
		// information for non-local declarations to the surrounding
		// context
		if (cachedParent != null) {
			//@ assert nestedFlowControlContext;
			cachedParent.replaceVariableInfoUpTo(parentIndex, variableInfo);
			cachedParent.setReachable(isReachable);
		}

		// pass definite assignment information for fields to the
		// surrounding context
		parent.replaceFieldInfoUpTo(fieldCount(), fieldInfo);
	}

	/**
	 * Checks that all declare local variables have been used now that they
	 * are going out of scope.
	 */
	protected void checkLocalVarUsage() {
		if (localVars != null) {
			Iterator iter = localVars.iterator();
			while (iter.hasNext()) {
				JLocalVariable var = (JLocalVariable) iter.next();

				if (!var.isUsed() && !var.ident().startsWith("_")) {
					MessageDescription mesg = null;

					switch (var.getDescription()) {
						case JLocalVariable.DES_PARAMETER:
							mesg = MjcMessages.UNUSED_PARAMETER;
							break;
						case JLocalVariable.DES_CATCH_PARAMETER:
							mesg = MjcMessages.UNUSED_CATCH_PARAMETER;
							break;
						case JLocalVariable.DES_LOCAL_VAR:
							mesg = MjcMessages.UNUSED_LOCALVAR;
							break;
						default:
							continue;
					}
					reportTrouble(new CWarning(var.getTokenReference(),
											   mesg, var.ident()));
				}
			}
		}
	}

	public TokenReference getTokenReference() {
		return where;
	}

	// ----------------------------------------------------------------------
	// ACCESSORS (NEW CONTEXT)
	// ----------------------------------------------------------------------

	// !FIXME!10/26 can we somehow lock this context until the created
	// context or contexts are closed?

	public CFlowControlContextType createFlowControlContext(TokenReference where) {
		return new CFlowControlContext(this, where);
	}

	public CFlowControlContextType createFlowControlContext(int params,
															TokenReference where) {
		return new CFlowControlContext(this, params, where);
	}

	public CLabeledContext createLabeledContext(JLabeledStatement self) {
		return new CLabeledContext(this, self);
	}

	public CLoopContext createLoopContext(JLoopStatement self) {
		return new CLoopContext(this, self);
	}

	public CSwitchBodyContext createSwitchBodyContext(JSwitchStatement stmt,
													  CType switchType) {
		return new CSwitchBodyContext(this, stmt, switchType);
	}

	public CTryContext createTryContext(TokenReference where) {
		return new CTryContext(this, where);
	}

	// don't create a CTryContext for a try-finally, since a
	// try-finally doesn't actually catch any exceptions
	public CFlowControlContextType createFinallyContext(
			CFlowControlContextType tryContext, TokenReference where) {
		return new CFinallyContext(this, tryContext, where);
	}

	public CExpressionContextType createExpressionContext() {
		return new CExpressionContext(this);
	}

	/**
	 * Create a clone of this context to handle divergent paths in the control
	 * flow.
	 */
	public CFlowControlContextType cloneContext() {
		assertTrue(cachedParent != null);
		CFlowControlContext c = new CFlowControlContext(cachedParent, this);
		c.innerClasses = this.innerClasses;
		return c;
	}


	/**
	 * Creates a set of child contexts for typechecking parallel control flows.
	 * After using the contexts they should be passed to the
	 * <code>adoptParallelContexts</code> method of this.
	 * <p/>
	 * <pre><jml>
	 * also
	 * requires count > 0;
	 * </jml></pre>
	 *
	 * @see #adoptParallelContexts(CFlowControlContextType[])
	 */
	public CFlowControlContextType[] createParallelContexts(int count,
															TokenReference where) {
		//@ assert count > 0;
		CFlowControlContextType[] result = new CFlowControlContext[count];
		for (int i = 0; i < count; i++) {
			result[i] = new CFlowControlContext(this, where);
		}
		return result;
	}

	/**
	 * Adopts the information from the given contexts.
	 * <p/>
	 * <pre><jml>
	 * also
	 * requires contexts != null && contexts.length > 0;
	 * </jml></pre>
	 */
	public void adoptParallelContexts(CFlowControlContextType[] contexts) {
		CFlowControlContextType continuationContext = null;
		for (int i = 1; i < contexts.length; i++) {
			CFlowControlContextType ctxt = contexts[i];
			if (ctxt.isReachable()) {
				if (continuationContext == null) {
					continuationContext = ctxt;
				} else {
					continuationContext.merge(ctxt);
				}
			}
		}

		if (continuationContext == null) {
			setReachable(false);
		} else {
			adopt(continuationContext);
		}
	}

	// -------------------------------------------------------------
	// COMBINING CONTEXTS
	// -------------------------------------------------------------

	public final void merge(CFlowControlContextType context) {

		isReachable |= context.isReachable();

		int varCount = numberOfLocalVars();
		variableInfo.mergeUpTo(varCount, context.variableInfo());

		fieldInfo.mergeUpTo(fieldCount(), context.fieldInfo());
		mergeNullityInfo(context);
	}

	public final void adopt(CFlowControlContextType context) {
		isReachable = context.isReachable();

		int varCount = numberOfLocalVars();
		variableInfo.replaceUpTo(varCount, context.variableInfo());

		fieldInfo.replaceUpTo(fieldCount(), context.fieldInfo());
		adoptNullityInfo(context);
	}

	// ----------------------------------------------------------------------
	// ACCESSORS (VARIABLE)
	// ----------------------------------------------------------------------

	/**
	 * Checks whether a local variable has a name that is fresh (i.e., no other
	 * local variable in scope has the same name).
	 *
	 * @param	var		the variable
	 */
	public boolean isFreshVariableName(JLocalVariable var) {
		if (localVars == null)
			return true;

		for (int i = 0; i < localVars.size(); i++) {
			if (((JLocalVariable) localVars.get(i))
					.ident().equals(var.ident())) {
				return false;
			}
		}

		// if cachedParent is not set then this is the outermost
		// context and we're done, otherwise this is a nested context
		// and we check the surrounding context
		return cachedParent == null ? true :
			   cachedParent.isFreshVariableName(var);
	}

	/**
	 * Adds a local variable to this block
	 *
	 * @throws UnpositionedError if the variable is a redeclaration
	 * @param	var		the name of the variable
	 */
	public void addVariable(JLocalVariable var) throws UnpositionedError {
		if (localVars == null) {
			localVars = new ArrayList();
		}

		if (!isFreshVariableName(var)) {
			throw new UnpositionedError(MjcMessages.VAR_RENAME, var.ident());
		}

		var.setPosition(localsPosition);
		var.setIndex(localsIndex + parentIndex);

		localVars.add(var);
		localsIndex++;
		assertTrue(localsIndex == localVars.size());
		localsPosition += var.getType().getSize();
	}

	/**
	 * Records that the Java keyword <CODE>this</CODE> is used in this block.
	 */
	public void addThisVariable() {
		localsPosition += 1;
	}

	/**
	 * Records that a variable slot must be reserved for a synthetic
	 * <code>this</code> parameter.  Recall that <code>this</code> is passed as a
	 * parameter for external methods.
	 */
	public void addSyntheticThisParameter() {
		localsPosition += 1;
	}

	/**
	 * Returns the number of stack positions required to store the local variables
	 * encountered thus far in this context.
	 */
	public int localsPosition() {
		return localsPosition;
	}

	public int numberOfLocalVars() {
		return parentIndex + localsIndex;
	}

	public CVariableInfoTable variableInfo() {
		return variableInfo;
	}

	public CVariableInfoTable fieldInfo() {
		return fieldInfo;
	}

	public ArrayList localVars() {
		return localVars;
	}

	public int parentIndex() {
		return parentIndex;
	}

	public int localsIndex() {
		return localsIndex;
	}

	/**
	 * For each variable (and field), if it is definitely assigned in this context,
	 * is not definitely assigned in the parent context, and is final, then report
	 * error.  This method should only be called by subclasses that wish to verify
	 * that a final slot is not assigned to within the local context.
	 */
	protected void checkAssignmentToFinals() {
		if (cachedParent != null) {
			// Checks all variables that existed in the parent context.
			// Does not need to check local variables.
			for (int i = 0; i < parentIndex; i++) {
				if (isVarDefinitelyAssigned(i) &&
					isVarFinal(i) &&
					!cachedParent.isVarDefinitelyAssigned(i)) {
					reportTrouble(new PositionedError(where,
													  MjcMessages.LOOP_ASSIGN_FINAL_VARIABLE,
													  varIdent(i)));
				}
			}
			// Checks all fields
			int fieldCount = fieldCount();
			for (int i = 0; i < fieldCount; i++) {
				if (isFieldDefinitelyAssigned(i) &&
					isFieldFinal(i) &&
					!cachedParent.isFieldDefinitelyAssigned(i)) {
					reportTrouble(new PositionedError(where,
													  MjcMessages.LOOP_ASSIGN_FINAL_FIELD,
													  fieldIdent(i)));
				}
			}
		}
	}

	// -------------------------------------------------------------
	// ACCESSORS (FLOW CONTROL)
	// -------------------------------------------------------------

	/**
	 * Mutates the reachability status of this context.
	 *
	 * @param reachable indicates whether the statements in this context are
	 *                  reachable after the current statement being checked.
	 */
	public final void setReachable(boolean reachable) {
		isReachable = reachable;
	}

	/**
	 * Indicates whether the statements in this context are reachable up to most
	 * recent statement that was typechecked.  (Assuming that statement's
	 * typechecking code calls <code>setReachable</code> appropriately.)
	 *
	 * @return if this context is still reachable
	 */
	public final /*@ pure @*/ boolean isReachable() {
		return isReachable;
	}

	/**
	 * Returns the statement with the specified label.
	 */
	public JLabeledStatement getLabeledStatement(String label) {
		return cachedParent == null ?
			   null : cachedParent.getLabeledStatement(label);
	}

	/**
	 * Returns the nearest breakable statement.
	 */
	public JStatement getNearestBreakableStatement() {
		return cachedParent == null ?
			   null : cachedParent.getNearestBreakableStatement();
	}

	/**
	 * Returns the nearest continuable statement.
	 */
	public JStatement getNearestContinuableStatement() {
		return cachedParent == null ?
			   null : cachedParent.getNearestContinuableStatement();
	}

	/**
	 * Returns the nearest control flow context (where control flow information is
	 * stored.)
	 *
	 * @return the nearest parent of type CFlowControlContextType
	 */
	public CFlowControlContextType getFlowControlContext() {
		return this;
	}

	// ----------------------------------------------------------------------
	// ACCESSORS (FIELD AND VARIABLE STATE)
	// ----------------------------------------------------------------------

	/**
	 * Marks the variable with the given descriptor as definitely assigned to in
	 * this context.
	 */
	public void initializeVariable(VariableDescriptor varDesc) {
		variableInfo.initialize(varDesc);
	}

	/**
	 * Indicates whether the variable in the given position is definitely assigned
	 * to in this context.
	 */
	public boolean isVarDefinitelyAssigned(int pos) {
		return varAt(pos).isDefinitelyAssigned();
	}

	public boolean isVarDefinitelyUnassigned(int i) {
		return varAt(i).isDefinitelyUnassigned();
	}

	public boolean isVarFinal(int pos) {
		return varAt(pos).isFinal();
	}

	public String varIdent(int pos) {
		return varAt(pos).ident();
	}

	private CVariableState varAt(int pos) {
		return variableInfo.getState(pos);
	}

	/**
	 * Replaces the local field info for fields in positions 0 up to
	 * <code>pos</code> with the info in <code>replacement</code>.
	 */
	public void replaceFieldInfoUpTo(int pos,
									 CVariableInfoTable replacement) {
		fieldInfo.replaceUpTo(pos, replacement);
	}

	/**
	 * Replaces the local variable info in positions 0 up to <code>pos</code> with
	 * the info in <code>replacement</code>.
	 */
	public void replaceVariableInfoUpTo(int pos,
										CVariableInfoTable replacement) {
		variableInfo.replaceUpTo(pos, replacement);
	}

	/**
	 * Marks the field with the given descriptor as definitely assigned to in this
	 * context.
	 */
	public void initializeField(VariableDescriptor varDesc) {
		fieldInfo.initialize(varDesc);
	}

	/**
	 * Indicates whether the field in the given position is definitely assigned to
	 * in this context.
	 */
	public boolean isFieldDefinitelyAssigned(int pos) {
		return fieldAt(pos).isDefinitelyAssigned();
	}

	public boolean isFieldFinal(int pos) {
		return fieldAt(pos).isFinal();
	}

	public String fieldIdent(int pos) {
		return fieldAt(pos).ident();
	}

	private CVariableState fieldAt(int pos) {
		return fieldInfo.getState(pos);
	}

	/**
	 * Returns the number of fields in the enclosing class.
	 */
	protected int fieldCount() {
		return parent.getClassContext().getHostClass().getFieldCount();
	}

	// ----------------------------------------------------------------------
	// ACCESSORS (TYPE DEFINITION)
	// ----------------------------------------------------------------------

	/**
	 * Adds to this context a class declared via a type declaration statement.
	 *
	 * @throws UnpositionedError if duplicate class exists in this lexical context
	 * @param	clazz		the clazz to add
	 */
	public void addLocalClass(CClass clazz) throws UnpositionedError {
		if (innerClasses == null) {
			innerClasses = new HashMap();
		}
		Object old = innerClasses.put(clazz.ident(), clazz);
		if (old != null) {
			throw new UnpositionedError(MjcMessages.CLAZZ_RENAME,
										clazz.ident());
		}
	}

	// ----------------------------------------------------------------------
	// ACCESSORS (LOOKUP)
	// ----------------------------------------------------------------------

	/**
	 * Searches for a class with the given simple name according the procedure in
	 * JLS2 6.5.5.  This method checks the first bullet point of that section: "If
	 * the simple type name occurs within the scope of a visible local class
	 * declaration with that name, then the simple type name denotes that local
	 * class type."  If no match is found, then this method calls to the
	 * surrounding context to check the remaining bullet points.
	 *
	 * @throws UnpositionedError if search fails
	 * @param	name	the class name, without qualifiers
	 * @return	the class if found, null otherwise
	 */
	public CClass lookupClass(String name) throws UnpositionedError {
		CClass clazz = lookupLocalClass(name);

		return (clazz != null) ? clazz : super.lookupClass(name);
	}

	/**
	 * Looks for a class with the given name declared via a type declaration
	 * statement in this context.
	 *
	 * @param	ident		the name of the class
	 * @return	a class from an ident in current context, or null if none
	 */
	private CClass lookupLocalClass(String ident) {
		return innerClasses == null ? null : (CClass) innerClasses.get(ident);
	}

	/**
	 * Returns the variable referred to by the given name in this context,
	 * recursing to surrounding contexts as appropriate.  If no variable of the
	 * given name is found then <code>null</code> is returned.
	 *
	 * @param	ident		the name of the variable
	 */
	public JLocalVariable lookupLocalVariable(String ident) {
		//  !FIXME!10/26 should be able to use a map instead
		if (localVars != null) {
			for (int i = 0; i < localsIndex; i++) {
				JLocalVariable var = (JLocalVariable) localVars.get(i);

				if (var.ident() == ident) {
					return var;
				}
			}
		}

		return super.lookupLocalVariable(ident);
	}

	/**
	 * Indicates whether this context is enclosed in a loop and the given variable
	 * is declared outside the inner-most loop context.
	 */
	public boolean declaredOutsideOfLoop(JLocalVariable var) {
		// Recursion terminates at the first loop context encountered,
		// so if the recursion reaches this context there was no inner
		// loop context.  If this context declares the var then return
		// false, otherwise recurse to outer context.
		if (declares(var)) {
			return false;
		} else {
			return super.declaredOutsideOfLoop(var);
		}
	}

	/**
	 * Returns true if the given local variable is declared exactly in this
	 * context, i.e., it is not declared in an outer context.
	 */
	public boolean declares(JLocalVariable var) {
		if (localVars != null) {
			for (int i = 0; i < localsIndex; i++) {
				if (localVars.get(i) == var) {
					return true;
				}
			}
		}
		return false;
	}

	// ----------------------------------------------------------------------
	// THROWABLES
	// ----------------------------------------------------------------------

	/**
	 * Registers that the given throwable can be thrown within this context.  The
	 * implementation passes the throwable up through the nested contexts until a
	 * surrounding context is found that can handle throwables.  For example, a
	 * CMethodContextType instance collects the throwables that are uncaught within
	 * a method body, while a CTryContext collects the throwables that are thrown
	 * within a try-block and are therefore subject to handling by associated
	 * catch-blocks.
	 *
	 * @param throwable the type of the new throwable
	 */
	public void addThrowable(CThrowableInfo throwable) {
		if (cachedParent != null) {
			cachedParent.addThrowable(throwable);
		} else {
			methodContext.addThrowable(throwable);
		}
	}

	// ----------------------------------------------------------------------
	// MISC
	// ----------------------------------------------------------------------
	public CFlowControlContextType cachedParent() {
		return cachedParent;
	}

	// ----------------------------------------------------------------------
	// PRIVATE DATA MEMBERS
	// ----------------------------------------------------------------------

	/*@ public ghost boolean nestedFlowControlContext;
	  @ protected invariant
	  @            !nestedFlowControlContext <==> methodContext != null;
	  @ protected invariant nestedFlowControlContext <==> cachedParent != null;
	  @ protected invariant cachedParent != null <==> parent == cachedParent;
	  @ protected invariant methodContext != null <==> parent == methodContext;
	  @*/

	/**
	 * This is an alias of the inherited parent field but stored as a subtype to
	 * avoid casts.
	 */
	protected final CFlowControlContextType cachedParent;

	/**
	 * This is an alias of the inherited parent field but stored as a subtype to
	 * avoid casts.
	 */
	protected final CMethodContextType methodContext;

	protected boolean isReachable = true;

	/**
	 * The definite assignment state of the variables in this context.
	 */
	protected CVariableInfoTable variableInfo;

	/**
	 * The definite assignment state of the fields in this context.
	 */
	protected CVariableInfoTable fieldInfo;

	/**
	 * A map from identifiers to class signatures for the classes declared via type
	 * declaration statements in this context.
	 */
	private HashMap innerClasses;

	/**
	 * A list of the local variables declared in this context.
	 * <p/>
	 * <pre><jml>
	 * private invariant localVars != null ==>
	 * 		 (\forall Object o; localVars.contains(o);
	 * 			  o instanceof JLocalVariable );
	 * </jml></pre>
	 */
	private ArrayList localVars;

	/**
	 * The total number of local variables declared in the surrounding contexts
	 * before the beginning of this context.
	 */
	private final int parentIndex;

	/**
	 * The total number of local variables declared in this context.
	 */
	private int localsIndex = 0;

	/**
	 * The number of stack positions required to store the local variables
	 * encountered thus far in this and surrounding contexts.
	 */
	private int localsPosition;

	/**
	 * Default estimate of the number of local variable slots to reserve.
	 */
	private static final int DEFAULT_VAR_ESTIMATE = 0;

	/**
	 * A source code position for the start of this context.
	 */
	private final TokenReference where;
}
