/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CFlowControlContextType.java,v 1.15 2004/09/14 19:01:37 cclifton Exp $
 */

package org.multijava.mjc;

import java.util.Set;
import java.util.HashMap;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CWarning;
import java.util.ArrayList;
import java.util.Iterator;
import org.multijava.util.MessageDescription;

/**
 * <P>This class is used during typechecking for control flow analysis
 * that maintains local variable definite assignment (JLS2, 16),
 * throwable, and reachability information (JLS2, 14.20).</P> */
public interface CFlowControlContextType extends CContextType {

    // -------------------------------------------------------------
    // "DESTRUCTOR"
    // -------------------------------------------------------------

    /**
     * <P>Registers that this context is no longer needed.  Passes
     * information collected by this context to the parent context.</P>
     *
     * <p>Passes this context's field information to the surrounding
     * context; passes local variable information to the surrounding
     * context if that context is one which maintains local variable
     * information.</p>
     *
     * <P>Checks that local variables are used and issues warnings if
     * not.</P> 
     *
     * <P>This should only be called on a linearly nested context!
     * !FIXME!10/26 If we refactor the context creation code as
     * planned (to lock the parent context) then this method is called
     * to unlock the parent context when we exit a linearly nested
     * context.</P> */
    void checkingComplete();

    TokenReference getTokenReference();

    // ----------------------------------------------------------------------
    // ACCESSORS (NEW CONTEXT)
    // ----------------------------------------------------------------------

    // !FIXME!10/26 can we somehow lock this context until the created
    // context or contexts are closed?

    CFlowControlContextType createFlowControlContext(TokenReference where);
  
    CFlowControlContextType createFlowControlContext(int params,
						     TokenReference where);

    CLabeledContext createLabeledContext(JLabeledStatement self);

    CLoopContext createLoopContext(JLoopStatement self);
  
    CSwitchBodyContext createSwitchBodyContext( JSwitchStatement stmt,
						CType switchType );

    CTryContext createTryContext(TokenReference where);

    CFlowControlContextType createFinallyContext(
			CFlowControlContextType tryContext, TokenReference where);

    CExpressionContextType createExpressionContext();

    /**
     * Create a clone of this context to handle divergent paths in the
     * control flow.  */
    CFlowControlContextType cloneContext();


    /**
     * Creates a set of child contexts for typechecking parallel
     * control flows.  After using the contexts they should be passed
     * to the <code>adoptParallelContexts</code> method of this.
     *
     * <pre><jml>
     * requires count > 0;
     * </jml></pre>
     *
     * @see #adoptParallelContexts(CFlowControlContextType[]) */
    CFlowControlContextType[] createParallelContexts(int count,
						     TokenReference where);

    /**
     * Adopts the information from the given contexts.
     *
     * <pre><jml>
     * requires contexts != null && contexts.length > 0;
     * </jml></pre>
     *
     */
    void adoptParallelContexts(CFlowControlContextType[] contexts);

    // -------------------------------------------------------------
    // COMBINING CONTEXTS
    // -------------------------------------------------------------

    void merge(CFlowControlContextType context);

    void adopt(CFlowControlContextType context);

    // ----------------------------------------------------------------------
    // ACCESSORS (VARIABLE)
    // ----------------------------------------------------------------------

    /**
     * Checks whether a local variable has a name that is fresh
     * (i.e., no other local variable in scope has the same name).
     * @param	var		the variable
     */
    boolean isFreshVariableName(JLocalVariable var);

    /**
     * Adds a local variable to this block
     * @param	var		the name of the variable
     * @exception UnpositionedError	if the variable is a redeclaration
     */
    void addVariable(JLocalVariable var) throws UnpositionedError;

    /**
     * Records that the Java keyword <CODE>this</CODE> is used in this
     * block.
     */
    void addThisVariable();

    /**
     * Records that a variable slot must be reserved for a synthetic
     * <code>this</code> parameter.  Recall that <code>this</code> is
     * passed as a parameter for external methods.  */
    void addSyntheticThisParameter();

    /**
     * Returns the number of stack positions required to store the
     * local variables encountered thus far in this context.  */
    int localsPosition();

    int numberOfLocalVars();

    /** Returns the variable info table */
    CVariableInfoTable variableInfo();

    /** Returns the field info table */
    CVariableInfoTable fieldInfo();

    /** Returns the local variables list */
    ArrayList localVars();
    
    int parentIndex();

    int localsIndex();
			  
    // -------------------------------------------------------------
    // ACCESSORS (FLOW CONTROL)
    // -------------------------------------------------------------

    /**
     * Mutates the reachability status of this context.
     *
     * @param reachable indicates whether the statements in this context 
     *			are reachable after the current statement being 
     *			checked.
     */
    void setReachable(boolean reachable);

    /**
     * Indicates whether the statements in this context are reachable
     * up to most recent statement that was typechecked.  (Assuming
     * that statement's typechecking code calls
     * <code>setReachable</code> appropriately.)
     *
     * @return if this context is still reachable */
    /*@ pure @*/ boolean isReachable();

    /**
     * Returns the statement with the specified label.
     */
    JLabeledStatement getLabeledStatement(String label);

    /**
     * Returns the nearest breakable statement.
     */
    JStatement getNearestBreakableStatement();

    /**
     * Returns the nearest continuable statement.
     */
    JStatement getNearestContinuableStatement();

    /** 
     * Returns the nearest control flow context (where control flow
     * information is stored.)
     * @return the nearest parent of type CFlowControlContextType
     */
    CFlowControlContextType getFlowControlContext();

    // ----------------------------------------------------------------------
    // ACCESSORS (DEFINITE ASSIGNMENT)
    // ----------------------------------------------------------------------

    /**
     * Marks the variable with the given descriptor as definitely assigned
     * to in this context.  */
    void initializeVariable(VariableDescriptor varDesc);

    /**
     * Indicates whether the variable in the given position is
     * definitely assigned to in this context.  */
    boolean isVarDefinitelyAssigned(int pos);

	/**
	 * Indicates whether the variable in the given position is definitely
	 * unassigned to in this context
	 * @param i stack position
	 * @return true iff the variable is definitely unassigned
	 */
	boolean isVarDefinitelyUnassigned(int i);

    /**
     * Replaces the local field info for fields in positions 0 up to
     * <code>pos</code> with the info in <code>replacement</code>.  */
    void replaceFieldInfoUpTo( int pos, 
			       CVariableInfoTable replacement );

    /**
     * Replaces the local variable info in positions 0 up to
     * <code>pos</code> with the info in <code>replacement</code>.  */
    void replaceVariableInfoUpTo( int pos, 
				  CVariableInfoTable replacement );

    /**
     * Marks the field with the given descriptor as definitely assigned
     * to in this context.  */
    void initializeField(VariableDescriptor varDesc);

    /**
     * Indicates whether the field in the given position is
     * definitely assigned to in this context.  */
    boolean isFieldDefinitelyAssigned(int pos);

    // ----------------------------------------------------------------------
    // ACCESSORS (TYPE DEFINITION)
    // ----------------------------------------------------------------------

    /**
     * Adds to this context a class declared via a type declaration
     * statement.
     * @param	clazz		the clazz to add
     * @exception UnpositionedError	if duplicate class exists in this
     * lexical context */
    void addLocalClass(CClass clazz) throws UnpositionedError;

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------

    /**
     * Searches for a class with the given simple name according the
     * procedure in JLS2 6.5.5.  This method checks the first bullet
     * point of that section: "If the simple type name occurs within
     * the scope of a visible local class declaration with that name,
     * then the simple type name denotes that local class type."  If
     * no match is found, then this method calls to the surrounding
     * context to check the remaining bullet points.
     * 
     * @param	name	the class name, without qualifiers
     * @return	the class if found, null otherwise
     * @exception UnpositionedError if search fails */
    CClass lookupClass(String name) throws UnpositionedError;

    // ----------------------------------------------------------------------
    // THROWABLES
    // ----------------------------------------------------------------------

    /**
     * Registers that the given throwable can be thrown within this
     * context.  The implementation passes the throwable up through
     * the nested contexts until a surrounding context is found that
     * can handle throwables.  For example, a CMethodContextType
     * instance collects the throwables that are uncaught within a
     * method body, while a CTryContext collects the throwables that are
     * thrown within a try-block and are therefore subject to handling
     * by associated catch-blocks.
     * @param throwable the type of the new throwable */
    void addThrowable(CThrowableInfo throwable);
}
