/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CGFCollectionMap.java,v 1.6 2004/01/16 23:21:09 cclifton Exp $
 */

package org.multijava.mjc;

import java.util.Map;
import java.util.Set;
import java.util.HashSet;


/**
 * <P>This represents a mapping from compilation units and strings to
 * sets of collections of generic functions: <code>CGFCollectionMap:
 * CCompilationUnit x String -> { CGFCollection }</code>.  Instances of
 * this are stored within <code>CCompilationUnit</code> instances and
 * so the mapping is directly from strings to sets of collections of
 * generic functions.</P>
 *
 * <P>This is used to process unveiling references to external generic
 * functions.  Each generic function import for a compilation unit
 * generates an entry in this map.  When an unveiling reference is
 * encountered during typechecking this map is consulted to find the
 * set of generic function collections that should be made visible.
 * This is a set since multiple generic functions with the same name
 * may be imported from different packages.  Each collection in the
 * set is then processed to augment the appropriate classes.</P>
 *
 * @see CCompilationUnit
 * @see JCompilationUnit 
 * @see CGenericFunctionCollection
 * @see CAugmentationMap */
public class CGFCollectionMap extends org.multijava.util.Utils {
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    // just uses generic constructor

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * Adds a mapping from the given <CODE>ident</CODE>ifier to the
     * given generic function <CODE>coll</CODE>ection.  
     */
    public void register(String ident, CGenericFunctionCollection coll) {
	Set s = backingMap.getSet( ident );
	if( s==null ) {
	    s = new HashSet();

	    backingMap.put( ident, s );
	}
	s.add( coll );
    }

    /**
     * Adds a mapping from the given ident to each of the collections
     * in the given set. 
     *
     * <pre><jml>
     * requires (\forall Object o; collSet.contains(o); 
     *			 o instanceof CGenericFunctionCollection);
     * </jml></pre>
     */
    public void registerAll( String ident, Set collSet ) {
	Set s = backingMap.getSet( ident );
	if (s==null) {
	    s = new HashSet(collSet);
	    backingMap.put(ident,s);
	} else {
	    s.addAll(collSet);
	}
    }

    /**
     * Returns all the generic function collections mapped to by
     * <CODE>ident</CODE>.  */
    public CGenericFunctionCollection[] getCollectionsFor( String ident ) {
	Set s = backingMap.getSet( ident );
	if( s==null ) {
	    return new CGenericFunctionCollection[0];
	} else {
	    CGenericFunctionCollection[] result =
		(CGenericFunctionCollection[]) 
		s.toArray( new CGenericFunctionCollection[0] );
	    return result;
	}
    }

    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------
     
    public static class SetMap extends java.util.HashMap {
	Set getSet( String ident ) {
	    return (Set) get( ident );
	}
    }    

    private SetMap backingMap = new SetMap();

}


