/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CGenericFunctionCollection.java,v 1.9 2004/01/16 23:21:09 cclifton Exp $
 */

package org.multijava.mjc;

import java.util.HashSet;

import org.multijava.util.compiler.UnpositionedError;

/**
 * This interface represents a single generic function collection,
 * which encapsulates all the top-level methods of the same name
 * declared in a given context.  It include a methods for adding the
 * methods of the collection to the classes that they augment.  A
 * nested class is provided that implementors can delegate to.
 *
 * @see CGFCollectionMap */
public interface CGenericFunctionCollection {

    /**
     * This is just a unique object implementing the interface but
     * doing nothing.  It is used as a marker in
     * <CODE>CTopLevel.gfCollSingletonMap</CODE> and as a return value
     * from <CODE>CTopLevel.lookupGFCollection</CODE> 
     * @see CTopLevel#gfCollSingletonMap
     * @see CTopLevel#lookupGFCollection(String)
     */
    CGenericFunctionCollection GF_UNDEFINED = 
	new CGenericFunctionCollection() {
	    public void applyAugmentationFor(CContextType context) {
	    }
	    public String qualifiedName() {
		return "[UndefinedGF]";
	    }
	};

    /**
     * For the given context, augments the appropriate classes
     * based on the methods in this collection.  For each method
     * <em>m</em> in this collection, a mapping from the given
     * context to <em>m</em> is added to the <code>CClass</code>
     * representing <em>m</em>'s receiver type.
     *
     * @param context	the context in which the augmentation occurs */
    void applyAugmentationFor( CContextType context );

    /**
     * Returns the fully qualified name of this generic function collection.
     */
    String qualifiedName();

    class Impl extends CMethodSet implements CGenericFunctionCollection {
	public Impl( CMethod[] methods ) {
	    super( methods );
	}

	//---------------------------------------------------------------------
	// ACCESSORS
	//---------------------------------------------------------------------

	/**
	 * For the given context, augments the appropriate classes
	 * based on the methods in this collection.  For each method
	 * <em>m</em> in this collection, a mapping from the given
	 * context to <em>m</em> is added to the <code>CClass</code>
	 * representing <em>m</em>'s receiver type.
	 *
	 * @param context	the context in which the augmentation occurs */
	public void applyAugmentationFor( CContextType context ) {
	    CCompilationUnitContextType cuCT = context.getCompilationUnit();
	    if( ! augmentationCompleteSet.contains( cuCT ) ) {
		// have not done this augmentation before, so get busy
		int collSize = size();
		for( int i=0; i < collSize; i++ ) {
		    CMethod m = getMethod( i );
		    CClass receiver = m.receiverType().getCClass();
		    receiver.augmentWithMethod( m, cuCT );
		}
		augmentationCompleteSet.add( cuCT );
	    }
	}

	/**
	 * Returns the fully qualified name of this generic function
	 * collection.  Clients of this implementation class can
	 * generally implement this more efficiently and so should not
	 * delegate to this method.
	 */
	public String qualifiedName() {
	    if (size() == 0) {
		return "[unknown qualified name]";
	    } else {
		return getMethod(0).qualifiedName();
	    }
	}

	//---------------------------------------------------------------------
	// PRIVATE FIELDS
	//---------------------------------------------------------------------

	/**
	 * The tracks the contexts for which the augmentations in this
	 * collection have been applied.  This is an efficiency hack
	 * to avoid repeating work in
	 * <Code>applyAugmentationFor</code> */
	private HashSet augmentationCompleteSet = new HashSet();
    }
}
