/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CLoopContext.java,v 1.10 2004/01/22 20:04:39 leavens Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.UnpositionedError;


/**
 * This class represents the context for a loop statement during
 * checking passes (checkInterface, checkInitializers, typecheck).  It
 * stores the loop statement and include methods for calculating to
 * where <code>break</code>s and <code>continue</code>s inside this
 * loop should be directed.
 * @see CContextType 
 */
public class CLoopContext extends CFlowControlContext {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a block context, it supports local variable allocation
     * throw statement and return statement
     * @param	parent		the parent context, it must be different
     *				than null except if called by the top level
     */
    CLoopContext(CFlowControlContextType parent, JLoopStatement self) {
	super( parent, self == null ? null : self.getTokenReference() );

	this.self = self;
    }


    /**
     * Used by cloneContext.  Should not be called by clients.  */
    protected CLoopContext( CFlowControlContextType cachedParent,
			    CLoopContext clone ) {
	super( cachedParent, clone );
	this.self = clone.self;
    }

    // -------------------------------------------------------------
    // "DESTRUCTOR"
    // -------------------------------------------------------------

    /**
     * <P>Registers that this context is no longer needed.  Passes
     * information collected by this context to the parent context.</P>
     *
     * <p>Passes this context's field information to the surrounding
     * context; passes local variable information to the surrounding
     * context if that context is one which maintains local variable
     * information.</p>
     *
     * <P>Checks that local variables are used and issues warnings if
     * not.</P> 
     *
     * <P>This should only be called on a linearly nested context!
     * !FIXME!10/26 If we refactor the context creation code as
     * planned (to lock the parent context) then this method is called
     * to unlock the parent context when we exit a linearly nested
     * context.</P> */
    //@ also requires bodyChecksCompleted;
    public final void checkingComplete() {
	assertTrue( bodyChecksCompleted, 
		    "You have encountered a bug in mjc.  " + 
		    "We apologize for the inconvenience.  " +
		    "Please report the bug, including this error message " + 
		    "and the stack trace, to www.multijava.org.  Thank you." );
	super.checkingComplete();
    }


    /**
     * Performs final checks on the main body of the loop, i.e., on
     * the context that represents a full execution of the loop body
     * (without break or continue statements).  Checks that the loop
     * body does not assign to final variables and fields within code
     * that might be repeated.
     */
    //@ ensures bodyChecksCompleted;
    public final void doneWithCheckingBody() {
	bodyChecksCompleted = true;
	if (!isReachable()) {
	    // Returns without performing checks if this context will
	    // just be discarded anyway.
	    return;
	}
	
	// Checks for assignments to final variables and fields from
	// within loop.
	checkAssignmentToFinals();
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (NEW CONTEXT)
    // ----------------------------------------------------------------------

    /**
     * Create a clone of this context to handle divergent paths in the
     * control flow.  */
    public CFlowControlContextType cloneContext() {
	assertTrue( cachedParent != null );
	return new CLoopContext( cachedParent, this );
    }

    // -------------------------------------------------------------
    // ACCESSORS (BREAKS AND CONTINUES)
    // -------------------------------------------------------------

    /**
     * Get the nearest breakable statement
     */
    public JStatement getNearestBreakableStatement() {
	return self;
    }

    /**
     * Get the nearest continuable statement.
     */
    public JStatement getNearestContinuableStatement() {
	return self;
    }

    /**
     * Indicates whether this context is in a loop.
     */
    public final boolean isInLoop() {
	return true;
    }

    /**
     * Indicates whether this context is enclosed in a loop and the
     * given variable is declared outside the inner-most loop
     * context. */
    public boolean declaredOutsideOfLoop(JLocalVariable var) {
	// If recursion has reached this loop context then no inner
	// context declares the var.  Thus, if this context does not
	// declare the var, it must be the case that the var is
	// declared outside this loop.
	return !declares(var);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private final JLoopStatement self;

    /**
     * Indicates that checks have been performed on this context
     * between the completion of the loop body and the merging of
     * break and continuation contexts.
     */
    private /*@ spec_public @*/ boolean bodyChecksCompleted = false;
}
