/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CMember.java,v 1.19 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

/**
 * This class represents a member of a class or compilation
 * unit.  It is primarily just a data structure, apart from methods
 * for generating the qualified name and for determining whether the
 * member is accessible from some class.  The classes in this
 * hierarchy are used to track the signatures of the members declared
 * in *.java files being compiled and *.class files that are read as 
 * needed.  The subclasses also include methods for code generation.
 *
 * @see	CMember
 * @see	CField
 * @see	CMethod
 * @see CClass 
 */
public class CMember extends org.multijava.util.Utils implements Constants {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a top-level member export
     * @param	accessControl	the owner, host, and modifiers for this member
     * @param	ident		the ident of this member
     * @param	deprecated	is this member deprecated
     */
    public CMember(MemberAccess accessControl, 
		   String ident,
		   boolean deprecated)
    {
	this.accessControl = accessControl;
	this.ident = ident;
	this.deprecated = deprecated;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the owner of this member or null if this is a top-level
     * class declaration.
     *
     * @return the class that owns this member */
    public /*@ pure @*/ CClass owner() {
	return accessControl.owner();
    }

    /**
     * Returns the host of this member's declaration, which is the
     * same value returned by <code>owner()</code> unless this is a
     * top-level declaration.
     *
     * @return the host of this member's declaration */
    public /*@ pure @*/ CMemberHost host() {
	return accessControl.host();
    }

    /**
     * @return the ident of this method
     */
    public /*@ pure @*/ String ident() {
	return ident;
    }

    public String getIdent() {
	return ident;
    }

    /**
     * @return the fully qualified name of this member
     */
    public /*@ pure @*/ String qualifiedName() {
	return getOwnerName() + JAV_NAME_SEPARATOR + ident;
    }

    /**
     * @return the fully qualified name of this member's owner
     */
    public /*@ pure @*/ String getOwnerName() {
	return owner().qualifiedName();
    }
    
    
    /**
     * @return the fully qualified name of this member
     */
    public /*@ pure @*/ String getJavaName() {
	return getQualNameWithSeparator('.');
    }

    /**
     * @return the fully qualified name of this member with the given
     * character as the separator
     */
    public /*@ pure @*/ String getQualNameWithSeparator(char c) {
	return qualifiedName().replace('/' , c);
    }

    /**
     * @return the modifiers of this member
     */
    public /*@ pure @*/ long modifiers() { 
	return accessControl.modifiers();
    }

    /**
     * Add (bit or) to the modifiers of this member.
     */
    public void addModifiers(long modifiers) { 
	accessControl.addModifiers(modifiers);
    }

    /**
     * Sets the modifiers of this member
     */
    public void setModifiers(long modifiers) { 
	accessControl.setModifiers(modifiers);
    }

    /**
     * Indicates if this member is deprecated.
     */
    public /*@ pure @*/ boolean deprecated() {
	return deprecated;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (QUICK)
    // ----------------------------------------------------------------------

    /**
     * @return	the interface
     */
    public /*@ pure @*/ CField getField() {
	fail();
	return null;
    }

    /**
     * @return	the interface
     */
    public /*@ pure @*/ CMethod getMethod() {
	fail();
	return null;
    }

    /**
     * @return	the interface
     */
    public /*@ pure @*/ CClass getCClass() {
	fail();
	return null;
    }

    /**
     * @return	true if this member is static
     */
    public /*@ pure @*/ boolean isStatic() {
	return accessControl.isStatic();
    }

    /**
     * @return	true if this member is public
     */
    public /*@ pure @*/ boolean isPublic() {
	return accessControl.isPublic();
    }

    /**
     * @return	true if this member is protected
     */
    public /*@ pure @*/ boolean isProtected() {
	return accessControl.isProtected();
    }

    /**
     * @return	true if this member is private
     */
    public /*@ pure @*/ boolean isPrivate() {
	return accessControl.isPrivate();
    }

    public /*@ pure @*/ boolean hasDefaultAccess() {
	return accessControl.hasDefaultAccess();
    }

    /**
     * @return	true if this member is final
     */
    public /*@ pure @*/ boolean isFinal() {
	return accessControl.isFinal( );
    }

    public /*@ pure @*/  boolean isDeclaredNonNull() {
	return hasFlag(accessControl.modifiers(), ACC_NON_NULL | IMPLICITLY_NON_NULL);
    }

    /**
     * @return	true if this member is deprecated
     */
    public /*@ pure @*/ boolean isDeprecated() {
	return deprecated;
    }

    /**
     * Indicates whether the declaration of this member is local to
     * the given host.  There are two cases: 
     *
     *<OL>
     *
     * <LI>If the given host is <em>not</em> a compilation unit, then
     * this is local to the given host if the declaration of this
     * appears in the same compilation unit in which the declaration
     * of the given host appears.</LI>
     *
     * <li>If the given host <em>is</em> a compilation unit, then this
     * is local to the given host if the declaration of this appears
     * within the given host.</LI>
     *
     * </OL>
     *
     * @param	from		the given host
     * @return	desc
     * 
     */
    public /*@ pure @*/ boolean isLocalTo( CMemberHost from ) {
	CMemberHost host = accessControl.host();
	return from.isDefinedInside( host ) || host.isDefinedInside( from );
    }

    /**
     * Indicates whether this is accessible from the given host.
     *
     * @param	from	the host that wants access
     * @return	true iff the given host is allowed access to this member
     */
    public /*@ pure @*/ boolean isAccessibleFrom( CMemberHost from ) {
	return accessControl.isMemberVisibleIn( this, from );
    }

    /**
     * Indicates whether this member is accessible from the given host 
     * when this member has protected visibility.
     *
     * @param	from	the host that wants access
     * @return	true iff the given host is allowed access to this member
     */
    public /*@ pure @*/ boolean hasProtectedVisibilityIn( CMemberHost from ) {
	return from.isClass() && from.descendsFrom( owner() );
    }

    /**
     * Returns the compilation unit that hosts this member.
     */
    public CCompilationUnit getCCompilationUnit() {
	/*
	if( hostCompilationUnit == null ) {
	    // climb up through the nested hosts until we reach a
	    // compilation unit
	    CMemberHost climbingHost = host;
	    while( climbingHost != null && 
		   !climbingHost.isCompilationUnit() ) {
		climbingHost = climbingHost.host();
	    }
	    hostCompilationUnit = (CCompilationUnit) climbingHost;
	}
	return hostCompilationUnit;
	 */
	return accessControl.getCCompilationUnit();
    }

    public /*@ pure @*/ MemberAccess access() {
	return accessControl;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private String			ident;
    private boolean			deprecated;
    private MemberAccess		accessControl;
}
