/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CMethod.java,v 1.62 2008/05/14 18:16:31 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;
import java.util.Set;
import org.multijava.util.classfile.*;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents an actual method, whether generated from
 * source code or read from a class file.  It includes the type
 * signature of the method and operations for checking the method's
 * applicability and whether it is more specfic than a given
 * method. */
public abstract class CMethod extends CMember implements Comparable {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a method member.
     * @param	access		contains the info used to determine accessibility of this method
     * @param	ident		the ident of this method
     * @param	returnType	the return type of this method
     * @param	parameters	the parameters type of this method
     * @param	exceptions	a list of all exceptions of the throws list
     * @param	deprecated	is this method deprecated
     * @param	declarationContext the context in which this method is declared
     */
    public CMethod( MemberAccess access,
		    String ident,
		    CType returnType,
		    CSpecializedType[] parameters,
		    CClassType[] exceptions,
		    CTypeVariable[] typevariables,
		    boolean deprecated,
		    CContextType declarationContext )
    {
	super( access, ident, deprecated );

	this.receiverType = access().owner().getType();
	this.returnType = returnType;
	this.parameters = parameters;
	this.exceptions = exceptions;
	this.typevariables = typevariables;
	this.declarationContext = declarationContext;
	this.specArgs = new BitSet( parameters.length + 1 );
	// receiver is always specialized probably unnecesszary for
	// static methods, but it doesn't seem to hurt anything
	this.specArgs.set( 0 );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the declaration context.
     */
    public /*@ pure @*/ CContextType declarationContext() {
	return declarationContext;
    }

    /**
     * Returns the type of this methods receiver.
     */
    public /*@ pure @*/ CType receiverType() {
	return receiverType;
    }

    /**
     * @return	the interface
     */
    public /*@ pure @*/ CMethod getMethod() {
	return this;
    }

    /**
     * @return the type of this field
     */
    public final /*@ pure @*/ CType returnType() {
	return returnType;
    }

    /**
     * Returns the static and dynamic types of this methods parameters.
     * @return the static and dynamic types of this methods parameters
     */
    public final /*@ pure @*/ CSpecializedType[] parameters() {
	return parameters;
    }


    /**
     * Returns the size (in stack slots) of the parameters of this.  */
    public final int parametersSize() {
	//  !FIXME!could cache this value, would need to update (or
	//  purge cache) on assignment to this.parameters
	int size = 0;
	
	for (int i = 0; i < parameters.length; i++) {
	    size += parameters[i].getSize();
	}
	return size;
    }

    //@ assignable this.parameters;
    protected final void setParameters( CSpecializedType[] parameters ) {
	this.parameters = parameters;
    }
    
    
    protected final void setReturnType(CType type){
	this.returnType = type;
    }

    /**
     * Returns the bytecode type signature of this method as seen by
     * callers.  I.e., the signature based on the generic function's
     * static argument types.  */
    public String getSignature() {
	CType[] parms = CSpecializedType.staticTypeTupleFrom( null, 
							      parameters );
	return CType.genMethodSignature( returnType, parms );
    }
    
    public String getGenericSignature() {
	CType[] parms = CSpecializedType.staticTypeTupleFrom( null, 
							      parameters );
	return CType.genGenericMethodSignature( getTypeVariable(),returnType, parms );
    }
    /**
     * @return	the exceptions that can be thrown by this method
     */
    public /*@ pure @*/ CClassType[] throwables() {
	return exceptions;
    }

    /**
     * Set the exceptions that may be thrown by this method.  Used by
     * the typecheck methods for static and instance initializers
     * since the throwable exceptions for such code are only know
     * after the body is checked.
     * @param throwables the exceptions that can be thrown by this method */
    public void setThrowables(Set throwables) {
	if (throwables != null) {
	    Iterator iter = throwables.iterator();
	    int count = 0;

	    exceptions = new CClassType[throwables.size()];
	    while (iter.hasNext()) {
		exceptions[count++] =
		    ((CThrowableInfo)iter.next()).getThrowable();
	    }
	}
    }

    /**
     * Indicates whether this method belongs to an external generic function.
     *
     * @return	true iff the top method of the generic function to which this 
     *		belongs is external
     * @exception UnpositionedError if the top method of the generic function
     * is ambiguous
     * <pre><jml>
     * assignable topConcreteMethod;
     * </jml></pre>
     */
    public boolean isInExternalGF() 
	throws UnpositionedError
    {
	return topConcreteMethod().isExternal();
    }

    
    /**
     * Has this method the same signature as the one given as argument ?
     * NOTE: return type not considered
     * @param	other		the method to compare to
     */
    public boolean hasSameSignature(CMethod other, CClassType[] substitution) {
	if (parameters.length != other.parameters.length) {
	    return false;
	} else {
	    for (int i = 0; i < parameters.length; i++) {
		if (!parameters[i].equals(other.parameters[i], substitution)) {
		    return false;
		}
	    }
	    return true;
	}
    }

    // WMD: version without Universes

    public boolean equalParametersNoUniverses( CMethod other ) {
	if( parameters.length != other.parameters.length ) {
	    return false;
	}

	for( int i = 0; i < parameters.length; i++ ) {
	    if( !parameters[i].equals( other.parameters[i], false ) ) {
		return false;
	    }
	}
	return true;
    }



    public boolean hasSameSignatureNoUniverses(CMethod other) {
	if( this==other ) {
	    return true;
	}
	
	if( !ident().equals( other.ident() ) ) {
	    return false;
	}

	if( !this.equalParametersNoUniverses( other )) {
	    return false;
	}

	if( !receiverType.equals( other.receiverType ) ) {
	    return false;
	}
	
	return true;
    }
    
    /**
     * Has this method the same signature as the one given as argument ?
     * NOTE: return type not considered
     * @param	other		the method to compare to
     */
    public boolean hasSameSignature(CMethod other) {
		
	if( !ident().equals( other.ident() ) ) {
	    return false;
	}

	CSpecializedType[] op = other.parameters();
	if (parameters.length != op.length) {
 	    return false;
 	}
	
	boolean ret = true;
	for (int i = 0; i < parameters.length; i++) {
	    //need to be changed later, here TJ
	    CType p1 = parameters[i].staticType();
	    CType p2 = op[i].staticType();
	    CType d1 = parameters[i].dynamicType();
	    CType d2 = op[i].dynamicType();

            if (p1.isArrayType() != p2.isArrayType()) {
                return false;
            } else if (p1.isArrayType() && p2.isArrayType()) {
                CArrayType a1 = (CArrayType) p1;
                CArrayType a2 = (CArrayType) p2;
                if (a1.getArrayBound() != a2.getArrayBound()) {
                    return false;
                } else {
                    p1 = a1.getBaseType();
                    p2 = a2.getBaseType();
                }
            }

            if (d1.isArrayType() != d2.isArrayType()) {
                return false;
            } else if (d1.isArrayType() && d2.isArrayType()) {
                CArrayType a1 = (CArrayType) d1;
                CArrayType a2 = (CArrayType) d2;
                if (a1.getArrayBound() != a2.getArrayBound()) {
                    return false;
                } else {
                    d1 = a1.getBaseType();
                    d2 = a2.getBaseType();
                }
            }

	    if ((!p1.isClassType() || !p2.isClassType()) && 
		!p1.equals(p2) ){
		return false;
	    }
	    
	    if ((!d1.isClassType() || !d2.isClassType()) && 
		!d1.equals(d2) ){
		return false;
	    }
	    
	    if (p1.isClassType() && p2.isClassType() 
		&&!(p1.getCClass().equals(p2.getCClass()))){
		// if(p1.isClassType() && p2.isClassType() &&
		// 		   !p1.equals(p2)) {
		return false;
	    } 
	    if (d1.isClassType() && d2.isClassType() 
		&&!(d1.getCClass().equals(d2.getCClass()))){
		// if(p1.isClassType() && p2.isClassType() &&
		// 		   !p1.equals(p2)) {
		return false;
	    } 
		
	}
	return true;
    }
    
    /**
     * Finds the top concrete method of the generic function to which
     * this belongs.  Registers any specialized argument positions of
     * this with the top method.
     *
     * <pre><jml>
     * requires topConcreteMethod != null || declarationContext != null || 
     *          (* this is overridden in subclass to not depend on 
     *		   declarationContext *);
     * assignable topConcreteMethod;
     * ensures topConcreteMethod != null;
     * </jml></pre>
     *
     * @return the top method of the generic function to which this belongs */
    public CMethod topConcreteMethod() 
	throws UnpositionedError
    {
	if (topConcreteMethod != null) {
	    return topConcreteMethod;
	}
	CClassType[] args = owner() == null? CClassType.EMPTY:
	    owner().getType().getArguments(); 
	assertTrue( declarationContext != null );
	topConcreteMethod =
	    owner().lookupTopConcreteMethod( this,args ,declarationContext );
	if (topConcreteMethod == null) {
	    topConcreteMethod = this;
	}

	// set specialized arguments for top method when we first find
	// the top method
	for (int i = 0; i < parameters.length; i++) {
	    if (parameters[i].isSpecialized()) {
		topConcreteMethod.specArgs.set(i+1);
	    }
	}
	
	return topConcreteMethod;
    }

    /** 
     * Return the bit set indicating which arguments of the associated
     * generic function are specialized.  This field is only used by
     * top methods. */
    public /*@ pure @*/ BitSet getSpecArgs() {
	return specArgs;
    }

    /**
     * Set the method's gfID and methodID, used to give its associated
     * $body method (if any) a unique name. */
    public void setIDs(int gfID, int mID) {
	this.gfID = gfID;
	this.methodID = mID;
    }
    
    // ----------------------------------------------------------------------
    // ACCESSORS (QUICK)
    // ----------------------------------------------------------------------

    /**
     * Returns true iff this method is native.
     */
    public /*@ pure @*/ boolean isNative() {
	return access().isNative();
    }

    /**
     * Returns true iff this method is abstract.
     */
    public /*@ pure @*/ boolean isAbstract() {
	return access().isAbstract();
    }

    /**
     * Returns true iff this method is a constructor.
     */
    public /*@ pure @*/ boolean isConstructor() {
	return ident() == JAV_CONSTRUCTOR;
    }

    /**
     * Indicates whether this method is external, i.e., whether its
     * host is different than its logical owner.
     *
     * @return	true iff this method is external */
    public /*@ pure @*/ boolean isExternal() {
	return access().isExternalMethod();
    }

    /**
     * Indicates whether this method is declared local to the given
     * context.
     *
     * @param	context	the context in which to check
     * @return whether this method is declared local to the given context */
    public boolean isLocalTo( CContextType context ) {
	return host().isDefinedInside( context.getCompilationUnit()
				       .findNearestHost() );
    }

    /**
     * Indicates that this method is pleomorphic with one external
     * generic function and one or more internal generic functions.
     * */
    protected void setIsMixedPleomorphic() {
	isMixedPleomorphic = true;
    }
    /**
     * Indicates whether this method is pleomorphic with one external
     * generic function and one or more internal generic functions.
     * */
    public boolean isMixedPleomorphic() {
	return isMixedPleomorphic;
    }


    // ----------------------------------------------------------------------
    // CHECK MATCHING
    // ----------------------------------------------------------------------

    /**
     * Returns true if the given <code>method</code> and
     * this method have parameters with the same types.
     *
     * @param other the <code>CMethod</code> to compare this to */
    public boolean equalParameters( CMethod other ) {

	if( parameters.length != other.parameters.length ) {
	    return false;
	}

	for( int i = 0; i < parameters.length; i++ ) {
	    if( !parameters[i].equals( other.parameters[i] ) ) {
		return false;
	    }
	}
	return true;
    }
    
    /**
     * Returns true if the given object is a <code>CMethod</code> and
     * the two methods have same signature.
     *
     * @param other the object to compare this to
     */
    public boolean equals(/*@ nullable @*/  Object other ) {
	if( this==other ) {
	    return true;
	}
	
	if( !(other instanceof CMethod) ) {
	    return false;
	}

	CMethod method = (CMethod) other;

	if( !ident().equals( method.ident() ) ) {
	    return false;
	}

	if( !this.equalParameters( method )) {
	    return false;
	}

	if( !receiverType.equals( method.receiverType ) ) {
	    return false;
	}
	
	return true;
    }

    /**
     * Returns the hash code for this, calculated from the hash codes
     * of the methods identifier, receiver, and argument types.
     *
     * @return the hash code for this */
    public int hashCode() {
	// !FIXME! count how many times this is called in a typical
	// compilation for the same object and decide if it makes
	// sense to cache this value.  Keep in mind that if some of
	// the fields are not final than the hashCode will change.
	int result = ident().hashCode();
	result = result * 31 + receiverType.hashCode();
	for( int i=0; i<parameters.length; i++ ) {
	    result = result * 31 + parameters[i].hashCode();
	}
	return result;
    }
    
    /**
     * Allows methods to be sorted.  Comparison is by identifier only.
     *
     * @param	o	the method to be compared with
     * @return	-1, 0, 1 based on identifiers, see String.compareTo
     */
    public int compareTo(/*@ non_null @*/ Object o) {
	CMethod other = (CMethod) o;
	if( this.equals( other ) ) {
	    return 0;
	} else {
	    return ident().compareTo( other.ident() );
	}
    }

    /**
     * Returns true if this method is applicable to a method call with
     * the given identifier and actual (static) argument types.
     *
     * @param	ident		the identifier to match
     * @param	recvType	receiver type	
     * @param	actuals		the method call static argument types
     * @return true if this method is applicable */
    public boolean isApplicable( String ident, 
				 CType recvType, CType[] actuals,
				 CClassType[] args) {
	return parameters.length == actuals.length &&
	    ident == ident() &&      
	    isApplicableArgTypes( recvType, actuals,
				  (ident==JAV_CONSTRUCTOR
				   || ident==JAV_INIT
				   || ident==JAV_STATIC_INIT),
				  args);
    }

    /**
     * Returns true if the given actual argument types are assignable
     * to this methods formals and receiver.
     * 
     * @param	recvType	receiver type
     * @param	actuals		the method call static argument types
     * @param	initStyleCheck	if true then perform an initializer
     *				applicability check (i.e., receiver types
     *				must be equal), otherwise perform a pointwise
     *				subtype check on the argument type tuples
     * @return true if the given actual argument types are assignable 
     * to this methods formals
     */
    private boolean isApplicableArgTypes( CType recvType, CType[] actuals,
					  boolean initStyleCheck,
					  CClassType[] args) 
    {
	if( (initStyleCheck && !recvType.equals(receiverType, false))
	    || (!initStyleCheck 
		&& !recvType.isAlwaysAssignableToIgnoreUniverses(receiverType))) {
	    return false;
	}
	for( int i = 0; i < actuals.length; i++ ) {
	    // method invocation conversion == assignment conversion
	    // without primitive narrowing [JLS2 5.2, 5.3].  Since the 
	    // primitive narrowing is only applicable to literals it is 
	    // checked in the JLiteral subclasses.  The types only check
	    // widening conversions.
	    // WMD: original check:
	    //	    if( !actuals[i].isAlwaysAssignableTo( parameters[i].dynamicType(), args ) ) {
	    // WMD: but the universe check on parameters is more complicated...

	    // if Universes are disabled or
            // it's a primitive type do the old check
            if( !(actuals[i] instanceof CClassType) ) {

		if( !actuals[i].isAlwaysAssignableTo( parameters[i].dynamicType(), args ) ) {
		    return false;
		}
            } else {
            	// WMD: we have to ignore the Universe modifiers in the method lookup,
            	// and therefore here, because we have to do viewpoint adaptation to get
            	// the correct types. Unfortunately, the receiver here is only of the correct
            	// class, not the actual receiver type of a method call.
            	
		// Universe checks are turned off
		if (!((CClassType)actuals[i]).isAlwaysAssignableToIgnoreUniverses(
				parameters[i].dynamicType(),
				args)) {
		    // compare without Universes
		    return false;
		}
		}
            
	} // for loop

	return true;
    }

    /**
     * Returns true if this method is strictly more specific than the
     * given method.  This does not mean that this method 
     * <em>specializes</em> the given method.
     * 
     * @param	other		the method to compare to
     * @return	true if this method is more specific than 
     * <code>other</code> */
    //@ pure
    public boolean isMoreSpecificThan( CMethod other, CClassType[] args ) {
    	// WMD Universes disabled!
	return isMoreSpecificThan( other, args,false,false );
    }

    //@ pure
    public boolean isMoreSpecificThan(CMethod other) {
    	// WMD Universes disabled!
	return isMoreSpecificThan(other, receiverType().getArguments(),false,false);
    }
    
    //@ pure
    public boolean isMoreSpecificThan(CMethod other, boolean b) {
    	// WMD Universes disabled!
	return isMoreSpecificThan(other, receiverType().getArguments(),b,false);
    }

    /**
     * Returns true if this method is strictly more specific than the
     * given method.  This does not mean that this method 
     * <em>specializes</em> the given method.
     * 
     * @param	other		the method to compare to
     * @param	intrsDoNotExtendObject true if we should not treat interfaces
     *					as inheriting from Object
     * @return	true if this method is more specific than 
     * <code>other</code> */
    public /*@ pure @*/
        boolean isMoreSpecificThan( CMethod other,
				    CClassType[] args,
				    boolean intrsDoNotExtendObject ) {

	return isMoreSpecificThan( other, args, intrsDoNotExtendObject, CTopLevel.getCompiler().universeChecks() );
    }

    // WMD: version with Universe switch
    public /*@ pure @*/
        boolean isMoreSpecificThan( CMethod other,
				    CClassType[] args,
				    boolean intrsDoNotExtendObject,
				    boolean enableUniv) {
	String ident = ident();
	if( ident != other.ident() 
	    || parameters.length != other.parameters.length ) {
	    return false;
	}

	    // if they have different receivers and are both static,
	    // then the methods are unrelated
	if ( !(receiverType.equals( other.receiverType )) &&
	     isStatic() && other.isStatic() ) {
	    return false;
	}

	    // if we shouldn't treat interfaces as inheriting from Object
	    // then don't
	if ( intrsDoNotExtendObject && owner().isInterface() &&
	     other.receiverType() == CStdType.Object ) {
	    return false;
	}
	
	// if true then perform an initializer
	// applicability check (i.e., receiver types must be equal),
	// otherwise perform a pointwise subtype check on the argument
	// type tuples
	boolean initStyleCheck = (ident==JAV_CONSTRUCTOR
				  || ident==JAV_INIT 
				  || ident==JAV_STATIC_INIT);

	boolean eqSpec = true;	// equally specific
	if( (initStyleCheck && ! receiverType.equals( other.receiverType ))
	    ||
	    (!initStyleCheck &&
	     !receiverType.isAssignableTo( other.receiverType ))) {
	    return false;
	} else {
	    eqSpec = eqSpec && receiverType.equals( other.receiverType );
	}

	CSpecializedType[] otherFormals = other.parameters;
	for( int i = 0; i < otherFormals.length; i++ ) {
	    // method invocation conversion == assigment conversion
	    // without literal narrowing we just look at the type and
	    // do not consider literal special cases
	    if( enableUniv && !parameters[i].isAlwaysAssignableTo( otherFormals[i],args ) ) {
	    	return false;
	    } else if( !enableUniv && !parameters[i].isAlwaysAssignableToIgnoreUniverses(otherFormals[i],args ) ) {
	    	return false;
	    } else {
		eqSpec = eqSpec && 
		    parameters[i]
		    .dynamicType().equals( otherFormals[i].dynamicType(),
					   args, enableUniv);
	    }
	}

	// if control flow reaches here, then this is at least as
	// specific as other
	return !eqSpec;
    }


    /**
     * <P>Checks whether this method apparently specializes a given
     * method.  This method apparently specializes <code>other</code>
     * iff this method is more specific than <code>other</code>, the
     * two methods' static parameter types are identical, and the
     * receiver type of this is a (possibly reflexive) subtype of the
     * receiver type of other.</P>
     *
     * <P>This relationship is <em>apparently</em> because
     * <code>other</code> is not necessarily visible to this.  For
     * example, <code>other</code> could be a method of an external
     * generic function that is not imported in the context of the
     * declaration of this.</P>
     *
     * @param	other		the method to compare to
     * @return true if this method apparently specializes <code>other</code>
     * @see #specializes(CMethod) */
    public /*@ pure @*/ boolean apparentlySpecializes( CMethod other,
						       CClassType[] args) {

	return apparentlySpecializes( other, args, CTopLevel.getCompiler().universeChecks() );
    }

    // WMD: version with Universe switch
    public /*@ pure @*/ boolean apparentlySpecializes( CMethod other,
						       CClassType[] args,
						       boolean enableUniv ) {
	if( !isMoreSpecificThan( other, args, false, enableUniv ) ) {
	    return false;
	}
	//@ assert parameters.length == other.parameters.length;
	for( int i=0; i<parameters.length; i++ ) {
	    if( !parameters[i].specializes( other.parameters[i], args )
		&& !parameters[i].equals( other.parameters[i], args, false ) ) {
		return false;
	    }
	}
	return true;
    }

    public /*@ pure @*/ boolean apparentlySpecializes( CMethod other ){
	return apparentlySpecializes(other,receiverType().getArguments(),true);
    }
    
    /**
     * <P>Checks whether this method specializes a given method.  This
     * method specializes <code>other</code> iff this method
     * apparently specializes <code>other</code>, and the two methods
     * have belong to the same generic function as determined by their
     * top methods.</P>
     *
     * @param	other		the method to compare to
     *
     * <pre><jml>
     * requires other != null;
     * assignable topConcreteMethod, other.topConcreteMethod;
     * ensures \result <==> this.topConcreteMethod == other.topConcreteMethod
     *                        && this.apparentlySpecializes( other );
     * </jml></pre>
     *
     * @return true if this method specializes <code>other</code> 
     * @exception UnpositionedError if the top method of either this or 
     * 				    <code>other</code> is ambiguous
     * @see #apparentlySpecializes(CMethod) */
    public boolean specializes( CMethod other ) 
	throws UnpositionedError
	       // !FIXME! this should be a positioned error, else the
	       // error will be mis-positioned when returned back
	       // through lookupTopConcreteMethod
    {
	assertTrue( other != null );
	return this.topConcreteMethod() == other.topConcreteMethod()
	    && this.apparentlySpecializes( other );
    }

    /**
     * Returns a human-readable string representation of this method
     * signature. 
     * 
     * Add token reference information once this is made available in
     * CMember (see TODO list).!FIXME!
     *
     * @return a human-readable string representation of this */
    public String toString() {
	return toString( false );
    }

    /**
     * Utility method allows the generation of a string representation
     * of this with or without the receiver type.  */
    protected String toString( boolean omitReceiver ) {
	StringBuffer result = new StringBuffer(200);
				// 200 is just something big enough to 
				// ordinarily avoid the buffer needing to be 
				// enlarged.  Pick any number you like, but
				// this method gets called frequently.
	if (!omitReceiver) {
		result.append(receiverType);
		result.append(".");
	}
	result.append(ident());
	result.append("( ");
	if (parameters.length != 0) result.append(parameters[0].toVerboseString());
	for( int i=1; i<parameters.length; i++ ) {
	    result.append(", ");
	    result.append(parameters[i].toVerboseString());
	}
	return result.toString() + " )";
    }

    /**
     *@return the type variables of this method
     */
    public CTypeVariable[] getTypeVariable(){
	return typevariables;
    }
    
    /**
     *whether this method is generic or not
     */
    public boolean isGenericMethod(){
	return typevariables.length >0;
    }
    
    public /*@nullable*/ CTypeVariable lookupTypeVariable(String ident){
	for (int i = 0; i < typevariables.length; i++) {
	    if (ident.equals(typevariables[i].getIdent())) {
		return typevariables[i];
	    }
	}
	if (isStatic()) {
	    return null;
	} else {
	    return owner().lookupTypeVariable(ident);
	}
    }


    // ----------------------------------------------------------------------
    // Purity of a method
    // ----------------------------------------------------------------------


    /**
     * Returns true if this method is pure. A method is pure if it is
     * annotated with the <code>pure</code> modifier or it
     * inherits pureness from its owner or supertypes. A non-static
     * method (including a constructor) is pure if its owner class or
     * interface is pure, or if it overrides a pure method inherited
     * from supertypes (classes and interfaces).
     */
    //@ implies_that
    //@   assignable objectState;
    public boolean isPure() {
        if (pureCached) {
            return cachedPure;
        }

        // calcluate and cache the pureness of this method
        cachedPure = access().isPure()
	    // inherited from supertypes (classes and interfaces)?
            || isPureInherited();
        pureCached = true;
        
	if( !access().isPure() &&
	    cachedPure ) {
	    // the modifier was changed
	    purityWasChanged = true;
	}

        return cachedPure;
    }

    /**
     * This method can be used to determine that the purity of this method
     * was not given explicitly by the programmer, but inherited from
     * an overridden method or the class/interface.
     */
    public boolean purityWasChanged() {
	return purityWasChanged;
    }

    /**
     * Returns true if this method inherits pureness from its owner or 
     * supertypes, including both superclasses and interfaces.
     */
    private boolean isPureInherited() {
	if ( hasFlag(modifiers(), ACC_STATIC) ) {
	    // static methods do not inherit pureness from their owner 
	    // type or from overridden methods (they cannot override 
	    // super class methods).
	    return false;
	}

	CClass owner = owner();
	// inherited from owner class or interface?
	if ( owner != null
	     && owner.access().isPure() ) {
	    return true;
	}

	try {
	    setOverriddenMethodSet();
	} 
	catch (UnpositionedError e) {
	    return true; // don't know, so be conservative
	}
        CMethodSet mset = overriddenMethodSet();
        if (mset != null) {
            for (int i = 0; i < mset.size(); i++) {
                CMethod meth = mset.getMethod(i);
                if ( meth.isPure() ) {
                    return true;
                } 
            }
        }
        return false;
    }


    /**
     * A version that uses default parameters.
     * This method is public, because it is also called from JML.
     */
    public CMethodSet setOverriddenMethodSet() 
	throws UnpositionedError
    {
	return setOverriddenMethodSet(receiverType().getArguments(), new ArrayList(),
				      declarationContext() );
    }


    /**
     * Calculates and returns the set of methods overridden by this
     * method.  Caches the result for later calls to 
     * {@link #overriddenMethodSet()}.
     */
    protected CMethodSet setOverriddenMethodSet( CClassType[] args,
						 ArrayList needbridge,
						 CContextType context ) 
	throws UnpositionedError
    {
	overriddenMethodSet = 
	    owner().lookupMethodsSpecializedByMethod( this, args, needbridge,
						      context );
	return overriddenMethodSet;
    }


    /**
     * Returns the set of methods overridden by this method, may be
     * null.
     */
    public /*@ pure nullable @*/ CMethodSet overriddenMethodSet() {
	return overriddenMethodSet;
    }


    /**
     * The pureness of this method. We cache this value because it's
     * expensive to calculate it to inheritance. An overriding method
     * inherits the pureness of the overridden methods.
     */
    private boolean cachedPure; //@ in objectState;


    /**
     * Indicates if the pureness of this method has been calculated
     * and cached in the field <code>cachedPure</code>.
     */    
    private boolean pureCached = false;  //@ in objectState;


    /**
     * Indicates if the pureness of this method was inherited
     * from a supermethod or a class/interface.
     */    
    private boolean purityWasChanged = false;  //@ in objectState;


    /**
     * The set of methods overridden by this method, may be null.
     */
    private CMethodSet overriddenMethodSet;



    // ----------------------------------------------------------------------
    // GENERATE BYTECODE INFO
    // ----------------------------------------------------------------------

    /**
     * Swallow the receiver, which is passed as the first argument
     * to an external generic function.
     *
     * <pre><jml>
     * requires topConcreteMethod != null && topConcreteMethod.isExternal();
     * assignable parameters;
     * ensures parameters.length == \old(parameters).length + 1;
     * ensures parameters[0].staticType() == topConcreteMethod.receiverType();
     * ensures parameters[0].dynamicType() == this.receiverType();
     * </jml></pre>
     *
     */
    protected void swallowReceiver() {
        CSpecializedType[] oldParms = parameters();
	CSpecializedType[] newParms = 
	    new CSpecializedType[ oldParms.length + 1 ];
	if (topConcreteMethod.receiverType() == receiverType() || 
	    this instanceof CSourceDispatcherMethod) {
	    newParms[0] = 
		new CSpecializedType( topConcreteMethod.receiverType() );
	} else {
	    newParms[0] = 
		new CSpecializedType( topConcreteMethod.receiverType(), 
                                      receiverType() );
	} // end of else
	System.arraycopy( oldParms, 0, newParms, 1, oldParms.length );
	
	setParameters( newParms );
    }

    /**
     * Generates a sequence of bytecode for invoking this method,
     * assuming appropriate argument values have already been pushed
     * to the JVM stack.
     * @param	code		the code sequence
     * @param nonVirtual force non-virtual dispatching if possible (ignored if
     *			 this method belongs to an interface or is external)
     * <pre></jml>
     * requires topConcreteMethod != null;
     * </jml></pre>
     */
    public void genGenFuncInvocation( CodeSequence code, boolean nonVirtual ) {
	if( owner().isInterface() || topConcreteMethod.isExternal() ) {
	    int size = parametersSize();
      
	    String interfaceName;
	    String id;
	    if( topConcreteMethod.isExternal() ) {
		interfaceName = dispatcherSignature().qualifiedName();
		id = "apply";
	    } else {
		interfaceName = getOwnerName();
		id = ident();
	    }
//  	    System.err.println( "interfaceName: " + interfaceName );
//  	    System.err.println( "id: " + id );
//  	    System.err.println( "sig: " + getSignature() );
//  	    System.err.println( "size+1: " + size+1 );
	    code.plantInstruction(
		new InvokeinterfaceInstruction( interfaceName,
						id,
						getSignature(),
						size + 1 ));
	} else {
	    int opcode;

	    try {
		if (isStatic()) {
		    opcode = opc_invokestatic;
		} else if ((nonVirtual || isPrivate()) &&
			   !topConcreteMethod().needsPrivacyModifierRemoved) {
			// JDK1.2 does not like ... || isFinal()
		    opcode = opc_invokespecial;
		} else {
		    opcode = opc_invokevirtual;
		}
		code.plantInstruction(
		    new MethodRefInstruction( opcode,
					      getOwnerName(),
					      ident(),
					      getSignature() ));
	    } catch(UnpositionedError e) {
		fail("shouldn't ever reach this code");
	    }

	}
    }

    /**
     * Generates bytecode to put a reference to the external
     * <code>function</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     *   requires this.isExternal();
     * </jml></pre> */
    public abstract void plantFunctionRef( CodeSequence code );

    /**
     * Generates bytecode to put a reference to the
     * <code>oldFunction</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     *   requires this.isExternal();
     * </jml></pre> 
     * @param code the code sequence to add the bytecode to
     * @param forSuper if true, then this is for a super send, which,
     *                 under RMJ, should start from the oldSuperFunction
     *                 rather than oldFunction
     * @param fromOuter if true, then generates a reference that works from the
     *                  outer class context, otherwise generates a reference that
     *			works from within the dispatcher class
     */
    public abstract void plantOldFunctionRef( CodeSequence code, 
					      boolean forSuper, 
					      boolean fromOuter );


    /**
     * Returns the $signature class for this.
     *
     * <pre><jml>
     *   requires this.isExternal();
     * </jml></pre> */
    public abstract CDispatcherSignature dispatcherSignature();

    /**
     * Returns the ambiguous-method dispatcher class for this. (RMJ only.)
     *
     * <pre><jml>
     *   requires this.isExternal();
     * </jml></pre> */
    public abstract CAmbiguousDispatcherClass ambiguousDispatcherClass();

    /**
     * Returns the $anchor class for this generic function.
     *
     * <pre><jml>
     * requires this.isExternal();
     * </jml></pre>
     *
     */
    public abstract CClass anchorClass();
    
    /**
     * Returns the function field number for this generic function
     * within its $anchor class.
     *
     * <pre><jml>
     * requires this.isExternal();
     * </jml></pre>
     * */
    public abstract int functionNumber();
    
    /**
     * Generates the dispatch code for the specialized arguments of
     * this method.  Generated code implements the following
     * pseudocode:
     *
     * <pre><code>
         if ( !(this is applicable) ) then goto nextAlternative;
	 execute this method body;
     * </code></pre> 
     *
     * @return true if any applicability test code was required */
    public boolean genDispatch( CodeSequence code, CodeLabel nextAlternative,
				/*@ non_null @*/ CMethod topConcreteMethod,
                                String dispatcherName ) {
	boolean testRequired = false;
	boolean isExternalGF = topConcreteMethod.isExternal();
	LocalVarInstruction varref;
	String instOfCheck;
	// Used to calculate position of parameters:
	int pos = isStatic() ? 0 : 1;
	
	// for external g.f. the receiver has already been added to
	// the parameter list, so this loop handles all necessary
	// dispatch tests
	for( int i=0; i < parameters.length; i++ ) {
	    CSpecializedType t = parameters[i];
	    if( t.isSpecialized() ) {
		CType dynType = t.dynamicType();
		    // load the actual parameter
		code.plantInstruction( new LocalVarInstruction(
					   dynType.getLoadOpcode(), pos ) );
		    // generate the dispatch test, branching
		    // to the nextAlternative if the test fails
		dynType.plantDispatchTest( code, nextAlternative );
		testRequired = true;
	    }
	    pos += t.getSize();	    
	}

	    // now generate the call to the appropriate $body method
	
	// plant arguments to the call
	if (!isStatic())
	    code.plantLoadThis();
	plantSelfArguments( code, topConcreteMethod );

	// plant invocation of multimethod body
	plantMMInvocation( code, dispatcherName );

	code.plantInstruction( new NoArgInstruction( returnType.
						     getReturnOpcode() ));
	return testRequired;
    }

    /**
     * Plants bytecode to push the arguments for a call to this
     * multimethod body from within a dispatcher of this generic
     * function.  After executing the generated code the JVM stack
     * will include the arguments of the generic function, cast to the
     * appropriate dynamic types.  */
    protected void plantSelfArguments( CodeSequence code,
                                       /*@ non_null @*/ CMethod topConcreteMethod ) {
	boolean isExternalGF = topConcreteMethod.isExternal();
	LocalVarInstruction varref;
	String instOfCheck;

	// Used to calculate position of parameters:
	int pos = isStatic() ? 0 : 1;
	
	for( int i=0; i < parameters.length; i++ ) {
	    CSpecializedType t = parameters[i];
	    varref = new LocalVarInstruction( t.staticType().getLoadOpcode(), 
					      pos );
	    code.plantInstruction( varref );
		// value dispatches do not require downcasts
	    if( t.isSpecialized() && !t.dynamicType().isValueType() ) {
		instOfCheck = 
		    ((CClassType) t.dynamicType()).qualifiedName();
		code.plantInstruction( new ClassRefInstruction( opc_checkcast,
								instOfCheck ));
	    }
	    pos += t.getSize();	    
	}
    }

    /**
     * Plants an invocation of the multimethod body represented by
     * this.  */
    protected void plantMMInvocation( CodeSequence code, String dispatcherName ) 
    {
	int opcode;

	opcode = isStatic() ? opc_invokestatic : opc_invokespecial;
	
	MethodRefInstruction invokeBody =
	    new MethodRefInstruction( opcode, dispatcherName, bodyIdent(),
				      bodySignature() );
	code.plantInstruction( invokeBody );
    }

    protected /*@ pure @*/ String bodyIdent() {
	return ident() + "$body" + gfID + "$" + methodID;
    }

    protected /*@ pure @*/ String bodySignature() {
	CType[] parms = CSpecializedType.dynamicTypeTupleFrom( parameters );
	return CType.genMethodSignature( returnType, parms );
    }
    
    protected String bodyGenericSignature() {
	CType[] parms = CSpecializedType.dynamicTypeTupleFrom(parameters);
	return CType.genGenericMethodSignature(getTypeVariable(), returnType,
					       parms);
    }


    /**
     * Returns the type signature of this method in the multimethod
     * descriptor format, which is like a regular bytecode method
     * descriptor but includes explicit specializers.  */
    protected String mmSignature() {
	return CType.genMultimethodSignature( returnType, receiverType, 
					      parameters, isExternal() );
    }
    
    protected String mmGenericSignature() {
	return CType.genGenericMethodSignature(getTypeVariable(),returnType,
					       receiverType,parameters, 
					       isExternal());
    }
    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Generate the method information for this for constructing a
     * class file [JVM2 4.6]. 
     *
     * @return an ArrayList of <code>MethodInfo</code>, just a single element
     * here, but overridden in <Code>CSourceDispatcherMethod</code> to include 
     * separate elements for the dispatcher and each multimethod body
     */
    public ArrayList genMethodInfo() {
	fail( "genMethodInfo() is not supported for an abstract CMethod" );
	//@ unreachable;
	return null;
    }

    /**
     * Generates the generic function information for this required
     * when generating a class file, or returns null if this method
     * requires no such information.  */
    public GenericFunctionInfo genGenericFunctionInfo() {
	fail( "genGenericFunctionInfo() is not supported for an abstract CMethod" );
	//@ unreachable;
	return null;
    }

    public CMethod getGenerateFrom() {
	return generatefrom;
    }
    
    public void setGenerateFrom(CMethod me) {
	generatefrom = me;
    }
					
    public boolean isSynthetic() {
	return synthetic;
    }

    public void setSynthetic(boolean syn) {
	synthetic = syn;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private CClassType		receiverType;
    private CType		returnType;
    private /*@ spec_protected @*/ CSpecializedType[]	parameters;
    private CClassType[]	exceptions;

    /**
     * The context in which this method was declared.  Used for
     * resolving the top method of the generic function to which this
     * belongs. */
    //@ spec_public
    protected CContextType declarationContext;

    /**
     * The top concrete method of the generic function to which this belongs.
     */
    //@ spec_public
    protected CMethod topConcreteMethod;

    /**
     * If this is a top method, we record which arguments of the
     * generic function are specialized.  An argument is
     * specialized if there is some method in the generic function
     * that specializes on it.
     */
    protected BitSet specArgs;

    /**
     * Indicates whether this method is pleomorphic with one external
     * generic function and one or more internal generic functions.
     * */
    private boolean isMixedPleomorphic = false;

    /**
     * Indicates whether this method should have its privacy modifier
     * removed for the purposes of code generation.  This is the case
     * if the method is private and either a constructor or in an
     * external gf, but it is also accessed by an enclosing or nested
     * class.  We should really probably use javac-style accessor methods,
     * as we do for all other kinds of privileged access scenarios.
     **/
    public boolean needsPrivacyModifierRemoved = false;

	/** Integers used in bodyIdent() to give a unique name
	 *  to a multimethod's associated $body method.  This is needed
	 *  to handle value dispatching, where the different multimethods
	 *  have the same type signature, so static overloading cannot be
	 *  used to disambiguate them.  The gfID is shared for all methods
	 *  of the same gf declared in the same class.  The methodID is
	 *  unique to this method (as compared to the methodIDs of other
	 *  methods of this gf declared in the same class).
	 **/
    private int gfID;
    private int methodID;
    private CTypeVariable[] typevariables;
    protected CMethod generatefrom = null;
    public boolean synthetic = false;
}
