/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CMethodContext.java,v 1.14 2006/12/11 05:55:56 chalin Exp $
 */

package org.multijava.mjc;

import java.util.Hashtable;
import java.util.Enumeration;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

/**
 * This class represents the context for a method during checking
 * passes (checkInterface, checkInitializers, typecheck).
 * @see CContextType 
 */
public class CMethodContext extends CContext implements CMethodContextType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Create a context in which to check a method.
     * Clients should not call this but should instead call
     * <code>CClassContextType.createMethodContext</code>.
     *
     * <pre><jml>
     * requires (* \caller instanceof CContextType *);
     * </jml></pre>
     *
     * @param	parent		the parent context
     * @param	self		the corresponding method interface
     *
     * @see CClassContextType#createMethodContext(CMethod)
     */
    public CMethodContext( CContextType parent, CMethod self ) {
	super(parent);
	this.self = self;
    }

    /**
     * Verifies that all checked exceptions are defined in the throw
     * list.
     *
     * @exception PositionedError if checks fail */
    public void verifyExceptions(TokenReference ref) throws PositionedError {
	Iterator iter = throwables.iterator();
	CClassType[]	checked = self.throwables();
	boolean[]		used = new boolean[checked.length];
    
    loop:
	while (iter.hasNext()) {
	    CThrowableInfo thrown = (CThrowableInfo)iter.next();
	    CClassType type = thrown.getThrowable();
      
	    // only checked exceptions need to be checked
	    if (! type.isCheckedException()) {
		continue loop;
	    }
      
	    for (int j = 0; j < checked.length; j++) {
		if (type.isAlwaysAssignableTo(checked[j])) {
		    used[j] = true;
		    continue loop;
		}
	    }
	    throw new PositionedError(thrown.getLocation().getTokenReference(),
				      MjcMessages.METHOD_UNCAUGHT_EXCEPTION,
				      type,
				      null);
	}
    
	for (int i = 0; i < checked.length; i++) {
	    if (!checked[i].isCheckedException()) {
		reportTrouble(
		     new CWarning(ref,
				  MjcMessages.METHOD_UNCHECKED_EXCEPTION,
				  checked[i],
				  null));
	    } else if (!used[i]) {
		reportTrouble(
		     new CWarning(ref,
				  MjcMessages.METHOD_UNTHROWN_EXCEPTION,
				  checked[i],
				  null));
	    }
	}
    }

    public /*@nullable*/ CTypeVariable lookupTypeVariable(String ident) throws UnpositionedError {
	CTypeVariable tv = getCMethod().lookupTypeVariable(ident);
	if(tv !=null) { 
	    return tv;
	} else { 
	    if (isStatic()) {
		return null;
	    } else {
		return getClassContext().lookupTypeVariable(ident);
	    }
	}
    }
    // ----------------------------------------------------------------------
    // ACCESSORS (NEW CONTEXT)
    // ----------------------------------------------------------------------

    public CFlowControlContextType createFlowControlContext(int localVars, TokenReference where) {
	return createFlowControlContext(localVars, false, where);
    }

    public CFlowControlContextType createFlowControlContext(
	int localVars, boolean isInExternalGF, TokenReference where) 
    {
	return new CFlowControlContext(this, localVars, isInExternalGF, where);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the signature of the method declaration in which this
     * context is enclosed, or null if this context is not enclosed in
     * a method declaration.  */
    //@ pure
    public CMethod getCMethod() {
	return self;
    }

    /**
     * Returns the nearest surrounding context of type CMethodContextType.  */
    public CMethodContextType getMethodContext() {
	return this;
    }

    /**
     * Indicates whether this context is enclosed in an instance or
     * static initializer.  */
    public boolean isInInitializer() {
	return false;
    }

    /**
     * Indicates whether this context is enclosed in a constructor.  */
    public boolean isInConstructor() {
	return false;
    }

    /**
     * Indicates whether this context is "static".
     *
     * @return true iff this context is enclosed in a context for a
     * static initializer or static method.  */
    public boolean isStatic() {
	return self.isStatic();
    }


    /**
     * Indicates whether this context is "pure".
     *
     * @return true iff this context is enclosed in a pure method.
     */
    //@ also
    //@ assignable objectState;
    public boolean isPure() {
	return self.isPure();
    }


    // ----------------------------------------------------------------------
    // THROWABLES
    // ----------------------------------------------------------------------

    /**
     * Registers that the given throwable can be thrown within this
     * context.  The implementation passes the throwable up through
     * the nested contexts until a surrounding context is found that
     * can handle throwables.  For example, a CMethodContextType
     * instance collects the throwables that are uncaught within a
     * method body, while a CTryContext collects the throwables that are
     * thrown within a try-block and are therefore subject to handling
     * by associated catch-blocks.
     * @param throwable the type of the new throwable */
    public void addThrowable(CThrowableInfo throwable) {
	throwables.add(throwable);
    }

    /**
     * @return the list of exception that may be thrown
     */
    public Set throwables() {
	return throwables;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	HashSet	throwables = new HashSet();
    protected	Hashtable	labels;		// Hashtable<String, String>
    private	CMethod		self;  //@ in objectState;
                                  //@ maps self.objectState \into objectState;
}
