/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CMethodSet.java,v 1.12 2004/06/12 14:43:36 leavens Exp $
 */

package org.multijava.mjc;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.multijava.util.Destination;
import org.multijava.util.classfile.ClassFileFormatException;
import org.multijava.util.classfile.GenericFunctionInfo;
import org.multijava.util.classfile.InnerClassInfo;
import org.multijava.util.classfile.MethodInfo;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class encapsulates a collection of method signatures.  It is
 * used to track the type signatures of methods declared in a class
 * and to track the apparent methods of a class as augmented by
 * external methods and as viewed by another compilation unit.  It
 * includes facilities for accumulating sets of methods that meet some
 * specified criteria, for example, methods that are specialized by a given
 * method.
 *
 * @see CClass
 * @see CAugmentationMap */
public class CMethodSet extends CAbstractMethodSet implements Constants {
    // !FIXME! There may be opportunities to improve the efficiency of
    // this by using an implementation that considers the subtype
    // relationships of the methods' tuples of specializers.  

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------
    public CMethodSet( CMethod[] methodArray ) {
	super( methodArray );
	if( methodArray.length < MIN_HASH_SIZE ) {
	    methodIdents = null;
	} else {
	    methodIdents = new HashSet( methodArray.length + 1, 1.1f );
	    for( int i=0; i < methodArray.length; i++ ) {
		methodIdents.add( methodArray[i].ident() );
	    }
	}
    }

    public CMethodSet() {
	super();
	methodIdents = null;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Processes the methods in this set using the given strategy and
     * mutates the accumulator according to the strategy.  Examples of
     * the use of this method can be found in CClass.
     *
     * @param	actor		the strategy used for selecting methods
     * @param	ident		the identifier of the methods to be selected,
     *				used in an efficiency hack for avoiding a 
     *				linear search in cases where such a search is 
     *				certain to fail
     * @param	accum		an accumulator for the result, is mutated
     * @exception	UnpositionedError	at the discretion of the
     * strategy */
    public void accumMethodSet( Strategy actor, String ident, 
				CMethodSet accum,
				CClassType[] args )
				
	throws UnpositionedError
    {
	if( methodIdents != null &&
	    ident != JAV_CONSTRUCTOR &&	// these will always be in the set...
	    ident != JAV_INIT &&        // ...so it is more efficient to...
	    ident != JAV_STATIC_INIT && // ...not look for them
	    !methodIdents.contains( ident ) ) {
	    return;		// leaving accum unchanged
	}

	int thisSize = size();
	for( int i=0; i < thisSize; i++ ) {
	    CMethod current = getMethod( i );
	    if ( current.isSynthetic()) {
		    continue;
	    }
	    if( actor.maybeInclude( current, args ) ) {
		boolean ignoreCurrent = false;
		Iterator iter = accum.iterator();
		while( iter.hasNext() ) {
		    CMethod accumMethod = iter.next();
		    byte r = actor.resultFor( current, accumMethod );
		    ignoreCurrent = ignoreCurrent 
			|| (r & Strategy.IGNORE_CURRENT)!=0;
		    if( (r & Strategy.REMOVE_ACCUM) != 0 ) {
			iter.remove();
		    }
		    if( (r & Strategy.DONE_WITH_ACCUM) != 0 ) {
			break;
		    }
		}

		if( !ignoreCurrent ) {
		    accum.addMethod( current );
		}

	    }
	}
    }

    /**
     * Represents the interface for strategies used in accumulating
     * method sets.  Typically a client of <code>CMethodSet</code>
     * will create an anonymous class of type <code>Strategy</code>
     * that is passed to <code>accumMethodSet</code> for adding some
     * subset of the methods of the <code>CMethodSet</code> to an
     * accumulator.
     *
     * @see CMethodSet#accumMethodSet(CMethodSet.Strategy,String,CMethodSet) */
    public interface Strategy {
	/**
	 * Indicates whether the given <code>candidate</code> method
	 * should be checked against the methods already in the
	 * accumulator for possible inclusion in the accumulator.
	 *
	 * @param	candidate	the candidate method for inclusion in
	 *				the accumulator
	 * @return	true if the method should be considered
	 * @exception	UnpositionedError usually the implementer
	 * will want to use <code>CContextType.reportTrouble()</code> for
	 * error reporting to avoid throwing an exception, the exception is
	 * allowed here for flexibility  */
	boolean maybeInclude( CMethod candidate, CClassType[] args) 
	    throws UnpositionedError;

	/**
	 * Checks the given <code>candidate</code> method against a
	 * single method in the accumulator and decides whether the
	 * candidate should be added to the accumulator, whether the
	 * method from the accumulator should be removed, and whether
	 * additional checks against the remaining methods in the
	 * accumulator are necessary.
	 *
	 * @param	candidate	the candidate method for inclusion in
	 *				the accumulator
	 * @param	accumMethod	a single method from the accumulator
	 * @return	a bit mask indicating the actions to be taken
	 * @exception	UnpositionedError if a problem is detected in the 
	 * source code.  */
	byte resultFor( CMethod candidate, CMethod accumMethod ) 
	    throws UnpositionedError;

	/**
	 * The bit mask representing that the <code>candidate</code>
	 * method passed to <code>resultFor()</code> should be
	 * ignored.
	 *
	 * @see #resultFor(CMethod,CMethod) */
	byte IGNORE_CURRENT = 0x01;

	/**
	 * The bit mask representing that the <code>accumMethod</code>
	 * method passed to <code>resultFor()</code> should be
	 * removed from the accumulator.
	 *
	 * @see #resultFor(CMethod,CMethod) */
	byte REMOVE_ACCUM = 0x02;

	/**
	 * The bit mask representing that no more checks of the
	 * <code>candidate</code> method need to be made against the
	 * accumulator by calls to <code>resultFor()</code>.
	 *
	 * @see #resultFor(CMethod,CMethod) */
	byte DONE_WITH_ACCUM = 0x04;
    }

    /**
     * Constructs a bit-mask for implementers of the
     * <code>resultFor()</code> method of <code>Strategy</code>.
     *
     * @see CMethodSet.Strategy
     * @see CMethodSet.Strategy#resultFor(CMethod,CMethod)
     * @see CMethodSet.Strategy#IGNORE_CURRENT
     * @see CMethodSet.Strategy#REMOVE_ACCUM
     * @see CMethodSet.Strategy#DONE_WITH_ACCUM
     */
    public static byte makeResult( boolean ignoreCurrent, 
				   boolean remove, boolean done ) 
    {
	return (byte) ((ignoreCurrent ? Strategy.IGNORE_CURRENT : 0 ) 
	    + (remove ? Strategy.REMOVE_ACCUM : 0 )
	    + (done ? Strategy.DONE_WITH_ACCUM : 0 ));
    }

    /**
     * Appends all the abstract methods in this collection to the given list.
     *
     * @param	v		mutated to include the CMethods for all
     *				abstract methods in this collection
     */
    public void appendAbstractMethodsTo( ArrayList v ) {
	int size =  size();
	for( int i=0; i < size; i++ ) {
	    CMethod current = getMethod( i );

	    if( current.isAbstract() ) {
		v.add( current );
	    }
	}
    }
    
    public void appendAbstractMethodsTo( ArrayList v, CClassType[] args ) {
	int size =  size();
	for( int i=0; i < size; i++ ) {
	    CMethod current = getMethod( i );
	    
	    if( current.isAbstract() ) {
		MethodArgsPair pair = new MethodArgsPair(current, args);
		v.add( pair );
	    }
	}
    }

    /**
     * Appends all the  methods in this collection to the given list.
     *
     * @param	v		mutated to include the CMethods for all
     *				methods in this collection
     */
    public void appendMethodsTo( ArrayList v ) {
	int size =  size();
	for( int i=0; i < size; i++ ) {
	    v.add( getMethod( i ) );
	}
    }
    //inner class
    public static class MethodArgsPair {
	CMethod me;
	CClassType[] args;
	MethodArgsPair(CMethod m, CClassType[] ar) {
	    this.me = m;
	    this.args = ar;
	}
	public CMethod getMethod() {
	    return me;
	}
	public CClassType[] getArgs() {
	    return args;
	}
	 
    }
    public void appendMethodsTo( ArrayList v, CClassType[] args) {
	int size = size();
	for( int i = 0; i< size; i++) {
	    MethodArgsPair pair = new MethodArgsPair(getMethod(i),args);
	    v.add(pair);
	}
    }

    public String toString() {
	return "CMethodSet" + super.toString();
    }

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    /**
     * Generates an array of method info for the methods in this collection.
     */
    public MethodInfo[] buildMethodInfo() {
	int size =  size();

	ArrayList result = new ArrayList( size );
	
	for( int i=0; i < size; i++ ) {
	    ArrayList mi = getMethod( i ).genMethodInfo();
	    result.addAll( mi );
	    
	}

	MethodInfo[] rA = new MethodInfo[ result.size() ];
	result.toArray( rA );
	
	return rA;
    }

    public void buildGenericFunctionInfo(ArrayList accum) {
	int size = size();
	
	for( int i=0; i < size; i++ ) {
	    GenericFunctionInfo gfInfo = getMethod(i).genGenericFunctionInfo();
	    if (gfInfo != null) {
		accum.add(gfInfo);
	    }
	}
    }

    //---------------------------------------------------------------------
    // STATIC MEMBERS
    //---------------------------------------------------------------------

    /**
     * The minimum size of the collection for which a
     * <code>Hashset</code> will be instantiated to track method
     * identifiers.  */
    private static int MIN_HASH_SIZE = 10;
    
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * A set of the identifiers of methods present in this collection.
     * This is used for efficient lookup; a linear search of
     * <code>methodArray</code> is only performed if this set
     * indicates that the search will be successful.  The search is
     * still needed since many lookup operations will yield sets of
     * applicable methods.  Also note that for small collections this
     * set is just left as null as it is more efficient in those cases
     * to just perform the linear search.  */
    private Set			methodIdents;

    private static final TokenReference NO_TOKEN = TokenReference.NO_REF;
}
