/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CModifier.java,v 1.13 2006/02/11 03:05:57 chalin Exp $
 */

package org.multijava.mjc;

import java.util.HashMap;
import java.util.ArrayList;

import org.multijava.util.compiler.ModifierUtility;


/**
 * This class provides utilities for operating on modifier bit masks.  
 * 
*/
public class CModifier extends org.multijava.util.Utils 
    implements Constants, ModifierUtility
{

/*@
 @ public invariant ( \forall int i; (\exists int j; 0<=j && j<codes.length;
 @				      	  i == codes[j] );
 @			  codes[((Integer) codesMap.get( new Integer( i ) )).
 @				    intValue()] == i );
 @ public invariant_redundantly (* codesMap maps from code values to the 
 @				   position of the corresponding code value  
 @				   in the codes array *);
 @*/
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Instantiate the modifier utilities for the default modifier 
     * bit-masks and names.  Equivalent to using constructor:
     *
     * <pre>
     * CModifier( Constants.ACCESS_FLAG_ARRAY, 
     *            Constants.ACCESS_FLAG_NAMES );
     * </pre>
     *
     * @see org.multijava.util.classfile.Constants
     * @see org.multijava.util.classfile.Constants#ACCESS_FLAG_ARRAY
     * @see org.multijava.util.classfile.Constants#ACCESS_FLAG_NAMES
     */
    public CModifier() {
	this( Constants.ACCESS_FLAG_ARRAY, Constants.ACCESS_FLAG_NAMES );
    }
    
    /**
     * Instantiate the modifier utilities for the given modifier 
     * bit-masks and names.
     *
     * @param	codes	an array of modifier bit-masks, typically from the
     *			<code>ACCESS_FLAG_ARRAY</code> field of the 
     *			appropriate <code>Constants</code> interface.
     * @param	names	an array of modifier names, typically from the
     *			<code>ACCESS_FLAG_NAMES</code> field of the 
     *			appropriate <code>Constants</code> interface.
     * @see org.multijava.util.classfile.Constants
     * @see org.multijava.util.classfile.Constants#ACCESS_FLAG_ARRAY
     * @see org.multijava.util.classfile.Constants#ACCESS_FLAG_NAMES
     */
    public CModifier( long[] codes, String[] names ) { 
	this.codes = codes;
	this.names = names;
	calculateCodesMap();
    }
    
    private void calculateCodesMap() {
	codesMap = new HashMap();
	for( int i=0; i<codes.length; i++ ) {
	    codesMap.put( new Long( codes[i] ), new Integer( i ) ); 
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------
    
    /** Returns the codes of modifiers */
    public long[] codes() { 
	return codes;
    }

    /** Returns the names of modifiers */
    public String[] names() {
	return names;
    }

    //---------------------------------------------------------------------
    // OPERATORS
    //---------------------------------------------------------------------

    /**
     * Returns a string with the names of all the modifiers in the
     * preferred order.  */
    public String asString( long modifiers ) { 
	String[] resultArr = asStrings( modifiers );
	StringBuffer result = new StringBuffer();
	for (int i = 0; i < resultArr.length; i++) {
	    result.append( resultArr[i] + " " );
	}
	return result.toString();
    }

    /**
     * Returns an array of the names of all the modifiers in the
     * preferred order. */
    public String[] asStrings( long modifiers ) { 
	ArrayList results = new ArrayList();
	for (int i = 0; i < codes.length; i++) {
	    if( hasFlag( modifiers, codes[i] ) && names[i].length() > 0 ) {
		results.add( names[i] );
	    }
	}
	String[] resultArr = new String[ results.size() ];
	results.toArray( resultArr );
	
	return resultArr;
    }

    /**
     * Returns true if no modifiers in <code>previousMods</code> are
     * supposed to come after the modifier in <code>newMod</code>. */
    public boolean inPreferredOrder( long previousMods, long newMod ) { 
	return highestPos( previousMods ) < modPos( newMod );
    }

    /**
     * Returns the name of the given modifier.
     *
     * <pre><jml>
     * also
     * {|
     *    requires modPos( mod ) != -1;
     *    ensures \result.equals( names[ modPos( mod ) ] );
     * also
     *    requires modPos( mod ) == -1;
     *    signals_only IllegalArgumentException;
     * |}
     * </jml></pre>
     */
    public String nameOf( long mod ) { 
	int pos = modPos( mod );
	if( pos == -1 ) {
	    throw new IllegalArgumentException( "argument must denote exactly one modifier, given argument denotes " + asString( mod ) );
	} else {
	    return names[pos];
	}
    }

    /**
     * Returns the highest of index from <code>codes</code> of a
     * modifier flag that is set in <code>mods</code>.
     *
     * <pre><jml>
     * ensures \result == ( \max int i; 
     *                          0 <= i && i < codes.length && 
     *					hasFlag( mods, codes[i] ); i );
     * </jml></pre>
     *
     * @param mods a modifier bit-mask */
    private int highestPos( long mods ) { 
	int max = -1;

	for (int pos = 0; pos < codes.length; pos++) {
	    if ( hasFlag( mods, codes[pos] ) ) {
		max = pos;
	    }
	}

	return max;
    }

    /**
     * Returns the index of <code>mod</code> in <code>codes</code>.
     *
     * <pre><jml>
     * {|
     *    requires (\exists int i; 0 <= i && i < codes.length; codes[i]==mod );
     *    ensures codes[\result] == mod;
     * also
     *    requires !(\exists int i; 0 <= i && i < codes.length; codes[i]==mod);
     *    ensures \result == -1;
     * |}
     * </jml></pre>
     */
    private /*@ spec_public pure @*/ int modPos( long mod ) { 
	Object maybeResult = codesMap.get( new Long( mod ) ); 
	if( maybeResult == null ) {
	    return -1;
	} else {
	    return ((Integer) maybeResult).intValue(); 
	}
    }

    private long stripInternalMods( long mod ) {
	return mod & ACC_MODIFIER_FLAGS_MASK;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected /*@ spec_public @*/ String[] names;
	    
    protected /*@ spec_public @*/ long[] codes; 

    /**
     * Maps from code values to preferred positions.
     */
    private /*@ spec_public @*/ HashMap codesMap;
}
