/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CNumericType.java,v 1.24 2004/03/21 19:14:49 davidcok Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.FastStringBuffer;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents java and kopi numericals types
 * Such as byte, short, int, long, float, double
 * @see	CType
 */
public class CNumericType extends CType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructor
     * @param	typeID		the ident (int value) of this type
     */
    /*@ protected normal_behavior
      @   requires isValidNumericTypeID(typeID);
      @   assignable type;
      @*/
    protected CNumericType(int typeID) {
	super(typeID);
	assertTrue(isValidNumericTypeID(typeID));
    }

    /*@ protected normal_behavior
      @   assignable \nothing;
      @   ensures
      @     \result <== (typeID == TID_BYTE ||
      @     typeID == TID_CHAR ||
      @     typeID == TID_DOUBLE ||
      @     typeID == TID_FLOAT ||
      @     typeID == TID_INT ||
      @     typeID == TID_LONG ||
      @     typeID == TID_SHORT);
      @*/
    protected /*@ pure @*/ boolean isValidNumericTypeID(int typeID) {
	return typeID == TID_BYTE ||
		typeID == TID_CHAR ||
		typeID == TID_DOUBLE ||
		typeID == TID_FLOAT ||
		typeID == TID_INT ||
		typeID == TID_LONG ||
		typeID == TID_SHORT;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Transforms this type to a string
     */
    public String toString() {
	switch (type) {
	case TID_BYTE:
	    return "byte";
	case TID_CHAR:
	    return "char";
	case TID_DOUBLE:
	    return "double";
	case TID_FLOAT:
	    return "float";
	case TID_INT:
	    return "int";
	case TID_LONG:
	    return "long";
	case TID_SHORT:
	    return "short";
	default:
	    fail();
	    return null;
	}
    }

        
    /**
     * Transforms this type to a string
     */
    public String getSignature() {
	switch (type) {
	case TID_BYTE:
	    return "B";
	case TID_CHAR:
	    return "C";
	case TID_DOUBLE:
	    return "D";
	case TID_FLOAT:
	    return "F";
	case TID_INT:
	    return "I";
	case TID_LONG:
	    return "J";
	case TID_SHORT:
	    return "S";
	default:
	    fail();
	    return null;
	}
    }

    /**
     * Transforms this type to a string
     */
    public void appendSignature(FastStringBuffer buff) {
	char	c;

	switch (type) {
	case TID_BYTE:
	    c = 'B';
	    break;
	case TID_CHAR:
	    c = 'C';
	    break;
	case TID_DOUBLE:
	    c = 'D';
	    break;
	case TID_FLOAT:
	    c = 'F';
	    break;
	case TID_INT:
	    c = 'I';
	    break;
	case TID_LONG:
	    c = 'J';
	    break;
	case TID_SHORT:
	    c = 'S';
	    break;
	default:
	    c = '!';
	    fail();
	}
	buff.append(c);
    }

    /**
     * Returns the size used in stack by value of this type
     */
    public int getSize() {
	return (type == TID_LONG || type == TID_DOUBLE) ? 2 : 1;
    }

    /**
     * Check if a type is a numeric type
     * @return is it a numeric type ?
     */
    //@ pure
    public boolean isNumeric() {
	return true;
    }

    /**
     * Check if this type is ... [TBC]
     * @return TBC ?
     */
    //@ pure
    public boolean isCodedAsInt() {
	switch (type) {
	case TID_BYTE:
	case TID_CHAR:
	case TID_INT:
	case TID_SHORT:
	    return true;
	}
	return false;
    }

    /**
     * Check if this type is an integral type.
     * @return is this a integral type ?
     */
    //@ pure
    public boolean isOrdinal() {
	switch (type) {
	case TID_BYTE:
	case TID_CHAR:
	case TID_SHORT:
	case TID_INT:
	case TID_LONG:
	    return true;
	default:
	    return false;
	}
    }

    /**
     * Check if this type is a floating point type.
     * @return is this a floating point number type ?
     */
    //@ pure
    public boolean isFloatingPoint() {
	switch (type) {
	case TID_FLOAT:
	case TID_DOUBLE:
	    return true;
	default:
	    return false;
	}
    }

    // ----------------------------------------------------------------------
    // BODY CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks that this type is valid, throwing an exception if it is
     * not.  For class types this check resolves type names into their
     * fully qualified versions and sets the class parameter so that
     * getCClass returns the appropriate singleton.
     *
     * @param		context		the context in which this type was
     *					declared, may be null
     * @exception	UnpositionedError	this unpositioned error
     *						should be positioned
     *						by the calling node of
     *						the AST, typically by
     *						calling the
     *						addPosition method of
     *						UnpositionedError 
     */
    public CType checkType(CContextType context)
	throws UnpositionedError {
	return this;
    }

    
    /**
     * Compute the value of a specializer expression used to dispatch on
     * a compile-time constant.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    public void resolveValueType(CExpressionContextType context)
	throws PositionedError {}

    /** Returns true iff <code>from</code> type can be converted by widening
     * primitive conversion to <code>this</code> type. [JLS2 5.1.2] */
    /*@ public normal_behavior
      @  requires true;
      @  ensures  \result <== (*
      @   (from is byte && this is short, int, long, float, or double) ||
      @   (from is short && this is int, long, float, or double) ||
      @   (from is char && this is int, long, float, or double) ||
      @   (from is int && this is long, float, or double) ||
      @   (from is long && this is float or double) ||
      @   (from is float && this is double) *)
      @  && !\result <== (*
      @    (from is byte && this is char) ||
      @    (from is short && this is char) 
      @  *);
      @*/
    public /*@pure@*/ boolean widening_primitive_conv_from(/*@non_null@*/ CType from) {
	if (!from.isNumeric() || from.isValueType()) {
	    return false;
	}
	return (from == CStdType.Byte  && this == CStdType.Char ||
		from == CStdType.Short && this == CStdType.Char)
	    ? false
	    : from.getTypeID() < type;
    }
	
    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2] [JLS2 5.1.2].  The only conversions
     * checked here are the widening primitive conversions and
     * identity conversion.  Narrowing primitive conversions are
     * treated as a special case in JExpression.isAssignableTo()
     *
     * @param	dest		the destination type
     * @return true iff assignment is valid 
     * @see org.multijava.mjc.JExpression#isAssignableTo(CType)
     */
    /*@ also public normal_behavior
      @  requires true;
      @  assignable objectState; // copied from superclass.
      @  ensures  \result <==> 
      @    (* this type can be converted to dest by identity
      @       conversion or widening primitive conversion *);
      @*/
    public boolean isAlwaysAssignableTo(CType dest) {
	if (dest == this) {
	    return true; // identity conversion.
	}
	if( !dest.isNumeric() || dest.isValueType()) {
	    return false;
	}
	return ((CNumericType)dest).widening_primitive_conv_from(this);
    }

    /**
     * isCastableTo
     * @param dest  the destination type
     * @return  cast is valid ?
     */
    public boolean isCastableTo(CType dest) {
	return dest.isNumeric();
    }

    /**
     * implicitlyCastTo
     * @param	dest	the expression type to cast to
     * @return	cast is valid ?
     */
    /*@ also public normal_behavior
      @  ensures  \result <==>
      @    dest.isNumeric() && getTypeID() <= dest.getTypeID() &&
      @    (getTypeID() == TID_SHORT ==> dest.getTypeID() != TID_CHAR);
      @*/
    public /*@pure@*/ boolean implicitlyCastTo(/*@non_null@*/CType dest) {
	return this == dest ||
	    dest.isNumeric() &&
	    ((CNumericType)dest).widening_primitive_conv_from(this);
	/*
	return 
	    this == dest ||
	    (dest.isNumeric() && type < dest.type &&
	     (type == TID_SHORT ? dest.getTypeID() != TID_CHAR : true));
	*/
	/* original code:
	if (this == dest) {
	    return true;
	} else if (!dest.isNumeric()) {
	    return false;
	} else {
	    switch (type) {
	    case TID_BYTE:
		if (dest.type == TID_SHORT || dest.type == TID_CHAR) {
		    return true;
		}
	    case TID_SHORT:
	    case TID_CHAR:
		if (dest.type == TID_INT) {
		    return true;
		}
	    case TID_INT:
		if (dest.type == TID_LONG) {
		    return true;
		}
	    case TID_LONG:
		if (dest.type == TID_FLOAT) {
		    return true;
		}
	    case TID_FLOAT:
		if (dest.type == TID_DOUBLE) {
		    return true;
		}
	    case TID_DOUBLE:
		break;
	    default:
		fail();
	    }
	    return false;
	}
	*/
    }

    /** 
     * This isn't used and isn't right at any rate.
     *  
     * Search the type corresponding to the unification of the two types.  !FIXME!
     * @param t1 the first type
     * @param t2 the second type
     * @return the corresponding type or null
     */
    public static CType unifyNumeric(CType t1, CType t2) {
	assertTrue(t1.isNumeric() && t2.isNumeric());
	return CNumericType.unifyNumericIn(t1, CNumericType.unifyNumericIn(t2, t1));
    }

    /**
     * unifyNumeric
     * search the type corresponding to the unification of the two types
     * @param t1 the first type
     * @param t2 the second type
     * @return the corresponding type or null
     */
    private static CType unifyNumericIn(CType t1, CType t2) {
	if (t1 == CStdType.Short && t2 == CStdType.Byte) {
	    return CStdType.Short;
	} else if (t1 == CStdType.Char && (t2 == CStdType.Char || t2 == CStdType.Byte)) {
	    return CStdType.Char;
	} else {
	    return null;
	}
    }

    /**
     * unaryPromote
     * search the type corresponding to the type after computation
     * @return the corresponding type after operation
     */
   public CType unaryPromote() {
	return (this.getTypeID() <= TID_INT) ? CStdType.Integer : this;
    }

    /**
     * binaryPromote
     * search the type corresponding to the promotion of the two types (this and other)
     * @param other the second type
     * @return the corresponding type or null
     */
    public CType binaryPromote(CType other) {
        assertTrue(other.isNumeric());
        CType maxType = (this.getTypeID() > other.getTypeID()) ? this : other;
        return (maxType.getTypeID() <= TID_INT) ? CStdType.Integer : maxType;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Generates a sequence of bytecodes
     * @param	dest		the destination type
     * @param	code		the code list
     */
    public void genCastTo(CNumericType dest, CodeSequence code) {
	if (dest != this) {
	    switch (dest.type) {
	    case TID_BYTE:
		genCastTo(CStdType.Integer, code);
		code.plantInstruction(new NoArgInstruction(opc_i2b));
		break;

	    case TID_CHAR:
		genCastTo(CStdType.Integer, code);
		code.plantInstruction(new NoArgInstruction(opc_i2c));
		break;

	    case TID_SHORT:
		genCastTo(CStdType.Integer, code);
		code.plantInstruction(new NoArgInstruction(opc_i2s));
		break;

	    case TID_INT:
		switch (type) {
		case TID_BYTE:
		case TID_CHAR:
		case TID_SHORT:
		    break;
		case TID_INT:
		    fail();
		case TID_LONG:
		    code.plantInstruction(new NoArgInstruction(opc_l2i));
		    break;
		case TID_FLOAT:
		    code.plantInstruction(new NoArgInstruction(opc_f2i));
		    break;
		case TID_DOUBLE:
		    code.plantInstruction(new NoArgInstruction(opc_d2i));
		    break;
		default:
		    fail();
		}
		break;

	    case TID_LONG:
		switch (type) {
		case TID_BYTE:
		case TID_CHAR:
		case TID_SHORT:
		case TID_INT:
		    code.plantInstruction(new NoArgInstruction(opc_i2l));
		    break;
		case TID_LONG:
		    fail();
		case TID_FLOAT:
		    code.plantInstruction(new NoArgInstruction(opc_f2l));
		    break;
		case TID_DOUBLE:
		    code.plantInstruction(new NoArgInstruction(opc_d2l));
		    break;
		default:
		    fail();
		}
		break;
	
	    case TID_FLOAT:
		switch (type) {
		case TID_BYTE:
		case TID_CHAR:
		case TID_SHORT:
		case TID_INT:
		    code.plantInstruction(new NoArgInstruction(opc_i2f));
		    break;
		case TID_LONG:
		    code.plantInstruction(new NoArgInstruction(opc_l2f));
		    break;
		case TID_FLOAT:
		    fail();
		case TID_DOUBLE:
		    code.plantInstruction(new NoArgInstruction(opc_d2f));
		    break;
		default:
		    fail();
		}
		break;
	
	    case TID_DOUBLE:
		switch (type) {
		case TID_BYTE:
		case TID_CHAR:
		case TID_SHORT:
		case TID_INT:
		    code.plantInstruction(new NoArgInstruction(opc_i2d));
		    break;
		case TID_LONG:
		    code.plantInstruction(new NoArgInstruction(opc_l2d));
		    break;
		case TID_FLOAT:
		    code.plantInstruction(new NoArgInstruction(opc_f2d));
		    break;
		case TID_DOUBLE:
		    fail();
		default:
		    fail();
		}
		break;

	    default:
		fail();
	    }
	}
    }
}
