/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CParseClassContext.java,v 1.6 2005/05/27 12:52:25 chalin Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Stack;

/**
 * This class is used by the parser to collect the members of a class
 * declaration.  For efficiency (and to avoid memory leaks caused by
 * poor garbage collection in the JVM) a factory method is used to
 * generate instances and old instances are store on a stack for
 * reuse.
 */
public class CParseClassContext {

    /**
     * Factory method returns an instance of CParseClassContext.  The
     * instance is recycled from the stack of discarded instances or a
     * new instance is minted if the stack is empty.
     * @return	a "new" CParseClassContext instance
     */
    public static CParseClassContext getInstance() {
	return stack.size() == 0 ?
	    new CParseClassContext() :
	    (CParseClassContext)stack.pop();
    }

    /**
     * Calls the static method <code>release(this)</code>
     */
    public void release() {
	release(this);
    }

    /**
     * Erases the data stored in <code>context</code> and pushes the
     * instance onto a stack for recycling.
     * @param	context	a used instance to be recycled
     */
    public static void release(CParseClassContext context) {
	context.clear();
	stack.push(context);
    }

    /**
     * Hides the default constructor except to subclasses.
     */ 
    protected CParseClassContext(){}

    /**
     * Prepares this instance for recycling by clearing all the data
     * stored in its ArrayLists.
     */
    protected void clear() {
	methods.clear();
	inners.clear();
	fieldsAndInits.clear();
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (ADD)
    // ----------------------------------------------------------------------

    public void addFieldDeclaration(JFieldDeclarationType decl) {
	fieldsAndInits.add(decl);
    }

    public void addMethodDeclaration(JMethodDeclarationType decl) {
	methods.add(decl);
    }
 
    public void addInnerDeclaration(JTypeDeclarationType decl) {
	inners.add(decl);
    }

    public void addBlockInitializer(JClassBlock block) {
	fieldsAndInits.add(block);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (GET)
    // ----------------------------------------------------------------------

    public ArrayList getMethods() {
	return (ArrayList) methods.clone();
    }

    public ArrayList getInnerClasses() {
	return (ArrayList) inners.clone();
    }

    public JPhylum[] getFieldsAndInits() {
	JPhylum[] temp = new JPhylum[fieldsAndInits.size()];

	return (JPhylum[]) fieldsAndInits.toArray(temp);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private /*@non_null*/ ArrayList methods = new ArrayList();
    private /*@non_null*/ ArrayList inners = new ArrayList();
    private /*@non_null*/ ArrayList fieldsAndInits = new ArrayList();

    private static /*@non_null*/ Stack stack = new Stack();
}
