/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CParseCompilationUnitContext.java,v 1.10 2004/01/16 23:21:09 cclifton Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Stack;

/**
 * This class is used by the parser to collect the contents of a
 * compilation unit.  For efficiency (and to avoid memory leaks caused
 * by poor garbage collection in the JVM) a factory method is used to
 * generate instances and old instances are stored on a stack for
 * reuse.  
 */
public class CParseCompilationUnitContext {
    public static CParseCompilationUnitContext getInstance() {
	return stack.size() == 0 ?
	    new CParseCompilationUnitContext() :
	    (CParseCompilationUnitContext)stack.pop();
    }

    /**
     * Calls the static method <code>release(this)</code>
     */
    public void release() {
	release(this);
    }

    /**
     * Erases the data stored in <code>context</code> and pushes the
     * instance onto a stack for recycling.
     * @param	context	a used instance to be recycled
     */
    public static void release(CParseCompilationUnitContext context) {
	context.clear();
	stack.push(context);
    }

    /**
     * Hides the default constructor and ensures that every instance
     * package <code>java.lang</code>.
     */ 
    private CParseCompilationUnitContext() {
	// 7.5.3 Automatic Imports
	addPackageImport(JAVA_LANG);
    }

    /**
     * prepares this instance for recycling by clearing all the data
     * stored in its ArrayLists.
     */
    private void clear() {
	packageImports.clear();
	singleImports.clear();
	typeDeclarations.clear();
	tlMethods.clear();

	// 7.5.3 Automatic Imports
	addPackageImport(JAVA_LANG);
	pack = null;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (ADD)
    // ----------------------------------------------------------------------

    public void setPackage(JPackageName pack) {
	packageName = (pack == JPackageName.UNNAMED) ? "" : pack.getName()+'/';
	this.pack = pack;
	export = new CCompilationUnit( packageName );
    }

    public void addPackageImport(JPackageImportType pack) {
	packageImports.add(pack);
    }

    public void addSingleImport(JClassOrGFImportType clazz) {
	singleImports.add(clazz);
    }

    public void addTypeDeclaration(Main compiler, JTypeDeclarationType decl) {
	typeDeclarations.add(decl);
	decl.generateInterface( compiler, null, export, packageName, 
				false, false );
    }

    public void addMJTopLevelMethodDeclaration( Main compiler, 
						MJTopLevelMethodDeclaration 
						    decl ) {
	tlMethods.add(decl);
	decl.registerPendingGFCollection(packageName);
    }

    

    // ----------------------------------------------------------------------
    // ACCESSORS (GET)
    // ----------------------------------------------------------------------

    public JPackageImportType[] getPackageImports() {
	JPackageImportType[] temp = 
	    new JPackageImportType[packageImports.size()];

	packageImports.toArray(temp);

	return temp;
    }

    /**
     * Returns a list representing all the classes and generic functions
     * imported explicitly in this compilation unit.
     *
     * @return	a list of JClassOrGFImportType instances
     */
    public ArrayList getSingleImports() {
	return (ArrayList) singleImports.clone();    
    }

    public JTypeDeclarationType[] getTypeDeclarations() {
	JTypeDeclarationType[] temp = 
	    new JTypeDeclarationType[typeDeclarations.size()];

	typeDeclarations.toArray(temp);

	return temp;    
    }

    /**
     * Returns a list of the top level declarations in this compilation
     * unit context.
     * @return	a ArrayList whose elements are all instances of 
     *		MJTopLevelMethodDeclaration
     */
    public ArrayList getMJTopLevelMethodDeclarations() {
	return (ArrayList) tlMethods.clone();
    }

    public JPackageName getPackageName() {
	return pack;
    }

    public CCompilationUnit compilationUnitExport() {
	return export;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    public static final JPackageImportType	JAVA_LANG =
	new JPackageImport(org.multijava.util.compiler.TokenReference.NO_REF, Constants.JAV_RUNTIME, null);

    private JPackageName pack;


    /**
     * The name of the package of the compilation unit represented by
     * this.  Formatted with '/' separators, not '.'.  */
    private String packageName;
    private CCompilationUnit export;
    private ArrayList packageImports = new ArrayList();
    private ArrayList singleImports = new ArrayList();
    private ArrayList typeDeclarations = new ArrayList();
    private ArrayList tlMethods = new ArrayList();

    private static Stack stack = new Stack();
}
