/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceAmbiguousDispatcherClass.java,v 1.10 2004/02/23 07:09:57 tongjie Exp $
 */

package org.multijava.mjc;

import java.util.*;
import org.multijava.relaxed.util.RMJAnnotation;

import org.multijava.util.classfile.*;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents an inner dispatcher class from which an
 * ambiguous error method can be generated.  RMJ only. */

public class CSourceAmbiguousDispatcherClass extends CSourceClass 
    implements CAmbiguousDispatcherClass {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    public CSourceAmbiguousDispatcherClass( Main compiler, CSourceClass outer,
					    TokenReference where, 
					    CDispatcherSignature dispatcherSig,
					    CSourceDispatcherMethod
					      dispatcherMethod,
					    int number ) {
	super( compiler, 
	       new MemberAccess(outer, outer, ACC_PUBLIC | ACC_STATIC), where, 
	       "ident() overridden", "qualifiedName() overridden", CTypeVariable.EMPTY,false, 
	       true, false );
	this.where = where;
	this.outer = outer;
	this.dispatcherSig = dispatcherSig;
	this.dispatcherMethod = dispatcherMethod;
	this.number = number;

	computeDispatchedPositions();

        CTopLevel.addClass( this );
	compiler.classToGenerate( this );

	Hashtable fields = computeFields();

	setSuperClass(CStdType.Object);
	setCheckedInterfaces( 
	    new CClassType[] { dispatcherSig.getType() });

	setCheckedMembers( fields, new CMethod[0] );

	CMethod[] methods = computeMethods();

	close( methods, new CClassType[0] );
    }

    private void computeDispatchedPositions()
    {
	// figure out which arguments are the dispatchable positions
	CSpecializedType[] params = dispatcherMethod.parameters();
	int numParams = params.length;
	for (int i = 0; i < numParams; i++) {
	    CSpecializedType param = params[i];
	    if (param.staticType().isClassType()) {
		// this position might be dispatched upon
		dispatchedPositions.add(new Integer(i));
	    }
	}
    }

    private Hashtable computeFields()
    {
	Hashtable fields = new Hashtable();
	int fieldPos = 0;

	// Need an oldFunction field for the chain of responsibility,
	// for if this error method doesn't apply.
	CSourceField oldFunction =
	    new CSourceField( new MemberAccess(this, ACC_PROTECTED), 
			      "oldFunction",
			      dispatcherSig.getType(),
			      false ); 
	oldFunction.setIndex(fieldPos++);
	fields.put( oldFunction.ident(), oldFunction );

	// add fields to hold the actual specializers (or null) for
	// each of the dispatchable argument positions
	Iterator iter = dispatchedPositions.iterator();
	while (iter.hasNext()) {
	    int pos = ((Integer) iter.next()).intValue();
	    CSourceField spec =
		new CSourceField( new MemberAccess(this, ACC_PROTECTED), 
				  "spec" + pos,
				  CStdType.Class, 
				  false ); 
	    spec.setIndex(fieldPos++);
	    fields.put( spec.ident(), spec );
	}

	return fields;
    }

    private CMethod[] computeMethods()
    {
	CMethod[] methods = new CMethod[3];
	methods[0] = dispatcherConstructor();
	methods[1] = installMethod();
	methods[2] = filterMethod();
	return methods;
    }

    /**
     * Returns the constructor for this ambiguous-method error
     * dispatcher class.  Takes the classes of the specializable
     * arguments and stores them into the corresponding fields.
     */
    private CMethod dispatcherConstructor() {
	// anonymous class overrides the code generation methods
	return new CSourceMethod( new MemberAccess(this, this, ACC_PUBLIC),
				  JAV_CONSTRUCTOR,
				  CStdType.Void, 
				  new CSpecializedType[0],
				  new CClassType[0],
				  CTypeVariable.EMPTY,
				  false, null, null, null ) {
		// instance initializer
		{
		    this.topConcreteMethod = this;
		}

		public ArrayList genMethodInfo() {
		    ArrayList result = new ArrayList( 1 );
		    result.add(nonEmptyMethodInfo( this.ident() ));
		    return result;
		}

		public CodeInfo genCode() {
		    CodeSequence code = CodeSequence.getCodeSequence();

		    // generate call to Object()
		    code.plantLoadThis();
		    MethodRefInstruction superConst =
			new MethodRefInstruction( 
			      opc_invokespecial,
			      CStdType.Object.qualifiedName(),
			      JAV_CONSTRUCTOR,
			      CType.genMethodSignature( CStdType.Void, 
							new CType[0] ) );
		    code.plantInstruction( superConst );

		    code.plantInstruction( new NoArgInstruction(opc_return) );

		    CodeInfo info = new CodeInfo( code.getInstructionArray(),
						  code.getHandlers(),
						  optimizingLevel() >  1 ? 
						  null : code.getLineNumbers(),
						  null);
		    code.release();

		    return setParmsAndOptimizeCode( info );
		}
	    };
    }



    /**
     * Returns the constructor for this ambiguous-method error
     * dispatcher class.  Takes the classes of the specializable
     * arguments and stores them into the corresponding fields.
     */
    private CMethod installMethod() {
	int numParams = dispatchedPositions.size();
	CSpecializedType[] params = new CSpecializedType[numParams];
	for (int i = 0; i < numParams; i++) {
	    params[i] = new CSpecializedType(CStdType.Class);
	}

	// anonymous class overrides the code generation methods
	return new CSourceMethod( new MemberAccess(this, this, 
						   ACC_PUBLIC | ACC_STATIC),
				  "install",
				  CStdType.Void, params, new CClassType[0],
				  CTypeVariable.EMPTY,
				  false, null, null, null ) {
		// instance initializer
		{
		    this.topConcreteMethod = this;
		}

		public ArrayList genMethodInfo() {
		    ArrayList result = new ArrayList( 1 );
		    result.add(nonEmptyMethodInfo( this.ident() ));
		    return result;
		}

		public CodeInfo genCode() {
		    CodeSequence code = CodeSequence.getCodeSequence();

		    String qualifiedName = 
			CSourceAmbiguousDispatcherClass.this.qualifiedName();

		    // create the dispatcher object
		    code.plantInstruction(
			  new ClassRefInstruction( opc_new, qualifiedName ));

		    // dup it
		    code.plantInstruction( new NoArgInstruction( opc_dup ) );

		    // invoke the initializer
		    MethodRefInstruction constructor =
			new MethodRefInstruction( 
			      opc_invokespecial,
			      qualifiedName,
			      JAV_CONSTRUCTOR,
			      CType.genMethodSignature(
			          CStdType.Void, 
				  new CType[0]));
		    code.plantInstruction( constructor );

		    // initialize all the instance variables
		    int argNum = 0;
		    String specTypeName = CStdType.Class.getSignature();
		    Iterator iter = dispatchedPositions.iterator();
		    while (iter.hasNext()) {
			int pos = ((Integer) iter.next()).intValue();
			// obj.spec_i = arg_j
			LocalVarInstruction loadNextArg =
			    new LocalVarInstruction( opc_aload, argNum++ );
			FieldRefInstruction setNextSpec =
			    new FieldRefInstruction( opc_putfield,
						     qualifiedName,
						     "spec" + pos,
						     specTypeName);
			code.plantInstruction( new NoArgInstruction( opc_dup ) );
			code.plantInstruction( loadNextArg );
			code.plantInstruction( setNextSpec );
		    }

		    String anchorClassName = 
			CSourceAmbiguousDispatcherClass.this.
			anchorClass().qualifiedName();
		    String functionName = "function" +
			CSourceAmbiguousDispatcherClass.this.
			functionNumber(); // !FIXME!function
		    String oldFunctionTypeName = 
			CSourceAmbiguousDispatcherClass.this.
			dispatcherSig.getType().getSignature();

		    // this.oldFunction = anchor.function
		    FieldRefInstruction getAnchorFunction =
			new FieldRefInstruction( opc_getstatic,
						 anchorClassName,
						 functionName,
						 oldFunctionTypeName );
		    FieldRefInstruction setOldFunction =
			new FieldRefInstruction( opc_putfield,
						 qualifiedName,
						 "oldFunction",
						 oldFunctionTypeName );
		    code.plantInstruction( new NoArgInstruction( opc_dup ) );
		    code.plantInstruction( getAnchorFunction );
		    code.plantInstruction( setOldFunction );

		    // finally, add this object to the
		    // chain-of-responsibility of the anchor
		    FieldRefInstruction setAnchorFunction =
			new FieldRefInstruction( opc_putstatic,
						 anchorClassName,
						 functionName,
						 oldFunctionTypeName );
		    code.plantInstruction( setAnchorFunction );

		    code.plantInstruction( new NoArgInstruction(opc_return) );

		    CodeInfo info = new CodeInfo( code.getInstructionArray(),
						  code.getHandlers(),
						  optimizingLevel() >  1 ? 
						  null : code.getLineNumbers(),
						  null);
		    code.release();

		    return setParmsAndOptimizeCode( info );
		}
	    };
    }


    /**
     * Returns the method that tests whether the arguments match the
     * error condition.
     */
    private CMethod filterMethod() {
	final CSpecializedType[] params = 
	    CSpecializedType.unspecializedTypeTupleFrom(
	        dispatcherMethod.parameters());
    
	// anonymous class overrides the code generation methods
	return new CSourceMethod( new MemberAccess(this, this, ACC_PUBLIC),
				  "apply",
				  dispatcherMethod.returnType(),
				  params,
				  dispatcherMethod.throwables(),
				  CTypeVariable.EMPTY,
				  false, null, null, null ) {
		// instance initializer
		{
		    this.topConcreteMethod = this;
		}

		public ArrayList genMethodInfo() {
		    ArrayList result = new ArrayList( 1 );
		    result.add(nonEmptyMethodInfo( this.ident() ));
		    return result;
		}

		public CodeInfo genCode() {
		    CodeSequence code = CodeSequence.getCodeSequence();

		    String specTypeName = CStdType.Class.getSignature();
		    String qualifiedName = 
			CSourceAmbiguousDispatcherClass.this.qualifiedName();

		    CodeLabel noMatchLabel = new CodeLabel();

		    // for each argument:
		    Iterator iter = dispatchedPositions.iterator();
		    int pos = -1;
		    int numParams = params.length;
		    int argPos = 1;
		    int i = 0;
		    for (; i < numParams; i++) {
			CType type = params[i].staticType();

			if (pos < i) {
			    // we need to find the next specializable position
			    if (! iter.hasNext()) {
				// no more; we're done!
				break;
			    }
			    pos = ((Integer) iter.next()).intValue();
			    assertTrue(pos >= i);
			}

			if (pos == i) {
			    // this position is a specializable one

			    // load this.spec
			    FieldRefInstruction nextSpec =
				new FieldRefInstruction( opc_getfield,
							 qualifiedName,
							 "spec" + pos,
							 specTypeName);
			    code.plantLoadThis();
			    code.plantInstruction(nextSpec);

			    code.plantInstruction(new NoArgInstruction(opc_dup));

			    // check if null
			    CodeLabel nullLabel = new CodeLabel();
			    code.plantLabelRef(opc_ifnull, nullLabel);

			    // if not null, then check if spec.isInstance(arg)

			    // load corresponding arg
			    CodeLabel matchedLabel = new CodeLabel();
			    LocalVarInstruction nextArg =
				new LocalVarInstruction( type.getLoadOpcode(),
							 argPos );
			    code.plantInstruction( nextArg );

			    // invoke isInstance
			    MethodRefInstruction isInstance =
				new MethodRefInstruction( 
				    opc_invokevirtual,
				    CStdType.Class.qualifiedName(),
				    "isInstance",
				    CType.genMethodSignature(
					CStdType.Boolean, 
					new CType[] { CStdType.Object } ));
			    code.plantInstruction( isInstance );

			    // if false (i.e., if = 0), didn't match
			    code.plantLabelRef(opc_ifeq, noMatchLabel);

			    // matched
			    code.plantLabelRef(opc_goto, matchedLabel);

			    code.plantLabel(nullLabel);
			    // specializer was null; drop it and continue
			    code.plantInstruction(new NoArgInstruction(opc_pop));

			    code.plantLabel(matchedLabel);
			    // this specializable position matched;
			    // continue on with the rest of the
			    // specializable positions
			}

			// compute the next arg stack pos, and continue
			argPos += type.getSize();
		    }
		    assertTrue(!iter.hasNext());
		    assertTrue(pos < i);

		    // all specializable positions matched; that means
		    // we had an error

		    // push new RMJRuntimeException
		    code.plantInstruction(
			  new ClassRefInstruction( opc_new, 
				CStdType.RMJRuntimeException.qualifiedName()));

		    // dup it
		    code.plantInstruction( new NoArgInstruction( opc_dup ) );

		    // push the string error message
		    String message = where + 
			": invoking method on ambiguously defined arguments";
		    code.plantInstruction(new PushLiteralInstruction(message));

		    // invoke the exn initializer
		    MethodRefInstruction exnConst =
			new MethodRefInstruction( 
			      opc_invokespecial,
			      CStdType.RMJRuntimeException.qualifiedName(),
			      JAV_CONSTRUCTOR,
			      CType.genMethodSignature(
			          CStdType.Void, 
				  new CType[] { CStdType.String } ));
		    code.plantInstruction( exnConst );

		    // throw the exn
		    code.plantInstruction(new NoArgInstruction(opc_athrow));

		    code.plantLabel(noMatchLabel);
		    // didn't match (that's good); invoke oldFunction

		    // push oldFunction
		    String oldFunctionTypeName = 
			CSourceAmbiguousDispatcherClass.this.
			dispatcherSig.getType().getSignature();
		    FieldRefInstruction getOldFunction =
			new FieldRefInstruction( opc_getfield,
						 qualifiedName,
						 "oldFunction",
						 oldFunctionTypeName );
		    code.plantLoadThis();
		    code.plantInstruction( getOldFunction );

		    // push all the arguments
		    argPos = 1;
		    for (i = 0; i < numParams; i++) {
			CType type = params[i].staticType();
			LocalVarInstruction nextArg =
			    new LocalVarInstruction( type.getLoadOpcode(), 
						     argPos );
			code.plantInstruction( nextArg );
			argPos += type.getSize();
		    }

		    // invoke oldFunction.apply(args)
		    String oldFunctionClassName = 
			CSourceAmbiguousDispatcherClass.this.
			dispatcherSig.getType().qualifiedName();
		    CType applyReturnType = dispatcherMethod.returnType();
		    CType[] applyArgTypes =
			CSpecializedType.staticTypeTupleFrom(null, params);

		    InvokeinterfaceInstruction apply =
			new InvokeinterfaceInstruction( 
				    oldFunctionClassName,
				    "apply",
				    CType.genMethodSignature(
					applyReturnType,
					applyArgTypes ),
				    argPos);
		    code.plantInstruction( apply );

		    if (applyReturnType.isVoid()) {
			code.plantInstruction(
			    new NoArgInstruction(opc_return));
		    } else {
			code.plantInstruction(
			    new NoArgInstruction(
				applyReturnType.getReturnOpcode()));
		    }

		    CodeInfo info = new CodeInfo( code.getInstructionArray(),
						  code.getHandlers(),
						  optimizingLevel() >  1 ? 
						  null : code.getLineNumbers(),
						  null);
		    code.release();

		    return setParmsAndOptimizeCode( info );
		}
	    };
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public String ident() {
	return outer.ident() + "$ambig" + number;
    }

    public String qualifiedName() {
	String result = host().packageName();
	return result.length() > 1 ? result + ident() : ident();
    }

    public CClass anchorClass() {
	return outer;
    }

    public int functionNumber() {
	return number;
    }

    //---------------------------------------------------------------------
    // DERIVED ATTRIBUTES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // INTERFACE CHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // TYPECHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    public void rmjAnnotate(RMJAnnotation rmjInfo)
    {
	// we don't generate any annotations for this class, since
	// it's loaded on demand by the loader to represent an
	// ambiguous method combination
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    private int number;

    private TokenReference where;

    private CSourceClass outer;

    protected final CDispatcherSignature dispatcherSig;

    protected final CSourceDispatcherMethod dispatcherMethod;

    protected final ArrayList dispatchedPositions = new ArrayList();

}// CSourceAmbiguousDispatcherClass
