/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceClass.java,v 1.20 2006/12/10 22:51:21 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.*;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError; 

import org.multijava.util.Destination;

import org.multijava.relaxed.util.RMJAnnotation;

import java.io.File;
import java.io.IOException;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Set;
import java.util.Map;
import java.util.Iterator;

/**
 * This class represents a class read from a *.java file.  It is
 * primarily just a data structure, apart from methods for generating
 * the qualified name and for determining whether the member is
 * accessible from some class.
 *
 * @see	CMember
 */
public class CSourceClass extends CClass {
    //$$$ OPTIMIZE OUTERS HANDLING

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a class export from source
     */
    public CSourceClass( Main compiler,
			 MemberAccess access,
			 TokenReference where,
			 String ident,
			 String qualifiedName, 
			 CTypeVariable[] typevariables,
			 boolean isAnonymous,
			 boolean isMember,
			 boolean deprecated )
    {
	super(compiler, access, where.file(), 
	      ident, qualifiedName, null, deprecated);
	setTypeVariables(typevariables);
	this.isAnonymous = isAnonymous;
	this.isMember = isMember;
    }
    
    /**
     * Stores information on member signatures gathered during
     * checkInterface pass prior to closing in the translateMJ
     * pass.  Method and field lookup is possible after this method
     * has been called, however, code generation must wait until after
     * the <code>close</code> method has been called. */
    public void setCheckedMembers( Hashtable fields_H,
				   CMethod[] protoMethods_A ) {
	super.setCheckedMembers( null, fields_H, protoMethods_A );
    }

    /**
     * Returns true if initialization pass is finished for this class
     */
    public boolean initializerChecked() {
	return initializerChecked;
    }

    /**
     * Sets initialization pass to be finished
     */
    public void setInitializerChecked() {
	initializerChecked = true;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * @return true if this class is anonymous
     */
    public boolean isAnonymous() {
	return isAnonymous;
    }

    /**
     * Indicates whether the class represented by this is a member
     * type.  A member type is a type that is declared inside another
     * type but that is not an anonymous class or a local type
     * declaration.  */
    public boolean isMember() {
	return isMember;
    }
    
    /**
     * Return the version of the encoding used to encode the universe
     * type modifiers.
     * By AHS.
     */
    public String getUniverseVersion() {
        return CTopLevel.getCompiler().universeVersion();
    }
    
    /**
     * Return true if the class is being compiled with run-time support
     * enabled.
     * By AHS.
     */
    public boolean hasUniverseRuntimeSupport() {
        return CTopLevel.getCompiler().universeDynChecks();
    }
    

    
    /**
     * Mutates the given array to add any synthetic parameters for
     * outer local variables referenced within the class represented
     * by this.
     *
     * <pre><jml>
     * also
     * requires (* this represents an anonymous class or 
     *		   a local class declaration *);
     * requires (* params is the parameter array for a constructor of this, 
     *		   already including any synthetic outer this parameter *);
     * ensures (* params includes the types of any outer local variables 
     *		  required to instantiate this *);
     * </jml></pre>
     *
     * @see #outers
     */
    public CType[] genConstructorArray(CType[] params) {
	int size = params.length 
	    + (outers == null ? 0 : outers.size());
	if( params.length != size ) {
	    CType[] ret = new CType[size];
	    
	    System.arraycopy( params, 0, ret, 0, params.length );
	    int pos = params.length;
      
	    if( outers != null ) {
		Enumeration kys = outers.keys();
		while( kys.hasMoreElements() ) {
		    JLocalVariable var = (JLocalVariable)kys.nextElement();
		    ret[pos++] = var.getType();
		}
	    }

	    return ret;
	}

	return params;
    }

    /**
     * Adds synthetic parameters to the code sequence for a call to
     * this class's constructor.  In other words, it does nothing,
     * except for anonymous classes and inner class declarations that
     * reference final local variables.  In those cases, this adds
     * code to push the local variable values onto the stack.
     * 
     * @param code the code sequence being generated */
    public void genSyntheticParams( CodeSequence code ) {
	if (outers != null) {
	    // Add code to push outer local variables used by this
	    // onto the stack
	    Enumeration	kys = outers.keys();
	    while ( kys.hasMoreElements() ) {
		JLocalVariable	var = (JLocalVariable) kys.nextElement();

		new JLocalVariableExpression( TokenReference.NO_REF, var ).
		    genCode( code );
	    }
	}
    }

    /**
     * Adds synthetic parameters to the code sequence for an explicit
     * constructor invocation to this class's constructor.  In other
     * words, it does nothing, except for anonymous classes and inner
     * class declarations that reference final local variables.  In
     * those cases, this adds code to push the local variable values
     * onto the stack.
     *
     * @param code 		the code sequence being generated 
     * @param constructor 	the signature of the calling constructor
     */
    public void genSyntheticParamsForExplicitThis(CodeSequence code,
						  CMethod constructor) {
	if (outers != null) {
	    Set	pairs = outers.entrySet();
	    Iterator pairIter = pairs.iterator();
	    while (pairIter.hasNext()) {
		Map.Entry varNamePair = (Map.Entry) pairIter.next();
		String name = (String) varNamePair.getValue();
		JLocalVariable	var = (JLocalVariable) varNamePair.getKey();

		generatedLocalVarHelper( constructor, 
					 (CSourceField) getDeclaredField(name),
					 var, 
					 TokenReference.NO_REF ).genCode(code);
	    }
	}
    }


    /**
     * For an explicit super class constructor invocation, returns the
     * expressions that must be evaluated in the body of the
     * constructor <code>caller</code> of the class
     * <code>callerClass</code> to add the outer local variable values
     * required by this to the stack.
     *
     * <pre><jml>
     * also
     * requires callerClass.descendsFrom(this);
     * requires caller.isConstructor();
     * ensures \result != null;
     * </jml></pre>
     * */
    public JExpression[] 
	getSyntheticParamsForExplicitSuper( CSourceClass callerClass,
					    CMethod caller ) {
	JExpression[] result = 
	    new JExpression[ outers == null ? 0 : outers.size() ];

	if (outers != null) {
	    Enumeration	kys = outers.keys();
	    int pos = 0;
	    while ( kys.hasMoreElements() ) {
		JLocalVariable	var = (JLocalVariable) kys.nextElement();

		result[pos++] = 
		    callerClass.getOuterLocalAccess( TokenReference.NO_REF, 
						     var, caller );
	    }
	}

	return result;
    }

    /**
     * Returns an expression that evaluates to the value of the outer
     * local variable <code>var</code>.  The <code>constructor</code>
     * parameter is used to generate a reference to a synthetic outer
     * variable parameter rather than a synthetic field when the original
     * reference appears in a constructor.
     * 
     * @param ref	a token reference to be added to the returned expr
     * @param var	the variable to be referenced
     * @param constructor the constructor in which this reference appears,
     *			or null if the reference is not in a constructor
     */
    public JExpression getOuterLocalAccess( TokenReference ref, 
					    JLocalVariable var, 
					    final CMethod constructor ) {
	CSourceField	field;

	if (outers == null) {
	    outers = new Hashtable();
	}
	
	String name = (String)outers.get(var);

	if (name == null) {
	    name = "val$" + var.ident(); // local vars uniq, so no collisions
	    outers.put(var, name);
	    field = new CSourceField(new MemberAccess(this, 
						      ACC_PRIVATE | ACC_FINAL), 
				     name, var.getType(), false);
	    // JOuterLocalVariable checks that the outer var is
	    // assigned to.  Code generation will ensure that the
	    // field is assigned to.
	    field.initializedUnlessFinalInstance(); 
	    addField(field);
	    countSyntheticsFields++;
	} else {
	    field = (CSourceField)getDeclaredField(name);
	}

	// A reference to the outer local var is via the local call
	// stack within the inner class's constructor.  Otherwise the
	// references are resolved via the synthetic field.
	if (constructor != null) {
	    return new JCheckedExpression(ref, 
					  generatedLocalVarHelper(constructor,
								  field, var,
								  ref));
	} else {
	    return new JClassFieldExpression(ref, 
					     new JThisExpression(ref, 
								 getCClass()), 
					     name);
	}
    }

    private JLocalVariableExpression 
	generatedLocalVarHelper(final CMethod constructor, 
				final CSourceField field,
				JLocalVariable var,
				TokenReference ref) 
    {
	JGeneratedLocalVariable local = 
	    new JGeneratedLocalVariable(org.multijava.util.compiler.TokenReference.NO_REF, 0, 
					var.getType(), 
					var.ident(), null) {
		    /**
		     * @return the local index in context variable table
		     */
		    public int getPosition() {
			return 1 /*this*/ 
			    + constructor.parameters().length 
			    + (isInnerClass() ? 1 : 0) 
			    + field.index() 
			    - (getFieldCount() - countSyntheticsFields);
		    }
		};
	return new JLocalVariableExpression(ref, local);
    }

    /**
     * Adds to the body of a constructor the storage of synthetic
     * parameters (for outer local variables) into synthetic fields.
     * Some nested classes (anonymous classes and local class
     * declarations) have access to final local variables.  This
     * access is accomplished in bytecode by passing the values of the
     * variables as synthetic parameters to the nested class's
     * constructor.  This code stores these parameters in synthetic
     * fields of the owner of this.  
     *
     * <pre><jml>
     * requires isNestedType();
     * requires (* code represents the body of a constructor of the class 
     *		   represented by this *);
     * requires (* countLocals == the stack space consumed by the actual
     *		   parameters of the constructor along with any synthetic
     *		   outer this parameter *);
     * </jml></pre>
     *
     */
    public void genInit(CodeSequence code, int countLocals) {
	countLocals++;		// move to next slot
	if (outers != null) {
	    Enumeration	kys = outers.keys();
	    while (kys.hasMoreElements()) {
		JLocalVariable ovar = (JLocalVariable)kys.nextElement();
		JGeneratedLocalVariable var = 
		    new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF, 
						 0, ovar.getType(), 
						 ovar.ident(), null );
		var.setPosition(countLocals);
		countLocals += var.getType().getSize();
		code.plantLoadThis();
		var.genLoad(code);
		FieldRefInstruction putField =
		    new FieldRefInstruction(opc_putfield, qualifiedName(),
					    "val$" + var.ident(),
					    var.getType().getSignature());
		code.plantInstruction( putField );
	    }
	}
    }

    public CMethodSet lookupMethodOrSet( String name,
					 CType[] params,
					 CClassType[] args,
					 CContextType context)
        throws UnpositionedError
    {
	if (this == CLS_UNDEFINED) {
	    // ??? !FIXME! System.out.println("Fatal error because of an attempt to lookup a method in an undefined class, possibly because of a failure to load a class because of a previous error");
	    // This can happen if a syntax error causes the processTaskQueue
	    // loop to abort without having loaded a class.
	    return null;
	} else {
	    return super.lookupMethodOrSet(name,params,args,context);
	}
    }


    public CClass lookupClass( String name, CContextType context )  
        throws UnpositionedError
    {
        if (this == CLS_UNDEFINED) {
            // ??? !FIXME! System.out.println("Fatal error because of an attempt to lookup a method in an undefined class, possibly because of a failure to load a class because of a previous error");
            // This can happen if a syntax error causes the processTaskQueue
            // loop to abort without having loaded a class. 
            return null;
        } else {
                return super.lookupClass(name,context);
        }
    }

    /**
     * Returns a unique number for a nested dispatcher class of this.
     */
    protected int nextDispatcherNumber() {
        return nextDispatcherNumber++;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Generate a class file for the class represented by this. */
    public void genCode( Destination destination )
	throws IOException, ClassFileFormatException
    {
	if (getCompiler().RMJ()) {
	    // Before the bytecode is generated, go ahead and get
	    // annotations and add to class
	    rmjAnnotate();
	}

	genClassFile( destination );
    }

    /** Provide an annotation for any dispatcher classes (RMJ only) */
    protected void rmjAnnotate()
    {
	CClassType[] inners = getInnerClasses();
	for (int i = 0; i < inners.length; i++) {
	    CClassType ctype = inners[i];
	    CClass cls = ctype.getCClass();
	    cls.rmjAnnotate(_rmjInfo);
	}
    }
    
    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Appends any custom attributes for this class to the attribute
     * list represented by <code>accum</code>.  */
    protected void appendCustomAttributes(AttributeList accum) {
	if (getCompiler().RMJ()) {
	    // I hope we've already called the genCode method....

	    if (!_rmjInfo.isEmpty()) {
		String rmjAnnote = _rmjInfo.getAnnotationString();
		accum.add( new RMJAttribute(rmjAnnote) );
	    }
	}
	super.appendCustomAttributes(accum);
    }

    public void addRMJMethodAnnotation( RMJAnnotation.Method annote )
    {
	_rmjInfo.add(annote);
    }

   
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	Hashtable		outers;

    private	int			countSyntheticsFields;
    private	boolean			initializerChecked;

    /**
     * Used to construct nested dispatcher class identifiers for
     * methods of external generic functions.  */
    private int nextDispatcherNumber;

    // RMJ 21Mar2002 Added general annotation mechanism for any class
    private RMJAnnotation _rmjInfo = new RMJAnnotation();

    /**
     * Indicates whether this represents the signature of an anonymous
     * class.  */
    private final boolean isAnonymous;
    
    /**
     * Indicates whether the class represented by this is a member
     * type.  A member type is a type that is declared inside another
     * type but that is not an anonymous class or a local type
     * declaration.  */
    private final boolean isMember;

}
