/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceDispatcherClass.java,v 1.33 2006/12/10 22:51:21 chalin Exp $
 */

package org.multijava.mjc;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import org.multijava.relaxed.util.RMJAnnotation;

import org.multijava.util.Destination;
import org.multijava.util.classfile.ClassFileFormatException;
import org.multijava.util.classfile.ClassRefInstruction;
import org.multijava.util.classfile.CodeInfo;
import org.multijava.util.classfile.FieldRefInstruction;
import org.multijava.util.classfile.HandlerInfo;
import org.multijava.util.classfile.LocalVarInstruction;
import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents an inner dispatcher class for a collection of
 * multimethods of the same generic function declared in a single
 * lexical context.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.33 $ */

public class CSourceDispatcherClass extends CSourceClass {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public CSourceDispatcherClass( Main compiler, CSourceClass outer,
                                   TokenReference where, 
                                   CSourceDispatcherMethod dispatcherMethod ) {
        super( compiler, 
               new MemberAccess(outer, outer, 
                                /* RMJ needs to have public access
                                 * from the class loader */
                                ACC_STATIC | (compiler.RMJ() ? 
                                              ACC_PUBLIC : ACC_PRIVATE)),
               where, 
               "ident() overridden", "qualifiedName() overridden", CTypeVariable.EMPTY,false, 
               true, false );
        this.number = outer.nextDispatcherNumber();
        this.outer = outer;
        this.where = where;     // cache here because superclass just extracts
                                // the filename and discards line number
        this.dispatcherMethod = dispatcherMethod;
        CTopLevel.addClass( this );
        compiler.classToGenerate( this );
        if (dispatcherMethod.isTopDispatcherInLattice()) {
            completeClassInfo( compiler, where );
        }
    }


    /**
     * Generates the $signature interface and (if RMJ) the
     * ambiguous-method dispatcher class for this generic function and
     * then calls <code>completeClassInfo()</code>.
     *
     * <pre><jml>
     * requires dispatcherMethod.isTopDispatcherInLattice();
     * </jml></pre>
     *
     */
    private void completeClassInfo( Main compiler, TokenReference where ) {
        assertTrue( dispatcherMethod.isTopDispatcherInLattice() );
        dispatcherSig = 
            new CSourceDispatcherSignature( compiler, outer, where,
                                            dispatcherMethod, number );
        if (compiler.RMJ()) {
            ambiguousDispatcherClass =
                new CSourceAmbiguousDispatcherClass(compiler, outer, where,
                                                    dispatcherSig, 
                                                    dispatcherMethod,
                                                    number );
        }
        completeClassInfo();
    }


    /**
     * Lazily completes the information necessary to generate the
     * bytecode for the dispatcher class represented by this.  The
     * info is completed lazily so that when this dispatcher class is
     * <em>not</em> the top dispatcher in the latice, the $signature
     * interface for the top dispatcher will already exist.  */
    private void completeClassInfo() {
        if (infoCompleted) {
            return;
        }

        boolean RMJ = getCompiler().RMJ();

        if (dispatcherSig == null) {
            // get signature from top dispatcher and cache it here
            dispatcherSig = dispatcherMethod.dispatcherSignature();
        }
        if (RMJ && ambiguousDispatcherClass == null) {
            // get ambiguous dispatcher from top dispatcher and cache it here
            ambiguousDispatcherClass =
                dispatcherMethod.ambiguousDispatcherClass();
        }
        
        Hashtable fields = new Hashtable();
        if( !dispatcherMethod.isTopDispatcherInLattice() ) {
            // Need an oldFunction field for the chain of
            // responsibility.  Ideally this would be private, but
            // since the JVM doesn't support privileged access on
            // inner classes we make this protected.
            CSourceField oldFunction =
                new CSourceField( new MemberAccess(this, ACC_PROTECTED), 
                                  "oldFunction",
                                  dispatcherSignature().getType(), false ); 
            fields.put( oldFunction.ident(), oldFunction );

            // XXX for now, we're not doing this, until we figure out
            // really how to support super sends from dispatcher
            // classes
            if (false && RMJ) {
                // also need an oldSuperFunction, to support super sends
                CSourceField oldSuperFunction =
                    new CSourceField( new MemberAccess(this, ACC_PROTECTED), 
                                      "oldSuperFunction",
                                      dispatcherSignature().getType(), 
                                      false ); 
                fields.put( oldSuperFunction.ident(), oldSuperFunction );
            }
        }

        setSuperClass(CStdType.Object);
        setCheckedInterfaces( 
            new CClassType[] { dispatcherSignature().getType() });
        setCheckedMembers( fields, new CMethod[0] );

        CMethod initMethod = null;
        if (RMJ) {
            // we include the initializer to put this dispatcher on
            // the anchor class's chain here
            initMethod = new CDispatcherInitializer(this);
        }

        CMethod[] methods;
        int baseSize = 2 + (initMethod == null ? 0 : 1);
        if (dispatcherMethod.isTopDispatcherInLattice()) {
            CMethodSet filteredDispatchers = 
                dispatcherMethod.filteredDispatchers();
            methods = new CMethod[filteredDispatchers.size() + baseSize];
            CAbstractMethodSet.Iterator i = filteredDispatchers.iterator();
            int pos = baseSize;
            while (i.hasNext()) {
                methods[pos] = i.next();
                pos++;
            }
        } else {
            methods = new CMethod[baseSize];
        }
        methods[0] = dispatcherMethod;
        methods[1] = dispatcherConstructor();
        if (initMethod != null) {
            methods[2] = initMethod;
        }

        close( methods, new CClassType[0] );
        infoCompleted = true;
    }


    /**
     * Returns the appropriate constructor for this dispatcher class.
     * For the top of the lattice the constructor simply calls the
     * super constructor <code>Object()</code>.  Otherwise the
     * constructor sets the <code>oldFunction</code> field of this
     * dispatcher to the next dispatcher in the
     * chain-of-responsibility.  
     *
     * <pre><jml>
     * requires dispatcherSig != null;
     * </jml></pre>
     *
     */
    private CMethod dispatcherConstructor() {
        final boolean chainNeeded = 
            !dispatcherMethod.isTopDispatcherInLattice();
        final boolean RMJ = getCompiler().RMJ();

        CSpecializedType[] params;
        if (chainNeeded) {
            CSpecializedType fnType = 
                new CSpecializedType( dispatcherSig.getType() );
            // XXX for now, we're not doing this, until we figure out
            // really how to support super sends from dispatcher
            // classes
            if (false && RMJ) {
                params = new CSpecializedType[] { fnType, fnType };
            } else {
                params = new CSpecializedType[] { fnType };
            }
        } else {
            params = new CSpecializedType[] { };
        }
        
        // anonymous class overrides the code generation methods
        return new CSourceMethod( new MemberAccess(this, this, ACC_PROTECTED), 
                                  JAV_CONSTRUCTOR,
                                  CStdType.Void, params, new CClassType[0],
                                  CTypeVariable.EMPTY,
                                  false, null, null, null ) {
                // instance initializer
                {
                    this.topConcreteMethod = this;
                }

                public ArrayList genMethodInfo() {
                    ArrayList result = new ArrayList( 1 );
                    result.add(nonEmptyMethodInfo( this.ident() ));
                    return result;
                }

                public CodeInfo genCode() {
                    CodeSequence code = CodeSequence.getCodeSequence();

                    // generate call to Object()
                    code.plantLoadThis();
                    MethodRefInstruction superConst =
                        new MethodRefInstruction( 
                              opc_invokespecial,
                              CStdType.Object.qualifiedName(),
                              JAV_CONSTRUCTOR,
                              CType.genMethodSignature( CStdType.Void, 
                                                        new CType[0] ) );
                    code.plantInstruction( superConst );

                    if (chainNeeded) {
                        // oldFunction = arg1
                        LocalVarInstruction loadArg1 =
                            new LocalVarInstruction( opc_aload, 1 );
                        FieldRefInstruction setOldFunction =
                            new FieldRefInstruction( opc_putfield,
                                                     CSourceDispatcherClass.
                                                     this.qualifiedName(),
                                                     "oldFunction",
                                                     dispatcherSig.
                                                     getType().getSignature());
                        code.plantLoadThis();
                        code.plantInstruction( loadArg1 );
                        code.plantInstruction( setOldFunction );

                        // XXX for now, we're not doing this, until we
                        // figure out really how to support super
                        // sends from dispatcher classes
                        if (false && RMJ) {
                            // oldSuperFunction = arg2
                            LocalVarInstruction loadArg2 =
                                new LocalVarInstruction( opc_aload, 2 );
                            FieldRefInstruction setOldSuperFunction =
                                new FieldRefInstruction( opc_putfield,
                                                         CSourceDispatcherClass.
                                                         this.qualifiedName(),
                                                         "oldSuperFunction",
                                                         dispatcherSig.
                                                         getType().getSignature());
                            code.plantLoadThis();
                            code.plantInstruction( loadArg2 );
                            code.plantInstruction( setOldSuperFunction );
                        }
                    }

                    code.plantInstruction( new NoArgInstruction(opc_return) );
                    CodeInfo info = new CodeInfo( code.getInstructionArray(),
                                                  code.getHandlers(),
                                                  optimizingLevel() >  1 ? 
                                                  null : code.getLineNumbers(),
                                                  null);
                    code.release();

                    return setParmsAndOptimizeCode( info );
                }
            };
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public String ident() {
        return outer.ident() + outersDispatcherFieldIdent();
    }

    public String qualifiedName() {
        String result = host().packageName();
        return result.length() > 1 ? result + ident() : ident();
    }

    /**
     * Returns the $signature interface for this generic function. */
    protected CDispatcherSignature dispatcherSignature() {
        if (dispatcherSig == null) {
            completeClassInfo();
        } // end of if 
        return dispatcherSig;
    }


    /**
     * Returns the ambiguous-method dispatcher class for this generic
     * function. (RMJ only) */
    protected CAmbiguousDispatcherClass ambiguousDispatcherClass() {
        if (ambiguousDispatcherClass == null) {
            completeClassInfo();
        } // end of if 
        return ambiguousDispatcherClass;
    }


    /**
     * Returns a field to be added to the outer class for storing a
     * reference to a singleton instance of this, or returns null if
     * the outer class is the anchor.  (The additional field is not
     * necessary in the anchor class.)
     */
    protected CSourceField outersDispatcherField() {
        if (dispatcherMethod.isTopDispatcherInLattice()) {
            return null;
        }
        /* RMJ needs to have public access from the class loader */
        long modifiers = 
            ACC_STATIC | (getCompiler().RMJ() ? ACC_PUBLIC : ACC_PRIVATE);
        return new CSourceField( new MemberAccess(outer, modifiers),
                                 outersDispatcherFieldIdent(), getType(), 
                                 false );
    }

    private String outersDispatcherFieldIdent() {
        return "$dispatcher" + number; //  !FIXME!function
    }

    //---------------------------------------------------------------------
    // DERIVED ATTRIBUTES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // INTERFACE CHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // TYPECHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    /**
     * Generate a class file for the class represented by this.
     * Overridden for external GF anchors.  */
    public void genCode( Destination destination )
        throws IOException, ClassFileFormatException
    {
        completeClassInfo();
        genClassFile( destination );
    }



    /**
     * Generates bytecode to put a reference to the external
     * <code>function</code> field for the external generic function
     * on the JVM stack. 
     *
     * <pre><jml>
     * requires dispatcherMethod.isTopDispatcherInLattice();
     * requires_redundantly dispatcherSig != null;
     * </jml></pre>
     *
     */
    public void plantFunctionRef( CodeSequence code ) {
        code.plantInstruction( 
                new MethodRefInstruction( opc_invokestatic,
                                          outer.qualifiedName(),
                                          "getFunction" + number, // !FIXME!function
                                          CType.genMethodSignature(
                                          	dispatcherSig.getType(),
                                          	CType.EMPTY) ));
    }

    /**
     * Creates a synchronized accessor method for the function field
     * for this dispatcher's generic function.
     *
     * <pre><jml>
     * requires dispatcherMethod.isTopDispatcherInLattice();
     * requires_redundantly dispatcherSig != null;
     * </jml></pre>
     */
    public CSourceMethod functionAccessorMethod() {
        CType type;
        if (getCompiler().RMJ()) {
            // call dispatcherSignature() insead of using
            // dispatcherSig in order to force a call to
            // completeClassInfo. This modification enables external
            // top level methods that extend other top level methods,
            // i.e., glue
            type = dispatcherSignature().getType();
        } else {
            assertTrue( dispatcherMethod.isTopDispatcherInLattice() && 
                        dispatcherSig != null );
            type = dispatcherSig.getType();
        }
        return new CFunctionAccessMethod( outer, dispatcherMethod.modifiers(),
                                 		 number, type );
    }

    /**
     * Creates a field data structure representing the function field
     * for this dispatcher's generic function.
     *
     * <pre><jml>
     * requires dispatcherMethod.isTopDispatcherInLattice();
     * requires_redundantly dispatcherSig != null;
     * </jml></pre>
     */
    public CSourceField functionField() {
        CType type;
        if (getCompiler().RMJ()) {
            // call dispatcherSignature() insead of using
            // dispatcherSig in order to force a call to
            // completeClassInfo. This modification enables external
            // top level methods that extend other top level methods,
            // i.e., glue
            type = dispatcherSignature().getType();
        } else {
            assertTrue( dispatcherMethod.isTopDispatcherInLattice() && 
                        dispatcherSig != null );
            type = dispatcherSig.getType();
        }
        return new CSourceField( new MemberAccess(outer, 
                                 dispatcherMethod.modifiers() | ACC_STATIC),
                                 "function" + number, // !FIXME!function
                                 type, false );
    }

    /**
     * Plants bytecode to load the oldFunction field of this
     * dispatcher class.  
     *
     * <pre><jml>
     * requires !dispatcherMethod.isTopDispatcherInLattice();
     * requires_redundantly dispatcherSig == null;
     * </jml></pre>
     * @param code the code sequence to add the bytecode to
     * @param forSuper if true, then this is for a super send, which,
     *                 under RMJ, should start from the oldSuperFunction
     *                 rather than oldFunction
     * @param fromOuter if true, then generates a reference that works from the
     *                  outer class context, otherwise generates a reference that
     *                  works from within the dispatcher class
     */
    public void plantOldFunctionRef( CodeSequence code, 
                                     boolean forSuper, boolean fromOuter ) {
        if (fromOuter) {
            FieldRefInstruction getDispatcher =
                new FieldRefInstruction( opc_getstatic, outer.qualifiedName(),
                                         outersDispatcherFieldIdent(),
                                         getType().getSignature() );
            code.plantInstruction( getDispatcher );
        } else {
            code.plantLoadThis();
        }
        
        String oldFunctionFieldName;
        // XXX for now, we're not doing this, until we figure out
        // really how to support super sends from dispatcher classes
        if (false && forSuper && getCompiler().RMJ()) {
            oldFunctionFieldName = "oldSuperFunction";
        } else {
            oldFunctionFieldName = "oldFunction";
        }

        FieldRefInstruction getOldFunction =
            new FieldRefInstruction( opc_getfield,
                                     qualifiedName(),
                                     oldFunctionFieldName,
                                     dispatcherSignature().getType().
                                     getSignature() );
        
        code.plantInstruction( getOldFunction );
    }

    /**
     * Plants the bytecode to add this to the appropriate
     * chain-of-responsibility and store a reference in a field of the
     * outer class.  */
    public void plantCORInitialization( CodeSequence code ) {
        code.setLineNumber( where.line() );
        // If this is the top of the lattice then we just set the
        // function field, otherwise we must link the oldFunction
        // field to the previous value.
        boolean isTop = dispatcherMethod.isTopDispatcherInLattice();
        
        // Enter the critical region for updating the chain of responsibility,
        // synchronized on the anchor class's Class instance.
        JLocalVariable lockVar = 
			new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF,
						     0, CStdType.Class, "lock", null ) {
				public int getPosition() {
					// Just assume we can use the second slot.  This works
					// because we're done with the slot before any code that
					// might use it tries to do so.
					return 1;
				}
			};
        plantGetAnchorClassObject(code);
        code.plantInstruction(new NoArgInstruction( opc_dup ));
		code.plantLocalVar(opc_astore, lockVar);
		code.plantInstruction(new NoArgInstruction(opc_monitorenter));
		int startPC = code.getPC(); // save start of critical region
                
        ClassRefInstruction createDispatcher =
            new ClassRefInstruction( opc_new, qualifiedName() );
        code.plantInstruction( createDispatcher );

        NoArgInstruction dup = new NoArgInstruction( opc_dup );
        code.plantInstruction( dup );

        if (!isTop) {
            // get current function field value
            plantGetFunctionField( code );
        }

        MethodRefInstruction initDispatcher =
            new MethodRefInstruction( 
                 opc_invokespecial,
                 qualifiedName(),
                 JAV_CONSTRUCTOR, 
                 CType.genMethodSignature( CStdType.Void,
                                           isTop ? new CType[0] : 
                                           new CType[] { 
                                               dispatcherSignature().getType()
                                           }));
        code.plantInstruction( initDispatcher );

        if (!isTop) {
            // Classes that specialize an external generic function
            // have a field that references the dispatcher.  This is
            // used for short-circuiting the chain of responsibility
            // on super calls.  (Note, this short-circuiting cannot be
            // used to optimize regular calls because it would
            // circumvent downcalls to subclasses.)
            dup = new NoArgInstruction( opc_dup );
            code.plantInstruction( dup );
            plantSetOutersDispatcherField( code );
        }

        plantSetFunctionField( code );
        
        // Exit critical region
		code.plantLocalVar(opc_aload, lockVar);
		code.plantInstruction(new NoArgInstruction(opc_monitorexit));
		int endPC = code.getPC();
		
		CodeLabel afterHandler = new CodeLabel();
		code.plantLabelRef(opc_goto, afterHandler);

		int	errorPC = code.getPC();
		code.plantLocalVar(opc_aload, lockVar);
		code.plantInstruction(new NoArgInstruction(opc_monitorexit));
		code.plantInstruction(new NoArgInstruction(opc_athrow));

		code.plantLabel(afterHandler);

		code.addExceptionHandler(new HandlerInfo(code.getInstructionAt(startPC),
								code.getInstructionAt(endPC),
								code.getInstructionAt(errorPC),
								null));

    }

    private void plantSetOutersDispatcherField( CodeSequence code ) {
        FieldRefInstruction setDispatcherField =
            new FieldRefInstruction( opc_putstatic, outer.qualifiedName(),
                                     outersDispatcherFieldIdent(),
                                     getType().getSignature() );
        code.plantInstruction( setDispatcherField );
    }

    private void plantGetFunctionField( CodeSequence code ) {
        FieldRefInstruction getFunction =
            new FieldRefInstruction( opc_getstatic,
                                     anchorClass().qualifiedName(),
                                     "function" + functionNumber(), //  !FIXME!function
                                     dispatcherSignature().getType().
                                     getSignature() );
        code.plantInstruction( getFunction );
    }

    private void plantSetFunctionField( CodeSequence code ) {
        FieldRefInstruction setFunction =
            new FieldRefInstruction( opc_putstatic,
                                     anchorClass().qualifiedName(),
                                     "function" +  functionNumber(), //  !FIXME!function
                                     dispatcherSignature().getType().
                                     getSignature() );
        code.plantInstruction( setFunction );
    }

	private void plantGetAnchorClassObject(CodeSequence code) {
		code.plantInstruction(
			new PushLiteralInstruction(anchorClass().qualifiedName().replace('/', '.')));
		code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
							   "java/lang/Class",
							   "forName",
							   "(Ljava/lang/String;)Ljava/lang/Class;"));
	}

    public CClass anchorClass() {
        if (dispatcherMethod.isTopDispatcherInLattice()) {
            return (CSourceGFCollection) outer;
        } else {
            return dispatcherMethod.anchorClass();
        }       
    }

    public int functionNumber() {
        if (dispatcherMethod.isTopDispatcherInLattice()) {
            return number;
        } else {
            return dispatcherMethod.functionNumber();
        }       
    }

    /**
     * Accumulates an array of all the required gf_info data
     * structures for this.  */
    protected void genGenericFunctions(ArrayList accum) {
        completeClassInfo();
        super.genGenericFunctions(accum);
    }

    /** Provide an annotation for any dispatcher classes (RMJ only) */
    public void rmjAnnotate(RMJAnnotation rmjInfo)
    {
        // compute the properties of the multimethods in this
        // dispatcher that are common across all multimethods:

        CDispatcherSignature dsig = dispatcherSignature();
        String sigName = dsig.qualifiedName();
        int pos = sigName.lastIndexOf("$");
        String baseAnchorClassName = sigName.substring(0,pos);

        String kind;
        String anchorClassName = outer.qualifiedName();
        if (baseAnchorClassName.equals(anchorClassName)) {
            // The prefixes match, this is a base anchor
            kind = "B";
        } else {
            // must be a subanchor (or glue, which will be
            // computed by the loader)
            kind = "S";
        }

        String dispatcherClassName = qualifiedName();
                
        String ambigErrorClassName =
            baseAnchorClassName + "$ambig" + 
            sigName.substring(pos + "$signature".length());

        // now build the multimethod-specific annotations
        RMJAnnotation.Method[] annotations = 
            dispatcherMethod.getMethodAnnotations(kind,
                                                  baseAnchorClassName, 
                                                  dispatcherClassName,
                                                  ambigErrorClassName);

        for (int j = 0; j < annotations.length; j++) {
            rmjInfo.add(annotations[j]);
        }
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * The unique number of this within it's outer class, used to
     * construct the identifier and qualified name of this.  */
    private int number;

    private CSourceClass outer;

    private TokenReference where;

    //@ spec_public
    protected final CSourceDispatcherMethod dispatcherMethod;

    /**
     * The $signature interface for this generic function.  */
    //@ spec_public
    protected CDispatcherSignature dispatcherSig;

    /** The ambiguous-method dispatcher class for this generic
     * function. (RMJ only) */
    protected CAmbiguousDispatcherClass ambiguousDispatcherClass;

    /*@ private invariant dispatcherMethod.isTopDispatcherInLattice() 
      @                 ==> infoCompleted;
      @ private invariant infoCompleted <==> dispatcherSig != null;
      @ protected invariant_redundantly dispatcherMethod.isTopDispatcherInLattice() 
      @                         ==> dispatcherSig != null;
      @*/

    /**
     * Indicates whether the info needed for code generation has been
     * completed. 
     *
     * @see #completeClassInfo()
     */
    private boolean infoCompleted = false;

}// CSourceDispatcherClass
