/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceDispatcherMethod.java,v 1.38 2004/04/12 21:00:42 tongjie Exp $
 */

package org.multijava.mjc;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Hashtable;
import org.multijava.relaxed.util.RMJAnnotation;

import org.multijava.util.Destination;
import org.multijava.util.DirectedAcyclicGraph;
import org.multijava.util.classfile.*;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents a dispatcher method wrapping a set of
 * multimethods of the same generic function as read from a source
 * file.  */
public class CSourceDispatcherMethod extends CSourceMethod 
    implements MJSpecialMethod, Cloneable
{
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Constructs a dispatcher method for the given set of
     * multimethods.  
     *
     * <pre><jml>
     * requires multimethods != null && multimethods.length > 0;
     * </jml></pre>
     *
     * @param where	a reference to the source code for annotating bytecode
     * @param outer	the class that hosts this dispatcher, either the
     *			declaring class for internal methods or the anchor
     *			class for external methods
     * @param topConcreteMethod the top concrete method of the generic function
     * @param multimethods the collection of multimethods for which this
     *			this will dispatch
     * @param gfID	the unique ID of this dispatcher's associated gf,
     *			used to give its $body methods a unique name
     */
    public CSourceDispatcherMethod( TokenReference where, Main compiler,
				    CSourceClass outer, 
				    CMethod topConcreteMethod, 
				    CSourceMethod[] multimethods, int gfID ) {
	// typechecking restrictions for dispatcher mates ensure that
	// the modifiers and exceptions for any one multimethod match
	// those of the others
	super( new MemberAccess(multimethods[0].owner(), 
				multimethods[0].host(), 
				multimethods[0].access().modifiers()), 
	       multimethods[0].ident(), 
	       multimethods[0].returnType(), 
	       CSpecializedType.unspecializedTypeTupleFrom( multimethods[0].
							    parameters() ),
	       multimethods[0].throwables(),
	       CTypeVariable.EMPTY,
	       multimethods[0].isDeprecated(), null, 
	       multimethods[0].declarationContext, null );

	this.multimethods = multimethods;
	this.topConcreteMethod = topConcreteMethod;
	this.where = where;

	// sort multimethods by the overrides relation
	DirectedAcyclicGraph dag = 
	    new DirectedAcyclicGraph( multimethods, OVERRIDES_RELATION );
	Object[] sorted = dag.inDFSOrder();
	System.arraycopy(sorted, 0, multimethods, 0, multimethods.length);

	// The most general multimethod of this dispatcher will be
	// the last one (since the multimethods are sorted).
	isTopDispatcherInLattice = 
	    multimethods[ multimethods.length - 1 ] == topConcreteMethod;

        // Add links from multimethods back to this.  Also, give each
	// multimethod its gfID and methodID, to disambiguate its
	// $body method from any other such method
        for (int i = 0; i < multimethods.length; i++) {
            multimethods[i].setDispatcherMethod( this );
	    multimethods[i].setIDs( gfID, i );
        } // end of for

        // Instantiate representations of an inner dispatcher class
        // and a dispatcher signature interface if necessary.
	if (topConcreteMethod.isExternal()) {
	    // make receiver an explicit parameter of the method
	    swallowReceiver();

	    // a dispatcher class is needed
	    dispatcherClass = new CSourceDispatcherClass( compiler, outer, 
							  where, this );
	    setOwner( dispatcherClass );

	    // RMJ: added this, because RMJ generates each method of
	    // an external GF into its own dispatcher class, and
	    // sometimes we want to know not whether a dispatcher is
	    // the top one, but whether it's in the base anchor class.
	    isInBaseAnchorClass = (topConcreteMethod.host() == host());

	    // isTopDispatcherInLattice => isInBaseAnchorClass, for an
	    // external GF
	    assertTrue(!isTopDispatcherInLattice || isInBaseAnchorClass);

	} // end of if 
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Sets the owner of this member.
     *
     * @see #owner
     */
    protected void setOwner( CClass owner ) {
	access().setOwner( owner );
    }

    /**
     * Indicates whether this dispatcher method is the top dispatcher
     * method in the generic function lattice.  */
    public /*@ pure @*/ boolean isTopDispatcherInLattice()  {
        return isTopDispatcherInLattice;
    }

    /* !FIXME! This should be a doc comment, but callable statement fails
     * <pre><jml>
     * also
     * protected code_contract
     *   callable multimethods[*].swallowReceiver();
     * </jml></pre>
     */
    protected void swallowReceiver() {
	super.swallowReceiver();
	for (int i = 0; i < multimethods.length; i++) {
	    multimethods[i].swallowReceiver();
	} // end of for
    }

    /**
     * Returns a string representation of this grouped multimethod.
     *
     * @return	a string representation of this 
     */
    public String toString() {
	ArrayList v = new ArrayList();
	v.addAll( Arrays.asList( multimethods ) );
	String result =  "/* grouped */ " + 
	    super.toString( topConcreteMethod.isExternal() ) + " " + v;
	
	return result;
    }

    /**
     * Returns the dispatcher class for this.
     *
     * <pre><jml>
     * requires dispatcherClass != null;
     * requires_redundantly topConcreteMethod.isExternal();
     * </jml></pre>
     *
     */
    protected /*@ pure @*/ CSourceDispatcherClass dispatcherClass() {
        return dispatcherClass;
    }
    
    /**
     * Returns the $signature class for this.
     *
     * <pre><jml>
     * also
     * requires dispatcherClass != null;
     * requires_redundantly topConcreteMethod.isExternal();
     * </jml></pre>
     *
     */
    public CDispatcherSignature dispatcherSignature() {
	if (isTopDispatcherInLattice) {
	     return dispatcherClass.dispatcherSignature();
	} // end of if 
        return topConcreteMethod.dispatcherSignature();
    }
    
    /**
     * Returns the ambiguous-method dispatcher class for this.  (RMJ
     * only)
     *
     * <pre><jml>
     * also
     * requires dispatcherClass != null;
     * requires_redundantly topConcreteMethod.isExternal();
     * </jml></pre>
     *
     */
    public CAmbiguousDispatcherClass ambiguousDispatcherClass() {
	if (isTopDispatcherInLattice) {
	     return dispatcherClass.ambiguousDispatcherClass();
	} // end of if 
        return topConcreteMethod.ambiguousDispatcherClass();
    }
    
    /**
     * Returns the $anchor class for this generic function.
     *
     * <pre><jml>
     * also
     * requires dispatcherClass != null;
     * requires_redundantly topConcreteMethod.isExternal();
     * </jml></pre>
     *
     */
    public CClass anchorClass() {
	if (isTopDispatcherInLattice) {
	     return dispatcherClass.anchorClass();
	} // end of if 
        return topConcreteMethod.anchorClass();
    }
    
    /**
     * Returns the function field number for this generic function
     * within its $anchor class.
     *
     * <pre><jml>
     * also
     * requires dispatcherClass != null;
     * requires_redundantly topConcreteMethod.isExternal();
     * </jml></pre>
     * */
    public int functionNumber() {
	if (isTopDispatcherInLattice) {
	     return dispatcherClass.functionNumber();
	} // end of if 
        return topConcreteMethod.functionNumber();
    }
    
    /**
     * Returns the filtered dispatcher methods for the external
     * generic function whose top dispatcher is this.  At code generation
     * time these filtered dispatcher methods are used to generate
     * special dispatchers for handling superclass method invocations
     * within the generic function.
     *
     * <pre><jml>
     * also 
     * requires topConcreteMethod instanceof CSourceMethod &&
     *          (\exists int i; i <= 0 && i < multimethods.length;
     *              multimethods[i] == topConcreteMethod);
     * </jml></pre>
     *
     */
    protected CMethodSet filteredDispatchers() {
	return ((CSourceMethod) topConcreteMethod).filteredDispatchers();
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Returns a list of <code>CMethodInfo</code> instances, one for
     * this dispatcher and one for each multimethod body referenced by
     * this.  */
    public ArrayList genMethodInfo() {
	removePrivacyModifierIfNecessary();
	
	boolean isExternalGF = topConcreteMethod.isExternal();

	// If this is a dispatcher for an external generic function
	// then the signature interface implemented by the dispatcher
	// class says that this method must be public.  The privileged
	// access modifiers of the dispatcherClass will maintain
	// encapsulation.
	if (isExternalGF) {
	    setModifiers( (access().modifiers() & ~ACC_PRIVATE & ~ACC_PROTECTED) |
			  ACC_PUBLIC );
	}
	
	ArrayList result = new ArrayList( multimethods.length + 1 );

	// add the method info for the dispatcher
	MethodInfo selfInfo = nonEmptyMethodInfo(isExternalGF ? "apply":ident());
	selfInfo.addAttribute( new DispatcherAttribute() );
	result.add( selfInfo );

	// add method info for the multimethod bodies, which are
	// private members
	short privateMods = 
	    (short) ( access().modifiers() & ~ACC_PUBLIC & ~ACC_PROTECTED 
		      | ACC_PRIVATE
		      //| (isExternalGF ? ACC_STATIC : 0 ) 
		      );
	for( int i = 0; i < multimethods.length; i++ ) {
	    MethodInfo bodyInfo = 
		new CMethodInfo( privateMods, multimethods[i].bodyIdent(),
				 multimethods[i].bodySignature(),
				 multimethods[i].bodyGenericSignature(),
				 multimethods[i].exceptionsAsStringArray(),
				 multimethods[i],
				 isDeprecated(), false );
	    bodyInfo.addAttribute(new MultimethodBodyAttribute());
	    result.add( bodyInfo );
	}

	return result;
    }

    // This method creates the RMJ annotation for each of the
    // multimethods embedded in this dispatcher (RMJ only)
    public RMJAnnotation.Method[]
	getMethodAnnotations(String kind,
			     String baseAnchorClassName, 
			     String dispatcherClassName,
			     String ambigErrorClassName)
    {
	RMJAnnotation.Method[] annotes = 
	    new RMJAnnotation.Method[multimethods.length];

	// the operation name and introducing receiver class is shared
	// by all multimethods in this dispatcher

	String opname = 
	    baseAnchorClassName.substring(0,
					  baseAnchorClassName.length() - 
					  MJ_ANCHOR.length());

	for (int i = 0; i < multimethods.length; i++) {
	    CSourceMethod mm = multimethods[i];

	    // the receiver and argument static types and specializers:
	    CSpecializedType[] parms = mm.parameters();
	    String[] types = new String[parms.length];
	    String[] specs = new String[parms.length];
	    for (int j = 0; j < parms.length; j++) {
		CSpecializedType parm = parms[j];

		CType stat = parm.staticType();
		String type;
		if (stat instanceof CClassType) {
		    // a class type
		    type = stat.toString();
		} else {
		    // a primitive type
		    type = "=" + stat.toString();
		}
		types[j] = type;

		String spec;
		if (parm.isSpecialized()) {
		    CType dyn = parm.dynamicType();
		    if (dyn instanceof CClassType) {
			// a class or array specializer
			spec = dyn.toString();
		    } else if (dyn instanceof CValueType) {
			// a value specializer
			spec = "=" + ((CValueType)dyn).getValue().toString();
		    } else {
			throw new RuntimeException(
			    "Internal error: unexpected specializer: " + dyn);
		    }
		} else {
		    // this position isn't specialized
		    spec = "-";
		}
		specs[j] = spec;
	    }

	    // whether or not this multimethod is abstract:
	    boolean isAbstract = mm.isTopLevelAbstractMethod();

	    annotes[i] = new RMJAnnotation.Method(
		opname.replace('/', '.'),
		types,
		specs, 
		isAbstract, 
		kind,
		baseAnchorClassName.replace('/', '.'),
		dispatcherClassName.replace('/', '.'),
		ambigErrorClassName.replace('/', '.'));
	}

	return annotes;
    }

    /**
     * Generates the generic function information for this required
     * when generating a class file, or returns null if this method
     * requires no such information.  */
    public GenericFunctionInfo genGenericFunctionInfo() {
	short fn;
	String topConcMethColl = null;
	if (dispatcherClass != null) {
	    fn = (short) functionNumber();
	    if (!isTopDispatcherInLattice && !isInBaseAnchorClass) {
		// [RMJ: added the second conjunct, because we also
		// want to skip over any other dispatchers in the base
		// anchor class]
		String anchorName = anchorClass().qualifiedName();
		topConcMethColl = anchorName.substring(0, 
						   anchorName.length() - 
						   MJ_ANCHOR.length());
	    }
	} else {
	    fn = Short.MAX_VALUE;
	}
	
	MultimethodInfo[] mmInfo = new MultimethodInfo[ multimethods.length ];
	for (int i = 0; i < multimethods.length; i++) {
	    CSourceMethod mm = multimethods[i];
	    mmInfo[i] = new MultimethodInfo( (short) mm.access().modifiers(), 
					     ident(),
					     mm.mmSignature(),
					     mm.mmGenericSignature(),
					     fn, 
					     topConcMethColl,
					     mm.exceptionsAsStringArray(),
					     mm.isDeprecated() );
	}
	
	return new GenericFunctionInfo(ident(), this.getSignature(),
				       this.getSignature(),
				       fn, 
				       topConcMethColl, mmInfo);
    }

    /**
     * Generates bytecode to put a reference to the external
     * <code>function</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     *   also
     *   requires isTopDispatcherInLattice();
     * </jml></pre> */
    public void plantFunctionRef( CodeSequence code ) {
        dispatcherClass.plantFunctionRef( code );
    }

    /**
     * Generates bytecode to put a reference to the
     * <code>oldFunction</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     *   also
     *   requires this.isExternal();
     * </jml></pre> 
     * @param code the code sequence to add the bytecode to
     * @param forSuper if true, then this is for a super send, which,
     *                 under RMJ, should start from the oldSuperFunction
     *                 rather than oldFunction
     * @param fromOuter if true, then generates a reference that works from the
     *                  outer class context, otherwise generates a reference that
     *			works from within the dispatcher class
     */
    public void plantOldFunctionRef( CodeSequence code, 
				     boolean forSuper, boolean fromOuter ) {
        dispatcherClass.plantOldFunctionRef( code, forSuper, fromOuter );
    }

    protected Object clone() {
	try {
	    return super.clone();
	} catch( CloneNotSupportedException e ) {
	    fail( "shouldn't get here: " + e.getMessage() );
            //@ unreachable;
            return null;
	}
    }

    /**
     * Plants the bytecode for a dispatcher method that dynamically
     * selects the appropriate multimethod body from the set
     * represented by this.  */
    protected void plantBodyBytecode( CodeSequence code ) {
	plantFilteredDispatcherBody( code, null );
    }

    /**
     * Plants the bytecode for a dispatcher method that dynamically
     * selects the appropriate multimethod body from the set
     * represented by this, but filtered so that only multimethods whose
     * receiver type is a proper supertype of <code>receiverSubclass</code>
     * will be dispatched to.  <code>receiverSubclass</code> may be null
     * in which case no filtering is done.  */
    protected void plantFilteredDispatcherBody( CodeSequence code,
						CClass receiverSubclass ) {
	boolean isExternalGF = topConcreteMethod.isExternal();
	CType receiverSubtype = null;
	if (receiverSubclass != null) {
	    receiverSubtype = receiverSubclass.getType();
	}

	code.setLineNumber( where.line() );

	// generate dispatcher code
    genDisp:
	{
	    // !FIXME!optimize currently just dumb code that repeats tests
	    for( int i=0; i < multimethods.length; i++ ) {
		CType mmRT = multimethods[i].receiverType();
		if ( receiverSubtype != null // Filtering on ...
		     && 	// ... and not a proper supertype, ...
		     !(receiverSubtype.isAlwaysAssignableTo( mmRT )
		       && receiverSubtype != mmRT) ) {
		    // ... so this multimethod should be filtered out.
		    continue;
		}
		
		CodeLabel nextAlternative = new CodeLabel();
		boolean testWasGenerated =
		    multimethods[i].genDispatch( code, nextAlternative, 
						 topConcreteMethod, 
						 getOwnerName() );
		if( !testWasGenerated ) {
		    // We have a problem if an alternative other than
		    // the last does not include a test.  We wouldn't
		    // be able to reach the other alternatives.
		    /* !FIXME!3/3 */
		    if (i != multimethods.length - 1) {
			System.err.println("Found the problem at: " + where);
		    }
		    
		    assertTrue( i == multimethods.length - 1 );
		    // The last alternative covers every possibility,
		    // so we can break past the code generation for
		    // the default alternative
		    break genDisp;
		}
		code.plantLabel( nextAlternative );
	    }
		
	    // Generate the default alternative
	    if( isTopDispatcherInLattice ) {
		// XXX (disabled for now)
		if (false && declarationContext().getCompiler().RMJ()) {
		    // we can get here, if we're at the top
		    // (multimethod) that overrides only abstract
		    // methods.  we must have ignored a load-time
		    // hazard warning, and now we get the ambiguity
		    // error
		    
		    // push new RMJRuntimeException
		    code.plantInstruction(
			  new ClassRefInstruction( opc_new, 
				CStdType.RMJRuntimeException.qualifiedName()));

		    // dup it
		    code.plantInstruction( new NoArgInstruction( opc_dup ) );

		    // push the string error message
		    String message = where + 
			": invoking abstract (possibly implicit) method";
		    code.plantInstruction(new PushLiteralInstruction(message));

		    // invoke the exn initializer
		    MethodRefInstruction exnConst =
			new MethodRefInstruction( 
			      opc_invokespecial,
			      CStdType.RMJRuntimeException.qualifiedName(),
			      JAV_CONSTRUCTOR,
			      CType.genMethodSignature(
			          CStdType.Void, 
				  new CType[] { CStdType.String } ));
		    code.plantInstruction( exnConst );

		    // throw the exn
		    code.plantInstruction(new NoArgInstruction(opc_athrow));

		} else {
		    // out of options
		    plantNoMethodThrow( code );
		    throw new RuntimeException( "!FIXME! Could not find top " +
					    "method for "
					    + topConcreteMethod.toString() + ".  " +
					    "Should not reach this " + 
					    "code because CMethod.genDispatch" 
					    + " should never generate a check"
					    + " for the top method.  If a " + 
					    "check was required for the top " +
					    "method then the generic function"
					    + " is incomplete.  Typechecking"
					    + " should catch that." );
		}
	    } else {
		// No matches here so...
		if( isExternalGF ) {
		    // ... try next closure in the chain of responsibility.
		    dispatcherClass.plantOldFunctionRef( code, true, false );
		    plantSelfArguments( code, topConcreteMethod );
		    topConcreteMethod.genGenFuncInvocation( code, false );
		} else {
		    // ... try super call.
		    code.plantLoadThis();
		    plantSelfArguments( code, topConcreteMethod );
		    MethodRefInstruction superCall =
			new MethodRefInstruction( opc_invokespecial,
						  owner().getSuperClass().
						  qualifiedName(),
						  ident(), getSignature() );
		    code.plantInstruction( superCall );
		} 
		NoArgInstruction returnInst =
		    new NoArgInstruction( returnType().getReturnOpcode() );
		code.plantInstruction( returnInst );
	    }
	} // end of genDisp
    }

    /** 
     * Plant instructions for throwing a NoSuchMethodException onto
     * the given code sequence.  */
    private void plantNoMethodThrow( CodeSequence code ) {
	String excep = "java/lang/Error";
	String msg = "binary incompatibility, incomplete generic function";
	code.plantInstruction( new ClassRefInstruction( opc_new, excep ) );
	code.plantInstruction( new NoArgInstruction( opc_dup ));
	code.plantInstruction( new PushLiteralInstruction( msg ));
	String sig = 
            CType.genMethodSignature( CStdType.Void, 
                                      new CType[] { CStdType.String } );
	code.plantInstruction( new MethodRefInstruction( opc_invokespecial,
							 excep,
							 JAV_CONSTRUCTOR,
							 sig ));
	code.plantInstruction( new NoArgInstruction(opc_athrow) );
    }

    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------

    //@ private invariant multimethods != null && multimethods.length > 0;
    //@ spec_protected
    private final CSourceMethod[] multimethods;

    private TokenReference where;

    private boolean isTopDispatcherInLattice;

    // true only for external methods 
    private boolean isInBaseAnchorClass = false;

    //@ invariant topConcreteMethod.isExternal() <==> dispatcherClass != null;

    /**
     * If this a method of an external generic function, then this
     * field represents the dispatcher inner class for this method.
     * Otherwise this field is <code>null</code>.  */
    //@ spec_public
    protected CSourceDispatcherClass dispatcherClass;

    private static final DirectedAcyclicGraph.EdgeCalculator OVERRIDES_RELATION=
	new DirectedAcyclicGraph.EdgeCalculator() {
	    public boolean edgeExists( Object o1, Object o2 ) {
		CMethod m1 = (CMethod) o1;
		CMethod m2 = (CMethod) o2;
		return m1.isMoreSpecificThan(m2);
	    }
	};
}
