/*
 * Copyright (C) 2004 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceFilteredDispatcherMethod.java,v 1.11 2004/02/23 07:15:07 tongjie Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;

import org.multijava.util.classfile.MethodRefInstruction;

/**
 * This class represents a special dispatcher method for superclass
 * method invocations whose caller is an external method and whose
 * target is the same external generic function to which the caller
 * belongs [see Clifton-01, Figure 24, case s-e4].  The special
 * dispatcher filters the applicable methods to just include those
 * whose receiver type is a proper superclass of the caller's receiver
 * type.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.11 $
 */
public class CSourceFilteredDispatcherMethod extends CSourceMethod 
    implements MJSpecialMethod 
{

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Creates a new <code>CSourceFilteredDispatcherMethod</code> instance.
     *
     * @param target the top method of the target generic function of
     * 			the dispatcher
     * @param host the <code>CSourceGFCollection</code> containing
     * 			target and the dispatcher
     * @param receiverSubclass only methods of the generic function whose
     *			receiver type is a proper supertype of the given class
     *			will be considered applicable by the dispatcher
     */
    public CSourceFilteredDispatcherMethod( CSourceMethod target, 
					    CSourceGFCollection host,
					    CClass receiverSubclass, 
					    int number ) {
	super( new MemberAccess(target.owner(), host, ACC_PRIVATE), 
	       "ident() overridden",
	       target.returnType(),
	       CSpecializedType.unspecializedTypeTupleFrom(target. parameters()),
	       target.throwables(),CTypeVariable.EMPTY,
	       target.isDeprecated(), null, 
	       null, null );

	this.topConcreteMethod = this;
	this.target = target;
	this.receiverSubclass = receiverSubclass;
	this.number = number;
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public String ident() {
	return target.ident() + "$filter" + number;
    }

    // -------------------------------------------------------------
    // GENERATE BYTECODE
    // -------------------------------------------------------------

    /**
     * Generates bytecode to put a reference to the external
     * <code>function</code> field for the external generic function
     * on the JVM stack.  Overridden here to plant a self reference
     * since for a local superclass method invocation on an external
     * generic function <code>this</code> is the last dispatcher in
     * the chain of responsibility. 
     */
    public void plantFunctionRef( CodeSequence code ) {
	code.plantLoadThis();
    }

    /**
     * Generates a sequence of bytecode for invoking this method,
     * assuming appropriate argument values have already been pushed
     * to the JVM stack.
     * @param	code		the code sequence
     * @param nonVirtual force non-virtual dispatching if possible (ignored if
     *			 this method belongs to an interface or is external)
     */
    public void genGenFuncInvocation( CodeSequence code, boolean nonVirtual ) {
	// always use invokespecial since the method represented by
	// this is private
	code.plantInstruction(
	    new MethodRefInstruction( opc_invokespecial, 
				      target.dispatcherMethod().
				      dispatcherClass.qualifiedName(),
				      ident(), target.getSignature() ));
    }


    /**
     * Generates the method info record for the .class file.
     */
    public ArrayList genMethodInfo() {
	swallowReceiver();
	ArrayList result = new ArrayList(1);
	result.add( nonEmptyMethodInfo( ident() ) );
	return result;
    }

    /**
     * Plants the bytecode for a dispatcher method that dynamically selects 
     * the appropriate multimethod body from the generic function to which
     * {@link #target} belongs, filtered by receivers that are proper supertypes
     * of {@link #receiverSubclass}.
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	target.dispatcherMethod().plantFilteredDispatcherBody(code, 
							      receiverSubclass);
    }
    
    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * Only multimethods in the same generic function as this and
     * whose receiver type is a proper supertype of
     * {@link #receiverSubclass} will be considered by the generated
     * dispatch code.
     */
    private final CSourceMethod target;

    /**
     * Only multimethods in the same generic function as
     * {@link #target} and whose receiver type is a proper supertype of
     * this will be considered by the generated dispatch code.
     */
    private final CClass receiverSubclass;


    /**
     * A unique integer for this filtered dispatcher within the host
     * generic function collection.  Used to generate a unique method
     * name.
     * @see #ident()
     */
    private final int number;
}
