/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceGFCollection.java,v 1.18 2005/04/26 02:39:31 kui_dai Exp $
 */

package org.multijava.mjc;

import java.io.IOException;

import java.util.Map;
import java.util.HashSet;

import org.multijava.util.classfile.ClassFileFormatException;
import org.multijava.util.classfile.MethodInfo;
import org.multijava.util.classfile.FieldInfo;
import org.multijava.util.classfile.AnchorAttribute;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

import org.multijava.util.Destination;
import org.multijava.util.classfile.AttributeList;
import org.multijava.util.classfile.GenericFunctionInfo;
import org.multijava.util.classfile.GenericFunctionsAttribute;

/**
 * This class represents a single generic function collection, which
 * encapsulates all the top-level methods declared in a given context.
 * It includes methods for adding the methods of the collection to the
 * classes that they augment. */
public class CSourceGFCollection extends CSourceClass 
    implements CGenericFunctionCollection 
{

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    public CSourceGFCollection( Main compiler,
                                CClass owner,
                                CMemberHost host,
                                TokenReference where,
                                long modifiers,
                                String ident,
                                String qualifiedName,
                                boolean isAnonymous, boolean isMember,
                                boolean deprecated,
                                CMethod[] tlMethods )
    {
        super( compiler, new MemberAccess(owner, host, modifiers), where, 
               ident, qualifiedName, CTypeVariable.EMPTY,isAnonymous, isMember, deprecated );
        this.collection = new CGenericFunctionCollection.Impl( tlMethods );
    }

    //---------------------------------------------------------------------
    // PROCESS AUGMENTING METHODS
    //---------------------------------------------------------------------

    /**
     * Augments the classes given by the receivers of methods in this
     * collection for the given context.  For each method <em>m</em>
     * in this collection, a mapping from the given context to
     * <em>m</em> is added to the <code>CClass</code> representing
     * <em>m</em>'s receiver type.
     *
     * @param   context         the context in which the augmentation occurs
     */
    public void applyAugmentationFor( CContextType context ) {
        collection.applyAugmentationFor( context );
    }

    public String toString() {
        return "CSourceGFCollection: " + collection;
    }


    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------


    /**
     * Generates a special superclass method invocation dispatcher for
     * the external generic function containing <CODE>target</CODE>.
     * The dispatcher filters out methods of the generic function
     * whose receiver type is not a proper superclass of
     * <CODE>receiverSubclass</CODE>.
     *
     */
    //@ requires target != null;
    //@ requires target.topConcreteMethod instanceof CSourceMethod;
    // FIXME: had    requires target.getHostClass() == this;
    // FIXME: had    requires target.topConcreteMethod.getHostClass() == this;
    protected CMethod superclassMethDispFor( CMethod target,
                                             CClass receiverSubclass ) 
        throws UnpositionedError
    {
        CSourceMethod targetTopMethod = 
            (CSourceMethod) target.topConcreteMethod();

        // Construct a special dispatcher method whose multimethods
        // are taken from target.topConcreteMethod()'s
        // dispatcherMethod but are filtered.  The bytecode for this
        // dispatcher method is generated lazily since the set of
        // multimethods used to construct target.topConcreteMethod()'s
        // dispatcherMethod is not formed until the translateMJ pass
        // whereas this method is called in the typecheck pass.
        CSourceFilteredDispatcherMethod m =
            new CSourceFilteredDispatcherMethod( 
                targetTopMethod, this, receiverSubclass, nextFilterNumber++ );
        
        targetTopMethod.addFilteredDispatcher( m );

        return m;
    }

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    /**
     * Appends any custom attributes for this class to the attribute
     * list represented by <code>accum</code>.  */
    protected void appendCustomAttributes(AttributeList accum) {
        accum.add(new AnchorAttribute());
        super.appendCustomAttributes(accum);
    }

    /**
     * Returns an array representing all the methods for bytecode. For
     * an external generic function collection such as this, the array
     * includes one <code>getFunction</code> accessor for each generic
     * function. */
    protected MethodInfo[] genMethods() {
        CMethodSet ms = methodsForCodeGen();
        
        // add an accessor method for each generic function
        CClassType[] inners = getInnerClasses();
        for (int i = 0; i < inners.length; i++) {
            CClass cl = inners[i].getCClass();
            if (cl instanceof CSourceDispatcherClass) {
                ms.addMethod(
                    ((CSourceDispatcherClass) cl).functionAccessorMethod() );
            }
        }
        
        return methods.buildMethodInfo();
    }

    /**
     * Returns an array representing all the fields for bytecode. For
     * an external generic function collection such as this, the array
     * includes one <code>function</code> field for each generic
     * function. */
    protected FieldInfo[] genFields() {
        CFieldTable fields = fieldsForCodeGen();

        // add a field for each generic function
        CClassType[] inners = getInnerClasses();
        for (int i = 0; i < inners.length; i++) {
            CClass cl = inners[i].getCClass();
            if (cl instanceof CSourceDispatcherClass) {
                fields.add( ((CSourceDispatcherClass) cl).functionField() );
            }
        }

        return fields.buildFieldInfo();
    }

    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------

    /**
     * Delegee for the storage of augmenting methods and the
     * application thereof. */
    private CGenericFunctionCollection.Impl collection;


    /**
     * Used to generate unique names for filtered dispatchers.
     */
    private int nextFilterNumber = 0;
}
