/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceMethod.java,v 1.33 2005/09/08 07:15:07 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.*;

import java.util.ArrayList;
import org.multijava.util.compiler.UnpositionedError;
import java.util.Hashtable;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.FastStringBuffer;

/**
 * This class represents a method read from a source code (*.java) file.
 * It stores the fully qualified type signature and the method body.
 */
public class CSourceMethod extends CMethod implements SigSourceMember {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a representation of the method as read from the source 
     * file including the type signature and the method body
     *
     * @param	access		contains the info used to determine accessibility of this method
     * @param	ident		the ident of this method
     * @param	returnType	the return type of this method
     * @param	paramTypes	the parameter types of this method
     * @param	exceptions	a list of all exceptions in the throws list
     * @param	deprecated	is this method deprecated
     * @param	body		the source code
     * @param	declarationContext the context in which this method is declared
     * @param declarationASTNode the AST node corresponding to the
     * 				declaration of this, may be null
     */
    public CSourceMethod( MemberAccess access, 
			  String ident, CType returnType,
			  CSpecializedType[] paramTypes, 
			  CClassType[] exceptions,
			  CTypeVariable[] typevariables,
			  boolean deprecated, JBlock body,
			  CContextType declarationContext, 
			  JMethodDeclaration declarationASTNode )
    {
	super( access, ident, returnType, 
	       paramTypes, exceptions, typevariables,
	       deprecated, declarationContext );

	this.body = body;
	this.optimizingLevel = CTopLevel.getCompiler().optimizeCode();
	this.declarationASTNode = declarationASTNode;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Checks whether this method is used.  Non-private methods are always
     * assumed to be used.
     *
     * @return		true if this method should be consider used
     */
    public /*@ pure @*/ boolean isUsed() {
	return used || !isPrivate() || ident().indexOf("$") >= 0; // $$$
    }

    /**
     * Sets a flag indicating that this method is used
     */
    public void setUsed() {
	used = true;
    }

    /**
     * Returns the body of this method
     */
    public /*@ pure @*/ JBlock body() {
	return body;
    }

    /**
     * Returns the optimizing level of this.  */
    public int optimizingLevel() {
	return optimizingLevel;
    }

    /**
     * Sets the dispatcher method for this multimethod.
     *
     * @param dispatcherMethod a <code>CSourceDispatcherMethod</code> value
     */
    protected void setDispatcherMethod(CSourceDispatcherMethod dispatcherMethod)
    {
        this.dispatcherMethod = dispatcherMethod;
    }


    /**
     * Returns the dispatcher method for this multimethod.
     */
    protected /*@ pure @*/ CSourceDispatcherMethod dispatcherMethod() {
	return dispatcherMethod; 
    }

    /**
     * Returns the name of the dispatcher class for this multimethod.
     *
     * <pre><jml>
     * requires topConcreteMethod.isExternal();
     * </jml></pre>
     *
     */
    protected String dispatcherClassName() {
	return dispatcherMethod.getOwnerName();
    }

    /**
     * Stores a reference to the given filtered dispatcher method.  At
     * code generation time this filtered dispatcher method will be
     * used to generate a special dispatcher for handling superclass
     * method invocations within the external generic function whose
     * top method this is.
     */
    protected void addFilteredDispatcher( CSourceFilteredDispatcherMethod fdm ) {
	filteredDispatchers.addMethod( fdm );
    }

    /**
     * Returns the filtered dispatcher methods for the external
     * generic function whose top method is this.  At code generation
     * time these filtered dispatcher method are used to generate
     * special dispatchers for handling superclass method invocations
     * within the generic function.
     */
    protected /*@ pure @*/ CMethodSet filteredDispatchers() {
	return filteredDispatchers;
    }


    /**
     * Return a string representation of this.
     */
    public String toString() {
	return toString( dispatcherMethod != null && 
			 dispatcherMethod.topConcreteMethod.isExternal() );
    }


    /**
     * Returns the AST node for the declaration corresponding to this
     * signature.  May return null if mapping to declaration is not
     * available.  For example, an objects representing a synthetic
     * method might be a subtype of CSourceMethod, yet such an object
     * does not have a corresponding AST node.
     *
     * <pre><jml>
     * also 
     * ensures \result instanceof JMethodDeclaration;
     * </jml></pre>
     *
     */
    public /*@ pure @*/ JMemberDeclaration declarationASTNode() {
	return declarationASTNode;
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    public ArrayList genMethodInfo() {
	MethodInfo mi;
	
        removePrivacyModifierIfNecessary();
	String stro = getGenericSignature();
	ArrayList result = new ArrayList( 1 );
	if( body != null ) {
	    mi = nonEmptyMethodInfo(ident());
	} else {
	    mi = createMethodInfo(access().modifiers(),
                                         ident(),
                                         getSignature(),
					 getGenericSignature(),
					 exceptionsAsStringArray(),
                                         null,
                                         isDeprecated(),
                                         false);
	}

	// AHS: choose either attributes or Java 5 annotations to store
	// JUTS modifiers
	if (CTopLevel.getCompiler().universeBytecode() ||
            CTopLevel.getCompiler().universeBytecodeAnnotations() ) {
	    
	    // generate byte constant array for parameter modifiers
	    CSpecializedType params[] = parameters();
	    byte paramcodes[] = new byte[params.length];
            
	    for (int i = 0; i < params.length; i++) {
	        if (params[i].staticType().isClassType()) {
	            paramcodes[i] = ((CClassType) params[i].staticType()).getUniverseByteConstant();
	        } else {
	            paramcodes[i] = CUniverseImplicitPeer.byteConstant;
	        }
	    }
            
	    if (CTopLevel.getCompiler().universeBytecodeAnnotations()) {
	        // build annotations for purity and return type universe
	        ArrayList annots = new ArrayList();

                if (isPure())
	            annots.add(new Annotation("org.multijava.universes.UniversePure"));
	        
	        if (returnType().isClassType()) {
	            Annotation retType = new Annotation("org.multijava.universes.UniverseType");
	            retType.addElement("value", new AnnotByteElementValue( (byte)
	                    ((CClassType)returnType()).getUniverseByteConstant()));	            
	            annots.add(retType);
	        }
	        
	        if (annots.size() > 0)
	            mi.addAttribute(new RuntimeVisibleAnnotationsAttribute(annots));
	        
	        // build parameter annotations
	        ArrayList[] paramAnnots = new ArrayList[params.length];
	        for (int i = 0; i < paramAnnots.length; i++) {
	            paramAnnots[i] = new ArrayList();
	            Annotation paramType = new Annotation("org.multijava.universes.UniverseType");
	            paramType.addElement("value", new AnnotByteElementValue(paramcodes[i]));
	            paramAnnots[i].add(paramType);
	        }
	        
	        if (paramAnnots.length > 0)
	            mi.addAttribute(new RuntimeVisibleParameterAnnotationsAttribute(paramAnnots));
	    }

            if (CTopLevel.getCompiler().universeBytecode() ) {
                // Use normal Universe method attribute
	        UniverseMethodAttribute uma;
                
	        if (returnType().isClassType()) {
	            uma = new UniverseMethodAttribute(isPure(), ((CClassType)returnType()).getUniverseByteConstant(), paramcodes);
	        } else {
	            uma = new UniverseMethodAttribute(isPure(), paramcodes);
	        }
	        mi.addAttribute(uma);	        
	    }
	}
	
	result.add(mi);
	return result;
    }
    
    protected MethodInfo nonEmptyMethodInfo( String id ) {
	return createCMethodInfo(access().modifiers(), 
				 id, getSignature(),
				 getGenericSignature(),
				 exceptionsAsStringArray(), this,
				 isDeprecated(), false );
    }

    /**
     * Creates a method info object. This is a factory method that may
     * be overridden by the subclass.
     */
    protected MethodInfo createCMethodInfo(long modifiers, 
					   String name,
					   String type,
					   String genericsignature,
					   String[] exceptions, 
					   CSourceMethod method,
					   boolean deprecated,
					   boolean synthetic) {
        return new CMethodInfo((short) modifiers, name, type, 
			       genericsignature, exceptions,
                               method, deprecated, synthetic);
    }
    
    /**
     * Creates a method info object. This is a factory method that may
     * be overridden by the subclass.
     */
    protected MethodInfo createMethodInfo(long modifiers, 
                                          String name,
                                          String type,
					  String genericsignature,
                                          String[] exceptions, 
                                          CodeInfo code,
                                          boolean deprecated,
                                          boolean synthetic) {
        return new MethodInfo((short) modifiers, name, type, 
			      genericsignature,exceptions,
			      code, deprecated, synthetic);
    }

    /**
     * Returns a string array of the fully qualified names of the
     * exceptions throwable by this.  Provided as a utility method so
     * subclasses of this can easily override
     * <Code>genMethodInfo</code>.
     * 
     * @see #genMethodInfo() */
    protected String[] exceptionsAsStringArray() {
	CClassType[]	excs = throwables();
	String[]	exceptions = new String[excs.length];
	for (int i = 0; i < excs.length; i++) {
	    exceptions[i] = excs[i].qualifiedName();
	}
	return exceptions;
    }

    /**
     * Generates the generic function information for this required
     * when generating a class file, or returns null if this method
     * requires no such information.  */
    public GenericFunctionInfo genGenericFunctionInfo() {
	// a method requires a generic function info entry in its class file
	// if and only if it is a dispatcher method
	return null;
    }

    /**
     * Generates bytecode to put a reference to the external
     * <code>function</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     * also
     * requires topConcreteMethod == this;
     * requires this.isExternal();
     * requires_redundantly dispatcherMethod != null;
     * </jml></pre> */
    public void plantFunctionRef( CodeSequence code ) {
        dispatcherMethod.plantFunctionRef( code );
    }

    /**
     * Generates bytecode to put a reference to the
     * <code>oldFunction</code> field for the external generic function
     * on the JVM stack.
     *
     * <pre><jml>
     *  also
     *   requires this.isExternal();
     * </jml></pre> 
     * @param code the code sequence to add the bytecode to
     * @param forSuper if true, then this is for a super send, which,
     *                 under RMJ, should start from the oldSuperFunction
     *                 rather than oldFunction
     * @param fromOuter if true, then generates a reference that works from the
     *                  outer class context, otherwise generates a reference that
     *			works from within the dispatcher class
     */
     public void plantOldFunctionRef( CodeSequence code,
				      boolean forSuper, boolean fromOuter ) {
        dispatcherMethod.plantOldFunctionRef( code, forSuper, fromOuter );
    }

   /**
     * Returns the $signature class for this.
     *
     * <pre><jml>
     * also 
     * requires this.isExternal();
     * requires_redundantly dispatcherMethod != null;
     * </jml></pre> */
    public CDispatcherSignature dispatcherSignature() {
        if (topConcreteMethod != this) {
            return topConcreteMethod.dispatcherSignature();
        } // end of if 
        return dispatcherMethod.dispatcherSignature();
    }

   /**
     * Returns the ambiguous-method dispatcher class for this.  (RMJ only.)
     *
     * <pre><jml>
     * also
     * requires this.isExternal();
     * requires_redundantly dispatcherMethod != null;
     * </jml></pre> */
    public CAmbiguousDispatcherClass ambiguousDispatcherClass() {
        if (topConcreteMethod != this) {
            return topConcreteMethod.ambiguousDispatcherClass();
        } // end of if 
        return dispatcherMethod.ambiguousDispatcherClass();
    }

    /**
     * Returns the $anchor class for this generic function.
     *
     * <pre><jml>
     * also 
     * requires this.isExternal();
     * requires_redundantly dispatcherMethod != null;
     * </jml></pre>
     *
     */
    public CClass anchorClass() {
	if (topConcreteMethod != this) {
	     return topConcreteMethod.anchorClass();
	} // end of if 
        return dispatcherMethod.anchorClass();
    }
    
    /**
     * Returns the function field number for this generic function
     * within its $anchor class.
     *
     * <pre><jml>
     * also
     * requires this.isExternal();
     * requires_redundantly dispatcherMethod != null;
     * </jml></pre>
     * */
    public int functionNumber() {
	if (topConcreteMethod != this) {
	     return topConcreteMethod.functionNumber();
	} // end of if 
        return dispatcherMethod.functionNumber();
    }

    /**
     * Generates a string giving the type signature of this method.
     *
     * @return the type signature of this method
     */
    public String getSignature() {
  	if( owner().isNestedType() && isConstructor() ) {
	    // Some nested classes (anonymous classes and local class
	    // declarations) have access to final local variables.
	    // This access is accomplished in bytecode by passing the
	    // values of the variables as synthetic parameters to the
	    // nested class's constructor.  This code adds these extra
	    // variables to the signature of the constructor
	    CType[] params = 
		CSpecializedType.staticTypeTupleFrom( null, parameters() );
  	    params = ((CSourceClass)owner()).genConstructorArray( params );
  	    return CType.genMethodSignature( returnType(), params );
  	} else {
  	    return super.getSignature();
  	}
    }

    public String getGenericSignature() {
   	if( owner().isNestedType() && isConstructor() ) {
	    CType[] params = 
		CSpecializedType.staticTypeTupleFrom( null, parameters() );
	    params = ((CSourceClass)owner()).genConstructorArray( params );
	    return CType.genGenericMethodSignature( getTypeVariable(),returnType(), params );
	} else {
  	    return super.getGenericSignature();
	}
    }
        
    /**
     * Generates the bytecode for this method
     *
     * @return the bytecode info for this method
     */
    public CodeInfo genCode() {
	CodeSequence code = CodeSequence.getCodeSequence();

	plantBodyBytecode( code );
	
	CodeInfo info = new CodeInfo( code.getInstructionArray(),
				      code.getHandlers(),
				      optimizingLevel >  1 ? 
				          null : code.getLineNumbers(),
				      null);
	code.release();
	body = null;

	return setParmsAndOptimizeCode( info );
    }


    /**
     * Plants the bytecode for the body of this method.  Separated
     * from <code>genCode()</code> so subclasses my replace standard
     * body bytecode with specialized code.  
     *
     * @see #genCode()
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	if( body == null ) {
	    fail( "trying to generate code for \"body-less\" method " +
		  this.toString() );
	}

	code.setLineNumber(body.getTokenReference().line());
	body.genCode(code);
	if (returnType() == CStdType.Void) {
	    code.ensureFinalReturn();
	}
    }


    protected CodeInfo setParmsAndOptimizeCode( CodeInfo info ) {
	CSpecializedType[] parameters = parameters();
	int paramCount = 0;

	for (int i = 0; i < parameters.length; i++) {
	    paramCount += parameters[i].getSize();
	}
	paramCount += returnType().getSize();
	paramCount += isStatic() ? 0 : 1;

	info.setParameterCount(paramCount);

	if (optimizingLevel >= 10) {
	    return 
		org.multijava.util.backend.Optimizer.
		optimize( null, info, 
			  new org.multijava.util.backend.BackendOptions() );
	} else if (optimizingLevel >= 1) {
	    return org.multijava.util.optimize.Optimizer.
		optimize( info, optimizingLevel);
	} else {
	    return info;
	}
    }

	/* If this method (or one of its dispatcher mates) is private but is
	   accessed by an enclosing or nested class, remove its privacy
	   modifier to hack around the Java verifier.
	   Currently used only for external methods and constructors.
	*/
    protected void removePrivacyModifierIfNecessary() {
	try {
	    if ( topConcreteMethod().needsPrivacyModifierRemoved ) {
		this.access().removePrivateModifier();
	    }
	    
	} catch(UnpositionedError e) {
	    fail("should never reach this code");
	}
    }

    public /*@ pure @*/ boolean isTopLevelAbstractMethod() {
	return _isTopLevelAbstractMethod;
    }
    
    public void markAsTopLevelAbstractMethod() {
	_isTopLevelAbstractMethod = true;
    }    

    protected JBlock getBody() {
        return body;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	JBlock		body;
    private	boolean		used;
    private	int		optimizingLevel;

    //@ spec_public
    private CSourceDispatcherMethod dispatcherMethod;
    
    private      boolean         _isTopLevelAbstractMethod = false;
    
    /**
     * Stores the filtered dispatcher methods for the external generic
     * function whose top method is this.  At code generation time
     * these filtered dispatcher method are used to generate special
     * dispatchers for handling superclass method invocations within
     * the generic function.
     */
    private CMethodSet filteredDispatchers = new CMethodSet();

    /**
     * The AST node corresponding to the declaration of this, may be
     * null.
     */
    private JMethodDeclaration declarationASTNode;
}
