/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSourceRedirectorMethod.java,v 1.13 2005/04/26 02:39:32 kui_dai Exp $
 */

package org.multijava.mjc;

import java.util.Arrays;
import java.util.ArrayList;

import org.multijava.util.classfile.MethodInfo;
import org.multijava.util.classfile.CodeInfo;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.LocalVarInstruction;
import org.multijava.util.classfile.RedirectorAttribute;

/**
 * This class represents a redirector method for a pleomorphic method
 * that belongs to an internal and external generic function.  */
public class CSourceRedirectorMethod extends CSourceMethod 
    implements MJSpecialMethod
{
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Constructs a redirector for the given method, assuming the
     * given declaration context. */
    public CSourceRedirectorMethod( CMethod target,
				    TokenReference where,
				    CContextType context ) {
	this( target.owner(), target.host(), target.access().modifiers(), 
	      target.ident(), target.returnType(), 
	      CSpecializedType.
	      unspecializedTypeTupleFrom( target.parameters() ),
	      target.throwables(), target.isDeprecated(), null, context );
	this.target = target;
	this.where = where;
    }

    /**
     * Constructs a redirector for the given source method. */
    public CSourceRedirectorMethod( CSourceMethod target,
				    TokenReference where ) {
	this( target, where, target.declarationContext );
    }

    /**
     * Constructs a redirector method.
     *
     * @param	owner		the owner of this method
     * @param	host		the lexical host for the decl of this method
     * @param	modifiers	the modifiers on this method
     * @param	ident		the ident of this method
     * @param	returnType	the return type of this method
     * @param	paramTypes	the parameter types of this method
     * @param	exceptions	a list of all exceptions in the throws list
     * @param	deprecated	is this method deprecated
     * @param	body		the dispatching source code
     * @param	declarationContext the context in which this is declared
     */
    private CSourceRedirectorMethod( CClass owner,
				     CMemberHost host,
				     long modifiers,
				     String ident,
				     CType returnType,
				     CSpecializedType[] paramTypes,
				     CClassType[] exceptions,
				     boolean deprecated,
				     JBlock body,
				     CContextType declarationContext)
    {
	super( new MemberAccess(owner, host, modifiers), ident, returnType, 
	       paramTypes, exceptions,CTypeVariable.EMPTY, deprecated, body, 
	       declarationContext, null );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns a string representation of this grouped multimethod.
     *
     * @return	a string representation of this 
     */
    public String toString() {
	String result =  "/* redirected */ " + super.toString();
	return result;
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Overrides superclass method so that the redirection code will
     * be generated for this method, even though <code>body</code> is
     * null.
     */
    public ArrayList genMethodInfo() {
	ArrayList result = new ArrayList( 1 );
	MethodInfo selfInfo = nonEmptyMethodInfo( this.ident() );
	selfInfo.addAttribute( new RedirectorAttribute() );
	result.add(selfInfo);
	return result;
    }

    /**
     * Plants the bytecode for the body of this method.  Separated
     * from <code>genCode()</code> so subclasses may replace standard
     * body bytecode with specialized code.
     *
     * @see #genCode()
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	code.setLineNumber( where.line() );

	// We know that the target generic function is external, else
	// this redirector would not be needed.  We also know that the
	// target method is not static or it wouldn't be pleomorphic
	// with an interface method.  (Ambiguity checking ensures no
	// pleomorphism between concrete generic functions.)

	// !FIXME!RMJ The interaction of RMJ and pleomorphic methods
	// !hasn't been studied and isn't handled here.

	code.plantLoadThis();
	target.topConcreteMethod.plantFunctionRef( code );
	// now swap the receiver and the dispatcher function so
	// they are in the correct order on the stack.  it's necessary
	// to generate them in the wrong order so that all relevant
	// dispatchers will be added to the chain before loading the
	// dispatcher function
	code.plantSwap();
	
	// put the arguments of the redirector onto the stack
	// Used to calculate position of parameters:
	int pos = isStatic() ? 0 : 1;
	CSpecializedType[] parms = parameters();
	for( int i=0; i < parms.length; i++ ) {
	    CSpecializedType t = parms[i];
	    LocalVarInstruction varref = 
		new LocalVarInstruction( t.staticType().getLoadOpcode(), pos );
	    code.plantInstruction( varref );
	    pos += t.getSize();	    
	}	
	boolean forceNonVirtual = false;
	target.genGenFuncInvocation( code, forceNonVirtual );

	NoArgInstruction returnInst =
	    new NoArgInstruction( returnType().getReturnOpcode() );
	code.plantInstruction( returnInst );
    }
	
    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------

    /**
     * The target method of this redirector. */
    private CMethod target;

    /**
     * The location in the source code that caused this redirector to
     * be generated. */
    private TokenReference where;
}
