/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSpecializedType.java,v 1.23 2008/03/16 15:52:44 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.Stack;

import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.FastStringBuffer;

/**
 * This class represents a specialized type like
 * <code>type1@type2</code>. */
/* Implementation: This is essentially a pair of <code>CType</code>
 * objects. Convenience methods call <code>CType</code> methods on the
 * appropriate half of the specialized type.  For example, the
 * <code>isAlwaysAssignableTo</code> method of this calls the
 * isAlwaysAssignableTo method of the dynamic type half of the
 * specialized type. */
public class CSpecializedType extends org.multijava.util.Utils 
    implements CTypeSignatureAppender
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a type specification without an explicit specializer. */
    public CSpecializedType( CType type ) {
	assertTrue( type != null );
	this.staticType = type;
	this.dynamicType = type;
	this.isSpecialized = false;
    }

    /**
     * Constructs a specialized type specification.
     *
     * @param staticType	the static type of the specification,
     *				must not be null
     * @param specializerType	the dynamic type of the specification, 
     *				must not be null */
    public CSpecializedType( CType staticType, CType specializerType ) {
	assertTrue( staticType != null );
	assertTrue( specializerType != null );
	this.staticType = staticType;
	this.dynamicType = specializerType;
	this.isSpecialized = true;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the static type part of this specialized type.
     */
    public /*@ pure @*/ CType staticType() {
	return staticType;
    }

    /**
     * Return the dynamic type of this specialized type.
     */
    public /*@ pure @*/ CType dynamicType() {
	return dynamicType;
    }

    /**
     * Checks whether this specialized type is made up of the same
     * pair of types as a given specialized type.
     *
     * @param other	another specialized type to compare with
     * @return	true iff this specialized type is made up of the same
     * 		pair of types as <code>other</code>     */
    public boolean equals( Object other, CClassType[] args ) {
	if( !(other instanceof CSpecializedType) ) {
	    return false;
	} else {
	    CSpecializedType ot = (CSpecializedType) other;
	    return staticType.equals( ot.staticType,args )
		&& (isSpecialized == ot.isSpecialized)
		&& (!isSpecialized || dynamicType.equals( ot.dynamicType, 
							  args));
	}
    }

    // WMD: a version of equals for which you can disable the
    // Universe checks
    public boolean equals( Object other, CClassType[] args,
			   boolean enableUniv ) {
	if( !(other instanceof CSpecializedType) ) {
	    return false;
	} else {
	    CSpecializedType ot = (CSpecializedType) other;
	    return staticType.equals( ot.staticType,args, enableUniv )
		&& (isSpecialized == ot.isSpecialized)
		&& (!isSpecialized || dynamicType.equals( ot.dynamicType, 
							  args, enableUniv));
	}
    }

    // WMD: a version of equals for which you can disable the
    // Universe checks
    public boolean equals( Object other, boolean enableUniv ) {
	//	return equals(other, CClassType.EMPTY);
	if( !(other instanceof CSpecializedType) ) {
	    return false;
	} else {
	    CSpecializedType ot = (CSpecializedType) other;
	    return staticType.equals( ot.staticType, enableUniv )
		&& (isSpecialized == ot.isSpecialized)
		&& (!isSpecialized || dynamicType.equals( ot.dynamicType, enableUniv )); 
							  
	}
    }
    
    public boolean equals(/*@ nullable @*/ Object other){
	// WMD: for efficiency I left this method, is equivalent to
	//	return equals(other, true);
	if( !(other instanceof CSpecializedType) ) {
	    return false;
	} else {
	    CSpecializedType ot = (CSpecializedType) other;
	    return staticType.equals( ot.staticType )
		&& (isSpecialized == ot.isSpecialized)
		&& (!isSpecialized || dynamicType.equals( ot.dynamicType)); 
							  
	}
    }

    /**
     * Returns the hashcode for this.
     *
     * @return	the hashcode for this
     */
    public int hashCode() {
	int result = staticType.hashCode();
	if( isSpecialized ) {
	    result = result * 31 + dynamicType.hashCode();
	}
	return result;
    }

    /**
     * Checks whether this specialized type is a specializing type
     * for the given specialized type
     *
     * @param other	another specialized type to compare with
     * @return	true iff this specialized type specializes the
     * 		type given by <code>other</code>     */
    public boolean specializes( CSpecializedType other, CClassType[] args ) {
	return isSpecialized	// this must be specialized
	    && staticType.equals( other.staticType, args )
	    && ( !other.isSpecialized()	 // case 1: other is not specialized
		 ||
		 isAlwaysAssignableTo( other,args ) // case 2: other is specialized
		 );
    }

    /**
     * Checks whether a method with this specialized type in a
     * parameter position could overload a method with
     * <code>other</code> in the same position.
     */
    public boolean couldOverload( CSpecializedType other ) {
	return !this.isSpecialized
	    && !other.isSpecialized
	    && (this.staticType.equals(other.staticType)
		|| (this.staticType.isReference() 
		    && !this.staticType.isInterface()
		    && this.staticType.isAlwaysAssignableTo(other.staticType)));
    }

    public /*@ pure @*/ boolean isSpecialized() {
	return isSpecialized;
    }

    /**
     * Transforms this type to a string
     */
    public String toString() {
	if( isSpecialized ) {
	    return staticType + dynamicType.specializerSymbol() + dynamicType;
	} else {
	    return staticType.toString();
	}
    }

    /**
     * Transforms this type to a verbose string including universe
     * annotations.
     * Added by WMD.
     */
    public String toVerboseString() {
	if( isSpecialized ) {
	    return staticType.toVerboseString() + dynamicType.specializerSymbol() + dynamicType.toVerboseString();
	} else {
	    return staticType.toVerboseString();
	}
    }

    /**
     * Returns an unspecialized type with the same static type as
     * this.  */
    public CSpecializedType unspecializedType() {
	return new CSpecializedType( staticType );
    }

    // !FIXME! House cleaning: all these methods are commented out so
    // I can convert them to calls on the appropriate half of the
    // specialized type.

    //      /**
    //       * Transforms this type to a string
    //       * @return	the class file form of this type
    //       */
    //      public String getSignature() {
    //  	FastStringBuffer buff = StringBuffers.getStringBuffer();

    //  	appendSignature(buff);
    //  	String s = buff.toString();
    //  	StringBuffers.release(buff);

    //  	return s;
    //      }

    /**
     * Transforms this type to a string
     */
    public void appendSignature(FastStringBuffer buff) {
	// for ease of parsing the specializer symbol comes first
	if (isSpecialized) {
	    buff.append(dynamicType.specializerSymbol());
	}
	staticType.appendSignature(buff);
	if (isSpecialized) {
	    dynamicType.appendSignature(buff);
	}
    }

    public void appendGenericSignature(FastStringBuffer buff) {
	if(isSpecialized) {
	    buff.append(dynamicType.specializerSymbol());
	}
	staticType.appendGenericSignature(buff);
	if(isSpecialized) {
	    dynamicType.appendGenericSignature(buff);
	}
    }
    /**
     * Returns the size used in stack by value of this type
     */
    public /*@ pure @*/ int getSize() {
	return dynamicType.getSize();
    }

    //      /**
    //       * Check if a type is a numeric type
    //       * @return is it a numeric type ?
    //       */
    //      public boolean isNumeric() {
    //  	return false;
    //      }

    //      public boolean isCodedAsInt() {
    //  	return false;
    //      }

    //      /**
    //       * Check if a type is an integer type
    //       * @return is it a integer type ?
    //       */
    //      public boolean isOrdinal() {
    //  	return false;
    //      }

    //      /**
    //       * Check if a type is an integer type
    //       * @return is it a real number type ?
    //       */
    //      public boolean isFloatingPoint() {
    //  	return false;
    //      }

    //      /**
    //       * Check if a type is a class type
    //       * @return is it a subtype of ClassType ?
    //       */
    //      public boolean isPrimitive() {
    //  	return isNumeric() || (this == CStdType.Boolean);
    //      }

    //      /**
    //       * Check if a type is a class type
    //       * @return is it a subtype of ClassType ?
    //       */
    //      public boolean isReference() {
    //  	return false;
    //      }

    //      /**
    //       * Check if a type is a class type
    //       * @return is it a subtype of ClassType ?
    //       */
    //      public boolean isClassType() {
    //  	return false;
    //      }

    //      /**
    //       * @return is this type an array ?
    //       */
    //      public boolean isArrayType() {
    //  	return false;
    //      }

    //      /**
    //       * @return true if this type is valid
    //       */
    //      public boolean checked() {
    //  	return true;
    //      }

    //      /**
    //       * Returns the ID of this type
    //       */
    //      public int getTypeID() {
    //  	return type;
    //      }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    //      /**
    //       * Checks that this type is valid, throwing an exception if it is
    //       * not.  For class types this check resolves type names into their
    //       * fully qualified versions and sets the class parameter so that
    //       * getCClass returns the appropriate singleton.
    //       *
    //       * @exception	UnpositionedError	this unpositioned error
    //       *						should be positioned
    //       *						by the calling node of
    //       *						the AST, typically by
    //       *						calling the
    //       *						addPosition method of
    //       *						UnpositionedError 
    //       */
    //      public abstract void checkType(CContextType context) throws UnpositionedError;

    //      /**
    //       * Can this type be converted to the specified type by casting conversion (JLS 5.5) ?
    //       * @param	dest		the destination type
    //       * @return	true iff the casting conversion is valid
    //       */
    //      public abstract boolean isCastableTo(CType dest);

    //      /**
    //       * implicitlyCastTo
    //       * @param	dest	the expression type to cast to
    //       * @return	cast is valid ?
    //       */
    //      public boolean implicitlyCastTo(CType dest) {
    //  	return dest.equals(this);
    //      }

    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2].
     * @param	dest		the destination type
     * @return true iff assignment is valid */
    public boolean isAlwaysAssignableTo( CSpecializedType dest,
			 CClassType[] args) {
    	return dynamicType.isAlwaysAssignableTo( dest.dynamicType,
				 args);
    }

    public boolean isAlwaysAssignableToIgnoreUniverses( CSpecializedType dest,
			 CClassType[] args) {
    	if( dynamicType instanceof CClassType ) {
    		return ((CClassType)dynamicType).isAlwaysAssignableToIgnoreUniverses( dest.dynamicType, args);
    	} else {
    		return dynamicType.isAlwaysAssignableTo( dest.dynamicType, args);
    	}
    }

    //      /**
    //       * @return	true if this type corrspond to a checked exception
    //       */
    //      public boolean isCheckedException() {
    //  	return false;
    //      }

    //      /**
    //       * @return the object class of this type
    //       */
    //      public CClass getCClass() {
    //  	fail();
    //  	return null;
    //      }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    //      /**
    //       * Returns the opcode to load a local variable of this type.
    //       */
    //      public int getLoadOpcode() {
    //  	switch (type) {
    //  	case TID_CLASS:
    //  	case TID_ARRAY:
    //  	    return opc_aload;
    //  	case TID_DOUBLE:
    //  	    return opc_dload;
    //  	case TID_FLOAT:
    //  	    return opc_fload;
    //  	case TID_LONG:
    //  	    return opc_lload;
    //  	case TID_BYTE:
    //  	case TID_SHORT:
    //  	case TID_CHAR:
    //  	case TID_INT:
    //  	case TID_BOOLEAN:
    //  	    return opc_iload;
    //  	default:
    //  	    throw new RuntimeException("INTERNAL ERROR: " + type);
    //  	}
    //      }

    //      /**
    //       * Returns the opcode to store a local variable of this type.
    //       */
    //      public int getStoreOpcode() {
    //  	switch (type) {
    //  	case TID_CLASS:
    //  	case TID_ARRAY:
    //  	    return opc_astore;
    //  	case TID_DOUBLE:
    //  	    return opc_dstore;
    //  	case TID_FLOAT:
    //  	    return opc_fstore;
    //  	case TID_LONG:
    //  	    return opc_lstore;
    //  	case TID_BYTE:
    //  	case TID_SHORT:
    //  	case TID_CHAR:
    //  	case TID_INT:
    //  	case TID_BOOLEAN:
    //  	    return opc_istore;
    //  	default:
    //  	    throw new RuntimeException("INTERNAL ERROR: " + type);
    //  	}
    //      }

    //      /**
    //       * Returns the opcode to return a value of this type.
    //       */
    //      public int getReturnOpcode() {
    //  	switch (type) {
    //  	case TID_VOID:
    //  	    return opc_return;
    //  	case TID_CLASS:
    //  	case TID_ARRAY:
    //  	    return opc_areturn;
    //  	case TID_DOUBLE:
    //  	    return opc_dreturn;
    //  	case TID_FLOAT:
    //  	    return opc_freturn;
    //  	case TID_LONG:
    //  	    return opc_lreturn;
    //  	case TID_BYTE:
    //  	case TID_SHORT:
    //  	case TID_CHAR:
    //  	case TID_INT:
    //  	case TID_BOOLEAN:
    //  	    return opc_ireturn;
    //  	default:
    //  	    throw new RuntimeException("INTERNAL ERROR: " + type);
    //  	}
    //      }



    // -------------------------------------------------------------
    // STATIC UTILITY METHODS
    // -------------------------------------------------------------

    /**
     * Converts the given array of specialized types to an array of
     * specializers.
     *
     * @param	specTypes	the formal parameters
     * @return	the tuple of specializers */
    public static CType[] 
	tupleOfSpecializersFrom( CSpecializedType[] specTypes ) 
    {
	CType[] res = new CType[ specTypes.length ];
	for( int i=0; i<specTypes.length; i++ ) {
	    res[i] = specTypes[i].dynamicType;
	}
	return res;
    }

    public static CType[] 
	tupleOfSpecializersFrom( CSpecializedType[] specTypes, 
				 CClassType[] args ) {
	CType[] res = new CType[ specTypes.length ];
	for ( int i=0; i<specTypes.length; i++ ) {
	    CType dt = specTypes[i].dynamicType;
	    if(dt.isTypeVariable()) {
		res[i] = args[((CTypeVariable)dt).getIndex()];
	    } else { 
		res[i] = dt;
	    }
	}
	return res;
    }
    /**
     * Converts the given array of specialized types to an array of
     * the dynamic types of arguments sent to formals with these
     * specialized types.  This method differs from the method
     * tupleOfSpecializersFrom in that the dynamic type for a value
     * specializer (e.g. int@@3) is the same as the static type.
     *
     * @param	specTypes	the formal parameters
     * @return	the dynamic type tuple
     **/
    public static /*@ pure @*/ CType[] 
	dynamicTypeTupleFrom( CSpecializedType[] specTypes ) 
    {
	return dynamicTypeTupleFrom(null, specTypes);
    }

    /**
     * Converts the given array of specialized types to an array of
     * the dynamic types of arguments sent to formals with these
     * specialized types.  This method differs from the method
     * tupleOfSpecializersFrom in that the dynamic type for a value
     * specializer (e.g. int@@3) is the same as the static type.
     *
     * @param	recvType	the receiver type, or null
     * @param	specTypes	the formal parameters
     * @return	the dynamic type tuple
     **/
    public static /*@ pure @*/ CType[] 
	dynamicTypeTupleFrom( CType recvType,
			      CSpecializedType[] specTypes ) 
    {
	int offset = recvType == null ? 0 : 1;
	CType[] res = new CType[ specTypes.length + offset ];
	if( recvType != null ) {
	    res[0] = recvType;
	}
	for( int i=0; i<specTypes.length; i++, offset++ ) {
	    if ( specTypes[i].dynamicType.isValueType() ) {
		res[offset] = specTypes[i].staticType;
	    }
	    else {
		res[offset] = specTypes[i].dynamicType;
	    }
	}
	return res;
    }

    /**
     * Converts the given array of specialized types to an array of
     * the static type parts.
     *
     * @param	recvType	the receiver type, or null
     * @param	specTypes	the formal parameters
     * @return	the tuple of specializers */
    public static /*@ pure @*/ CType[] 
	staticTypeTupleFrom( CType recvType, 
			     CSpecializedType[] specTypes ) 
    {
	int offset = recvType == null ? 0 : 1;
	CType[] res = new CType[ specTypes.length + offset ];
	if( recvType != null ) {
	    res[0] = recvType;
	}
	for( int i=0; i<specTypes.length; i++, offset++ ) {
	    res[offset] = specTypes[i].staticType;
	}
	return res;
    }

    /**
     * Converts the given array of specialized types to an array of
     * unspecialized types.  That is, effectively forms the returned
     * array by removing the explicit specializers from each item in
     * the input array. */
    public static /*@ pure @*/ CSpecializedType[] 
	unspecializedTypeTupleFrom( CSpecializedType[] specTypes ) 
    {
	CSpecializedType[] res = new CSpecializedType[ specTypes.length ];
	for( int i=0; i<specTypes.length; i++ ) {
	    res[i] = specTypes[i].unspecializedType();
	}
	return res;
    }

    public static /*@ pure @*/ CSpecializedType[] 
	unspecializedTypeTupleFrom( CType[] simpleTypes )
    {
	CSpecializedType[] res = new CSpecializedType[ simpleTypes.length ];
	for( int i=0; i<simpleTypes.length; i++ ) {
	    res[i] = new CSpecializedType( simpleTypes[i] );
	}
	return res;

    }

    // ----------------------------------------------------------------------
    // PUBLIC CONSTANTS
    // ----------------------------------------------------------------------

    public static final CSpecializedType[] EMPTY = new CSpecializedType[0];

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private CType staticType;
    private CType dynamicType;
    private boolean isSpecialized = true;
}
