/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CStdType.java,v 1.16 2008/03/16 15:52:44 wdietl Exp $
 */

package org.multijava.mjc;

/**
 * This class is a singleton that provides variables for the various
 * built-in and java.lang types.  
 */
public class CStdType extends org.multijava.util.Utils implements Constants {

    // ----------------------------------------------------------------------
    // PRIMITIVE TYPES
    // ----------------------------------------------------------------------

    public static final CVoidType     Void;
    public static final CNullType     Null;

    public static final CNumericType  Byte;
    public static final CNumericType  Integer;
    public static final CNumericType  Long;
    public static final CNumericType  Double;
    public static final CNumericType  Float;
    public static final CNumericType  Short;
    public static final CNumericType  Char;

    public static final CBooleanType  Boolean;

    /**
     * An array of all the primitive types */
    static final CType[] PRIMITIVE_TYPES;
    
    //---------------------------------------------------------------------
    // REFERENCE TYPES
    //---------------------------------------------------------------------

    public static CClassType    Object;
    public static CClassType    WildcardBoundObject;
    public static CClassType    Class;
    public static CClassType    String;
    public static CClassType    ClassLoader;
    public static CClassType    Throwable;
    public static CClassType    Exception;
    public static CClassType    Error;
    public static CClassType    RuntimeException;
    public static CClassType    RMJRuntimeException;    /* RMJ */
    public static CClassType    Cloneable;
    public static CClassType    Serializable;

    public static CArrayType    CharArray;

    
    // ----------------------------------------------------------------------
    // INITIALIZERS
    // ----------------------------------------------------------------------

    static {
        PRIMITIVE_TYPES = new CType[ 8 ];
        PRIMITIVE_TYPES[0] = Byte = new CNumericType(TID_BYTE);
        PRIMITIVE_TYPES[1] = Integer = new CNumericType(TID_INT);
        PRIMITIVE_TYPES[2] = Long = new CNumericType(TID_LONG);
        PRIMITIVE_TYPES[3] = Double = new CNumericType(TID_DOUBLE);
        PRIMITIVE_TYPES[4] = Float = new CNumericType(TID_FLOAT);
        PRIMITIVE_TYPES[5] = Short = new CNumericType(TID_SHORT);
        PRIMITIVE_TYPES[6] = Char = new CNumericType(TID_CHAR);
        PRIMITIVE_TYPES[7] = Boolean = new CBooleanType();
        
        Void = new CVoidType();
        Null = new CNullType();

        // Class types must be lazily initialized by the call to
        // initSession, else CTopLevel is not properly initialized.
        // initClassTypes();
    }

    /** 
     * Initializes Java class types such as Object, Class, etc. */
    private static void initClassTypes() {
	if( CTopLevel.getCompiler() != null &&
		CTopLevel.getCompiler().universeChecks() ) {
        // Need Object first since others depend on it.
        Object = CTopLevel.getTypeRep(Constants.JAV_OBJECT, true);
        WildcardBoundObject  = CTopLevel.getTypeRep(Constants.JAV_OBJECT, CUniverseReadonly.getUniverse(), true);
        
        Class = CTopLevel.getTypeRep(Constants.JAV_CLASS, CUniverseReadonly.getUniverse(), true);
        String = CTopLevel.getTypeRep(Constants.JAV_STRING, CUniverseReadonly.getUniverse(), true);
        ClassLoader = CTopLevel.getTypeRep(Constants.JAV_CLASSLOADER, true);

	    // WMD: made all Throwable subtypes readonly
	    Throwable = CTopLevel.getTypeRep(Constants.JAV_THROWABLE,
					     CUniverseReadonly.getUniverse(),
					     true);
	    Exception = CTopLevel.getTypeRep(Constants.JAV_EXCEPTION,
					     CUniverseReadonly.getUniverse(),
					     true);
	    Error = CTopLevel.getTypeRep(Constants.JAV_ERROR,
					 CUniverseReadonly.getUniverse(),
					 true);
	    RuntimeException = 
		CTopLevel.getTypeRep(Constants.JAV_RUNTIME_EXCEPTION,
				     CUniverseReadonly.getUniverse(),
				     true);
	    RMJRuntimeException = 
		CTopLevel.getTypeRep(Constants.JAV_RMJ_RUNTIME_EXCEPTION,
				     CUniverseReadonly.getUniverse(),
				     true);
	} else {
        // Need Object first since others depend on it.
        Object = CTopLevel.getTypeRep(Constants.JAV_OBJECT, true);
        WildcardBoundObject  = Object;
        
        Class = CTopLevel.getTypeRep(Constants.JAV_CLASS, true);
        String = CTopLevel.getTypeRep(Constants.JAV_STRING, true);
        ClassLoader = CTopLevel.getTypeRep(Constants.JAV_CLASSLOADER, true);

		Throwable = CTopLevel.getTypeRep(Constants.JAV_THROWABLE, true);
	    Exception = CTopLevel.getTypeRep(Constants.JAV_EXCEPTION, true);
	    Error = CTopLevel.getTypeRep(Constants.JAV_ERROR, true);
	    RuntimeException = 
		CTopLevel.getTypeRep(Constants.JAV_RUNTIME_EXCEPTION, true);
	    RMJRuntimeException = 
		CTopLevel.getTypeRep(Constants.JAV_RMJ_RUNTIME_EXCEPTION, true);
	}

        Cloneable =  CTopLevel.getTypeRep(Constants.JAV_CLONEABLE, true);
        Serializable =  CTopLevel.getTypeRep(Constants.JAV_SERIALIZABLE, true);

        CharArray = new CArrayType(Char, 1, null);
    }

    /** 
     * Initializes a new session by clearing internal caches. This method
     * is useful for implementing multi-phase tools such as jmlrac. */
    public static void initSession() {
        initClassTypes();
    }
}
