/*
 * Copyright (C) 2000-2001 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;

import org.multijava.util.classfile.*;


/**
 * This class represents a value specializer for Strings.
 */
public class CStringValueType extends CValueType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructor
     * @param 	dispatchExpr	the AST of the specializer expression
     */
    public CStringValueType(JExpression dispatchExpr) {
	super(CStdType.String, dispatchExpr);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Transforms this type's value to a string for use in
     * bytecode attributes for this value specializer.
     */
    public String signatureStringValue() {
	    // this differs from dispatchValue.toString() in that
	    // the surrounding quotes must be escaped, so that we can
	    // uniquely find the trailing quote that signals the end of
	    // a bytecode attribute for a value specializer
	return escapeString(
	    "\"" + dispatchValue.getStringLiteral().stringValue() + "\"" );
    }

	/**
	 * Is the value of this specializer equal to the value of the
	 * given value specializer?
	 *
	 * @param	t	the type to compare against
	 * @return	true iff the values are equal
	 */
    protected boolean valueEquals(CValueType t) {
	if (!(t instanceof CStringValueType))
	    return false;
	CStringValueType ot = (CStringValueType) t;
	return dispatchValue.getStringLiteral().stringValue().equals(
	    ot.dispatchValue.getStringLiteral().stringValue());
    }

	/**
	 * Returns an integer representation of the specializer's value,
	 * for use in computing its hashCode.
	 */
    protected int intValue() {
	return dispatchValue.getStringLiteral().stringValue().hashCode();
    }
    

    	/*
	 * Is this value specializer of the appropriate type?
	 * requires dispatchValue != null	
	 */
    protected boolean isValid() {
	return dispatchValue.isStringLiteral();
    }
    

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

	/* Performs the instanceof / equality test for specializers of this
	 * type.  We assume that the actual argument has already been loaded.
	 *
	 * @param	code		the bytecode sequence to append to
	 * @param	nextInstruction	the instruction to branch to if the
	 *				dispatch test fails
	 */
    public void plantDispatchTest( CodeSequence code,
				   CodeLabel nextInstruction ) {
	    // load the specializer's value
	dispatchValue.getStringLiteral().genCode(code);

	CType[] parms = new CType[1];
	parms[0] = CStdType.Object;
	String sig = CType.genMethodSignature(CStdType.Boolean, parms);
	
	    // do the equality test
	    // 1 will be put on the stack if the test succeeds; 0 otherwise
	code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
						       "java/lang/String",
						       "equals",
						       sig));

	    // go to the next alternative if the test fails
	code.plantLabelRef( opc_ifeq, nextInstruction );
    }   
}
