/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.

 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSwitchBodyContext.java,v 1.7 2004/01/22 20:04:39 leavens Exp $
 */

package org.multijava.mjc;

import java.util.Hashtable;

import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents the context for the body of a switch
 * statement during checking passes (checkInterface,
 * checkInitializers, typecheck).  It includes methods for collecting
 * labeled cases and the default case and for determining where a
 * break should take the control flow.
 * @see CContextType 
 */
public class CSwitchBodyContext extends CFlowControlContext {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a context for a switch statement.
     * Clients should not call this but should instead call
     * <code>CFlowControlContextType.createSwitchBodyContext</code>.
     *
     * <pre><jml>
     * requires (* \caller instanceof CContextType *);
     * </jml></pre>
     * @param	parent		the parent context, it must be different
     *				than null
     * @param	switchType	the size of switch (byte, short, char...)
     *
     * @see CFlowControlContextType#createSwitchBodyContext(JSwitchStatement,CType)
     */
    CSwitchBodyContext(CFlowControlContextType parent, 
		       JSwitchStatement stmt, CType switchType) {
	super( parent, stmt.getTokenReference() );
	this.stmt = stmt;
	this.switchType = switchType;
    }


    // ----------------------------------------------------------------------
    // ACCESSORS (SWITCHES)
    // ----------------------------------------------------------------------

    /**
     * Adds a default label to this switch.
     * @exception UnpositionedError if a default label already exists
     */
    public void addDefault() throws UnpositionedError {
	if (defaultExist) {
	    throw new UnpositionedError(MjcMessages.SWITCH_DEFAULT_DOUBLE);
	}
	defaultExist = true;
    }

    /**
     * Adds a label to this switch and checks that it is a new one.
     * @param	lit	the literal value of this label
     * @exception UnpositionedError if the label is bad
     */
    public void addLabel(JOrdinalLiteral lit) throws UnpositionedError {
	if (!lit.isAssignableTo(switchType)) {
	    throw new UnpositionedError(MjcMessages.SWITCH_LABEL_OVERFLOW, 
					lit.getValue());
	}

	if (labels.put(lit.getValue(), "*") != null) {
	    throw new UnpositionedError(MjcMessages.SWITCH_LABEL_EXIST, 
					lit.getValue());
	}
    }

    /**
     * @return	true is a default label has already happen
     */
    public boolean defaultExists() {
	return defaultExist;
    }

    /**
     * remove a label from the list of labels
     */
    public JStatement getNearestBreakableStatement() {
	return stmt;
    }

    // -------------------------------------------------------------
    // CONTEXT FACTORIES AND ADOPTORS
    // -------------------------------------------------------------

    /**
     * Creates a context for checking the statements in a labelled
     * block statement group within this switch.  Each object returned
     * by this method should eventually be passed back to this object
     * as the <code>previousContext</code> argument of this method or 
     * <code>adoptSwitchGroupContext</code>.
     *
     * <pre><jml>
     * requires previousContext == null <==> (* the returned context is
     *					  	for the first group *);
     * </jml></pre>
     *
     *
     * @see #adoptSwitchGroupContext(CSwitchGroupContext)
     */
    public CSwitchGroupContext 
	createSwitchGroupContext(CSwitchGroupContext previousContext,
				 TokenReference where) {

	if (previousContext == null) {
	     return new CSwitchGroupContext( this, where );
	}
	
	if (previousContext.isReachable()) {
	    return new CSwitchGroupContext( this, previousContext, where );
	} else {
	    return new CSwitchGroupContext( this, where );
	}
    }


    /**
     * Records that <code>previousContext</code> is no longer being
     * used and that all block statements inside the switch body have
     * been checked.  Prepares this context for adoption by its parent
     * context.  
     *
     * <pre><jml>
     * requires previousContext != null && 
     *		(* previousContext is for last group in switch *);
     * </jml></pre>
     *
     */
    public void adoptSwitchGroupContext(CSwitchGroupContext previousContext) {
	assertTrue( previousContext != null );

	adopt( previousContext );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	boolean			defaultExist;
    private	Hashtable		labels = new Hashtable();
    private	JSwitchStatement	stmt;
    private	CType			switchType;
}
