/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CSwitchGroupContext.java,v 1.8 2004/01/16 23:21:09 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.TokenReference;


/**
 * This class represents the context for a switch group during
 * checking passes.  An instance of this is always the child context
 * of a CSwitchBodyContext.  This context maintains independent
 * definite assignment and reachability information from its parent
 * context but delegates variable declaration information to the
 * parent context.  This reflects the fact that the scope of a
 * variable declaration in the body of a switch statement is the
 * entire remainder of the body, <em>regardless of any intervening
 * case labels<em>.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.8 $ */

public class CSwitchGroupContext extends CFlowControlContext {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /*@ invariant cachedParent != null && 
      @           cachedParent instanceof CSwitchBodyContext;
      @*/

    /**
     * Construct a context for checking control flow through a switch
     * group where fall through from the previous group is not
     * possible.
     * Clients should not call this but should instead call
     * <code>CSwitchBodyContext.createSwitchGroupContext</code>.
     *
     * <pre><jml>
     * requires (* \caller instanceof CContextType *);
     * ensures cachedParent != null && 
     *         cachedParent instanceof CSwitchBodyContext;
     * </jml></pre>
     *
     * @see CSwitchBodyContext#createSwitchGroupContext(CSwitchGroupContext)
     */
    CSwitchGroupContext(CSwitchBodyContext parent, TokenReference where) {
	super(parent,0,where);
    }
    
    /**
     * Construct a context for checking control flow through a switch
     * group where fall through from the previous group is
     * possible.
     * Clients should not call this but should instead call
     * <code>CSwitchBodyContext.createSwitchGroupContext</code>.
     *
     * <pre><jml>
     * requires (* \caller instanceof CContextType *);
     * requires previousContext != null && previousContext.isReachable();
     * ensures cachedParent != null && 
     *         cachedParent instanceof CSwitchBodyContext;
     * </jml></pre>
     *
     * @see CSwitchBodyContext#createSwitchGroupContext(CSwitchGroupContext)
     */
    CSwitchGroupContext(CSwitchBodyContext parent,
			CSwitchGroupContext previousContext,
			TokenReference where) {
	super(parent,0,where);
	merge(previousContext);
    }
    
    // ---------------------------------------------------------------------
    // ACCESSORS (VARIABLE)
    // ---------------------------------------------------------------------
    // All local variable operations are delegated to the parent
    // switch body context.

    /**
     * Adds a local variable to this block
     * @param	var		the name of the variable
     * @exception UnpositionedError	if the variable is a redeclaration
     */
    public void addVariable(JLocalVariable var) throws UnpositionedError {
	cachedParent.addVariable(var);
    }

    /**
     * Records that the Java keyword <CODE>this</CODE> is used in this
     * block.
     */
    public void addThisVariable() {
	cachedParent.addThisVariable();
    }

    /**
     * Records that a variable slot must be reserved for a synthetic
     * <code>this</code> parameter.  Recall that <code>this</code> is
     * passed as a parameter for external methods.  */
    public void addSyntheticThisParameter() {
	cachedParent.addSyntheticThisParameter();
    }

    /**
     * Returns the number of stack positions required to store the
     * local variables encountered thus far in this context.  */
    public int localsPosition() {
	return cachedParent.localsPosition();
    }

    public int numberOfLocalVars() {
	return cachedParent.numberOfLocalVars();
    }

    /**
     * Returns true if the given local variable is declared exactly in
     * this context, i.e., it is not declared in an outer context. */
    public boolean declares(JLocalVariable var) {
	return cachedParent.declares(var);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (TYPE DEFINITION)
    // ----------------------------------------------------------------------
    // All local type declaration methods delegate to the surrounding
    // switch body context.

    /**
     * Adds to this context a class declared via a type declaration
     * statement.
     * @param	clazz		the clazz to add
     * @exception UnpositionedError	if duplicate class exists in this
     * lexical context */
    public void addLocalClass(CClass clazz) throws UnpositionedError {
	cachedParent.addLocalClass(clazz);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (LOOKUP)
    // ----------------------------------------------------------------------
    // All lookup methods delegate to the surrounding switch body
    // context.

    /**
     * Searches for a class with the given simple name according the
     * procedure in JLS2 6.5.5.  This method checks the first bullet
     * point of that section: "If the simple type name occurs within
     * the scope of a visible local class declaration with that name,
     * then the simple type name denotes that local class type."  If
     * no match is found, then this method calls to the surrounding
     * context to check the remaining bullet points.
     * 
     * @param	name	the class name, without qualifiers
     * @return	the class if found, null otherwise
     * @exception UnpositionedError if search fails */
    public CClass lookupClass(String name) throws UnpositionedError {
	return cachedParent.lookupClass(name);
    }

    /**
     * Returns the variable referred to by the given name in this
     * context, recursing to surrounding contexts as appropriate.  If
     * no variable of the given name is found then <code>null</code>
     * is returned.
     *
     * @param	ident		the name of the variable */
    public JLocalVariable lookupLocalVariable(String ident) {
	return cachedParent.lookupLocalVariable(ident);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (LABELS)
    // ----------------------------------------------------------------------
    
    /**
     * Adds a default label to this switch.
     * @exception UnpositionedError if a default label already exists
     */
    public void addDefault() throws UnpositionedError {
	((CSwitchBodyContext) cachedParent).addDefault();
    }

    /**
     * Add a label to this switch and check that it is a new one.
     * @param	lit	the literal value of this label
     * @exception UnpositionedError if the label is bad
     */
    public void addLabel(JOrdinalLiteral lit) throws UnpositionedError {
	((CSwitchBodyContext) cachedParent).addLabel(lit);
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

}
