/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CTopLevel.java,v 1.46 2008/03/16 15:52:45 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.ClassInfo;
import org.multijava.util.classfile.ClassPath;
import org.multijava.util.classfile.ClassFileReadException;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.Map;
import java.io.File;
import java.util.HashSet;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.ListIterator;

/**
 * This class is a facade for an instance of TypeLoader (or a
 * subclass).  It acts like a "class loader"; it keeps track of which
 * classes and external generic functions referenced in a compilation
 * session are being generated from source and which are being read
 * from disk as class files.  It also handles the loading of such
 * classes and external generic functions from disk.  in general all
 * cached operations on types and type signatures are directed through
 * this facade. */
public final class CTopLevel extends org.multijava.util.Utils 
    implements Constants 
{

    // -------------------------------------------------------------
    // HANDLING OF LIGHTWEIGHT TYPE REPRESENTATIONS (JUST NAMES WITHOUT SIGS)
    // -------------------------------------------------------------

    /**
     * Returns the lightweight type representation of the class with
     * the given name, generating a new one if necessary.
     * @param	name	the name of the class
     * @param	isFullyQualified  the name is known to be fully qualified
     * @return an object representing the type of the named class
     * (singleton within this compilation session if isFullyQualified)
     */ 
    public static CClassType getTypeRep(String name, CClassType[][] arguments,
					boolean isFullyQualified) 
    {
	// orig: return typeLoader.getTypeRep(name,null,arguments,isFullyQualified);
	// WMD: call the other method
	return getTypeRep(name, null, arguments, isFullyQualified);
    }

    /** return the lightweight type representation of the class with the given
     * name, generating a new one if necessary.
     */
    public static CClassType getTypeRep(String name, boolean isFullyQualified){
	return getTypeRep(name, null, CClassType.EMPTY_ARG, isFullyQualified);
    }


    /** getTypeRep with a universe parameter and no arguments.
     * By WMD.
     */
    public static CClassType getTypeRep(String name, CUniverse univ,
                                        boolean isFullyQualified){
        return getTypeRep(name, univ, CClassType.EMPTY_ARG, isFullyQualified);
    }


    /** getTypeRep with a universe parameter and arguments.
     *
     * We get a cached version of the class and then clone it
     * and set the correct universe modifier.
     * The clone is "shallow" and most of the data can be shared
     * between different universes.
     *
     * By WMD.
     */
    public static CClassType getTypeRep(String name, CUniverse univ,
                                        CClassType[][] arguments,
                                        boolean isFullyQualified) {

        CClassType cache;
	cache = typeLoader.getTypeRep(name, univ, arguments, isFullyQualified);

	if( CTopLevel.getCompiler() != null &&
	    CTopLevel.getCompiler().universeChecks() ) {
	    // WMD only do this if we parse or check the universes

		// RZU - Use this as a universe modifier since no given 
		//       universe modifier means that it is used in
		//       JThisExpression where we need it to be "this"
	    if( univ == null )
		univ = CUniverseThis.getUniverse();

	    if( ! cache.getCUniverse().equals( univ ) ) { // WMD Check
		try {
		    cache = (CClassType) cache.clone();
		} catch(CloneNotSupportedException e) {
		    fail("Clone of cached CClassType failed!");
		}

		cache.setCUniverse( univ );
	    }
	} else {
	    // set the universe to this!
	    // this is important if the parsing is enabled, but the
	    // checking is disabled, to prevent mixing with old
	    // universe information that might still be in the cache!
	    cache.setCUniverse( CUniverseThis.getUniverse() );
	}

	return cache;
    }





    /**
     * Adds the given type representation, which must represent a
     * fully qualified type name, to the cache.
     *
     * <pre><jml>
     * requires qName != null && typeRep != null &&
     *          (* all separators in qName are '/' not '.' *);
     * </jml></pre>
     *
     */
    public static void addTypeRep(String qName, CClassType typeRep) {
	typeLoader.addTypeRep(qName, typeRep);
    }

    // -------------------------------------------------------------
    // HANDLING OF TYPE SIGNATURES
    // -------------------------------------------------------------

    /**
     * Loads type definition; uses the FileFinder supplied in initSession
     * to determine which file is loaded for the given name.
     * Returns CClass.CLS_UNDEFINED or null if the type could not be loaded.
     * 
     * <pre><jml>
     *  requires qName != null && 
     *		(* any separators in qName are slashes ('/'), not dots ('.') *);
     *  modifies typeLoader;
     * </jml></pre>
     * @param qName	the fully qualified name of the type to load.
     */
    public static CClass loadClass(String qName) {
	CClass typ = typeLoader.loadType(qName);
	return typ;
    }

    /**
     * Determines whether the given Class, using its fully qualified name, 
     * has been declared in more one file. 
     * This method is needed by JML in order to handle classes that 
     * have been refined, e.g., when its specifications have intentionally 
     * been declared in a separate file.
     *
     * Returns true if the class has been declared in more than 
     * one source file (unless refined), and false otherwise.
     *
     * <pre><jml>
     *  requires clazz != null; 
     * </jml></pre>
     *
     */
    public static boolean isDeclaredInDifferentSourceFiles( boolean isUnique, 
							    CClass clazz ) {
	return typeLoader.isDeclaredInDifferentSourceFiles( isUnique, 
							    clazz );
    }

    /**
     * Reloads the specified class if this class has been defined
     * during a previous compilation session.  Note that this method
     * can return either null or CLS_UNDEFINED or a valid CClass
     * value.
     */
    public static CClass reloadClass(CClass clazz) {
	return typeLoader.reloadType(clazz);
    }

    /** 
     * Returns true if the class with the given fully-qualified name
     * has already been read and loaded.
     *
     * <pre><jml>
     *  requires qName != null && 
     *		(* any separators in qName are slashes ('/'), not dots ('.') *);
     * </jml></pre>
     * @param qName	the fully qualified name of the type to load.
     */
    public static boolean isClassLoaded(String qName) {
	return typeLoader.isTypeLoaded(qName);
    }

    /**
     * Adds the given class to the table of available classes. 
     * Returns true if the class was not declared in more than one 
     * source file and false otherwise.
     *
     * <pre><jml>
     requires clazz != null;
     modifies typeLoader;
     * </jml></pre>
     */
    public static boolean addClass( CClass clazz ) {
	return typeLoader.addToTypeCache(clazz);
    }

    /**
     * Checks whether a class with the given name is available in the class
     * path.  Note that this is true if the class has already been successfully
     * loaded or if there is a file that can be expected to hold its
     * definition.  The file may actually not hold a valid definition, so
     * classExists is not a guarantee that the class with the given name 
     * actually does exist.
     * @return true iff the class is in the class path
     * @param qName the fully qualified name of the type with slash separators
     *			not dots
     */
    public static boolean classExists( String qName ) {
	return typeLoader.classExists(qName);
    }

    // -------------------------------------------------------------
    // HANDLING OF GENERIC FUNCTION SIGNATURES
    // -------------------------------------------------------------

    /**
     * Registers that the external generic function with the given
     * fully qualified name has been parsed and, barring an error
     * during the preprocessDependencies pass, will be available to other
     * compilation units.
     *
     * <pre><jml>
     * requires qName != null && qName.length() > 0
     *	    && (* any separators in qName are slashes ('/'), not dots ('.') *);
     * </jml></pre>
     *
     * @see MJTopLevelMethodDeclaration#registerPendingGFCollection(String)
     */
    public static void registerPendingGFCollection(String qName) {
	pendingGFs.add(qName);
    }

    /**
     * Adds the given generic function collection singleton to the map.
     *
     * @return	true if this collection is OK, i.e., is not a redeclaration.
     */
    public static boolean addGFCollection( CGenericFunctionCollection coll ) {
	CGenericFunctionCollection previous = (CGenericFunctionCollection) 
	    gfCollSingletonMap.put( coll.qualifiedName(), coll );

	return previous == null || 
	    previous == CGenericFunctionCollection.GF_UNDEFINED || 
	    !(previous instanceof CSourceGFCollection) ||
	    ((CClass) previous).isNestedType();

    }

    /**
     * Finds the generic function collection singleton for the given
     * fully qualified name.  If no singleton exists then tries to
     * read the collection from disk (either from binary or source).
     * If no disk file is found then returns
     * CGenericFunctionCollection.GF_UNDEFINED.
     *
     * @param	qualifiedIdent	the fully qualified name of the collection
     * @return the singleton representing the collection, or
     * CGenericFunctionCollection.GF_UNDEFINED if one cannot be found
     * */
    public static CGenericFunctionCollection 
	lookupGFCollection( String qualifiedIdent )
    {
	String anchorName = qualifiedIdent + MJ_ANCHOR;
	CGenericFunctionCollection coll = 
	    (CGenericFunctionCollection) gfCollSingletonMap.get( anchorName );

	if( coll == null ) {
	    // For collections that may be generated during this
	    // compilation session, singletons were creating during
	    // parsing and the get above would not have returned null.
	    // So we know that we aren't generating the requested
	    // collection during this compilation session.  Also since
	    // the get returned null we know we haven't read the
	    // requested collection from disk yet.  So let's try to do
	    // that...

	    ClassPath.ClassDescription desc = typeLoader.find(anchorName);
	    if (desc == null)  {
		// No appropriate file exists to define the anchor
		// class.  That may be OK because lookupGFCollection
		// is often called as part of a fishing expedition for
		// implicitly imported generic functions.  We put a
		// marker object in the singleton map at this
		// anchorName to avoid repeating work the next time
		// the compiler fishes for qualifiedIdent.
		coll = CGenericFunctionCollection.GF_UNDEFINED;
	    } else if (desc.isClass()) {
		ClassInfo info;
		try {
		    info = ClassPath.getClassInfo(anchorName, true);
		} catch (ClassFileReadException e) {
		    compiler.reportTrouble(e);
		    info = null;
		}
		
		if (info != null && info.isAnchor()) {
		    // Initialization of the instance includes a
		    // callback that adds the instance to the
		    // gfCollSingletonMap.
		    new CBinaryGFCollection(FROMCLASSES, info);
		    coll = (CGenericFunctionCollection) 
			gfCollSingletonMap.get( anchorName );
		} else {
		    coll = CGenericFunctionCollection.GF_UNDEFINED;
		}
	    } else {
		// System.out.println("DESC SOURCE " + name);
		File f = ((ClassPath.FileClassDescription) desc).file();

		// This triggers compiling of the source code for
		// the generic function up to the current pass of
		// the main compiler invocation.  Remaining passes
		// handle this new anchor class along with all the
		// other classes being compiled.  At the end of
		// the parsing pass the newly loaded anchor class
		// is added to gfCollSingletonMap by a call to
		// CTopLevel.addGFCollection (declared
		// above).  Note that we don't use the
		// name-checking version of catchUp because we
		// don't really care if the file happens to not
		// declare a generic function with the correct
		// name--this might just be a fishing expedition.
		compiler.catchUpGF(f,qualifiedIdent); 
		coll = (CGenericFunctionCollection) 
		    gfCollSingletonMap.get( anchorName );
	    }

	    // a bit of defensive programming to make sure we don't
	    // forget to set coll to non-null value in some branch
	    // above
	    if (coll == null) {
		coll = CGenericFunctionCollection.GF_UNDEFINED;
	    }

	    gfCollSingletonMap.put( anchorName, coll );
	}

	return coll;
    }

    public static CClassType lookupTypeRep(String qName) {
	return typeLoader.lookupTypeRep(qName);
    }
    /** 
     * Returns true if the generic function with the given fully-qualified name
     * has already been read and loaded.
     *
     * <pre><jml>
     * 	requires qName != null && 
     *           (* any separators in qName are slashes ('/'), not dots ('.') *);
     * </jml></pre>
     */
    public static boolean isGFLoaded(String qName) {
	if (pendingGFs.contains(qName)) {
	    // GF is in the task queue to be compiled during this session
	    return true;
	}
	
	String anchorName = qName + MJ_ANCHOR;
	CGenericFunctionCollection coll = 
	    (CGenericFunctionCollection) gfCollSingletonMap.get( anchorName );

	return coll != null && coll != CGenericFunctionCollection.GF_UNDEFINED;
    }


    /**
     * Checks whether the source code contain in the given file
     * declares a type or an external generic function of the given
     * name.
     * @param qName the fully qualified name of the class or generic function
     *		    expected in <code>f</code>
     * @param f the source code file to check
     * @return {@link #IS_TYPE} if the code in <code>f</code> declares a type
     *	       named <code>qName</code>,
     *	       {@link #IS_GEN_FUNC} if the code in <code>f</code> declares 
     *	       an external generic function named <code>qName</code>, and
     *	       {@link #IS_NEITHER} otherwise.
     */
    public static int whichIsThis(String qName, File f) {
	// try to catch-up the file, if it has already been processed
	// then the task will be skipped my main, but we leave that
	// optimization to Main
	compiler.catchUp(f);

	// the catching up process should have registered either a
	// compilation unit or a class
	if (isClassLoaded(qName)) {
	    return IS_TYPE;
	} else if (isGFLoaded(qName)) {
	    return IS_GEN_FUNC;
	} else {
	    return IS_NEITHER;
	}
    }

    /**
     * A constant return value for {@link #whichIsThis(String,java.io.File)}.
     */
    public static final int IS_TYPE = 0;

    /**
     * A constant return value for {@link #whichIsThis(String,java.io.File)}.
     */
    public static final int IS_GEN_FUNC = 1;

    /**
     * A constant return value for {@link #whichIsThis(String,java.io.File)}.
     */
    public static final int IS_NEITHER = -1;
    
    // -------------------------------------------------------------
    // UTILITY METHODS FOR SESSION CONTROL
    // -------------------------------------------------------------

    /**
     * Removes all non-trusted, cached information from this
     * compilation sessions so that new declarations can be processed.
     * This is package protected so that tools inheriting from
     * mjc.Main must use the {@link Main#initSession(TypeLoader)}
     * method for session initialization.
     *
     * <pre><jml>
     requires loader != null;
     modifies CTopLevel.compiler;
     modifies typeLoader;
     modifies gfCollSingletonMap, pendingGFs;
     ensures CTopLevel.compiler == compiler &&
     CTopLevel.typeLoader == loader;
     * </jml></pre>
     *
     */
    static void initSession( Main compiler, TypeLoader loader ) {
	CTopLevel.compiler = compiler;
	CTopLevel.typeLoader = loader;
	if (loader == null) {
	    throw new RuntimeException("Implementation error: CTopLevel.initSession called with a null loader argument");
	}

	// clear any cached files already found, since we may have a
	// new TypeLoader
	ClassPath.initSession();

        // clear all loaded classes such as .java, .spec, .classes etc.
        typeLoader.initSession();

	// clear loaded generic functions
	gfCollSingletonMap.clear();
	pendingGFs.clear();
    }

    /**
     * Returns the top level compiler.  Result may be
     * <CODE>null</CODE> if called after <CODE>initSession()</CODE> is
     * invoked with a <CODE>null</CODE> first argument.  */
    //@ ensures \result == compiler;
    //@ ensures \result != null || \result == null;
    public static Main getCompiler() {
	return compiler;
    }

    public static Main getFromClassesCompiler() {
	return FROMCLASSES;
    }

    /**
     * Returns the FileFinder instance that is used by the current
     * session when searching for classes and generic functions.  */
    //@ ensures \result != null;
    public static FileFinder fileFinder() {
	return typeLoader.fileFinder();
    }

    public static File findSourceFile(String name) {
	ClassPath.ClassDescription c = typeLoader.findSourceFile(name);
	return c == null ? null : ((ClassPath.FileClassDescription)c).file();
    }

    /**
     * Called by Main when the given file, from which the given
     * compilation unit AST was derived, did not contain an expected
     * result; this method removes all cached info. for the file.
     */
    public static void forgetEverythingAbout(File f,JCompilationUnitType cunit) {
	typeLoader.forgetEverythingAbout(f, cunit);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    //@ spec_public
    private static Main	compiler;

    private static final Main FROMCLASSES = 
	new Main( new CModifier() ) {
	    public boolean run(String[] args) {
		return true;
	    }
	    public void reportTrouble(Exception a) {
	    }
	    public boolean parseComments() {
		return true;
	    }
	    public boolean verboseMode() {
		return false;
	    }
	    public void catchUp(File file) {
	    }
	    public void catchUp(JTypeDeclarationType decl) {
	    }
	};


    /**
     * Maps from qualified names to
     * <code>CGenericFunctionCollection</code> singletons.  */
    //@ spec_public // actually only need spec_package visibility
    private static java.util.Map gfCollSingletonMap = new java.util.HashMap();

    //@ spec_public // actually only need spec_package visibility
    private static java.util.HashSet pendingGFs = new java.util.HashSet();

    //@ spec_public
    private static TypeLoader typeLoader = TypeLoader.getSingleton();
}
