/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CType.java,v 1.52 2008/05/14 18:16:19 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.*;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.FastStringBuffer;

/**
 * <P>This class is the root for type hierarchy.  It contains methods for
 * generating type-signatures in bytecode, finding appropriate
 * op-codes for operations on various types, and answering questions
 * about the appropriateness of casts and assignments.</P>
 */
public abstract class CType extends org.multijava.util.Utils 
    implements Constants, Comparable, CTypeSignatureAppender
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a type signature
     */
    protected CType(int type) {
	this.type = type;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Compares two objects for equality.
     *
     * @param other	the object to compare with
     * @return true iff <code>other</code> is a <code>CType</code> object 
     *		representing the same Java type as this represents
     */
    public boolean equals(/*@ nullable @*/ Object other) {
	// non-reference types are singletons so this is the right check,
	// need to override for CClassType to consider CClass singletons
	return this == other;
    }

    /**
     * Compares two objects for equality.
     *
     * @param other	the object to compare with
     * @param enableUniv whether to also check the universe part
     *          this is needed sometimes to allow relaxed checks
     * @param allowUniversesCovariance Allow limited covariance from GUT iff true. 
     * @return true iff <code>other</code> is a <code>CType</code> object 
     *		representing the same Java type as this represents
     */
    public boolean equals(/*@ nullable @*/ Object other, boolean enableUniv,
    		boolean allowUniversesCovariance) {
	return equals(other);
    }

    public boolean equals(/*@ nullable @*/ Object other, boolean enableUniv ) {
	return equals(other, enableUniv, false);
    }

    public boolean equals(/*@ nullable @*/ CType other,
			  /*@ non_null @*/ CClassType[] substitution) {
	return equals(other);
    }

    public boolean equals(/*@ nullable @*/ CType other,
			  /*@ non_null @*/ CClassType[] substitution,
			  boolean enableUniv) {
	return equals(other, enableUniv, false);
    }
    
    /**
     * Compares this to another type alphabetically by
     * identifier.</em> */
    public int compareTo(/*@ non_null @*/ Object o) throws ClassCastException {
	return toString().compareTo(o.toString());
    }

    /**
     * Transforms this type to a string
     */
    public abstract String toString();

    /** A more verbose version of toString only used for error reporting.
     * Adds the universe modifier to the output.
     * Made this a separate function, to make sure that we do not
     * mess up any bytecode generation routines that use toString()
     *
     * By WMD.
     */
    public String toVerboseString() {
	return toString();
    }

    /**
     * Indicates whether this type is a nested type, i.e., a member
     * type.  */
    public /*@ pure @*/ boolean isNestedType() {
	return false;
    }

    /**
     * Transforms this type to a string
     * @return	the class file form of this type
     */
    public String getSignature() {
	FastStringBuffer buff = StringBuffers.getStringBuffer();

	appendSignature(buff);
	String s = buff.toString();
	StringBuffers.release(buff);

	return s;
    }

    public String getGenericSignature() {
	FastStringBuffer buff = StringBuffers.getStringBuffer();
	appendGenericSignature(buff);
	String s = buff.toString();
	StringBuffers.release(buff);
	
	return s;
    }

    public void appendGenericSignature(FastStringBuffer buff) {
	appendSignature(buff);
    }
    /**
     * Returns the size used in stack by value of this type
     */
    public /*@ pure @*/ abstract int getSize();

    /**
     * Check if a type is a numeric type
     * @return is it a numeric type ?
     */
    public /*@ pure @*/ boolean isNumeric() {
	return false;
    }

    /**
     * Check if a type is void type
     * @return is it void type ?
     */
    public /*@ pure @*/ boolean isVoid() {
	return false;
    }

    public /*@ pure @*/ boolean isCodedAsInt() {
	return false;
    }

    /**
     * Check if a type is an integer type
     * @return is it a integer type ?
     */
    public /*@ pure @*/ boolean isOrdinal() {
	return false;
    }

    /**
     * Check if a type is a real number type
     * @return is it a real number type ?
     */
    public /*@ pure @*/ boolean isFloatingPoint() {
	return false;
    }

    /**
     * Check if a type is the boolean type
     * @return is it the boolean type ?
     */
    public /*@ pure @*/ boolean isBoolean() {
	return false;
    }

    /**
     * Check if a type is a primitive scalar type
     * @return is it a numeric or boolean type ?
     */
    public /*@ pure @*/ boolean isPrimitive() {
	return isNumeric() || isBoolean();
    }

    /**
     * Check if a type is a pointer type (which is the same as a class type)
     * @return is it a subtype of ClassType ?
     */
    public /*@ pure @*/ boolean isReference() {
	return false;
    }

    /**
     * Check if a type is a reference type (a class, interface, or array)
     * @return is it a subtype of ClassType ?
     */
    public /*@ pure @*/ boolean isClassType() {
	return false;
    }
    
    /** 
     * Indicates whether this type is a Generic class type
     */
    //@ assignable \nothing;
    public boolean isGenericType() {
	return false;
    }

    /** 
     * Returns whether this is a raw type (JLS3 4.8).
     *
     * @return  Whether this is a raw type.
     */
    public boolean isRawType() {
        return false;
    }

    /** 
     * Returns whether this is a reifiable type (JLS3 4.7).
     *
     * @return  Whether this is a reifiable type.
     */
    public boolean isReifiableType() {
        return true;
    }

    /**
     * Returns whether this type differs from its type erasure (JLS3 4.6).
     *
     * @return  Whether this type differs from its type erasure.
     */
    public boolean changesByErasure() {
        return false;
    }

    public CType getErasure(CContextType context) throws UnpositionedError {
    	return getErasure().checkType(context);
    }

	public CType getErasure() {
        return this;
    }

    /**
     * Returns the result of applying capture conversion (JLS3 5.1.10) to this
     * type.
     * <p>
     * Note that capture conversion only changes a parametrized type with at
     * least one wildcard in its type arguments.
     * </p>
     *
     * @param context  The type checking context.
     * @return         The result of applying capture conversion to this type.
     *                 Is never <code>null</code>.
     *
     * @throws UnpositionedError if some error occurs during type checking of
     *                           the returned type.
     */
    public /*@ non_null @*/ CType getCapture(CContextType context) throws UnpositionedError {
        return this;
    }

    public CClassType[][] getAllArguments() {
	throw new RuntimeException("no Arguments in this type");
    }
    
    public CClassType[] getArguments() {
	throw new RuntimeException("no Arguments in this type");
    }

    public String getIdent() {
	return "";
    }
     
    /**
     * @return the object class of this type
     */
    public CClass getCClass() {
	fail();
	return null;
    }

    /**
     * Check if a type is a type variable
     */
    public boolean isTypeVariable() {
	return isTypeVariable;
    }

    /**
     * Returns whether this type is a type variable introduced by a generic
     * class declaration (JLS3 8.1.2) or a generic interface declaration
     * (JLS3 9.1.2) as opposed to a method type variable.
     *
     * @return  Whether this type is a class type variable.
     */
    public /*@ pure @*/ boolean isClassTypeVariable() {
        return false;
    }
    
    //@ assignable \nothing;
    public boolean isMethodTypeVariable() {
	return false;
    }
    
    public void setMethodTypeVariable(boolean isMTV) {
	isMethodTypeVariable = isMTV;
    }

    public void setTypeVariable(boolean isTV) {
	isTypeVariable = isTV;
    }

    /**
     * Returns whether this is a wildcard type.
     *
     * @return  Whether this is a wildcard type.
     */
    public /*@ pure @*/ boolean isWildcard() {
        return false;
    }

    /**
     * Returns whether this is a wildcard capture type as resulting from
     * capture conversion (JLS3 5.1.10).
     *
     * @return  Whether this is a wildcard capture type.
     */
    public /*@ pure @*/ boolean isCapture() {
        return false;
    }

    /**
     * Check if a type is an interface
     * @return is it a class type representing an interface ?
     */
    public boolean isInterface() {
	return false;
    }

    /**
     * @return is this type an array ?
     */
    public /*@ pure @*/ boolean isArrayType() {
	return false;
    }


    /**
     * @return is this type representing a value specializer ?
     */
    public /*@ pure @*/ boolean isValueType() {
	return false;
    }

    /**
     * @return true if this type is valid
     */
    public /*@ pure @*/ boolean checked() {
	return true;
    }

    /**
     * Returns the ID of this type
     */
    public /*@ pure @*/ int getTypeID() {
	return type;
    }

    /**
     * Indicates whether this type is concrete.  All types are
     * concrete except abstract class types and interface types.
     *
     * @return	whether this type is concrete
     *
     * <pre><jml>
     * assignable objectState;
     * </jml></pre>
     */
    public boolean isConcrete() {
	return true;
    }

    /**
     * Indicates whether this type is declared local to the given
     * context.  Can only be true of reference types.
     *
     * @param	context	the context in which to check
     * @return whether this type is declared local to the given context
     *
     * <pre><jml>
     * assignable objectState;
     * </jml></pre>
     */
    public boolean isLocalTo( CContextType context ) {
	return false;
    }

	/**
	 * Returns the specializer symbol to use when this type is the
	 * specializer.
	 */
    public /*@ pure @*/ String specializerSymbol() {
	return "@";
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks that this type is valid, throwing an exception if it is
     * not.  For class types this check resolves type names into their
     * fully qualified versions and sets the class parameter so that
     * getCClass returns the appropriate singleton.
     *
     * @exception	UnpositionedError	this unpositioned error
     *						should be positioned
     *						by the calling node of
     *						the AST, typically by
     *						calling the
     *						addPosition method of
     *						UnpositionedError 
     */
    public abstract CType checkType(CContextType context) throws UnpositionedError;

    /**
     * Compute the value of a specializer expression used to dispatch on
     * a compile-time constant.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     */
    public abstract void resolveValueType(CExpressionContextType context)
	throws PositionedError;

    /**
     * Can this type be converted to the specified type by casting conversion (JLS 5.5) ?
     * @param	dest		the destination type
     * @return	true iff the casting conversion is valid
     *
     * <pre><jml>
     * assignable objectState;
     * </jml></pre>
     */
    public abstract boolean isCastableTo(CType dest);

    /**
     * implicitlyCastTo
     * @param	dest	the expression type to cast to
     * @return	cast is valid ?
     */
    public boolean implicitlyCastTo(/*@ non_null */ CType dest) {
	return dest.equals(this);
    }

    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2].
     * @param	dest		the destination type
     * @return true iff assignment is valid
     *
     * <pre><jml>
     * //assignable objectState;
     * </jml></pre>
     */
    public abstract boolean isAlwaysAssignableTo(CType dest);
    
    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2].
     * 
     * The only difference to
     * {@link #isAlwaysAssignableTo(CType) isAlwaysAssignableTo} is
     * that this version ignores the universe modifiers.
     * 
     * @param dest The destination type 
     * @return true iff assignment is valid
     * 
     * By RZU
     */
    public boolean isAlwaysAssignableToIgnoreUniverses(CType dest) {
    	return isAlwaysAssignableTo(dest);
    }
    
    // assignable objectState;
    public boolean isAlwaysAssignableTo(CType dest, boolean inst) {
	return isAlwaysAssignableTo(dest);
    }

    // assignable objectState;
    public boolean isAlwaysAssignableTo(CType dest,
					CClassType[] substitution){
	return isAlwaysAssignableTo(dest);
    }
    
    public boolean isAssignableTo(CType dest) {
	return false;
    }

    /**
     * Returns whether unchecked conversion (JLS3 5.1.9) is required in order
     * to convert from this type to the given type <code>dest</code>.
     * 
     * @param dest  The target type.
     * @return      Whether unchecked conversion is required to convert this
     *              type to the type <code>dest</code>.
     *
     * @see #isAlwaysAssignableTo(CType)
     */
    public boolean needsUncheckedConversion(/*@ non_null @*/ CType dest) {
        return false;
    }

    /**
     * Returns whether unchecked conversion (JLS3 5.1.9) is required in order
     * to convert from this type to the given type <code>dest</code>, eventually
     * using the given generic <code>substitution</code>.
     * 
     * @param dest          The target type.
     * @param substitution  The generic substitution for class type variables.
     * @return              Whether unchecked conversion is required to convert
     *                      this type to the type <code>dest</code>.
     *
     * @see #isAlwaysAssignableTo(CType)
     */
    public boolean needsUncheckedConversion(
            /*@ non_null @*/ CType dest,
            /*@ non_null @*/ CClassType[] substitution) {
        return needsUncheckedConversion(dest);
    }

    /**
     * @return	true if this type corrsponds to a checked exception
     */
    public boolean isCheckedException() {
	return false;
    }

    
    /**
     * @return the value expression of this type (for value types only)
     */
    public JExpression getValue() {
	fail();
	return null;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Returns the opcode to load a local variable of this type.
     */
    public /*@ pure @*/ int getLoadOpcode() {
	switch (type) {
	case TID_CLASS:
	case TID_ARRAY:
	    return opc_aload;
	case TID_DOUBLE:
	    return opc_dload;
	case TID_FLOAT:
	    return opc_fload;
	case TID_LONG:
	    return opc_lload;
	case TID_BYTE:
	case TID_SHORT:
	case TID_CHAR:
	case TID_INT:
	case TID_BOOLEAN:
	    return opc_iload;
	default:
	    throw new RuntimeException("INTERNAL ERROR: " + type);
	}
    }

    /**
     * Returns the opcode to store a local variable of this type.
     */
    public /*@ pure @*/ int getStoreOpcode() {
	switch (type) {
	case TID_CLASS:
	case TID_ARRAY:
	    return opc_astore;
	case TID_DOUBLE:
	    return opc_dstore;
	case TID_FLOAT:
	    return opc_fstore;
	case TID_LONG:
	    return opc_lstore;
	case TID_BYTE:
	case TID_SHORT:
	case TID_CHAR:
	case TID_INT:
	case TID_BOOLEAN:
	    return opc_istore;
	default:
	    throw new RuntimeException("INTERNAL ERROR: " + type);
	}
    }

    /**
     * Returns the opcode to return a value of this type.
     */
    public int getReturnOpcode() {
	switch (type) {
	case TID_VOID:
	    return opc_return;
	case TID_CLASS:
	case TID_ARRAY:
	    return opc_areturn;
	case TID_DOUBLE:
	    return opc_dreturn;
	case TID_FLOAT:
	    return opc_freturn;
	case TID_LONG:
	    return opc_lreturn;
	case TID_BYTE:
	case TID_SHORT:
	case TID_CHAR:
	case TID_INT:
	case TID_BOOLEAN:
	    return opc_ireturn;
	default:
	    throw new RuntimeException("INTERNAL ERROR: " + type);
	}
    }

	/* Performs the instanceof / equality test for specializers of this
	 * type.  We assume that the actual argument has already been loaded.
	 *
	 * @param	code		the bytecode sequence to append to
	 * @param	nextInstruction	the instruction to branch to if the
	 *				dispatch test fails
	 */
    public void plantDispatchTest( CodeSequence code,
				   CodeLabel nextInstruction ) {
	fail();
    }
	

    // ----------------------------------------------------------------------
    // UTILITIES
    // ----------------------------------------------------------------------

    /**
     * Parse a java type signature
     *  Description : Attempts to parse the provided string as if it started with
     *    the Java VM-standard signature for a type.
     */
    public static CType parseSignature(String signature) {
	// AHS: call the other method
        return parseSignature(signature, null);
    }
    
    public static CType parseSignature(String signature,
                                       CUniverseTypeAnnotation univAnnot) {
        return instance.parseSignature(signature, univAnnot);
    }
    
    
    public static CType parseGenericTypeSignature(String signature) {
        // AHS: call the other method
	return parseGenericTypeSignature(signature, null);
    }
    
    public static CType parseGenericTypeSignature(String signature,
                                                  CUniverseTypeAnnotation univAnnot) {
	return instance.parseGenericTypeSignature(signature, univAnnot);
    }    

    /**
     * Returns a MethodSignature data structure by parsing the given
     * method or multimethod signature as read from bytecode.  */
    public static MethodSignature parseMethodSignature(String sig) {
        // AHS: call the other method
        return parseMethodSignature(sig, null);
    }
    
    public static MethodSignature parseMethodSignature(String sig,
            					       CUniverseMethodAnnotation univAnnot) {
        return instance.parseMethodSignature(sig, univAnnot);
    }
    

    /**
     * Generates a method signature as used in regular Java bytecode.  */
    public static String genMethodSignature(CType returnType, 
					    CType[] parameters) 
    {
	return genMethodSignatureHelper(returnType, null, parameters, false);
    }
    
    public static String genGenericMethodSignature(CTypeVariable[] tvs,
						   CType returnType, 
						   CType[] parameters) 
    {
	return genGenericMethodSignatureHelper(tvs, returnType, null, parameters, false);
    }
    
    /**
     * Generates a multimethod signature as used in custom bytecode
     * attributes.  */
    public static String genMultimethodSignature(CType returnType,
						 CType receiverType,
						 CSpecializedType[] parameters,
						 boolean skipSynthParameter) 
    {
	return genMethodSignatureHelper(returnType, receiverType, parameters,
					skipSynthParameter);
    }

    /**
     * Generates a multimethod signature as used in custom bytecode
     * attributes.  */
    public static String genGenericMethodSignature(CTypeVariable[] tvs,
						   CType returnType,
						   CType receiverType,
						   CSpecializedType[] parameters,
						   boolean skipSynthParameter) 
    {
	return genGenericMethodSignatureHelper(tvs, returnType, receiverType,
					       parameters,
					       skipSynthParameter);
    }
    
    
    public static String genGenericMethodSignatureHelper(CTypeVariable[] tvs,
							 CType returnType,
							 CType receiverType,
							 CTypeSignatureAppender[] parameters,
							 boolean skipSynthParameter)
    {
	String		result;
	FastStringBuffer buffer = CType.StringBuffers.getStringBuffer();
	if (tvs.length > 0) {
	    buffer.append('<');
	    for (int i = 0; i < tvs.length; i++) {
		tvs[i].appendDefinitionSignature(buffer);
	    }
	    buffer.append('>');
	}
	int start = skipSynthParameter ? 1 : 0;
	
	buffer.append('(');
	for (int i = start; i < parameters.length; i++) {
	    parameters[i].appendGenericSignature(buffer);
	}
	buffer.append(')');
	returnType.appendGenericSignature(buffer);
	if (receiverType != null) {
    	    receiverType.appendGenericSignature(buffer);
	}
	result = buffer.toString();
	CType.StringBuffers.release(buffer);
	return result;
    }
    
    /**
     * Helper method generates a method signature as used in regular
     * Java bytecode if receiverType is null, or a multimethod
     * signature otherwise. */
    private static String 
	genMethodSignatureHelper(CType returnType, CType receiverType,
				 CTypeSignatureAppender[] parameters,
				 boolean skipSynthParameter) 
    {
	FastStringBuffer buff = StringBuffers.getStringBuffer();
	
	buff.append('(');
	int start = skipSynthParameter ? 1 : 0;
	for (int i = start; i < parameters.length; i++) {
	    parameters[i].appendSignature(buff);
	}
	buff.append(')');

	returnType.appendSignature(buff);

	if (receiverType != null) {
	    receiverType.appendSignature(buff);
	}
	
	String s = buff.toString();
	StringBuffers.release(buff);
	return s;
    }

    
    /**
     * Constructs a nicely formatted string representing the type tuple given
     * by the argument array.
     *
     * @param	argTuple	length > 0
     * @return	a nicely format string
     */
    public static String tupleToString( CType[] argTuple ) {
	StringBuffer result = new StringBuffer( "<[ " 
						+ argTuple[0].toString() );
	for( int i=1; i<argTuple.length; i++ ) {
	    result.append( ", " + argTuple[i].toString() );
	}
	result.append( " ]>" );
	return result.toString();
    }

    /**
     * Constructs a nicely format string representing the type tuple given
     * by the arguments.
     *
     * @param 	recvType	the type of the first element in the tuple
     * @param	actuals		length >= 0
     * @return	a nicely format string
     */
    public static String tupleToString( CType recvType, CType[] actuals ) {
	StringBuffer result = new StringBuffer( "<[ " 
						+ recvType.toString() );
	for( int i=0; i<actuals.length; i++ ) {
	    result.append( ", " + actuals[i].toString() );
	}
	result.append( " ]>" );
	return result.toString();
    }

    // -------------------------------------------------------------
    // HELPER CLASSES
    // -------------------------------------------------------------
    

    // optimization
    /**
     * This inner class optimizes performance by maintaining a stack of
     * FastStringBuffers to be recycled.
     */
    public static class StringBuffers {
	static FastStringBuffer getStringBuffer() {
	    if (!ENV_USE_CACHE || stack.empty()) {
		return new FastStringBuffer();
	    }
	    FastStringBuffer buff = (FastStringBuffer)stack.pop();
	    buff.reset();

	    return buff;
	}

	static void release(FastStringBuffer buff) {
	    if (ENV_USE_CACHE) {
		stack.push(buff);
	    }
	}

	// -------------------------------------------------------------------
	// PRIVATE DATA MEMBERS
	// -------------------------------------------------------------------

	private static Stack stack = new Stack();
    }

    /**
     * This data structure represents a method type signature as an
     * array of specialized parameter types, a return type, and a
     * receiver type (which may be null for regular Java method
     * signatures).  */
    public static class MethodSignature {
	MethodSignature(CSpecializedType[] parameterTypes, CType returnType,
			CClassType receiverType, CTypeVariable[] tvs) {
	    this.parameterTypes = parameterTypes;
	    this.returnType = returnType;
	    this.receiverType = receiverType;
	    this.tvs = tvs;
	}
	public final CSpecializedType[] parameterTypes;
	public final CType returnType;
	public final CClassType receiverType;
	public final CTypeVariable[] tvs;
    }

    /**
     * This inner class provides methods for converting a bytecode
     * type signature into an instance of (a subclass of) CType for
     * use in compilation.  This conversion is used to determine the
     * types of members read from a *.class file.
     */
    public static class MethodSignatureParser {
	/**
	 * Parses a java type signature. Attempts to parse the
	 * provided string as if it started with the Java VM-standard
	 * signature for a type.
	 * @param	signature	a string containing the
	 * 				signature to be parsed
	 * @return	CType		a CType instance for the type
	 * 				given by signature
	 */
	public final CType parseSignature(String signature) {
	    // AHS: call the other method
	    return parseSignature(signature, null);
	}

	public final CType parseSignature(String signature,
	                                  CUniverseTypeAnnotation univAnnot) {
	    return parseSignature(signature, 0, signature.length(), univAnnot);
	}
	
	private CType parseSignature(String signature, int from, int to) {
	    // AHS: call the other method
	    return parseSignature(signature, from, to, null);
	}

	private CType parseSignature(String signature, int from, int to,
	                             CUniverseTypeAnnotation univAnnot) {
	    return parseSignature(signature, from, to, null, univAnnot);
	}
	
	// the staticType is used to parse value specializers
	// AHS: extended to to call getTypeRep with universe modifier
	private CType parseSignature(String signature, int from, int to,
				     CType staticType, CUniverseTypeAnnotation univAnnot) {
	    CType	type;
	    int	cursor = from;
	    while (signature.charAt(cursor) == '[') {
		cursor++;
	    }

	    switch (signature.charAt(cursor)) {
	    case 'V':
		type = CStdType.Void;
		break;
	    case 'B':
		type = CStdType.Byte;
		break;
	    case 'C':
		type = CStdType.Char;
		break;
	    case 'D':
		type = CStdType.Double;
		break;
	    case 'F':
		type = CStdType.Float;
		break;
	    case 'I':
		type = CStdType.Integer;
		break;
	    case 'J':
		type = CStdType.Long;
		break;
	    case 'L':
	        // AHS
	        if (CTopLevel.getCompiler().universeBytecode() &&
                    univAnnot != null) {
	            if (univAnnot.isArrayType()) {
	                type = CTopLevel.getTypeRep(signature.substring(cursor + 1, to - 1),
	                        		    univAnnot.getBaseUniverse(), true);
	            } else {
	                type = CTopLevel.getTypeRep(signature.substring(cursor + 1, to - 1),
	                        		    univAnnot.getUniverse(), true); 
	            }
	        } else {
	            type = CTopLevel.getTypeRep(signature.substring(cursor + 1, to - 1), true);
	        }
	            
		break;
	    case 'S':
		type = CStdType.Short;
		break;
	    case 'X':
		  /* a value specializer! */
	      type = parseValueSpecSignature(
		  signature.substring(cursor + 1, to - 1), staticType );
	      break;
	    case 'Z':
		type = CStdType.Boolean;
		break;
	    case 'T':
		type = new CTypeVariableAlias(signature.substring(cursor + 1, to - 1));
		break;
	    default:
		fail();
		return null;
	    }

	    cursor -= from;
	    
	    
	    if( type instanceof CClassType ) {
	    	// make sure that the Universe modifiers are correctly defaulted
	    	type = CUniverseServices.setDefaultVariableAgain(type);
	    } 
	    
	    // AHS
            
            if (cursor == 0) {
                return type;
            }
            if (type.isClassType() || cursor == 1) {
                if (CTopLevel.getCompiler().universeBytecode() &&
                        univAnnot != null) {
                    return new CArrayType(type, cursor, univAnnot.getUniverse());
                } else {
                    return new CArrayType(type, cursor, null);
                }
            }
            // reach here if we are a primitive array of >1 dimension
            if (CTopLevel.getCompiler().universeBytecode() &&
                    univAnnot != null) {
                return new CArrayType(type, cursor, univAnnot.getUniverse(), null);
            } else {
                return new CArrayType(type, cursor, null, null);
            }
            
            // TODO for primitive arrays we need a second modifier!
	    /* if (CTopLevel.getCompiler().universeBytecode() &&
                univAnnot != null)
	        return cursor > 0 ? new CArrayType(type, cursor, univAnnot.getUniverse()) : type;
	    else
	        return cursor > 0 ? new CArrayType(type, cursor, null) : type;
             */
	}
	
	public  CTypeVariable[] parseTypeParameter(String signature, 
							 char[] sig){
	    if (sig[current] == '<') {
		ArrayList    tvVect = new ArrayList(10);
		
		current++;
		while (sig[current] != '>') {
		    int             end = current; 
		    String          ident; // type var name
		    
		    while (sig[end] != ':') {
			end += 1;
		    }
		    ident = signature.substring(current, end);
		    current = ++end;
		    
		    ArrayList          bounds = new ArrayList(5); // type bound

                    // If no class but some interfaces have been specified in
                    // the bounds of a type variable, the signature does not
                    // simply contain no class bound, but an empty one, so let's
                    // account for this.
                    if (sig[current] == ':') {
                        current++;
                    }

		    bounds.add(parseGenericTypeSignature(signature, sig, null));
		    while (sig[current] == ':') { // another bound ?
			current++;
			bounds.add(parseGenericTypeSignature(signature, sig, null));
		    }
		    
		    CClassType[]    tvBound = (CClassType[])bounds.toArray(new CClassType[bounds.size()]);
		    
		    tvVect.add(new CTypeVariable(ident, tvBound));
		}
		
		CTypeVariable[]   typeVariable = (CTypeVariable[])tvVect.toArray(new CTypeVariable[tvVect.size()]);
		
		for (int i = 0; i < typeVariable.length; i++) {
		    typeVariable[i].setIndex(i);
		}
		current++;
		return typeVariable;
	    } else {
		return CTypeVariable.EMPTY;
	    }
	}
	
	public  CType parseGenericTypeSignature(String signature) {
	    // AHS: call the other method
	    return parseGenericTypeSignature(signature, null);
	}
	
	public  CType parseGenericTypeSignature(String signature,
	                                        CUniverseTypeAnnotation univAnnot) {
	    char[] sig = signature.toCharArray();
	    current = 0;
	    return parseGenericTypeSignature(signature, sig, univAnnot);
	}
	
	public CType parseGenericTypeSignature(String signature, char[] sig,
	                                       CUniverseTypeAnnotation univAnnot) {
	    int                 from = current;
	    CType               type = null;
	    
	    while (sig[current] == '[') {
		current += 1;
	    }
	    
	    if ((sig[current] != 'L') && (sig[current] != 'T')) {
		current += 1;
	    }
	    
	    else {
		while ((sig[current] != ';') && (sig[current] != '<')) {
		    current += 1;
		}
		current += 1;
	    }
	    if (sig[current-1] != '<') {
	        type = parseSignature(signature, from, current, univAnnot);
	    } else {
		String            ident = String.valueOf(sig, from+1, current - from-2);
		CClassType[]      types = parseTypeArgumentSignature(signature, sig);
		
		// make sure that the Universe modifiers are correctly defaulted
		type = CUniverseServices.setDefaultVariableAgain( 
			CTopLevel.getTypeRep(ident, new CClassType[][]{ types}, true) 
			 )
			;
	    }
	    return type;
	}
	
	public CClassType[] parseTypeArgumentSignature(String signature, char[] sig) {
	    ArrayList vect = new ArrayList(10);

	    while (sig[current] != '>') {
                CType type = null;
                switch (sig[current]) {
                case '*':
                    current++;
                    type = CWildcardType.createWildcard();
                    break;
                case '+':
                    current++;
                    CType upperBound = parseGenericTypeSignature(signature, sig, null);
                    assertTrue(upperBound instanceof CClassType,
                               "Bad classfile or internal compiler error: "
                               + "Wildcard must have a reference type as bound. "
                               + "Error occurred in signature: " + signature);
                    type = CWildcardType.createUpperBoundedWildcard(
                            (CClassType) upperBound);
                    break;
                case '-':
                    current++;
                    CType lowerBound = parseGenericTypeSignature(signature, sig, null);
                    assertTrue(lowerBound instanceof CClassType,
                               "Bad classfile or internal compiler error: "
                               + "Wildcard must have a reference type as bound. "
                               + "Error occurred in signature: " + signature);
                    type = CWildcardType.createLowerBoundedWildcard(
                            (CClassType) lowerBound);
                    break;
                default:
                    type = parseGenericTypeSignature(signature, sig, null);
                    break;
                }
                vect.add(type);
	    }

	    current++;
	    // verify(sig[current] == ';');
	    current++;

	    return (CClassType[]) vect.toArray(new CClassType[vect.size()]); 
	}
	
	private CType parseValueSpecSignature(/*@ non_null @*/ String value, 
					      /*@ non_null @*/ CType staticType) {
	    CType valType = null;
	    if (staticType.isBoolean()) {
		boolean b = value.equals("true");
		valType = new CBooleanValueType(
		    new JBooleanLiteral(org.multijava.util.compiler.TokenReference.NO_REF, b) );
	    }
	    else if (staticType.isOrdinal()) {
		JLiteral lit;
		if (staticType.equals(CStdType.Char)) {
			// char has the form '<escaped-char>'
			// unescape it and then strip the quotes
		    value = unescapeString(value);
		    lit = new JCharLiteral(org.multijava.util.compiler.TokenReference.NO_REF,
					   value.substring(1, value.length() - 1) );
		} else {
		    lit = new JOrdinalLiteral(org.multijava.util.compiler.TokenReference.NO_REF, value);
		}
		valType = new COrdinalValueType(staticType, lit);
	    }
	    else if (staticType.isFloatingPoint()) {
		valType = new CRealValueType(
		    staticType,
		    new JRealLiteral(org.multijava.util.compiler.TokenReference.NO_REF, value) );
	    }
	    else if (staticType.equals(CStdType.String)) {
		    // string has the form \"<escaped-string>\"
		    // unescape it and then strip the quotes
		value = unescapeString(value);
		valType = new CStringValueType(
			// strip off the quotes
		    new JStringLiteral(org.multijava.util.compiler.TokenReference.NO_REF,
				       value.substring(1, value.length() - 1)
				       ));
	    }
	    else {
		fail("did not find an appropriate static type for value specializer " + value);
	    }
		// set the specializer's value
	    try {
		valType.resolveValueType(null);
	    } catch (PositionedError e) {
		fail("could not reconstruct value specializer from bytecode attribute " + value);
	    }
	    
	    return valType;
	}

	/**
	 * Parses a Multijava method signature.
	 *
	 * @param	signature	a string containing the
	 * 				signature to be parsed
	 */
	public synchronized MethodSignature parseMethodSignature(String 
								 signature) {
	    // AHS: call the other method
	    return parseMethodSignature(signature, null);
	}
		
	public synchronized MethodSignature parseMethodSignature(String signature,
	        			CUniverseMethodAnnotation univAnnot) 
	{
	    CTypeVariable[] typevariables;
	    int paramCnt = 0;
	    
	    // assert(sig.charAt(0) == '(');
	    char[] sig = signature.toCharArray();
	    current = 0 ;
	    typevariables = parseTypeParameter(signature, sig);
	    current++;
	    while (sig[current] != ')') {
		paramCnt++;
		
	        if (sig[current] == '@') {
		    current++;
		    if (sig[current] == '@') {
			current++;
		    }
		    //int end = findEnd(current,sig);
		    
		    CType statType = (univAnnot == null || univAnnot.isEmptyParam() ) ? 
		            		parseGenericTypeSignature(signature,sig,null) :
		            		parseGenericTypeSignature(signature,sig,
		            		        univAnnot.getParamAnnotation(paramCnt-1));
		    
		    int end = findEnd(current,sig);
		    
		    CType dynType = parseSignature(signature, current, end, statType, null);
		    
		    current = end;
		    res.add(new CSpecializedType(statType,dynType));
		} else {
		    // AHS
		    if (CTopLevel.getCompiler().universeBytecode() &&
                        univAnnot != null &&
			!univAnnot.isEmptyParam() ) {
		        res.add(new CSpecializedType(
		                	parseGenericTypeSignature(signature, sig,
		                	        univAnnot.getParamAnnotation(paramCnt-1))));
		    } else {
		        res.add(new CSpecializedType(parseGenericTypeSignature(signature, sig, null)));
		    }
		}
	    }
	    
	    current++;		// skip trailing ')'
	    
	    CType ret;
	    if (CTopLevel.getCompiler().universeBytecode() &&
                univAnnot != null) { // AHS
	        ret = parseGenericTypeSignature(signature, sig, univAnnot.getRettypeAnnotation());
	    } else {
	        ret = parseGenericTypeSignature(signature, sig, null);
	    }
	    
	    int end = signature.length();
	    CClassType rec = null; // will remain null for regular Java sig
	    if (current < end) {
		rec = (CClassType) parseSignature( signature,current,end);
	    }

	    // store into array
	    CSpecializedType[] arr = new CSpecializedType[res.size()];

	    res.toArray(arr);
	    res.clear();

	    return new MethodSignature(arr,ret,rec,typevariables);
	}
	
	    // note: this code is essentially a copy of
	    // findParamDescriptionEnd in dis/Disassembler
	private int findEnd(int current, char[] sig) {
	    int end = current;
	    while (sig[end] == '[') {
		end += 1;
	    }
	    if (sig[end] == 'L') {
		while (sig[end] != ';') {
		    end += 1;
		}
		end += 1;
	    } else if (sig[end] == 'X') {
		    // we're at the end when we see a quote symbol with
		    // an even number of backslashes preceding it.
		boolean evenBackslashes = true;
		boolean notDone = true;
		while (notDone) {
		    while (sig[end] != '\"') {
			if (sig[end] == '\\')
			    evenBackslashes = !evenBackslashes;
			else
			    evenBackslashes = true;
			end += 1;
		    }
		    if (evenBackslashes) {
			notDone = false;
		    } else {
			evenBackslashes = true;
		    }
		    end += 1;
		}		
	    } else {
		end += 1;
	    }
	    return end;
	}

	ArrayList res = new ArrayList();
	int current;
    }


    /**
     * This nested class defines a collection of tuples where only the
     * most general type tuples are retained in the collection.  This
     * class is useful in gathering a set of tuples to be checked
     * during generic function completeness checking.  */
    public static class TupleCollection {
	public void add(CType[] tuple) {
	    // Scan through existing members, if tuple is more
	    // specific, then ignore it, if tuple is less specific
	    // then add it and remove existing member, if tuple is
	    // incomparable then add it.
	    Iterator iter = members.iterator();
	    boolean addTuple = true;
	loop:
	    while (iter.hasNext()) {
		CType[] existing = (CType[]) iter.next();
		switch (compareTuples(tuple,existing)) {
		case LESS_SPECIFIC:
		    iter.remove();
		    break;
		    
		case MORE_SPECIFIC:
		    // tuple does not belong, no need to keep checking
		    addTuple = false;
		    break loop;

		default:	// INCOMPARABLE
		    break;
		}
	    }
	    
	    if (addTuple) {
		members.add(tuple);
	    }
	}

	public CType[][] members() {
	    CType[][] result = new CType[members.size()][];
	    members.toArray(result);
	    return result;
	}
	
	public /*@ pure @*/ boolean isEmpty() {
	    return members.isEmpty();
	}

	public void clear() {
	    members.clear();
	}

	private int compareTuples( CType[] first, CType[] second ) {
	    if (first.length != second.length) {
		return INCOMPARABLE;
	    }
	    int countFirstSub = 0;
	    int countSecondSub = 0;
	    int countEqual = 0;
	    for (int i = 0; i < first.length; i++) {
		if (first[i].equals(second[i])) {
		    countEqual++;
		} else if (first[i].isAlwaysAssignableTo(second[i])) {
		    countFirstSub++;
		} else if (second[i].isAlwaysAssignableTo(first[i])) {
		    countSecondSub++;
		} else {
		    // incomparable in some position
		    return INCOMPARABLE;
		}
		
	    }
	    if (countEqual == first.length) {
		return EQUAL;
	    }
	    if (countFirstSub > 0 && countSecondSub == 0) {
		return MORE_SPECIFIC;
	    }
	    if (countSecondSub > 0 && countFirstSub == 0) {
		return LESS_SPECIFIC;
	    }
	    return INCOMPARABLE;
	}

	private ArrayList members = new ArrayList();

	private static final int LESS_SPECIFIC = 0;
	private static final int MORE_SPECIFIC = 1;
	private static final int INCOMPARABLE = 2;
	private static final int EQUAL = 3;
    }

    // ----------------------------------------------------------------------
    // PUBLIC DATA MEMBERS
    // ----------------------------------------------------------------------

	public static final CType[] EMPTY = new CType[0];

    // ----------------------------------------------------------------------
    // PRIVILEGED DATA MEMBERS
    // ----------------------------------------------------------------------

    protected int type;
    protected boolean isTypeVariable = false;
    protected boolean isMethodTypeVariable = false;
    protected static final MethodSignatureParser instance = 
	new MethodSignatureParser();

}
