package org.multijava.mjc;

/**
 * <P>This class performs nullity type information checks.
 */
public class CTypeNullity {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

   private CTypeNullity(long modifiers, CType type) {
        final boolean isNullable = (0!=(modifiers & Constants.ACC_NULLABLE))
                                || (0!=(modifiers & Constants.IMPLICITLY_NON_NULL)) ;
        final boolean isNonNull  = (0!=(modifiers & Constants.ACC_NON_NULL))
                                || (0!=(modifiers & Constants.ACC_NON_NULL_BY_DEFAULT));

        if (!(isNullable || isNonNull)) 
           throw new RuntimeException("modifier("+Long.toHexString(modifiers)+") is neither nullable nor non_null");
        if (isNullable && isNonNull) 
           throw new RuntimeException("modifier("+Long.toHexString(modifiers)+") is both nullable and non_null");
        this.isNonNull=isNonNull;
        this.type=type;
   }

   private CTypeNullity(boolean isNonNull, CType type) {
       this.isNonNull=isNonNull;
       this.type=type;
   }

    public static /*@non_null@*/ CTypeNullity getAlwaysNonNull(/*@non_null*/ CType type) {
       return new CTypeNullity(Constants.ACC_NON_NULL, type);
    }
    public static /*@non_null@*/ CTypeNullity getAlwaysNullable(/*@non_null*/ CType type) {
       return new CTypeNullity(Constants.ACC_NULLABLE, type);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public boolean equals(/*@nullable*/ Object that) {
       if (that instanceof CTypeNullity) 
          return this.isNonNull == ((CTypeNullity) that).isNonNull;
       else
	  return false;
    }

   /**
    *  Returns true if the nullity of left and right would
    *  allow the assignment left=right.
    *  Note that other aspects of the type must be taken
    *  into consideration.
    */
//   public boolean _isAssignable(CTypeNullity that) {
//      return _isAssignable(this.isNonNull, that.isNonNull, this.type.isArrayType());
//   }

   public static boolean isAssignable(/*@non_null*/ JLocalVariable left, /*@non_null*/ JExpression right, CContextType context) {
//System.out.println("\nleft is "+left.toString());
	CType    leftType =  left.getType();
	return isAssignable(left.isDeclaredNonNull(), right, leftType, context);
   }

   public static boolean isAssignable(/*@non_null*/ JMethodDeclaration left, /*@non_null*/ JExpression right, CContextType context) {
//System.out.println("\nleft is "+left.toString());
	CType    leftType =  left.returnType();
	return isAssignable(left.isDeclaredNonNull(), right, leftType, context);
   }

   public static boolean isAssignable(/*@non_null*/ JExpression left, /*@non_null*/ JExpression right, CContextType context) {
//System.out.println("\nleft is "+left.toString());
	CType    leftType =  left.getType();
	return isAssignable(left.isDeclaredNonNull(), right, leftType, context);
   }

   private static boolean isAssignable(boolean leftIsDeclaredNonNull, JExpression right, CType leftType, CContextType context) {
	boolean rightIsDeclaredNonNull = right.isDeclaredNonNull();
	boolean rightIsNonNull = right.isNonNull(context);

	if (CTopLevel.getCompiler().experimentalArrayHandling()) {
	   // assignment is only invariant for arrays of reference types
	   boolean isInvariant = leftType.isArrayType() && ! ((CArrayType) leftType).getBaseType().isPrimitive();
	   if (isInvariant) {
//System.out.println("isInvariant");
//System.out.println("isArrayExpression = "+isArrayExpression(right));
//System.out.println("leftIsDeclaredNonNull = "+leftIsDeclaredNonNull);
//System.out.println("rightIsNonNull = "+rightIsNonNull);
//System.out.println("rightIsDeclaredNonNull = "+rightIsDeclaredNonNull);
//System.out.println("right is a "+right.getClass());

	      if (isArrayExpression(right)) {
	        return leftIsDeclaredNonNull == rightIsNonNull;
	      } else {
	        return leftIsDeclaredNonNull == rightIsDeclaredNonNull;
	      }
	   } else {
//System.out.println("not isInvariant");
//System.out.println("leftIsDeclaredNonNull = "+leftIsDeclaredNonNull);
//System.out.println("rightIsNonNull = "+rightIsNonNull);
	      return !leftIsDeclaredNonNull || rightIsNonNull;
	   }
	}
	else
	   return !leftIsDeclaredNonNull || rightIsNonNull;
   }
   private static /*@pure*/ boolean isArrayExpression(JExpression expr) {
	return expr instanceof JArrayInitializer 
            || expr instanceof JNewArrayExpression;
   }

   public static String toString(boolean isNonNull) {
      if (isNonNull) return "non_null";
      else           return "nullable";
   }
   public String toString() {
      return toString(this.isNonNull);
   }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected final	boolean	isNonNull;      // not used
    protected final	CType	type;           // not used
}
