/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CTypeVariable.java,v 1.19 2008/03/16 15:52:45 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.FastStringBuffer;

public class CTypeVariable extends CClassType{
    public CTypeVariable(String name, CClassType[] bounds){
	super();
	this.bounds = bounds;
	this.name = name;
    }
    public CTypeVariable(CTypeVariable tv){
	super();
	bounds = tv.bounds;
	name = tv.name;
    }
    
    public String getIdent() {
	return name;
    }
    
    public CClassType[] getBounds() {
	return bounds;
    }

    public CType checkType(CContextType context) throws UnpositionedError {
	if(checked) {
	    return this;
	} 
	setTypeVariable(true);
	checked = true;
	for (int i = 0; i < bounds.length; i++) {
	    bounds[i].checkType(context);
	}
	
    // Checks for Generic Universe Types
    if (CTopLevel.getCompiler().universeChecks()) {
    	checkTypeUniverse(context);
    }
	
	if (bounds.length == 0) {
	    setClass(CTopLevel.getTypeRep(Constants.JAV_OBJECT, CUniverseServices.getDefaultReadonly(), true).getCClass());
	} else {
	    if (bounds.length == 1) {
	    bounds[0] = (CClassType)bounds[0].checkType(context);
	    setClass(bounds[0].getCClass());
	    } else { //more than one bound
		CClass first =null;
		bounds[0]= (CClassType)bounds[0].checkType(context);
		if(!bounds[0].getCClass().isInterface()){
		    setClass(bounds[0].getCClass());
		} else {
		    first = bounds[0].getCClass(); //is interface
		}
		for (int i = 1; i<bounds.length; i++) {
		    bounds[i] = (CClassType) bounds[i].checkType(context);
		    bounds[i] = (CClassType) CUniverseServices.setDefaultReadonlyAgain(bounds[i]); 

		    if (!bounds[i].getCClass().isInterface()) {
			throw new UnpositionedError(MjcMessages.TV_NOT_AN_INTERFACE, bounds[i]);
		    }
		    if ((first != null) && (first.qualifiedName().compareTo(bounds[0].getCClass().qualifiedName()) > 0)) {
			first = bounds[0].getCClass();
		    }
		}
		if(first!=null) {
		    setClass(first);
		}
	    }
	}

        if (context != null) {
            context.registerVisibleType(this);
        }
	checked = true;
	return this;
    }

    /**
     * Performs checks for Generic Universe Types.
     * 
     * Checks performed:
     * - rep modifiers must not occur in upper bounds of non-method
     *   type variables
     *   
     * by RZU
     */
    private void checkTypeUniverse(CContextType context)
    		throws UnpositionedError{
	    if (!isMethodTypeVariable() &&
	    	!isCapture() ) {
	    	if (CUniverseServices
	    			.containsRepModifier(bounds)) {
	    		throw new UnpositionedError(
	    				CUniverseMessages.REP_FORBIDDEN_IN_BOUND);
	    	}
    	}
	}
    
	public /*@ pure @*/ boolean isClassTypeVariable() {
        return !isMethodTypeVariable();
    }

    public void setMethodTypeVariable(boolean isMethodTypeVariable){
	this.isMethodTypeVariable = isMethodTypeVariable;
    }
    
    public boolean isMethodTypeVariable() {
	return isMethodTypeVariable;
    }

    public boolean isReifiableType() {
        return false;
    }

    public boolean changesByErasure() {
        return true;
    }

    /**
     * Can this type be converted to the specified type by casting
     * conversion (JLS 5.5) ?
     */
    public boolean isCastableTo(CType dest) {
    	// In case the type variable has no bounds, it is of type Object
    	// which everything can be casted to
    	if (getBounds().length == 0) {
    		return true;
    	}
    	
    	// If the type variable has bounds, we check if any of the bounds
    	// is castable to dest
		for (int i = 0; i < getBounds().length; i++) {
			if (getBounds()[i].isCastableTo(dest)) {
				return true;
			}
		}
			
		// Otherwise we decline
		return false;
    }
    
    public int hashCode() {
	if(isMethodTypeVariable()) {
	    return getCClass().hashCode();
	} else {
	    return super.hashCode();
	}
    }

	/**
	 * Checks for equality of <code>this</code> and <code>other</code>
	 *
	 * @param other The object to compare it to
	 * @param enableUniv Ignored in this version
	 * @return true iff they two are equal
	 */
    public boolean equals(/*@ nullable @*/ Object other,
    		boolean enableUniv,
    		boolean allowUniversesCovariance) {
    	return equals(other);
    }
    
	/**
	 * Checks for equality of <code>this</code> and <code>other</code>
	 *
	 * @param other The object to compare it to
	 * @return true iff they two are equal
	 */
    public boolean equals(/*@ nullable @*/ Object other) {
	if(!(other instanceof CClassType)) {
	    return false;
	}
	if(isMethodTypeVariable()) {
	    return getCClass().equals(((CClassType) other).getCClass());
	} else {
		// We compare type variables based on their name. This is correct since
		// we only compare type variables that belong to the same class, i.e.
		// substitution was performed before this method is called.
		
		// The safe alternative, equality operator ==, does not work because
		// we have multiple instances of the same type variable. Why?
		return (other instanceof CTypeVariable)
				&& ((CTypeVariable) other).getIdent().equals(getIdent());
	}
    }
    
	/**
	 * Checks for equality of <code>this</code> and <code>other</code>
	 * Performs substitution of <code>other</code> first
	 *
	 * @param other The object to compare it to
	 * @param subst The type arguments
	 * @return true iff they two are equal
	 */
    public boolean equals(/*@ nullable @*/ CType other,
			  /*@ non_null @*/ CClassType[] subst) {
	//method typevariable can not be substituted
	if(!(other instanceof CClassType)) {
	    return false;
	}
	if(isMethodTypeVariable()) {
	    return getCClass().equals(((CClassType)other).getCClass());
	}
	if (!other.isTypeVariable()) {
	    return false;
	} else {
	    return subst[((CTypeVariable) other).getIndex()] == this;
	}
    }
    
    public boolean isAlwaysAssignableTo(CType dest) {
    	return isAlwaysAssignableTo(dest, false);
    }
    
    public boolean isAlwaysAssignableTo(CType dest, boolean inst) {
    	return isAlwaysAssignableToImpl(dest, inst, true);
    }
    
    public boolean isAlwaysAssignableToIgnoreUniverses(CType dest) {
    	return isAlwaysAssignableToImpl(dest, false, false);
    }

    public boolean isAlwaysAssignableTo(CType dest, CClassType[] subst) {
    	return isAlwaysAssignableToImpl(dest, subst, true);
    }

    public boolean isAlwaysAssignableToIgnoreUniverses(CType dest, CClassType[] subst) {
    	return isAlwaysAssignableToImpl(dest, subst, false);
    }
    
    /**
     * The implementation of all isAssignableTo... methods that
     * take a parameter <code>inst</code>.
     * 
     * @param dest The destination of the assignment
     * @param inst ??? (could not find a description)
     * @param enableUniv Check universes iff true
     * @return True if <code>this</code> is assignable to <code>dest</code>
     */
	private boolean isAlwaysAssignableToImpl(CType dest,
			boolean inst,
			boolean enableUniv) {
		if (this == dest) {
		    return true;
		}

		if (dest.isCapture()) {
		    // A type is a subtype of a capture type variable iff it is a
		    // subtype of the capture's lower bound (see end of JLS3 4.10.2).
		    CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
		    if (lowerBound == null) {
		        return false;
		    } else {
		        return enableUniv
		        		? isAlwaysAssignableTo(lowerBound, inst)
		        		: isAlwaysAssignableToIgnoreUniverses(lowerBound);
		    }
		}

		if (dest.isTypeVariable()) {
		    if (!inst) {
			return false;
		    } else {
			CClassType[]    destBounds = ((CTypeVariable)dest).getBounds();
			for(int i = 0 ; i< destBounds.length ; i++) {
			    boolean isAssignable = false;
			    for (int j=0; j< bounds.length; j++ ) {
				if (bounds[j].isAlwaysAssignableTo(destBounds[i], true)){
				    isAssignable = true;
				    break;
				}
			    }
			    if (!isAssignable) return false;
			}
			return true;
		    }
		} else { 
		    if (!dest.isClassType()) {
			return false;
		    }
		    if (bounds.length == 0) {
    			return enableUniv
    					? CTopLevel.getTypeRep(Constants.JAV_OBJECT,
    							CUniverseImplicitReadonly.getUniverse(),
    							true)
    							.isAlwaysAssignableTo(dest)
    					: CTopLevel.getTypeRep(Constants.JAV_OBJECT, true)
    							.isAlwaysAssignableToIgnoreUniverses(dest);
		    } else {
			for (int k = 0; k<bounds.length; k++) {
				if ((enableUniv && bounds[k].isAlwaysAssignableTo(dest))
						|| (!enableUniv && bounds[k]
						.isAlwaysAssignableToIgnoreUniverses(dest))) {
					return true;
			    }
			}
		    }
		    return false;
		}
	}
        
    /**
     * The implementation of all isAssignableTo... methods that
     * perform substitution.
     * 
     * @param dest The destination of the assignment
     * @param subst The type arguments used for substitution
     * @param enableUniv Check universes iff true
     * @return True if <code>this</code> is assignable to <code>dest</code>
     */
    private boolean isAlwaysAssignableToImpl(CType dest,
    		CClassType[] subst,
    		boolean enableUniv) {
    	if (this == dest) {
    		return true;
    	}
    	if (dest.isClassTypeVariable()) {
    		int       index = ((CTypeVariable)dest).getIndex();
    		dest = subst[index];
    	} 

    	if (dest.isCapture()) {
    		// A type is a subtype of a capture type variable iff it is a
    		// subtype of the capture's lower bound (see end of JLS3 4.10.2).
    		CClassType lowerBound = ((CCaptureType) dest).getLowerBound();
    		if (lowerBound == null) {
    			return false;
    		} else {
    			dest = lowerBound;
    		}
    	}

    	if (dest.isTypeVariable()) {
    		return dest.equals(this, enableUniv);
    	} else {
    		if (bounds.length ==  0) {
    			return enableUniv
    					? CTopLevel.getTypeRep(Constants.JAV_OBJECT,
    							CUniverseImplicitReadonly.getUniverse(),
    							true)
    							.isAlwaysAssignableTo(dest)
    					: CTopLevel.getTypeRep(Constants.JAV_OBJECT, true)
    							.isAlwaysAssignableToIgnoreUniverses(dest);
    		} else { 
    			for (int i = 0 ; i< bounds.length ; i++) {
    				if ((enableUniv && bounds[i].isAlwaysAssignableTo(dest, subst))
    						|| (!enableUniv && bounds[i]
    						.isAlwaysAssignableToIgnoreUniverses(dest, subst))) {
    					return true;
    				}
    			}
    			return false;
    		}
    	}
    }

    public boolean needsUncheckedConversion(CType dest) {
        for (int i = 0; i < bounds.length; i++) {
            if (bounds[i].needsUncheckedConversion(dest)) {
                return true;
            }
        }
        return false;
    }

    /**
     *return the index of this type variable
     */
    public int getIndex(){
	return index;
    }

    
    
    /**
     *set the index of the type variable to the given int
     */
    public void setIndex(int i){
	index = i;
    }	
	
    public boolean hasSameBound(CType other){
	if (!(other instanceof CTypeVariable)){
	    return false;
	}
	CClassType[] otherbounds = ((CTypeVariable)other).getBounds();
	if (bounds.length != otherbounds.length){
	    return false;
	}
	for(int i = 0 ; i < bounds.length; i++){
	    if (!(bounds[i].equals(otherbounds[i]))){
		return false;
	    }
	}
	return true;
    }
    
    
    public String toString() {
    	return name;
    }
        
    public String toVerboseString() {
       	return toString();
    }

    
    /** Return the universe modifier as string.
     * Always returns a Universe string, even if the Universe checks
     * are turned off!
     * By WMD.
     */
    public String getCUniverseString() {
    	assertTrue(getCUniverse().equals(CUniverseThis.getUniverse()));
    	return "";
    }
    
    /** Return the universe modifier as string, for output in JML
     * specifications. The only difference is that implicit annotations
     * are escaped with a backslash.
     * By AHS, WMD.
     */
    public String getCUniverseJMLString() {
    	assertTrue(getCUniverse().equals(CUniverseThis.getUniverse()));
    	return "";
    }

    /**
     * Returns the universe modifier as String, for output in source
     * code parsed by MultiJava, i.e. the modifiers are written as is.
     * By RZU, WMD.
     */
    public String getCUniverseMJString() {
    	assertTrue(getCUniverse().equals(CUniverseThis.getUniverse()));
	return "";
    }
    
    
    // ----------------------------------------------------------------------
    // SIGNATURE GENERATING
    // ----------------------------------------------------------------------
    
    /**
     * Appends the generic signature (attribute) of this type variable to the specified buffer.
     */
    public void appendGenericSignature(FastStringBuffer buffer) {
 	buffer.append('T');
 	buffer.append(name);
 	buffer.append(';');
    }
    /**
     * Appends the VM signature of this type variable to the specified buffer.
     */
    public void appendSignature(FastStringBuffer buffer) {
 	buffer.append('L');
 	buffer.append(getCClass().qualifiedName());
 	buffer.append(';');
    }
    
    /**
     * Appends the generic signature of the definition of this type variable to the specified buffer.
     */
    protected void appendDefinitionSignature(FastStringBuffer buffer) {
 	buffer.append(name);
 	buffer.append(':');
 	if (bounds.length == 0) {
 	    buffer.append(CStdType.Object.getSignature());
 	} else {
            // If we have no class in the bounds but an interface, we still have
            // to provide an empty class bound in the signature.
            if (bounds[0].isInterface()) {
                buffer.append(':');
            }

 	    bounds[0].appendGenericSignature(buffer);
 	    for (int i = 1; i < bounds.length; i++) {
 		buffer.append(':');
 		bounds[i].appendGenericSignature(buffer);
 	    }
 	}
    }
    
    private CClassType[] bounds;
    private String name;
    private boolean checked;  //@ in thisClassInfo;

    public static final CTypeVariable[] EMPTY = new CTypeVariable[0];
}
