/*
 * Copyright (C) 2004-2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package org.multijava.mjc;

import org.multijava.util.Utils;

/**
 * This class represents a universe annotation for a method. This includes
 * purity and universe annotations for both parameters and return type.
 *
 * @author AHS
 */
public class CUniverseMethodAnnotation {

    /**
     * Construct a Universe method annotation from byte constants.
     * 
     * @param purity		the method's purity
     * @param retTypeConstant 	the return type constant
     * @param paramConstants	the parameter constants (captured)
     */
    public CUniverseMethodAnnotation(boolean purity, byte retTypeConstant,
                                     byte[] paramConstants) {
        this.purity = purity;
        this.retTypeConstant = retTypeConstant;
        this.paramConstants = paramConstants;
    }
    
    /**
     * Construct a Universe method annotation from byte constants for a method
     * with void or primitive return type.
     * 
     * @param purity		the method's purity
     * @param paramConstants	the parameter constants (captured)
     */
    public CUniverseMethodAnnotation(boolean purity, byte[] paramConstants) {
        this(purity, CUniverseImplicitPeer.byteConstant, paramConstants);
    }
    
    /**
     * Construct a Universe method annotation from byte constants for a
     * method with no or only primitive parameters.
     * 
     * @param purity		the method's purity
     * @param retTypeConstant	the return type constant
     */
    public CUniverseMethodAnnotation(boolean purity, byte retTypeConstant) {
        this(purity, retTypeConstant, EMPTY);
    }
    
    /**
     * Construct a Universe method annotation for a method with only primitive
     * parameter and return types.
     *  
     * @param purity	the method's purity
     */
    public CUniverseMethodAnnotation(boolean purity) {
        this(purity, CUniverseImplicitPeer.byteConstant, EMPTY); 
    }

    
    /**
     * Return the method's purity.
     */
    public boolean isPure() {
        return purity;
    }
    
    /**
     * Get the universe annotation for the i-th parameter.
     *
     * @param index the parameter's index
     */
    public CUniverseTypeAnnotation getParamAnnotation(int index) {
        if (index < 0 || index >= paramConstants.length)
            Utils.fail("parameter index out of range");
        
        return new CUniverseTypeAnnotation(paramConstants[index]);
    }

    /**
     * Check whether this is the dummy annotation for empty
     * universe annotations.
     */
    public boolean isEmptyParam() {
	return paramConstants == EMPTY;
    }


    /**
     * Get the universe annotation for the return value.
     */
    public CUniverseTypeAnnotation getRettypeAnnotation() {
        return new CUniverseTypeAnnotation(retTypeConstant);
    }
    
    

    /** the method's purity. */
    private boolean	purity;
    
    /** the byte constant for the return type. */
    private byte	retTypeConstant;
    
    /** the byte constants for the parameters. */
    private byte[]	paramConstants; 
    
    
    /** dummy value for empty parameter array. */
    private static final byte[]	EMPTY = new byte[0];
}
