/*
 * Copyright (C) 2003-2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CUniverseRuntimeHelper.java,v 1.3 2008/03/16 15:52:50 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;


/**
 * This is a helper class for the universe runtime checks.
 *
 * Its purpose is to store some information at a central place
 * and therefore make it easier to maintain the transformations
 * that are used to implement the runtime checks.
 *
 * @author scdaniel
 */
public class CUniverseRuntimeHelper {

    /// do not instantiate this class
    private CUniverseRuntimeHelper() {}
	
    /**
     * Creates a JNameExpression form a String.
     * Example String: "org.multijava.universes.rt.UniverseRuntime.handler"
     * 
     * @param ident	the String
     * @return	the corresponding JNameExpression
     */
    public static JNameExpression getNameExpression( String ident ) {
	int end;
	String str = ident;
	JNameExpression expr = null;
		
	while ( (end = str.indexOf('.')) != -1 ) {
	    if ( expr != null )
		expr = new JNameExpression( TokenReference.NO_REF, expr,
					    str.substring(0, end));
	    else
		expr = new JNameExpression( TokenReference.NO_REF,
					    str.substring(0, end));
	    str = str.substring(end + 1);
	}
		
	if ( expr != null )
	    expr = new JNameExpression(TokenReference.NO_REF, expr, str);
	else
	    expr = new JNameExpression(TokenReference.NO_REF, str);
		
	return expr;
    }
	
    /**
     * @return a JMethodCallExpression, that calls Thread.currentThread()
     */
    public static JMethodCallExpression getCurrentContext() {
	return new JMethodCallExpression( TokenReference.NO_REF,
					  getHashTable(),
					  "getContext",
					  new JExpression[0],
					  false);
    }
	
    /**
     * @return	a JNameExpression representing the hashTable class
     * 			used at runtime.
     */
    public static JNameExpression getHashTable() {
	return getNameExpression("org.multijava.universes.rt.UniverseRuntime.handler");
    }
	
    /**
     * @return	a JNameExpression representing the policy class
     * 			used at runtime.
     */
    public static JNameExpression getPolicy() {
	return getNameExpression("org.multijava.universes.rt.UniverseRuntime.policy");
    }
	
    /**
     * Get the owner of the current universe.
     * 
     * @param context	the current context
     * @return	the owner of the current universe
     */
    public static JExpression getCurrentUniverseOwner( CExpressionContextType context) {
	JExpression owner;

	if ( context.isStatic() )
	    owner = CUniverseRuntimeHelper.getCurrentContext();
	else
	    owner = new JThisExpression(TokenReference.NO_REF);
		
	return owner;
    }

    /**
     * Get a reference to a temporary variable to store expressions that are
     * used more than once, but should not be evaluated more than once of 
     * course.
     * 
     * The type of the variable returned is Object[] for array expressions
     * and Object else. 
     * 
     * @param context		the current context
     * @param currentNode	the current node ("this" wherever you call it from)
     * @param type			the type of the expression
     * 
     * @return	a reference to a temporary variable
     * @throws PositionedError
     */
    public static JLocalVariable getTmpVariable(
				      CExpressionContextType context,
				      JExpression currentNode,
				      CType type )
	throws PositionedError
    {
	JLocalVariable tmp;
	CUniverse univ = CUniverseReadonly.getUniverse();
		
	/* get variable (already defined in this context?) */
	if ( type instanceof CArrayType )
	    tmp = context.lookupLocalVariable(Constants.UNIV_ARRAY_TMP);
	else
	    tmp = context.lookupLocalVariable(Constants.UNIV_TMP);
		
	/* create if not defined already */
	if( tmp == null ) {
	    if ( type instanceof CArrayType ) {
		tmp = new JGeneratedLocalVariable( TokenReference.NO_REF, 0,
				  new CArrayType( CStdType.Object,
						  ((CArrayType) type).getArrayBound(),
						  univ),
				  Constants.UNIV_ARRAY_TMP, null);
	    } else {
		tmp = new JGeneratedLocalVariable( TokenReference.NO_REF, 0,
				  CTopLevel.getTypeRep( Constants.JAV_OBJECT,
							univ, true),
				  Constants.UNIV_TMP, null);
	    }

	    /* register */
	    try {
		context.getFlowControlContext().addVariable(tmp);
	    } catch( UnpositionedError e ) {
		throw e.addPosition(currentNode.getTokenReference());
	    }
			
	    context.getFlowControlContext().initializeVariable(tmp);
	}
		
	return tmp;
    }
    
    /**
     * Get the owner test that tests if testObject is compatible (can be casted
     * to) the type described by dest.
     * The return value can be null. Then the test is not necessary (testing
     * for readonly for example).
     * 
     * @param context		the current context
     * @param dest			the destination type (what to test for)
     * @param testObject	the location where the expression is found (ie
     * 						maybe a temporary variable)
     * @param expr			the expression itself
     * 
     * @return	the test for this conversion
     */
    public static JExpression getOwnerTest( CExpressionContextType context,
					    CType dest,
					    JExpression testObject,
					    JExpression expr ) {
	JExpression owner = CUniverseRuntimeHelper.getCurrentUniverseOwner(context);
		
	JExpression args[] = {owner, testObject};
		
	/* owner test */
    	JExpression ownerTest;
    	
    	/* if the type of the expression is != readonly it is
	 * statically checked */
	if ( ((CClassType) expr.getType()).getCUniverse() instanceof CUniverseReadonly ) {
	    CUniverse universe = ((CClassType) dest).getCUniverse();

	    if ( universe instanceof CUniversePeer ) {
		ownerTest = new JMethodCallExpression( TokenReference.NO_REF,
				       CUniverseRuntimeHelper.getHashTable(),
						       "isPeer",
						       args, false);
    	
	    } else if ( universe instanceof CUniverseRep ) {
		ownerTest = new JMethodCallExpression( TokenReference.NO_REF,
				       CUniverseRuntimeHelper.getHashTable(),
						       "isOwner",
						       args, false);
    	
	    } else {
			// ownerTest might be set below... hopefully
	    	// System.out.println("CURH 1; expr: " + expr.getType().toVerboseString() );
	    	// System.out.println("dest: " + dest.toVerboseString());
			ownerTest = null;
		}
	} else {
		// ownerTest might be set below... hopefully
		// System.out.println("CURH 2; expr: " + expr.getType().toVerboseString() );
    	// System.out.println("dest: " + dest.toVerboseString());
		ownerTest = null;
	}
    	
    	/* additional test for arrays */
    	while ( (dest instanceof CArrayType) &&
		(expr.getType() instanceof CArrayType) ) {

	    CArrayType type = (CArrayType) expr.getType();

	    /* only handle arrays of reference types */
	    if ( !(type.getBaseType() instanceof CClassType) )
		break;

	    /* if element type is != readonly it is statically checked */
	    CClassType baseType = (CClassType) type.getBaseType();
	    if ( !(baseType.getCUniverse() instanceof CUniverseReadonly) )
		break;
    		
	    /* get element type to check for */
	    int elementType;
	    CUniverse universe = ((CClassType) ((CArrayType) dest).getBaseType()).getCUniverse();

	    if ( universe instanceof CUniversePeer )
		elementType = MjcTokenTypes.LITERAL_peer;
	    else if ( universe instanceof CUniverseRep )
		elementType = MjcTokenTypes.LITERAL_rep;
	    else
		break;
    		
	    args = new JExpression[] { testObject,
			       new JOrdinalLiteral(TokenReference.NO_REF,
						   elementType,
						   CStdType.Integer)};
    		 
	    JMethodCallExpression m =
		new JMethodCallExpression(TokenReference.NO_REF,
				  CUniverseRuntimeHelper.getHashTable(),
					  "checkArrayType",
					  args, false);

	    if ( ownerTest != null )
		ownerTest =
		    new JConditionalAndExpression(TokenReference.NO_REF,
						  ownerTest, m);
	    else
		ownerTest = m;
    		
	    break;
    	}
    	   	
    	// org.multijava.util.Utils.assertTrue( ownerTest != null, "CURH: The ownertest should not be null!" );
    	return ownerTest;
    }
}

