/*
 * Copyright (C) 2003-2007 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
package org.multijava.mjc;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.multijava.util.Utils;

/**
 * Service which handles viewpoint adaptation for Generic Universe Types.
 * No instantiation possible.
 * 
 * Passed types are not changed. Instead, they are cloned, the universes
 * updated and returned.
 * 
 * Implemented according to Werner Dietl, Sophia Drossopoulou, Peter Mueller.
 * Generic Universe Types. ECOOP 2007. Section 3.2 Viewpoint Adaptation
 * 
 * @author RZU and WMD
 */
public class CUniverseServices extends Utils {

    /**
     * Private constructor to disallow instatiation.
     */
    private CUniverseServices() {
    }

    /**
     * Adapts a Type w.r.t. a Type.
     * 
     * Note: Call {@link #substituteInheritedTypeVariables(CClassType, CClassType, CContextType)}
     *       beforehand!
     *       
     * @param first The type to adapt from
     * @param second The type to adapt
     * @return <code>second</code> adapted from the viewpoint of
     *         <code>first</code>
     */
    public static CClassType combine(CClassType first, CClassType second) {
        // u \triangleright T
        CClassType adaptedType = combine(first.getCUniverse(), second);

        // Substitution happens in the caller because this is specific to
        // the caller
        // - JClassFieldExpression#typecheck(CExpressionContextType)
        // - JMethodCallExpression#checkParameterUniverse(CType,
        //		 	 JExpression,CContextType)

        return adaptedType;
    }

    /**
     * Adapts an Ownership Modifier w.r.t. an Ownership Modifier.
     * 
     * 
     * @param first The ownership modifier to adapt from
     * @param second The ownership modifier to adapt
     * @return The combination of the two given universe modifiers
     */
    private static CUniverse combine(CUniverse first,
            CUniverse second) {
        // this \triangleright u' = u'
        if (first instanceof CUniverseThis) {
            return second;
        // u' \triangleright this = u'
        } else if (second instanceof CUniverseThis) {
            return first;
        // rep \triangleright peer = rep
        } else if (first instanceof CUniverseRep && second instanceof CUniversePeer) {
            return CUniverseRep.getUniverse();
        // peer \triangleright peer = peer
        } else if (first instanceof CUniversePeer && second instanceof CUniversePeer) {
            return CUniversePeer.getUniverse();
        // all other cases evaluate to any (readonly)
        } else {
            return CUniverseReadonly.getUniverse();
        }
    }

    /**
     * Adapts a Type w.r.t. an Ownership Modifier.
     * 
     * @param modifier The ownership modifier to adapt <code>type</code> to
     * @param type The type to adapt
     * @return <code>type</code> adapted to <code>modifier</code>
     */
    public static CClassType combine(CUniverse modifier, CClassType type) {
        // combining with "this" never changes anything, shortcurcuit.
        if (modifier instanceof CUniverseThis) {
            return type;
        }

        // Type variables don't need to be adapted
        // u \triangleright X = X
        if (type instanceof CTypeVariable) {
            // WMD TODO what should happen to captured types?
            return type;
        }

        if (type instanceof CWildcardType) {
            CClassType clonedType = null;
            try {
                clonedType = (CClassType) type.clone();
            } catch (CloneNotSupportedException e) {
                fail("Clone of CClassType failed! Exception: " + e);
            }

            ((CWildcardType) clonedType).setBound(combine(modifier, ((CWildcardType) type).getBound()));
            return clonedType;
        }

        // u \triangleright N = (u \triangleright_m N) C<\bar{u \triangleright T}>
        // In order to not modify the existing type, we clone it first
        CClassType clonedType = null;
        try {
            clonedType = (CClassType) type.clone();
        } catch (CloneNotSupportedException e) {
            fail("Clone of CClassType failed! Exception: " + e);
        }

        // Set the type's main modifier
        clonedType.setCUniverse(combineM(modifier, clonedType));

        // Adapt type arguments
        clonedType.setAllArguments(adaptTypeArguments(modifier, type));

        return clonedType;
    }

    /**
     * Determines the main modifier when adapting a Type w.r.t. an 
     * Ownership Modifier.
     * 
     * @param modifier The ownership modifier
     * @param type The type whose main modifier has to be determined.
     *             Needs to have adapted viewpoints.
     * @return The main modifier of <code>type</code>
     */
    private static CUniverse combineM(CUniverse modifier, CClassType type) {
        // any    if (u = rep \/ u = peer) /\ u' = peer /\ rep \in \bar{T}
        if ((modifier instanceof CUniverseRep ||
                modifier instanceof CUniversePeer ||
                modifier instanceof CUniverseThis) && (type.getCUniverse() instanceof CUniversePeer ||
                type.getCUniverse() instanceof CUniverseThis) && containsRepModifier(type.getArguments())) {
            return CUniverseReadonly.getUniverse();
        } else {
            // u \triangleright u'    otherwise
            return combine(modifier, type.getCUniverse());
        }
    }

    /**
     * Looks for a rep modifier, either as main modifier or recursively
     * in the type arguments.
     * 
     * @param types Types to be searched
     * @return True iff at least one "rep" modifier occurs in the transitive
     *         closure of all types
     */
    public static boolean containsRepModifier(CClassType[] types) {
        if (types == null) {
            return false;
        }

        // Search each type
        boolean repFound = false;
        for (int i = 0; i < types.length; i++) {
            repFound |= containsRepModifier(types[i]);
        }

        return repFound;
    }

    /**
     * Looks for a rep modifier, either as main modifier or recursively
     * in the type arguments.
     * 
     * @param type Type to be searched
     * @return True iff at least one "rep" modifier occurs in the transitive
     *         closure of the type
     */
    public static boolean containsRepModifier(CClassType type) {
        if (type == null) {
            return false;
        }

        // Type variables contain no universe modifier
        if (type.isTypeVariable()) {
            return false;
        }

        if (type.isWildcard()) {
            return containsRepModifier(((CWildcardType) type).getBound());
        }

        // Check the main modifier first
        if (type.getCUniverse() instanceof CUniverseRep) {
            return true;
        }

        // DFS through all type arguments
        if (type.getAllArguments().length != 0) {
            CClassType[] typeArguments = type.getArguments();
            for (int i = 0; i < typeArguments.length; i++) {
                // Check if one of its type arguments is "rep"
                if (containsRepModifier(typeArguments[i])) {
                    return true;
                }
            }
        }

        // In case we can't find a "rep" there is none -> return false
        return false;
    }

    /**
     * Adapts the viewpoint of the type arguments when adapting a Type w.r.t.
     * an Ownership modifier
     * 
     * @param modifier The ownership modifier to adapt <code>type</code> to
     * @param type The type to adapt
     * @return A copied version of the arguments array with a replaced type
     *         arguments array having updated viewpoints
     */
    private static CClassType[][] adaptTypeArguments(CUniverse modifier,
            CClassType type) {
        // Recursively combine the type parameters, if any
        if (type.getAllArguments().length != 0 && type.getArguments() != null) {
            // Copy the type arguments array, to not touch the existing
            CClassType[] copiedTypeArguments = new CClassType[type.getArguments().length];
            System.arraycopy(type.getArguments(),
                    0,
                    copiedTypeArguments,
                    0,
                    type.getArguments().length);

            // Apply viewpoint adaptation to type parameters
            for (int i = 0; i < type.getArguments().length; i++) {
                copiedTypeArguments[i] = combine(modifier,
                        type.getArguments()[i]);
            }

            // Set type arguments of cloned type to copied version (to which
            // viewpoint adaptation has been applied)
            // Requires copying the whole arguments array since modifying the
            // existing would also affect the original type
            CClassType[][] copiedArguments = new CClassType[type.getAllArguments().length][];
            System.arraycopy(type.getAllArguments(),
                    0,
                    copiedArguments,
                    0,
                    type.getAllArguments().length);

            // Replace the type arguments which are the last entry in the
            // arguments array
            copiedArguments[type.getAllArguments().length - 1] = copiedTypeArguments;

            return copiedArguments;
        } else {
            // If there are no type arguments, we simply return the existing
            // arguments array since there have and will be no changes
            return type.getAllArguments();
        }
    }

    /**
     * Substitutes all inherited type variables of the passed type.
     * 
     * This is required to successfully perform viewpoint adaptation, since
     * inherited type variables are meant respectively to 
     * 
     * @param context The context.
     * @return The substituted field type.
     */
    public static CClassType substituteInheritedTypeVariables(
            CClassType type,
            CClassType prefixType,
            CContextType context) {
        CClass local = context.getClassContext().getHostClass();

        // If the type is a type variable (T a), we use the receiver's type
        // arguments to replace it
        if (type.isTypeVariable()) { // T
            return prefixType.getCClass().getSubstitution(
                    (CTypeVariable) type,
                    prefixType.isTypeVariable()
                    ? ((CTypeVariable) prefixType).getBounds()[0].getAllArguments()
                    : prefixType.getAllArguments(),
                    true);

        // If the type is an array whose type is a type variable (T[] a),
        // we use the receiver's type arguments to replace it
        } else if (type.isArrayType() && ((CArrayType) type).getBaseType().isTypeVariable()) {
            CClassType subtype = prefixType.getCClass().getSubstitution((CTypeVariable) ((CArrayType) type).getBaseType(),
                    prefixType.getAllArguments(),
                    true);
            return new CArrayType(subtype,
                    ((CArrayType) type).getArrayBound(),
                    ((CArrayType) type).getCUniverse());

        // If the type is an array with a generic base type (List<Integer>[] a),
        // we use the receiver's type arguments to replace type variables in
        // the type arguments, if any
        } else if (type.isArrayType() && ((CArrayType) type).getBaseType().isGenericType()) {
            CType basetype = ((CArrayType) type).getBaseType();
            CClassType newType = ((CClassType) basetype).createSubstitutedType(local,
                    (CClassType) prefixType,
                    basetype.getAllArguments(),
                    true);
            return new CArrayType(newType,
                    ((CArrayType) type).getArrayBound(),
                    ((CArrayType) type).getCUniverse());

        // If the type is an array with a generic base type (List<Integer>[] a),
        // we use the receiver's type arguments to replace type variables in
        // the type arguments, if any
        } else if (type.isGenericType()) {
            return ((CClassType) type).createSubstitutedType(local,
                    (CClassType) prefixType,
                    local.getType().getAllArguments(),
                    true);
        }

        // Otherwise, we return the existing field type
        return type;
    }

    /** Returns the default ownership modifier that should be used
     * for variables.
     * 
     * @return The ownership modifier for variables.
     */
    public static CUniverse getDefaultVariable() {
        if (CTopLevel.getCompiler() != null &&
                CTopLevel.getCompiler().universeChecks()) {
            return CUniverseImplicitPeer.getUniverse();
        } else {
            return CUniverseThis.getUniverse();
        }
    }

    /**
     * Set the ownership modifiers to default values for a variable.
     * This method can be called, even if the type is not completely resolved yet.
     * But beware, only the main Universe modifier will be set correctly and you
     * have to ensure that the xxxAgain method is also called!
     * 
     * @param toReplace the type for which the default Universe modifier should be set
     * @return A clone of the input type
     */
    public static CType setDefaultVariable(CType toReplace) {
        // return toReplace;
        if (toReplace != null &&
                toReplace instanceof CClassType &&
                CTopLevel.getCompiler() != null &&
                CTopLevel.getCompiler().universeChecks()) {
            // return toReplace;
            return cloneReplaceThis((CClassType) toReplace, CUniverseImplicitPeer.getUniverse());
        } else {
            return toReplace;
        }
    }

    /**
     * Replaces "this" as a main modifier by implicit "peer", if applicable.
     * 
     * Reason:
     * In some places, e.g. definitions of local variables or method parameters,
     * if we're not given universe modifiers, we simply put "this" there. This
     * is obviously wrong.
     * 
     * @param toReplace Type to check and replace, if applicable.
     * @return The updated type.
     */
    // public static CClassType replaceThisReference(CClassType toReplace) {
    //	return replaceMainMod( toReplace, CUniverseThis.getUniverse(), CUniverseImplicitPeer.getUniverse());
    //}
    public static CType setDefaultVariableAgain(CType toReplace) {
        // return toReplace;
        if (toReplace != null &
                toReplace instanceof CClassType &&
                CTopLevel.getCompiler() != null &&
                CTopLevel.getCompiler().universeChecks()) {
            // return toReplace;
            return cloneReplaceRec((CClassType) toReplace, CUniverseImplicitPeer.getUniverse());
        } else {
            return toReplace;
        }
    }

    /** Determine whether the given CClassType is immutable.
     * At the moment we just use a fixed set of classes, but this could
     * be extended to take JML immutability into account. However, this
     * would create a dependency onto JML.
     * This method has to be called after name resolution of the type.
     * 
     * @param cls The class to test.
     * @return True, iff the class should be considered immutable.
     */
    private static boolean isImmutable(CClassType cls) {
        String id = cls.qualifiedName();
        // System.out.println("Testing: " + id);

        String[] imm = {Constants.JAV_STRING,
            Constants.JAV_CLASS,
            "java/lang/Boolean",
            "java/lang/Byte",
            "java/lang/Character",
            "java/lang/Double",
            "java/lang/Float",
            "java/lang/Integer",
            "java/lang/Long",
            "java/lang/Number",
            "java/lang/Short",
            "java/math/BigInteger"
        };

        for (int i = 0; i < imm.length; ++i) {
            if (id.equals(imm[i])) {
                return true;
            }
        }
        return false;
    }

    /** Get the default ownership modifier for variables that are
     * readonly by default.
     * 
     * @return Default readonly modifier.
     */
    public static CUniverse getDefaultReadonly() {
        if (CTopLevel.getCompiler() != null &&
                CTopLevel.getCompiler().universeChecks()) {
            return CUniverseImplicitReadonly.getUniverse();
        } else {
            return CUniverseThis.getUniverse();
        }
    }

    /**
     * Set the ownership modifiers to default values for a readonly variable.
     * This method can be called, even if the type is not completely resolved yet.
     * But beware, only the main Universe modifier will be set correctly and you
     * have to ensure that the xxxAgain method is also called!
     * 
     * @param toReplace the type for which the default Universe modifier should be set
     * @return A clone of the input type
     */
    public static CType setDefaultReadonly(CType toReplace) {
        // return toReplace;
        if (toReplace != null &&
                toReplace instanceof CClassType &&
                CTopLevel.getCompiler() != null &&
                CTopLevel.getCompiler().universeChecks()) {
            // return toReplace;
            return cloneReplaceThis((CClassType) toReplace, CUniverseImplicitReadonly.getUniverse());
        } else {
            return toReplace;
        }

    }

    /**
     * Replaces "this" as a main modifier by implicit "readonly", if applicable.
     * 
     * Reason:
     * In some places, e.g. definitions of local variables or method parameters,
     * if we're not given universe modifiers, we simply put "this" there. This
     * is obviously wrong.
     * 
     * @param toReplace Type to check and replace, if applicable.
     * @return The updated type.
     */
    public static CType setDefaultReadonlyAgain(CType toReplace) {
        // return toReplace;
        if (toReplace != null &&
                toReplace instanceof CClassType &&
                CTopLevel.getCompiler() != null &&
                CTopLevel.getCompiler().universeChecks()) {
            // return toReplace;
            return cloneReplaceRec((CClassType) toReplace, CUniverseImplicitReadonly.getUniverse());
        } else {
            return toReplace;
        }
    }

    /** Recursively replace the implicit ownership modifiers in toReplace by the 
     * modifier to.
     * 
     * @param toReplace Input type to replace.
     * @param to Ownersip modifier to use.
     * @return A clone of toReplace that uses to instead of implicit modifiers.
     */
    private static CClassType cloneReplaceRec(CClassType toReplace, CUniverse to) {
        if (toReplace instanceof CTypeVariable ||
                toReplace instanceof CWildcardType) {
            // System.out.println("Ignoring: " + toReplace.toVerboseString());
            return toReplace;
        }

        CClassType replacedType;

        if (isImmutable(toReplace)) {
            replacedType = cloneReplaceThisAndImplPeer(toReplace, CUniverseImplicitReadonly.getUniverse());
        } else {
            replacedType = cloneReplaceThis(toReplace, to);
        }


        // If it is an array type, we have to look at the declared base type
        if (replacedType instanceof CArrayType) {
            CType baseType = ((CArrayType) replacedType).getBaseType();

            // We only care about base types of class type
            if (baseType instanceof CClassType) {
                CClassType cbaseType = (CClassType) baseType;

                if (isImmutable(cbaseType) &&
                        !(cbaseType.getCUniverse() instanceof CUniverseThis) &&
                        !(cbaseType.getCUniverse() instanceof CUniverseImplicitPeer) &&
                        !(cbaseType.getCUniverse() instanceof CUniverseImplicitReadonly) &&
                        (replacedType.getCUniverse() instanceof CUniverseImplicitPeer ||
                        replacedType.getCUniverse() instanceof CUniverseImplicitReadonly)) {
                    
                    // the array is of an immutable type, has a modifier for the
                    // base type, but no explicit main modifier.
                    // we switch the modifiers then.
                    // TODO direct modification of base type might not be ok!
                    replacedType.setCUniverse(cbaseType.getCUniverse());
                    cbaseType.setCUniverse(CUniverseImplicitReadonly.getUniverse());
                }

                cbaseType = cloneReplaceRec(cbaseType, to);

                ((CArrayType) replacedType).setBaseType(cbaseType);
            }
        }

        CClassType[][] args = replacedType.getAllArguments();
        if (args != null) {
            CClassType[] argsi;
            for (int i = 0; i < args.length; ++i) {
                if (args[i] != null) {
                    argsi = args[i];
                    for (int j = 0; j < argsi.length; ++j) {
                        argsi[j] = cloneReplaceRec(argsi[j], to);
                    }
                }
            }
        }

        return replacedType;
    }

    /** Clone toReplace and set the main modifier to to, if it is "this".
     * 
     * @param toReplace Input type.
     * @param to New ownership modifier.
     * @return A clone of toReplace that uses to as main modifier.
     */
    private static CClassType cloneReplaceThis(CClassType toReplace, CUniverse to) {
        if (toReplace instanceof CTypeVariable ||
                toReplace instanceof CWildcardType) {
            return toReplace;
        }

        CClassType replacedType = null;

        // We clone the type if the main modifier has to be replaced
        if (toReplace.getCUniverse() == null ||
                toReplace.getCUniverse() instanceof CUniverseThis) {
            try {
                replacedType = (CClassType) toReplace.clone();
            } catch (CloneNotSupportedException e) {
                fail("Clone of CClassType failed! Exception: " + e);
            }
            replacedType.setCUniverse(to);
            return replacedType;
        } else {
            return toReplace;
        }
    }

    /** Clone toReplace and set the main modifier to to, if it is "this" or 
     * implicit "peer".
     * 
     * @param toReplace Input type.
     * @param to New ownership modifier.
     * @return A clone of toReplace that uses to as main modifier.
     */
    private static CClassType cloneReplaceThisAndImplPeer(CClassType toReplace, CUniverse to) {
        if (toReplace instanceof CTypeVariable ||
                toReplace instanceof CWildcardType) {
            // System.out.println("Ignoring: " + toReplace.toVerboseString());
            return toReplace;
        }

        CClassType replacedType = null;

        // We clone the type if the main modifier has to be replaced
        if (toReplace.getCUniverse() == null ||
                toReplace.getCUniverse() instanceof CUniverseThis ||
                toReplace.getCUniverse() instanceof CUniverseImplicitPeer) {
            try {
                replacedType = (CClassType) toReplace.clone();
            } catch (CloneNotSupportedException e) {
                fail("Clone of CClassType failed! Exception: " + e);
            }
            replacedType.setCUniverse(to);
            return replacedType;
        } else {
            return toReplace;
        }
    }

    public static boolean isPartial(CUniverse u) {
        return u instanceof CUniverseThis ||
                u instanceof CUniverseImplicitPeer ||
                u instanceof CUniverseImplicitReadonly;
    }

    /** Check whether the given type contains only partial Universe modifiers,
     * i.e. some of the modifiers remained implicit.
     * 
     * @param type The type to check.
     * @return true if the type contained an implicit Universe modifier.
     */
    public static boolean isPartial(CClassType type) {
        if (type.isTypeVariable()) {
            return false;
        }
        /* if( type instanceof CTypeVariable ) {
        CClassType[] bs = ((CTypeVariable)type).getBounds();
        if( bs!=null && bs[0]!=null) {
        type = bs[0];
        } else {    			
        return false;
        }
        } */

        if (type instanceof CWildcardType) {
            type = ((CWildcardType) type).getBound();
        }

        CUniverse u = type.getCUniverse();

        if (isPartial(u)) {
            return true;
        }

        if (type.isArrayType()) {
            CType bt = ((CArrayType) type).getBaseType();

            if (bt instanceof CClassType) {
                return isPartial((CClassType) bt);
            } else {
                if (((CArrayType) type).getArrayBound() > 1) {
                    return isPartial(((CArrayType) type).getCUniverseElem());
                }
            }
        }

        if (type.isRawType()) {
            return false;
        }
        CClassType[] args = type.getArguments();

        if (args != null) {
            for (int i = 0; i < args.length; ++i) {
                if (isPartial(args[i])) {
                    return true;
                }
            }
        }

        return false;
    }

    /** Propagate the Universe modifiers from template into target.
     * Existing modifiers in target are preserved.
     * The returned object is a clone of target.
     * 
     * @param target The type that should get additional Universe modifiers.
     * @param template The input type that contains additional modifiers.
     * @return A clone of target that contains more modifiers.
     * 
     * By WMD.
     */
    public static CClassType propagate(CClassType target, CClassType template) {
        if (target instanceof CTypeVariable) {
            // TVs do not take any Universe modifier
            return target;
        }
        if (template instanceof CTypeVariable) {
            CClassType[] bs = ((CTypeVariable) template).getBounds();
            if (bs != null && bs[0] != null) {
                template = bs[0];
            } else {
                template = CStdType.WildcardBoundObject;
            }
        }

        CClassType res = null;
        CWildcardType reswc = null;

        try {
            res = (CClassType) target.clone();
        } catch (CloneNotSupportedException e) {
            fail("Clone of CClassType failed! Exception: " + e);
        }

        if (res instanceof CWildcardType) {
            CClassType wcb = ((CWildcardType) res).getBound();
            reswc = (CWildcardType) res;
            if (wcb != null) {
                res = wcb;
            } else {
                res = CStdType.WildcardBoundObject;
            }
        }


        // make sure that both have the same number of type arguments
        CClassType eq = raise(res, template);

        if (eq == null) {
            eq = lower(res, template);
        }
        
        if (eq == null) {
        // Could not determine common class, just ignore.
        // fail("Could not determine common supertype of " + res.toVerboseString() + " and " + template.toVerboseString());
        } else {
            CUniverse u = res.getCUniverse();

            if (isPartial(u)) {
                res.setCUniverse(eq.getCUniverse());
            }

            if (target.isArrayType() &&
                    eq.isArrayType()) {
                CType bta = ((CArrayType) target).getBaseType();
                CType btb = ((CArrayType) eq).getBaseType();
                CArrayType resa = (CArrayType) res;

                if (bta instanceof CClassType &&
                        btb instanceof CClassType) {
                    resa.setBaseType(propagate((CClassType) bta, (CClassType) btb));
                } else if (bta instanceof CClassType &&
                        btb.isPrimitive()) {
                    if (((CArrayType) eq).getArrayBound() > 1) {
                        ((CClassType) resa.getBaseType()).setCUniverse(((CArrayType) eq).getCUniverseElem());
                    }
                } else if (bta.isPrimitive() &&
                        btb instanceof CClassType) {
                    fail("Should never reach this point.");
                } else {
                    // both base types should be primitive now
                    if (resa.getArrayBound() > 1) {
                        resa.setCUniverseElem(((CArrayType) eq).getCUniverseElem());
                    }

                }
            }

            CClassType[] argsa = res.getArguments();
            CClassType[] argsb = eq.getArguments();

            if (argsa != null) {
                assertTrue(argsa.length == argsb.length);

                for (int i = 0; i < argsb.length; ++i) {
                    if (argsb[i] != null) {
                        argsa[i] = propagate(argsa[i], argsb[i]);
                    }
                }
            }
        }

        if (reswc != null) {
            reswc.setBound(res);
            return reswc;
        } else {
            return res;
        }
    }

    /** Starting from template, go up the subtype tree and ensure that the
     * returned CClassType and the target CClassType have the same underlying class.
     * 
     * @param target The type that determines to which class to raise. 
     * @param template The type from which to go up the subtype tree.
     * @return A type that has the same CClass as target.
     */
    private static CClassType raise(CClassType target, CClassType template) {
        if (target.getCClass().equals(template.getCClass())) // This is also true for arrays. This is okay, as the
        // base types will be propagated later.
        {
            return template;
        }

        CClassType sc = template.getSuperClassType();
        if (sc != null) {
            if (target.getCClass().equals(sc.getCClass())) {
                return sc;
            } else {
                CClassType es = raise(target, sc);
                if (es != null) {
                    return es;
                }
            }
        }

        CClassType[] si = template.getSuperInterfaceTypes();
        if (si != null) {
            for (int i = 0; i < si.length; ++i) {
                if (si[i] != null) {
                    if (target.getCClass().equals(si[i].getCClass())) {
                        return si[i];
                    } else {
                        CClassType es = raise(target, si[i]);
                        if (es != null) {
                            return es;
                        }
                    }
                }
            }
        }

        // if we reach here, we did not find target in the supertypes
        // just stay the same... alternative would be to return null
        return null;
    }

    /** Starting from template, go down the subtype tree and ensure that the
     * returned CClassType and the target CClassType have the same underlying class.
     * 
     * @param target The type that determines to which class to raise. 
     * @param template The type from which to go down the subtype tree.
     * @return A type that has the same CClass as target.
     */
    private static CClassType lower(CClassType target, CClassType template) {
        if (target.getCClass().equals(template.getCClass())) // This is also true for arrays. This is okay, as the
        // base types will be propagated later.
        {
            return template;
        }
        

        // what is an efficient way to achieve this?
        // at the moment just propagate the main modifier.
        // Example where more would be good:
        // peer List<rep O> rl = ...;
        // ArrayList<O> o = (ArrayList<O>) rl;
        // we would like the modifiers from rl propagated into the cast.

        // at the moment let's just add the main modifier
        CClassType res = null;
        try {
            res = (CClassType) target.clone();
            res.setCUniverse(template.getCUniverse());
        } catch (CloneNotSupportedException ex) {
            res = null;
        }
        return res;
    }
}
