/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.FastStringBuffer;

/**
 * This class represents a value specializer "type."  It is the base class
 * for a hierarchy representing different kinds of value specializers, e.g.
 * String literals, integer constants, etc.
 */
public abstract class CValueType extends CType {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructor
     * @param	staticType	the associated static type of the formal
     * @param 	dispatchExpr	the AST of the specializer expression
     */
    public CValueType(CType staticType, JExpression dispatchExpr) {
	super(staticType.getTypeID());
	this.staticType = staticType;
	this.dispatchExpr = dispatchExpr;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /* Is this value specializer equal to the given one? */
    public boolean equals(/*@ nullable @*/ Object o) {
	    // we could write this method (and its overriding methods)
	    // much nicer with multimethods!
	if (!(o instanceof CValueType))
	    return false;
	
	CValueType other = (CValueType) o;

	if (!isResolvedAndValid() || !other.isResolvedAndValid())
	    return super.equals(other);

	return staticType.equals(other.staticType) &&
	    this.valueEquals(other);
    }

    public int hashCode() {
	    // if we haven't yet checked the specializer expression,
	    // default to inherited hashCode
	if (!isResolvedAndValid())
	    return super.hashCode();

	return staticType.hashCode() * 31 + this.intValue();
    }
    
    /**
     * Is this type assignable to the given type.  The original
     * dispatch expression must have the right type.  Also, if both
     * types are CValueTypes, then their values must be the same.
     *
     * @param	dest		the destination type
     * @return true iff assignment is valid 
     * @see org.multijava.mjc.JExpression#isAssignableTo(CType) */
    // also modifies \everything; 
    public boolean isAlwaysAssignableTo(CType dest) {
	if (!isResolvedAndValid()) {
	    // not sure if this is the right thing to do in all cases
	    // but it should be conservative
	    return false;
	}
	if (!(dest instanceof CValueType)) {
	    return dispatchValue.isAssignableTo(dest);
	}
	CValueType dt = (CValueType) dest;
	if (!dt.isResolvedAndValid()) {
	    return false;
	}
	return dispatchValue.isAssignableTo(dt.staticType) &&
	    this.valueEquals(dt);
    }

    /**
     * Transforms this type to a string
     */
    public String toString() {
	return dispatchExpr.toString();
    }

    
    /**
     * Transforms this type's value to a string for use in
     * bytecode attributes for this value specializer.
     */
    public String signatureStringValue() {
	return dispatchValue.toString();
    }
    
    
    /**	
     * Transforms this type to a string for bytecode attributes
     */
    public void appendSignature(FastStringBuffer buff) {
	assertTrue(isResolvedAndValid());
	buff.append('X');
	buff.append(signatureStringValue());
	buff.append('"');
    }

    /**
     * Returns the size used in stack by value of this type
     */
    public int getSize() {
	return staticType.getSize();
    }

    /**
     * Check if a type is a numeric type
     * @return is it a numeric type?
     */
    public /*@ pure @*/ boolean isNumeric() {
	return staticType.isNumeric();
    }

    public /*@ pure @*/ boolean isCodedAsInt() {
	return staticType.isCodedAsInt();
    }

    /**
     * Check if a type is an integer type
     * @return is it a integer type ?
     */
    public /*@ pure @*/ boolean isOrdinal() {
	return staticType.isOrdinal();
    }

    /**
     * Check if a type is an integer type
     * @return is it a real number type ?
     */
    public /*@ pure @*/ boolean isFloatingPoint() {
	return staticType.isFloatingPoint();
    }

    /**
     * Check if a type is the boolean type
     * @return is it the boolean type ?
     */
    public /*@ pure @*/ boolean isBoolean() {
	return staticType.isPrimitive();
    }

    
    /**
     * Check if a type is a class type
     * @return is it a subtype of ClassType ?
     */
    public /*@ pure @*/ boolean isPrimitive() {
	return staticType.isPrimitive();
    }

    /**
     * Check if a type is a class type
     * @return is it a subtype of ClassType ?
     */
    public /*@ pure @*/ boolean isReference() {
	return staticType.isReference();
    }


    /**
     * @return is this type representing a value specializer ?
     */
    public /*@ pure @*/ boolean isValueType() {
	return true;
    }

	/**
	 * Returns the specializer symbol to use when this type is the
	 * specializer.
	 */
    public /*@ pure @*/ String specializerSymbol() {
	return "@@";
    }
    
    /**
     * @return the value expression of this type 
     */
    public /*@ pure @*/ JExpression getValue() {
	return dispatchValue;
    }

    
    /**
     * Is the value of this specializer equal to the value of the
     * given value specializer?  Subclasses provide this method,
     * which is used in equals and isAlwaysAssignableTo above.
     *
     * @param	t	the type to compare against
     * @return	true iff the values are equal
     */
    //assignable \nothing;
    protected abstract boolean valueEquals(CValueType t);

    /**
     * Returns an integer representation of the specializer's value,
     * for use in computing its hashCode.
     */
    protected abstract int intValue();

    /*
     * Has this specializer value been resolved, and is it of the
     * appropriate type?
     */
    protected /*@ pure @*/ boolean isResolvedAndValid() {
	return dispatchValue != null && isValid();
    }

    /*
     * Is this value specializer of the appropriate type?
     * @requires dispatchValue != null	
     */
    protected /*@ pure @*/ abstract boolean isValid();

    
    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Register this type as visible, so it can be checked during ITC
     * on the compilation unit
     *
     * @param		context		the context in which this type was
     *					declared, may be null
     * @exception	UnpositionedError	this unpositioned error
     *						should be positioned
     *						by the calling node of
     *						the AST, typically by
     *						calling the
     *						addPosition method of
     *						UnpositionedError 
     */
    public CType checkType(CContextType context)
	throws UnpositionedError {
	context.registerVisibleType(this);
	return this;
    }

    
    /**
     * Compute the value of a specializer expression used to dispatch on
     * a compile-time constant.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    public void resolveValueType(CExpressionContextType context) throws
	PositionedError {
	dispatchValue = dispatchExpr.typecheck( context );
	    // make sure the value is converted to the right type
	if (dispatchValue.isAssignableTo(staticType)) {
	    dispatchValue = dispatchValue.convertType(staticType, context);
	}
    }

        /**
     * Can this type be converted to the specified type by casting conversion (JLS 5.5) ?
     * @param	dest		the destination type
     * @return	true iff the casting conversion is valid
     */
    public boolean isCastableTo(CType dest) {
	return staticType.isCastableTo( dest );
    }

    

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    // the AST representation of the specializer expression
    protected JExpression dispatchExpr;
    // the value of the dispatchExpr (which must be a compile-time constant)
    // this field is computed during typechecking
    protected JExpression dispatchValue;

    protected CType staticType;
}
