/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CVariableInfoTable.java,v 1.7 2004/09/14 19:01:37 cclifton Exp $
 */

package org.multijava.mjc;

/**
 * This class stores information about the set of local variables in a given
 * context.  This information is collected during typechecking. It includes the
 * information necessary to replace the source code's variable identifiers with
 * references to the appropriately numbered positions in a bytecode method's
 * frame (JVM2, 3.6.1).  The information also includes whether the variable has
 * been assigned to (JLS2 16).
 *
 * @see CFlowControlContextType
 */
public final class CVariableInfoTable implements Cloneable {

	// -------------------------------------------------------------
	// CONSTRUCTORS
	// -------------------------------------------------------------

	public CVariableInfoTable() {
		states = new CVariableState[DEFAULT_SIZE];
	}

	private CVariableInfoTable(CVariableState[] states) {
		this.states = states;
	}

	public Object clone() {
		CVariableState[] newStates = new CVariableState[states.length];
		for (int i = 0; i < states.length; i++) {
			if (states[i] != null) {
				newStates[i] = states[i].duplicate();
			}
		}

		return new CVariableInfoTable(newStates);
	}

	public String toString() {
		StringBuffer result = new StringBuffer(states.length * 2 + 2);
		result.append("[");
		for (int i = 0; i < states.length; i++) {
			CVariableState state = states[i];
			if (state == null) {
				result.append("-");
			} else {
				result.append(state.toString());
			}
			if (i < states.length - 1) {
				result.append(",");
			}
		}
		result.append("]");
		return result.toString();
	}

	public CVariableInfoTable duplicate() {
		return (CVariableInfoTable) clone();
	}

	// ----------------------------------------------------------------------
	// ACCESSORS
	// ----------------------------------------------------------------------

	public void initialize(VariableDescriptor varDesc) {
		CVariableState state = getState(varDesc.index());
		state.initialize();
		state.setDescriptor(varDesc);
	}

	/**
	 * Returns the state of the variable at the given position in the local frame.
	 *
	 * @return an alias to the variable state object stored in the given position
	 *         of this table
	 */
	public CVariableState getState(int pos) {
		if (pos >= states.length) {
			expandArrayToInclude(pos);
		}

		//@ assert pos < states.length;

		if (states[pos] == null) {
			CVariableState result = new CVariableState();
			states[pos] = result;
			return result;
		}

		return states[pos];
	}

	/**
	 * Merges the given state information with the stored information in the given
	 * position.
	 */
	public void mergeState(int pos, CVariableState state) {
		//  !FIXME!10/26 this method should eventually be private
		if (pos >= states.length) {
			expandArrayToInclude(pos);
		}
		//@ assert pos < states.length;

		if (states[pos] == null) {
			if (state == null) {
				// both states are null so nothing to do
				return;
			}
			states[pos] = new CVariableState();
		}
		states[pos].mergeWith(state);
	}

	private void tryFinallyMergeState(int pos, CVariableState tryVariableState) {
		// FIXME: implement method body
		//@ assert pos < states.length;
		if (states[pos] == null) {
			if (tryVariableState == null) {
				return;
			}
			states[pos] = new CVariableState();
		}
		states[pos].tryFinallyMergeWith(tryVariableState);
	}



	/**
	 * Merges the state information from other into this for positions from 0 to
	 * <code>numberOfPos - 1</code>.
	 */
	public void mergeUpTo(int numberOfPos, CVariableInfoTable other) {
		if (numberOfPos > states.length) {
			expandArrayToInclude(numberOfPos - 1);
		}

		CVariableState[] otherStates = other.states;
		int otherStatesLength = otherStates.length;
		for (int i = 0; i < numberOfPos; i++) {
			mergeState(i, (i >= otherStatesLength) ? null : otherStates[i]);
		}
	}

	/**
	 * Merges the state information from other into this for positions from 0 to
	 * <code>numberOfPos - 1</code> using the function for merging a finally block
	 * (this) with a try block.
	 * @param numberOfPos number of positions that should be merged
	 * @param tryInfoTable the variable state information for the try block
	 */
	public void tryFinallyMergeUpTo(int numberOfPos,
									CVariableInfoTable tryInfoTable) {
		if (numberOfPos > states.length) {
			expandArrayToInclude(numberOfPos - 1);
		}

		CVariableState[] tryStates = tryInfoTable.states;
		for (int i = 0; i < numberOfPos; i++) {
			tryFinallyMergeState(i, (i >= tryStates.length) ? null : tryStates[i]);
		}
	}

	/**
	 * Replaces the state information for the variable at the given position with
	 * the given state information.
	 */
	public void replaceState(int pos, CVariableState state) {
		//  !FIXME!10/26 this should eventually be vestigial
		if (pos >= states.length) {
			expandArrayToInclude(pos);
		}
		//@ assert pos < states.length;

		states[pos] = state;
	}

	/**
	 * Replaces the state information of this with information from other for
	 * positions from 0 to <code>numberOfPos - 1</code>.
	 */
	public void replaceUpTo(int numberOfPos, CVariableInfoTable other) {
		if (numberOfPos > states.length) {
			expandArrayToInclude(numberOfPos - 1);
		}
		CVariableState[] otherStates = other.states;
		int otherStatesLength = otherStates.length;
		for (int i = 0; i < numberOfPos; i++) {
			// this implementation uses knowledge of the
			// implementation to avoid instantiating CVariableState
			// objects if not necessary
			states[i] = (i >= otherStatesLength) ? null : otherStates[i];
		}
	}

	/**
	 * Expands the states array to hold state information for at least
	 * <code>pos</code> variables.
	 *
	 * @param pos the highest position for variable that this table needs to
	 *            accomodate (so far)
	 *            <p/>
	 *            <pre><jml>
	 *            requires pos >= states.length;
	 *            ensures states.length > pos;
	 *            ensures (\forall int i; 0<=i && i<\old(states.length);
	 *            			states[i] == \old(states[i]) );
	 *            </jml></pre>
	 */
	private void expandArrayToInclude(int pos) {
		CVariableState[] newStates = new CVariableState[pos + 1];
		System.arraycopy(states, 0, newStates, 0, states.length);
		states = newStates;
	}

	// ----------------------------------------------------------------------
	// PRIVATE DATA MEMBERS
	// ----------------------------------------------------------------------

	/**
	 * The default size of the array constructed to hold local variable
	 * information.  Corresponds roughly to the number of local variable
	 * declarations expected in a given lexical contour.
	 */
	private static final int DEFAULT_SIZE = 10;

	/**
	 * The states of the variables represented by this information table.  Note
	 * that for efficiency to following invariant holds.
	 * <p/>
	 * <pre><jml>
	 * private invariant states != null &&
	 * 		(\forall int i; 0 <= i && i < states.length;
	 * 			states[i]==null ==>
	 *                   (* the corresponding variable is definitely not
	 *                      assigned *));
	 * </jml></pre>
	 */
	private CVariableState[] states;
}
