/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CVariableState.java,v 1.16 2005/12/23 18:14:20 chalin Exp $
 */

package org.multijava.mjc;

/**
 * The class stores information about the definitely assigned state of a
 * variable or field.  See JLS2, chapter 16.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.16 $
 */

public class CVariableState implements Cloneable {

	//---------------------------------------------------------------------
	// CONSTRUCTORS
	//---------------------------------------------------------------------

	/**
	 * Create a variable state corresponding to a newly allocated variable.
	 */
	public CVariableState() {
	}

	public Object clone() {
		try {
			return super.clone(); // shallow copy is OK
		} catch (CloneNotSupportedException e) {
			//@ unreachable;
			return null;
		} // end of try-catch
	}

	public String toString() {
		if (isDefinitelyAssigned) {
			return "A";
		} else if (maybeAssigned) {
			return "M";
		} else {
			return "U";
		}
	}

	/**
	 * Returns a duplicate of this.  Utility method allows clients to avoid
	 * typecast when cloning this.
	 * <p/>
	 * <pre><jml>
	 * ensures \fresh(\result) && \result.equals(this);
	 * </jml></pre>
	 */
	public CVariableState duplicate() {
		return (CVariableState) clone();
	}

	//---------------------------------------------------------------------
	// ACCESSORS
	//---------------------------------------------------------------------

	/**
	 * Returns true if the variable represented by this is definitely initialized
	 * (i.e., is definitely assigned according to JLS2, 16).
	 */
	public /*@ pure @*/ boolean isDefinitelyAssigned() {
		return isDefinitelyAssigned;
	}

	/**
	 * Returns true if the variable represented by this is definitely not
	 * initialized (i.e., is definitely unassigned according to JLS2, 16).
	 */
	public /*@ pure @*/ boolean isDefinitelyUnassigned() {
		return !maybeAssigned;
	}


	/**
	 * Mutates the state of this to indicate that the corresponding variable has
	 * been initialized in the local control flow.
	 */
	public void initialize() {
		maybeAssigned = true;
		isDefinitelyAssigned = true;
	}

	/**
	 * Mutates the state of this to indicate the result of merging with the given
	 * variable state.
	 * 
	 * <pre><jml>
	 * {|
	 *   requires other == null;
	 *   modifiable objectState;
	 *   ensures (* result is equivalent to the case where other != null &&
	 *              other.isDefinitelyUnassigned() *);
	 *   ensures \result == this;
	 * |}
	 * also
	 * {|
	 *   requires other != null;
	 *   modifiable objectState;
	 *   ensures (* mergeWith is symmetric *);
	 * 
	 *   ensures \old(this.isDefinitelyAssigned()) &&
	 * 		other.isDefinitelyAssigned()
	 *   		==> \result.isDefinitelyAssigned();
	 * 
	 *   ensures \old(this.isDefinitelyAssigned()) &&
	 * 		!other.isDefinitelyAssigned()
	 *   	 	==> !\result.isDefinitelyUnassigned() &&
	 * 				!\result.isDefinitelyAssigned();
	 * 
	 *   ensures \old(this.isDefinitelyUnassigned()) &&
	 * 		other.isDefinitelyUnassigned()
	 * 			==> \result.isDefinitelyUnassigned();
	 * 
	 *   ensures \old(!this.isDefinitelyAssigned() &&
	 * 			!this.isDefinitelyUnassigned()) &&
	 * 		!other.isDefinitelyAssigned()
	 *   	 	==> !\result.isDefinitelyUnassigned() &&
	 * 				!\result.isDefinitelyAssigned();
	 *   ensures \result == this;
	 * |}
	 * </jml></pre>
	 */
	public CVariableState mergeWith(CVariableState other) {
		if (other == null) {
			isDefinitelyAssigned = false;
			maybeAssigned = isDefinitelyAssigned || maybeAssigned;
		} else {
			isDefinitelyAssigned =
					isDefinitelyAssigned && other.isDefinitelyAssigned;
			maybeAssigned = isDefinitelyAssigned || maybeAssigned ||
							other.maybeAssigned;
			// Adopts variable descriptor of other if ours is null
			if (variableDesc == null) {
				variableDesc = other.variableDesc;
			}

		}
		return this;
	}

	/**
	 * Sets the definite assignment state of this using the special merge logic
	 * appropriate for combining the definite assignment information from a
	 * finally block (as represented by this) with the definite assignment
	 * information from a try-block.
	 * @param tryVariableState definite assignment state from a try block
	 */
	public void tryFinallyMergeWith(CVariableState tryVariableState) {
		if (tryVariableState != null && tryVariableState.isDefinitelyAssigned()) {
			initialize();
		}
		// else just use the state from the finally block, i.e., do nothing
	}

	public /*@ pure @*/ boolean equals(/*@ nullable @*/ Object other) {
		try {
			CVariableState oState = (CVariableState) other;
			return maybeAssigned == oState.maybeAssigned &&
				   isDefinitelyAssigned == oState.isDefinitelyAssigned;
		} catch (ClassCastException e) {
			return false;
		} // end of try-catch
	}

	// doc comment and specification inherited
	public int hashCode() {
		return 2 * (maybeAssigned ? 1 : 0) + (isDefinitelyAssigned ? 1 : 0);
	}

	/**
	 * Sets the descriptor of the slot whose state is tracked by this.
	 */
	public void setDescriptor(VariableDescriptor varDesc) {
		this.variableDesc = varDesc;
	}

	/**
	 * Returns true if the slot tracked by this is declared final. Result is only
	 * valid if maybeAssigned.
	 */
	public boolean isFinal() {
		return variableDesc != null && variableDesc.isFinal();
	}

	/**
	 * Returns the (unqualified) identifier of the slot tracked by this. Result is
	 * only valid if maybeAssigned.
	 */
	public String ident() {
		if (variableDesc == null) {
			return "<unknown>";
		} else {
			return variableDesc.ident();
		}
	}

	//---------------------------------------------------------------------
	// PRIVILEGED DATA
	//---------------------------------------------------------------------

	//@ private invariant isDefinitelyAssigned ==> maybeAssigned;
	//@ private invariant isDefinitelyAssigned ==> isDefinitelyAssigned();
	//@ private invariant maybeAssigned ==> !isDefinitelyUnassigned();
	//@ public invariant isDefinitelyAssigned() ==> !isDefinitelyUnassigned();
	//@ public invariant isDefinitelyUnassigned() ==> !isDefinitelyAssigned();

	/**
	 * Indicates that the variable represented by this may have been initialized.
	 */
	private boolean maybeAssigned = false;
	//@                 in objectState;

	/**
	 * Indicates that the variable represented by this definitely has been
	 * initialized.
	 */
	private boolean isDefinitelyAssigned = false;
	//@                 in objectState;

	/**
	 * A descriptor for the field or variable (i.e., slot) tracked by this state.
	 * Descriptor is only initialized if the slot may be initialized.
	 */
	private VariableDescriptor variableDesc;
	//@                        in objectState;
}
