/*
 * Copyright (C) 2006 Swiss Federal Institute of Technology Zurich
 *
 * Part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2006 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CWildcardType.java,v 1.5 2008/03/17 16:59:46 wdietl Exp $
 */
package org.multijava.mjc;

import org.multijava.util.compiler.FastStringBuffer;
import org.multijava.util.compiler.UnpositionedError;


/**
 * Type representing a wildcard (JLS3 4.5.1).
 *
 * The semantics of a wildcard in the generic Java type system are fully
 * determined by its optional upper/lower bound and by the formal type variable
 * (JLS3 4.4) for which the wildcard is a type argument (JLS3 4.5.1).
 */
public class CWildcardType extends CClassType {

    /**
     * Indicates whether the wildcard has been explicitely <i>or</i> implicitely
     * upper bounded.
     */
    private boolean isUpperBounded = true;

    /**
     * The (upper or lower) bound of the wildcard. This is <code>null</code>
     * if and only if the wildcard was declared without an explicit bound.
     */
    private /*@ nullable @*/ CClassType bound = null;

    /** The formal type variable for which this wildcard is a type argument. */
    private /*@ nullable @*/ CTypeVariable typeVariable = null;

    /**
     * Constructor for creating an eventually bounded wildcard.
     *
     * As not all combinations of parameters are meaningful, this constructor
     * is declared to be <code>private</code>. That way, client code must use
     * the provided <code>static</code> factory methods which indirectly invoke
     * this constructor with the correct parameters.
     *
     * @param isUpperBounded  Whether the wildcard is upper bounded (explicitly
     *                        or implicitly) or not (i.e. lower bounded).
     * @param bound           The upper/lower bound (according to the flag
     *                        <code>isUpperBounded</code>) of the wildcard or
     *                        <code>null</code> if the wildcard is not bounded.
     *
     * @see #createWildcard()
     * @see #createUpperBoundedWildcard(CClassType)
     * @see #createLowerBoundedWildcard(CClassType)
     */
    private CWildcardType(boolean isUpperBounded, /*@ nullable @*/ CClassType bound) {
        super();
        this.isUpperBounded = isUpperBounded;
        this.bound = bound;
        this.bound = (CClassType)CUniverseServices.setDefaultReadonly(bound);
    }

    /**
     * Creates a wildcard implicitely upper bounded by the
     * <code>java.lang.Object</code> class.
     *
     * @return  An implicitely upper bounded wildcard.
     */
    public static /*@ non_null @*/ CWildcardType createWildcard() {
        return new CWildcardType(true, null);
    }

    /**
     * Creates an upper bounded wildcard with the given <code>upperBound</code>.
     *
     * @param upperBound  The upper bound of the wildcard to create.
     * @return            A wildcard having the given <code>upperBound</code>.
     */
    public static /*@ non_null @*/ CWildcardType createUpperBoundedWildcard(
            /*@ non_null @*/ CClassType upperBound) {
        // Check for the correctness of the provided arguments.
        assertTrue(upperBound != null,
                   "The upper bound of the wildcard may not be null at this point.");

        return new CWildcardType(true, upperBound);
    }

    /**
     * Creates a lower bounded wildcard with the given <code>lowerBound</code>.
     *
     * @param lowerBound  The lower bound of the wildcard to create.
     * @return            A wildcard having the given <code>lowerBound</code>.
     */
    public static /*@ non_null @*/ CWildcardType createLowerBoundedWildcard(
            /*@ non_null @*/ CClassType lowerBound) {
        // Check for the correctness of the provided arguments.
        assertTrue(lowerBound != null,
                   "The lower bound of the wildcard may not be null at this point.");

        return new CWildcardType(false, lowerBound);
    }

    public /*@ pure @*/ boolean isWildcard() {
        return true;
    }

    /**
     * Returns whether an explicit bound has been specified in the declaration
     * of this wildcard.
     *
     * @return  Whether this wildcard is explicitely bounded.
     *
     * @see #createWildcard()
     * @see #createUpperBoundedWildcard(CClassType)
     * @see #createLowerBoundedWildcard(CClassType)
     */
    public /*@ pure @*/ boolean isBounded() {
        return getBound() != null;
    }

    /**
     * Returns whether this wildcard type has an explicit <i>or</i> implicit
     * upper bound.
     *
     * @return  Whether this wildcard is explicitely <i>or</i> implicitely
     *          upper bounded.
     */
    public /*@ pure @*/ boolean isUpperBounded() {
        return isUpperBounded;
    }

    /**
     * Returns the (upper or lower) bound of this wildcard which is
     * <code>null</code> if and only if the wildcard was declared without an
     * explicit bound.
     *
     * @return  The bound of this wildcard.
     */
    public /*@ pure @*/ /*@ nullable @*/ CClassType getBound() {
        return bound;
    }

    /**
     * Update the bound of the wildcard.
     * Added by WMD to allow viewpoint adaptation.
     * 
     * @param b the new bound
     */
    public void setBound( CClassType b ) {
    	bound = b;
    }
    
    
    /**
     * Convenience method returning either the bound explicitely provided in the
     * wildcard declaration, if any, or, otherwise, the implicit bound
     * <code>java.lang.Object</code>.
     *
     * @return  The explicit or implicit bound of this wildcard.
     *          Is never <code>null</code>.
     *
     * @see #getBound()
     * @see #isBounded()
     */
    public /*@ pure @*/ /*@ non_null @*/ CClassType getExplicitOrImplicitBound() {
    	return isBounded() ? getBound() : CStdType.WildcardBoundObject;
    	
    	/*
    	if( typeVariable != null ) {
    		CClassType[] tvb = typeVariable.getBounds();
    		if( tvb != null && tvb.length > 0 )
    			return tvb[0];
    	}*/
    	
        // return isBounded() ? getBound() : CStdType.Object;
    }

	/**
	 * If this wildcard contains <code>type</code>
	 *
	 * @param type The type to check if it is contained
	 * @param considerUniverses igonred
	 * @param allowUniversesCovariance ignored
	 * @return true iff this wildcard contains the type
	 */
    public boolean contains(/*@ non_null @*/ CClassType type,
    		boolean considerUniverses,
    		boolean allowUniversesCovariance) {
        type = type.capture();
        if (isUpperBounded()) {
            // An upper bounded wildcard contains another type iff we have
            // covariant subtyping between its upper bound and that type.
            return type.isAlwaysAssignableTo(getExplicitOrImplicitBound());
        } else {
            // A lower bounded wildcard contains another type iff we have
            // contravariant subtyping between its lower bound and that type.
            return getExplicitOrImplicitBound().isAlwaysAssignableTo(type);
        }
    }

    /**
     * This wildcard is assignable to the given <code>type</code> if and only
     * if its capture type is assignable to it.
     *
     * @see #capture()
     */
    public boolean isAlwaysAssignableTo(CType dest) {
        return capture().isAlwaysAssignableTo(dest);
    }

    /**
     * This wildcard is assignable to the given <code>type</code> if and only
     * if its capture type is assignable to it.
     *
     * @see #capture()
     */
    public boolean isAlwaysAssignableTo(CType dest, boolean inst) {
        return capture().isAlwaysAssignableTo(dest, inst);
    }

    /**
     * This wildcard is assignable to the given <code>type</code> if and only
     * if its capture type is assignable to it.
     *
     * @see #capture()
     */
    public boolean isAlwaysAssignableTo(CType dest, CClassType[] substitution) {
        return capture().isAlwaysAssignableTo(dest, substitution);
    }

    /**
     * This wildcard is assignable to the given <code>type</code> if and only
     * if its capture type is assignable to it.
     *
     * @see #capture()
     */
    public boolean isAlwaysAssignableToIgnoreUniverses(CType dest, CClassType[] substitution) {
        return capture().isAlwaysAssignableToIgnoreUniverses(dest, substitution);
    }

	/**
	 * Checks if this wildcard is a valid argument for type variable
	 * <code>typeVariable</code>.
	 *
	 * @param typeVariable The type variable
	 * @param substitution The type arguments for substitution
	 * @param type The type <code>typeVariable</code> belongs to
	 * @return true iff this wildcard is a valid type argument for
	 *		   <code>typeVariable</code>.
	 */
    public boolean isValidTypeArgumentFor(
            CTypeVariable typeVariable,
            CClassType[] substitution,
            CClassNameType type) {
        CClassType wildcardBound = getExplicitOrImplicitBound();

        // For lower bounded wildcards, the lower bound itself must be a valid
        // type argument for the type variable.
        if (!isUpperBounded()) {
            return wildcardBound.isValidTypeArgumentFor(typeVariable,
            		substitution,
            		type);
        }

        // For upper bounded wildcards, the upper bound must be castable to all
        // types in the bounds of the type variable.
        CClassType[] typeVariableBounds = typeVariable.getBounds();
        for (int i = 0; i < typeVariableBounds.length; i++) {
            if (!wildcardBound.isCastableTo(typeVariableBounds[i])) {
                return false;
            }
        }
        return true;
    }

    public boolean equals(/*@ nullable @*/ Object other) {
        return equals(other, CTopLevel.getCompiler().universeChecks(), false);
    }

    public boolean equals(/*@ nullable @*/ Object other, boolean enableUniv,
    		boolean allowUniversesCovariance) {
        if (!(other instanceof CWildcardType)) {
            return false;
        }

        // If the two wildcards don't have the same kind of bound, they cannot
        // be equal.
        CWildcardType wildcardType = (CWildcardType) other;
        if ((isBounded() != wildcardType.isBounded())
            || (isUpperBounded() != wildcardType.isUpperBounded())) {
            return false;
        }

        // We have the same kind of bound, so check whether the bounds are equal.
        return getExplicitOrImplicitBound()
                .equals(wildcardType.getExplicitOrImplicitBound(), enableUniv, allowUniversesCovariance);
    }

    public boolean equals(
            /*@ nullable @*/ CType other,
            /*@ non_null @*/ CClassType[] substitution) {
        return equals(other, substitution, CTopLevel.getCompiler().universeChecks());
    }

    public boolean equals(
            /*@ nullable @*/ CType other,
            /*@ non_null @*/ CClassType[] substitution,
            boolean enableUniv) {
        if (!(other instanceof CWildcardType)) {
            return false;
        }

        // If the two wildcards don't have the same kind of bound, they cannot
        // be equal.
        CWildcardType wildcardType = (CWildcardType) other;
        if ((isBounded() != wildcardType.isBounded())
            || (isUpperBounded() != wildcardType.isUpperBounded())) {
            return false;
        }

        // We have the same kind of bound, so check whether the bounds are equal.
        return getExplicitOrImplicitBound()
                .equals(wildcardType.getExplicitOrImplicitBound(), substitution, enableUniv);
    }

    /**
     * Returns the formal type variable for which this wildcard is a type
     * argument.
     *
     * @return  The type variable for which this wildcard is a type argument.
     */
    public /*@ pure @*/ CTypeVariable getTypeVariable() {
        return typeVariable;
    }

    /**
     * Sets the formal type variable for which this wildcard is a type argument.
     *
     * @param typeVariable  The type variable for which this wildcard is a type
     *                      argument.
     */
    public void setTypeVariable(CTypeVariable typeVariable) {
        this.typeVariable = typeVariable;
    }

    /**
     * During type checking, the wildcard's optional bound is type checked, too.
     * Note that, a non-null type variable must have been set on this wildcard
     * via <code>setTypeVariable(CTypeVariable)</code> before invoking this
     * method. Furthermore, this type variable must already have been type
     * checked to be able to query its <code>CClass</code>.
     */
    //@ also requires (getTypeVariable() != null) && (getTypeVariable().isChecked());
    public CType checkType(CContextType context) throws UnpositionedError {
        // Explicitely check the preconditions of this method.
        assertTrue(getTypeVariable() != null,
                   "The formal type variable of this wildcard must be set "
                   + "before querying its capture type.");

        // Type check the bound, if any.
        if (bound != null) {
            bound = (CClassType) bound.checkType(context);
            bound = (CClassType)CUniverseServices.setDefaultReadonlyAgain(bound);
        }

        // The class of a wildcard should be set to the least specific class
        // which can be captured by the wildcard.
        setClass(capture().getCClass());

        return this;
    }

    /**
     * The capture type depends on the optional bound of the wildcard as well
     * as on the formal type variable for which this wildcard is a type
     * argument. Therefore, a non-null type variable must have been set on this
     * wildcard via <code>setTypeVariable(CTypeVariable)</code> before invoking
     * this method.
     *
     * @return  The capture type of this wildcard.
     *          Is never <code>null</code>.
     *
     * @see CCaptureType
     * @see #getTypeVariable()
     * @see #setTypeVariable(CTypeVariable)
     */
    //@ also requires getTypeVariable() != null;
    public /*@ non_null @*/ CClassType capture() {
        // Explicitely check the preconditions of this method.
        assertTrue(getTypeVariable() != null,
                   "The formal type variable of a wildcard must be set "
                   + "before querying its capture type.");

        return CCaptureType.create(this);
    }

    /**
     * Note that a non-generic signature should never be used for a wildcard
     * since a wildcard is always a generic type argument. Therefore, an
     * internal compiler error is triggered upon invoking this method in order
     * to immediately indicate the misuse of this method.
     *
     * @see #appendGenericSignature(FastStringBuffer)
     */
    public void appendSignature(FastStringBuffer buff) {
        fail("A non-generic signature of a wildcard should never be queried.");
    }

    public /*@ pure @*/ void appendGenericSignature(FastStringBuffer buffer) {
        if (getBound() == null) {
            buffer.append('*');
        } else {
            buffer.append(isUpperBounded() ? '+' : '-');
            getBound().appendGenericSignature(buffer);
        }
    }

    /**
     * Returns the erasure of this wildcard's capture type.
     *
     * @see #capture()
     */
    public CType getErasure() {
        return capture().getErasure();
    }

    public String toString() {
        StringBuffer buffer = new StringBuffer("?");
        if (getBound() != null) {
            buffer.append(isUpperBounded() ? " extends " : " super ");
            buffer.append(getBound().toString());
        }
        return buffer.toString();
    }

    public String toVerboseString() {
        StringBuffer buffer = new StringBuffer("?");
        if (getBound() != null) {
            buffer.append(isUpperBounded() ? " extends " : " super ");
            buffer.append(getBound().toVerboseString());
        }
        return buffer.toString();
    }

}
