/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CWrapperMethod.java,v 1.12 2005/04/26 02:39:32 kui_dai Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.classfile.LocalVarInstruction;
import org.multijava.util.classfile.NoArgInstruction;

/**
 * This class represents a wrapper method for a method that an inner
 * or enclosing class has privileged access to.  The wrapper method is
 * necessary to work around the Java verifier, which doesn't know that
 * the inner/enclosing class has privileged access.  */

public class CWrapperMethod extends CSourceMethod {
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Constructs a wrapper method for the given method.
     * @param	target	the method to wrap
     * @param	owner	the receiver of the wrapper method
     * @param   isSuper true ==> wrapper is for a super method invocation
     */
    public CWrapperMethod( CMethod target, CClass owner, boolean isSuper ) {
	super( new MemberAccess(owner, owner,
		   /* !FIXME! Wrapper methods in javac are always static,
		      allowing non-virtual dispatch.  We may want to do that
		      for performance reasons at some point, but it's easier
		      not to, so that the caller doesn't have to worry about
		      the fact that it's calling a wrapper method.
		   */
				target.isStatic() ? ACC_STATIC : 0L),
	       CClass.getNextAccessName(), target.returnType(), 
	       CSpecializedType.
	       unspecializedTypeTupleFrom( target.parameters() ),  
	       target.throwables(), CTypeVariable.EMPTY,target.isDeprecated(),
		   // fake a body so that code will be generated
		   // for this method
	       JBlock.DUM_BLOCK,
	       null, null );
	this.target = target;
	this.isSuper = isSuper;
	this.topConcreteMethod = this;
    }

    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    // check whether this wrapper method can be re-used for the
    // given method

    /**
     * Checks whether this wrapper method can be re-used for the given
     * method.
     *
     * @param m		the method to be wrapped
     * @param mIsSuper	whether a super method wrapper is to be created for 
     *			<code>m</code>
     */
    public boolean isApplicableTo(CMethod m, boolean mIsSuper) 
	throws UnpositionedError 
    {
	return target.topConcreteMethod().equals( m.topConcreteMethod() ) && 
	    (mIsSuper == isSuper);
    }

    // ----------------------------------------------------------------------
    // GENERATE CLASSFILE INFO
    // ----------------------------------------------------------------------

    /**
     * Plants the bytecode for the body of this method.  Separated
     * from <code>genCode()</code> so subclasses may replace standard
     * body bytecode with specialized code.  
     *
     * @see #genCode()
     */
    protected void plantBodyBytecode( CodeSequence code ) {
	int offset = 0;
	if ( !isStatic() ) {
		// put the receiver on the stack
	    code.plantLoadThis();
	    offset = 1;
	}
	    // put the args on the stack
	for (int i = 0; i < parameters().length; i++) {
	    code.plantInstruction(new LocalVarInstruction(
		parameters()[i].staticType().getLoadOpcode(), i+offset )
		);
	}
	    // call the wrapped method
	target.genGenFuncInvocation( code, isSuper );
	    // return
	code.plantInstruction(
	    new NoArgInstruction( returnType().getReturnOpcode() ) );
    }
    

    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------

    /**
     * The target method of this wrapper. */
    private CMethod target;


    /**
     * Indicates whether this wrapper is for a super method
     * invocation.  */
    private boolean isSuper;
}
