/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CodeSequence.java,v 1.12 2004/01/22 20:04:41 leavens Exp $
 */

/**
 * Code sequence is used as a bag to hold lists of instructions
 * until it is time to put them into a class file.
 *
 * Note: Local variables are currently not used
 */
package org.multijava.mjc;

import java.util.Stack;
import java.util.ArrayList;
import org.multijava.util.classfile.*;

public final class CodeSequence
    extends org.multijava.util.Utils
    implements org.multijava.util.classfile.Constants
{
    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Constructs a code attribute.
     */
    private CodeSequence() {
	instructions = 
	    new Instruction[org.multijava.util.classfile.Constants.MAX_CODE_PER_METHOD];
	handlers = new ArrayList();
	lines = new ArrayList();
	/* LOCAL VARIABLES NOT USED
	   locals = new Hashtable();
	   LOCAL VARIABLES NOT USED */
    }
    
    /**
     * Constructs a code sequence.
     */
    public static CodeSequence getCodeSequence() {
	CodeSequence	seq;
	if (!Constants.ENV_USE_CACHE || stack.empty()) {
	    seq = new CodeSequence();
	} else {
	    seq = (CodeSequence)stack.pop();
	}
	seq.pc = 0;
	seq.discardValue = false;
	seq.labelAtEnd = false;
	seq.lineNumber = 0;
	seq.lastLine = -1;
	return seq;
    }
    
    /**
     * Release a code sequence
     */
    public void release() {
	if (Constants.ENV_USE_CACHE) {
	    stack.push(this);
	    handlers.clear();
	    lines.clear();
	    /* LOCAL VARIABLES NOT USED
	       seq.locals.clear();
	       LOCAL VARIABLES NOT USED */      
	}
    }
    
    public static void initSession() {
	while (!stack.empty()) {
	    stack.pop();
	}
    }
    
    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------
    
    /**
     * Adds an instruction to the code of the current method.
     *
     * @param	insn		the instruction to append
     */
    public final void plantInstruction(Instruction insn) {
	instructions[pc++] = insn;
	
	discardValue = false;
	labelAtEnd = false;
	if (lineNumber != lastLine) {
	    lastLine = lineNumber;
	    lines.add(new LineNumberInfo((short)lineNumber, insn));
	}
    }
    
    /**
     * Adds a local var instruction to the code of the current method.
     *
     */
    public final void plantLocalVar(int opcode, JLocalVariable var) {
	LocalVarInstruction insn = new LocalVarInstruction(opcode,
							   var.getPosition());
	plantInstruction(insn);
	addLocalVarInfo(insn, var);
    }
    
    /**
     * Adds a load of this (local var 0)
     *
     */
    public final void plantLoadThis() {
	plantInstruction(new LocalVarInstruction(opc_aload, 0));
    }
    
    /**
     * Adds an instruction to the code of the current method.
     *
     */
    public final void plantLabelRef(int opcode, CodeLabel label) {
	plantInstruction(new JumpInstruction(opcode, label));
    }
    
    /**
     * Adds an instruction to the code of the current method.
     *
     */
    public final void plantLabel(CodeLabel label) {
	label.setAddress(pc);
	labelAtEnd = true;
    }
    
    /**
     * Plants a swap instruction on the end of this code sequence.
     */
    public final void plantSwap() {
	plantInstruction(new NoArgInstruction(opc_swap));
    }

    // --------------------------------------------------------------------
    // ENVIRONMENT
    // --------------------------------------------------------------------
    
    /**
     * @param	lineNumber		the current line number in source code
     */
    public final void setLineNumber(int lineNumber) {
	if (lineNumber != 0) {
	    this.lineNumber = lineNumber;
	}
    }
    
    /**
     * @return	an array of line number information
     */
    public final LineNumberInfo[] getLineNumbers() {
	LineNumberInfo[]	info = new LineNumberInfo[lines.size()];
	lines.toArray(info);
	return info;
    }
    
    /**
     * @return	an array of local vars information
     */
    public final LocalVariableInfo[] getLocalVariableInfos() {
	return null;
    }
    
    /**
     * @param	discardValue		set if the value of subexpression
     *					should be discarded
     */
    public final void setDiscardValue(boolean discardValue) {
	this.discardValue = discardValue;
    }
    
    /**
     * @return	true if super expression use value of sub expression
     */
    public final /*@ pure @*/ boolean discardValue() {
	return discardValue;
    }
    
    /**
     * Ask the code handler to generate the necessary code to call every
     * finally clause of all try statements
     */
    public final void plantReturn(JReturnStatement ret) {
	for (int i = contexts.size() - 1; i >= 0; i--) {
	    JStatement stmt = (JStatement)contexts.elementAt(i);
	    
	    if (stmt instanceof JTryFinallyStatement) {
		((JTryFinallyStatement)stmt).genFinallyCall(this, ret);
	    } else if (stmt instanceof JSynchronizedStatement) {
		((JSynchronizedStatement)stmt).genMonitorExit(this);
	    }
	}
    }
    
    /**
     * Ask the code handler to generate the necessary code to call every
     * finally and monitorexit
     */
    public final void plantBreak(JStatement top) {
	for (int i = contexts.size() - 1; 
	     i >= 0 && contexts.elementAt(i) != top; i--) {
	    JStatement stmt = (JStatement)contexts.elementAt(i);
	    
	    if (stmt instanceof JTryFinallyStatement) {
		((JTryFinallyStatement)stmt).genFinallyCall(this, null);
	    } else if (stmt instanceof JSynchronizedStatement) {
		((JSynchronizedStatement)stmt).genMonitorExit(this);
	    }
	}
    }
    

    /**
     * Ensures that the code block ends with a no-argument return
     * statement.  This method should be called as the last step of
     * code generation for the body of a method with a return type of
     * void.
     */
    //@ requires contexts.size() == 0;
    public final void ensureFinalReturn() {
	if (pc == 0 || labelAtEnd || instructions[pc-1].getOpcode() != opc_return) {
	    plantInstruction(new NoArgInstruction(opc_return));
	}
    }

    // --------------------------------------------------------------------
    // PUSH CONTEXT
    // --------------------------------------------------------------------
    
    /**
     * Informs the code handlers that we begin a portion of breakable code.
     */
    public final void pushContext(JStatement stmt) {
	contexts.push(stmt);
    }
    
    /**
     * Informs the code handlers that we exit a breakable code.
     * Checks that contexts match.
     */
    public final void popContext(JStatement stmt) {
	assertTrue((JStatement)contexts.pop() == stmt);
    }
    
    
    // --------------------------------------------------------------------
    // EXCEPTIONS
    // --------------------------------------------------------------------
    
    /**
     * Add an exception handler to this code
     */
    public final void addExceptionHandler(HandlerInfo info) {
	handlers.add(info);
    }
    
    /**
     * Returns an array of all exception handler
     */
    public final HandlerInfo[] getHandlers() {
	HandlerInfo[]	info = new HandlerInfo[handlers.size()];
	handlers.toArray(info);
	return info;
    }
    
    // --------------------------------------------------------------------
    // PC
    // --------------------------------------------------------------------
    
    /**
     * Gets the location in code sequence
     */
    public final int getPC() {
	return pc;
    }
    
    /**
     * Returns the actual size of code (number of instruction)
     */
    public final int size() {
	return pc;
    }
    
    /**
     * Returns the instruction at a given position
     */
    public final Instruction getInstructionAt(int pc) {
	return instructions[pc];
    }
    
    /**
     * Return the instruction as a list
     *
     * WARNING: AFTER a call to release() this array will be reused
     */
    public Instruction[] getInstructionArray() {
	// we resolve the labels here, since the array will be reused
	resolveLabels();
	
	Instruction[]	code = new Instruction[pc];
	System.arraycopy(instructions, 0, code, 0, pc);
	return code;
    }
    
    private void resolveLabels() {
	// if there is a label planted as last instruction, add a dummy
	// instruction at the end: it will never be reached
	if (labelAtEnd) {
	    plantInstruction(new NoArgInstruction(Constants.opc_nop));
	}
	
	try {
	    AccessorTransformer	transformer = 
		new AccessorTransformer() {
			public InstructionAccessor
			    transform(InstructionAccessor accessor,
				     AccessorContainer container)
			{
			    // the only accessors to resolve are labels
			    return 
				getInstructionAt(((CodeLabel)accessor)
						 .getAddress());
			}
		    };
	    
	    for (int i = 0; i < pc; i++) {
		if (instructions[i] instanceof AccessorContainer) {
		    ((AccessorContainer)instructions[i]).transformAccessors(transformer);
		}
	    }
	} catch (BadAccessorException e) {
	    throw new RuntimeException("INTERNAL ERROR");
	}
    }
    
    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------
    
    /**
     * Add a local variable name information
     */
    private final void addLocalVarInfo(LocalVarInstruction insn,
				       JLocalVariable var) {
    }
    
    
    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------
    
    private Instruction[]		instructions;
    private ArrayList			handlers;
    private ArrayList			lines;
    /* LOCAL VARIABLES NOT USED
       private Hashtable			locals;
       LOCAL VARIABLES NOT USED */
    private int				pc;
    private boolean			discardValue;
    // is a label after the last instruction ?
    private boolean			labelAtEnd;
    private int				lineNumber;
    private int				lastLine;
    private /*@ spec_public @*/ Stack	contexts = new Stack();
    
    private static final Stack		stack = new Stack();
}





