/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * $Id: Debug.java,v 1.8 2004/01/16 23:21:10 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.MessageDescription;

import java.io.PrintStream;

import java.text.BreakIterator;

import java.util.HashMap;
import java.util.Map;
import java.util.Stack;
import java.util.ArrayList;
import java.io.IOException;
import java.io.File;
import java.io.PrintWriter;
import java.io.FileWriter;

/**
 * This singleton class localizes the reporting of debugging
 * information.  It includes features like indenting to match depth of
 * the call stack and prefixes to indicate what section of code is
 * being executed.  It also allows users to specify an output stream
 * other than standard out (e.g., a file).  */
public class Debug {		// !FIXME! this should probably be synchronized
    /**
     * Overrides the default initialize so this class cannot be instantiated.
     */
    private Debug() {
    }

    //---------------------------------------------------------------------
    // INITIALIZATION METHODS
    //---------------------------------------------------------------------

    /**
     * Resets the debug reporting methods to turn off debugging and use
     * standard error as the output stream
     */
    public static void initialize() {
	initialize( false );
    }

    /**
     * Resets the debug reporting methods to use
     * standard error as the output stream.
     *
     * @param	debug		indicates whether debug reporting methods 
     *				should report
     */
    public static void initialize(boolean debug) {
	initialize( debug, System.err );
    }

    /**
     * Resets the debug reporting methods.
     *
     * @param	debug		indicates whether debug reporting methods 
     *				should report
     * @param	out		the stream to write debug messages to
     */
    public static void initialize(boolean debug, PrintStream out) {
	Debug.debug = debug;
	Debug.out = out;
    }

    /**
     * Sets the prefix that is written before each debugging message.
     * This is useful, for example, in differentiating the debugging 
     * messages from different compilation passes.
     *
     * @param	prefix	the prefix that should precede all output lines
     */
    public static void setPrefix( String prefix ) {
	Debug.prefix = prefix;
	indent = 0;
	if( debug ) {
	    output( "============================================================" );
	}
    }

    /**
     * Indicates whether debugging output is currently activated.
     *
     * @return	a flag indicating whether debugging output is currently 
     *		activated
     */
    public static boolean isDebugOn() {
	return debug;
    }

    //---------------------------------------------------------------------
    // METHODS TO WRITE DEBUG MESSAGES
    //---------------------------------------------------------------------

    /**
     * Closes a debugging session and print the indent stack if the
     * outdents did not match the indents.  */
    public static void close() {
	out.println();
	out.println( "Debugging complete. " );
	java.util.Iterator keyIter = metaDebugMap.keySet().iterator();
	while( keyIter.hasNext() ) {
	    Object caller = keyIter.next();
	    Stack st = (Stack) metaDebugMap.get( caller );
	    if( st != null && !st.empty() ) {
		out.println( "!!! " + st.size() 
			     + (st.size() > 1 ? " entries " : " entry ")
			     + "left on stack for caller: " + caller );
		out.println( "caller's stack: " + st );
	    }
	}
	metaDebugMap.clear();
    }
    
    /**
     * Writes the given debugging message to the debugging output stream.
     *
     * @param	msg		the string to write
     */
    public static void msg( String msg ) {
	if( debug )
	    output( msg );
    }

    /**
     * Writes the given debugging message, without parameter
     * substitution, to the debugging output stream.
     *
     * @param	msg		the MessageDescription to write
     */
    public static void msg( MessageDescription msg ) {
	if( debug )
	    output( msg.format( null ) );
    }

    /**
     * Writes the given debugging message, with appropriate parameter
     * substitution, to the debugging output stream.
     *
     * @param	msg		the MessageDescription to write
     * @param	p0		the object to substitute for {0} in msg
     */
    public static void msg( MessageDescription msg, Object p0 ) {
	if( debug )
	    output( msg.format( new Object[] { p0 } ) );
    }

    /**
     * Writes the given debugging message, with appropriate parameter
     * substitution, to the debugging output stream.
     *
     * @param	msg		the MessageDescription to write
     * @param	p0		the object to substitute for {0} in msg
     * @param	p1		the object to substitute for {1} in msg
     */
    public static void msg( MessageDescription msg, Object p0, Object p1 ) {
	if( debug )
	    output( msg.format( new Object[] { p0, p1 } ) );
    }

    /**
     * Writes the given debugging message, with appropriate parameter
     * substitution, to the debugging output stream.
     *
     * @param	msg		the MessageDescription to write
     * @param	ps		the objects to substitute msg
     */
    public static void msg( MessageDescription msg, Object[] ps ) {
	if( debug )
	    output( msg.format( ps ) );
    }

    /**
     * Writes the given string to the debug output stream handling
     * indenting, prefixes and line wrapping.
     *
     * @param fullMsg the String to write */
    private static void output( String fullMsg ) {
	// calculate the space available for the message once the
	// prefix and indentation are considered
	int availableWidth = 
	    lineWidth - (prefix.length() + 2) - ((indent+1) * TAB_LENGTH);

	// set up the line BreakIterator for this message
	lineBreaker.setText( fullMsg );

	// break the message into lines
	int start = 0;
	int end = start + availableWidth;
	while( end < fullMsg.length() ) {
	    end = lineBreaker.preceding( end );
	    if( end==start ) {	
		// no breaks available within availableWidth
		// characters so just grab that many characters and
		// let the break fall where it may
		end += availableWidth;
	    }
	    lines.add( fullMsg.substring( start, end ) );
	    start = end;
	    end += availableWidth;
	}
	lines.add( fullMsg.substring( start ) );

	// write the lines to the output stream
	for( int l=0; l < lines.size(); l++) {
	    out.print( prefix + ":" );
	    for( int i=0; i<indent; i++) {
		out.print( TAB );
	    }
	    if( l>0 ) {
		out.print( BLANK );
	    }
	    out.println( lines.get( l ) );
	}

	// reset the lines list so it's ready for next time
	lines.clear();
    }

    //---------------------------------------------------------------------
    // METHODS FOR PROFILING
    //---------------------------------------------------------------------

    /**
     * Begins profiling for a block of code with the given profileID object.
     *
     * @see #profileEnd( Object )
     */
    public static void profileStart( Object profileID ) {
	profileMap.put( profileID, new Long( System.currentTimeMillis() ) );
    }

    /**
     * Finishes profiling for a block of code with the given profileID
     * object by returning the time (in milliseconds) since
     * profileStart was called for the given ID.
     *
     * @see #profileStart( Object ) */
    public static long profileEnd( Object profileID ) {
	long endTime = System.currentTimeMillis();
	Long startTime = (Long) profileMap.get( profileID );
	profileMap.remove( profileID );
	long result = endTime - startTime.longValue();
	return result;
    }

    /**
     * Finishes profiling for a block of code with the given profileID
     * object by returning a String of the form (profileID.toString +
     * "took x ms") where x is the time (in milliseconds) since
     * profileStart was called for the given ID.
     *
     * @see #profileStart( Object ) */
    public static String profileEndAsString( Object profileID ) {
	long result = profileEnd( profileID );
	return profileID.toString() + " took " + result + " ms";
    }

    // -------------------------------------------------------------
    // METHODS FOR WRITING MESSAGES TO DISK FILE
    // -------------------------------------------------------------

    /**
     * Appends the given string to a file named
     * MultiJavaDebugMessages.txt in the current temporary directory,
     * creating the file if it does not exist.
     */
    public static void recordMessage( String msg ) {
	try {
	    File f = new File(System.getProperty("java.io.tmpdir"),
			      "MultiJavaDebugMessages.txt");
	    PrintWriter fileOut = 
		new PrintWriter( new FileWriter( f.getAbsolutePath(), true ) );
	    fileOut.println(msg);
	    fileOut.close();
	} catch (IOException e) {
	    System.err.println("IOException while recording debug message \"" + 
			       msg + "\".");
	    e.printStackTrace(System.err);
	}
    }

    //---------------------------------------------------------------------
    // META-DEBUGGING METHODS
    //---------------------------------------------------------------------

    private static void registerIndent( Object caller, String msg ) {
	Stack maybeStack = (Stack) metaDebugMap.get( caller );
	if( maybeStack==null ) {
	    maybeStack = new Stack();
	    maybeStack.push( msg );
	    metaDebugMap.put( caller, maybeStack );
	} else {
	    maybeStack.push( msg );
	}
    }

    private static void registerOutdent( Object caller ) {
	Stack maybeStack = (Stack) metaDebugMap.get( caller );
	if( maybeStack==null || maybeStack.empty() ) {
	    out.println( "!!! Debugging output error !!!" );
	    out.println( "outdenting without matching indent for " 
			 + caller.getClass() );
	    Thread.dumpStack();
	} else {
	    try {
		maybeStack.pop();
	    } catch( java.util.EmptyStackException e ) {
		// This is guarded by the if, so we don't get here
		out.println( "!!! Debugging output error !!!" );
		e.printStackTrace( out );
	    }
	}
    }

    /**
     * This method does nothing.  It is a hack for using jdb.
     * Sometimes jdb won't set a break point, complaining "no code at
     * line x" when you know good and well that there is code there,
     * dammit.  So, set a break point on this (stop in
     * org.multijava.mjc.Debug.breakMe()) and add a call Debug.breakMe()
     * where you want to break.  This also supports conditional
     * breaks.  Cool, huh? */
    public static void breakMe() {
    }

    //---------------------------------------------------------------------
    // STATIC FIELDS
    //---------------------------------------------------------------------

    private static boolean debug = false;
    private static PrintStream out = System.err;
    private static String prefix = "DBG";
    private static int indent = 0;
    private static int lineWidth = 80;
    private static long messageMask = 1;
    
    private static final String TAB = "|  ";
    private static final String BLANK = "   ";
    private static final int TAB_LENGTH = TAB.length();
    private static final BreakIterator lineBreaker = 
	BreakIterator.getLineInstance();
    private static final ArrayList lines = new ArrayList();
    private static final Map metaDebugMap = new HashMap();
    private static final Map profileMap = new HashMap();

    /**
     * A cache of System.err so we can hack around code that uses
     * System.setErr() */
    public static final PrintStream stderr = System.err;

}
