/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: FunctionalTestSuite.java,v 1.13 2008/03/16 15:52:50 wdietl Exp $
 */

package org.multijava.mjc;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Stack;
import java.util.StringTokenizer;

public class FunctionalTestSuite extends junit.framework.TestSuite {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------
    
    public FunctionalTestSuite() {
	super();
    }

    public FunctionalTestSuite( String name ) {
	super( name );
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------

    public static junit.framework.Test suite() {
        junit.framework.TestSuite suite = 
            new junit.framework.TestSuite() {
                    public String toString() { return TEST_DESC; }
                };
        suite.addTestSuite( org.multijava.mjc.TestMain_Typecheck.class );
        suite.addTestSuite( org.multijava.mjc.TestMain_Runtime.class );
        suite.addTestSuite( org.multijava.mjc.TestMain_CodeGen.class );
        suite.addTestSuite( org.multijava.mjc.TestMain_TransCompile.class );
        return suite;
    }

    public static final String TEST_DESC = 
        "Functional tests";

    //---------------------------------------------------------------------
    // NESTED HELPER CLASS FOR FUNCTIONAL TESTS
    //---------------------------------------------------------------------

    public static class TestCase extends org.multijava.util.testing.TestCase {

	public TestCase( String name ) {
	    super( name );
	    String filesep = System.getProperty( "file.separator" );
	    this.workingDir = 
		new File( System.getProperty( "java.io.tmpdir" ) +
			  filesep +
			  "mjc" );
	    // create the working directory and mark it for deletion
	    // after all tests are run
	    this.workingDir.mkdir();
	    this.workingDir.deleteOnExit();
	    //	    this.dir = workingDir.toString() + filesep;
	    this.tempLoader = 
		new ClassLoader() {
			public Class findClass( String clName )
			    throws ClassNotFoundException
			{
			    try {
				byte[] b = loadClassData( clName );
				return defineClass( clName, b, 0, b.length );
			    } catch( IOException e ) {
				throw new ClassNotFoundException( clName + 
								  " not found" 
								  + " in " + 
								  workingDir );
			    }
			}
			private byte[] loadClassData( String clName ) 
			    throws IOException
			{
			    String fn = 
				clName.replace( '.', 
						System.
						getProperty( "file.separator" )
						.charAt(0)
						) + ".class";

			    File inFile = 
				new File( workingDir, fn );
			    BufferedInputStream in = null;
                            byte[] result = null;
                            try {
                                in = new BufferedInputStream( 
					 new FileInputStream( inFile ) );
                                result = new byte[ in.available() ];
                                in.read( result );
                            } finally {
                                if (in != null) {
                                    in.close();
                                }
                            }
			    return result;
			}
		    };
	}

	/**
	 * This is a hack to blast all the *.class files out of the
	 * temp directory.
	 */
	public void setUp() { cleanWorkingDir(); }

	/**
	 * This is a hack to blast all the *.class files out of the
	 * temp directory.
	 */
	public void tearDown() { cleanWorkingDir(); }


	/**
	 * Deletes all *.class files in workingDir and its
	 * subdirectories.
	 */
	protected final void cleanWorkingDir() {
	    ArrayList accum = new ArrayList();
	    ArrayList accumDirs = new ArrayList();
	    Stack dirs = new Stack();

	    dirs.push(workingDir);
	    while (!dirs.empty()) {
		File dir = (File) dirs.pop();
		// get files
                File [] files = dir.listFiles(CLASS_FILE_FILTER);
                if (files != null) {
                    accum.addAll( Arrays.asList( files) );
                }
                File [] subdirs = dir.listFiles(SUBDIR_FILTER);
                if (subdirs != null) {
                    // get subdirectories
		    List sd = Arrays.asList( subdirs );
		    dirs.addAll( sd );
		    // append to front to get deletion order right
                    accumDirs.addAll( 0, sd );
                }
	    }
	
	    for (int i = 0; i < accum.size(); i++) {
		((File) accum.get(i)).delete();
	    }
	    for (int i = 0; i < accumDirs.size(); i++) {
		((File) accumDirs.get(i)).delete();
	    }
	    // we can not remove the working directory itself,
	    // because the following testcases still need it
	    //	    this.workingDir.delete();
	}
	
	/**
	 * Used by {@link #cleanWorkingDir()}.
	 */
	private static final FilenameFilter CLASS_FILE_FILTER = 
	    new FilenameFilter() {
		public boolean accept(File dir, String name) {
		    return name.endsWith(".class");
		}
	    };
	
	/**
	 * Used by {@link #cleanWorkingDir()}.
	 */
	private static final FileFilter SUBDIR_FILTER = 
	    new FileFilter() {
		public boolean accept(File pathname) {
		    return pathname.isDirectory();
		}
	    };

	/**
	 * Creates a file in the temporary directory with the given
	 * name and contents.  To keep test methods from having to
	 * catch an IOException, these are converted to (unchecked)
	 * RuntimeExceptions.
	 *
	 * @return the file name so that operations may be cascaded */
	protected String saveAs( String fileName, String contents ) {
	    try {
		FileWriter out = 
		    new FileWriter( new File( workingDir, fileName ));
		out.write( contents );
		out.close();
	    } catch( IOException e ) {
		throw new RuntimeException( e.getMessage() );
	    }
	    return fileName;
	}

	/**
	 * Deletes files of the given names from the current directory. 
	 *
	 * @param fileNames a space delimited list of file names
	 */
	protected void remove( String fileNames ) {
	    String[] names = split( fileNames );
	    for( int i=0; i < names.length; i++ ) {
		new File( workingDir, names[i] ).delete();
	    }
	}

	/**
	 * Invokes the compiler, passing it the given command line.
	 * Standard output and error streams are captured so they can be
	 * included in the results. */
	protected CompilationResults compile( String cmdLine ) {
	    OutputStream[] streams = captureStreams();
	    String[] oldDirs = setWorkingDir();

	    try {
		String[] args;
		if( cmdLine.indexOf( "-d" ) < 0 ) {
		    args = split( "-d " + workingDir + " " + cmdLine );
		} else {
		    args = split( cmdLine );
		}
		boolean success = invokeCompiler(args);

		return new CompilationResults( success, streams[0].toString(),
					       streams[1].toString() );
	    } finally {
		releaseStreams( streams );
		resetWorkingDir( oldDirs );
	    }
	}

	protected boolean invokeCompiler( String[] args ) {
	    return Main.compile( args );
	}

	/**
	 * Establishes a new working directory and class path and
	 * returns an array containing the old values.  The array may
	 * be passed to <code>resetWorkingDir( String[] )</code> to
	 * reset to the original values. 
	 *
	 * @see #resetWorkingDir(String[])
	 */
	private String[] setWorkingDir() {
	    String oldWorkingDir = System.getProperty( WORKING_DIR_PROP );
	    System.setProperty( WORKING_DIR_PROP, workingDir.toString() );
	    String oldClassPath = System.getProperty( CLASS_PATH_PROP );
	    System.setProperty( CLASS_PATH_PROP, workingDir.toString() );
	    return new String[] { oldWorkingDir, oldClassPath };
	}

	/**
	 * Resets the working directory and class path based on the argument
	 * array returned by <code>setWorkingDir()</code>. 
	 *
	 * @see #setWorkingDir()
	 */
	private void resetWorkingDir( String[] old ) {
	    System.setProperty( WORKING_DIR_PROP, old[0] );
	    System.setProperty( CLASS_PATH_PROP, old[1] );
	}

	private static final String WORKING_DIR_PROP = "user.dir";
	private static final String CLASS_PATH_PROP = "java.class.path";

	/**
	 * Establishes new output streams to capture text written to
	 * standard out and standard error.  Returns an array such the
	 * \result[0] is the new standard error stream, \result[1] is
	 * the new standard output stream, \result[2] is the old
	 * standard error stream, and \result[3] is the old standard
	 * output stream.
	 *
	 * @see #releaseStreams(OutputStream[]) */
	private OutputStream[] captureStreams() {
	    PrintStream oldErrorOut = System.err;
	    PrintStream oldStdOut = System.out;
	    ByteArrayOutputStream newErrorOut = new ByteArrayOutputStream();
	    System.setErr( new PrintStream( newErrorOut ) );
	    ByteArrayOutputStream newStdOut = new ByteArrayOutputStream();
	    System.setOut( new PrintStream( newStdOut ) );
	    return new OutputStream[] {
		newErrorOut, newStdOut, oldErrorOut, oldStdOut
	    };
	}

	/**
	 * Reverses the action for <code>captureStreams()</code>.
	 *
	 * @param	streams		the array returned by 
	 *				<code>captureStreams()</code>
	 * @see #captureStreams() */
	private void releaseStreams( OutputStream[] streams ) {
	    System.setErr( (PrintStream) streams[2] );
	    System.setOut( (PrintStream) streams[3] );
	}

	/**
	 * Tests the class with the given name.  Loads the class and
	 * executes a method in it with signature <code>public static
	 * void main( String[] )</code>, passing an empty array.
	 * Captures the values written to standard out by the method
	 * and returns these in the result.  The result also contains
	 * a record of the exception thrown if the execution
	 * failed. */
	protected RuntimeResults run( String className ) {
	    String[] oldDirs = setWorkingDir();
	    OutputStream[] streams = captureStreams();
	    boolean success = false;
	    Throwable excep = null;
	    try {
		Class clazz = Class.forName( className, true, tempLoader );
		Method runMethod = clazz.getMethod( "main", STRING_ARRAY );
		runMethod.invoke( null, EMPTY_ARRAY );
		success = true;
	    } catch( InvocationTargetException ivte ) {
		excep = ivte.getTargetException();
	    } catch( Throwable e ) {
		excep = e;
	    } finally {
		resetWorkingDir( oldDirs );
		releaseStreams( streams );
	    }
		return new RuntimeResults( success, excep, 
					   streams[0].toString(),
					   streams[1].toString() );
	}

	/**
	 * Disassembles the class with the given name. */
	protected CodeGenResults disasm( String className, boolean details ) {
	    String[] oldDirs = setWorkingDir();
	    OutputStream[] streams = captureStreams();
	    boolean success = false;
	    String errors = "";
	    String results = "";
	    try {
		String[] args;
		if( details ) {
		    args = new String[] { "--stdout", className };
		} else {
		    args = new String[] { "--stdout", "--inter", className };
		}

		success = org.multijava.dis.Main.run( args );

		errors = streams[0].toString();
		results = streams[1].toString();

	    } catch( Exception e ) {
		errors = e.toString();
		success = false;
	    } finally {
		resetWorkingDir( oldDirs );
		releaseStreams( streams );
	    }
		return new CodeGenResults( success, errors, results );
	}

	/**
	 * Splits the given text into an array of words separated by
	 * whitespace.  Requires that splitCache be empty when this is
	 * called.  Ensures the splitCache is empty.  
	 *
	 * @see #splitCache
	 */
	protected String[] split( String text ) {
	    StringTokenizer toker = new StringTokenizer( text );
	    while( toker.hasMoreTokens() ) {
		splitCache.add( toker.nextToken() );
	    }
	    
	    String[] result = (String[]) splitCache.toArray( new String[0] );
	    splitCache.clear();
	    return result;
	}
    
	/**
	 * Called with the result of compilation or runtime when
	 * success is expected, this generates the appropriate
	 * assertion and message. */
	protected void expectSuccess( CompilationResults results ) {
	    if( results.success ) {
		return;
	    }
	    StringBuffer msg = 
		new StringBuffer( "expected success, had failure: " );
	    msg.append( "error = <" + results.errorOut + ">" );
	    if( results instanceof RuntimeResults ) { // !CONVERT! 
		RuntimeResults rr = (RuntimeResults) results;
		if( rr.excep != null ) {
		    StringWriter writer = new StringWriter();
		    rr.excep.printStackTrace( new PrintWriter( writer ) );
		    msg.append( " exception = <" + writer.toString() + ">" );
		}
	    }
	    assertTrue( msg.toString(), false );
	}
    
	/**
	 * Called with the result of compilation when failure (i.e.,
	 * compilation error) is expected, this generates the appropriate
	 * assertion and message. */
	protected void expectFailure( CompilationResults results,
				    String expectedError ) {
	    assertTrue( "expected failure, had success", !results.success );
	    assertEquals( expectedError, results.errorOut, true );
	}
    
	//---------------------------------------------------------------------
	// COMMON FIXTURE
	//---------------------------------------------------------------------

	/**
	 * Used by split() to split strings into words.  Instead of
	 * creating a new list we fill and flush this one.  
	 *
	 * @see #split( String )
	 */
	private ArrayList splitCache = new ArrayList();    

	/**
	 * The temporary working directory.  */
	protected final File workingDir;

	/**
	 * A string representation of the working directory, used in
	 * matching expected error messages.  */
	// protected String dir;

	/**
	 * Represents the results of a compilation run, including a
	 * success flag and any error messages.  */
	protected static class CompilationResults {
	    CompilationResults( boolean success, String errorOut, 
				String stdOut ) {
		this.success = success;
		this.errorOut = errorOut;
		this.stdOut = stdOut;
	    }
	    final boolean success;
	    final String errorOut;
	    final String stdOut;
	}

	/**
	 * Represents the results of disassembling, including a
	 * success flag and any error messages.  */
	protected static class CodeGenResults extends CompilationResults{
	    CodeGenResults( boolean success, String errorOut, String stdOut ) {
		super( success, errorOut, stdOut );
	    }
	}

	/**
	 * Represents the results of a runtime test, including a
	 * success flag, any exception thrown and the values written
	 * to standard error and standard out. */
	protected static class RuntimeResults extends CompilationResults {
	    RuntimeResults( boolean success, Throwable excep,
			    String errorOut, String stdOut ) {
		super( success, errorOut, stdOut );
		this.excep = excep;
	    }
	    final Throwable excep;
	}

	/* !FIXME!vestigial? */
	protected static class Processor implements Runnable {
	    Processor( String[] cmd, String[] env ) {
		this.cmd = cmd;
		this.env = env;
	    }

	    public void run() {
		try {
		    final int BUF_SIZE = 1000;
		    char[] buffer = new char[ BUF_SIZE ];
		    Runtime rt = Runtime.getRuntime();
		    Process p = rt.exec( cmd, env );
		    BufferedReader errReader = 
			new BufferedReader( 
				new InputStreamReader( p.getErrorStream() ));
		    BufferedReader inReader = 
			new BufferedReader( 
				new InputStreamReader( p.getInputStream() ));
		    int errCount, inCount;
		    // int x = 1;	// !FIXME! 
		    do {
			errCount = errReader.read( buffer, 0, BUF_SIZE );
			if( errCount > 0 ) {
			    errors.append( buffer, 0, errCount );
			}
			inCount = inReader.read( buffer, 0, BUF_SIZE );
			if( inCount > 0 ) {
			    results.append( buffer, 0, inCount );
			}
			//errors.append( "iteration " + (x++) + " errCount="
			//		 + errCount + " inCount=" + inCount
			//		 + NEWLINE );
			Thread.yield();
		    } while( errCount != -1 || inCount != -1 && !killed );
		    if( !killed ) {
			completed = true;
		    }
		} catch( Exception e ) {
		    errors = new StringBuffer( "exception running process: " 
					       + e.toString() );
		    completed = true;
		} finally {
		    cleanUp();
		}
	    }

	    public String errors() {
		return errors.toString();
	    }

	    public String results() {
		return results.toString();
	    }

	    public void cleanUp() {
		try {
		    if( errReader != null ) {
			errReader.close();
			errReader = null;
		    }
		} catch( IOException e ) {
		    // ignored
		}
		try {
		    if( inReader != null ) {
			inReader.close();
			inReader = null;
		    }
		} catch( IOException e ) {
		    // ignored
		}
	    }
	    
	    protected void finalize() {
		cleanUp();
	    }

	    public void kill() {
		killed = true;
	    }

	    String[] cmd;
	    String[] env;
	    boolean completed = false;
	    StringBuffer errors = new StringBuffer();
	    StringBuffer results = new StringBuffer();
	    private BufferedReader errReader;
	    private BufferedReader inReader;
	    private boolean killed;
	}

	private static Object[] EMPTY_ARRAY = new Object[] { new String[0] };
	private static Class[] STRING_ARRAY = new Class[] { String[].class };

	private ClassLoader tempLoader;
    }
}
