/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JAddExpression.java,v 1.14 2005/02/12 23:24:12 xi_hao Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.ClassRefInstruction;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the addition binary expression.
 */
public class JAddExpression extends JBinaryArithmeticExpression implements Constants{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JAddExpression( TokenReference where,
			   JExpression left,
			   JExpression right ) {
	super( where, left, right );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return left.toString() + " + " + right.toString();
    }
    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	isInSafeMathMode  = context.arithmeticMode().equals(AMID_SAFE_MATH);

	left = left.typecheck( context );
	right = right.typecheck( context );

	check( context, 
	       left.getType() != CStdType.Void 
	       && right.getType() != CStdType.Void,
	       MjcMessages.ADD_BADTYPE, 
	       left.getType().toVerboseString(), 
	       right.getType().toVerboseString() );

	check( context, 
	       !(left instanceof JTypeNameExpression) 
	       && !(right instanceof JTypeNameExpression),
	       MjcMessages.ADD_BADTYPE, 
	       left.getType().toVerboseString(), 
	       right.getType().toVerboseString() );

	JExpression expr = typecheckBinaryArith(context);
	if( (expr == this) && !type.equals(CStdType.String) ) {
	    left = left.convertType( type, context ); 
	    right = right.convertType( type, context );
	}
	return expr;
    }

    /**
     * Typechecks a binary arithmetic expression and handles simple
     * optimizations like constant folding.
     *
     * @param context the actual context of analysis
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if checks fail */
    public JExpression typecheckBinaryArith( CExpressionContextType context ) 
	throws PositionedError 
    {
	try {
	    type = computeType( left.getType(), right.getType() );
	    // no cst fold on string to save constant pool size ???
	    return ( left.isConstant() && right.isConstant() 
		     && !type.equals(CStdType.String) )
		? constantFolding( context )
		: verifyOperation( this, context );
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
    }
    
    /**
     * Verify the operation. Overridden in JML to issue warnings.
     */
    public JExpression verifyOperation(JExpression expr,  CExpressionContextType context ){
    	return expr;
    }
    
    /**
     * compute the type of this expression according to operands
     * @param	leftType	the type of left operand
     * @param	rightType	the type of right operand
     * @return	the type computed for this binary operation
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public static CType computeType( CType leftType, 
				     CType rightType) 
	throws UnpositionedError 
    {
	// WMD: we ignore the Universe part, because we want the
	// String "+" operator to be (readonly, readonly) -> peer.
	if( leftType.equals(CStdType.String, false) 
	    || rightType.equals(CStdType.String, false) ) {
	    return CStdType.String;
	}

	if( leftType.isNumeric() && rightType.isNumeric() ) {
	    return ((CNumericType)leftType).binaryPromote( rightType );	    
	}

	throw new UnpositionedError( MjcMessages.ADD_BADTYPE, 
				     leftType.toVerboseString(),
				     rightType.toVerboseString() );
    }

    /**
     * Converts two constant expressions into a single literal that is
     * the result of applying this operation to the expressions.
     *
     * @return	a literal resulting from an operation over two literals
     */
    public JExpression constantFolding( CExpressionContextType context ) 
	throws UnpositionedError 
    {
	if( type == CStdType.Long ) {
	    long result,
		left_value = left.getNumberLiteral().numberValue().longValue(),
		right_value = right.getNumberLiteral().numberValue().longValue();

	    result = !context.arithmeticMode().equals(AMID_JAVA_MATH)
		? safe_compute( left_value, right_value )
		: compute( left_value, right_value, context );
	    return new JOrdinalLiteral( getTokenReference(), 
					new Long(result), CStdType.Long);
	} else if( type == CStdType.Float ) {
	    float result;
      
	    result = 
		compute( left.getNumberLiteral().numberValue().floatValue(), 
			 right.getNumberLiteral().numberValue().floatValue() );
	    return new JRealLiteral( getTokenReference(), 
				     new Float(result), CStdType.Float);
	} else if( type == CStdType.Double ) {
	    double result;
      
	    result = 
		compute( left.getNumberLiteral().numberValue().doubleValue(), 
			 right.getNumberLiteral().numberValue().doubleValue());
	    return new JRealLiteral( getTokenReference(), 
				     new Double(result), CStdType.Double);
	} else if( type.implicitlyCastTo(CStdType.Integer) ) {
	    int result,
		left_value = left.getNumberLiteral().numberValue().intValue(),
		right_value = right.getNumberLiteral().numberValue().intValue();

	    result = !context.arithmeticMode().equals(AMID_JAVA_MATH)
		? safe_compute( left_value, right_value )
		: compute( left_value, right_value, context );
	    return new JOrdinalLiteral( getTokenReference(), 
					new Integer(result), CStdType.Integer);
	} else {
	    fail();
	    return null;
	}
    }

    public int safe_compute(int left, int right)
	throws UnpositionedError 
    {
	int result = left + right;
	if(left > 0 && right > 0 && result < 0 || left < 0 && right < 0 && result >= 0)	{
	    throw new UnpositionedError( MjcMessages.INTEGRAL_ARITHMETIC_OVERFLOW );
	}
	return result;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public int compute(int left, int right, CExpressionContextType context) {
	return left + right;
    }

    public long safe_compute(long left, long right) 
	throws UnpositionedError 
    {
	long result = left + right;
	if(left > 0 && right > 0 && result < 0 || left < 0 && right < 0 && result >= 0) {
	    throw new UnpositionedError( MjcMessages.INTEGRAL_ARITHMETIC_OVERFLOW );
	}

	return result;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public long compute(long left, long right, CExpressionContextType context) {
	return left + right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public float compute(float left, float right) {
	return left + right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public double compute(double left, double right) {
	return left + right;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitAddExpression( this );
    }

    /**
     * @param	type		the type of result
     * @return	the type of opcode for this operation
     */
    public static int getOpcode(CType type) {
	switch (type.getTypeID()) {
	case TID_FLOAT:
	    return opc_fadd;
	case TID_LONG:
	    return opc_ladd;
	case TID_DOUBLE:
	    return opc_dadd;
	default:
	    return opc_iadd;
	}
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {

	code.setLineNumber(getTokenReference().line());
	if (type.equals(CStdType.String)) {
	    code.plantInstruction(new ClassRefInstruction(opc_new, "java/lang/StringBuffer"));
	    code.plantInstruction(new NoArgInstruction(opc_dup));
	    code.plantInstruction(new MethodRefInstruction(opc_invokespecial,
							   "java/lang/StringBuffer",
							   JAV_CONSTRUCTOR,
							   "()V"));
	    appendToStringBuffer(code, left);
	    appendToStringBuffer(code, right);
	    code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
							   "java/lang/StringBuffer",
							   "toString",
							   "()Ljava/lang/String;"));
	} else {
	    left.genCode(code);
	    right.genCode(code);

	    if (isInSafeMathMode){

		if (getType().equals(CStdType.Long)) {
		    
		    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
								   "org/multijava/mjc/SafeIntegralArithmetic",
								   "add",
								   "(JJ)J"));
		} else  if ((getType().equals(CStdType.Double))||(getType().equals(CStdType.Float))){
		    code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
		} else {   
		    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "add",
                                                                   "(II)I"));
		}
	    } else {
		code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
	    }
	}
    }
    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    private void appendToStringBuffer(CodeSequence code, JExpression expr) {
	if ((expr instanceof JAddExpression) && (expr.getType().equals(CStdType.String))) {
	    ((JAddExpression)expr).appendToStringBuffer(code, ((JAddExpression)expr).left);
	    ((JAddExpression)expr).appendToStringBuffer(code, ((JAddExpression)expr).right);
	} else {
	    expr.genCode(code);

	    CType	type = expr.getType();
	    if (!type.isReference() || type.equals(CStdType.String) || type.equals(CStdType.CharArray)) {
		if (type == CStdType.Byte || type == CStdType.Short) {
		    type = CStdType.Integer;
		}

		code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
							       "java/lang/StringBuffer",
							       "append",
							       "(" + type.getSignature() + ")Ljava/lang/StringBuffer;"));
	    } else {
		code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
							       "java/lang/StringBuffer",
							       "append",
							       "(Ljava/lang/Object;)Ljava/lang/StringBuffer;"));
	    }
	}
    }

    // private data
    private boolean isInSafeMathMode  = false;
}




