/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JArrayAccessExpression.java,v 1.20 2006/12/13 20:59:34 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class implements an access to an array.
 */
public class JArrayAccessExpression extends JExpression 
    implements CInitializable {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	prefix		denotes the array to access
     * @param	accessor	denotes a natural integer indicating the 
     *				element to access
     */
    public JArrayAccessExpression(TokenReference where,
				  JExpression prefix,
				  JExpression accessor)
    {
	super( where );

	this.prefix = prefix;
	this.accessor = accessor;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression prefix() { return prefix; }
    public JExpression accessor() { return accessor; }

    /**
     * @return	the type of this expression
     */
    public CType getType() {
	return type;
    }

    /**
     * Indicates whether this expression can appear on the left-hand
     * side of an assignment or in an increment or decrement
     * operation.
     *
     * <pre><jml>
     * also ensures true;
     * </jml></pre>
     *
     */
    public boolean isMaybeInitializable() {
	return true;
    }

    /**
     * @return	true if this expression is a variable already valued
     */
    public boolean isDefinitelyAssigned( CContextType context) {
	// nothing to do in array access 15.12 Array Access Expressions
	return true;
    }

    /**
     * Initialize this object
     */
    public void initialize( CContextType ctxt ) {
	// nothing to do in array access 15.12 Array Access Expressions
    }

    /** Determine whether the array access is a LValue in this context.
     * WMD: A readonly array is never a LValue.
     *
     */
    public boolean isLValue(CExpressionContextType context) {
	// nothing to do in array access 15.12 Array Access Expressions
	// WMD old: return true;

	// WMD
	if( CTopLevel.getCompiler().universePurity() &&
	    context.isPure() ) {

	    return false;
	}

	// WMD: readonly array references are no LValue
	if( CTopLevel.getCompiler().universeChecks() ) {

	    boolean prefix_io_ro =
		((CArrayType) prefix.getType()).getCUniverse() instanceof CUniverseReadonly;

	    if( prefix_io_ro ) {
		return false;
	    } else
		// scdaniel: only warn if dynamic checks are disabled
		if( type instanceof CClassType &&
		    ((CClassType) type).getCUniverse() 
		    instanceof CUniverseReadonly &&
		    !CTopLevel.getCompiler().universeDynChecks() )
		    {
			context.reportTrouble(
			    new CWarning( getTokenReference(),
				      CUniverseMessages.COVARIANT_ARRAY_WRITE,
				      prefix.getType().toVerboseString() ));
		    }
	}

	return true;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return prefix.isDeclaredNonNull() || context.isFANonNull(this);
    }

    /**
     * Returns true iff the variable, field or method represente by this expression
     * was declared non_null.
     */
    public /*@pure*/ boolean isDeclaredNonNull() {
	   return prefix.isDeclaredNonNull();
   }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking. JLS2 15.13 Array
     * Access Expressions
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	accessor = accessor.typecheck( context, false, false, false );
	check( context,
	       accessor.getType().implicitlyCastTo( CStdType.Integer ),
	       MjcMessages.ARRAY_EXPRESSION_INT, 
	       accessor.getType() );

	prefix = prefix.typecheck( context, false, false, false );
	CType exprType = prefix.getType();
	check( context, 
	       exprType.isArrayType(), 
	       MjcMessages.ARRAY_PREFIX,
	       exprType.toString() );

	type = ((CArrayType) prefix.getType()).getIndexedType();

	// Unless we are on the LHS of an assignment, we have to apply
	// capture conversion (JLS3 5.1.10) to the type of an array access
	// expression (see JLS3 15.13).
	if (!context.isLeftSide()) {
	    try {
	        type = getType().getCapture(context);
	    } catch (UnpositionedError e) {
	        throw e.addPosition(getTokenReference());
	    }
	}

	// check nullity
	if (CTopLevel.getCompiler().nonNullTypes()) {
           check( context,
                  prefix.isNonNull(context),
                 MjcMessages.ARRAYACCESS_NULL_REF,
                 CTypeNullity.toString(prefix.isNonNull(context)));
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitArrayAccessExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	boolean discarded = code.discardValue();

	// Evaluates prefix and accessor for side effects
	code.setDiscardValue(false);
	prefix.genCode(code);
	accessor.genCode(code);
	
	code.setDiscardValue(discarded);

	if (discarded) {
	    // pops prefix and accessor
	    code.plantInstruction(new NoArgInstruction(opc_pop2));
	} else {
	    // generates appropriate load opcode
	    int opcode;
	    switch (getType().getTypeID()) {
	    case TID_CLASS:
	    case TID_ARRAY:
		opcode = opc_aaload;
		break;
	    case TID_INT:
		opcode = opc_iaload;
		break;
	    case TID_LONG:
		opcode = opc_laload;
		break;
	    case TID_FLOAT:
		opcode = opc_faload;
		break;
	    case TID_DOUBLE:
		opcode = opc_daload;
		break;
	    case TID_BYTE:
	    case TID_BOOLEAN:
		opcode = opc_baload;
		break;
	    case TID_CHAR:
		opcode = opc_caload;
		break;
	    case TID_SHORT:
		opcode = opc_saload;
		break;
	    default:
		fail();
		opcode = -1;
	    }

	    code.plantInstruction(new NoArgInstruction(opcode));
	}
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genStartStoreCode(CodeSequence code) {
	prefix.genCode(code);
	accessor.genCode(code);
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genEndStoreCode(CodeSequence code, boolean discardValue) {
	if (!discardValue) {
	    if (getType().getSize() == 2) {
		code.plantInstruction(new NoArgInstruction(opc_dup2_x2));
	    } else if (type.getSize() == 1) {
		code.plantInstruction(new NoArgInstruction(opc_dup_x2));
	    }
	}

	int		opcode;

	switch (getType().getTypeID()) {
	case TID_CLASS:
	case TID_ARRAY:
	    opcode = opc_aastore;
	    break;
	case TID_DOUBLE:
	    opcode = opc_dastore;
	    break;
	case TID_FLOAT:
	    opcode = opc_fastore;
	    break;
	case TID_LONG:
	    opcode = opc_lastore;
	    break;
	case TID_BYTE:
	case TID_BOOLEAN:
	    opcode = opc_bastore;
	    break;
	case TID_SHORT:
	    opcode = opc_sastore;
	    break;
	case TID_CHAR:
	    opcode = opc_castore;
	    break;
	default:
	    opcode = opc_iastore;
	}
	code.plantInstruction(new NoArgInstruction(opcode));
    }

	/* 
	 * By rzakhejm
	 */
	public String toString() {
		return prefix.toString() + "[" + accessor.toString() + "]";
	}
    
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	JExpression	prefix;
    private	JExpression	accessor;
    private	CType		type;
}
