/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JArrayDimsAndInits.java,v 1.10 2008/05/20 17:41:05 leavens Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.MultiarrayInstruction;
import org.multijava.util.classfile.ClassRefInstruction;
import org.multijava.util.classfile.NewarrayInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class implements an AST node representing a list of
 * expressions used in array dimensioning and an array initialization
 * expression, as in the <code>[1][2][3]</code> in <code>new
 * Integer[1][2][3]</code> or the <code>{ 1, 2, 3 }</code> in <code>
 * Integer[] foo = { 1, 2, 3 }</code>.  */
public class JArrayDimsAndInits extends JPhylum {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs an array dimension and initializer node in the AST
     * @param	where	a reference to the source file generating this node
     * @param   array_univ      the universe of the whole array WMD
     *                  Maybe better in JNewArrayExpression.
     * @param	dims	an array of the dimension expression AST nodes
     * @param	init	an initializer AST node
     */
    public JArrayDimsAndInits(TokenReference where, 
			      CUniverse array_univ, // WMD
			      JExpression[] dims, 
			      JArrayInitializer init) {
	super(where);
	
	this.array_universe = array_univ;
        this.elem_universe = null;
        
	this.init = init;
	this.dims = dims;
    }

     /**
     * Constructs an array dimension and initializer node in the AST
     * @param	where	a reference to the source file generating this node
     * @param   array_univ      the universe of the whole array WMD
     * @param   elem_univ  the universe for arrays of primitive types WMD
     * @param	dims	an array of the dimension expression AST nodes
     * @param	init	an initializer AST node
     */
    public JArrayDimsAndInits(TokenReference where, 
			      CUniverse array_univ,
                              CUniverse elem_univ,
			      JExpression[] dims, 
			      JArrayInitializer init) {
	super(where);
	
	this.array_universe = array_univ;
        this.elem_universe = elem_univ;
	this.init = init;
	this.dims = dims;
    }

    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this array dimension and initialization instruction
     * in the given context.
     * @param	context		the context in which this instruction appears
     * @param	type		type of the array which we are creating, 
     *				ignored if this is part of an array dimension 
     *				statement like <code>new Integer[1][2][3], 
     *				used if this is part of an array 
     *				initialization like <code>Integer[] foo = 
     *				{ 1, 2, 3 }</code>
     * @return	the type of the array that is being created
     * @exception PositionedError if a check fails */
    public CType typecheck(CExpressionContextType context, CType type) 
	throws PositionedError {
	boolean	expr = false;
	boolean	hole = true;

	// determine whether at least one dimension of the array is
	// specified and that all specified dimensions are integers
	for( int i = 0; i < dims.length; i++ ) {
	    if( dims[i] != null ) {
		expr = true;
		dims[i] = dims[i].typecheck( context );
		check( context, 
		       dims[i].getType().implicitlyCastTo( CStdType.Integer ), 
		       MjcMessages.ARRAY_BADTYPE, 
		       dims[i].getType() );
		hole = false;
	    }
	}

	// It must be the case that some dimension is given or that an
	// initializer is present (from which the array dimensions can
	// be determined.
	check(context, !hole || init != null, MjcMessages.ARRAY_NEW_NOSIZE);

	if( this.type == null ) {
            if( type.isPrimitive() && dims.length > 1 ) {
                this.type = new CArrayType(type, dims.length, array_universe, elem_universe);
            } else {
	        this.type = new CArrayType(type, dims.length, array_universe);
            }
            
	    this.type = (CArrayType)CUniverseServices.setDefaultVariableAgain(this.type);
            if( CTopLevel.getCompiler().universeChecks() &&
                    CUniverseServices.isPartial(this.type) &&
                    init!=null &&
                    init.elems()!=null &&
                    init.elems().length > 0  ) {
                // also see JArrayInitializer
                try {
                    init.elems()[0].typecheck(context);
                    if (init.elems()[0].getType() != null) {
                        CArrayType templ = new CArrayType(init.elems()[0].getType(), ((CArrayType) this.type).getArrayBound(), CUniverseImplicitPeer.getUniverse());
                        this.type = (CArrayType) CUniverseServices.propagate(this.type, templ);
                    }
                } catch (Exception e) {
                    // forget about it.
                }
             
            }
	}

	// If there is an initializer then there cannot be an
	// expression giving the array dimensions.  Checks this and
	// that the initializer has the correct type.
	if( init != null ) {
	    check( context, !expr, MjcMessages.ARRAY_NEW_NOSIZE );
	    init.setType( this.type );
	    init.typecheck( context );
	}
	return this.type;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitArrayDimsAndInit( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	if (code.discardValue()) {
	    return;
	}

	if (init == null) {
	    allocArray(code, type, dims);
	} else {
	    init.genCode(code);
	}
    }


    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public static void allocArray(CodeSequence code, 
				  CArrayType type, 
				  JExpression[] dims) {
	for (int i = 0; i < dims.length; i++) {
	    if (dims[i] != null) {
		dims[i].genCode(code);
	    }
	}

	if (type.getArrayBound() > 1) {
	    int		count = 0;
	    for (int i = 0; i < dims.length && dims[i] != null; i++) {
		count++;
	    }

	    code.plantInstruction(
		   new MultiarrayInstruction(type.getSignature(), count));
	} else {
	    if (type.getIndexedType().isReference()) {
		code.plantInstruction(
		       new ClassRefInstruction(opc_anewarray,
					       ((CClassType)type
						.getIndexedType())
					       .qualifiedName()));
	    } else {
		byte	num;

		switch (type.getIndexedType().getTypeID()) {
		case TID_BYTE:
		    num = 8;
		    break;
		case TID_BOOLEAN:
		    num = 4;
		    break;
		case TID_CHAR:
		    num = 5;
		    break;
		case TID_SHORT:
		    num = 9;
		    break;
		case TID_INT:
		    num = 10;
		    break;
		case TID_LONG:
		    num = 11;
		    break;
		case TID_FLOAT:
		    num = 6;
		    break;
		case TID_DOUBLE:
		    num = 7;
		    break;
		default:
		    fail();
		    num = -1;
		}
		code.plantInstruction(new NewarrayInstruction(num));
	    }
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression[] dims() { return dims; }
    public JArrayInitializer init() { return init; }

    /* 
     * By rzakhejm
     */
    public String toString() {
    	StringBuffer out = new StringBuffer();
    	for (int i = 0; i < dims.length; i++) {
			out.append("[" + (dims[i] != null ? dims[i].toString() : "")+ "]");
		}
    	if (init != null)
    		out.append(init.toString());
    	return out.toString();
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return (init!=null && init.isNonNull(context))
		|| (dims!=null && dims.length==1 && dims[0] instanceof JOrdinalLiteral 
			       && ((JOrdinalLiteral) dims[0]).numberValue().longValue() == 0);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	CArrayType		type;
    private	JExpression[]		dims;
    protected	JArrayInitializer	init;
    private	CUniverse		array_universe; // WMD
    private	CUniverse		elem_universe; // WMD
}
