/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JArrayInitializer.java,v 1.10 2008/05/14 18:16:26 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class implements a constant list of expressions used in array
 * initialisation.  The method isAssignableTo is redefined in order to
 * follow (JLS 5.2) */
public class JArrayInitializer extends JExpression {

    /**
     * Constructs a node in the parsing tree.  This method is directly
     * called by the parser.
     *
     * @param where the line of this node in the source code
     * @param elems an array of expressions {e1, e2, e3, e4} */
    public JArrayInitializer( TokenReference where, JExpression[] elems ) {
	super(where);

	this.elems = elems;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression[] elems() { return elems; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return type;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    for (int i = 0; i < elems.length; i++) {
		if (!elems[i].isNonNull(context)) {
		   return false;
		}
	    }
	    return true;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Sets the type of this array initializer.
     * @param	type		the type of this array
     *				assume type was checked
     */
    public void setType(CType type) {
	assertTrue( type.checked() );
	this.type = type;
    }

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context )  
	throws PositionedError {
	assertTrue( type != null );
	assertTrue( type.isArrayType() );


        this.type = (CArrayType)CUniverseServices.setDefaultVariableAgain(this.type);
        if (CTopLevel.getCompiler().universeChecks() &&
                CUniverseServices.isPartial((CClassType)this.type) &&
                elems != null &&
                elems.length > 0 ) {
            // if there is an initializer, we try to use the first element to
            // improve the defaulting.
            // sometimes the type of the initializer expression still is null,
            // eg in:
            //    rep Object po = new rep Object();
            //    Object[] c = { po, "he!" };
            // maybe because the field type was not determined yet.
            // Also, some stupid expressions throw a NPE, because we didn't call
            // typecheck yet.
            try {
                elems[0].typecheck(context);
                if (elems[0].getType() != null) {
                    CArrayType templ = new CArrayType(elems[0].getType(), ((CArrayType) this.type).getArrayBound(), CUniverseImplicitPeer.getUniverse());
                    this.type = CUniverseServices.propagate((CClassType) this.type, templ);
                }
            } catch  (Exception e) {
                // forget about it.
            }
        }
        
        CType elemsType = ((CArrayType) type).getIndexedType();

	if( elemsType.isArrayType() ) {
	    for( int i = 0; i < elems.length; i++ ) {
		if( elems[i] instanceof JArrayInitializer) {
		    ((JArrayInitializer) elems[i]).setType( elemsType );
		}
		elems[i] = elems[i].typecheck(context);
		check( context, 
		       elems[i].isAssignableTo( elemsType ),
		       MjcMessages.ARRAY_INIT_BADTYPE, 
		       elemsType.toVerboseString(), 
		       elems[i].getType().toVerboseString() );
	    }
	} else {
	    for( int i = 0; i < elems.length; i++ ) {
		elems[i] = elems[i].typecheck( context );
		check( context, 
		       elems[i].isAssignableTo(elemsType),
		       MjcMessages.ARRAY_INIT_BADTYPE, 
		       elemsType.toVerboseString(), 
		       elems[i].getType().toVerboseString() );
		elems[i] = elems[i].convertType( elemsType, context );
	    }
	}

	return this;
    }

    
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitArrayInitializer( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	JArrayDimsAndInits.allocArray(code,
				      (CArrayType)getType(),
				      new JExpression[] { new JOrdinalLiteral(getTokenReference(),
									      elems.length,
									      CStdType.Integer)});
	int		opcode;

	switch (((CArrayType)getType()).getIndexedType().getTypeID()) {
	case TID_CLASS:
	case TID_ARRAY:
	    opcode = opc_aastore;
	    break;
	case TID_INT:
	    opcode = opc_iastore;
	    break;
	case TID_LONG:
	    opcode = opc_lastore;
	    break;
	case TID_FLOAT:
	    opcode = opc_fastore;
	    break;
	case TID_DOUBLE:
	    opcode = opc_dastore;
	    break;
	case TID_BYTE:
	case TID_BOOLEAN:
	    opcode = opc_bastore;
	    break;
	case TID_CHAR:
	    opcode = opc_castore;
	    break;
	case TID_SHORT:
	    opcode = opc_sastore;
	    break;
	default:
	    fail();
	    opcode = -1;
	}

	for (int i = 0; i < elems.length; i++) {
	    //if (i != elems.length - 1)
	    code.plantInstruction(new NoArgInstruction(opc_dup));
	    code.plantInstruction(new PushLiteralInstruction(i));
	    elems[i].genCode(code);
	    code.plantInstruction(new NoArgInstruction(opcode));
	}
    }
    
    
    /* 
     * By rzakhejm
     */
    public String toString() {
    	StringBuffer out = new StringBuffer("{");
    	for (int i = 0; i < elems.length; i++) {
			out.append(elems[i].toString());
			if (i < elems.length - 1)
				out.append(", ");
		}
    	return out.append("}").toString();
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	CType		type;
    private	JExpression[]	elems;
}
