/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JArrayLengthExpression.java,v 1.9 2006/08/04 18:22:38 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents an array length expression in the AST.
 */
public class JArrayLengthExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	prefix		the left expression like t.this
     */
    public JArrayLengthExpression( TokenReference where, JExpression prefix ) {
	super( where );
	this.prefix = prefix;
     }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression prefix() { return prefix; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return CStdType.Integer;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * prj 04 August 2006 this code does not seem to be called.
     * the only place it would be is JNameExpression.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	assertTrue( prefix != null );
	prefix = prefix.typecheck( context );
	check( context, 
	       prefix.getType().isArrayType(), 
	       MjcMessages.ARRAY_LENGTH_BADTYPE );

	//check nullity
	return typecheckNullity( context );
    }
    public JExpression typecheckNullity( CExpressionContextType context ) 
	throws PositionedError 
    {
	if (CTopLevel.getCompiler().nonNullTypes()) {
	   check( context,
                  prefix.isNonNull( context ),
                  MjcMessages.FIELDACCESS_NULL_REF,
                  "length" );
	}


	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitArrayLengthExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());
	boolean	discardValue = code.discardValue();

	code.setDiscardValue(false);
	prefix.genCode(code);
	if (discardValue) {
	    code.plantInstruction(new NoArgInstruction(opc_pop));
	} else {
	    code.plantInstruction(new NoArgInstruction(opc_arraylength));
	}
    }
    
	/*
	 * By rzakhejm
	 */
	public String toString() {
			return prefix().toString() + ".length";
	}
	
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	JExpression	prefix;

}
