/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JAssertStatement.java,v 1.14 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * The syntax for Java assert statements is defined as follows.
 *
 * <pre>
 *  assert-statement ::= "assert" expression [ ":" expression ] ";"
 * </pre>
 *
 * @author Curtis Clifton
 * @version $Revision: 1.14 $
 */

public class JAssertStatement extends JStatement {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public JAssertStatement( TokenReference where, JExpression predicate,
			     JavaStyleComment[] comments ) {
	this( where, predicate, null, comments );
    }
    
    public JAssertStatement( TokenReference where, JExpression predicate,
			     JExpression throwMessage, 
			     JavaStyleComment[] comments ) {
	super( where, comments );
	this.predicate = predicate;
	this.throwMessage = throwMessage;
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression predicate() {
	return predicate;
    }

    public JExpression throwMessage() {
	return throwMessage;
    }
    
    public boolean hasThrowMessage() {
	return throwMessage != null;
    }

    //---------------------------------------------------------------------
    // DERIVED ATTRIBUTES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // INTERFACE CHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // TYPECHECKING
    //---------------------------------------------------------------------

    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {

	    // get the directly enclosing class.	
	    // NOTE: If this 'assert' is inside a method belonging to
	    // an external gf, currCls below will not be the directly
	    // enclosing class of the method.  This is because the
	    // dispatcher classes have yet to be created.  The
	    // consequence is that a dispatcher class will have
	    // assertions enabled iff its surrounding class does (an
	    // anchor class, or a regular class containing an internal
	    // method of an external gf).
	CClass currCls = context.getClassContext().getHostClass();
	
	if (currCls.noAssertionsYet()) {
		// this is the first assertion we've seen in the directly
		// enclosing class, so we have some work to do first...
	    createSyntheticFieldsAndMethods(currCls, context);
	}
	    
	    /* now create the desugared statement, which looks as follows:
	       do {
	         if (!(<assertDisabledName> || <predicate>))
		   throw new java.lang.AssertionError(<throwMessage>);
	       } while (false);

	       the do loop is used just to convert the if statement to
	       the kind of statement of asserts
	       (StatementWithoutTrailingSubstatement)  See JLS2 14.5.
	       Without this "hack" an assert statement inside the "then" 
	       clause of an if-else would capture the dangling else.
	    */
	
	TokenReference tokenRef = getTokenReference();
	
	JExpression orExpr = new JConditionalOrExpression(
	    tokenRef,
	    new JNameExpression(tokenRef, assertDisabledName),
	    predicate);
	JExpression ifGuard = new JUnaryExpression(tokenRef, OPE_LNOT,
						   orExpr);

	JExpression[] params;
	if (hasThrowMessage()) {
	    params = new JExpression[1];
	    params[0] = throwMessage;
	}
	else {
	    params = JExpression.EMPTY;
	}	
	JExpression assertionError = new JNewObjectExpression(
	    tokenRef, CTopLevel.getTypeRep(JAV_ASSERTION_ERROR, true),
	    null, params );
	JStatement ifBody = new JThrowStatement(tokenRef, assertionError,
						JavaStyleComment.EMPTY);

	JStatement ifStmt = new JIfStatement( tokenRef, ifGuard, ifBody,
					      null,
					      JavaStyleComment.EMPTY );

	desugared_stmt = new JDoStatement(
	    tokenRef, new JBooleanLiteral(tokenRef, false),
	    ifStmt, getComments() );

	desugared_stmt.typecheck(context);
//	// don't gather nullity information from assert statements
	Object[] nonNulls = predicate.typecheck(context.createExpressionContext()).getFANonNulls(context);
	context.addFANonNulls(nonNulls);
    }


    private void createSyntheticFieldsAndMethods(
	CClass currCls, CFlowControlContextType context) {
	    // find the top-level type in which this assert resides
	CClass topType = currCls;
	while(topType.isNestedType()) {
	    topType = topType.owner();
	}

	    // the class that helper methods and fields for computing
	    // and storing the top-level type's associated Class instance
	    // should be put in.
	CClass helperCls;
	
	    // if the top type is a class, then necessary synthetic
	    // fields and methods for computing the top type's Class
	    // instance go there.  if the top type is an interface,
	    // then we put the stuff in the current class instead of
	    // the top-level type.  javac handles this situation of a
	    // top-level interface slightly differently: it makes a
	    // dummy class to hold the stuff, which can then be shared
	    // by all inner classes of the interface.
	if (topType.isInterface()) {
	    helperCls = currCls;
	} else {
	    helperCls = topType;
	}
	
	if (helperCls.noPossiblyInnerAssertionsYet()) {
		// the helper class has not yet been prepared to handle
		// asserts in its enclosing classes or itself
	    
	    // create a method that computes a String's associated
	    // java.lang.Class instance
	    CMethod assertHelperMeth = new CAssertHelperMethod(helperCls);
	    helperCls.setAssertHelperMethod(assertHelperMeth);

	    // create the field that will hold the result of calling the above
	    // method with the name of the top type.
	    helperCls.addField(new CSourceField(
				   new MemberAccess(helperCls, ACC_STATIC),
				   assertSyntheticPrefix +
				   topType.getQualNameWithSeparator('$'),
				   CStdType.Class, false ) );
	}
	
	    // create the field in the current class that will hold
	    // whether or not the class has assertions disabled
	    CSourceField assertDisFld 
		= new CSourceField( new MemberAccess( currCls, 
						      ACC_STATIC | ACC_FINAL ),
				    assertDisabledName, 
				    CStdType.Boolean, false );
		// make sure to mark the field as having been
		// initialized.  otherwise we can get spurious
		// definite assignment errors when the field is used
		// (as part of the desugaring of asserts)
	    assertDisFld.initializedUnlessFinalInstance();
	    currCls.addField(assertDisFld);

	    
	    // create the static initialization code that sets the
	    // above fields appropriately
	currCls.setAssertStaticInitMethod(
	    new CAssertStaticInitMethod(currCls, topType, helperCls) );
    }
	
	
	

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    public void genCode( CodeSequence code ) {
	desugared_stmt.genCode(code);
    }

    /**
     * Accepts the specified visitor
     * @param p	the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitAssertStatement( this );
    }
    
    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    protected JExpression predicate;
    protected JExpression throwMessage;

	// the desugared version of this assert statement
    protected JStatement desugared_stmt;

    // a prefix used by some of the synthetic fields and methods needed
    // in compiling asserts
    static final String assertSyntheticPrefix = "class$";

	// the name of the field saying whether or not assertions are
	// disabled
    static final String assertDisabledName = "$assertionsDisabled";
}
