/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JAssignmentExpression.java,v 1.20 2008/03/16 15:52:51 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class implements the assignment operation
 */
public class JAssignmentExpression extends JBinaryExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JAssignmentExpression( TokenReference where,
				  JExpression left,
				  JExpression right ) {
	super( where, left, right );
    }

    /**
     * Factory method creates an assignment expression or a compound
     * assignment expression (like <code>x += 3</code>) depending on
     * the assignment operator.  
     *
     * @param	where	the line of this node in the source code
     * @param	oper	the sort of assignment operator
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public static JExpression create( TokenReference where, 
				      int oper, 
				      JExpression left, 
				      JExpression right ) {
	if (oper == OPE_SIMPLE) {
	    return new JAssignmentExpression(where, left, right);
	} else {
	    return new JCompoundAssignmentExpression(where, oper, left, right);
	}
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns true iff this expression can be used as a statement (JLS 14.7)
     */
    public /*@ pure @*/ boolean isStatementExpression() {
	return true;
    }

    /** 
     * Returns the assignment operator of this assignment expression.
     */
    public /*@ pure @*/ int oper() { 
        return OPE_SIMPLE;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return left.isNonNull(context);
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck(CExpressionContextType context) 
	throws PositionedError 
    {
	left = left.typecheck(context, false, true, true);

	// Perform universe checks if enabled
	if (CTopLevel.getCompiler().universeChecks()) {
		typecheckUniverse(context);
	}

	// Assignment must only happen to LValues
    check(context, 
    		left.isMaybeInitializable() 
    		&& ((CInitializable) left).isLValue(context),
    		MjcMessages.ASSIGNMENT_NOTLVALUE );
	
	right = right.typecheck( context, false, false, false );
	
	try {
	    ((CInitializable) left).initialize( context );
	} catch (UnpositionedError e) {
	    throw e.addPosition(getTokenReference());
	}

	if( left.equals(right) ) {
	    context.reportTrouble(
		      new CWarning( getTokenReference(), 
				    MjcMessages.SELF_ASSIGNMENT ));
	}
        if (!right.isAssignableTo(left.getType())) {
            check( context, 
                    false,
                    MjcMessages.ASSIGNMENT_BADTYPE, 
                    right.getType().toVerboseString(), // WMD
                    left.getType().toVerboseString() );
        } else if (right.getType().needsUncheckedConversion(left.getType())) {
            warn(context,
                 MjcMessages.UNCHECKED_CONVERSION,
                 right.getType(),
                 left.getType());
        }

	// check nullity
	if (CTopLevel.getCompiler().nonNullTypes()) {
//if (!CTypeNullity.isAssignable(left, right, context))
//    System.out.println("left is "+left.toString());

            check( context, 
                  CTypeNullity.isAssignable(left, right, context),
	          (left.getType().isArrayType() && CTopLevel.getCompiler().experimentalArrayHandling() ?
	           MjcMessages.ASSIGNMENT_ARRAY_NULLITY_INVARIANT :
                   MjcMessages.ASSIGNMENT_OF_NULL));
	}
	if (!left.isDeclaredNonNull()) {
	   if (right.isNonNull(context)) {
	      context.addFANonNull(left);
	   } else {
	     context.removeFANonNull(left);
	   }
	}

	if( right instanceof JTypeNameExpression ) {
	    check( context, 
		   false, 
		   MjcMessages.VAR_UNKNOWN, 
		   ((JTypeNameExpression) right).qualifiedName().
		   replace('/','.') );
	}

        // We have to apply capture conversion (JLS3 5.1.10) to the type of the
        // assignment expression (see JLS3 15.26).
	try {
	    type = left.getType().getCapture(context);
	} catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
	}
	right = right.convertType( left.getType(), context );

	// scdaniel: universe runtime checks
	while ( CTopLevel.getCompiler().universeDynChecks() ) {
	    arrayStoreTest = null;

	    /*
	     * Create test for array store if the lefthandside is an
	     * array element and the element type of the array
	     * variable is readonly.
	     */ 
	    while ( left instanceof JArrayAccessExpression ) {
		/* get the actual array variable */
		JExpression arrayVar = ((JArrayAccessExpression) left).prefix();
			
		if ( !(arrayVar.getType() instanceof CArrayType) )
		    break;
			
		CType baseType = ((CArrayType) arrayVar.getType()).getBaseType();
		if ( !(baseType instanceof CClassType) )
		    break;
			
		CUniverse universe = ((CClassType) baseType).getCUniverse();
		if ( !(universe instanceof CUniverseReadonly) )
		    break;
			
		/* arguments for call to checkArrayType */
		JExpression args[] = { arrayVar,
		       new JOrdinalLiteral( TokenReference.NO_REF,
					    MjcTokenTypes.LITERAL_readonly,
					    CStdType.Integer) };
			
		/* (array.$type != readonly) */
		JExpression test =
		    new JUnaryExpression( TokenReference.NO_REF,
					  OPE_LNOT,
			  new JMethodCallExpression(TokenReference.NO_REF,
				    CUniverseRuntimeHelper.getHashTable(),
						    "checkArrayType",
						    args, false) );
	    	
		/*
		 * first we do the store, then the check
		 * this way we don't need a temporary variable
		 * if the righthandside creates a new object
		 * (and also does not evaluate expressions on
		 * the righthandside twice, which would be bad)
		 */
	    	args = new JExpression[] {arrayVar, left};
	    	
	    	/* ((array.$type != readonly) &&
		 * (array.$owner != accessedArrayElement.$owner)) */
	    	test = new JConditionalAndExpression(TokenReference.NO_REF,
			   test,
			   new JUnaryExpression(TokenReference.NO_REF,
			       OPE_LNOT,
			       new JMethodCallExpression(TokenReference.NO_REF,
					 CUniverseRuntimeHelper.getHashTable(),
							 "isPeer",
							 args, false)));

	    	/* if the above expression is true, the store was illegal */
		arrayStoreTest =
		    new JIfStatement(TokenReference.NO_REF, test,
		       new JExpressionStatement(TokenReference.NO_REF,
			    new JMethodCallExpression(TokenReference.NO_REF,
					CUniverseRuntimeHelper.getPolicy(),
						      "illegalArrayStore",
						      new JExpression[0],
						      false),
			    null),
		       null, null);

		// temporarily disable universe checks ...
		boolean universeChecks = CTopLevel.getCompiler().universeChecks();
		boolean universePurity = CTopLevel.getCompiler().universePurity();
		CTopLevel.getCompiler().setUniverseChecks(false);
		CTopLevel.getCompiler().setUniversePurity(false);

		// ... typecheck ...
		arrayStoreTest.typecheck(context.getFlowControlContext());

		// ... enable universe checks again.
		CTopLevel.getCompiler().setUniverseChecks(universeChecks);
		CTopLevel.getCompiler().setUniversePurity(universePurity);
		
		break;
	    }
		
	    /* find out if and what JNewObjectExpression is involved */
	    JExpression newObject = right;
		
	    if ( right instanceof JUnaryPromote )
		newObject = ((JUnaryPromote) right).expr();
		
	    /* find out where it is saved and we may access it later */
	    JExpression ident;
	    if ( left instanceof JLocalVariableExpression )
		ident = ((JLocalVariableExpression) left);
	    else if ( left instanceof JClassFieldExpression )
		ident = left;
	    else if ( left instanceof JArrayAccessExpression )
		ident = left;
	    else
		break;
		
	    /* create transformation */
	    newObject.buildUniverseDynChecks(context, ident);
	    
	    break;
	}
	return this;
    }

    /**
     * Ensure static correctness of Generic Universe Types.
     * 
     * @param context The current context
     * 
     * by WMD
     * by RZU
     */
	private void typecheckUniverse(CExpressionContextType context)
			throws PositionedError {
		// Get the type of the prefix
		CClassType prefixType = getPrefixType(left);
		
		// Variable indicating whether we are operating on a readonly reference
		boolean readonlyAccess = prefixType != null && prefixType
				.getCUniverse() instanceof CUniverseReadonly;

		// If the context is pure but the access is not readonly
		// no assignment must take place
		if (CTopLevel.getCompiler().universePurity()
				&& context.isPure() && !readonlyAccess) {
			check(context, 
					left.isMaybeInitializable() 
					&& ((CInitializable) left).isLValue(context),
					CUniverseMessages.ASSIGNMENT_IN_PURE);

		// Otherwise, if the context is not pure but we have a readonly
		// reference, no assignment must take place either
		} else {
			check(context,
					!readonlyAccess,
					CUniverseMessages.ASSIGNMENT_THROUGH_READONLY );
		}
		
		// Ensure that the update either happens on this, or no type in the
		// sequence contain a rep modifier
		boolean cond = prefixType == null
			|| prefixType.getCUniverse() instanceof CUniverseThis
			|| !(left instanceof JClassFieldExpression)
			|| !(CUniverseServices.containsRepModifier(
					((JClassFieldExpression) left).getTypeBeforeViewpointAdaptation() ) );
		// When the condition is false, we know that it is a JClassFieldExpression.
		// The arguments are always evaluated, so check again.
		check(context,
			  cond,
			  CUniverseMessages.REP_NOT_ON_THIS_IN_UPDATE,
			  cond ? "" : ((JClassFieldExpression) left).getField().ident(),
			  cond ? "" : ((JClassFieldExpression) left).getTypeBeforeViewpointAdaptation().toVerboseString());
	}

	/**
	 * Gets the type of a prefix from an expression.
	 * 
	 * This is a convenience method. Its a the wrong place.
	 * 
	 * @return The type of the prefix of the given expression. Null if
	 *         the expression doesn't have a prefix.
	 */
	private CClassType getPrefixType(JExpression expression) {
		if (expression instanceof JClassFieldExpression) {
			return (CClassType) ((JClassFieldExpression) expression)
					.prefix().getType();
		} else if (expression instanceof JArrayAccessExpression) {
			return (CClassType) ((JArrayAccessExpression) expression)
					.prefix().getType();
		} else if (expression instanceof JMethodCallExpression) {
			return (CClassType) ((JMethodCallExpression) left)
					.prefix().getType();
		} else {
			return null;
		}
	}

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitAssignmentExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	boolean	discardValue = code.discardValue();

	code.setDiscardValue(false);
	((CInitializable)left).genStartStoreCode(code);
	right.genCode(code);
	((CInitializable)left).genEndStoreCode(code, discardValue);
	code.setDiscardValue(discardValue);

	/*
	 * scdaniel: universe runtime checks
	 * 
	 * Register newly created objects to the UrtHashtable after
	 * they have been stored into their variable slot.
	 */
	while ( CTopLevel.getCompiler().universeDynChecks() ) {
	    JExpression newObject = right;
		
	    if ( right instanceof JUnaryPromote )
		newObject = ((JUnaryPromote) right).expr();
		
	    /* if ( !(left instanceof JLocalVariableExpression) &&
	            !(left instanceof JClassFieldExpression) &&
	            !(left instanceof JArrayAccessExpression) )
			break;
	    */

	    /* generate code */
	    newObject.genUniverseDynCheckCode(code);
		
	    /*
	     * insert check for assignment to array elements if the
	     * lefthandside is an array element.
	     */
	    if ( arrayStoreTest != null ) {
		discardValue = code.discardValue();
		
		code.setDiscardValue(false);
		arrayStoreTest.genCode(code);
		code.setDiscardValue(discardValue);
	    }
	    break;
	}
    }
    
    // scdaniel: universe runtime checks
    private JIfStatement arrayStoreTest;
}
