/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JBinaryExpression.java,v 1.8 2006/08/29 03:28:19 ye-cui Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class is an abstract root class for binary expressions
 * Here are conversion method following JLS 5.6.2
 */
public abstract class JBinaryExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	left		left operand
     * @param	right		right operand
     */
    public JBinaryExpression( TokenReference where,
			      JExpression left,
			      JExpression right ) {
	super( where );
	this.left = left;
	this.right = right;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public /*@ pure @*/ JExpression left() { return left; }
    public /*@ pure @*/ JExpression right() { return right; }
    public void setLeft(JExpression e) { left = e; }
    public void setRight(JExpression e) { right = e; }

    /**
     * @return the type of this expression
     */
    public /*@ pure @*/ CType getType() {
	return type;
    }

    // ----------------------------------------------------------------------
    // OPERATIONS
    // ----------------------------------------------------------------------

    public void setType(CType t){
	type = t;
    }

    /**
     * Apply binary numeric promotion (JLS 5.6.2)
     * @param	context		the context in which this expression appears
     * @return	the type after promotion
     * @exception	PositionedError	if the promotion fails to typecheck
     */
    public CType promote( CExpressionContextType context ) 
	throws PositionedError 
    {
	CType promoted;

	promoted = ((CNumericType)left.getType()).binaryPromote( right.getType() );

	left = left.convertType( promoted, context );
	right = right.convertType( promoted, context );	

	return promoted;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genBooleanResultCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	CodeLabel		okayLabel = new CodeLabel();
	CodeLabel		nextLabel = new CodeLabel();

	genBranch(true, code, okayLabel);		//		LEFT CODE IFNE okay
	code.plantInstruction(new PushLiteralInstruction(0)); //	FALSE
	code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
	code.plantLabel(okayLabel);			//	okay:
	code.plantInstruction(new PushLiteralInstruction(1)); //	TRUE
	code.plantLabel(nextLabel);			//	next	...
    }

    /**
     * Generates a sequence of bytecodes to branch on a label
     * This method helps to handle heavy optimizables conditions
     * @param	code		the code list
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
	// here, we try to optimize
	left = left.unParenthesize();
	right = right.unParenthesize();

	genBranch(left, right, cond, code, label);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
			     JExpression right,
			     boolean cond,
			     CodeSequence code,
			     CodeLabel label)
    {
	genCode(code);
	code.plantLabelRef(cond ? opc_ifne : opc_ifeq, label);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	CType			type;
    protected	JExpression		left;
    protected	JExpression		right;
}
