/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JBlock.java,v 1.12 2006/12/10 22:51:21 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * A block is a sequence of statements and local variable declaration
 * statements within braces.  JLS 14.2: Block
 */
public class JBlock extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     */
    public JBlock(/*@ non_null @*/ TokenReference where, 
		  /*@ non_null @*/ JStatement[] body, 
		  /*@ non_null @*/ JavaStyleComment[] comments ) {
	super( where, comments );

	this.body = body;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	CFlowControlContextType inside = 
	    context.createFlowControlContext(getTokenReference());

	// scdaniel: universe runtime checks
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     genUniverseDynChecks ) {
	    JExpression args[] =
		{ new JMethodCallExpression(TokenReference.NO_REF,
				       CUniverseRuntimeHelper.getHashTable(),
				       "getRootSetMember",
				       new JExpression[0], false) };

	    // generate method calls
	    setContext = new JMethodCallExpression(TokenReference.NO_REF,
				CUniverseRuntimeHelper.getHashTable(),
						   "setContext",
						   args,
						   false);

	    resetContext = new JMethodCallExpression(TokenReference.NO_REF,
				CUniverseRuntimeHelper.getHashTable(),
						     "resetContext",
						     new JExpression[0],
						     false);
		
	    // temporarily disable universe checks ...
	    boolean universeChecks = CTopLevel.getCompiler().universeChecks();
	    boolean universePurity = CTopLevel.getCompiler().universePurity();
	    CTopLevel.getCompiler().setUniverseChecks(false);
	    CTopLevel.getCompiler().setUniversePurity(false);

	    // ... typecheck ...
	    setContext.typecheck(context.createExpressionContext());
	    resetContext.typecheck(context.createExpressionContext());

	    //added for assignable checking pass
	    cachedExpContext = context.createExpressionContext();

	    // ... enable universe checks again.
	    CTopLevel.getCompiler().setUniverseChecks(universeChecks);
	    CTopLevel.getCompiler().setUniversePurity(universePurity);
	}
	
	if (body != null) {
	    for( int i = 0; i < body.length; i++ ) {
		if( !inside.isReachable() ) {
		    throw new CLineError( body[i].getTokenReference(), 
					  MjcMessages.STATEMENT_UNREACHABLE );
		}

		try {
		    body[i].typecheck( inside );
		} catch( CLineError le ) {
		    context.reportTrouble( le );
		}
	    }
	}

	inside.checkingComplete();
	context.adoptNullityInfo(inside);
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitBlockStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());
	
	// scdaniel: universe runtime checks
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     genUniverseDynChecks ) {
	    setContext.genCode(code);
	}
	
	for (int i = 0; i < body.length; i++) {
	    body[i].genCode(code);
	}

	// scdaniel: universe runtime checks
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     genUniverseDynChecks ) {
	    resetContext.genCode(code);
	}
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public /*@ pure non_null @*/ JStatement[] body() { return body; }

    public void setGenUniverseDynChecks( boolean b ) {
	this.genUniverseDynChecks = b;
    }

    public CExpressionContextType cachedExpContext() {
	return cachedExpContext;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    public static final /*@ non_null @*/ JBlock DUM_BLOCK = 
	new JBlock(org.multijava.util.compiler.TokenReference.NO_REF,
		   new JStatement[0],
		   new JavaStyleComment[0]);

    protected /*@ non_null @*/ JStatement[] body;
    
    // scdaniel: Universe runtime checks
    private boolean genUniverseDynChecks = false;
    private JMethodCallExpression setContext, resetContext;
    private CExpressionContextType cachedExpContext = null;
}
