/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JBooleanLiteral.java,v 1.7 2004/01/22 20:04:41 leavens Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.compiler.TokenReference;

/**
 * Root class for all expressions
 */
public class JBooleanLiteral extends JLiteral {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	value		the value of this literal
     */
    public JBooleanLiteral(TokenReference where, boolean value) {
	super(where);
	this.value = value;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return value ? "true" : "false";
    }
    
    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return CStdType.Boolean;
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before typecheck
     */
    public /*@ pure @*/ boolean booleanValue() {
	return value;
    }

    /**
     * Returns the literal value of an expression
     */
    public /*@ pure @*/ boolean isDefault() {
	return !booleanValue();
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before typecheck
     */
    public Object getValue() {
	return new Integer(value ? 1 : 0);
    }

    /**
     * set the value of this literal
     * @param	newValue		the new value
     */
    public void setValue(boolean newValue) {
	value = newValue;
    }

    /**
     * Does this expression represent a boolean literal?
     */
    public boolean isBooleanLiteral() {
	return true;
    }
    
    
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitBooleanLiteral( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	code.plantInstruction(new PushLiteralInstruction(value ? 1 : 0));
    }

    /**
     * Generates a sequence of bytecodes to branch on a label
     * This method helps to handle heavy optimizables conditions
     * @param	code		the code list
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
	code.setLineNumber(getTokenReference().line());

	if (value == cond) {
	    code.plantLabelRef(opc_goto, label);
	}
    }


    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private boolean		value;
}
