/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCastExpression.java,v 1.20 2008/05/14 18:16:27 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.ClassRefInstruction;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a cast expression '((byte)2)'
 */
public class JCastExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a node in the parsing tree.
     * This method is directly called by the parser.
     * @param   where   the line of this node in the source code
     * @param   expr    the expression to be casted
     * @param   dest    the type of this expression after cast
     */
    public JCastExpression(
            TokenReference where, 
            JExpression expr,
            CType dest) {
        this(where, expr, dest, false);
    }

    /**
     * Constructs a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where	the line of this node in the source code
     * @param	expr	the expression to be casted
     * @param	dest	the type of this expression after cast
     * @param   isSynthetic  whether this is a synthetic cast introduced by
     *                       the compiler (e.g. a generics cast) 
     */
    public JCastExpression( TokenReference where, 
			    JExpression expr, CType dest, boolean isSynthetic) {
	super( where );
	this.expr = expr;
	this.dest = CUniverseServices.setDefaultVariable(dest);
        this.isSynthetic = isSynthetic;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return "(" + dest.toString() + ") " + expr.toString();
    }

    public JExpression expr() { return expr; }
    public void setExpr(JExpression e) { expr = e; }

    /**
     * Returns whether this is a synthetic cast introduced by the compiler.
     * An example of such a cast is a generics cast.
     *
     * @return  Whether this is a synthetic cast introduced by the compiler.
     */
    public boolean isSynthetic() {
        return isSynthetic;
    }

    /**
     * Compute the type of this expression.
     *
     * @return the type of this expression
     */
    public CType getType() {
	return dest;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	return expr.isNonNull(context);
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails
     *
     * WMD: issue a warning if there is a downcast from readonly to peer
     * or rep.
     */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError {
	expr = expr.typecheck( context );

        return typecheckHelper( context );
    }

    protected JExpression typecheckHelper( CExpressionContextType context ) 
	throws PositionedError {
	try {
	    dest = dest.checkType( context );
	    dest = CUniverseServices.setDefaultVariableAgain(dest);
	} catch (UnpositionedError e) {
	throw e.addPosition(getTokenReference());
	}

	// WMD some additional checks and changes
        // RZU - Restructured the part handling universe checks to emit
	//       runtime checks even if static universe checks are off
	//       This is necessary to have runtime checks work with JML
	if (expr.getType() instanceof CClassType
			&& dest instanceof CClassType ) {
	    CClassType destclass = (CClassType)dest;
	    CClassType exprclass = ((CClassType)expr.getType());

	    if ( CTopLevel.getCompiler().universeChecks() ) {
	    	String orig_dest = destclass.toVerboseString();

		    // Casts in a static context must not contain rep modifiers
			check(context,
					!context.isStatic()
					|| !CUniverseServices
							.containsRepModifier(destclass),
					CUniverseMessages.REP_IN_STATIC_FORBIDDEN);

	    	if ( CUniverseServices.isPartial(destclass) ) {
	    		// there is a cast without complete Universe specification
	    		// issue warning and change the universe
	    		destclass = CUniverseServices.propagate(destclass, exprclass);
                        dest = destclass;
	    		context.reportTrouble(new CWarning(getTokenReference(), 
						 CUniverseMessages.INCOMPLETE_CAST,
						 orig_dest,
						 dest.toVerboseString()));
	    	}

	    	/* This is now combined into isPartial/propagate
	    	if (destclass instanceof CArrayType
	    			&& exprclass instanceof CArrayType) {
	    		CArrayType destarray = (CArrayType) dest;
	    		CArrayType exprarray = (CArrayType) exprclass;

	    		if (destarray.getBaseType() instanceof CClassType
	    				&& exprarray.getBaseType() instanceof CClassType
	    				&& ((CClassType)destarray.getBaseType()).getCUniverse() instanceof CUniverseImplicitPeer ) {
	    			// if both are arrays we can also try the element type
	    			cast_msg = true;
	    			((CClassType) destarray.getBaseType())
	    					.setCUniverse(((CClassType) exprarray.getBaseType()).getCUniverse());
	    		}
	    	} */

	    	// WMD: issue a warning if there is a cast from readonly,
	    	// if the dynamic checks are disabled
	    	if (exprclass.getCUniverse() instanceof CUniverseReadonly
	    			&& !(destclass.getCUniverse() instanceof CUniverseReadonly)
	    			&& !CTopLevel.getCompiler().universeDynChecks()) {
	    		context.reportTrouble(new CWarning(getTokenReference(), 
	    				CUniverseMessages.READONLY_CAST,
	    				exprclass.toVerboseString(),
	    				dest.toVerboseString()));
	    	}
	    
	    }

	    /*
	     * scdaniel: universe runtime checks
	     */
	    if (CTopLevel.getCompiler().universeDynChecks() 
	    		&& exprclass.getCUniverse() instanceof CUniverseReadonly
	    		&& !(destclass.getCUniverse() instanceof CUniverseReadonly)) {
	    	/* get reference to a temporary variable */
	    	tmpLocalVar = CUniverseRuntimeHelper.getTmpVariable(context, this, dest);

	    	/* build the test */
	    	JExpression ownerTest = CUniverseRuntimeHelper.getOwnerTest( context,
	    			dest, 
	    			new JLocalVariableExpression(TokenReference.NO_REF, tmpLocalVar),
	    			expr);

	    	/* what to do if the test fails */
	    	JMethodCallExpression m = new JMethodCallExpression( TokenReference.NO_REF,
	    			CUniverseRuntimeHelper.getPolicy(),
	    			"illegalCast",
	    			new JExpression[0], false);

	    	/* complete the test */
	    	test = new JIfStatement( TokenReference.NO_REF,
	    			new JUnaryExpression(TokenReference.NO_REF, OPE_LNOT, ownerTest),
	    			new JExpressionStatement(TokenReference.NO_REF, m, null), null, null);

	    	// temporarily disable universe checks ...
	    	boolean universeChecks = CTopLevel.getCompiler().universeChecks();
	    	boolean universePurity = CTopLevel.getCompiler().universePurity();
	    	CTopLevel.getCompiler().setUniverseChecks(false);
	    	CTopLevel.getCompiler().setUniversePurity(false);

	    	// ... typecheck ...
	    	test.typecheck(context.getFlowControlContext());

	    	// ... enable universe checks again.
	    	CTopLevel.getCompiler().setUniverseChecks(universeChecks);
	    	CTopLevel.getCompiler().setUniversePurity(universePurity);
	    } 
	}

	if (!expr.getType().isCastableTo(dest)) {
    	    check( context, 
    	           false, 
    	           MjcMessages.CAST_CANT,
    	           expr.getType().toVerboseString(), // WMD
    	           dest.toVerboseString() );
        } else if (!isSynthetic() && !dest.isReifiableType()) {
            warn( context, 
                  MjcMessages.UNCHECKED_CAST,
                  dest.toVerboseString());
        }

	if( !expr.getType().isPrimitive() 
	    && expr.getType().implicitlyCastTo(dest) 
	    && expr.getType() != CStdType.Null ) {
	    context.reportTrouble( new CWarning( getTokenReference(), 
				 MjcMessages.UNNECESSARY_CAST,
				 expr.getType().toVerboseString(), // WMD
				 dest.toVerboseString() ));
	}

        try {
            // We have to apply capture conversion (JLS3 5.1.10) to the type of
            // the cast expression (see JLS3 15.16).
            dest = dest.getCapture( context );
        } catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }

	if( expr.isConstant() && expr.getType().isPrimitive() ) {
	    return expr.convertType( dest, context );
	}

	if( !dest.isAlwaysAssignableTo(expr.getType()) ) {
	    return expr.convertType(dest, context);
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitCastExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	expr.genCode(code);

	if (dest.isNumeric()) {
	    ((CNumericType)expr.getType()).genCastTo((CNumericType)dest, code);
	} else if (dest instanceof CClassType) {
	    // scdaniel: universe runtime checks
	    if ( CTopLevel.getCompiler().universeDynChecks() &&
		 (test != null) ) {
	    	/* Store expression in temporary variable ... */
	    	tmpLocalVar.genStore(code);
	    	/* ... run test ... */
	    	test.genCode(code);
	    	/* ... put expression back onto the stack. */
	    	tmpLocalVar.genLoad(code);
	    }
	    code.plantInstruction(new ClassRefInstruction(opc_checkcast, ((CClassType)dest).qualifiedName()));
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected JExpression	expr;

    // dest can be null in instances of JmlCastExpressions
    protected CType		dest;

    /** Whether this is a synthetic cast introduced by the compiler. */
    boolean                     isSynthetic;

    // scdaniel: universe runtime checks
    private JIfStatement        test;
    private JLocalVariable      tmpLocalVar;
}
