/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCatchClause.java,v 1.9 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.HandlerInfo;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents a catch clause in the syntax tree.
 */
public class JCatchClause extends JPhylum {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	exception	the exception to be caught
     * @param	body		the block to be executed if the exception is
     *				caught
     */
    public JCatchClause(TokenReference where, 
			JFormalParameter exception, 
			JBlock body) {
	super(where);

	this.exception = exception;
	this.body = body;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JFormalParameter exception() { return exception; }
    public JBlock body() { return body; }

    /**
     * getType
     * @return	the type of exception catched by this clause
     */
    public CClassType getType() {
	return (CClassType)exception.getType();
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Indicates that this catch clause has been reached during
     * control flow analysis.  */
    public boolean isReached() {
	return reached;
    }

    /**
     * Checks the type of the caught exception.  Called during the
     * typecheck pass for a try-catch statement.
     *
     * @param	context		the actual context of analysis
     * @exception PositionedError if the check fails */
    public void checkInterface(CFlowControlContextType context) 
	throws PositionedError
    {
	exception.checkInterface( context.getClassContext() ); 
	
	// This checks that the exception was not dynamically
	// dispatched.  More checking is done in later passes
	check( context,
	       !exception.isSpecialized(),
	       MjcMessages.DISPATCHED_EXCEPTION,
	       exception );


	// WMD
	// if it was rep or peer, then change it to readonly
	// and issue a warning
	if( CTopLevel.getCompiler().universeChecks() && 
	    ! (((CClassType) exception.getType()).getCUniverse() 
	       instanceof CUniverseReadonly) ) {

	    warn(context, CUniverseMessages.CATCH_PARAMETER_NOW_READONLY,
		 ((CClassType)exception.getType()).toVerboseString());

	    ((CClassType) exception.getType()).setCUniverse(
  		       CUniverseReadonly.getUniverse() );
	}

    }

    /**
     * Typechecks this catch clause and modifies the context to
     * indicate that this catch clause is reachable.
     * @param	context		the context in which this clause appears
     * @exception PositionedError if the check fails */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	reached = true;

	CFlowControlContextType inside = 
	    context.createFlowControlContext(1,getTokenReference());

	inside.setReachable( true );
	exception.typecheck( inside );
	inside.addFANonNull( exception );
	body.typecheck( inside );

	inside.checkingComplete();

	if( body.body().length == 0 ) {
	    context.reportTrouble(new CWarning(getTokenReference(),
					       MjcMessages.EMPTY_CATCH_BLOCK));
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitCatchClause( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	catchPC = code.getPC();
	exception.genStore(code);
	body.genCode(code);
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genExceptionHandler(CodeSequence code, int startPC, int endPC) {
	code.addExceptionHandler(new HandlerInfo(code.getInstructionAt(startPC),
						 code.getInstructionAt(endPC),
						 code.getInstructionAt(catchPC),
						 exception.getType().getCClass().qualifiedName()));
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JFormalParameter	exception;
    private JBlock		body;

    /**
     * Indicates that this catch clause has been reached during
     * control flow analysis.  */
    private boolean		reached;
    private int			catchPC;
}
