/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCharLiteral.java,v 1.10 2004/01/22 20:04:41 leavens Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.NumberParser;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * A simple character constant
 */
public class JCharLiteral extends JOrdinalLiteral {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	image		the string representation of this literal
     */
    public JCharLiteral(TokenReference where, String image) {
	super(where, image);
	value = new Character(image.charAt(0));
    }


    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	val		the value of this literal
     */
    public JCharLiteral(TokenReference where, char val) {
	super(where, "");

	value = new Character(val);
    }

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	val		the value of this literal
     */
    public JCharLiteral(TokenReference where, Object val) {
	super(where, "");
	if( val instanceof Character) {
	    value = (Character)val;
	} else {
	    value = new Character((char)((Number)val).intValue());
	}
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return "'" + escapeString(value.toString()) + "'";
    }
	
    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return CStdType.Char;
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before typecheck
     */
    public Number numberValue() {
 	return new Integer(value.charValue());
    }

    /**
     * get a Character that represents the value of this literal
     * Do not call before typecheck
     */
    public Object getValue() {
	return value;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck(CExpressionContextType context) 
	throws PositionedError 
    {
	if( image == null ) {
	    return this;
	}
	if( image.startsWith("\\u") ) {
	    value = new Character( (char) NumberParser
				   .decodeHexInt( false, image.substring(2) ));
	} else if( image.startsWith("\\") && image.length() > 1 ) {
	    if( (image.charAt(1) >= '0') && (image.charAt(1) <= '7') ) {
		value = new Character( (char)NumberParser
				       .decodeOctInt( false, 
						      image.substring(1) ));
		check( context, value.charValue() <= 0377,
		       MjcMessages.INVALID_OCTAL_CHAR, image );
	    } else if( image.equals("\\b") ) {
		value = new Character('\b');
	    } else if( image.equals("\\r") ) {
		value = new Character('\r');
	    } else if( image.equals("\\t") ) {
		value = new Character('\t');
	    } else if( image.equals("\\n") ) {
		value = new Character('\n');
	    } else if( image.equals("\\f") ) {
		value = new Character('\f');
	    } else if( image.equals("\\\"") ) {
		value = new Character('\"');
	    } else if( image.equals("\\\'") ) {
		value = new Character('\'');
	    } else if( image.equals("\\\\") ) {
		value = new Character('\\');
	    } else {
		check( context, false, 
		       MjcMessages.INVALID_ESCAPE_SEQUENCE, image );
	    }
	} else {
	    value = new Character( image.charAt(0) );
	}

	image = null;
	return this;
    }

    /**
     * convertType
     * changes the type of this expression to an other
     * @param  dest the destination type
     */
    public JExpression convertType( CType dest, CExpressionContextType context ) 
	throws PositionedError 
    {
	JExpression result;

	if( dest == CStdType.Char ) {
	    result = this;
	} else if( dest.isOrdinal() ) {
	    long v = value.charValue();
	    // Adjust value, if required:
	    v = dest == CStdType.Integer ? (int) v
		: dest == CStdType.Short ? (short) v
		: dest == CStdType.Byte  ? (byte) v
		: dest == CStdType.Char  ? (char) v
		: v;
	    result = new JOrdinalLiteral( getTokenReference(),
					  v,
					  (CNumericType) dest );
	} else if( dest.isFloatingPoint() ) {
	    Number n = (dest == CStdType.Float) 
		? (Number) new Float((float) value.charValue())
		: (Number) new Double((double) value.charValue());
	    result = new JRealLiteral(getTokenReference(), n, dest);
	} else {
	    fail( "Attempting to convert char literal to " + dest );
	    result = null;
	} 
	return result;
    }

    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2].  This considers both widening and
     * narrowing conversions.
     * @param	dest		the destination type
     * @return true iff assignment is valid */
    public boolean isAssignableTo(CType dest) {
	// check widening conversion
	if( CStdType.Char.isAlwaysAssignableTo( dest ) ) {
	    return true;
	}

	// check narrowing conversion [JLS2 5.2]
	if( dest.type >= TID_INT ) {
	    return false;
	} else {
	    // if no information loss results, then narrowing
	    // conversion is permitted
	    int val = value.charValue();
	    if (dest == CStdType.Byte) {
		return (byte)val == val;
	    } else if (dest == CStdType.Short) {
		return (short)val == val;
	    }
	    return false;
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
//  	if( value != null) {
//  	    p.visitCharLiteral(value.charValue());
//  	} else {
//  	    p.visitCharLiteral(image.charAt(0)); // !!! false
//  	}
	p.visitCharLiteral( this );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	Character		value;
}
