/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JClassDeclaration.java,v 1.34 2008/03/16 15:52:51 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Iterator;
import org.multijava.javadoc.JavadocComment;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CompilationAbortedError;

/**
 * This class represents a java class in the syntax tree
 */
public class JClassDeclaration extends JTypeDeclaration 
    implements JClassDeclarationType
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a class declaration in the parsing tree.
     * @param   where           the line of this node in the source code
     * @param   modifiers       the list of modifiers of this class
     * @param   ident           the short name of this class
     * @param   typevariables   the type variables of this class
     * @param   superType       the type this class's superclass
     * @param   interfaces      the names of this types's interfaces
     * @param   methods         a list of JMethodDeclarationTypes giving the
     *                          methods of this type
     * @param   inners          a list of JTypeDeclarationTypes giving the
     *                          inner classes (and interfaces) of this type
     * @param   fieldsAndInits  the fields and initializers of this type, 
     *                          passed together because the order matters 
     *                          for class and object initialization, members 
     *                          of the array should be instances of 
     *                          <code>JFieldDeclarationType</code> or 
     *                          <code>JClassBlock</code>
     * @param   javadoc         javadoc comments including whether this 
     *                          type declaration is deprecated
     * @param   comment         regular java comments
     * @see JMethodDeclarationType
     * @see JFieldDeclarationType
     * @see JClassBlock
     */
    public JClassDeclaration( TokenReference where,
                              long modifiers, 
                              String ident,
                              CTypeVariable[] typevariables,
                              CClassType superType,
                              CClassType[] interfaces,
                              ArrayList methods,
                              ArrayList inners,
                              JPhylum[] fieldsAndInits,
                              JavadocComment javadoc,
                              JavaStyleComment[] comment ) {
        super( where, modifiers, ident, typevariables, interfaces, methods, 
               inners, fieldsAndInits, javadoc, comment );
        
        this.superType = (CClassType) CUniverseServices.setDefaultVariable(superType);
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Sets the super class
     */
    public void setSuperClass(CClassType superType) {
        this.superType = superType;
    }

    /**
     * Sets the interfaces
     */
    public void setInterfaces(CClassType[] interfaces) {
        this.interfaces = interfaces;
    }

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    public void preprocessDependencies( CContextType context )
        throws PositionedError
    {
        CClassContextType self = createContext(context);
        CClass superClass = null;
        //CClassType superType = null;
        
        if (superType == null) {
            if (sourceClass.getType().equals(CStdType.Object)) {
                // java/lang/Object
                // superClass = null;
                // superType = null;
            } else {
                superClass = CStdType.Object.getCClass();
                superType = CStdType.Object;
            }
        } else {
            // find super class
            try {
            	// This is a hack to not crash if a class occurs in an upper bound
            	// of one of its own type variables
            	if (superType instanceof CClassNameType
            			// && !superType.qualifiedName().contains("/")
			        // contains is only available in Java >= 1.5
                                && (superType.qualifiedName().indexOf("/") == -1)
            			&& self.lookupTypeVariable(superType.qualifiedName()) == null) {
        			CClass clazz = context.lookupClass(superType.qualifiedName());
        			if (clazz == null) {
        				throw new UnpositionedError(MjcMessages.TYPE_UNKNOWN, 
        						superType.toString());
            			}
            		
            		((CClassNameType) superType).setClass(clazz);
            		getCClass().setSuperClass(superType);
            		getCClass().setCheckedInterfaces(new CClassType[0]);
            	}
            	
            	superType = (CClassType)(superType.checkType(self));
            	superType = (CClassType)CUniverseServices.setDefaultVariableAgain(superType);
                superClass = superType.getCClass();
                //superClass = CTopLevel.loadClass(superName());
                //superType = superClass.getType();
                //superType.checkType(context);
            } catch (UnpositionedError cue) {
                throw cue.addPosition( getTokenReference() );
            }
        }
        
        if (superClass != null) {
            check(context, superClass.isAccessibleFrom(sourceClass),
//                superClass.isPublic() 
//                || sourceClass.packageName() == superClass.packageName(),
                  MjcMessages.CLASS_ACCESSPARENT, superName());
            check(context, !superClass.isFinal(), 
                  MjcMessages.CLASS_PARENT_FINAL, superName());
            check(context, !superClass.isInterface(), 
                  MjcMessages.CLASS_EXTENDS_INTERFACE, superName());
        }

        if (superType != null) {
            // Top-level wildcard type arguments are not allowed in a supertype
            // (see JLS3 8.1.4).
            check(context,
                  !superType.hasWildcardTypeArgument(),
                  MjcMessages.WILDCARD_IN_SUPER_TYPE,
                  superType.toVerboseString());
        }

        super.preprocessDependencies( context, superType );
    }

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type
     * signatures of everything (imported class files, classes being
     * compiled, methods, fields, etc...)  needed for the later
     * passes.  This information is stored in a CCompilationUnit
     * instance and instances of CMember that are bound to the AST.
     * Also adds things like the default constructor and the initializer
     * method to the AST (these are suppressed during pretty-printing).
     *
     * @return          iff sub tree is correct enought to check code
     * @exception       PositionedError an error with reference to the 
     *                  source file 
     */
    public void checkInterface(final CContextType context) 
        throws PositionedError
    {
        CClassContextType self = createContext(context);

        statInit = constructInitializers(true);
        if (statInit != null) {
            statInit.checkInterface(self);
        }

        // create the default constructor if necessary
        if (!hasConstructor() && getDefaultConstructor() == null) {
            setDefaultConstructor(constructDefaultConstructor());
        }

        instanceInit = constructInitializers(false);
        if (instanceInit != null) {
            instanceInit.checkInterface(self);
        }

        if (getCClass().getSuperClass() != null) {
            CClass clazz = getCClass();
            CClass superClazz = clazz.getSuperClass();
            if (clazz.descendsFrom(superClazz) &&
                superClazz.descendsFrom(clazz)) {
                throw new PositionedError(getTokenReference(), 
                                          MjcMessages.CLASS_CIRCULARITY,
                                          ident );
            }
        }
        super.checkInterface(context);
    }

    /**
     * Returns true if this class declaration contains an 
     * explicit constructor declaration. 
     * This method is used by <code>checkInterface</code> 
     * to determine whether or not to create a default 
     * constructor for this type.
     */
    public boolean hasConstructor() {
        Iterator iter = methods.iterator();
        while( iter.hasNext() ) {
            if (iter.next() instanceof JConstructorDeclarationType) {
                // usually constructors are at the top of the file
                // so its not as inefficient as it looks
                return true;
            }
        }
        return false;
    }

    /**
     * Checks the static initializers created during the
     * checkInterface pass and performs some other checks that can be
     * performed simply before full blown typechecking.
     *
     * @param   context         the context in which this class 
     *                          declaration appears
     * @exception PositionedError if check fails */
    public void checkInitializers(CContextType context) 
        throws PositionedError 
    {
        self = createContext( context );
        
        if (statInit != null) {
            statInit.checkInitializer(self);
            staticInfo = self.fieldInfoTable();
        }

        // Check inners
        Iterator iter = inners.iterator();
        while( iter.hasNext() ) {
            ((JTypeDeclarationType) iter.next()).checkInitializers( self );
        }

        super.checkInitializers( context );
    }

    /**
     * Computes the values of specializer expressions used to dispatch on
     * compile-time constants.
     *
     * @param   context         the context in which this class 
     *                          declaration appears
     * @exception PositionedError if the check fails */
    public void resolveSpecializers(CContextType context)
        throws PositionedError {
        self = createContext( context );
        
        Iterator iter = methods.iterator();
        while( iter.hasNext() ) {
            ((JMethodDeclarationType) iter.next()).
                resolveSpecializers( self );
        }
        iter = inners.iterator();
        while( iter.hasNext() ) {
            ((JTypeDeclarationType) iter.next()).
                resolveSpecializers( self );
        }
    }


    /**
     * Typechecks this type declaration in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context the context in which this type declaration appears
     * @exception PositionedError if any checks fail */
    public void typecheck(CContextType context) throws PositionedError {
        try {
            CVariableInfoTable  instanceInfo = null;
            // For each constructor there is an entry in this list of
            // the definite assignment status of the fields after
            // executing the constructor.  constructors is the data
            // structure that is used to build this list.  After
            // building the list, the data is moved to
            // constructorsInfo for efficiency.
            ArrayList           constructors = 
                getDefaultConstructor() == null 
                ? new ArrayList() : null;
            CVariableInfoTable[] constructorsInfo;

            Iterator            iter;

            // Mark all non-final and static fields as having been initialized.
            JFieldDeclarationType[] fields = fields();
            for( int i = 0; i < fields.length; i++ ) {
                ((CSourceField) fields[i].getField()).
                    initializedUnlessFinalInstance();
            }

            if (instanceInit != null) {
                instanceInit.checkInitializer(self);
                instanceInfo = self.fieldInfoTable();
            }

            // Check the inner classes
            iter = inners.iterator();
            while( iter.hasNext() ) {
                try {
                    ((JTypeDeclarationType) iter.next()).typecheck(self);
                } catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            // Extract the constructors from the methods list and check them. 
            iter = methods.iterator();
            while( iter.hasNext() ) {
                JMethodDeclarationType method = 
                    (JMethodDeclarationType) iter.next();
                try {
                    if (method instanceof JConstructorDeclarationType) {
                        ((JConstructorDeclarationType) method).
                            typecheck(self,instanceInfo);
                        constructors.add(self.fieldInfoTable());
                    }
                } catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            if (getDefaultConstructor() == null) {
                constructorsInfo = (CVariableInfoTable[]) constructors
                    .toArray(new CVariableInfoTable[constructors.size()]);
                
            } else {
                getDefaultConstructor().typecheck(self,instanceInfo);
                constructorsInfo = 
                    new CVariableInfoTable[] {self.fieldInfoTable()};
            }

            // Now we check methods
            iter = methods.iterator();
            while( iter.hasNext() ) {
                JMethodDeclarationType method = 
                    (JMethodDeclarationType) iter.next();
                try {
                    if (!(method instanceof JConstructorDeclarationType)) {
                        method.typecheck(self);
                    }
                } catch (CBlockError e) {
                    context.reportTrouble(e);
                }
            }

            // Now we check for unused private members
            iter = methods.iterator();
            while( iter.hasNext() ) {
                JMethodDeclarationType method = ((JMethodDeclarationType) iter.next());
                if (!((CSourceMethod)method.getMethod()).isUsed() &&
                    !method.getMethod().ident().equals(JAV_CONSTRUCTOR)) {
                    context.reportTrouble(
                        new CWarning( method.getTokenReference(), 
                                      MjcMessages.UNUSED_PRIVATE_METHOD, 
                                      method.getMethod().ident()));
                }
            }

            for( int i = 0; i < fields.length; i++ ) {
                if (!((CSourceField) fields[i].getField()).isUsed()) {
                    context.reportTrouble(
                         new CWarning( fields[i].getTokenReference(), 
                                       MjcMessages.UNUSED_PRIVATE_FIELD, 
                                       fields[i].getField().ident() ));
                }
            }

            

            // mutate the context to record the information gathered
            // by this method, verifies that all necessary
            // declarations exist (for example, implementations of all
            // abstract super methods if this is concrete)
            self.checkingComplete( this, staticInfo, instanceInfo, 
                                   constructorsInfo );
            super.typecheck(context);
        } catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
    
        self = null;
    }

    /**
     * Creates a class context for this class declaration.
     * @param   parent          the parent context or null
     * @return  returns a CClassContextType that represents this context
     */
    public CClassContextType createContext(CContextType parent) {
        return parent.createClassContext( sourceClass );
    }

    /**
     * Builds an AST node representing the default constructor for this 
     * class.
     *
     * @return  an AST node
     */
    protected JConstructorDeclarationType constructDefaultConstructor() {
        long modifiers; 
        CClass clz = getCClass();
        if (clz.isPublic()) {
            modifiers = ACC_PUBLIC;
        } else if (clz.isProtected()) {
            // Only nested classes may be protected
            modifiers = ACC_PROTECTED;
        } else {
            // Only nested classes may be private.  To allow outer
            // classes to instantiate their nested classes we make the
            // nested classes' constructors package-accessible.  See
            // testcase/codegen/InnerProtected1.java for the javac 
            // strategy, which is different.
            modifiers = 0;
        }
        
        return new JConstructorDeclaration( getTokenReference(), modifiers,
                                            ident );
    }

    /**
     * Collects all initializers and builds a single method.
     * @param isStatic  build class (static) or instance initializers?
     */
    protected JInitializerDeclaration constructInitializers(boolean isStatic) {
        ArrayList       elems = new ArrayList();
        boolean needGen = false;

        for( int i = 0; i < fieldsAndInits.length; i++ ) {
            if( (fieldsAndInits[i] instanceof JClassBlock)
                && 
                (((JClassBlock) fieldsAndInits[i]).isStaticInitializer() == 
                 isStatic) ) {
                elems.add( fieldsAndInits[i] );
                needGen = true;
            } else {
                if( (fieldsAndInits[i] instanceof JFieldDeclarationType)
                    && 
                    (((JFieldDeclarationType) fieldsAndInits[i]).variable().
                     isStatic() == isStatic) ) {
                    needGen |= ((JFieldDeclarationType) fieldsAndInits[i]).
                        needInitialization();
                    elems.add(
                        new JClassFieldDeclarator( getTokenReference(), 
                                                   (JFieldDeclarationType) 
                                                   fieldsAndInits[i]) );
                }
            }
        }

        JStatement[] stmts = new JStatement[elems.size()];
        elems.toArray( stmts );

        
        if (stmts.length > 0) {
            return new JInitializerDeclaration(getTokenReference(),
                                               new JClassBlock(
                                                     getTokenReference(), 
                                                     false, stmts ),
                                               isStatic,
                                               !needGen);
        } else {
            return null;
        }
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param   p               the visitor
     */
    public void accept(MjcVisitor p) {
        p.visitClassDeclaration( this );
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    //@ pure
    public String ident() { return getCClass().ident(); }

    public String superName() { 
        if (superType != null) {
            String qn = superType.qualifiedName();
            if (qn != null) {
                return qn;
            }
            return superType.getCClass().qualifiedName();
        } else {
            return null;
        }
    }
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private     CVariableInfoTable              staticInfo;
    protected   CClassContextType               self;
    protected   CClassType                      superType;
}
