/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JClassExpression.java,v 1.9 2004/06/30 09:04:38 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.FieldRefInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents an AST node for a class literal expression,
 * e.g., <code>int.class</code> and <code>String.class</code>.  See
 * JLS2 15.8.2.  */
public class JClassExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param where 	the line of this node in the source code
     * @param prefixType	the type of the primitive preceding the '.' in 
     *			<code>int.class</code>
     */
    public JClassExpression( TokenReference where, CType prefixType ) {
	super( where );
	this.prefixType = prefixType;
    }

    /**
     * Construct a node in the parsing tree.
     * @param where 	the line of this node in the source code
     * @param expr	the expression preceding the '.' in 
     *			<code>int.class</code>
     * @param bounds	the number of dimensions, i.e., <code>[]</code>,
     *			between the expression preceding the '.' and the
     *			'.' in an expression like <code>String[].class</code>
     */
    public JClassExpression( TokenReference where, 
			     JExpression expr, int bounds) {
	super( where );

	// expr must represent a type name, JLS2 6.5.1, "A name is
	// syntactically classified as a type name as the type in a
	// class literal."
	prefixType = CTopLevel.getTypeRep( 
	    ((JNameExpression) expr).qualifiedName().replace('.', '/'),false );
	prefixType = bounds > 0 ? 
	    new CArrayType(prefixType, bounds, null) : prefixType;
	// WMD TODO do we ever get a universe here?
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public CType prefixType() { return prefixType; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return CStdType.Class;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck(CExpressionContextType context) 
	throws PositionedError 
    {
	try {
	    prefixType = prefixType.checkType( context );
	} catch (UnpositionedError e) {
	    throw e.addPosition( getTokenReference() );
	}
	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitClassExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	String		className = "UNKNOWN";
	switch (prefixType.getTypeID()) {
	case TID_CLASS:
	case TID_ARRAY: // $$$ may be optimized and exception changed
	    code.plantInstruction(new PushLiteralInstruction(((CClassType)prefixType).qualifiedName().replace('/', '.')));
	    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
							   "java/lang/Class",
							   "forName",
							   "(Ljava/lang/String;)Ljava/lang/Class;"));
	    return;
	case TID_INT:
	    className = "java/lang/Integer";
	    break;
	case TID_LONG:
	    className = "java/lang/Long";
	    break;
	case TID_FLOAT:
	    className = "java/lang/Float";
	    break;
	case TID_DOUBLE:
	    className = "java/lang/Double";
	    break;
	case TID_BYTE:
	    className = "java/lang/Byte";
	    break;
	case TID_BOOLEAN:
	    className = "java/lang/Boolean";
	    break;
	case TID_CHAR:
	    className = "java/lang/Character";
	    break;
	case TID_SHORT:
	    className = "java/lang/Short";
	    break;
	}

	code.plantInstruction(new FieldRefInstruction(opc_getstatic,
						      className,
						      "TYPE",
						      CStdType.Class.getSignature()));
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	CType		prefixType;
}
