/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JClassFieldExpression.java,v 1.38 2008/03/16 15:52:51 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the AST node for a field access, e.g.
 * <code>getExpression().t</code> or <code>t</code>.  */
public class JClassFieldExpression extends JExpression 
    implements CInitializable {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param where 	 the line of this node in the source code
     * @param prefix 	 the expression preceding the field name
     * @param ident 	 the field name
     * @param sourceName the name of this reference as originally
     *                   appeared in source file. It may be null if generated.
     *
     * <pre><jml>
     * requires prefix != null;
     * ensures !prefixWasBlank;
     * </jml></pre>
     *
     */
    public JClassFieldExpression(TokenReference where,
				 JExpression prefix,
				 String ident,
                                 JNameExpression sourceName)
    {
	super(where);
	assertTrue( prefix != null );
	this.prefix = prefix;
	this.ident = ident;
	prefixWasBlank = false; // perhaps, can eliminate now?
        this.sourceName = sourceName;
    }

    /**
     * Construct a node in the parsing tree.
     * @param	where	the line of this node in the source code
     * @param	ident 	the field name
     * @param sourceName the name of this reference as originally 
     *                   appeared in source file. It may be null if generated.
     *
     * <pre><jml>
     * ensures prefixWasBlank;
     * </jml></pre>
     *
     */
    public JClassFieldExpression(TokenReference where, String ident,
                                 JNameExpression sourceName) {
	super(where);
	this.prefix = null;
	this.ident = ident;
	prefixWasBlank = true;
        this.sourceName = sourceName;
    }

    /**
     * Construct a node in the parsing tree.
     * @param where 	 the line of this node in the source code
     * @param prefix 	 the expression preceding the field name
     * @param ident 	 the field name
     *
     * <pre><jml>
     * requires prefix != null;
     * ensures !prefixWasBlank;
     * </jml></pre>
     *
     */
    public JClassFieldExpression(TokenReference where,
				 JExpression prefix,
				 String ident) {
        this(where, prefix, ident, null);
    }

    /**
     * Construct a node in the parsing tree.
     * @param	where	the line of this node in the source code
     * @param	ident 	the field name
     *
     * <pre><jml>
     * ensures prefixWasBlank;
     * </jml></pre>
     *
     */
    public JClassFieldExpression(TokenReference where, String ident) {
	this(where, ident, null);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------
    
    /**
     * Returns the name of this field reference as originally appeared
     * in the source file. The return value may be null if this field
     * reference was built by the system.
     */
    public /*@ pure @*/ JNameExpression sourceName() {
        return sourceName;
    }

    /**
     * Returns true if the prefix was blank.
     * 
     * <pre><jml>
     * ensures \result == prefixWasBlank;
     * </jml></pre>
     */
    public /*@ pure @*/ boolean prefixWasBlank() {
        return prefixWasBlank;
    }

    public JExpression prefix() { return prefix; }

    public /*@ pure @*/ String ident() {
	return ident;
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public /*@ pure @*/ CType getType() {
	return fieldtype;
    }
    
    /**
     * Returns the type before performing viewpoint adaptation.
     * 
     * This method is specific to Generic Universe Types.
     * 
     * @return Type before viewpoint adaptation. Null if not yet performed
     *         or not applicable.
     */
    public /*@ pure @*/ CClassType getTypeBeforeViewpointAdaptation() {
    	return fieldTypeBeforeViewpointAdaptation;
    }
    
    public void setType(CType type) {
	this.fieldtype = type;

	// WMD check: should we reset the universe modifier?
	// maybe we need to clone the type and modify it again.
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return isDeclaredNonNull() || isFinalNonNull() || context.isFANonNull(field);
    }

    public /*@pure*/ boolean isDeclaredNonNull() {
        return field.getField().isDeclaredNonNull();
    }

    private /*@pure*/ boolean isFinalNonNull() {
	    boolean retVal = false;
	    JExpression value = field.getValue();
	    try { retVal = field.isFinal() ;
		  retVal = retVal      && value != null;
	     	  retVal = retVal      && value.isNonNull(null);
	    }
	    catch (NullPointerException npe) {
		  retVal = false;
	    }
	    return retVal;
    }

    /**
     * @return is this expression constant ?
     */
    public /*@ pure @*/ boolean isConstant() {
	return field.isFinal() 
	    && field.getValue() != null 
	    && field.getValue().isConstant();
    }

    /**
     * Indicates whether this expression can appear on the left-hand
     * side of an assignment or in an increment or decrement
     * operation.
     *
     * <pre><jml>
     * also ensures true;
     * </jml></pre>
     *
     */
    public boolean isMaybeInitializable() {
	return true;
    }

    /**
     * Returns true if this field accepts assignment.
     *
     * WMD: additional checks for the universe modifiers and whether 
     * we are in a pure context are necessary.
     */
    public boolean isLValue(CExpressionContextType context) {
	//  !FIXME!10/26 something is funky here, when the field is a
	//  binary field or is being assigned to from a context other
	//  than its owners, we permit the assignment, regardless of
	//  things like final status or actual state of the field.
//  	if( !field.isFinal() || !(field instanceof CSourceField) ) {
//  	    return true;
//  	} else if( context.findNearestHost() == field.owner()
//  		   && !((CSourceField)field).isDefinitelyAssigned() ) {
//  	    return context.isFieldDefinitelyAssigned( ((CSourceField) field)
//  						      .getPosition() );
//  	} else {
//  	    return true;
//  	}

	// WMD
	if( CTopLevel.getCompiler().universePurity() &&
	    context.isPure() ) {

	    if( prefix instanceof JThisExpression &&
		context.isInConstructor() &&
		!field.isFieldStatic() ) {

		// assignments to instance variables are allowed
		// in a pure constructor
		return true;
	    }
		/*
		  if( prefix instanceof JLocalVariableExpression &&
		    field.owner() == context.findNearestHost() ) {
		    // local variables are also allowed
		    // second part makes sure that no fields of
		    // outer classes are changed
		    return true;
		    }
		*/

	    return false;
	}

	return true;
    }

    /**
     * Returns true if this field is already initialized
     */
    public boolean isDefinitelyAssigned( CContextType context ) {
	if(context!=null){
	CMethodContextType mContext = context.getMethodContext();
	if( !(field instanceof CSourceField) || 
	    (!prefixWasBlank && !(prefix instanceof JThisExpression)) ||
	    !(context.isInInitializer() || context.isInConstructor()) ||
	    (context.isInInitializer() && field.ident().indexOf('$') >= 0)) { //  !FIXME!1/22 hack
	    // We are not able to determine whether the field was
	    // assigned to because it might belong to a different
	    // object than "this" or because we are in a regular
	    // method (as opposed to a constructor).  So we assume
	    // that the field is definitely assigned.  Also, the check
	    // for "$" is a hack so the synthetic outer fields can be
	    // used in initializers even though the typechecking of
	    // initializers assumes that no local fields have been
	    // assigned to.  The check for "$" also means that field
	    // getter/setter methods vacuously pass the check for
	    // definite assignment
	    return true;
	} else if( context.findNearestHost() == field.owner() 
		   && !((CSourceField)field).isDefinitelyAssigned() ) {
	    return context.isFieldDefinitelyAssigned( 
		((CSourceField) field).index() );
	} else {
	    return true;
	}
	}else return false;
    }

    /**
     * Records that this field is initialized in the given context.
     *
     * @exception UnpositionedError if this is a final field and has
     * already been assigned to */
    public void initialize( CContextType context ) 
	throws UnpositionedError
    {
	if( (isDefinitelyAssigned( context )) && field.isFinal() ) {
	    throw new UnpositionedError( MjcMessages.REASSIGNMENT_FINAL_FIELD, 
					 ident() );
	}
	if( (field instanceof CSourceField) 
	    && (context.findNearestHost() == field.owner() 
		&& !((CSourceField) field).isDefinitelyAssigned())) {
	    context.initializeField( (CSourceField) field );
	}
    }

    /**
     * Returns the CField represented by this expression.
     */
    public /*@ pure @*/ CFieldAccessor getField() {
	return field;
    }

    public void setField(CFieldAccessor field){
	this.field = field;
    }

    /**
     * Returns the literal value of this field.  Will fail with a
     * class-cast exception if this field is not representable as a
     * literal. */
    public /*@ pure @*/ JLiteral getLiteral() {
	return (JLiteral)field.getValue();
    }

    public String toString() {
	return (prefix != null ? prefix.toString() + "." : "") +
	    ident;
    }
    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context )  
	throws PositionedError 
    {
	CMemberHost localHost = context.findNearestHost();
	CClass local = context.getClassContext().getHostClass();
	findPrefixAndField( localHost, context );
	checkAccess( localHost, context );
	
	if (context.isInConstructor() || context.isInInitializer()) {
	    check( context,
		   isDefinitelyAssigned(context)
		   || (context.isLeftSide() && context.discardValue()),
		   MjcMessages.UNINITIALIZED_VARIABLE, field.ident());
	}

	check( context,
	       field.isFieldStatic() ||
	       !(prefix instanceof JTypeNameExpression),
	       MjcMessages.FIELD_NOSTATIC, ident);

	if( field.isFieldStatic() &&
	    !(prefix instanceof JTypeNameExpression)) {
	    context.reportTrouble(
		new CWarning( getTokenReference(),
			      MjcMessages.INSTANCE_PREFIXES_STATIC_FIELD,
			      ident ));
	}

	// WMD: at the moment we do not really support static fields!
	if( CTopLevel.getCompiler().universeChecks() && 
	    field.isFieldStatic() &&
	    field.getType() instanceof CClassType ) {

	    CClassType ft = (CClassType) field.getType();
	    CClassType roft = (CClassType) CUniverseServices.combine(CUniverseReadonly.getUniverse(),
	    									ft);
	    
	    if( !ft.equals(roft) ) {
	    	context.reportTrouble(
	    		new CWarning(getTokenReference(),
    					CUniverseMessages.ACCESS_TO_STATIC_FIELD,
    					field.ident() ) );
	    }
	}


	if( field instanceof CSourceField && !context.discardValue() ) {
	    ((CSourceField) field).setUsed();
	}

	// check nullity of prefix
	if (CTopLevel.getCompiler().nonNullTypes()) {
	   check(context, 
                 prefix.isNonNull(context) || JMethodCallExpression.isInnerThis(prefix), 
                 MjcMessages.FIELDACCESS_NULL_REF, ident);
	}
	
	if( isConstant() ) {
	    return field.getValue();
	} else {
	    setType(field.getType());

	    // Unless we are on the LHS of an assignment, we have to apply
	    // capture conversion (JLS3 5.1.10) to the type of a field access
	    // expression (see JLS3 6.5.6).
	    if (!context.isLeftSide()) {
	        try {
	            setType(getType().getCapture(context));
	        } catch (UnpositionedError e) {
	            throw e.addPosition(getTokenReference());
	        }
	    }

	    // WMD determine universe type
	    if( CTopLevel.getCompiler().universeChecks() ) {
	    	typecheckUniverse( context );
	    }
	    //   System.out.println("we have: " + fieldtype.toVerboseString());

            handlePossibleRawTypeAccess(context);

	    if (prefix instanceof JThisExpression) {
		return this;
	    }
	    //if field type is typevariable , cast it to proper subtype
	    if(fieldtype.isTypeVariable()) {
	    	CClassType subtype = prefix.getType().getCClass().getSubstitution(
	    			(CTypeVariable) fieldtype,
	    			prefix.getType().isTypeVariable()
	    					? ((CTypeVariable) prefix.getType()).getBounds()[0].getAllArguments()
	    					: prefix.getType().getAllArguments());
		try{
                    // Avoid inserting a cast on the LHS of an assignment.
		    if(!context.isLeftSide()
                       && !(field.getType().getErasure(context).equals
			 (subtype.getErasure(context)))) {
			JCastExpression castexpr =
			    new JCastExpression(getTokenReference(),
						new JCheckedExpression(getTokenReference(),
								       this),
						subtype,
                                                true);
			castexpr.typecheck(context);
			return castexpr;
		    } else {
		        setType(subtype);
                    }
		}catch(UnpositionedError e) {
		    throw e.addPosition(getTokenReference());
		}
	    } else if (fieldtype.isArrayType()  // T[] a; 
		       && ((CArrayType)fieldtype).getBaseType().isTypeVariable())
		{ 
		    CClassType subtype = prefix.getType().getCClass().getSubstitution
			((CTypeVariable)((CArrayType)fieldtype).getBaseType(),
			 prefix.getType().getAllArguments());
		    subtype = new CArrayType(subtype,((CArrayType)fieldtype).getArrayBound(), ((CArrayType)fieldtype).getCUniverse()); // WMD
		    try {
                        // Avoid inserting a cast on the LHS of an assignment.
			if (!context.isLeftSide()
                            && !(fieldtype.getErasure(context).equals
			      (subtype.getErasure(context)))){
			    JCastExpression castexpr = 
				new JCastExpression
				(getTokenReference(),
				 new JCheckedExpression(getTokenReference(),
							this),
				 subtype,
                                 true);
			    castexpr.typecheck(context);
			    return castexpr;
			} else {
			    setType(subtype);
			}
		    }catch (UnpositionedError e) {
			throw e.addPosition(getTokenReference());
		    }
		} else {
		    if (fieldtype.isArrayType()
			&& ((CArrayType)fieldtype).getBaseType().isGenericType()) //Stack<Integer>[] a;
			{ 
			    CType basetype = ((CArrayType)fieldtype).getBaseType();
			    CClassType newType = 
				((CClassType)basetype).createSubstitutedType
				(local,
				 (CClassType)prefix.getType(),
				 basetype.getAllArguments());
			    CArrayType newarray = 
				new CArrayType
				(newType,
				 ((CArrayType)fieldtype).getArrayBound(),
				 ((CArrayType)fieldtype).getCUniverse()); // WMD
			    setType(newarray);
			} 
		}
	    if(fieldtype.isGenericType()) { //Stack<Integer> a
		
		if(prefix instanceof JThisExpression) {
		    return this;}
		else {
		    CClassType newType = ((CClassType)fieldtype).createSubstitutedType(local,(CClassType)prefix.getType(),local.getType().getAllArguments());
		    //field.setType(newType);
		    setType(newType);
		}
	    }
	    return this;
	}
    }
    
    /**
     * Finds the type of the prefix as part of typechecking.  Sets
     * <code>field</code> to the signature of the field referenced. */
    private void findPrefixAndField( CMemberHost localHost, 
				     CExpressionContextType context ) 
	throws PositionedError 
    {
	if( prefix != null ) {
	    prefix = prefix.typecheck( context, false, false, false );
	    check( context,
		   prefix.getType().isClassType(),
		   MjcMessages.FIELD_BADACCESS, prefix.getType() );
	    try {
		field = 
		    prefix.getType().getCClass().lookupField( ident, context );
		
	    } catch( UnpositionedError cue ) {
		throw cue.addPosition( getTokenReference() );
	    }
		// if we are looking for but can't find the synthetic outer
		// this parameter, then it was not generated because it is
		// inaccessible
	    check( context, !(ident == JAV_OUTER_THIS && field == null),
		   MjcMessages.CONSTRUCTOR_EXPLICIT_CALL );
	    check( context, field != null, MjcMessages.FIELD_UNKNOWN, ident);
	    if( prefix instanceof JSuperExpression ) {
		// An outer super field access always needs a wrapper
		// method since the verifier checks that directed
		// field reference only comes from a subclass.
		if (prefix.getType() != context.getClassContext().
		    getOwnerClass().getSuperClass().getType()) {
		    field =
			((JSuperExpression) prefix).targetSubclass().
			getterOrSetterMethodFor((CField)field,context);
		    // prefix will be set appropriately below
		    prefix = null;
		}
	    }
	} else {
	    // prefix was null so this must be a simple-name field reference
	    try {
		field = context.lookupField( ident );
		if (field == null) { // didn't find the field in the immediate
				// surrounding class so try the class outside
				// that
		    field = context.lookupOuterField(ident);
		}
	    } catch (UnpositionedError e) {
		throw e.addPosition(getTokenReference());
	    }
		// if we are looking for but can't find the synthetic outer
		// this parameter, then it was not generated because it is
		// inaccessible
	    check( context, !(ident == JAV_OUTER_THIS && field == null),
		   MjcMessages.CONSTRUCTOR_EXPLICIT_CALL );
	    
	    check( context, field != null, MjcMessages.FIELD_UNKNOWN, ident);
	}
	//@ assert field != null;

	// If prefix was null or was set to null for an outer super
	// field access then generate the appropriate value for the
	// prefix
	if (prefix == null) {
	    if( field.isFieldStatic() ) {
		// build an AST node that references the class of the static
		// field (either this class or a superclass)
		prefix = new JTypeNameExpression( getTokenReference(), 
						  field.owner().getType(),
                                                  null);
	    } else {
		CClass fieldOwner = field.owner();
		if( fieldOwner == localHost // local field, or ...
		    || ( localHost instanceof CClass && // inherited field
			 ((CClass) localHost).descendsFrom( fieldOwner ) ) ) {
		    prefix = new JThisExpression( getTokenReference() );
		} else {
		    // Reference to a field of an outer or augmented
		    // class.  In either case the prefix reference
		    // should be to a generated variable (either an
		    // "outer this" or a parameter of the external
		    // method).  The typechecking and code generation
		    // of the "this" expression handles appropriate
		    // conversion to a generated variable reference.
		    prefix = new JThisExpression( getTokenReference(),
						  fieldOwner );
		}
	    }
	    prefix = prefix.typecheck( context, false, false, false );
	}
    }

    /**
     * If a non-static field is accessed as a member of a raw type, the
     * following actions are performed (see JLS3 4.8):
     * <ul>
     *   <li>
     *     Replace the type of the field access expression by its erasure.
     *   </li>
     *   <li>
     *     If the field's type changes by erasure, issue an unchecked warning.
     *   </li>
     * </ul>
     *
     * @param context  The typechecking context.
     *
     * @throws PositionedError if an error occurs during typechecking.
     */
    private void handlePossibleRawTypeAccess(CExpressionContextType context)
            throws PositionedError {
        // A static field retains its type even if accessed as a member of a
        // raw type.
        if (field.isStatic()) {
            return;
        }

        CClassType prefixType = (CClassType) prefix.getType();
        // If a non-static field is accessed on a raw type, the expression's
        // type is the erasure of the field type (see JLS3 4.8).
        if (prefixType.isRawTypeInvocationOf(field.owner())) {
            try {
                setType(getType().getErasure(context));

                // If the field's type changes by erasure and the field is being
                // assigned, we have to issue an unchecked warning (JLS3 4.8).
                if (field.getType().changesByErasure() && context.isLeftSide()) {
                    warn(context,
                         MjcMessages.UNCHECKED_FIELD_ASSIGNMENT,
                         field.ident(),
                         field.owner().qualifiedName().replace('/', '.'));
                }
            } catch (UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }
        }
    }

    /**
     * Verifies that the field is accessible in this context.
     *
     * @exception	PositionedError	if a check fails
     */
    private void checkAccess( CMemberHost localHost, 
			     CExpressionContextType context ) 
	throws PositionedError 
    {
	    // Is this field reference in the class in which this
	    // field is defined (or inherited)?
	if( prefix instanceof JThisExpression 
	    && !field.isFieldStatic() ) {
	    // make sure the super constructor has already been called
	    // else the programmer is trying to pass an uninitialized
	    // field to the superclass constructor
	    check( context, !context.isBeforeSuperConstructorCall(),
		   MjcMessages.CONSTRUCTOR_EXPLICIT_CALL );
	}

	try {
	    field.setType(field.getType().checkType( context ));
	} catch( UnpositionedError cue ) {
	    throw cue.addPosition( getTokenReference() );
	}

    }

    public boolean equals(/*@ nullable @*/ Object o) {
	return o instanceof JClassFieldExpression &&
	    field.equals(((JClassFieldExpression)o).field) &&
	    prefix.equals(((JClassFieldExpression)o).prefix);
    }

    /**
     * Throws a runtime exception because hashCode isn't properly
     * implemented yet. */
    public int hashCode() {
	fail( "An instance of " + getClass() + " cannot be stored in "
	      + " a HashTable or HashSet because equals is overridden but "
	      + " hashCode is not" );
	return 0;
    }



    /**
     * Performs checks to ensure the class field expression is
     * valid according to Generic Universe Types.
     * Also sets fieldtype_with_universe to a clone of the original
     * field type and sets the correct result universe.
     *
     * By WMD.
     */
    private void typecheckUniverse(CExpressionContextType context)
    throws PositionedError {
    	// If the type of the field is not a class type, we don't care
    	// about universes since values are copied
		if (!(fieldtype instanceof CClassType)) {
			return;
		}

		// HANDLES REFERENCES FROM INNER TO OUTER CLASSES
		// DO NOT SUPPORT THIS YET WITH GENERIC UNIVERSE TYPES
		// Required information for the 
		// CMemberHost localHost = context.findNearestHost();
		// CClass fieldOwner = field.owner();

		// Get type of the prefix
		CClassType prefixType = (CClassType) prefix.getType();
		
		// Get type of the field
		CClassType fieldType = (CClassType) fieldtype;

		// Substitute inherited type variables
		fieldTypeBeforeViewpointAdaptation = CUniverseServices
				.substituteInheritedTypeVariables(fieldType, prefixType, context);
		
		// Determine the resulting type of the expression
		fieldType = CUniverseServices
				.combine(prefixType, fieldTypeBeforeViewpointAdaptation);

		setType(fieldType);
   }
    
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitFieldExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	boolean		discardValue = code.discardValue();

	if (field.isFieldStatic()) {
	    if (prefix != null) {
		code.setDiscardValue(true);
		prefix.genCode(code);
		code.setDiscardValue(discardValue);
	    }
	} else {
	    prefix.genCode(code);
	}
	if (!discardValue) {
	    field.genLoad(code);
	}
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genStartStoreCode(CodeSequence code) {
	if (field.isFieldStatic()) {
	    if (prefix != null) {
		boolean		discardValue = code.discardValue();

		code.setDiscardValue(true);
		prefix.genCode(code);
		code.setDiscardValue(discardValue);
	    }
	} else {
	    prefix.genCode(code);
	}
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genEndStoreCode(CodeSequence code, boolean discardValue) {
	if (!discardValue) {
	    int	opcode;

	    if (field.isFieldStatic()) {
		if (getType().getSize() == 2) {
		    opcode = opc_dup2;
		} else {
		    opcode = opc_dup;
		}
	    } else {
		if (getType().getSize() == 2) {
		    opcode = opc_dup2_x1;
		} else {
		    opcode = opc_dup_x1;
		}
	    }

	    code.plantInstruction(new NoArgInstruction(opcode));
	}

	field.genStore(code);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    //@ spec_public
    private JExpression	prefix;
    private String	ident;
    private CFieldAccessor	field;
    private CType fieldtype;
    private CClassType fieldTypeBeforeViewpointAdaptation = null;

    /**
     * If true, indicates that this was a simple field reference
     * without an explicit prefix.  */
    //@ spec_public
    private boolean prefixWasBlank;

    /** the name of this field reference as originally appeared in
     * the source file. It may be null if generated by the system. */
    private JNameExpression sourceName;
}
