/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JClassOrGFImport.java,v 1.11 2006/12/20 06:15:51 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.classfile.ClassPath;
import org.multijava.util.compiler.CompilationAbortedError;
import org.multijava.util.compiler.PositionedError;

/**
 * This class represents (in the AST) import statements for single
 * classes or generic functions, e.g., <code>import
 * java.util.ArrayList;</code> or <code>import
 * org.multijava.samples.typecheck</code>. */
public class JClassOrGFImport extends JPhylum 
    implements JClassOrGFImportType
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs an AST node for a class or external member import statement.
     * @param where	the token reference of this node
     * @param name	the class or external member name
     * @param comments	the non-javadoc comments for this statement
     */
    public JClassOrGFImport( TokenReference where, 
			     /*@non_null@*/ String name, 
			     JavaStyleComment[] comments ) {
	super(where);

	this.name = name;
	this.comments = comments;
	this.used = false;

	int index = name.lastIndexOf('/');

	this.ident = index == -1 ? name : name.substring(index + 1).intern();
    }

    // ----------------------------------------------------------------------
    // ACCESSORS & MUTATORS
    // ----------------------------------------------------------------------

    /**
     * @return the class name defined by this statement
     */
    public /*@non_null@*/ String getName() {
	return name;
    }

    /**
     * @return the class name defined by this statement
     */
    public /*@non_null@*/ String ident() {
	return ident;
    }

    /**
     * States that specified class is used.
     */
    public void setUsed() {
	used = true;
    }

    /**
     * Indicates whether this simple import statement imports a type.
     *
     * @return	true iff this imports a type
     */
    public boolean isClassImport() {
	return isClassImport;
    }

    /**
     * Indicates whether this simple import statement imports an external
     * generic function.
     *
     * @return	true iff this imports an external generic function
     */
    public boolean isGFImport() {
	return !isClassImport;
    }

    /**
     * Registers whether this imports a type or an external generic function.
     */
    public void setImportKind(/*@non_null@*/ FileFinder finder) throws PositionedError {
	// If we find a matching class definition file, then assume it
	// is that.
	ClassPath.ClassDescription classDesc = finder.find( name );
	ClassPath.ClassDescription gfDesc = finder.find( name + MJ_ANCHOR );

	// there are several possible cases
	switch ((descKind(classDesc) << SHIFT) + descKind(gfDesc)) {
	case             (NULL       << SHIFT) +          NULL:
	    // Since no source code or bytecode exists the source code
	    // is not on the classpath or sourcepath, however it may
	    // be that the source was actually named on the command
	    // line and is being compiled.
	    if (CTopLevel.isClassLoaded(name)) {
		isClassImport = true;
	    } else if (CTopLevel.isGFLoaded(name)) {
		isClassImport = false;
	    } else {
		// This will eventually be an error since neither a
		// class or generic function of the given name exists, but
		// we mark it as a class so as to get the same sort of
		// error message as regular Java would provide.
		isClassImport = true;
	    }
	    break;
	    
	case             (NULL       << SHIFT) +          BYTECODE:
	    isClassImport = false;
	    break;

	case             (NULL       << SHIFT) +          SOURCE:
	    // if one search returns source then the other should also
	    //@ unreachable;
	    isClassImport = false;
	    break;

	case             (BYTECODE   << SHIFT) +          NULL:
	    isClassImport = true;
	    break;

	case             (BYTECODE   << SHIFT) +          BYTECODE:
	    // this shouldn't happen unless someone names a class with
	    // $anchor, assume the class declaration is correct
	    isClassImport = true;
	    break;

	case             (BYTECODE   << SHIFT) +          SOURCE:
	    // this is the typical case for an up-to-date class
	    // declaration, the source code probably declares the
	    // class defined in the bytecode
	    isClassImport = true;
	    break;

	case             (SOURCE     << SHIFT) +          NULL:
	    // if one search returns source then the other must also
	    //@ unreachable;
	    isClassImport = true;
	    break;

	case             (SOURCE     << SHIFT) +          SOURCE:
	    // Ambiguous without reading the source file since the
	    // source file might contain a generic function
	    // declaration or a class declaration.
	    int kind = 
		CTopLevel.whichIsThis(name, ((ClassPath.FileClassDescription) 
					     classDesc).file());
	    switch (kind) {
	    case CTopLevel.IS_GEN_FUNC:
		isClassImport = false;
		break;

	    case CTopLevel.IS_TYPE:
		isClassImport = true;
		break;

	    default:
		throw new PositionedError( getTokenReference(),
					   MjcMessages.TYPE_OR_GF_UNKNOWN,
					   name.replace('/', '.') );
	    }
	    
	    break;

	case             (SOURCE     << SHIFT) +          BYTECODE:
	    // this is the typical case for an up-to-date generic
	    // function declaration, the source code probably declares
	    // the generic function defined in the bytecode
	    isClassImport = false;
	    break;

	default:
	    throw new CompilationAbortedError( "unreachable" );
	}
    }

    // These constants are used to make the case analyis in
    // setImportKind easier to read.
    private static final int NULL = 0;
    private static final int BYTECODE = 1;
    private static final int SOURCE = 2;
    private static final int SHIFT = 2; // bits

    private static int descKind( ClassPath.ClassDescription desc ) 
    {
	if (desc == null) {
	    return NULL;
	}
	if (desc.isClass()) {
	    return BYTECODE;
	}
	return SOURCE;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks whether this class import statement names a class (or
     * generic function) that is used in the source code.  Reports a
     * warning if not.  */
    public void typecheck(/*@non_null@*/ Main compiler) {
	if( !used && getTokenReference() != TokenReference.NO_REF ) {
	    compiler.reportTrouble(
			  new CWarning( getTokenReference(),
					MjcMessages.UNUSED_CLASS_IMPORT,
					name.replace('/', '.'),
					null ));
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(/*@non_null@*/ MjcVisitor p) {
//  !FIXME! 	if (comments != null) {
//  	    p.visitComments(comments);
//  	}
//  	p.visitClassImport(name);
	p.visitClassOrGFImport( this );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private final /*@non_null@*/ String		name;
    private final /*@non_null@*/ String		ident;
    private final JavaStyleComment[]	comments;
    private boolean			used;
    private boolean			isClassImport;
}
