/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCompilationUnit.java,v 1.66 2007/02/08 14:04:54 leavens Exp $
 */

package org.multijava.mjc;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.TreeSet;

import org.multijava.util.Destination;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a single Java compilation unit (typically a file
 * in a file-based compiler like this) in the AST.  The production for a
 * compilation unit is the main entry point in Java parser grammar.
 */
public class JCompilationUnit extends JPhylum 
    implements Constants, JCompilationUnitType 
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a CompilationUnit with the specified top level context
     * @param	where		the position of this token
     * @param	packageName	the package to which this c.u. belongs
     * @param	importedPackages	the packages imported by this c.u.
     * @param	importedUnits	the classes and external members imported by 
     *				this compilation unit
     * @param	typeDeclarations	the classes and interfaces declared 
     *					in this c.u.
     * @param	tlMethods	the MultiJava top-level declarations
     *				in this c.u.
     */
    public JCompilationUnit( /*@non_null*/ TokenReference where,
			     /*@non_null*/ JPackageName packageName,
			     /*@non_null*/ CCompilationUnit export,
			     /*@non_null*/ JPackageImportType[] importedPackages,
			     /*@non_null*/ ArrayList importedUnits,
			     /*@non_null*/ JTypeDeclarationType[] typeDeclarations,
			     /*@non_null*/ ArrayList tlMethods //of MJTopLevelMethodDeclaration
			     ) {
	super(where);
	this.packageName = packageName;
	this.export = export;
	this.importedPackages = importedPackages;
	this.importedUnits = importedUnits;
	this.typeDeclarations = typeDeclarations;
	this.tlMethods = tlMethods;

	String baseName = where.name();
	int pos = baseName.indexOf( "." );
	if( pos < 0 ) {
	    fileIdent =  baseName;
	} else {
	    fileIdent = baseName.substring( 0, pos );
	}
    }

    /**
     * Separates the import statements into those for classes and
     * those for generic functions.  Assigns to importedClasses and
     * importedGFs. */
    private void separateImports() throws PositionedError {
	FileFinder finder = CTopLevel.fileFinder();

	ArrayList importedClassesArrayList = new ArrayList();
	ArrayList importedGFsArrayList = new ArrayList();
	Iterator iter = importedUnits.iterator();
	while( iter.hasNext() ) {
	    JClassOrGFImportType iu = (JClassOrGFImportType) iter.next();
	    iu.setImportKind(finder);
	    if( iu.isClassImport() ) {
		importedClassesArrayList.add( iu );
	    } else {
		importedGFsArrayList.add( iu );
	    }
	}
	
	importedClasses = 
	    new JClassOrGFImportType[importedClassesArrayList.size()];
	importedClassesArrayList.toArray( importedClasses );
	importedGFs = new JClassOrGFImportType[importedGFsArrayList.size()];
	importedGFsArrayList.toArray( importedGFs );
	
	
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * @return the package name of this compilation unit
     */
    public /*@non_null@*/ String packageNameAsString() {
	assertTrue(packageName != null);
	return packageName == JPackageName.UNNAMED ? 
	    "" : packageName.getName() + '/';
    }

    /**
     * Returns the file associated with this compilation unit.
     */
    public File file() {
	return getTokenReference().file();
    }

    /**
     * Returns the Java identifier associated with the name of the
     * file containing this compilation unit.  For example, if the
     * file is named &quot;typecheck.java&quot; then this method
     * returns &quot;typecheck&quot;.
     *
     * @return the Java identifier associated with the file containing this */
    public String fileNameIdent() {
	return fileIdent;
    }

    public /*@non_null*/ JPackageName packageName() { return packageName; }

    public void setPackage(/*@non_null@*/ JPackageName n) {
	packageName = n;
    }

    public /*@non_null@*/ JPackageImportType[] importedPackages() { 
	return importedPackages==null ? 
	    new JPackageImportType[0] : importedPackages;
    }

    public void setImportedPackages(/*@non_null*/ JPackageImportType[] packages) {
	importedPackages = packages;
    }

    public /*@non_null@*/ JClassOrGFImportType[] importedClasses() { 
	return importedClasses==null ? 
	    new JClassOrGFImportType[0] : importedClasses;
    }

    public /*@ pure non_null@*/ JTypeDeclarationType[] typeDeclarations() { 
	return typeDeclarations==null ? 
	    new JTypeDeclarationType[0] : typeDeclarations;
    }

    public /*@non_null@*/ MJGenericFunctionDecl[] gfDeclarations() { 
	return gfDeclarations==null ? 
	    new MJGenericFunctionDecl[0] : gfDeclarations;
    }

    /**
     * Returns the CSourceClass objects representing the type
     * signatures of the types declared in this compilatoin unit
     * <em>and as nested types</em>.
     */
    public /*@non_null@*/ CSourceClass[] allTypeSignatures() {
	if (typeDeclarations == null) {
	    return new CSourceClass[0];
	}
	
	ArrayList accum = new ArrayList();
	for (int i = 0; i < typeDeclarations.length; i++) {
	    typeDeclarations[i].accumAllTypeSignatures(accum);
	}
	CSourceClass[] result = new CSourceClass[accum.size()];
	accum.toArray(result);
	return result;
    }

    /**
     * Returns true iff this compilation unit contains a declaration
     * for a top-level type with the given fully qualified name.
     *
     * @param qualifiedName a fully qualified name with parts
     * separated by '/' not '.'
     */
    public boolean declaresType(/*@non_null@*/ String qualifiedName) {
	String nonPackagePart = matchPackageName(qualifiedName);
	if (nonPackagePart == null || typeDeclarations == null ||
	    nonPackagePart.indexOf('/') >= 0) {
	    return false;
	}
	for (int i = 0; i < typeDeclarations.length; i++) {
	    if (typeDeclarations[i].ident().equals(nonPackagePart)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Returns true iff this compilation unit contains a declaration
     * for a generic function with the given fully qualified name.
     *
     * @param qualifiedName a fully qualified name with parts
     * separated by '/' not '.'
     */
    public boolean declaresGF(/*@non_null@*/ String qualifiedName) {
	String nonPackagePart = matchPackageName(qualifiedName);
	if (nonPackagePart == null || tlMethods == null ||
	    nonPackagePart.indexOf('/') >= 0) {
	    return false;
	}
	Iterator iter = tlMethods.iterator();
	while (iter.hasNext()) {
	    MJTopLevelMethodDeclaration m = 
		(MJTopLevelMethodDeclaration) iter.next();
	    if (m.ident().equals(nonPackagePart)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * If qualifiedName begins with the package name of this
     * compilation unit, then returns the remainder of qualifiedName;
     * otherwise returns null.
     */
    private String matchPackageName(/*@non_null@*/ String qualifiedName) {
	String packageName = packageNameAsString();
	if (qualifiedName.startsWith(packageName)) {
	    return qualifiedName.substring(packageName.length());
	} else {
	    return null;
	}
    }

    /**
     * Returns the AST subtrees for the augmenting method
     * declarations.  Each element of the list is of type
     * MJTopLevelMethodDeclaration */
    public /*@non_null*/ ArrayList tlMethods() { return tlMethods; }


    /**
     * Returns an array of the class and generic function imports for
     * this compilation unit.  Typically code should use
     * importedClasses or importedGFs instead.  This method is
     * available as a testing harness since the other two methods
     * report empty results until after the preprocessDependencies pass.
     */
    public /*@non_null@*/ JClassOrGFImportType[] importedUnits() { 
	if (importedUnits == null) {
	    return new JClassOrGFImportType[0];
	} else {
	    JClassOrGFImportType[] result = 
		new JClassOrGFImportType[importedUnits.size()];
	    importedUnits.toArray(result);
	    return result;
	}
    }

    public ArrayList importedUnitsAsArrayList() { 
	return importedUnits;
    }

    public /*@non_null@*/ JClassOrGFImportType[] importedGFs() { 
	return importedGFs==null ? 
	    new JClassOrGFImportType[0] : importedGFs;
    }

    public /*@non_null@*/ CCompilationUnitContextType createCompilationUnitContext( 
        /*@non_null@*/ Main compiler, 
	CCompilationUnit cunit ) 
    {
	return compiler.createCompilationUnitContext(this, cunit);
    }


    // ---------------------------------------------------------------------
    // REVISE AST FOR MULTIJAVA
    // ---------------------------------------------------------------------

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved. Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    public void preprocessDependencies( /*@non_null@*/ Main compiler )
	throws PositionedError 
    {
	separateImports();

	// update the CCompilationUnit instance that manages the name space
	export.update( importedClasses,
		       importedGFs,
		       importedPackages,
		       allLoadedClasses,
		       inScopeGenFuncMap );

	CCompilationUnitContextType context = 
	    createCompilationUnitContext( compiler, export );

	try {
	    // JLS 7.5.1 Single-Type-Import Declaration
	    // (Type-Import-on-Demand Declaration (JLS 7.5.2) are handled
	    // separately, see lookupClass generic function)

	    for (int i = 0; i < importedClasses.length; i++) {
		JClassOrGFImportType ic = importedClasses[i];

		// gets a CClassType for the given name
		CClassType impClass = CTopLevel.getTypeRep(ic.getName(),true);

		try {
		    impClass = (CClassType)(impClass.checkType(context));
		} catch (org.multijava.util.compiler.UnpositionedError e) {
		    throw e.addPosition(getTokenReference());
		} catch (RuntimeException e) {
		    // This is a debugging aid that can be removed
		    // when all bugs are fixed.  It helps by noting
		    // the position in the file being parsed/checked
		    // that triggers a RuntimeException that ought to
		    // be fixed. !FIXME! - DRCok
		    warn(context, MjcMessages.RUNTIME_EXCEPTION,
			 "while attempting to check the type of " + 
			 ic.getName());
		    throw e;
		}

		// check whether there are two single-type-imports that
		// refer to classes with the same simple name appearing in
		// different packages
		Object clazz = allLoadedClasses.put(ic.ident(), impClass);
		if( clazz != null ) {
		    if( impClass != clazz ) {
			compiler
			    .reportTrouble(
				new PositionedError( getTokenReference(), 
						     MjcMessages.CUNIT_RENAME, 
						     impClass.ident() ));
		    }
		}
	    }

	    groupByAnchor( compiler, context );

	    // -------------------------------------------------------------
	    // recurse into child nodes of this AST node
	    for (int i = 0; i < typeDeclarations.length; i++) {
		typeDeclarations[i].preprocessDependencies( context );
	    }

	    for (int i = 0; i < gfDeclarations.length ; i++) {
		gfDeclarations[i].preprocessDependencies( context );
	    }

	} catch (RuntimeException e) {
	    // This is a debugging aid that can be removed when all bugs
	    // are fixed.  It helps by noting the position in the file
	    // begin parsed/checked that triggers a RuntimeException that
	    // ought to be fixed. !FIXME! - DRCok
	    warn(context, MjcMessages.RUNTIME_EXCEPTION,
		 "while grouping compilation unit");
	    throw e;
	} finally {
	    compiler.adoptCompilationUnitContext( context );
	}
    }
    
    /**
     * Groups top-level methods into collections (of type
     * <code>MJGenericFunctionDecl</code>) that correspond to the
     * anchor classes.  
     *
     * @see MJGenericFunctionDecl
     */
    private void groupByAnchor( /*@non_null@*/ Main compiler, 
				CCompilationUnitContextType context ) 
	throws PositionedError
    {
	// First sort by identifiers and private modifier: Good style
	// dictates that all the identifiers should be the same and
	// match the c.u. name.  We warn on this inside the method's
	// checkInterface.  However, we permit different names for
	// this version.  !HERE!filenaming
	Collections.sort( tlMethods,
			  new Comparator() {
			      public int compare( Object o1, Object o2 ) {
				  JMethodDeclarationType m1 =
				      (JMethodDeclarationType) o1;
				  JMethodDeclarationType m2 =
				      (JMethodDeclarationType) o2;
				  long mod1 = m1.modifiers() & ACC_PRIVATE; 
				  long mod2 = m2.modifiers() & ACC_PRIVATE;
				  if( mod1 != mod2 ) {
				      // exactly one is private
				      return (mod1 == ACC_PRIVATE) ? 
					  -1 : 1;
				  } else {
				      // both private or both are not
				      String s1 = m1.ident();
				      String s2 = m2.ident();
				      return s1.compareTo( s2 );
				  }
			      }
			  } );
	/* 
	   assert (* tlMethods is sorted first by privacy and 
	   then by identifier *);
	*/

	// Group top-level methods into collections that will share a
	// single anchor class.  As the collections are formed we
	// create new AST child nodes, corresponding to anchor
	// classes, to hold the method declarations.  These nodes are
	// JTypeDeclarationType subtypes that will be augmented during
	// the typechecking pass with the information needed for code
	// generation.
	int methodCount = tlMethods.size();
	ArrayList anchorMates = new ArrayList();
	ArrayList anchors = new ArrayList();
	ArrayList privateAnchors = new ArrayList();
	String lastIdent = null;
	boolean lastWasPrivate = false;
	for( int i=0; i < methodCount; i++ ) {
	    MJTopLevelMethodDeclaration methDecl = 
		(MJTopLevelMethodDeclaration) tlMethods.get(i);
	    CMethod m = (CMethod) methDecl.checkInterface( context );
	    if( lastIdent != null &&
		( !lastIdent.equals( m.ident() )
		  || lastWasPrivate != m.isPrivate() ) ) {
		// different anchor so we now have the complete set of
		// methods that belong to the previous anchor
		makeAndStoreAnchorDecl( anchorMates, 
					lastIdent, lastWasPrivate,
					anchors, privateAnchors );
		// start a new list for the new anchor
		anchorMates.clear();
	    }
	    anchorMates.add( methDecl );
	    lastIdent = m.ident();
	    lastWasPrivate = m.isPrivate();
	}
	
	if( !anchorMates.isEmpty() ) {
	    makeAndStoreAnchorDecl( anchorMates, 
				    lastIdent, lastWasPrivate,
				    anchors, privateAnchors );
	}
	
	// ----------------------------------------------------------------
	// Attach the anchor ASTs to the appropriate parent nodes

	// non-private anchors are attached to this
	gfDeclarations = new MJGenericFunctionDecl[ anchors.size() ];
	anchors.toArray( gfDeclarations );

	// Add the private decls to all the decls in gfDeclarations
	// and in typeDeclarations. These private external methods
	// will be typechecked repeatedly, once for each declaration
	// that they are inserted into here.  That is OK though since
	// any type declaration statements and anonymous classes
	// within the private GFs will cause the appropriate synthetic
	// classes to be generated.
	privateGFDeclarations = 
	    new MJGenericFunctionDecl[ privateAnchors.size() ];
	privateAnchors.toArray( privateGFDeclarations );
	for( int j=0; j < typeDeclarations.length; j++ ) {
	    clonePrivateAnchorsAndAddTo( compiler, typeDeclarations[j], true );
	}
	for( int j=0; j<gfDeclarations.length; j++ ) {
	    clonePrivateAnchorsAndAddTo( compiler, gfDeclarations[j], false );
	}

	for( int i=0; i < gfDeclarations.length; i++ ) {
	    gfDeclarations[i].generateInterface( compiler, null, export, 
						 packageNameAsString(), 
						 false, false );
	}
    }

    /**
     * Utility method generates a <code>MJGenericFunctionDecl</code>
     * anchor for the given methods and stores it in the appropriate
     * anchor collection.
     *
     * @param	anchorMates	the method declarations
     * @param	id		the shared identifier of the methods
     * @param	priv		indicates that all the methods are private
     * @param	anchors		non-private anchors declared in this
     * @param	privateAnchors	private anchors declared in this
     */
    private void makeAndStoreAnchorDecl( /*@non_null@*/ ArrayList anchorMates,
					 /*@non_null@*/String id, 
					 boolean priv,
					 /*@non_null@*/ ArrayList anchors, 
					 /*@non_null@*/ ArrayList privateAnchors ) {
	MJTopLevelMethodDeclaration[] methodDecls = 
	    new MJTopLevelMethodDeclaration[ anchorMates.size() ];
	anchorMates.toArray( methodDecls );
	MJGenericFunctionDecl anchorDecl = 
	    new MJGenericFunctionDecl( getTokenReference(),
				       priv ? ACC_PRIVATE : ACC_PUBLIC, id, 
				       methodDecls );
	if( priv ) {
	    privateAnchors.add( anchorDecl );
	} else {
	    anchors.add( anchorDecl );
	}
    }

    /**
     * Clones every private generic function anchor and adds each as a
     * child of the given parent node.
     *
     * @param	parent		the AST node to add the private generic 
     *				function anchor AST to
     * @param	genInterface	if true then generate the interface for the
     * cloned anchors */
    private void clonePrivateAnchorsAndAddTo( /*@non_null@*/ Main compiler,
					      /*@non_null@*/ JTypeDeclarationType parent,
					      boolean genInterface ) {
	for( int i=0; i < privateGFDeclarations.length; i++ ) {
	    MJGenericFunctionDecl copy = 
		(MJGenericFunctionDecl) privateGFDeclarations[i].clone();
	    parent.addMember( copy );
	    if( genInterface ) {
		String prefix = parent.getCClass().qualifiedName() + "$";
		copy.generateInterface( compiler, null, parent.getCClass(),
					prefix, false, true );
	    }
	}
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type signatures of
     * everything (imported generic functions, classes being compiled,
     * methods, fields, etc...)  needed for the later passes.  This
     * information is stored in a CCompilationUnit instance and
     * instances of CMember that are bound to the AST.  Also adds
     * things like the default constructor and common initializer
     * methods (which group all field initializers and initializer
     * blocks) to the AST.
     *
     * @exception	PositionedError		an error with reference 
     * 						to the source file */
    public void checkInterface( /*non_null@@*/ Main compiler ) throws PositionedError {
	CCompilationUnitContextType context = 
	    createCompilationUnitContext( compiler, export );

	try {
	    // -------------------------------------------------------------
	    // Single-GF-import declarations
	    //
	    // These must be processed in a separate pass from the
	    // generation of anchor classes so that appropriate singletons
	    // are available for generic functions declared in compilation
	    // units being compiled in this session.
	    for (int i = 0; i < importedGFs.length; i++) {
		JClassOrGFImportType gfImport = importedGFs[i];

		// gets a CGenericFunctionCollection singleton for the
		// given name
		CGenericFunctionCollection impGF = 
		    CTopLevel.lookupGFCollection( gfImport.getName() );
		if( impGF == CGenericFunctionCollection.GF_UNDEFINED ) {
		    throw new PositionedError( getTokenReference(),
					       MjcMessages.EXT_GF_UNKNOWN, 
					       gfImport.getName() );
		}
	    
		// Registers that this compilation unit imports a generic
		// function with the given name.  Unlike with type
		// imports, it is permissible to import two different
		// generic function collections with the same simple name.
		// The checks for lack of ambiguity over visible legal
		// argument tuples or method call targets will resolve any
		// problems that might arise from the dual import.
		inScopeGenFuncMap.register( gfImport.ident(), impGF );
	    }

	    // -------------------------------------------------------------
	    // locally declared external generic functions are
	    // registered in MJGenericFunctionDecl.checkInterface()

	    // -------------------------------------------------------------
	    // check uniqueness of type declarations
	    HashSet declaredIdents = new HashSet();
	    //Debug.breakMe();
	    for (int i = 0; i < typeDeclarations.length ; i++) {
		CClass declaredClass = typeDeclarations[i].getCClass();
		CType declaredType = declaredClass.getType();
		String declaredIdent = declaredClass.ident();

		boolean duplicate = false;

		if (declaredIdents.contains(declaredIdent)) {
		    // same type is declared twice in this c.u.
		    duplicate = true;
		}
		declaredIdents.add(declaredIdent);

		CType expectedType = 
		    (CType) allLoadedClasses.get( declaredIdent );

		if (expectedType == null) {
		    allLoadedClasses.put( declaredIdent, declaredType );
		    export.registerVisibleType( declaredType );
		} else {
  		    if( expectedType != declaredType ) {
			// there is an import of a type with the same short name
			duplicate = true;
  		    }
		}

		if (duplicate) {
		    compiler.reportTrouble(
			new PositionedError( getTokenReference(),
					     MjcMessages.CUNIT_RENAME,
					     declaredIdent ));
		}
		
	    }

	    // -------------------------------------------------------------
	    // check for public external generic function and public
	    // type of the same name
	    HashSet publicNames = new HashSet(typeDeclarations.length);
	    // all the type names are unique per checks above
	    for (int i = 0; i < typeDeclarations.length; i++) {
		if (hasFlag(typeDeclarations[i].modifiers(),ACC_PUBLIC)) {
		    publicNames.add(typeDeclarations[i].ident());
		}
	    }
	    for (int i = 0; i < gfDeclarations.length; i++) {
		if (hasFlag(gfDeclarations[i].modifiers(),ACC_PUBLIC) && 
		    publicNames.contains(gfDeclarations[i].methodIdent())) {
		    compiler.reportTrouble(
			new PositionedError( getTokenReference(),
					     MjcMessages.TYPE_AND_FUNCTION,
					     gfDeclarations[i].methodIdent() ) );
		}
	    }
	    

	    

	    // -------------------------------------------------------------
	    // descend into child nodes of this in the AST; check
	    // interface of child type declarations
	    for (int i = 0; i < typeDeclarations.length ; i++) {
		typeDeclarations[i].checkInterface( context );
	    }

	    for (int i = 0; i < gfDeclarations.length ; i++) {
		gfDeclarations[i].checkInterface( context );
	    }

	} finally {
	    compiler.adoptCompilationUnitContext( context );
	}
    }

    /**
     * Checks the static initializers created during the
     * checkInterface pass and performs some other checks that can be
     * performed simply before full blown typechecking.  Among these
     * other checks are those for appropriateness of subclassing, such
     * as inheritance circularities and &quot;reimplementation&quot;
     * of interfaces.
     *
     * @param	compiler	the compiler that is calling (passed down 
     *				through the AST via CContextType subtypes
     *				and used for error reporting)
     * @exception PositionedError if check fails */
    public void checkInitializers(/*@non_null@*/ Main compiler )  
	throws PositionedError 
    {
	CCompilationUnitContextType	context = 
	    createCompilationUnitContext( compiler, export );
	try {
	    for( int i = 0; i < typeDeclarations.length ; i++ ) {
		typeDeclarations[i].checkInitializers( context );
	    }
	    for( int i=0; i < gfDeclarations.length; i++ ) {
		gfDeclarations[i].checkInitializers( context );
	    }

	} finally {
	    compiler.adoptCompilationUnitContext( context );
	}
    }

    /**
     * Resolves value specializer expressions to the compile-time
     * constants they represent.  Must come after the
     * checkInitializers phase.
     *
     * @param	compiler	the compiler that is calling (passed down 
     *				through the AST via CContextType subtypes
     *				and used for error reporting)
     */
    public void resolveSpecializers(Main compiler) throws PositionedError {
	CCompilationUnitContextType	context = 
	    createCompilationUnitContext( compiler, export );
	try {
	    // compute the values being dispatched upon, for
	    // dispatching on constants (e.g. int@2)
	    for( int i = 0; i < typeDeclarations.length ; i++ ) {
		typeDeclarations[i].resolveSpecializers( context );
	    }
	    for( int i=0; i < gfDeclarations.length; i++ ) {
		gfDeclarations[i].resolveSpecializers( context );
	    }
	} finally {
	    compiler.adoptCompilationUnitContext( context );
	}
    }


    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Finds the top method of every declared method.  Called after
     * the resolveSpecializers pass.  This cannot be done before then
     * because the external generic function mappings are not complete
     * until the end of the checkInterface pass and the constant value
     * specializers are not known until after the resolveSpecializers
     * pass.  This must be done before the typecheck pass so that all
     * specialized argument positions for generic functions are known
     * for ambiguity checking. */
    public void resolveTopMethods() throws PositionedError {
	for( int i = 0; i < typeDeclarations.length ; i++ ) {
	    typeDeclarations[i].resolveTopMethods();
	}
	
	for( int i=0; i < gfDeclarations.length; i++ ) {
	    gfDeclarations[i].resolveTopMethods();
	}
    }

    /**
     * Typechecks this compilation unit.
     *
     * @param	compiler	the compiler that is calling
     * @exception PositionedError 	if a check fails that we cannot
     * 					recover from */
    public void typecheck( /*@non_null@*/ Main compiler ) 
	throws PositionedError 
    {
	CCompilationUnitContextType	context = 
	    createCompilationUnitContext( compiler, export );

	try {
	    if (packageName == JPackageName.UNNAMED) {
		compiler.reportTrouble(
		    new CWarning( getTokenReference(),
				  MjcMessages.PACKAGE_IS_MISSING ));
	    }

	    for( int i = 0; i < typeDeclarations.length ; i++ ) {
		typeDeclarations[i].typecheck( context );
	    }

	    for( int i=0; i < gfDeclarations.length; i++ ) {
		gfDeclarations[i].typecheck( context );
	    }

	    //-------------------------------------------------------------
	    // Check for unambiguous and complete implementation of
	    // generic functions.  That is, "for each visible generic
	    // function, each visible legal argument tuple [must have]
	    // a visible most-specific applicable method to invoke
	    // [CLCM 4.1.3.1]"

	    // Make sure the top methods have been resolved for any
	    // methods read from bytecode during the previous passes,
	    // and during the typechecking of this compilation unit.
	    CBinaryMethod.resolveTops();

	    // First get the visible generic functions that must be checked
	    CMethod[] visibleGenericFunctions = calcVisibleGenericFunctions();

	    if( visibleGenericFunctions.length > 0 ) {

		// Stores references to the <code>CType</code>s of all types
		// visible in this compilation unit.  A type is visible if it is
		// declared in or referred to in this compilation unit, or if it
		// is a primitive type.
		HashSet visibleTypesSet = 
		    (HashSet) PRIMITIVE_TYPES_SET.clone();
		visibleTypesSet.addAll( export.visibleTypes() );
		
		// Find types visible in external generic functions
		for (int i = 0; i < visibleGenericFunctions.length; i++) {
		    CMethod m = visibleGenericFunctions[i];
		    if (m.isExternal()) {
			CCompilationUnit mCU = m.getCCompilationUnit();
			if (mCU != export) {
			    visibleTypesSet.addAll( mCU.visibleTypes() );
			}
		    }
		}

		// More binary methods were probably loaded during the
		// transitive search for visible types.  We need to
		// resolve their top methods also.
		CBinaryMethod.resolveTops();
		
		// Convert the visible type set to an array for efficiency
		CType[] visibleTypes = 
		    (CType[]) visibleTypesSet
		    .toArray( new CType[ visibleTypesSet.size()] );

		// For every type t that appears in a top argument tuple of
		// some top method in visibleGenericFunctions, we
		// calculate the set of visible subtypes, 
		// Sub(t) = { u \in visibleTypes | u <: t }
		HashMap sub = new HashMap();
		for( int i = 0; i < visibleGenericFunctions.length; i++ ) {
		    CMethod m = visibleGenericFunctions[i];
		    CType[] stt = 
			CSpecializedType.staticTypeTupleFrom(m.receiverType(),
							     m.parameters() );
		    for( int j = 0; j < stt.length; j++ ) {
			CType t = stt[j];
			ArrayList sub_of_t = (ArrayList) sub.get( t );
			if( sub_of_t != null ) {
			    continue;
			}

			sub_of_t = new ArrayList();
			sub.put( t, sub_of_t );
			for( int k = 0; k < visibleTypes.length; k++ ) {
			    CType u = visibleTypes[k];
			    if( u.isAlwaysAssignableTo( t ) ) {
				sub_of_t.add( u );
			    } 
			}
			Collections.sort(sub_of_t);
		    }
		}

		// For each generic function, f, we check every legal argument
		// tuple according to CLCM 4.1.3.1 and 4.2, R1 and R2, with one
		// modification -- see the first case in the definition of
		// Sub'(t) below.  Let the top argument type tuple of f be
		// <tf0,tf1,...,tfn> and let the set of legal argument tuples
		// be L_f.  
		// L_f={<u0,u1,...,un> \in 
		//                   Sub'(tf0) x Sub''(tf1) x...x Sub''(tfn)}
		// where
		// Sub'(t) = Sub(t)
		//	     if the top method of f is not abstract
		//           [using all subtypes Sub(t) in this case
		//           ensures that we will fully check concrete
		//           methods of abstract classes for incompleteness
		//           and ambiguity.  This obviates errors "waiting to
		//           happen", which would be found when a concrete
		//           subclass comes along.  Using this semantics
		//           fixes two problems that we'd otherwise have to
		//           address:  1) the potential for a "top method"
		//           to be a multimethod, which screws up our compiler
		//           royally; 2) the potential for super sends to be
		//           ambiguous.]
		// else
		// Sub'(t) = { u \in Sub(t) | (isConcrete(u) || !isLocal(u)) }
		//           if R1 applies, i.e., isLocal(m) && isExternal(m); 
		// else
		// Sub'(t) = { u \in Sub(t) | isConcrete(u) } 
		//           otherwise;
		// and where
		// Sub''(tfi) = Sub(tfi) if parameter i is specialized in any 
		//                       of the visible methods of the generic 
		//                       function
		// else
		// Sub''(tfi) = tfi      otherwise
		for( int i=0; i < visibleGenericFunctions.length; i++ ) {
		    CMethod m = visibleGenericFunctions[i];
		    if( m.ident() == JAV_CONSTRUCTOR
			|| m.ident() == JAV_INIT
			|| m.ident() == JAV_STATIC_INIT ) {
			continue;
		    }

		    CType[] stt = 
			CSpecializedType.
			staticTypeTupleFrom( m.receiverType(),
					     m.parameters() );
		
		    // find the specialized argument positions
		    BitSet specArgs = m.getSpecArgs();
		    if( specArgs.length() <= 1 ) {
			// no ambiguities possible
			continue;
		    }

		    ArrayList[] l_f = new ArrayList[ stt.length ];
		    for( int j = 0; j < stt.length; j++ ) {
			// if-else handles the Sub'' calculation for
			// non-receiver argument positions (see comment
			// above)
			if( j==0 || specArgs.get( j ) ) {
			    l_f[j] = (ArrayList) sub.get( stt[j] );
			} else {
			    l_f[j] = new ArrayList( 1 );
			    l_f[j].add( stt[j] );
			}
		    }

		    // convert to Sub' (see comment above)
		    boolean r1Applies = m.isLocalTo(context) && m.isExternal();
		    l_f[0] = (ArrayList) l_f[0].clone();
		    // only prune out some abstract classes if the
		    // given gf is abstract (see Sub' definition above)
		    if (m.isAbstract()) {
			Iterator iter = l_f[0].iterator();
			while( iter.hasNext() ) {
			    CType u = (CType) iter.next();
			    if( !(u.isConcrete()
				  || r1Applies && !u.isLocalTo(context)) ) {
				iter.remove();
			    }
			}
		    }

		    CType.TupleCollection incompletions =
			new CType.TupleCollection();

		    try {
			Iterator legalArgTuples = makeIterator( l_f );
			while( legalArgTuples.hasNext() ) {
			    CType[] argTuple = (CType[]) legalArgTuples.next();
			    CMethodSet msamset =
				argTuple[0].getCClass().
				lookupMSAMethod( m, argTuple, 
						 CClassType.EMPTY,
						 context);

			    if( msamset.size() == 0 ) {
				incompletions.add(argTuple);
			    }
			}
			if (!incompletions.isEmpty()) {
			    CType[][] badTuples = incompletions.members();
			    for (int j = 0; j < badTuples.length; j++) {
				check( context, false,
				       MjcMessages.MSAM_NOT_FOUND, 
				       m,
				       CType.tupleToString( badTuples[j] ) );
			    }
			    incompletions.clear();
			}
		    } catch( UnpositionedError cue ) {
			throw cue.addPosition( getTokenReference() );
		    }
		}
	    }

	    // Check for unused class imports
	    for (int i = 0; i < importedClasses.length; i++) {
		importedClasses[i].typecheck( compiler );
	    }

	    // Check for unused package imports
	    for (int i = 0; i < importedPackages.length; i++) {
		importedPackages[i].typecheck( compiler, packageName );
	    }

	    // Check for unused generic function imports !FIXME!usedGF
	    //  	for (int i = 0; i < importedGFs.length; i++) {
	    //  	    importedGFs[i].typecheck( compiler );
	    //  	}

	} catch (PositionedError e) {
	    throw e;
	} catch (RuntimeException e) {
	    // This is a debugging aid that can be removed when all bugs
	    // are fixed.  It helps by noting the position in the file
	    // begin parsed/checked that triggers a RuntimeException that
	    // ought to be fixed. !FIXME! - DRCok
	    warn(context, MjcMessages.RUNTIME_EXCEPTION,
		 "while typechecking compilation unit");
	    throw e;
	} finally {
	    compiler.adoptCompilationUnitContext( context );
	}
    }

    /**
     * Finds the top methods of every generic function visible in this
     * compilation unit.  Works by building a set, adding the top
     * method of every visible method from
     * <code>export.visibleMethods()</code>.  Also calculates which arguments
     * of the generic function are specialized on in this compilation.
     * The visible generic functions are stored in
     * <code>visibleGenericFunctions</code>.  Each top method stores
     * a bit set indicating which arguments are specialized for its generic
     * function.
     *
     * @exception PositionedError if a top method cannot be found 
     */

    private /*@non_null@*/ CMethod[] calcVisibleGenericFunctions() throws PositionedError {
	try {
	    HashSet visibleGFs = new HashSet();

	    // for each visible method find the top method of its
	    // generic function and update the specialized argument
	    // positions
	    Iterator iter = export.visibleMethods().iterator();
	    while( iter.hasNext() ) {
		CMethod visibleMethod = (CMethod) iter.next();
		CMethod top = visibleMethod.topConcreteMethod();
		int parmCount = top.parameters().length;
		BitSet specArgs = top.getSpecArgs();

		// if some non-receiver position is specialized,
		// we need to perform ambiguity checking
		for( int i=0; i<parmCount; i++ ) {
		    if( specArgs.get( i+1 ) ) {
			visibleGFs.add(top);
			break;
		    }
		}
	    }

	    // convert to an array for efficiency
	    CMethod[] visibleGenericFunctions = new CMethod[visibleGFs.size()];
	    visibleGFs.toArray( visibleGenericFunctions );

	    // and sort the array for consistent error reporting
	    Arrays.sort(visibleGenericFunctions,VGF_COMPARATOR);

	    return visibleGenericFunctions;
	} catch( UnpositionedError cue ) {
	    throw cue.addPosition( getTokenReference() );
	}
    }

    /**
     * Generates an iterator for looping through the cartesian
     * cross-product of the sets of types.
     *
     * @param	l_f	each element should be a set of <code>CType</code>s
     * @return	an iterator */
    private static /*@non_null@*/ Iterator makeIterator( final ArrayList[] l_f ) {
	// if one of the members of the cross product is empty then
	// return empty iterator
	for( int i=0; i < l_f.length; i++ ) {
	    if( l_f[i].isEmpty() ) {
		return EMPTY_ITERATOR;
	    }
	}
	return new Iterator() {
		int[] pos;	// the position of the next elt to be returned 
		CType[][] tupleSpace;
		boolean hasNext = true;

		// instance initializer
		{
		    pos = new int[ l_f.length ];
		    tupleSpace = new CType[ l_f.length ][];
		    for( int i=0; i<pos.length; i++ ) {
			pos[i] = 0; 
			tupleSpace[i] = new CType[ l_f[i].size() ];
			l_f[i].toArray( tupleSpace[i] );
			if (tupleSpace[i].length == 0) {
			    hasNext = false;
			    break;
			}
		    }
		}

		public boolean hasNext() {
		    return hasNext;
		}

		public Object next() throws NoSuchElementException {
		    if (!hasNext) {
			throw new NoSuchElementException("passed end of set");
		    }
		    
		    // construct tuple by using pos as an index
		    // into the dimensions of tuple space
		    CType[] tuple = new CType[ pos.length ];
		    for( int i=0; i<pos.length; i++) {
			tuple[i] = tupleSpace[i][pos[i]];
		    }

		    // construct the next pos list by
		    // incrementing the first position and rippling
		    // the carries
		    boolean carry = true;
		    int dimn = 0;
		    while( carry && dimn < pos.length) {
			pos[dimn]++;
			pos[dimn] %= tupleSpace[dimn].length;
			carry = (pos[dimn] == 0);
			dimn++;
		    }
			
		    if (carry && dimn >= pos.length) {
			// carried off the end, so done
			hasNext = false;
		    }

		    return tuple;
			 
		}

		public void remove() throws UnsupportedOperationException {
		    throw new UnsupportedOperationException();
		}
	    };
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept( /*@non_null@*/ MjcVisitor p ) {
	p.visitCompilationUnit( this );
    }

    /**
     * Refactors this to include dispatchers for multimethods and
     * other code necessary for running MultiJava code on a standard
     * JVM.  */
    public void translateMJ( /*@non_null@*/ Main compiler ) {
	CCompilationUnitContextType	context = 
	    createCompilationUnitContext( compiler, export );

	try {
	    for( int i = 0; i < typeDeclarations.length ; i++ ) {
		typeDeclarations[i].translateMJ( context );
	    }
	    
	    for( int i=0; i < gfDeclarations.length; i++ ) {
		gfDeclarations[i].translateMJ( context );
	    }
	} finally {
	    compiler.adoptCompilationUnitContext( context );
	}
    }

    // -------------------------------------------------------------
    // CompilerPassEnterable
    // -------------------------------------------------------------

    /**
     * Compares this to a given object.
     *
     * <pre><jml>
     also
     ensures (* \result == lexically ordered according to the file name 
     without the suffix *);
     also
     requires o != null && !(o instanceof JCompilationUnit);
     signals_only ClassCastException;
     * </jml></pre>
     *
     *
     * @param o an <code>Object</code> value to compare to this
     * @return based on a lexical ordering according to the type name 
     * @exception ClassCastException if <code>o</code> is incomparable to this
     */
    public int compareTo(/*@ non_null @*/ Object o) throws ClassCastException {
	JCompilationUnit other = (JCompilationUnit) o;
	int res = fileIdent.compareTo(other.fileNameIdent());
	return res;
    }
    
    /**
     * Caches the arguments for the compiler passes.
     *
     * <pre><jml>
     * also
     * ensures arePassParametersCached;
     * </jml></pre> 
     *
     * @see CompilerPassEnterable
     */
    public void cachePassParameters( Main compiler, 
				     Destination destination ) {
	cachedCompiler = compiler;
	//cachedDestination = destination;
    }

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    public void preprocessDependencies() throws PositionedError {
	try {
	    preprocessDependencies(cachedCompiler);
	} catch (RuntimeException e) {
	    CCompilationUnitContextType context = 
		createCompilationUnitContext( cachedCompiler, export );
	    // This is a debugging aid that can be removed when all bugs
	    // are fixed.  It helps by noting the position in the file
	    // begin parsed/checked that triggers a RuntimeException that
	    // ought to be fixed. !FIXME! - DRCok
	    warn(context, MjcMessages.RUNTIME_EXCEPTION,
		 "while typechecking compilation unit");
	    cachedCompiler.adoptCompilationUnitContext(context);
	    throw e;
	}
    }

    public void checkInterface() throws PositionedError {
	checkInterface(cachedCompiler);
    }

    public void checkInitializers() throws PositionedError {
	checkInitializers(cachedCompiler);
    }

    public void resolveSpecializers() throws PositionedError {
	resolveSpecializers(cachedCompiler);
    }
    
    public void typecheck() throws PositionedError {
	typecheck(cachedCompiler);
    }

    public void translateMJ() {
	translateMJ(cachedCompiler);
    }

    private Main cachedCompiler;
    //@             in arePassParametersCached;
    //@ private represents arePassParametersCached <- cachedCompiler != null;

    //    private Destination cachedDestination;

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private /*@non_null*/ JPackageName	packageName;
    private /*@non_null*/ JClassOrGFImportType[]	importedClasses;
    private /*@non_null*/ JClassOrGFImportType[]	importedGFs;
    private /*@non_null*/ ArrayList		importedUnits; // of JClassOrGFImportType
    private /*@non_null*/ JPackageImportType[]	importedPackages;
    private /*@non_null*/ JTypeDeclarationType[]	typeDeclarations;
    private /*@non_null*/ String fileIdent;

    /**
     * AST subtrees for the augmenting method declarations.  Each
     * element of the list is of type MJTopLevelMethodDeclaration */
    private /*@non_null*/ ArrayList		tlMethods;

    /**
     * Stores references to the collections of private generic
     * functions declared in this compilation unit.  */
    private /*@non_null*/ MJGenericFunctionDecl[]	privateGFDeclarations;
				
    /**
     * Stores references to the collections of non-private generic
     * functions declared in this compilation unit.  */
    private /*@non_null*/ MJGenericFunctionDecl[]	gfDeclarations;
				
    /**
     * Maps from identifiers (of type <code>String</code>) to class
     * type singletons (of type <code>CClassType</code>).  Aliased in
     * CCompilationUnit. 
     *
     * @see CClassType
     * @see CCompilationUnit
     * @see CCompilationUnit#loadedClasses 
     */
    private /*@non_null*/ Hashtable allLoadedClasses = new Hashtable();

    /**
     * Maps from identifiers to sets of external generic function
     * collections that are in scope for this compilation unit.
     * Single-gf-import-on-demand clauses and local external generic
     * function declarations are loaded as early as possible.  Generic
     * functions that are in scope because of the package declaration
     * or import-on-demand clauses are added to the map lazily via
     * alias in CCompilationUnit. 
     * @see CCompilationUnit#loadedGFCollections
     */
    private /*@non_null*/ CGFCollectionMap inScopeGenFuncMap = new CGFCollectionMap();

    /**
     * Provides name space management, e.g., resolving references to
     * external types.  */
    private final /*@non_null*/ CCompilationUnit export;

    private static /*@non_null*/ HashSet PRIMITIVE_TYPES_SET =
	new HashSet( Arrays.asList( CStdType.PRIMITIVE_TYPES ) );

    private static final /*@non_null*/ Iterator EMPTY_ITERATOR =
	new Iterator() {
	    public boolean hasNext() {
		return false;
	    }

	    public Object next() throws NoSuchElementException {
		throw new NoSuchElementException("beyond end of set");
	    }

	    public void remove() throws UnsupportedOperationException {
		throw new UnsupportedOperationException();
	    }
	};

    /**
     * Compares two CMethod objects based on identifiers and parameter
     * types, but only comparing these alphabetically. This is useful
     * for ensuring consistency of output, but not for much else. */
    private static final /*@non_null@*/ Comparator VGF_COMPARATOR =
	new Comparator() {
	    public int compare(/*@non_null@*/ Object o1, /*@non_null@*/ Object o2) { // !CONVERT!
		CMethod m1 = (CMethod) o1;
		CMethod m2 = (CMethod) o2;
		int result = m1.ident().compareTo(m2.ident());
		if (result != 0) {
		    return result;
		}
		CSpecializedType[] m1Parms = m1.parameters();
		CSpecializedType[] m2Parms = m2.parameters();
		result = m2Parms.length - m1Parms.length;
		if (result != 0) {
		    return result;
		}
		for (int i = 0; i < m1Parms.length; i++) {
		    result = m1Parms[i].staticType().
			compareTo(m2Parms[i].staticType());
		    if (result != 0) {
			return result;
		    }
		}
		for (int i = 0; i < m1Parms.length; i++) {
		    result = m1Parms[i].dynamicType().
			compareTo(m2Parms[i].dynamicType());
		    if (result != 0) {
			return result;
		    }
		}
		return 0;
	    }
	};
}
