/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCompilationUnitType.java,v 1.16 2006/12/20 06:15:52 perryjames Exp $
 */

package org.multijava.mjc;

import java.io.File;
import java.util.ArrayList;
import org.multijava.util.Destination;
import org.multijava.util.compiler.PositionedError;

/**
 * This type represents a single Java compilation unit (typically a file
 * in a file-based compiler like this) in the AST.  The production for a
 * compilation unit is the main entry point in Java parser grammar.
 */
public interface JCompilationUnitType extends CompilerPassEnterable {
    /**
     * @return the package name of this compilation unit
     */
    /*@non_null@*/ String packageNameAsString();

    /**
     * Returns the file associated with this compilation unit.
     */
    File file();

    /**
     * Returns the Java identifier associated with the name of the
     * file containing this compilation unit.  For example, if the
     * file is named &quot;typecheck.java&quot; then this method
     * returns &quot;typecheck&quot;.
     *
     * @return the Java identifier associated with the file containing this */
    /*@non_null@*/ String fileNameIdent();

    /*@non_null@*/ JPackageName packageName();

    JPackageImportType[] importedPackages();

    JClassOrGFImportType[] importedClasses();

    JTypeDeclarationType[] typeDeclarations();

    MJGenericFunctionDecl[] gfDeclarations();

    /**
     * Returns the CSourceClass objects representing the type
     * signatures of the types declared in this compilatoin unit
     * <em>and as nested types</em>.
     */
    CSourceClass[] allTypeSignatures();

    /**
     * Returns true iff this compilation unit contains a declaration
     * for a type with the given fully qualified name.
     *
     * @param qualifiedName a fully qualified name with parts
     * separated by '/' not '.'
     */
    boolean declaresType(/*@non_null@*/ String qualifiedName);

    /**
     * Returns true iff this compilation unit contains a declaration
     * for a generic function with the given fully qualified name.
     *
     * @param qualifiedName a fully qualified name with parts
     * separated by '/' not '.'
     */
    boolean declaresGF(/*@non_null@*/ String qualifiedName);

    /**
     * Returns the AST subtrees for the augmenting method
     * declarations.  Each element of the list is of type
     * MJTopLevelMethodDeclaration */
    ArrayList tlMethods();

    /**
     * Returns an array of the class and generic function imports for
     * this compilation unit.  Typically code should use
     * importedClasses or importedGFs instead.  This method is
     * available as a testing harness since the other two methods
     * report empty results until after the preprocessDependencies pass.
     */
    JClassOrGFImportType[] importedUnits();

    JClassOrGFImportType[] importedGFs();

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    void preprocessDependencies( /*@non_null@*/ Main compiler ) 
	throws PositionedError;

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type signatures of
     * everything (imported generic functions, classes being compiled,
     * methods, fields, etc...)  needed for the later passes.  This
     * information is stored in a CCompilationUnit instance and
     * instances of CMember that are bound to the AST.  Also adds
     * things like the default constructor and common initializer
     * methods (which group all field initializers and initializer
     * blocks) to the AST.
     *
     * @exception	PositionedError		an error with reference 
     * 						to the source file */
    void checkInterface( /*@non_null@*/ Main compiler ) throws PositionedError;

    /**
     * Checks the static initializers created during the
     * checkInterface pass and performs some other checks that can be
     * performed simply before full blown typechecking.  Among these
     * other checks are those for appropriateness of subclassing, such
     * as inheritance circularities and &quot;reimplementation&quot;
     * of interfaces.
     *
     * @param	compiler	the compiler that is calling (passed down 
     *				through the AST via CContextType subtypes
     *				and used for error reporting)
     * @exception PositionedError if check fails */
    void checkInitializers( /*@non_null@*/ Main compiler ) throws PositionedError;

    /**
     * Resolves value specializer expressions to the compile-time
     * constants they represent.  Must come after the
     * checkInitializers phase.
     */
    void resolveSpecializers() throws PositionedError;

    
    /**
     * Finds the top method of every declared method.  Called after
     * the resolveSpecializers pass.  This cannot be done before then
     * because the external generic function mappings are not complete
     * until the end of the checkInterface pass and the constant value
     * specializers are not known until after the resolveSpecializers
     * pass.  This must be done before the typecheck pass so that all
     * specialized argument positions for generic functions are known
     * for ambiguity checking. */
    public void resolveTopMethods() throws PositionedError;

    /**
     * Typechecks this compilation unit.
     *
     * @param	compiler	the compiler that is calling
     * @exception PositionedError 	if a check fails that we cannot
     * 					recover from */
    void typecheck( /*@non_null@*/ Main compiler ) throws PositionedError;

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    void accept( /*@non_null@*/ MjcVisitor p );

    /**
     * Refactors this to include dispatchers for multimethods and
     * other code necessary for running MultiJava code on a standard
     * JVM.  */
    void translateMJ( /*@non_null@*/ Main compiler );

    /**
     * Caches the arguments for the compiler passes.
     *
     * <pre><jml>
     * ensures arePassParametersCached;
     * </jml></pre> 
     *
     * @see CompilerPassEnterable
     */
    void cachePassParameters( /*@non_null@*/ Main compiler, /*@non_null@*/ Destination destination );
}
