/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JCompoundAssignmentExpression.java,v 1.8 2005/03/30 03:45:45 xi_hao Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the compound assignment operation, e.g.,
 * <code>x += 3</code> and <code>y *= 4</code>, in the AST.  */
public class JCompoundAssignmentExpression extends JAssignmentExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	oper		the assignment operator
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JCompoundAssignmentExpression( TokenReference where,
					  int oper,
					  JExpression left,
					  JExpression right ) {
	super( where, left, right );
	this.oper = oper;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.  Per JLS 15.25.2 this
     * expression is converted from something like <code>x += 3</code>
     * to something like <Code>x = (x.getType()) (x + 3)</code>.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError {
	isInSafeMathMode  = context.arithmeticMode().equals(AMID_SAFE_MATH);

	// discardValue = false: check if initialized
	left = left.typecheck( context, false, true, false );
	check( context,
	       left.isMaybeInitializable() 
	       && ((CInitializable) left).isLValue( context ),
	       MjcMessages.ASSIGNMENT_NOTLVALUE );

	// try to assign: will throw error if lhs is final and not
	// definitely unassigned (JLS2 16)
	try {
	    ((CInitializable) left).initialize( context );
	} catch (UnpositionedError e) {
	    throw e.addPosition( getTokenReference() );
	}

	right = right.typecheck( context, false, false, false );

	boolean convertRight = true;
	try {
	    switch (oper) {
	    case OPE_STAR:
		type = JMultExpression.computeType( left.getType(),
						    right.getType() );
		break;
	    case OPE_SLASH:
		type = JDivideExpression.computeType( left.getType(),
						      right.getType() );
		break;
	    case OPE_PERCENT:
		type = JModuloExpression.computeType( left.getType(),
						      right.getType() );
		break;
	    case OPE_PLUS:
		type = JAddExpression.computeType( left.getType(),
						   right.getType() );
		    // don't convert the right type if the result type
		    // is a String because we're doing one of those
		    // funky String += operations, which allow any
		    // type on the right-hand side
		if ( type.equals(CStdType.String) ) {
		    convertRight = false;
		}
		break;
	    case OPE_MINUS:
		type = JMinusExpression.computeType( left.getType(),
						     right.getType() );
		break;
	    case OPE_SL:
	    case OPE_SR:
	    case OPE_BSR:
		type = JShiftExpression.computeType( left.getType(),
						     right.getType() );
		// right need not match left, but it must be an integer
		convertRight = false;
		right = right.convertType( CStdType.Integer, context );
		break;
	    case OPE_BAND:
	    case OPE_BXOR:
	    case OPE_BOR:
		type = JBitwiseExpression.computeType( left.getType(),
						       right.getType() );
		break;
	    }
	} catch (UnpositionedError e) {
	    throw e.addPosition(getTokenReference());
	}
	check( context,
	       type.isCastableTo( left.getType() ),
	       MjcMessages.ASSIGNMENT_BADTYPE,
	       right.getType().toVerboseString(), // WMD
	       left.getType().toVerboseString() );
	if( convertRight && !right.isAssignableTo( left.getType() ) ) {
	    context.reportTrouble(
		      new CWarning( getTokenReference(),
				    MjcMessages.NARROWING_COMPOUND_ASSIGNMENT,
				    right.getType(), left.getType()));
	}

	type = left.getType();

	if( convertRight ) {
	    right = right.convertType( type, context );
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitCompoundAssignmentExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	boolean	discardValue = code.discardValue();

	code.setDiscardValue(false);

	((CInitializable)left).genStartStoreCode(code);

	code.setDiscardValue(false);	//!!! graf 991210 ? IS IT A BUG ?

	if (oper == OPE_PLUS && type.equals(CStdType.String)) {
	    left.genCode(code);
	    right.genCode(code);
	    if (!right.getType().equals(CStdType.String)) {
		code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
							       "java/lang/String",
							       "valueOf",
							       "(" + right.getType().getSignature() + ")Ljava/lang/String;"));
	    }

	    code.plantInstruction(new MethodRefInstruction(opc_invokevirtual,
							   "java/lang/String",
							   "concat",
							   "(Ljava/lang/String;)Ljava/lang/String;"));

	    ((CInitializable)left).genEndStoreCode(code, discardValue);
	    return;
	}

	left.genCode(code);
	right.genCode(code);

	int		opcode = -1;

	switch (oper) {
	case OPE_STAR:
	    if (isInSafeMathMode){
                if (getType().equals(CStdType.Long)) {
		    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "multiply",
                                                                   "(JJ)J"));
                } else  if ((getType().equals(CStdType.Double))||(getType().equals(CStdType.Float))){
                    code.plantInstruction(new NoArgInstruction(JMultExpression.getOpcode(getType())));
                } else {
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "multiply",
                                                                   "(II)I"));
                }

	    }else{
		opcode = JMultExpression.getOpcode(getType());
	    }
	    break;
	case OPE_SLASH:
	    if (isInSafeMathMode){
                if (getType().equals(CStdType.Long)) {
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "divide",
                                                                   "(JJ)J"));
                } else  if ((getType().equals(CStdType.Double))||(getType().equals(CStdType.Float))){
                    code.plantInstruction(new NoArgInstruction(JDivideExpression.getOpcode(getType())));
                } else {
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "divide",
                                                                   "(II)I"));
                }

            }else{
		opcode = JDivideExpression.getOpcode(getType());
	    }
	    break;
	case OPE_PERCENT:
            if (isInSafeMathMode){
		code.plantInstruction(new NoArgInstruction(JModuloExpression.getOpcode(getType())));
	    } else {
		opcode = JModuloExpression.getOpcode(getType());
	    }
	    break;
	case OPE_PLUS:
            if (isInSafeMathMode){
                if (getType().equals(CStdType.Long)) {
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "add",
                                                                   "(JJ)J"));
                } else  if ((getType().equals(CStdType.Double))||(getType().equals(CStdType.Float))){
                    code.plantInstruction(new NoArgInstruction(JAddExpression.getOpcode(getType())));
                } else {
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "add",
                                                                   "(II)I"));
                }

            }else{
		opcode = JAddExpression.getOpcode(getType());
	    }
	    break;
	case OPE_MINUS:
	    if (isInSafeMathMode){
                if (getType().equals(CStdType.Long)) {
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "subtract",
                                                                   "(JJ)J"));
                } else  if ((getType().equals(CStdType.Double))||(getType().equals(CStdType.Float))){
                    code.plantInstruction(new NoArgInstruction(JMinusExpression.getOpcode(getType())));
                } else {
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "subtract",
                                                                   "(II)I"));
                }

            } else {
		opcode = JMinusExpression.getOpcode(getType());
	    }
	    break;
	case OPE_SL:
	case OPE_SR:
	case OPE_BSR:
            if (isInSafeMathMode){
                code.plantInstruction(new NoArgInstruction(JShiftExpression.getOpcode(oper, getType())));
            } else {
                opcode = JShiftExpression.getOpcode(oper,getType());
            }
	    break;
	case OPE_BAND:
	case OPE_BXOR:
	case OPE_BOR:
            if (isInSafeMathMode){
                code.plantInstruction(new NoArgInstruction(JBitwiseExpression.getOpcode(oper,getType())));
            } else {
                opcode = JBitwiseExpression.getOpcode(oper,getType());
            }
	    break;
	}
	if (!isInSafeMathMode) {
	    code.plantInstruction(new NoArgInstruction(opcode));
	}
	((CInitializable)left).genEndStoreCode(code, discardValue);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    //@ pure
    public int oper() { return oper; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected  int		oper;
    private boolean isInSafeMathMode  = false;
}
