/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JConditionalAndExpression.java,v 1.9 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class implements the conditional and operation
 */
public class JConditionalAndExpression extends JBinaryExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JConditionalAndExpression(TokenReference where,
				     JExpression left,
				     JExpression right) {
	super(where, left, right);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return left.toString() + " && " + right.toString();
    }

    /**
     * Returns a list of expressions known to be non-null (null) in this context
     */
    public /*@pure non_null*/ Object[] getFANonNulls(/*@non_null@*/ CContextType context) {
	    Object[] fromLeft =left.getFANonNulls(context);
	    Object[] fromRight=right.getFANonNulls(context);
	    Object[] retVal=new JExpression[(fromLeft.length+fromRight.length)];
	    concatArrays(fromLeft, fromRight, retVal);
	    return retVal;
    }
    // no need to override getFANulls since no useful information can be 
    // gathered about nulls

    /*@
      @ requires retVal.length >= a.length+b.length;
      @*/
    public static void concatArrays(/*non_null*/ Object[] a, /*non_null*/ Object[] b, /*non_null*/ Object[] retVal) {
	    for (int i=0; i<a.length; i++) {
		    retVal[i]=a[i];
	    }
	    for (int i=0; i<b.length; i++) {
		    retVal[a.length+i]=b[i];
	    }
    }
 
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError {
	left = left.typecheck( context );
        CExpressionContextType rightContext = context.createExpressionContext_();
	Object[] nonNulls=left.getFANonNulls(context);
	rightContext.addFANonNulls(nonNulls);
	right = right.typecheck( rightContext );

	CType leftType = left.getType();
	CType rightType = right.getType();

	check( context,
	       leftType == CStdType.Boolean,
	       MjcMessages.BOOLEAN_EXPECTED, leftType );
	
	check( context,
	       rightType == CStdType.Boolean,
	       MjcMessages.BOOLEAN_EXPECTED, rightType );

	type = CStdType.Boolean;

	// Optimize constants--need to be very careful here to
	// preserve side-effects.  If both left and right are constant
	// we can create a single new boolean literal from them.  If
	// just left is constant then, because of short-circuit
	// evaluation, we return left if left is false and return
	// right if left is true.  If just right is constant and right
	// is true we return left.  If just right is constant and
	// right is false we cannot optimize because left needs to be
	// evaluated for its side-effects.
	if( left.isConstant() && right.isConstant() ) {
	    JBooleanLiteral leftLit = left.getBooleanLiteral();
	    JBooleanLiteral rightLit = right.getBooleanLiteral();
	    return new JBooleanLiteral( getTokenReference(), 
					leftLit.booleanValue() 
					&& rightLit.booleanValue() );
	} else if( left.isConstant() ) { // && !right.isConstant()
	    if( left.getBooleanLiteral().booleanValue() ) {
		return right;
	    } else {
		return left;
	    }
	} else if( right.isConstant() ) { // && !left.isConstant()
	    if( right.getBooleanLiteral().booleanValue() ) {
		return left;
	    }
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitConditionalAndExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
			     JExpression right,
			     boolean cond,
			     CodeSequence code,
			     CodeLabel label)
    {
	if (cond) {
	    CodeLabel		skip = new CodeLabel();
	    left.genBranch(false, code, skip);
	    right.genBranch(true, code, label);
	    code.plantLabel(skip);
	} else {
	    left.genBranch(false, code, label);
	    right.genBranch(false, code, label);
	}
    }
}
