/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JConditionalExpression.java,v 1.14 2006/12/13 20:59:34 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class implements expressions using the conditional operator.
 * I.e., for expressions using <code> ? :</code>. */
public class JConditionalExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	cond		the condition operand
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JConditionalExpression( TokenReference where,
				   JExpression cond,
				   JExpression left,
				   JExpression right )
    {
	super( where );
	this.cond = cond;
	this.left = left;
	this.right = right;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression cond() { return cond; }
    public JExpression left() { return left; }
    public JExpression right() { return right; }

    public String toString() {
	return cond.toString() + " ? " + left.toString()
	    + " : " + right.toString();
    }
    
    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return type;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return left.isNonNull(context) && right.isNonNull(context);
    }

    public void setType(CType t){
	type = t;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	cond = cond.typecheck( context );

	Object[] nonNulls=cond.getFANonNulls(context);
	Object[] nulls   =cond.getFANulls(context);

        CExpressionContextType rightContext = context.createExpressionContext_();

	context.addFANonNulls(nonNulls);
	rightContext.addFANonNulls(nulls);

	left = left.typecheck( context );
	right = right.typecheck( rightContext );
	context.mergeNullityInfo(rightContext);

	check( context, 
	       cond.getType() == CStdType.Boolean, 
	       MjcMessages.TRINARY_BADCOND );

	CType leftType = left.getType();
	CType rightType = right.getType();

	if( leftType.equals( rightType ) ) {
	    if( leftType == CStdType.Null ) {
		type = rightType;
	    } else {
		type = leftType;
	    }
	} else if( leftType.isNumeric() && rightType.isNumeric() ) {
	    type = unifyNumeric( leftType, rightType );
	    if (type == null) {
		type = ((CNumericType)leftType).binaryPromote( rightType );
	    }
	    // !FIXME! this check will always succeed
//  	    check( context,	
//  		   type != null, 
//  		   MjcMessages.TRINARY_INCOMP, leftType, rightType);
	    left = left.convertType( type, context );
	    right = right.convertType( type, context );
	} else if( leftType.isReference() && rightType.isReference() ) {
	    if( leftType == CStdType.Null ) {
		type = rightType;
	    } else if( rightType == CStdType.Null ) {
		type = leftType;
	    } else if( leftType.isAlwaysAssignableTo(rightType) ) {
		type = rightType;
	    } else if( rightType.isAlwaysAssignableTo(leftType) ) {
		type = leftType;
	    } else {
		check( context, false, 
		       MjcMessages.TRINARY_INCOMP, leftType, rightType);
	    }
	} else {
	    check( context, false, 
		   MjcMessages.TRINARY_INCOMP, leftType, rightType);
	}

        // We have to apply capture conversion (JLS3 5.1.10) to the type of the
        // conditional expression (see JLS3 15.25).
        try {
            type = type.getCapture(context);
        } catch (UnpositionedError e) {
            throw e.addPosition(getTokenReference());
        }

	if( cond.isConstant() ) {
	    if( cond.getBooleanLiteral().booleanValue() ) {
		return left;
	    } else {
		return right;
	    }
	}

	return this;
    }

    /**
     * unifyNumeric
     * search the type corresponding to the unification of the two types
     * @param t1 the first type
     * @param t2 the second type
     * @return the corresponding type or null
     */
    private CType unifyNumeric(CType t1, CType t2) {
	if (t1 == CStdType.Short && t2 == CStdType.Byte) {
	    return CStdType.Short;
	} else if (t2 == CStdType.Short && t1 == CStdType.Byte) {
	    return CStdType.Short;
	} else if (t1 == CStdType.Char && (t2 == CStdType.Char || t2 == CStdType.Byte)) {
	    return CStdType.Char;
	} else if (t2 == CStdType.Char && (t1 == CStdType.Char || t1 == CStdType.Byte)) {
	    return CStdType.Char;
	} else {
	    return null;
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitConditionalExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	CodeLabel		rightLabel = new CodeLabel();
	CodeLabel		nextLabel = new CodeLabel();

	cond.genBranch(false, code, rightLabel);		//		COND IFEQ right
	left.genCode(code); 			//		LEFT CODE
	code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
	code.plantLabel(rightLabel);		//	right:
	right.genCode(code);			//		RIGHT CODE
	code.plantLabel(nextLabel);			//	next	...
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private CType		type;
    private JExpression   cond;
    private JExpression   left;
    private JExpression   right;
}
