/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JConditionalOrExpression.java,v 1.9 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class implements the conditional or operation
 */
public class JConditionalOrExpression extends JBinaryExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JConditionalOrExpression(TokenReference where,
				    JExpression left,
				    JExpression right) {
	super(where, left, right);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return left.toString() + " || " + right.toString();
    }
 
     /**
      * Returns a list of expressions known to be null in this context
      */
     public /*@pure non_null*/ Object[] getFANulls(/*@non_null@*/ CContextType context) {
 	    Object[] fromLeft =left.getFANulls(context);
 	    Object[] fromRight=right.getFANulls(context);
 	    Object[] retVal=new Object[(fromLeft.length+fromRight.length)];
 	    JConditionalAndExpression.concatArrays(fromLeft, fromRight, retVal);
 	    return retVal;
     }
     // no need to override FANonNulls since no useful information can be 
     // gathered about non-nulls
  
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck(CExpressionContextType context) 
	throws PositionedError {
	left = left.typecheck(context);
	context.addFANonNulls(left.getFANulls(context));
	right = right.typecheck(context);

	CType	leftType = left.getType();
	CType	rightType = right.getType();

	check( context,
	       leftType == CStdType.Boolean,
	       MjcMessages.BOOLEAN_EXPECTED, leftType );
	
	check( context,
	       rightType == CStdType.Boolean,
	       MjcMessages.BOOLEAN_EXPECTED, rightType );
	
	type = CStdType.Boolean;

	// Optimize constants--need to be very careful here to
	// preserve side-effects.  If both left and right are constant
	// we can create a single new boolean literal from them.  If
	// just left is constant then, because of short-circuit
	// evaluation, we return left if left is true and return
	// right if left is false.  If just right is constant and right
	// is false we return left.  If just right is constant and
	// right is true we cannot optimize because left needs to be
	// evaluated for its side-effects.
	if( left.isConstant() && right.isConstant() ) {
	    JBooleanLiteral leftLit = left.getBooleanLiteral();
	    JBooleanLiteral rightLit = right.getBooleanLiteral();
	    return new JBooleanLiteral( getTokenReference(), 
					leftLit.booleanValue() 
					|| rightLit.booleanValue() );
	} else if( left.isConstant() ) { // && !right.isConstant()
	    if( left.getBooleanLiteral().booleanValue() ) {
		return left;
	    } else {
		return right;
	    }
	} else if( right.isConstant() ) { // && !left.isConstant()
	    if( !right.getBooleanLiteral().booleanValue() ) {
		return left;
	    }
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitConditionalOrExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
			     JExpression right,
			     boolean cond,
			     CodeSequence code,
			     CodeLabel label)
    {
	if (cond) {
	    left.genBranch(true, code, label);
	    right.genBranch(true, code, label);
	} else {
	    CodeLabel		skip = new CodeLabel();
	    left.genBranch(true, code, skip);
	    right.genBranch(false, code, label);
	    code.plantLabel(skip);
	}
    }
}
