/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JConstructorBlock.java,v 1.13 2005/06/12 19:37:48 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.LocalVarInstruction;
import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents the block of a constructor.  A block is a
 * sequence of statements and local variable declaration statements
 * within braces.  JLS 14.2: Block
 */
public class JConstructorBlock extends JBlock {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param   where           the line of this node in the source code
     */
    public JConstructorBlock( TokenReference where, JStatement[] body ) {
        super( where, body, null );
    }

    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context   the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( /*@non_null@*/ CFlowControlContextType context ) 
        throws PositionedError 
    {
        CFlowControlContextType inside = context;
        CClassContextType classContext = context.getClassContext();

        // set explicitSuper to be an explicit constructor invocation
        // (JLS2, 8.8.5.1), either by creating a new call or by
        // extracting the first expression of the body and replacing
        // it with a "NOP" (that's No OPeration for those of you that
        // didn't write 6502 assembly code)
        sourceClass = (CSourceClass) classContext.getHostClass();
        if( body == null || body.length == 0
            || !(body[0] instanceof JExpressionStatement
                 && ((JExpressionStatement) body[0]).getExpression() 
                    instanceof JExplicitConstructorInvocation) ) {
            if( sourceClass.getType().equals(CStdType.Object) ) {
                explicitSuper = null;
            } else {
                explicitSuper = 
                    new JExplicitConstructorInvocation( getTokenReference(),
                                                        null, JAV_SUPER,
                                                        JExpression.EMPTY );
            }
        } else if( body[0] instanceof JExpressionStatement
                   && ((JExpressionStatement)body[0]).getExpression() 
                   instanceof JExplicitConstructorInvocation ) {
            explicitSuper = 
                (JExplicitConstructorInvocation)
                ((JExpressionStatement) body[0]).getExpression();
            body[0] = new JEmptyStatement( getTokenReference(), null );
        }

        // The following is informal because getType() is not pure.
        /*@ assert (* sourceClass.getType().equals(CStdType.Object) *)
          @        || explicitSuper != null; @*/

        paramsSize = context.getMethodContext().getCMethod().parametersSize();

        // If the explicit constructor call to start this constructor
        // is not an alternative constructor call, then we have some
        // work to do.
        if (explicitSuper == null || !explicitSuper.isThisInvoke()) {
            //  Store the synthetic outer-this parameter if there is one.
            if (sourceClass.hasSyntheticOuterThis()) {
                TokenReference where = getTokenReference();
                JExpression left = 
                    new JNameExpression( where,
                                         new JThisExpression( where ),
                                         JAV_OUTER_THIS );
                JExpression right =
                    new JNameExpression( where, JAV_OUTER_THIS );
                JExpression assn = 
                    new JAssignmentExpression( where, left, right );
                outerFieldInitializer = 
                    new JExpressionStatement( where, assn, null );
            }

            // If the class containing this constructor has an
            // instance initializer, then add a call to the
            // initializer.
            if (classContext.hasInitializer()) {
                addInitializerBlockCall();
            }
        }

        if( explicitSuper != null ) {
            explicitSuper.typecheck( inside.createExpressionContext() );
        }

        if( explicitSuper != null && explicitSuper.isThisInvoke() ) {
            classContext.markAllFieldsAsInitialized();
        }

        if (outerFieldInitializer != null) {
            outerFieldInitializer.typecheck( inside );
        }

        super.typecheck( inside );

        inside.checkingComplete();
		
	// scdaniel: universe runtime checks
	if ( CTopLevel.getCompiler().universeDynChecks() ) {
	    /*
	     * Call setOwner with the current object (this) as parameter.
	     * Before the constructor is called, the currently active object
	     * (owner of this object) and the modifier (rep or peer) are saved
	     * in the current thread's record in the UrtHashtable
	     * by a call to setConstructorData.
	     */
	    universeTransformation =
		new JMethodCallExpression(TokenReference.NO_REF,
				  CUniverseRuntimeHelper.getHashTable(),
					  "setOwner",
					  new JExpression[] { new JThisExpression(TokenReference.NO_REF) },
					  false);
	    
	    // temporarily disable universe checks ...
	    boolean universeChecks = CTopLevel.getCompiler().universeChecks();
	    boolean universePurity = CTopLevel.getCompiler().universePurity();
	    CTopLevel.getCompiler().setUniverseChecks(false);
	    CTopLevel.getCompiler().setUniversePurity(false);

	    // ... typecheck ...
	    universeTransformation.typecheck(inside.createExpressionContext());

	    // ... enable universe checks again.
	    CTopLevel.getCompiler().setUniverseChecks(universeChecks);
	    CTopLevel.getCompiler().setUniversePurity(universePurity);
	}
    }

    protected void addInitializerBlockCall() {
        initBlockCallNeeded = true;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Generates a sequence of bytecodes
     * @param   code            the code list
     */
    public void genCode(CodeSequence code) {
        code.setLineNumber(getTokenReference().line());

	if (explicitSuper != null) {
	    explicitSuper.genCode(code);
	} 
	
	/*
	 * scdaniel: universe runtime checks
	 * 
	 * Create the code for the transformation before anything else
	 * happens in the constructor. We need the newly created object
	 * to be registered to the UrtHashtable before any
	 * Objects (fields with initializers or variables in the
	 * constructor) are created.
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() )
	    universeTransformation.genCode(code);

        // Some nested classes (anonymous classes and local class
        // declarations) have access to final local variables.  This
        // access is accomplished in bytecode by passing the values of
        // the variables as synthetic parameters to the nested class's
        // constructor.  This code stores these parameters in
        // synthetic fields of the owner of this.
        if (sourceClass.isNestedType()) {
            sourceClass.genInit(code, paramsSize);
        }

        if (outerFieldInitializer != null) {
            outerFieldInitializer.genCode(code);
        }

        if (initBlockCallNeeded) {
            code.plantInstruction( new LocalVarInstruction(opc_aload, 0) );
            code.plantInstruction(
                new MethodRefInstruction(opc_invokespecial, 
                                         sourceClass.qualifiedName(),
                                         JAV_INIT,
                                         CType.genMethodSignature(CStdType.Void,
                                                                  CType.EMPTY) ));            
        }

        if (body.length == 0) {
            new JReturnStatement( getTokenReference(), null, null ).
                genCode(code);
        }

        for (int i = 0; i < body.length; i++) {
            body[i].genCode(code);
        }
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExplicitConstructorInvocation explicitSuper() {
        return explicitSuper;
    }

    /**
     * This lame accessor method is provided for JML visitor code.
     */
    public JStatement blockCall() {
        if (initBlockCallNeeded) {
            JNameExpression ident =
                new JNameExpression( getTokenReference(), JAV_INIT );
            JExpression expr = 
                new JMethodCallExpression( getTokenReference(), ident,
                                           JExpression.EMPTY );
            return new JExpressionStatement( getTokenReference(),
                                             expr, null );
        } else {
            return null;
        }
    }

    public void accept( /*@non_null@*/ MjcVisitor v ) {
        v.visitConstructorBlock( this );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected JExplicitConstructorInvocation    explicitSuper;
    private boolean initBlockCallNeeded = false;

    /**
     * Contains the AST for code to initialize local fields of inner
     * classes that hold references to outer classes.  */
    private JStatement                          outerFieldInitializer;
    private CSourceClass                        sourceClass;
    private int                                 paramsSize;
    
    // scdaniel: universe runtime checks
    private JMethodCallExpression universeTransformation;
}
