/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JConstructorDeclaration.java,v 1.20 2008/03/17 16:59:46 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents a constructor in the AST
 */
public class JConstructorDeclaration extends JMethodDeclaration 
    implements JConstructorDeclarationType
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	modifiers	list of modifiers
     * @param	ident		the name of this method
     * @param	parameters	the parameters of this method
     * @param	exceptions	the exceptions throw by this method
     * @param	body		the body of this method, or null if abstract
     * @param	javadoc		is this constructor deprecated
     * @param	comments	regular Java comments 
     */
    public JConstructorDeclaration( TokenReference where,
				    long modifiers, 
				    String ident,
				    JFormalParameter[] parameters,
				    CClassType[] exceptions,
				    JConstructorBlock body,
				    JavadocComment javadoc,
				    JavaStyleComment[] comments ) {
	super( where, modifiers,CTypeVariable.EMPTY, CStdType.Void, ident,
	       parameters, exceptions, body, javadoc, comments );
    }

    /**
     * Constructs an AST node for a default constructor
     *
     * @param	where		the line of this node in the source code
     * @param	modifiers	list of modifiers
     * @param	ident		the name of this method
     */
    public JConstructorDeclaration( TokenReference where, 
				    long modifiers, 
				    String ident ) {
	this( where, modifiers, ident, JFormalParameter.EMPTY,
	      CClassType.EMPTY,
	      new JConstructorBlock( where, new JStatement[0] ), null, null );
    }
    
    /**
     * Constructs an AST node for a compiler-generated constructor.
     * Constructor includes no comments.
     *
     * <pre><jml>
     * ensures areParametersComplete == true;
     * </jml></pre>
     *
     * @param	where		the line of this node in the source code
     * @param	modifiers	list of modifiers
     * @param	ident		the name of this method
     * @param	parameters	the parameters of this method
     * @param	exceptions	the exceptions throw by this method
     * @param body the body of this method, or null if abstract */
    public JConstructorDeclaration(TokenReference where,
				   long modifiers, 
				   String ident,
				   JFormalParameter[] parameters,
				   CClassType[] exceptions,
				   JConstructorBlock body ) {
	this( where, modifiers, ident, parameters, exceptions, 
	      body, null, null );
	this.areParametersComplete = true;
    }
    

    public /*@ pure @*/ boolean hasBody() { 
	return body().body() != null;
    }
    

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things
     * generally look OK.  This pass gathers information about
     * the type signatures of everything (imported class
     * files, classes being compiled, methods, fields, etc...)
     * needed for the later passes.  This information is stored 
     * in a context hierarchy that is bound to the AST.
     *
     * <pre><jml>
     * also ensures areParametersComplete == true;
     * </jml></pre>
     *
     *
     * @return true iff sub tree is correct enought to check code
     * @exception	PositionedError	an error with reference to the source file
     */
    public CMember checkInterface(CContextType context) 
	throws PositionedError 
    {
	CClassContextType classContext = context.getClassContext();

	check( context,
	       !hasOtherFlags( stripNonJavaModifiers(modifiers()),
			       ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE ),
	       MjcMessages.CONSTRUCTOR_FLAGS,
	       ident() );

	check( context,
	       ident().equals(classContext.getOwnerClass().ident()),
	       MjcMessages.CONSTRUCTOR_BAD_NAME,
	       ident(),
	       classContext.getOwnerClass().ident() );

	if (!areParametersComplete && 
	    classContext.getOwnerClass().hasSyntheticOuterThis()) {
	    CClass outerClass = classContext.getOwnerClass().owner();
	    CType outerType = outerClass.getType();

	    if( CTopLevel.getCompiler().universeChecks() ) {
		// WMD: get a readonly version of the outer type
		outerType = CTopLevel.getTypeRep(
                              ((CClassType)outerType).qualifiedName(),
			      CUniverseImplicitReadonly.getUniverse(), true );
	    }

	    JFormalParameter outerThisParm =
		new JFormalParameter( getTokenReference(),
				      JLocalVariable.DES_GENERATED,
				      new CSpecializedType(outerType), 
				      JAV_OUTER_THIS, true );
	    addParameter( outerThisParm );
	}
	areParametersComplete = true;
	
	CMember result = super.checkInterface( context );
	return result;
    }

    /**
     * Makes sure that all in-scope external generic functions are
     * added to the appropriate augmentation maps before top method
     * searches occur in later passes.  */
    public void resolveExtMethods( CContextType context ) {
	// nothing to do (except override superclass method to do
	// nothing) since constructors cannot belong to external
	// generic functions
    }

    /**
     * Typechecks this method declaration.  Mutates the context to
     * record the information gathered during the checks.
     *
     * @param context the context in which this method appears
     * @exception PositionedError if the checks fail and the failure
     * cannot be recovered from */
    public void typecheck(CContextType context) throws PositionedError {
	check( context, body() != null, 
	       MjcMessages.CONSTRUCTOR_NOBODY, ident() );

	// Cast will succeed since it was checked in checkInterface pass.
	CClassContextType cContext = (CClassContextType) context;
	CMethodContextType	self = createSelfContext(cContext);
	CFlowControlContextType inside = 
	    self.createFlowControlContext(parameters().length, 
					  getTokenReference());
    
	inside.addThisVariable();
    
	for (int i = 0; i < parameters().length; i++) {
	    parameters()[i].typecheck( inside );
	}
	
	// Checks for Generic Universe Types, if enabled
	if (CTopLevel.getCompiler().universeChecks()) {
		typecheckUniverse(context);
	}

    
	if (body() != null) body().typecheck( inside );
    
	inside.checkingComplete();
	self.verifyExceptions( getTokenReference() );
    }

	/**
	 * Ensures static correctness of Generic Universe Types.
	 * 
	 * Checks performed:
	 * - Types of parameters must not contain rep modifiers
	 * 
	 * @param context The current typechecking context.
	 * 
	 * By WMD
	 * By RZU
	 */
	private void typecheckUniverse(CContextType context) throws PositionedError {
		// Types of parameters should not contain rep modifiers
		for (int i = 0; i < parameters().length; i++) {		
			if( (parameters()[i].getType() instanceof CClassType)
	    		&& CUniverseServices
	    				.containsRepModifier((CClassType) parameters()[i].getType())) {
				warn( context,
						CUniverseMessages.CONSTRUCTOR_REP_IN_SIGNATURE_WARNING, parameters()[i].ident());
			}
		}
	}

    /**
     * Typechecks this method declaration.  Mutates the context to
     * record the information gathered during the checks.
     *
     * @param cContext the context in which this method appears
     * @param instanceInfo the definite-assignment state of the surrounding
     *                     class after instance initializers but before any
     *                     constructors
     * @exception PositionedError if the checks fail and the failure
     * cannot be recovered from */
    public void typecheck( CClassContextType cContext, 
			   CVariableInfoTable instanceInfo) 
	throws PositionedError
    {
	cContext.setFieldInfoTable(instanceInfo);
	typecheck(cContext);
    }
	
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Creates a context for this constructor AST node.
     *
     * @param	parent		the parent context
     */
    public CMethodContextType createSelfContext(CClassContextType parent) {
	return parent.createConstructorContext(getMethod());
    }

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
				// !FIXME! genComments(p);
	p.visitConstructorDeclaration( this );
    }

    // -------------------------------------------------------------
    //  PRIVILEGED DATA
    // -------------------------------------------------------------
    

    /**
     * Flag indicates whether any required changes to the parameters
     * for accepting outer this instance have been completed.  */
    private /*@ spec_public @*/ boolean areParametersComplete = false;

}
