/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JContinueStatement.java,v 1.4 2004/01/16 23:21:10 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * A continue statement may occur only in a while, do, or for statement;
 * statements of these three kinds are called iteration statements.
 * Control passes to the loop-continuation point of an iteration statement.
 * JLS 14.14: Continue Statement
 *
 */
public class JContinueStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param where	the line of this node in the source code
     * @param label	the label of the enclosing labeled statement
     * @param comments	the comments that go with this statement
     */
    public JContinueStatement( TokenReference where, String label, 
			       JavaStyleComment[] comments ) {
	super( where, comments );
	this.label = label;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	if( label != null ) {
	    target = context.getLabeledStatement( label );
	    check( context, target != null, 
		   MjcMessages.LABEL_UNKNOWN, label );
	    check( context, target.acceptsContinue(), 
		   MjcMessages.CONTINUE_NOTLOOP );
	} else {
	    target = context.getNearestContinuableStatement();
	    check( context, target != null, MjcMessages.CANNOT_CONTINUE );
	}

	target.addContinue( context );
	context.setReachable(false);
 
	continueLabel = target.getContinueLabel();
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitContinueStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	code.plantBreak(target);

	code.plantLabelRef(opc_goto, continueLabel);

	continueLabel = null;
	target = null;
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public String label() { return label; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private CodeLabel		continueLabel;
    private String		label;
    private JStatement		target;
}
