/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JDoStatement.java,v 1.9 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents a do statement in the AST.  The do statement
 * executes a statement and an expression repeatedly until the value
 * of the expression is false.  JLS 14.11: Do Statement */
public class JDoStatement extends JLoopStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	cond		the expression to evaluate
     * @param	body		the loop body
     * @param	comments	the comments associated with this statement
     */
    public JDoStatement( TokenReference where, JExpression cond, 
			 JStatement body, JavaStyleComment[] comments ) {
	super( where, comments );

	this.cond = cond;
	this.body = body;
    }


    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    /**
     * Registers that there is a continue statement within this
     * loop. */
    public void addContinue(CFlowControlContextType context) {
	super.addContinue(context);
	// continue in do-statements allows the while condition to be
	// reached, as opposed to while and for statements where
	// continue prevents the following code from being reached
	continueContext.setReachable( true );
    }


    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	try {
	    CLoopContext self = context.createLoopContext( this );

	    body.typecheck( self );

	    if (continueContext != null) {
		if (self.isReachable()) {
		    self.merge( continueContext );
		} else {
		    self.adopt( continueContext );
		}
	    }

	    self.doneWithCheckingBody();
	    self.checkingComplete();

	    cond = cond.typecheck( context.createExpressionContext() );
	    check( context, 
		   cond.getType() == CStdType.Boolean, 
		   MjcMessages.DO_COND_NOTBOOLEAN, cond.getType() );

	    if( cond instanceof JAssignmentExpression ) {
		context.reportTrouble(
			  new CWarning( getTokenReference(),
					MjcMessages.ASSIGNMENT_IN_CONDITION ));
	    }
	} catch( CBlockError e ) {
	    context.reportTrouble( e );
	}

	if( cond.isConstant() && cond.getBooleanLiteral().booleanValue() ) {
	    // Cannot exit a do loop whose condition is a true literal...
	    context.setReachable( false );
	}

	// ...unless there is an appropriate break.
	if (hasBreak()) {
	    if (context.isReachable()) {
		context.merge(getBreakContextSummary());
	    } else {
		context.adopt(getBreakContextSummary());
	    }
	} else {
	    context.addFANonNulls(cond.getFANulls(context));
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitDoStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	code.pushContext(this);

	CodeLabel		bodyLabel = new CodeLabel();

	code.plantLabel(bodyLabel);				//	body:
	body.genCode(code);					//		BODY CODE
	code.plantLabel(getContinueLabel());			//	cont:
	cond.genBranch(true, code, bodyLabel);		//		EXPR CODE; IFNE body
	code.plantLabel(getBreakLabel());				//	end:	...

	code.popContext(this);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression cond() { return cond; }
    public JStatement body() { return body; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		cond;
    private JStatement		body;


}
