/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JEqualityExpression.java,v 1.11 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents the AST node for the equality operators.
 * I.e., for <code>==</code> and <code>!=</code>. JLS2 15.21 */
public class JEqualityExpression extends JBinaryExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	oper		the operator
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JEqualityExpression( TokenReference where,
				int oper,
				JExpression left,
				JExpression right )
    {
	super( where, left, right );
	this.oper = oper;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	String s = (oper == OPE_EQ) ? " == " : " != ";
	return left.toString() + s + right.toString();
    }
    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck(CExpressionContextType context) 
	throws PositionedError {
	left = left.typecheck(context);
	right = right.typecheck(context);
	return typecheckHelper(context,left.getType(),right.getType());
}

    protected JExpression typecheckHelper(CExpressionContextType context, CType leftType, CType rightType) 
	throws PositionedError {

	if( leftType.isNumeric() ) {
	    // JLS 15.20.1: Numerical Equality Operators
	    check( context, 
		   rightType.isNumeric(), 
		   MjcMessages.EQUALITY_TYPE, leftType.toVerboseString(), rightType.toVerboseString());
	    promote( context );
	} else if( leftType == CStdType.Boolean ) {
	    // JLS 15.20.2: Boolean Equality Operators
	    check( context, 
		   rightType == CStdType.Boolean, 
		   MjcMessages.EQUALITY_TYPE, leftType.toVerboseString(), rightType.toVerboseString());
	    if( left instanceof JBooleanLiteral 
		|| right instanceof JBooleanLiteral ) {
		context.reportTrouble(
		   new CWarning( getTokenReference(),
				 MjcMessages.COMPARING_BOOLEAN_CONSTANT ));
	    }
	} else {
	    // JLS 15.20.3: Reference Equality Operators
	    check( context,
		   leftType.isReference() && rightType.isReference(),
		   MjcMessages.EQUALITY_TYPE, leftType.toVerboseString(), rightType.toVerboseString());
	    check( context,
		   rightType.isCastableTo(leftType) 
		   || leftType.isCastableTo(rightType),
		   MjcMessages.EQUALITY_TYPE, leftType.toVerboseString(), rightType.toVerboseString());
	    if( left.getType().equals(CStdType.String) 
		&& right.getType().equals(CStdType.String) 
		&& (left.isConstant() || right.isConstant())) {
		context.reportTrouble( 
			    new CWarning( getTokenReference(),
					  MjcMessages.STRING_COMPARISON ));
	    }
	}

	type = CStdType.Boolean;

	if( left.isConstant() && right.isConstant() ) {
	    return constantFolding( left.getLiteral(), right.getLiteral() );
	} else {
	    return this;
	}
    }

    /**
     * Calculates the boolean literal that is the result of comparing
     * the two given literals for equality (or inequality, depending
     * on <code>oper</code>)
     *
     * @param	left	the literal value of the left expression
     * @param	right	the literal value of the right expression
     * @return a literal resulting of from an operation over two literals */
    public JExpression constantFolding(JLiteral left, JLiteral right) {
	if( left.getType().isNumeric() ) { 
	    if (left.getType() == CStdType.Long) {
		long leftVal =
		    left.getNumberLiteral().numberValue().longValue();
		long rightVal =
		    right.getNumberLiteral().numberValue().longValue();
		if( oper == OPE_EQ ) {
		    return new JBooleanLiteral( getTokenReference(),
						leftVal == rightVal );
		} else {
		    return new JBooleanLiteral( getTokenReference(),
					    leftVal != rightVal );
		}
	    } else if (left.getType() == CStdType.Float) {
		float leftVal =
		    left.getNumberLiteral().numberValue().floatValue();
		float rightVal =
		    right.getNumberLiteral().numberValue().floatValue();
		if( oper == OPE_EQ ) {
		    return new JBooleanLiteral( getTokenReference(),
						leftVal == rightVal );
		} else {
		    return new JBooleanLiteral( getTokenReference(),
						leftVal != rightVal );
		}
	    } else if (left.getType() == CStdType.Double) {
		double leftVal =
		    left.getNumberLiteral().numberValue().doubleValue();
		double rightVal =
		    right.getNumberLiteral().numberValue().doubleValue();
		if( oper == OPE_EQ ) {
		    return new JBooleanLiteral( getTokenReference(),
						leftVal == rightVal );
		} else {
		    return new JBooleanLiteral( getTokenReference(),
						leftVal != rightVal );
		}
	    } else if (left.getType().implicitlyCastTo(CStdType.Integer)) {
		int leftVal =
		    left.getNumberLiteral().numberValue().intValue();
		int rightVal =
		    right.getNumberLiteral().numberValue().intValue();
		if( oper == OPE_EQ ) {
		    return new JBooleanLiteral( getTokenReference(),
						leftVal == rightVal );
		} else {
		    return new JBooleanLiteral( getTokenReference(),
						leftVal != rightVal );
		}
	    } else {
		fail();
		return null;
	    }
	} else if( left.getType() == CStdType.Boolean ) {
	    boolean leftVal = left.getBooleanLiteral().booleanValue();
	    boolean rightVal = right.getBooleanLiteral().booleanValue();
	    if (oper == OPE_EQ) {
		return new JBooleanLiteral( getTokenReference(),
					    leftVal == rightVal );
	    } else {
		return new JBooleanLiteral( getTokenReference(),
					    leftVal != rightVal );

	    }
	} else if( left instanceof JNullLiteral 
		   && right instanceof JNullLiteral ) {
	    return new JBooleanLiteral( getTokenReference(), 
					(oper == OPE_EQ) );
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitEqualityExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
			     JExpression right,
			     boolean cond,
			     CodeSequence code,
			     CodeLabel label)
    {
	code.setLineNumber(getTokenReference().line());

	if ((left.isConstant() || right.isConstant())) {
	    if (left instanceof JNullLiteral || right instanceof JNullLiteral) {
		if (left instanceof JNullLiteral) {
		    JExpression temp = left;
		    left = right;
		    right = temp;
		}
		left.genCode(code);
		code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_ifnull : opc_ifnonnull, label);
		return;
	    } else if (((left instanceof JOrdinalLiteral && ((JOrdinalLiteral)left).numberValue().intValue() == 0) ||
			(right instanceof JOrdinalLiteral && ((JOrdinalLiteral)right).numberValue().intValue() == 0)) &&
		       left.getType().isCodedAsInt()) {
		if ((left instanceof JOrdinalLiteral) &&((JOrdinalLiteral)left).numberValue().intValue() == 0) {
		    JExpression temp = left;
		    left = right;
		    right = temp;
		}
		left.genCode(code);
		code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_ifeq : opc_ifne, label);
		return;
	    }
	}

	left.genCode(code);
	right.genCode(code);

	switch (left.getType().getTypeID()) {
	case TID_ARRAY:
	case TID_CLASS:
	    code.plantLabelRef(cond  == (oper == OPE_EQ) ? opc_if_acmpeq : opc_if_acmpne, label);
	    return;
	case TID_FLOAT:
	    code.plantInstruction(new NoArgInstruction(opc_fcmpl));
	    break;
	case TID_LONG:
	    code.plantInstruction(new NoArgInstruction(opc_lcmp));
	    break;
	case TID_DOUBLE:
	    code.plantInstruction(new NoArgInstruction(opc_dcmpl));
	    break;
	default:
	    code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_if_icmpeq : opc_if_icmpne, label);
	    return;
	}

	code.plantLabelRef(cond == (oper == OPE_EQ) ? opc_ifeq : opc_ifne, label);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public int oper() { return oper; }

    /**
     * Returns a list of expressions known to be non-null in this context
     * Any expression can be returned, but adding them to the contexts will
     * filter it out
     */
    public /*@pure*/ Object[] getFANonNulls(/*@non_null@*/ CContextType context) {
	    if (oper == Constants.OPE_NE) {
		    if (right instanceof JNullLiteral) {
		       return new Object[]{left};
		    } else if (left instanceof JNullLiteral) {
		       return new Object[]{right};
		    }
	    }
	    return EMPTY;
    }
    /**
     * Returns a list of expressions known to be null in this context
     */
    public /*@pure*/ Object[] getFANulls(/*@non_null@*/ CContextType context) {
	    if (oper == Constants.OPE_EQ) {
		    if (right instanceof JNullLiteral) {
		       return new Object[]{left};
		    } else if (left instanceof JNullLiteral) {
		       return new Object[]{right};
		    }
	    }
	    return EMPTY;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected final int			oper;
}

