/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JExplicitConstructorInvocation.java,v 1.16 2008/05/14 18:16:30 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a explicit call to a super or self constructor.
 */
public class JExplicitConstructorInvocation extends JExpression {

    /**
     * Construct a node in the parsing tree.
     * This constructor is directly called by the parser.
     *
     * <pre><jml>
     * ensures argsCompleted == false;
     * </jml></pre>
     *
     * @param	where		the line of this node in the source code
     * @param	prefix		a qualifying outer class for subclasses 
     *				of inner class, see JLS2, 8.8.5.1
     * @param	ident		the method name
     * @param	args		the argument of the call
     */
    public JExplicitConstructorInvocation( TokenReference where,
					   JExpression prefix,
					   String ident,
					   JExpression[] args )
    {
	super( where );

	this.prefix = prefix;
	this.ident = ident;
	this.args = args;

	this.argsCompleted = false;
    }

     /**
     * Construct an AST node for an explicit constructor invocation.
     *
     * <pre><jml>
     * ensures argsCompleted == true;
     * </jml></pre>
     *
     * @param	where		the line of this node in the source code
     * @param	ident		the method name
     * @param args the arguments of the call, including any outer
     * classes for inner class constructors */
    public JExplicitConstructorInvocation( TokenReference where,
					   String ident,
					   JExpression[] args )
    {
	super( where );

	this.prefix = null;
	this.ident = ident;
	this.args = args;

	this.argsCompleted = true;
    }

   

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns true if it's this() else it's super().
     */
    //@ pure
    final boolean isThisInvoke() {
	return ident.equals(JAV_THIS);
    }

    /**
     * Returns true if the prefix is a synthesized one, e.g., JAV_OUTER_THIS.
     */
    public boolean isPrefixSynthesized() {
	return isPrefixSynthesized;
    }

    /**
     * Returns the constructor being called. This method must be called
     * after typechecking.
     */
    public CMethod method() {
	return method;
    }

    /**
     *
     */
    public CType getType() {
	return null;
    }

    public JExpression prefix() { return prefix; }
    public String ident() { return ident; }
    public JExpression[] params() { return args; }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	// Typechecking is according to JLS2, 8.8.5.1

	check( context,  
	       context.isInConstructor(),
	       MjcMessages.CONSTRUCTOR_ILLEGAL_EXPLICIT );

	callingClass = 
	    (CSourceClass) context.getClassContext().getOwnerClass();
	callingConstructor = context.getMethodContext().getCMethod();

	if( isThisInvoke() ) {
	    // the grammar ensures that a alternative constructor
	    // invocation is not qualified
	    assertTrue( prefix==null );

	    clazz = callingClass;
	    if (clazz.hasSyntheticOuterThis()) {
		// The constructor in which this alternate constructor
		// invocation appears must have a synthetic outer-this
		// parameter.  That parameter will be passed to the
		// next constructor.
		prefix = 
		    new JNameExpression( getTokenReference(), JAV_OUTER_THIS ).
			typecheck(context);
		isPrefixSynthesized = true;
	    }
	    
	} else {
	    clazz = context.getClassContext().getOwnerClass().getSuperClass();
	    if (!clazz.isInnerClass()) {
		check( context, prefix==null, 
		       MjcMessages.CONSTRUCTOR_QUALIFIED_EXPLICIT );
	    } else {
		if (prefix != null) {
		    prefix = prefix.typecheck( context );
		    check( context, 
			   prefix.getType().getCClass().
			   descendsFrom(clazz.owner()),
			   MjcMessages.CONSTRUCTOR_BAD_QUALIFIER, 
			   clazz.owner() );
		} else if (!argsCompleted) {
		    prefix = 
			new JThisExpression(getTokenReference(),clazz.owner()).
			    typecheck(context);
		    isPrefixSynthesized = true;
		}
	    }
	}

	assertTrue( clazz != null );
	context.registerVisibleType( clazz.getType() );

	    // make sure the arguments to the explicit constructor call
	    // don't access instance variables or methods, or 'this' or
	    // 'super'
	((CConstructorContextType) context.getMethodContext())
	    .setSuperConstructorCalled( false );

	/*@ assert clazz.isInnerClass() ==> prefix != null;
	    assert (* prefix is an expression that generates the appropriate
	   	      outer class for the explicit constructor invocation *);
	@*/
	CType[]	argsType = new CType[args.length];

	// also need to build a type array that considers synthetic
	// outer this parameter
	final int offset = (prefix == null) ? 0 : 1;
	CType[] searchArgsType = new CType[args.length + offset];
	if (prefix != null) {
	    searchArgsType[0] = prefix.getType();
	}
	
	for( int i = 0; i < argsType.length; i++ ) {
	    args[i] = args[i].typecheck( context );

	    argsType[i] = args[i].getType();
	    searchArgsType[i + offset] = argsType[i];
	    assertTrue( argsType[i] != null );
	}

	try {
	    // CClassType[] sub = prefix.getType().isGenericType()?
	    //prefix.getType().getArguments():CClassType.EMPTY;
	    method = clazz.lookupMethod( JAV_CONSTRUCTOR, searchArgsType, 
					 CClassType.EMPTY,context );
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}

	if( method == null || ! method.owner().equals(clazz) ) {
	    // The lookupMethod generic function does not handle the
	    // special case of constructors and so may return a
	    // constructor from a superclass of clazz.  That would be
	    // bad.  Hence the second term in the disjunction.
	    throw new CMethodNotFoundError( getTokenReference(), 
					    clazz.getType().toString(), 
					    argsType, true );
	}

	CClassType[] exceptions = method.throwables();
	for( int i = 0; i < exceptions.length; i++ ) {
	    if( exceptions[i].isCheckedException() ) {
		context.getFlowControlContext().addThrowable(
		 new CThrowableInfo( exceptions[i],
				     this ));
	    }
	}

	check( context, 
	       !context.isStatic(), 
	       MjcMessages.BAD_THIS_STATIC );

    // WMD
    if( CTopLevel.getCompiler().universeChecks() ) {
    	typecheckUniverseArguments(context, argsType);
    }

	searchArgsType = 
	    CSpecializedType.tupleOfSpecializersFrom( method.parameters() );

	for( int i = 0; i < args.length; i++ ) {
	    args[i] = args[i].convertType(searchArgsType[i + offset], context);
	}

	((CConstructorContextType) context.getMethodContext())
	    .setSuperConstructorCalled(true);

	//  If this is a this() constructor invocation then we need to
	//  register that all final instance fields are initialized in
	//  the flow control context
	if (isThisInvoke()) {
	    CSourceField[] fields = clazz.fieldsForCodeGen().sourceFields();
	    for (int i = 0; i < fields.length; i++) {
		CSourceField f = fields[i];
		if (f.isFinal() && !f.isStatic()) {
		    context.getFlowControlContext().initializeField( f );
		}
	    }
	} else {
	    // We also need to get the expressions for generating the
	    // synthetic outer local variable parameters of clazz.
	    // These expressions are currently just thrown away and
	    // then regenerated at during code generation.  Though
	    // this may be optimizable since the only case where these
	    // expressions exist is when the super class, clazz, is a
	    // local type declaration, in which case it must be
	    // typechecked before callerClass.
	    syntheticSuperParms = 
		clazz.getSyntheticParamsForExplicitSuper( callingClass,
							  callingConstructor );
	    for (int i = 0; i < syntheticSuperParms.length; i++) {
		syntheticSuperParms[i] = 
		    syntheticSuperParms[i].typecheck(context);
	    }
	    
	}
	
	return this;
    }

    
	/**
	 * Check the Universe modifiers of the constructor call.
	 * 
	 * @param context the type checking context
	 * @param argTypes the arguments to the constructor
	 * @throws PositionedError
	 */
	private void typecheckUniverseArguments(CExpressionContextType context,
				CType[] argTypes) throws PositionedError
	{
		// Check the parameters and arguments by iterating over them
		CType[] parameterTypes = CSpecializedType.staticTypeTupleFrom(null,
				method.parameters());
				
		CClassType currentParameterType;
		CClassType currentArgumentType;
		
		// should it be host or owner class?
		CClassType type = context.getClassContext().getHostClass().getType();

		// if it's the constructor of an inner class, then the compiler automatically
		// adds a parameter.
		int i, j;
		if( clazz.isInnerClass() ) {
			i = 1; j = 0;
			assertTrue(parameterTypes.length == argTypes.length + 1);
		} else {
			i = 0; j = 0;
			assertTrue(parameterTypes.length == argTypes.length);
		}
		
		
		for (; i < parameterTypes.length; ++i, ++j) {
			// We only care about parameters of class type
			if (parameterTypes[i] instanceof CClassType) {
				currentParameterType = (CClassType) parameterTypes[i];
				// arg should also be a class type, as we do this check
				// after the normal type checks and a mismatch would have
				// been found already.
				currentArgumentType = (CClassType) argTypes[j];
			} else {
				continue;
			}

			CClassType substitutedParameterType = CUniverseServices
					.substituteInheritedTypeVariables(
							(CClassType) currentParameterType, type, context);

			// Adapt viewpoint of method parameters
			CClassType combinedType = CUniverseServices
					.combine(type, substitutedParameterType);

			// Perform the substitution
			CClassType substitutedType = getFullySubstitutedType(combinedType, context);

			check( context,
				   currentArgumentType.isAlwaysAssignableTo(
							substitutedType),
							CUniverseMessages.ACTUAL_NOT_ASSIGNABLE_TO_FORMAL_PARAMETER,
							new Object[] { new Integer(i+1),
								currentArgumentType.toVerboseString(),
								method.ident(),
								substitutedType.toVerboseString() });
		}
	}
    
    /**
	 * Substitutes the type variables of a method parameter.
	 * 
	 * @param inputtype
	 *            The type of the parameter to substitute
	 * @param context
	 *            The context used for type checking
	 * @return The substituted type containing no type variables
	 * 
	 * copied by WMD from JNewObjectExpression.
	 * 
	 * by RZU
	 */
	private CClassType getFullySubstitutedType(CClassType inputtype,
			CContextType context) throws PositionedError {
		CClassType substitutedType = inputtype;

		// removed possibility of being a method type variable
		// Get the host class
		CClass hostClass = context.getClassContext().getHostClass();

		CClassType type = hostClass.getType(); 
		
		// If the type is a type variable (T a), we use the receiver's type
		// arguments to replace it
		if (inputtype.isTypeVariable()) {
			substitutedType = type.getCClass().getSubstitution(
					(CTypeVariable) inputtype,
					type.getAllArguments());

			// If the type is an array whose type is a type variable (T[] a),
			// we use the receiver's type arguments to replace it
		} else if (inputtype.isArrayType()
				&& ((CArrayType) inputtype).getBaseType().isTypeVariable()) {
			// Get substituted base type...
			substitutedType = type.getCClass().getSubstitution(
					(CTypeVariable) ((CArrayType) inputtype).getBaseType(),
					type.getAllArguments());
			// ... and create an array out of it
			substitutedType = new CArrayType(substitutedType,
					((CArrayType) inputtype).getArrayBound(),
					((CArrayType) inputtype).getCUniverse());

			// If the type is an array with a generic base type (List<Integer>[]
			// a),
			// we use the receiver's type arguments to replace type variables in
			// the type arguments, if any
		} else if (inputtype.isArrayType()
				&& ((CArrayType) inputtype).getBaseType().isGenericType()) {
			CType baseType = ((CArrayType) inputtype).getBaseType();
			CClassType newType = ((CClassType) baseType).createSubstitutedType(
					hostClass, type, baseType
							.getAllArguments());
			substitutedType = new CArrayType(newType, ((CArrayType) inputtype)
					.getArrayBound(), ((CArrayType) inputtype).getCUniverse());

			// If the type is an array with a generic base type (List<Integer>[]
			// a),
			// we use the receiver's type arguments to replace type variables in
			// the type arguments, if any
		} else if (inputtype.isGenericType()) {
			substitutedType = inputtype.createSubstitutedType(hostClass,
					type, hostClass.getType()
							.getAllArguments());
		}

		return substitutedType;
	}

	
	
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitExplicitConstructorInvocation( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());
	code.plantLoadThis();

	if (prefix != null) {
	    prefix.genCode(code);
	}

	for (int i = 0; i < args.length; i++) {
	    args[i].genCode(code);
	}

	if (isThisInvoke()) {
	    clazz.genSyntheticParamsForExplicitThis(code,callingConstructor);
	} else {
	    for (int i = 0; i < syntheticSuperParms.length; i++) {
		syntheticSuperParms[i].genCode(code);
	    }
	}
	

	code.plantInstruction(new MethodRefInstruction(opc_invokespecial,
						       method.getOwnerName(),
						       method.ident(),
						       method.getSignature()));
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------


    /**
     * The class of the constructor invoked by this expression.  
     *
     * <pre><jml>
     * private invariant clazz != null && method != null ==>
     *			method.owner() == clazz;
     * </jml></pre>
     *
     */
    private	CClass		clazz;


    /**
     * The class of the constructor that contains this expression.  */
    private CSourceClass callingClass;

    /**
     * The constructor that contains this expression. */
    private CMethod callingConstructor;

    private	CMethod		method;
    private	JExpression	prefix;
    private	String		ident;
    private	JExpression[]	args;

    /** <code>true</code> if the prefix is a synthesized one (e.g.,
     * JAV_OUTER_THIS) as the result of typechecking. */
    private boolean isPrefixSynthesized;

    /**
     * Flag indicates whether the arguments have been processed to
     * include outer this parameters for inner classes.  */
    //@ spec_public
    private boolean argsCompleted;

    /**
     * Stores the expressions needed to push outer local variables
     * onto the call stack when invoking a superclass constructor.
     *
     * <pre><jml>
     * private invariant isThisInvoke() <==> syntheticSuperParms == null;
     * </jml></pre>
     * */
    private JExpression[] syntheticSuperParms;
}
