/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JExpression.java,v 1.21 2006/12/20 06:15:52 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.MessageDescription;

import java.util.List;
import java.util.ArrayList;

/**
 * This class is the root class for all classes representing expression 
 * nodes in the AST.
 */
public abstract class JExpression extends JPhylum  {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------
  
    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     */
    public JExpression(TokenReference where) {
	super(where);
    }

    public /*@ non_null @*/ Object clone() {
	try {
	    return super.clone();
	} catch ( CloneNotSupportedException e ) {
	    // Overridden method declares CloneNotSupportedException
	    // so not all subclasses are required to support clone,
	    // however, it won't be thrown.
	    throw new IllegalStateException("unreachable code reached!");
	}
    }
    
    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the type of this expression (call after parsing only)
     */
    public abstract /*@ non_null @*/  CType getType();

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return true;
    }

    /**
     * Returns true iff the variable, field or method represente by this expression
     * was declared non_null.
     */
    public /*@pure*/ boolean isDeclaredNonNull() {
	    return false;
//FIXME          throw new RuntimeException("Unexpected call by an instance of "+this.getClass()+".");
    }

    /**
     * Returns a list of expressions known to be non-null if this evaluates
     * to true.
     * Should only be called on boolean expressions.
     * Returns a list of expressions known to be non-null (null) in this context
     */
    public /*@pure non_null*/ Object[] getFANonNulls(/*@non_null@*/ CContextType context) {
	    return EMPTY;
    }

    /**
     * Print the message followed by the array's elements.
     * Useful for debugging, e.g., when the array is getFANonNulls().
     */
    public void dumpArray(/*@non_null*/ String msg, /*@non_null*/ Object[] exprs) {
	    System.out.print(msg+"[");
	    for (int i=0; i<exprs.length; i++) {
		if (i>0) 
		   System.out.print(", ");
		System.out.print(exprs[i]);
	    }
	    System.out.println("]");
    }

    /**
     * Returns a list of expressions known to be null if this evaluates
     * to false.
     * Should only be called on boolean expressions.
     */
    public /*@pure non_null*/ Object[] getFANulls(/*@non_null@*/ CContextType context) {
	    return EMPTY;
    }

    /**
     * Returns the type of this expression (called after parsing).
     * This method typically returns the same type as returned by
     * <code>getType()</code>; but for nodes such as
     * {@link JNewAnonymousClassExpression}, it returns the type
     * written by the programmer, e.g., <code>T</code> in 
     * <code>new T(...) {...}</code>.
     *
     * @return the apparent type of this expression 
     * @see #getType()
     */
    public /*@pure non_null@*/ CType getApparentType() {
	return getType();
    }

    /**
     * @return true if this expression is constant
     */
    public /*@ pure @*/ boolean isConstant() {
	return false;
    }

    /**
     * Returns true iff this expression can be used as a statement (JLS 14.7)
     */
    public /*@ pure @*/ boolean isStatementExpression() {
	return false;
    }

    /**
     * Indicates whether this expression can appear on the left-hand
     * side of an assignment or in an increment or decrement
     * operation.  Default value is false.  This method should be
     * overridden in subclasses that implement CInitializable.  This
     * allows us to eliminate <code>instanceof</code> checks in
     * several places during typechecking.  
     *
     * <pre><jml>
     * ensures \result ==> (this instanceof CInitializable);
     * </jml></pre>
     *
     * The converse is not true.
     *
     * @see JParenthesedExpression
     */
    public /*@ pure @*/ boolean isMaybeInitializable() {
	return false;
    }

    // ----------------------------------------------------------------------
    // LITERAL HANDLING
    // ----------------------------------------------------------------------

    /**
     * Returns the literal value of this expression
     */
    public /*@ pure @*/ JLiteral getLiteral() {
	throw new RuntimeException("THIS IS NOT A LITERAL");
    }

    /**
     * Returns a typed literal for this expression
     *
     * The expression must be a constant
     */
    public /*@ pure @*/ final JOrdinalLiteral getOrdinalLiteral() {
	return (JOrdinalLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression
     *
     * The expression must be a constant
     */
    public /*@ pure @*/ final JRealLiteral getRealLiteral() {
	return (JRealLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression
     *
     * The expression must be a constant
     */
    public /*@ pure @*/ final JBooleanLiteral getBooleanLiteral() {
	return (JBooleanLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression
     *
     * The expression must be a constant
     */
    public /*@ pure @*/ final JNumberLiteral getNumberLiteral() {
	return (JNumberLiteral)getLiteral();
    }

    /**
     * Returns a typed literal for this expression
     *
     * The expression must be a constant
     */
    public /*@ pure @*/ final JStringLiteral getStringLiteral() {
	return (JStringLiteral)getLiteral();
    }

    /**
     * Does this expression represent a literal?
     */
    public /*@ pure @*/ boolean isLiteral() {
	return false;
    }


    /**
     * Does this expression represent an ordinal literal?
     */
    public /*@ pure @*/ boolean isOrdinalLiteral() {
	return false;
    }

    /**
     * Does this expression represent a real literal?
     */
    public /*@ pure @*/ boolean isRealLiteral() {
	return false;
    }

    /**
     * Does this expression represent a boolean literal?
     */
    public /*@ pure @*/ boolean isBooleanLiteral() {
	return false;
    }

    /**
     * Does this expression represent a string literal?
     */
    public /*@ pure @*/ boolean isStringLiteral() {
	return false;
    }
    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.  Returns a Java
     * expression that has been simplified for direct conversion to
     * bytecode.  The simplifications include evaluating compile-time
     * constant expressions, performing necessary conversions (JLS 5),
     * and desugaring compound-assignment.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression
     * @exception PositionedError if the check fails */
    public abstract /*@ non_null @*/ JExpression typecheck( /*@non_null@*/ CExpressionContextType context ) 
	throws PositionedError;

    /** 
     * <p>Typechecks the expression and mutates the context to record
     * information gathered during typechecking.  This overloaded
     * method is used when the expression is an increment/decrement or
     * appears as one entire side of an assignment.  While at first
     * glance it appears that the last two boolean arguments would
     * vary together (either both true or both false) in the case of
     * compound assignments, like <code>x += 3</code>, the expression
     * for <code>x</code> is on the left side but its value is not
     * discarded.</p>
     *
     * <P>The context tracks the boolean values of these flags so
     * subexpressions have access to this information.  These flags in
     * the current context are reset to their old values at the
     * conclusion of this method.</P>
     *
     * @param	context		the context in which this expression appears
     * @param	isIncDec	indicates whether this expression appears in 
     *				an increment or decrement expression
     * @param	isLeftSide	indicates whether this expression appears on 
     *				the left side of an assignment
     * @param	discardValue	indicates whether the result of an 
     *				expression should be discarded
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public final /*@ non_null @*/ JExpression typecheck( /*@non_null@*/ CExpressionContextType context,
					boolean isIncDec,
					boolean isLeftSide,
					boolean discardValue )
	throws PositionedError
    {
	boolean	isIncDecSaved = context.isIncDec();
	boolean	isLeftSideSaved = context.isLeftSide();
	boolean	discardValueSaved = context.discardValue();

	try {
	    context.setIncDec( isIncDec );
	    context.setLeftSide( isLeftSide );
	    context.setDiscardValue( discardValue );
	    return typecheck( context );
	} finally {
	    context.setIncDec( isIncDecSaved );
	    context.setLeftSide( isLeftSideSaved );
	    context.setDiscardValue( discardValueSaved );
	}
    }


    /**
     * Adds a compiler error.
     * @param	context		the context in which the error occurred
     * @param	key		the message ident to be displayed
     * @param	params		the array of parameters
     *
     */
    protected void fail(/*@ non_null @*/ CContextType context, 
			/*@ non_null @*/ MessageDescription key, 
			/*@ nullable @*/ Object[] params) throws PositionedError
    {
	throw new CLineError(getTokenReference(), key, params);
    }

    /**
     * Changes the type of this expression to the given type.  The
     * caller is responsible for using <code>isAssignableTo</code> to
     * verify that the conversion is valid.
     *
     * @param	dest		the destination type
     * @param	context		the context in which this expression appears
     * @exception PositionedError 	if a type promotion is required and
     * 					the promotion fails to typecheck 
     *					!FIXME! it seems like this
     * 					will never happen */
    public JExpression convertType(/*@non_null@*/ CType dest, /*@non_null@*/ CExpressionContextType context) 
	throws PositionedError 
    {
	if( getType().equals(dest) ) {
	    return this;
	} else if( isConstant() && getType().isNumeric() ) {
	    // not null
	    return JLiteral.createLiteral( dest, 
					   this.getLiteral().getValue() );
	} else {
	    return new JUnaryPromote( this, dest ).typecheck( context );
	}
    }

    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2].  This method is overridden for literals to
     * implement primitive narrowing conversion.
     * @param	dest		the destination type
     * @return true iff assignment is valid */
    public boolean isAssignableTo( /*@non_null@*/ CType dest ) {
	return getType().isAlwaysAssignableTo(dest);
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public abstract void accept(/*@non_null@*/ MjcVisitor p);

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public abstract void genCode(/*@ non_null */ CodeSequence code);

    /**
     * Generates a sequence of bytecodes to branch on a label
     * This method helps to handle heavy optimizables conditions
     * @param	code		the code list
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
	genCode(code);
	code.plantLabelRef(cond ? opc_ifne : opc_ifeq, label);
    }

    /**
     * Returns this with outer parentheses removed.
     */
    public /*@ pure non_null @*/ JExpression unParenthesize() {
	return this;
    }

    // ----------------------------------------------------------------------
    // UNIVERSE RUNTIME CHECKS
    // ----------------------------------------------------------------------

    /**
     * Creates the expressions for the dynamic universe typechecks.
     * 
     * This method is declared here to make the code simpler. Instead
     * of using a lot of if's etc. dynamic binding can be used.
     * 
     * @param context	context used to typecheck the expressions created
     * @param var	the variable where the new object is found
     * @throws PositionedError
     * 
     * By scdaniel.
     */
    public void buildUniverseDynChecks( CExpressionContextType context,
					JExpression var )
	throws PositionedError
    {
    	// nothing (it's overriden in JNewArrayExpression
	// and JNewObjectExpression)
    }
    
    /**
     * Generate the code for the dynamic universe typechecks.
     * 
     * This method is declared here to make the code simpler. Instead
     * of using a lot of if's etc. dynamic binding can be used.
     * 
     * @param code	the CodeSequence for the codegeneration
     * 
     * By scdaniel.
     */
    public void genUniverseDynCheckCode( CodeSequence code ) {
    	// nothing (it's overriden in JNewArrayExpression
	// and JNewObjectExpression)
    }

    // ----------------------------------------------------------------------
    // PUBLIC CONSTANTS
    // ----------------------------------------------------------------------
    
    public static final /*@non_null@*/ JExpression[]		EMPTY = new JExpression[0];
}
