/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JExpressionListStatement.java,v 1.7 2006/08/08 17:18:43 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents an expression list, a comma-separated list of expression statements used in the initializer and iterator of a for-loop statement.
 * JLS2 14.8: Expression Statement, JLS 14.13 For statement.
 */
public class JExpressionListStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param where	the line of this node in the source code
     * @param exprs	the expressions in this list
     * @param comments	the comments that go with this statement
     */
    public JExpressionListStatement( TokenReference where, 
				     JExpression[] exprs, 
				     JavaStyleComment[] comments ) {
	super( where, comments );
	this.exprs = exprs;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Returns an array of expression
     */
    public JExpression[] getExpressions() {
	return exprs;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) throws PositionedError {
	for( int i = 0; i < exprs.length; i++ ) {
	    //!!! ADD PARAMETERS !!! graf 991208 XXX
	    CExpressionContextType expContext = context.createExpressionContext();

	    expContext.setDiscardValue( true );
	    exprs[i] = exprs[i].typecheck( expContext );
	    context.adoptNullityInfo(expContext);
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitExpressionListStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	for (int i = 0; i < exprs.length; i++) {
	    code.setDiscardValue(true);
	    exprs[i].genCode(code);
	    code.setDiscardValue(false);
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression[] exprs() { return exprs; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression[]		exprs;
}
