/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JExpressionStatement.java,v 1.7 2006/08/08 17:18:43 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * Certain kinds of expressions may be used as statements by following
 * them with semicolon.  An expression statement is executed by
 * evaluating the expression; if the expression has a value, the value
 * is discarded.  JLS 14.7: Expression Statement
 * */
public class JExpressionStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	expr		the expression to evaluate.
     * @param	comments	the comments associated with this expression
     */
    public JExpressionStatement( TokenReference where, JExpression expr, 
				 JavaStyleComment[] comments ) {
	super( where, comments );
	this.expr = expr;
    }

    public JExpression getExpression() {
	return expr;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) throws PositionedError {
	CExpressionContextType expContext = context.createExpressionContext();

	expContext.setDiscardValue( true );
	expr = expr.typecheck( expContext );
	check( context, expr.isStatementExpression(), 
	       MjcMessages.INVALID_EXPRESSION_STATEMENT );
	context.adoptNullityInfo(expContext);
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitExpressionStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	boolean	discardValue = code.discardValue();

	code.setDiscardValue(true);
	expr.genCode(code);
	code.setDiscardValue(discardValue);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression expr() { return expr; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		expr;
}

