/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JFieldDeclaration.java,v 1.22 2008/03/16 15:52:52 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents a field declaration in the syntax tree.
 */
public class JFieldDeclaration extends JMemberDeclaration 
    implements JFieldDeclarationType 
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	variable	the variable definition
     * @param	javadoc		is this field deprecated
     * @param	comments	non-javadoc comments for this field
     */
    public JFieldDeclaration(/*@non_null@*/ TokenReference where,
    		/*@non_null@*/ JVariableDefinition variable,
    		/*@nullable@*/ JavadocComment javadoc,
    		/*@nullable@*/ JavaStyleComment[] comments) {
	super(where, javadoc, comments);

	this.variable = variable;
    }

    // ----------------------------------------------------------------------
    // ACCESSOR
    // ----------------------------------------------------------------------

    /**
     * Returns true if this field declarator has initializer (should be initialized)
     */
    public /*@ pure @*/ boolean hasInitializer() {
	return variable.hasInitializer();
    }

    /**
     * @return	the value of this field at initialization
     */
    public /*@ pure @*/ JExpression getInitializer() {
	return variable.getValue();
    }

    /**
     * Sets the initialization expression for this field.
     */
    public void setInitializer(JExpression expr) {
	variable.setValue(expr);
    }

    /**
     * Returns the type of this field
     */
    public /*@ pure @*/ CType getType() {
	return variable.getType();
    }

    public /*@pure*/ boolean isDeclaredNonNull() {
	return variable.isDeclaredNonNull();
    }

    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	return variable.isNonNull(context);
    }

    /**
     * Make this declarator (implicitly) non-null.
     */
    public void setNonNull() {
	variable.setNonNull();
	// Also adjust the signature of this method.
	if(getField() != null) {
	    getField().addModifiers(IMPLICITLY_NON_NULL);
	}
    }

    /**
     * Returns true if this field need to be initialized WARNING: this
     * method returns true when initial value corresponds to a default
     * value ====> a second check should be made after typecheck
     * to ensure that an initialization is really needed */
    public /*@ pure @*/ boolean needInitialization() {
//  	if( hasInitializer() ) {
//  	    return (!variable.getValue().isConstant() 
//  		    || getType().isReference() 
//  		    || isDefaultValue());
//  	}

	return hasInitializer();
    }

    public /*@ pure @*/ JVariableDefinition variable() {
	return variable;
    }

    /** Returns the modifiers of this field declaration */
    public /*@ pure @*/ long modifiers() { 
	return variable.modifiers();
    }

   /** Returns the identifier of this field declaration */
    public /*@ pure @*/ String ident() {
	return variable.ident();
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things
     * generally look OK.  This pass gathers information about
     * the type signatures of everything (imported class
     * files, classes being compiled, methods, fields, etc...)
     * needed for the later passes.  This information is stored 
     * in a context hierarchy that is bound to the AST.
     *
     * @exception	PositionedError	an error with reference to the source file
     */
    public /*@non_null@*/ CSourceField checkInterface(/*@non_null@*/ CClassContextType context) 
	throws PositionedError 
    {
	variable.checkInterface(context);
	CSourceField field = makeFieldSignature( context.getOwnerClass() );
	setInterface( field );

	MemberAccess access = field.access();
	access.checkFieldModifiers(context, this);
	long newModifiers = access.checkInterfaceFieldModifiers(context, this); 
	variable.setModifiers(newModifiers);

	return field;
    }

    /**
     * Generates the signature object for this class declaration. 
     *
     * @param owner	the class signature singleton for the logical outer 
     *			class of this, or null if this is a top level 
     *			declaration
     */
    protected CSourceField makeFieldSignature( /*@non_null@*/ CClass owner )
    {
	long modifiers = variable.modifiers(); 

	MemberAccess access = new MemberAccess( owner, modifiers );

	return ( new CSourceField( access,
				   variable.ident(),
				   variable.getType(),
				   isDeprecated() ));
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks whether this field declaration includes a field
     * initializer and mutates the context to store this information
     * about the field.  Records the value of a compile-time constant
     * initializer.
     *
     * @param	context	the context in which this field is declared
     * @exception PositionedError if any checks fail */
    public void typecheck( /*@non_null@*/ CFlowControlContextType context ) 
	throws PositionedError 
    {  	
	variable.typecheck(context, getField().wasTypeAlreadyAsked());

	// type checking the variable might have propagated additional 
	// Universe modifiers. Make sure the field relects this change.
	if(	CTopLevel.getCompiler().universeChecks() &&
		!getField().wasTypeAlreadyAsked() ) {
		getField().setType(variable.getType());
	}

	if( hasInitializer() && getField().isFinal() ) {
	    JExpression	value = variable.getValue();
	    if (variable.isStatic() && variable.getType().isTypeVariable()){
		context.reportTrouble(new PositionedError(getTokenReference(),
							  MjcMessages.TYPE_VARIABLE_STATIC_FIELD,
							  getField().ident()));
	    
				      
	    }
	    if( value.isConstant() ) {
		getField().setValue(value.getLiteral());

		if( !getField().isStatic() ) {
		    context.reportTrouble(
		       new CWarning( getTokenReference(),
				     MjcMessages.FINAL_FIELD_IMPLICITLY_STATIC,
				     getField().ident() ) );
		}
	    }
	}

	// WMD: report a warning for all non-readonly static fields
	if( CTopLevel.getCompiler().universeChecks() && 
	    getField().isStatic() &&
	    getField().getType() instanceof CClassType ) {

	    CClassType ft = (CClassType) getField().getType();
    	CClassType roft = (CClassType) CUniverseServices.combine(
    			CUniverseReadonly.getUniverse(), ft);
	    
    	if( !ft.equals(roft) &&
    		!CUniverseServices.containsRepModifier(ft) ) {
    		// if the field type contains a rep modifier, we already issue an error
    		// message in JVariableDefinition
    		context.reportTrouble(
   				new CWarning( getTokenReference(),
   					CUniverseMessages.DECLARATION_OF_STATIC_FIELD,
   					getField().ident() ) );
	    }
	}

	if( hasInitializer() ) {
	    context.initializeField( (CSourceField) getField() );
	}

	if( !(getField().isPublic() || getField().isPrivate()) ) {
	    context.reportTrouble(
			new CWarning( getTokenReference(),
				      MjcMessages.PACKAGE_PROTECTED_ATTRIBUTE,
				      getField().ident() ) );
	}

	// scdaniel: universe runtime checks
	if ( CTopLevel.getCompiler().universeDynChecks() ) {
	    JExpression newObject = variable.getValue();

	    /* if (newObject == null) the variable has no initializer */
	    if ( newObject != null ) {
		if ( newObject instanceof JUnaryPromote )
		    newObject = ((JUnaryPromote) variable.getValue()).expr();
	
		/* create transformation */
		newObject.buildUniverseDynChecks(
				  context.createExpressionContext(),
				  new JNameExpression(TokenReference.NO_REF,
						      variable.ident()));
	    }
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(/*@non_null@*/ MjcVisitor p) {
	p.visitFieldDeclaration( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	if (variable.getValue() != null) {
	    code.setLineNumber(getTokenReference().line());

	    if (!getField().isStatic()) {
		code.plantLoadThis();
	    }
	    variable.getValue().genCode(code);
	    getField().genStore(code);

	    /*
	     * scdaniel: universe runtime checks
	     * 
	     * Register newly created objects to the UrtHashtable after
	     * they have been stored into their variable slot.
	     */
	    if ( CTopLevel.getCompiler().universeDynChecks() ) {
		JExpression newObject = variable.getValue();

		/* if (newObject == null) the variable has no initializer */
		if ( newObject != null ) {
		    if ( newObject instanceof JUnaryPromote )
			newObject = ((JUnaryPromote) variable.getValue()).expr();
				
		    /* generate code */
		    newObject.genUniverseDynCheckCode(code);
		}
	    }
	}
    }
    
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected JVariableDefinition		variable;
}
