/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JForStatement.java,v 1.8 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents a for statement in the AST.  JLS2 14.13 */
public class JForStatement extends JLoopStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	init		the init part
     * @param	cond		the cond part
     * @param	incr		the increment part
     * @param	body		the loop body
     * @param	comments	the comments associated with this statement
     */
    public JForStatement( TokenReference where,
			  JStatement init,
			  JExpression cond,
			  JStatement incr,
			  JStatement body,
			  JavaStyleComment[] comments )
    {
	super( where, comments );

	this.init = init;
	this.cond = cond;
	this.incr = incr;
	if( body instanceof JBlock ) {
	    this.body = 
		new JCompoundStatement( where, ((JBlock) body).body() );
	} else {
	    this.body = body;
	}
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	try {
	    CFlowControlContextType statementContext = 
		context.
		createFlowControlContext(init instanceof 
					 JVariableDeclarationStatement ? 
					 ((JVariableDeclarationStatement) init)
					 .getVars().length : 0,
					 getTokenReference());

	    if( init != null ) {
		init.typecheck( statementContext );
	    }

	    CExpressionContextType condContext = statementContext.createExpressionContext();
	    if( cond != null ) {
		cond = cond.typecheck(condContext);
		check( statementContext, 
		       cond.getType() == CStdType.Boolean, 
		       MjcMessages.FOR_COND_NOTBOOLEAN );
	    }

	    CFlowControlContextType neverContext = null;
	    if( cond != null 
		&& (!cond.isConstant() 
		    || !((JBooleanLiteral)cond.getLiteral()).booleanValue())) {
		// if there is a condition and it is not the constant
		// value true, then one control flow path skips the
		// body and increment
		neverContext = statementContext.cloneContext();
	    }

	    CLoopContext loopContext = 
		statementContext.createLoopContext( this );

	    // JLS2 14.20: The contained statement is reachable iff
	    // the for statement is reachable and the condition
	    // expression is not a constant expression with value
	    // falsex.
	    loopContext.setReachable(cond==null || !cond.isConstant() ||
				     ((JBooleanLiteral)cond.getLiteral()).
				     booleanValue());
	    
	    if( init instanceof JVariableDeclarationStatement ) {
		((JVariableDeclarationStatement) init).setIsInFor();
	    }

	    Object[] nonNulls;
	    Object[] nulls;
	    if (cond==null) {
	       nonNulls = JPhylum.EMPTY;
	       nulls = JPhylum.EMPTY;
	    } else {
	       nonNulls = cond.getFANonNulls(condContext);
	       nulls    = cond.getFANulls(condContext);
	    }
	    loopContext.addFANonNulls(nonNulls);

	    CFlowControlContextType secondLoopContext = loopContext.cloneContext();
	    body.typecheck( loopContext );
	    if( init instanceof JVariableDeclarationStatement ) {
		((JVariableDeclarationStatement) init).unsetIsInFor();
	    }

	    if (continueContext != null) {
		if (loopContext.isReachable()) {
		    loopContext.merge(continueContext);
		} else {
		    loopContext.adopt(continueContext);
		}
	    }
	    
	    if( incr != null ) {
		incr.typecheck( loopContext );
	    }

	    // check again to see if problems could be caused in the cond
	    // (or body) because of changes in the body (or incr).  
	    // these should be calls to just the nullity part of typechecking,
	    // once that is factored out.
	    if (cond==null) {
	       nonNulls = JPhylum.EMPTY;
	    } else {
	       cond.typecheck( loopContext.createExpressionContext() );
	       nonNulls = cond.getFANonNulls(condContext);
	    }
	    secondLoopContext.adoptNullityInfo(loopContext);
	    secondLoopContext.addFANonNulls(nonNulls);
// the following call to body.typecheck()
// causes a cascade of java.lang.NullPointerException
// in org.multijava.mjc.CClass,
// starting at genMethods(CClass.java:2458)
// and genGenericFunctions(CClass.java:2524)
//	    if (body != null) body.typecheck( secondLoopContext );
	    //end extra nullity checks

	    loopContext.doneWithCheckingBody();

	    if (neverContext == null) {
		// Condition does not allow loop to terminate even
		// though all the code in the loop body and incr may
		// complete normally.  The only way out is via a
		// break, return, or throw.
		loopContext.setReachable(false);
	    }
	    
	    if( hasBreak() ) {
		if (loopContext.isReachable()) {
		    loopContext.merge( getBreakContextSummary() );
		} else {
		    loopContext.adopt( getBreakContextSummary() );
		}
	    }
      
	    loopContext.checkingComplete();

	    if (neverContext != null) {
		if (statementContext.isReachable()) {
		    statementContext.merge( neverContext );
		} else {
		    statementContext.adopt( neverContext );
		}
	    }
	    if (!hasBreak()) {
		context.addFANonNulls(nulls);
	    }	
	    
	    statementContext.checkingComplete();

	} catch( CBlockError e ) {
	    context.reportTrouble( e );
	}

    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitForStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	CodeLabel		startLabel = new CodeLabel();
	CodeLabel		condLabel = new CodeLabel();

	code.pushContext(this);
	if (init != null) {
	    init.genCode(code);			//		INIT
	}
	code.plantLabelRef(opc_goto, condLabel);	//		GOTO cond
	code.plantLabel(startLabel);		//	start:
	if (body != null) {
	    body.genCode(code);			//		BODY
	}
	code.plantLabel(getContinueLabel());		//	incr:
	if (incr != null) {
	    incr.genCode(code);			//		INCR
	}
	code.plantLabel(condLabel);			//	cond:
	if (cond != null) {
	    cond.genBranch(true, code, startLabel);	//		COND IFNE start
	} else {
	    code.plantLabelRef(opc_goto, startLabel);	//		GOTO start
	}
	code.plantLabel(getBreakLabel());			//	end:

	code.popContext(this);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JStatement init() { return init; }
    public JExpression cond() { return cond; }
    public JStatement incr() { return incr; }
    public JStatement body() { return body; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JStatement		init;
    private JExpression		cond;
    private JStatement		incr;
    private JStatement		body;
}

