/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JFormalParameter.java,v 1.23 2008/03/16 15:52:52 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a parameter declaration in the syntax tree
 */
public class JFormalParameter extends JLocalVariable {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a formal parameter node in the parsing tree.  
     * @param where		the line of this node in the source code
     * @param modifiers		the modifiers on this parameter
     * @param desc		the sort of variable that this is, see
     *				constants in JLocalVariable
     * @param type		the static and dynamic types of this formal
     * @param ident		the name of this variable
     */
    public JFormalParameter( TokenReference where,
			     long modifiers,
			     int desc,
			     CSpecializedType type,
			     String ident ) {
    	super( where, modifiers, desc, type.dynamicType(), ident, null );
    	
    	// WMD Note: the dynamic type is adapted by the superclass.
    	if( type.staticType() == type.dynamicType() )
    		this.overriddenType = super.type;
    	else
    		this.overriddenType = CUniverseServices.setDefaultVariable(type.staticType());
    	
    	this.metaDebugKey = "JFormalParameter " + toString();    	
    }

    /**
     * Constructs a formal parameter node in the parsing tree.  
     * @param where		the line of this node in the source code
     * @param desc		the sort of variable that this is, see
     *				constants in JLocalVariable
     * @param type		the static and dynamic types of this formal
     * @param ident		the name of this variable
     * @param isFinal		a flag indicating whether this is declared 
     *				final
     */
    public JFormalParameter( TokenReference where,
			     int desc,
			     CSpecializedType type,
			     String ident,
			     boolean isFinal ) {
	this( where, isFinal ? ACC_FINAL : 0L, desc, type, ident );
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * @return the type of this variable
     */
    //@ pure
    public CType getType() {
	    // a value specializer is not a real type
	if (type.isValueType())
	    return overriddenType;
	else
	    return type;
    }
    
    /**
     * Sets the type of this parameter
	 *
	 * @param type The type to set it to
     */
    public void setType(CType type) {
    	this.type = type;
    	this.overriddenType = type;
    }

    /**
     * Returns the static type part of this formal parameter type.
     *
     * @return	the static type part of this formal parameter type
     */
    public /*@ pure @*/ CType staticType() {
	return overriddenType;
    }

    /**
     * Returns the dynamic type part of this formal parameter type.
     *
     * @return  the dynamic type part of this formal parameter type
     */
    public /*@ pure @*/ CType dynamicType() {
        return type;
    }
    
    /**
     * Returns the overridden and dynamic types for this formal.
     *
     * @return the overridden and dynamic types for this formal */
    public CSpecializedType specializedType() {
	assertTrue( interfaceChecked );
	return specializedType;
    }

    public /*@ pure @*/ boolean isSpecialized() {
    	return overriddenType!=null && overriddenType != type;
    }

    public String toString() {
	StringBuffer buf = new StringBuffer( );
	buf.append( modToString() );
	buf.append( typeToString() );
	buf.append( paramToString() );
	return buf.toString();
    }
    
    public String modToString() {
	return ( isFinal() ? "final " : "" );
    }

    public String typeToString() {
	StringBuffer buf = new StringBuffer();
	if( isSpecialized() ) {	// specialized formal
	    buf.append( overriddenType.toString() +
			type.specializerSymbol() );
	}
	buf.append(type.toString());
	return buf.toString();
    }

    public String paramToString() {
	if (ident().indexOf("$") == -1) { // FIXME - is this right - print the type but not the ident if it is a generated  parameter ???
	    return ( " " + ident() );
	}
	return "";
    }

    /** Returns any parameter modifiers as a String; if the String is
     * not empty, it will have a trailing space. 
     */
    public /*@ non_null @*/ String modifiersAsString() {
	return "";
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things
     * generally look OK.  This pass gathers information about
     * the type signatures of everything (imported class
     * files, classes being compiled, methods, fields, etc...)
     * needed for the later passes.  This information is stored 
     * in a context hierarchy that is bound to the AST.
     *
     * @param context	the context in which this appears
     * @return	the overridden and dynamic types for this formal
     * @exception PositionedError if the check fails */
    public CSpecializedType checkInterface( CContextType context ) throws PositionedError
    {
	interfaceChecked = true;
	try {
	    type =  type.checkType( context );
	    overriddenType = overriddenType.checkType(context);
	    
	    // WMD reset Universe defaults
	    if( type == this.overriddenType ) {
			this.type = CUniverseServices.setDefaultVariableAgain(this.type);
			this.overriddenType = this.type;
		} else {
			this.type = CUniverseServices.setDefaultVariableAgain(this.type);
			this.overriddenType = CUniverseServices.setDefaultVariableAgain(this.overriddenType);
		}

	    if( isSpecialized() ) {
		//overriddenType = overriddenType.checkType( context );
		if( type.equals( overriddenType ) ) {
		    context.reportTrouble( 
			   new PositionedError( getTokenReference(),
						MjcMessages.SAME_DYNAMIC_TYPE,
						type ) );
		}
		specializedType = new CSpecializedType( overriddenType, type );
	    } else {
		specializedType = new CSpecializedType( type );
	    }
	} catch( UnpositionedError cue ) {
	    context.reportTrouble( cue.addPosition( getTokenReference() ) );
	    specializedType =  
		new CSpecializedType( CStdType.Object );
	}
	return specializedType;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Compute the value of a specializer expression used to dispatch on
     * a compile-time constant.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    public void resolveSpecializer(CFlowControlContextType context)
	throws PositionedError {
	type.resolveValueType( context.createExpressionContext() );
    }

    /**
     * Typechecks the parameter and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @exception PositionedError if the check fails */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	super.typecheck(context);
	if( !interfaceChecked ) {
	    checkInterface( context );
	}
	try {
	    context.addVariable( this );
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
		
	// For the specialized formal "S@T i", S must be a proper
	// supertype of T and T must be a class type [CLCM 4.1.3.1].
	// RMJ: allow T to be an interface.
	// This check is also extended to handle value dispatches.
	if( isSpecialized() ) {
	    if (type.isValueType()) {
		check( context,
		       isValidSpecializerValueType( type ),
		       MjcMessages.NON_VALUE_SPECIALIZER,
		       type, overriddenType );
		check( context,
		       type.isAlwaysAssignableTo( overriddenType ),
		       MjcMessages.BAD_VALUE_SPECIALIZER, type,
		       overriddenType );
	    }
	    else {
		//RMJ: allow interfaces as specializers
		if (context.getCompiler().RMJ() &&
		    type.isClassType() && type.getCClass().isInterface()) {
		    warn( context,
			  MjcMessages.INTERFACE_SPECIALIZER,
			  type, overriddenType );
		} else {
		    check( context,
			   isValidSpecializerClassType( type ),
			   MjcMessages.NON_CLASS_SPECIALIZER,
			   type, overriddenType );
		}
		check( context,
		       type.isAlwaysAssignableTo( overriddenType ), 
		       MjcMessages.BAD_CLASS_SPECIALIZER, type, overriddenType );
	    }
	    
	}
	       
	context.initializeVariable( this );
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept( /*@non_null@*/ MjcVisitor p ) {
	p.visitFormalParameters( this );
    }
 
    //---------------------------------------------------------------------
    // STATIC METHODS
    //---------------------------------------------------------------------

    /**
     * Creates from an array of formal parameters a String of the
     * dynamic types of the parameters.  This string can be used as a
     * key into a hashtable for verifying unique method signatures.
     *
`     * @param	parms	an array of the formal parameters
     * @return a string that is the concatenation of the dynamic parm types */
    public static String specializerString( JFormalParameter[] parms ) {
	String metaDebugContext = 
	    "static JFormalParameter.specializerString()";
	StringBuffer result = new StringBuffer();
	for( int i=0; i<parms.length; i++ ) {
	    CType ddType = parms[i].getType();
	    result.append( (i==0 ? "" : ",") 
			    + ((ddType==null) ? "null" : ddType.toString()) );
	}
	return result.toString();
    }

    /**
     * Indicates whether the given type can ever appear as an explicit
     * class specializer.  Does not check whether the given type is a
     * proper subtype of the static type.  Only checks the type is a
     * non-interface, non-array reference type or is an array (of any
     * dimension) of non-interface, non-array reference types.  */
    private static boolean isValidSpecializerClassType( CType type ) {
	return type.isClassType() && !type.getCClass().isInterface()
	    && (!type.isArrayType() || 
		isValidSpecializerClassType(
		    ((CArrayType) type).getBaseType() ));
    }

    /**
     * Indicates whether the given type can ever appear as an explicit
     * value specializer.  It can be, as long as the value is a compile-time
     * constant.
     **/
    private static boolean isValidSpecializerValueType( CType type ) {
	    // we use isLiteral() instead of isConstant(), which can
	    // be true of things that are not literals, thereby screwing
	    // things up
	return type.isValueType() && type.getValue().isLiteral();
    }

    // ----------------------------------------------------------------------
    // PUBLIC CONSTANTS
    // ----------------------------------------------------------------------

    public static final JFormalParameter[] EMPTY = new JFormalParameter[0];

    //---------------------------------------------------------------------
    // PRIVATE FIELDS
    //---------------------------------------------------------------------

    private CType overriddenType; // the static type part of this formal
    private boolean interfaceChecked = false;
    private CSpecializedType specializedType;
    /**
     * A string to be used as a key for meta-debugging the Debug
     * operations.  Because the equals method inherited from
     * JLocalVariable is screwy, we need a key that is reasonable to
     * use in HashMaps.  */
    private String metaDebugKey;
}
