/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JIfStatement.java,v 1.10 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents an if statement in the AST.
 * JLS 14.10: If Statement
 */
public class JIfStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param where		the line of this node in the source code
     * @param cond		the expression to evaluate
     * @param thenClause	the statement to execute if the condition 
     *				is true
     * @param elseClause	the statement to execute if the condition 
     *				is false
     * @param comments		the Java comments associated with this
     *				statement
     */
    public JIfStatement( TokenReference where,
			 JExpression cond,
			 JStatement thenClause,
			 JStatement elseClause,
			 JavaStyleComment[] comments ) {
	super( where, comments );

	this.cond = cond;
	this.thenClause = thenClause;
	this.elseClause = elseClause;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
        CExpressionContextType condContext = context.createExpressionContext();
	cond = cond.typecheck( condContext );
	check( context, cond.getType() == CStdType.Boolean, 
	       MjcMessages.IF_COND_NOTBOOLEAN, cond.getType() );

	// create a clone of the context so we can check the
	// else-clause in a context that doesn't include the results
	// of checking the then-clause
	CFlowControlContextType elseContext = context.cloneContext();
	Object[] nonNulls=cond.getFANonNulls(context);
  	Object[] nulls   =cond.getFANulls(context);

//prj 02 aug 2006 -- work in progress to allow reversing the 2 lines below. 
	context.addFANonNulls(nonNulls);
//	context.adoptFANonNull(condContext);

	thenClause.typecheck( context );

        elseContext.addFANonNulls(nulls);
	if( elseClause != null ) {
	    elseClause.typecheck( elseContext );

	    if( elseContext.isReachable() && context.isReachable() ) {
		context.merge( elseContext );
	    } else if( elseContext.isReachable() ) {
		context.adopt(elseContext);
	    }
	} else {
	    if (!(cond.isConstant() && cond.getBooleanLiteral().booleanValue())) {
		// Since there is no else branch, and since the
		// condition is not a constant value of 'true' it is
		// possible that the thenClause will be skipped at
		// runtime.  The never-used elseContext represents the
		// path where the thenClause is skipped.
		final boolean contextWasReachable=context.isReachable();
		context.merge( elseContext );
		// if context was not reachable, we left from the thenClause
		// and we can add the cond's nulls as non-nulls to context
		if (!contextWasReachable) {
		   context.addFANonNulls(nulls);
		}
	    } else {
		// Java has a special case for constant valued
		// conditions to enable the use of final fields as
		// conditional compilation flags.  See JLS2 14.20.
		context.setReachable(true);
	    }

	}

	if( cond instanceof JAssignmentExpression ) {
	    context.reportTrouble( 
			  new CWarning( getTokenReference(),
					MjcMessages.ASSIGNMENT_IN_CONDITION ));
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitIfStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	if (cond.isConstant()) {
	    if (cond.getBooleanLiteral().booleanValue()) {
		thenClause.genCode(code);
	    } else if (elseClause != null) {
		elseClause.genCode(code);
	    }
	} else {
	    CodeLabel		elseLabel = new CodeLabel();
	    CodeLabel		nextLabel = new CodeLabel();

	    cond.genBranch(false, code, elseLabel);   //		COND IFEQ else
	    thenClause.genCode(code);			//		THEN CODE
	    code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
	    code.plantLabel(elseLabel);		//	else:
	    if (elseClause != null) {
		elseClause.genCode(code);		//		ELSE CODE
	    }
	    code.plantLabel(nextLabel);		//	next	...
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression cond() { return cond; }
    public JStatement thenClause() { return thenClause; }
    public JStatement elseClause() { return elseClause; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		cond;
    private JStatement		thenClause;
    private JStatement		elseClause;
}
