/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JInitializerDeclaration.java,v 1.8 2006/12/11 05:55:57 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents an initializer (either static or instance) in
 * the AST.  Each class declaration has at most two of these nodes in
 * its AST, one representing a static initializer and one an instance
 * initializer.  Separate AST nodes are maintained for the user
 * declared initializer blocks under the class AST node.  These
 * separate nodes are used for pretty-printing and other tasks.  The
 * JInitializerDeclaration have aliases to the blocks for code
 * generation and typechecking purposes.  */
public class JInitializerDeclaration extends JMethodDeclaration {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     */
    public JInitializerDeclaration(TokenReference where,
				   JBlock body,
				   boolean isStatic,
				   boolean isDummy) {
	super( where, (isStatic ? ACC_STATIC : ACC_PRIVATE),CTypeVariable.EMPTY,
	       CStdType.Void, (isStatic ? JAV_STATIC_INIT : JAV_INIT),
	       JFormalParameter.EMPTY, CClassType.EMPTY, body, null, null );

	this.isDummy = isDummy;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Return true if this initialiser declaration is used only to check code
     * and that it should not generate code
     */
    public boolean isDummy() {
	return isDummy;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Creates a context for this initializer AST node.
     *
     * @param	parent		the parent context
     */
    public CMethodContextType createSelfContext(/*@non_null@*/ CClassContextType parent) {
	return parent.createInitializerContext(getMethod());
    }

    protected void checkAccess( CClassContextType context, 
				CMethodSet superMethods ) {
	// nothing to do here
    }

    /**
     * Checks the static initializers and informs the context that it
     * has a valid initializer.
     *
     * @param	context		the context in which this initializer appears
     * @exception PositionedError if the checks fail
     */
    public void checkInitializer(CClassContextType context) 
	throws PositionedError 
    {
	if (!getMethod().isStatic() && !isDummy()) {
	    context.addInitializer();
	}
	// This initializer is type checked as a method in this pass.
	// This should mutate the context so that the information from
	// the initializers is available when checking the regular
	// methods.

	super.typecheck(context);
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitInitializerDeclaration( this );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private boolean	isDummy;
}
