/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JInstanceofExpression.java,v 1.11 2008/03/16 15:52:52 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.ClassRefInstruction;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents an <code>instanceof</code> expression.
 */
public class JInstanceofExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     *
     * @param	where	the line of this node in the source code
     * @param	expr	the expression to be casted
     * @param	dest	the type to test for
     */
    public JInstanceofExpression( TokenReference where, 
				  JExpression expr, CType dest) {
	super( where );
	this.expr = expr;
	this.dest = CUniverseServices.setDefaultVariable(dest);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression expr() { return expr; }
    public CType dest() { return dest; }

    /**
     * Compute the type of this expression.
     *
     * @return the type of this expression
     */
    public CType getType() {
	return CStdType.Boolean;
    }

    /**
     * Returns a list of expressions known to be non-null (null) in this context
     */
    public /*@pure non_null*/ Object[] getFANonNulls(/*@non_null@*/ CContextType context) {
	    return new Object[] {expr};
    }
    // don't want override getFANulls, since failing instanceof does not 
    // tell anything about the nullity.


    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	expr = expr.typecheck( context );
	try {
	    dest = dest.checkType( context );
	    dest = CUniverseServices.setDefaultVariableAgain(dest);
	    
	    if( dest != null &&
	    	dest instanceof CClassType &&
    		expr.getType() instanceof CClassType &&
    		CTopLevel.getCompiler().universeChecks() &&
	    	CUniverseServices.isPartial( (CClassType)dest ) ) {
	    	String orig_dest = dest.toVerboseString();
	    	dest = CUniverseServices.propagate( (CClassType)dest, (CClassType)expr.getType());

	    	context.reportTrouble(new CWarning(getTokenReference(), 
					 CUniverseMessages.INCOMPLETE_INSTANCEOF,
					 orig_dest,
					 dest.toVerboseString()));
	    }
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
	check( context, expr.getType().isReference(),
	       MjcMessages.INSTANCEOF_BADTYPE, expr.getType().toVerboseString(), dest.toVerboseString() );

	check( context, dest.isCastableTo(expr.getType()),
	       MjcMessages.INSTANCEOF_BADTYPE, expr.getType().toVerboseString(), dest.toVerboseString() );

	if( expr.getType().implicitlyCastTo(dest) ) {
	    context.reportTrouble( 
		           new CWarning( getTokenReference(), 
					 MjcMessages.UNNECESSARY_INSTANCEOF, 
					 null ));
	}

        // instanceof may not be applied to non-reifiable types. (JLS3 15.20.2)
	check(context, 
	      dest.isReifiableType(), 
	      MjcMessages.NON_REIFIABLE_INSTANCEOF,
	      dest.toVerboseString());

	// scdaniel: universe runtime checks
	while ( CTopLevel.getCompiler().universeDynChecks() ) {
	    /* get reference to a temporary variable */
	    tmpLocalVar = CUniverseRuntimeHelper.getTmpVariable(context, this, dest);

	    /* create test */
	    JExpression ownerTest =
		CUniverseRuntimeHelper.getOwnerTest(context,
						    dest,
						    new JLocalVariableExpression(TokenReference.NO_REF, tmpLocalVar),
						    expr);

	    if ( ownerTest == null ) {
    		visited = true;
    		break;
	    }

	    // temporarily disable universe checks ...
	    boolean universeChecks = CTopLevel.getCompiler().universeChecks();
	    boolean universePurity = CTopLevel.getCompiler().universePurity();
	    CTopLevel.getCompiler().setUniverseChecks(false);
	    CTopLevel.getCompiler().setUniversePurity(false);

	    // ... typecheck ...
	    ownerTest.typecheck(context);

	    // ... enable universe checks again.
	    CTopLevel.getCompiler().setUniverseChecks(universeChecks);
	    CTopLevel.getCompiler().setUniversePurity(universePurity);
    	
	    /* 
	     * Combine with this node (do it after typecheck because it would
	     * typecheck this node again (and again ...)).
	     */
	    universeDynCheck =
		new JConditionalAndExpression(TokenReference.NO_REF,
					      this, ownerTest);

	    break;
	}
	
	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitInstanceofExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
    	/*
    	 * scdaniel: universe runtime checks
    	 * 
    	 * Generate code for the checks. The real instanceof code
    	 * is a child of the check and will be generated a little later.
    	 * (that's what visited is used for)
    	 */
        if ( CTopLevel.getCompiler().universeDynChecks() &&
	     !visited ) {
	    /* next time we reach this node generate normal instanceof code */
	    visited = true;

	    /* save result of expression in temporary variable */
	    expr.genCode(code);
	    tmpLocalVar.genStore(code);

	    /* gen code for checks */
	    universeDynCheck.genCode(code);
        } else {
	    code.setLineNumber(getTokenReference().line());

	    /*
	     * scdaniel: universe runtime checks
	     * 
	     * Load value of expression from temporary variable or evaluate
	     * expression if dynChecks are disabled.
	     */
	    if ( CTopLevel.getCompiler().universeDynChecks() &&
		 (universeDynCheck != null) )
		tmpLocalVar.genLoad(code);
	    else
		expr.genCode(code); 
	    
	    code.plantInstruction( new ClassRefInstruction(opc_instanceof,
			 	       ((CClassType)dest).qualifiedName()) );
        }
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression	expr;
    private CType	dest;
    
    // scdaniel: universe runtime checks
    private JConditionalAndExpression universeDynCheck;
    private JLocalVariable tmpLocalVar;
    private boolean visited = false;
}
