/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JInterfaceDeclaration.java,v 1.18 2006/12/20 14:35:21 perryjames Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Iterator;
import org.multijava.javadoc.JavadocComment;

import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a java interface in the syntax tree
 */
public class JInterfaceDeclaration extends JTypeDeclaration 
    implements JInterfaceDeclarationType
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs an interface declaration in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	modifiers	the list of modifiers of this class
     * @param	ident		the short name of this class
     * @param   typevariables   the type variables of this type
     * @param	interfaces	the names of this types's interfaces
     * @param	methods	        a list of JMethodDeclarationTypes giving the
     *				methods of this type
     * @param	inners	        a list of JTypeDeclarationTypes giving the
     *				inner classes (and interfaces) of this type
     * @param	fieldsAndInits	the fields and initializers of this type, 
     *				passed together because the order matters 
     *				for class and object initialization, members 
     *				of the array should be instances of 
     *				<code>JFieldDeclarationType</code> or 
     *				<code>JClassBlock</code>
     * @param	javadoc	        javadoc comments including whether this 
     *				type declaration is deprecated
     * @param	comments        regular java comments
     * @see JMethodDeclarationType
     * @see JFieldDeclarationType
     * @see JClassBlock
     */
    public JInterfaceDeclaration( /*@ non_null */ TokenReference where,
				  long modifiers, 
				  /*@ non_null */  String ident,
				  /*@ non_null */ CTypeVariable[] typevariables,
				  /*@ non_null */ CClassType[] interfaces,
				  /*@ non_null */ ArrayList methods,
				  /*@ non_null */ ArrayList inners,
				  /*@ non_null */ JPhylum[] fieldsAndInits,
				  /*@ nullable */ JavadocComment javadoc,
				  /*@ nullable */ JavaStyleComment[] comments ) {
	super( where, modifiers | ACC_INTERFACE | ACC_ABSTRACT,
	       ident, typevariables, interfaces, methods, inners, fieldsAndInits, 
	       javadoc, comments );
    }
    
    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    public void preprocessDependencies( /*@non_null@*/ CContextType context ) 
	throws PositionedError
    {
	super.preprocessDependencies( context, CStdType.Object );
    }

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type
     * signatures of everything (imported class files, classes being
     * compiled, methods, fields, etc...)  needed for the later
     * passes.  This information is stored in a CCompilationUnit
     * instance and instances of CMember that are bound to the AST.
     * Also adds things like the default constructor and the initializer
     * method to the AST (these are suppressed during pretty-printing).
     *
     * @param		context		the context in which this
     *					declaration appears
     * @exception	PositionedError	an error with reference to 
     *					the source file
     */
    public void checkInterface( /*@non_null@*/ CContextType context ) throws PositionedError {
	statInit = constructStaticInitializers();

	super.checkInterface(context);
    }

    /**
     * Checks the static initializers created during the
     * checkInterface pass and performs some other checks that can be
     * performed simply before full blown typechecking.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if check fails */
    public void checkInitializers( /*@non_null@*/ CContextType context) throws PositionedError {
	CInterfaceContextType self = (CInterfaceContextType) createContext( context );

	if (statInit != null) {
	    statInit.checkInitializer( self );
	    staticInfo = self.fieldInfoTable();
	}

	// Check inners
	Iterator iter = inners.iterator();
	while( iter.hasNext() ) {
	    ((JTypeDeclarationType) iter.next()).checkInitializers( self );
	}

	super.checkInitializers( context );
    }

    /**
     * Computes the values of specializer expressions used to dispatch on
     * compile-time constants.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    public void resolveSpecializers(/*@non_null@*/ CContextType context) throws
	PositionedError {
	CInterfaceContextType self =
	    (CInterfaceContextType) createContext( context );
 	Iterator iter = inners.iterator();
	while( iter.hasNext() ) {
	    ((JTypeDeclarationType) iter.next()).
		resolveSpecializers( self );
	}
    }
    
    /**
     * Typechecks this type declaration in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context the context in which this type declaration appears
     * @exception PositionedError if any checks fail */
    public void typecheck(/*@non_null@*/ CContextType context) throws PositionedError {
	CInterfaceContextType self = (CInterfaceContextType)createContext(context);
	Iterator iter;

	iter = inners.iterator();
	while( iter.hasNext() ) {
	    try {
		((JTypeDeclarationType) iter.next()).typecheck(self);
	    } catch (CBlockError e) {
		context.reportTrouble(e);
	    }
	}

	iter = methods.iterator();
	while( iter.hasNext() ) {
	    try {
		((JMethodDeclarationType) iter.next()).typecheck(self);
	    } catch (PositionedError ce) {
		context.reportTrouble(ce);
	    }
	}

	try {
	    // mutate the context to record the information gathered by
	    // this method, verifies that all necessary declarations exist
	    // (for example, implementations of all abstract super methods
	    // if this is concrete)
	    self.checkingComplete( this, staticInfo, null, 
				   new CVariableInfoTable[0] );
	} catch (UnpositionedError cue) {
	    throw cue.addPosition(getTokenReference());	     
	} // end of catch
	
	super.typecheck(context);
    }

    /**
     * Creates an interface context for this interface declaration.
     * @param	parent		the parent context or null
     * @return	returns a CClassContextType that represents this context
     */
    public /*@non_null@*/ CClassContextType createContext(/*@non_null@*/ CContextType parent) {
	return parent.createInterfaceContext( sourceClass );
    }

    /**
     * check static initializers
     */
    protected /*@ nullable */ JInitializerDeclaration constructStaticInitializers() {
	// collect all static initializers and build a method
	ArrayList	elems = new ArrayList();
	boolean	needGen = false;
	for (int i = 0; i < fieldsAndInits.length; i++) {
	    if (fieldsAndInits[i] instanceof JClassBlock) {
		elems.add(fieldsAndInits[i]);
		needGen = true;
	    } else {
		if (fieldsAndInits[i] instanceof JFieldDeclarationType) {
		    needGen |= ((JFieldDeclarationType)fieldsAndInits[i]).
			needInitialization();
		    elems.add(
			    new JClassFieldDeclarator( getTokenReference(), 
						       (JFieldDeclarationType)
						       fieldsAndInits[i]));
		}
	    }
	}

	JStatement[] stmts = new JStatement[elems.size()];
	stmts = (JStatement[]) elems.toArray(stmts);

	
	if (stmts.length > 0) {
	    return new JInitializerDeclaration(getTokenReference(),
					       new JClassBlock(
						     getTokenReference(), 
						     false, stmts ),
					       true,
					       !needGen);
	} else {
	    return null;
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public /*@pure non_null@*/ String ident() { return sourceClass.ident(); }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(/*@non_null@*/ MjcVisitor p) {
	p.visitInterfaceDeclaration( this );
    }


    // -------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // -------------------------------------------------------------
    private	CVariableInfoTable		staticInfo;
}
