/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLabeledStatement.java,v 1.11 2004/06/14 05:09:49 leavens Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents a labeled statement.
 * JLS 14.6: Labeled Statement
 */
public class JLabeledStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	label		the label of the enclosing labeled statement
     * @param	stmt		the statement
     * @param	comments	comments in the source text
     */
    public JLabeledStatement( TokenReference where, String label, 
			      JStatement stmt, JavaStyleComment[] comments ) {
	super( where, comments );

	this.label = label;
	this.stmt = stmt;
	this.isBreakTarget = false;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the label of this statement.
     */
    public String getLabel() {
	return label;
    }

    public /*@ pure @*/ JStatement stmt() { return stmt; }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	check( context,
	       context.getLabeledStatement( label ) == null,
	       MjcMessages.LABEL_ALREADY_EXISTS, label );

	CLabeledContext	self = context.createLabeledContext( this );

	stmt.typecheck( self );

	if (breakContext != null) {
	    if (self.isReachable()) {
		self.merge( breakContext );
	    } else {
		self.adopt( breakContext );
	    }
	    self.setReachable( true );
	}

	if( !isBreakTarget ) {
	    context.reportTrouble( new CWarning( getTokenReference(), 
						 MjcMessages.UNUSED_LABEL, 
						 label ));
	}

	self.checkingComplete();
    }

    // ----------------------------------------------------------------------
    // BREAK/CONTINUE HANDLING
    // ----------------------------------------------------------------------

    /**
     * Returns true iff this statement is a possible target for a
     * break statement.  */
    public /*@ pure @*/ boolean acceptsBreak() {
	return true;
    }

    /**
     * Returns true iff this statement is a possible target for a
     * continue statement.  */
    public /*@ pure @*/ boolean acceptsContinue() {
	return stmt.acceptsContinue();
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitLabeledStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	stmt.genCode(code);
	code.plantLabel(endLabel);

	endLabel = null;
    }

    /**
     * Registers that this statement is the target of a break
     * statement.  */
    public void addBreak(CFlowControlContextType context) {
	isBreakTarget = true;
	// If the contained statement can handle a break then we let
	// it do so (since it might have special behavior, e.g., a
	// switch statement where a case can follow the break.
	// Otherwise we need to keep track of the break context.
	if (stmt.acceptsBreak()) {
	    stmt.addBreak(context);
	} else {
	    if (breakContext == null) {
		breakContext = context.cloneContext();
	    } else {
		breakContext.merge( context );
	    }
	}
    }

    /**
     * Registers that this statement is the target of a continue
     * statement.
     *
     * <pre><jml>
     * also 
     * requires stmt.acceptsContinue();
     * </jml></pre>
     *
     */
    public void addContinue(CFlowControlContextType context) {
	stmt.addContinue(context);
    }

    /**
     * Return the end of this block (for break statement)
     */
    public CodeLabel getBreakLabel() {
	if (stmt.acceptsBreak()) {
	    return stmt.getBreakLabel();
	} else {
	    return endLabel;
	}
    }

    /**
     * Return the beginning of this block (for continue statement)
     */
    public CodeLabel getContinueLabel() {
	return stmt.getContinueLabel();
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------


    /**
     * The control flow context representing paths in this statement
     * that can jump to the end of the statement via a break.  */
    private CFlowControlContextType breakContext;

    private String		label;
    private /*@ spec_public */ JStatement		stmt;
    private CodeLabel		endLabel = new CodeLabel();
    private boolean		isBreakTarget;
}
