/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLiteral.java,v 1.8 2008/03/16 15:52:52 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.PositionedError;

/**
 * Root class for all literals expression
 */
public abstract class JLiteral extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param where the line of this node in the source code
     */
    public JLiteral(TokenReference where) {
	super(where);
    }

    /**
     * Create a literal from a type and a value
     */
    public static JLiteral createLiteral(CType type, Object value) {
	if (CStdType.Char.equals(type,false)) {
	    return new JCharLiteral(TokenReference.NO_REF, value);
	}

	if (value instanceof Character) {
	    value = new Integer(((Character)value).charValue());
	}

	if (type.isOrdinal()) {
	    switch (type.getTypeID()) {
	    case TID_BYTE:
		return new JOrdinalLiteral( TokenReference.NO_REF, 
					    ((Number) value).byteValue(), 
					    CStdType.Byte );
	    case TID_SHORT:
		return new JOrdinalLiteral( TokenReference.NO_REF, 
					    ((Number) value).shortValue(), 
					    CStdType.Short );
	    case TID_INT:
		return new JOrdinalLiteral( TokenReference.NO_REF, 
					    ((Number) value).intValue(), 
					    CStdType.Integer );
	    case TID_LONG:
		return new JOrdinalLiteral( TokenReference.NO_REF, 
					    ((Number) value).longValue(), 
					    CStdType.Long );
	    }
	}

	if( CStdType.Double.equals(type,false) ) {
	    return new JRealLiteral( TokenReference.NO_REF, 
				     (Number)value, CStdType.Double );
	}

	if( CStdType.Float.equals(type,false) ) {
	    return new JRealLiteral( TokenReference.NO_REF, 
				     (Number)value, CStdType.Float );
	}

	if( CStdType.Boolean.equals(type,false) ) {
	    return new JBooleanLiteral( TokenReference.NO_REF, 
					((Number) value).intValue() >= 1);
	}

	if( CStdType.String.equals(type,false) ) {
	    return new JStringLiteral( TokenReference.NO_REF, 
				       value.toString() );
	}

	fail();
	return null;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * @return is this expression a lvalue ?
     */
    public boolean isLValue(CExpressionContextType context) {
	return false;
    }

    /**
     * @return is this expression constant ?
     */
    public /*@ pure @*/ boolean isConstant() {
	return true;
    }

    /**
     * Returns the literal value of an expression
     */
    public /*@ pure @*/ JLiteral getLiteral() {
	return this;
    }

    /**
     * Returns the literal value of an expression
     */
    public /*@ pure @*/ abstract boolean isDefault();

    /**
     * Does this expression represent a literal?
     */
    public /*@ pure @*/ boolean isLiteral() {
	return true;
    }
    
    /**
     * Returns the value of this literal
     */
    public abstract Object getValue();

    //---------------------------------------------------------------------
    // CODE CHECKING
    //---------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	return this;
    }


}
