/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLocalVariable.java,v 1.15 2008/03/16 15:52:53 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;

/**
 * This class represents a local variable declaration in the AST.
 */
public abstract class JLocalVariable extends JPhylum 
    implements VariableDescriptor 
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a local variable definition AST node.
     * @param where	the location of the declaration of this variable
     * @param modifiers	the modifiers on this variable
     * @param desc	the sort of variable this is (see constants)
     * @param type	the type of this variable
     * @param name	the name of this variable
     * @param expr	the initializing expression for this variable
     */
    public JLocalVariable( TokenReference where,
			   long modifiers, 
			   int desc,
			   CType type,
			   String name,
			   JExpression expr ) {
	super( where );

	this.modifiers = modifiers;
	this.desc = desc;
	
	// WMD: could do this, but then we would not get a warning when we change it.
	// if( desc == JLocalVariable.DES_CATCH_PARAMETER)
	// this.type = CUniverseServices.setDefaultReadonly(type);
	// else
	// if( desc == JLocalVariable.DES_GENERATED)
	//	this.type = type;
	// else
	this.type = CUniverseServices.setDefaultVariable(type);
	
	
	this.name = name;
	this.expr = expr;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns if this variable is final
     */
    public /*@ pure @*/ boolean isFinal() {
	return (modifiers() & ACC_FINAL) != 0;
    }

    /**
     * Returns if this variable is static
     */
    public /*@ pure @*/ boolean isStatic() {
	return (modifiers() & ACC_STATIC) != 0;
    }

    /**
     * Returns the variable definied by this formal parameter
     */
    public /*@ pure @*/ boolean isGenerated() {
	return (getDescription() & DES_GENERATED) != 0;
    }

    /**
     * @return the modifiers of this variable
     */
    public /*@ pure @*/ long modifiers() { 
	return modifiers;
    }

    /**
     * set the modifiers of this variable
     */
    public void setModifiers(long modifiers) { 
	this.modifiers = modifiers;
    }

    /**
     * @return the name of this variable
     */
    public /*@ pure @*/ String ident() {
	return name;
    }

    /**
     * Sets the identifier of this local variable.
     */
    public void setIdent(String name) {
	this.name = name;
    }

    /**
     * @return the type of this variable
     */
    public /*@ pure @*/ CType getType() {
	return type;
    }

    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return isDeclaredNonNull() || context.isFANonNull(this) 
 	        || isFinalNonNull() || (!assigned && exprIsArrayExpression());
    }
    public /*@ pure @*/ boolean isDeclaredNonNull() {
	return hasFlag(modifiers, ACC_NON_NULL | IMPLICITLY_NON_NULL);
    }
    private /*@pure*/ boolean isFinalNonNull() {
	    boolean retVal = false;
	    JExpression value = expr;
	    try { retVal = isFinal() 
		        && value != null
	     	        && value.isNonNull(null);
	    }
	    catch (NullPointerException npe) {
		  retVal = false;
	    }
	    return retVal;
    }
    private /*@pure*/ boolean exprIsArrayExpression() {
        return (expr instanceof JNewArrayExpression)
	    || (expr instanceof JArrayInitializer);
    }

    /**
     * Make this declarator (implicitly) non-null.
     */
    public void setNonNull() {
	// if(!returnType.isNonNull()) {
	//    returnType = ((CClassType) returnType).getNonNullType();
	// }
	modifiers |= IMPLICITLY_NON_NULL;
    }

    /**
     * Sets the initialization expression for this local variable.
     */
    public void setValue(JExpression expr) {
	this.expr = expr;
    }

    /**
     * @return	the value of this local variable at initialization
     * $$$ why not dynamically...
     */
    public /*@ pure @*/ JExpression getValue() {
	return expr;
    }

    /**
     * @return	true if this variable is constant
     */
    public /*@ pure @*/ boolean isConstant() {
	return hasFlag( modifiers, Constants.ACC_FINAL ) 
	    && expr != null && expr.isConstant();
	// int (final int i) {}
    }

    /**
     * Returns the index of this variable in the local block in which
     * it is declared. */
    public /*@ pure @*/ int index() {
	return index;
    }

    /**
     * Sets the index of this variable in the local block in which it
     * is declared. */
    public void setIndex(int index) {
	this.index = index;
    }

    public /*@ pure @*/ int getDescription() {
	return desc;
    }

    /**
     * Returns the index of this variable in the stack frame for the
     * method in which it is declared.  */
    public /*@ pure @*/ int getPosition() {
	return position;
    }

    /**
     * Sets the index of this variable in the stack frame for the
     * method in which it is declared.  */
    public void setPosition(int position) {
	this.position = position;
    }

    public void setUsed() {
	used = true;
    }

    public /*@ pure @*/ boolean isUsed() {
	return used;
    }


    /**
     * Records that this variable is assigned to and reports a warning
     * if the assignment is to a for-loop variable in a location other
     * than the increment of the loop.
     *
     * @see #setIsLoopVariable()
     * @see #unsetIsLoopVariable()
     * @see #setAssigned(TokenReference,CContextType)
     */
    public void setAssigned(TokenReference ref, CContextType context) {
	assigned = true;
	if (isLoopVariable()) {
	    context.
		reportTrouble(new CWarning( ref,
					    MjcMessages.ASSIGNS_LOOP_VARIABLE,
					    ident() ));
	}
    }

    public boolean isAssigned() {
	return assigned;
    }


    /**
     * Registers that this variable is an index variable for a
     * for-loop.  This is used to warn if the variable is assigned to
     * apart from in the increment operation of the loop.  
     *
     * @see #unsetIsLoopVariable()
     * @see #setAssigned(TokenReference,CContextType)
     */
    public void setIsLoopVariable() {
	loopVariable = true;
    }

    /**
     * De-registers that this variable is an index variable for a
     * for-loop.  
     *
     * @see #setIsLoopVariable()
     * @see #setAssigned(TokenReference,CContextType)
     */
    public void unsetIsLoopVariable() {
	loopVariable = false;
    }

    public boolean isLoopVariable() {
	return loopVariable;
    }

    /**
     * Returns true if the given object should be consider equal to
     * this.  Be warned that the equality comparison is true if
     * <code>o</code> is any instance of <code>JLocalVariable</code>
     * with the same <code>index</code> as this.  Thus &quot;false
     * positives&quot; may result if this method is called on two
     * <code>JLocalVariable</code>s from different lexical
     * contexts. !FIXME!
     *
     * @param	o	the object to compare against
     * @return something reasonably useful if handled with care */

    public boolean equals(/*@ nullable @*/ Object o) {
	return o instanceof JLocalVariable &&
	    index == ((JLocalVariable)o).index;
    }

    public int hashCode() {
	return index;
    }

    public JExpression expr() { return expr; }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the local variable and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @exception PositionedError if the check fails */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	//check nullity
	if (CTopLevel.getCompiler().nonNullTypes() && expr != null) {
	   final boolean exprIsNonNullInContext = expr.isNonNull(context);
	   final boolean thisIsDeclaredNonNull = this.isDeclaredNonNull();
	   check( context, 
                  CTypeNullity.isAssignable(this, expr, context),
	          (this.getType().isArrayType() && CTopLevel.getCompiler().experimentalArrayHandling() ?
	           MjcMessages.ASSIGNMENT_ARRAY_NULLITY_INVARIANT :
                   MjcMessages.ASSIGNMENT_OF_NULL));
	   if (!thisIsDeclaredNonNull) {
	      if (exprIsNonNullInContext || exprIsArrayExpression()) {
	         context.addFANonNull(this);
	      } else {
	         context.removeFANonNull(this);
	      }
	   }
	}

	
    }


    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Generates a sequence of bytecode to load
     * the value of the variable on the stack
     * @param	code		the code sequence
     */
    public void genLoad(CodeSequence code) {
	code.plantLocalVar(type.getLoadOpcode(), this);
    }
  
    /**
     * Generates a sequence of bytecodes
     * @param	code		the code sequence
     */
    public void genStore(CodeSequence code) {
	code.plantLocalVar(type.getStoreOpcode(), this);
    }

    // ----------------------------------------------------------------------
    // PUBLIC CONSTANTS
    // ----------------------------------------------------------------------

    public static final int	DES_LOCAL_VAR		= 1 << 0;
    public static final int	DES_PARAMETER		= 1 << 1;
    public static final int	DES_CATCH_PARAMETER	= 1 << 2;
    public static final int	DES_GENERATED		= 1 << 3;

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------
    public String toString() {return "JLocalVariable['"+name+"']";}

    private int			desc;
    private boolean		used;
    private boolean		assigned;
    private boolean		loopVariable;
    protected long		modifiers; 
    protected String		name;
    protected CType		type;
    protected JExpression	expr;

    // index in block local vars array
    private int			index;

    // position in method local vars
    private int			position;
}
