/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLocalVariableExpression.java,v 1.18 2008/03/16 15:52:53 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the AST node for local variable references.
 */
public class JLocalVariableExpression extends JExpression
		implements CInitializable {

	// ----------------------------------------------------------------------
	// CONSTRUCTORS
	// ----------------------------------------------------------------------

	/**
	 * Construct a node in the parsing tree
	 *
	 * @param where the line of this node in the source code
	 */
	public JLocalVariableExpression(TokenReference where,
									JLocalVariable variable) {
		super(where);
		this.variable = variable;
	}

	// ----------------------------------------------------------------------
	// ACCESSORS
	// ----------------------------------------------------------------------

	/**
	 * Indicates whether this expression can appear on the left-hand side of an
	 * assignment or in an increment or decrement operation.
	 * <p/>
	 * <pre><jml>
	 * also ensures true;
	 * </jml></pre>
	 */
	public boolean isMaybeInitializable() {
		return true;
	}

	/**
	 * Returns true if this field accept assignment
	 */
	public boolean isLValue(CExpressionContextType context) {
		return !variable.isFinal() || !isDefinitelyAssigned(context);
	}

	/**
	 * Returns true if this field is already initialized. </pre>
	 */
	public boolean isDefinitelyAssigned(CContextType context) {
		return context.getFlowControlContext().
				isVarDefinitelyAssigned(variable.index());
	}

	private boolean isDefinitelyUnassigned(CContextType context) {
		return context.getFlowControlContext().isVarDefinitelyUnassigned(
				variable.index());
	}

	/**
	 * Records that this variable is initialized in the given context.
	 *
	 * @throws UnpositionedError if this is a final field and has already been
	 *                           assigned to
	 */
	public void initialize(CContextType context)
			throws UnpositionedError {
		if (variable.isFinal() && !isDefinitelyUnassigned(context)) {
			// Cannot assign to a final variable that has either
			// already been assigned to in this context.
			throw
					new UnpositionedError(MjcMessages.REASSIGNMENT_FINAL_VARIABLE,
										  ident());
		}
		context.getFlowControlContext().initializeVariable(variable);
	}

	/**
	 * Returns the position of this variable in the sets of local vars
	 */
	public /*@ pure @*/ int getPosition() {
		return variable.getPosition();
	}

	/**
	 * Compute the type of this expression (called after parsing)
	 *
	 * @return the type of this expression
	 */
	public /*@ pure @*/ CType getType() {
                return (type != null) ? type : variable.getType();
	}

        /**
         * Returns true iff the value represented by this expression is non-null
         */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    return variable.isNonNull(context);
    }

    /**
     * Returns true iff the variable, field or method represente by this expression
     * was declared non_null.
     */
    public /*@pure*/ boolean isDeclaredNonNull() {
	   return variable.isDeclaredNonNull();
    }

	public /*@ pure @*/ String ident() {
		return variable.ident();
	}

	/**
	 * Returns true iff this expression is constant.
	 */
	public /*@ pure @*/ boolean isConstant() {
		return variable.isConstant();
	}

	/**
	 * Returns the literal value of this field
	 */
	public /*@ pure @*/ JLiteral getLiteral() {
		return (JLiteral) variable.getValue();
	}

	public /*@ pure @*/ JLocalVariable variable() {
		return variable;
	}

	public String toString() {
		return variable.toString();
	}

	// ----------------------------------------------------------------------
	// CODE CHECKING
	// ----------------------------------------------------------------------

	/**
	 * Typechecks the expression and mutates the context to record information
	 * gathered during typechecking.
	 *
	 * @param context the context in which this expression appears
	 *
	 * @return a desugared Java expression (see <code>JExpression.typecheck()</code>)
	 *
	 * @throws PositionedError if the check fails
	 */
	public JExpression typecheck(CExpressionContextType context)
			throws PositionedError {
		if (!context.isLeftSide() || !context.discardValue()) {
			variable.setUsed();
		}
		if (context.isLeftSide()) {
			variable.setAssigned(getTokenReference(), context);
		}

		check(context,
			  context.getFlowControlContext().
			  isVarDefinitelyAssigned(variable.index())
			  || variable.isConstant()
			  || (context.isLeftSide() && context.discardValue()),
			  MjcMessages.UNINITIALIZED_VARIABLE, variable.ident());

		if (variable.isConstant() && !context.isLeftSide()) {
			return variable.getValue();
		}

        type = variable.getType();

                // Unless we are on the LHS of an assignment, we have to apply
                // capture conversion (JLS3 5.1.10) to the type of the
                // expression (see JLS3 6.5.6).
		if (!context.isLeftSide()) {
                        try {
                                type = type.getCapture(context);
                        } catch (UnpositionedError e) {
                                throw e.addPosition(getTokenReference());
                        }
                }

		return this;
	}

	// ----------------------------------------------------------------------
	// CODE GENERATION
	// ----------------------------------------------------------------------

	/**
	 * Accepts the specified visitor
	 *
	 * @param	p		the visitor
	 */
	public void accept(MjcVisitor p) {
		p.visitLocalVariableExpression(this);
	}

	public boolean equals(/*@ nullable @*/ Object o) {
		return (o instanceof JLocalVariableExpression) &&
			   variable.equals(((JLocalVariableExpression) o).variable);
	}

	/**
	 * Throws a runtime exception because hashCode isn't properly implemented yet.
	 */
	public int hashCode() {
		fail("An instance of " + getClass() + " cannot be stored in "
			 + " a HashTable or HashSet because equals is overridden but "
			 + " hashCode is not");
		return 0;
	}

	/**
	 * Generates a sequence of bytecodes
	 *
	 * @param	code		the code list
	 */
	public void genCode(CodeSequence code) {
		code.setLineNumber(getTokenReference().line());

		if (!code.discardValue()) {
			variable.genLoad(code);
		}
	}

	/**
	 * Generates a sequence of bytecodes
	 *
	 * @param	code		the code list
	 */
	public void genStartStoreCode(CodeSequence code) {
		// nothing to do here
	}

	/**
	 * Generates a sequence of bytecodes
	 *
	 * @param	code		the code list
	 */
	public void genEndStoreCode(CodeSequence code, boolean discardValue) {
		if (!discardValue) {
			int opcode;

			if (getType().getSize() == 2) {
				opcode = opc_dup2;
			} else {
				opcode = opc_dup;
			}
			code.plantInstruction(new NoArgInstruction(opcode));
		}
		variable.genStore(code);
	}

	// ----------------------------------------------------------------------
	// PRIVATE DATA MEMBERS
	// ----------------------------------------------------------------------

	JLocalVariable variable;

        /**
         * The type of this local variable expression.
         * If the local variable is the target of an assignment, the type
         * simply corresponds to its declared type. Otherwise, the type of this
         * expression is the type resulting from applying capture conversion
         * (JLS3 5.1.10) to the declared type of the variable.
         */
        private CType type;
}
