/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JLoopStatement.java,v 1.4 2004/01/16 23:21:11 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This abstract class is the superclass for all the classes
 * representing loop statements in the AST.  */
public abstract class JLoopStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	comments	the comments associated with this statement
     */
    public JLoopStatement( TokenReference where, 
			   JavaStyleComment[] comments ) {
	super( where, comments );
    }

    // ----------------------------------------------------------------------
    // BREAK/CONTINUE HANDLING
    // ----------------------------------------------------------------------

    /**
     * Returns true iff this statement is a possible target for a
     * break statement.  */
    public boolean acceptsBreak() {
	return true;
    }

    /**
     * Returns true iff this statement is a possible target for a
     * continue statement.  */
    public boolean acceptsContinue() {
	return true;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Registers that there is a break statement within this loop.
     * Stores a clone of the context state up to the break statement
     * that invoked this.  In the event of multiple breaks within this
     * loop the contexts are merged to perform reachability and
     * definite assignment analysis. */
    public void addBreak(CFlowControlContextType context) {
	if (breakContextSummary == null) {
	    breakContextSummary = context.cloneContext();
	} else {
	    breakContextSummary.merge(context);
	}

	breakContextSummary.setReachable( true );
	breaked = true;
    }

    /**
     * Registers that there is a continue statement within this
     * loop. */
    public void addContinue(CFlowControlContextType context) {
	if (continueContext == null) {
	    continueContext = context.cloneContext();
	} else {
	    continueContext.merge( context );
	}
	continueContext.setReachable( false ); // overridden in do-statement
    }

    /**
     * Returns true if this loop statement may be broken
     */
    public boolean hasBreak() {
	return breaked;
    }

    /**
     * Returns a summary of all break context
     */
    public CFlowControlContextType getBreakContextSummary() {
	return breakContextSummary;
    }

    /**
     * Return the end of this block (for break statement)
     */
    public CodeLabel getBreakLabel() {
	return endLabel;
    }

    /**
     * Return the beginning of this block (for continue statement)
     */
    public CodeLabel getContinueLabel() {
	return contLabel;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private CFlowControlContextType	breakContextSummary;
    private boolean		breaked;
    private CodeLabel		contLabel = new CodeLabel();
    private CodeLabel		endLabel = new CodeLabel();

    /**
     * Tracks the flow control analysis information for all control
     * paths that may exist the body of the loop via a continue
     * statement.  */
    protected  CFlowControlContextType continueContext = null;

}
