/**
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMethodCallExpression.java,v 1.67 2008/05/14 18:16:34 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import java.util.HashMap;
import java.util.Map;

/**
 * This class represents method calls
 * <code>methodname( e<sub>1</sub>, e<sub>2</sub>, ..., e<sub>n</sub> )</code>
 */
public class JMethodCallExpression extends JExpression {

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where	the line of this node in the source code
     * @param	prefix	an expression that is a member of an object 
     *			or class representing a method
     * @param	args	the arguments of the call
     */
    public JMethodCallExpression( TokenReference where,
				  JExpression prefix,
				  JExpression[] args)
    {
	this( where, ((JNameExpression) prefix).getPrefix(),
	      ((JNameExpression) prefix).getName(), args );
	this.sourceName = (JNameExpression) prefix;
    }


    /**
     * Constructs a method call node of an AST.  This method is not
     * directly called by the parser.
     * @param where the location of this node in the source code
     * @param prefix the receiver of the method call, or null
     * @param ident the name of the method to be invoked, must be non-null
     * @param args the non-receiver arguments of the call
     */
    protected JMethodCallExpression( TokenReference where, JExpression prefix,
				     String ident, JExpression[] args ) {
	super( where );
	this.prefix = prefix;
	this.ident = ident.intern();
	this.args = args;
    }

    /**
     * Constructs a method call node of an AST.
     * This method is not directly called by the parser.
     * 
     * Used for the universe runtime checks. Before and after calls to static
     * methods additional calls are generated. As these calls are static too,
     * we need a way to stop the recursion so all method calls generated for
     * the universe runtime checks use this constructor and set doSetContext to
     * false.
     * 
     * @param where the location of this node in the source code
     * @param prefix the receiver of the method call, or null
     * @param ident the name of the method to be invoked, must be non-null
     * @param args the non-receiver arguments of the call
     * @param doSetContext whether or not to create the calls to save the current
     * context (will be generated only when the universeDynChecks are enabled)
     * 
     * By scdaniel.
     */ 
    public JMethodCallExpression (TokenReference where, JExpression prefix,
		     String ident, JExpression[] args, boolean doSetContext) {
    	this(where, prefix, ident, args);
    	this.doSetContext = doSetContext;
    	setContext = null;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the name of this field reference as originally appeared
     * in the source file. The return value may be null if this field
     * reference was built by the system.
     */
    public JNameExpression sourceName() {
	return sourceName;
    }

    public JExpression prefix() { 
	return prefix; 
    }

    public String ident() { 
	return ident;
    }

    public JExpression[] args() { return args; }

    /**
     * @return the type of this expression
     */
    public CType getType() {
	//return type == null? method.returnType():type;
	if(type != null){
	    return type;
	}else if(method != null){
	    type = method.returnType();
	    return type;
	}else if(cachedType != null){
	    type = cachedType;
	    return type;
	}
	return null;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	if (method==null) return false;
	return method.isDeclaredNonNull();
    }

    public void setType(CType type) {
	this.type = type;
    }

    public void setPrefix(JExpression expr){
	prefix = expr;
    }

    /**
     * Returns the method being called. This method must be called
     * after typechecking.
     */
    public CMethod method() {
	return method;
    }

    /**
     * Returns true iff this expression can be used as a statement (JLS 14.7)
     */
    public boolean isStatementExpression() {
	return true;
    }

    /**
     * Returns a string representation of this.
     *
     * @return a string representation of this
     */
    public String toString() {
    	/* rzakhejm begin */
    	/* old: return ident + "(...)";*/
    	StringBuffer params = new StringBuffer();
    	for (int i = 0; i < args().length; i++) {
			params.append(args[i].toString());
			if (i < args.length - 1)
				params.append(", ");
		}
    	return prefix() + "." + ident + "(" + params.toString() + ")";
    	/* rzakhejm end */
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see 
     *		<code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	try {
	    CType[] argTypes = checkArgumentTypes( context );

	    if (context.getCompiler().options().multijava()) resolveExternalMethods( context );

	    if( prefix != null ) {
		prefix = prefix.typecheck( context, false, false, false );
		if( prefix instanceof JNameExpression ) {
		    check( context,
			   false,
			   MjcMessages.BAD_METHOD_NAME, 
			   ((JNameExpression) prefix).getName());
		}
		check( context,
		       prefix.getType().isReference(),
		       MjcMessages.METHOD_BADPREFIX, ident, prefix.getType() );
	    }
	    CClass local = context.getClassContext().getHostClass();
	    method = lookupMethod( argTypes, context );
	    CSpecializedType[] paraTypes = method.parameters();
	    //CType[] actualType = new CType[argTypes.length];  
	    HashMap hash = new HashMap();
	    if (prefix instanceof JThisExpression) {
	    	//doing nothing 
	    } else {
	    	for (int i = 0; i < argTypes.length; i++) {		 
	    		CType tt = paraTypes[i].staticType();
	    		if (tt.isTypeVariable() && tt.isMethodTypeVariable()) {
	    			CClassType temp
	    					= (CClassType) paraTypes[i].staticType();
	    			String ident = temp.getIdent();
	    			if (hash.containsKey(ident)) {
	    				temp = (CClassType) hash.get(ident);
	    				
	    				if (!argTypes[i].isAlwaysAssignableTo(temp)) {
	    					if (temp.isAlwaysAssignableToIgnoreUniverses(argTypes[i])) {
	    						hash.put(ident, argTypes[i]);
	    					} else {
	    	    				throw new UnpositionedError(
	    	    						MjcMessages.METHOD_TV_INCONSISTENT,
	    	    						((CTypeVariable) tt).getIdent());
	    					}
	    				}
	    			} else {
	    				temp = (CClassType) argTypes[i];
	    				hash.put(ident, temp);
	    			}
	    		}
	    	}
	    }
	    	    
	    // RMJ generates calls to external gfs differently
	    // from MJ.  In particular, RMJ first computes all
	    // args and assigns them to local vars.  This is
	    // necessary so that any external methods that
	    // becomes reachable by computing the args will get
	    // loaded before we access the gf's top dispatcher.
	    // (For regular MJ to work properly, it suffices that
	    // the code for the receiver, rather than all args,
	    // is generated before the code for the gf.)
	    if (context.getCompiler().RMJ() && method.isInExternalGF()) {
		CType[] localVarTypes =
		    CSpecializedType.tupleOfSpecializersFrom(
			method.parameters() );
		argLocalVars = new JLocalVariable[localVarTypes.length];
		for (int i = 0; i < argLocalVars.length; i++) {
		    argLocalVars[i] = 
			new JGeneratedLocalVariable(org.multijava.util.compiler.TokenReference.NO_REF,
						    0, localVarTypes[i],
						    "arg" + hashCode() + i /* should be unique */,
						    null );
		    try {
			context.getFlowControlContext().
			    addVariable( argLocalVars[i] );
		    } catch( UnpositionedError e) {
			throw e.addPosition( getTokenReference() );
		    }
		}
	    }

	    if (prefix instanceof JSuperExpression) {
		handleSuperPrefix( argTypes, context );
	    }

	    // record the top method of the generic function,
	    // useful for code generation
	    topConcreteMethod = method.topConcreteMethod();

	    processExceptions( context );

	    ensurePrefixIsInitialized( context );

	    // -------------------------------------------------------------
	    // do some typechecks
	    check( context, 
		   method.access().isStatic() || 
		   !(prefix instanceof JTypeNameExpression), 
		   MjcMessages.METHOD_STATIC_BAD, method.ident() );
	    
	    if (method.isDeprecated()) {
		context.reportTrouble(
		    new CWarning( getTokenReference(),
				  MjcMessages.USE_OF_DEPRECATED,
				  method.ident() ) );
	    }

	    if( method.access().isStatic() && prefix != null 
		&& !(prefix instanceof JTypeNameExpression) ) {
		context.reportTrouble( 
		    new CWarning( getTokenReference(),
				  MjcMessages.INSTANCE_PREFIXES_STATIC_METHOD,
				  method.ident() ));
	    }

	    // Is this method call in the class in which this
	    // method is defined?
	    if( prefix instanceof JThisExpression 
		&& !method.isStatic() ) {
		// make sure the super constructor has already been called
		// else it is an error to call an instance method
		check( context, !context.isBeforeSuperConstructorCall(),
		       MjcMessages.CONSTRUCTOR_EXPLICIT_CALL );
	    }	    

	    // Is this a self call to an external method from a constructor
	    // or initializer?  If so, warn that the chain-of-responsibility
	    // may not be initialized.
        if( prefix == null || prefix.getType().isAssignableTo(local.getType()) ) {
            if( method.isInExternalGF() 
                && (context.isInConstructor() || context.isInInitializer()) ) {
                context.reportTrouble(
                    new CWarning( getTokenReference(),
                          MjcMessages.EXTERNAL_SELF_CALL_FROM_INIT,
                          method.ident() ) );
            }
        }

    	// Checks for Generic Universe Types
	    if (CTopLevel.getCompiler().universeChecks()) {
	    	typecheckUniverse(context, hash);
	    }
	    
	    // scdaniel: universe runtime checks
	    if ( CTopLevel.getCompiler().universeDynChecks() &&
		 method.isStatic() &&
		 !context.isStatic() &&
		 doSetContext ) {

	    	createUniverseDynChecks(context);
	    }	    

            if (CTopLevel.getCompiler().Generic()) {
                typecheckUncheckedWarnings(argTypes, context);
            }
		
	    
	    // add necessary casts to the argument expressions
	    argTypes = 
		CSpecializedType.tupleOfSpecializersFrom( method.parameters());
	    for( int i = 0; i < argTypes.length; i++ ) {
		if( args[i] instanceof JTypeNameExpression ) {
		    check( context, false,
			   MjcMessages.VAR_UNKNOWN, 
			   ((JTypeNameExpression) args[i]).qualifiedName().
			   replace('/','.'));
		}
		args[i] = args[i].convertType( argTypes[i], context );
	    }

	    // Mark method as used if it is a source method
	    if( method instanceof CSourceMethod ) {
		((CSourceMethod) method).setUsed();
	    }

	    // check nullity
	    if (CTopLevel.getCompiler().nonNullTypes()) {
    	       typecheckNullity(context);
	    }

            // We have to apply capture conversion (JLS3 5.1.10) to the return
            // type of a method call expression (see JLS3 15.12.2.6).
	    setType(getType().getCapture(context));

            handlePossibleRawTypeAccess(context);

	    if(prefix instanceof JThisExpression) {
		return this;
	    }
	    
	    //	    type = method.returnType();
	    CType returntype = getType();
	    if(getType().isTypeVariable()) {
		CClassType subtype;
		if (getType().isMethodTypeVariable()){
		    String ident = getType().getIdent();
 		    CType temp = (CType)hash.get(ident);
 		    if (hash.containsKey(ident)) {
 			subtype = (CClassType)temp;
 		    } else {
 			//method.setReturnType(getType());
			subtype = (CClassType)returntype;
		    } 
		} else { // is not methodTypeVariable but typevariable
		    subtype = prefix.getType().getCClass().getSubstitution
					((CTypeVariable) getType(),
					prefix.getType().isTypeVariable()
							? ((CTypeVariable) prefix.getType()).getBounds()[0].getAllArguments()
							: prefix.getType().getAllArguments());

		}
		if(!(subtype.getErasure(context).equals
		     (returntype.getErasure(context)))) {
		    JCastExpression castexpr =
			new JCastExpression(getTokenReference(), 
					    new JCheckedExpression(getTokenReference(), this), 
					    subtype,
                                            true);
		    castexpr.typecheck(context);
		    return castexpr;
		} else {
		    setType(subtype);
		}
	    }
	    if(!method.isStatic() && getType().isGenericType()) {  //return type is Stack<T>
		CClassType newType = 
		    ((CClassType)getType()).createSubstitutedType
		    (local,
		     (CClassType)prefix.getType(),
		     getType().getAllArguments());
		setType(newType);
	    }
	    
	    if(!method.isStatic() && getType().isArrayType()) {
		if(((CArrayType)getType()).getBaseType().isGenericType()) {
		    CType basetype = ((CArrayType)getType()).getBaseType();
		    CClassType newbase = 
			((CClassType)basetype).createSubstitutedType
			(local,
			 (CClassType)prefix.getType(),
			 basetype.getAllArguments());
		    CArrayType newarray = 
			new CArrayType(newbase, 
				       ((CArrayType)getType()).getArrayBound(),
				       ((CArrayType)getType()).getCUniverse());
		    setType(newarray);
		}else if (((CArrayType)getType()).getBaseType().isTypeVariable()) {
		    CClassType subtype;
		    if(((CArrayType)getType()).getBaseType().isMethodTypeVariable()) {
			String ident = ((CArrayType)getType()).getBaseType().getIdent();
			CType temp = (CType)hash.get(ident);
			if (hash.containsKey(ident)) {
			    subtype = (CClassType)temp;
			} else {
			    //method.setReturnType(getType());
			    subtype = (CClassType)((CArrayType)getType()).getBaseType();
			} 
		    } else { //is a type variable but not a method typevariable
			subtype = 
			    prefix.getType().getCClass().getSubstitution
			    ((CTypeVariable)((CArrayType)getType()).getBaseType(),
			     prefix.getType().getAllArguments());
			subtype = 
			    new CArrayType(subtype,((CArrayType)getType()).getArrayBound(),((CArrayType)getType()).getCUniverse());
		    }
		    if(!(getType().getErasure(context).equals
			 (subtype.getErasure(context)))) {
			JCastExpression castexp = 
			    new JCastExpression(getTokenReference(),
						new JCheckedExpression
						(getTokenReference(),
						 this),
						subtype,
                                                true);
			castexp.typecheck(context);
			return castexp;
		    }
		}
		//array
		
	    }
	    if ((prefix!=null)&&(args!=null)){
	    
                boolean isSiaChecked = (ident.compareTo("checked")==0)&&(prefix.toString().indexOf("SafeIntegralArithmetic")>0)&& method.access().isStatic()&& method.access().isFinal();
	    
                if ( isSiaChecked && (!args[0].getType().equals(CStdType.Double))&&(!args[0].getType().equals(CStdType.Float))) {
		    MJMathModeExpression mmExp = new MJMathModeExpression(getTokenReference(),"\\safe_math",args[0],Constants.AMID_SAFE_MATH);
		    return mmExp.typecheck(context);
                }
		
		boolean isSiaUnchecked = (ident.compareTo("unchecked")==0)&&(prefix.toString().indexOf("SafeIntegralArithmetic")>0)&& method.access().isStatic()&& method.access().isFinal();

		if (isSiaUnchecked && (!args[0].getType().equals(CStdType.Double))&&(!args[0].getType().equals(CStdType.Float))) {
                    MJMathModeExpression mmExp = new MJMathModeExpression(getTokenReference(),"\\java_math",args[0],Constants.AMID_JAVA_MATH);
                    return mmExp.typecheck(context);
		}
	    }
	    if(type != null) {cachedType = type;}
	    return this;
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
    }

    
    /** Create the code for the dynamic checks.
     *
     * @param context the expression context from typecheck
     * @throws PositionedError something happened
     * 
     * by scdaniel.
     */
    private void createUniverseDynChecks(CExpressionContextType context) throws PositionedError {
		JExpression args[] = new JExpression[1];
		if ( (prefix != null) &&
		     !(prefix instanceof JThisExpression) ) {

		    CUniverse universe = ((CClassType) prefix.getType()).getCUniverse();
		    if ( !(universe instanceof CUniversePeer) ) {
			/* get reference to temporary variable ... */
			tmpLocalVar = CUniverseRuntimeHelper.getTmpVariable(context, this, type);
			JLocalVariableExpression tmpVar =
			    new JLocalVariableExpression(TokenReference.NO_REF, tmpLocalVar);
					
			/*
			 * Create new rep object used as the context for the
			 * called static method if called on a typename or if
			 * the variable is null.
			 */
			repObject = new JNewObjectExpression(
					    TokenReference.NO_REF,
					    CTopLevel.getTypeRep(
					       Constants.JAV_OBJECT,
					       CUniverseRep.getUniverse(),
					       true),
					    null,
					    new JExpression[0]);
					
			if ( (prefix instanceof JTypeNameExpression) ||
			     ((prefix instanceof JParenthesedExpression) &&
			      (((JParenthesedExpression) prefix).expr() instanceof JTypeNameExpression)) ) {
			    repObject.typecheck(context);
			} else {
			    repTest =
				new JIfStatement( TokenReference.NO_REF,
				    new JEqualityExpression(
						     TokenReference.NO_REF,
						     OPE_EQ,
						     tmpVar,
						     new JNullLiteral(TokenReference.NO_REF)),
				    new JExpressionStatement(
						     TokenReference.NO_REF,
					     new JAssignmentExpression(
							TokenReference.NO_REF,
							tmpVar, repObject),
						     null),
				    null, null);

			    repTest.typecheck(context.getFlowControlContext());
			}

			args[0] = tmpVar;
		    } else {
			args[0] = new JThisExpression(TokenReference.NO_REF);
		    }
		} else {
		    args[0] = new JThisExpression(TokenReference.NO_REF);
		}

		// generate method calls
		setContext =
		    new JMethodCallExpression(TokenReference.NO_REF,
				      CUniverseRuntimeHelper.getHashTable(),
					      "setContext",
					      args, false);

		resetContext =
		    new JMethodCallExpression(TokenReference.NO_REF,
				      CUniverseRuntimeHelper.getHashTable(),
					      "resetContext",
					      new JExpression[0], false);
			
		// temporarily disable universe checks ...
		boolean universeChecks = CTopLevel.getCompiler().universeChecks();
		boolean universePurity = CTopLevel.getCompiler().universePurity();
		CTopLevel.getCompiler().setUniverseChecks(false);
		CTopLevel.getCompiler().setUniversePurity(false);

		// ... typecheck ...
		setContext.typecheck(context);
		resetContext.typecheck(context);

		// ... enable universe checks again.
		CTopLevel.getCompiler().setUniverseChecks(universeChecks);
		CTopLevel.getCompiler().setUniversePurity(universePurity);
	}


	/**
     * Issues unchecked warnings for the actual parameters which need unchecked
     * conversion (JLS3 5.1.9) to conform to their corresponding formal
     * parameters.
     *
     * @param actuals  The actual arguments of this method call.
     * @param context  The type checking context.
     *
     * @throws PositionedError if some error occurs during type checking.
     */
    private void typecheckUncheckedWarnings(
            CType[] actuals,
            CExpressionContextType context) throws PositionedError {
        if (prefix == null) {
            return;
        }

        CClassType prefixType = (CClassType) prefix.getType();

        // The types of a non-static method accessed on a raw type are erased.
        // Therefore, no unchecked conversion is ever necessary.
        if (!method.isStatic()
            && prefixType.isRawTypeInvocationOf(method.owner())) {
            return;
        }

        CClass clazz = prefixType.getCClass();
        CClassType[][] args = prefixType.getAllArguments();

        CSpecializedType[] formals = method.parameters();
        for (int i = 0; i < formals.length; i++) {
            CType formal = formals[i].dynamicType();

            // Substitute type variables by its actual type arguments.
            if (formal.isClassTypeVariable()) {
                formal = clazz.getSubstitution((CTypeVariable) formal, args);
            } else if (formal.isArrayType()
                       && ((CArrayType) formal).getBaseType().isClassTypeVariable()) {
                CArrayType array = (CArrayType) formal;
                CTypeVariable var = (CTypeVariable) array.getBaseType();
                CClassType subst = clazz.getSubstitution(var, args);
                formal = new CArrayType(subst, array.getArrayBound(), array.getCUniverse());
            }

            if (actuals[i].needsUncheckedConversion(formal)) {
                warn(context,
                     MjcMessages.UNCHECKED_CONVERSION,
                     actuals[i],
                     formal);
            }
        }
    }

    /**
     * If a non-static method is accessed as a member of a raw type, the
     * following actions are performed (see JLS3 4.8):
     * <ul>
     *   <li>
     *     Replace the type of the method return value by its erasure.
     *   </li>
     *   <li>
     *     If any of the formal parameter types of the method changes by
     *     erasure, issue an unchecked warning.
     *   </li>
     * </ul>
     *
     * @param context  The typechecking context.
     *
     * @throws PositionedError if an error occurs during typechecking.
     */
    private void handlePossibleRawTypeAccess(CExpressionContextType context)
            throws PositionedError {
        if (prefix == null) {
            return;
        }

        if (method.isStatic()) {
            return;
        }

        CClassType prefixType = (CClassType) prefix.getType();
        // If a non-static method is accessed on a raw type, the expression's
        // type is the erasure of the method return type (see JLS3 4.8).
        if (prefixType.isRawTypeInvocationOf(method.owner())) {
            try {
                setType(getType().getErasure(context));
            } catch (UnpositionedError e) {
                throw e.addPosition(getTokenReference());
            }

            // As we are invoking the method on a raw type, if any of the method's
            // formal parameter types (not the return type!) changes by erasure,
            // we have to issue an unchecked warning (JLS3 4.8).
            CSpecializedType[] params = method.parameters();
            for (int i = 0; i < params.length; i++) {
                if (params[i].staticType().changesByErasure()
                    || params[i].dynamicType().changesByErasure()) {
                    warn(context,
                         MjcMessages.UNCHECKED_METHOD_CALL,
                         method.toString(true),
                         method.owner().qualifiedName().replace('/', '.'));
                    break;
                }
            }
        }
    }

    private void typecheckNullity(CExpressionContextType context)
	throws PositionedError 
    {
    	    // check nullity of prefix
//if (true || prefix!=null && !prefix.isNonNull(context)) {
//System.out.println("-------------------");
//System.out.println("prefix is "+prefix);
//if (prefix!=null) System.out.println("prefix is a "+prefix.getClass());
//if (prefix!=null) System.out.println("prefix is declared = "+prefix.isDeclaredNonNull());
//if (prefix!=null) System.out.println("prefix is non-null = "+prefix.isNonNull(context));
//if (prefix instanceof org.multijava.mjc.JClassFieldExpression) {
//System.out.println("prefix.prefix is  = "+((org.multijava.mjc.JClassFieldExpression)prefix).prefix());
//System.out.println("prefix.prefix is  a "+((org.multijava.mjc.JClassFieldExpression)prefix).prefix().getClass());
//System.out.println("prefix.field is  = "+((org.multijava.mjc.JClassFieldExpression)prefix).ident());
//System.out.println("prefix.field is  a "+((org.multijava.mjc.JClassFieldExpression)prefix).getType());
//}
//}
	    check(context, prefix==null || prefix.isNonNull(context) || isInnerThis(prefix), MjcMessages.METHODCALL_NULL_REF, ident);

	    // check nullity of parameters
            final JFormalParameter[] formals = getFormalParameters();
	    if (formals==null) {
	       context.reportTrouble( new CWarning( getTokenReference(),
					  MjcMessages.METHOD_NULLITYUNCHECKABLE, 
					  ident ));
            } else {
               for (int i=0; i<args.length; i++) {
                   final JExpression actual = args[i];
	           check( context, 
			  CTypeNullity.isAssignable(formals[i], actual, context),
	                  (formals[i].getType().isArrayType()            ?
			   MjcMessages.PARAMETER_ARRAY_NULLITY_INVARIANT :
                           MjcMessages.METHODPARAM_CANNOT_BE_NULL), 
			  new String[] {""+(i+1), formals[i].ident()} );
	       }
	    } 
    }
    
    static boolean isInnerThis(JExpression expr) {
	    if (expr instanceof JClassFieldExpression) { 
	       JClassFieldExpression classField = (JClassFieldExpression) expr;
	       JExpression prefix = classField.prefix();
	       if (prefix instanceof JThisExpression) {
		  String ident = classField.ident();
		  return ident!=null && ident.equals("this$0");
	       }
	       return isInnerThis(prefix);
	    }
	    return false;
    }

    /**
     * Typechecks args and returns an array of the argument types.
     *
     * <pre><jml>
     * requires context != null;
     * modifiable args;
     * ensures \result != null && \result.length == args.length;
     * </jml></pre>
     *
     */
    protected CType[] checkArgumentTypes( CExpressionContextType context ) 
	throws PositionedError
    {
	CType[] argTypes = new CType[args.length];
	
	for( int i = 0; i < argTypes.length; i++ ) {
	    args[i] = args[i].typecheck( context, false, false, false );
	    argTypes[i] = args[i].getType();
	}

	return argTypes;
    }

    /**
     * Makes sure any external generic functions that are implicitly
     * imported are actually available in the data structures.
     */
    protected void resolveExternalMethods( CExpressionContextType context ) {
	context.resolveMaybeExtMethodRef( ident );
    }

    protected JFormalParameter[] getFormalParameters() {
        if (method instanceof CSourceMethod) {
           CSourceMethod sourceMethod =  (CSourceMethod) method;
	   JMemberDeclaration memberDecl = sourceMethod.declarationASTNode();
	   JMethodDeclaration methodDecl = (JMethodDeclaration) memberDecl;
           if (methodDecl == null) // how can this ever be true? 
              return null;         // but it is when compiling ESC/Java2
	   else
              return methodDecl.parameters();
        }
        return null;
    }

    /**
     * Returns the apparent target method for this call.  (We say apparent 
     * because it may be that some synthetic redirection is needed in the
     * generated code to account for compilation techniques of MultiJava or
     * Java workarounds for the JVM's lack of support for inner classes.)
     */
    protected CMethod lookupMethod( CType[] argTypes, 
				    CExpressionContextType context ) 
	throws UnpositionedError, PositionedError
    {
	CMethodSet results;
	if (prefix != null) {
	    CClassType[] sub = prefix.getType().isGenericType()?
		prefix.getType().getArguments():CClassType.EMPTY;
	    results = prefix.getType().getCClass().lookupMethodOrSet(
		ident, argTypes, sub,context);
	} else {
	    results = context.lookupMethodOrSet( ident, argTypes );
	}
	
	if (results != null && results.size() > 1) {
	    // we should only get here if we're doing RMJ compilation
	    assertTrue(context.getCompiler().RMJ());

	    // RMJ: this can happen if we're invoking an external GF,
	    // and there are multiple ambiguous callee methods.
	    // Later glue might fix the problem.  So warn for now.
	    warn ( context,
		   MjcMessages.RMJ_METHOD_AMBIG,
		   results.getMethod(0), results.getMethod(1) );

	    // the result of this code uses the "invoked" method just
	    // as a representative of the method that will be invoked;
	    // at run-time, any overriding method might be invoked.
	    // If we just pick one of the target methods, e.g. the
	    // first, then this static interface invariant should be
	    // preserved.
	}

	CMethod result = results == null ? null : results.getFirst();

	checkMethodFound( result, argTypes );
	return result;
    }

    /**
     * Builds an error message and throws an exception if
     * <code>found</code> is null.
     *
     * <pre><jml>
     * ensures found != null ==> true;
     * signals_only CMethodNotFoundError;
     * signals (CMethodNotFoundError) found == null;
     * </jml></pre>
     * */
    private void checkMethodFound(CMethod found, CType[] argTypes) 
	throws CMethodNotFoundError 
    {
	if( found == null ) {
	    // report error, the message takes some cooking...
	    String prefixName;

	    if( prefix instanceof JNameExpression ) {
		prefixName = 
		    ((JNameExpression) prefix).qualifiedName() + ".";
	    } else if( prefix instanceof JTypeNameExpression ) {
		prefixName =  
		    ((JTypeNameExpression) prefix).qualifiedName() + ".";
	    } else if (prefix != null) {
		prefixName = prefix.getType().toString() + ".";
	    } else {
		prefixName = "";
	    }

	    throw new CMethodNotFoundError( getTokenReference(), 
					    prefixName + ident, argTypes,
					    false );
	}
    }

    /**
     * Handles processing of cases where prefix is a
     * <code>super</code> expression, including various hacks for
     * MultiJava super sends and for super sends to inner and outer
     * classes.
     *
     * <pre><jml>
     * modifies method, prefix, localMethod, skipLocalFunction, isCase_s_e2;
     * </jml></pre>
     *
     */
    protected void handleSuperPrefix( CType[] argTypes, 
				      CExpressionContextType context ) 
	throws UnpositionedError, PositionedError
    {
	CClass callerOwner = 
	    context.getClassContext().getOwnerClass();
	CMethod caller = context.getCMethod();
	// An outer super call (the first disjunct below)
	// always needs a wrapper method since the
	// verifier checks that invokespecial only calls a
	// superclass method.  The second disjunct handles
	// case s-e1 from Clifton-01, Figure 24, where the
	// calling method will be compiled into a nested
	// dispatcher class.
	if (prefix.getType() != callerOwner.getSuperClass().getType()
	    || (!caller.isExternal() && caller.isInExternalGF() &&
		method != null && !method.isInExternalGF())
	    ) {
	    method =
		((JSuperExpression) prefix).targetSubclass().
		wrapSuperCallTo(method);
	    // prefix will be set appropriately below
	    prefix = null;
	}

	// If method is null we report the error below.
	if( method != null ) {
	    // If method is not null we want to be sure it
	    // is concrete:
	    check( context, !method.isAbstract(),
		   MjcMessages.SUPER_METHOD_ABSTRACT, method );

	    CMethodSet maybeLocalMethods =
		callerOwner.lookupMethodOrSet(ident,argTypes,
					      CClassType.EMPTY,context);

	    if (maybeLocalMethods != null && maybeLocalMethods.size() > 1) {
		// we should only get here if we're doing RMJ compilation
		assertTrue(context.getCompiler().RMJ());

		// RMJ: we have multiple ambiguous methods (which
		// we assume are part of an external GF and involve
		// specializing on interfaces).  This problem could be
		// fixed if an appropriate glue method is added later, but
		// for now, just report an error.
		warn( context, 
		      MjcMessages.RMJ_AMBIGUOUS_SUPER, 
		      maybeLocalMethods.getMethod(0),
		      maybeLocalMethods.getMethod(1) );
	    }

	    CMethod maybeLocalMethod = 
		maybeLocalMethods == null ? null : 
		maybeLocalMethods.getFirst();

	    if (method.topConcreteMethod().equals( caller.topConcreteMethod )) {
		// super send to the same generic function

		// if method is not in caller.overriddenMethodSet()
		// then give warning that super call is skipping a
		// portion of the generic function hierarchy
		CMethodSet directlyOverriddenMethodSet = 
		    ((CSourceMethod) caller).overriddenMethodSet();

		// Because the super class includes 'method' and the
		// local class contains 'caller' it must at least be
		// the case the caller specializes some method in the
		// superclass (although not necessarily 'method'!), so...

		//@ assert directlyOverriddenMethodSet != null;
		
		// filter out interface methods so we don't
		// give a spurious caution below
		CMethodSet overriddenMethodSet = new CMethodSet();
		CMethodSet.Iterator iter =
		    directlyOverriddenMethodSet.iterator();
		while (iter.hasNext()) {
		    CMethod m = iter.next();
		    if (!(m.isAbstract() && m.owner().isInterface())) {
			overriddenMethodSet.addMethod(m);
		    }
		}

		// we issue a caution if there is at least one
		// non-interface method that we are skipping over
		if (! (overriddenMethodSet.size() == 0 ||
		       overriddenMethodSet.contains( method )) ) {
		    if (context.getCompiler().RMJ()) {
			// XXX for now, we don't allow this in RMJ,
			// because we don't generate code to define
			// the oldSuperFunction field yet.
			fail(context,
			     overriddenMethodSet.size() == 1 ?
			     MjcMessages.RMJ_SUPER_SKIPS_METHOD :
			     MjcMessages.RMJ_SUPER_SKIPS_METHODS);
		    } else {
			warn(context,
			     overriddenMethodSet.size() == 1 ?
			     MjcMessages.SUPER_SKIPS_METHOD :
			     MjcMessages.SUPER_SKIPS_METHODS);
		    }
		}
		
		if (caller.isExternal()) {
		    if (context.getCompiler().RMJ()) {
			// under RMJ, each top-level method goes into
			// its own dispatcher (the non-RMJ code below
			// seems to assume that all top-level methods
			// are in the same base-anchor dispatcher).
			// so treat it like a call from a subclass
			// dispatcher.
			isCase_s_e2 = true;

			// we need to skip the local members
			// of the GF (see uses in genCode())
			skipLocalFunction = true;
			localMethod = maybeLocalMethod;
		    } else {
			// super send to a portion of the
			// local external generic function
			// (Clifton-01, Figure 24, case s-e4)
			CSourceGFCollection callerHost =
			    (CSourceGFCollection) 
			    context.getClassContext().getHostClass();
			method = callerHost.superclassMethDispFor( 
			    method, callerOwner );
		    }
		    prefix = null;
		} else if (caller.isInExternalGF()) {
		    // super send from an internal method
		    // of an external generic function
		    // (Clifton-01, Figure 24, case s-e2)
		    isCase_s_e2 = true;

		    // we need to skip the local members
		    // of the GF (see uses in genCode())
		    skipLocalFunction = true;
		    localMethod = maybeLocalMethod;
				
		    // prefix will be set appropriately below
		    prefix = null;
		}
	    } else {
		// super send to a different generic function

		// It is illegal for an external method to use
		// a super send to call a different generic
		// function, since that can break
		// encapsulation [see Clifton-01, section 2.3.1].
		check( context, !caller.isExternal(),
		       MjcMessages.ILLEGAL_EXTERNAL_SUPER );

		if (!caller.isExternal() && method.isInExternalGF()) {
		    // For cases s-i3 and s-e3 from
		    // Clifton-01, Figure 24, we need to skip
		    // any local members of the target GF.
		    if (maybeLocalMethod.owner() == callerOwner) {
			// used in genCode() to skip the
			// locally declared members of the GF
			skipLocalFunction = true;
			localMethod = maybeLocalMethod;
		    }
		    // prefix will be set appropriately in typecheck
		    prefix = null;
		}
	    }
	}
    }

    /**
     * Processes the exceptions thrown by the target method and
     * registers them with the context.
     */
    private void processExceptions( CExpressionContextType context ) {
	// process the exceptions thrown by the method and register
	// them with the context
	CClassType[] exceptions = method.throwables();
	for( int i = 0; i < exceptions.length; i++ ) {
	    // Only checked exception need to be added to the
	    // surrounding context.  This also handles a.clone() where
	    // 'a' references an array as a special case. We don't need
	    // to catch CloneNotSupportedException for an invocation
	    // of clone() on an array.
	    if( exceptions[i].isCheckedException() 
		&& !(prefix != null
		     && prefix.getType().isArrayType() 
		     && ident == Constants.JAV_CLONE
		     && exceptions[i].getCClass().qualifiedName().
		     equals("java/lang/CloneNotSupportedException") ) ) {
		context.getFlowControlContext().addThrowable(
		    new CThrowableInfo( exceptions[i], this ));
	    }
	}
    }


    /**
     * Ensures that, if the target method is not static, then the
     * prefix field contains an appropriate expression that evaluates
     * to the method's receiver.
     *
     * <pre><jml>
     * requires context != null;
     * modifies prefix;
     * ensures prefix != null || method.access().isStatic();
     * </jml></pre>
     *
     */
    private void ensurePrefixIsInitialized( CExpressionContextType context ) 
	throws PositionedError
    {
	// For non-static methods the last argument to the method call
	// in bytecode is the logical receiver of the method.  For
	// internal generic functions this receiver is used to begin
	// the virtual method search.  For external generic functions
	// this receiver is the last argument to the static apply()
	// method.  The following code sets prefix to be an expression
	// that yields a reference to the appropriate receiver.
	if( (prefix == null) && !method.access().isStatic() ) {
	    CMemberHost callerHost = context.findNearestHost();
	    CClass targetOwner = method.owner();

	    if( targetOwner == callerHost // local call, or ...
		|| (callerHost instanceof CClass && // call to inherited method
		    ((CClass) callerHost).descendsFrom( targetOwner )) ) {
		// so just use a normal "this" expression
		prefix = new JThisExpression( getTokenReference() ).
		    typecheck( context );
	    } else {
		// This is either a call to a method of an outer class
		// or a call from within an external method to a
		// method of the external method's receiver class.  In
		// either case the prefix reference should be to a
		// generated variable (either an "outer this" or a
		// parameter of the external method).  The
		// typechecking and code generation of the "this"
		// expression handles appropriate conversion to a
		// generated variable reference.
		prefix = 
		    new JThisExpression(getTokenReference(), targetOwner).
		    typecheck( context );
	    }
	    //System.out.println( "prefix = " + prefix );
	}
    }



    /**
     * Typecheck the universe properties of the method call.
     * Some general checks are performed here and then the specialised
     * functions for the different method types are invoked.
     * In the end the attribute result_universe will
     * contain the correct universe modifier.
     * The method setType has to be called to set the type and universe
     * modifier!
     *
     * by WMD
     * by RZU
     */
	private void typecheckUniverse(CExpressionContextType context,
			Map methodTypeVariablesMap)
			throws PositionedError {
        // TODO why is the prefix null sometimes???
		if( prefix==null ) return;
        // Get the type of the prefix
		CClassType prefixType = (CClassType) prefix.getType();
    	
    	// If we're dealing with a type variable, we have to take
    	// the upper bound
    	if (prefixType.isTypeVariable()) {
    		prefixType = ((CTypeVariable) prefixType).getBounds()[0];
    	}
    	
    	// Get universe of the prefix
    	CUniverse prefixUniverse = prefixType.getCUniverse();
    	
    	// Calls of the form "rep T.m()" within a static method are forbidden
    	// static calls can not contain type arguments, so checking the main
    	// modifier should be enough
	 	check( context,
			!( context.isStatic() &&
		     prefixUniverse instanceof CUniverseRep),
		     CUniverseMessages.REP_CALL_IN_STATIC_METHOD_FORBIDDEN,
		     method.ident() );
    	
    	// Non-pure methods must not be called in the body of pure methods
	    if (CTopLevel.getCompiler().universePurity()) {
	    	check(context,
	    			!(context.isPure() && !method.isPure()),
	    			CUniverseMessages.NON_PURE_CALL_IN_PURE_METHOD_FORBIDDEN,
	    			method.ident());
		}
    	
    	// Ensure non-pure method isn't called on readonly type
    	check(context,
    		method.isPure() ||
    				!(prefixUniverse instanceof CUniverseReadonly),
   			CUniverseMessages.NON_PURE_ON_READONLY_FORBIDDEN,
   			method.ident());
    			
		// Determine the resulting type of the method call on the given
		// expression - only if a class type
    	if (method.returnType() instanceof CClassType) {
    		CClassType substitutedReturnType = CUniverseServices
    				.substituteInheritedTypeVariables(
    						(CClassType) method.returnType(),
    						(CClassType) prefix.getType(),
    						context);
 
    		substitutedReturnType = CUniverseServices
    				.combine(prefixType,
    						substitutedReturnType);
    		
    		type = getFullySubstitutedType(substitutedReturnType,
    				methodTypeVariablesMap,
    				context);
    	}

    	// Does the call happen on this?
    	boolean onThis = prefixUniverse instanceof CUniverseThis;
    	
    	// If the call doesn't happen on this, the upper bounds of the
    	// method type variables mustn't contain rep modifiers
    	if (!onThis) {
    		CTypeVariable[] methodTypeVariables = method.getTypeVariable();
    		
    		for (int i = 0; i < methodTypeVariables.length; i++) {
    			check(context,
    					!CUniverseServices
    							.containsRepModifier(methodTypeVariables[i].getBounds()),
    							CUniverseMessages.REP_FORBIDDEN_IN_BOUND);
    		}
    	}
    	
    	// Check the parameters and arguments by iterating over them
    	CType[] parameterTypes
    			= CSpecializedType.staticTypeTupleFrom(null, method.parameters());
    	CClassType currentParameterType;
    	for (int i = 0; i < parameterTypes.length; i++) {
    		// We only care about parameters of class type
    		if (parameterTypes[i] instanceof CClassType) {
    			currentParameterType = (CClassType) parameterTypes[i];
    		} else {
    			continue;
    		}
    		
    		// Ensure the argument respects the universe modifiers
    		typecheckUniverseArgument(currentParameterType,
    				args[i],
    				methodTypeVariablesMap,
    				context,
    				i + 1);
    		
    		// If the call doesn't happen on this, no rep modifiers must occur
    		check(context,
    				onThis || !CUniverseServices
    					.containsRepModifier(currentParameterType),
    				CUniverseMessages.REP_NOT_ON_THIS_IN_PARAMETER,
    				new Object[] { method.ident(),
    				  new Integer(i+1),
    				  currentParameterType.toVerboseString() });

    	}
    }

	/**
	 * Checks the correctness of the universe modifiers of an argument
	 * passed to a method as a parameter.
	 * 
	 * This is done by applying viewpoint adaptation and performing the
	 * necessary checks.
	 *
	 * @param parameterType The type of the method parameter, without
	 *                      adapted viewpoint
	 * @param argument The argument passed to this parameter
	 * @param methodTypeVariablesMap Mapping of method type variables
	 *                               to types
	 * @param context The context used to report errors
	 * @param parameterPosition Position of parameter for error reporting
	 * 
	 * by WMD
	 * by RZU
	 */
	private void typecheckUniverseArgument(CClassType parameterType,
			JExpression argument,
			Map methodTypeVariablesMap,
			CContextType context,
			int parameterPosition) throws PositionedError {
		CClassType substitutedParameterType = CUniverseServices
				.substituteInheritedTypeVariables(
						(CClassType) parameterType,
						(CClassType) prefix.getType(),
						context);
		
		// Adapt viewpoint of method parameters
		CClassType combinedType
				= CUniverseServices.combine(
						(CClassType) prefix.getType(),
						substitutedParameterType);

		// Perform the substitution
		CClassType substitutedType = getFullySubstitutedType(
				(CClassType) combinedType,
				methodTypeVariablesMap,
				context);

		// Ensure the argument is a valid argument for the parameter
		// type. Since we already know that the type itself is okay,
		// it can only be the universe modifier that disallow it
		if( method.isStatic() ) {
			check(context,
					argument.getType().isAlwaysAssignableTo(substitutedType),
					CUniverseMessages.ACTUAL_NOT_ASSIGNABLE_TO_FORMAL_PARAMETER_STATIC,
					new Object[] {new Integer(parameterPosition),
							argument.getType().toVerboseString(),
							method.ident(),
							substitutedType.toVerboseString()}
			);
		} else {
			check(context,
					argument.getType().isAlwaysAssignableTo(substitutedType),
					CUniverseMessages.ACTUAL_NOT_ASSIGNABLE_TO_FORMAL_PARAMETER,
					new Object[] {new Integer(parameterPosition),
							argument.getType().toVerboseString(),
							method.ident(),
							substitutedType.toVerboseString()}
			);
		}
	}

    /**
     * Substitutes the type variables of a method parameter.
     * 
     * @param type The type of the parameter to substitute
	 * @param methodTypeVariablesMap Mapping of method type variables
	 *                               to types
     * @param context The context used for type checking
     * @return The substituted type containing no type variables
     * 
     * by RZU
     */
    private CClassType getFullySubstitutedType(CClassType type,
			Map methodTypeVariablesMap,
    		CContextType context) throws PositionedError{
    	CClassType substitutedType = type;
    	
		if (type.isTypeVariable() && type.isMethodTypeVariable()) {			
			
			// Use mapping of method type variables if it is a
			// method type variable
			substitutedType = (CClassType) methodTypeVariablesMap
					.get(type.getIdent());
			
			// If there's a match, we have to check that it respects
			// the bounds
			if (substitutedType != null) {
				CTypeVariable typeVariable = (CTypeVariable) type;
		        for (int i = 0 ; i < typeVariable.getBounds().length; i++) {
		        	// Adapt viewpoint of bound
		        	CClassType combinedBound
		        			= CUniverseServices.combine(
		        					(CClassType) prefix.getType(),
		        					typeVariable.getBounds()[i]);
		        	
		        	check(context,
		        			substitutedType.isAlwaysAssignableTo(
		        					combinedBound),
							CUniverseMessages.METHOD_TYPE_VARIABLE_BOUND_VIOLATED,
    						new Object[] {substitutedType.toVerboseString(),
		        					typeVariable.getIdent()});
		        }
			}
		} else {

			// Get the host class
			CClass hostClass = context.getClassContext().getHostClass();

			// If the type is a type variable (T a), we use the receiver's type
			// arguments to replace it
			if (type.isTypeVariable()) {
				substitutedType = prefix.getType().getCClass()
						.getSubstitution((CTypeVariable) type,
								prefix.getType().getAllArguments());

				// If the type is an array whose type is a type variable (T[] a),
				// we use the receiver's type arguments to replace it
			} else if (type.isArrayType()
					&& ((CArrayType) type).getBaseType().isTypeVariable()) { 
				// Get substituted base type...
				substitutedType = prefix.getType().getCClass().getSubstitution(
						(CTypeVariable) ((CArrayType) type).getBaseType(),
								prefix.getType().getAllArguments());
				// ... and create an array out of it
				substitutedType = new CArrayType(substitutedType,
						((CArrayType) type).getArrayBound(),
						((CArrayType) type).getCUniverse());

				// If the type is an array with a generic base type (List<Integer>[] a),
				// we use the receiver's type arguments to replace type variables in
				// the type arguments, if any
			} else if (type.isArrayType()
					&& ((CArrayType) type).getBaseType().isGenericType()) {
				CType baseType = ((CArrayType) type).getBaseType();
				CClassType newType = ((CClassType) baseType).createSubstitutedType(
						hostClass,
						(CClassType) prefix.getType(),
						baseType.getAllArguments());
				substitutedType = new CArrayType(
						newType,
						((CArrayType) type).getArrayBound(),
						((CArrayType) type).getCUniverse());

				// If the type is an array with a generic base type (List<Integer>[] a),
				// we use the receiver's type arguments to replace type variables in
				// the type arguments, if any
			} else if (type.isGenericType()) {
				substitutedType = type.createSubstitutedType(
						hostClass,
						(CClassType) prefix.getType(),
						hostClass.getType().getAllArguments());
			}
		}
	    
	    return substitutedType;
	}


	// ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitMethodCallExpression( this );
    }

    /**
     * Generates a sequence of bytecodes.
     * <pre><jml>
     * also
     * public normal_behavior
     *   requires code != null;
     *   {|
     *     requires code.discardValue();
     *     assignable \not_specified;
     *     ensures (* after executing generated code JVM stack is unchanged *);
     *   also
     *     requires !code.discardValue();
     *     assignable \not_specified;
     *     ensures (* after executing generated code JVM stack contains 
     *		      result *);
     *   |}
     * </jml></pre>
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	boolean discardValue = code.discardValue();
	boolean targetIsExternal = topConcreteMethod.isExternal();

	code.setDiscardValue(false); // leave receiver or function ref on stack
	genPrefix( code );
	
	/*
	 * scdaniel: we need a copy for the setContext call below and
	 * we don't want to reevaluate the expression
	 * (ie. no side-effects please)
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     (setContext != null) &&
	     (tmpLocalVar != null) ) {

	    if ( repObject != null ) {
		if ( repTest != null ) {
		    tmpLocalVar.genStore(code);
		    repTest.genCode(code);
		} else {
		    repObject.genCode(code);
		    tmpLocalVar.genStore(code);
		}
	    } else
		tmpLocalVar.genStore(code);
	}

	// if this is an external gf call and we're using RMJ,
	// generate code for the args now, and store them in their
	// associated local vars
	if ( argLocalVars != null) {
	    genActualArgs(code);
	    genActualArgStores(code);
	}
	
	if( targetIsExternal ) {
	    genExternalTargetRef( code );
	}

	// if we stored the args in local variables earlier,
	// just load them on the stack now
	if ( argLocalVars != null) {
	    genActualArgLoads(code);
	} else {
	    genActualArgs(code);
	}

	/*
	 * scdaniel: universe runtime checks
	 * 
	 * Save the current object before a call to a static function
	 * from a non-static context.
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     (setContext != null) ) {
	    setContext.genCode(code);
	}

	genInvocationOpcode( code );

	if( discardValue ) {
	    if (getType().getSize() == 2) {
		code.plantInstruction(new NoArgInstruction(opc_pop2));
	    } else if (getType().getSize() == 1) {
		code.plantInstruction(new NoArgInstruction(opc_pop));
	    }
	}

	/*
	 * scdaniel: universe runtime checks
	 * 
	 * Reset the saved context (to have the correct context on top
	 * of the context stack again). Somewhere in the call hierarchy
	 * might be another static function call that needs the context
	 * on top after the reset.
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     (resetContext != null) ) {
	    resetContext.genCode(code);
	}
    }

    /**
     * Generates the bytecode for adding the prefix to the JVM
     * stack. */
    private void genPrefix( CodeSequence code ) {
	if( prefix != null ) {
	    /*
	     * scdaniel: universe runtime checks
	     * 
	     * we need a copy of the prefix if (tmpLocalVar != null) to push
	     * the reciever object onto the stack without re-evaluating
	     * the expression.
	     */
	    if ( method.access().isStatic() &&
		 !(CTopLevel.getCompiler().universeDynChecks() &&
		   (setContext != null) &&
		   (tmpLocalVar != null)) ) {
		code.setDiscardValue(true);
	    }
	    
	    prefix.genCode(code);
	}
    }
     

    /**
     * Plants the code for putting the args to the call on the stack.
     *
     */
    protected void genActualArgs( CodeSequence code ) {
	for (int i = 0; i < args.length; i++) {
	    code.setDiscardValue(false);
	    args[i].genCode(code);
	}
    }

     
    /**
     * Plants the code for storing the actual args to local variables.
     * Used by RMJ for calls to external generic functions.
     * Assumes the actual args are currently on the stack, with the last
     * one on top.
     *
     */
    protected void genActualArgStores( CodeSequence code ) {
	for (int i = args.length - 1; i >= 0; i--) {
	    code.plantLocalVar(argLocalVars[i].getType().getStoreOpcode(),
			       argLocalVars[i]);
	}
    }

     
    /**
     * Plants the code for loading the actual args from their local vars.
     * Used by RMJ for calls to external generic functions.
     *
     */
    protected void genActualArgLoads( CodeSequence code ) {
	for (int i = 0; i < argLocalVars.length; i++) {
	    code.plantLocalVar(argLocalVars[i].getType().getLoadOpcode(),
			       argLocalVars[i]);
	}
    }

    /**
     * Plants the code for putting a reference to the appropriate
     * dispatcher in the chain-of-responsibility onto the stack.
     *
     * <pre><jml>
     * requires topConcreteMethod.isExternal();
     * requires (* code already contains instructions for leaving any
     *             necessary prefix object on the stack *);
     * </jml></pre>
     *
     */
    protected void genExternalTargetRef( CodeSequence code ) {
	if (skipLocalFunction) {
	    // get function reference from local method
	    localMethod.plantOldFunctionRef( code, true, !isCase_s_e2 );
	} else {
	    // get function reference from top method because it
	    // is inside the anchor class of the GF
	    topConcreteMethod.plantFunctionRef( code );
	}
	// now swap the receiver and the dispatcher function so
	// they are in the correct order on the stack.  it's necessary
	// to generate them in the wrong order so that all relevant
	// dispatchers will be added to the chain before loading the
	// dispatcher function
	code.plantSwap();
    }

    /**
     * Generates the bytecode for invoking the actual target method.
     * Assumes code sequence already contains code for pushing all the
     * necessary arguments onto the stack.  This is a helper method of
     * {@link #genCode(CodeSequence)}.
     */
    protected void genInvocationOpcode( CodeSequence code ) {
	boolean forceNonVirtual = prefix instanceof JSuperExpression
	    && !(method instanceof CWrapperMethod);
        method.genGenFuncInvocation( code, forceNonVirtual );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    //@ spec_public
    protected JExpression	prefix;
    protected String		ident;
    protected JExpression[]	args;
    private CType type;
    private CType cachedType;
    /**
     * The method invoked by this method call, as determined by the
     * static receiver and argument types.  Dynamic dispatch may cause
     * an overriding method to be invoked at runtime.
     */
    //@ spec_public
    protected CMethod method;

    /**
     * The top concrete method of the generic function invoked by this
     * method call.
     */
    //@ spec_public
    protected CMethod topConcreteMethod;

    /**
     * Tracks whether invocation is of a super class method that
     * belongs to an external generic function (Clifton-01, Figure 24,
     * cases s-i3, s-e2, or s-e3) where a local method of the target
     * generic function must be skipped in the
     * chain-of-responsibility.
     */
    //@ spec_public
    private boolean skipLocalFunction = false;

    /**
     * Tracks whether the invocation is of a super class method that
     * belongs to the same external generic function as the caller but
     * is declared in a different compilation unit (Clifton-01, Figure
     * 24, case s-e2).
     */
    //@ spec_public
    private boolean isCase_s_e2 = false;

    /**
     * A local method of the generic function targeted by this invocation.
     *
     * <pre><jml>
     * private invariant skipLocalFunction ==> localMethod != null;
     * </jml></pre>
     *
     */
    //@ spec_public
    private CMethod localMethod;

    /** the name of this field reference as originally appeared in
     * the source file. It may be null if generated by the system. */
    private JNameExpression sourceName;

    /** an array of local variables used as temps for the args of the call.
     *  necessary if the call is external and we're using RMJ. */
    private JLocalVariable[] argLocalVars;
    
    // scdaniel: universe runtime checks
    private JMethodCallExpression setContext, resetContext;
    private JLocalVariable tmpLocalVar;
    private boolean doSetContext = true;
    private JExpression repObject;
    private JIfStatement repTest;
}
