/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMethodDeclaration.java,v 1.76 2008/05/14 18:16:24 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a java method in the syntax tree.  This
 * classes natural ordering is inconsistent with equals.  It is
 * consistent with sameGenericFunction.  */
public class JMethodDeclaration extends JMemberDeclaration 
    implements JMethodDeclarationType
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	modifiers	list of modifiers
     * @param   typevariables   the type variable of this method
     * @param	returnType	the return type of this method
     * @param	ident		the name of this method
     * @param	parameters	the parameters of this method
     * @param	exceptions	the exceptions throw by this method
     * @param	body		the body of this method
     * @param	javadoc		javadoc comments, including whether this 
     *				method is deprecated
     * @param	comments	regular java style comments
     */
    public JMethodDeclaration(TokenReference where,
			      long modifiers, 
			      CTypeVariable[] typevariables,
			      CType returnType,
			      String ident,
			      JFormalParameter[] parameters,
			      CClassType[] exceptions,
			      JBlock body,
			      JavadocComment javadoc,
			      JavaStyleComment[] comments ) {
	super(where, javadoc, comments);

	this.modifiers = modifiers; 
	this.typevariables = typevariables;

	// WMD
    // We have to replace the this modifier by peer since it is not
    // appropriate for method parameters
    this.returnType = CUniverseServices.setDefaultVariable(returnType);

	this.ident = ident.intern();
	this.body = body;

	// scdaniel: universe runtime checks
	if ( CTopLevel.getCompiler() != null && 
	     CTopLevel.getCompiler().universeDynChecks() &&
	     ident.equals(JAV_STATIC_INIT) &&
	     this.body != null )
	    this.body.setGenUniverseDynChecks( true );

	this.parameters = parameters;
	this.exceptions = exceptions;
	
	assertTrue(parameters != null);
	assertTrue(exceptions != null);
	
	for(int i=0; i<exceptions.length; ++i) {
		this.exceptions[i] = (CClassType) CUniverseServices.setDefaultReadonly(exceptions[i]);
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------
    public /*@ pure @*/ JFormalParameter[] parameters() {
	return parameters;
    }

    public void setParameters( JFormalParameter[] parameters ) {
	this.parameters = parameters;
    }


    /**
     * Adds an additional formal parameter to this method, appending
     * it to the end of the existing parameter list.  
     *
     * <pre><jml>
     * also requires param != null;
     * </jml></pre>
     */
    public void addParameter( JFormalParameter param ) {
	JFormalParameter[] newParms;
	if (parameters == null) {
	    newParms = new JFormalParameter[1];
	} else {
	    newParms = new JFormalParameter[parameters.length + 1];
	}
	System.arraycopy( parameters, 0, newParms, 1, parameters.length );
	newParms[0] = param;
	parameters = newParms;
    }

    public /*@ pure @*/ String ident() {
	return ident;
    }

    public /*@ pure @*/ CType returnType() {
	return returnType;
    }

    public /*@ pure @*/ CClassType[] getExceptions() {
	return exceptions;
    }

    public /*@ pure @*/ long modifiers() { 
	return modifiers;
    }

    /** Returns the type variables of this method. */
    public /*@ pure @*/ CTypeVariable[] typevariables() {
        return typevariables;
    }

    /** Sets the modifiers of this method declaration */
    public void setModifiers(long modifiers) { 
	this.modifiers = modifiers;
    }

    public /*@ pure @*/ JBlock body() { 
	return body; 
    }

    public /*@ pure @*/ boolean hasBody() { 
	return this.body != null;
    }

    public void setIdent(String name) {
	this.ident = name;
    }

    public void setBody(JBlock body) {
	this.body = body;
    }

    public /*@pure*/ boolean isDeclaredNonNull() {
	return hasFlag(modifiers, ACC_NON_NULL | IMPLICITLY_NON_NULL);
    }

    /**
     * Set the nullity status of this declarator.
     */
    public void setNonNull() {
	// if(!returnType.isNonNull())
	//    returnType = ((CClassType) returnType).getNonNullType();

	modifiers |= IMPLICITLY_NON_NULL;
	// Also adjust the signature of this method.
	if(getMethod() != null) {
	    getMethod().addModifiers(IMPLICITLY_NON_NULL);
	}
    }

    /** Return <code>true</code> if this method declaration overrides
     * any of its superclass (or interfaces) method declarations.
     */
    public /*@ pure @*/ boolean isOverriding() {
	/*@ nullable @*/ CMethodSet methodSet = overriddenMethods();
	return methodSet != null && !methodSet.isEmpty();
    }    

    /** Return the set of methods that are directly overriden (specialized)
     * by this method declaration.
     */
    public /*@ pure nullable @*/ CMethodSet overriddenMethods() {
	/*@ nullable @*/ CSourceMethod meth = (CSourceMethod) getMethod();
	return meth == null ? null : meth.overriddenMethodSet();
    }   

    /**
     * Indicates whether this method uses multiple dispatch
     *
     * @return	true iff a parameter of this method has a dynamic dispatch
     *		annotation or, for an external method, this methods receiver
     *		is of a different type than the generic function's top method
     */
    public boolean usesMultipleDispatch() {
	for( int i=0; i<parameters.length; i++ ) {
	    if( parameters[i].isSpecialized() ) {
		return true;
	    }
	}
	// if control flow reaches here then no parameters are specialized
	return false;
    }

    /**
     * Compares this method to a given method and returns 0 if the
     * methods belong to the same generic function, otherwise returns
     * -1 or +1 to sort the methods.  This comparator is used to
     * separate methods by generic function in the
     * MJGenericFunctionDispatcher class.
     *
     * @param	o	the object to be compared against, must be a
     *			JMethodDeclarationType
     * @return	-1, +1, or 0
     * @exception ClassCastException if <code>o</code> is not an instance of
     *					CType */
    public int compareTo(/*@ non_null @*/ Object o) throws ClassCastException {
	// must override in MJTopLevelMethodDeclaration subclass to
	// consider receiver static types
	JMethodDeclarationType m2 = (JMethodDeclarationType) o;

	JFormalParameter[] p2 = m2.parameters();
	    
	    // first compare argument counts
	int result = parameters.length - p2.length;
	if( result != 0 ) {
	    return result;
	}

	// next compare identifiers
	result = ident.compareTo( m2.ident() );
	if( result != 0 ) {
	    return result;
	}

	// if control flow reaches here then the receivers,
	// identifiers, and argument counts match so we check the
	// remaining arguments in lexicographic order
	for( int i=0; i<parameters.length; i++ ) {
	    CType t1 = parameters[i].staticType();
	    CType t2 = p2[i].staticType();
	    if( !(t1.equals( t2 )) ) {
		return t1.compareTo( t2 );
	    }
	}

	// if control flow reaches here then everything matches
	return 0;
    }

    /**
     * Indicates whether this member is external.
     * Always returns false for this but is overridden for external
     * subclasses.
     * @return	a flag indicating whether this member is external
     */
    public /*@ pure @*/ boolean isExternal() {
	return false;
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things
     * generally look OK.  This pass gathers information about
     * the type signatures of everything (imported class
     * files, classes being compiled, methods, fields, etc...)
     * needed for the later passes.  This information is stored 
     * in a context hierarchy that is bound to the AST.
     *
     * @return	a source code representation of the method, iff sub tree 
     *		is correct enough to check code
     * @exception	PositionedError	an error with reference to the 
     *					source file
     */
    public CMember checkInterface( CContextType context ) 
	throws PositionedError 
    {
	
	String	ident = (this instanceof JConstructorDeclarationType) ? 
	    JAV_CONSTRUCTOR : this.ident;

	check( context,
	       !(context instanceof CExtMethodContext),
	       MjcMessages.REGULAR_METHOD_MISPLACED,
	       ident );	

	CClass owner = context.getClassContext().getOwnerClass();
	try{ 
	    for (int i = 0; i < typevariables.length; i++) {
	    	typevariables[i].setMethodTypeVariable(true);		
	    	typevariables[i] = (CTypeVariable)(typevariables[i].checkType(context));
	    }
	}catch (UnpositionedError cue) {
	    throw cue.addPosition(getTokenReference());
	}    
	CContextType self = createExtendedClassContext(context,typevariables);
	MemberAccess access = makeMemberAccess( self, owner );

	// check the modifier combinations common to all declarations
        access.checkAccessModifiers( self, this );

	modifiers = access.checkInterfaceMethodModifiers(self, this);

	return checkInterfaceType( self, access, ident );
    }

    /**
     * Performs the interface checks that are common to all sorts of
     * methods. Gets the signatures of the return type, parameters and
     * exceptions, and generates the source code representation,
     * CSourceMethod.  Registers that the generic function <code>ident</code>
     * might be extended by this in the given <code>context</code>.
     *
     * @param	context		the context in which this method appears
     * @param	access		the <code>MemberAccess</code> object for this method
     * @param	ident		the method name (passed as a parameter 
     *				instead of using the field to properly 
     *				handle constructors where the field is 
     *				the class name but ident is &lt;init&gt;
     * @return	a source code representation of the method, iff sub tree 
     *		is correct enough to check code
     * @exception		PositionedError, an error with reference to the 
     * source file 
     */
    public CSourceMethod checkInterfaceType( CContextType context,
					     MemberAccess access, 
					     String ident) 
	throws PositionedError
    {
	try {
	    // 8.4.3 Method Modifiers
	    access.checkMethodModifiers(context, this);

	    // resolve the return type
	    returnType = returnType.checkType( context );
	    returnType = CUniverseServices.setDefaultVariableAgain(returnType);
	    
	    // get the static and dynamic types of the parameters
	    CSpecializedType[] parameterTypes = 
		new CSpecializedType[parameters.length];
	    for (int i = 0; i < parameters.length; i++) {
		parameterTypes[i] = parameters[i].checkInterface( context );
	    }

	    if( this instanceof JConstructorDeclarationType
		|| access.isAbstract() ) {

		boolean isSpecialized = false;
		
		for( int i=0; i < parameters.length; i++ ) {
		    if( parameters[i].isSpecialized() ) {
			isSpecialized = true;
			break;
		    }			   
		}
		check( context, 
		       !(this instanceof JConstructorDeclarationType 
			 && isSpecialized),
		       MjcMessages.SPECIALIZED_CONSTRUCTOR );
		
		//  !HERE!note6
		check( context, 
		       !(access.isAbstract() && isSpecialized),
		       MjcMessages.SPECIALIZED_ABSTRACT );

	    }

	    for (int i = 0; i < exceptions.length; i++) {
		exceptions[i] = (CClassType)(exceptions[i].checkType(context));

	    }

	    CSourceMethod method = makeMethodSignature( context, access, ident, 
							parameterTypes );
	    setInterface( method );

	    return method;
	} catch (UnpositionedError cue) {
	    throw cue.addPosition(getTokenReference());
	}
    }

     /**
     * Create a Extended Class context that include the typevariables
     */
    public CExtendedClassContext createExtendedClassContext(CContextType parent,
							    CTypeVariable[] typevariables) {
	return new CExtendedClassContext(parent,typevariables);
    }
	
    /**
     * Generates the signature object for this method declaration. 
     */
    protected CSourceMethod makeMethodSignature( CContextType context,
						 MemberAccess access, 
						 String ident, 
					    CSpecializedType[] parameterTypes ) 
    {
	return 
	    new CSourceMethod(access, ident, returnType, parameterTypes, 
			      exceptions, typevariables, isDeprecated(), 
			      body, context, this);
    }

    protected MemberAccess makeMemberAccess( CContextType context,
					     CClass owner )
    {
	return( new MemberAccess( owner, 
				  context.findNearestHost(), 
				  modifiers ) ); 
    }
    
    /**
     * Computes the values of specializer expressions used to dispatch on
     * compile-time constants.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    public void resolveSpecializers(CContextType context)
	throws PositionedError {
	CMethodContextType selfC = 
	    createSelfContext( (CClassContextType) context );

	CFlowControlContextType inside = 
	    selfC.createFlowControlContext(parameters.length,
					   getTokenReference());

	for( int i = 0; i < parameters.length; i++ ) {
	    parameters[i].resolveSpecializer( inside );
	}
	
    }

    /**
     * Makes sure that all in-scope external generic functions are
     * added to the appropriate augmentation maps before top method
     * searches occur in later passes.  */
    public void resolveExtMethods( CContextType context ) {
	context.resolveMaybeExtMethodRef( ident );
    }

    /**
     * Finds the top method of every declared method.  Called after
     * the resolveSpecializers pass.  This cannot be done before then
     * because the external generic function mappings are not complete
     * until the end of the checkInterface pass and the constant value
     * specializers are not known until after the resolveSpecializers
     * pass.  This must be done before the typecheck pass so that all
     * specialized argument positions for generic functions are known
     * for ambiguity checking. */
    public void resolveTopMethods() throws PositionedError {
	try {
	    getMethod().topConcreteMethod();
	} catch (UnpositionedError e) {
	    throw e.addPosition(getTokenReference());
	}
    }

	/**
	 * Checks static correctness of Generic Universe Types.
	 * 
	 * Checks performed:
	 * - Parameters of pure methods must be readonly
	 * - Parameters and return type of static methods must not contain
	 *   readonly modifiers
	 * 
	 * @param context The current typechecking context.
	 * 
	 * by WMD
	 * by RZU
	 */
    private void typecheckUniverses(CContextType context)
    		throws PositionedError {
    	CMethod self = getMethod();
    	
    	// If we're dealing with a pure method that is not a constructor,
    	// we have to ensure that all parameters are readonly
    	// Note that for pure constructors, we also allow peer and rep references
    	if (self.isPure()
    			&& !(this instanceof JConstructorDeclarationType)) {
    		for (int i = 0; i < parameters.length; i++) {
    			if (parameters[i].getType() instanceof CClassType) {
    				CClassType currentParameterType
    						= (CClassType) parameters[i].getType();
    				
    				// Combine it with any to see if anything changes
    				CClassType adaptedType
    						= CUniverseServices
    								.combine(CUniverseReadonly.getUniverse(),
    										currentParameterType);
    				
    				// If the two are not equal, and the main modifier wasn't
    				// implicit, we complain
    				check(context,
    						currentParameterType.getCUniverse()
    								instanceof CUniverseImplicitPeer
    						|| currentParameterType.equals(adaptedType),
    						CUniverseMessages.PURE_METHOD_NON_READONLY_PARAMETER_ERROR,
    						new Object[] {parameters[i].ident(),
    								currentParameterType.toVerboseString()});
    			
    				// Update parameter type
    				parameters[i].setType(adaptedType);
    				
    				// If it was implicit, set it to implicit again
    				if (currentParameterType.getCUniverse()
    								instanceof CUniverseImplicitPeer) {
    					adaptedType.setCUniverse(CUniverseImplicitReadonly
    							.getUniverse());
    					warn(context,
    							CUniverseMessages.PURE_METHOD_NON_READONLY_PARAMETER_CHANGED,
    							new Object[] {parameters[i].ident(),
    									self.ident(),
    									currentParameterType.toVerboseString(),
    									adaptedType.toVerboseString() });
    				}
    			}
    		}
    	}

    	// Warn if purity changes from non-pure to pure
    	if (self.purityWasChanged()) {	    
    		warn(context,
    				CUniverseMessages.OVERRIDING_PURE_METHOD_NON_PURE_CHANGED,
    				ident());
    	}

    	// Rep modifiers are neither allowed in parameter types nor in the return
    	// type of static methods

		if( parameters!=null )
		for(int i=0; i<parameters.length; ++i) {
			if( (parameters[i].getType() instanceof CClassType)
		    		&& CUniverseServices
		    				.containsRepModifier((CClassType) parameters[i].getType())) {

				if(self.isStatic()) {
					check(context,
						false,
						CUniverseMessages.REP_IN_STATIC_FORBIDDEN);
				} else {
					warn( context,
						CUniverseMessages.METHOD_REP_IN_SIGNATURE_WARNING,
						self.ident(),
						parameters[i].ident());
				}
			}
		}

		if (self.isStatic()) {
		    check(context,
					!(returnType instanceof CClassType)
					|| !CUniverseServices
							.containsRepModifier(
									(CClassType) returnType),
					CUniverseMessages.REP_IN_STATIC_FORBIDDEN);
		}
    }


    /**
     * Check whether the method signature of the current method is consistent
     * with the supermetod.
     * A public method now, in order for JmlMethodDeclaration to call it on
     * its delegee. Otherwise not meant for the public.
     * 
     * By WMD.
     */
    public void checkUniverseOverriding( CContextType context,
					  JFormalParameter[] thisparams,
					  CMethod superMethod )
	throws PositionedError
    {
	CMethod self = getMethod();

	CSpecializedType[] smspec = superMethod.parameters();
	CType[] smparams;
        org.multijava.util.MessageDescription msg;
        
	smparams = new CType[ smspec.length ];
	for( int spc = 0; spc < smparams.length; ++spc ) {
	    smparams[spc] = smspec[spc].staticType();
	}

	for (int i = 0; i < thisparams.length; i++) {
	    if( !(thisparams[i].getType().equals(smparams[i]) ||
		  smparams[i].equals(thisparams[i].getType()) ) ) {
                
                // The parameter types in the supermethod and the current method
                // are not equal... look whether we can do something about it.
                
                if( (smparams[i] instanceof CClassType &&
                    CUniverseServices.isPartial((CClassType)smparams[i])) ||
                    smparams[i].isPrimitive() ) {
                    // the super signature is also not complete or primitive...
                    // let's just ignore the parameter.
                    msg = null;
                } else if( thisparams[i].getType() instanceof CClassType &&
			!self.isPure() &&
		    CUniverseServices.isPartial((CClassType)(thisparams[i].getType())) ) {
		    // the parameter did not have a modifier
		    // -> if the super-parameter only uses readonly
                    // we can take the super-parameter. Otherwise give error. 
                    CClassType smany = CUniverseServices.combine(CUniverseReadonly.getUniverse(), (CClassType)smparams[i]);
                    if( smany.equals(smparams[i])) {
                        thisparams[i].setType(CUniverseServices.propagate((CClassType) (thisparams[i].getType()), (CClassType) smparams[i]));
                        msg = CUniverseMessages.OVERRIDING_METHOD_UNIVERSE_CHANGED;
                    } else {
                        msg = CUniverseMessages.OVERRIDING_METHOD_UNIVERSE_ERROR;
                    }
		} else if( (thisparams[i].getType()) instanceof CClassType &&
			   ((CClassType)(thisparams[i].getType())).getCUniverse() instanceof CUniverseImplicitReadonly ) {
		    // the parameter was changed, because the method is pure
		    // but maybe the super-method was not changed yet.
		    // ignore.
		    msg = null;
		} else {
		    // there was an explicit modifier present
		    // -> error message
		    msg = CUniverseMessages.OVERRIDING_METHOD_UNIVERSE_ERROR;
		}

                if (msg != null) {
                    warn(context,
			 msg,
			 new Object[] { superMethod.getMethod(),
					parameters[i].ident(), // new Integer(i+1),
					smparams[i].toVerboseString(),
				    thisparams[i].getType().toVerboseString() }
			 );
                }
            }
        }
    }
    




    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Creates a context for this method declaration AST node.
     *
     * @param	parent		the parent context
     */
    public CMethodContextType createSelfContext(CClassContextType parent) {
	return parent.createMethodContext(getMethod());
    }
    
    protected boolean noBodyOK(CContextType context, CMethod self) {
	CClass receiver = self.owner();
	boolean b = self.isAbstract() || self.isNative() 
		       || receiver.isInterface();
	return b || Main.contextBehavior().noBodyOK(context) ;
    }
    
    /**
     * Typechecks this method declaration.  Mutates the context to
     * record the information gathered during the checks.
     *
     * @param context the context in which this method appears
     * @exception PositionedError if the checks fail and the failure
     * cannot be recovered from
     */
    public void typecheck( CContextType context ) 
	throws PositionedError 
    {
	CMemberHost host = context.findNearestHost();
	CSourceMethod self = (CSourceMethod) getMethod();
	CClass receiver = self.owner();

	try {
	    if( self.isAbstract() ) {
		// Won't get here for external methods because they
		// cannot be abstract (by R1).  This fact is verified
		// in checkInterface for MJTopLevelMethodDeclaration
		check( context, receiver.isAbstract(),
		       MjcMessages.METHOD_ABSTRACT_CLASSNOT, ident );
	    }
	    Debug.breakMe();

	    CClassType[] argss = receiver.getType().getArguments();
	    ArrayList needbridges = new ArrayList();
	    CMethodSet superMethods = self.setOverriddenMethodSet(argss, 
								  needbridges,
								  context );

	    // WMD: additional Universe checks
	    if( CTopLevel.getCompiler().universeChecks() ) {
		typecheckUniverses( context );
	    }

	    if( !(superMethods.isEmpty()) ) {
		checkOverriding( context, superMethods );
	    }

	    // Checks whether this is an overload that could be a
	    // multimethod:
	    if (context.getCompiler().options.multijava() ||
		CTopLevel.getCompiler().universeChecks()) {
		CMethod overloaded = 
		    receiver.lookupOverloadedMethod(self, 
						    CClassType.EMPTY,
						    context);
		if (overloaded != null) {

		    if( CTopLevel.getCompiler().universeChecks() ) {
			if( self.hasSameSignatureNoUniverses( overloaded ) ) {
			    // WMD: check parameter types and make sure that
			    // the Universe annotation was not changed
			    fail( context,
				  CUniverseMessages.OVERLOADING_METHOD_UNIVERSE_ERROR,
				  new Object[] { self.ident(),
						     overloaded,
						     self } );
			}
		    } else if (context.getCompiler().options.multijava()) {
			warn(context, MjcMessages.OVERLOAD_COULD_BE_OVERRIDE,
			     overloaded);
		    } 
		}
	    }

	    for (int i =0; i < needbridges.size(); i++) {
		addBridge(context, (CMethod)needbridges.get(i));
	    }
	    
	    if( body != null ) {

		check( context,
                       noBodyOK(context,self)
		       || !receiver.isInterface() 
		       || (this instanceof JInitializerDeclaration)
		       || self.isExternal(),
		       MjcMessages.METHOD_BODY_IN_INTERFACE, 
		       ident );

		check( context,
		       !self.isAbstract() && !self.isNative() ,
		       MjcMessages.METHOD_BODY_NATIVE_ABSTRACT, 
		       ident );

		CMethodContextType selfC = 
		    createSelfContext( (CClassContextType) context );
		CFlowControlContextType inside = 
		    selfC.createFlowControlContext(parameters.length,
						   self.isInExternalGF(),
						   getTokenReference());

		if( !self.isStatic() ) {
		    // add this local var
		    inside.addThisVariable();
		}

		for( int i = 0; i < parameters.length; i++ ) {
		    parameters[i].typecheck( inside );
		}
		

		body.typecheck( inside );

		inside.checkingComplete();
		selfC.verifyExceptions( getTokenReference() );

		if( inside.isReachable() 
		    && self.returnType() != CStdType.Void ) {
		    context.reportTrouble( 
			      new CLineError( getTokenReference(),
					      MjcMessages.METHOD_NEED_RETURN, 
					      self.ident() ) );
		}
	    } else {
		// body is null
		check( context,
                       noBodyOK(context,self),
		       MjcMessages.METHOD_NOBODY_NOABSTRACT, 
		       ident );
	    }

	    context.registerVisibleMethod( self );
	} catch( UnpositionedError ce ) {
	    context.reportTrouble( ce.addPosition( getTokenReference() ));
	} catch( PositionedError pe ) {
	    context.reportTrouble( pe );
	}
    }

    public void addBridge(CContextType context, CMethod method)
	throws PositionedError{
	CClass local = context.getClassContext().getOwnerClass();
	CType bridgeType = null;
	JBlock bridgeBody = null;
	bridgeType = method.returnType();
	if(bridgeType.isTypeVariable()){
	    try{
		bridgeType = bridgeType.getErasure(context);
	    }catch (UnpositionedError e) {
		e.addPosition(getTokenReference());
	    }
	}
	JExpression[] args = new JExpression[parameters.length];
	CSpecializedType[] superParams = method.parameters();
	JFormalParameter[] bridgeParameter =
	    new JFormalParameter[parameters.length];
	for (int i = 0; i < args.length; i++) {
	    CType             btType = null;
	    
	    try {
		btType = superParams[i].staticType().getErasure(context);
	    } catch (UnpositionedError e) {
		throw e.addPosition(getTokenReference());
	    }
	    
	    bridgeParameter[i] = new JFormalParameter(getTokenReference(), 
						      JLocalVariable.DES_LOCAL_VAR, 
						      new CSpecializedType(btType),
						      parameters[i].ident(), 
						      parameters[i].isFinal());
	    
	    if ((parameters[i].getType().isClassType()) 
		&& (parameters[i].getType().getCClass() != superParams[i].staticType().getCClass())) {
		try {
		    args[i] = new 
			JCastExpression(getTokenReference(),
					new JNameExpression(getTokenReference(),
							    parameters[i].ident()), 
					parameters[i].getType().getErasure(context));
		} catch (UnpositionedError e) {
		    throw e.addPosition(getTokenReference());
		}
	    } else {
		args[i] = new JNameExpression(getTokenReference(), 
					      parameters[i].ident());
	    }
	}
	
	JExpression         callRealMethod = new JMethodCallExpression(getTokenReference(),
								       new JThisExpression(getTokenReference()),
								       ident, 
								       args);
	bridgeBody = new JBlock(getTokenReference(),
				new JStatement[] {
				    (bridgeType.getTypeID() == TID_VOID) ? 
				    (JStatement)new JExpressionStatement(getTokenReference(), 
									 callRealMethod,
									 null) 
				    :  new JReturnStatement(getTokenReference(), 
							    callRealMethod,
							    null)
				},
				null);
	
	JMethodDeclaration  bridgeMethod = new JMethodDeclaration(getTokenReference(),
								  ((modifiers | Constants.ACC_ABSTRACT) ^ Constants.ACC_ABSTRACT),
								  CTypeVariable.EMPTY,
								  bridgeType, 
								  ident, 
								  bridgeParameter, 
								  exceptions, 
								  bridgeBody, 
								  null,
								  null) {
		protected void checkOverriding(CClassContext contxt) 
		    throws PositionedError {
		    // nothing to do!!
		}
	    };
	CSourceMethod            bridge = (CSourceMethod)bridgeMethod.checkInterface(context);
	bridge.setGenerateFrom(getMethod());
	bridge.setSynthetic(true);
	try {
	    ((CSourceClass)local).addBridgeMethod(bridge,context);
	}
	catch (UnpositionedError ce) {
	     context.reportTrouble( ce.addPosition( getTokenReference() ));
	}
	//bridgeMethod.typecheck(context);
    }

    /**
     * Checks that this method appropriately overrides the given
     * superclass methods.  The checks are those given in JLS2,
     * section 8.4.6.3 and CLCM 00 4.1.3.1 (extended per Clifton's
     * thesis).
     *
     * @param	context		the context in which this appears
     * @param	superMethods	the super type methods that this
     *                          <em>may</em> override.
     * @exception PositionedError if a check fails
     */
    public void checkOverriding( CContextType context, CMethodSet superMethods ) 
	throws PositionedError 
    {
	try {
	    CMethod self = getMethod();
	    CMethod topConcreteMethod = self.topConcreteMethod();

		// check that the current method does not have an interface
		// for a receiver, if the method is part of an external gf.
		// only top methods of external gfs may do so.
		//x@ assert self != topConcreteMethod
	    if (context.getCompiler().RMJ()) {
		/* RMJ: changed this to a warning from a check */
		if (self.isInExternalGF() && self.owner().isInterface()) {
		    warn( context,
			  MjcMessages.RMJ_ENHANCES_INTERFACE,
			  ident() );
		}
	    } else {
		check( context,
		       !(self.isInExternalGF() && self.owner().isInterface()),
		       MjcMessages.ENHANCES_INTERFACE,
		       ident() );
	    }

	    int size = superMethods.size();
	    boolean inExternalGF = topConcreteMethod.isExternal();	    
	    for( int i=0; i<size; i++ ) {

		CMethod superMethod = superMethods.getMethod( i );

		// return types must match
		if(!context.getCompiler().Generic()){
		    check( context,
			   returnType.equals( superMethod.returnType() ),
			   MjcMessages.METHOD_RETURN_DIFFERENT, ident);
		} else {
		    if(returnType.isClassType()
		       && superMethod.returnType().isClassType()) {
			check(context,
			       returnType.isAlwaysAssignableTo(superMethod.returnType(),
			    		   true),
			       MjcMessages.METHOD_RETURN_NOT_SUBTYPE,
			       ident);
		    } else {
			check(context,
			      returnType.equals(superMethod.returnType() ),
			      MjcMessages.METHOD_RETURN_DIFFERENT, ident);
		    }
		}
		    // it's an error if one is static and the other is not
		if (self.isStatic()) {
		    check( context,
			   superMethod.isStatic(),
			   MjcMessages.METHOD_HIDE_NON_STATIC, ident,
			   superMethod);
		} else {
		    check( context,
			   !(superMethod.isStatic()),
			   MjcMessages.METHOD_OVERRIDE_STATIC, ident,
			   superMethod);
		}

		// WMD: additional Universe checks
		if( CTopLevel.getCompiler().universeChecks() ) {
		    checkUniverseOverriding( context, parameters, superMethod );
		}

		if (context.getCompiler().RMJ()) {
		    // RMJ: allow glue methods (methods that override
		    // non-local external GFs).
		    check( context, 
			   !isExternal() || superMethod.isInExternalGF(),
			   MjcMessages.RMJ_EXT_MM_BAD_OVERRIDE, ident, 
			   superMethod );
		    if(isExternal() && !superMethod.isLocalTo( context )) {
			Object[] parms = { ident, superMethod };
			warn(context, 
			     MjcMessages.EXT_METHOD_OVERRIDES_NON_LOCAL,
			     parms);
		    }
		} else {
		    // restriction R3
		    if (isExternal()) {
				check( context, 
					   superMethod.isLocalTo( context ),
					   MjcMessages.EXT_MM_BAD_OVERRIDE, ident, 
					   superMethod );
				check( context, 
					   superMethod.isExternal(),
					   MjcMessages.EXT_MM_BAD_OVERRIDE2, ident, 
					   superMethod );
			}
		}
		
		if( inExternalGF && !superMethod.isInExternalGF() ) {
		    // this method is pleomorphic, overriding internal
		    // and external methods of different generic
		    // functions
		    self.setIsMixedPleomorphic();
		}
		
		// JLS 8.4.3.3 A method can be declared final to prevent
		// subclasses from overriding or hiding it.  It is a
		// compile-time error to attempt to override or hide a
		// final method.
		check( context, 
		       !superMethod.isFinal(), 
		       MjcMessages.METHOD_OVERRIDE_FINAL, ident, superMethod );

		// privilege access modifiers cannot be made more restrictive
		// JLS2, p. 179
		// recall that (A => B), i.e., A implies B, is equivalent to 
		// (B || !A)
	    
		// superMethod.isPublic() => self.isPublic()
		check( context,
		       self.isPublic() || !superMethod.isPublic(),
		       MjcMessages.METHOD_ACCESS_DIFFERENT, ident );

		// superMethod.isProtected() => 
		// (self.isPublic() || self.isProtected())
		check( context,
		       (self.isPublic() || self.isProtected()) 
		       || !superMethod.isProtected(),
		       MjcMessages.METHOD_ACCESS_DIFFERENT, ident );

		// superMethod.hasDefaultAccess() => !(self.isPrivate())
		check( context,
		       !self.isPrivate() || !superMethod.hasDefaultAccess(),
		       MjcMessages.METHOD_ACCESS_DIFFERENT, ident );
      

		// JLS2 8.4.4 A method that overrides or hides another
		// method (JLS2 8.4.6), including methods that
		// implement abstract methods defined in interfaces,
		// may not be declared to throw more CHECKED
		// exceptions than the overridden or hidden method.
	
		CClassType[] exc = superMethod.throwables();
	
	    loop:
		for (int j = 0; j < exceptions.length; j++) {
		    if (exceptions[j].isCheckedException()) {
			for (int k = 0; k < exc.length; k++) {
			    if (exceptions[j].isAlwaysAssignableTo(exc[k])) {
				continue loop;
			    }
			}
			Object[] params = new Object[3];
			params[0] = ident;
			params[1] = exceptions[j];
			params[2] = superMethod.toString();
			check( context, false, 
			       MjcMessages.METHOD_THROWS_DIFFERENT, 
			       params
			   );
		    }
		}		
		
	    }	

	} catch( UnpositionedError cue ) {
	    throw cue.addPosition( getTokenReference() );
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitMethodDeclaration( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	fail(); // nothing to do here
    }

    //---------------------------------------------------------------------
    // USEFUL FOR DEBUGGING
    //---------------------------------------------------------------------
    public String toString() {
	return "JMethodDeclaration for " + ident;
    }


    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private long		modifiers; 
    private CType		returnType;
    private String		ident;
    private JFormalParameter[]	parameters;
    private CClassType[]	exceptions;
    private JBlock		body;
    protected CTypeVariable[]   typevariables;
}
