/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMethodDeclarationType.java,v 1.19 2007/02/08 14:04:54 leavens Exp $
 */

package org.multijava.mjc;

import java.util.HashSet;
import java.util.Arrays;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This type represents a java method in the syntax tree.  */
public interface JMethodDeclarationType extends JMemberDeclarationType,
						Comparable
{
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------
    JFormalParameter[] parameters();

    void setParameters( JFormalParameter[] parameters );

    /**
     * Adds an additional formal parameter to this method, appending
     * it to the end of the existing parameter list.  
     *
     * <pre><jml>
     * requires param != null;
     * </jml></pre>
     */
    void addParameter( JFormalParameter param );

    String ident();

    CType returnType();

    CClassType[] getExceptions();

    long modifiers(); 
    /*@pure*/ boolean isDeclaredNonNull();

    /** Sets the modifiers of this method declaration */
    void setModifiers(long modifiers); 


    /** WMD TODO remove */
    // no longer needed
    // boolean isFunctional();

    /** WMD TODO remove */
    // not needed so far
    //    void setFunctional(boolean isFunctional);


    JBlock body();

    //@ pure
    boolean hasBody();

    /**
     * Indicates whether this method uses multiple dispatch
     *
     * @return	true iff a parameter of this method has a dynamic dispatch
     *		annotation or, for an external method, this methods receiver
     *		is of a different type than the generic function's top method
     */
    boolean usesMultipleDispatch();

    /**
     * Compares this method to a given method and returns 0 if the
     * methods belong to the same generic function, otherwise returns
     * -1 or +1 to sort the methods.  This comparator is used to
     * separate methods by generic function in the
     * MJGenericFunctionDispatcher class.
     *
     * @param	o	the object to be compared against, must be a
     *			JMethodDeclarationType
     * @return	-1, +1, or 0
     * @exception ClassCastException if <code>o</code> is not an instance of
     *					CType */
    int compareTo(/*@ non_null @*/ Object o) throws ClassCastException;

    /**
     * Indicates whether this member is external.
     * Always returns false for this but is overridden for external
     * subclasses.
     * @return	a flag indicating whether this member is external
     */
    boolean isExternal();

    /** Return <code>true</code> if this method declaration overrides
     * any of its superclass (or interfaces) method declarations.
     */
    boolean isOverriding();

    /** Return the set of methods that are directly overriden (specialized)
     * by this method declaration.
     */
    /*@ nullable @*/ CMethodSet overriddenMethods();

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things
     * generally look OK.  This pass gathers information about
     * the type signatures of everything (imported class
     * files, classes being compiled, methods, fields, etc...)
     * needed for the later passes.  This information is stored 
     * in a context hierarchy that is bound to the AST.
     *
     * @return	a source code representation of the method, iff sub tree 
     *		is correct enough to check code
     * @exception	PositionedError	an error with reference to the 
     *					source file
     */
    CMember checkInterface( CContextType context ) throws PositionedError;

    /**
     * Performs the interface checks that are common to all sorts
     * of methods. Gets the signatures of the return type,
     * parameters and exceptions, and generates the source code
     * representation, CSourceMethod.
     *
     * @param	context		the context in which this method appears
     * @param	access		the <code>MemberAccess</code> object for this method
     * @param	ident		the method name (passed as a parameter 
     *				instead of using the field to properly 
     *				handle constructors where the field is 
     *				the class name but ident is &lt;init&gt;
     * @return	a source code representation of the method, iff sub tree 
     *		is correct enough to check code
     * @exception	PositionedError	an error with reference to the 
     *					source file
     */
    public CSourceMethod checkInterfaceType( CContextType context,
					     MemberAccess access, 
					     String ident) 
	throws PositionedError;

    /**
     * Computes the values of specializer expressions used to dispatch on
     * compile-time constants.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    void resolveSpecializers(CContextType context) throws PositionedError;

    
    /**
     * Makes sure that all in-scope external generic functions are
     * added to the appropriate augmentation maps before top method
     * searches occur in later passes.  */
    public void resolveExtMethods( CContextType context );

    /**
     * Finds the top method of every declared method.  Called after
     * the resolveSpecializers pass.  This cannot be done before then
     * because the external generic function mappings are not complete
     * until the end of the checkInterface pass and the constant value
     * specializers are not known until after the resolveSpecializers
     * pass.  This must be done before the typecheck pass so that all
     * specialized argument positions for generic functions are known
     * for ambiguity checking. */
    public void resolveTopMethods() throws PositionedError;

	
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Creates a context for this method declaration AST node.
     *
     * @param	parent		the parent context
     */
    CMethodContextType createSelfContext(CClassContextType parent);

    /**
     * Typechecks this method declaration.  Mutates the context to
     * record the information gathered during the checks.
     *
     * @param context the context in which this method appears
     * @exception PositionedError if the checks fail and the failure
     * cannot be recovered from */
    void typecheck( CContextType context ) throws PositionedError;

    /**
     * Checks that this method appropriately overrides the given
     * superclass methods.  The checks are those given in JLS2,
     * section 8.4.6.3 and CLCM 00 4.1.3.1 (extended per Clifton's
     * thesis).
     *
     * @param	context		the context in which this appears
     * @param	superMethods	the super type methods that this
     *                          <em>may</em> override
     * @exception PositionedError if a check fails */
    void checkOverriding( CContextType context, CMethodSet superMethods ) 
	throws PositionedError;

}
