/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JMinusExpression.java,v 1.15 2005/02/12 23:24:12 xi_hao Exp $
 */

package org.multijava.mjc;
import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the subtraction binary expression.
 */
public class JMinusExpression extends JBinaryArithmeticExpression implements Constants{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JMinusExpression( TokenReference where,
			     JExpression left,
			     JExpression right ) {
	super( where, left, right );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return left.toString() + " - " + right.toString();
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks a binary arithmetic expression and handles simple
     * optimizations like constant folding.
     *
     * @param context the actual context of analysis
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if checks fail */
    public JExpression typecheckBinaryArith( CExpressionContextType context )
	throws PositionedError
    {
	isInSafeMathMode  = context.arithmeticMode().equals(AMID_SAFE_MATH);

	try {
	    type = computeType( left.getType(), right.getType() );
	    return ( left.isConstant() && right.isConstant() )
		? constantFolding( context )
		: verifyOperation( this, context );
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
    }
    
    /**
     * Verify the operation. Overridden in JML to issue warnings.
     */
    public JExpression verifyOperation(JExpression expr,  CExpressionContextType context ){
    	return expr;
    }

    /**
     * compute the type of this expression according to operands
     * @param	leftType		the type of left operand
     * @param	rightType		the type of right operand
     * @return	the type computed for this binary operation
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public static CType computeType(CType leftType, CType rightType)
	throws UnpositionedError
    {
	if( leftType.isNumeric() && rightType.isNumeric() ) {
	    return ((CNumericType)leftType).binaryPromote( rightType );
	}
	throw new UnpositionedError( MjcMessages.MINUS_BADTYPE,
				     leftType.toVerboseString(),
				     rightType.toVerboseString() );
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @return	a literal resulting of an operation over two literals
     */
    public JExpression constantFolding( CExpressionContextType context )
	throws UnpositionedError
    {
	if (type == CStdType.Long) {
	    long result,
		left_value = left.getNumberLiteral().numberValue().longValue(),
		right_value = right.getNumberLiteral().numberValue().longValue();

	    result = !context.arithmeticMode().equals(AMID_JAVA_MATH)
		? safe_compute( left_value, right_value )
		: compute( left_value, right_value, context );
	    return new JOrdinalLiteral( getTokenReference(),
					new Long(result), CStdType.Long);
	} else if (type == CStdType.Float) {
	    float		result;

	    result = compute(left.getNumberLiteral().numberValue().floatValue(), right.getNumberLiteral().numberValue().floatValue());
	    return new JRealLiteral(getTokenReference(), new Float(result), CStdType.Float);
	} else if (type == CStdType.Double) {
	    double		result;

	    result = compute(left.getNumberLiteral().numberValue().doubleValue(), right.getNumberLiteral().numberValue().doubleValue());
	    return new JRealLiteral(getTokenReference(), new Double(result), CStdType.Double);
	} else if (type.implicitlyCastTo(CStdType.Integer)) {
	    int result,
		left_value = left.getNumberLiteral().numberValue().intValue(),
        	right_value = right.getNumberLiteral().numberValue().intValue();

	    result = !context.arithmeticMode().equals(AMID_JAVA_MATH)
		? safe_compute( left_value, right_value )
		: compute( left_value, right_value, context );
	    return new JOrdinalLiteral( getTokenReference(),
					new Integer(result), CStdType.Integer);
	} else {
	    fail();
	    return null;
	}
    }

    public int safe_compute(int left, int right)
	throws UnpositionedError
    {
    	int result = left - right;
	if((left >= 0 && right < 0 && result < 0) || (left < 0 && right > 0 && result > 0)){
	    throw new UnpositionedError( MjcMessages.INTEGRAL_ARITHMETIC_OVERFLOW );
	}
	return result;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public int compute(int left, int right, CExpressionContextType context) {
	return left - right;
    }

    public long safe_compute(long left, long right)
	throws UnpositionedError
    {
	long result = left - right;
	if((left >= 0 && right < 0 && result < 0) || (left < 0 && right > 0 && result > 0)){
	    throw new UnpositionedError( MjcMessages.INTEGRAL_ARITHMETIC_OVERFLOW );
	}
	return result;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public long compute(long left, long right, CExpressionContextType context) {
	return left - right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public float compute(float left, float right) {
	return left - right;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public double compute(double left, double right) {
	return left - right;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitMinusExpression( this );
    }

    /**
     * @param	type		the type of result
     * @return	the type of opcode for this operation
     */
    public static int getOpcode(CType type) {
	switch (type.getTypeID()) {
	case TID_FLOAT:
	    return opc_fsub;
	case TID_LONG:
	    return opc_lsub;
	case TID_DOUBLE:
	    return opc_dsub;
	default:
	    return opc_isub;
	}
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	left.genCode(code);
	right.genCode(code);
	if (isInSafeMathMode){

	    if (getType().equals(CStdType.Long)) {
                    
		code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
							       "org/multijava/mjc/SafeIntegralArithmetic",
							       "subtract",
							       "(JJ)J"));
	    } else  if ((getType().equals(CStdType.Double))||(getType().equals(CStdType.Float))){
		code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
	    } else {                  
		code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
							       "org/multijava/mjc/SafeIntegralArithmetic",
							       "subtract",
							       "(II)I"));
	    }
	} else {
	    code.plantInstruction(new NoArgInstruction(getOpcode(getType())));
        }
    }
    // private data
    private boolean isInSafeMathMode  = false;
}
