/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JNameExpression.java,v 1.22 2008/03/16 15:52:54 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a name within an expression.  This could be a
 * variable reference, a field reference, a type name, a package name.
 * The meaning of the name is determined according to the rules of
 * JLS2 6.5.  */
public class JNameExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a node in the parsing tree.
     *
     * <pre><jml>
     * requires ident != null;
     * ensures this.ident != null;
     * </jml></pre>
     *
     *
     * @param	where	the line of this node in the source code
     * @param	ident	the short name of this name
     */
    public JNameExpression( TokenReference where, String ident ) {
	this( where, ident, null );
    }


    /**
     * Constructs a node in the parsing tree.
     *
     * <pre><jml>
     * requires ident != null;
     * ensures this.ident != null;
     * </jml></pre>
     *
     *
     * @param	where	the line of this node in the source code
     * @param	ident	the short name of this name
     * @param   univ    the universe modifier WMD
     */
    public JNameExpression( TokenReference where, String ident,
			    CUniverse univ ) {
	super( where );

	assertTrue( ident != null );
	this.ident = ident;
	this.universe = univ;
        this.sourceIdent = ident;
	assertTrue(ident.indexOf('.') == -1); // $$$
	assertTrue(ident.indexOf('/') == -1);
    }

    /**
     * Constructs a node in the parsing tree.
     *
     * <pre><jml>
     * requires ident != null;
     * ensures this.ident != null;
     * </jml></pre>
     *
     *
     * @param	where	the line of this node in the source code
     * @param	prefix	the primary expression qualifying this name
     * @param	ident	the short name of this name
     */
    public JNameExpression( TokenReference where, 
			    JExpression prefix, String ident ) {
	this( where, prefix, ident, null );
    }

    /**
     * Constructs a node in the parsing tree.
     *
     * <pre><jml>
     * requires ident != null;
     * ensures this.ident != null;
     * </jml></pre>
     *
     *
     * @param	where	the line of this node in the source code
     * @param	prefix	the primary expression qualifying this name
     * @param	ident	the short name of this name
     * @param   univ    the universe modifier WMD
     */
    public JNameExpression( TokenReference where, 
			    JExpression prefix, String ident,
			    CUniverse univ ) {
	super(where);

	assertTrue( ident != null );
	this.prefix = prefix;
	this.ident = ident;
	this.universe = univ;
        this.sourcePrefix = prefix;
        this.sourceIdent = ident;
	assertTrue(ident.indexOf('.') == -1); // $$$
	assertTrue(ident.indexOf('/') == -1);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return qualifiedName();
    }
    
    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return null;
    }

    /**
     * Returns the name of this name expression.
     *
     * <pre><jml>
     * ensures \result != null;
     * </jml></pre>
     */
    public String getName() {
	return ident;
    }

    /**
     * @return the prefix of this name expression
     */
    public JExpression getPrefix() {
	return prefix;
    }

    /**
     * @return the prefix of this name expression as it appeared in
     * the source code.
     */
    public JExpression sourcePrefix() {
	return sourcePrefix;
    }


    /**
     * @return the identifier of this name expression as it appeared in
     * the source code.
     */
    public String sourceIdent() {
		return sourceIdent;
    }

    /**
     * Returns the longest name available
     */
    public String qualifiedName() {
	String str = getName();

	if (prefix == null) {
	    return str;
	} else if (prefix instanceof JNameExpression) {
	    return ((JNameExpression)prefix).qualifiedName() + "." + str;
	} else if (prefix instanceof JTypeNameExpression) {
	    return ((JTypeNameExpression)prefix) + "." + str;
	} else {
	    return str;
	}
		// WMD what should be done with the universe modifier?
    }

    public /*@pure*/ CUniverse getCUniverse() {
    	return this.universe;
    }
    
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	    throw new RuntimeException("unexpected call to JNameExpression.isNonNull()");
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see 
     *		<code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails
     * @see JExpression#typecheck(CExpressionContextType)
     */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError {
	try {
	    // 6.5.2 Reclassification of Contextually Ambiguous Names
	    // If the AmbiguousName is a simple name, consisting of a
	    // single Identifier:
	    if( prefix == null ) {
		// If the Identifier appears within the scope of a
		// local variable declaration or parameter
		JLocalVariable	var = context.lookupLocalVariable(ident);
		if( var != null ) {
		    return
			new JLocalVariableExpression( getTokenReference(), var)
			    .typecheck( context );
		}
	
		// Otherwise, consider the class or interface C within
		// whose declaration the Identifier occurs
		if( context.lookupField(ident) != null ) {
		    return createClassField( getTokenReference(), ident )
			.typecheck( context );
		}

		// If the Identifier appears within (an outer) scope
		// of a local variable declaration or parameter
		JExpression outer = 
		    context.lookupOuterLocalVariable( getTokenReference(), 
						      ident );
		if( outer != null ) {
		    return outer.typecheck(context);
		}
	
		// Otherwise, outers
		if( context.lookupOuterField(ident) != null ) {
		    return createClassField( getTokenReference(), ident )
			.typecheck( context );
		}
	
		try {
		    // Otherwise: TypeName
		    // WMD: added the universe modifier
		    CClassType type = CTopLevel.getTypeRep(ident,universe,false);

		    type = (CClassType)(type.checkType( context ));
		    return new JTypeNameExpression( getTokenReference(), 
						    type,
                                                    this).typecheck(context);
		} catch (UnpositionedError cue) {
		    // Otherwise: PackageName
		    throw new CLineError( getTokenReference(),
					  MjcMessages.VAR_UNKNOWN, 
					  ident.replace('/','.'));
		}
	    }

	    // assertTrue( prefix != null );
      
	    try {
		prefix = prefix.typecheck( context );
	    } catch( CLineError cue ) {
		if( cue.hasDescription( MjcMessages.VAR_UNKNOWN ) ) {
		    return convertToPackageName( context );
		} else {
		    throw cue;
		}
	    }

	    // If the name to the left of the "." is reclassified as a
	    // TypeName
	    if( prefix instanceof JTypeNameExpression ) {
		if( ((JTypeNameExpression) prefix)
		    .getClassType().getCClass().lookupField( ident, context ) 
		    != null ) {
		    // there is a field
		    return createClassField( getTokenReference(), prefix, 
					     ident )
			.typecheck( context );
		} else {
		    try {
			// WMD: added universe
			// no field => should be a type name
			String qualifiedName 
			    = ((JTypeNameExpression) prefix).qualifiedName()
			      + "/" + ident;
			return 
			    new JTypeNameExpression( getTokenReference(),
						     qualifiedName, universe,
                                                     this).typecheck(context);
		    } catch (PositionedError cue) {
			// Otherwise: PackageName
			throw new CLineError( getTokenReference(),
					      MjcMessages.VAR_UNKNOWN, 
					      ((JTypeNameExpression) prefix)
					      .qualifiedName().replace('/','.')
					      + "." + ident);
		    }
		}
	    }
      
	    // If the name to the left of the "." is reclassified as
	    // an ExpressionName
      	    if( ident == Constants.JAV_LENGTH ) {
		// is it an array access ?
		if( prefix.getType().isArrayType() ) {
		    JArrayLengthExpression lengthExpression = 
			new JArrayLengthExpression( getTokenReference(), prefix );
		    return lengthExpression.typecheckNullity(context);
		}
	    }
      
	    return createClassField( getTokenReference(), prefix, ident )
		.typecheck( context );
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
    }

    /**
     * Called if the prefix doesn't typecheck, this tries to use the
     * prefix as a package name to be prepended to the identifier
     * hopefully yielding a fully qualified name.
     *
     * @param	context	the context in which this occurs
     * @return a fresh type name expression
     * @exception PositionedError	if the resulting fully qualified 
     *					name is not a valid type name */
    private JExpression convertToPackageName( CExpressionContextType context ) 
	throws PositionedError 
    {
	try {
	    if( prefix instanceof JNameExpression ) {
		CClassType type =
		    CTopLevel.getTypeRep( ((JNameExpression) prefix).getName() 
					  + '/' + ident, false);
		type = (CClassType)(type.checkType( context ));
		return new JTypeNameExpression( getTokenReference(), 
                                                type, this ).
		    typecheck(context);
	    } else if( !(prefix instanceof JTypeNameExpression) ) {
		throw new CLineError( getTokenReference(), 
				      MjcMessages.VAR_UNKNOWN, 
				      ident.replace('/','.'));
	    } else {
		return prefix;
	    }
	} catch (UnpositionedError cue2) {
	    // Otherwise: PackageName
	    ident = ((JNameExpression) prefix).getName() + '/' +  ident;
	    throw new CLineError( getTokenReference(), 
				  MjcMessages.VAR_UNKNOWN, 
				  ident.replace('/','.'));
	}
    }

    /**
     * Since class field may be overloaded in sub compiler, this method allow
     * you to specifie the type of class field without needed to touch
     * the huge method above !
     */
    protected JClassFieldExpression createClassField(TokenReference ref,
						     JExpression prefix,
						     String ident)
    {
	return new JClassFieldExpression(ref, prefix, ident, this);
    }

    /**
     * Since class field may be overloaded in sub compiler, this method allow
     * you to specifie the type of class field without needed to touch
     * the huge method above !
     */
    protected JClassFieldExpression createClassField(TokenReference ref,
						     String ident)
    {
	return new JClassFieldExpression(ref, ident, this);
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitNameExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	fail();
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		prefix;

    /** The prefix as appeared in the source code; note that prefix
     * may be changed as the result of typechecking. This may be null. */
    private JExpression	sourcePrefix;

    /** The identifier as appeared in the source code; note that ident
     * may be changed as the result of typechecking. This may be null. */
    private String sourceIdent;


    /**
     * The short name of this, i.e., the name following the dot
     * between prefix and this.
     *
     * <pre><jml>
     * private invariant ident != null;
     * </jml></pre>
     * */
    //@ spec_public
    private String ident;


    /** The universe associated with the type name.
     *
     * By WMD.
     */
    private     CUniverse               universe;

}
