/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JNewAnonymousClassExpression.java,v 1.16 2005/04/26 02:39:32 kui_dai Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a new anonymous class allocation expression.
 * For example, this is <code>new type(...) {...}</code>.  JLS2 15.9.5 */
public class JNewAnonymousClassExpression extends JNewObjectExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	objectType	the type of this object allocator
     * @param	explicitThis	a TypeName expression
     * @param	params		parameters to be passed to constructor
     */
    public JNewAnonymousClassExpression( TokenReference where,
					 CClassType objectType,
					 JExpression explicitThis,
					 JExpression[] params,
					 JClassDeclarationType decl ) {
	super( where, objectType, explicitThis, params );
	
	this.decl = decl;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JClassDeclarationType decl() { 
	return decl; 
    }

    /**
     * Returns the type of this expression (called after parsing).
     * This method typically returns the same type as returned by
     * <code>getType()</code>; but for nodes such as
     * {@link JNewAnonymousClassExpression}, it returns the type
     * written by the programmer, e.g., <code>T</code> in 
     * <code>new T(...) {...}</code>.
     *
     * @return the apparent type of this expression 
     * @see #getType()
     */
    public CType getApparentType() {
	return CTopLevel.getTypeRep(originalTypeName(),false);
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Resolves and checks the types of the actual parameters to the
     * constructor.  The returned data structure also carries
     * information on the types of any explicit and implicit outer
     * this parameters for various sorts of inner classes.
     * 
     * Builds the declaration for an anonymous inner class that
     * implements or subclasses the interface or class given by
     * <code>type</code>.
     *
     * <pre><jml>
     * also
     * requires typeState == IS_NON_ANON_TYPE;
     * ensures typeState == IS_ACTUAL;
     * </jml></pre>
     *
     * @param	context		the context in which this expression occurs
     * @return	an array of CType singletons for the argument types
     * @exception PositionedError if the checks fail 
     * @see #type */
    protected ArgTypes checkTypes( CExpressionContextType context ) 
	throws PositionedError 
    {
	try {
	    ArgTypes argTypes = super.checkTypes( context );
	    CClassContextType classContext = context.getClassContext();
	    CClass host = classContext.getHostClass();
	    CClass owner = classContext.getOwnerClass();
	    CClassType superType;

	    if( type.getCClass().isInterface() ) {
		decl.setInterfaces( new CClassType[] { type } );
		superType = CStdType.Object;
	    } else {
		superType = type;
	    }
	
	    CClass superClass = superType.getCClass();
	    //decl.setSuperClass( superClass.qualifiedName() );
	    decl.setSuperClass(superType);
	    // An anonymous class instantiation in a "static context",
	    // i.e., in a static method or static initializer, is
	    // considered to be a nested class.  An anonymous class
	    // instantiation in a "non-static context", is an inner
	    // class.
	    if (context.isStatic()) {
		decl.setStatic();
	    }

	    decl.generateInterface( context.getCompiler(), owner, host,
				    host.qualifiedName() + "$" 
				    + host.getNextSyntheticIndex(),
				    true, false );

	    // this call will set the anonymous class's associated
	    // CSourceClass instance's super class appropriately
	    decl.preprocessDependencies( context );

	    // add implicit constructor with appropriate explicit
	    // constructor call to the supertype based on the
	    // parameters given
	    JConstructorDeclarationType	cstr;

	    CMethod superCstr = 
		superClass.lookupMethod( JAV_CONSTRUCTOR,
					 argTypes.constructorTypes(), 
					 CClassType.EMPTY,
					 decl.createContext( context ) );

	    if( superCstr == null ) {
		throw new CMethodNotFoundError( getTokenReference(), 
						type.toString(), 
						argTypes.regularArgs(), 
						true );
	    }

	    CType ownerType = owner.getType();

	    // Next we add an additional synthetic outer this
	    // argument, unless the anonymous class is declared in a static
	    // context or in a constructor before the superclass constructor
	    // has been called.  In the latter case, it is not legal to
	    // access the synthetic outer this arg, so it should not be
	    // generated.

	    // ------- primary expr given (i.e., explicitThis != null)
	    // | 
	    // | ----- superClass.isInnerClass()
	    // | |
	    // | | --- owner.descendsFrom( superClass.owner() )
	    // | | |
	    // V V V
	    // n n -	1. Add ownerType as the only irregular parameter.
	    // n y n	2. Owner must be a _member_ of a class that descends 
	    //		from superClass.owner() (else inner inheritance 
	    //		checking in super.checkTypes() would have failed).
	    //		Replace existing irregular parameter.
	    // n y y    3. Replace existing irregular parameter with ownerType.
	    // y n -	4. Not applicable, typechecking would have failed  
	    //		since superClass would be searched for within the  
	    //		type of the primary expression and given that 
	    //		superClass is not an inner class the search would have 
	    //		failed.
	    // y y -	5. Since the outer this for superClass is given 
	    //		explicitly we must add ownerType as an additional 
	    //		irregular parameter.

	    CMethodContextType methodContext = context.getMethodContext();
	    boolean generateSyntheticOuterThis =
		!(context.isStatic() ||
		  context.isBeforeSuperConstructorCall());

	    if (generateSyntheticOuterThis) {
		argTypes.setSyntheticOuterThis( ownerType );
	    } else {
		decl.syntheticOuterThisInaccessible();
	    }
	    
	    	    
	    // ========================================================
	    // Next we generate an AST for the implicitly declared
	    // constructor of the anonymous class.

	    // --------------------------------------------------------
	    // Declare a bunch of variables for intermediate values

	    // The exceptions thrown by the implicit constructor are
	    // those thrown by the super constructor.
	    CClassType[] throwables = superCstr.throwables();

            // The types of the implicit constructor's parameters.
	    // This is an initial best guess and will be mutated to
	    // match the super constructor's parameter types.
	    CSpecializedType[] cstrParamTypes = CSpecializedType.
		unspecializedTypeTupleFrom( argTypes.constructorTypes() );

	    // The super constructor's parameter types
	    CSpecializedType[] superCstrParamTypes = superCstr.parameters();

	    // ASTs for the implicit constructor's parameter
	    // declarations
	    JFormalParameter[] fparams = 
		new JFormalParameter[cstrParamTypes.length];

	    // ASTs for the actual arguments to the super constructor
	    // invocation
	    JExpression[] superArgs = 
		new JExpression[superCstrParamTypes.length];

	    // ------------------------------------------------------
	    // Populate fparams and superArgs arrays...
	    int implToSuperOffset = 
		superCstrParamTypes.length - cstrParamTypes.length;
	    // see case analysis above
	    boolean case2or3 = generateSyntheticOuterThis && 
		explicitThis == null && superClass.isInnerClass();
	    boolean case2 = case2or3 && 
		!owner.descendsFrom( superClass.owner() );
	    boolean case3 = case2or3 && 
		owner.descendsFrom( superClass.owner() );

	    for( int i = 0; i < cstrParamTypes.length; i++ ) {
		int aPos = i + implToSuperOffset;
		if (aPos >= (case2or3 ? 1 : 0)) {
		    // mutate cstrParamTypes[i] to match super
		    // constructor type
		    cstrParamTypes[i] = 
			superCstrParamTypes[aPos];		    
		}
		
		// For parameters that will just be passed on to
		// superclass constructor we can use any name, but the
		// local outer this, if present, must be named
		// correctly or the code to store the parameter into
		// the appropriate local field will be wrong.
		String paramName;
		if (i==0 && argTypes.hasSyntheticOuterThis()) {
		    paramName = JAV_OUTER_THIS;
		} else {
		    paramName = "dummy" + i;
		}

		fparams[i] = 
		    new JFormalParameter( getTokenReference(),
					  JLocalVariable.DES_GENERATED,
					  cstrParamTypes[i], paramName, true );
		if (aPos >= (case2 ? 1 : 0)) {
		    superArgs[aPos] = 
			new JLocalVariableExpression( getTokenReference(),
						      fparams[i] );
		}
	    }

	    if (case2) {
		// We need to use the outer this field to get to the
		// enclosing instance (of this anonymous class) that
		// is a subclass of superClass's outer type.
		superArgs[0] = new JThisExpression( getTokenReference(), 
						    superClass.owner() );
	    }

	    cstr = 
		new JConstructorDeclaration( 
		       getTokenReference(),
		       0L, decl.getCClass().ident(), 
		       fparams, throwables,
		       new JConstructorBlock(
			      getTokenReference(), 
			      new JStatement[] {
				  new JExpressionStatement(
					getTokenReference(),
					new JExplicitConstructorInvocation( 
					      getTokenReference(), JAV_SUPER, 
					      superArgs), 
					null )
			      } ) );
	    decl.setDefaultConstructor( cstr );
	    // ============================================================

	    
	    // Type check the anonymous class and add its information
	    // to the global symbol table.
	    context.catchUp(decl);

	    type = decl.getCClass().getType();
	    type = (CClassType)(type.checkType( context ));
	    //@ set typeState = IS_ACTUAL;

	    return argTypes;
	} catch (UnpositionedError cue) {
	    throw cue.addPosition( getTokenReference() );
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitNewAnonymousClassExpression( this );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	JClassDeclarationType	decl;

}
