/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JNewArrayExpression.java,v 1.11 2008/03/16 15:52:54 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a new array allocation expression 'new type[...]'
 */
public class JNewArrayExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	arrayType	the type of this array allocator
     * @param	dims		dimensions and initializers
     */
    public JNewArrayExpression(TokenReference where,
			       CType arrayType,
			       JArrayDimsAndInits dims) {
	super(where);

	// this.type = arrayType;
	this.type = CUniverseServices.setDefaultVariable(arrayType);
	
	this.dims = dims;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JArrayDimsAndInits dims() { return dims; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return type;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
    	return dims.isNonNull(context);
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError {
	try {
	    type = type.checkType(context);
	} catch (UnpositionedError e) {
	    throw e.addPosition(getTokenReference());
	}

        // Arrays of non-reifiable types cannot be instantiated. (JLS3 15.10)
        check(context, 
              type.isReifiableType(), 
              MjcMessages.NON_REIFIABLE_ARRAY_INSTANTIATION,
              type.toVerboseString());

	type = dims.typecheck( context, type );

	try {
	    type = type.checkType( context );

	    type = CUniverseServices.setDefaultVariableAgain(type);
	    // WMD defaulting might set the main modifier to any,
	    // e.g. in "new any Integer[5]"
	    if( type instanceof CClassType &&
	    	((CClassType)type).getCUniverse() instanceof CUniverseReadonly ) {
	    	// at the moment change it back to peer
	    	// once we allow "new any" this will be ok again
	    	((CClassType)type).setCUniverse(CUniverseImplicitPeer.getUniverse());
	    }

	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}

	if( CTopLevel.getCompiler().universeDynChecks() ) {
	    /* get reference to temporary variable ... */
	    tmpLocalVar = CUniverseRuntimeHelper.getTmpVariable( context,
								 this, type);
		
	    setAnonymousOwner =
		buildNewArrayUniverseDynChecks( context,
		       (CArrayType) type,
		       new JLocalVariableExpression( TokenReference.NO_REF,
						     tmpLocalVar) );
	}
	
	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitNewArrayExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	dims.genCode(code);
	
	/*
	 * scdaniel: universe runtime checks
	 * 
	 * This part generates the code to register anonymous arrays
	 * to the hashtable.
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     (setOwner == null) &&
	     (tmpLocalVar != null)) {
	    tmpLocalVar.genStore(code);
	    setAnonymousOwner.genCode(code);
	    tmpLocalVar.genLoad(code);
	}
    }


    /**
     * Creates the expressions for the dynamic universe typechecks. 
     * 
     * @param context	context used to typecheck the expressions created
     * @param var	the variable where the new object is found
     * @throws PositionedError
     * 
     * By scdaniel.
     */
    public void buildUniverseDynChecks( CExpressionContextType context,
					JExpression var)
	throws PositionedError
    {
	setOwner = buildNewArrayUniverseDynChecks( context,
						   (CArrayType) this.type,
						   var );
    }

    /**
     * Creates the method call for the dynamic universe typechecks. 
     * 
     * @param context	context used to typecheck the expressions created
     * @param var	the variable where the new object is found
     * @throws PositionedError
     * 
     * By scdaniel.
     */
    private static JMethodCallExpression buildNewArrayUniverseDynChecks(
				      CExpressionContextType context,
				      CArrayType type,
				      JExpression var)
	throws PositionedError
    {
	JMethodCallExpression retVal = null;

	// loop to make break possible
	while ( true ) {
	    /* get owner of the current universe */
	    JExpression owner =
		CUniverseRuntimeHelper.getCurrentUniverseOwner(context);
		
	    /* arrays of basic types are handled like normal objects */
	    if ( !(type.getBaseType() instanceof CClassType) ) {
		JExpression args[] = { var,
				       owner };

		/* create the function calls to insert into the bytecode */
		CUniverse universe = type.getCUniverse();
		if ( universe instanceof CUniversePeer ) {
		    retVal =
			new JMethodCallExpression( TokenReference.NO_REF,
						   CUniverseRuntimeHelper.getHashTable(),
						   "setOwnerPeer",
						   args, false);
		} else {
		    retVal =
			new JMethodCallExpression( TokenReference.NO_REF,
						   CUniverseRuntimeHelper.getHashTable(),
						   "setOwnerRep",
						   args, false);
		}
		
    		break;
	    }

	    CClassType baseType = (CClassType) type.getBaseType();
	    
	    /* get element owner */
	    JOrdinalLiteral elementOwner;
	    if ( baseType.getCUniverse() instanceof CUniversePeer ) {
		elementOwner =
		    new JOrdinalLiteral( TokenReference.NO_REF,
					 MjcTokenTypes.LITERAL_peer,
					 CStdType.Integer);
	    } else if ( baseType.getCUniverse() instanceof CUniverseRep ) {
		elementOwner =
		    new JOrdinalLiteral( TokenReference.NO_REF,
					 MjcTokenTypes.LITERAL_rep,
					 CStdType.Integer);
	    } else {
		elementOwner =
		    new JOrdinalLiteral( TokenReference.NO_REF,
					 MjcTokenTypes.LITERAL_readonly,
					 CStdType.Integer);
	    }

	    JExpression args[] = { var, owner, elementOwner};

	    /* create the function calls to insert into the bytecode */
	    CUniverse universe = type.getCUniverse();
	    if ( universe instanceof CUniversePeer ) {
		retVal =
		    new JMethodCallExpression( TokenReference.NO_REF,
					       CUniverseRuntimeHelper.getHashTable(),
					       "setArrayOwnerPeer",
					       args, false);
	    } else {
		retVal =
		    new JMethodCallExpression( TokenReference.NO_REF,
					       CUniverseRuntimeHelper.getHashTable(),
					       "setArrayOwnerRep",
					       args, false);
	    }
				
	    break;
	}

	if( retVal != null ) {
	    // temporarily disable universe checks ...
	    boolean universeChecks = CTopLevel.getCompiler().universeChecks();
	    boolean universePurity = CTopLevel.getCompiler().universePurity();
	    CTopLevel.getCompiler().setUniverseChecks(false);
	    CTopLevel.getCompiler().setUniversePurity(false);

	    // ... typecheck ...
	    retVal.typecheck(context);

	    // ... enable universe checks again.
	    CTopLevel.getCompiler().setUniverseChecks(universeChecks);
	    CTopLevel.getCompiler().setUniversePurity(universePurity);
	}

	return retVal;
    }


    /**
     * Generate the code for the dynamic universe typechecks.
     * 
     * @param code	the CodeSequence for the codegeneration
     * 
     * By scdaniel.
     */
    public void genUniverseDynCheckCode( CodeSequence code ) {
    	if ( setOwner != null )
	    setOwner.genCode(code);
    }

    /* 
     * By rzakhejm
     */
    public String toString() {
    	return "new " + ((CArrayType) getType()).getBaseType().toString() + dims().toString();
    }
    
    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	CType			type;
    protected	JArrayDimsAndInits	dims;

    // scdaniel: universe runtime checks
    private JMethodCallExpression setOwner;
    private JLocalVariable        tmpLocalVar;
    private JMethodCallExpression setAnonymousOwner;
}
