/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JNewObjectExpression.java,v 1.28 2008/05/14 18:16:28 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.Map;

import org.multijava.util.classfile.*;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents an object instantiation expression 'new type(...)'
 */
public class JNewObjectExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     *
     * <pre><jml>
     * ensures typeState == IS_PARSER_GUESS;
     * </jml></pre>
     *
     * @param	where		the line of this node in the source code
     * @param	objectType	the type of this object to be instantiated,
     *				initially just the parser's best guess
     * @param	explicitThis	a Primary expression (i.e., an explicit
     *				outer class for the instantiation of an 
     *				inner class)
     * @param	params		parameters to be passed to constructor
     */
    public JNewObjectExpression( TokenReference where,
				 CClassType objectType,
				 JExpression explicitThis,
				 JExpression[] params ) {
	super(where);

	this.originalTypeName = objectType.originalQualifiedName();
	this.type = objectType;
	
	this.type = (CClassType) CUniverseServices.setDefaultVariable(this.type);
	
	this.params = params;
	this.explicitThis = explicitThis;
	//@ set typeState = IS_PARSER_GUESS;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression[] params() { return params; }

    /**
     * Returns the type of this expression (called after parsing).
     * The type evolves over the course of the compilation passes.
     * Immediately after parsing it is just the parser's best guess.
     * After typechecking the type returned by this method is the
     * actual type of the object instantiated.
     *
     * @return the type of this expression */
    public CType getType() {
	return type;
    }

    /**
     * Returns true iff this expression can be used as a statement (JLS 14.7)
     */
    public boolean isStatementExpression() {
	return true;
    }

    /**
     * Returns the JExpression that is the explicit-this portion of
     * the new Object construct.  */
    public JExpression thisExpr() {
	return explicitThis;
    }

    /**
     * Returns the actual text following the "new" and preceding the
     * parameters.  */
    public String originalTypeName() {
	return originalTypeName;
    }

    
    /**
     * Returns the specific constructor by this object nstantiation
     * expression. This method must be called after typechecking.
     */
    public /*@ pure @*/ CMethod constructor() {
        return constructor;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * <pre><jml>
     * also 
     * requires typeState == IS_PARSER_GUESS;
     * ensures typeState == IS_ACTUAL;
     * </jml></pre>
     *
     *
     * @param context the context in which this expression appears
     * @return a desugared Java expression (see
     * <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails 
     * @see JExpression#typecheck(CExpressionContextType)
     */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	try {	
	    hostClass = context.getClassContext().getHostClass();
	    owner = context.getCMethod().owner();
	     
	    // find the type of the object being instantiated when
	    // there is a qualifying primary expression according to
	    // JLS2, 15.9
	    if (explicitThis != null) {
		// process qualified class instance creation expression
		explicitThis = explicitThis.typecheck( context );
	    
		// search for inner classes of explicitThis's type
		CClass outerClass = explicitThis.getType().getCClass();
		String qualName = type.qualifiedName();
		int index = qualName.lastIndexOf( "$" ) + 1;
		String shortName = qualName.substring( index );
		CClass innerClass = 
		    outerClass.lookupClass( shortName, context );
	    
		check( context, innerClass != null, 
		       MjcMessages.TYPE_UNKNOWN, shortName );
	    
		type = innerClass.getType();
	    }
	    
	    type = (CClassType)(type.checkType( context ));
	    
	    // WMD set the defaults again
	    type = (CClassType)CUniverseServices.setDefaultVariableAgain(type);
	    
	    // WMD defaulting might set the main modifier to any,
	    // e.g. in "new Integer(5)"
	    if( type.getCUniverse() instanceof CUniverseReadonly ) {
	    	// at the moment change it back to peer
	    	// once we allow "new any" this will be ok again
	    	type.setCUniverse(CUniverseImplicitPeer.getUniverse());
	    }
	    
	    check( context,
		   !type.isTypeVariable(),
		   MjcMessages.NEW_TYPE_VARIABLE,
		   originalTypeName());

            // Wildcard type arguments are not allowed in object instantiations.
            check(context,
                  !type.hasWildcardTypeArgument(),
                  MjcMessages.WILDCARD_IN_OBJECT_INSTANTIATION,
                  type);

	    typecheckUniverse(context);
	} catch (UnpositionedError e) {
	    throw e.addPosition(getTokenReference());
	}
	
	//@ set typeState = IS_NON_ANON_TYPE;
	
	ArgTypes argTypes = checkTypes( context );
	
	/*
	 * scdaniel: universe runtime checks
	 * 
	 * This part is for anonymous objects.
	 * It could be used for all kinds of other objects too. This would
	 * simplify the compiler a little but it would also cause some overhead
	 * for using a temporary variable even when there is no need.
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() ) {
		check( context,
			   !context.isBeforeSuperConstructorCall(),
			   CUniverseMessages.NEW_IN_CONSTR_NO_DYNCHECKS );

	    /* get reference to temporary variable ... */
	    tmpLocalVar = CUniverseRuntimeHelper.getTmpVariable(context,
								this, type);

	    JMethodCallExpression[] mce =
		buildNewObjectUniverseDynChecks( context,
			  type,
			  new JLocalVariableExpression( TokenReference.NO_REF,
							tmpLocalVar) );

	    setAnonymousOwner = mce[0];
	    setConstructorData = mce[1];
	}
	
	//@ set typeState = IS_ACTUAL;
	return checkCall( context, argTypes );
    }

	/**
	 * Ensure static correctness of Generic Universe Types.
	 * 
	 * Checks performed:
	 * - No rep must occur in static contexts
	 * 
	 * @param context The current typechecking context.
	 * 
	 * by WMD
	 * by RZU
	 */
	private void typecheckUniverse(CExpressionContextType context)
	throws PositionedError {
		check(context,
				!context.isStatic()
				|| !CUniverseServices.containsRepModifier(type),
				CUniverseMessages.REP_IN_STATIC_FORBIDDEN);
	}

	
	/**
	 * Check the Universe modifiers of the constructor call.
	 * 
	 * @param context the type checking context
	 * @param argTypes the arguments to the constructor
	 * @throws PositionedError
	 */
	private void typecheckUniverseArguments(CExpressionContextType context,
				ArgTypes argTypes) throws PositionedError
	{
		// Check the parameters and arguments by iterating over them
		CType[] parameterTypes = CSpecializedType.staticTypeTupleFrom(null,
				constructor.parameters());
		CType[] argumentTypes = argTypes.regularArgs();
		
		CClassType currentParameterType;
		CClassType currentArgumentType;

		// inner classes have one more formal parameter
		int i,j;
		if( type.getCClass().isInnerClass() ) {
			i = 1; j=0;
			assertTrue(parameterTypes.length == argumentTypes.length + 1);
		} else {
			i = 0; j=0;
			assertTrue(parameterTypes.length == argumentTypes.length);
		}

		for (; i < parameterTypes.length; ++i, ++j) {
			// We only care about parameters of class type
			if (parameterTypes[i] instanceof CClassType) {
				currentParameterType = (CClassType) parameterTypes[i];
				// arg should also be a class type, as we do this check
				// after the normal type checks and a mismatch would have
				// been found already.
				currentArgumentType = (CClassType) argumentTypes[j];
			} else {
				continue;
			}

			// If the call doesn't happen on this, no rep modifiers must occur
			check(context, !CUniverseServices
							.containsRepModifier(currentParameterType),
					CUniverseMessages.REP_NOT_ON_THIS_IN_CONSTRUCTOR_PARAMETER,
					new Object[] { new Integer(i+1),
								currentParameterType.toVerboseString() });

			CClassType substitutedParameterType = CUniverseServices
					.substituteInheritedTypeVariables(
							(CClassType) currentParameterType, type, context);

			// Adapt viewpoint of method parameters
			CClassType combinedType = CUniverseServices
					.combine(type, substitutedParameterType);

			// Perform the substitution
			CClassType substitutedType = getFullySubstitutedType(combinedType, context);

			check( context,
				   currentArgumentType.isAlwaysAssignableTo(
							substitutedType),
							CUniverseMessages.ACTUAL_NOT_ASSIGNABLE_TO_FORMAL_PARAMETER,
							new Object[] { new Integer(i+1),
								currentArgumentType.toVerboseString(),
								constructor.ident(),
								substitutedType.toVerboseString() });
		}
	}

	
	
    /**
	 * Substitutes the type variables of a method parameter.
	 * 
	 * @param inputtype
	 *            The type of the parameter to substitute
	 * @param context
	 *            The context used for type checking
	 * @return The substituted type containing no type variables
	 * 
	 * copied by WMD from JMethodCallExpression and adapted, i.e. removed
	 * method type variables.
	 * 
	 * by RZU
	 */
	private CClassType getFullySubstitutedType(CClassType inputtype,
			CContextType context) throws PositionedError {
		CClassType substitutedType = inputtype;

		// removed possibility of being a method type variable
		// Get the host class
		CClass hostClass = context.getClassContext().getHostClass();

		// If the type is a type variable (T a), we use the receiver's type
		// arguments to replace it
		if (inputtype.isTypeVariable()) {
			substitutedType = type.getCClass().getSubstitution(
					(CTypeVariable) inputtype,
					type.getAllArguments());

			// If the type is an array whose type is a type variable (T[] a),
			// we use the receiver's type arguments to replace it
		} else if (inputtype.isArrayType()
				&& ((CArrayType) inputtype).getBaseType().isTypeVariable()) {
			// Get substituted base type...
			substitutedType = type.getCClass().getSubstitution(
					(CTypeVariable) ((CArrayType) inputtype).getBaseType(),
					type.getAllArguments());
			// ... and create an array out of it
			substitutedType = new CArrayType(substitutedType,
					((CArrayType) inputtype).getArrayBound(),
					((CArrayType) inputtype).getCUniverse());

			// If the type is an array with a generic base type (List<Integer>[]
			// a),
			// we use the receiver's type arguments to replace type variables in
			// the type arguments, if any
		} else if (inputtype.isArrayType()
				&& ((CArrayType) inputtype).getBaseType().isGenericType()) {
			CType baseType = ((CArrayType) inputtype).getBaseType();
			CClassType newType = ((CClassType) baseType).createSubstitutedType(
					hostClass, type, baseType
							.getAllArguments());
			substitutedType = new CArrayType(newType, ((CArrayType) inputtype)
					.getArrayBound(), ((CArrayType) inputtype).getCUniverse());

			// If the type is an array with a generic base type (List<Integer>[]
			// a),
			// we use the receiver's type arguments to replace type variables in
			// the type arguments, if any
		} else if (inputtype.isGenericType()) {
			substitutedType = inputtype.createSubstitutedType(hostClass,
					type, hostClass.getType()
							.getAllArguments());
		}

		return substitutedType;
	}
	
	
	
    /**
	 * This nested class is used to track the argument types of the actual
	 * constructor call. The actual arguments include explicit primary
	 * expressions and synthetic outer this arguments for inner classes.
	 */
    protected static class ArgTypes {
	protected ArgTypes( CType[] regularArgs ) {
	    this.regularArgs = regularArgs;
	}

	protected void setPrimaryExpr( CType primaryExpr ) {
	    this.primaryExpr = primaryExpr;
	}

	protected void setSyntheticOuterThis( CType synthOuterThis ) {
	    this.synthOuterThis = synthOuterThis;
	}

	protected CType[] constructorTypes() {
	    CType[] result = new CType[ typeCount() ];
	    int pos = 0;

	    if (synthOuterThis != null) {
		result[pos++] = synthOuterThis;
	    }

	    if (primaryExpr != null) {
		result[pos++] = primaryExpr;
	    }
	    
	    System.arraycopy(regularArgs, 0, result, pos, regularArgs.length);
		
	    return result;
	}

	protected CType[] regularArgs() {
	    return regularArgs;
	}

	protected boolean hasSyntheticOuterThis() {
	    return synthOuterThis != null;
	}

	/**
	 * Returns the types to which to cast the regular arguments
	 * for the given constructor.
	 *
	 * <pre><jml>
	 * requires constructor.parameters().length == typeCount();
	 * ensures \result.length == regularArgs.length;
	 * </jml></pre>
	 *
	 */
	protected CType[] castToTypes(CMethod constructor) {
	    CType[] result = new CType[regularArgs.length];
	    CType[] cArgs = 
		CSpecializedType.tupleOfSpecializersFrom( constructor.
							  parameters() );

	    // irregular args are all at the beginning of the
	    // constructors parameter list, so we just the end.
	    System.arraycopy( cArgs, cArgs.length - result.length,
			      result, 0, result.length );

	    return result;
	}

	protected /*@ pure @*/ int typeCount() {
	    return  regularArgs.length + 
		(synthOuterThis == null ? 0 : 1) +
		(primaryExpr == null ? 0 : 1);
	}
	
	private /*@ spec_protected @*/ final CType[] regularArgs;

	private CType primaryExpr;

	private CType synthOuterThis;
    }

    /**
     * Resolves and checks the types of the actual parameters to the
     * constructor.  The returned data structure also carries
     * information on the types of any explicit and implicit outer
     * this parameters for various sorts of inner classes.
     *
     * <pre><jml>
     * requires typeState == IS_NON_ANON_TYPE;
     * </jml></pre>
     *
     * @param	context		the context in which this expression occurs
     * @return	an array of CType singletons for the argument types
     * @exception PositionedError if the checks fail */
    protected ArgTypes checkTypes( CExpressionContextType context ) 
	throws PositionedError 
    {
	CType[] regularArgs = new CType[params.length];

	for (int i = 0; i < regularArgs.length; i++) {
	    params[i] = params[i].typecheck( context );
	    regularArgs[i] = params[i].getType();
	    assertTrue( regularArgs[i] != null );
	}

	ArgTypes result = new ArgTypes( regularArgs );

	if (explicitThis != null) {
	    result.setPrimaryExpr( explicitThis.getType() );
	} else if (type.getCClass().isInnerClass()) {
	    CClass outerClass = type.getCClass().owner();
	    // Before adding the synthetic outer-this parameter, we
	    // check that such a parameter is available in this
	    // context.  The presence of an type-correct explicitThis
	    // (as checked in typecheck() above) ensures that the
	    // appropriate synthetic outer-this parameter is
	    // available, because the search that refines the field
	    // type to inner classes of explicitThis's type would fail
	    // otherwise.

	    // a static method doesn't have a "this" to
	    // provide to inner class
	    check( context, !context.isStatic(),
		   MjcMessages.INNER_INHERITENCE, type, hostClass );

	    // make sure a "this" of the correct type is available
	    check( context,
		   owner.nestedDescendsFrom( outerClass ),
		   MjcMessages.INNER_INHERITENCE, type, hostClass );

	    result.setSyntheticOuterThis( type.getCClass().owner().getType() );
	}
	
	return result;
    }

    /**
     * Typechecks the invocation of the actual constructor of the
     * object instantiated.  
     *
     * <pre><jml>
     * requires typeState == IS_ACTUAL;
     * </jml></pre>
     *
     * @param context	the context in which this expression appears
     * @param argTypes  the types of the arguments to the constructor of
     *			<code>type</code>, including any synthetic outer 
     *			this parameters
     */
    protected JExpression checkCall( CExpressionContextType context, 
				     ArgTypes argTypes) 
	throws PositionedError 
    {
	context.setDiscardValue( false );

	try {
	    CClass typeClass = type.getCClass();
	    CClass outerClass = typeClass.owner();
	    
	    check( context,
		   !type.isTypeVariable(),
		   MjcMessages.NEW_TYPE_VARIABLE,
		   originalTypeName());

	    check( context, 
		   !typeClass.isInterface(), 
		   MjcMessages.NEW_INTERFACE, type );
	    check( context, 
		   !typeClass.isAbstract(), 
		   MjcMessages.NEW_ABSTRACT, type );
	    CClassType[] args = type.isGenericType()?
		type.getArguments():CClassType.EMPTY;
	    //Debug.breakMe();
	    constructor = typeClass.lookupMethod( JAV_CONSTRUCTOR,
						  argTypes.constructorTypes(), 
						  args,
						  context );

	    if( constructor != null && constructor.owner() != typeClass ) {
		// Since the lookupMethod generic function used to
		// find the constructor treats constructor inheritance
		// like method inheritance it is possible to find a
		// superclass constructor.  This is not the
		// appropriate semantics and so such a constructor
		// should be discarded.
		constructor = null;
	    }

	    if( constructor == null ) {
		throw new CMethodNotFoundError( getTokenReference(), 
						type.toString(), 
						argTypes.regularArgs(), 
						true );
	    }

            if (CTopLevel.getCompiler().Generic()) {
                checkForUncheckedWarnings(argTypes, context);
            }

            // If we are instantiating a raw type and any of the constructor's
            // formal parameter types changes by erasure, we have to issue an
            // unchecked warning (JLS3 4.8).
            if (type.isRawType()) {
                CSpecializedType[] ctorParams = constructor.parameters();

                // Skip a possible synthetic outer this parameter.
                int firstParam = argTypes.hasSyntheticOuterThis() ? 1 : 0;

                for (int i = firstParam; i < ctorParams.length; i++) {
                    if (ctorParams[i].staticType().changesByErasure()
                        || ctorParams[i].dynamicType().changesByErasure()) {
                        warn(context,
                             MjcMessages.UNCHECKED_CONSTRUCTOR_CALL,
                             type,
                             constructor.toString(true));
                        break;
                    }
                }
            }

	    // WMD: only pure constructors may be used in a pure context
	    if( CTopLevel.getCompiler().universePurity() ) {
		check( context,
		       !context.isPure() ||
		       constructor.isPure(),
		       CUniverseMessages.NON_PURE_CONSTRUCTOR_IN_PURE_METHOD_FORBIDDEN,
		       constructor.ident() );
	    }

	    // construct the expression for the
	    // synthetic outer-this if needed
	    if (argTypes.hasSyntheticOuterThis()) {
		if (hostClass.descendsFrom( outerClass )) {
		    // Instantiation expression appears in a
		    // (possibly reflexive) subclass of the outertype, so we
		    // can just use a regular this expression.
		    outerThisParam = 
			new JThisExpression( getTokenReference() ).
			    typecheck(context);
		} else {
		    // Some enclosing instance of the object in
		    // which this instantiation expression appears
		    // must be used as the outer type.
		    outerThisParam = 
			new JThisExpression( getTokenReference(), 
					     outerClass ).typecheck(context);
		    }

	    }
	    
	    // WMD
	    if( CTopLevel.getCompiler().universeChecks() ) {
	    	typecheckUniverseArguments(context, argTypes);
	    }
	    
	    // add the appropriate type coercions to the parameters to
	    // match the constructor
	    CType[] castToTypes = argTypes.castToTypes( constructor );
	    for( int i = 0; i < params.length; i++ ) {
		params[i] = params[i].convertType( castToTypes[i], context );
	    }

	    // add the exceptions of the constructor being invoked to
	    // those that must be caught or thrown in this context
	    CClassType[] exceptions = constructor.throwables();
	    for( int i = 0; i < exceptions.length; i++ ) {
		context.getFlowControlContext().addThrowable(
		   new CThrowableInfo( exceptions[i], this ));
	    }
    
	    return this;
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
    }

    /**
     * Issues unchecked warnings for the actual parameters which need unchecked
     * conversion (JLS3 5.1.9) to conform to their corresponding formal
     * parameters.
     *
     * @param argTypes  The actual arguments of this constructor call.
     * @param context   The type checking context.
     *
     * @throws PositionedError if some error occurs during type checking.
     */
    private void checkForUncheckedWarnings(
            ArgTypes argTypes,
            CExpressionContextType context) throws PositionedError {
        // In a raw type constructor, the constructor parameter types are erased.
        // Therefore, no unchecked conversion is ever necessary.
        if (type.isRawType()) {
            return;
        }

        CClass clazz = type.getCClass();
        CClassType[][] args = type.getAllArguments();

        CSpecializedType[] formals = constructor.parameters();
        CType[] actuals = argTypes.constructorTypes();

        // Skip possible synthetic arguments.
        int firstParam = actuals.length - argTypes.regularArgs().length;

        for (int i = firstParam; i < formals.length; i++) {
            CType formal = formals[i].dynamicType();

            // Substitute type variables by its actual type arguments.
            if (formal.isClassTypeVariable()) {
                formal = clazz.getSubstitution((CTypeVariable) formal, args);
            } else if (formal.isArrayType()
                       && ((CArrayType) formal).getBaseType().isClassTypeVariable()) {
                CArrayType array = (CArrayType) formal;
                CTypeVariable var = (CTypeVariable) array.getBaseType();
                CClassType subst = clazz.getSubstitution(var, args);
                formal = new CArrayType(subst, array.getArrayBound(), array.getCUniverse());
            }

            if (actuals[i].needsUncheckedConversion(formal)) {
                warn(context,
                     MjcMessages.UNCHECKED_CONVERSION,
                     actuals[i],
                     formal);
            }
        }
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitNewObjectExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	boolean	discardValue = code.discardValue();
	code.setLineNumber( getTokenReference().line() );

	/*
	 * scdaniel: universe runtime checks
	 * 
	 * This code inserts a call to setConstructorData into the bytecode.
	 * setConstructorData saves the current object and the modifier (peer or rep)
	 * for the object to be created into the thread's entry in the
	 * UrtHashtable to be used in the constructor of the new object.
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     (setConstructorData != null) ) {
	    setConstructorData.genCode(code);
	}
	
	code.plantInstruction(
	    new ClassRefInstruction( opc_new, 
				     type.getCClass().qualifiedName() ));

	if ( !discardValue ) {
	    code.plantInstruction( new NoArgInstruction( opc_dup ) );
	}
	
	/* scdaniel: we need a copy for the setOwnerXXX 
	 * call below (this is for anonymous objects) */
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     (setOwner == null) ) {
	    code.plantInstruction( new NoArgInstruction( opc_dup ) );
	}

	code.setDiscardValue( false );

	if (outerThisParam != null) {
	    outerThisParam.genCode( code );
	}
	
	if (explicitThis != null) {
	    explicitThis.genCode( code );
	}
	
	for ( int i = 0; i < params.length; i++ ) {
	    params[i].genCode( code );
	}

	// Generate references to any outer local variables used
	// within the class (only applicable for anonymous inner
	// classes and local class declarations).

	constructor.owner().genSyntheticParams( code );

	MethodRefInstruction invokeConstructor =
	    new MethodRefInstruction( opc_invokespecial,
				      constructor.getOwnerName(),
				      constructor.ident(),
				      constructor.getSignature() );
	code.plantInstruction( invokeConstructor );
	
	/*
	 * scdaniel: universe runtime checks
	 * 
	 * This part generates the code to register anonymous
	 * objects to the hashtable.
	 */
	if ( CTopLevel.getCompiler().universeDynChecks() &&
	     (setOwner == null) ) {
	    tmpLocalVar.genStore(code);
	    setAnonymousOwner.genCode(code);
	}
    }


    /**
     * Creates the expressions for the dynamic universe typechecks. 
     * 
     * @param context	context used to typecheck the expressions created
     * @param var		the variable where the new object is found
     * @throws PositionedError
     * 
     * By scdaniel.
     */
    public void buildUniverseDynChecks( CExpressionContextType context,
					JExpression var )
	throws PositionedError
    {
	JMethodCallExpression[] mce =
	    buildNewObjectUniverseDynChecks( context, type, var );

	setOwner = mce[0];
	setConstructorData = mce[1];
    }

    /**
     * Creates the expressions for the dynamic universe typechecks. 
     * 
     * @param context	context used to typecheck the expressions created
     * @param var		the variable where the new object is found
     * @return Array mce with the two method call expressions.
     *         mce[0] is the call to use for setOwner or setAnonymousOwner
     *         mce[1] is the call for setConstructorData
     * @throws PositionedError
     * 
     * By scdaniel.
     */
    /*@ ensures \result != null &&
      @         \result.length == 2 &&
      @         \result[0] != null &&
      @         \result[1] != null;
      @*/
    private static JMethodCallExpression[] buildNewObjectUniverseDynChecks(
			     CExpressionContextType context,
			     CClassType type,
			     JExpression var )
	throws PositionedError
    {
	JMethodCallExpression[] retVal = new JMethodCallExpression[2];

	/* get owner of the current universe */
	JExpression owner =
	    CUniverseRuntimeHelper.getCurrentUniverseOwner(context);
		
	JExpression args[] = { var, owner };

	/* create the function calls to insert into the bytecode */
	CUniverse universe = type.getCUniverse();
	if ( universe instanceof CUniversePeer ) {
	    retVal[0] =
		new JMethodCallExpression( TokenReference.NO_REF,
				   CUniverseRuntimeHelper.getHashTable(),
					   "setOwnerPeer",
					   args, false );
			
	    args = new JExpression[] { owner,
	               new JClassExpression(TokenReference.NO_REF, type),
		       new JOrdinalLiteral(TokenReference.NO_REF,
					   MjcTokenTypes.LITERAL_peer,
					   CStdType.Integer) };

	} else {
	    retVal[0] =
		new JMethodCallExpression( TokenReference.NO_REF,
				   CUniverseRuntimeHelper.getHashTable(),
					   "setOwnerRep",
					   args, false );
	    
	    args = new JExpression[] { owner,
		       new JClassExpression(TokenReference.NO_REF, type),
		       new JOrdinalLiteral(TokenReference.NO_REF,
					   MjcTokenTypes.LITERAL_rep,
					   CStdType.Integer) };
	}
	    
	retVal[1] =
	    new JMethodCallExpression( TokenReference.NO_REF,
				       CUniverseRuntimeHelper.getHashTable(),
				       "setConstructorData",
				       args, false);
		

	// temporarily disable universe checks ...
	boolean universeChecks = CTopLevel.getCompiler().universeChecks();
	boolean universePurity = CTopLevel.getCompiler().universePurity();
	CTopLevel.getCompiler().setUniverseChecks(false);
	CTopLevel.getCompiler().setUniversePurity(false);

	// ... typecheck ...
	retVal[0].typecheck(context);
	retVal[1].typecheck(context);

	// ... enable universe checks again.
	CTopLevel.getCompiler().setUniverseChecks(universeChecks);
	CTopLevel.getCompiler().setUniversePurity(universePurity);

	return retVal;
    }

    /**
     * Generate the code for the dynamic universe typechecks.
     * 
     * @param code	the CodeSequence for the codegeneration
     * 
     * By scdaniel.
     */
    public void genUniverseDynCheckCode(CodeSequence code) {
    	setOwner.genCode(code);
    }

    public String toString() {
	return "new " + constructor().ident() + "(" + paramString() + ")";
    }

    private String paramString() {
	String paramString = " ";
	if (params != null) {
	    for(int p = 0; p<params.length; p++) {
		JExpression param = params[p];
		if (param != null)
		    paramString += param + (p < params.length - 1 ? ", " : "");
	    }
	}
	
	return paramString;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * The primary expression qualifying this class instance creation
     * expression, or null if this is an unqualified class instance
     * creation expression.  */
    protected JExpression explicitThis;

    /**
     * The expressions that, when evaluated, yield the actual
     * parameters for the constructor invocation.  */
    protected	JExpression[]		params;

    /**
     * The type of object instantiated by this expression.  Initially
     * this is just the parser's best guess.  The valued stored in
     * this field changes during the typechecking pass when the
     * optional leading primary expression (stored in
     * <code>explicitThis</code> is considered and when the
     * declaration AST for an anonymous classes is defined.
     *
     * @see #explicitThis
     * @see JNewAnonymousClassExpression
     */
    protected	CClassType		type;

    /*@
      @ public ghost int typeState;
      @ public final static ghost int IS_PARSER_GUESS = 0;
      @ public final static ghost int IS_NON_ANON_TYPE = 1;
      @ public final static ghost int IS_ACTUAL = 2;
      @
      @ invariant typeState == IS_PARSER_GUESS 
      @			==> (* type is parser's best guess *);
      @ invariant typeState ==  IS_NON_ANON_TYPE
      @			==> (* type is refined to consider explicitThis, 
      @			       but for anonymous class type is still 
      @			       the supertype *);
      @ invariant typeState == IS_ACTUAL 
      @			==> (* type is refined to the actual
      @			       type of the object instantiated
      @			       at runtime *);
      @*/

    /**
     * A reference to the host class of this expression.
     * */
    protected CClass hostClass;

    /**
     * A reference to the logical owner of this expression.
     * */
    protected CClass owner;

    /**
     * The specific constructor that is called in this object
     * instantiation expression.  */
    protected	CMethod			constructor;

    /**
     * Stores the actual text following the <code>new</code> and
     * preceding the parameters.  */
    private final String originalTypeName;


    /**
     * Stores the expression to generate a possible synthetic outer
     * this parameter to the constructor call.  */
    protected JExpression outerThisParam;

    // scdaniel: universe runtime checks
    private JMethodCallExpression setOwner;
    private JMethodCallExpression setConstructorData;
    private JLocalVariable tmpLocalVar;
    private JMethodCallExpression setAnonymousOwner;
}
