/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JOrdinalLiteral.java,v 1.23 2005/02/28 07:04:05 xi_hao Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.PushLiteralInstruction;

import org.multijava.util.compiler.NumberParser;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents literals of primitive integral types (byte, short,
 * int, long).  Instances of this class are also used to hold the results of
 * computations of constant expressions over integral types.
 */
public class JOrdinalLiteral extends JNumberLiteral implements Constants {

    /*@ protected invariant // ? TBC - Chalin
      @  (image != null || value != null) &&
      @  type != null <==> value != null;
      @*/

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param image  the string representation of this literal
     */
    public JOrdinalLiteral(TokenReference where, String image) {
	super(where);
	this.image = image;
    }

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param value  the value of this literal
     * @param type   the type of this literal
     */
    public JOrdinalLiteral(TokenReference where, Number value, CNumericType type) {
	super(where);
	setValue(value, type);
    }

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param value  the value of this literal
     * @param type   the type of this literal
     */
    public JOrdinalLiteral(TokenReference where, long value, CNumericType type) {
	this(where, new Long(value), type);
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * This is needed to handle -2147483648 that is allowed against
     * 2147483648 that is not allowed, so we need to know if there is
     * an unary minus before us */
    public void negate() { // setSignToMinus
	if (image != null) {
	    image = "-" + image;
	}
	if (value != null) {
	    value = new Long(-value.longValue());
	}
    }

    public boolean isImageOnly() {
	return value == null; // and hence image != null;
    }

    /**
     * Returns the type of this expression.
     */
    //@ also
    //@ protected behavior 
    //@     modifies value, type;
    public CType getType() {
	if( type == null ) {
	    calculateType();
	}
	return type;
    }

    /**
     * Returns a Number that represents the value of this literal.
     */
    /*@ also protected normal_behavior
      @   assignable value, type;
      @   ensures value != null &&
      @     \result == value;
      @*/
    public Number numberValue() {
	if( value == null ) {
	    calculateType();
	}
	return value;
    }

    /**
     * Returns a Number that represents the value of this literal.
     * Do not call before typecheck.
     */
    /*@ also protected normal_behavior
      @   assignable value, type;
      @   ensures value != null &&
      @     \result == value;
      @*/
    public Object getValue() {
	return numberValue();
    }

    /**
     * Sets value to the smallest numeric type that can contain
     * newValue and sets type to newType.  */
    private void setValue(Number newValue, CNumericType newType) {
	long	val = newValue.longValue();

	if (val == newValue.byteValue()) {
	    value = new Byte(newValue.byteValue());
	} else if (val == newValue.shortValue()) {
	    value = new Short(newValue.shortValue());
	} else if (val <= Character.MAX_VALUE && val >= Character.MIN_VALUE) {
	    value = new Integer(newValue.intValue());
	} else if (val == newValue.intValue()) {
	    value = new Integer(newValue.intValue());
	} else {
	    value = newValue;
	}

	type = newType;
    }

    public String image() { return image; }

    /*
     * Does this expression represent an ordinal literal?
     */
    public /*@ pure @*/ boolean isOrdinalLiteral() {
	return true;
    }    

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck(CExpressionContextType context) 
	throws PositionedError {
	if( value == null ) {
	    try {
		if( (image.indexOf("L") > 0) || (image.indexOf("l") > 0) ) {
		    setValue(
		       new Long( NumberParser
				 .decodeLong( image
					      .substring( 0, 
							  image.length() 
							  - 1) )), 
		       CStdType.Long );
		} else {
		    setValue( new Integer( NumberParser.decodeInt(image) ), 
			      CStdType.Integer );
		}
		image = null;
	    } catch( NumberFormatException e ) {
		check( context, false, MjcMessages.LITER_INT_BAD, image );
	    }
	}

	return this;
    }

    /**
     * convertType
     * changes the type of this expression to another
     * @param	dest	the destination type
     * @param	context	the context in which this expression appears
     */
    /*@ also public normal_behavior
      @  requires dest.isNumeric();
      @  assignable \nothing;
      @  ensures \result != null &&
      @    (\result instanceof JOrdinalLiteral) ==>
      @      (* ((JOrdinalLiteral)\result).numberValue().equals(this.numberValue()) *);
      @ also protected normal_behavior
      @  assignable type, value;
      @  ensures type != null && value != null;
      @*/
    public JExpression convertType( CType dest, CExpressionContextType context) 
	throws PositionedError 
    {
	if( type == null ) {
	    calculateType();
	}

	JExpression result;

	if( type.equals(dest) ) {
	    result = this;
	} else if( dest.isOrdinal() )
	{
	    long v = value.longValue();

	    if(type.getTypeID() > dest.getTypeID()) {
		long original_v = v;
		// Adjust, i.e. narrow v based on dest type:
		v =   dest == CStdType.Integer ? (int) v
		    : dest == CStdType.Short ? (short) v
		    : dest == CStdType.Byte  ? (byte) v
		    : dest == CStdType.Char  ? (char) v
		    : v;

		/*check( context, 
		       context.arithmeticMode().equals(AMID_JAVA_MATH) ||
		       original_v == v,
		       MjcMessages.INTEGRAL_ARITHMETIC_OVERFLOW );*/
	    }
	    result = new JOrdinalLiteral( getTokenReference(),
					  v,
					  (CNumericType) dest );
	}
	else if( dest.isFloatingPoint() )
	{
	    Number n = (dest == CStdType.Float) 
		? (Number) new Float((float)value.longValue())
		: (Number) new Double((double)value.longValue());
	    result = new JRealLiteral( getTokenReference(), n, dest);
	} else {
	    fail( "Attempting to convert ordinal literal to " + dest );
	    result = null;
	} 
	return result;
    }

    /**
     * Is this type assignable to the given type by assignment type
     * conversion [JLS2 5.2].  This considers both widening and
     * narrowing conversions.
     * @param	dest		the destination type
     * @return true iff assignment is valid */
    public boolean isAssignableTo(CType dest) {
	// It is possible that type==null when this is called. If so
	// we set type to be the narrowest type possible.  Since
	// widening conversions are implicit this strategy will work.
	// !FIXME! a JLS2 reference for this should be found.
	if( type == null ) {
	    calculateType();
	}
	
	// check widening conversion
	if( type.isAlwaysAssignableTo( dest ) ) {
	    return true;
	}

	// check narrowing conversion [JLS2 5.2]
	if( type.type > TID_INT || dest.type >= TID_INT ) {
	    return false;
	} else {
	    // if no information loss results, then narrowing
	    // conversion is permitted
	    if (dest == CStdType.Byte) {
		long	val = value.longValue();

		return (byte)val == val;
	    } else if (dest == CStdType.Short) {
		long	val = value.longValue();

		return (short)val == val;
	    } else if (dest == CStdType.Char) {
		long	val = value.longValue();

		return (char)val == val;
	    }
	    return false;
	}
    }

    /**
     * Determines the smallest magnitude ordinal type (byte, short, int, or
     * long) that can represent the value of this literal. The
     * <code>value</code> is computed, if it is <code>null</code>. */
    /*@ protected normal_behavior
      @  modifies value, type;
      @  ensures value != null && type != null;
      @*/
    protected void calculateType() {
	if( value == null ) {
	    try {
		String s = image;
		int p = s.indexOf('L');
		if (p != -1) s = s.substring(0,p);
		p = s.indexOf('l');
		if (p != -1) s = s.substring(0,p);
		value = new Long(NumberParser.decodeLong(s));
	    } catch (NumberFormatException e) {
		System.out.println(e);
		value = new Long(0);
	    }
	}
	long val = value.longValue();
	if( Byte.MIN_VALUE <= val && val <= Byte.MAX_VALUE ) {
	    type = CStdType.Byte;
	} else if( Short.MIN_VALUE <= val && val <= Short.MAX_VALUE ) {
	    type = CStdType.Short;
	} else if( Integer.MIN_VALUE <= val && val <= Integer.MAX_VALUE ) {
	    type = CStdType.Integer;
	} else {
	    type = CStdType.Long;
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
//  	assert(value != null);

//  	if (type == CStdType.Byte) {
//  	    p.visitByteLiteral(value.byteValue());
//  	} else if (type == CStdType.Integer) {
//  	    p.visitIntLiteral(value.intValue());
//  	} else if (type == CStdType.Long) {
//  	    p.visitLongLiteral(value.longValue());
//  	} else if (type == CStdType.Short) {
//  	    p.visitShortLiteral(value.shortValue());
//  	} else {
//  	    fail();
//  	}
	p.visitOrdinalLiteral( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	if (type == CStdType.Long) {
	    code.plantInstruction(new PushLiteralInstruction(numberValue().longValue()));
	} else {
	    code.plantInstruction(new PushLiteralInstruction(numberValue().intValue()));
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected String	image;
    protected Number	value;
    protected CType	type;

}
