/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JPackageImport.java,v 1.8 2004/01/16 23:21:11 cclifton Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import org.multijava.util.classfile.ClassPath;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents (in the AST) full-package import statements, 
 * an asterisk.  For example, <code>import java.util.*;</code>.
 */
public class JPackageImport extends JPhylum 
    implements JPackageImportType
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * @param	where		the token reference of this node
     * @param	name		the package name
     * @param	comments	non-javadoc style comments for this statement
     */
    public JPackageImport(TokenReference where, 
			  String name, 
			  JavaStyleComment[] comments) {
	super(where);

	this.name = name;
	this.comments = comments;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS & MUTATORS
    // ----------------------------------------------------------------------

    /**
     * Returns the package name defined by this declaration.
     *
     * @return	the package name defined by this declaration
     */
    public String getName() {
	return name;
    }

    /**
     * States that specified class in imported package is used.
     * @param	clazz		the class that is used.
     */
    public void setClassUsed(String clazz) {
	if (classesUsed == null) {
	    classesUsed = new ArrayList();
	}
	classesUsed.add(clazz);
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the use of package import statements.  Issues error if
     * package cannot be found on classpath.  Issues warnings for
     * things like unused packages being imported or a package import
     * being used when only a small number of classes from the package
     * are actually referenced.  In general these are stylistic
     * issues.
     * @param	compiler	the compiler calling this method
     * @param	pack		the name of the package being imported */
    public void typecheck( Main compiler, JPackageName pack ) {
	// avoid checking automatically generated import of
	// java.lang.*
	if( getTokenReference() != TokenReference.NO_REF ) { 
	    String dottedName = name.replace('/','.');
	    String loc = ClassPath.findPackageSource( dottedName );
	    if (loc == null) {
		loc = ClassPath.findPackage( dottedName );
	    }
	    if (loc == null) {
		compiler.reportTrouble(
		    new CWarning( getTokenReference(),
				  MjcMessages.NON_EXISTENT_PACKAGE,
				  dottedName ) );
	    }
	    
	    
	    if( classesUsed == null ) {
		compiler.reportTrouble( 
			   new CWarning( getTokenReference(),
					 MjcMessages.UNUSED_PACKAGE_IMPORT,
					 dottedName, null ));
	    } else if( classesUsed.size() < 5 ) {
		StringBuffer buffer = new StringBuffer();

		for( int i = 0; i < classesUsed.size(); i++ ) {
		    if( i != 0 ) {
			buffer.append(", ");
		    }
		    buffer.append( (name + "/" 
				    + (String)classesUsed
				    .get(i)).replace('/', '.') );
		}
		compiler.reportTrouble(
			  new CWarning( getTokenReference(),
					MjcMessages.REPLACE_PACKAGE_IMPORT,
					dottedName, buffer.toString() ));
	    }
    
	    if( name.equals("java/lang") ) {
		compiler.reportTrouble(
			  new CWarning( getTokenReference(),
					MjcMessages.IMPORT_JAVA_LANG,
					name ));
	    }
      
	    if( name.equals( pack.getName() ) ) {
		compiler.reportTrouble(
			   new CWarning( getTokenReference(),
					 MjcMessages.IMPORT_CURRENT_PACKAGE,
					 name ));
	    }
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
//  !FIXME! 	if (comments != null) {
//  	    p.visitComments(comments);
//  	}
//  	p.visitPackageImport(name);
  	p.visitPackageImport( this );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private final String		name;
    private final JavaStyleComment[]	comments;
    private ArrayList			classesUsed;
}
