/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JParenthesedExpression.java,v 1.13 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.mjc.CodeSequence;
import org.multijava.mjc.CInitializable;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.mjc.CExpressionContextType;
import org.multijava.mjc.CContextType;

/**
 * This class represents expression within parentheses.
 */
public class JParenthesedExpression extends JExpression 
    implements CInitializable 
{

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	expr		the expression beetwen parenthese
     */
    public JParenthesedExpression( TokenReference where, JExpression expr ) {
	super( where );
	this.expr = expr;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return "(" + expr.toString() + ")";
    }
    /**
     * Returns a list of expressions known to be non-null (null) in this context
     */
    public /*@pure non_null*/ Object[] getFANonNulls(/*@non_null@*/ CContextType context) {
	    return expr.getFANonNulls(context);
    }
    public /*@pure non_null*/ Object[] getFANulls(/*@non_null@*/ CContextType context) {
	    return expr.getFANulls(context);
    }

 
    /**
     * Indicates whether this expression can appear on the left-hand
     * side of an assignment or in an increment or decrement
     * operation.
     *
     * <pre><jml>
     * also 
     * ensures \result <==> expr.isMaybeInitializable();
     * </jml></pre>
     *
     */
    public /*@ pure @*/ boolean isMaybeInitializable() {
	return expr.isMaybeInitializable();
    }

    /**
     * Delegates to <code>expr</code>.  */
    public void initialize(CContextType context) throws UnpositionedError
    {
	((CInitializable) expr).initialize(context);
    }

    /**
     * Delegates to <code>expr</code>.  */
    public boolean isLValue(CExpressionContextType context)
    {
	return ((CInitializable) expr).isLValue(context);
    }

    /**
     * Delegates to <code>expr</code>.  */
    public boolean isDefinitelyAssigned(CContextType context)
    {
	return ((CInitializable) expr).isDefinitelyAssigned(context);
    }

    /**
     * Delegates to <code>expr</code>.  */
    public void genStartStoreCode(CodeSequence code)
    {
	((CInitializable) expr).genStartStoreCode(code);
    }

    /**
     * Delegates to <code>expr</code>.  */
    public void genEndStoreCode(CodeSequence code, boolean discard)
    {
	((CInitializable) expr).genEndStoreCode(code, discard);
    }

    public JExpression expr() { return expr; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return expr.getType();
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	return expr.isNonNull(context);
    }
    /**
     * Returns true iff the variable, field or method represente by this expression
     * was declared non_null.
     */
    public /*@pure*/ boolean isDeclaredNonNull() {
	return expr.isDeclaredNonNull();
    }


    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	expr = expr.typecheck( context );
	if( expr.isConstant() || expr instanceof JThisExpression ) {
	    return expr;
	}

	// From a code-generation standpoint we could just return expr
	// since the AST encodes precedence.  But we may want to
	// pretty-print a typechecked AST and the parens are need
	// there.
	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitParenthesedExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	expr.genCode(code);
    }

    /**
     * Generates a sequence of bytecodes to branch on a label
     * This method helps to handle heavy optimizables conditions
     * @param	code		the code list
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
	expr.genBranch(cond, code, label);
    }

    /**
     * Returns this with outer parentheses removed, i.e., returns the
     * results of recursively calling <Code>unParenthesize</code> on
     * the <code>JExpression</code> passed to the constructor of this.  */
    public JExpression unParenthesize() {
	return expr.unParenthesize();
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    //@ spec_public
    private JExpression expr;
}
