/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JPhylum.java,v 1.13 2008/05/14 18:16:24 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.MessageDescription;

/**
 * This class is a superclass for all elements of the parsing tree.
 * That is, all classes for nodes of the AST descend from this class.
 */
public abstract class JPhylum extends org.multijava.util.compiler.Phylum 
    implements Constants, Cloneable {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * construct an element of the parsing tree
     * @param where the token reference of this node
     */
    public JPhylum(/*@ non_null */ TokenReference where) {
	super(where);
    }

    // ----------------------------------------------------------------------
    // ERROR HANDLING
    // ----------------------------------------------------------------------

    /**
     * Adds a compiler error.
     * Redefine this method to change error handling behaviour.
     * @param	context		the context in which the error occurred
     * @param	description	the message ident to be displayed
     * @param	params		the array of parameters
     *
     */
    //@ assignable \nothing;
    protected void fail(/*@non_null*/ CContextType context, 
			/*@non_null*/ MessageDescription description, 
			/*@nullable*/ /*@\readonly \readonly@*/ Object[] params)
	throws PositionedError
    {
	context.reportTrouble(new PositionedError(getTokenReference(), 
						  description, params));
    }

    /**
     * Adds a compiler error.
     * Redefine this method to change error handling behaviour.
     * @param	context		the context in which the error occurred
     * @param	description	the message ident to be displayed
     */
    //@ assignable \nothing;
    protected void fail(/*@non_null*/ CContextType context, 
			/*@non_null*/ MessageDescription description)
	throws PositionedError
    {
	fail(context, description, new Object[] { });
    }

    /**
     * Adds a compiler error.
     * Redefine this method to change error handling behaviour.
     * @param	context		the context in which the error occurred
     * @param	description	the message ident to be displayed
     * @param	param		the parameter
     *
     */
    //@ assignable \nothing;
    protected void fail(/*@non_null*/ CContextType context, 
			/*@non_null*/ MessageDescription description, 
			/*@\readonly@*/ Object param)
	throws PositionedError
    {
	fail(context, description, new Object[] { param });
    }

    /**
     * Verifies that the condition is true; otherwise adds an error.
     * @param	context		the context in which the check occurred
     * @param	cond		the condition to verify
     * @param	description	the message ident to be displayed
     * @param	params		the array of parameters
     */
    //@ assignable \nothing;
    public final void check(/*@non_null*/ CContextType context, boolean cond, 
			    /*@non_null*/ MessageDescription description, 
			    /*@\readonly \readonly@*/ Object[] params)
	throws PositionedError
    {
	if (!cond) {
	    fail(context, description, params);
	}
    }

    /**
     * Verifies that the condition is true; otherwise adds an error.
     * @param	context		the context in which the check occurred
     * @param	cond		the condition to verify
     * @param	description	the message ident to be displayed
     * @param	param1		the first parameter
     * @param	param2		the second parameter
     */
    //@ assignable \nothing;
    public final void check(/*@non_null*/ CContextType context, boolean cond, 
			    /*@non_null*/ MessageDescription description, 
			    /*@\readonly@*/ Object param1, /*@\readonly@*/ Object param2)
	throws PositionedError
    {
	if (!cond) {
	    fail(context, description, new Object[] { param1, param2 });
	}
    }

    /**
     * Verifies that the condition is true; otherwise adds an error.
     * @param	context		the context in which the check occurred
     * @param	cond		the condition to verify
     * @param	description	the message ident to be displayed
     * @param	param		the parameter
     */
    //@ assignable \nothing;
    public final void check(/*@non_null*/ CContextType context, boolean cond, 
			    /*@non_null*/ MessageDescription description,
                            /*@\readonly@*/ Object param)
	throws PositionedError
    {
	if (!cond) {
	    fail(context, description, new Object[] { param });
	}
    }

    /**
     * Verifies that the condition is true; otherwise adds an error.
     * @param	context		the context in which the check occurred
     * @param	cond		the condition to verify
     * @param	description	the message ident to be displayed
     */
    //@ assignable \nothing;
    public final void check(/*@non_null*/ CContextType context, boolean cond, 
			    /*@non_null*/ MessageDescription description)
	throws PositionedError
    {
	if (!cond) {
	    fail(context, description, null);
	}
    }

    /**
     * Gives a warning message.
     * @param	context	the context in which the warning occurred
     * @param	description the message ident to be displayed
     * @param	params	the array of parameters
     */
    //@ assignable \nothing;
    public final void warn(/*@non_null*/ CContextType context, 
			   /*@non_null*/ MessageDescription description,
			   /*@nullable*/ /*@\readonly \readonly@*/ Object[] params)
	throws PositionedError
    {
	context.reportTrouble(new CWarning(getTokenReference(), description, 
					   params));
    }

    /**
     * Gives a warning message.
     * @param	context	the context in which the warning occurred
     * @param	description the message ident to be displayed
     */
    //@ assignable \nothing;
    public final void warn(/*@non_null*/ CContextType context, 
			   /*@non_null*/ MessageDescription description)
	throws PositionedError
    {
	warn(context, description, null);
    }

    /**
     * Gives a warning message.
     * @param	context	the context in which the warning occurred
     * @param	description the message ident to be displayed
     * @param	param the parameter
     */
    //@ assignable \nothing;
    public final void warn(/*@non_null*/ CContextType context, 
			   /*@non_null*/ MessageDescription description,
			   /*@\readonly@*/ Object param)
	throws PositionedError
    {
	warn(context, description, new Object[] { param });
    }

    /**
     * Gives a warning message.
     * @param	context	the context in which the warning occurred
     * @param	description the message ident to be displayed
     * @param	param1 the parameter
     * @param	param2 the parameter
     */
    //@ assignable \nothing;
    public final void warn(/*@non_null*/ CContextType context, 
			   /*@non_null*/ MessageDescription description,
			   /*@\readonly@*/ Object param1,
			   /*@\readonly@*/ Object param2)
	throws PositionedError
    {
	warn(context, description, new Object[] { param1, param2 });
    }

    // ----------------------------------------------------------------------
    // JAVA CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public abstract void accept(/*@non_null@*/ MjcVisitor p);

    // ----------------------------------------------------------------------
    // PUBLIC CONSTANTS
    // ----------------------------------------------------------------------
    
    public static final /*@non_null@*/ JPhylum[] EMPTY = new JPhylum[0];
}
