/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JPrefixExpression.java,v 1.7 2005/03/30 03:45:45 xi_hao Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.IincInstruction;
import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.classfile.MethodRefInstruction;

/**
 * This class represents prefix increment and decrement expressions. 
 * JLS 15.13.2 Prefix Increment Operator <code>++</code>.
 * JLS 15.13.3 Prefix Decrement Operator <code>--</code>.
 */
public class JPrefixExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	oper		the operator
     * @param	expr		the operand
     */
    public JPrefixExpression( TokenReference where, 
			      int oper, JExpression expr ) {
	super( where );
	this.oper = oper;
	this.expr = expr;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public int oper() { return oper; }
    public JExpression expr() { return expr; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return type;
    }

    /**
     * Returns true iff this expression can be used as a statement (JLS 14.7)
     */
    public boolean isStatementExpression() {
	return true;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	isInSafeMathMode  = context.arithmeticMode().equals(AMID_SAFE_MATH);
	expr = expr.typecheck( context, true, true, context.discardValue() );
	check( context, 
	       expr.getType().isNumeric(), 
	       MjcMessages.PREFIX_BADTYPE, expr.getType() );
	check( context,
	       expr.isMaybeInitializable()
	       && ((CInitializable)expr).isLValue( context ),
	       MjcMessages.PREFIX_NOTLVALUE );
	check( context,
	       ((CInitializable)expr).isDefinitelyAssigned( context ),
	       MjcMessages.PREFIX_NOTINITIALIZED );

	type = expr.getType();

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitPrefixExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());
	boolean		discardValue = code.discardValue();

	int			val = oper == OPE_PREINC ? 1 : -1;

	if ((expr.getType() == CStdType.Integer) &&
	    (expr instanceof JLocalVariableExpression)&&(!isInSafeMathMode)) {

	    code.plantInstruction(
		new IincInstruction(
		    ((JLocalVariableExpression)expr).getPosition(), val));
	    code.setDiscardValue(discardValue);
	    expr.genCode(code);
	    code.setDiscardValue(false);

	} else {
	    code.setDiscardValue(false);

	    ((CInitializable)expr).genStartStoreCode(code);

	    expr.genCode(code);
	    switch (expr.getType().getTypeID()) {
	    case TID_FLOAT:
		code.plantInstruction(new PushLiteralInstruction((float)val));
		code.plantInstruction(new NoArgInstruction(opc_fadd));
		break;
	    case TID_LONG:
		code.plantInstruction(new PushLiteralInstruction((long)val));
		if (isInSafeMathMode){
		    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "add",
                                                                   "(JJ)J"));
		}else{
		    code.plantInstruction(new NoArgInstruction(opc_ladd));
		}
		break;
	    case TID_DOUBLE:
		code.plantInstruction(new PushLiteralInstruction((double)val));
		code.plantInstruction(new NoArgInstruction(opc_dadd));
		break;

	    case TID_BYTE:
	    case TID_CHAR:
	    case TID_SHORT:
	    case TID_INT:
		code.plantInstruction(new PushLiteralInstruction(val));
		if (isInSafeMathMode){
                    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
                                                                   "org/multijava/mjc/SafeIntegralArithmetic",
                                                                   "add",
                                                                   "(II)I"));
                }else{
                    code.plantInstruction(new NoArgInstruction(opc_iadd));
                }
		switch (expr.getType().getTypeID()) {
		case TID_BYTE:
		    code.plantInstruction(new NoArgInstruction(opc_i2b));
		    break;
		case TID_CHAR:
		    code.plantInstruction(new NoArgInstruction(opc_i2c));
		    break;
		case TID_SHORT:
		    code.plantInstruction(new NoArgInstruction(opc_i2s));
		}
	    }

	    ((CInitializable)expr).genEndStoreCode(code, discardValue);
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------
    private boolean isInSafeMathMode  = false;
    protected int		oper;
    protected JExpression	expr;
    protected CType		type;
}
