/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JRealLiteral.java,v 1.12 2004/01/16 23:21:11 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents real-valued literals (float, double)
 */
public class JRealLiteral extends JNumberLiteral {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param where  the line of this node in the source code
     * @param image  the string representation of this literal
     */
    public JRealLiteral(TokenReference where, String image) {
	super(where);
	this.image = image;
    }

    /**
     * Construct a node in the parsing tree
     */
    public JRealLiteral(TokenReference where, Number value, CType type) {
	super(where);
	assertTrue(type.isFloatingPoint());
	setValue(value);
	this.type = type;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public void negate() {
	if (image != null) {
	    image = "-" + image;
	}
	if (value != null) {
	    value = (type == CStdType.Float) 
		? (Number) new Float( -value.floatValue() )
		: (Number) new Double( -value.doubleValue() );
	}
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	calculateType();
	return type;
    }

    /**
     * Get a Number that represents the value of this literal.
     */
    public /*@ pure @*/ Number numberValue() {
	return value != null ? value : valueFromImage();
    }

    /**
     * Get a Number that represents the value of this literal.
     */
    public Object getValue() {
	return numberValue();
    }

    /**
     * Sets the value of this to the given value.
     */
    private void setValue(Number newValue) {
	value = newValue;
    }

    /**
     * Sets the value of this from the image.
     */
    private /*@ pure @*/ Double valueFromImage() 
	throws NumberFormatException 
    {
	return Double.valueOf( image );
    }

    /**
     * Sets the type of this from the image.
     */
    private void calculateType() {
	if( type == null ) {
	    // The only way for type to be null is if the constructor
	    // that takes an image is used.
	    if( image.endsWith("f") || image.endsWith("F") ) {
		type = CStdType.Float;
	    } else {
		type = CStdType.Double;
	    }
	}
    }

    /**
     * Does this expression represent a real literal?
     */
    public boolean isRealLiteral() {
	return true;
    }    

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
     {

	if (value != null) {
	    return this; // already checked
	}

	try {
	    setValue( valueFromImage() );
	} catch ( NumberFormatException ex ) {
	    check( context, false, MjcMessages.LITER_DOUBLE_BAD, image );
	}

	calculateType();

	return this;
    }

    /**
     * convertType
     * changes the type of this expression to an other
     * @param  dest the destination type
     * @param	context	the context in which this expression appears
     */
    public JExpression convertType( CType dest, CExpressionContextType context ) 
	throws PositionedError 
    {
	// trying to follow JLS 5.1.3
	// Why is there no check for type == null here like in JOrdinalLiteral?

	JExpression result;

	if( type.equals(dest) ) {
	    result = this;
	} else if( dest.isOrdinal() ) {
	    long v = value.longValue();
	    // Adjust (narrow) value, if required:
	    v = dest == CStdType.Integer ? (int) v
		: dest == CStdType.Short ? (short) v
		: dest == CStdType.Byte  ? (byte) v
		: dest == CStdType.Char  ? (char) v
		: v;
	    result = new JOrdinalLiteral( getTokenReference(),
					  v,
					  (CNumericType) dest );
	} else if( dest.isFloatingPoint() ) {
	    Number n = (dest == CStdType.Float) 
		? (Number) new Float( type == CStdType.Float
				      ? value.floatValue()
				      : (float) value.doubleValue())
		: (Number) new Double( type == CStdType.Float
				       ? (double) value.floatValue()
				       : value.doubleValue());
	    result = new JRealLiteral(getTokenReference(), n, dest);
	} else {
	    fail( "Attempting to convert floating point literal to " + dest );
	    result = null;
	} 
	return result;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitRealLiteral( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	if( type == CStdType.Float) {
	    code.plantInstruction(new PushLiteralInstruction(value.floatValue()));
	} else {
	    code.plantInstruction(new PushLiteralInstruction(value.doubleValue()));
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private String	image;
    private Number	value;
    private CType	type;
}
