/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JRelationalExpression.java,v 1.10 2004/01/16 23:21:11 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents the AST node for the relational operators,
 * <code>&lt;</code>, <code>&gt;</code>, etc. JLS2 15.20 */
public class JRelationalExpression extends JBinaryExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	oper		the operator
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JRelationalExpression( TokenReference where,
				  int oper,
				  JExpression left,
				  JExpression right ) {
	super( where, left, right );
	this.oper = oper;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /*@ pure @*/ public String opString() {
	String s = "";
	switch (oper) {
	  case OPE_LT:
	    s = " < ";
	    break;
	  case OPE_LE:
	    s = " <= ";
	    break;
	  case OPE_GT:
	    s = " > ";
	    break;
	  case OPE_GE:
	    s = " >= ";
	    break;
	  default:
	    fail();
        }
	return s;
    }

    public String toString() {
	return left.toString() + opString() + right.toString();
    }
    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	left = left.typecheck(context);
	right = right.typecheck(context);
	return typecheck_helper(context);
    }

    /** This is just used by subclasses to call super class type checking
	when the arguments are already typechecked.
    */
    protected JExpression typecheck_helper( CExpressionContextType context ) 
	throws PositionedError 
    {
	check( context, 
	       left.getType().isNumeric() && right.getType().isNumeric(), 
	       MjcMessages.RELATIONAL_TYPE, left.getType(), right.getType());

	type = CStdType.Boolean;

	promote( context );

	if( left.isConstant() && right.isConstant() ) {
	    return constantFolding( left.getNumberLiteral(), 
				    right.getNumberLiteral() );
	}

	return this;
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the left literal
     * @param	right		the right literal
     * @return	a literal resulting of an operation over two literals
     */
    public JExpression constantFolding( JNumberLiteral left, 
					JNumberLiteral right ) {
	if (left.getType() == CStdType.Long) {
	    long leftVal = left.numberValue().longValue();
	    long rightVal = right.numberValue().longValue();
	    switch (oper) {
	      case OPE_LT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal < rightVal);
	      case OPE_LE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal <= rightVal);
	      case OPE_GT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal > rightVal);
	      case OPE_GE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal >= rightVal);
	      default:
		fail();
		return null;
	    }
	} else if (left.getType() == CStdType.Float) {
	    float leftVal = left.numberValue().floatValue();
	    float rightVal = right.numberValue().floatValue();
	    switch (oper) {
	      case OPE_LT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal < rightVal);
	      case OPE_LE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal <= rightVal);
	      case OPE_GT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal > rightVal);
	      case OPE_GE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal >= rightVal);
	      default:
		fail();
		return null;
	    }
	} else if (left.getType() == CStdType.Double) {	
	    double leftVal = left.numberValue().doubleValue();
	    double rightVal = right.numberValue().doubleValue();
	    switch (oper) {
	      case OPE_LT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal < rightVal);
	      case OPE_LE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal <= rightVal);
	      case OPE_GT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal > rightVal);
	      case OPE_GE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal >= rightVal);
	      default:
		fail();
		return null;
	    }
	} else if (left.getType().implicitlyCastTo(CStdType.Integer)) {
	    int leftVal = left.numberValue().intValue();
	    int rightVal = right.numberValue().intValue();
	    switch (oper) {
	      case OPE_LT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal < rightVal);
	      case OPE_LE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal <= rightVal);
	      case OPE_GT:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal > rightVal);
	      case OPE_GE:
		return new JBooleanLiteral(getTokenReference(),
					   leftVal >= rightVal);
	      default:
		fail();
		return null;
	    }
	} else {
	    fail();
	    return null;
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitRelationalExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	genBooleanResultCode(code);
    }

    /**
     * Optimize a bi-conditional expression
     */
    protected void genBranch(JExpression left,
			     JExpression right,
			     boolean cond,
			     CodeSequence code,
			     CodeLabel label)
    {
	code.setLineNumber(getTokenReference().line());

	// Perform constant folding if one of the two compared values
	// is a constant 0 and both are ints or longs (with floats and
	// doubles we have to worry about NaN)
	if (((left instanceof JOrdinalLiteral
	      && ((JOrdinalLiteral)left).numberValue().intValue() == 0) 
	     ||
	     (right instanceof JOrdinalLiteral 
	      && ((JOrdinalLiteral)right).numberValue().intValue() == 0)) 
	    && left.getType().isCodedAsInt()) {
	    int opcode = -1;
	    if (left instanceof JOrdinalLiteral 
		&& ((JOrdinalLiteral)left).numberValue().intValue() == 0) {
		right.genCode(code);
		switch (oper) {
		case OPE_LT:
		    opcode = cond ? opc_ifgt : opc_ifle;
		    break;
		case OPE_LE:
		    opcode = cond ? opc_ifge : opc_iflt;
		    break;
		case OPE_GT:
		    opcode = cond ? opc_iflt : opc_ifge;
		    break;
		case OPE_GE:
		    opcode = cond ? opc_ifle : opc_ifgt;
		    break;
		}
	    } else {
		left.genCode(code);
		switch (oper) {
		case OPE_LT:
		    opcode = cond ? opc_iflt : opc_ifge;
		    break;
		case OPE_LE:
		    opcode = cond ? opc_ifle : opc_ifgt;
		    break;
		case OPE_GT:
		    opcode = cond ? opc_ifgt : opc_ifle;
		    break;
		case OPE_GE:
		    opcode = cond ? opc_ifge : opc_iflt;
		    break;
		}
	    }
	    code.plantLabelRef(opcode, label);
	    return;
	}

	left.genCode(code);
	right.genCode(code);

	int opcode = -1;
	switch (left.getType().getTypeID()) {
	case TID_FLOAT:
	    // Use float comparison that fails in the "right way" for NaN
	    // (see JVM2, 7.5)
	    opcode = (oper == OPE_LT || oper == OPE_LE) ? 
		opc_fcmpg : opc_fcmpl;
	    code.plantInstruction(new NoArgInstruction(opcode));
	    break;
	case TID_LONG:
	    code.plantInstruction(new NoArgInstruction(opc_lcmp));
	    break;
	case TID_DOUBLE:
	    // Use double comparison that fails in the "right way" for NaN
	    // (see JVM2, 7.5)
	    opcode = (oper == OPE_LT || oper == OPE_LE) ? 
		opc_dcmpg : opc_dcmpl;
	    code.plantInstruction(new NoArgInstruction(opcode));
	    break;
	default:
	    switch (oper) {
	    case OPE_LT:
		opcode = cond ? opc_if_icmplt : opc_if_icmpge;
		break;
	    case OPE_LE:
		opcode = cond ? opc_if_icmple : opc_if_icmpgt;
		break;
	    case OPE_GT:
		opcode = cond ? opc_if_icmpgt : opc_if_icmple;
		break;
	    case OPE_GE:
		opcode = cond ? opc_if_icmpge : opc_if_icmplt;
		break;
	    }
	    code.plantLabelRef(opcode, label);
	    return;
	}
	switch (oper) {
	case OPE_LT:
	    opcode = cond ? opc_iflt : opc_ifge;
	    break;
	case OPE_LE:
	    opcode = cond ? opc_ifle : opc_ifgt;
	    break;
	case OPE_GT:
	    opcode = cond ? opc_ifgt : opc_ifle;
	    break;
	case OPE_GE:
	    opcode = cond ? opc_ifge : opc_iflt;
	    break;
	}

	code.plantLabelRef(opcode, label);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public int oper() { return oper; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	int			oper;
}
