/*
 * Copyright (C) 2004 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JResendExpression.java,v 1.18 2008/03/16 15:52:54 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CompilationAbortedError;

/**
 * This class represents a MultiJava resend expression, for invoking a
 * directly overridden method of the caller's generic function.  That is:
 *
 * <CODE>e0.resend(e1,...en)</CODE>
 *
 * @author Curtis Clifton
 * @version $Revision: 1.18 $
 */
public class JResendExpression extends JMethodCallExpression {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public JResendExpression( TokenReference where, JExpression prefix,
			      JExpression[] args ) {
	super( where, prefix, "resend", args );
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks args and returns an array of the argument types.
     *
     * <pre><jml>
     * also 
     * requires context != null && args != null;
     * modifiable args, callingMethod;
     * ensures callingMethod != null;
     * ensures \result != null && \result.length == args.length;
     * </jml></pre>
     *
     */
    protected CType[] checkArgumentTypes( CExpressionContextType context ) 
	throws PositionedError
    {
	callingMethod = (CSourceMethod) context.getCMethod();

	// -------------------------------------------------------------
	// check the receiver
	if (prefix != null) {
	    check( context, !callingMethod.isStatic(), 
		   MjcMessages.STATIC_RESEND_HAS_RECEIVER );
	    check( context, prefix instanceof JThisExpression && 
		   ((JThisExpression) prefix).prefix() == null,
		   MjcMessages.BAD_RESEND_RECEIVER );
	}

	// -------------------------------------------------------------
	// check the argument positions corresponding to specialized
	// formal parameters of the caller
	JMethodDeclaration callingMethodDecl = 
	    (JMethodDeclaration) callingMethod.declarationASTNode();

	if (callingMethodDecl == null) {
	    final String msg = "We're sorry but you've encountered a bug " +
		"in mjc.  Please report this to www.multijava.org.  Thanks.";
	    throw new CompilationAbortedError( msg );
	}
	
	final JFormalParameter[] callingMethodParms = 
	    callingMethodDecl.parameters();

	check( context, args.length == callingMethodParms.length,
	       MjcMessages.RESEND_TO_DIFFERENT_GF );

	for (int i = 0; i < callingMethodParms.length; i++) {
	    JFormalParameter parm = callingMethodParms[i];
	    /* !FIXME!resend 
	       if (parm.isSpecialized()) { */
	    String parmIdent = parm.ident();
		
	    // specialized formals must be final
	    check( context, parm.isFinal(),
		   parm.isSpecialized() ? MjcMessages.RESEND_PARM_MUST_BE_FINAL
		   : MjcMessages.TEMP_RESEND_PARM_MUST_BE_FINAL, 
		   parmIdent );
		
	    // resend arguments in these positions must be name
	    // expressions naming the formal (using
	    // context.reportTrouble in this block so that error
	    // messages can be positioned based on argument
	    // expressions.)
	    JExpression arg = args[i];
	    if (!(arg instanceof JNameExpression)) {
		context.reportTrouble( 
		    new PositionedError( arg.getTokenReference(),
					 parm.isSpecialized() ? 
					 MjcMessages.RESEND_PARM_MUST_BE_NAME :
					 MjcMessages.TEMP_RESEND_PARM_MUST_BE_NAME,
					 parmIdent ) );
	    } else {
		JNameExpression nameArg = (JNameExpression) arg;
		if (!(nameArg.getName().equals(parmIdent) && 
		      nameArg.getPrefix() == null)) {
		    context.reportTrouble( 
			new PositionedError( arg.getTokenReference(),
					     parm.isSpecialized() ? 
					     MjcMessages.RESEND_PARM_MUST_BE_NAME :
					     MjcMessages.TEMP_RESEND_PARM_MUST_BE_NAME,
					     parmIdent ) );
		}
	    }
	    /* !FIXME!resend 
	       } */
	}

	return super.checkArgumentTypes(context);
    }

    /**
     * Makes sure any external generic functions that are implicitly
     * imported are actually available in the data structures.  
     *
     * <pre><jml>
     * also 
     * requires callingMethod != null && context != null;
     * </jml></pre>
     *
     */
    protected void resolveExternalMethods( CExpressionContextType context ) {
	context.resolveMaybeExtMethodRef( callingMethod.ident() );
    }

    /**
     * Returns the apparent target method for this call.  (We say apparent 
     * because it may be that some synthetic redirection is needed in the
     * generated code to account for compilation techniques of MultiJava or
     * Java workarounds for the JVM's lack of support for inner classes.)
     *
     * <pre><jml>
     * also 
     * requires callingMethod != null;
     * </jml></pre>
     *
     */
    protected CMethod lookupMethod( CType[] argTypes, 
				    CExpressionContextType context ) 
	throws PositionedError, UnpositionedError
    {
	this.RMJ = context.getCompiler().RMJ();

	CMethod result;

	CMethodSet possibleTargets = callingMethod.overriddenMethodSet();

	// filter out interface methods so we don't give a spurious
	// error below
	CMethodSet overriddenMethodSet = new CMethodSet();
	if (possibleTargets != null) {
	    CMethodSet.Iterator iter =
		possibleTargets.iterator();
	    while (iter.hasNext()) {
		CMethod m = iter.next();
		if (!(m.isAbstract() && m.owner().isInterface())) {
		    overriddenMethodSet.addMethod(m);
		}
	    }
	}

	int size = overriddenMethodSet.size();
	
	if (size == 0) {
	    if (callingMethod.isStatic()) {
		throw new PositionedError( getTokenReference(),
					   MjcMessages.RESEND_FROM_TOP_STATIC );
	    } else {
		throw new PositionedError( getTokenReference(),
					   MjcMessages.RESEND_FROM_TOP );
	    }
	} else if (size == 1) {
	    result = overriddenMethodSet.getFirst();
	} else {
	    // XXX In RMJ-land, it's possible for glue to come along
	    // and fix the problem with the ambiguity, but for now,
	    // let's not allow that.  (An analogous issue exists with
	    // regular super.)
	    throw new PositionedError( getTokenReference(),
				       MjcMessages.AMBIGUOUS_RESEND, 
				       overriddenMethodSet.getMethod(0),
				       overriddenMethodSet.getMethod(1) );
	}
	
	// -------------------------------------------------------------
	// check the applicability given that arbitrary expressions
	// can appear in argument positions not corresponding to
	// specialized formal parameters of the caller
	CType receiverType = callingMethod.owner().getType();
	//  by checks in checkArgumentTypes() above we have:
	//x@ assert prefix != null ==> prefix.getType() == receiverType;
	// !FIXME! method 'getType()' is not pure (benevolent side-effects)
	// so it cannot be used in an assertion.
	check( context, result.isApplicable( result.ident(), receiverType,
					     argTypes,CClassType.EMPTY ),
	       MjcMessages.RESEND_TO_DIFFERENT_GF );

	// -------------------------------------------------------------
	// tweak the prefix for a special implementation case
	if ( !result.isInExternalGF() 
	     && callingMethod.owner() != result.owner() 
	     && callingMethod.owner().descendsFrom(result.owner()) ) {
	    // case o-i2 from Clifton-01, section 3.4.2
	    prefix = 
		new JSuperExpression( getTokenReference() ).
		typecheck( context, false, false, false );
	}

	return result;
    }


    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Plants the code for putting a reference to the appropriate
     * dispatcher in the chain-of-responsibility onto the stack.
     *
     * <pre><jml>
     * also
     * requires topConcreteMethod.isExternal();
     * requires (* code already contains instructions for leaving any
     *             necessary prefix object on the stack *);
     * </jml></pre>
     *
     */
    protected void genExternalTargetRef( CodeSequence code ) {
	if (method.host() == callingMethod.host()) {
	    /* if in same host as callingMethod, we can assume method
	     * is a source method, so we can ask for its dispatcher */
	    if (((CSourceMethod)method).dispatcherMethod() != 
		callingMethod.dispatcherMethod()) {
		/* RMJ: added the test for same dispatcher methods,
		 * because RMJ separates external methods with the same
		 * host into different dispatcher methods, while MJ
		 * doesn't.  this test suffices for both languages. */
		// if they're different dispatchers, then we need to
		// treat it like a regular call, but skipping the
		// current method
		callingMethod.plantOldFunctionRef( code, false, false );
	    } else {
		// method.host() == callingMethod.host() ==>
		//   Clifton-01, case o-e3;
		// method.owner() == callingMethod.owner() ==>
		//   method.host() == callingMethod.host() ==>
		//   Clifton-01, case o-e1;
		// will use the current dispatcher object, so no reference
		// to the chain-of-responsibility is needed
		code.plantLoadThis();
	    }
	    code.plantSwap();
	} else if (!callingMethod.isExternal() || RMJ) {
	    // Clifton-01, case o-e2
	    // get function reference from local method
	    callingMethod.plantOldFunctionRef( code, false, false );
	    code.plantSwap();
	} else {
	    super.genExternalTargetRef( code );
	}
    }

    /**
     * Generates the bytecode for invoking the actual target method.
     * Assumes code sequence already contains code for pushing all the
     * necessary arguments onto the stack.  This is a helper method of
     * {@link #genCode(CodeSequence)}.
     */
    protected void genInvocationOpcode( CodeSequence code ) {
	if (method.host() == callingMethod.host() &&
	    /* if in same host as callingMethod, we can assume method
	     * is a source method */
	    ((CSourceMethod)method).dispatcherMethod() == 
	    callingMethod.dispatcherMethod()) {
	    /* RMJ: see comment above */
	    String runtimeHostName = null;
	    if (topConcreteMethod.isExternal()) {
		// Clifton-01, case o-e1
		// (callingMethod instanceof CSourceMethod) &&
		// method.owner() == callingMethod.owner()
		// ==>
		// method instanceof CSourceMethod
		runtimeHostName = 
		    ((CSourceMethod) method).dispatcherClassName();
	    } else {
		// Clifton-01, case o-i1
		runtimeHostName = method.getOwnerName();
	    }
	    
	    // generate call directly to ident$body method
	    method.plantMMInvocation( code, runtimeHostName );
	} else {
	    super.genInvocationOpcode(code);
	}
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * The method signature for the method from which this resend
     * originates, initialized during typechecking.
     */
    //@ spec_public
    private CSourceMethod callingMethod;

    // whether we're compiling RMJ
    private boolean RMJ = false;
}
