/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JReturnStatement.java,v 1.13 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents a return statement in the AST.
 * JLS 14.15: Return Statement
 */
public class JReturnStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	expr		the expression to return, null for 
     *				<code>return;</code> in a void method
     * @param	comments	the comments associated with this statement
     */
    public JReturnStatement( TokenReference where, JExpression expr, 
			     JavaStyleComment[] comments ) {
	super( where, comments );
	this.expr = expr;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the type of this return statement
     */
    public CType getType() {
	return expr != null ? expr.getType() : CStdType.Void;
    }

    public JExpression expr() { return expr; }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	check( context, !context.isInInitializer(),
	       MjcMessages.RETURN_IN_INIT );

	CType returnType = 
	    context.getMethodContext().getCMethod().returnType();

	if( expr != null ) {
	    check( context, returnType != CStdType.Void, 
		   MjcMessages.RETURN_NONEMPTY_VOID);

	    CExpressionContextType expContext = 
		context.createExpressionContext();

	    expr = expr.typecheck( expContext );
	    check( context,
		   expr.isAssignableTo( returnType ),
		   MjcMessages.RETURN_BADTYPE,
		   expr.getType().isTypeVariable()?((CTypeVariable)expr.getType()).getIdent():
		       (expr.getType().toVerboseString()), // WMD
		   returnType.isTypeVariable()?((CTypeVariable)returnType).getIdent():
		   returnType.toVerboseString()); // WMD
	    expr = expr.convertType(returnType, expContext);
	    // check nullity
	    if (CTopLevel.getCompiler().nonNullTypes()) {
               checkNullity(context);
	    }
	} else {
	    check( context, returnType == CStdType.Void, 
		   MjcMessages.RETURN_EMPTY_NONVOID );
	}
	context.setReachable(false);
    }

    public void checkNullity( CContextType context )
	throws PositionedError 
    {
	final JMemberDeclaration memberDecl = ((CSourceMethod)context.getMethodContext().getCMethod()).declarationASTNode();
        final JMethodDeclaration methodDecl = (JMethodDeclaration) memberDecl;
        
        final boolean isDeclaredNonNull = methodDecl.isDeclaredNonNull();
        final boolean isExpressionNonNull = expr.isNonNull(context);
	check( context, 
               CTypeNullity.isAssignable(methodDecl, expr, context),
	       (methodDecl.returnType().isArrayType()      ?
	        MjcMessages.RETURN_ARRAY_NULLITY_INVARIANT :
	        MjcMessages.RETURN_CAN_BE_NULL ));
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitReturnStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	if (expr != null) {
	    expr.genCode(code);

	    code.plantReturn(this);
	    code.plantInstruction(new NoArgInstruction(expr.getType().getReturnOpcode()));
	} else {
	    code.plantReturn(this);
	    code.plantInstruction(new NoArgInstruction(opc_return));
	}
    }

    /**
     * Load the value from a local var (after finally)
     */
    public void load(CodeSequence code, JLocalVariable var) {
	code.plantLocalVar(expr.getType().getLoadOpcode(), var);
    }

    /**
     * Load the value from a local var (after finally)
     */
    public void store(CodeSequence code, JLocalVariable var) {
	code.plantLocalVar(expr.getType().getStoreOpcode(), var);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected JExpression		expr;
}
