/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JShiftExpression.java,v 1.6 2004/01/16 23:21:11 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the shift (left, right, boolean-right) binary
 * expressions.  */
public class JShiftExpression extends JBinaryArithmeticExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	oper		the operator
     * @param	left		the left operand
     * @param	right		the right operand
     */
    public JShiftExpression( TokenReference where,
			     int oper,
			     JExpression left,
			     JExpression right ) {
	super( where, left, right );
	this.oper = oper;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	String s = "";
	switch (oper) {
	  case OPE_SL:
	    s = " << ";
	    break;
	  case OPE_SR:
	    s = " >> ";
	    break;
	  case OPE_BSR:
	    s = " >>> ";
	    break;
	  default:
	    fail();
	}
	return left.toString() + s + right.toString();
    }

    
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	left = left.typecheck( context );
	right = right.typecheck( context );

	JExpression expr = typecheckBinaryArith( context );
	if( expr == this ) {
	    left = left.convertType( type, context );
	    right = right.convertType( CStdType.Integer, context );
	}
	return expr;
    }

    /**
     * Typechecks a binary arithmetic expression and handles simple
     * optimizations like constant folding.
     *
     * @param context the actual context of analysis
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if checks fail */
    public JExpression typecheckBinaryArith( CExpressionContextType context ) 
	throws PositionedError 
    {
	try {
	    type = computeType( left.getType(), right.getType() );
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}

	if( left.isConstant() && right.isConstant() ) {
	    return constantFolding();
	} else {
	    return this;
	}
    }

    /**
     * compute the type of this expression according to operands
     * @param	leftType		the type of left operand
     * @param	rightType		the type of right operand
     * @return	the type computed for this binary operation
     * @exception	UnpositionedError	this error will be positioned soon
     */
    public static CType computeType(CType leftType, CType rightType) 
	throws UnpositionedError 
    {
	//if (rightType == CStdType.Long)
	// !FIXME! check the spec  throw new UnpositionedError(MjcMessages.shift-badtype, leftType, rightType);

	if( leftType.isOrdinal() && rightType.isOrdinal() ) {
	    return leftType = ((CNumericType)leftType).unaryPromote();
	}

	throw new UnpositionedError (MjcMessages.SHIFT_BADTYPE, 
				     leftType, rightType );
    }

    /**
     * Converts two constant expressions into a single literal that is
     * the result of applying this operation to the expressions.
     *
     * @return a literal resulting from an operation over two literals */
    public JExpression constantFolding() {
	if( left.getType() == CStdType.Long ) {
	    long result;

	    if( right.getType() == CStdType.Long ) {
		result = 
		    compute( left.getNumberLiteral().numberValue().longValue(),
			     (int) right.getNumberLiteral().numberValue()
			           .longValue() );
	    } else {
		result = 
		    compute(left.getNumberLiteral().numberValue().longValue(),
			    right.getNumberLiteral().numberValue().intValue());
	    }
	    return new JOrdinalLiteral( getTokenReference(), 
					new Long(result), CStdType.Long);
	} else {
	    int	result;

	    if( right.getType() == CStdType.Long ) {
		result = 
		    compute( left.getNumberLiteral().numberValue().intValue(),
			     (int) right.getNumberLiteral().numberValue()
			           .longValue());
	    } else {
		result = 
		    compute(left.getNumberLiteral().numberValue().intValue(),
			    right.getNumberLiteral().numberValue().intValue());
	    }
	    return new JOrdinalLiteral(	getTokenReference(), 
					new Integer(result), CStdType.Integer);
	}
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public int compute(int left, int right) {
	switch (oper) {
	case OPE_SL:
	    return left << right;
	case OPE_SR:
	    return left >> right;
	case OPE_BSR:
	    return left >>> right;
	default:
	    fail();
	    return 0;
	}
    }

    /**
     * Computes the result of the operation at compile-time (JLS 15.27).
     * @param	left		the first operand
     * @param	right		the seconds operand
     * @return	the result of the operation
     */
    public long compute(long left, int right) {
	switch (oper) {
	case OPE_SL:
	    return left << right;
	case OPE_SR:
	    return left >> right;
	case OPE_BSR:
	    return left >>> right;
	default:
	    fail();
	    return 0L;
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitShiftExpression( this );
    }


    /**
     * @param	type		the type of result
     * @return	the type of opcode for this operation
     */
    public static int getOpcode(int oper, CType type) {
	if (type == CStdType.Long) {
	    switch (oper) {
	    case OPE_SL:
		return opc_lshl;
	    case OPE_SR:
		return opc_lshr;
	    case OPE_BSR:
		return opc_lushr;
	    }
	} else {
	    switch (oper) {
	    case OPE_SL:
		return opc_ishl;
	    case OPE_SR:
		return opc_ishr;
	    case OPE_BSR:
		return opc_iushr;
	    }
	}
	return -1;
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	left.genCode(code);
	right.genCode(code);

	code.plantInstruction(new NoArgInstruction(getOpcode(oper, getType())));
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public int oper() { return oper; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected	int			oper;
}
