/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JStatement.java,v 1.10 2006/12/20 06:15:52 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.MessageDescription;

/**
 * This class is the root class for all classes representing statement 
 * nodes in the AST.  See JLS 14.4.
 */
public abstract class JStatement extends JPhylum {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param where 	the line of this node in the source code
     * @param comments	the Java comments that go with this statement
     */
    public JStatement( /*@ non_null */ TokenReference where, /*@ non_null */ JavaStyleComment[] comments ) {
	super( where );
	this.comments = comments;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public abstract void typecheck( /*@non_null@*/ CFlowControlContextType context ) 
	throws PositionedError;

    /**
     * Adds a compiler error.
     * @param	context		the context in which the error occurred
     * @param	key		the message ident to be displayed
     * @param	params		the array of parameters
     *
     */
    protected void fail( CContextType context, 
			 MessageDescription key, 
			 /*@ nullable @*/ Object[] params )
	throws PositionedError
    {
	throw new CLineError(getTokenReference(), key, params);
    }

    // ----------------------------------------------------------------------
    // BREAK/CONTINUE HANDLING
    // ----------------------------------------------------------------------

    /**
     * Returns true iff this statement is a possible target for a
     * break statement.  This method is overridden in classes that
     * accept breaks.  */
    public /*@ pure @*/ boolean acceptsBreak() {
	return false;
    }

    /**
     * Returns true iff this statement is a possible target for a continue statement.
     * This method is overridden in classes that accept continues.
     */
    public /*@ pure @*/ boolean acceptsContinue() {
	return false;
    }

    /**
     * Returns a label at end of this statement (for break statement)
     */
    public CodeLabel getBreakLabel() {
	throw new RuntimeException("NO END LABEL");
    }

    /**
     * Returns the beginning of this block (for continue statement)
     */
    public CodeLabel getContinueLabel() {
	throw new RuntimeException("NO CONTINUE LABEL");
    }

    /**
     * Registers that this statement is the target of a break
     * statement.  This default implementation throws an
     * <code>UnsupportedOperationException</code> since only some
     * sorts statements are valid break targets.  The classes
     * representing those sorts of statements override this method.  */
    public void addBreak(CFlowControlContextType context) {
	throw new UnsupportedOperationException("CANNOT BREAK");
    }

    /**
     * Registers that this statement is the target of a continue
     * statement.  This default implementation throws an
     * <code>UnsupportedOperationException</code> since only some
     * sorts statements are valid continue targets.  The classes
     * representing those sorts of statements override this method.  */
    public void addContinue(CFlowControlContextType context) {
	throw new UnsupportedOperationException("CANNOT CONTINUE");
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public abstract void genCode( /*@ non_null */ CodeSequence code);

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(/*@non_null@*/ MjcVisitor p) {
//  !FIXME! 	if (comments != null) {
//  	    p.visitComments(comments);
//  	}
    }

    /**
     * Returns the comments
     */
    public JavaStyleComment[] getComments() {
	return comments;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    private /*@ non_null */ JavaStyleComment[]	comments;
}
