/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JStringLiteral.java,v 1.12 2006/08/04 18:22:39 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.compiler.TokenReference;

/**
 * A simple character constant
 */
public class JStringLiteral extends JLiteral {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	image		the string representation of this literal
     */
    public JStringLiteral(TokenReference where, String image) {
	this(where, image, false);
    }

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	image		the string representation of this literal
     * @param	escaped		there are escapes in the image
     */
    public JStringLiteral( TokenReference where, String image, 
			   boolean escaped) {
	super(where);

	if (image == null) {
	    throw new RuntimeException();
	}

	if (escaped) {
	    value = unescapeString(image);
	} else {
	    value = image;
	}
    }
    
    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return "\"" + escapeString(value, false) + "\"";
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return CStdType.String;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	return value!=null;
    }

    /**
     * get a Number that represents the value of this literal
     * Do not call before typecheck
     */
    public Object getValue() {
	return value;
    }

    /**
     * get a String that represents the value of this literal
     * Do not call before typecheck
     */
    public /*@ pure @*/ String stringValue() {
	return value;
    }

    /**
     * Returns the literal value of an expression
     */
    public boolean isDefault() {
	return false;
    }

    /**
     * Does this expression represent a string literal?
     */
    public boolean isStringLiteral() {
	return true;
    }

    
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
//  !FIXME! 	StringBuffer s = new StringBuffer();
//  	for (int i = 0; i < value.length(); i++) {
//  	    char c = value.charAt(i);
//  	    switch (c) {
//  	    case '\n' : s.append("\\n"); break;
//  	    case '\r' : s.append("\\r"); break;
//  	    case '\t' : s.append("\\t"); break;
//  	    case '\b' : s.append("\\b"); break;
//  	    case '\f' : s.append("\\f"); break;
//  	    case '\"' : s.append("\\\""); break;
//  	    case '\'' : s.append("\\\'"); break;
//  	    case '\\' : s.append("\\\\"); break;
//  	    default:
//  		s.append(c);
//  	    }
//  	}
//  	value = s.toString();
//  	p.visitStringLiteral(value);
	p.visitStringLiteral( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());
	code.plantInstruction(new PushLiteralInstruction(value));
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	String		value;
}
