/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JSuperExpression.java,v 1.10 2005/12/22 20:27:22 darvasa Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a "super" primary expression in an AST. See
 * JLS2 15.11.2, 15.12.
 * @author Curtis Clifton
 * @version $Id: JSuperExpression.java,v 1.10 2005/12/22 20:27:22 darvasa Exp $
 */
public class JSuperExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node for an unqualified super expression.
     * @param	where		the line of this node in the source code
     */
    public JSuperExpression( TokenReference where ) {
	super( where );
    }

    /**
     * Construct a qualified "super" expression node in the AST,
     * where only the qualifying expression is known, not its
     * type.  Note that to be semantically correct <code>prefix</code>
     * must be a <code>JNameExpression</code> where the qualified name
     * denoted by <code>prefix</code> is a type name [JLS2 6.5.1].
     *
     * @param	where		the line of this node in the source code
     * @param	prefix		the left expression like <code>T</code> 
     * in <code>T.super</code> */
    public JSuperExpression( TokenReference where, JExpression prefix ) {
	this( where );

	// We just store the prefix initially.  It will be checked
	// during typechecking.  It is possible that the parser will
	// discard this object and extract the prefix if it discovers
	// that the "this" that this object was created for is really
	// an explicit this-constructor invocation.
	this.prefix = prefix;
    }
    
    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression prefix() { return prefix; }

    /**
     * Returns the type of this expression.  Only valid after typechecking.
     * @return	the type of this expression
     */
    public CType getType() {
	return self.getType();
    }


    /**
     * Returns the local subclass of the class to which this super
     * expression refers.  For an unqualified super expression, the
     * result is just the owner of this expression.  For a qualified
     * super expression the result is the class named by the
     * qualifying expression.  */
    public CClass targetSubclass() {
	return sub;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {

	    // get the context of the surrounding class
	CClassContextType classContext = context.getClassContext();

	check( context, !context.isStatic(), MjcMessages.BAD_SUPER_STATIC );

	// Set self to be the CClass singleton that is the target of this super
	if( prefix != null ) {
	    // prefix must represent a type name, JLS2 6.5.1, "A name is
	    // syntactically classified as a type name as the qualifying
	    // type of a qualified this expression."
	    check( context, prefix instanceof JNameExpression, 
		   MjcMessages.SUPER_BADACCESS );

	    try {
		CClassType prefixType = 
		    CTopLevel.getTypeRep( ((JNameExpression) prefix)
					  .qualifiedName().replace('.', '/'),
					  false );
		prefixType = (CClassType)(prefixType.checkType(context));
		sub = prefixType.getCClass();
		self = sub.getSuperClass();
	    } catch (UnpositionedError e) {
		throw e.addPosition( getTokenReference() );
	    }
	    
	} else {
	    sub = classContext.getOwnerClass();
	    self = sub.getSuperClass();
	}

	if( classContext.getOwnerClass() == sub ) {
	    // the target of this "super" expression is the superclass
	    // of the logical owner of this context

		// make sure the super constructor has already been called
		// else the programmer is trying to use 'super' before
		// the new object is properly constructed
	    check( context,
		   !context.isBeforeSuperConstructorCall(),
		   MjcMessages.CONSTRUCTOR_EXPLICIT_CALL );
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitSuperExpression(this);
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	// This expression only survives into code generation if this
	// is an unqualified super.  The method of field access
	// expression handles the "super send" in those cases, we just
	// need to put this on the stack.
	code.plantLoadThis();
    }

    /*
	 * By rzakhejm
	 */
	public String toString() {
			return (prefix() != null ? prefix().toString() : sub.toString()) + ".super";
	}

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /*@ private invariant self != null ==> 
      @			sub != null && sub.descendsFrom(self);
      @*/

    /**
     * The class to which this super expression refers.
     */
    private CClass self;

    /**
     * The local subclass of the class to which this super expression
     * refers.  For an unqualified super expression, <code>sub</code>
     * is just the owner of this expression.  For a qualified super
     * expression <code>sub</code> is the class named by the
     * qualifying expression.  */
    private CClass sub;

    private	JExpression	prefix;
}
