/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JSwitchGroup.java,v 1.6 2004/01/16 23:21:11 cclifton Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents an AST node for a group in a switch
 * statement.  A group is a set of case labels and a sequence of
 * statements for those labels.  */
public class JSwitchGroup extends JPhylum {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	labels		a group of labels
     * @param	stmts		a group of statements
     */
    public JSwitchGroup( TokenReference where, 
			 JSwitchLabel[] labels, 
			 JStatement[] stmts ) {
	super( where );

	this.labels = labels;
	this.stmts = stmts;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns a list of statements
     */
    public JStatement[] getStatements() {
	return stmts;
    }

    public JSwitchLabel[] labels() { return labels; }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the group and mutates the context to record
     * information gathered during typechecking.
     *
     * <pre><jml>
     * requires context != null && context.isReachable();
     * </jml></pre>
     *
     * @param context	the context for the control flow in which the 
     *			statements of the group are checked
     * @exception PositionedError if the check fails */
    public void typecheck( CSwitchGroupContext context ) 
	throws PositionedError 
    {
	assertTrue( context.isReachable() );
	
	// This method may be called multiple times for multiple
	// control flow paths.  We only need to check the labels once.
	if (!labelsChecked) {
	    labelsChecked = true;
	    for( int i = 0; i < labels.length; i++ ) {
		labels[i].typecheck( context );
	    }
	}

	for( int i = 0; i < stmts.length; i++ ) {
	    try {
		if( !context.isReachable() ) {
		    throw new CLineError( stmts[i].getTokenReference(), 
					  MjcMessages.STATEMENT_UNREACHABLE );
		}
		stmts[i].typecheck( context );
	    } catch( CLineError le ) {
		context.reportTrouble( le );
	    }
	}

	return;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitSwitchGroup( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	matches			a list of values to match
     * @param	targets			a list of target labels
     */
    public void collectLabels( CodeLabel deflab, 
			       ArrayList matches, ArrayList targets ) {
	pos = null;

	// check if one of the labels is "default:"
	for (int i = 0; pos == null && i < this.labels.length; i++) {
	    if (this.labels[i].isDefault()) {
		pos = deflab;
	    }
	}

	if (pos == null) {
	    // no default: define a new label
	    pos = new CodeLabel();

	    for (int i = 0; i < this.labels.length; i++) {
		matches.add(this.labels[i].getLabel());
		targets.add(pos);
	    }
	}
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	code.plantLabel(pos);
	for (int i = 0; i < stmts.length; i++) {
	    stmts[i].genCode(code);
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JSwitchLabel[]	labels;

    /**
     * Flag indicates whether the labels for this group have been
     * typechecked.  */
    private boolean labelsChecked = false;

    private JStatement[]	stmts;
    private CodeLabel		pos;
}
