/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JSwitchLabel.java,v 1.5 2005/12/23 18:14:20 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.MessageDescription;

/**
 * This class represents an AST node for the label for a single case of a
 * switch statement.
 */
public class JSwitchLabel extends JPhylum {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * This method is directly called by the parser.
     * @param	where		the line of this node in the source code
     * @param	expr		the expression (null if default label)
     */
    public JSwitchLabel( TokenReference where, JExpression expr ) {
	super( where );

	this.expr = expr;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * @return	true if this label is a "default:" label
     */
    public boolean isDefault() {
	return expr == null;
    }

    /**
     * @return	the value of this label
     */
    public Integer getLabel() {
	return label;
    }

    public JExpression expr() { return expr; }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the group and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context		the context for the entire switch in which the 
     *				labels are checked
     * @exception PositionedError if the check fails */
    public void typecheck( CSwitchGroupContext context ) 
	throws PositionedError 
    {
	try {
	    if( expr != null ) {
		expr = expr.typecheck( context.createExpressionContext() );
		check( context,
		       expr.getType().implicitlyCastTo( CStdType.Integer ),
		       MjcMessages.SWITCH_LABEL_EXPR_NOTINT, expr.getType() );

		check( context, 
		       expr.isConstant(), 
		       MjcMessages.SWITCH_LABEL_EXPR_NOTCONST );
		context.addLabel( expr.getOrdinalLiteral() );
		label = new Integer( expr.getOrdinalLiteral()
				     .numberValue().intValue() );
	    } else {
		context.addDefault();
	    }
	} catch( UnpositionedError cue ) {
	    throw cue.addPosition( getTokenReference() );
	}
    }

    /**
     * Adds a compiler error.
     * @param	context		the context in which the error occurred
     * @param	key		the message ident to be displayed
     * @param	params		the array of parameters
     *
     */
    protected void fail( CContextType context, MessageDescription key, 
			 /*@ nullable @*/ Object[] params )
	throws PositionedError
    {
	throw new CLineError( getTokenReference(), key, params );
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitSwitchLabel( this );
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private Integer		label;
    private JExpression		expr;
}
