/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JSwitchStatement.java,v 1.9 2008/03/16 15:52:54 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.SwitchInstruction;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.JavaStyleComment;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class represents a switch statement in the AST.
 * JLS 14.9: Switch Statement
 */
public class JSwitchStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	expr		the expr part
     * @param	groups		the cases of the switch
     * @param	comments	the comments associated with this statement
     */
    public JSwitchStatement( TokenReference where,
			     JExpression expr,
			     JSwitchGroup[] groups,
			     JavaStyleComment[] comments )
    {
	super( where, comments );

	this.expr = expr;
	this.groups = groups;
	this.hasBreak = false;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	expr = expr.typecheck( context.createExpressionContext() );
	check( context, expr.getType().implicitlyCastTo(CStdType.Integer), 
	       MjcMessages.SWITCH_BADEXPR);

	CSwitchBodyContext self = 
	    context.createSwitchBodyContext( this, expr.getType() );
	CSwitchGroupContext groupContext = null;

	// thread through the groups and check them
	for( int i = 0; i < groups.length; i++ ) {
	    // for each group we create a context the considers
	    // potential fall through from the previous group
	    groupContext = 
		self.createSwitchGroupContext(groupContext, 
					      groups[i].getTokenReference());

	    // now we typecheck the group
	    groups[i].typecheck( groupContext );

	    // issue warning on fall through, except for last group
	    if (groupContext.isReachable() && i != groups.length - 1) {
		context.reportTrouble( 
		     new CWarning( getTokenReference(),
				   MjcMessages.CASE_FALL_THROUGH ));
	    }
	}

	if (!self.defaultExists()) {
	    // if there is no default group then there exist a control
	    // flow path that skips all groups
	    groupContext = 
		self.createSwitchGroupContext(groupContext, getTokenReference());
	    context.reportTrouble(new CWarning(getTokenReference(),
					       MjcMessages.SWITCH_NO_DEFAULT));
	}

	self.adoptSwitchGroupContext( groupContext );

	if (hasBreak) {
	    if (self.isReachable()) {
		// last context fell through so we need to merge the
		// definite assignment info in self with the
		// breakContext
		self.merge( breakContext );
	    } else {
		// Last context did not fall through, but if it had a
		// break that targets this switch then we already have
		// its definite assignment data in breakContext.  If
		// it had a return or something else to skip the end
		// of this statement then we don't care about its
		// definite assignment data (as the target of the
		// abnormal termination should be worried about that).
		// So we adopt the breakContext info here.
		self.adopt( breakContext );
	    }
	    self.setReachable(true);
	}

	self.checkingComplete();
    }

    // ----------------------------------------------------------------------
    // BREAK/CONTINUE HANDLING
    // ----------------------------------------------------------------------

    /**
     * Returns true iff this statement is a possible target for a break statement.
     */
    public boolean acceptsBreak() {
	return true;
    }

    /**
     * Returns true iff this statement is a possible target for a continue statement.
     */
    public boolean acceptsContinue() {
	return false;
    }

    /**
     * Registers that this statement is the target of a break
     * statement.  */
    public void addBreak(CFlowControlContextType context) {
	hasBreak = true;
	if (breakContext == null) {
	    breakContext = context.cloneContext();
	} else {
	    breakContext.merge(context);
	}
    }

    /**
     * Return the end of this block (for break statement)
     */
    public CodeLabel getBreakLabel() {
	return endLabel;
    }

    /**
     * Return the beginning of this block (for continue statement)
     */
    public CodeLabel getContinueLabel() {
	fail();
	return null;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitSwitchStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	CodeLabel defaultLabel = new CodeLabel();

	ArrayList matches = new ArrayList();
	ArrayList targets = new ArrayList();

	for (int i = 0; i < groups.length; i++) {
	    groups[i].collectLabels(defaultLabel, matches, targets);
	}

	expr.genCode(code);

	code.pushContext(this);
	code.plantInstruction(new SwitchInstruction(defaultLabel, matches, targets));
	for (int i = 0; i < groups.length; i++) {
	    groups[i].genCode(code);
	}

	if (!defaultLabel.hasAddress()) {
	    code.plantLabel(defaultLabel);
	}

	code.plantLabel(endLabel);
	code.popContext(this);

	
	

	endLabel = null;
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression expr() { return expr; }
    public JSwitchGroup[] groups() { return groups; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		expr;
    private JSwitchGroup[]	groups;
    private boolean		hasBreak;
    private CodeLabel		endLabel = new CodeLabel();

    /**
     *  Stores the  definite assignment  information for  control flow
     * paths that reach the end of the switch via a break statement.  */
    private CFlowControlContextType breakContext;
}
