/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JSynchronizedStatement.java,v 1.6 2006/12/10 22:51:21 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.HandlerInfo;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents a synchronized statement in an AST.
 * A synchronized statement acquires a mutual-exclusion lock on behalf
 * of the executing thread, executes a block, then releases the lock.
 * JLS2 14.18: Synchronized Statement
 */
public class JSynchronizedStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	cond		the expression to evaluate
     * @param	body		the loop body
     * @param	comments	the comments associated with this statement
     */
    public JSynchronizedStatement( TokenReference where,
				   JExpression cond,
				   JBlock body,
				   JavaStyleComment[] comments )
    {
	super( where, comments );
	this.cond = cond;
	this.body = body;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) throws PositionedError {
	localVar = new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF,
						0,
						CStdType.Integer,
						"sync" + toString() /*unique*/,
						null );
	try {
	    context.addVariable( localVar );
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}
	cond = cond.typecheck( context.createExpressionContext() );
	check( context, 
	       cond.getType().isReference(), 
	       MjcMessages.SYNCHRONIZED_NOTREFERENCE );
	body.typecheck( context );
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitSynchronizedStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genMonitorExit(CodeSequence code) {
	code.plantLocalVar(opc_aload, localVar);
	code.plantInstruction(new NoArgInstruction(opc_monitorexit));
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	CodeLabel		nextLabel = new CodeLabel();

	cond.genCode(code);
	code.plantInstruction(new NoArgInstruction(opc_dup));
	code.plantLocalVar(opc_astore, localVar);
	code.plantInstruction(new NoArgInstruction(opc_monitorenter));

	code.pushContext(this);
	int		startPC = code.getPC();
	body.genCode(code);
	code.popContext(this);
	genMonitorExit(code);
	int		endPC = code.getPC();
	code.plantLabelRef(opc_goto, nextLabel);

	int	errorPC = code.getPC();
	genMonitorExit(code);
	code.plantInstruction(new NoArgInstruction(opc_athrow));

	code.plantLabel(nextLabel);			//	next:	...

	// protect
	code.addExceptionHandler(new HandlerInfo(code.getInstructionAt(startPC),
						 code.getInstructionAt(endPC),
						 code.getInstructionAt(errorPC),
						 null));
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression cond() { return cond; }
    public JBlock body() { return body; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		cond;
    private JBlock		body;
    private JLocalVariable	localVar;
}
