/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JThisExpression.java,v 1.14 2006/12/10 22:51:21 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * A 'this' expression
 */
public class JThisExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct an unqualified "this" expression node in the parsing tree.
     * @param where the line of this node in the source code
     */
    public JThisExpression( TokenReference where ) {
	super( where );
    }

    /**
     * Construct a qualified "this" expression node in the parsing
     * tree, where the actual <code>CClass</code> singleton of the
     * qualifier is known.  This is an abuse for referencing inherited
     * fields or methods without qualifiers.  The compiler actually
     * cooks "this" expressions for such references, passing (in
     * <code>self</code>) the <em>declaring</em> class of the
     * inherited field or method.  When typechecking this "this" we
     * need to consider this abuse.  Briefly, this is abusive because
     * within a class <code>B</code> that extends <code>A</code> the
     * expression <code>A.this</code> is illegal in regular Java.
     *
     * @param	where		the line of this node in the source code
     * @param self the class onto which this suffix is applied */
    public JThisExpression( TokenReference where, CClass self ) {
	this( where );
	this.self = self;
	this.generated = true;
	// WMD TODO Do we get a problem from direct access to CClass???
    }

    /**
     * Construct a qualified "this" expression node in the parsing
     * tree, where only the qualifying expression is known, not its
     * type.  Note that to be semantically correct <code>prefix</code>
     * must be a <code>JNameExpression</code> where the qualified name
     * denoted by <code>prefix</code> is a type name [JLS2 6.5.1].
     *
     * @param	where		the line of this node in the source code
     * @param	prefix		the left expression like <code>T</code> 
     * in <code>T.this</code> */
    public JThisExpression( TokenReference where, JExpression prefix ) {
	this( where );

	// We just store the prefix initially.  It will be checked
	// during typechecking.  It is possible that the parser will
	// discard this object and extract the prefix if it discovers
	// that the "this" that this object was created for is really
	// an explicit this-constructor invocation.
	this.prefix = prefix;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression prefix() { return prefix; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
        if( selfType == null ) {
            selfType = self.getType();
        } 
        
	return selfType;
    }

    /**
     * @return is this expression a lvalue ?
     */
    public boolean isLValue(CExpressionContextType context) {
	return false;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared java expression (see 
     *		<code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( final CExpressionContextType context ) 
	throws PositionedError 
    {
	// get the context of the surrounding class
	CClassContextType classContext = context.getClassContext();

	// determine if this "this" refers to the the logical receiver
	// passed as a synthetic parameter for methods of an external
	// generic function
	boolean directlyInExtMethod = false;
	try {
	    directlyInExtMethod = context.getCMethod().isInExternalGF();
	} catch (UnpositionedError e) {
	    throw e.addPosition( getTokenReference() );
	}
	    
	// Set self to be the CClass singleton that is the target of this this
	if( prefix != null ) {
	    // prefix must represent a type name, JLS2 6.5.1, "A name is
	    // syntactically classified as a type name as the qualifying
	    // type of a qualified this expression."
	    check( context, prefix instanceof JNameExpression, 
		   MjcMessages.THIS_BADACCESS );

	    try {
                if( CTopLevel.getCompiler().universeChecks() ) {
                    // WMD outer this is readonly
                    selfType = 
                        CTopLevel.getTypeRep( ((JNameExpression) prefix)
					  .qualifiedName().replace('.', '/'),
                                          CUniverseImplicitReadonly.getUniverse(),
					  false );
                } else {
                    selfType = 
                        CTopLevel.getTypeRep( ((JNameExpression) prefix)
					  .qualifiedName().replace('.', '/'),
                                          CUniverseImplicitPeer.getUniverse(),
					  false );                   
                }

		selfType = (CClassType)(selfType.checkType(context));
		self = selfType.getCClass();
	    } catch (UnpositionedError e) {
		throw e.addPosition( getTokenReference() );
	    }
	    
	} else if( self == null ) {
	    self = classContext.getOwnerClass();
            selfType = self.getType();
	}
	
	if( classContext.getOwnerClass() == self 
	    || (generated 
		&& classContext.getOwnerClass().descendsFrom( self )) ) {
	    // the target of this "this" expression is the logical
	    // owner of this context

	    // make sure the reference is not from a static method,
	    // since "this" is undefined in such a method
	    check( context, !context.isStatic(), MjcMessages.BAD_THIS_STATIC );

		// make sure the super constructor has already been called
		// else the programmer is trying to use the new object before
		// it is properly constructed
	    check( context, !context.isBeforeSuperConstructorCall(),
		   MjcMessages.CONSTRUCTOR_EXPLICIT_CALL );

	    if( directlyInExtMethod ) {
		// for external methods we must change the "this"
		// reference to a local variable reference
		return makeThisVariableReference( context );
	    } else {
		// no changes necessary for regular methods
		return this;
	    }
	}

	// The target of the "this" expression must be an outer class
	// (or else an error)
	JExpression expr;	// the result expression
	CClass clazz;		// the CClass of expr.getType()

	if( context.isInConstructor() ) {
	    // Within a constructor the outer-this field is not
	    // necessarily initialized, so we begin with a reference
	    // to the outer-this parameter.
	    expr = null;
	    try {
		expr = 
		    new JNameExpression( getTokenReference(), JAV_OUTER_THIS ).
			typecheck( context );
	    } catch (PositionedError e) {
		throw new PositionedError( getTokenReference(),
					   MjcMessages.THIS_UNKNOWN_OUTER, 
					   self.ident() );
	    }
	    //@ assert expr != null;
	} else {
	    expr = 
		new JThisExpression( getTokenReference() ).typecheck(context);
	}

	clazz = expr.getType().getCClass();

	// Climb out through the outer classes looking for the
	// qualifier of this "this".  Check for static contexts along
	// the way and build the appropriate expression to get at the
	// target object.
	while( clazz != self 
	       && !(/*generated &&*/ clazz.descendsFrom( self ))) {

	    check( context, 
		   classContext != null 
		   && classContext.getHostClass().owner() != null,
		   MjcMessages.THIS_UNKNOWN_OUTER, self.ident() );

	    CClass currentClass = classContext.getOwnerClass();
	    
	    check( context, 
		   !currentClass.access().isStatic(),
		   MjcMessages.THIS_INVALID_OUTER, currentClass.ident() );

	    classContext = classContext.getParentContext().getClassContext();
	
	    // build the expression to reference the new outer this
	    expr = new JClassFieldExpression( getTokenReference(), 
					      expr, JAV_OUTER_THIS );

	    // We expect that this expression should typecheck, since
	    // we just built it.  However, we need to make the call to
	    // resolve some references in the expression that will be
	    // used during code generation
	    expr = expr.typecheck( context );
	    clazz = expr.getType().getCClass();
	    // assert( clazz == classContext.getOwnerClass() ); // !FIXME! 
	}

	return expr;
    }

    /**
     * Generates a synthetic variable reference to a local variable
     * that will resolve to "this" at run time.  This synthetic
     * reference is used for "this" expressions that appear in
     * methods of external gfs.
     *
     * @param	context		the context in which the declaration appears,
     *				used to get correct variable slot number for
     *				code generation
     * @return	a synthetic variable reference expression */
    private JExpression makeThisVariableReference( CContextType context ) 
    {
	JGeneratedLocalVariable local = 
	    new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF, 0, 
					 getType(), "toto", 
					 null ) {
		    /**	
		     * @return the local index in the context variable
		     * table */
		public int getPosition() {
			// we put the receiver as the first argument for
			// methods of external gfs, to ease code generation
// !FIXME!codegen parameter shuffling problem for multimethods bodies that become static
		    return 1;
		}
	    };
	
	return new JLocalVariableExpression( getTokenReference(), local ) 
	    {
		public JExpression typecheck( CExpressionContextType ctxt) {
			// no checking required
		    return this;
		}
	    };
    }


    public boolean equals(/*@ nullable @*/  Object o ) {
	return o instanceof JThisExpression &&
	    self.equals(((JThisExpression)o).self);
    }

    /**
     * Throws a runtime exception because hashCode isn't properly
     * implemented yet. */
    public int hashCode() {
	fail( "An instance of " + getClass() + " cannot be stored in "
	      + " a HashTable or HashSet because equals is overridden but "
	      + " hashCode is not" );
	return 0;
    }

    public String toString() {
	return (self != null ? self.toString() + "." :
		(prefix != null) ? prefix.toString() + "." : "") + "this";
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitThisExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	if (!code.discardValue()) {
	    code.setLineNumber(getTokenReference().line());

	    code.plantLoadThis();
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	CClass		self;
    
    // WMD: safe the CClassType too, b/c it must be readonly
    // for outer this expressions
    private     CClassType      selfType;
    
    private	JExpression	prefix;

    /**
     * Indicates whether this this expression was generated by the
     * compiler.  */
    private boolean generated = false;
}
