/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTryCatchStatement.java,v 1.9 2004/06/28 13:52:05 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.Iterator;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import java.util.Hashtable;
import java.util.HashSet;
import java.util.Set;
import java.util.Collections;

/**
 * This class represents a try-catch statement in the AST.
 * JLS2 14.19: Try Statement
 *
 * <P>A try statement executes a block.  If a value is thrown and the
 * try statement has one or more catch clauses that can catch it, then
 * control will be transferred to the first such catch clause.  If the
 * try statement has a finally clause, then another block of code is
 * executed, no matter whether the try block completes normally or
 * abruptly, and no matter whether a catch clause is first given
 * control.</P>
 *
 * <P>In this implementation, the Try Statement is split into a
 * Try-Catch Statement and a Try-Finally Statement. A Try Statement
 * where both catch and finally clauses are present is rewritten as a
 * Try-Catch Statement enclosed in a Try-Finally Statement.</P>  */
public class JTryCatchStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	tryClause	the body
     * @param	catchClauses	the catch clause of this statement
     * @param	comments	the comments associated with this statement
     */
    public JTryCatchStatement( TokenReference where,
			       JBlock tryClause,
			       JCatchClause[] catchClauses,
			       JavaStyleComment[] comments )
    {
	super( where, comments );

	this.tryClause = tryClause;
	this.catchClauses = catchClauses;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	check( context, catchClauses.length > 0, MjcMessages.TRY_NOCATCH );

	// It is possible for all of the try-block to be skipped;
	// e.g., an unchecked exception may be thrown on the first
	// expression.
	CFlowControlContextType	catchEntryContext = context.cloneContext();

	// Create a context for typechecking the try-block.  This
	// context will collect the set of exceptions that could be
	// thrown within the try-block.  This is needed for checking
	// the catch-clauses.
	CTryContext tryContext = 
	    context.createTryContext( getTokenReference() );

	tryClause.typecheck( tryContext );

	for (int i = 0; i < catchClauses.length; i++) {
	    catchClauses[i].checkInterface( context );
	}
	
	// The context for checking the clauses must include the
	// information from the try-block if it can complete
	// successfully or just from skipping the try-block otherwise.
	// The context for completion of the whole try-catch statement
	// must include the definite assignment information for the
	// try-block if it can complete successfully.
	if (tryContext.isReachable()) {
	    context.adopt(tryContext);
	    catchEntryContext.merge(tryContext);
	}

	boolean isReachable = tryContext.isReachable();
	Iterator iter = tryContext.throwables().iterator();
    loop:
	while( iter.hasNext() ) {
	    CThrowableInfo info = (CThrowableInfo)iter.next();
	    CClassType type = info.getThrowable();

	    for( int i = 0; i < catchClauses.length; i++ ) {
		if( type.isAlwaysAssignableTo( catchClauses[i].getType() ) 
		    || catchClauses[i].getType().isAlwaysAssignableTo( type )) {
		    // The catch clause could catch some or all of
		    // the possible exceptions of type "type" so
		    // mark it as reached.
		    if( !catchClauses[i].isReached() ) {
			isReachable |= checkClause( catchEntryContext, context, 
						    catchClauses[i] );
		    }
		    if (!catchClauses[i].getType().equals(type) &&
			catchClauses[i].getType().isAlwaysAssignableTo(type)) {
			// It is possible that some of the possible
			// exceptions of type "type" are not caught by
			// this catch clause, so continue checking
			// later catch clauses against this type.
			continue;
		    } else {
			// All possible remaining exceptions of type
			// "type" will be caught by this catch clause,
			// so do not check later catch clauses against
			// this type.
			continue loop;
		    }
		}
	    }
	    // not caught so pass up to the outer context
	    context.addThrowable( info );
	}
    
	// CHECK ORDER
	for( int i = 0; i < catchClauses.length; i++ ) {
	    CType earlierType = catchClauses[i].getType();
	    for( int j = i + 1; j < catchClauses.length; j++ ) {
		if( catchClauses[j].getType().
		    isAlwaysAssignableTo( earlierType ) ) {
		    throw new CBlockError(catchClauses[j].getTokenReference(), 
					  MjcMessages.CATCH_UNREACHABLE_1);
		}
	    }
	}

	// At this point unreachable catch clauses are unreachable
	// because either (1) they attempt to catch a checked
	// exception that is never thrown or (2) they catch a checked
	// exception that may be thrown but is caught by an earlier
	// catch clause.
	for( int i = 0; i < catchClauses.length; i++ ) {
	    if( !catchClauses[i].isReached() ) {
		CClassType catchType = catchClauses[i].getType();
		if( catchType.isCheckedException() 
		    && !catchType.equals(CStdType.Throwable) 
		    && !catchType.equals(CStdType.Exception)) {
		    // see if there is a thrown exception that is
		    // assignable to the type of this catch clause
		    boolean isThrown = false;
		    iter = tryContext.throwables().iterator();
		    while( iter.hasNext() ) {
			CThrowableInfo info = (CThrowableInfo)iter.next();
			CClassType excType = info.getThrowable();
			if( excType.isAlwaysAssignableTo( catchType ) ) {
			    isThrown = true;
			    break;
			}
		    }
		    if( isThrown ) {
			context.reportTrouble(
			    new CWarning( catchClauses[i].getTokenReference(), 
					  MjcMessages.CATCH_UNREACHABLE_2 ));
			// Checks the clause despite the warning,
			// since per the JLS the code is legal and we
			// need to compile it.
			isReachable |= checkClause( catchEntryContext, context, 
						    catchClauses[i] );
		    } else {
			throw
			    new CBlockError( catchClauses[i].
					     getTokenReference(), 
					     MjcMessages.CATCH_UNREACHABLE_3, 
					     catchType.toVerboseString() );
		    }
		} else {
		    isReachable |= checkClause( catchEntryContext, context, 
						catchClauses[i] );
		}
	    }
	}

	context.setReachable( isReachable );
    }


    /**
     * Checks the given catch-block, <code>clause</code>, using the
     * definite assignment information from <code>entryContext</code>.
     * Returns <code>true</code> if the catch-block can complete
     * normally.  Mutates the <CODE>exitContext</CODE> to record
     * definite assignment and throwable information.  */
    private boolean checkClause( CFlowControlContextType entryContext, 
				 CFlowControlContextType exitContext, 
				 JCatchClause clause ) 
	throws PositionedError 
    {
	CFlowControlContextType body = entryContext.cloneContext();
	clause.typecheck(body);
	  
	if( body.isReachable() ) {
	    exitContext.merge( body );
	}
    
	return body.isReachable();
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitTryCatchStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	CodeLabel		nextLabel = new CodeLabel();

	int		startPC = code.getPC();
	tryClause.genCode(code);			//		TRY CODE
	int		endPC = code.getPC();
	code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
	for (int i = 0; i < catchClauses.length; i++) {
	    catchClauses[i].genCode(code);
	    catchClauses[i].genExceptionHandler(code, startPC, endPC);
	    code.plantLabelRef(opc_goto, nextLabel);	//		GOTO next
	}
	code.plantLabel(nextLabel);			//	next:	...
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JBlock tryClause() { return tryClause; }

    public JCatchClause[] catchClauses() { return catchClauses; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JBlock		tryClause;
    private JCatchClause[]	catchClauses;
}
