/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTryFinallyStatement.java,v 1.10 2006/12/10 22:51:21 chalin Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.HandlerInfo;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.JavaStyleComment;

/**
 * This class represents a try-catch statement in the AST.
 * JLS2 14.19: Try Statement
 *
 * <P>A try statement executes a block.  If a value is thrown and the
 * try statement has one or more catch clauses that can catch it, then
 * control will be transferred to the first such catch clause.  If the
 * try statement has a finally clause, then another block of code is
 * executed, no matter whether the try block completes normally or
 * abruptly, and no matter whether a catch clause is first given
 * control.</P>
 *
 * <P>In this implementation, the Try Statement is split into a
 * Try-Catch Statement and a Try-Finally Statement. A Try Statement
 * where both catch and finally clauses are present is rewritten as a
 * Try-Catch Statement enclosed in a Try-Finally Statement.</P.  */
public class JTryFinallyStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param   where           the line of this node in the source code
     * @param   tryClause       the body
     * @param   finallyClause   the finally clause
     * @param   comments        the comments associated with this statement
     */
    public JTryFinallyStatement( TokenReference where,
                                 JBlock tryClause,
                                 JBlock finallyClause,
                                 JavaStyleComment[] comments )
    {
        super( where, comments );

        this.tryClause = tryClause;
        this.finallyClause = finallyClause;
        this.finallyLabel = new CodeLabel();
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context   the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
        throws PositionedError 
    {
        localVar = new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF,
                                                0,
                                                CStdType.Integer,
                                                "try" + toString() /*unique*/,
                                                null );
        try {
            context.addVariable( localVar );
        } catch( UnpositionedError e ) {
            throw e.addPosition( getTokenReference() );
        }
        finallyVar = new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF,
                                                  0,
                                                  CStdType.Integer,
                                                  "try_finally_" + toString(),
                                                  null );
        try {
            context.addVariable( finallyVar );
        } catch( UnpositionedError e ) {
            throw e.addPosition( getTokenReference() );
        }
        returnVar = 
            new JGeneratedLocalVariable( org.multijava.util.compiler.TokenReference.NO_REF,
                                         0,
                                         CStdType.Integer,
                                         "try_finally_ret" + toString(),
                                         null );
        try {
            context.addVariable( returnVar );
        } catch( UnpositionedError e ) {
            throw e.addPosition( getTokenReference() );
        }

		CFlowControlContextType tryContext = context.cloneContext();
        tryClause.typecheck( tryContext );

        CFlowControlContextType finallyContext =
        		context.createFinallyContext(tryContext, getTokenReference());
		finallyClause.typecheck( finallyContext );
		finallyContext.checkingComplete();
	}

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * @return the finally label or null
     */
    public CodeLabel getFinallyLabel() {
        return finallyLabel;
    }

    /**
     * Accepts the specified visitor
     * @param   p               the visitor
     */
    public void accept(MjcVisitor p) {
        p.visitTryFinallyStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param   code            the code list
     */
    public void genCode(CodeSequence code) {
        code.setLineNumber(getTokenReference().line());

        CodeLabel               nextLabel       = new CodeLabel();

        int             startPC = code.getPC();
        code.pushContext(this);
        tryClause.genCode(code);                        //              TRY CODE
        int             endPC = code.getPC();
        code.plantLabelRef(opc_jsr, finallyLabel);      //              JSR finally
        code.plantLabelRef(opc_goto, nextLabel);        //              GOTO next

        // HANDLER FOR OTHER EXCEPTIONS
        int     finallyPC = code.getPC();
        code.plantLocalVar(opc_astore, localVar);
        code.plantLabelRef(opc_jsr, finallyLabel);      //              JSR finally
        code.plantLocalVar(opc_aload, localVar);
        code.plantInstruction(new NoArgInstruction(opc_athrow));

        // FINALLY CODE
        code.plantLabel(finallyLabel);          //      finally:...
        code.plantLocalVar(opc_astore, finallyVar);
        generatingFinallyClause = true;
        finallyClause.genCode(code);                    //              FINALLY CLAUSE
        generatingFinallyClause = false;

        code.plantLocalVar(opc_ret, finallyVar);
        code.addExceptionHandler(new HandlerInfo(code.getInstructionAt(startPC),
                                                 code.getInstructionAt(endPC - 1 == -1 ? 0 : endPC - 1),
                                                 code.getInstructionAt(finallyPC),
                                                 null));
        //CStdType.Throwable.getCClass().qualifiedName()));
        code.plantLabel(nextLabel);                     //      next:   ...
        code.popContext(this);

        finallyLabel = null;
    }

    /**
     * Generates a sequence of bytecodes
     * @param   code            the code list
     */
    public void genFinallyCall(CodeSequence code, JReturnStatement ret) {
        // nothing to do if we are already in the finally clause
        if (generatingFinallyClause) {
            return;
        }
        
        if (ret != null && ret.getType().getSize() > 0) {
            ret.store(code, returnVar);
            code.plantLabelRef(Constants.opc_jsr, finallyLabel);
            ret.load(code, returnVar);
        } else {
            code.plantLabelRef(Constants.opc_jsr, finallyLabel);
        }
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JBlock tryClause() { return tryClause; }

    public JBlock finallyClause() { return finallyClause; }


    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private CodeLabel           finallyLabel;
    private JBlock              tryClause;
    private JBlock              finallyClause;


    /**
     * Set to true while invoking
     * <code>finallyClause.genCode()</code>, this flag is used to
     * prevent recursive <code>jsr</code>s to
     * <code>finallyClause</code> */
    private boolean generatingFinallyClause;

    private JLocalVariable      localVar;
    private JLocalVariable      finallyVar;
    private JLocalVariable      returnVar;
}
