/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTypeDeclaration.java,v 1.81 2008/05/14 18:16:22 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

import org.multijava.javadoc.JavadocComment;



import org.multijava.relaxed.util.RMJAnnotation;

/**
 * This class represents a java class or interface in the syntax tree
 */
public abstract class JTypeDeclaration extends JMemberDeclaration 
    implements JTypeDeclarationType
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a type declaration in the parsing tree.
     * @param   where           the line of this node in the source code
     * @param   modifiers       the list of modifiers of this class
     * @param   ident           the short name of this type
     * @param   typevariables   the type variables of this type
     * @param   interfaces      the names of this types's interfaces
     * @param   methods         a list of <Code>JMethodDeclarationTypes</code> 
     *                          giving the methods of this type
     * @param   inners          a list of JTypeDeclarationTypes giving the
     *                          inner classes (and interfaces) of this type
     * @param   fieldsAndInits  the fields and initializers of this type, 
     *                          passed together because the order matters 
     *                          for class and object initialization, members 
     *                          of the array should be instances of 
     *                          <code>JFieldDeclarationType</code> or 
     *                          <code>JClassBlock</code>
     * @param   javadoc         javadoc comments
     * @param   comments        regular java comments
     * @see JMethodDeclarationType
     * @see JFieldDeclarationType
     * @see JClassBlock
     */
    public JTypeDeclaration( /*@ non_null */ TokenReference where,
                             long modifiers, 
                             /*@ non_null */ String ident,
                             /*@ non_null */ CTypeVariable[] typevariables,
                             /*@ non_null */ CClassType[] interfaces,
                             /*@ non_null */ ArrayList methods,
                             /*@ non_null */ ArrayList inners,
                             /*@ non_null */ JPhylum[] fieldsAndInits,
                             /*@ nullable */ JavadocComment javadoc,
                             /*@ nullable */ JavaStyleComment[] comments ) {
        super( where, javadoc, comments );

        this.modifiers = modifiers; 
        this.ident = ident.intern();
        this.typevariables = typevariables;
        this.interfaces = interfaces;
        this.methods = methods;
        this.inners = inners;
        this.fieldsAndInits = fieldsAndInits;
        extractFields();
        
        if( interfaces != null )
        for(int i=0; i<interfaces.length; ++i)
        	interfaces[i] = (CClassType) CUniverseServices.setDefaultVariable(interfaces[i]);
    }

    /**
     * Generates a <code>CSourceClass</code> class signature singleton
     * for this declaration and then registers the signature with 
     * CTopLevel.
     *
     * @param compiler  the compiler instance for which this singleton is 
     *                  generated
     * @param owner     the class signature singleton for the logical outer 
     *                  class of this, or null if this is a top level 
     *                  declaration
     * @param host      the signature singleton of the context in which this
     *                  is declared, a <code>CCompilationUnit</code> for 
     *                  top-level declarations
     * @param prefix    the prefix prepended to this declaration's
     *                  identifier to achieve the fully qualified
     *                  name, just the package name (using '/'
     *                  separators) for top-level classes, package
     *                  name plus $-delimited outer class names plus
     *                  synthetic index for inner classes
     * @param isAnon    true if this is an anonymous class, in which 
     *                  case the fully qualified name is just 
     *                  <code>prefix</code> 
     * @param isMember  true if this is a member type, i.e., a nested type
     *                  that is not a local type or an anonymous class
     */
    public void generateInterface( /*@non_null@*/ Main compiler, CClass owner, 
                                   CMemberHost host, String prefix,
                                   boolean isAnon, boolean isMember ) 
    {
        if (owner != null) {
            if (owner.isInterface()) {
                // Nested types in an interface have some modifiers
                // implicitly, which we make explicit here.  This is done
                // here (rather than in checkInterface) so that the
                // generated signature has the same modifiers as this.
                modifiers |= ACC_PUBLIC | ACC_STATIC;
            }
            if (hasFlag(modifiers,ACC_INTERFACE)) {
                // Nested interfaces are implicitly static.
                modifiers |= ACC_STATIC;
            }
        }
        
        sourceClass = makeSignature( compiler, owner, host, prefix, isAnon, 
                                     isMember );
        setInterface( sourceClass );
        
        CClassType[] innerClasses = new CClassType[inners.size()];
        Iterator iter = inners.iterator();
        int i = 0;
        while( iter.hasNext() ) {
            JTypeDeclarationType inner = (JTypeDeclarationType) iter.next();
            inner.generateInterface( compiler, sourceClass, sourceClass,
                                     sourceClass.qualifiedName() + "$",
                                     false, true );
            innerClasses[i++] = inner.getCClass().getType();
        }
        sourceClass.setInnerClasses(innerClasses);
        uniqueSourceClass = registerSignature();
    }

    /**
     * Generates the signature object for this.  Overridden in
     * subclasses for generic function anchor declarations. 
     *
     * @param compiler  the compiler instance for which this signature is 
     *                  generated
     * @param owner     the class signature singleton for the logical outer 
     *                  class of this, or null if this is a top level 
     *                  declaration
     * @param host      the signature singleton of the context in which this
     *                  is declared, a <code>CCompilationUnit</code> for 
     *                  top-level declarations
     * @param prefix    the prefix prepended to this declaration's
     *                  identifier to achieve the fully qualified
     *                  name, just the package name (using '/'
     *                  separators) for top-level classes, package
     *                  name plus $-delimited outer class names plus
     *                  synthetic index for inner classes
     * @param isAnon    true if this is an anonymous class, in which 
     *                  case the fully qualified name is just 
     *                  <code>prefix</code> 
     * @param isMember  true if this is a member type, i.e., a nested type
     *                  that is not a local type or an anonymous class
     */
    protected /*@non_null@*/ CSourceClass makeSignature( /*@non_null@*/ Main compiler, /*@ nullable */ CClass owner, 
                                          /*@ non_null */ CMemberHost host, /*@non_null@*/ String prefix,
                                          boolean isAnon, boolean isMember ) {
        String id;              // identifier of signature
        String qualName;        // qualified name of signature

        if (isAnon) {
            // anonymous classes don't use the identifier
            qualName = prefix;

            // strip any package names from the prefix to get the identifier
            int index = prefix.lastIndexOf('/');
            id = index < 0 ? prefix : prefix.substring(index + 1);
        } else {
            id = ident;
            qualName = prefix + ident;
        }

        MemberAccess access = new MemberAccess(owner, host, modifiers);
        return new CSourceClass( compiler, access,
                                 getTokenReference(), id,
                                 qualName, typevariables ,isAnon,
                                 isMember, isDeprecated() );
    }

    /**
     * Registers the signature of this declaration with the name space
     * manager in <code>CTopLevel</code>.
     *
     * <pre><jml>
         requires sourceClass != null;
     * </jml></pre>
     *
     * @see CTopLevel#addClass(CClass) */
    protected boolean registerSignature() {
        return CTopLevel.addClass( sourceClass );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /** Returns the type variables of this class. */
    public CTypeVariable[] typevariables() {
        return typevariables;
    }

    //@ pure 
    public /*@non_null*/ CClassType[] interfaces() { 
        return interfaces==null ? 
            new CClassType[0] : interfaces;
    }

    /**
     * Returns an array of the static and instance initializers for
     * the type represented by this.  */
    //@ pure
    public /*@non_null*/ JPhylum[] fieldsAndInits() { 
        return fieldsAndInits==null ? 
            new JPhylum[0] : fieldsAndInits;
    }

    /**
     * Walks up the extends hierarchy and adds all methods
     * to the returned ArrayList.
     *
     * @return          a ArrayList that includes the CMethods for all 
     *                  methods in this class and all direct ancestors
     */
    //@ pure 
    public /*@non_null*/ ArrayList getAllMethods() {
        ArrayList allMethods = new ArrayList();
        getCClass().getAllMethods(allMethods);
        return allMethods;
    }

    //@ pure
    public /*@non_null@*/ ArrayList methods() { return (ArrayList) methods.clone(); }

    public void setMethods(ArrayList m) { methods = m; }

    //@ pure
    public ArrayList inners() { return inners; }
    
    public void setInners(ArrayList i) {
        inners = i;
    }

    /**
     * @return  true if this class is at top level
     */
    //@ pure
    public boolean isAtTopLevel() {
        return sourceClass.host().isCompilationUnit();
    }

    //@ pure
    public JFieldDeclarationType[] fields() {
        return fields;
    }

    //@ pure
    public JConstructorDeclarationType getDefaultConstructor() {
        return defaultConstructor;
    }

    public void setDefaultConstructor(JConstructorDeclarationType 
                                      defaultConstructor) {
        this.defaultConstructor = defaultConstructor;
    }

    public void setIdent(/*@non_null@*/ String ident) {
        this.ident = ident;
    }

    /**
     * Returns the identifier for this type declaration.
     * @return  the unqualified identifier for this type declaration
     */
    //@ pure
    public /*@non_null@*/ String ident() {
        return ident;
    }

    /**
     * Adds the given member to this type's interface and modifies 
     * sourceClass to include the new member
     *
     * @param   newMember       a member to be added to this type
     */
    public void addMember( JMemberDeclarationType newMember ) {
        // !CONVERT! this should use dynamic dispatch on the type of
        // newMember
        if( newMember instanceof JMethodDeclarationType ) {
            methods.add( newMember );
        } else if( newMember instanceof JTypeDeclarationType ) {
            inners.add( newMember );
        } else {
            throw new RuntimeException( "JTypeDeclaration.addMember called " +
                                        "with invalid type: " +
                                        newMember.getClass() );
        }
    }

    public void setFields( JPhylum[] fieldsAndInits ) {
            this.fieldsAndInits = fieldsAndInits;
            extractFields();
    }

    public void setFieldsOnly( JFieldDeclarationType[] fields ) {
        this.fields = fields;
        this.fieldsAndInits = new JPhylum[fields.length];
        for (int i=0; i<fields.length; i++) {
            this.fieldsAndInits[i] = (JPhylum)fields[i];
        }
    }

    /**
     * Returns the logical owner of this type.  */
    public /*@pure@*/ /*@non_null*/ CClass owner() {
        return sourceClass.owner();
    }

    public /*@ pure @*/ long modifiers() {  
        return getCClass().access().modifiers();
    }

    /**
     * Marks this type as static.
     *
     * <pre><jml>
     * also 
     * requires (* this.generateInterface(...) has not been invoked *);
     * requires_redundantly sourceClass == null;
     * ensures hasFlag(modifiers,ACC_STATIC);
     * </jml></pre>
     *
     */
    public void setStatic() {
        modifiers |= ACC_STATIC;
    }

    /**
     * Marks this type as non-static.
     *
     * <pre><jml>
     * also 
     * ensures !hasFlag(modifiers,ACC_STATIC);
     * </jml></pre>
     */
    public void unsetStatic() {
        modifiers = (modifiers & ~ACC_STATIC);
        getCClass().access().removeStaticModifier();
    }

        /**
         * Records the fact that even though the class is an inner
         * class, we should not generate an outer this for it, because
         * the outer this is not accessible.
         * The sourceClass field should not be null.
         */
    public void syntheticOuterThisInaccessible() {
        assertTrue(sourceClass != null);
        sourceClass.syntheticOuterThisInaccessible();
    }

    /**
     * Adds the signature CSourceClass of this, and of all nested
     * types, to accum.
     */
    public void accumAllTypeSignatures(/*@non_null@*/ ArrayList accum) {
        accum.add(getCClass());
        Iterator iter = inners.iterator();
        while (iter.hasNext()) {
            JTypeDeclarationType t = (JTypeDeclarationType) iter.next();
            t.accumAllTypeSignatures(accum);
        }
        
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    public abstract void preprocessDependencies( /*@non_null@*/ CContextType context ) 
        throws PositionedError;

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    public void preprocessDependencies( /*@non_null@*/ CContextType context, 
                                     CClassType superClassType ) 
        throws PositionedError 
    {
        CClassContextType self = createContext( context );
        sourceClass.setSuperClass(superClassType);

        // Check inners
        Iterator iter = inners.iterator();
        while( iter.hasNext() ) {
            ((JTypeDeclarationType) iter.next()).preprocessDependencies(self);
        }

    }

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type
     * signatures of everything (imported class files, classes being
     * compiled, methods, fields, etc...)  needed for the later
     * passes.  This information is stored in a CCompilationUnit
     * instance and instances of CMember that are bound to the AST.
     * Also adds things like the default constructor and the initializer
     * method to the AST (these are suppressed during pretty-printing).
     *
     * @param           context         the context in which this
     *                                  decl appears
     * @exception       PositionedError an error with reference to the 
     *                                  source file
     */
    public void checkInterface( /*@non_null@*/ CContextType context ) throws PositionedError {
        CClassContextType self = createContext( context );

        Iterator iter;

        if ( isDeclaredInDifferentSourceFiles() ) {
            context.reportTrouble( uniqueSourceWarning() );
        }

        // 8.1 Class Modifiers
        MemberAccess access = sourceClass.access();
        access.checkClassModifiers(context, this);

        // !HERE!filenaming
        if( isAtTopLevel() && sourceClass.isPublic() 
            && !(this instanceof MJGenericFunctionDecl) // will warn when 
                                                        // checking methods
            && !getTokenReference().name().startsWith( ident + "." ) ) {
            context.reportTrouble(new CWarning(getTokenReference(), 
                                               MjcMessages.CLASS_NAME_FILENAME,
                                               ident, 
                                               getTokenReference().name()));
        }
        try {
            for(int i = 0; i<typevariables.length; i++ ) {
                typevariables[i].checkType(self);
            }
        } catch(UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
        
        // resolve interfaces
        try {
            for (int i = 0; i < interfaces.length; i++) {
                interfaces[i].checkType(self);
                interfaces[i] = (CClassType)CUniverseServices.setDefaultVariableAgain(interfaces[i]);
                if (! interfaces[i].isInterface() ) {
                    context.reportTrouble(
                        new PositionedError(getTokenReference(), 
                                    MjcMessages.CLASS_IMPLEMENTS_NON_INTERFACE,
                                            interfaces[i].ident())
                        );
                }

                // Top-level wildcard type arguments are not allowed in a
                // superinterface (see JLS3 8.1.5).
                check(context,
                      !interfaces[i].hasWildcardTypeArgument(),
                      MjcMessages.WILDCARD_IN_SUPER_TYPE,
                      interfaces[i].toVerboseString());
            }
        } catch (UnpositionedError cue) {
            throw cue.addPosition(getTokenReference());
        }
        sourceClass.setCheckedInterfaces( interfaces );


        // Add fields of this class
        Hashtable hashField;
        int generatedFields = 0;
        if (sourceClass.hasSyntheticOuterThis()) {
            addOuterThis();
            generatedFields = 1;
        }

        hashField = new Hashtable(fields.length + generatedFields + 1, 1.5f);
        for( int i = 0; i < fields.length; i++ ) {
            CSourceField field = fields[i].checkInterface(self);

            field.setIndex(i);
            check(context, hashField.put(field.ident(), field) == null,
                  MjcMessages.FIELD_RENAME, field.ident());
        }
        if (generatedFields > 0) {
            CSourceField field = outerThis.checkInterface(self);

            field.setIndex(hashField.size());
            check(context, hashField.put(JAV_OUTER_THIS, field) == null,
                  MjcMessages.FIELD_RENAME, JAV_OUTER_THIS);
        }

        // Check the declared methods individually and build an array
        // of CMethod instances that is used to register the method
        // signatures with the CClass representation stored in
        // sourceClass
        CMethod[] methodList;
        int generatedMethods = 0;
        generatedMethods += defaultConstructor != null ? 1 : 0;
        generatedMethods += statInit != null && !statInit.isDummy() ? 1 : 0;
        generatedMethods += 
            instanceInit != null && !instanceInit.isDummy() ? 1 : 0;

        methodList = new CMethod[ methods.size() + generatedMethods ];

        int count = 0;          // used to track position in methodList
        iter = methods.iterator();
        while( iter.hasNext() ) {
            JMethodDeclarationType method = 
                (JMethodDeclarationType) iter.next();

            methodList[count++] = (CMethod) method.checkInterface( self );
        }

        if (defaultConstructor != null) {
            methodList[count] = 
                (CMethod) (defaultConstructor.checkInterface(self));
            count++;
        }
        if (statInit != null) {
            if( statInit.isDummy() ) {
                statInit.checkInterface(self);
            } else {
                methodList[count] = (CMethod) (statInit.checkInterface(self));
                count++;
            }
        }
        if (instanceInit != null) {
            if( instanceInit.isDummy() ) {
                instanceInit.checkInterface(self);
            } else {
                methodList[count] = 
                    (CMethod) (instanceInit.checkInterface(self));
                count++;
            }
        }
        
        JMethodDeclarationType dup = 
            checkForDups( (JMethodDeclarationType[]) 
                          methods.toArray( new JMethodDeclarationType[0] ) );

        
        if( dup != null ) {
            throw new PositionedError( dup.getTokenReference(), 
                                       MjcMessages.METHOD_REDEFINE,
                                       dup.getMethod().ident() );
        }
        
        if(context.getCompiler().Generic()) {
            JMethodDeclarationType sameSig =
                checkForSameSignature((JMethodDeclarationType[] )
                                      methods.toArray(new JMethodDeclarationType[0]));         
            if(sameSig != null) {
                throw new PositionedError(sameSig.getTokenReference(),
                                          MjcMessages.METHOD_HAS_DUPLICATE_SIG,
                                          sameSig.getMethod().ident() );
            }  
        } 
            // Check inners
        iter = inners.iterator();
        while( iter.hasNext() ) {
            ((JTypeDeclarationType) iter.next()).checkInterface(self);
        }

        sourceClass.setCheckedMembers( hashField, methodList );

        resolveExtMethods( context );
    }

    /**
     * Returns the first method in the given array of methods that
     * duplicates the signature of an earlier method or null if no
     * duplicates exist.
     *
     * <pre><jml>
         requires (* checkInterface has already been called on each element
                        of methods *);
     * </jml></pre> */
    protected static JMethodDeclarationType checkForDups( JMethodDeclarationType[] 
                                                      methods ) 
    {
	if( !CTopLevel.getCompiler().universeChecks() ) {
	    HashSet duplicateCheckSet = new HashSet( methods.length );
	    for( int i=0; i < methods.length; i++ ) {
		CMethod m = methods[i].getMethod();
		if( !duplicateCheckSet.add( m ) ) {
		    return methods[i];
		}
	    }
	} else {
	    // WMD: disregard the Universe annotations for this check
	    for (int i = 0; i < methods.length; i++ ) {
		CMethod m = methods[i].getMethod();
		for (int j = 0; j < i; j++) {
		    CMethod m2 = methods[j].getMethod();
		    if (m.hasSameSignatureNoUniverses(m2)) {
			return methods[i];
		    }
		}
	    }
	}
	return null;
    }

    protected static JMethodDeclarationType checkForSameSignature
        (JMethodDeclarationType[] methods) {
        for (int i = 0; i< methods.length; i++ ) {
            CMethod m = methods[i].getMethod();
            for (int j = 0; j < i; j++) {
                CMethod m2 = methods[j].getMethod();
                if (m.hasSameSignature(m2)) {
                    return methods[i];
                }
            }
        }
        return null;
    }
    
    /**
     * Returns the appropriate warning message when another class has
     * been declared whose signature matches the signature of this.
     * Overridden in subclass.  */
    protected /*@non_null@*/ CWarning uniqueSourceWarning() {
        return new CWarning( getTokenReference(), MjcMessages.TOP_RENAME, 
                             sourceClass.qualifiedName() );
    }

    /**
     * Makes sure that all in-scope external generic functions are
     * added to the appropriate augmentation maps before top method
     * searches occur in later passes.  */
    public void resolveExtMethods( /*@non_null@*/ CContextType context ) {
        CClassContextType selfC = createContext( context );

        Iterator iter = methods.iterator();
        while (iter.hasNext()) {
            ((JMethodDeclarationType) iter.next()).resolveExtMethods(selfC);
        }
    }

    /**
     * Checks the static initializers created during the
     * checkInterface pass and performs some other checks that can be
     * performed simply before full blown typechecking.
     *
     * @param   context         the context in which this class 
     *                          declaration appears
     * @exception PositionedError if check fails */
    public void checkInitializers(/*@non_null@*/ CContextType context) throws PositionedError {
        // check for duplicate interfaces
        for( int i = 0; i < interfaces.length; i++ ) {
            for( int j = 0; j < i; j++ ) {
                check( context,
                       !interfaces[i].equals(interfaces[j]),
                       MjcMessages.INTERFACES_DUPLICATE, 
                       ident, interfaces[i] );
            }
        }

        // Checks that specified interfaces are not inherited
        CClass  parent = sourceClass.getSuperClass();

        if( parent != null ) {
            for( int i = 0; i < interfaces.length; i++ ) {
                if ( parent.descendsFrom( interfaces[i].getCClass() ) ) {
                    context.reportTrouble(
                      new CWarning( getTokenReference(), 
                                    MjcMessages
                                    .INTERFACE_IMPLEMENTED_BY_SUPERCLASS, 
                                    interfaces[i], 
                                    parent.ident() ));
                }
            }
        }

        for( int i = 0; i < interfaces.length; i++ ) {
            for( int j = 0; j < interfaces.length; j++ ) {
                if( j != i 
                    && interfaces[j].getCClass()
                    .descendsFrom( interfaces[i].getCClass() ) ) {
                    context.reportTrouble(
                          new CWarning( getTokenReference(),
                                        MjcMessages
                                        .INTERFACE_IMPLEMENTED_BY_SUPERCLASS,
                                        interfaces[i], 
                                        interfaces[j] ));
                }
            }
        }

        sourceClass.setInitializerChecked();

    }

    /**
     * Determines whether this class is declared in more than one 
     * source file and returns a flag indicating this.
     *
     * <pre><jml>
         ensures (* \result == true if declared in two different source 
                    files, otherwise \result == false *);
     * </jml></pre>
     */
    // FIXME - the above contained "requires clazz != null;" which does not check
    public boolean isDeclaredInDifferentSourceFiles() {
        return CTopLevel.isDeclaredInDifferentSourceFiles(uniqueSourceClass, 
                                                          sourceClass);
    }
    
    /**
     * Finds the top method of every declared method.  Called after
     * the resolveSpecializers pass.  This cannot be done before then
     * because the external generic function mappings are not complete
     * until the end of the checkInterface pass and the constant value
     * specializers are not known until after the resolveSpecializers
     * pass.  This must be done before the typecheck pass so that all
     * specialized argument positions for generic functions are known
     * for ambiguity checking. */
    public void resolveTopMethods() throws PositionedError {
        Iterator iter = methods.iterator();
        while (iter.hasNext()) {
            ((JMethodDeclarationType) iter.next()).resolveTopMethods();
        }
    }

    /**
     * Computes the values of specializer expressions used to dispatch on
     * compile-time constants.
     *
     * @param   context         the context in which this class 
     *                          declaration appears
     * @exception PositionedError if the check fails */
    public abstract void resolveSpecializers(/*@non_null@*/ CContextType context)
        throws PositionedError;


    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * This method establishes a fabricated field to store a reference
     * to an outer class of this.  It is designed to be a helper
     * method to the checkInterface method of this.  The fabricated
     * field is stored in the <code>outerThis</code> field of this
     * declaration and is also passed to <code>sourceClass</code> for
     * code generation.  If this declaration is immediately enclosed
     * by an external method then the fabricated field will reference
     * the _this_ synthetic receiver argument of the external method.
     * The value of the field in either case is set at run-time via a
     * synthetic parameter of this class's constructor. */
    private void addOuterThis() {
        if( outerThis == null ) {
            TokenReference loc = getTokenReference();
            CType ownerType = owner().getType();

	    if( CTopLevel.getCompiler().universeChecks() ) {

		// WMD: get a readonly version of the outer type
		// assumes that the owner is always a CClassType
		ownerType = CTopLevel.getTypeRep(
			      ((CClassType)ownerType).qualifiedName(),
			      CUniverseImplicitReadonly.getUniverse(), true );
	    }

            outerThis = 
                new JFieldDeclaration( loc,
                                       new JVariableDefinition( loc,
                                                                ACC_PRIVATE |
                                                                ACC_FINAL,
                                                                ownerType,
                                                                JAV_OUTER_THIS,
                                                                null ),
                                       null,
                                       null );
        }
    }
    
    /**
     * Creates a lexical context for typechecking and control flow
     * analysis on this.  */
    public abstract /*@non_null@*/ CClassContextType createContext(/*@non_null@*/ CContextType parent) ;

    /**
     * Typechecks this type declaration in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.  Actually just updates the context by adding a
     * CSourceClass representation of this.  Overriding methods in
     * subclasses of this handle the actual checks.
     *
     * @param context the context in which this type declaration appears
     * @exception PositionedError if any checks fail */
    public void typecheck( /*@non_null@*/ CContextType context ) throws PositionedError {
        context.classToGenerate( sourceClass );
    }

    // ---------------------------------------------------------------------
    // CODE GENERATION
    // ---------------------------------------------------------------------

    /**
     * Refactors this to include dispatchers for multimethods and
     * other code necessary for running MultiJava code on a standard
     * JVM.  */
    public void translateMJ( /*@non_null@*/ CContextType context ) {

        CClassContextType self = createContext( context );

        int size = inners.size();
        for( int i=0; i < size; i++ ) {
            ((JTypeDeclarationType) inners.get(i)).translateMJ( self );
        }

        if (context.getCompiler().RMJ()) {
            // generate any method annotations needed by internal
            // methods that specialize on interfaces
            try {
                annotateInternalInterfaceMethods(context);
            } catch (UnpositionedError err) {
                // shouldn't get here because compilation would
                // have aborted on the typechecking pass if this
                // try-block is bad
                fail( "shouldn't reach this code" );
            }
        }

        WrapResult wrapRes = 
            dispatcherWrapMethods( context, methodsToBeWrapped() );

        sourceClass.close( wrapRes.methods, wrapRes.dispClassTypes );
    }

    /**
     * Returns an array of the methods to be considered for
     * refactoring into dispatchers during the translateMJ pass.
     *
     * @see #translateMJ(CContextType) */
    protected /*@non_null@*/ JMethodDeclarationType[] methodsToBeWrapped() {
        return (JMethodDeclarationType[])
            methods.toArray( new JMethodDeclarationType[0] );
    }

    /**
     * This nested class is used to return the results from <code>
     * dispatcherWrapMethods</code>
     *
     * @see #dispatcherWrapMethods(CContextType, JMethodDeclarationType[])
     */
    protected static class WrapResult {
        protected WrapResult( CMethod[] methods, CClassType[] dispClassTypes ) 
        {
            this.methods = methods;
            this.dispClassTypes = dispClassTypes;
        }
        protected CMethod[] methods;
        protected CClassType[] dispClassTypes;
    }

    /**
     * <P>Builds a collection of methods for code generation.  Finds
     * the sets of multimethods that belong to the same generic
     * function or otherwise require a dispatcher method and generates
     * such a method for each set.  Regular Java methods (those not
     * requiring a dispatcher) and generated methods (e.g., default
     * constructors) are also included in the result.  An array of
     * inner dispatcher classes, one for each overridden external
     * generic function, is also part of the result.</P>
     *
     * <P>(This method would be private but is declared protected and
     * final to allow for unit testing.)</P>
     *
     * @param context           the context in which these method declarations 
     *                          appear, used for finding the top method of  
     *                          each generic function
     * @param methodDecls       an array of the methods to be generated
     * @return an array of the methods to be created during code generation,
     *          and an array of the inner dispatcher classes to be created
     *          during code generation
     * @see #translateMJ(CContextType) */
    protected final WrapResult dispatcherWrapMethods(final CContextType context, 
                                                     JMethodDeclarationType[] 
                                                     methodDecls )
    {
        // -------------------------------------------------------------------
        // Some local class declarations
        // -------------------------------------------------------------------

        /**
         * For efficiency it is useful to be able to deal with method
         * signatures and their top method as a single object; this
         * local class provides that grouping.  */
        class MethodRecord implements Comparable {
            MethodRecord( JMethodDeclarationType methodDecl ) {
                try {
                    where = methodDecl.getTokenReference();
                    methodSig = (CSourceMethod) methodDecl.getMethod();
                    topConcreteMethod = methodSig.topConcreteMethod();
                    needsDispatcher = topConcreteMethod.isExternal() || 
                        methodDecl.usesMultipleDispatch();
                } catch( UnpositionedError e ) {
                    // shouldn't get here because compilation would
                    // have aborted on the typechecking pass if this
                    // try-block is bad
                    fail( "shouldn't reach this code" );
                }
            }

            /**
             * Imposes an ordering based on identity and signature of
             * the top methods.  This ensures that methods of the same
             * generic function are contiguous in a natural-order sort
             * on collections of MethodRecords.  */
            public int compareTo(/*@ non_null @*/ Object o) {
                MethodRecord other = (MethodRecord) o;
                if (topConcreteMethod == other.topConcreteMethod) {
                    return 0;
                } else {
                    return topSig().compareTo( other.topSig() );
                }
            }

            private String topSig() {
                if (topSig == null) {
                    topSig = topConcreteMethod.toString();
                }
                return topSig;
            }

            TokenReference where;

            /**
             * If true then this method requires a dispatcher; if
             * false this method may or may not require a dispatcher
             * depending on the other method in the context. */
            boolean needsDispatcher;

            /**
             * The signature of this method.  */
            CSourceMethod methodSig;

            /**
             * The top concrete method of this method's generic function.  */
            CMethod topConcreteMethod;

            /**
             * String representation of the signature of the top
             * method.  Initialized lazily by topSig(). */
            String topSig;
        };

        /**
         * A collection of dispatcher and signature classes that makes
         * the algorithm below more readable.  */
        class DispatcherClassList extends ArrayList {
            DispatcherClassList( int initialCapacity ) {
                super( initialCapacity );
            }
            boolean addDispatcherClass( CSourceDispatcherClass cl ) {
                return add( cl );
            }
            boolean addAmbiguousDispatcherClass( 
                    CAmbiguousDispatcherClass cl ) {
                // RMJ only
                return add( cl );
            }
            boolean addDispatcherSignature( CDispatcherSignature sig ) {
                return add( sig );
            }

            CClassType[] allClassTypes() {
                CClassType[] result = new CClassType[ size() ];
                for (int i = 0; i < result.length; i++) {
                    result[i] = ((CClass) get(i)).getType();
                } // end of for
                return result;
            }

            CSourceDispatcherClass[] dispatcherClassesArray() {
                ArrayList rList = new ArrayList( size() );
                int size = size();
                for (int i = 0; i < size; i++) {
                    CClass c = (CClass) get(i);
                    if (c instanceof CSourceDispatcherClass) {
                        rList.add(c);
                    } // end of if 
                } // end of for
                CSourceDispatcherClass[] result = 
                    new CSourceDispatcherClass[ rList.size() ];
                rList.toArray( result );
                return result;
            }

        }

        /**
         * A collection of methods makes the algorithm below more
         * readable.  */
        class MethodList extends ArrayList {
            MethodList( int initialCapacity ) {
                super( initialCapacity );
            }
            boolean addMethod( MethodRecord m ) {
                return add( m );
            }
            MethodRecord getMethod( int index ) {
                return (MethodRecord) get( index );
            }
            CSourceMethod[] methodArray() {
                CSourceMethod[] result = new CSourceMethod[ size() ];
                methodArray( result, 0 );
                return result;
            }
            void methodArray( CSourceMethod[] dst, int offset ) {
                int size = size();
                for( int i=0; i < size; i++ ) {
                    dst[i + offset] = getMethod( i ).methodSig;
                }
            }
            /**
             * Determines whether a dispatcher is needed for the
             * methods in this.  If so, creates the dispatcher.  If
             * the dispatcher is for an internal generic function,
             * then it is added to <code>methods</code>, otherwise the
             * constructor for the dispatcher will generate a
             * <Code>CSourceDispatcherClass</code> instance and, if
             * necessary, a <code>CSourceDispatcherSignature</code>
             * instance and add them to the list of classes to be
             * generated and to the list of inner classes for this.
             * If no dispatcher is required then this method adds the
             * (necessarily) single method in this group to
             * <code>methods</code>.
             * The gfID is used to give each gf's associated $body methods
             * a unique name.  */
            void addAppropriateMethodTo( ArrayList methods, 
                                         DispatcherClassList dispClasses,
                                         int gfID ) {
                MethodRecord oldGroupRepresentative = getMethod(0);
                if( oldGroupRepresentative.needsDispatcher ||
                    size() > 1 ) {
                    // a dispatcher is required
                    CSourceDispatcherMethod m =
                        new CSourceDispatcherMethod( 
                                 oldGroupRepresentative.where,
                                 context.getCompiler(),
                                 sourceClass,
                                 oldGroupRepresentative.topConcreteMethod,
                                 methodArray(), gfID );
                    if (m.topConcreteMethod.isExternal()) {
                        // add inner dispatcher class
                        dispClasses.addDispatcherClass( m.dispatcherClass() );
                        if (m.isTopDispatcherInLattice()) {
                            // add inner dispatcher signature class
                            dispClasses.
                                addDispatcherSignature(m.dispatcherSignature());
                            if (context.getCompiler().RMJ()) {
                                // also add inner ambiguous dispatcher
                                // class, in case the loader needs it
                                dispClasses.addAmbiguousDispatcherClass( 
                                    m.ambiguousDispatcherClass() );
                            }
                        } // end of if 
                    } else {
                        // add internal dispatcher method
                        methods.add( m );
                    }
                } else {
                    // no dispatcher is required and we are certain
                    // that group.size() == 1
                    methods.add( oldGroupRepresentative.methodSig );
                }
            }
        };

        class PleomorphSet extends HashSet {
            PleomorphSet( MethodList list ) {
                super( list.size() ); // est. capacity as maximum value
                for( int i=0; i < list.size(); i++ ) {
                    MethodRecord method = list.getMethod( i );
                    if( method.methodSig.isMixedPleomorphic() ) {
                        add( new CSourceRedirectorMethod( method.methodSig,
                                                          method.where));
                    }
                }
            }
        }

        // -------------------------------------------------------------------
        // The actual algorithm
        // -------------------------------------------------------------------

        int methodCount = methodDecls.length;

        // Build a list of methods to be generated in bytecode
        // within this type.  This list will include all the
        // redirectors, all the regular methods, and all the
        // dispatchers for internal generic functions.  The list
        // will not include any multimethods that are enclosed by
        // dispatchers or any dispatchers for external generic
        // functions.  The latter are generated in bytecode for inner
        // dispatcher classes (see CSourceDispatcherClass).
        ArrayList toBeGeneratedMethods = new ArrayList( methodCount );
        
        // Also build an array of the inner dispatcher classes for this.
        DispatcherClassList dispClasses = 
            new DispatcherClassList( methodCount );

        if (methodCount>0) {
             
            // Convert the methods into a form that is easier to work with.
            MethodList allMethods = new MethodList( methodCount );
            for( int i=0; i < methodCount; i++ ) {
                allMethods.addMethod( new MethodRecord( methodDecls[i] ) );
            }

            // Sort the methods so that methods of the same generic function
            // will be contiguous.
            Collections.sort( allMethods );

            // Create the redirector methods (for pleomorphic methods
            // belonging to internal and external generic functions) .
            Set redirectors = new PleomorphSet( allMethods );

            toBeGeneratedMethods.addAll( redirectors );
            MethodList group = new MethodList( methodCount );

            MethodRecord method = allMethods.getMethod( 0 );
            Object lastTop = method.topConcreteMethod;
                // give each generic function a unique id, for use in the
                // name of the gf's associated $body methods
            group.addMethod( method );
            for( int i = 1; i < methodCount; i++ ) {
                method = allMethods.getMethod( i );
                if ( lastTop != method.topConcreteMethod ||
                     /* RMJ: put each method of an external GF into
                      * its own dispatcher class */
                     context.getCompiler().RMJ() && 
                     method.topConcreteMethod.isExternal() ) {
                    // This method is the start of a new group, so see if
                    // a dispatcher is required for the old group and if
                    // so create it.
                        
                    group.addAppropriateMethodTo( toBeGeneratedMethods,
                                                  dispClasses, i );
                    group.clear();      // reset to begin next group
                    lastTop = method.topConcreteMethod;
                }
                group.addMethod( method );
            }
            // See if a dispatcher is required for the last group and if
            // so create it.
            group.addAppropriateMethodTo( toBeGeneratedMethods,
                                          dispClasses, methodCount );
        }

        // add the default constructors and generated initializers, if any
        if ( defaultConstructor != null ) {
            toBeGeneratedMethods.add( defaultConstructor.getMethod() );
        }

        boolean hasDispatchers = dispClasses.size() > 0;
        CMethod init = null;
        if (statInit != null && !statInit.isDummy()) {
            // get the regular static initializer
            init = statInit.getMethod();
        }

        CMethod instInit = null;
        if (instanceInit != null && !instanceInit.isDummy()) {
            // get the regular initializer
            instInit = instanceInit.getMethod();
        }

        if (hasDispatchers) {
            CSourceDispatcherClass[] dca = 
                dispClasses.dispatcherClassesArray();
            if ( context.getCompiler().RMJ() ) {
                // RMJ: don't use anchor initializers, but rather use
                // dispatcher initializers
            } else {
                // Create members for initializing the chains of responsibility
                // for the external generic functions extended by this.
                CCORInitializer corInitializer = 
                    new CCORInitializer(sourceClass, dca);
                toBeGeneratedMethods.add( corInitializer );
                sourceClass.addField( corInitializer.isInitializedFlag() );
                sourceClass.addField( corInitializer.isInitializedLock() );

                // wrap the regular initializers (if any) with calls to the
                // initializer method for the chain of responsibility
                init = new CAnchorInitializer(sourceClass, (CSourceMethod) init,
                                              true);
                instInit = new CAnchorInitializer(sourceClass, 
                                                  (CSourceMethod) instInit, 
                                                  false);
            }
            // add fields to this for the dispatcher classes
            for (int i = 0; i < dca.length; i++) {
                CSourceField odf = dca[i].outersDispatcherField();
                if (odf != null) {
                    sourceClass.addField(odf);
                }
            }
        }
        if ( init != null ) {
            toBeGeneratedMethods.add( init );
        }

        if ( instInit != null ) {
            toBeGeneratedMethods.add( instInit );
        }

        CSourceMethod[] resultMethods = 
            new CSourceMethod[ toBeGeneratedMethods.size() ];
        toBeGeneratedMethods.toArray( resultMethods );

        if (hasDispatchers && !context.getCompiler().RMJ()) {
            // An ugly kludge to make sure any constructors call the initializer 
            // block, which in turn ensures that the chain of responsibility has
            // been initialized before any instances are available.
            for (int i = 0; i < resultMethods.length; i++ ) {
                JBlock bod = resultMethods[i].body();
                if (bod instanceof JConstructorBlock) {
                    ((JConstructorBlock) bod).addInitializerBlockCall();
                }
            }        
        }

        return new WrapResult( resultMethods, dispClasses.allClassTypes() );
    }

    // ----------------------------------------------------------------------
    // PROTECTED UTILITIES
    // ----------------------------------------------------------------------

    /**
     * Extracts the field declarations from <code>fieldsAndInits</code>.  
     *
     * @see #fieldsAndInits
     */
    private void extractFields() {
        ArrayList results = new ArrayList();
        for( int i = 0; i < fieldsAndInits.length; i++ ) {
            if( fieldsAndInits[i] instanceof JFieldDeclarationType ) {
                results.add( fieldsAndInits[i] );
            }
        }
        fields = new JFieldDeclarationType[results.size()];
        fields = (JFieldDeclarationType[]) results.toArray( fields );
        
    }

    // -------------------------------------------------------------
    // CompilerPassEnterable
    // -------------------------------------------------------------


    /**
     * Compares this to a given object.
     *
     * <pre><jml>
       also
         ensures (* \result == lexically ordered according to the type name *);
       also
         requires o != null && !(o instanceof JTypeDeclaration);
         signals_only ClassCastException;
     * </jml></pre>
     *
     *
     * @param o an <code>Object</code> value to compare to this
     * @return based on a lexical ordering according to the type name 
     * @exception ClassCastException if <code>o</code> is incomparable to this
     */
    public int compareTo(/*@ non_null @*/ /*@\readonly@*/ Object o) throws ClassCastException {
        JTypeDeclaration other = (JTypeDeclaration) o;
        int res = ident.compareTo(other.ident());
        return res;
    }

    
    //@ protected represents arePassParametersCached <- arePassParametersCached();

    /*@ protected pure model boolean arePassParametersCached() {
      @    return cachedContext == null;
      @ }
      @*/
    
    /**
     * Caches the arguments for the compiler passes.
     *
     * <pre><jml>
     * also
     * ensures arePassParametersCached;
     * </jml></pre> 
     *
     * @see CompilerPassEnterable
     */
    public void cachePassParameters( CContextType context ) {
        cachedContext = context;
    }

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    public void preprocessDependencies() throws PositionedError {
        preprocessDependencies( cachedContext );
    }

    public void checkInterface() throws PositionedError {
        checkInterface(cachedContext);
    }

    public void checkInitializers() throws PositionedError {
        checkInitializers(cachedContext);
    }

    public void resolveSpecializers() throws PositionedError {
        resolveSpecializers(cachedContext);
    }
    
    public void typecheck() throws PositionedError {
        typecheck(cachedContext);
    }

    public void translateMJ() {
        translateMJ(cachedContext);
    }

    protected CContextType cachedContext;

    // ---------------------------------------------------------------------
    // RMJ Annotations
    // ---------------------------------------------------------------------

    // generate any method annotations needed by internal
    // methods that specialize on interfaces (RMJ only)
    protected void annotateInternalInterfaceMethods(CContextType context)
        throws UnpositionedError
    {
        // first find all the internal GFs (as represented by their
        // top methods)
        CMethodSet topConcreteMethods = new CMethodSet();
        Iterator iter = methods.iterator();
        while (iter.hasNext()) {
            JMethodDeclarationType methodDecl =
                (JMethodDeclarationType) iter.next();
            CSourceMethod methodSig = (CSourceMethod) methodDecl.getMethod();
            
            recordInternalTopMethod(methodSig, topConcreteMethods);
        }

        // then for each of the top methods (i.e., internal GFs)...
        int size = topConcreteMethods.size();
        for (int i = 0; i < size; i++) {
            CMethod topConcreteMethod = topConcreteMethods.getMethod(i);

            // generate any annotations for interface specializers necessary
            annotateInternalInterfaceMethodsFor(topConcreteMethod, context);
        }
    }

    // find the top method(s) that represent the gf(s) for methodSig,
    // and add them to topConcreteMethods
    private void recordInternalTopMethod(CMethod methodSig, 
                                         CMethodSet topConcreteMethods)
        throws UnpositionedError
    {
        CMethod topConcreteMethod = methodSig.topConcreteMethod();
        if (!topConcreteMethod.isExternal()) {
            // this is an internal GF, which we may need to
            // generate annotations for
            if (! topConcreteMethods.contains(topConcreteMethod)) {
                topConcreteMethods.addMethod(topConcreteMethod);
            }
        
        } else if (methodSig.isMixedPleomorphic()) {
            // this method is in both internal and external GFs; find
            // the internal ones
            CMethodSet superMethods =
                ((CSourceMethod)methodSig).overriddenMethodSet();
            int size = superMethods.size();
            for (int i = 0; i < size; i++) {
                CMethod sm = superMethods.getMethod(i);
                recordInternalTopMethod(sm, topConcreteMethods);
            }

        } else {
            // this method is only in an external GF, which has
            // its annotations handled a different way
        }
    }

    // generate any annotations necessary for the gf represented by
    // topConcreteMethod
    private void annotateInternalInterfaceMethodsFor(CMethod topConcreteMethod, 
                                                     CContextType context)
        throws UnpositionedError
    {
        // find all the methods in this GF
        CMethodSet methods = sourceClass.lookupAllMethods(topConcreteMethod,
                                                          CClassType.EMPTY,
                                                          context);

        // now see whether any of these methods specializes on an interface
        boolean foundInterfaceSpecializingMethod = false;
        int numMethods = methods.size();
        for (int i = 0; i < numMethods; i++) {
            CMethod m = methods.getMethod(i);
            CSpecializedType[] params = m.parameters();
            int numParams = params.length;
            for (int j = 0; j < numParams; j++) {
                CSpecializedType param = params[j];
                CType dt = param.dynamicType();
                if (dt.isClassType() && 
                    ((CClassType)dt).getCClass().isInterface()) {
                    // this argument is an interface; need to generate
                    // some annotations!
                    
                    // check to see whether this
                    // interface-specializing method is in the current
                    // class or inherited from a superclass
                    if (m.owner().equals(sourceClass)) {
                        // a method in this class
                        foundInterfaceSpecializingMethod = true;
                        break;
                    } else {
                        // a method inherited; we already know what to do
                        annotateInterfaceMethods(topConcreteMethod, methods, true);
                        return;
                    }
                }
            }
        }
        if (foundInterfaceSpecializingMethod) {
            // found at least one method that contained an interface
            // specializer, but all such methods were in the current
            // class
            annotateInterfaceMethods(topConcreteMethod, methods, false);
        } else {
            // no interface specializers for any method; no
            // annotations to generate
        }
    }

    // we call this if we've found an internal GF whose methods have
    // interface specializers; this should generate annotations that
    // are checked at load-time.  anyInherited is false iff all of the
    // interface-specializing methods are in the current class; in
    // this case, inherited non-interface methods need to have
    // annotations generated as well.
    private void annotateInterfaceMethods(CMethod topConcreteMethod,
                                          CMethodSet methods,
                                          boolean anyInherited)
    {
        // XXX probably we could filter out some of these methods as
        // not causing any ambiguities not already detectable by
        // existing methods, but we don't do that for now.  here are
        // some notes about what we have to include, and what we don't
        // have to include:

        //   if we have one method specializing on an interface, and
        //   another specializing on an interface or a class, then
        //   it's always possible for a new class to be loaded that
        //   inherits from the first interface and the second
        //   interface or class, causing the two methods to become
        //   ambiguous.  so we can always have ambiguities generated
        //   between unordered methods if at least one specializes on
        //   an interface.

        //   if a method overrides two ambiguous methods, resolving
        //   their ambiguity, then it should be included.

        //   if a method overrides another, but doesn't resolve an
        //   ambiguity (i.e., it has exactly one directly overridden
        //   method), then it doesn't need to be generated, since any
        //   ambiguities that would be detected for it would also be
        //   detected for its super-method.  [is this right?]

        // in the face of multiple dispatched arguments, one way to
        // think about this is to consider argument-tuple-space, with
        // methods associated with various points in this space.
        // then, we need to include a method's declaration only if (a)
        // (i) there exists some other method declaration with which
        // it has a non-empty intersection (considering all possible
        // future classes), and (ii) it doesn't have a single
        // super-method that is also ambiguous that that other method
        // declaration; or (b) it has more than one immediate
        // super-method.

        if (anyInherited) {
            // if this class is inheriting from a class that's already
            // generated its method annotations, then generate only
            // annotations for new methods
            int numMethods = methods.size();
            for (int i = 0; i < numMethods; i++) {
                CMethod m = methods.getMethod(i);
                if (m.owner().equals(sourceClass)) {
                    addRMJMethodAnnotation(m, topConcreteMethod, anyInherited);
                }
            }
        } else {
            // if this class is the first class that has any
            // interface-specializing methods, then we need to
            // generate all the methods in this GF [XXX what happens
            // if we have some other subclass of the same superclass
            // which also introduces interface-specializing methods?
            // then we get duplicate annotations for these inherited
            // methods.]
            int numMethods = methods.size();
            for (int i = 0; i < numMethods; i++) {
                CMethod m = methods.getMethod(i);
                addRMJMethodAnnotation(m, topConcreteMethod, anyInherited);
            }
        }
    }

    private void addRMJMethodAnnotation(CMethod method, 
                                        CMethod topConcreteMethod,
                                        boolean anyInherited)
    {
        String opname = method.ident();

        // the receiver and argument static types and specializers
        // (in this context, parameters *doesn't* include the receiver)
        CSpecializedType[] parms = method.parameters();
        String[] types = new String[parms.length + 1];
        String[] specs = new String[parms.length + 1];

        // the receiver:
        types[0] = topConcreteMethod.receiverType().toString();
        if (topConcreteMethod.receiverType().equals(method.receiverType())) {
            // the receiver isn't "specialized"
            specs[0] = "-";
        } else {
            // the receiver is "specialized"
            specs[0] = method.receiverType().toString();
        }

        // the arguments:
        for (int j = 0; j < parms.length; j++) {
            CSpecializedType parm = parms[j];

            CType stat = parm.staticType();
            String type;
            if (stat instanceof CClassType) {
                // a class type
                type = stat.toString();
            } else {
                // a primitive type
                type = "=" + stat.toString();
            }
            types[j+1] = type;

            String spec;
            if (parm.isSpecialized()) {
                CType dyn = parm.dynamicType();
                if (dyn instanceof CClassType) {
                    // a class or array specializer
                    spec = dyn.toString();
                } else if (dyn instanceof CValueType) {
                    // a value specializer
                    spec = "=" + ((CValueType)dyn).getValue().toString();
                } else {
                    throw new RuntimeException(
                        "Internal error: unexpected specializer: " + dyn);
                }
            } else {
                // this position isn't specialized
                spec = "-";
            }
            specs[j+1] = spec;
        }

        // whether or not this method is abstract:
        boolean isAbstract = method.isAbstract();

        String kind;
        if (anyInherited) {
            // we're extending what's come before, so it's as if we're
            // in a subanchor [XXX i guess]
            kind = "S";
        } else {
            // we're generating the first of the annotations, so we're
            // in the base anchor
            kind = "B";
        }

        // use the receiver class of the top method as the base anchor
        // class name
        String baseAnchorClassName = topConcreteMethod.owner().qualifiedName();

        // use the receiver class of this method as the dispatcher
        // class name (doesn't really matter)
        String dispatcherClassName = method.owner().qualifiedName();
                
        // use "-" as the ambig error class name, to indicate that
        // we're in an internal GF and we can't load an ambiguity
        // error class
        String ambigErrorClassName = "-";
                
        RMJAnnotation.Method annote = new RMJAnnotation.Method(
                opname.replace('/', '.'),
                types,
                specs, 
                isAbstract, 
                kind,
                baseAnchorClassName.replace('/', '.'),
                dispatcherClassName.replace('/', '.'),
                ambigErrorClassName.replace('/', '.'));
        
        sourceClass.addRMJMethodAnnotation(annote);
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    //@ spec_public
    protected   long                    modifiers; 
    protected   String                  ident;

    /**
     * The fields and initializers of this type, passed together
     * because the order matters for class and object initialization.
     * Members of the array should be instances of
     * <code>JFieldDeclarationType</code> or <code>JClassBlock</code>.  */
    protected   JPhylum[]               fieldsAndInits;

    /**
     * The fields declared within the type represented by this,
     * extracted from fieldsAndInits at initialization time. */
    private     JFieldDeclarationType[] fields;

    protected   ArrayList                       methods;

    /**
     * Stores the nested and inner types and private generic functions
     * of the type or generic function represented by this.  */
    protected   ArrayList                       inners;
    protected   CClassType[]            interfaces;

    private     JFieldDeclarationType   outerThis;
    private     JConstructorDeclarationType defaultConstructor;
    protected   JInitializerDeclaration statInit;
    protected   JInitializerDeclaration instanceInit;

    //@ spec_public
    protected   CSourceClass            sourceClass;

    /**
     * Used to detect whether there are duplicated declarations on the
     * same name. */
    protected   boolean                 uniqueSourceClass = true;
    protected   CTypeVariable[]         typevariables;

}

