/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTypeDeclarationStatement.java,v 1.8 2004/01/16 23:21:11 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a local type declaration statement.  A local
 * type declaration declaration statement declares one type
 * declaration in a body of a method.  JLS 14.3: Local type
 * declaration declaration statement 
 */
public class JTypeDeclarationStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where	the line of this node in the source code
     * @param	decl	the type declaration for the inner class
     */
    public JTypeDeclarationStatement( TokenReference where, 
				      JTypeDeclarationType decl ) {
	super( where, null );
	this.decl = decl;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	CClass owner = context.getClassContext().getOwnerClass();
	CClass host = context.getClassContext().getHostClass();
	String prefix = owner.qualifiedName() + "$" 
	    + owner.getNextSyntheticIndex() + "$";

	// An local type declaration in a "static context", i.e., in a
	// static method or static initializer, is considered to be a
	// nested class but not an inner class.
	if (context.isStatic()) {
	    decl.setStatic();
	}

	decl.generateInterface( context.getCompiler(), 
				owner, host, prefix, false, false );

	try {
	    context.addLocalClass( decl.getCClass() );
	} catch( UnpositionedError cue ) {
	    throw cue.addPosition( getTokenReference() );
	}

	context.catchUp(decl);
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitTypeDeclarationStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	// nothing to do here
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JTypeDeclarationType decl() { return decl; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JTypeDeclarationType		decl;
}
