/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTypeDeclarationType.java,v 1.26 2008/08/18 19:20:28 smshaner Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

import org.multijava.javadoc.JavadocComment;



/**
 * This type represents a java class or interface in the syntax tree
 */
public interface JTypeDeclarationType 
    extends JMemberDeclarationType, CompilerPassEnterable
{

    /**
     * Generates a <code>CSourceClass</code> class signature singleton
     * for this declaration.
     *
     * @param compiler	the compiler instance for which this singleton is 
     *			generated
     * @param owner	the class signature singleton for the logical outer 
     *			class of this, or null if this is a top level 
     *			declaration
     * @param host	the signature singleton of the context in which this
     *			is declared, a <code>CCompilationUnit</code> for 
     *			top-level declarations
     * @param prefix 	the prefix prepended to this declaration's
     *			identifier to achieve the fully qualified
     *			name, just the package name (using '/'
     *			separators) for top-level classes, package
     *			name plus $-delimited outer class names plus
     *			synthetic index for inner classes
     * @param isAnon	true if this is an anonymous class, in which 
     *			case the fully qualified name is just 
     *			<code>prefix</code> 
     * @param isMember	true if this is a member type, i.e., a nested type
     *			that is not a local type or an anonymous class
     */
    void generateInterface( Main compiler, CClass owner, CMemberHost host,
			    String prefix, boolean isAnon, boolean isMember );

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /** Returns the set of interfaces that this type implements or extends. */
    /*@ pure @*/ CClassType[] interfaces();

    /**
     * Returns an array of the static and instance initializers for
     * the type represented by this.  */
    /*@ pure @*/ JPhylum[] fieldsAndInits();

    /**
     * Walks up the extends hierarchy and adds all methods
     * to the ArrayList returned.
     *
     * @return		a ArrayList that includes the CMethods for all 
     *			methods in this class and all direct ancestors
     */
    /*@ pure @*/ ArrayList getAllMethods();

    /*@ pure @*/ ArrayList methods();

    /*@ pure @*/ ArrayList inners();
    
    void setInners(ArrayList v);

    /**
     * @return	true if this class is at top level
     */
    /*@ pure @*/ boolean isAtTopLevel();

    /*@ pure @*/ JFieldDeclarationType[] fields();

    /*@ pure @*/ JConstructorDeclarationType getDefaultConstructor();

    void setDefaultConstructor( JConstructorDeclarationType defaultConstructor);

    void setIdent(String ident);

    /**
     * Returns the identifier for this type declaration.
     * @return	the unqualified identifier for this type declaration
     */
    /*@ pure @*/ String ident();

    /**
     * Returns the type variables for this type declaration.
     * @return 
     */
    CTypeVariable[] typevariables();


    /**
     * Adds the given member to this type's interface and modifies 
     * sourceClass to include the new member
     *
     * @param	newMember	a member to be added to this type
     */
    void addMember( JMemberDeclarationType newMember );

    /**
     * Returns the logical owner of this type.  */
    /*@ pure @*/ CClass owner();

    /*@ pure @*/ long modifiers(); 

    /**
     * Marks this type as static.
     *
     * <pre><jml>
     * requires (* this.generateInterface(...) has not been invoked *);
     * requires_redundantly (*sourceClass == null*); // !FIXME! Fix annotation - sourceClass is defined in the implementing classes
     * ensures org.multijava.util.Utils.hasFlag(modifiers(),Constants.ACC_STATIC); 
     * </jml></pre>
     *
     */
    void setStatic();

    
    /**
     * Marks this type as non-static.
     */
    void unsetStatic();

    	/**
	 * Records the fact that even though the class is an inner
	 * class, we should not generate an outer this for it, because
	 * the outer this is not accessible.
	 */
    public void syntheticOuterThisInaccessible();

    /**
     * Adds the signature CSourceClass of this, and of all nested
     * types, to accum.
     */
    void accumAllTypeSignatures(ArrayList accum);

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Performs preliminary processing on compilation units and types.
     * Processes type imports so external methods' receiver types can
     * be analyzed and supertypes can be resolved.  Groups external
     * methods by name, corresponding to the anchor classes that will
     * eventually be generated.  Mutates the name space management in
     * CTopLevel to record a CGenericFunctionCollection singleton for
     * each anchor class.  */
    void preprocessDependencies( CContextType context ) throws PositionedError;

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type
     * signatures of everything (imported class files, classes being
     * compiled, methods, fields, etc...)  needed for the later
     * passes.  This information is stored in a CCompilationUnit
     * instance and instances of CMember that are bound to the AST.
     * Also adds things like the default constructor and the initializer
     * method to the AST (these are suppressed during pretty-printing).
     *
     * @param		context		the context in which this
     *					decl appears
     * @exception	PositionedError	an error with reference to 
     *					the source file
     */
    void checkInterface( CContextType context ) throws PositionedError;

    /**
     * Checks the static initializers created during the
     * checkInterface pass and performs some other checks that can be
     * performed simply before full blown typechecking.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if check fails */
    void checkInitializers(CContextType context) throws PositionedError;
    
    /**
     * Computes the values of specializer expressions used to dispatch on
     * compile-time constants.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    void resolveSpecializers(/*@non_null@*/ CContextType context)
	throws PositionedError;

    /**
     * Finds the top method of every declared method.  Called after
     * the resolveSpecializers pass.  This cannot be done before then
     * because the external generic function mappings are not complete
     * until the end of the checkInterface pass and the constant value
     * specializers are not known until after the resolveSpecializers
     * pass.  This must be done before the typecheck pass so that all
     * specialized argument positions for generic functions are known
     * for ambiguity checking. */
    void resolveTopMethods() throws PositionedError;

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Creates a lexical context for typechecking and control flow
     * analysis on this.  */
    CClassContextType createContext(/*@ non_null @*/ CContextType parent);

    /**
     * Typechecks this type declaration in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.  Actually just updates the context by adding a
     * CSourceClass representation of this.  Overriding methods in
     * subclasses of this handle the actual checks.
     *
     * @param context the context in which this type declaration appears
     * @exception PositionedError if any checks fail */
    void typecheck( CContextType context ) throws PositionedError;

    // ---------------------------------------------------------------------
    // CODE GENERATION
    // ---------------------------------------------------------------------

    /**
     * Refactors this to include dispatchers for multimethods and
     * other code necessary for running MultiJava code on a standard
     * JVM.  */
    void translateMJ( CContextType context );

    /**
     * Caches the arguments for the compiler passes.
     *
     * <pre><jml>
     * ensures arePassParametersCached;
     * </jml></pre> 
     *
     * @see CompilerPassEnterable
     */
    void cachePassParameters( CContextType context );

}

