/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JTypeNameExpression.java,v 1.10 2004/06/28 13:52:05 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents the AST node for a type name expression like
 * <code>Object</code> */
public class JTypeNameExpression extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	qualifiedName	the name of the type as a String
     * @param   univ            the universe modifier
     * @param   sourceName      the name as appeared in the source file;
     *                          useful for pretty printing.
     */
    public JTypeNameExpression( TokenReference where, 
                                String qualifiedName,
				CUniverse univ, // WMD
                                JNameExpression sourceName) {
	super( where );
	type = CTopLevel.getTypeRep( qualifiedName, univ, false );
        this.sourceName = sourceName;
    }

    /**
     * Constructs a node in the parsing tree.
     * @param	where	 the line of this node in the source code
     * @param	type	 the type that this expression references
     * @param sourceName the name as appeared in the source file;
     *                   useful for pretty printing.
     */
    public JTypeNameExpression(TokenReference where, 
                               CClassType type,
                               JNameExpression sourceName) {
	super( where );
	this.type = type;
        this.sourceName = sourceName;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public String toString() {
	return qualifiedName();
    }

    /**
     * Returns the name of this type as originally appeared in the source
     * file. The return value may be null if this type name is internally
     * generated by the system.
     */
    public JNameExpression sourceName() {
        return sourceName;
    }
    
    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return type;
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CClassType getClassType() {
	return type;
    }

    /**
     * Returns a qualified name for the type of this
     */
    public String qualifiedName() {
	return type.qualifiedName();
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	try {
	    type = (CClassType)(type.checkType( context ));
	} catch( UnpositionedError e ) {
	    throw e.addPosition( getTokenReference() );
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitTypeNameExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	// do nothing here
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private	CClassType		type;


    /**
     * The name of this type as originally appeared in the source
     * file. It may be null if this type name is generated by the
     * system.
     */
    private JNameExpression sourceName;
}
	
