/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JUnaryExpression.java,v 1.20 2006/08/29 03:28:19 ye-cui Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.MethodRefInstruction;
import org.multijava.util.classfile.NoArgInstruction;
import org.multijava.util.classfile.PushLiteralInstruction;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents unary expressions (unary plus, unary minus,
 * bitwise complement, and logical not).  */
public class JUnaryExpression extends JExpression implements Constants {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	oper		the operator
     * @param	expr		the operand
     */
    public JUnaryExpression( TokenReference where, 
			     int oper, JExpression expr ) {
	super( where );
	this.oper = oper;
	this.expr = expr;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public int oper() { return oper; }
    public JExpression expr() { return expr; }
    public void setExpr(JExpression e) { expr = e; }

    public String toString() {
	String s = "";
	
	switch(oper) {
	  case OPE_PLUS:
	    s = "+";
	    break;
	  case OPE_MINUS:
	    s = "-";
	    break;
	  case OPE_LNOT:
	    s = "!";
	    break;
	  case OPE_BNOT:
	    s = "~";
	    break;
	  default:
	    fail();
	}
	return s + expr.toString();
    }
    
    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return type;
    }

    public void setType(CType t){
	type = t;
    }
    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck( CExpressionContextType context ) 
	throws PositionedError 
    {
	// Expr.typecheck has the side-effect of actually computing an
	// integral literal's value from its image (lexeme).  This is a
	// problem, of course, for the constant expressions -2147483648
	// (Integer.MIN_VALUE) and -9223372036854775808L (Long.MIN_VALUE)
	// because 2147483648 and 9223372036854775808L are not in the range of
	// int and long, respectively. To get around this, we call
	// expr.negate() *before* typecheck() to append "-" to the front of
	// the token image (as if the original numeric image and the - were a
	// single token).  When expr is actually a computed ordinal value (and
	// hence not an image), this trick does not apply (if you still try to
	// do so then some safe math checks will fail to be treated properly).

	isInSafeMathMode  = context.arithmeticMode().equals(AMID_SAFE_MATH);

	if( (expr instanceof JOrdinalLiteral) && oper == OPE_MINUS &&
	    ((JOrdinalLiteral)expr).isImageOnly()) 
	{
	    ((JOrdinalLiteral) expr).negate();
	    oper = OPE_PLUS;
	}
	expr = expr.typecheck( context );

	switch( oper ) {
	case OPE_PLUS:
	case OPE_MINUS:
	    check( context, 
		   expr.getType().isNumeric(), 
		   MjcMessages.UNARY_BADTYPE_PM, expr.getType() );
	    type = ((CNumericType)expr.getType()).unaryPromote();

	    if( expr.isConstant() ) {
		return plusMinusConstantFolding( context );
	    }
	    break;
	    
	case OPE_BNOT:
	    check( context, 
		   expr.getType().isOrdinal(), 
		   MjcMessages.UNARY_BADTYPE_BNOT, expr.getType() );
	    type = ((CNumericType)expr.getType()).unaryPromote();

	    if( expr.isConstant() ) {
		return binaryNotConstantFolding();
	    }
	    break;
	    
	case OPE_LNOT:
	    check( context, 
		   expr.getType() == CStdType.Boolean, 
		   MjcMessages.UNARY_BADTYPE_LNOT, expr.getType() );
	    type = CStdType.Boolean;
	    if( expr.isConstant() ) {
		return logicalNotConstantFolding();
	    }
	    break;

	default:
	    fail();
	}
	return noFolding(this, context);
    }

   /**
    * No Constant Folding (to be overridden in JML)
    */
   protected JExpression noFolding( JExpression expr, CExpressionContextType context ) throws PositionedError{
   	return expr;
   }
    
   /**
    * Constant Folding (+, -)
    */
   /*@ protected normal_behavior
     @  requires (oper == OPE_PLUS || oper == OPE_MINUS)
     @           && expr.isConstant()
     @           // && type.getTypeID() >= TID_INT
     @           ;
     @  modifies expr, type;
     @  ensures  \result instanceof JNumberLiteral;
     @*/      
    protected JExpression plusMinusConstantFolding(CExpressionContextType context)
    	throws PositionedError 
    {	 	
    	if( oper == OPE_MINUS ) {
	    // Must create a new J*Literal as expr may be the value of a final
	    // field which we do not want to alter (via negate()).
	    JNumberLiteral lit = expr.getNumberLiteral();
	    if( type.isOrdinal() ) {
	    	
		if(!context.arithmeticMode().equals(AMID_JAVA_MATH)) {
		    long v = lit.numberValue().longValue();
		    check( context, 
			   ( type == CStdType.Integer
			     ? v != Integer.MIN_VALUE
			     : v != Long.MIN_VALUE ),
			   MjcMessages.INTEGRAL_ARITHMETIC_OVERFLOW );
		}
		return new JOrdinalLiteral( getTokenReference(),
					    -lit.numberValue().longValue(),
					    (CNumericType) type );
	    } else { // isFloatingPoint()
		Number n = (type == CStdType.Float) 
		    ? (Number) new Float( -lit.numberValue().floatValue() )
		    : (Number) new Double( -lit.numberValue().doubleValue() );
		return new JRealLiteral(getTokenReference(), n, type);
	    }
	}
	return expr.convertType(type, context);
    }
    
    /**
    * Constant Folding (~)
    */
    protected JExpression binaryNotConstantFolding() {
    	
    	assertTrue(oper == OPE_BNOT);
    	
    	JOrdinalLiteral	lit = expr.getOrdinalLiteral();
	if( type == CStdType.Long ) {
	    lit = 
		new JOrdinalLiteral( getTokenReference(),
				     new Long(
				      ~lit.numberValue().longValue()),
				     CStdType.Long );
	} else {
	    lit = 
		new JOrdinalLiteral( getTokenReference(),
				     new Integer(
				       ~lit.numberValue().intValue()),
				     CStdType.Integer );
	}
	return lit;	
    }

    /**
    * Constant Folding (!)
    */
    protected JExpression logicalNotConstantFolding() {
    	
    	assertTrue(oper == OPE_LNOT);
    	
    	return 
	    new JBooleanLiteral( getTokenReference(), 
				 !expr.getBooleanLiteral()
				 .booleanValue() );			 
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitUnaryExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	switch (oper) {
	case OPE_PLUS:
	    expr.genCode(code);
	    break;

	case OPE_MINUS:
	    expr.genCode(code);
	    switch (type.getTypeID()) {
	    case TID_FLOAT:
		code.plantInstruction(new NoArgInstruction(opc_fneg));
		break;
	    case TID_LONG:
		if (isInSafeMathMode){
		    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
								   "org/multijava/mjc/SafeIntegralArithmetic",
								   "negate",
								   "(J)J"));
		} else {
		    code.plantInstruction(new NoArgInstruction(opc_lneg));
		}
		break;
	    case TID_DOUBLE:
		code.plantInstruction(new NoArgInstruction(opc_dneg));
		break;
	    default:
                if (isInSafeMathMode){

		    code.plantInstruction(new MethodRefInstruction(opc_invokestatic,
								   "org/multijava/mjc/SafeIntegralArithmetic",
								   "negate",
								   "(I)I"));
		} else {
		    code.plantInstruction(new NoArgInstruction(opc_ineg));
		}
	    }	
	    break;

	case OPE_LNOT:
	    expr.genCode(code);
	    code.plantInstruction(new PushLiteralInstruction(1));
	    code.plantInstruction(new NoArgInstruction(opc_ixor));
	    break;

	case OPE_BNOT:
	    expr.genCode(code);
	    if (type == CStdType.Long) {
		code.plantInstruction(new PushLiteralInstruction((long)-1));
		code.plantInstruction(new NoArgInstruction(opc_lxor));
	    } else {
		code.plantInstruction(new PushLiteralInstruction((int)-1));
		code.plantInstruction(new NoArgInstruction(opc_ixor));
	    }
	    break;

	default:
	    fail();
	}
    }
    
    /**
     * Generates a sequence of bytecodes to branch on a label
     * This method helps to handle heavy optimizables conditions
     * @param	code		the code list
     */
    public void genBranch(boolean cond, CodeSequence code, CodeLabel label) {
	assertTrue(oper == OPE_LNOT);

	expr.genBranch(!cond, code, label);
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    protected int			oper;
    protected JExpression		expr;
    protected CType		type;
    private boolean isInSafeMathMode  = false;

}
