/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JUnaryPromote.java,v 1.7 2006/08/08 17:18:43 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.ClassRefInstruction;
import org.multijava.util.compiler.PositionedError;

/**
 * This class promotes an arithmetic expression to a new type.  See JLS2 5.2
 * A null litteral is wrapped in some circumstances (e.g., return null)
 */
public class JUnaryPromote extends JExpression {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a promotion of the given expression to the given type.
     * @param	expr	the expression to be promoted
     * @param	type	the type to which the expression should be promoted
     */
    public JUnaryPromote( JExpression expr, CType type ) {
	super( expr.getTokenReference() );

	this.expr = expr;
	this.type = type;

	if( !expr.isAssignableTo(type) ) {
	    needCheck = true;
	}
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression expr() { return expr; }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return type;
    }

    public boolean isCheckNeeded() {
	return needCheck;
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CContextType context) {
	return expr.isNonNull(context);
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>)
     * @exception PositionedError if the check fails */
    public JExpression typecheck(CExpressionContextType context) 
	throws PositionedError 
    {
	if( type.equals(CStdType.String) 
	    && expr.getType().isReference() 
	    && !(expr instanceof JNullLiteral) ) {
	    JNameExpression name = 
		new JNameExpression( getTokenReference(),
				     new JCheckedExpression( getTokenReference(), 
							     expr ),
				    "toString");

	    JMethodCallExpression invoke = 
		new JMethodCallExpression( getTokenReference(), 
					   name, 
					   JExpression.EMPTY );
	    return invoke.typecheck( context );
	}

	return this;
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitUnaryPromoteExpression( this );
   }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	expr.genCode(code);

	if (type.isNumeric()) {
	    ((CNumericType)expr.getType()).genCastTo((CNumericType)type, code);
	} else if (needCheck) {
	    code.plantInstruction(new ClassRefInstruction(opc_checkcast,
							  ((CClassType)type).qualifiedName()));
	}
    }

    /* 
     * By rzakhejm
     */
    public String toString() {
    	return expr.toString();
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		expr;
    private CType		type;
    private boolean		needCheck;
}
