/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JVariableDeclarationStatement.java,v 1.7 2005/06/12 19:37:48 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * A local variable declaration statement declares one or more local
 * variable names.  JLS 14.3: Local Variable Declaration Statement
 */
public class JVariableDeclarationStatement extends JStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	vars		the variables declared by this statement
     */
    public JVariableDeclarationStatement( TokenReference where, 
					  JVariableDefinition[] vars, 
					  JavaStyleComment[] comments ) {
	super( where, comments);

	this.vars = vars;
    }

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     * @param	var		the variable declared by this statement
     */
    public JVariableDeclarationStatement( TokenReference where, 
					  JVariableDefinition var, 
					  JavaStyleComment[] comments ) {
	super( where, comments );

	this.vars = new JVariableDefinition[] {var};
    }

    /**
     * Returns an array of variable definition declared by this statement
     */
    public JVariableDefinition[] getVars() {
	return vars;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Sets the variables to be for variables
     */
    public void setIsInFor() {
	for (int i = 0; i < this.vars.length; i++) {
	    vars[i].setIsLoopVariable();
	}
    }

    /**
     * Unsets the variables to be for variables
     */
    public void unsetIsInFor() {
	for (int i = 0; i < this.vars.length; i++) {
	    vars[i].unsetIsLoopVariable();
	}
    }

    /**
     * Typechecks the statement and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @exception PositionedError if the check fails */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	for( int i = 0; i < this.vars.length; i++ ) {
	    try {
		
		context.addVariable( vars[i] );
		vars[i].typecheck( context );
		if( vars[i].hasInitializer() ) {
		    context.initializeVariable( vars[i] );
		}
		
		// scdaniel: universe runtime checks
		if ( CTopLevel.getCompiler().universeDynChecks() ) {
		    JExpression newObject = vars[i].getValue();
			
		    /* if (newObject == null) then the variable
		     * has no initializer */
		    if ( newObject != null ) {
			if ( newObject instanceof JUnaryPromote )
			    newObject = ((JUnaryPromote) vars[i].getValue()).expr();
				
			/* create transformation */
			newObject.buildUniverseDynChecks(context.createExpressionContext(),
						 new JNameExpression(TokenReference.NO_REF, vars[i].ident()));
		    }
		}
		
	    } catch( UnpositionedError e ) {
		throw new CLineError( getTokenReference(), 
				      e.getFormattedMessage() );
	    }
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitVariableDeclarationStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	for (int i = 0; i < this.vars.length; i++) {
	    if (vars[i].getValue() != null) {
		vars[i].getValue().genCode(code);
		vars[i].genStore(code);

		/*
		 * scdaniel: universe runtime checks
		 * 
		 * Register newly created objects to the UrtHashtable after
		 * they have been stored into their variable slot.
		 */
		if ( CTopLevel.getCompiler().universeDynChecks() ) {
		    JExpression newObject = vars[i].getValue();
			
		    /* if (newObject == null) then the variable
		     * has no initializer */
		    if ( newObject != null ) {
			if ( newObject instanceof JUnaryPromote )
			    newObject = ((JUnaryPromote) vars[i].getValue()).expr();
	
			/* generate code */
			newObject.genUniverseDynCheckCode(code);
		    }
		}
		
	    }
	}
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JVariableDefinition[]		vars;
}
