/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JVariableDefinition.java,v 1.22 2008/05/14 18:16:33 wdietl Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a local variable definition in the syntax tree.
 */
public class JVariableDefinition extends JLocalVariable {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.  This method is directly
     * called by the parser.  */
	public JVariableDefinition( TokenReference where,
			long modifiers, 
			CType type,
			String ident,
			JExpression initializer ) {
		super( where, modifiers, DES_LOCAL_VAR, type, ident, initializer );
		assertTrue(type != null);
	}

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * hasInitializer
     * @return	true if there is an initializer
     */
    public /*@ pure @*/ boolean hasInitializer() {
	return getValue() != null;
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type
     * signatures of everything (imported class files, classes being
     * compiled, methods, fields, etc...)  needed for the later
     * passes.  This information is stored in a CCompilationUnit
     * instance and instances of CMember that are bound to the AST.
     * Also adds things like the default constructor and the initializer
     * method to the AST (these are suppressed during pretty-printing).
     *
     * @param	context		the current context
     * @return	true iff sub tree is correct enought to check code
     */
    public void checkInterface(CClassContextType context) {
	try {
	    type = type.checkType(context);
	    type = CUniverseServices.setDefaultVariableAgain(type);	    
	} catch (UnpositionedError cue) {
	    /* typecheck will catch error */
	}
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the definition and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @exception PositionedError if the check fails */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError {
    	typecheck(context, true);
    }
	
    public void typecheck( CFlowControlContextType context,
    		boolean FieldWasAlreadyAsked ) 
	throws PositionedError 
    {
	try {
	    type = type.checkType( context );
	    
	    type = CUniverseServices.setDefaultVariableAgain(type);

	    if( expr != null ) {
		// special case for array initializers, which need to
		// dimensions of the variable type for typechecking
		if( expr instanceof JArrayInitializer ) {
		    ((JArrayInitializer) expr).setType( type );
		    if ( !type.isArrayType() ) {
			check( context, false, MjcMessages.VAR_INIT_BADTYPE2, 
			       ident() );
			return;
		    }
		}

		CExpressionContextType expressionContext = 
		    context.createExpressionContext();
		expr = expr.typecheck( expressionContext );
		if( expr instanceof JTypeNameExpression ) {
		    check( context,
			   false,
			   MjcMessages.VAR_UNKNOWN,
			   ((JTypeNameExpression) expr).qualifiedName().
			   replace('/','.'));
		}
		// WMD if the programmer did not specify any Universe modifiers,
		// we try to propagate them from the initializer expression.
		if( type instanceof CClassType &&
			CTopLevel.getCompiler().universeChecks() &&
			expr.getType().isAlwaysAssignableToIgnoreUniverses(type) &&
			CUniverseServices.isPartial((CClassType)type) ) {
         
			if ( (!context.isInInitializer() || !FieldWasAlreadyAsked)  &&
				 !(expr instanceof JNullLiteral) &&
				 !(expr instanceof JThisExpression)) {
				// added the last condition above to ensure that we are not in a field initializer.
				// Otherwise, one might have cyclic dependencies that we don't take care of correctly.
				type = CUniverseServices.propagate((CClassType)type, (CClassType)expr.getType());

				warn( context,
						CUniverseMessages.TYPE_PROPAGATED,
						this.ident(), type.toVerboseString() );
			} else {
				if( !expr.isAssignableTo( type ) ) {
					warn( context,
							CUniverseMessages.TYPE_DEFAULTED_GIVES_ERROR,
							new Object[] { this.ident(),
								type.toVerboseString(), expr.getType().toVerboseString() } );			
				} else {
					warn( context,
							CUniverseMessages.TYPE_DEFAULTED,
							this.ident(), type.toVerboseString() );
				}
			}
		}
		// System.out.println("JVD UTS: " + CTopLevel.getCompiler().universeChecks());
		// System.out.println("JVD expr: " + expr.getType().toVerboseString());
		// System.out.println("JVD Type: " + type.toVerboseString());

		if( !expr.isAssignableTo( type ) ) {				
		    check( context,
			   false,
			   MjcMessages.VAR_INIT_BADTYPE,
			   ident(), 
			   expr.getType().isTypeVariable()?
			       ((CTypeVariable)expr.getType()).getIdent():(expr.getType().toVerboseString())); // WMD
		} else {
		    if (expr.getType().needsUncheckedConversion(type)) {
		        warn(context,
		             MjcMessages.UNCHECKED_CONVERSION,
                             expr.getType(),
                             type);
		    }

		    expr = expr.convertType( type, expressionContext );
		}
		if (expr.isNonNull(context) || expr instanceof JNewArrayExpression) {
		   context.addFANonNull(this);
		}
	    } else {
	    	// no initializer expression
			if( type instanceof CClassType &&
				CTopLevel.getCompiler().universeChecks() &&
				CUniverseServices.isPartial((CClassType)type) ) {
				warn( context,
					CUniverseMessages.TYPE_DEFAULTED,
					this.ident(), type.toVerboseString() );
			}
	    }

	    // Checks for Generic Universe Types
	    if (CTopLevel.getCompiler().universeChecks()) {
	    	typecheckUniverse(context);
	    }

	    super.typecheck(context);
	} catch( UnpositionedError cue ) {
	    throw cue.addPosition(getTokenReference());
	}
    }

	/**
	 * Checks static correctness of Generic Universe Types.
	 * 
	 * Checks performed:
	 * - Static variables must not have rep modifiers
	 * 
	 * @param context The current typechecking context.
	 * 
	 * By WMD
	 * By RZU
	 */
	private void typecheckUniverse(CFlowControlContextType context)
			throws PositionedError {
		// Ensure that static variables don't contain rep modifiers
	    if (context.isStatic()) {
	    	check(context,
	    			!(type instanceof CClassType)
	    			|| !CUniverseServices.containsRepModifier(
	    					(CClassType) type),
	    			CUniverseMessages.REP_IN_STATIC_FORBIDDEN);
	    }
	}

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitVariableDefinition( this );
    }
}
