/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: JWhileStatement.java,v 1.9 2006/11/27 15:36:26 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;

/**
 * This class represents a while statement in the AST.
 * JLS 14.10: While Statement
 */
public class JWhileStatement extends JLoopStatement {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * @param	where		the line of this node in the source code
     * @param	cond		the expression to evaluate
     * @param	body		the loop body
     * @param	comments	the comments associated with this statement
     */
    public JWhileStatement( TokenReference where, JExpression cond, 
			    JStatement body, JavaStyleComment[] comments ) {
	super( where, comments );
	this.cond = cond;
	this.body = body;
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks this statement in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context	the context in which this appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CFlowControlContextType context ) 
	throws PositionedError 
    {
	try {
	    CExpressionContextType condContext = context.createExpressionContext();
	    cond = cond.typecheck( condContext );
	    check( context, cond.getType() == CStdType.Boolean,
		   MjcMessages.WHILE_COND_NOTBOOLEAN, cond.getType() );

	    if( !cond.isConstant() ) {
		CFlowControlContextType neverContext = context.cloneContext();
		CLoopContext self = context.createLoopContext( this );

		Object[] nonNulls=cond.getFANonNulls(context);
		Object[] nulls   =cond.getFANulls(context);
		self.addFANonNulls(nonNulls);
		neverContext.addFANonNulls(nulls);

		body.typecheck( self );
		self.doneWithCheckingBody();

		if (hasBreak()) {
		    if (self.isReachable()) {
			self.merge( getBreakContextSummary() );
		    } else {
			self.adopt( getBreakContextSummary() );
		    }
		}
		
		self.checkingComplete();
		
		if (context.isReachable()) {
		    context.merge( neverContext );

		} else {
		    context.adopt( neverContext );
		}
		if (!hasBreak()) {
		    context.addFANonNulls(nulls);
		}
	    } else {
		if( !cond.getBooleanLiteral().booleanValue() ) {
		    // This is a 'while (false)'
		    throw new CLineError( body.getTokenReference(), 
					  MjcMessages.STATEMENT_UNREACHABLE );
		}

		// This is a 'while (true)'
		if( cond instanceof JAssignmentExpression ) {
		    context.reportTrouble(
		       new CWarning( getTokenReference(),
				     MjcMessages.ASSIGNMENT_IN_CONDITION ));
		}

		CLoopContext self = context.createLoopContext( this );

		body.typecheck( self );
		self.doneWithCheckingBody();

		if( hasBreak() ) {
		    self.adopt( getBreakContextSummary() );
		} 

		self.checkingComplete();
		context.setReachable( hasBreak() );
	    }
	} catch( CBlockError e ) {
	    context.reportTrouble( e );
	}
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitWhileStatement( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	code.setLineNumber(getTokenReference().line());

	code.pushContext(this);

	if (cond.isConstant() && cond.getBooleanLiteral().booleanValue()) {
	    // while (true)
	    code.plantLabel(getContinueLabel());
	    body.genCode(code);
	    code.plantLabelRef(opc_goto, getContinueLabel());
	    code.plantLabel(getBreakLabel());
	} else {
	    CodeLabel		bodyLabel = new CodeLabel();

	    code.plantLabelRef(opc_goto, getContinueLabel());	//		GOTO cont
	    code.plantLabel(bodyLabel);			//	body:
	    body.genCode(code);				//		BODY CODE
	    code.plantLabel(getContinueLabel());		//	cont:
	    cond.genBranch(true, code, bodyLabel);      	//		EXPR CODE; IFNE body
	    code.plantLabel(getBreakLabel());			//	end:	...
	}

	code.popContext(this);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public JExpression cond() { return cond; }
    public JStatement body() { return body; }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private JExpression		cond;
    private JStatement		body;
}
