/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MJGenericFunctionDecl.java,v 1.24 2006/12/20 14:35:21 perryjames Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;

/**
 * This class represents a group of method declarations, all sharing
 * the same name, that together form the top of one or more overloaded
 * external generic function lattices.  It corresponds to the anchor
 * class for an external generic function.  */
public class MJGenericFunctionDecl extends JClassDeclaration 
    implements Constants, Cloneable
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a generic function anchor declaration in the parsing tree.
     * @param	where	the line of this node in the source code
     * @param	modifiers the modifiers of this anchor
     * @param	ident	the short name of thie anchor
     * @param	methods	the MJTopLevelMethodDeclarations that make up this 
     *			generic function, must have non-zero length and all
     *			non-null members
     */
    public MJGenericFunctionDecl( TokenReference where, long modifiers, 
				  String ident, 
				  MJTopLevelMethodDeclaration[] methods ) {
	super( where, modifiers, ident + MJ_ANCHOR, CTypeVariable.EMPTY,null,
	       CClassType.EMPTY, new ArrayList(),
	       new ArrayList(), new JPhylum[0], null, null );
	this.externalMethods = methods;
    }

    /**
     * Generates the signature object for this.
     *
     * @param compiler	the compiler instance for which this signature is 
     *			generated
     * @param owner	the class signature singleton for the logical outer 
     *			class of this, or null if this is a top level 
     *			declaration
     * @param host	the signature singleton of the context in which this
     *			is declared, a <code>CCompilationUnit</code> for 
     *			top-level declarations
     * @param prefix 	the prefix prepended to this declaration's
     *			identifier to achieve the fully qualified
     *			name, just the package name (using '/'
     *			separators) for top-level classes, package
     *			name plus $-delimited outer class names plus
     *			synthetic index for inner classes
     * @param isAnon	true if this is an anonymous class, in which 
     *			case the fully qualified name is just 
     *			<code>prefix</code> 
     * @param isMember	true if this is a member type, i.e., a nested type
     *			that is not a local type or an anonymous class
     */
    protected CSourceClass makeSignature( /*@ non_null */ Main compiler, /*@ nullable */ CClass owner, 
					  /*@ non_null */ CMemberHost host, /*@ non_null */ String prefix,
					  boolean isAnon, boolean isMember ) {
	CMethod[] mSigs = new CMethod[ externalMethods.length ];
	for( int i=0; i < mSigs.length; i++ ) {
	    mSigs[i] = externalMethods[i].getMethod();
	}
	return new CSourceGFCollection( compiler, owner, host, 
					getTokenReference(), modifiers, ident, 
					prefix + ident, isAnon, isMember,
					isDeprecated(), mSigs );
    }

    /**
     * Registers the signature of this declaration with the name space
     * manager in <code>CTopLevel</code>.
     *
     * <pre><jml>
         also requires sourceClass != null && 
	 		sourceClass instanceof CSourceGFCollection;
     * </jml></pre>
     *
     * @return true if this is a new signature or false if this is a 
     *	       redeclaration
     *
     * @see CTopLevel#addGFCollection( CGFCollection ) */
    protected boolean registerSignature() {
	return CTopLevel.
	    addGFCollection( (CSourceGFCollection) sourceClass );
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /** Return the top-level methods in this GF */
    public MJTopLevelMethodDeclaration[] externalMethods()
    {
	return externalMethods;
    }
    
    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things generally look
     * OK.  This pass gathers information about the type
     * signatures of everything (imported class files, classes being
     * compiled, methods, fields, etc...)  needed for the later
     * passes.  This information is stored in a CCompilationUnit
     * instance and instances of CMember that are bound to the AST.
     * Also adds things like the default constructor and the initializer
     * method to the AST (these are suppressed during pretty-printing).
     *
     * @param		context		the context in which this
     *					decl appears
     * @exception	PositionedError	an error with reference to the 
     *					source file
     */
    public void checkInterface( CContextType context ) 
	throws PositionedError 
    {
	JMethodDeclarationType dup = 
	    checkForDups( externalMethods );

	if( dup != null ) {
	    throw new PositionedError( dup.getTokenReference(), 
				       MjcMessages.METHOD_REDEFINE,
				       dup.getMethod().ident() );
	}

	context.registerGFDecl(methodIdent(), (CSourceGFCollection) sourceClass);

	super.checkInterface( context );
    }

    
    /**
     * Computes the values of specializer expressions used to dispatch on
     * compile-time constants.
     *
     * @param	context		the context in which this class 
     *				declaration appears
     * @exception PositionedError if the check fails */
    public void resolveSpecializers(CContextType context)
	throws PositionedError {

	CContextType self = createContext( context );

	for (int i = 0; i < externalMethods.length; i++) {
	    externalMethods[i].resolveSpecializers( self );
	}
	
	super.resolveSpecializers(context);
    }

    
    /**
     * Makes sure that all in-scope external generic functions are
     * added to the appropriate augmentation maps before top method
     * searches occur in later passes.  */
    public void resolveExtMethods( CContextType context ) {
	CClassContextType selfC = createContext( context );

	for (int i = 0; i < externalMethods.length; i++) {
	    externalMethods[i].resolveExtMethods(selfC);
	}

	super.resolveExtMethods(context);
    }

    /**
     * Finds the top method of every declared method.  Called after
     * the resolveSpecializers pass.  This cannot be done before then
     * because the external generic function mappings are not complete
     * until the end of the checkInterface pass and the constant value
     * specializers are not known until after the resolveSpecializers
     * pass.  This must be done before the typecheck pass so that all
     * specialized argument positions for generic functions are known
     * for ambiguity checking. */
    public void resolveTopMethods() throws PositionedError {
	for (int i = 0; i < externalMethods.length; i++) {
	    externalMethods[i].resolveTopMethods();
	}

	super.resolveTopMethods();
    }

    
    /**
     * Returns the appropriate warning message when another class has
     * been declared whose signature matches the signature of this.
     * Overridden in subclass.  */
    protected CWarning uniqueSourceWarning() {
	return new CWarning( getTokenReference(), 
			     MjcMessages.EXT_GF_REDEFINED2 );
    }

    /**
     * Typechecks this generic function declaration in the context in which it
     * appears.  Mutates the context to record the information learned
     * during checking.
     *
     * @param context the context in which this type declaration appears
     * @exception PositionedError if any checks fail */
    public void typecheck( CContextType context ) throws PositionedError {
 	for( int i=0; i < externalMethods.length; i++ ) {
 	    MJTopLevelMethodDeclaration method = externalMethods[i];
	    CClass receiverClass = method.getOpenClassType().getCClass();
	    CExtMethodContext methodContext =
		context.createExtMethodContext( sourceClass, receiverClass );
	    method.typecheck( methodContext );
	}

	super.typecheck( context );
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * Produces a copy of this declaration, including a deep copy of
     * the external method declarations housed here-in.
     *
     * @return	a copy of this declaration. */
    public Object clone() {
	try {
  	    MJGenericFunctionDecl newSelf = 
  		(MJGenericFunctionDecl) super.clone();
 	    MJTopLevelMethodDeclaration[] newExternalMethods =
 		new MJTopLevelMethodDeclaration[ externalMethods.length ];
 	    for( int i=0; i<newExternalMethods.length; i++ ) {
 		newExternalMethods[i] = 
 		    (MJTopLevelMethodDeclaration) externalMethods[i].clone();
  	    }
 	    newSelf.externalMethods = newExternalMethods;
  	    return newSelf;
  	} catch( CloneNotSupportedException e ) {
	    // should never get here
	    e.printStackTrace();
	    return null;
	}
    }


    /**
     * Returns the ident of the methods contained in this generic
     * function.  This is distinct from the value returned by
     * <CODE>ident()</CODE> as that value represents the name of the
     * anchor class for the generic function.  
     * @see JTypeDeclaration#ident()
     */
    protected String methodIdent() {
	return ident.substring(0,ident.length()-MJ_ANCHOR.length());
    }

    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	v		the visitor
     */
    public void accept( /*@non_null@*/ MjcVisitor v ) {
	v.visitGenericFunctionDecl( this );
    }

    /**
     * Returns an array of the methods to be considered for
     * refactoring into dispatchers during the translateMJ pass.
     *
     * @see JTypeDeclarationType#translateMJ(CContextType) */
    protected JMethodDeclarationType[] methodsToBeWrapped() {
	return externalMethods;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * Stores the external methods declared within this anchor
     * declaration.  These are separate from the methods stored in a
     * regular Java type declaration, since typechecking performed on
     * these is distinct from that for regular method.  The generated
     * code will be accessed through the chain of responsibility
     * pattern.  */
    MJTopLevelMethodDeclaration[] externalMethods;
}
