/*
 * Copyright (C) 2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MJMathModeExpression.java,v 1.8 2006/08/04 18:22:39 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.PositionedError;

/**
 * This AST node is used to change the arithmetic mode for a given expression. */

public class MJMathModeExpression extends JExpression implements Constants {

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     */
    public MJMathModeExpression( TokenReference where,
			     String operatorName,
			     JExpression operand,
			     byte mode ) {
	super( where );
	this.operatorName = operatorName;
	this.operand = operand;
	this.arithmeticMode = mode;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression expr() { return operand; }

    public String operatorName() {
	return operatorName;
    }

    public String toString() {
	return operatorName() + "(" + expr().toString() + ")";
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return operand.getType();
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CExpressionContextType context) {
	    return operand.isNonNull(context);
    }

    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>) */
    public JExpression typecheck( CExpressionContextType context ) 
    	throws PositionedError 
    {
	byte orig_arithmetic_mode_in_context = context.arithmeticMode().getMathModeId();

	/*
	if(orig_arithmetic_mode_in_context == arithmeticMode){
	    // This WarnExpression is a noop in the given context.  Hence remove
	    // it from the AST (by returning the typecheck'ed operand.
	    return operand.typecheck( context );
	}
	*/

	context.setArithmeticMode(arithmeticMode);
	
	JExpression result = operand.typecheck( context );
	context.setArithmeticMode(orig_arithmetic_mode_in_context);
	// Now that type checking is done, discard this node and simply return
	// the type checked operand.
	return result;
    }
    
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitMathModeExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	throw new RuntimeException( "Method should not be called on a "
				    + "MJMathModeExpression" );
    }

    // ----------------------------------------------------------------------

    private String operatorName;
    private JExpression operand;
    private byte arithmeticMode;
}
