/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MJTopLevelAbstractMethodDeclaration.java,v 1.5 2005/04/26 02:39:33 kui_dai Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.MessageDescription;

public class MJTopLevelAbstractMethodDeclaration extends MJTopLevelMethodDeclaration
{
    
    /**
     *
     * This class represents an abstract MultiJava external method in
     * the syntax tree. Special behevior is required for abstract
     * external methods for Relaxed MultiJava 
     *
     */
    
    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------
    
    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	modifiers	list of modifiers
     * @param	returnType	the return type of this method
     * @param	openClassType	the class that this method extends
     * @param	ident		the name of this method
     * @param	parameters	the parameters of this method
     * @param	exceptions	the exceptions throw by this method
     * @param	body		the body of this method
     * @param	javadoc		javadoc comments, including deprecated flag
     * @param	comments	non-javadoc comments
     */
    public MJTopLevelAbstractMethodDeclaration( TokenReference where,
						long modifiers, 
						CType returnType,
						CType openClassType,
						String ident,
						JFormalParameter[] parameters,
						CClassType[] exceptions,
						JBlock body,
						JavadocComment javadoc,
						JavaStyleComment[] comments ) {
	// de-abstract-ify any abstract methods and create a
	// body to throw an exception
	super(where, modifiers & ~ACC_ABSTRACT, CTypeVariable.EMPTY, returnType, openClassType, 
	      ident, parameters, exceptions, 
	      make_abstract_body(where), 
	      javadoc, comments );
    }
    
    // Make a body to execute whenever an "abstract" method ends up
    // being called at runtime Simply throws an Error with a
    // (hopefully) useful message
    private static JBlock make_abstract_body(TokenReference where)
    {
	String message = where + ": invoking abstract top-level method";
  	JExpression args[] = { new JStringLiteral(where,message) };
  	JExpression expr =
	    new JNewObjectExpression(where, CStdType.RMJRuntimeException,
				     null, args);
  	JStatement stmt[] = { new JThrowStatement(where,expr,null) };
  	JBlock body = new JBlock(where,stmt,null);
	return body;
    }

    // Give a little warning whenever an abstract generic external
    // method is declared
    public CMember checkInterface( CContextType context ) 
	throws PositionedError
    {
	context.reportTrouble(
	      new CWarning( getTokenReference(), 
			    MjcMessages.RMJ_ABSTRACT_TOP_LEVEL_METHOD,
			    ident()) );
	
	CMember result = super.checkInterface(context);

	// Use this opportunity to decorate the CSourceMethod
	check(context, 
	      result instanceof CSourceMethod,
	      new MessageDescription("Internal compiler error in " + 
				     getClass().toString(), "", 0),
	      ident());
	CSourceMethod smet = (CSourceMethod) result;
	smet.markAsTopLevelAbstractMethod();

	return result;
    }
}
	

