/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MJTopLevelMethodDeclaration.java,v 1.19 2007/02/08 14:04:55 leavens Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.PositionedError;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.UnpositionedError;
import org.multijava.util.compiler.CWarning;

public class MJTopLevelMethodDeclaration extends JMethodDeclaration
    implements MJTopLevelDeclaration, Cloneable {

    /**
     * This class represents a MultiJava external method in the syntax tree
     */

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree
     * This method is directly called by the parser
     * @param	where		the line of this node in the source code
     * @param	modifiers	list of modifiers
     * @param	returnType	the return type of this method
     * @param	openClassType	the class that this method extends
     * @param	ident		the name of this method
     * @param	parameters	the parameters of this method
     * @param	exceptions	the exceptions throw by this method
     * @param	body		the body of this method
     * @param	javadoc		javadoc comments, including deprecated flag
     * @param	comments	non-javadoc comments
     */
    public MJTopLevelMethodDeclaration( TokenReference where,
					long modifiers, 
					CTypeVariable[] typevariables,
					CType returnType,
					CType openClassType,
					String ident,
					JFormalParameter[] parameters,
					CClassType[] exceptions,
					JBlock body,
					JavadocComment javadoc,
					JavaStyleComment[] comments ) {
	super(where, modifiers,typevariables, returnType, ident, parameters, 
	      exceptions, body, javadoc, comments );

	this.openClassType = openClassType;
    }


    /**
     * Registers with the compilation session that an external generic
     * function to which this top-level method will belong is being
     * compiled.  This registration occurs during the parsing pass.
     *
     * <pre><jml>
     * requires packageName != null && 
     *		(* any separators in packageName are 
     *             slashes ('/') not dots ('.') *);
     * </jml></pre>
     *
     */
    protected void registerPendingGFCollection(String packageName) {
	CTopLevel.registerPendingGFCollection(packageName + ident());
    }

    // ---------------------------------------------------------------------
    // ACCESSORS
    // ---------------------------------------------------------------------
    public CType getOpenClassType() {
	return openClassType;
    }

    /**
     * Returns the static type of this method's receiver, i.e., the
     * receiver type of the top method of this generic function.
     *
     * @return	the static type of this method's receiver
     */
    public CType receiverStaticType() {
	assertTrue( receiverStaticType != null); // This should be set in
			                         // the typecheck pass.
	return receiverStaticType;
    }

    /**
     * Indicates that this member is external.
     * @return	true
     */
    public boolean isExternal() {
	return true;
    }

    /**
     * Indicates whether this method uses multiple dispatch
     *
     * @return	true iff a parameter of this method has a dynamic dispatch
     *		annotation or, for an external method, this methods receiver
     *		is of a different type than the generic function's top method
     */
    public boolean usesMultipleDispatch() {
	return 
	    !openClassType.equals( receiverStaticType() )
	    || super.usesMultipleDispatch();
    }

    /**
     * Compares this method to a given method and returns 0 if the
     * methods belong to the same generic function, otherwise returns
     * -1 or +1 to sort the methods.  This comparator is used to
     * separate methods by generic function in the
     * MJGenericFunctionDispatcher class.
     *
     * @param	o	the object to be compared against, must be a
     *			JMethodDeclarationType
     * @return	-1, +1, or 0
     * @exception ClassCastException if <code>o</code> is not an instance of
     *					CType */
    public int compareTo(/*@ non_null @*/ Object o) throws ClassCastException {
	// if both o and this are 
	// MJTopLevelMethodDeclarations then algorithm considers
	// static receiver types before considering identifier and
	// other argument, otherwise just use the super compareTo
	// method that ignores the receiver static types
	if( o instanceof MJTopLevelMethodDeclaration )  {
	    MJTopLevelMethodDeclaration m2 = (MJTopLevelMethodDeclaration) o;
	    CType t1 = receiverStaticType();
	    CType t2 = m2.receiverStaticType();
	    // if the receiver static types are not identical then
	    // we sort by subtype relationship
	    if( !(t1.equals( t2 )) ) {
		return t1.compareTo( t2 );
	    }
	}
	return super.compareTo( o );
    }

    /**
     * Creates a shallow copy of this.
     */
    public Object clone() throws CloneNotSupportedException {
	return super.clone();
    }

    // ----------------------------------------------------------------------
    // INTERFACE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Checks the basic interfaces to make sure things
     * generally look OK.  This pass gathers information about
     * the type signatures of everything (imported class
     * files, classes being compiled, methods, fields, etc...)
     * needed for the later passes.  This information is stored 
     * in a context hierarchy that is bound to the AST.
     *
     * @param	context		the context in which this method appears
     * @return	a representation of the interface and body of this method 
     *		(iff no errors are detected)
     * @exception PositionedError if any checks fail 
     */
    public CMember checkInterface( CContextType context ) 
	throws PositionedError
    {
	check( context,
	       context instanceof CCompilationUnitContextType,
	       MjcMessages.TOP_LEVEL_METHOD_MISPLACED,
	       ident() );	
	CContextType self = createExtendedCompilationUnitContext
	    (context, typevariables);
	Debug.breakMe();
	try {
	   openClassType = openClassType.checkType( self );
	} catch (UnpositionedError e) {
	    throw e.addPosition(getTokenReference());
	}
	
	CClass openClass = openClassType.getCClass(); // owner
	MemberAccess access = makeMemberAccess( self, openClass );
        
	//special type checking when method receiver is a generic class
 
	if (openClass.isGenericClass()){
	    CTypeVariable[] originalparams = openClass.getTypeVariables();
	    CClassType[] openclassparams = openClassType.getArguments();
	    for (int i = 0 ; i < originalparams.length; i++){
		//each type variable has to have same bound
		//with the correspoing typevariable bounding of declaring
		//class;
		check(context,
		      originalparams[i].hasSameBound(openclassparams[i]),
		      MjcMessages.TOP_RECEIVER_INVALID,
		      ident()
		      );
	    }
	}
	access.checkExternalMethodModifiers( self, this );

	if ( access.isPublic()
	     && !getTokenReference().name().startsWith(ident() + ".") ) {
	    context.reportTrouble(
		new CWarning( getTokenReference(), 
			      MjcMessages.METHOD_NAME_FILENAME,
			      ident(), 
			      getTokenReference().name() ) );
	}
	
	CMember method = checkInterfaceType( self, access, ident() );
	
	return method;
    }

    public CExtendedCompilationUnitContext createExtendedCompilationUnitContext
	(CContextType parent,CTypeVariable[] typevariables) {
	return new CExtendedCompilationUnitContext(parent,typevariables);
    }
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	// !FIXME! genComments(p);
	p.visitTopLevelMethodDeclaration( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	fail(); // nothing to do here
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * The class that this method extends */
    private CType openClassType;

    /**
     * The static type of this method's receiver, i.e., the receiver
     * type of the top method of this generic function */
    private CType receiverStaticType = null;
}
	
