/*
 * Copyright (C) 2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MJWarnExpression.java,v 1.9 2006/08/04 18:22:39 perryjames Exp $
 */

package org.multijava.mjc;

import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.PositionedError;

/**
 * This AST node is used to enable or disable the compile-time and run-time
 * checking of integral arithmetic overflow. */

public class MJWarnExpression extends JExpression implements Constants {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     */
    public MJWarnExpression( TokenReference where,
			     String operatorName,
			     JExpression operand,
			     boolean enable_safe_math ) {
	this( where, operatorName, operand, enable_safe_math, false);
    }

    /**
     * Construct a node in the parsing tree.
     * @param	where		the line of this node in the source code
     */
    public MJWarnExpression( TokenReference where, 
			     String operatorName,
			     JExpression operand,
			     boolean enable_safe_math,
			     boolean scope_is_one_op) {
	super( where );
	this.operatorName = operatorName;
	this.operand = operand;
	this.enable_safe_math = enable_safe_math;
	this.scope_is_one_op = scope_is_one_op;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    public JExpression expr() { return operand; }

    public String operatorName() {
	return operatorName;
    }

    public String toString() {
	return operatorName() + "(" + expr().toString() + ")";
    }

    /**
     * Compute the type of this expression (called after parsing)
     * @return the type of this expression
     */
    public CType getType() {
	return operand.getType();
    }

    /**
     * Returns true iff the value represented by this expression is non-null
     */
    public /*@pure*/ boolean isNonNull(/*@non_null@*/ CExpressionContextType context) {
	    return operand.isNonNull(context);
    }


    // ----------------------------------------------------------------------
    // CODE CHECKING
    // ----------------------------------------------------------------------

    /**
     * Typechecks the expression and mutates the context to record
     * information gathered during typechecking.
     *
     * @param context the context in which this expression appears
     * @return  a desugared Java expression (see <code>JExpression.typecheck()</code>) */
    public JExpression typecheck( CExpressionContextType context ) 
    	throws PositionedError 
    {
	byte orig_arithmetic_mode_in_context = context.arithmeticMode().getMathModeId();

	if((orig_arithmetic_mode_in_context != AMID_JAVA_MATH && enable_safe_math)||
	   (orig_arithmetic_mode_in_context == AMID_JAVA_MATH && !enable_safe_math)) {
	    // This WarnExpression is a noop in the given context.  Hence remove
	    // it from the AST (by returning the typecheck'ed operand.
	    return operand.typecheck( context );
	}

	if(scope_is_one_op) {
	    partly_desugar( context );
	}

	if(enable_safe_math){
		context.setArithmeticMode((byte)17);	//!FIXME!
	} else {
		context.setArithmeticMode(AMID_JAVA_MATH);	
	}
	
	JExpression result = operand.typecheck( context );
	context.setArithmeticMode(orig_arithmetic_mode_in_context);
	// Now that type checking is done, discard this node and simply return
	// the type checked operand.
	return result;
    }

    /** Requires this.scope_is_one_op.  Desugar _*warn_op by inserting a
     * _warn/_nowarn operator in the operand expression. E.g.
     * unsafe_math_op(-E) becomes unsafe_math_op(-safe_math(E)), and
     * safe_math_op(E1 + E2) becomes safe_math_op(unsafe_math(E1) + unsafe_math(E2)).
     */
    private void partly_desugar( CExpressionContextType context ) 
    	throws PositionedError 
    {
	if(operand instanceof JUnaryExpression && 
	   ((JUnaryExpression)operand).oper() == OPE_MINUS)
	{
	    JUnaryExpression e = (JUnaryExpression) operand;
	    e.setExpr(wrap(e.expr()));
	} 
	else if(operand instanceof JCastExpression)
	{
	    JCastExpression e = (JCastExpression) operand;
	    e.setExpr(wrap(e.expr()));
	} 
	else if(operand instanceof JAddExpression
		|| operand instanceof JMinusExpression
		|| operand instanceof JMultExpression
		|| operand instanceof JDivideExpression)
	{
	    JBinaryExpression e = (JBinaryExpression) operand;
	    e.setLeft(wrap(e.left()));
	    e.setRight(wrap(e.right()));
	} else {
	    /* Report an error since operand is not one of: 
	     * cast, unary -, binary +, -, *, /. 
	     */
	    check(context, false, MjcMessages.SAFE_MATH_BAD_OP);
	}
	return;
    }

    private JExpression wrap(JExpression e) {
	return new MJWarnExpression(e.getTokenReference(),
				    "not"+operatorName(), // !FIXME!
				    e,
				    !enable_safe_math,
				    false // is _op
				    );
    }
    
    // ----------------------------------------------------------------------
    // CODE GENERATION
    // ----------------------------------------------------------------------

    /**
     * Accepts the specified visitor
     * @param	p		the visitor
     */
    public void accept(MjcVisitor p) {
	p.visitWarnExpression( this );
    }

    /**
     * Generates a sequence of bytecodes
     * @param	code		the code list
     */
    public void genCode(CodeSequence code) {
	throw new RuntimeException( "Method should not be called on a "
				    + "MJWarnExpression" );
    }

    // ----------------------------------------------------------------------

    private String operatorName;
    private JExpression operand;
    private boolean scope_is_one_op;
    private boolean enable_safe_math;
}
