/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.140 2008/05/14 18:16:26 wdietl Exp $
 */

package org.multijava.mjc;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.multijava.util.Destination;
import org.multijava.util.MessageDescription;
import org.multijava.util.Utils;
import org.multijava.util.classfile.AttributeList;
import org.multijava.util.classfile.ClassFileFormatException;
import org.multijava.util.classfile.ClassPath;
import org.multijava.util.classfile.MJAttributeParser;
import org.multijava.util.classfile.UniverseAttributeParser;
import org.multijava.util.compiler.*;
import org.multijava.util.FormattedException;

/**
 * This class implements the entry point of the MultiJava compiler.  Use
 * <code>java org.multijava.mjc.Main --help</code> to get usage options.
 */
public class Main extends org.multijava.util.compiler.Compiler {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    public Main() {
        this( new CModifier( Constants.ACCESS_FLAG_ARRAY,
                             Constants.ACCESS_FLAG_NAMES ) );
    }

    public Main( ModifierUtility modUtil ) {
        super( modUtil );
        appName = "mjc";
    }

    // ----------------------------------------------------------------------
    // RUN FROM COMMAND LINE
    // ----------------------------------------------------------------------

    /**
     * Entry point
     *
     * @param   args            the command line arguments
     */
    public static void main(String[] args) {
	try {
	    boolean success = compile(args);
	    //TaskTimes.report();
	    System.exit(success ? 0 : 1);
	} catch (Throwable e) {
	    bugReportRequest(e,args);
	    System.exit(1);
	}
    }

    /**
     * Second entry point
     */
    public static boolean compile(String[] args) {
        return new Main().run(args); 
    }

    /**
     * Runs a compilation session
     *
     * @param   args            the command line arguments
     */
    public boolean run(String[] args) {
        long startTime = System.currentTimeMillis();

        // Sets options and initializes compilation
        options = makeOptionsInstance();
        try {
            if (!runInitialization(args)) {
                errorFound = true;
                return false;
            }
        } catch (CompilationAbortedException e) {
            //  !FIXME! This doesn't seem right.  But test cases
            //  expected true result on -help or -version option, and
            //  Options throws an exception on these.  We have to
            //  handle the exception here so we don't continue to the
            //  compilation passes.
            return true;
        }
        

        return runCompilation(startTime);
    }
                
    // -------------------------------------------------------------
    // Generate a bug report on a Throwable that reaches the top level.
    public static void bugReportRequest(/*@\readonly@*/ Throwable e, String[] args) {
	java.io.PrintStream o = System.out;
	bugReportBoilerplate(true);
	o.println("---------------------------------------------------------");
	o.println("Arguments: ");
	for (int i = 0; i < args.length; i++) {
	    o.println("  " + args[i]);
	}
	bugReportProperty("java.version");
	bugReportProperty("java.vendor");
	bugReportProperty("java.home");
	bugReportProperty("java.io.tmpdir");
	bugReportProperty("os.name");
	bugReportProperty("os.arch");
	bugReportProperty("os.version");
	o.println("Exception cause: " + e.getCause());
	o.println("Exception message: " + e.getMessage());
	o.println("Exception stack trace: ");
	e.printStackTrace(o);
	o.println("---------------------------------------------------------");
	bugReportBoilerplate(false);
    }

    public static void bugReportBoilerplate(boolean isReportBelow) {
	java.io.PrintStream o = System.out;
	o.println("We're sorry, but there is a bug in the compiler that ");
	o.println("prevents it from successfully executing your command.");
	o.println("Please tell us about this bug at www.multijava.org.");
	o.println("Please include the information " +
		  (isReportBelow ? "below" : "above") + 
		  ", plus any additional");
	o.println("information that you think might be helpful for us to fix");
	o.println("this problem.  Thank you very much.");
    }

    public static void bugReportProperty(String key) {
	System.out.println(key + ": " + System.getProperty(key));
    }

    // ----------------------------------------------------------------------
    // GUI COMMAND LINE PARSER
    // ----------------------------------------------------------------------
        
    /**
     *  Runs the argument parser only so the GUI can process options from
     *  the command line
     */
    public boolean runParser(String[] args, MjcCommonOptions opt, 
                             ArrayList infiles) {

        if (!opt.parseCommandLine(args, infiles)) return false;

        return true;
    }
        
    // ----------------------------------------------------------------------
    // RUN FROM GUI
    // ----------------------------------------------------------------------

    /**
     * Entry point for the GUI
     *
     * @param   args            the command line arguments
     * @param   opt             the compiler options from the GUI
     * @param   os              the output stream for the compiler messages
     */
    public static boolean compile(String[] args, MjcCommonOptions opt, 
                                  OutputStream os) {
        return new Main().run(args, opt, os); 
    }
        
    /**
     * Runs a compilation session for the GUI
     *
     * @param   args      the non-options: files, packages, and directories
     * @param   opt       the compiler options
     * @param   os        the output stream to use for output messages - if null, no change is made
     */
    public boolean run(String[] args, MjcCommonOptions opt, OutputStream os) {
        final long startTime = System.currentTimeMillis();
         
        // Sets options and initializes compilation
        options = getOptionsInstance(opt);
        if (os != null) setOutputStream(os);
        if (!runInitialization(args)) {
            errorFound = true;
            return false;
        }
        
        return runCompilation(startTime);
    }
        
    /** Used as a hook for other subclasses of mjc to override.  In these
     *  subclasses, the opt variable must be assigned to the local instance
     *  of the specific option class for that tool (such as MjcOptions is
     *  specific to mjc).  Thus, methods that override this method must
     *  perform additional tasks before returning opt, because they must
     *  both assign opt to the local instance and to the protected 
     *  variable MjcCommonOptions options in this class.
     *
     *  @param    opt    The compiler options
     */
    protected MjcCommonOptions getOptionsInstance(MjcCommonOptions opt) {
        return opt;
    }   
             
    /**  
     * Interrupts an executing compilation process.  Typically used by
     * GUI to halt executing compilation process.
     */  
    public static final void interruptCompilation() {
        compilationInterrupted = true;
    }
    
    // ----------------------------------------------------------------------
    // METHODS FOR BOTH COMMAND LINE AND GUI
    // ----------------------------------------------------------------------
    
    /** Invoked from within the method run to do the argument parsing
        and compiler setup prior to actually parsing files via the
        task mechanism; returns true if no errors are encountered
        during initialization; if false is returned the initialization
        is not necessarily complete.
        @param  args   The array of command-line arguments
        @return true if no errors encountered during initialization
    */
    //@ requires args != null;
    protected boolean runInitialization(String[] args) {
        // Expand @files.
        String[] expandedArgs = expandAtFiles(args);

        // Parses the mjc command line options
        ArrayList infiles = new ArrayList();
        if (!parseArguments(expandedArgs,infiles)) {
            return false;
        }
                
        // Sets up the classpaths (at least), which we need to do before we can
        // process any package names that are on the command line.
        initialize();

        // Now process the items on the command line that are not options, namely
        // file, package, and directory names
        if (!handleNonOptions(infiles)) return false;

        /*@ assert (\forall Object o; infiles.contains(o); 
	  o instanceof String );
          @*/
        
        // Displays the usage message if no files were given to compile
        // Use filesFound since it is true if files/directories were given, even
        // if after stripping off non-existent files and empty directories there
        // is nothing left.
        if (!filesFound) {
            options.usage();
            reportTrouble(new FormattedException(MjcMessages.NO_INPUT_FILES));
            return false;
        }

        // establish reporting of debugging messages
        Debug.initialize( options.debug() );

        // initializes the compilation session
        Debug.setPrefix( "INIT" );

        if (verboseMode()) {
            inform( CompilerMessages.COMPILATION_STARTED, 
                    new Integer( infiles.size() ));
        }

        // Verifies that the destination is valid
        destination = Destination.check( options.destination() );

        // Initialize the "class loader" for this compilation session.
        // (see CTopLevel class docs for more detail)
        initSession();

        // Main parsing, typechecking, and printing routine
        runSetInitialTasks(infiles);

        infiles = null;
        return true;
    }

    /** Runs the compilation session for both the command line and GUI
        @param  startTime   The time the compilation started in the run method
        @return true if no errors encountered during compilation
    */
    protected boolean runCompilation(long startTime) {
        try {
            processTaskQueue();

            taskQueueEmptied();
            
            if( verboseMode() ) {
                inform( CompilerMessages.COMPILATION_ENDED );
            }

            // clear the stack maintained by the CodeSequence singleton
            CodeSequence.initSession();

            // Initialize the compilation session again.  This will be
            // repeated at the start of the next compilation session,
            // but the initialization is inexpensive and doing it here
            // frees up memory as soon as possible.
            initSession();

            if( options.debug() ) {
                Debug.close();
            }

            if (!quietMode()) {
                inform( MjcMessages.COMPILATION_TIME, 
                        new Long( System.currentTimeMillis() - startTime ) );
            } // end of if 

        } catch (CompilationAbortedException e) {
            if (!e.result()) errorFound = true;
        } catch (CompilationAbortedError e) {
            String s = e.getMessage();
            if (s != null && s.length()!=0) reportTrouble(new FormattedException(s));
            errorFound = true;
        }

        // If the compilation produced unchecked warnings but those warnings are
        // not enabled, we should nevertheless inform the user.
        if (uncheckedWarningsIssued
            && !RecommendedWarning("unchecked")) {
            inform(new FormattedException(MjcMessages.UNCHECKED_WARNINGS));
        }

        // Could use an option to set a error level limit
        // Note: A non-fatal error is categorized as a error in a .msg
        // file, but is reported as a warning so it does not abruptly
        // terminate compilation.  This check finds such errors.
        if (mostSevereWarningIssued <= MessageDescription.LVL_ERROR) 
            return false;
        return !errorFound;
    }

    /**
     * Expands a list of files containing the @files syntax into a
     * list of literal arguments.  See Sun's documentation for javac.
     */
    //@ requires infiles != null;
    //@ requires (\forall int i; 0<=i && i<infiles.length; infiles[i]!=null);
    //@ ensures infiles.equals(\old(infiles)) && infiles == \old(infiles);
    //@ ensures \result != null;
    //@ ensures (\forall int i; 0<=i && i<\result.length; infiles[i]!=null);
    protected String[] expandAtFiles(String[] infiles) {
        ArrayList v = new ArrayList();
        for (int i=0; i<infiles.length; i++) {
            String s = infiles[i];

            if (s.length() > 0 && s.charAt(0) == '@') {
                File f = new File(s.substring(1));
                if (f.isFile()) {
                    if (f.canRead()) {
                        parseAtFile(f, v);
                    } else {
                        // This behavior is more consistent with the
                        // behavior for other files specified on the
                        // command line but not found, although I
                        // think we should throw an error.
                        reportTrouble(new FormattedException(
				MjcMessages.ARG_NOT_READABLE_FILE,
                               Utils.relativePathTo(f)));
                    }
                } else {
                    reportTrouble(new CWarning(TokenReference.NO_REF,
			MjcMessages.ARG_NOT_EXIST, Utils.relativePathTo(f)));
                }
            } else {
                // The argument is not an @-file; just pass it through.
                v.add(s);
            }
        }
        return (String[])v.toArray(new String[]{});
    }

    /**
     * Does the actual tokenizing of the @-file, and puts each element
     * into the list v.
     *
     */
    //@ requires f != null;
    //@ requires v != null;
    //@ requires !v.containsNull && v.elementType <: \type(String);
    //@ ensures !v.containsNull && v.elementType <: \type(String);
    protected void parseAtFile(File f, ArrayList v) {
        try {
            FileReader r = null;
            try {
                r = new FileReader(f);
                StringBuffer buf = new StringBuffer();
                int next;
                while ((next = r.read()) != -1) {
                    char c = (char)next;
                    if (Character.isWhitespace(c)) {
                        if (buf.length() > 0) {
                            v.add(buf.toString());
                            buf.delete(0, buf.length());
                        }
                    } else {
                        buf.append(c);
                    }
                }
                if (buf.length() > 0) {
                    v.add(buf.toString());
                }
            } finally {
                if (r != null) {
                    r.close();
                }
            }
        } catch (IOException e) {
            throw new CompilationAbortedError(
                "Failure while parsing arguments from file: "
                + Utils.relativePathTo(f));
        }
    }

    /** Sets up the initial tasks that start the parsing.  
        Additional tasks may be generated if
        these files reference other classes that need parsing.
    */
    //@ requires infiles != null;
    //@ requires !infiles.containsNull && infiles.elementType <: \type(String);
    //@ ensures infiles.equals(\old(infiles)) && infiles == \old(infiles);
    protected void runSetInitialTasks(ArrayList infiles) {
        ParseTask first = firstTask( infiles );
        setMainSequenceID( first.sequenceID() );
        taskQueue.add( first );
    }

    /** 
     * Initializes this compilation session using the default "class
     * loader".
     */
    protected void initSession() {
        initSession( TypeLoader.getSingleton() );
        // Test that the classpath is at all reasonable, this test is
        // overridden in the JML tools because it works that way and
        // doesn't otherwise.
        if (CTopLevel.loadClass("java/lang/Object") == CClass.CLS_UNDEFINED) {
            throw new CompilationAbortedError(
                appName + " - Error: classpath+bootclasspath appears to be unusable (does not contain java.lang.Object): "
                + ClassPath.classPath());
        }
    }
   
    /** 
     * Initializes this compilation session
     * using the given type loader.
     */
    protected final void initSession( TypeLoader l ) {
        compilationInterrupted = false;
        CClass.initSession();
        currentlyParsingRelation.clear();
        failedParsingRelation.clear();
        alreadyParsedSet.clear();
        CTopLevel.initSession(this, l );
    }

    // The contextBehavior object allows some context specific behavior to be
    // defined without having to do a full-fledged derivation of new AST
    // classes, which gets complicated because one has to muck with the
    // parser as well. 

    static public class ContextBehavior {
        public boolean noBodyOK(CContextType c) { return false; }
    }

    static private ContextBehavior contextBehavior = new ContextBehavior();

    static public ContextBehavior contextBehavior() { return contextBehavior; }

    static protected void setContextBehavior(ContextBehavior cb) {
        contextBehavior = cb;
    }

    /**
     * This simply serves to pass some additional information out of
     * the call of parseArguments, in order to aid error reporting.
     */
    //@ spec_public
    protected  boolean filesFound = false; 

    /**
     * Parses the argument list.  Mutates infiles, options:
     * handles all options in the command-line, and puts all
     * non-option arguments into 'infiles'.
     */
    //@ requires args != null && infiles != null;
    //@ requires infiles.size() == 0;
    //@ modifies options;
    //@ modifies filesFound;
    public boolean parseArguments(String[] args, ArrayList infiles) {
        // the following mutates infiles
        if (!options.parseCommandLine(args, infiles)) return false;
        if (infiles.size() != 0) filesFound = true;
        return true;
    }

    /** 
     * This method does any processing of the non-option arguments on
     * the command-line, by mutating the given ArrayList (of String)
     * containing all those arguments; this implementation replaces
     * any directories by files within the directory that are
     * eligible for processing.
     * @return true if the arguments are well-formed
     */
    //@ requires infiles != null;
    //@ requires infiles.elementType <: \type(String) && !infiles.containsNull;
    //@ ensures infiles.elementType <: \type(String) && !infiles.containsNull;
    public boolean handleNonOptions(ArrayList infiles) {
        // Check for any directories or packages in the list of files, and replace
        // them by any files to be processed in them. (see method addFiles below)

        ArrayList dirs = new ArrayList();
        //@ set dirs.containsNull = false;
        //@ set dirs.elementType = \type(File);
        ArrayList packages = new ArrayList();
        //@ set packages.containsNull = false;
        //@ set packages.elementType = \type(String);
        Iterator i = infiles.iterator();
        while (i.hasNext()) {
            // iterating over command-line arguments that are not options
            String s = (String)(i.next());

            // Just in case mods to the arg list have left it in a bad state
            if (s == null || s.length() == 0) {
                i.remove();
                continue;
            }

            // First check if it is a file (whether or not it is a valid java file)
            File f = new File(s);
            if (f.isFile()) continue;

            // Also if the name ends in .java it is a file
            // (Primarily to avoid breaking all the tests that complain about 
            // non-existent files)
            if (s.endsWith(".java") &&
                s.indexOf('.') == s.lastIndexOf('.') &&
                validPackageName(s)) {
                // s looks like both (a) a package name and (b) a 
                // (non-existent) .java file naming a class.  We want to
                // produce a file missing error message, but only if it is
                // not a package.
                if (checkPackageName(s,true) == null) continue;
            }

            // Next check if it is a syntactically legitimate package name, 
            // whether or not the package exists.
            if (validPackageName(s)) {
                packages.add(s);
                i.remove();
                continue;
            }

            // Then check if it is a directory
            if (f.isDirectory()) {
                i.remove();
                dirs.add(f);
            } else {
                // We don't set the errorFound flag because we just
                // want to ignore non-existent files, not abort the
                // parsing of the others.
                // But turning on the warning bollixes up the regression tests.
                // FIXME - How?
                inform(MjcMessages.ARG_NOT_EXIST, Utils.relativePathTo(f));
                i.remove();
            }
        }

        // Do packages first since they may be converted to directories
        handlePackages(packages,dirs,infiles);

        // Now iterate over all directories found, to add any additional
        // files that should be processed
        handleDirectories(dirs,infiles);
        return true;
    }

    /** Returns true if the string is syntactically a valid package name. */
    //@ requires name != null;
    static public boolean validPackageName(String name) {
        int i = 0; 
        int n = name.length();
        while (i < n && Character.isJavaIdentifierStart(name.charAt(i))) {
            ++i;
            while (i < n && Character.isJavaIdentifierPart(name.charAt(i))) ++i;
            if (i==n) return true;
            if (name.charAt(i) != '.') return false;
            ++i;
        }       
        return false;   
    }

    /** This processes a list of directories found in the command-line, adding
        files to infiles as the overridable method sees fit; in this case any
        java files found in the directory are added.
        @param dirs a ArrayList of File objects
        @param infiles a ArrayList of String names of files
    */
    //@ requires dirs != null && infiles != null;
    //@ requires (\forall Object o; dirs.contains(o); o instanceof File);
    //@ requires (\forall Object o; infiles.contains(o); o instanceof String);
    //@ ensures (\forall Object o; dirs.contains(o); o instanceof File);
    //@ ensures (\forall Object o; infiles.contains(o); o instanceof String);
    public void handleDirectories(ArrayList dirs, ArrayList infiles) {
        Iterator i = dirs.iterator();
        while (i.hasNext()) {
            File f = (File)(i.next());
            addFiles(f,infiles);
        }
    }

    /** Processes all the package names found on the command-line, altering
        the list of directories or files as appropriate.  The implementation
        for mjc simply converts package names to the absolute path names
        of the directories in which they are found (per the sourcepath and
        classpath) and adds the directory to the list in dirs.
        @param packages the ArrayList of String containing a list of package names
        @param dirs a ArrayList of File containing a list of directory locations
        @param infiles a ArrayList of String containing a list of file paths
    */
    //@ requires packages != null && dirs != null && infiles != null;
    //@ requires packages.elementType <: \type(String) && !packages.containsNull;
    //@ requires dirs.elementType <: \type(File) && !dirs.containsNull;
    //@ requires infiles.elementType <: \type(String) && !infiles.containsNull;
    //@ ensures packages == \old(packages) && (* packages elements unchanged*);
    //@ ensures dirs.elementType <: \type(File) && !dirs.containsNull;
    //@ ensures infiles.elementType <: \type(String) && !infiles.containsNull;
    public void handlePackages(ArrayList packages, ArrayList dirs, ArrayList infiles) {
        Iterator i = packages.iterator();
        while (i.hasNext()) {
            String s = (String)(i.next());
            handlePackageName(s,dirs);
        }
    }

    /** Finds the package named by 'packageName' and adds it to the list
        of File objects in 'dirs'.
        @param packageName      the name of a package from the command-line
        @param dirs             a list of File objects that are directories to be processed
    */
    //@ requires packageName != null && dirs != null;
    //@ requires dirs.elementType <: \type(File) && !dirs.containsNull;
    //@ ensures dirs.elementType <: \type(File) && !dirs.containsNull;
    public File handlePackageName(String packageName, ArrayList dirs) {
        String location = checkPackageName(packageName,false);
        if (location == null) return null;

        String dirname = location + File.separator + 
            packageName.replace('.',File.separatorChar);
        File d = new File(dirname);
        dirs.add(d);
	return d;
        // Don't need to handle recursive packages, because that is handled 
        // in method addFiles
    }

    /** Checks whether a syntactically valid package name actually exists, 
        and returns the directory location of the root of the package
        or null if there is no package by that name on the sourcepath or
        classpath.
    */
    //@ requires packageName != null;
    public String checkPackageName(String packageName, boolean quiet) {
        String location = ClassPath.findPackageSource(packageName);
        if (location == null) location = ClassPath.findPackage(packageName);
        if (location == null) {
            if (!quiet) { 
		// FIXME - should this be a warning?
                reportTrouble(new FormattedException("Error: " + appName + 
                       " could not find file or package " + packageName));
            }
            return null;
        }

        if (! ((new File(location)).isDirectory())) {
            reportTrouble(new FormattedException("Not implemented by " + appName + 
                   ": processing of source files within a non-directory " + 
                   "(e.g. jar file): " + location));
            return null;
        }
        return location;

    }

    /** This class is used with the Directory.list method to list
        those files in a directory that this program is interested in
        processing - in this case, all those that end in '.java'
    */
    static public class Filter implements FilenameFilter {
        public Filter() {}

        //@ also
        //@ requires (* dir is a Directory, dir is non-null, name is non-null, name is an object in dir *);
        //@ ensures (* \result is true iff name ends with ".java" *);
        public boolean accept(File dir, String name) {
            return name.endsWith(".java");
        }
    }
    
    /** This class is used with the Directory.list method to list the
     *  subdirectories in a directory (ignoring the special case of
     *  the 'CVS' subdirectory).  */
    static public class DFilter implements FilenameFilter {
        public DFilter() {}

        //@ also
        //@ requires (* dir is a Directory, dir is non-null, name is non-null, name is an object in dir *);
        //@ ensures (* \result is true iff name is a subdirectory that is not named CVS *);
        public boolean accept(File dir, String name) {
            return (new File(dir,name)).isDirectory() && !name.equals("CVS");
        }
    }

    /** Overridable method that provides a filter to list all the relevant files in a directory;
        for mjc this is all files whose names end in ".java".
    */
    //@ ensures \result != null;
    protected FilenameFilter filenameFilter() {
        return new Filter();
    }

    /** Overridable method that provides a filter to list all the relevant subdirectories in a
        directory; for mjc this is all subdirectories that are not named 'CVS'.
    */
    //@ ensures \result != null;
    protected FilenameFilter subdirectoryFilter() {
        return new DFilter();
    }

    /** This is a helper function that (a) adds any relevant files in the
     *  directory to 'infiles', and (b) if we are searching directories
     *  recursively, calls addFiles on any subdirectories of 'directory'.
     *  "Relevant files" are those that are selected by the FilenameFilter
     *  object that is returned by the method filenameFilter().
     *  "Subdirectories" are those that are selected by the FilenameFilter
     *  object that is returned by the method subdirectoryFilter().  
     */
    //@ requires directory != null;
    //@ requires infiles.elementType <: \type(String) && !infiles.containsNull;
    //@ ensures infiles.elementType <: \type(String) && !infiles.containsNull;
    private void addFiles(File directory, ArrayList infiles) {
        // First get everything that might be a .java file.
        String[] javafiles = directory.list(filenameFilter());
        for (int i=0; i<javafiles.length; i++) {
            // System.out.println("ADDING FILE " + javafiles[i]);
            File f = new File(directory,javafiles[i]);
            // The file would not be returned by directory.list if it did not exist,
            // but check to make sure that it is a simple file.
            if (f.isFile()) {
                String path = f.getAbsolutePath();
                infiles.add(path);
            }
        }

        // Then, if we are searching files recursively, check for any
        // directories and add files from them.

        if (options.recursive()) {
            String[] dirs = directory.list(subdirectoryFilter());
            for (int i=0; i<dirs.length; i++) {
                // System.out.println("ADDING DIR " + directory + File.separator + dirs[i]);
                addFiles(new File(directory,dirs[i]),infiles);
            }
        }

    }

    /** Overridable method that returns an instance of an object 
        that will process command-line options.
    */
    //@ ensures \result != null;
    protected MjcCommonOptions makeOptionsInstance() {
        return new MjcOptions();
    }

    /**
     * Creates a compilation unit context for this compiler.  
     *
     * <pre><jml>
     * requires cunit != null;
     * ensures \result != null && contextsCreated.contains(\result);
     * </jml></pre>
     *
     */
    public CCompilationUnitContextType 
        createCompilationUnitContext(JCompilationUnitType jc, 
                                     CCompilationUnit cunit ) 
    {
        CCompilationUnitContextType result = 
            new CCompilationUnitContext( this, cunit );
        contextsCreated.add(result);
        return result;
    }
                                                

    /**
     * Signals to this compiler that the given context is no longer
     * needed.  
     *
     * <pre><jml>
     * requires childContext != null && contextsCreated.contains(childContext);
     * ensures !contextsCreated.contains(childContext);
     * </jml></pre>
     *
     */
    public void adoptCompilationUnitContext( CCompilationUnitContextType 
                                             childContext ) {
        assertTrue( contextsCreated.remove( childContext ) );
    }

    /**
     * Generates the first task in the compilation sequence.  This is a 
     * method so subclasses can modify the task sequence.
     *
     * <pre><jml>
     * requires infiles != null && (\forall Object o; infiles.contains(o);
     *                                  o instanceof String);
     * </jml></pre>
     *
     */
    public ParseTask firstTask( ArrayList infiles ) {
        return new ParseTask( infiles );
    }

    /**
     * Generates the first task in the compilation sequence.  This is a 
     * method so subclasses can modify the task sequence.
     *
     * <pre><jml>
     * requires filename != null && expected != null;
     * </jml></pre>
     *
     */
    public ParseTask firstTask( File filename, ExpectedResult expected ) {
        return new ParseTask( filename, expected );
    }

    /**
     * Reentrant procedure processes tasks from the task queue until
     * all tasks have been completed whose priority is higher than the
     * current priority when this was reentered.  The method returns true
     * if the compilation had to be aborted prematurely because of errors,
     * or false (with either errorFound true or false) if compilation ran
     * to completion, possibly finding errors along the way.  */
    protected final void processTaskQueue() {
        boolean aborted = false;
        int currentPriority = activeTaskPriority >= 0 ?
            activeTaskPriority : Integer.MAX_VALUE;
        int stopAtPriority = currentPriority;
        if (verboseMode()) {
            inform(CompilerMessages.MISC_STRING,
                   "[ Entering processTaskQueue with stopAtPriority = " 
                   + stopAtPriority + ", taskQueue: " + taskQueue + " ]");
        }
        
        processTasks:
        while (!taskQueue.isEmpty()) {
            Task first = (Task) taskQueue.first();
            Object currentActiveSeqID = activeSequenceID();

            int firstPriority = first.priority();

            if (firstPriority > stopAtPriority) {
                // we've gone far enough in this reentrant invocation
                if (verboseMode()) {
                    inform(CompilerMessages.MISC_STRING,
                           "[ catch up completed, stop priority is " + stopAtPriority + 
                           ", current priority of " + first + " is " + firstPriority + " ]");
                }
                break processTasks;
            }

            taskQueue.remove( first );

            if (firstPriority > errorLimit && !options.keepGoing()) {
                // We have had a previous error in a task with a lower priority,
                // so we abort this task.

                if (verboseMode()) {
                    inform(CompilerMessages.MISC_STRING,
                           "[ ABORTING TASK WITH PRIORITY " + firstPriority + 
                           ", error priority is " + errorLimit + " ]");
                }
                aborted = true;
                continue;
            }

            // We use currentErrorFound to enable us to detect the occurrence of
            // errors in this particular execution of a task (or any tasks that 
            // it spins off).  If there are any errors, we do not continue
            // with further tasks in the sequence.
            boolean accumulatedErrorFound = errorFound;
            errorFound = false;

            // Only code under the call graph of statements in the
            // try-catch-finally following the assignment to
            // activeTaskPriority can safely make reentrant calls to
            // this method.  We can't use Java synchronization
            // mechanisms to enforce this since those don't restrict
            // reentrant calls from a single thread.
            activeTaskPriority = firstPriority;
            try {
                if (verboseMode()) {
                    inform(CompilerMessages.MISC_STRING,
                           "[ processing task " + first + " with priority " + 
                           firstPriority + " ]");
                }
                activeSequenceID = first.sequenceID();
                long startTime = System.currentTimeMillis();
                first.execute(); // may result in reentrant invocation
                TaskTimes.add(first.getClass().toString(),System.currentTimeMillis()-startTime);
            } catch (CompilationAbortedException e) {
                noteError();
                aborted = true;
            } finally {
                // Need to restore the activeSequenceID rather than making 
                // it null!!
                activeSequenceID = currentActiveSeqID;
                activeTaskPriority = currentPriority;
            }

            Task next=null;
            if (!errorFound     // next task only if first is error free...
                && (!(first instanceof Trees) || // ...and not empty...
                    ((Trees)first).trees().length != 0) 
                && !(compilationInterrupted && // ...and user has not
                     currentPriority == Integer.MAX_VALUE) // ...pressed stop
                ) { 
                next = createTaskAfter( first );
            } else {
                try {
                    if (verboseMode()) {
                        inform(CompilerMessages.MISC_STRING,
                               "[ aborting compilation ]");
                    }
                } catch (CompilationAbortedException e) {
                    noteError();
                    aborted = true;
                    compilationInterrupted = false;
                }
            }
                
            if (verboseMode()) {
                inform(CompilerMessages.MISC_STRING,"[ done processing task " + first + 
                       ", resulting in task " + next + " ]");
            }
            aborted = aborted || errorFound;
            errorFound = accumulatedErrorFound || errorFound;

            if (next!=null) {
                if (verboseMode()) {
                    inform(CompilerMessages.MISC_STRING,
                           "[ adding task " + next + " to taskQueue ]");
                }
                taskQueue.add( next );
            }
        }
        
        if (verboseMode()) {
            inform(CompilerMessages.MISC_STRING,
                   "[ End of processTaskQueue loop: taskQueue.isEmpty() = "
                   + taskQueue.isEmpty() + ", errorFound = " + errorFound 
                   + ", errorLimit = " + errorLimit
                   + ", aborted = " + aborted 
                   + ", options.keepGoing = " + options.keepGoing() + " ]");
        }
        if (aborted) throw new CompilationAbortedException();
        return ;
    }


    /**
     * This method is called by {@link #run(String[])} when all
     * reentrant invocations of {@link #processTaskQueue()} have been
     * exited and in this implementation the method generates any
     * required bytecode.  This method should be overridden by
     * subclasses to perform final work after completion of all tasks.  */
    protected void taskQueueEmptied() {
        // Done with parsing, typechecking, and printing.  Time to
        // generate code.
        if (codeGenNeeded) {
            // write the bytecode
            Debug.setPrefix( "GENCODE" );
            CBinaryMethod.swallowReceivers();
            genCode();
        }
    }

    /**
     * Compiles the named file (using the options of the current
     * compilation session) up to the same compilation pass as the
     * currently active task in the task queue.  Adds the result
     * to the set of compilation units to be processed to completion.
     *
     * @param filename the file descriptor of the file to be compiled
     *
     * <pre><jml>
     * also
     * requires filename != null && (* filename.canRead() *); // !FIXME!
     * </jml></pre>
     * */
    public void catchUp( File filename ) {
        taskQueue.add( firstTask(filename, ExpectedIndifferent.instance()) );
        // reentrant invocation of processTaskQueue
        processTaskQueue();
    }    

    /** 
     * This version of catchUp also parses the given file through the
     * TaskQueue passes as far as other files have been processed,
     * but discards the result if the file did not contain a declaration 
     * of the type qualifiedName.
     *
     * <pre><jml>
     * requires qualifiedName != null;
     * requires filename != null;
     * </jml></pre>
     */
    public void catchUpType( File filename, String qualifiedName ) {

        taskQueue.add(firstTask(filename, ExpectedType.instance(qualifiedName)));
        processTaskQueue();
    }

    /** 
     * This version of catchUp also parses the given file through the
     * TaskQueue passes as far as other files have been processed,
     * but: after the ParseTask phase, checks that the given
     * qualifiedName actually has a successful generic function
     * definition that is now loaded.
     *
     * <pre><jml>
     requires qualifiedName != null;
     requires filename != null;
     * </jml></pre>
     */
    public void catchUpGF( File filename, String qualifiedName ) {
        taskQueue.add(firstTask(filename,ExpectedGF.instance(qualifiedName)));
        processTaskQueue();
    }

    /**
     * <P>Compiles the given AST (using the options of the current
     * compilation session) up to the same compilation pass as the
     * currently active task in the task queue.  Adds the result
     * to the set of compilation units to be processed to completion.</P>
     *
     * <P>Since in this statically overloaded method the argument is
     * an AST for a type declaration some of the initial passes
     * are skipped.</P>
     *
     * <P>This entry point is accessed for compiling type declaration
     * statements and anonymous class instantiation expressions.</P>
     *
     * @param decl the AST for the type declaration to be compiled
     *
     * <pre><jml>
     * requires decl != null;
     * </jml></pre>
     * */
    public void catchUp( JTypeDeclarationType decl ) {
        taskQueue.add( firstCheckingTask(decl) );
        processTaskQueue();
    }

    /**
     * Generates the first task in the compilation sequence for an
     * already formed AST.  This is a method so subclasses can modify
     * the task sequence.
     *
     * <pre><jml>
     * requires decl != null && decl.arePassParametersCached;
     * </jml></pre>
     * */
    public Task firstCheckingTask( JTypeDeclarationType decl ) {
        Object sid = 
            (activeSequenceID == null) ? new Object() : activeSequenceID;
        return new PreprocessTask( new CompilerPassEnterable[] {decl}, sid );
    }

    /**
     * Generates the code for all the classes.
     */
    public void genCode() {
        CSourceClass[]  classes = getClasses();
        this.classes.clear();

        try {
            for (int count = 0; count < classes.length; count++) {
                long            lastTime = System.currentTimeMillis();

                if (verboseMode()) {
                    inform( MjcMessages.STARTING,
                            classes[count].qualifiedName().replace('/', '.') );
                }

                classes[count].genCode( destination );
        
                if (verboseMode() && !classes[count].isNestedType()) {
                    inform(CompilerMessages.CLASSFILE_GENERATED, 
                           classes[count]
                           .qualifiedName().replace('/', '.'), 
                           new Long(System.currentTimeMillis() 
                                    - lastTime));
                }
                classes[count] = null;
            }
        } catch (ClassFileFormatException e) {
            reportTrouble(e);
            e.printStackTrace();
        } catch (IOException e) {
            reportTrouble(e);
        }
    }

    /**
     * Initialize the compiler (read classpath, initialize type descriptors)
     */
    protected void initialize() {
        // get the classpath, from the mjc command line option, the user's
        // CLASSPATH or java command line option, or by using the current
        // directory
        if (options.classpath() == null) {
            options.set_classpath(System.getProperty("java.class.path"));
            if (options.classpath() == null) {
                options.set_classpath(".");
            }
        }
        // ClassPath uses the Singleton pattern to provide utilities for
        // finding files based on fully-qualified class names
        org.multijava.util.classfile.ClassPath.init(options.classpath(),
                                                    options.sourcepath());

        AttributeList.addParser(new MJAttributeParser());
	
	// AHS: add the Universe attribute parser to the chain of responsiblity
        // This will allow us to recognize the universe attributes
        AttributeList.addParser(new UniverseAttributeParser());
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * @return true if compilation in verbose mode
     */
    public boolean verboseMode() {
        return options.verbose();
    }

    /**
     * @return the optimization level
     */
    public int optimizeCode() {
        return (options instanceof MjcOptions) ? 
	    ((MjcOptions)options).optimize() : 0;
    }


    /**
     * Returns true if quiet mode is enabled.
     */
    public boolean quietMode() {
        return options.quiet();
    }


    /**
     * @return true if supporting Relaxed MultiJava extensions
     */
    public boolean RMJ() {
        return options.relaxed();
    }

    public boolean Generic() {
	return options != null && options.generic();
    }

    // TODO[OJM]: Remove as soon as Java generics are supported!!!
    public boolean Generic(CClass host) {
        return Generic()
               && ((host == null) || (host.qualifiedName() == null)
                   || !(host.qualifiedName().startsWith("java/")
                        || host.qualifiedName().startsWith("javax/")
                        || host.qualifiedName().startsWith("sun/")
                        || host.qualifiedName().startsWith("sunw/")));
    }

    /**
     * Returns whether the recommended warnings type identified by
     * <code>warning</code> is enabled in the compiler.
     *
     * @param warning  The recommended warning type to check for.
     * @return         Whether the given warning type is enabled.
     */
    public boolean RecommendedWarning(String warning) {
        String[] warnings = options.Xlint().split(",");
        for (int i = 0; i < warnings.length; i++) {
            if (warnings[i].equals(warning)) {
                return true;
            }
        }
        return false;
    }

    /**
     * @return true iff "safe math", i.e. integral arithmetic overflow, is enabled.
     */
    public boolean safeMath() {
	return options != null && options.safemath();
    }


    /**
     * @return true iff the non-null type system is to be used.
     */
    public boolean nonNullTypes() {
	return options != null && options.nonnulltypes();
    }
    public boolean experimentalArrayHandling() {
	return options != null && options.xArrayNNTS();
    }



    /** Returns whether the checking of universes should be enabled.
     * By WMD.
     *
     * @return true iff "universes", i.e. universe type modifiers, are enabled.
     *
     * @see #allowUniverseChecks
     */
    public boolean universeChecks() {
	// return options != null && options.universes();
	return allowUniverseChecks;
    }
    
    /**
     * Used to temporarily disable universe checks, to not
     * cause any confusion with dynamic universe type checks.
     * (The dyn.checks call methods taking peer arguments with
     * rep variables. Universe typechecks don't allow this normally)
     * 
     * @param check enable/disable universe checks.
     * 
     * By scdaniel.
     */
    public void setUniverseChecks (boolean check) {
    	allowUniverseChecks = check;
    }

    /** Returns whether the new keywords for the universe type system
     * should be enabled.
     * By WMD.
     *
     * @see #allowUniverseKeywords
     */
    public boolean universeKeywords() {
	return allowUniverseKeywords;
    }

    /** Returns whether the (conservative) purity checking of methods
     * should be enabled.
     * By WMD.
     *
     * @see #allowUniversePurity
     */
    public boolean universePurity() {
	return allowUniversePurity;
    }

    /**
     * Used to temporarily disable purity checks, to not
     * cause any confusion with dynamic universe type checks.
     * (The dyn.checks call non-pure methods, the purity checks
     * would not allow this)
     * 
     * @param check enable/disable purity checks.
     */
    public void setUniversePurity (boolean check) {
    	allowUniversePurity = check;
    }

    /** Returns whether the dynamic checks for the universe type system
     * should be enabled.
     * By WMD.
     *
     * @see #allowUniverseBytecode
     */
    public boolean universeDynChecks() {
	return allowUniverseDynChecks;
    }

    /** Returns whether the extended bytecode for the universe type
     * system should be enabled.
     * By WMD.
     *
     * @see #allowUniverseBytecode
     */
    public boolean universeBytecode() {
	return allowUniverseBytecode;
    }
    
    /** Returns whether to use Java 5 annotations to represent universe
     * modifiers in generated classfiles.
     * By AHS.
     * 
     * @see #unverseBytecodeAnnotations
     */
    public boolean universeBytecodeAnnotations() {
        return allowUniverseAnnotations;
    }
    
    /** Returns the version of the encoding used to the encode the
     * universe type modifiers.
     * By AHS.
     * 
     * @see #universeVersion
     */ 
    public String universeVersion() {
        return universeVersion;
    }


    /** From the given String parameter determine what universe checks
     * should be enabled.
     *
     * By WMD.
     */
    protected void setAllowUniverses( String option ) {

	option = option.toLowerCase();

	if( option.equals(UNIVERSE_FULL) ) {

	    // enable everything
	    allowUniverseKeywords = allowUniverseChecks = true;
        allowUniversePurity = true;    
        allowUniverseDynChecks = true;

        // use attributes as default bytecode representation
        allowUniverseBytecode = true;
	    allowUniverseAnnotations = false;

	} else if( option.equals(UNIVERSE_NO) ) {

	    // disable everything
	    allowUniverseKeywords = allowUniverseChecks =
		allowUniversePurity =
		allowUniverseDynChecks = allowUniverseBytecode =
                allowUniverseAnnotations = false;

	} else {

	    // set everything to false first and then enable them
	    // one by one.
	    allowUniverseKeywords = allowUniverseChecks =
		allowUniversePurity =
		allowUniverseDynChecks = allowUniverseBytecode =
                allowUniverseAnnotations = false;
            
	    String[] params = option.split(",");
	    for (int indx=0; indx < params.length; ++indx) {
		if ( params[indx].equals(UNIVERSE_PARSE) ) {
		    allowUniverseKeywords = true;
		} else if ( params[indx].equals(UNIVERSE_CHECKS) ) {
		    allowUniverseChecks = true;
		} else if ( params[indx].equals(UNIVERSE_PURITY) ) {
		    allowUniversePurity = true;
		} else if ( params[indx].equals(UNIVERSE_DYNCHECKS) ) {
		    allowUniverseChecks = true;
		    allowUniverseDynChecks = true;
		} else if ( params[indx].equals(UNIVERSE_BYTECODE) ) {
		    allowUniverseChecks = true;
		    allowUniverseBytecode = true;		
		} else if ( params[indx].equals(UNIVERSE_ANNOTATIONS) ) {
		    allowUniverseChecks = true;
		    allowUniverseAnnotations = true;
		} else {
            inform(CompilerMessages.MISC_STRING,
                "Unknown Universe command-line option \"" + params[indx] + "\" ignored!");
		}
	    }
	}

	CStdType.initSession();
    }


    // --------------------------------------------------------------------
    // COMPILER
    // --------------------------------------------------------------------

    /**
     * Reports a trouble (error or warning).  All error and warning output
     * should go through this method so that subclasses can handle such
     * output as they see fit (it is best if all output other than the
     * progress output that is shut off by -quiet go through here - no direct
     * uses of System.out).
     *
     * @param   trouble         a description of the trouble to report.
     */
    public void reportTrouble(/*@\readonly@*/ Exception trouble ) {
        if( trouble instanceof CWarning ) {
            CWarning warning = (CWarning) trouble;
            // Handle unchecked warnings.
            // FIXME[OJM]: Handle recommended warnings more generally!
            if (warning.getSeverityLevel() == MessageDescription.LVL_WARNING_UNCHECKED) {
                uncheckedWarningsIssued = true;

                if (!RecommendedWarning("unchecked")) {
                    return;
                }
            } else if (options.warning() > 0 && !suppressWarning((CWarning)trouble)) {
                int i = ((CWarning)trouble).getSeverityLevel();
                if (i < mostSevereWarningIssued && 
                    i != MessageDescription.LVL_UNDEFINED) 
                    mostSevereWarningIssued = i;
            } else {
		// Warning filtered out
		return;
	    }
	} else {
	    noteError();
	}
	reportTroubleFiltered(trouble);
    }

    /** This method is intended for overriding by subclasses - all error and
        warning output (except that filtered out by the -w or -filter option)
        should pass through here.  The implementation here sends it to System.out
        or System.err, but subclasses can do whatever - write to a file, put
        up a dialog, etc.

        @param trouble An exception indicating the error or warning.  This 
		will often be a PositionedError or FormattedException
		containing information about a source location, but might
		also be an unexpected exception that is thrown somewhere.
     */
    public void reportTroubleFiltered(/*@\readonly@*/ Exception trouble) {
	inform(trouble);
    }

    /**
     * Returns true if the given warning should be suppressed.  */
    protected boolean suppressWarning(CWarning warning) {
        WarningFilter filter = getFilter();
        int value = filter.filter(warning);

        switch (value) {
        case WarningFilter.FLT_REJECT:
            return true;
        case WarningFilter.FLT_FORCE:
            return false;
        case WarningFilter.FLT_ACCEPT:
            // getSeverityLevel returns a larger number for less
            // severe warnings (go figure)
            return warning.getSeverityLevel() > options.warning();
        default:
            throw new RuntimeException("INTERNAL ERROR");
        }
    }

    /**
     * Return the warning filter. */
    protected WarningFilter getFilter() {
        if (filter == null) {
            String filterClassName = getWarningFilterNameFromOptions(options);
            if (filterClassName != null) {
                try {
                    Object f = Class.forName(filterClassName).newInstance();
                    filter = (WarningFilter) f;
                } catch (Exception e) {
                    reportTrouble(new FormattedException(MjcMessages.FILTER_NOT_FOUND, filterClassName));
                }
            }
            if (filter == null) {
                filter = getDefaultFilter();
            }
        }

        return filter;
    }

    /**
     * Get the warning filter's class name from the options structure.
     * This is a hook method that can be overridden by subclasses.
     * @return null if there is no warning filter option specified.
     */
    protected String getWarningFilterNameFromOptions(MjcCommonOptions opts)
    {
        return ((MjcOptions)opts).filter();
    }

    /**
     * Return an instance of the default warning filter.  This is a
     * hook method that can be overridden by subclasses. */
    protected /*@ non_null @*/ WarningFilter getDefaultFilter() {
        return new DefaultFilter();
    }

    /**
     * Returns true iff comments should be parsed (false if to be skipped)
     */
    public boolean parseComments() {
        return options.deprecation() || !(options instanceof MjcOptions);
    }

    /**
     * Returns the classes to generate
     */
    public CSourceClass[] getClasses() {
        return (CSourceClass[]) classes.toArray( new CSourceClass[0] );
    }


    /**
     * Adds a class to the list of classes to be generated in bytecode.
     *
     * <pre><jml>
     * requires clazz != null;
     * assignable classes;
     * ensures classes.contains(clazz);
     * </jml></pre>
     *
     */
    public void classToGenerate( CSourceClass clazz )  {
        classes.add( clazz );
    }

    // -------------------------------------------------------------
    // FILE AND AST CACHING
    // -------------------------------------------------------------

    /**
     * Registers with the current session that the given file is being
     * parsed in hopes of finding the given expected result. 
     *
     * <pre><jml>
     * requires f != null && expected != null;
     * </jml></pre>
     *
     */
    protected final void currentlyParsingFor(File f, ExpectedResult expected) {
        Set bindings = currentlyParsingSetFor(f);
        bindings.add(expected);
    }

    /**
     * Returns true iff the given file is currently being parsed in
     * hopes of finding the given expected result.
     *
     * <pre><jml>
     * requires f != null && expected != null;
     * </jml></pre>
     *
     */
    protected final boolean isCurrentlyParsingFor(File f, ExpectedResult expected) {
        Set bindings = currentlyParsingSetFor(f);
        if (bindings.isEmpty()) {
            return false;
        } else if (expected.isIndifferent()) {
            return true;
        } else {
            return bindings.contains(expected);
        }
    }

    /**
     * Returns true iff the given compilation unit contains the
     * results that were expected from parsing the given file.
     *
     * <pre><jml>
     * requires f != null;
     * requires (* currentlyParsingRelation.containsKey(getFilePath(f)) *);
     * </jml></pre>
     *
     */
    protected final boolean isAnExpectedResult(File f, JCompilationUnitType cu) {
        if (cu == null) {
            return false;
        }
        Set bindings = currentlyParsingSetFor(f);
        Iterator iter = bindings.iterator();
        while (iter.hasNext()) {
            ExpectedResult expected = (ExpectedResult) iter.next();
            if (expected.foundIn(cu)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns the set of expected results for processing the given
     * file.  The result is an alias to the set stored in
     * <CODE>currentlyParsingRelation</CODE> and can be mutated to
     * change the set of expected results mapped to be the given file.
     *
     * <pre><jml>
     * requires f != null;
     * </jml></pre>
     *
     */
    private final Set currentlyParsingSetFor(File f) {
        String key = getFilePath(f);
        Set bindings = (Set) currentlyParsingRelation.get(key);
        if (bindings == null) {
            bindings = new HashSet();
            currentlyParsingRelation.put(key,bindings);
        }
        return bindings;
    }

    /**
     * Returns true if we have already tried and failed to parse the
     * given file for the given expected result.
     */
    protected final boolean hasAlreadyFailedToParseFor(File f, ExpectedResult expected)
    {
        String key = getFilePath(f);
        Set bindings = (Set) failedParsingRelation.get(key);
        return bindings != null && bindings.contains(expected);
    }

    /**
     * Called when a file is parsed and found to not contain the
     * expected declarations.  This method removes the given file from
     * the set of those currently being parsed.  The method also
     * removes from the global cache all reference to the given
     * compilation unit.
     */
    protected final void failedParsing(File f, JCompilationUnitType cu) {
        CTopLevel.forgetEverythingAbout(f, cu);

        String key = getFilePath(f);
        Set bindings = (Set) currentlyParsingRelation.get(key);
        currentlyParsingRelation.remove(key);
        Set failedBindings = (Set) failedParsingRelation.get(key);
        if (failedBindings == null) {
            failedBindings = new HashSet();
            failedParsingRelation.put(key,bindings);
        }
        failedBindings.addAll(bindings);
    }

    /**
     * Called when a file is parsed and found to contain the expected
     * declarations.  This method removes the given file from the set
     * of those currently being parsed and adds it to the set of
     * already parsed files.
     */
    protected final void successfullyParsed(File f) {
        String key = getFilePath(f);
        currentlyParsingRelation.remove(key);
        failedParsingRelation.remove(key);
        alreadyParsedSet.add(key);
    }

    /**
     * Returns true iff the given file has already been parsed and the
     * results were not discarded.
     *
     * <pre><jml>
     * requires f != null;
     * </jml></pre>
     *
     */
    protected final boolean hasAlreadySuccessfullyParsed(File f) {
        String key = getFilePath(f);
        return alreadyParsedSet.contains(key);
    }

    // -------------------------------------------------------------
    // TASK AND TASK SEQUENCING CODE
    // -------------------------------------------------------------

    /**
     * This method uses the dynamic type of <code>oldTask</code> along
     * with the command line options to determine what task to add to
     * the task queue after the given task completes.  
     *
     * <pre><jml>
     * requires oldTask != null && oldTask.completed;
     * ensures \result == null || 
     *         (\result.sequenceID() == oldTask.sequenceID());
     * </jml></pre>
     *
     */
    protected Task createTaskAfter( Task oldTask ) { // !CONVERT! multimethods
        if (oldTask instanceof ParseTask) {
            if (!(options instanceof MjcOptions) 
		|| !((MjcOptions)options).notc() ) {
                return new PreprocessTask( ((ParseTask)oldTask).trees(),
					   oldTask.sequenceID() );
            }
            return null;
        } else if (oldTask instanceof PreprocessTask) {
            return new CheckInterfaceTask(((PreprocessTask) oldTask).trees(),
                                          oldTask.sequenceID() );
        } else if (oldTask instanceof CheckInterfaceTask) {
            return 
                new CheckInitializerTask( ((CheckInterfaceTask) oldTask).
                                          trees(),
                                          oldTask.sequenceID()  );
        } else if (oldTask instanceof CheckInitializerTask) {
            return 
                new ResolveSpecializerTask(((CheckInitializerTask) oldTask)
                                           .trees(),
                                           oldTask.sequenceID() );
        } else if (oldTask instanceof ResolveSpecializerTask) {
            return 
                new ResolveTopMethodTask(((ResolveSpecializerTask) oldTask)
                                         .trees(),
                                         oldTask.sequenceID() );
        } else if (oldTask instanceof ResolveTopMethodTask) {
            return 
                new TypecheckTask(((ResolveTopMethodTask) oldTask).trees(),
                                  oldTask.sequenceID() );
        } else if (oldTask instanceof TypecheckTask) {
            CompilerPassEnterable[] trees = ((TreeProcessingTask) oldTask).trees();
            if (options instanceof MjcOptions && ((MjcOptions)options).nowrite()) {
                return null;
            }
            
            if (!(options instanceof MjcOptions)) {
                return null;
            } else if (!((MjcOptions)options).notc()) {
                return new TranslateMJTask( trees, oldTask.sequenceID() );
            }
            return null;
        } else if (oldTask instanceof PrettyPrintTask) {
            return null;
        } else if (oldTask instanceof TranslateMJTask) {
            return null;
        }
        
        throw new IllegalArgumentException( "bug encountered: task following" +
                                            " a task of type " + 
                                            oldTask.getClass() + 
                                            " is not specified." );
    }


    /**
     * This class represents the expected result of parsing a file.
     */
    protected abstract static class ExpectedResult {
        protected ExpectedResult(String qualifiedName) {
            this.qualifiedName = qualifiedName;
        }

        /**
         * Returns true if this expected result is contained in cu.
         */
        public abstract boolean foundIn(JCompilationUnitType cu);

        public boolean isIndifferent() {
            return false;
        }

        public final String qualifiedName() {
            return qualifiedName;
        }

        public final int hashCode() {
            return qualifiedName.hashCode();
        }

        public abstract boolean equals(/*@ nullable @*/ Object other); //  !CONVERT!

        public /*@ pure @*/ boolean equals(/*@ nullable @*/ ExpectedResult other) {  // !CONVERT!
            return qualifiedName.equals(other.qualifiedName);
        }

        public String toString() {
            return qualifiedName;
        }

        private final String qualifiedName;
    }

    protected static class ExpectedType extends ExpectedResult {
        /**
         * Clients, including Main, should always use this factory
         * method to create instances and should not call the
         * constructor.  */
        public static ExpectedType instance(String qualifiedName) {
            return new ExpectedType(qualifiedName);
        }

        private ExpectedType(String qualifiedName) {
            super(qualifiedName);
        }

        /**
         * Returns true if this expected result is contained in cu.
         */
        public boolean foundIn(JCompilationUnitType cu) {
            return cu.declaresType(qualifiedName());
        }

        public boolean equals(/*@ nullable @*/ Object o) {
            if (o instanceof ExpectedType) {
                return super.equals( (ExpectedType) o );
            } else {
                return false;
            }
        }

        public String toString() {
            return "Type " + super.toString();
        }
    }

    protected static class ExpectedGF extends ExpectedResult {
        /**
         * Clients, including Main, should always use this factory
         * method to create instances and should not call the
         * constructor.  */
        public static ExpectedGF instance(String qualifiedName) {
            return new ExpectedGF(qualifiedName);
        }

        private ExpectedGF(String qualifiedName) {
            super(qualifiedName);
        }

        /**
         * Returns true if this expected result is contained in cu.
         */
        public boolean foundIn(JCompilationUnitType cu) {
            return cu.declaresGF(qualifiedName());
        }

        public boolean equals(/*@ nullable @*/ Object o) {
            if (o instanceof ExpectedGF) {
                return super.equals( (ExpectedGF) o );
            } else {
                return false;
            }
        }

        public String toString() {
            return "GF " + super.toString();
        }
    }

    protected static class ExpectedIndifferent extends ExpectedResult {
        /**
         * Clients, including Main, should always use this factory
         * method to create instances and should not call the
         * constructor.  */
        public static ExpectedIndifferent instance() {
            return SINGLETON;
        }

        private ExpectedIndifferent() {
            super("do not care");
        }

        public final boolean isIndifferent() {
            return true;
        }

        /**
         * Returns true if this expected result is contained in cu.
         */
        public boolean foundIn(JCompilationUnitType cu) {
            return true;
        }

        public boolean equals(/*@ nullable @*/ Object o) {
            if (o instanceof ExpectedIndifferent) {
                return true;    // singleton
            } else {
                return false;
            }
        }

        public String toString() {
            return "Indifferent";
        }

        private static final ExpectedIndifferent SINGLETON = 
            new ExpectedIndifferent();
    }

    /**
     * This abstract class represents a single task to be performed by
     * the compiler.  A task is a single pass of the compiler,
     * performed on one or (for efficiency) several compilation units.
     * Each concrete task has a priority which is used for ordering
     * tasks in a priority queue.  This technique allows the
     * compilation of transitively referenced files to be "caught up"
     * with the compilation of files named at the command line.  */
    public abstract class Task implements Comparable {
        /**
         * Creates a new task of the given priority and sequenceID.
         * The priority is used to schedule tasks in the taskQueue.
         * We use the term <em>task sequence</em> to refer to a
         * sequence of tasks from parsing through to the end of
         * compilation such that every task in the sequence except the
         * first was returned by a call to createTaskAfter(t) where t
         * is the previous task in the sequence.  The sequenceID is a
         * common identifier that is passed along to each task in a
         * given task sequence.  For example, suppose file F1.java is
         * given to the compile as the file to be processed and
         * further suppose that it refers to code found in a file
         * F2.java.  Initially a task will be create for compiling
         * F1.java.  Suppose this task has a sequenceID 'x'.  After
         * parsing a new task will be created for the next stage of
         * processing for F1.java; this new task will also have the
         * sequenceID 'x'.  Now if in this stage the compiler decides
         * it needs to process F2.java, then F2.java is added to the
         * task queue by creating a new task.  This task will have a
         * sequenceID 'y' where y != x.  Among other things the
         * compiler can use these sequenceIDs to differentiate tasks
         * in compilation sequences initiated by the command line
         * arguments vs. those initiated by calls to catchUp.
         *
         * <pre><jml>
         * ensures !completed;
         * </jml></pre>
         * */
        public Task( int priority, Object sequenceID ) {
            this.priority = priority;
            this.sequenceID = sequenceID;
        }

        /**
         * Returns the relative priority of this task.  Smaller
         * values signify higher priority.  */
        public final int priority() {
            return priority;
        }


        /**
         * Returns the unique sequenceID for tasks in this sequence.  */
        public final /*@ pure @*/ Object sequenceID() {
            return sequenceID;
        }

        public final Task setToMainSequenceID() {
	    sequenceID = mainSequenceID();
	    return this;
	}

        /** 
         * Compare tasks, ensuring that tasks of equal priority are not equal.
         *
         * <pre><jml>
         * also 
         * old Task otherTask = (Task) other;
         * requires other instanceof Task;
         * {|
         *   requires priority != otherTask.priority;
         *   ensures \result == priority - otherTask.priority;
         * also
         *   requires priority == otherTask.priority;
         *   ensures \result == 0 <==> this == otherTask;
         * |}
         * </jml></pre>
         */
        public final int compareTo(/*@ non_null @*/ Object other) {
            // this is tricky, we don't want tasks of equal priority
            // to compare equal, else only one will survive in the set
            Task otherTask = (Task) other;
            int result = priority - ((Task) other).priority;
            if (result == 0) {
                result = this.uniqueNumber - otherTask.uniqueNumber;
            } // end of if 
            return result;
        }

        /**
         * Executes the task represented by this.  
         *
         * <pre><jml>
         * requires !completed;
         * ensures completed;
         * </jml></pre>
         */
        // !FIXME! need some sort of JML set statement: completed = true;
        public abstract void execute();

        //@ ghost public boolean completed;

        //@ spec_public
        private int priority;

        private Object sequenceID;

        /**
         * This is a unique number assigned to each task and used in
         * {@link #compareTo(Object)} to differentiate between two
         * tasks with the same priority.
         */
        private int uniqueNumber = nextUniqueTaskNumber++;
    }

    public void setMainSequenceID( Object sequenceID ) {
        if (mainSequenceID != null) {
            throw new IllegalStateException( "cannot change the main sequence"
                                             + " ID after it has been set" );
        } else {
            mainSequenceID = sequenceID;
        }
    }

    public Object mainSequenceID() {
        return mainSequenceID;
    }

    /** Returns the sequence ID of the currently active task. If no task
     * is active, <code>null</code> is returned. */
    protected Object activeSequenceID() {
        return activeSequenceID;
    }

    public interface Trees {
        public CompilerPassEnterable[] trees();
    }

    /**
     * This class parses a group of files, given by filenames as
     * strings, and generates a forest of ASTs. */
    public class ParseTask extends Task implements Trees {
        public ParseTask( ArrayList infiles ) {
            super( PRI_PARSE, new Object() );
            expected = ExpectedIndifferent.instance();
            String[] fileNames = new String[ infiles.size() ];
            infiles.toArray( fileNames );
            try {
                ArrayList inputFilesArrayList = Main.this.verifyFiles( fileNames );
                files = new File[ inputFilesArrayList.size() ];
                inputFilesArrayList.toArray( files );
            } catch (UnpositionedError e) {
		// FIXME - UnpositionedError inherits Throwable - shouldn't
		// it be part of the PositionedError hierarchy?
                reportTrouble( new FormattedException(e.getFormattedMessage()) );
                files = new File[0];
            }
        }

        public ParseTask( File fileName, ExpectedResult expected ) {
            super( PRI_PARSE, new Object() );
            this.expected = expected;
            files = new File[] { fileName };
        }


        /**
         * This is final so that subclasses do not eliminate check for
         * file already being processed.  Subclasses change behavior
         * by overriding <CODE>parseFiles(File)</CODE> method of
         * this.*/
        public final void execute() {
            Debug.setPrefix( "PARSE" );
            long startTime = System.currentTimeMillis();

            // A list containing the forest of ASTs
            ArrayList trees = new ArrayList(files.length);

            // Parses the input files
            for (int i = 0; i < files.length; i++) {
                File f = files[i];
                if (isCurrentlyParsingFor(f, expected)
                    || hasAlreadySuccessfullyParsed(f)
                    || hasAlreadyFailedToParseFor(f,expected)) {
                    continue;   // nothing to do f
                } else if (isCurrentlyParsingFor(f,ExpectedIndifferent.instance())) {
                    // Already parsing for something else in an outer
                    // frame on the call stack.  Registers that we are
                    // now interested in expected as well.
                    currentlyParsingFor(f, expected);
                    continue;   // let outer frame handle parsing
                }
                
                currentlyParsingFor(f,expected);
                JCompilationUnitType cunit = parseFile(f);
                if (!isAnExpectedResult(f, cunit)) {
                    if (sequenceID() != mainSequenceID) {
                        // clear errors since this was just a
                        // failed fishing expedition
                        errorFound = false;
                    }
                    failedParsing(f,cunit);
                    continue; // don't add cunit to trees
                } else {
                    successfullyParsed(f);
                }
                
                trees.add(cunit);
            }

            if (!quietMode()) {
                inform( MjcMessages.PARSING_TIME,
                        new Long(System.currentTimeMillis() - startTime) );
            }

            this.trees = new CompilerPassEnterable[trees.size()];
            trees.toArray(this.trees);
        }

        /**
         * Parses the given file and returns an AST representing it.
         *
         * @param       file    the file to be parsed
         * @return      the compilation unit defined by this file
         *
         * <pre><jml>
         * requires file!=null && (* file.exists() *); // !FIXME!
         * </jml></pre>
         *
         */
        protected JCompilationUnitType parseFile(File file) {
            Reader buffer;

            try {
                buffer = new BufferedReader(new FileReader( file ));
            } catch (IOException e) {
                reportTrouble(e);
                return null;
            }

            MjcLexer mjLexer;
            JavadocLexer docLexer;
            MjcParser parser = null; 
            JCompilationUnitType unit;

            // Used to switch lexers.  Aliases are created within the
            // lexer instances.
            ParsingController parsingController;

            Long duration = null;
            long lastTime = System.currentTimeMillis();

	    setAllowUniverses( options.universesx() ); // WMD

            parsingController = new ParsingController( buffer, file );
            mjLexer = new MjcLexer( parsingController, 
                                    options.source().equals("1.4"),
                                    options.multijava(),
				    allowUniverseKeywords, // WMD
                                    Main.this );
            docLexer = new JavadocLexer( parsingController );

            try {
                parsingController.addInputStream( mjLexer, "multijava" );
                parsingController.addInputStream( docLexer, "javadoc" );
                parsingController.selectInitial( "multijava" );
                if (!parseJavadoc) {
                    parsingController.discardAllTokensFor( "javadoc" );
                }

                parser = 
                    new MjcParser( Main.this, 
                                   parsingController.initialOutputStream(),
                                   parsingController, 
                                   options.generic(),
                                   options.multijava(), 
                                   options.relaxed(), 
				   allowUniverseKeywords, // WMD
                                   parseJavadoc );

                unit = parser.jCompilationUnit();
                unit.cachePassParameters( Main.this, destination );
            } catch( ParsingController.ConfigurationException e ) {
                reportTrouble(new FormattedException( MjcMessages.PARSER_INITIALIZATION_PROBLEM,
                        file.getName(), e.getMessage() ));
                noteError();
                unit = null;
            } catch( antlr.ANTLRException e ) {
                reportTrouble( parser.beautifyParserError(e) );
                noteError();
                unit = null;
            } catch( Exception e ) {
		reportTrouble(e);
                e.printStackTrace();
                unit = null;
            } finally {
                duration = new Long( System.currentTimeMillis() - lastTime );
                try {
                    buffer.close();
                } catch( IOException e ) {
                    reportTrouble( e );
                }
            }

            if( verboseMode() ) {
                inform( CompilerMessages.FILE_PARSED, file.getPath(), 
                        duration );
            }

            return unit;
        }

        /**
         * Returns the result of parsing the file(s) on which this was
         * instantiated.  
         *
         * <pre><jml>
         * also
         * requires completed;
         * </jml></pre>
         *
         */
        public CompilerPassEnterable[] trees() {
            return trees;
        }

        protected File[] files;

        /**
         * The result of parsing the file(s) on which this was instantiated.
         */
        protected CompilerPassEnterable[] trees;

        /**
         * The expected result of parsing files[0]. 
         */
        private final ExpectedResult expected;
        /*@ private invariant files.length > 0 ==> 
          @               expected instanceof ExpectedIndifferent; @*/
    }

    /**
     * This class is subclasses by classes that represent tasks
     * operating on a forest of ASTs.  */
    public abstract class TreeProcessingTask extends Task implements Trees {

        /**
         * Constructs a task that operates on an AST forest.
         *
         * <pre><jml>
         * requires trees != null;
         * requires (\forall int i; 0 <= i && i < trees.length;
         *           trees[i] != null && trees[i].arePassParametersCached);
         * </jml></pre>
         *
         */
        public TreeProcessingTask( int priority, Object sequenceID,
                                   CompilerPassEnterable[] trees,
                                   MessageDescription passCompletedMessage )
        {
            super( priority, sequenceID );
            this.trees = trees;
            this.passCompletedMessage = passCompletedMessage;
        }

        public void execute() {
            long startTime = System.currentTimeMillis();
            for (int i = 0; i < trees.length; i++) {
                processTree( trees[i] );
            }

            if (!quietMode()) {
                inform( passCompletedMessage,
                        new Long(System.currentTimeMillis() - startTime) );
            }
        }

        /**
         * Returns the result of processing the trees on which this was
         * instantiated.  
         *
         * <pre><jml>
         * also
         * requires completed;
         * </jml></pre>
         *
         */
        public CompilerPassEnterable[] trees() {
            return trees;
        }

        protected abstract void processTree( CompilerPassEnterable tree );

        protected MessageDescription passCompletedMessage;

        protected CompilerPassEnterable[] trees;
    }

    /**
     * This class preprocesses type and generic function imports and
     * groups external methods.  The class is constructed on an AST
     * forest and generates a new AST forest.
     */
    public class PreprocessTask extends TreeProcessingTask {
        /**
         * Constructs a task for preprocessing the ASTs in the given
         * forest.  Stores an alias to trees.  */
        public PreprocessTask( CompilerPassEnterable[] trees, 
			       Object sequenceID ) {
            super( PRI_PREPROCESS, sequenceID, trees,
                   MjcMessages.PREPROCESSED );
        }

        public void execute() {
            Debug.setPrefix( "INTERN" );

            super.execute();
        }

        /**
         * Preprocesses type and generic function imports. Groups
         * external methods by name, corresponding to the anchor
         * classes that will eventually be generated.
         *
         * @param tree the compilation unit */
        protected void processTree( CompilerPassEnterable tree ) {
            long lastTime = System.currentTimeMillis();

            try {
                tree.preprocessDependencies();
            } catch( PositionedError e ) {
                reportTrouble(e);
            }

            if (verboseMode()) {
                inform( MjcMessages.FILE_PREPROCESSED, 
                        tree.getTokenReference().file(), 
                        new Long(System.currentTimeMillis() - lastTime));
            }

        }
    }

    /**
     * This class checks the interfaces of the declarations in the
     * source code.  The class is constructed on an AST forest.  */
    public class CheckInterfaceTask extends TreeProcessingTask {
        /**
         * Constructs a task for checking the interfaces in the given
         * forest.  Stores an alias to trees.  */
        public CheckInterfaceTask( CompilerPassEnterable[] trees, 
                                   Object sequenceID ) {
            super( PRI_CHECK_INTERFACE, sequenceID, trees,
                   CompilerMessages.INTERFACES_CHECKED);
        }

        public void execute() {
            Debug.setPrefix( "CHK_INT" );
            super.execute();
        }

        /**
         * Checks the basic interfaces to make sure things generally
         * look OK.  This pass gathers information about the type
         * signatures of everything (imported class files, classes
         * being compiled, methods, fields, etc...)  needed for the
         * later passes.  This information is stored in a
         * CCompilationUnit instance and instances of CMember that are
         * bound to the AST.  Also adds things like the default
         * constructor and the initializer method to the AST (these
         * are suppressed during pretty-printing).
         *
         * @param tree the compilation unit */
        protected void processTree(CompilerPassEnterable tree) {
            long lastTime = System.currentTimeMillis();

            try {
                tree.checkInterface();
            } catch (PositionedError e) {
                reportTrouble(e);
            }

            if (verboseMode()) {
                inform( MjcMessages.FILE_INTERFACE_CHECKED, 
                        tree.getTokenReference().file(),
                        new Long(System.currentTimeMillis() - lastTime));
            }
        }
    }

    /**
     * This class typechecks the initializers of the declarations in
     * the source code.  The class is constructed on an AST forest.
     * */
    public class CheckInitializerTask extends TreeProcessingTask {
        /**
         * Constructs a task for checking the initializers in the given
         * forest.  Stores an alias to trees.  */
        public CheckInitializerTask( CompilerPassEnterable[] trees, 
                                     Object sequenceID ) {
            super( PRI_CHECK_INITIALIZER, sequenceID, trees,
                   MjcMessages.INITIALIZERS_CHECKED );
        }

        public void execute() {
            Debug.setPrefix( "CHK_INIT" );
            super.execute();
        }

        /**
         * Checks the initializers created during the checkInterface
         * pass and performs some other checks that can be performed
         * simply before full blown typechecking.  Among these other
         * checks are those for appropriateness of subclassing, such
         * as inheritance circularities and
         * &quot;reimplementation&quot; of interfaces.
         *
         * @param tree the compilation unit to be checked */
        protected void processTree(CompilerPassEnterable tree) {
            long lastTime = System.currentTimeMillis();

            try {
                tree.checkInitializers();
            } catch (PositionedError e) {
                reportTrouble(e);
            }

            if (verboseMode()) {
                inform( MjcMessages.FILE_INITIALIZERS_CHECKED, 
                        tree.getTokenReference().file(),
                        new Long(System.currentTimeMillis() - lastTime));
            }
        }
    }

    /**
     * This class resolves value specializer expressions to the
     * compile-time constants they represent. The class is constructed
     * on an AST forest.
     */
    public class ResolveSpecializerTask extends TreeProcessingTask {
        /**
         * Constructs a task for resolving the specializers in the given
         * forest.  Stores an alias to trees.  */
        public ResolveSpecializerTask( CompilerPassEnterable[] trees, 
                                       Object sequenceID ) {
            super( PRI_RESOLVE_SPECIALIZER, sequenceID, trees,
                   MjcMessages.SPECIALIZERS_RESOLVED );
        }

        public void execute() {
            Debug.setPrefix( "SPEC_RES" );
            super.execute();
        }

        /**
         * Resolves value specializer expressions to the compile-time
         * constants they represent.
         *
         * @param tree the AST to be checked
         */
        protected void processTree(CompilerPassEnterable tree) {
            long lastTime = System.currentTimeMillis();

            try {
                tree.resolveSpecializers();
            } catch (PositionedError e) {
                reportTrouble(e);
            }

            if (verboseMode()) {
                inform( MjcMessages.FILE_SPECIALIZERS_RESOLVED,
                        tree.getTokenReference().file(),
                        new Long(System.currentTimeMillis() - lastTime));
            }
        }
    }

    /**
     * This class resolves the top methods of all method declarations.
     * The class is constructed on an AST forest.
     * */
    public class ResolveTopMethodTask extends TreeProcessingTask {
        /**
         * Constructs a task for resolving the top methods in the
         * given forest.  Stores an alias to trees.  */
        public ResolveTopMethodTask( CompilerPassEnterable[] trees, 
                                     Object sequenceID ) {
            super( PRI_TOP_METHODS, sequenceID, trees,
                   MjcMessages.TOP_METHODS_RESOLVED );
        }

        public void execute() {
            Debug.setPrefix( "TOP_METH" );
            super.execute();
        }

        protected void processTree(CompilerPassEnterable tree) {
            long lastTime = System.currentTimeMillis();

            try {
                tree.resolveTopMethods();
            } catch (PositionedError e) {
                reportTrouble(e);
            }

            if (verboseMode()) {
                inform( MjcMessages.FILE_TOP_METHODS_RESOLVED, 
                        tree.getTokenReference().file(),
                        new Long(System.currentTimeMillis() - lastTime));
            }
        }
    }

    /**
     * This class typechecks the source code.  The class is
     * constructed on an AST forest.  */
    public class TypecheckTask extends TreeProcessingTask {
        /**
         * Constructs a task for checking the initializers in the given
         * forest.  Stores an alias to trees.  */
        public TypecheckTask( CompilerPassEnterable[] trees, 
                              Object sequenceID ) {
            super( PRI_TYPECHECK, sequenceID, trees,
                   MjcMessages.TYPECHECK_TIME );
        }
        public TypecheckTask( int priority, Object sequenceID,
                              CompilerPassEnterable[] trees,
                              MessageDescription passCompletedMessage )
        {
            super( priority, sequenceID, trees, passCompletedMessage );
        }

        public void execute() {
            Debug.setPrefix( "TYPECHK" );
            super.execute();
        }

        /**
         * Check that body of a given compilation unit is correct.
         *
         * @param       tree           the compilation unit
         */
        protected void processTree( CompilerPassEnterable tree ) {
            long lastTime = System.currentTimeMillis();
            try {
                tree.typecheck();
            } catch( PositionedError e ) {
                reportTrouble( e );
            }
            if( verboseMode() ) {
                inform( CompilerMessages.BODY_CHECKED, 
                        tree.getTokenReference().file(),
                        new Long(System.currentTimeMillis() - lastTime) );
            }
        }
    }

    /**
     * This class pretty prints the trees in the AST forest.  It is
     * subclassed in org.jmlspecs.jmlrac.main.  */
    public class PrettyPrintTask extends TreeProcessingTask {
        /**
         * Constructs a task for pretty printer the trees in the given
         * forest.  Stores an alias to trees.  */
        public PrettyPrintTask( CompilerPassEnterable[] trees, 
                                Object sequenceID ) {
            super( PRI_PRETTY_PRINT, sequenceID, 
                   trees, MjcMessages.PRINT_TIME );
        }

        public void execute() {
            // just regurgitating the source code:
            Debug.setPrefix( "PRINT" );
            super.execute();
        }

        /**
         * generate the source code of parsed compilation unit
         * @param       tree           the compilation unit
         */
        protected void processTree( CompilerPassEnterable tree ) {
            long        lastTime = System.currentTimeMillis();

            //tree.prettyPrint();
            prettyPrint(tree);

            if (verboseMode()) {
                inform( CompilerMessages.JAVA_CODE_GENERATED, 
                        tree.getTokenReference().file(),
                        new Long(System.currentTimeMillis() - lastTime) );
            }
        }
    }

    public void prettyPrint(CompilerPassEnterable tree) {
        MjcPrettyPrinter pp =
            new MjcPrettyPrinter( destination.textTarget( tree.getTokenReference().
                                                          name() +
                                                          ".gen" ),
                                  new CModifier() );
        tree.accept( pp );
        pp.close();

    }

    /**
     * This class pretty prints the trees in the AST forest.  */
    public class TranslateMJTask extends TreeProcessingTask {
        /**
         * Constructs a task for pretty printer the trees in the given
         * forest.  Stores an alias to trees.  */
        public TranslateMJTask( CompilerPassEnterable[] trees, 
                                Object sequenceID ) {
            super( PRI_TRANSLATE_MJ, sequenceID, 
                   trees, MjcMessages.TRANSLATE_MJ_TIME );
        }

        public void execute() {
            // Next we translate MultiJava constructs into pure
            // Java (e.g., create dispatchers for multimethods,
            // create anchor classes)
            Debug.setPrefix( "TRANS" );
            super.execute();
            codeGenNeeded = true;
        }

        /**
         * Refactors the given compilation unit to include dispatchers
         * for multimethods and other code necessary for running
         * MultiJava code on a standard JVM.  */
        protected void processTree( CompilerPassEnterable tree ) {
            long lastTime = System.currentTimeMillis();

            tree.translateMJ();

            if( verboseMode() ) {
                inform( MjcMessages.FILE_TRANSLATING_COMPLETED,
                        tree.getTokenReference().file(),
                        new Long(System.currentTimeMillis() - lastTime) );
            }
        }
    }

    // New task priorities in subclasses of the compiler should use
    // values that split the difference between these values to leave
    // room for future expansion.  Mutually exclusive task can have
    // the same priority.
    public final static int PRI_PARSE =                 100;
    public final static int PRI_PREPROCESS =            200;
    public final static int PRI_CHECK_INTERFACE =       300;
    public final static int PRI_CHECK_INITIALIZER =     400;
    public final static int PRI_RESOLVE_SPECIALIZER =   425;
    public final static int PRI_TOP_METHODS =           450;
    public final static int PRI_TYPECHECK =             500;
    public final static int PRI_PRETTY_PRINT =          600;
    public final static int PRI_TRANSLATE_MJ =          600;

    // Class to support recording cumulative times

    static public class TaskTimes {
        private static TreeMap map = new TreeMap();

        public static void init() { map = new TreeMap(); }

        public static void add(Object o, long time) {
            Long oo = (Long)map.get(o);
            if (oo == null) {
                map.put(o, oo = new Long(time));
            } else {
                map.put(o, new Long(oo.longValue() + time));
            }
        }

        public static void report() {
            Iterator i = map.keySet().iterator();
            final String spaces = "                                             ";
            while (i.hasNext()) {
                Object k = i.next();
                Long d = (Long)(map.get(k));
                int digits = 1;
                long n = d.longValue();
                while (n > 9) { ++digits; n = n/10; }
                System.out.println(spaces.substring(0,12-digits) + d + "\t\t" + k.toString());
            }
        }
    }
        
    // ----------------------------------------------------------------------
    // PROTECTED DATA MEMBERS
    // ----------------------------------------------------------------------

    /*@ invariant taskQueue != null ==> 
      @                 (\forall Object o; taskQueue.contains(o);
      @                         o instanceof Task);
      @*/
    /**
     * Priority queue of tasks to be completed.
     * @see org.multijava.mjc.Main.Task
     */
    protected TreeSet taskQueue = new TreeSet();

    /**
     * The priority of the top-most active task, where top-most is
     * defined in terms of the recursive invocations of
     * processTaskQueue on the call stack.  */
    protected int activeTaskPriority = -1;

    /**
     * Records whether an error was found during processing of a task.
     */
    protected boolean           errorFound;
    protected int               errorLimit = Integer.MAX_VALUE;
    protected int               mostSevereWarningIssued = Integer.MAX_VALUE;
    // more severe warnings have lower severity values

    /** Whether any unchecked warning has been issued during compilation. */
    protected boolean uncheckedWarningsIssued = false;

    /**
     * Register that an error was found during processing of a task.
     */
    protected void noteError() {
        errorFound = true;
        //if (errorLimit > activeTaskPriority) System.out.println("SETTING ERRORLIMIT TO " + activeTaskPriority + " " + errorLimit);
        if (errorLimit > activeTaskPriority) errorLimit = activeTaskPriority;
    }

    /*@ spec_public @*/ protected MjcCommonOptions options;
    public MjcCommonOptions options() { return options; }

    private WarningFilter       filter;

    // all generated classes
    /*@ spec_public @*/ protected ArrayList classes = new ArrayList(100);

    protected Destination destination;

    protected boolean codeGenNeeded = false;

    /**
     * Stores all the compilation unit contexts created for this
     * compiler and not yet disposed of.  
     *
     * @see #createCompilationUnitContext(JCompilationUnitType, CCompilationUnit)
     * @see #adoptCompilationUnitContext(CCompilationUnitContextType)
     */
    /*@ spec_public @*/ protected HashSet contextsCreated = new HashSet();

    /** When the following flag is false, the parser ignores javadoc
     * comments (as it does regular comments); when the flag is true,
     * javadoc comments are collected as part of the AST; derived
     * classes should set this flag after the super class
     * constructor is called, but before parsing is begun.  */
    protected boolean parseJavadoc = false;

    /**
     * Stores an alias to the sequenceID of the task sequence created
     * for the files given as command line arguments. 
     *
     * <pre><jml>
     * private invariant (\forall Task t; t != null; 
     *                    t.sequenceID() == mainSequenceID <==>
     *                    (* t does not belong to a task sequence used to 
     *                       catch up recursively loaded files *));
     * </jml></pre>
     *
     */
    private Object mainSequenceID;

    /** SequenceID of the currently active Task. */
    protected Object activeSequenceID;

    /**
     * A relation mapping files currently being parsed by the
     * compiler to the expected results.
     */
    /*@ private invariant (\forall Object o; currentlyParsingRelation.containsValue(o);
      @                                      o instanceof Set); 
      @*/
    private /*@ spec_public @*/ HashMap currentlyParsingRelation = new HashMap();

    /**
     * A relation mapping files already parsed by the compiler to the
     * results that were expected but NOT found.
     */
    /*@ private invariant (\forall Object o; failedParsingRelation.containsValue(o);
      @                                      o instanceof Set); 
      @*/
    private /*@ spec_public @*/ HashMap failedParsingRelation = new HashMap();

    /**
     * A set of files previously parsed by the compiler for which the results were
     * not discarded.
     */
    /*@ private invariant (\forall Object o; alreadyParsedSet.contains(o);
      @                                      o instanceof String); 
      @*/
    private /*@ spec_public @*/ HashSet alreadyParsedSet = new HashSet();

    /**
     * A unique number that will be assigned to the next task created.
     * This really should be a static member of the Task inner class,
     * but Java does not allow inner classes to have static members.
     */
    private static int nextUniqueTaskNumber = 0;

    /** The name of this app as it should be given in error or information messages. */
    //@ invariant appName != null;
    protected String appName = "**Unspecified app**";

    private static boolean compilationInterrupted = false;



    /** Whether the special tokens for the universe type system should
     * be allowed.
     * This separation was added for JML support, where special tokens
     * with a preceding backslash will always be allowed.
     * Then we need to be able to just turn on the typechecking.
     * 
     * By WMD.
     */
    protected boolean allowUniverseKeywords = false;

    /** Whether we should check the program for correctness according
     * to the universe type system.
     * For MultiJava this should always be the same as
     * allowUniverseTokens, but for JML this might be different.
     * There it should be possible to typecheck a program without the
     * new keywords.
     * Keywords within JML-comments starting with backslash will be
     * used there.
     *
     * By WMD.
     */
    protected boolean allowUniverseChecks = false;

    /** Whether we should check the purity of methods.
     * We currently have a rather restrictive definition of purity
     * and there exist better, but more complex, algorithms.
     * You can combine your own purity checks with our Universe checks
     * by not enabling this option.
     * This only influences the purity checking, the other universe
     * rules are not influenced, i.e. we still only allow that
     * pure methods are called on readonly references.
     *
     * By WMD.
     */
    protected boolean allowUniversePurity = false;

    /** Enable the generation of dynamic checks that verify that the
     * universe invariant holds during the execution.
     * This creates checks for things like casts and instanceof tests.
     * Needs allowUniverseChecks to also be true, to make sure that 
     * the static properties hold.
     *
     * By WMD.
     */
    protected boolean allowUniverseDynChecks = false;

    /** Enable the generation of extended bytecode that contains the
     * Universe annotations.
     * This version generates attributes containing the Universe modifiers.
     * This allows easy compilation of programs without access to
     * all source files.
     * Needs allowUniverseChecks to also be true, to make sure that 
     * the static properties hold.
     *
     * By WMD.
     */
    protected boolean allowUniverseBytecode = false;
    
    /** Enable the generation of extended bytecode that contains the
     * Universe annotations.
     * This version generates Java 5 annotations for the Universe modifiers.
     *
     * By AHS.
     */
    protected boolean allowUniverseAnnotations = false;
    
    /** Universe encoding version.
     * By AHS.
     */
    protected String universeVersion = "1.0";




    /** Command line option to make all universe features enabled.
     */
    public static final String UNIVERSE_FULL = "full";

    /** Command line option to enable allowUniverseDynChecks.
     */
    public static final String UNIVERSE_DYNCHECKS = "dynchecks";

    /** Command line option to enable allowUniverseBytecode.
     */
    public static final String UNIVERSE_BYTECODE = "xbytecode";
    
    /** AHS: Command line option that chooses Java 5 annotations as output
     * style for universe modifiers. */
    public static final String UNIVERSE_ANNOTATIONS = "annotations";
    
    /** Command line option to enable allowUniverseKeywords.
     */
    public static final String UNIVERSE_PARSE = "parse";

    /** Command line option to enable allowUniverseChecks.
     */
    public static final String UNIVERSE_CHECKS = "check";

    /** Command line option to enable allowUniversePurity.
     */
    public static final String UNIVERSE_PURITY = "purity";

    /** Command line option to disable all universe features.
     */
    public static final String UNIVERSE_NO = "no";
}
