/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MemberAccess.java,v 1.10 2006/02/11 03:05:57 chalin Exp $
 */

package org.multijava.mjc;
import org.multijava.util.compiler.PositionedError;

/**
 * This class represents and contains the information needed to determine 
 * whether a member of a class or compilation unit can be accessed 
 * from some other member.  
 * It was factored out of <code>CMember</code> so it could more easily be 
 * extended by JML classes without having to repeat code 
 * over several subclasses.  It has the interface checking code for 
 * ensuring that modifiers are valid for the member containing it, 
 * e.g., class, field, and method signature classes.
 * The modifier checking code was factored out of 
 * <code>JTypeDeclaration</code>, <code>JFieldDeclaration</code>, and   
 * <code>JMethodDeclaration</code> rather than using callbacks so 
 * the modifier checking could all be done in one class.  
 * Callbacks are used in some of the accessibility computations.  
 *
 * @see	CMember
 */
public class MemberAccess extends org.multijava.util.Utils 
    implements Constants 
{

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a member export
     * @param	owner		the owner of this member
     * @param	modifiers	the modifiers on this member
     */
    public MemberAccess(CClass owner, long modifiers) 
    {
	this.owner = owner;
	this.host = owner;
	this.modifiers = modifiers;
    }

    /**
     * Constructs a top-level member export
     * @param	owner		the owner of this member
     * @param	host		the host of this member's declaration
     * @param	modifiers	the modifiers on this member
     */
    public MemberAccess(CClass owner, CMemberHost host, long modifiers ) 
    {
	this.owner = owner;
	this.host = host;
	this.modifiers = modifiers;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the owner of this member or null if this is a top-level
     * class declaration.
     *
     * @return the class that owns this member */
    public /*@ pure @*/ CClass owner() {
	return owner;
    }

    /**
     * Sets the owner of this member.
     *
     * @see #owner
     */
    protected void setOwner( CClass owner ) {
	this.owner = owner;
    }

    /**
     * Returns the host of this member's declaration, which is the
     * same value returned by <code>owner()</code> unless this is a
     * top-level declaration.
     *
     * @return the host of this member's declaration */
    public /*@ pure @*/ CMemberHost host() {
	return host;
    }

    /**
     * @return the fully qualified name of this member's owner
     */
    public /*@ pure @*/ String getOwnerName() {
	return owner.qualifiedName();
    }

    /**
     * @return the modifiers of this member
     */
    public /*@ pure @*/ long modifiers() { 
	return modifiers;
    }

    /**
     * Add (bit or) to the modifiers of this member.
     */
    public void addModifiers(long modifiers) { 
	this.modifiers |= modifiers;
    }

    /**
     * Sets the modifiers of this member
     */
    public void setModifiers(long modifiers) { 
	this.modifiers = modifiers;
    }

    // ----------------------------------------------------------------------
    // ACCESSORS (QUICK)
    // ----------------------------------------------------------------------

    /**
     * A member is static if it is annotated with the 
     * <code>static</code> modifier.
     *
     * @return	true if this member is static
     */
    public /*@ pure @*/ boolean isStatic() {
	return hasFlag(modifiers, ACC_STATIC);
    }

    /**
     * A member is pure if it is annotated with the 
     * <code>pure</code> modifier.
     * By WMD.
     *
     * @return	true if this member is pure
     */
    public /*@ pure @*/ boolean isPure() {
	return hasFlag(modifiers, ACC_PURE);
    }

    /**
     * A member is abstract if it is annotated with the 
     * <code>abstract</code> modifier.
     *
     * @return	true if this member is abstract
     */
    public /*@ pure @*/ boolean isAbstract() {
	return hasFlag(modifiers, ACC_ABSTRACT);
    }

    /**
     * A member is public if it is annotated with the 
     * <code>public</code> modifier.
     *
     * @return	true if this member is public
     */
    public /*@ pure @*/ boolean isPublic() {
	return hasFlag(modifiers, ACC_PUBLIC);
    }

    /**
     * A member is protected if it is annotated with the 
     * <code>protected</code> modifier.
     *
     * @return	true if this member is protected
     */
    public /*@ pure @*/ boolean isProtected() {
	return hasFlag(modifiers, ACC_PROTECTED);
    }

    /**
     * A member is private if it is annotated with the 
     * <code>private</code> modifier.
     *
     * @return	true if this member is private
     */
    public /*@ pure @*/ boolean isPrivate() {
	return hasFlag(modifiers, ACC_PRIVATE);
    }

    public /*@ pure @*/ boolean hasDefaultAccess() {
	return !isPublic() && !isProtected() && !isPrivate();
    }

    /**
     * A member is final if it is annotated with the 
     * <code>final</code> modifier.
     *
     * @return	true if this member is final.
     */
    public /*@ pure @*/ boolean isFinal() {
	return hasFlag(modifiers, ACC_FINAL);
    }

    /**
     * A member is synchronized if it is annotated with the 
     * <code>synchronized</code> modifier.
     *
     * @return	true if this member is synchronized. 
     */
    public boolean isSynchronized() {
        return hasFlag(modifiers, ACC_SYNCHRONIZED);
    }

    /**
     * A member is transient if it is annotated with the 
     * <code>transient</code> modifier.
     *
     * @return	true if this member is transient. 
     */
    public boolean isTransient() {
        return hasFlag(modifiers, ACC_TRANSIENT);
    }

    /**
     * A member is volatile if it is annotated with the 
     * <code>volatile</code> modifier.
     *
     * @return	true if this member is volatile. 
     */
    public boolean isVolatile() {
        return hasFlag(modifiers, ACC_VOLATILE);
    }

    /**
     * A member is native if it is annotated with the 
     * <code>native</code> modifier.
     *
     * @return	true if this member is native.
     */
    public /*@ pure @*/ boolean isNative() {
	return hasFlag(modifiers, ACC_NATIVE);
    }

    /**
     * A member is strictfp if it is annotated with the 
     * <code>strictfp</code> modifier.
     *
     * @return	true if this member is strictfp.
     */
    public /*@ pure @*/ boolean isStrictFP() {
	return hasFlag(modifiers, ACC_STRICT);
    }

    /**
     * Returns true iff this member is an interface.
     */
    public /*@ pure @*/ boolean isInterface() {
	return hasFlag(modifiers, ACC_INTERFACE);
    }

    /**
     * Indicates whether this is accessible from the given host.
     *
     * @param	member	the member that needs to be accessed
     * @param	from	the host that wants to access member
     * @return	true iff the given host is allowed access to this member
     */
    public /*@ pure @*/ boolean isMemberVisibleIn( CMember member, 
						   CMemberHost from ) 
    {
	// public members and members defined in lexically
	// nested contexts are accessible
	if( isPublic() || member.isLocalTo( from ) ) {
	    return true;
	}

	// nested types have the same access privileges as their host class
	// (except for protected members of the class, but these are 
	// considered below)
	if( from.isNestedType() && isMemberVisibleIn( member, from.host() )) {
	    return true;
	}

	// non-private members are accessible throughout their home package
	if( !isPrivate() && host.packageName() == from.packageName() ) {
	    return true;
	}

	// protected members are accessible from subclasses of the
	// logical owner, i.e., subclasses of an augmented class can
	// access protected augmenting methods
	// if( isProtected() && from.isClass() && from.descendsFrom( owner ) ) {
	if( isProtected() && member.hasProtectedVisibilityIn( from ) ) {
	    return true;
	}
	
	// no soup for you
	return false;
    }

    /**
     * Returns the compilation unit that hosts this member.
     */
    public CCompilationUnit getCCompilationUnit() {
	if( hostCompilationUnit == null ) {
	    // climb up through the nested hosts until we reach a
	    // compilation unit
	    CMemberHost climbingHost = host;
	    while( climbingHost != null && 
		   !climbingHost.isCompilationUnit() ) {
		climbingHost = climbingHost.host();
	    }
	    hostCompilationUnit = (CCompilationUnit) climbingHost;
	}
	return hostCompilationUnit;
    }

    /**
     * Used to determine whether a method is external, i.e., whether its
     * host is different than its logical owner.
     *
     * @return	true iff the method is external
     */
    public /*@ pure @*/ boolean isExternalMethod() {
	return host.isCompilationUnit() || !owner.equals( host );
    }

    /**
     * Removes the 'private' modifier when necessary.
     */
    public void removePrivateModifier() {
	modifiers = stripPrivateModifier( modifiers );
    }

    /**
     * Removes the 'static' modifier when necessary.
     */
    public void removeStaticModifier() {
	modifiers = (modifiers & ~ACC_STATIC);
    }

    /**
     * Removes the 'final' modifier when necessary.
     */
    public void removeFinalModifier() {
	modifiers = (modifiers & ~ACC_FINAL);
    }

    /**
     * Removes the 'abstract' modifier when necessary.
     */
    public void removeAbstractModifier() {
	modifiers = (modifiers & ~ACC_ABSTRACT);
    }

    // ----------------------------------------------------------------------
    // MODIFIER CHECKS
    // ----------------------------------------------------------------------

    /**
     * Check illegal combinations of modifiers common to classes,
     * interfaces, fields, and methods.
     */
    public void checkAccessModifiers( CContextType context,
				      JMemberDeclaration member )
	throws PositionedError
    {
	// 8.3.1 and 8.4.3 Access Modifiers
	boolean hasInvalidModifiers =
	    (isPublic() && (isProtected() || isPrivate())) 
	    || (isProtected() && isPrivate());
	
	member.check( context,
		      !hasInvalidModifiers,
		      MjcMessages.INVALID_MODIFIER_COMBINATION, 
		      member.ident() );
    }

    /**
     * Check for illegal combinations of modifiers disallowed in classes 
     * and interfaces.
     */
    public void checkClassModifiers(CContextType context,
				    JTypeDeclaration member )
	throws PositionedError
    {
	// check the modifier combinations common to all declarations
        this.checkAccessModifiers( context, member );

	// Check modifiers
	// 8.1.1 Class Modifiers
	member.check(context,
		     !hasFlag(modifiers, invalidClassModifiers),
		     MjcMessages.CLASS_FLAGS);

	// 8.1.2 Class Modifiers
	// protected and private are not allowed in top level classes
	// member.isAtTopLevel() => !(isProtected() || isPrivate()),
	member.check(context,
		     !member.isAtTopLevel() || !(isProtected() || isPrivate()),
		     MjcMessages.CLASS_FLAGS);
    }

    /**
     * Check for illegal combinations of modifiers disallowed in field 
     * declarations.
     */
    public void checkFieldModifiers( CContextType context,
				     JFieldDeclaration member )
	throws PositionedError
    {
	// check the modifier combinations common to all declarations
        this.checkAccessModifiers( context, member );

	// Check modifiers
	// 8.3.1 Field Modifiers
	member.check(context,
		     !hasFlag(modifiers, invalidFieldModifiers),
		     MjcMessages.FIELD_FLAGS);
    }

    /**
     * Check for illegal combinations of modifiers disallowed in 
     * interface field declarations.
     *
     * Adds the default public, static, and final modifiers (the only ones 
     * allowed) for fields declared in interfaces.
     *
     * @return	the new set of modifiers 
     */
    public long checkInterfaceFieldModifiers( CContextType context, 
					     JFieldDeclaration member )
	throws PositionedError
    {
	// Fields in an interface are automatically public, static, and final

	if (owner.isInterface()) {
	    modifiers |= interfaceFieldModifiers;
	    member.check(context,
			 !hasOtherFlags( stripNonJavaModifiers(modifiers),
					 interfaceFieldModifiers ),
			 MjcMessages.FIELD_FLAGS_IN_INTERFACE);
	}
	return modifiers;
    }

    /**
     * Check for illegal combinations of modifiers disallowed in method 
     * declarations.
     */
    public void checkMethodModifiers( CContextType context,
				      JMethodDeclaration member )
	throws PositionedError
    {
	// Check modifiers
	// 8.4.3 Method Modifiers

	member.check( context,
		      !hasFlag(modifiers, invalidMethodModifiers),
		      MjcMessages.METHOD_FLAGS );

        // Check illegal combinations of allowed modifiers

	member.check( context,
		      !( isAbstract()
			 && hasFlag(modifiers, 
				    invalidAbstractMethodModifiers) ),
		      MjcMessages.METHOD_MODIFIER_COMBINATION,
		      member.ident() );

	member.check( context,
		      !( isNative() && isStrictFP() ),
		      MjcMessages.METHOD_NATIVE_STRICT,
		      member.ident() );
    }

    /**
     * Check for illegal combinations of modifiers disallowed in external 
     * method declarations.
     */
    public void checkExternalMethodModifiers( CContextType context,
					      JMethodDeclaration member )
	throws PositionedError
    {
	if (context.getCompiler().RMJ()) {
	    // RMJ: don't disallow abstract top-level methods.
	} else {
	    // MultiJava, R1 consequence, external methods cannot be
	    // abstract
	    member.check( context, 
			  !isAbstract(),
			  MjcMessages.ABSTRACT_TOP_LEVEL_METHOD, 
			  member.ident() );
	}

	// External methods may not be static (yet)
	member.check( context, 
		      !isStatic(),
		      MjcMessages.STATIC_EXT_METHOD, member.ident() );
    }

    /**
     * Check for illegal combinations of modifiers disallowed in 
     * interface method declarations.
     *
     * Adds the default public and abstract modifiers (the only ones allowed) 
     * for methods declared in interfaces.
     *
     * @return	the new set of modifiers 
     */
    public long checkInterfaceMethodModifiers( CContextType context, 
					      JMethodDeclaration member )
	throws PositionedError
    {
	// Check for illegal combinations of modifiers disallowed in 
	// interface methods.

	boolean	isExported = !(member instanceof JInitializerDeclaration);
	// 9.4 Methods in interfaces are implicitly public and
	// abstract, no other modifiers are allowed
	if (isExported && owner.isInterface()) {
	    // Adds the public and abstract modifiers that are the default 
	    // for methods declared in interfaces.
	    modifiers |= interfaceMethodModifiers;
	    member.check( context,
			  !hasOtherFlags(stripNonJavaModifiers(modifiers),
					 interfaceMethodModifiers),
			  MjcMessages.METHOD_FLAGS_IN_INTERFACE, 
			  member.ident() );
	}
	return modifiers;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * Denotes the class in which this member appears logically.  This
     * is set to null for top-level classes.  It is set to the class
     * of the receiver for top-level method declarations.  It is set
     * to the outer class for inner classes.  It is set to the
     * containing class for fields. */
    protected CClass			owner;

    /**
     * Denotes the context in which this member is declared.  In most
     * cases it is the same as <code>owner</code>.  For top-level
     * declarations it is the compilation unit in which the member is
     * declared. */
    protected final CMemberHost		host;

    /** Subclasses may need to modify this field to hack around the
     *  problem of privileged access to private methods/fields from an
     *  enclosing/inner class.
     */
    protected long			modifiers; 

    /**
     * Stores a cached copy of the result of getCCompilationUnit.
     *
     * @see #getCCompilationUnit()
     */
    protected CCompilationUnit hostCompilationUnit;

    // ----------------------------------------------------------------------
    // STATIC PROTECTED DATA MEMBERS
    // ----------------------------------------------------------------------

    // the following are the valid class modifiers:
    //  ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE 
    //  + ACC_ABSTRACT + ACC_STATIC + ACC_FINAL 
    //  + ACC_STRICT;
    protected static final long invalidClassModifiers 
	= ACC_SYNCHRONIZED + ACC_VOLATILE + ACC_TRANSIENT + ACC_NATIVE;

    // WMD: have to allow pure classes here, b/c JML needs it
    //	+ ACC_PURE; // WMD: classes can not be pure

    // the following are the valid field modifiers:
    //  ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE 
    //  + ACC_STATIC + ACC_FINAL + ACC_TRANSIENT + ACC_VOLATILE;
    protected static final long invalidFieldModifiers 
	= ACC_ABSTRACT + ACC_SYNCHRONIZED + ACC_NATIVE + ACC_STRICT
	+ ACC_PURE; // WMD: fields can not be pure

    // the following are the valid method modifiers:
    //	ACC_PUBLIC + ACC_PROTECTED + ACC_PRIVATE 
    //	+ ACC_ABSTRACT + ACC_STATIC + ACC_FINAL 
    //	+ ACC_SYNCHRONIZED + ACC_NATIVE + ACC_STRICT;
    // WMD: + ACC_PURE;
    protected static final long invalidMethodModifiers 
	= ACC_VOLATILE + ACC_TRANSIENT;

    protected static final long invalidAbstractMethodModifiers 
	= ACC_PRIVATE + ACC_STATIC + ACC_FINAL 
	+ ACC_SYNCHRONIZED + ACC_NATIVE + ACC_STRICT;

    protected static final long interfaceFieldModifiers 
	= ACC_PUBLIC + ACC_STATIC + ACC_FINAL;

    protected static final long interfaceMethodModifiers 
	= ACC_PUBLIC + ACC_ABSTRACT;

}
