/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjcPrettyPrinter.java,v 1.34 2008/03/16 15:52:56 wdietl Exp $
 */

package org.multijava.mjc;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;
import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.JavaStyleComment;
import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.compiler.TabbedPrintWriter;

/**
 * This class implements a Java pretty printer
 */
public class MjcPrettyPrinter extends org.multijava.util.Utils 
    implements Constants, MjcVisitor {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * construct a pretty printer object for java code
     */
    public MjcPrettyPrinter( java.io.Writer wr, ModifierUtility modUtil ) {
	this.p = new TabbedPrintWriter( wr );
	this.modUtil = modUtil;
    }

    /**
     * construct a pretty printer object for java code
     * @param	fileName	the file into which the code is generated
     */
    public MjcPrettyPrinter( String fileName, ModifierUtility modUtil ) {
	this( new File( fileName ), modUtil );
    }
    
    /**
     * construct a pretty printer object for java code
     * @param	file	the file into which the code is generated
     */
    public MjcPrettyPrinter( File file, ModifierUtility modUtil ) {
	try {
	    file.getParentFile().mkdirs();
	    this.p = new TabbedPrintWriter(
			   new java.io.BufferedWriter(
				  new java.io.FileWriter( file )));
	    this.modUtil = modUtil;
	} catch( Exception e ) {
	    e.printStackTrace();
	    System.err.println("can't write: " + file + e);
	    // !FIXME! probably should throw an exception
	}
    }
    
    /**
     * construct a pretty printer object for java code
     * @param	p	the file into which the code is generated
     */
    public MjcPrettyPrinter( TabbedPrintWriter p ) {
	this.p = p;
    }
    
    /**
     * Close the stream at the end
     */
    public void close() {
	p.close();
    }

    public void setPos( int pos ) {
	this.pos = pos;
    }

    // ----------------------------------------------------------------------
    // COMPILATION UNIT
    // ----------------------------------------------------------------------

    /**
     * prints a compilation unit
     */
    public void visitCompilationUnit(/*@non_null@*/ JCompilationUnit self ) {
	JPackageName packageName = self.packageName();
	JPackageImportType[] importedPackages = self.importedPackages();
	JClassOrGFImportType[] importedClasses = self.importedClasses();
	JClassOrGFImportType[] importedGFs = self.importedGFs();
	JTypeDeclarationType[] typeDeclarations = self.typeDeclarations();
	ArrayList tlMethods = self.tlMethods();

	if (packageName.getName().length() > 0) {
	    packageName.accept(this);
	    if (importedPackages.length + importedClasses.length > 0) {
		newLine();
	    }
	}
	
	for (int i = 0; i < importedPackages.length ; i++) {
	    if (!importedPackages[i].getName().equals("java/lang")) {
		importedPackages[i].accept(this);
		newLine();
	    }
	}

	for (int i = 0; i < importedClasses.length ; i++) {
	    importedClasses[i].accept(this);
	    newLine();
	}

	for (int i = 0; i < importedGFs.length ; i++) {
	    importedGFs[i].accept(this);
	    newLine();
	}

	for (int i = 0; i < typeDeclarations.length ; i++) {
	    newLine();
	    typeDeclarations[i].accept(this);
	    newLine();
	}
	
	acceptAll( tlMethods );

    }

    // ----------------------------------------------------------------------
    // TYPE DECLARATION
    // ----------------------------------------------------------------------

    /**
     * prints a class declaration
     */
    public void visitClassDeclaration(/*@non_null@*/ JClassDeclaration self ) {
	// !FIXME!comment visit Comments
	long modifiers = self.modifiers();
	String ident = self.ident();
	String superName = self.superName();
	CClassType[] interfaces = self.interfaces();
	JPhylum[] fieldsAndInits = self.fieldsAndInits();
	ArrayList methods = self.methods();
	ArrayList inners = self.inners();

	newLine();
	printClassModifiers(self,modifiers);
	print("class " + ident);

	if (superName != null) {
	    print(" extends " + superName.replace('/', '.'));
	}

	if (interfaces.length != 0) {
	    print(" implements ");
	    for (int i = 0; i < interfaces.length; i++) {
		print((i == 0 ? "" : ", ") + interfaces[i]);
	    }
	}

	print(" ");
	visitClassBody(inners, methods, fieldsAndInits);
    }

    protected void printClassModifiers(/*@non_null@*/JClassDeclaration self, long modifiers) {
	print(modUtil.asString(modifiers));
    }

    /**
     *
     */
    protected void visitClassBody( ArrayList inners,
				 ArrayList methods,
				 JPhylum[] fieldsAndInits ) {
	print("{");
	pos += TAB_SIZE;
	acceptAll( inners );
	acceptAll( methods );

	for (int i = 0; i < fieldsAndInits.length ; i++) {
	    if (!(fieldsAndInits[i] instanceof JFieldDeclarationType)) {
		fieldsAndInits[i].accept(this);
	    }
	}
	for (int i = 0; i < fieldsAndInits.length ; i++) {
	    if (fieldsAndInits[i] instanceof JFieldDeclarationType) {
		fieldsAndInits[i].accept(this);
	    }
	}

	pos -= TAB_SIZE;
	newLine();
	print("}");
    }

    /**
     * prints an interface declaration
     */
    public void visitInterfaceDeclaration(/*@non_null@*/ JInterfaceDeclaration self ) {
	long modifiers = self.modifiers();
	String ident = self.ident();
	CClassType[] interfaces = self.interfaces();
	JPhylum[] fieldsAndInits = self.fieldsAndInits();
	ArrayList methods = self.methods();

	newLine();
	printInterfaceModifiers(self,modifiers);
	print("interface " + ident);

	if (interfaces.length != 0) {
	    print(" extends ");
	    for (int i = 0; i < interfaces.length; i++) {
		print((i == 0 ? "" : ",") + interfaces[i]);
	    }
	}
	visitInterfaceBody(methods,fieldsAndInits,self.inners());
    }
    
    protected void printInterfaceModifiers(/*@non_null@*/JInterfaceDeclaration self, long modifiers) {
	print(modUtil.asString(modifiers));
    }

    protected void visitInterfaceBody(ArrayList methods,
				 JPhylum[] fieldsAndInits,
				 ArrayList inners ) {
	print(" {");
	pos += TAB_SIZE;
	for (int i = 0; i < fieldsAndInits.length; i++) {
	    fieldsAndInits[i].accept(this);
	}

	acceptAll( methods );

	pos -= TAB_SIZE;
	newLine();
	print("}");
    }

    /**
     * prints a generic function anchor
     */
    public void visitGenericFunctionDecl(/*@non_null@*/ MJGenericFunctionDecl self ) {
	// !FIXME! anything to do here?
    }
    
    // ----------------------------------------------------------------------
    // METHODS AND FIELDS
    // ----------------------------------------------------------------------

    /**
     * prints a field declaration
     */
    public void visitFieldDeclaration(/*@non_null@*/ JFieldDeclaration self ) {
	// !FIXME!comment visit Comments
	JVariableDefinition variable = self.variable();
	long modifiers = variable.modifiers();
	CType type = variable.getType();
	String ident = variable.ident();
	JExpression expr = variable.getValue();

	if (ident.indexOf("$") != -1) {
	    return; // dont print generated elements
	}

	newLine();
	print(modUtil.asString(modifiers));
	print(type);
	print(" ");
	print(ident);
	if (expr != null) {
	    print(" = ");
	    expr.accept(this);
	}
	print(";");
    }

    /**
     * prints a method declaration
     */
    public void visitMethodDeclaration(/*@non_null@*/ JMethodDeclaration self ) {
	long modifiers = self.modifiers();
	CType returnType = self.returnType();
	String ident = self.ident();
	JFormalParameter[] parameters = self.parameters();
	CClassType[] exceptions = self.getExceptions();
	JBlock body = self.body();

	newLine();
	print(modUtil.asString(modifiers));
	print(returnType);
	print(" ");
	print(ident);
	print("(");
	int count = 0;

	for (int i = 0; i < parameters.length; i++) {
	    if (count != 0) {
		print(", ");
	    }

	    if (!parameters[i].isGenerated()) {
		parameters[i].accept(this);
		count++;
	    }
	}
	print(")");

	for (int i = 0; i < exceptions.length; i++) {
	    if (i != 0) {
		print(", ");
	    } else {
		print(" throws ");
	    }
	    print(exceptions[i].toString());
	}
   
	print(" ");
	if (body != null) {
	    body.accept(this);
	} else {
	    print(";");
	}
	newLine();
    }

    /**
     * prints an initializer declaration
     */
    public void visitInitializerDeclaration(/*@non_null@*/ JInitializerDeclaration self ) {
	// We do not want to generate these methods in source code
	// since they are the concatenation of the JClassBlocks and
	// JFieldDeclarationType initializers.  Those AST nodes are still
	// in the AST and can be printed in the correct place from the
	// original nodes.
    }

    /**
     * prints an external method declaration
     */
    public void 
	visitTopLevelMethodDeclaration(/*@non_null@*/ MJTopLevelMethodDeclaration self ) 
    {
	long modifiers = self.modifiers();
	CType returnType = self.returnType();
	CType openClassType = self.getOpenClassType();
	String ident = self.ident();
	JFormalParameter[] parameters = self.parameters();
	CClassType[] exceptions = self.getExceptions();
	JBlock body = self.body();

	newLine();
	print(modUtil.asString(modifiers));
	print(returnType);
	print(" ");
	if( self.isExternal() ) {
	    print(openClassType);
	    print(".");
	}
	print(ident);
	print("(");
	int count = 0;

  	for (int i = 0; i < parameters.length; i++) {
	    if (count != 0) {
		print(", ");
	    }

	    if (!parameters[i].isGenerated()) {
		parameters[i].accept(this);
		count++;
	    }
	}
	print(")");

	for (int i = 0; i < exceptions.length; i++) {
	    if (i != 0) {
		print(", ");
	    } else {
		print(" throws ");
	    }
	    print(exceptions[i].toString());
	}
   
	visitTLMethodBody(body);
	newLine();
    }
    
    public void visitTLMethodBody(JBlock body) {
	print(" ");
	if (body != null) {
	    body.accept(this);
	} else {
	    print(";");
	}
    }

    /**
     * prints a constructor declaration
     */
    public void visitConstructorDeclaration(/*@non_null@*/ JConstructorDeclaration self ) {
	long modifiers = self.modifiers();
	String ident = self.ident();
	JFormalParameter[] parameters = self.parameters();
	CClassType[] exceptions = self.getExceptions();
	JBlock body = self.body();

	newLine();
	print(modUtil.asString(modifiers));
	print(ident);
	print("(");
	int count = 0;
	for (int i = 0; i < parameters.length; i++) {
	    if (count != 0) {
		print(", ");
	    }
	    if (!parameters[i].isGenerated()) {
		parameters[i].accept(this);
		count++;
	    }
	}
	print(")");

	for (int i = 0; i < exceptions.length; i++) {
	    if (i != 0) {
		print(", ");
	    } else {
		print(" throws ");
	    }
	    print(exceptions[i].toString());
	}
	print(" ");
        body.accept(this);
	newLine();
    }

    // ----------------------------------------------------------------------
    // STATEMENT
    // ----------------------------------------------------------------------

    /**
     * prints a while statement
     */
    public void visitWhileStatement(/*@non_null@*/ JWhileStatement self ) { 
	JExpression cond = self.cond();
	JStatement body = self.body();

	print("while (");
	cond.accept(this);
	print(") ");

	body.accept(this);
	visitComments(self.getComments());
    }

    /**
     * prints a variable declaration statement
     */
    public void 
	visitVariableDeclarationStatement(/*@non_null@*/JVariableDeclarationStatement self) 
    {
	JVariableDefinition[] vars = self.getVars();

        //@ assume vars.length > 0;
        printVariableDefinition(vars[0], true); // with type and modifiers

	for (int i = 1; i < vars.length; i++) {
            print(", ");
            printVariableDefinition(vars[i], false); // no type and modifier
	}
	print(";");
	visitComments(self.getComments());
    }
    
    /**
     * Prints the given variable definition, <code>self</code>. If the
     * argument, <code>typeAndMod</code> is true, the type and
     * modifiers are also printed; otherwise, they are not printed.
     */
    private void printVariableDefinition(/*@non_null@*/ JVariableDefinition self,
                                          boolean typeAndMod ) {
        if (typeAndMod) {
            long modifiers = self.modifiers();
            CType type = self.getType();
            print(modUtil.asString(modifiers));
            print(toString(type));
            print(" ");
        }

	String ident = self.ident();
	JExpression expr = self.expr();
	print(ident);
	if (expr != null) {
	    print(" = ");
	    expr.accept(this);
	}
    }

    /**
     * prints a variable declaration statement
     */
    public void visitVariableDefinition(/*@non_null@*/ JVariableDefinition self ) {
	long modifiers = self.modifiers();
	CType type = self.getType();
	String ident = self.ident();
	JExpression expr = self.expr();

	print(modUtil.asString(modifiers));
	print(toString(type));
	print(" ");
	print(ident);
	if (expr != null) {
	    print(" = ");
	    expr.accept(this);
	}
	print(";");
    }

    /**
     * prints a try-catch statement
     */
    public void visitTryCatchStatement(/*@non_null@*/ JTryCatchStatement self ) {
	JBlock tryClause = self.tryClause();
	JCatchClause[] catchClauses = self.catchClauses();

	print("try ");
	tryClause.accept(this);
	for (int i = 0; i < catchClauses.length; i++) {
	    catchClauses[i].accept(this);
	}
    }

    /**
     * prints a try-finally statement
     */
    public void visitTryFinallyStatement(/*@non_null@*/ JTryFinallyStatement self ) {
	JBlock tryClause = self.tryClause();
	JBlock finallyClause = self.finallyClause();

	print("try ");
	tryClause.accept(this);
	if (finallyClause != null) {
	    print(" finally ");
	    finallyClause.accept(this);
	}
    }

    /**
     * prints a throw statement
     */
    public void visitThrowStatement(/*@non_null@*/ JThrowStatement self ) {
	JExpression expr = self.expr();
	print("throw ");
	expr.accept(this);
	print(";");
	visitComments(self.getComments());
    }

    /**
     * prints a synchronized statement
     */
    public void visitSynchronizedStatement(/*@non_null@*/ JSynchronizedStatement self ) {
	JExpression cond = self.cond();
	JBlock body = self.body();
	print("synchronized (");
	cond.accept(this);
	print(") ");
	body.accept(this);
	visitComments(self.getComments());
    }

    /**
     * prints a switch statement
     */
    public void visitSwitchStatement(/*@non_null@*/ JSwitchStatement self ) {
	JExpression expr = self.expr();
	JSwitchGroup[] groups = self.groups();
	print("switch (");
	expr.accept(this);
	print(") {");
	for (int i = 0; i < groups.length; i++) {
	    groups[i].accept(this);
	}
	newLine();
	print("}");
	visitComments(self.getComments());
    }

    /**
     * prints a return statement
     */
    public void visitReturnStatement(/*@non_null@*/ JReturnStatement self ) {
	JExpression expr = self.expr();
	print("return");
	if (expr != null) {
	    print(" ");
	    expr.accept(this);
	}
	print(";");
	visitComments(self.getComments());
    }

    /**
     * prints a labeled statement
     */
    public void visitLabeledStatement(/*@non_null@*/ JLabeledStatement self ) {
	String label = self.getLabel();
	JStatement stmt = self.stmt();
	print(label + ":");
	stmt.accept(this);
	visitComments(self.getComments());
    }

    /**
     * prints a if statement
     */
    public void visitIfStatement(/*@non_null@*/ JIfStatement self ) { 
	JExpression cond = self.cond();
	JStatement thenClause = self.thenClause();
	JStatement elseClause = self.elseClause();

	print("if (");
	cond.accept(this);
	print(") ");
	pos += thenClause instanceof JBlock ? 0 : TAB_SIZE;
	thenClause.accept(this);
	pos -= thenClause instanceof JBlock ? 0 : TAB_SIZE;
	if (elseClause != null) {
	    if ((elseClause instanceof JBlock) 
		|| (elseClause instanceof JIfStatement)) {
		print(" ");
	    } else {
		newLine();
	    }
	    print("else ");
	    pos += elseClause instanceof JBlock 
		|| elseClause instanceof JIfStatement ? 0 : TAB_SIZE;
	    elseClause.accept(this);
	    pos -= elseClause instanceof JBlock 
		|| elseClause instanceof JIfStatement ? 0 : TAB_SIZE;
	}
    }

    /**
     * prints a for statement
     */
    public void visitForStatement(/*@non_null@*/ JForStatement self ) { 
	JStatement init = self.init();
	JExpression cond = self.cond();
	JStatement incr = self.incr();
	JStatement body = self.body();

	print("for (");
	forInit = true;
	if (init != null) {
	    init.accept(this);
	} else {
	    print(";");
	}
	forInit = false;

	print(" ");
	if (cond != null) {
	    cond.accept(this);
	}
	print("; ");

	if (incr != null) {
	    incr.accept(this);
	}
	print(") ");

	print("{");
	pos += TAB_SIZE;
	body.accept(this);
	pos -= TAB_SIZE;
	newLine();
	print("}");
    }

    /**
     * prints a compound statement
     */
    public void visitCompoundStatement(/*@non_null@*/ JCompoundStatement self ) {
	JStatement[] body = self.body();
	visitCompoundStatement(body);
    }

    /**
     * prints a compound statement
     */
    public void visitCompoundStatement(JStatement[] body) {
	for (int i = 0; i < body.length; i++) {
	    if (body[i] instanceof JIfStatement && 
		i < body.length - 1 &&
		!(body[i + 1] instanceof JReturnStatement)) {
		newLine();
	    }
	    if (body[i] instanceof JReturnStatement && i > 0) {
		newLine();
	    }

	    newLine();
	    body[i].accept(this);

	    if (body[i] instanceof JVariableDeclarationStatement && 
		i < body.length - 1 &&
		!(body[i + 1] instanceof JVariableDeclarationStatement)) {
		newLine();
	    }
	}
    }

    /**
     * prints an expression statement
     */
    public void visitExpressionStatement(/*@non_null@*/ JExpressionStatement self ) {
	JavaStyleComment[] comments = self.getComments();
	if(comments != null){
	    visitComments(comments);
	}

	JExpression expr = self.expr();
	expr.accept(this);
	if (!forInit) {
	    print(";");
	}
    }

    /**
     * prints an expression list statement
     */
    public void visitExpressionListStatement(/*@non_null@*/ JExpressionListStatement self ) {
	JavaStyleComment[] comments = self.getComments();
	if(comments != null){
	    visitComments(comments);
	}

	JExpression[] exprs = self.exprs();
	for (int i = 0; i < exprs.length; i++) {
	    if (i != 0) {
		print(", ");
	    }
	    exprs[i].accept(this);
	}
	if (forInit) {
	    print(";");
	}
    }

    /**
     * prints a empty statement
     */
    public void visitEmptyStatement(/*@non_null@*/ JEmptyStatement self ) {
	JavaStyleComment[] comments = self.getComments();
	if(comments != null){
	    visitComments(comments);
	}

	newLine();
	print(";");
    }

    /**
     * prints a do statement
     */
    public void visitDoStatement(/*@non_null@*/ JDoStatement self ) {
	JavaStyleComment[] comments = self.getComments();
	if(comments !=null){ 
	    visitComments(comments);
	}

	JExpression cond = self.cond();
	JStatement body = self.body();
	newLine();
	print("do ");
	body.accept(this);
	print("");
	print("while (");
	cond.accept(this);
	print(");");
    }

    /**
     * prints a continue statement
     */
    public void visitContinueStatement(/*@non_null@*/ JContinueStatement self ) {
	String label = self.label();
	newLine();
	print("continue");
	if (label != null) {
	    print(" " + label);
	}
	print(";");
    }

    /**
     * prints a break statement
     */
    public void visitBreakStatement(/*@non_null@*/ JBreakStatement self ) {
	JavaStyleComment[] comments = self.getComments();
        if(comments!=null){
	    visitComments(comments);
	}

	String label = self.label();
	newLine();
	print("break");
	if (label != null) {
	    print(" " + label);
	}
	print(";");
    }

    /**
     * Prints a assert statement
     */
    public void visitAssertStatement(/*@non_null@*/ JAssertStatement self ) {
	JavaStyleComment[] comments = self.getComments();
	if(comments!=null){
	    visitComments(comments);
	}

	print("assert ");
        self.predicate().accept(this);
        if (self.hasThrowMessage()) {
	    print(" : ");
            self.throwMessage().accept(this);
	}
	print(";");
    }

    /**
     * prints an expression statement
     */
    public void visitBlockStatement(/*@non_null@*/ JBlock self ) {
	JStatement[] body = self.body();
	JavaStyleComment[] comments = self.getComments();
	print("{");
	pos += TAB_SIZE;
	visitCompoundStatement(body);
	if (comments != null) {
	    visitComments(comments);
	}
	pos -= TAB_SIZE;
	newLine();
	print("}");
    }

    /**
     * prints an expression statement
     */
    public void visitConstructorBlock(/*@non_null@*/ JConstructorBlock self ) {
	JStatement[] body = self.body();
	JExpression explicitSuper = self.explicitSuper();
	JStatement blockCall = self.blockCall();
	JavaStyleComment[] comments = self.getComments();
	print("{");
	pos += TAB_SIZE;
	if( explicitSuper != null ) {
	    explicitSuper.accept( this );
	}
	if( blockCall != null ) {
	    blockCall.accept( this );
	}
        if (body != null) {
            visitCompoundStatement(body);
        }
	if (comments != null) {
	    visitComments(comments);
	}
	pos -= TAB_SIZE;
	newLine();
	print("}");
    }

    /**
     * prints a class block (initializer)
     */
    public void visitClassBlock(/*@non_null@*/ JClassBlock self ) {
	newLine();
	if (self.isStaticInitializer()) {
	    print("static ");
	}
	visitBlockStatement( self );
    }

    /**
     * prints a type declaration statement
     */
    public void visitTypeDeclarationStatement(/*@non_null@*/ JTypeDeclarationStatement self) {
	JTypeDeclarationType decl = self.decl();

        boolean isStatic = hasFlag(decl.modifiers(), ACC_STATIC);

        // A local class should not have a "static" modifier. If one
        // has been added during type-checking (see
        // TypeDeclarationStatement.typecheck(), temporarily remove it
        // while pretty-printing and restore it later.
        if (isStatic) {
            decl.unsetStatic();
        }
	decl.accept(this);
        if (isStatic) {
            decl.setStatic();
        }

	visitComments(self.getComments());
    }

    // ----------------------------------------------------------------------
    // EXPRESSION
    // ----------------------------------------------------------------------

    /**
     * prints an unary expression
     */
    public void visitUnaryExpression(/*@non_null@*/  JUnaryExpression self ) {
        // The tailing space after the operator (e.g., "+ ") is
        // crucial not to print expressions like "+ ++x" as "+++x",
        // where the first is legal Java expression while the second
        // is not. Thank to Joe Kiniry for observing this.
	int oper = self.oper();
	JExpression expr = self.expr();
	switch (oper) {
	case OPE_PLUS:
	    print("+ ");
	    expr.accept(this);
	    break;

	case OPE_MINUS:
	    print("- ");
	    expr.accept(this);
	    break;

	case OPE_BNOT:
	    print("~ ");
	    expr.accept(this);
	    break;

	case OPE_LNOT:
	    print("! ");
	    expr.accept(this);
	    break;

	default:
	    fail();
	}
    }

    /**
     * prints a type name expression
     */
    public void visitTypeNameExpression(/*@non_null@*/  JTypeNameExpression self ) {
        // if available, print as it appeared in the source file.
        if (self.sourceName() != null) {
            self.sourceName().accept(this);
        } else {
            print(self.getType());
        }
    }

    /**
     * prints a this expression
     */
    public void visitThisExpression(/*@non_null@*/  JThisExpression self ) {
	JExpression prefix = self.prefix();
	if (prefix != null) {
	    prefix.accept(this);
	    print(".this");
	} else {
	    print("this");
	}
    }

    /**
     * prints a super expression
     */
    public void visitSuperExpression(/*@non_null@*/ JSuperExpression self ) {
	print("super");
    }

    /**
     * prints a shift expression
     */
    public void visitShiftExpression(/*@non_null@*/ JShiftExpression self ) {
	int oper = self.oper();
	JExpression left = self.left();
	JExpression right = self.right();
	left.accept(this);
	if (oper == OPE_SL) {
	    print(" << ");
	} else if (oper == OPE_SR) {
	    print(" >> ");
	} else {
	    print(" >>> ");
	}
	right.accept(this);
    }

    /**
     * prints a shift expressiona
     */
    public void visitRelationalExpression(/*@non_null@*/ JRelationalExpression self ) {
	int oper = self.oper();
	JExpression left = self.left();
	JExpression right = self.right();
	left.accept(this);
	switch (oper) {
	case OPE_LT:
	    print(" < ");
	    break;
	case OPE_LE:
	    print(" <= ");
	    break;
	case OPE_GT:
	    print(" > ");
	    break;
	case OPE_GE:
	    print(" >= ");
	    break;
	default:
	    fail();
	}
	right.accept(this);
    }

    /**
     * prints a prefix expression
     */
    public void visitPrefixExpression(/*@non_null@*/ JPrefixExpression self ) {
	int oper = self.oper();
	JExpression expr = self.expr();
	if (oper == OPE_PREINC) {
	    print("++");
	} else {
	    print("--");
	}
	expr.accept(this);
    }

    /**
     * prints a postfix expression
     */
    public void visitPostfixExpression(/*@non_null@*/ JPostfixExpression self ) {
	int oper = self.oper();
	JExpression expr = self.expr();
	expr.accept(this);
	if (oper == OPE_POSTINC) {
	    print("++");
	} else {
	    print("--");
	}
    }

    /**
     * prints a parenthesed expression
     */
    public void visitParenthesedExpression(/*@non_null@*/ JParenthesedExpression self ) {
	JExpression expr = self.expr();
	print("(");
	expr.accept(this);
	print(")");
    }

    /**
     * prints an array allocator expression
     */
    public void visitNewArrayExpression(/*@non_null@*/ JNewArrayExpression self ) {
	CType type = self.getType();
	JArrayDimsAndInits dims = self.dims();
	if (type instanceof CArrayType) {
		// WMD TODO check whether to use toJMLString.
    	String umod = CTopLevel.getCompiler().universeDynChecks() ? ((CClassType) type).getCUniverse().toMJString() + " " : "";
	    print("new " + umod +
	    		toString(((CArrayType)type).getBaseType()));
	} else {
	    print("new " + toString(type));
	}
	dims.accept(this);
    }

    /**
     * prints a name expression
     */
    public void visitNameExpression(/*@non_null@*/ JNameExpression self ) { 
	JExpression prefix = self.sourcePrefix();
	String ident = self.sourceIdent();

	if (prefix != null) {
		if( CTopLevel.getCompiler().universeChecks() && self.getCUniverse() != null ) {
		  print(self.getCUniverse().toMJString() + " ");
		}
	    prefix.accept(this);
	    print('.');
	}
	print(ident);
    }

    /**
     * prints a binary expression with the given operator
     */
    protected void visitBinaryExpression(/*@non_null@*/ JBinaryExpression self, String oper ) {
	JExpression left = self.left();
	JExpression right = self.right();
	left.accept(this);
	print(" ");
	print(oper);
	print(" ");
	right.accept(this);
    }

    /**
     * visits an add expression
     */
    public void visitAddExpression(/*@non_null@*/ JAddExpression self ) {
	visitBinaryExpression( self, "+" );
    }

    /**
     * visits a boolean AND expression
     */
    public void 
	visitConditionalAndExpression(/*@non_null@*/ JConditionalAndExpression self ) 
    {
	visitBinaryExpression( self, "&&" );
    }

     /**
     * visits a boolean OR expression
     */
    public void 
	visitConditionalOrExpression(/*@non_null@*/ JConditionalOrExpression self ) 
    {
	visitBinaryExpression( self, "||" );
    }

    /**
     * visits a divide expression
     */
    public void visitDivideExpression(/*@non_null@*/ JDivideExpression self ) {
	visitBinaryExpression( self, "/" );
    }

    /**
     * visits a minus expression
     */
    public void visitMinusExpression(/*@non_null@*/ JMinusExpression self ) {
	visitBinaryExpression( self, "-" );
    }

    /**
     * visits a modulo division expression
     */
    public void visitModuloExpression(/*@non_null@*/ JModuloExpression self ) {
	visitBinaryExpression( self, "%" );
    }

    /**
     * visits a multiplication expression
     */
    public void visitMultExpression(/*@non_null@*/ JMultExpression self ) {
	visitBinaryExpression( self, "*" );
    }

   /**
     * prints a method call expression
     */
    public void visitMethodCallExpression(/*@non_null@*/ JMethodCallExpression self ) {
	String ident = self.ident();
	JExpression[] args = self.args();

        // ignore the generated methods
	if (ident != null && ident.equals(JAV_INIT)) {
	    return; 
	}

	JExpression prefix = self.prefix();
        JNameExpression sourceName = self.sourceName();

        // if possible, print as it appeared in the source file.
        if (sourceName != null) {
            sourceName.accept(this);
        } else {
            if (prefix != null) {
                prefix.accept(this);
                print('.');
            }
            print(ident);
        }

	print("(");
	visitArgs(args);
	print(")");
    }

    /**
     * prints a local variable expression
     */
    public void visitLocalVariableExpression(/*@non_null@*/ JLocalVariableExpression self ) {
	String ident = self.ident();
	print(ident);
    }

    /**
     * prints an instanceof expression
     */
    public void visitInstanceofExpression(/*@non_null@*/ JInstanceofExpression self ) {
	JExpression expr = self.expr();
	CType dest = self.dest();
	expr.accept(this);
	print(" instanceof ");
	print(dest);
    }

    /**
     * prints an equality expression
     */
    public void visitEqualityExpression(/*@non_null@*/ JEqualityExpression self ) {
	int oper = self.oper();
	JExpression left = self.left();
	JExpression right = self.right();
	left.accept(this);
	if (oper == OPE_EQ) {
	    print(" == ");
	} else {
	    print(" != ");
	}
	right.accept(this);
    }

    /**
     * prints a conditional expression
     */
    public void visitConditionalExpression(/*@non_null@*/ JConditionalExpression self ) {
	JExpression cond = self.cond();
	JExpression left = self.left();
	JExpression right = self.right();
	cond.accept(this);
	print(" ? ");
	left.accept(this);
	print(" : ");
	right.accept(this);
    }

    /**
     * prints a compound expression
     */
    public void 
	visitCompoundAssignmentExpression(/*@non_null@*/JCompoundAssignmentExpression self) {
	int oper = self.oper();
	JExpression left = self.left();
	JExpression right = self.right();
	left.accept(this);
	switch (oper) {
	case OPE_STAR:
	    print(" *= ");
	    break;
	case OPE_SLASH:
	    print(" /= ");
	    break;
	case OPE_PERCENT:
	    print(" %= ");
	    break;
	case OPE_PLUS:
	    print(" += ");
	    break;
	case OPE_MINUS:
	    print(" -= ");
	    break;
	case OPE_SL:
	    print(" <<= ");
	    break;
	case OPE_SR:
	    print(" >>= ");
	    break;
	case OPE_BSR:
	    print(" >>>= ");
	    break;
	case OPE_BAND:
	    print(" &= ");
	    break;
	case OPE_BXOR:
	    print(" ^= ");
	    break;
	case OPE_BOR:
	    print(" |= ");
	    break;
	}
	right.accept(this);
    }

    /**
     * prints a field expression
     */
    public void visitFieldExpression(/*@non_null@*/ JClassFieldExpression self ) {
        // if available, print as it appeared in the source file.
        if (self.sourceName() != null) {
            self.sourceName().accept(this);
            return;
        }

	JExpression prefix = self.prefix();
	String ident = self.ident();
	if (ident.equals(JAV_OUTER_THIS) // don't generate generated fields
            && !self.prefixWasBlank()) {
	    print(prefix.getApparentType().getCClass().owner().getType()
                  + ".this");
	    return;
	}
	int index = ident.indexOf("val$");
	if (index != -1) {
            // skip the generated "val$"
	    print(ident.substring(index + 4)); // local var
	} else {
            if (!self.prefixWasBlank()) {
                prefix.accept(this);
                print(".");
            }
	    print(ident);
	}
    }

    /**
     * prints a class expression
     */
    public void visitClassExpression(/*@non_null@*/ JClassExpression self ) {
	CType prefixType = self.prefixType();
	print(prefixType);
	print(".class");
    }

    /**
     * prints a cast expression
     */
    public void visitCastExpression(/*@non_null@*/ JCastExpression self ) {
	JExpression expr = self.expr();
	CType dest = self.getType();
	print("(");
	print(dest);
	print(")");
	expr.accept(this);
    }

    /**
     * prints a cast expression
     */
    public void visitUnaryPromoteExpression(/*@non_null@*/ JUnaryPromote self ) {
	JExpression expr = self.expr();
        CType type = self.getType();

        print("(");
        print("(");
        print(type);
        print(")(");
        expr.accept(this);
        print("))");
    }

    /**
     * prints a compound assignment expression
     */
    public void visitBitwiseExpression(/*@non_null@*/ JBitwiseExpression self ) {
	int oper = self.oper();
	JExpression left = self.left();
	JExpression right = self.right();
	left.accept(this);
	switch (oper) {
	case OPE_BAND:
	    print(" & ");
	    break;
	case OPE_BOR:
	    print(" | ");
	    break;
	case OPE_BXOR:
	    print(" ^ ");
	    break;
	default:
	    fail();
	}
	right.accept(this);
    }

    /**
     * prints an assignment expression
     */
    public void visitAssignmentExpression(/*@non_null@*/ JAssignmentExpression self ) {
	JExpression left = self.left();
	JExpression right = self.right();
	if ((left instanceof JClassFieldExpression) &&
	    ((JClassFieldExpression)left).getField()
	    .ident().equals(Constants.JAV_OUTER_THIS)) {
	    return;
	}

	left.accept(this);
	print(" = ");
	right.accept(this);
    }

    /**
     * prints an array length expression
     */
    public void visitArrayLengthExpression(/*@non_null@*/ JArrayLengthExpression self ) {
	JExpression prefix = self.prefix();
	prefix.accept(this);
	print(".length");
    }

    /**
     * prints an array length expression
     */
    public void visitArrayAccessExpression(/*@non_null@*/ JArrayAccessExpression self ) {
	JExpression prefix = self.prefix();
	JExpression accessor = self.accessor();
	prefix.accept(this);
	print("[");
	accessor.accept(this);
	print("]");
    }

    /**
     * prints an array length expression
     */
    protected void visitComments(JavaStyleComment[] comments) {
	if (comments == null) return;
	for (int i = 0; i < comments.length; i++) {
	    if (comments[i] != null) {
		visitComment(comments[i]);
	    }
	}
    }

    /**
     * prints an array length expression
     */
    protected void visitComment(/*@non_null@*/JavaStyleComment comment) {
	StringTokenizer	tok = new StringTokenizer(comment.getText(), "\n");

	if (comment.hadSpaceBefore()) {
	    newLine();
	}

	if (comment.isLineComment()) {
	    print("// " + tok.nextToken().trim());
	    p.println();
	} else {
	    if (p.getLine() > 0) {
		if (!nl) {
		    newLine();
		}
		newLine();
	    }
	    print("/*");
	    while (tok.hasMoreTokens()){
		String comm = tok.nextToken().trim();
		if (comm.startsWith("*")) {
		    comm = comm.substring(1).trim();
		}
		if (tok.hasMoreTokens() || comm.length() > 0) {
		    newLine();
		    print(" * " + comm);
		}
	    }
	    newLine();
	    print(" */");
	    newLine();
	}

	if (comment.hadSpaceAfter()) {
	    newLine();
	}
    }

    /**
     * prints a javadoc comment
     */
    protected void visitJavadoc(/*@non_null@*/JavadocComment comment) {
	StringTokenizer	tok = new StringTokenizer(comment.text(), "\n");
	boolean		isFirst = true;

	if (!nl) {
	    newLine();
	}
	newLine();
	print("/**");
	while (tok.hasMoreTokens()) {
	    String	text = tok.nextToken().trim();
	    String	type = null;
	    boolean	param = false;
	    int	idx = text.indexOf("@param");
	    if (idx >= 0) {
		type = "@param";
		param = true;
	    }
	    if (idx < 0) {
		idx = text.indexOf("@exception");
		if (idx >= 0) {
		    type = "@exception";
		    param = true;
		}
	    }
	    if (idx < 0) {
		idx = text.indexOf("@exception");
		if (idx >= 0) {
		    type = "@exception";
		    param = true;
		}
	    }
	    if (idx < 0) {
		idx = text.indexOf("@author");
		if (idx >= 0) {
		    type = "@author";
		}
	    }
	    if (idx < 0) {
		idx = text.indexOf("@see");
		if (idx >= 0) {
		    type = "@see";
		}
	    }
	    if (idx < 0) {
		idx = text.indexOf("@version");
		if (idx >= 0) {
		    type = "@version";
		}
	    }
	    if (idx < 0) {
		idx = text.indexOf("@return");
		if (idx >= 0) {
		    type = "@return";
		}
	    }
	    if (idx < 0) {
		idx = text.indexOf("@deprecated");
		if (idx >= 0) {
		    type = "@deprecated";
		}
	    }
	    if (idx >= 0) {
		newLine();
		isFirst = false;
		if (param) {
		    text = text.substring(idx + type.length()).trim();
		    idx = Math.min(text.indexOf(" ") == -1 ? 1000 : text.indexOf(" "),
				   text.indexOf("\t") == -1 ? 1000 : text.indexOf("\t"));
		    String	before = text.substring(0, idx);
		    print(" * " + type);
		    pos += 12;
		    print(before);
		    pos += 20;
		    print(text.substring(idx).trim());
		    pos -= 20;
		    pos -= 12;
		} else {
		    text = text.substring(idx + type.length()).trim();
		    print(" * " + type);
		    pos += 12;
		    print(text);
		    pos -= 12;
		}
	    } else {
		text = text.substring(text.indexOf("*") + 1);
		if (tok.hasMoreTokens() || text.length() > 0) {
		    newLine();
		    print(" * ");
		    pos += isFirst ? 0 : 32;
		    print(text.trim());
		    pos -= isFirst ? 0 : 32;
		}
	    }
	}
	newLine();
	print(" */");
    }

    // ----------------------------------------------------------------------
    // UTILS
    // ----------------------------------------------------------------------

    /**
     * prints a switch label
     */
    public void visitSwitchLabel(/*@non_null@*/ JSwitchLabel self ) {
	JExpression expr = self.expr();
	newLine();
	if (expr != null) {
	    print("case ");
	    expr.accept(this);
	    print(": ");
	} else {
	    print("default: ");
	}
    }

    /**
     * prints a switch group
     */
    public void visitSwitchGroup(/*@non_null@*/ JSwitchGroup self ) {
	JSwitchLabel[] labels = self.labels();
	JStatement[] stmts = self.getStatements();
	for (int i = 0; i < labels.length; i++) {
	    labels[i].accept(this);
	}
	pos += TAB_SIZE;
	for (int i = 0; i < stmts.length; i++) {
	    newLine();
	    stmts[i].accept(this);
	}
	pos -= TAB_SIZE;
    }

    /**
     * prints a catch clause
     */
    public void visitCatchClause(/*@non_null@*/ JCatchClause self ) {
	JFormalParameter exception = self.exception();
	JBlock body = self.body();
	print(" catch (");
	exception.accept(this);
	print(") ");
	body.accept(this);
    }

    /**
     * prints a boolean literal
     */
    public void visitBooleanLiteral(/*@non_null@*/ JBooleanLiteral self ) {
	boolean value = self.booleanValue();
	print(value);
    }

    /**
     * prints a character literal
     */
    public void visitCharLiteral(/*@non_null@*/ JCharLiteral self ) {
	Character chValue = (Character) self.getValue();
	char value;
	if( chValue != null) {
	    value = chValue.charValue();
	} else {
	    value = self.image().charAt(0); // !!! false
	}
	switch (value) {
	case '\b':
	    print("'\\b'");
	    break;
	case '\r':
	    print("'\\r'");
	    break;
	case '\t':
	    print("'\\t'");
	    break;
	case '\n':
	    print("'\\n'");
	    break;
	case '\f':
	    print("'\\f'");
	    break;
	case '\\':
	    print("'\\\\'");
	    break;
	case '\'':
	    print("'\\''");
	    break;
	case '\"':
	    print("'\\\"'");
	    break;
	default:
	    print("'" + value + "'");
	}
    }

    /**
     * prints an ordinal literal
     */
    public void visitOrdinalLiteral(/*@non_null@*/ JOrdinalLiteral self ) {
	Number value = self.numberValue();
	CType type = self.getType();
	assertTrue(value != null);

	if (type == CStdType.Byte) {
	    visitByteLiteral(value.byteValue());
	} else if (type == CStdType.Integer) {
	    visitIntLiteral(value.intValue());
	} else if (type == CStdType.Long) {
	    visitLongLiteral(value.longValue());
	} else if (type == CStdType.Short) {
	    visitShortLiteral(value.shortValue());
	} else {
	    fail();
	}
    }

    /**
     * prints a byte literal
     */
    protected void visitByteLiteral( byte value ) {
	print("(byte)" + value);
    }

    /**
     * prints a int literal
     */
    protected void visitIntLiteral(int value) {
	print(value);
    }

    /**
     * prints a long literal
     */
    protected void visitLongLiteral(long value) {
	print(value + "L");
    }

    /**
     * prints a short literal
     */
    protected void visitShortLiteral(short value) {
	print("(short)" + value);
    }

    /**
     * prints a real literal
     */
    public void visitRealLiteral(/*@non_null@*/ JRealLiteral self ) {
	Number value = self.numberValue();
	CType type = self.getType();
	assertTrue(value != null);

	if( type == CStdType.Double) {
	    visitDoubleLiteral(value.doubleValue());
	} else if( type == CStdType.Float) {
	    visitFloatLiteral(value.floatValue());
	} else {
	    fail();
	}
    }

    /**
     * prints a double literal
     */
    protected void visitDoubleLiteral(double value) {
	print("(double)" + value);
    }

    /**
     * prints a float literal
     */
    protected void visitFloatLiteral(float value) {
	print(value + "F");
    }

    /**
     * prints a string literal
     */
    public void visitStringLiteral(/*@non_null@*/ JStringLiteral self) {
	String value = self.stringValue();
	StringBuffer s = new StringBuffer();
	for (int i = 0; i < value.length(); i++) {
	    char c = value.charAt(i);
	    switch (c) {
	    case '\n' : s.append("\\n"); break;
	    case '\r' : s.append("\\r"); break;
	    case '\t' : s.append("\\t"); break;
	    case '\b' : s.append("\\b"); break;
	    case '\f' : s.append("\\f"); break;
	    case '\"' : s.append("\\\""); break;
	    case '\'' : s.append("\\\'"); break;
	    case '\\' : s.append("\\\\"); break;
	    default:
		s.append(c);
	    }
	}
	value = s.toString();
	print('"' + value + '"');
    }

    /**
     * prints a null literal
     */
    public void visitNullLiteral(/*@non_null@*/ JNullLiteral self ) {
	print("null");
    }

    /**
     * prints a package name statement
     */
    public void visitPackageName(/*@non_null@*/ JPackageName self ) {
	String name = self.getName();
	print("package " + name.replace('/', '.') + ";");
	newLine();
    }

    /**
     * prints a package import declaration
     */
    public void visitPackageImport(/*@non_null@*/ JPackageImport self ) {
	String name = self.getName();
	print("import " + name.replace('/', '.') + ".*;");
    }

    /**
     * prints a class import declaration
     */
    public void visitClassOrGFImport(/*@non_null@*/ JClassOrGFImport self ) {
	String name = self.getName();
	print("import " + name.replace('/', '.') + ";");
    }

    /**
     * prints a formal parameter
     */
    public void visitFormalParameters(/*@non_null@*/ JFormalParameter self ) {
	print(self.toString());
    }

    /**
     * prints an explicit constructor invocation
     */
    public void 
	visitExplicitConstructorInvocation(/*@non_null@*/JExplicitConstructorInvocation self)
    {
	JExpression prefix = self.prefix();
	String ident = self.ident();
	JExpression[] params = self.params();
	newLine();

	// print prefix if it's not a synthesized one (i.e., JAV_OUTER_THIS)
	if (prefix != null && !self.isPrefixSynthesized()) {
	    prefix.accept(this);
	    print(".");
	}

	print(ident);

	print("(");
	visitArgs(params);
	print(");");
    }

    /**
     * prints an object allocator expression
     */
    public void visitNewObjectExpression(/*@non_null@*/ JNewObjectExpression self ) {
	CType type = self.getType();
	JExpression[] params = self.params();
	print("new " + toString(type) + "(");

	visitArgs(params);

	print(")");
    }

    /**
     * prints an object allocator expression for an anonymous class
     */
    public void 
	visitNewAnonymousClassExpression(/*@non_null@*/ JNewAnonymousClassExpression self ) 
    {
	String typeName = self.originalTypeName();
	JExpression[] params = self.params();
	JClassDeclarationType decl = self.decl();
	JExpression prefix = self.thisExpr();
	if (prefix != null) {
	    prefix.accept( this );
	    print(".");
	}
	print("new " + typeName + "(");

	visitArgs(params);

	print(")");
	if (decl != null) {
	    print(" ");
	    visitClassBody( decl.inners(),
			    decl.methods(),
			    decl.fieldsAndInits() );
	}
    }

    /**
     * prints an array initializer expression
     */
    public void visitArrayInitializer(/*@non_null@*/ JArrayInitializer self ) {
	JExpression[] elems = self.elems();
	newLine();
	print("{");
	for (int i = 0; i < elems.length; i++) {
	    if (i != 0) {
		print(", ");
	    }
	    elems[i].accept(this);
	}
	print("}");
    }

    /**
     * prints an array dimension and initialization expression
     */
    public void visitArrayDimsAndInit(/*@non_null@*/ JArrayDimsAndInits self ) {
	JExpression[] dims = self.dims();
	JArrayInitializer init = self.init();
	for (int i = 0; i < dims.length; i++) {
	    print("[");		   // note, the brackets are always printed...
	    if (dims[i] != null) { // ...even if some do not contain...
		dims[i].accept(this); // ...expressions, i.e., this is legal:
	    }                         // a = new int[][x][][y];
	    print("]");
	}
	if (init != null) {
	    init.accept(this);
	}
    }

    /**
     * prints a warn expression
     */
    public void visitWarnExpression(/*@non_null@*/ MJWarnExpression self ) {
	print(self.operatorName() + "(");
	self.expr().accept(this);
	print(")");
    }
    
    /**
     * prints a math mode expression
     */
    public void visitMathModeExpression(/*@non_null@*/ MJMathModeExpression self ) {
	print(self.operatorName() + "(");
	self.expr().accept(this);
	print(")");
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------

    /** Returns the string for the given type. E.g., if the type
     * represents a local local class such as <code>n1.n2.1.n3</code>, then
     * returned is the string <code>"n3"</code>.
     */
    protected String toString(/*@non_null@*/CType type) {
        if (!(type instanceof CClassType)) {
            return type.toString();
        }

        // fix for local class names; e.g., n1.n2.1.n3 is translated
        // into n3.
        StringTokenizer tokens = new StringTokenizer(type.toString(), ".");
        String result = "";
        
        while (tokens.hasMoreTokens()) {
            String name = tokens.nextToken();
            boolean isNumeric = true;
            for (int i = 0; i < name.length(); i++) {
                if (!Character.isDigit(name.charAt(i))) {
                    isNumeric = false;
                    break;
                }
            }
            result = isNumeric ? "" : (result + "." + name);
        }
        
        // remove the starting "."
        result = result.substring(1);
        
        // RZU - put universe modifier at the beginning if dynamic universe runtime checks 
        //       are enabled
    	String universeModifier = CTopLevel.getCompiler().universeDynChecks() ? ((CClassType) type).getCUniverseMJString() + " " : "";
        
    	return universeModifier + result;
    }

    /**
     * prints a list of arguments
     */
    protected void visitArgs(JExpression[] args) {
	if (args != null) {
	    for (int i = 0; i < args.length; i++) {
		if (i != 0) {
		    print(", ");
		}
		args[i].accept(this);
	    }
	}
    }

    //@ assignable \nothing;
    protected void newLine() {
	p.println();
    }

    //@ assignable \nothing;
    protected void print(/*@non_null@*/Object s) {
	print(s.toString());
    }

    //@ assignable \nothing;
    protected void print(String s) {
        p.setPos(pos);
        p.print(s);
    }

    //@ assignable \nothing;
    protected void print(CType t) {
        print(toString(t));
    }

    //@ assignable \nothing;
    protected void print(boolean s) {
	print("" + s);
    }

    //@ assignable \nothing;
    protected void print(int s) {
	print("" + s);
    }

    //@ assignable \nothing;
    protected void print(char s) {
	print("" + s);
    }

    //@ assignable \nothing;
    protected void print(double s) {
	print("" + s);
    }

    protected void acceptAll(/*@non_null@*/ ArrayList all ) {
	Iterator iter = all.iterator();
	while( iter.hasNext() ) {
	    ((JPhylum) iter.next()).accept(this);
	}
    }

    // ----------------------------------------------------------------------
    // PROTECTED DATA MEMBERS
    // ----------------------------------------------------------------------

    protected boolean		forInit;	// is on a for init
    protected int		TAB_SIZE = 2;
    protected int		WIDTH = 80;
    protected int		pos = 0;

    protected TabbedPrintWriter	p;
    protected boolean		nl = true;

    protected ModifierUtility modUtil;
}

