/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 * Copyright (C) 1990-2001 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjcSignatureParser.java,v 1.9 2008/05/14 18:16:21 wdietl Exp $
 */

package org.multijava.mjc;

import java.util.ArrayList;

import org.multijava.util.InconsistencyException;
import org.multijava.util.Utils;

public class MjcSignatureParser implements SignatureParser {

    /**
     * Parses a VM-standard type signature.
     *
     * @param	signature	the type signature
     * @return	the type represented by the signature
     */
    public final CType parseSignature(String signature) {
	return parseSignature(signature, 0, signature.length());
    }
    
    /**
     * Parses a VM-standard type signature within a signature string.
     *
     * @param	signature	the type signature
     * @param	from		the start index
     * @param	to		the end index
     * @return	the type represented by the signature 
     */
    protected CType parseSignature(String signature, int from, int to) {
	CType	type;
	int		bounds;
	
	bounds = 0;
	for (; signature.charAt(from) == '['; from++) {
	    bounds += 1;
	}
	
	switch (signature.charAt(from)) {
	case 'V':
	    type = CStdType.Void;
	    break;
	case 'B':
	    type = CStdType.Byte;
	    break;
	case 'C':
	    type = CStdType.Char;
	    break;
	case 'D':
	    type = CStdType.Double;
	    break;
	case 'F':
	    type = CStdType.Float;
	    break;
	case 'I':
	    type = CStdType.Integer;
	    break;
	case 'J':
	    type = CStdType.Long;
	    break;
	case 'L':
	    type =  CTopLevel.getTypeRep(signature.substring(from + 1, 
							     to -1),
					 true);
	    break;
	case 'S':
	    type = CStdType.Short;
	    break;
	case 'Z':
	    type = CStdType.Boolean;
	    break;
	case 'T': // JSR 14: TypeVariable
	    type = new CTypeVariableAlias(signature.substring(from + 1, to - 1));
	    break;
	default:
	    throw new InconsistencyException("Unknown signature: " + signature.charAt(from));
	}
	
        if( bounds==0 ) {
            return type;
        }
        if( type.isClassType() || bounds==1) {
            return new CArrayType(type, bounds, null);
        }
        // reach here if we are a primitive array of >1 dimension
	return new CArrayType(type, bounds, null, null);
    }
    
    protected CType parseGenericTypeSignature(String signature, char[] sig) {
	int                 from = current;
	CType               type = null;
	
	while (sig[current] == '[') {
	    current += 1;
	}
	
	if ((sig[current] != 'L') && (sig[current] != 'T')) {
	    current += 1;
	}
	
	else {
	    while ((sig[current] != ';') && (sig[current] != '<')) {
		current += 1;
	    }
	    current += 1;
	}
	if (sig[current-1] != '<') {
	    type = parseSignature(signature, from, current);
	} else {
	    String            ident = String.valueOf(sig, from+1, current - from-2);
	    CClassType[]      types = parseTypeArgumentSignature(signature, sig);
	    
	    type = CTopLevel.getTypeRep(ident, new CClassType[][]{types}, true);
	}
	return type;
    }
    
    protected CClassType[] parseTypeArgumentSignature(String signature, char[] sig) {
	ArrayList              vect = new ArrayList(10);
	
	while (sig[current] != '>') {
	    vect.add(parseGenericTypeSignature(signature, sig));
	}
	current++;
	// verify(sig[current] == ';');
	current++;
	
	return (CClassType[])vect.toArray(new CClassType[vect.size()]); 
    }
    
    protected CTypeVariable[] parseTypeParameter(String signature, char[] sig) {
	// '<' TypeVarName1 ':' bound1 ':' bound2 [...] TypeVarName2 : ... '>'
	if (sig[current] == '<') {
	    ArrayList    tvVect = new ArrayList(10);
	    
	    current++;
	    while (sig[current] != '>') {
		int             end = current; 
		String          ident; // type var name
		
		while (sig[end] != ':') {
		    end += 1;
		}
		ident = signature.substring(current, end);
		current = ++end;
		
		ArrayList          bounds = new ArrayList(5); // type bound

                // If no class but some interfaces have been specified in
                // the bounds of a type variable, the signature does not
                // simply contain no class bound, but an empty one, so let's
                // account for this.
                if (sig[current] == ':') {
                    current++;
                }

		bounds.add(parseGenericTypeSignature(signature, sig));
		while (sig[current] == ':') { // another bound ?
		    current++;
		    bounds.add(parseGenericTypeSignature(signature, sig));
		}
		
		CClassType[]    tvBound = (CClassType[])bounds.toArray(new CClassType[bounds.size()]);
		
		tvVect.add(new CTypeVariable(ident, tvBound));
	    }
	    
	    CTypeVariable[]   typeVariable = (CTypeVariable[])tvVect.toArray(new CTypeVariable[tvVect.size()]);
	    
	    for (int i = 0; i < typeVariable.length; i++) {
		typeVariable[i].setIndex(i);
	    }
	    current++;
	    return typeVariable;
	} else {
	    return CTypeVariable.EMPTY;
	}
    }
    
    /**
     * Returns an object representing the types the signature of a class
     *
     */
    public ClassSignature parseClassSignature(String signature) {
	CTypeVariable[]      tvs;
	CClassType           superType;
	CClassType[]         inter;
	char[]               sig = signature.toCharArray();
	
	current = 0;
	// typeVariable
	tvs = parseTypeParameter(signature, sig);
	// supertype
	superType = (CClassType) parseGenericTypeSignature(signature, sig);
	
	// interfaces
	ArrayList               interfaces = new ArrayList(10);
	
	while (current < sig.length) {
	    interfaces.add(parseGenericTypeSignature(signature, sig));
	}
	inter = (CClassType[])interfaces.toArray(new CClassType[interfaces.size()]);
	
	return new ClassSignature(superType, inter, tvs);
    }
    
    public static MjcSignatureParser getInstance() {
	return instance;
    }

    protected static final MjcSignatureParser instance = 
	new MjcSignatureParser();
    private int current;
}
