/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: ParserUtility.java,v 1.15 2005/12/22 17:54:05 tongjie Exp $
 */

package org.multijava.mjc;

import antlr.*;
import org.multijava.util.compiler.*;

import org.multijava.javadoc.JavadocComment;

/**
 * This class is delegated to my the parser classes and provides a
 * place to encapsulate the utility methods of those classes. */
public class ParserUtility implements TroubleReporter {
    public ParserUtility( Main compiler, 
			  ParsingController parsingController,
			  boolean allowGeneric,
			  boolean allowMultiJava, 
			  boolean allowRelaxedMultiJava, 
			  boolean allowUniverseKeywords, // WMD
			  boolean parseJavadocs ) {
	this.compiler = compiler;
	this.parsingController = parsingController;
	this.allowGeneric = allowGeneric;
	this.allowMultiJava = allowMultiJava;
	this.allowRelaxedMultiJava = allowRelaxedMultiJava;
	this.allowUniverseKeywords = allowUniverseKeywords; // WMD
	this.parseJavadocs = parseJavadocs;
    }

    /**
     * Returns the compiler that instantiated the parser.  */
    public final Main getCompiler() {
	return compiler;
    }

    /**
     * Constructs a token reference (i.e., a source code position
     * marker) for annotating the abstract syntax tree. */
    public final TokenReference buildTokenReference() {
	return TokenReference.build( parsingController.file(), 
				     parsingController.line(), 
				     parsingController.column() );
    }

    /**
     * Constructs a token reference (i.e., a source code position
     * marker) for annotating the abstract syntax tree.  This is more
     * efficient than the no-argument version.
     *
     * @param tok A token from which the row and column will be extracted */
    public final TokenReference buildTokenReference( Token tok ) {
	if (tok==null || tok.getLine()==0 && tok.getColumn()==0) {
	    // Token is unnumbered so instead of using it we use the
	    // current position of the parser, which may be up to k
	    // tokens ahead (for LL-k parsing).
	    return buildTokenReference();
	} // end of if 

	return TokenReference.build( parsingController.file(), 
				     tok.getLine(), tok.getColumn() );
    }

    /**
     * Constructs a token reference (i.e., a source code position
     * marker) for annotating the abstract syntax tree.  This is more
     * efficient than the no-argument version.
     *
     * @param wrap A wrapped token from which the row and column will be 
     * 		  extracted 
     *
     * <pre><jml>
     * requires wrap.unwrap() != null;
     * </jml></pre>
     *
     */
    public final TokenReference buildTokenReference( ParsingController.
						     TokenWrapper wrap ) {
	return buildTokenReference( wrap.unwrap() );
    }

    /**
     * Returns the non-Javadoc comments preceding the current
     * lookahead token.  */
    public final JavaStyleComment[] getStatementComment() {
	return null;			// !FIXME!comment
    }

    /**
     * Supplies an AST for the javadoc style comment preceding the
     * given token. */
    public final JavadocComment getJavadocComment( ParsingController.TokenWrapper 
					      declStart ) 
	throws TokenStreamException, RecognitionException
    {
	if( !parseJavadocs || declStart == null ) {
	    return null;
	}
	return getJavadocComment( declStart.unwrap() );
    }

    /**
     * Supplies an AST for the javadoc style comment preceding the
     * given token. */
    public JavadocComment getJavadocComment( Token tok ) 
	throws TokenStreamException, RecognitionException
    {
	if( !parseJavadocs ) {
	    return null;
	}
	try {
	    TokenStream tokStream = 
		parsingController.streamForBefore( "javadoc", tok );
	    JavadocParser parser = new JavadocParser( tokStream );
			
	    return parser.docComment();
	} catch( ParsingController.KeyException ke ) {
	    throw 
		new RecognitionException( "unable to establish javadoc parser." );
	}
    }

    /**
     * Flushes the javadoc token buffer of all javadoc tokens
     * preceding the given token.  Issues a warning if any tokens were
     * flushed.  */
    public final void flushJavadocTokensWithWarning( ParsingController.TokenWrapper
					       index ) 
	throws TokenStreamException
    {
	if( !parseJavadocs || index == null ) {
	    return;
	}
	flushJavadocTokensWithWarning( index.unwrap() );
    }
	
    /**
     * Flushes the javadoc token buffer of all javadoc tokens
     * preceding the given token.  Issues a warning if any tokens were
     * flushed.  */
    public final void flushJavadocTokensWithWarning( Token tok ) 
	throws TokenStreamException
    {
	if( !parseJavadocs ) {
	    return;
	}
	try {
	    if( parsingController.discardTokensBefore( "javadoc", tok ) ) {
		reportTrouble( 
			      new CWarning( buildTokenReference( tok ),
					    MjcMessages.MISPLACED_JAVADOC ));
		
	    }
	} catch( ParsingController.KeyException ke ) {
	    throw 
		new TokenStreamException( "unable to discard tokens for javadoc" );
	}
    }
	
    /**
     * If the given wrapper reference is non-null and the wrapper is
     * empty then wrap the given token in the wrapper.  */
    public final void wrapIfEmptyNonNullWrapper( ParsingController.TokenWrapper 
					   wrapper, Token tok ) {
	if( wrapper != null ) {
	    wrapper.ifEmptyWrap( tok );
	}
    }

    /**
     * Reports that an error has been detected in the lexical analyser.
     * The handling is delegated to the compiler driver.
     * @param	trouble		the error to report
     */
    public final void reportTrouble( Exception trouble ) {
	compiler.reportTrouble( trouble );
    }

    /**
     * Returns an exception from our exception hierarchy by wrapping
     * the ANTLR excepton.  The exception is really positioned, by the
     * positioning information is embedded in the message. */
    public final PositionedError beautifyParserError( ANTLRException e ) {
	String msg = e.getMessage();
	return new PositionedError( buildTokenReference(), 
				    CompilerMessages.SYNTAX_ERROR, msg );
    }

    /**
     * Returns the name of the modifier according to the
     * <code>ModifierUtility</code> of the compiler used to
     * instantiate this.
     *
     * @see org.multijava.util.compiler.ModifierUtility#nameOf(int) */
    public final String getModifierName( long mod ) {
	return compiler.modUtil().nameOf( mod );
    }

    /**
     * Returns the names of the modifiers according to the
     * <code>ModifierUtility</code> of the compiler used to
     * instantiate this.
     *
     * @see org.multijava.util.compiler.ModifierUtility#asString(int) */
    public final String getModifierNames( long mods ) {
	return compiler.modUtil().asString( mods );
    }

    /**
     * Returns the true if the modifiers are in the preferred order
     * according to the <code>ModifierUtility</code> of the compiler
     * used to instantiate this.
     *
     * @see org.multijava.util.compiler.ModifierUtility#inPreferredOrder(int,int) 
     */
    public final boolean modifiersInPreferredOrder( long prevMods, long mod ) {
	return compiler.modUtil().inPreferredOrder( prevMods, mod );
    }
    
    /**
     * get the value of unmatchedTypeLT, from the ParsingController
     */
    public int getUnmatchedTypeLT() {
      return this.parsingController.getUnmatchedTypeLT();
    } 

    /**
     * set the value of unmatchedTypeLT, from the ParsingController
     */
    public void setUnmatchedTypeLT(int count) {
      this.parsingController.setUnmatchedTypeLT(count);
    }

    // --------------------------------------------------------------------
    // DATA MEMBERS
    // --------------------------------------------------------------------

    protected Main compiler;

    /**
     * manages parsing of nested language features (like javadoc comments) */
    protected ParsingController parsingController;

    /**
     * Indicates whether multijava syntax should be recognized. */
    public final boolean allowMultiJava;
    /**
     * Indicates whether generics syntax should be recognized. */
    public final boolean allowGeneric;  

    /**
     * Indicates whether relaxed multijava syntax should be recognized. */
    public final boolean allowRelaxedMultiJava;

    /**
     * WMD
     * Indicates whether universe type modifier syntax should be recognized. */
    public final boolean allowUniverseKeywords;

    /**
     * Indicates whether javadoc comments should be parsed. */
    public final boolean parseJavadocs;

}
