/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: ParsingController.java,v 1.12 2005/12/22 17:54:05 tongjie Exp $
 */

package org.multijava.mjc;

import antlr.LexerSharedInputState;
import antlr.Token;
import antlr.TokenStream;
import antlr.TokenStreamException;

import java.io.Reader;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.io.File;

/**
 * This class manages the parsing of source code.  It supports the use
 * of multiple lexers and parsers to process nested languages (as in
 * javadoc comments in Java programs).  It includes most of the
 * functionality of antlr.TokenStreamSelector, though for efficiency
 * it duplicates the necessary code rather than inheriting it.  This
 * class buffers the tokens produced by multiple lexers and allows
 * client parsers to accept tokens from just a single lexer. */
public class ParsingController {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /**
     * Establish a parsing controller for the given reader.  The
     * reader is used to generate shared lexer state for the various
     * lexers tied to this.  A lexer tied to this parser should be
     * instantiated by passing it this.  The lexer's constructor will
     * call a super constructor passing it the result of calling
     * <code>sharedInputState()</code> on this.
     *
     * @see #sharedInputState() */
    public ParsingController( Reader r, File file ) {
	super();
	this.inputState = new InputState( r, file ); 
    }

    //---------------------------------------------------------------------
    // PUBLIC METHODS, INITIALIZATION
    //---------------------------------------------------------------------

    /**
     * Returns an input state that can be shared by the lexers
     * attached to this controller.  */
    public LexerSharedInputState sharedInputState() {
	return inputState;
    }

    /**
     * Adds the given input stream to the controller and binds it to
     * the given key.
     *
     * @exception ParsingController.KeyException the given key is already bound or if
     * either argument is null */
    public void addInputStream( TokenStream stream, String key ) 
	throws KeyException
    {
	if( stream == null || key == null ) {
	    throw new KeyException( "null argument to addInputStream" );
	}
	
	if( keyStreamBufferMap.containsKey( key ) ) {
	    throw new KeyException( "key already bound: " + key );
	}
	keyStreamBufferMap.put( key, new StreamBufferPair( stream ) );
    }

    /**
     * Returns the currently active input stream. */
    public TokenStream getCurrentStream() {
	return activeStreamBuffer.stream;
    }

    /**
     * Returns the named input stream (previously added with addInputStream). */
    public TokenStream getStream(String sname) {
	StreamBufferPair sbp = 
	    (StreamBufferPair) keyStreamBufferMap.get( sname );
        if (sbp == null) {
            throw new IllegalArgumentException("TokenStream " + sname + " not found");
        }
        return sbp.stream;
    }

    /**
     * Makes the lexer bound to the given key the initial lexer.
     *
     * @exception ParsingController.KeyException if the given key is not bound 
     * @exception ParsingController.ConfigurationException if the initial lexer has already 
     * been selected */
    public synchronized void selectInitial( String key ) 
	throws KeyException, ConfigurationException
    {
	if( initialSet ) {
	    throw new ConfigurationException( "initial input stream " + 
					      "already set" );
	}
	useKey( key );
	StreamBufferPair sbp = 
	    (StreamBufferPair) keyStreamBufferMap.get( key );
	setActive( key, sbp );
	initialFilteredStream = new FilteredTokenStream( sbp.buffer );
	initialSet = true;
    }

    /**
     * Returns a token stream that provides tokens generated by the
     * initial lexer.
     *
     * @exception ParsingController.ConfigurationException if the initial lexer has not been 
     * 		selected. 
     * @see #selectInitial(String)
     */
    public synchronized FilteredTokenStream initialOutputStream() 
	throws ConfigurationException 
    {
	if( initialFilteredStream == null ) {
	    throw new ConfigurationException( "no initial input " + 
					      "stream selected, see " + 
					      "ParsingController." + 
					      "selectInitial");

	} else {
	    return initialFilteredStream;
	}
    }

    /**
     * Indicates that tokens generated by the lexer bound to the given
     * key should be discarded instead of buffered.  This allows the
     * controller to make more efficient use of memory.
     *
     * @exception ParsingController.KeyException if the given key is not bound 
     * to a lexer or if a TokenStream for the key has already been exposed 
     */
    public synchronized void discardAllTokensFor( String key ) 
	throws KeyException 
    {
	useKey( key );
	keysToDiscard.add( key );
	// Destroy the reference to the buffer in the stream-buffer
	// pair for this key so the buffer can be garbage collected.
	// The discarding flag should ensure that this null buffer is
	// never referenced.
	StreamBufferPair sbp = 
	    (StreamBufferPair) keyStreamBufferMap.get( key );
	sbp.buffer = null;
    }

    //---------------------------------------------------------------------
    // PUBLIC METHODS, LEXING/PARSING
    //---------------------------------------------------------------------

    /**
     * Returns a token stream from the lexer bound to the given key
     * and containing all the tokens preceding the given token of the
     * surrounding language.
     * 
     * @exception ParsingController.KeyException if the given key is not bound to a
     * lexer */
    public FilteredTokenStream streamForBefore( String key, Token index ) 
	throws KeyException
    {
	try {
	    verifyHasKey( key );
	    if( keysToDiscard.contains( key ) ) {
		IndexedTokenBuffer buffer = newBlockedEmptyBuffer();
		return new FilteredTokenStream( buffer );
	    }
	    StreamBufferPair sbp = 
		(StreamBufferPair) keyStreamBufferMap.get( key );
	    return sbp.buffer.streamForTokensBefore( index );
	} catch( TokenStreamException e ) {
	    // cannot be thrown because no buffers in
	    // keyStreamBufferMap are blocked
	    throw new RuntimeException( "shouldn't get here" );
	}
    }

    /**
     * Discards all tokens from the lexer bound to the given key that
     * precede the given token of the surrounding language.
     *
     * @return true if any tokens are actually discarded 
     * @exception ParsingController.KeyException if the given key is not bound to a lexer */
    public boolean discardTokensBefore( String key, Token index )
	throws KeyException
    {
	try {
	    if( keysToDiscard.contains( key ) ) {
		return false;
	    }
	    verifyHasKey( key );
	    StreamBufferPair sbp = 
		(StreamBufferPair) keyStreamBufferMap.get( key );
	    return sbp.buffer.flushTokensBefore( index );
	} catch( TokenStreamException e ) {
	    // cannot be thrown because no buffers in
	    // keyStreamBufferMap are blocked
	    throw new RuntimeException( "shouldn't get here" );
	}
    }

    /**
     * Pops a stream from the stack of input streams.  Called when
     * exiting a nested language block.
     *
     * @exception antlr.TokenStreamException if a pending operation prevents 
     * popping */
    public synchronized void pop() throws TokenStreamException {
	// The actual pop must be delayed until the lexer returns a
	// value to bufferNextToken, otherwise the token returned
	// might be assigned to the wrong output stream.
	if( fetching ) {
	    if( operationPending == NONE ) {
		operationPending = POP;
	    } else {
		throw new TokenStreamException( "Operation pending. " + 
						"Cannot pop." );
	    }
	} else {
	    processPop();
	}
    }

	/** Returns the TokenStream corresponding to a given name.
		@param key		the name of the TokenStream to return
		@return		        the matching TokenStream
	*/
	//@ requires key != null;
	public synchronized TokenStream get(String key) throws KeyException, TokenStreamException {
		verifyHasKey(key);
		StreamBufferPair sbp = 
			(StreamBufferPair) keyStreamBufferMap.get( key );
		return sbp.stream;
	}
	
    /**
     * Switches to a new input stream for future token requests,
     * maintaining a stack of the old input streams.  Called when
     * entering a nested language block.
     *
     * @param	key	the unique identifier of the lexer to switch to
     * @exception antlr.TokenStreamException if a pending operation prevents 
     *                                       pushing
     * @exception ParsingController.KeyException if the given key is not bound to a lexer */
    public synchronized void push( String key ) 
	throws KeyException, TokenStreamException
    {
	verifyHasKey( key );
	// The actual push must be delayed until the lexer returns a
	// value to bufferNextToken, otherwise the token returned
	// might be assigned to the wrong output stream.
	if( fetching ) {
	    if( operationPending == NONE ) {
		pendingKey = key;
		operationPending = PUSH;
	    } else {
		throw new TokenStreamException( "Operation pending. " + 
						"Cannot push: " + key );
	    }
	} else {
	    processPush( key );
	}
    }

    /**
     * Returns the file name of the input.
     */
    public File file() {
	return inputState.file();
    }

    /**
     * Returns the current line number in the input.
     */
    public int line() {
	return inputState.line();
    }
    
    /**
     * Returns the current column number in the input.
     */
    public int column() {
	return inputState.column();
    }
    
    //---------------------------------------------------------------------
    // PRIVATE METHODS
    //---------------------------------------------------------------------

    /**
     * Retrieves the next token from the currently active lexer and
     * adds it to the appropriate buffer or discards it if necessary.
     *
     * @exception antlr.TokenStreamException passed on from underlying lexer */
    private synchronized void bufferNextToken() throws TokenStreamException {
	fetching = true;
	if( !initialSet ) {
	    throw new TokenStreamException( "no initial input " + 
					    "stream selected, see " + 
					    "ParsingController.selectInitial");
	}
    fetch:
	{
	    if( endOfFileReached ) {
		break fetch;
	    }
	    Token tok = activeStreamBuffer.stream.nextToken();
	    if( tok == IGNORE ) {
		break fetch;
	    }
	    if( tok.getType() == Token.EOF_TYPE ) {
		endOfFileReached = true;
		// add EOF token to every buffer that isn't discarding
		Iterator iter = keyStreamBufferMap.values().iterator();
		while( iter.hasNext() ) {
		    IndexedTokenBuffer itb = 
			((StreamBufferPair) iter.next()).buffer;
		    if( itb != null ) {
			itb.bufferToken( tok );
		    }
		}
		if( nestedLanguageBuffer != null ) {
		    nestedLanguageBuffer.recordIndexToken( tok );
		}
	    } else if( !discarding ) {
		activeStreamBuffer.buffer.bufferToken( tok );
		if( nestedLanguageBuffer != null ) {
		    nestedLanguageBuffer.recordIndexToken( tok );
		}
	    }
	} // end of 'fetch' block
	// done fetching, process pending pop or push
	switch( operationPending ) {
	case PUSH:
	    processPush( pendingKey );
	    break;
	case POP:
	    processPop();
	    break;
	case NONE:
	    break;
	default:
	    throw new TokenStreamException( "invalid operation pending" );
	}
	pendingKey = null;
	operationPending = NONE;
	fetching = false;
    }

    /**
     * Switches to a new input stream, maintaining a stack of the old
     * input streams.  Called when entering a nested language block.
     *
     * @param	key	the unique identifier of the lexer to switch to */
    private void processPush( String key ) {
	streamBufferStack.push( activeStreamBuffer );
	keyStack.push( activeKey );
	StreamBufferPair sbp = 
	    (StreamBufferPair) keyStreamBufferMap.get( key );
	setActive( key, sbp );
    }

    /**
     * Pops a stream from the stack of input streams.  Called when
     * exiting a nested language block.  */
    private void processPop() {
	// If we're popping from a discarded language then keep the
	// current nested language buffer for indexing.  Otherwise add
	// future indices to the buffer for the language that we are
	// popping from.
	if( !discarding ) {
	    nestedLanguageBuffer = activeStreamBuffer.buffer;
	}
	setActive( (String) keyStack.pop(),
		   (StreamBufferPair) streamBufferStack.pop() );
    }

    /**
     * Registers that tokens generated by the lexer bound to the given
     * key will be used or discarded.  Handles checks to prevent
     * assigning a lexer to multiple output streams or to both discard
     * and output.
     *
     * @param	key	the unique identifier of the lexer whose tokens are 
     *			to be discarded or passed to an output stream
     * @exception ParsingController.KeyException if the given key is not bound 
     * to a lexer or if a TokenStream for the key has already been exposed */
    private void useKey( String key ) throws KeyException {
	verifyHasKey( key );

	if( usedKeys.contains( key ) ) {
	    throw new KeyException( "key already used: " + key );
	}

	usedKeys.add( key );
    }
    
    /**
     * Verifies that the given key is bound to a lexer.
     *
     * @exception ParsingController.KeyException if the given key is not bound 
     */
    private void verifyHasKey( String key ) throws KeyException {
	if( !keyStreamBufferMap.containsKey( key ) ) {
	    throw new KeyException( "unknown key: " + key );
	}
    }

    /**
     * Sets the given key, input stream, and token buffer to be
     * active.  */
    private void setActive( String key, StreamBufferPair sbp ) {
	activeKey = key;
	activeStreamBuffer = sbp;
	discarding = keysToDiscard.contains( key );
    }
    
    /**
     * Factory method creates a new token buffer that is blocked and
     * empty.  */
    private IndexedTokenBuffer newBlockedEmptyBuffer() {
	LinkedList buffers = new LinkedList();
	buffers.add( new LinkedList() );
	return new IndexedTokenBuffer( buffers );
    }
    
    /**
     * get the value of unmatchedTypeLT
     */
    public int getUnmatchedTypeLT() {
      return unmatchedTypeLT;
    } 

    /**
     * set the value of unmatchedTypeLT
     */
    public void setUnmatchedTypeLT(int count) {
      this.unmatchedTypeLT = count;
    }

    //---------------------------------------------------------------------
    // NESTED AND INNER CLASSES
    //---------------------------------------------------------------------

    /**
     * This inner class represents exceptions that can be thrown
     * because of misconfiguration of the parsing controller.  */
    public static class ConfigurationException extends Exception {
	public ConfigurationException() {
	    super();
	}
	public ConfigurationException( String msg ) {
	    super( msg );
	}
    }

    /**
     * This inner class represents exceptions that can be thrown
     * because of bad key values passed as arguments to the outer
     * classes methods.  */
    public static class KeyException extends ConfigurationException {
	public KeyException() {
	    super();
	}
	public KeyException( String msg ) {
	    super( msg );
	}
    }

    /**
     * This inner class encapsulates a buffer and an input stream so
     * they can be the stored as a single value in a map.  */
// !FIXME! The following should be private when Sun fixes its javadoc bug - DRCok
    /*private*/ class StreamBufferPair {
	private StreamBufferPair( TokenStream stream ) {
	    this.stream = stream;
	}
	
	private TokenStream stream;
	private IndexedTokenBuffer buffer = new IndexedTokenBuffer();
    }

    /**
     * This inner class wraps a single token.  It can be used by
     * clients to find the first token in actually appearing in when
     * matching a rule that begins with optional non-terminals.  */
    public static class TokenWrapper {
	public void ifEmptyWrap( Token t ) {
	    if( contents == null ) {
		contents = t;
	    }
	}
	public /*@ pure @*/ Token unwrap() {
	    return contents;
	}
	private Token contents;
    }

    /**
     * This inner class buffers tokens of a nested language and allows
     * selection and discarding of a subset of these buffered tokens
     * based on a reference to an index token of the surrounding
     * language. */
// !FIXME! The following should be private when Sun fixes its javadoc bug - DRCok
    /*private*/ class IndexedTokenBuffer {
	/**
	 * Creates a new empty token buffer.
	 */
	public IndexedTokenBuffer() {
	    tokenBuffers = new LinkedList();
	    indexTokens = new LinkedList();
	    isBlocked = false;
	}

	/**
	 * Creates a blocked token buffer from the given list of token
	 * buffers.  A token buffer that is blocked does
	 * not accept new tokens or indices. */
	public IndexedTokenBuffer( LinkedList tokenBuffers ) {
	    this.tokenBuffers = tokenBuffers;
	    indexTokens = null;
	    isBlocked = true;
	}

	/**
	 * Adds the given token of the nested language to the buffer.  
	 *
	 * @exception antlr.TokenStreamException if buffer is blocked
	 */
	public void bufferToken( Token tok ) throws TokenStreamException {
	    if( isBlocked ) {
		throw new TokenStreamException( "cannot add token to " + 
						"blocked buffer" );
	    }
	    if( startNewBuffer ) {
		// start of new block of contiguous tokens
		tokenBuffers.addLast( new LinkedList() );
		indexTokens.addLast( new LinkedList() );
		startNewBuffer = false;
	    }
	    ((LinkedList) tokenBuffers.getLast()).addLast( tok );
	}

	/**
	 * Records the given index token.   
	 *
	 * @exception antlr.TokenStreamException if buffer is blocked
	 */
	public void recordIndexToken( Token index ) 
	    throws TokenStreamException 
	{
	    if( isBlocked ) {
		throw new TokenStreamException( "cannot add index to " + 
						"blocked buffer" );
	    }
	    if( !startNewBuffer ) {
		// end of block of contiguous tokens
		startNewBuffer = true;
	    }
	    if( tokenBuffers.size() > 0 ) {
		((LinkedList) indexTokens.getLast()).addLast( index );
	    }
	    // No buffered tokens before the index token so discard
	    // the index token.  (For unblocked buffers
	    // tokenBuffers.size() == indexTokens.size() and so we
	    // must discard the index token or the line above will
	    // throw a NoSuchElementException on the call to
	    // getLast().)
	}

	/**
	 * Returns a token stream that includes just the buffered
	 * tokens that came before the given index token.  Removes
	 * those buffered tokens from this.  If the index token is not
	 * recorded in this, then the stream returned will just
	 * contain an end-of-file token.    
	 *
	 * @exception antlr.TokenStreamException if buffer is blocked
	 */
	public FilteredTokenStream streamForTokensBefore( Token index ) 
	    throws TokenStreamException 
	{
	    if( isBlocked ) {
		throw new TokenStreamException( "cannot select tokens from " + 
						"blocked buffer" );
	    }
	    int removeCount = findIndexToken( index ) + 1;
	    if( removeCount == 0 ) {
		IndexedTokenBuffer buffer = newBlockedEmptyBuffer();
		return new FilteredTokenStream( buffer );
	    }
	    // Remove the buffered tokens (and index buffers) and
	    // collect the removed tokens into a new list of token buffers.
	    LinkedList newTokenBuffers = new LinkedList();
	    for( int i = 1; i <= removeCount; i++ ) {
		newTokenBuffers.addLast( tokenBuffers.removeFirst() );
		indexTokens.removeFirst();
	    }
	    IndexedTokenBuffer buffer = 
		new IndexedTokenBuffer( newTokenBuffers );
	    return new FilteredTokenStream( buffer );
	}

	/**
	 * Discards all buffered tokens before the given index token.
	 * If the index token is not recorded in this then no buffered
	 * tokens are discarded.
	 *
	 * @return true, iff some buffered tokens were discarded   
	 * @exception antlr.TokenStreamException if buffer is blocked
	 */
	public boolean flushTokensBefore( Token index )  
	    throws TokenStreamException 
	{
	    if( isBlocked ) {
		throw new TokenStreamException( "cannot select tokens from " + 
						"blocked buffer" );
	    }
	    int removeCount = findIndexToken( index ) + 1;
	    if( removeCount == 0 ) {
		return false;
	    }
	    for( int i=1; i<=removeCount; i++ ) {
		tokenBuffers.removeFirst();
		indexTokens.removeFirst();
	    }
	    return true;
	}

	/**
	 * Indicates whether this token buffer is blocked from
	 * accepting additinal tokens or indices.  */
	public boolean isBlocked() {
	    return isBlocked;
	}

	/**
	 * Returns the next token from the buffer.  */
	public Token next() {
	    LinkedList firstSubbuffer = (LinkedList) tokenBuffers.getFirst();
	    Token result = (Token) firstSubbuffer.removeFirst();
	    if( firstSubbuffer.size() == 0 &&
		( isBlocked || 
		  ((LinkedList) indexTokens.getFirst()).size() != 0 ) ) {
		// We have consumed all tokens in the oldest buffer
		// and no more tokens can be added to it.  Thus we can
		// remove the oldest buffer and the corresponding
		// indices.
		tokenBuffers.removeFirst();
		if( !isBlocked() ) {
		    indexTokens.removeFirst();
		}
	    }
	    return result;
	}

	/**
	 * Returns true if this buffer has additional tokens.  A
	 * buffer that is not blocked may return false and then return
	 * true on subsequent calls if tokens are added to the
	 * buffer. */
	public boolean hasNext() {
	    return tokenBuffers.size() > 0 && 
		((LinkedList) tokenBuffers.getFirst()).size() > 0;
	}

	/**
	 * Returns the position in tokenBuffers of the last contiguous
	 * block of buffered tokens that comes before the given index
	 * token.  Returns <code>-1</code> if the index token is not
	 * recorded in this.  */
	private int findIndexToken( Token index ) {
	    Iterator iter = indexTokens.iterator();
	    int position = 0;
	    while( iter.hasNext() ) {
		LinkedList indexBlock = (LinkedList) iter.next();
		if( indexBlock.contains( index ) ) {
		    return position;
		}
		position++;
	    }
	    return -1;
	}

	/**
	 * Adds an end-of-file token to the end of the token buffer.
	 * Should only be called on a blocked buffer.  */
	private void addEOF() {
	    ((LinkedList) tokenBuffers.getLast()).
		addLast( new Token( Token.EOF_TYPE ) );
	}
	
	/** 
	 * Indicates whether this is a blocked token buffer, to which
	 * new tokens or indices cannot be added.  */
	private boolean isBlocked;

	/**
	 * Indicates whether a new buffer is required to store the
	 * next buffered token.  */
	private boolean startNewBuffer = true;

	/**
	 * A list of token buffers.  Each token buffer stored in this
	 * list contains a group of contiguous tokens from the
	 * original source text.  Unless isBlocked, no token buffer
	 * stored in this list is empty. */
	private LinkedList tokenBuffers;
	/**
	 * A list of buffers containing index tokens of the
	 * surrounding language.  Unless isBlocked, for all i,
	 * 0<=i<tokenBuffers.size(), all tokens in tokenBuffers.get(i)
	 * precede all tokens in indexTokens.get(i).  Unless
	 * isBlocked, no buffer stored in this list is empty. */
	private LinkedList indexTokens;
    }
    
    /**
     * This inner class provides a token stream that accesses one of the
     * lexers managed by the controller. */
    public class FilteredTokenStream implements TokenStream {
	/**
	 * Creates a new token stream reading from the given buffer.
	 * If the buffer is blocked then reading more than the number
	 * of tokens in the buffer will yield an end-of-file token. */
	private FilteredTokenStream( IndexedTokenBuffer buffer ) {
	    if( buffer.isBlocked() ) {
		buffer.addEOF();
	    }
	    this.buffer = buffer;
	}

	/**
	 * Returns the next token from the appropriate lexer.  */
	public Token nextToken() throws TokenStreamException {
	    if( localEOFReached ) {
		throw new TokenStreamException( "past end-of-file" );
	    }
	    // the end-of-file token is added to every buffer, so
	    // repeated calls to bufferNextToken will eventually
	    // result in a non-zero buffer size
	    while( !buffer.hasNext() ) {
		bufferNextToken();
	    }

	    Token tok = (Token) buffer.next();
	    if( tok.getType() == Token.EOF_TYPE ) {
		localEOFReached = true;
	    }

	    return tok;
	}

	/**
	 * Maintains a link to the token buffer that is applicable for
	 * this output stream.  */
	private final IndexedTokenBuffer buffer;

	/**
	 * Indicates whether an end-of-file token has been processed
	 * on the attached input stream. */
	private boolean localEOFReached = false;
    }
    
    /**
     * This nested class is a hack to expose the line and column
     * fields of <code>LexerSharedInputState</code> for reading.  */
// !FIXME! The following should be private when Sun fixes its javadoc bug - DRCok
    /*private*/ static class InputState extends LexerSharedInputState {
	public InputState( Reader r, File file ) {
	    super( r );
	    this.file = file;
	    this.filename = file == null ? "unknown" : file.getName();
	}
	
	public int line() {
	    return line;
	}

	public int column() {
	    return column;
	}

	public File file() {
	    return file;
	}

	private final File file;
    }
    
    //---------------------------------------------------------------------
    // DATA MEMBERS
    //---------------------------------------------------------------------

    /**
     * The input state that should be shared by the lexers bound to
     * this parsing controller.  */
    private InputState inputState;

    /**
     * Indicates whether the initial input stream has been selected.  */
    private boolean initialSet = false;

    /**
     * The output stream for the initially selected lexer.  */
    private FilteredTokenStream initialFilteredStream;

    /**
     * Maps keys to lexers and buffers */
    private Map keyStreamBufferMap = new HashMap();

    /**
     * The set of keys corresponding to lexers whose output are being
     * discarded or have been bound to output token streams.  */
    private Set usedKeys = new HashSet();

    /**
     * The set of keys corresponding to lexers whose output are being
     * discarded. */
    private Set keysToDiscard = new HashSet();

    /**
     * Indicates whether tokens for the currently active stream are to
     * be discarded. */
    private boolean discarding = false;

    /**
     * For efficiency, used to track the keys when switching between
     * input streams.  The top of the the stack is the previous active
     * key, not the current active key. */
    private Stack keyStack = new Stack();

    /**
     * Tracks the active input streams.  The top of the the stack is
     * the previous active input stream and token buffer, not the
     * current. */
    private Stack streamBufferStack = new Stack();


    /**
     * The key of the currently active input stream. */
    private String activeKey;

    /**
     * The currently active input stream and token buffer. */
    private StreamBufferPair activeStreamBuffer;

    /**
     * The active buffer before the last pop operation.  Used to store
     * index tokens into the buffer of a nested language.  */
    private IndexedTokenBuffer nestedLanguageBuffer = null;

    /**
     * Indicates whether an end-of-file token has been received on
     * some input stream.  */
    private boolean endOfFileReached = false;

    /**
     * Indicates that a token fetch request is in progress for the
     * active input stream.  */
    private boolean fetching = false;

    /**
     * Indicates whether any input stream switching operations (i.e.,
     * push, pop) are pending on completion of the current
     * call to bufferNextToken.  */
    private int operationPending = NONE;
    private int ltCount = 0;
    /**
     * The key of any pending push operation.
     *
     * @see #operationPending */
    private String pendingKey = null;

    /**
     * Number of LT tokens that have not been matched by GT in Type Arguments
     * or Type Parameters
     */
    private int unmatchedTypeLT = 0 ;

    /**
     * When operationPending equals this, then no operations are
     * pending.
     *
     * @see #operationPending */
    private static final int NONE =   0;

    /**
     * When operationPending equals this, then a pop operation is
     * pending.
     *
     * @see #operationPending */
    private static final int POP =    1;

    /**
     * When operationPending equals this, then a push operation is
     * pending.
     *
     * @see #operationPending */
    private static final int PUSH =   2;

    /**
     * A token that may be returned by lexers upon detecting a lexical
     * pattern that triggers a switch between lexers (e.g.,
     * <code>/**</code> for switching to javadoc).  This token is not
     * passed to any client parsers but is swallowed by the parsing
     * controller.  Using this token in a lexer allows the associated
     * parser to ignore the lexical triggers.  Note that using a
     * "skip" token in ANTLR doesn't work for this purpose because the
     * generated lexer code will not return from a nextToken call
     * until a non-skip token is found.  Thus using a skip token
     * causes the lexer switch to be delayed. */
    public static final Token IGNORE = 
	new antlr.CommonToken( Token.INVALID_TYPE, "<ignored>" );
}



