/*
 * (C) 2004 Dependable Software Research Group, Concordia University.
 * 
 * This file is part of mjc, the MultiJava Compiler.
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: SafeIntegralArithmetic.java,v 1.2 2005/04/06 02:35:38 xi_hao Exp $
 */

/**
 * An implementation of safe integral arithmetic based on
 * "A safe variant of the unsafe integer arithmetic of Java(tm)" by
 * Juergen F.H. Winkler. Software-Practice and Experience 2002; 32:669-701.
 * 
 * Only arithmetic operators that can overflow are provided and only
 * for int and long arguments.  These are mean to emulate what corresponding
 * operations could do in the JVM if overflow were detected.
 *
 * @author P.Chalin, DSRG
 * @author Hao Xi, DSRG
 */

package org.multijava.mjc;

public class SafeIntegralArithmetic {

    private static ArithmeticException newArithException(String expr) {
	return new ArithmeticException(expr + " overflow");
    }

    private static ArithmeticException newArithException(
							 String operatorAndSignature, long arg1, long arg2) //
    {
	return newArithException(operatorAndSignature + //
				 "(" + arg1 + ", " + arg2 + ")");
    }

    // ------------------------------- ADDITION ------------------------------

    public static final int add(int arg1, int arg2) {
	long result = (long) arg1 + (long) arg2;
	if (result < Integer.MIN_VALUE || result > Integer.MAX_VALUE)
	    throw newArithException("int addition", arg1, arg2);
	return (int) result;
    }

    public static final long add(long arg1, long arg2) {
	long result = arg1 + arg2;
	if ((arg1 > 0 && arg2 > 0 && result < 0)
	    || (arg1 < 0 && arg2 < 0 && result >= 0))
	    throw newArithException("long addition", arg1, arg2);
	return result;
    }

    // ------------------------------- SUBTRACTION ------------------------

    public static final int subtract(int arg1, int arg2) {
	long result = (long) arg1 - (long) arg2;
	if (result < Integer.MIN_VALUE || result > Integer.MAX_VALUE)
	    throw newArithException("int subtraction", arg1, arg2);
	return (int) result;
    }

    public static final long subtract(long arg1, long arg2) {
	long result = arg1 - arg2;
	if ((arg1 >= 0 && arg2 < 0 && result < 0)
	    || (arg1 < 0 && arg2 > 0 && result > 0))
	    throw newArithException("long subtraction", arg1, arg2);
	return result;
    }

    // ------------------------------- NEGATION --------------------------

    public static final int negate(int arg) {
	if (arg == Integer.MIN_VALUE)
	    throw newArithException("int negation");
	return (int) (-arg);
    }

    public static final long negate(long arg) {
	if (arg == Long.MIN_VALUE)
	    throw newArithException("long negation");
	return -arg;
    }

    // ------------------------------- MULTIPLICATION -----------------------

    public static final int multiply(int arg1, int arg2) {
	long result = (long) arg1 * (long) arg2;
	if (result < Integer.MIN_VALUE || result > Integer.MAX_VALUE)
	    throw newArithException("int multiplication", arg1, arg2);
	return (int) result;
    }

    public static final long multiply(long arg1, long arg2) {
	long result = arg1 * arg2;
	if ((arg1 > 0 && (arg2 < Long.MIN_VALUE / arg1 || //
			  arg2 > Long.MAX_VALUE / arg1))
	    || (arg1 == -1 && arg2 == Long.MIN_VALUE)
	    || (arg1 < -1 && (arg2 < Long.MAX_VALUE / arg1 || arg2 > Long.MIN_VALUE
			      / arg1)))
	    throw newArithException("long multiplication", arg1, arg2);
	return result;
    }

    // ------------------------------- DIVISION -------------------------

    public static final int divide(int arg1, int arg2) {
	long result = (long) arg1 / (long) arg2;
	if (arg1 == Integer.MIN_VALUE && arg2 == -1)
	    throw newArithException("int division", arg1, arg2);
	else
	    return (int) result;
    }

    public static final long divide(long arg1, long arg2) {
	long result = arg1 / arg2;
	if (arg1 == Long.MIN_VALUE && arg2 == -1)
	    throw newArithException("long division", arg1, arg2);
	return result;
    }

    //--------------------------checked/unchecked----------------------------

    public static final int checked(int i) {
	return i;
    }

    public static final int unchecked (int i) {
	return i;
    }

    public static final long checked(long i) {
        return i;
    }

    public static final long unchecked (long i) {
        return i;
    }

    public static final float checked(float i) {
        return i;
    }

    public static final float unchecked (float i) {
        return i;
    }

    public static final double checked(double i) {
        return i;
    }

    public static final double unchecked (double i) {
        return i;
    }

}


