/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestMain_Runtime.java,v 1.10 2004/01/16 23:21:12 cclifton Exp $
 */

package org.multijava.mjc;

import org.multijava.util.testing.TestCase;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;

import java.util.StringTokenizer;

public class TestMain_Runtime extends FunctionalTestSuite.TestCase {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------
    
    public TestMain_Runtime( String name ) {
	super( name );
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // Some simple self-tests
    //---------------------------------------------------------------------

    /**
     * Should successfully compile and run the (very simple) source code.  */
    public void test_simple_class_1() {
	String className = "SimpleClass1";
	String filename = className + ".java";
	try {
	    CompilationResults compRes = 
		compile( saveAs( filename,
				 "public class SimpleClass1 {" + NEWLINE +
				 "  public static void main( String[] args ) {"
				 + NEWLINE +
				 "    System.out.print( \"Hello, world!\" );" 
				 + NEWLINE +
				 "  }" + NEWLINE +
				 "}" 
				 ) );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( className );
	    expectSuccess( runRes );
	    assertEquals( "Hello, world!", runRes.stdOut, true );
	} finally {
	    remove( filename );
	}
    }

    //---------------------------------------------------------------------
    // Internalizing
    //---------------------------------------------------------------------

    /**
     * Should successfully internalize the augmenting methods.  */
    public void test_internalize_1() {
	String className = "Internalize1";
	String filename = className + ".java";
	try {
	    CompilationResults compRes = 
		compile( saveAs( filename,
				 PACKAGE_DECL +
				 "public class Internalize1 {" + NEWLINE +
				 "  public static void main( String[] args ) {"
				 + NEWLINE +
				 "    Internalize1 i = new Internalize1();" +
				 NEWLINE + 
				 "    i.hello();" + NEWLINE +
				 "    i.world();" + NEWLINE +
				 "  }" + NEWLINE +
				 "}"  + NEWLINE +
				 "public void Internalize1.hello() {" + 
				 NEWLINE +
				 "  System.out.print( \"Hello, \" );" + 
				 NEWLINE +
				 "}" + NEWLINE +
				 "public void runtime.Internalize1.world() {" +
				 NEWLINE +
				 "  System.out.print( \"world!\" );" + 
				 NEWLINE +
				 "}" ) );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( PACKAGE + "." + className );
	    expectSuccess( runRes );
	    assertEquals( "Hello, world!", runRes.stdOut, true );
	} finally {
	    remove( filename );
	}
    }

    //---------------------------------------------------------------------
    // Multiple Dispatch
    //---------------------------------------------------------------------

    /**
     * Should successfully use multiple dispatch on a local internal
     * generic function. */
    public void test_dispatch_1() {
	try {
	    CompilationResults compRes =
		compile( saveAs( "Dispatch1.java",
				 "public class Dispatch1 {" + NEWLINE +
				 "  public static void main( String[] args ) {"
				 + NEWLINE +
				 "    Object o1 = new Object();" + NEWLINE +
				 "    Object o2 = new Dispatch1();" + NEWLINE +
				 "    Dispatch1 d = new Dispatch1();" + 
				 NEWLINE +
				 "    d.m( o1 );" + NEWLINE +
				 "    d.m( o2 );" + NEWLINE +
				 "    d.m( d );" + NEWLINE +
				 "  }" + NEWLINE +
				 "  public void m( Object o ) {" + NEWLINE +
				 "    System.out.print( \"o\" );" + NEWLINE +
				 "  }" + NEWLINE +
				 "  public void m( Object@Dispatch1 d ) {" + 
				 NEWLINE +
				 "    System.out.print( \"d\" );" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" ));
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "Dispatch1" );
	    expectSuccess( runRes );
	    assertEquals( "odd", runRes.stdOut, true );
	} finally {
	    remove( "Dispatch1.java" );
	}
    }

    /**
     * Should successfully use multiple dispatch on a non-local internal
     * generic function. */
    public void test_dispatch_2() {
	try {
	    CompilationResults compRes =
		compile( saveAs( "Dispatch2.java",
				 "public class Dispatch2 {" + NEWLINE +
				 "  public static void main( String[] args ) {"
				 + NEWLINE +
				 "    Object o1 = new Object();" + NEWLINE +
				 "    Object o2 = new Dispatch2('a');" + 
				 NEWLINE +
				 "    Object o3 = new Dispatch2('a');" + 
				 NEWLINE +
				 "    System.out.println( o1.equals(o1) );"
				 + NEWLINE +
				 "    System.out.println( o1.equals(o2) );"
				 + NEWLINE +
				 "    System.out.println( o1.equals(o3) );"
				 + NEWLINE +
				 "    System.out.println( o2.equals(o1) );"
				 + NEWLINE +
				 "    System.out.println( o2.equals(o2) );"
				 + NEWLINE +
				 "    System.out.println( o2.equals(o3) );"
				 + NEWLINE +
				 "    System.out.println( o3.equals(o1) );"
				 + NEWLINE +
				 "    System.out.println( o3.equals(o2) );"
				 + NEWLINE +
				 "    System.out.println( o3.equals(o3) );"
				 + NEWLINE +
				 "  }" + NEWLINE +
				 "  public Dispatch2( char c ) {" + NEWLINE +
				 "    this.c = c;" + NEWLINE +
				 "  }" + NEWLINE +
				 "  private char c;" + NEWLINE +
				 "  public boolean equals( Object@Dispatch2 d ) {" + 
				 NEWLINE +
				 "    return c == d.c;" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" ));
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "Dispatch2" );
	    expectSuccess( runRes );
	    assertEquals( "true" + NEWLINE +
			  "false" + NEWLINE +
			  "false" + NEWLINE +

			  "false" + NEWLINE +
			  "true" + NEWLINE +
			  "true" + NEWLINE +

			  "false" + NEWLINE +
			  "true" + NEWLINE +
			  "true" + NEWLINE, runRes.stdOut, true );
	} finally {
	    remove( "Dispatch2.java" );
	}
    }

    // -------------------------------------------------------------
    // OPEN CLASSES
    // -------------------------------------------------------------
    
    /**
     * Should successfully add open method to non-local class. */ 
    public void test_open_classes_1() {
	try {
	    saveAs( "openClasses1A.java",
		    "" + NEWLINE +
		    "public void Object.openClasses1A() {" + NEWLINE +
		    "  System.out.print( \"o\" );" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "OpenClasses1B.java",
		    "import openClasses1A;" + NEWLINE +
		    "public class OpenClasses1B {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    new Object().openClasses1A();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    CompilationResults compRes =
		compile( "openClasses1A.java OpenClasses1B.java" );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "OpenClasses1B" );
	    expectSuccess( runRes );
	    assertEquals( "o", runRes.stdOut, true );
	} finally {
	    remove( "openClasses1A.java OpenClasses1B.java" );
	}
    }

    /**
     * Should successfully add open methods to non-local classes and
     * use dynamic single dispatch. */
    public void test_open_classes_2() {
	try {
	    saveAs( "openClasses2A.java",
		    "" + NEWLINE +
		    "public void Object.openClasses2A() {" + NEWLINE +
		    "  System.out.print( \"o\" );" + NEWLINE +
		    "}" + NEWLINE +
		    "public void Number.openClasses2A() {" + NEWLINE +
		    "  System.out.print( \"n\" );" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "OpenClasses2B.java",
		    "import openClasses2A;" + NEWLINE +
		    "public class OpenClasses2B {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    new Object().openClasses2A();" + NEWLINE +
		    "    new Integer(1).openClasses2A();" + NEWLINE +
		    "    new Float(1.0f).openClasses2A();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    CompilationResults compRes =
		compile( "openClasses2A.java OpenClasses2B.java" );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "OpenClasses2B" );
	    expectSuccess( runRes );
	    assertEquals( "onn", runRes.stdOut, true );
	} finally {
	    remove( "openClasses2A.java OpenClasses2B.java" );
	}
    }

    /**
     * Should successfully add open methods to non-local classes and
     * use dynamic multiple dispatch. */
    public void test_open_classes_3() {
	try {
	    saveAs( "openClasses3A.java",
		    "" + NEWLINE +
		    "public void Object.openClasses3A( Object o ) {" + NEWLINE +
		    "  System.out.print( \"o\" );" + NEWLINE +
		    "}" + NEWLINE +
		    "public void Number.openClasses3A( Object@Number n ) {" + NEWLINE +
		    "  System.out.print( \"n\" + intValue() + n.intValue() );" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "OpenClasses3B.java",
		    "import openClasses3A;" + NEWLINE +
		    "public class OpenClasses3B {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    new Object().openClasses3A( new Object() );" + NEWLINE +
		    "    new Integer(1).openClasses3A( new Integer(2) );" + NEWLINE +
		    "    new Float(1.0f).openClasses3A( new Object() );" + NEWLINE +
		    "    new Object().openClasses3A( new Integer(2) );" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    CompilationResults compRes =
		compile( "openClasses3A.java OpenClasses3B.java" );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "OpenClasses3B" );
	    expectSuccess( runRes );
	    assertEquals( "on12oo", runRes.stdOut, true );
	} finally {
	    remove( "openClasses3A.java OpenClasses3B.java" );
	}
    }

    /**
     * Should successfully add open methods to non-local classes and
     * use dynamic types for (single) dispatch. */
    public void test_open_classes_4() {
	try {
	    saveAs( "openClasses4A.java",
		    "" + NEWLINE +
		    "public void Object.openClasses4A() {" + NEWLINE +
		    "  System.out.print( \"o\" );" + NEWLINE +
		    "}" + NEWLINE +
		    "public void Number.openClasses4A() {" + NEWLINE +
		    "  System.out.print( \"n\" + intValue() );" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "OpenClasses4B.java",
		    "import openClasses4A;" + NEWLINE +
		    "public class OpenClasses4B {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    Object o = new Object();" + NEWLINE +
		    "    Object i1 = new Integer(1);" + NEWLINE +
		    "    Object i2 = new Integer(2);" + NEWLINE +
		    "    Object f1 = new Float(1.0f);" + NEWLINE +
		    "    o.openClasses4A();" + NEWLINE +
		    "    i1.openClasses4A();" + NEWLINE +
		    "    i2.openClasses4A();" + NEWLINE +
		    "    f1.openClasses4A();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    CompilationResults compRes =
		compile( "openClasses4A.java OpenClasses4B.java" );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "OpenClasses4B" );
	    expectSuccess( runRes );
	    assertEquals( "on1n2n1", runRes.stdOut, true );
	} finally {
	    remove( "openClasses4A.java OpenClasses4B.java" );
	}
    }

    /**
     * Should successfully add open methods to non-local classes and
     * use dynamic types for (multiple) dispatch. */
    public void test_open_classes_5() {
	try {
	    saveAs( "openClasses5A.java",
		    "" + NEWLINE +
		    "public void Object.openClasses5A( Object o ) {" + NEWLINE +
		    "  System.out.print( \"o\" );" + NEWLINE +
		    "}" + NEWLINE +
		    "public void Number.openClasses5A( Object@Number n ) {" + NEWLINE +
		    "  System.out.print( \"n\" + intValue() + n.intValue() );" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "OpenClasses5B.java",
		    "import openClasses5A;" + NEWLINE +
		    "public class OpenClasses5B {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    Object o = new Object();" + NEWLINE +
		    "    Object i1 = new Integer(1);" + NEWLINE +
		    "    Object i2 = new Integer(2);" + NEWLINE +
		    "    Object f1 = new Float(1.0f);" + NEWLINE +
		    "    o.openClasses5A( o );" + NEWLINE +
		    "    i1.openClasses5A( i2 );" + NEWLINE +
		    "    f1.openClasses5A( o );" + NEWLINE +
		    "    o.openClasses5A( i2 );" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    CompilationResults compRes =
		compile( "openClasses5A.java OpenClasses5B.java" );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "OpenClasses5B" );
	    expectSuccess( runRes );
	    assertEquals( "on12oo", runRes.stdOut, true );
	} finally {
	    remove( "openClasses5A.java OpenClasses5B.java" );
	}
    }

    /**
     * Adding open methods to abstract classes (thanks to Todd Millstein). */
    public void test_open_classes_6() {
	try {
	    saveAs( "OpenClasses6A.java",
		    "public abstract class OpenClasses6A {" + NEWLINE +
		    "  public OpenClasses6A() { }" + NEWLINE +
		    "}" );
	    saveAs( "OpenClasses6B.java",
		    "public class OpenClasses6B extends OpenClasses6A {" + NEWLINE +
		    "  public OpenClasses6B() { }" + NEWLINE +
		    "}" );
	    saveAs( "openClasses6C.java",
		    "" + NEWLINE +
		    "public void OpenClasses6A.openClasses6C() {" + NEWLINE +
		    "  System.out.print( \"A\" );" + NEWLINE +
		    "}" + NEWLINE +
		    "public void OpenClasses6B.openClasses6C() {" + NEWLINE +
		    "  System.out.print( \"B\" );" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "OpenClasses6D.java",
		    "import openClasses6C;" + NEWLINE +
		    "public class OpenClasses6D {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    OpenClasses6A a = new OpenClasses6B();" + NEWLINE +
		    "    OpenClasses6B b = new OpenClasses6B();" + NEWLINE +
		    "    a.openClasses6C();" + NEWLINE +
		    "    b.openClasses6C();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    CompilationResults compRes =
		compile( "OpenClasses6A.java OpenClasses6B.java " + 
			 "openClasses6C.java OpenClasses6D.java" );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "OpenClasses6D" );
	    expectSuccess( runRes );
	    assertEquals( "BB", runRes.stdOut, true );
	} finally {
  	    remove( "OpenClasses6A.java OpenClasses6B.java " + 
  		    "openClasses6C.java OpenClasses6D.java" );
	}
    }

    /**
     * Should successfully add open methods to non-local classes and
     * use dynamic single dispatch. */
    public void test_open_classes_7() {
	try {
	    saveAs( "openClasses7A.java",
		    "package test;" + NEWLINE +
		    "" + NEWLINE +
		    "public void Object.openClasses7A() {" + NEWLINE +
		    "  System.out.print( \"o\" );" + NEWLINE +
		    "}" + NEWLINE +
		    "public void Number.openClasses7A() {" + NEWLINE +
		    "  System.out.print( \"n\" );" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "OpenClasses7B.java",
		    "package test;" + NEWLINE +
		    "import test.openClasses7A;" + NEWLINE +
		    "public class OpenClasses7B {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    new Object().openClasses7A();" + NEWLINE +
		    "    new Integer(1).openClasses7A();" + NEWLINE +
		    "    new Float(1.0f).openClasses7A();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    CompilationResults compRes =
		compile( "openClasses7A.java OpenClasses7B.java" );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( "test.OpenClasses7B" );
	    expectSuccess( runRes );
	    assertEquals( "onn", runRes.stdOut, true );
	} finally {
	    remove( "openClasses7A.java OpenClasses7B.java" );
	}
    }

    //---------------------------------------------------------------------
    // Unescaping String literals
    //---------------------------------------------------------------------

    /**
     * Test that octal and unicode escape sequences are properly handled. */
    public void test_unescape_string_1() {
	String className = "OctalUnicodeEscape1";
	String filename = className + ".java";
	try {
	    CompilationResults compRes = 
		compile( saveAs( filename,
				 "public class OctalUnicodeEscape1 {" +
				 NEWLINE +
				 "  public static void main( String[] args ) {"
				 + NEWLINE +
				 "    String s = \"\\777\\3777\\368\\uffffudf\\04\";" +
				 NEWLINE + 
				 "    char[] c = s.toCharArray();" + NEWLINE +
				 "    for (int i = 0; i < c.length; i++) {" +
				 NEWLINE +
				 "        System.out.print((int) c[i]);" +
				 NEWLINE +
				 "        System.out.print(\" \");" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" ) );
	    expectSuccess( compRes );
	    RuntimeResults runRes = run( className );
	    expectSuccess( runRes );
	    assertEquals( "63 55 255 55 30 56 65535 117 100 102 4 ",
			  runRes.stdOut, true );
	} finally {
	    remove( filename );
	}
    }
   
    //---------------------------------------------------------------------
    // PRIVATE HELPERS
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // COMMON FIXTURE
    //---------------------------------------------------------------------

    // public void setUp() { }

    // public void tearDown() { }

    private static final String PACKAGE = "runtime";
    private static final String PACKAGE_DECL = 
	"package " + PACKAGE + ";" + NEWLINE;
}
