/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestMain_Typecheck.java,v 1.28 2006/06/27 18:45:04 kiniry Exp $
 */

package org.multijava.mjc;

import org.multijava.util.classfile.ClassPath;

import org.multijava.util.testing.TestCase;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;

import java.util.StringTokenizer;

public class TestMain_Typecheck extends FunctionalTestSuite.TestCase {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------
    
    public TestMain_Typecheck( String name ) {
	super( name );
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // Some simple self-tests
    //---------------------------------------------------------------------

    public void test_split_method1() {
	String[] results = split( "SplitOne.java" );
	assertEquals( "SplitOne.java", results[0], true );
	results = split( "SplitOnea.java SplitOneb.java" );
	assertEquals( "SplitOnea.java", results[0], true );
	assertEquals( "SplitOneb.java", results[1], true );
    }

    /**
     * Should successfully compile the (very simple) source code.  */
    public void test_simple_class_1() {
	try {
	    CompilationResults results = 
		compile( saveAs( "SimpleClass1.java",
				 "public class SimpleClass1 { }" ) );
	    assertEquals( "", results.errorOut, true );
	    expectSuccess( results );
	} finally {
	    remove( "SimpleClass1.java" );
	}
    }

    /**
     * Should fail on the malformed class declaration.  */
    public void test_simple_class_2() {
	try {
	    CompilationResults results = 
		compile( saveAs( "SimpleClass2.java",
				 "public class { }" ) );
	    expectFailure( results, 
			    
			   "File \"SimpleClass2.java\", line 1, character 15 error: " + 
			   "Syntax error: expecting an identifier, " +
			   "found '{'" + NEWLINE );
	} finally {
	    remove( "SimpleClass2.java" );
	}
    }

    /**
     * Should fail on the malformed class declaration.  */
    public void test_simple_class_3() {
	try {
	    CompilationResults results = 
		compile( saveAs( "SimpleClass3.java",
				 "public class SimpleClass3;" ) );
	    expectFailure( results, 
			    
			   "File \"SimpleClass3.java\", line 1, character 27 error: Syntax error: unexpected token: ;" + NEWLINE );
	} finally {
	    remove( "SimpleClass3.java" );
	}
    }

    // -------------------------------------------------------------
    // constructors
    // -------------------------------------------------------------

    public void test_constructor_1() {
	try {
	    CompilationResults results = 
		compile( saveAs( "Constructor1.java",
				 "public class Constructor1 {" + NEWLINE +
				 "  public void m() {" + NEWLINE +
				 "    super();" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,
			    "File \"Constructor1.java\", line 3, character 12 error: Can only use explicit constructor invocation from within a constructor body [JLS2 8.8]" + NEWLINE );
	} finally {
	    remove( "Constructor1.java" );
	} // end of finally
	
    }

    //---------------------------------------------------------------------
    // Tests on implementation-side checks of pleomorphic methods.
    //---------------------------------------------------------------------

    /**
     * The pleomorphic method should be allowed since both generic
     * functions are internal. */
    public void test_pleomorphic_1() {
	try {
	    saveAs( "Pleomorphic1A.java",
		    "public class Pleomorphic1A {" + NEWLINE +
		    "    public void m() { }" + NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic1B.java",
		    "public interface Pleomorphic1B {" + NEWLINE + 
		    "    public void m();" + NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic1C.java",
		    "public class Pleomorphic1C extends Pleomorphic1A " + 
		    NEWLINE +
		    "    implements Pleomorphic1B" + NEWLINE +
		    "{" + NEWLINE +
		    "    public void m() { " + NEWLINE +
		    "	     super.m();" + NEWLINE +
		    "    }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
		compile( "Pleomorphic1A.java Pleomorphic1B.java " + 
			 "Pleomorphic1C.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Pleomorphic1A.java Pleomorphic1B.java " + 
		    "Pleomorphic1C.java" );
	}
    }

    /**
     * The pleomorphic method should be allowed by static
     * typechecking. */
    public void test_pleomorphic_2() {
	try {
	    saveAs( "Pleomorphic2A.java",
		    "public class Pleomorphic2A {" + NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic2B.java",
		    "public interface Pleomorphic2B {" + NEWLINE + 
		    "    public void pleomorphic2_method();" + NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic2C.java",
		    "" + NEWLINE +
		    "public class Pleomorphic2C extends Pleomorphic2A " + 
		    NEWLINE +
		    "    implements Pleomorphic2B" + NEWLINE +
		    "{" + NEWLINE +
		    "    public void pleomorphic2_method() {" 
		    + NEWLINE +
		    "	     super.pleomorphic2_method();" + NEWLINE +
		    "    }" + NEWLINE +
		    "}" );
	    saveAs( "pleomorphic2_method.java",
		    "public void Pleomorphic2A.pleomorphic2_method() { }" );
	    CompilationResults results = 
		compile( "Pleomorphic2A.java Pleomorphic2B.java " + 
			 "Pleomorphic2C.java pleomorphic2_method.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Pleomorphic2A.java Pleomorphic2B.java " + 
		    "Pleomorphic2C.java pleomorphic2_method.java" );
	}
    }

    /**
     * The pleomorphic method should be disallowed by static
     * typechecking since both top methods are concrete. */
    public void test_pleomorphic_3() {
	try {
	    saveAs( "Pleomorphic3A.java",
		    "public class Pleomorphic3A {" + NEWLINE +
		    "    public void pleomorphic3_method() { }" + NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic3B.java",
		    "" + NEWLINE +
		    "public class Pleomorphic3B extends Pleomorphic3A" + 
		    NEWLINE +
		    "{" + NEWLINE +
		    "    public void pleomorphic3_method() { // bad!" 
		    + NEWLINE +
		    "	     super.pleomorphic3_method();" + NEWLINE +
		    "    }" + NEWLINE +
		    "}" );
	    saveAs( "pleomorphic3_method.java",
		    "public void Pleomorphic3A.pleomorphic3_method() { }" );
	    CompilationResults results = 
		compile( "Pleomorphic3A.java Pleomorphic3B.java " + 
			 "pleomorphic3_method.java" );
	    expectFailure( results, 
			    "File \"Pleomorphic3B.java\", line 4, character 11 error: The " +
			   "top concrete method of the generic function is " +
			   "ambiguous in this context between (at least) " +
			   "Pleomorphic3A.pleomorphic3_method(  ) and " +
			   "Pleomorphic3A.pleomorphic3_method(  ). [MultiJava]"
			   + NEWLINE );
	} finally {
	    remove( "Pleomorphic3A.java Pleomorphic3B.java " + 
		    "pleomorphic3_method.java" );
	}
    }

    /**
     * The pleomorphic method should be disallowed by static
     * typechecking since both top methods are concrete. */
    public void test_pleomorphic_4() {
	try {
	    saveAs( "pleomorphic4Func.java",
		    "package p2;" + NEWLINE +
		    "public void p1.Pleomorphic4A.pleomorphic4Func() { }" );
	    saveAs( "Pleomorphic4A.java",
		    "package p1;" + NEWLINE +
		    "public class Pleomorphic4A { }" );
	    saveAs( "Pleomorphic4B.java",
		    "package p1;" + NEWLINE +
		    "public class Pleomorphic4B extends Pleomorphic4A {" + 
		    NEWLINE +
		    "  public void pleomorphic4Func() { }" + NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic4C.java",
		    "package p1;" + NEWLINE +
		    "import p2.pleomorphic4Func;" + NEWLINE +
		    "public class Pleomorphic4C extends Pleomorphic4B {" +
		    NEWLINE +
		    "  public void pleomorphic4Func() { } // internal or external top meth?" 
		    + NEWLINE +
		    "}" );
	    CompilationResults results = 
		compile( "Pleomorphic4A.java Pleomorphic4B.java " + 
			 "Pleomorphic4C.java pleomorphic4Func.java" );
	    expectFailure( results,  "File \"Pleomorphic4C.java\", line 4, character 9 error: The top concrete method of the generic function is ambiguous in this context between (at least) p1.Pleomorphic4B.pleomorphic4Func(  ) and p1.Pleomorphic4A.pleomorphic4Func(  ).  Should the compilation unit declaring p1.Pleomorphic4B.pleomorphic4Func(  ) import the generic function declaring p1.Pleomorphic4A.pleomorphic4Func(  )? [MultiJava]" + NEWLINE );
	} finally {
	    remove( "Pleomorphic4A.java Pleomorphic4B.java " + 
		    "Pleomorphic4C.java pleomorphic4Func.java" );
	}
    }

    /**
     * The pleomorphic method should be disallowed by static
     * typechecking since both top methods are concrete. */
    public void test_pleomorphic_5() {
	try {
	    saveAs( "pleomorphic5Func.java",
		    "package p2;" + NEWLINE +
		    "public void p1.Pleomorphic5B.pleomorphic5Func() { }" );
	    saveAs( "Pleomorphic5A.java",
		    "package p1;" + NEWLINE +
		    "public class Pleomorphic5A {" + 
		    NEWLINE +
		    "  public void pleomorphic5Func() { }" + NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic5B.java",
		    "package p1;" + NEWLINE +
		    "public class Pleomorphic5B extends Pleomorphic5A {" + 
		    NEWLINE +
		    "}" );
	    saveAs( "Pleomorphic5C.java",
		    "package p1;" + NEWLINE +
		    "import p2.pleomorphic5Func;" + NEWLINE +
		    "public class Pleomorphic5C extends Pleomorphic5B {" +
		    NEWLINE +
		    "  public void pleomorphic5Func() { } // internal or external top meth?" 
		    + NEWLINE +
		    "}" );
	    CompilationResults results = 
		compile( "Pleomorphic5A.java Pleomorphic5B.java " + 
			 "Pleomorphic5C.java pleomorphic5Func.java" );
	    expectFailure( results,  "File \"pleomorphic5Func.java\", line 2, character 12 error: Top-level method pleomorphic5Func specializes the non-local method p1.Pleomorphic5A.pleomorphic5Func(  ).  Top-level methods must only specialize local methods. [MultiJava R3]" + NEWLINE + "File \"pleomorphic5Func.java\", line 2, character 12 error: Top-level method pleomorphic5Func specializes the regular method p1.Pleomorphic5A.pleomorphic5Func(  ).  Top-level methods must only specialize other top-level methods. [MultiJava R3]" + NEWLINE);
	} finally {
	    remove( "Pleomorphic5A.java Pleomorphic5B.java " + 
		    "Pleomorphic5C.java pleomorphic5Func.java" );
	}
    }

    //---------------------------------------------------------------------
    // Tests for good and bad explicit specializers.
    //---------------------------------------------------------------------

    /**
     * Typechecker should allow specialization of Object to any
     * (non-primitive, non-interface) array type.  */
    public void test_goodSpec_1() {
	try {
	    saveAs( "GoodSpec1.java",
		    "public class GoodSpec1 {" + NEWLINE +
		    "    public void m( Object o ) { }" + NEWLINE +
		    "    public void m( Object@Number[] o ) { }" + NEWLINE +
		    "}" );
	    CompilationResults results = compile( "GoodSpec1.java" );
	    expectSuccess( results );
	} finally {
	    remove( "GoodSpec1.java" );
	}
    }

    /**
     * interface types can be specialized */
    public void test_goodSpec_3() {
	try {
	    saveAs( "GoodSpec3.java",
		    "import java.util.Set;" + NEWLINE +
		    "import java.util.HashSet;" + NEWLINE +
		    "public class GoodSpec3 {" + NEWLINE +
		    "	 public void foo( Set s ) { }" + NEWLINE +
		    "	 public void foo( Set@HashSet hs ) { }" + NEWLINE +
		    "}" );
	    CompilationResults result = compile( "GoodSpec3.java" );
	    expectSuccess( result );
	} finally {
	    remove( "GoodSpec3.java" );
	}
    }

    /**
     * Simplest specializer test. */
    public void test_goodSpec_4() {
	try {
	    saveAs( "GoodSpec4.java",
		    "public class GoodSpec4 {" + NEWLINE +
		    "	 public void m( Object o ) { }" + NEWLINE +
		    "	 public void m( Object@String s ) { }" + NEWLINE +
		    "}" );
	    CompilationResults result = compile( "GoodSpec4.java" );
	    expectSuccess( result );
	} finally {
	    remove( "GoodSpec4.java" );
	}
    }

    /**
     * specializer must be proper subtype */
    public void test_badSpec_1() {
	try{
	    saveAs( "BadSpec1.java",
		    "public class BadSpec1 {" + NEWLINE +
		    "	 public void foo( String s ) { }" + NEWLINE +
		    "	 public void foo( String@Integer i ) { }" + NEWLINE +
		    "}" );
	    CompilationResults result = compile( "BadSpec1.java" );
	    expectFailure( result,  "File \"BadSpec1.java\", line 3, character 33 error: The type \"java.lang.Integer\" is not assignable to \"java.lang.String\" and so is not a valid type specializer. [CLCM 00, 4.1.3.1]" + NEWLINE + "File \"BadSpec1.java\", line 1, character 7 error: No applicable method found for generic function BadSpec1.foo( java.lang.String@java.lang.Integer ) on visible argument tuple <[ BadSpec1, java.lang.String ]>. [MultiJava]" + NEWLINE);
	} finally {
	    remove( "BadSpec1.java" );
	}
    }

    /**
     * cannot specialize to non-reference type */
    public void test_badSpec_3() {
	try{
	    saveAs( "BadSpec3.java",
		    "public class BadSpec3 {" + NEWLINE +
		    "	 public void foo( double d ) { }" + NEWLINE +
		    "	 public void foo( double@float f ) { }" + NEWLINE +
		    "}" );
	    CompilationResults result = compile( "BadSpec3.java" );
	    expectFailure( result,  "File \"BadSpec3.java\", line 3, character 33 error: Only class types and arrays of class types may be explicit type specializers.  \"float\" cannot be a type specializer. [CLCM 00, 4.1.3.1]" + NEWLINE );
	} finally {
	    remove( "BadSpec3.java" );
	}
    }

    /**
     * cannot specialize to interface type */
    public void test_badSpec_5() {
	try{
	    saveAs( "BadSpec5.java",
		    "public class BadSpec5 {" + NEWLINE +
		    "	 public void foo( Object o ) { }" + NEWLINE +
		    "	 public void foo( Object@Cloneable c ) { }" + NEWLINE +
		    "}" );
	    CompilationResults result = compile( "BadSpec5.java" );
	    expectFailure( result,  "File \"BadSpec5.java\", line 3, character 33 error: Only class types and arrays of class types may be explicit type specializers.  \"java.lang.Cloneable\" cannot be a type specializer. [CLCM 00, 4.1.3.1]" + NEWLINE );
	} finally {
	    remove( "BadSpec5.java" );
	}
    }

    /**
     * specializer must be proper subtype */
    public void test_badSpec_6() {
	try{
	    saveAs( "BadSpec6.java",
		    "public class BadSpec6 {" + NEWLINE +
		    "	 public void foo( String s ) { }" + NEWLINE +
		    "	 public void foo( String@Object s ) { }" + NEWLINE +
		    "}" );
	    CompilationResults result = compile( "BadSpec6.java" );
	    expectFailure( result,  "File \"BadSpec6.java\", line 3, character 33 error: The type \"java.lang.Object\" is not assignable to \"java.lang.String\" and so is not a valid type specializer. [CLCM 00, 4.1.3.1]" + NEWLINE );
	} finally {
	    remove( "BadSpec6.java" );
	}
    }

    /**
     * specializer must be proper subtype */
    public void test_badSpec_7() {
	try{
	    saveAs( "BadSpec7.java",
		    "public class BadSpec7 {" + NEWLINE +
		    "	 public void foo( String s ) { }" + NEWLINE +
		    "	 public void foo( String@String s ) { }" + NEWLINE +
		    "}" );
	    CompilationResults result = compile( "BadSpec7.java" );
	    expectFailure( result,  "File \"BadSpec7.java\", line 3, character 33 error: A formal parameter may not have the same static and dynamic dispatch type (see parameter with type \"java.lang.String\") [CLCM 00, 4.1.3.1]" + NEWLINE );
	} finally {
	    remove( "BadSpec7.java" );
	}
    }

    /**
     * Typechecker should disallow specialization to an array of primitive
     * types.  */
    public void test_badSpec_8() {
	try {
	    saveAs( "BadSpec8.java",
		    "public class BadSpec8 {" + NEWLINE +
		    "    public void m( Object[] o ) { }" + NEWLINE +
		    "    public void m( Object[]@int[] o ) { }" + NEWLINE +
		    "}" );
	    CompilationResults results = compile( "BadSpec8.java" );
	    expectFailure( results,  "File \"BadSpec8.java\", line 3, character 26 error: Only class types and arrays of class types may be explicit type specializers.  \"int[]\" cannot be a type specializer. [CLCM 00, 4.1.3.1]" + NEWLINE );
	} finally {
	    remove( "BadSpec8.java" );
	}
    }

    /**
     * Typechecker should disallow specialization to an array of interface
     * types.  */
    public void test_badSpec_9() {
	try {
	    saveAs( "BadSpec9.java",
		    "public class BadSpec9 {" + NEWLINE +
		    "    public void m( Object[] o ) { }" + NEWLINE +
		    "    public void m( Object[]@Cloneable[] o ) { }" + 
		    NEWLINE +
		    "}" );
	    CompilationResults results = compile( "BadSpec9.java" );
	    expectFailure( results,  "File \"BadSpec9.java\", line 3, character 26 error: Only class types and arrays of class types may be explicit type specializers.  \"java.lang.Cloneable[]\" cannot be a type specializer. [CLCM 00, 4.1.3.1]" + NEWLINE);
	} finally {
	    remove( "BadSpec9.java" );
	}
    }

    //---------------------------------------------------------------------
    // Tests for detection of ambiguous generic function declarations.
    //---------------------------------------------------------------------

    /**
     * Should fail since augmenting method induces an ambiguity.  */
    public void test_ambiguity_1() {
	try {
	    saveAs( "Ambiguity1.java",
		    "import equals;" + NEWLINE +
		    "public class Ambiguity1 {" + NEWLINE +
		    "    public boolean m( Object o ) {" + NEWLINE +
		    "        return o.equals( o );" + NEWLINE +
		    "    }" + NEWLINE +
		    "}" );
	    saveAs( "equals.java",
		    "public boolean Object.equals( Object other ) {" + 
		    NEWLINE +
		    "    return true;" + NEWLINE +
		    "}" );

	    CompilationResults results = 
		compile( "Ambiguity1.java equals.java" );
	    expectFailure( results, 
			    "File \"Ambiguity1.java\", line 4, character 30 error: Call " + 
			   "of method is ambiguous between (at least) " + 
			   "java.lang.Object.equals( java.lang.Object ) " + 
			   "and java.lang.Object.equals( java.lang.Object " + 
			   "). [JLS 15.11]" + NEWLINE );
	} finally {
	    remove( "Ambiguity1.java equals.java" );
	}
    }

    /**
     * Typechecker should detect ambiguity on
     * < Ambiguity2,Number[],Number[] >. */
    public void test_ambiguity_2() {
	try {
	    saveAs( "Ambiguity2.java",
		    "public class Ambiguity2 {" + NEWLINE +
		    "    public void m( Object oA1, Object oA2 ) { }" + 
		    NEWLINE +
		    "    public void m( Object@Number[] iA, Object oA ) { }" + 
		    NEWLINE +
		    "    public void m( Object oA, Object@Number[] iA ) { }" + 
		    NEWLINE +
		    "}" );
	    CompilationResults results = compile( "Ambiguity2.java" );
	    expectFailure( results,  "File \"Ambiguity2.java\", line 1, character 7 error: The generic function \"m\" is ambiguous on the argument tuple <[ Ambiguity2, java.lang.Number[], java.lang.Number[] ]> between (at least) methods Ambiguity2.m( java.lang.Object@java.lang.Number[], java.lang.Object ) and Ambiguity2.m( java.lang.Object, java.lang.Object@java.lang.Number[] ). [MultiJava]" + NEWLINE);
	} finally {
	    remove( "Ambiguity2.java" );
	}
    }

    /**
     * Typechecker should not detect ambiguity on overloaded methods
     * because one is private. */
    public void test_ambiguity_3() {
	try {
	    saveAs( "Ambiguity3A.java",
		    "public class Ambiguity3A {" + NEWLINE +
		    "  public boolean equals( Object o ) {" + NEWLINE +
		    "    return equals( (Ambiguity3A) o );" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public boolean equals( Ambiguity3A o ) {" + NEWLINE +
		    "    return this==o;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    saveAs( "Ambiguity3B.java",
		    "public class Ambiguity3B {" + NEWLINE +
		    "  public void m() {" + NEWLINE +
		    "    System.out.println( new Ambiguity3A()." + NEWLINE +
		    "      equals( new Ambiguity3A() ));" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
		compile( "Ambiguity3A.java Ambiguity3B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Ambiguity3A.java Ambiguity3B.java" );
	}
    }

    /**
     * Typechecker should detect ambiguity on
     * < Ambiguity4B,Ambiguity4B > because Ambiguity4B has types
     * Ambiguity4A and Comparable. */
    public void test_ambiguity_4() {
	try {
	    saveAs( "Ambiguity4A.java",
		    "public class Ambiguity4A {" + NEWLINE +
		    "  public void m( Ambiguity4A a ) { }" + 
		    NEWLINE +
		    "  public void m( Comparable c ) { }" + 
		    NEWLINE +
		    "}" );
	    saveAs( "Ambiguity4B.java",
		    "public class Ambiguity4B extends Ambiguity4A" + NEWLINE +
		    "  implements Comparable" + NEWLINE +
		    "{" + NEWLINE +
		    "  public int compareTo( Object o ) {" + NEWLINE + 
		    "    return 0;" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public void foo() {" + NEWLINE +
		    "    this.m( this );" + NEWLINE + 
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
		compile( "Ambiguity4A.java Ambiguity4B.java" );
	    expectFailure( results,  "File \"Ambiguity4B.java\", line 8, character 19 error: Call of method is ambiguous between (at least) Ambiguity4A.m( Ambiguity4A ) and Ambiguity4A.m( java.lang.Comparable ). [JLS 15.11]" + NEWLINE);
	} finally {
	    remove( "Ambiguity4A.java Ambiguity4B.java" );
	}
    }

    /**
     * Typechecker should detect ambiguity on < Ambiguity5B,Ambiguity5E2 >
     */
    public void test_ambiguity_5() {
	try {
	    saveAs( "Ambiguity5A.java",
		    "class Ambiguity5E1 {}" + NEWLINE +
		    "class Ambiguity5E2 extends Ambiguity5E1 {}" + NEWLINE +
		    "public class Ambiguity5A {" + NEWLINE +
		    "  public void m(Ambiguity5E1 a) {}" + NEWLINE +
		    "  public void m(Ambiguity5E1@Ambiguity5E2 a) {}"
		    + NEWLINE + "}" );
	    saveAs( "Ambiguity5B.java",
		    "public class Ambiguity5B extends Ambiguity5A {"
		    + NEWLINE +
		    "  public void m(Ambiguity5E1 a) {}" + NEWLINE +
		    "}" );
	    CompilationResults results = 
		compile( "Ambiguity5A.java Ambiguity5B.java" );
	    expectFailure( results,  "File \"Ambiguity5B.java\", line 1, character 7 error: The generic function \"m\" is ambiguous on the argument tuple <[ Ambiguity5B, Ambiguity5E2 ]> between (at least) methods Ambiguity5B.m( Ambiguity5E1 ) and Ambiguity5A.m( Ambiguity5E1@Ambiguity5E2 ). [MultiJava]" + NEWLINE);
	} finally {
	    remove( "Ambiguity5A.java Ambiguity5B.java" );
	}
    }
	    
    // -------------------------------------------------------------
    // Tests for incompleteness of generic functions
    // -------------------------------------------------------------

    //  !HERE!note6
//      public void test_incomplete_1() {
//  	try {
//  	    saveAs( "Incomplete1A.java",
//  		    "public class Incomplete1A {" + NEWLINE +
//  		    "}" );
//  	    saveAs( "incomplete1B.java",
//  		    "public void Incomplete1A.incomplete1B() {" + NEWLINE +
//  		    "}" );
//  	    saveAs( "Incomplete1C.java",
//  		    "import incomplete1B;" + NEWLINE +
//  		    "public abstract class Incomplete1C extends Incomplete1A {"
//  		    + NEWLINE +
//  		    "  public abstract void incomplete1B();" + NEWLINE +
//  		    "}" );
//  	    CompilationResults results = 
//                  compile( "Incomplete1A.java incomplete1B.java " + 
//  			 "Incomplete1C.java" );
//  	    expectSuccess( results );
//  	} finally {
//  	    remove( "Incomplete1A.java incomplete1B.java Incomplete1C.java" );
//  	}
//      }

    //---------------------------------------------------------------------
    // Tests for the typechecking of catch clauses.
    //---------------------------------------------------------------------

    /**
     * Typechecker should allow properly ordered catches on exceptions
     * from the same hierarchy.  */
    public void test_catch_1() {
	try {
	    saveAs( "Catch1A.java",
		    "public class Catch1A {" + NEWLINE +
		    "    public void m() { " + NEWLINE +
                    "        try {" + NEWLINE +
                    "            n();" + NEWLINE +
                    "        } catch( Catch1C c ) {" + NEWLINE +
                    "        } catch( Catch1B b ) {" + NEWLINE +
                    "        }" + NEWLINE +
                    "    }" + NEWLINE +
		    "    public void n() throws Catch1B {" +NEWLINE+
                    "        throw new Catch1C();" + NEWLINE +
                    "    }" + NEWLINE +
		    "}" );
            saveAs( "Catch1B.java",
                    "public class Catch1B extends Exception {}" );
            saveAs( "Catch1C.java",
                    "public class Catch1C extends Catch1B {}" );
	    CompilationResults results = 
                compile( "Catch1A.java Catch1B.java Catch1C.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Catch1A.java Catch1B.java Catch1C.java" );
	}
    }

    /**
     * Typechecker should give a warning because catch for Catch2B is
     * unreachable.  */
    public void test_catch_2() {
	try {
	    saveAs( "Catch2A.java",
		    "public class Catch2A {" + NEWLINE +
		    "    public void m() { " + NEWLINE +
                    "        try {" + NEWLINE +
                    "            n();" + NEWLINE +
                    "        } catch( Catch2C c ) {" + NEWLINE +
                    "        } catch( Catch2B b ) {" + NEWLINE +
                    "        }" + NEWLINE +
                    "    }" + NEWLINE +
		    "    public void n() throws Catch2C {" +NEWLINE+
                    "        throw new Catch2C();" + NEWLINE +
                    "    }" + NEWLINE +
		    "}" );
            saveAs( "Catch2B.java",
                    "public class Catch2B extends Exception {}" );
            saveAs( "Catch2C.java",
                    "public class Catch2C extends Catch2B {}" );
	    CompilationResults results = 
                compile( "-w1 Catch2A.java Catch2B.java Catch2C.java" );
	    expectSuccess( results );
	    assertEquals(  "File \"Catch2A.java\", line 6, character 16 caution: Catch clause is unreachable since it catches a proper supertype of a checked exception that is caught earlier" + NEWLINE, results.errorOut, true );
	} finally {
	    remove( "Catch2A.java Catch2B.java Catch2C.java" );
	}
    }

    /**
     * Typechecker should signal error because second catch for
     * RuntimeException is unreachable.  */
    public void test_catch_3() {
	try {
	    saveAs( "Catch3.java",
		    "public class Catch3 {" + NEWLINE +
		    "    public void m() { " + NEWLINE +
                    "        try {" + NEWLINE +
                    "            n();" + NEWLINE +
                    "        } catch( RuntimeException r ) {" + NEWLINE +
                    "        } catch( RuntimeException r ) {" + NEWLINE +
                    "        }" + NEWLINE +
                    "    }" + NEWLINE +
		    "    public void n() { }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "Catch3.java" );
	    expectFailure( results,  "File \"Catch3.java\", line 6, character 16 error: Catch clause is unreachable since it catches an exception that is the same as or a subtype of an exception caught earlier [JLS2 14.20]" + NEWLINE );
	} finally {
	    remove( "Catch3.java" );
	}
    }

    /**
     * Typechecker should signal error because the catch for
     * CloneNotSupportedException is unreachable.  */
    public void test_catch_4() {
	try {
	    saveAs( "Catch4.java",
		    "public class Catch4 {" + NEWLINE +
		    "    public void m() { " + NEWLINE +
                    "        try {" + NEWLINE +
                    "            n();" + NEWLINE +
                    "        } catch( CloneNotSupportedException r ) {" + 
		    NEWLINE +
                    "        }" + NEWLINE +
                    "    }" + NEWLINE +
		    "    public void n() { }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "Catch4.java" );
	    expectFailure( results,  "File \"Catch4.java\", line 5, character 16 error: Catch clause is unreachable since no exception of the type \"java.lang.CloneNotSupportedException\" or a subtype can be thrown [JLS2 14.20]" + NEWLINE );
	} finally {
	    remove( "Catch4.java" );
	}
    }

    //---------------------------------------------------------------------
    // Tests for implicit and explicit use of "this" references within
    // augmenting methods.
    //---------------------------------------------------------------------

    /**
     * Regular methods of the augmented object can be accessed with
     * or without an explicit "this".  */
    public void test_augmenting_this_1() {
	try {
	    saveAs( "augmentingThis1.java",
		    "public int Number.augmentingThis1() {" + NEWLINE +
		    "	 return this.intValue();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis1.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis1.java" );
	}
	try {
	    saveAs( "augmentingThis1.java",
		    "public int Number.augmentingThis1() {" + NEWLINE +
		    "	 return intValue();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis1.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis1.java" );
	}
    }
    
    /**
     * Augmenting methods of the augmented object can be accessed with
     * or without an explicit "this".  */
    public void test_augmenting_this_2() {
	try {
	    saveAs( "augmentingThis2.java",
		    "" + NEWLINE +
		    "public int Number.augmentingThis2() {" + NEWLINE +
		    "	 return this.helper();" + NEWLINE +
		    "}" + NEWLINE +
		    "private int Number.helper() {" + NEWLINE +
		    "	 return this.intValue();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis2.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis2.java" );
	}
	try {
	    saveAs( "augmentingThis2.java",
		    "" + NEWLINE +
		    "public int Number.augmentingThis2() {" + NEWLINE +
		    "	 return helper();" + NEWLINE +
		    "}" + NEWLINE +
		    "private int Number.helper() {" + NEWLINE +
		    "	 return intValue();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis2.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis2.java" );
	}
    }
    
    /**
     * Methods of the receiver class of an augmenting method can be
     * accessed from within an anonymous class of the augmenting
     * method, with or without an explicit this.   */
    public void test_augmenting_this_3() {
	try {
	    saveAs( "augmentingThis3.java",
		    "public Cloneable Number.augmentingThis3() {" + NEWLINE +
		    "  return new Cloneable() {" + NEWLINE +
		    "	   public Object clone() { return this; }" + NEWLINE +
		    "	   public int hashCode() { " + NEWLINE +
		    "	     return Number.this.intValue(); " + NEWLINE +
		    "	   }" + NEWLINE +
		    "	 };" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis3.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis3.java" );
	}
	try {
	    saveAs( "augmentingThis3.java",
		    "public Cloneable Number.augmentingThis3() {" + NEWLINE +
		    "  return new Cloneable() {" + NEWLINE +
		    "	   public Object clone() { return this; }" + NEWLINE +
		    "	   public int hashCode() { " + NEWLINE +
		    "	     return intValue(); " + NEWLINE +
		    "	   }" + NEWLINE +
		    "	 };" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis3.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis3.java" );
	}
    }
    
    /**
     * Regular fields of the augmented object can be accessed with
     * or without an explicit "this".  */
    public void test_augmenting_this_4() {
	try {
	    saveAs( "AugmentingThis4A.java",
		    "public class AugmentingThis4A {" + NEWLINE +
		    "  int value = 4;" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis4B.java",
		    "public int AugmentingThis4A.augmentingThis4B() {" + 
		    NEWLINE +
		    "  return this.value;" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis4A.java augmentingThis4B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis4A.java augmentingThis4B.java" );
	}
	try {
	    saveAs( "AugmentingThis4A.java",
		    "public class AugmentingThis4A {" + NEWLINE +
		    "  int value = 4;" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis4B.java",
		    "public int AugmentingThis4A.augmentingThis4B() {" + 
		    NEWLINE +
		    "  return value;" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis4A.java augmentingThis4B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis4A.java augmentingThis4B.java" );
	}
    }
    
    /**
     * Fields of the receiver class of an augmenting method can be
     * accessed from within an anonymous class of the augmenting
     * method, with or without an explicit this.  */
    public void test_augmenting_this_5() {
	try {
	    saveAs( "AugmentingThis5A.java",
		    "public class AugmentingThis5A {" + NEWLINE +
		    "  int value = 5;" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis5B.java",
		    "public Cloneable AugmentingThis5A.augmentingThis5B() {" +
		    NEWLINE +
		    "  return new Cloneable() {" + NEWLINE +
		    "	   public Object clone() { return this; }" + NEWLINE +
		    "	   public int hashCode() { " + NEWLINE +
		    "	     return AugmentingThis5A.this.value; " + NEWLINE +
		    "	   }" + NEWLINE +
		    "	 };" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis5A.java augmentingThis5B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis5A.java augmentingThis5B.java" );
	}
	try {
	    saveAs( "AugmentingThis5A.java",
		    "public class AugmentingThis5A {" + NEWLINE +
		    "  int value = 5;" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis5B.java",
		    "public Cloneable AugmentingThis5A.augmentingThis5B() {" +
		    NEWLINE +
		    "  return new Cloneable() {" + NEWLINE +
		    "	   public Object clone() { return this; }" + NEWLINE +
		    "	   public int hashCode() { " + NEWLINE +
		    "	     return value; " + NEWLINE +
		    "	   }" + NEWLINE +
		    "	 };" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis5A.java augmentingThis5B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis5A.java augmentingThis5B.java" );
	}
    }
    
    /**
     * Static inner classes of the augmented object can be referenced with
     * or without the augmented object's class name.  */
    public void test_augmenting_this_6() {
	try {
	    saveAs( "AugmentingThis6A.java",
		    "public class AugmentingThis6A {" + NEWLINE +
		    "  public static class Inner { }" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis6B.java",
		    "public AugmentingThis6A.Inner " + NEWLINE + 
		    "    AugmentingThis6A.augmentingThis6B() {" + NEWLINE +
		    "  return new AugmentingThis6A.Inner();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis6A.java augmentingThis6B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis6A.java augmentingThis6B.java" );
	}
	try {
	    saveAs( "AugmentingThis6A.java",
		    "public class AugmentingThis6A {" + NEWLINE +
		    "  public static class Inner { }" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis6B.java", // this return type is interesting
		    "public AugmentingThis6A.Inner " + NEWLINE + 
		    "    AugmentingThis6A.augmentingThis6B() {" + NEWLINE +
		    "  return new Inner();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis6A.java augmentingThis6B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis6A.java augmentingThis6B.java" );
	}
    }
    
    /**
     * Inner classes of the augmented object can be instantiated with
     * or without an explicit this.  */
    public void test_augmenting_this_7() {
	try {
	    saveAs( "AugmentingThis7A.java",
		    "public class AugmentingThis7A {" + NEWLINE +
		    "  public class Inner { }" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis7B.java",
		    "public AugmentingThis7A.Inner " + NEWLINE + 
		    "    AugmentingThis7A.augmentingThis7B() {" + NEWLINE +
		    "  return this.new Inner();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis7A.java augmentingThis7B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis7A.java augmentingThis7B.java" );
	}
	try {
	    saveAs( "AugmentingThis7A.java",
		    "public class AugmentingThis7A {" + NEWLINE +
		    "  public class Inner { }" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis7B.java",
		    "public AugmentingThis7A.Inner " + NEWLINE + 
		    "    AugmentingThis7A.augmentingThis7B() {" + NEWLINE +
		    "  return new Inner();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis7A.java augmentingThis7B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis7A.java augmentingThis7B.java" );
	}
    }
    
    /**
     * Methods of the receiver class of an augmenting method can be
     * accessed from within a class declaration statement of the
     * augmenting method, with or without an explicit this.  */
    public void test_augmenting_this_8() {
	try {
	    saveAs( "augmentingThis8.java",
		    "public Cloneable Number.augmentingThis8() {" + NEWLINE +
		    "  class O implements Cloneable {" + NEWLINE +
		    "	 public Object clone() { return this; }" + NEWLINE +
		    "	 public int hashCode() { " + NEWLINE +
		    "	   return Number.this.intValue(); " + NEWLINE +
		    "	 }" + NEWLINE +
		    "  };" + NEWLINE +
		    "  return new O();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis8.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis8.java" );
	}
    }
    
    /**
     * Fields of the receiver class of an augmenting method can be
     * accessed from within a class declaration statement of the
     * augmenting method, with or without an explicit this.  */
    public void test_augmenting_this_9() {
	try {
	    saveAs( "AugmentingThis9A.java",
		    "public class AugmentingThis9A {" + NEWLINE +
		    "  int value = 9;" + NEWLINE +
		    "}" );
	    saveAs( "augmentingThis9B.java",
		    "public Cloneable AugmentingThis9A.augmentingThis9B() {" +
		    NEWLINE +
		    "  class O implements Cloneable {" + NEWLINE +
		    "	 public Object clone() { return this; }" + NEWLINE +
		    "	 public int hashCode() { " + NEWLINE +
		    "	   return AugmentingThis9A.this.value; " + NEWLINE +
		    "	 }" + NEWLINE +
		    "  };" + NEWLINE +
		    "  return new O();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "AugmentingThis9A.java augmentingThis9B.java" );
	    expectSuccess( results );
	} finally {
	    remove( "AugmentingThis9A.java augmentingThis9B.java" );
	}
    }
    
    /**
     * Final local variables of an augmenting method can be accessible
     * from within inner classes.  */
    public void test_augmenting_this_10() {
	try {
	    saveAs( "augmentingThis10.java",
		    "public Cloneable Number.augmentingThis10() {" + NEWLINE +
		    "  final Object o = this;" + NEWLINE +
		    "  class O implements Cloneable {" + NEWLINE +
		    "	 public Object clone() { return o; }" + NEWLINE +
		    "  };" + NEWLINE +
		    "  return new O();" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "augmentingThis10.java" );
	    expectSuccess( results );
	} finally {
	    remove( "augmentingThis10.java" );
	}
    }
    
    //---------------------------------------------------------------------
    // Tests for implicit and explicit "this" references from within
    // inner classes.
    //---------------------------------------------------------------------

    /**
     * Methods of the outer class can be accessed from within an
     * anonymous class, with or without an explicit "this". */
    public void test_inner_this_1() {
	try {
	    saveAs( "InnerThis1.java",
		    "public class InnerThis1 {" + NEWLINE +
		    "  public Cloneable m() {" + NEWLINE +
		    "  	 return new Cloneable() {" + NEWLINE +
		    "  	     public Object clone() { return this; }" + 
		    NEWLINE +
		    "  	     public int hashCode() { " + NEWLINE +
		    "  	       return InnerThis1.this.value(); " + NEWLINE +
		    "  	     }" + NEWLINE +
		    "  	   };" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public int value() { return 1; }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "InnerThis1.java" );
	    expectSuccess( results );
	} finally {
	    remove( "InnerThis1.java" );
	}
	try {
	    saveAs( "InnerThis1.java",
		    "public class InnerThis1 {" + NEWLINE +
		    "  public Cloneable m() {" + NEWLINE +
		    "  	 return new Cloneable() {" + NEWLINE +
		    "  	     public Object clone() { return this; }" + 
		    NEWLINE +
		    "  	     public int hashCode() { " + NEWLINE +
		    "  	       return value(); " + NEWLINE +
		    "  	     }" + NEWLINE +
		    "  	   };" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public int value() { return 1; }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "InnerThis1.java" );
	    expectSuccess( results );
	} finally {
	    remove( "InnerThis1.java" );
	}
    }
    
    /**
     * Fields of the outer class can be accessed from within an
     * anonymous class, with or without an explicit "this". */
    public void test_inner_this_2() {
	try {
	    saveAs( "InnerThis2.java",
		    "public class InnerThis2 {" + NEWLINE +
		    "  public Cloneable m() {" + NEWLINE +
		    "  	 return new Cloneable() {" + NEWLINE +
		    "  	     public Object clone() { return this; }" + 
		    NEWLINE +
		    "  	     public int hashCode() { " + NEWLINE +
		    "  	       return InnerThis2.this.value; " + NEWLINE +
		    "  	     }" + NEWLINE +
		    "  	   };" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public int value = 2;" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "InnerThis2.java" );
	    expectSuccess( results );
	} finally {
	    remove( "InnerThis2.java" );
	}
	try {
	    saveAs( "InnerThis2.java",
		    "public class InnerThis2 {" + NEWLINE +
		    "  public Cloneable m() {" + NEWLINE +
		    "  	 return new Cloneable() {" + NEWLINE +
		    "  	     public Object clone() { return this; }" + 
		    NEWLINE +
		    "  	     public int hashCode() { " + NEWLINE +
		    "  	       return value; " + NEWLINE +
		    "  	     }" + NEWLINE +
		    "  	   };" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public int value = 2;" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "InnerThis2.java" );
	    expectSuccess( results );
	} finally {
	    remove( "InnerThis2.java" );
	}
    }
    
    /**
     * References to fields of an outer class cannot cross a static
     * lexical contour. */
    public void test_inner_this_3() {
	try {
	    saveAs( "InnerThis3.java",
		    "public class InnerThis3 {" + NEWLINE +
		    "  public static class Inner {" + NEWLINE +
		    "    public Cloneable m() {" + NEWLINE +
		    "      return new Cloneable() {" + NEWLINE +
		    "          public Object clone() { return this; }" + 
		    NEWLINE +
		    "          public int hashCode() { " + NEWLINE +
		    "    	 return InnerThis3.this.value; " + 
		    NEWLINE +
		    "          }" + NEWLINE +
		    "        };" + NEWLINE +
		    "    }" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public int value = 3;" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "InnerThis3.java" );
	    expectFailure( results,  "File \"InnerThis3.java\", line 7, character 36 error: Invalid outer class for this operation; reference crosses the static context \"Inner\" [JLS2 15.8.4]" + NEWLINE +  "File \"InnerThis3.java\", line 6, character 17 error: Method \"hashCode\" must return a value [JLS 8.4.5]" + NEWLINE );
	} finally {
	    remove( "InnerThis3.java" );
	}
	try {
	    saveAs( "InnerThis3.java",
		    "public class InnerThis3 {" + NEWLINE +
		    "  public static class Inner {" + NEWLINE +
		    "    public Cloneable m() {" + NEWLINE +
		    "      return new Cloneable() {" + NEWLINE +
		    "          public Object clone() { return this; }" + 
		    NEWLINE +
		    "          public int hashCode() { " + NEWLINE +
		    "    	 return value; " + 
		    NEWLINE +
		    "          }" + NEWLINE +
		    "        };" + NEWLINE +
		    "    }" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public int value = 3;" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "InnerThis3.java" );
	    expectFailure( results,  "File \"InnerThis3.java\", line 7, character 27 error: Invalid outer class for this operation; reference crosses the static context \"Inner\" [JLS2 15.8.4]" + NEWLINE +  "File \"InnerThis3.java\", line 6, character 17 error: Method \"hashCode\" must return a value [JLS 8.4.5]" + NEWLINE );
	} finally {
	    remove( "InnerThis3.java" );
	}
    }
    
    //---------------------------------------------------------------------
    // Tests for appropriate privileged access warnings.
    //---------------------------------------------------------------------

    /**
     * Regular fields of the augmented object can be accessed with
     * or without an explicit "this".  */
    public void test_field_access_1() {
	try {
	    saveAs( "FieldAccess1A.java",
		    "public class FieldAccess1A {" + NEWLINE +
		    "  public int value() { return value; }" + NEWLINE +
		    "  private int value = 1;" + NEWLINE +
		    "}" );
	    saveAs( "FieldAccess1B.java",
		    "public class FieldAccess1B {" + NEWLINE +
		    "  public int m() {" + NEWLINE +
		    "    return new FieldAccess1A().value;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "-w3 FieldAccess1A.java FieldAccess1B.java" );
	    expectFailure( results, "warning: Reference to field \"value\" matches _non-accessible_ member declared in \"FieldAccess1A\".  This may be OK, but be sure right field is being referenced." + NEWLINE +  "File \"FieldAccess1B.java\", line 3, character 37 error: Cannot find field \"value\" [JLS2 15.11]" + NEWLINE +  "File \"FieldAccess1B.java\", line 2, character 9 error: Method \"m\" must return a value [JLS 8.4.5]" + NEWLINE );
	} finally {
	    remove( "FieldAccess1A.java FieldAccess1B.java" );
	}
    }

    /**
     * Classes inherit (non-private) classes from superclasses and
     * interfaces. JLS2 6.5.5 */
    public void test_class_access_1() {
	try {
	    saveAs( "ClassAccess1A.java",
		    "public class ClassAccess1A {" + NEWLINE +
		    "  private class B {}" + NEWLINE +
		    "}" );
	    saveAs( "ClassAccess1B.java",
		    "interface ClassAccess1B {" + NEWLINE +
		    "  class B {}" + NEWLINE +
		    "}" );
	    saveAs( "ClassAccess1C.java",
		    "public class ClassAccess1C extends ClassAccess1A " +
		    NEWLINE +
		    "  implements ClassAccess1B" + NEWLINE +
		    "{" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    Object o = new B();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "-w3 ClassAccess1A.java ClassAccess1B.java " + 
			 "ClassAccess1C.java" );
	    expectSuccess( results );
	    assertEquals( "warning: Reference to type \"B\" matches _non-accessible_ class ClassAccess1A$B.  This may be OK, but be sure right type is being found." + NEWLINE +  "File \"ClassAccess1C.java\", line 5, character 13 warning: Local variable \"o\" is not used [KOPI]" + NEWLINE, results.errorOut, true );
	} finally {
	    remove( "ClassAccess1A.java ClassAccess1B.java " + 
		    "ClassAccess1C.java" );
	}
    }
    
    /**
     * Classes inherit (non-private) classes from superclasses and
     * interfaces. JLS2 6.5.5.  Both inner classes are inherited resulting in
     * an ambiguity*/
    public void test_class_access_2() {
	try {
	    saveAs( "ClassAccess2A.java",
		    "public class ClassAccess2A {" + NEWLINE +
		    "  public class B {}" + NEWLINE +
		    "}" );
	    saveAs( "ClassAccess2B.java",
		    "interface ClassAccess2B {" + NEWLINE +
		    "  class B {}" + NEWLINE +
		    "}" );
	    saveAs( "ClassAccess2C.java",
		    "public class ClassAccess2C extends ClassAccess2A " +
		    NEWLINE +
		    "  implements ClassAccess2B" + NEWLINE +
		    "{" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    Object o = new B();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "-w3 ClassAccess2A.java ClassAccess2B.java " + 
			 "ClassAccess2C.java" );
	    expectFailure( results,  "File \"ClassAccess2C.java\", line 5, character 21 error: The type reference is ambiguous between (at least) ClassAccess2A$B and ClassAccess2B$B. [JLS2 6.5.5]" + NEWLINE );
	} finally {
	    remove( "ClassAccess2A.java ClassAccess2B.java " + 
		    "ClassAccess2C.java" );
	}
    }
    
    /**
     * Classes cannot instantiate private inner class of another
     * public class. */
    public void test_class_access_3() {
	try {
	    saveAs( "ClassAccess3A.java",
		    "public class ClassAccess3A {" + NEWLINE +
		    "  private class Inner {}" + NEWLINE +
		    "}" );
	    saveAs( "ClassAccess3B.java",
		    "public class ClassAccess3B" + NEWLINE +
		    "{" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    Object o = new ClassAccess3A().new Inner();" + 
		    NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "-w3 ClassAccess3A.java ClassAccess3B.java" );
	    expectFailure( results, 
"warning: Reference to type \"Inner\" matches _non-accessible_ class ClassAccess3A$Inner.  This may be OK, but be sure right type is being found." + NEWLINE +
 "File \"ClassAccess3B.java\", line 4, character 39 error: Cannot find type \"Inner\"" + NEWLINE +
 "File \"ClassAccess3B.java\", line 4, character 13 warning: Local variable \"o\" is not used [KOPI]" + NEWLINE );
	} finally {
	    remove( "ClassAccess3A.java ClassAccess3B.java" );
	}
    }

	/**
	 * Inner classes hide inner classes of the same name in superclasses
	 * and superinterfaces. */
    public void test_class_access_4() {
	try {
	    saveAs( "ClassAccess4A.java",
		    "  public interface ClassAccess4A {" + NEWLINE +
		    "    public static class Inner {" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    saveAs( "ClassAccess4B.java",
		    "public interface ClassAccess4B extends ClassAccess4A {"
		    + NEWLINE +
		    "  public static class Inner {" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results =
		compile( "ClassAccess4A.java ClassAccess4B.java" );
	    expectSuccess( results );
	} finally {
	    remove ( "ClassAccess4A.java ClassAccess4B.java" );
	}
    }
    
    /**
     * Private methods are not inherited and so subclass can change
     * return type.  */
        public void test_method_access_1() {
	try {
	    saveAs( "MethodAccess1A.java",
		    "public class MethodAccess1A {" + NEWLINE +
		    "  private boolean m() { return true; }" + NEWLINE +
		    "}" );
	    saveAs( "MethodAccess1B.java",
		    "public class MethodAccess1B extends MethodAccess1A {" + 
		    NEWLINE +
		    "  public int m() { return 0; }" + NEWLINE +
		    "}" );
	    saveAs( "MethodAccess1C.java",
		    "public class MethodAccess1C {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "    int i = new MethodAccess1B().m();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results = 
                compile( "-w3 MethodAccess1A.java MethodAccess1B.java " + 
			 "MethodAccess1C.java" );
	    expectSuccess( results );
	    assertEquals(  "File \"MethodAccess1A.java\", line 2, character 10 warning: Private method \"m\" is not used [KOPI]" + NEWLINE + "warning: Method declaration \"m\" matches _non-accessible_ member declared in \"MethodAccess1A\" but cannot override it.  This may be OK, but be sure overriding is correct." + NEWLINE +  "File \"MethodAccess1C.java\", line 3, character 10 warning: Local variable \"i\" is not used [KOPI]" + NEWLINE, 
			  results.errorOut, true );
	} finally {
	    remove( "MethodAccess1A.java MethodAccess1B.java " + 
		    "MethodAccess1C.java" );
	}
    }

    //---------------------------------------------------------------------
    // Tests for internal augmenting methods (i.e., internalizable methods)
    //---------------------------------------------------------------------

    /**
     * Private augmenting generic functions should be accessible to
     * other augmenting generic functions in the same compilation
     * unit. */
    public void test_internalize_3() {
	try {
	    saveAs( "internalize3.java",
		    "" + NEWLINE +
		    "void Object.internalize3() { this.helper(); }" + NEWLINE +
		    "void Number.internalize3() { this.helper(); }" + NEWLINE +
		    "private String Object.helper() {" + NEWLINE +
		    "  return this.toString();" + NEWLINE +
		    "}" + NEWLINE +
		    "private String Number.helper() {" + NEWLINE +
		    "  return \"\" + this.doubleValue();" + NEWLINE +
		    "}" );

	    CompilationResults results = 
		compile( "internalize3.java" );
	    expectSuccess( results );
	} finally {
	    remove( "internalize3.java" );
	}
    }

    //---------------------------------------------------------------------
    // Duplicate classes
    //---------------------------------------------------------------------

    /**
     * Cannot import two classes with the same name in the same
     * compilation unit.  */
    public void test_dup_class_2() {
	try {
	    saveAs( "Vector.java",
		    "package other;" + NEWLINE + 
		    "public class Vector { }" );
	    saveAs( "DupClass2.java",
		    "import java.util.Vector;" + NEWLINE +
		    "import other.Vector;	// error, duplicate" + NEWLINE+
		    "public class DupClass2 { }" );

	    CompilationResults results = 
		compile( "Vector.java DupClass2.java" );
	    expectFailure( results,  "File \"DupClass2.java\", line 1, character 7 error: Redeclaration or reimport of type \"Vector\" declared or imported in the same compilation unit [JLS 7.3]" + NEWLINE );
	} finally {
	    remove( "Vector.java DupClass2.java" );
	}
    }

    //---------------------------------------------------------------------
    // Duplicate methods
    //---------------------------------------------------------------------

    /**
     * Changing the order of multimethod declarations shouldn't impact
     * the results of duplicate method checks.  */
    public void test_reorder_1() {
	try {
	    saveAs( "reorder1.java",
		    "public void Integer.reorderOK1( Number n ) { }" + 
		    NEWLINE +
		    "public void Integer.reorderOK1( Number@Integer n ) { }" + 
		    NEWLINE +
		    "public void Float.reorderOK1( Number@Float n ) { }" + 
		    NEWLINE +
		    "public void Integer.reorderOK1( Number@Float n ) { }" + 
		    NEWLINE +
		    "public void Float.reorderOK1( Number n ) { }" + NEWLINE +
		    "public void Float.reorderOK1( Number@Integer n ) { }" );

	    CompilationResults results = 
		compile( "reorder1.java" );
	    expectSuccess( results );
	} finally {
	    remove( "reorder1.java" );
	}
    }

    /**
     * Changing the order of multimethod declarations shouldn't impact
     * the results of duplicate method checks.  */
    public void test_reorder_2() {
	try {
	    saveAs( "reorder2.java",
		    "public void Integer.reorderOK2( Number n ) { }" + 
		    NEWLINE +
		    "public void Float.reorderOK2( Number n ) { }" + NEWLINE +
		    "public void Integer.reorderOK2( Number@Float n ) { }" + 
		    NEWLINE +
		    "public void Float.reorderOK2( Number@Integer n ) { }" +
		    NEWLINE +
		    "public void Integer.reorderOK2( Number@Integer n ) { }" + 
		    NEWLINE + 
		    "public void Float.reorderOK2( Number@Float n ) { }" );

	    CompilationResults results = 
		compile( "reorder2.java" );
	    expectSuccess( results );
	} finally {
	    remove( "reorder2.java" );
	}
    }

    /**
     * Changing the order of multimethod declarations shouldn't impact
     * the results of duplicate method checks.  */
    public void test_reorder_3() {
	try {
	    saveAs( "reorder3.java",
		    "public void Float.reorderOK3( Number n ) { }" + NEWLINE +
		    "public void Integer.reorderOK3( Number n ) { }" + 
		    NEWLINE +
		    "public void Integer.reorderOK3( Number@Integer n ) { }" + 
		    NEWLINE +
		    "public void Float.reorderOK3( Number@Float n ) { }" + 
		    NEWLINE +
		    "public void Float.reorderOK3( Number@Integer n ) { }" + 
		    NEWLINE +
		    "public void Integer.reorderOK3( Number@Float n ) { }" );

	    CompilationResults results = 
		compile( "reorder3.java" );
	    expectSuccess( results );
	} finally {
	    remove( "reorder3.java" );
	}
    }

    //---------------------------------------------------------------------
    // Final fields and variables
    //---------------------------------------------------------------------

    /**
     * A final variable can only be assigned to once. JLS2 4.5.4 */
    public void test_final_1() {
	try {
	    saveAs( "Final1.java",
		    "public class Final1 {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "	 final float x;" + NEWLINE +
		    "	 for( int i=0; i<2; i++ ) {" + NEWLINE +
		    "	   x = 0f; // error, may be assigning more than once" + 
		    NEWLINE +
		    "	 }" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final1.java" );
	    expectFailure( results,  "File \"Final1.java\", line 4, character 13 error: Final variable \"x\" may be assigned to within loop [JLS2 16]" + NEWLINE );
	} finally {
	    remove( "Final1.java" );
	}
    }

    /**
     * A final field can only be assigned to once. JLS2 8.3.1.2 */
    public void test_final_2() {
	try {
	    saveAs( "Final2.java",
		    "public class Final2 {" + NEWLINE +
		    "  final float x;" + NEWLINE +
		    "  public Final2() {" + NEWLINE +
		    "	 for( int i=0; i<2; i++ ) {" + NEWLINE +
		    "	   x = 0f; // error, may be assigning more than once or not at all" + 
		    NEWLINE +
		    "	 }" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final2.java" );
	    expectFailure( results,  
			   "File \"Final2.java\", line 4, character 13 error: Final field \"x\" may be assigned to within loop [JLS2 16]" + NEWLINE +
			   "File \"Final2.java\", line 1, character 13 error: Final field \"x\" may not have been initialized [JLS2 8.3.1.2]" + NEWLINE
			   );
	} finally {
	    remove( "Final2.java" );
	}
    }

    /**
     * A final variable can only be assigned to once. JLS2 4.5.4 */
    public void test_final_3() {
	try {
	    saveAs( "Final3.java",
		    "public class Final3 {" + NEWLINE +
		    "  public static void main( String[] args ) {" + NEWLINE +
		    "	 final float x;" + NEWLINE +
		    "	 x = 0f;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final3.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final3.java" );
	}
    }

    /**
     * A final field can only be assigned to once. JLS4 8.3.1.4 */
    public void test_final_4() {
	try {
	    saveAs( "Final4.java",
		    "public class Final4 {" + NEWLINE +
		    "  final float x;" + NEWLINE +
		    "  public Final4() {" + NEWLINE +
		    "    x = 0f;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final4.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final4.java" );
	}
    }

    /**
     * A final field can only be assigned to once. JLS2 8.3.1.2 */
    public void test_final_5() {
	try {
	    saveAs( "Final5.java",
		    "public class Final5 {" + NEWLINE +
		    "  static final float x;" + NEWLINE +
		    "  static {" + NEWLINE +
		    "	 for( int i=0; i<2; i++ ) {" + NEWLINE +
		    "	   x = 0f; // error, may be assigning more than once" + 
		    NEWLINE +
		    "	 }" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final5.java" );
	    expectFailure( results,  "File \"Final5.java\", line 4, character 13 error: Final field \"x\" may be assigned to within loop [JLS2 16]" + NEWLINE  );
	} finally {
	    remove( "Final5.java" );
	}
    }

    /**
     * A final field can only be assigned to once. JLS4 8.3.1.4 */
    public void test_final_6() {
	try {
	    saveAs( "Final6.java",
		    "public class Final6 {" + NEWLINE +
		    "  static final float x;" + NEWLINE +
		    "  static {" + NEWLINE +
		    "    x = 0f;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final6.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final6.java" );
	}
    }

    /**
     * A final field in an interface must be initialized. */
    public void test_final_7() {
	try {
	    saveAs( "Final7.java",
		    "public interface Final7 {" + NEWLINE +
		    "  static final float x;" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final7.java" );
	    expectFailure( results,  "File \"Final7.java\", line 1, character 17 error: Final field \"x\" may not have been initialized [JLS2 8.3.1.2]" + NEWLINE  );
	} finally {
	    remove( "Final7.java" );
	}
    }

    /**
     * A final field in an interface must be initialized. */
    public void test_final_8() {
	try {
	    saveAs( "Final8.java",
		    "public interface Final8 {" + NEWLINE +
		    "  static final float x = 1.0f;" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final8.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final8.java" );
	}
    }

    /**
     * A final field must be assigned to before being used. */
    public void test_final_9() {
	try {
	    saveAs( "Final9.java",
		    "public class Final9 {" + NEWLINE +
		    "  final float x;" + NEWLINE +
		    "  public Final9() {" + NEWLINE +
		    "    float y = x;" + NEWLINE +
		    "    x = 0f;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final9.java" );
	    expectFailure( results,  "File \"Final9.java\", line 4, character 17 error: Variable \"x\" may not have been initialized before use [JLS 14.3]" + NEWLINE );
	} finally {
	    remove( "Final9.java" );
	}
    }

    /**
     * A final field must be assigned to before being used. */
    public void test_final_10() {
	try {
	    saveAs( "Final10.java",
		    "public class Final10 {" + NEWLINE +
		    "  final float x;" + NEWLINE +
		    "  {" + NEWLINE +
		    "    float y = x;" + NEWLINE +
		    "    x = 0f;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final10.java" );
	    expectFailure( results,  "File \"Final10.java\", line 4, character 17 error: Variable \"x\" may not have been initialized before use [JLS 14.3]" + NEWLINE );
	} finally {
	    remove( "Final10.java" );
	}
    }

    /**
     * Explicit this() constructor calls are assumed to result in all
     * final instance fields being initialized. */
    public void test_final_11() {
	try {
	    saveAs( "Final11.java",
		    "public class Final11 {" + NEWLINE +
		    "  final int x;" + NEWLINE +
		    "  final int y;" + NEWLINE +
		    "  public Final11() {" + NEWLINE +
		    "    this(0);" + NEWLINE +
		    "    y = 0;  // error, assign to y twice" + NEWLINE +
		    "  }" + NEWLINE +
		    "  public Final11(int i) {" + NEWLINE +
		    "    this.x = i;" + NEWLINE +
		    "    y = 0;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final11.java" );
	    expectFailure( results,  "File \"Final11.java\", line 6, character 8 error: Final field \"y\" may have already been initialized [JLS2 16]" + NEWLINE );
	} finally {
	    remove( "Final11.java" );
	}
    }

    /**
     * We were seeing some strange bugs with references to final
     * fields of other instances of the local class from within
     * constructors. */
    public void test_final_12() {
	try {
	    saveAs( "Final12.java",
		    "public class Final12 {" + NEWLINE +
		    "  public final int x;" + NEWLINE +
		    "  public Final12(Final12 other) {" + NEWLINE +
		    "    this.x = other.x;" + NEWLINE +
		    "  }" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final12.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final12.java" );
	}
    }

    /**
     * We were seeing some strange bugs with references to final
     * fields of other instances of the local class from within
     * constructors. */
    public void test_final_13() {
	try {
	    saveAs( "Final13.java",
		    "public class Final13 {" + NEWLINE +
		    "  public int x = C;" + NEWLINE +
		    "  public final static int C = 0;" + NEWLINE +
		    "}"
		    );

	    CompilationResults results = 
		compile( "Final13.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final13.java" );
	}
    }

    /**
     * A constructor can use a final field initialized in an initializer. */
    public void test_final_14() {
	try {
	    saveAs( "Final14.java",
		    "public class Final14 {" + NEWLINE +
		    "    private final Object o = new Object();" + NEWLINE +
		    "" + NEWLINE +
		    "    Final14() {" + NEWLINE +
		    "	Object m = o;" + NEWLINE +
		    "    }   " + NEWLINE +
		    "}" + NEWLINE
		    );

	    CompilationResults results = 
		compile( "Final14.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final14.java" );
	}
    }

    /**
     * Two disjoint constructors can initialize a blank final field. */
    public void test_final_15() {
	try {
	    saveAs( "Final15.java",
		    "public class Final15 {" + NEWLINE +
		    "    private final int x;" + NEWLINE +
		    "    Final15(int x) {" + NEWLINE +
		    "	this.x = x;" + NEWLINE +
		    "    }" + NEWLINE +
		    "    Final15() {" + NEWLINE +
		    "	this.x = 0;" + NEWLINE +
		    "    }" + NEWLINE +
		    "}" + NEWLINE
		    );

	    CompilationResults results = 
		compile( "Final15.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Final15.java" );
	}
    }

    //---------------------------------------------------------------------
    // Local variable scoping
    //---------------------------------------------------------------------

    /** 
     * A local variable cannot appear within the scope of another
     * local variable of the same name (JLS2 14.4.2). */
    public void test_local1() {
	try {
	    CompilationResults results =
		compile( saveAs( "Local1.java",
				 "public class Local1 {" + NEWLINE +
				 "  public void m() {" + NEWLINE +
				 "    int i = 1;" + NEWLINE +
				 "    { int i = 0; }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,
			    "File \"Local1.java\", line 4, character 12 error: Local variable \"i\" is already defined in current block or an enclosing one [JLS2 14.4.2]" + NEWLINE);
	} finally {
	    remove( "Local1.java" );
	} // end of finally

    }

    /**
     * A switch statement's body is a new block, so local variables
     *  defined in the body are not visible outside the switch. */
    public void test_local2() {
	try {
	    CompilationResults results =
		compile( saveAs( "Local2.java",
				 "public class Local2 {" + NEWLINE +
				 "  public void m() {" + NEWLINE +
				 "    int i = 1;" + NEWLINE +
				 "    switch(i) {" + NEWLINE +
				 "      case 0: " + NEWLINE +
				 "        int j = 3;" + NEWLINE +
				 "    }" + NEWLINE +
				 "    int j = 2;" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "Local2.java" );
	} // end of finally

    }
	    
    /**
     * A for statement is a new block, so local variables
     *  defined in the initializer are not visible outside the for loop. */
    public void test_local3() {
	try {
	    CompilationResults results =
		compile( saveAs( "Local3.java",
				 "public class Local3 {" + NEWLINE +
				 "  public void m() {" + NEWLINE +
				 "    for( int i=0; i<10; i++ ) {" + NEWLINE +
				 "      System.out.println(i);" + NEWLINE +
				 "    }" + NEWLINE +
				 "    System.out.println(i);" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,
			    "File \"Local3.java\", line 6, character 26 error: Variable \"i\" is not defined in current context" + NEWLINE);
	} finally {
	    remove( "Local3.java" );
	} // end of finally
    }
	    
    /**
     * A for statement is a new block, so local variables
     *  defined in the initializer are not visible outside the for loop. */
    public void test_local4() {
	try {
	    CompilationResults results =
		compile( saveAs( "Local4.java",
				 "public class Local4 {" + NEWLINE +
				 "  public void m() {" + NEWLINE +
				 "    for( int i=0, j=0; i<10; i++ ) {" + NEWLINE +
				 "      System.out.println(i + \", \" + j);" + NEWLINE +
				 "    }" + NEWLINE +
				 "    System.out.println(i + \", \" + j);" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,
			    "File \"Local4.java\", line 6, character 27 error: Variable \"i\" is not defined in current context" + NEWLINE);
	} finally {
	    remove( "Local4.java" );
	} // end of finally
    }
	    
    /**
     * The scope of a variable declared in a switch group is all of
     * the switch statement following the declaration. */
    public void test_local5() {
	try {
	    CompilationResults results =
		compile( saveAs( "Local5.java",
				 "public class Local5 {" + NEWLINE +
				 "  public void f(int x) {" + NEWLINE + 
				 "    switch (x) {" + NEWLINE +
				 "    	case 1:" + NEWLINE +
				 "    	  Object o = new Object();" + NEWLINE +
				 "    	case 2:" + NEWLINE +
				 "    	  o = new Integer(x);" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "Local5.java" );
	}
    }
	    
    /**
     * The scope of a variable declared in a switch group is all of
     * the switch statement following the declaration, even without
     * fall through. */
    public void test_local6() {
	try {
	    CompilationResults results =
		compile( saveAs( "Local6.java",
				 "public class Local6 {" + NEWLINE +
				 "  public void f(int x) {" + NEWLINE + 
				 "    switch (x) {" + NEWLINE +
				 "    	case 1:" + NEWLINE +
				 "    	  Object o = new Object();" + NEWLINE +
				 "        return;" + NEWLINE +
				 "    	case 2:" + NEWLINE +
				 "    	  o = new Integer(x);" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "Local6.java" );
	}
    }
	    
    /**
     * Typechecking of switches had a bug where a case with
     * fall-through entry and a declaration signalled a duplicate
     * declaration. */
    public void test_local7() {
	try {
	    CompilationResults results =
		compile( saveAs( "Local7.java",
				 "public class Local7 {" + NEWLINE +
				 "  public void f(int x) {" + NEWLINE + 
				 "    switch (x) {" + NEWLINE +
				 "    	case 1:" + NEWLINE +
				 "    	  System.out.println();" + NEWLINE +
				 "    	case 2:" + NEWLINE +
				 "    	  int y = x;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "Local7.java" );
	}
    }
	    
    // -------------------------------------------------------------
    // Definite assignment
    // -------------------------------------------------------------

    //  !FIXME!10/26 Organize and add cases based on JLS2, 16

    /**
     * A variable that is assigned to in both branches on an if-then-else
     * statement is definitely assigned. */
    public void test_defAssn1() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn1.java",
				 "public class DefAssn1 {" + NEWLINE +
				 "  public void m(boolean flag) {" + NEWLINE +
				 "    int i;" + NEWLINE +
				 "    if (flag)" + NEWLINE +
				 "      i=0;" + NEWLINE +
				 "    else" + NEWLINE +
				 "      i=1;" + NEWLINE +
				 "    " + NEWLINE +
				 "    System.out.println(i);" + NEWLINE +
				 // this exercises the context
				 // mechanism differently than above:
				 "    Object o1 = this;" + NEWLINE +
				 "    Object o2;" + NEWLINE +
				 "    if (flag) {" + NEWLINE +
				 "    	 o2 = o1;" + NEWLINE +
				 "    } else {" + NEWLINE +
				 "    	 o2 = this;" + NEWLINE +
				 "    }" + NEWLINE +
				 "    System.out.println(o2);" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "DefAssn1.java" );
	} // end of finally

    }

     /**
     * A variable that is assigned to after a reachable statement of a
     * block is definitely assigned after the block. */
    public void test_defAssn2() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn2.java",
				 "public class DefAssn2 {" + NEWLINE +
				 "  public int m() {" + NEWLINE +
				 "    int i;" + NEWLINE +
				 "    {" + NEWLINE +
				 "      i=0;" + NEWLINE +
				 "    }" + NEWLINE +
				 "    return i;" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "DefAssn2.java" );
	} // end of finally

    }

    /**
     * Final fields may only be assigned to once, but once in each
     * branch of an if-then-else statement is OK.  */
    public void test_defAssn3() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn3.java",
				 "public class DefAssn3 {" + NEWLINE +
     
				 "  private static final int JDK_11 = 1;" + NEWLINE +
				 "  private static final int JDK_12 = 2;" + NEWLINE +
				 "  private static final int version;" + NEWLINE +
				 "  static {" + NEWLINE +
				 "    String s = \"blah\";" + NEWLINE +
				 "    if (null == s) {" + NEWLINE +
				 "  	version = JDK_11;" + NEWLINE +
				 "    } else {" + NEWLINE +
				 "  	version = JDK_12;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "DefAssn3.java" );
	} // end of finally

    }

    /**
     * A variable is only definitely assigned if it is assigned in all
     * possible paths.  */
    public void test_defAssn4() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn4.java",
				 "public class DefAssn4 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    int i;" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case 1:" + NEWLINE +
				 "    	 i = 1;" + NEWLINE +
				 "    	 // falls through" + NEWLINE +
				 "    case 2:" + NEWLINE +
				 "    	 // no assignment to i if s==2" + NEWLINE +	    
				 "    	 break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 i = 3;" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "    System.out.println(i); // error" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"DefAssn4.java\", line 15, character 26 error: Variable \"i\" may not have been initialized before use [JLS 14.3]" + NEWLINE );
	} finally {
	    remove( "DefAssn4.java" );
	} // end of finally

    }

    /**
     * Static fields are assumed to be assigned after static
     * initializers have been checked.  */
    public void test_defAssn5() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn5.java",
				 "public class DefAssn5 {" + NEWLINE +
				 "  public DefAssn5() {" + NEWLINE +
				 "    this.f = SF;" + NEWLINE +
				 "  }" + NEWLINE +
				 "  private static final int SF = 0;" + NEWLINE +
				 "  private int f;" + NEWLINE +
     				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "DefAssn5.java" );
	}
    }

    /**
     * A variable is definitely assigned if it is assigned on all
     * paths that can reach the reference.  */
    public void test_defAssn6() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn6.java",
				 "public class DefAssn6 {" + NEWLINE +
				 "  public int m(int s) {" + NEWLINE +
				 "    int t;" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case 1:" + NEWLINE +
				 "    	  t = 1;" + NEWLINE +
				 "    	  break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	  return 0;" + NEWLINE +
				 "    }" + NEWLINE +
				 // the default path cannot reach the following
				 "    return t;" + NEWLINE +
				 "  }" + NEWLINE +
     				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "DefAssn6.java" );
	}
    }

    /**
     * A variable is definitely assigned if it is assigned on all
     * paths that can reach the reference.  */
    public void test_defAssn7() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn7.java",
				 "public class DefAssn7 {" + NEWLINE +
				 "  public void m(int i, boolean b) {" + NEWLINE +
				 "    int k;" + NEWLINE +
				 "    switch (i) {" + NEWLINE +
				 "    case 1:" + NEWLINE +
				 "      if (b) {" + NEWLINE +
				 "	  break;" + NEWLINE +
				 "      } else {" + NEWLINE +
				 "	  k = 1;" + NEWLINE +
				 "      }" + NEWLINE +
				 "    default:" + NEWLINE +
				 "      k = 2;" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "    System.out.println(k);" + NEWLINE +
				 "  }" + NEWLINE +
     				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"DefAssn7.java\", line 15, character 26 error: Variable \"k\" may not have been initialized before use [JLS 14.3]" + NEWLINE );
	} finally {
	    remove( "DefAssn7.java" );
	}
    }

    /**
     * A variable is definitely assigned if it is assigned on all
     * paths that can reach the reference.  */
    public void test_defAssn8() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn8.java",
				 "  public class DefAssn8 {" + NEWLINE +
				 "    public void m(boolean b) {" + NEWLINE +
				 "      int y;" + NEWLINE +
				 "    label:" + NEWLINE +
				 "      {" + NEWLINE +
				 "  	 if (b) {" + NEWLINE +
				 "  	   break label;" + NEWLINE +
				 "  	 } else {" + NEWLINE +
				 "  	   y = 2;" + NEWLINE +
				 "  	 }" + NEWLINE +
				 "  	 System.out.println(y);" + NEWLINE +
				 "      }" + NEWLINE +
				 "      System.out.println(y);  // error" + NEWLINE +
				 "    }" + NEWLINE +
     				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"DefAssn8.java\", line 13, character 28 error: Variable \"y\" may not have been initialized before use [JLS 14.3]" + NEWLINE );
	} finally {
	    remove( "DefAssn8.java" );
	}
    }

    /**
     * A variable is definitely assigned if it is assigned on all
     * paths that can reach the reference.  */
    public void test_defAssn9() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn9.java",
				 "public class DefAssn9 {" + NEWLINE +
				 "    public void m1(boolean b) {" + NEWLINE +
				 "	int k;" + NEWLINE +
				 "	do {" + NEWLINE +
				 "	    if (b) {" + NEWLINE +
				 "		k = 0;" + NEWLINE +
				 "		continue;" + NEWLINE +
				 "	    } else {" + NEWLINE +
				 "		return;" + NEWLINE +
				 "	    }" + NEWLINE +
				 "	} while (b);" + NEWLINE +
				 "	System.out.println(k);" + NEWLINE +
				 "    }" + NEWLINE +
				 "    public void m2(boolean b) {" + NEWLINE +
				 "	int k;" + NEWLINE +
				 "	do {" + NEWLINE +
				 "	    if (b) {" + NEWLINE +
				 "		k = 0;" + NEWLINE +
				 "	    } else {" + NEWLINE +
				 "		continue;" + NEWLINE +
				 "	    }" + NEWLINE +
				 "	} while (b);" + NEWLINE +
				 "	System.out.println(k);" + NEWLINE +
				 "    }" + NEWLINE +
     				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"DefAssn9.java\", line 23, character 30 error: Variable \"k\" may not have been initialized before use [JLS 14.3]" + NEWLINE );
	} finally {
	    remove( "DefAssn9.java" );
	}
    }

    /**
     * Access to outer variables.  */
    public void test_defAssn10() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn10.java",
				 "public class DefAssign10 {" + NEWLINE +
				 "  public void m(final String name) {" + NEWLINE +
				 "    Object o1 = new Object() {" + NEWLINE +
				 // following is OK
				 "	  String actualName = name;" + NEWLINE +
				 "      };" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE 
				 ) );
	    expectSuccess( results );
	} finally {
	    remove( "DefAssn10.java" );
	}
    }

    /**
     * Access to outer variables.  */
    public void test_defAssn11() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn11.java",
				 "public class DefAssign11 {" + NEWLINE +
				 "  public void m(final String name) {" + NEWLINE +
				 "    String emptyName = \"foo\";" + NEWLINE + 
				 "    Object o = new Object() {" + NEWLINE +
				 // following is bad, emptyName not final
				 "	  String actualName = emptyName;" + NEWLINE +
				 "      };" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE 
				 ) );
	    expectFailure( results,  "File \"DefAssn11.java\", line 5, character 41 error: Local variable \"emptyName\" should be final to be used in inner class [JLS D.1.2]" + NEWLINE );
	} finally {
	    remove( "DefAssn11.java" );
	}
    }

    /**
     * Access to outer variables.  */
    public void test_defAssn12() {
	try {
	    CompilationResults results =
		compile( saveAs( "DefAssn12.java",
				 "public class DefAssign12 {" + NEWLINE +
				 "  public void m(final String name) {" + NEWLINE +
				 // following should be error but isn't yet
				 "    final String eN2;" + NEWLINE +
				 "    Object o3 = new Object() {" + NEWLINE +
				 // following is bad, eN2 not assigned
				 "        String bad = eN2;" + NEWLINE +
				 "      };" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE 
				 ) );
	    expectFailure( results,  "File \"DefAssn12.java\", line 5, character 26 error: Variable \"eN2\" may not have been initialized before use [JLS 14.3]" + NEWLINE );
	} finally {
	    remove( "DefAssn12.java" );
	}
    }

    // -------------------------------------------------------------
    // Switch statement labels
    // -------------------------------------------------------------

    /**
     * Labels must be unique.  */
    public void test_switchLabels1() {
	try {
	    CompilationResults results =
		compile( saveAs( "SwitchLabels1.java",
				 "public class SwitchLabels1 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case 1:" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    case 1:" + NEWLINE +
				 "    	break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"SwitchLabels1.java\", line 6, character 11 error: Label \"1\" is already defined in current switch [JLS 14.9]" + NEWLINE );
	} finally {
	    remove( "SwitchLabels1.java" );
	}
    }

    /**
     * Labels must be unique.  */
    public void test_switchLabels2() {
	try {
	    CompilationResults results =
		compile( saveAs( "SwitchLabels2.java",
				 "public class SwitchLabels2 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case 2:" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"SwitchLabels2.java\", line 8, character 13 error: This default selector must be unique in switch [JLS 14.9]" + NEWLINE );
	} finally {
	    remove( "SwitchLabels2.java" );
	}
    }

    /**
     * Labels must be unique.  */
    public void test_switchLabels3() {
	try {
	    CompilationResults results =
		compile( saveAs( "SwitchLabels3.java",
				 "public class SwitchLabels3 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case 3:" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    case 3:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"SwitchLabels3.java\", line 8, character 11 error: Label \"3\" is already defined in current switch [JLS 14.9]" + NEWLINE );
	} finally {
	    remove( "SwitchLabels3.java" );
	}
    }

    /**
     * Labels must be unique. */
    public void test_switchLabels4() {
	try {
	    CompilationResults results =
		compile( saveAs( "SwitchLabels4.java",
				 "public class SwitchLabels4 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    case 4:" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"SwitchLabels4.java\", line 8, character 13 error: This default selector must be unique in switch [JLS 14.9]" + NEWLINE );
	} finally {
	    remove( "SwitchLabels4.java" );
	}
    }

    /**
     * Labels must be constants. */
    public void test_switchLabels5() {
	try {
	    CompilationResults results =
		compile( saveAs( "SwitchLabels5.java",
				 "public class SwitchLabels5 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case s:" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"SwitchLabels5.java\", line 4, character 11 error: Switch label must be constant value [JLS 14.9]" + NEWLINE );
	} finally {
	    remove( "SwitchLabels5.java" );
	}
    }

    /**
     * Labels must be constants. */
    public void test_switchLabels6() {
	try {
	    CompilationResults results =
		compile( saveAs( "SwitchLabels6.java",
				 "public class SwitchLabels6 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    final int t=0;" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case t:" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	break;" + NEWLINE +
				 "    }" + NEWLINE +
				 // the following line also checks reachability
				 "    System.out.println(s);" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "SwitchLabels6.java" );
	}
    }

    /**
     * Constant labels may be defined in unreachable code.  Very
     * weird.  */
    public void test_switchLabels7() {
	try {
	    CompilationResults results =
		compile( saveAs( "SwitchLabels7.java",
				 "public class SwitchLabels7 {" + NEWLINE +
				 "  public void m(int s) {" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case 0:" + NEWLINE +
				 "      final int t = 1;" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    case t:" + NEWLINE +
				 "      break;" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 break;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "SwitchLabels7.java" );
	}
    }


    // -------------------------------------------------------------
    // Return Statements
    // -------------------------------------------------------------

    /**
     * Return statements may not appear in initializers.  */
    public void test_Return1() {
	try {
	    CompilationResults results =
		compile( saveAs( "Return1.java",
				 "  public class Return1 {" + NEWLINE +
				 "    int x;" + NEWLINE +
				 "    {" + NEWLINE +
				 "      x = 0;" + NEWLINE +
				 "      return;  // error" + NEWLINE +
				 "    }" + NEWLINE +
				 "  } " + NEWLINE
				 ) );
	    expectFailure( results,  "File \"Return1.java\", line 5, character 13 error: Cannot use a return statement within an initializer [JLS2 14.16]" + NEWLINE );
	} finally {
	    remove( "Return1.java" );
	}
    }



    // -------------------------------------------------------------
    // Reachability
    // -------------------------------------------------------------
    
    /**
     * The end of the method block cannot be reached (i.e., every
     * control flow path reaches a statement that completes abruptly.
     * */
    public void test_reachable1() {
	try {
	    CompilationResults results =
		compile( saveAs( "Reachable1.java",
				 "public class Reachable1 {" + NEWLINE +
				 "  public int m(int s) {" + NEWLINE +
				 "    switch (s) {" + NEWLINE +
				 "    case 0:" + NEWLINE +
				 "      do {" + NEWLINE +
				 "        switch (s) {" + NEWLINE +
				 "        default:" + NEWLINE +
				 "          break;" + NEWLINE +
				 "        }" + NEWLINE +
				 "      } while (true);" + NEWLINE +
				 "    default:" + NEWLINE +
				 "    	 return s;" + NEWLINE +
				 "    }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "Reachable1.java" );
	}
    }

    /**
     * Conditional breaks inside switches. */
    public void test_reachable2() {
	try {
	    CompilationResults results =
		compile( saveAs( "Reachable2.java",
				 "public class Reachable2 {" + NEWLINE +
				 "  public void m1(int i, int j) {" + NEWLINE +
				 "  	 switch(i) {" + NEWLINE +
				 "  	 case 0:" + NEWLINE +
				 "  	   if (i == j) { return; } else { break; }" + NEWLINE +
				 "  	 default:" + NEWLINE +
				 "  	   return;" + NEWLINE +
				 "  	 }" + NEWLINE +
				 "  	 int k = i + j;	" + NEWLINE +    
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "Reachable2.java" );
	}
    }


    /**
     * Check for warning on a case that falls through, except for the
     * last case.  */
    public void test_reachable3() {
	try {
	    CompilationResults results =
		compile( "-w3 " +  
			 saveAs( "Reachable3.java",
				 "public class Reachable3 {" + NEWLINE +
				 "  public void m1(int i, int j) {" + NEWLINE +
				 "  	 switch(i) {" + NEWLINE +
				 "  	 case 0:" + NEWLINE +
				 "         System.out.println();" + NEWLINE +
				 "  	 default:" + NEWLINE +
				 "         System.out.println();" + NEWLINE +
				 "  	 }" + NEWLINE +
				 "  }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	    assertEquals(  "File \"Reachable3.java\", line 3, character 18 warning: Fall through from one case to the next is often a source of bugs and is strongly discouraged [KOPI]" + NEWLINE, results.errorOut, true );
	} finally {
	    remove( "Reachable3.java" );
	}
    }

    /**
     * Conditional continues inside for-loops. */
    public void test_reachable4() {
	try {
	    CompilationResults results =
		compile( saveAs( "Reachable4.java",
				 "public class Reachable4 {" + NEWLINE +
				 "  int m(int i, boolean b) {" + NEWLINE +
				 "  tryAgain:" + NEWLINE +
				 "	for (;;) {" + NEWLINE +
				 "	    if (b) continue tryAgain;" + NEWLINE +
				 "	    return i;" + NEWLINE +
				 "	}" + NEWLINE +
				 "  }" + NEWLINE +
				 "    void m1(int j) {" + NEWLINE +
				 "	for (int i=0; i<10; i++) {" + NEWLINE +
				 "	    if (j == 0)" + NEWLINE +
				 "		continue;" + NEWLINE +
				 "	    else" + NEWLINE +
				 "		continue;" + NEWLINE +
				 "	}" + NEWLINE +
				 "    }" + NEWLINE +
				 "    void m2(int j) {" + NEWLINE +
				 "	for (int i=0; i<10; i++) {" + NEWLINE +
				 "  	    try {" + NEWLINE +
				 "  	    } catch (Throwable x) {" + NEWLINE +
				 "  		continue;" + NEWLINE +
				 "  	    }" + NEWLINE +
				 "	}" + NEWLINE +
				 "    }" + NEWLINE +
				 "    void m3(boolean b) {" + NEWLINE +
				 "    label:" + NEWLINE +
				 "	if (b) {" + NEWLINE +
				 "	    System.out.println();" + NEWLINE +
				 "	} else {" + NEWLINE +
				 "	    break label;" + NEWLINE +
				 "	}" + NEWLINE +
				 "    }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectSuccess( results );
	} finally {
	    remove( "Reachable4.java" );
	}
    }

    /**
     * Conditional continues inside for-loops. */
    public void test_reachable5() {
	try {
	    CompilationResults results =
		compile( saveAs( "Reachable5.java",
				 "public class Reachable5 {" + NEWLINE +
				 "    void m1(boolean b) {" + NEWLINE +
				 "	int k;" + NEWLINE +
				 "	do {" + NEWLINE +
				 "	    if (b) {" + NEWLINE +
				 "		break;" + NEWLINE +
				 "	    } else {" + NEWLINE +
				 "		k = 0;" + NEWLINE +
				 "	    }" + NEWLINE +
				 "	} while (b);" + NEWLINE +
				 "	System.out.println(k);" + NEWLINE +
				 "    }" + NEWLINE +
				 "    void m2(boolean b) {" + NEWLINE +
				 "	int k;" + NEWLINE +
				 "	do {" + NEWLINE +
				 "	    if (b) {" + NEWLINE +
				 "		k = 0;" + NEWLINE +
				 "		break;" + NEWLINE +
				 "	    }" + NEWLINE +
				 "	    return;" + NEWLINE +
				 "	} while (b);" + NEWLINE +
				 "	System.out.println(k);" + NEWLINE +
				 "    }" + NEWLINE +
				 "}" + NEWLINE ) );
	    expectFailure( results,  "File \"Reachable5.java\", line 11, character 30 error: Variable \"k\" may not have been initialized before use [JLS 14.3]" + NEWLINE  );
	} finally {
	    remove( "Reachable5.java" );
	}
    }    

    //---------------------------------------------------------------------
    // Inner class resolution
    //---------------------------------------------------------------------

	/* This test ensures that we resolve references to inner classes
	   properly */
    public void test_inner_resolve1() {
	try {
	    saveAs( "Outer.java",
		    "package foo.bar;" + NEWLINE +
		    "public class Outer {" + NEWLINE +
		    "  public static class Inner {" + NEWLINE +
		    "    public static void blab() {}" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" + NEWLINE );
	    saveAs( "InnerResolve.java",
		    "public class InnerResolve {" + NEWLINE +
		    "  public void bub() { " + NEWLINE +
		    "    foo.bar.Outer.Inner.blab();" + NEWLINE +
		    "  }" + NEWLINE +
		    "}" );
	    CompilationResults results =
		compile( "Outer.java InnerResolve.java" );
	    expectSuccess( results );
	} finally {
	    remove( "Outer.java InnerResolve.java" );
	} // end of finally

    }

    
    //---------------------------------------------------------------------
    // PRIVATE HELPERS
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // COMMON FIXTURE
    //---------------------------------------------------------------------

//      public void setUp() { }

//      public void tearDown() { }

}
