/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestMjcParser.java,v 1.20 2004/08/14 01:51:19 davidcok Exp $
 */

package org.multijava.mjc;

import antlr.RecognitionException;
import antlr.Token;
import antlr.TokenStreamException;

import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.compiler.PositionedError;

import org.multijava.javadoc.Annotatable;
import org.multijava.javadoc.JavadocComment;

import org.multijava.util.testing.TestCase;

import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;

import java.util.ArrayList;

/**
 * Unit tests for MjcParser */
public class TestMjcParser extends TestCase {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    public TestMjcParser( String name ) {
	this( name, new CModifier() );
    }

    public TestMjcParser( String name, ModifierUtility modUtil ) {
	super( name );
	compiler = new TestCompiler( errorOut, modUtil );
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------

    /**
     * Should handle a class declaration without a javadoc comment.
     * */
    public void testClassDeclParsing() 
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( SAMPLE_CLASS );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );
	    
	assertEquals( SAMPLE_CLASS, w.toString(), true );
	    
    }

    /**
     * Should handle a class declaration preceded by a javadoc
     * comment.  */
    public void testClassDeclJavadocParsing1() 
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( SAMPLE_DOC_CLASS );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_CLASS, w.toString(), true );
    }

    /**
     * Should handle javadoc comment for a class declaration.  */
    public void testClassDeclJavadocParsing2()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( SAMPLE_DOC_CLASS );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	assertEquals( SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );

    
    }

    /**
     * Should handle empty javadoc comment.  */
    public void testClassDeclJavadocParsing3()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "/**" + NEWLINE + 
					" *" + NEWLINE +
					" */" + NEWLINE +
					SAMPLE_CLASS );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	assertEquals( "", jdoc.text(), true );
	    
    }

    /**
     * Should handle two javadoc comments for a single class declaration.  */
    public void testClassDeclJavadocParsing4()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = 
	    getAST( SAMPLE_DOC + NEWLINE + SAMPLE_DOC + NEWLINE +
		    SAMPLE_CLASS );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_CLASS, w.toString(), true );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();
	// The space in the following is because of the space added in JavadocComment.add
	assertEquals( SAMPLE_DOC_CONTENT + NEWLINE + " " + SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );
    }

    /**
     * Should get a null javadoc comment for a class declaration when
     * javadoc parsing is disabled.  */
    public void testClassDeclJavadocParsing5()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( SAMPLE_DOC_CLASS, false );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	if( jdoc != null ) {
	    assertTrue( "should get null javadoc, instead got " + jdoc.text(), 
			false );
	}
    }

    /**
     * Should handle javadoc comments for an interface declaration.  */
    public void testInterfaceDeclJavadocParsing1()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( SAMPLE_DOC + NEWLINE + 
					SAMPLE_INTERFACE );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_INTERFACE, w.toString(), true );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	assertEquals( SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );
    }

    /**
     * Should handle javadoc comments for an augmenting method declaration.  */
    public void testAugmentingMethodDeclJavadocParsing1()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( SAMPLE_DOC + NEWLINE + 
					SAMPLE_AUG_METHOD );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_AUG_METHOD, w.toString(), true );

	ArrayList decls = unit.tlMethods();

	assertEquals( "one type declaration", 1, decls.size() );

	JavadocComment jdoc = 
	    ((Annotatable) decls.get(0)).javadocComment();

	assertEquals( SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );
    }

    /**
     * Should handle javadoc comments for an inner class declaration.  */
    public void testInnerClassJavadocParsing1()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Outer {" + NEWLINE +
					SAMPLE_DOC + NEWLINE +
					"    public class Inner {" + NEWLINE +
					"    }" + NEWLINE + 
					"}");

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one outer type declaration", 1, decls.length );

	ArrayList inners = decls[0].inners();

	assertEquals( "one inner type declaration", 1, inners.size() );

	JavadocComment jdoc = 
	    ((Annotatable) inners.get(0)).javadocComment();

	assertEquals( SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );
    }

    /**
     * Should handle javadoc comments for a method declaration.  */
    public void testMethodJavadocParsing1()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE +
					SAMPLE_DOC + NEWLINE +
					"    public void m() {" + NEWLINE +
					"    }" + NEWLINE + 
					"}");

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	ArrayList methods = decls[0].methods();

	assertEquals( "one method declaration", 1, methods.size() );

	JavadocComment jdoc = 
	    ((Annotatable) methods.get(0)).javadocComment();

	assertEquals(SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );
    }

    /**
     * Should handle javadoc comments for a field declaration.  */
    public void testFieldJavadocParsing1()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE +
					SAMPLE_DOC + NEWLINE +
					"    public int f;" + NEWLINE +
					"}");

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JFieldDeclarationType[] fields = decls[0].fields();

	assertEquals( "one field declaration", 1, fields.length );

	JavadocComment jdoc = fields[0].javadocComment();

	assertEquals( SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );
    }

    /**
     * Should handle javadoc comments for an initializer declaration.  */
    public void testInitializerJavadocParsing1()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE +
					SAMPLE_DOC + NEWLINE +
					"    {" + NEWLINE +
					"        i = 0;" + NEWLINE + 
					"    }" + NEWLINE +
					"    int i;" + NEWLINE +
					"}" );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JPhylum[] initializers = decls[0].fieldsAndInits();

	JavadocComment jdoc = 
	    ((Annotatable) initializers[0]).javadocComment();

	assertEquals( SAMPLE_DOC_CONTENT + NEWLINE, jdoc.text(), true );
    }

    /**
     * Should ignore javadoc comment preceding the package declaration.  */
    public void test_ignore_misplaced_javadoc_parsing1()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( SAMPLE_DOC + NEWLINE + "package foo;" 
					+ NEWLINE + SAMPLE_CLASS );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	assertTrue( "no javadoc", jdoc==null );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 12 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
	    
    }

    /**
     * Should ignore javadoc comment between "public" and "class".  */
    public void test_ignore_misplaced_javadoc_parsing2()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public " + SAMPLE_DOC + 
					" class Sample { }" );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	assertTrue( "no javadoc", jdoc==null );
	    
	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 2, character 27 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comment between "class" and "Sample".  */
    public void test_ignore_misplaced_javadoc_parsing3()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "class " + SAMPLE_DOC + 
					" Sample { }" );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	assertTrue( "no javadoc", jdoc==null );
	    
	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 2, character 21 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comment between "Sample" and "{".  */
    public void test_ignore_misplaced_javadoc_parsing4()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample " + SAMPLE_DOC + 
					" { }" );

	JTypeDeclarationType[] decls = unit.typeDeclarations();

	assertEquals( "one type declaration", 1, decls.length );

	JavadocComment jdoc = decls[0].javadocComment();

	assertTrue( "no javadoc", jdoc==null );
	    
	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 2, character 14 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comment inside empty class body.  */
    public void test_ignore_misplaced_javadoc_parsing5()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + 
					SAMPLE_DOC + NEWLINE + "}" );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_CLASS, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 1 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore evil javadoc comment inside empty class body
     * while successfully parsing the class.  */
    public void test_ignore_misplaced_javadoc_parsing6()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE + 
					"    /********/" + NEWLINE +
					"}" );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_CLASS, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 1 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comments inside an interface declaration.  */
    public void test_ignore_misplaced_javadoc_parsing7()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST("public abstract interface Sample {" + 
				       NEWLINE +
				       "    /******/" + NEWLINE +
				       "}");

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_INTERFACE, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 1 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comments inside an augmenting method
     * declaration.  */
    public void test_ignore_misplaced_javadoc_parsing8()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public void Sample.m() {" + NEWLINE +
					"    /******/" + NEWLINE +
					"}" );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_AUG_METHOD, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 1 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }    

    /**
     * Should ignore javadoc comments inside an empty inner class
     * declaration.  */
    public void test_ignore_misplaced_javadoc_parsing9()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Outer {" + NEWLINE +
					"    public class Inner {" + NEWLINE +
					"          /******/" + NEWLINE +
					"    }" + NEWLINE + 
					"}");

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_INNER_CLASS, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 4, character 5 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comments inside a method declaration.  */
    public void test_ignore_misplaced_javadoc_parsing10()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE +
					"    public void m() {" + NEWLINE +
					"          /******/" + NEWLINE +
					"    }" + NEWLINE + 
					"}");

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_METHOD, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 4, character 5 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comments inside a field declaration.  */
    public void test_ignore_misplaced_javadoc_parsing11()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE +
					"    public /** ignored */ int f;" + 
					NEWLINE +
					"}");

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_FIELD, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 2, character 32 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comments inside an initializer declaration.  */
    public void test_ignore_misplaced_javadoc_parsing12()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE +
					"    {" + NEWLINE +
					"        /** ignored */" + NEWLINE +
					"        i = 0;" + NEWLINE + 
					"    }" + NEWLINE +
					"    int i;" + NEWLINE +
					"}" );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_INIT, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 5, character 5 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    /**
     * Should ignore javadoc comments inside an if-else statement.  */
    public void test_ignore_misplaced_javadoc_parsing13()  
	throws RecognitionException, TokenStreamException 
    {
	JCompilationUnitType unit = getAST( "public class Sample {" + NEWLINE +
					"  {" + NEWLINE +
					"    if( Math.abs(0.0) > 0.0 )" + 
					NEWLINE + 
					"        /** ignored */" + NEWLINE +
					"      Math.random();" + NEWLINE +
					"    else" + NEWLINE +
					"      Math.random();" + NEWLINE +
					"  }" + NEWLINE +
					"}" + NEWLINE );

	Writer w = new StringWriter();
	unit.accept( new MjcPrettyPrinter( w, new CModifier() ) );

	assertEquals( SAMPLE_INIT2, w.toString(), true );

	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 7 warning: Then part of if-statement should be enclosed in block statement [JCC 7.2, 7.4]" + NEWLINE + "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 7 warning: Else part of if-statement should be enclosed in block statement [JCC 7.2, 7.4]" + NEWLINE + "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 8, character 3 warning: A preceding javadoc documentation comment occurs in a position where it is treated as a regular comment." + NEWLINE, errorOut.toString(), true );
    }

    //---------------------------------------------------------------------
    // Initializer modifiers
    //---------------------------------------------------------------------

    public void test_initMods_1()  
	throws RecognitionException, TokenStreamException 
     {
	JCompilationUnitType unit = 
	    getAST( "public class Sample {" + NEWLINE +
		    "  int x;" + NEWLINE +
		    "  { x = 0; }" + NEWLINE +
		    "}" + NEWLINE );

	assertEquals( "", errorOut.toString(), true );

    }

    public void test_initMods_2()  
	throws RecognitionException, TokenStreamException 
     {
	JCompilationUnitType unit = 
	    getAST( "public class Sample {" + NEWLINE +
		    "  static int x;" + NEWLINE +
		    "  static { x = 0; }" + NEWLINE +
		    "}" + NEWLINE );

	assertEquals( "", errorOut.toString(), true );

    }

    public void test_initMods_3()  
	throws TokenStreamException 
    {
	try {
	    JCompilationUnitType unit = 
		getAST( "public class Sample {" + NEWLINE +
			"  int x;" + NEWLINE +
			"  public { x = 0; }" + NEWLINE +
			"}" + NEWLINE );
	    fail( "should get exception" );
	} catch( RecognitionException e ) {
	    assertEquals( "unexpected token: public", e.getMessage(), true );
	}
    }

    public void test_initMods_4()  
	throws RecognitionException, TokenStreamException 
    {
	try {
	    JCompilationUnitType unit = 
		getAST( "public class Sample {" + NEWLINE +
			"  static int x;" + NEWLINE +
			"  public static { x = 0; }" + NEWLINE +
			"}" + NEWLINE );
	    fail( "should get exception" );
	} catch( RecognitionException e ) {
	    assertEquals( "unexpected token: {", e.getMessage(), true );
	}
    }

    //---------------------------------------------------------------------
    // Initializer modifiers
    //---------------------------------------------------------------------

    // should handle /**/ as a legal comment
    public void test_comment_parsing_1()  
	throws RecognitionException, TokenStreamException 
     {
	JCompilationUnitType unit = 
	    getAST( "public class Sample {" + NEWLINE +
		    "  void m() { /**/ }" + NEWLINE +
		    "}" + NEWLINE );

	assertEquals( "", errorOut.toString(), true );

    }

    // should handle EOF coming on the same line as a single-line comment
    public void test_comment_parsing_2()  
	throws RecognitionException, TokenStreamException 
     {
	JCompilationUnitType unit = 
	    getAST( "public class Sample {" + NEWLINE +
		    "}" + NEWLINE +
		    "////" );

	assertEquals( "", errorOut.toString(), true );

    }    

    
    // -------------------------------------------------------------
    // Test generics
    // -------------------------------------------------------------
    
//     public void test_generic1() 
// 	throws RecognitionException, TokenStreamException 
//     {
// 	JCompilationUnitType unit = 
// 	    getAST( "public class Sample<T> { T foo() { return null; } }" );

// 	assertEquals( "", errorOut.toString(), true );

//     }

//     public void test_generic2() 
// 	throws RecognitionException, TokenStreamException 
//     {
// 	JCompilationUnitType unit = 
// 	    getAST( "public class Sample<T { T foo() { return null; } }" );

// 	assertEquals( "the correct error message", errorOut.toString(), true );

//     }

    //---------------------------------------------------------------------
    // PRIVATE AND PROTECTED MEMBERS
    //---------------------------------------------------------------------

    /**
     * Establish lexers and parsers for the given source code,
     * including parsing of javadoc comments.  */
    protected void establishTest( String sourceCode ) {
	establishTest( sourceCode, true );
    }

    /**
     * Establish lexers and parsers for the given source code, parsing
     * javadoc comments according to the given boolean. */
    protected void establishTest( String sourceCode, 
				  boolean parseJavadocs ) {
	errorOut.setLength( 0 );
	Reader r = new StringReader( sourceCode );
	parsingController = new ParsingController( r, null );
	mjLexer = new MjcLexer( parsingController, true, true,
				false, // WMD: per default no universes
				compiler);
	docLexer = new JavadocLexer( parsingController );
	try {
	    parsingController.addInputStream( mjLexer, "multijava" );
	    parsingController.addInputStream( docLexer, "javadoc" );
	    parsingController.selectInitial( "multijava" );
	    if( !parseJavadocs ) {
		parsingController.discardAllTokensFor( "javadoc" );
	    }

	    final boolean ACCEPT_MULTIJAVA = true;
	    final boolean ACCEPT_GENERIC = false;
	    final boolean ACCEPT_RELAXED_MULTIJAVA = false;
	    final boolean ACCEPT_UNIVERSES = false; // WMD

	    parser = new MjcParser( compiler, 
				    parsingController.initialOutputStream(),
				    parsingController,
				    ACCEPT_GENERIC,
				    ACCEPT_MULTIJAVA,
				    ACCEPT_RELAXED_MULTIJAVA,
				    ACCEPT_UNIVERSES, // WMD
				    parseJavadocs );
	} catch( ParsingController.ConfigurationException e ) {
	    // want the exception to pass out to the test driver
	    // without having to worry about passing it explicitly
	    throw new RuntimeException( e.getMessage() );
	}
    }

    /**
     * Returns an AST for the source code represented by the given
     * string.  The AST is annotated with any javadocs annotations
     * that appear in the string.  */
    protected JCompilationUnitType getAST( String sourceCode ) 
	throws RecognitionException, TokenStreamException 
    {
	return getAST( sourceCode, true );
    }

    /**
     * Returns an AST for the source code represented by the given
     * string.  If the <code>parseJavadocs</code> flag is true, then
     * the AST is annotated with any javadocs that appear in the
     * string.  */
    protected JCompilationUnitType getAST( String sourceCode, 
					   boolean parseJavadocs )
	throws RecognitionException, TokenStreamException 
    {
	establishTest( sourceCode, parseJavadocs );
	return parser.jCompilationUnit();
    }

    private static MjcLexer mjLexer;
    private static JavadocLexer docLexer;
    private static MjcParser parser; 

    /** 
     * Used to switch lexers.  Aliases are created within lexer
     * instances. */
    private static ParsingController parsingController;

    public static final StringBuffer errorOut = new StringBuffer();

    protected static class TestCompiler extends Main {
	TestCompiler( StringBuffer errorOut, ModifierUtility modUtil ) {
	    super( modUtil );
	    this.errorOut = errorOut;
	    this.options = new MjcOptions();
	}
	public boolean parseComments() { return true; }
	public void reportTrouble( Exception trouble ) {
	    errorOut.append( trouble.toString() ).append( NEWLINE );
	}
	public boolean run( String[] args ) { return true; }
	public boolean verboseMode() { return true; }
	public int optimizeCode() { return 0; }
	public boolean RMJ() { return false; }

	private StringBuffer errorOut;
    }

    protected Main compiler;

    // spaces and newlines are to match the output of the pretty printer
	private static final String SAMPLE_DOC_CONTENT = " sample ";
    private static final String SAMPLE_DOC = 
	"/**" + NEWLINE +
	" *" + SAMPLE_DOC_CONTENT + "*/";
    private static final String SAMPLE_CLASS = NEWLINE + NEWLINE +
	"public class Sample {" + NEWLINE +
	"}" + NEWLINE;
    private static final String SAMPLE_INTERFACE = NEWLINE + NEWLINE +
	"public abstract interface Sample {" + NEWLINE +
	"}" + NEWLINE;
    private static final String SAMPLE_AUG_METHOD = NEWLINE +
	"public void Sample.m() {" + NEWLINE +
	"}" + NEWLINE;
    private static final String SAMPLE_DOC_CLASS = SAMPLE_DOC + SAMPLE_CLASS;
    private static final String SAMPLE_INNER_CLASS =  NEWLINE + NEWLINE +
	"public class Outer {" + NEWLINE +
	"  public class Inner {" + NEWLINE +
	"  }" + NEWLINE +
	"}" + NEWLINE;
    private static final String SAMPLE_METHOD =  NEWLINE + NEWLINE +
	"public class Sample {" + NEWLINE +
	"  public void m() {" + NEWLINE +
	"  }" + NEWLINE + NEWLINE +
	"}" + NEWLINE;
    private static final String SAMPLE_FIELD = NEWLINE + NEWLINE +
	"public class Sample {" + NEWLINE +
	"  public int f;" + NEWLINE +
	"}" + NEWLINE;
    private static final String SAMPLE_INIT =  NEWLINE + NEWLINE +
	"public class Sample {" + NEWLINE +
	"  {" + NEWLINE +
	"    i = (byte)0;" + NEWLINE + 
	"  }" + NEWLINE +
	"  int i;" + NEWLINE +
	"}" + NEWLINE;
    private static final String SAMPLE_INIT2 =  NEWLINE + NEWLINE +
	"public class Sample {" + NEWLINE +
	"  {" + NEWLINE +
	"    if (Math.abs((double)0.0) > (double)0.0) Math.random();" + 
	NEWLINE +
	"    else Math.random();" + NEWLINE +
	"  }" + NEWLINE +
	"}" + NEWLINE;
}
