/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * $Id: TestParsingController.java,v 1.13 2004/12/30 03:08:02 davidcok Exp $
 */

package org.multijava.mjc;

import antlr.RecognitionException;
import antlr.Token;
import antlr.TokenStream;
import antlr.TokenStreamException;

import org.multijava.javadoc.JavadocComment;
import org.multijava.util.compiler.PositionedError;
import org.multijava.util.testing.TestCase;

import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import org.multijava.util.compiler.TroubleReporter;

/**
 * Unit tests for ParsingController */
public class TestParsingController extends TestCase {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    public TestParsingController( String name ) {
	super( name );
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------

    /**
     * Should get an exception if the first argument is null.  */
    public void test_addInputStream_method1() {
	try {
	    establishLexers( SAMPLE_CLASS );
	    parsingController.addInputStream( null, "multijava" );
	    fail( "Expected ParsingController.KeyException" );
	} catch( ParsingController.KeyException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should get an exception if the second argument is null.  */
    public void test_addInputStream_method2() {
	try {
	    establishLexers( SAMPLE_CLASS );
	    parsingController.addInputStream( mjLexer, null );
	    fail( "Expected ParsingController.KeyException" );
	} catch( ParsingController.KeyException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should get an exception if we try to bind two input streams to
     * the same key.  */
    public void test_addInputStream_method3() {
	try {
	    establishLexers( SAMPLE_CLASS );
	    parsingController.addInputStream( mjLexer, "multijava" );
	    parsingController.addInputStream( docLexer, "multijava" );
	    fail( "Expected ParsingController.KeyException" );
	} catch( ParsingController.KeyException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should get an exception if we try to select based on a key that
     * is not bound.  */
    public void test_selectInitial_method1() {
	try {
	    establishTest( SAMPLE_CLASS, false );
	    parsingController.selectInitial( "foo" );
	    fail( "Expected ParsingController.KeyException" );
	} catch( ParsingController.KeyException e ) {
	    assertTrue( true );
	} catch( ParsingController.ConfigurationException e ) {
	    fail( "Expected ParsingController.KeyException" );
	}
    }

    /**
     * Should get an exception if we try to select more than once.  */
    public void test_selectInitial_method2() {
	try {
	    establishTest( SAMPLE_CLASS, false );
	    parsingController.selectInitial( "multijava" );
	    parsingController.selectInitial( "multijava" );
	    fail( "Expected ParsingController.ConfigurationException" );
	} catch( ParsingController.KeyException e ) {
	    fail( "Expected ParsingController.ConfigurationException" );
	} catch( ParsingController.ConfigurationException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should get an exception if we try to get the initial output
     * stream without setting an initial lexer.  */
    public void test_initialOutputStream_method1() {
	try {
	    establishTest( SAMPLE_CLASS, false );
	    parsingController.initialOutputStream();
	    fail( "Expected ParsingController.ConfigurationException" );
	} catch( ParsingController.ConfigurationException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should get an exception when requesting discarding for an
     * undefined key. */
    public void test_discardAllTokensFor_method1() {
	try {
	    establishTest( SAMPLE_CLASS );
	    parsingController.discardAllTokensFor( "foo" );
	    fail( "Expected ParsingController.KeyException" );
	} catch( ParsingController.KeyException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should get an exception when requesting discarding for a key
     * for which a token stream has already been exposed */
    public void test_discardAllTokensFor_method2() {
	try {
	    establishTest( SAMPLE_CLASS, true );
	    parsingController.discardAllTokensFor( "multijava" );
	    fail( "Expected ParsingController.KeyException" );
	} catch( ParsingController.KeyException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should find no javadoc comments when discarding them, even
     * though they are present in the source.  */
    public void test_discardAllTokensFor_method3() 
	throws ParsingController.KeyException, TokenStreamException 
    {
	establishTest( SAMPLE_DOC_CLASS );
	parsingController.discardAllTokensFor( "javadoc" );

	Token last = discardAllMJTokens();

	TokenStream ts = parsingController.streamForBefore( "javadoc", last );
	last = ts.nextToken();
	assertTrue( "should be end-of-file: " + last, 
		    last.getType() == Token.EOF_TYPE );
    }

//      /**
//       * For a doubly nested language with tokens discarded for the
//       * middle language, tokens of the outer language should be used as
//       * indices for the inner language. */
//      public void test_discardAllTokensFor_methodX() {
//  	fail( "test not written" );
//      }

    /**
     * Should get an exception when requesting a stream for an
     * undefined key. */
    public void test_streamForBefore_method1() {
	try {
	    establishTest( SAMPLE_CLASS );
	    TokenStream ts = parsingController.streamForBefore( "foo", null );
	    fail( "Expected ParsingController.KeyException" );
	} catch( ParsingController.KeyException e ) {
	    assertTrue( true );
	}
    }

    /**
     * Should be able to request a stream twice for the same key and
     * index, but second stream should yield end-of-file token. */
    public void test_streamForBefore_method2() 
	throws ParsingController.KeyException, TokenStreamException
    {
	establishTest( SAMPLE_CLASS );
	Token last = discardAllMJTokens();
	parsingController.streamForBefore( "javadoc", last );
	TokenStream ts = parsingController.streamForBefore( "javadoc",
							    last );
	last = ts.nextToken();
	assertTrue( "should be end-of-file: " + last, 
		    last.getType() == Token.EOF_TYPE );
    }

    /**
     * Should successfully lex the MultiJava tokens in the input
     * stream.  No javadoc tokens are in the input stream.  */
    public void test_FilteredTokenStream_class1() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_CLASS );
	TokenStream ts = parsingController.initialOutputStream();
	parsingController.discardAllTokensFor( "javadoc" );

	StringBuffer result = new StringBuffer();
	Token tok;
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    result.append( tok.getText() + " " );
	}
	assertEquals( SAMPLE_CLASS, result.toString(), true );
    }
    
    /**
     * Should successfully lex the javadoc tokens in the input
     * stream, but no javadoc tokens are in the input stream. */
    public void test_FilteredTokenStream_class2() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_CLASS );
	Token last = discardAllMJTokens();
	TokenStream ts = parsingController.streamForBefore( "javadoc", last );

	StringBuffer result = new StringBuffer();
	Token tok;
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    result.append( tok.getText() + " " );
	}
	assertEquals( "", result.toString(), true );
    }
    
    /**
     * Should successfully lex just the multijava tokens in the input
     * stream.  */
    public void test_FilteredTokenStream_class3() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC_CLASS );
	TokenStream ts = parsingController.initialOutputStream();
	parsingController.discardAllTokensFor( "javadoc" );

	StringBuffer result = new StringBuffer();
	Token tok;
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    result.append( tok.getText() + " " );
	}
	assertEquals( SAMPLE_CLASS, result.toString(), true );
    }
    
    /**
     * Should successfully lex just the javadoc tokens in the input
     * stream.  */
    public void test_FilteredTokenStream_class4() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC_CLASS );
	Token last = discardAllMJTokens();
	TokenStream ts = parsingController.streamForBefore( "javadoc", last );

	StringBuffer result = new StringBuffer();
	Token tok;
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    result.append( tok.getText() + " " );
	}
	assertEquals( EXPECTED_SAMPLE_DOC, result.toString(), true );
    }
    
    /**
     * Should successfully lex both the multijava and javadoc tokens
     * in the input stream.  */
    public void test_FilteredTokenStream_class5() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC_CLASS );
	
	TokenStream mjTokStream = 
	    parsingController.initialOutputStream();

	StringBuffer mjResult = new StringBuffer();
	StringBuffer docResult = new StringBuffer();
	Token tok;

	// get the first MultiJava token
	tok = mjTokStream.nextToken();
	mjResult.append( tok.getText() + " " );

	// get the javadocs preceding the first MultiJava token
	TokenStream docTokStream = 
	    parsingController.streamForBefore( "javadoc", tok );

	// lex the javadocs
	while( (tok = docTokStream.nextToken()).getType() != Token.EOF_TYPE ) {
	    docResult.append( tok.getText() + " " );
	}
	assertEquals( EXPECTED_SAMPLE_DOC, docResult.toString(), true );

	// lex the remaining MultiJava
	while( (tok = mjTokStream.nextToken()).getType() != Token.EOF_TYPE ) {
	    mjResult.append( tok.getText() + " " );
	}
	assertEquals( SAMPLE_CLASS, mjResult.toString(), true );
    }
    
    /**
     * Should successfully lex both the multijava and javadoc tokens
     * in the input stream.  */
    public void test_FilteredTokenStream_class6() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC2 + NEWLINE + SAMPLE_CLASS );
	
	TokenStream mjTokStream = 
	    parsingController.initialOutputStream();

	StringBuffer mjResult = new StringBuffer();
	StringBuffer docResult = new StringBuffer();
	Token tok;

	// get the first MultiJava token
	tok = mjTokStream.nextToken();
	mjResult.append( tok.getText() + " " );

	// get the javadocs preceding the first MultiJava token
	TokenStream docTokStream = 
	    parsingController.streamForBefore( "javadoc", tok );

	// lex the javadocs
	while( (tok = docTokStream.nextToken()).getType() != Token.EOF_TYPE ) {
	    docResult.append( tok.getText() + " " );
	}
	assertEquals( EXPECTED_SAMPLE_DOC2, docResult.toString(), true );

	// lex the remaining MultiJava
	while( (tok = mjTokStream.nextToken()).getType() != Token.EOF_TYPE ) {
	    mjResult.append( tok.getText() + " " );
	}
	assertEquals( SAMPLE_CLASS, mjResult.toString(), true );
    }
    
    /**
     * Should successfully lex both the multijava and javadoc tokens
     * in the input stream.  */
    public void test_FilteredTokenStream_class7() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC3 + NEWLINE + SAMPLE_CLASS );
	
	TokenStream mjTokStream = 
	    parsingController.initialOutputStream();

	StringBuffer mjResult = new StringBuffer();
	StringBuffer docResult = new StringBuffer();
	Token tok;

	// get the first MultiJava token
	tok = mjTokStream.nextToken();
	mjResult.append( tok.getText() + " " );

	// get the javadocs preceding the first MultiJava token
	TokenStream docTokStream = 
	    parsingController.streamForBefore( "javadoc", tok );

	// lex the javadocs
	while( (tok = docTokStream.nextToken()).getType() != Token.EOF_TYPE ) {
	    docResult.append( tok.getText() + " " );
	}
	assertEquals( EXPECTED_SAMPLE_DOC3, docResult.toString(), true );

	// lex the remaining MultiJava
	while( (tok = mjTokStream.nextToken()).getType() != Token.EOF_TYPE ) {
	    mjResult.append( tok.getText() + " " );
	}
	assertEquals( SAMPLE_CLASS, mjResult.toString(), true );
    }
    
    /**
     * Should successfully lex just the multijava tokens in the input
     * stream despite the javadoc comment ending in '**' + '/'.  */
    public void test_FilteredTokenStream_class8() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( "/**" + NEWLINE +
		       " * sample " + NEWLINE +
		       " **/" + NEWLINE +
		       SAMPLE_CLASS );
	TokenStream ts = parsingController.initialOutputStream();
	parsingController.discardAllTokensFor( "javadoc" );

	StringBuffer result = new StringBuffer();
	Token tok;
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    result.append( tok.getText() + " " );
	}
	assertEquals( SAMPLE_CLASS, result.toString(), true );
    }
    
    /**
     * Should successfully give the line numbers of each token.  */
    public void test_getLine_method1() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC_CLASS );
	TokenStream ts = parsingController.initialOutputStream();

	Token tok;
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    assertEquals( 3, parsingController.line() );
	}
    }
    
    /**
     * Should successfully give the line numbers of each token.  */
    public void test_getLine_method2() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC_CLASS );
	Token tok = discardAllMJTokens();
	TokenStream ts = parsingController.streamForBefore( "javadoc", tok );

	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    assertEquals( 3, parsingController.line() );
	}
    }
    
    /**
     * Should successfully give the column numbers of each token.  */
    public void test_getColumn_method1() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC_CLASS );
	TokenStream ts = parsingController.initialOutputStream();

	Token tok;
	StringBuffer result = new StringBuffer();
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    result.append( parsingController.column() + " " );
	}
	assertEquals( "7 13 20 22 24 ", result.toString(), true );
    }
    
    /**
     * Should successfully give the column numbers of each token.  */
    public void test_getColumn_method2() 
	throws ParsingController.ConfigurationException, TokenStreamException
    {
	establishTest( SAMPLE_DOC_CLASS );
	Token tok = discardAllMJTokens();
	TokenStream ts = parsingController.streamForBefore( "javadoc", tok );

	StringBuffer result = new StringBuffer();
	while( (tok = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    result.append( parsingController.column() + " " );
	}
	assertEquals( "25 25 ", result.toString(), true );
    }
        
    //---------------------------------------------------------------------
    // COMMON FIXTURE
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // PRIVATE MEMBERS
    //---------------------------------------------------------------------
    
    /**
     * Establish lexers and the controller for the given source code,
     * without adding streams to the controller. */
    private void establishLexers( String sourceCode ) {
	Reader r = new StringReader( sourceCode );
	parsingController = new ParsingController( r, null );
	mjLexer = new MjcLexer( parsingController, true, true,
				false, // WMD: no universes
				SIMPLE_REPORTER );
	docLexer = new JavadocLexer( parsingController );
    }

    /**
     * Establish lexers and controller for the given source code. 
     * 
     * @param sourceCode 	the source code to be lexed
     * @param selectInitial 	indicates whether an initial stream should 
     *				be selected
     */
    private void establishTest( String sourceCode, boolean selectInitial ) {
	try {
	establishLexers( sourceCode );
	parsingController.addInputStream( mjLexer, "multijava" );
	parsingController.addInputStream( docLexer, "javadoc" );
	if( selectInitial ) {
	    parsingController.selectInitial( "multijava" );
	}
	} catch( ParsingController.ConfigurationException e ) {
	    // convert to unchecked so test methods don't need to handle
	    throw new RuntimeException( "ConfigurationException: " + 
					e.getMessage() );
	}
    }

    /**
     * Establish lexers and controller for the given source code. */
    private void establishTest( String sourceCode ) {
	establishTest( sourceCode, true );
    }

    /**
     * Discards all MultiJava tokens from the input and returns the
     * end-of-file token.  */
    private Token discardAllMJTokens() {
	try {
	TokenStream ts = parsingController.initialOutputStream();
	Token last;
	while( (last = ts.nextToken()).getType() != Token.EOF_TYPE ) {
	    // eats all multijava tokens
	}
	return last;
	} catch( TokenStreamException e ) {
	    // convert to unchecked so test methods don't need to handle
	    throw new RuntimeException( "TokenStreamException: " + 
					e.getMessage() );
	} catch( ParsingController.ConfigurationException e ) {
	    // convert to unchecked so test methods don't need to handle
	    throw new RuntimeException( "ConfigurationException: " + 
					e.getMessage() );
	}
    }

    private MjcLexer mjLexer;
    private JavadocLexer docLexer;
    private ParsingController parsingController;

    // spaces in these examples are important as we are matching the
    // output of the lexer
    private static final String SAMPLE_CLASS = "public class Sample { } ";
    private static final String SAMPLE_DOC = 
	"/**" + NEWLINE +
	" * Sample  doc. */";
    private static final String SAMPLE_DOC2 = 
	"/**" + NEWLINE +
	" * Sample" + NEWLINE +
	" * doc. */";
    private static final String EXPECTED_SAMPLE_DOC2 = 
	" Sample" + NEWLINE +
	"  doc.  */ ";
    private static final String SAMPLE_DOC3 = 
	"/**" + NEWLINE +
	" * Sample" + NEWLINE +
	" * doc. " + NEWLINE + 
	" */";
    private static final String EXPECTED_SAMPLE_DOC3 = 
	" Sample" + NEWLINE +
	"  doc. " + NEWLINE + " */ ";
    private static final String EXPECTED_SAMPLE_DOC =
	" Sample  doc.  */ ";	// we expect the closing token from the lexer
    private static final String SAMPLE_DOC_CLASS = 
	SAMPLE_DOC + NEWLINE + SAMPLE_CLASS;
    private static final String IGNORE_SAMPLE_DOC_CLASS =
	"/**" + NEWLINE +
	" * ignored */" + NEWLINE + // not really a javadoc because of package
	"package foo;" + NEWLINE +
	SAMPLE_DOC + NEWLINE +
	SAMPLE_CLASS;

    private static final TroubleReporter SIMPLE_REPORTER =
	new TroubleReporter() {
	    public void reportTrouble( Exception p ) { }
	};
}
