/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TypeLoader.java,v 1.31 2006/12/20 06:15:52 perryjames Exp $
 */

package org.multijava.mjc;

import java.io.File;
import java.util.HashMap;

import org.multijava.util.classfile.ClassInfo;
import org.multijava.util.classfile.ClassPath;
import org.multijava.util.classfile.ClassFileReadException;
import org.multijava.util.MjcHashRelation;

/**
 * This class acts as a symbol table and a cache for types, type
 * signatures, and external generic functions.  It includes methods to
 * load unknown type signatures.  It maintains a mapping from fully
 * qualified names to the CClass instances representing their
 * signatures.  Specifically the mapping is from a fully qualified
 * type name to a set of instances that can be sorted.  In MultiJava
 * that set is a singleton set; however, for tools like JML where a
 * single fully qualified name can refer to a specification and one or
 * more refinements to that specification, the set may contain
 * multiple items.  
 */
public class TypeLoader extends org.multijava.util.Utils implements CClass.Observer {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    public static TypeLoader getSingleton() {
	if (singletonInstance == null) {
	    singletonInstance = new TypeLoader();
	}
	return singletonInstance;
    }

    private TypeLoader() {
	this(new MjcFileFinder());
    }

    /**
     * This constructor is only for use by subclasses.  Clients should use 
     * the factory method getSingleton().
     * @see #getSingleton()
     */
    protected TypeLoader(FileFinder fileFinder) {
	this.fileFinder = fileFinder;
	compilationSessionTypeCache = new MjcHashRelation(2000);
	vmSessionTypeCache = new MjcHashRelation(2000);
    }

    // -------------------------------------------------------------
    // HANDLING OF LIGHTWEIGHT TYPE REPRESENTATIONS (JUST NAMES WITHOUT SIGS)
    // -------------------------------------------------------------

    /**
     * Returns the lightweight type representation of the class with
     * the given name, generating a new one if necessary.
     * @param	name	the name of the class
     * @param	isFullyQualified  the name is known to be fully qualified
     * @return an object representing the type of the named class
     * (singleton within this compilation session if isFullyQualified)
     *
     * <pre><jml>
     * requires name != null;
     * modifies compilationSessionRepCache;
     * ensures isFullyQualified ==> 
     *   compilationSessionRepCache.containsKey(name) ||
     *   vmSessionRepCache.containsKey(name);
     * </jml></pre>
     */ 
    public CClassType getTypeRep(String name, CUniverse univ,
                                 CClassType[][] arguments,
                                 boolean isFullyQualified) {
	CClassType type;
	if (isFullyQualified) {
	    type = (CClassType)vmSessionRepCache.get(name);
	    if (type == null) {
		type = (CClassType)compilationSessionRepCache.get(name);
	    }

            if ((type != null) && !mayShare(type, univ, arguments)) {
                type = null;
            }

	    if (type == null) {
		type = new CClassFQNameType(name,univ,arguments);
		if (isTrusted(name)) {
		    vmSessionRepCache.put(name, type);
		} else {
		    compilationSessionRepCache.put(name, type);
		}		
	    }

	} else {
	    type = new CClassNameType(name, univ, arguments);
	}
	return type;
    }

    /**
     * Returns whether the given <code>type</code> can be shared by another
     * type having the given <code>univ</code> and type <code>arguments</code>.
     * This is the case iff the Universe modifier and the type arguments of the
     * type are equivalent to the ones provided here.
     *
     * @param type       The target type.
     * @param univ       The Universe modifier.
     * @param arguments  The generic type arguments.
     * @return           Whether the given type may be shared.
     */
    private static boolean mayShare(
            /*@ non_null @*/ CClassType type,
            /*@ nullable @*/ CUniverse univ,
            /*@ non_null @*/ CClassType[][] arguments) {
        CClassType[][] typeArguments = type.getAllArguments();

        // If all the arguments are identical, we are done.
        if (arguments == typeArguments) {
            return true;
        }

        // The length of the type arguments do not necessarily match, so let's
        // first check the elements both arrays have in common...
        int i = arguments.length - 1;
        int j = typeArguments.length - 1;
        for (; (i >= 0) && (j >= 0); i--, j--) {
            // If this set of arguments is identical, we go to the next round.
            if (arguments[i] == typeArguments[j]) {
                continue;
            }

            // If we have arguments on one side but not on the other, the type
            // cannot be shared.

            if ((arguments[i] == null) && (typeArguments[j].length > 0)) {
                return false;
            }

            if ((typeArguments[j] == null) && (arguments[i].length > 0)) {
                return false;
            }

            // If we have arguments on both side but the number of arguments
            // do not match, the type cannot be shared.
            if (arguments[i].length != typeArguments[j].length) {
                return false;
            }

            // A this point, the number of arguments matches, so let's check
            // them individually.

            //@ assert arguments[i].length == arguments[j].length;
            for (int k = 0; k < arguments[i].length; k++) {
                if (arguments[i][k] != typeArguments[j][k]) {
                    return false;
                }
            }
        }

        // If we have additional type arguments on one side without any
        // counterpart on the other side, the type cannot be shared.

        for (; i >= 0; i--) {
            if ((arguments[i] != null) && (arguments[i].length > 0)) {
                return false;
            }
        }
        for (; j >= 0; j--) {
            if ((typeArguments[j] != null) && (typeArguments[j].length > 0)) {
                return false;
            }
        }

        return true;
    }

    public CClassType lookupTypeRep(String qName) {
	CClassType typ = (CClassType) vmSessionRepCache.get( qName );
	if (typ == null) {
	    // not found in VM session cache, so check compilation
	    // session cache
	    typ = (CClassType) compilationSessionRepCache.get( qName );
	}
	return typ;
    }
    /**
     * Adds the given type representation, which must represent a
     * fully qualified type name, to the cache.
     *
     * <pre><jml>
     * requires qName != null && typeRep != null &&
     *          (* all separators in qName are '/' not '.' *);
     * modifies compilationSessionRepCache, vmSessionRepCache;
     * ensures compilationSessionRepCache.containsKey(qName) ||
     *         vmSessionRepCache.containsKey(qName);
     * </jml></pre>
     *
     */
    public void addTypeRep(String qName, CClassType typeRep) {
	if (isTrusted(qName)) {
	    vmSessionRepCache.put(qName, typeRep);
	} else {
	    compilationSessionRepCache.put(qName, typeRep);
	}
	
    }

    // -------------------------------------------------------------
    // HANDLING OF TYPE SIGNATURES
    // -------------------------------------------------------------

    public /*@ pure @*/ ClassPath.ClassDescription findSourceFile(String name) {
	return fileFinder.findSourceFile(name);
    }

    /** 
     * This method finds a file for the given class name; it looks for
     * both a java file on the source path and a class file on the
     * class path, returning the newer one if both are found (and null
     * if neither is found).  */
    //@ requires name != null; 
    public ClassPath.ClassDescription find(String name) {
	return fileFinder.find(name);
    }

    /**
     * Adds the given source class to the table of loaded classes. 
     * Returns true if the class was not declared in more than one 
     * source file and false otherwise.
     *
     * <pre><jml>
         requires typ != null;
	 modifies compilationSessionTypeCache, vmSessionTypeCache;
     * </jml></pre>
     */
    public boolean addToTypeCache( CClass typ ) {
	String qName = typ.qualifiedName();
	boolean isBinary = typ instanceof CBinaryClass;
	CClass prevTyp = lookupType( qName );
	if (prevTyp != null) {
	    if (prevTyp.getCompiler() != typ.getCompiler() ) {
		// Different compilation session: remove it 
		removeFromTypeCache(prevTyp);
	    }
	}

	boolean isTrusted = isTrusted(qName);
	CClass last;
	if (isBinary) {
	    // binary classes are ready to use as symbol tables once
	    // their constructor completes, so we just put them in the
	    // appropriate cache
	    if (isTrusted) {
		 last = (CClass) vmSessionTypeCache.put(qName, typ);
	    } else {
		 last = (CClass) compilationSessionTypeCache.put(qName, typ);
	    }
	} else {
	    // Source classes must be processed a bit before they can
	    // be used as symbol tables, so we put them in the
	    // compilation session cache.  (That way if they aren't
	    // processed far enough this session they can be discarded
	    // to be reprocessed in the next session.)  We register as
	    // an observer for the trusted ones so we can move those
	    // to the VM session cache when they are ready.
	    last = (CClass) compilationSessionTypeCache.put(qName, typ);
	    if (isTrusted) {
		typ.register(this);
	    }
	}
	
	return checkUniqueness(typ,last);
    }

    /**
     * Returns true if typ is a legal unique type declaration.
     * @param typ  a type to be added to the cache
     * @param last the previous cache entry mapped to by typ.qualifiedName()
     */
    protected /*@ pure @*/ boolean checkUniqueness(CClass typ, CClass last) {
	// The last two disjuncts are needed to exclude the case when 
	// a type has been redeclared within the same source file.
	// Redeclaration in the same lexical context is checked explicitly 
	// when typechecking the context.
	return !(last instanceof CSourceClass) 
	    || typ.isNestedType() 
	    || last.sourceFile() == typ.sourceFile();
    }

    /**
     * Called by Main when the given file, from which the given
     * compilation unit AST was derived, did not contain an expected
     * result; this method removes all cached info. for the file.
     */
    protected void forgetEverythingAbout(File f, JCompilationUnitType cunit) {
	// Note that f is used in overriding method
	if (cunit == null) {
	    return;
	}
	CSourceClass[] allTypeSignatures = cunit.allTypeSignatures();
	for (int i = 0; i < allTypeSignatures.length; i++) {
	    removeFromTypeCache(allTypeSignatures[i]);
	    CClassNameType typeRep = 
		(CClassNameType) getTypeRep(allTypeSignatures[i].qualifiedName(), null, CClassType.EMPTY_ARG,true);
	    typeRep.forgetChecking();
	}
    }

    /**
     * Removes the qualified name of the given source class from 
     * the table of loaded classes (so it can be reloaded without 
     * causing conflicts). This is only done if the conflict 
     * occurs because the names were added during two different 
     * compilation sessions.
     *
     * <pre><jml>
     *    requires typ != null;
     *    modifies compilationSessionTypeCache, vmSessionTypeCache;
     * </jml></pre>
     */
    protected void removeFromTypeCache( CClass typ ) {
	compilationSessionTypeCache.remove(typ.qualifiedName());
	vmSessionTypeCache.remove(typ.qualifiedName());
    }

    /**
     * Returns the first CClass instance in the set mapped to by the
     * given fully qualified name, or null if no such instance exists.
     *
     * <pre><jml>
     *  requires qName != null && 
     *	    (* any separators in qName are slashes ('/'), not dots ('.') *);
     * </jml></pre>
     */
    public CClass lookupType(String qName) {
	CClass typ = (CClass) vmSessionTypeCache.get( qName );
	if (typ == null) {
	    // not found in VM session cache, so check compilation
	    // session cache
	    typ = (CClass) compilationSessionTypeCache.get( qName );
	}
	return typ;
    }

    public CClass loadType(String qName) {
	// look in the cache
	// System.out.println("LOADCLASS " + qName);
	CClass typ = lookupType( qName );

	if (typ == null) {
	    // not cached, so try to load it
	    // System.out.println("NOTCACHED " + qName);

	    ClassPath.ClassDescription desc = fileFinder.find(qName);
	    if (desc == null)  {
		// System.out.println("DESC NULL " + qName);
		// no appropriate file exists to define the class
		typ = CClass.CLS_UNDEFINED;
	    } else if (desc.isClass()) {
		// System.out.println("DESC CLASS " + qName);
		ClassInfo info = createClassInfo(qName);
                // The createClassInfo method also adds the instance
                // to the cache.
                typ = info == null ? CClass.CLS_UNDEFINED : lookupType(qName);
		
		// System.out.println("DESC CLASS " + qName + " " +
		// (info != null) + " " + (typ != null)); 
	    } else {
		// System.out.println("DESC SOURCE " + qName);
		File f = ((ClassPath.FileClassDescription)(desc)).file();

		// This triggers compiling of the source code for
		// the class up to the current pass of the main
		// compiler invocation.  Remaining passes handle
		// this new class along with all the others being
		// compiled.  At the end of the parsing pass the
		// newly loaded class is added to
		// compilationSessionTypeCache by a call to
		// addToTypeCache (declared above) which must be
		// called by CTopLevel.addClass .
		CTopLevel.getCompiler().catchUpType(f,qName); 
		typ = lookupType(qName);

		// System.out.println("COMP " + typ.qualifiedQName()
		// + " " + qName);
	    }

	    // a bit of defensive programming to make sure we don't
	    // forget to set typ to non-null value in some branch
	    // above
	    if (typ == null) {
		typ = CClass.CLS_UNDEFINED;
	    }

	    // System.out.println("PUT " + qName + " " + (typ!=null &&
	    // typ != CClass.CLS_UNDEFINED));
	    addToTypeCache(typ);
	} 
	// else {
	//     System.out.println("CACHED " + qName + " " +
	//     (typ!=CClass.CLS_UNDEFINED));
	// }
	
	return typ;
    }

    /**
     * Creates and returns a class info object by reading the symbol
     * file for the class with the given fully qualified name
     * <code>qName</code>.  If no symbol file is found, null is
     * returned. This method also create a binary class object and
     * adds it to the system cache. This is a factory method that may
     * be overridden by subclasses.
     */
    protected ClassInfo createClassInfo(String qName) {
	ClassInfo info;
	try {
	    info = ClassPath.getClassInfo(qName, true);
	} catch (ClassFileReadException e) {
	    CTopLevel.getCompiler().reportTrouble(e);
	    info = null;
	}
		
        if (info != null) {
            new CBinaryClass(CTopLevel.getFromClassesCompiler(), info);
        }
        return info;
    }

    /**
     * Reloads the specified class if this class has been defined
     * during a previous compilation session.  Note that this method
     * can return either null or CLS_UNDEFINED or a valid CClass
     * value.
     */
    public CClass reloadType(CClass clazz) {
	if (clazz == null) {
	    return null;
	} 
	Main clazzCompiler = clazz.getCompiler();

	if (clazzCompiler == CTopLevel.getCompiler()) { 
	    // Same compilation session: keep it 
	    return clazz;
	} else {
	    // Loaded in a previous compilation session: reload it
	    return loadType(clazz.qualifiedName());
	}
    }

    /**
     * Checks whether a class with the given name has already been successfully
     * loaded. 
     * @return true iff the class has already been loaded.
     * @param qName the fully qualified name of the type with slash separators
     *			not dots
     */
    public boolean isTypeLoaded( String qName ) {
	CClass typ = lookupType( qName );
	return typ != null && typ != CClass.CLS_UNDEFINED;
    }

    /**
     * Checks whether a class with the given name is available in the class
     * path.  Note that this is true if the class has already been successfully
     * loaded or if there is a file that can be expected to hold its
     * definition.  The file may actually not hold a valid definition, so
     * classExists is not a guarantee that the class with the given name 
     * actually does exist.
     * @return true iff the class is in the class path
     * @param qName the fully qualified name of the type with slash separators
     *			not dots
     */
    public boolean classExists( String qName ) {
    // System.out.println("CLASSEXISTS? " + qName + " " + isTypeLoaded(qName) + " " + (fileFinder.find(qName) != null));
	return isTypeLoaded(qName)
	    || (fileFinder.find(qName) != null);
    }

    /**
     * This method is needed so it can be overridden by JML in order to 
     * handle classes that have been refined, e.g., when its 
     * specifications have intentionally been declared in a separate 
     * file.
     * In MultiJava, this method always returns isUnique since addToTypeCache 
     * has already determined whether the Type has already been declared 
     * in another file. 
     */
    public boolean isDeclaredInDifferentSourceFiles( boolean isUnique, 
						     CClass clazz ) {
	return ! isUnique;
    }

    /**
     * Implements CClass.Observer by moving typ from
     * compilationSessionTypeCache to vmSessionTypeCache if typ is
     * trusted.
     *
     * <pre><jml>
     * also
     * old String qName = typ.qualifiedName();
     * requires typ != null && isTrusted(qName);
     * ensures \old(compilationSessionTypeCache.getImage(qName).contains(typ))
     *         ==> !compilationSessionTypeCache.getImage(qName).contains(typ) &&
     *             vmSessionTypeCache.getImage(qName).contains(typ);
     * </jml></pre>
     *
     */
    public void signatureCompleted(CClass typ) {
	String qName = typ.qualifiedName();
	if (compilationSessionTypeCache.getImage(qName).contains(typ)) {
	    // move the type's signature
	    compilationSessionTypeCache.remove(qName,typ);
	    vmSessionTypeCache.put(qName,typ);

	    // Also move the type representation
	    Object typeRep = compilationSessionRepCache.get(qName);
	    if (typeRep != null) {
		compilationSessionRepCache.remove(qName);
		vmSessionRepCache.put(qName,typeRep);
	    }
	}
    }

    // -------------------------------------------------------------
    // TRUSTWORTHINESS PREDICATE
    // -------------------------------------------------------------

    /**
     * Returns true if the information for the type or package of the
     * given qualified name should be retained for subsequent
     * compilation sessions.
     *
     * <pre><jml>
     *  requires qName != null && (* package separators in qName are '/' not '.' *);
     * </jml></pre>
     */
    protected /*@ spec_public pure @*/ boolean isTrusted( String qName ) {
 	return qName.startsWith("java/");
    }

    // -------------------------------------------------------------
    // COMPILATION SESSION MANAGEMENT
    // -------------------------------------------------------------

    /**
     * Empties the compilation session caches to prepare for a new
     * compilation session.
     * 
     *
     * <pre><jml>
     * requires (* ClassPath.init() has been called with to set the 
     *             class path for this upcoming session *);
     * </jml></pre>
     *
     */
    public void initSession( ) {
	String newClassPath = ClassPath.classPath();
	boolean resetVMSessionCache = oldClassPath != null &&
	    !oldClassPath.equals( newClassPath );
	oldClassPath = newClassPath;

	// Type Signature Caches
	compilationSessionTypeCache.clear();
	if (resetVMSessionCache) {
	    vmSessionTypeCache.clear();
	}
	
	// Type Rep Caches
	compilationSessionRepCache.clear();
	if (resetVMSessionCache) {
	    vmSessionRepCache.clear();
	}
	
	CStdType.initSession();
    }

    /**
     * Returns the FileFinder instance that is used by the current
     * loader when searching for classes and generic functions.  */
    //@ ensures \result != null;
    public FileFinder fileFinder() { 
	return fileFinder;
    }

    // ----------------------------------------------------------------------
    // PRIVILEGED DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * This relation stores a mapping from fully qualified type names
     * to their CClass instances.  The mappings are for the types
     * processed during this compilation session that are likely to
     * change for the next compilation session.
     * @see #vmSessionTypeCache
     * @see #isTrusted(String)
     */
    /*@ invariant (\forall Object qn; compilationSessionTypeCache.containsKey(qn);
      @				     qn instanceof String && qn != null &&
      @				     (* any separators in qn are slashes, 
      @					not dots ('.') *) );
      @*/
    /*@ spec_public @*/ protected MjcHashRelation compilationSessionTypeCache;

    /**
     * This relation stores a mapping from fully qualified type names
     * to their CClass instances.  The mappings are for the types
     * processed during this compilation session that are not likely
     * to change for subsequent compilation sessions and that have
     * been processed to the point that they can be used for field and
     * method lookup.
     * @see #compilationSessionTypeCache
     * @see #isTrusted(String)
     */
    /*@ invariant (\forall Object qn; vmSessionTypeCache.containsKey(qn);
      @				     qn instanceof String && qn != null &&
      @				     (* any separators in qn are slashes, 
      @					not dots ('.') *) );
      @*/
    /*@ spec_public @*/ protected MjcHashRelation vmSessionTypeCache;

    /**
     * Maintains a mapping of qualified names to type representations; 
     * that is, it maps a fully qualified name to the CClassType 
     * instance that represents the type.  The mappings are for the types
     * processed during this compilation session that are likely to
     * change for the next compilation session.
     * @see #vmSessionRepCache
     * @see #isTrusted(String)
     */
    private /*@ spec_public @*/ HashMap compilationSessionRepCache = new HashMap(2000);

    /**
     * Maintains a mapping of qualified names to type representations;
     * that is, it maps a fully qualified name to the CClassType
     * instance that represents the type.  The mappings are for the
     * types processed during this compilation session that are not
     * likely to change for the next compilation session.
     * @see #compilationSessionRepCache
     * @see #isTrusted(String)
     */
    private /*@ spec_public @*/ HashMap vmSessionRepCache = new HashMap(2000);

    protected FileFinder fileFinder = null;

    private String oldClassPath;

    private static TypeLoader singletonInstance;
}
