/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjClassDoc.java,v 1.5 2005/05/03 04:24:55 kui_dai Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_141;

import com.sun.javadoc.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import org.multijava.mjc.*;

import org.multijava.util.classfile.FieldInfo;

/** This class is an interface between the MJC syntax tree classes and the javadoc program;
    it wraps a CClass object.
    @author David R. Cok
 */
public class MjClassDoc extends MjProgramElementDoc implements com.sun.javadoc.ClassDoc, com.sun.javadoc.Type {

    /** Holds the identifier of the class. */
    private String className;
    
    /** Holds the MJ type that this wraps (when source code is available). */
    private JTypeDeclarationType jtype;
    
    /** The CClass object that this class wraps. */
    private CClass cclass;
    
    /** True if this object represents an interface. */
    private boolean isInterface;
    
    /** True if this object represents a class derived from java.lang.Error. */
    private boolean isError;
    
    /** True if this object represents a class derived from java.lang.Exception. */
    private boolean isException;
    
    /** True if this object represents a class (not an interface, error or exception). */
    private boolean isOrdinaryClass;
    
    /** Cached array of the methods (and external methods) of this class. */
    private ArrayList allmethods = null;
    
    /** ArrayList of the internal and external methods of this class. */
    //private ArrayList methods;

    /** ArrayList of the external methods of this class. */
    private ArrayList tlmethods = new ArrayList();

    /** Cached value indicating whether there is an @deprecated tag. */
    private boolean isDeprecated = false;
    
    /** A constructor for the case when source code is available (which makes javadoc comments
	and formal parameter identifiers available).
    */
    public MjClassDoc(JTypeDeclarationType jtype) {
	super(jtype.getCClass()); 
	init(jtype.getCClass());
	setJtype(jtype);
    }

    /** Constructs an object from a CClass object (to be used when source code is not available). */
    public MjClassDoc(CClass cc) {
	super(cc); 
	init(cc);
    }
    
    /** A method to set the source code parse tree object for this class after it has been constructed. 
	Note that the type of the argument must match the type of the already built object.
    */
    public void setJtype(JTypeDeclarationType jt) {
	if (jtype == jt) return;
	if (!jt.getCClass().equals(cclass)) { 
	//if (jt.getCClass() != cclass) {
	    throw new org.multijava.util.compiler.CompilationAbortedError("INTERNAL MJDOC ERROR: Calling setJtype on an object that does not match the argument (" + jt.getCClass() + " vs. " + cclass + ")"); 
	} 
	this.jtype = jt;
	if (jtype.javadocComment() != null) {
	    setRawCommentText(jtype.javadocComment().text(),jtype.getTokenReference());
	} else {
	    setRawCommentText();
	}
	makeAllMethods();
	Tag[] t = tags("@deprecated");
	isDeprecated = (t.length > 0);


	superclass(); // Forces the registration of the super class
	interfaces(); // Forces registration of the implemented interfaces
	setIncluded(containingPackage().isIncluded());
    }

    /** A helper method that completes the construction of an object. */
    private void init(CClass cc) {
	this.jtype = null;
	className = cc.ident(); 
	cclass = cc;
	
	makeAllMethods();

	isInterface = cc.isInterface();	
	isException = this.subclassOf("java.lang.Exception");
	isError = this.subclassOf("java.lang.Error");
	isOrdinaryClass = !(isError||isException||isInterface);

	MjPackageDoc pkg = (MjPackageDoc)containingPackage();
	if (isInterface) pkg.addInterface(this);
	else if (isError)   pkg.addErrorClass(this);
	else if (isException) pkg.addExceptionClass(this);
	else             pkg.addOrdinaryClass(this);
	
	superclass(); // Forces the registration of the super class
	interfaces(); // Forces registration of the implemented interfaces
	setIncluded(containingPackage().isIncluded());
	
    }

    /** Returns true if this is a class (rather than an interface, field, method or constructor). */
    //@ pure
    public boolean isClass() { return !isInterface; } 

    /** Returns true if the object represents a class type that is not an interface, 
	an exception (derived from java.lang.Exception) or an error (derived from java.lang.Error). 
     */
    //@ pure
    public boolean isOrdinaryClass() { return isOrdinaryClass; } 
    
    /** Returns true if the object represents an interface type (rather than a class type). */    
    //@ pure
    public boolean isInterface() { return isInterface; }

    /** Returns true if the object represents an interface type (rather than a class type). 
    */  
    //@ pure
    public boolean isError() { return isError; }

    /** Returns true if the object represents an interface type (rather than a class type). */    
    //@ pure
    public boolean isException() { return isException; }

    /** Returns the CClass object that corresponds to this MjClassDoc. */
    public CClass cclass() { return cclass; }

    /** Returns the name of the class, with any enclosing classes prefixed, and with '.' as separator; no package name is prefixed.
     */
    //@ pure
    public String name() { 
	return containingClass() == null ? className : (containingClass().name() + "." + className) ; 
    } 

    /** Returns the wrapped JTypeDeclarationType object. */
    public JTypeDeclarationType jtype() { return jtype; }
    
    /** Returns the name of the class, with any enclosing classes prefixed, and with '.' as separator; no package name is prefixed.
     */
    public String typeName() { return name(); } 
	
    /** Returns the name of the class, with the package name and any enclosing classes prefixed, and with '.' as separator.
     */
    public String qualifiedTypeName() { 
	return ((MjPackageDoc)containingPackage()).prefix() + name();
    }

    /** Returns the name of the class, with the package name and any enclosing classes prefixed, and with '.' as separator.
     */
    //@ pure
    public String qualifiedName() { 
	return ((MjPackageDoc)containingPackage()).prefix() + name();
    }

    /** Returns the same thing as qualifiedTypeName(). */
    public String toString() { return qualifiedTypeName(); }

    /** Returns the object, since it already is a ClassDoc. */
    public ClassDoc asClassDoc() { return this; }

    /** Returns an empty string since a ClassDoc is not an array. */
    public String dimension() { return ""; } 
	

    /** NOT IMPLEMENTED */ // !FIXME! - definesSerializableFields
    public boolean definesSerializableFields() { return false; }

    public static MemberFilter memberFilter = new MjMemberFilter();

    public static interface MemberFilter {
	public boolean use(CMember f);
	public boolean use(JMemberDeclarationType f);
    }
    
    public static class MjMemberFilter implements MemberFilter {
	// Do not include compiler generated fields (those with $ in them) 
	//!FIXME! - Better definition?  // ??? Use isSynthetic()

	public boolean use(CMember f) { 
	    return (f.ident().indexOf('$') == -1) && 
		MjdocWrapper.checkPrivacy(f.modifiers()); 
	}

	// Do not include compiler generated fields (those with $ in them) 
	//!FIXME! here too
	public boolean use(JMemberDeclarationType f)  { 
	    return MjdocWrapper.checkPrivacy(f.modifiers()) && 
			(f.ident().indexOf('$') == -1);
	}
    }
    
    /** Returns the fields that are defined for the class represented by this. */
    public FieldDoc[] fields(boolean filtered) {
        return fields();
    }

    public FieldDoc[] fields() {
	if (jtype == null) {
	    Collection fc = cclass.fields();
	    Iterator f = fc.iterator();
	    ArrayList v = new ArrayList(fc.size());
	    while (f.hasNext()) {
		CField ff = (CField)(f.next());
		if (memberFilter.use(ff)) {
		    v.add(new MjFieldDoc(this,ff));
		}
	    }
	    MjFieldDoc[] result = new MjFieldDoc[v.size()];
	    v.toArray(result);
	    
	    return result; 
	} else {
	    JFieldDeclarationType[] fc = jtype.fields();
	    ArrayList v = new ArrayList(fc.length);
	    for (int j=0; j<fc.length; ++j) {
		if (memberFilter.use(fc[j])) v.add(new MjFieldDoc(this,fc[j]));
	    }
	    MjFieldDoc[] result = new MjFieldDoc[v.size()];
	    v.toArray(result);
	    
	    return result; 
	}
    }

    /** NOT IMPLEMENTED - findClass */ // FIXME
    public ClassDoc findClass(String className) { 
	notImplemented("MjClassDoc.findClass");
	return null; 
    }

    /** A cached value (after being initialized) of the array of classes imported into the compilation unit
	this class is defined in.
    */
    private ClassDoc[] cachedImportedClasses = null;
    
    /** Returns the array of classes imported into the compilation unit this class is defined in. */
    public ClassDoc[] importedClasses() {
	if (cachedImportedClasses == null) {
	    JClassOrGFImportType[] jf = cclass.getCCompilationUnit().importedClasses();
	    ArrayList v = new ArrayList(jf.length);
	    for (int i=0; i<jf.length; ++i) {
		String n = jf[i].getName();
		CClass c = CTopLevel.loadClass(n);
		if (c == CClass.CLS_UNDEFINED) {
		    System.err.println("Could not find imported class " + n + " in " + qualifiedName());
		} else {
		    MjClassDoc mj = MjdocWrapper.lookup(c);
		    if (mj == null) {
			System.err.println("Could not find imported class " + n + " in " + qualifiedName());
		    } else {
			v.add(mj);
		    }
		}
	    }
	    cachedImportedClasses = new MjClassDoc[v.size()];
	    v.toArray(cachedImportedClasses);
	    
	}
	return cachedImportedClasses; 
    }
    
    /** A cached value (after being initialized) of the array of packages imported into the compilation unit
	this class is defined in.
    */
    private PackageDoc[] cachedImportedPackages = null;

    /** Returns the array of packages imported into the compilation unit this class is defined in. */
    public PackageDoc[] importedPackages() {
	if (cachedImportedPackages == null) {
	    JPackageImportType[] jf = cclass.getCCompilationUnit().importedPackages();
	    ArrayList v = new ArrayList(jf.length);
	    for (int i=0; i<jf.length; ++i) {
		String n = jf[i].getName(); // '/'-separated name of the package
		MjPackageDoc mj = MjPackageDoc.createPackage(n);
		v.add(mj);
	    }
	    cachedImportedPackages = new MjPackageDoc[v.size()];
	    v.toArray(cachedImportedPackages);
	    
	}
	return cachedImportedPackages; 
    }

    /** Returns an array of all the nested classes and nested interfaces of this class. */
    public ClassDoc[] innerClasses() { 
	if (isInterface) return new MjClassDoc[0];
	CClass cl = cclass();
	CClassType[] ct = cl.getInnerClasses(); // FIXME - Inner or Nested?  are they really inner?
	ArrayList r = new ArrayList(ct.length);
	for (int i=0; i<ct.length; ++i) {
	    CClass c = ct[i].getCClass();
	    if (memberFilter.use(c)) r.add(MjdocWrapper.lookup(c));
	}
	
	ClassDoc[] result = new ClassDoc[r.size()];
	r.toArray(result);
	
	return result;
    }
    
    /** Returns an array of all the interfaces implemented by this class. */
    public ClassDoc[] interfaces() {
	CClassType[] ct = cclass.getInterfaces();
	ArrayList r = new ArrayList(ct.length);
	for (int i=0; i<ct.length; ++i) {
	    CClass c = ct[i].getCClass();
	    r.add(MjdocWrapper.lookup(c));
	}
	ClassDoc[] result = new ClassDoc[r.size()];
	r.toArray(result);
	
	return result;
    }

    /** Returns the array of (internal and external) methods of this class (as a MjMethodDoc array), 
        according to the filter set in memberFilter. */
    public MethodDoc[] methods() {
	return (MjMethodDoc[])(methodsList().toArray(new MjMethodDoc[0])); 
    }
    
    /** Returns the array of (internal and external) methods of this class (as a MjMethodDoc array), 
        according to the filter set in memberFilter. */
    public ArrayList methodsList() {
	if (allmethods == null) makeAllMethods();
	// Now filter them according to the current memberFilter
	ArrayList a = new ArrayList(allmethods.size());
	Iterator i = allmethods.iterator();
	while (i.hasNext()) {
	    MjMethodDoc m = (MjMethodDoc)(i.next());
	    if (memberFilter.use(m.cmethod())) a.add(m);
	}
	return a; 
    }
    
    /** Returns the array of (internal and external) methods of this class (as a MjMethodDoc array), 
        according to the filter set in memberFilter. */
    public ArrayList allmethods() {
	return allmethods; 
    }
    
    /** This method converts the methods of the CClass into MjMethodDoc objects.
	This is done once and cached, because methods() is called often (whenever
	the class is a superclass or interface of a class.
    */
    private void makeAllMethods() {
	allmethods = new ArrayList();
	if (jtype != null) {
	    java.util.Iterator i = jtype.methods().iterator();
	    allmethods.addAll(tlmethods);
	    while (i.hasNext()) {
		Object o = i.next();
		if (o instanceof JMethodDeclarationType && 
		       !(o instanceof JConstructorDeclarationType)) {
		    JMethodDeclarationType j = (JMethodDeclarationType)o;
		    if (j.ident().charAt(0) != '<') {
			allmethods.add(new MjMethodDoc(j,this));
		    }
		}
	    }
	} else {
	    // method is called on a MjClassDoc that was not generated from a source file, so
	    // we get the methods from the CMethod.  These are
	    // used to report inherited methods in the html file.
	    allmethods.addAll(tlmethods);
	    CMethodSet s = cclass.methods();
	    if (s == null) return;
	    CAbstractMethodSet.Iterator i = s.iterator();
	    while (i.hasNext()) {
		CMethod cm = i.next();
		if (!cm.isConstructor()
			&& cm.ident().charAt(0) != '<') {
		    allmethods.add(new MjMethodDoc(cm,this));
		}
	    }
	}
	return;
    }
    
    /** Returns the array of constructors of this class
     *  (as a MjConstructorDoc array); the result of this
     *  method depends on the current memberFilter. 
     */
    public ConstructorDoc[] constructors() { 
	return (ConstructorDoc[])(constructors(addDefaultConstructors).toArray(new MjConstructorDoc[0]));
    }
    
    public static boolean addDefaultConstructors = true;
    
    public ArrayList constructors(boolean addDefault) { 
	if (jtype != null) {
	    java.util.Iterator i = jtype.methods().iterator();
	    ArrayList v = new ArrayList();
	    boolean hasConstructors = false;
	    while (i.hasNext()) {
		Object o = i.next();
		if (o instanceof JConstructorDeclarationType) {
		    hasConstructors = true;
		    JConstructorDeclarationType jc = (JConstructorDeclarationType)o;
		    if (memberFilter.use(jc)) v.add(new MjConstructorDoc(jc,this));
		}
	    }
	    // Do default constructor
	    if (!hasConstructors && !isInterface && addDefault) v.add(new MjConstructorDoc((CMethod)null,this));
	    return v;
	} else {
	    // method is called on a MjClassDoc that was not generated 
	    // from a source file, so
	    // we get the methods from the CMethod.  These are
	    // used to report inherited methods in the html file.
	    CMethodSet s = cclass.methods();
	    ArrayList v = new ArrayList();
	    if (s == null) return v;
	    CAbstractMethodSet.Iterator i = s.iterator();
	    while (i.hasNext()) {
		CMethod cm = i.next();
		if (cm.isConstructor() && memberFilter.use(cm)) v.add(new MjConstructorDoc(cm,this));
	    }
	    // Why no default here ?? FIXME
	    return v; 
	}
    }

    /** Adds an external method to the set of such methods for this class, so 
	appropriate information will be added to the generated html file.
	@return the MjMethodDoc object representing the external method
    */
    public MjMethodDoc addTLMethod(MJTopLevelMethodDeclaration tlm) {
	MjMethodDoc j = new MjMethodDoc(tlm,this);
	tlmethods.add(j);
	//methods.add(j); 
	allmethods = null;
	return j; 
    }

    /** Returns true if this class is abstract. */
    public boolean isAbstract() { return org.multijava.util.Utils.hasFlag(cclass.modifiers(),org.multijava.util.classfile.Constants.ACC_ABSTRACT); }

    /** UNKNOWN */
    public boolean isExternalizable() { return false; }// !FIXME!
    
    /** UNKNOWN */
    public boolean isSerializable() { return false; }// !FIXME!
    
    /** UNKNOWN */
    public FieldDoc[] serializableFields() { return new MjFieldDoc[0]; }// !FIXME!
    
    /** UNKNOWN */
    public MethodDoc[] serializationMethods() { return new MjMethodDoc[0]; }// !FIXME!

    /** Returns true if 'this' represents a class that is a subclass of the argument. */
    public boolean subclassOf(ClassDoc c) { 
	ClassDoc cc = this;
	// Walk up the chain of super class objects until we get to the top (null super class)
	// or we find a class that equals the argument. 
	while (true) {
	    CClass cl = ((MjClassDoc)cc).cclass().getSuperClass();
	    if (cl == null) return false;
	    cc = MjdocWrapper.lookup(cl);
	    if (c == cc) return true;
	}
    }

    /** Returns true if 'this' represents a class that is a subclass of the argument. 
	@param qualifiedName A fully qualified, '.'-separated class name.
    */
    public boolean subclassOf(String qualifiedName) { 
	ClassDoc cc = this;
	// Walk up the chain of super class objects until we get to the top (null super class)
	// or we find a class that equals the argument. 
	while (true) {
	    if (cc.qualifiedName().equals(qualifiedName)) return true;
	    CClass cl = ((MjClassDoc)cc).cclass().getSuperClass();
	    if (cl == null) { return false; }
	    cc = MjdocWrapper.lookup(cl);
	    if (cc == null) { return false; }
	}
    }

    /** Returns (as a MjClassDoc object) the superclass of the class represented by the calling object. */
    public ClassDoc superclass() {
	if (isInterface) return null;
	CClass cl = cclass().getSuperClass();
	if (cl == null) return null;
	return MjdocWrapper.lookup(cl);
    }

    /** Returns the modifiers of this class declaration encoded as bits in an int, 
	using the MJ encoding.
    */
    // The magic number is what javadoc uses to denote a deprecated member
    // We turn off the super bit since it masquerades as synchronized
    public long longModifiers() {
	return (cclass.modifiers() & ~Constants.ACC_SUPER) | (isDeprecated ? 262144 : 0); 
    }

    /** Looks for a method with matching name and signature to the argument in this
	class and in any superclasses, either until found (in which case the parent
	class method is returned) or not (null is returned).
    */
    public MjMethodDoc findMethodRecursive(MjMethodDoc method) {
	MjMethodDoc c = findMethod(method);
	if (c != null) return c;
	CClass cc = cclass().getSuperClass();
	if (cc == null) return null;
	return MjdocWrapper.lookup(cc).findMethodRecursive(method);
    }
    
    /** Looks for a method with matching name and signature to the argument in this
	class; if found the parent class method is returned; if not null is returned.
    */
    public MjMethodDoc findMethod(MjMethodDoc method) {
	if (allmethods == null) makeAllMethods();
	Iterator i = allmethods.iterator();
	while (i.hasNext()) {
	    MjMethodDoc mj = (MjMethodDoc)(i.next());
	    if (method.sameMethod(mj)) return mj;
	}
	return null;
    }
    
    /** Looks for a method (with an MjMethodDoc) with matching name and signature 
	to the argument that is registered in the MjClassDoc target object.  If
	found the appropriate MjMethodDoc object is returned; if not found, null is
	returned.
    */
    public MjMethodDoc findMethod(CMethod m) {
	if (allmethods == null) makeAllMethods();
	Iterator i = allmethods.iterator();
	while (i.hasNext()) {
	    MjMethodDoc mj = (MjMethodDoc)(i.next());
	    if (m == mj.cmethod()) return mj;
	}
	return null;
    }
    // toString() returns fully quallified name

    /** This generates the String representation used by compareTo.  We use a different method than toString,
	because toString is not unique in all cases (e.g. methods with the same identifier but different arguments).
	This is used for the index.
    */
    protected String compareString() {
	return name(); 
    }

    /** Returns 0 if the classes are not comparable, 1 if the argument supercedes the object,
       -1 if the object supercedes the argument.
    */
    public int compare(MjClassDoc mjd) {
       return 0;
    }


    /** Return the source position of the documented entity, or null
     *  if no position is available.
     *  Added for JDK 1.4 compatability.
     */
//      public SourcePosition position() {
//          return null; /* !FIXME!JDK 1.4 compatability */
//      }
}
