/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjDoc.java,v 1.3 2007/02/08 14:04:56 leavens Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_141;

import com.sun.javadoc.*;
import org.multijava.util.compiler.TokenReference;

/** This abstract class provides an implementation of the Doclet Doc class. 
    Each Doc object represents a syntatical item, such as a field, method, 
    or class and may have a javadoc comment associated with it.
    @author David R. Cok
*/
abstract public class MjDoc implements Doc {

    /** Holds an object that contains a representation of the javadoc
	comment associated with this Doc item (once initialized with
	the javadoc comment text).
    */
    protected MjTagParser parser = new MjTagParser(this);

    /** Returns the descriptive section of the javadoc comment as unparsed text. */
    //@ pure
    public String commentText() { return parser.commentText(); } 

    /** This compares two Doc objects.  There is no specification of this comparison for javadoc, so this is
	not guaranteed to be a good comparator or the same comparator as javadoc's.  This implementation
	compares String representations of the two objects.  It is used to provide ordered lists of members
	and classes in various places in the javadoc documentation.
    */ 
    public int compareTo(/*@ non_null @*/ Object o) { 
	//System.out.println("CT " + compareString() + " " + ((MjDoc)o).compareString() + " " + compareString().//compareTo(((MjDoc)o).compareString()));
	return compareString().compareToIgnoreCase(((MjDoc)o).compareString());
    }

    /** This generates the String representation used by compareTo.  We use a different method than toString,
	because toString is not unique in all cases (e.g. methods with the same identifier but different arguments).
    */
    protected String compareString() {
	return toString();
    }
    
    /** Returns the summary of the description of the javadoc comment 
	(as set by setRawCommentText) as a sequence of tags. 
    */
    //@ pure
    public Tag[] firstSentenceTags() { return parser.firstSentenceTags(); }

    /** Returns the full unparsed javadoc comment (as set by setRawCommentText). */
    //@ pure
    public String getRawCommentText() { return parser.getRawCommentText(); } 
    
    /** Returns the descriptive section of the javadoc comment (as set by setRawCommentText) as 
	parsed into a sequence of Text and inline tags. 
    */    
    //@ pure
    public Tag[] inlineTags() { return parser.inlineTags(); }

    /** Returns true if this Doc element represents a class
	(not an interface, field, method, constructor).
    */
    //@ pure
    public boolean isClass() { return false; }
    
    /** Returns true if this Doc element represents a constructor. */
    //@ pure
    public boolean isConstructor() { return false; }
    
    /** Returns true if this Doc element represents a class derived from java.lang.Error. 
    */
    //@ pure
    public boolean isError() { return false; }
    
    /** Returns true if this Doc element represents a class derived from java.lang.Exception. */
    //@ pure
    public boolean isException() { return false; } 
    
    /** Returns true if this Doc element represents a field. */
    //@ pure
    public boolean isField() { return false; }

    /** Holds the value indicating whether this file is included in the documentation being generated. */
    private boolean included = true;

    /** Returns true if this Doc is in the active set. */ 
    public boolean isIncluded() { return included; }

    /** Sets the included field. */
    public void setIncluded(boolean b) { 
	included = b;
    }
    
    /** Returns true if this Doc element represents an interface (as opposed to a class, method or field). */
    //@ pure
    public boolean isInterface() { return false; }
    
    /** Returns true if this Doc element represents a method (but not a constructor). */
    //@ pure
    public boolean isMethod() { return false; }
    
    /** Returns true if this Doc element represents a class that is not an interface, error or exception class. */
    //@ pure
    public boolean isOrdinaryClass() { return false; } 

    /** Returns the name of the item, as interpreted by derived classes. */
    //@ pure
    abstract public String name();

    /** Returns the package this Doc is in. */
    abstract public PackageDoc containingPackage();
    
    /** Returns the sequence of tags representing all the '@see' tags in the tag section of the javadoc comment as set in setRawCommentText. */
    //@ also
    //@ ensures \result != null;
    //@ pure
    public SeeTag[] seeTags() { return parser.seeTags(); } 

    /** Sets the javadoc comment text.
	@deprecated
    */
    //@ also requires rawDocumentation != null;
    public void setRawCommentText(String rawDocumentation) { 
	parser = new MjTagParser(this,rawDocumentation,null); 
    }  
    public void setRawCommentText() { 
	parser = new MjTagParser(this,"",(SourcePosition)null); 
    }  
    public void setRawCommentText(String rawDocumentation, TokenReference tok) { 
	parser = new MjTagParser(this,rawDocumentation, new MjSourcePosition(tok)); 
    }  
    public void setRawCommentText(String rawDocumentation, SourcePosition sp) { 
	parser = new MjTagParser(this,rawDocumentation, sp); 
    }  

    /** Returns the sequence of tags representing the tag section of the javadoc comment 
	as set in setRawCommentText. 
    */
    //@ also ensures \result != null;
    //@ pure
    public Tag[] tags() { return parser.tags(); } 
 
    /** Returns the sequence of tags representing all the tags with the given 'kind' in the tag section of the javadoc comment as set in setRawCommentText. */
    //@ also ensures \result != null;
    //@ pure
    public Tag[] tags(String tagkind) { return parser.tags(tagkind); } 
    
    /** Returns the sequence of tags representing all the tags with the given 'kind' in the tag section of the javadoc comment as set in setRawCommentText. The type field should be something like 'new MjZZZTag[0]',
	where MjZZZTag is the class of that represents tags with the given kind.  It is used to do the correct
	casting for the dynamic type of the returned array.
    */
    //@ ensures \result != null;
    //@ pure
    public Tag[] tags(String tagkind, Tag[] type) { return parser.tags(tagkind,type); } 

    /** This is called by derived classes in methods that are not implemented or are not expected to be
	called.  Printing the stack trace helps the debugging.  Report instances of this, with enough detail 
	to reproduce the problem.
    */
    protected void notImplemented(String s) {
	System.out.println("UNEXPECTEDLY CALLED AN UNSPECIFIED OR UNIMPLEMENTED METHOD: " + s);
	(new Exception()).printStackTrace(System.out);
	throw new org.multijava.util.compiler.CompilationAbortedError("");
    }
    
	//FIXME - can we do something usefule with this???
    public SourcePosition position() { return null; }

}

