/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjExecutableMemberDoc.java,v 1.2 2005/05/03 04:24:55 kui_dai Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_141;
import com.sun.javadoc.*;
import org.multijava.mjc.CMethod;
import org.multijava.mjc.CType;
import org.multijava.mjc.CClassType;
import org.multijava.mjc.CSpecializedType;
import org.multijava.mjc.JFormalParameter;
import org.multijava.mjc.JMethodDeclarationType;
import org.multijava.mjc.MJTopLevelMethodDeclaration;
import org.multijava.util.classfile.Constants;

/** This class is a superclass representing common functionality between both methods
    and constructors; it wraps appropriate objects from the MJ classes and parse tree.
    @author David R. Cok
*/
abstract public class MjExecutableMemberDoc extends MjMemberDoc implements ExecutableMemberDoc {

//!FIXME Needs documentation

    /** Holder for the MJC construct denoting the method being wrapped. */
    private CMethod cmethod;
    
    /** Holder for the object from the MJ parse tree, when source code is available. */
    private JMethodDeclarationType jmethod;

    /** Cached value indicating whether there is an @deprecated tag. */
    private boolean isDeprecated = false;
    
    /** Constructs a wrapper object; to be used when no source is available.
    @param cm The MJ object denoting the method to be wrapped by the constructed object
    @param container An object denoting the class that contains this member
    */
    public MjExecutableMemberDoc(CMethod cm, MjClassDoc container) {
	super(container);
	cmethod = cm;
	jmethod = null;
	setRawCommentText();
	if (cmethod == null) {
	    cachedFlatSignature = "()";
	    cachedSignature = "()";
	}
    }
    
    /** Constructs a wrapper object; to be used when source code is available; this makes
	available the javadoc comments and parameter identifiers.
	@param jmethod The MJ object denoting the method to be wrapped by the constructed object
	@param container An object denoting the class that contains this member
    */
    public MjExecutableMemberDoc(JMethodDeclarationType jmethod, MjClassDoc container) {
	super(container);
	this.jmethod = jmethod;
	cmethod = jmethod.getMethod();
	if (jmethod.javadocComment() != null) {
	    setRawCommentText(jmethod.javadocComment().text(),jmethod.getTokenReference());
	} else {
	    setRawCommentText();
	}
	Tag[] t = tags("@deprecated");
	isDeprecated = (t.length > 0);
    }

    /** Caches the signature (with fully-qualified type names in the parameters) of this member. */
    private String cachedSignature = null;
    
    /** Caches the signature (with non-qualified type names in the parameters) of this member. */
    private String cachedFlatSignature = null;

    /** Cached value of the parameter array for this member. */
    private Parameter[] cachedParameters = null;
    
    /** Returns the signature of the method, from ( to ), with a space after each comma,  
	including dimensions and containing types, but no package names.  
	This signature is used in the html index.
     */
    public String flatSignature() { 
	if (cachedFlatSignature == null) {
	    StringBuffer sb = new StringBuffer();
	    sb.append("(");
	    CSpecializedType[] p = cmethod.parameters();
	    for (int i=0; i<p.length; ++i) {
		if (i!=0) sb.append(", ");
		CSpecializedType ct = p[i];
		CType cs = ct.staticType();
		CType cd = ct.dynamicType();
		if (!ct.isSpecialized()) {
		    sb.append(MjType.name_noq_dim(cs));
		} else {
		    sb.append(cs.toString());
		    sb.append(cd.specializerSymbol());
		    sb.append(MjType.name_noq_dim(cd));
		}
	    }
	    sb.append(")");
	    cachedFlatSignature = sb.toString();
	}
	return cachedFlatSignature; 
    } 

    /** Returns the signature of the method, from ( to ), with a space after each comma,  
	including parameter names, dimensions and containing types, but no package names.  
     */
    public String parameterSignature() { 
	    StringBuffer sb = new StringBuffer();
	    sb.append("(");
	    CSpecializedType[] p = cmethod.parameters();
	    for (int i=0; i<p.length; ++i) {
		if (i!=0) sb.append(", ");
		CSpecializedType ct = p[i];
		CType cs = ct.staticType();
		CType cd = ct.dynamicType();
		if (!ct.isSpecialized()) {
		    sb.append(MjType.name_noq_dim(cs));
		} else {
		    sb.append(cs.toString());
		    sb.append(cd.specializerSymbol());
		    sb.append(MjType.name_noq_dim(cd));
		}
		MjParameter mp = (MjParameter)(parameters()[i]);
		sb.append(" ");
		sb.append(mp.name());
	    }
	    sb.append(")");
	    return sb.toString();
    } 
    
    /** Returns the signature of the method, from ( to ), with a space after each comma, with 
	dimensions and fully qualified type names (i.e. including package names).
	This signature is used in the HREF tag that enables navigation to a method directly.
     */
    public String signature() {
	if (cachedSignature == null) {
	    StringBuffer sb = new StringBuffer();
	    sb.append("(");
	    CSpecializedType[] p = cmethod.parameters();
	    for (int i=0; i<p.length; ++i) {
		if (i!=0) sb.append(", ");
		CSpecializedType ct = p[i];
		CType cs = ct.staticType();
		CType cd = ct.dynamicType();
		if (cs == cd || cs.equals(cd)) {
		    sb.append(cs.toString());
		} else {
		    sb.append(cs.toString());
		    sb.append(cd.specializerSymbol());
		    sb.append(cd.toString());
		}
	    }
	    sb.append(")");
	    cachedSignature = sb.toString();
	}
	//System.out.println("SIG " + cachedSignature);
	return cachedSignature; 
    } 

    /** Returns true if the represented method is a native method. */
    public boolean isNative() { return cmethod == null? false : cmethod.isNative(); }

    /** Returns true if this member has the 'synchronized' modifier in the java code. */
    public boolean isSynchronized() { 
	return org.multijava.util.Utils.hasFlag(longModifiers(),
			org.multijava.util.classfile.Constants.ACC_SYNCHRONIZED); 
    }

    /** Returns an array of the parameters of this member, as MjParameter objects that
	wrap the information from the MJC parse tree; identifiers for the parameters
	are available only if this object was constructed with a JMethod object providing
	a reference to the source code.
    */
    public Parameter[] parameters() {
	if (cachedParameters == null) {
	    if (jmethod != null) {
		JFormalParameter[] jfp = jmethod.parameters();
		MjParameter[] p = new MjParameter[jfp.length];
		for (int i=0; i<jfp.length; ++i) {
		    p[i] = new MjParameter(jfp[i]);
		}
		cachedParameters = p; 
	    } else if (cmethod != null) {
		CSpecializedType[] cst = cmethod.parameters();
		MjParameter[] p = new MjParameter[cst.length];
		for (int i=0; i<cst.length; ++i) {
		    p[i] = new MjParameter(cst[i],"");
		}	
		cachedParameters = p; 
	    } else {
		// This is for a default constructor
		//!FIXME! What about additional parameters for inner classes???
		cachedParameters = new MjParameter[0];
	    }
	}
	return cachedParameters;
    }
    
    /** Returns the fully qualified name (package + enclosing classes + class + method name) of the method. */
    //@ pure
    public String qualifiedName() { return ((MjPackageDoc)containingPackage()).prefix() + name(); } // !FIXME! CHeck that name() includes the enclosing class

    /** Returns the classes that are the exceptions declared to be thrown by this method. */
    public ClassDoc[] thrownExceptions() { 
	if (cmethod == null) return new ClassDoc[0];
	CClassType[] ct = cmethod.throwables();
	ClassDoc[] cd = new MjClassDoc[ct.length];
	for (int i=0; i<ct.length; ++i) {
	    cd[i] = MjdocWrapper.lookup(ct[i].getCClass());
	}
	return cd; 
    }
    
    /** Returns the '@param' javadoc tags of this member. */
    public ParamTag[] paramTags() { 
	return (ParamTag[])(tags(MjParamTag.KIND,new MjParamTag[0]));
    }

    /** Returns the '@throws' javadoc tags of this member. */
    public ThrowsTag[] throwsTags() { 
	return (ThrowsTag[])(tags(MjThrowsTag.KIND, new ThrowsTag[0])); 
    }

    /** Returns the identifier of the method, without package or enclosing classes. */
    //@ pure
    abstract public String name(); 

    /** A constant used only by {@link #longModifiers()} that selects those modifier bits that are passed on
	to the default constructor.
    */
    static final private long ACC = Constants.ACC_PUBLIC|Constants.ACC_PROTECTED|Constants.ACC_PRIVATE;

    /** Returns the (mjc) modifiers of the member.  Note that the modifiers of a default constructor
	are the same as the modifiers of the containing class.
    */
    public long longModifiers() {
	// The magic number is the number javadoc uses to flag a deprecated method
	int i = isDeprecated ? 262144 : 0;
	return ((cmethod == null)? 
	    (((MjClassDoc)containingClass()).longModifiers() & ACC)
	    : cmethod.modifiers()) | i; 
    }
    
    /** Accessor for the wrapped CMethod. */
    public CMethod cmethod() { return cmethod; }
	
    /** Accessor for the wrapped JMethodDeclarationType. */
    public JMethodDeclarationType jmethod() { return jmethod; }

    /** Returns a String representation, used for ordering. */
    protected String compareString() {
	return qualifiedName() + signature();
    }
    
    /** Returns a String representation, matches javadoc. */// FIXME - is this really right - no signature?
    public String toString() {
	return cmethod == null ? ((MjClassDoc)containingClass()).name() : cmethod.ident();
    }

    /** This class is used for sorting the gfs as listed on the 
	package summary page and in the overview index lists. 
    */ // FIXME to give a more natural order - put in MjMethodDoc ???
    static public class Comp implements java.util.Comparator {
	public int compare(Object o1, Object o2) {
	    if (!(o1 instanceof MjMethodDoc)) return ((Comparable)o1).compareTo(o2);
	    if (!(o2 instanceof MjMethodDoc)) return ((Comparable)o1).compareTo(o2);
	    MjMethodDoc m1 = (MjMethodDoc)o1;
	    MjMethodDoc m2 = (MjMethodDoc)o2;
	    int i = m1.name().compareTo(m2.name());
	    if (i != 0) return i;
	    i = m1.signature().compareTo(m2.signature());
	    if (i != 0) return i;
	    i = m1.cuPackage().name().compareTo(m2.cuPackage().name());
	    //System.out.println("COMPARE " + m1.qualifiedName() + " " + m2.qualifiedName() + " " + i);
	    return i;
	
	}
    }
	
    
}
