/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjFieldDoc.java,v 1.2 2005/05/03 04:24:55 kui_dai Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_141;
import com.sun.javadoc.*;
import org.multijava.mjc.JFieldDeclarationType;
import org.multijava.mjc.CField;

/** This class implements the Doclet FieldDoc interface by wrapping the MJC
    objects corresponding to a field of a class or interface.  For classes for
    which source code is available, this is the JFieldDeclarationType object
    for the field; if source code is not available, it is only the CField object.
    @author David R. Cok
*/
public class MjFieldDoc extends MjMemberDoc implements com.sun.javadoc.FieldDoc {

    /** Holder for the MJC construct that this class wraps, when source code is available.  */ 
    private JFieldDeclarationType jfield = null;
    
    /** Holder for the CField object (the MJC construct) that this class wraps. */
    private CField field;
    
    /** The type of the field, derived from the CField object. */
    private Type type = null;

    /** Constructor of an object wrapping a specific field, used when source code is not available.
	@param container the object representing the class that contains this field
	@param field the MJC construct representing the field
    */
    public MjFieldDoc(MjClassDoc container, CField field) {
	super(container);
	this.field = field;
	this.type = new MjType(field.getType());
	setRawCommentText();
    }

    /** Constructor of an object wrapping a specific field, used when source code is available,
	since information such as the javadoc comments are available from a JMemberDeclarationType
	object, but not from a CField object.
	@param container the object representing the class that contains this field
	@param field the MJC construct representing the field
    */
    public MjFieldDoc(MjClassDoc container, JFieldDeclarationType jfield) {
	super(container);
	this.jfield = jfield;
	this.field = jfield.getField();
	this.type = new MjType(field.getType());
	if (jfield.javadocComment() != null) setRawCommentText(jfield.javadocComment().text(),jfield.getTokenReference()); // FIXME -- improve position ?
	else setRawCommentText();
    }

    /** The (non-qualified), simple identifier for the field. */
    //@ pure
    public String name() { return field.ident(); }

    /** The (non-qualified), simple identifier for the field. */
    public String toString() { return field.ident(); }

    /** Returns a String representation, used for ordering. */
    protected String compareString() {
	return name();
    }
    
    /** The type of the field, as a doclet object. */
    public Type type() { 
	return type; 
    }
    
    /** Returns the wrapped field declaration. */
    public JFieldDeclarationType jfield() { return jfield; }
    
    /** Returns true because this class represents a field. */
    //@ pure
    public boolean isField() { return true; }

    /** Returns true if this field is marked transient in the java code. */
    public boolean isTransient() { 
	return org.multijava.util.Utils.hasFlag(longModifiers(),
			org.multijava.util.classfile.Constants.ACC_TRANSIENT); 
    }
    
    /** Returns true if this field is marked volatile in the java code. */
    public boolean isVolatile() { 
    	return org.multijava.util.Utils.hasFlag(longModifiers(),
			org.multijava.util.classfile.Constants.ACC_VOLATILE); 
    }

    /** Returns an array of any @serialField tags for this field. */ // NOT IMPLEMENTED //!FIXME!
    public SerialFieldTag[] serialFieldTags() { return new SerialFieldTag[0]; } 

    /** Returns the modifiers (e.g. public, private, static, final ...) of the field, encoded
	as bits of an int, as is done in MJC (cf. the Constants in 
	{@link org.multijava.util.classfile.Constants}).
    */
    public long longModifiers() { return field.modifiers(); }


    /** The fully-qualified name of the field. */
    //@ pure
    public String qualifiedName() {
	return containingClass().qualifiedName() + "." + name();
    }

    /** Get the value of a constant field.
     *  @return the value of a constant field.  The value is
     *  automatically wrapped in an object if it has a primitive type.
     *  If the field is not constant, returns null.
     *  Added for JDK 1.4 compatability.  */
    public Object constantValue() {
        return null; /* !FIXME!JDK 1.4 compatability */
    }

    /** Get the value of a constant field.
     *  @return the text of a Java langauge expression whose value is
     *  the value of the.  The expression uses no identifiers other
     *  than primitive literals. If the field is not constant, returns
     *  null.  Added for JDK 1.4 compatability.  */
    public String constantValueExpression() {
        return null; /* !FIXME!JDK 1.4 compatability */
    }

    /** Return the source position of the documented entity, or null
     *  if no position is available.
     *  Added for JDK 1.4 compatability.
     */
//      public SourcePosition position() {
//          return null; /* !FIXME!JDK 1.4 compatability */
//      }
}
