/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjPackageDoc.java,v 1.2 2004/03/12 03:14:53 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_141;

import com.sun.javadoc.*;
import java.io.File;
import java.util.ArrayList;
import org.multijava.mjc.JCompilationUnitType;
import org.multijava.util.classfile.ClassPath;

/** This class represents a java package and provides methods for finding a package object and
    the classes within a package.
    @author David R. Cok
*/
public class MjPackageDoc extends MjDoc implements com.sun.javadoc.PackageDoc {

    /** Holder for the String name of the package ('.' separated). */
    private String packageName;

    /** The location in the file system of the package root. */
    private String location;
    
    /** Returns the location of the root of the package. */
    public String location() { return location; }

    /** A map of the classes registered as in this package (String -> ClassDoc), 
	where the String is the name of the class (with enclosing classes, '.'-separated,
	but not the package name).
    */
    private java.util.Map classes = new java.util.HashMap();

    /** This is a list of the classes in the package in the order in which they have been added. */ 
    private java.util.ArrayList classesInOrder = new java.util.ArrayList(10);

    /** A collection of the compilation units containing external methods in this package. */
    private java.util.ArrayList extMethods = new java.util.ArrayList();
    
    /** A map of all the packages created (String -> MjPackageDoc), 
	where the String is the name ('/' separated and terminated) of the package. 
    */
    static private java.util.Map packages = new java.util.HashMap();
    
    /** Creates a new MjPackageDoc object from its '/' separated and terminated name; you should
	ordinarily use {@link #createPackage} so that distinct MjPackageDoc objects
	refer to distinct packages and all the classes for one package are registered
	in one MjPackageDoc object.  This constructor automatically registers the
	new MjPackageDoc object in the internal map of such objects.
    */ // !FIXME! Verify that we want names with terminating /
    private MjPackageDoc(String n) { 
	setIncluded(false);
	packageName = n.replace('/','.');
	if (packageName.length() > 0 && packageName.charAt(packageName.length()-1) == '.') 
		    packageName = packageName.substring(0,packageName.length()-1);
	packages.put(n,this);
	location = ClassPath.findPackageSource(packageName); // returns file system path to package root
	if (location == null) return;
	File f = new File(location + File.separator + n.replace('/',File.separatorChar) + "package.html");
	if (f.exists()) MjdocWrapper.setDocComment(this,f); 
    }

    /** Finds an MjPackageDoc from its '/'-separated and terminated 
	name, or returns null if none has yet been created. */
    public static MjPackageDoc findPackage(String p) { return (MjPackageDoc)(packages.get(p)); }

    /** Returns the MjPackageDoc object for the given '/'-separated and terminated
	name, creating a new one if none currently exists.
    */
    //@ requires s != null && s.length() != 0;
    public static MjPackageDoc createPackage(String s) {
	MjPackageDoc p = (MjPackageDoc)(packages.get(s));
	if (p == null) { p = new MjPackageDoc(s); }
	return p;
    }

    /** Returns the '.'-separated name of the package (with no terminating '.'). */
    public String name() { return packageName; }

    /** Returns the prefix to be appended on the front of a type name to make a fully-qualified name. */
    public String prefix() { 
	if (packageName.length() == 0) return ""; else return packageName + "."; 
    }

    /** Registers an MjClassDoc object as an ordinary (not error, exception, or interface) class. */
    public void addOrdinaryClass(ClassDoc d) { 
	MjClassDoc mjd = (MjClassDoc)classes.get(d.name());
	if (d instanceof MjClassDoc) {
	    int c = ((MjClassDoc)d).compare(mjd);
	    if (c == 1) return;
	    if (c == -1) {
		ordinaryClasses.remove(mjd);
		classesInOrder.remove(mjd);
	    }
	}
	classes.put(d.name(),d); 
	ordinaryClasses.add(d); 
	classesInOrder.add(d); 
    }
    
    /** Registers an MjClassDoc object as an error class. */    
    public void addErrorClass(ClassDoc d) { 
	MjClassDoc mjd = (MjClassDoc)classes.get(d.name());
	if (d instanceof MjClassDoc) {
	    int c = ((MjClassDoc)d).compare(mjd);
	    if (c == 1) return;
	    if (c == -1) {
		errorClasses.remove(mjd);
		classesInOrder.remove(mjd);
	    }
	}
	classes.put(d.name(),d); 
	errorClasses.add(d); 
	classesInOrder.add(d); 
    }

    /** Registers an MjClassDoc object as an exception class. */
    public void addExceptionClass(ClassDoc d) { 
	MjClassDoc mjd = (MjClassDoc)classes.get(d.name());
	if (d instanceof MjClassDoc) {
	    int c = ((MjClassDoc)d).compare(mjd);
	    if (c == 1) return;
	    if (c == -1) {
		exceptionClasses.remove(mjd);
		classesInOrder.remove(mjd);
	    }
	}
	classes.put(d.name(),d); 
	exceptionClasses.add(d); 
	classesInOrder.add(d); 
    }

    /** Registers an MjClassDoc object as an interface class. */
    public void addInterface(ClassDoc d) { 
	MjClassDoc mjd = (MjClassDoc)classes.get(d.name());
	if (d instanceof MjClassDoc) {
	    int c = ((MjClassDoc)d).compare(mjd);
	    if (c == 1) return;
	    if (c == -1) {
		interfaces.remove(mjd);
		classesInOrder.remove(mjd);
	    }
	}
	classes.put(d.name(),d); 
	interfaces.add(d); 
	classesInOrder.add(d); 
    }
    
    /** Adds a Doc representing the external methods in a compilation unit. */
    public void addExtMethodsUnit(MjExtMethodsDoc m) {
	extMethods.add(m);
    }
    
    /** Returns a collection of MjExtMethodsDoc objects for this package. */
    public ArrayList extMethodsUnits() {
	return extMethods;
    }

    /** Finds the MjExtMethodsDoc object corresponding to the given compilation unit
	(or creates one if none exists).
    */
    //FIXME -- use a faster lookup
    public MjExtMethodsDoc lookup(JCompilationUnitType unit) {
	java.util.Iterator i = extMethods.iterator();
	while(i.hasNext()) {
	    MjExtMethodsDoc m = (MjExtMethodsDoc)(i.next());
	    if (m.unit() == unit) return m;
	}
	MjExtMethodsDoc m = new MjExtMethodsDoc(this,unit);
	addExtMethodsUnit(m);
	return m;
    }
    
    /** Required but unused method. */
    public PackageDoc containingPackage() { return null; }

    /** A helper object to avoid having to create this all the time. */
    final private static ClassDoc[] carray = new ClassDoc[0];

    /** Holder for all ordinary classes registered as part of this package. */
    private ArrayList ordinaryClasses = new ArrayList(); // ArrayList of ClassDoc

    /** Holder for all error classes registered as part of this package. */
    private ArrayList errorClasses = new ArrayList(); // ArrayList of ClassDoc

    /** Holder for all exception classes registered as part of this package. */
    private ArrayList exceptionClasses = new ArrayList(); // ArrayList of ClassDoc

    /** Holder for all interfaces registered as part of this package. */
    private ArrayList interfaces = new ArrayList(); // ArrayList of ClassDoc

    /** Returns an array of the ordinary classes registered with this package. */
    public ClassDoc[] ordinaryClasses() { return (ClassDoc[])(ordinaryClasses.toArray(carray)); }

    /** Returns an array of the error classes registered with this package. */
    public ClassDoc[] errors() { return (ClassDoc[])(errorClasses.toArray(carray)); }

    /** Returns an array of the exception classes registered with this package. */
    public ClassDoc[] exceptions() { return (ClassDoc[])(exceptionClasses.toArray(carray)); }

    /** Returns an array of the interfaces registered with this package. */
    public ClassDoc[] interfaces() { return (ClassDoc[])(interfaces.toArray(carray)); }

    /** Returns an array of all the classes and interfaces registered with this package. */
    public ClassDoc[] allClasses() { 
	return (ClassDoc[])(classesInOrder.toArray(carray));
    } 

    /** Returns an array of all the classes and interfaces registered with this package.
     * Added for JDK 1.4 compatability.
     * (Note from Gary: I'm not sure from the JDK documentation what the following is supposed to mean...)
     * @return all classes and interfaces in this package, filtered to include only the included classes if filter==true.
     */
    public ClassDoc[] allClasses(boolean filter) { 
	return (ClassDoc[])(classesInOrder.toArray(carray));
    } 

    /** Returns the class or interface in this package given the class name ('.'-separated, including any
	enclosing classes, but not the package name). 
    */
    public ClassDoc findClass(String s) { 
	// notImplemented("findClass"); //!FIXME - when is this used
	return (ClassDoc)(classes.get(s)); 
    }

    /** Returns the '.'-separated name of the package. */
    public String toString() {
	return packageName;
    }

    /** This generates the String representation used by compareTo.  We use a different method than toString,
	because toString is not unique in all cases (e.g. methods with the same identifier but different arguments).
    */
    protected String compareString() {
	return packageName;
    }

    //FIXME - perhaps need to overload setIncluded to reset all contained classes as well as the package


    /** Return the source position of the documented entity, or null
     *  if no position is available.
     *  Added for JDK 1.4 compatability.
     */
//      public SourcePosition position() {
//          return null; /* !FIXME!JDK 1.4 compatability */
//      }

}
