/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjThrowsTag.java,v 1.1 2004/01/14 11:05:44 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_141;
import com.sun.javadoc.*;
import org.multijava.mjc.CTopLevel;
import org.multijava.mjc.CClass;

/** This class represents a 'throws' tag in a javadoc comment, which has
    an exception name and a description; the description itself is
    composed of a sequence of text and inline tags.
    @author David R. Cok
*/
public class MjThrowsTag extends MjTag implements ThrowsTag {

    /** A constant defining the 'kind' of this tag, as defined for Tag. */
    final public static String KIND = "@throws";

    /** An alternate tagname, a synonym for 'throws'. */
    final public static String KIND2 = "@exception";

    /** A holder for the name of the tag (@throws or @exception). */
    private String tagname;
    
    /** Cached value of the class of the exception. */
    private MjClassDoc exception;
    
    /** A holder for the name of the class of the exception. */
    private String name;
    
    /** A holder for the text of the description (but ordinarily you will
	want to use the sequence of tags provided by inlineTags() ).
    */
    private String comment;

    /** A constructor for the tag object; the argument is the String that
        follows '@throws'; it will be parsed into an exception name and the
	inline tags constituting the description.
    */ 
    public MjThrowsTag(Doc owner, String tagname, String t, SourcePosition pos) {
	super(owner,KIND,t,pos);
	this.tagname = tagname;
	int i = 0;
	int len = t.length();
	while (i<len && Character.isWhitespace(t.charAt(i))) ++i;
	int j = i;
	while (i<len && !Character.isWhitespace(t.charAt(i))) ++i;	    
	name = t.substring(j,i);
	fstags = i==j ? new Tag[0] : new Tag[]{new MjTextTag(owner,t.substring(j,len),pos)};
	while (i<len && Character.isWhitespace(t.charAt(i))) ++i;
	comment = t.substring(i,len);
	tags = inlineTags(comment,pos);
    }

    /** Returns the 'kind' of this tag, as required by the Tag interface. */
    public String kind() { return KIND; }

    /** Returns the actual name (@throws or @exception) used for the tag. */
    public String name() { return tagname; }

    /** The MjClassDoc object for java.lang.RuntimeException .*/
    static private MjClassDoc mjRuntimeException = null;

    /** The MjClassDoc object for java.lang.Error .*/
    static private MjClassDoc mjRuntimeError = null;

    /** Returns true if the argument represents an unchecked Java Throwable class, that is
	one that need not be declared in the throws-list of a method declaration.
    */
    static boolean unchecked(MjClassDoc exception) {
	if (mjRuntimeException == null) {
	    CClass c = CTopLevel.loadClass("java/lang/RuntimeException");
	    mjRuntimeException = MjdocWrapper.lookup(c);
	}
	if (mjRuntimeError == null) {
	    CClass c = CTopLevel.loadClass("java/lang/Error");
	    mjRuntimeError = MjdocWrapper.lookup(c);
	}
	return (exception.subclassOf(mjRuntimeException) || exception.subclassOf(mjRuntimeError));
    }
    
    /** Returns the exception referenced in the tag as a MjClassDoc object. */
    public ClassDoc exception() {
    	if (exception == null && name != null && name.length() != 0) {
	    // Note: The resolution of the exception class is deferred to this point -
	    // on demand rather than at construction time - so that all the classes
	    // being documented or referenced have already been registered as part
	    // of the RootDoc tree.
	    exception = resolveReference(name);
	    if (exception == null) {
		// The doclet code prints a warning upon receiving the null value, so we don't need to here
		//MjClassDoc mj = containingClass();
		//System.err.println("mjdoc: warning - Tag " + name() + ": Class or Package not found: " + name
		//    + ((mj==null)? "" : " in " + mj.qualifiedName()));;
	    } else if (owner instanceof MjExecutableMemberDoc) {
		MjExecutableMemberDoc m = (MjExecutableMemberDoc)owner;
		ClassDoc[] cc = m.thrownExceptions();
		int i;
		for (i=0; i<cc.length; ++i) {
		    if (exception.subclassOf(cc[i])) break;
		}
		if (i == cc.length && !unchecked(exception)) {
		    // !FIXME! Should we enable this warning?? - use error reporter
		    // FIXME - check that it is an exception ???
		    //System.err.println("mjdoc Warning: exception " + name + " in tag is not declared as thrown");
		}
	    } else {
		// FIXME - use warning reporter
		System.err.println("mjdoc Warning: " + tagname + " tag is not attached to a method or constructor");
	    }
	}
	return exception; 
    } 

    /** Returns the unparsed comment text; ordinarily you will want to use the inlineTags() . */
    public String exceptionComment() { return comment; }

    /** Returns the fully-qualified class name of the exception given in the javadoc comment. */
    public String exceptionName() { return name; }

    /** Return the source position of the documented entity, or null
     *  if no position is available.
     *  Added for JDK 1.4 compatability.
     */
//      public SourcePosition position() {
//          return null; /* !FIXME!JDK 1.4 compatability */
//      }
}
