/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjType.java,v 1.1 2004/01/14 11:05:44 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_141;
import com.sun.javadoc.*;
import org.multijava.mjc.*;

/** This class implements the javadoc Doclet Type interface, by wrapping appropriate
    type objects from the MJ compiler.  It can represent the regular java types and also
    the specialized A@B types used in multi-methods.
    @author David R. Cok
*/
public class MjType implements Type {

    /** The CType object from MJC that this MjType object wraps. */
    private CType ctype;
    
    /** The CType object from MJC that is the dynamic type (in static@dynamic); 
	equal to (==) ctype if this represents a conventional java type. */
    private CType dtype;
    
    /** The type of the dtype object, without any array dimension information. */
    private CType baseType; 
    
    /** The number of array dimensions. */
    private int depth;
    
    /** A constant String that is used to print pairs of brackets. */
    static final private String brackets = "[][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][][]";
	    // Note: The length of the above does not matter except for efficiency - see the
	    // implementation of dimension() below.
    

    /** Creates an instance of a javadoc Type object based on the MJC CType. */
    public MjType(CType t) { 
    	ctype = t;
	dtype = t;
	depth = 0;
	while (t.isArrayType()) { depth++; t = ((CArrayType)t).getIndexedType(); }
	baseType = t;
	//System.out.println("MJT " + dtype.toString() + " # " + name_noq_nodim(dtype) + " # " + baseType.toString() + " # " + isSpecializedType());
    }

    /** Creates an instance of a javadoc Type object based on the MJC CSpecializedType. */
    public MjType(CSpecializedType tt) { 
	ctype = tt.staticType();
	dtype = tt.dynamicType();
	depth = 0;
	CType t = dtype;
	while (t.isArrayType()) { depth++; t = ((CArrayType)t).getIndexedType(); }
	baseType = t;
	//System.out.println("MJTS " + dtype.toString() + " # " + name_noq_nodim(dtype) +  " # " + name_noq_dim(dtype) + " # " + baseType.toString() + " # " + isSpecializedType());
    }

    /** Returns the name of the type, including containing classes, but without package name,
     *  and without the dimension information, all separated by '.'.  This is used at least in
     *  writing the parameter types in an HTML page.  The result here may not be real useful for
     *  specialized types because the first type in A@B will have dimension information and the
     *  second will not.  In any case, for writing parameter types in method declarations, the
     *  html writer does the writing of the two types separately so that it can put html
     *  links in on the two types.
     */
    public String typeName() { 
	if (isSpecializedType()) {
	    return ctype.toString() + baseType.specializerSymbol() +
		name_noq_nodim(baseType); 
	} else {
	    return name_noq_nodim(baseType);
	}
    } 
    
    /** Returns the name of the type, including containing classes and its package name,
     *  and without the dimension information, all separated by '.';  the first type in
     *  A@B will have its dimension information.
     */
    public String qualifiedTypeName() {
	String s;
	if (isSpecializedType()) {
	    s = ctype.toString() + baseType.specializerSymbol() +
		baseType.toString();
	} else {
	    s = baseType.toString();
	}
	return s;
    }

    /** Returns the fully-qualified type name, including dimension information. */
    public String toString() {
	return qualifiedTypeName() + dimension();
    }
    
    /** Returns the dimension information, either as a zero-length string or as a
        String containing pairs of brackets like "[]".
     */
    public String dimension() { 
	if (!ctype.isArrayType()) return ""; 
	int d = 2*depth;
	if (d <= brackets.length()) return brackets.substring(0,d);
	String s = "";
	while (d > brackets.length()) { s = s + brackets; d -= brackets.length(); }
	s = s + brackets.substring(0,d);
	return s; 
    }

    /** Returns the MjClassDoc object for this type; returns null if the type is
	primitive; may create a new MjClassDoc object if there is none to represent
	this reference type.
     */
    public ClassDoc asClassDoc() { 
	CType c = ctype;
	if (c.isArrayType()) c = ((CArrayType)c).getBaseType();
	if (!c.isReference() || c.isValueType()) return null;
	return MjdocWrapper.lookup(c.getCClass()); 	
    }
    
    /** Returns the static type of a multi-method specialized type; this is the
	first in the A@B pair; in a regular java type, A and B are equal.
    */
    public CType staticType() { return ctype; }

    /** Returns the dynamic type of a multi-method specialized type; this is the
	second in the A@B pair; in a regular java type, A and B are equal.
    */
    public CType dynamicType() { return dtype; }

    /** Returns true if this is a specialized type as used in multi-method declarations. */
    public boolean isSpecializedType() { return ctype != dtype && !ctype.equals(dtype); }

    
    // NOTE: CType.toString() returns the fully qualified name, with dimensions

    /** Returns the type name without the package and without dimensions. */
    static String name_noq_nodim(CType t) {
	if (t.isArrayType()) { t = ((CArrayType)t).getBaseType(); }
	if (!t.isClassType()) return t.toString();
	return name_nested(((CClassType)t).getCClass());
    }
    
    /** Returns the type name without the package but with dimensions. */
    static String name_noq_dim(CType t) {
	if (t.isArrayType()) { return name_noq_dim(((CArrayType)t).getIndexedType()) + "[]"; }
	if (!t.isClassType()) return t.toString();
	return name_nested(((CClassType)t).getCClass());
    }
    
    /** Returns just the class name, with any enclosing classes - no dimensions, no package. */
    private static String name_nested(CClass c) {
	if (c.isNestedType()) return name_nested(c.owner()) + "." + c.ident();
	return c.ident();
    }
    
}
