/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.8 2004/08/14 01:51:20 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;

import org.multijava.mjdoc.*;
import org.multijava.mjc.*;
import org.multijava.util.compiler.ModifierUtility;
import org.multijava.util.compiler.CompilationAbortedError;
import org.multijava.util.compiler.CompilationAbortedException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.io.File;
import java.io.OutputStream;

/**
 * This class implements the entry point of the mjdoc tool.  It 
 * extends the Main class of the multijava compiler. Use
 * <code>java org.multijava.mjdoc.Main --help</code> to get usage
 * options. 
 * @author David R. Cok
 */
public class Main extends org.multijava.mjc.Main {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------

    /** The principal constructor for an object to do the work of the
        mjdoc tool.*/
    public Main() {
	super();
	parseJavadoc = true;
	appName = "mjdoc";
	MjdocWrapper.setWrapper(new MjdocWrapper(this));
	// In javadoc 1.4, stubs are allowed in java files
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    return true;
                }
            }
        );       
    }
    
    /** Construct a mjdoc tool (an object of <code>Main</code>) that
     * uses the given <code>ModifierUtility</code>. This constructor
     * allows the subclasses to have their own version of modifier 
     * utilities.
     */
    protected Main( ModifierUtility modUtil ) {
	super( modUtil );
	parseJavadoc = true;
	appName = "mjdoc";
	MjdocWrapper.setWrapper(new MjdocWrapper(this));
	// In javadoc 1.4, stubs are allowed in java files
        setContextBehavior(
            new ContextBehavior() {
                public boolean noBodyOK(CContextType context) {
                    return true;
                }
            }
        );       
    }

    // ----------------------------------------------------------------------
    // RUN FROM COMMAND LINE
    // ----------------------------------------------------------------------

    /**
     * The entry point when starting this program from the command line.
     *
     * @param	args		the command line arguments
     */
    public static void main(String[] args) {
	boolean success = compile(args);
	System.exit(success ? 0 : 1);
    }

    /**
     * Second entry point.  This function is overridden from the parent
     * so that we can instantiate a derived class Main object.  
     * @param	args		the command line arguments
     * @return true if run successfully with no errors
     */
    public static boolean compile(String[] args) {
	MjdocWrapper.setAppName("mjdoc");
	return new Main().run(args);
    }
	
    // ----------------------------------------------------------------------
    // GUI COMMAND LINE PARSER
    // ----------------------------------------------------------------------
        
    /**
     *  Runs the argument parser only so the GUI can process options from
     *  the command line
     */
    public boolean runParser(String[] args, MjcCommonOptions opt, 
                             ArrayList infiles) {

        // We need to extract all the javadoc (single -) style comments
        // in order to have compatibility with javadoc.
        args = MjdocWrapper.wrapper().setOptions(args, (MjdocOptions)opt);
        if (args == null) return true;
        
        // Now parse the remaining options in 'args'
        if (!super.runParser(args, opt, infiles)) return false;
        
        return true;
    }
        
    // ----------------------------------------------------------------------
    // RUN FROM GUI
    // ----------------------------------------------------------------------
    
    /**
     * Entry point for the GUI
     *
     * @param   args    the file, package, and directory names
     * @param   opt     the options for the tool
     * @param   os      the output stream for the compiler messages
     */
    public static boolean compile(String[] args, MjdocOptions opt, 
                                  OutputStream os) {
        MjdocWrapper.setAppName("mjdoc");
        return new Main().run(args, opt, os);
    }

    // ----------------------------------------------------------------------
    // METHODS FOR BOTH COMMAND LINE AND GUI
    // ----------------------------------------------------------------------

    /** This function creates an object to do the parsing of the
     *  command line arguments and to store the values of the flags
     *  and options so obtained (it does not actually do the argument
     *  parsing).  It is overridden here from mjc.Main so that we can
     *  instantiate a MjdocOptions object, in order to parse some new
     *  options specific to this compiler (cf. MjdocOptions.opt).
     *  @see MjdocOptions */
    protected MjcCommonOptions makeOptionsInstance() {
	return new MjdocOptions();
    }
	
    /** Returns opt as an instance of MjcCommonOptions so it can be assigned 
     *  to the options variable in mjc's Main.  This is done so we can access
     *  options that are only processed in MjcCommonOptions.
     */
    //@ also
    //@ requires opt instanceof MjdocOptions;
     protected MjcCommonOptions getOptionsInstance(MjcCommonOptions opt) {
        return opt;
     }

    /**
     * Parses the argument list.  Mutates infiles, options.  Returns true
     * if there were no problems parsing the arguments.  The list of 
     * command line arguments is parsed per the options defined in 
     * MjdocOptions.opt.  Values of given options are stored in the protected
     * variable super.options.  Remaining command-line arguments are 
     * considered to be filenames or directories.  
     */
    //@ also
    //@ requires args != null && infiles != null;
    //@ modifies options; // (* and the object that is referenced by infiles *)
    //@ ensures (* values set in the fields of options correspond to the contents of args *);
    //@ ensures (* items in infiles are non-null Strings corresponding to files that exist and were on the command line *);
    //@ ensures (* \result is true if arguments parsed successfully *);
    public boolean parseArguments(String[] args, ArrayList infiles) {
        
        // We need to extract all the javadoc (single -) style comments
        // in order to have compatibility with javadoc.
                
        args = MjdocWrapper.wrapper().setOptions(args, (MjdocOptions)options);
        if (args == null) return true;
                
        // Now parse the remaining options in 'args',  leaving in 'infiles' all
        // non-option command-line arguments (i.e. files, packages, directories).
        if (! options.parseCommandLine(args, infiles)) return false;
        String fcnsopt = ((MjdocOptions)options).fcns();
        if (fcnsopt.equals("ext")) {
            MjdocWrapper.doallfcns = false;
            MjdocWrapper.donofcns = false;
        } else if (fcnsopt.equals("none")) {
            MjdocWrapper.doallfcns = false;
            MjdocWrapper.donofcns = true;
        } else if (fcnsopt.equals("all")) {
            MjdocWrapper.doallfcns = true;
            MjdocWrapper.donofcns = false;
        } else {
            System.err.println("Invalid option argument for --fcns (should be ext, all, or none): " + fcnsopt);
            return false;
        }
        if (options.deprecation()) {
            System.err.println("Option --deprecation is not applicable and is ignored; 'use -nodeprecated to omit deprecated classes and methods from the html documentation");
        }
        
        if (infiles.size() > 0) filesFound = true;
        return true;
    }
    
    /** Handles the package names found on the command-line by making them 
	'selected' for the purpose of the javadoc API and converting them
	into directory names for the purpose of the parser.
    */
    //@ also
    //@ requires packages != null && dirs != null && infiles != null;
    //@ requires !packages.containsNull && packages.elementType <: \type(String);
    //@ requires !dirs.containsNull && dirs.elementType <: \type(File);
    //@ requires !infiles.containsNull && infiles.elementType <: \type(String);
    //@ modifies packageList;
    //@ ensures !packages.containsNull && packages.elementType <: \type(String);
    //@ ensures !dirs.containsNull && dirs.elementType <: \type(File);
    //@ ensures !infiles.containsNull && infiles.elementType <: \type(String);
    public void handlePackages(ArrayList packages, ArrayList dirs, ArrayList infiles) {
	packageList = new ArrayList(10);
	//@ set packageList.containsNull = false;
	//@ set packageList.elementType = \type(String);
	Iterator i = packages.iterator();
	while (i.hasNext()) {
	    String packageName = (String)(i.next());

	    File d = handlePackageName(packageName,dirs);

	    String nn = packageName.replace('.','/') + '/';
	    packageList.add(nn);

	    if (options.recursive()) addRecursivePackages(nn,d,dirs);

	}

	MjdocWrapper.selectPackages(packageList);

    }


    
    /** This adds all subdirectories (recursively) as packages to be processed. */
    // FIXME - One might rather only add those pakcages that have the same package root, or actually have java files in them,
    // but for now all subdirectories are added.

    //@ requires packageName != null;
    //@ requires dir != null;
    //@ modifies packageList;
    //@ ensures dirs != null && dirs.elementType <: \type(File) && !dirs.containsNull;
    protected void addRecursivePackages(String packageName, File dir, ArrayList dirs) {
        String[] ds = dir.list(subdirectoryFilter());
        for (int i=0; i<ds.length; ++i) {
            String n = ds[i];
            String pn = packageName + n + '/';
            File dn = new File(dir,n);
            packageList.add(pn);
            dirs.add(dn); 
            addRecursivePackages(pn,dn,dirs);
        }
    }

    // ----------------------------------------------------------------------
    // PROTECTED METHODS
    // ----------------------------------------------------------------------

    /**
     * Get the warning filter's class name from the options structure.
     * This is a hook method that can be overridden by subclasses.
     * @return null if there is no warning filter option specified.
     */
    //@ also
    //@ requires opts != null && opts instanceof MjdocOptions;
    protected String getWarningFilterNameFromOptions(MjcCommonOptions opts)
    {
        return ((MjdocOptions)opts).filter();
    }

    // -------------------------------------------------------------
    // TASK AND TASK SEQUENCING CODE
    // -------------------------------------------------------------

    /** 
     * This method uses the dynamic type of <code>oldTask</code> along
     * with the command line options to determine what task to add to
     * the task queue after the given task completes.  
     *
     * <pre><jml>
     * also
     * requires oldTask != null && oldTask.completed;
     * ensures \result == null || 
     *	       (\result.sequenceID() == oldTask.sequenceID());
     * </jml></pre>
     *
     */
    protected Task createTaskAfter( Task oldTask ) { // !CONVERT! multimethods
	if (oldTask instanceof MjdocTask) {
		return null;
	} else if (oldTask instanceof TypecheckTask) { 
	    // !FIXME! Does this work for mjdoc with external
	    // functions?? - without the false,
	    // MjMethodDoc.topConcreteMethod() does not work on external
	    // functions 

	    // abort the parsing after checking interfaces -
	    // don't check static initializers or the body
	    try {
		// All classes need to be registered with the doclet code,
		// so that appropriate links, qualified names etc. can be
		// used in the documentation.
		return new MjdocTask(((TreeProcessingTask) oldTask).trees(),
				     oldTask.sequenceID());
					
	    } catch (CompilationAbortedException e) {
		throw e;
	    } catch (Exception e) {
		System.out.println(e); //!FIXME! do something better with this error
		e.printStackTrace(System.out);
	    }
	    return null;
	} else {
	    // The oldTask is one of the superclass' tasks.  Do
	    // whatever it says next.  If it returns null, then it is
	    // finished.
	    return super.createTaskAfter( oldTask );
	}
    }

    /**
     * Implemented in mjdoc to cause the generation of all the html
     * files by the appropriate call into the doclet api.  This method
     * is called by {@link #run(String[])} when all reentrant
     * invocations of {@link #processTaskQueue()} have been exited and
     * in this implementation the method finalizes the generation of
     * the HTML.  This method should be overridden by subclasses to
     * perform final work after completion of all tasks.  */
    protected void taskQueueEmptied() {
	long startTime = System.currentTimeMillis();
	boolean b = MjdocWrapper.complete();
	if (!b) errorFound = true;
	if (!quietMode()) {
	    inform(MjdocMessages.HTML_GEN,
		   new Long(System.currentTimeMillis() - startTime));
	}
    }

    /**
     * This task creates MjDoc objects that hold all relevant information
     * about classes/methods/interfaces/packages/etc for generating html.
     * However, the actual html files are generated when taskQueueEmptied is
     * called, since that is initiated by a single call into the doclet api.
     */
    public class MjdocTask extends TreeProcessingTask {

	public MjdocTask( CompilerPassEnterable[] trees, 
				Object sequenceID) {
	    super( PRI_PRETTY_PRINT, sequenceID, trees, 
		   MjdocMessages.REGISTER ); 
	}

	protected void processTree( CompilerPassEnterable tree ) {
	    try {
		MjdocWrapper.wrapper().register(tree);
		if (sequenceID() == mainSequenceID()) {
		    MjdocWrapper.wrapper().selected(tree);
		}

	    } catch (Exception e) {
		// Report a problem, but continue to process other files
		// !FIXME! How to report it in the process return value???
		//errorFound = true; // Does this allow continuing with other files???
		reportTrouble(e);
	    }
	}
    }

    //---------------------------------------------------------------------
    // Fields
    //---------------------------------------------------------------------

    /** ArrayList (of String) of command-line arguments that are packages ('/'-separated and terminated name). */
    //@ spec_public
    protected ArrayList packageList;
    
}
