/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjExtMethodsDoc.java,v 1.2 2004/06/14 05:09:50 leavens Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;

import com.sun.javadoc.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import org.multijava.mjc.*;

import org.multijava.util.classfile.FieldInfo;

/** This object holds a list of external methods that are in a single compilation unit.
    @author David R. Cok
 */
public class MjExtMethodsDoc extends MjDoc {

    /** Holds the containing package. */
    private MjPackageDoc pkg;
    
    /** Holds the compilation unit that this set of external methods is in. */
    private JCompilationUnitType unit;
    
    /** ArrayList of the methods of this class, in the order of declaration. */
    private ArrayList methods = new ArrayList();

    /** A constructor for the case when source code is available (which makes javadoc comments
	and formal parameter identifiers available).
    */
    public MjExtMethodsDoc(MjPackageDoc p, JCompilationUnitType unit) {
	pkg = p;
	this.unit = unit;
	setIncluded(containingPackage().isIncluded());
    }

    /** Adds a method */
    public void add(MjMethodDoc m) { methods.add(m); }
    
    /** Returns the external methods in this compilation unit (ArrayList of MjMethodDoc). */
    public ArrayList methods() { return methods; }

    /** Returns the compilation unit the external methods are in. */
    public JCompilationUnitType unit() { return unit; }
    
    /** Returns the name of the compilation unit; no package name is prefixed. 
     */
    public String name() { 
	return unit.fileNameIdent() ; 
    } 

    /** Returns the name of the compilation unit, with the package name prefixed, and with '.' as separator. 
     */
    public String qualifiedName() { 
	return ((MjPackageDoc)containingPackage()).prefix() + name();
    }

    /** Returns the same thing as qualifiedName(). */
    public String toString() { return qualifiedName(); }
    
    /** This generates the String representation used by compareTo.  We use a different method than toString,
	because toString is not unique in all cases (e.g. methods with the same identifier but different arguments).
    */
    protected String compareString() {
	return name();
    }

    /** Returns the package that contains this compilation unit. */ 
    public PackageDoc containingPackage() { return pkg; }
    

    /** A cached value (after being initialized) of the array of classes imported into the compilation unit
	this class is defined in.
    */
    private ClassDoc[] cachedImportedClasses = null;
    
    /** Returns the array of classes imported into the compilation unit this class is defined in. */
    public ClassDoc[] importedClasses() {
	if (cachedImportedClasses == null) {
	    JClassOrGFImportType[] jf = unit.importedClasses();
	    ArrayList v = new ArrayList(jf.length);
	    for (int i=0; i<jf.length; ++i) {
		String n = jf[i].getName();
		CClass c = CTopLevel.loadClass(n);
		if (c == CClass.CLS_UNDEFINED) {
		    System.err.println("Could not find imported class " + n + " in " + qualifiedName());
		} else {
		    MjClassDoc mj = MjdocWrapper.lookup(c);
		    if (mj == null) {
			System.err.println("Could not find imported class " + n + " in " + qualifiedName());
		    } else {
			v.add(mj);
		    }
		}
	    }
	    cachedImportedClasses = new MjClassDoc[v.size()];
	    v.toArray(cachedImportedClasses);
	    
	}
	return cachedImportedClasses; 
    }
    
    /** A cached value (after being initialized) of the array of packages imported into the compilation unit
	this class is defined in.
    */
    private PackageDoc[] cachedImportedPackages = null;

    /** Returns the array of packages imported into the compilation unit this class is defined in. */
    public PackageDoc[] importedPackages() {
	if (cachedImportedPackages == null) {
	    JPackageImportType[] jf = unit.importedPackages();
	    ArrayList v = new ArrayList(jf.length);
	    for (int i=0; i<jf.length; ++i) {
		String n = jf[i].getName(); // '/'-separated name of the package
		MjPackageDoc mj = MjPackageDoc.createPackage(n);
		if (mj == null) {
		    // This is really an internal error, since createPackage ought always to produce something non-null
		    System.err.println("Could not find imported package " + n + " in " + qualifiedName());
		} else {
		    v.add(mj);
		}
	    }
	    cachedImportedPackages = new MjPackageDoc[v.size()];
	    v.toArray(cachedImportedPackages);
	    
	}
	return cachedImportedPackages; 
    }

}
