/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjMethodDoc.java,v 1.1 2004/01/14 11:05:44 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;
import com.sun.javadoc.*;
import org.multijava.mjc.CMethod;
import org.multijava.mjc.CClass;
import org.multijava.mjc.JMethodDeclarationType;
import org.multijava.mjc.MJTopLevelMethodDeclaration;
import org.multijava.mjc.CSpecializedType;
import java.util.ArrayList;
import java.util.Iterator;

/** This class represents a method (but not constructor) of a class; it is the bridge between the MJC syntax tree
    object and the MethodDoc interface of javadoc doclets.  It wraps a JMethodDeclarationType object, or
    simply CMethod object if only a binary representation is available.
    @author David R. Cok
*/
public class MjMethodDoc extends MjExecutableMemberDoc implements MethodDoc {

    /** Records if this method is an external method or not (only valid if
	the object is constructed using source code. 
    */
    private boolean isExternal = false;

    /** The associated MjExtMethodsDoc object if this is external. */
    private MjExtMethodsDoc unit = null;
    
    /** This constructs an object in cases where only binary code is available.
	@param cm the object wrapped by this class, representing a method in the java program
	@param container the class of which cm is a member
    */
    public MjMethodDoc(CMethod cm, MjClassDoc container) {
	super(cm,container);
    }

    /** This constructs an object in cases where source code is available and hence a
	JMethodDeclarationType object is available.
	@param jm the object wrapped by this class, representing a method in the java program
	@param container the class of which jm is a member
    */
    public MjMethodDoc(JMethodDeclarationType jm, MjClassDoc container) {
	super(jm,container);
	isExternal = jm instanceof MJTopLevelMethodDeclaration;
    }

    /** Returns the identifier of the method, without package or enclosing classes. */
    //@ also ensures \result != null;
    //@ pure
    public String name() { return cmethod().ident(); }

    /** Returns the fully-qualified identifier of the method. */
    //@ also ensures \result != null;
    //@ pure
    public String qualifiedName() { return containingClass().qualifiedName() + "." + cmethod().ident(); }

    /** Returns a String representation, used for ordering. */
    protected String compareString() {
	return name() + signature();
    }
    
    /** Returns true if the method is abstract. */
    //@ pure
    public boolean isAbstract() { return cmethod().isAbstract(); }

    /** Returns true since this object represents a method. */
    //@ pure
    public boolean isMethod() { return true; }

    /** Returns true if this method is declared outside its host class. */
    //@ pure
    public boolean isExternal() { return isExternal; } 
    
    /** Returns the super class that contains the method that this method overrides,
	or null if there is no such class.
    */
    public ClassDoc overriddenClass() {
	MjClassDoc c = (MjClassDoc)containingClass();
	MjClassDoc sp = MjdocWrapper.lookup(c.cclass().getSuperClass());
	if (sp == null) return null;
	MjMethodDoc m = sp.findMethodRecursive(this);
	if (m == null) return null;
	if (m.topConcreteMethod() != topConcreteMethod()) return null;
	return m.containingClass(); 
    }

    /** Return the method that this method overrides.
     *  Added for JDK 1.4 compatability.
     *  @return a MethodDoc representing a method definition in a
     *  superclass this method overrides, null if this method does
     *  not override.  */
    public MethodDoc overriddenMethod() {
        return null; /* !FIXME!JDK 1.4 compatability */
    }

    /** Returns the compilation unit the method is declared in; applicable only to external methods. */
    public MjExtMethodsDoc externalUnit() { return unit; }
    
    /** Records the compilation unit the method is declared in; applicable only to external methods. */
    public void setExternalUnit(MjExtMethodsDoc m) { unit = m; }

    /** Returns the package containing the compilation unit that declares this method. */
    public PackageDoc cuPackage() {
	if (unit == null) return containingPackage();
	else return unit.containingPackage();
    }
    
    /* Returns the return type of the method as a MjType object. */
    //@ also ensures \result != null;
    //@ pure
    public Type returnType() { return new MjType(cmethod().returnType()); }
    
    /* Overrides usual behavior to add information to the resulting javadoc 
	about the location of the method in the case where
	this method is a multijava external method.
    */ // FIXME - Test this.
//    public void setRawCommentText(String rawDocumentation) {
	//if (jmethod() instanceof MJTopLevelMethodDeclaration)
	//    parser = new ExternalMethodTagParser(this,rawDocumentation,jmethod().getTokenReference().file().//getAbsolutePath());
	//else
	//    super.setRawCommentText(rawDocumentation);
//    }
    
    // FIXME -- used ????
    public String location() {
	if (jmethod() == null) return null;
	else return  jmethod().getTokenReference().file().getAbsolutePath();
    }
	
    /** A modified parser class used by {@link #setRawCommentText} to add additional information about
	multijava external methods.
    */
/*    static protected class ExternalMethodTagParser extends MjTagParser {
	public ExternalMethodTagParser(Doc owner, String text, String location) {
	    super(owner, text);
	    //Tag t = new MjTextTag(owner,"<BR><B>External method declared in</B> <A HREF=" + location + ">"+location+"</A>");
	    //Tag[] n = new MjTag[firstSentenceTags.length+1];
	    //for (int i=0; i<firstSentenceTags.length; ++i) n[i] = firstSentenceTags[i];
	    //n[firstSentenceTags.length] = t;
	    //firstSentenceTags=n;
	    //n = new MjTag[inlineTags.length+1];
	    //for (int i=0; i<inlineTags.length; ++i) n[i] = inlineTags[i];
	    //n[inlineTags.length] = t;
	    //inlineTags=n;	
	}
    }
*/
    /** Returns true if the receiver and argument have the same identifier and static parameter types
	(but may have different receiver types). 
    */
    public boolean sameMethod(MjMethodDoc m) {
	if (! cmethod().ident().equals(m.cmethod().ident())) return false;
	CSpecializedType[] p = cmethod().parameters();
	CSpecializedType[] pp = m.cmethod().parameters();
	if (p.length != pp.length) return false;
	for (int i=0; i<p.length; ++i) {
	    if (! p[i].staticType().equals(pp[i].staticType())) return false;
	}
	return true;
    }
    
    //!FIXME! Need to copy inherited documentation to derived classes when there is none in the derived class.  But need to do this after all classes are loaded so we have all generic functions registered.  There seems to be a means to do that in the doclet code.

    /** Cached value of the top method that this method overrides. */
    private MjMethodDoc topConcreteMethod = null;

    /** Returns the top concrete method of this method. */
    public MjMethodDoc topConcreteMethod() {
	if (topConcreteMethod == null) {
	    try {
		CMethod top = cmethod().topConcreteMethod();
		CClass cc = top.owner();
		MjClassDoc mc = MjdocWrapper.lookup(cc);
		ArrayList md = mc.allmethods();
		Iterator i = md.iterator();
		while (i.hasNext()) {
		    MjMethodDoc mjm = (MjMethodDoc)(i.next());
		    if (mjm.cmethod() == top || mjm.cmethod().equals(top)) {
			topConcreteMethod = mjm;
			break;
		    }
		}
		if (topConcreteMethod == null) {
		    //!FIXME! Fix the error messages - these are really internal errors
		    System.err.println("COULD NOT FIND METHOD " + top + " FROM " + this.qualifiedName());
		    topConcreteMethod = this;
		}
	    } catch (Throwable e) {
		System.err.println("COULD NOT FIND METHOD " + e);
		topConcreteMethod = this;
	    }
/*
	    topConcreteMethod = this;
	    MjClassDoc c = (MjClassDoc)(containingClass().superclass());
	    if (c != null) { 
		MjMethodDoc t = c.findMethodRecursive(this);
		if (t != null) { 
		    topConcreteMethod = t.topConcreteMethod();
		}
	    }
*/
	}
	return topConcreteMethod;
    }

    /** Cached value of the local top method that this method overrides. */
    //private MjMethodDoc localTopMethod = null;
    
    /** Finds the topmost method that is still in this package. */
    // FIXME - this may be ambiguous if the ultimate top method is not in this package.
    /*
    public MjMethodDoc localTopMethod() {
	if (localTopMethod == null) {
	    localTopMethod = this;
	    MjClassDoc c = (MjClassDoc)(containingClass().superclass());
	    if (c != null) { 
		MjMethodDoc t = c.findMethodRecursive(this);
		if (t != null && t.containingPackage() == containingPackage()) {
		    localTopMethod = t.localTopMethod();
		}
	    }
	}
	return localTopMethod;
    }
    */
    
    /** Returns the name of the html file (without directory path) for the generic function that this method
	should be documented in.
    */
    public String gfFileName() {
	MjMethodDoc m = topConcreteMethod(); 
	String s = m.containingClass().qualifiedName().replace('.','-') + "$$" + m.name();
	Parameter[] pp = m.parameters();
	for (int i=0; i<pp.length; ++i) {
	    MjParameter p = (MjParameter)(pp[i]);
	    String t = ((MjType)(p.type())).staticType().toString(); 
	    // replace . by -
	    // replace [] by --
	    // replace @ by #
	    t = t.replace('.','-').replace('[','-').replace(']','-').replace('@','#');
	    s = s + "$" + t;
	}
	s = s + ".html";
	return s;
    }
    
    /** Return the source position of the documented entity, or null
     *  if no position is available.
     *  Added for JDK 1.4 compatability.
     */
//      public SourcePosition position() {
//          return null; /* !FIXME!JDK 1.4 compatability */
//      }
}
