/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjProgramElementDoc.java,v 1.2 2005/04/26 02:39:35 kui_dai Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;
import com.sun.javadoc.*;
import org.multijava.util.Utils;
import org.multijava.util.classfile.Constants;
import org.multijava.mjc.CClass;
import java.lang.reflect.Modifier;

/** This is an abstract class that is a super class for derived classes representing program elements
    such as classes, interfaces, constructors, methods, fields; it implements the javadoc ProgramElementDoc
    interface; derived classes wrap appropriate objects representing the program elements in the MJC class 
    structure.
    @author David R. Cok
*/
abstract public class MjProgramElementDoc extends MjDoc implements ProgramElementDoc {

    /** Holder for the package that contains this program element,
	as is available from the containing class.
    */
    //@ spec_public
    private MjPackageDoc pkg;
    
    /** Holder for the class that contains this program element 
	(a nested class is contained by its outer class). 
    */
    //@ spec_public
    private MjClassDoc containingClass;

    /** Creates this super class object for the case that the derived class 
	represents a class or interface; cl is the class or interface
	being wrapped.
    */
    //@ requires cl != null;
    public MjProgramElementDoc(CClass cl) {
	    containingClass = MjdocWrapper.lookup(cl.owner());
	    pkg = MjPackageDoc.createPackage(cl.packageName());
    }

    /** Creates this super class object for the situation that
	the derived class represents some program element other than
	a class or interface; the argument is the class or interface
	that contains the program element.
    */
    //@ requires container != null;
    public MjProgramElementDoc(MjClassDoc container) {
	    containingClass = container;
	    pkg = (MjPackageDoc)(container.containingPackage());	
    }

    /** Returns the class that contains this program element. */
    //@ also ensures \result == containingClass;
    //@ pure
    public ClassDoc containingClass() { return containingClass; }
    
    /** Returns the package that contains the class that contains this program element. */
    //@ also ensures \result == pkg;
    //@ pure
    public PackageDoc containingPackage() { return pkg; }

    /** Returns the identifer of this program element, as interpreted by derived classes.
    */
    //@ pure
    abstract public String name();

    /** Returns the long representing the java modifiers attached to this program element
	in the java program (cf. the constants such as ACC_PUBLIC in 
	{@link org.multijava.util.classfile.Constants}).
    */
    abstract public long longModifiers(); 

    /** Returns true if this program element has the 'final' modifier in the java program. */
    //@ pure
    public boolean isFinal() { return Utils.hasFlag(longModifiers(),Constants.ACC_FINAL); }  
    
    /** Returns true if this program element is package-private, that is it does not
	have the public, protected or private modifiers in the java program. 
    */
    //@ pure
    public boolean isPackagePrivate() { 
	long m = longModifiers(); 
	return !(Utils.hasFlag(m,Constants.ACC_PUBLIC)||Utils.hasFlag(m,Constants.ACC_PROTECTED)||Utils.hasFlag(m,Constants.ACC_PRIVATE));
    }
    
    /** Returns true if this program element has the 'private' modifier in the java program. */
    //@ pure
    public boolean isPrivate() { return Utils.hasFlag(longModifiers(),Constants.ACC_PRIVATE); }  
    
    /** Returns true if this program element has the 'public' modifier in the java program. */
    //@ pure
    public boolean isPublic() { return Utils.hasFlag(longModifiers(),Constants.ACC_PUBLIC); }  
    
    /** Returns true if this program element has the 'protected' modifier in the java program. */
    //@ pure
    public boolean isProtected() { return Utils.hasFlag(longModifiers(),Constants.ACC_PROTECTED); } 
    
    /** Returns true if this program element has the 'static' modifier in the java program. */
    //@ pure
    public boolean isStatic() { return Utils.hasFlag(longModifiers(),Constants.ACC_STATIC); }  

    /** This lists the modifiers of the program element as a String, with spaces separating 
	modifiers.  The order of the modifiers in the output is not specified, but the following
	seems to match what the Standard doclet does (and is in order of increasing bit position).
    */ 
    public String modifiers() { 
	// Not positive that this is the canonical order of modifiers - FIXME
	StringBuffer sb = new StringBuffer();
	long m = longModifiers(); 
	if ((m&Constants.ACC_PUBLIC)!=0) sb.append("public ");
	if ((m&Constants.ACC_PRIVATE)!=0) sb.append("private ");
	if ((m&Constants.ACC_PROTECTED)!=0) sb.append("protected ");
	if ((m&Constants.ACC_STATIC)!=0) sb.append("static ");
	if ((m&Constants.ACC_FINAL)!=0) sb.append("final ");
	if ((m&Constants.ACC_SYNCHRONIZED)!=0) sb.append("synchronized ");
	if ((m&Constants.ACC_VOLATILE)!=0) sb.append("volatile ");
	if ((m&Constants.ACC_TRANSIENT)!=0) sb.append("transient ");
	if ((m&Constants.ACC_NATIVE)!=0) sb.append("native ");
	if ((m&Constants.ACC_INTERFACE)!=0) sb.append("interface ");
	else if ((m&Constants.ACC_ABSTRACT)!=0) sb.append("abstract ");
	    // !FIXME! Note the else above - for some reason mjc adds abstract onto any interface.
	    // We remove it here, in part to match javadoc
	if ((m&Constants.ACC_STRICT)!=0) sb.append("strictfp ");
	if (sb.length() == 0) return "";
	return sb.substring(0,sb.length()-1); // Leave off the last space
    }
    
    /** This returns the modifiers of the program element encoded as bits of an int;
	the encoding is that of the javac compiler, not of the mj compiler.
    */ 
    public int modifierSpecifier() {  
	// longModifiers() depends on the bit encoding of modifiers as defined in org.multijava.util.classfile.Constants
	// modifierSpecifier() wants the bit encodings as they are defined in java.lang.reflect.Modifier
	// As long as the two encodings are the same we do not need to do any translation.
	// The static block below checks this.
	return (int)longModifiers(); 
    } 

    /** Returns the qualified name (package + containing classes + program element identifier) of
	the program element.
    */
    //@ pure
    abstract public String qualifiedName();
    
    static {
	// We check these bit patterns because the implementation of modifierSpecifier() above 
	// presumes that the encoding of flags is the same between MJC and Java.  If not,
	// that is OK, as long as you implement modifierSpecifier() to do the translation.
	// -- DRCok
	if (Constants.ACC_PUBLIC != Modifier.PUBLIC ||
	    Constants.ACC_PROTECTED != Modifier.PROTECTED ||
	    Constants.ACC_PRIVATE != Modifier.PRIVATE ||
	    Constants.ACC_ABSTRACT != Modifier.ABSTRACT ||
	    Constants.ACC_FINAL != Modifier.FINAL ||
	    Constants.ACC_INTERFACE != Modifier.INTERFACE ||
	    Constants.ACC_NATIVE != Modifier.NATIVE ||
	    Constants.ACC_STATIC != Modifier.STATIC ||
	    Constants.ACC_STRICT != Modifier.STRICT ||
	    Constants.ACC_SYNCHRONIZED != Modifier.SYNCHRONIZED ||
	    Constants.ACC_TRANSIENT != Modifier.TRANSIENT ||
	    Constants.ACC_VOLATILE != Modifier.VOLATILE) {
	System.out.println("IMPLEMENTATION ERROR: Bit encodings of modifiers are not equivalent between MJC and Java");
	System.out.println("PUBLIC " + Constants.ACC_PUBLIC + " " + Modifier.PUBLIC);
	System.out.println("PROTECTED " + Constants.ACC_PROTECTED + " " + Modifier.PROTECTED);
	System.out.println("PRIVATE " + Constants.ACC_PRIVATE + " " + Modifier.PRIVATE);
	System.out.println("ABSTRACT " + Constants.ACC_ABSTRACT + " " + Modifier.ABSTRACT);
	System.out.println("FINAL " + Constants.ACC_FINAL + " " + Modifier.FINAL);
	System.out.println("INTERFACE " + Constants.ACC_INTERFACE + " " + Modifier.INTERFACE);
	System.out.println("NATIVE " + Constants.ACC_NATIVE + " " + Modifier.NATIVE);
	System.out.println("STATIC " + Constants.ACC_STATIC + " " + Modifier.STATIC);
	System.out.println("STRICT " + Constants.ACC_STRICT + " " + Modifier.STRICT);
	System.out.println("SYNCHRONIZED " + Constants.ACC_SYNCHRONIZED + " " + Modifier.SYNCHRONIZED);
	System.out.println("TRANSIENT " + Constants.ACC_TRANSIENT + " " + Modifier.TRANSIENT);
	System.out.println("VOLATILE " + Constants.ACC_VOLATILE + " " + Modifier.VOLATILE);
	// We do a hard exit here instead of a graceful CompilationAbortedError.
	// Could do the latter, but if we have this condition, it isn't even
	// worth running the JUnit tests.
	System.exit(1);
	}
    }

    
}
