/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjRootDoc.java,v 1.5 2004/08/14 01:51:20 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;

import com.sun.javadoc.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import org.multijava.mjc.MjcMessages;

import org.multijava.util.compiler.Compiler;
import org.multijava.util.FormattedException;

/** This class implements the Doclet interface for a RootDoc by 
    containing lists of bridge classes for packages and classes.
    @author David R. Cok
*/
public class MjRootDoc extends MjDoc implements RootDoc {

    /** Holds a representation of the command-line options relevant to html generation. */
    //@ spec_public
    private String[][] opt;

    /** Holds a reference to the compiler, for printing error, warning, and informational messages. */
    //@ spec_public
    private Compiler compiler;
    
    /** Holds a mapping from the qualifiedTypeName of a class/interface to the MjClassDoc object that wraps it. */
    private Map classes = new HashMap();
    
    /** Holds a mapping from the name of a package to the MjPackageDoc object that wraps it. */
    private Map packages = new HashMap();

    /** Holds a list of all the methods in all the classes (not interfaces ? !FIXME!) that have been registered. */
    private java.util.SortedSet allMethods = new java.util.TreeSet();
    
    /** Holds a list of the classes that are 'specified classes', namely those that will have html generated. */
    private ArrayList specifiedClasses = new ArrayList(); // ArrayList of ClassDoc

    /** Holds a list of the packages that are 'specified packages', namely those that will have html generated. */
    private ArrayList specifiedPackages = new ArrayList(); // ArrayList of PackageDoc
    private HashSet specifiedPackagesHash = new HashSet();

    /** Holds a list of the ext methods units that are 'specified', namely those that will have html generated. */
    private ArrayList specifiedExtMethodUnits = new ArrayList(); // ArrayList of MjExtMethodsDoc

    /** Holds a list of the external methods compilation units. */
    private ArrayList extMethodsCU = new ArrayList();
        
    /** If true, no informational messages are printed. */
    private boolean quiet = false;

    //FIXME
    public void quiet(boolean b) { 
        quiet = b; 
    }
    
    /** Constructs a new RootDoc, innitialized with the relevant command-line options. 
        @param opts an array of String[] each of which has an option and the option's arguments in it
    */
    //@ requires opts != null;
    //@ ensures opt == opts;
    public MjRootDoc(Compiler compiler, String[][] opts) { 
        this.compiler = compiler;
        opt = opts;
        if (opt != null) for (int i=0; i<opts.length; ++i) {
            if (opts[i][0].equals("-quiet")) { quiet = true; break; }
        }
        setIncluded(false);
    }

    /** Adds an external method compilation unit. */
    public void add(MjExtMethodsDoc m) { extMethodsCU.add(m); }
    
    /** Returns the collection of all the registered external method compilation units. */
    public ArrayList extMethods() { return extMethodsCU; }

    /** Returns the collection of the specified external method compilation units. */
    public ArrayList specifiedExtMethods() { return specifiedExtMethodUnits; }

    /** Marks a ext. method unit as specified (it must also be added explicitly by {@link #add(ClassDoc)}). */
    public void specify(MjExtMethodsDoc c) { specifiedExtMethodUnits.add(c); c.setIncluded(true); }
    
    /** Adds a method to the registered set of methods. */
    public void add(MjMethodDoc m) { allMethods.add(m); }

    /** Returns the list of methods that have been registered. */
    public java.util.SortedSet allMethods() { return allMethods; }

    /** Returns the list of top concrete methods of all the registered
     * methods. */
    public java.util.SortedSet topConcreteMethods(PackageDoc p) { 
        java.util.Iterator i = allMethods().iterator(); 
        java.util.SortedSet tops = new java.util.TreeSet(new MjExecutableMemberDoc.Comp());
        while (i.hasNext()) {
            MjMethodDoc m = (MjMethodDoc)(i.next());
            // FIXME What about the fcns option
            if (m.isExternal() && (p==null || m.topConcreteMethod().cuPackage() == p) && m == m.topConcreteMethod()) tops.add(m);
        }
        return tops;
    }
    
    /** Adds a class or interface to the set of objects that the RootDoc knows about. */
    public void add(ClassDoc c) { classes.put(c.qualifiedTypeName(), c); }
    
    /** Marks a class or interface as specified (it must also be added explicitly by {@link #add(ClassDoc)}). */
    public void specify(ClassDoc c) { 
        specifiedClasses.add(c); 
        ((MjDoc)c).setIncluded(true); 
    }

    /** Marks a class or interface as NOT specified. */
    public void unspecify(ClassDoc c) { 
	specifiedClasses.remove(c);
	((MjDoc)c).setIncluded(false); 
    }
    
    /** Adds a package (but not its contents) to the set of objects that the RootDoc knows about. */
    public void add(PackageDoc c) { packages.put(c.name(), c); }

    /** Marks a package as specified (it must also be added explicitly by {@link #add(PackageDoc)}). */
    public void specify(PackageDoc c) { 
	specifiedPackages.add(c); 
	specifiedPackagesHash.add(c);
	((MjDoc)c).setIncluded(true); 
    }

    public boolean isSpecified(PackageDoc c) {
	return specifiedPackagesHash.contains(c);
    }
    
    /** Returns all the classes and interfaces specified, whether as a file or as a package. */
    public ClassDoc[] classes() { 
        ArrayList v = new ArrayList(specifiedClasses.size());
        v.addAll(specifiedClasses);
        java.util.Iterator i = specifiedPackages.iterator();
        while (i.hasNext()) {
            PackageDoc p = (PackageDoc)(i.next());
            ClassDoc[] c = p.allClasses();
            for (int j=0; j<c.length; ++j) {
                v.add(c[j]);
            }
        }
        ClassDoc[] result = new ClassDoc[v.size()];
        v.toArray(result);
        
        return result;
    }

    /** Returns the cached option structure, as provided to the constructor. */
    //@ also
    //@ ensures \result == opt;
    //@ pure
    public String[][] options() { return opt; }

    /** Returns the class or interface with a specific fully-qualified name, null if it is not known. 
        A class may be found even if it is not specified on the command line, if it has been referenced
        in parsing some otherwise parsed class.  It is not sufficient simply to be a legitimate class
        on the CLASSPATH. CAUTION: Not guaranteed to produce the same result as the doclet api. !FIXME! - should this have the classes inside registered packages ?
    */
    public ClassDoc classNamed(String qualifiedName) { 
        return (ClassDoc)(classes.get(qualifiedName)); 
    }

    /** Returns the package with a specific name, null if it is not known.  A package is known if a
        file has been parsed or class referenced that is in that package.
        NOTE: The above description is what the Doclet API specifies, but javadoc seems to recognize
        more packages than have been referenced in the input source code.  !FIXME!
    */
    public PackageDoc packageNamed(String name) { return (PackageDoc)(packages.get(name)); }

    /** Returns an array of all the specified classes (but not classes in specified packages). */
    public ClassDoc[] specifiedClasses() {
        return (ClassDoc[])specifiedClasses.toArray(new ClassDoc[0]);
    }

    /** Returns an array of all the specified packages. */
    public PackageDoc[] specifiedPackages() { 
        return (PackageDoc[])specifiedPackages.toArray(new PackageDoc[0]);
    }

    // The following are required by the RootDoc interface

    /** Required by the Doc interface, but not normally called. */
    //@ also
    //@ ensures \result == "*RootDoc*";
    //@ pure
    public String name() { return "*RootDoc*"; } 

    /** Conventional toString value, matches javadoc behavior. */
    public String toString() { return name(); }
    
    //FIXME 
    public PackageDoc containingPackage() { return null; }
    
    /** This method prints an error to diagnostic output. */
    public void printError(String s ) { 
        printError(null,s);
    }

    /** This method prints an error to diagnostic output. */
    public void printError(SourcePosition pos, String s ) { 
        compiler.reportTrouble(new FormattedException((pos!=null?pos + " ":"") + MjdocWrapper.appName() + ": error - "  + s)); 
    }


    /** This method prints an warning to diagnostic output. */
    public void printWarning(String s ) { 
        printWarning(null,s);
    }

    /** This method prints a warning to diagnostic output.
     */
    public void printWarning(SourcePosition pos, String s ) {
        compiler.inform((pos!=null?pos + " ":"") + MjdocWrapper.appName() + ": warning - "  + s); 
    }

    /** This method prints an notice to normal output. */
    public void printNotice(String s ) { 
        printNotice(null,s);
    }

    /** This method prints an notice to normal output. */ 
    public void printNotice(SourcePosition pos, String s ) { 
        compiler.inform((pos!=null?pos + " ":"") + s, Compiler.PRINT_TO_OUT); 
    }

    /** This method prints a notice to System.out.
        Added for JDK 1.4 compatability.
     */
//      public void printNotice(SourcePosition pos, String s )
//      {
//          if (!quiet) {
//              System.out.print(pos.toString() + " ");
//              printNotice(s);
//          }
//      }

    /** Return the source position of the documented entity, or null
     *  if no position is available.
     *  Added for JDK 1.4 compatability.
     */
//      public SourcePosition position() {
//          return null; /* !FIXME!JDK 1.4 compatability */
//      }
}
