/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjSeeTag.java,v 1.2 2004/06/14 05:09:50 leavens Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;

import com.sun.javadoc.*;
import java.io.File;
import java.util.ArrayList;

import org.multijava.util.classfile.ClassPath;

/** This class represents the content of an @see or @link tag. 
    @author David R. Cok
*/
public class MjSeeTag extends MjTag implements SeeTag {

    /** A constant that is the value of 'kind' for this tag. */
    public final static String KIND = "@see";

    /** This constructs a tag object, parsing the given text and looking up referenced classes.  The behavior
	of some of the fields is a bit erratic, but tuned to match javadoc 1.3.1.
	@param n - the name of the tag, e.g. @see or @link
	@param t - the text of the tag
    */
    public MjSeeTag(Doc owner, String n, String t, SourcePosition pos) { 
	super(owner,n,t,pos); 
	int i = 0;
	int len = t.length();
	while (i<len && Character.isWhitespace(t.charAt(i))) ++i; // skip white space
	if (i == len) {
	    // No argument
	    label = "";
	    tags = new Tag[0];
	    fstags = tags;
	} else if (t.charAt(i) == '"') {
	    // quoted style
	    int k = len;
	    do { --k; } while (Character.isWhitespace(t.charAt(k)));
	    if (t.charAt(k) != '"' || k == i) {
		// No closing " 
		// FIXME - should use the reporting mechanism in RootDoc
		System.err.println("mjdoc: warning : Tag @see: no final close quote: \"" + t + "\"");
	    }
	    label = "";
	    tags = new MjTag[]{ new MjTextTag(owner,t.trim(),pos) };
	    fstags = tags;
	} else if (t.charAt(i) == '<') {
	    // explicit link style
	    label = "";
	    tags = new MjTag[]{ new MjTextTag(owner,t.substring(i),pos) };
	    fstags = tags;
	} else {
	    int j = i;
	    int endtag = len;
	    while (i<len) {
		if (t.charAt(i) == '(') {
		    while (i<len && t.charAt(i) != ')') ++i;
		    if (i==len) {
			// FIXME - should use the reporting mechanism in RootDoc
			System.err.println("mjdoc: warning - no ')' to match the '(' in tag " + n  + " " + t);
			break;
		    }
		}
		if (!Character.isWhitespace(t.charAt(i))) { ++i; continue; }
		else {
		    endtag = i;
		    break;
		}
	    }
	    while (i<len && Character.isWhitespace(t.charAt(i))) ++i; // skip white space
			
	    String name = t.substring(j,endtag);
	    label = t.substring(i,len); 
	    tags = new MjTag[]{ new MjTextTag(owner,t.trim(),pos) }; 
	    fstags = tags;
	    
	    int p = name.indexOf('#');
	    if (p == -1) {
		prefix = name;
		memberName = null;
	    } else {
		prefix = name.substring(0,p);
		memberName = (p+1 == name.length())? null : name.substring(p+1);
	    }
	}
    }

    /** An accessor that returns the 'kind' of this tag, as required by the Tag interface. */
    public String kind() { return KIND; }

    /** Holder for the value of the label. */
    private String label;
    
    /** Holder for the string in the tag that indicates the class or package. */
    private String prefix;

    /** Flag to indicate whether the search for the class or package has been performed. */
    private boolean resolved = false;
    
    /** Cached value of the class indicated in the tag. */
    private MjClassDoc classDoc = null;
    
    /** Cached value of the package of the class indicated in the tag. */
    private MjPackageDoc pkgDoc = null;
    
    /** Cached value of the class member indicated in the tag, if any. */
    private MjMemberDoc memberDoc = null;
    
    /** Holder for the string in the tag that indicates the member, if any. */
    private String memberName;
    
    /** Returns the label part of the tag, if any. */
    public String label() { return label; } 

    /** Returns the class referenced in the tag, if any. */
    public ClassDoc referencedClass() { 
	if (!resolved) resolve();
	return classDoc;
    } 

    /** Returns the fully-qualified name of the class or package referenced in the tag. */
    public String referencedClassName() { return referencedClass() == null ? prefix : referencedClass().qualifiedName(); } 

    /** Returns the member referenced in the see or link tag, if any (otherwise null). */
    public MemberDoc referencedMember() { 
	if (!resolved) resolve();
	return memberDoc; 
    } 

    /** Returns the name of the member and signature as written in the tag. */
    public String referencedMemberName() { 
	return memberName; 
    }

    /** Returns the package if the target of the see tag is a package name,
	otherwise returns null.  Checks everything on CLASSPATH.
    */
    public PackageDoc referencedPackage() {
	if (!resolved) resolve();
	return pkgDoc;
    }

    /** Helper function to determine whether the class/package name that is in 
	the tag is a package or a class and to create the appropriate Doc object.
    */
    private void resolve() { // FIXME this routine needs checking and scrubbing
	resolved = true;
	if (prefix == null && memberName == null) return;
	if (prefix == null || prefix.length()==0) {
	    if (owner instanceof ClassDoc) {
		classDoc = (MjClassDoc)owner;
	    } else {
		classDoc = (MjClassDoc)(((MjMemberDoc)owner).containingClass());
	    }
	} else {
	    classDoc = resolveReference(prefix);
	    // if (classDoc == null) {
            // // The doclet code prints a warning upon receiving the null value, so we don't need to here
            // // MjClassDoc mj = containingClass();
            // // System.err.println("mjdoc: warning - Tag " + name() + ": Class or Package not found: " + prefix
            // //   + ((mj==null)? "" : " in " + mj.qualifiedName()));;
	    // }
	}
	if (classDoc != null) {
	    if (memberName == null) return;
	    int paren = memberName.indexOf('(');
	    boolean hasParen = paren != -1;
	    if (paren==-1) paren = memberName.length();
	    String ident = memberName.substring(0,paren);

	    String cname = prefix;
	    if (prefix == null) {
		cname = owner.name();
	    }
	    int i = cname.lastIndexOf('.');
	    cname = cname.substring(i+1);
	    if (cname.length() == 0) cname = classDoc.name();
	    
	    MjExecutableMemberDoc[] m;
	    if (ident.equals(cname)) { // Constructor
		m = (MjConstructorDoc[])(classDoc.constructors());
	    } else { // Method
		m = (MjMethodDoc[])(classDoc.methods());
	    }

	    if (memberDoc == null && !hasParen) {
		// Check fields as well
		MjMemberDoc[] mm =  (MjMemberDoc[])(classDoc.fields());
		for (int k = 0; k<mm.length; ++k) {
		    if (mm[k].name().equals(ident)) { memberDoc = mm[k]; break; }
		}

	    }

	    // Now we are going to check if the parsed method in the see tag
	    // matches any of the executable members in the array m.
	    if (memberDoc == null) for (i=0; i<m.length; i++) {
		MjExecutableMemberDoc md = m[i];

		// If the identifier does not match, then try the next one
		if (! md.name().equals(ident)) continue;

		// If no arguments are given for the identifier, then we
		// match - on the first one with the same identifier
		if (!hasParen) { // FIXME -- should warn
		    memberDoc = md;
		    break;
		}

		// Otherwise we see if the parameter types match
		Parameter[] p = md.parameters();
		String[] s = parseArgs(memberName);

		// If there are different numbers of parameters then
		// there is clearly not a match
		if (p.length != s.length) continue;

		int k;
		for (k=0; k<p.length; ++k) {
		    // NOTE: Make sure this comparison works for specialized types
		    MjType t = (MjType)(p[k].type());
		    if (t.toString().equals(s[k])) continue;
		    if (t.toString().endsWith(s[k])) continue;
		    // !FIXME! Sensitive to embedded whitespace in the strings
		    // !FIXME! The above requires the strings to be fully-qualified and does not
		    // emulate proper lookup procedures
		    break;		// DID NOT MATCH
		}
		if (k == p.length) { memberDoc = md; break; }
	    }
	    return;
	}
	if (memberName != null || prefix == null) {
	    return;
	}
    
	String n = prefix.replace('.',File.separatorChar);
	pkgDoc = MjPackageDoc.findPackage(n);
	if (pkgDoc != null) return;
	if (ClassPath.findPackage(prefix) == null) {
	    return;
	}
	pkgDoc = MjPackageDoc.createPackage(n);
	return;
    }
    
    /** Parses an argument string of a see or link tag into the individual strings
	giving the text of the argument type; returns an array of String; no lookup
	of types from the strings is performed.
    */
// FIXME - whitespace not allowed in the typename
    static public String[] parseArgs(String mem) {
	int p = mem.indexOf('(');
	if (p == -1) return null;
	++p;
	ArrayList v = new ArrayList();
	while (true) {
	    int pp = mem.indexOf(',',p);
	    if (pp == -1) break;
	    String s = mem.substring(p,pp).trim();
	    int ppp = s.indexOf(' ');
	    if (ppp != -1) s = s.substring(0,ppp);
	    if (s.length() == 0) {
		// FIXME - use error reporter in RootDoc
		System.err.println("mjdoc warning: an argument is all whitespace in a see/link tag: " + mem);
	    }
	    v.add(s);
	    p = pp+1;
	}
	int pp = mem.indexOf(')',p);
	String s;
	if (pp == -1) {
	    s = (mem.substring(p).trim());
	} else if (p == pp) {
	    s = "";
	} else {
	    s = (mem.substring(p,pp).trim());
	}
	pp = s.indexOf(' ');
	if (pp != -1) s = s.substring(0,pp);
	if (s.length() != 0) v.add(s);
	else if (v.size() != 0) {
		// FIXME - use error reporter in RootDoc
	    System.err.println("mjdoc warning: an argument is all whitespace in a see/link tag: " + mem);
	}
	String[] result = new String[v.size()];
	v.toArray(result);
	
	return result;
    }

}
