/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjTag.java,v 1.1 2004/01/14 11:05:44 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;

import com.sun.javadoc.*;
import java.util.ArrayList;
import org.multijava.mjc.CClass;
import org.multijava.mjc.CCompilationUnit;
import org.multijava.mjc.CTopLevel;

import org.multijava.util.classfile.ClassPath;

/** This class is the abstract parent task for the various special kinds of
    tag objects that represent tags in javadoc comments.  Generically speaking,
    a tag consists of a kind, a name, and a set of inline tags which, when 
    rendered sequentially, constitute the description of the tag.  Some tags
    in addition have other fields that are parsed from the javadoc comment, 
    such as the name of a formal parameter or class of a thrown exception.
    @author David R. Cok
*/
abstract public class MjTag implements com.sun.javadoc.Tag {

    /** The raw, unparsed text of the tag. */
    private String text;
    
    /** The name of the tag. */
    private String name;
    
    /** The sequence of inline tags that constitute the description of the tag.
	Derived class constructors must set this field directly.
    */
    //@ spec_public
    protected Tag[] tags = null;

    /** This holds the value for firstSentenceTags.  
	Derived class constructors must set this field directly.
    */
    //@ spec_public
    protected Tag[] fstags = null;

    /** The Doc object to which this tag belongs. */
    //@ spec_public
    protected MjDoc owner = null;

    /** The SourcePosition giving the location of this tag; currently the source 
	position of the javadoc comment.
    */
    //@ spec_public
    protected SourcePosition position = null;
    
    
    /** A constructor for this parent class, which sets the name and unparsed text.
	Derived class constructors are required in addition to compute and set the
	value of 'tags'.
	@param n the name of the javadoc comment tag
	@param t the unparsed text of the javadoc comment tag
    */
    public MjTag(Doc owner, String n, String t) { 
	this.owner = (MjDoc)owner; 
	name = n; 
	text = t;
	position = null; 
    }
    public MjTag(Doc owner, String n, String t, SourcePosition pos) { 
	this.owner = (MjDoc)owner; 
	name = n; 
	text = t; 
	position = pos;
    }

    /** Returns the class that this tag modifies or is in (note that 'owner' is the program element
	the tag modifies. 
    */
    public MjClassDoc containingClass() {
	if (owner instanceof MjClassDoc) return (MjClassDoc)owner;
	if (owner instanceof MjProgramElementDoc) return (MjClassDoc)(((ProgramElementDoc)owner).containingClass());
	return null;
    }
	
    /** This returns the value that javadoc returns for the dervied classes.  The value does 
	not appear to have any meaning or specification, and the result does not appear to 
	be useful, but for easy of comparison testing, we copy javadoc.
    */
    public Tag[] firstSentenceTags() { return fstags; } 

    /** Returns the descriptionm (the text) of the tag as a sequence of tags (typically
	text and inline tags such as link tags).  This requires that the protected 'tags'
	field have been set by derived class constructors.
    */
    //@ also
    //@ requires tags != null;
    //@ ensures \result == tags;
    public Tag[] inlineTags() { 
	return tags; 
    } 

    /** This method must be overridden by derived classes to return the 'kind' of the tag. */
    abstract public String kind();

    /** Accessor method to return the owner of the tag. */
    public Doc holder() { return owner; }

    /** Accessor method to return the name of the tag. */
    public String name() { return name; }
    
    /** Accessor method to return the unparsed text of tag. */    
    public String text() { return text; }
    
    /** Makes a String representation of the tag. */
    public String toString() { return name() + ":" + text; } 

    /** This static method parses the text from a javadoc comment tag 
	to create the correct derived class instance of tag; the argument
	must begin with whitespace and an '@' symbol representing the
	start of the tag.
    */
    static MjTag create(Doc owner, String rawText, SourcePosition pos) {
	char[] carray = rawText.toCharArray();
	int i = 0;
	int len = carray.length;
	while (i<len && Character.isWhitespace(carray[i])) ++i;
	if (i==len) return null;
	int j = i; 
	while (i<len && !Character.isWhitespace(carray[i])) ++i;
	String name = new String(carray,j,i-j); 
	while (i<len && Character.isWhitespace(carray[i])) ++i;
	String text = new String(carray,i,len-i);

	if (name.equals(MjSeeTag.KIND)) return new MjSeeTag(owner,name,text.trim(),pos);
	else if (name.equals(MjParamTag.KIND)) return new MjParamTag(owner,text,pos);  
	else if (name.equals(MjThrowsTag.KIND)) return new MjThrowsTag(owner,name,text,pos); 
	else if (name.equals(MjThrowsTag.KIND2)) return new MjThrowsTag(owner,name,text,pos); 
	else return new MjGenericTag(owner,name,text,pos);
    }

    /** This helper method is used by some derived classes to convert raw, unparsed
	javadoc comment text into a sequence of tags (text and inline tags such as
	link tags), as must be assigned to the 'tags' protected variable and is
	returned by the inlineTags() method.
    */
    public Tag[] inlineTags(String description, SourcePosition pos) {
	//System.out.println("MAKING INLINE TAGS ##" + description + "##");
	int i = 0;
	ArrayList v = new ArrayList();
	int len = description.length();
	while (i<len && Character.isWhitespace(description.charAt(i))) ++i;
	int j = i;
	boolean done = false;
	while (!done) {
	    i = description.indexOf("{@link",i);
	    if (i==-1) { done = true; i = len; }
	    if (j != i) v.add(new MjTextTag(owner,description.substring(j,i),pos));
	    if (done) break;
	    j=i+6; // 6 for length of "{@link"
	    i = description.indexOf("}",i);
	    if (i==-1) { done = true; i = len; }
	    v.add(new MjSeeTag(owner,"@link",description.substring(j,i).trim(),pos));
	    j=i+1;
	}
	MjTag[] result = new MjTag[v.size()];
	v.toArray(result);
	
	return result;
    }
    
    // !FIXME! - resolveReference() probably not right for nested classes; does it reference imported classes?
    /** This is used by tags to resolve references to classes as they are given in tags
	(such as @see, @link, @throws, @exception tags).  Javadoc only recognizes classes that have
	already been referenced.  Mention in a tag does not cause the class to be parsed or loaded.
	@param name identifier or '.'-separated name of a class
    */
    public MjClassDoc resolveReference(String name) {
	
	// First try name as a fully-qualified name that is already referenced
	MjClassDoc mj = (MjClassDoc)(MjdocWrapper.root().classNamed(name));
	if (mj != null) return mj;

	// !FIXME! Need to look in super classes, super interfaces, imported packages,
	// or perhaps the name is a partially/fully qualified name of a nested class.
	
	int p = name.indexOf('.');
	String ident = (p == -1) ? name : name.substring(0,p);
	String suffix = (p==-1) ? "" : name.substring(p);
	
	MjClassDoc mjcd;
	if (owner instanceof MjClassDoc) {
	    mjcd = (MjClassDoc)owner;
	} else if (owner instanceof MjProgramElementDoc) {
	    mjcd = (MjClassDoc)(((MjProgramElementDoc)owner).containingClass());
	} else if (owner instanceof MjPackageDoc) {
	    // javadoc does not find classes in package.html files unless
	    // they are fully qualilfied - this will find classes in the
	    // same package - a convenient extension to javadoc
	    ClassDoc ccd = ((MjPackageDoc)owner).findClass(name);
	    return (MjClassDoc)ccd;
	} else {
	    System.err.println("Invalid owner class " + owner.getClass());
	    return null;
	}
	//System.out.println("IMPORTING CLASSES");
	ClassDoc[] cd = mjcd.importedClasses();
	for (int i=0; i<cd.length; ++i) {
	    //System.out.println("TRYING " + cd[i].name() + " " + ident + " " + name);
	    if (cd[i].name().equals(ident)) {
		//System.out.println("FOUND MATCH " + cd[i].qualifiedName());
		if (ident == name) return (MjClassDoc)cd[i];
		String qname = cd[i].qualifiedName() + suffix;
		    // Why isn't the following successful for nested classes when the enclosing class is already an MjClassDoc? - FIXME
		mj = (MjClassDoc)(MjdocWrapper.root().classNamed(qname));
		if (mj != null) return mj;
		CClass c = CTopLevel.loadClass(cd[i].qualifiedName().replace('.','/') + suffix.replace('.','$'));
		if (c != CClass.CLS_UNDEFINED) return MjdocWrapper.lookup(c);
		System.out.println("STILL CAN'T FIND " + qname); // FIXME
		return null; 
	    }
	}
	// Now try imported packages
	//System.out.println("IMPORTING PACKAGES " + mjcd.containingPackage().name());
	    CClass c = CTopLevel.loadClass(mjcd.containingPackage().name().replace('.','/') + "/" + ident);
	    if (c != CClass.CLS_UNDEFINED) {
		MjClassDoc mjc = MjdocWrapper.lookup(c);
		//System.out.println("FOUND MATCH " + mjc);
		if (name == ident) return mjc;
		String qname = mjc.qualifiedName() + name.substring(p);
		mj = (MjClassDoc)(MjdocWrapper.root().classNamed(qname));
		if (mj != null) return mj;
		c = CTopLevel.loadClass(mjc.qualifiedName().replace('.','/') + suffix.replace('.','$'));
		if (c != CClass.CLS_UNDEFINED) return MjdocWrapper.lookup(c);
		System.out.println("STILL CAN'T FIND " + qname); // FIXME
		return null; 
	    }
	    
	    c = CTopLevel.loadClass("java/lang/" + ident);
	    if (c != CClass.CLS_UNDEFINED) {
		MjClassDoc mjc = MjdocWrapper.lookup(c);
		//System.out.println("FOUND MATCH " + mjc);
		if (name == ident) return mjc;
		String qname = mjc.qualifiedName() + name.substring(p);
		mj = (MjClassDoc)(MjdocWrapper.root().classNamed(qname));
		if (mj != null) return mj;
		c = CTopLevel.loadClass(mjc.qualifiedName().replace('.','/') + suffix.replace('.','$'));
		if (c != CClass.CLS_UNDEFINED) return MjdocWrapper.lookup(c);
		System.out.println("STILL CAN'T FIND " + qname); // FIXME
		return null; 
	    }

	PackageDoc[] pd = mjcd.importedPackages();
	for (int i=0; i<pd.length; ++i) {
	    //System.out.println("TRYING " + pd[i].name() + " " + ident);
	    c = CTopLevel.loadClass(pd[i].name().replace('.','/') + "/" + ident);
	    if (c != CClass.CLS_UNDEFINED) {
		MjClassDoc mjc = MjdocWrapper.lookup(c);
		//System.out.println("FOUND MATCH " + mjc);
		if (name == ident) return mjc;
		String qname = mjc.qualifiedName() + name.substring(p);
		mj = (MjClassDoc)(MjdocWrapper.root().classNamed(qname));
		if (mj != null) return mj;
		c = CTopLevel.loadClass(mjc.qualifiedName().replace('.','/') + suffix.replace('.','$'));
		if (c != CClass.CLS_UNDEFINED) return MjdocWrapper.lookup(c);
		System.out.println("STILL CAN'T FIND " + qname); // FIXME
		return null; 
	    }
	}
	
	return mj;
    }

	// FIXME -- record source position
    public SourcePosition position() { return position; }
    
}
