/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjTagParser.java,v 1.1 2004/01/14 11:05:44 davidcok Exp $
 * Author: David R. Cok
 */

package org.multijava.mjdoc.mjdoc_142;

import com.sun.javadoc.*;
import java.text.BreakIterator;
import java.util.ArrayList;


/** This class is a helper class to parse javadoc comments into the appropriate set
    of tags.  A javadoc comment consists of some decriptive text followed by a tag section.
    The descriptive text is represented by a sequence of inline tags (typically
    Text and link tags).  The tag section is a sequence of tags, each of which may have 
    its own descriptive text (consisting of a sequence of Text and link and other inline tags).
    <P>
    This class is constructed with raw unprocessed text (consisting of the entire javadoc comment
    minus the optional white space and sequence of '*' characters that is ignored at the beginning
    of each line).  The constructor does the parsing, so after successful construction, all the
    accessor methods provide various views into the structure of the javadoc comment.
    @author David R. Cok
*/
public class MjTagParser {

    // The initial values below are set to give innocuous behavior for an 'empty' javadoc comment.
    
    /** A holder for the raw javadoc comment. */
    protected String rawComment = "";
    
    /** The description portion of the javadoc comment (but ordinarily use the parsed version
	as contained in inlineTags).
    */
    protected String commentText = "";
    
    /** The SourcePosition for the text being parsed; the current architecture does not 
	preserve source position within a javadoc comment.
    */
    protected SourcePosition sourcePosition;
    
    /** The parsed descriptive text of the first sentence of the javadoc comment, 
	represented as a sequence of Text and inline tags. 
    */
    protected Tag[] firstSentenceTags = new MjTextTag[0];
    
    /** The parsed descriptive text of the javadoc comment, represented as a sequence of Text and inline tags. */
    protected Tag[] inlineTags = new MjTextTag[0];
    
    /** The tag portion of the javadoc comment, represented as a sequence of tags. */
    protected Tag[] tags = new MjTextTag[0];

    /** The Doc object to which the tag being parsed belongs. */
    protected Doc owner;
    
    /** This constructor parses the input string into a descriptive portion (retrieved as a sequence
	of inline tags via the inlineTags() method), and a sequence of tags (retrieved via tags() ).
	The input text should be the complete javadoc comment text, without the opening and closing
	comment markers or any ignored whitespace+'*' at the beginning of a line.
    */
    //@ requires s != null;
    public MjTagParser(Doc owner, String s, SourcePosition pos) { 
	this.owner = owner;
	this.sourcePosition = pos;
	int len = s.length();
	int i = 0;
	while (i<len && Character.isWhitespace(s.charAt(i))) ++i;
	rawComment = i<len? s.substring(i) : s; // This logic is just to match javadoc on comments that are all whitespace
	parse(); 
    }

    /** A constructor for an empty javadoc comment. */
    public MjTagParser(Doc owner) { this.owner = owner; }

    /** The helper method that does the parsing of the raw comment text. */
    protected void parse() {
	char[] c = rawComment.toCharArray();
	int len = c.length;
	int i = skipToFirstTag(c);
	commentText = (new String(c,0,i)).trim();
	if (commentText.length() > 0) inlineTags = (new MjGenericTag(owner,"TEMP",commentText,sourcePosition)).inlineTags();
	String first = firstSentence(commentText);
	if (first.length() > 0) firstSentenceTags = (new MjGenericTag(owner,"TEMP",first,sourcePosition)).inlineTags();
	ArrayList v = new ArrayList();
	while (i<len) {
	    int j = skipToTag(c,i);
	    MjTag t = MjTag.create(owner,new String(c,i,j-i),sourcePosition);
	    if (t != null) v.add(t);
	    i = j;
	}
	tags = new MjTag[v.size()];
	v.toArray(tags);
	
    }

    //!FIXME! - document skipToFirstTag
    protected int skipToFirstTag(char[] c) {
	int j = 0;
	int len = c.length;
	int i = 0;
	while (i < len) {
	    while (i<len && Character.isWhitespace(c[i])) ++i; // Whitespace includes newline characters
	    if (i==len || c[i] == '@') return j;
	    while (i<len && c[i] != '\n' && c[i] != '\r') ++i;
	    j = i; // Record a tentative end before a new line
		    // If this is followed by whitespace and an '@', we say this is the end of the tag
		    // otherwise we include the next line in the tag (except if it is the end of the string)
	}
	return len;
    }
    /** A helper function that finds the end of a tag or the initial descriptive portion
	of a javadoc comment.
	@param c a char array that contains the text of the javadoc comment
	@param j the index representing the end of the previous tag or descriptive section (or 0)
	@return the end of the next tag or descriptive section.
    */ // FIXME - better description needed
    protected int skipToTag(char[] c, int j) { // j is at end of previous tag
	int len = c.length;
	int i = j;
	while (i<len && Character.isWhitespace(c[i])) ++i; // Whitespace includes newline characters
	while (i<len) {
	    while (i<len && c[i] != '\n' && c[i] != '\r') ++i;
	    j = i; // Record a tentative end after a new line
		    // If this is followed by whitespace and an '@', we say this is the end of the tag
		    // otherwise we include the next line in the tag (except if it is the end of the string)
	    while (i<len && Character.isWhitespace(c[i])) ++i; // Whitespace includes newline characters
	    if (i==len || c[i] == '@') return j;
	}
	return len;
    }

    /** Returns the summary sentence of the comment text.  The argument is the comment text;
	the argument may have in-line tags, but not non-inline tags.
    */
    //@ requires comment != null;
    //@ ensures \result != null;
    protected String firstSentence(String comment) {
	BreakIterator boundary = BreakIterator.getSentenceInstance(); // Uses default Locale
	boundary.setText(comment);
	int start = boundary.first();
	int e = boundary.next();
	if (e == -1) e = comment.length();
	return comment.substring(start,e).trim();
    }

    /** An accessor providing the unparsed text of the descriptive section of the javadoc comment. */
    //@ ensures \result != null;
    //@ pure
    public String commentText() { return commentText; } 
    
    /** An accessor providing the summary sentence of the javadoc comment as a sequence of 
	Text and inline tags. 
    */
    //@ ensures \result != null;
    //@ pure
    public Tag[] firstSentenceTags() { return firstSentenceTags; } 
    
    /** An accessor providing the unparsed text of the full javadoc comment, as supplied to the constructor. */
    //@ ensures \result != null;
    //@ pure
    public String getRawCommentText() { return rawComment; } 
    
    /** An accessor providing the parsed version of the descriptive section of the javadoc comment,
	as a sequence of Text and inline tags. 
    */
    //@ ensures \result != null;
    //@ pure
    public Tag[] inlineTags() { return inlineTags; } 

    /** An accessor providing the subset of the tag section of the javadoc comment that is the
	'@see' tags. 
    */
    //@ ensures \result != null;
    //@ pure
    public SeeTag[] seeTags() { 
	return (SeeTag[])tags(MjSeeTag.KIND,new MjSeeTag[0]); 
    }
    
    /** An accessor providing the tags making up the tag section of the javadoc comment. 
    */
    //@ ensures \result != null;
    //@ pure
    public Tag[] tags() { return tags; } 
    
    /** An accessor providing the tags making up the tag section of the javadoc comment. 
    */
    //@ ensures \result != null;
    //@ pure
    public Tag[] tags(String tagkind) { 
	return tags(tagkind,new MjTag[0]); 
    }
	
    /** An helper method providing the tags making up the tag section of the javadoc comment that
	are of the given 'kind'. 
    */
    //@ ensures \result != null;
    //@ pure
    public Tag[] tags(String tagkind, Tag[] type) { 
	if (tagkind.charAt(0) != '@') tagkind = "@" + tagkind;
	ArrayList v = new ArrayList();
	for (int i=0; i<tags.length; ++i) {
	    if (tags[i].kind().equals(tagkind)) v.add(tags[i]);
	}
	// The type argument is needed to get the result array produced with the correct
	// dynamic type.
	Tag[] result = (Tag[])(v.toArray(type));
	
	return result;
    }
	


}
