

/*
 * @(#)PackageFrameWriter.java	1.19 00/02/02
 *
 * Copyright 1998-2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * This software is the proprietary information of Sun Microsystems, Inc.  
 * Use is subject to license terms.
 * 
 */

package org.multijava.mjdoc.mjdoc_142;

import com.sun.tools.doclets.standard.*;

import com.sun.tools.doclets.*;
import com.sun.javadoc.*;
import java.io.*;
import java.lang.*;

/**
 * Class to generate file for each package contents in the left-hand bottom 
 * frame. This will list all the Class Kinds in the package. A click on any
 * class-kind will update the right-hand frame with the clicked class-kind page.
 *
 * @author Atul M Dambalkar
 */
public class MjdocPackageFrameWriter extends PackageFrameWriter {

    protected PackageDoc packagedoc;
    
    /**
     * Constructor to construct PackageFrameWriter object and to generate 
     * "package-frame.html" file in the respective package directory. 
     * For example for package "java.lang" this will generate file 
     * "package-frame.html" file in the "java/lang" directory. It will also 
     * create "java/lang" directory in the current or the destination directory
     * if it doesen't exist.
     *
     * @param path Directories in this path will be created if they are not 
     * already there.
     * @param filename Name of the package summary file to be generated, 
     * "package-frame.html".
     * @param packagedoc PackageDoc under consideration.
     */
    public MjdocPackageFrameWriter(ConfigurationStandard configuration, String path, String filename, 
                              PackageDoc packagedoc) 
                              throws IOException, DocletAbortException {
        super(configuration, path, filename, packagedoc);
	this.packagedoc = packagedoc;
    }

    /**
     * Generate a package summary page for the left-hand bottom frame. Construct
     * the PackageFrameWriter object and then uses it generate the file.
     *
     * @param pkg The package for which "pacakge-frame.html" is to be generated.
     */
    public static void generate(PackageDoc pkg) throws DocletAbortException {
        MjdocPackageFrameWriter packgen;
        String path = DirectoryManager.getDirectoryPath(pkg);
        String filename = "package-frame" + ".html";
        try {
            packgen = new MjdocPackageFrameWriter(Standard.configuration(), path, filename, pkg);
            packgen.generatePackageFile(true); // CHANGED FOR 1.4.1_03
            packgen.close();
        } catch (IOException exc) {
            Standard.configuration().standardmessage.
                    error("doclet.exception_encountered", 
                           exc.toString(), filename);
            throw new DocletAbortException();
        }
    }

    /**
     * Generate class listing for all the classes in this package. Divide class
     * listing as per the class kind and generate separate listing for 
     * Classes, Interfaces, Exceptions and Errors.
     */
    protected void generateClassListing() {
        generateClassKindListing(packagedoc.interfaces(), 
                                 getText("doclet.Interfaces"));
        generateClassKindListing(packagedoc.ordinaryClasses(),
                                 getText("doclet.Classes"));
        generateClassKindListing(packagedoc.exceptions(),
                                 getText("doclet.Exceptions"));
        generateClassKindListing(packagedoc.errors(),
                                 getText("doclet.Errors"));
	generateGFListing("Generic Functions");
    }

    /** 
     * Generate specific class kind listing. Also add label to the listing.
     *
     * @param label Label for the listing
     */
    protected void generateGFListing(String label) {
	java.util.Iterator i = 
	    MjdocWrapper.root().topConcreteMethods(packagedoc).iterator();
	if (i.hasNext()) {
            printPackageTableHeader();
            fontSizeStyle("+1", "FrameHeadingFont");
            print(label);
            fontEnd();
            println("&nbsp;");
            fontStyle("FrameItemFont");
            while (i.hasNext()) {
		MjMethodDoc m = (MjMethodDoc)(i.next());
                //if (!isGeneratedDoc(m)) {
                //    continue;
                //}
                br();
		printTargetHyperLink(pathString(m.cuPackage(), m.gfFileName()),
				"classFrame", m.name()+m.signature());
            }
	    br();
            fontEnd();
            printPackageTableFooter();
            println();
        }
    }

    /**
     * Print the html file header. Also print Html page title and stylesheet
     * default properties.  Only overridden to change 'javadoc' to 'mjdoc'.
     * 
     * @param title String title for the generated html file.
     */ // FIXME Unify with other cases of this method???
    public void printHeader(String title) {
        println("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Frameset//EN\"" +
                "\"http://www.w3.org/TR/REC-html40/frameset.dtd\">");    
        println("<!--NewPage-->");
        html();
        head();
        print("<!-- Generated by " + MjdocWrapper.appName() + " on ");
        print(today());
        println(" -->");
        if (Standard.configuration().charset.length() > 0) {
            println("<META http-equiv=\"Content-Type\" content=\"text/html; " 
                    + "charset=" + Standard.configuration().charset + "\">");
        }
        title();
        println(title);
        titleEnd();
        printStyleSheetProperties();
        headEnd();
        body("white",true); // CHANGED FOR 1.4.1_03
    }
}



