package org.multijava.relaxed.runtime;

import org.multijava.relaxed.util.*;

import org.multijava.util.DirectedAcyclicGraph;
import org.multijava.mjc.Debug;
import org.multijava.mjc.Constants;

import java.io.*;
import java.util.*;
import java.net.URL;
import java.lang.reflect.*;

public class RMJClassLoader extends ClassLoader implements Constants
{
    public static void main(String args[])
	throws ClassNotFoundException, NoSuchMethodException, 
	       IllegalAccessException, InvocationTargetException
    {
	// Create the special RMJ-aware class loader
	RMJClassLoader loader = new RMJClassLoader();
	
	// run its main function
	loader.loaderMain(args);
    }

    // [make this a non-static method, so we can override key pieces
    // in subclasses]
    public void loaderMain(String args[])
	throws ClassNotFoundException, NoSuchMethodException, 
	       IllegalAccessException, InvocationTargetException
    {
	// Make sure we have at least the class to load
	if (args.length < 1) {
	    System.err.println(usage());
	    System.exit(1);
	}

	    // Register all the glue methods listed in the
	    // rmj.glue property
	if (!RMJOptions.nonothing) {
	    registerAllGlue();
	}

	// Load the specified main class
	Class cls;
	try {
	    cls = loadClass(args[0]);
	} catch (ClassNotFoundException e) {
	    System.err.println("Class " + args[0] + " not found");
	    throw e;
	}
	
	try {
	    runProgram(cls, args);
	} finally {
	    if (RMJDebug.classreport) {
		// Print out a class report
		printClassReport(System.out);
	    }
	}
    }

    protected String usage() 
    {
	return "usage: " +
	    "java org.multijava.relaxed.runtime.RMJClassLoader class [args]";
    }

    protected void registerAllGlue()
	throws ClassNotFoundException {
	String glue = System.getProperty("rmj.glue");
	if (!(glue == null || glue.equals(""))) {
	    String pathsep = System.getProperty("path.separator");
	    String[] glueClasses = glue.split(pathsep, 0);
	    for (int i = 0; i < glueClasses.length; i++) {
		registerGlueAnchor(glueClasses[i] + MJ_ANCHOR);
	    }
	}
    }
    
    protected void runProgram(Class cls, String[] args) 
	throws ClassNotFoundException, NoSuchMethodException, 
	       IllegalAccessException, InvocationTargetException
    {
	// Reconstruct the argument list without the loaded class
	String mainArgs[] = new String[args.length - 1];
	System.arraycopy(args,1,mainArgs,0,args.length - 1);
	Object mainArgArray[] = { mainArgs };
	Class mainArgType[] = { mainArgs.getClass() };

	// Get root class's main method
	Method mainMethod = null;
	try {
	    mainMethod = cls.getMethod("main", mainArgType);
	} catch (NoSuchMethodException e) {
	    System.err.println("Class " + args[0] + " has no main method");
	    throw e;
	}

	// Invoke "main", i.e., run the RMJ program
	try {
	    mainMethod.invoke(null, mainArgArray);
	} catch (IllegalAccessException e) {
	    System.err.println("Main method of class " + args[0] +
			       " cannot be accessed");
	    throw e;
	} catch (InvocationTargetException e) {
	    // the callee method exited with an exception
	    System.err.println(e.getTargetException());
	    e.printStackTrace();
	    throw e;
	}
    }


    ///////////////////////////////////////////////////////////////////////

    public RMJClassLoader()
    {
	if (RMJDebug.verbose) {
	    RMJDebug.out("RMJClassLoader constructed");
	}
    }
    
    public RMJClassLoader(ClassLoader parent)
    {
	super(parent);
	if (RMJDebug.verbose) {
	    RMJDebug.out("RMJClassLoader constructed with parent " + parent);
	}
    }


    ///////////////////////////////////////////////////////////////////////

    public Class loadClass(String name)
	throws ClassNotFoundException
    {
	if (RMJDebug.loadtrace || RMJDebug.verbose) {
	    if (RMJDebug.verbose) {
		RMJDebug.out("");
	    }
	    RMJDebug.indent();
	    RMJDebug.out("Loading " + name);
	}
	    
	try {
	    if (name.startsWith("java.")) {
		// such classes are in privileged packages, so they
		// must be system classes
	    } else {
		try {
		    Class cls = loadUserClass(name);
		    // Success!
		    if (RMJDebug.loadtrace || 
			RMJDebug.verbose) {
			RMJDebug.out("Completed loading user class " + name + 
				     " (loader=" + cls.getClassLoader() + ")");
		    }
		    return cls;

		} catch (SecurityException e) {
		    if (RMJDebug.loadtrace || 
			RMJDebug.verbose) {
			RMJDebug.out("Exception loading " + name + 
				     " as user class\n    (" + e + ")");
		    }
		    // try to load as a system class
		}
	    }
	    
	    Class cls = loadSystemClass(name);
	    // Success!
	    if (RMJDebug.loadtrace || RMJDebug.verbose) {
		RMJDebug.out("Completed loading system class " + name + 
			     " (loader=" + cls.getClassLoader() + ")");
	    }
	    return cls;

	} catch (ClassNotFoundException e) {
	    if (RMJDebug.loadtrace || RMJDebug.verbose) {
		RMJDebug.out("Unable to load " + name + "\n    (" + e + ")");
	    }
	    throw e;
	} catch (Error e) {
	    if (RMJDebug.loadtrace || RMJDebug.verbose) {
		RMJDebug.out("Unable to load " + name + "\n    (" + e + ")");
	    }
	    throw e;
	} finally {
	    if (RMJDebug.loadtrace || RMJDebug.verbose) {
		RMJDebug.undent();
		if (RMJDebug.verbose) {
		    RMJDebug.out("");
		}
	    }
	}
    }

    protected Class loadUserClass(String name)
	throws ClassNotFoundException
    {
	byte[] classBytes = readClassByteCode(name);
	    
	if (RMJDebug.classdump) {
	    Readdump.dump(classBytes);
	}

	Class cls = defineClass(name, classBytes, 0, classBytes.length);
	
	if (RMJDebug.verbose) {
	    RMJDebug.out("Loaded " + cls.getName() + " (loader=" + 
			 cls.getClassLoader() + ")");
	}

	if(!RMJOptions.nonothing) {
		// see if this is a glue anchor class that we've
		// already scanned
	    RMJAnnotation anno = 
		(RMJAnnotation) glueAnchorClassAnnotationsMap.get(name);
	    if (anno == null) {
		    // first time reading this class
		anno = new RMJAnnotation(classBytes);
	    } else {
		    // we've already read this one
		if (RMJDebug.verbose) {
		    RMJDebug.out("Reusing annotations of glue anchor " + name);
		}
		    // clear out the data, since we're not going to load this
		    // class again
		glueAnchorClassAnnotationsMap.remove(name);
	    }
	    if (!anno.isEmpty()) {
		if (RMJDebug.ann) {
		    RMJDebug.out("Annotations of " + name + ": " + anno);
		}
	    }
	
	    // Do what we need to do to process the newly loaded class
	    processLoadedClass(cls, anno);
	}
	
	return cls;
    }

    public byte[] readClassByteCode(String name)
	throws ClassNotFoundException
    {
	try {
	    String path = name.replace('.', '/') + ".class";
	    if (RMJDebug.verbose) {
		RMJDebug.out("Path = " + path);
	    }
	    
	    URL url = getResource(path);
	    if (RMJDebug.verbose) {
		RMJDebug.out("URL = " + url);
	    }

	    if (url == null) {
		// we didn't find the file
		if (RMJDebug.verbose) {
		    RMJDebug.out("Error: readClassByteCode didn't find " + 
				 path);
		}
		throw new ClassNotFoundException(name);
	    }
	    
	    InputStream strm = url.openStream();

	    byte[] classBytes = readAll(strm);

	    return classBytes;

	} catch (IOException e) {
	    // We could not find the class, so indicate the problem
	    // with an exception
	    if (RMJDebug.exn) {
		RMJDebug.out("Error: readClassByteCode: caught exception " + 
			     e);
		e.printStackTrace();
	    }
	    throw new ClassNotFoundException(name);
	}
    }
   
    protected byte[] readAll(InputStream strm)
	throws IOException
    {
	// Just using an InputStream doesn't seem to work on Mac OS.
	// Apparently the native implementation of BufferedInputStream
	// handles things correctly.
	BufferedInputStream bin = new BufferedInputStream(strm);
	int numBytes = bin.available();
	byte[] classBytes = new byte[numBytes];
	int start = 0;
	int numRead;
	while ((numRead = bin.read(classBytes, start, numBytes)) > 0) {
	    start += numRead;
	    numBytes -= numRead;
	    if (numBytes == 0) {
		// Used all the allocated array space, but, since the
		// loop condition was true, there must be more to
		// read.  We try allocating for just getting the
		// number of bytes available.  (In profiling under Mac
		// OS, I found that a strategy of allocating a new
		// array of double the current size resulted in more
		// array allocations and copies, 0.38 additional per
		// class file, than the current strategy.)
		numBytes = bin.available();
		if (numBytes == 0) {
		    break;
		}
		byte[] moreBytes = new byte[start + numBytes];
		System.arraycopy(classBytes,0,moreBytes,0,start);
		classBytes = moreBytes;
	    }
	}

	if (start < classBytes.length) {
	    // Allocated too much space, so truncate array
	    byte[] fewerBytes = new byte[start];
	    System.arraycopy(classBytes,0,fewerBytes,0,start);
	    classBytes = fewerBytes;
	}
	
	return classBytes;
    }

    protected Class loadSystemClass(String name)
	throws ClassNotFoundException
    {
	Class cls = findSystemClass(name);
	if (RMJDebug.verbose) {
	    RMJDebug.out("Loaded system class " + cls.getName() + 
			 " (loader=" + cls.getClassLoader() + ")");
	}

	if (!RMJOptions.nonothing) {
	    registerSystemLoadAncestors(cls);

		// Do what we need to do to process the newly loaded
		// class (we know it has no annotations in it)
	    processLoadedClass(cls, null);
	}
	
	return cls;
    }

    protected void registerSystemLoadAncestors(Class cls)
	throws ClassNotFoundException
    {
	// for system classes, we need to visit all their ancestor
	// classes & interfaces, because otherwise we might not know
	// they they've been loaded, if they're not referenced
	// explicitly by the RMJ program
	Class intf[] = cls.getInterfaces();
	for (int i = 0; i < intf.length; i++) {
	    Class in = intf[i];
	    registerSystemLoad(in);
	}
	Class scls = cls.getSuperclass();
	if (scls != null) {
	    registerSystemLoad(scls);
	}
    }

    protected void registerSystemLoad(Class cls)
	throws ClassNotFoundException
    {
	if (! isLoaded(cls.getName())) {
	    // this class hasn't been noticed by RMJ yet

	    // first, recursively process its ancestors
	    registerSystemLoadAncestors(cls);

	    // then process this class
	    if (RMJDebug.verbose) {
		RMJDebug.out("Loading system class ancestor " + cls.getName());
	    }
	    processLoadedClass(cls, null);
	}
    }

    // process a newly loaded class
    protected void processLoadedClass(Class cls, RMJAnnotation anno)
	throws ClassNotFoundException
    {
	// register this newly loaded class
	registerClass(cls, anno);

	// load any newly reachable methods, and verify the
	// non-ambiguity of glue methods vs. other non-base methods
	// (and of interface methods)
	verifyNonAmbiguityAndLoadExternalMethods();

	if (!RMJOptions.nocheck) {
		// verify the completeness of the resulting system
	    verifyCompleteness(cls, anno);
	}
    }


    // register each glue anchor listed in the rmj.glue property
    protected void registerGlueAnchor(String glueAnchorClassName)
	throws ClassNotFoundException
    {
	if (glueAnchorClassAnnotationsMap.get(glueAnchorClassName) != null) {
	    // we've already registered this glue anchor class, so
	    // ignore this redundant glue include directive
	    if (RMJDebug.verbose) {
		RMJDebug.out("Ignoring redundant include of " +
			     "glue anchor class " + glueAnchorClassName);
	    }
	    return;
	}

	RMJAnnotation anno = readGlueAnchorAnnotations(glueAnchorClassName);

	// process each of the method annotations in this file
	if (anno != null) {
	    Iterator iter = anno.methodItems().iterator();
	    while (iter.hasNext()) {
		RMJAnnotation.Method methodAnno = 
		    (RMJAnnotation.Method) iter.next();

		// record that this signature is a glue signature, not
		// a subclass signature
		methodAnno.foundInGlueAnchor();

		// process the signature
		registerGlueSignature(methodAnno);

		// remove this annotation, since we've processed it,
		// and don't want to process it again when we load
		// this anchor class for real
		iter.remove();
	    }
	}

	// remember that we've already processed the method
	// annotations of this glue anchor class, so that we don't do
	// it again.  also avoid reading the annotations again, when
	// we look for any glue annotations.
	glueAnchorClassAnnotationsMap.put(glueAnchorClassName, anno);
    }

    protected RMJAnnotation readGlueAnchorAnnotations(String glueAnchorClassName)
	throws ClassNotFoundException
    {
	byte[] classBytes = readClassByteCode(glueAnchorClassName);
	if (RMJDebug.classdump) {
	    Readdump.dump(classBytes);
	}
	RMJAnnotation anno = new RMJAnnotation(classBytes);
	if (!anno.isEmpty()) {
	    if (RMJDebug.ann) {
		RMJDebug.out("Annotations of " + glueAnchorClassName +
			     ": " + anno);
	    }
	}
	return anno;
    }

    // do this for each method annotation found in the glue's anchor class
    protected void registerGlueSignature(RMJAnnotation.Method anno)
    {
	Set unloadedClassNames = RMJSignature.unloadedClassNames(anno, this);
	if (unloadedClassNames.isEmpty()) {
	    throw new RuntimeException("Registering a glue method that " +
				       "does not depend on any " +
				       "unloaded classes");
	} else {
	    registerUnreachableSignature(anno, unloadedClassNames);
	}
    }

    protected void registerUnreachableSignature(RMJAnnotation.Method anno,
						Set unloadedClassNames)
    {
	RMJUnreachableSignature sig =
	    new RMJUnreachableSignature(anno, unloadedClassNames);

	if (RMJDebug.verbose) {
	    RMJDebug.out("Registering unreachable signature " + sig + 
			 "; depends on " + unloadedClassNames);
	}

	// for each unloaded class name, put this unreachable signature
	// descriptor on its list
	Iterator iter = unloadedClassNames.iterator();
	while (iter.hasNext()) {
	    String unloadedClassName = (String) iter.next();
	    List unreachableSigs = 
		(List) unreachableSigsMap.get(unloadedClassName);
	    if (unreachableSigs == null) {
		// the first one for this class name
		unreachableSigs = new LinkedList();
		unreachableSigsMap.put(unloadedClassName, unreachableSigs);
	    }
	    unreachableSigs.add(sig);
	}
    }


    ///////////////////////////////////////////////////////////////////////

    // register a newly loaded class
    protected void registerClass(Class cls, RMJAnnotation anno)
    {
	// register it
	loadedClassesAndInterfaces.put(cls.getName(), cls);

	if (!RMJOptions.nocheck) {
	    if (isAbstractClassOrInterface(cls)) {
		    // nothing to do for an abstract class or interface
		if (RMJDebug.verbose) {
		    RMJDebug.out("No registration needed for " +
				 "abstract class or interface " +
				 cls.getName());
		}
		
	    } else {
		if (RMJDebug.verbose) {
		    RMJDebug.out("Registering concrete class " + 
				 cls.getName() +
				 " with any abstract ancestors");
		}

	    
		    // a concrete class needs to be registered on all its
		    // abstract superclasses & -interfaces, if any
		registerWithAbstractAncestors(cls);
	    }
	}

	// also register any abstract or top concrete method
	// signatures introduced in this class, if it's an anchor
	// class
	if (anno != null) {
	    Iterator iter = anno.methodItems().iterator();
	    while (iter.hasNext()) {
		RMJAnnotation.Method methodAnno = 
		    (RMJAnnotation.Method) iter.next();
		registerSignature(methodAnno);
	    }
	}

	// also see if any unreachable signatures have become reachable
	String className = cls.getName();
	List unreachableSigs = (List) unreachableSigsMap.get(className);
	if (unreachableSigs != null) {
	    Iterator iter = unreachableSigs.iterator();
	    while (iter.hasNext()) {
		RMJUnreachableSignature unsig =
		    (RMJUnreachableSignature) iter.next();
		unsig.registerLoadedClass(className);
		if (unsig.isReachable()) {
		    // the signature is now reachable.  promote it to
		    // a real signature, register it
		    if (RMJDebug.verbose) {
			RMJDebug.out("Unreachable sig " + unsig + 
				     " has become reachable");
		    }
		    registerReachableSignature(unsig.annotation());
		}
	    }
	    // forget this list, since it's no longer needed
	    unreachableSigsMap.remove(className);
	}

    }

    // add a concrete class to the lists of top concrete subclasses of
    // all its abstract superclasses and interfaces, transitively
    protected void registerWithAbstractAncestors(Class cls)
    {
	registerWithAbstractAncestorsOf(cls, cls, new LinkedList());
    }

    protected void registerWithAbstractAncestorsOf(Class cls, Class of,
						   List alreadyVisited)
    {
	alreadyVisited.add(of);
	Class intf[] = of.getInterfaces();
	for (int i = 0; i < intf.length; i++) {
	    Class in = intf[i];
	    if (!alreadyVisited.contains(in)) {
		alreadyVisited.add(in);
		registerWithAbstractClassOrInterface(cls, in);
		    // recursively add cls to i	n's superinterfaces's lists
		registerWithAbstractAncestorsOf(cls, in, alreadyVisited);
	    }
	}
	Class scls = of.getSuperclass();
	if (scls != null && isAbstractClassOrInterface(scls) &&
	    !alreadyVisited.contains(scls)) {
	    alreadyVisited.add(scls);
	    // it's an abstract superclass
	    registerWithAbstractClassOrInterface(cls, scls);
		// recursively add cls to scls's
		// superclass's/superinterfaces's lists
	    registerWithAbstractAncestorsOf(cls, scls, alreadyVisited);
	}
    }

    // register a concrete class (cls) on the list of concrete
    // subclasses of an abstract class or interface (abs) and its
    // transitive abstract ancestors
    protected void registerWithAbstractClassOrInterface(Class cls, Class abs)
    {
	if (RMJDebug.verbose) {
	    RMJDebug.out("Registering top concrete class " + 
			 cls.getName() + " with abstract ancestor " + 
			 abs.getName());
	}

	List topConcreteSubclasses = (List) topConcreteSubclassesMap.get(abs);

	if (topConcreteSubclasses == null) {
	    if (! isLoaded(abs.getName())) {
		// this abstract class or interface hasn't been loaded yet!
		throw new RuntimeException("Encountered unloaded abstract " +
					   "superclass/interface " +
					   abs.getName());
	    }

	    // this is the first time we've loaded a top concrete
	    // subclass of this abstract class => create its list
	    topConcreteSubclasses = new LinkedList();
	    topConcreteSubclassesMap.put(abs, topConcreteSubclasses);
	}

	topConcreteSubclasses.add(cls);

	if (isInterface(abs)) { 
	    List interfaceSpecializedSignatures =
		(List) interfaceSpecializedSignaturesMap.get(abs);
	    if (interfaceSpecializedSignatures != null) {
		// add all these signatures on a temporary list of
                // sigs to check again for non-ambiguity (XXX ideally,
 		// we'd only put something on a list if we introduced
 		// MI, but we can have abstract classes introduce MI,
 		// and then their singly-inheriting concrete
 		// subclasses would need to be considered here.
 		// perhaps the right answer is to identify which top
 		// concrete subclasses have any MI ancestors, and
 		// which interfaces are above the MI point, and only
 		// revisit signatures of those.)
		Iterator iter = interfaceSpecializedSignatures.iterator();
		while (iter.hasNext()) {
		    RMJSignature sig = (RMJSignature) iter.next();
		    if (RMJDebug.verbose) {
			RMJDebug.out("  Will recheck non-ambiguity of " + sig);
		    }
		    sigsRequiringAmbiguityChecking.add(sig);
		}
	    }
	}

    }

    protected void registerSignature(RMJAnnotation.Method anno)
    {
	Set unloadedClassNames = RMJSignature.unloadedClassNames(anno, this);
	if (! unloadedClassNames.isEmpty()) {
	    // we have to save this annotation until it becomes
	    // reachable before we consider it (a lot like a glue
	    // method!)
	    registerUnreachableSignature(anno, unloadedClassNames);

	} else {
	    // a reachable method.  register it on the right tables
	    // for later loading and completeness checking.
	    registerReachableSignature(anno);
	}
    }

    protected void registerReachableSignature(RMJAnnotation.Method anno)
    {
	RMJSignature sig = new RMJSignature(anno, this);

	if (!RMJOptions.nocheck) {
	    RMJOperation op = sig.operation();

		// look up the previously loaded sigs for this operation
	    List sigs = (List) signaturesMap.get(op);
	    if (sigs == null) {
		    // first time we've loaded an operation like this;
		    // create the list
		sigs = new LinkedList();
		signaturesMap.put(op, sigs);
	    }

		// do nothing if we've already seen this sig before.
		// that can happen because classes with interface specs
		// include annotations for methods of superclasses.  therefore,
		// multiple classes can include an annotation for the same
		// method declaration.
	    if (sigs.contains(sig)) {
		if (RMJDebug.verbose) {
		    RMJDebug.out("Skipping duplicate sig " + sig);
		}
		return;
	    }

	    if (RMJDebug.verbose) {
		RMJDebug.out("Registering reachable sig " + sig);
	    }
	    
	    if (sig.isAbstract()) {
		    // an abstract method signature
		    // register it with all its abstract class &
		    // interface specializers, so that it will get
		    // rechecked whenever a top concrete subclass of them
		    // is loaded
		registerAbstractSignature(sig);

		    // in case we're registering this signature after one
		    // of its concrete implementations, we search the list
		    // appropriately.
		Iterator iter = sigs.iterator();
		while (iter.hasNext()) {
		    RMJSignature s = (RMJSignature) iter.next();
		    if (s.isConcrete() && s.overrides(sig)) {
			registerConcreteSig(s, sig);
		    }
		}

		    // remember that we need to do completeness checking
		    // for this signature
		newlyLoadedAbstractSignatures.add(sig);
		
	    } else {
		    // a concrete method signature

		    // if it overrides some previously loaded abstract
		    // method signature, which is not already covered by
		    // some existing method signature, then this is a top
		    // concrete method signature, and we need to record it
		    // on the abstract method signature's list.
		Iterator iter = sigs.iterator();
		while (iter.hasNext()) {
		    RMJSignature s = (RMJSignature) iter.next();
		    if (s.isAbstract() && sig.overrides(s)) {
			registerConcreteSig(sig, s);
		    }
		}
	    }

		// finally, actually register the new signature
	    sigs.add(sig);
	}
	
	// remember that this signature needs to be loaded
	newlyReachableSigs.add(sig);
    }

    // record this abstract signature with each of its abstract
    // argument classes & interfaces
    protected void registerAbstractSignature(RMJSignature asig)
    {
	Object[] specs = asig.specializersOrTypes();
	for (int i = 0; i < specs.length; i++) {
	    Object spec = specs[i];
	    if (spec == null) {
		// a prim type; not abstract
	    } else if (spec instanceof Class) {
		// a class type
		Class cls = (Class) spec;
		if (isAbstractClassOrInterface(cls)) {
		    // register the abstract sig on the specializer's list
		    if (RMJDebug.verbose) {
			RMJDebug.out("Adding abstract sig " + asig +
				     " to list of " + cls.getName());
		    }
		    List sigsRequiringCompletenessChecking =
			(List) sigsRequiringCompletenessCheckingMap.get(cls);
		    if (sigsRequiringCompletenessChecking == null) {
			// first time; create the list
			sigsRequiringCompletenessChecking = new LinkedList();
			sigsRequiringCompletenessCheckingMap
			    .put(cls, sigsRequiringCompletenessChecking);
		    }
		    sigsRequiringCompletenessChecking.add(asig);
		}
	    } else {
		// a value specializer; not abstract.
	    }
	}
    }

    // record that we've now seen a concrete sig and an abstract sig,
    // and the first overrides the second
    protected void registerConcreteSig(RMJSignature csig, RMJSignature asig)
    {
	List csigs = (List) sigImplementationsMap.get(asig);
	if (csigs == null) {
	    // the first time we've seen this abstract signature;
	    // create the list of concrete implementations
	    csigs = new LinkedList();
	    sigImplementationsMap.put(asig, csigs);

	} else {
	    // sig might be a new top concrete signature.  check to
	    // see whether it overrides some existing concrete
	    // signature associated with this abstract sig, in which
	    // case it's not the top
	    Iterator iter = csigs.iterator();
	    while (iter.hasNext()) {
		RMJSignature csig2 = (RMJSignature) iter.next();
		if (csig.overrides(csig2)) {
		    // there's already a signature that's there that
		    // is the top concrete one; ignore this one
		    if (RMJDebug.verbose) {
			RMJDebug.out("Concrete sig " + csig +
				     " dominated by " + csig2 + "; ignoring");
		    }
		    return;

		} else if (csig2.overrides(csig)) {
		    // this one is more top than the existing one, so
		    // remove the existing one (the new one will replace it)
		    if (RMJDebug.verbose) {
			RMJDebug.out("Concrete sig " + csig +
				     " dominates " + csig2 + "; replacing");
		    }
		    iter.remove();
		}
	    }
	}

	// this concrete sig is a new top concrete implementation of
	// this abstract sig, so add it to the list
	if (RMJDebug.verbose) {
	    RMJDebug.out("Registering top concrete sig " + csig +
			 " of abstract sig " + asig);
	}
	csigs.add(csig);
    }


    ///////////////////////////////////////////////////////////////////////

    // verify the non-ambiguity and then load any newly reachable
    // methods
    protected void verifyNonAmbiguityAndLoadExternalMethods()
	throws ClassNotFoundException
    {
	List toBeLoaded = new LinkedList();

	if (!RMJOptions.nocheck) {
		// first iterate through any previously loaded signatures that
		// may have become ambiguous due to later class loading
	    Iterator iter = sigsRequiringAmbiguityChecking.iterator();
	    while (iter.hasNext()) {
		RMJSignature sig = (RMJSignature) iter.next();
		
		    // remove this from the list of unchecked sigs
		iter.remove();
		
		    // ensure that this signature isn't ambiguous with any
		    // previously loaded signature, or is resolved by a
		    // signature not yet checked
		verifyNonAmbiguityOfOldSig(sig, toBeLoaded);
	    }
		// clear out the list so we don't recheck again unnecessarily
	    sigsRequiringAmbiguityChecking.clear();
	}
	
	    // sort the newly reachable sigs, in least-specific-first order.
	DirectedAcyclicGraph dag = 
	    new DirectedAcyclicGraph(newlyReachableSigs.toArray(), 
				     OVERRIDES_RELATION);
	Object[] sorted = dag.inDFSOrder();
	    // now rebuild the newlyReachableSigs list, in the right order
	newlyReachableSigs.clear();
	for (int i = 0; i < sorted.length; i++) {
	    newlyReachableSigs.add(sorted[i]);
	}

	// then iterate through the sorted newly reachable sigs, in
	// least-specific-first order, doing the nessary checking
	Iterator iter = newlyReachableSigs.iterator();
	while (iter.hasNext()) {
	    RMJSignature sig = (RMJSignature) iter.next();

	    // remove this from the list of unchecked sigs (before
	    // doing the checking, so that we don't use this signature
	    // as one that can resolve its own ambiguities)
	    iter.remove();

	    // add to list of sigs to be loaded
	    toBeLoaded.add(sig);

	    if(!RMJOptions.nocheck) {
		    // ensure that this signature isn't ambiguous with any
		    // previously loaded signature, or is resolved by a
		    // signature not yet checked
		verifyNonAmbiguityOfNewSig(sig, toBeLoaded);
		
		if (sig.hasInterfaceSpecializer()) {
			// record that this sig's operation has methods that
			// specialize on interfaces
		    opsWithInterfaceSpecializers.add(sig.operation());
		    
			// remember this sig on each of its interface
			// specializers, so that it can be rechecked for
			// ambiguity whenever a top concrete subclass is
			// loaded later
		    Object[] specializers = sig.specializers();
		    for (int i = 0; i < specializers.length; i++) {
			Object spec = specializers[i];
			if (spec instanceof Class) {
			    Class cls = (Class) spec;
			    if (isInterface(cls)) {
				if (RMJDebug.verbose) {
				    RMJDebug.out("Registering sig " + sig +
						 " on interface specializer " +
						 cls);
				}
				List interfaceSpecializedSignatures =
				    (List)
				    interfaceSpecializedSignaturesMap.get(cls);
				if (interfaceSpecializedSignatures == null) {
					// first one; create the list
				    interfaceSpecializedSignatures =
					new LinkedList();
				    interfaceSpecializedSignaturesMap
					.put(cls,
					     interfaceSpecializedSignatures);
				}
				interfaceSpecializedSignatures.add(sig);
			    }
			}
		    }
		}
	    }
	}

	// now load all the dispatchers for these newly reachable
	// methods.  (do it in in a second pass so that we atomically
	// verify all the non-ambiguity and clear out the list, before
	// we load any classes which might lead to recursive
	// invocation of this method.)
	iter = toBeLoaded.iterator();
	while (iter.hasNext()) {
	    Object entry = iter.next();
	    if (entry instanceof RMJSignature) {
		// a regular glue method to be loaded
		RMJSignature sig = (RMJSignature) entry;
		loadDispatcherClass(sig);
	    } else {
		// an error case to be introduced
		RMJErrorSignature esig = (RMJErrorSignature) entry;
		loadErrorSig(esig);
	    }
	}
    }

    private static final DirectedAcyclicGraph.EdgeCalculator 
	OVERRIDES_RELATION = new DirectedAcyclicGraph.EdgeCalculator() {
		public boolean edgeExists( Object o1, Object o2 ) {
		    RMJSignature s1 = (RMJSignature) o1;
		    RMJSignature s2 = (RMJSignature) o2;
		    return 
			s1.operation().equals(s2.operation()) &&
			s2.overrides(s1);
		}
	    };

    protected void loadDispatcherClass(RMJSignature sig)
	throws ClassNotFoundException
    {
	String className = sig.dispatcherClassName();

	if (isLoaded(className)) {
	    // we've already loaded this dispatcher class (true if an
	    // internal GF, for instance)
	    return;
	}

	if (RMJDebug.verbose) {
	    RMJDebug.out("Loading the dispatcher class for " + sig);
	}
	
	Class dispatcherClass = loadClass(className);

	// Invoke the dispatcher's static init method
	try {
	    Method initMethod = 
		dispatcherClass.getMethod("init", new Class[]{});
	    initMethod.invoke(null, new Object[]{});
	} catch (Exception e) {
	    e.printStackTrace();
	    throw new RuntimeException(
	        "unexpected errors in initializing dispatcher class");
	}
    }

    protected void loadErrorSig(RMJErrorSignature esig)
	throws ClassNotFoundException
    {
	if (esig.isInternalGF()) {
	    // we shouldn't be trying to load an ambiguity error
	    // method for this
	    throw new RuntimeException("shouldn't be trying to load an " +
				       "ambiguity error method for an " +
				       "internal sig");
	}

	if (RMJDebug.verbose) {
	    RMJDebug.out("Creating ambiguity-reporting glue method " + 
			 esig.bodyString());
	}

	String ambigErrorClassName = esig.ambigErrorClassName();

	Class ambigErrorClass = 
	    (Class) loadedClassesAndInterfaces.get(ambigErrorClassName);
	if (ambigErrorClass == null) {
	    // this is the first time we've referenced this ambiguity
	    // error class, so load it now
	    ambigErrorClass = loadClass(ambigErrorClassName);
	}

	// figure out what the arguments should be
	Class[] errorSpecializers = esig.errorSpecializers();
	Class[] errorSpecializerTypes = new Class[errorSpecializers.length];
	Arrays.fill(errorSpecializerTypes, Class.class);

	// Invoke the ambiguity error class's install method to create
	// a new instance and add it to the chain of responsibility
	// for this gf
	try {
	    Method installMethod = 
		ambigErrorClass.getMethod("install", errorSpecializerTypes);
	    installMethod.invoke(null, errorSpecializers);
	} catch (Exception e) {
	    e.printStackTrace();
	    throw new RuntimeException(
	        "unexpected errors in installing ambiguity error method");
	}
    }

    protected void verifyNonAmbiguityOfNewSig(RMJSignature sig,
					      List toBeLoaded)
    {
	verifyNonAmbiguity(sig, false, toBeLoaded);
    }

    protected void verifyNonAmbiguityOfOldSig(RMJSignature sig, 
					      List toBeLoaded)
    {
	verifyNonAmbiguity(sig, true, toBeLoaded);
    }

    protected void verifyNonAmbiguity(RMJSignature sig, boolean isOld,
				      List toBeLoaded)
    {
	if (RMJDebug.verbose) {
	    RMJDebug.out((isOld ? "Rechecking" : "Checking") + 
			 " non-ambiguity of " + sig);
	}

	boolean sigHasInterfaceSpecializer = sig.hasInterfaceSpecializer();

	    // see if an earlier method on the same operation has
	    // interface specializers
	boolean opHasInterfaceSpecializer =
	    opsWithInterfaceSpecializers.contains(sig.operation());

	    // a method needs to be checked against all other methods
	    // that have interface specs, even if the new method doesn't
	    // have interface specs itself, as long as there is some
	    // potential multiple inheritance ambiguity.
	boolean hasInterfaceSpecializerToCheck =
	    sigHasInterfaceSpecializer || opHasInterfaceSpecializer;

	if (hasInterfaceSpecializerToCheck) {
	    // check ambiguity of this signature against each previously
	    // loaded base anchor signature
	    verifyNonAmbiguityAgainstAll(sig, loadedBaseAnchorSigsMap, isOld, 
					 toBeLoaded,
					     // if this method doesn't
					     // specialize on an
					     // interface, then we
					     // need only check this
					     // method against methods
					     // that have interface
					     // specs
					 !sigHasInterfaceSpecializer);
	}

	if (hasInterfaceSpecializerToCheck || sig.isGlue()) {
	    // check ambiguity of this signature against each previously
	    // loaded subanchor signature
	    verifyNonAmbiguityAgainstAll(sig, loadedSubAnchorSigsMap, isOld, 
					 toBeLoaded,
					     // if this method doesn't
					     // specialize on an
					     // interface and is not
					     // glue, then we need
					     // only check this method
					     // against methods that
					     // have interface specs
					 !(sigHasInterfaceSpecializer ||
					   sig.isGlue()));
	}

	if (hasInterfaceSpecializerToCheck || sig.isGlue()
	    || sig.isSubAnchor()) {
	    // check ambiguity of this signature against each previously
	    // loaded glue signature
	    verifyNonAmbiguityAgainstAll(sig, loadedGlueSigsMap, isOld, 
					 toBeLoaded, false);
	}

	if (! isOld) {
	    // now that we've passed the non-ambiguity checks, register
	    // this new signature
	    if (RMJDebug.verbose) {
		RMJDebug.out(" Registering non-ambiguity of " + sig);
	    }

	    if (sig.isBaseAnchor()) {
		registerNonAmbiguousSig(sig, loadedBaseAnchorSigsMap);
	    } else if (sig.isSubAnchor()) {
		registerNonAmbiguousSig(sig, loadedSubAnchorSigsMap);
	    } else if (sig.isGlue()) {
		registerNonAmbiguousSig(sig, loadedGlueSigsMap);
	    } else {
		throw new RuntimeException("unexpected type of sig: " + sig);
	    }
	}
    }

    protected void verifyNonAmbiguityAgainstAll(RMJSignature sig, 
						Map loadedSigsMap, 
						boolean isOld,
						List toBeLoaded,
						boolean onlyCheckIfHasIntSpec)
    {
	RMJOperation op = sig.operation();
	List loadedSigs = (List) loadedSigsMap.get(op);
	if (loadedSigs != null) {
	    Iterator iter = loadedSigs.iterator();
	    while (iter.hasNext()) {
		RMJSignature loadedSig = (RMJSignature) iter.next();
		    // filter out any sigs that don't specialize on an
		    // interface, if the flag is set
		if (onlyCheckIfHasIntSpec &&
		   !loadedSig.hasInterfaceSpecializer()) {
		    if (RMJDebug.verbose) {
			RMJDebug.out("  " + "Not checking against " +
				     loadedSig + " because it has no " +
				     "interface specializers");
		    }
		} else {
		    verifyNonAmbiguityOfSigs(sig, loadedSig,
					     isOld, toBeLoaded);
		}
	    }
	}
    }

    protected void verifyNonAmbiguityOfSigs(RMJSignature nsig, 
					    RMJSignature osig,
					    boolean isOld, List toBeLoaded)
    {
	if (RMJDebug.verbose) {
	    RMJDebug.out("  " + (isOld ? "Rechecking" : "Checking") + 
			 " against " + osig);
	}

	if (isOld && nsig == osig) {
	    // don't check against myself
	    if (RMJDebug.verbose) {
		RMJDebug.out("    (same)");
	    }
	    return;
	}

	try {
	    Collection intersections = nsig.intersect(osig, this);

	    // if we get here, the two signatures are ambiguous;
	    // search for a resolving signature for each intersection,
	    // which won't have been loaded yet, since we're visiting
	    // reachable signatures in top-down order (unless we're
	    // rechecking an old signature, in which case we need to
	    // look in the old signatures too)

	    Iterator iter = intersections.iterator();
	    while (iter.hasNext()) {
		Object[] intersection = (Object[]) iter.next();

		if (RMJDebug.verbose) {
		    RMJDebug.out("    Checking intersection " + 
				 tupleToString(intersection));
		}

		if (! checkForResolvingSig(intersection, isOld, nsig, osig)) {
		    // we didn't find anything that will resolve the
		    // ambiguity
		    String addendum;
		    if (nsig.isInternalGF()) {
			addendum = 
			    "\n\t(Error cannot be detected at " + 
			    "message-invocation time)";
		    } else {
			addendum = "";
			toBeLoaded.add(
			    new RMJErrorSignature(nsig, intersection));
		    }
		    warn("Method " + prettyString(nsig) + 
			 " ambiguous with " + (isOld ? "" : "previous ") + 
			 "method " + prettyString(osig) +
			 " on arguments " + prettyString(intersection) +
			 addendum);
		}
	    }

	} catch (RMJSignature.OverridesArgument e) {
	    // this method overrides the other; OK
	    if (RMJDebug.verbose) {
		RMJDebug.out("    (overrides)");
	    }
	} catch (RMJSignature.IncomparableSignatures e) {
	    // the two signatures are incomparable; OK
	    if (RMJDebug.verbose) {
		RMJDebug.out("    (incomparable)");
	    }
	} catch (RMJSignature.EqualSignatures e) {
	    // the two signatures are duplicates
	    if (isOld) {
		// we should have caught this already
		throw new RuntimeException(
		    "Method " + nsig + " duplicates method " + osig);
	    } else {
		warn("Method " + prettyString(nsig) +
		     " duplicates previous method " + prettyString(osig));
		toBeLoaded.add(new RMJErrorSignature(nsig, e.intersection));
	    }
	} catch (RMJSignature.ArgumentOverrides e) {
	    if (isOld) {
		// the other method overrides this one; OK for old signatures
		if (RMJDebug.verbose) {
		    RMJDebug.out("    (overridden)");
		}
	    } else {
		// the earlier method overrides this one; but that
		// shouldn't happen, since we should be loading glue
		// methods in least-specific-first order
		throw new RuntimeException(
		    "earlier glue method incorrectly overrides later one");
	    }
	}
    }

    protected boolean checkForResolvingSig(Object[] intersection,
					   boolean isOld,
					   RMJSignature nsig, 
					   RMJSignature osig) {
	RMJOperation op = nsig.operation();

	if (isOld) {
	    // we're rechecking an old signature that specializes on
	    // an interface, since we've loaded a new concrete class
	    // that might have created an MI ambiguity.  we've found
	    // an intersection.  first check existing signatures to
	    // see if they (still) resolve it
	    List loadedBaseAnchorSigs =
		(List) loadedBaseAnchorSigsMap.get(op);
	    if (checkCollectionForResolvingSig(loadedBaseAnchorSigs, null,
					       intersection,
					       nsig, osig)) {
		// found a resolver
		return true;
	    }
	    List loadedSubAnchorSigs =
		(List) loadedSubAnchorSigsMap.get(op);
	    if (checkCollectionForResolvingSig(loadedSubAnchorSigs, null,
					       intersection,
					       nsig, osig)) {
		// found a resolver
		return true;
	    }
	    List loadedGlueSigs =
		(List) loadedGlueSigsMap.get(op);
	    if (checkCollectionForResolvingSig(loadedGlueSigs, null,
					       intersection,
					       nsig, osig)) {
		// found a resolver
		return true;
	    }
	}

	// check to-be-loaded signatures to see if they will resolve
	// the ambiguity
	if (checkCollectionForResolvingSig(newlyReachableSigs, op,
					   intersection, null, null)) {
	    // found a resolver
	    return true;
	}

	// didn't find any resolving method
	return false;
    }

    protected boolean checkCollectionForResolvingSig(Collection sigs,
						     RMJOperation op,
						     Object[] intersection,
						     RMJSignature nsig,
						     RMJSignature osig) {
	if (sigs != null) {
	    Iterator iter = sigs.iterator();
	    while (iter.hasNext()) {
		RMJSignature sig = (RMJSignature) iter.next();
		if ((op == null || sig.operation().equals(op)) &&
		    sig.appliesTo(intersection) && 
		    (nsig == null || sig.overrides(nsig)) &&
		    (osig == null || sig.overrides(osig))) {
		    // we found a method that resolves the ambiguity
		    if (RMJDebug.verbose) {
			RMJDebug.out("      Intersection covered by " + sig);
		    }
		    return true;
		}
	    }
	}
	return false;
    }

    protected void registerNonAmbiguousSig(RMJSignature sig, Map loadedSigsMap)
    {
	RMJOperation op = sig.operation();
	List loadedSigs = (List) loadedSigsMap.get(op);
	if (loadedSigs == null) {
	    loadedSigs = new LinkedList();
	    loadedSigsMap.put(op, loadedSigs);
	}
	loadedSigs.add(sig);
    }


    ///////////////////////////////////////////////////////////////////////

    // verify that the whole system is again complete, after loading
    // cls and any newly reachable glue
    protected void verifyCompleteness(Class cls, RMJAnnotation anno)
    {
	if (isAbstractClassOrInterface(cls)) {
	    // nothing to do for an abstract class or interface; they
	    // can't cause the system to become incomplete

	} else if (isTopConcreteClass(cls)) {
	    // a top concrete class requires new completeness checking
	    verifyCompletenessOfTopConcreteClass(cls);

	} else {
	    // nothing to do for a non-top concrete class; previous
	    // top concrete classes (e.g. its superclasses) have
	    // already ensured that the system will be complete for
	    // all their non-top concrete subclasses
	}

	// now check any abstract method signatures introduced in this
	// class, if it's an anchor class, and then clear them out
	Iterator iter = newlyLoadedAbstractSignatures.iterator();
	while (iter.hasNext()) {
	    RMJSignature asig = (RMJSignature) iter.next();
	    verifyCompletenessOfAbstractSignature(asig);
	    iter.remove();
	}
    }

    protected void verifyCompletenessOfTopConcreteClass(Class cls)
    {
	// we're loading a top concrete class.  we need to verify that
	// all previously loaded abstract signatures are implemented
	// for this class

	if (RMJDebug.verbose) {
	    RMJDebug.out("Verifying completeness of top concrete class " + 
			 cls.getName());
	}

	// first, find all the abstract signatures that this top
	// concrete class "inherits"
	HashSet asigs = new HashSet();
	collectAbstractSignaturesAbove(cls, asigs);
	
	// then check each of them for completeness
	Iterator iter = asigs.iterator();
	while (iter.hasNext()) {
	    RMJSignature asig = (RMJSignature) iter.next();
	    verifyCompletenessOfAbstractSignature(asig, cls);
	}
    }

    protected void collectAbstractSignaturesAbove(Class cls, HashSet asigs)
    {
	Class intf[] = cls.getInterfaces();
	for (int i = 0; i < intf.length; i++) {
	    Class in = intf[i];
	    collectAbstractSignaturesOf(in, asigs);
	}
	Class scls = cls.getSuperclass();
	if (scls != null && isAbstractClassOrInterface(scls)) {
	    // it's an abstract superclass
	    collectAbstractSignaturesOf(scls, asigs);
	}
    }

    protected void collectAbstractSignaturesOf(Class cls, HashSet asigs)
    {
	// find all the abstract signatures stored with this abstract
	// class/interface
	List sigsRequiringCompletenessChecking = 
	    (List) sigsRequiringCompletenessCheckingMap.get(cls);

	// add each of them to the asigs set
	if (sigsRequiringCompletenessChecking != null) {
	    asigs.addAll(sigsRequiringCompletenessChecking);
	}

	// continue accumulating abstract signatures from additional
	// abstract superclasses/interfaces
	collectAbstractSignaturesAbove(cls, asigs);
    }


    protected void verifyCompletenessOfAbstractSignature(RMJSignature asig)
    {
	verifyCompletenessOfAbstractSignature(asig, null);
    }

    // here, if cls is non-null, then only bother checking tuples including cls
    protected void verifyCompletenessOfAbstractSignature(RMJSignature asig, 
							 Class cls)
    {
	// we detected an abstract signature.  we need to verify that
	// it's implemented for every combination of top concrete
	// classes of receiver + arguments.
	if (RMJDebug.verbose) {
	    RMJDebug.out("Verifying completeness of abstract signature " + 
			 asig + 
			 (cls == null ? "" : " for " + cls.getName()));
	}
	try {
	    List tuples = generateTopConcreteTuplesBelow(asig);
	    verifyCompletenessOfTopConcreteTuples(tuples, asig, cls);
	} catch (EmptyTupleSet e) {
	    // no concrete tuples to check
	    if (RMJDebug.verbose) {
		RMJDebug.out("  No concrete argument tuples to check");
	    }
	}
    }

    protected List generateTopConcreteTuplesBelow(RMJSignature asig)
	throws EmptyTupleSet
    {
	// compute the tuple of sets:
	List[] tupleOfSets = generateTupleOfTopConcreteSetsBelow(asig);

	// now generate their cross-product, the set of tuples:
	List setOfTuples = generateCrossProduct(tupleOfSets);

	return setOfTuples;
    }

    // this method generates a set of top concrete specializers (which
    // are either class or value specializers) for each argument
    // position of the signature
    protected List[] generateTupleOfTopConcreteSetsBelow(RMJSignature asig)
	throws EmptyTupleSet
    {
	Object[] specializers = asig.specializersOrTypes();
	List[] tuple = new List[specializers.length];
	for (int i = 0; i < specializers.length; i++) {
	    tuple[i] = generateTopConcreteSetsBelow(specializers[i]);
	}
	return tuple;
    }

    // this method generates a set of top concrete specializers (which
    // are either class or value specializers, or null) for a specializer
    protected List generateTopConcreteSetsBelow(Object specializer)
	throws EmptyTupleSet
    {
	List topConcreteSpecs = new LinkedList();
	if (specializer == null) {
	    // an unspecialized position of prim type
	    topConcreteSpecs.add(null);
	} else if (specializer instanceof Class) {
	    // a regular specialized position
	    Class cls = (Class) specializer;
	    if (isAbstractClassOrInterface(cls)) {
		// add each of this abstract class's top concrete children
		List topConcreteSubclasses =
		    (List) topConcreteSubclassesMap.get(cls);
		if (topConcreteSubclasses == null) {
		    // this abstract class doesn't have any
		    // implementations yet, so the abstract signature
		    // need not be implemented at all
		    throw new EmptyTupleSet();
		}
		topConcreteSpecs.addAll(topConcreteSubclasses);
	    } else {
		// the class is concrete, so it's its own top concrete subclass
		topConcreteSpecs.add(cls);
	    }
	} else {
	    // a value specializer; we have just this one case to check
	    topConcreteSpecs.add(specializer);
	}
	return topConcreteSpecs;
    }

    protected List generateCrossProduct(List[] tupleOfSets)
    {
	return generateCrossProductTo(tupleOfSets, tupleOfSets.length-1);
    }

    protected List generateCrossProductTo(List[] tupleOfSets, int pos)
    {
	List setOfTuples = new LinkedList();

	if (pos < 0) {
	    // we're generating the cross-product of 0 sets => return
	    // the singleton set of the empty tuple
	    setOfTuples.add(new Object[0]);
	} else {
	    // first recursively generate the cross-product of all earlier
	    // positions
	    List setOfEarlierTuples = 
		generateCrossProductTo(tupleOfSets, pos-1);

	    // then extend each of these for each element at this position
	    Iterator iter = setOfEarlierTuples.iterator();
	    while (iter.hasNext()) {
		Object[] earlierTuple = (Object[]) iter.next();

		Iterator iter2 = tupleOfSets[pos].iterator();
		while (iter2.hasNext()) {
		    Object newSpec = iter2.next();

		    Object[] newTuple = new Object[pos+1];
		    for (int i = 0; i < pos; i++) {
			newTuple[i] = earlierTuple[i];
		    }
		    newTuple[pos] = newSpec;

		    setOfTuples.add(newTuple);
		}
	    }
	}

	return setOfTuples;
    }

    // a way to abort tuple generation if we determine that the tuple
    // set is empty
    protected static class EmptyTupleSet extends Exception {
    }

    protected void verifyCompletenessOfTopConcreteTuples(List tuples,
							 RMJSignature asig,
							 Class cls)
    {
	// fetch the list of top concrete implementations of this signature
	List implementations = 
	    (List) sigImplementationsMap.get(asig);
	if (RMJDebug.verbose) {
	    RMJDebug.out(" Checking against " + implementations);
	}

	// for each tuple of top concrete argument specializers...
	Iterator iter = tuples.iterator();
	while (iter.hasNext()) {
	    Object[] tuple = (Object[]) iter.next();

	    if (cls != null) {
		// ignore this tuple unless it contains cls, since
		// we've already checked it
		boolean foundIt = false;
		for (int i = 0; i < tuple.length; i++) {
		    if (cls.equals(tuple[i])) {
			foundIt = true;
			break;
		    }
		}
		if (! foundIt) {
		    // skip this tuple
		    if (RMJDebug.verbose) {
			RMJDebug.out("  Skipping previously checked tuple " + 
				     tupleToString(tuple));
		    }
		    continue;
		}
	    }

	    if (RMJDebug.verbose) {
		RMJDebug.out("  Verifying completeness of tuple " + 
			     tupleToString(tuple));
	    }

	    // see if there exists some method implementation that handles it
	    if (implementations != null) {
		boolean foundOne = false;
		Iterator iter2 = implementations.iterator();
		while (iter2.hasNext()) {
		    RMJSignature impl = (RMJSignature) iter2.next();
		    if (impl.appliesTo(tuple)) {
			// we found a match
			if (RMJDebug.verbose) {
			    RMJDebug.out("    Covered by " + impl);
			}
			foundOne = true;
			break;
		    }
		}
		if (foundOne) {
		    // we found a match; go on to the next tuple
		    continue;
		}
	    }

	    // we didn't find any concrete implementations for this
	    // argument tuple
	    warn("Abstract method " + prettyString(asig) +
		 " not implemented for " + prettyString(tuple));
	}
    }

    ///////////////////////////////////////////////////////////////////////

    // compute the collection of classes that are the GLBs of the two
    // argument classes, i.e., subclasses of both that have no
    // concrete superclasses that are subclasses of both.  (the
    // classes in the list can be concrete or abstract, as is
    // convenient for this code.)  the two argument classes are
    // assumed to be incomparable.  if there is no intersection, then
    // the empty collection (or null) is returned.
    /* package */ Collection computeIntersection(Class cls1, 
						 Class cls2) 
    {
	if (isInterface(cls1)) {
	    // find all the top concrete subclasses of this interface
	    List topSubclasses1 = 
		(List) topConcreteSubclassesMap.get(cls1);
	    if (topSubclasses1 == null) {
		// no concrete subclasses => an empty intersection
		return null;
	    }

	    List intersection = new LinkedList();

	    List topSubclasses2;
	    if (isInterface(cls2)) {
		// find all the top concrete subclasses of this interface
		topSubclasses2 = 
		    (List) topConcreteSubclassesMap.get(cls2);
		if (topSubclasses2 == null) {
		    // no concrete subclasses => an empty intersection
		    return null;
		}
	    } else {
		// make this class its own top subclass
		topSubclasses2 = new LinkedList();
		topSubclasses2.add(cls2);
	    }

	    // now, for every combination of top subclass of each
	    // argument class, find their intersection, if any, and
	    // add it to the list of intersections
	    Iterator iter1 = topSubclasses1.iterator();
	    while (iter1.hasNext()) {
		Class subcls1 = (Class) iter1.next();

		Iterator iter2 = topSubclasses2.iterator();
		while (iter2.hasNext()) {
		    Class subcls2 = (Class) iter2.next();

		    // see if one inherits from the other.  if so,
		    // it's the intersection, otherwise there isn't
		    // one.
		    if (subcls2.isAssignableFrom(subcls1)) {
			// subcls1 is the more specific (or they're equal)
			intersection.add(subcls1);
		    } else if (subcls1.isAssignableFrom(subcls2)) {
			// subcls2 is the more specific
			intersection.add(subcls2);
		    } else {
			// they're incomparable
		    }
		}
	    }

	    return intersection;

	} else if (isInterface(cls2)) {
	    // reverse the arguments and retry
	    return computeIntersection(cls2, cls1);

	} else {
	    // they're both classes, which (since we're assuming
	    // they're incomparable) have no intersection
	    return null;
	}
    }


    ///////////////////////////////////////////////////////////////////////

    public String prettyString(RMJSignature sig) {
	String s = "";
	Class[] staticTypes = sig.staticTypes();
	Object[] specializers = sig.specializers();
	for (int i = 0; i < specializers.length; i++) {
	    Class staticType = staticTypes[i];
	    Object specializer = specializers[i];
	    if (i > 1) s += ",";
	    if (specializer == null) {
		// unspecialized in this position
		if (staticType == null) {
		    // a prim type
		    s += "<prim>";
		} else {
		    s += staticType.getName();
		}
	    } else if (specializer instanceof Class) {
		// a regular specializer
		s += "@" + ((Class)specializer).getName();
	    } else {
		// a value specializer
		s += "@@" + specializer.toString();
	    }
	    if (i == 0) {
		// this was the receiver
		s += "$" + sig.opName() + "(";
	    }
	}
	s += ")";
	return s;
    }

    public String prettyString(RMJAnnotation.Method sig) {
	String s = "";
	String[] staticTypeDescs = sig.staticTypeDescs();
	String[] specializerDescs = sig.specializerDescs();
	for (int i = 0; i < specializerDescs.length; i++) {
	    String staticTypeDesc = staticTypeDescs[i];
	    String specializerDesc = specializerDescs[i];
	    if (i > 1) s += ",";
	    char c = specializerDesc.charAt(0);
	    if (c == '=') {
		// a value specializer
		String specializerValue = specializerDesc.substring(1);
		s += "@@" + specializerValue;
	    } else if (c == '-') {
		// an unspecialized argument position
		c = staticTypeDesc.charAt(0);
		if (c == '=') {
		    // a primitive type
		    String staticType = staticTypeDesc.substring(1);
		    s += staticType;
		} else {
		    // a class type
		    s += staticTypeDesc;
		}
	    } else {
		// a class specializer
		s += "@" + specializerDesc;
	    }
	    if (i == 0) {
		// that was the receiver
		s += "$" + sig.opName() + "(";
	    }
	}
	s += ")";
	return s;
    }

    protected String prettyString(Object[] tuple)
    {
	String s = "";
	for (int i = 0; i < tuple.length; i++) {
	    if (i > 1) s += ",";
	    Object spec = tuple[i];
	    if (spec == null) {
		// an unspecialized argument
		s += "-";
	    } else if (spec instanceof Class) {
		// a class specializer
		s += ((Class) spec).getName();
	    } else {
		// a value specializer
		s += spec;
	    }
	    if (i == 0) {
		// that was the receiver
		s += "(";
	    }
	}
	s += ")";
	return s;
    }

    protected String tupleToString(Object[] tuple)
    {
	String s = "(";
	for (int i = 0; i < tuple.length; i++) {
	    if (i > 0) s += ",";
	    Object spec = tuple[i];
	    if (spec == null) {
		// an unspecialized argument
		s += "-";
	    } else if (spec instanceof Class) {
		// a class, specialized or not
		s += ((Class) spec).getName();
	    } else {
		// a value specializer
		s += spec;
	    }
	}
	s += ")";
	return s;
    }

    protected void warn(String msg)
    {
	// for now, print warnings to stderr and continue executing
	System.err.println("** RMJ Hazard: " + msg);
    }


    ///////////////////////////////////////////////////////////////////////

    /* package */ Class lookupClass(String className)
	throws ClassNotFoundException
    {
	Class cls = (Class) loadedClassesAndInterfaces.get(className);
	if (cls == null) {
	    throw new ClassNotFoundException("class " + className + 
					     " not found");
	}
	return cls;
    }

    /* package */ boolean isLoaded(String className)
    {
	Class cls = (Class) loadedClassesAndInterfaces.get(className);
	return cls != null;
    }


    ///////////////////////////////////////////////////////////////////////

    protected boolean isInterface(Class cls)
    {
	return cls.isInterface();
    }

    protected boolean isAbstractClassOrInterface(Class cls)
    {
	return Modifier.isAbstract(cls.getModifiers());
    }

    protected boolean isAbstractClass(Class cls)
    {
	return isAbstractClassOrInterface(cls) && !isInterface(cls);
    }

    protected boolean isConcreteClass(Class cls)
    {
	return !isAbstractClassOrInterface(cls);
    }

    
    protected boolean isTopConcreteClass(Class cls)
    {
	if (isAbstractClassOrInterface(cls)) return false;

	Class scls = cls.getSuperclass();
	if (scls != null && isAbstractClassOrInterface(scls)) return true;

	Class intf[] = cls.getInterfaces();
	if (intf.length > 0) return true;

	return false;
    }


    ///////////////////////////////////////////////////////////////////////

    public void printClassReport(PrintStream strm)
    {
	// XXX write me
    }


    ///////////////////////////////////////////////////////////////////////

    // INSTANCE VARIABLES

    // a map from the names of loaded classes and interfaces (Strings)
    // to the corresponding Class objects
    protected Map loadedClassesAndInterfaces = new HashMap();

    // THINGS FOR COMPLETENESS CHECKING:

    // a map from each loaded abstract class or interface (a Class) to
    // its loaded top concrete subclasses (a List of Class).
    protected Map topConcreteSubclassesMap = new HashMap();

    // a map from each loaded abstract class or interface to a list of
    // abstract method signatures requiring completeness checking (a
    // List of RMJMethodSig) whenever a top concrete subclass of this
    // thing is loaded
    protected Map sigsRequiringCompletenessCheckingMap = new HashMap();

    // a map from each loaded operation (an RMJOperation) to the list
    // of reachable method signatures (a List of RMJSignature) for the
    // operation
    protected Map signaturesMap = new HashMap();

    // a map from each abstract method signature (an RMJSignature) to
    // the list of top concrete implementations (a List of
    // RMJSignature)
    protected Map sigImplementationsMap = new HashMap();

    // a list of newly loaded abstract signatures (RMJSignature) that
    // need completeness checking; removed after they've been checked.
    protected List newlyLoadedAbstractSignatures = new LinkedList();

    // THINGS FOR GLUE METHOD LOADING:

    // a map of from the name of a glue anchor class (a String) to its
    // annotations (RMJAnnotation), which have already been read
    // during eager reading of glue includes and whose method
    // annotations have already been processed.
    protected Map glueAnchorClassAnnotationsMap = new HashMap();

    // a map from each unloaded class or interface name (String) to a
    // list of unloaded signatures (List of RMJUnreachableSignature)
    // awaiting the class/interface before becoming reachable
    protected Map unreachableSigsMap = new HashMap();

    // a list of newly reachable signatures (RMJSignature) that need
    // to be verified for non-ambiguity, and if a glue method, loaded;
    // removed after they've been checked.
    protected List newlyReachableSigs = new LinkedList();

    // THINGS FOR GLUE METHOD AMBIGUITY CHECKING:

    // a map from operations (RMJOperation) to lists of previously
    // loaded signatures in base anchors (List of RMJSignature)
    protected Map loadedBaseAnchorSigsMap = new HashMap();

    // a map from operations (RMJOperation) to lists of previously
    // loaded signatures in subanchors, i.e., internal methods on
    // external GFs (List of RMJSignature)
    protected Map loadedSubAnchorSigsMap = new HashMap();

    // a map from operations (RMJOperation) to lists of previously
    // loaded glue signatures (List of RMJSignature)
    protected Map loadedGlueSigsMap = new HashMap();

    // THINGS FOR MULTIPLE INHERITANCE AMBIGUITY CHECKING:

    // a map from loaded interfaces (Class) to a list of signatures
    // (List of RMJSignature) that specialize on that interface, so
    // that these signatures can be rechecked for ambiguity whenever a
    // new top concrete class implementing the interface is loaded
    protected Map interfaceSpecializedSignaturesMap = new HashMap();

    // a list of signatures (RMJSignature) that might have become
    // ambiguous due to loading a top concrete subclass of an
    // interface
    protected List sigsRequiringAmbiguityChecking = new LinkedList();

	// a list of operations (RMJOperation) that have at least one method
	// specializing on an interface
    protected List opsWithInterfaceSpecializers = new LinkedList();

}
