// An object of this class represents a particular reachable method
// signature.

package org.multijava.relaxed.runtime;

import org.multijava.relaxed.util.*;

import java.util.*;

public class RMJSignature
{
    public RMJSignature(RMJAnnotation.Method anno, RMJClassLoader loader)
    {
	try {
	    String[] staticTypeDescs = anno.staticTypeDescs();
	    String staticTypesString = "";
	    _staticTypes = new Class[staticTypeDescs.length];
	    for (int i = 0; i < staticTypeDescs.length; i++) {
		String staticTypeDesc = staticTypeDescs[i];
		Class staticType;
		String staticTypeName;
		char c = staticTypeDesc.charAt(0);
		if (c == '=') {
		    // a prim type
		    staticTypeName = staticTypeDesc.substring(1);
		    staticType = null;
		} else {
		    // a class type
		    staticTypeName = staticTypeDesc;
		    staticType = loader.lookupClass(staticTypeName);
		}
		_staticTypes[i] = staticType;
		if (i > 0) staticTypesString += ",";
		staticTypesString += staticTypeName;
	    }

	    _op = new RMJOperation(anno.opName(), staticTypesString);

	    String[] specializerDescs = anno.specializerDescs();
	    _specializers = new Object[specializerDescs.length];
	    _specializersOrTypes = new Object[specializerDescs.length];
	    _hasInterfaceSpecializer = false;
	    for (int j = 0; j < specializerDescs.length; j++) {
		String specializerDesc = specializerDescs[j];
		Object specializer;
		char c = specializerDesc.charAt(0);
		if (c == '=') {
		    // a value specializer
		    String specializerValue = specializerDesc.substring(1);
		    try {
			specializer = new Integer(specializerValue);
		    } catch (NumberFormatException e) {
			try {
			    specializer = new Double(specializerValue);
			} catch (NumberFormatException e2) {
			    if (specializerValue.equals("true")) {
				specializer = new Boolean(true);
			    } else if (specializerValue.equals("false")) {
				specializer = new Boolean(false);
			    } else if (specializerValue.charAt(0) == '"') {
				// remove the outer quotes
				int len = specializerValue.length();
				specializer = 
				    specializerValue.substring(1, len-1);
			    } else {
				System.err.println(
				    "unexpected specializer value: " + 
				    specializerValue);
				specializer = null;
			    }
			}
		    }
		} else if (c == '-') {
		    // an unspecialized argument position
		    specializer = null;
		} else {
		    // a class-specialized argument position
		    String className = specializerDesc;
		    Class cls = loader.lookupClass(className);
		    if (cls.isInterface()) {
			_hasInterfaceSpecializer = true;
		    }
		    specializer = cls;
		}
		_specializers[j] = specializer;
		if (specializer != null) {
		    _specializersOrTypes[j] = specializer;
		} else {
		    _specializersOrTypes[j] = _staticTypes[j];
		}
	    }

	    _isAbstract = anno.isAbstract();

	    _kind = anno.kind(); 

	    // don't load the base anchor or dispatcher classes until
	    // they're explicitly requested
	    _baseAnchorClassName = anno.baseAnchorClassName();
	    _dispatcherClassName = anno.dispatcherClassName();
	    _ambigErrorClassName = anno.ambigErrorClassName();

	} catch (ClassNotFoundException e) {
	    // unexpected error
	    System.err.println("Couldn't find class: " + e.getMessage());
	    e.printStackTrace();
	    System.exit(1);
	}
    }


    // OPERATIONS

    // return the set of names of unloaded classes that must be loaded
    // before this signature becomes reachable
    static public Set unloadedClassNames(RMJAnnotation.Method anno,
					 RMJClassLoader loader)
    {
	HashSet unloadedClassNames = new HashSet();

	String[] specializerDescs = anno.specializerDescs();
	for (int i = 0; i < specializerDescs.length; i++) {
	    String specializerDesc = specializerDescs[i];
	    char c = specializerDesc.charAt(0);
	    String className;
	    if (c == '=') {
		// a value specializer, of a prim type.  always reachable
		continue;
	    } else if (c == '-') {
		// an unspecialized argument position.  reachable when the
		// static type is
		String staticTypeDesc = (anno.staticTypeDescs())[i];
		char first = staticTypeDesc.charAt(0);
		if (first == '=') {
			// a prim type; always reachable
		    continue;
		} else {
			// a class type
		    className = staticTypeDesc;
		}
	    } else {
		// a (non-value) specialized argument position; not
		// reachable until the specializer is
		className = specializerDesc;
	    }
	    if (! loader.isLoaded(className)) {
		unloadedClassNames.add(className);
	    }
	}

	String baseAnchorClassName = anno.baseAnchorClassName();
	if (! loader.isLoaded(baseAnchorClassName)) {
	    unloadedClassNames.add(baseAnchorClassName);
	}

	return unloadedClassNames;
    }


        // return whether this signature represents the exact same
	// method as another signature.  note: duplicate glue methods
	// and similar ambiguities will not be "equals" under this
	// semantics -- this method only returns true when the two
	// signatures represent the same syntactic method declaration.
	// such duplication can occur because subclasses that
	// specialize on interfaces include method annotations for
	// methods in their superclasses.
    public boolean equals(/*@ nullable @*/ Object other) {
	if (other instanceof RMJSignature) {
	    return this.bodyString().equals(
		((RMJSignature)other).bodyString());
	} else {
	    return false;
	}
    }

    public int hashCode() {
        return bodyString().hashCode();
    }
    
    // return whether this signature is equal to or overrides the
    // other signature; we assume the two signatures are in the same
    // operation
    public boolean overrides(RMJSignature other)
    {
	if (! operation().equals(other.operation())) {
	    throw new RuntimeException(
	        "calling overrides on signatures from different operations");
	}
	// the below will check that the # of specializers is the same
	return overrides(specializers(), other.specializers());
    }

    // return whether this signature covers the given tuple of
    // receiver & argument specializers (which are classes, objects,
    // or null).  we assume that the signature is in the same GF as
    // the specializers.
    public boolean appliesTo(Object[] argSpecializers)
    {
	return overrides(argSpecializers, specializers());
    }

    // return whether the first tuple of specializers is at least as
    // specific as the second.  we assume they're from the same GF.
    private static boolean overrides(Object[] specializers, 
				     Object[] otherSpecializers)
    {
	if (specializers.length != otherSpecializers.length) {
	    throw new RuntimeException(
		"calling overrides on specializers of different lengths");
	}
	for (int i = 0; i < specializers.length; i++) {
	    Object thisSpec = specializers[i];
	    Object otherSpec = otherSpecializers[i];
	    if (thisSpec == null) {
		// this is unspecialized
		if (otherSpec == null) {
		    // the other is, too; they're the same
		    continue;
		} else {
		    // the other is specialized, somehow, so this
		    // doesn't override
		    return false;
		}
	    } else if (otherSpec == null) {
		// this is specialized, but the other isn't, so this overrides
		continue;
	    } else {
		// they're both specialized, somehow
		if (thisSpec instanceof Class) {
		    // this is a regular class specializer
		    if (otherSpec instanceof Class) {
			// they're two classes or interfaces, meaning
			// that we should compare using regular class
			// overrides comparisons
			Class thisClass = (Class) thisSpec;
			Class otherClass = (Class) otherSpec;
			if (otherClass.isAssignableFrom(thisClass)) {
			    // this is a subtype of the other, so this
			    // overrides
			    continue;
			} else {
			    // this isn't a subtype, so it doesn't override
			    return false;
			}
		    } else {
			// this is a class, but the other is a value
			// specializer, so this doesn't override
			return false;
		    }
		} else {
		    // this is a value specializer
		    if (otherSpec instanceof Class) {
			// the other is a class, but this is a value
			// specializer, which overrides
			continue;
		    } else {
			// they're both value specializers
			if (thisSpec.equals(otherSpec)) {
			    // they're the same value specializer, so we're OK
			    continue;
			} else {
			    // they're different, and incomparable, so
			    // this doesn't override
			    return false;
			}
		    }
		}
	    }
	}

	// all this signature's specializers were equal to or more
	// specific than the other's, so it does override
	return true;
    }

    // compute the collection of intersections (an array of
    // Class/Objects) of the specializers of this and the argument
    // signature, for ambiguity checking.  if one signature overrides
    // the other, then throw the appropriate exception.  if the two
    // signatures have the same specializers, then throw that
    // exception.  in the face of interfaces with multiple top
    // concrete subclasses, there can be multiple independent
    // intersections, leading to a non-singleton result list.  we
    // assume that the two signatures are in the same operation.
    public Collection intersect(RMJSignature other, RMJClassLoader loader) 
	throws OverridesArgument, ArgumentOverrides,
	       EqualSignatures, IncomparableSignatures
    {
	if (! operation().equals(other.operation())) {
	    throw new RuntimeException(
	        "calling intersect on signatures from different operations");
	}

	Object[] specializers = specializers();
	Object[] otherSpecializers = other.specializers();

	if (specializers.length != otherSpecializers.length) {
	    throw new RuntimeException(
		"calling intersect on signatures with specializers " +
		"of different lengths");
	}

	// the state after 0 arguments:
	int state = EQUAL;

	// whether there are lists of objects in any position of the
	// intersection
	boolean needsExpansion = false;

	// the intersection, eventually, if we need it
	Object[] intersection = new Object[specializers.length];

	for (int i = 0; i < specializers.length; i++) {
	    Object thisSpec = specializers[i];
	    Object otherSpec = otherSpecializers[i];

	    try {
		// compare the specializers, throwing Incomparable if this
		// position is Incomparable
		int comparison = intersect_one(thisSpec, otherSpec, loader);

		// update the state based on the comparison
		if (comparison == EQUAL) {
		    // they were the same; no change to the state
		    intersection[i] = thisSpec;
		} else if (comparison == THIS_OVERRIDES) {
		    // this overrides the other
		    if (state == EQUAL) {
			// weaken state
			state = THIS_OVERRIDES;
		    } else if (state == OTHER_OVERRIDES) {
			// weaken state
			state = AMBIGUOUS;
		    }
		    intersection[i] = thisSpec;
		} else {
		    // the other overrode this
		    if (state == EQUAL) {
			// weaken state
			state = OTHER_OVERRIDES;
		    } else if (state == THIS_OVERRIDES) {
			// weaken state
			state = AMBIGUOUS;
		    }
		    intersection[i] = otherSpec;
		}
	    } catch (HasIntersection e) {
		Collection one_intersection = e.getIntersection();
		if (one_intersection.size() == 1) {
		    // just store the single element
		    intersection[i] = one_intersection.iterator().next();
		} else {
		    // store the whole collection of things in the
		    // intersection
		    intersection[i] = one_intersection;
		    needsExpansion = true;
		}
		state = AMBIGUOUS;
	    }
	}

	// return the right final result
	if (state == EQUAL) throw new EqualSignatures(intersection);
	else if (state == THIS_OVERRIDES) throw new OverridesArgument();
	else if (state == OTHER_OVERRIDES) throw new ArgumentOverrides();

	// state == AMBIGUOUS; return a non-trivial intersection
	if (needsExpansion) {
	    // we have to expand a tuple of sets into a set of tuples
	    return generateCrossProduct(intersection);
	} else {
	    // a single intersection tuple
	    List intersections = new LinkedList();
	    intersections.add(intersection);
	    return intersections;
	}
    }

    public static class OverridesArgument extends Exception { }
    public static class ArgumentOverrides extends Exception { }
    public static class EqualSignatures extends Exception {
	public Object[] intersection;
	public EqualSignatures(Object[] i) { intersection = i; }
    }
    public static class IncomparableSignatures extends Exception { }

    // the possible states of comparison
    private static final int EQUAL = 0;
    private static final int THIS_OVERRIDES = 1;
    private static final int OTHER_OVERRIDES = 2;
    private static final int AMBIGUOUS = 3;

    private int intersect_one(Object thisSpec, Object otherSpec, 
			      RMJClassLoader loader) 
	throws IncomparableSignatures, HasIntersection
    {
	if (thisSpec == null) {
	    // this is unspecialized
	    if (otherSpec == null) {
		// the other is, too; they're the equal.
		// no change to the state.
		return EQUAL;
	    } else {
		// the other is specialized, somehow, so the other
		// overrides this
		return OTHER_OVERRIDES;
	    }
	} else if (otherSpec == null) {
	    // this is specialized, but the other isn't, so this overrides
	    return THIS_OVERRIDES;
	} else {
	    // they're both specialized, somehow
	    if (thisSpec instanceof Class) {
		// this is a regular class/interface specializer
		if (otherSpec instanceof Class) {
		    // they're two classes/interfaces, meaning
		    // that we should compare using regular class
		    // overrides comparisons
		    Class thisClass = (Class) thisSpec;
		    Class otherClass = (Class) otherSpec;
		    if (thisClass.equals(otherClass)) {
			// they're the same
			return EQUAL;
		    } else if (otherClass.isAssignableFrom(thisClass)) {
			// this is a subtype of the other, so this
			// overrides
			return THIS_OVERRIDES;
		    } else if (thisClass.isAssignableFrom(otherClass)) {
			// the overriding works the other way
			return OTHER_OVERRIDES;
		    } else {
			// the two classes/interfaces are
			// incomparable.  but if at least one is
			// an interface, then there may be an
			// intersection.
			Collection intersection = 
			    loader.computeIntersection(thisClass, otherClass);
			if (intersection == null || intersection.isEmpty()) {
			    throw new IncomparableSignatures();
			} else {
			    throw new HasIntersection(intersection);
			}
		    }
		} else {
		    // this is a class, but the other is a value
		    // specializer, so it overrides this
		    return OTHER_OVERRIDES;
		}
	    } else {
		// this is a value specializer
		if (otherSpec instanceof Class) {
		    // the other is a class, but this is a value
		    // specializer, which overrides
		    return THIS_OVERRIDES;
		} else {
		    // they're both value specializers
		    if (thisSpec.equals(otherSpec)) {
			// they're the same value specializer
			return EQUAL;
		    } else {
			// they're different, and incomparable
			throw new IncomparableSignatures();
		    }
		}
	    }
	}
    }

    private static class HasIntersection extends Exception
    { 
	private Collection intersection;
	public HasIntersection(Collection i)
	{
	    intersection = i;
	}
	public Collection getIntersection() { return intersection; }
    }

    private Collection generateCrossProduct(Object[] tupleOfSets)
    {
	return generateCrossProductTo(tupleOfSets, tupleOfSets.length-1);
    }

    private Collection generateCrossProductTo(Object[] tupleOfSets, int pos)
    {
	List setOfTuples = new LinkedList();

	if (pos < 0) {
	    // we're generating the cross-product of 0 sets => return
	    // the singleton set of the empty tuple
	    setOfTuples.add(new Object[0]);
	} else {
	    // first recursively generate the cross-product of all earlier
	    // positions
	    Collection setOfEarlierTuples = 
		generateCrossProductTo(tupleOfSets, pos-1);

	    // then extend each of these for each element at this position
	    Iterator iter = setOfEarlierTuples.iterator();
	    while (iter.hasNext()) {
		Object[] earlierTuple = (Object[]) iter.next();

		Object elem = tupleOfSets[pos];
		if (elem instanceof Collection) {
		    // a list of specializers at this position
		    Iterator iter2 = ((Collection) elem).iterator();
		    while (iter2.hasNext()) {
			Object newSpec = iter2.next();
			extendTuple(earlierTuple, pos, newSpec, setOfTuples);
		    }
		} else {
		    // a single specializer at this position
		    extendTuple(earlierTuple, pos, elem, setOfTuples);
		}
	    }
	}

	return setOfTuples;
    }

    private void extendTuple(Object[] earlierTuple, int pos, Object spec, 
			     List setOfTuples)
    {
	Object[] newTuple = new Object[pos+1];
	for (int i = 0; i < pos; i++) {
	    newTuple[i] = earlierTuple[i];
	}
	newTuple[pos] = spec;
	setOfTuples.add(newTuple);
    }


    public String toString()
    {
	return "RMJSignature[" + bodyString() + "]";
    }

    public String bodyString()
    {
	String s = operation().bodyString() + " (";
	Class[] staticTypes = staticTypes();
	Object[] specializers = specializers();
	for (int i = 0; i < specializers.length; i++) {
	    Class staticType = staticTypes[i];
	    Object specializer = specializers[i];
	    if (i > 0) s += ",";
	    if (staticType == null) {
		// a prim type
		s += "<prim>";
	    } else {
		s += staticType.getName();
	    }
	    if (specializer == null) {
		// unspecialized in this position
	    } else if (specializer instanceof Class) {
		// a regular specializer
		s += "@" + ((Class)specializer).getName();
	    } else {
		// a value specializer
		s += "@@" + specializer.toString();
	    }
	}
	s += "), " +
	    (isAbstract() ? "abstract" : "concrete") + 
	    ", kind=" + kind() + 
	    ", baseAnchor=" + baseAnchorClassName() +
	    ", dispatcher=" + dispatcherClassName() +
	    ", ambigError=" + ambigErrorClassName();
	return s;
    }


    // ACCESSORS

    public RMJOperation operation()
    {
	return _op;
    }

    public String opName()
    {
	return _op.opName();
    }

    public Class[] staticTypes()
    {
	return _staticTypes;
    }

    public Object[] specializers()
    {
	return _specializers;
    }

    public Object[] specializersOrTypes()
    {
	return _specializersOrTypes;
    }

    public boolean hasInterfaceSpecializer()
    {
	return _hasInterfaceSpecializer;
    }

    public boolean isAbstract() {
	return _isAbstract;
    }
    public boolean isConcrete() {
	return !isAbstract();
    }

    public String kind()
    {
	return _kind;
    }
    public boolean isBaseAnchor() {
	return _kind.equals("B");
    }
    public boolean isSubAnchor() {
	return _kind.equals("S");
    }
    public boolean isGlue() {
	return _kind.equals("G");
    }

    public String baseAnchorClassName() {
	return _baseAnchorClassName;
    }

    public String dispatcherClassName()
    {
	return _dispatcherClassName;
    }

    public String ambigErrorClassName()
    {
	return _ambigErrorClassName;
    }

    public boolean isInternalGF() {
	return _ambigErrorClassName.equals("-");
    }

    // INSTANCE VARIABLES

    private RMJOperation _op;

    // for each of the receiver and argument positions, either:
    //   null (if this position has a prim static type)
    //   a Class (if this position has a class static type)
    private Class[] _staticTypes;

    // for each of the receiver and argument positions, either:
    //   null (if this position is not specialized),
    //   a Class (if this position is specialized for a particular class),
    //   an Integer, or some other boxed value (if this position uses 
    //                                          value (@@) specialization)
    private Object[] _specializers;

    // for each of the receiver and argument positions, either:
    //   the specializer (if this position is specialized)
    //   the static type (if not)
    private Object[] _specializersOrTypes;

    // whether any of the specializers are interfaces
    private boolean _hasInterfaceSpecializer;

    private boolean _isAbstract;
    private String _kind;
    private String _baseAnchorClassName;
    private String _dispatcherClassName;
    private String _ambigErrorClassName;
}
