package org.multijava.relaxed.util;

import org.multijava.util.compiler.TokenReference;
import org.multijava.mjc.*;

import org.multijava.relaxed.util.classfile.ClassInfo;

import java.io.File;
import java.util.Vector;
import java.util.List;
import java.util.Iterator;

public class RMJAnnotation
{
    /* build an empty annotation object */
    public RMJAnnotation()
    {
    }

    /* build an annotation object from the annotation strings embedded
       in a buffer */
    public RMJAnnotation(byte[] buffer)
    {
	parseBuffer(buffer);
    }

    static private class ParseError extends Exception {}

    private void parseBuffer(byte[] buffer)
    {
	ClassInfo info = new ClassInfo();
	info.process(buffer);
	Iterator iter = info.rmjAttributes().iterator();
	while (iter.hasNext()) {
	    String attribute = (String) iter.next();
	    if (RMJDebug.verbose) {
		RMJDebug.out("Parsing annotation {" + attribute + "}");
	    }
	    try {
		parseInterior(attribute);
	    } catch (ParseError e) {
		// ignore this, if it's malformed
		if (RMJDebug.ann || RMJDebug.verbose) {
		    if (RMJDebug.verbose) {
			// already printed the annotation being parsed
			RMJDebug.out("Parse error!");
		    } else {
			RMJDebug.out("Parse error in annotation:\n  {" + 
				     attribute + "}");
		    }
		    e.printStackTrace();
		}
	    }
	}
    }

    private void parseInterior(String buffer)
	throws ParseError
    {
	int pos = 0;
	int end = buffer.length();
	while (pos < end) {
	    if (pos + 3 > end) {
		// not enough room for an annotation => malformed
		throw new ParseError();
	    }
	    char kind = buffer.charAt(pos);
	    if (buffer.charAt(pos + 1) != '[') {
		// no opening bracket
		throw new ParseError();
	    }
	    int close = buffer.indexOf(']', pos + 2);
	    if (close < 0) {
		// no ending bracket; error
		throw new ParseError();
	    }
	    switch (kind) {
	    case 'M':
		// a method annotation
		Method.parseBody(this, buffer, pos+2, close);
		break;
	    default:
		// unknown annotation kind
		throw new ParseError();
	    }
	    pos = close + 1;
	}
    }

    public String getAnnotationString()
    {
	String annotation = "";

	Iterator it = methodItems.iterator();
	while (it.hasNext()) {
	    Method item = (Method) it.next();
	    annotation += item.getAnnotationString();
	}
	return annotation;
    }

    public String toString()
    {
	return methodItems.toString();
    }

    public boolean isEmpty()
    {
	return methodItems.isEmpty();
    }

    public Vector methodItems()
    {
	return methodItems;
    }

    public void add(Method method)
    {
	methodItems.add(method);
    }

    static public class Method
    {
	public Method(String opName,
		      String[] staticTypeDescs, 
		      String[] specializerDescs, 
		      boolean isAbstract, String kind, 
		      String baseAnchorClassName, 
		      String dispatcherClassName, 
		      String ambigErrorClassName) {
	    _opName = opName;
	    _staticTypeDescs = staticTypeDescs;
	    _specializerDescs = specializerDescs;
	    _isAbstract = isAbstract;
	    _kind = kind; 
	    _baseAnchorClassName = baseAnchorClassName;
	    _dispatcherClassName = dispatcherClassName;
	    _ambigErrorClassName = ambigErrorClassName;
	}
	
	// conversion to the encoded form:
	private String getAnnotationString() {
	    String staticTypeDescsString = _staticTypeDescs.length + ":";
	    for (int i = 0; i < _staticTypeDescs.length; i++) {
		if (i > 0) staticTypeDescsString += ",";
		staticTypeDescsString += _staticTypeDescs[i];
	    }
	    String specializerDescsString = _specializerDescs.length + ":";
	    for (int j = 0; j < _specializerDescs.length; j++) {
		if (j > 0) specializerDescsString += ",";
		specializerDescsString += _specializerDescs[j];
	    }
	    return ("M[" +
		    _opName + "|" +
		    staticTypeDescsString + "|" +
		    specializerDescsString + "|" +
		    (_isAbstract ? "a" : "c") + "|" +
		    _kind + "|" +
		    _baseAnchorClassName + "|" +
		    _dispatcherClassName + "|" +
		    _ambigErrorClassName + "]");
	}

	// parsing from the encoded form:
	static public void parseBody(RMJAnnotation annos, 
				     String buffer, int start, int end)
	    throws ParseError
	{
	    int pos = start;
	    int sep;
	    
	    sep = buffer.indexOf('|', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String opName = buffer.substring(pos, sep);
	    pos = sep + 1;

	    sep = buffer.indexOf(':', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String numStaticTypeDescsString = buffer.substring(pos, sep);
	    int numStaticTypeDescs;
	    try {
		numStaticTypeDescs =
		    Integer.parseInt(numStaticTypeDescsString);
	    } catch (NumberFormatException e) {
		throw new ParseError();
	    }
	    pos = sep + 1;

	    String[] staticTypeDescs = new String[numStaticTypeDescs];
	    for (int i = 0; i < numStaticTypeDescs; i++) {
		char endChar = (i == numStaticTypeDescs-1) ? '|' : ',';
		sep = buffer.indexOf(endChar, pos);
		if (sep < 0 || sep > end) throw new ParseError();
		String staticTypeDesc = buffer.substring(pos, sep);
		staticTypeDescs[i] = staticTypeDesc;
		pos = sep + 1;
	    }
	    
	    sep = buffer.indexOf(':', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String numSpecsString = buffer.substring(pos, sep);
	    int numSpecs;
	    try {
		numSpecs = Integer.parseInt(numSpecsString);
	    } catch (NumberFormatException e) {
		throw new ParseError();
	    }
	    pos = sep + 1;

	    String[] specializerDescs = new String[numSpecs];
	    for (int i = 0; i < numSpecs; i++) {
		char endChar = (i == numSpecs-1) ? '|' : ',';
		sep = buffer.indexOf(endChar, pos);
		if (sep < 0 || sep > end) throw new ParseError();
		String specializerDesc = buffer.substring(pos, sep);
		specializerDescs[i] = specializerDesc;
		pos = sep + 1;
	    }
	    
	    sep = buffer.indexOf('|', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String isAbstractString = buffer.substring(pos, sep);
	    boolean isAbstract;
	    if (isAbstractString.equals("a")) {
		isAbstract = true;
	    } else if (isAbstractString.equals("c")) {
		isAbstract = false;
	    } else {
		throw new ParseError();
	    }
	    pos = sep + 1;
	    
	    sep = buffer.indexOf('|', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String kind = buffer.substring(pos, sep);
	    pos = sep + 1;
	    
	    sep = buffer.indexOf('|', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String baseAnchorClassName = buffer.substring(pos, sep);
	    pos = sep + 1;
	    
	    sep = buffer.indexOf('|', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String dispatcherClassName = buffer.substring(pos, sep);
	    pos = sep + 1;
	    
	    sep = buffer.indexOf(']', pos);
	    if (sep < 0 || sep > end) throw new ParseError();
	    String ambigErrorClassName = buffer.substring(pos, sep);
	    pos = sep + 1;

	    if (pos != end+1) throw new ParseError();

	    // add the new method item
	    annos.add(new Method(opName, staticTypeDescs, 
				 specializerDescs, isAbstract, kind,
				 baseAnchorClassName, 
				 dispatcherClassName, 
				 ambigErrorClassName));
	}

	public String toString() 
	{ 
	    return getAnnotationString();
	}

	// accessors:

	public String opName()
	{
	    return _opName;
	}

	public String[] staticTypeDescs()
	{
	    return _staticTypeDescs;
	}

	public String[] specializerDescs()
	{
	    return _specializerDescs;
	}

	public boolean isAbstract() {
	    return _isAbstract;
	}

	public String kind()
	{
	    return _kind;
	}

	public boolean isBaseAnchor() {
	    return _kind.equals("B");
	}
	public boolean isSubAnchor() {
	    return _kind.equals("S");
	}
	public boolean isGlue() {
	    return _kind.equals("G");
	}

	// if not in a base anchor, record that this method was in a
	// glue anchor, not sub-anchor
	public void foundInGlueAnchor() {
	    if (isSubAnchor()) {
		_kind = "G";
	    }
	}

	public String baseAnchorClassName() {
	    return _baseAnchorClassName;
	}

	public String dispatcherClassName()
	{
	    return _dispatcherClassName;
	}

	public String ambigErrorClassName()
	{
	    return _ambigErrorClassName;
	}

	private String _opName;
	private String[] _staticTypeDescs;
	private String[] _specializerDescs;
	private boolean _isAbstract;
	private String _kind;
	private String _baseAnchorClassName;
	private String _dispatcherClassName;
	private String _ambigErrorClassName;
    }

    private Vector methodItems = new Vector();
}
