/*
 * Copyright (c) 2002, Iowa State University
 */

package org.multijava.samples.cmb.parser;

import java.io.PrintWriter;
import java.io.File;
import java.util.Date;
import java.io.FileNotFoundException;

/**
 * This class represents an entire compilation unit in a VST.  An
 * instance of this class is the root node of every complete
 * well-formed VST.
 * @author Curtis Clifton
 * @version $Id: CompilationUnit.java,v 1.2 2002/06/25 17:08:29 cclifton Exp $ 
 */
public class CompilationUnit extends Node {

    private static boolean debugOn = true;

    // -------------------------------------------------------------
    // CONSTRUCTORS
    // -------------------------------------------------------------
    
    /**
     * Instantiate a compilation unit node in a VST with the given
     * data.  Used by parser to build VST from a Reader.  Client
     * should subsequently call setFile and setTimestamp to complete
     * the instance's state.  
     *
     * @see #setFile
     * @see #setTimestamp
     */
    public CompilationUnit( PackageClause packageClause,
			    ImportClause[] imports, TypeDecl[] types,
			    String postText ) {
	this.packageClause = packageClause;
	this.imports = imports == null ? new ImportClause[0] : imports;
	this.types = types == null ? new TypeDecl[0] : types;
	this.postText = postText;
	this.kind = COMPILATION_UNIT;
    }

    /**
     * Instantiate a compilation unit node in a VST with the given
     * data.  */
    public CompilationUnit( File file, long timestamp, 
			    PackageClause packageClause,
			    ImportClause[] imports, TypeDecl[] types,
			    String postText ) {
	this( packageClause, imports, types, postText );
	this.file = file;
	this.timestamp = timestamp;
    }

    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    /**
     * Sets the value of <CODE>file</CODE>
     *
     * @param argFile Value to assign to this.file
     */
    public void setFile(File argFile){
	this.file = argFile;
    }


    /**
     * Returns the value of <CODE>file</CODE>.
     */
    public File file() {	//  !FIXME! once addToIndices is moved to this package make this method package-private
	return file;
    }

    /**
     * Sets the value of timestamp
     *
     * @param argTimestamp Value to assign to this.timestamp
     */
    public void setTimestamp(long argTimestamp){
	this.timestamp = argTimestamp;
    }

    /**
     * Returns the type declaration VST node inside this compilation
     * unit having the given name, or null if no such declaration
     * exists */
    public TypeDecl typeDeclFor(String typeName) {
	//  !FIXME!inners This should recurse into each types[i].
	for (int i = 0; i < types.length; i++) {
	    if (typeName.equals(types[i].name())) {
		return types[i];
	    }
	}
	return null;
    }

    /**
     * Returns a string representing the package to which this
     * compilation unit belongs.  The string is a series 0 or more
     * Java identifiers, separated with "dots".  For a compilation
     * unit without a package clause the result is the empty string
     * (i.e., a series of 0 identifiers).  */
    public String packageName() {
	if (packageClause==null) {
	    return "";
	} else {
	    return packageClause.name();
	}
	
    }

    /**
     * Returns a string representing the short name of the first type
     * declared in this compilation unit.    */
    public String firstTypeName() {
	if (types.length==0) {
	    return "";
	} else {
	    return types[0].name();
	}
	
    }

    /**
     * Returns the package clause of this, or null if none.  This
     * method is package protected because it exposes the
     * representation of this. */
    PackageClause packageClause() {
	return packageClause;
    }

    /**
     * Returns the import clauses of this.  This method is package
     * protected because it exposes the representation of this. */
    ImportClause[] imports() {
	return imports;
    }

    /**
     * Returns the type declarations of this.  This method is package
     * protected because it exposes the representation of this. */
    public TypeDecl[] types() {	//  !FIXME! once addToIndices is moved to this package make this method package-private
	return types;
    }

    /**
     * Two compilation units always compare as equal since only one
     * compilation unit may be declared in the same VST.  
     * @see Node#equals
     */
    public boolean equals(Object other) {
	return other instanceof CompilationUnit;
    }

    public int hashCode() {
	return 13;		// any constant will do, see equals()
    }

    // -------------------------------------------------------------
    // PRETTY PRINTING
    // -------------------------------------------------------------

    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * VST anchored at this.  */
    public void prettyPrint(PrintWriter wrtr) {
	if (packageClause != null) {
	    packageClause.prettyPrint(wrtr);
	}

	for (int i = 0; i < imports.length; i++) {
	    imports[i].prettyPrint(wrtr);
	}
	
	for (int i = 0; i < types.length; i++) {
	    types[i].prettyPrint(wrtr);
	}
	
	wrtr.print(postText);
    }

    // -------------------------------------------------------------
    // ABSTRACT REPRESENTATION OUTPUT
    // -------------------------------------------------------------

    /**
     * Appends an abstract representation of this node onto the given
     * accumulator buffer.
     *
     * @see Node#abstractRep()
     */
    protected void abstractRepBuilder(StringBuffer accum) {
	accum.append("[CUn");
	if (packageClause != null) {
	    packageClause.abstractRepBuilder(accum);
	}
	for (int i = 0; i < imports.length; i++) {
	    imports[i].abstractRepBuilder(accum);
	}
	for (int i = 0; i < types.length; i++) {
	    types[i].abstractRepBuilder(accum);
	}
	accum.append(']');
    }

    // -------------------------------------------------------------
    // MUTATION
    // -------------------------------------------------------------
    
    // -------------------------------------------------------------
    // REVERSE ENGINEERING
    // -------------------------------------------------------------

    /**
     * Parses the source code corresponding to this VST and returns a
     * a new VST corresponding to the current state of the source
     * code.  */
    public CompilationUnit getNewVSTFromSource() {
	final String MSG = "Cannot read new source code from file " + file;
	try {
	    if (file.lastModified() <= timestamp) {
		// no change to file since last parsing
		return this;
	    }

	    return VSTBuilder.parse(file);
	} catch (FileNotFoundException e) {
	    throw new IllegalStateException(MSG); //  !FIXME!3/9
	} catch (VSTBuilder.SyntaxErrorException e) {
	    throw new IllegalStateException("Syntax error parsing file " + 
					    file +
					    ":\n" + e.getMessage());	//  !FIXME!3/9
	} 
    }

    // -------------------------------------------------------------
    // PRIVILEGED DATA MEMBERS
    // -------------------------------------------------------------

    /**
     * The file corresponding to this compilation unit, or null if not
     * known.  */
    private File file;

    /**
     * The time at which this VST was built.  If manually constructed
     * the timestamp should match the construction time.  If built
     * from a source code file the timestamp should match the date
     * stamp on the file.  */
    private long timestamp;

    /**
     * The package of this compilation unit, or null if not given.  */
    private PackageClause packageClause;

    /**
     * The imports of this compilation unit.
     *
     * <pre><jml>
     * invariant imports != null;
     * </jml></pre>
     *
     */
    private ImportClause[] imports;

    /**
     * The types declared in this compilation unit.
     *
     * <pre><jml>
     * invariant types != null;
     * </jml></pre>
     *
     */
    private TypeDecl[] types;

    /**
     * The text following all the types declared in this compilation
     * unit.  */
    private String postText;

}
