/*
 * Copyright (c) 2002, Iowa State University
 */

package org.multijava.samples.cmb.parser;

import java.io.PrintWriter;

/**
 * This is the abstract superclass of all VST nodes representing
 * members in Java.  */
public abstract class Member extends Node implements Javadocable {

    /**
     * Generate a new class declaration node for a VST.
     * @param preText source code text before any SREAnnotations
     * @param midText source code text between any SREAnnotations and the
     *			first child node text
     * @param postText source code text following the last child node
     * @param stereotype a <code>StereotypeAnnotation</code> value, or null
     * @param modifiers the modifiers of this
     * @param name the short name of this member
     */
    public Member( String preText, String midText, String postText, 
		   StereotypeAnnotation stereotype, Modifiers modifiers, 
		   String name ) 
    {
	this.preText = preText;
	this.midText = midText;
	this.postText = postText;
	this.stereotype = stereotype;
	this.modifiers = modifiers;
	this.name = name;
    }
    
    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    /**
     * Returns the modifiers of this member. */
    Modifiers modifiers() {
	return modifiers;
    }

    /**
     * Sets the value of postText.
     *
     * <pre><jml>
     * requires this.postText == null;
     * ensures this.postText == postText;
     * </jml></pre>
     *
     */
    void setPostText(String postText) {
	if (this.postText == null) {
	    this.postText = postText;
	}
    }


    /**
     * Returns the name of this member.  */
    public String name() {
	return name;
    }
    

    /**
     * Returns the text preceding any Javadoc tags for this member.  */
    public String preText() {
	return preText;
    }

    /**
     * Returns the text following any Javadoc tags but before the
     * first child node of this member.  */
    public String midText() {
	return midText;
    }


    /**
     * Returns the text following any child node of this member.  */
    public String postText() {
	return postText;
    }

    // -------------------------------------------------------------
    // PRETTY-PRINTING
    // -------------------------------------------------------------

    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * VST anchored at this.  */
    public void prettyPrint(PrintWriter wrtr) {
	wrtr.print(preText);
	if (stereotype != null) {
	    stereotype.prettyPrint(wrtr);
	}
	prettyPrintAnnotations(wrtr);
	wrtr.print(midText);
	prettyPrintChildren(wrtr);
	wrtr.print(postText);
    }

    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * annotation child nodes of this, excluding the common stereotype
     * annotation.  The default implementation here does nothing, some
     * subclasses will override this implementation. */
    void prettyPrintAnnotations(PrintWriter wrtr) {
    }

    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * non-annotation child nodes of this. */
    abstract void prettyPrintChildren(PrintWriter wrtr);


    // -------------------------------------------------------------
    // ABSTRACT REPRESENTATION OUTPUT
    // -------------------------------------------------------------

    /**
     * Appends an abstract representation of this node onto the given
     * accumulator buffer.
     *
     * @see Node#abstractRep()
     */
    protected void abstractRepBuilder(StringBuffer accum) {
	accum.append('[');
	accum.append(codon());
	accum.append(':');
	accum.append(name);
	accum.append(':');
	buildAttributeReps(accum);
	if (stereotype != null) {
	    stereotype.abstractRepBuilder(accum);
	}
	buildAnnotationReps(accum);
	buildChildrenReps(accum);
	accum.append(']');
    }

    /**
     * Appends to <code>accum</code> an abstract representation of the 
     * attributes of this. */
    protected void buildAttributeReps(StringBuffer accum) {
    }

    /**
     * Appends to <code>accum</code> an abstract representation of the 
     * annotation child nodes of this, excluding the common stereotype
     * annotation.  The default implementation here does nothing, some
     * subclasses will override this implementation. */
    protected void buildAnnotationReps(StringBuffer accum) {
    }

    /**
     * Appends to <code>accum</code> an abstract representation of the
     * non-annotation child nodes of this. */
    protected abstract void buildChildrenReps(StringBuffer accum);


    /**
     * Returns an identifier for this node type.  */
    protected abstract String codon();

    // -------------------------------------------------------------
    // PRIVILEDGED DATA MEMBERS
    // -------------------------------------------------------------

    /**
     * Source code text preceding any SRE-specific javadoc
     * annotations.  */
    private String preText;

    /**
     * Source code text between any SRE-specific javadoc annotations
     * and the text associated with the first child node of this.  */
    private String midText;

    /**
     * Source code text following the text of the last child node of
     * this.  */
    private String postText;

    /**
     * Any stereotype annotation associate with this member, or null
     * if none.  */
    private StereotypeAnnotation stereotype;

    /**
     * The modifiers of this.  */
    private Modifiers modifiers;

    /**
     * The identifier of this member.
     */
    private String name;
}
