/*
 * Copyright (c) 2002, Iowa State University
 */

package org.multijava.samples.cmb.parser;

import java.io.PrintWriter;
import org.multijava.samples.cmb.parser.*;
import java.util.Arrays;


/**
 * This class represents a constructor in a VST.
 *
 * @author Curtis Clifton
 * @version $Id: MethodDecl.java,v 1.2 2002/06/25 17:08:29 cclifton Exp $
 */
public class MethodDecl extends Member {

    // -------------------------------------------------------------
    // CONSTRUCTORS AND FACTORIES
    // -------------------------------------------------------------

    /**
     * Construct a VST node representing a method declaration.
     *
     * @param preText text preceding any SRE-specific javadoc annotations
     * @param midText text between any SRE-specific and the first left paren
     * @param postText text from the end of the last parameter name to the 
     *			end of the line following the closing right brace
     * @param stereotype a stereotype for this, or null
     * @param modifiers the constructor's modifiers
     * @param name 	the constructor's identifier
     * @param inputAnnotation an input annotation for this, or null
     * @param outputAnnotation an output annotation for this, or null
     * @param params an array of the parameters of this
     * @param returnType the return type of this
     * @param thrown an array of the names of the exceptions declared to be 
     *			thrown by this
     */
    public MethodDecl( String preText, String midText, String postText, 
		       StereotypeAnnotation stereotype, Modifiers modifiers, 
		       String name, TextValuedAnnotation inputAnnotation,
		       TextValuedAnnotation outputAnnotation, 
		       Parameter[] params, String returnType, 
		       String[] thrown ) 
    {
	super( preText, midText, postText, stereotype, modifiers, name );
	this.inputAnnotation = inputAnnotation;
	this.outputAnnotation = outputAnnotation;
	this.params = params == null ? new Parameter[0] : params;
	this.returnType = returnType;
	this.thrown = thrown == null ? new String[0] : thrown;
    }

    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    public boolean equals(Object other) {
	if (other instanceof MethodDecl) {
	    MethodDecl o = (MethodDecl) other;
	    return name().equals( o.name() ) &&
		Arrays.equals( params, o.params );
	} else {
	    return false;
	}
    }

    public int hashCode() {
	int result = name().hashCode();
	for (int i = 0; i < params.length; i++) {
	    result += params.hashCode();
	}
	return result;
    }

    // -------------------------------------------------------------
    // PRETTY-PRINTING
    // -------------------------------------------------------------

    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * annotation child nodes of this, excluding the common stereotype
     * annotation.  The default implementation here does nothing, some
     * subclasses will override this implementation. */
    void prettyPrintAnnotations(PrintWriter wrtr) {
	if (inputAnnotation != null) {
	    inputAnnotation.prettyPrint(wrtr);
	}
	if (outputAnnotation != null) {
	    outputAnnotation.prettyPrint(wrtr);
	}
    }
    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * non-annotation child nodes of this. */
    void prettyPrintChildren(PrintWriter wrtr) {
	for (int i = 0; i < params.length; i++) {
	    params[i].prettyPrint(wrtr);
	}
    }

    // -------------------------------------------------------------
    // ABSTRACT REPRESENTATION OUTPUT
    // -------------------------------------------------------------

    /**
     * Appends to <code>accum</code> an abstract representation of the 
     * annotation child nodes of this, excluding the common stereotype
     * annotation.  The default implementation here does nothing, some
     * subclasses will override this implementation. */
    protected void buildAnnotationReps(StringBuffer accum) {
	if (inputAnnotation != null) {
	    inputAnnotation.abstractRepBuilder(accum);
	}
	if (outputAnnotation != null) {
	    outputAnnotation.abstractRepBuilder(accum);
	}
    }

    /**
     * Appends to <code>accum</code> an abstract representation of the
     * non-annotation child nodes of this. */
    protected void buildChildrenReps(StringBuffer accum) {
	for (int i = 0; i < params.length; i++) {
	    params[i].abstractRepBuilder(accum);
	}
    }

    /**
     * Returns an identifier for this node type.  */
    protected String codon() { return "Mth"; }

    // -------------------------------------------------------------
    // REVERSE ENGINEERING
    // -------------------------------------------------------------

    // -------------------------------------------------------------
    // PRIVILEDGED DATA MEMBERS
    // -------------------------------------------------------------

    /**
     * The annotation giving the parameters of input kind.
     */
    private TextValuedAnnotation inputAnnotation;

    /**
     * The annotation giving the parameters of output kind.
     */
    private TextValuedAnnotation outputAnnotation;

    /**
     * The parameters of this.
     *
     * <pre><jml>
     * invariant params != null;
     * </jml></pre>
     *
     */
    private Parameter[] params;

    /**
     * The return type of this.
     */
    private String returnType;


    /**
     * The names of the exceptions declared to be thrown by this.  */
    private String[] thrown;
}
