/*
 * Copyright (c) 2002, Iowa State University
 */

package org.multijava.samples.cmb.parser;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;

/**
 * This is the abstract superclass of every verbatim syntax tree node.
 * @author Curtis Clifton
 * @version $Id: Node.java,v 1.4 2007/02/08 14:04:56 leavens Exp $
 */
public abstract class Node implements NodeKinds, Comparable {

    protected static final boolean debugOn = true;

    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    /**
     * Returns a string that is the result of prettyprinting the VST
     * rooted at this node.  
     *
     * @see #prettyPrint(java.io.PrintWriter)
     */
    public String toString() {
	StringWriter stWr = new StringWriter();
	PrintWriter wrtr = new PrintWriter( stWr );
	prettyPrint( wrtr );
	wrtr.close();
	return stWr.toString();
    }

    public int compareTo(/*@ non_null @*/ Object other) {
	Node o = (Node) other;
	if (this.kind != o.kind) {
	    return this.kind - o.kind;
	} else {
	    if (this.equals(o)) {
		return 0;
	    } else {
		return (this.hashCode() < o.hashCode()) ? -1 : 1;
	    }
	}
    }

    /**
     * <P>Returns <CODE>true</CODE> if <CODE>this</CODE> and
     * <CODE>other</CODE> are both VST nodes and <CODE>this</CODE>
     * should be considered a possibly edited version of
     * <CODE>other</CODE> for reverse engineering purposes.</P>
     *
     * <P>Compares <CODE>this</CODE> to <CODE>other</CODE> returning
     * true if the two VST instances being compared cannot legally
     * appear as children of the same parent VST node (i.e., they are
     * enough the same as to create ambiguity in Java), or false
     * otherwise.</P>
     *
     * <P>For example two FieldDecl nodes with the same name will
     * compare as equal.  However, two MethodDecl nodes with the same
     * name will only compare as equal if they have the same number of
     * parameters and the parameters are of the same types.</P> 
     * */
    public abstract boolean equals(Object other);

    /**
     * <pre><jml>
     * ensures (\forall Node n1;;
     *		(\forall Node n2;;
     *			n1.equals(n2) ==> n1.hashCode() == n2.hashCode()));
     * </jml></pre>
     *
     */
    public abstract int hashCode();

    // -------------------------------------------------------------
    // PRETTY PRINTING
    // -------------------------------------------------------------

    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * VST anchored at this.  */
    public abstract void prettyPrint(PrintWriter wrtr);

    // -------------------------------------------------------------
    // ABSTRACT REPRESENTATION OUTPUT
    // -------------------------------------------------------------

    /**
     * Returns a string representing the structure of the tree, useful
     * for testing and debugging purposes.  Nodes in the tree are
     * represented by the string
     * &quot;<code>[</code><em>XXXch1ch2...</em><code>]</code>&quot;
     * where XXX is a three letter abbreviation for the node type and
     * ch1, ch2, etc... are representations for the child nodes or
     * attributes. */
    public String abstractRep() {
	StringBuffer accum = new StringBuffer();
	abstractRepBuilder(accum);
	return accum.toString();
    }


    /**
     * Appends an abstract representation of this node onto the given
     * accumulator buffer.
     *
     * @see #abstractRep()
     */
    protected abstract void abstractRepBuilder(StringBuffer accum);

    // -------------------------------------------------------------
    // REVERSE ENGINEERING
    // -------------------------------------------------------------

    // -------------------------------------------------------------
    // PRIVILEGED DATA
    // -------------------------------------------------------------

    /**
     * An integer indicating the type of this object.  Used to avoid
     * instanceof checks and make switch statements more readable.
     * Package protected to allow representation exposure for
     * efficiency.  
     * @see NodeKinds
     */
    int kind;
}

