/*
 * Copyright (c) 2002, Iowa State University
 */

package org.multijava.samples.cmb.parser;

import java.util.HashMap;

/**
 * This class is the abstract superclass of all VST nodes representing
 * SRE-specific javadoc annotations.
 * @author Curtis Clifton
 * @version $Id: SREAnnotation.java,v 1.3 2002/06/25 17:08:29 cclifton Exp $
 */
public abstract class SREAnnotation extends SimpleNode {

    // -------------------------------------------------------------
    // CONSTRUCTORS
    // -------------------------------------------------------------

    /**
     * Generates an SRE-specific javadoc annotation node for a VST.
     * @param text the verbatim text of this annotation starting with the 
     *			beginning of the line containing the annotation and 
     *			continuing to the trailing newline
     * @param kind an integer describing the sort of annotation this represents
     */
    public SREAnnotation( String text, int kind ) {
	super(text);
	this.kind = kind;
    }

    // -------------------------------------------------------------
    // STATIC UTILITY METHODS AND FACTORY
    // -------------------------------------------------------------

    /**
     * Returns true if the given javadoc tag is one that indicates
     * SRE-specific information.  */
    public static boolean careAbout(String tag) {
	return kindMap.containsKey(tag);
    }


    /**
     * Constructs an appropriate concrete subclass of this for the
     * given arguments.
     * @param tag	the javadoc tag for the annotation, including leading
     *			<code>@</code> sign
     * @param value 	the value of the annotation, from the first 
     *			non-whitespace character after the tag to the
     *			newline preceding the start of the next tag or the 
     *			end of the javadoc comment, omitting whitespace and
     *			asterisks at the beginning of lines
     * @param verbatim	the verbatim text of this annotation, from the 
     *			beginning of the line containing <code>tag</code> 
     *			through the newline (if any) following the end 
     *			of <code>value</code>
     * @return an appropriate concrete subclass of this
     */
    public static SREAnnotation buildAnnotation( String tag,
						 String value,
						 String verbatim ) {
	int kind = ((Integer) kindMap.get(tag)).intValue();
	switch (kind) {
	case STEREOTYPE:
	    return new StereotypeAnnotation( verbatim, value );
	    
	default:
	    return new TextValuedAnnotation( verbatim, value, kind );
	}
    }
    
    // -------------------------------------------------------------
    // ABSTRACT REPRESENTATION OUTPUT
    // -------------------------------------------------------------

    /**
     * Returns an identifier for this node type.  */
    protected String codon() {
	return (String) codonMap.get(new Integer(kind));
    }

    // -------------------------------------------------------------
    // REVERSE ENGINEERING
    // -------------------------------------------------------------

    // -------------------------------------------------------------
    // PRIVILEDGED DATA MEMBERS
    // -------------------------------------------------------------

    /**
     * An index indicating the sort of annotation represented by this.  */
    public final int kind;	

    /**
     * Describe constant <code>STEREOTYPE</code> here.
     *
     */
    public static final int STEREOTYPE = 0;
    /**
     * Describe constant <code>INVARIANT</code> here.
     *
     */
    public static final int INVARIANT = 1;
    /**
     * Describe constant <code>INPUT</code> here.
     *
     */
    public static final int INPUT = 2;
    /**
     * Describe constant <code>OUTPUT</code> here.
     *
     */
    public static final int OUTPUT = 3;

    /**
     * Maps constant annotation 'kind' values to three character
     * codons for use in abstract representation.  */
    private static final HashMap codonMap = new HashMap();

    /**
     * Maps the javadoc tags that indicate SRE-specific
     * information to their 'kind' values.  */
    private static final HashMap kindMap = new HashMap();

    static {
	//  !FIXME! for CMB purposes we will fail to search any
	//  comments appearing after these tags
	addInfo( "@stereotype",    STEREOTYPE,    "Seo" );
	addInfo( "@invariant",     INVARIANT,     "Inv" );
	addInfo( "@input",         INPUT,         "Inp" );
	addInfo( "@output",        OUTPUT,        "Out" );
    }

    private static void addInfo( String tag, int kind, String codon ) {
	Integer kBox = new Integer(kind);
	kindMap.put(tag,kBox);
	codonMap.put(kBox, codon);
    }
}
