/*
 * Copyright (C) 2002 Iowa State University
 *
 * $Id: TextAndAnnotations.java,v 1.3 2002/12/20 03:08:13 leavens Exp $
 */

package org.multijava.samples.cmb.parser;

import java.util.ArrayList;


/**
 * This simple class serves as a data structure for communicating
 * between the Javadoc and Java parsers.  It collects the doc comments
 * parsed from a javadoc style annotation and provides methods to
 * access the text preceding the first SRE-specific annotation in a
 * javadoc comment, the SRE-specific annotations, and the text up to
 * the end of the comment.  Methods are included for accessing the
 * annotations by kind.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.3 $ */

class TextAndAnnotations {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Creates an instance of this class for accumulating items parsed
     * from a doc comment.  */
    TextAndAnnotations() {
	this.preText = new StringBuffer();
	this.notes = new ArrayList();
	this.postText = new StringBuffer();
    }
    
    // -------------------------------------------------------------
    // ACCUMULATION
    // -------------------------------------------------------------

    /**
     * Appends the given verbatim text to the accumulated verbatim
     * text preceding the first tag in the doc comment.  Verbatim text
     * includes whitespace and stars.  */
    void addVerbatimPreText( String text ) {
	preText.append(text);
    }

    /**
     * Appends the given verbatim text to the accumulated verbatim
     * text following the end of the last paragraph of the comment up
     * to the newline following the star-slash at the end of the
     * comment.  Verbatim text includes whitespace and stars.  */
    void addVerbatimPostText( String text ) {
	postText.append(text);
    }

    /**
     * Begins the accumulation of a tagged paragraph for the given tag.
     *
     * <pre><jml>
     * requires !inTaggedRegion;
     * ensures inTaggedRegion;
     * </jml></pre>
     *
     * @see #addTaggedText
     * @see #endTaggedRegion
     */
    void beginTaggedRegion( String tag ) throws IllegalStateException {
	if (inTaggedRegion) {
	    throw new IllegalStateException();
	}
	inTaggedRegion = true;
	careAboutTag = SREAnnotation.careAbout(tag);
	if (careAboutTag) {
	    currentTag = tag;
	    currentDocumentation = new StringBuffer();
	}
    }


    /**
     * Appends the documentation (non-verbatim) text to the current tag.
     *
     * <pre><jml>
     * requires inTaggedRegion;
     * </jml></pre>
     *
     * @see #beginTaggedRegion
     * @see #endTaggedRegion
     */
    void addTaggedText( String documentation ) throws IllegalStateException {
	if (!inTaggedRegion) {
	    throw new IllegalStateException();
	}
	if (careAboutTag) {
	    currentDocumentation.append(documentation);
	}
    }

    /**
     * Ends the accumulation of a tagged paragraph by recording the
     * verbatim text from the beginning of the line containing the tag
     * to the end of the line preceding the next tag or end of
     * comment.  
     *
     * <pre><jml>
     * requires inTaggedRegion;
     * ensures !inTaggedRegion;
     * </jml></pre>
     *
     * @see #beginTaggedRegion
     * @see #addTaggedText
     */
    void endTaggedRegion( String verbatim ) throws IllegalStateException {
	if (!inTaggedRegion) {
	    throw new IllegalStateException();
	}
	if (careAboutTag) {
	    notes.add(SREAnnotation.
		      buildAnnotation(currentTag, 
				      currentDocumentation.toString(), 
				      verbatim));
	} else {
	    preText.append(verbatim);
	}
	
	inTaggedRegion = false;
    }

    //---------------------------------------------------------------------
    // DERIVED ATTRIBUTES
    //---------------------------------------------------------------------

    /**
     * Returns the accumulated verbatim text preceding the first tag
     * in the doc comment.  Verbatim text includes whitespace and
     * stars.  */
    public String preText() {
	return preText.toString();
    }

    /**
     * Returns the accumulated verbatim text following the end of the
     * last paragraph of the comment up to the newline following the
     * star-slash at the end of the comment.  Verbatim text includes
     * whitespace and stars.  */
    public String postText() {
	return postText.toString();
    }

    /**
     * Returns the stereotype annotation for this, or null if one
     * isn't found.  */
    public StereotypeAnnotation stereotype() {
	return (StereotypeAnnotation) 
	    findFirstOfKind(SREAnnotation.STEREOTYPE);
    }

    /**
     * Returns an array of the invariant annotations for this, or an
     * empty array if none are found.  */
    public TextValuedAnnotation[] invariants() {
	ArrayList accum =  findAllOfKind(SREAnnotation.INVARIANT);
	return (TextValuedAnnotation[]) 
	    accum.toArray(new TextValuedAnnotation[accum.size()]);
    }

    /**
     * Returns the input annotation for this, or null if one
     * isn't found.  */
    public TextValuedAnnotation input() {
	return (TextValuedAnnotation) 
	    findFirstOfKind(SREAnnotation.INPUT);
    }

    /**
     * Returns the output annotation for this, or null if one
     * isn't found.  */
    public TextValuedAnnotation output() {
	return (TextValuedAnnotation) 
	    findFirstOfKind(SREAnnotation.OUTPUT);
    }

    /**
     * Returns the first annotation of the given kind for this, or
     * null if one isn't found.  */
    private SREAnnotation findFirstOfKind(int kind) {
	for (int i = 0; i < notes.size(); i++) {
	    SREAnnotation note = (SREAnnotation) notes.get(i);
	    if (note.kind == kind) {
		return note;
	    }
	}
	return null;
    }

    /**
     * Returns all the annotations of the given kind for this, or
     * an empty array if none are found.  */
    private ArrayList findAllOfKind(int kind) {
	ArrayList accum = new ArrayList(notes.size());
	for (int i = 0; i < notes.size(); i++) {
	    SREAnnotation note = (SREAnnotation) notes.get(i);
	    if (note.kind == kind) {
		accum.add(note);
	    }
	}
	return accum;
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    public final StringBuffer preText;
    public final StringBuffer postText;

    private final ArrayList notes;

    /**
     * Indicates whether text is being accumulated for a tag.  */
    private boolean inTaggedRegion = false;

    /**
     * Indicates whether the current tag for which text is being
     * accumulated is one that we care about recording.  */
    private boolean careAboutTag = false;

    /**
     * Tag for which text is being accumulated. */
    private String currentTag;

    /**
     * Accumulated non-verbatim documentation for the current tag.
     *
     * @see #currentTag
     */
    private StringBuffer currentDocumentation;
}
