/*
 * Copyright (c) 2002, Iowa State University
 */

package org.multijava.samples.cmb.parser;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

public abstract class TypeDecl extends Member {

    // -------------------------------------------------------------
    // CONSTRUCTORS
    // -------------------------------------------------------------

    /**
     * Generate a new class declaration node for a VST.
     * @param preText source code text before any SREAnnotations
     * @param midText source code text between any SREAnnotations and the
     *			first curly brace of the body
     * @param postText source code text following the last member
     * @param stereotype a <code>StereotypeAnnotation</code> value, or null
     * @param modifiers the modifiers of this
     * @param name the short name of this class
     * @param invariants the invariants of this class
     * @param members the members of this class
     * @param interfaces the interfaces implemented by this class
     */
    public TypeDecl( String preText, String midText, String postText, 
		     StereotypeAnnotation stereotype, Modifiers modifiers, 
		     String name, TextValuedAnnotation[] invariants, 
		     List members, String[] interfaces ) {
	super( preText, midText, postText, stereotype, modifiers, name );
	this.invariants = invariants != null ? 
	    invariants : new TextValuedAnnotation[0];
	this.members = members != null ?
	    members : new ArrayList();
	this.interfaces = interfaces != null ?
	    interfaces : new String[0];
    }

    // -------------------------------------------------------------
    // ACCESSORS
    // -------------------------------------------------------------

    public boolean equals(Object other) {
	return other instanceof TypeDecl &&
	    name().equals( ((TypeDecl)other).name() );
    }

    public int hashCode() {
	return name().hashCode();
    }


    /**
     * Returns the members of this type.  This method is package
     * protected because it exposes the representation.  */
    public List members() {	//  !FIXME! once addToIndices is moved to this package make this method package-private
	return members;
    }

    // -------------------------------------------------------------
    // PRETTY-PRINTING
    // -------------------------------------------------------------

    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * VST anchored at this.  */
    void prettyPrintAnnotations(PrintWriter wrtr) {
	for (int i = 0; i < invariants.length; i++) {
	    invariants[i].prettyPrint(wrtr);
	}
    }
    /**
     * Appends to <code>wrtr</code> the source code represented by the
     * non-annotation child nodes of this. */
    void prettyPrintChildren(PrintWriter wrtr) {
	for (int i = 0; i < members.size(); i++) {
	    ((Member) members.get(i)).prettyPrint(wrtr);
	}
    }

    // -------------------------------------------------------------
    // ABSTRACT REPRESENTATION OUTPUT
    // -------------------------------------------------------------

    /**
     * Appends to <code>accum</code> an abstract representation of the 
     * attributes of this. */
    protected void buildAttributeReps(StringBuffer accum) {
	if (interfaces.length > 0) {
	    accum.append("<I");
	    for (int i = 0; i < interfaces.length; i++) {
		accum.append(interfaces[i]);
		if (i < interfaces.length - 1) {
		    accum.append(",");
		}
		
	    }
	}
    }

    /**
     * Appends to <code>accum</code> an abstract representation of the
     * non-annotation child nodes of this. */
    protected void buildChildrenReps(StringBuffer accum) {
	for (int i = 0; i < members.size(); i++) {
	    ((Member) members.get(i)).abstractRepBuilder(accum);
	}
    }

    /**
     * Appends to <code>accum</code> an abstract representation of the 
     * annotation child nodes of this, excluding the common stereotype
     * annotation.  The default implementation here does nothing, some
     * subclasses will override this implementation. */
    protected void buildAnnotationReps(StringBuffer accum) {
	for (int i = 0; i < invariants.length; i++) {
	    invariants[i].abstractRepBuilder(accum);
	}
    }

    // -------------------------------------------------------------
    // MUTATION
    // -------------------------------------------------------------
	
    /**
     * Adds the given field declaration to this type.  */
    void addField( FieldDecl field ) {
	//  !FIXME!ordering Just adds new fields at end.
	members.add(field);
    }
	// remove the given field declaration from this type 
    void removeField(FieldDecl field){
	    members.remove(field);
        }
    /**
     * Adds the received method to members of this type.
     * Creation date: (3/11/2002 2:27:25 PM)
     * @param method org.multijava.samples.cmb.parser.MethodDecl
     * @author George Ushakov
     */
    void addMethod(MethodDecl method) {
	members.add(method);
    }

    // -------------------------------------------------------------
    // REVERSE ENGINEERING
    // -------------------------------------------------------------

    // -------------------------------------------------------------
    // PRIVILEDGED DATA MEMBERS
    // -------------------------------------------------------------

    /**
     * The class invariant annotations.  */
    private TextValuedAnnotation[] invariants;

    /**
     * The members of the type declared by this.  */
    private List members;
    //@ invariant members != null;
    //@ invariant (\forall Object o; members.contains(o); o instanceof Member);

    /**
     * The interfaces implemented or extends by this type declaration.  */
    private String[] interfaces;

}
