// Copyright (c) 2002, Iowa State University

package org.multijava.samples.cmb.parser.testcases;

import org.multijava.samples.cmb.parser.*;
import java.io.*;
import java.util.ArrayList;
import java.io.File;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;
import antlr.RecognitionException;
import antlr.TokenStreamException;

/**
 * This class is a simple test driver for testing the VST builder.
 */
public class Tester {

    /**
     * Run tests or produce expected results.  */
    public static void main (String[] args) {
	try {
	    Tester self = new Tester(args);
	    self.run();
	} catch (BadArgsException e) {
	    System.err.println(e.getMessage());
	    System.err.println(USAGE);
	    System.exit(1);
	} catch (Exception e) {
	    // redirect other exceptions to System.out
	    System.out.println(e.getMessage());
	    e.printStackTrace(System.out);
	    System.exit(3);
	}
    }

    protected Tester(String[] args) throws BadArgsException {
	files = parseArgs(args);
	stdout = new PrintWriter( System.out );
    }

    protected File[] parseArgs(String[] args) throws BadArgsException {
	final String NO_FILES = 
	    "This command requires at least one input file.";
	final String BAD_OPT = "Command does not recognize the option: ";
	final String BAD_FILE = "File does not exist or cannot be read: ";

	if (args.length < 1) {
	    throw new BadArgsException(NO_FILES);
	}

	// Separate args into possible options and filenames
	ArrayList options = new ArrayList(3);
	int firstFilename = -1;
	for (int i = 0; i < args.length; i++) {
	    if (args[i].length() < 1 || args[i].charAt(0) != '-' ) {
		firstFilename = i;
		break;
	    } else {
		// looks like an option
		options.add(args[i].substring(1));
	    }
	}
	
	// Check possible options
	for (int i = 0; i < options.size(); i++) {
	    String opt = (String) options.get(i);
	    if (opt.length() != 1) {
		throw new BadArgsException( BAD_OPT + opt );
	    }
	    switch (opt.charAt(0)) {
	    case 'a':
		abstractOutput = true;
		break;
	    case 'e':
		genExpected = true;
		break;
	    case 'p':
		processOnly = true;
		break;
		
	    case 'q':
		quiet = true;
		break;
		
	    default:
		throw new BadArgsException( BAD_OPT + opt );
	    }
	}
	
	// Sanity check the options
	if (processOnly && genExpected) {
	    System.out.println("Command excepts only one of '-e' and '-p', " + 
			       "ignoring '-e'.");
	    genExpected = false;
	}
	

	// Check for files
	if (firstFilename == -1 || firstFilename == args.length) {
	    throw new BadArgsException(NO_FILES);
	}
	File[] results = new File[args.length - firstFilename];
	for (int i = 0; i < results.length; i++) {
	    results[i] = new File(args[i+firstFilename]);
	    if (!results[i].canRead()) {
		throw new BadArgsException(BAD_FILE + args[i+firstFilename]);
	    }
	}
	
	return results;
    }

    public void run() 
	throws IOException
    {
	for (int i = 0; i < files.length; i++) {
	    Node vst = null;
	    try {
		vst = VSTBuilder.parse(files[i]);
	    } catch (VSTBuilder.SyntaxErrorException e) {
		System.out.println(e.getMessage());
		System.out.println("Skipping " + files[i]);
		continue;
	    }
	    
	    if (processOnly) {
		process(files[i],vst);
	    } else if (genExpected) {
		genExpected(files[i],vst);
	    } else {
 		test(files[i],vst);
	    }
	}
    }

    protected void process(File file, Node vst) {
	if (!quiet) {
	    report( "Results for " + file );
	}
	output(vst,stdout);
    }

    protected void output(Node vst, PrintWriter wrtr) {
	if (abstractOutput) {
	    String result = vst.abstractRep();
	    wrtr.println(result);
	} else {
	    vst.prettyPrint(wrtr);
	}
	wrtr.flush();
    }

    protected void genExpected(File file, Node vst) throws IOException {
	File newFile = suffixedFile(file,"-expected");
	File oldFile = suffixedFile(file,"-expected-old");
	if (!newFile.exists()) {
	    report("No previous expected results for " + file);
	    stdout.flush();
	} else {
	    if (!newFile.renameTo(oldFile)) {
		report("Unable to rename old expected " + 
		       "results from " + newFile + " to " + oldFile);
		report("Skipping " + file);
		stdout.flush();
		return;
	    }
	}
	PrintWriter wrtr = 
	    new PrintWriter( new BufferedWriter( new FileWriter(newFile)));
	output(vst,wrtr);
	
//  	if (compareEqual(newFile,oldFile)) {
//  	    oldFile.delete();
//  	}
    }

    protected void test(File file, Node vst) throws IOException {
	File newFile = suffixedFile(file,"-ckd");
	File oldFile = suffixedFile(file,"-expected");
	if (!oldFile.exists()) {
	    report("No expected results known for " + file);
	    oldFile = null;
	}
	PrintWriter wrtr = 
	    new PrintWriter(new BufferedWriter( new FileWriter(newFile)));
	output(vst,wrtr);
	
//  	if (oldFile != null && compareEqual(newFile,oldFile)) {
//  	    newFile.delete();
//  	}
    }

    protected static final File suffixedFile(File orig, String suffix) {
	return new File(orig.getParentFile(),orig.getName()+suffix);
    }

    protected final void report(String msg) {
	if (!quiet) {
	    stdout.println(msg);
	}
	
    }

//      /**
//       * Compares two files, writing diff output to System.out, and
//       * returning true if the files' contents are the same.  
//       *
//       * <pre><jml>
//       * requires newFile != null && oldFile != null;
//       * </jml></pre>
//       *
//       */
//      protected boolean compareEqual(File newFile, File oldFile) {
	
//      }

    // -------------------------------------------------------------
    // PRIVILEGED DATA MEMBERS
    // -------------------------------------------------------------
    
    protected static class BadArgsException extends Exception {
	protected BadArgsException(String msg) {
	    super(msg);
	}
    }

    /**
     * Indicate whether output should be abstract representation
     * instead of verbatim representation.  */
    private boolean abstractOutput = false;


    /**
     * Indicate whether expected result files should be generated
     * instead of running tests.  */
    private boolean genExpected = false;


    /**
     * Indicates whether files should just be processed with results
     * to System.out instead of running comparisons.  */
    private boolean processOnly = false;
    
    /**
     * Indicates whether messages should be silenced.  */
    private boolean quiet = false;

    /**
     * The files to be processed as read from the command line.  */
    private File[] files;

    private PrintWriter stdout;

    /**
     * A constant string used to display a usage message.  */
    private static final String USAGE =
	"\n"+
	"Usage: java org.multijava.samples.cmb.parser.testcases.Tester [opts] files...\n"+
	"  where the legal options are:\n"+
	"    -a  produce abstract representation instead of verbatim\n"+
	"    -e  generate expected results files, renaming previous\n" + 
	"        ~-expected results to ~-expected-old\n"+
	"    -p  process named files with results to System.out\n"+
	"    -q  quiet, do not display progress messages\n"+
	"\n"+
	"  With no options the default behavior is to process the named\n" +
	"  files, writing results to a file with the suffix \"-ckd\"\n" +
	"  appended.\n" +
	"\n"+
	"Currently client must use diff to compare the results. Eventually\n"+
	"the comparison functionality will be added to this utility."
	;
}
