/*
 * Copyright (C) 2002 Iowa State University
 *
 * $Id: UnitTestVSTBuilder.java,v 1.1 2002/03/26 00:31:34 cclifton Exp $
 */

package org.multijava.samples.cmb.parser.testcases;

import org.multijava.samples.cmb.parser.*;
import java.io.Reader;
import java.io.StringReader;
import java.io.File;

/**
 * This class performs unit tests of VST structure for the parse
 * methods of VSTBuilder.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.1 $ */

public class UnitTestVSTBuilder extends junit.framework.TestCase {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public UnitTestVSTBuilder(String name) {
	super(name);
    }

    // -------------------------------------------------------------
    // TEST CASES
    // -------------------------------------------------------------
    
    // CompilationUnit

    public void test_comp_unit1() {
	String src = "";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn]", rep);
    }

    public void test_comp_unit2() {
	String src = "package foo;";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac]]", rep);
    }

    public void test_comp_unit3() {
	String src = "import java.util.*;";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Imp:java.util.*]]", rep);
    }

    public void test_comp_unit4() {
	String src = "package foo; import java.io.File;";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Imp:java.io.File]]", rep);
    }

    public void test_comp_unit5() {
	// stray semi-colons allowed
	String src = "package foo; import java.io.File; ;";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Imp:java.io.File]]", rep);
    }

    // ClassDecl

    public void test_class_decl1() {
	String src = "class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Cls:Bar:]]", rep);
    }

    public void test_class_decl2() {
	String src = "package foo; class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:]]", rep);
    }

    public void test_class_decl3() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    "class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:[Seo]]]", rep);
    }

    public void test_class_decl4() {
	String src = 
	    " /** just a comment **/\n" + 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    "class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:[Seo]]]", rep);
    }

    public void test_class_decl5() {
	String src = 
	    " /** just a comment **/\n" + 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    "class Bar { };";	// stray semi-colons allowed
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:[Seo]]]", rep);
    }

    public void test_class_decl6() {
	String src = 
	    " /** just a comment **/\n" + 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    ";\n" +	// stray semi-colons screens javadoc
	    "class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:]]", rep);
    }

    public void test_class_decl7() {
	String src = "class Bar extends java.lang.Object { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Cls:Bar:<:java.lang.Object]]", rep);
    }

    public void test_class_decl8() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    "class Bar extends Monkey { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:<:Monkey[Seo]]]", rep);
    }

    public void test_class_decl9() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    "class Bar extends homo.Sapien { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:<:homo.Sapien[Seo]]]", rep);
    }

    public void test_class_decl10() {
	String src = "class Bar implements Comparable { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Cls:Bar:<IComparable]]", rep);
    }

    public void test_class_decl11() {
	String src = "class Bar extends Foo implements Comparable { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Cls:Bar:<:Foo<IComparable]]", rep);
    }

    public void test_class_decl12() {
	String src = 
	    "class Bar extends Foo implements Comparable,Cloneable\n" +
	    "{ }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Cls:Bar:<:Foo<IComparable,Cloneable]]", rep);
    }

    public void test_class_decl13() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @invariant 2==3\n" + 
	    "  */\n" + 
	    "class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:[Inv]]]", rep);
    }

     public void test_class_decl14() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  * @invariant 2==3\n" + 
	    "  * @invariant 2==3\n" + 
	    "  */\n" + 
	    "class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:[Seo][Inv][Inv]]]", rep);
    }

     public void test_class_decl15() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @invariant 2==3\n" + 
	    "  * @invariant 2==3\n" + 
	    "  */\n" + 
	    "class Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Cls:Bar:[Inv][Inv]]]", rep);
    }

   // InterfaceDecl

    public void test_interface_decl1() {
	String src = "interface Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Itr:Bar:]]", rep);
    }

    public void test_interface_decl2() {
	String src = "package foo; interface Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Itr:Bar:]]", rep);
    }

    public void test_interface_decl3() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    "interface Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Itr:Bar:[Seo]]]", rep);
    }

    public void test_interface_decl4() {
	String src = 
	    " /** just a comment **/\n" + 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  */\n" + 
	    "interface Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Itr:Bar:[Seo]]]", rep);
    }

    public void test_interface_decl5() {
	String src = "interface Bar extends Cloneable { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Itr:Bar:<ICloneable]]", rep);
    }

    public void test_interface_decl6() {
	String src = 
	    "package foo;\n" +
	    "interface Bar extends java.lang.Cloneable, Comparable { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Itr:Bar:<Ijava.lang.Cloneable,Comparable]]", 
		     rep);
    }

    public void test_interface_decl7() {
	String src = 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @invariant red==blue\n" + 
	    "  */\n" + 
	    "interface Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Itr:Bar:[Inv]]]", rep);
    }

    public void test_interface_decl8() {
	String src = 
	    " /** just a comment **/\n" + 
	    "package foo; /**\n" + 
	    " /**\n" +
	    "  * @stereotype utility\n" + 
	    "  * @invariant red==blue\n" + 
	    "  * @invariant foo==bar\n" + 
	    "  */\n" + 
	    "interface Bar { }";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Pac][Itr:Bar:[Seo][Inv][Inv]]]", rep);
    }

    // Constructor

    public void test_constructor_decl1() {
	String src = 
	    "class Bar {\n" +
	    "  Bar() { }\n" +
	    "  Bar(int x) { }\n" +
	    "  Bar(int x, Bar o) { }\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Cst:Bar:][Cst:Bar:[Par]][Cst:Bar:[Par][Par]]]]", 
	    rep);
    }

    public void test_constructor_decl2() { // mix in some modifiers
	String src = 
	    "class Bar {\n" +
	    "  private Bar() { }\n" +
	    "  protected Bar(int x) { }\n" +
	    "  public Bar(int x, final Bar o) { }\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Cst:Bar:][Cst:Bar:[Par]][Cst:Bar:[Par][Par]]]]", 
	    rep);
    }

    public void test_constructor_decl3() { // try some exceptions
	String src = 
	    "class Bar {\n" +
	    "  Bar() throws Exception { }\n" +
	    "  Bar(int x) throws Exception, java.io.IOException { }\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Cst:Bar:][Cst:Bar:[Par]]]]", 
	    rep);
    }

    // Nested and inner types

    public void test_nested_type1() {
	String src = 
	    "class Bar {\n" +
	    "  class Foo { }\n" + 
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Cls:Foo:]]]", 
	    rep);
    }

    public void test_nested_type2() {
	String src = 
	    "class Bar {\n" +
	    "  interface Foo { }\n" + 
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Itr:Foo:]]]", 
	    rep);
    }

    public void test_nested_type3() {
	String src = 
	    "interface Bar {\n" +
	    "  static class Foo { }\n" + 
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Itr:Bar:[Cls:Foo:]]]", 
	    rep);
    }

    public void test_nested_type4() {
	String src = 
	    "interface Bar {\n" +
	    "  interface Foo { }\n" + 
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Itr:Bar:[Itr:Foo:]]]", 
	    rep);
    }

    // Methods

    public void test_method_decl1() {
	String src = 
	    "class Bar {\n" +
	    "  void m() { }\n" +
	    "  int m(int x) { }\n" +
	    "  java.io.File m(int x, Bar o) { }\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Mth:m:][Mth:m:[Par]][Mth:m:[Par][Par]]]]", 
	    rep);
    }

    public void test_method_decl2() { // mix in some modifiers
	String src = 
	    "class Bar {\n" +
	    "  private void m() { }\n" +
	    "  protected final int m(int x) { }\n" +
	    "  public java.io.File m(int x, Bar o) { }\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Mth:m:][Mth:m:[Par]][Mth:m:[Par][Par]]]]", 
	    rep);
    }

    public void test_method_decl3() { // try some exceptions
	String src = 
	    "class Bar {\n" +
	    "  private void m() throws Exception { }\n" +
	    "  protected final int m(int x)\n" + 
	    "    throws Exception, java.io.IOException { }\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Mth:m:][Mth:m:[Par]]]]", 
	    rep);
    }

    // Fields

    public void test_field_decl1() {
	String src = 
	    "class Bar {\n" +
	    "  int x;\n" +
	    "  java.io.File f;\n" +
	    "  int[] x, y[], z;\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Fld:x:][Fld:f:][Fld:x:[Fld:y:][Fld:z:]]]]", 
	    rep);
    }

    public void test_field_decl2() { // mix in some modifiers
	String src = 
	    "class Bar {\n" +
	    "  int x;\n" +
	    "  java.io.File f;\n" +
	    "  int[] x, y[], z;\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Fld:x:][Fld:f:][Fld:x:[Fld:y:][Fld:z:]]]]", 
	    rep);
    }

    // Initializers

    public void test_initializer_decl1() {
	String src = 
	    "class Bar {\n" +
	    "  static { }\n" +
	    "  { }\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals(
	    "[CUn[Cls:Bar:[Izr:$Init:][Izr:$Init:]]]", 
	    rep);
    }

    // Javadocs
    public void test_javadocs1() {
	String src = 	    
	    "/*** comment ***/\n" +
	    "class Bar {\n" +
	    "}";
	String rep = abstractRepFrom(src);
	assertEquals("[CUn[Cls:Bar:]]", rep);
    }

    // -------------------------------------------------------------
    // HELPER METHODS
    // -------------------------------------------------------------

    static String abstractRepFrom(String sourceCode) {
	return parse(sourceCode).abstractRep();
    }

    static CompilationUnit parse(String sourceCode) {
	try {
	    Reader rdr = new StringReader( sourceCode );
	    return VSTBuilder.parse(rdr);
	} catch (VSTBuilder.SyntaxErrorException e) {
	    e.printStackTrace();
	    throw new RuntimeException(e.getMessage());
	}
    }

    // -------------------------------------------------------------
    // TEST FIXTURE
    // -------------------------------------------------------------

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

}
