/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: CommentTokenizer.java,v 1.5 2002/04/30 14:00:11 cclifton Exp $
 */

package org.multijava.samples.cmb.query;

import java.util.ArrayList;
import java.io.StringReader;
import antlr.Token;
import java.util.StringTokenizer;


/**
 * This class finds words appearing within the comments of a string of
 * source code text.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.5 $ */
class CommentTokenizer implements CommentFinderTokenTypes {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public CommentTokenizer(String text) {
	this.text = text;
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * Adds to <CODE>kwIndex</CODE> a mapping for <CODE>mW</CODE> from
     * each of the words appearing in a comment within
     * <CODE>this.text</CODE>.  */
    void addMappingsForTo( MemberWrapper mW, Index kwIndex ) {
	String[] words = keywords();

	for (int i = 0; i < words.length; i++) {
	    kwIndex.insert(words[i],mW);
	}
    }

    /**
     * Returns an array of the words appearing in comments, in the
     * order in which the words appear.  */
    String[] keywords() {
	if (keywords != null) {
	    return keywords;
	}

	ArrayList accum = new ArrayList();

	// Construct a lexer to extract the comment text
	CommentFinder finder = new CommentFinder( new StringReader( text ) );

	try {
	    Token tok = finder.nextToken();
	    while (tok.getType() != EOF) {
		switch (tok.getType()) {
		case SL_COMMENT:
		    lexSingleLine(tok.getText(), accum);
		    break;
		

		case ML_COMMENT:
		    lexMultiLine(tok.getText(), accum);
		    break;
		
		default:
		    // nothing to do
		    break;
		}
		tok = finder.nextToken();
	    }
	} catch (antlr.TokenStreamException e) {
	    //  Should not get here since the code was already
	    //  successfully lexed once.
	    throw new RuntimeException(e.getMessage());
	}
	
	// transfer contents of accum to instance field and return result
	keywords = new String[accum.size()];
	accum.toArray(keywords);
	return keywords;
    }

    // -------------------------------------------------------------
    // STATIC HELPER METHODS
    // -------------------------------------------------------------


    /**
     * Lexes the given single line <CODE>comment</CODE> and adds the
     * words appearing therein to <CODE>accum</CODE>.  */
    static void lexSingleLine(String comment, ArrayList accum) {
	StringTokenizer toker = new StringTokenizer(comment,WORD_DELIMS);
	while (toker.hasMoreTokens()) {
	    String word = stripPunctuation(toker.nextToken());
	    if (word != EMPTY_STRING) {
		accum.add(word);
	    }
	}
    }
    
    /**
     * Lexes the given multi-line <CODE>comment</CODE> and adds the
     * words appearing therein to <CODE>accum</CODE>.  */
    static void lexMultiLine(String comment, ArrayList accum) {
	// We strip out newlings and leading white-space and stars to
	// create a single long line of text.  That text can then be
	// processed as a single-line comment.
	StringBuffer buf = new StringBuffer();
	int pos = 0;
	while (pos < comment.length()) {
	    int lfPos = comment.indexOf('\n',pos);
	    int crPos = comment.indexOf('\r',pos);

	    if (lfPos < 0 && crPos < 0) {
		// no more newlines
		buf.append(comment.substring(pos));
		pos = comment.length();
		break;
	    }

	    // At least one of lfPos and crPos is non-negative.

	    // Guess that lfPos >= 0...
	    int nlPos = lfPos;
	    if (nlPos < 0) {
		// ... nope, only crPos is greater than zero.
		nlPos = crPos;
	    } else if (crPos > 0) {
		// both are greater than zero, so choose smallest (i.e., first)
		if (crPos < lfPos) {
		    nlPos = crPos;
		}
	    }

	    // append text between pos and nlPos (if any) to buffer
	    if (pos < nlPos) {
		buf.append(comment.substring(pos,nlPos)).append(" ");
		pos = nlPos;
	    }

	    // skip all white-space following the newline
	    while (pos < comment.length()) {
		if (!Character.isWhitespace( comment.charAt(pos) )) {
		    break;
		}
		pos++;
	    }
	    
	    // skip all stars following the whitespace following the newline
	    while (pos < comment.length()) {
		if (comment.charAt(pos) != '*') {
		    break;
		}
		pos++;
	    }
	    
	    // ready for next iteration
	}

	lexSingleLine(buf.toString(), accum);
    }
    

    /**
     * Removes any leading and trailing characters from orig for which
     * (Character.isLetterOrDigit() ||
     * Character.isJavaIdentifierPart()) is not true.  */
    static String stripPunctuation(String orig) {
	char[] chs = new char[orig.length()];
	orig.getChars(0,orig.length(),chs,0);
	int startPos;
	for (startPos = 0; startPos < chs.length; startPos++) {
	    if (Character.isLetterOrDigit(chs[startPos]) ||
		Character.isJavaIdentifierPart(chs[startPos])) {
		break;
	    }
	}

	if (startPos >= chs.length) {
	    return EMPTY_STRING;
	}

	int endPos;
	for (endPos = chs.length; endPos > startPos; endPos--) {
	    if (Character.isLetterOrDigit(chs[endPos-1]) ||
		Character.isJavaIdentifierPart(chs[endPos-1])) {
		break;
	    }
	}
	
	if (endPos < startPos) {
	    // should never get here because the first for loop should
	    // have consumed the entire string
	    return EMPTY_STRING;
	}
	
	// result is the text in the interval [startPos,endPos),
	// closed on the left, open on the right
	return orig.substring(startPos,endPos);
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * The verbatim text from which comment keywords are to be
     * extracted.  */
    private final String text;

    /**
     * The keywords appearing in this.text, or null if the text has
     * not yet been processed.  
     * @see #keywords()
     */
    private String[] keywords;

    private static final String WORD_DELIMS = " \t\n\r\f<>";

    private static final String EMPTY_STRING = "";
}
