/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Index.java,v 1.10 2002/06/25 17:08:29 cclifton Exp $
 */

package org.multijava.samples.cmb.query;

import java.util.HashMap;
import java.util.Map;
import java.io.File;
import org.multijava.samples.cmb.parser.*;
import java.io.FileNotFoundException;
import javax.swing.ProgressMonitor;

/**
 * This class represents an inverted-list style index mapping
 * vocabulary words to occurances of VSTs for declarations.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.10 $ */
public class Index {

    // -------------------------------------------------------------
    // CONSTRUCTORS
    // -------------------------------------------------------------
    Index( boolean isCaseSensitive ) {
	this.isCaseSensitive = isCaseSensitive;
    }

    // -------------------------------------------------------------
    // PUBLIC INTERFACE
    // -------------------------------------------------------------

    /**
     * Returns an identifier index (<CODE>\result[0]</CODE>) and a
     * keyword index (<CODE>\result[1]</CODE>) covering all the given
     * <CODE>files</CODE>.  */
    public static final Index[] buildIndexFor(File[] files) {
	return buildIndexFor(files,null);
    }

    /**
     * Returns an identifier index (<CODE>\result[0]</CODE>) and a
     * keyword index (<CODE>\result[1]</CODE>) covering all the given
     * <CODE>files</CODE>.  */
    public static final Index[] buildIndexFor(File[] files,
					      ProgressMonitor progMon) {
	Index idIndex = new Index(true);
	Index kwIndex = new Index(false);
	messages = new StringBuffer();

	// zap old parent field map
	parentLinksMap.clear();

    fileLoop:
	for (int i = 0; i < files.length; i++) {
	    try {
		if (progMon != null) {
		    progMon.setProgress(i);
		    progMon.setNote(files[i].getName());
		    if (progMon.isCanceled()) {
			messages.append("parsing cancelled after " + (i - 1) +
					" of " + files.length + " files");
			break fileLoop;
		    }
		    
		}
		
		CompilationUnit unit = VSTBuilder.parse(files[i]);
		unit.addToIndices(idIndex, kwIndex);
		unit.createParentLinks();
	    } catch (FileNotFoundException e) {
		messages.append( "Could not find file '" + files[i] + "'\n" );
	    } catch (VSTBuilder.SyntaxErrorException e) {
		messages.append( "Syntax error parsing file '" + 
				 files[i] + "':" + e.getMessage() + "\n" );
	    } catch (RuntimeException e) {
		messages.append( e.getClass() + " parsing file '" +
				 files[i] + "':" + e.getMessage() + "\n" );
	    } 
	}
	
	if (progMon != null) {
	    progMon.close();
	}
	return new Index[] { idIndex, kwIndex };
    }

    public static final String messages() {
	if (messages == null) {
	    return "";
	} else {
	    return messages.toString();
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * Adds an index entry mapping the given vocabulary word to the
     * given occurance.  */
    void insert(String vocabWord, MemberWrapper occurance) {
	String w = vocabWord;
	if (!isCaseSensitive) {
	    w = w.toLowerCase();
	}
	
	if (map.containsKey(w)) {
	    ((OccuranceList) map.get(w)).add(occurance);
	} else {
	    OccuranceList newOL = new OccuranceList();
	    newOL.add(occurance);
	    map.put(w,newOL);
	}
    }

    /**
     * Returns a list of all the occurance mapped to by the given
     * query term and satisfying the given matcher.  If <CODE>catMatcher ==
     * null</CODE> then all occurance matching the given query term are
     * returned.  */
    OccuranceList find(String queryTerm, Matcher catMatcher) {
	String qt = queryTerm;
	if (!isCaseSensitive) {
	    qt = qt.toLowerCase();
	}
	
	if (map.containsKey(qt)) {
	    OccuranceList result = (OccuranceList) map.get(qt);
	    if (catMatcher == null) {
		return result;
	    } else {
		return result.filter(catMatcher);
	    }
	} else {
	    return OccuranceList.EMPTY;
	}
    }

    public String toString() {
	return map.toString();
    }
    
    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    private Map map = new HashMap();

    private final boolean isCaseSensitive;

    /**
     * Used to store error messages accumulated during file parsing.  */
    private static StringBuffer messages;


    /**
     * This static member provides a mapping from VST nodes to VST
     * nodes.  <CODE>(parentLinksMap.get(c) == p) ==> (c is a child
     * node of p)</CODE> */
    final static Map parentLinksMap = new HashMap();
}
