/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MemberWrapper.java,v 1.3 2007/02/08 14:04:57 leavens Exp $
 */

package org.multijava.samples.cmb.query;

import java.io.File;
import org.multijava.samples.cmb.parser.Member;


/**
 * This class is used to wrap <CODE>Member</CODE> instances to modify
 * their behavior.  This wrapper &quot;overrides&quot;
 * <CODE>equals</CODE> and <CODE>hashCode</CODE> to be based on
 * <CODE>Member</CODE> object identity instead of the standard VST
 * node comparison.  (The <CODE>equals</CODE> methods of the VST nodes
 * are designed such that two members compare equal if they are
 * potentially ambiguous when appearing in the same compilation unit.)
 * This wrapper also adds file and ordering information to members to
 * proper ordering in displays.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.3 $ */
class MemberWrapper implements Comparable {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    MemberWrapper( Member mbr, File file, int position ) {
	this.mbr = mbr;
	this.file = file;
	this.position = position;
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------
    
    public boolean equals(Object other) {
	return other instanceof MemberWrapper &&
	    other.hashCode() == hashCode();
    }
    
    public int hashCode() {
	return System.identityHashCode(mbr);
    }
    
    public int compareTo(/*@ non_null @*/ Object other) {
	return hashCode() - other.hashCode();
    }
    
    /**
     * Returns an object that will compare less than or equal to
     * the successor of this.  
     *
     * <pre><jml>
     * ensures this.compareTo(\result) < 0;
     * ensures \result.compareTo(this) > 0;
     * </jml></pre>
     *
     */
    Comparable successor() {
	final int succId = hashCode() + 1;
	return new MemberWrapper(null,null,0) {
		public int hashCode() {
		    return succId;
		}
	    };
    }
    
    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------


    /**
     * The member wrapped by this.  */
    final Member mbr;

    /**
     * The File from which the wrapped member originated.  */
    final File file;


    /**
     * The relative position of the wrapped member in the file from
     * which it originated.  */
    final int position;

}
