/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: OccuranceList.java,v 1.3 2002/04/23 15:04:17 cclifton Exp $
 */

package org.multijava.samples.cmb.query;

import org.multijava.samples.cmb.parser.Member;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.SortedSet;
import java.util.Arrays;

/**
 * This class represents a list of partial VSTs for Java declarations.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.3 $
 */
public class OccuranceList {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Constructs a new, empty occurance list.  */
    public OccuranceList() {
	this.occurances = new TreeSet();
    }

    /**
     * Constructs a new occurance list containing the same members as
     * the given list.  */
    OccuranceList(OccuranceList clone) {
	this.occurances = (TreeSet) clone.occurances.clone();
    }
    
    // -------------------------------------------------------------
    // OPERATORS
    // -------------------------------------------------------------

    /**
     * Adds the given <CODE>wrappedMember</CODE> to this occurance list.  */
    void add(MemberWrapper wrappedMember) {
	occurances.add(wrappedMember);
    }

    /**
     * Returns an occurance list containing just those members of this
     * list that are accepted by the given <CODE>matcher</CODE>.  */
    public OccuranceList filter(Matcher matcher) {
	OccuranceList result = new OccuranceList();
	Iterator iter = occurances.iterator();
	while (iter.hasNext()) {
	    MemberWrapper mW = (MemberWrapper) iter.next();
	    if (matcher.matches(mW.mbr)) {
		result.occurances.add(mW);
	    }
	}
	return result;
    }

    /**
     * Calculates the set-theoretic intersection of <CODE>this</CODE>
     * and <CODE>other</CODE>. */
    public OccuranceList intersect(OccuranceList other) {
	OccuranceList result = new OccuranceList(this);
	result.occurances.retainAll(other.occurances);
	return result;
    }

    public OccuranceList intersect(OccuranceList@EmptyOccuranceList other) {
	return EMPTY;
    }

    /**
     * Calculates the set-theoretic union of <CODE>this</CODE> and
     * <CODE>other</CODE>. */
    public OccuranceList union(OccuranceList other) {
	OccuranceList result = new OccuranceList(this);
	result.occurances.addAll(other.occurances);
	return result;
    }

    public OccuranceList union(OccuranceList@EmptyOccuranceList other) {
	return new OccuranceList(this);
    }

    /**
     * Calculates the set-theoretic asymmetric difference of
     * <CODE>this</CODE> and <CODE>other</CODE>. */
    public OccuranceList minus(OccuranceList other) {
	OccuranceList result = new OccuranceList(this);
	result.occurances.removeAll(other.occurances);
	return result;
    }

    public OccuranceList minus(OccuranceList@EmptyOccuranceList other) {
	return new OccuranceList(this);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    public Member[] members() {
	MemberWrapper[] wrapped = new MemberWrapper[occurances.size()];
	occurances.toArray(wrapped);
	
	//  !FIXME! create custom comparator to sort result based on
	//  file and position.
	// Arrays.sort(result, new SourceOrder());

	Member[] result = new Member[wrapped.length];
	for (int i = 0; i < wrapped.length; i++) {
	    result[i] = wrapped[i].mbr;
	}
	
	return result;
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * Contains references to the members that are included in this
     * occurance list.
     *
     * <pre><jml>
     * invariant (\forall Object o; occurances.contains(o); 
     *				o instanceof MemberWrapper);
     * </jml></pre>
     *
     */
    private TreeSet occurances;

    /**
     * An empty, immutable occurance list.  */
    static final OccuranceList EMPTY = new EmptyOccuranceList();

    /**
     * This nested class represents an empty, immutable occurance
     * list.  It overrides various superclass methods for efficiency.
     * @see #emptyList()
     * @see #EMPTY */
    private static class EmptyOccuranceList extends OccuranceList {
	void add(MemberWrapper mbr) {
	    throw new IllegalStateException( "cannot add to immutable list" );
	}

	public OccuranceList filter(Matcher matcher) {
	    return this;
	}

	public OccuranceList intersect(OccuranceList other) {
	    return this;
	}

	public OccuranceList intersect(OccuranceList@EmptyOccuranceList other){
	    return this;
	}

	public OccuranceList union(OccuranceList other) {
	    return new OccuranceList(other);
	}
	
	public OccuranceList union(OccuranceList@EmptyOccuranceList other) {
	    return this;
	}

	public OccuranceList minus(OccuranceList other) {
	    return this;
	}
	
	public OccuranceList minus(OccuranceList@EmptyOccuranceList other) {
	    return this;
	}
    }
}


