/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: QueryLexer.java,v 1.2 2002/04/30 14:00:11 cclifton Exp $
 */

package org.multijava.samples.cmb.query;

import antlr.*;
import java.util.HashMap;

/**
 * This class implements a lexer for query expression strings.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.2 $
 */
public final class QueryLexer implements antlr.TokenStream, QueryTokenTypes {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Constructs a new lexer for the given query string.
     *
     * <pre><jml>
     * requires query != null;
     * </jml></pre>
     *
     */
    public QueryLexer(String query) {
	len = query.length();
	char[] queryChars = new char[len];
	query.getChars(0,len,queryChars,0);
	this.query = queryChars;
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * Returns the next token in the query string, or throws an
     * exception if the string does not match the lexical grammar.  */
    public Token nextToken() throws QueryProcessingException {
	tokType = Token.SKIP;
	while (tokType == Token.SKIP && scanPos < len) {
	    if (Character.isWhitespace(query[scanPos])) {
		consumeWhitespace();
	    } else if (query[scanPos] == '(') {
		matchLParen();
	    } else if (query[scanPos] == ')') {
		matchRParen();
	    } else {
		matchIdent();
	    }
	}
	
	return genToken();
    }

    // -------------------------------------------------------------
    // LEXER STATES
    // -------------------------------------------------------------

    void consumeWhitespace() {
	while (scanPos < len && Character.isWhitespace(query[scanPos])) {
	    scanPos++; 
	}
	startPos = scanPos;
	tokType = Token.SKIP;
    }

    void matchIdent() throws QueryProcessingException {
	if (!Character.isLetterOrDigit(query[scanPos]) &&
	    !Character.isJavaIdentifierPart(query[scanPos])) {
	    throw exception("letter or digit");
	}
	scanPos++;
	while (scanPos < len && 
	       (Character.isLetterOrDigit(query[scanPos]) ||
		Character.isJavaIdentifierPart(query[scanPos]))) {
	    scanPos++;
	}
	
	Object type = symbolTable.get(getText());
	if (type == null) {
	    tokType = IDENT;
	} else {
	    tokType = ((Integer) type).intValue();
	}
    }

    void matchLParen() throws QueryProcessingException {
	if (query[scanPos] != '(') {
	    throw exception("'('");
	}
	scanPos++;
	tokType = LPAREN;
    }
    
    void matchRParen() throws QueryProcessingException {
	if (query[scanPos] != ')') {
	    throw exception("')'");
	}
	scanPos++;
	tokType = RPAREN;
    }
    
    // -------------------------------------------------------------
    // HELPERS
    // -------------------------------------------------------------

    private QueryProcessingException exception(String expected) {
	String message = "Expected " + expected + " but found " + 
	    ((scanPos < len) ? ("'" + query[scanPos] + "'") : "end of text");
	return new QueryProcessingException(scanPos,message);
    }

    Token genToken() {
	Token result;
	if (startPos >= len) {
	    result = new CommonToken(Token.EOF_TYPE,"");
	} else {
	    result = new CommonToken(tokType,getText());
	}
	result.setColumn(startPos);
	startPos = scanPos;
	return result;
    }

    private String getText() {
	return new String(query, startPos, scanPos - startPos);
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA (package access for testing)
    //---------------------------------------------------------------------

    /**
     * The text of the query string.  */
    private final char[] query;

    /**
     * The length of the query string.  */
    private final int len;

    /**
     * The position in <CODE>query</CODE> of the next character to check.  */
    int scanPos = 0;

    /**
     * The position in <CODE>query</CODE> that is the beginning of the
     * next token to return.  */
    int startPos = 0;

    int tokType = Token.SKIP;

    static final HashMap symbolTable = new HashMap();

    static {
	symbolTable.put("AND",new Integer(AND));
	symbolTable.put("OR",new Integer(OR));
	symbolTable.put("NOT",new Integer(NOT));
    }
}
