/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestIndex.java,v 1.6 2002/06/25 17:08:29 cclifton Exp $
 */

package org.multijava.samples.cmb.query;

import org.multijava.util.testing.TestCase;
import org.multijava.samples.cmb.parser.*;
import java.io.*;

/**
 * TestIndex.java
 *
 * @author Curtis Clifton
 * @version $Revision: 1.6 $
 */
public class TestIndex extends TestCase implements TestConstants {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public TestIndex(String name) {
	super(name);
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------
    
    public void test_find() {
	Member[] result;

	// without filtering
	result = idx.find("list12",null).members();
	assertEquals(2,result.length);

	result = idx.find("list23",null).members();
	assertEquals(2,result.length);

	result = idx.find("foo",null).members();
	assertEquals(0,result.length);

	// with filtering
	result = idx.find("list12",new NameMatcher("i1")).members();
	assertEquals(1,result.length);

	result = idx.find("list12",new NameMatcher("i2")).members();
	assertEquals(1,result.length);

	result = idx.find("list12",new NameMatcher("i3")).members();
	assertEquals(0,result.length);

	result = idx.find("list23",new NameMatcher("i1")).members();
	assertEquals(0,result.length);

	result = idx.find("list23",new NameMatcher("i2")).members();
	assertEquals(1,result.length);

	result = idx.find("list23",new NameMatcher("i3")).members();
	assertEquals(1,result.length);

	result = idx.find("foo",new NameMatcher("i1")).members();
	assertEquals(0,result.length);
    }

    public void test_processQueryOn() {
	Member[] result;
	ParseNode query;

	// simple query
	query = new QueryTerm("list12");

	result = query.processQueryOn(idx, null).members();
	assertEquals(2,result.length);

	result = query.processQueryOn(idx, new NameMatcher("i1")).members();
	assertEquals(1,result.length);

	result = query.processQueryOn(idx, new NameMatcher("i2")).members();
	assertEquals(1,result.length);

	result = query.processQueryOn(idx, new NameMatcher("i3")).members();
	assertEquals(0,result.length);

	// and query
	query = new AndOp(new QueryTerm("list12"), new QueryTerm("list23"));
	result = query.processQueryOn(idx, null).members();
	assertEquals(1,result.length);

	result = query.processQueryOn(idx, new NameMatcher("i2")).members();
	assertEquals(1,result.length);

	// or query
	query = new OrOp(new QueryTerm("list12"), new QueryTerm("list23"));
	result = query.processQueryOn(idx, null).members();
	assertEquals(3,result.length);

	result = query.processQueryOn(idx, new NameMatcher("i2")).members();
	assertEquals(1,result.length);

	// but-not query
	query = new ButNotOp(new QueryTerm("list12"), new QueryTerm("list23"));
	result = query.processQueryOn(idx, null).members();
	assertEquals(1,result.length);

	result = query.processQueryOn(idx, new NameMatcher("i1")).members();
	assertEquals(1,result.length);
    }

    public void test_buildIndexFor() {
	String code = 
	    "package foo;\n" +
	    "\n" +
	    "/** class comment **/\n" +
	    "public class Foo {\n" +
	    "  /** method comment **/\n" +
	    "  void m() {\n" +
	    "    int x = 1;\n" +
	    "  } // endMethod\n" +
	    "} // endClass\n" + 
	    "\n"
	    ;
	File testFile = fileFor(code);

	File[] fs = new File[] { testFile };

	Index[] results = Index.buildIndexFor(fs);
	assertEquals( "", Index.messages() );

	ParseNode query;
	Member[] answer;

	// using identifier index
	query = new QueryTerm("Foo");
	answer = query.processQueryOn(results[0],null).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof ClassDecl);

	query = new QueryTerm("m");
	answer = query.processQueryOn(results[0],null).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof MethodDecl);

	// using keyword index
	query = new QueryTerm("class");
	answer = query.processQueryOn(results[1],null).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof ClassDecl);

	query = new QueryTerm("method");
	answer = query.processQueryOn(results[1],null).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof MethodDecl);

	query = new QueryTerm("endClass");
	answer = query.processQueryOn(results[1],null).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof ClassDecl);

	query = new QueryTerm("endMethod");
	answer = query.processQueryOn(results[1],null).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof MethodDecl);

	query = new QueryTerm("comment");
	answer = query.processQueryOn(results[1],null).members();
	assertEquals(2,answer.length);

	query = new QueryTerm("comment");
	answer = 
	    query.processQueryOn(results[1],
				 new Matcher() {
					 public boolean matches(Member m) {
					     return m instanceof TypeDecl;
					 }
				     } ).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof ClassDecl);

	query = new QueryTerm("comment");
	answer = 
	    query.processQueryOn(results[1],
				 new Matcher() {
					 public boolean matches(Member m) {
					     return m instanceof MethodDecl;
					 }
				     } ).members();
	assertEquals(1,answer.length);
	assertTrue(answer[0] instanceof MethodDecl);

    }


    // -------------------------------------------------------------
    // HELPERS
    // -------------------------------------------------------------

    private static File fileFor(String code) {
	try {
	    File testFile = File.createTempFile("cmb", ".java");
	    Writer wrtr = new FileWriter(testFile);
	    wrtr.write(code);
	    wrtr.close();
	    testFile.deleteOnExit();
	    return testFile;
	} catch (IOException e) {
	    return null;
	}
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    Index idx = new Index(false);
    {
	idx.insert("list12",mW1a);
	idx.insert("list12",mW1b);
	idx.insert("list12",mW2a);
	idx.insert("list12",mW2b);
	
	idx.insert("list23",mW2a);
	idx.insert("list23",mW2b);
	idx.insert("list23",mW3a);
	idx.insert("list23",mW3b);
    }
}
