/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestQueryLexer.java,v 1.2 2006/05/23 21:44:57 leavens Exp $
 */

package org.multijava.samples.cmb.query;

import org.multijava.util.testing.TestCase;
import antlr.Token;
import java.util.ArrayList;

/**
 * This class provides unit testing for the QueryLexer class.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.2 $
 */
public class TestQueryLexer extends TestCase implements QueryTokenTypes {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public TestQueryLexer(String name) {
	super(name);
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------
    
    // -------------------------------------------------------------
    public void test_consumeWhitespace_1() {
	lex = new QueryLexer("     ");
	lex.consumeWhitespace();
	assertEquals(5,lex.scanPos);
	assertEquals(5,lex.startPos);
 	lex.consumeWhitespace();
	assertEquals(5,lex.scanPos);
	assertEquals(5,lex.startPos);
   }

    public void test_consumeWhitespace_2() {
	lex = new QueryLexer("01234");
	lex.consumeWhitespace();
	assertEquals(0,lex.scanPos);
	assertEquals(0,lex.startPos);
	lex.consumeWhitespace();
	assertEquals(0,lex.scanPos);
	assertEquals(0,lex.startPos);
    }

    public void test_consumeWhitespace_3() {
	lex = new QueryLexer("     01234");
	lex.consumeWhitespace();
	assertEquals(5,lex.scanPos);
	assertEquals(5,lex.startPos);
	lex.consumeWhitespace();
	assertEquals(5,lex.scanPos);
	assertEquals(5,lex.startPos);
    }

    public void test_consumeWhitespace_4() {
	lex = new QueryLexer(" \t01234");
	lex.consumeWhitespace();
	assertEquals(2,lex.scanPos);
	assertEquals(2,lex.startPos);
	lex.consumeWhitespace();
	assertEquals(2,lex.scanPos);
	assertEquals(2,lex.startPos);
    }

    //-------------------------------------------------------------
    public void test_matchIdent_1() {
	try {
	    lex = new QueryLexer("#foo");
	    lex.matchIdent();
	    fail("should have thrown exception");
	} catch (QueryProcessingException e) {
	    assertEquals(0,e.position());
	    assertEquals("Expected letter or digit but found '#'",
			 e.getMessage());
	}
    }

    public void test_matchIdent_2() throws QueryProcessingException {
	lex = new QueryLexer("foo bar");

	lex.matchIdent();
	Token tok = lex.genToken();
	assertEquals("foo",tok.getText());
	assertEquals(0,tok.getColumn());
	assertEquals(IDENT,tok.getType());

	lex.consumeWhitespace();

	lex.matchIdent();
	tok = lex.genToken();
	assertEquals("bar",tok.getText());
	assertEquals(4,tok.getColumn());
	assertEquals(IDENT,tok.getType());
    }

    public void test_matchIdent_3() throws QueryProcessingException {
	lex = new QueryLexer("AND OR NOT");

	lex.matchIdent();
	Token tok = lex.genToken();
	assertEquals(0,tok.getColumn());
	assertEquals(AND,tok.getType());

	lex.consumeWhitespace();

	lex.matchIdent();
	tok = lex.genToken();
	assertEquals(4,tok.getColumn());
	assertEquals(OR,tok.getType());

	lex.consumeWhitespace();    

	lex.matchIdent();
	tok = lex.genToken();
	assertEquals(7,tok.getColumn());
	assertEquals(NOT,tok.getType());
    }

    //-------------------------------------------------------------
    public void test_matchLParen_1() {
	try {
	    lex = new QueryLexer("foo");
	    lex.matchLParen();
	    fail("should have thrown exception");
	} catch (QueryProcessingException e) {
	    assertEquals(0,e.position());
	    assertEquals("Expected '(' but found 'f'",
			 e.getMessage());
	}
    }

    public void test_matchLParen_2() throws QueryProcessingException {
	lex = new QueryLexer("(");
	lex.matchLParen();
	Token tok = lex.genToken();
	assertEquals("(",tok.getText());
	assertEquals(0,tok.getColumn());
	assertEquals(LPAREN,tok.getType());
    }

     public void test_matchRParen_1() {
	try {
	    lex = new QueryLexer("foo");
	    lex.matchRParen();
	    fail("should have thrown exception");
	} catch (QueryProcessingException e) {
	    assertEquals(0,e.position());
	    assertEquals("Expected ')' but found 'f'",
			 e.getMessage());
	}
    }

    public void test_matchRParen_2() throws QueryProcessingException {
	lex = new QueryLexer(")");
	lex.matchRParen();
	Token tok = lex.genToken();
	assertEquals(")",tok.getText());
	assertEquals(0,tok.getColumn());
	assertEquals(RPAREN,tok.getType());
    }
    
    //-------------------------------------------------------------
    public void test_nextToken_1() throws QueryProcessingException {
	String[] actual = getStringsFor("cat AND dog");
	String[] expected = new String[] { "cat", "AND", "dog" };
	assertEqualTokens(expected,actual);
    }

    public void test_nextToken_2() throws QueryProcessingException {
	String[] actual = getStringsFor("(cat AND dog)");
	String[] expected = new String[] { "(", "cat", "AND", "dog", ")" };
	assertEqualTokens(expected,actual);
    }

     public void test_nextToken_3() throws QueryProcessingException {
	String[] actual = getStringsFor("cat AND dog OR (horse NOT mule)");
	String[] expected = new String[] { "cat", "AND", "dog", "OR",
					   "(", "horse", "NOT", "mule", ")" };
	assertEqualTokens(expected,actual);
    }

    public void test_nextToken_4() throws QueryProcessingException {
	int[] actual = getTypesFor("cat AND dog");
	int[] expected = new int[] { IDENT,AND,IDENT,EOF };
	assertEqualTokTypes(expected,actual);
    }

    public void test_nextToken_5() throws QueryProcessingException {
	int[] actual = getTypesFor("(cat AND dog)");
	int[] expected = new int[] { LPAREN,IDENT,AND,IDENT,RPAREN,EOF };
	assertEqualTokTypes(expected,actual);
    }

     public void test_nextToken_6() throws QueryProcessingException {
	int[] actual = getTypesFor("cat AND dog OR (horse NOT mule)");
	int[] expected = new int[] { IDENT,AND,IDENT,OR,
				     LPAREN,IDENT,NOT,IDENT,RPAREN,EOF };
	assertEqualTokTypes(expected,actual);
    }

    // -------------------------------------------------------------
    // HELPERS
    // -------------------------------------------------------------

    private ArrayList getTokensFor(String query)
		throws QueryProcessingException 
    {
	ArrayList accum = new ArrayList();
	lex = new QueryLexer(query);
	Token tok = lex.nextToken();
	accum.add(tok);
	while (tok.getType() != Token.EOF_TYPE) {
	    tok = lex.nextToken();
	    accum.add(tok);
	}
	return accum;
    }

    private String[] getStringsFor(String query) 
	throws QueryProcessingException 
    {
	ArrayList accum = getTokensFor(query);
	// get strings from all but the EOF token
	String[] results = new String[accum.size() - 1];
	for (int i = 0; i < accum.size()-1; i++) {
	    results[i] = ((Token) accum.get(i)).getText();
	}
	return results;
    }

     private int[] getTypesFor(String query) 
	throws QueryProcessingException 
    {
	ArrayList accum = getTokensFor(query);
	int[] results = new int[accum.size()];
	for (int i = 0; i < accum.size(); i++) {
	    results[i] = ((Token) accum.get(i)).getType();
	}
	return results;
    }

   private void assertEqualTokens(String[] expected, String[] actual) {
	// we use the length of the longer array so an exception will
	// be thrown on array length mismatch
	int maxLength = expected.length > actual.length ? 
	    expected.length : actual.length;
	for (int i = 0; i < maxLength; i++) {
	    assertEquals(expected[i],actual[i]);
	}
    }

   private void assertEqualTokTypes(int[] expected, int[] actual) {
	// we use the length of the longer array so an exception will
	// be thrown on array length mismatch
	int maxLength = expected.length > actual.length ? 
	    expected.length : actual.length;
	for (int i = 0; i < maxLength; i++) {
	    assertEquals(expected[i],actual[i]);
	}
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    private QueryLexer lex;
    

}
