/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: FileTreeModel.java,v 1.4 2007/02/08 14:04:58 leavens Exp $
 */

package org.multijava.samples.cmb.ui;

import javax.swing.*;
import javax.swing.tree.*;
import java.io.File;
import java.util.TreeSet;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeModel;
import java.util.ArrayList;
import javax.swing.event.TreeModelEvent;
import javax.swing.filechooser.FileFilter;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.awt.Component;
import java.util.Collections;

/**
 * This class represents the program to be mined as a tree of files.
 * 
 * @author Curtis Clifton
 * @version $Revision: 1.4 $ */
public class FileTreeModel implements TreeModel {

    //---------------------------------------------------------------------
    // CONSTRUCTORS	
    //---------------------------------------------------------------------

    /**
     * Constructs a new empty <CODE>FileTreeModel</CODE>.  */
    public FileTreeModel() {
	this.root = new FileTreeRoot();
    }
    
    /**
     * Constructs a new <CODE>FileTreeModel</CODE> by performing a
     * deep copy of <CODE>old</CODE>.  */
    public FileTreeModel(FileTreeModel old) {
	this.root = new FileTreeRoot(old.root);
    }

    // -------------------------------------------------------------
    // IMPLEMENT TREEMODEL
    // -------------------------------------------------------------

    /**
     * Returns the <CODE>index</CODE>-th child node of
     * <CODE>parent</CODE>, or null if no such node exists.  */
    public Object getChild(Object parent, int index) {
	return ((FileTreeNode) parent).getChild(index);
    }

    /**
     * Returns the root node of this model */
    public Object getRoot() {
	return root;
    }

    /**
     * Returns the number of children of parent.  */
    public int getChildCount(Object parent) {
	return ((FileTreeNode) parent).getChildCount();
    }

    /**
     * Returns true of the node is a leaf node.  */
    public boolean isLeaf(Object node) {
	return ((FileTreeNode) node).isLeaf();
    }

    /**
     * This methods throws an UnsupportedOperationException since a
     * FileTreeModel should only be used in a non-editable JTree.  */
    public void valueForPathChanged(TreePath path, Object newValue) {
	throw new UnsupportedOperationException("A FileTreeModel should " +
						"only be used in a non-" +
						"editable JTree");
    }

    /**
     * Returns the index of child in parent.
     */
    public int getIndexOfChild(Object parent, Object child) {
	return ((FileTreeNode) parent).getIndexOfChild(child);
    }

    /**
     * Adds a listener for the tree model event posted after the tree
     * changes.  */
    public void addTreeModelListener(TreeModelListener l) {
	listeners.add(l);
    }

    /**
     * Removes a listener previously added with
     * <CODE>addTreeModelListener()</CODE>. */
    public void removeTreeModelListener(TreeModelListener l) {
	listeners.remove(l);
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * Adds the given file to the model as a child of the root.  */
    void add(File f) {
	root.add(f);
    }
    
    /**
     * Adds the given file to the model as a child of the root.  */
    void add(File f, FileFilter sel, boolean includeSubDirs) {
	root.add(f, sel, includeSubDirs);
    }


    /**
     * Returns true if the given file is a child of root.  */
    boolean contains(File f) {
	return root.contains(f);
    }


    /**
     * Removes the nodes specified by the given path.  */
    void remove(TreePath path) {
	int removePathLength = path.getPathCount();
	FileTreeNode parent = (FileTreeNode)
	    path.getPathComponent(removePathLength-2);
	FileTreeNode child = (FileTreeNode)
	    path.getPathComponent(removePathLength-1);
	parent.remove(child);
    }

    FileTreeModel copy() {
	return new FileTreeModel(this);
    }

    /**
     * Returns an array of all the regular files contained in this
     * model.  
     *
     * <pre><jml>
     * ensures \result != null &&
     *	       (\forall int i; 0 <= i && i < \result.length;
     *			\result[i].isFile());
     * </jml></pre>
     *
     */
    File[] regularFiles() {
	TreeSet accum = new TreeSet();
	root.accumFiles(accum);
	File[] result = new File[accum.size()];
	accum.toArray(result);
	return result;
    }

    //---------------------------------------------------------------------
    // HELPER CLASSES
    //---------------------------------------------------------------------

    /**
     * This inner class represents a node in the tree of files.  The
     * default implementation is for an inner node in the tree that
     * corresponds to an actual directory in the file system.
     * Subclasses handle leaf nodes for actual files and the root node
     * of the tree that does not correspond to an actual file or
     * directory.  */
    class FileTreeNode implements Comparable {

	// -------------------------------------------------------------
	// CONSTRUCTORS
	// -------------------------------------------------------------

	/**
	 * <pre><jml>
	 * requires f == null ==> this instanceof FileTreeRoot;
	 * requires f != null && !f.isDirectory() ==> 
	 *		this instanceof FileTreeLeaf;
	 * </jml></pre>
	 */
	FileTreeNode(File f, FileFilter sel, boolean includeSubDirs,
		     FileTreeNode parent) {
	    this.f = f;
	    this.sel = sel;
	    this.includeSubDirs = includeSubDirs;
	    this.parent = parent;
	    children = new TreeSet();
	}


	/**
	 * Creates a new file tree node that does not include
	 * subdirectories and without a file filter.  Equivalent to
	 * calling <CODE>FileTreeNode(f,null,false,parent)</CODE>. */
	FileTreeNode(File f, FileTreeNode parent) {
	    this(f,null,false,parent);
	}

	/**
	 * Creates a copy of the tree <CODE>old</CODE>. */
	FileTreeNode(FileTreeNode old, FileTreeNode parent) {
	    this(old.f, old.sel, old.includeSubDirs, parent);
	    Iterator iter = old.children.iterator();
	    while (iter.hasNext()) {
		FileTreeNode oldChild = (FileTreeNode) iter.next();
		// can't use factory method since we would get the
		// wrong outer class
		FileTreeNode childCopy = oldChild instanceof FileTreeLeaf ?
		    new FileTreeLeaf(oldChild,this) : 
		    new FileTreeNode(oldChild,this);
		children.add(childCopy);
	    }
	}

	/**
	 * Creates a copy of the tree rooted at this, giving the new
	 * tree the parent <CODE>newParent</CODE> */
	FileTreeNode copy(FileTreeNode newParent) {
	    FileTreeNode copy = new FileTreeNode(f, sel, includeSubDirs, 
						 newParent);
	    Iterator iter = this.children.iterator();
	    while (iter.hasNext()) {
		FileTreeNode thisChild = (FileTreeNode) iter.next();
		copy.children.add(thisChild.copy(copy));
	    }
	    return copy;
	}

	// -------------------------------------------------------------
	// TREE MODEL
	// -------------------------------------------------------------

	Object getChild(int index) {
	    if (index < 0 || index >= children.size()) {
		return null;
	    } else {
		// find the index-th child
		Iterator iter = children.iterator();
		int curIndex = 0;
		Object current = iter.next();
		while (curIndex < index) {
		    curIndex++;
		    current = iter.next();
		}
		return current;
	    }
	}

	int getChildCount() {
	    return children.size();
	}
	
	boolean isLeaf() {
	    return false;
	}

	int getIndexOfChild(Object child) {
	    Iterator iter = children.iterator();
	    int curIndex = -1;
	    while (iter.hasNext()) {
		curIndex++;
		Object current = iter.next();
		if (child == current) {
		    return curIndex;
		}
	    }
	    return -1;
	}

	// -------------------------------------------------------------
	// ACCESSORS
	// -------------------------------------------------------------

	FileTreeNode rebuildChildren() {
	    children.clear();
	    if (includeSubDirs) {
		// get all subdirs in f and for each add FileTreeNodes
		// to children, invoke rebuildChildren on each
		File[] subdirs = f.listFiles(DIRS_ONLY);
		for (int i = 0; i < subdirs.length; i++) {
		    children.add(new FileTreeNode(subdirs[i],sel,
						  includeSubDirs,this).
			rebuildChildren());
		}
	    }
	    
	    // get all regular files in f that sel matches and add
	    // FileTreeLeafs to children for each
	    File[] subfiles = f.listFiles(new java.io.FileFilter() {
		    public boolean accept(File subfile) {
			return subfile.isFile() &&
			    sel.accept(subfile);
		    }
		});
	    for (int i = 0; i < subfiles.length; i++) {
		children.add(new FileTreeLeaf(subfiles[i],this));
	    }
	    return this;
	}

	void accumFiles(TreeSet accum) {
	    Iterator iter = children.iterator();
	    while (iter.hasNext()) {
		((FileTreeNode) iter.next()).accumFiles(accum);
	    }
	}

 	void remove(FileTreeNode child) {
	    int pos = getIndexOfChild(child);
	    if (pos >= 0 && pos < children.size()) {
		children.remove(child);
		int ls = listeners.size();
		TreeModelEvent deleteEvent = 
		    new TreeModelEvent(FileTreeModel.this, 
				       pathTo(new ArrayList()),
				       new int[] { pos },
				       new Object[] { child } );
		for (int i = 0; i < ls; i++) {
		    ((TreeModelListener) listeners.get(i)).
			treeNodesRemoved(deleteEvent);
		}
	    }
	    
	}

	public Object[] pathTo(ArrayList reverseAccum) {
	    reverseAccum.add(this);
	    return parent.pathTo(reverseAccum);
	}

	public String toString() {
	    return f.getName();
	}

	public String description() {
	    return f.toString();
	}

	public boolean equals(/*@ nullable @*/ Object o) {
	    return false;
	}

	public boolean equals(/*@ nullable @*/ Object@FileTreeNode n) {
	    return n.f.equals(f);
	}

	public int compareTo(/*@ non_null @*/ Object o) {
	    throw new ClassCastException("cannot compare " + o.getClass() + 
					 " instance to a FileTreeNode");
	}

	public int compareTo(/*@ non_null @*/ Object@FileTreeNode n) {
	    return f.getName().compareTo(n.f.getName());
	}
	
	public int compareTo(/*@ non_null @*/ Object@FileTreeLeaf n) {
	    return -1;
	}
	
	// -------------------------------------------------------------
	// PRIVILEGED DATA
	// -------------------------------------------------------------

	File f;
	FileFilter sel;
	boolean includeSubDirs;
	FileTreeNode parent;

	/**
	 * The child nodes of this.
	 *
	 * <pre><jml>
	 * invariant (\forall Object o; children.contains(o);
	 *		o instanceof FileTreeNode);
	 * </jml></pre>
	 *
	 */
	TreeSet children;
    }


    /**
     * This inner class represents the root of the file tree model.  */
    final class FileTreeRoot extends FileTreeNode {

	// -------------------------------------------------------------
	// CONSTRUCTOR
	// -------------------------------------------------------------

	FileTreeRoot() {
	    super((File) null,null);
	}

	/**
	 * Creates a copy of the tree rooted at <CODE>old</CODE>. */
	FileTreeRoot(FileTreeRoot old) {
	    this();
	    Iterator iter = old.children.iterator();
	    while (iter.hasNext()) {
		FileTreeNode oldChild = (FileTreeNode) iter.next();
		// can't use factory method since we would get the
		// wrong outer class
		FileTreeNode childCopy = (oldChild instanceof FileTreeLeaf) ?
		    new FileTreeLeaf(oldChild,this) : 
		    new FileTreeNode(oldChild,this);
		children.add(childCopy);
	    }
	}

	// -------------------------------------------------------------
	// TREE MODEL
	// -------------------------------------------------------------

	// -------------------------------------------------------------
	// ACCESSORS
	// -------------------------------------------------------------
	
	FileTreeNode rebuildChildren() {
	    Iterator iter = children.iterator();
	    while (iter.hasNext()) {
		((FileTreeNode) iter.next()).rebuildChildren();
	    }
	    return this;
	}

	/**
	 * Adds f as an immediate child of this, unless it already is
	 * one.  
	 *
	 * <pre><jml>
	 * requires f.isFile() && !f.isDirectory();
	 * </jml></pre>
	 *
	 */
	void add(File f) {
	    FileTreeNode maybeNewNode = new FileTreeLeaf(f,this);
	    
	    if (!children.contains(maybeNewNode)) {
		children.add(maybeNewNode);
		int ls = listeners.size();
		TreeModelEvent addEvent = 
		    new TreeModelEvent(FileTreeModel.this, 
				       new Object[] { root },
				       new int[] { 
					   this.getIndexOfChild(maybeNewNode) 
				       },
				       new Object[] { maybeNewNode } );
		for (int i = 0; i < ls; i++) {
		    ((TreeModelListener) listeners.get(i)).
			treeNodesInserted(addEvent);
		}
	    }
	}

	/**
	 * Adds f as an immediate child of this, unless it already is
	 * one.  
	 *
	 * <pre><jml>
	 * requires !f.isFile() && f.isDirectory();
	 * </jml></pre>
	 *
	 */
	void add(File f, FileFilter sel, boolean includeSubDirs) {
	    FileTreeNode maybeNewNode = new FileTreeNode(f,sel,includeSubDirs,
							 this);
	    if (!children.contains(maybeNewNode)) {
		maybeNewNode.rebuildChildren();
		children.add(maybeNewNode);
		int ls = listeners.size();
		TreeModelEvent addEvent = 
		    new TreeModelEvent(this, new Object[] { root },
				       new int[] { 
					   this.getIndexOfChild(maybeNewNode) 
				       },
				       new Object[] { maybeNewNode } );
		for (int i = 0; i < ls; i++) {
		    ((TreeModelListener) listeners.get(i)).
			treeNodesInserted(addEvent);
		}
	    }
	}


	/**
	 * Returns true iff <CODE>children</CODE> includes a node
	 * containing <CODE>f</CODE>.  */
	boolean contains(File f) {
	    FileTreeNode compareNode;
	    if (f.isFile()) {
		compareNode = new FileTreeLeaf(f,this);
	    } else {
		compareNode = new FileTreeNode(f,this);
	    }
	    return children.contains(compareNode);
	}

	public Object[] pathTo(ArrayList reverseAccum) {
	    reverseAccum.add(this);
	    Collections.reverse(reverseAccum);
	    return reverseAccum.toArray();
	}

	public String toString() {
	    return "Files to process";
	}

	public String description() {
	    return "The files to be processed";
	}

	public boolean equals(/*@ nullable @*/ Object@FileTreeRoot r) {
	    return this == r;
	}

	public int compareTo(/*@ non_null @*/ Object@FileTreeRoot r) {
	    return (this == r) ? 0 : (this.hashCode() - r.hashCode());
	}
    }

    /**
     * This class represents a leaf in the file tree model.  It
     * corresponds to a normal file in the file system.  */
    final class FileTreeLeaf extends FileTreeNode {

	// -------------------------------------------------------------
	// CONSTRUCTOR
	// -------------------------------------------------------------

	FileTreeLeaf(File f, FileTreeNode parent) {
	    super(f,parent);
	    if (!f.isFile()) {
		throw new IllegalArgumentException("Tree leaves may only " +
						   "contain regular files.");
	    }
	}

	/**
	 * Creates a copy of the node <CODE>old</CODE>, giving the new
	 * node the parent <CODE>parent</CODE>. */
	FileTreeLeaf(FileTreeNode old, FileTreeNode parent) {
	    this(old.f, parent);
	}

	// -------------------------------------------------------------
	// TREE MODEL
	// -------------------------------------------------------------

	Object getChild(int index) {
	    return null;
	}

	int getChildCount() {
	    return 0;
	}
	
	boolean isLeaf() {
	    return true;
	}

	int getIndexOfChild(Object child) {
	    return -1;
	}

	// -------------------------------------------------------------
	// ACCESSORS
	// -------------------------------------------------------------

	FileTreeNode rebuildChildren() {
	    return this;
	}

	void accumFiles(TreeSet accum) {
	    accum.add(f);
	}

	public int compareTo(/*@ non_null @*/ Object@FileTreeNode n) {
	    return 1;
	}
	
	public int compareTo(/*@ non_null @*/ Object@FileTreeLeaf n) {
	    return f.getName().compareTo(n.f.getName());
	}
	
    }


    /**
     * Extends DefaultTreeCellRenderer to implement different tool tip
     * text for items in tree.  */
    final static class FileTreeCellRenderer extends DefaultTreeCellRenderer {

	/**
	 * Mutates this to be a component that when drawn represents
	 * value in tree with the given state information.  By
	 * mutating this instead of creating new components we avoid
	 * hundreds of object instantiations.  */
	public Component getTreeCellRendererComponent(JTree tree,
						      Object value,
						      boolean selected,
						      boolean expanded,
						      boolean leaf,
						      int row,
						      boolean hasFocus) {
	    super.getTreeCellRendererComponent(tree,value,selected,
					       expanded,leaf,row,hasFocus);
	    setToolTipText(((FileTreeNode)value).description());

	    return this;
	}
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    private final FileTreeRoot root;

    private final ArrayList listeners = new ArrayList();
    
    static final java.io.FileFilter DIRS_ONLY =
	new java.io.FileFilter() {
		public boolean accept(File subfile) {
		    return subfile.isDirectory();
		}
	    };
    
}
