/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: OpenDialog.java,v 1.3 2002/04/30 14:00:11 cclifton Exp $
 */

package org.multijava.samples.cmb.ui;

import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import javax.swing.filechooser.FileFilter;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import javax.swing.ToolTipManager;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * This class implements a dialog box for selecting the files to be
 * included in the index.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.3 $ */
public class OpenDialog extends JDialog 
    implements PropertyChangeListener, TreeSelectionListener
{

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public OpenDialog(File initialDirectory, final MainFrame owner,
		      FileTreeModel fileTree) {
	super(owner,"Open Program",true);
	setResizable(false);

	// content
	JComponent content = (JComponent) getContentPane();
	content.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

	// file selector on left
	fileChooser = new JFileChooser(initialDirectory);
	content.add(fileChooser,BorderLayout.WEST);
	//Enable tool tips.
	ToolTipManager.sharedInstance().registerComponent(fileChooser);
	fileChooser.setControlButtonsAreShown(false);
	fileChooser.addChoosableFileFilter(new FileFilter() {
		public boolean accept(File f) {
		    return f.isDirectory() ||
			f.getName().endsWith(".java");
		}
		public String getDescription() {
		    return "Java source files (*.java)";
		}
	    });
	fileChooser.addPropertyChangeListener(this);
	Border b =
	    BorderFactory.
	    createCompoundBorder(BorderFactory.
				 createTitledBorder("Files Available"),
				 BorderFactory.
				 createEmptyBorder(0,5,0,5));
	fileChooser.setBorder(b);

	// buttons in the middle
	Box buttonPane = Box.createVerticalBox();
	content.add(buttonPane,BorderLayout.CENTER);
	buttonPane.add(Box.createVerticalGlue());
	// GridLayout of 0,1 forces single column with as many rows as
	// necessary
	JPanel buttonGrid = new JPanel(new GridLayout(0,1));
	buttonPane.add(buttonGrid);
	//Enable tool tips.
	ToolTipManager.sharedInstance().registerComponent(buttonGrid);

	addButton = new JButton(">> Add >>");
	buttonGrid.add(addButton);
	addButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    addSelected();
		}
	    });
	addButton.setAlignmentX(CENTER_ALIGNMENT);
	addButton.setMnemonic(KeyEvent.VK_A);

	addAllButton = new JButton("Add all...");
	buttonGrid.add(addAllButton);
	addAllButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    addAllFiles();
		}
	    });
	addAllButton.setAlignmentX(CENTER_ALIGNMENT);
	addAllButton.setMnemonic(KeyEvent.VK_L);

	JLabel blankSpacerLabel = new JLabel();
	buttonGrid.add(blankSpacerLabel);

	removeButton = new JButton("Remove...");
	buttonGrid.add(removeButton);
	removeButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    removeSelected();
		}
	    });
	removeButton.setAlignmentX(CENTER_ALIGNMENT);
	removeButton.setMnemonic(KeyEvent.VK_R);

	buttonPane.add(Box.createVerticalGlue());

	// file tree on right
	this.fileTree = fileTree;
	progTree = new JTree(fileTree);
	//Enable tool tips.
	ToolTipManager.sharedInstance().registerComponent(progTree);
	progTree.setEditable(false);
	progTree.setRootVisible(false);
	progTree.setShowsRootHandles(true);
	progTree.setCellRenderer(new FileTreeModel.FileTreeCellRenderer());
	progTree.addTreeSelectionListener(this);
	progTree.getSelectionModel().
	    setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
	Dimension progSize = new Dimension(200,300);
	JScrollPane progPane = new JScrollPane(progTree);
	content.add(progPane,BorderLayout.EAST);
	progPane.setMinimumSize(progSize);
	progPane.setPreferredSize(progSize);
	b = BorderFactory.
	    createCompoundBorder(BorderFactory.
				 createTitledBorder("Files to Mine"),
				 BorderFactory.createEmptyBorder(5,5,5,5));
	progPane.setBorder(b);

	// control button at bottom
	JPanel controlPane = new JPanel();
	content.add(controlPane,BorderLayout.SOUTH);
	//Enable tool tips.
	ToolTipManager.sharedInstance().registerComponent(controlPane);

	JButton okButton = new JButton("OK");
	controlPane.add(okButton);
	okButton.setMnemonic(KeyEvent.VK_ACCEPT);
	okButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    changesMade = true;
		    OpenDialog.this.dispose();
		}
	    });
	JButton cancelButton = new JButton("Cancel");
	controlPane.add(cancelButton);
	cancelButton.setMnemonic(KeyEvent.VK_CANCEL);
	cancelButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    changesMade = false;
		    OpenDialog.this.dispose();
		}
	    });

	// handle button state
	handleAddButtonState();
	handleRemoveButtonState();

	// tidy things up
	pack();
	setLocationRelativeTo(owner);
    }
    

    /**
     * Displays a model dialog to select files. 
     * @return a new FileTreeModel reflecting the changes made by the
     * user, or null if no changes were made */
    static FileTreeModel show(File initialDirectory, final MainFrame owner,
			      FileTreeModel fileTree) {
	FileTreeModel copy = fileTree.copy();
	OpenDialog self = new OpenDialog(initialDirectory,owner,copy);
	// call to show blocks since dialog is modal
	self.show();
	if (self.changesMade) {
	    return copy;
	} else {
	    return null;
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    /**
     * Processes the add button.  */
    void addSelected() {
	File selected = fileChooser.getSelectedFile();
	//@ assert selected.isFile() && !selected.isDirectory();
	fileTree.add(selected);
	handleAddButtonState();
    }

    /**
     * Processes the remove button.  */
    void removeSelected() {
	String question = 
	    "Remove selected files from list of file to be processed?";
	int ans = JOptionPane.showConfirmDialog(this, question, "Remove",
						JOptionPane.YES_NO_OPTION);
	if (ans != JOptionPane.YES_OPTION) {
	    return;
	}
	
	TreePath selected = progTree.getSelectionPath();
	fileTree.remove(selected);
    }

    /**
     * Processes the add all button.  */
    void addAllFiles() {
	File curDir = fileChooser.getCurrentDirectory();
	FileFilter filter = fileChooser.getFileFilter();
	String question = "Add all " + filter.getDescription() + " in " + 
	    curDir.getName() + " directory?";
	String[] options = {
	    "Yes, without subdirectories",
	    "Yes, and include subdirectories",
	    "No"
	};
	int n = JOptionPane.showOptionDialog(this, question, "Add all", 
					     /* ignored */ 0, 
					     JOptionPane.QUESTION_MESSAGE, 
					     /* use default icon */ null, 
					     options, options[2]);
	boolean includeSubdirs = false;
	switch (n) {
	case 1:
	    includeSubdirs = true;
	    // fall through to next case!
	case 0:
	    fileTree.add( curDir, filter, includeSubdirs );
	    handleAddButtonState();
	    break;
	    
	    
	default:
	    break;
	}
    }

    private void handleAddButtonState() {
	File selFile = fileChooser.getSelectedFile();
	addButton.setEnabled(selFile != null && !fileTree.contains(selFile));

	File curDir = fileChooser.getCurrentDirectory();
	addAllButton.setEnabled(curDir != null && !fileTree.contains(curDir));
    }

    private void handleRemoveButtonState() {
	if (progTree.getLastSelectedPathComponent() == null) {
	    removeButton.setEnabled(false);
	} else {
	    removeButton.setEnabled(true);
	}
    }

    // -------------------------------------------------------------
    // IMPLEMENT LISTENERS
    // -------------------------------------------------------------

    public void propertyChange(PropertyChangeEvent evt) {
	String prop = evt.getPropertyName();
	if (prop.equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY) ||
	    prop.equals(JFileChooser.DIRECTORY_CHANGED_PROPERTY)) {
	    handleAddButtonState();
	}
    }

    public void valueChanged(TreeSelectionEvent evt) {
	handleRemoveButtonState();
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    private final JFileChooser fileChooser;
    private final JButton addButton;
    private final JButton addAllButton;
    private final JButton removeButton;
    private final JTree progTree;
    private final FileTreeModel fileTree;
    private boolean changesMade = false;
}
