/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: QueryPanel.java,v 1.5 2002/06/25 17:08:30 cclifton Exp $
 */

package org.multijava.samples.cmb.ui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.text.Caret;
import org.multijava.samples.cmb.query.*;
import org.multijava.samples.cmb.parser.*;

/**
 * This class represents the query panel for the Comment Mining Browser. 
 *
 * @author Curtis Clifton
 * @version $Revision: 1.5 $
 */
public class QueryPanel extends JPanel {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public QueryPanel(MainFrame mainFrame, ResultsPanel rPanel) {
	this.rPanel = rPanel;
	this.mainFrame = mainFrame;
	setLayout(new BorderLayout());
	
	// First the query parts
	{
	    JPanel queryPartsPanel = new JPanel();
	    queryPartsPanel.setLayout(new BoxLayout(queryPartsPanel,
						    BoxLayout.X_AXIS));
	    add(queryPartsPanel,BorderLayout.NORTH);

	    // Add the query text entries
	    JPanel queryEntryPanel = new JPanel();
	    queryEntryPanel.setLayout(new BoxLayout(queryEntryPanel,
						    BoxLayout.Y_AXIS));
	    Border b =
		BorderFactory.
		createCompoundBorder(BorderFactory.
				     createTitledBorder("Query expressions"),
				     BorderFactory.
				     createEmptyBorder(5,5,5,5));
	    queryEntryPanel.setBorder(b);
	    queryPartsPanel.add(queryEntryPanel);

	    // first the keyword query
	    JPanel keywordQueryPanel = new JPanel();
	    keywordQueryPanel.setLayout(new BoxLayout(keywordQueryPanel,
						      BoxLayout.X_AXIS));
	    queryEntryPanel.add(keywordQueryPanel);
	    JLabel kwLabel = new JLabel("Keyword query:");
	    kwQueryText = new JTextField(40);
	    keywordQueryPanel.add(kwLabel);
	    keywordQueryPanel.add(Box.createHorizontalStrut(5));
	    keywordQueryPanel.add(kwQueryText);
	    keywordQueryPanel.add(Box.createGlue());

	    // then the identifier query
	    JPanel identifierQueryPanel = new JPanel();
	    identifierQueryPanel.setLayout(new BoxLayout(identifierQueryPanel,
							 BoxLayout.X_AXIS));
	    queryEntryPanel.add(identifierQueryPanel);
	    JLabel idLabel = new JLabel("Identifier query:");
	    idQueryText = new JTextField(40);
	    identifierQueryPanel.add(idLabel);
	    identifierQueryPanel.add(Box.createHorizontalStrut(5));
	    identifierQueryPanel.add(idQueryText);
	    identifierQueryPanel.add(Box.createGlue());

	    // finally the buttons
	    JPanel buttonPanel = new JPanel();
	    buttonPanel.setBorder(BorderFactory.createEmptyBorder(5,0,5,0));
	    queryEntryPanel.add(buttonPanel);
	    JButton queryButton = new JButton("Query");
	    queryButton.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			processQuery();
		    }
		});
	    JButton clearButton = new JButton("Clear query");
	    clearButton.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			kwQueryText.setText("");
			idQueryText.setText("");
			selectAllDeclChecks();
		    }
		});
	    buttonPanel.add(queryButton);
	    buttonPanel.add(clearButton);

	    queryEntryPanel.add(Box.createGlue());

	    // Add the declaration kind checkboxes
	    queryPartsPanel.add(Box.createHorizontalStrut(5));
	    JPanel declKindPanel = new JPanel();
	    declKindPanel.setLayout(new GridLayout(2,2));
	    b = BorderFactory.
		createCompoundBorder(BorderFactory.
				     createTitledBorder("Declarations to " +
							"Display"),
				     BorderFactory.
				     createEmptyBorder(5,5,5,5));
	    declKindPanel.setBorder(b);
	    queryPartsPanel.add(declKindPanel);

	    classDeclCheck = new JCheckBox("Classes");
	    interfaceDeclCheck = new JCheckBox("Interfaces");
	    methodDeclCheck = new JCheckBox("Methods");
	    fieldDeclCheck = new JCheckBox("Fields");
	    selectAllDeclChecks();
	    declKindPanel.add(classDeclCheck);
	    declKindPanel.add(interfaceDeclCheck);
	    declKindPanel.add(methodDeclCheck);
	    declKindPanel.add(fieldDeclCheck);

	    // Left justify the subpanels
	    queryPartsPanel.add(Box.createGlue());
	}
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    // -------------------------------------------------------------
    // HELPERS
    // -------------------------------------------------------------

    /**
     * Called by the query button's action listener to process the
     * entered query.  */
    void processQuery() {
	final ProgressMonitor progMon =
	    new ProgressMonitor(this,"Processing Query","",0,4);
	progMon.setProgress(0);
	progMon.setMillisToDecideToPopup(1000);
	
	// create a thread to do the work
	Thread thrd = new Thread( new Runnable() {
		public void run() {
		    int p = 0;
		    // parse query text
		    ParseNode kwQuery;
		    ParseNode idQuery;
		    try {
			kwQuery = parseQuery(progMon,kwQueryText);
			progMon.setProgress(++p);
			idQuery = parseQuery(progMon,idQueryText);
			progMon.setProgress(++p);
		    } catch (QueryProcessingException e) {
			// parseQuery does the necessary work, we just need to
			// abort the query
			return;
		    }
		    
		    OccuranceList results = new OccuranceList();
		    if (kwQuery != null && idQuery != null) {
			// take intersection
			results = 
			    kwQuery.processQueryOn(mainFrame.kwIndex(),
						   catMatch);
			results = 
			    results.
			    intersect(idQuery.processQueryOn(mainFrame.idIndex(),
							     catMatch));
		    } else if (kwQuery != null) {
			// just use kw
			results = 
			    kwQuery.processQueryOn(mainFrame.kwIndex(),
						   catMatch);
		    } else if (idQuery != null) {
			// just use id
			results = 
			    idQuery.processQueryOn(mainFrame.idIndex(),
						   catMatch);
		    }
		    rPanel.display(results);
		}
	    });

	// Start the thread
	thrd.start();
    }

    private ParseNode parseQuery(ProgressMonitor progMon, JTextField queryText)
	throws QueryProcessingException 
    {
	ParseNode query;
	String qt = queryText.getText();
	if (qt.equals("")) {
	    query = null;
	} else {
	    try {
		query = QueryParser.parse(qt);
	    } catch (QueryProcessingException e) {
		progMon.close();
		Caret c = queryText.getCaret();
		c.setDot(qt.length());
		c.moveDot(e.position());
		queryText.requestFocus();
		JOptionPane.showMessageDialog(this, e.getMessage(),
					      "Bad Query Expression", 
					      JOptionPane.ERROR_MESSAGE);
		throw e;
	    }
	}
	
	return query;
    }

    void selectAllDeclChecks() {
	classDeclCheck.setSelected(true);
	interfaceDeclCheck.setSelected(true);
	methodDeclCheck.setSelected(true);
	fieldDeclCheck.setSelected(true);
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /**
     * The results panel that will display the results of this query
     * panel.  */
    private final ResultsPanel rPanel;
    private final MainFrame mainFrame;

    private JTextField kwQueryText;
    private JTextField idQueryText;

    private JCheckBox classDeclCheck;
    private JCheckBox interfaceDeclCheck;
    private JCheckBox methodDeclCheck;
    private JCheckBox fieldDeclCheck;

    Matcher catMatch = new Matcher() {
	    public boolean matches(Member m) {
		return false;
	    }
	    public boolean matches(Member@ClassDecl m) {
		return classDeclCheck.isSelected();
	    }
	    public boolean matches(Member@InterfaceDecl m) {
		return interfaceDeclCheck.isSelected();
	    }
	    public boolean matches(Member@MethodDecl m) {
		return methodDeclCheck.isSelected();
	    }
	    public boolean matches(Member@FieldDecl m) {
		return fieldDeclCheck.isSelected();
	    }
	};
}
