/*
 * Copyright (C) 2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: ArrayListCache.java,v 1.6 2003/12/27 04:49:13 davidcok Exp $
 */

package org.multijava.util;

import java.util.ArrayList;
import java.util.Collection;

/*@ model import org.jmlspecs.lang.JMLDataGroup; @*/

/**
 * This class implements a cache of ArrayLists.  It is an efficiency hack
 * to reduce the amount of object allocation and garbage collection do to 
 * short-term use of ArrayLists.  This class is _not_ synchronized.
 */
public final class ArrayListCache {

    // -------------------------------------------------------------
    // CONSTRUCTORS
    // -------------------------------------------------------------

    /**
     * Returns an empty ArrayList whose capacity is at least of the
     * given size.
     *
     * <pre><jml>
     * requires size > 0;
     * modifies cache;
     * ensures \result != null && !\reach(ArrayListCache.class).has(\result);
     * </jml></pre>
     *
     */
    public static ArrayList request(int size) {
	    return new ArrayList(size);
// 	if (pos < 0) {
// 	    return new ArrayList(size);
// 	} else {
// 	    ArrayList result = cacheArray[pos];
// 	    cacheArray[pos] = null;
// 	    pos--;
// 	    if (pos < shrinkAtSize && cacheSize > MIN_CACHE_SIZE) {
// 		shrinkCache();
// 	    }
// 	    result.ensureCapacity(size);
// 	    return result;
// 	}
    }

    /**
     * Returns an empty ArrayList of default capacity.
     *
     * <pre><jml>
     * modifies cache;
     * ensures \result != null && !\reach(ArrayListCache.class).has(\result);
     * </jml></pre>
     *
     */
    public static ArrayList request() {
	return new ArrayList();
// 	return request(DEFAULT_LIST_CAPACITY);
    }

    /**
     * Returns an ArrayList whose initial contents are the given collection.
     *
     * <pre><jml>
     * requires content != null;
     * modifies cache;
     * ensures \result != null && !\reach(ArrayListCache.class).has(\result);
     * </jml></pre>
     *
     */
    public static ArrayList request(Collection content) {
	    return new ArrayList(content);
// 	if (pos < 0) {
// 	    return new ArrayList(content);
// 	} else {
// 	    ArrayList result = request((int) (content.size() * 1.1));
// 	    result.addAll(content);
// 	    return result;
// 	}
    }

    /**
     * Adds the given ArrayList back to the cache.  Clients should not
     * clear the list before releasing it, as this method handles
     * that.  This method promises not to grab aliases to the list
     * members.
     *
     * <pre><jml>
     * requires al != null;
     * modifies cache;
     * ensures al.isEmpty() && 
     *   \reach(\old(al)).intersection(\reach(ArrayListCache.class)).isEmpty();
     * </jml></pre>
     */
    public static void release(ArrayList al) {
// 	al.clear();
// 	pos++;
// 	if (pos >= cacheSize) {
// 	    growCache();
// 	}
// 	cacheArray[pos] = al;
	/* !FIXME!8/27/2003: Profile the size of the cache.  We don't
	 * want it to get too big as that would curtail garbage
	 * collection.  Perhaps add a MAX_CACHE_SIZE and just discard
	 * once that is reached.*/
    }

    // !FIXME! Once we have generic types it would be handy to have
    // the following method:
    // public static <T> toArrayAndRelease(ArrayList al, <T> result) {
    // 	   al.toArray(result);
    // 	   release(al);
    // 	   return result;
    // }

    // -------------------------------------------------------------
    // CACHE ARRAY MANAGEMENT
    // -------------------------------------------------------------


    /**
     * Reduces the cache capacity.
     *
     * <pre><jml>
     * requires pos < shrinkAtSize && cacheSize > MIN_CACHE_SIZE;
     * ensures cacheSize < \old(cacheSize);
     * </jml></pre>
     */
    private static void shrinkCache() {
	int newSize = cacheSize / 2;
	if (newSize < MIN_CACHE_SIZE) {
	    newSize = MIN_CACHE_SIZE;
	}
	if (newSize < pos + 1) {
	    newSize = pos + 1;
	}
	
	ArrayList newCache[] = new ArrayList[newSize];
	System.arraycopy(cacheArray, 0, newCache, 0, newSize);

	cacheArray = newCache;
	cacheSize = newSize;
	shrinkAtSize = newSize / 4;
    }

    /**
     * Increases the cache capacity.
     *
     * <pre><jml>
     * ensures cacheSize > \old(cacheSize);
     * </jml></pre>
     *
     */
    private static void growCache() {
	int newSize = cacheSize * 2;
	
	ArrayList[] newCache = new ArrayList[newSize];
	System.arraycopy(cacheArray, 0, newCache, 0, cacheSize);

	cacheArray = newCache;
	cacheSize = newSize;
	shrinkAtSize = shrinkAtSize * 2;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    //@ public static model JMLDataGroup cache;

    static final int MIN_CACHE_SIZE = 16; // package protected for testing
    private static final int DEFAULT_LIST_CAPACITY = 10;

    /**
     * The cache of instantiated but currently unused ArrayList objects.
     */
    //@ private invariant cacheArray != null && cacheArray.length >= MIN_CACHE_SIZE;
    private static ArrayList[] cacheArray = new ArrayList[MIN_CACHE_SIZE];
    //@                            in cache;

    //@ private invariant cacheSize == cacheArray.length;
    private static int cacheSize = MIN_CACHE_SIZE;
    //@                    in cache;

    //@ private invariant shrinkAtSize < cacheArray.length;
    private static int shrinkAtSize = MIN_CACHE_SIZE / 4;
    //@                    in cache;

    /**
     * The highest index of a non-null object in the cache.
     */
    //@ private invariant pos >= -1;
    //@ private invariant cacheArray.length > pos;
    //@ private invariant pos >= 0 ==> cacheArray[pos] != null;
    //@ private invariant (\forall int i; i > pos && i < cacheArray.length; cacheArray[i] == null);
    private static int pos = -1;
    //@                  in cache;
}
