/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Destination.java,v 1.4 2004/01/16 16:26:56 cclifton Exp $
 */

package org.multijava.util;

import java.io.File;

/**
 * This class represents the destination of the output of one of the
 * tools included in the project.  If a destination is specified on
 * the command line and the input belongs to a package then the output
 * should be written to the concatenation of the destination directory
 * and the package names.  If no destination is specified then the
 * output should be written to the same directory as the input.  The
 * destination parameter from the command line is used to construct an
 * instance of this.  That instance provides File objects (via the
 * <code>...target()</code> methods) for writing particular
 * output files.  */
public class Destination extends File {
    private Destination( String destination, boolean unspecified ) {
	super( destination );
	this.unspecified = unspecified;
    }

    /**
     * Establish a new destination based on the command line argument.
     * If the argument is null then the destination is unspecified and
     * will be the current working directory.  */
    public static Destination check( String dest ) {
	if( dest == null ) {
	    return new Destination( System.getProperty( "user.dir" ), true );
	} else {
	    return new Destination( dest, false );
	}
    }

    /**
     * Returns the target for a text file, which is simply
     * <code>filename</code> appended to the destination directory. */
    public File textTarget( String filename ) {
	return new File( this, filename );
    }
    
    /**
     * Returns the target directory for writing a class file of the
     * given package and source file.  
     *
     * <pre><jml>
     * ensures unspecified && sourceFile != null ==> \result == sourceFile;
     * ensures unspecified && sourceFile == null ==> \result == this;
     * ensures !unspecified ==> \result.equals( new File( this, packageName ));
     * </jml></pre>
     *
     */
    public File classFileTargetDirectory( String packageName, 
					  File sourceFile ) {
	if( unspecified ) {
	    if (sourceFile == null) {
		return this;
	    } else {
		File parentDir = sourceFile.getParentFile();
		return parentDir == null ? this : parentDir;
	    }
	} else {
	    return new File( this, packageName );
	}
    }

    /**
     * Indicates whether the destination directory was unspecified on
     * the command line.  */
    //@ spec_public
    private boolean unspecified;
}
