/*
 * Copyright (C) 2003 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: DirectedAcyclicGraph.java,v 1.8 2004/01/16 16:26:56 cclifton Exp $
 */

package org.multijava.util;

/**
 * This utility class represents a directed acyclic graph.  It is
 * useful for sorting collections that contain some incomparable
 * members.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.8 $
 */
public class DirectedAcyclicGraph {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Constructs a directed acyclic graph with the given objects as
     * the vertices and the given calculator specifying the edges.
     *
     * <pre><jml>
     * requires vertices != null && calc != null &&
     *          acyclic(vertices, calc);
     * </jml></pre>
     *
     */
    public DirectedAcyclicGraph(Object[] vertices, EdgeCalculator calc) {
	this.vertex = vertices;
	edgeExists = new boolean[vertex.length][vertex.length];
	initEdges(calc);
    }

    /*@
      @ requires vertices != null && calc != null;
      @ ensures \result <==> 
      @         (\forall Object[] sequence; 
      @             sequence != null && sequence.length >= 2 &&
      @             sequence[0] == sequence[(int)(sequence.length - 1)] &&
      @             (\forall int i; 0 <= i && i < sequence.length;
      @                 (\exists int j; 0 <= j && j < vertices.length;
      @                      sequence[i] == vertices[j]));
      @             (\exists int k; 0 <= k && k < sequence.length - 1;
      @                 !calc.edgeExists(sequence[k],sequence[(int)(k+1)])));
      @ public pure model boolean acyclic(Object[] vertices, 
      @                                   EdgeCalculator calc); 
      @*/

    /**
     * Initializes the edges of the DAG representation using the given
     * calculator. */
    /*@ requires calc != null;
      @ assignable edgeExists[*];
      @ ensures (\forall int i,j; 0 <= i && i < vertex.length &&
      @                           0 <= j && j < vertex.length;
      @                  edgeExists[i][j] <==> 
      @                      calc.edgeExists(vertex[i],vertex[j]));
      @*/
    private /*@ helper @*/ void initEdges(EdgeCalculator calc) {
	for (int i = 0; i < vertex.length; i++) {
	    for (int j = 0; j < vertex.length; j++) {
		edgeExists[i][j] = calc.edgeExists(vertex[i],vertex[j]);
	    }
	}
    }

    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------

    //-------------------------------------------------------------
    // ACTIONS
    //-------------------------------------------------------------

    /**
     * Returns the vertices in depth-first search order.  Using
     * algorithm from Cormen, Leiserson, & Rivest, 1990, p. 478.
     */
    public Object[] inDFSOrder() {
	// for debugging
	// showDAG();

	result = new Object[vertex.length];
	nextResultSlot = vertex.length - 1;
	visited = new boolean[vertex.length];
	for (int i = 0; i < vertex.length; i++) {
	    if (!visited[i]) {
		visit(i);
	    }	    
	}
	return result;
    }

    /**
     * Helper method for depthFirstOrder.
     */
    private void visit(int pos) {
	visited[pos] = true;
	for (int j = 0; j < vertex.length; j++) {
	    if (edgeExists[pos][j] && !visited[j]) {
		visit(j);
	    }
	}
	result[nextResultSlot] = vertex[pos];
	nextResultSlot--;
    }

    private void showDAG() {
	System.out.println();
	for (int i = 0; i < vertex.length; i++) {
	    String header = i + ". " + vertex[i];
	    System.out.print(header);
	    for (int j = header.length(); j < 20; j++) {
		System.out.print(" ");
	    }
	    for (int j = 0; j < vertex.length; j++) {
		System.out.print(edgeExists[i][j] ? "Y " : "N ");
	    }
	    System.out.println();
	}
    }

    //-------------------------------------------------------------
    // INNER TYPES
    //-------------------------------------------------------------

    public interface EdgeCalculator {
	/**
	 * Returns true is a directed edge from o1 to o2 exists in the
	 * DAG. */
        //@ pure
	boolean edgeExists( Object o1, Object o2 );
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    /*@ private invariant vertex != null;
      @ private invariant edgeExists != null &&
      @           edgeExists.length == vertex.length &&
      @           (\forall int i; 0 <= i && i < vertex.length;
      @                           edgeExists[i] != null && 
      @                           edgeExists.length == vertex.length);
      @*/

    /**
     * The vertices of the DAG. */
    private final Object[] vertex;

    /**
     * The edges of the DAG.
     * edgeExists[i][j] ==> there is an edge from vertex[i] to vertex[j]
     */
    private final boolean[][] edgeExists;

    /**
     * Working space for the depth-first search algorithm
     */
    private Object[] result;

    /**
     * Index to the next entry to be filled in result.
     */
    private int nextResultSlot;

    /**
     * Tracks whether the depth-first search algorithm has visited a given vertex.
     */
    private boolean[] visited;
    
}
