/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: IntStack.java,v 1.6 2004/01/16 16:26:56 cclifton Exp $
 */

package org.multijava.util;
import java.util.EmptyStackException;

/**
 * This class represents a simple stack of integer values.  Unlike
 * java.util.Stack it is not synchronized and instead of storing
 * Objects it store int value.  Both of these changes make it faster
 * than using Stack.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.6 $ */

public final class IntStack {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------

    /**
     * Constructs a stack with default initial capacity.  */
    public IntStack() {
	this( DEFAULT_CAPACITY );
    }


    /**
     * Constructs a stack with the given initial capacity.  */
    public IntStack( int initialCapacity ) {
	minimumCapacity = initialCapacity;
	storage = new int[initialCapacity];
	top = -1;
    }
    
    //---------------------------------------------------------------------
    // ACCESSORS
    //---------------------------------------------------------------------


    /**
     * Returns true if this stack is empty.
     */
    public boolean isEmpty() {
	return top < 0;
    }


    /**
     * Returns the top item in the stack without modifying the stack.
     *
     * @exception EmptyStackException if the stack is empty
     */
    public int peek() {
	throwExceptionIfEmpty();
	return storage[top];
    }
    
    /**
     * Pops the top item from the stack and returns it.
     *
     * @exception EmptyStackException if the stack is empty
     */
    public int pop() {
	int result = peek();
	top -= 1;
	maybeShrinkStorage();
	return result;
    }


    /**
     * Pushes the given item onto the stack.
     */
    public void push(int i) {
	maybeExpandStorage();
	top += 1;
	storage[top] = i;
    }

    /**
     * Returns the size of this stack (i.e., the total number of items
     * in the stack.  */
    public /*@ pure @*/ int size() {
	return top + 1;
    }


    /**
     * Returns the capacity of the stack (i.e., the total number of
     * items that the stack can contains without automatically
     * allocating more storage space.  */
    public /*@ pure @*/ int capacity() {
	return storage.length;
    }

    /**
     * Throws an empty stack exception is the stack is empty.  This
     * avoids duplicating code in peek and pop.  Since the method is
     * private it can be inlined by compilers or JVMs.  */
    private void throwExceptionIfEmpty() {
	if (top == -1) {
	    throw new EmptyStackException();
	}
    }

    /**
     * Shrink the storage to release memory if the load factor of the
     * storage falls below MIN_LOAD_FACTOR.  */
    private void maybeShrinkStorage() {
	if ( storage.length <= minimumCapacity || 
	     top < shrinkIfTopLessThanThis ) {
	    return;
	}

	// need to shrink storage
	int newCapacity = storage.length / 2;
	if (newCapacity < minimumCapacity) {
	    newCapacity = minimumCapacity;
	}

	if (newCapacity < size() + 1) {
	    newCapacity = size() + 1;
	} // end of if 

	adjustCapacity(newCapacity);
    }


    /**
     * Expand the storage to accomodate more items.
     *
     * <pre><jml>
     * assignable storage, shrinkIfTopLessThanThis;
     * ensures size() < capacity();
     * </jml></pre>
     *
     */
    private void maybeExpandStorage() {
	if (size() + 1 < storage.length) {
	    return;
	} // end of if 
	
	// need to expand storage
	int newCapacity = storage.length * 2;
	adjustCapacity(newCapacity);
    }

    /**
     * Create new storage array of size newCapacity.  Copy existing
     * data to new array.  
     *
     * <pre><jml>
     * requires newCapacity >= size() + 1;
     * requires newCapacity >= minimumCapacity;
     * assignable storage, shrinkIfTopLessThanThis;
     * ensures storage.length == newCapacity;
     * </jml></pre>
     *
     */
    private void adjustCapacity( int newCapacity ) {
	int[] newStorage = new int[newCapacity];
	System.arraycopy( storage, 0, newStorage, 0, top + 1 );

	storage = newStorage;
	shrinkIfTopLessThanThis = (int) (MIN_LOAD_FACTOR * newCapacity);
    }

    //---------------------------------------------------------------------
    // DERIVED ATTRIBUTES
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // INTERFACE CHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // TYPECHECKING
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // CODE GENERATION
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------


    private int[] storage;
    private int top;
    private int minimumCapacity;
    private int shrinkIfTopLessThanThis;

    public static final int DEFAULT_CAPACITY = 10;
    public static final float MIN_LOAD_FACTOR = 0.25f;

}
