/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MjcHashRelation.java,v 1.14 2008/05/14 18:16:40 wdietl Exp $
 */

package org.multijava.util;

import java.util.Iterator;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Collection;

//@ model import org.jmlspecs.models.*;

/**
 * This class is used to map a key to a sequence of values, 
 * the image of the key. 
 * This allows a key to map to multiple values, i.e., 
 * this ADT represents a relation rather than a function. 
 */
public final class MjcHashRelation 
{
    /*@ public model JMLEqualsToEqualsRelation theRelation; 
      @ public initially theRelation.isEmpty();
      @*/
    /*@ public invariant
      @            (\forall Object k; k != null && theRelation.isDefinedAt(k);
      @			(\forall Object v; 
      @			    theRelation.elementImage(k).has(v);
      @			    v instanceof Comparable)); 
      @*/

    /**
     * <pre><jml>
     *    modifiable theRelation;
     *    ensures theRelation.isEmpty();
     * </jml></pre>
     */
    public MjcHashRelation() {
	theMap = new HashMap();
    }

    /**
     * <pre><jml>
     *    modifiable theRelation;
     *    ensures theRelation.isEmpty();
     * </jml></pre>
     */
    public MjcHashRelation(int size) {
	theMap = new HashMap(size);
    }

    /**
     *
     * <pre><jml>
     *    modifiable theRelation;
     *    ensures theRelation.isEmpty();
     * </jml></pre>
     */
    public void clear( ) {
	theMap.clear();
    }

    /**
     *
     * <pre><jml>
     *    ensures \result == theRelation.domain().int_size();
     * </jml></pre>
     */
    public /*@ pure @*/ int size( ) {
	return theMap.size();
    }

    /**
     *
     * <pre><jml>
     *    requires key != null;
     *    ensures \result == theRelation.isDefinedAt(key);
     * </jml></pre>
     */
    public /*@ pure @*/ boolean containsKey( /*@\readonly@*/ Object key ) {
	return theMap.containsKey(key);
    }

    /**
     * Returns true if there exists a key such that 
     * theRelation.elementImage(key).has( value ), 
     * otherwise it returns false.
     *
     * <pre><jml>
     *    requires value != null 
     *        && (\exists Object key; key != null;
     *                    theRelation.elementImage(key).has(value));
     *    ensures \result == true;
     *  also
     *    requires value != null 
     *        && (\forall Object key; key != null;
     *                    ! theRelation.elementImage(key).has(value));
     *    ensures \result == false;
     * </jml></pre>
     *
     */
    public /*@ pure @*/ boolean containsValue( /*@\readonly@*/ Object value ) {
	if ( ! (value instanceof Comparable) ) {
	    return false;
	}
	Iterator mapIter = theMap.values().iterator();
	while (mapIter.hasNext()) {
	    ArrayList valueList = (ArrayList) mapIter.next();
	    if (valueList.contains(value)) {
		return true;
	    }
	}
	return false;
    }

    public Object remove( /*@\readonly@*/ Object key ) {
	return theMap.remove(key);
    }

    /**
     *    requires key != null && value != null 
     *        && theRelation.isDefinedAt(key)
     *    modifiable theRelation;
     *    ensures  theRelation == 
     *                 \old(theRelation).removeFromDomain(key).add(key, value);
     *             && \result == null;
     */
    public Object remove( /*@\readonly@*/ Object key, /*@\readonly@*/ Comparable value ) {
	ArrayList valueList = (ArrayList) theMap.get(key);
	if ( valueList != null ) {
	    int ix = findValue(valueList, (Comparable) value);
	    if ( ix < valueList.size() ) {
		return valueList.remove(ix);
	    }
	}
	return null;
    }

    /**
     * Returns the least value in the set of values mapped to 
     * by the given key (if the image of the given key has been 
     * sorted), or null if this key is not in the domain 
     * of the theRelation.
     *
     * <pre><jml>
     *    requires key != null && theRelation.isDefinedAt(key);
     *    ensures (* \result == the first element in the list of values 
     *                          in theRelation.elementImage(key) *);
     *  also
     *    requires key != null && ! theRelation.isDefinedAt(key);
     *    ensures \result == null;
     * </jml></pre>
     *
     */
    public /*@ pure @*/ Comparable get( /*@\readonly@*/ Object key ) {
	ArrayList valueList = (ArrayList) theMap.get(key);
	if (valueList != null) {
	    return (Comparable) valueList.get(0);
	}
	return null;
    }

    /**
     * Returns the collection of values mapped to 
     * by the given key, or null if this key is not in the domain 
     * of the theRelation.
     *
     * <pre><jml>
     *    requires key != null && theRelation.isDefinedAt(key);
     *    ensures \result.equals(theRelation.elementImage(key));
     *  also
     *    requires key != null && ! theRelation.isDefinedAt(key);
     *    ensures \result == null;
     * </jml></pre>
     *
     */
    public /*@ pure @*/ Collection getImage( /*@\readonly@*/ Object key ) {
	return (ArrayList) ((ArrayList) theMap.get(key)).clone();
    }

    /**
     * Adds value to the set of values in the image of the given key.
     *
     * <pre><jml>
     *    requires key != null && value != null 
     *        && ! (\exists Object aValue; aValue != null; 
     *                      theRelation.elementImage(key).has( aValue )
     *                      && aValue == value);
     *    requires_redundantly (* the (key, value) pair is not already in 
     *                            theRelation *);
     *    modifiable theRelation;
     *    ensures theRelation.equals( \old(theRelation).add(key, value) )
     *         && \old(theRelation).elementImage(key).has(\result);
     *  also
     *    requires key != null && value != null 
     *        && (\exists Object aValue; aValue != null; 
     *                    theRelation.elementImage(key).has( aValue )
     *                    && aValue == value);
     *    requires_redundantly (* the (key, value) pair is already in 
     *                            theRelation *);
     *    ensures \result == null;
     * </jml></pre>
     *
     */
    public Comparable put( /*@\readonly@*/ Object key, /*@\readonly@*/ Comparable value ) {
	ArrayList valueList = (ArrayList) theMap.get(key);
	if (valueList == null ) {
	    valueList = new ArrayList();
	    valueList.add( value );
	    theMap.put(key, valueList);
	    return null;
	} else {
	    int ix = findValue(valueList, value);
	    if ( valueList.size() <= ix ) {
		// value is not in the valueList
		// (duplicates cannot be allowed).
		// The difference (ix - valueList.size()) is the 
		// position where the item should be inserted.
		valueList.add(ix-valueList.size(), value);
	    }
	    if ( valueList.size() > 1) {
		return (Comparable) valueList.get(valueList.size() - 2);
	    } else {
		return null;
	    }
	}
    }

    /**
     * Adds a mapping from the given key to the given image.
     *
     * <pre><jml>
     * requires key != null && image != null &&
     *          (\forall Object o; image.contains(o); o instanceof Comparable);
     * </jml></pre>
     */
    public Collection putImage( /*@\readonly@*/ Object key, /*@\readonly@*/ Collection image ) {
	ArrayList old = (ArrayList) theMap.get(key);
	ArrayList newImage = new ArrayList(image);
	Collections.sort(newImage);
	theMap.put(key, newImage);
	return old;
    }

   public boolean sortImage( /*@\readonly@*/ Object key ) {
	ArrayList valueList = (ArrayList) theMap.get(key);
	if ( valueList != null ) {
	    try {
		Collections.sort(valueList);
		return true;
	    } catch (RuntimeException e) {
		return false;
	    }
	} else {
	    return false;
	}
    }

    public Comparable getNextValue( /*@\readonly@*/ Object key, /*@\readonly@*/ Comparable value ) {
	return getNextValueFromN( new MjcHashRelation[] { this },
				  key, value );
    }

    public static Comparable getNextValueFromN( MjcHashRelation[] rels,
						/*@\readonly@*/ Object key,
						/*@\readonly@*/ Comparable value ) {
	ArrayList valueList = new ArrayList();
	for (int i = 0; i < rels.length; i++) {
	    ArrayList relsVals = (ArrayList) rels[i].theMap.get(key);
	    if (relsVals != null) {
		valueList.addAll(relsVals);
	    }
	}
	Collections.sort(valueList);
	if (valueList.size() > 1) {
	    try {
		Collections.sort(valueList);
	    } catch (RuntimeException e) {
		// elements are incomparable for some reason ???
		// should not happen
	    }
	    int ix = findValue(valueList, value);
	    if (ix < valueList.size() - 1) {
		return (Comparable) valueList.get(ix + 1);
	    } else {
		return null;
	    }
	}
	return null;
    }

    /**
     * This method is used to place the elements in theList in order. 
     * Returns the index where value is located in theList or where 
     * value could be inserted (by subtracting the size of theList) 
     * if it is not already in the list.
     * 
     * @param	value		the object being searched for 
     */
    protected static int findValue( ArrayList theList, Comparable value ) {
	int i;
	int size = theList.size();
	for (i = 0; i<size ;i++) {
	    Object currVal = theList.get(i);
	    if (currVal == value) {
		return i;
	    }
	    if ( 0 < value.compareTo(currVal) ) {
		// value > currVal
		// we add size to indicate that value was not found;
		// this also means that value can be inserted at 
		// the position (\result - size) and theList will 
		// still be sorted.
		return i + size;
	    }
	}
	return i;
    }

    private HashMap theMap = new HashMap();
    //@             in theRelation;

    //@ private represents theRelation <- abstractValue();
    /*@ private pure model JMLEqualsToEqualsRelation abstractValue() {
      @     JMLEqualsToEqualsRelation ret = new JMLEqualsToEqualsRelation();
      @     JMLEqualsSetEnumerator iter = theMap.theMap.elements();
      @     while (iter.hasMoreElements()) {
      @        java.util.Map.Entry e
      @           = (java.util.Map.Entry)(iter.nextElement());
      @        ret.add(e.getKey(), e.getValue());
      @     }
      @     return ret;
      @ }
      @*/

    // The following representation invariant is necessary so that 
    // getNextValue() will work properly.
    /*@ private invariant (* the representation does not relate an object, k,
      @                      to more another object, v, more than once. *);
      @*/

}
