/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Options.java,v 1.45 2006/08/18 03:46:16 leavens Exp $
 */

package org.multijava.util;

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedHashSet;
import org.multijava.util.compiler.CompilationAbortedException;



/**
 * This class implements basic option parsing.  Subclasses add tool-specific
 * options to the set of options that are accepted.  Subclasses are typically 
 * generated using the option generating utility 
 * {@link org.multijava.util.optgen.Main}.
 */
public abstract class Options {

    /**
     * Creates a new Option object.
     * @param   name            the command name to pass to getopt
     */
    public Options(String name) {
        this.name = name;
    }

    public boolean parseCommandLine(String[] argv) {
        Getopt  parser = new Getopt(name, argv, getShortOptions(), 
                                    getLongOptions(), true);

        for (;;) {
            int code;

            code = parser.getopt();
            if (code == -1) {
                break;
            }
            if (!processOption(code, parser)) {
                return false;
            }
        }

        // store remaining arguments
        nonOptions = new String[argv.length - parser.getOptind()];
        System.arraycopy(argv, parser.getOptind(), nonOptions, 0, 
                         nonOptions.length);

        return true;
    }

    /**
     * Parses and evaluates the command line arguments.
     * @param   argv            the command line arguments
     * @param   infiles         where to put non-option arguments
     * @return true iff the command line was parsed successfully
     */
    public boolean parseCommandLine(String[] argv, ArrayList infiles) {
        boolean result = parseCommandLine(argv);

        if (result) for (int i = 0; i < nonOptions.length; i++) {
            infiles.add(nonOptions[i]);
        }

        return result;
    }

    public boolean setOption(String name, Object newValue) {
        if (name.equals("help")) {
            help();
            throw new CompilationAbortedException(true);
        } else if (name.equals("version")) {
            printVersion();
            throw new CompilationAbortedException(true);
        } else if (name.equals("Xnoversion")) {
            Xnoversion = ((Boolean)newValue).booleanValue();
        } else {
            return false;
        }
        return true;
    }
        
    /**
     * Handles the option identified by the given <CODE>code</CODE>,
     * perhaps using <CODE>g</CODE> to process any arguments to the
     * option.  This method is typically overridden in subclasses to
     * process additional arguments with the subclass defering to the
     * superclass if <CODE>code</CODE> is unknown.
     *
     * @param   g            the command line arguments
     */
    public boolean processOption(int code, Getopt g) {
        switch (code) {
        case 'h':
            help();
            throw new CompilationAbortedException(true);
        case 'V':
            printVersion();
            throw new CompilationAbortedException(true);
        case NOVERSION_CASE:
            Xnoversion = true;
            return true;
        default:
            return false;
        }
    
    }

    public HashMap getOptions() {
        HashMap result = new HashMap();
        result.put("help",
                   "  --help, -h:                           Displays the help information");
        result.put("version",
                   "  --version, -V:                        Prints out the version information");
        result.put("Xnoversion",
                   "  --Xnoversion:                         Omits printing version, for testing");
        return result;
    }

    /**
     * Prints the options
     */
    public void printOptions() {
        HashMap optMap = getOptions();
        String[] options = new String[optMap.size()];
        optMap.values().toArray(options);
        Arrays.sort(options);

        System.out.println();
        System.out.println("Options:");

        for (int i = 0; i < options.length; i++) {
            System.out.println(options[i]);
        }
    }
    
    /**
     * Returns the longname
     */
    public LinkedHashSet getLongname() {
    LinkedHashSet longname = new LinkedHashSet();
    longname.add("help");
    longname.add("version");
    longname.add("Xnoversion");
        
    return longname;
    }
        
    /**
     * Returns the header for a table
     */
    public Hashtable getTableHeader() {
    Hashtable tableHeader = new Hashtable();
        
    return tableHeader;
    }
    
    /**
     * Returns the type
     */
    public Hashtable getType() {
    Hashtable type = new Hashtable();
    type.put("help", "null");
    type.put("version", "null");
    type.put("Xnoversion", "null");
        
    return type;
    }
    
    /**
     * Returns the default value
     */
    public Hashtable getDefaultValue() {
    Hashtable defaultValue = new Hashtable();
    defaultValue.put("help", "");
    defaultValue.put("version", "");
    defaultValue.put("Xnoversion", "");
        
    return defaultValue;
    }
        
    /**
     * Returns the current value
     */
    public Hashtable getCurrentValue() {
    Hashtable currentValue = new Hashtable();
    currentValue.put("help", "");
    currentValue.put("version", "");
    currentValue.put("Xnoversion", "");
        
    return currentValue;
    }
    
    /**
     * Returns the list of selections for an option
     */
    public Hashtable getSelectionList() {
    Hashtable selectionList = new Hashtable();
        
    return selectionList;
    }
    
    /**
     * Returns the help messages
     */
    public Hashtable getHelpString() {
    Hashtable helpString = new Hashtable();
    helpString.put("help", getOptions().get("help"));
    helpString.put("version", getOptions().get("version"));
    helpString.put("Xnoversion", getOptions().get("Xnoversion"));
        
    return helpString;
    }
    
    /**
     * Returns the GUI categories that this option is in
     */
    public Hashtable getGuiType() {
    Hashtable guiType = new Hashtable();
    guiType.put("help", "advanced");
    guiType.put("version", "advanced");
    guiType.put("Xnoversion", "advanced");
        
    return guiType;
    }

    /**
     * shows an usage message
     */
    public abstract void help();

    /**
     * Returns the version information
     */
    protected String version() { return VERSION; }

    /**
     * Prints the version information
     */
    public void printVersion() {
        System.out.println(Xnoversion?"--- VERSION INFO ---":version());
    }

    /**
     * shows a help message
     */
    protected abstract void usage();

    // ----------------------------------------------------------------------
    // UTILITIES
    // ----------------------------------------------------------------------

    protected int getInt(Getopt g, int defaultValue) {
        try {
            return g.getOptarg() != null ?
                new Integer(g.getOptarg()).intValue() :
                defaultValue;
        } catch (Exception e) {
            System.err.println("malformed option: " + g.getOptarg());
            throw new org.multijava.util.compiler.CompilationAbortedError();
        }
    }

    protected String getString(Getopt g, String defaultValue) {
        return g.getOptarg() != null ?
            g.getOptarg() :
            defaultValue;
    }
        
    protected String getString(String newValue) {
        return (newValue.equals("")) ?
            null :
            newValue;
    }
        
    protected String getNonNullString(String str) {
        return str != null ? 
            str : 
            "";
    }
        
    protected String[][] getNonNullStringArray(String[][] str) {
        return str != null ?
            str :
            new String[0][0];
    }
        
    protected String getSystemProperty(String property) {
        String classPath = "";
        java.util.Properties props = System.getProperties();
        
        if (property.equals("CLASSPATH")) {
            if (System.getProperty("java.class.path") != null) {
                classPath += System.getProperty("java.class.path");
            }
            
            String      version = System.getProperty("java.version");
            classPath = System.getProperty("java.class.path");
            if (System.getProperty("sun.boot.class.path") != null) {
                classPath += File.pathSeparator 
                    + System.getProperty("sun.boot.class.path");
            } else if (version.startsWith("1.2") 
                || version.startsWith("1.3")) {
                String  home = System.getProperty("java.home");
                classPath += File.pathSeparator + home + File.separatorChar +
                    "lib" + File.separatorChar + "rt.jar";
            }
        } else {
            classPath = System.getProperty(property);
        }
        return classPath;
    }
        
    protected String[] formOneColumnArray(String firstArg) {
        String[] newArray = { firstArg };
        return newArray;
    }
    
    protected String[] formTwoColumnArray(String firstArg, String secondArg) {
        String[] newArray = { firstArg, secondArg };
        return newArray;
    }

    // ----------------------------------------------------------------------
    // DEFAULT OPTIONS
    // ----------------------------------------------------------------------

    /**
     * Gets short options
     */
    public String getShortOptions() {
        return "hV";
    }

    /**
     * Gets long options
     */
    public LongOpt[] getLongOptions() {
        return new LongOpt[] {
            new LongOpt("help", LongOpt.NO_ARGUMENT, null, 'h'),
            new LongOpt("version", LongOpt.NO_ARGUMENT, null, 'V'),
            new LongOpt("Xnoversion", LongOpt.NO_ARGUMENT, null, NOVERSION_CASE)
        };
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * The array of non-option arguments.
     */
    private String[] nonOptions;
    public String[] nonOptions() { 
        return (String[]) nonOptions.clone();
    }

    private final String                name;

    private static final String VERSION = 
        // "Version 1.3.2 (released August 17, 2006)";
        "CVS version after release 1.3.2 (released August 17, 2006)";

    private boolean           Xnoversion = false;
    public boolean Xnoversion() { return Xnoversion; }

    private static final int           NOVERSION_CASE = 999;
}
