/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestUtils.java,v 1.6 2004/01/16 16:26:56 cclifton Exp $
 */

package org.multijava.util;

import java.io.File;
import java.io.FileFilter;


/**
 * TestMain.java
 *
 * @author Curtis Clifton
 * @version $Revision: 1.6 $
 */

public class TestUtils extends junit.framework.TestCase {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    public TestUtils( String name ) {
	super( name );
    }
    
    //---------------------------------------------------------------------
    // unescapeString
    //---------------------------------------------------------------------

    private void test_unescapeStrings_helper( String[] expect, String[] orig) {
	assertEquals( "test config. problem, array lengths should match", 
		      expect.length, orig.length );
	for (int i = 0; i < orig.length; i++) {
	    assertEquals( expect[i], Utils.unescapeString( orig[i] ));
	} // end of for
    }


    public void test_unescapeString_1() {
	String[] orig = new String[] { "plain", "vanilla", "strings" };
	String[] expect = orig;
	test_unescapeStrings_helper( expect, orig );
    }

    public void test_unescapeString_2() {
	String[] orig = new String[] { "\\\\slash", "\\\"quote", "\\'apost" };
	String[] expect = new String[] { "\\slash", "\"quote", "'apost" };
	test_unescapeStrings_helper( expect, orig );
    }

    public void test_unescapeString_3() {
	String[] orig = new String[] { "\\nnewline", "\\rreturn", "\\ttab" };
	String[] expect = new String[] { "\nnewline", "\rreturn", "\ttab" };
	test_unescapeStrings_helper( expect, orig );
    }

    public void test_unescapeString_4() {
	String[] orig = new String[] { "\\bbackspace", "\\fformfeed" };
	String[] expect = new String[] { "\bbackspace", "\fformfeed" };
	test_unescapeStrings_helper( expect, orig );
    }

   public void test_unescapeString_5() {
	String[] orig = new String[] { "\\060", "\\101", "\\172" };
	String[] expect = new String[] { "0", "A", "z" };
	test_unescapeStrings_helper( expect, orig );
    }

    public void test_unescapeString_6() {
	String[] orig = new String[] { "\\u0030", "\\u0041", "\\u007A" };
	String[] expect = new String[] { "0", "A", "z" };
	test_unescapeStrings_helper( expect, orig );
    }

    //---------------------------------------------------------------------
    // escapeString
    //---------------------------------------------------------------------

    private void test_escapeStrings_helper( String[] expect, String[] orig ) {
	assertEquals( "test config. problem, array lengths should match", 
		      expect.length, orig.length );
	for (int i = 0; i < orig.length; i++) {
	    assertEquals( expect[i], Utils.escapeString( orig[i] ));
	} // end of for
    }


    public void test_escapeString_1() {
	String[] orig = new String[] { "plain", "vanilla", "strings" };
	String[] expect = orig;
	test_escapeStrings_helper( expect, orig );
    }

    public void test_escapeString_2() {
	String[] orig = new String[] { "\\slash", "\"quote", "'apost" };
	String[] expect = new String[] { "\\\\slash", "\\\"quote", "\\'apost" };
	test_escapeStrings_helper( expect, orig );
    }

    public void test_escapeString_3() {
	String[] orig = new String[] { "\nnewline", "\rreturn", "\ttab" };
	String[] expect = new String[] { "\\nnewline", "\\rreturn", "\\ttab" };
	test_escapeStrings_helper( expect, orig );
    }

    public void test_escapeString_4() {
	String[] orig = new String[] { "\bbackspace", "\fformfeed" };
	String[] expect = new String[] { "\\bbackspace", "\\fformfeed" };
	test_escapeStrings_helper( expect, orig );
    }


    // -------------------------------------------------------------
    // relativePathTo
    // -------------------------------------------------------------
    
    public void test_relativePathTo_1() {
	assertEquals( ".", Utils.relativePathTo(userDir) );
	File[] files = userDir.listFiles();
	assertFiles( files, "" );
    }

    public void test_relativePathTo_2() {
	File[] files = userDir.getParentFile().listFiles();
	assertFiles( files, ".." + fileSep );
    }
    
    public void test_relativePathTo_3() {
	File[] subdirs = 
	    userDir.listFiles( new FileFilter() {
		    public boolean accept(File target) {
			return target.isDirectory();
		    }
		});
	File[] files = subdirs[0].listFiles();
	assertFiles( files, subdirs[0].getName() + fileSep );
    }

    // -------------------------------------------------------------
    // HELPERS
    // -------------------------------------------------------------

    /**
     * For each <CODE>i</CODE> in <CODE>0..files.length</CODE> this
     * asserts that <CODE>Utils.relativePathTo(files[i])</CODE> equals
     * <CODE>prefix + files[i].getName()</CODE> */
    private void assertFiles( File[] files, String prefix ) {
	for (int i = 0; i < files.length; i++) {
	    // special case to skip userDir
	    if (!files[i].equals(userDir)) {
		assertEquals( prefix + files[i].getName(),
			      Utils.relativePathTo(files[i]) );
	    }
	}
    }

    // -------------------------------------------------------------
    // SETUP
    // -------------------------------------------------------------

    public void setUp() {
	userDir = new File(System.getProperty("user.dir"));
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------
    
    private final static String fileSep = System.getProperty("file.separator");

    private File userDir;

}// TestMain
