/*
 * Copyright (C) 2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Annotation.java,v 1.1 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;

/**
 * This class represents an annotation structure in a classfile.
 *
 * @author AHS
 */
public class Annotation {
    
    /**
     * Read an annotation structure from a classfile.
     */
    public Annotation(ConstantPool cp, DataInput in)
    		throws IOException, ClassFileFormatException {
        int num_pairs;

        name = (AsciiConstant) cp.getEntryAt(in.readShort());
        typeName = name.getValue();
        num_pairs = in.readUnsignedShort();
        
        for (int i = 0; i < num_pairs; i++) {
            elementValuePairs.add(new ElementValuePair(cp, in));
        }
    }
    
    /**
     * Create an empty annotation of a given type. We can't use reflection yet,
     * as long as MJ doesn't support Java 5 stuff.
     */
    public Annotation(String typeName) {
        this.typeName = typeName;
        this.name = new AsciiConstant("L" + typeName.replace('.', '/') + ";");
    }
    
    
    /**
     * Add an element-value pair to the annotation.
     */
    public void addElement(String name, AnnotElementValue value) {
        elementValuePairs.add(new ElementValuePair(name, value));
    }
    
    /**
     * Return the value of the given element in the notation, or null
     * if the annotation does not contain such an element.
     */
    public AnnotElementValue getValue(String elementName) {
        for (int i = 0; i < elementValuePairs.size(); i++) {
            ElementValuePair pair = (ElementValuePair) elementValuePairs.get(i);
            if (pair.name().equals(elementName)) {
                return pair.value(); 
            }
        }
        return null;
    }
    
    /**
     * Return the name of the annotation, i.e. the name of the interface type.
     */
    public String getName() {
        return typeName;
    }
    
    
    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param cp	the constant pool for this class
     */    
    /* package */ void resolveConstants(ConstantPool cp) {
        cp.addItem(name);
        for (int i = 0; i < elementValuePairs.size(); i++) {
            ((ElementValuePair) elementValuePairs.get(i)).resolveConstants(cp);
        }
    }
    
    /**
     * Write the annotation structure to a classfile.
     */
    /* package */ void write(ConstantPool cp, DataOutput out) throws IOException {
        out.writeShort(name.getIndex());
        out.writeShort(elementValuePairs.size());
        for (int i = 0; i < elementValuePairs.size(); i++) {
            ((ElementValuePair)elementValuePairs.get(i)).write(cp, out);
        }
    }    

    
    /**
     * Return the size of this annotation structure in bytes.
     */ 
    /* package */ int getSize() {
        int size = 2 + 2;         
        for (int i = 0; i < elementValuePairs.size(); i++) {
            size += ((ElementValuePair) elementValuePairs.get(i)).getSize();
        }
        return size;
    }
    
    /**
     * Return a string representation of this annotation.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("@");
        sb.append(name.getValue());
        sb.append("(");
        for (int i = 0; i < elementValuePairs.size(); i++) {
            sb.append(elementValuePairs.get(i).toString());
            if (i != elementValuePairs.size()-1)
                sb.append(",");
        }
        sb.append(")");
        return sb.toString();
    }

    
    /**
     * Nested class representing name/value pair in an annotation structure.
     */
    private class ElementValuePair {
        
        /**
         * Construct a new element-value pair for an annotation structure.
         * 
         * @param name	the name of the pair
         * @param value	the value of the pair
         */
        public ElementValuePair(String name, AnnotElementValue value) {
            this.name = new AsciiConstant(name);
            this.value = value;
        }
        
        /**
         * Construct an element-value pair in an annotation structure from a
         * classfile stream.
         * 
         * @param cp	the constant pool
         * @param in	the stream to read from
         *
         * @throws IOException	an IO exception has occured
         * @throws ClassFileFormatException	the pair is malformed
         */
        public ElementValuePair(ConstantPool cp, DataInput in)
        	throws IOException, ClassFileFormatException
        {
            name = (AsciiConstant) cp.getEntryAt(in.readUnsignedShort());
            char tag = (char) in.readUnsignedByte();
            
            // TODO: Support for array, class, enum and annotation types
            switch (tag) { // read tag (VMS Table 4.2)
            case 'B': value = new AnnotByteElementValue(cp, in); break;
            case 'C': value = new AnnotCharElementValue(cp, in); break;
            case 'D': value = new AnnotDoubleElementValue(cp, in); break;
            case 'F': value = new AnnotFloatElementValue(cp, in); break;
            case 'I': value = new AnnotIntegerElementValue(cp, in); break;
            case 'J': value = new AnnotLongElementValue(cp, in); break;
            case 'S': value = new AnnotShortElementValue(cp, in); break;
            case 'Z': value = new AnnotBooleanElementValue(cp, in); break;
            case 's': value = new AnnotStringElementValue(cp, in); break;
            default:
                throw new ClassFileFormatException("Unsupported tag "
                        + tag + " in element-value pair");
            }
        }
        
        /**
         * Insert or check location of constant value on constant pool.
         *
         * @param cp	the constant pool for this class
         */
        /* package */ void resolveConstants(ConstantPool cp) {
            cp.addItem(name);
            value.resolveConstants(cp);
        }
        
        /**
         * Write this element-value pair to a stream.
         * 
         * @param cp	the constant pool
         * @param out	the output stream 
         * @throws IOException	an IO exception has occured
         */
        /* package */ void write(ConstantPool cp, DataOutput out) throws IOException {
            out.writeShort(name.getIndex());
            value.write(cp, out);
        }
        
        
        /**
         * Return the size in bytes of this element-value pair.
         */
        protected int getSize() {
            return 2 + 1 + value.getSize();
        }
        
        /**
         * Return a string representation of this element-value pair.
         */
        public String toString() {
            return name.getValue() + "=" + value.toString();
        }

            
        // ACCESSORS
        
        /**
         * Return the value of this element-value pair.
         */
        public AnnotElementValue value() {
            return value;
        }
        
        /**
         * Return the name of this element-value pair.
         */
        public String name() {
            return name.getValue();
        }

            
        // PRIVATE DATA MEMBERS

        /** the pooled constant holding the name of the pair */
        private AsciiConstant	name;
        
        /** the value of the pair */
        private AnnotElementValue	value;
    }


    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private String		typeName;
    private AsciiConstant	name;
    
    /** The element-value pairs associated with this annotation */
    private ArrayList		elementValuePairs = new ArrayList();
}
