/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: AttributeList.java,v 1.13 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.ListIterator;
import org.multijava.mjc.Debug;
import org.multijava.util.Utils;

/**
 * This class represents a list of attributes within a
 * <CODE>ClassInfo</CODE> structure (see JVM2, 4.7).  This class also
 * includes static helper methods for parsing attributes from
 * <CODE>.class</CODE> files.  This parsing is done using the
 * chain-of-responsibility pattern.  See the Attribute class for a
 * sample member of the chain.
 * */
public class AttributeList {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Constructs an empty attribute list
     */
    public AttributeList() {
	attributes = EMPTY;
    }

    /**
     * Constructs a one element attribute list
     */
    public AttributeList(Attribute att1) {
	attributes = att1 == null ? EMPTY : new Attribute[]{att1};
    }

    /**
     * Constructs a two elements attribute list
     */
    public AttributeList(Attribute att1, Attribute att2) {
	attributes = att1 == null && att2 == null ? EMPTY : 
	    att1 != null && att2 != null ? new Attribute[] {att1, att2} :
	    new Attribute[] {att1 == null ? att2 : att1};
    }

    /**
     * Constructs a three elements attribute list
     */
    public AttributeList(Attribute att1, Attribute att2, Attribute att3) {
	this(att1, att2, att3, null);
    }

    /**
     * Constructs a four elements attribute list
     */
    public AttributeList(Attribute att1, Attribute att2, Attribute att3, Attribute att4) {
	int count = (att1 == null ? 0 : 1) + 
	    (att2 == null ? 0 : 1) + (att3 == null ? 0 : 1) +
	    (att4 == null ? 0 : 1);
    
	attributes = new Attribute[count--];
    
	if (att4 != null) {
	    attributes[count--] = att4;
	}
	if (att3 != null) {
	    attributes[count--] = att3;
	}
	if (att2 != null) {
	    attributes[count--] = att2;
	}
	if (att1 != null) {
	    attributes[count--] = att1;
	}
    }

    /**
     * Constructs an attribute list from a class file stream
     *
     * @param	in		the stream to read from
     * @param	cp		the constant pool
     * @param	noCode		do not read code attribute
     *
     * @exception	java.io.IOException	an io problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					write a bad classfile info
     */
    AttributeList(DataInput in, ConstantPool cp, boolean noCode)
	throws IOException, ClassFileFormatException
    {
	int		count = in.readUnsignedShort();
	if (count > 0) {
	    ArrayList attributes = new ArrayList();
	    for (int i = 0; i < count; i += 1) {
		if (noCode) {
		    attributes.add(readInterfaceOnly(in, cp));
		} else {
		    attributes.add(read(in, cp));
		}
	    }
	    this.attributes = new Attribute[attributes.size()];
	    attributes.toArray(this.attributes);
	    
	} else {
	    attributes = EMPTY;
	}
    }

    /**
     * Constructs an sub-attribute list of CodeInfo from a class file stream
     *
     * @param	in		the stream to read from
     * @param	cp		the constant pool
     * @param	insns		(sparse) array of instructions
     *
     * @exception	java.io.IOException	an io problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					write a bad classfile info
     */
    AttributeList(DataInput in, ConstantPool cp, Instruction[] insns)
	throws IOException, ClassFileFormatException
    {
	int count = in.readUnsignedShort();

	ArrayList attributes = new ArrayList(count);
	for (int i = 0; i < count; i += 1) {
	    attributes.add(readCodeInfoAttribute(in,cp,insns));
	}
	this.attributes = new Attribute[attributes.size()];
	attributes.toArray(this.attributes);
	
    }

    // -------------------------------------------------------------
    // ATTRIBUTE PARSING CHAIN-OF-RESPONSIBILITY
    // -------------------------------------------------------------

    /**
     * This method calls each member of the attribute parsing
     * chain-of-responsibility in order to read the next attribute in
     * the given input stream.  Once a member of the chain matches no
     * additional members are called.  If no member matches then
     * returns a GenericAttribute. */
    static Attribute read(DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException
    {
        AsciiConstant name = 
	    (AsciiConstant)cp.getEntryAt(in.readUnsignedShort());

	String tag = name.getValue();

	ParserIterator iter = parserIterator();
	Attribute result = null;
	while (result == null && iter.hasNext()) {
	    result = iter.next().read(tag,in,cp);
	}
	
	return result == null ? new GenericAttribute(name, in, cp) : result;
    }

    /**
     * This method calls each member of the attribute parsing
     * chain-of-responsibility in order to read the next attribute in
     * the given input stream.  Once a member of the chain matches no
     * additional members are called.  If no member matches then
     * returns a GenericAttribute. */
    static Attribute readInterfaceOnly(DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException
    {
	AsciiConstant name = 
	    (AsciiConstant)cp.getEntryAt(in.readUnsignedShort());

	String tag = name.getValue();

	ParserIterator iter = parserIterator();
	Attribute result = null;
	while (result == null && iter.hasNext()) {
	    result = iter.next().readInterfaceOnly(tag,in,cp);
	}
	
	return result == null ? new GenericAttribute(name, in, cp) : result;
    }

    /**
     * This method calls each member of the attribute parsing
     * chain-of-responsibility, in order, to read the next CodeInfo
     * attribute in the given input stream.  Once a member of the
     * chain matches no additional members are called.  If no member
     * matches then returns a GenericAttribute. */
    static Attribute readCodeInfoAttribute(DataInput in, ConstantPool cp, 
					   Instruction[] insns)
	throws IOException, ClassFileFormatException
    {
	AsciiConstant name = 
	    (AsciiConstant)cp.getEntryAt(in.readUnsignedShort());

	String tag = name.getValue();

	ParserIterator iter = parserIterator();
	Attribute result = null;
	while (result == null && iter.hasNext()) {
	    result = iter.next().readCodeInfoAttribute(tag,in,cp, insns);
	}
	
	return result == null ? new GenericAttribute(name, in, cp) : result;
    }

    /**
     * Adds the given AttributeParser to the head of the
     * chain-of-responsibility.  */
    public static void addParser(AttributeParser prsr) {
	attrParsers.addFirst(prsr);
    }

    /**
     * Removes the given parser from the chain-of-responsibility.  */
    public static void removeParser(AttributeParser prsr) {
	attrParsers.remove(prsr);
    }

    /**
     * Removes all parsers, including the default tail element
     * <CODE>BaseAttributeParser</CODE>, from the
     * chain-of-responsibility.  */
    public static void removeAllParsers() {
	attrParsers.clear();
    }

    /**
     * Resets the chain-of-responsibility to just include the default
     * tail element <CODE>BaseAttributeParser</CODE>.  */
    public static void resetParsers() {
	attrParsers.clear();
	attrParsers.addFirst( new BaseAttributeParser() );
    }

    /**
     * Returns a specialized iterator that returns successive parsers
     * from the chain-of-responsibility.  */
    public static ParserIterator parserIterator() {
	return new ParserIterator(attrParsers);
    }

    /**
     * This nested class implements a specialized iterator that
     * returns successive parsers from the chain-of-responsibility and
     * keeps clients from having to perform casts.  It also prevents
     * representation exposure of the chain-of-responsibility data
     * structure (though parsers in the chain are exposed).  */
    public static class ParserIterator {
	ParserIterator( LinkedList prsrs ) {
	    this.prsrs = prsrs.listIterator();
	}
	
	public boolean hasNext() {
	    return prsrs.hasNext();
	}

	public AttributeParser next() {
	    return (AttributeParser) prsrs.next();
	}

	private ListIterator prsrs;
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    public boolean isEmpty() {
	return attributes.length > 0;
    }

    /**
     * Return the first attribute for this tag
     *
     * @param	tag		the tag that identifies the attribute
     */
    public final Attribute get(int tag) {
	for (int i = 0; i < attributes.length; i++) {
	    if (attributes[i].getTag() == tag) {
		return attributes[i];
	    }
	}

	return null;
    }

    /**
     * Adds an attribute to the list of attributes
     *
     * @param	attr	the attribute to add to the list, for convenience the
     *			argument may be null in which case this does nothing
     */
    public final void add(Attribute attr) {
	if (attr == null) {
	    return;
	}
	
	// if an attribute with same tag exists, replace it
	for (int i = 0; i < attributes.length; i++) {
	    if (attributes[i].getTag() == attr.getTag()) {
		attributes[i] = attr;
		return;
	    }
	}
    
	// add the new attribute at the end of the list
	Attribute[]	temp = new Attribute[attributes.length + 1];
	System.arraycopy(attributes, 0, temp, 0, attributes.length);
	attributes = temp;
	attributes[attributes.length - 1] = attr;
    }

    /**
     * Removes an attribute from the list of attributes
     *
     * @param	tag		the tag that identifies the attribute
     * @return	true iff an attribute with given tag was in the list
     */
    /*package*/ final boolean remove(int tag) {
	for (int i = 0; i < attributes.length; i++) {
	    if (attributes[i].getTag() == tag) {
		Attribute[]	temp = new Attribute[attributes.length - 1];

		System.arraycopy(attributes, 0, temp, 0, i);
		System.arraycopy(attributes, i + 1, temp, i, attributes.length - i);
		attributes = temp;

		return true;
	    }
	}

	return false;
    }

    /**
     * Returns the space in bytes used by this attribute in the classfile
     */
    /*package*/ int getSize() {
	int		size = 0;

	for (int i = 0; i < attributes.length; i++) {
	    size += attributes[i].getSize();
	}

	return size;
    }

    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant values in constant pool
     *
     * @param	cp		the constant pool for this class
     */
    /*package*/ void resolveConstants(ConstantPool cp) 
	throws ClassFileFormatException 
    {
	for (int i = 0; i < attributes.length; i++) {
            attributes[i].resolveConstants(cp);
	}
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool
     *
     * @param	cp		the constant pool that contain all data
     * @param	out		the file where to write this object info
     *
     * @exception	java.io.IOException	an io problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					write a bad classfile info
     */
    /*package*/ void write(ConstantPool cp, DataOutput out)
	throws IOException, ClassFileFormatException
    {
	out.writeShort(attributes.length);
	for (int i = 0; i < attributes.length; i++) {
	    attributes[i].write(cp, out);
	}
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private Attribute[]		attributes;

    private static final LinkedList attrParsers = new LinkedList();

    public static final Attribute[] EMPTY = new Attribute[0];

    static {
	resetParsers();
    }
}

