/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: AttributeParser.java,v 1.1 2002/05/31 21:51:34 cclifton Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;


/**
 * This interface is implemented by all classes which can participate
 * in the chain-of-responsibility for parsing Attributes from
 * bytecode.  The chain-of-responsibility is established via static
 * methods of {@link AttributeList}.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.1 $ */

public interface AttributeParser {

    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed */
    Attribute read(String tag, DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException;

    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized, but skips CodeInfo
     * attributes.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed */
    Attribute readInterfaceOnly(String tag, DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException;

    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized, but only parses attributes
     * that may appear within a CodeInfo attribute.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @param	insns		(sparse) array of instructions
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed or if the
     * next bytes are not valid within a CodeInfo attribute */
    Attribute readCodeInfoAttribute(String tag, DataInput in, ConstantPool cp,
				    Instruction[] insns)
	throws IOException, ClassFileFormatException;
}
