/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: BaseAttributeParser.java,v 1.6 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;
import org.multijava.mjc.Debug;

/**
 * This class implements a parser for regular Java bytecode attributes as
 * defined in the JVM specification.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.6 $
 */
public class BaseAttributeParser implements AttributeParser {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    // default constructor
    

    // -------------------------------------------------------------
    // IMPLEMENT AttributeParser
    // -------------------------------------------------------------
    
    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed */
    public Attribute read(String tag, DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException
    {
	switch (tag.charAt(0)) {
	case 'C':
	    if (tag.equals("Code")) {
		return new CodeInfo(in, cp);
	    } else if (tag.equals("ConstantValue")) {
		return new ConstantValueAttribute(in, cp);
	    }
	    break;

	case 'D':
	    if (tag.equals("Deprecated")) {
		return new DeprecatedAttribute(in, cp);
	    }
	    break;

	case 'E':
	    if (tag.equals("Exceptions")) {
		return new ExceptionsAttribute(in, cp);
	    }
	    break;

	case 'I':
	    if (tag.equals("InnerClasses")) {
		return new InnerClassTable(in, cp);
	    }
	    break;

	case 'L':
	    if (tag.equals("LineNumberTable")) {
		throw new ClassFileFormatException("Attribute \"LineNumberTable\" illegal outside of Attribute Code");
	    } else if (tag.equals("LocalVariableTable")) {
		throw new ClassFileFormatException("Attribute \"LocalVariableTable\" illegal outside of Attribute Code");
	    }
	    break;
	    
	// AHS: parse runtime-visible annotations
	case 'R':
	    if (tag.equals("RuntimeVisibleAnnotations")) {
	        return new RuntimeVisibleAnnotationsAttribute(in, cp);
	    } else if (tag.equals("RuntimeVisibleParameterAnnotations")) {
	        return new RuntimeVisibleParameterAnnotationsAttribute(in, cp);
	    }
	    break;

	case 'S':
	    if (tag.equals("SourceFile")) {
		return new SourceFileAttribute(in, cp);
	    } else if (tag.equals("Synthetic")) {
		return new SyntheticAttribute(in, cp);
	    } else if(tag.equals(SignatureAttribute.NAME)){
		return new SignatureAttribute(in, cp);
	    }
	    break;

	default:
	    break;
	}

	return null;
    }

    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized, but skips CodeInfo
     * attributes.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed */
    public Attribute readInterfaceOnly(String tag, DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException
    {
	switch (tag.charAt(0)) {
	case 'C':
	    if (tag.equals("Code")) {
		return new SkippedCodeInfo(in, cp);
	    } else if (tag.equals("ConstantValue")) {
		return new ConstantValueAttribute(in, cp);
	    }
	    break;

	case 'D':
	    if (tag.equals("Deprecated")) {
		return new DeprecatedAttribute(in, cp);
	    }
	    break;

	case 'E':
	    if (tag.equals("Exceptions")) {
		return new ExceptionsAttribute(in, cp);
	    }
	    break;

	case 'I':
	    if (tag.equals("InnerClasses")) {
		return new InnerClassTable(in, cp);
	    }
	    break;

	case 'L':
	    if (tag.equals("LineNumberTable")) {
		throw new ClassFileFormatException("Attribute \"LineNumberTable\" illegal outside of Attribute Code");
	    } else if (tag.equals("LocalVariableTable")) {
		throw new ClassFileFormatException("Attribute \"LocalVariableTable\" illegal outside of Attribute Code");
	    }
	    break;

	// AHS: parse RuntimeVisibleAnnotations attribute
	case 'R':
	    if (tag.equals("RuntimeVisibleAnnotations")) {
	        return new RuntimeVisibleAnnotationsAttribute(in, cp);
	    } else if (tag.equals("RuntimeVisibleParameterAnnotations")) {
	        return new RuntimeVisibleParameterAnnotationsAttribute(in, cp);
	    }
	    break;

	case 'S':
	    if (tag.equals("SourceFile")) {
		return new SourceFileAttribute(in, cp);
	    } else if (tag.equals("Synthetic")) {
		return new SyntheticAttribute(in, cp);
	    } else if (tag.equals(SignatureAttribute.NAME)){
		return new SignatureAttribute(in, cp);
	    }
	    break;

	default:
	    break;
	}

	return null;
    }

    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized, but only parses attributes
     * that may appear within a CodeInfo attribute.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @param	insns		(sparse) array of instructions
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed or if the
     * next bytes are not valid within a CodeInfo attribute */
    public Attribute readCodeInfoAttribute(String tag, DataInput in, 
					   ConstantPool cp, Instruction[] insns)
	throws IOException, ClassFileFormatException
    {
	switch (tag.charAt(0)) {
	case 'C':
	    if (tag.equals("Code")) {
		throw new ClassFileFormatException("Attribute \"Code\" illegal as sub-attribute of Attribute Code");
	    } else if (tag.equals("ConstantValue")) {
		throw new ClassFileFormatException("Attribute \"ConstantValue\" illegal as sub-attribute of Attribute Code");
	    }
	    break;

	case 'D':
	    if (tag.equals("Deprecated")) {
		throw new ClassFileFormatException("Attribute \"Deprecated\" illegal as sub-attribute of Attribute Code");
	    }
	    break;

	case 'E':
	    if (tag.equals("Exceptions")) {
		throw new ClassFileFormatException("Attribute \"Exceptions\" illegal as sub-attribute of Attribute Code");
	    }
	    break;

	case 'I':
	    if (tag.equals("InnerClasses")) {
		throw new ClassFileFormatException("Attribute \"InnerClasses\" "
						   + "illegal as sub-attribute "
						   + "of Attribute Code");
	    }
	    break;

	case 'L':
	    if (tag.equals("LineNumberTable")) {
		return new LineNumberTable(in, cp, insns);
	    } else if (tag.equals("LocalVariableTable")) {
		return new LocalVariableTable(in, cp, insns);
	    }
	    break;

	// AHS: parse RuntimeVisibleAnnotations attribute
	case 'R':
	    if (tag.equals("RuntimeVisibleAnnotations")) {
	        throw new ClassFileFormatException("Attribute \"RuntimeVisibleAnnotations\" " +
	                "illegal as sub-attribute of Attribute Code");
	    } else if (tag.equals("RuntimeVisibleParameterAnnotations")) {
	        throw new ClassFileFormatException("Attribute \"RuntimeVisibleParameterAnnotations\" " +
	                "illegal as sub-attribute of Attribute Code");
	    }
	    break;

	    
	case 'S':
	    if (tag.equals("SourceFile")) {
		throw new ClassFileFormatException("Attribute \"SourceFile\" " + 
						   "illegal as sub-attribute of"
						   + " Attribute Code");
	    } else if (tag.equals("Synthetic")) {
		throw new ClassFileFormatException("Attribute \"Synthetic\" " + 
						   "illegal as sub-attribute of"
						   + " Attribute Code");
	    }
	    break;

	default:
	    break;
	}

	return null;
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

}
