/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: ClassInfo.java,v 1.15 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.BufferedOutputStream;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;

import org.multijava.util.Destination;
import org.multijava.util.Utils;
import org.multijava.mjc.Debug;
/**
 * VMS 4.1 Class File
 *
 * This is the place where all information about the class defined
 * by this class file resides.
 */
public class ClassInfo extends Member {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Constructs a new class info structure from source.
     *
     * @param	modifiers	access permission to and properties of this class
     * @param	thisClass	the class or interface defined by this class file
     * @param	superClass	the superclass of this class
     * @param	interfaces	the interfaces implemented by this class
     * @param	fields		the fields which are members of this class
     * @param	methods		the methods which are members of this class
     * @param	innerClasses	the inner classes which are members of this class
     * @param	attributes	a list of the additional custom attributes for
     *				the classinfo, or null if none.  This reference
     *				is captured.
     * @param	sourceFile	the source file
     * @param	deprecated	is this class deprecated ?
     */
    public ClassInfo(short modifiers,
		     String thisClass,
		     String superClass,
		     String genericSignature,
		     ArrayList interfaces,
		     ArrayList fields,
		     ArrayList methods,
		     InnerClassInfo[] innerClasses,
		     AttributeList attributes,
		     File sourceFile,
		     boolean deprecated) {
	this(modifiers,
	     thisClass,
	     superClass,
	     genericSignature,
	     makeInterfacesArray(interfaces),
	     makeFieldInfoArray(fields),
	     makeMethodInfoArray(methods),
	     innerClasses,
	     attributes,
	     sourceFile,
	     deprecated);
    }

    /**
     * Constructs a new class info structure from source.
     *
     * @param	modifiers	access permission to and properties of this class
     * @param	thisClass	the class or interface defined by this class file
     * @param	superClass	the superclass of this class
     * @param	interfaces	the interfaces implemented by this class
     * @param	fields		the fields which are members of this class
     * @param	methods		the methods which are members of this class
     * @param	innerClasses	the inner classes which are members of this class
     * @param	attributes	a list of the additional custom attributes for
     *				the classinfo, or null if none. This reference is
     *				captured.
     * @param	sourceFile	the source file
     * @param	deprecated	is this class deprecated ?
     */
    public ClassInfo(short modifiers,
		     String thisClass,
		     String superClass,
		     String genericSignature,
		     ClassConstant[] interfaces,
		     FieldInfo[] fields,
		     MethodInfo[] methods,
		     InnerClassInfo[] innerClasses,
		     AttributeList attributes,
		     File sourceFile,
		     boolean deprecated) {
	minorVersion = JAVA_MINOR;
	majorVersion = JAVA_MAJOR;

	setModifiers((short)(modifiers | ACC_SUPER));
	this.thisClass = new ClassConstant(thisClass);
	this.superClass = superClass != null ?
	    new ClassConstant(superClass) :
	    null;
	this.fields = fields;
	this.methods = methods;
	this.interfaces = interfaces;
	this.sourceFile = sourceFile;

	InnerClassTable ict = innerClasses != null ? 
	    new InnerClassTable(innerClasses) : null;
	SourceFileAttribute sfa = sourceFile != null ? 
	    new SourceFileAttribute(sourceFile.getName()) : null;
	DeprecatedAttribute da = deprecated ? new DeprecatedAttribute() : null;
	SignatureAttribute sa = genericSignature != null? 
	    new SignatureAttribute(genericSignature):null;
	if (attributes == null) {
	    this.attributes = new AttributeList(ict, sfa, sa ,da);
	} else {
	    this.attributes = attributes;
	    this.attributes.add(ict);
	    this.attributes.add(sfa);
	    this.attributes.add(sa);
	    this.attributes.add(da);
	}
    }

    public ClassInfo(DataInput in, boolean interfaceOnly)
        throws IOException, ClassFileFormatException {
        this(in, interfaceOnly, ClassInfoCreator.getInstance());
    }

    /**
     * Constructs a class info structure from a class file
     *
     * @param	in		the stream to read the class from
     * @param	interfaceOnly	load only the interface, not the source code
     *
     * @exception	java.io.IOException	an io problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					write a bad classfile info
     */
    protected ClassInfo(DataInput in, 
                        boolean interfaceOnly, 
                        ClassInfoCreator cr)
        throws IOException, ClassFileFormatException {
	int magic = in.readInt();
	if (magic != JAVA_MAGIC) {
	    throw new ClassFileFormatException("Bad magic number: " + magic);
	}

	minorVersion = in.readUnsignedShort();
	majorVersion = in.readUnsignedShort();

	//  !FIXME! may want to search the attributes for a source
	//  file name and constructor a legitimate File, though it
	//  isn't clear how that File's path would be found, perhaps
	//  the sourceFile should be passed as an argument?
	sourceFile = null;

	ConstantPool constantPool = new ConstantPool(in);

	setModifiers((short)in.readUnsignedShort());
	thisClass    = 
	    (ClassConstant)constantPool.getEntryAt(in.readUnsignedShort());
	superClass   = 
	    (ClassConstant)constantPool.getEntryAt(in.readUnsignedShort());

	interfaces   = new ClassConstant[in.readUnsignedShort()];
	for (int i = 0; i < interfaces.length; i += 1) {
	    interfaces[i] = 
		(ClassConstant)constantPool.getEntryAt(in.readUnsignedShort());
	}

	fields = new FieldInfo[in.readUnsignedShort()];
	for (int i = 0; i < fields.length; i += 1) {
	    fields[i] = cr.createFieldInfo(in, constantPool);
	}
	methods = new MethodInfo[in.readUnsignedShort()];
	for (int i = 0; i < methods.length; i += 1) {
	    methods[i] = cr.createMethodInfo(in, constantPool, interfaceOnly);
	}
	this.attributes = new AttributeList(in, constantPool, false);

	constantPool.close();
	constantPool = null; // save memory
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Returns the name of the this class (fully qualified)
     */
    public String getName() {
	return thisClass.getName();
    }

    public String getGenericSignature() {
	Attribute att = attributes.get(Constants.ATT_SIGNATURE);
	return att ==null ? null:
	    ((SignatureAttribute)att).getSignature();
    }

    public void setGenericSignature(String type) {
	attributes.add(new SignatureAttribute(type));
    }
    /**
     * Sets the name of the this field (fully qualified)
     */
    public void setName(String name) {
	thisClass = new ClassConstant(name);
    }

    /**
     * Returns the type of the this field
     */
    public String getSignature() {
	return thisClass.getName();
    }

    /**
     * Returns the super class of the class in the file
     */
    public String getSuperClass() {
	return superClass == null ? null : superClass.getName();
    }

    /**
     * Sets the super class of the class in the file
     */
    public void setSuperClass(String superClass) {
	this.superClass = superClass == null ? null : new ClassConstant(superClass);
    }

    /**
     * Returns the version of the class in the file
     */
    public int getMajorVersion() {
	return majorVersion;
    }

    /**
     * Sets the version of the class in the file
     */
    public void setMajorVersion(int majorVersion) {
	this.majorVersion = majorVersion;
    }

    /**
     * Returns the version of the class in the file
     */
    public int getMinorVersion() {
	return minorVersion;
    }

    /**
     * Sets the version of the class in the file
     */
    public void setMinor(int minorVersion) {
	this.minorVersion = minorVersion;
    }

    /**
     * Returns the inner classes table of the class in the file
     */
    public InnerClassInfo[] getInnerClasses() {
	Attribute		attr = attributes.get(Constants.ATT_INNERCLASSES);

	return attr == null ? null : ((InnerClassTable)attr).getEntries();
    }

    /**
     * Sets the inner classes table of the class in the file
     */
    public void setInnerClasses(InnerClassInfo[] inners) {
	if (inners != null) {
	    attributes.add(new InnerClassTable(inners));
	} else {
	    attributes.remove(Constants.ATT_INNERCLASSES);
	}
    }

    /**
     * Returns the source file of the class in the file, or null if
     * the actual source file is unknown.
     *
     * @see #sourceFileName()
     */
    public File sourceFile() {
	return sourceFile;
    }


    /**
     * Returns the short name of the source file as stored in the
     * attributes of this.  */
    public String sourceFileName() {
	Attribute attr = attributes.get(Constants.ATT_SOURCEFILE);

	return attr == null ? null : ((SourceFileAttribute)attr).getValue();
    }
    
    /**
     * Return the version of the encoding for the universe modifiers in
     * the universe attributes or annotations. It will first check if the
     * universe_class attribute is present, if not it will check for
     * an annotation.
     * By AHS.
     */
    public String getUniverseVersion() {
        Attribute attr = attributes.get(Constants.ATT_UNIVERSE_VERSION);
	if (attr == null) { // attribute not found, look for annotation
	    attr = attributes.get(Constants.ATT_RUNTIME_VISIBLE_ANNOTATIONS);
	    if (attr == null)
	        return null;
	    Annotation annot = ((RuntimeVisibleAnnotationsAttribute)attr).get("org.multijava.universes.UniverseClass");
	    if (annot == null)
	        return null;
	    return ((AnnotStringElementValue) annot.getValue("version")).value();
	} else {
	    return ((UniverseClassAttribute)attr).getVersion();
	}
    }
    
    /**
     * Return true if this class was compiled with universe run-time
     * support enabled. It will first check if the universe_class attribute
     * is present, if not it will check for an annotation.
     * By AHS.
     */
    public boolean hasUniverseRuntimeSupport() {
        Attribute attr = attributes.get(Constants.ATT_UNIVERSE_VERSION);
	if (attr == null) { // attribute not found, look for annotation
	    attr = attributes.get(Constants.ATT_RUNTIME_VISIBLE_ANNOTATIONS);
	    if (attr == null)
	        return false;
	    Annotation annot = ((RuntimeVisibleAnnotationsAttribute)attr).get("org.multijava.universes.UniverseClass");
	    if (annot == null)
	        return false;
	    return ((AnnotBooleanElementValue) annot.getValue("hasRuntimeSupport")).value();
	} else {
	    return ((UniverseClassAttribute)attr).hasRuntimeSupport();
	}
    }

    /**
     * Returns true if the field is deprecated
     */
    public boolean isDeprecated() {
	return attributes.get(Constants.ATT_DEPRECATED) != null;
    }

    /**
     * Sets the deprecated attribute of this field
     */
    public void setDeprecated(boolean deprecated) {
	if (deprecated) {
	    attributes.add(new DeprecatedAttribute());
	} else {
	    attributes.remove(Constants.ATT_DEPRECATED);
	}
    }

    /**
     * Returns the interfaces of the class in the file
     */
    public String[] getInterfaces() {
	String[]	names = new String[interfaces.length];

	for (int i = 0; i < interfaces.length; i++) {
	    names[i] = interfaces[i].getName();
	}

	return names;
    }

    /**
     * Sets the interfaces of the class in the file
     */
    public void setInterfaces(String[] interfaces) {
	this.interfaces = new ClassConstant[interfaces.length];
	for (int i = 0; i < this.interfaces.length; i++) {
	    this.interfaces[i] = new ClassConstant(interfaces[i]);
	}
    }

    /**
     * Returns the fields info of the class in the file
     */
    public FieldInfo[] getFields() {
	return fields;
    }

    /**
     * Sets the fields info of the class in the file
     */
    public void setFields(FieldInfo[] fields) {
	this.fields = fields;
    }

    /**
     * Returns the methods info of the class in the file
     */
    public MethodInfo[] getMethods() {
	return methods;
    }

    /**
     * Sets the methods info of the class in the file
     */
    public void setMethods(MethodInfo[] methods) {
	this.methods = methods;
    }

    /**
     * Returns the method infos of the class but with synthetic
     * multimethod and redirector bodies and external methods removed
     * and internal multimethod signatures added. */
    public MethodInfo[] getMultimethods() {
	ArrayList result = new ArrayList();
	for (int i = 0; i < methods.length; i++) {
	    if (!(methods[i].isMultimethodBody() 
		  || methods[i].isDispatcher()
		  || methods[i].isRedirector()
		  )) {
		result.add(methods[i]);
	    }
	}

	GenericFunctionsAttribute attr = (GenericFunctionsAttribute) 
	    attributes.get(Constants.ATT_GENERIC_FUNCTIONS);
	// for a generic function anchor we don't want the
	// (necessarily external) multimethods included
	if (attr != null && !isAnchor()) {
	    attr.appendMultimethodInfo(result);
	}
	
	MethodInfo[] rA = new MethodInfo[result.size()];
	result.toArray(rA);
	return rA;
    }

    /**
     * Returns the method infos for the external methods of this.
     * Returns an empty array unless this is an anchor class. */
    public MultimethodInfo[] getExternalMethods() {
	ArrayList result = new ArrayList();
	GenericFunctionsAttribute attr = (GenericFunctionsAttribute) 
	    attributes.get(Constants.ATT_GENERIC_FUNCTIONS);
	if (attr != null && isAnchor()) {
	    attr.appendMultimethodInfo(result);
	}
	
	MultimethodInfo[] rA = new MultimethodInfo[result.size()];
	result.toArray(rA);
	return rA;
    }


    /**
     * Returns true if the class is an anchor class for an external
     * generic function.  */
    public boolean isAnchor() {
	return attributes.get(Constants.ATT_ANCHOR) != null;
    }

    // --------------------------------------------------------------------
    // OUTPUT (WRITE A CLASS FILE)
    // --------------------------------------------------------------------

    /**
     * Writes the content of the class to the specified output stream
     *
     * @param	out		the stream to write to
     *
     * @exception	IOException	an io problem has occured
     * @exception ClassFileFormatException attempt to write a bad classfile info */
    public void write(DataOutput out)
	throws IOException, ClassFileFormatException
    {
	ConstantPool constantPool = resolveConstants();

	// Headers
	out.writeInt(JAVA_MAGIC);
	out.writeShort(minorVersion);
	out.writeShort(majorVersion);

	constantPool.write(out);

	// Class hierarchy/access
	out.writeShort(getModifiers());
	out.writeShort(thisClass.getIndex());
	out.writeShort(superClass == null ? 0 : superClass.getIndex());

	// interfaces
	out.writeShort(interfaces.length);
	for (int i = 0; i < interfaces.length; i++) {
	    out.writeShort(interfaces[i].getIndex());
	}

	// fields
	out.writeShort(fields.length);
	for (int i = 0; i < fields.length; i++) {
	    fields[i].write(constantPool, out);
	}



	// methods
	out.writeShort(methods.length);
	for (int i = 0; i < methods.length; i++) {
	    methods[i].write(constantPool, out);
	}


	// attributes
	attributes.write(constantPool, out);

	constantPool.close();

    }

    /**
     * Writes the contents of the class to a file.
     *
     * @param	dest	the directory to write the .class file to, 
     *			if the directory doesn't exist it will be 
     *			created
     *
     * @exception IOException			an io problem occured
     * @exception ClassFileFormatException	attempt to write a bad 
     *						classfile info
     */
    public void write( Destination dest ) 
	throws IOException, ClassFileFormatException 
    {
	String[] packageAndTypeName = Utils.splitQualifiedName( getName() );

	File targetDir = dest.classFileTargetDirectory( packageAndTypeName[0], 
							sourceFile );

        targetDir.mkdirs();

	File outputFile = createOutputFile( targetDir, packageAndTypeName[1] );
	DataOutputStream outputStream =
	    new DataOutputStream( 
		  new BufferedOutputStream(
                         new FileOutputStream( outputFile )));

	write( outputStream );
	outputStream.flush();
	outputStream.close();
    }

    /** 
     * Creates an output file. This is a factory method that can be
     * overridden by subclasses. 
     */
    protected File createOutputFile(File dir, String className) {
        return new File( dir, className + ".class" );
    }

    // --------------------------------------------------------------------
    // PRIVATE METHODS
    // --------------------------------------------------------------------

    /**
     * Create a constant pool and insert in it all constants from from
     * all the element of the class (fields, method, attributes, ...)
     */
    private ConstantPool resolveConstants() throws ClassFileFormatException {

	ConstantPool constantPool = new ConstantPool();

	constantPool.addItem(thisClass);
	if (superClass != null) {
	    constantPool.addItem(superClass);
	}

	for (int i = 0; i < interfaces.length; i++) {
	    constantPool.addItem(interfaces[i]);
	}

	for (int i = 0; i < fields.length; i++) {
	    fields[i].resolveConstants(constantPool);
	}

	// methods (generate code, resolve constants, and pregen)
	for (int i = 0; i < methods.length; i++) {
	    methods[i].resolveConstants(constantPool);
	}

	attributes.resolveConstants(constantPool);

	return constantPool;
    }

    /*
     *
     */
    private static FieldInfo[] makeFieldInfoArray(ArrayList vector) {
	FieldInfo[] array;

	array = new FieldInfo[vector.size()];
	vector.toArray(array);
	return array;
    }

    /*
     *
     */
    private static MethodInfo[] makeMethodInfoArray(ArrayList vector) {
	MethodInfo[]	array;

	array = new MethodInfo[vector.size()];
	vector.toArray(array);
	return array;
    }

    /*
     *
     */
    private static ClassConstant[] makeInterfacesArray(ArrayList vector) {
	ClassConstant[]	array;

	array = new ClassConstant[vector.size()];
	for (int i = 0; i < array.length; i++) {
	    array[i] = new ClassConstant((String)vector.get(i));
	}
	return array;
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private int				majorVersion;
    private int				minorVersion;

    private ClassConstant			thisClass;
    private ClassConstant			superClass;
    private ClassConstant[]		interfaces;
    private FieldInfo[]			fields;
    private MethodInfo[]			methods;
    protected AttributeList			attributes;
    private final File sourceFile;
}
