/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: ClassPath.java,v 1.27 2008/03/16 15:58:57 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.*;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Stack;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;
import org.multijava.mjc.Debug;


/**
 * This class implements the conceptual directory structure for .class files
 */
public class ClassPath {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs the class path.  Private since this is a singleton.
     */
    private ClassPath(String classPath, String sourcePath) {
	// Source path does not need the system files
	sourceDirs = parseClassPath(sourcePath);
	String	version = System.getProperty("java.version");
	if (bootclasspath != null) {
	    classPath += File.pathSeparator + bootclasspath;
	} else if (System.getProperty("sun.boot.class.path") != null) {
	    classPath += File.pathSeparator + System.getProperty("sun.boot.class.path");
	} else if (version.startsWith("1.2") || version.startsWith("1.3")) {
	    String	home = System.getProperty("java.home");
	    classPath += File.pathSeparator + home + File.separatorChar +
		"lib" + File.separatorChar + "rt.jar";
	}
	finalclasspath = classPath;
	dirs = parseClassPath(classPath);
    }

    /**
     *
     */
    public static ClassDirectory[] parseClassPath(String classPath) {
	int start, length;

	ArrayList parsedDirs  = new ArrayList();
	length = classPath.length();
	start  = 0;
	while (start < length) {
	    int	end = classPath.indexOf(File.pathSeparatorChar, start);

	    if (end == -1) {
		end = length;
	    }

	    try {
		File	file = new File(classPath.substring(start, end));

		if (file.isDirectory()) {
		    parsedDirs.add(new DirClassDirectory(file));
		} else if (file.isFile()) {
		    // check if file is zipped (.zip or .jar)
                    String nam = file.getName();
		    if (nam.endsWith(".jar") || nam.endsWith(".JAR")
                        || nam.endsWith(".zip") || nam.endsWith(".ZIP")) {
			try {
			    parsedDirs.add(
				new ZipClassDirectory(new ZipFile(file)));
			} catch (ZipException e) {
			    // it was not a zip file, ignore it
				//System.out.println("ClassPath.parseClassPath - 1: " + e.toString());
				//e.printStackTrace();
			} catch (IOException e) {
			    // ignore it
				//System.out.println("ClassPath.parseClassPath - 2: " + e.toString());
				//e.printStackTrace();
			}
		    }
		}
	    } catch (SecurityException e) {
		// unreadable file, ignore it
		//System.out.println("ClassPath.parseClassPath - 3: " + e.toString());
		//e.printStackTrace();
	    }

	    start = end + 1;
	}

	ClassDirectory[]	cd = new ClassDirectory[parsedDirs.size()];
	parsedDirs.toArray(cd);
	
	return cd;
    }

    // ----------------------------------------------------------------------
    // SINGLETON
    // ----------------------------------------------------------------------

    /**
     * Initialization from strings that represents the class path and the source path
     * (if sourcePath is null, classPath is used for both files searches).
     * @param classPath		the classpath (used for finding .class files)
     * @param sourcePath	the sourcepath (used for finding .java and other source files)
     */
    //@ requires classPath != null;
    public static void init(String classPath, String sourcePath) {
	if (self != null && lastClassPath.equals(classPath) &&
	    ((lastSourcePath == null && sourcePath == null) ||
	     (lastSourcePath != null && lastSourcePath.equals(sourcePath)))) {
	    return;		// current value of self is fine
	}
	lastClassPath = classPath;
	lastSourcePath = sourcePath;
	self = new ClassPath(classPath, sourcePath == null ? classPath : sourcePath);
    }

    /** Sets the boot classpath (in palce of the default, which is the
	system classpath.  This is appended to the classpath.
    */
    static public void setBootClassPath(String b) {
	bootclasspath = b;	
    }

    /** Returns the combined classpath and bootclasspath.
    */
    static public String classPath() {
	return finalclasspath;
    }

    // ----------------------------------------------------------------------
    // LOAD CLASS
    // ----------------------------------------------------------------------

    /** Returns a ClassDescription for the java file containing the class with the given name; 
	returns null if can't be found. */
    static private String[] javasuffix = { ".java" };
    public static ClassPath.ClassDescription getJavaFile(String name) {
	return getSourceFile(name,javasuffix);
    }    

    /** Returns a ClassDescription for the source file containing the class with the given name
	and one of the given suffixes (searching for any suffix in each directory
	on the sourcepath in turn); 
	returns null if can't be found. 
	This is a helper function intended for use in subclasses of FileFinder. */
    public static ClassPath.ClassDescription getSourceFile(String name, String[] suffixes) {
	ClassPath.ClassDescription result = (ClassPath.ClassDescription)allLocatedSourceFiles.get(name);
	if (result == NO_CLASS_DESC) return null;
	if (result != null) return result;
	// no result cached, so search
	result = self.findSourceFile(name,suffixes); // returns null if not found
	allLocatedSourceFiles.put(name,result==null?NO_CLASS_DESC:result);
	return result;
    }    
    /** Returns a ClassDescription for the file with the given name+suffix; 
	returns null if can't be found. 
	This is a method intended for use when searching the class path 
	for a file with a specific suffix. */
    public static ClassPath.ClassDescription getFile(String name, String suffix) {
	String[] suffixes = { suffix };
	ClassPath.ClassDescription result 
	    = self.findSourceFile(name, suffixes); // returns null if not found
	return result;
    }    

    /** Returns a ClassDescription for the class file containing the class with the given name
	(searching each directory on the classpath in turn); 
	returns null if can't be found. 
	This is a helper function intended for use in subclasses of FileFinder. */
    public static ClassPath.ClassDescription getClassPathFile(String name) {
	return getClassPathFile(name, ".class");
    }

    public static ClassPath.ClassDescription getClassPathFile(String name,
                                                              String suffix) {
	return self.getClassFile(name, suffix);
    }

	
    /**
     * Creates a class info object by reading bytecode from the file
     * named <code>name</code>.class.
     *
     * @param name the name of the class
     * @param interfaceOnly read the signature, but not the code, of
     * the class if true
     * @return the content of a class file that contains the class
     * named name (null if no such class)
     * @exception ClassFileReadException if the class file is
     * malformed or an I/O error occurs
     */
    public static ClassInfo getClassInfo(String name, boolean interfaceOnly)
	throws ClassFileReadException
    {
        return getClassInfo(name, ".class", null, true, interfaceOnly);
    }

    /**
     * Creates a class info object by reading bytecode from the file
     * named <code>name.suffix</code>. If the argument
     * <code>creator</code> is not null, use it to create a class info
     * object; otherwise creates an instance of the class ClassInfo.
     *
     * @param name the name of the class
     * @param suffix the file suffix to look for
     * @param creator a factory, or null
     * @param markNotFound if true, then track files already checked
     * @param interfaceOnly read the signature, but not the code, of
     * the class if true
     * @return the content of a class file that contains the class
     * named name (null if no such class)
     * @exception ClassFileReadException if the class file is
     * malformed or an I/O error occurs
     */
    public static /*@ nullable */ClassInfo getClassInfo(/*@ non_null */ String name, 
                                         String suffix,
                                         /*@ nullable */ ClassInfoCreator creator,
                                         boolean markNotFound,
                                         boolean interfaceOnly) 
	throws ClassFileReadException
    {
	ClassPath.ClassDescription clazz 
            = (ClassPath.ClassDescription) allLoadedFile.get(name);
        // result == null --> the name was never looked up before
        // result == NO_CLASS_DESC --> the name was looked for,
        // and there was no such class 
        // otherwise --> the name was looked for and was found

	if (clazz == NO_CLASS_DESC) {
            // don't bother looking again
            return null;
        }

        // never looked before, so will now?
	if (clazz == null) {
	    ClassPath.ClassDescription desc = self.getClassFile(name, suffix);
	    // desc is null if the name cannot be found
	    if (desc == null) {
                if (markNotFound) {
                    allLoadedFile.put(name, NO_CLASS_DESC);
                }
		return null;
	    }

	    clazz = desc;
	    allLoadedFile.put(name, clazz);
	}

	allLoadedFile.remove(name);

	ClassPath.Data data = clazz.getData();
	try {
	    ClassInfo ci = creator == null ?
                new ClassInfo(data.getDataInput(), interfaceOnly) :
                creator.createClassInfo(data.getDataInput(), true);
	    data.release();
	    return ci;
	} catch (ClassFileFormatException e) {
	    data.release();
	    throw new ClassFileReadException(e, clazz.getName());
	} catch (IOException e) {
	    data.release();
	    throw new ClassFileReadException(e, clazz.getName());
	}
    }

    /** Returns the root directory (on the CLASSPATH) in which the
	given package can be found.  The package name is '.'
	separated.  Returns null if the package is not found as a
	directory on the on the CLASSPATH + BOOTCLASSPATH.  (The
	package could still exist in a jar.)
    */
    static public String findPackage(String name) {
	return self.findPackageLocation(name);
    }

    /** Returns the root directory in the sourcepath in which the
	given '.' separated package is found, or null if the package
	cannot be found.  (Though, the package could still exist in a
	jar.)
    */
    static public String findPackageSource(String name) {
	return self.findPackageSourceLocation(name);
    }

    /** Returns a String indicating the location of a package.  The
	argment is a '.'-separated fully-qualified package name.  For
	a normal directory-type package, the result is the path to the
	package root, or null if the package cannot be found.
    */
    private String findPackageLocation(String name) {
	String n = name.replace('.',File.separatorChar);
	for (int i = 0; i < dirs.length; i++) {
		if (dirs[i].containsDirectory(n)) return dirs[i].location();
	}
	return null;
    }
    
    /** Returns the root directory in the sourcepath in which the given
	'.' separated package is found.
    */
    private String findPackageSourceLocation(String name) {
	String n = name.replace('.',File.separatorChar);
	for (int i = 0; i < sourceDirs.length; i++) {
	    if (sourceDirs[i].containsDirectory(n)) {
		return sourceDirs[i].location();
	    }
	}
	return null;
    }
    
    /**
     * Returns a class file but don't read it or cache it.  Returns null if can't be found.
     */
    //@ requires name != null;
    //@ pure
    private ClassPath.ClassDescription getClassFile(String name, 
                                                    String suffix) {
	for (int i = 0; i < dirs.length; i++) {
	    ClassPath.ClassDescription desc = dirs[i].getClassFile(name, 
                                                                   suffix);

	    if (desc != null) return desc;

	}

	return null;
    }

    /** This function returns the first file found that has one of the given suffixes and is in the sourceDirs
	directory path.  Each directory is searched in turn; within each directory, each suffix is sought, in 
	order; if none are present searching proceeds with the next directory.
    	Returns null if the name cannot be found with one of the suffixes.
    */
    /*@ requires name != null;
        requires suffixes != null && \nonnullelements(suffixes);
        ensures (* \result == null || \result corresponds to a file that exists and is in the earliest directory, 
		  has the given name, and has the earliest suffix *);
	ensures (* \result == null implies no combination of directory and suffix is the name of an existing file *);
    */
    //@ pure
    private ClassPath.ClassDescription findSourceFile(String name, String[] suffixes) {
	for (int i = 0; i < sourceDirs.length; ++i) {
	    for (int j = 0; j < suffixes.length; ++j) {
		ClassPath.ClassDescription found = sourceDirs[i].findFile(name, suffixes[j]);

		if (found != null) return found;
	    }
	}

	return null;
    }

    /**
     * Prepares for a new compilation session by forgetting the
     * locations of known files (since the CLASSPATH and SOURCEPATH
     * may be different for the new session).
     */
    public static void initSession() {
	allLoadedFile = new HashMap();
	allLocatedSourceFiles = new HashMap();
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /**
     * A reference to the singleton instance of this class.
     */
    private static ClassPath	self = null;

    /**
     * The classPath argument that was used to construct
     * <CODE>self</CODE>.  Stored to avoid instantiating a new
     * singleton when it would have the same state as the old one.
     */
    private static String lastClassPath;

    /**
     * The sourcePath argument that was used to construct
     * <CODE>self</CODE>.  Stored to avoid instantiating a new
     * singleton when it would have the same state as the old one.
     */
    private static String lastSourcePath;

    private static HashMap	allLoadedFile = new HashMap();
    private static HashMap	allLocatedSourceFiles = new HashMap(); // any non-compiled file, not just .java
    private ClassDirectory[]	dirs;	// list of directories in class path
    private ClassDirectory[]	sourceDirs = null; // list of Directories in source path
    
    /** A ClassDescription value used to indicate that the is no file and no class description available. */
    private static ClassDescription NO_CLASS_DESC = new ClassDescription() {
	public long getTime() { return 0; }
	public String getName() { return ""; }
	public boolean isClass() { return true; }
	public ClassPath.Data getData() {
	    /**
	    * @return an DataInput to read this class
	    */
	    throw new RuntimeException("DUMMY");
	}
    };
    
    /** A File value used to indicate an undefined result. */
    //public final static File FILE_UNDEFINED = new File("");
    
    
    // ----------------------------------------------------------------------
    // Inner classes - declared here so that they can be used outside this file.
    // ----------------------------------------------------------------------

    /** This class wraps up needed information about a file-like entity that may
	or may not be an actual file in a directory on a disk.
    */
    public static abstract class  ClassDescription {
	/**
	* @return a DataInput to read this file-like object, or null if no such object exists or can be read
	*/
	public abstract ClassPath.Data getData();
	
	/** @return the last modification time of this file-like object */
	public abstract long getTime();
	
	/** @return the name of this file-like object */
	//@ ensures \result != null;
	public abstract String getName();
	
	public abstract boolean isClass();
    }


    static public class FileClassDescription extends ClassPath.ClassDescription {
	private long time;
	private File file;
	private boolean isClass;
	
	public File file() { return file; }
	
	FileClassDescription(File f, boolean isClass) {
	    time = f.lastModified();
	    file = f;
	    this.isClass = isClass;
	}
	
	public long getTime() { return time; }
	public String getName() { return file.getPath(); }
	public boolean isClass() { return isClass; }
	
			/**
			    * @return an DataInput to read this class
			    */
	public ClassPath.Data getData() {
	    try {
		return new ClassPath.Data(new FileInputStream(file));
	    } catch (FileNotFoundException e) {
		// REALLY BAD (file exist in classpath but is not accessible)
		// Perhaps ought to be a PositionedError FIXME?
		throw new org.multijava.util.compiler.CompilationAbortedError(e.toString());
	    }
	}
    };

    /** This class mimics a DataInput, but optimizes the reading of the supplied InputStream.
	The InputStream is read in one fell swoop into a data array - presumes that such reading
	is efficient and that the data array does not need to be too big.
    */
    public static class Data {

	//@ public ghost boolean ba_available = false;
	
	//@ requires is != null;
	//@ ensures !ba_available;
	public Data(InputStream is) {
	    this.is = is;
	}

	//@ requires ! ba_available;
	//@ requires is != null;
	//@ ensures \result != null;
	//@ ensures ba != null;
	//@ ensures is == null;
	//@ ensures ba_available;
	//@ ensures (* ba contains everything readable from is *);
	// Need to express that the Datainput produced by this function is only
	// valid until release is called on this ClassPath.Data object.
	public DataInput getDataInput() throws IOException {
	    ba = getByteArray();
	    int		n = 0;
	    int		available;
	    while (true) {
		int count = is.read(ba, n, ba.length - n);
		if (count < 0) {
		    break;
		}
		available = is.available();
		n += count;
		if (n + available > ba.length) {
		    byte[] temp = new byte[ba.length * 2];
		    System.arraycopy(ba, 0, temp, 0, ba.length);
		    ba = temp;
		} else if (available == 0) {
		    break;
		}
	    }
    
	    is.close();
	    is = null;

	    //@ set ba_available = true;
	    return new DataInputStream(new ByteArrayInputStream(ba, 0, n));
	}

	//@ requires ba_available;
	//@ modifies ba_available;
	//@ ensures ! ba_available;
	public void release() {
	    release(ba);
	    //@ set ba_available = false;
	}

	// These two functions manage a static stack of free 'ba' buffers.
	/** Returns a free byte array. */
	private static byte[] getByteArray() {
	    if (!Constants.ENV_USE_CACHE || stack.empty()) {
		return new byte[10000];
	    }
	    return (byte[])stack.pop();
	}

	/** Releases a byte array, adding it to the free list. */
	private static void release(byte[] arr) {
	    if (Constants.ENV_USE_CACHE) {
		stack.push(arr);
	    }
	}
    
	// ----------------------------------------------------------------------
	// PRIVATE DATA MEMBERS
	// ----------------------------------------------------------------------

	//@ spec_public
	private /*@nullable@*/ InputStream	is; // FIXME: should this really be nullable? [DSRG]
	
	//@ spec_public
	private byte[]	ba;
	
	private static /*@non_null@*/ Stack	stack = new Stack();
    }

    /** Holder for the user set boot class path; if null, the system value
	of the boot class path is used.
    */
    static private String	bootclasspath = null;

    /** Holder for the combined classpath and bootclasspath. */
    static private String	finalclasspath = null;
}

class ZipClassDescription extends ClassPath.ClassDescription {
    private long time;
    private ZipEntry entry;
    private ZipFile zip;
    
    ZipClassDescription(ZipFile z, ZipEntry e, long t) {
	time = t;
	entry = e;
	zip = z;
    }
    
    public long getTime() { return time; }
    public String getName() { return entry.getName(); }
    public boolean isClass() { return true; }
    
    public ClassPath.Data getData() {
	try {
	    return new ClassPath.Data(zip.getInputStream(entry));
	} catch (IOException e) {
	    return null; // REALLY BAD (file exist in cp but is not accessible)
	}
    }
}

/**
 * This class represents a conceptual directory which may hold
 * files. Since Java can use archived class files found in
 * a compressed ("zip") file, this entity may or may not correspond to
 * an actual directory on disk.
 */
abstract class ClassDirectory {

    /**
     * @return a ClassDescription of a class file that contain the class named name
     * @param name the name of the class
     */
    public abstract ClassPath.ClassDescription getClassFile(String name);

    public abstract ClassPath.ClassDescription getClassFile(String name,
                                                            String suffix);


    /**
     * Returns a file descriptor for the .java file that should store
     * the type of the given name.  
     *
     * <pre><jml>
     * {|
     *    requires (* a .java file that should store the type named "name"
     *                exists on the sourcepath *);
     *    ensures (* \result is the file descriptor for that file *);
     * also
     *    requires !(* a .java file that should store the type named "name"
     *                 exists on the sourcepath *);
     *    ensures \result == null;
     * |}
     * </jml></pre>
     *
     */
    //public abstract File findJavaFile(String name);
	
    // returns null if the name+suffix cannot be found
    public abstract ClassPath.ClassDescription findFile(String name, String suffix);

    /** returns true if the directory contains the given directory */
    public abstract boolean containsDirectory(String name);

    /** returns the root location of this directory element */
    public abstract String location();
}

class DirClassDirectory extends ClassDirectory {
    /**
     * Constructs a class directory representing a real directory
     */
    public DirClassDirectory(File dir) {
	this.dir = dir;
    }

    /**
     * @return a class file that contain the class named name ('/'-separated)
     * @param name the name of the class file
     */
     // returns null if the file cannot be found
    public ClassPath.ClassDescription getClassFile(String name,
                                                   String suffix) {
	name = name.replace('/', File.separatorChar) + suffix;

	final File file = new File(dir.getPath(), name);

	// canRead is most time-consuming call at this point.  If a file 
	// actually exists but is not readable, we'll let an io exception
	// be thrown later on
	if (!file.exists()) {
	    return null;
	} else {
	    return new ClassPath.FileClassDescription(file,true);
	}
    }

    public ClassPath.ClassDescription getClassFile(String name) {
        return getClassFile(name, ".class");
    }

    // returns null if the name+suffix cannot be found
    
    public ClassPath.ClassDescription findFile(String name, String suffix) {
	name = name.replace('/', File.separatorChar) + suffix;
	File file = new File(dir.getPath(), name);

	// canRead is most time-consuming call at this point.  If a file 
	// actually exists but is not readable, we'll let an io exception
	// be thrown later on
	if (!file.exists()) {
	    return null;
	} else {
	    return new ClassPath.FileClassDescription(file,
                                                      ".class".equals(suffix));
	}
    }
    
    /** returns true if the directory object contains the given name as a subdirectory.
	name is a File.separatorChar - separated multi-component path.
    */
    public boolean containsDirectory(String name) {
	File f = new File(dir.getAbsolutePath() + File.separator + name);
	return f.isDirectory();
    }

    /** Returns the location of this directory, in this case as its absolute path. */
    public String location() { return dir.getAbsolutePath(); }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private File dir;		// non null iff is a real directory
}

class ZipClassDirectory extends ClassDirectory {
    /**
     * Constructs a class directory representing a zipped file
     */
    public ZipClassDirectory(ZipFile zip) {
	this.zip = zip;
    }
    
  /**
   * @return a ClassDescription for a class file that contain the class named name
   * @param name the name of the class
   */
   // returns null if the name cannot be found
    public ClassPath.ClassDescription getClassFile(String name, 
                                                   String suffix) {
	name += suffix;

	final ZipEntry	entry = zip.getEntry(name);

	if (entry == null) {
	    return null;
	} else {
	    return new ZipClassDescription(zip,entry,entry.getTime());
	}
    }

    public ClassPath.ClassDescription getClassFile(String name) {
        return getClassFile(name, ".class");
    }
    
    /** Returns the location of the directory, in this case as the location of the zip file. */
    public String location() { return zip.getName(); }

    /**
     * Returns null since we do not support reading non-.class files from
     * zip or jar archives.  */
    public ClassPath.ClassDescription findFile(String name, String suffix) {
	final ZipEntry	entry = zip.getEntry(name + suffix);
	if (entry == null) {
	    return null;
	} else {
	    return new ZipClassDescription(zip,entry,entry.getTime());
	}
    }

    /** Returns true if the zip file contains the given directory. */
    public boolean containsDirectory(String name) {
	// Unfortunately, some of the JDK jar files don't include
	// actual directory entries for some items.  The directories
	// are implicitly created when unjarring.  But we need to know
	// that they would be, and so we have to do a linear search
	// over the Enumeration.
	if (dirsInZip == null) {
	    populateDirsInZip();
	}
	
	// separators in jar files are standardized
	String jarDirName = name.replace(File.separatorChar,'/');
	return dirsInZip.contains(jarDirName);
    }

    private void populateDirsInZip() {
	if (dirsInZip == null) {
	    dirsInZip = new HashSet(20);
	} else {
	    dirsInZip.clear();
	}
	
	for (Enumeration e = zip.entries(); e.hasMoreElements(); ) {
	    ZipEntry entry = (ZipEntry) e.nextElement();
	    String n = entry.getName();
	    // separators in jar files are standardized
	    int pos;
	    while ((pos = n.lastIndexOf('/')) >=0) {
		String dir = n.substring(0,pos);
		boolean alreadyThere = dirsInZip.add(dir);
		if (alreadyThere) {
		    break;	// no need to add prefixes since they
				// must be there
		}
		n = dir;
	    }
	}
	
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    private ZipFile	zip;		// non null iff is a zip or jar file
    private HashSet dirsInZip;	// lazily initialized by containsDirectory()

}

