/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: GenericFunctionInfo.java,v 1.6 2004/04/12 21:00:43 tongjie Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

import org.multijava.util.InternalError;
import java.util.ArrayList;

/**
 * This data structure groups a set of multimethods all declared in
 * the same class or generic function anchor and all belonging to the
 * same generic function.  */
public class GenericFunctionInfo {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Constructs a generic function info structure from source.
     *
     * @param	name		the simple name of the method
     * @param	type		the top method's signature
     * @param   number		the function number of the generic function
     * @param   topConcMethColl	the fully qualified name of the top method of the
     *				generic function to which these methods belong,
     *				or null if these methods are external or the
     *				generic function is internal
     * @param	multimethods	an array of the multimethods of this generic
     *				function declared in the local class
     */
    public GenericFunctionInfo(String name, String type, 
			       String generictype,
			       int number,
			       String topConcMethColl,
			       MultimethodInfo[] multimethods) 
    {
	this.name = new AsciiConstant(name);
	this.type = new AsciiConstant(type);
	this.generictype = new AsciiConstant(generictype);
	this.number = (short) number;
	if (topConcMethColl == null) {
	    this.topConcMethColl = new AsciiConstant("");
	} else {
	    this.topConcMethColl = new AsciiConstant(topConcMethColl);
	}
	this.multimethods = multimethods;
    }

    /**
     * Constructs a generic function info structure from a class file stream
     *
     * @param	in		the stream to read from
     * @param	cp		the constant pool
     *
     * @exception	java.io.IOException	an io problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					read a bad classfile info
     */
    public GenericFunctionInfo(DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException 
    {
	this.name = (AsciiConstant)cp.getEntryAt(in.readUnsignedShort());
	this.type = (AsciiConstant)cp.getEntryAt(in.readUnsignedShort());
	this.generictype = (AsciiConstant)cp.getEntryAt(in.readUnsignedShort());
	this.number = in.readShort();
	this.topConcMethColl = 
	    (AsciiConstant)cp.getEntryAt(in.readUnsignedShort());

	String tmcString = topConcMethColl.getValue();
	if (tmcString.length() == 0) {
	    tmcString = null;
	}
	
	short mmCount = in.readShort();
	this.multimethods = new MultimethodInfo[mmCount];
	for (int i = 0; i < mmCount; i++) {
	    multimethods[i] = new MultimethodInfo(in, cp, number, tmcString);
	}
	
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Returns the name of this generic function
     */
    public String getName() {
	return name.getValue();
    }

    /**
     * Sets the name of this generic function
     */
    public void setName(String name) {
	this.name = new AsciiConstant(name);
    }

    /**
     * Returns the type signature of this generic function
     */
    public String getSignature() {
	return type.getValue();
    }

    public String getGenericSignature() {
	return generictype.getValue();
    }
    /**
     * Sets the type of this generic function
     */
    public void setSignature(String type) {
	this.type = new AsciiConstant(type);
    }

    public void setGenericSignature(String generictype){
	this.generictype = new AsciiConstant(generictype);
    }

    protected int getSize() {
	int mmSize = 0;
	for (int i = 0; i < multimethods.length; i++) {
	    mmSize += multimethods[i].getSize();
	}
	
	return 2 + 2 + 2 + 2 + 2 + 2 + mmSize;
    }

    /**
     * Appends the MultimethodInfo objects for the multimethods
     * declared locally for this generic function to
     * <Code>result</code>. */
    protected void appendMultimethodInfo(ArrayList result) {
	for (int i = 0; i < multimethods.length; i++) {
	    result.add(multimethods[i]);
	}
    }

    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant value on constant pool
     *
     * @param	cp		the constant pool for this class
     */
    public void resolveConstants(ConstantPool cp) 
	throws ClassFileFormatException 
    {
	cp.addItem(name);
	cp.addItem(type);
	cp.addItem(generictype);
	cp.addItem(topConcMethColl);

	for (int i = 0; i < multimethods.length; i++) {
	    multimethods[i].resolveConstants(cp);
	}
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool
     *
     * @param	cp		the constant pool that contain all data
     * @param	out		the file where to write this object info
     *
     * @exception	java.io.IOException	an io problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					write a bad classfile info
     */
    public void write(ConstantPool cp, DataOutput out)
	throws IOException, ClassFileFormatException
    {
	out.writeShort(name.getIndex());
	out.writeShort(type.getIndex());
	out.writeShort(generictype.getIndex());
	out.writeShort(number);
	out.writeShort(topConcMethColl.getIndex());
	out.writeShort((short) multimethods.length);
	for (int i = 0; i < multimethods.length; i++) {
	    multimethods[i].write(cp,out);
	}
    }

    // --------------------------------------------------------------------
    // PRIVATE METHODS
    // --------------------------------------------------------------------

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private AsciiConstant name;
    private AsciiConstant type;
    private AsciiConstant generictype;
    private short number;
    private AsciiConstant topConcMethColl;
    private MultimethodInfo[] multimethods;
    
}
