/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: GenericFunctionsAttribute.java,v 1.4 2004/01/22 20:04:51 leavens Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.util.ArrayList;

/**
 * This attribute represents the entire set of generic functions
 * declared in or overridden in a single class or generic function
 * anchor.  */
public class GenericFunctionsAttribute extends Attribute {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Create a generic functions attribute from source.
     */
    public GenericFunctionsAttribute(GenericFunctionInfo[] genericFunctions) {
	this.genericFunctions = genericFunctions;
    }

    /**
     * Constructs a generic functions attribute from a class file stream.
     *
     * @param	in		the stream to read from
     * @param	cp		the constant pool
     *
     * @exception	java.io.IOException	an I/O problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					read a bad classfile info
     */
    public GenericFunctionsAttribute(DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException
    {
	if (in.readInt() < 2) {
	    throw new ClassFileFormatException("bad attribute length");
	}
	
	short gf_count = in.readShort();
	this.genericFunctions = new GenericFunctionInfo[gf_count];
	for (int i = 0; i < gf_count; i++) {
	    genericFunctions[i] = new GenericFunctionInfo(in, cp);
	}
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Returns the attribute's tag
     */
    protected int getTag() {
	return Constants.ATT_GENERIC_FUNCTIONS;
    }

    /**
     * Returns the space in bytes used by this attribute in the classfile
     */
    protected int getSize() {
	int genFuncSize = 0;
	for (int i = 0; i < genericFunctions.length; i++) {
	    genFuncSize += genericFunctions[i].getSize();
	}
	
	return 2 + 4 + 2 + genFuncSize;
    }


    /**
     * Appends the MultimethodInfo objects for the multimethods
     * declared in this class to <Code>result</code>. */
    protected void appendMultimethodInfo(ArrayList result) {
	for (int i = 0; i < genericFunctions.length; i++) {
	    genericFunctions[i].appendMultimethodInfo(result);
	}
	
    }

    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant value on constant pool
     *
     * @param	cp		the constant pool for this class
     */
    protected void resolveConstants(ConstantPool cp) 
	throws ClassFileFormatException
    {
	cp.addItem(attr);
	for (int i = 0; i < genericFunctions.length; i++) {
	    genericFunctions[i].resolveConstants(cp);
	}
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool
     *
     * @param	cp		the constant pool that contain all data
     * @param	out		the file where to write this object info
     *
     * @exception	java.io.IOException	an io problem has occured
     */
    protected void write(ConstantPool cp, DataOutput out) 
	throws IOException, ClassFileFormatException
    {
	out.writeShort(attr.getIndex());
	out.writeInt(getSize() - 6);
	out.writeShort((short) genericFunctions.length);
	for (int i = 0; i < genericFunctions.length; i++) {
	    genericFunctions[i].write(cp, out);
	}
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    static final String tagString = "org.multijava.generic_functions";
    private static AsciiConstant attr = new AsciiConstant(tagString);
    private GenericFunctionInfo[] genericFunctions;
}

