/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: InnerClassInfo.java,v 1.5 2007/02/08 14:04:59 leavens Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

/**
 * This class represents a single entry in the array of classes stored
 * inside an InnerClasses attribute.  See JVM2, 4.7.5.
 *
 * @see InnerClassTable
 */
public class InnerClassInfo implements Comparable {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Create a classes entry within an InnerClasses attribute.
     *
     * @param	innerClass	the encoded name of the (inner) class
     * @param	outerClass	the defining scope  of the (inner) class
     * @param	simpleName	the simple name of the (inner) class
     * @param	modifiers	access permission to and properties of the class
     */
    public InnerClassInfo(String innerClass,
			  String outerClass,
			  String simpleName,
			  short modifiers) {
	this.innerClass = new ClassConstant(innerClass);
	this.outerClass = outerClass == null ? null : new ClassConstant(outerClass);
	this.simpleName = simpleName == null ? null : new AsciiConstant(simpleName);
	this.modifiers = modifiers;
    }

    /**
     * Create a classes entry within an InnerClasses attribute.c
     *
     * @param	in		the stream to read from
     * @param	cp		the constant pool
     *
     * @exception	java.io.IOException	an io problem has occured
     */
    public InnerClassInfo(DataInput in, ConstantPool cp) throws IOException {
	this.innerClass = 
	    (ClassConstant)cp.getEntryAt(in.readUnsignedShort());

	int		outerClass = in.readUnsignedShort();
	this.outerClass = 
	    outerClass == 0 ? null : (ClassConstant)cp.getEntryAt(outerClass);

	int		simpleName = in.readUnsignedShort();
	this.simpleName = 
	    simpleName == 0 ? null : (AsciiConstant)cp.getEntryAt(simpleName);

	this.modifiers = (short)in.readUnsignedShort();
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Return the qualified name of this class
     */
    public String getQualifiedName() {
	return innerClass.getName();
    }

    /**
     * Return the name of the outer class of this.
     */
    public String outerClassName() {
	return outerClass != null ? outerClass.getName() : "";
    }

    /**
     * Return the modifiers bit-mask of this class
     */
    public short getModifiers() {
	return modifiers;
    }


    /**
     * Returns true if this represents an InnerClasses attribute
     * entry for an anonymous class.  That is, the simpleName field is
     * null.  */
    public boolean isAnonymous() {
	return simpleName == null;
    }

    /**
     * Returns true if this represents an InnerClasses attribute entry
     * for a member type of the class file in which the attribute
     * appears.  That is, the outerClass field is not null.  */
    public boolean isMember() {
	return outerClass != null;
    }

    public String toString() {
	StringBuffer result = new StringBuffer();
	result.append( "ICII=" + innerClass.toString() );
	result.append( ", OCII=" + 
		       (outerClass == null ? "null" : outerClass.toString()) );
	result.append( ", INI=" + 
		       (simpleName == null ? "null" : simpleName.toString()) );
	result.append( ", ICAF=" + modifiers );
	return result.toString();
    }

    public int compareTo(/*@ non_null @*/ Object o) throws ClassCastException {
	InnerClassInfo other = (InnerClassInfo) o;
	return innerClass.compareTo(other.innerClass);
    }

    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant value on constant pool
     *
     * @param	cp		the constant pool for this class
     */
    /*package*/ void resolveConstants(ConstantPool cp) {
	cp.addItem(innerClass);
	if (outerClass != null) {
	    cp.addItem(outerClass);
	}
	if (simpleName != null) {
	    cp.addItem(simpleName);
	}
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool
     *
     * @param	cp		the constant pool that contain all data
     * @param	out		the file where to write this object info
     *
     * @exception	java.io.IOException	an io problem has occured
     */
    /*package*/ void write(ConstantPool cp, DataOutput out) throws IOException {
	out.writeShort(innerClass.getIndex());
	out.writeShort(outerClass == null ? (short)0 : outerClass.getIndex());
	out.writeShort(simpleName == null ? (short)0 : simpleName.getIndex());
	out.writeShort(modifiers);
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private ClassConstant			innerClass;
    private ClassConstant			outerClass;
    private AsciiConstant			simpleName;
    private short				modifiers;
}

