/*
 * Copyright (C) 2002 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: MJAttributeParser.java,v 1.4 2004/04/12 21:00:43 tongjie Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;


/**
 * This class implements a parser for the additional bytecode
 * attributes defined in the MultiJava language specification.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.4 $ */
public class MJAttributeParser implements AttributeParser {

    //---------------------------------------------------------------------
    // CONSTRUCTORS 
    //---------------------------------------------------------------------
    // default constructor
    

    // -------------------------------------------------------------
    // IMPLEMENT AttributeParser
    // -------------------------------------------------------------
    
    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed */
    public Attribute read(String tag, DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException
    {
	if (!isOurs(tag)) {
	    return null;
	}
	
	switch (tag.charAt(firstInterestingPos)) {

	case 'a':		
	    if (tag.equals(AnchorAttribute.tagString)) {
		return new AnchorAttribute(in, cp);
	    }
	    break;

	case 'd':		
	    if (tag.equals(DispatcherAttribute.tagString)) {
		return new DispatcherAttribute(in, cp);
	    }
	    break;

	case 'r':		
	    if (tag.equals(RedirectorAttribute.tagString)) {
		return new RedirectorAttribute(in, cp);
	    }
	    break;

	case 'g':		
	    if (tag.equals(GenericFunctionsAttribute.tagString)) {
		return new GenericFunctionsAttribute(in, cp);
	    } 
	    break;

	case 'm':		
	    if (tag.equals(MultimethodBodyAttribute.tagString)) {
		return new MultimethodBodyAttribute(in, cp);
	    }
	    break;

	default:
	    break;
	}

	return null;
    }

    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized, but skips CodeInfo
     * attributes.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed */
    public Attribute readInterfaceOnly(String tag, DataInput in, ConstantPool cp)
	throws IOException, ClassFileFormatException
    {
	if (!isOurs(tag)) {
	    return null;
	}
	
	switch (tag.charAt(firstInterestingPos)) {

	case 'd':		
	    if (tag.equals(DispatcherAttribute.tagString)) {
		return new DispatcherAttribute(in, cp);
	    }
	    break;

	case 'r':		
	    if (tag.equals(RedirectorAttribute.tagString)) {
		return new RedirectorAttribute(in, cp);
	    }
	    break;

	case 'g':		
	    if (tag.equals(GenericFunctionsAttribute.tagString)) {
		return new GenericFunctionsAttribute(in, cp);
	    }
	    break;

	case 'm':		
	    if (tag.equals(MultimethodBodyAttribute.tagString)) {
		return new MultimethodBodyAttribute(in, cp);
	    }
	    break;

	default:
	    break;
	}

	return null;
    }

    /**
     * Checks the next attribute in the given input, <CODE>in</CODE>,
     * and parses it if it is recognized, but only parses attributes
     * that may appear within a CodeInfo attribute.
     * @param tag a String giving the name of the next attribute in
     * <CODE>in</CODE>
     * @param in an input stream for the bytecode ClassInfo being
     * parsed, whose read cursor is positioned just after the
     * attribute_name_index of the next attribute (see JVM2, 4.7)
     * @param cp the ConstantPool for the bytecode ClassInfo being parsed
     * @param	insns		(sparse) array of instructions
     * @return an <code>Attribute</code> value if the next bytes in
     * <CODE>in</CODE> represent an Attribute recognized by this, or
     * <CODE>null</CODE> if the next bytes are not recognized
     * @exception IOException if an error occurs while reading from
     * the bytecode file represented by <CODE>in</CODE>
     * @exception ClassFileFormatException if the next bytes begin
     * like a recognized Attribute but are then malformed or if the
     * next bytes are not valid within a CodeInfo attribute */
    public Attribute readCodeInfoAttribute(String tag, DataInput in, 
					   ConstantPool cp, Instruction[] insns)
	throws IOException, ClassFileFormatException
    {
	if (!isOurs(tag)) {
	    return null;
	}
	
	switch (tag.charAt(firstInterestingPos)) {

	case 'd':		
	    if (tag.equals(DispatcherAttribute.tagString)) {
		throw new ClassFileFormatException("Attribute \"" +
						   DispatcherAttribute.tagString +
						   "\" illegal as sub-attribute of Attribute Code");
	    }
	    break;

	case 'r':		
	    if (tag.equals(RedirectorAttribute.tagString)) {
		throw new ClassFileFormatException("Attribute \"" +
						   RedirectorAttribute.tagString +
						   "\" illegal as sub-attribute of Attribute Code");
	    }
	    break;

	case 'g':		
	    if (tag.equals(GenericFunctionsAttribute.tagString)) {
		throw new ClassFileFormatException("Attribute \"" +
						   GenericFunctionsAttribute.tagString +
						   "\" illegal as sub-attribute of Attribute Code");
	    }
	    break;

	case 'm':		
	    if (tag.equals(MultimethodBodyAttribute.tagString)) {
		throw new ClassFileFormatException("Attribute \"" +
						   MultimethodBodyAttribute.tagString +
						   "\" illegal as sub-attribute of Attribute Code");
	    }
	    break;

	default:
	    break;
	}

	return null;
    }

    // -------------------------------------------------------------
    // PRIVATE HELPERS
    // -------------------------------------------------------------

    /**
     * Returns true if the attribute named by the given tag String is
     * an attribute of org.multijava.  */
    private boolean isOurs(String tag) {
	return tag.startsWith(tagPrefix);
    }

    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------

    private static final String tagPrefix = "org.multijava.";
    private static final int firstInterestingPos = tagPrefix.length();
}
