/*
 * Copyright (C) 2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: RuntimeVisibleAnnotationsAttribute.java,v 1.1 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.util.ArrayList;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

/**
 * A class representing a RuntimeVisibleAnnotations attribute.
 * VMS 4.8.14
 *
 * @author AHS
 */
public class RuntimeVisibleAnnotationsAttribute extends Attribute {

    /**
     * Create a new RuntimeVisibleAnnotations attribute for the given
     * runtime-visible annotations.
     *  
     * @param annotations	the runtime-visible annotations, this reference
     * 				is captured
     */
    public RuntimeVisibleAnnotationsAttribute(ArrayList annotations) {
        this.annotations = annotations;
    }
    
    /**
     * Construct a new RuntimeVisibleAnnotations attribute from a classfile
     * stream.
     *
     * @param in	the stream to read from
     * @param cp	the constant pool
     * @throws IOException		an IO problem has occured
     * @throws ClassFileFormatException	the classfile is malformed
     */
    public RuntimeVisibleAnnotationsAttribute(DataInput in, ConstantPool cp)
	    throws IOException, ClassFileFormatException
    {
        int length = in.readInt();
        int num_annotations = in.readUnsignedShort();
        
        annotations = new ArrayList();
        for (int i = 0; i < num_annotations; i++) {
            Annotation annot = new Annotation(cp, in);
            annotations.add(annot);
        }
    }

        
    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------
    
    /**
     * Return the annotation given by the name, or null if no such annotation
     * has been attached to the element.
     */
    public Annotation get(String name) {
        for (int i = 0; i < annotations.size(); i++) {
            Annotation annot = (Annotation) annotations.get(i);
            if (annot.getName().equals("L" + name.replace('.', '/') + ";"))
                return annot;
        }
        return null;
    }

    /**
     * Returns the attribute's tag.
     */
    protected int getTag() {
        return Constants.ATT_RUNTIME_VISIBLE_ANNOTATIONS;
    }

    /**
     * Returns the space in bytes used by this attribute in the classfile.
     */
    protected int getSize() {
        int size = 2 + 4 + 2;
        for (int i = 0; i < annotations.size(); i++) {
            size += ((Annotation)annotations.get(i)).getSize();
        }
        return size;
    }

        
    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param	cp		the constant pool for this class
     */
    protected void resolveConstants(ConstantPool cp) {
        cp.addItem(attr);
        
        for (int i = 0; i < annotations.size(); i++) {
            ((Annotation)annotations.get(i)).resolveConstants(cp);
        }
    }

    /**
     * Write this attribute into the file (out) getting data position from
     * the constant pool.
     *
     * @param cp	the constant pool that contain all data
     * @param out	the file where to write this object info
     *
     * @exception java.io.IOException	an io problem has occured
     */
    protected void write(ConstantPool cp, DataOutput out) throws IOException {
        out.writeShort(attr.getIndex());        
        out.writeInt(getSize() - 6);
        out.writeShort(annotations.size());
        for (int i = 0; i < annotations.size(); i++) {
            ((Annotation)annotations.get(i)).write(cp, out);
        }
    }

        
    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    static final String 		tagString = "RuntimeVisibleAnnotations";
    private static AsciiConstant	attr = new AsciiConstant(tagString);
    
    /** the runtime-visisble annotations */
    private ArrayList annotations;
}

