/*
 * Copyright (C) 2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: RuntimeVisibleParameterAnnotationsAttribute.java,v 1.1 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.ArrayList;

/**
 * A class representing a RuntimeVisibleParameterAnnotations attribute.
 * VMS 4.8.16
 *
 * @author AHS
 */
public class RuntimeVisibleParameterAnnotationsAttribute extends Attribute {

    /**
     * Create a new RuntimeVisibleParameterAnnotations attribute for a given
     * array of annotations. Every i-th element of the array is a list of
     * annotations for the i-th parameter. The number of parameters is equal
     * to the length of the array.
     *  
     * @param annotations	parameter annotations, reference is captured
     */
    public RuntimeVisibleParameterAnnotationsAttribute(ArrayList[] annotations) {        
        this.annotations = annotations;
    }
    
    
    /**
     * Constructs a RuntimeVisibleParameterAnnotations attribute from a
     * classfile stream.
     *
     * @param in	the stream to read from
     * @param cp	the constant pool
     *
     * @exception java.io.IOException		an IO problem has occured
     * @exception ClassFileFormatException	the classfile is malformed
     */
    public RuntimeVisibleParameterAnnotationsAttribute(DataInput in, ConstantPool cp)
    		throws IOException, ClassFileFormatException
    {
        int length = in.readInt();
        int num_params = in.readUnsignedByte();
        annotations = new ArrayList[num_params];
        
        for (int i = 0; i < num_params; i++) {
            int num_annots = in.readUnsignedShort();
            annotations[i] = new ArrayList();
            for (int j = 0; j < num_annots; j++) {
                annotations[i].add(new Annotation(cp, in));
            }
        }
    }
    
    /**
     * Get the number of parameters as given in this attribute.
     */
    public int getParameterCount() {
        return annotations.length;
    }
    
    /**
     * Return the annotation of a given type for the specified parameter given
     * by its index, or null if no such annotation has been attached to the
     * parameter.
     * 
     * @param param	the index of the desired parameter
     * @param name	the name of the annotation (i.e. the interface type)
     */
    public Annotation get(int param, String name) {
        ArrayList annots = annotations[param];
        for (int i = 0; i < annots.size(); i++) {
            Annotation annot = (Annotation) annots.get(i);
            if (annot.getName().equals("L" + name.replace('.', '/') + ";"))
                return annot;
        }
        return null;
    }
    
    /**
     * Return the attribute's tag.
     */
    protected int getTag() {
        return Constants.ATT_RUNTIME_VISIBLE_PARAMETER_ANNOTATIONS;
    }

    /**
     * Return the space in bytes used by this attribute in the classfile.
     */
    protected int getSize() {
        int size = 2 + 4 + 1;
        for (int i = 0; i < annotations.length; i++) {
            size += 2; // u2 num_annotations
            for (int j = 0; j < annotations[i].size(); j++) {
                size += ((Annotation) annotations[i].get(j)).getSize();
            }
        }
        return size; 
    }

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param	cp		the constant pool for this class
     */
    protected void resolveConstants(ConstantPool cp)
            throws ClassFileFormatException {
        cp.addItem(attr);
        for (int i = 0; i < annotations.length; i++) {
            for (int j = 0; j < annotations[i].size(); j++) {
                ((Annotation) annotations[i].get(j)).resolveConstants(cp);
            }
        }
    }

    /**
     * Write this attribute into the file (out) getting data position from
     * the constant pool.
     *
     * @param cp	the constant pool that contain all data
     * @param out	the file where to write this object info
     *
     * @exception java.io.IOException	an io problem has occured
     */
    protected void write(ConstantPool cp, DataOutput out) throws IOException {
        out.writeShort(attr.getIndex());        
        out.writeInt(getSize() - 6);
        out.writeByte(annotations.length);
        
        for (int i = 0; i < annotations.length; i++) {
            out.writeShort(annotations[i].size());
            for (int j = 0; j < annotations[i].size(); j++) {
                ((Annotation) annotations[i].get(j)).write(cp, out);
            }
        }
    }

    static final String 		tagString = "RuntimeVisibleParameterAnnotations";
    private static AsciiConstant	attr = new AsciiConstant(tagString);
    
    /** the parameter annotations  */
    private ArrayList[] annotations;
}
