/*
 * Copyright (C) 2004-2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;
import org.multijava.mjc.Debug;

/**
 * This class implements an attribute parser that recognizes the JUTS
 * attributes. The parser is inserted into the attribute parsing chain of
 * responsibility.
 * 
 * @see AttributeList#addParser(AttributeParser)
 * 
 * @author AHS
 */
public class UniverseAttributeParser implements AttributeParser {
        
    /**
     * Checks the next attribute in the given input stream, and parses it if is
     * is recognized. Returns an Attribute instance if the attribute is
     * recognized, or null otherwise. 
     * 
     * @param tag	a String giving the name of the next attribute
     * @param in	the input stream for the the bytecode ClassInfo being
     * 			parsed whose read cursor is positioned just after
     * 			attribute_name_index of the next attribute (VMS 4.7)
     * @param cp	the constant pool for the ClassInfo being parsed
     * 
     * @exception IOException	an IO error has occurred
     * @exception ClassFileFormatException	the class file is malformed 
     */
    public Attribute read(String tag, DataInput in, ConstantPool cp)
            throws IOException, ClassFileFormatException
    {
	if (!isOurs(tag))
	    return null;
	
	switch (tag.charAt(firstInterestingPos)) {
	case 'u':
	    if (tag.equals(UniverseClassAttribute.tagString))
	        return new UniverseClassAttribute(in, cp);
	    else if (tag.equals(UniverseFieldAttribute.tagString))
	        return new UniverseFieldAttribute(in, cp);
	    else if (tag.equals(UniverseMethodAttribute.tagString))
	        return new UniverseMethodAttribute(in, cp);
	    break;
	default:
	    break;
	}
	
        return null;
    }

    /**
     * Checks the next attribute in the given input stream, and parses it if is
     * is recognized, but skips CodeInfo attributes Returns an Attribute instance
     * if the attribute is recognized, or null otherwise. 
     * 
     * @param tag	a String giving the name of the next attribute
     * @param in	the input stream for the the bytecode ClassInfo being
     * 			parsed whose read cursor is positioned just after
     * 			attribute_name_index of the next attribute (VMS 4.7)
     * @param cp	the constant pool for the ClassInfo being parsed
     * 
     * @exception IOException	an IO error has occurred
     * @exception ClassFileFormatException	the class file is malformed 
     */
    public Attribute readInterfaceOnly(String tag, DataInput in, ConstantPool cp)
            throws IOException, ClassFileFormatException
    {
	if (!isOurs(tag))
	    return null;
	
	switch (tag.charAt(firstInterestingPos)) {
	case 'u':
	    if (tag.equals(UniverseClassAttribute.tagString))
	        return new UniverseClassAttribute(in, cp);
	    else if (tag.equals(UniverseFieldAttribute.tagString))
	        return new UniverseFieldAttribute(in, cp);
	    else if (tag.equals(UniverseMethodAttribute.tagString))
	        return new UniverseMethodAttribute(in, cp);
	    break;

	default:
	    break;
	}
        
        return null;
    }

    /**
     * Checks the next attribute in the given input stream, and parses it if is
     * is recognized, but only parses attributes that may appear within a
     * CodeInfo attribute. Returns an Attribute instance if the attribute is
     * recognized, or null otherwise. 
     * 
     * @param tag	a String giving the name of the next attribute
     * @param in	the input stream for the the bytecode ClassInfo being
     * 			parsed whose read cursor is positioned just after
     * 			attribute_name_index of the next attribute (VMS 4.7)
     * @param cp	the constant pool for the ClassInfo being parsed
     * @param insns	(sparse) array of instructions
     * 
     * @exception IOException	an IO error has occurred
     * @exception ClassFileFormatException	the class file is malformed 
     */
    public Attribute readCodeInfoAttribute(String tag, DataInput in,
            ConstantPool cp, Instruction[] insns) throws IOException,
            ClassFileFormatException {
            
        if (!isOurs(tag))
	    return null;
	
	switch (tag.charAt(firstInterestingPos)) {
	case 'u':		
	    if (tag.equals(UniverseClassAttribute.tagString))
	        throw new ClassFileFormatException("Attribute \"" +
	                UniverseClassAttribute.tagString +
	        	"\" illegal as sub-attribute of Attribute Code");
	    else if (tag.equals(UniverseFieldAttribute.tagString))
	        throw new ClassFileFormatException("Attribute \"" +
	                UniverseFieldAttribute.tagString +
	        	"\" illegal as sub-attribute of Attribute Code");
	    else if (tag.equals(UniverseMethodAttribute.tagString))
	        throw new ClassFileFormatException("Attribute \"" +
	                UniverseMethodAttribute.tagString +
	        	"\" illegal as sub-attribute of Attribute Code");
	    break;
	    
	default:
	    break;
	}

	return null;
    }

    // -------------------------------------------------------------
    // PRIVATE HELPERS
    // -------------------------------------------------------------

    private boolean isOurs(String tag) {
	return tag.startsWith(tagPrefix);
    }

    //---------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    //---------------------------------------------------------------------

    /** the prefix of all names of attributes this parser may recognize */
    private static final String tagPrefix		= "org.multijava.";
    private static final int firstInterestingPos	= tagPrefix.length();
}
