/*
 * Copyright (C) 2004-2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: UniverseByteConstants.java,v 1.2 2008/03/16 15:58:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import org.multijava.mjc.CUniverse;
import org.multijava.mjc.CUniverseImplicitPeer;
import org.multijava.mjc.CUniversePeer;
import org.multijava.mjc.CUniverseRep;
import org.multijava.mjc.CUniverseReadonly;


/**
 * This class contains static helper functions that work on the byte
 * constants used to represent universe annotations in attributes.
 *
 * @author AHS
 */
public class UniverseByteConstants {
    
    /**
     * Returns if a given byte constant represents an array type.
     * 
     * @param byteConstant a byte constant representing a universe annotation
     */
    public static boolean isArray(byte byteConstant) {
        return (byteConstant & 0x80) != 0;
    }
    
    /**
     * Build byte constant from a given universe.
     * 
     * @param univ the universe
     */
    public static byte buildConstant(CUniverse univ) {
        return univ.getByteConstant();
    }
    
    /**
     * Build byte constant from a pair of universes (array type).
     * 
     * The hi bit is set to 1 to signify an array annotation.
     * 
     * @param univ the universe of the array itself
     * @param base_univ the universe of the array elements
     */
    public static byte buildConstant(CUniverse univ, CUniverse base_univ) {
        return (byte) (0x80 | base_univ.getByteConstant() << 2 | univ.getByteConstant());
    }
    
    /**
     * Returns an instance of the universe represented by a given byte
     * constant. For array annotations, this returns the universe of the
     * array object itself, not of the array elements.
     * The universe is stored in the low 2 bits of the byte constant.
     * 
     * @param byteConstant a byte constant representing a universe annotation
     */
    public static CUniverse toUniverse(byte byteConstant) {
        switch (byteConstant & 0x3) {
        case CUniversePeer.byteConstant:
            return CUniversePeer.getUniverse();
        case CUniverseRep.byteConstant:
            return CUniverseRep.getUniverse();
        case CUniverseReadonly.byteConstant:
            return CUniverseReadonly.getUniverse();
        default:
            return CUniverseImplicitPeer.getUniverse();
        	// WMD tried using the below, but then the build had a circle... 
        	// return CUniverseServices.getDefaultVariable();
        }
    }
    
    /**
     * Returns an instance of the universe represented by a a given
     * byte constant. This is for array annotations only, and returns the
     * universe of the array elements.
     * The universe of the array elements is stored in bits [3,2] of the
     * byte constant.
     * 
     * @param byteConstant a byte constant representing a universe annotation
     */
    public static CUniverse toBaseUniverse(byte byteConstant) {
        return toUniverse((byte) ((byteConstant & 0xC) >> 2));
    }
}

