/*
 * Copyright (C) 2004-2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: UniverseClassAttribute.java,v 1.1 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

/**
 * This class implements a JUTS class attribute. It contains the JUTS
 * version (e.g. "1.0") plus a boolean flag that indicates if the class has
 * been compiled with run-time checks enabled.
 * 
 * The format of the attribute (not including the standard initial six bytes):
 * - the first two bytes are an index into the constant pool. The constant
 *   pool entry at that index must be String constant
 * - the next byte is interpreted as a boolean value, and tells if the class
 *   has run-time checks enabled or not 
 * 
 * This attribute is recognized by UniverseAttributeParser.
 * 
 * VMS 4.7.1: The attribute's name is org.multijava.universe_class
 * 
 * @see UniverseAttributeParser
 * 
 * @author AHS
 */
public class UniverseClassAttribute extends Attribute {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Create a Universe version attribute.
     * 
     * @param version		the version of the Universe Type System used
     * @param runtimeSupport	does the class have run-time checks
     */
    public UniverseClassAttribute(String version, boolean runtimeSupport) {
        this.version = new AsciiConstant(version);
        this.runtimeSupport = runtimeSupport;
    }

    /**
     * Constructs a universe version attribute from a classfile stream.
     *
     * @param in	the stream to read from
     * @param cp	the constant pool
     *
     * @exception	java.io.IOException	an IO problem has occured
     * @exception	ClassFileFormatException the classfile is malformed
     */
    public UniverseClassAttribute(DataInput in, ConstantPool cp)
	    throws IOException, ClassFileFormatException {
        if (in.readInt() != 3)
            throw new ClassFileFormatException("Bad attribute length");
        this.version = (AsciiConstant) cp.getEntryAt(in.readUnsignedShort());
        this.runtimeSupport = in.readBoolean();
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Returns the attribute's tag.
     */
    protected int getTag() {
        return Constants.ATT_UNIVERSE_VERSION;
    }

    /**
     * Returns the space in bytes used by this attribute in the classfile.
     */
    protected int getSize() {
        return 2 + 4 + 3;
    }

    /**
     * Return the version of the Universe Type System used.
     * 
     * @see #version
     */
    public String getVersion() {
        return version.getValue();
    }
    
    /**
     * Return if the class was compiled with run-time support enabled.
     * 
     * @see #runtimeSupport
     */
    public boolean hasRuntimeSupport() {
        return runtimeSupport;
    }

    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param	cp		the constant pool for this class
     */
    protected void resolveConstants(ConstantPool cp) {
        cp.addItem(attr);
        cp.addItem(version);
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool.
     *
     * @param cp	the constant pool
     * @param out	the output stream
     *
     * @exception java.io.IOException	an IO problem has occurred
     */
    protected void write(ConstantPool cp, DataOutput out) throws IOException {
        out.writeShort(attr.getIndex());
        out.writeInt(3);
        out.writeShort(version.getIndex());
        out.writeBoolean(runtimeSupport);
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------
    
    /** the name of the attribute */
    static final String 		tagString = "org.multijava.universe_class";
    private static AsciiConstant	attr = new AsciiConstant(tagString);
    
    /** the version of JUTS used */
    private AsciiConstant		version;
    
    /** does the class have run-time checks enabled? */
    private boolean			runtimeSupport;
}

