/*
 * Copyright (C) 2004-2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: UniverseFieldAttribute.java,v 1.1 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;

/**
 * This class implements a JUTS field attribute. It is basically just a wrapper
 * for one byte that describes the universe type modifier for a field of
 * reference or array type.
 * 
 * The format of the attribute (not including the standard initial six bytes):
 * - the one and only byte is the encoded universe type modifier for this field
 * 
 * This attribute is recognized by UniverseAttributeParser.
 *  
 * VMS 4.7.1: The attribute's name is org.multijava.universe_field
 * 
 * @see UniverseAttributeParser
 * 
 * @author AHS
 */
public class UniverseFieldAttribute extends Attribute {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------
    
    /**
     * Create a universe field attribute for an encoded universe type modifier.
     * 
     * @param byteConstant	the encoded universe type modifier 
     */
    public UniverseFieldAttribute(byte byteConstant) {
        this.byteConstant = byteConstant;
    }   
    
    /**
     * Constructs a universe field attribute from a class file stream.
     *
     * @param in	the stream to read from
     * @param cp	the constant pool
     *
     * @exception java.io.IOException	an IO problem has occurred
     * @exception ClassFileFormatException	the classfile is malformed
     */
    public UniverseFieldAttribute(DataInput in, ConstantPool cp)
	    throws IOException, ClassFileFormatException {
        if (in.readInt() != 1)
            throw new ClassFileFormatException("Bad attribute length!");
        this.byteConstant = in.readByte();
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Returns the attribute's tag.
     */
    protected int getTag() {
        return Constants.ATT_UNIVERSE_FIELD;
    }

    /**
     * Returns the space in bytes used by this attribute in the classfile.
     */
    protected int getSize() {
        return 2 + 4 + 1;
    }
    
    /**
     * Return the universe of the reference type.
     * 
     * @see #byteConstant
     */
    public byte getByteConstant() {
        return byteConstant;
    }
    
    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param	cp		the constant pool for this class
     */
    protected void resolveConstants(ConstantPool cp) {
        cp.addItem(attr);
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool.
     *
     * @param	cp		the constant pool that contain all data
     * @param	out		the file where to write this object info
     *
     * @exception	java.io.IOException	an io problem has occured
     */
    protected void write(ConstantPool cp, DataOutput out) throws IOException {
        out.writeShort(attr.getIndex());
        out.writeInt(getSize() - 6);
        out.writeByte(byteConstant);
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    /** the attribute's name */
    static final String 		tagString = "org.multijava.universe_field";
    private static AsciiConstant	attr = new AsciiConstant(tagString);
    
    /** the byte holding the encoded universe type modifier */
    private byte			byteConstant;
}

