/*
 * Copyright (C) 2004-2005 Swiss Federal Institute of Technology Zurich
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: UniverseMethodAttribute.java,v 1.1 2005/06/24 12:37:58 wdietl Exp $
 */

package org.multijava.util.classfile;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

import org.multijava.mjc.CUniverseImplicitPeer;

/**
 * This class implements a JUTS method attribute. The attribute contains
 * information about the method's purity and the universe annotations of its
 * parameters and return type.
 * 
 * The format of the attribute (not including the standard initial six bytes):
 * - the first byte is either 0 or 1, and represents the method's purity
 * - the second byte is the return type's annotation. This is zero if the
 *   method has no return type
 * - the following bytes are the parameter's annotations
 * 
 * This attribute is recognized by UniverseAttributeParser.
 * 
 * VMS 4.7.1: The attribute's name is org.multijava.universe_method
 * 
 * @see UniverseAttributeParser
 * 
 * @author AHS
 */
public class UniverseMethodAttribute extends Attribute {
    
    /**
     * Construct a method attribute for a method with void return type.
     *
     * @param isPure the purity of the method 
     * @param paramConstants the byte constants for the parameters
     */
    public UniverseMethodAttribute(boolean isPure, byte paramConstants[]) {
        this(isPure, CUniverseImplicitPeer.byteConstant, paramConstants);
    }

    /**
     * Construct a method attribute for a method with non-void return type.
     *
     * @param isPure the purity of the method
     * @param retTypeByteConstant the byte constant for the return type 
     * @param paramConstants the byte constants for the parameters
     */
    public UniverseMethodAttribute(boolean isPure, byte retTypeByteConstant,
            			   byte paramConstants[]) {
        this.isPure = isPure;
        this.retConstant = retTypeByteConstant;
        this.paramConstants = (byte[]) paramConstants.clone();
    }
    
    
    /**
     * Constructs a Universe Method Attribute from a class file stream.
     *
     * @param	in		the stream to read from
     * @param	cp		the constant pool
     *
     * @exception	java.io.IOException	an io problem has occured
     * @exception	ClassFileFormatException	attempt to
     *					write a bad classfile info
     */
    public UniverseMethodAttribute(DataInput in, ConstantPool cp)
	    throws IOException, ClassFileFormatException {
        int length = in.readInt();
        int numParams = length - 1 - 1;
        
        isPure = in.readByte() != 0 ? true : false;
        retConstant = in.readByte();
        
        paramConstants = new byte[numParams];
        for (int i = 0; i < numParams; i++)
            paramConstants[i] = in.readByte();
    }
    
    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------
    
    /**
     * Return the method's purity.
     */
    public boolean getPurity() {
        return isPure;
    }
    
    /**
     * Return the byte constant representing the return type's annotation.
     */
    public byte getRetTypeConstant() {
        return retConstant;
    }
    
    /**
     * Return the byte constants representing the parameter annotations.
     */
    public byte[] getParamConstants() {
        return paramConstants;
    }
    
    /**
     * Return this attribute's tag.
     */
    protected int getTag() {
        return Constants.ATT_UNIVERSE_METHOD;
    }

    /**
     * Return the attribute's size, this attribute has variable size dependent
     * on the number of parameters.
     */
    protected int getSize() {
        return 2 + 4 + 1 + 1 + paramConstants.length;
    }

    // --------------------------------------------------------------------
    // WRITE
    // --------------------------------------------------------------------

    /**
     * Insert or check location of constant value on constant pool.
     *
     * @param	cp		the constant pool for this class
     */
    protected void resolveConstants(ConstantPool cp)
    		throws ClassFileFormatException {
        cp.addItem(attr);
    }

    /**
     * Write this class into the the file (out) getting data position from
     * the constant pool.
     *
     * @param	cp		the constant pool that contain all data
     * @param	out		the file where to write this object info
     *
     * @exception	java.io.IOException	an io problem has occured
     */
    protected void write(ConstantPool cp, DataOutput out) throws IOException,
            ClassFileFormatException {
        out.writeShort(attr.getIndex());
        out.writeInt(getSize() - 6);
        out.writeByte(isPure ? 1 : 0);
        out.writeByte(retConstant);
        for (int i = 0; i < paramConstants.length; i++)
            out.writeByte(paramConstants[i]);
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    /** the attribute's name */
    static final String 		tagString = "org.multijava.universe_method";
    private static AsciiConstant	attr = new AsciiConstant(tagString);
    
    /** the method's purity */
    private boolean			isPure;
    
    /** the encoded universe type modifier for the return type */
    private byte 			retConstant;
    
    /** the encoded universe type modifiers for the parameter types */
    private byte 			paramConstants[];
}
