/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Compiler.java,v 1.20 2008/05/14 18:17:10 wdietl Exp $
 */

package org.multijava.util.compiler;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.OutputStream;
import java.io.PrintWriter;

import java.util.ArrayList;
import java.util.Arrays;

import org.multijava.util.Message;
import org.multijava.util.MessageDescription;
import org.multijava.util.Utils;
import org.multijava.util.FormattedException;

/**
 * This class defines the common behaviour of all KOPI compilers.
 */
public abstract class Compiler extends Utils implements TroubleReporter {

    /**
     * Creates a new compiler instance.
     */
    protected Compiler( ModifierUtility modUtil ) {
        timestamp = System.currentTimeMillis();
        this.modUtil = modUtil;
    }

    /**
     * Returns the timestamp of the compilation session,
     * identifying it uniquely.
     */
    public final long getTimestamp() {
        return timestamp;
    }

    /**
     * Takes a list of file names and checks that each exists.
     * @param   names                   a list of names
     * @return  a list (of java.io.File) of files known to exist
     * @exception UnpositionedError at least one file does not exist */
    public ArrayList verifyFiles( ArrayList names ) throws UnpositionedError {
        String[] nameArray = new String[ names.size() ];
        names.toArray( nameArray );
        return verifyFiles( nameArray );
    }

    /**
     * Takes an array of file names and checks that each exists.
     * @param   names                   an array of names
     * @return  a list (of java.io.File) of files known to exist
     * @exception UnpositionedError at least one file does not exist */
    public ArrayList verifyFiles( String[] names ) throws UnpositionedError {
        ArrayList       accum = new ArrayList(Arrays.asList(names));
        ArrayList files = new ArrayList(accum.size());

        for (int i = 0; i < accum.size(); i++) {
            String name = (String) accum.get(i);
            File file = new File( name ).getAbsoluteFile();

            if( !file.exists() ) {
                throw new UnpositionedError( CompilerMessages.UNKNOWN_FILE, 
                                             name );
            }
            if (file.isDirectory()) {
                throw new UnpositionedError( CompilerMessages.DIRECTORY_FILE,
                                             name );
            }
            
            files.add(file);
        }
        
        return files;
    }

    /**
     * Parses the file referenced by <code>name</code>, extracting the
     * list of files that it contains.  The files are added to
     * <code>accum</code>
     *
     * @param   accum   a list in which to accumulate the results
     * @param   name    a string of the form "@file" where "file" is the 
     * name of a file containing a list of file names */
    private void readList( ArrayList accum, String name ) throws IOException {
        File file = new File( name.substring(1) );
        LineNumberReader reader = null;
        try {
            reader = new LineNumberReader(new FileReader(file));
            String str;

            while ((str = reader.readLine()) != null) {
                if (str.length() >  1) {
                    accum.add(str);
                }
            }
        } finally {
            if (reader != null) {
                reader.close();
            }
        }
    }

    /**
     * Runs a compilation session
     *
     * @param   err             the diagnostic output stream
     * @param   args            the arguments to the compiler
     * @return  true iff the compilation succeeded
     */
    public boolean run( PrintWriter err, String[] args ) {
        this.err = err;
    
        return run(args);
    }
        
    /**
     *  Sets the output stream for inform to <code>err</code>
     * @param   os the output stream to use for error and warning output
     *
     * <pre><jml>
     * requires os != null;
     * </jml></pre>
     */
    public void setOutputStream(OutputStream os) {
        this.err = new PrintWriter(os);
        this.out = new PrintWriter(os);
    }

    // --------------------------------------------------------------------
    // METHODS TO BE IMPLEMENTED BY SUBCLASSES
    // --------------------------------------------------------------------

    /**
     * Runs a compilation session
     *
     * @param   args            the arguments to the compiler
     * @return  true iff the compilation succeeded
     */
    public abstract boolean run(String[] args);


    /**
     * Compiles the named file (using the options of the current
     * compilation session) up to the same compilation pass as the
     * files named when the compilation session was initiated.  Then
     * adds the result to the set of compilation units to be processed
     * to completion.
     *
     * @param filename the file descriptor of the file to be compiled
     *
     * <pre><jml>
     * requires filename != null && filename.canRead();
     * </jml></pre>
     *
     */
    public abstract void catchUp(File filename);

    /**
     * Reports a trouble (error or warning).
     *
     * @param   trouble         a description of the trouble to report.
     */
    public abstract void reportTrouble(/*@\readonly@*/ Exception trouble);

    /**
     * Returns true iff comments should be parsed (false if to be skipped).
     */
    public abstract boolean parseComments();

    /**
     * Returns true iff compilation runs in verbose mode.
     */
    public abstract boolean verboseMode();

    // --------------------------------------------------------------------
    // DIAGNOSTIC OUTPUT
    // --------------------------------------------------------------------

    /**
     * Write a message to the diagnostic output.
     * @param   trouble         the error caught
     */
    public final void inform(/*@\readonly@*/ Throwable trouble) {
        inform(trouble.getMessage(), PRINT_TO_ERR);
    }

    /**
     * Write a message to the diagnostic output.
     * @param   trouble         the error caught
     */
    public final void inform(/*@\readonly@*/ FormattedException trouble) {
        inform(trouble.getErrorMessage(), PRINT_TO_ERR);
    }

    /**
     * Write a message to the diagnostic output.
     * @param   message         the formatted message
     */
    public final void inform(Message message) {
        inform(message.getMessage(), PRINT_TO_ERR);
    }

    /**
     * Write a message to the diagnostic output.
     * @param   description     the message description
     * @param   parameters      the array of parameters
     */
    public final void inform(MessageDescription description, /*@\readonly \readonly@*/ Object[] parameters) {
        inform(new Message(description, parameters));
    }

    /**
     * Write a message to the diagnostic output.
     * @param   description     the message description
     * @param   parameter1      the first parameter
     * @param   parameter2      the second parameter
     */
    public final void inform (MessageDescription description, 
                        /*@\readonly@*/ Object parameter1, /*@\readonly@*/ Object parameter2 ) 
    {
        inform(description, new /*@\peer \readonly@*/ Object[] { parameter1, parameter2 });
    }

    /**
     * Write a message to the diagnostic output.
     * @param   description     the message description
     * @param   parameter       the parameter
     */
    public final void inform(MessageDescription description, /*@\readonly@*/ Object parameter) {
        inform(description, new /*@\peer \readonly@*/ Object[] { parameter });
    }

    /**
     * Write a message to the diagnostic output.
     * @param   description     the message description
     */
    public final void inform(MessageDescription description) {
        inform(description, null);
    }
        
    /**
     * Write a message to the diagnostic output.
     * @param   message         the message text
     */
    public void inform(String message) {
        inform(message, PRINT_TO_ERR);
    }

    /**
     * Write text to the output stream if printDestination is PRINT_TO_OUT and
     * write text to the diagnostic stream if printDestination is PRINT_TO_ERR.
     * Subclasses should override setOutputStream() to change the output
     * destination.
     * @param   message          the message text
     * @param   printDestination the output stream that this message goes to
     * @see #setOutputStream(OutputStream)
     */
    public void inform(String message, int printDestination) {
	//(new Exception()).printStackTrace();
        switch (printDestination) {
        case 0:
            out.println(message);
            out.flush();
			break;
        case 1:
        default:
            err.println(message);
            err.flush();
        }
    }

    public /*@ pure @*/ final ModifierUtility modUtil() {
        return modUtil;
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    /** Indicates that the user wants to print to normal output. */
    public static final int PRINT_TO_OUT = 0;
        
    /** Indicates that the user wants to print to diagnostic output. */
    public static final int PRINT_TO_ERR = 1;
        
    /** Each compilation session is identified by a timestamp */
    private final long timestamp;

    /** The output stream for normal messages */
    private PrintWriter out = new PrintWriter(System.out);
        
    /** The output stream for diagnostic messages */
    private PrintWriter err = new PrintWriter(System.err);
        
    private final ModifierUtility modUtil;
}
