/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TokenReference.java,v 1.13 2008/05/14 18:17:11 wdietl Exp $
 */

package org.multijava.util.compiler;

import java.io.File;

import org.multijava.util.Utils;

/**
 * This class defines objets that hold a reference to a file and a position
 * in the file.
 */
public class TokenReference {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Construct a file and line reference
     * @param	file		the file
     * @param	line		the line number
     */
    private /*@pure@*/ TokenReference( /*@nullable*/ /*@\readonly@*/ File file, int line ) {
	this( file, line, DEF_COLUMN );
    }

    /**
     * Construct a file, line, and column reference
     * @param	file		the file
     * @param	line		the line number
     */
    private /*@pure@*/ TokenReference( /*@nullable*/ /*@\readonly@*/ File file, int line, int column ) {
	this.file = file;
	this.line = line;
	this.column = column;

	last = this;
    }

    /**
     * Construct a file and line reference
     * @param	file		the file
     * @param	line		the line number
     * WARNING: DOES NOT LIKE MULTITHREADING
     */
    public static /*@ pure @*/ TokenReference build( /*@nullable*/ File file, int line ) {
	// The comparison of files by object identity is intentional
	// here.  Comparing by equals might be more space efficient,
	// but this method is called so often that the time penalty
	// would be severe.  We can avoid this time penalty without
	// any space penalty by just passing the same File object to
	// successive calls to this method.  That is the typical usage
	// pattern anyway.
	if( line != last.line || file != last.file 
	    || last.column != DEF_COLUMN ) {
	    return new TokenReference( file, line );
	} else {
	    return last;
	}
    }

     /**
     * Construct a file, line, and column reference.
     * @param	file		the file
     * @param	line		the line number
     * @param	column		the column number
     * WARNING: DOES NOT LIKE MULTITHREADING
     */
    public static /*@ pure @*/ TokenReference build( /*@nullable*/ File file, 
						      int line, 
						      int column )
    {
	// The comparison of files by object identity is intentional
	// here.  Comparing by equals might be more space efficient,
	// but this method is called so often that the time penalty
	// would be severe.  We can avoid this time penalty without
	// any space penalty by just passing the same File object to
	// successive calls to this method.  That is the typical usage
	// pattern anyway.
	if( line != last.line || file != last.file || column != last.column ) {
	    return new TokenReference( file, line, column );
	} else {
	    return last;
	}
    }

   // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------

    /**
     * Returns the file to which this token reference refers.
     */
    public final /*@ pure nullable @*/ /*@\readonly@*/ File file() {
	return file;
    }

    /**
     * Returns the short file name of this reference (without any path
     * name).  Result is cached for efficient access.
     *
     * <pre><jml>
     * assignable cachedName;
     * ensures file() != null ==> \result.equals(file().getName());
     * ensures file() == null ==> \result == UNKNOWN;
     * </jml></pre>
     * */
    public final String name() {
	if (cachedName != null) {
	    return cachedName;
	}

	cachedName = file != null ? file.getName().intern() : UNKNOWN;
	return cachedName;
    }

    /**
     * Returns the line number of reference
     */
    public final /*@ pure @*/ int line() {
	return line;
    }

    /**
     * Returns the column number of reference
     */
    public final /*@ pure @*/ int column() {
	return column;
    }

    public /*@ pure @*/ String toString() {
	return "File \"" + relativeFilePath() + "\", line " 
	    + line + ((column!=DEF_COLUMN) ? (", character " + column) : "");
    }

    private /*@ pure @*/ String relativeFilePath() {
	if (file != null) {
	    return Utils.relativePathTo(file);
	    //return file.toString();
	} else {
	    return UNKNOWN;
	}
    }

    // ----------------------------------------------------------------------
    // DATA MEMBERS
    // ----------------------------------------------------------------------

    public static final /*@non_null*/ String UNKNOWN = "unknown_source_file";

    /** The default column number when none is specified */
    private static final int DEF_COLUMN = 0;

    public static /*@non_null*/ TokenReference NO_REF = 
	new TokenReference( new File("<GENERATED-BY-mjc>"), DEF_COLUMN );

    private static /*@non_null*/ TokenReference last = NO_REF;

    private final /*@nullable*/ /*@\readonly@*/ File file;
    private final int line;
    private final int column;

    /**
     * The name of the file referred to by this token, calculated
     * lazily.  The value is interned.
     *
     * <pre><jml>
     * private invariant (* cachedName != null ==> name() == cachedName *);
     * </jml></pre>
     *
     * @see java.lang.String#intern()
     */
    private /*@spec_public nullable*/ String cachedName;
}
