/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;

import java.util.ArrayList;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JTabbedPane;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.MenuElement;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import javax.swing.filechooser.FileFilter;

import org.multijava.util.Options;

/**
 *  A graphical user interface for running Mjc.
 *
 *@author     Kristina Boysen
 *@author     Gary T. Leavens
 *@author     David Behroozi
 *@version    $Revision: 1.6 $
 */

public abstract class GUI extends GUIFrame {

    // ----------------------------------------------------------------------
    // CONSTRUCTOR
    // ----------------------------------------------------------------------
        
    /**
     *  Initializes the user interface
     */
    protected GUI(String name, Options options, ArrayList infiles,
                  boolean commandLine, boolean standAlone) {
        super(name + "GUI");
        opt = options;
        this.standAlone = standAlone;
        
        // Creates the class that controls how the options are displayed.
        // Adds a listener to these options
        optionHandler = new GUIOptionHandler(opt, commandLine);
        optionHandler.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    // Sets whether the "Save Preferences" menu item is 
                    // selected, based on whether the normal or advanced
                    // options have changed.
                    setSavePrefsEnabled(optionHandler.prefsChanged());
                }
            });
                
        // Adds the infiles to a new instance of fileList, which keeps track
        // of the files that will be processed with the tool
        // If any files have come in from the command line, add them.
        fileModel = new DefaultListModel();
        fileList = new JList(fileModel);
        fileList.setVisibleRowCount(6);
        fileList.setSelectionMode(
            ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
        if (infiles.size() > 0) {
            for (int i = 0; i < infiles.size(); i++) {
                fileModel.addElement((infiles.get(i)).toString());
            }
            fileList.setSelectedIndex(infiles.size());
        }
                        
        // Checks the classpath to see whether the specs directory is included
        // If true, then initialization can continue
        if (specsDirectoryDefined()) {
            // Creates the class that controls the window preferences
            mainWindowPrefs = new GUIWindowPreferences(toolName, "Main", this);
                                  
            // Creates the GUI components
            getContentPane().setLayout(new BorderLayout());
            setJMenuBar(createMenus());
            getContentPane().add(createToolBar(), BorderLayout.NORTH);
            getContentPane().add(createBasicPanel());
        
            // Sets preferences and displays
            mainWindowPrefs.setAllPreferences();
            setResizable(true);
            setVisible(true);
        }
    }
        
    /**
     *  Opens a dialog box if the user has not specified a specs directory
     *  in the classpath.  Returns true when the directory is in the
     *  classpath.
     */
    private boolean specsDirectoryDefined() {
        if (!dirInClassPath(opt, "specs")) {
            final Object[] buttonNames = {"Browse...", "Ignore"};
            final JOptionPane optionPane = getFormattedOptionPane(50);
            optionPane.setMessage(
                "The sourcepath or classpath set in the following tool " +
                "does not contain an existing specs directory. " +
                "\n \n" +
                "Please browse the file system for an existing specs " +
                "directory or ignore this message to check the " +
                "correctness of your sourcepath or classpath." +
                "\n \n" +
                "Classpath of the tool: \n" + 
                getClasspath(opt) + 
                "\n \n" +
                "Sourcepath of the tool: \n" + 
                getSourcepath(opt));
            optionPane.setMessageType(JOptionPane.WARNING_MESSAGE);
            optionPane.setOptionType(JOptionPane.YES_NO_OPTION);
            optionPane.setIcon(null);
            optionPane.setOptions(buttonNames);
            optionPane.setInitialValue(buttonNames[0]);
            
            final JDialog dialog = new JDialog(this, "Specs Missing", true);
            dialog.setContentPane(optionPane);
            dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
            optionPane.addPropertyChangeListener(
                new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent e) {
                        String prop = e.getPropertyName();
                        if (dialog.isVisible() 
                            && (e.getSource() == optionPane)
                            && (prop.equals(JOptionPane.VALUE_PROPERTY))) {
                            dialog.setVisible(false);
                        }
                    }
                });

            dialog.pack();
            GUIUtil.centerWindow(dialog);
            dialog.setVisible(true);
                        
            String value = optionPane.getValue().toString();
            if (value == buttonNames[0]) {
                return new FindSpecsHandler().handleFileSelection(opt);
            }
        }
        return true;
    }
        
    /**
     *  Returns a subclass of JOptionPane that restricts the wordwrap
     *  on the dialog box is restricted to <code>paneWidth</code> characters 
     *  per line.
     */
    private static JOptionPane getFormattedOptionPane(int paneWidth) {

        class FormattedOptionPane extends JOptionPane {
            int paneWidth;
            
            FormattedOptionPane(int paneWidth) {
                this.paneWidth = paneWidth;
            }
                        
            public int getMaxCharactersPerLineCount() {
                return paneWidth;
            }
        }
        
        return new FormattedOptionPane(paneWidth);
    }


    // ----------------------------------------------------------------------
    // GUI CONSTRUCTOR METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Creates the general layout panels for the GUI.
     */
    private JPanel createBasicPanel() {
        JPanel basicPanel = new JPanel();
        basicPanel.setLayout(new BoxLayout(basicPanel, BoxLayout.Y_AXIS));
        basicPanel.add(createFileListPanel());
        basicPanel.add(createOptionsPanel());

        return basicPanel;
    }

    /**
     *  Create and initialize the file list panel on the main panel.
     */
    private JPanel createFileListPanel() {
        return new GUIFileListPanel(fileList, createOpenHandlerInstance());
    }

    /**
     *  Creates the panel which allows the user to change options.
     */
    private JTabbedPane createOptionsPanel() {
        // Create layout that the options sit in
        JTabbedPane optionsPane = new JTabbedPane();
        optionsPane.setBorder(
            BorderFactory.createTitledBorder("Tool Options"));

        // Create tab for the "Normal" options
        if (optionHandler.getNormalOptionPanel() != null) {
            optionsPane.addTab("Normal Options", null, 
                               optionHandler.getNormalOptionPanel(), 
                               "Frequently used options");
        }
        
        // Create tab for the "Advanced" options
        if (optionHandler.getAdvancedOptionPanel() != null) {
            optionsPane.addTab("Advanced Options", null,
                               optionHandler.getAdvancedOptionPanel(), 
                               "Other options");
        }
        
        // Create tab for the "Normal Javadoc" options
        if (optionHandler.getJavadocNormalOptionPanel() != null) {
            optionsPane.addTab("Normal Javadoc Options", null,
                               optionHandler.getJavadocNormalOptionPanel(),
                               "Frequently used javadoc options");
        }
        
        // Create tab for the "Tag Omission Javadoc" options
        if (optionHandler.getJavadocNoOptionPanel() != null) {
            optionsPane.addTab("Tag Omission Javadoc Options", null,
                               optionHandler.getJavadocNoOptionPanel(), 
                               "Options indicating not include certain tags " 
                               + "in the generated HTML");
        }
        
        // Create tab for the "Advanced Javadoc" options
        if (optionHandler.getJavadocAdvancedOptionPanel() != null) {
            optionsPane.addTab("Advanced Javadoc Options", null,
                               optionHandler.getJavadocAdvancedOptionPanel(), 
                               "Other javadoc options");
        }
        
        return optionsPane;
    }
    
    /**
     *  Create the toolbar
     */
    private JToolBar createToolBar() {
        JToolBar toolbar = new JToolBar();
        toolbar.setFloatable(false);

        JButton open = new JButton();
        open.setIcon(new ImageIcon(GUIUtil.loadIconFromJar("open.gif")));
        open.setToolTipText("Open File");
        open.addActionListener(createOpenHandlerInstance());

        JButton run = new JButton();
        run.setIcon(new ImageIcon(GUIUtil.loadIconFromJar("run.gif")));
        Font buttonFont = new Font("sansserif", Font.PLAIN, 10);
        run.setFont(buttonFont);
        run.setToolTipText("Run Tool");
        run.addActionListener(new RunEvent());

        toolbar.add(open);
        toolbar.addSeparator();
        toolbar.add(run);
                
        return toolbar;
    }

    /**
     *  Create the menus.
     */
    private JMenuBar createMenus() {
        JMenuBar menuBar = new JMenuBar();

        // Left justified menus
        menuBar.add(createFileMenu());
        menuBar.add(createViewMenu());
        menuBar.add(createPreferencesMenu());
        menuBar.add(createRunMenu());
                
        // Right justified menus
        menuBar.add(Box.createHorizontalGlue());
        menuBar.add(createHelpMenu());
                
        return menuBar;
    }

    /**
     *  Create the menu that opens a file for the tool or closes the 
     *  program
     */
    private JMenu createFileMenu() {
        // File menu
        JMenu fileMenu = new JMenu("File");
                
        // Add button
        JMenuItem addItem = new JMenuItem("Open File...");
        addItem.addActionListener(createOpenHandlerInstance());
                
        // Exit button
        JMenuItem exitItem = new JMenuItem("Exit");
        exitItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    closeWindowOperations();
                }
            });
                
        fileMenu.add(addItem);
        fileMenu.addSeparator();
        fileMenu.add(exitItem);
                
        return fileMenu;
    }

    /**
     *  Create the View Menu
     */
    private JMenu createViewMenu() {
        // View menu
        JMenu viewMenu = new JMenu("View");
                
        // Reset preferred window sizes button
        JMenuItem resetWindowItem = 
            new JMenuItem("Reset Preferred Window Sizes");
        resetWindowItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    mainWindowPrefs.setAllPreferences();
                }
            });
        
        viewMenu.add(resetWindowItem);
        
        return viewMenu;
    }

    /**
     *  Create the menu that runs the tool
     */
    private JMenu createRunMenu() {
        // Run Menu
        JMenu runMenu = new JMenu("Run");
                
        // Run button
        JMenuItem runItem = new JMenuItem("Run");
        runItem.addActionListener(new RunEvent());
                
        runMenu.add(runItem);
        
        return runMenu;
    }

    /**
     *  Create the preferences menu
     */
    private JMenu createPreferencesMenu() {
        // Preferences menu
        JMenu preferencesMenu = new JMenu("Preferences");
                
        // Look and Feel Submenu
        JMenu lafMenu = new JMenu("Select Look and Feel");
        UIManager.LookAndFeelInfo[] lafInfo = 
            UIManager.getInstalledLookAndFeels();
        ButtonGroup group = new ButtonGroup();
                        
        // Add each look and feel as a radio button to the menu
        for (int i = 0; i < lafInfo.length; i++) {
            String lafName = lafInfo[i].getName();
            final String lafClass = lafInfo[i].getClassName();
            JMenuItem lafMenuItem = new JRadioButtonMenuItem(lafName);
            lafMenuItem.addActionListener(
                new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        setLookAndFeel(lafClass);
                    }
                });
            if (lafClass.equals(preferences.get(toolName + "LAF",
                    UIManager.getSystemLookAndFeelClassName()))) {
                lafMenuItem.setSelected(true);
            }
            group.add(lafMenuItem);
            lafMenu.add(lafMenuItem);
        }
                
        // Save preferences button
        JMenuItem savePreferences = new JMenuItem("Save Preferences");
        savePreferences.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    savePreferences();
                    ((JMenuItem)(ae.getSource())).setEnabled(false);
                }
            });
        savePreferences.setEnabled(false);
                
        // Clear preferences button
        JMenuItem clearPreferences = new JMenuItem("Clear Saved Preferences");
        clearPreferences.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    clearPreferences();
                    ((JMenuItem)(ae.getSource())).setEnabled(false);
                }
            });
        clearPreferences.setEnabled(true);
                
        preferencesMenu.add(lafMenu);
        preferencesMenu.addSeparator();
        preferencesMenu.add(savePreferences);
        preferencesMenu.add(clearPreferences);
        
        return preferencesMenu;
    }

    /**
     *  Create the help menu.
     */
    private JMenu createHelpMenu() {
        // Help menu
        JMenu helpMenu = new JMenu("Help");
                
        // Help button
        JMenuItem helpItem = new JMenuItem("Help");
        helpItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    createHelpWindow();
                }
            });
                        
        // About button
        JMenuItem aboutItem = new JMenuItem(getWebpageName());
        aboutItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    createAboutWindow();
                }
            });
                
        helpMenu.add(helpItem);
        helpMenu.addSeparator();
        helpMenu.add(aboutItem);
        
        return helpMenu;
    }
        
    /**
     *  Sets the correct look and feel on the Preferences menu when it is
     *  changed indirectly.  When the preferences are cleared, the caller
     *  gets the new look and feel and sends it to this.  This searches and
     *  selects the correct menu item.
     */
    protected void setCurrentLookAndFeelEnabled(String name) {
        MenuElement[] menuBarElements = getJMenuBar().getSubElements();
                
        Component[] menuElements = null;
        for (int i = 0; i < menuBarElements.length; i++) {
            if (menuElements == null
                && (((JMenu)(menuBarElements[i])).getText()).equals(
                    "Preferences")) {
                menuElements = 
                    ((JMenu)(menuBarElements[i])).getMenuComponents();
            }
        }

        Component[] lafMenuElements = null;
        for (int i = 0; i < menuElements.length; i++) {
            if (lafMenuElements == null
                && (((JMenu)(menuElements[i])).getText()).equals(
                    "Select Look and Feel")) {
                lafMenuElements 
                    = ((JMenu)(menuElements[i])).getMenuComponents();
            }
        }

        for (int i = 0; i < lafMenuElements.length; i++) {
            if ((((JMenuItem)(lafMenuElements[i])).getText()).equals(
                name)) {
                ((JMenuItem)(lafMenuElements[i])).setSelected(true);
            }
        }
    }
        
    /**
     *  Set the "Save Preferences" menu item to be enabled/disabled.
     *  The menu item is disabled when the preferences have been saved
     *  and there are no more preferences to save.  The menu item is
     *  restored when more preferences are not saved.
     */
    protected void setSavePrefsEnabled(boolean enabled) {
        MenuElement[] menuBarElements = getJMenuBar().getSubElements();
                
        Component[] menuElements = null;
        for (int i = 0; i < menuBarElements.length; i++) {
            if ((((JMenu)(menuBarElements[i])).getText()).equals(
                "Preferences")) {
                menuElements = 
                    ((JMenu)(menuBarElements[i])).getMenuComponents();
            }
        }

        Component menuItem = null;
        for (int i = 0; i < menuElements.length; i++) {
            if (menuElements[i] instanceof JMenuItem
                && (((JMenuItem)(menuElements[i])).getText()).equals(
                "Save Preferences")) {
                menuItem = ((JMenuItem)(menuElements[i])).getComponent();
            }
        }

        JMenuItem savePrefs = (JMenuItem)(menuItem);
        savePrefs.setEnabled(enabled);
    }

    /**
     *  Set the "Clear Preferences" menu item to be enabled/disabled.
     *  The menu item is disabled when the preferences have been cleared
     *  and there are no more preferences to clear.  The menu item is
     *  restored when more preferences are set.
     */
    protected void setClearPrefsEnabled(boolean enabled) {
        MenuElement[] menuBarElements = getJMenuBar().getSubElements();
                
        Component[] menuElements = null;
        for (int i = 0; i < menuBarElements.length; i++) {
            if ((((JMenu)(menuBarElements[i])).getText()).equals(
                "Preferences")) {
                menuElements = 
                    ((JMenu)(menuBarElements[i])).getMenuComponents();
            }
        }

        Component menuItem = null;
        for (int i = 0; i < menuElements.length; i++) {
            if (menuElements[i] instanceof JMenuItem
                && (((JMenuItem)(menuElements[i])).getText()).equals(
                "Clear Saved Preferences")) {
                menuItem = ((JMenuItem)(menuElements[i])).getComponent();
            }
        }

        JMenuItem clearPrefs = (JMenuItem)(menuItem);
        clearPrefs.setEnabled(enabled);
    }
        
    // ----------------------------------------------------------------------
    // WINDOW CREATION METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Creates the window that holds the webpage
     */
    private void createAboutWindow() {
        if (aboutWindow == null) {
            aboutWindow = new GUIAboutWindow(this, getWebpageLocation());
        } else {
            aboutWindow.setVisible(true);
        }
    }
        
    /**
     *  Creates the window that holds the option specifications
     */
    private void createHelpWindow() {
        if (helpWindow == null) {
            helpWindow = new GUIHelpWindow(this, opt);
        } else {
            helpWindow.setVisible(true);
        }
    }
        
    // ----------------------------------------------------------------------
    // PREFERENCE SYSTEM METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Sets this window's look and feel to the preferred look and feel
     */
    public void setLookAndFeel() {
        try {
            UIManager.setLookAndFeel(preferences.get(toolName + "LAF",
                UIManager.getSystemLookAndFeelClassName()));
        } catch (Exception e) {
            // !FIXME! should do something with this... (also below)
        }
        
        updateTreeUI();
    }
        
    /**
     *  Sets this window's look and feel to the specificed look and feel
     *  class
     */
    public void setLookAndFeel(String lafClass) {
        try {
            UIManager.setLookAndFeel(lafClass);
        } catch (Exception e) {}
                 
        updateTreeUI();
    }
        
    /**
     *  Allows GUI to update the look and feel of all of its components
     */
    private void updateTreeUI() {
        // Update this window
        if (getContentPane() != null) {
            SwingUtilities.updateComponentTreeUI(getContentPane());
        }
        
        // Update the menu bar
        if (getJMenuBar() != null) {
            SwingUtilities.updateComponentTreeUI(getJMenuBar());
        }
                
        // Update the option panel
        if (optionHandler != null) {
            optionHandler.updateTreeUI();
        }
                
        // Update all initialized windows
        if (aboutWindow != null) {
            aboutWindow.setLookAndFeel();
        }
                
        if (helpWindow != null) {
            helpWindow.setLookAndFeel();
        }
    }
        
    /**
     * Returns the window's native preferred size, since the elements in each 
     * window are created dynamically and cannot have a static size.
     */
    public Dimension getMinimumSize() {
        int width = (int)(super.getPreferredSize().getWidth());
        int height = MINIMUM_WINDOW_HEIGHT;
        return new Dimension(width, height);
    }
        
    /**
     *  Returns the user's preferred size for this window.
     */
    public Dimension getPreferredSize() {
        return mainWindowPrefs.getPreferredSize();
    }
        
    /**
     *  Invoked when the tool is closing.  If the options have changed, the
     *  system asks the user whether to save the current options
     */
    protected void closeWindowOperations() {
        super.closeWindowOperations();
        mainWindowPrefs.putAllPreferences();

        // If any of the options have changed from the previous settings, 
        // query the user whether to save the current options
        if (optionHandler.prefsChanged()) {
            final JOptionPane optionPane = new JOptionPane(
                "Would you like to save this tool's \n" +
                "options for your next session?",
                JOptionPane.QUESTION_MESSAGE,
                JOptionPane.YES_NO_CANCEL_OPTION);
            final JDialog dialog = new JDialog(this, "Save Options?", true);
            dialog.setContentPane(optionPane);
            dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
            optionPane.addPropertyChangeListener(
                new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent e) {
                        String prop = e.getPropertyName();
                        if (dialog.isVisible() 
                            && (e.getSource() == optionPane)
                            && (prop.equals(JOptionPane.VALUE_PROPERTY))) {
                            dialog.setVisible(false);
                            int value 
                                = ((Integer)optionPane.getValue()).intValue();
                            if (value != JOptionPane.CANCEL_OPTION) {
                                // If user answers yes to the above 
                                // question, save the options that have 
                                // changed
                                if (value == JOptionPane.YES_OPTION) {
                                    optionHandler.putOptionPreferences();
                                }
                                if (standAlone) {
                                    System.exit(0);
                                } else {
                                    dispose();
                                }
                            } 
                        }
                    }
                });

            dialog.pack();
            GUIUtil.centerWindow(dialog);
            dialog.setVisible(true);
        } else {
            if (standAlone) {
                System.exit(0);
            } else {
                dispose();
            }
        }
    }

    /**
     *  Returns the preferences for this window
     */
    protected GUIWindowPreferences getGUIWindowPreferences() {
        return mainWindowPrefs;
    }
        
    /**
     *  Override of GUIFrame's savePreferences() so the options can be saved
     */
    public void savePreferences() {
        super.savePreferences();
        mainWindowPrefs.putAllPreferences();
        optionHandler.putOptionPreferences();
    }
        
    /**
     *  Override of GUIFrame's clearPreferences() so the options can be reset
     */
    public void clearPreferences() {
        super.clearPreferences();
        mainWindowPrefs.setAllPreferences();
        optionHandler.setOptionPreferences();
                
        // Update all initialized windows
        
        if (aboutWindow != null) {
            aboutWindow.clearPreferences();
        }
                
        if (helpWindow != null) {
            helpWindow.clearPreferences();
        }
    }
        
    /**
     *  Returns whether the preferences have changed for this window
     */
    protected boolean preferencesHaveChanged() {
        return (super.preferencesHaveChanged()
            || mainWindowPrefs.preferencesHaveChanged()
            || optionHandler.prefsChanged());
    }

    // ----------------------------------------------------------------------
    // ABSTRACT METHODS
    // ----------------------------------------------------------------------
    
    /**
     *  Allows the user to stop a tool process
     */
    public abstract void interruptCompilation();

    /**
     *  Returns whether a directory is in a specific classpath
     *  in the given options
     */
    protected abstract boolean dirInClassPath(Options opt, String dir);
    
    /**
     *  Returns the classpath property from the MjcCommonOptions instance
     *  of opt
     */
    protected abstract String getClasspath(Options opt);
    
    /**
     *  Returns the sourcepath property from the MjcCommonOptions instance
     *  of opt
     */
    protected abstract String getSourcepath(Options opt);
    /**
     *  Returns an instance of OpenHandler specific to the calling
     *  subclass
     */
    protected abstract OpenHandler createOpenHandlerInstance();
        
    /**
     *  Returns an instance of Compilation specific to the calling
     *  subclass
     */
    protected abstract Compilation createCompilationInstance(
        String[] files, Options opt, OutputStream os);
        
    /**
     *  Returns an instance of an output window for the compilation output
     */
    protected abstract GUIOutputWindow createOutputWindow(InputStream is);
                
    /**
     *  Returns the webpage name specific to the calling subclass
     */
    protected abstract String getWebpageName();
        
    /**
     *  Returns the webpage location specific to the calling subclass
     */
    protected abstract String getWebpageLocation();
        
    // ----------------------------------------------------------------------
    // PROTECTED AND PRIVATE CLASSES
    // ----------------------------------------------------------------------

    /**
     *  Handles the selection of files in an open file window
     */
    protected abstract class OpenHandler extends AbstractOpenHandler {
                
        /**
         *  Constructs the handler
         */
        protected OpenHandler() {
            super();
        }
                
        /**
         *  Returns the starting directory preferred by the user
         */
        protected String getCurrentDirectory() {
            return currentDirectory;
        }
                
        /**
         *  Returns the list of files that the tool is going to process
         */
        protected JList getFileList() {
            return fileList;
        }
                
        /**
         *  Set current directory location to last location accessed
         */
        protected void setDirectory(String file) {
            currentDirectory = file;
            putCurrentDirectory();
        }
    }


    /**
     *  File filter that determines which files the tool can accept
     */
    protected abstract class GUIFileFilter extends FileFilter {

        /**
         *  Determines whether to accept a file.  Returns true if a file
         *  should be selected for processing, false if the file does not
         *  meet criteria.
         */
        public boolean accept(java.io.File f) {
            String name = f.toString();
            return (f.isFile() && hasActiveSuffix(name))
                || (f.isDirectory() && !(name.endsWith("CVS")));
        }

        /**
         *  Determines whether the name of the file has the correct suffix
         */
        public abstract boolean hasActiveSuffix(String name);

        /**
         *  Returns the description of the acceptable files.  For example,
         *  for a Java file, the words "Java File" would appear instead of
         *  ".java".
         */
        public abstract String getDescription();
    }


    /**
     *  Runs the compilation in the Main method of the tool being used
     */
    private class RunEvent implements ActionListener {

        /**
         *  Runs the compilation using the user selected files
         */
        public void actionPerformed(ActionEvent event) {
            // Recombine all extra arguments into an array
            String[] fileArray = new String[fileModel.size()];
            System.arraycopy(fileModel.toArray(), 0, fileArray, 0, 
                                         fileModel.size());

            try {
                PipedOutputStream os = new PipedOutputStream();
                PipedInputStream is = new PipedInputStream(os);
                new Thread(createOutputWindow(is)).start();
                new Thread(createCompilationInstance(
                    fileArray, opt, os)).start();
            } catch (IOException ioe) {}
        }
    }


    /**
     *  Allows the compilation process to run seperately in another thread
     */
    protected abstract class Compilation implements Runnable {

        /**
         *  Runs the compilation process
         */
        public abstract void run();
    }


    /**
     *  Handles the selection of files when the user needs to specify
     *  a specs directory
     */
    private class FindSpecsHandler {
        
        /**
         *  The "Open File" dialog window
         */
        private JFileChooser fileChooser;

        /**
         *  The options used for the tool that needs the specs directory
         */
        private Options options = opt;

        /**
         *  Provides a way for the user to select files
         */
        protected FindSpecsHandler() {
            fileChooser = new JFileChooser(currentDirectory);
            fileChooser.setDialogTitle("Select the Specs Directory");
            fileChooser.setApproveButtonText("Select");
            fileChooser.setMultiSelectionEnabled(false);
            fileChooser.setFileSelectionMode(
                JFileChooser.DIRECTORIES_ONLY);
        }
                
        /**
         *  Gives the directory selected by the user to the optionHandler
         */
        public boolean handleFileSelection(Options opt) {
            int returnVal = fileChooser.showOpenDialog(null);
            if (returnVal == JFileChooser.APPROVE_OPTION) {
                File file = fileChooser.getSelectedFile();
/* FIXME: set either sourcepath or classpath */
                return optionHandler.setOptionField(
                    "classpath", file.getAbsolutePath());
            }
            return true;
        }
    }


    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------

    /**
     *  The options for this tool
     */
    private /*@ spec_public @*/ Options opt;
        
    /**
     *  The instance of GUIOptionHandler that encompasses all of the option
     *  panel factory classes
     */
    private /*@ spec_public @*/ GUIOptionHandler optionHandler = null;
        
    /**
     *  Window for the webpage
     */
    private /*@ spec_public @*/ GUIAboutWindow aboutWindow = null;
        
    /**
     *  Window for the option specifications
     */
    private /*@ spec_public @*/ GUIHelpWindow helpWindow = null;
        
    /**
     *  Window size and location preferences for the main window.
     */
    private /*@ spec_public @*/ GUIWindowPreferences mainWindowPrefs = null;

    /**
     *  List of files to process
     */
    private /*@ spec_public @*/ JList fileList;
        
    /**
     *  Represents the contents of fileList
     */
    private /*@ spec_public @*/ DefaultListModel fileModel;
        
    /**
     *  Flag that indicates whether this GUI has been launched on its own
     *  or with the launcher and determines how the GUI will close.
     */
    private /*@ spec_public @*/ boolean standAlone = true;
        
    /**
     *  The preferred width of the window
     */
    private /*@ spec_public @*/ int MINIMUM_WINDOW_HEIGHT = 650;
}
