/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.Dimension;

import javax.swing.JEditorPane;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;

import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLFrameHyperlinkEvent;

/**
 *  The frame that displays the tool's webpage
 *@author     Kristina Boysen
 *@version    $Revision: 1.2 $
 */
public class GUIAboutWindow extends GUIFrame {

    /**
     *  Creates the frame and displays the webpage
     */
    public GUIAboutWindow(GUI gui, String webpageLocation) {
        super("About");
        this.gui = gui;
        
        aboutWindowPrefs = new GUIWindowPreferences(toolName, "About", this);

        try {
            JEditorPane p =
                new JEditorPane(webpageLocation);
            p.setEditable(false);
            p.addHyperlinkListener(new LinkListener());
            JScrollPane s = new JScrollPane(p);
            getContentPane().add(s);
        } catch (java.io.IOException e) {}
                
        aboutWindowPrefs.setAllPreferences();
        setResizable(true);
        setVisible(true);
    }
        
    /**
     *  Sets the "Set Look and Feel" button in GUI.java to the value of
     *  name
     */
    protected void setCurrentLookAndFeelEnabled(String name) {
        gui.setCurrentLookAndFeelEnabled(name);
    }
        
    /**
     *  Sets the "Save Preferences" button in GUI.java to the value of
     *  enabled
     */
    protected void setSavePrefsEnabled(boolean enabled) {
        gui.setSavePrefsEnabled(enabled);
    }
        
    /**
     *  Sets the "Clear Preferences" button in GUI.java to the value of
     *  enabled
     */
    protected void setClearPrefsEnabled(boolean enabled) {
        gui.setClearPrefsEnabled(enabled);
    }
        
    // ----------------------------------------------------------------------
    // PREFERENCE METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Sets this window's look and feel to the main window's look and feel
     */
    protected void setLookAndFeel() {
        try {
            UIManager.setLookAndFeel(UIManager.getLookAndFeel());
        } catch (Exception e) {}
                
        if (getContentPane() != null) {
            SwingUtilities.updateComponentTreeUI(getContentPane());
        }
    }
        
    /**
     * Returns the window's minimum size.
     */
    public Dimension getMinimumSize() {
        int width = 500;
        int height = 600;
        return new Dimension(width, height);
    }
        
    /**
     *  Returns the user's preferred size for this window.
     */
    public Dimension getPreferredSize() {
        return aboutWindowPrefs.getPreferredSize();
    }

    /**
     *  Saves the preferences into memory when the window closes
     */
    protected void closeWindowOperations() {
        super.closeWindowOperations();
        savePreferences();
        
        setVisible(false);
    }
        
    /**
     *  Returns the preferences for this window
     */
    protected GUIWindowPreferences getGUIWindowPreferences() {
        return aboutWindowPrefs;
    }
        
    /**
     *  Override of GUIFrame's savePreferences() so the options can be saved
     */
    protected void savePreferences() {
        super.savePreferences();
        aboutWindowPrefs.putAllPreferences();
    }
        
    /**
     *  Override of GUIFrame's clearPreferences() so the options can be reset
     */
    protected void clearPreferences() {
        super.savePreferences();
        aboutWindowPrefs.setAllPreferences();
    }
        
    /**
     *  Returns whether the preferences have changed for this window
     */
    protected boolean preferencesHaveChanged() {
        return (super.preferencesHaveChanged()
            || aboutWindowPrefs.preferencesHaveChanged());
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASSES
    // ----------------------------------------------------------------------
        
    /**
     *  Allows for the HelpViewer to load the tool's webpage
     */
    private static class LinkListener implements HyperlinkListener {

        /**
         *  Displays the webpage from the provided hyperlink
         */
        public void hyperlinkUpdate(HyperlinkEvent e) {
            if (e.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                JEditorPane pane = (JEditorPane)e.getSource();
                if (e instanceof HTMLFrameHyperlinkEvent) {
                    HTMLFrameHyperlinkEvent evt = (HTMLFrameHyperlinkEvent)e;
                    HTMLDocument doc = (HTMLDocument)pane.getDocument();
                    doc.processHTMLFrameHyperlinkEvent(evt);
                } else {
                    try {
                        pane.setPage(e.getURL());
                    } catch (Throwable t) {
                        t.printStackTrace();
                    }
                }
            }
        }
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------
    /**
     *  Instance of the GUI that called this output window so the user can
     *  set "Save Preferences" and "Clear Preferences".
     */
    private GUI gui = null;
        
    /**
     *  Window size and location preferences for the help window.
     */
    private GUIWindowPreferences aboutWindowPrefs = null;
}

