/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 *  A GUI panel that allows users to select the files for the tool to
 *  process
 *
 *@author     Kristina Boysen
 *@author     Gary T. Leavens
 *@author     David Behroozi
 *@version    $Revision: 1.4 $
 */

public class GUIFileListPanel extends JPanel {
        
    // ----------------------------------------------------------------------
    // CONSTRUCTOR
    // ----------------------------------------------------------------------

    /**
     *  Initializes all variables for panel creation
     */
    public GUIFileListPanel(JList fileList, AbstractOpenHandler handler) {
        this.fileList = fileList;
        fileModel = (DefaultListModel)(fileList.getModel());
                
        createFileListPanel(handler);
    }
        
    // ----------------------------------------------------------------------
    // FILE LIST PANEL CREATION
    // ----------------------------------------------------------------------

    /**
     *  Creates the file list panel
     */
    private void createFileListPanel(AbstractOpenHandler handler) {
        GridBagLayout gbl = new GridBagLayout();
        GridBagConstraints gbc = new GridBagConstraints();
        setLayout(gbl);
        setBorder(BorderFactory.createTitledBorder(
            "Files, Directories, and Packages"));

        // The (big) list of selected files.
        gbc.gridy = 0;
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.gridheight = GridBagConstraints.RELATIVE;
        gbc.weighty = 1;
        gbc.weightx = 1;
        gbc.fill = GridBagConstraints.BOTH;
                
        //Add "delete" mnemonic
        fileList.getInputMap().put(KeyStroke.getKeyStroke(
            KeyEvent.VK_DELETE, 0), "delete");

        // Scrollbars on the file list panel.
        JScrollPane scrollPane = new JScrollPane(fileList);
        gbl.setConstraints(scrollPane, gbc);
        add(scrollPane);
                
        // The button for adding files from the file name text box.
        JButton add = new JButton("Add Manually");
        add.setToolTipText("Type in files to add");
        gbc.gridy = 1;
        gbc.gridwidth = 1;
        gbc.gridheight = GridBagConstraints.REMAINDER;
        gbc.weightx = 0;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbl.setConstraints(add, gbc);
        add(add);

        // The file name JTextBox
        final JTextField filename = new JTextField(20);
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.weightx = 1;
        gbl.setConstraints(filename, gbc);
        add(filename);
                
        // The popup edit menu for filename
        final GUIPopupMenu popup = new GUIPopupMenu(filename);
                
        // The button for adding files through the file system browser
        JButton browse = new JButton("Open File...");
        browse.setToolTipText("Select files from the file system to add");
        gbc.weightx = 0;
        gbc.gridwidth = GridBagConstraints.RELATIVE;
        gbl.setConstraints(browse, gbc);
        add(browse);

        // The button for removing a selection from the file list,
        // which is disabled if no element is selected.
        final JButton remove = new JButton("Remove");
        remove.setToolTipText("Remove selected item from list");
        remove.setEnabled(false);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbl.setConstraints(remove, gbc);
        add(remove);

        // The fileList listener
        fileList.addListSelectionListener(
            new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting() == false) {
                        if (fileList.getSelectedIndex() == -1) {
                            remove.setEnabled(false);
                        } else {
                            remove.setEnabled(true);
                            filename.setText(
                                fileList.getSelectedValue().toString());
                        }
                    }
                }
            });
                
        // "Delete" key listener
        getActionMap().put("delete", new RemoveFile(filename, remove));
                
        // TextBox listener
        filename.addActionListener(new AddManuallyEnteredFile(filename));
        
        // Popup menu listener
        filename.addMouseListener(
            new MouseAdapter() {
                public void mousePressed(MouseEvent e) {
                    maybeShowPopup(e);
                }

                public void mouseReleased(MouseEvent e) {
                    maybeShowPopup(e);
                }

                private void maybeShowPopup(MouseEvent e) {
                    if (e.isPopupTrigger())
                        popup.show(e.getComponent(), e.getX(), e.getY());
                }
            });

        // "Add" button listener
        add.addActionListener(new AddManuallyEnteredFile(filename));
                        
        // "Browse" button listener
        browse.addActionListener(handler);

        // "Remove" button listener
        remove.addActionListener(new RemoveFile(filename, remove));
    }
        
    // ----------------------------------------------------------------------
    // WINDOW SIZES
    // ----------------------------------------------------------------------
        
    /**
     * Returns the window's minimum size
     */
    public Dimension getMinimumSize() {
        int width = (int)(super.getPreferredSize().getWidth());
        int height = MINIMUM_WINDOW_HEIGHT;
        return new Dimension(width, height);
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASSES
    // ----------------------------------------------------------------------
        
    /**
     *  Adds a file manually typed in through a JTextField to the list of files
     */
    private class AddManuallyEnteredFile implements ActionListener {
        
        /**
         *  The JTextField where the filename was typed
         */
        private JTextField filename;
                
        /**
         *  Constructor
         */
        private AddManuallyEnteredFile(JTextField filename) {
            this.filename = filename;
        }
                
        /**
         *  Adds the file to the list if it is valid and not repeated
         */
        public void actionPerformed(ActionEvent ae) {
            int index = fileList.getSelectedIndex();
            int size = fileModel.getSize();
            if (!(filename.getText()).equals("")
                && !GUIUtil.filenameAlreadyAdded(
                    fileList, filename.getText())) {
                if (index == -1 || (index + 1 == size)) {
                    fileModel.addElement(filename.getText());
                    fileList.setSelectedIndex(size);
                } else {
                    fileModel.insertElementAt(filename.getText(), index + 1);
                    fileList.setSelectedIndex(index + 1);
                }
                filename.setText("");
            }
        }
    }


    /**
     *  Removes a file from the list of files
     */
    private class RemoveFile extends AbstractAction implements ActionListener {
       
        /**
         *  The JTextField where the filename was typed
         */
        private JTextField filename;
        
        /**
         *  The "Remove" button that needs to be disabled when there are no
         *  more files to remove
         */
        private JButton remove;
        
        /**
         *  Constructor
         */
        private RemoveFile(JTextField filename, JButton remove) {
            this.filename = filename;
            this.remove = remove;
        }
                
        /**
         *  Removes the file from the list
         */
        public void actionPerformed(ActionEvent ae) {
            int[] indices = fileList.getSelectedIndices();
            int index = indices[0];
                        
            if (fileModel.get(indices[indices.length - 1]).equals(
                fileModel.lastElement())) {
                index = indices[0] - 1;
            }
                        
            fileModel.removeRange(indices[0], indices[indices.length - 1]);
                        
            if (fileModel.getSize() == 0) {
                index = -1;
                remove.setEnabled(false);
                filename.setText("");
            }
                        
            fileList.setSelectedIndex(index);
        }
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------
        
    /**
     *  List of files to process
     */
    private /*@ spec_public @*/ JList fileList;
        
    /**
     *  Represents the contents of fileList
     */
    private /*@ spec_public @*/ DefaultListModel fileModel;
        
    /**
     *  The preferred width of the window
     */
    private /*@ spec_public @*/ int MINIMUM_WINDOW_HEIGHT = 200;
}
