/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowStateListener;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.util.prefs.BackingStoreException;
import java.util.prefs.PreferenceChangeEvent;
import java.util.prefs.PreferenceChangeListener;
import java.util.prefs.Preferences;

import javax.swing.JFrame;
import javax.swing.UIManager;

/**
 *  Superclass that defines all of the GUI windows
 *
 *@author     Kristina Boysen
 *@author     Gary T. Leavens
 *@version    $Revision: 1.1 $
 */
public abstract class GUIFrame extends JFrame {
    
    /**
     *  Constructs the common attributes between windows
     */
    public GUIFrame(String toolName) {
        super(toolName);
        this.toolName = toolName;
                                        
        // Set the current preferred directory
        setCurrentDirectory();
                
        // Set the current preferred look-and-feel
        setLookAndFeel();
                
        setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
                
        // Initialize listeners

        // Listens for this window closing
        addWindowListener(
            new WindowAdapter() {
                public void windowClosing(WindowEvent we) {
                    closeWindowOperations();
                }
            });

        // Listents for a change in look and feel
        UIManager.addPropertyChangeListener(
            new PropertyChangeListener() {
                public void propertyChange(PropertyChangeEvent pce) {
                    // Sets whether the "Save Preferences" menu item is 
                    // selected, based on whether the look and feel has changed
                    setSavePrefsEnabled(
                        lookAndFeelHasChanged());

                    // Selects the current look and feel on the menu.  Used
                    // if the look and feel is changed by clearing preferences.
                    setCurrentLookAndFeelEnabled(
                        UIManager.getLookAndFeel().getName());
                }
            });
                
        // Listens for a window state change (normal and maximize)
        addWindowStateListener(
            new WindowStateListener() {
                public void windowStateChanged(WindowEvent we) {
                    // Sets whether the "Save Preferences" menu item is 
                    // selected, based on whether preferences have changed.
                    setSavePrefsEnabled(
                        getGUIWindowPreferences().preferencesHaveChanged());

                    // Sets the window location, based on where it was last
                    // saved.
                    getGUIWindowPreferences().setWindowLocation();
                }

            });

        // Listens for window resizing events
        addComponentListener(
            new ComponentAdapter() {
                public void componentResized(ComponentEvent ce) {
                    // If the window size is more than the minimum size,
                    // then the new window size is allowable
                    setSavePrefsEnabled(
                        getGUIWindowPreferences().preferencesHaveChanged());

                    // Otherwise, both sizes are too small, so set both
                    // width and height back to the minimum size.
                    if (getWidth() <= getMinimumSize().getWidth()
                        && getHeight() <= getMinimumSize().getHeight()) {
                        setSize(
                            (int)(getMinimumSize().getWidth()),
                            (int)(getMinimumSize().getHeight()));
                    // If the width is too small, set the width back to the
                    // minimum size and keep the height
                    } else if (getWidth() <= getMinimumSize().getWidth()) {
                        setSize(
                            (int)(getMinimumSize().getWidth()), getHeight());  
                    // If the height is too small, set the height back to the
                    // minimum size and keep the width
                    } else if (getHeight() <= getMinimumSize().getHeight()) {
                        setSize(
                            getWidth(), (int)(getMinimumSize().getHeight()));
                    }
                }
                                
                public void componentMoved(ComponentEvent ce) {
                    // Sets whether the "Save Preferences" menu item is 
                    // selected, based on whether the preferences have changed.
                    setSavePrefsEnabled(
                        getGUIWindowPreferences().preferencesHaveChanged());
                }
            });
        
                // Listens for changes in preferences
        preferences.addPreferenceChangeListener(
            new PreferenceChangeListener() {
                public void preferenceChange(PreferenceChangeEvent pc) {
                    // If the changed node is null, then the preferences have
                    // been cleared, so setClearPrefsEnabled() should be set
                    // to false.  Otherwise, there is a change in structure
                    // that can be cleared, so set setClearPrefsEnabled() to
                    // true;
                    setClearPrefsEnabled(pc.getNode() != null);
                }
            });
    }

    // ----------------------------------------------------------------------
    // SYSTEM PREFERENCES
    // ----------------------------------------------------------------------
        
    /**
     *  Returns the absolute maximum size for this window, which is the size
     *  of the screen.
     */
    public Dimension getMaximumSize() {
        return Toolkit.getDefaultToolkit().getScreenSize();
    }
        
    /**
     *  Sets the current directory to the user's preferred location
     */
    protected void setCurrentDirectory() {
        currentDirectory = preferences.get(toolName + "CurrentDirectory", ".");
    }
        
    /**
     *  Sets the current directory location as the user's preferred location
     */
    protected void putCurrentDirectory() {
        preferences.put(toolName + "CurrentDirectory", currentDirectory);
    }
    
    /**
     *  Abstract method to allow all windows to update their look and feel
     *  simultaniously
     */
    protected abstract void setLookAndFeel();
    
    /**
     * Sets the current look and feel as the user's preferred look and feel
     */
    protected void putLookAndFeel() {
        if (lookAndFeelHasChanged()) {
            preferences.put(toolName + "LAF", 
                UIManager.getLookAndFeel().getClass().getName());
        }
    }
    
    /**
     * Reports whether the look and feel has changed from either the
     * look and feel in preferences or the default value
     */
    protected boolean lookAndFeelHasChanged() {
        return (!(preferences.get(toolName + "LAF", 
            UIManager.getSystemLookAndFeelClassName()).equals(
                UIManager.getLookAndFeel().getClass().getName())));
    }
        
    // ----------------------------------------------------------------------
    // LOCAL PREFERENCES
    // ----------------------------------------------------------------------
        
    /**
     *  Saves the preferences into memory when the window closes
     */
    protected void closeWindowOperations() {
        putLookAndFeel();
    }
        
    /**
     *  Save all preferences into memory
     */
    protected void savePreferences() {
        putLookAndFeel();
    }

    /**
     *  Clear all user preferences
     */
    protected void clearPreferences() {
        try {
            preferences.clear();
        } catch (BackingStoreException e) {
            e.printStackTrace();
        }
                
        setLookAndFeel();
        setCurrentDirectory();
    }
        
    /**
     *  Return whether any of the system preferences have changed
     */
    protected boolean preferencesHaveChanged() {
        return lookAndFeelHasChanged();
    }

    // ----------------------------------------------------------------------
    // ABSTRACT METHODS
    // ----------------------------------------------------------------------
        
    protected abstract void setCurrentLookAndFeelEnabled(String name);
    protected abstract void setSavePrefsEnabled(boolean enabled);
    protected abstract void setClearPrefsEnabled(boolean enabled);
    protected abstract GUIWindowPreferences getGUIWindowPreferences();

    // ----------------------------------------------------------------------
    // PROTECTED AND PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------
        
    /**
     *  The user preferences for the GUI
     */
    protected Preferences preferences = 
        Preferences.userNodeForPackage(GUIFrame.class);
        
    /**
     *  Name of the tool currently being used
     */
    protected String toolName;
        
    /**
     *  Current working directory, determined by what the user has last 
     *  selected
     */
    protected String currentDirectory;
}
