/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Toolkit;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.io.File;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashSet;

import java.util.prefs.Preferences;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;

import javax.swing.table.DefaultTableModel;

import org.multijava.util.Options;

/**
 *  Constructs the options lists from the options instance and uses the
 *  OptionsFactory to construct a JPanel of selectable options.
 *
 *@author    Kristina Boysen
 *@author    Gary T. Leavens
 *@author    Yoonsik Cheon
 *@version   $Revision 1.0 $
 */
public class GUIOptionHandler implements ActionListener {

    // ----------------------------------------------------------------------
    // CONSTRUCTOR
    // ----------------------------------------------------------------------

    /**
     *  Constructs the option handler and initializes the OptionsFactory
     *  for each of the option sets.
     */
    public GUIOptionHandler(Options options, boolean useCommandLineOptions) {
        this.options = options;
        this.useCommandLineOptions = useCommandLineOptions;
        
        listeners = new ArrayList();

        optionsFactoryArray = createOptionLists();
    }
        
    // ----------------------------------------------------------------------
    // HELPER METHODS
    // ----------------------------------------------------------------------
    
    /**
     *  Creates the options array lists from options so the user can select 
     *  the desired options from check boxes, lists, or text fields.
     *  Returns the array of all of the available factories.
     */
    private ArrayList createOptionLists() {
        ArrayList optionsFactoryArray = new ArrayList();
                
        ArrayList normalList = new ArrayList();
        ArrayList advancedList = new ArrayList();
        ArrayList javadocNormalList = new ArrayList();
        ArrayList javadocNoList = new ArrayList();
        ArrayList javadocAdvancedList = new ArrayList();
                
        // Get all the hashtables that hold the various option data
        LinkedHashSet longnames = options.getLongname();
        Hashtable tableHeaders = options.getTableHeader();
        Hashtable types = options.getType();
        Hashtable defaultValues = options.getDefaultValue();
        Hashtable currentValues = options.getCurrentValue();
        Hashtable selectionList = options.getSelectionList();
        Hashtable helpStrings = options.getHelpString();
        Hashtable guiTypes = options.getGuiType();

        // Iterate through all the longnames, checking the type for each.
        Iterator iterator = longnames.iterator();
        while (iterator.hasNext()) {
            String currentName = (String)iterator.next();
            String optionType = (String)(guiTypes.get(currentName));

            if (optionType.equals("normal")) {
                // In this case, the option type is "normal", so it is 
                // normal for the user to want to select/deselect this option
                                
                // Add all attributes of this option to the row array list
                ArrayList normalOptionsRow = new ArrayList();
                normalOptionsRow.add(currentName);
                normalOptionsRow.add(tableHeaders.get(currentName));
                normalOptionsRow.add(types.get(currentName));
                normalOptionsRow.add(defaultValues.get(currentName));
                normalOptionsRow.add(currentValues.get(currentName));
                normalOptionsRow.add(selectionList.get(currentName));
                normalOptionsRow.add(helpStrings.get(currentName));
                                
                // Add the array list to the normal option iterator
                normalList.add((Object)(normalOptionsRow));
            } else if (optionType.equals("advanced")) {
                // In this case, the option type is "advanced" so it is 
                // rare for the user to want to select/deselect this option
                                
                // Add all attributes of this option to the row array list
                ArrayList advancedOptionsRow = new ArrayList();
                advancedOptionsRow.add(currentName);
                advancedOptionsRow.add(tableHeaders.get(currentName));
                advancedOptionsRow.add(types.get(currentName));
                advancedOptionsRow.add(defaultValues.get(currentName));
                advancedOptionsRow.add(currentValues.get(currentName));
                advancedOptionsRow.add(selectionList.get(currentName));
                advancedOptionsRow.add(helpStrings.get(currentName));

                // Add the array list to the advanced option iterator
                advancedList.add((Object)(advancedOptionsRow));
            } else if (optionType.equals("javadocNormal")) {
                // In this case, the option type is "javadocNormal", so it is 
                // only used for documentation tools

                // Add all attributes of this option to the row array list
                ArrayList javadocNormalOptionsRow = new ArrayList();
                javadocNormalOptionsRow.add(currentName);
                javadocNormalOptionsRow.add(tableHeaders.get(currentName));
                javadocNormalOptionsRow.add(types.get(currentName));
                javadocNormalOptionsRow.add(defaultValues.get(currentName));
                javadocNormalOptionsRow.add(currentValues.get(currentName));
                javadocNormalOptionsRow.add(selectionList.get(currentName));
                javadocNormalOptionsRow.add(helpStrings.get(currentName));
                                
                // Add the array list to the javadocNormal option iterator
                javadocNormalList.add((Object)(javadocNormalOptionsRow));
            } else if (optionType.equals("javadocNo")) {
                // In this case, the option type is "javadocNo" so it is 
                // only used for documentation tools
                                
                // Add all attributes of this option to the row array list
                ArrayList javadocNoOptionsRow = new ArrayList();
                javadocNoOptionsRow.add(currentName);
                javadocNoOptionsRow.add(tableHeaders.get(currentName));
                javadocNoOptionsRow.add(types.get(currentName));
                javadocNoOptionsRow.add(defaultValues.get(currentName));
                javadocNoOptionsRow.add(currentValues.get(currentName));
                javadocNoOptionsRow.add(selectionList.get(currentName));
                javadocNoOptionsRow.add(helpStrings.get(currentName));

                // Add the array list to the javadocNo option iterator
                javadocNoList.add((Object)(javadocNoOptionsRow));
            } else if (optionType.equals("javadocAdvanced")) {
                // In this case, the option type is "javadocAdvanced" so it is 
                // only used for documentation tools
                                
                // Add all attributes of this option to the row array list
                ArrayList javadocAdvancedOptionsRow = new ArrayList();
                javadocAdvancedOptionsRow.add(currentName);
                javadocAdvancedOptionsRow.add(tableHeaders.get(currentName));
                javadocAdvancedOptionsRow.add(types.get(currentName));
                javadocAdvancedOptionsRow.add(defaultValues.get(currentName));
                javadocAdvancedOptionsRow.add(currentValues.get(currentName));
                javadocAdvancedOptionsRow.add(selectionList.get(currentName));
                javadocAdvancedOptionsRow.add(helpStrings.get(currentName));

                // Add the array list to the javadocAdvanced option iterator
                javadocAdvancedList.add((Object)(javadocAdvancedOptionsRow));
            } else {
                // In this case, the option type is not specified, so it
                // is not added to the options
            }
        }
                
        // Initialize the factories if there are any elements in the iterators
        if (!normalList.isEmpty()) {
            normalOF = new OptionsFactory(normalList);
            normalOF.addActionListener(this);
            optionsFactoryArray.add(normalOF);
        }
        
        if (!advancedList.isEmpty()) {
            advancedOF = new OptionsFactory(advancedList);
            advancedOF.addActionListener(this);
            optionsFactoryArray.add(advancedOF);
        }
        
        if (!javadocNormalList.isEmpty()) {
            javadocNormalOF = new OptionsFactory(javadocNormalList);
            javadocNormalOF.addActionListener(this);
            optionsFactoryArray.add(javadocNormalOF);
        }
        
        if (!javadocNoList.isEmpty()) {
            javadocNoOF = new OptionsFactory(javadocNoList);
            javadocNoOF.addActionListener(this);
            optionsFactoryArray.add(javadocNoOF);
        }
        
        if (!javadocAdvancedList.isEmpty()) {
            javadocAdvancedOF = new OptionsFactory(javadocAdvancedList);
            javadocAdvancedOF.addActionListener(this);
            optionsFactoryArray.add(javadocAdvancedOF);
        }
        
        return optionsFactoryArray;
    }
        
    // ----------------------------------------------------------------------
    // ACCESSORS
    // ----------------------------------------------------------------------
        
    /**
     *  Returns the instance of the normal options OptionFactory.
     */
    public JScrollPane getNormalOptionPanel() {
        if (normalOF != null) {
            return new JScrollPane(normalOF, 
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        } else {
            return null;
        }
    }
        
    /**
     *  Returns the instance of the advanced options OptionFactory.
     */
    public JScrollPane getAdvancedOptionPanel() {
        if (advancedOF != null) {
            return new JScrollPane(advancedOF,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        } else {
            return null;
        }
    }
        
    /**
     *  Returns the instance of the javadocNormal options OptionFactory.
     */
    public JScrollPane getJavadocNormalOptionPanel() {
        if (javadocNormalOF != null) {
            return new JScrollPane(javadocNormalOF,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        } else {
            return null;
        }
    }
        
    /**
     *  Returns the instance of the javadocNo options OptionFactory.
     */
    public JScrollPane getJavadocNoOptionPanel() {
        if (javadocNoOF != null) {
            return new JScrollPane(javadocNoOF,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        } else {
            return null;
        }
    }
        
    /**
     *  Returns the instance of the javadocAdvanced options OptionFactory.
     */
    public JScrollPane getJavadocAdvancedOptionPanel() {
        if (javadocAdvancedOF != null) {
            return new JScrollPane(javadocAdvancedOF,
                ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        } else {
            return null;
        }
    }
        
    /**
     *  Checks through each option set to see if either of them can be
     *  set to value.  If the option is found, then this returns true.
     *  Otherwise, it returns false. 
     */
    public boolean setOptionField(String longname, Object value) {
        for (int i = 0; i < optionsFactoryArray.size(); i++) {
            if (((OptionsFactory)(optionsFactoryArray.get(i)))
                .setOptionField(longname, value)) {
                return true;
            }
        }
        return false;
    }
        
    /**
     *  Sets the options to the current user preferences
     */
    public void setOptionPreferences() {
        for (int i = 0; i < optionsFactoryArray.size(); i++) {
            ((OptionsFactory)(optionsFactoryArray.get(i)))
                .setOptionPreferences();
        }
    }
    
    /**
     *  Sets the user preferences to the current options settings
     */
    public void putOptionPreferences() {
        for (int i = 0; i < optionsFactoryArray.size(); i++) {
            ((OptionsFactory)(optionsFactoryArray.get(i)))
                 .putOptionPreferences();
        }
    }
        
    /**
     *  Determines whether any of the normal options have changed
     */
    public boolean prefsChanged() {
        for (int i = 0; i < optionsFactoryArray.size(); i++) {
            if (((OptionsFactory)(optionsFactoryArray.get(i)))
                .prefsChanged()) {
                return true;
            }
        }
        return false;
    }
        
    /**
     *  Updates the components in each factory method when the preferences
     *  are changed
     */
    public void updateTreeUI() {
        for (int i = 0; i < optionsFactoryArray.size(); i++) {
            SwingUtilities.updateComponentTreeUI(
                (JPanel)(optionsFactoryArray.get(i)));
        }
    }
    
    /**
     *  Adds the specified action listener to the list of listeners
     */ 
    public void addActionListener(ActionListener al) {
        listeners.add(al);
    }
        
    /**
     *  Notifies all of this object's listeners that an event has occured
     */  
    protected void fireActionPerformed(ActionEvent ae) {
        for (int i = 0; i < listeners.size(); i++) {
            ((ActionListener)(listeners.get(i))).actionPerformed(ae);
        }
    }
        
    /**
     *  Invoked when an object has performed an action
     */ 
    public void actionPerformed(ActionEvent ae) {
        fireActionPerformed(ae);
    }

    // ----------------------------------------------------------------------
    // INTERFACE
    // ----------------------------------------------------------------------

    /**
     *  Provides a template for all of the different kinds of GUI options
     */
    public interface OptionsCommand {
        public GridBagLayout gridbaglayout = new GridBagLayout();
        public GridBagConstraints constraints = new GridBagConstraints();
        public Insets noInsets = new Insets(0, 0, 0, 0);
        public Insets rightInsets = new Insets(0, 0, 0, 50);
        
        /**
         *  Abstract method where implementing classes should create and add
         *  their object to the GUI
         */
        public abstract void createObject(JPanel optionPanel);
        
        /**
         *  Abstract method where implementing classes should set the default
         *  value for their specific option
         */
        public abstract void setDefault();
                
        /**
         *  Abstract method where implementing classes should set the value
         *  for the specific option and update the visual component
         */
        public abstract boolean setOptionField(String longname, Object value);
                
        /**
         *  Abstract method where implementing classes should set the value
         *  value for their specific option, only if the user requests it.
         */
        public abstract void setOptionPreferences();
                
        /**
         *  Abstract method where implementing classes should put the current
         *  value into the preferences instance for their specific option,
         *  only if the user requests it.
         */
        public abstract void putOptionPreferences();
        
        /**
         *  Abstract method where implementing classes should determine
         *  whether their preferences are different from the default settings
         *  or from user preferences
         */
        public abstract boolean prefsChanged();
    }


    // ----------------------------------------------------------------------
    // PRIVATE CLASSES
    // ----------------------------------------------------------------------

    /**
     *  Creates the GUI from the given option array list
     */
    private class OptionsFactory extends JPanel implements OptionsCommand, 
                                                           ActionListener {
        private Hashtable option = new Hashtable();
        private ArrayList optionList;
        private ArrayList listeners;
                
        /**
         *  Initializes the iterator and extra options, and sets the layout
         */
        public OptionsFactory(ArrayList optionList) {
            this.optionList = optionList;
            listeners = new ArrayList();
            
            setLayout(new BorderLayout());
                        
            JPanel optionPanel = new JPanel();
                        
            constraints.fill = GridBagConstraints.HORIZONTAL;
            constraints.anchor = GridBagConstraints.WEST;
            constraints.weighty = 1.0;
            optionPanel.setLayout(gridbaglayout);
                        
            createObject(optionPanel);
        }

        /**
         *  Creates the instances of the GUI creator classes and adds them
         *  to the layout
         */
        public void createObject(JPanel optionPanel) {
            String longname;
            String[] tableHeader;
            String type;
            Object defaultValue;
            Object currentValue;
            ArrayList selection;
            String help;
            
            for (int i = 0; i < optionList.size(); i++) {
                ArrayList currentList = (ArrayList)(optionList.get(i));
                                
                // Recreates the option attributes from the given array list
                                
                // The longname is the full name of this option
                longname = currentList.get(0).toString();
                                
                // The tableHeader is the headers for any option represented
                // by a table
                tableHeader = (String[])currentList.get(1);
                                
                // The type is what kind of option this is.  The type can
                // be either an boolean, String, or int.  All other types
                // are ignored.
                type = currentList.get(2).toString();
                                
                // The default value is the value that the option is set
                // to in the .opt file
                defaultValue = currentList.get(3);
                                
                // The current value is the value that the option is set
                // to currently.
                currentValue = currentList.get(4);
                                
                // The selection is an array of different close-ended
                // choices for this option
                selection = (ArrayList)(currentList.get(5));
                                
                // Help is the string that tells the user what this option
                // does.
                help = currentList.get(6).toString();
                
                if (type.equals("boolean")) {
                                        
                    // In this case, the option is a boolean, so a true/false
                    // check box makes the most sense.
                    GUICheckBox gcb = new GUICheckBox(
                        longname,
                        ((Boolean)defaultValue).booleanValue(),
                        ((Boolean)currentValue).booleanValue(),
                        help, i);
                    gcb.createObject(optionPanel);
                    gcb.addActionListener(this);
                    option.put(longname, gcb);

                } else if (type.equals("String") || type.equals("int")) {
                                        
                    // In this case, the option is a String or an int, 
                    // so there has to be room for more than two option
                    // like the GUICheckBox took care of
                    if (selection != null) {

                        // In this case, the ArrayList selection is not null, 
                        // so a selection list exists; thus we use a combo box
                        GUIComboBox gcb = new GUIComboBox(
                            longname, defaultValue.toString(),
                            currentValue.toString(), selection, help, i);
                        gcb.createObject(optionPanel);
                        gcb.addActionListener(this);
                        option.put(longname, gcb);
                    } else {
                        // In this case, the ArrayList selection is null, so a
                        // selection list does not exist; thus we use a text 
                        // field
                        GUITextField gtf = new GUITextField(
                            longname, defaultValue.toString(),
                            currentValue.toString(), help, i);
                        gtf.createObject(optionPanel);
                        gtf.addActionListener(this);
                        option.put(longname, gtf);
                    }
                } else if (type.equals("StringList")) {
                    // In this case, the option is a list with one or
                    // two dimensions, so a table makes the most sense.
                    GUITableDialog gst = new GUITableDialog(
                        longname, tableHeader, (String[][])defaultValue,
                        (String[][])currentValue, help, i);
                    gst.createObject(optionPanel);
                    gst.addActionListener(this);
                    option.put(longname, gst);
                } else {
                    // Type is an optional function in the Optgen.g grammar,
                    // so it is assumed that if no type is listed, then there
                    // exists another option that is the reverse of that
                    // option.  For example, multijava is a boolean option
                    // and nomultijava (which has no type) sets multijava to 
                    // false.  The option nomultijava will never be added
                    // to the GUI since this is redundant.
                }
            }
            
            add(optionPanel, BorderLayout.NORTH);
                        
            // Create "Restore Defaults" button
            JButton restoreDefault 
                = new JButton("Restore Tool Option Defaults");
            restoreDefault.addActionListener(
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        setDefault();
                    }
                });
            
            JPanel buttonPanel = new JPanel(new FlowLayout());
            buttonPanel.add(restoreDefault, BorderLayout.SOUTH);
            
            add(buttonPanel);
        }

        /**
         *  Overrides JFrame's getMaximumSize to redefine the maximum size
         *  of the JFrame.  The minimum size is the window's smallest packed
         *  size (so that all components are shown), so we just get the
         *  preferred size from JFrame here.
         */
        public Dimension getMinimumSize() {
            int width = (int)(super.getPreferredSize().getWidth());
            int height = MINIMUM_HEIGHT;
            return new Dimension(width, height);
        }

        /**
         *  Overrides JFrame's getPreferredSize to redefine the preferred
         *  size of the JFrame.  The width is restricted to a set number,
         *  while the height is the preferred height of each component.
         */
        public Dimension getPreferredSize() {
            int width = PREFERRED_WIDTH;
            int height = (int)(super.getPreferredSize().getHeight());
            return new Dimension(width, height);
        }
                
        /**
         *  Overrides JFrame's getMaximumSize to redefine the maximum size
         *  of the JFrame.  The maximum size is the size of the screen.
         */
        public Dimension getMaximumSize() {
            Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
            int width = (int)(screen.getWidth());
            int height = 300; //(int)(screen.getHeight());
            return new Dimension(width, height);
        }
                
        /**
         *  Sets the defaults for all of the components in this option set
         */
        public void setDefault() {
            for (Enumeration e = option.elements(); e.hasMoreElements();) {
                ((OptionsCommand)(e.nextElement())).setDefault();
            }
        }
                
        /**
         *  Finds the option that needs to be changed and calls
         *  setOptionField() on that option
         */
        public boolean setOptionField(String longname, Object value) {
            return ((OptionsCommand)
                    (option.get(longname))).setOptionField(longname, value);
        }
                
        /**
         *  Sets the user's preferred value for all the options in this
         *  option set
         */
        public void setOptionPreferences() {
            for (Enumeration e = option.elements(); e.hasMoreElements();) {
                ((OptionsCommand)(e.nextElement())).setOptionPreferences();
            }
        }
                
        /**
         *  Puts the current value into the preferences instance for all the 
         *  options in this option set.
         */
        public void putOptionPreferences() {
            for (Enumeration e = option.elements(); e.hasMoreElements();) {
                ((OptionsCommand)(e.nextElement())).putOptionPreferences();
            }
        }
                
        /**
         *  Determines whether any of the preferences have changed for any
         *  of the options in this option set.
         */
        public boolean prefsChanged() {
            for (Enumeration e = option.elements(); e.hasMoreElements();) {
                if (((OptionsCommand)(e.nextElement())).prefsChanged()) {
                    return true;
                }
            }
            return false;
        }
        
        /**
         *  Adds the specified action listener to the list of listeners
         */
        public void addActionListener(ActionListener al) {
            listeners.add(al);
        }
        
        /**
         *  Notifies all of the listener that an action has taken place
         */
        protected void fireActionPerformed(ActionEvent ae) {
            for (int i = 0; i < listeners.size(); i++) {
                ((ActionListener)(listeners.get(i))).actionPerformed(ae);
            }
        }
        
        /**
         *  Invoked when an object has performed an action 
         */
        public void actionPerformed(ActionEvent ae) {
            fireActionPerformed(ae);
        }
    }


    /**
     * Creates the option object for true/false option
     */
    private class GUICheckBox implements OptionsCommand {
        private JCheckBox box;

        private String longname;
        private boolean defaultValue;
        private boolean currentValue;
        private boolean prefValue;
        private String help;
        private int row;
                
        private ArrayList listeners;
                
        /**
         *  Initializes the variables for the class
         */
        public GUICheckBox(String lname, boolean dval, 
                           boolean cval, String h, int r) {
            longname = lname;
            defaultValue = dval;
            currentValue = cval;
            help = h;
            row = r;
                        
            listeners = new ArrayList();
        }
                
        /**
         *  Adds the checkBox to the layout
         */
        public void createObject(JPanel optionPanel) {
            prefValue = preferences.getBoolean(longname, defaultValue);
                        
            // If we are using command line options and the value currently
            // loaded into the options set by the options parser is different
            // from BOTH the default value and the preferred value, then
            // the current value should be loaded.  
            if (useCommandLineOptions && (currentValue != defaultValue)
                                      && (currentValue != prefValue)) {
                box = new JCheckBox(longname, currentValue);
            // If the current value is the same as the default value 
            // or the preferred value, then one of those should be loaded 
            // (the default value will be loaded if there are no 
            // preferences to load).
            } else {
                box = new JCheckBox(longname, prefValue);
                options.setOption(longname, 
                                  Boolean.valueOf(box.isSelected()));
            }
                        
            box.setToolTipText(help);

            box.addActionListener(
                new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        Boolean newValue = Boolean.valueOf(box.isSelected());
                        options.setOption(longname, newValue);
                        fireActionPerformed();
                    }
                });

            constraints.gridx = 0;
            constraints.gridy = row;
            constraints.insets = noInsets;
            constraints.weightx = 0.1;
            gridbaglayout.setConstraints(box, constraints);
            optionPanel.add(box);
        }
                
        /**
         *  Sets the default variable for this checkBox
         */
        public void setDefault() {
            box.setSelected(defaultValue);
        }
                
        /**
         *  Sets the field for this checkbox to value
         */
        public boolean setOptionField(String longname, Object value) {
            box.setSelected(((Boolean)value).booleanValue());
            return true;
        }
                
        /**
         *  Sets the user's preferred value for this checkBox.
         *  If there is no user value, sets the checkBox to the default.
         */
        public void setOptionPreferences() {
            prefValue = preferences.getBoolean(longname, defaultValue);
            box.setSelected(prefValue);
        }
                
        /**
         *  Puts the current value into the preferences instance
         */
        public void putOptionPreferences() {
            if (prefsChanged()) {
                preferences.putBoolean(longname, box.isSelected());
                prefValue = preferences.getBoolean(longname, defaultValue);
            }
        }
                
        /**
         *  Returns whether the current value is different from the 
                 *  preferences set before.
         */
        public boolean prefsChanged() {
            return (!(box.isSelected()) == prefValue);
        }
        
        /**
         *  Adds the specified action listener to the list of listeners
         */
        public void addActionListener(ActionListener al) {
            listeners.add(al);
        }
        
        /**
         *  Creates a new action event and calls the
         *  fireActionPerformed(ActionEvent ae) method below
         */     
        public void fireActionPerformed() {
            fireActionPerformed(new ActionEvent(this, 
            ActionEvent.ACTION_PERFORMED, longname));
        }
        
        /**
         * Notifies the listeners that an action has been performed
         */
        protected void fireActionPerformed(ActionEvent ae) {
            for (int i = 0; i < listeners.size(); i++) {
                ((ActionListener)(listeners.get(i))).actionPerformed(ae);
            }
        }
    }


    /**
     *  Defines the commonalities between GUIComboBox and GUITextField
     */
    private abstract class GUISelection implements OptionsCommand {
        protected String longname;
        protected String defaultValue;
        protected String currentValue;
        protected String prefValue;
        protected String help;
        protected int row;
                
        protected ArrayList listeners;

        /**
         *  Adds the label of the object to the layout
         */
        public void createObject(JPanel optionPanel) {
            JLabel label = new JLabel(longname);
            label.setToolTipText(help);
                        
            constraints.gridx = 0;
            constraints.gridy = row;
            constraints.insets = noInsets;
            constraints.weightx = 0.1;
            gridbaglayout.setConstraints(label, constraints);
            optionPanel.add(label);
        }
                
        /**
         *  Abstract method for setDefault()
         */
        public abstract void setDefault();
                
        /**
         *  Abstract method for setOptionField()
         */
        public abstract boolean setOptionField(String longname, Object value);
                
        /**
         *  Abstract method for setOptionPreference()
         */
        public abstract void setOptionPreferences();
                
        /**
         *  Abstract method for putOptionPreference()
         */
        public abstract void putOptionPreferences();
                
        /**
         *  Abstract method for prefsChanged()
         */
        public abstract boolean prefsChanged();
        
        /**
         *  Add the specified action listener to the set of listeners
         */
        public void addActionListener(ActionListener al) {
            listeners.add(al);
        }
        
        /**
         *  Creates a new ActionEvent and calls the
         *  fireActionPerformed(ActionEvent ae) method
         */
        public void fireActionPerformed() {
            if (prefsChanged()) {
                fireActionPerformed(new ActionEvent(this, 
                    ActionEvent.ACTION_PERFORMED, longname));
            }
        }
        
        /**
         *  Notifies all the listeners that an action has been performed
         */
        protected void fireActionPerformed(ActionEvent ae) {
            for (int i = 0; i < listeners.size(); i++) {
                ((ActionListener)(listeners.get(i))).actionPerformed(ae);
            }
        }
    }


    /**
     *  Creates the option object for option with several set selections
     */
    private class GUIComboBox extends GUISelection implements OptionsCommand {
        private JComboBox comboBox;
                
        private int index;
        private ArrayList selection;
                
        /**
         *  Initializes the variables for the class
         */
        public GUIComboBox(String lname, String dval, String cval,
                           ArrayList sel, String h, int r) {
            longname = lname;
            defaultValue = dval;
            currentValue = cval;
            selection = sel;
            help = h;
            row = r;
                        
            listeners = new ArrayList();
        }
                
        /**
         *  Adds the comboBox to the layout
         */
        public void createObject(JPanel optionPanel) {
            super.createObject(optionPanel);

            Object[] selectionData = ((ArrayList)(selection)).toArray();
            comboBox = new JComboBox(selectionData);
            comboBox.setToolTipText(help);
                        
            prefValue = preferences.get(longname, defaultValue);
            
            // If we are using command line options and the value currently
            // loaded into the options set by the options parser is different
            // from BOTH the default value and the preferred value, then
            // the current value should be loaded.  
            if (useCommandLineOptions && !currentValue.equals(defaultValue)
                                      && !currentValue.equals(prefValue)) {
                index = getIndex(comboBox, currentValue);
                comboBox.setSelectedIndex(index != -1 ? index : 0);
            // If the current value is the same as the default value 
            // or the preferred value, then one of those should be loaded 
            // (the default value will be loaded if there are no preferences 
            // to load).
            } else {
                index = getIndex(comboBox, prefValue);
                comboBox.setSelectedIndex(index != -1 ? index : 0);
                options.setOption(longname, comboBox.getSelectedItem());
            }

            comboBox.addActionListener(
                new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        Object newValue = comboBox.getSelectedItem();
                        options.setOption(longname, newValue);
                        fireActionPerformed();
                    }
                });
            
            constraints.gridx = GridBagConstraints.RELATIVE;
            constraints.gridy = row;
            constraints.insets = rightInsets;
            constraints.weightx = 1.0;
            gridbaglayout.setConstraints(comboBox, constraints);
            optionPanel.add(comboBox);
        }
        
        /**
         *  Sets the default variable for this comboBox
         */
        public void setDefault() {
            index = getIndex(comboBox, defaultValue);
            comboBox.setSelectedIndex(index != -1 ? index : 0);
        }
                
        /**
         *  Sets the selected value for this combo box to value
         */
        public boolean setOptionField(String longname, Object value) {
            index = getIndex(comboBox, (String)value);
            comboBox.setSelectedIndex(index != -1 ? index : 0);
            return true;
        }
                
        /**
         *  Sets the user's preferred value for this comboBox.
         *  If there is no user value, sets the comboBox to the default.
         */
        public void setOptionPreferences() {
            prefValue = preferences.get(longname, defaultValue);
            index = getIndex(comboBox, prefValue);
            comboBox.setSelectedIndex(index != -1 ? index : 0);
        }
                
        /**
         *  Sets the user's preferred value for this comboBox
         */
        public void putOptionPreferences() {
            if (prefsChanged()) {
                preferences.put(longname, 
                    (String)(comboBox.getSelectedItem()));
                prefValue = preferences.get(longname, defaultValue);
            }
        }
                
        /**
         *  Returns whether the current value is different from the 
         *  preferences set before.
         */
        public boolean prefsChanged() {
            return (!((String)(comboBox.getSelectedItem())).equals(prefValue));
        }

        /**
         *  Determines the default value in the comboBox based on an 
         *  input string.
         */
        private int getIndex(JComboBox box, String value) {
            for (int i = 0; i < box.getItemCount(); i++) {
                if (value.equals((String)box.getItemAt(i))) {
                    return i;
                }
            }
            return -1;
        }
    }


    /**
     *  Creates the option object for options with open-ended responces
     */
    private class GUITextField extends GUISelection implements OptionsCommand {
        private JTextField field;
                
        /**
         *  Initializes the variables for the class
         */
        public GUITextField(String lname, String dval, String cval,
                            String h, int r) {
            longname = lname;
            defaultValue = dval;
            currentValue = cval;
            help = h;
            row = r;
                        
            listeners = new ArrayList();
        }
                
        /**
         *  Adds the textField to the layout
         */
        public void createObject(JPanel optionPanel) {
            super.createObject(optionPanel);
                        
            prefValue = preferences.get(longname, defaultValue);
                        
            // If we are using command line options and the value currently
            // loaded into the options set by the options parser is different
            // from BOTH the default value and the preferred value, then
            // the current value should be loaded.
            if (useCommandLineOptions && !currentValue.equals(defaultValue)
                                      && !currentValue.equals(prefValue)) {
                field = new JTextField(currentValue);
            // If the current value is the same as the default value 
            // or the preferred value, then one of those should be loaded 
            // (the default value will be loaded if there are no preferences 
            // to load).
            } else {
                field = new JTextField(prefValue);
                options.setOption(longname, field.getText());
            }
            
            field.setToolTipText(help);
                        
            field.addFocusListener(
                new FocusListener() {
                    public void focusGained(FocusEvent g) {
                    }
                                        
                    public void focusLost(FocusEvent l) {
                        Object newValue = field.getText();
                        options.setOption(longname, newValue);
                        fireActionPerformed();
                    }
                });

            final GUIPopupMenu popup = new GUIPopupMenu(field);
            field.addMouseListener(new MouseAdapter() {
                    public void mousePressed(MouseEvent e) {
                        maybeShowPopup(e);
                    }

                    public void mouseReleased(MouseEvent e) {
                        maybeShowPopup(e);
                    }
            
                    private void maybeShowPopup(MouseEvent e) {
                        if (e.isPopupTrigger())
                            popup.show(e.getComponent(), e.getX(), e.getY());
                    }
                });


            constraints.gridx = GridBagConstraints.RELATIVE;
            constraints.gridy = row;
            constraints.insets = rightInsets;
            constraints.weightx = 1.0;
            gridbaglayout.setConstraints(field, constraints);
            optionPanel.add(field);
        }

        /**
         *  Sets the default variable for this textField
         */
        public void setDefault() {
            field.setText(defaultValue);
            field.dispatchEvent(new FocusEvent(
                                    field, FocusEvent.FOCUS_LOST, true));
        }
                
        /**
         *  Sets the selected value for this text field to value
         */
        public boolean setOptionField(String longname, Object value) {
            field.setText(field.getText() + File.pathSeparator 
                                          + (String)value);
            field.dispatchEvent(new FocusEvent(
                                    field, FocusEvent.FOCUS_LOST, true));
            return true;
        }
                
        /**
         *  Sets the user's preferred value for this textField.
         *  If there is no user value, sets the textField to the default.
         */
        public void setOptionPreferences() {
            prefValue = preferences.get(longname, defaultValue);
            field.setText(prefValue);
            field.dispatchEvent(new FocusEvent(
                                    field, FocusEvent.FOCUS_LOST, true));
        }
                
        /**
         *  Sets the user's preferred value for this textField
         */
        public void putOptionPreferences() {
            if (prefsChanged()) {
                preferences.put(longname, field.getText());
                prefValue = preferences.get(longname, defaultValue);
            }
        }

        /**
         *  Returns whether the current value is different from the 
         *  preferences set before.
         */
        public boolean prefsChanged() {
            return (!(field.getText()).equals(prefValue));
        }
    }
    
        
    /**
     *  Creates the options object for any one or two dimension
     *  lists
     */
    private class GUITableDialog implements OptionsCommand {
        private JTable table;
        private DefaultTableModel dtm;
        
        private String longname;
        private String[] tableHeader;
        private String defaultValue;
        private String currentValue;
        private String prefValue;
        private String help;
        private int row;
                
        private ArrayList listeners;
                
        /**
         *  Initializes all variables
         */
        public GUITableDialog(String lname, String[] tableHeader, 
                              String[][] dval, String[][] cval, 
                              String h, int r) {
            this.longname = lname;
            this.tableHeader = tableHeader;
            this.defaultValue = arrayToString(dval);
            this.currentValue = arrayToString(cval);
            this.help = h;
            this.row = r;
            
            listeners = new ArrayList();
        }
                
        /**
         *  Adds the table to the layout
         */
        public void createObject(JPanel optionPanel) {
            JButton button = new JButton(longname);
            button.setToolTipText(help);
            button.addActionListener(
                new ActionListener() {
                    public void actionPerformed(ActionEvent ae) {
                        initDialog();
                    }
                });
                        
            constraints.gridx = 0;
            constraints.gridy = row;
            constraints.insets = noInsets;
            constraints.weightx = 0.1;
            gridbaglayout.setConstraints(button, constraints);
            optionPanel.add(button);
                        
            prefValue = preferences.get(longname, defaultValue);
            // If we are using command line options and the extra option taken
            // from the command line is different from BOTH the default value 
            // and the preferred value, then the current value should be loaded.
            initTable();
            if (useCommandLineOptions 
                && !currentValue.equals(defaultValue)
                && !currentValue.equals(prefValue)) {
                setTableData(currentValue);
            // If the current value is the same as the default value 
            // or the preferred value, then one of those should be loaded 
            // (the default value will be loaded if there are no 
            // preferences to load).
            } else {
                setTableData(prefValue);
                options.setOption(longname, getTableData());
            }
        }
                
        /**
         *  Builds the dialog for displaying this table
         */
        private void initDialog() {
            final JOptionPane pane = new JOptionPane(new JScrollPane(table), 
                JOptionPane.PLAIN_MESSAGE, JOptionPane.OK_CANCEL_OPTION,
                null, null, null);
            final JDialog dialog = new JDialog();
            dialog.setContentPane(pane);
            dialog.setTitle(longname);
            dialog.setModal(true);
            dialog.pack();
            pane.addPropertyChangeListener(
                new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent e) {
                        String prop = e.getPropertyName();
                        if (dialog.isVisible() 
                            && (e.getSource() == pane)
                            && (prop.equals(JOptionPane.VALUE_PROPERTY))) {
                            dialog.setVisible(false);
                            int value 
                                = ((Integer)pane.getValue()).intValue();
                            if (value == JOptionPane.OK_OPTION) {
                                if (table.isEditing()) {
                                    table.getCellEditor().stopCellEditing();
                                }
                                options.setOption(longname, getTableData());
                                fireActionPerformed();
                            }
                        }
                    }
                });
            GUIUtil.centerWindow(dialog);
            dialog.setVisible(true);
        }

        /**
         * Initialize the table and table model so that the table options can be
         * reloaded without interference from the current contents
         */
        private void initTable() {
            dtm = new DefaultTableModel(tableHeader, 10);
            table = new JTable(dtm);
            table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        }

        /**
         *  Sets the default variable for this table
         */
        public void setDefault() {
            initTable();
            setTableData(defaultValue);
        }
                
        /**
         *  Sets the selected value for this table to value
         */
        public boolean setOptionField(String longname, Object value) {
            initTable();
            setTableData((String[][])value);
            return true;
        }
                
        /**
         *  Sets the user's preferred value for this table.
         *  If there is no user value, sets the table to the default.
         */
        public void setOptionPreferences() {
            initTable();
            prefValue = preferences.get(longname, defaultValue);
            setTableData(prefValue);
        }
                
        /**
         *  Sets the user's preferred value for this table
         */
        public void putOptionPreferences() {
            if (prefsChanged()) {
                preferences.put(longname, arrayToString(getTableData()));
                prefValue = preferences.get(longname, defaultValue);
            }
        }

        /**
         *  Returns whether the current value is different from the 
         *  preferences set before.
         */
        public boolean prefsChanged() {
            return !(arrayToString(getTableData())).equals(prefValue);
        }
                
        /**
         *  Invoked when an object has performed an action 
         */
        public void actionPerformed(ActionEvent ae) {
            fireActionPerformed(ae);
        }
                
        /**
         *  Add the specified action listener to the set of listeners
         */
        public void addActionListener(ActionListener al) {
            listeners.add(al);
        }
        
        /**
         *  Creates a new ActionEvent and calls the
         *  fireActionPerformed(ActionEvent ae) method
         */
        public void fireActionPerformed() {
            if (prefsChanged()) {
                fireActionPerformed(new ActionEvent(this, 
                    ActionEvent.ACTION_PERFORMED, longname));
            }
        }
        
        /**
         *  Notifies all the listeners that an action has been performed
         */
        protected void fireActionPerformed(ActionEvent ae) {
            for (int i = 0; i < listeners.size(); i++) {
                ((ActionListener)(listeners.get(i))).actionPerformed(ae);
            }
        }
                
        /**
         *  Sets the table value to the data array
         */
        private void setTableData(String[][] data) {
            for (int i = 0; i < data.length; i++) {
                for (int j = 0; j < data[i].length; j++) {
                    dtm.setValueAt(data[i][j], i, j);
                }
            }
        }
        
        /**
         *  Sets the table value to the string
         */
        private void setTableData(String string) {
            if (string == null || string.equals("")) return;
            int index = 0; // Initialize index
            int i = 0;
            while (index < string.length() && string.charAt(index) == '{') {
                index++;  // eat {
                int j = 0;
                while (index < string.length() 
                    && string.charAt(index) != '}') {
                    while (index < string.length() 
                        && string.charAt(index) == '"') {
                        index++;  // eat "
                        StringBuffer ele = new StringBuffer();
                        while (index < string.length() 
                            && string.charAt(index) != '"') {
                            if (string.charAt(index) == '\\') {
                                index++;  // eat \
                            }
                            ele.append(string.charAt(index));
                            index++;  // eat char
                        }
                        dtm.setValueAt(ele.toString(), i, j);
                        index += 3;  // eat ", and the space
                        j++;
                    }
                    index += 3;  // eat }, and the space
                }
                i++;
            }
        }
        
        /**
         *  Returns the current data in the table
         */
        private String[][] getTableData() {
            String[][] data 
                = new String[dtm.getRowCount()][dtm.getColumnCount()];
            for (int i = 0; i < data.length; i++) {
                for (int j = 0; j < data[i].length; j++) {
                    data[i][j] = (String)dtm.getValueAt(i, j);
                }
            }
            return data;
        }

        /**
         *  Converts a data array into the format expected by 
         *  <code>setTableData()</code>
         */
        private String arrayToString(String[][] list) {
            if (list == null) return "";
            StringBuffer newString = new StringBuffer();
            for (int i = 0; i < list.length; i++) {
                newString.append("{");
                for (int j = 0; j < list[i].length; j++) {
                    String item = list[i][j];
                    if (!((item == null) || (item.equals("null")) 
                          || item.equals(""))) {
                        newString.append(quoteString(item));
                        newString.append(", ");
                    }
                }
                newString.append("}, ");
            }
            return newString.toString();
        }

        /**
         * Adds quotes to incoming strings
         */
        private String quoteString(String str) {
            return "\"" + str.replaceAll("\"", "\\\"") + "\"";
        }
    }


    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------

    /**
     *  All of the option available to the user
     */
    private Options options;
        
    /**
     *  Flag to indicate whether the handler should load the preferred
     *  options or the options from the command line
     */
    private boolean useCommandLineOptions;
        
    /**
     *  The user preferences for the GUI
     */
    private Preferences preferences 
            = Preferences.userNodeForPackage(GUIFrame.class);
        
    /**
     *  List of all of the displayable options sets
     */
    private ArrayList optionsFactoryArray;
        
    /**
     *  Instance of the OptionsFactory for the normal options
     */
    private OptionsFactory normalOF = null;
        
    /**
     *  Instance of the OptionsFactory for the advanced options
     */
    private OptionsFactory advancedOF = null;
        
    /**
     *  Instance of the OptionsFactory for the javadocNormal options
     */
    private OptionsFactory javadocNormalOF = null;
    
        /**
     *  Instance of the OptionsFactory for the javadocNo options
     */
    private OptionsFactory javadocNoOF = null;
        
    /**
     *  Instance of the OptionsFactory for the javadocAdvanced options
     */
    private OptionsFactory javadocAdvancedOF = null;

    /**
     *  List of listeners for this object
     */
    private ArrayList listeners = null;
        
    /**
     *  The preferred width of the window
     */
    private int PREFERRED_WIDTH = 500;
        
    /**
     *  The preferred width of the window
     */
    private int MINIMUM_HEIGHT = 600;
}
