/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import java.util.prefs.Preferences;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

/** 
 *  Creates the compiler output window
 *@author     Kristina Boysen
 *@version    $Revision: 1.2 $
 */
public class GUIOutputWindow extends GUIFrame implements Runnable {

    /** 
     *  Builds the GUI for the output window
     */
    public GUIOutputWindow(GUI gui, InputStream is) {
        super("Output ");
        this.gui = gui;
        this.is = is;
        
        // Initialize preference system for the output window
        outputWindowPrefs = new GUIWindowPreferences(toolName, "Output", this);

        outputTextArea = new JTextArea();
        setJMenuBar(createMenus());

        getContentPane().setLayout(new BorderLayout());
                                        
        getContentPane().add(createControlPanel(), BorderLayout.NORTH);
        getContentPane().add(createOutputPanel(), BorderLayout.CENTER);
                
        outputWindowPrefs.setAllPreferences();
        setResizable(true);
        setVisible(true);
    }
        
    /**
     *  Create the menu bar for the output window
     */
    private JMenuBar createMenus() {
        // Menu bar
        JMenuBar menuBar = new JMenuBar();
        
        // File menu
        JMenu fileMenu = new JMenu("File");
                
        // Save button
        JMenuItem saveItem = new JMenuItem("Save Output File...");
        saveItem.addActionListener(new SaveOutputHandler(outputTextArea));
                                        
        // Exit button
        JMenuItem exitItem = new JMenuItem("Exit");
        exitItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent event) {
                    gui.interruptCompilation();
                    outputWindowPrefs.putAllPreferences();
                    setVisible(false);
                }
            });
                        
        fileMenu.add(saveItem);
        fileMenu.addSeparator();
        fileMenu.add(exitItem);
                
        menuBar.add(fileMenu);
                                        
        // Edit menu
        JMenu editMenu = new JMenu("Edit");
                                        
        // Copy button
        JMenuItem copyItem = new JMenuItem("Copy");
        copyItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent a) {
                    outputTextArea.copy();
                }
            });
                                        
        // Select all button
        JMenuItem selectAllItem = new JMenuItem("Select All");
        selectAllItem.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent a) {
                    outputTextArea.selectAll();
                }
            });

        editMenu.add(copyItem);
        editMenu.addSeparator();
        editMenu.add(selectAllItem);
                                        
        menuBar.add(editMenu);
                                        
        return menuBar;
    }
        
    /**
     *  Create the controlPanel that holds the compilation date and the
     *  "Stop" button
     */
    private JPanel createControlPanel() {
        JPanel controlPanel = new JPanel(new BorderLayout());
                
        JPanel outputPanel = new JPanel(new BorderLayout());
        outputPanel.add(
            new JLabel("Output: " + new java.util.Date()),
            BorderLayout.WEST);
        controlPanel.add(outputPanel, BorderLayout.WEST);
        
        JPanel buttonPanel = new JPanel(new BorderLayout());
        
        scrollBarOption 
            = new JCheckBox("Scrollbar always at bottom of output");
        scrollBarOption.setSelected(preferences.getBoolean(
                    "ScrollBarAtBottom", true));
        scrollBarOption.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    scrollBarAtBottom 
                        = ((JCheckBox)(ae.getSource())).isSelected();
                }
            });
        buttonPanel.add(scrollBarOption, BorderLayout.WEST);
        
        JButton stop = new JButton("Stop!",
            new ImageIcon(GUIUtil.loadIconFromJar("stop.gif")));
        stop.addActionListener(
            new ActionListener() {
                public void actionPerformed(ActionEvent ae) {
                    gui.interruptCompilation();
                }
            });
        buttonPanel.add(stop, BorderLayout.EAST);
                
        controlPanel.add(buttonPanel, BorderLayout.EAST);
                
        return controlPanel;
    }
        
    /**
     *  Create the outputPanel that holds the compilation results
     */
    private JPanel createOutputPanel() {
        JPanel outputPanel = new JPanel(new BorderLayout());
        outputTextArea.setEditable(false);
        JScrollPane scrollPane = new JScrollPane(outputTextArea);
        outputPanel.add(scrollPane, BorderLayout.CENTER);
                
        return outputPanel;
    }
                         
    /**
     *  Gets the error messages during runtime
     */
    public void run() {
        BufferedReader br = null;
        try {
            try {
                br = new BufferedReader(new InputStreamReader(is));
                String str = "";
                while (((str = br.readLine()) != null)) {
                    outputTextArea.append(str + '\n');
                    if (scrollBarAtBottom) {
                        outputTextArea.setCaretPosition(
                            outputTextArea.getDocument().getLength());
                    }
                }
            } finally {
                if (br != null) {
                    br.close();
                }
            }
        } catch (IOException ioe) {
            // !FIXME! should do something with this...
        }

        outputTextArea.append("Done");
    }
        
    /**
     *  Sets the "Set Look and Feel" button in GUI.java to the value of
     *  name
     */
    protected void setCurrentLookAndFeelEnabled(String name) {
        gui.setCurrentLookAndFeelEnabled(name);
    }
        
    /**
     *  Sets the "Save Preferences" button in GUI.java to the value of
     *  enabled
     */
    protected void setSavePrefsEnabled(boolean enabled) {
        gui.setSavePrefsEnabled(enabled);
    }
        
    /**
     *  Sets the "Clear Preferences" button in GUI.java to the value of
     *  enabled
     */
    protected void setClearPrefsEnabled(boolean enabled) {
        gui.setClearPrefsEnabled(enabled);
    }
        
    // ----------------------------------------------------------------------
    // PREFERENCE METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Sets this window's look and feel to the main window's look and feel
     */
    protected void setLookAndFeel() {
        try {
            UIManager.setLookAndFeel(UIManager.getLookAndFeel());
        } catch (Exception e) {}
                
        if (getContentPane() != null) {
            SwingUtilities.updateComponentTreeUI(getContentPane());
        }
                
        if (getJMenuBar() != null) {
            SwingUtilities.updateComponentTreeUI(getJMenuBar());
        }
    }
        
    /**
     * Returns the window's minimum size.
     */
    public Dimension getMinimumSize() {
        int width = 500;
        int height = 200;
        return new Dimension(width, height);
    }
        
    /**
     *  Returns the user's preferred size for this window.
     */
    public Dimension getPreferredSize() {
        return outputWindowPrefs.getPreferredSize();
    }

    /**
     *  Saves the preferences into memory when the window closes
     */
    protected void closeWindowOperations() {
        super.closeWindowOperations();
        savePreferences();
        gui.interruptCompilation();
                
        setVisible(false);
    }
        
    /**
     *  Returns the preferences for this window
     */
    protected GUIWindowPreferences getGUIWindowPreferences() {
        return outputWindowPrefs;
    }
        
    /**
     *  Override of GUIFrame's savePreferences() so the options can be saved
     */
    protected void savePreferences() {
        super.savePreferences();
        outputWindowPrefs.putAllPreferences();
                preferences.putBoolean("ScrollBarAtBottom", 
                    scrollBarOption.isSelected());
    }
        
    /**
     *  Override of GUIFrame's clearPreferences() so the options can be reset
     */
    protected void clearPreferences() {
        super.savePreferences();
        outputWindowPrefs.setAllPreferences();
                scrollBarOption.setSelected(preferences.getBoolean(
                    "ScrollbarAtBottom", true));
    }
        
    /**
     *  Returns whether the preferences have changed for this window
     */
    protected boolean preferencesHaveChanged() {
        return (super.preferencesHaveChanged()
            || outputWindowPrefs.preferencesHaveChanged()
            || (scrollBarOption.isSelected() 
                != preferences.getBoolean("ScrollBarAtBottom", true)));
    }

    // ----------------------------------------------------------------------
    // PRIVATE CLASSES
    // ----------------------------------------------------------------------
        
    /**
     *  Handles the selection of files in an save file window
     */
    protected class SaveOutputHandler implements ActionListener {
    
        protected JFileChooser fileChooser = 
            new JFileChooser(currentDirectory);
        private JTextArea outputArea;

        /**
         *  Provides a way for the user to select a location to save to
         */
        protected SaveOutputHandler(JTextArea output) {
            outputArea = output;
                        
            fileChooser.setDialogTitle("Save Output File");
            fileChooser.setApproveButtonText("Save");
            fileChooser.setMultiSelectionEnabled(false);
            fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        }

        /**
         *  Saves the output file
         */
        public void actionPerformed(ActionEvent event) {
            int returnVal = fileChooser.showSaveDialog(null);
            if (returnVal == JFileChooser.APPROVE_OPTION) {
                File file = fileChooser.getSelectedFile();
                try {
                    FileWriter writer;
                    if (file != null) {
                        writer = new FileWriter(file);
                    } else {
                        file = new File(fileChooser.getName(file));
                        writer = new FileWriter(file);
                    }
                    String outputText = outputArea.getText();
                    writer.write(outputText, 0, outputText.length());
                    writer.close();
                } catch (java.io.IOException io) {
                }
                                
                //Set dialog location to last location accessed
                putCurrentDirectory();
            }
        }
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------
    
    /**
     *  Instance of the GUI that called this output window so the user can
     *  set "Save Preferences" and "Clear Preferences", and also so the
     *  user can stop the compilation
     */
    private GUI gui = null;
        
    /**
     *  Window size and location preferences for the output window.
     */
    private GUIWindowPreferences outputWindowPrefs = null;
        
    /**
     *  Other user preferences for the window
     */
    private Preferences preferences 
            = Preferences.userNodeForPackage(GUIFrame.class);
    
    /**
     *  The text area where the about information is put
     */
    private JTextArea outputTextArea;
    
    /**
     *  Stream that pipes the output from the compiler
     */
    private InputStream is;
        
    /**
     *  Allows the user to select whether the scrollbar of the output window
     *  should always scroll to the bottom or stay at the current position
     */
    private JCheckBox scrollBarOption;
        
    /**
     *  A flag that indicates whether the scroll knob is at the bottom
     *  of the track
     */
    private boolean scrollBarAtBottom = true;
}
