/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import java.util.HashMap;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.KeyStroke;

import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;

import javax.swing.text.DefaultEditorKit;
import javax.swing.text.JTextComponent;

/**
 * A class to create a JPopupMenu with various edit-related actions.
 *
 *@author    Yoonsik Cheon
 *@author    Kristina Boysen
 *@version   $Revision 1.0 $
 */
public class GUIPopupMenu extends JPopupMenu {
        
    /**
     *  Creates a new JPopup window for <code>textField<\code>
     */
    public GUIPopupMenu(JTextField textField) {
        super("Edit");

        EditActionFactory ef = new EditActionFactory(textField);
        textField.addCaretListener(ef.getCaretListener());

        JMenuItem item = new JMenuItem(ef.getCopy());
        item.setMnemonic(KeyEvent.VK_C);
        item.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_C, ActionEvent.CTRL_MASK));
        add(item);

        item = new JMenuItem(ef.getCut());
        item.setMnemonic(KeyEvent.VK_T);
        item.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_X, ActionEvent.CTRL_MASK));
        add(item);

        item = new JMenuItem(ef.getPaste());
        item.setMnemonic(KeyEvent.VK_P);
        item.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_V, ActionEvent.CTRL_MASK));
        add(item);

        item = new JMenuItem(ef.getDelete());
        item.setMnemonic(KeyEvent.VK_D);
        item.setAccelerator(KeyStroke.getKeyStroke(
            KeyEvent.VK_DELETE, 0));
        add(item);
    }


    private static class EditActionFactory {
        
        /**
         * Creates a new instance for the given editor <code>tc</code>.
         */
        public EditActionFactory(/*@ non_null @*/ JTextComponent tc) {
            editor = tc;
            createActionTable(tc);
        }
        
        /**
         * Returns the action for <em>copy</em> menu item. Initially the
         * action is disabled, but it gets enabled whenever something is
         * selected on the editor.
         * 
         * <pre><jml>
         *  assignable actCopy;
         * </jml></pre>
         *
         * @see #getCaretListener()
         */    
        public /*@ non_null @*/ Action getCopy() {
            if (actCopy != null) {
                return actCopy;
            }
            actCopy = new PasteEnablingAction("Copy", getIcon("Copy.gif"), 
                getAction(DefaultEditorKit.copyAction));
            return actCopy;
        }
        
        /**
         * Returns the action for <em>cut</em> menu item. Initially the
         * action is disabled, but it gets enabled whenever something is
         * selected on the editor.
         * 
         * <pre><jml>
         *  assignable actCut;
         * </jml></pre>
         *
         * @see #getCaretListener()
         */    
        public /*@ non_null @*/ Action getCut() {
            if (actCut != null) {
                return actCut;
            }
            actCut = new PasteEnablingAction("Cut", getIcon("Cut.gif"), 
                getAction(DefaultEditorKit.cutAction));
            return actCut;
        }
        
        /**
         * Returns the action for <em>paste</em> menu item. Initially the
         * action is disabled, and gets enabled whenever something is
         * copied or cut by the editor.
         * 
         * <pre><jml>
         *  assignable actPaste;
         * </jml></pre>
         */    
        public /*@ non_null @*/ Action getPaste() {
            if (actPaste != null) {
                return actPaste;
            }
            actPaste = new PasteEnablingAction("Paste", getIcon("Paste.gif"), 
                getAction(DefaultEditorKit.pasteAction));
            return actPaste;
        }
        
        /**
         * Returns the action for <em>delete</em> menu item. Initially the
         * action is disabled, and gets enabled whenever something is
         * selected on the editor.
         * 
         * <pre><jml>
         *  assignable actDelete;
         * </jml></pre>
         *
         * @see #getCaretListener()
         */    
        public /*@ non_null @*/ Action getDelete() {
            if (actDelete != null) {
                return actDelete;
            }
        
            actDelete = new AbstractAction("Delete", getIcon("Delete.gif")) {
                    public void actionPerformed(ActionEvent e) {
                        editor.replaceSelection(null);
                    }
                };
        
            actDelete.setEnabled(false);
            return actDelete;
        }
        
        /** Returns the caret listener. */
        public /*@ non_null @*/ CaretListener getCaretListener() {
            return new GUICaretListener();
        }
        
        /** Returns an image icon whose image is read from the file named
         * <code>name</code>. A null is returned if reading the file fails.
         */
        protected ImageIcon getIcon(/*@ non_null @*/ String name) {
            return null; //Util.getIcon(name);
        }
        
        /**
         * Initializes the hash map <code>actions</code> to populate it
         * with the edit actions of the given text component provided by
         * the default editor tool kit.
         *
         * @see #actions
         */
        private void createActionTable(JTextComponent textComponent) {
            actions = new HashMap();
            Action[] acts = textComponent.getActions();
            for (int i = 0; i < acts.length; i++) {
                Action a = acts[i];
                actions.put(a.getValue(Action.NAME), a);
            }
        }
        
        /**
         * Returns the action of the given name from the hash map
         * <code>actions</code>.
         */
        private Action getAction(/*@ non_null @*/ String name) {
            return (Action)(actions.get(name));
        }
        
        
        /**
         * An action class to represent edit actions that enables the
         * paste action. This is a wrapper (or declarator) class that
         * delegates its action to some edit action and then enables the
         * paste edit action.
         */
        private static class PasteEnablingAction extends AbstractAction {
            
            /** 
             * Creates a new instance by wrapping the given edit action
             * <code>action</code>. 
             */
            public PasteEnablingAction(String name, Icon icon, Action action) {
                super(name, icon);
                this.action = action;
                setEnabled(false);
            }
        
            /** 
             * Handles the given action. The action is delegated to the
             * action <code>action</code> and the paste action is
             * enabled. 
             */
            public void actionPerformed(ActionEvent e) {
                // !FIXME! also enable on CTRL-X, etc.
                action.actionPerformed(e);
                actPaste.setEnabled(true);
            }
        
            /** The delegee action. */
            private Action action;
        }
        
        /**
         * A caret listener class for the editor. This class is used to
         * turn on and off such editing actions as copy, cut, and delete,
         * which should be enabled only when something is selected.
         */
        private class GUICaretListener implements CaretListener {
            
            /**
             * Enables or disables such actions as copy, cut, and delete
             * based on the current section on the editor. If nothing is
             * selected, they are all disabled; otherwise, they are all
             * enabled.
             */
            public void caretUpdate(CaretEvent e) {
                // no selection made?
                if (e.getDot() == e.getMark()) { 
                    // disable copy, cut and delete
                    actCopy.setEnabled(false);
                    actCut.setEnabled(false);
                    actDelete.setEnabled(false);
                } else {
                    // enable copy, cut and delete
                    actCopy.setEnabled(true);
                    actCut.setEnabled(true);
                    actDelete.setEnabled(true);
                }
            }
        }
        
        /** The action for the <em>copy</em> menu. */
        private /*@ spec_public @*/ Action actCopy = null;
        
        /** The action for the <em>cut</em> menu. */
        private /*@ spec_public @*/ Action actCut = null;
        
        /** The action for the <em>Paste</em> menu. This is shared by all. */
        private /*@ spec_public @*/ static Action actPaste = null;
        
        /** The action for the <em>Delete</em> menu. */
        private /*@ spec_public @*/ Action actDelete = null;
        
        /** The editor which this factory creates edit actions for. */
        private JTextComponent editor;
        
        /** 
         * The edit actions of the text component <code>editor</code>
         * provided by the default editor tool kit. The edit actions will
         * be implemented by using these default actions.
         *
         * @see #createActionTable(JTextComponet)
         */
        private HashMap actions;
    }
}
