/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package org.multijava.util.gui;

import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Toolkit;

import java.util.prefs.Preferences;

import javax.swing.JFrame;

/**
 *  Contains all of the window preferences
 *@author     Kristina Boysen
 *@version    $Revision: 1.1 $
 */
public class GUIWindowPreferences {
        
    /**
     *  Initializes the class variables
     */
    public GUIWindowPreferences(String toolName, String windowName, 
        JFrame window) {
        this.toolName = toolName;
        this.windowName = windowName;
        this.window = window;
    }
        
    // ----------------------------------------------------------------------
    // PREFERRED SIZE METHOD
    // ----------------------------------------------------------------------
        
    public Dimension getPreferredSize() {
        int width = preferences.getInt(
            toolName + windowName + "WindowWidth", 
            (int)(window.getMinimumSize().getWidth()));
        int height = preferences.getInt(
            toolName + windowName + "WindowHeight", 
            (int)(window.getMinimumSize().getHeight()));
        return new Dimension(width, height);
    }

    // ----------------------------------------------------------------------
    // SET METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Set the window to the user's preferred window state.
     */
    public void setWindowState() {
        // Set the window state to the preferred window state.  If there
        // is no preference set, the window will appear as normal (not
        // maximized)
        int windowState = preferences.getInt(
            toolName + windowName + "WindowState", 
            Frame.NORMAL);
        window.setExtendedState(windowState);
    }
        
    /**
     *  Set the window to the user's preferred window dimensions 
     */
    public void setWindowDimensions() {
        // If the window state is normal (not maximized), the dimensions
        // of the window will be preferred width and height.  If there
        // are no preferences specified, the dimensions of the window
        // will be determined by the preferred size of the window.
        if (window.getExtendedState() == Frame.NORMAL) {
            window.setSize(getPreferredSize());
        }
    }
        
    /**
     *  Sets the window location to the user's preferred window location
     */
    public void setWindowLocation() {
        // If the window state is normal (not maximized), the location
        // of the window will be preferred x and y locations.  If there
        // are no preferences specified, the window will appear in the
        // center of the screen
        if (window.getExtendedState() == Frame.NORMAL) {
            Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
            int x = preferences.getInt(toolName + windowName + "WindowX", 
                (int)((screen.getWidth() - window.getWidth()) / 2));
            int y = preferences.getInt(toolName + windowName + "WindowY", 
                (int)((screen.getHeight() - window.getHeight()) / 2));
            window.setLocation(x, y);
        }
    }
        
    /**
     *  Sets all of the window values to the user's preferred values
     */
    public void setAllPreferences() {
        window.pack();
                
        setWindowState();
        setWindowDimensions();
        setWindowLocation();
    }
        
    // ----------------------------------------------------------------------
    // PUT METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Set the current window state as the user's preferred size
     */
    public void putWindowState() {
        if (windowStateHasChanged()) {
            // The current window state is set as the preferred window state
            preferences.putInt(toolName + windowName + "WindowState", 
                window.getExtendedState());
        }
    }

    /**
     *  Set the current window size as the user's preferred size
     */
    public void putWindowDimensions() {
        if (windowDimensionsHaveChanged()) {
            // If the window state is normal (not maximized) then the 
            // width and height preferences will be set.
            if (window.getExtendedState() == Frame.NORMAL) {
                preferences.putInt(toolName + windowName + "WindowWidth", 
                    window.getWidth());
                preferences.putInt(toolName + windowName + "WindowHeight", 
                    window.getHeight());
            }
        }
    }
        
    /**
     *  Sets the current window's location as the user's preferred location
     */
    public void putWindowLocation() {
        if (windowLocationHasChanged()) {
            // If the window state is normal (not maximized) then the 
            // x and y location preferences will be set.
            if (window.getExtendedState() == Frame.NORMAL) {
                preferences.putInt(toolName + windowName + "WindowX", 
                    window.getX());
                preferences.putInt(toolName + windowName + "WindowY", 
                    window.getY());
            }
        }
    }
        
    /**
     *  Sets all of the current values as the user's preferred values
     */
    public void putAllPreferences() {
        putWindowState();
        putWindowDimensions();
        putWindowLocation();
    }
        
    // ----------------------------------------------------------------------
    // CHECK FOR PREFERENCE CHANGED METHODS
    // ----------------------------------------------------------------------
        
    /**
     *  Returns whether the state of the window has changed from either
     *  the preferences settings or the default
     */
    public boolean windowStateHasChanged() {
        return (window.getExtendedState() != preferences.getInt(
            toolName + windowName + "WindowState", Frame.NORMAL));
    }
        
    /**
     *  Returns whether the window dimensions have changed from either
     *  the preferences settings or the default
     */
    public boolean windowDimensionsHaveChanged() {
        return (window.getWidth() != preferences.getInt(
                    toolName + windowName + "WindowWidth", 
                    (int)(getPreferredSize().getWidth()))
                || window.getHeight() != preferences.getInt(
                    toolName + windowName + "WindowHeight", 
                    (int)(getPreferredSize().getHeight())));
    }
        
    /**
     *  Returns whether the location of the window has changed from either
     *  the preferences settings or the default
     */
    public boolean windowLocationHasChanged() {
        Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
        return (window.getX() != preferences.getInt(
                    toolName + windowName + "WindowX",
                    (int)((screen.getWidth() - window.getWidth()) / 2))
                || window.getY() != preferences.getInt(
                    toolName + windowName + "WindowY",
                    (int)((screen.getHeight() - window.getHeight()) / 2)));
        }
        
    /**
     *  Returns whether any of the window preferences have changed
     */
    public boolean preferencesHaveChanged() {
        return (windowStateHasChanged() 
            || windowDimensionsHaveChanged()
            || windowLocationHasChanged());
    }
        
    // ----------------------------------------------------------------------
    // PRIVATE CLASS VARIABLES
    // ----------------------------------------------------------------------

    /**
     *  Name of the tool using these preferences
     */ 
    private String toolName;

    /**
     *  Name of the window that the preferences are for
     */ 
    private String windowName;

    /**
     *  The window that the preferences are for
     */         
    private JFrame window;

    /**
     *  Where the preferences are stored
     */ 
    private Preferences preferences = 
        Preferences.userNodeForPackage(GUIFrame.class);
}
