/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: DefinitionFile.java,v 1.23 2004/07/08 01:07:10 netgecko Exp $
 */

package org.multijava.util.guigen;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.PrintWriter;

import java.util.ArrayList;
import java.util.Iterator;

class DefinitionFile {

    /**
     * Constructs a definition file
     */
    public DefinitionFile(String sourceFile,
                          String packageName,
                          ArrayList imports,
                          String prefix,
                          String readableFiles,
                          String webpageName,
                          String webpageLocation,
                          ArrayList acceptedSuffixes) {

        this.sourceFile         = sourceFile;
        this.packageName        = packageName;
        this.imports            = imports;
        this.prefix             = prefix;
        this.readableFiles      = readableFiles;
        this.webpageName        = webpageName;
        this.webpageLocation    = webpageLocation;
        this.acceptedSuffixes   = trailList(acceptedSuffixes);
    }
    
    /**
     *  Strips the leading and trailing quotes from a string
     */
    private static final String trail(String s) {
        if (s == null) {
            return null;
        } else if (s.length() < 2) {
            return s;
        } else {
            return s.substring(1, s.length() - 1);
        }
    }
    
    /**
     *  Strips the leading and trailing quotes from an ArrayList of strings
     */
    private static final ArrayList trailList(ArrayList list) {
        if (list == null) {
            return list;
        } else {
            // strip leading and trailing quotes
            ArrayList returnList = new ArrayList(list.size());
            for (int i = 0; i < list.size(); i++) {
                String s = (String)list.get(i);
                if (s.length() < 2) {
                    returnList.add(i, s);
                } else {
                    returnList.add(i, s.substring(1, s.length() - 1));
                }
            }
            return returnList;
        }
    }
        
    /**
     * Reads and parses an token definition file
     *
     * @param   sourceFile              the name of the source file
     * @return  a class info structure holding the information from the source
     *
     */
    public static DefinitionFile read(String sourceFile) throws GuigenError {
        try {
            InputStream input = new BufferedInputStream(
                                    new FileInputStream(sourceFile));
            GuigenLexer scanner = new GuigenLexer(input);
            GuigenParser        parser = new GuigenParser(scanner);
            DefinitionFile      defs = parser.aCompilationUnit(sourceFile);

            input.close();

            return defs;
        } catch (java.io.FileNotFoundException e) {
            throw new GuigenError(GuigenMessages.UNKNOWN_FILE, sourceFile);
        } catch (java.io.IOException e) {
            throw new GuigenError(GuigenMessages.IO_EXCEPTION, sourceFile, 
                                  e.getMessage());
        } catch (antlr.RecognitionException e) {
            throw new GuigenError(GuigenMessages.PARSING_ERROR,
                                  sourceFile,
                                  e.getLine() + " : " + e.getMessage());
        } catch (antlr.TokenStreamException e) {
            throw new GuigenError(GuigenMessages.LEXING_ERROR,
                                  sourceFile, e.getMessage());
        }

    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Prints the token definition to interface file (java)
     * @param   out             the output stream
     */
    public void printFile(PrintWriter out) {
        out.print("// Generated from " + sourceFile);
        out.println();
        if (packageName != null) {
            out.println("package " + packageName + ";");
            out.println();
        }
                
        //IMPORTS
        out.println("import java.util.ArrayList;");
        out.println("import java.io.IOException;");
        out.println("import java.io.InputStream;");
        out.println("import java.io.OutputStream;");
        out.println("import javax.swing.JFileChooser;");
        out.println();
        out.println("import org.multijava.util.gui.GUIOutputWindow;");
        out.println("import org.multijava.util.Options;");
        out.println();
        if (imports != null) {
            Iterator ii = imports.iterator();
            while (ii.hasNext()) {
                out.println("import " + ii.next() + ";");
            }
        }
                
        //DESCRIPTION
        out.println();
        out.println("/** This class is automatically generated from " 
                    + sourceFile);
        out.println(" *  and contains member fields corresponding to " 
                    + "tool-specific GUI specifications.");
        out.println(" */");
        out.println("public class " + prefix + "GUI"
                    + " extends org.multijava.util.gui.GUI {");
        out.println();

        // MAIN
        out.println();
        out.println("  public static void main(String[] args) {");
        out.println("    init(args, true);");
        out.println("  }");
        out.println();
        
        //INIT
        out.println();
        out.println("  public static " + prefix + "GUI init(String[] args, "
                    + "boolean standAlone) {");
        out.println("    " + prefix + "Options options = new " 
                    + prefix + "Options();");
        out.println("    ArrayList infiles = new ArrayList();");
        out.println("    ArrayList extraArguments = new ArrayList();");
        out.println("    boolean commandLine = false;");
        out.println("    if (args.length != 0) {");
        out.println("      commandLine = true;");
        out.println("      new Main().runParser(args, options, infiles);");
        out.println("    }");
        out.println("    return new " + prefix 
                    + "GUI(options, infiles, commandLine, standAlone);");
        out.println("  }");
        out.println();

        // CONSTRUCTOR
        out.println();
        out.println("  public " + prefix 
                    + "GUI(Options options, ArrayList infiles, "
                    + "boolean commandLine, boolean standAlone) {");
        out.println("    super(\"" + prefix + "\", options, infiles, "
                    + "commandLine, standAlone);");
        out.println("  }");
        out.println();

        // INTERRUPT COMPILATION
        out.println();
        out.println("  public void interruptCompilation() {");
        out.println("    Main.interruptCompilation();");
        out.println("  }");
        out.println();
                
        // DIRINCLASSPATH
        out.println();
        out.println("  protected boolean "
                    + "dirInClassPath(Options opt, String dir) {");
        if (prefix.equals("Jml") || prefix.equals("Jmldoc")
            || prefix.equals("Rac")) {
            out.println("    if (((" + prefix 
                        + "Options)opt).sourcepath() != null) {");
            out.println("      return GUIUtil.dirInPath(((" + prefix 
                        + "Options)opt).sourcepath(), dir);");
            out.println("    } else if (((" + prefix 
                        + "Options)opt).classpath() != null) {");
            out.println("      return GUIUtil.dirInPath(((" + prefix 
                        + "Options)opt).classpath(), dir);");
            out.println("    } else {");
            out.println("      return false;");
            out.println("    }");
        } else {
            out.println("    return true;");
        }
        out.println("  }");
        out.println();
        
        // GET CLASSPATH 
        out.println();
        out.println("  protected String getClasspath(Options opt) {");
        out.println("    return ((" + prefix + "Options)opt).classpath();");
        out.println("  }");
        out.println();

        // GET SOURCEPATH
        out.println();
        out.println("  protected String getSourcepath(Options opt) {");
        out.println("    return ((" + prefix + "Options)opt).sourcepath();");
        out.println("  }");
        out.println();
        
        // CREATE OPENHANDLER INSTANCE
        out.println();
        out.println("  protected OpenHandler "
                    + "createOpenHandlerInstance() {");
        out.println("    return new " + prefix + "OpenHandler();");
        out.println("  }");
        out.println();
        
        // CREATE COMPILATION INSTANCE
        out.println();
        out.println("  protected Compilation createCompilationInstance("
                    + "String[] files, Options opt, OutputStream os) {");
        out.println("    return new " + prefix + "Compilation(files, (" 
                    + prefix + "Options)opt, os);");
        out.println("  }");
        out.println();
                
        // GET WEBPAGE NAME
        out.println();
        out.println("  protected String getWebpageName() {");
        out.println("    return " + webpageName + ";");
        out.println("  }");
        out.println();
                
        // GET WEBPAGE LOCATION
        out.println();
        out.println("  protected String getWebpageLocation() {");
        out.println("    return " + webpageLocation + ";");
        out.println("  }");
        out.println();
                
        // CREATE OUTPUT WINDOW
        out.println();
        out.println("  protected GUIOutputWindow createOutputWindow("
                    + "InputStream is) {");
        out.println("    return new GUIOutputWindow(this, is);");
        out.println("  }");
        out.println();
        
        // TOOL OPENHANDLER
        out.println();
        out.println("  protected class " + prefix
                    + "OpenHandler extends OpenHandler {");
        out.println("    protected " + prefix + "OpenHandler() {");
        out.println("    }");
        out.println("    protected JFileChooser createFileChooser() {");
        out.println("      JFileChooser fileChooser "
                                   + "= super.createFileChooser();");
        out.println("      AllFilesGUIFileFilter all " 
                    + "= new AllFilesGUIFileFilter();");
        out.println("      fileChooser.addChoosableFileFilter(all);");
        for (int i = 0; i < acceptedSuffixes.size(); i++) {
            out.println("      fileChooser.addChoosableFileFilter(new "
                        + prefix + "GUIFileFilter("
                        + "\"." + acceptedSuffixes.get(i) + "\"));");
        }
        out.println("      fileChooser.setFileFilter(all);");
        out.println("      return fileChooser;");
        out.println("    }");
        out.println("  }");
        out.println();
                
        // ALLFILES FILEFILTER
        out.println();
        out.println("  protected class " 
                    + "AllFilesGUIFileFilter extends GUIFileFilter {");
        out.println("    public boolean hasActiveSuffix(String name) {");
        if (acceptedSuffixes != null) {
            out.print("      return name.endsWith(\".");
            out.print(acceptedSuffixes.get(0));
            out.print("\")");
            for (int i = 1; i < acceptedSuffixes.size(); i++) {
                out.println();
                out.print("        || name.endsWith(\".");
                out.print(acceptedSuffixes.get(i));
                out.print("\")");
            }
            out.println(";");
        }
        out.println("    }");
        out.println("    public String getDescription() {");
        out.println("      return " + readableFiles + ";");
        out.println("    }");
        out.println("  }");
        out.println();
                
        // PREFIX GUIFILEFILTER
        out.println();
        out.println("  protected class " + prefix
                    + "GUIFileFilter extends GUIFileFilter {");
        out.println("    String suffix;");
        out.println("    public " + prefix 
                    + "GUIFileFilter(String suffix) {");
        out.println("      this.suffix = suffix;");
        out.println("    }");
        out.println("    public boolean hasActiveSuffix(String name) {");
        out.println("      return name.endsWith(suffix);");
        out.println("    }");
        out.println("    public String getDescription() {");
        out.println("      return suffix;");
        out.println("    }");
        out.println("  }");
        out.println();

        // TOOL COMPILATION
        out.println();
        out.println("  protected class " + prefix + "Compilation extends "
                    + "Compilation implements Runnable {");
        out.println("    private String[] files;");
        out.println("    private " + prefix + "Options options;");
        out.println("    private OutputStream os;");
        out.println();
        out.println("    protected " + prefix + "Compilation(String[] files, "
                    + prefix + "Options options, OutputStream os) {");
        out.println("      this.files = files;");
        out.println("      this.options = options;");
        out.println("      this.os = os;");
        out.println("    }");
        out.println();
        out.println("    public void run() {");
        out.println("      Main.compile(files, options, os);");
        out.println("    }");
        out.println("  }");
        out.println();
                
        //EOF
        out.println();
        out.println("}");
    }

    /**
     * Returns the class name
     */
    public String getClassName() {
        return packageName + "." + prefix + "GUI";
    }

    /**
     * Returns the package name
     */
    public String getPackageName() {
        return packageName;
    }

    /**
     * Returns the literal prefix
     */
    public String getPrefix() {
        return prefix;
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private final String    sourceFile;
    private final String    packageName;
    private final ArrayList imports;
    private final String    prefix;
    private final String    readableFiles;
    private final String    webpageName;
    private final String    webpageLocation;
    private final ArrayList acceptedSuffixes;
}
