/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: DefinitionFile.java,v 1.6 2004/01/16 16:26:57 cclifton Exp $
 */

package org.multijava.util.lexgen;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.ArrayList;

import org.multijava.util.compiler.CompilerMessages;
import java.util.Collections;
import java.util.List;

class DefinitionFile {

    /**
     * Constructs a token definition file
     */
    public DefinitionFile(String sourceFile, String packageName,
			  String vocabulary, String prefix, List flags,
			  ArrayList definitions) {
	this.sourceFile	= sourceFile;
	this.packageName	= packageName;
	this.vocabulary	= vocabulary;
	this.prefix		= prefix == null ? DEFAULT_PREFIX : prefix;

	this.flags = Collections.unmodifiableList( flags );
	this.definitions	= new TokenDefinition[definitions.size()];
	definitions.toArray(this.definitions);
    }

    /**
     * Reads and parses an token definition file
     *
     * @param	sourceFile		the name of the source file
     * @return	a class info structure holding the information from the source
     *
     */
    public static DefinitionFile read(String sourceFile) throws LexgenError {
	try {
	    InputStream	input = new BufferedInputStream(new FileInputStream(sourceFile));
	    LexgenLexer	scanner = new LexgenLexer(input);
	    LexgenParser	parser = new LexgenParser(scanner);
	    DefinitionFile	defs = parser.aCompilationUnit(sourceFile);

	    input.close();

	    return defs;
	} catch (java.io.FileNotFoundException e) {
	    throw new LexgenError(LexgenMessages.FILE_NOT_FOUND, sourceFile);
	} catch (java.io.IOException e) {
	    throw new LexgenError(CompilerMessages.IO_EXCEPTION, sourceFile, e.getMessage());
	} catch (antlr.RecognitionException e) {
	    throw new LexgenError(LexgenMessages.PARSING_ERROR,
				  sourceFile,
				  e.getLine() + " : " + e.getMessage());
	} catch (antlr.TokenStreamException e) {
	    throw new LexgenError( LexgenMessages.LEXING_ERROR,
				   sourceFile, e.getMessage() );
	}

    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Check for duplicate identifiers
     * @param	identifiers	a table of all token identifiers
     * @param	prefix		the literal prefix
     * @param	id		the id of the first token
     * @return	the id of the last token + 1
     */
    public int checkIdentifiers(Hashtable identifiers, String prefix, int id)
	throws LexgenError
    {
	for (int i = 0; i < definitions.length; i++) {
	    definitions[i].checkIdentifiers(identifiers, prefix, id++, sourceFile);
	}

	return id;
    }

    /**
     * Prints token definitions to definition file (txt)
     * @param	out		the output stream
     * @param	prefix		the literal prefix
     */
    public void printDefinition(PrintWriter out, String prefix) {
	out.println();
	out.println("// Definitions from " + sourceFile);
	for (int i = 0; i < definitions.length; i++) {
	    definitions[i].printDefinition(out, prefix);
	}
    }

    /**
     * Prints the token definition to interface file (java)
     * @param	out		the output stream
     * @param	parent		the super interface
     */
    public void printInterface(PrintWriter out, String parent) {
	out.print("// Generated from " + sourceFile);
	out.println();
	out.println("package " + packageName + ";");
	out.println();
	out.print("public interface " + vocabulary + "TokenTypes");
	if (parent != null) {
	    out.print(" extends " + parent);
	}
	out.println(" {");

	if (parent == null) {
	    out.println("  int\tEOF = 1;");
	    out.println("  int\tNULL_TREE_LOOKAHEAD = 3;");
	}

	for (int i = 0; i < definitions.length; i++) {
	    definitions[i].printInterface(out, this.prefix);
	}

	out.println("}");
    }

    /**
     * Accumulates a list of keyword token information in <code>accum</code>.
     *
     * <pre><jml>
     * requires (\forall Object o; accum.contains(o); 
     *				   o instanceof KeywordTokenInfo);
     * ensures (\forall Object o; accum.contains(o); 
     *				  o instanceof KeywordTokenInfo);
     * also
     *   ensures \result != true ==> (* an error was reported to System.err *);
     * </jml></pre>
     *
     */
    boolean accumKeywords(List accum, String prefix, List legalFlags) {
	boolean ok = true;
	for (int i = 0; i < definitions.length; i++) {
	    ok &= definitions[i].accumKeyword(accum, prefix, legalFlags);
	}
	return ok;
    }

    /**
     * This nested class serves as a data structure for accumulating
     * information in {@link #accumKeywords(List,String,List)}.
     */
    static class KeywordTokenInfo {
	KeywordTokenInfo( String keyword, String type, int flags ) {
	    this.keyword = keyword;
	    this.type = type;
	    this.flags = flags;
	}
	final String keyword;
	final String type;
	final int flags;
    }

    /**
     * Returns the package name
     */
    public String getClassName() {
	return packageName + "." + vocabulary + "TokenTypes";
    }

    /**
     * Returns the package name
     */
    public String getPackageName() {
	return packageName;
    }

    /**
     * Returns the vocabulary name
     */
    public String getVocabulary() {
	return vocabulary;
    }

    /**
     * Returns the literal prefix
     */
    public String getPrefix() {
	return prefix;
    }

    /**
     * Returns an immutable list of identifier Strings
     * representing the valid flags for keywords and literals in this
     * vocabulary.
     */
    public List flags() {
	return flags;
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private static final String		DEFAULT_PREFIX = "LITERAL_";

    private final String		sourceFile;
    private final String		packageName;
    private final String		vocabulary;
    private final String		prefix;
    

    /**
     * An immutable list of identifier Strings representing the
     * valid flags for keywords and literals in this vocabulary.
     */
    private final List flags;

    private final TokenDefinition[]	definitions;
}
