/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.8 2004/01/16 16:26:57 cclifton Exp $
 */

package org.multijava.util.lexgen;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.ArrayList;
import org.multijava.util.Utils;

import org.multijava.util.jperf.JPerf;
import java.util.TreeSet;
import java.util.ArrayList;

/**
 * This class is the entry point for the keyword lookup generation facility.
 */
public class Main {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    public Main(String[] args) {
	this.args = args;
    }
    
    public boolean run() {
	if (!parseArguments(args)) {
	    return false;
	}

	if (! options.definition() && !options.inter() && !options.keywords()) {
	    options.usage();
	    return false;
	}

	String[] infiles = options.nonOptions();

	if (infiles.length == 0) {
	    options.usage();
	    System.err.println(LexgenMessages.NO_SOURCE_FILE.getFormat());
	    return false;
	} else {
	    boolean ok = true;
	    ok &= parseSourceFiles(infiles);
	    ok &= checkIdentifiers();
	    if (options.definition()) {
		ok &= buildDefinitionFile();
	    }
	    if (options.inter()) {
		ok &= buildInterfaceFile();
	    }
	    if (options.keywords()) {
		ok &= buildKeywordFile();
	    }
	    return ok;
	}
    }

    /**
     * Parses the named *.t files and returns true iff successful.
     */
    private boolean parseSourceFiles(String[] infiles) {
	boolean	errorsFound = false;

	definitions = new DefinitionFile[infiles.length];
	for (int i = 0; i < infiles.length; i++) {
	    try {
		definitions[i] = DefinitionFile.read(infiles[i]);
	    } catch (LexgenError e) {
		System.err.println(e.getMessage());
		errorsFound = true;
	    }
	}

	return !errorsFound;
    }


    /**
     * Checks the identifiers in the definitions and returns true
     * iff they are OK.
     */
    private boolean checkIdentifiers() {
	Hashtable		identifiers = new Hashtable();
	boolean		errorsFound = false;
	String		prefix = definitions[definitions.length - 1].getPrefix();
	int			cnt = MIN_TOKEN_ID;

	for (int i = 0; i < definitions.length; i++) {
	    try {
		cnt = definitions[i].checkIdentifiers(identifiers, prefix, cnt);
	    } catch (LexgenError e) {
		System.err.println(e.getMessage());
		errorsFound = true;
	    }
	}

	return !errorsFound;
    }

    /**
     * Builds the ANTLR TokenTypes.txt file and returns true iff successful.
     */
    private boolean buildDefinitionFile() {
	String		vocabulary = definitions[definitions.length - 1].getVocabulary();
	String		prefix = definitions[definitions.length - 1].getPrefix();
	File		outputFile = new File(vocabulary + "TokenTypes.txt");

	try {
	    PrintWriter	out = new PrintWriter(new BufferedWriter(new FileWriter(outputFile)));

	    // !!! print file header

	    // print vocabulary name
	    out.println(vocabulary + " // output vocabulary");

	    for (int i = 0; i < definitions.length; i++) {
		definitions[i].printDefinition(out, prefix);
	    }

	    out.flush();
	    out.close();
	    return true;
	} catch (java.io.IOException e) {
	    System.err.println("I/O Exception on " + outputFile.getPath() + ": " + e.getMessage());
	    return false;
	}
    }

    /**
     * Builds the TokenTypes.java file and returns true iff successful.
     */
    private boolean buildInterfaceFile() {
	DefinitionFile	definition = definitions[definitions.length - 1];
	String		vocabulary = definition.getVocabulary();
	File		outputFile = new File(vocabulary + "TokenTypes.java");

	try {
	    PrintWriter	out = new PrintWriter(new BufferedWriter(new FileWriter(outputFile)));

	    // !!! print file header

	    definition.printInterface(out,
				      definitions.length == 1
				      ? null
				      : definitions[definitions.length - 2].getClassName());

	    out.flush();
	    out.close();
	    return true;
	} catch (java.io.IOException e) {
	    System.err.println("I/O Exception on " + outputFile.getPath() + ": " + e.getMessage());
	    return false;
	}
    }

    /**
     * Generates the keyword file that instantiates tokens for each
     * literal and keyword and that provides the hashtable lookup of
     * IDENTs.
     */
    private boolean buildKeywordFile() {
	// get the summary data for the keyword file from the definition files
	String vocabulary = definitions[definitions.length - 1].getVocabulary();
	String prefix = definitions[definitions.length - 1].getPrefix();
	String packageName = 
	    definitions[definitions.length - 1].getPackageName();
	
	// get the set of flags that tokens may have
	ArrayList legalFlags = new ArrayList();
	for (int i = 0; i < definitions.length; i++) {
	    legalFlags.addAll( definitions[i].flags() );
	}

	ArrayList accum = new ArrayList();

	ArrayList header = new ArrayList();
	ArrayList footer = new ArrayList();

	// retrieve keywords to hash
	boolean ok = true;
	for (int i = 0; i < definitions.length; i++) {
	    ok &= definitions[i].accumKeywords(accum, prefix, legalFlags);
	}
	if (!ok) {
	    System.out.println("Errors found, not generating keyword file");
	    return false;
	}
	

	// Generate the header information, and construct a keyword
	// array while we're at it.
	header.add("// Generated by lexgen");
	header.add("package " + packageName + ";");
	header.add("import org.multijava.util.compiler.CToken;\n");
	header.add("/*package*/ final class " + vocabulary + "Keywords implements " + vocabulary + "TokenTypes {");
	header.add("  public static CToken lookup(final char[] data, int offset, int length) {");
	header.add("    return tokens[find(data, offset, length) + 1];");
	header.add("  }");
	header.add("  private static final CToken[] tokens = new CToken[] {\n    null,");
	int size = accum.size();
	String[] keywords = new String[size];
	for (int i = 0; i < size; i++) {
	    DefinitionFile.KeywordTokenInfo kti = 
		(DefinitionFile.KeywordTokenInfo) accum.get(i);
	    keywords[i] = kti.keyword;
	    String line = "    new CToken(" + 
		kti.type + ", \"" + Utils.escapeString( kti.keyword ) + "\", " +
		kti.flags + ")";

	    if (i < size - 1) {
		header.add(line + ",");
	    } else {
		header.add(line);
	    }
	}
	header.add("  };");

	for (int i = 0; i < legalFlags.size(); i++) {
	    header.add("  public final static int " + legalFlags.get(i) +
			      " = " + (1 << i) + ";");
	}

	// footer
	footer.add("}");

	// generate perfect hash function
	JPerf perf = new JPerf( keywords, makeStringArray(header),
			        makeStringArray(footer));
	perf.build();

	try {
	    perf.genCode(vocabulary + "Keywords.java");
	    return true;
	} catch (java.io.IOException e) {
	    System.err.println("I/O Exception on " + vocabulary + "Keywords.java" + ": " + e.getMessage());
	    return false;
	}
    }

    /**
     * Transforms a list of strings into a string array.
     */
    private static String[] makeStringArray(ArrayList vec) {
	String[] arr = new String[vec.size()];
	vec.toArray( arr );
	return arr;
    }


    // --------------------------------------------------------------------
    // ENTRY POINT
    // --------------------------------------------------------------------

    /**
     * Entry point to the assembler
     */
    public static void main(String[] args) {
	if (new Main(args).run()) {
	    System.exit(0);
	} else {
	    System.exit(1);
	}
	
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /*
     * Parse command line arguments.
     */
    private boolean parseArguments(String[] args) {
	options = new LexgenOptions();
	return options.parseCommandLine(args);
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private LexgenOptions options;
    private DefinitionFile[] definitions;

    private static final int MIN_TOKEN_ID = 4;

    private final String[] args;
}
