/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TokenDefinition.java,v 1.8 2004/01/16 16:26:57 cclifton Exp $
 */

package org.multijava.util.lexgen;

import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Vector;
import org.multijava.util.Utils;
import java.util.List;

class TokenDefinition {

    /**
     * Token types
     */
    public static final int	LITERAL = 1;
    public static final int	KEYWORD = 2;
    public static final int	OTHER = 3;

    /**
     * Creates a new <code>TokenDefinition</code> instance.
     *
     * @param type 	the kind of token represented by this, one of 
     *			{@link #LITERAL}, {@link #KEYWORD}, {@link #OTHER}
     * @param name 	an identifier for this token
     * @param value 	the source code text matched by this token, will match
     *			<code>name</code> if this parameter is <CODE>null</CODE>
     * @param flags 	a list of {@link String Strings} giving the 
     *			flags that are true for this token
     */
    public TokenDefinition(int type, String name, String value, List flags) {
	this.type		= type;
	this.name		= name;
	this.value		= value;
	this.flags = new String[flags.size()];
	flags.toArray( this.flags );
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Check for duplicate identifiers
     * @param	identifiers	a table of all token identifiers
     * @param	prefix		the literal prefix
     * @param	id		the id of the token
     * @param	sourceFile	the file where the token is defined
     */
    public void checkIdentifiers(Hashtable identifiers, String prefix, 
				 int id, String sourceFile) throws LexgenError
    {
	String		ident  = type == LITERAL ? prefix + name : name;
	String		stored = (String)identifiers.get(ident);

	if (stored != null) {
	    throw new LexgenError(LexgenMessages.DUPLICATE_DEFINITION, 
				  new Object[] { ident, sourceFile, stored });
	}
	identifiers.put(ident, sourceFile);

	this.id = id;
    }

    /**
     * Prints the token definition to definition file (txt)
     * @param	out		the output stream
     * @param	prefix		the literal prefix
     */
    public void printDefinition(PrintWriter out, String prefix) {
	if (type == LITERAL) {
	    out.print(prefix + name);
	    out.print(" = ");
	    out.print("\"" + 
		      (value == null ? name : Utils.escapeString(value,false)) + 
		      "\"");
	} else {
	    out.print(name);
	    if (value != null) {
		out.print("(\"" + Utils.escapeString(value,false) + "\")");
	    }
	}

	out.println(" = " + id);
    }

    /**
     * Prints the token definition to interface file (java)
     * @param	out		the output stream
     * @param	prefix		the literal prefix
     */
    public void printInterface(PrintWriter out, String prefix) {
	out.print("  int\t");
	if (type == LITERAL) {
	    out.print(prefix + name);
	} else {
	    out.print(name);
	}
	out.println(" = " + id + ";");
    }

    /**
     * Adds the token information for this to <code>accum</code> if
     * this is a keyword or literal token.
     *
     * <pre><jml>
     *   requires (\forall Object o; accum.contains(o); 
     *			o instanceof DefinitionFile.KeywordTokenInfo);
     *   ensures (\forall Object o; accum.contains(o); 
     *			o instanceof DefinitionFile.KeywordTokenInfo);
     * also
     *   requires type == LITERAL || type == KEYWORD;
     *   ensures \old(accum.size() + 1) == accum.size();
     * also
     *   requires !(type == LITERAL || type == KEYWORD);
     *   ensures \old(accum.equals(accum));
     * also
     *   ensures \result != true ==> (* an error was reported to System.err *);
     * </jml></pre>
     *
     */
    boolean accumKeyword(List accum, String prefix, List legalFlags) {
	if (type == LITERAL || type == KEYWORD) {
	    int f = flagsAsInt(legalFlags);
	    accum.add(
		new DefinitionFile.KeywordTokenInfo(value == null ? name : value,
						    type == LITERAL ? 
						    prefix + name : name,
						    f ));
	    if (f == -1) {
		return false;
	    }
	} else {
	    if (flags.length != 0) {
		System.err.println(
		    LexgenMessages.FLAG_ON_OTHER.format(
			new Object[] { name } ));
		return false;
	    }
	}
	return true;
    }

    private int flagsAsInt( List legalFlags ) {
	int result = 0;
	for (int i = 0; i < flags.length; i++) {
	    int pos = legalFlags.indexOf( flags[i] );
	    if (pos == -1) {
		System.err.println(
		    LexgenMessages.UNKNOWN_FLAG.format(
			new Object[] { flags[i], name } ));
		return -1;
	    }
	    result += (1 << pos);
	}
	return result;
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------


    //@ spec_public
    private final int type;

    private final String name;

    /**
     * The text value of this token in pure string format (with escape
     * codes evaluated).  */
    private final String value;


    /**
     * An array of the token flags that are true for this token.
     */
    private final String[] flags;

    private int id;
}
