/*
 * Copyright (C) 2004 Iowa State University
 *
 * This file is part of the MultiJava project.  More information is 
 * available from www.multijava.org.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestBuilder.java,v 1.4 2004/01/16 16:26:57 cclifton Exp $
 */

package org.multijava.util.lexgen.testcase;

import java.io.*;
import java.util.Arrays;
import java.util.ArrayList;
import org.multijava.util.testing.TestCase;
import org.multijava.util.testing.Diff;
import org.multijava.util.Utils;
import java.util.HashSet;
import java.util.TreeSet;

/**
 * This utility class generates the JUnit testcases for this package.
 * The generated Java file contains test* methods for each element of
 * the <CODE>args</CODE> array passed to this class's main method.
 * The generated Java file also contains a main method that will
 * generate the expected results of each of the tests.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.4 $
 */
public class TestBuilder extends TestCase {

    // -------------------------------------------------------------
    // COMMAND-LINE ENTRY POINT
    // -------------------------------------------------------------
    
    /**
     * Generates (to stdout) the source code for a JUnit TestCase on
     * the input files from the directory named in <code>args[0]</code>.
     *
     * <pre><jml>
     * requires args != null;
     * </jml></pre>
     *
     */
    public static void main (String[] args) {
	if (args.length == 0) {
	    System.err.println( "No input files specified");
	    System.exit(1);
	}
	
	Arrays.sort(args);

	try {
	    String action = System.getProperty( "action" );
	    if (action != null && action.equals("expected")) {
		new TestBuilder(args).generateExpected();
	    } else {
		new TestBuilder(args).outputTestCaseSource();
	    }
	} catch (TestBuilderException e) {
	    System.err.println( e.getMessage() );
	    System.exit(2);
	}
    }

    // -------------------------------------------------------------
    // CONSTRUCTORS
    // -------------------------------------------------------------

    /**
     * Constructs a testcase and expected results generator.
     *
     * <pre><jml>
         requires (* property 'package' names the package for generated code *)
	       && (* property 'projClassRoot' names the project root directory
		     in a format acceptable to the local JDK *);
         ensures this.inputs != null && this.inputs.length > 0;
     * </jml></pre>
     */
    private TestBuilder(String[] inputs) throws TestBuilderException {
	super( "TestBuilder" );

	this.genPackage = System.getProperty( "package" );
	if (genPackage == null) {
	    throw new TestBuilderException( 
		"Must specify a \"package\" property" );
	}
	
	String projClassRoot = System.getProperty( "projclassroot" );
	if (projClassRoot == null) {
	    throw new TestBuilderException( 
		"Must specify a \"projClassRoot\" property" );
	}

	String workingDirName = projClassRoot + File.separatorChar +
	    genPackage.replace('.', File.separatorChar);
	    
	this.workingDir = new File( workingDirName );
	if (!(workingDir.exists() && workingDir.isDirectory())) {
	    throw new TestBuilderException(
		"The path formed from concatenating projClassRoot, package, " +
		"and testdir (i.e., " + 
		'"' + workingDir + '"' + "), must name an existing directory." );
	}

	this.inputs = inputs;

// 	try {
// 	    this.inputs = calculateInputs();
// 	} catch (IOException e) {
// 	    throw new TestBuilderException( "Unable to calculate test inputs: " +
// 					    e.getMessage() );
// 	}
    }

    public TestBuilder( String workingDirName ) {
	super( "TestBuilder" );
	this.inputs = null;
	this.genPackage = null;
	this.workingDir = new File( workingDirName );
    }

//     protected String[] calculateInputs() throws IOException {
	
// 	// get all files with appropriate suffix
// 	File[] files = workingDir.listFiles( 
// 	    new FilenameFilter() {
// 		public boolean accept( File dir, String name ) {
// 		    return name.endsWith(".t");
// 		}
// 	    });

// 	// get files to be skipped
// 	File skip = new File(workingDir, "skip");
// 	HashSet skipNames = new HashSet();
// 	if (skip.exists()) {
// 	    BufferedReader r = new BufferedReader( new FileReader( skip ) );
// 	    String line;
// 	    while ((line = r.readLine()) != null) {
// 		if (!line.startsWith("#")) {
// 		    skipNames.add(line);
// 		}
// 	    }
// 	}

// 	// remove skipped files
// 	TreeSet names = new TreeSet();
// 	for (int i = 0; i < files.length; i++) {
// 	    String n = files[i].getName();
// 	    if (!skipNames.contains(n)) {
// 		names.add(n);
// 	    }
// 	}

// 	// convert to array
// 	String[] result = new String[names.size()];
// 	names.toArray(result);
// 	return result;
//     }

    // -------------------------------------------------------------
    // TEST CASE GENERATION
    // -------------------------------------------------------------

    /**
     * Generates (to stdout) the source code for a JUnit TestCase on
     * <code>inputs</code>.
     * <pre><jml>
         requires inputs != null && inputs.length > 0;
     * </jml></pre>
     */
    private void outputTestCaseSource() throws TestBuilderException {
	PrintWriter out = 
	    new PrintWriter( 
		new BufferedWriter(
		    new OutputStreamWriter( System.out ) ) );

	String genClass = System.getProperty( "class" );
	if (genClass == null) {
	    throw new TestBuilderException( 
		"Must specify a \"class\" property" );
	}
	// header
	out.println( "// Code generated by " + this.getClass().getName() );
	out.println( "// DO NOT EDIT THIS FILE, YOUR CHANGES WILL BE LOST" );
	out.println();
	if (genPackage != null && genPackage.length() > 0) {
	    out.println( "package " + genPackage + ";" );
	}
	out.println();
	out.println( "public class " + genClass + 
		     " extends org.multijava.util.testing.TestCase {" );
	
	// constructors
	out.println( "  public " + genClass + "() {" );
	out.println( "    this( " + '"' + genClass + '"' + ");" );
	out.println( "  }" );

	out.println( "  public " + genClass + "( String name ) {" );
	out.println( "    super( name );" );
	out.println( "    this.tb = new TestBuilder( \"" +
		     Utils.escapeString(workingDir.toString(),false) + "\" );" );
	out.println( "  }" );

	// methods
	for (int i = 0; i < inputs.length; i++) {
	    String ident = getIdentFrom(inputs[i]);
	    out.println();
	    out.println( "  public void test_" + ident + "() {" );
	    out.println( "    tb.runTest( \"" + ident + "\" );" );
	    out.println( "  }" );
	}

	// fields
	out.println( "  private final TestBuilder tb;" );

	// footer
	out.println( "}" );
	out.close();
    }
    

    /**
     * Returns a Java identifier from the given testcase input name.
     */
    private String getIdentFrom( String input ) 
	throws TestBuilderException
    {
	if (!input.endsWith(".t")) {
	    throw new TestBuilderException( "All inputs must be *.t files.  " +
					    '"' + input + '"' + " is not" );
	}
	String result = input.substring(0,input.length() - 2);
	char[] resultAsArray = result.toCharArray();
	if (resultAsArray.length == 0) {
	    throw new TestBuilderException( '"' + ".t" + '"' + 
					    " is not a valid input file name " + 
					    "because it has no basename" );
	}
	boolean valid = true;
	valid &= Character.isJavaIdentifierStart(resultAsArray[0]);
	for (int i = 1; i < resultAsArray.length; i++) {
	    valid &= Character.isJavaIdentifierPart(resultAsArray[i]);
	}
	if (!valid) {
	    throw new TestBuilderException( '"' + input + '"' +
					    " is not a valid input file name " + 
					    "because its basename is not a " +
					    "valid Java identifier" );
	}
	return result;
    }

    // -------------------------------------------------------------
    // EXPECTED RESULTS GENERATION
    // -------------------------------------------------------------

    /**
     * Generates the expected results for the <code>inputs</code>.
     */
    private void generateExpected() throws TestBuilderException {
	try {
	    for (int i = 0; i < inputs.length; i++) {
		String testName = getIdentFrom( inputs[i] );
		File expected = expectedFileFor( testName );
		File oldExpected = fileFor( testName, ".t-expected.old" );

		// rename the old expected results file
		if (expected.exists()) {
		    if (!expected.renameTo(oldExpected)) {
			throw new TestBuilderException(
			    "Unable to rename " + expected );
		    }
		} else {
		    System.out.println( "No previous expected results for " + 
					testName);
		}
	    
		// get the new expected results
		String expectedString = resultFor( testName );
		writeStringToFile( expectedString, expected );

		// compare the new results to the old
		if (oldExpected.exists()) {
		    String oldExpectedString = stringFromFile( oldExpected );
		    Diff diff = new Diff( "old expected", oldExpectedString, 
					  "new expected", expectedString );
		    System.out.print(testName);
		    if (diff.areDifferent()) {
			System.out.println( " changed!");
			System.out.println( diff.result() );
		    } else {
			System.out.println( " unchanged");
			if (!oldExpected.delete()) {
			    throw new TestBuilderException(
				"Unable to delete " + oldExpected);
			}
		    }
		}
	    }
	} catch (IOException e) {
	    throw new TestBuilderException( "IOException: " + e.getMessage() );
	}
    }


    /**
     * Returns a file that should store the expected results for the
     * test of the given name.
     */
    private File expectedFileFor( String testName ) {
	return fileFor(testName, ".t-expected" );
    }

    /**
     * Returns a file that should store the input for the
     * test of the given name.
     */
    private File inputFileFor( String testName ) {
	return fileFor(testName, ".t" );
    }

    /**
     * Returns a file that should store the actual resutlts for the
     * test of the given name.
     */
    private File resultFileFor( String testName ) {
	return fileFor(testName, ".t-ckd" );
    }

    /**
     * Returns a file in the working directory whose name is the
     * concatenation of testName and suffix.
     */
    private File fileFor( String testName, String suffix ) {
	return new File( workingDir, testName + suffix );
    }

    // -------------------------------------------------------------
    // TEST CASE EXECUTION
    // -------------------------------------------------------------

    /**
     * Executes the test of the given name.
     */
    public void runTest( String testName ) {
	try {
	    // get results for named test
	    String actual = resultFor( testName );
	    
	    // compare results to expected
	    String expected = stringFromFile( expectedFileFor( testName ) );
	    Diff diff = new Diff( "expected", expected, "actual", actual );
	    if (diff.areDifferent()) {
		writeStringToFile( actual, resultFileFor( testName ) );
		fail( diff.result() );
	    }
	} catch (IOException e) {
	    fail( e.getMessage() );
	}
    }

    /**
     * Executes the named test and returns a String containg
     * the results.  Redirects stdout and stderr to generate the results.
     */
    private String resultFor( String testName ) {
	// capture standard output and error streams
	PrintStream stdout = System.out;
	PrintStream stderr = System.err;
	ByteArrayOutputStream capturedOutput = new ByteArrayOutputStream();
	PrintStream capturedStream = 
	    new PrintStream(new BufferedOutputStream( capturedOutput ));
	System.setOut(capturedStream);
	System.setErr(capturedStream);
	try {
	    // run tool on test case
	    String[] args = new String[] {
		"--keywords",
		inputFileFor(testName).toString()
	    };
	    new org.multijava.util.lexgen.Main(args).run();

	    // clean up any possible lint
	    /* !FIXME! */
	} finally {
	    // restore standard output and error streams and return
	    // captured results
	    capturedStream.close();
	    System.setOut(stdout);
	    System.setErr(stderr);
	}
	return capturedOutput.toString();
    }


    // -------------------------------------------------------------
    // HELPERS
    // -------------------------------------------------------------

    public static final String stringFromFile( File f ) throws IOException {
	BufferedReader r = new BufferedReader( new FileReader( f ) );
	StringBuffer result = new StringBuffer();
	String line;
	while ((line = r.readLine()) != null) {
	    result.append(line + NEWLINE);
	}
	r.close();
	return result.toString();
    }

    public static final void writeStringToFile( String s, File f ) 
	throws IOException
    {
	PrintWriter p = 
	    new PrintWriter( new BufferedWriter( new FileWriter( f ) ) );
	p.print(s);
	p.close();
    }

    // -------------------------------------------------------------
    // PRIVILEGED DATA MEMBERS
    // -------------------------------------------------------------
    
    /**
     * The testcase input files.
     */
    private final String[] inputs;


    /**
     * The package to which the generated test cases belong.
     */
    private final String genPackage;

    /**
     * The directory where input and output files are found.
     *
     * <pre><jml>
       private invariant workingDir != null && workingDir.exists() && 
		 workingDir.isDirectory();
     * </jml></pre>
     *
     */
    private final File workingDir;

    static final class TestBuilderException extends Exception {
	TestBuilderException( String msg ) {
	    super( msg );
	}
    }
}
