/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: DefinitionFile.java,v 1.4 2004/01/22 20:05:02 leavens Exp $
 */

package org.multijava.util.msggen;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.ArrayList;

class DefinitionFile {

  /**
   * Constructs a token definition file
   */
  public DefinitionFile(String sourceFile,
			String packageName,
			String prefix,
			ArrayList definitions) {
    this.sourceFile	= sourceFile;
    this.packageName	= packageName;
    this.prefix		= prefix == null ? DEFAULT_PREFIX : prefix;

    this.definitions	= new MessageDefinition[definitions.size()];
    definitions.toArray(this.definitions);
  }

  /**
   * Reads and parses an token definition file
   *
   * @param	sourceFile		the name of the source file
   * @return	a class info structure holding the information from the source
   *
   */
  public static DefinitionFile read(String sourceFile) throws MsggenError {
    try {
      InputStream	input = new BufferedInputStream(new FileInputStream(sourceFile));
      MsggenLexer	scanner = new MsggenLexer(input);
      MsggenParser	parser = new MsggenParser(scanner);
      DefinitionFile	defs = parser.aCompilationUnit(sourceFile);

      input.close();

      return defs;
    } catch (java.io.FileNotFoundException e) {
      throw new MsggenError(MsggenMessages.UNKNOWN_FILE, sourceFile);
    } catch (java.io.IOException e) {
      throw new MsggenError(MsggenMessages.IO_EXCEPTION, sourceFile, e.getMessage());
    } catch (antlr.RecognitionException e) {
      throw new MsggenError(MsggenMessages.PARSING_ERROR,
			    sourceFile,
			    e.getLine() + " : " + e.getMessage());
    } catch (antlr.TokenStreamException e) {
      throw new MsggenError(MsggenMessages.LEXING_ERROR,
			    sourceFile,
			    e.getMessage());
    }

  }

  // --------------------------------------------------------------------
  // ACCESSORS
  // --------------------------------------------------------------------

  /**
   * Check for duplicate identifiers
   * @param	identifiers	a table of all token identifiers
   * @return	the id of the last token + 1
   */
  public void checkIdentifiers(Hashtable identifiers) throws MsggenError {
    for (int i = 0; i < definitions.length; i++) {
      definitions[i].checkIdentifiers(identifiers, sourceFile);
    }
  }

  /**
   * Prints the token definition to interface file (java)
   * @param	out		the output stream
   */
  public void printFile(PrintWriter out) {
    out.print("// Generated from " + sourceFile);
    out.println();
    out.println("package " + packageName + ";");
    out.println();
    out.println("import org.multijava.util.MessageDescription;");
    out.println();
    out.print("public class " + prefix + "Messages");
    out.println(" {");

    for (int i = 0; i < definitions.length; i++) {
      definitions[i].printInterface(out, this.prefix);
    }

    out.println("}");
  }

  /**
   * Returns the package name
   */
  public String getClassName() {
    return packageName + "." + prefix + "Messages";
  }

  /**
   * Returns the package name
   */
  public String getPackageName() {
    return packageName;
  }

  /**
   * Returns the literal prefix
   */
  public String getPrefix() {
    return prefix;
  }

  // --------------------------------------------------------------------
  // PRIVATE DATA MEMBERS
  // --------------------------------------------------------------------

  private static final String		DEFAULT_PREFIX = "";

  private final String			sourceFile;
  private final String			packageName;
  private final String			prefix;
  private final MessageDefinition[]	definitions;
}
