/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: DefinitionFile.java,v 1.19 2004/07/08 01:08:49 netgecko Exp $
 */

package org.multijava.util.optgen;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.ArrayList;

class DefinitionFile implements SelectionVariables {

    /**
     * Constructs a  definition file
     */
    public DefinitionFile(String sourceFile,
                          String packageName,
                          String parent,
                          String implement,
                          String prefix,
                          String version,
                          String usage,
                          String url,
                          ArrayList definitions) {
        
        this.sourceFile         = sourceFile;
        this.packageName        = packageName;
        this.parent             = parent;
        this.implement          = implement;
        this.version            = version;
        this.usage              = usage;
        this.prefix             = prefix;
        this.url                = url;
        this.definitions        = new OptionDefinition[definitions.size()];
        definitions.toArray(this.definitions);
    }

    /**
     * Reads and parses an token definition file
     *
     * @param   sourceFile              the name of the source file
     * @return  a class info structure holding the information from the source
     *
     */
    public static DefinitionFile read(String sourceFile) throws OptgenError {
        try {
            InputStream input = new BufferedInputStream(new FileInputStream(sourceFile));
            OptgenLexer scanner = new OptgenLexer(input);
            OptgenParser        parser = new OptgenParser(scanner);
            DefinitionFile      defs = parser.aCompilationUnit(sourceFile);

            input.close();

            return defs;
        } catch (java.io.FileNotFoundException e) {
            throw new OptgenError(OptgenMessages.UNKNOWN_FILE, sourceFile);
        } catch (java.io.IOException e) {
            throw new OptgenError(OptgenMessages.IO_EXCEPTION, sourceFile, 
                                  e.getMessage());
        } catch (antlr.RecognitionException e) {
            throw new OptgenError(OptgenMessages.PARSING_ERROR,
                                  sourceFile,
                                  e.getLine() + " : " + e.getMessage());
        } catch (antlr.TokenStreamException e) {
            throw new OptgenError(OptgenMessages.LEXING_ERROR,
                                  sourceFile, e.getMessage());
        }

    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Check for duplicate identifiers
     * @param   identifiers     a table of all token identifiers
     * @return  the id of the last token + 1
     */
    public void checkIdentifiers(Hashtable identifiers) throws OptgenError {
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].checkIdentifiers(identifiers, sourceFile);
        }
    }

    /**
     * Check for duplicate identifiers
     * @param   identifiers     a table of all token identifiers
     * @return  the id of the last token + 1
     */
    public void checkShortcuts(Hashtable identifiers) throws OptgenError {
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].checkShortcuts(identifiers, sourceFile);
        }
    }

    /**
     * Prints the token definition to interface file (java)
     * @param   out             the output stream
     */
    public void printFile(PrintWriter out) {
        out.print("// Generated from " + sourceFile);
        out.println();
        out.println("package " + packageName + ";");
        out.println();
        out.println("import gnu.getopt.Getopt;");
        out.println("import gnu.getopt.LongOpt;");
        out.println("import java.util.ArrayList;");
        out.println("import java.util.Hashtable;");
        out.println("import java.util.HashMap;");
        out.println("import java.util.LinkedHashSet;");
        out.println();
        out.println("/** This class is automatically generated from " + 
                    sourceFile);
        out.println(" *  and contains member fields corresponding to " + 
                    "command-line options.");
        out.println(" */");
        out.print("public class " + prefix + "Options");
        out.print(parent == null ? "" : " extends " + parent);
        out.print(implement == null ? "" : " implements " + implement);
        out.println(" {");

        // CONSTRUCTORS
        out.println();
        out.println("  public " + prefix + "Options(String name) {");
        out.println("    super(name);");
        out.println("  }");
        out.println();
        out.println("  public " + prefix + "Options() {");
        out.println("    this(\"" + prefix + "\");");
        out.println("  }");

        // FIELDS
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printFieldAndAccessors(out);
        }
        
        // PROCESSOPTION
        out.println();
        out.println("  public boolean processOption(int code, Getopt g) {");
        out.println("    switch (code) {");
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printParseArgument(out);
        }
        out.println("    default:");
        out.println("      return super.processOption(code, g);");
        out.println("    }");
        if (definitions.length > 0) {
            out.println("    return true;");
        }
        out.println("  }");
        
        // SETOPTION
        out.println();
        out.println("  public boolean setOption(String name, Object newValue) {");
        out.print("    ");
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printSetOption(out);
        }
        out.println("{");
        out.println("      return super.setOption(name, newValue);");
        out.println("    }");
        if (definitions.length > 0) {
            out.println("    return true;");
        }
        out.println("  }");
        
        // GETOPTIONS
        out.println();
        out.println("  public HashMap getOptions() {");
        out.println("    HashMap result = super.getOptions();");
        for (int i = 0; i < definitions.length; i++) {
            out.print("    result.put(");
            definitions[i].printLongname(out);
            out.print(", ");
            String dashes;
            if (prefix.equals("Javadoc")) {
                dashes = "\"  -";
            } else {
                dashes = "\"  --";
            }
            definitions[i].printUsage(out, dashes);
            out.println(");");
        }
        out.println("    ");
        out.println("    return result;");
        out.println("  }");
        
        // GETLONGNAME
        out.println("");
        out.println("");
        out.println("  public LinkedHashSet getLongname() {");
        out.println("    LinkedHashSet longname = super.getLongname();");
        for (int i = 0; i < definitions.length; i++) {
            out.print("    longname.add(");
            definitions[i].printLongname(out);
            out.println(");");
        }
        out.println("");
        out.println("    return longname;");
        out.println("  }");
        
        // GETTYPE
        out.println("");
        out.println("  public Hashtable getType() {");
        out.println("    Hashtable type = super.getType();");
        for (int i = 0; i < definitions.length; i++) {
            out.print("    type.put(");
            definitions[i].printLongname(out);
            out.print(", ");
            definitions[i].printType(out);
            out.println(");");
        }
        out.println("");
        out.println("    return type;");
        out.println("  }");

        // GETDEFAULTVALUE
        out.println("");
        out.println("");
        out.println("  public Hashtable getDefaultValue() {");
        out.println("    Hashtable defaultValue = super.getDefaultValue();");
        for (int i = 0; i < definitions.length; i++) {
            out.print("    defaultValue.put(");
            definitions[i].printLongname(out);
            out.print(", ");
            definitions[i].printDefaultValue(out);
            out.println(");");
        }
        out.println("");
        out.println("    return defaultValue;");
        out.println("  }");
                
        // GETCURRENTVALUE
        out.println("");
        out.println("");
        out.println("  public Hashtable getCurrentValue() {");
        out.println("    Hashtable currentValue = super.getCurrentValue();");
        for (int i = 0; i < definitions.length; i++) {
            out.print("    currentValue.put(");
            definitions[i].printLongname(out);
            out.print(", ");
            definitions[i].printCurrentValue(out);
            out.println(");");
        }
        out.println("");
        out.println("    return currentValue;");
        out.println("  }");
                
        // GETTABLEHEADER
        out.println("");
        out.println("");
        out.println("  public Hashtable getTableHeader() {");
        out.println("    Hashtable tableHeader = super.getTableHeader();");
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printTableHeader(out);
        }
        out.println("");
        out.println("    return tableHeader;");
        out.println("  }");

        // GETSELECTION
        out.println("");
        out.println("");
        out.println("  public Hashtable getSelectionList() {");
        out.print("    Hashtable");
        out.println(" " + structureName + " = super.getSelectionList();");
        out.println("    "+ selectionType + " " + selectionName + ";");
        for (int i = 0; i < definitions.length; i++) {
             definitions[i].printSelection(out);
        }
        out.println("");
        out.println("    return " + structureName + ";");
        out.println("  }");
        
        // GETHELPSTRINGS
        out.println("");
        out.println("");
        out.println("  public Hashtable getHelpString() {");
        out.println("    Hashtable helpString = super.getHelpString();");
        for (int i = 0; i < definitions.length; i++) {
            out.print("    helpString.put(");
            definitions[i].printLongname(out);
            out.print(", getOptions().get(");
                        definitions[i].printLongname(out);
            out.println("));");
        }
        out.println("");
        out.println("    return helpString;");
        out.println("  }");
        
        //GETGUITYPE
        out.println("");
        out.println("");
        out.println("  public Hashtable getGuiType() {");
        out.println("    Hashtable guiType = super.getGuiType();");
        for (int i = 0; i < definitions.length; i++) {
            out.print("    guiType.put(");
            definitions[i].printLongname(out);
            out.print(", ");
            definitions[i].printGuiType(out);
            out.println(");");
        }
        out.println("");
        out.println("    return guiType;");
        out.println("  }");

        // GETSHORTOPTIONS
        out.println("");
        out.println("");
        out.println("  public String getShortOptions() {");
        out.print("    return \"");
        for (int i = 0; i < definitions.length; i++) {
            definitions[i].printShortOption(out);
        }
        out.println("\" + super.getShortOptions();");
        out.println("  }");

        // VERSION
        if (version != null) {
            out.println("");
            out.println("");
            out.println("  public String version() {");
            out.print("    return ");
            out.print(version);
            out.println(";");
            out.println("  }");
        }

        // USAGE
        out.println("");
        out.println("");
        out.println("  public void usage() {");
        out.print("    System.err.println(");
        out.print(usage);
        out.println(");");
        out.print("    printVersion();");
        out.println("  }");

        // HELP
        out.println("");
        out.println("");
        out.println("  public void help() {");
        out.print("    System.err.println(");
        out.print(usage);
        out.println(");");
        out.println("    printOptions();");
        out.println("    System.err.println();");
        
        // URL
        if (url != null) {
            out.print("    System.err.println(");
            out.print(url);
            out.println(");");
        }

        out.print("    System.err.println(");
        out.print("\"This program includes some code from the Kopi project: http://www.dms.at/kopi\"");
        out.println(");");
        out.print("    System.err.println(");
        out.print("\"Some code for this program was generated with ANTLR: http://www.antlr.org\"");
        out.println(");");
        out.println("    printVersion();");
        out.println("  }");

        // GETLONGOPTIONS
        out.println();
        out.println("  public LongOpt[] getLongOptions() {");
        out.println("    LongOpt[]      parent = super.getLongOptions();");
        out.println("    LongOpt[]      total = new LongOpt[parent.length + LONGOPTS.length];");
        out.println("    ");
        out.println("    System.arraycopy(parent, 0, total, 0, parent.length);");
        out.println("    System.arraycopy(LONGOPTS, 0, total, parent.length, LONGOPTS.length);");
        out.println("    ");
        out.println("    return total;");
        out.println("  }");

        // LONGOPTS
        out.println();
        out.println("  private static final LongOpt[] LONGOPTS = {");
        for (int i = 0; i < definitions.length; i++) {
            if (i != 0) {
                out.println(",");
            }
            definitions[i].printLongOpts(out);
        }
        out.println();
        out.println("  };");

        out.println("}");
    }

    /**
     * Returns the package name
     */
    public String getClassName() {
        return packageName + "." + prefix + "Options";
    }

    /**
     * Returns the package name
     */
    public String getPackageName() {
        return packageName;
    }

    /**
     * Returns the literal prefix
     */
    public String getPrefix() {
        return prefix;
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private final String             sourceFile;
    private final String             packageName;
    private final String             parent;
    private final String             implement;
    private final String             prefix;
    private final String             version;
    private final String             usage;
    private final String             url;
    private final OptionDefinition[] definitions;
}
