/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.5 2004/01/16 16:26:57 cclifton Exp $
 */

package org.multijava.util.optgen;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Hashtable;

/**
 * This class is the entry point for the option generator.
 */
public class Main {

    // --------------------------------------------------------------------
    // CONSTRUCTORS
    // --------------------------------------------------------------------

    /**
     * Only main can construct Main
     */
    private Main(String[] args) {
	parseArgs(args);

	for (int i = 0; i < sourceFiles.length; i++) {
	    parseSourceFiles(sourceFiles[i]);
	    checkIdentifiers();
	    checkShortcuts();
	    buildInterfaceFile();
	}

	System.exit(0);
    }

    /**
     * Sets <CODE>this.definition</CODE> to a new
     * <CODE>DefinitionFile</CODE> instance formed by parsing the
     * named <CODE>sourceFile</CODE>.  */
    private void parseSourceFiles(String sourceFile) {
	boolean		errorsFound = false;

	try {
	    definition = DefinitionFile.read(sourceFile);
	} catch (OptgenError e) {
	    System.err.println(e.getMessage());
	    errorsFound = true;
	}

	if (errorsFound) {
	    System.exit(1);
	}
    }


    /**
     * Checks for unique option identifiers in
     * <CODE>this.definition</CODE>.  */
    private void checkIdentifiers() {
	Hashtable		identifiers = new Hashtable();
	boolean		errorsFound = false;

	try {
	    definition.checkIdentifiers(identifiers);
	} catch (OptgenError e) {
	    System.err.println(e.getMessage());
	    errorsFound = true;
	}

	if (errorsFound) {
	    System.exit(1);
	}
    }

    /**
     * Checks for unique shortcuts in <CODE>this.definition</CODE>.
     * */
    private void checkShortcuts() {
	Hashtable		identifiers = new Hashtable();
	boolean		errorsFound = false;

	try {
	    definition.checkShortcuts(identifiers);
	} catch (OptgenError e) {
	    System.err.println(e.getMessage());
	    errorsFound = true;
	}

	if (errorsFound) {
	    System.exit(1);
	}
    }

    /**
     * Generates the source code and writes it to disk for the option
     * definitions in * <CODE>this.definition</CODE>.  */
    private void buildInterfaceFile() {
	String		prefix = definition.getPrefix();
	File		outputFile = new File(prefix + "Options.java");

	try {
	    PrintWriter	out = new PrintWriter(new BufferedWriter(
						  new FileWriter(outputFile)));

	    definition.printFile(out);

	    out.flush();
	    out.close();
	} catch (java.io.IOException e) {
	    System.err.println("I/O Exception on " + outputFile.getPath() + ": " + e.getMessage());
	    System.exit(1);
	}
    }

    // --------------------------------------------------------------------
    // ENTRY POINT
    // --------------------------------------------------------------------

    /**
     * Entry point to the assembler
     */
    public static void main(String[] args) {
	new Main(args);
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /*
     * Parse command line arguments.
     */
    private void parseArgs(String[] args) {
	if (args.length == 0 || 
	    args[0].equals("--help") || 
	    args[0].equals("-h")) {
	    if (args.length == 0) {
		System.err.println(USAGE);
	    } else {
		System.err.println(HELP);
	    }
	    
	    this.sourceFiles = new String[0];
	    return;
	}
	
	this.sourceFiles = new String[args.length];
	System.arraycopy(args, 0, 
			 this.sourceFiles, 0, 
			 this.sourceFiles.length);
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private static final String USAGE = 
	"usage: org.multijava.util.optgen.Main [--help] <opt-files>";
    private static final String HELP = USAGE + "\n" +
	"\n" +
	"The file org/multijava/util/optgen/Optgen.g gives the format of the opt-files\n" +
	"\n" +
	"  --help, -h:\tDisplays the help information\n" +
	"\n" +
	"This program is part of the MultiJava project: http://www.multijava.org\n" +
	"This program is based in part on the Kopi project: http://www.dms.at/kopi\n" +
	"Code in the project is generated using ANTLR: http://www.antlr.org\n";


    private String[]		sourceFiles;
    private DefinitionFile	definition;
}
