/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Optgen.g,v 1.13 2004/05/17 05:34:27 netgecko Exp $
 */

// Import the necessary classes
header { package org.multijava.util.optgen; }

{
	import java.util.ArrayList;
}


//-----------------------------------------------------------------------------
// THE PARSER STARTS HERE
//-----------------------------------------------------------------------------

class OptgenParser extends Parser;

options {
	k = 2;				// two token lookahead
	exportVocab=Optgen;			// call the vocabulary
	codeGenMakeSwitchThreshold = 2;	// Some optimizations
	codeGenBitsetTestThreshold = 3;
	defaultErrorHandler = false;		// Don't generate parser error handlers
}

aCompilationUnit [String sourceFile]
returns [DefinitionFile self]
{
	String		prefix = null;
	String		parent = null;
	String      implement = null;
	String		packageName = null;
	String		version = null;
	String		usage = null;
	String		url = null;
	ArrayList	definitions = new ArrayList();
	OptionDefinition	def;
}
	:

		// First a prefix to for the generated file, i.e. <prefix>Options.java
		"prefix" prefix = aIdentifier[]

		// The fully qualified name of the class extended by the option file
		( "parent" parent = aName[] ) ?
		
		// Name of an interface for this file to implement
		( "implement" implement = aName[] )?

		// The version information string for the utility that the
		// options are for.  Defaults to the same version as this
		// MultiJava release (see org.multijava.util.Options)
		( "version" version = aString[] ) ?

		// A string giving usage information, printed when the utility
		// is invoked with illegal arguments.  Printed at the top of
		// the output when the utility is invoked with the --help
		// option.
		( "usage" usage = aString[] ) ?

		// A string giving URL reference for the utility or project,
		// printed near the end of the --help output.
		( "url" url = aString[] ) ?

		// The package to which the generated class should belong
		( "package" packageName = aName[] )?

		// The legal options for the utility
		(
			def = aOptionDefinition[]
			{ definitions.add(def); }
		)*
		{
			self = new DefinitionFile(sourceFile, packageName, parent, 
				implement, prefix, version, usage, url, definitions); 
		}
	;

aIdentifier []
returns [String self = null]
	:
		t : IDENT
		{ self = t.getText(); }
	;

aName []
returns [String self]
{
	String	ident;
}
	:
		ident = aIdentifier[]
		{ self = ident; }
		(
			DOT ident = aIdentifier[]
			{ self += "." + ident; }
		)*
	;

aString []
returns [String self]
	:
		t : STRING
		{ self = t.getText(); }
	;

aStringArray[]
returns [ArrayList self]
{
	String val = "";
	String val2 = "";
}
	:
		val = aString[]
		{
			self = new ArrayList();
		    self.add(val);
		}
		(
			COMMA val2 = aString[]
			{ self.add(val2); }
		)*
		
	;

aInt []
returns [int self]
	:
		t : INT
		{ self = (new Integer(t.getText())).intValue(); }
	;

aOptionDefinition []
returns [OptionDefinition self]
{
	String      longname;
	String      shortname = null;
	String      action = null;
	String      type = null;
	String      defaultValue = null;
	String      argument = null;
	ArrayList	tableHeader = null;
	ArrayList   selection = null;
	ArrayList   help = null;
    String      guiType = null;
	int         caseValue = 0;
}
	:
		// The long name of the option, users prefix this with "--" on
		// the command line when invoking the utility
		"longname"	longname = aString[]

		// A one-character shortcut for the option, users prefix this
		// with "-" on the command line when invoking the utility
		( "shortcut"	shortname = aString[] |
		  "casevalue"   caseValue = aInt[]
		) 

		// The action to do.
		( "action" 	action = aString[] ) ?

		// The type of the option in the generated code
		( "type"	type = aOptionType[]

		// The default value for the option if the user does not
		// specify the option on the command line.
		"default"	defaultValue = aString[]
		)?

		// If the user specifies the option, the following term
		// controls arguments to the option.
		( 
			// user may omit option argument, this value will be used
			"optionalDefault" argument = aString[] 
		| 
			// user must give option argument
			"requireArgument" { argument = ""; }
		| 
			// option does not take an argument
			/* empty */ 
		)
		
		// A list of table headers for options with more than one
		// argument
		(
			"tableHeader" tableHeader = aStringArray[]
		)?
		
		// A list of values that are possible for this option
		(
			"selection" selection = aStringArray[]
		)?

		// A description of the option, including the format of how the 
		// option should be structured and a description of the option
		"help"	help = aStringArray[]

        // Indicates the GUI type for this option
        (
            "normal"      { guiType = "normal"; }
        |
            "advanced"    { guiType = "advanced"; }
		|
            "javadocNormal"    { guiType = "javadocNormal"; }
		|
            "javadocAdvanced"    { guiType = "javadocAdvanced"; }
		|
            "javadocNo"    { guiType = "javadocNo"; }
        )?

        { self = new OptionDefinition(longname, shortname, action,
            type, defaultValue, argument, tableHeader, selection, help, 
			caseValue, guiType ); 
		}
	;

aOptionType []
returns [String self]
	:
		"int" { self = "int"; }
	|
		"boolean" { self = "boolean"; }
	|
		"String" { self = "String"; }
	|
		"StringList" { self = "StringList"; }
	|
	    "StringSystemProperty" { self = "StringSystemProperty"; }
	;

//-----------------------------------------------------------------------------
// THE SCANNER STARTS HERE
//-----------------------------------------------------------------------------

class OptgenLexer extends Lexer;

options {
	importVocab=Optgen;       // call the vocabulary
	testLiterals=false;    // don't automatically test for literals
	k=2;                   // 2 characters of lookahead
}

WS :
		(
			' '
		| '\t'
		| '\f'
		| '\r' ('\n')?	{ newline(); }
		| '\n'		{ newline(); }
		)
		{ _ttype = Token.SKIP; }
	;

DOT		: '.';
SHARP		: '#';
COMMA		: ',';

// Single-line comments
SL_COMMENT :
		"//"
		(~('\n'|'\r'))* ('\n'|'\r'('\n')?)
		{ _ttype = Token.SKIP; newline(); }
	;

// multiple-line comments
ML_COMMENT :
		"/*"
		(
			'\n' { newline(); }
		| '*' ~'/'
		| ~'*'
		)*
		"*/"
		{ _ttype = Token.SKIP; }
	;

STRING :
		'"' (ESC|~('"'|'\\'))* '"'
	;
	
INT:
		('0' .. '9')+
	;

protected
ESC :
		'\\'
		(
		  'n'
		| 'r'
		| 't'
		| 'b'
		| 'f'
		| '"'
		| '\''
		| '\\'
		| ('0'..'3') OCT_DIGIT OCT_DIGIT
		| 'u' HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT
		)
	;

protected
OCT_DIGIT :
		'0'..'7'
	;

protected
HEX_DIGIT :
		'0'..'9'
	| 'A'..'F'
	| 'a'..'f'
	;

// an identifier.  Note that testLiterals is set to true!  This means
// that after we match the rule, we look in the literals table to see
// if it's a literal or really an identifer
IDENT
options { testLiterals=true; }
	:
		('a'..'z'|'A'..'Z'|'_'|'$') ('a'..'z'|'A'..'Z'|'_'|'0'..'9'|'$')*
	;

// an identifier preceded by a '@' Note that testLiterals is set to true!  This means
// that after we match the rule, we look in the literals table to see
// if it's a literal or really an identifer
DUMMY
options { testLiterals=true; }
	:
		'@' IDENT
	;

// a dummy rule to force vocabulary to be all characters except special
// ones that ANTLR uses internally (0 to 2)
protected
VOCAB :
		'\3'..'\377'
	;
