/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * based in part on work:
 *
 * Copyright (C) 1990-99 DMS Decision Management Systems Ges.m.b.H.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: OptionDefinition.java,v 1.19 2004/06/12 14:43:37 leavens Exp $
 */

package org.multijava.util.optgen;

import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.ArrayList;

class OptionDefinition implements SelectionVariables {

    /**
     * Constructs an option definition
     */
    public OptionDefinition(String longname,
                            String shortname,
                            String action,
                            String type,
                            String defaultValue,
                            String argument,
                            ArrayList tableHeader,
                            ArrayList selection,
                            ArrayList help,
                            int    caseValue,
                            String guiType)
    {
        this.longname = trail(longname);
        this.shortname = trail(shortname);
        this.action = trail(action);
        this.type = type;
        this.defaultValue = trail(defaultValue);
        this.argument = trail(argument);
        this.tableHeader = tableHeader;
        this.selection = selection;
        this.help = trailArrayList(help);
        this.caseValue = caseValue;
        this.guiType = guiType;
    }

    private static final String trail(String s) {
        // strip leading and trailing quotes
        if (s == null) {
            return null;
        } else if (s.length() < 2) {
            return s;
        } else {
            return s.substring(1, s.length() - 1);
        }
    }
        
    private static final ArrayList trailArrayList(ArrayList a) {
        // strip leading and trailing quotes from array list
        if (a == null) return null;
        for (int i = 0; i < a.size(); i++) {
            String ele = trail((String)(a.get(i)));
            a.remove(i);
            a.add(i, ele);
        }
        return a;
    }

    // --------------------------------------------------------------------
    // ACCESSORS
    // --------------------------------------------------------------------

    /**
     * Check for duplicate identifiers
     * @param   identifiers     a table of all token identifiers
     * @param   sourceFile      the file where the token is defined
     */
    public void checkIdentifiers(Hashtable identifiers, String sourceFile)
        throws OptgenError
    {
        String          stored = (String)identifiers.get(longname);

        if (stored != null) {
            throw new OptgenError(OptgenMessages.DUPLICATE_DEFINITION, new Object[] { longname, sourceFile, stored });
        }

        identifiers.put(longname, sourceFile);
    }

    /**
     * Check for duplicate identifiers
     * @param   shortcuts     a table of all token identifiers
     * @param   sourceFile      the file where the token is defined
     */
    public void checkShortcuts(Hashtable shortcuts, String sourceFile)
        throws OptgenError
    {
        if (shortname == null) return;
        String stored = (String)shortcuts.get(shortname);

        if (stored != null) {
            throw new OptgenError(OptgenMessages.DUPLICATE_SHORTCUT, 
                                  new Object[] { shortname, sourceFile, 
                                                 stored });
        }

        shortcuts.put(shortname, sourceFile);
    }

    /**
     * Prints the parse argument case statement to interface file (java)
     * @param   out             the output stream
     */
    public void printParseArgument(PrintWriter out) {
        out.print("    case ");
        out.print( shortname!=null ? 
            ("\'" + shortname +"\'") : 
            (new Integer(caseValue)).toString());
        out.println(":");
    
        out.print("      ");
        if (action != null) {
            out.print(action);
        } else {
        out.print(longname);
        out.print(" = ");
        if (type.equals("boolean")) {
            out.print("!" + defaultValue);
            out.print(";");
        } else {
            String      methodName;
            String      argument = this.argument;

            if (type.equals("int")) {
                methodName = "getInt";
                if (argument == null || argument.equals("")) {
                    argument = "0";
                }
            } else {
                methodName = "getString";
                argument = "\"" + argument + "\"";
            }

            out.print(methodName + "(g, " + argument + ")");
            out.print(";");
        }
        }
        out.println(" break;");
    }
        
    /**
     * Prints the set field case statement to interface file (java)
     * @param   out             the output stream
     */
    public void printSetOption(PrintWriter out) {
        if (type == null) return;
        out.println("if (name.equals(\"" + longname + "\")) {");
        if (type.equals("boolean")) {
            out.println("      set_" + longname
                        + "(((Boolean)newValue).booleanValue());");
        } else if (type.equals("int")) {
            out.println("      set_" + longname
                        + "(Integer.parseInt((String)newValue));");
        } else if (type.equals("String")
                   || type.equals("StringSystemProperty")) {
            out.println("      set_" + longname
                        + "(getString((" + getType() + ")newValue));");
        } else if (type.equals("StringList")) {
            out.println("      set_" + longname
                        + "((" + getType() + ")newValue);");
        }
        out.print("    } else ");
    }
        
    /**
     * Prints the source-code for this option's field declaration and
     * accessor method.
     * @param   out             the output stream
     */
    public void printFieldAndAccessors(PrintWriter out) {
        if (type == null) return;
        String typeName = getType();
        out.print("  private ");
        out.print(typeName);
        out.print(" ");
        out.print(longname);
        out.print(" = ");
        if (defaultValue != null && !defaultValue.equals("null")) {
            if (type.equals("String")) {
                out.print("\"" + defaultValue + "\"");
            } else if (type.equals("StringSystemProperty")) {
                out.print("getSystemProperty(\"" + defaultValue + "\")");
            } else {
                out.print(defaultValue);
            }
        } else {
            out.print("null");
        }
        out.println(";");

        // the getter method
        out.print("  public /*@ pure @*/ ");
        out.print(typeName);
        out.print(" "); 
        out.print(longname); 
        out.print("() { ");
        out.print("return "); 
        out.print(longname); 
        out.println("; }");

        // the setter method
        out.print("  public ");
        out.print(typeName);
        out.print(" set_"); 
        out.print(longname); 
        out.print("("); 
        out.print(typeName); 
        out.print(" "); 
        out.print(longname);
        out.println(") { ");
        out.print("    return "); 
        out.print("this."); 
        out.print(longname); 
        out.print(" = "); 
        out.print(longname); 
        out.println(";");
        out.println("  }");

        out.println("");
    }

    /**
     * Prints the token definition to interface file (java)
     * @param   out             the output stream
     * @param   dashes                  the dashes to put in fron of the option
     */
    public void printUsage(PrintWriter out, String dashes) {
        String prefix = dashes + longname;
        if (shortname != null) {
            prefix += ", -";
            prefix += shortname;
        }
        if (argument != null) {
            prefix += " <" + help.get(0) + ">";
        }
        prefix += ": ";
        for (int i = prefix.length(); i <= 40; i++) {
            prefix += " ";
        }
        out.print(prefix + help.get(1));
                
        if (defaultValue != null && !defaultValue.equals("null")) {
            out.print(" [");
            out.print(defaultValue);
            out.print("]");
        }
        out.print("\"");
    }

    /**
     * Prints the longname string to interface file (java)
     * @param   out             the output stream
     */
    public void printLongname(PrintWriter out) {
        out.print("\"");
        out.print(longname);
        out.print("\"");
    }
    
    /**
     * Prints the option type strings to interface file (java)
     * @param   out             the output stream
     */
    public void printType(PrintWriter out) {
        out.print("\"");
        if (type != null && type.equals("StringList")) {
            out.print(type);
        } else {
            out.print(getType());
        }
        out.print("\"");
    }
    
    /**
     * Prints the option default values to interface file (java)
     * @param   out             the output stream
     */
    public void printDefaultValue(PrintWriter out) {
        if (type != null) {
            if (type.equals("boolean")) {
                out.print("Boolean.valueOf(" + defaultValue + ")");
            } else if (type.equals("int")) {
                out.print("new Integer(" + defaultValue + ")");
            } else if (type.equals("String")) {
                if (defaultValue != null && !defaultValue.equals("null")) {
                    out.print("\"");
                    out.print(defaultValue);
                    out.print("\"");
                } else {
                    out.print("\"\"");
                }
            } else if (type.equals("StringSystemProperty")) {
                if (defaultValue != null && !defaultValue.equals("null")) {
                    out.print("getSystemProperty(\"" + defaultValue + "\")");
                } else {
                    out.print("\"\"");
                }
            } else if (type.equals("StringList")) {
                if (defaultValue != null && !defaultValue.equals("null")) {
                    out.print(defaultValue);
                } else {
                    out.print("new String[0][0]");
                }
            }
        } else {
            out.print("\"\"");
        }
    }
        
    /**
     * Prints the current value to interface file (java)
     * @param   out             the output stream 
     */
    public void printCurrentValue(PrintWriter out) {
        if (type != null) {
            if (type.equals("boolean")) {
                out.print("Boolean.valueOf(");
            } else if (type.equals("int")) {
                out.print("new Integer(");
            } else if (type.equals("StringList")) {
                out.print("getNonNullStringArray(");
            } else {
                out.print("getNonNullString(");
            }
            out.print(longname);
            out.print(")");
        } else {
            out.print("\"\"");
        }
    }
    
    /**
     * Prints the table header of the option to interface file (java)
     * @param   out             the output stream
     */
    public void printTableHeader(PrintWriter out) {
        if ((tableHeader != null)  && (tableHeader.size() == 1 
                                       || tableHeader.size() == 2)) {
            out.print("    tableHeader.put(");
            out.print("\"" + longname + "\", ");
            if (tableHeader.size() == 1) {
                out.print("formOneColumnArray(");
                out.print(tableHeader.get(0));
            } else {
                out.print("formTwoColumnArray(");
                out.print(tableHeader.get(0) + ", ");
                out.print(tableHeader.get(1));
            }
            out.println("));");
        }
    }
        
    /**
     * Prints the selection of the option to interface file (java)
     * @param   out             the output stream
     */
    public void printSelection(PrintWriter out) {
        if (selection != null) {
            out.println("      " + selectionName + " = new " 
                                 + selectionType + "();");
            for (int i = 0; i < selection.size(); i++) {
             out.print("      " + selectionName);
             out.print(".add(");
             out.print(selection.get(i));
             out.println(");");
            }
            out.print("    " + structureName + ".put(");
            out.print("\"" + longname + "\", ");
            out.println(selectionName + ");");
        }
    }

    /**
     * Prints the help strings to interface file (java)
     * @param   out             the output stream
     */
    public void printHelpString(PrintWriter out) {
        out.print("\"");
        out.print(help);
        out.print("\"");
    }
    
    /**
     * Prints the value of isNormal to interface file (java)
     * @param   out             the output stream
     */
    public void printGuiType(PrintWriter out) {
        out.print("\"");
        out.print(guiType);
        out.print("\"");
    }

    public void printLongOpts(PrintWriter out) {
        out.print("    new LongOpt(\"");
        out.print(longname);
        out.print("\", ");
        if (argument == null) {
            out.print("LongOpt.NO_ARGUMENT");
        } else if (argument.equals("")) {
            out.print("LongOpt.REQUIRED_ARGUMENT");
        } else {
            out.print("LongOpt.OPTIONAL_ARGUMENT");
        }
        out.print(", null, ");
        out.print( shortname != null ? 
                   ("\'" + shortname +"\'") : 
                   (new Integer(caseValue)).toString());    
        out.print(")");

   }

    public void printShortOption(PrintWriter out) {
        if (shortname == null) return;
        out.print(shortname);
        if (argument != null) {
            if (argument.equals("")) {
                out.print(":");
            } else {
                out.print("::");
            }
        }
    }
        
    // --------------------------------------------------------------------
    // HELPER METHODS
    // --------------------------------------------------------------------
    
    /**
     *  Returns the correct type, based on whether it is of the type
     *  StringList, StringSystemProperty, or not
     */
    private String getType() {
        if (type == null) {
            return "";
        } else if (type.equals("StringList")) {
            return "String[][]";
        } else if (type.equals("StringSystemProperty")) {
            return "String";
        } else {
            return type;
        }
    }

    // --------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // --------------------------------------------------------------------

    private final String    longname;
    private final String    shortname;
    private final String    action;
    private final String    type;
    private final String    defaultValue;
    private final String    argument;
    private final ArrayList tableHeader;
    private final ArrayList selection;
    private final ArrayList help;
    private final int       caseValue;
    private final String    guiType;
}
