/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * Modified, 2005-2006 Gregory Dyke, EPFL, Switzerland
 *
 * This file is part of mjc, the MultiJava Compiler, and the JML Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Diff.java,v 1.9 2006/08/16 19:18:13 leavens Exp $
 */

package org.multijava.util.testing;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.StringTokenizer;

/**
 * Class for calculating a (somewhat) detailed comparison of two strings.
 *
 * @author Curtis Clifton
 * @version $Revision: 1.9 $
 */

public class Diff {
    
    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------
    /**
     * Calculate a difference between the given strings.
     *
     * @param oldTextLabel a label for the <code>oldText</code> parameter
     * @param oldText      a value to be compared
     * @param newTextLabel a label for the <code>newText</code> parameter
     * @param newText      a value to be compared
     */
    public Diff(String oldTextLabel, String oldText,
		String newTextLabel, String newText) {
	this(oldTextLabel, oldText, newTextLabel, newText, null);
    }

    /**
     * Calculate a difference between the given strings ignoring the portions
     * of each line that match the given regular expression pattern.
     *
     * @param oldTextLabel a label for the <code>oldText</code> parameter
     * @param oldText      a value to be compared
     * @param newTextLabel a label for the <code>newText</code> parameter
     * @param newText      a value to be compared
     * @param regExpr      a regular expression, matches of which are removed from lines
     *                     before comparison
     */
    public Diff(String oldTextLabel, String oldText,
		String newTextLabel, String newText, Pattern regExpr) {
	this(oldTextLabel, oldText, newTextLabel, newText, regExpr, "");
    }
    
    /**
     * Calculate a difference between the given strings ignoring the portions
     * of each line that match the given regular expression pattern, and treating any
     * character from the <code>equivChars</code> String as equal.
     *
     * @param oldTextLabel a label for the <code>oldText</code> parameter
     * @param oldText      a value to be compared
     * @param newTextLabel a label for the <code>newText</code> parameter
     * @param newText      a value to be compared
     * @param regExpr      a regular expression, matches of which are removed from lines
     *                     before comparison
     * @param equivChars   a string whose characters are considered equivalent
     *                     for comparison purposes
     */
    public Diff(String oldTextLabel, String oldText, String newTextLabel,
		String newText, Pattern regExpr, String equivChars) {
	this.oldText = oldText;
	this.newText = newText;
	this.regExpr = regExpr;
	if (equivChars != null && equivChars.length() > 1) {
	    this.equivChars = Pattern.compile("[" + equivChars + "]");
	    this.equivReplaceChar = equivChars.substring(0,1);
	} else {
	    this.equivChars = null;
	    this.equivReplaceChar = "";
	}
	calculate(oldTextLabel, newTextLabel);
    }
    
    //---------------------------------------------------------------------
    // OPERATIONS
    //---------------------------------------------------------------------
 

    private boolean areSame(String a, String b) {
	a = replaceChars(a);
	b = replaceChars(b);
	return a.equals(b);
    }
    
    private String replaceChars(String s) {
	if (equivChars != null) {
	    Matcher eq = equivChars.matcher(s);
	    s = eq.replaceAll(equivReplaceChar);
	}
	
	if (regExpr != null) {
	    Matcher mat = regExpr.matcher(s);
	    s = mat.replaceAll(IGNORED_CHARS_FILLER);
	}
	
	return s;
    }
    
    private void calculate(String oldTextLabel, String newTextLabel) {
	
	String[] oLines = splitByLine(oldText);
	String[] nLines = splitByLine(newText);

	// accumulate cost in terms of number of changes
	int[] costs = new int[oLines.length+1];
	// accumulate text representation of changes
	String[] results = new String[costs.length];
	
	// algorithm is adaptation of Dynamic Time Walk
	// move through the table of oLines x nLines
	// outerloop is nlines. Calculate best cost of going
	// through each square.
	
	// initialise costs and results for first column
	int x=0;
	int y=0;
	costs[y] = 0;
	results[y] = "";
	for (y=1; y<costs.length; y++) {
	    costs[y]=costs[y-1]+1;
	    results[y] = results[y-1] + (x+1) + OLD_CH + oLines[y-1] + NEWLINE;
	}
	
	// calculate next costs and results from previous column
	// loop invariant: results[y] and costs[y] are the cheapest 
	// way of reaching square (x,y). nLines.length+1 is end of file.
	for (x=1; x<nLines.length+1; x++) {
	    String tempstr = results[0]; //keep a buffer of one previous result
	    int tempint = costs[0];

	    // initialise the new column
	    costs[0] = costs[0]+1; 
	    results[0] = results[0] + (x) + NEW_CH + nLines[x-1] + NEWLINE;

	    // calculate the rest of the new column
	    for (y=1; y<costs.length; y++) {

		// this order of supposition produces a oldlines first, 
		// newlines next. inverse moving from below and from left
		// to produce inverse.

		// suppose best result comes from moving from square to left
		int nextcost=costs[y]+1;
		String nextresult=results[y]+(x)+NEW_CH+nLines[x-1]+NEWLINE;

		// if square to bottom left is cheaper and is a match
		if (areSame(oLines[y-1],nLines[x-1]) && tempint<nextcost) {
		    nextcost=tempint;
		    nextresult=tempstr.toString();
		} 

		// if moving from square below is cheaper.
		if (costs[y-1]+1 < nextcost) {
		    nextcost=costs[y-1]+1;
		    nextresult=results[y-1]+ (x+1) +OLD_CH+oLines[y-1] + NEWLINE;
		}

		tempstr = results[y];
		tempint = costs[y];

		results[y]=nextresult;
		costs[y]=nextcost;

	    }

	}

	y--;

	if (results[y].length() > 0) {
	    // Some diffs accumulated, so the strings are different
	    areDifferent = true;
	    
	    // Prepend a key to the the results.
	    // if the keys aren't "".
	    
	    if (oldTextLabel.compareTo("")!=0 && newTextLabel.compareTo("")!=0) {
		result = NEWLINE +
		    OLD_CH + oldTextLabel + NEWLINE +
		    NEW_CH + newTextLabel + NEWLINE +
		    NEWLINE +
		    results[y];
	    } else {
		result = results[y];
	    }
	} else {
	    // No diffs accumulated, so the strings must be the same
	    areDifferent = false;
	    result = "";
	}
    }
    
    private String[] splitByLine(String text) {
	// thanks to Windows ridiculous two character newlines it is
	// hard to detect blank lines, so we don't bother trying
	StringTokenizer toker = new StringTokenizer(text, DELIM, false);
	ArrayList lines = new ArrayList(oldText.length() / 60);
	while (toker.hasMoreTokens()) {
	    String tok = toker.nextToken();
	    lines.add(tok);
	}
	String[] result = new String[lines.size()];
	lines.toArray(result);
	return result;
    }
    
    /**
     * Returns true if strings on which this was constructed are different.
     */
    public /*@ pure @*/ boolean areDifferent() {
	return areDifferent;
    }
    
    /**
     * Returns the differences between the given strings.
     * <p/>
     * <pre><jml>
     * ensures !areDifferent() <==> \result.equals( "" );
     * </jml></pre>
     */
    public String result() {
	return result;
    }
    
    //---------------------------------------------------------------------
    // PRIVILEGED DATA
    //---------------------------------------------------------------------
    
    private final String oldText;
    private final String newText;
    private final Pattern regExpr;
    private final Pattern equivChars;
    private final String equivReplaceChar;
    private boolean areDifferent;
    private String result;
    
    private static final String DELIM = "\n\r\f";
    private static final String NEWLINE =
	System.getProperty("line.separator");
    private static final String OLD_CH = "< ";
    private static final String NEW_CH = "> ";
    private static final String IGNORED_CHARS_FILLER = "...";
}
