/*
 * Copyright (C) 2000-2002 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: FileCompareTestCase.java,v 1.3 2004/01/16 16:26:58 cclifton Exp $
 * Author: David R. Cok
 */
package org.multijava.util.testing;

import java.io.*;

/** This class is a jutils.framework.Test that compares two files.
    The user needs to make a derived class that overrides the (null)
    implementation of action, in order to do the relevant checking
    of the result of the compare.
*/
public class FileCompareTestCase extends TestCase {

    /** Holds the command String supplied in the constructor. */
    //@ non_null
    String command;
    
    /** Holds the file name supplied in the constructor. */
    //@ non_null
    String file1;
    
    /** Holds the second file name supplied in the constructor. */
    //@ non_null
    String file2;

    /** Constructs the appropriate Test.
	@param name	The name of the test
	@param file1	The first of the files to be compared
	@param file2	The second of the files to be compared
	@param command	The external command to be run to compare them, e.g. "diff"
    */
    //@ requires name != null;
    //@ requires file1 != null;
    //@ requires file2 != null;
    //@ requires command != null;
    public FileCompareTestCase(String name, String file1, String file2, String command) {
	super(name);
	this.command = command;
	this.file1 = file1;
	this.file2 = file2;
    }

    /** The runTest method required by a Test - calling this executes the
	test.  The actual effect is carried out in the 'action' method, which
	the user is expected to override.
    */
    public void runTest() throws java.io.IOException {
	String command = this.command + " " + file1 + " " + file2; 
	//System.out.println("COMPARING " + command);
	Process p = null;
	StringBuffer output = new StringBuffer();
	StringBuffer error = new StringBuffer();
	char[] carray = new char[1000];
	
	try {
	    p = Runtime.getRuntime().exec(command);
	    BufferedReader out = new BufferedReader(new InputStreamReader(p.getInputStream())); // the output of the process
	    BufferedReader err = new BufferedReader(new InputStreamReader(p.getErrorStream()));
	    
	    // FIXME - This is just a hack to get something working some of the time.
	    // We actually have two streams to read, while the process carries on in 
	    // another thread.  We don't know which streams are blocked, etc.  So we really
	    // should read each one on a separate thread, and then get the results back.
	    // ... work for a later time...
	    // We do the output stream first, since the error stream is not as likely to block ??? ...
	    
	    int k = 0;
	    while (k != -1) {
	    	k = out.read(carray,0,1000);
		if (k>0) output.append(carray,0,k);
	    }
	    k = 0;
	    while (k != -1) {
	    	k = err.read(carray,0,1000);
		if (k>0) error.append(carray,0,k);
	    }
	    
	    p.waitFor();
	    int result = p.exitValue();
	    action(result,output.toString(),error.toString());
	} catch (java.lang.InterruptedException e) {
	    fail("EXCEPTION while comparing output between " + file1 + " and " + file2 + ": " + e);
	} finally {
	    if (p != null) p.destroy();
	}
    }

    /** The user should override this to carry out an action (e.g. 'assertTrue(result==0)').
	The method is called by the runTest method after the files have been compared.
	@param result	The exit value produced by the external command supplied to the 
			constructor.  For example, a 'diff' command yields 0 for a match, 1
			for a non-match, and 2 for an execution error.
	@param outputText   The text produced by the output stream of the external process
	@param errorText    The text produced by the error stream of the external process
    */
    public void action(int result, String outputText, String errorText) {
    }
    
}




