/*
 * Copyright (C) 2000-2004 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: Main.java,v 1.9 2004/01/16 16:26:58 cclifton Exp $
 */

package org.multijava.util.testing;

import org.multijava.util.compiler.TabbedPrintWriter;
import org.multijava.util.compiler.CompilationAbortedException;

import java.io.Writer;
import java.io.FileWriter;
import java.io.OutputStreamWriter;
import java.io.IOException;

import java.util.ArrayList;

/**
 * This class provides utilities for automatically generating JUnit
 * test suites. */
public class Main {

    /**
     * Accept command line arguments */
    public static void main( String[] args ) {
	boolean success = new Main().run( args );
	System.exit( success ? 0 : 1 );
    }

    /**
     * Generate JUnit test suite based on the given arguments.  See
     * file TestingOptions.opt for more information on arguments. 
     *
     * @return true iff generation is successful */
    public boolean run( String[] args ) {
        try {
            if( !parseArguments( args ) ) {
                return false;
            }
        } catch( CompilationAbortedException e ) {
            // don't die when called with -help or -version
            return true;
        }

	out.println( "/* Generated by org.multijava.util.testing.Main */" );
	out.println();
	if( options.destPackage().length() > 0 ) {
	    out.println( "package " + options.destPackage() + ";" );
	    out.println();
	}

	out.println( "/**" );
	out.println( " * This class is automatically generated using " );
	out.println( " * org.multijava.util.testing.Main and is used to " );
	out.println( " * group a collection of JUnit tests for the local " );
	out.println( " * package and perhaps some subpackages. " );
	out.println( " */" );

	out.println( "public class " + SUITE_NAME + 
		     " extends junit.framework.TestSuite {" );
	out.add( TAB );
	    
	out.println( "public " + SUITE_NAME + "() {" );
	out.println( "    super();" );
	out.println( "}" );
	out.println();
	out.println( "public " + SUITE_NAME + "( String name ) {" );
	out.println( "    super( name );" );
	out.println( "}" );
	out.println();
	    
	out.println( "public static junit.framework.Test suite() {" );
	out.add( TAB );
	out.println( "junit.framework.TestSuite suite = " );
	out.add( TAB );
	out.println( "new junit.framework.TestSuite() {" );
	out.add( 2*TAB );
	out.println( "public String toString() { return TEST_DESC; }" );
	out.sub( -TAB );
	out.println( "};" );
	out.sub( -2*TAB );
	for( int i=0; i<inClasses.length; i++ ) {
	    String testCase = inClasses[i];
            if( options.useSuite() || isSuite( testCase ) ) {
		out.println( "suite.addTest( " + 
			     options.sourcePackage() + testCase +
			     ".suite() );" );
	    } else {
		out.println( "suite.addTestSuite( " + 
			     options.sourcePackage() + testCase +
			     ".class );" );
	    }
	}
        out.println( "return suite;" );
	out.sub( -TAB );
	out.println( "}" );

	out.println();
	out.println( "public static final String TEST_DESC = " );
	out.add( TAB);
	out.println( "\"" + options.testDescription() + "\";" );
	out.sub( -TAB );

	out.sub( -TAB );
	out.println( "}" );

	out.close();
	    
	return true;	
    }

    /**
     * Processes command line arguments and sets up object state.  */
    public boolean parseArguments( String[] args ) {
	ArrayList inClasses = new ArrayList();
	options = new TestingOptions();
	boolean parseOK = options.parseCommandLine( args, inClasses );
	this.inClasses = (String[]) inClasses.toArray( new String[0] );
	options.set_destPackage(fixPackage( options.destPackage() ));
	options.set_sourcePackage(fixPackage( options.sourcePackage() ));
	if (!options.sourcePackage().equals("")) {
	    options.set_sourcePackage(options.sourcePackage() + ".");
	}
	

	try {
	    Writer wr;
	    if( options.outFile().equals( DEFAULT_OUT ) ) {
		wr = new OutputStreamWriter( System.out );
	    } else {
		wr = new FileWriter( options.outFile() );
	    }
	    out = new SuiteWriter( wr );
	    return true;
	} catch( IOException e ) {
	    System.err.println( "Unable to open output file: " );
	    System.err.println( e.getMessage() );
	    return false;
	}
    }

    String fixPackage( String pack ) {
	return pack.replace( '/', '.' );
    }

    boolean isSuite( String testCase ) {
	return testCase.endsWith( SUITE_NAME );
    }

    private TestingOptions options;
    private String[] inClasses;
    private SuiteWriter out;

    private static final int TAB = 4;
    private static final String DEFAULT_OUT = "System.out";
    private static final String SUITE_NAME = "TestSuite";

    class SuiteWriter extends TabbedPrintWriter {
	SuiteWriter( Writer wr ) {
	    super( wr );
	}
	void println( String s ) {
	    print( s );
	    println();
	}
    }
}
