/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestFilesCompileAllRunEachTestSuite.java,v 1.4 2005/01/14 21:51:00 cheon Exp $
 */

package org.multijava.util.testing;

import org.multijava.util.testing.TestCase;
import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.lang.reflect.Method;

/** This is a JUnit TestSuite that is created from a number of tests
 *  as follows.  Each TestCase is an instance of the inner class
 *  Helper, instantiated with a name of a file.  The file names are
 *  read from the file named by the parameter
 *  <code>fileOfTestFilenames</code>.  The argument to the constructor
 *  named <code>args</code> provides a set of command-line arguments; the
 *  filename for the TestCase is added on to the end of the list of
 *  command-line arguments.  Then the static compile method of the
 *  given class is called on those command-line arguments.
 *
 *  <P> The standard output and error output is captured from the
 *  execution of the compile method. The TestCase succeeds if If the
 *  output is empty; otherwise, it fails.
 *
 *  <P> The test must be run from the directory in which it resides -
 *  because it runs external processes that are expected to be in the
 *  current directory and creates and opens files in the current
 *  directory.
 *
 *  @author David R. Cok
 *  @author Yoonsik Cheon
 */
public class TestFilesCompileAllRunEachTestSuite 
    extends org.multijava.util.testing.TestSuite {

    /** Constructs a new test suite.
     *
     * @param testName The name of the test suite
     * @param fileOfTestFilenames The file to be read for filenames of
     *        tests
     * @param args The command-line arguments that the static compile
     *        method will be applied to, with the filename added on
     * @param cls The class in which to find the static compile method
     */
    public TestFilesCompileAllRunEachTestSuite(String testName, 
                                        String fileOfTestFilenames,
                                        String[] args, 
                                        Class cls) {
	super(testName);
        addTest(new CompileTestCase(fileOfTestFilenames, args, cls));
        
	try { 
	    Iterator i = new FileIterator(fileOfTestFilenames);
	    while (i.hasNext()) {
		String s = (String)i.next();
                if (s.trim().length() > 0) {
                    addTest(new RunTestCase(s, args));	    
                }
	    }
	} catch (java.io.IOException e) {
	    throw new RuntimeException(e.toString());
	}
    }

    // -------------------------------------------------------------
    // INNER CLASSES
    // -------------------------------------------------------------

    /** This is a helper class that is actually a TestCase; it is run
     * repeatedly with different constructor arguments.
     */
    public class CompileTestCase extends org.multijava.util.testing.TestCase {

	/** The constructor called by the JUnit framework. 
         * The argument is used as the name of the test as well as 
         * the name of the file to be tested.
         */
	public CompileTestCase(String fileName,
                               String[] args, 
                               Class cls) {
	    super(fileName);
            this.fileName = fileName;
            this.flags = args;
            try {
                compiler = cls.getMethod("compile", 
                                       new Class[] { String[].class });
            } catch (NoSuchMethodException e) {
                throw new RuntimeException(e.toString());
            }
        }

        private String fileName;
        private Method compiler;
        private String[] flags;

        /** The method that is to be executed on the command-line arguments. */
        protected Method method;

	public void runTest() throws java.io.IOException {

            // First get all the files that are to be tested.
            // Run the external process, which must produce a list of the
            // test filenames, one per line.  A reader reads them into
            // an ArrayList. 
            ArrayList a = new ArrayList(500);
            Iterator i = new FileIterator(fileName);
            while (i.hasNext()) {
                String fn = (String)i.next();
                if (fn.trim().length() > 0) {
                    a.add(fn);
                }
            }

            // Now create a command-line with the options and all the
            // filenames.
            String[] args = new String[flags.length + a.size()];
            System.arraycopy(flags, 0, args, 0, flags.length);
            int j = flags.length;
            i = a.iterator();
            while (i.hasNext()) {
                args[j++] = (String)(i.next());
            }

            // Set the output streams to write into the ByteArray
            // and then execute the compile
            ByteArrayOutputStream ba = new ByteArrayOutputStream();
	    boolean b = false;
            try {
                Utils.setStreams(new PrintStream(ba));
                b = ((Boolean)(compiler.invoke(null, new Object[]{args}))).booleanValue();
            } catch (IllegalAccessException e) {
		fail(e.toString());
            }  catch (java.lang.reflect.InvocationTargetException e) {
		java.io.StringWriter sw = new StringWriter();
		sw.write(e.toString());
		e.printStackTrace(new PrintWriter(sw));
		fail(sw.toString());
	    } finally {
                Utils.restoreStreams();
            }
            String s = ba.toString();

            // Check the the result value is true and that there is no
            // output.
            assertTrue(s, s.length() == 0);

	    assertTrue("The compile produced an invalid return value.  It should be true and instead is " + b, b);

        }
    }
    
    /** This is a helper class that is actually a TestCase; it is run
     * repeatedly with different constructor arguments.
     */
    public class RunTestCase extends org.multijava.util.testing.TestCase {

	/** The constructor called by the JUnit framework. 
         * The argument is used as the name of the test as well as 
         * the name of the file to be tested.
         */
	public RunTestCase(String filename, String[] args) {
	    super(filename);
	    fileToTest = filename;
	}

	/** Executes this test case. This method compiles the given
	 * file and runs the main method. If the compilation or the
	 * exectution of the main method produces any output, then
	 * test fails; otherwise, test succeeds.
         */
	public void runTest() throws java.io.IOException {
	    PrintStream ps = null;
	    try {
		// redirect the output to a file
		ByteArrayOutputStream ba = new ByteArrayOutputStream(5000);
		ps = new PrintStream(ba);
		Utils.setStreams(ps);

                // run the test file
                String cname = fileToTest.substring(0,fileToTest.indexOf('.'));
                Class cls = Class.forName(cname);
                Method meth = cls.getMethod("main", 
                                            new Class[] { String[].class });
                if (meth != null) {
                    meth.invoke(null, new Object[] { new String[0] });
                    String s = ba.toString();
                    assertTrue(s, "".equals(s) || ".".equals(s));
		    // There is no way to test the exit status, nor is
		    // it relevant to the test.
                }
		ps.close();
            } catch (ClassNotFoundException e) {
		fail(e.toString());
            } catch (NoSuchMethodException e) {
		fail(e.toString());
	    } catch (IllegalAccessException e) {
		fail(e.toString());
	    } catch (IllegalArgumentException e) {
                Utils.restoreStreams();
                e.printStackTrace();
		fail(e.toString());
	    } catch (java.lang.reflect.InvocationTargetException e) {
		java.io.StringWriter sw = new StringWriter();
		sw.write(e.toString());
		e.printStackTrace(new PrintWriter(sw));
		fail(sw.toString());
	    } finally {
		Utils.restoreStreams();
	    }
	    if (ps != null) ps.close();
	}

	/** The name of the file to be tested. */
	private String fileToTest;
     }

    // -------------------------------------------------------------
    // DATA MEMBERS
    // -------------------------------------------------------------

    /** The name of this test suite. */
    private String testName;

}
