/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of mjc, the MultiJava Compiler.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestFilesTestSuite.java,v 1.14 2008/03/16 15:58:59 wdietl Exp $
 * Author: David R. Cok
 */
package org.multijava.util.testing;

import org.multijava.util.testing.TestCase;
import org.multijava.util.testing.*;
import junit.framework.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.lang.reflect.Method;

/** This is a JUnit TestSuite that is created from a number of tests as follows.
    Each TestCase is an instance of the inner class Helper, instantiated with
    a name of a file.  The file names are read from the file named by the
    parameter 'fileOfTestFilenames'.  The argument to the constructor named
    'args' provides a set of command-line arguments; the filename for the TestCase
    is added on to the end of the list of command-line arguments.  Then the static
    compile method of the given class is called on those command-line arguments.
<P>
    The standard output and error output is captured from the execution of the
    compile method.  This is compared to the output in filename + "-expected".
    The TestCase succeeds if these match; if they do not match, the test fails and
    the actual output is saved in filename + "-ckd".
<P>
    The test must be run from the directory in which it resides - because
    it creates and opens files in the current directory.

    @author David R. Cok
*/
public class TestFilesTestSuite  extends org.multijava.util.testing.TestSuite {

    /*
    Create derived classes with alternate tests by deriving a class from this 
    one. It should contain an inner Helper class derived from 
    TestFilesTestSuite.Helper.  The method TestFilesTestSuite.makeHelper 
    should be overridden to return an instance of the derived Helper class.  
    The derived Helper class should override Helper.dotest to do the actual 
    test.
    */
    
    // -------------------------------------------------------------
    // DATA MEMBERS
    // -------------------------------------------------------------

    /** The name of this test suite. */
    protected String testName;

    /** The method that is to be executed on the command-line arguments. */
    protected Method method;

    final static String SAVED_SUFFIX = "-ckd";
    final static String ORACLE_SUFFIX = "-expected";

    // -------------------------------------------------------------
    // CONSTRUCTOR
    // -------------------------------------------------------------

    /** A constructor for this test suite.
	@param testName	The name of the test suite
	@param fileOfTestFilenames The file to be read for filenames of tests
	@param args	The command-line arguments that the static compile 
                        method will be applied to, with the filename added on
	@param cls	The class in which to find the static compile method
    */
    public TestFilesTestSuite(String testName, String fileOfTestFilenames,
				String[] args,
				Class cls
				) { 
	super(testName);
	this.testName = testName;
	try {
	    method = cls.getMethod("compile", new Class[] { String[].class });
	} catch (NoSuchMethodException e) {
	    throw new RuntimeException(e.toString());
        }

	try { 
	    Iterator i = new FileIterator(fileOfTestFilenames);
	    while (i.hasNext()) {
		String s = (String)i.next();
		addTest(makeHelper(s,args));	    
	    }
	} catch (java.io.IOException e) {
	    throw new RuntimeException(e.toString());
	}
    }

    /** Factory method for the helper class object. */
    protected Helper makeHelper(String filename, String[] args) {
	return new Helper(filename,args);
    }

    public boolean expectedStatus(String filename, String ss) {
	return ss.length()==0;
    }


    // FIXME - This test does not do the equivalent of FIXTILT or PATHTOFILES
    // that is performed in the Makefile to canonicalize the outputs.  So far we
    // have not needed it.

    /** This is a helper class that is actually a TestCase; it is run repeatedly
	with different constructor arguments.
    */
    public class Helper extends org.multijava.util.testing.TestCase {
    
	/** 
	    The first argument is used as the name of the test as well as 
	    the names of the files to be tested.
	*/
	public Helper(String filenames, String[] args) {
	    super(filenames);
	    filesToTest = filenames.split(" ");
	    this.args = new String[args.length + filesToTest.length];
	    System.arraycopy(args, 0, this.args, 0, args.length);
	    System.arraycopy(filesToTest,
	    		0,
	    		this.args,
	    		args.length,
	    		filesToTest.length);
	}
	
	/** The names of the files to be tested. */
	protected String[] filesToTest;

	/** Command-line arguments (including filename) for this test. */
	protected String[] args;
    
	/** This is the framework around the test.  It sets up the streams to
	    capture output, and catches all relevant exceptions.
	*/
	public void runTest() throws java.io.IOException {
	    //System.out.println("\nTest suite " + testName + ": "  + fileToTest);
	    //for (int kk=0; kk<args.length; ++kk) System.out.println(args[kk]);
	    //System.out.println();
    
	    PrintStream ps = null;
	    try {
		// Redirect the output to a file, and then do the compile
		ByteArrayOutputStream ba = new ByteArrayOutputStream(10000);
		ps = new PrintStream(ba);
		Utils.setStreams(ps);
		
		dotest(ba);
		
	    } catch (IllegalAccessException e) {
		fail(e.toString());
	    } catch (IllegalArgumentException e) {
		fail(e.toString());
	    } catch (java.lang.reflect.InvocationTargetException e) {
		java.io.StringWriter sw = new StringWriter();
		sw.write(e.toString());
		e.printStackTrace(new PrintWriter(sw));
		fail(sw.toString());
//	    } catch (Throwable e) {  // THIS JUST FOR DEBUG
//		Utils.restoreStreams(); // Have to have this before the use of System.out on the next line
//		System.out.println(e);
//		e.printStackTrace();
	    } finally {
		Utils.restoreStreams();
		if (ps != null) ps.close();
	    }
	}
     
	/** This is the actual test; it compiles the given file and compares its output
	    to the expected result (in fileToTest+ORACLE_SUFFIX); the output is expected to 
	    match and the result of the compile to be true or false, depending on
	    whether errors or warnings were reported.  Override this method in derived tests.
	    
	    @param ba contains stuff written to the output stream in this test
	*/
	protected void dotest(ByteArrayOutputStream ba) 
			    throws IllegalAccessException, IllegalArgumentException, 
						java.io.IOException,
						java.lang.reflect.InvocationTargetException {
						
	    boolean b = ((Boolean)(method.invoke(null,new Object[]{args}))).booleanValue();
	    
	    String s = ba.toString();
	    
		String ss = Utils.readFile(filesToTest[0] + ORACLE_SUFFIX);
    
	    Diff df = new Diff("expected", ss, "actual", s);
    
	    if (!df.areDifferent()) {
		// If the two strings match, the test succeeds and we make sure
		// that there is no -ckd file to confuse anyone.
		(new File(filesToTest[0]+SAVED_SUFFIX)).delete();
	    } else {
		// If the strings do not match, we save the actual string and
		// fail the test.
		FileWriter f = new FileWriter(filesToTest[0]+SAVED_SUFFIX);
		f.write(s);
		f.close();
		
		fail(df.result());
	    }
	    
	    expectedStatusReport(b,ss);
	}

        public void expectedStatusReport(boolean b, String ss) {
	    boolean status = expectedStatus(filesToTest[0],ss);
 	    if (status) {
 		assertTrue("The compile produced an invalid return value.  It should be true since there was no error output, and instead is " + b, b);
 	    } else {
 		assertTrue("The compile produced an invalid return value.  It should be false since there was error output, and instead is " + b, !b);
 	    }
        }
     }

}


