/*
 * Copyright (C) 2000-2003 Iowa State University
 *
 * This file is part of JML, the Java Modeling Language.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: TestJmlParser.java,v 1.49 2008/03/16 16:03:50 wdietl Exp $
 */

package org.jmlspecs.checker;

import antlr.RecognitionException;
import antlr.TokenStreamException;
import antlr.TokenStreamRecognitionException;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import org.multijava.mjc.*;
import org.multijava.util.Utils;
import org.multijava.util.testing.TestCase;

/**
 * Unit tests for JmlParser */
public class TestJmlParser extends TestCase implements Constants {

    //---------------------------------------------------------------------
    // CONSTRUCTORS
    //---------------------------------------------------------------------
    
    /** The constructor called by the JUnit framework. */
    public TestJmlParser( String name ) {
	super( name );		// cached for the benefit of JUnit
	parserHelper = new Helper( name );
    }

    //---------------------------------------------------------------------
    // SUITE METHOD
    //---------------------------------------------------------------------

    /** The suite method called by the junit framework */
    public static junit.framework.Test suite() {
        return new junit.framework.TestSuite(TestJmlParser.class);
    }

    //---------------------------------------------------------------------
    // TEST CASES
    //---------------------------------------------------------------------
    // use names like test_memberName_memberKindX,
    // e.g. test_hashCode_method1

    //=====================================================================
    // Compilation Unit Syntax
    //=====================================================================

    //---------------------------------------------------------------------
    // Refine clauses
    //---------------------------------------------------------------------

    public void test_refine_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
  	    parserHelper.getJmlAST( "//@ refine \"foo.java-refined\";" +
				    NEWLINE+
				    "public class Sample { }" );
	assertTrue( "should find a refine-prefix", 
		    unit.refinePrefix() != null );
    }

    public void test_refine_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
  	    parserHelper.getJmlAST( "//@ refine \"foo.java-refined\";" +
				    NEWLINE+
				    "public class Sample { }" );
	assertTrue( "should find a refine-prefix", 
		    unit.refinePrefix() != null );
    }

    //---------------------------------------------------------------------
    // Import clauses
    //---------------------------------------------------------------------

    public void test_import_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( "//@ model import org.jmlspecs.models.*;" + NEWLINE +
				    "public class Sample { }" );
	assertTrue( "should find a model package import",
		    unit.importedPackages().length == 3 &&
		    ((JmlPackageImport) unit.importedPackages()[2]).isModel() );
	assertTrue( "should not find any class imports",
		    unit.importedUnits().length == 0 );
    }

    public void test_import_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( "//@ model import org.jmlspecs.models.JMLObjectSet;" + NEWLINE +
				    "public class Sample { }" );
	assertTrue( "should find a model class import",
		    unit.importedUnits().length == 1 &&
		    ((JmlClassOrGFImport) unit.importedUnits()[0]).isModel() );
	assertTrue( "should find two package imports (for java.lang and org.jmlspecs.lang)",
		    unit.importedPackages().length == 2 &&
		    unit.importedPackages()[0] instanceof JPackageImport &&
		    unit.importedPackages()[1] instanceof JmlPackageImport );
    }

    public void test_import_3()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( "//@ import org.jmlspecs.models.*;" + NEWLINE +
				    "public class Sample { }" );
	assertTrue( "should find a non-model package import",
		    unit.importedPackages().length == 3 &&
		    !((JmlPackageImport) unit.importedPackages()[2]).isModel() );
	assertTrue( "should not find any class imports",
		    unit.importedUnits().length == 0 );
    }

    public void test_import_4()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( "//@ import org.jmlspecs.models.JMLObjectSet;" + NEWLINE +
				    "public class Sample { }" );
	assertTrue( "should find a non-model class import",
		    unit.importedUnits().length == 1 &&
		    !((JmlClassOrGFImport) unit.importedUnits()[0]).isModel() );
	assertTrue( "should find two package imports (for java.lang and org.jmlspecs.lang)",
		    unit.importedPackages().length == 2 &&
		    unit.importedPackages()[0] instanceof JPackageImport &&
		    unit.importedPackages()[1] instanceof JmlPackageImport );
    }

    //=====================================================================
    // Type Definition Syntax
    //=====================================================================

    //---------------------------------------------------------------------
    // JML Modifiers
    //---------------------------------------------------------------------

    public void test_modifier_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( "public /*@ model @*/ class Sample { }" );
	assertTrue( "should find a model class declaration",
		    Utils.hasFlag( getTypeModifiers( unit ), ACC_MODEL ) );
    }

    public void test_modifier_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( "public /*@ pure @*/ class Sample { }" );
	assertTrue( "should find a pure class declaration",
		    Utils.hasFlag( getTypeModifiers( unit ), ACC_PURE ) );
    }

     public void test_modifier_3()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( "public /*@ pure model @*/ class Sample { }" );
	assertTrue( "should find a pure class declaration",
		    Utils.hasFlag( getTypeModifiers( unit ), ACC_PURE ) );
	assertTrue( "should find a model class declaration",
		    Utils.hasFlag( getTypeModifiers( unit ), ACC_MODEL ) );
    }

    //---------------------------------------------------------------------
    // type-spec
    //---------------------------------------------------------------------

    public void test_type_spec_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "	//@ model \\TYPE t;" + NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
    }

    public void test_type_spec_2()
	throws RecognitionException, TokenStreamException 
    {
	try {
	    JmlCompilationUnit unit =
		parserHelper.getJmlAST( 
				       "public class Sample {" + NEWLINE +
				       "	\\TYPE t;" + NEWLINE +
				       "}" 
				       );
	    fail( "should find an error, \\TYPE outside of annotation" );
	} catch( TokenStreamException e ) {
	    assertEquals( "exception message was wrong",
			  "unexpected char: '\\'", 
			  e.getMessage() );
	}
    }

    //---------------------------------------------------------------------
    // weakly
    //---------------------------------------------------------------------

    public void test_weakly_1() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample extends Object /*@ weakly @*/ {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find weakly flag",
		    ((JmlClassDeclaration) unit.typeDeclarations()[0]).
		    isWeakSubtype() );
    }

    public void test_weakly_2() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample extends Object {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertFalse( "should not find weakly flag",
		    ((JmlClassDeclaration) unit.typeDeclarations()[0]).
		    isWeakSubtype() );
    }

    public void test_weakly_3() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample implements Cloneable /*@ weakly @*/ {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { true } ));
    }

    public void test_weakly_4() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample implements Cloneable /*@ weakly @*/, Comparable {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { true, false } ));
    }

    public void test_weakly_5() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample implements Cloneable, Comparable /*@ weakly @*/ {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { false, true } ));
    }

    public void test_weakly_6() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample implements Cloneable /*@ weakly @*/, Comparable /*@ weakly @*/ {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { true, true } ));
    }

    public void test_weakly_7() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public interface Sample extends Cloneable /*@ weakly @*/ {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { true } ));
    }

    public void test_weakly_8() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public interface Sample extends Cloneable /*@ weakly @*/, Comparable {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { true, false } ));
    }

    public void test_weakly_9() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public interface Sample extends Cloneable, Comparable /*@ weakly @*/ {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { false, true } ));
    }

    public void test_weakly_10() 
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public interface Sample extends Cloneable /*@ weakly @*/, Comparable /*@ weakly @*/ {" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "shouldn't find an error",
		      "", parserHelper.errorOut.toString() );
	assertTrue( "should find matching weakly flags",
		    java.util.Arrays.equals(
			((JmlTypeDeclaration) unit.typeDeclarations()[0]).
			interfaceWeaklyFlags(),
			new boolean[] { true, true } ));
    }

    //=====================================================================
    // Field Syntax
    //=====================================================================

    // -------------------------------------------------------------
    // method-decl
    // -------------------------------------------------------------

    public void test_methodDecl_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  void m() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_methodDecl_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  public void m() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_methodDecl_3()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  public static void m() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_methodDecl_4()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public abstract class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  public void m();" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_methodDecl_5()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertFalse( methods[0].hasSpecification() );
    }

    public void test_methodDecl_6()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  public void m() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertFalse( methods[0].hasSpecification() );
    }

    public void test_methodDecl_7()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  public static void m() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertFalse( methods[0].hasSpecification() );
    }

    public void test_methodDecl_8()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public abstract class Sample {" + NEWLINE +
				   "  public void m();" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertFalse( methods[0].hasSpecification() );
    }

    public void test_methodDecl_9()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m()" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_methodDecl_10()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  public void m()" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_methodDecl_11()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  public static void m()" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_methodDecl_12()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public abstract class Sample {" + NEWLINE +
				   "  public void m()" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  ;" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_constructorDecl_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  Sample() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0] instanceof JmlConstructorDeclaration );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_constructorDecl_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  public Sample() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0] instanceof JmlConstructorDeclaration );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_constructorDecl_3()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  Sample() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0] instanceof JmlConstructorDeclaration );
	assertFalse( methods[0].hasSpecification() );
    }

    public void test_constructorDecl_4()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  public Sample() { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0] instanceof JmlConstructorDeclaration );
	assertFalse( methods[0].hasSpecification() );
    }

    public void test_constructorDecl_5()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  Sample()" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0] instanceof JmlConstructorDeclaration );
	assertTrue( methods[0].hasSpecification() );
    }

    public void test_constructorDecl_6()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  public Sample()" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  { }" + NEWLINE +
				   "}" 
				   );
	JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	assertEquals( "should find a single method", 1, methods.length );
	assertTrue( methods[0] instanceof JmlConstructorDeclaration );
	assertTrue( methods[0].hasSpecification() );
    }

    // -------------------------------------------------------------
    // jml-declarations are tested under "behavioral specification
    // syntax for types" below.
    // -------------------------------------------------------------

    // -------------------------------------------------------------
    // initializer and static_initializer
    // -------------------------------------------------------------

    public void test_inits_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  //@ static_initializer" + NEWLINE +
				   "}" 
				   );
	JmlClassBlock[] inits = initializersFrom( unit );
	assertEquals( "should find a single initializer", 1, inits.length );
	assertTrue( inits[0].hasSpecification() );
	assertTrue( inits[0].isStaticInitializer() );
	assertEquals( 0, inits[0].body().length );
    }

    public void test_inits_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  //@ initializer" + NEWLINE +
				   "}" 
				   );
	JmlClassBlock[] inits = initializersFrom( unit );
	assertEquals( "should find a single initializer", 1, inits.length );
	assertTrue( inits[0].hasSpecification() );
	assertTrue( !inits[0].isStaticInitializer() );
	assertEquals( 0, inits[0].body().length );
    }

    public void test_inits_3()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  static int x;" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  static {" + NEWLINE +
				   "    x = 1;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" 
				   );
	JmlClassBlock[] inits = initializersFrom( unit );
	assertEquals( "should find a single initializer", 1, inits.length );
	assertTrue( inits[0].hasSpecification() );
	assertTrue( inits[0].isStaticInitializer() );
	assertEquals( 1, inits[0].body().length );
    }

    public void test_inits_4()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x;" + NEWLINE +
				   "  //@ requires true;" + NEWLINE +
				   "  {" + NEWLINE +
				   "    x = 1;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" 
				   );
	JmlClassBlock[] inits = initializersFrom( unit );
	assertEquals( "should find a single initializer", 1, inits.length );
	assertTrue( inits[0].hasSpecification() );
	assertFalse( inits[0].isStaticInitializer() );
	assertEquals( 1, inits[0].body().length );
    }

    public void test_inits_5()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  static int x;" + NEWLINE +
				   "  static {" + NEWLINE +
				   "    x = 1;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" 
				   );
	JmlClassBlock[] inits = initializersFrom( unit );
	assertEquals( "should find a single initializer", 1, inits.length );
	assertFalse( inits[0].hasSpecification() );
	assertTrue( inits[0].isStaticInitializer() );
	assertEquals( 1, inits[0].body().length );
    }

    public void test_inits_6()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x;" + NEWLINE +
				   "  {" + NEWLINE +
				   "    x = 1;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" 
				   );
	JmlClassBlock[] inits = initializersFrom( unit );
	assertEquals( "should find a single initializer", 1, inits.length );
	assertFalse( inits[0].hasSpecification() );
	assertFalse( inits[0].isStaticInitializer() );
	assertEquals( 1, inits[0].body().length );
    }


    // -------------------------------------------------------------
    // variable-decls
    // -------------------------------------------------------------

    public void test_field_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x; int y; /*@ in x; @*/" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 2, fields.length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 1, fields[1].inGroupClauses().length );

	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
    }

    public void test_field_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x; " + 
				   "protected int y[]; /*@ in x; @*/" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 2, fields.length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 1, fields[1].inGroupClauses().length );

	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
    }

    public void test_field_3()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x; " + 
				   "protected int y=1; /*@ in x; @*/" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 2, fields.length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 1, fields[1].inGroupClauses().length );

	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
    }

    public void test_field_4()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x; " + 
				   "protected int y,z; /*@ in x; @*/" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 3, fields.length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 1, fields[1].inGroupClauses().length );
	assertEquals( 1, fields[2].inGroupClauses().length );

	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
	assertEquals( 0, fields[2].mapsIntoClauses().length );
    }

    public void test_field_5()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x; " + 
				   "protected int[] y[],z; /*@ in x; @*/" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 3, fields.length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 1, fields[1].inGroupClauses().length );
	assertEquals( 1, fields[2].inGroupClauses().length );

	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
	assertEquals( 0, fields[2].mapsIntoClauses().length );
    }

    public void test_field_6()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x; " + 
				   NEWLINE +
				   "protected static int y=1,z=2; /*@ in x; @*/" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 3, fields.length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 1, fields[1].inGroupClauses().length );
	assertEquals( 1, fields[2].inGroupClauses().length );

	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
	assertEquals( 0, fields[2].mapsIntoClauses().length );
    }

    public void test_field_7()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int x; " + 
				   NEWLINE +
				   "  Object y;" + 
				   NEWLINE +
				   "  //@ maps y.owner \\into x;" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 2, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 1, fields[1].mapsIntoClauses().length );

	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
    }

    public void test_field_8()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int g;  " + 
				   NEWLINE +
				   "  public int x[];" + 
				   "  //@ maps x[0..5] \\into g;" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 2, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 1, fields[1].mapsIntoClauses().length );

	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
    }

    public void test_field_9()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  public static int x=1;" + 
				   NEWLINE +
				   "  Object y = new Object();" + 
				   "  //@ maps y.owner \\into x;" + 
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 2, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 1, fields[1].mapsIntoClauses().length );

	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
    }

    public void test_field_10()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int g;  " + 
				   NEWLINE +
				   "  Object x, y;" + 
				   NEWLINE +
				   "  /*@ " +
				   "     maps x.owner \\into g;" + 
				   "     maps y.owner \\into g;" + 
				   "  @*/" +
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 3, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 1, fields[1].mapsIntoClauses().length );
	assertEquals( 1, fields[2].mapsIntoClauses().length );

	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
	assertEquals( 0, fields[2].inGroupClauses().length );
    }

    public void test_field_11()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int g;  " + 
				   NEWLINE +
				   "  public int[] x[], y;" + 
				   NEWLINE +
				   "  /*@ " +
				   "     maps x[*][*] \\into g;" + 
				   "     maps y[*] \\into g;" + 
				   "  @*/" +
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 3, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 1, fields[1].mapsIntoClauses().length );
	assertEquals( 1, fields[2].mapsIntoClauses().length );

	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
	assertEquals( 0, fields[2].inGroupClauses().length );
    }

    public void test_field_12()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int g;  " + 
				   NEWLINE +
				   "public static Object " + 
				   " x = new Object()," +
				   " y = new Object();" + 
				   NEWLINE +
				   "  /*@ " +
				   "     maps x.owner \\into g;" + 
				   "     maps y.owner \\into g;" + 
				   "  @*/" +
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 3, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 1, fields[1].mapsIntoClauses().length );
	assertEquals( 1, fields[2].mapsIntoClauses().length );

	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
	assertEquals( 0, fields[2].inGroupClauses().length );
    }

    public void test_field_13()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int i1, i2;  " + 
				   NEWLINE +
				   "  Object x, y;" + 
				   NEWLINE +
				   "  /*@ " +
				   "     in i1;" + 
				   "     maps x.owner \\into i1;" + 
				   "     maps y.owner \\into i2;" + 
				   "     in i2;" + 
				   "  @*/" +
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 4, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );

	// Only one 'maps-into' clause applies to each of the x and y fields.
	// However, both 'in' clauses apply to both x and y.
	assertEquals( 1, fields[2].mapsIntoClauses().length );
	assertEquals( 1, fields[3].mapsIntoClauses().length );
	assertEquals( 2, fields[2].inGroupClauses().length );
	assertEquals( 2, fields[3].inGroupClauses().length );
    }

    public void test_field_14()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int i1, i2;  " + 
				   "  int g;  " + 
				   NEWLINE +
				   "  public int[] x[], y;" + 
				   NEWLINE +
				   "  /*@ " +
				   "     maps x[*][*] \\into g;" + 
				   "     in i1;" + 
				   "     maps y[*] \\into g;" + 
				   "     in i2;" + 
				   "  @*/" +
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 5, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
	assertEquals( 0, fields[2].mapsIntoClauses().length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
	assertEquals( 0, fields[2].inGroupClauses().length );

	// Only one 'maps-into' clause applies to each of the x and y arrays.
	// However, both 'in' clauses apply to both x and y fields.
	assertEquals( 1, fields[3].mapsIntoClauses().length );
	assertEquals( 1, fields[4].mapsIntoClauses().length );
	assertEquals( 2, fields[3].inGroupClauses().length );
	assertEquals( 2, fields[4].inGroupClauses().length );
    }

    public void test_field_15()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int i1, i2;  " + 
				   "  int g;" + 
				   NEWLINE +
				   "  public Object x, y;" + 
				   NEWLINE +
				   "  /*@ " +
				   "     maps x.* \\into g;" + 
				   "     in i1;" + 
				   "     maps y.* \\into g;" + 
				   "     in i2;" + 
				   "  @*/" +
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 5, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
	assertEquals( 0, fields[2].mapsIntoClauses().length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
	assertEquals( 0, fields[2].inGroupClauses().length );

	// Only one 'maps-into' clause applies to each of the x and y arrays.
	// However, both 'in' clauses apply to both x and y fields.
	assertEquals( 1, fields[3].mapsIntoClauses().length );
	assertEquals( 1, fields[4].mapsIntoClauses().length );
	assertEquals( 2, fields[3].inGroupClauses().length );
	assertEquals( 2, fields[4].inGroupClauses().length );
    }

    public void test_field_16()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  int i1, i2;  " + 
				   "  int g;" + 
				   NEWLINE +
				   "  public Object x[][], y[];" + 
				   NEWLINE +
				   "  /*@ " +
				   "     maps x[1][*].* \\into g;" + 
				   "     in i1;" + 
				   "     maps y[1..5].* \\into g;" + 
				   "     in i2;" + 
				   "  @*/" +
				   NEWLINE +
				   "}" 
				   );

	JmlFieldDeclaration[] fields = fieldsFrom( unit );
	assertEquals( 5, fields.length );
	assertEquals( 0, fields[0].mapsIntoClauses().length );
	assertEquals( 0, fields[1].mapsIntoClauses().length );
	assertEquals( 0, fields[2].mapsIntoClauses().length );
	assertEquals( 0, fields[0].inGroupClauses().length );
	assertEquals( 0, fields[1].inGroupClauses().length );
	assertEquals( 0, fields[2].inGroupClauses().length );

	// Only one 'maps-into' clause applies to each of the x and y arrays.
	// However, both 'in' clauses apply to both x and y fields.
	assertEquals( 1, fields[3].mapsIntoClauses().length );
	assertEquals( 1, fields[4].mapsIntoClauses().length );
	assertEquals( 2, fields[3].inGroupClauses().length );
	assertEquals( 2, fields[4].inGroupClauses().length );
    }

    // -------------------------------------------------------------
    // axiom
    // -------------------------------------------------------------

    public void test_axiom_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  //@ axiom true;" + NEWLINE +
				   "}" 
				   );
	JmlAxiom[] axes = axiomsFrom( unit );
	assertEquals( "should find a single axiom", 1, axes.length );
    }


    // -------------------------------------------------------------
    // parameter-declaration
    // -------------------------------------------------------------

    public void test_parameterDeclaration_1() 
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( int x ) { }" + NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 1, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertFalse( p.isFinal() );
	assertFalse( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_2()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( final int x ) { }" + NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 1, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertTrue( p.isFinal() );
	assertFalse( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_3()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( /*@ non_null @*/ int x ) { }" + NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 1, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertFalse( p.isFinal() );
	assertTrue( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_4()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( final /*@ non_null @*/ int x ) { }" + 
				   NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 1, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertTrue( p.isFinal() );
	assertTrue( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_5()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( /*@ non_null @*/ final int x ) { }" + 
				   NEWLINE +
				   "}" 
				   );
	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 2, character 23 notice: Modifiers appear in non-standard order [final] [JLS 8.1.2, 8.3.1, 8.4.3]" + NEWLINE, 
		      parserHelper.errorOut.toString(), true );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 1, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertTrue( p.isFinal() );
	assertTrue( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_6()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( int x, final Object y ) { }" + 
				   NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 2, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertFalse( p.isFinal() );
	assertFalse( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
	p = (JmlFormalParameter) parms[1];
	assertTrue( p.isFinal() );
	assertFalse( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_7()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( final int x, /*@ non_null @*/ Object y ) { }" + NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 2, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertTrue( p.isFinal() );
	assertFalse( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
	p = (JmlFormalParameter) parms[1];
	assertFalse( p.isFinal() );
	assertTrue( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_8()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( /*@ non_null @*/ int x, final /*@ non_null @*/ Object y ) { }" + NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 2, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[0];
	assertFalse( p.isFinal() );
	assertTrue( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
	p = (JmlFormalParameter) parms[1];
	assertTrue( p.isFinal() );
	assertTrue( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    public void test_parameterDeclaration_9()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
				   "public class Sample {" + NEWLINE +
				   "  void m( final /*@ non_null @*/ int x, Object y ) { }" + 
				   NEWLINE +
				   "}" 
				   );
	JFormalParameter[] parms =
	    parameterDeclarationsFrom(unit);
	assertEquals( 2, parms.length );
	JmlFormalParameter p = (JmlFormalParameter) parms[1];
	assertFalse( p.isFinal() );
	assertFalse( Utils.hasFlag(p.modifiers(), ACC_NON_NULL) );
    }

    //=====================================================================
    // Behavioral Specification Syntax for Types
    //=====================================================================

    // -------------------------------------------------------------
    // jml-var-assertion
    // -------------------------------------------------------------
    
    public void test_varAssertion_1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ initially true; @*/" + 
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertTrue( va[0]instanceof JmlInitiallyVarAssertion );
    }

    public void test_varAssertion_2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ readable x if true; @*/" + 
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertTrue( va[0] instanceof JmlReadableIfVarAssertion );
    }

    public void test_varAssertion_3()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = this; @*/" + 
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertTrue( va[0] instanceof JmlMonitorsForVarAssertion );
    }

    public void test_varAssertion_4()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x <- this; " + 
                                    "    readable x if true; @*/ " +
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertEquals( 2, va.length);
	assertTrue( va[0] instanceof JmlMonitorsForVarAssertion );
	assertTrue( va[1] instanceof JmlReadableIfVarAssertion );
    }

    public void test_varAssertion_5()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = this; " + 
                                    "    initially true; @*/ " +
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertEquals( 2, va.length);
	assertTrue( va[0] instanceof JmlMonitorsForVarAssertion );
	assertTrue( va[1] instanceof JmlInitiallyVarAssertion );
    }

    public void test_varAssertion_6()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ readable x if true; " + 
                                    "    initially true; @*/ " +
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertEquals( 2, va.length);
	assertTrue( va[0] instanceof JmlReadableIfVarAssertion );
	assertTrue( va[1] instanceof JmlInitiallyVarAssertion );
    }

    public void test_varAssertion_7()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x, y; /*@ readable x if true; " + 
                                    "    readable y if true; @*/ " +
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertEquals( 2, va.length);
	assertTrue( va[0] instanceof JmlReadableIfVarAssertion );
	assertTrue( va[1] instanceof JmlReadableIfVarAssertion );
    }

    public void test_varAssertion_8()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = this; @*/" + 
				    NEWLINE +
				    "}" );
	JmlVarAssertion[] va = varAssertionFrom(unit);
	assertTrue( va[0] instanceof JmlMonitorsForVarAssertion );
    }

    //---------------------------------------------------------------------
    // invariant (along with some abuse of the lexer)
    //---------------------------------------------------------------------

    /**
     * Should find type specification annotation. */
    public void test_invariant_1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ invariant true; */" + NEWLINE +
				 "}" );
	JmlInvariant inv = invariantsFrom( unit )[0];

	assertFalse( inv.isRedundantly() );
    }

    /**
     * Should find type specification annotation. */
    public void test_invariant_2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public interface Sample {" + NEWLINE +
				 "//@ invariant_redundantly 1 + 1 == 2;" + NEWLINE +
				 "}" );
	JmlInvariant inv = invariantsFrom( unit )[0];

	assertTrue( inv.isRedundantly() );
    }

    /**
     * Should find type specification annotation spread over multiple lines. */
    public void test_invariant_3()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ invariant" + NEWLINE +
				 "  @     true;" + NEWLINE + 
				 "  @*/" + NEWLINE +
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find one invariant", 1, invs.length );
    }

    /**
     * Should find type specification annotation spread over multiple lines. */
    public void test_invariant_4()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ invariant" + NEWLINE +
				 "  @     true; @*/" + NEWLINE + 
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find one invariant", 1, invs.length );
    }

    /**
     * Should allow multiple '@'. */
    public void test_invariant_5()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@@@@@@@@@@@@@@@@@@@@@@@" + NEWLINE +
				 "  @@ invariant" + NEWLINE +
				 "  @@     true;" + NEWLINE + 
				 "  @@@@@@@@@@@@@@@@@@@@@@@*/" + NEWLINE +
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find one invariant", 1, invs.length );
    }

    /**
     * Should allow multiple '@'. */
    public void test_invariant_6()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//@@@@ invariant true;" + NEWLINE +
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find one invariant", 1, invs.length );
    }

    /**
     * Should find type specification annotation using '+' marker. */
    public void test_invariant_7()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ invariant true; @*/" + NEWLINE +
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find one invariant", 1, invs.length );
    }

    /**
     * Should find type specification annotation using '+' marker. */
    public void test_invariant_8()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//@ invariant true;" + NEWLINE +
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find one invariant", 1, invs.length );
    }

    /**
     * Should allow multiple type specification annotations. */
    public void test_invariant_9()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@" + NEWLINE +
				 "  @ invariant true;" + NEWLINE +
				 "  @*/" + NEWLINE +
				 "/*@" + NEWLINE +
				 "  @ invariant true;" + NEWLINE +
				 "  @*/" + NEWLINE +
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find two invariants", 2, invs.length );
    }

    /**
     * Should allow multiple type specification members in a single
     * annotation. */
    public void test_invariant_10()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@" + NEWLINE +
				 "  @ invariant true;" + NEWLINE +
				 "  @ invariant true;" + NEWLINE +
				 "  @*/" + NEWLINE +
				 "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find two invariants", 2, invs.length );
    }

    /**
     * Should allow a annotation across multiple single-line annotations. */
    public void test_invariant_11()  
	throws TokenStreamException, RecognitionException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "//@ invariant" + NEWLINE +
				    "//@     true;" + NEWLINE +
				    "}" );
	JmlInvariant[] invs = invariantsFrom( unit );

	assertEquals( "should find one invariant", 1, invs.length );
    }


    // -------------------------------------------------------------
    // constraint
    // -------------------------------------------------------------

    private JmlConstraint[] constraintsFrom( JmlCompilationUnit unit ) {
	return ((JmlTypeDeclaration) unit.typeDeclarations()[0]).constraints();
    }

    public void test_constraint_1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true; */" + NEWLINE +
				    "}" );
	JmlConstraint con = constraintsFrom( unit )[0];

	assertFalse( con.isRedundantly() );
	assertTrue( con.isForEverything() );
    }

    public void test_constraint_2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for \\everything; */" + NEWLINE +
				    "}" );
	JmlConstraint con = constraintsFrom( unit )[0];

	assertFalse( con.isRedundantly() );
	assertTrue( con.isForEverything() );
    }

    public void test_constraint_3()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for m; */" + NEWLINE +
				    "}" );
	JmlConstraint con = constraintsFrom( unit )[0];

	assertFalse( con.isRedundantly() );
	assertFalse( con.isForEverything() );
    }

    public void test_constraint_4()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint_redundantly true; */" + NEWLINE +
				    "}" );
	JmlConstraint con = constraintsFrom( unit )[0];

	assertTrue( con.isRedundantly() );
	assertTrue( con.isForEverything() );
    }

    public void test_constraint_5()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint_redundantly true for \\everything; */" + NEWLINE +
				    "}" );
	JmlConstraint con = constraintsFrom( unit )[0];

	assertTrue( con.isRedundantly() );
	assertTrue( con.isForEverything() );
    }

    public void test_constraint_6()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint_redundantly true for m; */" + NEWLINE +
				    "}" );
	JmlConstraint con = constraintsFrom( unit )[0];

	assertTrue( con.isRedundantly() );
	assertFalse( con.isForEverything() );
    }

    // -------------------------------------------------------------
    // method-name-list
    // -------------------------------------------------------------

    private JmlMethodName[] methodNamesFrom( JmlCompilationUnit unit ) {
	return constraintsFrom( unit )[0].methodNames().methodNameList();
    }

    public void test_methodNameList_1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for Outer.this; */" + 
				    NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 2, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "Outer", subnames[0].ident() );
	assertTrue( subnames[1].isThis() );
    }

    public void test_methodNameList_2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for Outer.*; */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 2, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "Outer", subnames[0].ident() );
	assertTrue( subnames[1].isWildcard() );
    }

    public void test_methodNameList_3()  
	throws RecognitionException, TokenStreamException 
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"/*@ constraint true for Outer.super; */" + NEWLINE +
					"}" );
	    fail( "expected exception" );
	} catch (RecognitionException e) {
	    assertEquals( "unexpected token: super", e.getMessage() );
	} // end of catch
    }

    public void test_methodNameList_4()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for new Sample; */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	assertTrue( mNames[0] instanceof JmlConstructorName );
    }

    public void test_methodNameList_5()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  public void m() { }" + NEWLINE +
				    "/*@ constraint true for this.m; */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 2, subnames.length );
	assertTrue( subnames[0].isThis() );
	assertTrue( subnames[1].isIdent() );
	assertEquals( "m", subnames[1].ident() );
    }

    public void test_methodNameList_6()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for super.toString(); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 2, subnames.length );
	assertTrue( subnames[0].isSuper() );
	assertTrue( subnames[1].isIdent() );
	assertEquals( "toString", subnames[1].ident() );
    }

    public void test_methodNameList_7()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for Object.toString(); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 2, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "Object", subnames[0].ident() );
	assertTrue( subnames[1].isIdent() );
	assertEquals( "toString", subnames[1].ident() );
    }

    public void test_methodNameList_8()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for java.lang.Object.toString(); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 4, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "java", subnames[0].ident() );
	assertTrue( subnames[1].isIdent() );
	assertEquals( "lang", subnames[1].ident() );
	assertTrue( subnames[2].isIdent() );
	assertEquals( "Object", subnames[2].ident() );
	assertTrue( subnames[3].isIdent() );
	assertEquals( "toString", subnames[3].ident() );
    }

    public void test_methodNameList_9()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "/*@ constraint true for Sample.toString(); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 2, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertTrue( subnames[1].isIdent() );
	assertEquals( "toString", subnames[1].ident() );
    }

    public void test_methodNameList_10()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  public void m( int i ) { }" + NEWLINE +
				    "  public void m() { }" + NEWLINE +
				    "/*@ constraint true for m(int); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 1, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "m", subnames[0].ident() );
    }

    public void test_constraint11()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  public void m( int i ) { }" + NEWLINE +
				    "  public void m() { }" + NEWLINE +
				    "/*@ constraint true for m(int i); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 1, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "m", subnames[0].ident() );
    }

    public void test_methodNameList_12()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  public void m( int[] i ) { }" + NEWLINE +
				    "  public void m() { }" + NEWLINE +
				    "/*@ constraint true for m(int[]); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 1, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "m", subnames[0].ident() );
    }

    public void test_methodNameList_13()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  public void m( int[] i ) { }" + NEWLINE +
				    "  public void m() { }" + NEWLINE +
				    "/*@ constraint true for m(int i[]); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 1, mNames.length );
	JmlName[] subnames = mNames[0].subnames();
	assertEquals( 1, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "m", subnames[0].ident() );
    }

    public void test_methodNameList_14()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  public void m( int[] i ) { }" + NEWLINE +
				    "  public void m() { }" + NEWLINE +
				    "/*@ constraint true for m(int[]),m(),toString,new Sample(); */" + NEWLINE +
				    "}" );
	JmlMethodName[] mNames = methodNamesFrom( unit );
	assertEquals( 4, mNames.length );
	JmlName[] subnames;
	subnames = mNames[0].subnames();
	assertEquals( 1, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "m", subnames[0].ident() );

	subnames = mNames[1].subnames();
	assertEquals( 1, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "m", subnames[0].ident() );

	subnames = mNames[2].subnames();
	assertEquals( 1, subnames.length );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "toString", subnames[0].ident() );

	assertTrue( mNames[3] instanceof JmlConstructorName );
    }

    public void test_methodNameList_15()  
	throws RecognitionException, TokenStreamException 
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"  public void m() { }" + NEWLINE +
					"  /*@ constraint true for \\everything,m; */" + NEWLINE +
					"}" );
	    JmlConstraint[] cons = constraintsFrom( unit );
	    fail( "should get exception, got cons[0] = " + cons[0] );
	} catch (RecognitionException e) {
	    assertEquals( "expecting ';', found ','", e.getMessage(), true );
	} // end of try-catch
    }

    // -------------------------------------------------------------
    // represents-decls
    // -------------------------------------------------------------

    private JmlRepresentsDecl[] representsDeclsFrom( JmlCompilationUnit unit )
    {
	return 
	    ((JmlTypeDeclaration) unit.typeDeclarations()[0]).
	    representsDecls();
    }

    public void test_represents_decl_1()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  //@ represents size <- 0;" + NEWLINE +
				    "}" );
	JmlRepresentsDecl repdecl = representsDeclsFrom( unit )[0];

      	assertTrue( repdecl.usesAbstractionFunction() );
	assertFalse( repdecl.usesAbstractionRelation() );
	assertFalse( repdecl.isRedundantly() );
    }

    public void test_represents_decl_2()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  //@ represents size \\such_that true;" +
				    NEWLINE +
				    "}" );
	JmlRepresentsDecl repdecl = representsDeclsFrom( unit )[0];

      	assertFalse( repdecl.usesAbstractionFunction() );
	assertTrue( repdecl.usesAbstractionRelation() );
	assertFalse( repdecl.isRedundantly() );
    }

    public void test_represents_decl_3()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  //@ represents_redundantly size <- 0;" 
				    + NEWLINE +
				    "}" );
	JmlRepresentsDecl repdecl = representsDeclsFrom( unit )[0];

      	assertTrue( repdecl.usesAbstractionFunction() );
	assertFalse( repdecl.usesAbstractionRelation() );
	assertTrue( repdecl.isRedundantly() );
    }

    public void test_represents_decl_4()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  //@ represents_redundantly size \\such_that true;" 
				    + NEWLINE +
				    "}" );
	JmlRepresentsDecl repdecl = representsDeclsFrom( unit )[0];

      	assertFalse( repdecl.usesAbstractionFunction() );
	assertTrue( repdecl.usesAbstractionRelation() );
	assertTrue( repdecl.isRedundantly() );
    }

    //=====================================================================
    // Store Ref Syntax
    //=====================================================================

    private JmlStoreRef storeRefFromCC( JmlSpecBodyClause bc ) {
        JmlCapturesClause cc = (JmlCapturesClause) bc;
	JmlStoreRef csref = cc.storeRefs()[0];
	return csref;
    }

    private JmlStoreRef storeRefFrom( JmlSpecBodyClause bc ) {
        JmlAssignableClause ac = (JmlAssignableClause) bc;
	return ac.storeRefs()[0];
    }

    private JmlStoreRef storeRefFrom( JmlCompilationUnit unit ) {
        JmlMethodDeclaration[] meths = methodDeclsFrom(unit);
        JmlMethodSpecification mspec = meths[0].methodSpecification();
        JmlSpecCase[] speccases = mspec.specCases();
        JmlGenericSpecCase gsc = (JmlGenericSpecCase)speccases[0];
        JmlSpecBodyClause bc = gsc.genericSpecBody().simpleSpecBodies()[0];
	return storeRefFrom( bc );
    }
    
    public void test_store_ref_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable \\nothing;");
	JmlStoreRef ref = storeRefFrom( bc );
	assertTrue( ((JmlStoreRefKeyword) ref).isNothing() );
    }

    public void test_store_ref_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable \\everything;");
	JmlStoreRef ref = storeRefFrom( bc );
	assertTrue( ((JmlStoreRefKeyword) ref).isEverything() );
    }

    public void test_store_ref_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable \\not_specified;");
	JmlStoreRef ref = storeRefFrom( bc );
	assertTrue( ((JmlStoreRefKeyword) ref).isNotSpecified() );
    }

    public void test_store_ref_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable (* some things *);");
	JmlStoreRef ref = storeRefFrom( bc );
	assertEquals( " some things ", ((JmlInformalStoreRef) ref).text() );
    }

    /**
     * Informal descriptions may not include newlines in single line
     * annotations. */
    public void test_store_ref_5()
	throws RecognitionException
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"  //@ assignable (* inf" + NEWLINE+ 
					" *) = 0;" + NEWLINE +
                                        " void foo() {}" + NEWLINE +
					"}" );
	    fail( "expected exception" );
	} catch (TokenStreamException e) {
	    assertTrue( "expected message beginning <expecting '*'...>, but got <" + e.getMessage() + ">",
			e.getMessage().startsWith( "expecting '*'" ));
	} // end of try-catch
    }

    /**
     * Informal descriptions may include newlines in multi-line
     * annotations. */
    public void test_store_ref_6()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable (* inf" + NEWLINE + 
				    "    @ desc *); */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRef ref = storeRefFrom( unit );
	assertEquals( " inf desc ", ((JmlInformalStoreRef) ref).text() );
    }

    /**
     * Informal descriptions may include leading stars in multi-line
     * annotations. */
    public void test_store_ref_7()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable (**** inf" + NEWLINE + 
				    "    @ desc *); */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRef ref = storeRefFrom( unit );

	assertEquals( " inf desc ", ((JmlInformalStoreRef) ref).text() );
    }
    
    public void test_store_ref_8()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable foo; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find one name ", 1, names.length );
	assertEquals( "name should be foo ", "foo", names[0].ident() );
    }
    
    public void test_store_ref_9()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable this; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find one name ", 1, names.length );
	assertTrue( "name should be this ", names[0].isThis() );
    }
    
    public void test_store_ref_10()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable super; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find one name ", 1, names.length );
	assertTrue( "name should be super ", names[0].isSuper() );
    }
    
    public void test_store_ref_11()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable foo.bar; */" +NEWLINE+
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find two names ", 2, names.length );
	assertEquals( "foo", names[0].ident() );
	assertEquals( "bar", names[1].ident() );
    }
    
    public void test_store_ref_12()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable this.bar; */" +NEWLINE+
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find two names ", 2, names.length );
	assertTrue( "name should be this ", names[0].isThis() );
	assertEquals( "bar", names[1].ident() );
    }
    
    public void test_store_ref_13()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable super.bar; */"+NEWLINE+
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find two names ", 2, names.length );
	assertTrue( "name should be super ", names[0].isSuper() );
	assertEquals( "bar", names[1].ident() );
    }
    
    public void test_store_ref_14()
	throws TokenStreamException
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"  /*@ assignable foo.super; */" + NEWLINE +
                                        " void foo() {}" + NEWLINE +
					"}" );
	    fail( "should get exception" );
	} catch (RecognitionException e) {
	    assertEquals( "unexpected token: .",
			  e.getMessage() );
	} // end of try-catch
    }
    
    public void test_store_ref_15()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable foo[2]; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find two names ", 2, names.length );
	assertEquals( "foo", names[0].ident() );
	assertTrue( "second \"name\" should be a spec-array-ref-expr", 
		    names[1].isSpecArrayRefExpr() );
	assertTrue( "second \"name\" should be a single position", 
		    names[1].isPos() );
    }
    
    public void test_store_ref_16()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable foo[2..4]; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find two names ", 2, names.length );
	assertEquals( "foo", names[0].ident() );
	assertTrue( "second \"name\" should be a spec-array-ref-expr", 
		    names[1].isSpecArrayRefExpr() );
	assertTrue( "second \"name\" should be a range position", 
		    names[1].isRange() );
    }
    

    /**
     * Make sure adding ".." to lexer didn't break detection of 1..2
     * error outside JML expressions.  */
    public void test_store_ref_17()
	throws TokenStreamException
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"  void m() {" + NEWLINE +
					"    double d = 1..2;" + NEWLINE +
					"  }" + NEWLINE +
					"}" );
	    fail( "should get exception" );
	} catch (RecognitionException e) {
	    assertEquals( "unexpected token: .2",
			  e.getMessage() );
	    
	} // end of try-catch
    }
    
    public void test_store_ref_18()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable foo[*]; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find two names ", 2, names.length );
	assertEquals( "foo", names[0].ident() );
	assertTrue( "second \"name\" should be a spec-array-ref-expr", 
		    names[1].isSpecArrayRefExpr() );
	assertTrue( "second \"name\" should be a star", 
		    names[1].isAll() );
    }
    
    public void test_store_ref_19()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ assignable Sample.this[*]; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
	JmlStoreRefExpression ref = 
	    (JmlStoreRefExpression) storeRefFrom( unit );

	JmlName[] names = ref.names();

	assertEquals( "should find three names ", 3, names.length );
	assertEquals( "Sample", names[0].ident() );
	assertTrue( "second \"name\" should be this ", names[1].isThis() );
	assertTrue( "third \"name\" should be a spec-array-ref-expr", 
		    names[2].isSpecArrayRefExpr() );
	assertTrue( "third \"name\" should be a star", 
		    names[2].isAll() );
    }
    
    public void test_store_ref_20()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  /*@ captures x, y; */" + NEWLINE +
                                    " void foo() {}" + NEWLINE +
				    "}" );
        JmlMethodDeclaration[] meths = methodDeclsFrom(unit);
        JmlMethodSpecification mspec = meths[0].methodSpecification();
        JmlSpecCase[] speccases = mspec.specCases();
        JmlGenericSpecCase gsc = (JmlGenericSpecCase)speccases[0];
        JmlSpecBodyClause bc = gsc.genericSpecBody().simpleSpecBodies()[0];
        JmlCapturesClause cc = (JmlCapturesClause) bc;
	JmlStoreRef[] csref = cc.storeRefs();

	assertEquals( "should find two names ", 2, csref.length );
	assertEquals( "name should be x ", "x", 
		      ((JmlStoreRefExpression)csref[0]).getName() );
	assertEquals( "name should be y ", "y", 
		      ((JmlStoreRefExpression)csref[1]).getName() );
    }
    
    public void test_store_ref_21()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "captures \\nothing;");
	JmlStoreRef ref = storeRefFromCC( bc );
	assertTrue( ((JmlStoreRefKeyword) ref).isNothing() );
    }

    public void test_store_ref_22()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "captures \\everything;");
	JmlStoreRef ref = storeRefFromCC( bc );
	assertTrue( ((JmlStoreRefKeyword) ref).isEverything() );
    }

    public void test_store_ref_23()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "captures \\not_specified;");
	JmlStoreRef ref = storeRefFromCC( bc );
	assertTrue( ((JmlStoreRefKeyword) ref).isNotSpecified() );
    }

    public void test_store_ref_24()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "captures (* some things *);");
	JmlStoreRef ref = storeRefFromCC( bc );
	assertEquals( " some things ", ((JmlInformalStoreRef) ref).text() );
    }
    
    //=====================================================================
    // Behavioral Specification Syntax for Methods
    //=====================================================================

//      //---------------------------------------------------------------------
//      // Method annotation members
//      //---------------------------------------------------------------------

//      /**
//       * Should find method specification annotation. */
//      public void test_method_annotation_1()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find no type annotations", 0, tNotes.length );
//  	assertEquals( "should find one method annotation", 1, mNotes.length );
//      }

//      /**
//       * Should find method specification annotations. */
//      public void test_method_annotation_2()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ ensures true;" + NEWLINE +
//  				 "    @ ensures true; */" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find no type annotations", 0, tNotes.length );
//  	assertEquals( "should find one method annotation", 2, mNotes.length );
//      }
    
//      /**
//       * Should find constructor specification annotations. */
//      public void test_method_annotation_3()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  Sample() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find no type annotations", 0, tNotes.length );
//  	assertEquals( "should find one method annotation", 1, mNotes.length );
//      }
    
//      /**
//       * Should find instance initializer specification annotation. */
//      public void test_method_annotation_4()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  { int x = 0; }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JmlAnnotatable m = 
//  	    (JmlAnnotatable) unit.typeDeclarations()[0].fieldsAndInits()[0];
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find no type annotations", 0, tNotes.length );
//  	assertEquals( "should find one method annotation", 1, mNotes.length );
//      }
    
//      /**
//       * Should find instance initializer specification annotation. */
//      public void test_method_annotation_5()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  static { int x = 0; }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JmlAnnotatable m = 
//  	    (JmlAnnotatable) unit.typeDeclarations()[0].fieldsAndInits()[0];
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find no type annotations", 0, tNotes.length );
//  	assertEquals( "should find one method annotation", 1, mNotes.length );
//      }


    //=====================================================================
    // Behavioral Specification Syntax for Methods
    //=====================================================================

    // -------------------------------------------------------------
    // method-specification
    // -------------------------------------------------------------

    private JmlMethodSpecification methodSpecificationFrom( String mspec) {
	try {
	    JmlCompilationUnit unit =
		parserHelper.getJmlAST( 
		   "public class Sample {" + NEWLINE +
		   "	//@ " + mspec + NEWLINE +
		   "    void foo() { }" + NEWLINE +
		   "}"
		   );
	    JmlMethodDeclaration[] methods = methodDeclsFrom( unit );
	    return methods[0].methodSpecification();
	}
	catch (Exception e) {
	    return null;
	}
    }

    public void test_methodSpecification_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;");

	assertNotNull( mspec );
	assertTrue ( mspec instanceof JmlSpecification );
    }

    public void test_methodSpecification_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "also ensures true;");

	assertNotNull( mspec );
	assertTrue ( mspec instanceof JmlExtendingSpecification );
    }

    // -------------------------------------------------------------
    // specification
    // -------------------------------------------------------------
    
    public void test_specification_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;" +
	       "implies_that ensures true;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNotNull( spec.specCases() );
	assertEquals( 1, spec.specCases().length);
	assertNotNull( spec.redundantSpec() );
    }

    public void test_specification_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNotNull( spec.specCases() );
	assertEquals( 1, spec.specCases().length);
	assertNull( spec.redundantSpec() );
    }

    public void test_specification_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;" +
	       "implies_that ensures true;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNotNull( spec.specCases() );
	assertEquals( 1, spec.specCases().length);
	assertNotNull( spec.redundantSpec() );
    }

    public void test_specification_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;" +
	       "implies_that ensures true;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNotNull( spec.specCases() );
	assertEquals( 1, spec.specCases().length);
	assertNotNull( spec.redundantSpec() );
    }

    public void test_specification_5()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNotNull( spec.specCases() );
	assertEquals( 1, spec.specCases().length);
	assertNull( spec.redundantSpec() );
    }
    public void test_specification_6()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "implies_that ensures true;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNull( spec.specCases() );
	assertNotNull( spec.redundantSpec() );
    }

    public void test_specification_7()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "code behavior assignable x; callable m;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNotNull( spec.specCases() );
	assertEquals( 1, spec.specCases().length);
	assertNull( spec.redundantSpec() );
    }

    public void test_specification_8()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "implies_that ensures true;"
	       );
	JmlSpecification spec = (JmlSpecification) mspec;
	assertNull( spec.specCases() );
	assertNotNull( spec.redundantSpec() );
    }

    // -------------------------------------------------------------
    // spec-case-seq
    // -------------------------------------------------------------
    
    public void test_specCaseSeq_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;");
	JmlSpecCase cases[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( cases );
	assertEquals( 1, cases.length);
    }

    public void test_specCaseSeq_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true; also ensures true;");
	JmlSpecCase cases[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( cases );
	assertEquals( 2, cases.length);
    }

    public void test_specCaseSeq_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true; also ensures true; also ensures true;");
	JmlSpecCase cases[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( cases );
	assertEquals( 3, cases.length);
    }

    // -------------------------------------------------------------
    // spec-case
    // -------------------------------------------------------------
    
    public void test_specCase_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "ensures true;");
	JmlSpecCase cases[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( cases );
	assertEquals( 1, cases.length);
	assertTrue( cases[0] instanceof JmlGenericSpecCase );
    }

    public void test_specCase_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "behavior ensures true;");
	JmlSpecCase cases[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( cases );
	assertEquals( 1, cases.length);
	assertTrue( cases[0] instanceof JmlBehaviorSpec );
    }

    public void test_specCase_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "model_program { return 3; }");
	JmlSpecCase cases[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( cases );
	assertEquals( 1, cases.length);
	assertTrue( cases[0] instanceof JmlModelProgram );
    }

    // -------------------------------------------------------------
    // extending-specification
    // -------------------------------------------------------------

    public void test_extending_specification_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "also ensures true;");
	assertTrue( mspec instanceof JmlExtendingSpecification );
    }

    // -------------------------------------------------------------
    // redundant-spec
    // -------------------------------------------------------------
    
    public void test_redundant_spec_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "implies_that ensures true;" +
	       "for_example example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.implications() );
	assertEquals( 1, r.implications().length );
	assertNotNull( r.examples() );
	assertEquals( 1, r.examples().length );
    }

    public void test_redundant_spec_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNull( r.implications() );
	assertNotNull( r.examples() );
	assertEquals( 1, r.examples().length );
    }

    public void test_redundant_spec_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom("implies_that ensures true;");

	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.implications() );
	assertEquals( 1, r.implications().length );
    }

    // -------------------------------------------------------------
    // implications
    // -------------------------------------------------------------
    
    public void test_immplications_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "implies_that ensures true;" +
	       "for_example example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.implications() );
	assertEquals( 1, r.implications().length );
    }

    public void test_immplications_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "implies_that ensures true; also ensures true;" +
	       "for_example example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.implications() );
	assertEquals( 2, r.implications().length );
    }

    public void test_immplications_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "implies_that ensures true;" +
	       "  also ensures true; also ensures true;" +
	       "for_example example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.implications() );
	assertEquals( 3, r.implications().length );
    }

    // -------------------------------------------------------------
    // examples
    // -------------------------------------------------------------
    
    public void test_examples_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.examples() );
	assertEquals( 1, r.examples().length );
    }

    public void test_examples_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example ensures true;" +
	       "  also example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.examples() );
	assertEquals( 2, r.examples().length );
    }

    public void test_examples_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example ensures true;" +
	       "  also example ensures true;" +
	       "  also example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertNotNull( r );
	assertNotNull( r.examples() );
	assertEquals( 3, r.examples().length );
    }

    // -------------------------------------------------------------
    // generic-spec-case
    // -------------------------------------------------------------

    public void test_genericSpecCase_1()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "ensures true;");

	assertNull( gc.specVarDecls() );
	assertNull( gc.specHeader() );
	assertNotNull( gc.genericSpecBody() );
    }

    public void test_genericSpecCase_2()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "requires true; ensures true;");

	assertNull( gc.specVarDecls() );
	assertNotNull( gc.specHeader() );
	assertEquals( 1, gc.specHeader().length );
	assertNotNull( gc.genericSpecBody() );
    }

    public void test_genericSpecCase_3()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "old int x = 0; ensures true;");

	assertNotNull( gc.specVarDecls() ); 
	assertNull( gc.specHeader() );
	assertNotNull( gc.genericSpecBody() );
    }

    public void test_genericSpecCase_4()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "old int x = 0; requires true; ensures true;");

	assertNotNull( gc.specVarDecls() );
	assertNotNull( gc.specHeader() );
	assertEquals( 1, gc.specHeader().length );
	assertNotNull( gc.genericSpecBody() );
    }

    public void test_genericSpecCase_5()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase( "continues true;" );
	
	assertEquals( "org.multijava.util.compiler.PositionedError: File \"unknown_source_file\", line 1, character 10 error: One of simple spec body clauses is expected [JML]" + NEWLINE, 
		      parserHelper.errorOut.toString(), true );
    }

    // -------------------------------------------------------------
    // spec-header
    // -------------------------------------------------------------

    public void test_specHeader_1()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "requires true; ensures true;");
	JmlRequiresClause[] h = gc.specHeader();
	assertEquals( 1, h.length );
    }

    public void test_specHeader_2()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "requires true; requires true; ensures true;");
	JmlRequiresClause[] h = gc.specHeader();
	assertEquals( 2, h.length );
    }

    public void test_specHeader_3()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "requires true; requires true; requires true; ensures true;");
	JmlRequiresClause[] h = gc.specHeader();
	assertEquals( 3, h.length );
    }

    // -------------------------------------------------------------
    // generic-spec-body
    // -------------------------------------------------------------

    public void test_genericSpecBody_1()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase(
	       "{| requires true; ensures true; |}");
	assertNull( gc.genericSpecBody().simpleSpecBodies() );
	JmlGenericSpecCase c[] = gc.genericSpecBody().genericSpecCases();
	assertNotNull( c );
	assertEquals( 1, c.length);
    }

    public void test_genericSpecBody_2()
	throws RecognitionException, TokenStreamException
    {
	JmlGenericSpecCase gc = 
	    parserHelper.getGenericSpecCase("ensures true;");
	JmlSpecBodyClause b[] = gc.genericSpecBody().simpleSpecBodies();
	assertNotNull( b );
	assertEquals( 1, b.length );
	assertTrue( b[0] instanceof JmlEnsuresClause );
	assertNull( gc.genericSpecBody().genericSpecCases() );
    }

    // -------------------------------------------------------------
    // spec-body
    // -------------------------------------------------------------

    public void test_specBody_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause sb[] = 
	    parserHelper.getSpecBody(
	       "diverges false;" +
	       "assignable x;" +
	       "when true;" + 
	       "working_space i;" +
	       "duration i;" +
	       "ensures true;" +
	       "signals (Exception);" +
	       "continues;" +
	       "breaks;" +
	       "returns;");
	assertEquals( 10, sb.length );
	for (int i = 1; i < sb.length; i++)
	    assertTrue( sb[i-1].preferredOrder() <= sb[i].preferredOrder() );
    }

    public void test_specBody_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause sb[] = 
	    parserHelper.getSpecBody(
                "diverges false;" +          // 1
                "assignable x;" +            // 2
                "when true;" +               // 3
                "working_space i;" +         // 4
                "duration i;" +              // 5
                "ensures true;" +            // 6
                "signals (Exception);" +     // 7
                "continues;" +               // 8
                "breaks;" +                  // 9
                "returns;");                 //10
	assertEquals( 10, sb.length );

	assertTrue( sb[0].isSimpleSpecBody() );
	assertTrue( sb[1].isSimpleSpecBody() );
	assertTrue( sb[2].isSimpleSpecBody() );
	assertTrue( sb[3].isSimpleSpecBody() );
	assertTrue( sb[4].isSimpleSpecBody() );
	assertTrue( sb[5].isSimpleSpecBody() );
	assertTrue( sb[6].isSimpleSpecBody() );
	assertFalse( sb[7].isSimpleSpecBody() );
	assertFalse( sb[8].isSimpleSpecBody() );
	assertFalse( sb[9].isSimpleSpecBody() );

	assertTrue( sb[0].isExceptionalSpecBody() );
	assertTrue( sb[1].isExceptionalSpecBody() );
	assertTrue( sb[2].isExceptionalSpecBody() );
	assertTrue( sb[3].isExceptionalSpecBody() );
	assertTrue( sb[4].isExceptionalSpecBody() );
	assertFalse( sb[5].isExceptionalSpecBody() );
	assertTrue( sb[6].isExceptionalSpecBody() );
	assertFalse( sb[7].isExceptionalSpecBody() );
	assertFalse( sb[8].isExceptionalSpecBody() );
	assertFalse( sb[9].isExceptionalSpecBody() );

	assertTrue( sb[0].isNormalSpecBody() );
	assertTrue( sb[1].isNormalSpecBody() );
	assertTrue( sb[2].isNormalSpecBody() );
	assertTrue( sb[3].isNormalSpecBody() );
	assertTrue( sb[4].isNormalSpecBody() );
	assertTrue( sb[5].isNormalSpecBody() );
	assertFalse( sb[6].isNormalSpecBody() );
	assertFalse( sb[7].isNormalSpecBody() );
	assertFalse( sb[8].isNormalSpecBody() );
	assertFalse( sb[9].isNormalSpecBody() );

	assertTrue( sb[0].isSimpleSpecStatementBody() );
	assertTrue( sb[1].isSimpleSpecStatementBody() );
	assertTrue( sb[2].isSimpleSpecStatementBody() );
	assertTrue( sb[3].isSimpleSpecStatementBody() );
	assertTrue( sb[4].isSimpleSpecStatementBody() );
	assertTrue( sb[5].isSimpleSpecStatementBody() );
	assertTrue( sb[6].isSimpleSpecStatementBody() );
	assertTrue( sb[7].isSimpleSpecStatementBody() );
	assertTrue( sb[8].isSimpleSpecStatementBody() );
	assertTrue( sb[9].isSimpleSpecStatementBody() );

	assertTrue( sb[0].isAbruptSpecBody() );
	assertTrue( sb[1].isAbruptSpecBody() );
	assertTrue( sb[2].isAbruptSpecBody() );
	assertTrue( sb[3].isAbruptSpecBody() );
	assertTrue( sb[4].isAbruptSpecBody() );
	assertFalse( sb[5].isAbruptSpecBody() );
	assertFalse( sb[6].isAbruptSpecBody() );
	assertTrue( sb[7].isAbruptSpecBody() );
	assertTrue( sb[8].isAbruptSpecBody() );
	assertTrue( sb[9].isAbruptSpecBody() );
    }

    // -------------------------------------------------------------
    // behavior-spec
    // -------------------------------------------------------------
    
    public void test_behavior_spec_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "behavior ensures true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
    }

    public void test_behavior_spec_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior ensures true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
    }

    public void test_behavior_spec_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior ensures true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
    }

    // -------------------------------------------------------------
    // exceptional-spec-case
    // -------------------------------------------------------------

    public void test_exceptional_spec_case_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior signals (Exception e);");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNull( gc.specHeader());
	assertNull( gc.specVarDecls());
	assertNotNull( gc.specBody());
    }
    
    public void test_exceptional_spec_case_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior requires true;" +
	       "signals (Exception e);");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specHeader());
	assertEquals( 1, gc.specHeader().length);
	assertNull( gc.specVarDecls());
	assertNotNull( gc.specBody());
    }

    public void test_exceptional_spec_case_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior old int x = 0;" +
	       "signals (Exception e);");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNull( gc.specHeader());
	assertNotNull( gc.specVarDecls());
	assertEquals( 1, gc.specVarDecls().length);
	assertNotNull( gc.specBody());
    }

    public void test_exceptional_spec_case_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior old int x = 0;" +
	       "requires true;" +
	       "signals (Exception e);");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specHeader());
	assertEquals( 1, gc.specHeader().length);	
	assertNotNull( gc.specVarDecls());
	assertEquals( 1, gc.specVarDecls().length);
	assertNotNull( gc.specBody());
    }

    // -------------------------------------------------------------
    // exceptional-spec-body
    // -------------------------------------------------------------

    public void test_exceptional_spec_body_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior diverges true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 1, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlDivergesClause );
	assertNull( gc.specBody().specCases() );
    }

    public void test_exceptional_spec_body_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior {| diverges true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNull( gc.specBody().specClauses() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 1, gc.specBody().specCases().length );
	assertTrue( gc.specBody().specCases()[0] 
		    instanceof JmlExceptionalSpecCase );
    }

    // -------------------------------------------------------------
    // exceptional-spec-case-seq
    // -------------------------------------------------------------

    public void test_exceptional_spec_case_seq_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior {| diverges true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 1, gc.specBody().specCases().length );
    }

    public void test_exceptional_spec_case_seq_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior " +
	       "{| diverges true; also diverges true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 2, gc.specBody().specCases().length );
    }

    public void test_exceptional_spec_case_seq_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "exceptional_behavior " +
	       "{| diverges true; also diverges true; also diverges true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlExceptionalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 3, gc.specBody().specCases().length );
    }

    // -------------------------------------------------------------
    // normal-spec-case
    // -------------------------------------------------------------

    public void test_normal_spec_case_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior ensures true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNull( gc.specHeader());
	assertNull( gc.specVarDecls());
	assertNotNull( gc.specBody());
    }
    
    public void test_normal_spec_case_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior requires true;" +
	       "ensures true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specHeader());
	assertEquals( 1, gc.specHeader().length);
	assertNull( gc.specVarDecls());
	assertNotNull( gc.specBody());
    }

    public void test_normal_spec_case_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior old int x = 0;" +
	       "ensures true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNull( gc.specHeader());
	assertNotNull( gc.specVarDecls());
	assertEquals( 1, gc.specVarDecls().length);
	assertNotNull( gc.specBody());
    }

    public void test_normal_spec_case_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior old int x = 0;" +
	       "requires true;" +
	       "ensures true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specHeader());
	assertEquals( 1, gc.specHeader().length);	
	assertNotNull( gc.specVarDecls());
	assertEquals( 1, gc.specVarDecls().length);
	assertNotNull( gc.specBody());
    }

    // -------------------------------------------------------------
    // normal-spec-body
    // -------------------------------------------------------------

    public void test_normal_spec_body_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior diverges true;");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 1, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlDivergesClause );
	assertNull( gc.specBody().specCases() );
    }

    public void test_normal_spec_body_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior {| diverges true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNull( gc.specBody().specClauses() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 1, gc.specBody().specCases().length );
	assertTrue( gc.specBody().specCases()[0] 
		    instanceof JmlNormalSpecCase );
    }

    // -------------------------------------------------------------
    // normal-spec-case-seq
    // -------------------------------------------------------------

    public void test_normal_spec_case_seq_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior {| ensures true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 1, gc.specBody().specCases().length );
    }

    public void test_normal_spec_case_seq_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior " +
	       "{| ensures true; also ensures true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 2, gc.specBody().specCases().length );
    }

    public void test_normal_spec_case_seq_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "normal_behavior " +
	       "{| ensures true; also ensures true; also ensures true; |}");
	JmlSpecCase specs[] = 
	    ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertEquals( 1, specs.length);
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlNormalBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specCases() );
	assertEquals( 3, gc.specBody().specCases().length );
    }

    // -------------------------------------------------------------
    //  clauses that used to be in the subclassing-contract
    // -------------------------------------------------------------
    
    public void test_subclassing_contract_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible x; captures x;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 3, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertTrue( gc.specBody().specClauses()[2] 
		    instanceof JmlCapturesClause );
	assertNull( gc.specBody().specCases() );
    }

    public void test_subclassing_contract_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible x;" +
	       "  callable m;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 3, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertTrue( gc.specBody().specClauses()[2] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
    }

    public void test_subclassing_contract_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable m; captures x;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 3, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertTrue( gc.specBody().specClauses()[2] 
		    instanceof JmlCapturesClause );
	assertNull( gc.specBody().specCases() );
    }

    public void test_subclassing_contract_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable m;" +
	       "  callable n; captures x, m;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 4, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertTrue( gc.specBody().specClauses()[2] 
		    instanceof JmlCallableClause );
	assertTrue( gc.specBody().specClauses()[3] 
		    instanceof JmlCapturesClause );
	assertNull( gc.specBody().specCases() );
    }

    public void test_subclassing_contract_5()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible x;" +
	       "  accessible y;" +
	       "  callable m; captures x, m;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 5, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertTrue( gc.specBody().specClauses()[2] 
		    instanceof JmlAccessibleClause );
	assertTrue( gc.specBody().specClauses()[3] 
		    instanceof JmlCallableClause );
	assertTrue( gc.specBody().specClauses()[4] 
		    instanceof JmlCapturesClause );
	assertNull( gc.specBody().specCases() );
    }

    // -------------------------------------------------------------
    // accessible-clause
    // -------------------------------------------------------------


    public void test_accessibleClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible x;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertFalse( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_accessibleClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible x, y;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertFalse( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );
    }

    public void test_accessibleClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible \\everything;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertFalse( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	assertTrue( sref.isEverything() );
    }

    public void test_accessibleClause_3b()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible \\nothing;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertFalse( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	assertTrue( sref.isNothing() );
    }

    public void test_accessibleClause_3c()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible \\not_specified;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertFalse( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	assertTrue( sref.isNotSpecified() );
    }

    public void test_accessibleClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible_redundantly x;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertTrue( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_accessibleClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible_redundantly \\everything;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertTrue( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	assertTrue( sref.isEverything() );
    }

    public void test_accessibleClause_6b()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible_redundantly \\nothing;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertTrue( ac.isRedundantly());
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	assertTrue( sref.isNothing() );
    }

    // -------------------------------------------------------------
    // object-ref
    // -------------------------------------------------------------

    public void test_objectRef_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible x;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref instanceof JmlStoreRefExpression );
    }
    
    public void test_objectRef_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable T .y;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
	JmlCallableClause cc 
	    = (JmlCallableClause) gc.specBody().specClauses()[1];

	assertFalse( cc.isRedundantly());
	assertNotNull( cc.methodNames().methodNameList() );
	assertNull( cc.methodNames().storeRefKeyword() );
    }

    public void test_objectRef_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; accessible T .y .z;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlAccessibleClause );
	assertNull( gc.specBody().specCases() );
	JmlAccessibleClause ac 
	    = (JmlAccessibleClause) gc.specBody().specClauses()[1];

	assertTrue( ac.storeRefs().length == 1 );
    }

    // -------------------------------------------------------------
    // callable-clause
    // -------------------------------------------------------------

    public void test_callableClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable m;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
	JmlCallableClause a
	    = (JmlCallableClause) gc.specBody().specClauses()[1];

	assertFalse( a.isRedundantly());
	assertNotNull( a.methodNames().methodNameList() );
	assertNull( a.methodNames().storeRefKeyword() );
    }

    public void test_callableClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable m, n;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
	JmlCallableClause a
	    = (JmlCallableClause) gc.specBody().specClauses()[1];

	assertFalse( a.isRedundantly());
	assertNotNull( a.methodNames().methodNameList() );
	assertNull( a.methodNames().storeRefKeyword() );
    }

    public void test_callableClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable \\everything;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
	JmlCallableClause a
	    = (JmlCallableClause) gc.specBody().specClauses()[1];

	assertFalse( a.isRedundantly());
	assertNull( a.methodNames().methodNameList() );
	assertNotNull( a.methodNames().storeRefKeyword() );
    }


    public void test_callableClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable_redundantly m;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
	JmlCallableClause a
	    = (JmlCallableClause) gc.specBody().specClauses()[1];

	assertTrue( a.isRedundantly());
	assertNotNull( a.methodNames().methodNameList() );
	assertNull( a.methodNames().storeRefKeyword() );
    }

    public void test_callableClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable_redundantly m, n;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
	JmlCallableClause a
	    = (JmlCallableClause) gc.specBody().specClauses()[1];

	assertTrue( a.isRedundantly());
	assertNotNull( a.methodNames().methodNameList() );
	assertNull( a.methodNames().storeRefKeyword() );
    }

    public void test_callableClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "  code behavior assignable x; callable_redundantly \\everything;"
	       );
	JmlSpecCase specs[] = ((JmlSpecification)mspec).specCases();
	assertNotNull( specs );
	assertTrue( specs[0] instanceof JmlHeavyweightSpec );
	assertTrue( specs[0] instanceof JmlBehaviorSpec );
	assertTrue( ((JmlHeavyweightSpec)specs[0]).isCodeSpec() );
	JmlGeneralSpecCase gc = ((JmlHeavyweightSpec)specs[0]).specCase();
	assertNotNull( gc.specBody() );
	assertNotNull( gc.specBody().specClauses() );
	assertEquals( 2, gc.specBody().specClauses().length );
	assertTrue( gc.specBody().specClauses()[0] 
		    instanceof JmlAssignableClause );
	assertTrue( gc.specBody().specClauses()[1] 
		    instanceof JmlCallableClause );
	assertNull( gc.specBody().specCases() );
	JmlCallableClause a
	    = (JmlCallableClause) gc.specBody().specClauses()[1];

	assertTrue( a.isRedundantly());
	assertNull( a.methodNames().methodNameList() );
	assertNotNull( a.methodNames().storeRefKeyword() );
    }

    //=====================================================================
    // Model Program Syntax
    //=====================================================================

    // -------------------------------------------------------------
    // model-program
    // -------------------------------------------------------------
    
    public void test_modelProgram_1()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( 0, "model_program { }" );
	assertTrue( prog.modifiers() == 0 );
	assertEquals( 0, prog.statements().length );
    }
        
    public void test_modelProgram_2()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( Constants.ACC_PRIVATE,
					  "model_program { }" );
	assertTrue( prog.modifiers() == Constants.ACC_PRIVATE );
	assertEquals( 0, prog.statements().length );
    }
        
    public void test_modelProgram_3()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( Constants.ACC_PROTECTED,
					  "model_program { }" );
	assertTrue( prog.modifiers() == Constants.ACC_PROTECTED );
	assertEquals( 0, prog.statements().length );
    }
        
    public void test_modelProgram_4()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( Constants.ACC_PUBLIC,
					  "model_program { }" );
	assertTrue( prog.modifiers() == Constants.ACC_PUBLIC );
	assertEquals( 0, prog.statements().length );
    }
        
    public void test_modelProgram_5()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( 0, "model_program { return; }" );
	assertTrue( prog.modifiers() == 0 );
	assertEquals( 1, prog.statements().length );
    }
        
    public void test_modelProgram_6()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( Constants.ACC_PRIVATE,
					  "model_program { return; }" );
	assertTrue( prog.modifiers() == Constants.ACC_PRIVATE );
	assertEquals( 1, prog.statements().length );
    }
        
    public void test_modelProgram_7()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( ACC_PROTECTED,
					  "model_program { return; }" );
	assertTrue( prog.modifiers() == Constants.ACC_PROTECTED );
	assertEquals( 1, prog.statements().length );
    }
        
    public void test_modelProgram_8()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( ACC_PUBLIC,
					  "model_program { return; }" );
	assertTrue( prog.modifiers() == Constants.ACC_PUBLIC );
	assertEquals( 1, prog.statements().length );
    }
        
    public void test_modelProgram_9()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( 0, 
					  "model_program { int x; return; }" );
	assertTrue( prog.modifiers() == 0 );
	assertEquals( 2, prog.statements().length );
    }
        
    public void test_modelProgram_10()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( ACC_PRIVATE,
					  "model_program { int x; return; }" );
	assertTrue( prog.modifiers() == Constants.ACC_PRIVATE );
	assertEquals( 2, prog.statements().length );
    }
        
    public void test_modelProgram_11()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( ACC_PROTECTED,
					  "model_program { int x; return; }" );
	assertTrue( prog.modifiers() == Constants.ACC_PROTECTED );
	assertEquals( 2, prog.statements().length );
    }
        
    public void test_modelProgram_12()
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( ACC_PUBLIC,
					  "model_program { int x; return; }" );
	assertTrue( prog.modifiers() == Constants.ACC_PUBLIC );
	assertEquals( 2, prog.statements().length );
    }
        
    // -------------------------------------------------------------
    // model-prog-statement
    // -------------------------------------------------------------

    private JmlModelProgStatement getModelProgStatement( String sourceCode ) 
	throws RecognitionException, TokenStreamException
    {
	JmlModelProgram prog = 
	    parserHelper.getModelProgram( 0, sourceCode );
	return (JmlModelProgStatement) prog.statements()[0];
    }

    // all the tests for model-prog-statements "done right" are below,
    // here we check for error message on misplaced
    // model-prog-statements
    public void test_modelProgStatement_1() 
	throws RecognitionException, TokenStreamException
    {
	parserHelper.getStatement( "choose { }" );
	assertDiff( "org.multijava.util.compiler.PositionedError: File \"unknown_source_file\", line 1, character 7 error: A model program statement cannot appear outside of a model program [JML]" + NEWLINE, parserHelper.errorOut.toString() );
    }

    public void test_modelProgStatement_2() 
	throws RecognitionException, TokenStreamException
    {
	parserHelper.getStatement( "choose_if { assume true; int x=0; } {" );
	assertDiff( "org.multijava.util.compiler.PositionedError: File \"unknown_source_file\", line 1, character 10 error: A model program statement cannot appear outside of a model program [JML]" + NEWLINE, parserHelper.errorOut.toString() );
    }

    public void test_modelProgStatement_3() 
	throws RecognitionException, TokenStreamException
    {
	parserHelper.getStatement( "invariant true;" );
	assertDiff( "org.multijava.util.compiler.PositionedError: File \"unknown_source_file\", line 1, character 10 error: A model program statement cannot appear outside of a model program [JML]" + NEWLINE, parserHelper.errorOut.toString() );
    }

    // -------------------------------------------------------------
    // invariant-statement
    // -------------------------------------------------------------
    
    public void test_invariantStatement_1() 
    	throws RecognitionException, TokenStreamException
    {
	JmlInvariantStatement statement = 
	    (JmlInvariantStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  invariant true;" + NEWLINE +
				   "}" );
	assertFalse( statement.isRedundantly() );
    }

    public void test_invariantStatement_2() 
    	throws RecognitionException, TokenStreamException
    {
	JmlInvariantStatement statement = 
	    (JmlInvariantStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  invariant_redundantly true;" + NEWLINE +
				   "}" );
	assertTrue( statement.isRedundantly() );
    }

    // -------------------------------------------------------------
    // nondeterministic-choice
    // -------------------------------------------------------------

    public void test_nondetChoice_1()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetChoiceStatement statement = 
	    (JmlNondetChoiceStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose {" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 1, statement.alternativeStatements().length );
	assertEquals( 0, statement.alternativeStatements()[0].body().length );
    }
        
    public void test_nondetChoice_2()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetChoiceStatement statement = 
	    (JmlNondetChoiceStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 1, statement.alternativeStatements().length );
	assertEquals( 1, statement.alternativeStatements()[0].body().length );
    }
        
    public void test_nondetChoice_3()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetChoiceStatement statement = 
	    (JmlNondetChoiceStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose {" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 2, statement.alternativeStatements().length );
	assertEquals( 0, statement.alternativeStatements()[0].body().length );
	assertEquals( 0, statement.alternativeStatements()[1].body().length );
    }
        
    public void test_nondetChoice_4()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetChoiceStatement statement = 
	    (JmlNondetChoiceStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose {" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 2, statement.alternativeStatements().length );
	assertEquals( 0, statement.alternativeStatements()[0].body().length );
	assertEquals( 1, statement.alternativeStatements()[1].body().length );
    }
        
    public void test_nondetChoice_5()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetChoiceStatement statement = 
	    (JmlNondetChoiceStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 2, statement.alternativeStatements().length );
	assertEquals( 1, statement.alternativeStatements()[0].body().length );
	assertEquals( 0, statement.alternativeStatements()[1].body().length );
    }
        
    public void test_nondetChoice_6()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetChoiceStatement statement = 
	    (JmlNondetChoiceStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    int x=1;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 2, statement.alternativeStatements().length );
	assertEquals( 1, statement.alternativeStatements()[0].body().length );
	assertEquals( 1, statement.alternativeStatements()[1].body().length );
    }
        
    public void test_nondetChoice_7()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetChoiceStatement statement = 
	    (JmlNondetChoiceStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    int x=1;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    int y=2;" + NEWLINE +
				   "    int x=y;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 3, statement.alternativeStatements().length );
	assertEquals( 1, statement.alternativeStatements()[0].body().length );
	assertEquals( 1, statement.alternativeStatements()[1].body().length );
	assertEquals( 2, statement.alternativeStatements()[2].body().length );
    }
        

    // -------------------------------------------------------------
    // nondeterministic-if
    // -------------------------------------------------------------

    public void test_nondetIf_1()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetIfStatement statement = 
	    (JmlNondetIfStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose_if {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 1, statement.guardedStatements().length );
	assertTrue( statement.elseStatements() == null );
	JmlGuardedStatement gs = statement.guardedStatements()[0];
	assertEquals( 1, gs.statements().length );
    }
        
    public void test_nondetIf_2()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetIfStatement statement = 
	    (JmlNondetIfStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose_if {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=1;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 2, statement.guardedStatements().length );
	assertTrue( statement.elseStatements() == null );
	JmlGuardedStatement gs = statement.guardedStatements()[0];
	assertEquals( 1, gs.statements().length );
    }
        
    public void test_nondetIf_3()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetIfStatement statement = 
	    (JmlNondetIfStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose_if {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=1;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int y=2;" + NEWLINE +
				   "    int x=y;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 3, statement.guardedStatements().length );
	assertTrue( statement.elseStatements() == null );
	JmlGuardedStatement gs = statement.guardedStatements()[0];
	assertEquals( 1, gs.statements().length );
    }
        
    public void test_nondetIf_4()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetIfStatement statement = 
	    (JmlNondetIfStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose_if {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } else {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 1, statement.guardedStatements().length );
	assertEquals( 1, statement.elseStatements().length );
	JmlGuardedStatement gs = statement.guardedStatements()[0];
	assertEquals( 1, gs.statements().length );
    }
        
    public void test_nondetIf_5()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetIfStatement statement = 
	    (JmlNondetIfStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose_if {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=1;" + NEWLINE +
				   "  } else {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 2, statement.guardedStatements().length );
	assertEquals( 1, statement.elseStatements().length );
	JmlGuardedStatement gs = statement.guardedStatements()[0];
	assertEquals( 1, gs.statements().length );
    }
        
    public void test_nondetIf_6()
	throws RecognitionException, TokenStreamException
    {
	JmlNondetIfStatement statement = 
	    (JmlNondetIfStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  choose_if {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int x=1;" + NEWLINE +
				   "  } or {" + NEWLINE +
				   "    assume true;" + NEWLINE +
				   "    int y=2;" + NEWLINE +
				   "    int x=y;" + NEWLINE +
				   "  } else {" + NEWLINE +
				   "    int x=0;" + NEWLINE +
				   "  }" + NEWLINE +
				   "}" );
	assertEquals( 3, statement.guardedStatements().length );
	assertEquals( 1, statement.elseStatements().length );
	JmlGuardedStatement gs = statement.guardedStatements()[0];
	assertEquals( 1, gs.statements().length );
    }
        
    // -------------------------------------------------------------
    // guarded-statement
    // -------------------------------------------------------------

    private JmlGuardedStatement getGuardedStatementFrom( String sourceCode ) 
	throws RecognitionException, TokenStreamException
    {
	JmlNondetIfStatement statement = 
	    (JmlNondetIfStatement) getModelProgStatement( sourceCode );
	return statement.guardedStatements()[0];
    }

    public void test_guardedStatement_1()
	throws RecognitionException, TokenStreamException
    {
	JmlGuardedStatement statement = 
	    getGuardedStatementFrom( "model_program {" + NEWLINE +
				     "  choose_if {" + NEWLINE +
				     "    assume true;" + NEWLINE +
				     "    int x=0;" + NEWLINE +
				     "  }" + NEWLINE +
				     "}" );
	assertEquals( 1, statement.statements().length );
    }
        
    public void test_guardedStatement_2()
	throws RecognitionException, TokenStreamException
    {
	JmlGuardedStatement statement = 
	    getGuardedStatementFrom( "model_program {" + NEWLINE +
				     "  choose_if {" + NEWLINE +
				     "    assume true;" + NEWLINE +
				     "    int x=0;" + NEWLINE +
				     "    int y=0;" + NEWLINE +
				     "  }" + NEWLINE +
				     "}" );
	assertEquals( 2, statement.statements().length );
    }
        
    public void test_guardedStatement_3()
	throws RecognitionException, TokenStreamException
    {
	JmlGuardedStatement statement = 
	    getGuardedStatementFrom( "model_program {" + NEWLINE +
				     "  choose_if {" + NEWLINE +
				     "    assume true;" + NEWLINE +
				     "    int x=0;" + NEWLINE +
				     "    int y=0;" + NEWLINE +
				     "    int z=0;" + NEWLINE +
				     "  }" + NEWLINE +
				     "}" );
	assertEquals( 3, statement.statements().length );
    }
        
    public void test_guardedStatement_4()
	throws RecognitionException, TokenStreamException
    {
	try {
	    JmlGuardedStatement statement = 
		getGuardedStatementFrom( "model_program {" + NEWLINE +
					 "  choose_if {" + NEWLINE +
					 "    int x=0;" + NEWLINE +
					 "  }" + NEWLINE +
					 "}" );
	    fail( "should get exception" );
	} catch (RecognitionException e) {
	    assertEquals( "unexpected token: int", e.getMessage() );
	} // end of catch
    }


    // -------------------------------------------------------------
    // spec-statement
    // -------------------------------------------------------------

    public void test_specStatement_1() 
    	throws RecognitionException, TokenStreamException
    {
	JmlSpecStatement statement = 
	    (JmlSpecStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  behavior requires true;" + NEWLINE +
				   "}" );
	assertTrue ( statement.isGeneric() );
	assertFalse( statement.isExceptional() );
	assertFalse( statement.isNormal() );
	assertFalse( statement.isAbrupt() );
    }

    public void test_specStatement_2() 
    	throws RecognitionException, TokenStreamException
    {
	JmlSpecStatement statement = 
	    (JmlSpecStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  exceptional_behavior requires true;" 
				   + NEWLINE +
				   "}" );
	assertFalse( statement.isGeneric() );
	assertTrue ( statement.isExceptional() );
	assertFalse( statement.isNormal() );
	assertFalse( statement.isAbrupt() );
    }

    public void test_specStatement_3() 
    	throws RecognitionException, TokenStreamException
    {
	JmlSpecStatement statement = 
	    (JmlSpecStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  normal_behavior requires true;" 
				   + NEWLINE +
				   "}" );
	assertFalse( statement.isGeneric() );
	assertFalse( statement.isExceptional() );
	assertTrue ( statement.isNormal() );
	assertFalse( statement.isAbrupt() );
    }

    public void test_specStatement_4() 
    	throws RecognitionException, TokenStreamException
    {
	JmlSpecStatement statement = 
	    (JmlSpecStatement)
	    getModelProgStatement( "model_program {" + NEWLINE +
				   "  abrupt_behavior requires true;" + 
				   NEWLINE +
				   "}" );
	assertFalse( statement.isGeneric() );
	assertFalse( statement.isExceptional() );
	assertFalse( statement.isNormal() );
	assertTrue ( statement.isAbrupt() );
    }

    // -------------------------------------------------------------
    // abrupt-spec-case
    // -------------------------------------------------------------

    private JmlGeneralSpecCase getSpecCase( String sourceCode ) 
	throws RecognitionException, TokenStreamException
    {
	JmlSpecStatement statement = 
	    (JmlSpecStatement) getModelProgStatement( sourceCode );
	return statement.specCase();
    }

    public void test_abruptSpecCase_1() 
    	throws RecognitionException, TokenStreamException
    {
	JmlAbruptSpecCase specCase = 
	    (JmlAbruptSpecCase)
	    getSpecCase( "model_program {" + NEWLINE +
			 "  abrupt_behavior assignable x;" + NEWLINE +
			 "}" );
	assertNull   ( specCase.specHeader());
	assertNull   ( specCase.specVarDecls());
	assertNotNull( specCase.specBody());
    }

    public void test_abruptSpecCase_2() 
    	throws RecognitionException, TokenStreamException
    {
	JmlAbruptSpecCase specCase = 
	    (JmlAbruptSpecCase)
	    getSpecCase( "model_program {" + NEWLINE +
			 "  abrupt_behavior" + NEWLINE +
			 "  requires true;" + NEWLINE +
			 "  assignable x;" + NEWLINE +
			 "}" );
	assertNotNull( specCase.specHeader());
	assertEquals ( 1, specCase.specHeader().length);
	assertNull( specCase.specVarDecls());
	assertNotNull( specCase.specBody());
    }


    public void test_abruptSpecCase_3() 
    	throws RecognitionException, TokenStreamException
    {
	JmlAbruptSpecCase specCase = 
	    (JmlAbruptSpecCase)
	    getSpecCase( "model_program {" + NEWLINE +
			 "  abrupt_behavior old int x = 0;" + NEWLINE +
			 "  assignable x;" + NEWLINE +
			 "}" );
	assertNull( specCase.specHeader());
	assertNotNull( specCase.specVarDecls());
	assertEquals ( 1, specCase.specVarDecls().length);
	assertNotNull( specCase.specBody());
    }


    public void test_abruptSpecCase_4() 
    	throws RecognitionException, TokenStreamException
    {
	JmlAbruptSpecCase specCase = 
	    (JmlAbruptSpecCase)
	    getSpecCase( "model_program {" + NEWLINE +
			 "  abrupt_behavior old int x = 0;" + NEWLINE +
			 "  requires true;" + NEWLINE +
			 "  assignable x;" + NEWLINE +
			 "}" );
	assertNotNull( specCase.specHeader());
	assertEquals ( 1, specCase.specHeader().length);
	assertNotNull( specCase.specVarDecls());
	assertEquals ( 1, specCase.specVarDecls().length);
	assertNotNull( specCase.specBody());
    }

    // -------------------------------------------------------------
    // abrupt-spec-body
    // -------------------------------------------------------------

    public void test_abruptSpecBody_1()
	throws RecognitionException, TokenStreamException
    {
	JmlAbruptSpecCase gc = 
	    parserHelper.getAbruptSpecCase(
	       "{| requires true; assignable x; |}");
	assertNull( gc.abruptSpecBody().specClauses() );
	JmlSpecCase c[] = gc.abruptSpecBody().specCases();
	assertNotNull( c );
	assertEquals( 1, c.length);
    }

    public void test_abruptSpecBody_2()
	throws RecognitionException, TokenStreamException
    {
	JmlAbruptSpecCase gc = 
	    parserHelper.getAbruptSpecCase("assignable x;");
	JmlSpecBodyClause b[] = gc.abruptSpecBody().specClauses();
	assertNotNull( b );
	assertEquals( 1, b.length );
	assertTrue( b[0] instanceof JmlAssignableClause );
	assertNull( gc.abruptSpecBody().specCases() );
    }

    // -------------------------------------------------------------
    // continues-clause
    // -------------------------------------------------------------
    
    public void test_continuesClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_continuesClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues true;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_continuesClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues \\not_specified;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_continuesClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues_redundantly;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_continuesClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues_redundantly true;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_continuesClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "continues_redundantly \\not_specified;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_continuesClause_7()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues -> ( l );");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_continuesClause_8()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues -> ( l ) true;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_continuesClause_9()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
               "continues -> ( l ) \\not_specified;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_continuesClause_10()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("continues_redundantly -> ( l ) ;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_continuesClause_11()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
               "continues_redundantly -> (l) true;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_continuesClause_12()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "continues_redundantly -> ( l ) \\not_specified;");
	JmlContinuesClause ec = (JmlContinuesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    // -------------------------------------------------------------
    // breaks-clause
    // -------------------------------------------------------------
    
    public void test_breaksClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_breaksClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks true;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_breaksClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks \\not_specified;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_breaksClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks_redundantly;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_breaksClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks_redundantly true;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_breaksClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "breaks_redundantly \\not_specified;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNull( ec.label() );
    }

    public void test_breaksClause_7()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks -> ( l );");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_breaksClause_8()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks -> ( l ) true;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_breaksClause_9()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks -> ( l ) \\not_specified;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_breaksClause_10()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks_redundantly -> ( l ) ;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_breaksClause_11()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("breaks_redundantly -> (l) true;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    public void test_breaksClause_12()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "breaks_redundantly -> ( l ) \\not_specified;");
	JmlBreaksClause ec = (JmlBreaksClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
	assertNotNull( ec.label() );
    }

    // -------------------------------------------------------------
    // returns-clause
    // -------------------------------------------------------------
    
    public void test_returnsClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("returns;");
	JmlReturnsClause ec = (JmlReturnsClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
    }

    public void test_returnsClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("returns true;");
	JmlReturnsClause ec = (JmlReturnsClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
    }

    public void test_returnsClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("returns \\not_specified;");
	JmlReturnsClause ec = (JmlReturnsClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
    }

    public void test_returnsClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("returns_redundantly;");
	JmlReturnsClause ec = (JmlReturnsClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
    }

    public void test_returnsClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("returns_redundantly true;");
	JmlReturnsClause ec = (JmlReturnsClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
	assertNotNull( ec.predOrNot() );
    }

    public void test_returnsClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "returns_redundantly \\not_specified;");
	JmlReturnsClause ec = (JmlReturnsClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
	assertNull( ec.predOrNot() );
    }

    //=====================================================================
    // Example Syntax
    //=====================================================================

    // -------------------------------------------------------------
    // example
    // -------------------------------------------------------------
    
    public void test_example_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertEquals( 0, r.examples()[0].privacy() );
    }

    public void test_example_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example public example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertEquals( Constants.ACC_PUBLIC, r.examples()[0].privacy() );
    }

    public void test_example_2_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example static example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( parserHelper.errorOut.toString().indexOf(
		       "One of privacy modifiers") > 0);
    }

    public void test_example_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example protected example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertEquals( Constants.ACC_PROTECTED, r.examples()[0].privacy() );
    }

    public void test_example_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example private example ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertEquals( Constants.ACC_PRIVATE, r.examples()[0].privacy() );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_5()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example" +
	       "  old int x = 0;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlExample );
	assertFalse( r.examples()[0] instanceof JmlExceptionalExample );
	assertFalse( r.examples()[0] instanceof JmlNormalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNotNull( r.examples()[0].specVarDecls() );
	assertEquals( 1, r.examples()[0].specVarDecls().length );
	assertNull( r.examples()[0].specHeader() );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_6()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example" +
	       "  requires true;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertEquals( 0, r.examples()[0].privacy() );
	assertNull( r.examples()[0].specVarDecls() );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_6_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example" +
	       "  requires true;");
	assertNull( mspec );
    }

    public void test_example_7()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example example" +
	       "  old int x = 0;" +
	       "  requires true;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertEquals( 0, r.examples()[0].privacy() );
	assertNotNull( r.examples()[0].specVarDecls() );
	assertEquals( 1, r.examples()[0].specVarDecls().length );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_8()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example public exceptional_example" +
	       "  old int x = 0;" +
	       "  requires true;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlExceptionalExample );
	assertEquals( Constants.ACC_PUBLIC, r.examples()[0].privacy() );
	assertNotNull( r.examples()[0].specVarDecls() );
	assertEquals( 1, r.examples()[0].specVarDecls().length );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_9()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example exceptional_example" +
	       "  requires true;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlExceptionalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNull( r.examples()[0].specVarDecls() );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_9_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example exceptional_example" +
	       "  requires true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlExceptionalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNull( r.examples()[0].specVarDecls() );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNull( r.examples()[0].specClauses() );
    }

    public void test_example_10()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example exceptional_example" +
	       "  old int x = 0;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlExceptionalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNotNull( r.examples()[0].specVarDecls() );
	assertEquals( 1, r.examples()[0].specVarDecls().length );
	assertNull( r.examples()[0].specHeader() );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_11()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example exceptional_example" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlExceptionalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNull( r.examples()[0].specVarDecls() );
	assertNull( r.examples()[0].specHeader() );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_12()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example public normal_example" +
	       "  old int x = 0;" +
	       "  requires true;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlNormalExample );
	assertEquals( Constants.ACC_PUBLIC, r.examples()[0].privacy() );
	assertNotNull( r.examples()[0].specVarDecls() );
	assertEquals( 1, r.examples()[0].specVarDecls().length );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_13()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example normal_example" +
	       "  requires true;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlNormalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNull( r.examples()[0].specVarDecls() );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_14()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example normal_example" +
	       "  old int x = 0;" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlNormalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNotNull( r.examples()[0].specVarDecls() );
	assertEquals( 1, r.examples()[0].specVarDecls().length );
	assertNull( r.examples()[0].specHeader() );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_15()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "for_example normal_example" +
	       "  ensures true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlNormalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNull( r.examples()[0].specVarDecls() );
	assertNull( r.examples()[0].specHeader() );
	assertNotNull( r.examples()[0].specClauses() );
	assertEquals( 1, r.examples()[0].specClauses().length );
    }

    public void test_example_16()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = methodSpecificationFrom(
	       "for_example normal_example" +
	       "  requires true;");
	JmlSpecification spec = (JmlSpecification) mspec;
	JmlRedundantSpec r = spec.redundantSpec;
	assertTrue( r.examples()[0] instanceof JmlNormalExample );
	assertEquals( 0, r.examples()[0].privacy() );
	assertNull( r.examples()[0].specVarDecls() );
	assertNotNull( r.examples()[0].specHeader() );
	assertEquals( 1, r.examples()[0].specHeader().length );
	assertNull( r.examples()[0].specClauses() );
    }

    public void test_example_17()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit =
	    parserHelper.getJmlAST( 
	       "public class Sample {" + NEWLINE +
               "  //@ for_example private requires true;" + NEWLINE +
               "  //@ ensures true;" + NEWLINE +
               "  void m() { }" + NEWLINE +
               "}" 
               );
	assertEquals( "org.multijava.util.compiler.PositionedError: File \"unknown_source_file\", line 2, character 26 error: A light-weighted example specification can not have an access modifier [JML]" + NEWLINE, 
		      parserHelper.errorOut.toString(), true );
    }


    //=====================================================================
    // Method Specification Clause Syntax
    //=====================================================================

    // -------------------------------------------------------------
    // spec-var-decls
    // -------------------------------------------------------------

    public void test_specVarDecls_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "forall int i; ensures" );
	assertEquals( 1, decls.length );
	assertTrue( decls[0] instanceof JmlForAllVarDecl );
    }
        
    public void test_specVarDecls_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "forall int i; forall int j; ensures" );
	assertEquals( 2, decls.length );
	assertTrue( decls[0] instanceof JmlForAllVarDecl );
	assertTrue( decls[1] instanceof JmlForAllVarDecl );
    }
        
    public void test_specVarDecls_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "forall int i; forall int j; forall int k; ensures" );
	assertEquals( 3, decls.length );
	assertTrue( decls[0] instanceof JmlForAllVarDecl );
	assertTrue( decls[1] instanceof JmlForAllVarDecl );
	assertTrue( decls[2] instanceof JmlForAllVarDecl );
    }
        
    public void test_specVarDecls_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "forall int i; old int j=0;" );
	assertEquals( 2, decls.length );
	assertTrue( decls[0] instanceof JmlForAllVarDecl );
	assertTrue( decls[1] instanceof JmlLetVarDecl );
    }
        
    public void test_specVarDecls_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "forall int i; forall int j; old int j=0;" );
	assertEquals( 3, decls.length );
	assertTrue( decls[0] instanceof JmlForAllVarDecl );
	assertTrue( decls[1] instanceof JmlForAllVarDecl );
	assertTrue( decls[2] instanceof JmlLetVarDecl );
    }
        
    public void test_specVarDecls_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "forall int i; old int x=0; old int y=0;" );
	assertEquals( 3, decls.length );
	assertTrue( decls[0] instanceof JmlForAllVarDecl );
	assertTrue( decls[1] instanceof JmlLetVarDecl );
 	assertTrue( decls[2] instanceof JmlLetVarDecl );
   }
     
   
    public void test_specVarDecls_7()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "forall int i; forall int j; old int j=0; old int y=0;" );
	assertEquals( 4, decls.length );
	assertTrue( decls[0] instanceof JmlForAllVarDecl );
	assertTrue( decls[1] instanceof JmlForAllVarDecl );
	assertTrue( decls[2] instanceof JmlLetVarDecl );
	assertTrue( decls[3] instanceof JmlLetVarDecl );
    }
        
    public void test_specVarDecls_8()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "old int j=0;" );
	assertEquals( 1, decls.length );
	assertTrue( decls[0] instanceof JmlLetVarDecl );
    }
        
    public void test_specVarDecls_9()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecVarDecl[] decls = 
	    parserHelper.getSpecVarDecls( "old int x=0; old int y=0;" );
	assertEquals( 2, decls.length );
	assertTrue( decls[0] instanceof JmlLetVarDecl );
 	assertTrue( decls[1] instanceof JmlLetVarDecl );
    }
    

    // -------------------------------------------------------------
    // forall-var-decl
    // -------------------------------------------------------------
    
    public void test_forAllVarDecl_1()
	throws RecognitionException, TokenStreamException
    {
	JmlForAllVarDecl decl = 
	    (JmlForAllVarDecl) parserHelper.
	    getSpecVarDecls( "forall int i; ensures" )[0];

	assertEquals( 1, decl.quantifiedVarDecls().length );
    }
        
    public void test_forAllVarDecl_2()
	throws RecognitionException, TokenStreamException
    {
	JmlForAllVarDecl decl = 
	    (JmlForAllVarDecl) parserHelper.
	    getSpecVarDecls( "forall int i, j; ensures" )[0];

	assertEquals( 2, decl.quantifiedVarDecls().length );
    }
        
    // -------------------------------------------------------------
    // let-var-decls, tested in specVariableDeclarators below
    // -------------------------------------------------------------

    private JmlLetVarDecl[] letVarDeclsFrom( JmlSpecVarDecl[] oldDecls ) {
	JmlLetVarDecl[] newDecls = new JmlLetVarDecl[ oldDecls.length ];
	System.arraycopy( oldDecls, 0, newDecls, 0, oldDecls.length );
	return newDecls;
    }

    // -------------------------------------------------------------
    // requires-clause
    // -------------------------------------------------------------

    public void test_requiresClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "requires true;" );
	assertFalse( rc.isRedundantly() );
	assertFalse( rc.isNotSpecified() );
    }
        
    public void test_requiresClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "requires_redundantly true;");
	assertTrue( rc.isRedundantly() );
	assertFalse( rc.isNotSpecified() );
    }
        
    public void test_requiresClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "requires \\not_specified;");
	assertFalse( rc.isRedundantly() );
	assertTrue( rc.isNotSpecified() );
    }
        
    public void test_requiresClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "requires_redundantly \\not_specified;" );
	assertTrue( rc.isRedundantly() );
	assertTrue( rc.isNotSpecified() );
    }
        
    public void test_requiresClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "pre true;" );
	assertFalse( rc.isRedundantly() );
	assertFalse( rc.isNotSpecified() );
    }
        
    public void test_requiresClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "pre_redundantly true;" );
	assertTrue( rc.isRedundantly() );
	assertFalse( rc.isNotSpecified() );
    }
        
    public void test_requiresClause_7()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "pre \\not_specified;" );
	assertFalse( rc.isRedundantly() );
	assertTrue( rc.isNotSpecified() );
    }
        
    public void test_requiresClause_8()
	throws RecognitionException, TokenStreamException
    {
	JmlRequiresClause rc = 
	    parserHelper.getRequiresClause( "pre_redundantly \\not_specified;" );
	assertTrue( rc.isRedundantly() );
	assertTrue( rc.isNotSpecified() );
    }

    // -------------------------------------------------------------
    // when-clause
    // -------------------------------------------------------------
    
    public void test_whenClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("when true;");
	JmlWhenClause ec = (JmlWhenClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_whenClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("when \\not_specified;");
	JmlWhenClause ec = (JmlWhenClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    public void test_whenClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("when_redundantly true;");
	JmlWhenClause ec = (JmlWhenClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_whenClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"when_redundantly \\not_specified;");
	JmlWhenClause ec = (JmlWhenClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    // -------------------------------------------------------------
    // working-space-clause
    // -------------------------------------------------------------
    
    public void test_workingSpaceClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "working_space \\not_specified;");
	JmlWorkingSpaceClause mc = (JmlWorkingSpaceClause) bc;
	assertFalse( mc.isRedundantly() );
	assertTrue( mc.isNotSpecified() );
	assertTrue( mc.specExp() == null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_workingSpaceClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "working_space i + 1;");
	JmlWorkingSpaceClause mc = (JmlWorkingSpaceClause) bc;
	assertFalse( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_workingSpaceClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "working_space i if true;");
	JmlWorkingSpaceClause mc = (JmlWorkingSpaceClause) bc;
	assertFalse( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() != null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_workingSpaceClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "working_space_redundantly \\not_specified;");
	JmlWorkingSpaceClause mc = (JmlWorkingSpaceClause) bc;
	assertTrue( mc.isRedundantly() );
	assertTrue( mc.isNotSpecified() );
	assertTrue( mc.specExp() == null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_workingSpaceClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "working_space_redundantly i + 1;");
	JmlWorkingSpaceClause mc = (JmlWorkingSpaceClause) bc;
	assertTrue( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_workingSpaceClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "working_space_redundantly i if true;");
	JmlWorkingSpaceClause mc = (JmlWorkingSpaceClause) bc;
	assertTrue( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() != null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    // -------------------------------------------------------------
    // duration-clause
    // -------------------------------------------------------------
    
    public void test_durationClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "duration \\not_specified;");
	JmlDurationClause mc = (JmlDurationClause) bc;
	assertFalse( mc.isRedundantly() );
	assertTrue( mc.isNotSpecified() );
	assertTrue( mc.specExp() == null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_durationClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "duration i + 1;");
	JmlDurationClause mc = (JmlDurationClause) bc;
	assertFalse( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_durationClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "duration i if true;");
	JmlDurationClause mc = (JmlDurationClause) bc;
	assertFalse( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() != null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_durationClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "duration_redundantly \\not_specified;");
	JmlDurationClause mc = (JmlDurationClause) bc;
	assertTrue( mc.isRedundantly() );
	assertTrue( mc.isNotSpecified() );
	assertTrue( mc.specExp() == null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_durationClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "duration_redundantly i + 1;");
	JmlDurationClause mc = (JmlDurationClause) bc;
	assertTrue( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_durationClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "duration_redundantly i if true;");
	JmlDurationClause mc = (JmlDurationClause) bc;
	assertTrue( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() != null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    // -------------------------------------------------------------
    // measured-clause
    // -------------------------------------------------------------
    
/* !FIXME! - reinstate in some form when the measured clause is defined
    public void test_measuredClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "measured_by \\not_specified;");
	JmlMeasuredClause mc 
	    = (JmlMeasuredClause) mspec.subclassingContract()[0];
	assertFalse( mc.isRedundantly() );
	assertTrue( mc.isNotSpecified() );
	assertTrue( mc.specExp() == null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_measuredClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "measured_by i + 1;");
	JmlMeasuredClause mc 
	    = (JmlMeasuredClause) mspec.subclassingContract()[0];
	assertFalse( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_measuredClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "measured_by i if true;");
	JmlMeasuredClause mc 
	    = (JmlMeasuredClause) mspec.subclassingContract()[0];
	assertFalse( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() != null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_measuredClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "measured_by_redundantly \\not_specified;");
	JmlMeasuredClause mc 
	    = (JmlMeasuredClause) mspec.subclassingContract()[0];
	assertTrue( mc.isRedundantly() );
	assertTrue( mc.isNotSpecified() );
	assertTrue( mc.specExp() == null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_measuredClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "measured_by_redundantly i + 1;");
	JmlMeasuredClause mc 
	    = (JmlMeasuredClause) mspec.subclassingContract()[0];
	assertTrue( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() == null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }

    public void test_measuredClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlMethodSpecification mspec = 
	    methodSpecificationFrom(
	       "measured_by_redundantly i if true;");
	JmlMeasuredClause mc 
	    = (JmlMeasuredClause) mspec.subclassingContract()[0];
	assertTrue( mc.isRedundantly() );
	assertFalse( mc.isNotSpecified() );
	assertTrue( mc.specExp() != null);
	assertTrue( mc.predOrNot() != null);
	//@ assert mc.specExp() == null ==> mc.predOrNot() == null;
	assertTrue( mc.specExp() != null || mc.predOrNot() == null);
    }
*/


    // -------------------------------------------------------------
    // assignable-clause
    // -------------------------------------------------------------
    
    public void test_assignableClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable i;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertFalse( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_assignableClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable i, j;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertFalse( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );
    }

    public void test_assignableClause_7()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable_redundantly i;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertTrue( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_assignableClause_9()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable_redundantly i, j;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertTrue( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );
    }

    public void test_assignableClause_13()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifiable i;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertFalse( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_assignableClause_15()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifiable i, j;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertFalse( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );
    }

    public void test_assignableClause_19()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifiable_redundantly i;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertTrue( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_assignableClause_21()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifiable_redundantly i, j;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertTrue( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );
    }

    public void test_assignableClause_25()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifies i;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertFalse( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_assignableClause_27()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifies i, j;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertFalse( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );
    }

    public void test_assignableClause_31()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifies_redundantly i;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertTrue( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 1 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
    }

    public void test_assignableClause_33()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "modifies_redundantly i, j;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertTrue( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );
    }

    public void test_assignableClause_37()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	       "assignable i.*, j.*;");
	JmlAssignableClause ac = (JmlAssignableClause) bc;
	assertFalse( ac.isRedundantly() );
	assertTrue( ac.storeRefs().length == 2 );
	JmlStoreRef sref = ac.storeRefs()[0];
	assertTrue( sref != null );
	JmlStoreRefExpression srefExpr 
	    = (JmlStoreRefExpression) sref;
	JmlName[] subnames = srefExpr.names();
	assertTrue( subnames.length == 2 );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "i", subnames[0].ident() );
	assertTrue( subnames[1].isFields() );
	sref = ac.storeRefs()[1];
	assertTrue( sref != null );

	srefExpr = (JmlStoreRefExpression) sref;
	subnames = srefExpr.names();
	assertTrue( subnames.length == 2 );
	assertTrue( subnames[0].isIdent() );
	assertEquals( "j", subnames[0].ident() );
	assertTrue( subnames[1].isFields() );
    }

    // -------------------------------------------------------------
    // ensures-clause
    // -------------------------------------------------------------
    
    public void test_ensuresClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("ensures true;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_ensuresClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("ensures \\not_specified;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    public void test_ensuresClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("ensures_redundantly true;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_ensuresClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"ensures_redundantly \\not_specified;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    public void test_ensuresClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("post true;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_ensuresClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("post \\not_specified;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    public void test_ensuresClause_7()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("post_redundantly true;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_ensuresClause_8()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"post_redundantly \\not_specified;");
	JmlEnsuresClause ec = (JmlEnsuresClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    // -------------------------------------------------------------
    // signals-clause
    // -------------------------------------------------------------
    
    public void test_signalsClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "signals ( Exception );");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"signals_redundantly ( Exception );");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "exsures ( Exception );");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures_redundantly ( Exception );");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_5()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "signals ( Exception ) \\not_specified;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertTrue( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_6()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"signals_redundantly ( Exception ) \\not_specified;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertTrue( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_7()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "exsures ( Exception ) \\not_specified;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertTrue( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_8()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures_redundantly ( Exception ) \\not_specified;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertTrue( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_9()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "signals ( Exception ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_10()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"signals_redundantly ( Exception ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_11()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "exsures ( Exception ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_12()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures_redundantly ( Exception ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertTrue( sc.ident() == null );
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_13()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "signals ( Exception e ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_14()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"signals_redundantly ( Exception e ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_15()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
	        "exsures ( Exception e ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_16()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures_redundantly ( Exception e ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_17()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures_redundantly ( Exception e );");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_18()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures ( Exception e ) true;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertFalse( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() != null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_19()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures_redundantly ( Exception e ) \\not_specified;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertTrue( sc.isRedundantly() );
	assertTrue( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    public void test_signalsClause_20()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"exsures ( Exception e ) \\not_specified;");
	JmlSignalsClause sc = (JmlSignalsClause) bc;
	assertFalse( sc.isRedundantly() );
	assertTrue( sc.isNotSpecified() );
	assertTrue( sc.type() != null );
	assertEquals( "e", sc.ident());
	assertTrue( sc.predOrNot() == null );
	//@ assert sc.isNotSpecified() ==> sc.predOrNot() == null;
	assertTrue( !sc.isNotSpecified() || sc.predOrNot() == null);
    }

    // -------------------------------------------------------------
    // diverges-clause
    // -------------------------------------------------------------
    
    public void test_divergesClause_1()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("diverges true;");
	JmlDivergesClause ec = (JmlDivergesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_divergesClause_2()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("diverges \\not_specified;");
	JmlDivergesClause ec = (JmlDivergesClause) bc;
	assertFalse( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    public void test_divergesClause_3()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause("diverges_redundantly true;");
	JmlDivergesClause ec = (JmlDivergesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertFalse( ec.isNotSpecified() );
    }

    public void test_divergesClause_4()
	throws RecognitionException, TokenStreamException
    {
	JmlSpecBodyClause bc = 
	    parserHelper.getSpecBodyClause(
		"diverges_redundantly \\not_specified;");
	JmlDivergesClause ec = (JmlDivergesClause) bc;
	assertTrue( ec.isRedundantly() );
	assertTrue( ec.isNotSpecified() );
    }

    //=====================================================================
    // Predicate and Specification Expression Syntax
    //=====================================================================

    // -------------------------------------------------------------
    // spec-expression-list
    // -------------------------------------------------------------

    public void test_specExpressionList_1()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x =this; @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecExpression[] list = specExpressionListFrom( unit );
	assertEquals( 1, list.length );
    }
        
    public void test_specExpressionList_2()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x= this, that; @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecExpression[] list = specExpressionListFrom( unit );
	assertEquals( 2, list.length );
    }
        


    // -------------------------------------------------------------
    // jml-primary
    // -------------------------------------------------------------

    public void test_jmlPrimary_1()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x =\\result; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlResultExpression );
    }

    public void test_jmlPrimary_2()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x=\\old( this ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlOldExpression) expr).specExpression().expression()
		    instanceof JThisExpression );
    }

    public void test_jmlPrimary_3()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\not_modified( x ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertEquals( 1, ((JmlNotModifiedExpression) expr).storeRefList().length);
    }

    public void test_jmlPrimary_4()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\fresh( x ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertEquals( 1, ((JmlFreshExpression) expr).specExpressionList().length);
    }

    public void test_jmlPrimary_5()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\reach( this ); @*/" + 
				    NEWLINE +
				    "}" );
	JmlReachExpression expr = (JmlReachExpression) expressionFrom( unit );
	assertTrue( expr.specExpression().expression()
		    instanceof JThisExpression );
	assertTrue( expr.referenceType() == null );
	assertTrue( expr.storeRefExpression() == null );
    }

    public void test_jmlPrimary_6()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\reach( this, Sample ); @*/" + 
				    NEWLINE +
				    "}" );
	JmlReachExpression expr = (JmlReachExpression) expressionFrom( unit );
	assertTrue( expr.specExpression().expression()
		    instanceof JThisExpression );
	assertEquals( "Sample",
		      ((CClassType) expr.referenceType()).qualifiedName(),
		      true );
	assertTrue( expr.storeRefExpression() == null );
    }

    public void test_jmlPrimary_7()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\reach( this, Sample, foo[*] ); @*/" + 
				    NEWLINE +
				    "}" );
	JmlReachExpression expr = (JmlReachExpression) expressionFrom( unit );
	assertTrue( expr.specExpression().expression()
		    instanceof JThisExpression );
	assertEquals( "Sample",
		      ((CClassType) expr.referenceType()).qualifiedName(),
		      true );
	assertTrue( expr.storeRefExpression() != null );
    }

    public void test_jmlPrimary_8()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = (* some thing *); */" + NEWLINE +
				    "}" );
	JmlInformalExpression expr = 
	    (JmlInformalExpression) expressionFrom(unit);
	assertEquals( " some thing ", expr.text() );
    }

    public void test_jmlPrimary_9()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\nonnullelements( this ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlNonNullElementsExpression) expr).specExpression().expression()
		    instanceof JThisExpression );
    }

    public void test_jmlPrimary_10()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\typeof( this ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlTypeOfExpression) expr).specExpression().expression()
		    instanceof JThisExpression );
    }

    public void test_jmlPrimary_11()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\elemtype( \\typeof(this) ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlElemTypeExpression) expr).specExpression().expression()
		    instanceof JmlTypeOfExpression );
    }

    public void test_jmlPrimary_12()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\type( int ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue(((JmlTypeExpression) expr).type().equals(CStdType.Integer));
    }

    public void test_jmlPrimary_13()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\lockset; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlLockSetExpression );
    }

    public void test_jmlPrimary_14()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\is_initialized( Sample ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertEquals( "Sample", 
		      ((CClassType) ((JmlIsInitializedExpression) expr).referenceType()).qualifiedName(), 
		       true );
    }

    public void test_jmlPrimary_15()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\invariant_for( this ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlInvariantForExpression) expr).specExpression().expression()
		    instanceof JThisExpression );
    }

    public void test_jmlPrimary_16()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\lblneg foo this); @*/" + 
				    NEWLINE +
				    "}" );
	JmlLabelExpression expr = (JmlLabelExpression) expressionFrom( unit );
	assertTrue( expr.specExpression().expression()
		    instanceof JThisExpression );
	assertFalse( expr.isPosLabel() );
	assertEquals( "foo", expr.ident() );
    }

    public void test_jmlPrimary_17()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\lblpos foo this); @*/" + 
				    NEWLINE +
				    "}" );
	JmlLabelExpression expr = (JmlLabelExpression) expressionFrom( unit );
	assertTrue( expr.specExpression().expression()
		    instanceof JThisExpression );
	assertTrue( expr.isPosLabel() );
	assertEquals( "foo", expr.ident() );
    }

    public void test_jmlPrimary_18()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\duration( this.m() ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlDurationExpression) expr).expression()
		    instanceof JMethodCallExpression );
    }

    public void test_jmlPrimary_19()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\working_space( this.m() ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlWorkingSpaceExpression) expr).expression()
		    instanceof JMethodCallExpression );
    }

    public void test_jmlPrimary_20()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\space( this ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( ((JmlSpaceExpression) expr).specExpression().expression()
		    instanceof JThisExpression );
    }

    public void test_jmlPrimary_21()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\not_assigned( x ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertEquals( 1, ((JmlNotAssignedExpression) expr).storeRefList().length);
    }

    public void test_jmlPrimary_22()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = \\only_assigned( x, y ); @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertEquals( 2, ((JmlOnlyAssignedExpression) expr).storeRefList().length);
    }

    // -------------------------------------------------------------
    // set-comprehension
    // -------------------------------------------------------------
    
    public void test_setComprehension_1()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i | this.has(i) && i != null }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_2()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i | this.f.getSet().has(i) && i != null }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_3()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer[] i | this.has(i) && i != null }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_4()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer[] i | this.f.getSet().has(i) && i != null }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_5()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i[] | this.has(i) && i != null }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_6()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i[] | this.f.getSet().has(i) && i != null }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_7()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i | this.has(i) && i != null && i.intValue() > 3 }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_8()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i | this.f.getSet().has(i) && i != null && i.intValue() > 3 }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_9()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer[] i | this.has(i) && i != null && i.intValue() > 3 }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_10()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer[] i | this.f.getSet().has(i) && i != null && i.intValue() > 3 }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_11()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i[] | this.has(i) && i != null && i.intValue() > 3 }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    public void test_setComprehension_12()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = new JmlObjectSet { Integer i[] | this.f.getSet().has(i) && i != null && i.intValue() > 3 }; @*/" + 
				    NEWLINE +
				    "}" );
	JExpression expr = expressionFrom( unit );
	assertTrue( expr instanceof JmlSetComprehension );
	assertEquals( "i", ((JmlSetComprehension) expr).ident() );
    }

    // -------------------------------------------------------------
    // spec-quantified-expr
    // -------------------------------------------------------------

    public void test_specQuantifiedExpr_1()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isForAll() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_2()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i;;true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isForAll() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_3()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i; false; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isForAll() );
	assertTrue( expr.hasPredicate() );
	assertTrue( expr.predicate().specExpression().expression()
		    instanceof JBooleanLiteral );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_4()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\exists int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isExists() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_5()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\exists int i;;true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isExists() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_6()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\exists int i; false; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isExists() );
	assertTrue( expr.hasPredicate() );
	assertTrue( expr.predicate().specExpression().expression()
		    instanceof JBooleanLiteral );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_7()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\max int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isMax() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_8()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\max int i;;true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isMax() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_9()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\max int i; false; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isMax() );
	assertTrue( expr.hasPredicate() );
	assertTrue( expr.predicate().specExpression().expression()
		    instanceof JBooleanLiteral );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_10()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\min int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isMin() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_11()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\min int i;;true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isMin() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_12()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\min int i; false; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isMin() );
	assertTrue( expr.hasPredicate() );
	assertTrue( expr.predicate().specExpression().expression()
		    instanceof JBooleanLiteral );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_13()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\num_of int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isNumOf() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_14()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\num_of int i;;true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isNumOf() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_15()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\num_of int i; false; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isNumOf() );
	assertTrue( expr.hasPredicate() );
	assertTrue( expr.predicate().specExpression().expression()
		    instanceof JBooleanLiteral );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_16()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\product int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isProduct() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_17()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\product int i;;true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isProduct() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_18()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\product int i; false; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isProduct() );
	assertTrue( expr.hasPredicate() );
	assertTrue( expr.predicate().specExpression().expression()
		    instanceof JBooleanLiteral );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_19()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\sum int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isSum() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_20()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\sum int i;;true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isSum() );
	assertFalse( expr.hasPredicate() );
	assertTrue( expr.predicate() == null );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    public void test_specQuantifiedExpr_21()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\sum int i; false; true); @*/" + 
				    NEWLINE +
				    "}" );
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	assertTrue( expr.isSum() );
	assertTrue( expr.hasPredicate() );
	assertTrue( expr.predicate().specExpression().expression()
		    instanceof JBooleanLiteral );
	assertTrue( expr.specExpression().expression() 
		    instanceof JBooleanLiteral );
    }

    // -------------------------------------------------------------
    // quantified-var-decl
    // -------------------------------------------------------------

    private JVariableDefinition[] 
	quantifiedVarDeclsFrom( JmlCompilationUnit unit ) 
    {
	JmlSpecQuantifiedExpression expr = 
	    (JmlSpecQuantifiedExpression) expressionFrom( unit );
	return expr.quantifiedVarDecls();
    }

    public void test_quantifiedVarDecl_1()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 1, decls.length );
	assertEquals( decls[0].getType(), CStdType.Integer );
	assertEquals( "i", decls[0].ident() );
    }

    public void test_quantifiedVarDecl_2()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i, j; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 2, decls.length );
	assertEquals( decls[0].getType(), CStdType.Integer );
	assertEquals( "i", decls[0].ident() );
	assertEquals( decls[1].getType(), CStdType.Integer );
	assertEquals( "j", decls[1].ident() );
    }

     public void test_quantifiedVarDecl_3()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i, j, k; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 3, decls.length );
	assertEquals( decls[0].getType(), CStdType.Integer );
	assertEquals( "i", decls[0].ident() );
	assertEquals( decls[1].getType(), CStdType.Integer );
	assertEquals( "j", decls[1].ident() );
	assertEquals( decls[2].getType(), CStdType.Integer );
	assertEquals( "k", decls[2].ident() );
    }

    public void test_quantifiedVarDecl_4()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int[] i; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 1, decls.length );
	assertEquals( 1, ((CArrayType) decls[0].getType()).getArrayBound() );
	assertEquals( "i", decls[0].ident() );
    }

    public void test_quantifiedVarDecl_5()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int[] i, j; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 2, decls.length );
	assertEquals( 1, ((CArrayType) decls[0].getType()).getArrayBound() );
	assertEquals( "i", decls[0].ident() );
	assertEquals( 1, ((CArrayType) decls[1].getType()).getArrayBound() );
	assertEquals( "j", decls[1].ident() );
    }

     public void test_quantifiedVarDecl_6()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int[] i, j, k; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 3, decls.length );
	assertEquals( 1, ((CArrayType) decls[0].getType()).getArrayBound() );
	assertEquals( "i", decls[0].ident() );
	assertEquals( 1, ((CArrayType) decls[1].getType()).getArrayBound() );
	assertEquals( "j", decls[1].ident() );
	assertEquals( 1, ((CArrayType) decls[2].getType()).getArrayBound() );
	assertEquals( "k", decls[2].ident() );
    }

    public void test_quantifiedVarDecl_7()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i[]; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 1, decls.length );
	assertEquals( 1, ((CArrayType) decls[0].getType()).getArrayBound() );
	assertEquals( "i", decls[0].ident() );
    }

    public void test_quantifiedVarDecl_8()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i[], j[][]; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 2, decls.length );
	assertEquals( 1, ((CArrayType) decls[0].getType()).getArrayBound() );
	assertEquals( "i", decls[0].ident() );
	assertEquals( 2, ((CArrayType) decls[1].getType()).getArrayBound() );
	assertEquals( "j", decls[1].ident() );
    }

     public void test_quantifiedVarDecl_9()
	throws RecognitionException, TokenStreamException
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "  int x; /*@ monitors_for x = (\\forall int i[], j[][], k[][][]; true); @*/" + 
				    NEWLINE +
				    "}" );
	JVariableDefinition[] decls = quantifiedVarDeclsFrom( unit );
	assertEquals( 3, decls.length );
	assertEquals( 1, ((CArrayType) decls[0].getType()).getArrayBound() );
	assertEquals( "i", decls[0].ident() );
	assertEquals( 2, ((CArrayType) decls[1].getType()).getArrayBound() );
	assertEquals( "j", decls[1].ident() );
	assertEquals( 3, ((CArrayType) decls[2].getType()).getArrayBound() );
	assertEquals( "k", decls[2].ident() );
    }


    // -------------------------------------------------------------
    // spec-variable-declarators
    // -------------------------------------------------------------

    public void test_specVariableDeclarators_1()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old int x=0;" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 1, vars.length );
	assertEquals( CStdType.Integer, vars[0].getType() );
	assertEquals( "x", vars[0].ident() );
	assertTrue( vars[0].hasInitializer() );
	assertTrue( vars[0].getValue() instanceof JmlSpecExpression );
    }
        
    public void test_specVariableDeclarators_2()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old int x;" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 1, vars.length );
	assertEquals( CStdType.Integer, vars[0].getType() );
	assertEquals( "x", vars[0].ident() );
	assertFalse( vars[0].hasInitializer() );
    }
        
    public void test_specVariableDeclarators_3()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old int[] x={ 0 };" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 1, vars.length );
	assertEquals( 1, ((CArrayType) vars[0].getType()).getArrayBound() );
	assertEquals( "x", vars[0].ident() );
	assertTrue( vars[0].hasInitializer() );
	assertTrue( vars[0].getValue() instanceof JArrayInitializer );
    }
        
    public void test_specVariableDeclarators_4()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old Object[] x={ 0, {1} };" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 1, vars.length );
	assertEquals( 1, ((CArrayType) vars[0].getType()).getArrayBound() );
	assertEquals( "x", vars[0].ident() );
	assertTrue( vars[0].hasInitializer() );
	assertTrue( vars[0].getValue() instanceof JArrayInitializer );
    }
        
    public void test_specVariableDeclarators_5()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old int x=0, y=0;" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 2, vars.length );
	assertEquals( CStdType.Integer, vars[0].getType() );
	assertEquals( "x", vars[0].ident() );
	assertTrue( vars[0].hasInitializer() );
	assertTrue( vars[0].getValue() instanceof JmlSpecExpression );
	assertEquals( CStdType.Integer, vars[1].getType() );
	assertEquals( "y", vars[1].ident() );
	assertTrue( vars[1].hasInitializer() );
	assertTrue( vars[1].getValue() instanceof JmlSpecExpression );
    }
        
    public void test_specVariableDeclarators_6()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old int x, y;" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 2, vars.length );
	assertEquals( CStdType.Integer, vars[0].getType() );
	assertEquals( "x", vars[0].ident() );
	assertFalse( vars[0].hasInitializer() );
 	assertEquals( CStdType.Integer, vars[1].getType() );
	assertEquals( "y", vars[1].ident() );
	assertFalse( vars[1].hasInitializer() );
   }
        
    public void test_specVariableDeclarators_7()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old int[] x={ 0 }, y={0};" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 2, vars.length );
	assertEquals( 1, ((CArrayType) vars[0].getType()).getArrayBound() );
	assertEquals( "x", vars[0].ident() );
	assertTrue( vars[0].hasInitializer() );
	assertTrue( vars[0].getValue() instanceof JArrayInitializer );
	assertEquals( 1, ((CArrayType) vars[1].getType()).getArrayBound() );
	assertEquals( "y", vars[1].ident() );
	assertTrue( vars[1].hasInitializer() );
	assertTrue( vars[1].getValue() instanceof JArrayInitializer );
    }
        
    public void test_specVariableDeclarators_8()
	throws RecognitionException, TokenStreamException
    {
	JmlLetVarDecl decl = 
	    letVarDeclsFrom( parserHelper.
			     getSpecVarDecls( "old Object[] x={ 0, {1} }, y={ 0, {1} };" ) )[0];

	JVariableDefinition[] vars = decl.specVariableDeclarators();
	assertEquals( 2, vars.length );
	assertEquals( 1, ((CArrayType) vars[0].getType()).getArrayBound() );
	assertEquals( "x", vars[0].ident() );
	assertTrue( vars[0].hasInitializer() );
	assertTrue( vars[0].getValue() instanceof JArrayInitializer );
	assertEquals( 1, ((CArrayType) vars[1].getType()).getArrayBound() );
	assertEquals( "y", vars[1].ident() );
	assertTrue( vars[1].hasInitializer() );
	assertTrue( vars[1].getValue() instanceof JArrayInitializer );
    }
        
    //=====================================================================
    // Statement and Annotation Statement Syntax
    //=====================================================================


    // -------------------------------------------------------------
    // loop-stmt
    // -------------------------------------------------------------

    public void test_loopStmt_1()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement( "for( int i=0; i<2; i++ );" );
	assertFalse( statement instanceof JmlLoopStatement );
    }

    public void test_loopStmt_2()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement( "while(true) {}" );
	assertFalse( statement instanceof JmlLoopStatement );
    }

    public void test_loopStmt_3()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement( "do {} while(true);" );
	assertFalse( statement instanceof JmlLoopStatement );
    }

    public void test_loopStmt_4()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_5()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_6()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_7()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_8()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_9()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_10()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_11()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_12()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_13()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "maintaining true;" + NEWLINE + 
				      "for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 2, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_14()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "maintaining true;" + NEWLINE + 
				      "while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 2, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_15()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "maintaining true;" + NEWLINE + 
				      "do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 2, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_16()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 2, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_17()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 2, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_18()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 2, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_19()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 2, loopStmt.loopInvariants().length );
	assertEquals( 2, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_20()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 2, loopStmt.loopInvariants().length );
	assertEquals( 2, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_21()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 2, loopStmt.loopInvariants().length );
	assertEquals( 2, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_22()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "l1: for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_23()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "l1: while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_24()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "l1: do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 0, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_25()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "l1: for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_26()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "l1: while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_27()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("decreasing x;" + NEWLINE + 
				      "l1: do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 0, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_28()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "l1: for( int i=0; i<2; i++ );" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_29()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "l1: while(true) {}" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    public void test_loopStmt_30()
	throws RecognitionException, TokenStreamException 
    {
	JStatement statement = 
	    parserHelper.getStatement("maintaining true;" + NEWLINE + 
				      "decreasing x;" + NEWLINE + 
				      "l1: do {} while(true);" );
	JmlLoopStatement loopStmt = (JmlLoopStatement) statement;
	assertEquals( 1, loopStmt.loopInvariants().length );
	assertEquals( 1, loopStmt.variantFunctions().length );
    }

    // -------------------------------------------------------------
    // hence-by-statement
    // -------------------------------------------------------------

    public void test_henceByStatement_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlHenceByStatement statement = (JmlHenceByStatement)
	    parserHelper.getStatement( "hence_by true;" );
	assertFalse( statement.isRedundantly() );
    }

    public void test_henceByStatement_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlHenceByStatement statement = (JmlHenceByStatement)
	    parserHelper.getStatement( "hence_by_redundantly true;" );
	assertTrue( statement.isRedundantly() );
    }

    // -------------------------------------------------------------
    // assert-statement
    // -------------------------------------------------------------
    
    public void test_assertStatement_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssertStatement statement = (JmlAssertStatement)
	    parserHelper.getStatement( "assert 1==2;" );
	assertFalse( statement.hasThrowMessage() );
    }

    public void test_assertStatement_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssertStatement statement = (JmlAssertStatement)
	    parserHelper.getStatement( "assert 1==2 : true;" );
	assertTrue( statement.hasThrowMessage() );
    }

    // -------------------------------------------------------------
    // assert-redundantly-statement
    // -------------------------------------------------------------
    
    public void test_assertRedundantlyStatement_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssertStatement statement = (JmlAssertStatement)
	    parserHelper.getStatement( "assert_redundantly 1==2;" );
	assertTrue( statement.isRedundantly() );
	assertFalse( statement.hasThrowMessage() );
    }

    public void test_assertRedundantlyStatement_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssertStatement statement = (JmlAssertStatement)
	    parserHelper.getStatement( "assert_redundantly 1==2 : true;" );
	assertTrue( statement.isRedundantly() );
	assertTrue( statement.hasThrowMessage() );
    }


    // -------------------------------------------------------------
    // assume-statement
    // -------------------------------------------------------------

    public void test_assumeStatement_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssumeStatement statement = (JmlAssumeStatement)
	    parserHelper.getStatement( "assume 1==2;" );
	assertFalse( statement.hasThrowMessage() );
	assertFalse( statement.isRedundantly() );
    }

    public void test_assumeStatement_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssumeStatement statement = (JmlAssumeStatement)
	    parserHelper.getStatement( "assume 1==2 : true;" );
	assertTrue( statement.hasThrowMessage() );
	assertFalse( statement.isRedundantly() );
    }

    public void test_assumeStatement_3()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssumeStatement statement = (JmlAssumeStatement)
	    parserHelper.getStatement( "assume_redundantly 1==2;" );
	assertFalse( statement.hasThrowMessage() );
	assertTrue( statement.isRedundantly() );
    }

    public void test_assumeStatement_4()
	throws RecognitionException, TokenStreamException 
    {
	JmlAssumeStatement statement = (JmlAssumeStatement)
	    parserHelper.getStatement( "assume_redundantly 1==2 : true;" );
	assertTrue( statement.hasThrowMessage() );
	assertTrue( statement.isRedundantly() );
    }


    // -------------------------------------------------------------
    // set-statement
    // -------------------------------------------------------------

    public void test_setStatement_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlSetStatement statement = (JmlSetStatement)
	    parserHelper.getStatement( "set x = 1;" );
    }

    public void test_setStatement_2()
	throws RecognitionException, TokenStreamException 
    {
	JmlSetStatement statement = (JmlSetStatement)
	    parserHelper.getStatement( "set x += 1;" );
    }


    // -------------------------------------------------------------
    // unreachable-statement
    // -------------------------------------------------------------

    public void test_unreachableStatement_1()
	throws RecognitionException, TokenStreamException 
    {
	JmlUnreachableStatement statement = (JmlUnreachableStatement)
	    parserHelper.getStatement( "unreachable;" );
    }



    //=====================================================================
    // Expression Syntax
    //=====================================================================

    // -------------------------------------------------------------
    // conditional-expr
    // -------------------------------------------------------------

    public void test_conditionalExpr_1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <==> false ? this : null; */" + NEWLINE +
				    "}" );
	JExpression expr = expressionFrom(unit);
	assertTrue( expr instanceof JConditionalExpression );
    }

    public void test_conditionalExpr_2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <==> false ? true ? this : null : false ? this : null; */" + NEWLINE +
				    "}" );
	JExpression expr = expressionFrom(unit);
	assertTrue( expr instanceof JConditionalExpression );
    }

    // -------------------------------------------------------------
    // equivalence-expr
    // -------------------------------------------------------------

    public void test_equivalenceExpr_1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <==> false; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isEquivalence() );
    }

    public void test_equivalenceExpr_2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <==> false <==> 1 == 2; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isEquivalence() );
    }

    public void test_equivalenceExpr_3()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <=!=> false; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isNonEquivalence() );
    }

    public void test_equivalenceExpr_4()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <=!=> false <==> 1 == 1; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isNonEquivalence() );
    }

    public void test_equivalenceExpr_5()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <==> false <=!=> 1 == 1; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isEquivalence() );
    }


    // -------------------------------------------------------------
    // implies-expr
    // -------------------------------------------------------------

    public void test_impliesExpr_1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = false ==> true; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isImplication() );
    }

    public void test_impliesExpr_2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = false <== 1 == 2; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isBackwardImplication() );
    }

    public void test_impliesExpr_3()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true ==> false ==> 1 == 2; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isImplication() );
    }

    public void test_impliesExpr_4()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				    "int x; /*@ monitors_for x = true <== false <== 1 == 2; */" + NEWLINE +
				    "}" );
	JmlRelationalExpression expr = 
	    (JmlRelationalExpression) expressionFrom(unit);
	assertTrue( expr.isBackwardImplication() );
    }

    public void test_impliesExpr_5()  
	throws RecognitionException, TokenStreamException 
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"int x; /*@ monitors_for x = true ==> false <== 1 == 2; */" + NEWLINE +
					"}" );
	    fail( "should get exception" );
	} catch (RecognitionException e) {
	    assertEquals( "unexpected token: <==", e.getMessage() );
	} // end of catch
    }

    public void test_impliesExpr_6()  
	throws RecognitionException, TokenStreamException 
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"int x; /*@ monitors_for x = true <== false ==> 1 == 2; */" + NEWLINE +
					"}" );
	    fail( "should get exception" );
	} catch (RecognitionException e) {
	    assertEquals( "unexpected token: ==>", e.getMessage() );
	} // end of catch
    }


    // -------------------------------------------------------------
    // primary-expr, see jml-primary
    // -------------------------------------------------------------

    //=====================================================================
    // Various Combinations of Syntax
    //=====================================================================

//      //---------------------------------------------------------------------
//      // Mixed annotation members
//      //---------------------------------------------------------------------

//      /**
//       * Should find method and type specification annotation. */
//      public void test_mixed_annotation_1()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ invariant true; */" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find one type annotations", 1, tNotes.length );
//  	assertEquals( "should find one method annotation", 1, mNotes.length );
//      }

//      /**
//       * Should find method and type specification annotation. */
//      public void test_mixed_annotation_2()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ invariant true;" + NEWLINE +
//  				 "    @ ensures true; */" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find one type annotations", 1, tNotes.length );
//  	assertEquals( "should find one method annotation", 1, mNotes.length );
//      }

    //=====================================================================
    // Bad Annotations
    //=====================================================================

    /**
     * Should not allow multi-line annotation within single-line
     * annotation. */
    public void test_bad_annotation_2()  
	throws RecognitionException 
    {
	try {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//@ invariant /*@ invariant true; */ true;" 
				 + NEWLINE +
				 "}" );
	assertFalse( "should get exception", true );
	} catch( TokenStreamException e ) {
	    assertEquals( "error: An annotation may not contain another annotation [JML]", e.getMessage(), true );
	}
    }

    /**
     * Should allow single-line annotation within single-line
     * annotation. */
    public void test_annotation_3()  
	throws RecognitionException 
    {
	try {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//@ invariant //@ true;" 
				 + NEWLINE +
				 "}" );
	// will throw an exception if parses incorrectly
		assertTrue( "should not get exception", true );
	} catch( TokenStreamException e ) {
	    assertEquals( "error: An annotation may contain another annotation [JML]", e.getMessage(), false );
	}
    }

    /**
     * Should not allow multi-line annotation within multi-line
     * annotation. */
    public void test_bad_annotation_4()  
	throws RecognitionException 
    {
	try {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ invariant /*@ invariant true; */" + 
				 NEWLINE +
				 "  @ true;" + NEWLINE +
				 "  @*/" + NEWLINE +
				 "}" );
	assertFalse( "should get exception", true );
	} catch( TokenStreamException e ) {
	    assertEquals( "error: An annotation may not contain another annotation [JML]", e.getMessage(), true );
	}
    }

    /**
     * Should allow single-line annotation within multi-line
     * annotation. */
    public void test_annotation_5()  
	throws RecognitionException 
    {
	try {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ invariant //@ true;" + 
				 NEWLINE +
				 "  @*/" + NEWLINE +
				 "}" );
	// will throw exception if parses incorrectly
	} catch( TokenStreamException e ) {
	    assertEquals( "error: An annotation may contain another SL annotation [JML]", e.getMessage(), false );
	}
    }

//      /**
//       * Should get warning on misplaced method annotation. */
//      public void test_bad_annotation_6()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  /*@ invariant true; */" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find one type annotations", 1, tNotes.length );
//  	assertEquals( "should find one method annotations", 1, mNotes.length );
//  	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 4, character 7 error: Behavioral specification for method is not followed by a method declaration; there is an intervening type specification member [JML]" + NEWLINE, 
//  		      parserHelper.errorOut.toString(), true );
//      }

//      /**
//       * Should get warning on misplaced method annotation. */
//      public void test_bad_annotation_7()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ invariant true; */" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find one type annotations", 1, tNotes.length );
//  	assertEquals( "should find no method annotations", 0, mNotes.length );
//  	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 5, character 1 caution:Behavioral specification(s) for a method discarded because the end of the type declaration block was reached without an intervening method [JML]" + NEWLINE, 
//  		      parserHelper.errorOut.toString(), true );
//      }

//      /**
//       * Should get warning on misplaced method annotation. */
//      public void test_bad_annotation_8()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ invariant true; */" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  ;" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find one type annotations", 1, tNotes.length );
//  	assertEquals( "should find no method annotations", 0, mNotes.length );
//  	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 4, character 3 caution:Behavioral specification(s) for a method discarded because of stray semicolon [JML]" + NEWLINE + "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 4, character 3 warning: Stray semicolon [KOPI]" + NEWLINE, 
//  		      parserHelper.errorOut.toString(), true );
//      }

//      /**
//       * Should get warning on misplaced method annotation. */
//      public void test_bad_annotation_9()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ invariant true; */" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  int x;" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find one type annotations", 1, tNotes.length );
//  	assertEquals( "should find no method annotations", 0, mNotes.length );
//  	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 19 error: Behavioral specification for method is not followed by a method declaration; there is a field or type declaration instead [JML]" + NEWLINE, 
//  		      parserHelper.errorOut.toString(), true );
//      }

//      /**
//       * Should get warning on misplaced method annotation. */
//      public void test_bad_annotation_10()  
//  	throws RecognitionException, TokenStreamException 
//      {
//  	JmlCompilationUnit unit = 
//  	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
//  				 "  /*@ invariant true; */" + NEWLINE +
//  				 "  /*@ ensures true; */" + NEWLINE +
//  				 "  class In { }" + NEWLINE +
//  				 "  void m() { }" + NEWLINE +
//  				 "}" );
//  	JmlNode[] tNotes = unit.typeDeclarations()[0].jmlAnnotations();
//  	JMethodDeclaration m = 
//  	    (JMethodDeclaration) unit.typeDeclarations()[0].methods().get(0);
//  	JmlNode[] mNotes = m.jmlAnnotations();

//  	assertEquals( "should find one type annotations", 1, tNotes.length );
//  	assertEquals( "should find no method annotations", 0, mNotes.length );
//  	assertEquals( "org.multijava.util.compiler.CWarning: File \"unknown_source_file\", line 3, character 19 error: Behavioral specification for method is not followed by a method declaration; there is a field or type declaration instead [JML]" + NEWLINE, 
//  		      parserHelper.errorOut.toString(), true );
//      }

    //=====================================================================
    // Comments
    //=====================================================================

    public void test_comment1()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//+ comment" + NEWLINE +
				 "}" );
	assertEquals( "should find one class declaration",
		      unit.typeDeclarations().length, 1 );
    }

    public void test_comment2()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//+" + NEWLINE +
				 "}" );
	assertEquals( "should find one class declaration",
		      unit.typeDeclarations().length, 1 );
    }

    public void test_comment3()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*+ comment */" + NEWLINE +
				 "}" );
	assertEquals( "should find one class declaration",
		      unit.typeDeclarations().length, 1 );
    }

    /**
     * Should allow comments within annotation. */
    public void test_comment_4()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ invariant   // comment" + NEWLINE + 
				 "  @      true;" + NEWLINE +	
				 "  @*/" + NEWLINE +
				 "}" );
	JmlInvariant[] inv = invariantsFrom( unit );

	assertEquals( "should find one annotation", 1, inv.length );
    }

    /**
     * Should allow comments within annotation. */
    public void test_comment_5()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//@ invariant /* comment */ true;" + 
				 NEWLINE + 
				 "}" );
	JmlInvariant[] inv = invariantsFrom( unit );

	assertEquals( "should find one annotation", 1, inv.length );
    }

    /**
     * Should not allow multi-line comment within single-line
     * annotation. */
    public void test_comment_6()
	throws RecognitionException, TokenStreamException
    {
	try {
	    JmlCompilationUnit unit = 
		parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
					"//@ invariant /* comment " + NEWLINE +
					" comment */ true;" + NEWLINE + 
					"}" );
	    assertFalse( "should get an exception", true );
	} catch (TokenStreamException e) {
	    assertTrue( "expected message that began <expecting '*'...> but message was <" + e.getMessage() + ">",
			e.getMessage().startsWith( "expecting '*'" ) );
	} // end of try-catch
    }

    /**
     * Should allow empty annotations. */
    public void test_comment_7()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "//@" + 
				 NEWLINE + 
				 "}" );
	assertEquals( "should find one class declaration",
		      unit.typeDeclarations().length, 1 );
    }

    /**
     * Should allow empty annotations. */
    public void test_comment_8()  
	throws RecognitionException, TokenStreamException 
    {
	JmlCompilationUnit unit = 
	    parserHelper.getJmlAST( "public class Sample {" + NEWLINE +
				 "/*@ */" + 
				 NEWLINE + 
				 "}" );
	assertEquals( "should find one class declaration",
		      unit.typeDeclarations().length, 1 );
    }

    //---------------------------------------------------------------------
    // PRIVATE HELPERS
    //---------------------------------------------------------------------

    private long getTypeModifiers( JmlCompilationUnit unit ) {
	return unit.typeDeclarations()[0].modifiers();
    }

    private JmlClassBlock[] initializersFrom( JmlCompilationUnit unit ) {
	JPhylum[] fieldsAndInits = 
	    ((JmlTypeDeclaration) unit.typeDeclarations()[0]).fieldsAndInits();
	ArrayList inits = new ArrayList( fieldsAndInits.length );
	for (int i = 0; i < fieldsAndInits.length; i++) {
	    if (fieldsAndInits[i] instanceof JmlClassBlock) {
		inits.add( fieldsAndInits[i] );
	    } // end of if 
	} // end of for
	JmlClassBlock[] result = new JmlClassBlock[ inits.size() ];
	inits.toArray( result );
	return result;
    }

    private JmlMethodDeclaration[] methodDeclsFrom( JmlCompilationUnit unit )
    {
	ArrayList v = ((JmlTypeDeclaration) unit.typeDeclarations()[0]).methods();
	JmlMethodDeclaration[] result = new JmlMethodDeclaration[ v.size() ];
	v.toArray( result );
	return result;
    }

    private JFormalParameter[] 
	parameterDeclarationsFrom( JmlCompilationUnit unit ) 
    {
	ArrayList mV = 
	    ((JmlTypeDeclaration) unit.typeDeclarations()[0]).methods();
	JMethodDeclarationType m = (JMethodDeclarationType) mV.get(0);
	return m.parameters();
    }

    private JmlFieldDeclaration[] fieldsFrom( JmlCompilationUnit unit ) {
	JFieldDeclarationType[] fields = 
	    ((JmlTypeDeclaration) unit.typeDeclarations()[0]).fields();
	JmlFieldDeclaration[] result = 
	    new JmlFieldDeclaration[ fields.length ];
	System.arraycopy( fields, 0, result, 0, fields.length );
	return result;
    }

    private JmlAxiom[] axiomsFrom( JmlCompilationUnit unit ) {
	return ((JmlTypeDeclaration) unit.typeDeclarations()[0]).axioms();
    }

    private JmlInvariant[] invariantsFrom( JmlCompilationUnit unit ) {
	return ((JmlTypeDeclaration) unit.typeDeclarations()[0]).invariants();
    }

    private JmlVarAssertion[] varAssertionFrom( JmlCompilationUnit unit ) {
	return ((JmlTypeDeclaration) unit.typeDeclarations()[0]).varAssertions();
    }

    // -------------------------------------------------------------
    // spec-expression, all spec-expressions are checked below, but a
    // helper is helpful here
    // -------------------------------------------------------------

    private JmlSpecExpression[] 
	specExpressionListFrom( JmlCompilationUnit unit )
    {
	try {
	return 
	    ((JmlMonitorsForVarAssertion) varAssertionFrom(unit)[0]).
	    specExpressionList();
	} catch (ClassCastException e) {
	    e.printStackTrace();
	    System.out.println(varAssertionFrom(unit)[0].getClass());
	    System.exit(-1);
	    return null;
	}
    }

    private JExpression expressionFrom( JmlCompilationUnit unit ) {
	return specExpressionListFrom( unit )[0].expression();
    }

    protected static class Helper extends TestMjcParser {
	public Helper( String name ) {
	    super( name, new CModifier( Constants.ACCESS_FLAG_ARRAY,
					Constants.ACCESS_FLAG_NAMES ));
	}
	/**
	 * Establish lexers and parsers for the given source code,
	 * parsing javadoc comments according to the given boolean. */
	protected void establishTest( String sourceCode, 
				      boolean parseJavadocs ) {
	    errorOut.setLength( 0 );
	    Reader r = new StringReader( sourceCode );
	    parsingController = new ParsingController( r, null );
	    lexingController = new TokenStreamSelector();
            boolean allowUniverses = false; // WMD
	    jmlLexer = new JmlLexer( parsingController, lexingController, 
				     true, true, allowUniverses, compiler );
	    docLexer = new JavadocLexer( parsingController );
	    jmlMLLexer = new JmlMLLexer( parsingController, lexingController, 
					 true, true, allowUniverses, compiler );
	    jmlSLLexer = new JmlSLLexer( parsingController, lexingController, 
					 true, true, allowUniverses, compiler );
	    try {
		lexingController.addInputStream( jmlLexer, "jmlTop" );
		lexingController.addInputStream( jmlMLLexer, "jmlML" );
		lexingController.addInputStream( jmlSLLexer, "jmlSL" );
		lexingController.addInputStream( docLexer, "javadoc" );
		lexingController.select( "jmlTop" );
		parsingController.addInputStream( lexingController, "jml" );
		parsingController.addInputStream( docLexer, "javadoc" );
		parsingController.selectInitial( "jml" );

		final boolean ACCEPT_MULTIJAVA = true;
		final boolean ACCEPT_RELAXEDMULTIJAVA = false;
		parser = 
		    new JmlParser( compiler, 
				   parsingController.initialOutputStream(),
				   parsingController,
                                   false,
				   ACCEPT_MULTIJAVA, 
				   ACCEPT_RELAXEDMULTIJAVA,
				   allowUniverses );
	    } catch( ParsingController.ConfigurationException e ) {
		// want the exception to pass out to the test driver
		// without having to worry about passing it explicitly
		throw new RuntimeException( e.getMessage() );
	    }
	}

	protected JmlCompilationUnit getJmlAST( String sourceCode ) 
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    return (JmlCompilationUnit) parser.jCompilationUnit();
	}

	protected JmlMethodSpecification getMethodSpecification ( 
		   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    long modifiers = 0;
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlMethodSpecification(modifiers); 
	}

	protected JmlGenericSpecCase getGenericSpecCase ( 
		   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlGenericSpecCase(false); 
	}

	protected JmlGenericSpecCase getGenericSpecStatementCase ( 
		   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlGenericSpecCase(true); 
	}

	protected JmlAbruptSpecCase getAbruptSpecCase ( 
		   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlAbruptSpecCase(); 
	}

	protected JmlSpecBodyClause[] getSpecBody ( 
		   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlSpecBody(); 
	}

	protected JmlRequiresClause getRequiresClause ( 
		   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlRequiresClause(); 
	}

	protected JmlSpecBodyClause getSpecBodyClause ( 
		   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlSpecBodyClause(); 
	}

	protected JmlModelProgram getModelProgram( long modifiers, 
						   String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlModelProgram(modifiers); 
	}

	protected JmlSpecVarDecl[] getSpecVarDecls( String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jmlSpecVarDecls(); 
	}

	protected JStatement getStatement( String sourceCode )
	    throws RecognitionException, TokenStreamException 
	{
	    establishTest( sourceCode );
	    lexingController.push( "jmlML" );
	    return parser.jStatement(); 
	}

	private ParsingController parsingController;
	private TokenStreamSelector lexingController;
	private JmlLexer jmlLexer;
	private JavadocLexer docLexer;
	private JmlMLLexer jmlMLLexer;
	private JmlSLLexer jmlSLLexer;
	private JmlParser parser;
    }

    //---------------------------------------------------------------------
    // COMMON FIXTURE
    //---------------------------------------------------------------------

//      public void setUp() {     
//      }

//        public void tearDown() { 
//        }

    private Helper parserHelper;
}
