/*
 * @(#)$Id: ExecPrettyPrinter.java,v 1.8 2008/05/02 01:17:47 wahlst Exp $
 *
 * Copyright (C) 2006 Iowa State University
 *
 * This file is part of JML
 *
 * JML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * JML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.jmlspecs.jmlexec;

import org.multijava.mjc.*;
import org.jmlspecs.checker.*;
import org.jmlspecs.jmlrac.*;
import java.io.File;
import java.util.*;
import org.multijava.util.compiler.TokenReference;
import org.multijava.util.compiler.CWarning;
import org.multijava.util.MessageDescription;

/** 
 * do not reuse variable names when processing multiple specification files
 */
class VarGen {
  public static VarGenerator varGen =
                           VarGenerator.forMethod(VarGenerator.forClass());
}

/**
 * A visitor class for translating JML specifications to Java programs
 * that use the Java Constraint Kit (JCK) to execute the specification.
 * Each JML class specification is translated to a Java class that
 * implements that specification.  Code for any specifications contained
 * in interfaces that the class implements is generated into the Java class.
 * Each node of AST is visited and printed to an external file
 * or writer object.
 *
 * <p>
 *
 * @author Tim Wahls
 * @author Ben Krause
 * @version $Revision: 1.8 $
 *
 */
public class ExecPrettyPrinter extends RacPrettyPrinter {

    // ----------------------------------------------------------------------
    // CONSTRUCTORS
    // ----------------------------------------------------------------------

    /**
     * Constructs a pretty printer object for JML specifications
     * @param file the file into which the code is to be printed
     * @param modUtil the modifier utility to print Java and JML modifiers
     */
    public ExecPrettyPrinter(File file, JmlModifier modUtil) {
	super( file, modUtil );
    }
    
    // ----------------------------------------------------------------------
    // COMPILATION UNIT
    // ----------------------------------------------------------------------

    private void getAllImplementedInterfaces(CClassType [] interfaces) {
       for (int i = 0; i < interfaces.length; i++) {
          if (interfaces[i].getCClass() instanceof JmlSourceClass) {
             JmlMemberDeclaration pintf = ((JmlSourceClass) 
                 interfaces[i].getCClass()).getAST();
             if (pintf instanceof JmlInterfaceDeclaration) {
                JmlInterfaceDeclaration intf = (JmlInterfaceDeclaration) pintf;
                if (!jmlInterfaces.contains(intf)) {
                   jmlInterfaces.add(intf);
                   getAllImplementedInterfaces(intf.interfaces());
                }
             }
          }
       }
    }

    private boolean isPrivate(long mod) {
       // return (mod & org.multijava.util.classfile.Constants.ACC_PRIVATE) != 0;
       if ((mod & org.multijava.util.classfile.Constants.ACC_PRIVATE) != 0) {
          if ((mod & ACC_SPEC_PUBLIC) != 0 || (mod & ACC_SPEC_PROTECTED) != 0) {
            return false;
          } else {
            return true;
          }
       } else {
         return false;
       }
    }

    private void getCS() {
       print("    final " + EPATH + "ConstraintSystem ");
       printCS();
       print(" = ");
       printEnv();
       println(".getCS();");
    }

    private void generateInvariantMethods(JmlTypeDeclaration self) {
       newLine();
       inPre = false;
       inAssert = false;
       CClass superClass =  self.getCClass().getSuperClass();
       print("  public void " + PREF + "callInvariants(final ");
       println(ENPATH + "Environment " + PREF + "env, " +
               "final boolean " + PREF + "inOld) {");
       getCS();
       acceptAll(allInvariants);
       if (hasSuperClass) {
          println("    super." + PREF + "callInvariants" +
                  "(" + PREF + "env, " + PREF + "inOld);");
       }
       println("  }");
       newLine();
       print("  public static void " + PREF + "callStaticInvariants(final ");
       println(ENPATH + "Environment " + PREF + "env, " +
               "final boolean " + PREF + "inOld) {");
       getCS();
       Iterator it = allInvariants.iterator();
       while (it.hasNext()) {
//       for (JmlInvariant jinv : allInvariants) {
          JmlInvariant jinv = (JmlInvariant) it.next(); 
          if (jinv.isStatic()) {
             jinv.accept(this);
          }
       }
       if (hasSuperClass) {
          println("    " + superClassName + 
           "." + PREF + "callStaticInvariants(" + PREF + "env, " + 
           PREF + "inOld);");
       }
       println("  }");
       newLine();
       inInvariantFor = true;
       print("  public static void " + PREF + "callInvariantsFor(final ");
       print("java.lang.Object " + PREF + "this, final " + ENPATH);
       println("Environment " + PREF + "env, " +
               "final boolean " + PREF + "inOld) {");
       getCS();
       acceptAll(allInvariants);
       if (hasSuperClass) {
          print("  " + superClassName);
          println("." + PREF + "callInvariantsFor(" + PREF + 
                  "this, " + PREF + "env, " + PREF + "inOld);");
       }
       println("  }");
       newLine();
       newLine();
       inAssert = true;
       print("  public static void " + PREF + "callInvariantsForAssert(final ");
       print("java.lang.Object " + PREF + "this, final " + ENPATH);
       println("Environment " + PREF + "env, " +
              "final boolean " + PREF + "inOld, " +
             "java.lang.Object " + PREF + "result) {");
       getCS();
       String res1, res2;
       result = PREF + "result";
       it = allInvariants.iterator();
       while (it.hasNext()) {
//       for (JmlInvariant jinv : allInvariants) {
          JmlInvariant jinv = (JmlInvariant) it.next(); 
          res1 = varGen.freshPreVar();
          declareVar(res1, RPATH + "MyBoolean");
          String oldRes = result;
          result = res1;
          jinv.accept(this);
          res2 = varGen.freshPreVar();
          declareVar(res2, RPATH + "MyBoolean");
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "CANDConstraint(");
          print(res1 + ", " + res2 + ", " + oldRes + ")");
          println(".setInAssert(true, false));");
          result = res2;
       }
       if (hasSuperClass) {
          print("    " + superClassName);
          println("." + PREF + "callInvariantsForAssert" +
                  "(" + PREF + "this, " + PREF + "env, " + PREF + "inOld, " + 
                  result + ");");
       } else {
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + EPATH + "Equality(" + result);
          println(", " + RPATH + "MyBoolean.TRUE()));");
       }
       println("  }");
       newLine();

       inPre = true;
       inInvariantFor = false;
       inAssert = false;
       result = null;
    }

    private String stripPackage(String nm) {
      int pos = nm.lastIndexOf('.');
      if (pos == -1) {
        return nm;
      } else {
        return nm.substring(pos + 1, nm.length());
      }
    }

    private void generateInitiallyMethod() {
      if (hasStaticInitially()) {
        isStatic = true;
        inPre = false;
        println("  static {");
        if (trace) {
          print("System.out.println(\"executing static initially clauses ");
          println("for " + className + "\");");
        }
        println("    " + ENPATH + "Environment " + PREF + "env = new " +
                 ENPATH + "Environment();");
        getCS();
        makeAllAssignable(getFields(curClass.getCClass()), true);
        JmlVarAssertion [] varAssertions = curClass.varAssertions();
        inPre = false;
        if (varAssertions != null) {
          for (int i = 0; i < varAssertions.length; i++) {
            if (varAssertions[i] instanceof JmlInitiallyVarAssertion) {
              varAssertions[i].accept(this);
            }
          }
        }
        Iterator it = jmlInterfaces.iterator();
        while (it.hasNext()) {
//        for (JmlInterfaceDeclaration jid : jmlInterfaces) {
          JmlInterfaceDeclaration jid = (JmlInterfaceDeclaration) it.next();
          varAssertions = jid.varAssertions();
          if (varAssertions != null) {
            for (int i = 0; i < varAssertions.length; i++) {
              if (varAssertions[i] instanceof JmlInitiallyVarAssertion) {
                varAssertions[i].accept(this);
              }
            }
          }
        }
  
        println("      " + PREF + "callStaticInvariants(" +
                PREF + "env, false);");
        println("  try {");
        println("  try {");
        print("    if (!" + CPATH + "JChoice.runGoal(");
        printCS();
        println(", " + trace + ")) {");
        print("      throw new RuntimeException(\"Error: ");
        println("bad static initially clause\");");
        println("    }");
        extractValues();
        println("    } finally {");
        println("      " + PREF + 
                "env.returnConstraintSystem();");
        println("    }");
        println("  }");
        inPre = true;
      }

      isStatic = false;
      inPre = true;
      println("  public void " + PREF + className + "$initially(" + ENPATH + 
        "Environment " + PREF + "env) {");
      getCS();
      if (hasSuperClass) {
        println("    " + PREF + stripPackage(superClassName) + 
                "$initially(" + PREF + "env);");
      }
      if (trace) {
        print("System.out.println(\"executing initially clauses for ");
        println(className + "\");");
      }
      inPre = false;
      JmlVarAssertion [] varAssertions = curClass.varAssertions();
      if (varAssertions != null) {
        for (int i = 0; i < varAssertions.length; i++) {
          if (varAssertions[i] instanceof JmlInitiallyVarAssertion) {
            varAssertions[i].accept(this);
          }
        }
      }
      Iterator it = jmlInterfaces.iterator();
      while (it.hasNext()) {
//        for (JmlInterfaceDeclaration jid : jmlInterfaces) {
        JmlInterfaceDeclaration jid = (JmlInterfaceDeclaration) it.next();
        varAssertions = jid.varAssertions();
        if (varAssertions != null) {
          for (int i = 0; i < varAssertions.length; i++) {
            if (varAssertions[i] instanceof JmlInitiallyVarAssertion) {
              varAssertions[i].accept(this);
            }
          }
        }
      }
      inPre = true;
      println("  }");
    }

    /** is self a class declaration with no default constructor? */
    private boolean needsConstructor(JmlTypeDeclaration self) {
      if (self instanceof JmlClassDeclaration) {
        Iterator it = self.methods().iterator();
        while (it.hasNext()) {
          Object next = it.next();
          if (next instanceof JmlConstructorDeclaration) {
            JmlConstructorDeclaration con = (JmlConstructorDeclaration) next;
            if (con.parameters().length == 0) {
              return false;
            }
          }
        }
        return true;
      } else {
        return false;
      }
    }

    private void makeAllAssignable(Set fields, boolean onlyStatic) {
      Iterator it = fields.iterator();
      while (it.hasNext()) {
        CField fld = (CField) it.next();
        // not sure where this$0 field is coming from - jmlrac must be adding
        // it to nested classes
        if (!fld.ident().startsWith("this$")) {
          if (!fld.isFinal() && !(onlyStatic && !fld.isStatic())) {
            String owner = fld.owner().ident();
            if (isJMLE(fld.owner().getType())) {
              if (fld.isStatic() && fld.owner().isInterface()) {
                print(owner + "." + PREF + owner + ".");
              } 
              println("  " + owner + "$" + PREF + "set" + fld.ident() + 
                      "(" + PREF + "env, true);");
            }
          }
        }
      }
    }

    /** auto-generate a default constructor */
    private void printDefaultConstructor(JmlTypeDeclaration self) {
      inPre = false;
      println("  public " + self.ident() + "() {");
      if (trace) {
      println("    System.out.println(\"executing auto-generated default " +
              "constructor for " + self.ident() + "\");");
      };
      println("  final " + ENPATH + "Environment " + PREF + "env = " +
              "new " + ENPATH + "Environment();");
      getCS();
      println("    try {");
      println("    " + PREF + self.ident() + "PPV(" + PREF + 
              "env, false, null);");
      print("    if (!" + CPATH + "JChoice.runGoal(");
      printCS();
      println(", " + trace + ")) {");
      println("      throw new " +
              "org.jmlspecs.jmlexec.runtime.PostconditionException(\"" +
              "Ensures clause (or invariant) not " +
              "satisfied (in automatically generated default constructor)." +
              "\");");
      println("  }");

 
      println("  try {");
      extractValues();
      println("    } finally {");
      println("      " + PREF + "env.returnConstraintSystem();");
      println("    }");
      println("  }");

      println("  public void " + PREF + self.ident() + "PPV(final " + ENPATH +
              "Environment " + PREF + "env, final boolean " + PREF +
              "inOld, java.util.ArrayList " + PREF + "allCases) {");
      getCS();
      visitFieldInits();

      // generate assignable clause
      if (hasInitially()) {
// START: won't work if initially clause uses static fields from an 
//        implemented interface
        makeAllAssignable(getFields(self.getCClass()), false);
        println("  " + PREF + className + "$initially(" + 
                PREF + "env);");
//        visitFieldInitsAsConstraints();
      }
      println("    if (this.getClass().equals(" + className + ".class)) {");
      println("      " + PREF + "callInvariants(" + PREF + "env, false);");
      println("      " + PREF + "callRepresents(" + PREF + "env);");
      println("    }");
      println("  }");
      newLine();
      inPre = true;
    }

    /** auto-generate a constructor for \fresh clauses*/
    private void printFreshConstructor(JmlTypeDeclaration self) {
      inPre = false;
      println("  public " + self.ident() + "(" + RPATH + 
              "FreshConstructor disambig, final " + 
              ENPATH + "Environment " + PREF + "env) {");
      getCS();
      if (trace) {
        println("    System.out.println(\"executing fresh " +
                "constructor for " + self.ident() + "\");");
      }
      visitFieldInits();

      // generate assignable clause
// START: won't work if initially clause uses static fields from an 
//        implemented interface

      makeAllAssignable(getFields(self.getCClass()), false);
      println("  " + PREF + className + "$initially(" + PREF + "env);");
      println("  " + PREF + "env.freshObjs().add(this);");
      println("      " + PREF + "callInvariants(" + PREF + "env, false);");
      println("      " + PREF + "callRepresents(" + PREF + "env);");
      println("  }");
      newLine();
      println("  public " + self.ident() + "(" + RPATH + 
              "InternalConstructor disambig, " + ENPATH + "Environment " +
              PREF + "env) {");
      if (hasSuperClass) {
         println("    super(disambig, " + PREF + "env);");
      }
      if (trace) {
        println("    System.out.println(\"executing internal constructor " +
                "call for " + className + ".\");");
      }
      println("  " + PREF + "env.freshObjs().add(this);");
      println("  }");
      inPre = true;
    }

  private boolean containsField(JmlFieldDeclaration self) {
    Iterator it = combinedFields.iterator();
    String ident = self.variable().ident();
    while (it.hasNext()) {
      if (ident.equals(((JmlFieldDeclaration) it.next()).variable().ident())) {
        return true;
      }
    }
    return false;
  }

  private String formatToken(JPhylum self) {
    return self.getTokenReference().toString().replace('"', '\'').replace('\\', '/');
  }

  private void addField(JmlFieldDeclaration self) {
    
    if (containsField(self)) {
         throw new 
           org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
  "Error: redeclaration of instance model field inherited from interface, " +
           formatToken(self) + ".");
    } else {
       combinedFields.add(self);
    }
  }

    private boolean isModelOrGhost(JmlFieldDeclaration self) {
      return self.isModel() || self.jmlAccess().isGhost();
    }

    private void generateRepresents() {
       JmlRepresentsDecl [] rdecls = curClass.representsDecls();
       println("  protected void " + PREF + "callRepresents("
             + ENPATH + "Environment " + PREF + "env) {");
       getCS();
       acceptAll(rdecls);
       if (hasSuperClass) {
         println("  super." + PREF + "callRepresents(" + PREF + "env);");
       }
       println("  }");
       println("");
       
       println("  protected static void " + PREF + "callStaticRepresents(" + 
              ENPATH + "Environment " + PREF + "env) {");
       getCS();
       for (int i = 0; i < rdecls.length; i++) {
         JmlRepresentsDecl rdec = rdecls[i];
         if (rdec.isStatic()) {
           rdec.accept(this);
         }
       }
       if (hasSuperClass) {
         println("  " + superClassName + "." + PREF + "callStaticRepresents("
             + PREF + "env);");
       }
       println("  }");
       println("");
    }

    /** Prints a JML type declaration. */
    protected void visitClassBody( JmlTypeDeclaration self ) 
    {
        boolean isModel = 
          self.getCClass().packageName().equals("org/jmlspecs/models/");
        if (!isModel) {
          if (self.interfaces() == null || self.interfaces().length == 0) {
            if (self instanceof JmlInterfaceDeclaration) {
              print(" extends ");
            } else {
              print(" implements ");
            }
          } else {
            print(", ");
          }
          print(RPATH);
	  print("JMLEType ");
        } 
        print("{");
	pos += TAB_SIZE;
        String oldName = className;
        className = self.ident();
        JmlTypeDeclaration oldCurClass = curClass;
        curClass = self;
        newLine();
        combinedFields = new ArrayList();

        if (self instanceof JmlInterfaceDeclaration) {

	   JPhylum[] fieldsAndInits = self.fieldsAndInits();
	   for (int i = 0; i < fieldsAndInits.length ; i++) {
              if (fieldsAndInits[i] instanceof JmlFieldDeclaration) {
                 JmlFieldDeclaration jField = 
                    (JmlFieldDeclaration) fieldsAndInits[i];
                 if (!isModelOrGhost(jField) &&
                     !jField.jmlAccess().isInstance()) {
                    print(modUtil.asString(jField.variable().modifiers()));
                    print(toString(jField.variable().getType()));
                    print(" ");
                    print(jField.variable().ident());
                    if (jField.variable().getValue() != null) {
                        print(" = " + jField.variable().getValue());
                    }
                    println(";");
                 }
                } 
	   }
           println("    public static class " + PREF + className + " {");
	   for (int i = 0; i < fieldsAndInits.length ; i++) {
              if (fieldsAndInits[i] instanceof JmlFieldDeclaration) {
                 JmlFieldDeclaration jField = 
                    (JmlFieldDeclaration) fieldsAndInits[i];
//                 if (isModelOrGhost(jField) &&
                   if (!jField.jmlAccess().isInstance()) {
// can't treat this as a normal field (i.e. no set method) because
// its in an interface
// need to make field nonfinal for visitation
                    visitJmlFieldDeclaration(jField);
                 } 
              } 
	   }
           println("    }");

           Iterator it = self.methods().iterator();
           while (it.hasNext()) {
              JmlMethodDeclaration meth = (JmlMethodDeclaration) it.next();
              super.visitJmlMethodDeclaration(meth);
              if (!isModel) {
                print("  public abstract void ");
                print(PREF);
                printMangled(meth.ident(), meth.parameters(), meth.returnType());
                print("(");
                printParamsAsObjects(meth.parameters());
                print(ENPATH + "Environment " + PREF + "env, boolean " +
                      PREF + "inOld");
                if (meth.returnType() != null &&
                  ! meth.returnType().isVoid()) {
                  print(", final java.lang.Object " + PREF + "result");
                }
                print(", final java.util.ArrayList " + PREF + "allCases)");
                printThrows(meth.getExceptions());
                println(";");
              }
           }
           newLine();
        } else { 

           jmlInterfaces = new ArrayList();
           getAllImplementedInterfaces(self.interfaces());

           Iterator it = jmlInterfaces.iterator();
           while (it.hasNext()) {
              JmlInterfaceDeclaration jid = 
                (JmlInterfaceDeclaration) it.next();
	      JPhylum[] intFields = jid.fieldsAndInits();
	      for (int i = 0; i < intFields.length ; i++) {
                 if (intFields[i] instanceof JmlFieldDeclaration) {
                    JmlFieldDeclaration jField = 
                       (JmlFieldDeclaration) intFields[i];
                    if (isModelOrGhost(jField) && 
                        jField.jmlAccess().isInstance()) {
//                       jField.accept(this);
                         addField(jField);
                    }
                 }
              }
           }

	   JPhylum[] fieldsAndInits = self.fieldsAndInits();
	   for (int i = 0; i < fieldsAndInits.length ; i++) {
//	      fieldsAndInits[i].accept(this);
              JPhylum fld = fieldsAndInits[i];
              if (fld instanceof JmlFieldDeclaration) {
                addField((JmlFieldDeclaration) fld);
              } else {
                fld.accept(this);
                if (1 == 1) throw new RuntimeException("ERROR!");
              }
	   }
           acceptAll(combinedFields);
           visitSuperDatagroups();
           println("  private static boolean " + PREF + "inOld = false;");

// 	acceptAll( self.axioms() );
           allInvariants = new ArrayList();
           JmlInvariant [] invs = self.invariants();
           for (int i = 0; i < invs.length; i++) {
              allInvariants.add(invs[i]);
           }
           it = jmlInterfaces.iterator();
           while (it.hasNext()) {
              JmlInterfaceDeclaration jid = (JmlInterfaceDeclaration)
                                              it.next();
              invs = jid.invariants();
              for (int i = 0; i < invs.length; i++) {
                 allInvariants.add(invs[i]);
              }
           }
           allConstraints = new ArrayList();
           JmlConstraint [] constraints = self.constraints();
           for (int i = 0; i < constraints.length; i++) {
              allConstraints.add(constraints[i]);
           }
           it = jmlInterfaces.iterator();
           while (it.hasNext()) {
              JmlInterfaceDeclaration jid = (JmlInterfaceDeclaration)
                                              it.next();
              constraints = jid.constraints();
              for (int i = 0; i < constraints.length; i++) {
                 allConstraints.add(constraints[i]);
              }
           }
           hasSuperClass = false;
           if (self instanceof JmlClassDeclaration) {
              String sName = ((JmlClassDeclaration) self).superName();
              if (sName != null) {
                superClassName = sName.replace('/', '.');
                hasSuperClass = !superClassName.startsWith("java.lang");
              }
           }

 	   acceptAll( allConstraints );
// 	acceptAll( self.dependsDecls() );

           // generate methods for represents decls
           if (needsConstructor(self)) {
             printDefaultConstructor(self);
           }
           printFreshConstructor(self);

           curClass = self;
           className = self.ident();
	   acceptAll( self.methods() );

           // add methods specified in implemented interfaces

           it = jmlInterfaces.iterator();
           while (it.hasNext()) {
              JmlInterfaceDeclaration jid = (JmlInterfaceDeclaration)
                                              it.next();
              curClass = jid;
              if (!isJMLType(jid.getCClass())) {
                Iterator methIt = jid.methods().iterator();
                while (methIt.hasNext()) {
                   JmlMethodDeclaration meth = 
                      (JmlMethodDeclaration) methIt.next();
                   exceptions = meth.getExceptions();
                   generateJMLMethod(meth);
                }
              }
           }
           curClass = self;
           className = self.ident();

           generateInvariantMethods(self);
           generateInitiallyMethod();
           generateRepresents();
           if (!self.jmlAccess().isAbstract()) {
             println("  public String " + PREF + "toString() {");
             print("   return \"");
             print(className + "@addr");
             println("\";");
             println("  }");
             println("");
           }

           print("  protected void " + PREF + "constraints(final ");
           println(ENPATH + "Environment " + PREF + "env) {");
           getCS();
           for (int i = 0; i < allConstraints.size(); i++) {
              JmlConstraint constraint = (JmlConstraint) allConstraints.get(i);
              if (constraint.isForEverything()) {  
                 print("  " + PREF + "constraint" + i + "(");
                 println(PREF + "env);");
              }
           }
           if (hasSuperClass) {
             println(className + ".super." + PREF + "constraints(" + 
                     PREF + "env);");
           }
           println("}");
           newLine();
           print("  protected static void " + PREF + 
                 "staticConstraints(");
           println(ENPATH + "Environment " + PREF + "env) {");
           getCS();
           for (int i = 0; i < allConstraints.size(); i++) {
              JmlConstraint constraint = (JmlConstraint) allConstraints.get(i);
              if (constraint.isForEverything() && 
                  constraint.isStatic()) {
                 println("  " + PREF + "constraint" + i + "(" + 
                         PREF + "env);");
              }
           }
           if (hasSuperClass) {
             print(superClassName + "." + PREF + "staticConstraints(");
             println(PREF + "env);");
           }
           println("}");
           newLine();

	   acceptAll( self.inners() );

        }
	println("}");
        className = oldName;
        curClass = oldCurClass;
    }

    private boolean hasSignalsClause(JmlSpecBodyClause [] clauses) {
      if (clauses != null) {
        for (int i = 0; i < clauses.length; i++) {
          if (clauses[i] instanceof JmlSignalsClause ||
              clauses[i] instanceof JmlSignalsOnlyClause) {
            return true;
          }
        }
        return false;
      } else {
        return false;
      }
    }

    private boolean hasSignalsClause(JmlGeneralSpecCase self) {
       JmlSpecBody body = self.specBody();
       if (body != null) {
         if (hasSignalsClause(body.specClauses())) {
           return true;
         }
         JmlGeneralSpecCase [] cases = body.specCases();
         if (cases != null) {
           for (int i = 0; i < cases.length; i++) {
             if (hasSignalsClause(cases[i])) {
               return true;
             }
           }
         }
       }
       return false;
    }

    private boolean hasSignalsClause(JmlMethodDeclaration self) {
       if (self.hasSpecification()) {
         JmlSpecCase[] specCases = self.methodSpecification().specCases();
         for (int i = 0; i < specCases.length; i++) {
           if (specCases[i] instanceof JmlGeneralSpecCase) {
             if (hasSignalsClause((JmlGeneralSpecCase) specCases[i])) {
               return true;
             }
           } else if (specCases[i] instanceof JmlHeavyweightSpec) {
             if (hasSignalsClause(
               ((JmlHeavyweightSpec) specCases[i]).specCase())) {
               return true;
             }
           } else if (specCases[i] instanceof JmlModelProgram) {
               return false;
           } else {
             fail("Error: invalid specification type");
           }
         }
         return false;
       } else {
         return false;
       }
    }

   private void printMangled(String name, Object [] params, CType ret) {
     // print mangled method name, using parameter types to resolve
     // overloading
     print(name.replace('/', '.'));
     CType [] types = new CType [params.length];
     for (int i = 0; i < params.length; i++) {
       if (params[i] instanceof JFormalParameter) {
         types[i] = ((JFormalParameter) params[i]).getType();
       } else if (params[i] instanceof JExpression) {
         types[i] = ((JExpression) params[i]).getType();
       } else if (params[i] instanceof CType) {
         types[i] = (CType) params[i];
       } else {
         fail("invalid parameter type: " + params[i].getClass());
       }
     }
     if (ret == null) {
       ret = CStdType.Void;
     } 
     String buff = CType.genMethodSignature(ret, types);
     buff = buff.replace('/', '$');
     buff = buff.replace(';', '$');
     buff = buff.replace('[', 'A');
     buff = buff.replace('(', 'P');
     buff = buff.replace(')', 'P');
     print(buff);
   }
      
   private boolean hasRequires(JmlSpecCase spec) {
     if (spec instanceof JmlHeavyweightSpec) {
       return hasRequires(((JmlHeavyweightSpec) spec).specCase().specHeader());
     }
     return true;
   }

    private String simpleName(/*@ non_null @*/ JmlName[] subnames) {
        String n = null;
        for (int i = 0; i < subnames.length; i++) {
            //@ assert (* subnames[i] is ident, this, super, or \other *);
            if (subnames[i].isThis())
                continue;
            if (subnames[i].isIdent()) {
                if (n != null)
                    return null; // can't handle names such as n1.n2
                else
                    n = subnames[i].ident();
            }
            else
                return null; // super or \other
        }
        return n;
    }

  private boolean staticCheck(JmlConstraint constraint,
                              JmlMethodDeclaration decl) {
    if (decl.isStatic()) {
      return constraint.isStatic();
    } else {
      return true;
    }
  }

  private boolean isApplicable(JmlConstraint constraint, 
                               JmlMethodDeclaration decl) {
     if (constraint.isForEverything()) {
       if (decl instanceof JmlConstructorDeclaration) {
         return constraint.isStatic();
       } else {
         return staticCheck(constraint, decl);
       }
     } else {
        JmlMethodName [] names = constraint.methodNames().methodNameList();
        for (int i = 0; i < names.length; i++) {
           JmlMethodName name = names[i];
           if (decl.ident().equals(simpleName(name.subnames()))) {
              // names match, check parameters
              if (!name.hasParamDisambigList()) {
                 return staticCheck(constraint, decl);
              } else {
                 CType [] params = name.paramDisambigList();
                 JFormalParameter [] formals = decl.parameters();
                 if (params.length == formals.length) {
                    boolean match = true;
                    for (int j = 0; j < params.length && match; j++) {
                       match = (params[j].equals(formals[j].getType()));
                    }
                    if (match && staticCheck(constraint, decl)) {
                       return true;
                    }
                 }
              }
           } 
        }
        return false;
     }
  }

   private boolean isOverriding(CMethodSet ms) {
      if (ms == null) {return false;}
      for (int i = 0; i < ms.size(); i++) {
         if (! TransUtils.excludedFromInheritance(ms.getMethod(i).owner())) {
            if (ms.getMethod(i).owner() instanceof JmlSourceClass) {
               JmlMemberDeclaration mem = ((JmlSourceClass) 
                                     ms.getMethod(i).owner()).getAST();
               if (mem instanceof JmlClassDeclaration) {
                  return true;
               } else if (mem instanceof JmlInterfaceDeclaration) {
                  if (!jmlInterfaces.contains((JmlInterfaceDeclaration) mem)) {
                     return true;
                  }
               }
            }
         }
      }
      return false;
   }

   private boolean hasInitially() {
     JmlVarAssertion [] varAssertions = curClass.varAssertions();
     if (varAssertions != null) {
       for (int i = 0; i < varAssertions.length; i++) {
         if (varAssertions[i] instanceof JmlInitiallyVarAssertion) {
           return true;
         }
       }
     }
     Iterator it = jmlInterfaces.iterator();
     while (it.hasNext()) {
       JmlInterfaceDeclaration jid = (JmlInterfaceDeclaration) it.next();
       varAssertions = jid.varAssertions();
       if (varAssertions != null) {
         for (int i = 0; i < varAssertions.length; i++) {
           if (varAssertions[i] instanceof JmlInitiallyVarAssertion) {
             return true;
           }
         }
       }
     }
     return false;
   }

   private boolean hasStaticInitially() {
     JmlVarAssertion [] varAssertions = curClass.varAssertions();
     if (varAssertions != null) {
       for (int i = 0; i < varAssertions.length; i++) {
         if (varAssertions[i] instanceof JmlInitiallyVarAssertion &&
             varAssertions[i].isStatic()) {
           return true;
         }
       }
     }
     Iterator it = jmlInterfaces.iterator();
     while (it.hasNext()) {
       JmlInterfaceDeclaration jid = (JmlInterfaceDeclaration) it.next();
       varAssertions = jid.varAssertions();
       if (varAssertions != null) {
         for (int i = 0; i < varAssertions.length; i++) {
           if (varAssertions[i] instanceof JmlInitiallyVarAssertion &&
             varAssertions[i].isStatic()) {
             return true;
           }
         }
       }
     }
     return false;
   }
 
   // pull all values from constraint system, assign to assignables
   private void extractValues() {
     String postVal = varGen.freshPreVar();
     String iter = varGen.freshPreVar();
     println("  " + RPATH + "Map.Entry " + postVal + ";");
     println("  java.util.Iterator " + iter + 
             " = " + PREF + "env.assignables().iterator();");
     println("  while (" + iter + ".hasNext()) {");
     println("    " + postVal + 
      " = (" + RPATH + "Map.Entry) " + iter + ".next();");
     println("  try {");
     String val = varGen.freshPreVar();
     println("    Object " + val + " = " + PREF + "env.getCS().getVarValue(" +
             postVal + ".getKey());");
     println("    if (" + val + " instanceof " + RPATH + "MyArray) {");
     print("      ((" + RPATH + "MyArray) " + val + ").setValues(");
     printCS();
     println(");");
     println("    }");
     print("    ((" + RPATH + "SetMethod)");
     println(" " + PREF + "env.setMethods().get(" + postVal + 
       ".getKey())).set(" + val + ");");
     println("  } catch (" + RPATH + "NoValueException " + PREF + "e) {");
     print("    ((" + RPATH + "SetMethod)");
     println(" " + PREF + "env.setMethods().get(" + postVal + 
             ".getKey())).set(" + postVal + ".getValue());");
     println("    }");
     println("  }");
     String keys = varGen.freshPreVar();
     println("  java.util.Set " + keys + " = " + PREF +
             "env.arrays().keySet();");
     println("  " + iter + " = " + keys + ".iterator();");
     String arr = varGen.freshPreVar();
     String myArr = varGen.freshPreVar();
     println("  while(" + iter + ".hasNext()) {");
     println("    java.lang.Object " + arr + " = " + iter + ".next();");
     println("    " + RPATH + "MyArray " + myArr + " = (" + RPATH + 
             "MyArray) " + PREF + "env.arrays().get(" + arr + ");");
     print("    " + myArr + ".convertBack(" + arr + ".getClass(), ");
     printCS();
     println(");");
     println("  }");
     if (!isStatic) {
       Iterator it = combinedFields.iterator();
       while (it.hasNext()) {
         JmlFieldDeclaration decl = (JmlFieldDeclaration) it.next();
         CType type = decl.getType();
         if (type.isArrayType()) {
           String ident = decl.ident();
           println("  if (" + SPECFLD + ident + " != null && " + SPECFLD +
                   ident + " != " + RPATH + "Null.NullObj) {");
           println("  try {");
           print("    ((" + RPATH + "MyArray) " + SPECFLD + ident + 
                      ").convertBack(Class.forName(\"" + getSignature(type) +
                      "\"), ");
           printCS();
           println(");");
           println("  } catch (ClassNotFoundException " + 
                   varGen.freshPreVar() + ") {");
           print("        throw new " + RPATH + "JmleException(\"Error:");
           println(" invalid type " + type + "\");");
           println("  }");
           println("  }");
         }
       }
     }
//     println("  } catch (RuntimeException " + PREF + "e) {");
     println("  } catch (" + RPATH + "JmleException " + PREF + "e) {");
     println("    " + RPATH + "Map.Entry " + PREF + "entry;");
     println("    java.util.Iterator " + PREF + "iter " +  
             " = " + PREF + "env.assignables().iterator();");
     println("    while (" + PREF + "iter.hasNext()) {");
     println("      " + PREF + "entry" + 
           " = (" + RPATH + "Map.Entry) " + PREF + "iter.next();");
     print("        ((" + RPATH + "SetMethod)");
     println(" " + PREF + "env.setMethods().get(" + PREF + "entry" + 
            ".getKey())).set(" + PREF + "entry.getValue());");
     println("    }");
     println("    throw " + PREF + "e;");
     println("  }");
   }

   private void printParamsAsObjects(JFormalParameter [] params) {
     // non-interface version - called only internally and converted
     // to world of JCK variables
     int pNum = 0;
     for (int i = 0; i < params.length; i++) {
       JFormalParameter param = params[i];
       if (!param.isGenerated()) {
         if (pNum > 0) {
           print(", ");
         }
         print("final java.lang.Object " + params[i].ident());
         pNum++;
       }
     }
     if (params.length > 0) print(", ");
   }

/*   
   private void allOverriddenMethods(CMethod self, ArrayList allMeths) {
     if (!allMeths.contains(self)) {
       allMeths.add(self);
     }
     CMethodSet ms = self.overriddenMethodSet();
     if (ms != null) {
       for (int i = 0; i < ms.size(); i++) {
         CMethod meth = ms.getMethod(i);
         allOverriddenMethods(meth, allMeths);
       }
     }
   }

   private ArrayList allOverriddenMethods(JmlMethodDeclaration self) {
     CMethodSet ms = self.overriddenMethods();
     ArrayList allMeths = new ArrayList();
     if (ms != null) {
       for (int i = 0; i < ms.size(); i++) {
         CMethod meth = (CMethod) ms.getMethod(i);
         allOverriddenMethods(meth, allMeths);
       }
     }
     return allMeths;
   }
*/

   private void visitFieldInits() {
      Iterator it = fieldInits.iterator();
      while (it.hasNext()) {
         JmlFieldDeclaration decl = (JmlFieldDeclaration) it.next();
         getInitValue(decl.variable().getValue(), decl.ident());
      }
   }

/*
   private void visitFieldInitsAsConstraints() {
      boolean backOld = inOld;
      inOld = false;
      for (JmlFieldDeclaration decl : fieldInits) {
        result = SPECFLD + decl.ident();
        decl.variable().getValue().accept(this);
      }
      result = null;
      inOld = backOld;
   }
*/

   private boolean isJMLType(CClass clz) {
     boolean result = clz.isInterface() && 
                   clz.qualifiedName().equals("org/jmlspecs/models/JMLType");
     return result;
   }

   private boolean implementsJMLType() {
     Iterator it = jmlInterfaces.iterator();
     while (it.hasNext()) {
       if (isJMLType(((JmlInterfaceDeclaration) it.next()).getCClass())) {
         return true;
       }
     }
     return false;
   }
     
   private void printDefaultSpecCaseHeader() {
      print("  " + PREF + "allCases.add(");
      println("new " + RPATH + "SpecCase(");
      println(RPATH + "MyBoolean.TRUE(),");
      println("  new " + RPATH + "GenMethod() {");
      print("     public void run(" + EPATH + "ConstraintSystem ");
      printCS();
      print(") ");
      if (exceptions != null) {
        printThrows(exceptions);
      }
      println(" {");
      println("}}, ");
      println("new " + RPATH + "GenMethod() {");
      print("      public void run(" + EPATH + "ConstraintSystem ");
      printCS();
      print(") ");
      if (exceptions != null) {
        printThrows(exceptions);
      }
      println("{");
   }
     
   private void printDefaultSpecCaseTrailer() {
       println("}}));");
   }

   private void generateJMLMethod(JmlMethodDeclaration self) {

       print("  public ");
       if (self.isStatic()) {
         print("static ");
       }
       CClass methOwner = self.getMethod().owner();
       if (methOwner.isInterface()) {
         print("void " + methOwner.toString().replace('/', '$') + "$");
       } else {
         print("void " + PREF);
       }
       CType returnType = self.returnType();
       if (returnType.isVoid()) {
         returnType = null;
       }
       printMangled(self.ident(), self.parameters(), returnType);
       print("(");
       printParamsAsObjects(self.parameters());
       print("final " + ENPATH + "Environment " + PREF + 
             "env, final boolean " + PREF + "inOld");
       if (returnType != null) {
         print(", final java.lang.Object " + PREF + "result");
       }
       print(", java.util.ArrayList " + PREF + "allCases) ");
       printThrows(self.getExceptions());
       println(" {");
       getCS();

       if (self instanceof JmlConstructorDeclaration) {
          visitFieldInits();
       }

       isStatic = self.isStatic();
       hasSignals = false;
       JmlAssignableFieldSet assign = self.getAssignableFieldSet();
       isPure = self.jmlAccess().isPure() || 
                  (curClass.jmlAccess().isPure() && 
                   !(self instanceof JmlConstructorDeclaration)) || 
                   assign == null ||
                   (assign.size() == 0 && !assign.isUniversalSet());
       curMeth = self;
        //self.acceptDelegee(this);
    

       CMethodSet ms2 = self.overriddenMethods();
       if (ms2 != null) {
         for (int i = 0; i < ms2.size(); i++) {
           CMethod meth = ms2.getMethod(i);
           CClass owner = meth.owner();
           if (! TransUtils.excludedFromInheritance(owner) && 
               !isJMLType(owner)) {
             if (owner.isInterface()) {
                print(owner.toString().replace('/','$'));
                print("$");
             } else {
                if (isStatic) {
                  print(owner);
                } else {
                  print("super");
                }
                print("." + PREF);
             }
             printMangled(self.ident(), self.parameters(), returnType);
             print("(");
             printFormalsAsActuals(self.parameters());
             if (self.parameters().length > 0) print(", ");
             print(PREF + "env, " + PREF + "inOld");
             if (returnType != null) {
               print(", " + PREF + "result");
             }
             println(", " + PREF + "allCases);");
           }
         }
       }

       if (!(curMeth instanceof JmlConstructorDeclaration)) {
         print("  " + PREF + "call");
         if (curMeth.isStatic()) {
           print("Static");
         }
         println("Invariants(" + PREF + "env, true);");
       }
       if (trace) {
          println("  System.out.println(\"executing precondition for " + 
                  self.ident() + " of " + curClass.ident() + "\");");
       }

       JmlMethodSpecification desugaredSpec = null;
       JmlSpecCase[] specCases = null;
       hasSignals = false;
       if (self.hasSpecification()) {
          hasSignals = hasSignalsClause(self);
          DesugarSpec desugar = new DesugarSpec();
          desugaredSpec = desugar.perform(self.methodSpecification(),
                                          self.getExceptions());
	  specCases = desugaredSpec.specCases();
          acceptAll(specCases);
       } 
       if (implementsJMLType()) {
         String ident = self.ident();
         JFormalParameter [] params = self.parameters();
         if (ident.equals("equals") && params != null && 
             params.length == 1 &&
             params[0].getType().equals(CStdType.Object)) {
           String param = params[0].ident();
           printDefaultSpecCaseHeader();
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + CPATH);
           println("DELAYConstraint(new Object [] {" + param + 
                   "}, new " + RPATH + "GenMethod() {");
           print("      public void run(" + EPATH + "ConstraintSystem ");
           printCS();
           println(") {");
           String ob2 = varGen.freshPreVar();
           getValue(param, ob2, CStdType.Object);
           println("  if (" + ob2 + " == null) {");
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + EPATH + "Equality(" + PREF);
           println("result, " + RPATH + "MyBoolean.getBoolean(false)));");
           println("  } else if (" + ob2 + " == " + className + ".this) {");
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + EPATH + "Equality(" + PREF);
           println("result, " + RPATH + "MyBoolean.getBoolean(true)));");
           println("  }");
           printDefaultSpecCaseTrailer();
           printDefaultSpecCaseTrailer();
         } else if (ident.equals("clone") && 
                    (params == null || params.length == 0)) {
           printDefaultSpecCaseHeader();
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + CPATH);
           println("DELAYConstraint(new Object [] {" + PREF + 
                   "result}, new " + RPATH + "GenMethod() {");
           print("      public void run(" + EPATH + "ConstraintSystem ");
           printCS();
           println(") {");
           String res = varGen.freshPreVar();
           getValue(PREF + "result", res, CStdType.Object);
           println("  if (" + res + " == null || !(" + res + " instanceof " +
                   MPATH + "JMLType)) {");
           print("    ");
           printCS();
           println(".addGoalConstraint(new " + EPATH + "Bool(\"false\"));");
           println("  } else {"); 
           String cases = varGen.freshPreVar();
           println("  java.util.ArrayList " + cases + 
                   " = new java.util.ArrayList();");
           println("  try {");
           String equalRes = RPATH + "MyBoolean.TRUE()";
           print(className + ".this.");
           print(PREF); 
           printMangled("equals", new Object [] {CStdType.Object}, 
                        CStdType.Boolean);
           print("(" + res + ", ");
           println(PREF + "env, " + PREF + "inOld || " + (inOld || inPre) + 
                   ", " + equalRes + ", " + cases + ");");
           printCS();
           println(".addGoalConstraint(new " + CPATH + 
                   "ALLCASESConstraint(" + cases + "));");
           catchException(equalRes, CStdType.Boolean);
           println("  }");
           printDefaultSpecCaseTrailer();
           printDefaultSpecCaseTrailer();
         }
       }
       printDefaultSpecCaseHeader();
       if (!methOwner.isInterface()) {
         callInvariants();
         print("  " + PREF + "call");
         if (isStatic) {
           print("Static");
         }
         println("Represents(" + PREF + "env);");
      }
      printDefaultSpecCaseTrailer();
 
      println("}");
      newLine();

      if (hasSignals) {
        printExceptionInterface();
      }
   }

   private void callInvariants() {
          // don't check invariants or history constraints (yet) if 
          // autocalled by subclass constructor
          boolean isConstructor = (curMeth instanceof 
                                     JmlConstructorDeclaration);
          if (isConstructor) {
            println("  " + PREF + className + "$initially(" + PREF +
                    "env);");
            println("  if (" + className + ".this.getClass().equals(" + 
                    className + ".class)) {");
          }
          if (curMeth.isStatic()) {
            println("    " + PREF + "callStaticInvariants(" + PREF + 
                    "env, false);");
          } else {
            println("    " + PREF + "callInvariants(" + PREF + "env, false);");
          }
          if (isConstructor) {
            println("    " + PREF + "staticConstraints(" + PREF + "env);");
          } else {
            if (isStatic) {
              print(className);
            } else {
              print(className + ".this");
            }
            print(".");
            printMangled(curMeth.ident(), curMeth.parameters(),
                         curMeth.returnType());
            println("$constraints(" + PREF + "env);");
          }

          if (isConstructor) {
            println("  }");
          }
   }

   private void printWrappedFormals() {
     int pNum = 0;
     for (int i = 0; i < formalParameters.length; i++) {
       JFormalParameter param = formalParameters[i];
       if (!param.isGenerated()) {
         if (pNum > 0) {
           print(", ");
         }
         CType type = param.getType();
         String pIdent = param.ident();
         if (type.isPrimitive()) {
           printWrapped(pIdent, type);
         } else {
           print(pIdent);
         }
         pNum++;
       }
     }
   }

   private void visitJmlSubprogram(JmlMethodDeclaration self) {
          generateJMLMethod(self);
          newLine();
          CType returnType = self.returnType();
          if (returnType.isVoid()) {
            returnType = null;
          }
          if (self instanceof JmlConstructorDeclaration) {
            visitConstructorDeclaration((JmlConstructorDeclaration) self);
          } else {
            visitMethodDeclaration(self);
          }
          println("  final " + ENPATH + "Environment " + PREF + "env = " +
                  "new " + ENPATH + "Environment();");
          getCS();
          println("  " + RPATH + "JMLTool.freshObjs = " + PREF + 
                  "env.freshObjs();");
          for (int i = 0; i < formalParameters.length; i++) {
            JFormalParameter param = formalParameters[i];
            CType type = param.getType();
            if (type.isArrayType()) {
              CArrayType arType = (CArrayType) type;
              println("  if ("+ RPATH + "ObjUtil.nonNull(" + param.ident() +
                  ")) {");
              print("    " + PREF + "env.arrays().put(" + param.ident() +
                    ", new " + RPATH + "MyArray(" + param.ident() + ", ");
              printCS();
              print("));");
              println("  }");
            }
          }
          println("  final java.util.ArrayList " + PREF + "allCases = " + 
                  "new java.util.ArrayList();");
          inPre = false;
          if (returnType != null) {
            declareVar(PREF + "result", returnType);
            printFD(PREF + "result", returnType, "null");
          }
          println("  try {");
          println("  try {");
          print("  ");
          print(PREF);
          printMangled(self.ident(), self.parameters(), returnType);
          print("(");
          printWrappedFormals();
          if (self.parameters().length > 0) print(", ");
          print(PREF + "env, false");
          if (returnType != null) {
            print(", " + PREF + "result");
          }
          println(", " + PREF + "allCases);");
          print("  ");
          printCS();
          println(".addGoalConstraint(new " + CPATH + "ALLCASESConstraint(" +
                  PREF + "allCases));");

          print("  if (!" + CPATH + "JChoice.runGoal(");
          printCS();
          println(", " + trace + ")) {");
          println("    throw new " +
             "org.jmlspecs.jmlexec.runtime.PostconditionException(\"" + 
              "Ensures clause (or invariant or history constraint) not " +
              "satisfied, " + formatToken(self) + ".\");");
          println("  }");

          extractValues();

          // return value
          if (returnType != null) {
            String resVal = varGen.freshPreVar();
            getValue(PREF + "result", resVal, returnType);
            println("  return " + resVal + ";");
          } 
          String ex = varGen.freshPreVar();
          println("  } catch (Throwable " + ex + ") {");
          println("  if (" + ex + " instanceof " + RPATH + "JmleException) {");
          println("    throw (" + RPATH + "JmleException) " + ex + ";");
          println("  } else {");
          println("  try {");
          extractValues();
          println("  if (" + ex + " instanceof " + RPATH + 
                  "ExceptionException) {");
          String ex2 = varGen.freshPreVar();
          println("  " + RPATH + "ExceptionException " + ex2 + " = (" +
                  RPATH + "ExceptionException) " + ex + ";");
          for (int i = 0; i < exceptions.length; i++) {
            println("  if (" + ex2 + ".getException() instanceof " + 
                    exceptions[i] + ") {");
            println("    throw (" + exceptions[i] + ") " + ex2 + 
                    ".getException();");
            println("  }");
          }
          for (int i = 0; i < exceptions.length; i++) {
            println("  else if (" + ex + " instanceof " + 
                    exceptions[i] + ") {");
            println("    throw (" + exceptions[i] + ") " + ex + ";");
            println("  }");
          }
          println("  } else if (" + ex + " instanceof RuntimeException) {");
          println("    throw (RuntimeException) " + ex + ";");
          println("  }");
          println("  throw new " + RPATH + "ExceptionException(" + ex + ");");
          println("  }");
          println("  } finally {");
          println("    " + PREF + "env.returnConstraintSystem();");
          println("  }");
          newLine();
          println("}");
          newLine();
    }

    private void printExceptionInterface() {
      println("interface ExceptionMethod" + exMethNum + "{");
      print("  public void run(" + EPATH + "ConstraintSystem ");
      printCS();
      print(") ");
      printThrows(exceptions);
      println(";");
      println("}");
      newLine();
      exMethNum++;
    }

    /** Prints a JML method declaration. */
    public void visitJmlMethodDeclaration( JmlMethodDeclaration self ) {

          isStatic = self.isStatic();
          JmlAssignableFieldSet assign = self.getAssignableFieldSet();
          isPure = self.jmlAccess().isPure() || 
                  (curClass.jmlAccess().isPure() && 
                   !(self instanceof JmlConstructorDeclaration)) || 
                   assign == null || 
                   (assign.size() == 0 && !assign.isUniversalSet());
          CType returnType = self.returnType();
          if (returnType.isVoid()) returnType = null;
          exceptions = self.getExceptions();
          formalParameters = self.parameters();
          newLine();
          visitJmlSubprogram(self);

          // generate constraint method 
          print("protected ");
          if (isStatic) {
            print("static ");
          }
          print("void ");
          printMangled(self.ident(), self.parameters(), returnType);
          print("$constraints(");
          println(ENPATH + "Environment " + PREF + "env) {");
          getCS();
          for (int i = 0; i < allConstraints.size(); i++) {
             JmlConstraint constraint = (JmlConstraint) allConstraints.get(i);
             if (isApplicable(constraint, self)) { 
                print("  " + PREF + "constraint" + i + "(");
                println(PREF + "env);");
             }
          }
          if (isOverriding(curMeth.overriddenMethods())) {
             if (isStatic) {
               print(superClassName);
             } else {
               print(className + ".super");
             }
             print(".");
             printMangled(curMeth.ident(), curMeth.parameters(), returnType);
             println("$constraints(" + PREF + "env);");
          } else if (hasSuperClass) {
             if (curMeth.isStatic()) {
               print(superClassName + "." + PREF + "staticConstraints(");
               println(PREF + "env);");
             } else {
               println(className + ".super." + PREF + "constraints(" + 
                       PREF + "env);");
             }
          }
          println("}");
          newLine();
          formalParameters = null;

    }

    /** Prints a JML constructor declaration. */
    public void visitJmlConstructorDeclaration( 
	JmlConstructorDeclaration self ) {
        exceptions = self.getExceptions();
        formalParameters = self.parameters();
	newLine();
        visitJmlSubprogram(self);
        formalParameters = null;
    }

    private void printFormals(JFormalParameter[] parameters) {
        int count = 0;
                                                                                
        for (int i = 0; i < parameters.length; i++) {
            if (count != 0) {
                print(", ");
            }
                                                                                
            if (!parameters[i].isGenerated()) {
                if (!parameters[i].isFinal()) {
                  print("final ");
                }
                parameters[i].accept(this);
                count++;
            }
        }
    }

    private void printFormalsAsActuals(JFormalParameter[] parameters) {
        int count = 0;
                                                                                
        for (int i = 0; i < parameters.length; i++) {
            if (count != 0) {
                print(", ");
            }
                                                                                
            if (!parameters[i].isGenerated()) {
                print(parameters[i].ident());
                count++;
            }
        }
    }

    private void printThrows(CClassType[] exceptions) {
        for (int i = 0; i < exceptions.length; i++) {
            if (i != 0) {
                print(", ");
            } else {
                print(" throws ");
            }
            print(exceptions[i].toString());
        }
    }

    public void visitMethodDeclaration(/*@non_null@*/ JmlMethodDeclaration self )
{
        long modifiers = self.modifiers();
//        if (isInterface) {
           modifiers = modifiers & 
              ~org.multijava.util.classfile.Constants.ACC_ABSTRACT &
              ~org.multijava.util.classfile.Constants.ACC_NATIVE; 
//        }
        CType returnType = self.returnType();
        String ident = self.ident();
                                                                                
        newLine();
        print(modUtil.asString(modifiers));
        print(returnType);
        print(" ");
        print(ident);
        print("(");
        printFormals(self.parameters());
        print(")");
        printThrows(self.getExceptions());
                                                                                
        print(" {");
        newLine();
    }

   public void visitConstructorDeclaration(/*@non_null@*/ JmlConstructorDeclaration self ) {
        long modifiers = self.modifiers();
        modifiers = modifiers & 
           ~org.multijava.util.classfile.Constants.ACC_ABSTRACT; 
        String ident = self.ident();
                                                                                
        newLine();
        print(modUtil.asString(modifiers));
        print(ident);
        print("(");
        printFormals(self.parameters());
        print(")");
                                                                                
        printThrows(self.getExceptions());
        print(" {");
        newLine();
    }

    /** Translates a JML method specification. */
    public void visitJmlMethodSpecification(JmlMethodSpecification self ) {
    }


    /** Prints a JML predicate. */
    public void visitJmlPredicate(JmlPredicate self) {
	self.specExpression().accept(this);
    }

    /** Prints a RAC predicate. */
/*
    public void visitRacPredicate(RacPredicate self) {
	self.specExpression().accept(this);
    }
*/

    private void printList(java.util.List lst, boolean leadComma) {
      Iterator it = lst.iterator();
      while (it.hasNext()) {
        String val = (String) it.next();
        if (val != null && !val.equals("")) {
          if (leadComma) print(", ");
          leadComma = true;
          print(val);
        }
      }
    }

    private class BoundQuantVariables {

       public class boundVar {
         public String id; // name of the bound variable
         public String resVar; // name of JMLObjectSet bounding the quantified
                               // variable
         public boolean isBound = false; // is the variable bound yet?
         public String lb = null, ub = null; // lower and upper bounds
         public CType type;
         public String boundingType;

         public boundVar(String i, String r, CType t) { 
           id = i; resVar = r; type = t;
         }

         public void setBound() { isBound = true; }
       }

       private java.util.Stack bVars;

       public BoundQuantVariables() { bVars = new java.util.Stack(); }

       public void push(String id, String res, CType t) {
         bVars.push(new boundVar(id, res, t));
       }

       public boundVar pop() {
         return (boundVar) bVars.pop();
       }
 
       public int size() { return bVars.size(); }

       public boundVar get(int i) {
         return (boundVar) bVars.get(i);
       }

       public boolean isQuantVar(String id) {
         for (int i = 0; i < size(); i++) {
            if (get(i).id.equals(id)) return true;
         }
         return false;
       }
    }

    /** return index of referenced variable (bound by quantifier) */
    private int boundQuantVar(JExpression [] args) {
      JExpression act; int res;
      for (int j = 0; j < args.length; j++) {
        act = args[j];
        res = boundQuantVar(act);
        if (res != -1) return res;
      }
      return -1;
    }

    /** return index of referenced variable (bound by quantifier) */
    private int boundQuantVar(JExpression arg) {
      String id;
      if (arg instanceof JUnaryPromote) 
        arg = ((JUnaryPromote) arg).expr();
      if (arg instanceof JLocalVariableExpression) 
        id = ((JLocalVariableExpression) arg).ident();
      else
        id = arg.toString();
      for (int i = quantVars.size() - 1; i > -1; i--) {
        if (id.equals(quantVars.get(i).id)) return i;
      }
      return -1;
    }

     /**
     * should we generate a constraint for this call of a JML model type
     * method?
     */
   private boolean useJMLCollectionConstraint(JMethodCallExpression self) {
       String ident = self.ident();
       boolean res = false;
       CType prefType;
       if (self.prefix() != null) {
         prefType = self.prefix().getType();
       } else {
         prefType = CStdType.Null;
       }
       if (isJMLCollectionType(prefType.toString()) && isConstraint(ident)) {
         if (inBoundaryPred && ident.equals("has") && 
             boundQuantVar(self.args()) != -1) {
           res = true;
         } else {
           String selfType = 
             prefType.getCClass().toString().substring(MODPATHLEN);
           if (ident.equals("remove") && self.args().length == 2 && 
               org.jmlspecs.jmlexec.runtime.JMLTool.isBag(selfType)) {
             res = false;
           } else if (self.getType().isBoolean() && result != null) {
             res = false;
           } else {
             res = true;
           }
         }
       }
       return res;
   }

     /**
     * visits a method call expression when the calling object is an 
     * instance of a JML model type
     */
    private void visitJMLModelTypeMethod(JMethodCallExpression self,
                                         String thisVar) {
       JExpression prefix = self.prefix();
       CClass typ = prefix.getType().getCClass();
       String selfType = ""; // set to the type of the calling object
       String ident = self.ident();
       String resVar;
       
       int qVar;
       selfType = typ.toString().substring(MODPATHLEN);
       if (inBoundaryPred && ident.equals("has") && 
           (qVar = boundQuantVar(self.args())) != -1) {
         if (secondTime) {
           equalTrue();
         } else {
           resVar = thisVar;
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + EPATH + "Equality(" + resVar);
           println(", " + quantVars.get(qVar).resVar + "));");
           quantVars.get(qVar).setBound();
           quantVars.get(qVar).boundingType = selfType;
         }
       } else {
         String constraint;
         if (ident.equals("int_length")) {
           constraint  = CPATH + "INT_SIZE";
         } else {
           constraint  = CPATH + ident.toUpperCase();
         }
         boolean twoArgs = actualArgs.size() == 2 && (ident.equals("has") || 
            (ident.equals("remove") && 
             !org.jmlspecs.jmlexec.runtime.JMLTool.isBag(selfType)));
         String pair = null;
         if (twoArgs) {
            pair = varGen.freshPreVar();
            declareVar(pair, MPATH + 
                    org.jmlspecs.jmlexec.runtime.JMLTool.pairType(selfType));
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + "TOPAIRConstraint(");
            printList(actualArgs, false);
            println(", \"" + selfType + "\", " + pair + "));");
         }
         print("  ");
         printCS();
         print(".addGoalConstraint((new " + constraint + "Constraint(" + 
               thisVar);
         if (twoArgs) {
           print(", " + pair);
         } else {
           printList(actualArgs, true);
         }
         print(", \"" + selfType + "\"");
         if (result != null) {
           print(", " + result);
         }
         if (self.getType().isBoolean() && result == null) {
           println(")).setInAssert(" + inAssert + "));");
         } else {
          // all but the last param (the "result") must be ground
           println(")).setInAssert(" + inAssert + ", false));");
         }
         result = null;
      } 
    }

    private boolean isJMLCollectionType(String typ) {
      boolean res =  typ.startsWith("org.jmlspecs.models.") && (
                     typ.endsWith("Sequence") || typ.endsWith("Set") ||
                     typ.endsWith("Bag") || typ.endsWith("Relation") ||
                     typ.endsWith("Map"));
      return res;
    }

    private String getClassName(String name) {
       int pos = name.lastIndexOf(".");
       if (pos == -1) {
         return name;
       } else {
         return name.substring(pos + 1, name.length());
       }
     }

     /*@
     * determine whether or not a JMLCollection method has an associated
     * constraint
     */
     private static boolean isConstraint(String ident) {
       // return !ident.equals("singleton");
       return nonConstraints.get(ident) == null;
     }

     private void catchException(String res, CType type) {
        String exName = varGen.freshPreVar();
        println("      } catch (Throwable " + exName + ") {");
        println("        if (" + exName + " instanceof " + RPATH +
                "JmleException) {");
        println("          throw (" + RPATH + "JmleException) " + 
                exName + ";");
        println("        } else {");
        print("          ");
        printCS();
        print(".addGoalConstraint(new " + EPATH + "Equality(" + res + ", ");
        if (type.isClassType()) {
          print(RPATH + "Null.NullObj");
        } else {
          printDefaultObject(type);
        }
        println("));");
/*
        println("          throw new " + RPATH + 
                "IllegalCallException(\"Error: a method called by this " +
                "specification threw exception \" + " + exName + ");");
*/
        println("        }");
        println("      }");
     }

    private boolean isWrapper(CType type) {
       String tStr = type.toString();
       return tStr.equals("java.lang.Byte") || tStr.equals("java.lang.Char") ||
              tStr.equals("java.lang.Short") || 
              tStr.equals("java.lang.Integer") ||
              tStr.equals("java.lang.Long") ||
              tStr.equals("java.lang.Float") ||
              tStr.equals("java.lang.Double") ||
              tStr.equals("java.lang.Boolean");
    }

    private boolean isValueMethod(String name) {
      return name.equals("byteValue") || name.equals("shortValue") ||
             name.equals("charValue") || name.equals("intValue") ||
             name.equals("longValue") || name.equals("floatValue") ||
             name.equals("doubleValue");
    }

     /*@
     * visits a method call expression
     */
    public void visitMethodCallExpression(/*@non_null@*/ JMethodCallExpression self ) {

       JExpression prefix = self.prefix();
       String ident = self.ident();
       if (prefix != null && prefix.getType().equals(CStdType.String) &&
           ident.equals("equals") && self.args().length == 1 && 
           result == null) {
           result = varGen.freshPreVar();
           declareVar(result, "java.lang.String");
           String thisVar = result;
           prefix.accept(this);
           result = varGen.freshPreVar();
           declareVar(result, "java.lang.String");
           String argVar = result;
           self.args()[0].accept(this);
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + CPATH + "STREQUALSConstraint(");
           println(thisVar + ", " + argVar + "));");
           result = null;
       } else if (prefix != null && isWrapper(prefix.getType()) &&
                  isValueMethod(ident)) {
           String oldRes = result;
           result = varGen.freshPreVar();
           declareVar(result, prefix.getType());
           String thisVar = result;
           prefix.accept(this);
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + CPATH + 
                 "VALUEMETHODConstraint(");
           println(thisVar + ", \"" + toString(prefix.getType()) + "\", " +
                   oldRes + "));");
           result = null;
       } else if (prefix != null && 
           prefix.getType().isAssignableTo(CStdType.Throwable) &&
           ident.equals("getMessage") && self.args().length == 0) {
// START - may be able to get rid of this when method calls can be used as
//         constraints
           String oldRes = result;
           result = varGen.freshPreVar();
           declareVar(result, prefix.getType());
           String thisVar = result;
           prefix.accept(this);
           print("  ");
           printCS();
           print(".addGoalConstraint(new " + CPATH + 
                 "EXCEPTIONMESSAGEConstraint(");
           println(thisVar + ", " + toString(prefix.getType()) + ".class, " +
                   oldRes + "));");
           result = null;
       } else if (prefix != null && 
                prefix instanceof org.multijava.mjc.JSuperExpression &&
                inInvariantFor) {
           fail("Error: use of super in an \\invariant_for expression is " +
                "not executable.");
       } else {
           CType prefType = null;
           if (prefix != null) {
              prefType = prefix.getType();
           } else {
              prefType = CStdType.Null;
           }
           boolean hasResult = true;
           String methRes = null;
           if (result == null) {
             hasResult = false;
             methRes = RPATH + "MyBoolean.TRUE()";
           } else {
             methRes = result;
           }
           String thisVar = null;
           String oldRes = result;
           if (prefix != null && !(prefix instanceof JSuperExpression) &&
               !(prefix instanceof JThisExpression) &&
               !(prefix instanceof JTypeNameExpression)) {
             result = varGen.freshPreVar();
             thisVar = result;
             declareVar(thisVar, prefType);
             prefix.accept(this);
             result = oldRes;
           }
           java.util.List prevArgs = actualArgs;
           visitArgs(self.args());
           if (useJMLCollectionConstraint(self)) {
             visitJMLModelTypeMethod(self, thisVar);
           } else {
             JmlTypeDeclaration cls;
             if (prefix == null) {
                cls = curClass;
             } else {
                cls = JmlTypeLoader.getJmlSingleton().typeDeclarationOf(
                 prefType.getCClass());
             }
             if (cls != null && isJMLE(cls.getCClass().getType())) {
               CType type = self.getType();
               if (type == null) {
                 type = cls.getCClass().getType();
               }
               if (type == null) {
                 fail("type is not set");
               }
               String thisNm = null;
               if (thisVar != null) {
                 print("  ");
                 printCS();
                 print(".addGoalConstraint(new " + CPATH);
                 print("DELAYConstraint(new Object [] {");
                 print(thisVar);
                 println("}, new " + RPATH + "GenMethod() {");
                 print("      public void run(" + EPATH + "ConstraintSystem ");
                 printCS();
                 println(") {");
                 thisNm = varGen.freshPreVar();
                 getValue(thisVar, thisNm, prefType);
               }            

               String cases = varGen.freshPreVar();
               println("  java.util.ArrayList " + cases + 
                       " = new java.util.ArrayList();");
               println("  try {");
               if (prefix == null) {
                 print(className + ".");
               } else if (prefix instanceof JSuperExpression) {
                 print(className + ".super.");
               } else if (prefix instanceof JThisExpression) {
                 print(className + ".this.");
               } else if (thisVar != null) {
                 print(thisNm + ".");
               } else if (prefix instanceof JTypeNameExpression) {
                 print(prefType.toString() + ".");
               }
               print(PREF); 
               printMangled(self.ident(), self.args(), type);
               print("(");
               printList(actualArgs, false);
               if (self.args().length > 0) print(", ");
               println(PREF + "env, " + PREF + "inOld || " + 
                       (inOld || inPre) + ", " + 
                       methRes + ", " + cases + ");");
               printCS();
               println(".addGoalConstraint(new " + CPATH + 
                       "ALLCASESConstraint(" + cases + "));");
               catchException(methRes, self.getType());
               if (thisVar != null) {
                 println("    }}));");
               }

             } else {
               String thisNm = null;
               print("  ");
               printCS();
               print(".addGoalConstraint(new " + CPATH + 
                     "DELAYConstraint(new Object [] {");
               boolean leadComma = false;
               if (thisVar != null) {
                 print(thisVar);
                 leadComma = true;
               }
               printList(actualArgs, leadComma);
               println("}, new " + RPATH + "GenMethod() {");
               print("      public void run(" + EPATH + "ConstraintSystem ");
               printCS();
               println(") {");
               println("      try {");
               ArrayList newParams = new ArrayList();
               Iterator it = actualArgs.iterator();
               int count = 0;
               while (it.hasNext()) {
                 String arg = (String) it.next();
                 CType aType = self.args()[count].getType();
                 String nm2 = varGen.freshPreVar();
                 newParams.add(nm2);
                 getValue(arg, nm2, aType);
                 count++;
               }
               if (thisVar != null) {
                 thisNm = varGen.freshPreVar();
                 getValue(thisVar, thisNm, prefType);
               }
               print(" ");
               printCS();
               print(".addGoalConstraint(new " + EPATH + "Equality(" + methRes 
                     + ", ");
               CType resType = self.getType();
               if (resType.isPrimitive()) {
                 print("new ");
                 printType(resType);
                 print("(");
               } else if (resType.isArrayType()) {
                 print("new " + RPATH + "MyArray(");
               }
               if (prefix == null) {
                 print(className + ".");
               } else if (prefix instanceof JSuperExpression) {
                 print(className + ".super.");
               } else if (prefix instanceof JThisExpression) {
                 print(className + ".this.");
               } else if (thisVar != null) {
                 print( thisNm + ".");
               } else if (prefix instanceof JTypeNameExpression) {
                 print(prefType.toString() + ".");
               }
               print(ident + "(");
               printList(newParams, false);
               if (resType.isPrimitive()) {
                 print(")");
               } else if (resType.isArrayType()) {
                 print("), ");
                 printCS();
               }
               println(")));");
               catchException(methRes, resType);
               println("    }}));");
             }
          }
          // when all finished
          actualArgs = prevArgs;
          result = null;
       }
   }

    /**
     * prints a list of arguments
     */
    protected void visitArgs(JExpression[] args) {
        actualArgs = new ArrayList();
        String oldres = result;
        if (args != null) {
            for (int i = 0; i < args.length; i++) {
                if (args[i] != null) {
                    result = varGen.freshPreVar();
                    declareVar(result, args[i].getType());
                    actualArgs.add(result);
                    args[i].accept(this);
                } else {
                    // for an array with no dimension specified
                    actualArgs.add("");
                }
            }
        }
        result = oldres;
    }


    /** Prints a JML specification expression. */
    public void visitJmlSpecExpression(JmlSpecExpression self) {
// START Useful - figuring out with methods to override!
 /*
try {
 System.out.println("Expression is: " +self.expression());
 System.out.println("of type: " + self.expression().getClass());
} catch (Exception e) {}
 */
	self.expression().accept(this);
    }

    /** Prints a JML abrupt spec body. */
    public void visitJmlAbruptSpecBody( JmlAbruptSpecBody self ) {}

    /** Prints a JML abrupt spec case. */
    public void visitJmlAbruptSpecCase( JmlAbruptSpecCase self ) {}

    /** Prints a JML assignable clause. If runtime assertion check code
     * has been generated, it is also printed. */
    public void visitJmlAssertStatement( JmlAssertStatement self ) 
    {
    }

    private Set getFields(CClass clazz, boolean noPrivate) {
       if (clazz == null || clazz.toString().equals("java/lang/Object")) {
         return new HashSet();
       } 
       Set fields = new HashSet();
       Iterator it = clazz.fields().iterator();
       while (it.hasNext()) {
         CField fld = (CField) it.next();
         if (!isPrivate(fld.modifiers()) || !noPrivate) {
           fields.add(fld);
         }
       }
       CClassType [] interfaces = clazz.getInterfaces();
       for (int i = 0; i < interfaces.length; i++) {
          fields.addAll(getFields(interfaces[i].getCClass(), noPrivate));
       }
       fields.addAll(getFields(clazz.getSuperClass(), true));
       if (clazz instanceof JmlSourceClass) { 
           JmlMemberDeclaration ref = 
             ((JmlSourceClass) clazz).getAST().getRefinedMember();
           if (ref != null) {
             fields.addAll(getFields(ref.getCClass(), noPrivate));
           }
       }
       return fields;
    }

    private Set getFields(CClass clazz) {
      return getFields(clazz, false);
    }

    private CField getField(CClass clazz, String fldName) {
       if (clazz == null || clazz.toString().equals("java/lang/Object")) {
         return null;
       } 
       Iterator it = clazz.fields().iterator();
       while (it.hasNext()) {
         CField fld = (CField) it.next();
         if (fld.ident().equals(fldName)) {
           return fld;
         }
       } 
       CClassType [] interfaces = clazz.getInterfaces();
       for (int i = 0; i < interfaces.length; i++) {
          CField fld = getField(interfaces[i].getCClass(), fldName);
          if (fld != null) return fld;
       }
       CField fld = getField(clazz.getSuperClass(), fldName);
       if (fld != null) return fld;
       if (clazz instanceof JmlSourceClass) { 
           JmlMemberDeclaration ref = 
             ((JmlSourceClass) clazz).getAST().getRefinedMember();
           if (ref != null) {
             return getField(ref.getCClass(), fldName);
           }
       }
       return null;
    }

    private CType getType(CClass clazz, String fldName) {
      CField fld = getField(clazz, fldName);
      if (fld == null) {
        return null;
      } else {
        return fld.getType();
      }
    }

    private boolean isStaticFieldOfReceiver(CClass clazz, JmlName[] names) {
      int ind = 0;
      if (names[0].isThis() || names[0].isSuper()) ind++;
      String name = names[ind].ident();
      CField fld = getField(clazz, name);
      if (fld == null) {
        return false;
      } else {
        return fld.isStatic();
      }
    }

    private boolean isStaticField(CClass clazz, String name) {
      if (clazz == null) {
        return false;
      } else {
        CField fld = getField(clazz, name);
        if (fld == null) {
          return false;
        } else {
          return fld.isStatic();
        }
      }
    }

    private void getInitValue(JExpression expr, String res) {
       result = varGen.freshPreVar();
       boolean backPre = inPre;
       inPre = false;
       declareVar(result, expr.getType());
       String res1 = result;
       expr.accept(this);
       result = null;
       print("    if (!");
       printCS();
       print(".callGoal(" + trace + ")) {");
       print("      throw new RuntimeException(\"Error: ");
       println("bad field initialization\");");
       println("    }");
       print("  " + SPECFLD + res + " = ");
       printCS();
       CType type = expr.getType();
       println(".getVarValue(" + res1 + ");");
       if (type.isClassType()) {
         println("  if (" + SPECFLD + res + " == null) {");
         println("    " + SPECFLD + res + " = " + RPATH + "Null.NullObj;");
         println("  } else {");
         if (type.isArrayType()) {
           print("      ((" + RPATH + "MyArray) " + SPECFLD + res +
                 ").setValues(");
           printCS();
           println(");");
         }
         println("  }");
       } 
       convertValue(SPECFLD + res, res, type);
       inPre = backPre;
    }

    private void getValue(JExpression expr, String res, boolean run) {
       result = varGen.freshPreVar();
       declareVar(result, expr.getType());
       String res1 = result;
       expr.accept(this);
       result = null;
       if (run) {
         print("    if (!" + CPATH + "JChoice.runGoal(");
         printCS();
         println(", " + trace + ", false)) {");
         print("      throw new RuntimeException(\"Error: ");
         println("bad assignable clause\");");
          
         println("    }");
       }
       getValue(res1, res, expr.getType());
    }

//    private JmlAssignableFieldSet getDataGroup(JmlStoreRefExpression storeRef) {
/*
System.out.println("STORE REF");
System.out.println(storeRef.expression());
System.out.println(storeRef.expression().getClass());
*/
/*
      if (storeRef.expression() instanceof JClassFieldExpression) {
        JClassFieldExpression expr = 
          (JClassFieldExpression) storeRef.expression();
        CField fld = expr.getField().getField();
        if (fld instanceof JmlSourceField) {
          JmlSourceField sFld = (JmlSourceField) fld;
          return curClass.getDataGroupMap().getMembers(sFld);
        }
      }
      return null;
    }
*/

    /** visit a store reference
        visitType == 0 => visiting for assignable clause
        visitType == 1 => visiting for \not_modified expression
        visitType == 2 => visiting for represents clause
     */
 
    private void visitStoreRef(JmlStoreRefExpression storeRef,
                   int visitType, String modRes) {
      boolean backOld = inOld;
      inOld = true;
      String nextRes = null;
      if (modRes != null) {
        nextRes = varGen.freshPreVar();
        println("  java.lang.Object " + nextRes + " = null;");
      }
      JmlName [] names = storeRef.names();
      JmlTypeDeclaration cur = curClass;
      CClass clazz = curClass.getCClass();
      CType type = clazz.getType();
      if (names[0].isSuper()) {
         clazz = clazz.getSuperClass();
         type = clazz.getType();
      } else if (isFormalParameter(names[0])) {
        type = getFormalType(names[0]);
      }
      int i = 0;
      String prefix = "";
      if (names[0].isThis() || names[0].isSuper()) {
         prefix = className;
         if (isStaticFieldOfReceiver(clazz, names)) {
           i++;
         }
      }
      if (visitType == 0) {
        println("  try {");
      }
      visitStoreRefHelper(storeRef, i, type, prefix,
                          visitType, modRes, nextRes);
      if (visitType == 0) {
         println("  } catch (ClassCastException " + varGen.freshPreVar() +
                 ") {");
         println("  throw new " + RPATH + "UnsupportedOperationException("
               + "\"Error: for executablility, an assignable clause must not" +
               " refer to an object that did not exist in the pre-state, " +
               formatToken(storeRef) + ".\");");
         println("  }");
      }
      if (modRes != null) {
        printCS();
        print(".addGoalConstraint(new " + EPATH + "Equality(" + modRes);
        println(", " + RPATH + "MyBoolean.TRUE()));");
      }
      inOld = backOld;
      result = null;
    }
 
    private String chopThis(String prefix) {
      String fldPrefix = "";
      if (prefix.endsWith(")")) {
        fldPrefix = prefix.substring(0, prefix.length() - 5);
        int ind = fldPrefix.indexOf(")");
        if (ind > -1) {
          fldPrefix = fldPrefix.substring(ind + 2, 
                                          fldPrefix.length());
        }
      } else {
        fldPrefix = prefix.substring(0, prefix.length() - 4);
      }
      if (fldPrefix.length() > 0 && !fldPrefix.endsWith(".")) {
         fldPrefix += ".";
      }
      return fldPrefix;
    }

    private void printWrapper(CType type) {
      if (type == CStdType.Byte) {
        print("Byte");
      } else if (type == CStdType.Short) {
        print("Short");
      } else if (type == CStdType.Char) {
        print("Character");
      } else if (type == CStdType.Integer) {
        print("Integer");
      } else if (type == CStdType.Long) {
        print("Long");
      } else if (type == CStdType.Boolean) {
        print("Boolean");
      } else {
        fail("" + type + " has no wrapper class!");
      }
    }

    private void printFD(String newVar, CType type, String oldVal) {
      if (type == null) {
        fail("fatal error: unknown type!");
      } else if (type.isOrdinal()) {
        printCS();
        print(".addGoalConstraint(new " + CPATH + "FDVARConstraint(" + newVar +
              ", \"" + type.toString() + "\", new " + RPATH + "FDomain(");
        printWrapper(type);
        print(".MIN_VALUE, ");        
        printWrapper(type);
        println(".MAX_VALUE, " + oldVal + ")));");
      }
    }

    private void genAssignable(String ident, String set, CType type) {
      String fakeIdent = (set.length() == 0) ? SPECFLD + ident : ident;
      print("if (" + fakeIdent);
      print(" == null || !(" + fakeIdent);
      println(" instanceof " + RPATH + "Variable)) {");
      String oldVal = varGen.freshPreVar();
      String newVar = varGen.freshPreVar();
      declareVar(newVar, type);
      print("  ");
      println("java.lang.Object " + oldVal + " = " + fakeIdent + ";");
      if (set.length() > 0) {
        println(set + "new " + RPATH + "Variable(" +
                newVar + ", " + oldVal + "));");
      } else {
        println(fakeIdent + " = new " + RPATH + "Variable(" +
                newVar + ", ((" + PREF + "inConstructor && " + oldVal + 
                " == null) ? " + newVar + ": " +oldVal + "));");
      }
      printFD(newVar, type, oldVal);
      println("    " + PREF + "env.assignables().put(" + newVar + ", " + 
               oldVal + ");");
      print("  " + PREF + "env.setMethods().put(" + newVar);
      println(", new org.jmlspecs.jmlexec.runtime.SetMethod() {");
      println("  public void set(Object " + PREF + "finalValue) {");
      if (set.length() > 0) {
        print(set + PREF + "finalValue);");
      } else {
//        println("  if (" + RPATH + "ObjUtil.isNull(" + PREF + 
//                "finalValue) || " + PREF + "finalValue instanceof " + RPATH +
//                "VarObject) {");
        println("  if (" +  PREF + "finalValue instanceof " + RPATH +
                "VarObject) {");
        print("  " + fakeIdent + " = ");
        if (type.isPrimitive()) {
          printDefaultObject(type);
        } else {
          print(RPATH + "Null.NullObj");
        }
        println(";");
        println("  } else {");
        println("  " + fakeIdent + " = " + PREF + "finalValue;");
        println("  }");
        convertValue(fakeIdent, ident, type);
      }
      println("}});");
      println("  }");
    }

    private CType getClass(String maybeQFileName) {
      maybeQFileName = maybeQFileName.replace('.', '/');
      CClass ret = CTopLevel.loadClass(maybeQFileName);
      if (ret == CClass.CLS_UNDEFINED) ret = null;
      if (ret == null) {
         // type name defined in same package?
        String qualName = curClass.getCClass().packageName();
        if (qualName != null && qualName != "") {
           qualName = qualName.replace('.', '/') + maybeQFileName;
           ret = CTopLevel.loadClass(qualName);
        }
        if (ret == CClass.CLS_UNDEFINED) ret = null;
        if (ret == null)  {
           // type name defined in imported package?
           Iterator iter = imports.iterator();
           while (iter.hasNext() && ret == null) {
             qualName = (String) iter.next();
             if (!qualName.endsWith(maybeQFileName)) {
               qualName = qualName + "/" + maybeQFileName;
             }
             ret = CTopLevel.loadClass(qualName);
             if (ret == CClass.CLS_UNDEFINED) ret = null;
           }
        }
      }
      if (ret != null) {
        return new CErasedClassType(ret);
      } else {
        return null;
      }
    }


    private void visitStoreRefHelper(JmlStoreRefExpression storeRef, int index, 
      CType type, String prefix, int visitType, String modRes, String nextRes) {
      JmlName [] names = storeRef.names();
      if (index < names.length) {
        if (prefix.length() > 0 && !prefix.equals(className) && 
            type != null && !prefix.equals(type.toString()) && 
            !(index > 0 && names[index - 1].isSuper())) {
          if (type.isArrayType()) {
            prefix = "((" + RPATH + "MyArray) " + prefix + ")";
          } else {
            prefix = "((" + toString(type) + ") " + prefix + ")";
          }
        }
      }
      if (visitType == 0 && index == names.length - 2 && 
          names[index].isSuper() && (names[index + 1].isIdent() ||
                                     names[index + 1].isFields())) {
          if (prefix.length() > 0) prefix = prefix + ".";
          prefix = prefix + "this";
          visitStoreRefHelper(storeRef, index + 1, type, prefix, visitType,
                              modRes, nextRes);
      } else if (index == names.length) {
        if (visitType == 1) {
          // visiting store-ref for \not_modified
          String res = null;
          if (modRes != null) {
            res = varGen.freshPreVar();
            declareVar(res, RPATH + "MyBoolean");
          } 
          print("  if (" + prefix);
          println(" instanceof " + RPATH + "Variable) {");
          if (modRes == null) {
            printCS();
            print(".addGoalConstraint(new ");
            print(EPATH + "Equality(");
            print("((" + RPATH + "Variable) " + prefix + ").var, ");
            print("((" + RPATH + "Variable) " + prefix + ").oldVal");
            println("));");
            println(" }");
          } else {
            printCS();
            println(".addGoalConstraint(new " + CPATH + 
                    "EQUALEQUALRConstraint(" +
                    "((" + RPATH + "Variable) " + prefix + ").oldVal, " +
                    "((" + RPATH + "Variable) " + prefix + ").var, " +
                     res + "));");
            println("  } else {");
            printCS();
            print(".addGoalConstraint(new " + EPATH + "Equality(");
            println(res + ", " + RPATH + "MyBoolean.TRUE()));");
            println(" }");
            println("  " + nextRes + " = new " + RPATH + "VarObject();");
            printCS();
            println(".addVariable(" + nextRes + ", \"" + RPATH + 
                    "MyBoolean\");");
            printCS();
            print(".addGoalConstraint(new " + CPATH + "CANDConstraint(");
            println(res + ", " + nextRes + ", " + modRes + "));");
            println(modRes + " = " + nextRes + ";");
          }
        } else if (visitType == 2) {
          // visiting for represents clause
          print("(" + prefix);
          print(" instanceof " + RPATH + "Variable ? ((");
          print(RPATH + "Variable) " + prefix + ").var : ");
          println(prefix + ")");
        } else {
          fail("invalid way of visiting store reference");
        }
      } else {
        JmlName self = names[index];
	if (self.isIdent()) {
          if (index == names.length - 1 && visitType == 0) {
            // visiting store-ref for \assignable clause
            print("  ");
            if (prefix.length() > 0) {
             print( prefix + ".");
            }
            CFieldAccessor field = storeRef.getField();
            if (field == null) {
              throw new 
                org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                   "Error: attempt to use a final field in an assignable " +
                   "clause or invalid store reference, " +
                    formatToken(self) + ".");
            } else {
              String owner = field.owner().ident();
              String fieldRef = "";
              if (field.isStatic() && field.owner().isInterface()) {
                fieldRef = owner + "." + PREF + owner + ".";
              } 
              fieldRef += owner + "$" + PREF + "set" + self.ident() + 
                      "(" + PREF + "env, " + 
                      (curMeth instanceof JmlConstructorDeclaration) + ");";
              println(fieldRef);
            }
          } else {
            if (prefix.length() > 0) prefix += ".";
            String ident = self.ident();
            if (index == 0 && isFormalParameter(self)) {
              // type is already set correctly
              if (type.isArrayType()) {
                ident = PREF + "env.arrays().get(" + ident + ")";
              }
	      prefix += ident;
            } else if (type == null || (index == 0 && 
                               getType(type.getCClass(), ident) == null)) {
            // accessing static field of another class?
              type = getClass(prefix + ident);
              if (type != null) {
                prefix = type.toString();
              } else if (index == names.length - 1) {
                fail("Error: could not find type " + prefix);
              } else {
                prefix = prefix + SPECFLD + ident;
              }
            } else if (type.isClassType() && !type.isArrayType()) {
              type = getType(type.getCClass(), ident);
	      prefix += SPECFLD + ident;
              if (type == null) {
                fail("Error: field " + ident + " not found");
              }
            } else {
              fail("Error: accessing field " + ident + " of nonclass type");
            }
            visitStoreRefHelper(storeRef, index + 1, type, prefix,
                                visitType, modRes, nextRes);
          }
	} else if (self.isThis()) {
          if (index == names.length - 1 && visitType == 0) {
            fail("Error: invalid JmlName at end of store reference.");
          } else {
            if (prefix.length() > 0) prefix += ".";
	    prefix += "this";
            visitStoreRefHelper(storeRef, index + 1, type, prefix,
                                visitType, modRes, nextRes);
          }
	} else if (self.isSuper()) {
          if (index == names.length - 1 && visitType == 0) {
            fail("Error: invalid JmlName at end of store reference.");
          } else {
            if (prefix.length() > 0) prefix += ".";
	    prefix += "super";
            visitStoreRefHelper(storeRef, index + 1, type, prefix,
                                visitType, modRes, nextRes);
          }
// START isWildcard  ???
	} else if (self.isWildcard()) {
            throw new 
              org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                 "Error: for executablility, an assignable clause must not" +
                 " use a wildcard, " + formatToken(self) + ".");
	} else if (self.isAll()) {
          String ind = varGen.freshPreVar();
          String ind2 = varGen.freshPreVar();
          println("  for (int " + ind + " = 0; " + ind + " < " + prefix +
                  ".length(); " + ind + "++) {");
          println("  final int " + ind2 + " = " + ind + ";");
          if (type.isArrayType()) {
            type = ((CArrayType) type).getIndexedType();
          } else {
            fail("Error: indexing into non-array");
          }
          if (visitType == 0 && index == names.length - 1) {
            genAssignable(prefix + ".get(" + ind + ")", 
                          prefix + ".set(" + ind2 + ", ", type);
          } else {
            prefix += ".get(" + ind2 + ")";
            visitStoreRefHelper(storeRef, index + 1, type, prefix,
                                visitType, modRes, nextRes);
          }
          println("  }");
	    // print("[*]");
	} else if (self.isPos()) {
          String ind = varGen.freshPreVar();
          if (type.isArrayType()) {
            type = ((CArrayType) type).getIndexedType();
          } else {
            fail("Error: indexing into non-array");
          }
          if (visitType == 1) {
            String indE = varGen.freshPreVar();
            String oldRes = result;
            declareVar(indE, self.start().getType());
            result = indE;
            self.start().accept(this);
            result = oldRes;
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + 
                  "DELAYConstraint(new Object [] {");
            print(indE);
            println("}, new " + RPATH + "GenMethod() {");
            print("      public void run(" + EPATH + "ConstraintSystem ");
            printCS();
            println(") {");
            getValue(indE, ind, self.start().getType());
          } else {
            getValue(self.start(), ind, true);
          }
          if (visitType == 0 && index == names.length - 1) {
            genAssignable(prefix + ".get(" + ind + ")", 
                          prefix + ".set(" + ind + ", ", type);
          } else {
            prefix += ".get(" + ind + ")";
            visitStoreRefHelper(storeRef, index + 1, type, prefix,
                                visitType, modRes, nextRes);
          }
          if (visitType == 1) {
             println("    }}));");
          }
	} else if (self.isRange()) {
          String lo = varGen.freshPreVar();
          String hi = varGen.freshPreVar();
          String oldRes = result;
          if (visitType == 1) {
            String loE = varGen.freshPreVar();
            String hiE = varGen.freshPreVar();
            declareVar(loE, self.start().getType());
            declareVar(hiE, self.end().getType());
            result = loE;
            self.start().accept(this);
            result = hiE;
            self.end().accept(this);
            result = oldRes;
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + 
                  "DELAYConstraint(new Object [] {");
            print(loE + ", " + hiE);
            println("}, new " + RPATH + "GenMethod() {");
            print("      public void run(" + EPATH + "ConstraintSystem ");
            printCS();
            println(") {");
            getValue(loE, lo, self.start().getType());
            getValue(hiE, hi, self.end().getType());
          } else {
            getValue(self.start(), lo, true);
            getValue(self.end(), hi, true);
          }
          String ind = varGen.freshPreVar();
          String ind2 = varGen.freshPreVar();
          println("  for (int " + ind + " = " + lo + "; " + ind + " <= " + 
                  hi + "; " + ind + "++) {");
          println("  final int " + ind2 + " = " + ind + ";");
          if (type.isArrayType()) {
            type = ((CArrayType) type).getIndexedType();
          } else {
            fail("Error: indexing into non-array");
          }
          if (visitType == 0 && index == names.length - 1) {
            genAssignable(prefix + ".get(" + ind2 + ")", 
                          prefix + ".set(" + ind2 + ", ", type);
          } else {
            prefix += ".get(" + ind2 + ")";
            visitStoreRefHelper(storeRef, index + 1, type, prefix,
                                visitType, modRes, nextRes);
          }
          println("  }");
          if (visitType == 1) {
             println("    }}));");
          }
        } else if (self.isFields()) {
           Set fields = getFields(type.getCClass());
           Iterator it = fields.iterator();
           while (it.hasNext()) {
             CField fld = (CField) it.next();
             String owner = fld.owner().ident();
             CType ownerCls = fld.owner().getType();
             String fldPrefix = "";
             if (fld.isStatic()) {
               fldPrefix = fld.owner().qualifiedName().replace('/', '.') + ".";
               if (visitType == 0 && index == names.length - 1) {
                   if (isJMLE(ownerCls)) {
                     print("  " + fldPrefix + fld.owner().ident());
                     println("$" + PREF + "set" + fld.ident() + 
                             "(" + PREF + "env, " +
                      (curMeth instanceof JmlConstructorDeclaration) + ");");
                 }
               } else {
                 visitStoreRefHelper(storeRef, index + 1, fld.getType(), 
                                     fldPrefix + SPECFLD + fld.ident(),
                                     visitType, modRes, nextRes);
               }
             } else {
               if (prefix.length() > 0) {
                 if (fld.isStatic() && index > 0 && names[index - 1].isThis()) {
                   fldPrefix = chopThis(prefix);
                 } else {
                   fldPrefix = prefix + ".";
                 }
               }
               if (visitType == 0 && index == names.length - 1) {
                 if (isJMLE(ownerCls)) {
                   print("  " + fldPrefix);
                   print(fld.owner().ident());
                   println("$" + PREF + "set" + fld.ident() + 
                           "(" + PREF + "env, " +
                      (curMeth instanceof JmlConstructorDeclaration) + ");");
                 }
               } else {
                 fldPrefix += SPECFLD + fld.ident();
                 CType fldType = fld.getType();
                 visitStoreRefHelper(storeRef, index + 1, fldType, fldPrefix, 
                                     visitType, modRes, nextRes);
               }
             }
           }
	} else {
           fail("Error: invalid JmlName in store reference.");
        }
      }

    }

    /** Prints a JML assignable clause. */
    public void visitJmlAssignableClause( JmlAssignableClause self ) 
    {
	Debug.msg("In JmlAssignableClause");
// START - need to modify this for reach etc.
	//newLineOffset();
        String oldVal, newVar;
	JmlStoreRef[] storeRefs = self.storeRefs();
	for (int i = 0; i < storeRefs.length; i++) {
          if (storeRefs[i].isEverything()) {
            throw new 
              org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                 "Error: for executablility, an assignable clause must give" +
                 " a specific list of locations, " + formatToken(self) + ".");
          } else if(storeRefs[i].isNotSpecified()) {
            if (!inOld) {
              warn(self.getTokenReference(), 
                   ExecMessages.ASSIGNABLE_NOT_SPECIFIED);
            }
             
          } else if (storeRefs[i].isThisExpression() ||
                     storeRefs[i].isSuperExpression()) {
            throw new 
              org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                 "Error: \"this\" and \"super\" are not assignable, " +
                  formatToken(self) + ".");
          } else if (!storeRefs[i].isNothing()) {
            if (storeRefs[i].isStoreRefExpression()) {
              JmlStoreRefExpression storeRef = 
                (JmlStoreRefExpression) storeRefs[i];
              visitStoreRef(storeRef, 0, null);
            } else {
              throw new 
                org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                   "Error: invalid store reference for executablility, " +
                    formatToken(self) + ".");
            }
        }
      }
    }

    /** Prints a JML assume statement. If runtime assertion check code
     * has been generated, it is also printed. */
/*
    public void visitJmlAssumeStatement( JmlAssumeStatement self ) 
    {
	mayStartAnnotation();
	print(self.isRedundantly() ?
	       "assume_redundantly " : "assume ");
	self.predicate().accept(this);
	if (self.hasThrowMessage()) {
	    print(" : ");
	    self.throwMessage().accept(this);
	}
	print(";");
	mayEndAnnotation();

	// print RAC code if exists
	if (self.hasAssertionCode()) {
	    newLine();
	    self.assertionCode().accept(this);
	}
    }
*/

    /** Prints a JML axim. */
/*
    public void visitJmlAxiom( JmlAxiom self ) 
    {
	mayStartAnnotation();
	print("axiom ");
	self.predicate().accept(this);
	print(";");
	mayEndAnnotation();
    }
*/

    /** Prints a JML behavior spec. */
    public void visitJmlBehaviorSpec( JmlBehaviorSpec self ) 
    {
	Debug.msg("In JmlBehaviorSpec");

/*
        if (!inPre) {
          hasSignals = false;
          println("  " + PREF + "exceptions.clear();");
        }
*/
	offset += TAB_SIZE;
	self.specCase().accept(this);
	offset -= TAB_SIZE;	
    }

    /** Prints static or instance initializer */
    public void visitJmlClassBlock( JmlClassBlock self ) 
    {
	visitClassBlock(self);
    }

    /** Prints a JML classOrGFImport clause. */
    public void visitJmlClassOrGFImport( JmlClassOrGFImport self ) {
        if (self.isModel()) {
            print("/*@ model */");
        }
	String name = self.getName().replace('/', '.');
	print("import " + name + ";");
        imports.add(self.getName());
    }

    /** Prints a JML constraint clause. */
    public void visitJmlConstraint( JmlConstraint self )
    {
        inPre = false;
        print("  private ");
        if (self.isStatic()) {
          print("static ");
        }
        print("void " + PREF + "constraint");
        print(consNum);
        consNum++;
        println("(" + ENPATH + "Environment " + PREF + "env) {");
        getCS();
        self.predicate().accept(this);
        println("}");
        inPre = true;
    }

    /** Prints a JML code contract. */
/*
    public void visitJmlCodeContract( JmlCodeContract self )
    {
	//newLineOffset();
	print("code_contract");
	offset += TAB_SIZE;
	JmlSpecBodyClause[] subclassingContract = self.accessibleClauses();
	for (int i = 0; i < subclassingContract.length; i++) {
	    subclassingContract[i].accept(this);
	}
	subclassingContract = self.callableClauses();
	for (int i = 0; i < subclassingContract.length; i++) {
	    subclassingContract[i].accept(this);
	}
	subclassingContract = self.measuredClauses();
	for (int i = 0; i < subclassingContract.length; i++) {
	    subclassingContract[i].accept(this);
	}
	offset -= TAB_SIZE;    
    }
*/

    /** Prints a JML constructor name. */
    public void visitJmlConstructorName( JmlConstructorName self ) 
    {
	print("new " + toString(self.ownerType()));
	print("(");
	if (self.hasParamDisambigList()) {
	    CType[] params = self.paramDisambigList();
	    for (int i = 0; i < params.length; i++) {
		if (i != 0)
		    print(", ");
		print(params[i]);
	    }
	}
	print(")");
    }

    /** Prints a JML diverges clause. */
/*
    public void visitJmlDivergesClause( JmlDivergesClause self ) {
	Debug.msg("In JmlDivergesClause");
	visitSpecBodyClause(self, "diverges");
    }
*/

    /** Prints a JML <code>elemtype</code> expression. */
    public void visitJmlElemTypeExpression( JmlElemTypeExpression self ) 
    {
/*
	print("\\elemtype(");
	self.specExpression().accept(this);
	print(")");
*/
        if (result != null) {
          String oldRes = result;
          result = varGen.freshPreVar();
          String newRes = result;
          declareVar(result, "java.lang.Object");
	  self.specExpression().accept(this);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "DELAYConstraint("
              + "new Object [] {" + newRes + "},");
          println("  new " + RPATH + "GenMethod() {");
          print("    public void run(" + EPATH + "ConstraintSystem ");
          printCS();
          println(") {");
          String typ = varGen.freshPreVar();
          getValue(newRes, typ, self.specExpression().getType());

          String typType = varGen.freshPreVar();
          println("  java.lang.Object " + typType + ";");
          println("  if (" + typ + ".isArray()) {");
          println("    " + typType + " = " + typ + ".getComponentType();");
          println("  } else {");
          println("    " + typType + " = " + RPATH + "Null.NullObj;");
          println("  }");

          print("  ");
          printCS();
          println(".addGoalConstraint(new " + EPATH + "Equality(" + oldRes +
                  ", " + typType + "));");
          println("}}));");
        }
    }

    /** Prints a JML ensures clause. */
/*
    public void visitJmlEnsuresClause( JmlEnsuresClause self ) 
    {
	Debug.msg("In JmlEnsuresClause");
	visitSpecBodyClause(self, "ensures");
    }
*/

    /** Prints a JML example. */
/*
    public void visitJmlExample( JmlExample self ) 
    {
	Debug.msg("In JmlExample"); 
	//newLineOffset();
	print(modUtil.asString(self.privacy()));
	print("example");
	offset += TAB_SIZE;
	visitExample(self);
	offset -= TAB_SIZE;
    }
*/

    /** Prints a JML exceptional behavior specification. */
/*
    public void visitJmlExceptionalBehaviorSpec( 
       JmlExceptionalBehaviorSpec self ) 
    {
	Debug.msg("In JmlExceptionalBehaviorSpec");

	//newLineOffset();
	print(modUtil.asString(self.privacy()));
	print("exceptional_behavior");
	offset += TAB_SIZE;
	self.specCase().accept(this);
	offset -= TAB_SIZE;	
    }
*/

    /** Prints a JML exceptional example. */
/*
    public void visitJmlExceptionalExample( JmlExceptionalExample self )
    {
	Debug.msg("In JmlExceptionalExample");

	//newLineOffset();
	print(modUtil.asString(self.privacy()));
	print("exceptional_example");
	offset += TAB_SIZE;
	visitExample( self );
	offset -= TAB_SIZE;
    }
*/

    /** Prints a JML exceptional spec body. */
/*
    public void visitJmlExceptionalSpecBody( JmlExceptionalSpecBody self )
    {
	Debug.msg("In JmlExceptionalSpecBody");
	visitSpecBody( self );
    }
*/

    /** Prints a JML exceptional spec case. */
/*
    public void visitJmlExceptionalSpecCase( JmlExceptionalSpecCase self )
    {
	Debug.msg("In JmlExceptionalSpecCase");
	visitGeneralSpecCase(self);
    }
*/

    /** Prints a JML method extending specification. */
    public void visitJmlExtendingSpecification(
       JmlExtendingSpecification self ) 
    {
	Debug.msg("In JmlExtendingSpecification");
	visitMethodSpecification(self);
    }

    /** Prints a JML method extending specification. */
    private void visitMethodSpecification(JmlSpecification self) 
    {

	JmlSpecCase[] specCases = self.specCases();
	if (specCases != null) {
	    for (int i = 0; i < specCases.length; i++) {
		specCases[i].accept(this);
	    }
	    if (specCases.length > 1)
		offset -= TAB_SIZE;
	}
	
	if (self.redundantSpec() != null)
	    self.redundantSpec().accept(this);
    }

    // get the represents clauses for this field that are not in the
    // current class
    private ArrayList getRepresents(CField fld) {
      ArrayList res = new ArrayList();
      JmlTypeDeclaration cls = curClass.findTypeWithRepresentsFor(fld);
      if (cls != null) {
        JmlRepresentsDecl [] rdecls = cls.representsDecls();
        if (rdecls != null) {
          for (int i = 0; i < rdecls.length; i++) {
            JmlRepresentsDecl rep = rdecls[i];
            CField rFld = rep.getField();
            if (fld == rFld || 
                (rFld == null && fld.ident().equals(rep.ident()))) {
              res.add(rep);
            }
          }
        }
      }
      return res;
    }

    /** requires combinedFields has been built
        ensures all MapInto clauses are returned */
    private ArrayList getAllMapsInto() {
      Iterator it = combinedFields.iterator();
      ArrayList mapsInto = new ArrayList();
      while (it.hasNext()) {
        JmlFieldDeclaration decl = (JmlFieldDeclaration) it.next();
        JmlMapsIntoClause [] maps = decl.getCombinedMapsIntoClauses();
        if (maps != null && maps.length > 0) {
          for (int i = 0; i < maps.length; i++) {
            mapsInto.add(maps[i]);
          }
        }
      }
      return mapsInto;
    }

    private class GroupPair {
      public JmlFieldDeclaration fld;
      public JmlInGroupClause group;

      public GroupPair(JmlFieldDeclaration fld, JmlInGroupClause group) {
        this.fld = fld;
        this.group = group;
      }
    }

    /** requires combinedFields has been built
        ensures all inGroup clauses are returned */
    private ArrayList getAllInGroups() {
      Iterator it = combinedFields.iterator();
      ArrayList inGroups = new ArrayList();
      while (it.hasNext()) {
        JmlFieldDeclaration decl = (JmlFieldDeclaration) it.next();
        JmlInGroupClause [] grps = decl.getCombinedInGroupClauses();
        for (int i = 0; grps != null && i < grps.length; i++) {
          inGroups.add(new GroupPair(decl, grps[i]));
        }
      }
      return inGroups;
    }

    private void visitSuperInGroup(JmlStoreRefExpression storeRefFld,
                                   ArrayList inGroup) {
      inPre = false;
      Iterator it = inGroup.iterator();
      String ident = storeRefFld.names()[1].ident();
      while (it.hasNext()) {
        GroupPair grp = (GroupPair) it.next();
        JmlStoreRefExpression[] ingrps = grp.group.groupList();
        for (int i = 0; ingrps != null && i < ingrps.length; i++) {
          JmlStoreRefExpression ref = ingrps[i];
          JmlName [] names = ref.names();
          if (names.length == 2 && names[0].isSuper() &&
            ident.equals(names[1].ident())) {
            genAssignable(grp.fld.ident(), "", grp.fld.getType());
          }
        }
      }
      inPre = true;
    }

    private void printSetHeader(CField cFld, String ident) {
      print("  public ");
      if (cFld.isStatic()) {
        print("static ");
      }
      println("void " + cFld.owner().ident() + 
              "$" + PREF + "set" + ident + "(final " + ENPATH + "Environment " +
               PREF + "env, boolean " + PREF + "inConstructor) {");
      getCS();
      if (cFld.isStatic()) {
        print("  " + cFld.owner().ident() + ".");
      } else {
        print("  super.");
      }
      println(cFld.owner().ident() +
              "$" + PREF + "set" + ident + "(" + PREF + "env, " + 
              PREF + "inConstructor);");
    }

    private void printSuperDatagroups(JmlStoreRefExpression storeRefFld,
                                    ArrayList mapsInto, ArrayList inGroup) {
      CField cFld = storeRefFld.getField().getField();
      String ident = storeRefFld.names()[1].ident();
      Iterator it = mapsInto.iterator();
      printSetHeader(cFld, ident);

      while (it.hasNext()) {
        JmlMapsIntoClause map = (JmlMapsIntoClause) it.next();
        JmlStoreRefExpression [] grps = map.groupList();
        if (grps != null && grps.length > 0) {
          for (int j = 0; j < grps.length; j++) {
            JmlStoreRefExpression storeRef = grps[j];
            JmlName [] names = storeRef.names();
            if (names.length == 2 && names[0].isSuper()) {
              if (ident.equals(names[1].ident())) {
               JmlStoreRef ref = map.memberRef();
               if (ref instanceof JmlStoreRefExpression) {
                 inPre = false;
                 visitStoreRef((JmlStoreRefExpression) ref, 0, null);
                 inPre = true;
               } else {
                 fail("Error: invalid store reference in maps into clause");
               }

              }
            }
          }
        }
      }
      visitSuperInGroup(storeRefFld, inGroup);
      println("  }");
      newLine();
    }

    /** requires combinedFields has been built
        ensures "set" methods for superclass fields are printed */
    private void visitSuperDatagroups() {
      ArrayList mapsInto = getAllMapsInto();
      ArrayList inGroup = getAllInGroups();
      Iterator it = mapsInto.iterator();
      Iterator itGrp = inGroup.iterator();
      ArrayList visited = new ArrayList();
      while (it.hasNext()) {
        JmlMapsIntoClause map = (JmlMapsIntoClause) it.next();
        JmlStoreRefExpression [] grps = map.groupList();
        if (grps != null && grps.length > 0) {
          for (int j = 0; j < grps.length; j++) {
            JmlStoreRefExpression storeRef = grps[j];
            JmlName [] names = storeRef.names();
            if (names.length == 2 && names[0].isSuper()) {
              String ident = names[1].ident();
              if (!visited.contains(ident)) {
                visited.add(ident);
                printSuperDatagroups(storeRef, mapsInto, inGroup);
              }
            }
          }
        }
      }
      while (itGrp.hasNext()) {
        GroupPair grp = (GroupPair) itGrp.next();
        JmlStoreRefExpression[] ingrps = grp.group.groupList();
        for (int i = 0; ingrps != null && i < ingrps.length; i++) {
          JmlStoreRefExpression ref = ingrps[i];
          JmlName [] names = ref.names();
          if (names.length == 2 && names[0].isSuper()) {
            String ident = names[1].ident();
            if (!visited.contains(ident)) {
              visited.add(ident);
              printSetHeader(ref.getField().getField(), ident);
              visitSuperInGroup(ref, inGroup);
              println("  }");
              newLine();
            }
          }
        }
      }
    }

    private ArrayList getMapsInto(JmlFieldDeclaration self) {
      ArrayList mapsClauses = new ArrayList();
      Iterator it = getAllMapsInto().iterator();
      while (it.hasNext()) {
        JmlMapsIntoClause map = (JmlMapsIntoClause) it.next();
        JmlStoreRefExpression [] grps = map.groupList();
        if (grps != null && grps.length > 0) {
          for (int j = 0; j < grps.length; j++) {
            JmlStoreRefExpression storeRef = grps[j];
            JmlName [] names = storeRef.names();
            String ident = null;
            if (names.length == 1 && names[0].isIdent()) {
              ident = names[0].ident();
            } else if (names.length == 2 && names[0].isThis()
                    && names[1].isIdent()) {
              ident = names[1].ident();
            } else if (names.length == 2 && names[0].isSuper()
                    && names[1].isIdent()) {
              // handled externally - of course, the superclass field does not
              //                      occur in the current class
              ident = "";
            } else {
              fail("Invalid store reference in a maps into group list");
            }
            if (ident.equals(self.variable().ident())) {
              mapsClauses.add(map);
            }
          }
        }
      }
      return mapsClauses;
    }

    /** Prints the given JML field declaration. If this field
     * declaration has a generated initialzer block (e.g., one for a
     * ghost field), it is also printed. */
    public void visitJmlFieldDeclaration( JmlFieldDeclaration self ) {
        inInitializer = true;
	JVariableDefinition variable = self.variable();
	long modifiers = variable.modifiers();
        long specmodifiers = (modifiers &  
                      ~org.multijava.util.classfile.Constants.ACC_FINAL &
                      ~org.multijava.util.classfile.Constants.ACC_PRIVATE &
                      ~org.multijava.util.classfile.Constants.ACC_PROTECTED) |
                      org.multijava.util.classfile.Constants.ACC_PUBLIC;
	CType type = variable.getType();
	String ident = variable.ident();
	JExpression expr = variable.getValue();
        if (ident.indexOf("$") != -1) {
           return; // don't print generated elements
       } else {

	   newLine();
	   print(modUtil.asString(modifiers & 
                          ~org.multijava.util.classfile.Constants.ACC_FINAL));
           print(" " + type + " " + ident);
           println(";");
           
	   newLine();
/*
	final boolean startAnnotation = hasFlag(modifiers, ACC_MODEL) 
	    || hasFlag(modifiers, ACC_GHOST);
	mayStartAnnotation(startAnnotation);
*/
	   print(modUtil.asString(specmodifiers));
           print(" java.lang.Object ");
	   print(SPECFLD + ident);
	   if (expr != null) {
              if (!self.getField().isStatic()) {
                fieldInits.add(self);
              }
   //	    expr.accept(this);
           } else if (type.isPrimitive()) {
             print(" = ");
             printDefaultObject(type);
           }
	   println(";");
           if (self.getField().isStatic() && expr != null) {
             println("  static {");
             if (trace) {
               println("System.out.println(\"executing"  + 
                 " static initializer for " + ident + " of " + 
                 curClass.ident() + "\");");
             }
             inPre = true;
             println("  try {");
             println("  " + EPATH + "ConstraintSystem " + PREF + "csPost = " +
                ENPATH + "ConstraintSystemStack.getConstraintSystem();");
             println("  if (" + RPATH + "JMLTool.freshObjs == null) {");
             println("    " + RPATH + "JMLTool.freshObjs = new " + RPATH + 
                    "ObjectArrayList();");
             println("  }");
             getInitValue(expr, ident);
             println("  } finally {");
             println("    " + ENPATH + 
                     "ConstraintSystemStack.returnConstraintSystem();");
             println("  }");
             println("  }");
           }
   /*
	   mayEndAnnotation(startAnnotation);
   */
   
           // print assertion check code, if any, e.g., private variable
           // declarations for ghost variables and their initializers
           // (see TransClass.ghostFieldAccessors).
           if (self.hasAssertionCode()) {
               self.assertionCode().accept(this);
           }
           newLine();
           CField cFld = self.getField();
           print("  public ");
           if (cFld.isStatic()) {
           print("static ");
           }
           println("void " + cFld.owner().ident() + 
                   "$" + PREF + "set" + ident + "(final " + ENPATH + 
                   "Environment "
                   + PREF + "env, boolean " + PREF + "inConstructor) {");
           getCS();
           inPre = false;
           if (self.isModel() && cFld instanceof JmlSourceField) {
             JmlSourceField sFld = (JmlSourceField) cFld;
             CClass owner = cFld.owner();
             JmlAssignableFieldSet map = null;
             // if field was actually declared in an interface, get the 
             // data group map from the interface instead of the current class
             if (owner.isInterface() && owner instanceof JmlSourceClass) {
               JmlMemberDeclaration intf = ((JmlSourceClass) owner).getAST();
               if (intf instanceof JmlTypeDeclaration) {
                 map = ((JmlTypeDeclaration) 
                   intf).getDataGroupMap().getMembers(sFld);
               } else {
                 fail("Error: model field " + ident + 
                    " declared in something other than a JmlTypeDeclaration!");
               }
             } else {
               map = curClass.getDataGroupMap().getMembers(sFld);
             }
             if (map == null) {
               fail("Error: model field " + ident + " with no data group!");
             } else {
               Iterator it = map.iterator();
               while (it.hasNext()) {
                 CField fld = (CField) it.next();
                 genAssignable(fld.ident(), "", fld.getType());
               }
               acceptAll(getRepresents(cFld));
             }
             Iterator it = getMapsInto(self).iterator();
             while (it.hasNext()) {
               JmlStoreRef ref = ((JmlMapsIntoClause) it.next()).memberRef();
               if (ref instanceof JmlStoreRefExpression) {
                 inPre = false;
                 visitStoreRef((JmlStoreRefExpression) ref, 0, null);
                 inPre = true;
               } else {
                 fail("Error: invalid store reference in maps into clause");
               }
             }
           } else {
             genAssignable(ident, "", type);
           }
           inPre = true;
           println("  }");
       }
       inInitializer = false;
    }

    /** Prints a JML forall variable declaration. */
    public void visitJmlForAllVarDecl( JmlForAllVarDecl self ) 
    {
/*
	//newLineOffset();
	print("forall ");
	visitVarDecls(self.quantifiedVarDecls());
	print(";");
*/
    }

    /** Prints a Java variable declaration. */
    // used only for JmlLetVarDecls
    private void visitVarDecls(JVariableDefinition[] varDefs)
    {
        if (inPre) {
	  for (int i = 0; i < varDefs.length; i++) {
            String ident = varDefs[i].ident();
            CType type = varDefs[i].getType();
            declareVar(ident, type);
            JExpression expr = varDefs[i].expr();
            if (expr != null) {
               String oldRes = result;
               result = ident;
               expr.accept(this);
               result = oldRes;
            }
	  }
        }
    }

    private boolean isThis(JmlSpecExpression self) {
      return (self.expression() instanceof JThisExpression);
    }

    private boolean isJMLE(CType type) {
       String typeString = type.toString();
       if (type.isArrayType() || typeString.startsWith("java.util.") || 
           typeString.startsWith("org.jmlspecs.models.") ||
           typeString.startsWith("java.lang.")) {
          return false;
       } else {
          return true;
       } 
    }

    private void warn(TokenReference tref,
                      MessageDescription description,
                      Object obj) {
       CTopLevel.getCompiler().reportTrouble(
          new CWarning(tref, description, obj));
    }

    private void warn(TokenReference tref, MessageDescription description) {
       CTopLevel.getCompiler().reportTrouble(new CWarning(tref, description));
    }
    
    /** Prints a JML fresh expression. */
    public void visitJmlFreshExpression( JmlFreshExpression self ) 
    {
	Debug.msg("In JmlFreshExpression");
        JmlSpecExpression[] expressions = self.specExpressionList();
        for (int i = 0; i < expressions.length; i++) {
          if (!inBoundaryPred) {
            CType type = expressions[i].getType();
            String typeString = type.toString();
            if (isThis(expressions[i])) {
              warn(self.getTokenReference(), ExecMessages.THIS_NOT_FRESH);
              equalTrue();
            } else if (result != null) {
              String oldres = result;
              result = varGen.freshPreVar();
              declareVar(result, type);
              String expres = result;
              expressions[i].accept(this);
              print("  ");
              printCS();
              print(".addGoalConstraint(new " + CPATH + "DELAYConstraint("
                  + "new Object [] {" + expres + "},");
              println("  new " + RPATH + "GenMethod() {");
              print("    public void run(" + EPATH + "ConstraintSystem ");
              printCS();
              println(") {");
  
              String val = varGen.freshPreVar();
              print("    java.lang.Object " + val + " = ");
              printCS();
              println(".getVarValue(" + expres + ");");
  
              print("  ");
              printCS();
              println(".addGoalConstraint(new " + EPATH + "Equality(" + oldres +
                      ", new " + RPATH + "MyBoolean(" + PREF + 
                      "env.freshObjs().contains(" + val + "))));");
              println("}}));");

              result = null;
            } else {
              if (isJMLE(type)) {
                result = varGen.freshPreVar();
                declareVar(result, type);
                String expres = result;
                expressions[i].accept(this);
                println("  try {");
                String val = varGen.freshPreVar();
                print("    java.lang.Object " + val + " = ");
                printCS();
                println(".getVarValue(" + expres + ");");
                println("    if (" + PREF + "env.freshObjs().contains(" +
                        val + ")) {");
                equalTrue();
                println("    } else {");
                print("  ");
                printCS();
                println(".addGoalConstraint(new " + EPATH + "Bool(\"false\"));");
                println("    }");
                println("  } catch (" + RPATH + "NoValueException " +
                        varGen.freshPreVar() + ")  {");
                print("  ");
                printCS();
                println(".addGoalConstraint(new " + EPATH + "Equality(" + expres +
                      ", ");
                print("  new ");
                printType(type);
                println("(" + RPATH + "FreshConstructor.disambig, " + PREF + 
                        "env)));");
                println("  }");
              } else {
                result = varGen.freshPreVar();
                declareVar(result, type);
                String expres = result;
                expressions[i].accept(this);
                print("  ");
                printCS();
                print(".addGoalConstraint(new " + CPATH + "DELAYConstraint("
                    + "new Object [] {" + expres + "},");
                println("  new " + RPATH + "GenMethod() {");
                print("    public void run(" + EPATH + "ConstraintSystem ");
                printCS();
                println(") {");
    
                String val = varGen.freshPreVar();
                print("    java.lang.Object " + val + " = ");
                printCS();
                println(".getVarValue(" + expres + ");");
    
                println("  if (!" + PREF + "env.freshObjs().contains(" + val +
                        ")) {");
                print("  ");
                printCS();
                println(".addGoalConstraint(new " + EPATH + "Bool(\"false\"));");
                println("  }");
                println("}}));");

                result = null;
              }
            }
          }
        }
    }
    
    /** Prints a JML generic spec body. */
    public void visitJmlGenericSpecBody( JmlGenericSpecBody self ) 
    {
	Debug.msg("In JmlGenericSpecBody");
	visitSpecBody(self);
    }

    /** Prints a JML generic spec case. */
    public void visitJmlGenericSpecCase( JmlGenericSpecCase self ) 
    {
	Debug.msg("In JmlGenericSpecCase");

	visitGeneralSpecCase(self);
    }

    /** Prints a JML guarded statement. */
    public void visitJmlGuardedStatement( JmlGuardedStatement self ) {}

    /** Prints a JML henceby statement. */
    public void visitJmlHenceByStatement( JmlHenceByStatement self ) 
    {
/*
	mayStartAnnotation();
	print(self.isRedundantly() ?
	       "hence_by_redundantly " : "hence_by ");
	self.predicate().accept(this);
	print(";");
	mayEndAnnotation();

	// print RAC code if exists
	if (self.hasAssertionCode()) {
	    newLine();
	    self.assertionCode().accept(this);
	}
*/
    }

    /** Prints a JML in data group clause. */
    public void visitJmlInGroupClause( JmlInGroupClause self ) 
    {
    }

    /** Prints a JML informal expression. */
    public void visitJmlInformalExpression( JmlInformalExpression self ) 
    {
        warn(self.getTokenReference(), ExecMessages.INFORMAL_NOT_EXECUTED,
             "expressions");
        if (result != null) {
           equalTrue();
        }
    }

    /** Prints a JML informal store reference. */
    public void visitJmlInformalStoreRef( JmlInformalStoreRef self )
    {
        warn(self.getTokenReference(), ExecMessages.INFORMAL_NOT_EXECUTED,
             "store references");
    }

    /** Prints a JML initially variable assertion. */
    public void visitJmlInitiallyVarAssertion( 
       JmlInitiallyVarAssertion self ) {
      if (isStatic == self.isStatic()) {
        if (trace && isStatic) {
          println("System.out.println(\"executing"  + 
            " initially clause for " + curClass.ident() + "\");");
        }
        self.predicate().accept(this);
      }
    }

    /** Prints a JML invariant. */
    public void visitJmlInvariant( JmlInvariant self ) 
    {
	self.predicate().accept(this);
    }

    /** Prints a JML invariant for expression. */
    public void visitJmlInvariantForExpression( 
       JmlInvariantForExpression self ) 
    {
        String oldRes = result;
        result = varGen.freshPreVar();
        declareVar(result, self.getType());
        String res = result;
        self.specExpression().accept(this);
        print("  ");
        printType(self.specExpression().getType());
        print("." + PREF + "callInvariantsFor");
        if (oldRes != null) {
           print("Assert");
        }
        print("(");
        print(res + ", " + PREF + "env, ");
        print(inOld || inPre);
        if (oldRes != null) {
           print(", " + oldRes);
        }
        println(");");
        result = null;
    }

    /** Prints a JML invariant statement. */
    public void visitJmlInvariantStatement( JmlInvariantStatement self ) 
    {
    }

    /** Prints a JML isinitialized expression. */
    public void visitJmlIsInitializedExpression( 
       JmlIsInitializedExpression self ) 
    {
	print("\\is_initialized(" + toString(self.referenceType()) + ")");
    }

    /** Prints a JML label expression. */
    public void visitJmlLabelExpression( JmlLabelExpression self ) 
    {
	print(self.isPosLabel() ? "(\\lblpos " : "(\\lblneg ");
	print(self.ident() + " ");
	self.specExpression().accept(this);
	print(")");
    }

    /** Prints a JML old variable declaration. */
    public void visitJmlLetVarDecl( JmlLetVarDecl self ) 
    {
	//newLineOffset();
        boolean backOld = inOld;
        inOld = true;
	visitVarDecls(self.specVariableDeclarators());
        inOld = backOld;
    }

    /** Prints a JML old variable declaration. */
    public void visitJmlVariableDefinition( JmlVariableDefinition self ) 
    {
/*
	visitVariableDefinition(self);
*/
    }

    /** Prints a JML lockset expression. */
    public void visitJmlLockSetExpression( JmlLockSetExpression self ) 
    {
/*
	print("\\lockset");
*/
    }

    /** Prints a JML loop invariant. */
    public void visitJmlLoopInvariant( JmlLoopInvariant self ) 
    {
/*
	mayStartAnnotation();
	print(self.isRedundantly() ?
	       "maintaining_redundantly " : "maintaining ");
	self.predicate().accept(this);
	print(";");
	mayEndAnnotation();
*/
    }

    /** Prints a JML loop statement. A JML loop statement is a Java
     * loop statement with optional loop invariants and variants.
     */
    public void visitJmlLoopStatement( JmlLoopStatement self ) {
/*
        JmlLoopInvariant[] invs = self.loopInvariants();
        for (int i = 0; i < invs.length; i++) {
            invs[i].accept(this);
        }

        JmlVariantFunction[] vars = self.variantFunctions();
        for (int i = 0; i < vars.length; i++) {
            vars[i].accept(this);
        }

        if (invs.length > 0 || vars.length > 0) {
            //newLineOffset();
        }

	if (self.hasAssertionCode()) {
            // print instrumeneted code
	    newLine();
	    self.assertionCode().accept(this);
	} else {
            self.stmt().accept(this);
        }
*/
    }

    /** Prints a JML working space clause. */
    public void visitJmlWorkingSpaceClause( JmlWorkingSpaceClause self ) 
    {
/*
	Debug.msg("In JmlWorkingSpaceClause");
	
	String keyword = self.isRedundantly() ?
	    "working_space_redundantly " : "working_space ";	    
	//newLineOffset();
	if (self.isNotSpecified()) {
	    print(keyword + "\\not_specified;");
	} else {
	    print(keyword);
	    self.specExp().accept(this);
	    if (self.predOrNot() != null) {
		print(" if ");
		self.predOrNot().accept(this);
	    }
	    print(";");
	}
*/
    }

    /** Prints a JML duration clause. */
    public void visitJmlDurationClause( JmlDurationClause self ) 
    {
/*
	Debug.msg("In JmlDurationClause");
	
	String keyword = self.isRedundantly() ?
	    "duration_redundantly " : "duration ";	    
	//newLineOffset();
	if (self.isNotSpecified()) {
	    print(keyword + "\\not_specified;");
	} else {
	    print(keyword);
	    self.specExp().accept(this);
	    if (self.predOrNot() != null) {
		print(" if ");
		self.predOrNot().accept(this);
	    }
	    print(";");
	}
*/
    }

    /** Prints a JML maps into data group clause. */
    public void visitJmlMapsIntoClause( JmlMapsIntoClause self ) 
    {
/*
	print("visitJmlMapsIntoClause not implemented in RacPrettyPrinter");
*/
    }

    /** Prints a JML measured clause. */
    public void visitJmlMeasuredClause( JmlMeasuredClause self ) 
    {
/*
	Debug.msg("In JmlMeasuredClause");
	
	String keyword = self.isRedundantly() ?
	    "measured_by_redundantly " : "measured_by ";	    
	//newLineOffset();
	if (self.isNotSpecified()) {
	    print(keyword + "\\not_specified;");
	} else {
	    print(keyword);
	    self.specExp().accept(this);
	    if (self.predOrNot() != null) {
		print(" if ");
		self.predOrNot().accept(this);
	    }
	    print(";");
	}
*/
    }

    /** Prints a JML method name. */
    public void visitJmlMethodName( JmlMethodName self ) 
    {
	JmlName[] subnames = self.subnames();
	for (int i = 0; i < subnames.length; i++) {
	    if (i != 0)
		print(".");
	    subnames[i].accept(this);
	}
	print("(");
	if (self.hasParamDisambigList()) {
	    CType[] params = self.paramDisambigList();
	    for (int i = 0; i < params.length; i++) {
		if (i != 0)
		    print(", ");
		print(toString(params[i]));
	    }
	}
	print(")");
    }

    /** Prints a JML model program. */
    public void visitJmlModelProgram( JmlModelProgram self ) {}

    /** Prints a JML MonitorsFor variable assertion. */
    public void visitJmlMonitorsForVarAssertion( 
       JmlMonitorsForVarAssertion self ) {}

    /** Prints a JML <code>\nonnullelements</code> expression. */
    public void visitJmlNonNullElementsExpression( 
         JmlNonNullElementsExpression self ) 
    {
/*
	print("\\nonnullelements(");
	self.specExpression().accept(this);
	print(")");
*/
        String oldRes = result;
        result = varGen.freshPreVar();
        String exprRes = result;
        CType type = self.specExpression().getType();
        declareVar(result, type);
        self.specExpression().accept(this);
        print("  ");
        printCS();
        print(".addGoalConstraint(new " + CPATH + "DELAYConstraint("
            + "new Object [] {" + exprRes + "},");
        println("  new " + RPATH + "GenMethod() {");
        print("    public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        println(") {");
        String res = varGen.freshPreVar();
        getValue(exprRes, res, type);
        println("    if (" + res + " == null) {");
        printCS();
        print(".addGoalConstraint(new ");
        print(CPATH + "EQUALEQUALConstraint(");
        if (oldRes == null) {
          print(RPATH + "MyBoolean.TRUE(), ");
        } else {
          print(oldRes + ", ");
        }
        println(RPATH + "MyBoolean.FALSE()));");
        println("    } else {");
        print("  ");
        printCS();
        print(".addGoalConstraint(new " + CPATH + "DELAYConstraint("
            + res + ",");
        println("  new " + RPATH + "GenMethod() {");
        print("    public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        println(") {");
        String flag = varGen.freshPreVar();
        String ind = varGen.freshPreVar();
        println("    boolean " + flag + " = true;");
        println("    for (int " + ind + " = 0; " + ind + " < " + res + 
                ".length && " + flag + "; " + ind + "++) {");
        println("      if (" + res + "[" + ind + "] == null) {");
        println("        " + flag + " = false;");
        println("      }");
        println("    }");
        printCS();
        print(".addGoalConstraint(new ");
        print(CPATH + "EQUALEQUALConstraint(");
        if (oldRes == null) {
          print(RPATH + "MyBoolean.TRUE(), ");
        } else {
          print(oldRes + ", ");
        }
        println("new " + RPATH + "MyBoolean(" + flag + ")));");
        println("}}));");
        println("    }");
        println("}}));");
        result = null;
    }

    /** Prints a JML nondeterministic choice statement. */
    public void visitJmlNondetChoiceStatement( 
       JmlNondetChoiceStatement self ) {}

    /** Prints a JML nondeterministic if statement. */
    public void visitJmlNondetIfStatement( JmlNondetIfStatement self ) {}

    /** Prints a JML normal behavior specification. */
    public void visitJmlNormalBehaviorSpec( JmlNormalBehaviorSpec self ) 
    {
	//newLineOffset();
/*
	print(modUtil.asString(self.privacy()));
	print("normal_behavior");
	offset += TAB_SIZE;
*/
	self.specCase().accept(this);
/*
	offset -= TAB_SIZE;	
*/
    }

    /** Prints a JML normal example. */
/*
    public void visitJmlNormalExample( JmlNormalExample self ) 
    {
	//newLineOffset();
	print(modUtil.asString(self.privacy()));
	print("normal_example");
	offset += TAB_SIZE;
	visitExample(self);
	offset -= TAB_SIZE;
    }
*/

    /** Prints a JML normal spec body. */
    public void visitJmlNormalSpecBody( JmlNormalSpecBody self ) 
    {
	Debug.msg("In JmlNormalSpecBody");
	visitSpecBody( self );
    }

    /** Prints a JML normal spec case. */
    public void visitJmlNormalSpecCase( JmlNormalSpecCase self ) 
    {
	Debug.msg("In JmlExceptionalBehaviorSpec");
	visitGeneralSpecCase(self);
    }

    /** Prints a JML <code>\not_modified</code> expression. */
    public void visitJmlNotModifiedExpression(JmlNotModifiedExpression self) 
    {
	JmlStoreRef[] storeRefs = self.storeRefList();
        boolean backOld = inOld;
        boolean isPrim = false;
        String oldRes = null;
        if (result != null) {
          oldRes = varGen.freshPreVar();
          declareVar(oldRes, RPATH + "MyBoolean", false);
          printCS();
          print(".addGoalConstraint(new " + EPATH + "Equality(" + result);
          println(", " + oldRes + "));");
        }
        String nextRes = null;
        if (oldRes != null) {
          nextRes = varGen.freshPreVar();
          println("  java.lang.Object " + nextRes + " = null;");
        }
        for (int i = 0; i < storeRefs.length; i++) {
          JmlStoreRef ref = storeRefs[i];
          if (ref instanceof JmlStoreRefExpression) {
            String modRes = null;
            if (oldRes != null) {
              modRes = varGen.freshPreVar();
              declareVar(modRes, RPATH + "MyBoolean", false);
              println("  " + nextRes + " = new " + RPATH + "VarObject();");
              printCS();
              println(".addVariable(" + nextRes + ", \"" + RPATH + 
                    "MyBoolean\");");
              printCS();
              print(".addGoalConstraint(new " + CPATH + "CANDConstraint(");
              println(modRes + ", " + nextRes + ", " + oldRes + "));");
              println("  " + oldRes + " = " + nextRes + ";");
            }
            visitStoreRef((JmlStoreRefExpression) ref, 1, modRes);
          } else {
            throw new 
              org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
             "Error: invalid store reference for \\not_modified, " +
              formatToken(self) + ".");
          }
        }
        if (oldRes != null) {
          printCS();
          print(".addGoalConstraint(new " + EPATH + "Equality(");
          println(oldRes + ", " + RPATH + "MyBoolean.TRUE()));");
        }
        inOld = backOld;
        result = null;
    }


    /** Prints a JML <code>\old</code> expression. */
    public void visitJmlOldExpression( JmlOldExpression self ) 
    {
        boolean backOld = inOld;
        inOld = true;
	self.specExpression().accept(this);
        inOld = backOld;
    }

    /** Prints a JML <code>\max</code> expression. */
    public void visitJmlMaxExpression( JmlMaxExpression self ) 
    {
	print("\\max(");
	self.expression().accept(this);
	print(")");
    }

    /** Prints a JML <code>\duration</code> expression. */
    public void visitJmlDurationExpression( JmlDurationExpression self ) 
    {
/*
	print("\\duration(");
	self.expression().accept(this);
	print(")");
*/
    }

    /** Prints a JML <code>\working_space</code> expression. */
    public void visitJmlWorkingSpaceExpression( JmlWorkingSpaceExpression self ) 
    {
/*
	print("\\working_space(");
	self.expression().accept(this);
	print(")");
*/
    }

    /** Prints a JML <code>\space</code> expression. */
    public void visitJmlSpaceExpression( JmlSpaceExpression self ) 
    {
/*
	print("\\space(");
	self.specExpression().accept(this);
	print(")");
*/
    }

    /** Prints a JML package import statement. */
    public void visitJmlPackageImport( JmlPackageImport self ) { 
        if (self.isModel()) {
            print("/*@ model */");
        }
	String name = self.getName().replace('/', '.');
	print("import " + name + ".*;");
        imports.add(self.getName());
    }      

    /** Prints a JML reach expression. */
    public void visitJmlReachExpression( JmlReachExpression self ) 
    {
        throw new 
          org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
         "Error: \\reach expression is not executable, " +
          formatToken(self) + ".");
/*
	CType type = self.referenceType();
	JmlStoreRefExpression storeRef = self.storeRefExpression();

	print("\\reach(");
	self.specExpression().accept(this);
	if (type != null)
	    print(", " + toString(type));
	if (storeRef != null) {
	    print(", ");
	    storeRef.accept(this);
	}
	print(")");	
*/
    }

    /** Prints a JML readableif variable assertion. */
    public void visitJmlReadableIfVarAssertion( 
       JmlReadableIfVarAssertion self ) {}

    /** Prints a JML redundant specifications including
     * <code>implies_that</code> and <code>for_example</code>.
     */
    public void visitJmlRedundantSpec( JmlRedundantSpec self ) 
    {
      if (self != null) {
	JmlSpecCase[] implications = self.implications();
        if (implications != null) {
           DesugarSpec desugar = new DesugarSpec();
           JmlMethodSpecification desugaredSpec = null;
           JmlMethodSpecification transform = 
             new JmlSpecification(self.getTokenReference(), implications, null);
           desugaredSpec = desugar.perform(transform, exceptions);
           desugaredSpec.accept(this);
         }
      }
                                      //    self.getExceptions());
	  //specCases = desugaredSpec.specCases();
/*
	JmlSpecCase[] implications = self.implications();
	if (implications != null) {
	    //newLineOffset();
//	    print("implies_that");
	    offset += TAB_SIZE;	    
	    for (int i = 0; i < implications.length; i++) {
		implications[i].accept(this);
	    }
	}
*/

	JmlExample[] examples = self.examples();
	if (examples != null) {
          warn(self.getTokenReference(), ExecMessages.EXAMPLES_NOT_EXECUTED);
/*
	    //newLineOffset();
	    print("for_example");
	    offset += TAB_SIZE;	    
	    for (int i = 0; i < examples.length; i++) {
		if (i != 0) {
		    offset -= TAB_SIZE;
		    //newLineOffset();
		    print("also");
		    offset += TAB_SIZE;
		}
		examples[i].accept(this);
	    }
	    offset -= TAB_SIZE;
*/
	}
    }

    /** Prints a JML refine prefix. */
    public void visitJmlRefinePrefix( JmlRefinePrefix self ) {}

    /** Prints a JML <code>represents</code> declaration. */
    public void visitJmlRepresentsDecl( JmlRepresentsDecl self ) 
    {
       if (self.isRedundantly()) {
        warn(self.getTokenReference(), ExecMessages.REDUNDANT_NOT_EXECUTED,
             "represents declarations");
       } else {
         inPre = false;
	 if (self.usesAbstractionFunction()) {
           if (!(self.storeRef() instanceof JmlStoreRefExpression)) {
             fail("invalid store reference for represents declaration");
           } else {
	     JmlStoreRefExpression storeRef = 
               (JmlStoreRefExpression) self.storeRef();
             result = varGen.freshPreVar();
             declareVar(result, storeRef.getType());
             String res = result;
	     self.specExpression().accept(this);
             print("  ");
             printCS();
             print(".addGoalConstraint(new " + EPATH + "Equality(");
             visitStoreRef(storeRef, 2, null);
             println(", " + res + "));");
             result = null;
           }
         } else {
	   self.predicate().accept(this);
         }
         inPre = true;
       }
    }

    /** Prints a JML requires clause. */
    public void visitJmlRequiresClause( JmlRequiresClause self ) 
    {
	Debug.msg("In JmlRequiresClause");

	visitSpecBodyClause(self, "requires");
    }

    /** Prints a JML <code>\result</code> expression. 
        requires result != null */
    public void visitJmlResultExpression( JmlResultExpression self ) 
    {
//	print("\\result");
       print("  ");
       printCS();
       print(".addGoalConstraint(new " + EPATH + "Equality(" + PREF + 
             "result, ");
       if (result == null) {
         print(RPATH + "MyBoolean.TRUE()");
       } else {
         print(result);
       }
       println("));");
       result = null;
    }

    /** Prints a JML return clause. */
    public void visitJmlReturnsClause( JmlReturnsClause self ) {}

    /** Prints a JML set comprehension expression. */
    public void visitJmlSetComprehension( JmlSetComprehension self ) 
    {
	String ident = self.ident();
	CType type = self.getApparentType();
	CType memberType = self.memberType();
	JExpression supersetPred = self.supersetPred();
	JmlPredicate predicate = self.predicate();
        String typ =  type.toString().substring(MODPATHLEN);

        if (supersetPred instanceof JMethodCallExpression) {
           JExpression prefix = ((JMethodCallExpression) supersetPred).prefix();
           String oldRes = result;
           result = varGen.freshPreVar();
           declareVar(result, prefix.getType());
           String boundRes = result;
           prefix.accept(this);
           result = PREF + "condRes";
           print("  ");
           printCS();
           print(".addGoalConstraint(");
           print("new " + CPATH + "SETCOMPConstraint(" + boundRes + ",");
           println("  new " + RPATH + "GenMethod2() {");
           print("      public void run(" + EPATH + "ConstraintSystem ");
           printCS();
           print(", final java.lang.Object " + ident);
           println(", final java.lang.Object " + PREF + "condRes) {");
	   if (predicate != null)
	       predicate.accept(this);
           else {
             equalTrue();
           }
           println("    }}, \"" + typ + "\", new " + RPATH + "Thunk(" + 
                   oldRes + ")));");
           result = null;
        } else {
            fail();
        }
    }

    /** Prints the given JML set statement. */
    public void visitJmlSetStatement( JmlSetStatement self ) {}

    private static boolean isTrueOrNull(JmlPredicate self) {
       if (self == null) {
          return true;
       } else {
          JExpression expr = self.specExpression().expression();
          if (expr instanceof JBooleanLiteral) {
            return ((JBooleanLiteral) expr).booleanValue();
          } else {
            return false;
          }
       }
    }

    private static boolean isFalse(JmlPredicate self) {
       if (self == null) {
          return false;
       } else {
          JExpression expr = self.specExpression().expression();
          if (expr instanceof JBooleanLiteral) {
            return !((JBooleanLiteral) expr).booleanValue();
          } else {
            return false;
          }
       }
    }

    private static boolean hasDefaultConstructor(CClass typ) {
      JmlTypeDeclaration cls = 
          JmlTypeLoader.getJmlSingleton().typeDeclarationOf(typ);
      ArrayList meths = cls.methods();
      Iterator it = meths.iterator();
      while (it.hasNext()) {
//      for (Object meth : meths) {
        Object meth = it.next();
        if (meth instanceof JmlConstructorDeclaration) {
          JmlConstructorDeclaration con = (JmlConstructorDeclaration) meth;
          if (con.parameters().length == 0) {
            return true;
          }
        }
      }
      return false;
    }

    /** Prints a JML signals_only clause */
    public void visitJmlSignalsOnlyClause( JmlSignalsOnlyClause self )  {
      println("    throw new " + self.exceptions()[0].toString() + "();");
    }

    private boolean isGeneratedSignals(JExpression self) {
       if (self instanceof JInstanceofExpression) {
         return true;
       } else if (self instanceof JConditionalOrExpression) {
         JExpression left = ((JConditionalOrExpression) self).left();
         JExpression right = ((JConditionalOrExpression) self).right();
         return isGeneratedSignals(left) && isGeneratedSignals(right);
       }
       return false;
    }

    private boolean isGeneratedSignals(JmlSignalsClause self) {
      if (toString(self.type()).equals("java.lang.Throwable") &&
          self.predOrNot() instanceof JmlPredicate &&
          self.predOrNot().specExpression() instanceof JmlSpecExpression) {
         return 
           isGeneratedSignals(self.predOrNot().specExpression().expression());
      }
      return false;
    }

    /** Prints a JML signals clause */
    public void visitJmlSignalsClause( JmlSignalsClause self ) 
    {
      Debug.msg("In JmlSignalsClause");
/*
      if (!hasSignals) {
        return;
      } else if (self.isRedundantly()) {
	System.out.println("WARNING: redundant signals clauses are" +
                           " not checked, " +
          self.getTokenReference().toString().replace('"', '\'') + ".");
      } else if (self.isNotSpecified()) {
        throw new 
          org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
      "Error: signals clause with unspecified predicate is not executable, " +
          self.getTokenReference().toString().replace('"', '\'') + ".");
      } else if (isGeneratedSignals(self)) {
	System.out.println("WARNING: ignoring generated signals clause, " +
          self.getTokenReference().toString().replace('"', '\'') + ".");
      } else if (!isFalse(self.predOrNot())) {
*/
        if (trace) {
           println("  System.out.println(\"executing signals clause for "
                   + curMeth.ident() + " method of " + curClass.ident() 
                   + "\");");
        }
        String ident = self.ident();
        String newVar = varGen.freshPreVar();
        declareVar(newVar, self.type());
        String newMeth = varGen.freshPreVar();
        print("  ExceptionMethod" + exMethNum + " " + newMeth + " = ");
//        print("  " + PREF + "exceptions.put(" + newVar + ",");
        println(" new ExceptionMethod" + exMethNum + "() {");
        print("      public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        print(") ");
        printThrows(exceptions);
        println(" {");
        println("  {");
        print("  ");
        printCS();
        print(".addGoalConstraint(new " + CPATH + "NEWEXCEPTIONConstraint(");
        println(newVar + ", " + toString(self.type()) + ".class));");
        if (ident != null) {
          declareVar(ident, self.type());
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + EPATH + "Equality(");
          println(newVar + ", " + ident + "));");
        }
        if (self.predOrNot() != null) {
          self.predOrNot().accept(this);
        }
        println("  }");
        println("    }};");
        print("    " + newMeth + ".run(");
        printCS();
        println(");");
        result = null;
    }

   private void equalTrue() {
     if (result != null) {
       print("  ");
       printCS();
       println(".addGoalConstraint(new " + EPATH + "Equality(");
       println(result + ", " + RPATH + "MyBoolean.TRUE()));");
       result = null;
     }
   }

    /** Prints a Java quantified variable declaration. */
    private void visitQuantVarDecls(JVariableDefinition[] varDefs)
    {
	for (int i = 0; i < varDefs.length; i++) {
            declareVar(varDefs[i].ident(), varDefs[i].getType());
            String bpRes = varGen.freshPreVar();
            declareVar(bpRes, "org.jmlspecs.models.JMLCollection");
            quantVars.push(varDefs[i].ident(), bpRes, varDefs[i].getType());
	    if (varDefs[i].expr() != null) {
              fail("variables bound by quantifiers can not be initialized");
	    }
	}
    }
 
    private void printIdentity(JmlSpecQuantifiedExpression self, CType type) {
      if (self.isMax()) {
	if (type == CStdType.Byte) {
          printWrapped(Byte.toString(Byte.MIN_VALUE), type);
	} else if (type == CStdType.Integer) {
          printWrapped(Integer.toString(Integer.MIN_VALUE), type);
	} else if (type == CStdType.Long) {
          printWrapped(Long.toString(Long.MIN_VALUE), type);
	} else if (type == CStdType.Short) {
          printWrapped(Short.toString(Short.MIN_VALUE), type);
	} else if (type == CStdType.Char) {
          printWrapped(Character.toString(Character.MIN_VALUE), type);
        } else {
          fail("Error: invalid type for variable bound by generalized " +
               "quantifier");
        }
      } else if (self.isMin()) {
	if (type == CStdType.Byte) {
          printWrapped(Byte.toString(Byte.MAX_VALUE), type);
	} else if (type == CStdType.Integer) {
          printWrapped(Integer.toString(Integer.MAX_VALUE), type);
	} else if (type == CStdType.Long) {
          printWrapped(Long.toString(Long.MAX_VALUE), type);
	} else if (type == CStdType.Short) {
          printWrapped(Short.toString(Short.MAX_VALUE), type);
	} else if (type == CStdType.Char) {
          printWrapped(Character.toString(Character.MAX_VALUE), type);
        } else {
          fail("Error: invalid type for variable bound by generalized " +
               "quantifier");
        }
      } else if (self.isProduct()) {
        printWrapped("1", type);
      } else if (self.isSum() || self.isNumOf()) {
        printWrapped("0", type);
      } else {
        fail("Error: invalid type of quantifier");
      }
    }
 
    private boolean isImplication(JExpression self) {
      if (self instanceof JmlSpecExpression) {
//System.out.println("ANT");
        JExpression expr = ((JmlSpecExpression) self).expression();
//System.out.println(((JmlSpecExpression) self).expression());
//System.out.println(((JmlSpecExpression) self).expression().getClass());
        if (expr instanceof JmlRelationalExpression) {
            return ((JmlRelationalExpression) expr).isImplication();
        } else if (expr instanceof JConditionalExpression) {
/*
System.out.println("COND");
System.out.println(((JConditionalExpression) expr).cond());
System.out.println();
System.out.println(((JConditionalExpression) expr).left());
System.out.println();
System.out.println(((JConditionalExpression) expr).right());
System.out.println();
*/
        }
      } else if (self instanceof JmlRelationalExpression) {
         return ((JmlRelationalExpression) self).isImplication();
      }
      return false;
    }

    /** Prints a JML specification quantified expression. */
    public void visitJmlSpecQuantifiedExpression( 
       JmlSpecQuantifiedExpression self ) 
    {
        boolean backAssert = inAssert;
        if (inBoundaryPred) {
          equalTrue();
        } else if (self.isForAll()) {
//	    print("\\forall ");
            andOnly = true;
            println("  {");
	    visitQuantVarDecls(self.quantifiedVarDecls());
            int numVars = self.quantifiedVarDecls().length;
            boolean oldPred = inBoundaryPred;
            inBoundaryPred = true;
            String oldRes = result;
            result = null;
            JExpression pred = self.predicate();
            JExpression ant = self.specExpression();
            if (pred == null && isImplication(ant)) {
              ant = ((JmlSpecExpression) ant).expression();
              while (isImplication(ant)) {
                JExpression cons = ((JmlRelationalExpression) ant).left();
                if (pred == null) {
                  pred = cons;
                } else {
                  pred = new JConditionalAndExpression(
                            self.getTokenReference(), pred, cons);
                }
                ant = ((JmlRelationalExpression) ant).right();
              }
            }
            secondTime = false;
	    if (pred != null) {
	        pred.accept(this);
            } else {
                throw new 
                  org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                 "Error: only partial bounds for quantified variable, " +
                  formatToken(self) + ".");
            }
            if (!andOnly) {
              throw new 
                 org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
               "Error: only partial bounds for quantified variable, " +
                formatToken(self) + ".");
            } 
            String resObj = null, nextRes = null;
            Stack results = null;
            if (oldRes != null) {
              resObj = oldRes;
              results = new Stack();
              results.push(resObj);
            }
            for (int i = quantVars.size() - numVars; i < quantVars.size();
                 i++) {
              BoundQuantVariables.boundVar bv = quantVars.get(i);
              if (oldRes == null) {
                print("  ");
                printCS();
                print(".addGoalConstraint(");
                println("new " + CPATH + "FORALLConstraint(" + bv.resVar + ",");
                println("  new " + RPATH + "GenMethod1() {");
                print("      public void run(" + EPATH + "ConstraintSystem ");
                printCS();
                println(", final java.lang.Object " + bv.id + ") {");
              } else {
                backAssert = inAssert;
                inAssert = true;
                print("  ");
                printCS();
                print(".addGoalConstraint(");
                println("new " + CPATH + "FORALLRConstraint(" + bv.resVar +
                         ",");
                println("  new " + RPATH + "GenMethod2() {");
                print("      public void run(" + EPATH + "ConstraintSystem ");
                printCS();
                print(", final java.lang.Object " + bv.id);
                nextRes = varGen.freshPreVar();
                results.push(nextRes);
                println(", final java.lang.Object " + nextRes + ") {");
                resObj = nextRes;
              }
            }
            result = varGen.freshPreVar();
            declareVar(result, RPATH + "MyBoolean");
            String antRes = result;
            secondTime = true;
            inBoundaryPred = oldPred;
	    if (pred != null)
	        pred.accept(this);
            else {
              equalTrue();
            }
            secondTime = false;
            if (oldRes == null) {
              print("  ");
              printCS();
              println(".addGoalConstraint(new " + CPATH + "IMPLYConstraint(");
              println(antRes + ",");
              println("  new " + RPATH + "GenMethod() {");
              print("      public void run(" + EPATH + "ConstraintSystem ");
              printCS();
              println(") {");
	      ant.accept(this);
              println("    }}));");
            } else {
              result = varGen.freshPreVar();
              declareVar(result, RPATH + "MyBoolean");
              String consRes = result;
	      ant.accept(this);
              print("  ");
              printCS();
              print(".addGoalConstraint(new " + CPATH + "IMPLYRConstraint(");
              println(antRes + ", " + consRes + ", " + resObj + "));");
              results.pop();
            }
            print("  ");
            for (int i = 0; i < numVars; i++) {
              if (oldRes == null) {
                print("}}));");
              } else {
                print("}}, ");
                print("new " + RPATH + "Thunk(" + results.pop() + ")));");
              }
            }
            println(" ");
            result = null;

            for (int i = quantVars.size() - numVars; i < quantVars.size();
                 i++) {
              BoundQuantVariables.boundVar bv = quantVars.get(i);
              if (bv.lb != null) {
                if (bv.ub != null) {
                  print("  ");
                  printCS();
                  print(".addGoalConstraint(new " + CPATH + 
                      "BSETConstraint(" + bv.lb + ", " + bv.ub);
                  println(", " + bv.resVar + "));");
                } else {
                throw new 
                  org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                 "Error: only partial bounds for quantified variable, " +
                  formatToken(self) + ".");
                }
              } else if (bv.ub != null) {
                throw new 
                  org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                 "Error: only partial bounds for quantified variable, " +
                  formatToken(self) + ".");
              }
            }
            println("  }");
            for (int i = 0; i < numVars; i++) {
              quantVars.pop();
            }
            inAssert = backAssert;
            result = null;
	} else if (self.isExists()) {
            andOnly = true;
            println("  {");
	    visitQuantVarDecls(self.quantifiedVarDecls());
            boolean oldPred = inBoundaryPred;
            inBoundaryPred = true;
            String oldRes = result;
            result = null;
            boolean hasPred = true;
            secondTime = false;
	    if (self.hasPredicate()) {
	        self.predicate().accept(this);
            } else {
                hasPred = false;
                self.specExpression().accept(this);
            }
            int numVars = self.quantifiedVarDecls().length;
            boolean predAnd = andOnly;
            String resObj = null, nextRes = null;
            Stack results = null;
            if (oldRes != null) {
              backAssert = inAssert;
              inAssert = true;
              resObj = oldRes;
              results = new Stack();
              results.push(resObj);
              for (int i = quantVars.size() - numVars; i < quantVars.size();
                   i++) {
                BoundQuantVariables.boundVar bv = quantVars.get(i);
                print("  ");
                printCS();
                print(".addGoalConstraint(");
                println("new " + CPATH + "EXISTSRConstraint(" + bv.resVar +
                         ",");
                println("  new " + RPATH + "GenMethod2() {");
                print("      public void run(" + EPATH + "ConstraintSystem ");
                printCS();
                print(", final java.lang.Object " + bv.id);
                nextRes = varGen.freshPreVar();
                results.push(nextRes);
                println(", final java.lang.Object " + nextRes + ") {");
                resObj = nextRes;
              }
              secondTime = true;
              inBoundaryPred = oldPred;
              result = varGen.freshPreVar();
              declareVar(result, RPATH + "MyBoolean");
              String res1 = result;
	      if (self.hasPredicate()) {
	        self.predicate().accept(this);
              } else {
//                equalTrue();
                self.specExpression().accept(this);
              }
              if (hasPred) {
                result = varGen.freshPreVar();
                declareVar(result, RPATH + "MyBoolean");
                String res2 = result;
	        self.specExpression().accept(this);
                print("  ");
                printCS();
                print(".addGoalConstraint(new " + CPATH + "CANDConstraint(");
                println(res1 + ", " + res2 + ", " + resObj + "));");
              } else {
                printCS();
                print(".addGoalConstraint(new " + EPATH + "Equality(");
                println(res1 + ", " + resObj + "));");
              }
              secondTime = false;
              results.pop();
              print("  ");
              for (int i = 0; i < numVars; i++) {
                print("}}, ");
                print("new " + RPATH + "Thunk(" + results.pop() + ")));");
              }
              println(" ");
              result = null;
            } else {
              inBoundaryPred = oldPred;
	      if (self.hasPredicate())
	        self.predicate().accept(this);
	      self.specExpression().accept(this);
            }
            for (int i = quantVars.size() - numVars; i < quantVars.size();
                 i++) {
              BoundQuantVariables.boundVar bv = quantVars.get(i);
              if (bv.isBound) {
                  if (!predAnd) {
                    throw new 
                 org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                     "Error: only partial bounds for quantified variable, " +
                      formatToken(self) + ".");
		}
                if (oldRes == null) {
                  print("  ");
                  printCS();
                  print(".addGoalConstraint(new " + CPATH + "HASConstraint(");
                  println(bv.resVar + ", " + bv.id + ", \"" + bv.boundingType +
                          "\"));");
                }
              } else if (bv.type.isOrdinal() && oldRes == null) {
                  printFD(bv.id, bv.type, "null");
              } else if (bv.lb != null && bv.ub != null && predAnd) {
                print("  ");
                printCS();
                print(".addGoalConstraint(new " + CPATH + 
                    "BSETConstraint(" + bv.lb + ", " + bv.ub);
                println(", " + bv.resVar + "));");
              } else if (oldRes != null) {
                  throw new 
               org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                   "Error: only partial bounds for quantified variable, " +
                    formatToken(self) + ".");
              }
            }
            println("  }");
            for (int i = 0; i < numVars; i++) {
              quantVars.pop();
            }
            inAssert = backAssert;
	} else if (self.isMax() || self.isMin() || self.isProduct() || 
                   self.isSum() || self.isNumOf()) {
            int numVars = self.quantifiedVarDecls().length;
            if (numVars > 1) {
                throw new 
                  org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                 "Error: generalized quantifiers over more than one variable" +
                 " are not executable, " +
                  formatToken(self) + ".");
            } else {
              andOnly = true;
              println("  {");
	      visitQuantVarDecls(self.quantifiedVarDecls());
              boolean oldPred = inBoundaryPred;
              inBoundaryPred = true;
              String oldRes = result;
              result = null;
              secondTime = false;
	      if (self.hasPredicate())
	          self.predicate().accept(this);
              if (!andOnly) {
                throw new 
                org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                 "Error: only partial bounds for quantified variable, " +
                  formatToken(self) + ".");
              } 
              BoundQuantVariables.boundVar bv = 
                quantVars.get(quantVars.size() - 1);
              if (bv.lb != null) {
                if (bv.ub != null) {
                  print("  ");
                  printCS();
                  print(".addGoalConstraint(new " + CPATH + 
                      "BSETConstraint(" + bv.lb + ", " + bv.ub);
                  println(", " + bv.resVar + "));");
                } else 
                throw new 
                  org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                 "Error: only partial bounds for quantified variable, " +
                  formatToken(self) + ".");
              } else if (bv.ub != null) {
                throw new 
                  org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                 "Error: only partial bounds for quantified variable, " +
                  formatToken(self) + ".");
              }

              print("  ");
              printCS();
              print(".addGoalConstraint(");
              print("new " + CPATH);
	      if (self.isMax()) {
                print("MAXOF");
              } else if (self.isMin()) {
                print("MINOF");
              } else if (self.isProduct()) {
                print("PRODUCT");
              } else if (self.isSum() || self.isNumOf()) {
                print("SUM");
              }
              print("Constraint(" + bv.resVar + ",");
              println("  new " + RPATH + "GenMethod2() {");
              print("      public void run(" + EPATH + "ConstraintSystem ");
              printCS();
              print(", final java.lang.Object " + bv.id);
              String bodRes = varGen.freshPreVar();
              println(", final java.lang.Object " + bodRes + ") {");

              result = varGen.freshPreVar();
              declareVar(result, RPATH + "MyBoolean");
              String condRes = result;
              inBoundaryPred = oldPred;
              secondTime = true;
	      if (self.hasPredicate())
	          self.predicate().accept(this);
              else {
                equalTrue();
              }
              if (self.isNumOf()) {
                result = varGen.freshPreVar();
                declareVar(result, RPATH + "MyBoolean");
                String exprRes = result;
	        self.specExpression().accept(this);
                String newRes = varGen.freshPreVar();
                declareVar(newRes, RPATH + "MyBoolean");
                print("  ");
                printCS();
                print(".addGoalConstraint(new " + CPATH + "CANDConstraint(");
                println(condRes + ", " + exprRes + ", " + newRes + "));");
                condRes = newRes;
              }
              secondTime = false;

              print("  ");
              printCS();
              print(".addGoalConstraint(new " + CPATH + "CONDEXPRConstraint(");
              println(condRes + ", ");
              println("  new " + RPATH + "GenMethod() {");
              print("      public void run(" + EPATH + "ConstraintSystem ");
              printCS();
              println(") {");
              if (self.isNumOf()) {
                print("  ");
                printCS();
                print(".addGoalConstraint(new " + EPATH + "Equality(" + bodRes);
                print(", ");
                printWrapped("1", CStdType.Long);
                println("));");
              } else {
                result = bodRes;
	        self.specExpression().accept(this);
              }
              println("    }},");
              println("  new " + RPATH + "GenMethod() {");
              print("      public void run(" + EPATH + "ConstraintSystem ");
              printCS();
              println(") {");
              print("  ");
              printCS();
              print(".addGoalConstraint(new " + EPATH + "Equality(" + bodRes);
              print(", ");
              CType varType;
              if (self.isNumOf()) {
                varType = CStdType.Long;
              } else {
                varType = self.quantifiedVarDecls()[0].getType();
              }
              printIdentity(self, varType);
              println("));");
              println("    }}));");

              print("}}, \"");
	      printType(varType);
              println("\", new " + RPATH + "Thunk(" + oldRes + ")));");
              println("}");

              quantVars.pop();
            }
	} else
            fail("Error: invalid type of quantifier");
    }

    /** Prints a JML specification statement. */
    public void visitJmlSpecStatement( JmlSpecStatement self ) {}

    /** Prints a JML specification. */
    public void visitJmlSpecification( JmlSpecification self ) 
    {
	Debug.msg("In JmlSpecification");
	visitMethodSpecification(self);
    }

    private boolean isFormalParameter(JmlName param) {
      boolean result = false;
      for (int i = 0; formalParameters != null && i < formalParameters.length;
           i++) {
        if (formalParameters[i].ident().equals(param.getName())) {
          result = true;
        }
      }
      return result;
    }

    private CType getFormalType(JmlName param) {
      for (int i = 0; formalParameters != null && i < formalParameters.length;
           i++) {
        if (formalParameters[i].ident().equals(param.getName())) {
          return formalParameters[i].getType();
        }
      }
      return null;
    }

    /** print a constructor call to build object of specified type */
    private void printDefaultObject(CType type) {
      print("new ");
      if (type == null) {
         System.out.println("fatal error: no type!");
         System.exit(0);
/*
      } else if (type.isArrayType()) {
         printType(((CArrayType) type).getBaseType());
         for (int i = 0; i < ((CArrayType) type).getArrayBound(); i++) {
           print("[0]");
         }
*/
      } else {
         // assume default constructor exists
         printType(type);
         print("()");
      }
    }

    /** Prints a JML store reference expression. */
    public void visitJmlStoreRefExpression( JmlStoreRefExpression self ) 
    {
	JmlName[] names = self.names();
        if (names[0].isThis()) {
           print(className + ".");
        }
        boolean first = true;
        for (int i = 0; i < names.length; i++) {
            if (!first && !names[i].isSpecArrayRefExpr())
                print(".");
            names[i].accept(this);
            first = false;
        }
    }

    /** Prints a JML store reference keywords such as
     * <code>\everything</code>, <code>\nothing</code>, and
     * <code>\not_specified</code>.
     */
    public void visitJmlStoreRefKeyword( JmlStoreRefKeyword self ) 
    {
/*
	if (self.isEverything())
	    print("\\everything");
	else if (self.isNothing())
	    print("\\nothing");
	else
	    print("\\not_specified");
*/
    }

    /** Prints a JML <code>\type</code> expression. */
    public void visitJmlTypeExpression( JmlTypeExpression self ) 
    {
/*
	print("\\type(" + toString(self.type()) + ")");
*/
        if (result != null) {
          print("  ");
          printCS();
          println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
            ", " + self.type().toString() + ".class));");
           result = null;
        }
    }

    /** Prints a JML <code>\typeof</code> expression. */
    public void visitJmlTypeOfExpression( JmlTypeOfExpression self ) 
    {
       if (result != null) {
          String oldRes = result;
          result = varGen.freshPreVar();
          String newRes = result;
          declareVar(result, "java.lang.Object");
	  self.specExpression().accept(this);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "DELAYConstraint("
              + "new Object [] {" + newRes + "},");
          println("  new " + RPATH + "GenMethod() {");
          print("    public void run(" + EPATH + "ConstraintSystem ");
          printCS();
          println(") {");

          String var = varGen.freshPreVar();
          getValue(newRes, var, self.specExpression().getType());

          print("  ");
          printCS();
          println(".addGoalConstraint(new " + EPATH + "Equality(" + oldRes +
                  ", " + var + ".getClass()));");
          println("}}));");
        }
        result = null;
    }

    /** Prints a JML unreachable statement. */
    public void visitJmlUnreachableStatement( JmlUnreachableStatement self ) 
    {
/*
	mayStartAnnotation();
	print("unreachable;");
	mayEndAnnotation();

	// print RAC code if exists
	if (self.hasAssertionCode()) {
	    newLine();
	    self.assertionCode().accept(this);
	}
*/
    }

    /** Prints a JML variant function. */
    public void visitJmlVariantFunction( JmlVariantFunction self ) 
    {
/*
	mayStartAnnotation();
	print(self.isRedundantly() ?
	      "decreasing_redundantly " : "decreasing ");
	self.specExpression().accept(this);
	print(";");
	mayEndAnnotation();
*/
    }
    
    /** Prints a JML when clause. */
    public void visitJmlWhenClause( JmlWhenClause self ) 
    {
	Debug.msg("In JmlWhenClause");
	visitSpecBodyClause(self, "when");
    }

    // ----------------------------------------------------------------------
    // OVERRIDDEN METHODS
    // ----------------------------------------------------------------------

    /**
     * Prints a cast expression.
     */
    public void visitCastExpression( JCastExpression self ) {
	JExpression expr = self.expr();
	CType dest = self.getType();
        CType source = expr.getType();
        if (dest.isNumeric() && source.isNumeric()) {
          String oldRes = result;
          result = varGen.freshPreVar();
          declareVar(result, source);
          String exprRes = result;
          expr.accept(this);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "CASTConstraint(");
          print("\"" + source.toString() + "\", \"" + dest.toString() + "\", ");
          println(exprRes + ", " + oldRes + "));");
          result = null;
        } else {
          expr.accept(this);
        }
    }

    /**
     * prints a cast expression
     */
    public void visitUnaryPromoteExpression(/*@non_null@*/ JUnaryPromote self ) {
	JExpression expr = self.expr();
        CType dest = self.getType();
        CType source = expr.getType();
        if (dest.isNumeric() && source.isNumeric()) {
          String oldRes = result;
          result = varGen.freshPreVar();
          declareVar(result, source);
          String exprRes = result;
          expr.accept(this);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "CASTConstraint(");
          print("\"" + source.toString() + "\", \"" + dest.toString() + "\", ");
          println(exprRes + ", " + oldRes + "));");
          result = null;
        } else {
          expr.accept(this);
        }
    }

    private static String getSignature(CType type) {
      if (type.isArrayType()) {
        CType indType = ((CArrayType) type).getIndexedType();
        if (indType.isNumeric()) {
          return "[" + ((CNumericType) indType).getSignature();
        } else if (indType.isBoolean()) {
          return "[" + ((CBooleanType) indType).getSignature();
        } else if (indType.isArrayType()) {
          return "[" + getSignature(indType);
        } else {
          return "[L" + indType + ";";
        }
      } else {
        return type.getSignature();
      }
    }

    // convert a jmle value to a java value
    private void convertValue(String val, String res, CType type) {
        if (type.isClassType()) {
          println("  if (" + val + " == null || " + val + " == " + RPATH +
                  "Null.NullObj) {");
          println("    " + res + " = null;");
          println("  } else {");
          if (type.isArrayType()) {
            println("      try {");
            print("      ((" + RPATH + "MyArray) " + val + 
                    ").convertBack(Class.forName(\"" + getSignature(type) +
                    "\"), ");
            printCS();
            println(");");
            println("      } catch (ClassNotFoundException " + 
                    varGen.freshPreVar() + ") {");
            print("        throw new " + RPATH + "JmleException(\"Error:");
            println(" invalid type " + type + "\");");
            println("      }");
            println("      " + res + " = (" + type + ") ((" + RPATH + 
                    "MyArray) " + val + ").oldArray();");
          } else {
            print("  " + res + " = (");
            printType(type);
            println(") " + val + ";");
          }
          println("  }");
        } else {
          print("  " + res + " = ((");
          printType(type);
          println(") " + val + ")." + type + "Value();");
        }
    }

    private void getValue(String val, String res, CType type) {
        String nm1 = varGen.freshPreVar();
        print("  java.lang.Object " + nm1 + " = ");
        printCS();
        println(".getVarValue(" + val + ");");
        print("  final ");
        if (type instanceof org.jmlspecs.checker.CTypeType) {
          print("java.lang.Class");
        } else if (type == null) {
          fail("Error: no type!");
        } else {
          print(type);
        }
        println(" " + res + ";");
        convertValue(nm1, res, type);
    }


    /**
     * Prints an object allocator expression.
       requires result != null
     */
    public void visitNewObjectExpression( JNewObjectExpression self ) {
	CType type = self.getType();
	JExpression[] params = self.params();
        String oldres = result;
        java.util.List prevArgs = actualArgs;
	visitArgs(params);

        JmlTypeDeclaration cls = 
           JmlTypeLoader.getJmlSingleton().typeDeclarationOf(type.getCClass());
        if (type == null) {
          type = cls.getCClass().getType();
        }
        if (type == null) {
          fail("type is not set");
        }
        if (cls != null && !inInitializer && isJMLE(type)) {
          String freshObj = varGen.freshPreVar();
          print("  ");
          printType(type);
          print(" " + freshObj + " = new ");
          printType(type);
          println("(" + RPATH + "InternalConstructor.disambig, " +
            PREF + "env);");
          String cases = varGen.freshPreVar();
          println("  java.util.ArrayList " + cases + 
                  " = new java.util.ArrayList();");
          println("  try {");
          print("  " + freshObj + ".");
          print(PREF);
          printMangled(cls.ident(), params, null);
          print("(");
          printList(actualArgs, false);
          if (params.length > 0) print(", ");
          println(PREF + "env, " + (inOld || inPre) + " || " + PREF + 
                  "inOld, " + cases + ");");
          printCS();
          println(".addGoalConstraint(new " + CPATH + 
                  "ALLCASESConstraint(" + cases + "));");
          catchException(result, type);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "EQUALEQUALConstraint(" + result + 
                ", " + freshObj + "));");
        } else {
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH +
              "DELAYConstraint(new Object [] {");
          printList(actualArgs, false);
          println("}, new " + RPATH + "GenMethod() {");
          print("      public void run(" + EPATH + "ConstraintSystem ");
          printCS();
          println(") {");
          ArrayList newParams = new ArrayList();
          int count = 0;
          Iterator it = actualArgs.iterator();
          while (it.hasNext()) {
            String val = (String) it.next();
            String res = varGen.freshPreVar();
            getValue(val, res, params[count].getType());
            newParams.add(res);
            count++;
          }
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
                ", ");
          print("new " + toString(type) + "(");
          printList(newParams, false);
          println(")));");
          println("  }}));");
        }
        actualArgs = prevArgs;
        result = null;
    }

    private void printArrayDims(java.util.List dims) {
      Iterator it = dims.iterator();
      while (it.hasNext()) {
        String dim = (String) it.next();
        print("[" + dim + "]");
      }
    }

    public void visitArrayInitializer(JArrayInitializer self) {
      JExpression [] elems = self.elems();
      String oldRes = result;
      if (elems != null) {
        print("  ");
        printCS();
        println(".addGoalConstraint((new " + CPATH + 
                "ARRAYLENGTHConstraint(" + oldRes + ", \"" + 
                elems[0].getType() + 
                "\", new " + RPATH  + "MyInteger(" + elems.length + "))).setInAssert(" + inAssert + ", false));");
        for (int i = 0; i < elems.length; i++) {
          result = varGen.freshPreVar();
          String elemRes = result;
          declareVar(result, elems[i].getType());
          String objRes = varGen.freshPreVar();
          declareVar(objRes, elems[i].getType());
          elems[i].accept(this);
          printCS();
          println(".addGoalConstraint((new " + CPATH + "ARRAYINDEXConstraint(" 
                + oldRes + ", new " + RPATH + "MyInteger(" + i + "), " 
                + objRes + ")).setInAssert(" + inAssert + ", false));");
          printCS();
          print(".addGoalConstraint(new " + CPATH + 
                "EQUALNEWORNULLConstraint(" + objRes + ", " + elemRes);
          println(", new java.lang.Boolean(" + PREF + "inOld || " + 
                  (inOld || inPre) + ")));");
        }
      }
      result = null;
    }

    /**
     * Prints an array allocator expression.
     */
    public void visitNewArrayExpression( JNewArrayExpression self ) {
	CType type = self.getType();
	JArrayDimsAndInits dims = self.dims();
        JExpression [] arrDims = dims.dims();
        java.util.List prevArgs = actualArgs;
        if (arrDims != null) {
          JArrayInitializer init = dims.init();
          if (init != null) {
            init.accept(this);
          } else {
	  visitArgs(arrDims);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH +
                "DELAYConstraint(new Object [] {");
          printList(actualArgs, false);
          println("}, new " + RPATH + "GenMethod() {");
          print("      public void run(" + EPATH + "ConstraintSystem ");
          printCS();
          println(") {");
          ArrayList newParams = new ArrayList();
          int count = 0;
          Iterator it = actualArgs.iterator();
          while (it.hasNext()) {
            String val = (String) it.next();
            if (arrDims[count] != null) {
              String res = varGen.freshPreVar();
              getValue(val, res, arrDims[count].getType());
              newParams.add(res);
            } else {
              newParams.add("");
            }
            count++;
          }
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + EPATH + "Equality(" + result + ", ");
          print("new " + RPATH + "MyArray(new ");
	  if (type instanceof CArrayType) {
	      print(toString(((CArrayType)type).getBaseType()));
	  } else {
	      print(toString(type));
	  }
          printArrayDims(newParams);
          print(", ");
          printCS();
          println(")));");
          println("  }}));");
        }
        }

        result = null;
    }

    // ----------------------------------------------------------------------
    // HELPER METHODS
    // ----------------------------------------------------------------------

    private boolean hasRequires(JmlRequiresClause [] specHeader) {
       if (specHeader == null) {
          return false;
       } else if (specHeader.length > 1) {
          return true;
       } else {
         JExpression pre = 
           specHeader[0].predOrNot().specExpression().expression();
         if (pre instanceof JBooleanLiteral) {
            return !((JBooleanLiteral) pre).booleanValue();
         } else {
           return true;
         }
       }
    }

    public boolean isNormal(JmlSpecBody body) {
      if (!(body instanceof JmlGenericSpecBody)) {
         fail("Error: should be JmlGenericSpecBody after desugaring");
      }
      JmlGenericSpecBody bod = (JmlGenericSpecBody) body;
      JmlSpecBodyClause [] clauses = bod.simpleSpecBodies();
      JmlGenericSpecCase [] cases = bod.genericSpecCases();
      if (cases != null) {
         fail("Error: cases should be removed by desugaring");
      }
      if (clauses != null) {
         return cantThrowException(clauses);
      }
      return false;
    }

    public boolean isExceptional(JmlSpecBody body) {
      if (!(body instanceof JmlGenericSpecBody)) {
         fail("Error: should be JmlGenericSpecBody after desugaring");
      }
      JmlGenericSpecBody bod = (JmlGenericSpecBody) body;
      JmlSpecBodyClause [] clauses = bod.simpleSpecBodies();
      JmlGenericSpecCase [] cases = bod.genericSpecCases();
      if (cases != null) {
         fail("Error: cases should be removed by desugaring");
      }
      if (clauses != null) {
        return hasFalseEnsures(clauses) && hasSignalsClause(clauses);
      }
      return false;
    }

    /** Prints the given general spec case. */
    public void visitGeneralSpecCase( JmlGeneralSpecCase self ) 
    {
	Debug.msg("In helper visitSpecCase");

	JmlSpecVarDecl[] specVarDecls = self.specVarDecls();
	JmlRequiresClause[] specHeader = self.specHeader();
	JmlSpecBody specBody = self.specBody();
        boolean isExceptional = false;
        boolean isNormal = false;
        if (specBody != null) {
          isNormal = isNormal(specBody);
          isExceptional =  isExceptional(specBody);
        }
        inPre = true;
        inAssert = true;
        boolean hasPre = hasRequires(specHeader);
        println("  {");
	if (specVarDecls != null) {
	    for (int i = 0; i < specVarDecls.length; i++) {
		specVarDecls[i].accept(this);
	    }
	}

        result = varGen.freshPreVar();
        declareVar(result, RPATH + "MyBoolean");
        String preExpr = result;

        String res1, res2;
	if (specHeader != null) {
	    for (int i = 0; i < specHeader.length; i++) {
                if (i == specHeader.length - 1) {
	       	  specHeader[i].accept(this);
                } else {
                  res1 = varGen.freshPreVar();
                  declareVar(res1, RPATH + "MyBoolean");
                  String oldRes = result;
                  result = res1;
	       	  specHeader[i].accept(this);
                  res2 = varGen.freshPreVar();
                  declareVar(res2, RPATH + "MyBoolean");
                  print("  ");
                  printCS();
                  print(".addGoalConstraint(new " + CPATH + "CANDConstraint(");
                  println(res1 + ", " + res2 + ", " + oldRes + "));");
                  result = res2;
                }
	    }
	}
        newLine();

        print("  " + PREF + "allCases.add(new ");
        println(RPATH + "SpecCase(");
        println(preExpr + ",");
        println("  new " + RPATH + "GenMethod() {");
        print("     public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        print(") ");
        if (exceptions != null) {
          printThrows(exceptions);
        }
        println(" {");
        inAssert = false;
        result = null;
        visitAssignableClauses(specBody);
        println("}}, ");
        println("new " + RPATH + "GenMethod() {");
        print("      public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        print(") ");
        if (exceptions != null) {
          printThrows(exceptions);
        }
        println(" {");
        inPre = false;
        if (specBody != null) specBody.accept(this);
        println("    }}, " + isNormal + ", " + isExceptional + "));");
        result = null;

        if (!hasSignals && exceptions != null && exceptions.length != 0) {
           warn(self.getTokenReference(), ExecMessages.NO_SIGNALS);
        }

        println("  }");
    }

    /** Prints spec cases. */
    protected void visitSpecCases( JmlSpecCase[] specCases ) 
    {
	for (int i = 0; i < specCases.length; i++) {
	    specCases[i].accept(this);
	}
    }

    /** Prints example specifications. */
/*
    protected void visitExample( JmlExample self )
    {
	Debug.msg("In JmlExample");

	JmlSpecVarDecl[] specVarDecls = self.specVarDecls();
	JmlRequiresClause[] specHeader = self.specHeader();
	JmlSpecBodyClause[] specBody = self.specBody();
	
	if (specVarDecls != null) {
	    for (int i = 0; i < specVarDecls.length; i++) {
		specVarDecls[i].accept(this);
	    }
	}
	if (specHeader != null) {
	    for (int i = 0; i < specHeader.length; i++) {
		specHeader[i].accept(this);
	    }
	}
	if (specBody != null) {
	    for (int i = 0; i < specBody.length; i++) {
		specBody[i].accept(this);
	    }
	}
    }
*/

    private boolean hasFalseEnsures(JmlSpecBodyClause[] cases) {
      if (cases != null) {
        for (int i = 0; i < cases.length; i++) {
          if (cases[i] instanceof JmlEnsuresClause) {
            return isFalse(((JmlEnsuresClause) cases[i]).predOrNot());
          }
        }
      }
      return false;
    }

    private boolean isException(CType type) {
      if (type instanceof CClassFQNameType) {
        return ((CClassFQNameType) type).qualifiedName().equals(
             "java/lang/Exception");
      }
      return false;
    }

    private boolean cantThrowException(JmlSpecBodyClause[] cases) {
      if (cases != null) {
        for (int i = 0; i < cases.length; i++) {
          if (cases[i] instanceof JmlSignalsClause) {
             JmlSignalsClause signals = (JmlSignalsClause) cases[i];
             if (signals.hasPredicate() && isFalse(signals.predOrNot()) &&
                 isException(signals.type())) {
                 return true;
            }
          }
        }
      }
      return false;
    }

    private void visitAssignableClauses(JmlSpecBody self) {
      if (self != null) {
	JmlSpecBodyClause[] specClauses = self.specClauses();
        boolean hasAssignable = false;
        if (specClauses != null) {
          for (int i = 0; i < specClauses.length; i++) {
             if (specClauses[i] instanceof JmlAssignableClause) {
                specClauses[i].accept(this);
                hasAssignable = true;
             }
          }
        }
        if (!hasAssignable && curMeth instanceof JmlConstructorDeclaration &&
            !isExceptional(self)) {
          makeAllAssignable(getFields(curClass.getCClass()), false);
        }
      }
    }

    /** Prints specification body. */
    protected void visitSpecBody( JmlSpecBody self )
    {
	Debug.msg("In helper visitSpecBody");

	JmlSpecBodyClause[] specClauses = self.specClauses();
	JmlGeneralSpecCase[] specCases = self.specCases();
        boolean hasAssignable = isPure;
        boolean inExceptional = hasFalseEnsures(specClauses);
        ArrayList signals = new ArrayList();
        ArrayList post = new ArrayList();
	if (specClauses != null) {
	    for (int i = 0; i < specClauses.length; i++) {
                JmlSpecBodyClause clause = specClauses[i];
                if (clause instanceof JmlAssignableClause) {
                   hasAssignable = true;
                } else if (clause instanceof JmlSignalsOnlyClause) {
                   JmlSignalsOnlyClause so = (JmlSignalsOnlyClause) clause;
                   if (inExceptional && so.exceptions() != null && 
                       so.exceptions().length == 1 && 
                       hasDefaultConstructor(so.exceptions()[0].getCClass()) &&
                       !so.isRedundantly()) {
                     signals.add(clause);
                   } else {
                      warn(so.getTokenReference(), ExecMessages.SIGNALS_ONLY);
                   }
                } else if (clause instanceof JmlEnsuresClause) {
                   if (!inExceptional) {
                     post.add(clause);
                   }
                } else if (clause instanceof JmlSignalsClause) {
                   JmlSignalsClause sc = (JmlSignalsClause) clause;
                   if (sc.isRedundantly()) {
                      warn(sc.getTokenReference(), 
                           ExecMessages.REDUNDANT_NOT_EXECUTED,
                           "signals clauses");
                   } else if (sc.isNotSpecified()) {
                     throw new 
                    org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                    "Error: signals clause with unspecified predicate is" +
                       " not executable, " + formatToken(sc) + ".");
                   } else if (isGeneratedSignals(sc)) {
                      warn(sc.getTokenReference(), 
                           ExecMessages.GENERATED_NOT_EXECUTED,
                           "signals clauses");
                   } else if (!isFalse(sc.predOrNot())) {
                     signals.add(sc);
                   }
                } else {
                    warn(self.getTokenReference(), 
                         ExecMessages.GENERATED_NOT_EXECUTED,
                         self.getClass());
                }
	    }
	}
        if (!hasAssignable && !(curMeth instanceof JmlConstructorDeclaration)) {
          throw new 
            org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
               "Error: for executablility, a method specification must have " +
               "an explicit assignable clause (and not use \\everything), " +
                formatToken(self) + ".");
        }
        if (signals.isEmpty()) {
          if (trace) {
            println("  System.out.println(\"executing postcondition for " + 
                    curMeth.ident() + " of " + curClass.ident() + "\");");
          }
          acceptAll(post);
        } else if (signals.size() == 1 && post.isEmpty()) {
          acceptAll(signals);
        } else {
          int closeBracket = 0;
          print("  ");
          printCS();
          print(".addGoalConstraint(");
          println("  new " + CPATH + "ORConstraint(");
          int size = signals.size() + 1;
          for (int i = 0; i < signals.size(); i++) {
            println("  // Signals clause " + i);
            println("  new " + RPATH + "GenMethod() {");
            print("      public void run(" + EPATH + "ConstraintSystem ");
            printCS();
            println(") throws Exception {");
            ((JmlSpecBodyClause) signals.get(i)).accept(this);
            print("    }}");
            if (i < size - 2 ) {
              closeBracket++;
              println(",");
              println(" new " + RPATH + "GenMethod() {");
              print("      public void run(" + EPATH + "ConstraintSystem ");
              printCS();
              println(") {");
              print("  ");
              printCS();
              println(".addGoalConstraint(new " + CPATH + "ORConstraint(");
            } else if (i < size - 1) {
              println(",");
            }
          }
          if (!post.isEmpty()) {
            println("  // normal post");
            println("  new " + RPATH + "GenMethod() {");
            print("      public void run(" + EPATH + "ConstraintSystem ");
            printCS();
            println(") {");
            if (trace) {
              println("  System.out.println(\"executing postcondition for " + 
                      curMeth.ident() + " of " + curClass.ident() + "\");");
            }
            acceptAll(post);
            println("    }}");
          } else {
            println("  new " + RPATH + "GenMethod() {");
            print("      public void run(" + EPATH + "ConstraintSystem ");
            printCS();
            println(") {");
            if (trace) {
              println("  System.out.println(\"executing generated " +
                      "postcondition for " + 
                      curMeth.ident() + " of " + curClass.ident() + "\");");
            }
            if (inExceptional) {
              println("  // false (exceptional postcondition)");
              print("  ");
              printCS();
              println(".addGoalConstraint(new " + EPATH + "Bool(\"false\"));");
            } else {
              println("  // true");
            }
            println("    }}");
          }
          for (int i = 0; i < closeBracket; i++) {
            //print(")");
            print("));}}");
          }
          print("))");
          println(";");
        }

// START - may need to do something with these when more specification
//         syntax is supported
	if (specCases != null) {
	  //  visitSpecCases(specCases);
          fail("desugaring should remove specCases");
        }
    }

    /** Prints specfication body clause. */
    public void visitSpecBodyClause(JmlPredicateClause self, 
				    String keyword) 
    {
	Debug.msg("In helpe visitSpecBodyClause");
	
/*
	if (self.isRedundantly())
	    keyword = keyword + "_redundantly";
	    
	//newLineOffset();
	if (self.isNotSpecified()) {
	    print(keyword + " \\not_specified;");
	} else {
	    print(keyword + " ");
	    self.predOrNot().accept(this);
	    print(";");
	}
*/
        newLine();
	if (!self.isNotSpecified()) {
	    self.predOrNot().accept(this);
	}
      
    }

/*
    protected void print(String s) {
        super.print(s);
    }

    protected void println(String code) { 
      print(code + "\n");
    }
*/

    /**
     * prints a relational expression
     */
    public void visitRelationalExpression(/*@non_null@*/ JRelationalExpression self ) {
        int oper = self.oper();
        JExpression left = self.left();
        JExpression right = self.right();
        int lb = -1, rb = -1;
  
        if (inBoundaryPred && ((lb = boundQuantVar(left)) != -1 
                           | (rb = boundQuantVar(right)) != -1)) {
          if (secondTime) {
            equalTrue();
          } else {
            boolean lower = false, offByOne, addOne = false, leftVar = false;
            leftVar = lb > rb;  // true when bound variable is on the left
/*
            BoundQuantVariables.boundVar bv = quantVars.get((lb > rb) ? 
                                                lb : rb);
*/
            BoundQuantVariables.boundVar bv = quantVars.get(leftVar ? 
                                                lb : rb);
            if (!bv.type.isOrdinal()) {
               throw new 
                 org.jmlspecs.jmlexec.runtime.InsufficientInformationException(
                "Error: non-ordinal quantified variable, " +
                 formatToken(self) + ".");
            }
            offByOne = (oper == OPE_LT || oper == OPE_GT);
            addOne = (leftVar && oper == OPE_GT) || 
                     (!leftVar && oper == OPE_LT);
            lower = (leftVar && (oper == OPE_GT || oper == OPE_GE))
                 || (!leftVar && (oper == OPE_LT || oper == OPE_LE));

            CType type;
            String oldRes = result;
            result = varGen.freshPreVar();
            String inRes1 = result;
            if (leftVar) {
              type = right.getType();
            } else {
              type = left.getType();
            }
            declareVar(result, type);
            if (leftVar) {
              right.accept(this);
            } else {
              left.accept(this);
            }
            String inRes2;
            if (offByOne) {
              inRes2 = varGen.freshPreVar();
              declareVar(inRes2, type);
              print("  ");
              printCS();
              print(".addGoalConstraint(new " + CPATH);
              print("ADDNUMS");
              print("Constraint(" + inRes1 + ", ");
              if (addOne) {
                printWrapped("1", type);
              } else {
                printWrapped("-1", type);
              }
              println(", " + inRes2 + "));");
            } else {
              inRes2 = inRes1;
            }
            String inRes3;
            if (lower && !type.equals(bv.type)) {
              inRes3 = varGen.freshPreVar();
              declareVar(inRes3, bv.type);
              print("  ");
              printCS();
              print(".addGoalConstraint(new " + CPATH + "CASTConstraint(");
              print("\"" + type.toString() + "\", \"" + bv.type.toString() 
                    + "\", ");
              println(inRes2 + ", " + inRes3 + "));");
            } else {
              inRes3 = inRes2; 
            }
            if (lower) bv.lb = inRes3;
            else bv.ub = inRes3;
            result = oldRes;
          }
        } else if (!inBoundaryPred || secondTime) {
          boolean flip = false;
          String opString = "";
          switch (oper) {
          case OPE_LT:
              opString = "LT";
              break;
          case OPE_LE:
              opString = "LE";
              break;
          case OPE_GT:
              opString = "LT";
              flip = true;
              break;
          case OPE_GE:
              opString = "LE";
              flip = true;
              break;
          default:
              fail();
          }
          if (flip) {
            visitBinaryBooleanExpression(opString, right, left);
          } else {
            visitBinaryBooleanExpression(opString, left, right);
          }
       } else {
          equalTrue();
       }
    }

    /** Prints a JML relational expression. */
    public void visitJmlRelationalExpression( JmlRelationalExpression self ) 
    {
	int oper = self.oper();
        if (oper == OPE_LT || oper == OPE_LE || oper == OPE_GT || 
            oper == OPE_GE) {
            visitRelationalExpression(self);
        } else {
          JExpression left = self.left();
          JExpression right = self.right();
          String opString;
          switch (oper) {
            case OPE_EQUIV:
                andOnly = false;
                opString = "EQUALEQUAL";
                visitBinaryBooleanExpression(opString, left, right);
                break;
            case OPE_NOT_EQUIV:
                andOnly = false;
                opString = "INEQUALITY";
                visitBinaryBooleanExpression(opString, left, right);
                break;
            case OPE_BACKWARD_IMPLIES:
                JExpression tmp = left;
                left = right;
                right = tmp;
                // break intentionally omitted
            case OPE_IMPLIES:
                andOnly = false;
                if (result == null) {
                   result = varGen.freshPreVar();
                   String leftRes = result;
                   declareVar(result, left.getType());
                   boolean backAssert = inAssert;
                   inAssert = true;
                   left.accept(this);
                   inAssert = backAssert;
                   print("  ");
                   printCS();
                   println(".addGoalConstraint(new " + CPATH + "IMPLYConstraint(");
                   println(leftRes + ",");
                   println("  new " + RPATH + "GenMethod() {");
                   print("     public void run(" + EPATH + "ConstraintSystem ");
                   printCS();
                   println(") {");
                   right.accept(this);
                   println("    }}));");
                 } else {
                   boolean backAssert = inAssert;
                   inAssert = true;
                   visitBinaryExpression("IMPLYR", left, right);
                   inAssert = backAssert;
                 }
                break;
            case OPE_SUBTYPE:
                String oldRes = result;
                result = varGen.freshPreVar();
                declareVar(result, "java.lang.Class");
                String leftRes = result;
                left.accept(this);
                result = varGen.freshPreVar();
                declareVar(result, "java.lang.Class");
                String rightRes = result;
                right.accept(this);
                print("  ");
                printCS();
                print(".addGoalConstraint(new ");
                print(CPATH + "DELAYConstraint(new Object [] {");
                println(leftRes + ", " + rightRes + "}, ");
                println("  new " + RPATH + "GenMethod() {");
                print("    public void run(" + EPATH + "ConstraintSystem ");
                printCS();
                println(") {");

                String lVar = varGen.freshPreVar();
                String rVar = varGen.freshPreVar();
                getValue(leftRes, lVar, left.getType());
                getValue(rightRes, rVar, right.getType());

                print("  ");
                printCS();
                print(".addGoalConstraint(new " + EPATH + "Equality(");
                print("new " + RPATH + "MyBoolean(" + rVar + 
                      ".isAssignableFrom(" + lVar + ")), ");
                if (oldRes != null) {
                   println(oldRes + "));");
                } else {
                   println(RPATH + "MyBoolean.TRUE()));");
                }
                println("}}));");
                result = null;
                break;
            default:
                opString = "";
                throw new 
                  org.jmlspecs.jmlexec.runtime.UnsupportedOperationException(
                 "Error: specified operation is not executable, " +
                  formatToken(self) + ".");
          }
        }
    }



    /** 
    * print appropriate constraint system variable (pre or post-state)
    */

    private void printCS() {
      print(PREF + "csPost");
    }

    private void printEnv() {
      print(PREF + "env");
    }

    /** 
    * print a type, wrapping primitives and null
    */
    private void printType(CType typ) {
      print(typeString(typ));
    }

    private String typeString(CType typ) {
      if (typ.isArrayType()) {
        return RPATH + "MyArray";
      } else if (typ == CStdType.Null) {
        return "java.lang.Object";
      } else if (typ == CStdType.Boolean) {
        return RPATH + "MyBoolean";
      } else if (typ == CStdType.Void) {
        return "java.lang.Void";
      } else if (typ.isNumeric()) {
        if (typ == CStdType.Byte) {
          return RPATH + "MyByte"; 
        } else if (typ == CStdType.Char) {
          return RPATH + "MyChar"; 
        } else if (typ == CStdType.Short) {
          return RPATH + "MyShort"; 
        } else if (typ == CStdType.Integer) {
          return RPATH + "MyInteger"; 
        } else if (typ == CStdType.Long) {
          return RPATH + "MyLong"; 
        } else if (typ == CStdType.Float) {
          return RPATH + "MyFloat"; 
        } else if (typ == CStdType.Double) {
          return RPATH + "MyDouble"; 
        } else {
          fail();
          return null;
        }
      } else if (typ instanceof CTypeType) {
          return "java.lang.Class"; 
      } else if (typ.isClassType()) {
        CClass cls = typ.getCClass();
        if (cls.isNestedType()) {
          return typeString(cls.owner()) + "$" + cls.ident();
        } else {
          return toString(typ);
        }
      } else {
        return toString(typ);
      }
    }

    private void printType(CClass cls) {
      print(typeString(cls));
    }

    private String typeString(CClass cls) {
      if (cls.isNestedType()) {
        return typeString(cls.owner()) + "$" + cls.ident();
      } else {
        return cls.toString().replace('/', '.');
      }
    }

    /**
    *  print a primitive value as a wrapped value
    */

    private void printWrapped(String val, CType typ) {
      print("new ");
      printType(typ);
      println("(" + val + ")");
    }
    

    /** 
    * declare constraint variable of specified type
    */
    private void declareVar(String var, CType typ, boolean fin) {
      print("  ");
      if (fin) print("final ");
      println(RPATH + "VarObject " + var + 
                " = new " + RPATH + "VarObject();");
      print("  ");
      printCS();
      print(".addVariable(" + var + ", \"");
      if (typ == null || typ.isArrayType()) {
        // hack because Class.forName does not recognize array types
        print("java.lang.Object");
      } else {
        printType(typ);
      }
      println("\", \"" + var + "\");");
    }

    private void declareVar(String var, CType typ) {
      declareVar(var, typ, true);
    }

    /** 
    * declare constraint variable of specified type
    */
    private void declareVar(String var, String typ, boolean fin) {
      print("  ");
      if (fin) print("final ");
      println("java.lang.Object " + var + 
                " = new " + RPATH + "VarObject();");
      print("  ");
      printCS();
      print(".addVariable(" + var + ", \"");
      print(typ);
      println("\", \"" + var + "\");");
    }

    private void declareVar(String var, String typ) {
      declareVar(var, typ, true);
    }

    private void visitAndExpression(JExpression left, JExpression right) {
        if (result == null) {
          left.accept(this);
          print("\n");
          right.accept(this);
        } else {
          // evaluate as assertion
          boolean backAssert = inAssert;
          inAssert = true;
          visitBinaryExpression("CAND", left, right);
          inAssert = backAssert;
        }
     }

     /**
     * visits a boolean AND expression
     */
    public void visitConditionalAndExpression(/*@non_null@*/ 
       JConditionalAndExpression self )
    {
        visitAndExpression(self.left(), self.right());
    }

    private boolean equalExprs(JExpression left, JExpression right) {
      if (left == null) {
        return right == null;
      } else if (left instanceof JUnaryPromote) {
        if (right instanceof JUnaryPromote) {
          JUnaryPromote lft = (JUnaryPromote) left;
          JUnaryPromote rht = (JUnaryPromote) right;
          return lft.getType().equals(rht.getType()) &&
                 equalExprs(lft.expr(), rht.expr());
        }
      }
      return left.equals(right);
    }

    private JExpression conditionalSetOfValues(JExpression self, 
                                               JExpression compare) {
      if (self instanceof JConditionalOrExpression) {
        JConditionalOrExpression orExpr = (JConditionalOrExpression) self;
        JExpression rht = orExpr.right();
        if (rht instanceof JParenthesedExpression) {
          rht = rht.unParenthesize();
        }
        if (rht instanceof JmlEqualityExpression) {
           JmlEqualityExpression right = (JmlEqualityExpression) rht;
           if (right.left().getType().isArrayType() ||
               right.right().getType().isArrayType()) {
             return null;
           }
           if (compare != null) {
             if (equalExprs(compare, right.left()) || 
                 equalExprs(compare, right.right())) {
               return conditionalSetOfValues(orExpr.left(), compare);
             } else {
               return null;
             }
           } else {
             JExpression res = conditionalSetOfValues(orExpr.left(), 
                                                      right.left());
             if (res != null) {
               return res;
             } else {
               return conditionalSetOfValues(orExpr.left(), right.right());
             }
           }
        }
      } else if (compare != null) {
        if (self instanceof JParenthesedExpression) {
          self = self.unParenthesize();
        }
        if (self instanceof JmlEqualityExpression) {
          JmlEqualityExpression expr = (JmlEqualityExpression) self;
           if (expr.left().getType().isArrayType() ||
               expr.right().getType().isArrayType()) {
             return null;
           }
          if (equalExprs(expr.left(), compare) || 
              equalExprs(expr.right(), compare)) {
            return compare;
          }
        }
      }
      return null;
    }

    public void visitConditionalSetOfValues(JExpression self, 
                                            JExpression compare,
                                            ArrayList elems) {
      JExpression visitMe = null;
      if (self instanceof JConditionalOrExpression) {
        JConditionalOrExpression orExpr = (JConditionalOrExpression) self;
        visitMe = orExpr.right();
        visitConditionalSetOfValues(orExpr.left(), compare, elems);
      }  else if (self instanceof JParenthesedExpression) {
          visitMe = self.unParenthesize();
      } else {
          visitMe = self;
      }
      if (visitMe == null || !(visitMe instanceof JmlEqualityExpression)) {
        fail();
      } else {
        JmlEqualityExpression expr = (JmlEqualityExpression) visitMe;
        JExpression visit = equalExprs(expr.left(), compare) ? expr.right() : 
                                                               expr.left();
        result = varGen.freshPreVar();
        elems.add(0, result);
        declareVar(result, visit.getType());
        visit.accept(this);
        result = null; 
      }
    }

     /**
     * visits a boolean OR expression
     */
    public void visitConditionalOrExpression(/*@non_null@*/ 
       JConditionalOrExpression self )
    {
        JExpression left = self.left();
        JExpression right = self.right();
        JExpression compare = null;
        if (result == null) {
          compare = conditionalSetOfValues(self, null);
        }
        andOnly = false;
        if (compare != null) {
          ArrayList elems = new ArrayList();
          visitConditionalSetOfValues(self, compare, elems);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + 
                "DELAYConstraint(new Object [] {");
          printList(elems, false);
          println("}, new " + RPATH + "GenMethod() {");
          print("      public void run(" + EPATH + "ConstraintSystem ");
          printCS();
          println(") {");

          String set = varGen.freshPreVar();
          print("  " + MPATH + "JMLObjectSet " + set + " = " + MPATH +
                  "JMLObjectSet.convertFrom(new Object [] {");
          boolean leadcomma = false;
          Iterator it = elems.iterator();
          while (it.hasNext()) {
//          for (String var : elems) {
            String var = (String) it.next();
            if (leadcomma) {
              print(", ");
            } else {
              leadcomma = true;
            }
            printCS();
            print(".getVarValue(" + var + ")");
          }
          println("});");
          result = varGen.freshPreVar();
          declareVar(result, compare.getType());
          String res = result;
          compare.accept(this);
          print("  ");
          printCS();
          println(".addGoalConstraint(new " + CPATH + "HASConstraint(" + 
                  set + ", " + res + ", \"JMLObjectSet\"));");
          println("  }}));");
          result = null;
        } else if (result == null) {
          print("  ");
          printCS();
          println(".addGoalConstraint(new " + CPATH + "ORConstraint(");
          println("  new " + RPATH + "GenMethod() {");
          print("      public void run(" + EPATH + "ConstraintSystem ");
          printCS();
          println(") {");
          left.accept(this);
          println("    }},");
          println("  new " + RPATH + "GenMethod() {");
          print("      public void run(" + EPATH + "ConstraintSystem ");
          printCS();
          println(") {");
          right.accept(this);
          println("    }}));");
        } else {
          boolean backAssert = inAssert;
          inAssert = true;
          visitBinaryExpression("COR", left, right);
          inAssert = backAssert;
        }
     }

    /**
     * visits a conditional expression: e?v1:v2
     */
    public void visitConditionalExpression(JConditionalExpression self) {
        JExpression cond = self.cond();
        JExpression left = self.left();
        JExpression right = self.right();
        boolean backAssert = inAssert;
        inAssert = true;
        String oldRes = result;
        result = varGen.freshPreVar();
        declareVar(result, RPATH + "MyBoolean");
        String condRes = result;
        cond.accept(this);
        inAssert = backAssert;
        print("  ");
        printCS();
        print(".addGoalConstraint(new " + CPATH + "CONDEXPRConstraint(");
        println(condRes + ", ");
        println("  new " + RPATH + "GenMethod() {");
        print("      public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        println(") {");
        String leftRes = null;
        if (oldRes != null) {
          result = varGen.freshPreVar();
          declareVar(result, left.getType());
          leftRes = result;
        } else {
          result = null;
        }
        left.accept(this);
        if (oldRes != null) {
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "EQUALEQUALConstraint(");
          print(leftRes + ", " + oldRes + "));");
        }
        println("    }},");
        println("  new " + RPATH + "GenMethod() {");
        print("      public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        println(") {");
        String rightRes = null;
        if (oldRes != null) {
          result = varGen.freshPreVar();
          declareVar(result, right.getType());
          rightRes = result;
        } else {
          result = null;
        }
        right.accept(this);
        if (oldRes != null) {
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "EQUALEQUALConstraint(");
          print(rightRes + ", " + oldRes + "));");
        }
        println("    }}));");
        result = null;
    }

    /**
     * prints a parenthesed expression
     */
    public void visitParenthesedExpression(/*@non_null@*/ JParenthesedExpression self ) {
        JExpression expr = self.expr();
        expr.accept(this);
    }

      /**
     * prints an instanceof expression
     */
    public void visitInstanceofExpression(/*@non_null@*/ JInstanceofExpression self ) {
        JExpression expr = self.expr();
        CType dest = self.dest();
        String oldRes = result;
        result = varGen.freshPreVar();
        CType exprType = expr.getType();
        declareVar(result, exprType);
        String leftRes = result;
        expr.accept(this);
        print("  ");
        printCS();
        print(".addGoalConstraint(new ");
        print(CPATH + "DELAYConstraint(new Object [] {");
        println(leftRes + "}, ");
        println("  new " + RPATH + "GenMethod() {");
        print("    public void run(" + EPATH + "ConstraintSystem ");
        printCS();
        println(") {");

        String lVar = varGen.freshPreVar();
/*
        print("  " + exprType + " " + lVar + " = (" + exprType + ") ");
        printCS();
        println(".getVarValue(" + leftRes + ");");
*/
        getValue(leftRes, lVar, exprType);

        print("  ");
        printCS();
        print(".addGoalConstraint(new " + EPATH + "Equality(");
        print("new " + RPATH + "MyBoolean(" + lVar + 
              " instanceof " + dest + "), ");
        if (oldRes != null) {
           println(oldRes + "));");
        } else {
           println(RPATH + "MyBoolean.TRUE()));");
        }
        println("}}));");
    }
 
    /**
     * prints an equality (or inequality) expression
     */

    public void visitEqualityExpression(/*@non_null@*/ 
       JEqualityExpression self) {
        if (!inBoundaryPred || secondTime) {
          int oper = self.oper();
          JExpression left = self.left();
          JExpression right = self.right();
          String opString;
          if (oper == OPE_EQ) {
             opString = "EQUALEQUAL";
          } else { // inequality
             opString = "INEQUALITY";
          }
          visitBinaryBooleanExpression(opString, left, right);
        } else {
          equalTrue();
        }
    }
 
    /**
     * prints a relational expression
     */

    public void visitBinaryBooleanExpression(String constraintType,
        JExpression left, JExpression right) {
        String oldRes = result;
        boolean backAssert = false;
        if (oldRes != null) {
          backAssert = inAssert;
          inAssert = true;
          if (constraintType.equals("EQUALEQUAL")) {
            constraintType = "EQUALEQUALR";
          }
        }
        result = varGen.freshPreVar();
        String leftRes = result;
        declareVar(result, left.getType());
        String rightRes = varGen.freshPreVar();
        declareVar(rightRes, right.getType());
        left.accept(this);
        result = rightRes;
        right.accept(this);
        print("  ");
        printCS();
        print(".addGoalConstraint((new " + CPATH + constraintType);
        print("Constraint(");
        print(leftRes);
        print(", " + rightRes);
        if (oldRes != null) {
          print(", " + oldRes);
        } else if (!constraintType.equals("EQUALEQUAL")) {
          print(", org.jmlspecs.jmlexec.runtime.MyBoolean.TRUE()");
        }
        if (constraintType.equals("EQUALEQUAL")) {
          println(")));");
        } else {
          println(")).setInAssert(" + inAssert + ", false));");
        }
        inAssert = backAssert;
        result = null;
    } 

    /**
     * visits a binary  expression
     */
    public void visitBinaryExpression(String constraintType,
        /*@non_null@*/ JExpression left, /*@non_null@*/ JExpression right) {

      String oldRes = result;
      if (constraintType.equals("ADDNUMS") || constraintType.equals("SUB")
       || constraintType.equals("MULT") || constraintType.equals("DIVD")) {
        printFD(oldRes, left.getType(), "null");
      }
      result = varGen.freshPreVar();
      String leftRes = result;
      declareVar(result, left.getType());
      String rightRes = varGen.freshPreVar();
      declareVar(rightRes, right.getType());
      left.accept(this);
      result = rightRes;
      right.accept(this);
      print("  ");
      printCS();
      print(".addGoalConstraint((new " + CPATH);
      if (constraintType.equals("SUB") && !inAssert) {
        print("ADDNUMSConstraint(" + oldRes + ", " + rightRes + ", " +
              leftRes + "))");
        println(".setInAssert(" + inAssert  + ", false));");
      } else {
        print(constraintType + "Constraint(" 
               + leftRes + ", ");
        print(rightRes + ", " + oldRes);
        print(")).setInAssert(");
        if (constraintType.equals("CAND") || constraintType.equals("COR")
         || constraintType.equals("IMPLYR")) {
          print("false");
        } else {
          print(inAssert);
        }
        println(", false));");
      }
      result = null;
    }

    public void visitBinaryExpression(/*@non_null@*/ JBinaryExpression self,
      String constraintType) {

      JExpression left = self.left();
      JExpression right = self.right();
      visitBinaryExpression(constraintType, left, right);
    }

    /**
     * visits an add expression
     */
    public void visitAddExpression(/*@non_null@*/ JAddExpression self ) {
      String cons;
      if (self.left().getType().toString().equals("java.lang.String") ||
          self.right().getType().toString().equals("java.lang.String")) {
        cons = "STRCAT";
      } else {
        cons = "ADDNUMS";
      }
      visitBinaryExpression(self, cons);
    }

    /**
     * visits a divide expression
     */
    public void visitDivideExpression(/*@non_null@*/ JDivideExpression self ) {
        visitBinaryExpression( self, "DIVD");
    }
                                                                                
    /**
     * visits a minus expression
     */
    public void visitMinusExpression(/*@non_null@*/ JMinusExpression self ) {
        visitBinaryExpression( self, "SUB");
    }

    /**
     * visits a modulo division expression
     */
    public void visitModuloExpression(/*@non_null@*/ JModuloExpression self ) {
        visitBinaryExpression( self, "MOD");
    }
                                                                                
    /**
     * visits a multiplication expression
     */
    public void visitMultExpression(/*@non_null@*/ JMultExpression self ) {
        visitBinaryExpression( self, "MULT");
    }

    /**
     * prints a shift expression
     */
    public void visitShiftExpression(/*@non_null@*/ JShiftExpression self ) {
        int oper = self.oper();
        if (oper == OPE_SL) {
            visitBinaryExpression( self, "SHIFTLEFT");
        } else if (oper == OPE_SR) {
            visitBinaryExpression( self, "SHIFTRIGHT");
        } else {
            visitBinaryExpression( self, "SHIFTRIGHTPAD");
        }
    }

    /**
     * prints a compound assignment expression
     */
    public void visitBitwiseExpression(/*@non_null@*/ JBitwiseExpression self ) {
        int oper = self.oper();
        switch (oper) {
        case OPE_BAND:
            if (self.left().getType().equals(CStdType.Boolean)) {
              visitAndExpression(self.left(), self.right());
            } else {
              visitBinaryExpression( self, "BITAND");
            }
            break;
        case OPE_BOR:
            if (self.left().getType().equals(CStdType.Boolean)) {
              visitConditionalOrExpression(new JConditionalOrExpression(
                self.getTokenReference(), self.left(), self.right()));
            } else {
              visitBinaryExpression( self, "BITOR");
            }
            break;
        case OPE_BXOR:
            visitBinaryExpression( self, "BITXOR");
            break;
        default:
            fail();
        }
    }

    /** 
     * prints a field expression 
    */
   public void visitFieldExpression(/*@non_null@*/ JClassFieldExpression self) {
       JExpression prefix = self.prefix();
       boolean boolField = inAssert && result == null;
        if (prefix instanceof org.multijava.mjc.JSuperExpression &&
            inInvariantFor) {
           fail("Error: use of super in an \\invariant_for expression " +
                "is not executable.");
        } else 
        if (prefix instanceof org.multijava.mjc.JTypeNameExpression ||
            prefix instanceof org.multijava.mjc.JSuperExpression ||
            (!inInvariantFor && (prefix == null || 
            prefix instanceof org.multijava.mjc.JThisExpression))) {
          String field = "";
          if (inInvariantFor && (prefix == null || 
                prefix instanceof org.multijava.mjc.JThisExpression)) {
            field = PREF + "this.";
          } else if (prefix != null) {
             field = prefix.toString().replace('/', '.') + ".";
          }
          if (prefix instanceof org.multijava.mjc.JTypeNameExpression &&
              self.getField().owner().isInterface()) {
              field += PREF + self.getField().owner().ident() + ".";
          }
          field = field + SPECFLD + self.ident();
          String fldVar = varGen.freshPreVar();
          print("  Object " + fldVar + " = " + RPATH + "ObjUtil.getVar(");
          print(field + ", new java.lang.Boolean(");
          print(PREF + "inOld || " + (inOld || inPre) + "), ");
          printCS();
          println(");");
          print("  ");
          printCS();
          print(".addGoalConstraint(new ");
          if (boolField) {
            print(CPATH + "EQUALEQUALRConstraint(");
          } else {
            print(EPATH + "Equality(");
          }
          if (result != null) {
            print(result);
          } else {
            print(RPATH + "MyBoolean.TRUE()");
          }
          print(", " + fldVar);

          if (boolField) {
            print(", " + RPATH + "MyBoolean.TRUE()");
          }
          println("));");
          result = null;
        } else {
          String obj = PREF + "this";
          if (prefix != null) {
            String oldres = result;
            result = varGen.freshPreVar();
            obj = result;
            declareVar(result, prefix.getType());
            prefix.accept(this);
            result = oldres;
          }
          String fld = varGen.freshPreVar();
          declareVar(fld, "java.lang.Object");
          print("  ");
          printCS();
          println(".addGoalConstraint(new " + CPATH + 
             "FIELDConstraint(" + obj
             + ", \"" + SPECFLD + self.ident() + "\", " + fld + "));");
          if (boolField) {
            result = varGen.freshPreVar();
            declareVar(result, RPATH + "MyBoolean");
          }
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + 
                "EQUALNEWORNULLConstraint(");
          print(fld + ", ");
          if (result == null) {
            print(RPATH + "MyBoolean.TRUE()");
          } else {
            print( result);
          }
          print(", new java.lang.Boolean(" + (inOld || inPre));
          print(" || " + PREF + "inOld");
          println(")));");
          if (boolField) {
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + 
                  "EQUALEQUALRConstraint(" + result + ", ");
            print(RPATH + "MyBoolean.TRUE(), ");
            print(RPATH + "MyBoolean.TRUE()));");
          }
  
          result = null;
        }
   }

    /**
     * prints an array length expression
     */
    public void visitArrayLengthExpression(/*@non_null@*/ JArrayLengthExpression self ) {
        JExpression prefix = self.prefix();
        String oldRes = result;
        result = varGen.freshPreVar();
        String arVar = result;
        declareVar(result, prefix.getType());
        prefix.accept(this);
        CType elemType = ((CArrayType) prefix.getType()).getIndexedType();
        if (elemType.isArrayType()) {
          elemType = CStdType.Object;
        }
        print("  ");
        printCS();
        println(".addGoalConstraint((new " + CPATH + 
                "ARRAYLENGTHConstraint(" + arVar + ", \"" + elemType + 
                "\", " + oldRes + ")).setInAssert(" + inAssert + ", false));");
        result = null;
    }

    /**
     * prints an array access (index) expression
     */
    public void visitArrayAccessExpression(/*@non_null@*/ JArrayAccessExpression self ) {
        JExpression prefix = self.prefix();
        JExpression accessor = self.accessor();
        String oldRes = result;
        if (oldRes == null) {
          oldRes = RPATH + "MyBoolean.TRUE()";
        }
        result = varGen.freshPreVar();
        String arVar = result;
        declareVar(result, prefix.getType());
        prefix.accept(this);
        result = varGen.freshPreVar();
        String indVar = result;
        declareVar(result, accessor.getType());
        accessor.accept(this);
        String castInd;
        if (accessor.getType().equals(CStdType.Integer)) {
          castInd = indVar;
        } else {
          castInd = varGen.freshPreVar();
          declareVar(castInd, CStdType.Integer);
          print("  ");
          printCS();
          print(".addGoalConstraint(new " + CPATH + "CASTConstraint(");
          println("\"" + accessor.getType() + "\", \"" + CStdType.Integer + 
                  "\", " + indVar + ", " + castInd + "));");
        }
        print("  ");
        result = varGen.freshPreVar();
        String elemRes = result;
        declareVar(elemRes, ((CArrayType) prefix.getType()).getIndexedType());
        printCS();
        println(".addGoalConstraint((new " + CPATH + "ARRAYINDEXConstraint(" + 
                arVar + ", " + castInd + ", " + elemRes +
                ")).setInAssert(" + inAssert + ", false));");
        printCS();
        print(".addGoalConstraint(new " + CPATH + 
              "EQUALNEWORNULLConstraint(" + elemRes + ", " + oldRes);
        print(", new java.lang.Boolean(" + (inOld || inPre));
        print(" || " + PREF + "inOld");
        println(")));");
        result = null;
    }



    /**
     * prints a this expression
     * requires result != null
     */
    public void visitThisExpression(/*@non_null@*/  JThisExpression self ) {
        print("  ");
        printCS();
        print(".addGoalConstraint(new " + EPATH + "Equality(");
        if (!inInvariantFor) {
          print(className + ".this, ");
        } else {
          print(PREF + "this, ");
        }
        println(result + "));");
        result = null;
    }

    /**
     * prints a super expression
     * requires result != null && the caller has accounted for the superclass
     */
    public void visitSuperExpression(/*@non_null@*/  JSuperExpression self ) {
       if (inInvariantFor) {
         fail("Error: use of super in an \\invariant_for expression " +
              "is not executable.");
       } else {
         print("  ");
         printCS();
         print(".addGoalConstraint(new " + EPATH + "Equality(this, ");
         println(result + "));");
         result = null;
       }
    }


    /**
     * prints an unary expression
     */
    public void visitUnaryExpression(/*@non_null@*/  JUnaryExpression self ) {
        int oper = self.oper();
        JExpression expr = self.expr();
        String oldres;
        String exprRes;

        switch (oper) {
        case OPE_PLUS:
            oldres = result;
            result = varGen.freshPreVar();
            exprRes = result;
            declareVar(result, self.getType());
            expr.accept(this);
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + "CASTConstraint(");
            println("\"" + expr.getType() + "\", \"" + self.getType() + 
                    "\", " + exprRes + ", " + oldres + "));");
            result = null;
            break;
        case OPE_MINUS:
            oldres = result;
            printFD(oldres, self.getType(), "null");
            result = varGen.freshPreVar();
            exprRes = result;
            declareVar(result, self.getType());
            expr.accept(this);
            String castRes = varGen.freshPreVar();
            declareVar(castRes, self.getType());
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + "CASTConstraint(");
            println("\"int\", \"" + self.getType() + "\", new " + RPATH +
                    "MyInteger(-1), " + castRes + "));");
            String castRes2 = varGen.freshPreVar();
            declareVar(castRes2, self.getType());
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + "CASTConstraint(");
            println("\"" + expr.getType() + "\", \"" + self.getType() + 
                    "\", " + exprRes + ", " + castRes2 + "));");
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + "MULTConstraint(");
            println(castRes + ", " + castRes2 + ", " + oldres + "));");
            result = null;
            break;
        case OPE_BNOT:
            oldres = result;
            result = varGen.freshPreVar();
            exprRes = result;
            declareVar(result, self.getType());
            expr.accept(this);
            result = oldres;
            print("  ");
            printCS();
            print(".addGoalConstraint(new " + CPATH + "BITNOTConstraint(");
            println(exprRes + ", " + result + "));");
            result = null;
            break;
                                                                                
        case OPE_LNOT:
            andOnly = false;
            boolean backAssert = inAssert;
            if (result == null) {
              result = RPATH + "MyBoolean.FALSE()";
              inAssert = true;
              expr.accept(this);
              result = null;
            } else {
              String oldRes = result;
              result = varGen.freshPreVar();
              inAssert = true;
              declareVar(result, expr.getType());
              String expRes = result;
              expr.accept(this);
              print("  ");
              printCS();
              println(".addGoalConstraint(new " + CPATH + "NOTConstraint(" +
                      expRes + ", " + oldRes + "));");
            }
            inAssert = backAssert;
            break;
                                                                                
        default:
            fail();
        }
    }


    /**
     * prints a boolean literal
     */
    public void visitBooleanLiteral(/*@non_null@*/ JBooleanLiteral self ) {
        if (result != null) {
          print("  ");
          printCS();
          println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
            ", " + RPATH + "MyBoolean.getBoolean(" +
            self.booleanValue() + ")));");
          result = null;
        } else if (self.booleanValue()) {
          // true is easy to satisfy
        } else {
          // ensure failure
          print("  ");
          printCS();
          println(".addGoalConstraint(new " + EPATH + "Bool(\"false\"));");
        }
    }

    /**
     * prints a byte literal
     */
    protected void visitByteLiteral( byte value ) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", new " + RPATH + "MyByte((byte)" + value + ")));");
        result = null;
    }

    /**
     * prints an int literal
     */
    protected void visitIntLiteral(int value) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", new " + RPATH + "MyInteger(" + value + ")));");
        result = null;
    }

    /**
     * prints a long literal
     */
    protected void visitLongLiteral(long value) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", new " + RPATH + "MyLong(" + value + "L)));");
        result = null;
    }
                                                                                
    /**
     * prints a short literal
     */
    protected void visitShortLiteral(short value) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", new " + RPATH + "MyShort((short)" + value + ")));");
        result = null;
    }
                                                                                
    /**
     * prints a char literal
     */
    protected void visitCharLiteral(char value) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", new " + RPATH + "MyChar(" + value + ")));");
        result = null;
    }

    /**
     * Prints a double literal. This method is overridden here to take
     * care of special values such as INFINITY, NaN, MIN_VALUE, and
     * MAX_VALUE.
     */

    protected void visitDoubleLiteral(double value) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", new " + RPATH + "MyDouble(" + TransUtils.toString(value) + ")));");
        result = null;
    }
                                                                                
    /**
     * Prints a float literal. This method is overridden here to take
     * care of special values such as Float.POSTIVE_INFINITY,
     * Float.NaN, etc.
     */
    protected void visitFloatLiteral(float value) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", new " + RPATH + "MyFloat(" + TransUtils.toString(value) + ")));");
        result = null;
    }

    /**
     * prints a string literal
     */
    public void visitStringLiteral(/*@non_null@*/ JStringLiteral self) {
        String value = self.stringValue();
        StringBuffer s = new StringBuffer();
        for (int i = 0; i < value.length(); i++) {
            char c = value.charAt(i);
            switch (c) {
            case '\n' : s.append("\\n"); break;
            case '\r' : s.append("\\r"); break;
            case '\t' : s.append("\\t"); break;
            case '\b' : s.append("\\b"); break;
            case '\f' : s.append("\\f"); break;
            case '\"' : s.append("\\\""); break;
            case '\'' : s.append("\\\'"); break;
            case '\\' : s.append("\\\\"); break;
            default:
                s.append(c);
            }
        }
        value = s.toString();
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
          ", \"" + value + "\"));");
        result = null;
    }


    /**
     * prints a null literal
     */

    public void visitNullLiteral(/*@non_null@*/ JNullLiteral self ) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        println(".addGoalConstraint(new " + EPATH + "Equality(" + result + 
           ", " + RPATH + "Null.NullObj));");
        result = null;
    }

    /**
     * prints a local variable expression
     */
    public void visitLocalVariableExpression(/*@non_null@*/ JLocalVariableExpression self ) {
        // pre: result contains variable to hold expression result
        print("  ");
        printCS();
        String ident = self.ident();
        CType type = self.getType();
        print(".addGoalConstraint(new ");
        print(EPATH + "Equality(");

        if (type.isArrayType()) {
          print("(" + PREF + "env.arrays().get(" + ident + ")) == null ? " +
                ident + " : " + PREF + "env.arrays().get(" + ident + ")");
        } else {
          print(ident);
        }

        println(", " + result + "));");
        result = null;
    }

    public void visitNameExpression(/*@non_null@*/ JNameExpression self ) {
        print("  ");
        printCS();
        print(".addGoalConstraint(new " + EPATH + "Equality(");
        String instVar = "";
        JExpression prefix = self.sourcePrefix();
        if (prefix != null) {
            instVar = prefix.toString() + ".";
        }
        instVar += self.sourceIdent();
        if (self.getType() != null && self.getType().isVoid()) {
          printWrapped(instVar, self.getType());
        } else {
          print(instVar);
        }
        print(", " + result);
        println("));");
        result = null;
    }

    // ----------------------------------------------------------------------
    // PRIVATE DATA MEMBERS
    // ----------------------------------------------------------------------

    /** the space from '@' to the current print position for specification. */
    private int offset;

    /** indicates if currently printing inside a JML annotation marker. 
     * <pre><jml>
     * private invariant inAnnotation <==> annotationDepth > 0;
     * </jml></pre>
     */
    /*@ spec_protected @*/ private boolean inAnnotation;

    /** the depth of annotations.
     * 
     * <pre><jml>
     * private invariant annotationDepth >= 0;
     * </jml></pre>
     */
    /*@ spec_protected @*/ private int annotationDepth = 0;

    /** the column number where annotation markers were written. */
    /*@ spec_protected @*/ private int atMarkerPos;

    /**
     * Modifier utility to manipulating modifiers, e.g., to get string 
     * representations of modifiers from bit mask encoding.
     * 
     * <pre><jml>
     * private invariant jmlModUtil != null && jmlModUtil == modUtil &&
     *           inAnnotation <==> jmlModUtil.withoutMarkers();
     * </jml></pre>
     * */


     private boolean inPre = true;   // visiting precondition
     private String result = null;  // variable to return result into
     private VarGenerator varGen = VarGen.varGen;
     private boolean inOld = false; // are we within an \old expr?
     private JFormalParameter[] formalParameters = null;
     private java.util.List actualArgs; // build list of variables for
                                                // values of actuals
     private boolean trace = Debug.isDebugOn(); // whether or not to print JCK trace
     // exceptions for method currently being compiled
     private CClassType [] exceptions;
     private boolean inAssert = false; // are we in an assertion context?
     // if so, then nonlocals should NOT be bound
     // quantified variables - not treated as Java primitives
     private BoundQuantVariables quantVars = new BoundQuantVariables();
     // holds name and boundary set for each quantified variable

     // visiting a bounding predicate for a quantified assertion?
     private boolean inBoundaryPred = false;
     private boolean andOnly = true; // only and (no other relational ops)
                                     // encounted while visiting a subtree
     private boolean secondTime = false; // visiting subtree for second time
     // paths where constraints and solvers can be found
     private static final String CPATH = "org.jmlspecs.jmlexec.constraints.";
     private static final String EPATH = "org.jmlspecs.jmlexec.jack.evaluator.";
     private static final String RPATH = "org.jmlspecs.jmlexec.runtime.";
     private static final String ENPATH = "org.jmlspecs.jmlexec.environment.";
     private static final String MPATH = "org.jmlspecs.models.";
     // length of CLASSPATH to models directory, i.e. "org.jmlspecs.models."
     private static final int MODPATHLEN = MPATH.length();
     // name of class currently being compiled
     private String className;
     // name of super class of class currently being compiled
     private String superClassName;
     // method currently being compiled
     private JmlMethodDeclaration curMeth;
     // # of constraints
     private int consNum = 0;
     // # of methods with signals clauses
     private static int exMethNum = 0;
     // is the current method static?
     private boolean isStatic = false;
     // executing static initializer?
     private boolean inInitializer = false;
     // does the current method have at least one signals clause?
     private boolean hasSignals = false;
     // current class or interface being compiled
     private JmlTypeDeclaration curClass;
     // collection of all fields (including those inherited from
     // interfaces)
     private ArrayList combinedFields = null;
     // generating invariant_for methods?
     private boolean inInvariantFor = false;
     // list of all constraints
     private ArrayList allConstraints;
     // list of all invariants
     private ArrayList allInvariants;
     // all implemented interfaces
     private ArrayList jmlInterfaces;
     private boolean hasSuperClass = false;
     // is the current method declared pure?
     private boolean isPure = false;
     // cache imported packages
     private ArrayList imports = new ArrayList();
     // cache field initializers
     private ArrayList fieldInits = new ArrayList();
     // prefix to distinguish variables added by jmle
     private static final String PREF = "spec$";
     // prefix to distinguish fields added by jmle
     private static final String SPECFLD = "specFld_";
     // which JCK handlers to use
     private static final String [] JMLHANDLERS = {
        CPATH + "JMLHandler",
        CPATH + "QUANTHandler"
     };
     // do not treat the following JmlCollection methods as constraints
     private static HashMap nonConstraints;
     static {
       nonConstraints = new HashMap();
       nonConstraints.put("singleton", "singleton");
       nonConstraints.put("convertFrom", "convertFrom");
       nonConstraints.put("clone", "clone");
       nonConstraints.put("hashCode", "hashCode");
       nonConstraints.put("toString", "toString");
       nonConstraints.put("toBag", "toBag");
       nonConstraints.put("toSet", "toSet");
       nonConstraints.put("toSequence", "toSequence");
       nonConstraints.put("toArray", "toArray");
       nonConstraints.put("compose", "compose");
       nonConstraints.put("range", "range");
       nonConstraints.put("domain", "domain");
       nonConstraints.put("image", "image");
       nonConstraints.put("imagePairSet", "imagePairSet");
       nonConstraints.put("removeFromDomain", "removeFromDomain");
       nonConstraints.put("inverseImage", "inverseImage");
       nonConstraints.put("inverseElementImage", "inverseElementImage");
       nonConstraints.put("extend", "extend");
       nonConstraints.put("extendUnion", "extendUnion");
       nonConstraints.put("clashReplaceUnion", "clashReplaceUnion");
       nonConstraints.put("disjointUnion", "disjointUnion");
       nonConstraints.put("removeAll", "removeAll");
       
     }
}
